# This file is generated by build_data.py.
LANGUAGES_WITH_NAME_DATA = {
    'ab',
    'af',
    'agq',
    'ak',
    'am',
    'an',
    'ang',
    'ann',
    'apc',
    'ar',
    'arn',
    'as',
    'asa',
    'ast',
    'az',
    'ba',
    'bal',
    'bas',
    'be',
    'bem',
    'bez',
    'bg',
    'bgc',
    'bgn',
    'bho',
    'blo',
    'blt',
    'bm',
    'bn',
    'bo',
    'br',
    'brx',
    'bs',
    'bss',
    'ca',
    'cch',
    'ccp',
    'ce',
    'ceb',
    'cgg',
    'cho',
    'chr',
    'cic',
    'ckb',
    'co',
    'cs',
    'csw',
    'cv',
    'cy',
    'da',
    'dav',
    'de',
    'dje',
    'doi',
    'dsb',
    'dua',
    'dyo',
    'dz',
    'ebu',
    'ee',
    'el',
    'en',
    'eo',
    'es',
    'et',
    'eu',
    'ewo',
    'fa',
    'ff',
    'fi',
    'fil',
    'fo',
    'fr',
    'frr',
    'fur',
    'fy',
    'ga',
    'gaa',
    'gd',
    'gl',
    'gn',
    'gsw',
    'gu',
    'guz',
    'gv',
    'ha',
    'haw',
    'he',
    'hi',
    'hil',
    'hnj',
    'hr',
    'hsb',
    'ht',
    'hu',
    'hy',
    'ia',
    'id',
    'ie',
    'ig',
    'ii',
    'ilo',
    'io',
    'is',
    'it',
    'ja',
    'jbo',
    'jgo',
    'jmc',
    'jv',
    'ka',
    'kaa',
    'kab',
    'kaj',
    'kam',
    'kcg',
    'kcm',
    'kde',
    'kea',
    'ken',
    'kgp',
    'khq',
    'ki',
    'kk',
    'kkj',
    'kl',
    'kln',
    'km',
    'kn',
    'ko',
    'kok',
    'kpe',
    'ks',
    'ksb',
    'ksf',
    'ksh',
    'ku',
    'kw',
    'kxv',
    'ky',
    'la',
    'lag',
    'lah',
    'lb',
    'lg',
    'lij',
    'lkt',
    'lld',
    'lmo',
    'ln',
    'lo',
    'lrc',
    'lt',
    'lu',
    'luo',
    'luy',
    'lv',
    'mai',
    'mas',
    'mdf',
    'mer',
    'mfe',
    'mg',
    'mgh',
    'mgo',
    'mi',
    'mic',
    'mk',
    'ml',
    'mn',
    'mni',
    'moh',
    'mr',
    'ms',
    'mt',
    'mua',
    'mus',
    'my',
    'myv',
    'mzn',
    'nah',
    'naq',
    'nb',
    'nd',
    'nds',
    'ne',
    'nl',
    'nn',
    'nnh',
    'no',
    'nqo',
    'nso',
    'nus',
    'nv',
    'ny',
    'nyn',
    'oc',
    'om',
    'or',
    'os',
    'osa',
    'pa',
    'pap',
    'pcm',
    'pis',
    'pl',
    'prg',
    'ps',
    'pt',
    'qu',
    'quc',
    'raj',
    'rhg',
    'rif',
    'rm',
    'rn',
    'ro',
    'rof',
    'ru',
    'rup',
    'rw',
    'rwk',
    'sa',
    'sah',
    'saq',
    'sat',
    'sbp',
    'sc',
    'scn',
    'sd',
    'sdh',
    'se',
    'seh',
    'ses',
    'sg',
    'sh',
    'shi',
    'shn',
    'si',
    'sk',
    'skr',
    'sl',
    'sma',
    'smj',
    'smn',
    'sms',
    'sn',
    'so',
    'sq',
    'sr',
    'ss',
    'ssy',
    'st',
    'su',
    'sv',
    'sw',
    'syr',
    'szl',
    'ta',
    'te',
    'teo',
    'tg',
    'th',
    'ti',
    'tk',
    'tn',
    'to',
    'tok',
    'tpi',
    'tr',
    'trv',
    'trw',
    'tt',
    'twq',
    'tzm',
    'ug',
    'uk',
    'ur',
    'uz',
    'vai',
    've',
    'vec',
    'vi',
    'vmw',
    'vo',
    'vun',
    'wa',
    'wae',
    'wbp',
    'wo',
    'xh',
    'xnr',
    'xog',
    'yav',
    'yi',
    'yo',
    'yrl',
    'yue',
    'za',
    'zgh',
    'zh',
    'zu',
}

CODE_TO_NAMES = {
    '001': {'af': 'wêreld', 'ak': 'wiase', 'am': 'ዓለም', 'ar': 'العالم', 'as': 'বিশ্ব', 'ast': 'Mundu', 'az': 'Dünya', 'az-Arab': 'دۆنیا', 'az-Cyrl': 'Дүнја', 'az-Latn': 'Dünya', 'bal-Latn': 'Donyá', 'be': 'Свет', 'be-tarask': 'Свет', 'bg': 'свят', 'bgn': 'دونیا/جهان', 'blo': 'nɖulinya', 'bn': 'পৃথিবী', 'bo': 'འཛམ་གླིང་།', 'br': 'Bed', 'brx': 'बुहुम', 'bs': 'Svijet', 'bs-Cyrl': 'Свијет', 'bs-Latn': 'Svijet', 'ca': 'Món', 'ca-ES-valencia': 'Món', 'ccp': '𑄛𑄨𑄖𑄴𑄗𑄨𑄟𑄨', 'ce': 'Дерригдуьненан', 'ceb': 'kalibutan', 'chr': 'ᎡᎶᎯ', 'ckb': 'جیهان', 'cs': 'svět', 'csw': 'ᐊᐢᑭᕀ', 'cv': 'тӗнче', 'cy': 'Y Byd', 'da': 'Verden', 'de': 'Welt', 'dsb': 'swět', 'dz': 'འཛམ་གླིང༌', 'ee': 'xexeme', 'el': 'Κόσμος', 'el-polyton': 'Κόσμος', 'en': 'world', 'en-Dsrt': '𐐎𐐲𐑉𐑊𐐼', 'eo': 'mondo', 'es': 'Mundo', 'et': 'maailm', 'eu': 'Mundua', 'fa': 'جهان', 'ff-Adlm': '𞤀𞤣𞤵𞤲𞤢', 'fi': 'maailma', 'fil': 'Mundo', 'fo': 'heimur', 'fr': 'Monde', 'fur': 'Mont', 'fy': 'Wrâld', 'ga': 'an Domhan', 'gd': 'An Saoghal', 'gl': 'Mundo', 'gsw': 'Wält', 'gu': 'વિશ્વ', 'ha': 'Duniya', 'he': 'העולם', 'hi': 'विश्व', 'hi-Latn': 'world', 'hr': 'Svijet', 'hsb': 'swět', 'ht': 'Monde', 'hu': 'Világ', 'hy': 'Աշխարհ', 'ia': 'mundo', 'id': 'Dunia', 'ie': 'munde', 'ig': 'Uwa', 'ii': 'ꋧꃅ', 'is': 'Heimurinn', 'it': 'Mondo', 'ja': '世界', 'jgo': 'Mbí', 'jv': 'Donya', 'ka': 'მსოფლიო', 'kaa': 'дүнья', 'kaa-Cyrl': 'дүнья', 'kea': 'Mundu', 'kgp': 'Tã mĩ ke kar', 'kk': 'әлем', 'kk-Cyrl': 'әлем', 'km': 'ពិភពលោក', 'kn': 'ಪ್ರಪಂಚ', 'ko': '세계', 'kok': 'जग', 'kok-Deva': 'जग', 'ks': 'دُنیا', 'ks-Arab': 'دُنیا', 'ksh': 'de Ääd', 'ku': 'dinya', 'kxv': 'raji, purti', 'kxv-Deva': 'राजि, पृती', 'kxv-Latn': 'raji, purti', 'kxv-Telu': 'ప్రపంచం', 'ky': 'Дүйнө', 'lb': 'Welt', 'lkt': 'Makȟásitomni', 'lld': 'monn', 'lo': 'ໂລກ', 'lrc': 'دونیا', 'lt': 'pasaulis', 'lv': 'pasaule', 'mai': 'विश्व', 'mi': 'te ao', 'mk': 'Свет', 'ml': 'ലോകം', 'mn': 'Дэлхий', 'mni': 'মালেম', 'mni-Beng': 'মালেম', 'mr': 'जग', 'ms': 'Dunia', 'mt': 'Dinja', 'my': 'ကမ္ဘာ', 'mzn': 'جهون', 'nb': 'verden', 'ne': 'विश्व', 'nl': 'wereld', 'nn': 'verda', 'no': 'verden', 'nqo': 'ߞߌߢߍ߲߫', 'om': 'addunyaa', 'or': 'ବିଶ୍ୱ', 'os': 'Дуне', 'pa': 'ਸੰਸਾਰ', 'pa-Guru': 'ਸੰਸਾਰ', 'pcm': 'Wọld', 'pl': 'świat', 'prg': 'swītai', 'ps': 'نړۍ', 'pt': 'Mundo', 'qu': 'Pacha', 'rm': 'mund', 'ro': 'Lume', 'ru': 'весь мир', 'sah': 'Аан дойду', 'sc': 'Mundu', 'sd': 'دنيا', 'sd-Arab': 'دنيا', 'se': 'máilbmi', 'si': 'ලෝකය', 'sk': 'svet', 'sl': 'svet', 'so': 'dunida', 'sq': 'Bota', 'sr': 'свет', 'sr-Cyrl': 'свет', 'sr-Latn': 'svet', 'sv': 'världen', 'sw': 'Dunia', 'syr': 'ܬܐܒܝܠ', 'ta': 'உலகம்', 'te': 'ప్రపంచం', 'tg': 'ҷаҳонӣ', 'th': 'โลก', 'ti': 'ዓለም', 'tk': 'Dünýä', 'to': 'Māmani', 'tok': 'ma ale', 'tr': 'Dünya', 'tt': 'дөнья', 'ug': 'دۇنيا', 'uk': 'Світ', 'ur': 'دنیا', 'uz': 'Dunyo', 'uz-Cyrl': 'Дунё', 'uz-Latn': 'Dunyo', 'vec': 'mondo', 'vi': 'Thế giới', 'wae': 'Wäld', 'wo': 'àddina', 'xh': 'ihlabathi', 'xnr': 'दुनिया', 'yi': 'וועלט', 'yo': 'Agbáyé', 'yrl': 'Iwi', 'yue': '世界', 'yue-Hans': '世界', 'yue-Hant': '世界', 'zh': '世界', 'zh-Hans': '世界', 'zh-Hant': '世界', 'zu': 'umhlaba'},
    '002': {'af': 'Afrika', 'ak': 'Abibirem', 'am': 'አፍሪካ', 'ar': 'أفريقيا', 'as': 'আফ্ৰিকা', 'ast': 'África', 'az': 'Afrika', 'az-Cyrl': 'Африка', 'az-Latn': 'Afrika', 'bal-Latn': 'Afriká', 'be': 'Афрыка', 'be-tarask': 'Афрыка', 'bg': 'Африка', 'bgn': 'افریقا', 'blo': 'Garɩɖontǝna', 'bn': 'আফ্রিকা', 'br': 'Afrika', 'brx': 'आफ्रिका', 'bs': 'Afrika', 'bs-Cyrl': 'Африка', 'bs-Latn': 'Afrika', 'ca': 'Àfrica', 'ca-ES-valencia': 'Àfrica', 'ccp': '𑄃𑄜𑄳𑄢𑄨𑄇', 'ce': 'Африка', 'ceb': 'Africa', 'chr': 'ᎬᎿᎨᏍᏛ', 'ckb': 'ئەفریقا', 'cs': 'Afrika', 'cv': 'Африка', 'cy': 'Affrica', 'da': 'Afrika', 'de': 'Afrika', 'dsb': 'Afrika', 'dz': 'ཨཕ་རི་ཀ', 'ee': 'Afrika nutome', 'el': 'Αφρική', 'el-polyton': 'Αφρική', 'en': 'Africa', 'en-Dsrt': '𐐈𐑁𐑉𐐲𐐿𐐲', 'es': 'África', 'et': 'Aafrika', 'eu': 'Afrika', 'fa': 'افریقا', 'ff-Adlm': '𞤀𞤬𞤪𞤭𞤳𞤢𞥄', 'fi': 'Afrikka', 'fil': 'Africa', 'fo': 'Afrika', 'fr': 'Afrique', 'fur': 'Afriche', 'fy': 'Afrika', 'ga': 'an Afraic', 'gd': 'Afraga', 'gl': 'África', 'gsw': 'Afrika', 'gu': 'આફ્રિકા', 'ha': 'Afirka', 'he': 'אפריקה', 'hi': 'अफ़्रीका', 'hi-Latn': 'Africa', 'hr': 'Afrika', 'hsb': 'Afrika', 'ht': 'Afrique', 'hu': 'Afrika', 'hy': 'Աֆրիկա', 'ia': 'Africa', 'id': 'Afrika', 'ie': 'Africa', 'ig': 'Afrika', 'ii': 'ꃏꍏ', 'is': 'Afríka', 'it': 'Africa', 'ja': 'アフリカ', 'jgo': 'Afɛlîk', 'jv': 'Afrika', 'ka': 'აფრიკა', 'kaa': 'Африка', 'kaa-Cyrl': 'Африка', 'kea': 'Áfrika', 'kgp': 'Afrika', 'kk': 'Африка', 'kk-Cyrl': 'Африка', 'km': 'អាហ្វ្រិក', 'kn': 'ಆಫ್ರಿಕಾ', 'ko': '아프리카', 'kok': 'आफ्रिका', 'kok-Deva': 'आफ्रिका', 'ks': 'اَفریٖکا', 'ks-Arab': 'اَفریٖکا', 'ksh': 'Affrika', 'ku': 'Afrîka', 'ky': 'Африка', 'lb': 'Afrika', 'lkt': 'Hásapa Makȟáwita', 'lld': 'Africa', 'lo': 'ອາຟຣິກາ', 'lrc': 'ئفریقا', 'lt': 'Afrika', 'lv': 'Āfrika', 'mai': 'अफ्रीका', 'mi': 'Āwherika', 'mk': 'Африка', 'ml': 'ആഫ്രിക്ക', 'mn': 'Африк', 'mni': 'অফ্রিকা', 'mni-Beng': 'অফ্রিকা', 'mr': 'आफ्रिका', 'ms': 'Afrika', 'mt': 'Affrika', 'my': 'အာဖရိက', 'mzn': 'آفریقا', 'nb': 'Afrika', 'ne': 'अफ्रिका', 'nl': 'Afrika', 'nn': 'Afrika', 'no': 'Afrika', 'nqo': 'ߝߘߊ߬ߝߌ߲߬ߠߊ߫', 'om': 'Afrikaa', 'or': 'ଆଫ୍ରିକା', 'os': 'Африкӕ', 'pa': 'ਅਫ਼ਰੀਕਾ', 'pa-Guru': 'ਅਫ਼ਰੀਕਾ', 'pcm': 'Áfríka', 'pl': 'Afryka', 'ps': 'افريقا', 'pt': 'África', 'qu': 'Africa', 'rm': 'Africa', 'ro': 'Africa', 'ru': 'Африка', 'sah': 'Аапырыка', 'sc': 'Àfrica', 'sd': 'آفريڪا', 'sd-Arab': 'آفريڪا', 'se': 'Afrihkká', 'si': 'අප්\u200dරිකාව', 'sk': 'Afrika', 'sl': 'Afrika', 'so': 'Afrika', 'sq': 'Afrikë', 'sr': 'Африка', 'sr-Cyrl': 'Африка', 'sr-Latn': 'Afrika', 'sv': 'Afrika', 'sw': 'Afrika', 'syr': 'ܐܦܪܝܩܐ', 'ta': 'ஆப்பிரிக்கா', 'te': 'ఆఫ్రికా', 'tg': 'Африқо', 'th': 'แอฟริกา', 'ti': 'ኣፍሪቃ', 'tk': 'Afrika', 'to': 'ʻAfilika', 'tr': 'Afrika', 'tt': 'Африка', 'ug': 'ئافرىقا', 'uk': 'Африка', 'ur': 'افریقہ', 'uz': 'Afrika', 'uz-Cyrl': 'Африка', 'uz-Latn': 'Afrika', 'vec': 'Àfrega', 'vi': 'Châu Phi', 'wae': 'Afrika', 'wo': 'Africa', 'xh': 'IAfrika', 'xnr': 'अफ़्रीका', 'yi': 'אַפֿריקע', 'yo': 'Áfíríkà', 'yrl': 'Afirika', 'yue': '非洲', 'yue-Hans': '非洲', 'yue-Hant': '非洲', 'zh': '非洲', 'zh-Hans': '非洲', 'zh-Hant': '非洲', 'zu': 'i-Africa'},
    '003': {'af': 'Noord-Amerika', 'ak': 'Amɛrika Atifi', 'am': 'ሰሜን አሜሪካ', 'ar': 'أمريكا الشمالية', 'as': 'উত্তৰ আমেৰিকা', 'ast': 'Norteamérica', 'az': 'Şimali Amerika', 'az-Cyrl': 'Шимали Америка', 'az-Latn': 'Şimali Amerika', 'bal-Latn': 'Shemáli Amriká', 'be': 'Паўночная Амерыка', 'be-tarask': 'Паўночная Амерыка', 'bg': 'Северноамерикански континент', 'blo': 'Gamalɩkatǝna gʊnyɩpɛnɛlaŋ', 'bn': 'উত্তর আমেরিকা', 'br': 'Norzhamerika', 'brx': 'साहा आमेरिका', 'bs': 'Sjeverna Amerika', 'bs-Cyrl': 'Сјеверноамерички континент', 'bs-Latn': 'Sjeverna Amerika', 'ca': 'Amèrica del Nord', 'ca-ES-valencia': 'Amèrica del Nord', 'ccp': '𑄅𑄪𑄖𑄴𑄖𑄮𑄢𑄴 𑄃𑄟𑄬𑄢𑄨𑄇', 'ce': 'Къилбаседа Америка', 'ceb': 'North America', 'chr': 'ᏧᏴᏢ ᎠᎹᏰᏟ', 'ckb': 'ئەمەریکای باکوور', 'cs': 'Severní Amerika', 'csw': 'ᒥᐦᑭᓇᕽ ᒥᓂᐢᑎᐠ', 'cv': 'Ҫурҫӗр Америка', 'cy': 'Gogledd America', 'da': 'Nordamerika', 'de': 'Nordamerika', 'dsb': 'Pódpołnocna Amerika', 'dz': 'བྱང་ཨ་མི་རི་ཀ', 'ee': 'Dziehe Amerika nutome', 'el': 'Βόρεια Αμερική', 'el-polyton': 'Βόρεια Αμερική', 'en': 'North America', 'en-Dsrt': '𐐤𐐱𐑉𐑃 𐐊𐑋𐐯𐑉𐐲𐐿𐐲', 'es': 'América del Norte', 'et': 'Põhja-Ameerika', 'eu': 'Ipar Amerika', 'fa': 'امریکای شمالی', 'ff-Adlm': '𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫 𞤀𞤥𞤫𞤪𞤭𞤳𞤢𞥄', 'fi': 'Pohjois-Amerikka', 'fil': 'Hilagang Amerika', 'fo': 'Norðuramerika', 'fr': 'Amérique du Nord', 'fur': 'Americhe dal Nord', 'fy': 'Noard-Amerika', 'ga': 'Meiriceá Thuaidh', 'gd': 'Aimeireaga a Tuath', 'gl': 'América do Norte', 'gsw': 'Nordameerika', 'gu': 'ઉત્તર અમેરિકા', 'ha': 'Amurka ta Arewa', 'he': 'צפון אמריקה', 'hi': 'उत्तर अमेरिका', 'hi-Latn': 'North America', 'hr': 'Sjevernoamerički kontinent', 'hsb': 'Sewjerna Amerika', 'ht': 'Amérique du Nord', 'hu': 'Észak-Amerika', 'hy': 'Հյուսիսային Ամերիկա', 'ia': 'America del Nord', 'id': 'Amerika Utara', 'ig': 'Mpaghara Ugwu Amerịka', 'ii': 'ꀙꂰꍏ', 'is': 'Norður-Ameríka', 'it': 'Nord America', 'ja': '北アメリカ大陸', 'jv': 'Amérika Lor', 'ka': 'ჩრდილოეთ ამერიკა', 'kaa': 'Арқа Америка', 'kaa-Cyrl': 'Арқа Америка', 'kea': 'Merka di Norti', 'kgp': 'Nãrti-Amẽrika', 'kk': 'Солтүстік Америка', 'kk-Cyrl': 'Солтүстік Америка', 'km': 'អាមេរិក\u200bខាង\u200bជើង', 'kn': 'ಉತ್ತರ ಅಮೇರಿಕಾ', 'ko': '북아메리카', 'kok': 'उत्तर अमेरिका', 'kok-Deva': 'उत्तर अमेरिका', 'ks': 'شُمٲلی اَمریٖکا', 'ks-Arab': 'شُمٲلی اَمریٖکا', 'ksh': 'Nood-Amärrika', 'ku': 'Amerîkaya Bakur', 'ky': 'Түндүк Америка', 'lb': 'Nordamerika', 'lo': 'ອາເມລິກາເໜືອ', 'lrc': 'ئمریکا شومالی', 'lt': 'Šiaurės Amerika', 'lv': 'Ziemeļamerika', 'mai': 'उत्तरी अमेरिका', 'mi': 'Amerika ki te Raki', 'mk': 'Северна Америка', 'ml': 'വടക്കേ അമേരിക്ക', 'mn': 'Хойд Америк', 'mr': 'उत्तर अमेरिका', 'ms': 'Amerika Utara', 'my': 'မြောက် အမေရိကတိုက်', 'mzn': 'شمالی آمریکا', 'nb': 'Nord-Amerika', 'ne': 'उत्तर अमेरिका', 'nl': 'Noord-Amerika', 'nn': 'Nord-Amerika', 'no': 'Nord-Amerika', 'nqo': 'ߊߡߋߙߌߞߌ߬ ߞߐ߬ߘߎ߮', 'om': 'Ameerikaa Kaabaa', 'or': 'ଉତ୍ତର ଆମେରିକା', 'pa': 'ਉੱਤਰ ਅਮਰੀਕਾ', 'pa-Guru': 'ਉੱਤਰ ਅਮਰੀਕਾ', 'pcm': 'Nọ́t Amẹ́ríka', 'pl': 'Ameryka Północna', 'ps': 'شمالی امریکا', 'pt': 'América do Norte', 'qu': 'Norte America', 'rm': 'America dal Nord', 'ro': 'America de Nord', 'ru': 'Северная Америка', 'sah': 'Хотугу Эмиэрикэ', 'sc': 'Amèrica de su Nord', 'sd': 'اتر آمريڪا', 'sd-Arab': 'اتر آمريڪا', 'se': 'dávvi-Amerihkká ja gaska-Amerihkká', 'si': 'උතුරු ඇමෙරිකාව', 'sk': 'Severná Amerika', 'sl': 'Severna Amerika', 'so': 'Waqooyi Ameerika', 'sq': 'Amerika e Veriut', 'sr': 'Северноамерички континент', 'sr-Cyrl': 'Северноамерички континент', 'sr-Latn': 'Severnoamerički kontinent', 'sv': 'Nordamerika', 'sw': 'Amerika Kaskazini', 'syr': 'ܐܡܪܝܟܐ ܓܪܒܝܝܬܐ', 'ta': 'வட அமெரிக்கா', 'te': 'ఉత్తర అమెరికా', 'tg': 'Америкаи Шимолӣ', 'th': 'อเมริกาเหนือ', 'ti': 'ሰሜን ኣመሪካ', 'tk': 'Demirgazyk Amerika', 'to': 'ʻAmelika tokelau', 'tr': 'Kuzey Amerika', 'tt': 'Төньяк Америка', 'ug': 'شىمالىي ئامېرىكا', 'uk': 'Північна Америка', 'ur': 'شمالی امریکہ', 'uz': 'Shimoliy Amerika', 'uz-Cyrl': 'Шимолий Америка', 'uz-Latn': 'Shimoliy Amerika', 'vec': 'Mèrega de’l nord', 'vi': 'Bắc Mỹ', 'wae': 'Nordamerika', 'wo': 'North America', 'xh': 'UMntla Melika', 'xnr': 'उत्तर अमरिका', 'yi': 'צפון־אַמעריקע', 'yo': 'Àríwá Amẹ́ríkà', 'yrl': 'Amerika Nuti suí', 'yue': '北美洲', 'yue-Hans': '北美洲', 'yue-Hant': '北美洲', 'zh': '北美洲', 'zh-Hans': '北美洲', 'zh-Hant': '北美洲', 'zu': 'i-North America'},
    '005': {'af': 'Suid-Amerika', 'ak': 'Amɛrika Anaafoɔ', 'am': 'ደቡብ አሜሪካ', 'ar': 'أمريكا الجنوبية', 'as': 'দক্ষিণ আমেৰিকা', 'ast': 'América del Sur', 'az': 'Cənubi Amerika', 'az-Cyrl': 'Ҹәнуби Америка', 'az-Latn': 'Cənubi Amerika', 'bal-Latn': 'Zerbári Amriká', 'be': 'Паўднёвая Амерыка', 'be-tarask': 'Паўднёвая Амерыка', 'bg': 'Южна Америка', 'blo': 'Gamalɩkatǝna gʊnyɩsonolaŋ', 'bn': 'দক্ষিণ আমেরিকা', 'br': 'Suamerika', 'brx': 'खोला आमेरिका', 'bs': 'Južna Amerika', 'bs-Cyrl': 'Јужна Америка', 'bs-Latn': 'Južna Amerika', 'ca': 'Amèrica del Sud', 'ca-ES-valencia': 'Amèrica del Sud', 'ccp': '𑄘𑄨𑄉𑄨𑄚𑄴 𑄃𑄟𑄬𑄢𑄨𑄇', 'ce': 'Къилба Америка', 'ceb': 'South America', 'chr': 'ᏧᎦᏃᏮ ᎠᎺᎵᎦ', 'ckb': 'ئەمەریکای باشوور', 'cs': 'Jižní Amerika', 'csw': 'ᓵᐘᓄᕽ ᒥᐦᑭᓇᕽ ᒥᓂᐢᑎᐠ', 'cv': 'Кӑнтӑр Америка', 'cy': 'De America', 'da': 'Sydamerika', 'de': 'Südamerika', 'dsb': 'Pódpołdnjowa Amerika', 'dz': 'ལྷོ་ཨ་མི་རི་ཀ', 'ee': 'Anyiehe Amerika nutome', 'el': 'Νότια Αμερική', 'el-polyton': 'Νότια Αμερική', 'en': 'South America', 'en-Dsrt': '𐐝𐐵𐑃 𐐊𐑋𐐯𐑉𐐲𐐿𐐲', 'es': 'Sudamérica', 'et': 'Lõuna-Ameerika', 'eu': 'Hego Amerika', 'fa': 'امریکای جنوبی', 'ff-Adlm': '𞤙𞤢𞤥𞤲𞤢𞥄𞤲𞤺𞤫 𞤀𞤥𞤫𞤪𞤭𞤳𞤢𞥄', 'fi': 'Etelä-Amerikka', 'fil': 'Timog Amerika', 'fo': 'Suðuramerika', 'fr': 'Amérique du Sud', 'fur': 'Americhe meridionâl', 'fy': 'Sûd-Amerika', 'ga': 'Meiriceá Theas', 'gd': 'Aimeireaga a Deas', 'gl': 'América do Sur', 'gsw': 'Süüdameerika', 'gu': 'દક્ષિણ અમેરિકા', 'ha': 'Amurka ta Kudu', 'he': 'דרום אמריקה', 'hi': 'दक्षिण अमेरिका', 'hi-Latn': 'South America', 'hr': 'Južna Amerika', 'hsb': 'Južna Amerika', 'ht': 'Amérique du Sud', 'hu': 'Dél-Amerika', 'hy': 'Հարավային Ամերիկա', 'ia': 'America del Sud', 'id': 'Amerika Selatan', 'ie': 'Sud-America', 'ig': 'Mpaghara Mgbada Ugwu America', 'ii': 'ꆆꂰꍏ', 'is': 'Suður-Ameríka', 'it': 'America del Sud', 'ja': '南アメリカ', 'jv': 'Amérika Kidul', 'ka': 'სამხრეთ ამერიკა', 'kaa': 'Қубла Америка', 'kaa-Cyrl': 'Қубла Америка', 'kea': 'Merka di Sul', 'kgp': 'Sur-Amẽrika', 'kk': 'Оңтүстік Америка', 'kk-Cyrl': 'Оңтүстік Америка', 'km': 'អាមេរិក\u200bខាង\u200bត្បូង', 'kn': 'ದಕ್ಷಿಣ ಅಮೇರಿಕಾ', 'ko': '남아메리카', 'kok': 'दक्षिण अमेरिका', 'kok-Deva': 'दक्षिण अमेरिका', 'ks': 'جَنوٗنی اَمرٖیٖکا', 'ks-Arab': 'جَنوٗنی اَمرٖیٖکا', 'ksh': 'Söhd-Amärrika', 'ku': 'Amerîkaya Başûr', 'ky': 'Түштүк Америка', 'lb': 'Südamerika', 'lld': 'America dl Süd', 'lo': 'ອາເມລິກາໃຕ້', 'lrc': 'ئمریکا ھارگە', 'lt': 'Pietų Amerika', 'lv': 'Dienvidamerika', 'mai': 'दक्षिण अमेरिका', 'mi': 'Amerika ki te Tonga', 'mk': 'Јужна Америка', 'ml': 'തെക്കേ അമേരിക്ക', 'mn': 'Өмнөд Америк', 'mr': 'दक्षिण अमेरिका', 'ms': 'Amerika Selatan', 'mt': 'Amerika t’Isfel', 'my': 'တောင် အမေရိက', 'mzn': 'جنوبی آمریکا', 'nb': 'Sør-Amerika', 'ne': 'दक्षिण अमेरिका', 'nl': 'Zuid-Amerika', 'nn': 'Sør-Amerika', 'no': 'Sør-Amerika', 'nqo': 'ߊߡߋߙߌߞߌ߬ ߥߙߏ߬ߘߎ߮', 'om': 'Ameerikaa Kibbaa', 'or': 'ଦକ୍ଷିଣ ଆମେରିକା', 'pa': 'ਦੱਖਣ ਅਮਰੀਕਾ', 'pa-Guru': 'ਦੱਖਣ ਅਮਰੀਕਾ', 'pcm': 'Saút Amẹ́ríka', 'pl': 'Ameryka Południowa', 'ps': 'سويلي امريکا', 'pt': 'América do Sul', 'qu': 'Sud America', 'rm': 'America dal Sid', 'ro': 'America de Sud', 'ru': 'Южная Америка', 'sah': 'Соҕуруу Эмиэрикэ', 'sc': 'Amèrica de su Sud', 'sd': 'ڏکڻ آمريڪا', 'sd-Arab': 'ڏکڻ آمريڪا', 'se': 'mátta-Amerihkká', 'si': 'දකුණු ඇමෙරිකාව', 'sk': 'Južná Amerika', 'sl': 'Južna Amerika', 'so': 'Koonfur Ameerika', 'sq': 'Amerika e Jugut', 'sr': 'Јужна Америка', 'sr-Cyrl': 'Јужна Америка', 'sr-Latn': 'Južna Amerika', 'sv': 'Sydamerika', 'sw': 'Amerika Kusini', 'syr': 'ܐܡܪܝܟܐ ܬܝܡܢܝܬܐ', 'ta': 'தென் அமெரிக்கா', 'te': 'దక్షిణ అమెరికా', 'tg': 'Америкаи Ҷанубӣ', 'th': 'อเมริกาใต้', 'ti': 'ደቡብ ኣመሪካ', 'tk': 'Günorta Amerika', 'to': 'ʻAmelika tonga', 'tr': 'Güney Amerika', 'tt': 'Көньяк Америка', 'ug': 'جەنۇبىي ئامېرىكا', 'uk': 'Південна Америка', 'ur': 'جنوبی امریکہ', 'uz': 'Janubiy Amerika', 'uz-Cyrl': 'Жанубий Америка', 'uz-Latn': 'Janubiy Amerika', 'vec': 'Mèrega de’l sud', 'vi': 'Nam Mỹ', 'wae': 'Südamerika', 'wo': 'Amerique du Sud', 'xh': 'UMzantsi Melika', 'xnr': 'दखण अमरिका', 'yi': 'דרום־אַמעריקע', 'yo': 'Gúúṣù Amẹ́ríkà', 'yrl': 'Amerika Su suí', 'yue': '南美洲', 'yue-Hans': '南美洲', 'yue-Hant': '南美洲', 'zh': '南美洲', 'zh-Hans': '南美洲', 'zh-Hant': '南美洲', 'zu': 'i-South America'},
    '009': {'af': 'Oseanië', 'ak': 'Osiana', 'am': 'ኦሺያንያ', 'ar': 'أوقيانوسيا', 'as': 'অ’চেনীয়া', 'ast': 'Oceanía', 'az': 'Okeaniya', 'az-Cyrl': 'Океанија', 'az-Latn': 'Okeaniya', 'bal-Latn': 'Ushiáná', 'be': 'Акіянія', 'be-tarask': 'Акіянія', 'bg': 'Океания', 'bgn': 'اوقیانوسیه', 'blo': 'Oseyanii', 'bn': 'ওশিয়ানিয়া', 'br': 'Oseania', 'brx': 'ओशेआनिया', 'bs': 'Okeanija', 'bs-Cyrl': 'Океанија', 'bs-Latn': 'Okeanija', 'ca': 'Oceania', 'ca-ES-valencia': 'Oceania', 'ccp': '𑄃𑄮𑄥𑄨𑄠𑄚𑄨𑄠', 'ce': 'Океани', 'ceb': 'Oceania', 'chr': 'ᎣᏏᏰᏂᎠ', 'ckb': 'ئۆقیانووسیا', 'cs': 'Oceánie', 'cv': 'Океани', 'cy': 'Oceania', 'da': 'Oceanien', 'de': 'Ozeanien', 'dsb': 'Oceaniska', 'dz': 'ཨོཤི་ཡཱན་ན', 'ee': 'Oceania nutome', 'el': 'Ωκεανία', 'el-polyton': 'Ωκεανία', 'en': 'Oceania', 'en-Dsrt': '𐐄𐑇𐐨𐐰𐑌𐐨𐐲', 'es': 'Oceanía', 'et': 'Okeaania', 'eu': 'Ozeania', 'fa': 'اقیانوسیه', 'ff-Adlm': '𞤌𞤧𞤭𞤴𞤢𞤲𞤭𞥅', 'fi': 'Oseania', 'fil': 'Oceania', 'fo': 'Osiania', 'fr': 'Océanie', 'fur': 'Oceanie', 'fy': 'Oceanië', 'ga': 'an Aigéine', 'gd': 'Roinn a’ Chuain Sèimh', 'gl': 'Oceanía', 'gsw': 'Ozeaanie', 'gu': 'ઓશનિયા', 'ha': 'Osheniya', 'he': 'אוקיאניה', 'hi': 'ओशिआनिया', 'hi-Latn': 'Oceania', 'hr': 'Oceanija', 'hsb': 'Oceaniska', 'ht': 'Océanie', 'hu': 'Óceánia', 'hy': 'Օվկիանիա', 'ia': 'Oceania', 'id': 'Oseania', 'ie': 'Oceania', 'ig': 'Oceania', 'ii': 'ꄊꑸꍏ', 'is': 'Eyjaálfa', 'it': 'Oceania', 'ja': 'オセアニア', 'jv': 'Oséania', 'ka': 'ოკეანეთი', 'kaa': 'Океания', 'kaa-Cyrl': 'Океания', 'kea': 'Oseania', 'kgp': 'Oseanĩja', 'kk': 'Океания', 'kk-Cyrl': 'Океания', 'km': 'អូសេអានី', 'kn': 'ಓಶಿಯೇನಿಯಾ', 'ko': '오세아니아', 'kok': 'ओसेनिया', 'kok-Deva': 'ओसेनिया', 'ks': 'اوشَنیا', 'ks-Arab': 'اوشَنیا', 'ksh': 'Ozejahneje', 'ku': 'Okyanûsya', 'ky': 'Океания', 'lb': 'Ozeanien', 'lld': 'Ozeania', 'lo': 'ໂອຊີອານີ', 'lrc': 'ھوم پئڤأند جأھوٙن آڤ', 'lt': 'Okeanija', 'lv': 'Okeānija', 'mai': 'ओसेनिया', 'mi': 'Ngā Moutere-a-Kiwa', 'mk': 'Океанија', 'ml': 'ഓഷ്യാനിയ', 'mn': 'Номхон далайн орнууд', 'mr': 'ओशनिया', 'ms': 'Oceania', 'mt': 'Oċejanja', 'my': 'သမုဒ္ဒရာဒေသ', 'mzn': 'اوقیانوسیه', 'nb': 'Oseania', 'ne': 'ओसनिया', 'nl': 'Oceanië', 'nn': 'Oseania', 'no': 'Oseania', 'nqo': 'ߟߌ߲ߓߊ߲ߘߎ߯', 'om': 'Oshiiniyaa', 'or': 'ଓସିନିଆ', 'os': 'Океани', 'pa': 'ਓਸ਼ੇਨੀਆ', 'pa-Guru': 'ਓਸ਼ੇਨੀਆ', 'pcm': 'Oshẹnia', 'pl': 'Oceania', 'ps': 'اوقيانوسيه', 'pt': 'Oceania', 'qu': 'Oceania', 'rm': 'Oceania', 'ro': 'Oceania', 'ru': 'Океания', 'sc': 'Otzeània', 'sd': 'اوشنيا', 'sd-Arab': 'اوشنيا', 'se': 'Oseania', 'si': 'ඕෂනියාව', 'sk': 'Oceánia', 'sl': 'Oceanija', 'so': 'Osheeniya', 'sq': 'Oqeani', 'sr': 'Океанија', 'sr-Cyrl': 'Океанија', 'sr-Latn': 'Okeanija', 'sv': 'Oceanien', 'sw': 'Oceania', 'syr': 'ܐܘܩܝܢܘܣܝܐ', 'ta': 'ஓஷியானியா', 'te': 'ఓషియానియా', 'tg': 'Уқёнус', 'th': 'โอเชียเนีย', 'ti': 'ኦሽያንያ', 'tk': 'Okeaniýa', 'to': 'ʻOsēnia', 'tr': 'Okyanusya', 'tt': 'Океания', 'ug': 'ئوكيانىيە', 'uk': 'Океанія', 'ur': 'اوشیانیا', 'uz': 'Okeaniya', 'uz-Cyrl': 'Океания', 'uz-Latn': 'Okeaniya', 'vec': 'Oseania', 'vi': 'Châu Đại Dương', 'wae': 'Ozeanie', 'wo': 'Oseani', 'xh': 'E-Oceania', 'xnr': 'ओशिआनिया', 'yi': 'אקעאַניע', 'yo': 'Òsọ́ọ́níà', 'yrl': 'Useãniya', 'yue': '大洋洲', 'yue-Hans': '大洋洲', 'yue-Hant': '大洋洲', 'zh': '大洋洲', 'zh-Hans': '大洋洲', 'zh-Hant': '大洋洲', 'zu': 'i-Oceania'},
    '011': {'af': 'Wes-Afrika', 'ak': 'Abibirem Atɔeɛ Fam', 'am': 'ምዕራባዊ አፍሪካ', 'ar': 'غرب أفريقيا', 'as': 'পাশ্চাত্য আফ্ৰিকা', 'ast': 'África Occidental', 'az': 'Qərbi Afrika', 'az-Cyrl': 'Гәрби Африка', 'az-Latn': 'Qərbi Afrika', 'bal-Latn': 'Rónendi Apriká', 'be': 'Заходняя Афрыка', 'be-tarask': 'Заходняя Афрыка', 'bg': 'Западна Афирка', 'blo': 'Garɩɖontǝna gɩteŋshilelaŋ', 'bn': 'পশ্চিম আফ্রিকা', 'br': 'Afrika ar Cʼhornôg', 'brx': 'सोनाबारि आफ्रिका', 'bs': 'Zapadna Afrika', 'bs-Cyrl': 'Западна Африка', 'bs-Latn': 'Zapadna Afrika', 'ca': 'Àfrica occidental', 'ca-ES-valencia': 'Àfrica occidental', 'ccp': '𑄛𑄧𑄏𑄨𑄟𑄴 𑄃𑄜𑄳𑄢𑄨𑄇', 'ce': 'Малхбузен Африка', 'ceb': 'Western Africa', 'chr': 'ᏭᏕᎵᎬ ᏗᏜ ᎬᎿᎨᏍᏛ', 'ckb': 'ڕۆژاوای ئەفریقا', 'cs': 'západní Afrika', 'cv': 'Анӑҫ Африка', 'cy': 'Gorllewin Affrica', 'da': 'Vestafrika', 'de': 'Westafrika', 'dsb': 'Pódwjacorna Afrika', 'dz': 'ནུབ་ཕྱོགས་ཀྱི་ཨཕ་རི་ཀ', 'ee': 'Ɣetoɖoƒelɔƒo Afrika nutome', 'el': 'Δυτική Αφρική', 'el-polyton': 'Δυτική Αφρική', 'en': 'Western Africa', 'en-Dsrt': '𐐎𐐯𐑅𐐻𐐲𐑉𐑌 𐐈𐑁𐑉𐐲𐐿𐐲', 'es': 'África occidental', 'et': 'Lääne-Aafrika', 'eu': 'Mendebaldeko Afrika', 'fa': 'غرب افریقا', 'ff-Adlm': '𞤖𞤭𞥅𞤪𞤲𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭 𞤀𞤬𞤪𞤭𞤳𞤢𞥄', 'fi': 'Länsi-Afrikka', 'fil': 'Kanlurang Africa', 'fo': 'Vesturafrika', 'fr': 'Afrique occidentale', 'fur': 'Afriche ocidentâl', 'fy': 'West-Afrika', 'ga': 'Iarthar na hAfraice', 'gd': 'Afraga an Iar', 'gl': 'África Occidental', 'gsw': 'Weschtafrika', 'gu': 'પશ્ચિમી આફ્રિકા', 'ha': 'Afirka ta Yamma', 'he': 'מערב אפריקה', 'hi': 'पश्चिमी अफ़्रीका', 'hi-Latn': 'Western Africa', 'hr': 'Zapadna Afrika', 'hsb': 'zapadna Afrika', 'ht': 'Afrique occidentale', 'hu': 'Nyugat-Afrika', 'hy': 'Արևմտյան Աֆրիկա', 'ia': 'Africa occidental', 'id': 'Afrika Bagian Barat', 'ig': 'Mpaghara Ọdịda Anyanwụ Afrịka', 'is': 'Vestur-Afríka', 'it': 'Africa occidentale', 'ja': '西アフリカ', 'jv': 'Afrika Kulon', 'ka': 'დასავლეთ აფრიკა', 'kaa': 'Батыс Африка', 'kaa-Cyrl': 'Батыс Африка', 'kea': 'Áfrika Osidental', 'kgp': 'Afrika Rãpur', 'kk': 'Батыс Африка', 'kk-Cyrl': 'Батыс Африка', 'km': 'អាហ្វ្រិក\u200bខាង\u200bលិច', 'kn': 'ಪಶ್ಚಿಮ ಆಫ್ರಿಕಾ', 'ko': '서부 아프리카', 'kok': 'अस्तंत आफ्रिका', 'kok-Deva': 'अस्तंत आफ्रिका', 'ks': 'مَغریٖبی اَفریٖکا', 'ks-Arab': 'مَغریٖبی اَفریٖکا', 'ksh': 'Wäß-Affrika', 'ku': 'Rojavayê Afrîkayê', 'ky': 'Батыш Африка', 'lb': 'Westafrika', 'lld': 'Africa ozidentala', 'lo': 'ອາຟຣິກາຕາເວັນຕົກ', 'lt': 'Vakarų Afrika', 'lv': 'Rietumāfrika', 'mai': 'पश्चिमी अफ्रीका', 'mi': 'Āwherika ki te Uru', 'mk': 'Западна Африка', 'ml': 'പശ്ചിമ ആഫ്രിക്ക', 'mn': 'Баруун Африк', 'mr': 'पश्चिम आफ्रिका', 'ms': 'Afrika Barat', 'mt': 'Affrika tal-Punent', 'my': 'အနောက် အာဖရိက', 'mzn': 'غربی آفریقا', 'nb': 'Vest-Afrika', 'ne': 'पश्चिमी अफ्रिका', 'nl': 'West-Afrika', 'nn': 'Vest-Afrika', 'no': 'Vest-Afrika', 'nqo': 'ߝߘߊ߬ߝߌ߲߬ߠߊ߫ ߕߟߋ߬ߓߋ', 'om': 'Afrikaa Dhihaa', 'or': 'ପଶ୍ଚିମ ଆଫ୍ରିକା', 'pa': 'ਪੱਛਮੀ ਅਫ਼ਰੀਕਾ', 'pa-Guru': 'ਪੱਛਮੀ ਅਫ਼ਰੀਕਾ', 'pcm': 'Wẹ́stán Áfríka', 'pl': 'Afryka Zachodnia', 'ps': 'لویدیځ افریقا', 'pt': 'África Ocidental', 'qu': 'Africa Occidental', 'rm': 'Africa dal Vest', 'ro': 'Africa Occidentală', 'ru': 'Западная Африка', 'sc': 'Àfrica otzidentale', 'sd': 'اولهه آفريقا', 'sd-Arab': 'اولهه آفريقا', 'se': 'oarji-Afrihkká', 'si': 'බටහිරදිග අප්\u200dරිකාව', 'sk': 'západná Afrika', 'sl': 'Zahodna Afrika', 'so': 'Galbeeka Afrika', 'sq': 'Afrika Perëndimore', 'sr': 'Западна Африка', 'sr-Cyrl': 'Западна Африка', 'sr-Latn': 'Zapadna Afrika', 'sv': 'Västafrika', 'sw': 'Afrika ya Magharibi', 'syr': 'ܐܦܪܝܩܐ ܡܥܪܒܝܬܐ', 'ta': 'மேற்கு ஆப்பிரிக்கா', 'te': 'పశ్చిమ ఆఫ్రికా భూభాగం', 'tg': 'Африқои Ғарбӣ', 'th': 'แอฟริกาตะวันตก', 'ti': 'ምዕራባዊ ኣፍሪቃ', 'tk': 'Günbatar Afrika', 'to': 'ʻAfilika fakahihifo', 'tr': 'Batı Afrika', 'tt': 'Көнбатыш Африка', 'ug': 'غەربىي ئافرىقا', 'uk': 'Західна Африка', 'ur': 'مغربی افریقہ', 'uz': 'G‘arbiy Afrika', 'uz-Cyrl': 'Ғарбий Африка', 'uz-Latn': 'G‘arbiy Afrika', 'vec': 'Àfrega osidentale', 'vi': 'Tây Phi', 'wae': 'Weštafrika', 'wo': 'Sowwu Afrique', 'xh': 'INtshona Afrika', 'xnr': 'पश्चिमी अफ़्रीका', 'yo': 'Ìwọ̀ oorùn Afíríkà', 'yrl': 'Afirika Usidẽtawara', 'yue': '西非', 'yue-Hans': '西非', 'yue-Hant': '西非', 'zh': '西非', 'zh-Hans': '西非', 'zh-Hant': '西非', 'zu': 'i-Western Africa'},
    '013': {'af': 'Sentraal-Amerika', 'ak': 'Amɛrika Mfimfini', 'am': 'መካከለኛው አሜሪካ', 'ar': 'أمريكا الوسطى', 'as': 'মধ্য আমেৰিকা', 'ast': 'América Central', 'az': 'Mərkəzi Amerika', 'az-Cyrl': 'Мәркәзи Америка', 'az-Latn': 'Mərkəzi Amerika', 'bal-Latn': 'Myáni Amriká', 'be': 'Цэнтральная Амерыка', 'be-tarask': 'Цэнтральная Амерыка', 'bg': 'Централна Америка', 'blo': 'Gamalɩkatǝna gɩcɩɩca', 'bn': 'মধ্য আমেরিকা', 'br': 'Kreizamerika', 'brx': 'मिरु आमेरिका', 'bs': 'Srednja Amerika', 'bs-Cyrl': 'Централна Америка', 'bs-Latn': 'Srednja Amerika', 'ca': 'Amèrica Central', 'ca-ES-valencia': 'Amèrica Central', 'ccp': '𑄟𑄧𑄖𑄴𑄙𑄳𑄠 𑄃𑄜𑄳𑄢𑄨𑄇', 'ce': 'Юккъера Америка', 'ceb': 'Central America', 'chr': 'ᎠᏰᏟ ᎠᎹᏰᏟ', 'ckb': 'ئەمریکای ناوەڕاست', 'cs': 'Střední Amerika', 'csw': 'ᐋᐱᐦᑕᐤ ᒥᐦᑭᓇᕽ ᒥᓂᐢᑎᐠ', 'cv': 'Тӗп Америка', 'cy': 'Canolbarth America', 'da': 'Mellemamerika', 'de': 'Mittelamerika', 'dsb': 'Srjejźna Amerika', 'dz': 'བར་ཕྱོགས་ཨ་མི་རི་ཀ', 'ee': 'Titina Amerika nutome', 'el': 'Κεντρική Αμερική', 'el-polyton': 'Κεντρική Αμερική', 'en': 'Central America', 'en-Dsrt': '𐐝𐐯𐑌𐐻𐑉𐐲𐑊 𐐊𐑋𐐯𐑉𐐲𐐿𐐲', 'es': 'Centroamérica', 'et': 'Kesk-Ameerika', 'eu': 'Erdialdeko Amerika', 'fa': 'امریکای مرکزی', 'ff-Adlm': '𞤚𞤵𞤥𞤦𞤮 𞤀𞤥𞤫𞤪𞤭𞤳𞤢𞥄', 'fi': 'Väli-Amerikka', 'fil': 'Gitnang Amerika', 'fo': 'Miðamerika', 'fr': 'Amérique centrale', 'fur': 'Americhe centrâl', 'fy': 'Midden-Amerika', 'ga': 'Meiriceá Láir', 'gd': 'Meadhan Aimeireaga', 'gl': 'América Central', 'gsw': 'Mittelameerika', 'gu': 'મધ્ય અમેરિકા', 'ha': 'Amurka ta Tsakiya', 'he': 'מרכז אמריקה', 'hi': 'मध्य अमेरिका', 'hi-Latn': 'Central America', 'hr': 'Centralna Amerika', 'hsb': 'Srjedźna Amerika', 'ht': 'Amérique centrale', 'hu': 'Közép-Amerika', 'hy': 'Կենտրոնական Ամերիկա', 'ia': 'America central', 'id': 'Amerika Tengah', 'ie': 'central America', 'ig': 'Etiti America', 'is': 'Mið-Ameríka', 'it': 'America Centrale', 'ja': '中央アメリカ', 'jv': 'Amérika Tengah', 'ka': 'ცენტრალური ამერიკა', 'kaa': 'Орайлық Америка', 'kaa-Cyrl': 'Орайлық Америка', 'kea': 'Merka Sentral', 'kgp': 'Amẽrika-Kuju', 'kk': 'Орталық Америка', 'kk-Cyrl': 'Орталық Америка', 'km': 'អាមេរិក\u200bកណ្ដាល', 'kn': 'ಮಧ್ಯ ಅಮೇರಿಕಾ', 'ko': '중앙 아메리카', 'kok': 'मध्य अमेरिका', 'kok-Deva': 'मध्य अमेरिका', 'ks': 'مرکٔزی اَمریٖکا', 'ks-Arab': 'مرکٔزی اَمریٖکا', 'ksh': 'Meddelammärrika', 'ku': 'Amerîkaya Navîn', 'ky': 'Борбордук Америка', 'lb': 'Mëttelamerika', 'lld': 'America Zentrala', 'lo': 'ອາເມລິກາກາງ', 'lrc': 'مینجا ئمریکا', 'lt': 'Centrinė Amerika', 'lv': 'Centrālamerika', 'mai': 'मध्य अमेरिका', 'mi': 'Te Puku o Amerika', 'mk': 'Централна Америка', 'ml': 'മദ്ധ്യ അമേരിക്ക', 'mn': 'Төв Америк', 'mr': 'मध्य अमेरिका', 'ms': 'Amerika Tengah', 'mt': 'Amerika Ċentrali', 'my': 'ဗဟို အမေရိက', 'mzn': 'میونی آمریکا', 'nb': 'Mellom-Amerika', 'ne': 'केन्द्रीय अमेरिका', 'nl': 'Midden-Amerika', 'nn': 'Sentral-Amerika', 'no': 'Mellom-Amerika', 'nqo': 'ߊߡߋߙߌߞߌ߬ ߕߊ߲ߓߊ߲', 'om': 'Ameerikaa Gidduugaleessaa', 'or': 'ମଧ୍ୟ ଆମେରିକା', 'pa': 'ਕੇਂਦਰੀ ਅਮਰੀਕਾ', 'pa-Guru': 'ਕੇਂਦਰੀ ਅਮਰੀਕਾ', 'pcm': 'Mídúl Amẹ́ríka', 'pl': 'Ameryka Środkowa', 'ps': 'منخنۍ امريکا', 'pt': 'América Central', 'qu': 'America Central', 'rm': 'America Centrala', 'ro': 'America Centrală', 'ru': 'Центральная Америка', 'sc': 'Amèrica tzentrale', 'sd': 'وچ آمريڪا', 'sd-Arab': 'وچ آمريڪا', 'se': 'gaska-Amerihkká', 'si': 'මධ්\u200dයම ඇමෙරිකාව', 'sk': 'Stredná Amerika', 'sl': 'Srednja Amerika', 'so': 'Bartamaha Ameerika', 'sq': 'Amerika Qendrore', 'sr': 'Централна Америка', 'sr-Cyrl': 'Централна Америка', 'sr-Latn': 'Centralna Amerika', 'sv': 'Centralamerika', 'sw': 'Amerika ya Kati', 'syr': 'ܐܡܪܝܟܐ ܡܨܥܝܬܐ', 'ta': 'மத்திய அமெரிக்கா', 'te': 'మధ్య అమెరికా', 'tg': 'Америкаи Марказӣ', 'th': 'อเมริกากลาง', 'ti': 'ማእከላይ ኣመሪካ', 'tk': 'Orta Amerika', 'to': 'ʻAmelika lotoloto', 'tr': 'Orta Amerika', 'tt': 'Үзәк Америка', 'ug': 'ئوتتۇرا ئامېرىكا', 'uk': 'Центральна Америка', 'ur': 'وسطی امریکہ', 'uz': 'Markaziy Amerika', 'uz-Cyrl': 'Марказий Америка', 'uz-Latn': 'Markaziy Amerika', 'vec': 'Mèrega sentrale', 'vi': 'Trung Mỹ', 'wae': 'Zentralamerika', 'wo': 'Amerique Centrale', 'xh': 'UMbindi Melika', 'xnr': 'बिचला अमरिका', 'yi': 'צענטראַל־אַמעריקע', 'yo': 'Ààrin Gbùgbùn Àmẹ́ríkà', 'yrl': 'Amerika Piterapura', 'yue': '中美', 'yue-Hans': '中美', 'yue-Hant': '中美', 'zh': '中美洲', 'zh-Hans': '中美洲', 'zh-Hant': '中美', 'zu': 'i-Central America'},
    '014': {'af': 'Oos-Afrika', 'ak': 'Abibirem Apueiɛ Fam', 'am': 'ምስራቃዊ አፍሪካ', 'ar': 'شرق أفريقيا', 'as': 'প্ৰাচ্য আফ্ৰিকা', 'ast': 'África Oriental', 'az': 'Şərqi Afrika', 'az-Cyrl': 'Шәрги Африка', 'az-Latn': 'Şərqi Afrika', 'bal-Latn': 'Ródarátki Apriká', 'be': 'Усходняя Афрыка', 'be-tarask': 'Усходняя Афрыка', 'bg': 'Източна Африка', 'blo': 'Garɩɖontǝna gajakalaŋ', 'bn': 'পূর্ব আফ্রিকা', 'br': 'Afrika ar Reter', 'brx': 'सानजायारि आफ्रिका', 'bs': 'Istočna Afrika', 'bs-Cyrl': 'Источна Африка', 'bs-Latn': 'Istočna Afrika', 'ca': 'Àfrica oriental', 'ca-ES-valencia': 'Àfrica oriental', 'ccp': '𑄛𑄪𑄇𑄴𑄘𑄩 𑄃𑄜𑄳𑄢𑄨𑄇', 'ce': 'Малхбален Африка', 'ceb': 'Eastern Africa', 'chr': 'ᏗᎧᎸᎬ ᏗᏜ ᎬᎿᎨᏍᏛ', 'ckb': 'ڕۆژھەڵاتی ئەفریقا', 'cs': 'východní Afrika', 'cv': 'Хӗвелтухӑҫ Африка', 'cy': 'Dwyrain Affrica', 'da': 'Østafrika', 'de': 'Ostafrika', 'dsb': 'pódzajtšna Afrika', 'dz': 'ཤར་ཕྱོགས་ཀྱི་ཨཕ་རི་ཀ', 'ee': 'Ɣedzeƒe Afrika nutome', 'el': 'Ανατολική Αφρική', 'el-polyton': 'Ανατολική Αφρική', 'en': 'Eastern Africa', 'en-Dsrt': '𐐀𐑅𐐻𐐲𐑉𐑌 𐐈𐑁𐑉𐐲𐐿𐐲', 'es': 'África oriental', 'et': 'Ida-Aafrika', 'eu': 'Ekialdeko Afrika', 'fa': 'شرق افریقا', 'ff-Adlm': '𞤊𞤵𞤯𞤲𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭 𞤀𞤬𞤪𞤭𞤳𞤢𞥄', 'fi': 'Itä-Afrikka', 'fil': 'Silangang Africa', 'fo': 'Eysturafrika', 'fr': 'Afrique orientale', 'fur': 'Afriche orientâl', 'fy': 'East-Afrika', 'ga': 'Oirthear na hAfraice', 'gd': 'Afraga an Ear', 'gl': 'África Oriental', 'gsw': 'Oschtafrika', 'gu': 'પૂર્વીય આફ્રિકા', 'ha': 'Afirka ta Gabas', 'he': 'מזרח אפריקה', 'hi': 'पूर्वी अफ़्रीका', 'hi-Latn': 'Eastern Africa', 'hr': 'Istočna Afrika', 'hsb': 'wuchodna Afrika', 'ht': 'Afrique orientale', 'hu': 'Kelet-Afrika', 'hy': 'Արևելյան Աֆրիկա', 'ia': 'Africa oriental', 'id': 'Afrika Bagian Timur', 'ig': 'Mpaghara Ọwụwa Anyanwụ Afrịka', 'is': 'Austur-Afríka', 'it': 'Africa orientale', 'ja': '東アフリカ', 'jv': 'Afrika Wétan', 'ka': 'აღმოსავლეთ აფრიკა', 'kaa': 'Шығыс Африка', 'kaa-Cyrl': 'Шығыс Африка', 'kea': 'Áfrika Oriental', 'kgp': 'Afrika Rãjur', 'kk': 'Шығыс Африка', 'kk-Cyrl': 'Шығыс Африка', 'km': 'អាហ្វ្រិកខាងកើត', 'kn': 'ಪೂರ್ವ ಆಫ್ರಿಕಾ', 'ko': '동부 아프리카', 'kok': 'उदेंत आफ्रिका', 'kok-Deva': 'उदेंत आफ्रिका', 'ks': 'مَشرِقی اَفریٖکا', 'ks-Arab': 'مَشرِقی اَفریٖکا', 'ksh': 'Oß-Affrika', 'ku': 'Rojhilatê Afrîkayê', 'ky': 'Чыгыш Африка', 'lb': 'Ostafrika', 'lld': 'Africa orientala', 'lo': 'ອາຟຣິກາຕາເວັນອອກ', 'lt': 'Rytų Afrika', 'lv': 'Austrumāfrika', 'mai': 'पूरबी अफ्रीका', 'mi': 'Āwherika ki te Rāwhiti', 'mk': 'Источна Африка', 'ml': 'കിഴക്കൻ ആഫ്രിക്ക', 'mn': 'Зүүн Африк', 'mr': 'पूर्व आफ्रिका', 'ms': 'Afrika Timur', 'mt': 'Affrika tal-Lvant', 'my': 'အရှေ့ အာဖရိက', 'mzn': 'شرقی آفریقا', 'nb': 'Øst-Afrika', 'ne': 'पूर्वी अफ्रिका', 'nl': 'Oost-Afrika', 'nn': 'Aust-Afrika', 'no': 'Øst-Afrika', 'nqo': 'ߝߘߊ߬ߝߌ߲߬ߠߊ߫ ߕߟߋ߬ߓߐ', 'om': 'Afrikaa Bahaa', 'or': 'ପୂର୍ବ ଆଫ୍ରିକା', 'pa': 'ਪੂਰਬੀ ਅਫ਼ਰੀਕਾ', 'pa-Guru': 'ਪੂਰਬੀ ਅਫ਼ਰੀਕਾ', 'pcm': 'Ístán Áfríká', 'pl': 'Afryka Wschodnia', 'ps': 'ختیځ افریقا', 'pt': 'África Oriental', 'qu': 'Africa Oriental', 'rm': 'Africa da l’Ost', 'ro': 'Africa Orientală', 'ru': 'Восточная Африка', 'sc': 'Àfrica orientale', 'sd': 'اوڀر آفريڪا', 'sd-Arab': 'اوڀر آفريڪا', 'se': 'nuorta-Afrihkká', 'si': 'පෙරදිග අප්\u200dරිකාව', 'sk': 'východná Afrika', 'sl': 'Vzhodna Afrika', 'so': 'Afrikada Bari', 'sq': 'Afrika Lindore', 'sr': 'Источна Африка', 'sr-Cyrl': 'Источна Африка', 'sr-Latn': 'Istočna Afrika', 'sv': 'Östafrika', 'sw': 'Afrika ya Mashariki', 'syr': 'ܐܦܪܝܩܐ ܡܕܢܚܝܬܐ', 'ta': 'கிழக்கு ஆப்பிரிக்கா', 'te': 'తూర్పు ఆఫ్రికా', 'tg': 'Африқои Шимолӣ', 'th': 'แอฟริกาตะวันออก', 'ti': 'ምብራቓዊ ኣፍሪቃ', 'tk': 'Gündogar Afrika', 'to': 'ʻAfilika fakahahake', 'tr': 'Doğu Afrika', 'tt': 'Көнчыгыш Африка', 'ug': 'شەرقىي ئافرىقا', 'uk': 'Східна Африка', 'ur': 'مشرقی افریقہ', 'uz': 'Sharqiy Afrika', 'uz-Cyrl': 'Шарқий Африка', 'uz-Latn': 'Sharqiy Afrika', 'vec': 'Àfrega orientale', 'vi': 'Đông Phi', 'wae': 'Oštafrika', 'wo': 'Penku Afrique', 'xh': 'IMpuma Afrika', 'xnr': 'पूर्वी अफ़्रीका', 'yo': 'Ìlà Oorùn Áfíríkà', 'yrl': 'Afirika Uriẽtawara', 'yue': '東非', 'yue-Hans': '东非', 'yue-Hant': '東非', 'zh': '东非', 'zh-Hans': '东非', 'zh-Hant': '東非', 'zu': 'i-Eastern Africa'},
    '015': {'af': 'Noord-Afrika', 'ak': 'Abibirem Atifi Fam', 'am': 'ሰሜናዊ አፍሪካ', 'ar': 'شمال أفريقيا', 'as': 'উদীচ্য আফ্ৰিকা', 'ast': 'África del Norte', 'az': 'Şimali Afrika', 'az-Cyrl': 'Шимали Африка', 'az-Latn': 'Şimali Afrika', 'bal-Latn': 'Shemáli Apriká', 'be': 'Паўночная Афрыка', 'be-tarask': 'Паўночная Афрыка', 'bg': 'Северна Африка', 'blo': 'Garɩɖontǝna gʊnyɩpɛnɛlaŋ', 'bn': 'উত্তর আফ্রিকা', 'br': 'Afrika an Norzh', 'brx': 'साहायारि आफ्रिका', 'bs': 'Sjeverna Afrika', 'bs-Cyrl': 'Сјеверна Африка', 'bs-Latn': 'Sjeverna Afrika', 'ca': 'Àfrica septentrional', 'ca-ES-valencia': 'Àfrica septentrional', 'ccp': '𑄅𑄪𑄖𑄴𑄖𑄮𑄢𑄴 𑄃𑄜𑄳𑄢𑄨𑄇', 'ce': 'Къилбаседа Африка', 'ceb': 'Northern Africa', 'chr': 'ᏧᏴᏢ ᏗᏜ ᎬᎿᎨᏍᏛ', 'ckb': 'باکووری ئەفریقا', 'cs': 'severní Afrika', 'cv': 'Ҫурҫӗр Африка', 'cy': 'Gogledd Affrica', 'da': 'Nordafrika', 'de': 'Nordafrika', 'dsb': 'pódpołnocna Afrika', 'dz': 'བྱང་ཕྱོགས་ཀྱི་ཨཕ་རི་ཀ', 'ee': 'Dziehe Afrika nutome', 'el': 'Βόρεια Αφρική', 'el-polyton': 'Βόρεια Αφρική', 'en': 'Northern Africa', 'en-Dsrt': '𐐤𐐱𐑉𐑄𐐲𐑉𐑌 𐐈𐑁𐑉𐐲𐐿𐐲', 'es': 'África septentrional', 'et': 'Põhja-Aafrika', 'eu': 'Ipar Afrika', 'fa': 'شمال افریقا', 'ff-Adlm': '𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭 𞤀𞤬𞤪𞤭𞤳𞤢𞥄', 'fi': 'Pohjois-Afrikka', 'fil': 'Hilagang Africa', 'fo': 'Norðurafrika', 'fr': 'Afrique septentrionale', 'fur': 'Afriche setentrionâl', 'fy': 'Noard-Afrika', 'ga': 'Tuaisceart na hAfraice', 'gd': 'Afraga a Tuath', 'gl': 'África Setentrional', 'gsw': 'Nordafrika', 'gu': 'ઉત્તરી આફ્રિકા', 'ha': 'Arewacin Afirka', 'he': 'צפון אפריקה', 'hi': 'उत्तरी अफ़्रीका', 'hi-Latn': 'Northern Africa', 'hr': 'Sjeverna Afrika', 'hsb': 'sewjerna Afrika', 'ht': 'Afrique septentrionale', 'hu': 'Észak-Afrika', 'hy': 'Հյուսիսային Աֆրիկա', 'ia': 'Africa septentrional', 'id': 'Afrika Bagian Utara', 'ig': 'Mpaghara Ugwu Afrịka', 'is': 'Norður-Afríka', 'it': 'Nordafrica', 'ja': '北アフリカ', 'jv': 'Afrika Lor', 'ka': 'ჩრდილოეთ აფრიკა', 'kaa': 'Арқа Африка', 'kaa-Cyrl': 'Арқа Африка', 'kea': 'Norti di Áfrika', 'kgp': 'Nãrti-Afrika', 'kk': 'Солтүстік Африка', 'kk-Cyrl': 'Солтүстік Африка', 'km': 'អាហ្វ្រិក\u200bខាង\u200bជើង', 'kn': 'ಉತ್ತರ ಆಫ್ರಿಕಾ', 'ko': '북부 아프리카', 'kok': 'उत्तरीय आफ्रिका', 'kok-Deva': 'उत्तरीय आफ्रिका', 'ks': 'شُمٲلی اَفریٖکا', 'ks-Arab': 'شُمٲلی اَفریٖکا', 'ksh': 'Nood-Affrika', 'ku': 'Bakurê Afrîkayê', 'ky': 'Түндүк Африка', 'lb': 'Nordafrika', 'lld': 'Africa dl Nord', 'lo': 'ອາຟຣິກາເໜືອ', 'lt': 'Šiaurės Afrika', 'lv': 'Ziemeļāfrika', 'mai': 'उत्तरी अफ्रीका', 'mi': 'Āwherika ki te Raki', 'mk': 'Северна Африка', 'ml': 'ഉത്തരാഫ്രിക്ക', 'mn': 'Хойд Африк', 'mr': 'उत्तर आफ्रिका', 'ms': 'Afrika Utara', 'mt': 'Affrika ta’ Fuq', 'my': 'မြောက် အာဖရိက', 'mzn': 'شمالی ۀفریقا', 'nb': 'Nord-Afrika', 'ne': 'उत्तरी अफ्रिका', 'nl': 'Noord-Afrika', 'nn': 'Nord-Afrika', 'no': 'Nord-Afrika', 'nqo': 'ߝߘߊ߬ߝߌ߲߬ߠߊ߫ ߞߐ߬ߘߎ߮', 'om': 'Afrikaa Kabaa', 'or': 'ଉତ୍ତର ଆଫ୍ରିକା', 'pa': 'ਉੱਤਰੀ ਅਫ਼ਰੀਕਾ', 'pa-Guru': 'ਉੱਤਰੀ ਅਫ਼ਰੀਕਾ', 'pcm': 'Nọ́tán Áfríka', 'pl': 'Afryka Północna', 'ps': 'شمالي افریقا', 'pt': 'África Setentrional', 'qu': 'Africa del Norte', 'rm': 'Africa dal Nord', 'ro': 'Africa Septentrională', 'ru': 'Северная Африка', 'sc': 'Àfrica setentrionale', 'sd': 'اتر آفريڪا', 'sd-Arab': 'اتر آفريڪا', 'se': 'davvi-Afrihkká', 'si': 'උතුරුදිග අප්\u200dරිකාව', 'sk': 'severná Afrika', 'sl': 'Severna Afrika', 'so': 'Waqooyiga Afrika', 'sq': 'Afrika Veriore', 'sr': 'Северна Африка', 'sr-Cyrl': 'Северна Африка', 'sr-Latn': 'Severna Afrika', 'sv': 'Nordafrika', 'sw': 'Afrika ya Kaskazini', 'syr': 'ܐܦܪܝܩܐ ܓܪܒܝܝܬܐ', 'ta': 'வடக்கு ஆப்பிரிக்கா', 'te': 'ఉత్తర ఆఫ్రికా', 'tg': 'Африқои Шарқӣ', 'th': 'แอฟริกาเหนือ', 'ti': 'ሰሜናዊ ኣፍሪቃ', 'tk': 'Demirgazyk Afrika', 'to': 'ʻAfilika fakatokelau', 'tr': 'Kuzey Afrika', 'tt': 'Төньяк Африка', 'ug': 'شىمالىي ئافرىقا', 'uk': 'Північна Африка', 'ur': 'شمالی افریقہ', 'uz': 'Shimoliy Afrika', 'uz-Cyrl': 'Шимолий Африка', 'uz-Latn': 'Shimoliy Afrika', 'vec': 'Àfrega setentrionale', 'vi': 'Bắc Phi', 'wae': 'Nordafrika', 'wo': 'Afrique du Nord', 'xh': 'UMntla Afrika', 'xnr': 'उत्तरी अफ़्रीका', 'yo': 'Àríwá Afíríkà', 'yrl': 'Afirika Nuti suí', 'yue': '北非', 'yue-Hans': '北非', 'yue-Hant': '北非', 'zh': '北非', 'zh-Hans': '北非', 'zh-Hant': '北非', 'zu': 'i-Northern Africa'},
    '017': {'af': 'Midde-Afrika', 'ak': 'Abibirem Mfimfini', 'am': 'መካከለኛው አፍሪካ', 'ar': 'وسط أفريقيا', 'as': 'মধ্য আফ্ৰিকা', 'ast': 'África Central', 'az': 'Mərkəzi Afrika', 'az-Cyrl': 'Мәркәзи Африка', 'az-Latn': 'Mərkəzi Afrika', 'bal-Latn': 'Myáni Apriká', 'be': 'Сярэдняя Афрыка', 'be-tarask': 'Сярэдняя Афрыка', 'bg': 'Централна Африка', 'blo': 'Garɩɖontǝna gɩcɩɩca', 'bn': 'মধ্য আফ্রিকা', 'br': 'Afrika ar Cʼhreiz', 'brx': 'गेजेरारि आफ्रिका', 'bs': 'Srednja Afrika', 'bs-Cyrl': 'Централна Африка', 'bs-Latn': 'Srednja Afrika', 'ca': 'Àfrica central', 'ca-ES-valencia': 'Àfrica central', 'ccp': '𑄟𑄧𑄖𑄴𑄙𑄳𑄠𑄧 𑄃𑄜𑄳𑄢𑄨𑄇', 'ce': 'Юккъера Африка', 'ceb': 'Middle Africa', 'chr': 'ᎠᏰᏟ ᎬᎿᎨᏍᏛ', 'ckb': 'ناوەڕاستی ئەفریقا', 'cs': 'střední Afrika', 'cv': 'Тӗп Африка', 'cy': 'Canol Affrica', 'da': 'Centralafrika', 'de': 'Zentralafrika', 'dsb': 'srjejźna Afrika', 'dz': 'སྦུག་ཕྱོགས་ཀྱི་ཨཕ་རི་ཀ', 'ee': 'Titina Afrika nutome', 'el': 'Μέση Αφρική', 'el-polyton': 'Μέση Αφρική', 'en': 'Middle Africa', 'en-Dsrt': '𐐣𐐮𐐼𐑊 𐐈𐑁𐑉𐐮𐐿𐐲', 'es': 'África central', 'et': 'Kesk-Aafrika', 'eu': 'Erdialdeko Afrika', 'fa': 'مرکز افریقا', 'ff-Adlm': '𞤚𞤵𞤥𞤦𞤮 𞤀𞤬𞤪𞤭𞤳𞤢𞥄', 'fi': 'Keski-Afrikka', 'fil': 'Gitnang Africa', 'fo': 'Miðafrika', 'fr': 'Afrique centrale', 'fur': 'Afriche di mieç', 'fy': 'Sintraal-Afrika', 'ga': 'an Afraic Láir', 'gd': 'Meadhan Afraga', 'gl': 'África Central', 'gsw': 'Zentraalafrika', 'gu': 'મધ્ય આફ્રિકા', 'ha': 'Afirka ta Tsakiya', 'he': 'מרכז אפריקה', 'hi': 'मध्य अफ़्रीका', 'hi-Latn': 'Middle Africa', 'hr': 'Središnja Afrika', 'hsb': 'srjedźna Afrika', 'ht': 'Afrique centrale', 'hu': 'Közép-Afrika', 'hy': 'Կենտրոնական Աֆրիկա', 'ia': 'Africa central', 'id': 'Afrika Bagian Tengah', 'ig': 'Etiti Afrịka', 'is': 'Mið-Afríka', 'it': 'Africa centrale', 'ja': '中部アフリカ', 'jv': 'Afrika Sisih Tengah', 'ka': 'შუა აფრიკა', 'kaa': 'Орайлық Африка', 'kaa-Cyrl': 'Орайлық Африка', 'kea': 'Áfrika Sentral', 'kgp': 'Afrika-Kuju', 'kk': 'Орталық Африка', 'kk-Cyrl': 'Орталық Африка', 'km': 'អាហ្វ្រិក\u200bកណ្តាល', 'kn': 'ಮಧ್ಯ ಆಫ್ರಿಕಾ', 'ko': '중부 아프리카', 'kok': 'मध्य आफ्रिका', 'kok-Deva': 'मध्य आफ्रिका', 'ks': 'وسطی اَفریٖکا', 'ks-Arab': 'وسطی اَفریٖکا', 'ksh': 'Meddel-Affrika', 'ku': 'Afrîkaya Navîn', 'ky': 'Борбордук Африка', 'lb': 'Zentralafrika', 'lld': 'Africa zentrala', 'lo': 'ອາຟຣິກາກາງ', 'lt': 'Vidurio Afrika', 'lv': 'Vidusāfrika', 'mai': 'मध्य अफ्रीका', 'mi': 'Te Pokapū o Āwherika', 'mk': 'Средна Африка', 'ml': 'മദ്ധ്യ ആഫ്രിക്ക', 'mn': 'Төв Африк', 'mr': 'मध्य आफ्रिका', 'ms': 'Afrika Tengah', 'mt': 'Affrika Nofsani', 'my': 'အလယ် အာဖရိက', 'mzn': 'میونی آفریقا', 'nb': 'Sentral-Afrika', 'ne': 'मध्य अफ्रिका', 'nl': 'Centraal-Afrika', 'nn': 'Sentral-Afrika', 'no': 'Sentral-Afrika', 'nqo': 'ߝߘߊ߬ߝߌ߲߬ߠߊ߫ ߕߊ߲ߓߊ߲', 'om': 'Afrikaa Gidduugaleessaa', 'or': 'ମଧ୍ୟ ଆଫ୍ରିକା', 'pa': 'ਮੱਧ ਅਫ਼ਰੀਕਾ', 'pa-Guru': 'ਮੱਧ ਅਫ਼ਰੀਕਾ', 'pcm': 'Mídúl Áfríka', 'pl': 'Afryka Środkowa', 'ps': 'منځنۍ افریقا', 'pt': 'África Central', 'qu': 'Africa Media', 'rm': 'Africa Centrala', 'ro': 'Africa Centrală', 'ru': 'Центральная Африка', 'sc': 'Àfrica tzentrale', 'sd': 'وچ آفريڪا', 'sd-Arab': 'وچ آفريڪا', 'se': 'gaska-Afrihkká', 'si': 'මධ්\u200dයම අප්\u200dරිකාව', 'sk': 'stredná Afrika', 'sl': 'Srednja Afrika', 'so': 'Afrikada Dhexe', 'sq': 'Afrika e Mesme', 'sr': 'Централна Африка', 'sr-Cyrl': 'Централна Африка', 'sr-Latn': 'Centralna Afrika', 'sv': 'Centralafrika', 'sw': 'Afrika ya Kati', 'syr': 'ܐܦܪܝܩܐ ܡܨܥܝܬܐ', 'ta': 'மத்திய ஆப்பிரிக்கா', 'te': 'మధ్య ఆఫ్రికా', 'tg': 'Африқои Миёна', 'th': 'แอฟริกากลาง', 'ti': 'ማእከላይ ኣፍሪቃ', 'tk': 'Orta Afrika', 'to': 'ʻAfilika lotoloto', 'tr': 'Orta Afrika', 'tt': 'Урта Африка', 'ug': 'ئوتتۇرا ئافرىقا', 'uk': 'Центральна Африка', 'ur': 'وسطی افریقہ', 'uz': 'Markaziy Afrika', 'uz-Cyrl': 'Марказий Африка', 'uz-Latn': 'Markaziy Afrika', 'vec': 'Àfrega sentrale', 'vi': 'Trung Phi', 'wae': 'Mittelafrika', 'wo': 'Moyen Afrique', 'xh': 'UMbindi Afrika', 'xnr': 'बिचला अफ़्रीका', 'yo': 'Ààrín gbùngbùn Afíríkà', 'yrl': 'Afirika Piterapura', 'yue': '中非', 'yue-Hans': '中非', 'yue-Hant': '中非', 'zh': '中非', 'zh-Hans': '中非', 'zh-Hant': '中非', 'zu': 'i-Middle Africa'},
    '018': {'af': 'Suider-Afrika', 'ak': 'Abibirem Anaafoɔ Fam', 'am': 'ደቡባዊ አፍሪካ', 'ar': 'أفريقيا الجنوبية', 'as': 'দাক্ষিণাত্য আফ্ৰিকা', 'ast': 'África del Sur', 'az': 'Cənubi Afrika', 'az-Cyrl': 'Ҹәнуби Африка', 'az-Latn': 'Cənubi Afrika', 'bal-Latn': 'Jonubi Apriká', 'be': 'Паўднёвая Афрыка', 'be-tarask': 'Паўднёвая Афрыка', 'bg': 'Южноафрикански регион', 'blo': 'Garɩɖontǝna gʊnyɩsonolaŋ', 'bn': 'দক্ষিন আফ্রিকা', 'br': 'Afrika ar Su', 'brx': 'खोलायारि आफ्रिका', 'bs': 'Južna Afrika', 'bs-Cyrl': 'Јужна Африка', 'bs-Latn': 'Južna Afrika', 'ca': 'Àfrica meridional', 'ca-ES-valencia': 'Àfrica meridional', 'ce': 'Къилба Африка', 'ceb': 'Southern Africa', 'chr': 'ᏧᎦᎾᏮ ᏗᏜ ᎬᎿᎨᏍᏛ', 'ckb': 'باشووری ئەفریقا', 'cs': 'jižní Afrika', 'cv': 'Кӑнтӑр Африка', 'cy': 'Deheudir Affrica', 'da': 'Det sydlige Afrika', 'de': 'Südliches Afrika', 'dsb': 'pódpołdnjowa Afrika', 'dz': 'ལྷོའི་ཨཕ་རི་ཀ', 'ee': 'Anyiehelɔƒo Afrika nutome', 'el': 'Νότιος Αφρική', 'el-polyton': 'Νότιος Αφρική', 'en': 'Southern Africa', 'en-Dsrt': '𐐝𐐲𐑄𐐲𐑉𐑌 𐐈𐑁𐑉𐐲𐐿𐐲', 'es': 'África meridional', 'et': 'Lõuna-Aafrika', 'eu': 'Hegoaldeko Afrika', 'fa': 'جنوب افریقا', 'ff-Adlm': '𞤙𞤢𞥄𞤥𞤲𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭 𞤀𞤬𞤪𞤭𞤳𞤢𞥄', 'fi': 'eteläinen Afrikka', 'fil': 'Katimugang Africa', 'fo': 'sunnari partur av Afrika', 'fr': 'Afrique australe', 'fur': 'Afriche meridionâl', 'fy': 'Sûdelijk Afrika', 'ga': 'Deisceart na hAfraice', 'gd': 'Ceann a Deas Afraga', 'gl': 'África Meridional', 'gsw': 'Süüdlichs Afrika', 'gu': 'સધર્ન આફ્રિકા', 'ha': 'Kudancin Afirka', 'he': 'דרום יבשת אפריקה', 'hi': 'दक्षिणी अफ़्रीका', 'hi-Latn': 'Southern Africa', 'hr': 'Južna Afrika', 'hsb': 'južna Afrika', 'ht': 'Afrique australe', 'hu': 'Afrika déli része', 'hy': 'Հարավային Աֆրիկա', 'ia': 'Africa meridional', 'id': 'Afrika Bagian Selatan', 'ig': 'Mpaghara Mgbada Ugwu Afrịka', 'is': 'Suðurhluti Afríku', 'it': 'Africa del Sud', 'ja': '南部アフリカ', 'jv': 'Afrika Sisih Kidul', 'ka': 'სამხრეთ აფრიკა', 'kaa': 'Қубла Африка', 'kaa-Cyrl': 'Қубла Африка', 'kea': 'Sul di Áfrika', 'kgp': 'Afrika Mẽrinhonỹr', 'kk': 'Оңтүстік Африка аймағы', 'kk-Cyrl': 'Оңтүстік Африка аймағы', 'km': 'អាហ្វ្រិកភាគខាងត្បូង', 'kn': 'ಆಫ್ರಿಕಾದ ದಕ್ಷಿಣ ಭಾಗ', 'ko': '남부 아프리카', 'kok': 'दक्षिण आफ्रिका', 'kok-Deva': 'दक्षिण आफ्रिका', 'ks': 'جنوٗبی اَفریٖکا', 'ks-Arab': 'جنوٗبی اَفریٖکا', 'ksh': 'Söhd-Affrika', 'ku': 'Başûrê Afrîkayê', 'ky': 'Түштүк Африка', 'lb': 'Südlecht Afrika', 'lld': 'Africa dl Süd', 'lo': 'ອາຟຣິກາຕອນໃຕ້', 'lt': 'Pietinė Afrika', 'lv': 'Dienvidāfrika', 'mai': 'दक्षिणी अफ्रीका', 'mi': 'Āwherika Whakatetonga', 'mk': 'Јужна Африка', 'ml': 'തെക്കേ ആഫ്രിക്ക', 'mn': 'Өмнөд Африк тив', 'mr': 'दक्षिणी आफ्रिका', 'ms': 'Selatan Afrika', 'mt': 'Affrika t’Isfel', 'my': 'အာဖရိက တောင်ပိုင်း', 'mzn': 'جنوبی آفریقا', 'nb': 'Sørlige Afrika', 'ne': 'दक्षिणी अफ्रिका', 'nl': 'Zuidelijk Afrika', 'nn': 'Sørlege Afrika', 'no': 'Sørlige Afrika', 'nqo': 'ߝߘߊ߬ߝߌ߲߬ߠߊ߫ ߥߙߏ߬ߘߎ߮', 'om': 'Kibba Afrikaa', 'or': 'ଦକ୍ଷିଣସ୍ଥ ଆଫ୍ରିକା', 'pa': 'ਦੱਖਣੀ ਅਫ਼ਰੀਕਾ', 'pa-Guru': 'ਦੱਖਣੀ ਅਫ਼ਰੀਕਾ', 'pcm': 'Saútán Áfríka', 'pl': 'Afryka Południowa', 'ps': 'سويلي افريقا', 'pt': 'África Meridional', 'qu': 'Sud Africa', 'rm': 'Africa Meridiunala', 'ro': 'Africa Meridională', 'ru': 'Южная Африка', 'sc': 'Àfrica meridionale', 'sd': 'ڏاکڻي آمريڪا', 'sd-Arab': 'ڏاکڻي آمريڪا', 'se': 'mátta-Afrihkká', 'si': 'දකුණුදිග අප්\u200dරිකාව', 'sk': 'južné územia Afriky', 'sl': 'Južna Afrika', 'so': 'Afrikada Koonfureed', 'sq': 'Afrika Jugore', 'sr': 'Јужна Африка', 'sr-Cyrl': 'Јужна Африка', 'sr-Latn': 'Južna Afrika', 'sv': 'södra Afrika', 'sw': 'Afrika ya Kusini', 'syr': 'ܐܦܪܝܩܐ ܬܝܡܢܝܬܐ', 'ta': 'தெற்கு ஆப்பிரிக்கா', 'te': 'దక్షిణ ఆఫ్రికా భూభాగం', 'tg': 'Африқои Ҷанубӣ', 'th': 'แอฟริกาตอนใต้', 'ti': 'ደቡባዊ ኣፍሪቃ', 'tk': 'Afrikanyň günorta sebitleri', 'to': 'ʻAfilika fakatonga', 'tr': 'Afrika’nın Güneyi', 'tt': 'Көньяктагы Африка', 'ug': 'جەنۇبىي ئافرىقا رايونى', 'uk': 'Південноафриканський регіон', 'ur': 'جنوبی افریقہ کا علاقہ', 'uz': 'Janubiy Afrika', 'uz-Cyrl': 'Жануби-Африка', 'uz-Latn': 'Janubiy Afrika', 'vec': 'Àfrega meridionale', 'vi': 'Miền Nam Châu Phi', 'wae': 'Südličs Afrika', 'wo': 'Afrique du Sud', 'xh': 'EMazantsi e-Afrika', 'xnr': 'दखणे बखें अफ़्रीका', 'yo': 'Apágúúsù Áfíríkà', 'yrl': 'Afirika Meridiyunau', 'yue': '非洲南部', 'yue-Hans': '非洲南部', 'yue-Hant': '非洲南部', 'zh': '南部非洲', 'zh-Hans': '南部非洲', 'zh-Hant': '非洲南部', 'zu': 'i-Southern Africa'},
    '019': {'af': 'Amerikas', 'ak': 'Amɛrikafoɔ', 'am': 'አሜሪካ', 'ar': 'الأمريكتان', 'as': 'আমেৰিকাছ্', 'ast': 'América', 'az': 'Amerika', 'az-Cyrl': 'Америка', 'az-Latn': 'Amerika', 'bal-Latn': 'Amriká', 'be': 'Паўночная і Паўднёвая Амерыкі', 'be-tarask': 'Паўночная і Паўднёвая Амерыкі', 'bg': 'Америка', 'bgn': 'امریکای براعظم', 'blo': 'Gamalɩkatǝna', 'bn': 'আমেরিকা', 'br': 'Amerikaoù', 'brx': 'आमेरिकानि', 'bs': 'Amerika', 'bs-Cyrl': 'Сјеверна и Јужна Америка', 'bs-Latn': 'Amerika', 'ca': 'Amèrica', 'ca-ES-valencia': 'Amèrica', 'ccp': '𑄃𑄟𑄬𑄢𑄨𑄇𑄥𑄴', 'ce': 'Къилбаседа а, къилба а Америка', 'ceb': 'Americas', 'chr': 'ᎠᎺᎵᎦᎢ', 'ckb': 'ئەمەریکای باکوور و باشوور', 'cs': 'Amerika', 'cv': 'Америка', 'cy': 'Yr Amerig', 'da': 'Nord-, Mellem- og Sydamerika', 'de': 'Amerika', 'dsb': 'Amerika', 'dz': 'ཨ་མི་རི་ཀ་ཚུ', 'ee': 'Amerika nutome', 'el': 'Αμερική', 'el-polyton': 'Αμερική', 'en': 'Americas', 'en-Dsrt': '𐐊𐑋𐐯𐑉𐐲𐐿𐐲𐑆', 'es': 'América', 'et': 'Ameerika', 'eu': 'Amerika', 'fa': 'امریکا', 'ff-Adlm': '𞤀𞤥𞤫𞤪𞤭𞤳𞤢𞥄', 'fi': 'Amerikka', 'fil': 'Americas', 'fo': 'Amerika', 'fr': 'Amériques', 'fur': 'Americhis', 'fy': 'Amerika', 'ga': 'Críocha Mheiriceá', 'gd': 'An Dà Aimeireaga', 'gl': 'América', 'gsw': 'Nord-, Mittel- und Süüdameerika', 'gu': 'અમેરિકા', 'ha': 'Nahiyoyin Amurka', 'he': 'אמריקה', 'hi': 'अमेरिकाज़', 'hi-Latn': 'Americas', 'hr': 'Amerike', 'hsb': 'Amerika', 'ht': 'Amériques', 'hu': 'Amerika', 'hy': 'Ամերիկա', 'ia': 'Americas', 'id': 'Amerika', 'ie': 'Americas', 'ig': 'Amerịka', 'ii': 'ꂰꍏ', 'is': 'Ameríka', 'it': 'Americhe', 'ja': 'アメリカ大陸', 'jgo': 'Amɛlîk', 'jv': 'Amérika', 'ka': 'ამერიკები', 'kaa': 'Америка', 'kaa-Cyrl': 'Америка', 'kea': 'Merkas', 'kgp': 'Amẽrika ag', 'kk': 'Америка', 'kk-Cyrl': 'Америка', 'km': 'អាមេរិក', 'kn': 'ಅಮೆರಿಕಾಸ್', 'ko': '아메리카 대륙', 'kok': 'अमेरिकास', 'kok-Deva': 'अमेरिकास', 'ks': 'اَمریٖکَس', 'ks-Arab': 'اَمریٖکَس', 'ksh': 'Ammärrika', 'ku': 'Amerîka', 'ky': 'Америка', 'lb': 'Amerika', 'lkt': 'Khéya Wíta', 'lld': 'Americhes', 'lo': 'ອາເມຣິກາ', 'lrc': 'ئمریکا', 'lt': 'Amerika', 'lv': 'Amerika', 'mai': 'अमेरिका', 'mi': 'Ngā Amerika', 'mk': 'Америки', 'ml': 'അമേരിക്കകൾ', 'mn': 'Америк', 'mni': 'অমেরিকাশিং', 'mni-Beng': 'অমেরিকাশিং', 'mr': 'अमेरिका', 'ms': 'Amerika', 'mt': 'Amerika', 'my': 'အမေရိကန်', 'mzn': 'آمریکا', 'nb': 'Amerika', 'ne': 'अमेरिकास', 'nl': 'Amerika', 'nn': 'Amerika', 'no': 'Amerika', 'nqo': 'ߊߡߋߙߌߞߌ߬', 'om': 'Ameerikaa', 'or': 'ଆମେରିକା', 'os': 'Америкӕ', 'pa': 'ਅਮਰੀਕਾ', 'pa-Guru': 'ਅਮਰੀਕਾ', 'pcm': 'Amẹ́ríkas', 'pl': 'Ameryka', 'ps': 'امريکې', 'pt': 'Américas', 'qu': 'America', 'rm': 'americas', 'ro': 'Americi', 'ru': 'Америка', 'sc': 'Amèricas', 'sd': 'آمريڪا', 'sd-Arab': 'آمريڪا', 'se': 'Amerihkká', 'si': 'ඇමරිකාව', 'sk': 'Amerika', 'sl': 'Amerike', 'so': 'Ameerikaas', 'sq': 'Amerikë', 'sr': 'Северна и Јужна Америка', 'sr-Cyrl': 'Северна и Јужна Америка', 'sr-Latn': 'Severna i Južna Amerika', 'sv': 'Nord- och Sydamerika', 'sw': 'Amerika', 'syr': 'ܐܡܪ̈ܝܟܐ', 'ta': 'அமெரிக்காஸ்', 'te': 'అమెరికాస్', 'tg': 'Америка', 'th': 'อเมริกา', 'ti': 'ኣመሪካታት', 'tk': 'Amerika', 'to': 'Ongo ʻAmelika', 'tr': 'Amerika', 'tt': 'Америка', 'ug': 'ئامېرىكا', 'uk': 'Американський регіон', 'ur': 'امیریکاز', 'uz': 'Amerika', 'uz-Cyrl': 'Америка', 'uz-Latn': 'Amerika', 'vec': 'Mèreghe', 'vi': 'Châu Mỹ', 'wae': 'Amerikaniš Kontinänt', 'wo': 'Amerika', 'xh': 'EMzantsi Melika', 'xnr': 'अमरिका दा', 'yi': 'אַמעריקע', 'yo': 'Amẹ́ríkà', 'yrl': 'America-ita', 'yue': '美洲', 'yue-Hans': '美洲', 'yue-Hant': '美洲', 'zh': '美洲', 'zh-Hans': '美洲', 'zh-Hant': '美洲', 'zu': 'i-Americas'},
    '021': {'af': 'Noordelike Amerika', 'ak': 'Amɛrika Atifi Fam', 'am': 'ሰሜናዊ አሜሪካ', 'ar': 'شمال أمريكا', 'as': 'উদীচ্য আমেৰিকা', 'ast': 'América del Norte', 'az': 'Şimal Amerikası', 'az-Cyrl': 'Шимал Америкасы', 'az-Latn': 'Şimal Amerikası', 'bal-Latn': 'Shemáli Amrika', 'be': 'Паўночнаамерыканскі рэгіён', 'be-tarask': 'Паўночнаамерыканскі рэгіён', 'bg': 'Северна Америка', 'blo': 'Gamalɩkatǝna kagʊnyɩpɛnɛlaŋ', 'bn': 'উত্তরাঞ্চলীয় আমেরিকা', 'br': 'Amerika an Norzh', 'brx': 'साहायारि आमेरिका', 'bs': 'Sjeverni dio Amerike', 'bs-Cyrl': 'Сјеверна Америка', 'bs-Latn': 'Sjeverni dio Amerike', 'ca': 'Amèrica septentrional', 'ca-ES-valencia': 'Amèrica septentrional', 'ccp': '𑄅𑄪𑄖𑄴𑄖𑄮𑄢𑄴 𑄎𑄉𑄢𑄴 𑄃𑄟𑄬𑄢𑄨𑄇', 'ce': 'Къилбаседа Америка – АЦШ а, Канада а', 'ceb': 'Northern America', 'chr': 'ᏧᏴᏢ ᏗᏜ ᎠᎹᏰᏟ', 'ckb': 'ئەمریکای باکوور', 'cs': 'Severní Amerika (oblast)', 'csw': 'ᑮᐍᑎᓄᕽ ᒥᐦᑭᓇᕽ ᒥᓂᐢᑎᐠ', 'cv': 'Ҫурҫӗр Америка регион', 'cy': 'America i’r Gogledd o Fecsico', 'da': 'Det nordlige Amerika', 'de': 'Nördliches Amerika', 'dsb': 'pódpołnocny ameriski kontinent', 'dz': 'བྱང་ཕྱོགས་ཀྱི་ཨ་མི་རི་ཀ', 'ee': 'Dziehelɔƒo Amerika nutome', 'el': 'Βόρειος Αμερική', 'el-polyton': 'Βόρειος Αμερική', 'en': 'Northern America', 'en-Dsrt': '𐐤𐐱𐑉𐑄𐐲𐑉𐑌 𐐊𐑋𐐯𐑉𐐲𐐿𐐲', 'es': 'Norteamérica', 'et': 'Ameerika põhjaosa', 'eu': 'Amerikako iparraldea', 'fa': 'شمال امریکا', 'ff-Adlm': '𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭 𞤀𞤥𞤫𞤪𞤭𞤳𞤢𞥄', 'fi': 'pohjoinen Amerikka', 'fil': 'Northern America', 'fo': 'Amerika norðanfyri Meksiko', 'fr': 'Amérique septentrionale', 'fur': 'Americhe setentrionâl', 'fy': 'Noardlik Amerika', 'ga': 'Tuaisceart Mheiriceá', 'gd': 'Ceann a Tuath Aimeireaga', 'gl': 'América Setentrional', 'gsw': 'Nördlichs Ameerika', 'gu': 'ઉત્તરી અમેરિકા', 'ha': 'Arewacin Amurka', 'he': 'אמריקה הצפונית', 'hi': 'उत्तरी अमेरिका', 'hi-Latn': 'Northern America', 'hr': 'Sjeverna Amerika', 'hsb': 'sewjerny ameriski kontinent', 'ht': 'Amérique septentrionale', 'hu': 'Amerika északi része', 'hy': 'Հյուսիսային Ամերիկա - ԱՄՆ և Կանադա', 'ia': 'America septentrional', 'id': 'Amerika Bagian Utara', 'ie': 'Nord-America', 'ig': 'Mpaghara Ugwu America', 'is': 'Ameríka norðan Mexíkó', 'it': 'America del Nord', 'ja': '北アメリカ', 'jv': 'Amérika Sisih Lor', 'ka': 'ამერიკის ჩრდილოეთი', 'kaa': 'Арқа Америка – AҚШ ҳәм Канада', 'kaa-Cyrl': 'Арқа Америка – AҚШ ҳәм Канада', 'kea': 'Norti di Merka', 'kgp': 'Amẽrika Setẽntrionỹr', 'kk': 'Солтүстік Америка (аймақ)', 'kk-Cyrl': 'Солтүстік Америка (аймақ)', 'km': 'អាមេរិក\u200bភាគ\u200bខាង\u200bជើង', 'kn': 'ಅಮೇರಿಕಾದ ಉತ್ತರ ಭಾಗ', 'ko': '북부 아메리카', 'kok': 'उत्तरीय अमेरिका', 'kok-Deva': 'उत्तरीय अमेरिका', 'ks': 'شُمٲلی اَمریٖکا خٕطہٕ', 'ks-Arab': 'شُمٲلی اَمریٖکا خٕطہٕ', 'ksh': 'der Norde vun Amärrika', 'ku': 'Bakurê Amerîkayê', 'ky': 'Түндүк Америка (чөлкөм)', 'lb': 'Nërdlecht Amerika', 'lld': 'America dl Nord', 'lo': 'ພາກເໜືອອາເມລີກາ', 'lrc': 'ئمریکا ڤارو', 'lt': 'Šiaurinė Amerika', 'lv': 'Amerikas ziemeļu daļa', 'mai': 'उत्तर अमेरिका', 'mi': 'Te Raki o Amerika', 'mk': 'Северна континентална Америка', 'ml': 'വടക്കൻ അമേരിക്ക', 'mn': 'Хойд Америк тив', 'mr': 'उत्तरी अमेरिका', 'ms': 'Utara Amerika', 'my': 'မြောက် အမေရိက', 'mzn': 'شمالی امریکا', 'nb': 'Nordlige Amerika', 'ne': 'उत्तरी अमेरिका', 'nl': 'Noordelijk Amerika', 'nn': 'Nordlege Amerika', 'no': 'Nordlige Amerika', 'nqo': 'ߞߐ߰ߘߎ߮ ߊߡߋߙߌߞߌ߬', 'om': 'Ameerikaa Kaaba', 'or': 'ଉତ୍ତରସ୍ଥ ଆମେରିକା', 'pa': 'ਉੱਤਰੀ ਅਮਰੀਕਾ', 'pa-Guru': 'ਉੱਤਰੀ ਅਮਰੀਕਾ', 'pcm': 'Nọ́tán Amẹ́ríka', 'pl': 'Ameryka Północna (USA, Kanada)', 'ps': 'شمالي امریکا', 'pt': 'América Setentrional', 'qu': 'America del Norte', 'rm': 'Amercia dal Nord', 'ro': 'America Septentrională', 'ru': 'Североамериканский регион', 'sc': 'Amèrica setentrionale', 'sd': 'اترين آمريڪا', 'sd-Arab': 'اترين آمريڪا', 'se': 'dávvi-Amerihkká', 'si': 'උතුරුදිග ඇමෙරිකාව', 'sk': 'severné územia Ameriky', 'sl': 'severnoameriška celina', 'so': 'Waqooyiga Ameerika', 'sq': 'Amerika Veriore', 'sr': 'Северна Америка', 'sr-Cyrl': 'Северна Америка', 'sr-Latn': 'Severna Amerika', 'sv': 'norra Amerika', 'sw': 'Amerika ya Kaskazini', 'syr': 'ܓܪܒܝܐ ܐܡܪܝܟܐ', 'ta': 'வடக்கு அமெரிக்கா', 'te': 'ఉత్తర అమెరికా భూభాగం', 'tg': 'Амрикои Шимолӣ', 'th': 'อเมริกาตอนเหนือ', 'ti': 'ሰሜናዊ ኣመሪካ', 'tk': 'Amerikanyň demirgazyk ýurtlary', 'to': 'ʻAmelika fakatokelau', 'tr': 'Amerika’nın Kuzeyi', 'tt': 'Төньяктагы Америка', 'ug': 'شىمالىي ئامېرىكا رايونى', 'uk': 'Північноамериканський регіон', 'ur': 'شمالی امریکہ کا علاقہ', 'uz': 'Shimoliy Amerika – AQSH va Kanada', 'uz-Cyrl': 'Шимоли-Америка', 'uz-Latn': 'Shimoliy Amerika – AQSH va Kanada', 'vec': 'Mèrega setentrionale', 'vi': 'Miền Bắc Châu Mỹ', 'wae': 'Nördličs Amerika', 'wo': 'amerique du nord', 'xh': 'EMantla Melika', 'xnr': 'उत्तरी अमरिका', 'yi': 'צפונדיקע אַמעריקע', 'yo': 'Apáàríwá Amẹ́ríkà', 'yrl': 'Amerika Setẽtiriunau', 'yue': '北美', 'yue-Hans': '北美', 'yue-Hant': '北美', 'zh': '美洲北部', 'zh-Hans': '美洲北部', 'zh-Hant': '北美', 'zu': 'i-Northern America'},
    '029': {'af': 'Karibiese streek', 'ak': 'Karibia', 'am': 'ካሪቢያን', 'ar': 'الكاريبي', 'as': 'কেৰিবিয়ান', 'ast': 'Caribe', 'az': 'Karib', 'az-Cyrl': 'Кариб', 'az-Latn': 'Karib', 'bal-Latn': 'Kerébian', 'be': 'Карыбскі рэгіён', 'be-tarask': 'Карыбскі рэгіён', 'bg': 'Карибски регион', 'blo': 'Karayiib', 'bn': 'ক্যারিবিয়ান', 'br': 'Karib', 'brx': 'केरिबियान', 'bs': 'Karibi', 'bs-Cyrl': 'Кариби', 'bs-Latn': 'Karibi', 'ca': 'Carib', 'ca-ES-valencia': 'Carib', 'ccp': '𑄇𑄳𑄠𑄢𑄝𑄨𑄠𑄚𑄴', 'ce': 'Карибаш', 'ceb': 'Caribbean', 'chr': 'ᎨᏆᏙᏯ', 'ckb': 'کاریبی', 'cs': 'Karibik', 'cv': 'Карибсем', 'cy': 'Y Caribî', 'da': 'Caribien', 'de': 'Karibik', 'dsb': 'Karibiska', 'dz': 'ཀེ་རི་བི་ཡེན', 'ee': 'Karibbea nutome', 'el': 'Καραϊβική', 'el-polyton': 'Καραϊβική', 'en': 'Caribbean', 'en-Dsrt': '𐐗𐐯𐑉𐐲𐐺𐐨𐐲𐑌', 'es': 'Caribe', 'et': 'Kariibi piirkond', 'eu': 'Karibea', 'fa': 'کارائیب', 'ff-Adlm': '𞤑𞤢𞤪𞤦𞤭𞤴𞤢𞥄', 'fi': 'Karibia', 'fil': 'Carribbean', 'fo': 'Karibia', 'fr': 'Caraïbes', 'fur': 'caraibic', 'fy': 'Karibysk gebiet', 'ga': 'an Mhuir Chairib', 'gd': 'Am Muir Caraibeach', 'gl': 'Caribe', 'gsw': 'Karibik', 'gu': 'કેરિબિયન', 'ha': 'Karebiyan', 'he': 'האיים הקריביים', 'hi': 'कैरिबियन', 'hi-Latn': 'Caribbean', 'hr': 'Karibi', 'hsb': 'Karibika', 'ht': 'Caraïbes', 'hu': 'Karib-térség', 'hy': 'Կարիբներ', 'ia': 'Caribes', 'id': 'Kepulauan Karibia', 'ig': 'Caribbean', 'is': 'Karíbahafið', 'it': 'Caraibi', 'ja': 'カリブ', 'jv': 'Karibia', 'ka': 'კარიბის ზღვა', 'kaa': 'Кариб бассейни', 'kaa-Cyrl': 'Кариб бассейни', 'kea': 'Karaibas', 'kgp': 'Karime', 'kk': 'Кариб', 'kk-Cyrl': 'Кариб', 'km': 'ការ៉ាប៊ីន', 'kn': 'ಕೆರೀಬಿಯನ್', 'ko': '카리브 제도', 'kok': 'कॅरिबियन', 'kok-Deva': 'कॅरिबियन', 'ks': 'کَرِببیٖن', 'ks-Arab': 'کَرِببیٖن', 'ksh': 'de Karribik', 'ku': 'Karayîb', 'ky': 'Кариб аралдары', 'lb': 'Karibik', 'lld': 'Caraibi', 'lo': 'ຄາຣິບບຽນ', 'lrc': 'کارائیب', 'lt': 'Karibai', 'lv': 'Karību jūras reģions', 'mai': 'कैरिबियाई', 'mi': 'Karapīana', 'mk': 'Кариби', 'ml': 'കരീബിയൻ', 'mn': 'Карибын тэнгис', 'mr': 'कॅरीबियन', 'ms': 'Caribbean', 'mt': 'Karibew', 'my': 'ကာရစ်ဘီယံ', 'mzn': 'کاراییب', 'nb': 'Karibia', 'ne': 'क्यारिबियन', 'nl': 'Caribisch gebied', 'nn': 'Karibia', 'no': 'Karibia', 'nqo': 'ߞߙߊߦߌߓ', 'om': 'Kariibiyaan', 'or': 'କାରିବିଆନ୍', 'pa': 'ਕੈਰੇਬੀਆਈ', 'pa-Guru': 'ਕੈਰੇਬੀਆਈ', 'pcm': 'Karíbián', 'pl': 'Karaiby', 'ps': 'کیریبین', 'pt': 'Caribe', 'qu': 'Caribe', 'rm': 'Caribica', 'ro': 'Caraibe', 'ru': 'Карибы', 'sc': 'Caràibes', 'sd': 'ڪيريبين', 'sd-Arab': 'ڪيريبين', 'se': 'Karibia', 'si': 'කැරීබියන්', 'sk': 'Karibik', 'sl': 'Karibi', 'so': 'Karibiyaan', 'sq': 'Karaibe', 'sr': 'Кариби', 'sr-Cyrl': 'Кариби', 'sr-Latn': 'Karibi', 'sv': 'Karibien', 'sw': 'Karibiani', 'syr': 'ܟܐܪܝܒܝܢ', 'ta': 'கரீபியன்', 'te': 'కరీబియన్', 'tg': 'Кариб', 'th': 'แคริบเบียน', 'ti': 'ካሪብያን', 'tk': 'Karib basseýni', 'to': 'Kalipiane', 'tr': 'Karayipler', 'tt': 'Кариб бассейны', 'ug': 'كارىب دېڭىزى', 'uk': 'Карибський басейн', 'ur': 'کریبیائی', 'uz': 'Karib havzasi', 'uz-Cyrl': 'Кариб ҳавзаси', 'uz-Latn': 'Karib havzasi', 'vec': 'Caràibi', 'vi': 'Ca-ri-bê', 'wae': 'Karibik', 'wo': 'Caraïbe', 'xh': 'ECaribbean', 'xnr': 'कैरिबियन', 'yi': 'קאַראַאיבע', 'yo': 'Káríbíànù', 'yrl': 'Karíbi', 'yue': '加勒比海', 'yue-Hans': '加勒比海', 'yue-Hant': '加勒比海', 'zh': '加勒比地区', 'zh-Hans': '加勒比地区', 'zh-Hant': '加勒比海', 'zu': 'i-Caribbean'},
    '030': {'af': 'Oos-Asië', 'ak': 'Asia Apueiɛ', 'am': 'ምስራቃዊ እስያ', 'ar': 'شرق آسيا', 'as': 'প্ৰাচ্য এছিয়া', 'ast': 'Asia Oriental', 'az': 'Şərqi Asiya', 'az-Cyrl': 'Шәрги Асија', 'az-Latn': 'Şərqi Asiya', 'bal-Latn': 'Ródarátki Ásiá', 'be': 'Усходняя Азія', 'be-tarask': 'Усходняя Азія', 'bg': 'Източна Азия', 'blo': 'Gacǝlǝŋtǝna gajakalaŋ', 'bn': 'পূর্ব এশিয়া', 'br': 'Azia ar Reter', 'brx': 'सानजायारि एसिया', 'bs': 'Istočna Azija', 'bs-Cyrl': 'Источна Азија', 'bs-Latn': 'Istočna Azija', 'ca': 'Àsia oriental', 'ca-ES-valencia': 'Àsia oriental', 'ccp': '𑄛𑄪𑄉𑄬𑄘𑄩 𑄃𑄬𑄥𑄨𑄠', 'ce': 'Юккъера Ази', 'ceb': 'Eastern Asia', 'chr': 'ᏗᎧᎸᎬ ᏗᏜ ᏓᎶᏂᎨᏍᏛ', 'ckb': 'ڕۆژهەڵاتی ئاسیا', 'cs': 'východní Asie', 'cv': 'Хӗвелтухӑҫ Ази', 'cy': 'Dwyrain Asia', 'da': 'Østasien', 'de': 'Ostasien', 'dsb': 'pódzajtšna Azija', 'dz': 'ཤར་ཕྱོགས་ཀྱི་ཨེ་ཤི་ཡ', 'ee': 'Ɣedzeƒe Asia nutome', 'el': 'Ανατολική Ασία', 'el-polyton': 'Ανατολική Ασία', 'en': 'Eastern Asia', 'en-Dsrt': '𐐀𐑅𐐻𐐲𐑉𐑌 𐐁𐑈𐐲', 'es': 'Asia oriental', 'et': 'Ida-Aasia', 'eu': 'Ekialdeko Asia', 'fa': 'شرق آسیا', 'ff-Adlm': '𞤊𞤵𞤯𞤲𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭 𞤀𞥄𞤧𞤭𞤴𞤢', 'fi': 'Itä-Aasia', 'fil': 'Silangang Asya', 'fo': 'Eysturasia', 'fr': 'Asie de l’Est', 'fur': 'Asie orientâl', 'fy': 'East-Azië', 'ga': 'Oirthear na hÁise', 'gd': 'Àisia an Ear', 'gl': 'Asia Oriental', 'gsw': 'Oschtaasie', 'gu': 'પૂર્વીય એશિયા', 'ha': 'Gabashin Asiya', 'he': 'מזרח אסיה', 'hi': 'पूर्वी एशिया', 'hi-Latn': 'Eastern Asia', 'hr': 'Istočna Azija', 'hsb': 'wuchodna Azija', 'ht': 'Asie de l’Est', 'hu': 'Kelet-Ázsia', 'hy': 'Արևելյան Ասիա', 'ia': 'Asia oriental', 'id': 'Asia Bagian Timur', 'ig': 'Mpaghara Ọwụwa Anyanwụ Asia', 'is': 'Austur-Asía', 'it': 'Asia orientale', 'ja': '東アジア', 'jv': 'Asia Wétan', 'ka': 'აღმოსავლეთ აზია', 'kaa': 'Шығыс Азия', 'kaa-Cyrl': 'Шығыс Азия', 'kea': 'Ázia Oriental', 'kgp': 'Ajia Rãjur', 'kk': 'Шығыс Азия', 'kk-Cyrl': 'Шығыс Азия', 'km': 'អាស៊ី\u200bខាង\u200bកើត', 'kn': 'ಪೂರ್ವ ಏಷ್ಯಾ', 'ko': '동아시아', 'kok': 'उदेंत आशिया', 'kok-Deva': 'उदेंत आशिया', 'ks': 'مَشرِقی ایشیا', 'ks-Arab': 'مَشرِقی ایشیا', 'ksh': 'Oß-Aasije', 'ku': 'Rojhilatê Asyayê', 'ky': 'Чыгыш Азия', 'lb': 'Ostasien', 'lld': 'Asia orientala', 'lo': 'ອາຊີຕາເວັນອອກ', 'lt': 'Rytų Azija', 'lv': 'Austrumāzija', 'mai': 'पूरबी एशिया', 'mi': 'Āhia ki te Rāwhiti', 'mk': 'Источна Азија', 'ml': 'കിഴക്കൻ ഏഷ്യ', 'mn': 'Зүүн Ази', 'mr': 'पूर्व आशिया', 'ms': 'Asia Timur', 'mt': 'Asja tal-Lvant', 'my': 'အရှေ့အာရှ', 'mzn': 'شرقی آسیا', 'nb': 'Øst-Asia', 'ne': 'पूर्वी एशिया', 'nl': 'Oost-Azië', 'nn': 'Aust-Asia', 'no': 'Øst-Asia', 'nqo': 'ߊߖ߭ߌ߫ ߕߟߋ߬ߓߐ', 'om': 'Eeshiyaa Bahaa', 'or': 'ପୂର୍ବ ଏସିଆ', 'pa': 'ਪੂਰਬੀ ਏਸ਼ੀਆ', 'pa-Guru': 'ਪੂਰਬੀ ਏਸ਼ੀਆ', 'pcm': 'Ístán Éshia', 'pl': 'Azja Wschodnia', 'ps': 'ختیځ آسیا', 'pt': 'Ásia Oriental', 'qu': 'Asia Oriental', 'rm': 'Asia da l’Ost', 'ro': 'Asia Orientală', 'ru': 'Восточная Азия', 'sc': 'Àsia orientale', 'sd': 'اوڀر ايشيا', 'sd-Arab': 'اوڀر ايشيا', 'se': 'nuorta-Ásia', 'si': 'නැගෙනහිර ආසියාව', 'sk': 'východná Ázia', 'sl': 'Vzhodna Azija', 'so': 'Aasiyada Bari', 'sq': 'Azia Lindore', 'sr': 'Источна Азија', 'sr-Cyrl': 'Источна Азија', 'sr-Latn': 'Istočna Azija', 'sv': 'Östasien', 'sw': 'Asia ya Mashariki', 'syr': 'ܐܣܝܐ ܡܕܢܚܝܬܐ', 'ta': 'கிழக்கு ஆசியா', 'te': 'తూర్పు ఆసియా', 'tg': 'Осиёи Шарқӣ', 'th': 'เอเชียตะวันออก', 'ti': 'ምብራቓዊ ኤስያ', 'tk': 'Gündogar Aziýa', 'to': 'ʻĒsia fakahahake', 'tr': 'Doğu Asya', 'tt': 'Көнчыгыш Азия', 'ug': 'شەرقىي ئاسىيا', 'uk': 'Східна Азія', 'ur': 'مشرقی ایشیا', 'uz': 'Sharqiy Osiyo', 'uz-Cyrl': 'Шарқий Осиё', 'uz-Latn': 'Sharqiy Osiyo', 'vec': 'Azia orientale', 'vi': 'Đông Á', 'wae': 'Oštasie', 'wo': 'Asie penku', 'xh': 'IMpuma Asia', 'xnr': 'पूर्वी एशिया', 'yi': 'מזרח אַזיע', 'yo': 'Ìlà Òòrùn Eṣíà', 'yrl': 'Ásiya Uriẽtawara', 'yue': '東亞', 'yue-Hans': '东亚', 'yue-Hant': '東亞', 'zh': '东亚', 'zh-Hans': '东亚', 'zh-Hant': '東亞', 'zu': 'i-Eastern Asia'},
    '034': {'af': 'Suid-Asië', 'ak': 'Asia Anaafoɔ', 'am': 'ደቡባዊ እስያ', 'ar': 'جنوب آسيا', 'as': 'দাক্ষিণাত্য এছিয়া', 'ast': 'Asia del Sur', 'az': 'Cənubi Asiya', 'az-Cyrl': 'Ҹәнуби Асија', 'az-Latn': 'Cənubi Asiya', 'bal-Latn': 'Zerbári Ásiá', 'be': 'Паўднёвая Азія', 'be-tarask': 'Паўднёвая Азія', 'bg': 'Южна Азия', 'blo': 'Gacǝlǝŋtǝna gʊnyɩsonolaŋ', 'bn': 'দক্ষিণ এশিয়া', 'br': 'Azia ar Su', 'brx': 'खोलायारि एसिया', 'bs': 'Južna Azija', 'bs-Cyrl': 'Јужна Азија', 'bs-Latn': 'Južna Azija', 'ca': 'Àsia meridional', 'ca-ES-valencia': 'Àsia meridional', 'ccp': '𑄘𑄧𑄉𑄨𑄚𑄬 𑄃𑄬𑄥𑄨𑄠', 'ce': 'Къилба Ази', 'ceb': 'Southern Asia', 'chr': 'ᏧᎦᎾᏮ ᏗᏜ ᏓᎶᏂᎨᏍᏛ', 'ckb': 'باشووری ئاسیا', 'cs': 'jižní Asie', 'cv': 'Кӑнтӑр Ази', 'cy': 'De Asia', 'da': 'Sydasien', 'de': 'Südasien', 'dsb': 'pódpołdnjowa Azija', 'dz': 'ལྷོའི་ཨེ་ཤི་ཡ', 'ee': 'Anyiehelɔƒo Asia nutome', 'el': 'Νότια Ασία', 'el-polyton': 'Νότια Ασία', 'en': 'Southern Asia', 'en-Dsrt': '𐐝𐐲𐑄𐐲𐑉𐑌 𐐁𐑈𐐲', 'es': 'Asia meridional', 'et': 'Lõuna-Aasia', 'eu': 'Hego Asia', 'fa': 'جنوب آسیا', 'ff-Adlm': '𞤙𞤢𞤥𞤲𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭 𞤀𞥄𞤧𞤭𞤴𞤢', 'fi': 'Etelä-Aasia', 'fil': 'Katimugang Asya', 'fo': 'Suðurasia', 'fr': 'Asie du Sud', 'fur': 'Asie meridionâl', 'fy': 'Sûd-Azië', 'ga': 'Deisceart na hÁise', 'gd': 'Àisia a Deas', 'gl': 'Asia Meridional', 'gsw': 'Süüdaasie', 'gu': 'દક્ષિણ એશિયા', 'ha': 'Kudancin Asiya', 'he': 'דרום אסיה', 'hi': 'दक्षिणी एशिया', 'hi-Latn': 'Southern Asia', 'hr': 'Južna Azija', 'hsb': 'južna Azija', 'ht': 'Asie du Sud', 'hu': 'Dél-Ázsia', 'hy': 'Հարավային Ասիա', 'ia': 'Asia meridional', 'id': 'Asia Bagian Selatan', 'ig': 'Mpaghara Mgbada Ugwu Asia', 'is': 'Suður-Asía', 'it': 'Asia del Sud', 'ja': '南アジア', 'jv': 'Asia Kidul', 'ka': 'სამხრეთ აზია', 'kaa': 'Қубла Азия', 'kaa-Cyrl': 'Қубла Азия', 'kea': 'Sul di Ázia', 'kgp': 'Ajia Mẽrinhonỹr', 'kk': 'Оңтүстік Азия', 'kk-Cyrl': 'Оңтүстік Азия', 'km': 'អាស៊ី\u200bខាង\u200bត្បូង', 'kn': 'ದಕ್ಷಿಣ ಏಷ್ಯಾ', 'ko': '남아시아', 'kok': 'दक्षिण आशिया', 'kok-Deva': 'दक्षिण आशिया', 'ks': 'جنوٗبی ایشیا', 'ks-Arab': 'جنوٗبی ایشیا', 'ksh': 'Söhd-Aasije', 'ku': 'Başûrê Asyayê', 'ky': 'Түштүк Азия', 'lb': 'Südasien', 'lld': 'Asia dl Süd', 'lo': 'ອາຊີໃຕ້', 'lt': 'Pietų Azija', 'lv': 'Dienvidāzija', 'mai': 'दक्षिणी एशिया', 'mi': 'Āhia ki te Tonga', 'mk': 'Јужна Азија', 'ml': 'തെക്കേ ഏഷ്യ', 'mn': 'Өмнөд Ази', 'mr': 'दक्षिण आशिया', 'ms': 'Asia Selatan', 'mt': 'Asja t’Isfel Ċentrali', 'my': 'တောင်အာရှ', 'mzn': 'جنوبی آسیا', 'nb': 'Sør-Asia', 'ne': 'दक्षिणी एशिया', 'nl': 'Zuid-Azië', 'nn': 'Sør-Asia', 'no': 'Sør-Asia', 'nqo': 'ߊߖ߭ߌ߫ ߥߙߏ߬ߘߎ߮', 'om': 'Eeshiyaa Kibbaa', 'or': 'ଦକ୍ଷିଣ ଏସିଆ', 'pa': 'ਦੱਖਣੀ ਏਸ਼ੀਆ', 'pa-Guru': 'ਦੱਖਣੀ ਏਸ਼ੀਆ', 'pcm': 'Saútán Éshia', 'pl': 'Azja Południowa', 'ps': 'سويلي آسيا', 'pt': 'Ásia Meridional', 'qu': 'Asia del Sur', 'rm': 'Asia dal Sid', 'ro': 'Asia Meridională', 'ru': 'Южная Азия', 'sc': 'Àsia meridionale', 'sd': 'ڏکڻ ايشيا', 'sd-Arab': 'ڏکڻ ايشيا', 'se': 'mátta-Ásia', 'si': 'දකුණු ආසියාව', 'sk': 'južná Ázia', 'sl': 'Južna Azija', 'so': 'Aasiyada Koonfureed', 'sq': 'Azia Jugore', 'sr': 'Јужна Азија', 'sr-Cyrl': 'Јужна Азија', 'sr-Latn': 'Južna Azija', 'sv': 'Sydasien', 'sw': 'Asia ya Kusini', 'syr': 'ܐܣܝܐ ܬܝܡܢܝܬܐ', 'ta': 'தெற்கு ஆசியா', 'te': 'దక్షిణ ఆసియా', 'tg': 'Осиёи Ҷанубӣ', 'th': 'เอเชียใต้', 'ti': 'ደቡባዊ ኤስያ', 'tk': 'Günorta Aziýa', 'to': 'ʻĒsia fakatonga', 'tr': 'Güney Asya', 'tt': 'Көньяк Азия', 'ug': 'جەنۇبىي ئاسىيا', 'uk': 'Південна Азія', 'ur': 'جنوبی ایشیا', 'uz': 'Janubiy Osiyo', 'uz-Cyrl': 'Жанубий Осиё', 'uz-Latn': 'Janubiy Osiyo', 'vec': 'Azia meridionale', 'vi': 'Nam Á', 'wae': 'Südasie', 'wo': 'Asie du Sud', 'xh': 'UMzantsi Asia', 'xnr': 'दखणे बखें एशिया', 'yi': 'דרום־אַזיע', 'yo': 'Gúúṣù Eṣíà', 'yrl': 'Ásiya Meridiyunau', 'yue': '南亞', 'yue-Hans': '南亚', 'yue-Hant': '南亞', 'zh': '南亚', 'zh-Hans': '南亚', 'zh-Hant': '南亞', 'zu': 'i-Southern Asia'},
    '035': {'af': 'Suidoos-Asië', 'ak': 'Asia Anaafoɔ Apuieɛ', 'am': 'ደቡብ ምስራቅ እስያ', 'ar': 'جنوب شرق آسيا', 'as': 'দক্ষিণ-পূব এছিয়া', 'ast': 'Sureste Asiáticu', 'az': 'Cənub-Şərqi Asiya', 'az-Cyrl': 'Ҹәнуб-Шәрги Асија', 'az-Latn': 'Cənub-Şərqi Asiya', 'bal-Latn': 'Zerbárródarátki Ásiá', 'be': 'Паўднёва-Усходняя Азія', 'be-tarask': 'Паўднёва-Усходняя Азія', 'bg': 'Югоизточна Азия', 'blo': 'Gacǝlǝŋtǝna gʊsono na gajakayɛlaŋ kʊfɔɔ nɩ', 'bn': 'দক্ষিণ পূর্ব এশিয়া', 'br': 'Azia ar Gevred', 'brx': 'खोला-सानजा एसिया', 'bs': 'Jugoistočna Azija', 'bs-Cyrl': 'Југоисточна Азија', 'bs-Latn': 'Jugoistočna Azija', 'ca': 'Àsia sud-oriental', 'ca-ES-valencia': 'Àsia sud-oriental', 'ccp': '𑄘𑄧𑄉𑄨𑄚𑄴 𑄛𑄪𑄇𑄴 𑄃𑄬𑄥𑄨𑄠', 'ce': 'Къилба-малхбален Ази', 'ceb': 'Southeast Asia', 'chr': 'ᏧᎦᎾᏮ ᏗᎧᎸᎬ ᏓᎶᏂᎨᏍᏛ', 'ckb': 'باشووری ڕۆژھەڵاتی ئاسیا', 'cs': 'jihovýchodní Asie', 'cv': 'Кӑнтӑр хӗвелтухӑҫ Ази', 'cy': 'De-Ddwyrain Asia', 'da': 'Sydøstasien', 'de': 'Südostasien', 'dsb': 'krotkozajtšna Azija', 'dz': 'ལྷོ་ཤར་ཕྱོགས་ཀྱི་ཨེ་ཤི་ཡ', 'ee': 'Anyiehe Ɣedzeƒe Afrika nutome', 'el': 'Νοτιοανατολική Ασία', 'el-polyton': 'Νοτιοανατολική Ασία', 'en': 'Southeast Asia', 'en-Dsrt': '𐐝𐐵𐑃-𐐀𐑅𐐻𐐲𐑉𐑌 𐐁𐑈𐐲', 'es': 'Sudeste asiático', 'et': 'Kagu-Aasia', 'eu': 'Hego-ekialdeko Asia', 'fa': 'جنوب شرق آسیا', 'ff-Adlm': '𞤙𞤢𞤥𞤬𞤭𞤯𞤲𞤢𞥄𞤲𞤺𞤫 𞤀𞥄𞤧𞤭𞤴𞤢', 'fi': 'Kaakkois-Aasia', 'fil': 'Timog-Silangang Asya', 'fo': 'Útsynningsasia', 'fr': 'Asie du Sud-Est', 'fur': 'Asie sud orientâl', 'fy': 'Sûdoost-Azië', 'ga': 'an Áise Thoir Theas', 'gd': 'Àisia an Ear-dheas', 'gl': 'Sueste Asiático', 'gsw': 'Süüdoschtaasie', 'gu': 'દક્ષિણપૂર્વ એશિયા', 'ha': 'Kudu Maso Gabashin Asiya', 'he': 'דרום־מזרח אסיה', 'hi': 'दक्षिण-पूर्व एशिया', 'hi-Latn': 'Southeast Asia', 'hr': 'Jugoistočna Azija', 'hsb': 'juhowuchodna Azija', 'ht': 'Asie du Sud-Est', 'hu': 'Délkelet-Ázsia', 'hy': 'Հարավարևելյան Ասիա', 'ia': 'Asia del sud-est', 'id': 'Asia Tenggara', 'ig': 'Mpaghara Mgbada Ugwu Asia dị na Ọwụwa Anyanwụ', 'is': 'Suðaustur-Asía', 'it': 'Sud-est asiatico', 'ja': '東南アジア', 'jv': 'Asia Kidul-wétan', 'ka': 'სამხრეთ-აღმოსავლეთ აზია', 'kaa': 'Қубла-шығыс Азия', 'kaa-Cyrl': 'Қубла-шығыс Азия', 'kea': 'Sudesti Aziátiku', 'kgp': 'Ajia Sur-Rãjur', 'kk': 'Оңтүстік-Шығыс Азия', 'kk-Cyrl': 'Оңтүстік-Шығыс Азия', 'km': 'អាស៊ីអាគ្នេយ៍', 'kn': 'ಆಗ್ನೇಯ ಏಷ್ಯಾ', 'ko': '동남아시아', 'kok': 'आग्नेय आशिया', 'kok-Deva': 'आग्नेय आशिया', 'ks': 'جنوٗبہِ مَشرِقی ایشیا', 'ks-Arab': 'جنوٗبہِ مَشرِقی ایشیا', 'ksh': 'Söhd-Oß-Aasije', 'ku': 'Başûrrojhilatê Asyayê', 'ky': 'Түштүк-Чыгыш Азия', 'lb': 'Südostasien', 'lld': 'Asia dl Süd-Ost', 'lo': 'ອາຊີຕາເວັນອອກສຽງໃຕ້', 'lt': 'Pietryčių Azija', 'lv': 'Centrālaustrumāzija', 'mai': 'दक्षिण-पूरबी एशिया', 'mi': 'Āhia ki te Tonga-mā-rāwhiti', 'mk': 'Југоисточна Азија', 'ml': 'തെക്ക്-കിഴക്കൻ ഏഷ്യ', 'mn': 'Зүүн өмнөд Ази', 'mr': 'दक्षिण पूर्व आशिया', 'ms': 'Asia Tenggara', 'mt': 'Asja tax-Xlokk', 'my': 'အရှေ့တောင်အာရှ', 'mzn': 'آسیای ِجنوب\u200cشرقی\u200cوَر', 'nb': 'Sørøst-Asia', 'ne': 'दक्षिण पूर्वी एशिया', 'nl': 'Zuidoost-Azië', 'nn': 'Søraust-Asia', 'no': 'Sørøst-Asia', 'nqo': 'ߊߖ߭ߌ߫ ߥߙߏ߬ߘߎ߮-ߕߟߋ߬ߓߐ', 'om': 'Kibba baha Eeshiyaa', 'or': 'ଦକ୍ଷିଣପୂର୍ବ ଏସିଆ', 'pa': 'ਦੱਖਣ-ਪੂਰਬੀ ਏਸ਼ੀਆ', 'pa-Guru': 'ਦੱਖਣ-ਪੂਰਬੀ ਏਸ਼ੀਆ', 'pcm': 'Saútíst Éshiá', 'pl': 'Azja Południowo-Wschodnia', 'ps': 'سويلي ختيځ آسيا', 'pt': 'Sudeste Asiático', 'qu': 'Sureste de Asia', 'rm': 'Asia dal Sidost', 'ro': 'Asia de Sud-Est', 'ru': 'Юго-Восточная Азия', 'sc': 'Sud-est asiàticu', 'sd': 'ڏکڻ اوڀر ايشيا', 'sd-Arab': 'ڏکڻ اوڀر ايشيا', 'se': 'mátta-nuorta-Ásia', 'si': 'අග්නිදිග ආසියාව', 'sk': 'juhovýchodná Ázia', 'sl': 'Jugovzhodna Azija', 'so': 'Aasiyada Koonfur-galbeed', 'sq': 'Azia Juglindore', 'sr': 'Југоисточна Азија', 'sr-Cyrl': 'Југоисточна Азија', 'sr-Latn': 'Jugoistočna Azija', 'sv': 'Sydostasien', 'sw': 'Asia ya Kusini Mashariki', 'syr': 'ܬܝܡܢ ܡܕܢܚ ܐܣܝܐ', 'ta': 'தென்கிழக்கு ஆசியா', 'te': 'ఆగ్నేయ ఆసియా', 'tg': 'Осиёи Ҷанубу Шарқӣ', 'th': 'เอเชียตะวันออกเฉียงใต้', 'ti': 'ደቡባዊ ምብራቕ ኤስያ', 'tk': 'Günorta-gündogar Aziýa', 'to': 'ʻĒsia fakatongahahake', 'tr': 'Güneydoğu Asya', 'tt': 'Көньяк-Көнчыгыш Азия', 'ug': 'شەرقىي جەنۇبىي ئاسىيا', 'uk': 'Південно-Східна Азія', 'ur': 'جنوب مشرقی ایشیا', 'uz': 'Janubi-sharqiy Osiyo', 'uz-Cyrl': 'Жанубий-Шарқий Осиё', 'uz-Latn': 'Janubi-sharqiy Osiyo', 'vec': 'Azia de’l sudest', 'vi': 'Đông Nam Á', 'wae': 'Südoštasie', 'wo': 'Asie Sud-est', 'xh': 'UMzantsi-mpuma Asia', 'xnr': 'दखण-पूर्व एशिया', 'yi': 'דרום־מזרח אַזיע', 'yo': 'Gúúṣù ìlà òòrùn Éṣíà', 'yrl': 'Sudeti Ásiyatiku', 'yue': '東南亞', 'yue-Hans': '东南亚', 'yue-Hant': '東南亞', 'zh': '东南亚', 'zh-Hans': '东南亚', 'zh-Hant': '東南亞', 'zu': 'i-South-Eastern Asia'},
    '039': {'af': 'Suid-Europa', 'ak': 'Yuropu Anaafoɔ', 'am': 'ደቡባዊ አውሮፓ', 'ar': 'جنوب أوروبا', 'as': 'দাক্ষিণাত্য ইউৰোপ', 'ast': 'Europa del Sur', 'az': 'Cənubi Avropa', 'az-Cyrl': 'Ҹәнуби Авропа', 'az-Latn': 'Cənubi Avropa', 'bal-Latn': 'Zerbári Yurop', 'be': 'Паўднёвая Еўропа', 'be-tarask': 'Паўднёвая Еўропа', 'bg': 'Южна Европа', 'blo': 'Garɩfɔntǝna gʊnyɩsonolaŋ', 'bn': 'দক্ষিণ ইউরোপ', 'br': 'Europa ar Su', 'brx': 'खोलायारि यूरोप', 'bs': 'Južna Evropa', 'bs-Cyrl': 'Јужна Европа', 'bs-Latn': 'Južna Evropa', 'ca': 'Europa meridional', 'ca-ES-valencia': 'Europa meridional', 'ccp': '𑄘𑄧𑄉𑄨𑄚𑄴 𑄄𑄃𑄪𑄢𑄮𑄛𑄴', 'ce': 'Къилба Европа', 'ceb': 'Southern Europe', 'chr': 'ᏧᎦᎾᏮ ᏗᏜ ᏳᎳᏛ', 'ckb': 'باشووری ئەورووپا', 'cs': 'jižní Evropa', 'cv': 'Кӑнтӑр Европа', 'cy': 'De Ewrop', 'da': 'Sydeuropa', 'de': 'Südeuropa', 'dsb': 'pódpołdnjowa Europa', 'dz': 'ལྷོའི་ཡུ་རོབ', 'ee': 'Anyiehelɔƒo Europa nutome', 'el': 'Νότια Ευρώπη', 'el-polyton': 'Νότια Ευρώπη', 'en': 'Southern Europe', 'en-Dsrt': '𐐝𐐲𐑄𐐲𐑉𐑌 𐐏𐐲𐑉𐐲𐐹', 'es': 'Europa meridional', 'et': 'Lõuna-Euroopa', 'eu': 'Hego Europa', 'fa': 'جنوب اروپا', 'ff-Adlm': '𞤙𞤢𞤥𞤲𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭 𞤓𞤪𞤨𞤢', 'fi': 'Etelä-Eurooppa', 'fil': 'Katimugang Europe', 'fo': 'Suðurevropa', 'fr': 'Europe du Sud', 'fur': 'Europe meridionâl', 'fy': 'Sûd-Europa', 'ga': 'Deisceart na hEorpa', 'gd': 'An Roinn-Eòrpa a Deas', 'gl': 'Europa Meridional', 'gsw': 'Süüdeuropa', 'gu': 'દક્ષિણ યુરોપ', 'ha': 'Kudancin Turai', 'he': 'דרום אירופה', 'hi': 'दक्षिणी यूरोप', 'hi-Latn': 'Southern Europe', 'hr': 'Južna Europa', 'hsb': 'južna Europa', 'ht': 'Europe du Sud', 'hu': 'Dél-Európa', 'hy': 'Հարավային Եվրոպա', 'ia': 'Europa meridional', 'id': 'Eropa Bagian Selatan', 'ig': 'Mpaghara Mgbada Ugwu Europe', 'is': 'Suður-Evrópa', 'it': 'Europa meridionale', 'ja': '南ヨーロッパ', 'jv': 'Éropah Kidul', 'ka': 'სამხრეთ ევროპა', 'kaa': 'Қубла Европа', 'kaa-Cyrl': 'Қубла Европа', 'kea': 'Europa di Sul', 'kgp': 'Orópa Mẽrinhonỹr', 'kk': 'Оңтүстік Еуропа', 'kk-Cyrl': 'Оңтүстік Еуропа', 'km': 'អឺរ៉ុប\u200bខាង\u200bត្បូង', 'kn': 'ದಕ್ಷಿಣ ಯೂರೋಪ್', 'ko': '남유럽', 'kok': 'दक्षिण युरोप', 'kok-Deva': 'दक्षिण युरोप', 'ks': 'جنوٗبی یوٗرَپ', 'ks-Arab': 'جنوٗبی یوٗرَپ', 'ksh': 'Söhd-Europpa', 'ku': 'Başûrê Ewropayê', 'ky': 'Түштүк Европа', 'lb': 'Südeuropa', 'lld': 'Europa dl Süd', 'lo': 'ຢູໂຣບໃຕ້', 'lt': 'Pietų Europa', 'lv': 'Dienvideiropa', 'mai': 'दक्षिणी यूरोप', 'mi': 'Ūropi ki te Tonga', 'mk': 'Јужна Европа', 'ml': 'തെക്കേ യൂറോപ്പ്', 'mn': 'Өмнөд Европ', 'mr': 'दक्षिण युरोप', 'ms': 'Eropah Selatan', 'mt': 'Ewropa t’Isfel', 'my': 'တောင်ဥရောပ', 'mzn': 'جنوبی اروپا', 'nb': 'Sør-Europa', 'ne': 'दक्षिणी युरोप', 'nl': 'Zuid-Europa', 'nn': 'Sør-Europa', 'no': 'Sør-Europa', 'nqo': 'ߋߙߐߔߎ߬ ߥߙߏ߬ߘߎ߮', 'om': 'Awurooppaa Kibbaa', 'or': 'ଦକ୍ଷିଣ ୟୁରୋପ୍', 'pa': 'ਦੱਖਣੀ ਯੂਰਪ', 'pa-Guru': 'ਦੱਖਣੀ ਯੂਰਪ', 'pcm': 'Saútán Yúrop', 'pl': 'Europa Południowa', 'ps': 'سويلي اروپا', 'pt': 'Europa Meridional', 'qu': 'Europa del Sur', 'rm': 'Europa dal Sid', 'ro': 'Europa Meridională', 'ru': 'Южная Европа', 'sc': 'Europa meridionale', 'sd': 'ڏکڻ يورپ', 'sd-Arab': 'ڏکڻ يورپ', 'se': 'mátta-Eurohpá', 'si': 'දකුණුදිග යුරෝපය', 'sk': 'južná Európa', 'sl': 'Južna Evropa', 'so': 'Yurubta Koonfureed', 'sq': 'Evropa Jugore', 'sr': 'Јужна Европа', 'sr-Cyrl': 'Јужна Европа', 'sr-Latn': 'Južna Evropa', 'sv': 'Sydeuropa', 'sw': 'Ulaya ya Kusini', 'syr': 'ܐܘܪܘܦܐ ܬܝܡܢܝܬܐ', 'ta': 'தெற்கு ஐரோப்பா', 'te': 'దక్షిణ యూరోప్', 'tg': 'Аврупои Ҷанубӣ', 'th': 'ยุโรปใต้', 'ti': 'ደቡባዊ ኤውሮጳ', 'tk': 'Günorta Ýewropa', 'to': 'ʻEulope fakatonga', 'tr': 'Güney Avrupa', 'tt': 'Көньяк Европа', 'ug': 'جەنۇبىي ياۋروپا', 'uk': 'Південна Європа', 'ur': 'جنوبی یورپ', 'uz': 'Janubiy Yevropa', 'uz-Cyrl': 'Жанубий Европа', 'uz-Latn': 'Janubiy Yevropa', 'vec': 'Europa meridionale', 'vi': 'Nam Âu', 'wae': 'Südeuropa', 'wo': 'Sud Europe', 'xh': 'UMzantsi Yurophu', 'xnr': 'दखणे बखें यूरोप', 'yi': 'דרום־אייראפּע', 'yo': 'Gúúṣù Yúróòpù', 'yrl': 'Eurupa Meridiyunau', 'yue': '南歐', 'yue-Hans': '南欧', 'yue-Hant': '南歐', 'zh': '南欧', 'zh-Hans': '南欧', 'zh-Hant': '南歐', 'zu': 'i-Southern Europe'},
    '053': {'af': 'Australasië', 'ak': 'Ɔstrelia ne Asia', 'am': 'አውስትራሌዥያ', 'ar': 'أسترالاسيا', 'as': 'অষ্ট্ৰেলেছিয়া', 'ast': 'Australasia', 'az': 'Avstralaziya', 'az-Cyrl': 'Австралазија', 'az-Latn': 'Avstralaziya', 'bal-Latn': 'Ástrálásiá', 'be': 'Аўстралазія', 'be-tarask': 'Аўстралазія', 'bg': 'Австралазия', 'bgn': 'استرالیا', 'blo': 'Ɔstracǝlǝŋtǝna', 'bn': 'অস্ট্রেলেশিয়া', 'br': 'Aostralazia', 'brx': 'अस्ट्रेलएसिया', 'bs': 'Australazija', 'bs-Cyrl': 'Аустралија и Нови Зеланд', 'bs-Latn': 'Australazija', 'ca': 'Australàsia', 'ca-ES-valencia': 'Australàsia', 'ccp': '𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄣𑄬𑄥𑄨𑄠', 'ce': 'Австралази', 'ceb': 'Australasia', 'chr': 'ᎠᏍᏔᎴᏏᎠ', 'ckb': 'ئۆسترالیا', 'cs': 'Australasie', 'cv': 'Австралази', 'cy': 'Awstralasia', 'da': 'Australasien', 'de': 'Australasien', 'dsb': 'Awstralazija', 'dz': 'ཨཱོས་ཊྲེལ་ཨེ་ཤི་ཡ', 'ee': 'Australia kple New Zealand nutome', 'el': 'Αυστραλασία', 'el-polyton': 'Αυστραλασία', 'en': 'Australasia', 'en-Dsrt': '𐐉𐑅𐐻𐑉𐐩𐑊𐐨𐐲 𐐰𐑌𐐼 𐐤𐐭 𐐞𐐨𐑊𐐲𐑌𐐼', 'es': 'Australasia', 'et': 'Australaasia', 'eu': 'Australasia', 'fa': 'استرالزی', 'ff-Adlm': '𞤀𞤧𞤼𞤢𞤪𞤤𞤢𞥄𞤧𞤭𞤴𞤢𞥄', 'fi': 'Australaasia', 'fil': 'Australasia', 'fo': 'Avstralasia', 'fr': 'Australasie', 'fur': 'Australie e Gnove Zelande', 'fy': 'Australazië', 'ga': 'an Astraláise', 'gd': 'Astràilia is Sealainn Nuadh', 'gl': 'Australasia', 'gsw': 'Auschtraalie und Nöiseeland', 'gu': 'ઓસ્ટ્રેલેશિયા', 'ha': 'Asturelasiya', 'he': 'אוסטרלאסיה', 'hi': 'ऑस्ट्रेलेशिया', 'hi-Latn': 'Australasia', 'hr': 'Australazija', 'hsb': 'Awstralazija', 'ht': 'Australasie', 'hu': 'Ausztrálázsia', 'hy': 'Ավստրալասիա', 'ia': 'Australasia', 'id': 'Australasia', 'ig': 'Australasia', 'is': 'Ástralasía', 'it': 'Australasia', 'ja': 'オーストララシア', 'jv': 'Australasia', 'ka': 'ავსტრალაზია', 'kaa': 'Австралазия', 'kaa-Cyrl': 'Австралазия', 'kea': 'Australázia', 'kgp': 'Ausitrarajia', 'kk': 'Аустралазия', 'kk-Cyrl': 'Аустралазия', 'km': 'អូស្ត្រាឡាស៊ី', 'kn': 'ಆಸ್ಟ್ರೇಲೇಷ್ಯಾ', 'ko': '오스트랄라시아', 'kok': 'ऑस्ट्रेलेसिया', 'kok-Deva': 'ऑस्ट्रेलेसिया', 'ks': 'آسٹریلیشیا', 'ks-Arab': 'آسٹریلیشیا', 'ksh': 'de Rejjohn öm Austrahlije', 'ku': 'Awistralasya', 'ky': 'Австралазия', 'lb': 'Australien an Neiséiland', 'lld': 'Australasia', 'lo': 'ໂອດສະຕາລີ', 'lt': 'Australazija', 'lv': 'Austrālāzija', 'mai': 'आस्ट्रेलेशिया', 'mi': 'Atareiria', 'mk': 'Австралазија', 'ml': 'ഓസ്\u200cട്രേലിയയും ന്യൂസിലാൻഡും', 'mn': 'Австрали-Ази', 'mr': 'ऑस्\u200dट्रेलेशिया', 'ms': 'Australasia', 'mt': 'Awstralja u New Zealand', 'my': 'ဩစတြေးလျနှင့် နယူးဇီလန်', 'mzn': 'اوسترالزی', 'nb': 'Australasia', 'ne': 'अष्ट्रालासिया', 'nl': 'Australazië', 'nn': 'Australasia', 'no': 'Australasia', 'nqo': 'ߐߛߑߕߙߊߟߊߖ߭ߌ߫', 'om': 'Awustiralashiyaa', 'or': 'ଅଷ୍ଟ୍ରେଲେସିଆ', 'pa': 'ਆਸਟਰੇਲੇਸ਼ੀਆ', 'pa-Guru': 'ਆਸਟਰੇਲੇਸ਼ੀਆ', 'pcm': 'Ọstraléshia', 'pl': 'Australazja', 'ps': 'آسترالیا', 'pt': 'Australásia', 'qu': 'Australasia', 'rm': 'Australia e Nova Zelanda', 'ro': 'Australasia', 'ru': 'Австралазия', 'sc': 'Australàsia', 'sd': 'آسٽریلیشیا', 'sd-Arab': 'آسٽریلیشیا', 'se': 'Austrália ja Ođđa-Selánda', 'si': 'ඕස්ට්\u200dරලේෂියාව', 'sk': 'Australázia', 'sl': 'Avstralija in Nova Zelandija', 'so': 'Austraalaasiya', 'sq': 'Australazia', 'sr': 'Аустралија и Нови Зеланд', 'sr-Cyrl': 'Аустралија и Нови Зеланд', 'sr-Latn': 'Australija i Novi Zeland', 'sv': 'Australasien', 'sw': 'Australasia', 'syr': 'ܐܘܣܛܪܐܠܐܣܝܐ', 'ta': 'ஆஸ்திரலேசியா', 'te': 'ఆస్ట్రేలేసియా', 'tg': 'Австралияву Осиё', 'th': 'ออสตราเลเซีย', 'ti': 'ኣውስትራሌዥያ', 'tk': 'Awstralaziýa', 'to': 'ʻAositelēlēsia', 'tr': 'Avustralasya', 'tt': 'Австралазия', 'ug': 'ئاۋسترالئاسىيا', 'uk': 'Австралазія', 'ur': 'آسٹریلیشیا', 'uz': 'Avstralaziya', 'uz-Cyrl': 'Австралазия', 'uz-Latn': 'Avstralaziya', 'vec': 'Australazia', 'vi': 'Australasia', 'wae': 'Auštralie und Niwséland', 'wo': 'Ostralasi', 'xh': 'I-Australasia', 'xnr': 'ऑस्ट्रेलेशिया', 'yo': 'Ọṣirélaṣíà', 'yrl': 'Ausitarasia', 'yue': '澳洲同紐西蘭', 'yue-Hans': '澳洲同纽西兰', 'yue-Hant': '澳洲同紐西蘭', 'zh': '澳大拉西亚', 'zh-Hans': '澳大拉西亚', 'zh-Hant': '澳洲與紐西蘭', 'zu': 'i-Australasia'},
    '054': {'af': 'Melanesië', 'ak': 'Melanesia', 'am': 'ሜላኔዥያ', 'ar': 'ميلانيزيا', 'as': 'মেলানেছিয়া', 'ast': 'Melanesia', 'az': 'Melaneziya', 'az-Cyrl': 'Меланезија', 'az-Latn': 'Melaneziya', 'bal-Latn': 'Melanésiá', 'be': 'Меланезія', 'be-tarask': 'Меланезія', 'bg': 'Меланезия', 'blo': 'Melanesiya', 'bn': 'মেলানেশিয়া', 'br': 'Melanezia', 'brx': 'मेलानेसिया', 'bs': 'Melanezija', 'bs-Cyrl': 'Меланезија', 'bs-Latn': 'Melanezija', 'ca': 'Melanèsia', 'ca-ES-valencia': 'Melanèsia', 'ccp': '𑄟𑄳𑄠𑄣𑄬𑄚𑄬𑄥𑄨𑄠', 'ce': 'Меланези', 'ceb': 'Melanesia', 'chr': 'ᎺᎳᏁᏏᎠ', 'ckb': 'میلانێزیا', 'cs': 'Melanésie', 'cv': 'Меланези', 'cy': 'Melanesia', 'da': 'Melanesien', 'de': 'Melanesien', 'dsb': 'Melaneziska', 'dz': 'མེ་ལ་ནི་ཤི་ཡ', 'ee': 'Melanesia nutome', 'el': 'Μελανησία', 'el-polyton': 'Μελανησία', 'en': 'Melanesia', 'en-Dsrt': '𐐣𐐯𐑊𐐲𐑌𐐨𐑈𐐲', 'es': 'Melanesia', 'et': 'Melaneesia', 'eu': 'Melanesia', 'fa': 'ملانزی', 'ff-Adlm': '𞤃𞤭𞤤𞤢𞤲𞤭𞥅𞤧𞤴𞤢', 'fi': 'Melanesia', 'fil': 'Melanesia', 'fo': 'Melanesia', 'fr': 'Mélanésie', 'fur': 'Melanesie', 'fy': 'Melanesië', 'ga': 'an Mheilinéis', 'gd': 'Na h-Eileanan Dubha', 'gl': 'Melanesia', 'gsw': 'Melaneesie', 'gu': 'મેલાનેશિયા', 'ha': 'Melanesia', 'he': 'מלנזיה', 'hi': 'मेलानेशिया', 'hi-Latn': 'Melanesia', 'hr': 'Melanezija', 'hsb': 'Melaneziska', 'ht': 'Mélanésie', 'hu': 'Melanézia', 'hy': 'Մելանեզիա', 'ia': 'Melanesia', 'id': 'Melanesia', 'ig': 'Melanesia', 'is': 'Melanesía', 'it': 'Melanesia', 'ja': 'メラネシア', 'jv': 'Melanesia', 'ka': 'მელანეზია', 'kaa': 'Меланезия', 'kaa-Cyrl': 'Меланезия', 'kea': 'Melanézia', 'kgp': 'Mẽranẽjia', 'kk': 'Меланезия', 'kk-Cyrl': 'Меланезия', 'km': 'មេឡាណេស៊ី', 'kn': 'ಮೆಲನೇಷಿಯಾ', 'ko': '멜라네시아', 'kok': 'मेलानेसिया', 'kok-Deva': 'मेलानेसिया', 'ks': 'مؠلَنیٖشِیا', 'ks-Arab': 'مؠلَنیٖشِیا', 'ksh': 'Mellanehsije', 'ku': 'Melanezya', 'ky': 'Меланезия', 'lb': 'Melanesien', 'lld': 'Melanesia', 'lo': 'ເມລານີເຊຍ', 'lt': 'Melanezija', 'lv': 'Melanēzija', 'mai': 'मेलानेशिया', 'mi': 'Meranīhia', 'mk': 'Меланезија', 'ml': 'മെലനേഷ്യ', 'mn': 'Меланези', 'mr': 'मेलानेशिया', 'ms': 'Melanesia', 'mt': 'Melanesja', 'my': 'မီလာနီးရှား', 'mzn': 'ملانزی', 'nb': 'Melanesia', 'ne': 'मेलानेसिया', 'nl': 'Melanesië', 'nn': 'Melanesia', 'no': 'Melanesia', 'nqo': 'ߡߋߟߊߣߋߖ߭ߌ߫', 'om': 'Melaaneeshiyaa', 'or': 'ମେଲାନେସିଆ', 'pa': 'ਮੇਲਾਨੇਸ਼ੀਆ', 'pa-Guru': 'ਮੇਲਾਨੇਸ਼ੀਆ', 'pcm': 'Mẹlanẹíshia', 'pl': 'Melanezja', 'ps': 'ملانشیا', 'pt': 'Melanésia', 'qu': 'Melanesia', 'rm': 'Melanesia', 'ro': 'Melanezia', 'ru': 'Меланезия', 'sc': 'Melanèsia', 'sd': 'میلانیشیا', 'sd-Arab': 'میلانیشیا', 'se': 'Melanesia', 'si': 'මෙලනීසියාව', 'sk': 'Melanézia', 'sl': 'Melanezija', 'so': 'Melaneesiya', 'sq': 'Melanezia', 'sr': 'Меланезија', 'sr-Cyrl': 'Меланезија', 'sr-Latn': 'Melanezija', 'sv': 'Melanesien', 'sw': 'Melanesia', 'syr': 'ܡܝܠܐܢܝܣܝܐ', 'ta': 'மெலனேஷியா', 'te': 'మెలనేశియ', 'tg': 'Меланезия', 'th': 'เมลานีเซีย', 'ti': 'መላነዥያ', 'tk': 'Melaneziýa', 'to': 'Melanīsia', 'tr': 'Melanezya', 'tt': 'Меланезия', 'ug': 'مېلانېسىيە', 'uk': 'Меланезія', 'ur': 'مالینیشیا', 'uz': 'Melaneziya', 'uz-Cyrl': 'Меланезия', 'uz-Latn': 'Melaneziya', 'vec': 'Melanezia', 'vi': 'Melanesia', 'wae': 'Melanesie', 'wo': 'Melanesi', 'xh': 'IMelanasia', 'xnr': 'मेलानेशिया', 'yo': 'Mẹlanéṣíà', 'yrl': 'Meranésiya', 'yue': '美拉尼西亞', 'yue-Hans': '美拉尼西亚', 'yue-Hant': '美拉尼西亞', 'zh': '美拉尼西亚', 'zh-Hans': '美拉尼西亚', 'zh-Hant': '美拉尼西亞', 'zu': 'i-Melanesia'},
    '057': {'af': 'Mikronesiese streek', 'ak': 'Micronesia Mantam', 'am': 'የማይክሮኔዥያን ክልል', 'ar': 'الجزر الميكرونيزية', 'as': 'মাইক্ৰোনেচিয় ক্ষেত্ৰ', 'ast': 'Rexón de Micronesia', 'az': 'Mikroneziya Regionu', 'az-Cyrl': 'Микронезија Реҝиону', 'az-Latn': 'Mikroneziya Regionu', 'bal-Latn': 'Máekrónési Damag', 'be': 'Мікранезійскі рэгіён', 'be-tarask': 'Мікранезійскі рэгіён', 'bg': 'Микронезийски регион', 'blo': 'Mikronesiya kagʊsaʊ', 'bn': 'মাইক্রোনেশিয়া অঞ্চল', 'br': 'Rannved Mikronezia', 'brx': 'माईक्र’नेसियायारि ओनसोल', 'bs': 'Mikronezijska regija', 'bs-Cyrl': 'Микронезијски регион', 'bs-Latn': 'Mikronezijska regija', 'ca': 'Regió de la Micronèsia', 'ca-ES-valencia': 'Regió de la Micronèsia', 'ccp': '𑄟𑄭𑄇𑄳𑄢𑄮𑄚𑄬𑄥𑄨𑄠 𑄎𑄉', 'ce': 'Микронези', 'ceb': 'Micronesian Region', 'chr': 'ᎠᏰᏟ ᏧᎾᎵᎪᎯ ᎾᎿ ᎹᎢᏉᏂᏏᏯ ᎢᎬᎾᏕᎾ', 'ckb': 'ناوچەی مایکرۆنیزیا', 'cs': 'Mikronésie (region)', 'cv': 'Микронези регион', 'cy': 'Rhanbarth Micronesia', 'da': 'Mikronesiske område', 'de': 'Mikronesisches Inselgebiet', 'dsb': 'Mikroneziska (kupowy region)', 'dz': 'ལུང་ཕྱོགས་མའི་ཀྲོ་ནི་ཤི་ཡ', 'ee': 'Mikronesia', 'el': 'Περιοχή Μικρονησίας', 'el-polyton': 'Περιοχή Μικρονησίας', 'en': 'Micronesian Region', 'en-Dsrt': '𐐣𐐴𐐿𐑉𐐲𐑌𐐨𐑈𐐲𐑌 𐐡𐐨𐐾𐐲𐑌', 'es': 'Región de Micronesia', 'et': 'Mikroneesia (piirkond)', 'eu': 'Mikronesia eskualdea', 'fa': 'ناحیهٔ میکرونزی', 'ff-Adlm': '𞤖𞤭𞤤𞥆𞤮 𞤃𞤭𞤳𞤪𞤮𞤲𞤭𞥅𞤧𞤸𞤮', 'fi': 'Mikronesian alue', 'fil': 'Rehiyon ng Micronesia', 'fo': 'Mikronesi øki', 'fr': 'région micronésienne', 'fur': 'Regjon de Micronesie', 'fy': 'Micronesyske regio', 'ga': 'an Réigiún Micrinéiseach', 'gd': 'Roinn nam Meanbh-Eileanan', 'gl': 'Rexión de Micronesia', 'gsw': 'Mikroneesischs Inselgebiet', 'gu': 'માઈક્રોનેશિયન ક્ષેત્ર', 'ha': 'Yankin Micronesiya', 'he': 'אזור מיקרונזיה', 'hi': 'माइक्रोनेशियाई क्षेत्र', 'hi-Latn': 'Micronesian Region', 'hr': 'Mikronezijsko područje', 'hsb': 'Mikroneziska (kupowy region)', 'ht': 'région micronésienne', 'hu': 'Mikronéziai régió', 'hy': 'Միկրոնեզյան տարածաշրջան', 'ia': 'Region micronesian', 'id': 'Wilayah Mikronesia', 'ig': 'Mpaghara Micronesian', 'is': 'Míkrónesíusvæðið', 'it': 'Regione micronesiana', 'ja': 'ミクロネシア', 'jv': 'Daerah Mikronesia', 'ka': 'მიკრონეზიის რეგიონი', 'kaa': 'Микронезия', 'kaa-Cyrl': 'Микронезия', 'kea': 'Rejion di Mikronézia', 'kgp': 'Mĩkronẽjia Pénĩn', 'kk': 'Микронезия аймағы', 'kk-Cyrl': 'Микронезия аймағы', 'km': 'តំបន់\u200bមីក្រូណេស៊ី', 'kn': 'ಮೈಕ್ರೋನೇಶಿಯನ್ ಪ್ರದೇಶ', 'ko': '미크로네시아 지역', 'kok': 'मायक्रोनेशियन प्रांत', 'kok-Deva': 'मायक्रोनेशियन प्रांत', 'ks': 'مَیکرونَیشِیَن خٕطہٕ', 'ks-Arab': 'مَیکرونَیشِیَن خٕطہٕ', 'ksh': 'de Rejohn vun Mikronehseje', 'ku': 'Herêma Mîkronezyayê', 'ky': 'Микронезия чөлкөмү', 'lb': 'Mikronesescht Inselgebitt', 'lld': 'Isoles dla Micronesia', 'lo': 'ເຂດໄມໂຄຣເນຊຽນ', 'lt': 'Mikronezijos regionas', 'lv': 'Mikronēzijas reģions', 'mai': 'माइक्रोनेशियाई क्षेत्र', 'mi': 'Te Rohe o Mekanēhia', 'mk': 'Микронезиски регион', 'ml': 'മൈക്രോനേഷ്യൻ പ്രദേശം', 'mn': 'Микронезийн бүс', 'mr': 'मायक्रोनेशियन प्रदेश', 'ms': 'Wilayah Mikronesia', 'mt': 'Reġjun ta’ Mikroneżja', 'my': 'မိုက်ခရိုနီးရှား ဒေသ', 'mzn': 'میکرونزی منقطه', 'nb': 'Mikronesia', 'ne': 'माइक्रोनेसियाली क्षेत्र', 'nl': 'Micronesische regio', 'nn': 'Mikronesia', 'no': 'Mikronesia', 'nqo': 'ߡߌߞߙߏߣߋߖ߭ߌ߫ ߕߌ߲߬ߞߎߘߎ߲', 'om': 'Naannoo Maleeshiyaa', 'or': 'ମାଇକ୍ରୋନେସିଆନ୍ ଅଞ୍ଚଳ', 'pa': 'ਮਾਇਕ੍ਰੋਨੇਸ਼ੀਆਈ ਇਲਾਕਾ', 'pa-Guru': 'ਮਾਇਕ੍ਰੋਨੇਸ਼ੀਆਈ ਇਲਾਕਾ', 'pcm': 'Maikroníshia Ríjọn', 'pl': 'Region Mikronezji', 'ps': 'د مایکرونیسینین سیمه', 'pt': 'Região da Micronésia', 'qu': 'Región Micronesia', 'rm': 'Regiun Micronesica', 'ro': 'Regiunea Micronezia', 'ru': 'Микронезия', 'sc': 'Regione micronesiana', 'sd': 'مائڪرونيشائي خطو', 'sd-Arab': 'مائڪرونيشائي خطو', 'se': 'Mikronesia guovllus', 'si': 'මයික්\u200dරෝනීසියානු කළාපය', 'sk': 'oblasť Mikronézie', 'sl': 'mikronezijska regija', 'so': 'Gobolka Aasiyada yar', 'sq': 'Rajoni Mikronezian', 'sr': 'Микронезијски регион', 'sr-Cyrl': 'Микронезијски регион', 'sr-Latn': 'Mikronezijski region', 'sv': 'Mikronesiska öarna', 'sw': 'Eneo la Mikronesia', 'syr': 'ܡܝܟܪܘܢܝܙܝܐ', 'ta': 'மைக்ரோனேஷியன் பிரதேசம்', 'te': 'మైక్రోనేసియన్ ప్రాంతం', 'tg': 'Минтақаи Микронезия', 'th': 'เขตไมโครนีเซีย', 'ti': 'ዞባ ማይክሮነዥያ', 'tk': 'Mikroneziýa sebti', 'to': 'Potu fonua Mikolonīsia', 'tr': 'Mikronezya Bölgesi', 'tt': 'Микронезия төбәге', 'ug': 'مىكرونېزىيە رايونى', 'uk': 'Мікронезійський регіон', 'ur': 'مائکرونیشیائی علاقہ', 'uz': 'Mikroneziya mintaqasi', 'uz-Cyrl': 'Микронезия минтақаси', 'uz-Latn': 'Mikroneziya mintaqasi', 'vec': 'Rejon microneziana', 'vi': 'Vùng Micronesia', 'wae': 'Mikronesišes Inselgebiet', 'wo': 'Mikronesi', 'xh': 'UMmandla weMicronesia', 'xnr': 'माइक्रोनेशियाई लाका', 'yo': 'Agbègbè Maikironéṣíà', 'yrl': 'Micuronesiya Tetãma', 'yue': '密克羅尼西亞', 'yue-Hans': '密克罗尼西亚', 'yue-Hant': '密克羅尼西亞', 'zh': '密克罗尼西亚地区', 'zh-Hans': '密克罗尼西亚地区', 'zh-Hant': '密克羅尼西亞群島', 'zu': 'i-Micronesian Region'},
    '061': {'af': 'Polinesië', 'ak': 'Pɔlenesia', 'am': 'ፖሊኔዥያ', 'ar': 'بولينيزيا', 'as': 'পলিনেচিয়া', 'ast': 'Polinesia', 'az': 'Polineziya', 'az-Cyrl': 'Полинезија', 'az-Latn': 'Polineziya', 'bal-Latn': 'Pólinisiá', 'be': 'Палінезія', 'be-tarask': 'Палінезія', 'bg': 'Полинезия', 'blo': 'Polinesiya', 'bn': 'পলিনেশিয়া', 'br': 'Polinezia', 'brx': 'पलिनेसिया', 'bs': 'Polinezija', 'bs-Cyrl': 'Полинезија', 'bs-Latn': 'Polinezija', 'ca': 'Polinèsia', 'ca-ES-valencia': 'Polinèsia', 'ccp': '𑄛𑄧𑄣𑄨𑄚𑄬𑄥𑄨𑄠', 'ce': 'Полинези', 'ceb': 'Polynesia', 'chr': 'ᏆᎵᏂᏏᎠ', 'ckb': 'پۆلینیزیا', 'cs': 'Polynésie', 'cv': 'Полинези', 'cy': 'Polynesia', 'da': 'Polynesien', 'de': 'Polynesien', 'dsb': 'Polyneziska', 'dz': 'པོ་ལི་ནི་ཤི་ཡ', 'ee': 'Pɔlinesia nutome', 'el': 'Πολυνησία', 'el-polyton': 'Πολυνησία', 'en': 'Polynesia', 'en-Dsrt': '𐐑𐐪𐑊𐐲𐑌𐐨𐑈𐐲', 'es': 'Polinesia', 'et': 'Polüneesia', 'eu': 'Polinesia', 'fa': 'پلی\u200cنزی', 'ff-Adlm': '𞤆𞤮𞤤𞤭𞤲𞤭𞥅𞤧𞤴𞤢', 'fi': 'Polynesia', 'fil': 'Polynesia', 'fo': 'Polynesia', 'fr': 'Polynésie', 'fur': 'Polinesie', 'fy': 'Polynesië', 'ga': 'an Pholainéis', 'gd': 'Poilinèis', 'gl': 'Polinesia', 'gsw': 'Polineesie', 'gu': 'પોલિનેશિયા', 'ha': 'Kasar Polynesia', 'he': 'פולינזיה', 'hi': 'पोलिनेशिया', 'hi-Latn': 'Polynesia', 'hr': 'Polinezija', 'hsb': 'Polyneziska', 'ht': 'Polynésie', 'hu': 'Polinézia', 'hy': 'Պոլինեզիա', 'ia': 'Polynesia', 'id': 'Polinesia', 'ig': 'Polynesia', 'is': 'Pólýnesía', 'it': 'Polinesia', 'ja': 'ポリネシア', 'jv': 'Polinesia', 'ka': 'პოლინეზია', 'kaa': 'Полинезия', 'kaa-Cyrl': 'Полинезия', 'kea': 'Polinézia', 'kgp': 'Porinẽjia', 'kk': 'Полинезия', 'kk-Cyrl': 'Полинезия', 'km': 'ប៉ូលីណេស៊ី', 'kn': 'ಪಾಲಿನೇಷ್ಯಾ', 'ko': '폴리네시아', 'kok': 'पोलिनेशिया', 'kok-Deva': 'पोलिनेशिया', 'ks': 'پالنیشِیا', 'ks-Arab': 'پالنیشِیا', 'ksh': 'Polinehsije', 'ku': 'Polînezya', 'ky': 'Полинезия', 'lb': 'Polynesien', 'lld': 'Polinesia', 'lo': 'ໂພລີນີເຊຍ', 'lt': 'Polinezija', 'lv': 'Polinēzija', 'mai': 'पोलिनेशिया', 'mi': 'Te Moana-nui-a-Kiwa', 'mk': 'Полинезија', 'ml': 'പോളിനേഷ്യ', 'mn': 'Полинези', 'mr': 'पॉलिनेशिया', 'ms': 'Polinesia', 'mt': 'Polinesja', 'my': 'ပိုလီနီရှား', 'mzn': 'پولی\u200cنزی', 'nb': 'Polynesia', 'ne': 'पोलिनेशिया', 'nl': 'Polynesië', 'nn': 'Polynesia', 'no': 'Polynesia', 'nqo': 'ߔߏߟߣߋߖ߭ߌ߫', 'om': 'Poolineeshiyaa', 'or': 'ପଲିନେସିଆ', 'pa': 'ਪੋਲੀਨੇਸ਼ੀਆ', 'pa-Guru': 'ਪੋਲੀਨੇਸ਼ੀਆ', 'pcm': 'Poliníshiá', 'pl': 'Polinezja', 'ps': 'پولنيسيا', 'pt': 'Polinésia', 'qu': 'Polinesia', 'rm': 'Polinesia', 'ro': 'Polinezia', 'ru': 'Полинезия', 'sc': 'Polinèsia', 'sd': 'پولینیشیا', 'sd-Arab': 'پولینیشیا', 'se': 'Polynesia', 'si': 'පොලිනීසියාව', 'sk': 'Polynézia', 'sl': 'Polinezija', 'so': 'Booliyneesiya', 'sq': 'Polinezia', 'sr': 'Полинезија', 'sr-Cyrl': 'Полинезија', 'sr-Latn': 'Polinezija', 'sv': 'Polynesien', 'sw': 'Polynesia', 'syr': 'ܦܘܠܢܝܣܝܐ', 'ta': 'பாலினேஷியா', 'te': 'పాలినేషియా', 'tg': 'Полинезия', 'th': 'โปลินีเซีย', 'ti': 'ፖሊነዥያ', 'tk': 'Polineziýa', 'to': 'Polinīsia', 'tr': 'Polinezya', 'tt': 'Полинезия', 'ug': 'پولىنىزىيە', 'uk': 'Полінезія', 'ur': 'پولینیشیا', 'uz': 'Polineziya', 'uz-Cyrl': 'Полинезия', 'uz-Latn': 'Polineziya', 'vec': 'Polinezia', 'vi': 'Polynesia', 'wae': 'Polinesie', 'wo': 'Polineesi', 'xh': 'I-Polynesia', 'xnr': 'पोलिनेशिया', 'yi': 'פּאלינעזיע', 'yo': 'Polineṣíà', 'yrl': 'Pulinesiya', 'yue': '玻里尼西亞', 'yue-Hans': '玻里尼西亚', 'yue-Hant': '玻里尼西亞', 'zh': '玻利尼西亚', 'zh-Hans': '玻利尼西亚', 'zh-Hant': '玻里尼西亞', 'zu': 'i-Polynesia'},
    '142': {'af': 'Asië', 'ak': 'Asia', 'am': 'እስያ', 'ar': 'آسيا', 'as': 'এছিয়া', 'ast': 'Asia', 'az': 'Asiya', 'az-Cyrl': 'Асија', 'az-Latn': 'Asiya', 'bal-Latn': 'Ásiá', 'be': 'Азія', 'be-tarask': 'Азія', 'bg': 'Азия', 'bgn': 'اسیا', 'blo': 'Gacǝlǝŋtǝna', 'bn': 'এশিয়া', 'br': 'Azia', 'brx': 'एसिया', 'bs': 'Azija', 'bs-Cyrl': 'Азија', 'bs-Latn': 'Azija', 'ca': 'Àsia', 'ca-ES-valencia': 'Àsia', 'ccp': '𑄃𑄬𑄥𑄨𑄠', 'ce': 'Ази', 'ceb': 'Asia', 'chr': 'ᏓᎶᎾᎨᏍᏛ', 'ckb': 'ئاسیا', 'cs': 'Asie', 'cv': 'Ази', 'cy': 'Asia', 'da': 'Asien', 'de': 'Asien', 'dsb': 'Azija', 'dz': 'ཨེ་ཤི་ཡ', 'ee': 'Asia nutome', 'el': 'Ασία', 'el-polyton': 'Ασία', 'en': 'Asia', 'en-Dsrt': '𐐁𐑈𐐲', 'es': 'Asia', 'et': 'Aasia', 'eu': 'Asia', 'fa': 'آسیا', 'ff-Adlm': '𞤀𞥄𞤧𞤭𞤴𞤢', 'fi': 'Aasia', 'fil': 'Asya', 'fo': 'Asia', 'fr': 'Asie', 'fur': 'Asie', 'fy': 'Azië', 'ga': 'an Áise', 'gd': 'Àisia', 'gl': 'Asia', 'gsw': 'Aasie', 'gu': 'એશિયા', 'ha': 'Asiya', 'he': 'אסיה', 'hi': 'एशिया', 'hi-Latn': 'Asia', 'hr': 'Azija', 'hsb': 'Azija', 'ht': 'Asie', 'hu': 'Ázsia', 'hy': 'Ասիա', 'ia': 'Asia', 'id': 'Asia', 'ie': 'Asia', 'ig': 'Asia', 'ii': 'ꑸꍏ', 'is': 'Asía', 'it': 'Asia', 'ja': 'アジア', 'jgo': 'Azî', 'jv': 'Asia', 'ka': 'აზია', 'kaa': 'Азия', 'kaa-Cyrl': 'Азия', 'kea': 'Ázia', 'kgp': 'Ajia', 'kk': 'Азия', 'kk-Cyrl': 'Азия', 'km': 'អាស៊ី', 'kn': 'ಏಷ್ಯಾ', 'ko': '아시아', 'kok': 'आशिया', 'kok-Deva': 'आशिया', 'ks': 'ایشیا', 'ks-Arab': 'ایشیا', 'ksh': 'Aasije', 'ku': 'Asya', 'ky': 'Азия', 'lb': 'Asien', 'lkt': 'Hazíla Makȟáwita', 'lld': 'Asia', 'lo': 'ອາຊີ', 'lrc': 'آسیا', 'lt': 'Azija', 'lv': 'Āzija', 'mai': 'एशिया', 'mi': 'Āhia', 'mk': 'Азија', 'ml': 'ഏഷ്യ', 'mn': 'Ази', 'mr': 'आशिया', 'ms': 'Asia', 'mt': 'Asja', 'my': 'အာရှ', 'mzn': 'آسیا', 'nb': 'Asia', 'ne': 'एशिया', 'nl': 'Azië', 'nn': 'Asia', 'no': 'Asia', 'nqo': 'ߊߖ߭ߌ߫', 'om': 'Eeshiyaa', 'or': 'ଏସିଆ', 'os': 'Ази', 'pa': 'ਏਸ਼ੀਆ', 'pa-Guru': 'ਏਸ਼ੀਆ', 'pcm': 'Éshia', 'pl': 'Azja', 'ps': 'آسيا', 'pt': 'Ásia', 'qu': 'Asia', 'rm': 'Asia', 'ro': 'Asia', 'ru': 'Азия', 'sc': 'Àsia', 'sd': 'ايشيا', 'sd-Arab': 'ايشيا', 'se': 'Ásia', 'si': 'ආසියාව', 'sk': 'Ázia', 'sl': 'Azija', 'so': 'Aasiya', 'sq': 'Azi', 'sr': 'Азија', 'sr-Cyrl': 'Азија', 'sr-Latn': 'Azija', 'sv': 'Asien', 'sw': 'Asia', 'syr': 'ܐܣܝܐ', 'ta': 'ஆசியா', 'te': 'ఆసియా', 'tg': 'Осиё', 'th': 'เอเชีย', 'ti': 'ኤስያ', 'tk': 'Aziýa', 'to': 'ʻĒsia', 'tr': 'Asya', 'tt': 'Азия', 'ug': 'ئاسىيا', 'uk': 'Азія', 'ur': 'ایشیا', 'uz': 'Osiyo', 'uz-Cyrl': 'Осиё', 'uz-Latn': 'Osiyo', 'vec': 'Azia', 'vi': 'Châu Á', 'wae': 'Asie', 'wo': 'Asia', 'xh': 'E-Asia', 'xnr': 'एशिया', 'yi': 'אַזיע', 'yo': 'Áṣíà', 'yrl': 'Ásiya', 'yue': '亞洲', 'yue-Hans': '亚洲', 'yue-Hant': '亞洲', 'zh': '亚洲', 'zh-Hans': '亚洲', 'zh-Hant': '亞洲', 'zu': 'i-Asia'},
    '143': {'af': 'Sentraal-Asië', 'ak': 'Asia Mfimfini', 'am': 'መካከለኛው እስያ', 'ar': 'وسط آسيا', 'as': 'মধ্য এছিয়া', 'ast': 'Asia Central', 'az': 'Mərkəzi Asiya', 'az-Cyrl': 'Мәркәзи Асија', 'az-Latn': 'Mərkəzi Asiya', 'bal-Latn': 'Myáni Ásiá', 'be': 'Цэнтральная Азія', 'be-tarask': 'Цэнтральная Азія', 'bg': 'Централна Азия', 'blo': 'Gacǝlǝŋtǝna gɩcɩɩca', 'bn': 'মধ্য এশিয়া', 'br': 'Azia ar Cʼhreiz', 'brx': 'मिरु एसिया', 'bs': 'Srednja Azija', 'bs-Cyrl': 'Централна Азија', 'bs-Latn': 'Srednja Azija', 'ca': 'Àsia central', 'ca-ES-valencia': 'Àsia central', 'ccp': '𑄟𑄧𑄖𑄴𑄙𑄳𑄠𑄧 𑄃𑄬𑄥𑄨𑄠', 'ce': 'Юккъера Малхбале', 'ceb': 'Central Asia', 'chr': 'ᎠᏰᏟ ᏓᎶᏂᎨᏍᏛ', 'ckb': 'ناوەڕاستی ئاسیا', 'cs': 'Střední Asie', 'cv': 'Тӗп Ази', 'cy': 'Canol Asia', 'da': 'Centralasien', 'de': 'Zentralasien', 'dsb': 'centralna Azija', 'dz': 'སྦུག་ཕྱོགས་ཀྱི་ཨེ་ཤི་ཡ', 'ee': 'Titina Asia nutome', 'el': 'Κεντρική Ασία', 'el-polyton': 'Κεντρική Ασία', 'en': 'Central Asia', 'en-Dsrt': '𐐝𐐯𐑌𐐻𐑉𐐲𐑊 𐐁𐑈𐐲', 'es': 'Asia central', 'et': 'Kesk-Aasia', 'eu': 'Erdialdeko Asia', 'fa': 'آسیای مرکزی', 'ff-Adlm': '𞤚𞤵𞤥𞤦𞤮𞥅𞤪𞤭 𞤀𞥄𞤧𞤭𞤴𞤢', 'fi': 'Keski-Aasia', 'fil': 'Gitnang Asya', 'fo': 'Miðasia', 'fr': 'Asie centrale', 'fur': 'Asie centrâl', 'fy': 'Sintraal-Azië', 'ga': 'an Áise Láir', 'gd': 'Meadhan Àisia', 'gl': 'Asia Central', 'gsw': 'Zentraalaasie', 'gu': 'મધ્ય એશિયા', 'ha': 'Asiya ta Tsakiya', 'he': 'מרכז אסיה', 'hi': 'मध्य एशिया', 'hi-Latn': 'Central Asia', 'hr': 'Srednja Azija', 'hsb': 'centralna Azija', 'ht': 'Asie centrale', 'hu': 'Közép-Ázsia', 'hy': 'Կենտրոնական Ասիա', 'ia': 'Asia central', 'id': 'Asia Tengah', 'ig': 'Etiti Asia', 'is': 'Mið-Asía', 'it': 'Asia centrale', 'ja': '中央アジア', 'jv': 'Asia Tengah', 'ka': 'ცენტრალური აზია', 'kaa': 'Орайлық Азия', 'kaa-Cyrl': 'Орайлық Азия', 'kea': 'Ázia Sentral', 'kgp': 'Ajia-Kuju', 'kk': 'Орталық Азия', 'kk-Cyrl': 'Орталық Азия', 'km': 'អាស៊ី\u200bកណ្ដាល', 'kn': 'ಮಧ್ಯ ಏಷ್ಯಾ', 'ko': '중앙 아시아', 'kok': 'मध्य आशिया', 'kok-Deva': 'मध्य आशिया', 'ks': 'مرکٔزی ایشیا', 'ks-Arab': 'مرکٔزی ایشیا', 'ksh': 'Meddelaasije', 'ku': 'Asyaya Navîn', 'ky': 'Борбор Азия', 'lb': 'Zentralasien', 'lld': 'Asia zentrala', 'lo': 'ອາຊີກາງ', 'lt': 'Centrinė Azija', 'lv': 'Centrālāzija', 'mai': 'मध्य एशिया', 'mi': 'Te Puku o Āhia', 'mk': 'Централна Азија', 'ml': 'മദ്ധ്യേഷ്യ', 'mn': 'Төв Ази', 'mr': 'मध्य आशिया', 'ms': 'Asia Tengah', 'mt': 'Asja Ċentrali', 'my': 'အလယ်အာရှ', 'mzn': 'میونی آسیا', 'nb': 'Sentral-Asia', 'ne': 'केन्द्रीय एशिया', 'nl': 'Centraal-Azië', 'nn': 'Sentral-Asia', 'no': 'Sentral-Asia', 'nqo': 'ߊߖ߭ߌ߫ ߕߊ߲ߓߊ߲', 'om': 'Eeshiyaa Gidduugaleessaa', 'or': 'ମଧ୍ୟ ଏସିଆ', 'pa': 'ਕੇਂਦਰੀ ਏਸ਼ੀਆ', 'pa-Guru': 'ਕੇਂਦਰੀ ਏਸ਼ੀਆ', 'pcm': 'Mídúl Éshia', 'pl': 'Azja Środkowa', 'ps': 'منځنۍ آسيا', 'pt': 'Ásia Central', 'qu': 'Asia Central', 'rm': 'Asia Centrala', 'ro': 'Asia Centrală', 'ru': 'Центральная Азия', 'sc': 'Àsia tzentrale', 'sd': 'وچ ايشيا', 'sd-Arab': 'وچ ايشيا', 'se': 'gaska-Ásia', 'si': 'මධ්\u200dයම ආසියාව', 'sk': 'stredná Ázia', 'sl': 'Osrednja Azija', 'so': 'Bartamaha Aasiya', 'sq': 'Azia Qendrore', 'sr': 'Централна Азија', 'sr-Cyrl': 'Централна Азија', 'sr-Latn': 'Centralna Azija', 'sv': 'Centralasien', 'sw': 'Asia ya Kati', 'syr': 'ܐܣܝܐ ܡܨܥܝܬܐ', 'ta': 'மத்திய ஆசியா', 'te': 'మధ్య ఆసియా', 'tg': 'Осиёи Марказӣ', 'th': 'เอเชียกลาง', 'ti': 'ማእከላይ ኤስያ', 'tk': 'Merkezi Aziýa', 'to': 'ʻĒsia lotoloto', 'tr': 'Orta Asya', 'tt': 'Үзәк Азия', 'ug': 'ئوتتۇرا ئاسىيا', 'uk': 'Центральна Азія', 'ur': 'وسطی ایشیا', 'uz': 'Markaziy Osiyo', 'uz-Cyrl': 'Марказий Осиё', 'uz-Latn': 'Markaziy Osiyo', 'vec': 'Azia sentrale', 'vi': 'Trung Á', 'wae': 'Zentralasie', 'wo': 'Asie centrale', 'xh': 'UMbindi Asia', 'xnr': 'बिचला एशिया', 'yi': 'צענטראַל־אַזיע', 'yo': 'Ààrin Gbùngbùn Éṣíà', 'yrl': 'Ásiya Piterapura', 'yue': '中亞', 'yue-Hans': '中亚', 'yue-Hant': '中亞', 'zh': '中亚', 'zh-Hans': '中亚', 'zh-Hant': '中亞', 'zu': 'i-Central Asia'},
    '145': {'af': 'Wes-Asië', 'ak': 'Asia Atɔeɛ', 'am': 'ምዕራባዊ እስያ', 'ar': 'غرب آسيا', 'as': 'পাশ্চাত্য এছিয়া', 'ast': 'Asia Occidental', 'az': 'Qərbi Asiya', 'az-Cyrl': 'Гәрби Асија', 'az-Latn': 'Qərbi Asiya', 'bal-Latn': 'Rónendi Ásiá', 'be': 'Заходняя Азія', 'be-tarask': 'Заходняя Азія', 'bg': 'Западна Азия', 'blo': 'Gacǝlǝŋtǝna gɩteŋshilelaŋ', 'bn': 'পশ্চিম এশিয়া', 'br': 'Azia ar Cʼhornôg', 'brx': 'सोनाबारि एसिया', 'bs': 'Zapadna Azija', 'bs-Cyrl': 'Западна Азија', 'bs-Latn': 'Zapadna Azija', 'ca': 'Àsia occidental', 'ca-ES-valencia': 'Àsia occidental', 'ccp': '𑄛𑄧𑄎𑄨𑄟𑄴 𑄃𑄬𑄥𑄨𑄠', 'ce': 'Юккъера а, Гергара а Малхбале', 'ceb': 'Western Asia', 'chr': 'ᏭᏕᎵᎬ ᏗᏜ ᏓᎶᏂᎨᏍᏛ', 'ckb': 'ڕۆژاوای ئاسیا', 'cs': 'západní Asie', 'cv': 'Анӑҫ Ази', 'cy': 'Gorllewin Asia', 'da': 'Vestasien', 'de': 'Westasien', 'dsb': 'pódwjacorna Azija', 'dz': 'ནུབ་ཕྱོགས་ཀྱི་ཨེ་ཤི་ཡ', 'ee': 'Ɣetoɖoƒelɔƒo Asia nutome', 'el': 'Δυτική Ασία', 'el-polyton': 'Δυτική Ασία', 'en': 'Western Asia', 'en-Dsrt': '𐐎𐐯𐑅𐐻𐐲𐑉𐑌 𐐁𐑈𐐲', 'es': 'Asia occidental', 'et': 'Lääne-Aasia', 'eu': 'Mendebaldeko Asia', 'fa': 'غرب آسیا', 'ff-Adlm': '𞤖𞤭𞥅𞤪𞤲𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭 𞤀𞥄𞤧𞤭𞤴𞤢', 'fi': 'Länsi-Aasia', 'fil': 'Kanlurang Asya', 'fo': 'Vesturasia', 'fr': 'Asie de l’Ouest', 'fur': 'Asie ocidentâl', 'fy': 'West-Azië', 'ga': 'Iarthar na hÁise', 'gd': 'Àisia an Iar', 'gl': 'Asia Occidental', 'gsw': 'Weschtaasie', 'gu': 'પશ્ચિમી એશિયા', 'ha': 'Yammacin Asiya', 'he': 'מערב אסיה', 'hi': 'पश्चिमी एशिया', 'hi-Latn': 'Western Asia', 'hr': 'Zapadna Azija', 'hsb': 'zapadna Azija', 'ht': 'Asie de l’Ouest', 'hu': 'Nyugat-Ázsia', 'hy': 'Արևմտյան Ասիա', 'ia': 'Asia occidental', 'id': 'Asia Bagian Barat', 'ig': 'Mpaghara Ọdịda Anyanwụ Asia', 'is': 'Vestur-Asía', 'it': 'Asia occidentale', 'ja': '西アジア', 'jv': 'Asia Kulon', 'ka': 'დასავლეთ აზია', 'kaa': 'Батыс Азия', 'kaa-Cyrl': 'Батыс Азия', 'kea': 'Ázia Osidental', 'kgp': 'Ajia-Rãpur', 'kk': 'Батыс Азия', 'kk-Cyrl': 'Батыс Азия', 'km': 'អាស៊ី\u200bខាង\u200bលិច', 'kn': 'ಪಶ್ಚಿಮ ಏಷ್ಯಾ', 'ko': '서아시아', 'kok': 'अस्तंत आशिया', 'kok-Deva': 'अस्तंत आशिया', 'ks': 'مَغرِبی ایشیا', 'ks-Arab': 'مَغرِبی ایشیا', 'ksh': 'Wäß-Aasije', 'ku': 'Rojavayê Asyayê', 'ky': 'Батыш Азия', 'lb': 'Westasien', 'lld': 'Asia ozidentala', 'lo': 'ອາຊີຕາເວັນຕົກ', 'lt': 'Vakarų Azija', 'lv': 'Rietumāzija', 'mai': 'पश्चिमी एशिया', 'mi': 'Āhia ki te Uru', 'mk': 'Западна Азија', 'ml': 'പശ്ചിമേഷ്യ', 'mn': 'Баруун Ази', 'mr': 'पश्चिम आशिया', 'ms': 'Asia Barat', 'mt': 'Asja tal-Punent', 'my': 'အနောက်အာရှ', 'mzn': 'غربی آسیا', 'nb': 'Vest-Asia', 'ne': 'पश्चिमी एशिया', 'nl': 'West-Azië', 'nn': 'Vest-Asia', 'no': 'Vest-Asia', 'nqo': 'ߊߖ߭ߌ߫ ߕߟߋ߬ߓߋ', 'om': 'Eeshiyaa Dhihaa', 'or': 'ପଶ୍ଚିମ ଏସିଆ', 'pa': 'ਪੱਛਮੀ ਏਸ਼ੀਆ', 'pa-Guru': 'ਪੱਛਮੀ ਏਸ਼ੀਆ', 'pcm': 'Wẹ́stán Éshia', 'pl': 'Azja Zachodnia', 'ps': 'لویدیځ آسیا', 'pt': 'Ásia Ocidental', 'qu': 'Asia Occidental', 'rm': 'Asia dal Vest', 'ro': 'Asia Occidentală', 'ru': 'Западная Азия', 'sc': 'Àsia otzidentale', 'sd': 'اولهه ايشيا', 'sd-Arab': 'اولهه ايشيا', 'se': 'oarji-Ásia', 'si': 'බටහිර ආසියාව', 'sk': 'západná Ázia', 'sl': 'Zahodna Azija', 'so': 'Aasiyada Galbeed', 'sq': 'Azia Perëndimore', 'sr': 'Западна Азија', 'sr-Cyrl': 'Западна Азија', 'sr-Latn': 'Zapadna Azija', 'sv': 'Västasien', 'sw': 'Asia ya Magharibi', 'syr': 'ܐܣܝܐ ܡܥܪܒܝܬܐ', 'ta': 'மேற்கு ஆசியா', 'te': 'పశ్చిమ ఆసియా', 'tg': 'Осиёи Ғарбӣ', 'th': 'เอเชียตะวันตก', 'ti': 'ምዕራባዊ ኤስያ', 'tk': 'Günbatar Aziýa', 'to': 'ʻĒsia fakahihifo', 'tr': 'Batı Asya', 'tt': 'Көнбатыш Азия', 'ug': 'غەربىي ئاسىيا', 'uk': 'Західна Азія', 'ur': 'مغربی ایشیا', 'uz': 'G‘arbiy Osiyo', 'uz-Cyrl': 'Ғарбий Осиё', 'uz-Latn': 'G‘arbiy Osiyo', 'vec': 'Azia osidentale', 'vi': 'Tây Á', 'wae': 'Weštasie', 'wo': 'Asie sowwu jàng', 'xh': 'INtshona Asia', 'xnr': 'पश्चिमी एशिया', 'yi': 'מערב־אַזיע', 'yo': 'Ìwọ̀ Òòrùn Eṣíà', 'yrl': 'Ásiya Usidẽtawara', 'yue': '西亞', 'yue-Hans': '西亚', 'yue-Hant': '西亞', 'zh': '西亚', 'zh-Hans': '西亚', 'zh-Hant': '西亞', 'zu': 'i-Western Asia'},
    '150': {'af': 'Europa', 'ak': 'Yuropu', 'am': 'አውሮፓ', 'ar': 'أوروبا', 'as': 'ইউৰোপ', 'ast': 'Europa', 'az': 'Avropa', 'az-Cyrl': 'Авропа', 'az-Latn': 'Avropa', 'bal-Latn': 'Yurop', 'be': 'Еўропа', 'be-tarask': 'Еўропа', 'bg': 'Европа', 'bgn': 'اورورپا', 'blo': 'Garɩfɔntǝna', 'bn': 'ইউরোপ', 'br': 'Europa', 'brx': 'यूरोप', 'bs': 'Evropa', 'bs-Cyrl': 'Европа', 'bs-Latn': 'Evropa', 'ca': 'Europa', 'ca-ES-valencia': 'Europa', 'ccp': '𑄄𑄃𑄪𑄢𑄮𑄛𑄴', 'ce': 'Европа', 'ceb': 'Europe', 'chr': 'ᏳᎳᏛ', 'ckb': 'ئەورووپا', 'cs': 'Evropa', 'cv': 'Европа', 'cy': 'Ewrop', 'da': 'Europa', 'de': 'Europa', 'dsb': 'Europa', 'dz': 'ཡུ་རོབ', 'ee': 'Europa nutome', 'el': 'Ευρώπη', 'el-polyton': 'Ευρώπη', 'en': 'Europe', 'en-Dsrt': '𐐏𐐲𐑉𐐲𐐹', 'es': 'Europa', 'et': 'Euroopa', 'eu': 'Europa', 'fa': 'اروپا', 'ff-Adlm': '𞤓𞤪𞤨𞤢', 'fi': 'Eurooppa', 'fil': 'Europe', 'fo': 'Evropa', 'fr': 'Europe', 'fur': 'Europe', 'fy': 'Europa', 'ga': 'an Eoraip', 'gd': 'An Roinn-Eòrpa', 'gl': 'Europa', 'gsw': 'Euroopa', 'gu': 'યુરોપ', 'ha': 'Turai', 'he': 'אירופה', 'hi': 'यूरोप', 'hi-Latn': 'Europe', 'hr': 'Europa', 'hsb': 'Europa', 'ht': 'Europe', 'hu': 'Európa', 'hy': 'Եվրոպա', 'ia': 'Europa', 'id': 'Eropa', 'ie': 'Europa', 'ig': 'Europe', 'ii': 'ꉩꍏ', 'is': 'Evrópa', 'it': 'Europa', 'ja': 'ヨーロッパ', 'jgo': 'Ʉlôp', 'jv': 'Éropah', 'ka': 'ევროპა', 'kaa': 'Европа', 'kaa-Cyrl': 'Европа', 'kea': 'Europa', 'kgp': 'Orópa', 'kk': 'Еуропа', 'kk-Cyrl': 'Еуропа', 'km': 'អឺរ៉ុប', 'kn': 'ಯೂರೋಪ್', 'ko': '유럽', 'kok': 'युरोप', 'kok-Deva': 'युरोप', 'ks': 'یوٗرَپ', 'ks-Arab': 'یوٗرَپ', 'ksh': 'Europpa', 'ku': 'Ewropa', 'ky': 'Европа', 'lb': 'Europa', 'lkt': 'Wašíču Makȟáwita', 'lld': 'Europa', 'lo': 'ຢູໂຣບ', 'lrc': 'ئوروٙپا', 'lt': 'Europa', 'lv': 'Eiropa', 'mai': 'यूरोप', 'mi': 'Ūropi', 'mk': 'Европа', 'ml': 'യൂറോപ്പ്', 'mn': 'Европ', 'mni': 'য়ুরোপ', 'mni-Beng': 'য়ুরোপ', 'mr': 'युरोप', 'ms': 'Eropah', 'mt': 'Ewropa', 'my': 'ဥရောပ', 'mzn': 'اروپا', 'nb': 'Europa', 'ne': 'युरोप', 'nl': 'Europa', 'nn': 'Europa', 'no': 'Europa', 'nqo': 'ߋߙߐߔߎ߬', 'om': 'Awurooppaa', 'or': 'ୟୁରୋପ୍', 'os': 'Европӕ', 'pa': 'ਯੂਰਪ', 'pa-Guru': 'ਯੂਰਪ', 'pcm': 'Yúrop', 'pl': 'Europa', 'ps': 'اروپا', 'pt': 'Europa', 'qu': 'Europa', 'rm': 'Europa', 'ro': 'Europa', 'ru': 'Европа', 'sc': 'Europa', 'sd': 'يورپ', 'sd-Arab': 'يورپ', 'se': 'Eurohpá', 'si': 'යුරෝපය', 'sk': 'Európa', 'sl': 'Evropa', 'so': 'Yurub', 'sq': 'Evropë', 'sr': 'Европа', 'sr-Cyrl': 'Европа', 'sr-Latn': 'Evropa', 'sv': 'Europa', 'sw': 'Ulaya', 'syr': 'ܐܘܪܘܦܐ', 'ta': 'ஐரோப்பா', 'te': 'యూరప్', 'tg': 'Аврупо', 'th': 'ยุโรป', 'ti': 'ኤውሮጳ', 'tk': 'Ýewropa', 'to': 'ʻEulope', 'tr': 'Avrupa', 'tt': 'Европа', 'ug': 'ياۋروپا', 'uk': 'Європа', 'ur': 'یورپ', 'uz': 'Yevropa', 'uz-Cyrl': 'Европа', 'uz-Latn': 'Yevropa', 'vec': 'Europa', 'vi': 'Châu Âu', 'wae': 'Europa', 'wo': 'Europe', 'xh': 'EYurophu', 'xnr': 'यूरोप', 'yi': 'אייראפּע', 'yo': 'Yúróòpù', 'yrl': 'Eurupa', 'yue': '歐洲', 'yue-Hans': '欧洲', 'yue-Hant': '歐洲', 'zh': '欧洲', 'zh-Hans': '欧洲', 'zh-Hant': '歐洲', 'zu': 'i-Europe'},
    '151': {'af': 'Oos-Europa', 'ak': 'Yuropu Apuieɛ', 'am': 'ምስራቃዊ አውሮፓ', 'ar': 'شرق أوروبا', 'as': 'প্ৰাচ্য ইউৰোপ', 'ast': 'Europa Oriental', 'az': 'Şərqi Avropa', 'az-Cyrl': 'Шәрги Авропа', 'az-Latn': 'Şərqi Avropa', 'bal-Latn': 'Ródarátki Yurop', 'be': 'Усходняя Еўропа', 'be-tarask': 'Усходняя Еўропа', 'bg': 'Източна Европа', 'blo': 'Garɩfɔntǝna gajakalaŋ', 'bn': 'পূর্ব ইউরোপ', 'br': 'Europa ar Reter', 'brx': 'सानजायारि यूरोप', 'bs': 'Istočna Evropa', 'bs-Cyrl': 'Источна Европа', 'bs-Latn': 'Istočna Evropa', 'ca': 'Europa oriental', 'ca-ES-valencia': 'Europa oriental', 'ccp': '𑄛𑄪𑄉𑄬𑄘𑄨 𑄄𑄃𑄪𑄢𑄮𑄛𑄴', 'ce': 'Малхбален Европа', 'ceb': 'Eastern Europe', 'chr': 'ᏗᎧᎸᎬ ᏗᏜ ᏳᎳᏛ', 'ckb': 'ڕۆژهەڵاتی ئەورووپا', 'cs': 'východní Evropa', 'cv': 'Хӗвелтухӑҫ Европа', 'cy': 'Dwyrain Ewrop', 'da': 'Østeuropa', 'de': 'Osteuropa', 'dsb': 'pódzajtšna Europa', 'dz': 'ཤར་ཕྱོགས་ཀྱི་ཡུ་རོབ', 'ee': 'Ɣedzeƒe Europa nutome', 'el': 'Ανατολική Ευρώπη', 'el-polyton': 'Ανατολική Ευρώπη', 'en': 'Eastern Europe', 'en-Dsrt': '𐐀𐑅𐐻𐐲𐑉𐑌 𐐏𐐲𐑉𐐲𐐹', 'es': 'Europa oriental', 'et': 'Ida-Euroopa', 'eu': 'Ekialdeko Europa', 'fa': 'شرق اروپا', 'ff-Adlm': '𞤊𞤵𞤯𞤲𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭 𞤓𞤪𞤨𞤢', 'fi': 'Itä-Eurooppa', 'fil': 'Silangang Europe', 'fo': 'Eysturevropa', 'fr': 'Europe de l’Est', 'fur': 'Europe orientâl', 'fy': 'East-Europa', 'ga': 'Oirthear na hEorpa', 'gd': 'An Roinn-Eòrpa an Ear', 'gl': 'Europa do Leste', 'gsw': 'Oschteuroopa', 'gu': 'પૂર્વીય યુરોપ', 'ha': 'Gabashin Turai', 'he': 'מזרח אירופה', 'hi': 'पूर्वी यूरोप', 'hi-Latn': 'Eastern Europe', 'hr': 'Istočna Europa', 'hsb': 'wuchodna Europa', 'ht': 'Europe de l’Est', 'hu': 'Kelet-Európa', 'hy': 'Արևելյան Եվրոպա', 'ia': 'Europa oriental', 'id': 'Eropa Bagian Timur', 'ig': 'Mpaghara Ọwụwa Anyanwụ Europe', 'is': 'Austur-Evrópa', 'it': 'Europa orientale', 'ja': '東ヨーロッパ', 'jv': 'Éropah Wétan', 'ka': 'აღმოსავლეთ ევროპა', 'kaa': 'Шығыс Европа', 'kaa-Cyrl': 'Шығыс Европа', 'kea': 'Europa Oriental', 'kgp': 'Orópa-Rãjur', 'kk': 'Шығыс Еуропа', 'kk-Cyrl': 'Шығыс Еуропа', 'km': 'អឺរ៉ុប\u200bខាង\u200bកើត', 'kn': 'ಪೂರ್ವ ಯೂರೋಪ್', 'ko': '동유럽', 'kok': 'उदेंत युरोप', 'kok-Deva': 'उदेंत युरोप', 'ks': 'مشرِقی یوٗرَپ', 'ks-Arab': 'مشرِقی یوٗرَپ', 'ksh': 'Oß-Europpa', 'ku': 'Rojhilatê Ewropayê', 'ky': 'Чыгыш Европа', 'lb': 'Osteuropa', 'lld': 'Europa orientala', 'lo': 'ຢູໂຣບຕາເວັນອອກ', 'lt': 'Rytų Europa', 'lv': 'Austrumeiropa', 'mai': 'पूरबी यूरोप', 'mi': 'Ūropi ki te Rāwhiti', 'mk': 'Источна Европа', 'ml': 'കിഴക്കൻ യൂറോപ്പ്', 'mn': 'Зүүн Европ', 'mr': 'पूर्व युरोप', 'ms': 'Eropah Timur', 'mt': 'Ewropa tal-Lvant', 'my': 'အရှေ့ ဥရောပ', 'mzn': 'شرقی اروپا', 'nb': 'Øst-Europa', 'ne': 'पूर्वी युरोप', 'nl': 'Oost-Europa', 'nn': 'Aust-Europa', 'no': 'Øst-Europa', 'nqo': 'ߋߙߐߔߎ߬ ߕߟߋ߬ߓߐ', 'om': 'Awurooppaa Bahaa', 'or': 'ପୂର୍ବ ୟୁରୋପ୍', 'pa': 'ਪੂਰਬੀ ਯੂਰਪ', 'pa-Guru': 'ਪੂਰਬੀ ਯੂਰਪ', 'pcm': 'Ístán Yúrop', 'pl': 'Europa Wschodnia', 'ps': 'ختيځ اروپا', 'pt': 'Europa Oriental', 'qu': 'Europa Oriental', 'rm': 'Europa Orientala', 'ro': 'Europa Orientală', 'ru': 'Восточная Европа', 'sc': 'Europa orientale', 'sd': 'اوڀر يورپ', 'sd-Arab': 'اوڀر يورپ', 'se': 'nuorta-Eurohpá', 'si': 'නැගෙනහිර යුරෝපය', 'sk': 'východná Európa', 'sl': 'Vzhodna Evropa', 'so': 'Yurubta Bari', 'sq': 'Evropa Lindore', 'sr': 'Источна Европа', 'sr-Cyrl': 'Источна Европа', 'sr-Latn': 'Istočna Evropa', 'sv': 'Östeuropa', 'sw': 'Ulaya ya Mashariki', 'syr': 'ܐܘܪܘܦܐ ܡܕܢܚܝܬܐ', 'ta': 'கிழக்கு ஐரோப்பா', 'te': 'తూర్పు యూరోప్', 'tg': 'Аврупои Шарқӣ', 'th': 'ยุโรปตะวันออก', 'ti': 'ምብራቓዊ ኤውሮጳ', 'tk': 'Gündogar Ýewropa', 'to': 'ʻEulope fakahahake', 'tr': 'Doğu Avrupa', 'tt': 'Көнчыгыш Европа', 'ug': 'شەرقىي ياۋروپا', 'uk': 'Східна Європа', 'ur': 'مشرقی یورپ', 'uz': 'Sharqiy Yevropa', 'uz-Cyrl': 'Шарқий Европа', 'uz-Latn': 'Sharqiy Yevropa', 'vec': 'Europa orientale', 'vi': 'Đông Âu', 'wae': 'Ošteuropa', 'wo': 'Europe bu penku', 'xh': 'IMpuma Yurophu', 'xnr': 'पूर्वी यूरोप', 'yi': 'מזרח־אייראפּע', 'yo': 'Ìlà Òrùn Yúrópù', 'yrl': 'Eurupa Uriẽtawara', 'yue': '東歐', 'yue-Hans': '东欧', 'yue-Hant': '東歐', 'zh': '东欧', 'zh-Hans': '东欧', 'zh-Hant': '東歐', 'zu': 'i-Eastern Europe'},
    '154': {'af': 'Noord-Europa', 'ak': 'Yuropu Atifi', 'am': 'ሰሜናዊ አውሮፓ', 'ar': 'شمال أوروبا', 'as': 'উদীচ্য ইউৰোপ', 'ast': 'Europa del Norte', 'az': 'Şimali Avropa', 'az-Cyrl': 'Шимали Авропа', 'az-Latn': 'Şimali Avropa', 'bal-Latn': 'Shemáli Yurop', 'be': 'Паўночная Еўропа', 'be-tarask': 'Паўночная Еўропа', 'bg': 'Северна Европа', 'blo': 'Garɩfɔntǝna gʊnyɩpɛnɛlaŋ', 'bn': 'উত্তর ইউরোপ', 'br': 'Europa an Norzh', 'brx': 'साहायारि यूरोप', 'bs': 'Sjeverna Evropa', 'bs-Cyrl': 'Сјеверна Европа', 'bs-Latn': 'Sjeverna Evropa', 'ca': 'Europa septentrional', 'ca-ES-valencia': 'Europa septentrional', 'ccp': '𑄅𑄪𑄖𑄴𑄖𑄮𑄢𑄴 𑄄𑄃𑄪𑄢𑄮𑄛𑄴', 'ce': 'Къилбаседа Европа', 'ceb': 'Northern Europe', 'chr': 'ᏧᏴᏢ ᏗᏜ ᏳᎳᏛ', 'ckb': 'باکووری ئەورووپا', 'cs': 'severní Evropa', 'cv': 'Ҫурҫӗр Европа', 'cy': 'Gogledd Ewrop', 'da': 'Nordeuropa', 'de': 'Nordeuropa', 'dsb': 'pódpołnocna Europa', 'dz': 'བྱང་ཕྱོགས་ཀྱི་ཡུ་རོབ', 'ee': 'Dziehelɔƒo Europa nutome', 'el': 'Βόρεια Ευρώπη', 'el-polyton': 'Βόρεια Ευρώπη', 'en': 'Northern Europe', 'en-Dsrt': '𐐤𐐱𐑉𐑄𐐲𐑉𐑌 𐐏𐐲𐑉𐐲𐐹', 'es': 'Europa septentrional', 'et': 'Põhja-Euroopa', 'eu': 'Ipar Europa', 'fa': 'شمال اروپا', 'ff-Adlm': '𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭 𞤓𞤪𞤨𞤢', 'fi': 'Pohjois-Eurooppa', 'fil': 'Hilagang Europe', 'fo': 'Norðurevropa', 'fr': 'Europe du Nord', 'fur': 'Europe setentrionâl', 'fy': 'Noard-Europa', 'ga': 'Tuaisceart na hEorpa', 'gd': 'An Roinn-Eòrpa a Tuath', 'gl': 'Europa Setentrional', 'gsw': 'Nordeuroopa', 'gu': 'ઉત્તરીય યુરોપ', 'ha': 'Arewacin Turai', 'he': 'צפון אירופה', 'hi': 'उत्तरी यूरोप', 'hi-Latn': 'Northern Europe', 'hr': 'Sjeverna Europa', 'hsb': 'sewjerna Europa', 'ht': 'Europe du Nord', 'hu': 'Észak-Európa', 'hy': 'Հյուսիսային Եվրոպա', 'ia': 'Europa septentrional', 'id': 'Eropa Bagian Utara', 'ig': 'Mpaghara Ugwu Europe', 'is': 'Norður-Evrópa', 'it': 'Europa settentrionale', 'ja': '北ヨーロッパ', 'jv': 'Éropah Lor', 'ka': 'ჩრდილოეთ ევროპა', 'kaa': 'Арқа Европа', 'kaa-Cyrl': 'Арқа Европа', 'kea': 'Europa di Norti', 'kgp': 'Orópa Setẽntrionỹr', 'kk': 'Солтүстік Еуропа', 'kk-Cyrl': 'Солтүстік Еуропа', 'km': 'អឺរ៉ុប\u200bខាង\u200bជើង', 'kn': 'ಉತ್ತರ ಯೂರೋಪ್', 'ko': '북유럽', 'kok': 'उत्तर युरोप', 'kok-Deva': 'उत्तर युरोप', 'ks': 'شُمٲلی یوٗرَپ', 'ks-Arab': 'شُمٲلی یوٗرَپ', 'ksh': 'Nood-Europpa', 'ku': 'Bakurê Ewropayê', 'ky': 'Түндүк Европа', 'lb': 'Nordeuropa', 'lld': 'Europa dl Nord', 'lo': 'ຢູໂຣບເໜືອ', 'lt': 'Šiaurės Europa', 'lv': 'Ziemeļeiropa', 'mai': 'उत्तरी यूरोप', 'mi': 'Ūropi ki te Raki', 'mk': 'Северна Европа', 'ml': 'വടക്കേ യൂറോപ്പ്', 'mn': 'Хойд Европ', 'mr': 'उत्तर युरोप', 'ms': 'Eropah Utara', 'mt': 'Ewropa ta’ Fuq', 'my': 'မြောက် ဥရောပ', 'mzn': 'شمالی اروپا', 'nb': 'Nord-Europa', 'ne': 'उत्तरी युरोप', 'nl': 'Noord-Europa', 'nn': 'Nord-Europa', 'no': 'Nord-Europa', 'nqo': 'ߋߙߐߔߎ߬ ߞߐ߬ߘߎ߮', 'om': 'Awurooppaa Kaabaa', 'or': 'ଉତ୍ତର ୟୁରୋପ୍', 'pa': 'ਉੱਤਰੀ ਯੂਰਪ', 'pa-Guru': 'ਉੱਤਰੀ ਯੂਰਪ', 'pcm': 'Nọ́tán Yúrop', 'pl': 'Europa Północna', 'ps': 'شمالي اروپا', 'pt': 'Europa Setentrional', 'qu': 'Europa del Norte', 'rm': 'Europa dal Nord', 'ro': 'Europa Septentrională', 'ru': 'Северная Европа', 'sc': 'Europa setentrionale', 'sd': 'اترين يورپ', 'sd-Arab': 'اترين يورپ', 'se': 'davvi-Eurohpá', 'si': 'උතුරු යුරෝපය', 'sk': 'severná Európa', 'sl': 'Severna Evropa', 'so': 'Yurubta Waqooyi', 'sq': 'Evropa Veriore', 'sr': 'Северна Европа', 'sr-Cyrl': 'Северна Европа', 'sr-Latn': 'Severna Evropa', 'sv': 'Nordeuropa', 'sw': 'Ulaya ya Kaskazini', 'syr': 'ܐܘܪܘܦܐ ܓܪܒܝܝܬܐ', 'ta': 'வடக்கு ஐரோப்பா', 'te': 'ఉత్తర యూరోప్', 'tg': 'Аврупои Шимолӣ', 'th': 'ยุโรปเหนือ', 'ti': 'ሰሜናዊ ኤውሮጳ', 'tk': 'Demirgazyk Ýewropa', 'to': 'ʻEulope fakatokelau', 'tr': 'Kuzey Avrupa', 'tt': 'Төньяк Европа', 'ug': 'شىمالىي ياۋروپا', 'uk': 'Північна Європа', 'ur': 'شمالی یورپ', 'uz': 'Shimoliy Yevropa', 'uz-Cyrl': 'Шимолий Европа', 'uz-Latn': 'Shimoliy Yevropa', 'vec': 'Europa setentrionale', 'vi': 'Bắc Âu', 'wae': 'Nordeuropa', 'wo': 'Europe du nord', 'xh': 'UMntla Yurophu', 'xnr': 'उत्तरी यूरोप', 'yi': 'צפֿון־אייראפּע', 'yo': 'Àríwá Yúróòpù', 'yrl': 'Eurupa Setẽtiriunau', 'yue': '北歐', 'yue-Hans': '北欧', 'yue-Hant': '北歐', 'zh': '北欧', 'zh-Hans': '北欧', 'zh-Hant': '北歐', 'zu': 'i-Northern Europe'},
    '155': {'af': 'Wes-Europa', 'ak': 'Yuropu Atɔeɛ', 'am': 'ምዕራባዊ አውሮፓ', 'ar': 'غرب أوروبا', 'as': 'পাশ্চাত্য ইউৰোপ', 'ast': 'Europa Occidental', 'az': 'Qərbi Avropa', 'az-Cyrl': 'Гәрби Авропа', 'az-Latn': 'Qərbi Avropa', 'bal-Latn': 'Rónendi Yurop', 'be': 'Заходняя Еўропа', 'be-tarask': 'Заходняя Еўропа', 'bg': 'Западна Европа', 'blo': 'Garɩfɔntǝna gɩteŋshilelaŋ', 'bn': 'পশ্চিম ইউরোপ', 'br': 'Europa ar Cʼhornôg', 'brx': 'सोनाबारि यूरोप', 'bs': 'Zapadna Evropa', 'bs-Cyrl': 'Западна Европа', 'bs-Latn': 'Zapadna Evropa', 'ca': 'Europa occidental', 'ca-ES-valencia': 'Europa occidental', 'ccp': '𑄛𑄧𑄎𑄨𑄟𑄴 𑄄𑄃𑄪𑄢𑄮𑄛𑄴', 'ce': 'Малхбузен Европа', 'ceb': 'Western Europe', 'chr': 'ᏭᏕᎵᎬ ᏗᏜ ᏳᎳᏛ', 'ckb': 'ڕۆژاوای ئەورووپا', 'cs': 'západní Evropa', 'cv': 'Анӑҫ Европа', 'cy': 'Gorllewin Ewrop', 'da': 'Vesteuropa', 'de': 'Westeuropa', 'dsb': 'pódwjacorna Europa', 'dz': 'ནུབ་ཕྱོགས་ཀྱི་ཡུ་རོབ', 'ee': 'Ɣetoɖoƒelɔƒo Europa nutome', 'el': 'Δυτική Ευρώπη', 'el-polyton': 'Δυτική Ευρώπη', 'en': 'Western Europe', 'en-Dsrt': '𐐎𐐯𐑅𐐻𐐲𐑉𐑌 𐐏𐐲𐑉𐐲𐐹', 'es': 'Europa occidental', 'et': 'Lääne-Euroopa', 'eu': 'Mendebaldeko Europa', 'fa': 'غرب اروپا', 'ff-Adlm': '𞤖𞤭𞥅𞤪𞤲𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭 𞤓𞤪𞤨𞤢', 'fi': 'Länsi-Eurooppa', 'fil': 'Kanlurang Europe', 'fo': 'Vesturevropa', 'fr': 'Europe de l’Ouest', 'fur': 'Europe ocidentâl', 'fy': 'West-Europa', 'ga': 'Iarthar na hEorpa', 'gd': 'An Roinn-Eòrpa an Iar', 'gl': 'Europa Occidental', 'gsw': 'Weschteuroopa', 'gu': 'પશ્ચિમી યુરોપ', 'ha': 'Yammacin Turai', 'he': 'מערב אירופה', 'hi': 'पश्चिमी यूरोप', 'hi-Latn': 'Western Europe', 'hr': 'Zapadna Europa', 'hsb': 'zapadna Europa', 'ht': 'Europe de l’Ouest', 'hu': 'Nyugat-Európa', 'hy': 'Արևմտյան Եվրոպա', 'ia': 'Europa occidental', 'id': 'Eropa Bagian Barat', 'ig': 'Mpaghara Ọdịda Anyanwụ Europe', 'is': 'Vestur-Evrópa', 'it': 'Europa occidentale', 'ja': '西ヨーロッパ', 'jv': 'Éropah Kulon', 'ka': 'დასავლეთ ევროპა', 'kaa': 'Батыс Европа', 'kaa-Cyrl': 'Батыс Европа', 'kea': 'Europa Osidental', 'kgp': 'Orópa Rãpur', 'kk': 'Батыс Еуропа', 'kk-Cyrl': 'Батыс Еуропа', 'km': 'អឺរ៉ុប\u200bខាង\u200bលិច', 'kn': 'ಪಶ್ಚಿಮ ಯೂರೋಪ್', 'ko': '서유럽', 'kok': 'अस्तंत युरोप', 'kok-Deva': 'अस्तंत युरोप', 'ks': 'مغرِبی یوٗرَپ', 'ks-Arab': 'مغرِبی یوٗرَپ', 'ksh': 'Wäß-Europpa', 'ku': 'Rojavayê Ewropayê', 'ky': 'Батыш Европа', 'lb': 'Westeuropa', 'lld': 'Europa ozidentala', 'lo': 'ຢູໂຣບຕາເວັນຕົກ', 'lt': 'Vakarų Europa', 'lv': 'Rietumeiropa', 'mai': 'पश्चिमी यूरोप', 'mi': 'Ūropi ki te Uru', 'mk': 'Западна Европа', 'ml': 'പശ്ചിമ യൂറോപ്പ്', 'mn': 'Баруун Европ', 'mr': 'पश्चिम युरोप', 'ms': 'Eropah Barat', 'mt': 'Ewropa tal-Punent', 'my': 'အနောက် ဥရောပ', 'mzn': 'غربی اروپا', 'nb': 'Vest-Europa', 'ne': 'पश्चिमी युरोप', 'nl': 'West-Europa', 'nn': 'Vest-Europa', 'no': 'Vest-Europa', 'nqo': 'ߋߙߐߔߎ߬ ߕߟߋ߬ߓߋ', 'om': 'Awurooppaa Dhihaa', 'or': 'ପଶ୍ଚିମ ୟୁରୋପ୍', 'pa': 'ਪੱਛਮੀ ਯੂਰਪ', 'pa-Guru': 'ਪੱਛਮੀ ਯੂਰਪ', 'pcm': 'Wẹ́stán Yúrop', 'pl': 'Europa Zachodnia', 'ps': 'لوېديځ اروپا', 'pt': 'Europa Ocidental', 'qu': 'Europa Occidental', 'rm': 'Europa dal Vest', 'ro': 'Europa Occidentală', 'ru': 'Западная Европа', 'sc': 'Europa otzidentale', 'sd': 'اولهه يورپ', 'sd-Arab': 'اولهه يورپ', 'se': 'oarji-Eurohpá', 'si': 'බටහිර යුරෝපය', 'sk': 'západná Európa', 'sl': 'Zahodna Evropa', 'so': 'Yurubta Galbeed', 'sq': 'Evropa Perëndimore', 'sr': 'Западна Европа', 'sr-Cyrl': 'Западна Европа', 'sr-Latn': 'Zapadna Evropa', 'sv': 'Västeuropa', 'sw': 'Ulaya ya Magharibi', 'syr': 'ܐܘܪܘܦܐ ܡܥܪܒ݂ܝܬܐ', 'ta': 'மேற்கு ஐரோப்பா', 'te': 'పశ్చిమ యూరోప్', 'tg': 'Аврупои Ғарбӣ', 'th': 'ยุโรปตะวันตก', 'ti': 'ምዕራባዊ ኤውሮጳ', 'tk': 'Günbatar Ýewropa', 'to': 'ʻEulope fakahihifo', 'tr': 'Batı Avrupa', 'tt': 'Көнбатыш Европа', 'ug': 'غەربىي ياۋروپا', 'uk': 'Західна Європа', 'ur': 'مغربی یورپ', 'uz': 'G‘arbiy Yevropa', 'uz-Cyrl': 'Ғарбий Европа', 'uz-Latn': 'G‘arbiy Yevropa', 'vec': 'Europa osidentale', 'vi': 'Tây Âu', 'wae': 'Wešteuropa', 'wo': 'Europe sowwu jàng', 'xh': 'INtshona Yurophu', 'xnr': 'पश्चिमी यूरोप', 'yi': 'מערב־אייראפּע', 'yo': 'Ìwọ̀ Òòrùn Yúrópù', 'yrl': 'Eurupa Usidẽtawara', 'yue': '西歐', 'yue-Hans': '西欧', 'yue-Hant': '西歐', 'zh': '西欧', 'zh-Hans': '西欧', 'zh-Hant': '西歐', 'zu': 'i-Western Europe'},
    '202': {'af': 'Afrika suid van die Sahara', 'ak': 'Abibirem Mpaprɛ Anaafoɔ', 'am': 'ከሰሃራ በታች አፍሪካ', 'ar': 'أفريقيا جنوب الصحراء الكبرى', 'as': 'উপ-ছাহাৰান আফ্ৰিকা', 'az': 'Saharadan cənub', 'az-Latn': 'Saharadan cənub', 'bal-Latn': 'Sab-Sahári Apriká', 'be': 'Трапічная Афрыка', 'be-tarask': 'Трапічная Афрыка', 'bg': 'Субсахарска Африка', 'blo': 'Garɩɖontǝna Sahara katǝntǝn', 'bn': 'উপ সাহারান আফ্রিকা', 'br': 'Afrika issaharat', 'brx': 'साहारायारि लोब्बा गोनां आफ्रिका', 'bs': 'Subsaharska Afrika', 'bs-Cyrl': 'Подсахарска Африка', 'bs-Latn': 'Subsaharska Afrika', 'ca': 'Àfrica subsahariana', 'ca-ES-valencia': 'Àfrica subsahariana', 'ceb': 'Sub-Saharan Africa', 'chr': 'ᎭᏫᏂ-ᏌᎭᏩ ᎬᎿᎨᏍᏛ', 'ckb': 'ئەفریقای ژێر سەحرا', 'cs': 'subsaharská Afrika', 'cv': 'Тропик Африка', 'cy': 'Affrica Is-Sahara', 'da': 'Subsaharisk Afrika', 'de': 'Subsahara-Afrika', 'dsb': 'subsaharojska Afrika', 'el': 'Υποσαχάρια Αφρική', 'el-polyton': 'Υποσαχάρια Αφρική', 'en': 'Sub-Saharan Africa', 'es': 'África subsahariana', 'et': 'Sahara-tagune Aafrika', 'eu': 'Saharaz hegoaldeko Afrika', 'fa': 'افریقای جنوب صحرا', 'ff-Adlm': '𞤀𞤬𞤪𞤭𞤳𞤢𞥄 𞤂𞤫𞤧-𞤅𞤢𞥄𞤸𞤢𞤪𞤢', 'fi': 'Saharan eteläpuolinen Afrikka', 'fil': 'Sub-Saharan Africa', 'fo': 'Afrika sunnanfyri Sahara', 'fr': 'Afrique subsaharienne', 'ga': 'an Afraic fho-Shahárach', 'gd': 'Afraga Deas air an t-Sathara', 'gl': 'África subsahariana', 'gu': 'સબ-સહારન આફ્રિકા', 'ha': 'Afirka ta Kudancin Sahara', 'he': 'אפריקה שמדרום לסהרה', 'hi': 'उप-सहारा अफ़्रीका', 'hi-Latn': 'Sub-Saharan Africa', 'hr': 'Subsaharska Afrika', 'hsb': 'subsaharaska Afrika', 'ht': 'Afrique subsaharienne', 'hu': 'Szubszaharai Afrika', 'hy': 'Արևադարձային Աֆրիկա', 'ia': 'Africa subsaharian', 'id': 'Afrika Sub-Sahara', 'ig': 'Sub-Saharan Afrịka', 'is': 'Afríka sunnan Sahara', 'it': 'Africa subsahariana', 'ja': 'サブサハラアフリカ', 'jv': 'Afrika Kidule Sahara', 'ka': 'სუბსაჰარული აფრიკა', 'kaa': 'Тропикалық Африка', 'kaa-Cyrl': 'Тропикалық Африка', 'kea': 'Áfrika Subisariana', 'kgp': 'Afrika Sahara-Jẽgu', 'kk': 'Субсахаралық Африка', 'kk-Cyrl': 'Субсахаралық Африка', 'km': 'អនុតំបន់សាហារ៉ាអាហ្វ្រិក', 'kn': 'ಉಪ-ಸಹಾರನ್ ಆಫ್ರಿಕಾ', 'ko': '사하라 사막 이남 아프리카', 'kok': 'उप-सहाराई आफ्रिका', 'kok-Deva': 'उप-सहाराई आफ्रिका', 'ks': 'ذیلی سہارن افریقہ', 'ks-Arab': 'ذیلی سہارن افریقہ', 'ku': 'Afrîkaya Jêra Sahrayê', 'ky': 'Африка өлкөлөрү', 'lld': 'Africa sot-sahariana', 'lo': 'ຊັບ ຊາຮາຣານ ອາຟຣິກາ', 'lt': 'Užsachario Afrika', 'lv': 'Subsahāras Āfrika', 'mai': 'उप सहारा अफ्रीका', 'mi': 'Āwherika ki te Tonga o Te Hahāra', 'mk': 'Супсахарска Африка', 'ml': 'സബ്-സഹാറൻ ആഫ്രിക്ക', 'mn': 'Сахарын цөлийн урд хэсгийн Африк', 'mr': 'उप-सहारा आफ्रिका', 'ms': 'Afrika Sub-Sahara', 'my': 'ဆာဟာရ-အောက်ပိုင်း အာဖရိက', 'nb': 'Afrika sør for Sahara', 'ne': 'उप-साहारा अफ्रिका', 'nl': 'Sub-Saharaans Afrika', 'nn': 'Afrika sør for Sahara', 'no': 'Afrika sør for Sahara', 'nqo': 'ߝߘߊ߬ߝߌ߲߬ߠߊ߫ ߞߌ߬ߢߍ߬ߞߏ߲ߞߏ߫ ߘߎ߰ߟߊ߬-ߖߡߊߣߊ', 'om': 'Afrikaa Sahaaraan Gadii', 'or': 'ଉପ-ସାହାରା ଆଫ୍ରିକା', 'pa': 'ਉਪ-ਸਹਾਰਾ ਅਫ਼ਰੀਕਾ', 'pa-Guru': 'ਉਪ-ਸਹਾਰਾ ਅਫ਼ਰੀਕਾ', 'pcm': 'Áfríka Éria Biló Sahára', 'pl': 'Afryka Subsaharyjska', 'ps': 'سب سهارن افريقا', 'pt': 'África Subsaariana', 'qu': 'Africa Sub-Sahariana', 'rm': 'Africa Subsaharica', 'ro': 'Africa Subsahariană', 'ru': 'Тропическая Африка', 'sc': 'Àfrica sub-sahariana', 'sd': 'سب-سهارا آفريڪا', 'sd-Arab': 'سب-سهارا آفريڪا', 'si': 'උප-සහරානු අප්\u200dරිකාව', 'sk': 'subsaharská Afrika', 'sl': 'podsaharska Afrika', 'so': 'Afrikada ka hooseysa Saxaraha', 'sq': 'Afrika Subsahariane', 'sr': 'Подсахарска Африка', 'sr-Cyrl': 'Подсахарска Африка', 'sr-Latn': 'Podsaharska Afrika', 'sv': 'Subsahariska Afrika', 'sw': 'Afrika ya Kusini mwa Jangwa la Sahara', 'syr': 'ܐܦܪܝܩܐ ܨܚܪܐ ܬܝܡܢܝܬܐ', 'ta': 'துணை சஹாரா ஆப்பிரிக்கா', 'te': 'ఉప సెహరన్ ఆఫ్రికా', 'tg': 'Африқои Суб-Сахара', 'th': 'แอฟริกาใต้สะฮารา', 'ti': 'ንኡስ ሰሃራዊ ኣፍሪቃ', 'tk': 'Saharadan aşakdaky Afrika', 'to': 'ʻAfilika fakasahala-tonga', 'tr': 'Sahra Altı Afrika', 'tt': 'Сахарадан көньякта Африка', 'uk': 'Африка на південь від Сахари', 'ur': 'ذیلی صحارن افریقہ', 'uz': 'Sahro janubidagi Afrika', 'uz-Latn': 'Sahro janubidagi Afrika', 'vec': 'Àfrega subsahariana', 'vi': 'Châu Phi hạ Sahara', 'wo': 'Afrique sub-saharienne', 'xh': 'UMzantsi weSahara', 'xnr': 'उप-सहारा अफ़्रीका', 'yo': 'Apá Sàhárà Áfíríkà', 'yrl': 'Afirika Subisariana', 'yue': '撒哈拉以南非洲', 'yue-Hans': '撒哈拉以南非洲', 'yue-Hant': '撒哈拉以南非洲', 'zh': '撒哈拉以南非洲', 'zh-Hans': '撒哈拉以南非洲', 'zh-Hant': '撒哈拉撒沙漠以南非洲', 'zu': 'Sub-Saharan Africa'},
    '419': {'af': 'Latyns-Amerika', 'ak': 'Laaten Amɛrika', 'am': 'ላቲን አሜሪካ', 'ar': 'أمريكا اللاتينية', 'as': 'লেটিন আমেৰিকা', 'ast': 'América Llatina', 'az': 'Latın Amerikası', 'az-Cyrl': 'Латын Америкасы', 'az-Latn': 'Latın Amerikası', 'bal-Latn': 'Látini Amriká', 'be': 'Лацінская Амерыка', 'be-tarask': 'Лацінская Амерыка', 'bg': 'Латинска Америка', 'blo': 'Latɛŋ kaAmalɩkatǝna', 'bn': 'লাতিন আমেরিকা', 'br': 'Amerika Latin', 'brx': 'लेटीन आमेरिका', 'bs': 'Latinska Amerika', 'bs-Cyrl': 'Латинска Америка', 'bs-Latn': 'Latinska Amerika', 'ca': 'Amèrica Llatina', 'ca-ES-valencia': 'Amèrica Llatina', 'ccp': '𑄣𑄳𑄠𑄑𑄨𑄚𑄴 𑄃𑄟𑄬𑄢𑄨𑄇', 'ce': 'Латинан Америка', 'ceb': 'Latin America', 'chr': 'ᎳᏘᏂ ᎠᎹᏰᏟ', 'ckb': 'ئەمەریکای لاتین', 'cs': 'Latinská Amerika', 'cv': 'Латинла Америка', 'cy': 'America Ladin', 'da': 'Latinamerika', 'de': 'Lateinamerika', 'dsb': 'Łatyńska Amerika', 'dz': 'ལེ་ཊིནཨ་མི་རི་ཀ', 'ee': 'Latin Amerika nutome', 'el': 'Λατινική Αμερική', 'el-polyton': 'Λατινική Αμερική', 'en': 'Latin America', 'en-Dsrt': '𐐢𐐰𐐻𐑌 𐐊𐑋𐐯𐑉𐐲𐐿𐐲 𐐰𐑌𐐼 𐑄 𐐗𐐯𐑉𐐲𐐺𐐨𐐲𐑌', 'es': 'Latinoamérica', 'et': 'Ladina-Ameerika', 'eu': 'Latinoamerika', 'fa': 'امریکای لاتین', 'ff-Adlm': '𞤀𞤥𞤫𞤪𞤭𞤳𞤢𞥄 𞤂𞤢𞤼𞤭𞤲𞤳𞤮', 'fi': 'Latinalainen Amerikka', 'fil': 'Latin America', 'fo': 'Latínamerika', 'fr': 'Amérique latine', 'fur': 'Americhe latine', 'fy': 'Latynsk-Amearika', 'ga': 'Meiriceá Laidineach', 'gd': 'Aimeireaga Laidinneach', 'gl': 'América Latina', 'gsw': 'Latiinameerika', 'gu': 'લેટિન અમેરિકા', 'ha': 'Latin Amurka', 'he': 'אמריקה הלטינית', 'hi': 'लैटिन अमेरिका', 'hi-Latn': 'Latin America', 'hr': 'Latinska Amerika', 'hsb': 'Łaćonska Amerika', 'ht': 'Amérique latine', 'hu': 'Latin-Amerika', 'hy': 'Լատինական Ամերիկա', 'ia': 'America latin', 'id': 'Amerika Latin', 'ie': 'latin America', 'ig': 'Latin America', 'is': 'Rómanska Ameríka', 'it': 'America Latina', 'ja': 'ラテンアメリカ', 'jv': 'Amérika Latin', 'ka': 'ლათინური ამერიკა', 'kaa': 'Латын Америкасы', 'kaa-Cyrl': 'Латын Америкасы', 'kea': 'Merka Latinu', 'kgp': 'Amẽrika Ratinỹ', 'kk': 'Латын Америкасы', 'kk-Cyrl': 'Латын Америкасы', 'km': 'អាមេរិក\u200bឡាទីន', 'kn': 'ಲ್ಯಾಟಿನ್ ಅಮೇರಿಕಾ', 'ko': '라틴 아메리카', 'kok': 'लॅटीन अमेरिका', 'kok-Deva': 'लॅटीन अमेरिका', 'ks': 'لاطیٖنی اَمریٖکا', 'ks-Arab': 'لاطیٖنی اَمریٖکا', 'ksh': 'Lateinamärrika', 'ku': 'Amerîkaya Latîn', 'kxv': 'laṭin amerika', 'kxv-Deva': 'लातिन आमेरिका ति', 'kxv-Latn': 'laṭin amerika', 'kxv-Orya': 'ଲାଟିନ୍\u200c ଆମେରିକା', 'kxv-Telu': 'లాటిన్ ఆమెరిక', 'ky': 'Латын Америкасы', 'lb': 'Latäinamerika', 'lld': 'America latina', 'lo': 'ລາຕິນ ອາເມລິກາ', 'lrc': 'ئمریکا لاتین', 'lt': 'Lotynų Amerika', 'lv': 'Latīņamerika', 'mai': 'लातिन अमेरिका', 'mi': 'Amerika Rātini', 'mk': 'Латинска Америка', 'ml': 'ലാറ്റിനമേരിക്ക', 'mn': 'Латин Америк', 'mr': 'लॅटिन अमेरिका', 'ms': 'Amerika Latin', 'mt': 'Amerika Latina', 'my': 'လက်တင်အမေရိက', 'mzn': 'لاتین آمریکا', 'nb': 'Latin-Amerika', 'ne': 'ल्याटिन अमेरिका', 'nl': 'Latijns-Amerika', 'nn': 'Latin-Amerika', 'no': 'Latin-Amerika', 'nqo': 'ߊߡߋߙߞߌ߬ ߟߊ߬ߕߍ߬ߡߊ߬ ߦߙߐ', 'om': 'Laatin Ameerikaa', 'or': 'ଲାଟିନ୍\u200c ଆମେରିକା', 'pa': 'ਲਾਤੀਨੀ ਅਮਰੀਕਾ', 'pa-Guru': 'ਲਾਤੀਨੀ ਅਮਰੀਕਾ', 'pcm': 'Látín Amẹ́ríka', 'pl': 'Ameryka Łacińska', 'ps': 'لاتیني امریکا', 'pt': 'América Latina', 'qu': 'AmericaLatina', 'rm': 'America Latina', 'ro': 'America Latină', 'ru': 'Латинская Америка', 'sc': 'Amèrica latina', 'sd': 'لاطيني آمريڪا', 'sd-Arab': 'لاطيني آمريڪا', 'se': 'lulli-Amerihkká', 'si': 'ලතින් ඇමෙරිකාව', 'sk': 'Latinská Amerika', 'sl': 'Latinska Amerika', 'so': 'Laatiin Ameerika', 'sq': 'Amerika Latine', 'sr': 'Латинска Америка', 'sr-Cyrl': 'Латинска Америка', 'sr-Latn': 'Latinska Amerika', 'sv': 'Latinamerika', 'sw': 'Amerika ya Kilatini', 'syr': 'ܐܡܪܝܟܐ ܠܬܝܢܝܬܐ', 'ta': 'லத்தீன் அமெரிக்கா', 'te': 'లాటిన్ అమెరికా', 'tg': 'Америкаи Лотинӣ', 'th': 'ละตินอเมริกา', 'ti': 'ላቲን ኣመሪካ', 'tk': 'Latyn Amerikasy', 'to': 'ʻAmelika fakalatina', 'tr': 'Latin Amerika', 'tt': 'Латин Америка', 'ug': 'لاتىن ئامېرىكا', 'uk': 'Латинська Америка', 'ur': 'لاطینی امریکہ', 'uz': 'Lotin Amerikasi', 'uz-Cyrl': 'Лотин Америкаси', 'uz-Latn': 'Lotin Amerikasi', 'vec': 'Mèrega Latina', 'vi': 'Châu Mỹ La-tinh', 'wae': 'Latíamerika', 'wo': 'Amerique Latine', 'xh': 'ILatin America', 'xnr': 'लैटिन अमरिका', 'yi': 'לאַטיין־אַמעריקע', 'yo': 'Látín Amẹ́ríkà', 'yrl': 'Amerika Latina', 'yue': '拉丁美洲', 'yue-Hans': '拉丁美洲', 'yue-Hant': '拉丁美洲', 'zh': '拉丁美洲', 'zh-Hans': '拉丁美洲', 'zh-Hant': '拉丁美洲', 'zu': 'i-Latin America'},
    'AA': {'en': 'Private use'},
    'AC': {'af': 'Ascensioneiland', 'ak': 'Asɛnhyin', 'am': 'አሴንሽን ደሴት', 'ar': 'جزيرة أسينشيون', 'as': 'এচেনচিয়ন দ্বীপ', 'ast': 'Islla Ascensión', 'az': 'Askenson adası', 'az-Cyrl': 'Аскенсон адасы', 'az-Latn': 'Askenson adası', 'bal-Latn': 'Asenshan Islánd', 'be': 'Востраў Узнясення', 'be-tarask': 'Востраў Узнясення', 'bg': 'остров Възнесение', 'blo': 'Asɛnsiyɔɔn kaAtukǝltǝna', 'bn': 'অ্যাসেনশন দ্বীপপুঞ্জ', 'br': 'Enez Ascension', 'brx': 'एसेनसिअन द्वीप', 'bs': 'Ostrvo Ascension', 'bs-Cyrl': 'Острво Асенсион', 'bs-Latn': 'Ostrvo Ascension', 'ca': 'Illa de l’Ascensió', 'ca-ES-valencia': 'Illa de l’Ascensió', 'ccp': '𑄃𑄳𑄠𑄥𑄴𑄥𑄬𑄚𑄴𑄥𑄧𑄚𑄴 𑄃𑄭𑄣𑄳𑄠𑄚𑄴𑄓𑄴', 'ce': 'Айъадаларан гӀайре', 'ceb': 'Ascension Island', 'chr': 'ᎤᎵᏌᎳᏓᏅ ᎤᎦᏚᏛᎢ', 'ckb': 'دوورگەی ئاسینسیۆن', 'cs': 'Ascension', 'cv': 'Вознесени утравӗ', 'cy': 'Ynys Ascension', 'da': 'Ascensionøen', 'de': 'Ascension', 'dsb': 'Ascension', 'dz': 'ཨེ་སེན་ཤུན་ཚོ་གླིང༌', 'ee': 'Ascension ƒudomekpo nutome', 'el': 'Νήσος Ασενσιόν', 'el-polyton': 'Νήσος Ασενσιόν', 'en': 'Ascension Island', 'es': 'Isla de la Ascensión', 'et': 'Ascensioni saar', 'eu': 'Ascension uhartea', 'fa': 'جزایر آسنسیون', 'ff-Adlm': '𞤅𞤵𞤪𞤭𞥅𞤪𞤫 𞤀𞤧𞤢𞤲𞤧𞤮𞥅𞤲', 'fi': 'Ascension-saari', 'fil': 'Acsencion island', 'fo': 'Ascension', 'fr': 'Île de l’Ascension', 'fy': 'Ascension', 'ga': 'Oileán na Deascabhála', 'gd': 'Eilean na Deasgabhalach', 'gl': 'Illa de Ascensión', 'gu': 'એસેન્શન આઇલેન્ડ', 'ha': 'Tsibirin Ascension', 'he': 'האי אסנשן', 'hi': 'असेंशन द्वीप', 'hi-Latn': 'Ascension Island', 'hr': 'Otok Uzašašća', 'hsb': 'Ascension', 'ht': 'Île de l’Ascension', 'hu': 'Ascension-sziget', 'hy': 'Համբարձման կղզի', 'ia': 'Insula del Ascension', 'id': 'Pulau Ascension', 'ig': 'Ascension Island', 'is': 'Ascension-eyja', 'it': 'Isola Ascensione', 'ja': 'アセンション島', 'jv': 'Pulo Ascension', 'ka': 'ამაღლების კუნძული', 'kea': 'Ilha di Asenson', 'kgp': 'Asẽnsỹv Goj-vẽso', 'kk': 'Әскенжін аралы', 'kk-Cyrl': 'Әскенжін аралы', 'km': 'កោះ\u200bអាសេនសិន', 'kn': 'ಅಸೆನ್ಶನ್ ದ್ವೀಪ', 'ko': '어센션섬', 'kok': 'असेंशन आयलँड', 'kok-Deva': 'असेंशन आयलँड', 'ks': 'ایسنشن جزیرٕ', 'ks-Arab': 'ایسنشن جزیرٕ', 'ksh': 'Aßension', 'ku': 'Girava Ascensionê', 'ky': 'Вознесение аралы', 'lb': 'Ascension', 'lld': 'Isola dl’Ascensiun', 'lo': 'ເກາະອາເຊນຊັນ', 'lt': 'Dangun Žengimo sala', 'lv': 'Debesbraukšanas sala', 'mai': 'असेंशन द्वीप', 'mi': 'Te Moutere Aupikinga', 'mk': 'Остров Асенсион', 'ml': 'അസൻഷൻ ദ്വീപ്', 'mn': 'Асенсион арал', 'mr': 'अ\u200dॅसेन्शियन बेट', 'ms': 'Pulau Ascension', 'mt': 'Ascension Island', 'my': 'အဆန်းရှင်းကျွန်း', 'mzn': 'آسنسیون جزیره', 'nb': 'Ascension', 'ne': 'असेन्सन टापु', 'nl': 'Ascension', 'nn': 'Ascension', 'no': 'Ascension', 'nqo': 'ߊߛߊ߲ߛߌߦߐ߲߫ ߕߌ߲', 'om': 'Odola Asenshiin', 'or': 'ଆସେନସିଅନ୍\u200c ଦ୍ୱୀପ', 'pa': 'ਅਸੈਂਸ਼ਨ ਟਾਪੂ', 'pa-Guru': 'ਅਸੈਂਸ਼ਨ ਟਾਪੂ', 'pcm': 'Asẹ́nshọ́n Aíland', 'pl': 'Wyspa Wniebowstąpienia', 'ps': 'اسينشان ټاپو', 'pt': 'Ilha de Ascensão', 'qu': 'Islas Ascensión', 'rm': 'Insla d’Ascensiun', 'ro': 'Insula Ascension', 'ru': 'о-в Вознесения', 'sc': 'Ìsula de s’Ascensione', 'sd': 'طلوع ٻيٽ', 'sd-Arab': 'طلوع ٻيٽ', 'se': 'Ascension', 'si': 'ඇසෙන්ෂන් දිවයින', 'sk': 'Ascension', 'sl': 'Otok Ascension', 'smn': 'Ascension-suálui', 'so': 'Jasiiradda Asensiyoon', 'sq': 'Ishulli Asenshion', 'sr': 'Острво Асенсион', 'sr-Cyrl': 'Острво Асенсион', 'sr-Latn': 'Ostrvo Asension', 'sv': 'Ascension', 'sw': 'Kisiwa cha Ascension', 'ta': 'அஷன்ஷியன் தீவு', 'te': 'అసెన్షన్ దీవి', 'tg': 'Асунсон', 'th': 'เกาะแอสเซนชัน', 'ti': 'ደሴት ኣሰንስዮን', 'tk': 'Beýgeliş adasy', 'to': 'Motu ʻAsenisini', 'tr': 'Ascension Adası', 'tt': 'Вознесение утравы', 'ug': 'ئاسسېنسىيون ئارىلى', 'uk': 'Острів Вознесіння', 'ur': 'اسینشن آئلینڈ', 'uz': 'Me’roj oroli', 'uz-Cyrl': 'Меърож ороли', 'uz-Latn': 'Me’roj oroli', 'vai': 'ꗻꗡ ꕒꕡꕌ ꗏ ꔳꘋꗣ', 'vai-Vaii': 'ꗻꗡ ꕒꕡꕌ ꗏ ꔳꘋꗣ', 'vec': 'Ìzola Asension', 'vi': 'Đảo Ascension', 'wae': 'Himmelfártsinsla', 'wo': 'Ile Ascension', 'xh': 'E-Ascension Island', 'xnr': 'असेंशन द्वीप', 'yo': 'Erékùsù Ascension', 'yrl': 'Asesãu Kapuãma', 'yue': '阿森松島', 'yue-Hans': '阿森松岛', 'yue-Hant': '阿森松島', 'zh': '阿森松岛', 'zh-Hans': '阿森松岛', 'zh-Hant': '阿森松島', 'zu': 'i-Ascension Island'},
    'AD': {'af': 'Andorra', 'agq': 'Àndolà', 'ak': 'Andora', 'am': 'አንዶራ', 'ar': 'أندورا', 'as': 'আন্দোৰা', 'asa': 'Andora', 'ast': 'Andorra', 'az': 'Andorra', 'az-Cyrl': 'Андорра', 'az-Latn': 'Andorra', 'bal-Latn': 'Andorrá', 'bas': 'Àŋdɔ̂r', 'be': 'Андора', 'be-tarask': 'Андора', 'bez': 'Huandola', 'bg': 'Андора', 'blo': 'Anɖɔraa', 'bm': 'Andɔr', 'bn': 'আন্ডোরা', 'br': 'Andorra', 'brx': 'एन्डर्रा', 'bs': 'Andora', 'bs-Cyrl': 'Андора', 'bs-Latn': 'Andora', 'ca': 'Andorra', 'ca-ES-valencia': 'Andorra', 'ccp': '𑄃𑄚𑄴𑄓𑄮𑄢', 'ce': 'Андорра', 'ceb': 'Andorra', 'cgg': 'Andora', 'chr': 'ᎠᏂᏙᎳ', 'ckb': 'ئاندۆرا', 'cs': 'Andorra', 'cv': 'Андорра', 'cy': 'Andorra', 'da': 'Andorra', 'dav': 'Andora', 'de': 'Andorra', 'dje': 'Andoora', 'dsb': 'Andorra', 'dyo': 'Andorra', 'dz': 'ཨཱན་དོ་ར', 'ebu': 'Andora', 'ee': 'Andorra nutome', 'el': 'Ανδόρα', 'el-polyton': 'Ἀνδόρα', 'en': 'Andorra', 'en-Dsrt': '𐐈𐑌𐐼𐐱𐑉𐐲', 'eo': 'Andoro', 'es': 'Andorra', 'et': 'Andorra', 'eu': 'Andorra', 'ewo': 'Andór', 'fa': 'آندورا', 'ff': 'Anndoora', 'ff-Adlm': '𞤀𞤲𞤣𞤮𞤪𞤢𞥄', 'ff-Latn': 'Anndoora', 'fi': 'Andorra', 'fil': 'Andorra', 'fo': 'Andorra', 'fr': 'Andorre', 'fur': 'Andorra', 'fy': 'Andorra', 'ga': 'Andóra', 'gd': 'Andorra', 'gl': 'Andorra', 'gsw': 'Andorra', 'gu': 'ઍંડોરા', 'guz': 'Andora', 'ha': 'Andora', 'he': 'אנדורה', 'hi': 'एंडोरा', 'hi-Latn': 'Andorra', 'hr': 'Andora', 'hsb': 'Andorra', 'ht': 'Andorre', 'hu': 'Andorra', 'hy': 'Անդորրա', 'ia': 'Andorra', 'id': 'Andorra', 'ig': 'Andorra', 'is': 'Andorra', 'it': 'Andorra', 'ja': 'アンドラ', 'jmc': 'Andora', 'jv': 'Andora', 'ka': 'ანდორა', 'kaa': 'Андорра', 'kaa-Cyrl': 'Андорра', 'kab': 'Undura', 'kam': 'Andora', 'kde': 'Andola', 'kea': 'Andora', 'kgp': 'Ỹnora', 'khq': 'Andoora', 'ki': 'Andora', 'kk': 'Андорра', 'kk-Cyrl': 'Андорра', 'kln': 'Emetab Andorra', 'km': 'អង់ដូរ៉ា', 'kn': 'ಅಂಡೋರಾ', 'ko': '안도라', 'kok': 'अंडोरा', 'kok-Deva': 'अंडोरा', 'ks': 'اینڈورا', 'ks-Arab': 'اینڈورا', 'ksb': 'Andola', 'ksf': 'andɔrǝ', 'ksh': 'Andorra', 'ku': 'Andorra', 'kxv': 'ānḍorā', 'kxv-Deva': 'अंडोरा', 'kxv-Latn': 'ānḍorā', 'kxv-Orya': 'ଆଣ୍ଡୋରା', 'kxv-Telu': 'ఆండొర', 'ky': 'Андорра', 'lag': 'Andóra', 'lb': 'Andorra', 'lg': 'Andora', 'lld': 'Andorra', 'ln': 'Andorɛ', 'lo': 'ອັນດໍຣາ', 'lt': 'Andora', 'lu': 'Andore', 'luo': 'Andorra', 'luy': 'Andora', 'lv': 'Andora', 'mai': 'एंडोरा', 'mas': 'Andora', 'mer': 'Andora', 'mfe': 'Andor', 'mg': 'Andorra', 'mgh': 'Uandora', 'mi': 'Anatōra', 'mk': 'Андора', 'ml': 'അൻഡോറ', 'mn': 'Андорра', 'mr': 'अँडोरा', 'ms': 'Andorra', 'mt': 'Andorra', 'mua': 'andorra', 'my': 'အန်ဒိုရာ', 'mzn': 'آندورا', 'naq': 'Andorrab', 'nb': 'Andorra', 'nd': 'Andora', 'ne': 'अन्डोर्रा', 'nl': 'Andorra', 'nmg': 'Andɔ́ra', 'nn': 'Andorra', 'no': 'Andorra', 'nqo': 'ߊ߲ߘߐߙ', 'nus': 'Andora', 'nyn': 'Andora', 'om': 'Andooraa', 'or': 'ଆଣ୍ଡୋରା', 'pa': 'ਅੰਡੋਰਾ', 'pa-Guru': 'ਅੰਡੋਰਾ', 'pcm': 'Andọ́ra', 'pl': 'Andora', 'ps': 'اندورا', 'pt': 'Andorra', 'qu': 'Andorra', 'rm': 'Andorra', 'rn': 'Andora', 'ro': 'Andorra', 'rof': 'Andoro', 'ru': 'Андорра', 'rwk': 'Andora', 'saq': 'Andora', 'sbp': 'Andola', 'sc': 'Andorra', 'sd': 'اندورا', 'sd-Arab': 'اندورا', 'se': 'Andorra', 'seh': 'Andorra', 'ses': 'Andoora', 'sg': 'Andôro', 'shi': 'ⴰⵏⴷⵓⵔⴰ', 'shi-Latn': 'andura', 'shi-Tfng': 'ⴰⵏⴷⵓⵔⴰ', 'si': 'ඇන්ඩෝරාව', 'sk': 'Andorra', 'sl': 'Andora', 'smn': 'Andorra', 'sn': 'Andora', 'so': 'Andora', 'sq': 'Andorrë', 'sr': 'Андора', 'sr-Cyrl': 'Андора', 'sr-Latn': 'Andora', 'sv': 'Andorra', 'sw': 'Andorra', 'syr': 'ܐܢܕܘܪܐ', 'ta': 'அன்டோரா', 'te': 'ఆండోరా', 'teo': 'Andora', 'tg': 'Андорра', 'th': 'อันดอร์รา', 'ti': 'ኣንዶራ', 'tk': 'Andorra', 'to': 'ʻAnitola', 'tr': 'Andorra', 'tt': 'Андорра', 'twq': 'Andoora', 'tzm': 'Anḍurra', 'ug': 'ئاندوررا', 'uk': 'Андорра', 'ur': 'انڈورا', 'uz': 'Andorra', 'uz-Cyrl': 'Андорра', 'uz-Latn': 'Andorra', 'vai': 'ꕉꖆꕟ', 'vai-Latn': 'Aŋdóra', 'vai-Vaii': 'ꕉꖆꕟ', 'vec': 'Andora', 'vi': 'Andorra', 'vun': 'Andora', 'wae': 'Andorra', 'wo': 'Andoor', 'xh': 'E-Andorra', 'xnr': 'एंडोरा', 'xog': 'Andora', 'yav': 'Aŋtúla', 'yi': 'אַנדארע', 'yo': 'Ààndórà', 'yrl': 'Ãdura', 'yue': '安道爾', 'yue-Hans': '安道尔', 'yue-Hant': '安道爾', 'zgh': 'ⴰⵏⴷⵓⵔⴰ', 'zh': '安道尔', 'zh-Hans': '安道尔', 'zh-Hant': '安道爾', 'zu': 'i-Andorra'},
    'AE': {'af': 'Verenigde Arabiese Emirate', 'agq': 'Yùnaetɛ Alab ɛmelɛ̀', 'ak': 'United Arab Emirates', 'am': 'የተባበሩት ዓረብ ኤምሬትስ', 'ar': 'الإمارات العربية المتحدة', 'as': 'সংযুক্ত আৰব আমিৰাত', 'asa': 'Falme dha Kiarabu', 'ast': 'Emiratos Árabes Xuníos', 'az': 'Birləşmiş Ərəb Əmirlikləri', 'az-Cyrl': 'Бирләшмиш Әрәб Әмирликләри', 'az-Latn': 'Birləşmiş Ərəb Əmirlikləri', 'bal-Latn': 'Emárát', 'bas': 'Àdnà i Bilɔ̀ŋ bi Arābìà', 'be': 'Аб’яднаныя Арабскія Эміраты', 'be-tarask': 'Аб’яднаныя Арабскія Эміраты', 'bez': 'Hufalme dza Hihalabu', 'bg': 'Обединени арабски емирства', 'bgn': 'متحدین عربین امارات', 'blo': 'Emiir baGanɔ gaɖɔŋkɔnɔ kaAlaaributǝna', 'bm': 'Arabu mara kafoli', 'bn': 'সংযুক্ত আরব আমিরাত', 'br': 'Emirelezhioù Arab Unanet', 'brx': 'जथाय आराब एमिरेत', 'bs': 'Ujedinjeni Arapski Emirati', 'bs-Cyrl': 'Уједињени Арапски Емирати', 'bs-Latn': 'Ujedinjeni Arapski Emirati', 'ca': 'Emirats Àrabs Units', 'ca-ES-valencia': 'Emirats Àrabs Units', 'ccp': '𑄎𑄧𑄙 𑄃𑄢𑄧𑄝𑄴 𑄃𑄟𑄨𑄢𑄖𑄴', 'ce': 'Ӏарбийн Цхьанатоьхна Эмираташ', 'ceb': 'United Arab Emirates', 'cgg': 'Amahanga ga Buharabu ageeteereine', 'chr': 'ᏌᏊ ᎢᏳᎾᎵᏍᏔᏅ ᎡᎳᏈ ᎢᎹᎵᏘᏏ', 'ckb': 'میرنشینە یەکگرتووە عەرەبییەکان', 'cs': 'Spojené arabské emiráty', 'cv': 'Арапсен Пӗрлешӳллӗ Эмирачӗ', 'cy': 'Emiradau Arabaidd Unedig', 'da': 'De Forenede Arabiske Emirater', 'dav': 'Falme za Kiarabu', 'de': 'Vereinigte Arabische Emirate', 'dje': 'Laaraw Imaarawey Margantey', 'dsb': 'Zjadnośone arabiske emiraty', 'dz': 'ཡུ་ནཱའི་ཊེཌ་ ཨ་རབ་ ཨེ་མེ་རེཊས', 'ebu': 'Falme za Kiarabu', 'ee': 'United Arab Emirates nutome', 'el': 'Ηνωμένα Αραβικά Εμιράτα', 'el-polyton': 'Ἠνωμένα Ἀραβικὰ Ἐμιράτα', 'en': 'United Arab Emirates', 'en-Dsrt': '𐐏𐐭𐑌𐐴𐐼𐐮𐐼 𐐇𐑉𐐲𐐺 𐐇𐑋𐐲𐑉𐐩𐐻𐑅', 'eo': 'Unuiĝintaj Arabaj Emirlandoj', 'es': 'Emiratos Árabes Unidos', 'et': 'Araabia Ühendemiraadid', 'eu': 'Arabiar Emirerri Batuak', 'ewo': 'Bemirá yá Arábə uní', 'fa': 'امارات متحدهٔ عربی', 'ff': 'Emiraat Araab Denntuɗe', 'ff-Adlm': '𞤁𞤫𞤲𞤼𞤢𞤤 𞤋𞤥𞤪𞤢𞥄𞤼𞤭 𞤀𞥄𞤪𞤢𞤦𞤵', 'ff-Latn': 'Emiraat Araab Denntuɗe', 'fi': 'Arabiemiirikunnat', 'fil': 'United Arab Emirates', 'fo': 'Sameindu Emirríkini', 'fr': 'Émirats arabes unis', 'fur': 'Emirâts araps unîts', 'fy': 'Verenigde Arabyske Emiraten', 'ga': 'Aontas na nÉimíríochtaí Arabacha', 'gd': 'Na h-Iomaratan Arabach Aonaichte', 'gl': 'Emiratos Árabes Unidos', 'gsw': 'Veräinigti Arabischi Emirate', 'gu': 'યુનાઇટેડ આરબ અમીરાત', 'guz': 'Falme za Kiarabu', 'ha': 'Haɗaɗɗiyar Daular Larabawa', 'he': 'איחוד האמירויות הערביות', 'hi': 'संयुक्त अरब अमीरात', 'hi-Latn': 'United Arab Emirates', 'hr': 'Ujedinjeni Arapski Emirati', 'hsb': 'Zjednoćene arabske emiraty', 'ht': 'Émirats arabes unis', 'hu': 'Egyesült Arab Emírségek', 'hy': 'Արաբական Միացյալ Էմիրություններ', 'ia': 'Emiratos Arabe Unite', 'id': 'Uni Emirat Arab', 'ig': 'United Arab Emirates', 'is': 'Sameinuðu arabísku furstadæmin', 'it': 'Emirati Arabi Uniti', 'ja': 'アラブ首長国連邦', 'jmc': 'Falme za Kiarabu', 'jv': 'Uni Émirat Arab', 'ka': 'არაბთა გაერთიანებული საამიროები', 'kaa': 'Бирлескен Араб Әмирликлери', 'kaa-Cyrl': 'Бирлескен Араб Әмирликлери', 'kab': 'Tigeldunin Yedduklen Taɛrabin', 'kam': 'Falme za Kiarabu', 'kde': 'Dimiliki dya Vakulungwa va Chalabu', 'kea': 'Emiradus Árabi Unidu', 'kgp': 'Emĩrano Arame Unĩno', 'khq': 'Laaraw Imaarawey Margantey', 'ki': 'Falme za Kiarabu', 'kk': 'Біріккен Араб Әмірліктері', 'kk-Cyrl': 'Біріккен Араб Әмірліктері', 'kln': 'Emetab kibagenge nebo arabuk', 'km': 'អេមីរ៉ាត\u200bអារ៉ាប់\u200bរួម', 'kn': 'ಯುನೈಟೆಡ್ ಅರಬ್ ಎಮಿರೇಟ್ಸ್', 'ko': '아랍에미리트', 'kok': 'युनाइटेड अरब इमीरात', 'kok-Deva': 'युनाइटेड अरब इमीरात', 'ks': 'مُتحدہ عرَب امارات', 'ks-Arab': 'مُتحدہ عرَب امارات', 'ksb': 'Falme za Kialabu', 'ksf': 'bǝlɔŋ bǝ kaksa bɛ táatáaŋzǝn', 'ksh': 'Vereinschte Arrabesche Emmirate', 'ku': 'Mîrgehên Erebî yên Yekbûyî', 'kxv': 'aanḍiti arab emireṭs', 'kxv-Deva': 'आण्डी ति अरब एमिरेट्स', 'kxv-Latn': 'aanḍiti arab emireṭs', 'kxv-Orya': 'ଆଣ୍ଡିତି ଆରବ ଏମିରେଟସ୍', 'kxv-Telu': 'యునైటెడ్ ఆరబ్ ఎమిరేబ్స్', 'ky': 'Бириккен Араб Эмираттары', 'lag': 'Ʉtemi wa Kɨaráabu', 'lb': 'Vereenegt Arabesch Emirater', 'lg': 'Emireeti', 'lld': 'Emirac Arabics Unis', 'ln': 'Lɛmila alabo', 'lo': 'ສະຫະລັດອາຣັບເອມິເຣດ', 'lt': 'Jungtiniai Arabų Emyratai', 'lu': 'Lemila alabu', 'luo': 'United Arab Emirates', 'luy': 'Falme za Kiarabu', 'lv': 'Apvienotie Arābu Emirāti', 'mai': 'संयुक्त अरब अमीरात', 'mas': 'Falme za Kiarabu', 'mer': 'Falme cia Kiarabu', 'mfe': 'Emira arab ini', 'mg': 'Emirà Arabo mitambatra', 'mi': 'Kotahitanga o ngā Whenua o Ārapi', 'mk': 'Обединети Арапски Емирати', 'ml': 'യുണൈറ്റഡ് അറബ് എമിറൈറ്റ്\u200cസ്', 'mn': 'Арабын Нэгдсэн Эмират Улс', 'mr': 'संयुक्त अरब अमीरात', 'ms': 'Emiriah Arab Bersatu', 'mt': 'l-Emirati Għarab Magħquda', 'mua': 'Sǝr Arabiya ma taini', 'my': 'ယူအေအီး', 'mzn': 'متحده عربی امارات', 'naq': 'United Arab Emirates', 'nb': 'De forente arabiske emirater', 'nd': 'United Arab Emirates', 'ne': 'संयुक्त अरब इमिराट्स', 'nl': 'Verenigde Arabische Emiraten', 'nmg': 'Minlambɔ́ Nsaŋ́nsa mí Arabia', 'nn': 'Dei sameinte arabiske emirata', 'no': 'De forente arabiske emirater', 'nqo': 'ߋߡߌߙߊߕ ߊߙߊߓߎ߫ ߘߍ߬ߣߍ߲', 'nyn': 'Amahanga ga Buharabu ageeteereine', 'om': 'Yuunaatid Arab Emereet', 'or': 'ସଂଯୁକ୍ତ ଆରବ ଏମିରେଟସ୍', 'pa': 'ਸੰਯੁਕਤ ਅਰਬ ਅਮੀਰਾਤ', 'pa-Guru': 'ਸੰਯੁਕਤ ਅਰਬ ਅਮੀਰਾਤ', 'pcm': 'Yunaítẹ́d Áráb Ẹ́mírets', 'pl': 'Zjednoczone Emiraty Arabskie', 'ps': 'متحده عرب امارات', 'pt': 'Emirados Árabes Unidos', 'qu': 'Emiratos Árabes Unidos', 'rm': 'Emirats Arabs Unids', 'rn': 'Leta Zunze Ubumwe z’Abarabu', 'ro': 'Emiratele Arabe Unite', 'rof': 'Falme za Kiarabu', 'ru': 'ОАЭ', 'rwk': 'Falme za Kiarabu', 'saq': 'Falme za Kiarabu', 'sbp': 'Wutwa wa shiyalabu', 'sc': 'Emirados Àrabos Unidos', 'sd': 'متحده عرب امارات', 'sd-Arab': 'متحده عرب امارات', 'se': 'Ovttastuvvan Arábaemiráhtat', 'seh': 'Emirados Árabes Unidos', 'ses': 'Laaraw Imaarawey Margantey', 'sg': 'Arâbo Emirâti Ôko', 'shi': 'ⵍⵉⵎⴰⵔⴰⵜ', 'shi-Latn': 'limarat', 'shi-Tfng': 'ⵍⵉⵎⴰⵔⴰⵜ', 'si': 'එක්සත් අරාබි එමිර් රාජ්\u200dයය', 'sk': 'Spojené arabské emiráty', 'sl': 'Združeni arabski emirati', 'smn': 'Arabiemirkodeh', 'sn': 'United Arab Emirates', 'so': 'Midawga Imaaraatka Carabta', 'sq': 'Emiratet e Bashkuara Arabe', 'sr': 'Уједињени Арапски Емирати', 'sr-Cyrl': 'Уједињени Арапски Емирати', 'sr-Latn': 'Ujedinjeni Arapski Emirati', 'sv': 'Förenade Arabemiraten', 'sw': 'Falme za Kiarabu', 'syr': 'ܐܡܝܪ̈ܘܬܐ ܡܚܝܕ̈ܬܐ ܥܪ̈ܒܝܐ', 'ta': 'ஐக்கிய அரபு எமிரேட்ஸ்', 'te': 'యునైటెడ్ అరబ్ ఎమిరేట్స్', 'teo': 'Falme za Kiarabu', 'tg': 'Аморатҳои Муттаҳидаи Араб', 'th': 'สหรัฐอาหรับเอมิเรตส์', 'ti': 'ሕቡራት ኢማራት ዓረብ', 'tk': 'Birleşen Arap Emirlikleri', 'to': 'ʻAlepea Fakatahataha', 'tr': 'Birleşik Arap Emirlikleri', 'tt': 'Берләшкән Гарәп Әмирлекләре', 'twq': 'Laaraw Imaarawey Margantey', 'tzm': 'Imarat Tiεrabin Tidduklin', 'ug': 'ئەرەب بىرلەشمە خەلىپىلىكى', 'uk': 'Обʼєднані Арабські Емірати', 'ur': 'متحدہ عرب امارات', 'uz': 'Birlashgan Arab Amirliklari', 'uz-Cyrl': 'Бирлашган Араб Амирликлари', 'uz-Latn': 'Birlashgan Arab Amirliklari', 'vai': 'ꖳꕯꔤꗳ ꕉꕟꔬ ꗡꕆꔓꔻ', 'vai-Latn': 'Yunaitɛ Arabhi Ɛmire', 'vai-Vaii': 'ꖳꕯꔤꗳ ꕉꕟꔬ ꗡꕆꔓꔻ', 'vec': 'Emirài Àrabi Unìi', 'vi': 'Các Tiểu Vương quốc Ả Rập Thống nhất', 'vun': 'Falme za Kiarabu', 'wae': 'Vereinigti Arabiše Emirat', 'wo': 'Emira Arab Ini', 'xh': 'E-United Arab Emirates', 'xnr': 'संयुक्त अरब अमीरात', 'xog': 'Emireeti', 'yav': 'imiláat i paaláap', 'yo': 'Ẹmirate ti Awọn Arabu', 'yrl': 'Emiradu Árabi Yepewasuwaitá', 'yue': '阿拉伯聯合大公國', 'yue-Hans': '阿拉伯联合大公国', 'yue-Hant': '阿拉伯聯合大公國', 'zgh': 'ⵍⵉⵎⴰⵔⴰⵜ', 'zh': '阿拉伯联合酋长国', 'zh-Hans': '阿拉伯联合酋长国', 'zh-Hant': '阿拉伯聯合大公國', 'zu': 'i-United Arab Emirates'},
    'AF': {'af': 'Afganistan', 'agq': 'Àfɨ̀ganìsɨ̀tân', 'ak': 'Afganistan', 'am': 'አፍጋኒስታን', 'ar': 'أفغانستان', 'as': 'আফগানিস্তান', 'asa': 'Afuganistani', 'ast': 'Afganistán', 'az': 'Əfqanıstan', 'az-Cyrl': 'Әфганыстан', 'az-Latn': 'Əfqanıstan', 'bal-Latn': 'Awghánestán', 'bas': 'Àfgànìstâŋ', 'be': 'Афганістан', 'be-tarask': 'Афганістан', 'bez': 'Huafuganistani', 'bg': 'Афганистан', 'blo': 'Afganistan', 'bm': 'Afiganistaŋ', 'bn': 'আফগানিস্তান', 'br': 'Afghanistan', 'brx': 'आफगानिस्तान', 'bs': 'Afganistan', 'bs-Cyrl': 'Афганистан', 'bs-Latn': 'Afganistan', 'ca': 'Afganistan', 'ca-ES-valencia': 'Afganistan', 'ccp': '𑄃𑄛𑄴𑄉𑄚𑄨𑄌𑄴𑄖𑄚𑄴', 'ce': 'ОвхӀан мохк', 'ceb': 'Afghanistan', 'cgg': 'Afuganistani', 'chr': 'ᎠᏫᎨᏂᏍᏖᏂ', 'ckb': 'ئەفغانستان', 'cs': 'Afghánistán', 'cv': 'Афганистан', 'cy': 'Afghanistan', 'da': 'Afghanistan', 'dav': 'Afuganistani', 'de': 'Afghanistan', 'dje': 'Afgaanistan', 'dsb': 'Afghanistan', 'dyo': 'Afganistan', 'dz': 'ཨཕ་གྷ་ནི་སཏཱན', 'ebu': 'Afuganistani', 'ee': 'Afghanistan nutome', 'el': 'Αφγανιστάν', 'el-polyton': 'Ἀφγανιστάν', 'en': 'Afghanistan', 'en-Dsrt': '𐐈𐑁𐑀𐐰𐑌𐐲𐑅𐐻𐐰𐑌', 'eo': 'Afganujo', 'es': 'Afganistán', 'et': 'Afganistan', 'eu': 'Afganistan', 'ewo': 'Afəganisətán', 'fa': 'افغانستان', 'ff': 'Afganistaan', 'ff-Adlm': '𞤀𞤬𞤺𞤢𞤲𞤭𞤧𞤼𞤢𞥄𞤲', 'ff-Latn': 'Afganistaan', 'fi': 'Afganistan', 'fil': 'Afghanistan', 'fo': 'Afganistan', 'fr': 'Afghanistan', 'fur': 'Afghanistan', 'fy': 'Afghanistan', 'ga': 'an Afganastáin', 'gd': 'Afghanastàn', 'gl': 'Afganistán', 'gsw': 'Afganischtan', 'gu': 'અફઘાનિસ્તાન', 'guz': 'Afuganistani', 'ha': 'Afaganistan', 'he': 'אפגניסטן', 'hi': 'अफ़गानिस्तान', 'hi-Latn': 'Afghanistan', 'hr': 'Afganistan', 'hsb': 'Afghanistan', 'ht': 'Afghanistan', 'hu': 'Afganisztán', 'hy': 'Աֆղանստան', 'ia': 'Afghanistan', 'id': 'Afganistan', 'ig': 'Afghanistan', 'is': 'Afganistan', 'it': 'Afghanistan', 'ja': 'アフガニスタン', 'jmc': 'Afuganistani', 'jv': 'Afganistan', 'ka': 'ავღანეთი', 'kaa': 'Аўғанстан', 'kaa-Cyrl': 'Аўғанстан', 'kab': 'Afɣanistan', 'kam': 'Afuganistani', 'kde': 'Afuganistani', 'kea': 'Afeganistãu', 'kgp': 'Afeganĩtã', 'khq': 'Afgaanistan', 'ki': 'Afuganistani', 'kk': 'Ауғанстан', 'kk-Cyrl': 'Ауғанстан', 'kln': 'Emetab Afghanistan', 'km': 'អាហ្វហ្កានីស្ថាន', 'kn': 'ಅಫಘಾನಿಸ್ಥಾನ', 'ko': '아프가니스탄', 'kok': 'अफगानिस्तान', 'kok-Deva': 'अफगानिस्तान', 'ks': 'اَفغانَستان', 'ks-Arab': 'اَفغانَستان', 'ksb': 'Afuganistani', 'ksf': 'afganistáŋ', 'ksh': 'Afjaanistahn', 'ku': 'Efxanistan', 'kxv': 'aapganistan', 'kxv-Deva': 'अप़गानिस्तान', 'kxv-Latn': 'aapganistan', 'kxv-Orya': 'ଆପଗାନିସ୍ତାନ୍', 'kxv-Telu': 'ఆప్గనిస్తాన్', 'ky': 'Афганистан', 'lag': 'Afuganisitáani', 'lb': 'Afghanistan', 'lg': 'Afaganisitani', 'lld': 'Afghanistan', 'ln': 'Afiganisitá', 'lo': 'ອາຟການິດສະຖານ', 'lt': 'Afganistanas', 'lu': 'Afuganisita', 'luo': 'Afghanistan', 'luy': 'Afuganistani', 'lv': 'Afganistāna', 'mai': 'अफगानिस्तान', 'mas': 'Afuganistani', 'mer': 'Afuganistani', 'mfe': 'Afganistan', 'mg': 'Afghanistan', 'mgh': 'Ufugustani', 'mi': 'Awhekenetāna', 'mk': 'Авганистан', 'ml': 'അഫ്\u200cഗാനിസ്ഥാൻ', 'mn': 'Афганистан', 'mr': 'अफगाणिस्तान', 'ms': 'Afghanistan', 'mt': 'l-Afganistan', 'mua': 'afghanistaŋ', 'my': 'အာဖဂန်နစ္စတန်', 'mzn': 'افغانستون', 'naq': 'Afghanistanni', 'nb': 'Afghanistan', 'nd': 'Afghanistan', 'ne': 'अफगानिस्तान', 'nl': 'Afghanistan', 'nmg': 'Afganistaŋ', 'nn': 'Afghanistan', 'no': 'Afghanistan', 'nqo': 'ߊߝߎߜ߭ߊߣߌߛߑߕߊ߲߫', 'nus': 'Abganithtan', 'nyn': 'Afuganistani', 'om': 'Afgaanistaan', 'or': 'ଆଫଗାନିସ୍ତାନ୍', 'pa': 'ਅਫ਼ਗਾਨਿਸਤਾਨ', 'pa-Guru': 'ਅਫ਼ਗਾਨਿਸਤਾਨ', 'pcm': 'Afgánístan', 'pl': 'Afganistan', 'ps': 'افغانستان', 'pt': 'Afeganistão', 'qu': 'Afganistán', 'rm': 'Afghanistan', 'rn': 'Afuganisitani', 'ro': 'Afganistan', 'rof': 'Afuganistani', 'ru': 'Афганистан', 'rwk': 'Afuganistani', 'saq': 'Afuganistani', 'sbp': 'Afuganisitani', 'sc': 'Afghànistan', 'sd': 'افغانستان', 'sd-Arab': 'افغانستان', 'se': 'Afghanistan', 'seh': 'Afeganistão', 'ses': 'Afgaanistan', 'sg': 'Faganïta, Afganïstäan', 'shi': 'ⴰⴼⵖⴰⵏⵉⵙⵜⴰⵏ', 'shi-Latn': 'afɣanistan', 'shi-Tfng': 'ⴰⴼⵖⴰⵏⵉⵙⵜⴰⵏ', 'si': 'ඇෆ්ගනිස්ථානය', 'sk': 'Afganistan', 'sl': 'Afganistan', 'smn': 'Afganistan', 'sn': 'Afuganistani', 'so': 'Afgaanistaan', 'sq': 'Afganistan', 'sr': 'Авганистан', 'sr-Cyrl': 'Авганистан', 'sr-Latn': 'Avganistan', 'sv': 'Afghanistan', 'sw': 'Afghanistan', 'syr': 'ܐܦܓܐܢܣܬܐܢ', 'ta': 'ஆஃப்கானிஸ்தான்', 'te': 'ఆఫ్ఘనిస్తాన్', 'teo': 'Afuganistani', 'tg': 'Афғонистон', 'th': 'อัฟกานิสถาน', 'ti': 'ኣፍጋኒስታን', 'tk': 'Owganystan', 'to': 'ʻAfikānisitani', 'tr': 'Afganistan', 'tt': 'Әфганстан', 'twq': 'Afgaanistan', 'tzm': 'Afɣanistan', 'ug': 'ئافغانىستان', 'uk': 'Афганістан', 'ur': 'افغانستان', 'uz': 'Afgʻoniston', 'uz-Arab': 'افغانستان', 'uz-Cyrl': 'Афғонистон', 'uz-Latn': 'Afgʻoniston', 'vai': 'ꕉꔱꕭꔕꔻꕚꘋ', 'vai-Latn': 'Afigándesitaŋ', 'vai-Vaii': 'ꕉꔱꕭꔕꔻꕚꘋ', 'vec': 'Afghanistan', 'vi': 'Afghanistan', 'vun': 'Afuganistani', 'wae': 'Afganištan', 'wo': 'Afganistaŋ', 'xh': 'E-Afghanistan', 'xnr': 'अफ़गानिस्तान', 'xog': 'Afaganisitani', 'yav': 'Afkanistáŋ', 'yi': 'אַפֿגהאַניסטאַן', 'yo': 'Àfùgànístánì', 'yrl': 'Afegãniretãma', 'yue': '阿富汗', 'yue-Hans': '阿富汗', 'yue-Hant': '阿富汗', 'zgh': 'ⴰⴼⵖⴰⵏⵉⵙⵜⴰⵏ', 'zh': '阿富汗', 'zh-Hans': '阿富汗', 'zh-Hant': '阿富汗', 'zu': 'i-Afghanistan'},
    'AG': {'af': 'Antigua en Barbuda', 'agq': 'Àntigwà à Bàbudà', 'ak': 'Antigua ne Baabuda', 'am': 'አንቲጓ እና ባርቡዳ', 'ar': 'أنتيغوا وبربودا', 'as': 'এণ্টিগুৱা আৰু বাৰ্বুডা', 'asa': 'Antigua na Barbuda', 'ast': 'Antigua y Barbuda', 'az': 'Antiqua və Barbuda', 'az-Cyrl': 'Антигуа вә Барбуда', 'az-Latn': 'Antiqua və Barbuda', 'bal-Latn': 'Antiga o Barbuda', 'bas': 'Àŋtigà ɓɔ Bàrbudà', 'be': 'Антыгуа і Барбуда', 'be-tarask': 'Антыгуа і Барбуда', 'bez': 'Huantigua na Hubarubuda', 'bg': 'Антигуа и Барбуда', 'blo': 'Antiguwaa na Barbuɖaa', 'bm': 'Antiga-ni-Barbuda', 'bn': 'অ্যান্টিগুয়া ও বারবুডা', 'br': 'Antigua ha Barbuda', 'brx': 'एन्टिगुआ आरो बारबुडा', 'bs': 'Antigva i Barbuda', 'bs-Cyrl': 'Антигва и Барбуда', 'bs-Latn': 'Antigva i Barbuda', 'ca': 'Antigua i Barbuda', 'ca-ES-valencia': 'Antigua i Barbuda', 'ccp': '𑄆𑄚𑄴𑄖𑄨𑄉𑄱 𑄃𑄮 𑄝𑄢𑄴𑄟𑄪𑄓', 'ce': 'Антигуа а, Барбуда а', 'ceb': 'Antigua & Barbuda', 'cgg': 'Angiguwa na Babuda', 'chr': 'ᎤᏪᏘ & ᏆᏊᏓ', 'ckb': 'ئانتیگوا و باربودا', 'cs': 'Antigua a Barbuda', 'cv': 'Антигуа тата Барбуда', 'cy': 'Antigua a Barbuda', 'da': 'Antigua og Barbuda', 'dav': 'Antigua na Barbuda', 'de': 'Antigua und Barbuda', 'dje': 'Antigua nda Barbuuda', 'dsb': 'Antigua a Barbuda', 'dyo': 'Antigua di Barbuda', 'dz': 'ཨན་ཊི་གུ་ཝ་ ཨེནཌ་ བྷར་བྷུ་ཌ', 'ebu': 'Antigua na Barbuda', 'ee': '́Antigua kple Barbuda nutome', 'el': 'Αντίγκουα και Μπαρμπούντα', 'el-polyton': 'Ἀντίγκουα καὶ Μπαρμπούντα', 'en': 'Antigua & Barbuda', 'en-Dsrt': '𐐈𐑌𐐻𐐨𐑀𐐶𐐲 𐐰𐑌𐐼 𐐒𐐪𐑉𐐺𐐷𐐭𐐼𐐲', 'eo': 'Antigvo kaj Barbudo', 'es': 'Antigua y Barbuda', 'et': 'Antigua ja Barbuda', 'eu': 'Antigua eta Barbuda', 'ewo': 'Antígwa ai Barəbúda', 'fa': 'آنتیگوا و باربودا', 'ff': 'Antiguwaa e Barbudaa', 'ff-Adlm': '𞤀𞤲𞤼𞤭𞤺𞤵𞤱𞤢 & 𞤄𞤢𞤪𞤦𞤵𞥅𞤣𞤢', 'ff-Latn': 'Antiguwaa e Barbudaa', 'fi': 'Antigua ja Barbuda', 'fil': 'Antigua & Barbuda', 'fo': 'Antigua & Barbuda', 'fr': 'Antigua-et-Barbuda', 'fur': 'Antigua e Barbuda', 'fy': 'Antigua en Barbuda', 'ga': 'Antigua agus Barbúda', 'gd': 'Aintìoga is Barbuda', 'gl': 'Antigua e Barbuda', 'gsw': 'Antigua und Barbuda', 'gu': 'ઍન્ટિગુઆ અને બર્મુડા', 'guz': 'Antigua na Barbuda', 'ha': 'Antigua da Barbuda', 'he': 'אנטיגואה וברבודה', 'hi': 'एंटिगुआ और बरबुडा', 'hi-Latn': 'Antigua & Barbuda', 'hr': 'Antigva i Barbuda', 'hsb': 'Antigua a Barbuda', 'ht': 'Antigua-et-Barbuda', 'hu': 'Antigua és Barbuda', 'hy': 'Անտիգուա և Բարբուդա', 'ia': 'Antigua e Barbuda', 'id': 'Antigua dan Barbuda', 'ig': 'Antigua na Barbuda', 'is': 'Antígva og Barbúda', 'it': 'Antigua e Barbuda', 'ja': 'アンティグア・バーブーダ', 'jmc': 'Antigua na Barbuda', 'jv': 'Antigua lan Barbuda', 'ka': 'ანტიგუა და ბარბუდა', 'kaa': 'Антигуа ҳәм Барбуда', 'kaa-Cyrl': 'Антигуа ҳәм Барбуда', 'kab': 'Untiga d Barbuda', 'kam': 'Antigua na Barbuda', 'kde': 'Antigua na Balbuda', 'kea': 'Antigua i Barbuda', 'kgp': 'Ỹntiguva kar Marmuna', 'khq': 'Antigua nda Barbuuda', 'ki': 'Antigua na Barbuda', 'kk': 'Антигуа және Барбуда', 'kk-Cyrl': 'Антигуа және Барбуда', 'kln': 'Emetab Antigua ak Barbuda', 'km': 'អង់ទីហ្គា និង បាប៊ុយដា', 'kn': 'ಆಂಟಿಗುವಾ ಮತ್ತು ಬರ್ಬುಡಾ', 'ko': '앤티가 바부다', 'kok': 'एँटिगुआ आनी बारबुडा', 'kok-Deva': 'एँटिगुआ आनी बारबुडा', 'ks': 'اؠنٹِگُوا تہٕ باربوڑا', 'ks-Arab': 'اؠنٹِگُوا تہٕ باربوڑا', 'ksb': 'Antigua na Balbuda', 'ksf': 'antiga ri barbúda', 'ksh': 'Antigwa un Barbuda', 'ku': 'Antîgua û Berbûda', 'kxv': 'eṇṭigaa aḍe barbuḍaa', 'kxv-Deva': 'एंटिगुआ ऑड़े बरबुडा', 'kxv-Latn': 'eṇṭigaa aḍe barbuḍaa', 'kxv-Orya': 'ଏଣ୍ଟିଗୁଆ ଅଡ଼େ ବାରବୁଦା', 'kxv-Telu': 'ఆంటిగ్వా మరియు బార్బుడా', 'ky': 'Антигуа жана Барбуда', 'lag': 'Antigúua na Baribúuda', 'lb': 'Antigua a Barbuda', 'lg': 'Antigwa ne Barabuda', 'lld': 'Antigua y Barbuda', 'ln': 'Antiga mpé Barbuda', 'lo': 'ແອນທິກົວ ແລະ ບາບູດາ', 'lt': 'Antigva ir Barbuda', 'lu': 'Antiga ne Barbuda', 'luo': 'Antigua gi Barbuda', 'luy': 'Antigua na Barbuda', 'lv': 'Antigva un Barbuda', 'mai': 'एंटिगुआ आ बारबुडा', 'mas': 'Antigua na Barbuda', 'mer': 'Antigua na Barbuda', 'mfe': 'Antigua-ek-Barbuda', 'mg': 'Antiga sy Barboda', 'mi': 'Motu Nehe me Pāputa', 'mk': 'Антига и Барбуда', 'ml': 'ആൻറിഗ്വയും ബർബുഡയും', 'mn': 'Антигуа ба Барбуда', 'mr': 'अँटिग्वा आणि बर्बुडा', 'ms': 'Antigua dan Barbuda', 'mt': 'Antigua u Barbuda', 'mua': 'antiguan ne Barbuda', 'my': 'အန်တီဂွါနှင့် ဘာဘူဒါ', 'mzn': 'آنتیگوا و باربودا', 'naq': 'Antiguab tsî Barbudab', 'nb': 'Antigua og Barbuda', 'nd': 'Antigua le Barbuda', 'ne': 'एन्टिगुआ र बारबुडा', 'nl': 'Antigua en Barbuda', 'nmg': 'Antíga bá Barbúda', 'nn': 'Antigua og Barbuda', 'no': 'Antigua og Barbuda', 'nqo': 'ߊ߲ߕߌߞߎߥߊ߫ ߣߌ߫ ߓߊߙߑߓߎߘߊ߫', 'nus': 'Antiguaa kɛnɛ Barbuda', 'nyn': 'Angiguwa na Babuda', 'om': 'Antiiguyaa fi Barbuudaa', 'or': 'ଆଣ୍ଟିଗୁଆ ଏବଂ ବାରବୁଦା', 'pa': 'ਐਂਟੀਗੁਆ ਅਤੇ ਬਾਰਬੁਡਾ', 'pa-Guru': 'ਐਂਟੀਗੁਆ ਅਤੇ ਬਾਰਬੁਡਾ', 'pcm': 'Antígwua & Barbúda', 'pl': 'Antigua i Barbuda', 'ps': 'انټيګوا او باربودا', 'pt': 'Antígua e Barbuda', 'qu': 'Antigua y Barbuda', 'rm': 'Antigua e Barbuda', 'rn': 'Antigwa na Baribuda', 'ro': 'Antigua și Barbuda', 'rof': 'Antigua na Babuda', 'ru': 'Антигуа и Барбуда', 'rwk': 'Antigua na Barbuda', 'saq': 'Antigua na Barbuda', 'sbp': 'Anitiguya ni Balubuda', 'sc': 'Antigua e Barbuda', 'sd': 'انٽيگا ۽ باربد', 'sd-Arab': 'انٽيگا ۽ باربد', 'se': 'Antigua ja Barbuda', 'seh': 'Antígua e Barbuda', 'ses': 'Antigua nda Barbuuda', 'sg': 'Antîgua na Barbûda', 'shi': 'ⴰⵏⵜⵉⴳⴰ ⴷ ⴱⵔⴱⵓⴷⴰ', 'shi-Latn': 'antiga d brbuda', 'shi-Tfng': 'ⴰⵏⵜⵉⴳⴰ ⴷ ⴱⵔⴱⵓⴷⴰ', 'si': 'ඇන්ටිගුවා සහ බාබියුඩාව', 'sk': 'Antigua a Barbuda', 'sl': 'Antigva in Barbuda', 'smn': 'Antigua já Barbuda', 'sn': 'Antigua ne Barbuda', 'so': 'Antigua & Barbuuda', 'sq': 'Antigua e Barbuda', 'sr': 'Антигва и Барбуда', 'sr-Cyrl': 'Антигва и Барбуда', 'sr-Latn': 'Antigva i Barbuda', 'sv': 'Antigua och Barbuda', 'sw': 'Antigua na Barbuda', 'syr': 'ܐܢܬܝܓܘܐ ܘܒܐܪܒܘܕܐ', 'ta': 'ஆண்டிகுவா மற்றும் பார்புடா', 'te': 'ఆంటిగ్వా మరియు బార్బుడా', 'teo': 'Antigua na Barbuda', 'tg': 'Антигуа ва Барбуда', 'th': 'แอนติกาและบาร์บูดา', 'ti': 'ኣንቲጓን ባርቡዳን', 'tk': 'Antigua we Barbuda', 'to': 'Anitikua mo Palaputa', 'tr': 'Antigua ve Barbuda', 'tt': 'Антигуа һәм Барбуда', 'twq': 'Antigua nda Barbuuda', 'tzm': 'Antigwa d Barbuda', 'ug': 'ئانتىگۇئا ۋە باربۇدا', 'uk': 'Антигуа і Барбуда', 'ur': 'انٹیگوا اور باربودا', 'uz': 'Antigua va Barbuda', 'uz-Cyrl': 'Антигуа ва Барбуда', 'uz-Latn': 'Antigua va Barbuda', 'vai': 'ꕉꘋꔳꖶꕎ ꗪ ꕑꖜꕜ', 'vai-Latn': 'Aŋtígwa Ɓahabhuda', 'vai-Vaii': 'ꕉꘋꔳꖶꕎ ꗪ ꕑꖜꕜ', 'vec': 'Antigua e Barbuda', 'vi': 'Antigua và Barbuda', 'vun': 'Antigua na Barbuda', 'wae': 'Antigua und Barbuda', 'wo': 'Antiguwa ak Barbuda', 'xh': 'E-Antigua & Barbuda', 'xnr': 'एंटिगुआ कने बरबुडा', 'xog': 'Antigwa ni Barabuda', 'yav': 'Aŋtíka na Palpúta', 'yi': 'אַנטיגוע און באַרבודע', 'yo': 'Ààntígúà àti Báríbúdà', 'yrl': 'Ãtigua asuí Babuda', 'yue': '安提瓜同巴布達', 'yue-Hans': '安提瓜同巴布达', 'yue-Hant': '安提瓜同巴布達', 'zgh': 'ⴰⵏⵜⵉⴳⴰ ⴷ ⴱⵔⴱⵓⴷⴰ', 'zh': '安提瓜和巴布达', 'zh-Hans': '安提瓜和巴布达', 'zh-Hant': '安地卡及巴布達', 'zu': 'i-Antigua ne-Barbuda'},
    'AI': {'af': 'Anguilla', 'agq': 'Àŋgwilà', 'ak': 'Anguila', 'am': 'አንጉይላ', 'ar': 'أنغويلا', 'as': 'এনগুইলা', 'asa': 'Anguilla', 'ast': 'Anguila', 'az': 'Angilya', 'az-Cyrl': 'Анҝилја', 'az-Latn': 'Angilya', 'bal-Latn': 'Angwila', 'bas': 'Àŋgiyà', 'be': 'Ангілья', 'be-tarask': 'Ангілья', 'bez': 'Huanguila', 'bg': 'Ангуила', 'bgn': 'انگویلا', 'blo': 'Aŋguwilaa', 'bm': 'Angiya', 'bn': 'অ্যাঙ্গুইলা', 'br': 'Anguilla', 'brx': 'एंगुइल्ला', 'bs': 'Angvila', 'bs-Cyrl': 'Ангвила', 'bs-Latn': 'Angvila', 'ca': 'Anguilla', 'ca-ES-valencia': 'Anguilla', 'ccp': '𑄄𑄳𑄠𑄋𑄴𑄉𑄪𑄃𑄨𑄣', 'ce': 'Ангилья', 'ceb': 'Anguilla', 'cgg': 'Angwira', 'chr': 'ᎠᏂᎩᎳ', 'ckb': 'ئانگویلا', 'cs': 'Anguilla', 'cv': 'Ангилья', 'cy': 'Anguilla', 'da': 'Anguilla', 'dav': 'Anguilla', 'de': 'Anguilla', 'dje': 'Angiiya', 'dsb': 'Anguilla', 'dyo': 'Angiiya', 'dz': 'ཨང་གི་ལ', 'ebu': 'Anguilla', 'ee': 'Anguilla nutome', 'el': 'Ανγκουίλα', 'el-polyton': 'Ἀνγκουίλα', 'en': 'Anguilla', 'en-Dsrt': '𐐈𐑍𐑀𐐶𐐮𐑊𐐲', 'eo': 'Angvilo', 'es': 'Anguila', 'et': 'Anguilla', 'eu': 'Aingira', 'ewo': 'Angíyə', 'fa': 'آنگویلا', 'ff': 'Anngiyaa', 'ff-Adlm': '𞤀𞤲𞤺𞤭𞤤𞤢𞥄', 'ff-Latn': 'Anngiyaa', 'fi': 'Anguilla', 'fil': 'Anguilla', 'fo': 'Anguilla', 'fr': 'Anguilla', 'fur': 'Anguilla', 'fy': 'Anguilla', 'ga': 'Angaíle', 'gd': 'Anguillia', 'gl': 'Anguila', 'gsw': 'Anguilla', 'gu': 'ઍંગ્વિલા', 'guz': 'Anguilla', 'ha': 'Angila', 'he': 'אנגווילה', 'hi': 'एंग्विला', 'hi-Latn': 'Anguilla', 'hr': 'Angvila', 'hsb': 'Anguilla', 'ht': 'Anguilla', 'hu': 'Anguilla', 'hy': 'Անգուիլա', 'ia': 'Anguilla', 'id': 'Anguilla', 'ig': 'Anguilla', 'is': 'Angvilla', 'it': 'Anguilla', 'ja': 'アンギラ', 'jmc': 'Anguilla', 'jv': 'Anguilla', 'ka': 'ანგილია', 'kaa': 'Ангилья', 'kaa-Cyrl': 'Ангилья', 'kab': 'Ungiya', 'kam': 'Anguilla', 'kde': 'Angwila', 'kea': 'Angila', 'kgp': 'Ỹngira', 'khq': 'Angiiya', 'ki': 'Anguilla', 'kk': 'Ангилья', 'kk-Cyrl': 'Ангилья', 'kln': 'Emetab Anguilla', 'km': 'អង់ហ្គីឡា', 'kn': 'ಆಂಗ್ವಿಲ್ಲಾ', 'ko': '앵귈라', 'kok': 'अंगुला', 'kok-Deva': 'अंगुला', 'ks': 'انگوئیلا', 'ks-Arab': 'انگوئیلا', 'ksb': 'Anguilla', 'ksf': 'angiya', 'ksh': 'Anggwilla', 'ku': 'Anguîla', 'kxv': 'anguila', 'kxv-Deva': 'एंग्विला', 'kxv-Latn': 'anguila', 'kxv-Orya': 'ଏଙ୍ଗ୍ୱିଲା', 'kxv-Telu': 'ఆంగ్విల్లా', 'ky': 'Ангилья', 'lag': 'Anguíila', 'lb': 'Anguilla', 'lg': 'Angwila', 'lld': 'Anguilla', 'ln': 'Angiyɛ', 'lo': 'ແອນກຸຍລາ', 'lt': 'Angilija', 'lu': 'Angiye', 'luo': 'Anguilla', 'luy': 'Anguilla', 'lv': 'Angilja', 'mai': 'एंग्विला', 'mas': 'Anguilla', 'mer': 'Anguilla', 'mfe': 'Anguilla', 'mg': 'Anguilla', 'mi': 'Anguira', 'mk': 'Ангвила', 'ml': 'ആൻഗ്വില്ല', 'mn': 'Ангилья', 'mr': 'अँग्विला', 'ms': 'Anguilla', 'mt': 'Anguilla', 'mua': 'anguiya', 'my': 'အန်ဂီလာ', 'mzn': 'آنگویلا', 'naq': 'Anguillab', 'nb': 'Anguilla', 'nd': 'Anguilla', 'ne': 'आङ्गुइला', 'nl': 'Anguilla', 'nmg': 'Anguílla', 'nn': 'Anguilla', 'no': 'Anguilla', 'nqo': 'ߊ߲ߞߎ߳ߟߊ߫', 'nus': 'Aŋguɛla', 'nyn': 'Angwira', 'om': 'Anguyilaa', 'or': 'ଆଙ୍ଗୁଇଲ୍ଲା', 'pa': 'ਅੰਗੁਇਲਾ', 'pa-Guru': 'ਅੰਗੁਇਲਾ', 'pcm': 'Angwíla', 'pl': 'Anguilla', 'ps': 'انګیلا', 'pt': 'Anguila', 'qu': 'Anguila', 'rm': 'Anguilla', 'rn': 'Angwila', 'ro': 'Anguilla', 'rof': 'Anguila', 'ru': 'Ангилья', 'rwk': 'Anguilla', 'saq': 'Anguilla', 'sbp': 'Anguilla', 'sc': 'Anguilla', 'sd': 'انگويلا', 'sd-Arab': 'انگويلا', 'se': 'Anguilla', 'seh': 'Anguilla', 'ses': 'Angiiya', 'sg': 'Angûîla', 'shi': 'ⴰⵏⴳⵉⵍⴰ', 'shi-Latn': 'angila', 'shi-Tfng': 'ⴰⵏⴳⵉⵍⴰ', 'si': 'ඇන්ගුයිලාව', 'sk': 'Anguilla', 'sl': 'Angvila', 'smn': 'Anguilla', 'sn': 'Anguila', 'so': 'Anguula', 'sq': 'Anguilë', 'sr': 'Ангвила', 'sr-Cyrl': 'Ангвила', 'sr-Latn': 'Angvila', 'sv': 'Anguilla', 'sw': 'Anguilla', 'syr': 'ܐܢܓܘܝܠܐ', 'ta': 'அங்கியுலா', 'te': 'ఆంగ్విల్లా', 'teo': 'Anguilla', 'tg': 'Ангилия', 'th': 'แองกวิลลา', 'ti': 'ኣንጒላ', 'tk': 'Angilýa', 'to': 'Anikuila', 'tr': 'Anguilla', 'tt': 'Ангилья', 'twq': 'Angiiya', 'tzm': 'Angwilla', 'ug': 'ئانگۋىللا', 'uk': 'Ангілья', 'ur': 'انگوئیلا', 'uz': 'Angilya', 'uz-Cyrl': 'Ангилья', 'uz-Latn': 'Angilya', 'vai': 'ꕉꕄꕞ', 'vai-Latn': 'Aŋgíla', 'vai-Vaii': 'ꕉꕄꕞ', 'vec': 'Anguila', 'vi': 'Anguilla', 'vun': 'Anguilla', 'wae': 'Anguilla', 'wo': 'Angiiy', 'xh': 'E-Anguilla', 'xnr': 'एंग्विला', 'xog': 'Angwila', 'yav': 'Aŋkíla', 'yo': 'Ààngúlílà', 'yrl': 'Ãgira', 'yue': '安圭拉', 'yue-Hans': '安圭拉', 'yue-Hant': '安圭拉', 'zgh': 'ⴰⵏⴳⵉⵍⴰ', 'zh': '安圭拉', 'zh-Hans': '安圭拉', 'zh-Hant': '安奎拉', 'zu': 'i-Anguilla'},
    'AL': {'af': 'Albanië', 'agq': 'Àabɛnìa', 'ak': 'Albenia', 'am': 'አልባኒያ', 'ar': 'ألبانيا', 'as': 'আলবেনিয়া', 'asa': 'Albania', 'ast': 'Albania', 'az': 'Albaniya', 'az-Cyrl': 'Албанија', 'az-Latn': 'Albaniya', 'bal-Latn': 'Albániá', 'bas': 'Àlbanìà', 'be': 'Албанія', 'be-tarask': 'Албанія', 'bez': 'Hualbania', 'bg': 'Албания', 'blo': 'Albanii', 'bm': 'Alibani', 'bn': 'আলবেনিয়া', 'br': 'Albania', 'brx': 'आल्बानिया', 'bs': 'Albanija', 'bs-Cyrl': 'Албанија', 'bs-Latn': 'Albanija', 'ca': 'Albània', 'ca-ES-valencia': 'Albània', 'ccp': '𑄃𑄣𑄴𑄝𑄬𑄚𑄨𑄠', 'ce': 'Албани', 'ceb': 'Albania', 'cgg': 'Arubania', 'chr': 'ᎠᎵᏇᏂᏯ', 'ckb': 'ئەڵبانیا', 'cs': 'Albánie', 'cv': 'Албани', 'cy': 'Albania', 'da': 'Albanien', 'dav': 'Albania', 'de': 'Albanien', 'dje': 'Albaani', 'dsb': 'Albańska', 'dyo': 'Albani', 'dz': 'ཨཱལ་བེ་ནི་ཡ', 'ebu': 'Albania', 'ee': 'Albania nutome', 'el': 'Αλβανία', 'el-polyton': 'Ἀλβανία', 'en': 'Albania', 'en-Dsrt': '𐐈𐑊𐐺𐐩𐑌𐐨𐐲', 'eo': 'Albanujo', 'es': 'Albania', 'et': 'Albaania', 'eu': 'Albania', 'ewo': 'Aləbánia', 'fa': 'آلبانی', 'ff': 'Albanii', 'ff-Adlm': '𞤀𞤤𞤦𞤢𞤲𞤭𞤴𞤢𞥄', 'ff-Latn': 'Albanii', 'fi': 'Albania', 'fil': 'Albania', 'fo': 'Albania', 'fr': 'Albanie', 'fur': 'Albanie', 'fy': 'Albanië', 'ga': 'an Albáin', 'gd': 'Albàinia', 'gl': 'Albania', 'gsw': 'Albaanie', 'gu': 'અલ્બેનિયા', 'guz': 'Albania', 'ha': 'Albaniya', 'he': 'אלבניה', 'hi': 'अल्बानिया', 'hi-Latn': 'Albania', 'hr': 'Albanija', 'hsb': 'Albanska', 'ht': 'Albanie', 'hu': 'Albánia', 'hy': 'Ալբանիա', 'ia': 'Albania', 'id': 'Albania', 'ie': 'Albania', 'ig': 'Albania', 'is': 'Albanía', 'it': 'Albania', 'ja': 'アルバニア', 'jmc': 'Albania', 'jv': 'Albani', 'ka': 'ალბანეთი', 'kaa': 'Албания', 'kaa-Cyrl': 'Албания', 'kab': 'Lalbani', 'kam': 'Albania', 'kde': 'Albania', 'kea': 'Albánia', 'kgp': 'Armánĩja', 'khq': 'Albaani', 'ki': 'Albania', 'kk': 'Албания', 'kk-Cyrl': 'Албания', 'kln': 'Emetab Albania', 'km': 'អាល់បានី', 'kn': 'ಅಲ್ಬೇನಿಯಾ', 'ko': '알바니아', 'kok': 'अल्बानीया', 'kok-Deva': 'अल्बानीया', 'ks': 'البانیا', 'ks-Arab': 'البانیا', 'ksb': 'Albania', 'ksf': 'albaní', 'ksh': 'Albaanije', 'ku': 'Albanya', 'kxv': 'albania', 'kxv-Deva': 'अल्बानिया', 'kxv-Latn': 'albania', 'kxv-Orya': 'ଆଲବାନିଆ', 'kxv-Telu': 'ఆల్లేనియా', 'ky': 'Албания', 'lag': 'Alubánia', 'lb': 'Albanien', 'lg': 'Alibaniya', 'lld': 'Albania', 'ln': 'Alibani', 'lo': 'ແອວເບເນຍ', 'lt': 'Albanija', 'lu': 'Alubani', 'luo': 'Albania', 'luy': 'Albania', 'lv': 'Albānija', 'mai': 'अलबानिया', 'mas': 'Albania', 'mer': 'Alubania', 'mfe': 'Albani', 'mg': 'Albania', 'mgh': 'Ualbania', 'mi': 'Arapeinia', 'mk': 'Албанија', 'ml': 'അൽബേനിയ', 'mn': 'Албани', 'mr': 'अल्बानिया', 'ms': 'Albania', 'mt': 'l-Albanija', 'mua': 'albaniya', 'my': 'အယ်လ်ဘေးနီးယား', 'mzn': 'آلبانی', 'naq': 'Albaniab', 'nb': 'Albania', 'nd': 'Albania', 'ne': 'अल्बेनिया', 'nl': 'Albanië', 'nmg': 'Albania', 'nn': 'Albania', 'no': 'Albania', 'nqo': 'ߊߟߑߓߊߣߌ߫', 'nus': 'Albänia', 'nyn': 'Arubania', 'om': 'Albaaniyaa', 'or': 'ଆଲବାନିଆ', 'pa': 'ਅਲਬਾਨੀਆ', 'pa-Guru': 'ਅਲਬਾਨੀਆ', 'pcm': 'Albénia', 'pl': 'Albania', 'ps': 'البانیه', 'pt': 'Albânia', 'qu': 'Albania', 'rm': 'Albania', 'rn': 'Alubaniya', 'ro': 'Albania', 'rof': 'Albania', 'ru': 'Албания', 'rwk': 'Albania', 'saq': 'Albania', 'sbp': 'Alubaniya', 'sc': 'Albania', 'sd': 'البانيا', 'sd-Arab': 'البانيا', 'se': 'Albánia', 'seh': 'Albânia', 'ses': 'Albaani', 'sg': 'Albanïi', 'shi': 'ⴰⵍⴱⴰⵏⵢⴰ', 'shi-Latn': 'albanya', 'shi-Tfng': 'ⴰⵍⴱⴰⵏⵢⴰ', 'si': 'ඇල්බේනියාව', 'sk': 'Albánsko', 'sl': 'Albanija', 'smn': 'Albania', 'sn': 'Albania', 'so': 'Albaaniya', 'sq': 'Shqipëri', 'sr': 'Албанија', 'sr-Cyrl': 'Албанија', 'sr-Latn': 'Albanija', 'sv': 'Albanien', 'sw': 'Albania', 'syr': 'ܐܠܒܢܝܐ', 'ta': 'அல்பேனியா', 'te': 'అల్బేనియా', 'teo': 'Albania', 'tg': 'Албания', 'th': 'แอลเบเนีย', 'ti': 'ኣልባንያ', 'tk': 'Albaniýa', 'to': 'ʻAlipania', 'tr': 'Arnavutluk', 'tt': 'Албания', 'twq': 'Albaani', 'tzm': 'Albanya', 'ug': 'ئالبانىيە', 'uk': 'Албанія', 'ur': 'البانیہ', 'uz': 'Albaniya', 'uz-Cyrl': 'Албания', 'uz-Latn': 'Albaniya', 'vai': 'ꕉꔷꕑꕇꕩ', 'vai-Latn': 'Abhaniya', 'vai-Vaii': 'ꕉꔷꕑꕇꕩ', 'vec': 'Albania', 'vi': 'Albania', 'vun': 'Albania', 'wae': 'Albanie', 'wo': 'Albani', 'xh': 'E-Albania', 'xnr': 'अल्बानिया', 'xog': 'Alibaniya', 'yav': 'Alpaní', 'yi': 'אַלבאַניע', 'yo': 'Àlùbàníánì', 'yrl': 'Aubãniya', 'yue': '阿爾巴尼亞', 'yue-Hans': '阿尔巴尼亚', 'yue-Hant': '阿爾巴尼亞', 'zgh': 'ⴰⵍⴱⴰⵏⵢⴰ', 'zh': '阿尔巴尼亚', 'zh-Hans': '阿尔巴尼亚', 'zh-Hant': '阿爾巴尼亞', 'zu': 'i-Albania'},
    'AM': {'af': 'Armenië', 'agq': 'Àmɛnyìa', 'ak': 'Aamenia', 'am': 'አርሜኒያ', 'ar': 'أرمينيا', 'as': 'আৰ্মেনিয়া', 'asa': 'Armenia', 'ast': 'Armenia', 'az': 'Ermənistan', 'az-Cyrl': 'Ермәнистан', 'az-Latn': 'Ermənistan', 'bal-Latn': 'Árminiá', 'bas': 'Àrmenìà', 'be': 'Арменія', 'be-tarask': 'Арменія', 'bez': 'Huamenia', 'bg': 'Армения', 'blo': 'Armenii', 'bm': 'Arimeni', 'bn': 'আর্মেনিয়া', 'br': 'Armenia', 'brx': 'आर्मेनिया', 'bs': 'Armenija', 'bs-Cyrl': 'Арменија', 'bs-Latn': 'Armenija', 'ca': 'Armènia', 'ca-ES-valencia': 'Armènia', 'ccp': '𑄃𑄢𑄴𑄟𑄬𑄚𑄨𑄠', 'ce': 'Эрмалойчоь', 'ceb': 'Armenia', 'cgg': 'Arimeniya', 'chr': 'ᎠᎵᎻᏂᎠ', 'ckb': 'ئەرمەنستان', 'cs': 'Arménie', 'cv': 'Армени', 'cy': 'Armenia', 'da': 'Armenien', 'dav': 'Armenia', 'de': 'Armenien', 'dje': 'Armeeni', 'dsb': 'Armeńska', 'dyo': 'Armeni', 'dz': 'ཨར་མི་ནི་ཡ', 'ebu': 'Armenia', 'ee': 'Armenia nutome', 'el': 'Αρμενία', 'el-polyton': 'Ἀρμενία', 'en': 'Armenia', 'en-Dsrt': '𐐂𐑉𐑋𐐨𐑌𐐨𐐲', 'eo': 'Armenujo', 'es': 'Armenia', 'et': 'Armeenia', 'eu': 'Armenia', 'ewo': 'Arəménia', 'fa': 'ارمنستان', 'ff': 'Armenii', 'ff-Adlm': '𞤀𞤪𞤥𞤫𞤲𞤭𞤴𞤢𞥄', 'ff-Latn': 'Armenii', 'fi': 'Armenia', 'fil': 'Armenia', 'fo': 'Armenia', 'fr': 'Arménie', 'fur': 'Armenie', 'fy': 'Armenië', 'ga': 'an Airméin', 'gd': 'Airmeinea', 'gl': 'Armenia', 'gsw': 'Armeenie', 'gu': 'આર્મેનિયા', 'guz': 'Armenia', 'ha': 'Armeniya', 'he': 'ארמניה', 'hi': 'आर्मेनिया', 'hi-Latn': 'Armenia', 'hr': 'Armenija', 'hsb': 'Armenska', 'ht': 'Arménie', 'hu': 'Örményország', 'hy': 'Հայաստան', 'ia': 'Armenia', 'id': 'Armenia', 'ig': 'Armenia', 'is': 'Armenía', 'it': 'Armenia', 'ja': 'アルメニア', 'jmc': 'Armenia', 'jv': 'Arménia', 'ka': 'სომხეთი', 'kaa': 'Армения', 'kaa-Cyrl': 'Армения', 'kab': 'Arminya', 'kam': 'Armenia', 'kde': 'Almenia', 'kea': 'Arménia', 'kgp': 'Armẽnĩja', 'khq': 'Armeeni', 'ki': 'Armenia', 'kk': 'Армения', 'kk-Cyrl': 'Армения', 'kln': 'Emetab Armenia', 'km': 'អាមេនី', 'kn': 'ಆರ್ಮೇನಿಯ', 'ko': '아르메니아', 'kok': 'आर्मीनीया', 'kok-Deva': 'आर्मीनीया', 'ks': 'اَرمانِیا', 'ks-Arab': 'اَرمانِیا', 'ksb': 'Almenia', 'ksf': 'armɛní', 'ksh': 'Arrmeenije', 'ku': 'Ermenistan', 'kxv': 'aarmenia', 'kxv-Deva': 'आर्मेनिया', 'kxv-Latn': 'aarmenia', 'kxv-Orya': 'ଆର୍ମେନିଆ', 'kxv-Telu': 'ఆర్మేనియా', 'ky': 'Армения', 'lag': 'Ariménia', 'lb': 'Armenien', 'lg': 'Arameniya', 'lld': 'Armenia', 'ln': 'Amɛni', 'lo': 'ອາເມເນຍ', 'lt': 'Armėnija', 'lu': 'Ameni', 'luo': 'Armenia', 'luy': 'Armenia', 'lv': 'Armēnija', 'mai': 'आरमेनिया', 'mas': 'Armenia', 'mer': 'Armenia', 'mfe': 'Armeni', 'mg': 'Armenia', 'mi': 'Āmenia', 'mk': 'Ерменија', 'ml': 'അർമേനിയ', 'mn': 'Армени', 'mr': 'अर्मेनिया', 'ms': 'Armenia', 'mt': 'l-Armenja', 'mua': 'armeniya', 'my': 'အာမေးနီးယား', 'mzn': 'ارمنستون', 'naq': 'Armeniab', 'nb': 'Armenia', 'nd': 'Armenia', 'ne': 'आर्मेनिया', 'nl': 'Armenië', 'nmg': 'Arménia', 'nn': 'Armenia', 'no': 'Armenia', 'nqo': 'ߊߙߑߡߋߣߌ߫', 'nus': 'Aɛrmänia', 'nyn': 'Arimeniya', 'om': 'Armeeniyaa', 'or': 'ଆର୍ମେନିଆ', 'pa': 'ਅਰਮੀਨੀਆ', 'pa-Guru': 'ਅਰਮੀਨੀਆ', 'pcm': 'Armẹ́niá', 'pl': 'Armenia', 'ps': 'ارمنستان', 'pt': 'Armênia', 'qu': 'Armenia', 'rm': 'Armenia', 'rn': 'Arumeniya', 'ro': 'Armenia', 'rof': 'Amenia', 'ru': 'Армения', 'rwk': 'Armenia', 'saq': 'Armenia', 'sbp': 'Alimeniya', 'sc': 'Armènia', 'sd': 'ارمینیا', 'sd-Arab': 'ارمینیا', 'se': 'Armenia', 'seh': 'Armênia', 'ses': 'Armeeni', 'sg': 'Armenïi', 'shi': 'ⴰⵔⵎⵉⵏⵢⴰ', 'shi-Latn': 'arminya', 'shi-Tfng': 'ⴰⵔⵎⵉⵏⵢⴰ', 'si': 'ආර්මේනියාව', 'sk': 'Arménsko', 'sl': 'Armenija', 'smn': 'Armenia', 'sn': 'Armenia', 'so': 'Armeeniya', 'sq': 'Armeni', 'sr': 'Јерменија', 'sr-Cyrl': 'Јерменија', 'sr-Latn': 'Jermenija', 'sv': 'Armenien', 'sw': 'Armenia', 'syr': 'ܐܪܡܢܝܐ', 'ta': 'அர்மேனியா', 'te': 'ఆర్మేనియా', 'teo': 'Armenia', 'tg': 'Арманистон', 'th': 'อาร์เมเนีย', 'ti': 'ኣርሜንያ', 'tk': 'Ermenistan', 'to': 'ʻĀmenia', 'tr': 'Ermenistan', 'tt': 'Әрмәнстан', 'twq': 'Armeeni', 'tzm': 'Arminya', 'ug': 'ئەرمېنىيە', 'uk': 'Вірменія', 'ur': 'آرمینیا', 'uz': 'Armaniston', 'uz-Cyrl': 'Арманистон', 'uz-Latn': 'Armaniston', 'vai': 'ꕉꕆꕯ', 'vai-Latn': 'Améniya', 'vai-Vaii': 'ꕉꕆꕯ', 'vec': 'Armenia', 'vi': 'Armenia', 'vun': 'Armenia', 'wae': 'Armenie', 'wo': 'Armeni', 'xh': 'E-Armenia', 'xnr': 'आर्मेनिया', 'xog': 'Arameniya', 'yav': 'Almanía', 'yi': 'אַרמעניע', 'yo': 'Améníà', 'yrl': 'Arimẽniya', 'yue': '亞美尼亞', 'yue-Hans': '亚美尼亚', 'yue-Hant': '亞美尼亞', 'zgh': 'ⴰⵔⵎⵉⵏⵢⴰ', 'zh': '亚美尼亚', 'zh-Hans': '亚美尼亚', 'zh-Hant': '亞美尼亞', 'zu': 'i-Armenia'},
    'AN': {'en': 'Netherlands Antilles'},
    'AO': {'af': 'Angola', 'agq': 'Àŋgolà', 'ak': 'Angola', 'am': 'አንጐላ', 'ar': 'أنغولا', 'as': 'এংগোলা', 'asa': 'Angola', 'ast': 'Angola', 'az': 'Anqola', 'az-Cyrl': 'Ангола', 'az-Latn': 'Anqola', 'bal-Latn': 'Angólá', 'bas': 'Àŋgolà', 'be': 'Ангола', 'be-tarask': 'Ангола', 'bez': 'Huangola', 'bg': 'Ангола', 'bgn': 'انگولا', 'blo': 'Aŋgolaa', 'bm': 'Angola', 'bn': 'অ্যাঙ্গোলা', 'br': 'Angola', 'brx': 'आंगला', 'bs': 'Angola', 'bs-Cyrl': 'Ангола', 'bs-Latn': 'Angola', 'ca': 'Angola', 'ca-ES-valencia': 'Angola', 'ccp': '𑄃𑄳𑄠𑄋𑄴𑄉𑄮𑄣', 'ce': 'Ангола', 'ceb': 'Angola', 'cgg': 'Angora', 'chr': 'ᎠᏂᎪᎳ', 'ckb': 'ئەنگۆلا', 'cs': 'Angola', 'cv': 'Ангола', 'cy': 'Angola', 'da': 'Angola', 'dav': 'Angola', 'de': 'Angola', 'dje': 'Angoola', 'dsb': 'Angola', 'dyo': 'Angola', 'dz': 'ཨང་གྷོ་ལ', 'ebu': 'Angola', 'ee': 'Angola nutome', 'el': 'Αγκόλα', 'el-polyton': 'Ἀνγκόλα', 'en': 'Angola', 'en-Dsrt': '𐐈𐑌𐑀𐐬𐑊𐐲', 'eo': 'Angolo', 'es': 'Angola', 'et': 'Angola', 'eu': 'Angola', 'ewo': 'Angolá', 'fa': 'آنگولا', 'ff': 'Anngolaa', 'ff-Adlm': '𞤀𞤲𞤺𞤮𞤤𞤢𞥄', 'ff-Latn': 'Anngolaa', 'fi': 'Angola', 'fil': 'Angola', 'fo': 'Angola', 'fr': 'Angola', 'fur': 'Angola', 'fy': 'Angola', 'ga': 'Angóla', 'gd': 'Angòla', 'gl': 'Angola', 'gsw': 'Angoola', 'gu': 'અંગોલા', 'guz': 'Angola', 'ha': 'Angola', 'he': 'אנגולה', 'hi': 'अंगोला', 'hi-Latn': 'Angola', 'hr': 'Angola', 'hsb': 'Angola', 'ht': 'Angola', 'hu': 'Angola', 'hy': 'Անգոլա', 'ia': 'Angola', 'id': 'Angola', 'ig': 'Angola', 'is': 'Angóla', 'it': 'Angola', 'ja': 'アンゴラ', 'jgo': 'Aŋgɔ́la', 'jmc': 'Angola', 'jv': 'Angola', 'ka': 'ანგოლა', 'kaa': 'Ангола', 'kaa-Cyrl': 'Ангола', 'kab': 'Ungula', 'kam': 'Angola', 'kde': 'Angola', 'kea': 'Angola', 'kgp': 'Ỹgóra', 'khq': 'Angoola', 'ki': 'Angola', 'kk': 'Ангола', 'kk-Cyrl': 'Ангола', 'kln': 'Emetab Angola', 'km': 'អង់ហ្គោឡា', 'kn': 'ಅಂಗೋಲಾ', 'ko': '앙골라', 'kok': 'अंगोला', 'kok-Deva': 'अंगोला', 'ks': 'انگولا', 'ks-Arab': 'انگولا', 'ksb': 'Angola', 'ksf': 'angóla', 'ksh': 'Angjoola', 'ku': 'Angola', 'kxv': 'angola', 'kxv-Deva': 'अंगोला', 'kxv-Latn': 'angola', 'kxv-Orya': 'ଅଙ୍ଗଲା', 'kxv-Telu': 'ఆంగోలా', 'ky': 'Ангола', 'lag': 'Angóola', 'lb': 'Angola', 'lg': 'Angola', 'lld': 'Angola', 'ln': 'Angóla', 'lo': 'ແອງໂກລາ', 'lt': 'Angola', 'lu': 'Angola', 'luo': 'Angola', 'luy': 'Angola', 'lv': 'Angola', 'mai': 'अंगोला', 'mas': 'Angola', 'mer': 'Angola', 'mfe': 'Angola', 'mg': 'Angola', 'mi': 'Anakora', 'mk': 'Ангола', 'ml': 'അംഗോള', 'mn': 'Ангол', 'mr': 'अंगोला', 'ms': 'Angola', 'mt': 'l-Angola', 'mua': 'angola', 'my': 'အန်ဂိုလာ', 'mzn': 'آنگولا', 'naq': 'Angolab', 'nb': 'Angola', 'nd': 'Angola', 'ne': 'अङ्गोला', 'nl': 'Angola', 'nmg': 'Angola', 'nn': 'Angola', 'no': 'Angola', 'nqo': 'ߊ߲ߜߏߟߊ߫', 'nus': 'Aŋgola', 'nyn': 'Angora', 'om': 'Angoolaa', 'or': 'ଆଙ୍ଗୋଲା', 'pa': 'ਅੰਗੋਲਾ', 'pa-Guru': 'ਅੰਗੋਲਾ', 'pcm': 'Angóla', 'pl': 'Angola', 'ps': 'انګولا', 'pt': 'Angola', 'qu': 'Angola', 'rm': 'Angola', 'rn': 'Angola', 'ro': 'Angola', 'rof': 'Angolo', 'ru': 'Ангола', 'rwk': 'Angola', 'saq': 'Angola', 'sbp': 'Angola', 'sc': 'Angola', 'sd': 'انگولا', 'sd-Arab': 'انگولا', 'se': 'Angola', 'seh': 'Angola', 'ses': 'Angoola', 'sg': 'Angoläa', 'shi': 'ⴰⵏⴳⵓⵍⴰ', 'shi-Latn': 'angula', 'shi-Tfng': 'ⴰⵏⴳⵓⵍⴰ', 'si': 'ඇන්ගෝලාව', 'sk': 'Angola', 'sl': 'Angola', 'smn': 'Angola', 'sn': 'Angola', 'so': 'Angoola', 'sq': 'Angolë', 'sr': 'Ангола', 'sr-Cyrl': 'Ангола', 'sr-Latn': 'Angola', 'sv': 'Angola', 'sw': 'Angola', 'syr': 'ܐܢܓܘܠܐ', 'ta': 'அங்கோலா', 'te': 'అంగోలా', 'teo': 'Angola', 'tg': 'Ангола', 'th': 'แองโกลา', 'ti': 'ኣንጎላ', 'tk': 'Angola', 'to': 'ʻAngikola', 'tr': 'Angola', 'tt': 'Ангола', 'twq': 'Angoola', 'tzm': 'Angula', 'ug': 'ئانگولا', 'uk': 'Ангола', 'ur': 'انگولا', 'uz': 'Angola', 'uz-Cyrl': 'Ангола', 'uz-Latn': 'Angola', 'vai': 'ꕉꖐꕞ', 'vai-Latn': 'Aŋgóla', 'vai-Vaii': 'ꕉꖐꕞ', 'vec': 'Angola', 'vi': 'Angola', 'vun': 'Angola', 'wae': 'Angola', 'wo': 'Àngolaa', 'xh': 'E-Angola', 'xnr': 'अंगोला', 'xog': 'Angola', 'yav': 'Aŋkúla', 'yi': 'אַנגאלע', 'yo': 'Ààngólà', 'yrl': 'Ãgura', 'yue': '安哥拉', 'yue-Hans': '安哥拉', 'yue-Hant': '安哥拉', 'zgh': 'ⴰⵏⴳⵓⵍⴰ', 'zh': '安哥拉', 'zh-Hans': '安哥拉', 'zh-Hant': '安哥拉', 'zu': 'i-Angola'},
    'AQ': {'af': 'Antarktika', 'ak': 'Antaatika', 'am': 'አንታርክቲካ', 'ar': 'أنتاركتيكا', 'as': 'এণ্টাৰ্কটিকা', 'ast': 'L’Antártida', 'az': 'Antarktika', 'az-Cyrl': 'Антарктика', 'az-Latn': 'Antarktika', 'bal-Latn': 'Antárktiká', 'be': 'Антарктыка', 'be-tarask': 'Антарктыка', 'bg': 'Антарктика', 'blo': 'Gatutaltǝna', 'bn': 'অ্যান্টার্কটিকা', 'br': 'Antarktika', 'brx': 'एन्टार्कटिका', 'bs': 'Antarktika', 'bs-Cyrl': 'Антарктик', 'bs-Latn': 'Antarktika', 'ca': 'Antàrtida', 'ca-ES-valencia': 'Antàrtida', 'ccp': '𑄃𑄳𑄠𑄚𑄴𑄑𑄢𑄴𑄇𑄧𑄑𑄨𑄇', 'ce': 'Антарктида', 'ceb': 'Antarctica', 'chr': 'ᏧᏁᏍᏓᎸ', 'ckb': 'ئانتارکتیکا', 'cs': 'Antarktida', 'cv': 'Антарктида', 'cy': 'Antarctica', 'da': 'Antarktis', 'de': 'Antarktis', 'dsb': 'Antarktis', 'dz': 'འཛམ་གླིང་ལྷོ་མཐའི་ཁྱགས་གླིང', 'ee': 'Antartica nutome', 'el': 'Ανταρκτική', 'el-polyton': 'Ἀνταρκτική', 'en': 'Antarctica', 'en-Dsrt': '𐐈𐑌𐐻𐐪𐑉𐐿𐐻𐐮𐐿𐐲', 'eo': 'Antarkto', 'es': 'Antártida', 'et': 'Antarktika', 'eu': 'Antartika', 'fa': 'جنوبگان', 'ff-Adlm': '𞤀𞤲𞤼𞤢𞤪𞤼𞤭𞤳𞤢𞥄', 'fi': 'Antarktis', 'fil': 'Antarctica', 'fo': 'Antarktis', 'fr': 'Antarctique', 'fur': 'Antartic', 'fy': 'Antarctica', 'ga': 'Antartaice', 'gd': 'An Antartaig', 'gl': 'Antártida', 'gsw': 'Antarktis', 'gu': 'એન્ટાર્કટિકા', 'ha': 'Antatika', 'he': 'אנטארקטיקה', 'hi': 'अंटार्कटिका', 'hi-Latn': 'Antarctica', 'hr': 'Antarktika', 'hsb': 'Antarktika', 'ht': 'Antarctique', 'hu': 'Antarktisz', 'hy': 'Անտարկտիդա', 'ia': 'Antarctica', 'id': 'Antarktika', 'ie': 'Antarctica', 'ig': 'Antarctica', 'is': 'Suðurskautslandið', 'it': 'Antartide', 'ja': '南極', 'jv': 'Antartika', 'ka': 'ანტარქტიკა', 'kea': 'Antártika', 'kgp': 'Ỹntartina', 'kk': 'Антарктида', 'kk-Cyrl': 'Антарктида', 'km': 'អង់តាក់ទិក', 'kn': 'ಅಂಟಾರ್ಟಿಕಾ', 'ko': '남극 대륙', 'kok': 'अंटार्क्टिका', 'kok-Deva': 'अंटार्क्टिका', 'ks': 'اینٹارٹِکا', 'ks-Arab': 'اینٹارٹِکا', 'ksh': 'der Södpol', 'ku': 'Antarktîka', 'kxv': 'ānṭarkṭikā', 'kxv-Deva': 'अंटार्कटिका', 'kxv-Latn': 'ānṭarkṭikā', 'kxv-Orya': 'ଆଣ୍ଟାର୍କଟିକା', 'kxv-Telu': 'ఆంటార్కటికా', 'ky': 'Антарктида', 'lb': 'Antarktis', 'lld': 'Antartida', 'ln': 'Antarctique', 'lo': 'ແອນຕາດຕິກາ', 'lt': 'Antarktida', 'lv': 'Antarktika', 'mai': 'अंटार्टिका', 'mi': 'Te Kōpakatanga ki te Tonga', 'mk': 'Антарктик', 'ml': 'അന്റാർട്ടിക്ക', 'mn': 'Антарктид', 'mr': 'अंटार्क्टिका', 'ms': 'Antartika', 'mt': 'l-Antartika', 'my': 'အန်တာတိက', 'mzn': 'جنوبی یخ\u200cبزه قطب', 'nb': 'Antarktis', 'ne': 'अन्टारटिका', 'nl': 'Antarctica', 'nn': 'Antarktis', 'no': 'Antarktis', 'nqo': 'ߊ߲ߕߊߙߑߛߕߌߞ', 'om': 'Antaarkitikaa', 'or': 'ଆଣ୍ଟାର୍କାଟିକା', 'pa': 'ਅੰਟਾਰਕਟਿਕਾ', 'pa-Guru': 'ਅੰਟਾਰਕਟਿਕਾ', 'pcm': 'Antáktíka', 'pl': 'Antarktyda', 'ps': 'انتارکتیکا', 'pt': 'Antártida', 'qu': 'Antártida', 'rm': 'Antarctica', 'ro': 'Antarctica', 'ru': 'Антарктида', 'sc': 'Antàrticu', 'sd': 'انٽارڪٽيڪا', 'sd-Arab': 'انٽارڪٽيڪا', 'se': 'Antárktis', 'si': 'ඇන්ටාක්ටිකාව', 'sk': 'Antarktída', 'sl': 'Antarktika', 'smn': 'Antarktis', 'so': 'Antaarktika', 'sq': 'Antarktikë', 'sr': 'Антарктик', 'sr-Cyrl': 'Антарктик', 'sr-Latn': 'Antarktik', 'sv': 'Antarktis', 'sw': 'Antaktiki', 'syr': 'ܐܢܬܪܬܝܟܐ', 'ta': 'அண்டார்டிகா', 'te': 'అంటార్కిటికా', 'tg': 'Антарктида', 'th': 'แอนตาร์กติกา', 'ti': 'ኣንታርክቲካ', 'tk': 'Antarktika', 'to': 'ʻAnitātika', 'tr': 'Antarktika', 'tt': 'Антарктика', 'ug': 'ئانتاركتىكا', 'uk': 'Антарктика', 'ur': 'انٹارکٹیکا', 'uz': 'Antarktida', 'uz-Cyrl': 'Антарктида', 'uz-Latn': 'Antarktida', 'vai': 'ꕉꘋꕚꔳꕪ', 'vai-Vaii': 'ꕉꘋꕚꔳꕪ', 'vec': 'Antàrtide', 'vi': 'Nam Cực', 'wae': 'Antarktis', 'wo': 'Antarktik', 'xh': 'E-Antarctica', 'xnr': 'अंटार्कटिका', 'yi': 'אַנטאַרקטיקע', 'yo': 'Antakítíkà', 'yrl': 'Ãtartida', 'yue': '南極洲', 'yue-Hans': '南极洲', 'yue-Hant': '南極洲', 'zh': '南极洲', 'zh-Hans': '南极洲', 'zh-Hant': '南極洲', 'zu': 'i-Antarctica'},
    'AR': {'af': 'Argentinië', 'agq': 'Àdzɛ̀ntinà', 'ak': 'Agyɛntina', 'am': 'አርጀንቲና', 'ar': 'الأرجنتين', 'as': 'আৰ্জেণ্টিনা', 'asa': 'Ajentina', 'ast': 'Arxentina', 'az': 'Argentina', 'az-Cyrl': 'Арҝентина', 'az-Latn': 'Argentina', 'bal-Latn': 'Arjentiná', 'bas': 'Àrgàŋtinà', 'be': 'Аргенціна', 'be-tarask': 'Аргенціна', 'bez': 'Huajendina', 'bg': 'Аржентина', 'bgn': 'ارجنٹاین', 'blo': 'Arjantin', 'bm': 'Arizantin', 'bn': 'আর্জেন্টিনা', 'br': 'Arcʼhantina', 'brx': 'आर्जेन्टिना', 'bs': 'Argentina', 'bs-Cyrl': 'Аргентина', 'bs-Latn': 'Argentina', 'ca': 'Argentina', 'ca-ES-valencia': 'Argentina', 'ccp': '𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚', 'ce': 'Аргентина', 'ceb': 'Argentina', 'cgg': 'Arigentina', 'chr': 'ᎠᏥᏂᏘᏂᎠ', 'ckb': 'ئەرژەنتین', 'cs': 'Argentina', 'cv': 'Аргентина', 'cy': 'Yr Ariannin', 'da': 'Argentina', 'dav': 'Ajentina', 'de': 'Argentinien', 'dje': 'Argentine', 'dsb': 'Argentinska', 'dyo': 'Arsantin', 'dz': 'ཨར་ཇེན་ཊི་ན', 'ebu': 'Ajentina', 'ee': 'Argentina nutome', 'el': 'Αργεντινή', 'el-polyton': 'Ἀργεντινή', 'en': 'Argentina', 'en-Dsrt': '𐐂𐑉𐐾𐐲𐑌𐐻𐐨𐑌𐐲', 'eo': 'Argentino', 'es': 'Argentina', 'et': 'Argentina', 'eu': 'Argentina', 'ewo': 'Arəhenətína', 'fa': 'آرژانتین', 'ff': 'Arjantiin', 'ff-Adlm': '𞤀𞤪𞤶𞤢𞤲𞤼𞤭𞤲𞤢𞥄', 'ff-Latn': 'Arjantiin', 'fi': 'Argentiina', 'fil': 'Argentina', 'fo': 'Argentina', 'fr': 'Argentine', 'fur': 'Argjentine', 'fy': 'Argentinië', 'ga': 'an Airgintín', 'gd': 'An Argantain', 'gl': 'Arxentina', 'gsw': 'Argentiinie', 'gu': 'આર્જેન્ટીના', 'guz': 'Ajentina', 'ha': 'Ajentina', 'he': 'ארגנטינה', 'hi': 'अर्जेंटीना', 'hi-Latn': 'Argentina', 'hr': 'Argentina', 'hsb': 'Argentinska', 'ht': 'Argentine', 'hu': 'Argentína', 'hy': 'Արգենտինա', 'ia': 'Argentina', 'id': 'Argentina', 'ig': 'Argentina', 'is': 'Argentína', 'it': 'Argentina', 'ja': 'アルゼンチン', 'jgo': 'Ajɛntîn', 'jmc': 'Ajentina', 'jv': 'Argèntina', 'ka': 'არგენტინა', 'kaa': 'Аргентина', 'kaa-Cyrl': 'Аргентина', 'kab': 'Arjuntin', 'kam': 'Ajentina', 'kde': 'Adyentina', 'kea': 'Arjentina', 'kgp': 'Arjẽtinỹ', 'khq': 'Argentine', 'ki': 'Ajentina', 'kk': 'Аргентина', 'kk-Cyrl': 'Аргентина', 'kln': 'Emetab Argentina', 'km': 'អាហ្សង់ទីន', 'kn': 'ಅರ್ಜೆಂಟಿನಾ', 'ko': '아르헨티나', 'kok': 'अर्जेंटिना', 'kok-Deva': 'अर्जेंटिना', 'ks': 'أرجَنٹینا', 'ks-Arab': 'أرجَنٹینا', 'ksb': 'Ajentina', 'ksf': 'arjǝntín', 'ksh': 'Ajjentiinije', 'ku': 'Arjantîn', 'kxv': 'aarhenṭina', 'kxv-Deva': 'अर्जेंटीना', 'kxv-Latn': 'aarhenṭina', 'kxv-Orya': 'ଅର୍ଜେଣ୍ଟିନା', 'kxv-Telu': 'ఆర్జెంటినా', 'ky': 'Аргентина', 'lag': 'Ajentíina', 'lb': 'Argentinien', 'lg': 'Arigentina', 'lld': 'Argentina', 'ln': 'Arizantinɛ', 'lo': 'ອາເຈນທິນາ', 'lt': 'Argentina', 'lu': 'Alijantine', 'luo': 'Argentina', 'luy': 'Ajentina', 'lv': 'Argentīna', 'mai': 'अर्जेंटिना', 'mas': 'Ajentina', 'mer': 'Ajentina', 'mfe': 'Larzantinn', 'mg': 'Arzantina', 'mi': 'Āketina', 'mk': 'Аргентина', 'ml': 'അർജന്റീന', 'mn': 'Аргентин', 'mr': 'अर्जेंटिना', 'ms': 'Argentina', 'mt': 'l-Arġentina', 'mua': 'argentiniya', 'my': 'အာဂျင်တီးနား', 'mzn': 'آرژانتین', 'naq': 'Argentinab', 'nb': 'Argentina', 'nd': 'Ajentina', 'ne': 'अर्जेन्टिना', 'nl': 'Argentinië', 'nmg': 'Argentína', 'nn': 'Argentina', 'no': 'Argentina', 'nqo': 'ߊߙߑߖ߭ߊ߲ߕߌߣ', 'nus': 'Aɛrgentin', 'nyn': 'Arigentina', 'om': 'Arjentiinaa', 'or': 'ଆର୍ଜେଣ୍ଟିନା', 'pa': 'ਅਰਜਨਟੀਨਾ', 'pa-Guru': 'ਅਰਜਨਟੀਨਾ', 'pcm': 'Ajẹntína', 'pl': 'Argentyna', 'ps': 'ارجنټاين', 'pt': 'Argentina', 'qu': 'Argentina', 'rm': 'Argentinia', 'rn': 'Arijantine', 'ro': 'Argentina', 'rof': 'Ajentina', 'ru': 'Аргентина', 'rwk': 'Ajentina', 'saq': 'Ajentina', 'sbp': 'Ajentina', 'sc': 'Argentina', 'sd': 'ارجنٽينا', 'sd-Arab': 'ارجنٽينا', 'se': 'Argentina', 'seh': 'Argentina', 'ses': 'Argentine', 'sg': 'Arzantîna', 'shi': 'ⴰⵔⵊⴰⵏⵜⵉⵏ', 'shi-Latn': 'arjantin', 'shi-Tfng': 'ⴰⵔⵊⴰⵏⵜⵉⵏ', 'si': 'ආර්ජෙන්ටිනාව', 'sk': 'Argentína', 'sl': 'Argentina', 'smn': 'Argentina', 'sn': 'Ajentina', 'so': 'Arjentiina', 'sq': 'Argjentinë', 'sr': 'Аргентина', 'sr-Cyrl': 'Аргентина', 'sr-Latn': 'Argentina', 'sv': 'Argentina', 'sw': 'Ajentina', 'syr': 'ܐܪܓܢܬܝܢܐ', 'ta': 'அர்ஜென்டினா', 'te': 'అర్జెంటీనా', 'teo': 'Ajentina', 'tg': 'Аргентина', 'th': 'อาร์เจนตินา', 'ti': 'ኣርጀንቲና', 'tk': 'Argentina', 'to': 'ʻAsenitina', 'tr': 'Arjantin', 'tt': 'Аргентина', 'twq': 'Argentine', 'tzm': 'Arjuntin', 'ug': 'ئارگېنتىنا', 'uk': 'Аргентина', 'ur': 'ارجنٹینا', 'uz': 'Argentina', 'uz-Cyrl': 'Аргентина', 'uz-Latn': 'Argentina', 'vai': 'ꕉꘀꘋꔳꕯ', 'vai-Latn': 'Ajɛŋtína', 'vai-Vaii': 'ꕉꘀꘋꔳꕯ', 'vec': 'Arjentina', 'vi': 'Argentina', 'vun': 'Ajentina', 'wae': 'Argentinie', 'wo': 'Arsàntin', 'xh': 'E-Argentina', 'xnr': 'अर्जेंटीना', 'xog': 'Arigentina', 'yav': 'Alsaŋtín', 'yi': 'אַרגענטינע', 'yo': 'Agentínà', 'yrl': 'Argẽtina', 'yue': '阿根廷', 'yue-Hans': '阿根廷', 'yue-Hant': '阿根廷', 'zgh': 'ⴰⵔⵊⴰⵏⵜⵉⵏ', 'zh': '阿根廷', 'zh-Hans': '阿根廷', 'zh-Hant': '阿根廷', 'zu': 'i-Argentina'},
    'AS': {'af': 'Amerikaanse Samoa', 'agq': 'Àmɛlekan Samwà', 'ak': 'Amɛrika Samoa', 'am': 'የአሜሪካ ሳሞአ', 'ar': 'ساموا الأمريكية', 'as': 'আমেৰিকান চামোৱা', 'asa': 'Thamoa ya Marekani', 'ast': 'Samoa Americana', 'az': 'Amerika Samoası', 'az-Cyrl': 'Америка Самоасы', 'az-Latn': 'Amerika Samoası', 'bal-Latn': 'Amriki Samóá', 'be': 'Амерыканскае Самоа', 'be-tarask': 'Амерыканскае Самоа', 'bez': 'Husamoa ya Humalekani', 'bg': 'Американска Самоа', 'blo': 'Samowa Amalɩka kaja', 'bm': 'Samowa amerikani', 'bn': 'আমেরিকান সামোয়া', 'br': 'Samoa Amerikan', 'brx': 'आमेरिकान सामआ', 'bs': 'Američka Samoa', 'bs-Cyrl': 'Америчка Самоа', 'bs-Latn': 'Američka Samoa', 'ca': 'Samoa Americana', 'ca-ES-valencia': 'Samoa Americana', 'ccp': '𑄃𑄟𑄬𑄢𑄨𑄇𑄚𑄴 𑄥𑄟𑄮𑄠', 'ce': 'Американ Самоа', 'ceb': 'American Samoa', 'cgg': 'Samowa ya Ameerika', 'chr': 'ᎠᎺᎵᎧ ᏌᎼᎠ', 'ckb': 'ساموای ئەمەریکایی', 'cs': 'Americká Samoa', 'cv': 'Америка Самоа', 'cy': 'Samoa America', 'da': 'Amerikansk Samoa', 'dav': 'Samoa ya Marekani', 'de': 'Amerikanisch-Samoa', 'dje': 'Ameriki Samoa', 'dsb': 'Ameriska Samoa', 'dyo': 'Samoa yati Amerik', 'dz': 'ས་མོ་ཨ་ཡུ་ཨེས་ཨེ་མངའ་ཁོངས', 'ebu': 'Samoa ya Marekani', 'ee': 'Amerika Samoa nutome', 'el': 'Αμερικανική Σαμόα', 'el-polyton': 'Ἀμερικανικὴ Σαμόα', 'en': 'American Samoa', 'en-Dsrt': '𐐊𐑋𐐯𐑉𐐲𐐿𐐲𐑌 𐐝𐐲𐑋𐐬𐐲', 'es': 'Samoa Americana', 'et': 'Ameerika Samoa', 'eu': 'Samoa Estatubatuarra', 'ewo': 'Bəsamóa yá Amə́rəka', 'fa': 'ساموآی امریکا', 'ff': 'Samowa Amerik', 'ff-Adlm': '𞤅𞤢𞤥𞤵𞤱𞤢 𞤀𞤥𞤫𞤪𞤭𞤳𞤭𞤴𞤢𞤲𞤳𞤮', 'ff-Latn': 'Samowa Amerik', 'fi': 'Amerikan Samoa', 'fil': 'American Samoa', 'fo': 'Amerikanska Samoa', 'fr': 'Samoa américaines', 'fur': 'Samoa merecanis', 'fy': 'Amerikaansk Samoa', 'ga': 'Samó Mheiriceá', 'gd': 'Samotha na h-Aimeireaga', 'gl': 'Samoa Americana', 'gsw': 'Amerikaanisch-Samoa', 'gu': 'અમેરિકન સમોઆ', 'guz': 'Samoa ya Marekani', 'ha': 'Samowa Ta Amurka', 'he': 'סמואה האמריקנית', 'hi': 'अमेरिकी समोआ', 'hi-Latn': 'American Samoa', 'hr': 'Američka Samoa', 'hsb': 'Ameriska Samoa', 'ht': 'Samoa américaines', 'hu': 'Amerikai Szamoa', 'hy': 'Ամերիկյան Սամոա', 'ia': 'Samoa american', 'id': 'Samoa Amerika', 'ig': 'American Samoa', 'is': 'Bandaríska Samóa', 'it': 'Samoa Americane', 'ja': '米領サモア', 'jmc': 'Samoa ya Marekani', 'jv': 'Samoa Amerika', 'ka': 'ამერიკის სამოა', 'kab': 'Samwa Tamarikanit', 'kam': 'Samoa ya Marekani', 'kde': 'Samoa ya Malekani', 'kea': 'Samoa Merkanu', 'kgp': 'Samãva Amẽrikynỹ', 'khq': 'Ameriki Samoa', 'ki': 'Samoa ya Marekani', 'kk': 'Америкалық Самоа', 'kk-Cyrl': 'Америкалық Самоа', 'kln': 'Emetab American Samoa', 'km': 'សាម័រ អាមេរិកាំង', 'kn': 'ಅಮೇರಿಕನ್ ಸಮೋವಾ', 'ko': '아메리칸 사모아', 'kok': 'अमेरिकी सामोआ', 'kok-Deva': 'अमेरिकी सामोआ', 'ks': 'اَمریٖکَن سَموا', 'ks-Arab': 'اَمریٖکَن سَموا', 'ksb': 'Samoa ya Malekani', 'ksf': 'samɔa a amɛrika', 'ksh': 'Ammerikaanesch Sammoha', 'ku': 'Samoaya Amerîkanî', 'kxv': 'amerikati samva', 'kxv-Deva': 'आमेरिका-ति समोआ', 'kxv-Latn': 'amerikati samva', 'kxv-Orya': 'ଅମେରିକା ତି ସମୱା', 'kxv-Telu': 'ఆమెరికన్ సమూవా', 'ky': 'Америкалык Самоа', 'lag': 'Samóoa ya Amerɨ́ka', 'lb': 'Amerikanesch-Samoa', 'lg': 'Samowa omumerika', 'lld': 'Samoa americana', 'ln': 'Samoa ya Ameriki', 'lo': 'ອາເມຣິກາ ຊາມົວ', 'lt': 'Amerikos Samoa', 'lu': 'Samoa wa Ameriki', 'luo': 'American Samoa', 'luy': 'Samoa lya Marekani', 'lv': 'ASV Samoa', 'mai': 'अमेरिकी समोआ', 'mas': 'Samoa ya Marekani', 'mer': 'Samoa ya Amerika', 'mfe': 'Samoa amerikin', 'mg': 'Samoa amerikanina', 'mgh': 'Usamoa ya Marekani', 'mi': 'Hāmoa-Amerika', 'mk': 'Американска Самоа', 'ml': 'അമേരിക്കൻ സമോവ', 'mn': 'Америкийн Самоа', 'mr': 'अमेरिकन सामोआ', 'ms': 'Samoa Amerika', 'mt': 'is-Samoa Amerikana', 'mua': 'samoa Amerika', 'my': 'အမေရိကန် ဆမိုးအား', 'mzn': 'آمریکای ِساموآ', 'naq': 'Americab Samoab', 'nb': 'Amerikansk Samoa', 'nd': 'Samoa ye Amelika', 'ne': 'अमेरिकी समोआ', 'nl': 'Amerikaans-Samoa', 'nmg': 'Samoa m ́Amɛ́rka', 'nn': 'Amerikansk Samoa', 'no': 'Amerikansk Samoa', 'nqo': 'ߛߊߡߏߥߊ߫ ߊߡߋߙߞߌߞߊ', 'nus': 'Amerika thamow', 'nyn': 'Samowa ya Ameerika', 'om': 'Saamowa Ameerikaa', 'or': 'ଆମେରିକାନ୍ ସାମୋଆ', 'pa': 'ਅਮੈਰੀਕਨ ਸਮੋਆ', 'pa-Guru': 'ਅਮੈਰੀਕਨ ਸਮੋਆ', 'pcm': 'Amẹ́ríká Samoa', 'pl': 'Samoa Amerykańskie', 'ps': 'امریکایی ساماوا', 'pt': 'Samoa Americana', 'qu': 'Samoa Americana', 'rm': 'Samoa Americana', 'rn': 'Samowa nyamerika', 'ro': 'Samoa Americană', 'rof': 'Samoa ya Marekani', 'ru': 'Американское Самоа', 'rwk': 'Samoa ya Marekani', 'saq': 'Samoa ya Marekani', 'sbp': 'Samoya ya Malekani', 'sc': 'Samoa americanas', 'sd': 'آمريڪي ساموا', 'sd-Arab': 'آمريڪي ساموا', 'se': 'Amerihká Samoa', 'seh': 'Samoa Americana', 'ses': 'Ameriki Samoa', 'sg': 'Samöa tî Amerîka', 'shi': 'ⵙⴰⵎⵡⴰ ⵜⴰⵎⵉⵔⵉⴽⴰⵏⵉⵜ', 'shi-Latn': 'samwa tamirikanit', 'shi-Tfng': 'ⵙⴰⵎⵡⴰ ⵜⴰⵎⵉⵔⵉⴽⴰⵏⵉⵜ', 'si': 'ඇමරිකානු සැමෝවා', 'sk': 'Americká Samoa', 'sl': 'Ameriška Samoa', 'smn': 'Amerika Samoa', 'sn': 'Samoa ye Amerika', 'so': 'Samowa Ameerika', 'sq': 'Samoa Amerikane', 'sr': 'Америчка Самоа', 'sr-Cyrl': 'Америчка Самоа', 'sr-Latn': 'Američka Samoa', 'sv': 'Amerikanska Samoa', 'sw': 'Samoa ya Marekani', 'syr': 'ܣܡܘܐ ܐܡܝܖ̈ܟܝܐ', 'ta': 'அமெரிக்க சமோவா', 'te': 'అమెరికన్ సమోవా', 'teo': 'Samoa ya Marekani', 'tg': 'Самоаи Америка', 'th': 'อเมริกันซามัว', 'ti': 'ኣመሪካዊት ሳሞኣ', 'tk': 'Amerikan Samoasy', 'to': 'Haʻamoa ʻAmelika', 'tr': 'Amerikan Samoası', 'tt': 'Америка Самоасы', 'twq': 'Ameriki Samoa', 'tzm': 'Samwa Imirikaniyyin', 'ug': 'ئامېرىكا ساموئا', 'uk': 'Американське Самоа', 'ur': 'امریکی ساموآ', 'uz': 'Amerika Samoasi', 'uz-Cyrl': 'Америка Самоаси', 'uz-Latn': 'Amerika Samoasi', 'vai': 'ꕶꕱ ꕢꕹꕎ', 'vai-Latn': 'Poo Sambowa', 'vai-Vaii': 'ꕶꕱ ꕢꕹꕎ', 'vec': 'Samòa meregane', 'vi': 'Samoa thuộc Mỹ', 'vun': 'Samoa ya Marekani', 'wae': 'Amerikaniš Samoa', 'wo': 'Samowa bu Amerig', 'xh': 'E-American Samoa', 'xnr': 'अमरिकी समोआ', 'xog': 'Samowa omumerika', 'yav': 'Sámua u Amelíka', 'yo': 'Sámóánì ti Orílẹ́ède Àméríkà', 'yrl': 'Samua Amerikiwara', 'yue': '美屬薩摩亞', 'yue-Hans': '美属萨摩亚', 'yue-Hant': '美屬薩摩亞', 'zgh': 'ⵙⴰⵎⵡⴰ ⵜⴰⵎⵉⵔⵉⴽⴰⵏⵉⵜ', 'zh': '美属萨摩亚', 'zh-Hans': '美属萨摩亚', 'zh-Hant': '美屬薩摩亞', 'zu': 'i-American Samoa'},
    'AT': {'af': 'Oostenryk', 'agq': 'Usɨtɨ̀là', 'ak': 'Ɔstria', 'am': 'ኦስትሪያ', 'ar': 'النمسا', 'as': 'অষ্ট্ৰিয়া', 'asa': 'Authtria', 'ast': 'Austria', 'az': 'Avstriya', 'az-Cyrl': 'Австрија', 'az-Latn': 'Avstriya', 'bal-Latn': 'Ástriá', 'bas': 'Òstrǐk', 'be': 'Аўстрыя', 'be-tarask': 'Аўстрыя', 'bez': 'Huastlia', 'bg': 'Австрия', 'blo': 'Otrish', 'bm': 'Otirisi', 'bn': 'অস্ট্রিয়া', 'br': 'Aostria', 'brx': 'अस्ट्रिया', 'bs': 'Austrija', 'bs-Cyrl': 'Аустрија', 'bs-Latn': 'Austrija', 'ca': 'Àustria', 'ca-ES-valencia': 'Àustria', 'ccp': '𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄨𑄠', 'ce': 'Австри', 'ceb': 'Austria', 'cgg': 'Osituria', 'chr': 'ᎠᏍᏟᏯ', 'ckb': 'نەمسا', 'cs': 'Rakousko', 'cv': 'Австри', 'cy': 'Awstria', 'da': 'Østrig', 'dav': 'Austria', 'de': 'Österreich', 'dje': 'Otriši', 'dsb': 'Awstriska', 'dyo': 'Otris', 'dz': 'ཨཱོས་ཊྲི་ཡ', 'ebu': 'Austria', 'ee': 'Austria nutome', 'el': 'Αυστρία', 'el-polyton': 'Αὐστρία', 'en': 'Austria', 'en-Dsrt': '𐐉𐑅𐐻𐑉𐐨𐐲', 'eo': 'Aŭstrujo', 'es': 'Austria', 'et': 'Austria', 'eu': 'Austria', 'ewo': 'Osətəlía', 'fa': 'اتریش', 'ff': 'Otiriis', 'ff-Adlm': '𞤌𞤼𞤭𞤪𞤧𞤢', 'ff-Latn': 'Otiriis', 'fi': 'Itävalta', 'fil': 'Austria', 'fo': 'Eysturríki', 'fr': 'Autriche', 'fur': 'Austrie', 'fy': 'Eastenryk', 'ga': 'an Ostair', 'gd': 'An Ostair', 'gl': 'Austria', 'gsw': 'Ööschtriich', 'gu': 'ઑસ્ટ્રિયા', 'guz': 'Austria', 'ha': 'Ostiriya', 'he': 'אוסטריה', 'hi': 'ऑस्ट्रिया', 'hi-Latn': 'Austria', 'hr': 'Austrija', 'hsb': 'Awstriska', 'ht': 'Autriche', 'hu': 'Ausztria', 'hy': 'Ավստրիա', 'ia': 'Austria', 'id': 'Austria', 'ie': 'Austria', 'ig': 'Austria', 'is': 'Austurríki', 'it': 'Austria', 'ja': 'オーストリア', 'jmc': 'Austria', 'jv': 'Ostenrik', 'ka': 'ავსტრია', 'kaa': 'Австрия', 'kaa-Cyrl': 'Австрия', 'kab': 'Ustriya', 'kam': 'Austria', 'kde': 'Austlia', 'kea': 'Áustria', 'kgp': 'Agtirija', 'khq': 'Otriši', 'ki': 'Austria', 'kk': 'Аустрия', 'kk-Cyrl': 'Аустрия', 'kln': 'Emetab Austria', 'km': 'អូទ្រីស', 'kn': 'ಆಸ್ಟ್ರಿಯಾ', 'ko': '오스트리아', 'kok': 'ऑस्ट्रिया', 'kok-Deva': 'ऑस्ट्रिया', 'ks': 'آسٹریا', 'ks-Arab': 'آسٹریا', 'ksb': 'Austlia', 'ksf': 'otric', 'ksh': 'Öösterisch', 'ku': 'Awistirya', 'kxv': 'austria', 'kxv-Deva': 'ऑस्ट्रिया', 'kxv-Latn': 'austria', 'kxv-Orya': 'ଅଷ୍ଟ୍ରିଆ', 'kxv-Telu': 'ఆస్ట్రీయా', 'ky': 'Австрия', 'lag': 'Áusitiria', 'lb': 'Éisträich', 'lg': 'Awusituriya', 'lld': 'Austria', 'ln': 'Otilisi', 'lo': 'ອອສເທຣຍ', 'lt': 'Austrija', 'lu': 'Otilisi', 'luo': 'Austria', 'luy': 'Austria', 'lv': 'Austrija', 'mai': 'आस्ट्रिया', 'mas': 'Austria', 'mer': 'Austiria', 'mfe': 'Lostris', 'mg': 'Aotrisy', 'mi': 'Ataria', 'mk': 'Австрија', 'ml': 'ഓസ്ട്രിയ', 'mn': 'Австри', 'mr': 'ऑस्ट्रिया', 'ms': 'Austria', 'mt': 'l-Awstrija', 'mua': 'austriya', 'my': 'ဩစတြီးယား', 'mzn': 'اتریش', 'naq': 'Austriab', 'nb': 'Østerrike', 'nd': 'Austria', 'ne': 'अष्ट्रिया', 'nl': 'Oostenrijk', 'nmg': 'Ötrish', 'nn': 'Austerrike', 'no': 'Østerrike', 'nqo': 'ߏߕߑߙߌߛ', 'nus': 'Athtɛria', 'nyn': 'Osituria', 'om': 'Awustiriyaa', 'or': 'ଅଷ୍ଟ୍ରିଆ', 'pa': 'ਆਸਟਰੀਆ', 'pa-Guru': 'ਆਸਟਰੀਆ', 'pcm': 'Ọ́stria', 'pl': 'Austria', 'ps': 'اتریش', 'pt': 'Áustria', 'qu': 'Austria', 'rm': 'Austria', 'rn': 'Otirishe', 'ro': 'Austria', 'rof': 'Ostria', 'ru': 'Австрия', 'rwk': 'Austria', 'saq': 'Austria', 'sbp': 'Awusitiliya', 'sc': 'Àustria', 'sd': 'آسٽريا', 'sd-Arab': 'آسٽريا', 'se': 'Nuortariika', 'seh': 'Áustria', 'ses': 'Otriši', 'sg': 'Otrîsi', 'shi': 'ⵏⵏⵎⵙⴰ', 'shi-Latn': 'nnmsa', 'shi-Tfng': 'ⵏⵏⵎⵙⴰ', 'si': 'ඔස්ට්\u200dරියාව', 'sk': 'Rakúsko', 'sl': 'Avstrija', 'smn': 'Nuorttâriijkâ', 'sn': 'Austria', 'so': 'Awsteriya', 'sq': 'Austri', 'sr': 'Аустрија', 'sr-Cyrl': 'Аустрија', 'sr-Latn': 'Austrija', 'sv': 'Österrike', 'sw': 'Austria', 'syr': 'ܐܘܣܛܪܝܐ', 'ta': 'ஆஸ்திரியா', 'te': 'ఆస్ట్రియా', 'teo': 'Austria', 'tg': 'Австрия', 'th': 'ออสเตรีย', 'ti': 'ኦስትርያ', 'tk': 'Awstriýa', 'to': 'ʻAositulia', 'tr': 'Avusturya', 'tt': 'Австрия', 'twq': 'Otriši', 'tzm': 'Ustriyya', 'ug': 'ئاۋىستىرىيە', 'uk': 'Австрія', 'ur': 'آسٹریا', 'uz': 'Avstriya', 'uz-Cyrl': 'Австрия', 'uz-Latn': 'Avstriya', 'vai': 'ꖺꔻꖤꕎ', 'vai-Latn': 'Ɔ́situwa', 'vai-Vaii': 'ꖺꔻꖤꕎ', 'vec': 'Àustria', 'vi': 'Áo', 'vun': 'Austria', 'wae': 'Öštrič', 'wo': 'Ótiriis', 'xh': 'E-Austria', 'xnr': 'ऑस्ट्रिया', 'xog': 'Awusituriya', 'yav': 'Otilís', 'yi': 'עסטרייך', 'yo': 'Asítíríà', 'yrl': 'Ausitiriya', 'yue': '奧地利', 'yue-Hans': '奥地利', 'yue-Hant': '奧地利', 'zgh': 'ⵏⵏⵎⵙⴰ', 'zh': '奥地利', 'zh-Hans': '奥地利', 'zh-Hant': '奧地利', 'zu': 'i-Austria'},
    'AU': {'af': 'Australië', 'agq': 'Ùsɨ̀tɛ̀lɛlìa', 'ak': 'Ɔstrelia', 'am': 'አውስትራልያ', 'ar': 'أستراليا', 'as': 'অষ্ট্ৰেলিয়া', 'asa': 'Authtralia', 'ast': 'Australia', 'az': 'Avstraliya', 'az-Cyrl': 'Австралија', 'az-Latn': 'Avstraliya', 'bal-Latn': 'Ásthréliá', 'bas': 'Òstralìà', 'be': 'Аўстралія', 'be-tarask': 'Аўстралія', 'bez': 'Huaustlalia', 'bg': 'Австралия', 'bgn': 'اسٹرالیا', 'blo': 'Ɔstraliya', 'bm': 'Ositirali', 'bn': 'অস্ট্রেলিয়া', 'br': 'Aostralia', 'brx': 'अस्ट्रेलिया', 'bs': 'Australija', 'bs-Cyrl': 'Аустралија', 'bs-Latn': 'Australija', 'ca': 'Austràlia', 'ca-ES-valencia': 'Austràlia', 'ccp': '𑄃𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠', 'ce': 'Австрали', 'ceb': 'Australia', 'cgg': 'Ositureeriya', 'chr': 'ᎡᎳᏗᏜ', 'ckb': 'ئوسترالیا', 'cs': 'Austrálie', 'cv': 'Австрали', 'cy': 'Awstralia', 'da': 'Australien', 'dav': 'Australia', 'de': 'Australien', 'dje': 'Ostraali', 'dsb': 'Awstralska', 'dyo': 'Ostraalia', 'dz': 'ཨཱོས་ཊྲེལ་ལི་ཡ', 'ebu': 'Australia', 'ee': 'Australia nutome', 'el': 'Αυστραλία', 'el-polyton': 'Αὐστραλία', 'en': 'Australia', 'en-Dsrt': '𐐉𐑅𐐻𐑉𐐩𐑊𐐨𐐲', 'eo': 'Aŭstralio', 'es': 'Australia', 'et': 'Austraalia', 'eu': 'Australia', 'ewo': 'Osətəlalí', 'fa': 'استرالیا', 'ff': 'Ostaraalii', 'ff-Adlm': '𞤌𞤧𞤼𞤢𞤪𞤤𞤭𞤴𞤢𞥄', 'ff-Latn': 'Ostaraalii', 'fi': 'Australia', 'fil': 'Australia', 'fo': 'Avstralia', 'fr': 'Australie', 'fur': 'Australie', 'fy': 'Australië', 'ga': 'an Astráil', 'gd': 'Astràilia', 'gl': 'Australia', 'gsw': 'Auschtraalie', 'gu': 'ઑસ્ટ્રેલિયા', 'guz': 'Australia', 'ha': 'Ostareliya', 'haw': 'Nūhōlani', 'he': 'אוסטרליה', 'hi': 'ऑस्ट्रेलिया', 'hi-Latn': 'Australia', 'hr': 'Australija', 'hsb': 'Awstralska', 'ht': 'Australie', 'hu': 'Ausztrália', 'hy': 'Ավստրալիա', 'ia': 'Australia', 'id': 'Australia', 'ig': 'Australia', 'is': 'Ástralía', 'it': 'Australia', 'ja': 'オーストラリア', 'jmc': 'Australia', 'jv': 'Ostrali', 'ka': 'ავსტრალია', 'kaa': 'Австралия', 'kaa-Cyrl': 'Австралия', 'kab': 'Ustrali', 'kam': 'Australia', 'kde': 'Austlalia', 'kea': 'Austrália', 'kgp': 'Avotyraria', 'khq': 'Ostraali', 'ki': 'Australia', 'kk': 'Аустралия', 'kk-Cyrl': 'Аустралия', 'kln': 'Emetab Australia', 'km': 'អូស្ត្រាលី', 'kn': 'ಆಸ್ಟ್ರೇಲಿಯಾ', 'ko': '오스트레일리아', 'kok': 'ऑस्ट्रेलीया', 'kok-Deva': 'ऑस्ट्रेलीया', 'ks': 'آسٹریلِیا', 'ks-Arab': 'آسٹریلِیا', 'ksb': 'Austlalia', 'ksf': 'ɔstralí', 'ksh': 'Austraalije', 'ku': 'Awistralya', 'kxv': 'astrelia', 'kxv-Deva': 'ऑस्ट्रेलिया', 'kxv-Latn': 'astrelia', 'kxv-Orya': 'ଅଷ୍ଟ୍ରେଲିଆ', 'kxv-Telu': 'ఆస్ట్రేలియా', 'ky': 'Австралия', 'lag': 'Ausiteréelia', 'lb': 'Australien', 'lg': 'Awusitureliya', 'lld': 'Australia', 'ln': 'Ositáli', 'lo': 'ອອສເຕຣເລຍ', 'lt': 'Australija', 'lu': 'Ositali', 'luo': 'Australia', 'luy': 'Australia', 'lv': 'Austrālija', 'mai': 'आस्ट्रेलिया', 'mas': 'Australia', 'mer': 'Austrĩlia', 'mfe': 'Lostrali', 'mg': 'Aostralia', 'mi': 'Ahitereiria', 'mk': 'Австралија', 'ml': 'ഓസ്\u200cട്രേലിയ', 'mn': 'Австрали', 'mr': 'ऑस्ट्रेलिया', 'ms': 'Australia', 'mt': 'l-Awstralja', 'mua': 'australiya', 'my': 'ဩစတြေးလျ', 'mzn': 'استرالیا', 'naq': 'Australieb', 'nb': 'Australia', 'nd': 'Australia', 'ne': 'अस्ट्रेलिया', 'nl': 'Australië', 'nmg': 'Östraliá', 'nn': 'Australia', 'no': 'Australia', 'nqo': 'ߐߛߑߕߙߊߟߌ߫', 'nus': 'Athɔra̱lia', 'nyn': 'Ositureeriya', 'om': 'Awustiraaliyaa', 'or': 'ଅଷ୍ଟ୍ରେଲିଆ', 'pa': 'ਆਸਟ੍ਰੇਲੀਆ', 'pa-Guru': 'ਆਸਟ੍ਰੇਲੀਆ', 'pcm': 'Ọstrélia', 'pl': 'Australia', 'ps': 'آسټرالیا', 'pt': 'Austrália', 'qu': 'Australia', 'rm': 'Australia', 'rn': 'Ositaraliya', 'ro': 'Australia', 'rof': 'Australia', 'ru': 'Австралия', 'rwk': 'Australia', 'saq': 'Australia', 'sbp': 'Awusitilaliya', 'sc': 'Austràlia', 'sd': 'آسٽريليا', 'sd-Arab': 'آسٽريليا', 'se': 'Austrália', 'seh': 'Austrália', 'ses': 'Ostraali', 'sg': 'Ostralïi, Sotralïi', 'shi': 'ⵓⵙⵜⵔⴰⵍⵢⴰ', 'shi-Latn': 'ustralya', 'shi-Tfng': 'ⵓⵙⵜⵔⴰⵍⵢⴰ', 'si': 'ඕස්ට්\u200dරේලියාව', 'sk': 'Austrália', 'sl': 'Avstralija', 'smn': 'Australia', 'sn': 'Australia', 'so': 'Awstaraaliya', 'sq': 'Australi', 'sr': 'Аустралија', 'sr-Cyrl': 'Аустралија', 'sr-Latn': 'Australija', 'sv': 'Australien', 'sw': 'Australia', 'syr': 'ܐܘܣܬܪܠܝܐ', 'ta': 'ஆஸ்திரேலியா', 'te': 'ఆస్ట్రేలియా', 'teo': 'Australia', 'tg': 'Австралия', 'th': 'ออสเตรเลีย', 'ti': 'ኣውስትራልያ', 'tk': 'Awstraliýa', 'to': 'ʻAositelēlia', 'tr': 'Avustralya', 'tt': 'Австралия', 'twq': 'Ostraali', 'tzm': 'Usṭralya', 'ug': 'ئاۋسترالىيە', 'uk': 'Австралія', 'ur': 'آسٹریلیا', 'uz': 'Avstraliya', 'uz-Cyrl': 'Австралия', 'uz-Latn': 'Avstraliya', 'vai': 'ꖺꖬꖤꔃꔷꕩ', 'vai-Latn': 'Ɔsituwéeliya', 'vai-Vaii': 'ꖺꖬꖤꔃꔷꕩ', 'vec': 'Australia', 'vi': 'Australia', 'vun': 'Australia', 'wae': 'Australie', 'wo': 'Ostarali', 'xh': 'E-Australia', 'xnr': 'ऑस्ट्रेलिया', 'xog': 'Awusitureliya', 'yav': 'Otalalí', 'yi': 'אויסטראַליע', 'yo': 'Austrálíà', 'yrl': 'Ausitaraliya', 'yue': '澳洲', 'yue-Hans': '澳洲', 'yue-Hant': '澳洲', 'zgh': 'ⵓⵙⵜⵔⴰⵍⵢⴰ', 'zh': '澳大利亚', 'zh-Hans': '澳大利亚', 'zh-Hant': '澳洲', 'zu': 'i-Australia'},
    'AW': {'af': 'Aruba', 'agq': 'Àlubà', 'ak': 'Aruba', 'am': 'አሩባ', 'ar': 'أروبا', 'as': 'আৰুবা', 'asa': 'Aruba', 'ast': 'Aruba', 'az': 'Aruba', 'az-Cyrl': 'Аруба', 'az-Latn': 'Aruba', 'bal-Latn': 'Aruba', 'bas': 'Àrubà', 'be': 'Аруба', 'be-tarask': 'Аруба', 'bez': 'Hualuba', 'bg': 'Аруба', 'blo': 'Arubaa', 'bm': 'Aruba', 'bn': 'আরুবা', 'br': 'Aruba', 'brx': 'आरूबा', 'bs': 'Aruba', 'bs-Cyrl': 'Аруба', 'bs-Latn': 'Aruba', 'ca': 'Aruba', 'ca-ES-valencia': 'Aruba', 'ccp': '𑄃𑄢𑄪𑄝', 'ce': 'Аруба', 'ceb': 'Aruba', 'cgg': 'Aruba', 'chr': 'ᎠᎷᏆ', 'ckb': 'ئارووبا', 'cs': 'Aruba', 'cv': 'Аруба', 'cy': 'Aruba', 'da': 'Aruba', 'dav': 'Aruba', 'de': 'Aruba', 'dje': 'Aruuba', 'dsb': 'Aruba', 'dyo': 'Aruba', 'dz': 'ཨ་རུ་བཱ', 'ebu': 'Aruba', 'ee': 'Aruba nutome', 'el': 'Αρούμπα', 'el-polyton': 'Ἀρούμπα', 'en': 'Aruba', 'en-Dsrt': '𐐊𐑉𐐭𐐺𐐲', 'eo': 'Arubo', 'es': 'Aruba', 'et': 'Aruba', 'eu': 'Aruba', 'ewo': 'Arúba', 'fa': 'آروبا', 'ff': 'Aruuba', 'ff-Adlm': '𞤀𞤪𞤵𞤦𞤢𞥄', 'ff-Latn': 'Aruuba', 'fi': 'Aruba', 'fil': 'Aruba', 'fo': 'Aruba', 'fr': 'Aruba', 'fur': 'Aruba', 'fy': 'Aruba', 'ga': 'Arúba', 'gd': 'Arùba', 'gl': 'Aruba', 'gsw': 'Aruba', 'gu': 'અરુબા', 'guz': 'Aruba', 'ha': 'Aruba', 'he': 'ארובה', 'hi': 'अरूबा', 'hi-Latn': 'Aruba', 'hr': 'Aruba', 'hsb': 'Aruba', 'ht': 'Aruba', 'hu': 'Aruba', 'hy': 'Արուբա', 'ia': 'Aruba', 'id': 'Aruba', 'ig': 'Aruba', 'is': 'Arúba', 'it': 'Aruba', 'ja': 'アルバ', 'jmc': 'Aruba', 'jv': 'Aruba', 'ka': 'არუბა', 'kaa': 'Аруба', 'kaa-Cyrl': 'Аруба', 'kab': 'Aruba', 'kam': 'Aruba', 'kde': 'Aluba', 'kea': 'Aruba', 'kgp': 'Aruma', 'khq': 'Aruuba', 'ki': 'Aruba', 'kk': 'Аруба', 'kk-Cyrl': 'Аруба', 'kln': 'Emetab Aruba', 'km': 'អារូបា', 'kn': 'ಅರುಬಾ', 'ko': '아루바', 'kok': 'अरुबा', 'kok-Deva': 'अरुबा', 'ks': 'اَروٗبا', 'ks-Arab': 'اَروٗبا', 'ksb': 'Aluba', 'ksf': 'aruba', 'ksh': 'Aruba', 'ku': 'Arûba', 'kxv': 'aruba', 'kxv-Deva': 'अरूबा', 'kxv-Latn': 'aruba', 'kxv-Orya': 'ଅରୁବା', 'kxv-Telu': 'ఆరుబా', 'ky': 'Аруба', 'lag': 'Arúuba', 'lb': 'Aruba', 'lg': 'Aruba', 'lld': 'Aruba', 'ln': 'Aruba', 'lo': 'ອາຣູບາ', 'lt': 'Aruba', 'lu': 'Aruba', 'luo': 'Aruba', 'luy': 'Aruba', 'lv': 'Aruba', 'mai': 'अरूबा', 'mas': 'Aruba', 'mer': 'Aruba', 'mfe': 'Aruba', 'mg': 'Arobà', 'mi': 'Arūpa', 'mk': 'Аруба', 'ml': 'അറൂബ', 'mn': 'Аруба', 'mr': 'अरुबा', 'ms': 'Aruba', 'mt': 'Aruba', 'mua': 'aruba', 'my': 'အာရူးဗား', 'mzn': 'آروبا', 'naq': 'Arubab', 'nb': 'Aruba', 'nd': 'Arubha', 'ne': 'अरुबा', 'nl': 'Aruba', 'nmg': 'Árúba', 'nn': 'Aruba', 'no': 'Aruba', 'nqo': 'ߊߙߎߓߊ߫', 'nus': 'Aruba', 'nyn': 'Aruba', 'om': 'Arubaa', 'or': 'ଆରୁବା', 'pa': 'ਅਰੂਬਾ', 'pa-Guru': 'ਅਰੂਬਾ', 'pcm': 'Arúba', 'pl': 'Aruba', 'ps': 'آروبا', 'pt': 'Aruba', 'qu': 'Aruba', 'rm': 'Aruba', 'rn': 'Aruba', 'ro': 'Aruba', 'rof': 'Aruba', 'ru': 'Аруба', 'rwk': 'Aruba', 'saq': 'Aruba', 'sbp': 'Aluba', 'sc': 'Aruba', 'sd': 'عروبا', 'sd-Arab': 'عروبا', 'se': 'Aruba', 'seh': 'Aruba', 'ses': 'Aruuba', 'sg': 'Arûba', 'shi': 'ⴰⵔⵓⴱⴰ', 'shi-Latn': 'aruba', 'shi-Tfng': 'ⴰⵔⵓⴱⴰ', 'si': 'අරූබා', 'sk': 'Aruba', 'sl': 'Aruba', 'smn': 'Aruba', 'sn': 'Arubha', 'so': 'Aruba', 'sq': 'Arubë', 'sr': 'Аруба', 'sr-Cyrl': 'Аруба', 'sr-Latn': 'Aruba', 'sv': 'Aruba', 'sw': 'Aruba', 'syr': 'ܐܪܘܒܐ', 'ta': 'அரூபா', 'te': 'అరుబా', 'teo': 'Aruba', 'tg': 'Аруба', 'th': 'อารูบา', 'ti': 'ኣሩባ', 'tk': 'Aruba', 'to': 'ʻAlupa', 'tr': 'Aruba', 'tt': 'Аруба', 'twq': 'Aruuba', 'tzm': 'Aruba', 'ug': 'ئارۇبا', 'uk': 'Аруба', 'ur': 'اروبا', 'uz': 'Aruba', 'uz-Cyrl': 'Аруба', 'uz-Latn': 'Aruba', 'vai': 'ꕉꖩꕑ', 'vai-Latn': 'Arubha', 'vai-Vaii': 'ꕉꖩꕑ', 'vec': 'Aruba', 'vi': 'Aruba', 'vun': 'Aruba', 'wae': 'Aruba', 'wo': 'Aruba', 'xh': 'E-Aruba', 'xnr': 'अरूबा', 'xog': 'Aruba', 'yav': 'Alúpa', 'yi': 'אַרובאַ', 'yo': 'Árúbà', 'yrl': 'Aruba', 'yue': '荷屬阿魯巴', 'yue-Hans': '荷属阿鲁巴', 'yue-Hant': '荷屬阿魯巴', 'zgh': 'ⴰⵔⵓⴱⴰ', 'zh': '阿鲁巴', 'zh-Hans': '阿鲁巴', 'zh-Hant': '荷屬阿魯巴', 'zu': 'i-Aruba'},
    'AX': {'af': 'Ålandeilande', 'ak': 'Aland Aeland', 'am': 'የአላንድ ደሴቶች', 'ar': 'جزر آلاند', 'as': 'আলণ্ড দ্বীপপুঞ্জ', 'ast': 'Islles Aland', 'az': 'Aland adaları', 'az-Cyrl': 'Аланд адалары', 'az-Latn': 'Aland adaları', 'bal-Latn': 'Áwlánd Islánd', 'be': 'Аландскія астравы', 'be-tarask': 'Аландскія астравы', 'bg': 'Оландски острови', 'blo': 'Ɔɔlanɖ kaBʊtǝlǝltǝna', 'bn': 'অলান্ড দ্বীপপুঞ্জ', 'br': 'Inizi Åland', 'brx': 'आलाण्ड द्वीपफोर', 'bs': 'Olandska ostrva', 'bs-Cyrl': 'Оландска острва', 'bs-Latn': 'Olandska ostrva', 'ca': 'Illes Åland', 'ca-ES-valencia': 'Illes Åland', 'ccp': '𑄃𑄣𑄚𑄴𑄓𑄧 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Аландан гӀайренаш', 'ceb': 'Aland Islands', 'chr': 'ᎣᎴᏅᏓ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی ئالاند', 'cs': 'Ålandy', 'cv': 'Аланди утравӗсем', 'cy': 'Ynysoedd Åland', 'da': 'Åland', 'de': 'Ålandinseln', 'dsb': 'Åland', 'dz': 'ཨ་ལནཌ་གླིང་ཚོམ', 'ee': 'Åland ƒudomekpo nutome', 'el': 'Νήσοι Όλαντ', 'el-polyton': 'Νήσοι Όλαντ', 'en': 'Åland Islands', 'en-Dsrt': '𐐈𐑊𐐰𐑌𐐼 𐐌𐑊𐐲𐑌𐐼𐑆', 'es': 'Islas Aland', 'et': 'Ahvenamaa', 'eu': 'Åland', 'fa': 'جزایر آلاند', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤀𞤤𞤢𞤲𞤣', 'fi': 'Ahvenanmaa', 'fil': 'Åland Islands', 'fo': 'Áland', 'fr': 'Îles Åland', 'fur': 'Isulis Aland', 'fy': 'Ålân', 'ga': 'Oileáin Åland', 'gd': 'Na h-Eileanan Åland', 'gl': 'Illas Åland', 'gsw': 'Aaland-Insle', 'gu': 'ઑલેન્ડ આઇલેન્ડ્સ', 'ha': 'Tsibirai na Åland', 'he': 'איי אולנד', 'hi': 'एलैंड द्वीपसमूह', 'hi-Latn': 'Aland Islands', 'hr': 'Ålandski otoci', 'hsb': 'Åland', 'ht': 'Îles Åland', 'hu': 'Åland-szigetek', 'hy': 'Ալանդյան կղզիներ', 'ia': 'Insulas Åland', 'id': 'Kepulauan Aland', 'ig': 'Åland Islands', 'is': 'Álandseyjar', 'it': 'Isole Åland', 'ja': 'オーランド諸島', 'jv': 'Kapuloan Alan', 'ka': 'ალანდის კუნძულები', 'kaa': 'Аланд атаўлары', 'kaa-Cyrl': 'Аланд атаўлары', 'kea': 'Ilhas Åland', 'kgp': 'Gojga Goj-vẽso', 'kk': 'Аланд аралдары', 'kk-Cyrl': 'Аланд аралдары', 'km': 'កោះ\u200bអាឡង់', 'kn': 'ಆಲ್ಯಾಂಡ್ ದ್ವೀಪಗಳು', 'ko': '올란드 제도', 'kok': 'अलांड जुवे', 'kok-Deva': 'अलांड जुवे', 'ks': 'ایلینڑ جٔزیٖرٕ', 'ks-Arab': 'ایلینڑ جٔزیٖرٕ', 'ksh': 'de Ohland-Enselle', 'ku': 'Giravên Alandê', 'kxv': 'eleṇḍ dīp', 'kxv-Deva': 'एलेंड द्वीप', 'kxv-Latn': 'eleṇḍ dīp', 'kxv-Orya': 'ଏଲେଣ୍ଡ', 'kxv-Telu': 'ఆలాండ్ దీపులు', 'ky': 'Аланд аралдары', 'lb': 'Ålandinselen', 'lld': 'Isoles Åland', 'lo': 'ຫມູ່ເກາະໂອລັນ', 'lt': 'Alandų Salos', 'lv': 'Olandes salas', 'mai': 'अलैंड द्वीपसमूह', 'mi': 'Motu Ōrana', 'mk': 'Оландски Острови', 'ml': 'അലൻഡ് ദ്വീപുകൾ', 'mn': 'Аландын арлууд', 'mr': 'अ\u200dॅलँड बेटे', 'ms': 'Kepulauan Aland', 'mt': 'il-Gżejjer Aland', 'my': 'အာလန်ကျွန်း', 'mzn': 'آلند جزیره', 'nb': 'Åland', 'ne': 'अलान्ड टापुहरु', 'nl': 'Åland', 'nn': 'Åland', 'no': 'Åland', 'nqo': 'ߊߟߊ߲ߘ ߕߌ߲', 'om': 'Odoloota Alaand', 'or': 'ଅଲାଣ୍ଡ ଦ୍ଵୀପପୁଞ୍ଜ', 'pa': 'ਅਲੈਂਡ ਟਾਪੂ', 'pa-Guru': 'ਅਲੈਂਡ ਟਾਪੂ', 'pcm': 'Ọ́lánd Aílands', 'pl': 'Wyspy Alandzkie', 'ps': 'الاند ټاپوان', 'pt': 'Ilhas Aland', 'qu': 'Islas Åland', 'rm': 'Inslas Aland', 'ro': 'Insulele Åland', 'ru': 'Аландские о-ва', 'sc': 'Ìsulas Åland', 'sd': 'الند ٻيٽ', 'sd-Arab': 'الند ٻيٽ', 'se': 'Ålánda', 'si': 'ඕලන්ඩ් දූපත්', 'sk': 'Alandy', 'sl': 'Ålandski otoki', 'smn': 'Vuáskueennâm', 'so': 'Jasiiradda Aland', 'sq': 'Ishujt Alandë', 'sr': 'Оландска Острва', 'sr-Cyrl': 'Оландска Острва', 'sr-Latn': 'Olandska Ostrva', 'sv': 'Åland', 'sw': 'Visiwa vya Aland', 'syr': 'ܓܙܝܖ̈ܐ ܕܐܠܐܢܕ', 'ta': 'ஆலந்து தீவுகள்', 'te': 'ఆలాండ్ దీవులు', 'tg': 'Ҷазираҳои Аланд', 'th': 'หมู่เกาะโอลันด์', 'ti': 'ደሴታት ኣላንድ', 'tk': 'Aland adalary', 'to': 'ʻOtumotu ʻAlani', 'tr': 'Åland Adaları', 'tt': 'Аланд утраулары', 'ug': 'ئالاند ئاراللىرى', 'uk': 'Аландські Острови', 'ur': 'آلینڈ آئلینڈز', 'uz': 'Aland orollari', 'uz-Cyrl': 'Аланд ороллари', 'uz-Latn': 'Aland orollari', 'vai': 'ꕉꕞꔺ', 'vai-Vaii': 'ꕉꕞꔺ', 'vec': 'Ìzole Aland', 'vi': 'Quần đảo Åland', 'wae': 'Alandinslä', 'wo': 'Duni Aalànd', 'xh': 'E-Åland Islands', 'xnr': 'एलैंड द्वीपां दा झुंड', 'yo': 'Àwọn Erékùsù ti Aland', 'yrl': 'Kapuãma-ita Arãdi', 'yue': '奧蘭群島', 'yue-Hans': '奥兰群岛', 'yue-Hant': '奧蘭群島', 'zh': '奥兰群岛', 'zh-Hans': '奥兰群岛', 'zh-Hant': '奧蘭群島', 'zu': 'i-Åland Islands'},
    'AZ': {'af': 'Azerbeidjan', 'agq': 'Àzɨbɛ̀dzân', 'ak': 'Asabegyan', 'am': 'አዘርባጃን', 'ar': 'أذربيجان', 'as': 'আজাৰবেইজান', 'asa': 'Adhabajani', 'ast': 'Azerbaixán', 'az': 'Azərbaycan', 'az-Cyrl': 'Азәрбајҹан', 'az-Latn': 'Azərbaycan', 'bal-Latn': 'Ázerbáiján', 'bas': 'Àzɛ̀rbajàŋ', 'be': 'Азербайджан', 'be-tarask': 'Азербайджан', 'bez': 'Huazabajani', 'bg': 'Азербайджан', 'bgn': 'آزربایجان', 'blo': 'Asɛrbaɩjaŋ', 'bm': 'Azɛrbayjaŋ', 'bn': 'আজারবাইজান', 'br': 'Azerbaidjan', 'brx': 'आजेरबाईजान', 'bs': 'Azerbejdžan', 'bs-Cyrl': 'Азербејџан', 'bs-Latn': 'Azerbejdžan', 'ca': 'Azerbaidjan', 'ca-ES-valencia': 'Azerbaidjan', 'ccp': '𑄃𑄎𑄢𑄴𑄝𑄭𑄎𑄚𑄴', 'ce': 'Азербайджан', 'ceb': 'Azerbaijan', 'cgg': 'Azabagyani', 'chr': 'ᎠᏎᏆᏣᏂ', 'ckb': 'ئازەربایجان', 'cs': 'Ázerbájdžán', 'cv': 'Азербайджан', 'cy': 'Aserbaijan', 'da': 'Aserbajdsjan', 'dav': 'Azabajani', 'de': 'Aserbaidschan', 'dje': 'Azerbaayijaŋ', 'dsb': 'Azerbajdžan', 'dyo': 'Aserbaysan', 'dz': 'ཨ་ཛར་བྷའི་ཇཱན', 'ebu': 'Azabajani', 'ee': 'Azerbaijan nutome', 'el': 'Αζερμπαϊτζάν', 'el-polyton': 'Ἀζερμπαϊτζάν', 'en': 'Azerbaijan', 'en-Dsrt': '𐐈𐑆𐐲𐑉𐐺𐐴𐑈𐐪𐑌', 'eo': 'Azerbajĝano', 'es': 'Azerbaiyán', 'et': 'Aserbaidžaan', 'eu': 'Azerbaijan', 'ewo': 'Azɛrəbaidzáŋ', 'fa': 'جمهوری آذربایجان', 'ff': 'Ajerbayjaan', 'ff-Adlm': '𞤀𞥁𞤫𞤪𞤦𞤢𞤭𞤶𞤢𞥄𞤲', 'ff-Latn': 'Ajerbayjaan', 'fi': 'Azerbaidžan', 'fil': 'Azerbaijan', 'fo': 'Aserbadjan', 'fr': 'Azerbaïdjan', 'fur': 'Azerbaigian', 'fy': 'Azerbeidzjan', 'ga': 'an Asarbaiseáin', 'gd': 'Asarbaideàn', 'gl': 'Acerbaixán', 'gsw': 'Aserbäidschan', 'gu': 'અઝરબૈજાન', 'guz': 'Azabajani', 'ha': 'Azarbaijan', 'he': 'אזרבייג׳ן', 'hi': 'अज़रबैजान', 'hi-Latn': 'Azerbaijan', 'hr': 'Azerbajdžan', 'hsb': 'Azerbajdźan', 'ht': 'Azerbaïdjan', 'hu': 'Azerbajdzsán', 'hy': 'Ադրբեջան', 'ia': 'Azerbaidzhan', 'id': 'Azerbaijan', 'ig': 'Azerbaijan', 'is': 'Aserbaídsjan', 'it': 'Azerbaigian', 'ja': 'アゼルバイジャン', 'jmc': 'Azabajani', 'jv': 'Azerbaijan', 'ka': 'აზერბაიჯანი', 'kaa': 'Әзербайжан', 'kaa-Cyrl': 'Әзербайжан', 'kab': 'Azrabijan', 'kam': 'Azabajani', 'kde': 'Azabadyani', 'kea': 'Azerbaidjan', 'kgp': 'Ajermajjáv', 'khq': 'Azerbaayijaŋ', 'ki': 'Azabajani', 'kk': 'Әзірбайжан', 'kk-Cyrl': 'Әзірбайжан', 'kln': 'Emetab Azerbaijan', 'km': 'អាស៊ែបៃហ្សង់', 'kn': 'ಅಜರ್ಬೈಜಾನ್', 'ko': '아제르바이잔', 'kok': 'अजरबैजान', 'kok-Deva': 'अजरबैजान', 'ks': 'آذربائیجان', 'ks-Arab': 'آذربائیجان', 'ksb': 'Azabajani', 'ksf': 'azabecán', 'ksh': 'Asserbaidschahn', 'ku': 'Azerbeycan', 'kxv': 'ajerbaijan', 'kxv-Deva': 'अज़रबेजान', 'kxv-Latn': 'ajerbaijan', 'kxv-Orya': 'ଅଜେର୍\u200d\u200dବାଇଜାନ୍', 'kxv-Telu': 'ఆజర్బైజాన్', 'ky': 'Азербайжан', 'lag': 'Azabajáani', 'lb': 'Aserbaidschan', 'lg': 'Azerebayijaani', 'lld': 'Azerbaijan', 'ln': 'Azɛlɛbaizá', 'lo': 'ອາເຊີໄບຈານ', 'lt': 'Azerbaidžanas', 'lu': 'Ajelbayidja', 'luo': 'Azerbaijan', 'luy': 'Azabajani', 'lv': 'Azerbaidžāna', 'mai': 'अजरबैजान', 'mas': 'Azabajani', 'mer': 'Azebaijani', 'mfe': 'Azerbaïdjan', 'mg': 'Azerbaidjan', 'mgh': 'Uazabajani', 'mi': 'Atepaihānia', 'mk': 'Азербејџан', 'ml': 'അസർബൈജാൻ', 'mn': 'Азербайжан', 'mr': 'अझरबैजान', 'ms': 'Azerbaijan', 'mt': 'l-Ażerbajġan', 'mua': 'azerbaijaŋ', 'my': 'အဇာဘိုင်ဂျန်', 'mzn': 'آذربایجون', 'naq': 'Azerbaijanni', 'nb': 'Aserbajdsjan', 'nd': 'Azerbaijan', 'ne': 'अजरबैजान', 'nl': 'Azerbeidzjan', 'nmg': 'Azerbaïjaŋ', 'nn': 'Aserbajdsjan', 'no': 'Aserbajdsjan', 'nqo': 'ߊߖߊߙߑߓߊߦߌߖߊ߲', 'nus': 'Adhe̱rbe̱ja̱n', 'nyn': 'Azabagyani', 'om': 'Azerbaajiyaan', 'or': 'ଆଜେରବାଇଜାନ୍', 'pa': 'ਅਜ਼ਰਬਾਈਜਾਨ', 'pa-Guru': 'ਅਜ਼ਰਬਾਈਜਾਨ', 'pcm': 'Azẹrbaijan', 'pl': 'Azerbejdżan', 'ps': 'اذربايجان', 'pt': 'Azerbaijão', 'qu': 'Azerbaiyán', 'rm': 'Aserbaidschan', 'rn': 'Azerubayijani', 'ro': 'Azerbaidjan', 'rof': 'Azabajani', 'ru': 'Азербайджан', 'rwk': 'Azabajani', 'saq': 'Azabajani', 'sbp': 'Asabajani', 'sc': 'Azerbaigiàn', 'sd': 'آذربائيجان', 'sd-Arab': 'آذربائيجان', 'se': 'Aserbaižan', 'seh': 'Azerbaijão', 'ses': 'Azerbaayijaŋ', 'sg': 'Zerebaidyäan, Azerbaidyäan,', 'shi': 'ⴰⴷⵔⴰⴱⵉⵊⴰⵏ', 'shi-Latn': 'adrabijan', 'shi-Tfng': 'ⴰⴷⵔⴰⴱⵉⵊⴰⵏ', 'si': 'අසර්බයිජානය', 'sk': 'Azerbajdžan', 'sl': 'Azerbajdžan', 'smn': 'Azerbaidžan', 'sn': 'Azabajani', 'so': 'Asarbajan', 'sq': 'Azerbajxhan', 'sr': 'Азербејџан', 'sr-Cyrl': 'Азербејџан', 'sr-Latn': 'Azerbejdžan', 'sv': 'Azerbajdzjan', 'sw': 'Azerbaijani', 'syr': 'ܐܙܪܒܝܓܐܢ', 'ta': 'அசர்பைஜான்', 'te': 'అజర్బైజాన్', 'teo': 'Azabajani', 'tg': 'Озарбойҷон', 'th': 'อาเซอร์ไบจาน', 'ti': 'ኣዘርባጃን', 'tk': 'Azerbaýjan', 'to': 'ʻAsapaisani', 'tr': 'Azerbaycan', 'tt': 'Әзәрбайҗан', 'twq': 'Azerbaayijaŋ', 'tzm': 'Azerbiǧan', 'ug': 'ئەزەربەيجان', 'uk': 'Азербайджан', 'ur': 'آذربائیجان', 'uz': 'Ozarbayjon', 'uz-Cyrl': 'Озарбайжон', 'uz-Latn': 'Ozarbayjon', 'vai': 'ꕉꕤꕑꔤꕧꘋ', 'vai-Latn': 'Azabhaijaŋ', 'vai-Vaii': 'ꕉꕤꕑꔤꕧꘋ', 'vec': 'Azerbaijan', 'vi': 'Azerbaijan', 'vun': 'Azabajani', 'wae': 'Aserbaidšan', 'wo': 'Aserbayjaŋ', 'xh': 'E-Azerbaijan', 'xnr': 'अज़रबैजान', 'xog': 'Azerebayijaani', 'yav': 'Asɛlpaisáŋ', 'yo': 'Asẹ́bájánì', 'yrl': 'Aseriretãma', 'yue': '阿塞拜疆', 'yue-Hans': '亚塞拜然', 'yue-Hant': '阿塞拜疆', 'zgh': 'ⴰⴷⵔⴰⴱⵉⵊⴰⵏ', 'zh': '阿塞拜疆', 'zh-Hans': '阿塞拜疆', 'zh-Hant': '亞塞拜然', 'zu': 'i-Azerbaijan'},
    'Adlm': {'af': 'Adlam', 'ak': 'Adlam kasa', 'am': 'አድላም', 'ar': 'أدلم', 'as': 'এডলাম', 'az': 'adlam', 'az-Latn': 'adlam', 'bal-Latn': 'Adlam', 'be': 'адлам', 'be-tarask': 'адлам', 'bg': 'адлам', 'bn': 'আদলাম', 'br': 'adlam', 'brx': 'आदलाम', 'bs': 'adlam pismo', 'bs-Cyrl': 'адламанско писмо', 'bs-Latn': 'adlam pismo', 'ca': 'adlam', 'ca-ES-valencia': 'adlam', 'chr': 'ᎠᏓᎳᎻ', 'cs': 'adlam', 'cy': 'Adlam', 'da': 'adlam', 'de': 'Adlam', 'dsb': 'adlamske pismo', 'el': 'Άντλαμ', 'el-polyton': 'Άντλαμ', 'en': 'Adlam', 'es': 'ádlam', 'et': 'adlami', 'eu': 'adlama', 'fa': 'آدلام', 'ff-Adlm': '𞤀𞤁𞤂𞤢𞤃', 'fi': 'fulanin adlam-aakkosto', 'fil': 'Adlam', 'fo': 'adlam', 'fr': 'adlam', 'ga': 'Adlam', 'gd': 'Adlam', 'gl': 'adlam', 'gu': 'એડલમ', 'ha': 'Adlam', 'he': 'אדלם', 'hi': 'ऐडलम', 'hi-Latn': 'Adlam', 'hr': 'adlam pismo', 'hsb': 'adlamske pismo', 'ht': 'adlam', 'hu': 'Adlam', 'hy': 'ադլամ', 'id': 'Adlam', 'ig': 'Adlam', 'is': 'adlam', 'it': 'adlam', 'ja': 'アドラム文字', 'jv': 'Adlam', 'ka': 'ადლამი', 'kk': 'адлам жазуы', 'kk-Cyrl': 'адлам жазуы', 'km': 'អាតឡាម', 'kn': 'ಆ್ಯಡ್ಲಮ್', 'ko': '아들람 문자', 'kok': 'अडलाम', 'kok-Deva': 'अडलाम', 'ky': 'Адлам (жазуу)', 'lo': 'ແອດລາມ', 'lt': 'ADLAM', 'lv': 'adlama', 'mai': 'अदलाम', 'mi': 'Atarama', 'mk': 'адламско', 'ml': 'അദ്\u200cലാം', 'mn': 'Адлам бичиг', 'mr': 'अदलम', 'ms': 'Adlam', 'my': 'အက်ဒ်လမ်', 'nb': 'adlam', 'ne': 'एडलाम', 'nl': 'Adlam', 'nn': 'adlam', 'no': 'adlam', 'nqo': 'ߊߘߑߟߊߡ', 'or': 'ଆଡଲମ୍', 'pa': 'ਅਦਲਾਮ', 'pa-Guru': 'ਅਦਲਾਮ', 'pcm': 'Ádlam', 'pl': 'adlam', 'ps': 'اډلام', 'pt': 'adlam', 'qu': 'Adlam Simi', 'ro': 'adlam', 'ru': 'адлам', 'sc': 'adlam', 'sd': 'ايڊلام', 'sd-Arab': 'ايڊلام', 'si': 'ඇඩ්ලම්', 'sk': 'adlam', 'sl': 'adlamski', 'so': 'Adlam', 'sq': 'adlam', 'sr': 'адлам', 'sr-Cyrl': 'адлам', 'sr-Latn': 'adlam', 'sv': 'adlamiska', 'sw': 'Kiadlamu', 'ta': 'அட்லாம்', 'te': 'అద్లామ్', 'tg': 'Адламӣ', 'th': 'อัดลัม', 'ti': 'አድላም', 'tk': 'adlam', 'to': 'tohinima fakaʻatilami', 'tr': 'Adlam', 'uk': 'адлам', 'ur': 'ایڈلم', 'uz': 'adlam', 'uz-Latn': 'adlam', 'vec': 'adlam', 'vi': 'Chữ Adlam', 'yo': 'Èdè Adam', 'yue': '阿德拉姆文', 'yue-Hans': '阿德拉姆文', 'yue-Hant': '阿德拉姆文', 'zh': '阿德拉姆文', 'zh-Hans': '阿德拉姆文', 'zh-Hant': '富拉文', 'zu': 'isi-Adlam'},
    'Afak': {'ast': 'afaka', 'bs': 'afaka pismo', 'bs-Latn': 'afaka pismo', 'ca': 'afaka', 'ca-ES-valencia': 'afaka', 'cs': 'afaka', 'da': 'afaka', 'de': 'Afaka', 'en': 'Afaka', 'et': 'afaka', 'fi': 'afaka', 'fy': 'Defaka', 'gd': 'Afaka', 'hi-Latn': 'Afaka', 'hr': 'afaka pismo', 'id': 'Afaka', 'it': 'afaka', 'ja': 'アファカ文字', 'ka': 'აფაკა', 'ko': '아파카 문자', 'lo': 'ອັບຟາກາ', 'lt': 'Afaka', 'mk': 'афака', 'nb': 'afaka', 'nl': 'Defaka', 'nn': 'afaka', 'no': 'afaka', 'ru': 'афака', 'sv': 'afakiska', 'th': 'อะฟาคา', 'to': 'tohinima fakaʻafaka', 'tr': 'Afaka', 'ug': 'ئافاكا', 'uk': 'афака', 'vi': 'Chữ Afaka', 'yue': '阿法卡文字', 'yue-Hans': '阿法卡文字', 'yue-Hant': '阿法卡文字', 'zh': '阿法卡文', 'zh-Hans': '阿法卡文', 'zh-Hant': '阿法卡文字'},
    'Aghb': {'ast': 'cáucaso-albanés', 'bal-Latn': 'Kákáshi Albáni', 'bs': 'kavkazijsko albansko pismo', 'bs-Cyrl': 'бијело албанско писмо', 'bs-Latn': 'kavkazijsko albansko pismo', 'ca': 'albanès caucàsic', 'ca-ES-valencia': 'albanès caucàsic', 'cs': 'kavkazskoalbánské', 'de': 'Kaukasisch-Albanisch', 'en': 'Caucasian Albanian', 'et': 'albaani', 'eu': 'Kaukasoko albaniera', 'fa': 'آلبانیایی قفقازی', 'ff-Adlm': '𞤀𞤹𞤦𞤢𞤲𞤭𞤴𞤢', 'fi': 'kaukasianalbanialainen', 'ga': 'Albánach Cugasach', 'gd': 'Albàinis Chabhcasach', 'he': 'אלבני קווקזי', 'hi-Latn': 'Caucasian Albanian', 'hu': 'Kaukázusi albaniai', 'id': 'Albania Kaukasia', 'ig': 'Caucasian Albanian', 'it': 'albanese caucasico', 'ja': 'カフカス・アルバニア文字', 'kk': 'агван жазуы', 'kk-Cyrl': 'агван жазуы', 'km': 'អាល់បានីកៅកាស៊ី', 'ko': '코카시안 알바니아 문자', 'lt': 'Kaukazo Albanijos', 'mk': 'кавкаскоалбански', 'mn': 'Кавказийн албани', 'ms': 'Kaukasia Albania', 'nb': 'kaukasus-albansk', 'nl': 'Kaukasisch Albanees', 'nn': 'kaukasus-albansk', 'no': 'kaukasus-albansk', 'nqo': 'ߊߜ߭ߑߓߊߣߌ', 'rm': 'albanais dal Caucasus', 'ro': 'albaneză caucaziană', 'sc': 'albanesu caucàsicu', 'so': 'Qoraalka Luuqada Caucasian Albanian', 'sq': 'albanishte e Kaukazit', 'sv': 'kaukasiska albanska', 'tg': 'Албани Қафқозӣ', 'th': 'แอลเบเนีย คอเคเซีย', 'to': 'tohinima fakaʻalapēnia-kaukasia', 'tr': 'Kafkas Albanyası', 'uk': 'кавказька албанська', 'vec': 'albaneze caucàzego', 'yue': '高加索阿爾巴尼亞文', 'yue-Hans': '高加索阿尔巴尼亚文', 'yue-Hant': '高加索阿爾巴尼亞文', 'zh': '高加索阿尔巴尼亚文', 'zh-Hans': '高加索阿尔巴尼亚文', 'zh-Hant': '高加索阿爾巴尼亞文', 'zu': 'isi-Caucasian Albanian'},
    'Ahom': {'bs': 'ahom pismo', 'bs-Cyrl': 'ахом писмо', 'bs-Latn': 'ahom pismo', 'et': 'ahomi', 'eu': 'ahomera', 'ff-Adlm': '𞤀𞤸𞤮𞤥', 'he': 'אהום', 'kk': 'ахом жазуы', 'kk-Cyrl': 'ахом жазуы', 'km': 'អាហូម', 'mn': 'Ахом', 'nqo': 'ߊߤߏ߲', 'so': 'Dadka Ahom', 'sq': 'ahomisht', 'tg': 'Ахомӣ', 'uk': 'ахом', 'zh': '阿豪姆文', 'zh-Hans': '阿豪姆文', 'zh-Hant': '阿洪姆文', 'zu': 'isi-Ahom', 'en': 'Ahom'},
    'Arab': {'af': 'Arabies', 'ak': 'Arabeke', 'am': 'ዓረብኛ', 'ar': 'العربية', 'as': 'আৰবী', 'ast': 'árabe', 'az': 'ərəb', 'az-Arab': 'عربجه', 'az-Latn': 'ərəb', 'bal-Latn': 'Arabi', 'be': 'арабскае', 'be-tarask': 'арабскае', 'bg': 'арабска', 'bgn': 'عربی', 'blo': 'laaribu kʊja', 'bn': 'আরবি', 'br': 'arabek', 'brx': 'आराबिक', 'bs': 'arapsko pismo', 'bs-Cyrl': 'арапско писмо', 'bs-Latn': 'arapsko pismo', 'ca': 'àrab', 'ca-ES-valencia': 'àrab', 'ccp': '𑄃𑄢𑄧𑄝𑄨', 'ce': 'Ӏаьрбийн', 'ceb': 'Arabiko', 'chr': 'ᎡᎳᏈᎩ', 'ckb': 'عەرەبی', 'cs': 'arabské', 'cv': 'арап', 'cy': 'Arabaidd', 'da': 'arabisk', 'de': 'Arabisch', 'doi': 'अरबी', 'dsb': 'arabski', 'dz': 'ཨེ་ར་བིཀ་ཡིག་གུ', 'ee': 'Arabiagbeŋɔŋlɔ', 'el': 'Αραβικό', 'el-polyton': 'Ἀραβικό', 'en': 'Arabic', 'en-Dsrt': '𐐇𐑉𐐲𐐺𐐮𐐿', 'es': 'árabe', 'et': 'araabia', 'eu': 'arabiarra', 'fa': 'عربی', 'ff-Adlm': '𞤀𞥄𞤪𞤢𞤦𞤵', 'fi': 'arabialainen', 'fil': 'Arabic', 'fo': 'arabisk', 'fr': 'arabe', 'fur': 'arap', 'fy': 'Arabysk', 'ga': 'Arabach', 'gd': 'Arabais', 'gl': 'árabe', 'gsw': 'Arabisch', 'gu': 'અરબી', 'ha': 'Larabci', 'he': 'ערבי', 'hi': 'अरबी', 'hi-Latn': 'Arabic', 'hr': 'arapsko pismo', 'hsb': 'arabsce', 'ht': 'arabe', 'hu': 'Perzsa-arab', 'hy': 'արաբական', 'ia': 'arabe', 'id': 'Arab Persia', 'ie': 'arabic', 'ig': 'Mkpụrụ Okwu Arabic', 'ii': 'ꀊꇁꀨꁱꂷ', 'is': 'arabískt', 'it': 'arabo', 'ja': 'アラビア文字', 'jv': 'hija’iyah', 'ka': 'არაბული', 'kaa': 'араб жазыўы', 'kaa-Cyrl': 'араб жазыўы', 'kea': 'arábiku', 'kgp': 'arame', 'kk': 'араб жазуы', 'kk-Cyrl': 'араб жазуы', 'km': 'អារ៉ាប់', 'kn': 'ಅರೇಬಿಕ್', 'ko': '아랍 문자', 'kok': 'आरबी', 'kok-Deva': 'आरबी', 'ks': 'عربی', 'ks-Arab': 'عربی', 'ks-Deva': 'अरबी', 'ksh': 'arraabesche Schreff', 'ku': 'erebî', 'kxv': 'aarabic', 'kxv-Deva': 'अरबी', 'kxv-Latn': 'aarabic', 'kxv-Orya': 'ଆରବିକ୍', 'kxv-Telu': 'ఆరబిక్', 'ky': 'Араб', 'lb': 'Arabesch', 'lld': 'arabich', 'lo': 'ອາຣາບິກ', 'lrc': 'عأرأڤی', 'lt': 'arabų', 'lv': 'arābu', 'mai': 'अरबी', 'mi': 'Arapika', 'mk': 'арапско писмо', 'ml': 'അറബിക്', 'mn': 'араб', 'mni': 'আরবিক', 'mni-Beng': 'আরবিক', 'mr': 'अरबी', 'ms': 'Perso-Arab', 'mt': 'Għarbi', 'my': 'အာရေဗျ', 'mzn': 'عربی', 'nb': 'arabisk', 'ne': 'अरबी', 'nl': 'Arabisch', 'nn': 'arabisk', 'no': 'arabisk', 'nqo': 'ߊߙߊߓߎ', 'om': 'Arabiffa', 'or': 'ଆରବିକ୍', 'os': 'Араббаг', 'pa': 'ਅਰਬੀ', 'pa-Arab': 'عربی', 'pa-Guru': 'ਅਰਬੀ', 'pcm': 'Arábík', 'pl': 'arabskie', 'ps': 'عربي', 'pt': 'árabe', 'qu': 'Arabe Simi', 'ro': 'arabă', 'ru': 'арабица', 'sa': 'अरबी', 'sah': 'Арааптыы', 'sat': 'ᱟᱨᱵᱤᱠ', 'sat-Olck': 'ᱟᱨᱵᱤᱠ', 'sc': 'àrabu', 'sd': 'عربي', 'sd-Arab': 'عربي', 'sd-Deva': 'अरबी', 'se': 'arába', 'si': 'අරාබි', 'sk': 'arabské', 'sl': 'arabski', 'so': 'Carabi', 'sq': 'arabik', 'sr': 'арапско писмо', 'sr-Cyrl': 'арапско писмо', 'sr-Latn': 'arapsko pismo', 'sv': 'arabiska', 'sw': 'Kiarabu', 'syr': 'ܥܪܒܝܬ', 'ta': 'அரபிக்', 'te': 'అరబిక్', 'tg': 'Арабӣ', 'th': 'อาหรับ', 'ti': 'ዓረብኛ', 'tk': 'Arap elipbiýi', 'to': 'tohinima fakaʻalepea', 'tok': 'sitelen Alapi', 'tpi': 'Arabik', 'tr': 'Arap', 'tt': 'гарәп', 'ug': 'ئەرەب', 'uk': 'арабиця', 'ur': 'عربی', 'uz-Arab': 'عربی', 'uz-Cyrl': 'Араб', 'vec': 'àrabo', 'vi': 'Chữ Ả Rập', 'wae': 'Arabiš', 'wo': 'Araab', 'xh': 'Isi-Arabhu', 'xnr': 'अरबी', 'yi': 'אַראַביש', 'yo': 'èdè Lárúbáwá', 'yrl': 'arabika', 'yue': '阿拉伯文', 'yue-Hans': '阿拉伯文', 'yue-Hant': '阿拉伯文', 'zh': '阿拉伯文', 'zh-Hans': '阿拉伯文', 'zh-Hant': '阿拉伯字母', 'zu': 'isi-Arabic'},
    'Aran': {'af': 'Nastaliq', 'ak': 'Nastaliki kasa', 'am': 'ናስታሊክ', 'ar': 'نستعليق', 'as': 'নাষ্টালিক', 'bal-Latn': 'Nastalik', 'be': 'насталік', 'be-tarask': 'насталік', 'bg': 'aранска', 'bn': 'নাস্তালিক', 'brx': 'नास्तालिक', 'bs': 'nastalik pismo', 'bs-Cyrl': 'насталик писмо', 'bs-Latn': 'nastalik pismo', 'ca': 'nasta’liq', 'ca-ES-valencia': 'nasta’liq', 'chr': 'ᎾᏍᏔᎵᏆ', 'cs': 'nastalik', 'cy': 'Nastaliq', 'da': 'nastaliq', 'de': 'Nastaliq', 'dsb': 'nastaliqske pismo', 'el': 'Νασταλίκ', 'el-polyton': 'Νασταλίκ', 'en': 'Nastaliq', 'es': 'nastaliq', 'et': 'nastaliik', 'eu': 'nastaliqa', 'fa': 'آران', 'ff-Adlm': '𞤐𞤢𞤧𞤼𞤢𞤤𞤭𞤹', 'fi': 'nastaliq', 'fil': 'Nastaliq', 'fo': 'nastaliq', 'fr': 'nastaliq', 'ga': 'Nastaliq', 'gd': 'Nastaliq', 'gl': 'nastaliq', 'gu': 'નસ્તાલીક', 'ha': 'Rubutun Nastaliq', 'he': 'נסתעליק', 'hi': 'नस्तालीक़', 'hi-Latn': 'Nastaliq', 'hr': 'nastaliq pismo', 'hsb': 'nastaliqske pismo', 'ht': 'nastaliq', 'hu': 'Nasztalik', 'hy': 'նաստալիք', 'id': 'Nastaliq', 'ig': 'Nastaliq', 'is': 'nastaliq', 'it': 'nastaliq', 'ja': 'ナスタアリーク体', 'jv': 'Nastalik', 'ka': 'ნასტალიქი', 'kk': 'насталик жазуы', 'kk-Cyrl': 'насталик жазуы', 'km': 'អារ៉ាន', 'kn': 'ನಸ್ಟಾಲಿಕ್', 'ko': '나스탈리크체', 'kok': 'नास्तालिक', 'kok-Deva': 'नास्तालिक', 'ks': 'نستعلیق', 'ks-Arab': 'نستعلیق', 'ku': 'nestalîq', 'ky': 'Насталик (Араб жазуусу)', 'lo': 'ນາສຕໍລິກ (ຄຳໃບ້ການແປ: ລະຫັດພິເສດກຳນົດຮູບແບບຕົວໜັງສືອາຣັບ.)', 'lt': 'Nastalik', 'lv': 'nastaliku', 'mai': 'नास्तलिक', 'mi': 'Nātarika', 'mk': 'насталик', 'ml': 'നസ്\u200cറ്റാലിക്', 'mn': 'Насталик бичиг', 'mr': 'नस्तालीक', 'ms': 'Nastaliq', 'my': 'နက်စ်တာလိခ်', 'nb': 'nastaliq', 'ne': 'नास्टालिक', 'nl': 'Nastaliq', 'nn': 'nastaliq', 'no': 'nastaliq', 'nqo': 'ߣߊߛߑߕߊߟߌߞ', 'or': 'ନାଷ୍ଟାଲିକ୍\u200c', 'pa': 'ਨਸਤਾਲੀਕ', 'pa-Arab': 'نستعلیق', 'pa-Guru': 'ਨਸਤਾਲੀਕ', 'pcm': 'Aran Lángwej', 'pl': 'nastaliq', 'ps': 'نستعلیق', 'pt': 'nastaliq', 'qu': 'Nastaliq qillqa', 'ro': 'nastaaliq', 'ru': 'насталик', 'sc': 'nastaʿlīq', 'sd': 'نستعلیق', 'sd-Arab': 'نستعلیق', 'si': 'නස්ටාලික්', 'sk': 'nastaliq', 'sl': 'nastalik', 'so': 'Farta Luuqada Faarsiga', 'sq': 'nastalik', 'sr': 'насталик', 'sr-Cyrl': 'насталик', 'sr-Latn': 'nastalik', 'sv': 'nastaliq', 'sw': 'Kinastaliki', 'ta': 'நஸ்டாலிக்', 'te': 'నాస్టాలిక్', 'tg': 'Насталиқӣ', 'th': 'นาสตาลิก', 'ti': 'ናስታሊ', 'tk': 'Nastalik ýazuwy', 'to': 'tohinima fakanasatalīki', 'tr': 'Nestâlik', 'uk': 'насталік', 'ur': 'نستعلیق', 'uz': 'nastaʼliq', 'uz-Latn': 'nastaʼliq', 'vec': 'nastaliq', 'vi': 'Chữ Nastaliq', 'yo': 'Èdè Aran', 'yue': '納斯塔利克文', 'yue-Hans': '纳斯塔利克文', 'yue-Hant': '納斯塔利克文', 'zh': '波斯体', 'zh-Hans': '波斯体', 'zh-Hant': '波斯體', 'zu': 'i-Nastaliq'},
    'Armi': {'ast': 'aramaicu imperial', 'bal-Latn': 'Shahensháhi Aramái', 'bg': 'Арамейска', 'bn': 'আরমি', 'br': 'arameek impalaerel', 'brx': 'शहनशाही आरामाईक', 'bs': 'imperijsko aramejsko pismo', 'bs-Cyrl': 'империјско арамејско писмо', 'bs-Latn': 'imperijsko aramejsko pismo', 'ca': 'arameu imperial', 'ca-ES-valencia': 'arameu imperial', 'ccp': '𑄃𑄢𑄧𑄟𑄨', 'cs': 'aramejské (imperiální)', 'el': 'Αυτοκρατορικό Αραμαϊκό', 'el-polyton': 'Αυτοκρατορικό Αραμαϊκό', 'en': 'Imperial Aramaic', 'en-Dsrt': '𐐆𐑋𐐹𐐮𐑉𐐨𐐲𐑊 𐐁𐑉𐐲𐑋𐐩𐐮𐐿', 'et': 'vanaaramea', 'eu': 'aramiera inperiarra', 'fa': 'آرامی هخامنشی', 'ff-Adlm': '𞤀𞤪𞤢𞤥𞤭𞤴𞤢 𞤉𞤥𞤨𞤫𞤪𞤭𞤴𞤢𞤤', 'fi': 'valtakunnanaramealainen', 'fr': 'araméen impérial', 'fy': 'Keizerlijk Aramees', 'ga': 'Aramach Impiriúil', 'gd': 'Aramais impireil', 'gu': 'ઇમ્પિરિયલ આર્મનિક', 'he': 'ארמית רשמית', 'hi': 'इम्पिरियल आर्मेनिक', 'hi-Latn': 'Imperial Aramaic', 'hr': 'aramejsko pismo', 'ht': 'araméen impérial', 'hu': 'Birodalmi arámi', 'id': 'Aram Imperial', 'ig': 'Imperial Aramaic', 'it': 'aramaico imperiale', 'ja': '帝国アラム文字', 'ka': 'იმპერიული არამეული', 'kgp': 'armĩ', 'kk': 'арамей жазуы', 'kk-Cyrl': 'арамей жазуы', 'km': 'អារ៉ាម៉ាអ៊ីមភើរៀល', 'kn': 'ಇಂಪೀರಿಯಲ್ ಅರೆಮಾಯಿಕ್', 'ko': '아랍제국 문자', 'lo': 'ອິມພີຮຽນ ອາເມອິກ', 'lt': 'imperinė aramaikų', 'lv': 'aramiešu', 'mk': 'царскоарамејски', 'ml': 'അർമി', 'mn': 'Арамайк', 'mr': 'इम्पिरियल आर्मेनिक', 'ms': 'Aramia Imperial', 'nb': 'arameisk', 'ne': 'आर्मी', 'nl': 'Keizerlijk Aramees', 'nn': 'armisk', 'no': 'arameisk', 'nqo': 'ߝߊ߰ߡߊ߬ߟߊ-ߊߙߊߡߋ', 'or': 'ଇମ୍ପେରିଆଲ୍ ଆରମିକ୍', 'rm': 'arameic imperial', 'ro': 'aramaică imperială', 'ru': 'арамейская', 'sc': 'aramàicu imperiale', 'sl': 'imperialno-aramejski', 'so': 'Luuqada Imperial Aramaic', 'sq': 'aramaishte perandorake', 'sr': 'империјско арамејско писмо', 'sr-Cyrl': 'империјско арамејско писмо', 'sr-Latn': 'imperijsko aramejsko pismo', 'sv': 'imperisk arameiska', 'syr': 'ܐܪܡܝܐ', 'ta': 'இம்பேரியல் அரமெய்க்', 'te': 'ఇంపీరియల్ అరామాక్', 'tg': 'Арамейкии Империалӣ', 'th': 'อิมพีเรียล อราเมอิก', 'to': 'tohinima fakaʻalāmiti-ʻemipaea', 'tr': 'İmparatorluk Aramicesi', 'ug': 'خان جەمەتى ئارامۇ', 'uk': 'армі', 'vec': 'amàrego inperiale', 'vi': 'Chữ Imperial Aramaic', 'xnr': 'इम्पिरियल आर्मेनिक', 'yrl': 'arimi', 'yue': '皇室亞美尼亞文', 'yue-Hans': '皇室亚美尼亚文', 'yue-Hant': '皇室亞美尼亞文', 'zh': '皇室亚拉姆文', 'zh-Hans': '皇室亚拉姆文', 'zh-Hant': '皇室亞美尼亞文', 'zu': 'isi-Imperial Aramaic'},
    'Armn': {'af': 'Armeens', 'ak': 'Amenia kasa', 'am': 'አርሜንያዊ', 'ar': 'الأرمينية', 'as': 'আৰ্মেনীয়', 'ast': 'armeniu', 'az': 'erməni', 'az-Arab': 'ائرمنیجه', 'az-Latn': 'erməni', 'bal-Latn': 'Arman', 'be': 'армянскае', 'be-tarask': 'армянскае', 'bg': 'арменска', 'bgn': 'ارمنی', 'bn': 'আর্মেনীয়', 'br': 'armenianek', 'brx': 'आर्मेनियान', 'bs': 'armensko pismo', 'bs-Cyrl': 'јерменско писмо', 'bs-Latn': 'armensko pismo', 'ca': 'armeni', 'ca-ES-valencia': 'armeni', 'ccp': '𑄃𑄢𑄴𑄟𑄬𑄚𑄩𑄠𑄧', 'ce': 'эрмалойн', 'chr': 'ᎠᎳᎻᎠᏂ', 'ckb': 'ئەرمەنی', 'cs': 'arménské', 'cy': 'Armenaidd', 'da': 'armensk', 'de': 'Armenisch', 'dsb': 'armeński', 'dz': 'ཨར་མི་ནི་ཡཱན་ཡིག་གུ', 'ee': 'armeniagbeŋɔŋlɔ', 'el': 'Αρμενικό', 'el-polyton': 'Ἀρμενικό', 'en': 'Armenian', 'en-Dsrt': '𐐂𐑉𐑋𐐨𐑌𐐨𐐲𐑌', 'es': 'armenio', 'et': 'armeenia', 'eu': 'armeniarra', 'fa': 'ارمنی', 'ff-Adlm': '𞤀𞤪𞤥𞤫𞤲𞤭𞤴𞤢𞤲', 'fi': 'armenialainen', 'fil': 'Armenian', 'fo': 'armenskt', 'fr': 'arménien', 'fur': 'armen', 'fy': 'Armeens', 'ga': 'Airméanach', 'gd': 'Airmeinis', 'gl': 'armenio', 'gsw': 'Armenisch', 'gu': 'અર્મેનિયન', 'ha': 'Armeniyanci', 'he': 'ארמני', 'hi': 'आर्मेनियाई', 'hi-Latn': 'Armenian', 'hr': 'armensko pismo', 'hsb': 'armensce', 'ht': 'arménien', 'hu': 'Örmény', 'hy': 'հայկական', 'ia': 'armenie', 'id': 'Armenia', 'ig': 'Mkpụrụ ọkwụ Armenịan', 'is': 'armenskt', 'it': 'armeno', 'ja': 'アルメニア文字', 'jv': 'Armenia', 'ka': 'სომხური', 'kea': 'arméniu', 'kgp': 'armẽnjo', 'kk': 'армян жазуы', 'kk-Cyrl': 'армян жазуы', 'km': 'អាមេនី', 'kn': 'ಅರ್ಮೇನಿಯನ್', 'ko': '아르메니아 문자', 'kok': 'आर्मेनियन', 'kok-Deva': 'आर्मेनियन', 'ks': 'اَرمانیَن', 'ks-Arab': 'اَرمانیَن', 'ksh': 'armeenesche Schreff', 'ku': 'ermenî', 'ky': 'Армян', 'lb': 'Armenesch', 'lo': 'ອາເມນຽນ', 'lrc': 'أرمأنی', 'lt': 'armėnų', 'lv': 'armēņu', 'mai': 'आर्मेनियाई', 'mi': 'Āmeniana', 'mk': 'ерменско писмо', 'ml': 'അർമേനിയൻ', 'mn': 'армени', 'mr': 'अर्मेनियन', 'ms': 'Armenia', 'my': 'အာမေးနီးယား', 'mzn': 'ارمنی', 'nb': 'armensk', 'ne': 'आर्मेनियाली', 'nl': 'Armeens', 'nn': 'armensk', 'no': 'armensk', 'nqo': 'ߊߙߌߡߋߣߌ', 'or': 'ଆର୍ମେନୀୟ', 'pa': 'ਅਰਮੀਨੀਆਈ', 'pa-Guru': 'ਅਰਮੀਨੀਆਈ', 'pcm': 'Armẹ́nia', 'pl': 'ormiańskie', 'ps': 'ارمانیایي', 'pt': 'armênio', 'qu': 'Armenio Simi', 'rm': 'armen', 'ro': 'armeană', 'ru': 'армянская', 'sah': 'Эрмээннии', 'sc': 'armenu', 'sd': 'عرماني', 'sd-Arab': 'عرماني', 'si': 'ආර්මේනියානු', 'sk': 'arménske', 'sl': 'armenski', 'so': 'Armeeniyaan', 'sq': 'armen', 'sr': 'јерменско писмо', 'sr-Cyrl': 'јерменско писмо', 'sr-Latn': 'jermensko pismo', 'sv': 'armeniska', 'sw': 'Kiarmenia', 'syr': 'ܐܪܡܢܝܬ', 'ta': 'அர்மேனியன்', 'te': 'అర్మేనియన్', 'tg': 'Арманӣ', 'th': 'อาร์เมเนีย', 'ti': 'ዓይቡቤን', 'tk': 'Ermeni elipbiýi', 'to': 'tohinima fakaʻāmenia', 'tr': 'Ermeni', 'ug': 'ئەرمەن', 'uk': 'вірменська', 'ur': 'آرمینیائی', 'uz': 'arman', 'uz-Cyrl': 'Арман', 'uz-Latn': 'arman', 'vec': 'armen', 'vi': 'Chữ Armenia', 'wae': 'Armeniš', 'xnr': 'आर्मेनियाई', 'yo': 'Àmẹ́níà', 'yrl': 'arimẽniu', 'yue': '亞美尼亞文', 'yue-Hans': '亚美尼亚文', 'yue-Hant': '亞美尼亞文', 'zh': '亚美尼亚文', 'zh-Hans': '亚美尼亚文', 'zh-Hant': '亞美尼亞文', 'zu': 'isi-Armenian'},
    'Avst': {'ast': 'avésticu', 'az': 'avestan', 'az-Latn': 'avestan', 'bal-Latn': 'Awestái', 'bg': 'Авестанска', 'bn': 'আভেসতান', 'br': 'avestek', 'brx': 'अवस्तन्', 'bs': 'avestansko pismo', 'bs-Cyrl': 'авестанско писмо', 'bs-Latn': 'avestansko pismo', 'ca': 'avèstic', 'ca-ES-valencia': 'avèstic', 'ccp': '𑄃𑄞𑄬𑄥𑄧𑄖𑄚𑄴', 'cs': 'avestánské', 'da': 'avestansk', 'de': 'Avestisch', 'el': 'Αβεστάν', 'el-polyton': 'Αβεστάν', 'en': 'Avestan', 'en-Dsrt': '𐐊𐑂𐐯𐑅𐐻𐐲𐑌', 'es': 'avéstico', 'et': 'avesta', 'eu': 'avestera', 'fa': 'اوستایی', 'ff-Adlm': '𞤀𞤾𞤫𞤧𞤼𞤢𞤲', 'fi': 'avestalainen', 'fo': 'avestanskt', 'fr': 'avestique', 'fy': 'Avestaansk', 'ga': 'Aivéisteach', 'gd': 'Avestanais', 'gsw': 'Aveschtisch', 'gu': 'અવેસ્તન', 'he': 'אווסטן', 'hi': 'अवेस्तन', 'hi-Latn': 'Avestan', 'hr': 'avestansko pismo', 'ht': 'avestique', 'hu': 'Avesztán', 'id': 'Avesta', 'ig': 'Avestan', 'it': 'avestico', 'ja': 'アヴェスター文字', 'ka': 'ავესტური', 'kgp': 'avénhko', 'kk': 'авеста жазуы', 'kk-Cyrl': 'авеста жазуы', 'km': 'អាវេស្ថាន', 'kn': 'ಅವೆಸ್ತಾನ್', 'ko': '아베스타 문자', 'ks': 'اَویستَن', 'ks-Arab': 'اَویستَن', 'lb': 'Avestesch', 'lo': 'ອະເວສຕະ', 'lt': 'avestano', 'mk': 'авестанско', 'ml': 'അവെസ്ഥൻ', 'mn': 'Авестан', 'mr': 'अवेस्तान', 'ms': 'Avestan', 'nb': 'avestisk', 'ne': 'आभेस्टान', 'nl': 'Avestaans', 'nn': 'avestisk', 'no': 'avestisk', 'nqo': 'ߊߝ߭ߋߛߑߕߌ', 'or': 'ଆବେସ୍ଥାନ୍', 'pl': 'awestyjskie', 'pt': 'avéstico', 'rm': 'avestic', 'ro': 'avestică', 'ru': 'авестийская', 'sc': 'avèsticu', 'sl': 'avestanski', 'so': 'Luuqada Avestan', 'sq': 'avestanisht', 'sr': 'авестанско писмо', 'sr-Cyrl': 'авестанско писмо', 'sr-Latn': 'avestansko pismo', 'sv': 'avestiska', 'ta': 'அவெஸ்தான்', 'te': 'అవేస్టాన్', 'tg': 'Авестоӣ', 'th': 'อเวสตะ', 'to': 'tohinima fakaʻavesitani', 'tr': 'Avesta', 'ug': 'ئاۋېستا', 'uk': 'авестійський', 'vec': 'avèstego', 'vi': 'Chữ Avestan', 'xnr': 'अवेस्तन', 'yrl': 'awétiku', 'yue': '阿維斯陀文', 'yue-Hans': '阿维斯陀文', 'yue-Hant': '阿維斯陀文', 'zh': '阿维斯陀文', 'zh-Hans': '阿维斯陀文', 'zh-Hant': '阿維斯陀文', 'zu': 'isi-Avestan'},
    'BA': {'af': 'Bosnië en Herzegowina', 'agq': 'Bosɨnyìa à Hɛ̀zɛ̀gòvinà', 'ak': 'Bosnia ne Hɛzegovina', 'am': 'ቦስኒያ እና ሄርዞጎቪኒያ', 'ar': 'البوسنة والهرسك', 'as': 'ব’ছনিয়া আৰু হাৰ্জেগ’ভিনা', 'asa': 'Bothnia na Hedhegovina', 'ast': 'Bosnia y Herzegovina', 'az': 'Bosniya və Herseqovina', 'az-Cyrl': 'Боснија вә Һерсеговина', 'az-Latn': 'Bosniya və Herseqovina', 'bal-Latn': 'Bósniá', 'bas': 'Bòhnià Ɛrzègòvinà', 'be': 'Боснія і Герцагавіна', 'be-tarask': 'Боснія і Герцагавіна', 'bez': 'Hubosinia na Huhezegovina', 'bg': 'Босна и Херцеговина', 'blo': 'Bɔsniya na Hɛrsegɔfina', 'bm': 'Bozni-Ɛrizigovini', 'bn': 'বসনিয়া ও হার্জেগোভিনা', 'br': 'Bosnia ha Herzegovina', 'brx': 'बसनिया आरो होर्जेगभिना', 'bs': 'Bosna i Hercegovina', 'bs-Cyrl': 'Босна и Херцеговина', 'bs-Latn': 'Bosna i Hercegovina', 'ca': 'Bòsnia i Hercegovina', 'ca-ES-valencia': 'Bòsnia i Hercegovina', 'ccp': '𑄝𑄧𑄥𑄴𑄚𑄨𑄠 𑄃𑄮 𑄦𑄢𑄴𑄎𑄬𑄉𑄮𑄞𑄨𑄚', 'ce': 'Босни а, Герцеговина а', 'ceb': 'Bosnia & Herzegovina', 'cgg': 'Boziniya na Hezegovina', 'chr': 'ᏉᏏᏂᎠ & ᎲᏤᎪᏫᎾ', 'ckb': 'بۆسنیا و ھەرزەگۆڤینا', 'cs': 'Bosna a Hercegovina', 'cv': 'Боснипе Герцеговина', 'cy': 'Bosnia a Herzegovina', 'da': 'Bosnien-Hercegovina', 'dav': 'Bosnia na Hezegovina', 'de': 'Bosnien und Herzegowina', 'dje': 'Bosni nda Herzegovine', 'dsb': 'Bosniska a Hercegowina', 'dyo': 'Bosni di Hersegovin', 'dz': 'བྷོས་ནི་ཡ་ ཨེནཌ་ ཧར་ཛི་གྷོ་བི་ན', 'ebu': 'Bosnia na Hezegovina', 'ee': 'Bosnia kple Herzergovina nutome', 'el': 'Βοσνία - Ερζεγοβίνη', 'el-polyton': 'Βοσνία - Ἐρζεγοβίνη', 'en': 'Bosnia & Herzegovina', 'en-Dsrt': '𐐒𐐱𐑆𐑌𐐨𐐲 𐐰𐑌𐐼 𐐐𐐲𐑉𐐻𐑅𐐲𐑀𐐬𐑂𐐨𐑌𐐲', 'eo': 'Bosnujo kaj Hercegovino', 'es': 'Bosnia y Herzegovina', 'et': 'Bosnia ja Hertsegoviina', 'eu': 'Bosnia-Herzegovina', 'ewo': 'Bosəní ai ɛrəzegovín', 'fa': 'بوسنی و هرزگوین', 'ff': 'Bosnii Hersegowiin', 'ff-Adlm': '𞤄𞤮𞤧𞤲𞤭𞤴𞤢 & 𞤖𞤫𞤪𞤧𞤫𞤳𞤮𞤾𞤭𞤲𞤢𞥄', 'ff-Latn': 'Bosnii Hersegowiin', 'fi': 'Bosnia ja Hertsegovina', 'fil': 'Bosnia and Herzegovina', 'fo': 'Bosnia-Hersegovina', 'fr': 'Bosnie-Herzégovine', 'fur': 'Bosnie e Ercegovine', 'fy': 'Bosnië en Herzegovina', 'ga': 'an Bhoisnia agus an Heirseagaivéin', 'gd': 'Bosna is Hearsagobhana', 'gl': 'Bosnia e Hercegovina', 'gsw': 'Bosnie und Herzegowina', 'gu': 'બોસ્નિયા અને હર્ઝેગોવિના', 'guz': 'Bosnia na Hezegovina', 'ha': 'Bosniya da Harzagobina', 'he': 'בוסניה והרצגובינה', 'hi': 'बोस्निया और हर्ज़ेगोविना', 'hi-Latn': 'Bosnia & Herzegovina', 'hr': 'Bosna i Hercegovina', 'hsb': 'Bosniska a Hercegowina', 'ht': 'Bosnie-Herzégovine', 'hu': 'Bosznia-Hercegovina', 'hy': 'Բոսնիա և Հերցեգովինա', 'ia': 'Bosnia e Herzegovina', 'id': 'Bosnia dan Herzegovina', 'ie': 'Bosnia e Herzegovina', 'ig': 'Bosnia & Herzegovina', 'is': 'Bosnía og Hersegóvína', 'it': 'Bosnia ed Erzegovina', 'ja': 'ボスニア・ヘルツェゴビナ', 'jmc': 'Bosnia na Hezegovina', 'jv': 'Bosnia lan Hèrségovina', 'ka': 'ბოსნია და ჰერცეგოვინა', 'kaa': 'Босния ҳәм Герцеговина', 'kaa-Cyrl': 'Босния ҳәм Герцеговина', 'kab': 'Busna d Hersek', 'kam': 'Bosnia na Hezegovina', 'kde': 'Bosnia na Hezegovina', 'kea': 'Bósnia i Erzegovina', 'kgp': 'Mósinĩja', 'khq': 'Bosni nda Herzegovine', 'ki': 'Bosnia na Hezegovina', 'kk': 'Босния және Герцеговина', 'kk-Cyrl': 'Босния және Герцеговина', 'kln': 'Emetab Bosnia ak Herzegovina', 'km': 'បូស្ន៊ី និងហឺហ្ស៊ីហ្គូវីណា', 'kn': 'ಬೋಸ್ನಿಯಾ ಮತ್ತು ಹರ್ಜೆಗೋವಿನಾ', 'ko': '보스니아 헤르체고비나', 'kok': 'बोस्निया आनी हेर्जेगोविना', 'kok-Deva': 'बोस्निया आनी हेर्जेगोविना', 'ks': 'بوسنِیا تہٕ ہَرزِگووِنا', 'ks-Arab': 'بوسنِیا تہٕ ہَرزِگووِنا', 'ksb': 'Bosnia na Hezegovina', 'ksf': 'bɔsnyɛ ri hɛrsǝgɔvín', 'ksh': 'Boßnije un Herzegovina', 'ku': 'Bosna û Hersek', 'kxv': 'bajnia aḍe hertsegovina', 'kxv-Deva': 'बोस्निया ऑड़े हर्ज़ेगोविना', 'kxv-Latn': 'bajnia aḍe hertsegovina', 'kxv-Orya': 'ବସ୍ନିଆ ଅଡ଼େ ହର୍ଜଗୱିନା', 'kxv-Telu': 'బోస్నియా మరియు బెర్జిగోవినా', 'ky': 'Босния жана Герцеговина', 'lag': 'Bósinia', 'lb': 'Bosnien an Herzegowina', 'lg': 'Boziniya Hezegovina', 'lld': 'Bosnia y Herzegovina', 'ln': 'Bosini mpé Hezegovine', 'lo': 'ບອດສະເນຍ ແລະ ແຮສໂກວີນາ', 'lt': 'Bosnija ir Hercegovina', 'lu': 'Mbosini ne Hezegovine', 'luo': 'Bosnia gi Herzegovina', 'luy': 'Bosnia na Hezegovina', 'lv': 'Bosnija un Hercegovina', 'mai': 'बोस्निया आ हर्जेगोबिना', 'mas': 'Bosnia na Hezegovina', 'mer': 'Bosnia na Hezegovina', 'mfe': 'Bosni-Herzegovinn', 'mg': 'Bosnia sy Herzegovina', 'mi': 'Pōngia-Herekōwini', 'mk': 'Босна и Херцеговина', 'ml': 'ബോസ്നിയയും ഹെർസഗോവിനയും', 'mn': 'Босни-Херцеговин', 'mr': 'बोस्निया अणि हर्जेगोविना', 'ms': 'Bosnia dan Herzegovina', 'mt': 'il-Bożnija-Ħerzegovina', 'mua': 'bosniya ne Herzegovina', 'my': 'ဘော့စနီးယားနှင့် ဟာဇီဂိုဗီနား', 'mzn': 'بوسنی و هرزگوین', 'naq': 'Bosniab tsî Herzegovinab', 'nb': 'Bosnia-Hercegovina', 'nd': 'Bhosnia le Herzegovina', 'ne': 'बोस्निया एण्ड हर्जगोभिनिया', 'nl': 'Bosnië en Herzegovina', 'nmg': 'Bosnia na Ɛrzegovina', 'nn': 'Bosnia-Hercegovina', 'no': 'Bosnia-Hercegovina', 'nqo': 'ߓߐߛߑߣߌ߫-ߍߙߑߖ߭ߋߜ߭ߏߝ߭ߌߣ', 'nus': 'Bothnia kɛnɛ ɣärgobinia', 'nyn': 'Boziniya na Hezegovina', 'om': 'Bosiiniyaa fi Herzoogovinaa', 'or': 'ବୋସନିଆ ଏବଂ ହର୍ଜଗୋଭିନା', 'pa': 'ਬੋਸਨੀਆ ਅਤੇ ਹਰਜ਼ੇਗੋਵੀਨਾ', 'pa-Guru': 'ਬੋਸਨੀਆ ਅਤੇ ਹਰਜ਼ੇਗੋਵੀਨਾ', 'pcm': 'Bọ́zniá & Hẹzẹgovína', 'pl': 'Bośnia i Hercegowina', 'ps': 'بوسنيا او هېرزګوينا', 'pt': 'Bósnia e Herzegovina', 'qu': 'Bosnia y Herzegovina', 'rm': 'Bosnia ed Erzegovina', 'rn': 'Bosiniya na Herigozevine', 'ro': 'Bosnia și Herțegovina', 'rof': 'Bosnia na Hezegovina', 'ru': 'Босния и Герцеговина', 'rwk': 'Bosnia na Hezegovina', 'saq': 'Bosnia na Hezegovina', 'sbp': 'Bosiniya ni Hesegovina', 'sc': 'Bòsnia e Erzegòvina', 'sd': 'بوسنيا ۽ هرزوگووينا', 'sd-Arab': 'بوسنيا ۽ هرزوگووينا', 'se': 'Bosnia-Hercegovina', 'seh': 'Bósnia-Herzegovina', 'ses': 'Bosni nda Herzegovine', 'sg': 'Bosnïi na Herzegovînni', 'shi': 'ⴱⵓⵙⵏⴰ ⴷ ⵀⵉⵔⵙⵉⴽ', 'shi-Latn': 'busna d hirsik', 'shi-Tfng': 'ⴱⵓⵙⵏⴰ ⴷ ⵀⵉⵔⵙⵉⴽ', 'si': 'බොස්නියාව සහ හර්සගොවීනාව', 'sk': 'Bosna a Hercegovina', 'sl': 'Bosna in Hercegovina', 'smn': 'Bosnia já Herzegovina', 'sn': 'Boznia ne Herzegovina', 'so': 'Boosniya & Harsegofina', 'sq': 'Bosnjë-Hercegovinë', 'sr': 'Босна и Херцеговина', 'sr-Cyrl': 'Босна и Херцеговина', 'sr-Latn': 'Bosna i Hercegovina', 'sv': 'Bosnien och Hercegovina', 'sw': 'Bosnia na Hezegovina', 'syr': 'ܒܘܣܢܐ ܘܗܪܬܣܓܘܒܝܢܐ', 'ta': 'போஸ்னியா & ஹெர்ஸகோவினா', 'te': 'బోస్నియా మరియు హెర్జిగోవినా', 'teo': 'Bosnia na Hezegovina', 'tg': 'Босния ва Ҳерсеговина', 'th': 'บอสเนียและเฮอร์เซโกวีนา', 'ti': 'ቦዝንያን ሄርዘጎቪናን', 'tk': 'Bosniýa we Gersegowina', 'to': 'Posinia mo Hesikōvina', 'tr': 'Bosna-Hersek', 'tt': 'Босния һәм Герцеговина', 'twq': 'Bosni nda Herzegovine', 'tzm': 'Busna-d-Hirsik', 'ug': 'بوسىنىيە ۋە گېرتسېگوۋىنا', 'uk': 'Боснія і Герцеговина', 'ur': 'بوسنیا اور ہرزیگووینا', 'uz': 'Bosniya va Gertsegovina', 'uz-Cyrl': 'Босния ва Герцеговина', 'uz-Latn': 'Bosniya va Gertsegovina', 'vai': 'ꕷꔻꕇꕰ ꗪ ꗥꕤꖑꔲꕯ', 'vai-Latn': 'Bhɔsiniya', 'vai-Vaii': 'ꕷꔻꕇꕰ ꗪ ꗥꕤꖑꔲꕯ', 'vec': 'Boznia e Erzegòvina', 'vi': 'Bosnia và Herzegovina', 'vun': 'Bosnia na Hezegovina', 'wae': 'Bosnie und Herzegovina', 'wo': 'Bosni Ersegowin', 'xh': 'EBosnia & Herzegovina', 'xnr': 'बोस्निया कने हर्ज़ेगोविना', 'xog': 'Boziniya Hezegovina', 'yav': 'Pusiní-ɛlkofína', 'yi': 'באסניע הערצעגאווינע', 'yo': 'Bọ̀síníà àti Ẹtisẹgófínà', 'yrl': 'Businiya asuí Eseguwina', 'yue': '波斯尼亞同黑塞哥維那', 'yue-Hans': '波斯尼亚同黑塞哥维那', 'yue-Hant': '波斯尼亞同黑塞哥維那', 'zgh': 'ⴱⵓⵙⵏⴰ ⴷ ⵀⵉⵔⵙⵉⴽ', 'zh': '波斯尼亚和黑塞哥维那', 'zh-Hans': '波斯尼亚和黑塞哥维那', 'zh-Hant': '波士尼亞與赫塞哥維納', 'zu': 'i-Bosnia ne-Herzegovina'},
    'BB': {'af': 'Barbados', 'agq': 'Bàbadòs', 'ak': 'Baabados', 'am': 'ባርቤዶስ', 'ar': 'بربادوس', 'as': 'বাৰ্বাডোচ', 'asa': 'Babadothi', 'ast': 'Barbados', 'az': 'Barbados', 'az-Cyrl': 'Барбадос', 'az-Latn': 'Barbados', 'bal-Latn': 'Barbadós', 'bas': 'Bàrbadò', 'be': 'Барбадас', 'be-tarask': 'Барбадас', 'bez': 'Hubabadosi', 'bg': 'Барбадос', 'blo': 'Barbaɖɔɔsɩ', 'bm': 'Barbadi', 'bn': 'বার্বাডোজ', 'br': 'Barbados', 'brx': 'बारबाड’स', 'bs': 'Barbados', 'bs-Cyrl': 'Барбадос', 'bs-Latn': 'Barbados', 'ca': 'Barbados', 'ca-ES-valencia': 'Barbados', 'ccp': '𑄝𑄢𑄴𑄝𑄘𑄮𑄌𑄴', 'ce': 'Барбадос', 'ceb': 'Barbados', 'cgg': 'Babadosi', 'chr': 'ᏆᏇᏙᏍ', 'ckb': 'باربادۆس', 'cs': 'Barbados', 'cv': 'Барбадос', 'cy': 'Barbados', 'da': 'Barbados', 'dav': 'Babadosi', 'de': 'Barbados', 'dje': 'Barbaados', 'dsb': 'Barbados', 'dyo': 'Barbad', 'dz': 'བྷར་བེ་ཌོས', 'ebu': 'Babadosi', 'ee': 'Barbados nutome', 'el': 'Μπαρμπέιντος', 'el-polyton': 'Μπαρμπέιντος', 'en': 'Barbados', 'en-Dsrt': '𐐒𐐪𐑉𐐺𐐩𐐼𐐬𐑅', 'eo': 'Barbado', 'es': 'Barbados', 'et': 'Barbados', 'eu': 'Barbados', 'ewo': 'Barəbád', 'fa': 'باربادوس', 'ff': 'Barbadoos', 'ff-Adlm': '𞤄𞤢𞤪𞤦𞤫𞥅𞤣𞤮𞥅𞤧', 'ff-Latn': 'Barbadoos', 'fi': 'Barbados', 'fil': 'Barbados', 'fo': 'Barbados', 'fr': 'Barbade', 'fur': 'Barbados', 'fy': 'Barbados', 'ga': 'Barbadós', 'gd': 'Barbados', 'gl': 'Barbados', 'gsw': 'Barbados', 'gu': 'બારબાડોસ', 'guz': 'Babadosi', 'ha': 'Barbadas', 'he': 'ברבדוס', 'hi': 'बारबाडोस', 'hi-Latn': 'Barbados', 'hr': 'Barbados', 'hsb': 'Barbados', 'ht': 'Barbade', 'hu': 'Barbados', 'hy': 'Բարբադոս', 'ia': 'Barbados', 'id': 'Barbados', 'ig': 'Barbados', 'is': 'Barbados', 'it': 'Barbados', 'ja': 'バルバドス', 'jmc': 'Babadosi', 'jv': 'Barbadhos', 'ka': 'ბარბადოსი', 'kaa': 'Барбадос', 'kaa-Cyrl': 'Барбадос', 'kab': 'Barbadus', 'kam': 'Babadosi', 'kde': 'Babadosi', 'kea': 'Barbadus', 'kgp': 'Juvã-mág', 'khq': 'Barbaados', 'ki': 'Babadosi', 'kk': 'Барбадос', 'kk-Cyrl': 'Барбадос', 'kln': 'Emetab Barbados', 'km': 'បាបាដុស', 'kn': 'ಬಾರ್ಬಡೋಸ್', 'ko': '바베이도스', 'kok': 'बारबाडोस', 'kok-Deva': 'बारबाडोस', 'ks': 'باربیڈوس', 'ks-Arab': 'باربیڈوس', 'ksb': 'Babadosi', 'ksf': 'baabaadǝ', 'ksh': 'de Ensel Barbados', 'ku': 'Barbados', 'kxv': 'barbaḍos', 'kxv-Deva': 'बारबाडोस', 'kxv-Latn': 'barbaḍos', 'kxv-Orya': 'ବାର୍\u200dବଡସ', 'kxv-Telu': 'బార్బడోస్', 'ky': 'Барбадос', 'lag': 'Babadóosi', 'lb': 'Barbados', 'lg': 'Barabadosi', 'lld': 'Barbados', 'ln': 'Barɛbadɛ', 'lo': 'ບາບາໂດສ', 'lt': 'Barbadosas', 'lu': 'Barebade', 'luo': 'Barbados', 'luy': 'Babadosi', 'lv': 'Barbadosa', 'mai': 'बारबाडोस', 'mas': 'Babadosi', 'mer': 'Babadosi', 'mfe': 'Barbad', 'mg': 'Barbady', 'mi': 'Papatohe', 'mk': 'Барбадос', 'ml': 'ബാർബഡോസ്', 'mn': 'Барбадос', 'mr': 'बार्बाडोस', 'ms': 'Barbados', 'mt': 'Barbados', 'mua': 'barbadiya', 'my': 'ဘာဘေးဒိုးစ်', 'mzn': 'باربادوس', 'naq': 'Barbados', 'nb': 'Barbados', 'nd': 'Bhabhadosi', 'ne': 'बार्बाडोस', 'nl': 'Barbados', 'nmg': 'Barbado', 'nn': 'Barbados', 'no': 'Barbados', 'nqo': 'ߓߊߙߑߓߊߘ', 'nus': 'Bärbadoth', 'nyn': 'Babadosi', 'om': 'Barbaaros', 'or': 'ବାରବାଡୋସ୍', 'pa': 'ਬਾਰਬਾਡੋਸ', 'pa-Guru': 'ਬਾਰਬਾਡੋਸ', 'pcm': 'Barbédọs', 'pl': 'Barbados', 'ps': 'باربادوس', 'pt': 'Barbados', 'qu': 'Barbados', 'rm': 'Barbados', 'rn': 'Barubadosi', 'ro': 'Barbados', 'rof': 'Babado', 'ru': 'Барбадос', 'rwk': 'Babadosi', 'saq': 'Babadosi', 'sbp': 'Babadosi', 'sc': 'Barbados', 'sd': 'باربڊوس', 'sd-Arab': 'باربڊوس', 'se': 'Barbados', 'seh': 'Barbados', 'ses': 'Barbaados', 'sg': 'Barabâda', 'shi': 'ⴱⴰⵔⴱⴰⴷ', 'shi-Latn': 'barbad', 'shi-Tfng': 'ⴱⴰⵔⴱⴰⴷ', 'si': 'බාබඩෝස්', 'sk': 'Barbados', 'sl': 'Barbados', 'smn': 'Barbados', 'sn': 'Barbados', 'so': 'Baarbadoos', 'sq': 'Barbados', 'sr': 'Барбадос', 'sr-Cyrl': 'Барбадос', 'sr-Latn': 'Barbados', 'sv': 'Barbados', 'sw': 'Babadosi', 'syr': 'ܒܪܒܐܕܘܣ', 'ta': 'பார்படாஸ்', 'te': 'బార్బడోస్', 'teo': 'Babadosi', 'tg': 'Барбадос', 'th': 'บาร์เบโดส', 'ti': 'ባርባዶስ', 'tk': 'Barbados', 'to': 'Pāpeitosi', 'tr': 'Barbados', 'tt': 'Барбадос', 'twq': 'Barbaados', 'tzm': 'Barbadus', 'ug': 'باربادوس', 'uk': 'Барбадос', 'ur': 'بارباڈوس', 'uz': 'Barbados', 'uz-Cyrl': 'Барбадос', 'uz-Latn': 'Barbados', 'vai': 'ꕑꔆꖁꔻ', 'vai-Latn': 'Bhabhedo', 'vai-Vaii': 'ꕑꔆꖁꔻ', 'vec': 'Barbados', 'vi': 'Barbados', 'vun': 'Babadosi', 'wae': 'Barbados', 'wo': 'Barbad', 'xh': 'EBarbados', 'xnr': 'बारबाडोस', 'xog': 'Barabadosi', 'yav': 'Palpatós', 'yi': 'באַרבאַדאס', 'yo': 'Bábádósì', 'yrl': 'Babadu', 'yue': '巴貝多', 'yue-Hans': '巴贝多', 'yue-Hant': '巴貝多', 'zgh': 'ⴱⴰⵔⴱⴰⴷ', 'zh': '巴巴多斯', 'zh-Hans': '巴巴多斯', 'zh-Hant': '巴貝多', 'zu': 'i-Barbados'},
    'BD': {'af': 'Bangladesj', 'agq': 'Baŋgɨ̀làdɛ̂', 'ak': 'Bangladɛhye', 'am': 'ባንግላዲሽ', 'ar': 'بنغلاديش', 'as': 'বাংলাদেশ', 'asa': 'Bangladeshi', 'ast': 'Bangladex', 'az': 'Banqladeş', 'az-Cyrl': 'Бангладеш', 'az-Latn': 'Banqladeş', 'bal-Latn': 'Bangaladésh', 'bas': 'Bàŋglàdɛ̂s', 'be': 'Бангладэш', 'be-tarask': 'Бангладэш', 'bez': 'Hubangaladeshi', 'bg': 'Бангладеш', 'blo': 'Baŋglaɖɛɛshɩ', 'bm': 'Bɛngiladɛsi', 'bn': 'বাংলাদেশ', 'br': 'Bangladesh', 'brx': 'बांलादेश', 'bs': 'Bangladeš', 'bs-Cyrl': 'Бангладеш', 'bs-Latn': 'Bangladeš', 'ca': 'Bangladesh', 'ca-ES-valencia': 'Bangladesh', 'ccp': '𑄝𑄁𑄣𑄘𑄬𑄌𑄴', 'ce': 'Бангладеш', 'ceb': 'Bangladesh', 'cgg': 'Bangaradeshi', 'chr': 'ᏆᏂᎦᎵᏕᏍ', 'ckb': 'بەنگلادیش', 'cs': 'Bangladéš', 'cv': 'Бангладеш', 'cy': 'Bangladesh', 'da': 'Bangladesh', 'dav': 'Bangladeshi', 'de': 'Bangladesch', 'dje': 'Bangladeši', 'dsb': 'Bangladeš', 'dyo': 'Banglades', 'dz': 'བངྒ་ལ་དེཤ', 'ebu': 'Bangladeshi', 'ee': 'Bangladesh nutome', 'el': 'Μπανγκλαντές', 'el-polyton': 'Μπανγκλαντές', 'en': 'Bangladesh', 'en-Dsrt': '𐐒𐐪𐑍𐑀𐑊𐐲𐐼𐐯𐑇', 'eo': 'Bangladeŝo', 'es': 'Bangladés', 'et': 'Bangladesh', 'eu': 'Bangladesh', 'ewo': 'Bangaladɛ́s', 'fa': 'بنگلادش', 'ff': 'Banglaadees', 'ff-Adlm': '𞤄𞤢𞤲𞤺𞤭𞤤𞤢𞤣𞤫𞥅𞤧', 'ff-Latn': 'Banglaadees', 'fi': 'Bangladesh', 'fil': 'Bangladesh', 'fo': 'Bangladesj', 'fr': 'Bangladesh', 'fur': 'Bangladesh', 'fy': 'Bangladesh', 'ga': 'an Bhanglaidéis', 'gd': 'Bangladais', 'gl': 'Bangladesh', 'gsw': 'Bangladesch', 'gu': 'બાંગ્લાદેશ', 'guz': 'Bangladeshi', 'ha': 'Bangladesh', 'he': 'בנגלדש', 'hi': 'बांग्लादेश', 'hi-Latn': 'Bangladesh', 'hr': 'Bangladeš', 'hsb': 'Bangladeš', 'ht': 'Bangladesh', 'hu': 'Banglades', 'hy': 'Բանգլադեշ', 'ia': 'Bangladesh', 'id': 'Bangladesh', 'ig': 'Bangladesh', 'is': 'Bangladess', 'it': 'Bangladesh', 'ja': 'バングラデシュ', 'jmc': 'Bangladeshi', 'jv': 'Banggaladésa', 'ka': 'ბანგლადეში', 'kaa': 'Бангладеш', 'kaa-Cyrl': 'Бангладеш', 'kab': 'Bangladac', 'kam': 'Bangladeshi', 'kde': 'Bangladeshi', 'kea': 'Bangladexi', 'kgp': 'Mágranési', 'khq': 'Bangladeši', 'ki': 'Bangladeshi', 'kk': 'Бангладеш', 'kk-Cyrl': 'Бангладеш', 'kln': 'Emetab Bangladesh', 'km': 'បង់ក្លាដែស', 'kn': 'ಬಾಂಗ್ಲಾದೇಶ', 'ko': '방글라데시', 'kok': 'बांगलादेश', 'kok-Deva': 'बांगलादेश', 'ks': 'بَنگلادیش', 'ks-Arab': 'بَنگلادیش', 'ksb': 'Bangladeshi', 'ksf': 'baŋladɛ́c', 'ksh': 'Bangladesch', 'ku': 'Bengladeş', 'kxv': 'banglades', 'kxv-Deva': 'बांग्लादेस', 'kxv-Latn': 'banglades', 'kxv-Orya': 'ବାଂଲାଦେଶ', 'kxv-Telu': 'బంగ్లాదేస్', 'ky': 'Бангладеш', 'lag': 'Bangaladéeshi', 'lb': 'Bangladesch', 'lg': 'Bangaladesi', 'lld': 'Bangladesc', 'ln': 'Bengalidɛsi', 'lo': 'ບັງກະລາເທດ', 'lt': 'Bangladešas', 'lu': 'Benguladeshi', 'luo': 'Bangladesh', 'luy': 'Bangladeshi', 'lv': 'Bangladeša', 'mai': 'बंगलादेश', 'mas': 'Bangladeshi', 'mer': 'Bangiradeshi', 'mfe': 'Banglades', 'mg': 'Bangladesy', 'mi': 'Pākaratēhi', 'mk': 'Бангладеш', 'ml': 'ബംഗ്ലാദേശ്', 'mn': 'Бангладеш', 'mr': 'बांगलादेश', 'ms': 'Bangladesh', 'mt': 'il-Bangladesh', 'mua': 'bangladeshiya', 'my': 'ဘင်္ဂလားဒေ့ရှ်', 'mzn': 'بنگلادش', 'naq': 'Banglades', 'nb': 'Bangladesh', 'nd': 'Bhangiladeshi', 'ne': 'बङ्गलादेश', 'nl': 'Bangladesh', 'nmg': 'Bɛŋgladɛsh', 'nn': 'Bangladesh', 'no': 'Bangladesh', 'nqo': 'ߓߊ߲ߜ߭ߑߟߊߘߍߛ', 'nus': 'Bengeladiec', 'nyn': 'Bangaradeshi', 'om': 'Banglaadish', 'or': 'ବାଂଲାଦେଶ', 'pa': 'ਬੰਗਲਾਦੇਸ਼', 'pa-Guru': 'ਬੰਗਲਾਦੇਸ਼', 'pcm': 'Bangladẹsh', 'pl': 'Bangladesz', 'ps': 'بنگله دېش', 'pt': 'Bangladesh', 'qu': 'Bangladesh', 'rm': 'Bangladesch', 'rn': 'Bangaladeshi', 'ro': 'Bangladesh', 'rof': 'Bangladeshi', 'ru': 'Бангладеш', 'rwk': 'Bangladeshi', 'saq': 'Bangladeshi', 'sbp': 'Bangiladeshi', 'sc': 'Bangladèsh', 'sd': 'بنگلاديش', 'sd-Arab': 'بنگلاديش', 'se': 'Bangladesh', 'seh': 'Bangladesh', 'ses': 'Bangladeši', 'sg': 'Bengladêshi', 'shi': 'ⴱⴰⵏⴳⵍⴰⴷⵉⵛ', 'shi-Latn': 'bangladic', 'shi-Tfng': 'ⴱⴰⵏⴳⵍⴰⴷⵉⵛ', 'si': 'බංග්ලාදේශය', 'sk': 'Bangladéš', 'sl': 'Bangladeš', 'smn': 'Bangladesh', 'sn': 'Bangladeshi', 'so': 'Bangaladhesh', 'sq': 'Bangladesh', 'sr': 'Бангладеш', 'sr-Cyrl': 'Бангладеш', 'sr-Latn': 'Bangladeš', 'sv': 'Bangladesh', 'sw': 'Bangladeshi', 'syr': 'ܒܢܓܠܐܕܝܫ', 'ta': 'பங்களாதேஷ்', 'te': 'బంగ్లాదేశ్', 'teo': 'Bangladeshi', 'tg': 'Бангладеш', 'th': 'บังกลาเทศ', 'ti': 'ባንግላደሽ', 'tk': 'Bangladeş', 'to': 'Pengilātesi', 'tr': 'Bangladeş', 'tt': 'Бангладеш', 'twq': 'Bangladeši', 'tzm': 'Bangladic', 'ug': 'بېنگال', 'uk': 'Бангладеш', 'ur': 'بنگلہ دیش', 'uz': 'Bangladesh', 'uz-Cyrl': 'Бангладеш', 'uz-Latn': 'Bangladesh', 'vai': 'ꕑꕅꕞꗵꔼ', 'vai-Latn': 'Bhangiladɛ̀shi', 'vai-Vaii': 'ꕑꕅꕞꗵꔼ', 'vec': 'Bangladesh', 'vi': 'Bangladesh', 'vun': 'Bangladeshi', 'wae': 'Bangladeš', 'wo': 'Bengalades', 'xh': 'EBangladesh', 'xnr': 'बांग्लादेश', 'xog': 'Bangaladesi', 'yav': 'Paŋkalatɛs', 'yi': 'באַנגלאַדעש', 'yo': 'Bángáládésì', 'yrl': 'Bãgaradexi', 'yue': '孟加拉', 'yue-Hans': '孟加拉', 'yue-Hant': '孟加拉', 'zgh': 'ⴱⴰⵏⴳⵍⴰⴷⵉⵛ', 'zh': '孟加拉国', 'zh-Hans': '孟加拉国', 'zh-Hant': '孟加拉', 'zu': 'i-Bangladesh'},
    'BE': {'af': 'België', 'agq': 'Bɛɛdzwùm', 'ak': 'Bɛlgyium', 'am': 'ቤልጄም', 'ar': 'بلجيكا', 'as': 'বেলজিয়াম', 'asa': 'Ubelgiji', 'ast': 'Bélxica', 'az': 'Belçika', 'az-Cyrl': 'Белчика', 'az-Latn': 'Belçika', 'bal-Latn': 'Béljiam', 'bas': 'Bɛlgyùm', 'be': 'Бельгія', 'be-tarask': 'Бельгія', 'bez': 'Huubelgiji', 'bg': 'Белгия', 'bgn': 'بیلجیم', 'blo': 'Bɛljiiki', 'bm': 'Bɛliziki', 'bn': 'বেলজিয়াম', 'br': 'Belgia', 'brx': 'बेलजियाम', 'bs': 'Belgija', 'bs-Cyrl': 'Белгија', 'bs-Latn': 'Belgija', 'ca': 'Bèlgica', 'ca-ES-valencia': 'Bèlgica', 'ccp': '𑄝𑄬𑄣𑄴𑄎𑄨𑄠𑄟𑄴', 'ce': 'Бельги', 'ceb': 'Belgium', 'cgg': 'Bubirigi', 'chr': 'ᏇᎵᏥᎥᎻ', 'ckb': 'بەلژیک', 'cs': 'Belgie', 'cv': 'Бельги', 'cy': 'Gwlad Belg', 'da': 'Belgien', 'dav': 'Ubelgiji', 'de': 'Belgien', 'dje': 'Belgiiki', 'dsb': 'Belgiska', 'dyo': 'Belsik', 'dz': 'བྷེལ་ཇམ', 'ebu': 'Ubelgiji', 'ee': 'Belgium nutome', 'el': 'Βέλγιο', 'el-polyton': 'Βέλγιο', 'en': 'Belgium', 'en-Dsrt': '𐐒𐐯𐑊𐐾𐐲𐑋', 'eo': 'Belgujo', 'es': 'Bélgica', 'et': 'Belgia', 'eu': 'Belgika', 'ewo': 'Bɛləhíg', 'fa': 'بلژیک', 'ff': 'Beljik', 'ff-Adlm': '𞤄𞤫𞤤𞤶𞤭𞤳𞤢𞥄', 'ff-Latn': 'Beljik', 'fi': 'Belgia', 'fil': 'Belgium', 'fo': 'Belgia', 'fr': 'Belgique', 'fur': 'Belgjiche', 'fy': 'België', 'ga': 'an Bheilg', 'gd': 'A’ Bheilg', 'gl': 'Bélxica', 'gsw': 'Belgie', 'gu': 'બેલ્જીયમ', 'guz': 'Ubelgiji', 'ha': 'Belgiyom', 'he': 'בלגיה', 'hi': 'बेल्जियम', 'hi-Latn': 'Belgium', 'hr': 'Belgija', 'hsb': 'Belgiska', 'ht': 'Belgique', 'hu': 'Belgium', 'hy': 'Բելգիա', 'ia': 'Belgica', 'id': 'Belgia', 'ie': 'Belgia', 'ig': 'Belgium', 'ii': 'ꀘꆹꏃ', 'is': 'Belgía', 'it': 'Belgio', 'ja': 'ベルギー', 'jmc': 'Ubelgiji', 'jv': 'Bèlgi', 'ka': 'ბელგია', 'kaa': 'Бельгия', 'kaa-Cyrl': 'Бельгия', 'kab': 'Belǧik', 'kam': 'Ubelgiji', 'kde': 'Ubelgidi', 'kea': 'Béljika', 'kgp': 'Mérjika', 'khq': 'Belgiiki', 'ki': 'Ubelgiji', 'kk': 'Бельгия', 'kk-Cyrl': 'Бельгия', 'kln': 'Emetab Belgium', 'km': 'បែលហ្ស៊ិក', 'kn': 'ಬೆಲ್ಜಿಯಮ್', 'ko': '벨기에', 'kok': 'बेल्जियम', 'kok-Deva': 'बेल्जियम', 'ks': 'بیلجِیَم', 'ks-Arab': 'بیلجِیَم', 'ksf': 'bɛljík', 'ksh': 'Bellje', 'ku': 'Belçîka', 'kxv': 'beljiym', 'kxv-Deva': 'बेल्जियम', 'kxv-Latn': 'beljiym', 'kxv-Orya': 'ବେଲଜିୟମ୍', 'kxv-Telu': 'బెల్జియం', 'ky': 'Бельгия', 'lag': 'Ʉbeligíiji', 'lb': 'Belsch', 'lg': 'Bubirigi', 'lld': 'Belgio', 'ln': 'Beleziki', 'lo': 'ເບວຢຽມ', 'lt': 'Belgija', 'lu': 'Belejiki', 'luo': 'Belgium', 'luy': 'Ubelgiji', 'lv': 'Beļģija', 'mai': 'बेल्जियम', 'mas': 'Ubelgiji', 'mer': 'Beronjiamu', 'mfe': 'Belzik', 'mg': 'Belzika', 'mi': 'Peretiama', 'mk': 'Белгија', 'ml': 'ബെൽജിയം', 'mn': 'Бельги', 'mr': 'बेल्जियम', 'ms': 'Belgium', 'mt': 'il-Belġju', 'mua': 'belgika', 'my': 'ဘယ်လ်ဂျီယမ်', 'mzn': 'بلژیک', 'naq': 'Belgiummi', 'nb': 'Belgia', 'nd': 'Bhelgium', 'ne': 'बेल्जियम', 'nl': 'België', 'nmg': 'Bɛlgik', 'nn': 'Belgia', 'no': 'Belgia', 'nqo': 'ߓߍߟߑߖ߭ߌߞ', 'nus': 'Be̱lgim', 'nyn': 'Bubirigi', 'om': 'Beeljiyeem', 'or': 'ବେଲଜିୟମ୍', 'pa': 'ਬੈਲਜੀਅਮ', 'pa-Guru': 'ਬੈਲਜੀਅਮ', 'pcm': 'Bẹ́ljọm', 'pl': 'Belgia', 'ps': 'بیلجیم', 'pt': 'Bélgica', 'qu': 'Bélgica', 'rm': 'Belgia', 'rn': 'Ububiligi', 'ro': 'Belgia', 'rof': 'Ubelgiji', 'ru': 'Бельгия', 'rwk': 'Ubelgiji', 'saq': 'Ubelgiji', 'sbp': 'Ubeligiji', 'sc': 'Bèlgiu', 'sd': 'بيلجيم', 'sd-Arab': 'بيلجيم', 'se': 'Belgia', 'seh': 'Bélgica', 'ses': 'Belgiiki', 'sg': 'Bêleze, Belezîki', 'shi': 'ⴱⵍⵊⵉⴽⴰ', 'shi-Latn': 'bljika', 'shi-Tfng': 'ⴱⵍⵊⵉⴽⴰ', 'si': 'බෙල්ජියම', 'sk': 'Belgicko', 'sl': 'Belgija', 'smn': 'Belgia', 'sn': 'Beljium', 'so': 'Biljam', 'sq': 'Belgjikë', 'sr': 'Белгија', 'sr-Cyrl': 'Белгија', 'sr-Latn': 'Belgija', 'sv': 'Belgien', 'sw': 'Ubelgiji', 'syr': 'ܒܠܓܝܩܐ', 'ta': 'பெல்ஜியம்', 'te': 'బెల్జియం', 'teo': 'Ubelgiji', 'tg': 'Белгия', 'th': 'เบลเยียม', 'ti': 'ቤልጅዩም', 'tk': 'Belgiýa', 'to': 'Pelesiume', 'tr': 'Belçika', 'tt': 'Бельгия', 'twq': 'Belgiiki', 'tzm': 'Beljika', 'ug': 'بېلگىيە', 'uk': 'Бельгія', 'ur': 'بیلجیم', 'uz': 'Belgiya', 'uz-Cyrl': 'Бельгия', 'uz-Latn': 'Belgiya', 'vai': 'ꗩꕀꗚꘋ', 'vai-Latn': 'Bhɛgiyɔŋ', 'vai-Vaii': 'ꗩꕀꗚꘋ', 'vec': 'Beljo', 'vi': 'Bỉ', 'vun': 'Ubelgiji', 'wae': 'Belgie', 'wo': 'Belsig', 'xh': 'EBelgium', 'xnr': 'बेल्जियम', 'xog': 'Bubirigi', 'yav': 'Pɛlsíik', 'yi': 'בעלגיע', 'yo': 'Bégíọ́mù', 'yrl': 'Beujika', 'yue': '比利時', 'yue-Hans': '比利时', 'yue-Hant': '比利時', 'zgh': 'ⴱⵍⵊⵉⴽⴰ', 'zh': '比利时', 'zh-Hans': '比利时', 'zh-Hant': '比利時', 'zu': 'i-Belgium'},
    'BF': {'af': 'Burkina Faso', 'agq': 'Bùkinà Fasò', 'ak': 'Bɔkina Faso', 'am': 'ቡርኪና ፋሶ', 'ar': 'بوركينا فاسو', 'as': 'বুৰকিনা ফাচো', 'asa': 'Bukinafatho', 'ast': 'Burkina Fasu', 'az': 'Burkina Faso', 'az-Cyrl': 'Буркина Фасо', 'az-Latn': 'Burkina Faso', 'bal-Latn': 'Burkiná Pásó', 'bas': 'Bùrkìnà Fasò', 'be': 'Буркіна-Фасо', 'be-tarask': 'Буркіна-Фасо', 'bez': 'Hubukinafaso', 'bg': 'Буркина Фасо', 'blo': 'Burkinaa', 'bm': 'Burukina Faso', 'bn': 'বুরকিনা ফাসো', 'br': 'Burkina Faso', 'brx': 'बुर्किना फास’', 'bs': 'Burkina Faso', 'bs-Cyrl': 'Буркина Фасо', 'bs-Latn': 'Burkina Faso', 'ca': 'Burkina Faso', 'ca-ES-valencia': 'Burkina Faso', 'ccp': '𑄝𑄪𑄢𑄴𑄇𑄨𑄚 𑄜𑄥𑄮', 'ce': 'Буркина- Фасо', 'ceb': 'Burkina Faso', 'cgg': 'Bokina Faso', 'chr': 'ᏋᎩᎾ ᏩᏐ', 'ckb': 'بورکینافاسۆ', 'cs': 'Burkina Faso', 'cv': 'Буркина-Фасо', 'cy': 'Burkina Faso', 'da': 'Burkina Faso', 'dav': 'Bukinafaso', 'de': 'Burkina Faso', 'dje': 'Burkina faso', 'dsb': 'Burkina Faso', 'dyo': 'Burukiina Faso', 'dz': 'བྷར་ཀི་ན་ ཕེ་སོ', 'ebu': 'Bukinafaso', 'ee': 'Burkina Faso nutome', 'el': 'Μπουρκίνα Φάσο', 'el-polyton': 'Μπουρκίνα Φάσο', 'en': 'Burkina Faso', 'en-Dsrt': '𐐒𐐲𐑉𐐿𐐩𐑌𐐲 𐐙𐐰𐑅𐐬', 'eo': 'Burkino', 'es': 'Burkina Faso', 'et': 'Burkina Faso', 'eu': 'Burkina Faso', 'ewo': 'Buləkiná Fasó', 'fa': 'بورکینافاسو', 'ff': 'Burkibaa Faaso', 'ff-Adlm': '𞤄𞤵𞤪𞤳𞤭𞤲𞤢 𞤊𞤢𞤧𞤮𞥅', 'ff-Latn': 'Burkibaa Faaso', 'fi': 'Burkina Faso', 'fil': 'Burkina Faso', 'fo': 'Burkina Faso', 'fr': 'Burkina Faso', 'fur': 'Burkina Faso', 'fy': 'Burkina Faso', 'ga': 'Buircíne Fasó', 'gd': 'Buirciona Faso', 'gl': 'Burkina Faso', 'gsw': 'Burkina Faaso', 'gu': 'બુર્કિના ફાસો', 'guz': 'Bukinafaso', 'ha': 'Burkina Faso', 'he': 'בורקינה פאסו', 'hi': 'बुर्किना फ़ासो', 'hi-Latn': 'Burkina Faso', 'hr': 'Burkina Faso', 'hsb': 'Burkina Faso', 'ht': 'Burkina Faso', 'hu': 'Burkina Faso', 'hy': 'Բուրկինա Ֆասո', 'ia': 'Burkina Faso', 'id': 'Burkina Faso', 'ig': 'Burkina Faso', 'is': 'Búrkína Fasó', 'it': 'Burkina Faso', 'ja': 'ブルキナファソ', 'jgo': 'Mbulukína Fásɔ', 'jmc': 'Bukinafaso', 'jv': 'Burkina Faso', 'ka': 'ბურკინა-ფასო', 'kaa': 'Буркина-Фасо', 'kaa-Cyrl': 'Буркина-Фасо', 'kab': 'Burkina Fasu', 'kam': 'Bukinafaso', 'kde': 'Buchinafaso', 'kea': 'Burkina Fasu', 'kgp': 'Murkinỹ Faso', 'khq': 'Burkina faso', 'ki': 'Bukinafaso', 'kk': 'Буркина-Фасо', 'kk-Cyrl': 'Буркина-Фасо', 'kln': 'Emetab Burkina Faso', 'km': 'បួគីណាហ្វាសូ', 'kn': 'ಬುರ್ಕಿನಾ ಫಾಸೊ', 'ko': '부르키나파소', 'kok': 'बुर्किना फॅसो', 'kok-Deva': 'बुर्किना फॅसो', 'ks': 'بُرکِنا فیسو', 'ks-Arab': 'بُرکِنا فیسو', 'ksb': 'Bukinafaso', 'ksf': 'bukína fǝ́ asɔ', 'ksh': 'Bukkinna-Faase', 'ku': 'Burkîna Faso', 'kxv': 'burkina paso', 'kxv-Deva': 'बुर्किना प़ासो', 'kxv-Latn': 'burkina paso', 'kxv-Orya': 'ବୁର୍କିନା ପାସ୍ୱ', 'kxv-Telu': 'బుర్కినా పాసో', 'ky': 'Буркина-Фасо', 'lag': 'Bukinafáaso', 'lb': 'Burkina Faso', 'lg': 'Burukina Faso', 'lld': 'Burkina Faso', 'ln': 'Bukina Faso', 'lo': 'ເບີກິນາ ຟາໂຊ', 'lt': 'Burkina Fasas', 'lu': 'Bukinafaso', 'luo': 'Burkina Faso', 'luy': 'Bukinafaso', 'lv': 'Burkinafaso', 'mai': 'बुरकिना फासो', 'mas': 'Bukinafaso', 'mer': 'Bukinafaso', 'mfe': 'Burkina Faso', 'mg': 'Borkina Faso', 'mi': 'Pākina Wharo', 'mk': 'Буркина Фасо', 'ml': 'ബർക്കിന ഫാസോ', 'mn': 'Буркина Фасо', 'mr': 'बुर्किना फासो', 'ms': 'Burkina Faso', 'mt': 'il-Burkina Faso', 'mua': 'burkina Faso', 'my': 'ဘာကီးနား ဖားဆို', 'mzn': 'بورکینا فاسو', 'naq': 'Burkina Fasob', 'nb': 'Burkina Faso', 'nd': 'Bhukina Faso', 'ne': 'बुर्किना फासो', 'nl': 'Burkina Faso', 'nmg': 'Burkina Faso', 'nn': 'Burkina Faso', 'no': 'Burkina Faso', 'nqo': 'ߓߙߎߞߌߣߊ߫ ߝߊ߬ߛߏ߫', 'nus': 'Burkinɛ pa̱thu', 'nyn': 'Bokina Faso', 'om': 'Burkiinaa Faasoo', 'or': 'ବୁର୍କିନା ଫାସୋ', 'pa': 'ਬੁਰਕੀਨਾ ਫ਼ਾਸੋ', 'pa-Guru': 'ਬੁਰਕੀਨਾ ਫ਼ਾਸੋ', 'pcm': 'Burkína Fáso', 'pl': 'Burkina Faso', 'ps': 'بورکینا فاسو', 'pt': 'Burquina Faso', 'qu': 'Burkina Faso', 'rm': 'Burkina Faso', 'rn': 'Burukina Faso', 'ro': 'Burkina Faso', 'rof': 'Bukinafaso', 'ru': 'Буркина-Фасо', 'rwk': 'Bukinafaso', 'saq': 'Bukinafaso', 'sbp': 'Bukinafaso', 'sc': 'Burkina Faso', 'sd': 'برڪينا فاسو', 'sd-Arab': 'برڪينا فاسو', 'se': 'Burkina Faso', 'seh': 'Burquina Faso', 'ses': 'Burkina faso', 'sg': 'Burkina Faso', 'shi': 'ⴱⵓⵔⴽⵉⵏⴰ ⴼⴰⵙⵓ', 'shi-Latn': 'burkina fasu', 'shi-Tfng': 'ⴱⵓⵔⴽⵉⵏⴰ ⴼⴰⵙⵓ', 'si': 'බර්කිනා ෆාසෝ', 'sk': 'Burkina Faso', 'sl': 'Burkina Faso', 'smn': 'Burkina Faso', 'sn': 'Bukinafaso', 'so': 'Burkiina Faaso', 'sq': 'Burkina-Faso', 'sr': 'Буркина Фасо', 'sr-Cyrl': 'Буркина Фасо', 'sr-Latn': 'Burkina Faso', 'sv': 'Burkina Faso', 'sw': 'Bukinafaso', 'syr': 'ܒܘܪܩܝܢܐ ܦܐܣܘ', 'ta': 'புர்கினா ஃபாஸோ', 'te': 'బుర్కినా ఫాసో', 'teo': 'Bukinafaso', 'tg': 'Буркина-Фасо', 'th': 'บูร์กินาฟาโซ', 'ti': 'ቡርኪና ፋሶ', 'tk': 'Burkina-Faso', 'to': 'Pekano Faso', 'tr': 'Burkina Faso', 'tt': 'Буркина-Фасо', 'twq': 'Burkina faso', 'tzm': 'Burkina Fasu', 'ug': 'بۇركىنا فاسو', 'uk': 'Буркіна-Фасо', 'ur': 'برکینا فاسو', 'uz': 'Burkina-Faso', 'uz-Cyrl': 'Буркина-Фасо', 'uz-Latn': 'Burkina-Faso', 'vai': 'ꕷꕃꕯ ꕘꖇ', 'vai-Latn': 'Bhokina Fáso', 'vai-Vaii': 'ꕷꕃꕯ ꕘꖇ', 'vec': 'Burkina Fazo', 'vi': 'Burkina Faso', 'vun': 'Bukinafaso', 'wae': 'Burkina Faso', 'wo': 'Burkina Faaso', 'xh': 'EBurkina Faso', 'xnr': 'बुर्किना फ़ासो', 'xog': 'Burukina Faso', 'yav': 'Pulikínafásó', 'yi': 'בורקינע פֿאַסא', 'yo': 'Bùùkíná Fasò', 'yrl': 'Bukina Fasu', 'yue': '布吉納法索', 'yue-Hans': '布吉纳法索', 'yue-Hant': '布吉納法索', 'zgh': 'ⴱⵓⵔⴽⵉⵏⴰ ⴼⴰⵙⵓ', 'zh': '布基纳法索', 'zh-Hans': '布基纳法索', 'zh-Hant': '布吉納法索', 'zu': 'i-Burkina Faso'},
    'BG': {'af': 'Bulgarye', 'agq': 'Bùugɛlìa', 'ak': 'Bɔlgeria', 'am': 'ቡልጋሪያ', 'ar': 'بلغاريا', 'as': 'বুলগেৰিয়া', 'asa': 'Bulgaria', 'ast': 'Bulgaria', 'az': 'Bolqarıstan', 'az-Cyrl': 'Болгарыстан', 'az-Latn': 'Bolqarıstan', 'bal-Latn': 'Balgáriá', 'bas': 'Bùlgarìà', 'be': 'Балгарыя', 'be-tarask': 'Балгарыя', 'bez': 'Hubulgaria', 'bg': 'България', 'bgn': 'بولغاریه', 'blo': 'Bulgarii', 'bm': 'Buligari', 'bn': 'বুলগেরিয়া', 'br': 'Bulgaria', 'brx': 'बुल्गारिया', 'bs': 'Bugarska', 'bs-Cyrl': 'Бугарска', 'bs-Latn': 'Bugarska', 'ca': 'Bulgària', 'ca-ES-valencia': 'Bulgària', 'ccp': '𑄝𑄪𑄣𑄴𑄉𑄬𑄢𑄨𑄠', 'ce': 'Болгари', 'ceb': 'Bulgaria', 'cgg': 'Burugariya', 'chr': 'ᏊᎵᎨᎵᎠ', 'ckb': 'بولگاریا', 'cs': 'Bulharsko', 'cv': 'Болгари', 'cy': 'Bwlgaria', 'da': 'Bulgarien', 'dav': 'Bulgaria', 'de': 'Bulgarien', 'dje': 'Bulgaari', 'dsb': 'Bulgarska', 'dyo': 'Bulgari', 'dz': 'བུལ་ག་རི་ཡ', 'ebu': 'Bulgaria', 'ee': 'Bulgaria nutome', 'el': 'Βουλγαρία', 'el-polyton': 'Βουλγαρία', 'en': 'Bulgaria', 'en-Dsrt': '𐐒𐐲𐑊𐑀𐐯𐑉𐐨𐐲', 'eo': 'Bulgarujo', 'es': 'Bulgaria', 'et': 'Bulgaaria', 'eu': 'Bulgaria', 'ewo': 'Buləgarí', 'fa': 'بلغارستان', 'ff': 'Bulgarii', 'ff-Adlm': '𞤄𞤵𞥅𞤤𞤺𞤢𞤪𞤭𞤴𞤢𞥄', 'ff-Latn': 'Bulgarii', 'fi': 'Bulgaria', 'fil': 'Bulgaria', 'fo': 'Bulgaria', 'fr': 'Bulgarie', 'fur': 'Bulgarie', 'fy': 'Bulgarije', 'ga': 'an Bhulgáir', 'gd': 'A’ Bhulgair', 'gl': 'Bulgaria', 'gsw': 'Bulgaarie', 'gu': 'બલ્ગેરિયા', 'guz': 'Bulgaria', 'ha': 'Bulgariya', 'he': 'בולגריה', 'hi': 'बुल्गारिया', 'hi-Latn': 'Bulgaria', 'hr': 'Bugarska', 'hsb': 'Bołharska', 'ht': 'Bulgarie', 'hu': 'Bulgária', 'hy': 'Բուլղարիա', 'ia': 'Bulgaria', 'id': 'Bulgaria', 'ie': 'Bulgaria', 'ig': 'Bulgaria', 'is': 'Búlgaría', 'it': 'Bulgaria', 'ja': 'ブルガリア', 'jmc': 'Bulgaria', 'jv': 'Bulgari', 'ka': 'ბულგარეთი', 'kaa': 'Болгария', 'kaa-Cyrl': 'Болгария', 'kab': 'Bulgari', 'kam': 'Bulgaria', 'kde': 'Bulgalia', 'kea': 'Bulgária', 'kgp': 'Murgarjia', 'khq': 'Bulgaari', 'ki': 'Bulgaria', 'kk': 'Болгария', 'kk-Cyrl': 'Болгария', 'kln': 'Emetab Bulgaria', 'km': 'ប៊ុលហ្ការី', 'kn': 'ಬಲ್ಗೇರಿಯಾ', 'ko': '불가리아', 'kok': 'बल्गेरीया', 'kok-Deva': 'बल्गेरीया', 'ks': 'بَلجیرِیا', 'ks-Arab': 'بَلجیرِیا', 'ksb': 'Bulgalia', 'ksf': 'bulgarí', 'ksh': 'Bulljaarije', 'ku': 'Bulgaristan', 'kxv': 'bulgaria', 'kxv-Deva': 'बुल्गारिया', 'kxv-Latn': 'bulgaria', 'kxv-Orya': 'ବୁଲ୍\u200dଗାରିଆ', 'kxv-Telu': 'బుల్గేరియ', 'ky': 'Болгария', 'lag': 'Buligaría', 'lb': 'Bulgarien', 'lg': 'Bulugariya', 'lld': 'Bulgaria', 'ln': 'Biligari', 'lo': 'ບັງກາເຣຍ', 'lt': 'Bulgarija', 'lu': 'Biligari', 'luo': 'Bulgaria', 'luy': 'Bulgaria', 'lv': 'Bulgārija', 'mai': 'बुल्गारिया', 'mas': 'Bulgaria', 'mer': 'Bulgaria', 'mfe': 'Bilgari', 'mg': 'Biolgaria', 'mi': 'Purukāria', 'mk': 'Бугарија', 'ml': 'ബൾഗേറിയ', 'mn': 'Болгар', 'mr': 'बल्गेरिया', 'ms': 'Bulgaria', 'mt': 'il-Bulgarija', 'mua': 'bulgariya', 'my': 'ဘူလ်ဂေးရီးယား', 'mzn': 'بلغارستون', 'naq': 'Bulgariab', 'nb': 'Bulgaria', 'nd': 'Bhulgariya', 'ne': 'बुल्गेरिया', 'nl': 'Bulgarije', 'nmg': 'Bulgaria', 'nn': 'Bulgaria', 'no': 'Bulgaria', 'nqo': 'ߓߌߟߑߜ߭ߊ߯ߙߌ߫', 'nus': 'Bulga̱a̱ria', 'nyn': 'Burugariya', 'om': 'Bulgaariyaa', 'or': 'ବୁଲଗେରିଆ', 'pa': 'ਬੁਲਗਾਰੀਆ', 'pa-Guru': 'ਬੁਲਗਾਰੀਆ', 'pcm': 'Bọlgéria', 'pl': 'Bułgaria', 'ps': 'بلغاریه', 'pt': 'Bulgária', 'qu': 'Bulgaria', 'rm': 'Bulgaria', 'rn': 'Buligariya', 'ro': 'Bulgaria', 'rof': 'Bulgaria', 'ru': 'Болгария', 'rwk': 'Bulgaria', 'saq': 'Bulgaria', 'sbp': 'Buligaliya', 'sc': 'Bulgaria', 'sd': 'بلغاريا', 'sd-Arab': 'بلغاريا', 'se': 'Bulgária', 'seh': 'Bulgária', 'ses': 'Bulgaari', 'sg': 'Bulugarïi', 'shi': 'ⴱⵍⵖⴰⵔⵢⴰ', 'shi-Latn': 'blɣara', 'shi-Tfng': 'ⴱⵍⵖⴰⵔⵢⴰ', 'si': 'බල්ගේරියාව', 'sk': 'Bulharsko', 'sl': 'Bolgarija', 'smn': 'Bulgaria', 'sn': 'Bulgaria', 'so': 'Bulgaariya', 'sq': 'Bullgari', 'sr': 'Бугарска', 'sr-Cyrl': 'Бугарска', 'sr-Latn': 'Bugarska', 'sv': 'Bulgarien', 'sw': 'Bulgaria', 'syr': 'ܒܘܠܓܐܪܝܐ', 'ta': 'பல்கேரியா', 'te': 'బల్గేరియా', 'teo': 'Bulgaria', 'tg': 'Булғория', 'th': 'บัลแกเรีย', 'ti': 'ቡልጋርያ', 'tk': 'Bolgariýa', 'to': 'Pulukalia', 'tr': 'Bulgaristan', 'tt': 'Болгария', 'twq': 'Bulgaari', 'tzm': 'Belɣarya', 'ug': 'بۇلغارىيە', 'uk': 'Болгарія', 'ur': 'بلغاریہ', 'uz': 'Bolgariya', 'uz-Cyrl': 'Болгария', 'uz-Latn': 'Bolgariya', 'vai': 'ꗂꔠꔸꕩ', 'vai-Latn': 'Bhɔgeriya', 'vai-Vaii': 'ꗂꔠꔸꕩ', 'vec': 'Bulgarìa', 'vi': 'Bulgaria', 'vun': 'Bulgaria', 'wae': 'Bulgarie', 'wo': 'Bilgari', 'xh': 'EBulgaria', 'xnr': 'बुल्गारिया', 'xog': 'Bulugariya', 'yav': 'Pulukalíi', 'yi': 'בולגאַריע', 'yo': 'Bùùgáríà', 'yrl': 'Bugáriya', 'yue': '保加利亞', 'yue-Hans': '保加利亚', 'yue-Hant': '保加利亞', 'zgh': 'ⴱⵍⵖⴰⵔⵢⴰ', 'zh': '保加利亚', 'zh-Hans': '保加利亚', 'zh-Hant': '保加利亞', 'zu': 'i-Bulgaria'},
    'BH': {'af': 'Bahrein', 'agq': 'Bàlaen', 'ak': 'Baren', 'am': 'ባህሬን', 'ar': 'البحرين', 'as': 'বাহৰেইন', 'asa': 'Bahareni', 'ast': 'Baḥréin', 'az': 'Bəhreyn', 'az-Cyrl': 'Бәһрејн', 'az-Latn': 'Bəhreyn', 'bal-Latn': 'Bahren', 'bas': 'Bàraìn', 'be': 'Бахрэйн', 'be-tarask': 'Бахрэйн', 'bez': 'Hubahaleni', 'bg': 'Бахрейн', 'bgn': 'بحرین', 'blo': 'Barɛɛn', 'bm': 'Bareyini', 'bn': 'বাহারিন', 'br': 'Bahrein', 'brx': 'बाहरैन', 'bs': 'Bahrein', 'bs-Cyrl': 'Бахреин', 'bs-Latn': 'Bahrein', 'ca': 'Bahrain', 'ca-ES-valencia': 'Bahrain', 'ccp': '𑄝𑄦𑄧𑄢𑄭𑄚𑄴', 'ce': 'Бахрейн', 'ceb': 'Bahrain', 'cgg': 'Bahareni', 'chr': 'ᏆᎭᎴᎢᏂ', 'ckb': 'بەحرەین', 'cs': 'Bahrajn', 'cv': 'Бахрейн', 'cy': 'Bahrain', 'da': 'Bahrain', 'dav': 'Bahareni', 'de': 'Bahrain', 'dje': 'Bahareen', 'dsb': 'Bahrain', 'dyo': 'Bahrayn', 'dz': 'བྷ་རེན', 'ebu': 'Bahareni', 'ee': 'Bahrain nutome', 'el': 'Μπαχρέιν', 'el-polyton': 'Μπαχρέιν', 'en': 'Bahrain', 'en-Dsrt': '𐐒𐐪𐑉𐐩𐑌', 'eo': 'Barejno', 'es': 'Baréin', 'et': 'Bahrein', 'eu': 'Bahrain', 'ewo': 'Bahərɛ́n', 'fa': 'بحرین', 'ff': 'Bahreyn', 'ff-Adlm': '𞤄𞤢𞤸𞤢𞤪𞤢𞤴𞤲', 'ff-Latn': 'Bahreyn', 'fi': 'Bahrain', 'fil': 'Bahrain', 'fo': 'Barein', 'fr': 'Bahreïn', 'fur': 'Bahrain', 'fy': 'Bahrein', 'ga': 'Bairéin', 'gd': 'Bachrain', 'gl': 'Bahrain', 'gsw': 'Bachräin', 'gu': 'બેહરીન', 'guz': 'Bahareni', 'ha': 'Baharen', 'he': 'בחריין', 'hi': 'बहरीन', 'hi-Latn': 'Bahrain', 'hr': 'Bahrein', 'hsb': 'Bahrain', 'ht': 'Bahreïn', 'hu': 'Bahrein', 'hy': 'Բահրեյն', 'ia': 'Bahrain', 'id': 'Bahrain', 'ig': 'Bahrain', 'is': 'Barein', 'it': 'Bahrein', 'ja': 'バーレーン', 'jmc': 'Bahareni', 'jv': 'Bahrain', 'ka': 'ბაჰრეინი', 'kaa': 'Бахрейн', 'kaa-Cyrl': 'Бахрейн', 'kab': 'Baḥrin', 'kam': 'Bahareni', 'kde': 'Bahaleni', 'kea': 'Barain', 'kgp': 'Marẽj', 'khq': 'Bahareen', 'ki': 'Bahareni', 'kk': 'Бахрейн', 'kk-Cyrl': 'Бахрейн', 'kln': 'Emetab Bahrain', 'km': 'បារ៉ែន', 'kn': 'ಬಹ್ರೇನ್', 'ko': '바레인', 'kok': 'बेहरेन', 'kok-Deva': 'बेहरेन', 'ks': 'بحریٖن', 'ks-Arab': 'بحریٖن', 'ksb': 'Bahaleni', 'ksf': 'barǝ́n', 'ksh': 'Bachrain', 'ku': 'Behreyn', 'kxv': 'bāren', 'kxv-Deva': 'बहरीन', 'kxv-Latn': 'bāren', 'kxv-Orya': 'ବହରେନ୍', 'kxv-Telu': 'బహరిన్', 'ky': 'Бахрейн', 'lag': 'Baharéeni', 'lb': 'Bahrain', 'lg': 'Baareeni', 'lld': 'Bahrain', 'ln': 'Bahrɛnɛ', 'lo': 'ບາເຣນ', 'lt': 'Bahreinas', 'lu': 'Bahrene', 'luo': 'Bahrain', 'luy': 'Bahareni', 'lv': 'Bahreina', 'mai': 'बहरीन', 'mas': 'Bahareni', 'mer': 'Baharini', 'mfe': 'Bahreïn', 'mg': 'Bahrain', 'mi': 'Pāreina', 'mk': 'Бахреин', 'ml': 'ബഹ്റിൻ', 'mn': 'Бахрейн', 'mr': 'बहारीन', 'ms': 'Bahrain', 'mt': 'il-Bahrain', 'mua': 'bahraiŋ', 'my': 'ဘာရိန်း', 'mzn': 'بحرین', 'naq': 'Bahrain', 'nb': 'Bahrain', 'nd': 'Bhahareni', 'ne': 'बहराइन', 'nl': 'Bahrein', 'nmg': 'Bahrain', 'nn': 'Bahrain', 'no': 'Bahrain', 'nqo': 'ߓߤߊ߬ߙߊߦߌ߬ߣ', 'nus': 'Ba̱reen', 'nyn': 'Bahareni', 'om': 'Baahireen', 'or': 'ବାହାରିନ୍', 'pa': 'ਬਹਿਰੀਨ', 'pa-Guru': 'ਬਹਿਰੀਨ', 'pcm': 'Barein', 'pl': 'Bahrajn', 'ps': 'بحرين', 'pt': 'Barein', 'qu': 'Baréin', 'rm': 'Bahrain', 'rn': 'Bahareyini', 'ro': 'Bahrain', 'rof': 'Bahareni', 'ru': 'Бахрейн', 'rwk': 'Bahareni', 'saq': 'Bahareni', 'sbp': 'Bahaleni', 'sc': 'Bahrein', 'sd': 'بحرين', 'sd-Arab': 'بحرين', 'se': 'Bahrain', 'seh': 'Bahrain', 'ses': 'Bahareen', 'sg': 'Bahrâina', 'shi': 'ⴱⵃⵔⴰⵢⵏ', 'shi-Latn': 'bḥrayn', 'shi-Tfng': 'ⴱⵃⵔⴰⵢⵏ', 'si': 'බහරේන්', 'sk': 'Bahrajn', 'sl': 'Bahrajn', 'smn': 'Bahrain', 'sn': 'Bahareni', 'so': 'Baxreyn', 'sq': 'Bahrejn', 'sr': 'Бахреин', 'sr-Cyrl': 'Бахреин', 'sr-Latn': 'Bahrein', 'sv': 'Bahrain', 'sw': 'Bahareni', 'syr': 'ܒܚܪܝܢ', 'ta': 'பஹ்ரைன்', 'te': 'బహ్రెయిన్', 'teo': 'Bahareni', 'tg': 'Баҳрайн', 'th': 'บาห์เรน', 'ti': 'ባሕሬን', 'tk': 'Bahreýn', 'to': 'Paleini', 'tr': 'Bahreyn', 'tt': 'Бәхрәйн', 'twq': 'Bahareen', 'tzm': 'Baḥrayn', 'ug': 'بەھرەين', 'uk': 'Бахрейн', 'ur': 'بحرین', 'uz': 'Bahrayn', 'uz-Cyrl': 'Баҳрайн', 'uz-Latn': 'Bahrayn', 'vai': 'ꕑꗸꘋ', 'vai-Latn': 'Bharɛŋ', 'vai-Vaii': 'ꕑꗸꘋ', 'vec': 'Barein', 'vi': 'Bahrain', 'vun': 'Bahareni', 'wae': 'Bačrain', 'wo': 'Bahreyin', 'xh': 'EBahrain', 'xnr': 'बहरीन', 'xog': 'Baareeni', 'yav': 'Palɛŋ', 'yo': 'Báránì', 'yrl': 'Barẽi', 'yue': '巴林', 'yue-Hans': '巴林', 'yue-Hant': '巴林', 'zgh': 'ⴱⵃⵔⴰⵢⵏ', 'zh': '巴林', 'zh-Hans': '巴林', 'zh-Hant': '巴林', 'zu': 'i-Bahrain'},
    'BI': {'af': 'Burundi', 'agq': 'Bùlundì', 'ak': 'Burundi', 'am': 'ብሩንዲ', 'ar': 'بوروندي', 'as': 'বুৰুণ্ডি', 'asa': 'Burundi', 'ast': 'Burundi', 'az': 'Burundi', 'az-Cyrl': 'Бурунди', 'az-Latn': 'Burundi', 'bal-Latn': 'Borondi', 'bas': 'Bùrundì', 'be': 'Бурундзі', 'be-tarask': 'Бурундзі', 'bez': 'Huburundi', 'bg': 'Бурунди', 'bgn': 'بروندی', 'blo': 'Burunɖii', 'bm': 'Burundi', 'bn': 'বুরুন্ডি', 'br': 'Burundi', 'brx': 'बुरूण्डी', 'bs': 'Burundi', 'bs-Cyrl': 'Бурунди', 'bs-Latn': 'Burundi', 'ca': 'Burundi', 'ca-ES-valencia': 'Burundi', 'ccp': '𑄝𑄪𑄢𑄪𑄚𑄴𑄘𑄨', 'ce': 'Бурунди', 'ceb': 'Burundi', 'cgg': 'Burundi', 'chr': 'ᏋᎷᏂᏗ', 'ckb': 'بوروندی', 'cs': 'Burundi', 'cv': 'Бурунди', 'cy': 'Burundi', 'da': 'Burundi', 'dav': 'Burundi', 'de': 'Burundi', 'dje': 'Burundi', 'dsb': 'Burundi', 'dyo': 'Burundi', 'dz': 'བྷུ་རུན་ཌི', 'ebu': 'Burundi', 'ee': 'Burundi nutome', 'el': 'Μπουρούντι', 'el-polyton': 'Μπουρούντι', 'en': 'Burundi', 'en-Dsrt': '𐐒𐐲𐑉𐐳𐑌𐐼𐐨', 'eo': 'Burundo', 'es': 'Burundi', 'et': 'Burundi', 'eu': 'Burundi', 'ewo': 'Burundí', 'fa': 'بوروندی', 'ff': 'Burunndi', 'ff-Adlm': '𞤄𞤵𞤪𞤵𞤲𞤣𞤭', 'ff-Latn': 'Burunndi', 'fi': 'Burundi', 'fil': 'Burundi', 'fo': 'Burundi', 'fr': 'Burundi', 'fur': 'Burundi', 'fy': 'Burundi', 'ga': 'an Bhurúin', 'gd': 'Burundaidh', 'gl': 'Burundi', 'gsw': 'Burundi', 'gu': 'બુરુંડી', 'guz': 'Burundi', 'ha': 'Burundi', 'he': 'בורונדי', 'hi': 'बुरुंडी', 'hi-Latn': 'Burundi', 'hr': 'Burundi', 'hsb': 'Burundi', 'ht': 'Burundi', 'hu': 'Burundi', 'hy': 'Բուրունդի', 'ia': 'Burundi', 'id': 'Burundi', 'ig': 'Burundi', 'is': 'Búrúndí', 'it': 'Burundi', 'ja': 'ブルンジ', 'jgo': 'Mbulundí', 'jmc': 'Burundi', 'jv': 'Burundi', 'ka': 'ბურუნდი', 'kaa': 'Бурунди', 'kaa-Cyrl': 'Бурунди', 'kab': 'Burandi', 'kam': 'Mbulundi', 'kde': 'Bulundi', 'kea': 'Burundi', 'kgp': 'Murũni', 'khq': 'Burundi', 'ki': 'Mburundi', 'kk': 'Бурунди', 'kk-Cyrl': 'Бурунди', 'kln': 'Emetab Burundi', 'km': 'ប៊ូរុនឌី', 'kn': 'ಬುರುಂಡಿ', 'ko': '부룬디', 'kok': 'बुरुंडी', 'kok-Deva': 'बुरुंडी', 'ks': 'بورَنڈِ', 'ks-Arab': 'بورَنڈِ', 'ksb': 'Bulundi', 'ksf': 'burundí', 'ksh': 'Burundi', 'ku': 'Bûrûndî', 'kxv': 'buruṇḍī', 'kxv-Deva': 'बुरुंडी', 'kxv-Latn': 'buruṇḍī', 'kxv-Orya': 'ବୁରୁନ୍ଦି', 'kxv-Telu': 'బురుండి', 'ky': 'Бурунди', 'lag': 'Burúundi', 'lb': 'Burundi', 'lg': 'Burundi', 'lld': 'Burundi', 'ln': 'Burundi', 'lo': 'ບູຣຸນດິ', 'lt': 'Burundis', 'lu': 'Burundi', 'luo': 'Burundi', 'luy': 'Burundi', 'lv': 'Burundija', 'mai': 'बुरूंडी', 'mas': 'Burundi', 'mer': 'Burundi', 'mfe': 'Burundi', 'mg': 'Borondi', 'mgh': 'Urundi', 'mi': 'Puruniti', 'mk': 'Бурунди', 'ml': 'ബറുണ്ടി', 'mn': 'Бурунди', 'mr': 'बुरुंडी', 'ms': 'Burundi', 'mt': 'il-Burundi', 'mua': 'burundi', 'my': 'ဘူရွန်ဒီ', 'mzn': 'بوروندی', 'naq': 'Burundib', 'nb': 'Burundi', 'nd': 'Bhurundi', 'ne': 'बुरूण्डी', 'nl': 'Burundi', 'nmg': 'Burundi', 'nn': 'Burundi', 'no': 'Burundi', 'nqo': 'ߓߎߙߎ߲ߘߌ߫', 'nus': 'Burundi', 'nyn': 'Burundi', 'om': 'Burundii', 'or': 'ବୁରୁଣ୍ଡି', 'pa': 'ਬੁਰੁੰਡੀ', 'pa-Guru': 'ਬੁਰੁੰਡੀ', 'pcm': 'Burúndi', 'pl': 'Burundi', 'ps': 'بروندي', 'pt': 'Burundi', 'qu': 'Burundi', 'rm': 'Burundi', 'rn': 'Uburundi', 'ro': 'Burundi', 'rof': 'Burundi', 'ru': 'Бурунди', 'rwk': 'Burundi', 'saq': 'Burundi', 'sbp': 'Bulundi', 'sc': 'Burundi', 'sd': 'برونڊي', 'sd-Arab': 'برونڊي', 'se': 'Burundi', 'seh': 'Burundi', 'ses': 'Burundi', 'sg': 'Burundïi', 'shi': 'ⴱⵓⵔⵓⵏⴷⵉ', 'shi-Latn': 'burundi', 'shi-Tfng': 'ⴱⵓⵔⵓⵏⴷⵉ', 'si': 'බුරුන්දි', 'sk': 'Burundi', 'sl': 'Burundi', 'smn': 'Burundi', 'sn': 'Burundi', 'so': 'Burundi', 'sq': 'Burundi', 'sr': 'Бурунди', 'sr-Cyrl': 'Бурунди', 'sr-Latn': 'Burundi', 'sv': 'Burundi', 'sw': 'Burundi', 'syr': 'ܒܘܪܘܢܕܝ', 'ta': 'புருண்டி', 'te': 'బురుండి', 'teo': 'Burundi', 'tg': 'Бурунди', 'th': 'บุรุนดี', 'ti': 'ብሩንዲ', 'tk': 'Burundi', 'to': 'Puluniti', 'tr': 'Burundi', 'tt': 'Бурунди', 'twq': 'Burundi', 'tzm': 'Burundi', 'ug': 'بۇرۇندى', 'uk': 'Бурунді', 'ur': 'برونڈی', 'uz': 'Burundi', 'uz-Cyrl': 'Бурунди', 'uz-Latn': 'Burundi', 'vai': 'ꖜꖩꔺ', 'vai-Latn': 'Bhurundi', 'vai-Vaii': 'ꖜꖩꔺ', 'vec': 'Burundi', 'vi': 'Burundi', 'vun': 'Burundi', 'wae': 'Burundi', 'wo': 'Burundi', 'xh': 'EBurundi', 'xnr': 'बुरुंडी', 'xog': 'Burundi', 'yav': 'Púlúndí', 'yi': 'בורונדי', 'yo': 'Bùùrúndì', 'yrl': 'Burũdi', 'yue': '蒲隆地', 'yue-Hans': '蒲隆地', 'yue-Hant': '蒲隆地', 'zgh': 'ⴱⵓⵔⵓⵏⴷⵉ', 'zh': '布隆迪', 'zh-Hans': '布隆迪', 'zh-Hant': '蒲隆地', 'zu': 'i-Burundi'},
    'BJ': {'af': 'Benin', 'agq': 'Bɛ̀nɨ̂ŋ', 'ak': 'Bɛnin', 'am': 'ቤኒን', 'ar': 'بنين', 'as': 'বেনিন', 'asa': 'Benini', 'ast': 'Benín', 'az': 'Benin', 'az-Cyrl': 'Бенин', 'az-Latn': 'Benin', 'bal-Latn': 'Benin', 'bas': 'Bènɛ̂ŋ', 'be': 'Бенін', 'be-tarask': 'Бенін', 'bez': 'Hubenini', 'bg': 'Бенин', 'blo': 'Benɛɛ', 'bm': 'Benɛn', 'bn': 'বেনিন', 'br': 'Benin', 'brx': 'बेनिन', 'bs': 'Benin', 'bs-Cyrl': 'Бенин', 'bs-Latn': 'Benin', 'ca': 'Benín', 'ca-ES-valencia': 'Benín', 'ccp': '𑄝𑄬𑄚𑄨𑄚𑄴', 'ce': 'Бенин', 'ceb': 'Benin', 'cgg': 'Benini', 'chr': 'ᏆᏂᎢᏂ', 'ckb': 'بێنین', 'cs': 'Benin', 'cv': 'Бенин', 'cy': 'Benin', 'da': 'Benin', 'dav': 'Benini', 'de': 'Benin', 'dje': 'Beniŋ', 'dsb': 'Benin', 'dyo': 'Bene', 'dz': 'བྷེ་ནིན', 'ebu': 'Benini', 'ee': 'Benin nutome', 'el': 'Μπενίν', 'el-polyton': 'Μπενίν', 'en': 'Benin', 'en-Dsrt': '𐐒𐐲𐑌𐐨𐑌', 'eo': 'Benino', 'es': 'Benín', 'et': 'Benin', 'eu': 'Benin', 'ewo': 'Bəníŋ', 'fa': 'بنین', 'ff': 'Benee', 'ff-Adlm': '𞤄𞤫𞤲𞤫𞤲', 'ff-Latn': 'Benee', 'fi': 'Benin', 'fil': 'Benin', 'fo': 'Benin', 'fr': 'Bénin', 'fur': 'Benin', 'fy': 'Benin', 'ga': 'Beinin', 'gd': 'Beinin', 'gl': 'Benín', 'gsw': 'Benin', 'gu': 'બેનિન', 'guz': 'Benini', 'ha': 'Binin', 'he': 'בנין', 'hi': 'बेनिन', 'hi-Latn': 'Benin', 'hr': 'Benin', 'hsb': 'Benin', 'ht': 'Bénin', 'hu': 'Benin', 'hy': 'Բենին', 'ia': 'Benin', 'id': 'Benin', 'ig': 'Benin', 'is': 'Benín', 'it': 'Benin', 'ja': 'ベナン', 'jgo': 'Mbɛnɛ̂ŋ', 'jmc': 'Benini', 'jv': 'Bénin', 'ka': 'ბენინი', 'kaa': 'Бенин', 'kaa-Cyrl': 'Бенин', 'kab': 'Binin', 'kam': 'Benini', 'kde': 'Benini', 'kea': 'Benin', 'kgp': 'Menĩnh', 'khq': 'Beniŋ', 'ki': 'Benini', 'kk': 'Бенин', 'kk-Cyrl': 'Бенин', 'kln': 'Emetab Benin', 'km': 'បេណាំង', 'kn': 'ಬೆನಿನ್', 'ko': '베냉', 'kok': 'बेनीन', 'kok-Deva': 'बेनीन', 'ks': 'بِنِن', 'ks-Arab': 'بِنِن', 'ksb': 'Benini', 'ksf': 'bɛnǝ́n', 'ksh': 'dä Benin', 'ku': 'Bênîn', 'kxv': 'benīn', 'kxv-Deva': 'बेनिन', 'kxv-Latn': 'benīn', 'kxv-Orya': 'ବେନିନ୍', 'kxv-Telu': 'బెనిన్', 'ky': 'Бенин', 'lag': 'Beníini', 'lb': 'Benin', 'lg': 'Benini', 'lld': 'Benin', 'ln': 'Benɛ', 'lo': 'ເບນິນ', 'lt': 'Beninas', 'lu': 'Bene', 'luo': 'Benin', 'luy': 'Benini', 'lv': 'Benina', 'mai': 'बेनिन', 'mas': 'Benini', 'mer': 'Benini', 'mfe': 'Benin', 'mg': 'Benin', 'mgh': 'Ubelin', 'mi': 'Penīna', 'mk': 'Бенин', 'ml': 'ബെനിൻ', 'mn': 'Бенин', 'mr': 'बेनिन', 'ms': 'Benin', 'mt': 'il-Benin', 'mua': 'beniŋ', 'my': 'ဘီနင်', 'mzn': 'بنین', 'naq': 'Benins', 'nb': 'Benin', 'nd': 'Bhenini', 'ne': 'बेनिन', 'nl': 'Benin', 'nmg': 'Benin', 'nn': 'Benin', 'no': 'Benin', 'nqo': 'ߓߋߣߍ߲߫', 'nus': 'Be̱ni̱n', 'nyn': 'Benini', 'om': 'Beenii', 'or': 'ବେନିନ୍', 'pa': 'ਬੇਨਿਨ', 'pa-Guru': 'ਬੇਨਿਨ', 'pcm': 'Binin', 'pl': 'Benin', 'ps': 'بینن', 'pt': 'Benin', 'qu': 'Benín', 'rm': 'Benin', 'rn': 'Bene', 'ro': 'Benin', 'rof': 'Benini', 'ru': 'Бенин', 'rwk': 'Benini', 'saq': 'Benini', 'sbp': 'Benini', 'sc': 'Benin', 'sd': 'بينن', 'sd-Arab': 'بينن', 'se': 'Benin', 'seh': 'Benin', 'ses': 'Beniŋ', 'sg': 'Benëen', 'shi': 'ⴱⵉⵏⵉⵏ', 'shi-Latn': 'binin', 'shi-Tfng': 'ⴱⵉⵏⵉⵏ', 'si': 'බෙනින්', 'sk': 'Benin', 'sl': 'Benin', 'smn': 'Benin', 'sn': 'Benini', 'so': 'Biniin', 'sq': 'Benin', 'sr': 'Бенин', 'sr-Cyrl': 'Бенин', 'sr-Latn': 'Benin', 'sv': 'Benin', 'sw': 'Benin', 'syr': 'ܒܢܝܢ', 'ta': 'பெனின்', 'te': 'బెనిన్', 'teo': 'Benini', 'tg': 'Бенин', 'th': 'เบนิน', 'ti': 'ቤኒን', 'tk': 'Benin', 'to': 'Penini', 'tr': 'Benin', 'tt': 'Бенин', 'twq': 'Beniŋ', 'tzm': 'Binin', 'ug': 'بېنىن', 'uk': 'Бенін', 'ur': 'بینن', 'uz': 'Benin', 'uz-Cyrl': 'Бенин', 'uz-Latn': 'Benin', 'vai': 'ꗩꕇꘋ', 'vai-Latn': 'Bhɛni', 'vai-Vaii': 'ꗩꕇꘋ', 'vec': 'Benin', 'vi': 'Benin', 'vun': 'Benini', 'wae': 'Benin', 'wo': 'Benee', 'xh': 'EBenin', 'xnr': 'बेनिन', 'xog': 'Benini', 'yav': 'Penɛŋ', 'yi': 'בענין', 'yo': 'Bẹ̀nẹ̀', 'yrl': 'Benĩ', 'yue': '貝南', 'yue-Hans': '贝南', 'yue-Hant': '貝南', 'zgh': 'ⴱⵉⵏⵉⵏ', 'zh': '贝宁', 'zh-Hans': '贝宁', 'zh-Hant': '貝南', 'zu': 'i-Benin'},
    'BL': {'af': 'Sint Barthélemy', 'ak': 'St. Baatilemi', 'am': 'ሴንት ባርቴሌሚ', 'ar': 'سان بارتليمي', 'as': 'ছেইণ্ট বাৰ্থলেমে', 'ast': 'San Bartolomé', 'az': 'Sent-Bartelemi', 'az-Cyrl': 'Сент-Бартелеми', 'az-Latn': 'Sent-Bartelemi', 'bal-Latn': 'St. Bartélémi', 'be': 'Сен-Бартэльмі', 'be-tarask': 'Сен-Бартэльмі', 'bg': 'Сен Бартелеми', 'blo': 'Sɛŋ-Batolomayɔ', 'bn': 'সেন্ট বার্থেলেমি', 'br': 'Saint Barthélemy', 'brx': 'सैन्ट बार्थेलेमी', 'bs': 'Sveti Bartolomej', 'bs-Cyrl': 'Свети Бартоломеј', 'bs-Latn': 'Sveti Bartolomej', 'ca': 'Saint-Barthélemy', 'ca-ES-valencia': 'Saint-Barthélemy', 'ccp': '𑄥𑄬𑄚𑄴𑄑𑄴 𑄝𑄢𑄴𑄗𑄬𑄣𑄨𑄟𑄨', 'ce': 'Сен-Бартельми', 'ceb': 'St. Barthelemy', 'chr': 'ᎤᏓᏅᏘ ᏆᏕᎳᎻ', 'ckb': 'سەن بارتێلێمی', 'cs': 'Svatý Bartoloměj', 'cv': 'Сен-Бартелеми', 'cy': 'Saint Barthélemy', 'da': 'Saint Barthélemy', 'de': 'St. Barthélemy', 'dsb': 'St. Barthélemy', 'dz': 'སེནཊ་ བར་ཐོ་ལོམ་མིའུ', 'ee': 'Saint Barthélemy nutome', 'el': 'Άγιος Βαρθολομαίος', 'el-polyton': 'Άγιος Βαρθολομαίος', 'en': 'St. Barthélemy', 'en-Dsrt': '𐐝𐐩𐑌𐐻 𐐒𐐪𐑉𐐻𐐩𐑊𐐲𐑋𐐨', 'es': 'San Bartolomé', 'et': 'Saint-Barthélemy', 'eu': 'Saint Barthélemy', 'fa': 'سن بارتلمی', 'ff-Adlm': '𞤅𞤼. 𞤄𞤢𞤪𞤼𞤫𞤤𞤭𞤥𞤭', 'fi': 'Saint-Barthélemy', 'fil': 'St. Barthélemy', 'fo': 'St. Barthélemy', 'fr': 'Saint-Barthélemy', 'fur': 'Sant Barthélemy', 'fy': 'Saint Barthélemy', 'ga': 'Saint Barthélemy', 'gd': 'Saint Barthélemy', 'gl': 'Saint Barthélemy', 'gsw': 'St. Barthelemi', 'gu': 'સેંટ બાર્થેલેમી', 'ha': 'San Barthélemy', 'he': 'סנט ברתולומיאו', 'hi': 'सेंट बार्थेलेमी', 'hi-Latn': 'St. Barthelemy', 'hr': 'Saint Barthélemy', 'hsb': 'St. Barthélemy', 'ht': 'Saint-Barthélemy', 'hu': 'Saint-Barthélemy', 'hy': 'Սուրբ Բարդուղիմեոս', 'ia': 'Sancte Bartholomeo', 'id': 'Saint Barthélemy', 'ig': 'St. Barthélemy', 'is': 'Sankti Bartólómeusareyjar', 'it': 'Saint-Barthélemy', 'ja': 'サン・バルテルミー', 'jv': 'Saint Barthélémi', 'ka': 'სენ-ბართელმი', 'kaa': 'Сен-Бартелеми', 'kaa-Cyrl': 'Сен-Бартелеми', 'kea': 'San Bartolomeu', 'kgp': 'Sỹ Martoromeu', 'kk': 'Сен-Бартелеми', 'kk-Cyrl': 'Сен-Бартелеми', 'km': 'សាំង\u200bបាថេឡេមី', 'kn': 'ಸೇಂಟ್ ಬಾರ್ಥೆಲೆಮಿ', 'ko': '생바르텔레미', 'kok': 'सॅंट बार्थेल्मी', 'kok-Deva': 'सॅंट बार्थेल्मी', 'ks': 'سینٹ بارتَھیلمی', 'ks-Arab': 'سینٹ بارتَھیلمی', 'ksh': 'de Zint Battälmi-Enselle', 'ku': 'Saint Barthelemy', 'kxv': 'se barṭelemi', 'kxv-Deva': 'सेंट बार्तेलेमी', 'kxv-Latn': 'se barṭelemi', 'kxv-Orya': 'ସେଣ୍ଟ ବାର୍ତେଲେମି', 'kxv-Telu': 'సెంట్ బర్తెలిమి', 'ky': 'Сент Бартелеми', 'lb': 'Saint-Barthélemy', 'lld': 'St. Barthélemy', 'lo': 'ເຊນ ບາເທເລມີ', 'lt': 'Sen Bartelemi', 'lv': 'Senbartelmī', 'mai': 'सेंट बार्थेलेमी', 'mi': 'Hato Pāteremi', 'mk': 'Свети Вартоломеј', 'ml': 'സെന്റ് ബാർത്തലമി', 'mn': 'Сент-Бартельми', 'mr': 'सेंट बार्थेलेमी', 'ms': 'St. Barthelemy', 'mt': 'Saint Barthélemy', 'my': 'စိန့်ဘာသယ်လ်မီ', 'mzn': 'سنت بارتلمی', 'nb': 'Saint-Barthélemy', 'ne': 'सेन्ट बार्थेलेमी', 'nl': 'Saint-Barthélemy', 'nn': 'Saint-Barthélemy', 'no': 'Saint-Barthélemy', 'nqo': 'ߛߍ߲ߕ-ߓߌߙߑߕߟߋߡߌ߫', 'om': 'St. Barzeleemii', 'or': 'ସେଣ୍ଟ ବାର୍ଥେଲେମି', 'pa': 'ਸੇਂਟ ਬਾਰਥੇਲੇਮੀ', 'pa-Guru': 'ਸੇਂਟ ਬਾਰਥੇਲੇਮੀ', 'pcm': 'Sént Batẹlẹ́mi', 'pl': 'Saint-Barthélemy', 'ps': 'سينټ بارتيلمي', 'pt': 'São Bartolomeu', 'qu': 'San Bartolomé', 'rm': 'Son Barthélemy', 'ro': 'Saint-Barthélemy', 'ru': 'Сен-Бартелеми', 'sc': 'Santu Bartolomeu', 'sd': 'سینٽ برٿلیمی', 'sd-Arab': 'سینٽ برٿلیمی', 'se': 'Saint Barthélemy', 'si': 'ශාන්ත බර්තලෙමි', 'sk': 'Svätý Bartolomej', 'sl': 'Saint Barthélemy', 'smn': 'St. Barthélemy', 'so': 'St. Baathelemiy', 'sq': 'Sen-Bartelemi', 'sr': 'Свети Бартоломеј', 'sr-Cyrl': 'Свети Бартоломеј', 'sr-Latn': 'Sveti Bartolomej', 'sv': 'S:t Barthélemy', 'sw': 'St. Barthelemy', 'syr': 'ܡܪܬܝ ܒܪ ܬܘܠܡܝ', 'ta': 'செயின்ட் பார்தேலெமி', 'te': 'సెయింట్ బర్థెలిమి', 'tg': 'Сент-Бартелми', 'th': 'เซนต์บาร์เธเลมี', 'ti': 'ቅዱስ ባርተለሚ', 'tk': 'Sen-Bartelemi', 'to': 'Sā Patēlemi', 'tr': 'Saint Barthelemy', 'tt': 'Сен-Бартельми', 'ug': 'ساينت بارتېلېمى', 'uk': 'Сен-Бартелемі', 'ur': 'سینٹ برتھلیمی', 'uz': 'Sen-Bartelemi', 'uz-Cyrl': 'Сен-Бартелеми', 'uz-Latn': 'Sen-Bartelemi', 'vai': 'ꕪꘋꕓ ꗞꗢ ꕒꕚꕞꕆ', 'vai-Vaii': 'ꕪꘋꕓ ꗞꗢ ꕒꕚꕞꕆ', 'vec': 'S. Bartolomèo', 'vi': 'St. Barthélemy', 'wae': 'St. Bartholomäus-Insla', 'wo': 'Saŋ Bartalemi', 'xh': 'ESt. Barthélemy', 'xnr': 'सेंट बार्थेलेमी', 'yo': 'Ìlú Bátílẹ́mì', 'yrl': 'Sã Batulumeu', 'yue': '聖巴瑟米', 'yue-Hans': '圣巴瑟米', 'yue-Hant': '聖巴瑟米', 'zh': '圣巴泰勒米', 'zh-Hans': '圣巴泰勒米', 'zh-Hant': '聖巴瑟米', 'zu': 'i-Saint Barthélemy'},
    'BM': {'af': 'Bermuda', 'agq': 'Bɛ̀mudà', 'ak': 'Bɛmuda', 'am': 'ቤርሙዳ', 'ar': 'برمودا', 'as': 'বাৰ্মুডা', 'asa': 'Bermuda', 'ast': 'Les Bermudes', 'az': 'Bermud adaları', 'az-Cyrl': 'Бермуд адалары', 'az-Latn': 'Bermud adaları', 'bal-Latn': 'Barmudá', 'bas': 'Bɛ̀rmudà', 'be': 'Бермудскія астравы', 'be-tarask': 'Бермудскія астравы', 'bez': 'Hubelmuda', 'bg': 'Бермудски острови', 'bgn': 'بیرمودا', 'blo': 'Bɛrmuɖaa', 'bm': 'Bermudi', 'bn': 'বারমুডা', 'br': 'Bermuda', 'brx': 'बारमूडा', 'bs': 'Bermuda', 'bs-Cyrl': 'Бермуди', 'bs-Latn': 'Bermuda', 'ca': 'Bermudes', 'ca-ES-valencia': 'Bermudes', 'ccp': '𑄝𑄢𑄴𑄟𑄪𑄓', 'ce': 'Бермудан гӀайренаш', 'ceb': 'Bermuda', 'cgg': 'Berimuda', 'chr': 'ᏆᏊᏓ', 'ckb': 'بێرموودا', 'cs': 'Bermudy', 'cv': 'Бермуд утравӗсем', 'cy': 'Bermuda', 'da': 'Bermuda', 'dav': 'Bermuda', 'de': 'Bermuda', 'dje': 'Bermuda', 'dsb': 'Bermudy', 'dyo': 'Bermud', 'dz': 'བར་མུ་ཌ', 'ebu': 'Bermuda', 'ee': 'Bermuda nutome', 'el': 'Βερμούδες', 'el-polyton': 'Βερμοῦδες', 'en': 'Bermuda', 'en-Dsrt': '𐐒𐐲𐑉𐑋𐐷𐐭𐐼𐐲', 'eo': 'Bermudoj', 'es': 'Bermudas', 'et': 'Bermuda', 'eu': 'Bermuda', 'ewo': 'Bɛrəmúd', 'fa': 'برمودا', 'ff': 'Bermudaa', 'ff-Adlm': '𞤄𞤭𞤪𞤥𞤵𞤣𞤢', 'ff-Latn': 'Bermudaa', 'fi': 'Bermuda', 'fil': 'Bermuda', 'fo': 'Bermuda', 'fr': 'Bermudes', 'fur': 'Bermuda', 'fy': 'Bermuda', 'ga': 'Beirmiúda', 'gd': 'Bearmùda', 'gl': 'Illas Bermudas', 'gsw': 'Bermuuda', 'gu': 'બર્મુડા', 'guz': 'Bermuda', 'ha': 'Barmuda', 'he': 'ברמודה', 'hi': 'बरमूडा', 'hi-Latn': 'Bermuda', 'hr': 'Bermudi', 'hsb': 'Bermudy', 'ht': 'Bermudes', 'hu': 'Bermuda', 'hy': 'Բերմուդներ', 'ia': 'Bermuda', 'id': 'Bermuda', 'ig': 'Bemuda', 'is': 'Bermúdaeyjar', 'it': 'Bermuda', 'ja': 'バミューダ', 'jmc': 'Bermuda', 'jv': 'Bermuda', 'ka': 'ბერმუდა', 'kaa': 'Бермуд атаўлары', 'kaa-Cyrl': 'Бермуд атаўлары', 'kab': 'Bermuda', 'kam': 'Bermuda', 'kde': 'Belmuda', 'kea': 'Bermudas', 'kgp': 'Mermũna', 'khq': 'Bermuda', 'ki': 'Bermuda', 'kk': 'Бермуд аралдары', 'kk-Cyrl': 'Бермуд аралдары', 'kln': 'Emetab Bermuda', 'km': 'ប៊ឺមុយដា', 'kn': 'ಬರ್ಮುಡಾ', 'ko': '버뮤다', 'kok': 'बर्मुडा', 'kok-Deva': 'बर्मुडा', 'ks': 'برمودا', 'ks-Arab': 'برمودا', 'ksb': 'Belmuda', 'ksf': 'bɛɛmúdǝ', 'ksh': 'de Bermudas', 'ku': 'Bermûda', 'kxv': 'bermuḍā', 'kxv-Deva': 'बरमूडा', 'kxv-Latn': 'bermuḍā', 'kxv-Orya': 'ବର୍\u200dମୁଣ୍ଡା', 'kxv-Telu': 'బర్ముడా', 'ky': 'Бермуд аралдары', 'lag': 'Berimúuda', 'lb': 'Bermuda', 'lg': 'Beremuda', 'lld': 'Bermuda', 'ln': 'Bermuda', 'lo': 'ເບີມິວດາ', 'lt': 'Bermuda', 'lu': 'Bermuda', 'luo': 'Bermuda', 'luy': 'Bermuda', 'lv': 'Bermudu salas', 'mai': 'बरमुडा', 'mas': 'Bermuda', 'mer': 'Bamuda', 'mfe': 'Bermid', 'mg': 'Bermioda', 'mi': 'Pāmura', 'mk': 'Бермуди', 'ml': 'ബർമുഡ', 'mn': 'Бермуда', 'mr': 'बर्मुडा', 'ms': 'Bermuda', 'mt': 'Bermuda', 'mua': 'bermudiya', 'my': 'ဘာမြူဒါ', 'mzn': 'برمودا', 'naq': 'Bermudas', 'nb': 'Bermuda', 'nd': 'Bhemuda', 'ne': 'बर्मुडा', 'nl': 'Bermuda', 'nmg': 'Bɛrmuda', 'nn': 'Bermuda', 'no': 'Bermuda', 'nqo': 'ߓߍߙߑߓߎߘ', 'nus': 'Be̱rmudaa', 'nyn': 'Berimuda', 'om': 'Beermudaa', 'or': 'ବର୍ମୁଡା', 'pa': 'ਬਰਮੂਡਾ', 'pa-Guru': 'ਬਰਮੂਡਾ', 'pcm': 'Bẹmiúda', 'pl': 'Bermudy', 'ps': 'برمودا', 'pt': 'Bermudas', 'qu': 'Bermudas', 'rm': 'Bermudas', 'rn': 'Berimuda', 'ro': 'Bermuda', 'rof': 'Bermuda', 'ru': 'Бермудские о-ва', 'rwk': 'Bermuda', 'saq': 'Bermuda', 'sbp': 'Belimuda', 'sc': 'Bermudas', 'sd': 'برمودا', 'sd-Arab': 'برمودا', 'se': 'Bermuda', 'seh': 'Bermudas', 'ses': 'Bermuda', 'sg': 'Beremûda', 'shi': 'ⴱⵔⵎⵓⴷⴰ', 'shi-Latn': 'brmuda', 'shi-Tfng': 'ⴱⵔⵎⵓⴷⴰ', 'si': 'බර්මියුඩා', 'sk': 'Bermudy', 'sl': 'Bermudi', 'smn': 'Bermuda', 'sn': 'Bermuda', 'so': 'Barmuuda', 'sq': 'Bermude', 'sr': 'Бермуди', 'sr-Cyrl': 'Бермуди', 'sr-Latn': 'Bermudi', 'sv': 'Bermuda', 'sw': 'Bermuda', 'syr': 'ܒܪܡܘܕܐ', 'ta': 'பெர்முடா', 'te': 'బెర్ముడా', 'teo': 'Bermuda', 'tg': 'Бермуда', 'th': 'เบอร์มิวดา', 'ti': 'በርሙዳ', 'tk': 'Bermuda', 'to': 'Pēmuta', 'tr': 'Bermuda', 'tt': 'Бермуд утраулары', 'twq': 'Bermuda', 'tzm': 'Birmuda', 'ug': 'بېرمۇدا', 'uk': 'Бермудські Острови', 'ur': 'برمودا', 'uz': 'Bermuda orollari', 'uz-Cyrl': 'Бермуда', 'uz-Latn': 'Bermuda orollari', 'vai': 'ꗩꖷꕜ', 'vai-Latn': 'Bhɛmuda', 'vai-Vaii': 'ꗩꖷꕜ', 'vec': 'Bermuda', 'vi': 'Bermuda', 'vun': 'Bermuda', 'wae': 'Bermuda', 'wo': 'Bermid', 'xh': 'EBermuda', 'xnr': 'बरमूडा', 'xog': 'Beremuda', 'yav': 'Pɛlmúta', 'yi': 'בערמודע', 'yo': 'Bémúdà', 'yrl': 'Bemuda', 'yue': '百慕達', 'yue-Hans': '百慕达', 'yue-Hant': '百慕達', 'zgh': 'ⴱⵔⵎⵓⴷⴰ', 'zh': '百慕大', 'zh-Hans': '百慕大', 'zh-Hant': '百慕達', 'zu': 'i-Bermuda'},
    'BN': {'af': 'Broenei', 'agq': 'Bɨ̀lunè', 'ak': 'Brunae', 'am': 'ብሩኒ', 'ar': 'بروناي', 'as': 'ব্ৰুনেই', 'asa': 'Brunei', 'ast': 'Brunéi', 'az': 'Bruney', 'az-Cyrl': 'Брунеј', 'az-Latn': 'Bruney', 'bal-Latn': 'Brunái', 'bas': 'Brunei', 'be': 'Бруней', 'be-tarask': 'Бруней', 'bez': 'Hubrunei', 'bg': 'Бруней Даруссалам', 'bgn': 'برونی', 'blo': 'Brunɛɩ', 'bm': 'Burinɛyi', 'bn': 'ব্রুনেই', 'br': 'Brunei', 'brx': 'ब्रूने', 'bs': 'Brunej', 'bs-Cyrl': 'Брунеј', 'bs-Latn': 'Brunej', 'ca': 'Brunei', 'ca-ES-valencia': 'Brunei', 'ccp': '𑄝𑄳𑄢𑄪𑄚𑄬𑄭', 'ce': 'Бруней-Даруссалам', 'ceb': 'Brunei', 'cgg': 'Burunei', 'chr': 'ᏊᎾᎢ', 'ckb': 'بروونای', 'cs': 'Brunej', 'cv': 'Бруней-Даруссалам', 'cy': 'Brunei', 'da': 'Brunei', 'dav': 'Brunei', 'de': 'Brunei Darussalam', 'dje': 'Bruunee', 'dsb': 'Brunei', 'dyo': 'Buruney', 'dz': 'བྷྲུ་ནའི', 'ebu': 'Brunei', 'ee': 'Brunei nutome', 'el': 'Μπρουνέι', 'el-polyton': 'Μπρουνέι', 'en': 'Brunei', 'en-Dsrt': '𐐒𐑉𐐭𐑌𐐴', 'eo': 'Brunejo', 'es': 'Brunéi', 'et': 'Brunei', 'eu': 'Brunei', 'ewo': 'Buluné', 'fa': 'برونئی', 'ff': 'Burnaay', 'ff-Adlm': '𞤄𞤵𞤪𞤲𞤢𞥄𞤴', 'ff-Latn': 'Burnaay', 'fi': 'Brunei', 'fil': 'Brunei', 'fo': 'Brunei', 'fr': 'Brunei', 'fur': 'Brunei', 'fy': 'Brunei', 'ga': 'Brúiné', 'gd': 'Brùnaigh', 'gl': 'Brunei', 'gsw': 'Brunäi Tarussalam', 'gu': 'બ્રુનેઇ', 'guz': 'Brunei', 'ha': 'Burune', 'he': 'ברוניי', 'hi': 'ब्रूनेई', 'hi-Latn': 'Brunei', 'hr': 'Brunej', 'hsb': 'Brunei', 'ht': 'Brunei', 'hu': 'Brunei', 'hy': 'Բրունեյ', 'ia': 'Brunei', 'id': 'Brunei', 'ig': 'Brunei', 'is': 'Brúnei', 'it': 'Brunei', 'ja': 'ブルネイ', 'jmc': 'Brunei', 'jv': 'Brunéi', 'ka': 'ბრუნეი', 'kaa': 'Бруней', 'kaa-Cyrl': 'Бруней', 'kab': 'Bruney', 'kam': 'Brunei', 'kde': 'Blunei', 'kea': 'Brunei', 'kgp': 'Mrunẽj', 'khq': 'Bruunee', 'ki': 'Brunei', 'kk': 'Бруней', 'kk-Cyrl': 'Бруней', 'kln': 'Emetab Brunei', 'km': 'ព្រុយណេ', 'kn': 'ಬ್ರೂನಿ', 'ko': '브루나이', 'kok': 'ब्रूनेई', 'kok-Deva': 'ब्रूनेई', 'ks': 'برونے', 'ks-Arab': 'برونے', 'ksb': 'Blunei', 'ksf': 'brunǝ́', 'ksh': 'Brunei', 'ku': 'Brûney', 'kxv': 'brunae', 'kxv-Deva': 'ब्रूनेई', 'kxv-Latn': 'brunae', 'kxv-Orya': 'ବ୍ରୂନେଇ', 'kxv-Telu': 'బ్రునేఇ', 'ky': 'Бруней', 'lag': 'Burunéei', 'lb': 'Brunei', 'lg': 'Burunayi', 'lld': 'Brunei', 'ln': 'Brineyi', 'lo': 'ບຣູໄນ', 'lt': 'Brunėjus', 'lu': 'Brineyi', 'luo': 'Brunei', 'luy': 'Brunei', 'lv': 'Bruneja', 'mai': 'ब्रूनेई', 'mas': 'Brunei', 'mer': 'Brunei', 'mfe': 'Brunei', 'mg': 'Brunei', 'mi': 'Poronai', 'mk': 'Брунеј', 'ml': 'ബ്രൂണൈ', 'mn': 'Бруней', 'mr': 'ब्रुनेई', 'ms': 'Brunei', 'mt': 'il-Brunei', 'mua': 'bruniya', 'my': 'ဘရူနိုင်း', 'mzn': 'برونئی', 'naq': 'Brunei', 'nb': 'Brunei', 'nd': 'Brunei', 'ne': 'ब्रुनाइ', 'nl': 'Brunei', 'nmg': 'Brunɛi', 'nn': 'Brunei', 'no': 'Brunei', 'nqo': 'ߓߙߎߣߋ߫', 'nus': 'Burunɛy', 'nyn': 'Burunei', 'om': 'Biruniyee', 'or': 'ବ୍ରୁନେଇ', 'pa': 'ਬਰੂਨੇਈ', 'pa-Guru': 'ਬਰੂਨੇਈ', 'pcm': 'Brunẹi', 'pl': 'Brunei', 'ps': 'برونائي', 'pt': 'Brunei', 'qu': 'Brunéi', 'rm': 'Brunei', 'rn': 'Buruneyi', 'ro': 'Brunei', 'rof': 'Brunei', 'ru': 'Бруней', 'rwk': 'Brunei', 'saq': 'Brunei', 'sbp': 'Buluneyi', 'sc': 'Brunei', 'sd': 'برونائي', 'sd-Arab': 'برونائي', 'se': 'Brunei', 'seh': 'Brunei', 'ses': 'Bruunee', 'sg': 'Brunêi', 'shi': 'ⴱⵔⵓⵏⵉ', 'shi-Latn': 'bruni', 'shi-Tfng': 'ⴱⵔⵓⵏⵉ', 'si': 'බෲනායි', 'sk': 'Brunej', 'sl': 'Brunej', 'smn': 'Brunei', 'sn': 'Burunei', 'so': 'Buruneey', 'sq': 'Brunei', 'sr': 'Брунеј', 'sr-Cyrl': 'Брунеј', 'sr-Latn': 'Brunej', 'sv': 'Brunei', 'sw': 'Brunei', 'syr': 'ܒܪܘܢܐܝ', 'ta': 'புருனே', 'te': 'బ్రూనే', 'teo': 'Brunei', 'tg': 'Бруней', 'th': 'บรูไน', 'ti': 'ብሩነይ', 'tk': 'Bruneý', 'to': 'Pulunei', 'tr': 'Brunei', 'tt': 'Бруней', 'twq': 'Bruunee', 'tzm': 'Brunay', 'ug': 'بىرۇنېي', 'uk': 'Бруней', 'ur': 'برونائی', 'uz': 'Bruney', 'uz-Cyrl': 'Бруней', 'uz-Latn': 'Bruney', 'vai': 'ꖜꖩꘉꔧ', 'vai-Latn': 'Bhurunɛĩ', 'vai-Vaii': 'ꖜꖩꘉꔧ', 'vec': 'Brunéi', 'vi': 'Brunei', 'vun': 'Brunei', 'wae': 'Brunei', 'wo': 'Burney', 'xh': 'eBrunei', 'xnr': 'ब्रूनेई', 'xog': 'Burunayi', 'yav': 'Pulunéy', 'yi': 'ברוניי', 'yo': 'Búrúnẹ́lì', 'yrl': 'Burunei', 'yue': '汶萊', 'yue-Hans': '汶莱', 'yue-Hant': '汶萊', 'zgh': 'ⴱⵔⵓⵏⵉ', 'zh': '文莱', 'zh-Hans': '文莱', 'zh-Hant': '汶萊', 'zu': 'i-Brunei'},
    'BO': {'af': 'Bolivië', 'agq': 'Bòlevà', 'ak': 'Bolivia', 'am': 'ቦሊቪያ', 'ar': 'بوليفيا', 'as': 'বলিভিয়া', 'ast': 'Bolivia', 'az': 'Boliviya', 'az-Cyrl': 'Боливија', 'az-Latn': 'Boliviya', 'bal-Latn': 'Boliwiá', 'bas': 'Bòlivìà', 'be': 'Балівія', 'be-tarask': 'Балівія', 'bez': 'Hubolivia', 'bg': 'Боливия', 'bgn': 'بولیویه', 'blo': 'Bolifiya', 'bm': 'Bolivi', 'bn': 'বলিভিয়া', 'br': 'Bolivia', 'brx': 'बलिभिया', 'bs': 'Bolivija', 'bs-Cyrl': 'Боливија', 'bs-Latn': 'Bolivija', 'ca': 'Bolívia', 'ca-ES-valencia': 'Bolívia', 'ccp': '𑄝𑄧𑄣𑄨𑄞𑄨𑄠', 'ce': 'Боливи', 'ceb': 'Bolivia', 'cgg': 'Boriiviya', 'chr': 'ᏉᎵᏫᎠ', 'ckb': 'بۆلیڤیا', 'cs': 'Bolívie', 'cv': 'Боливи', 'cy': 'Bolifia', 'da': 'Bolivia', 'dav': 'Bolivia', 'de': 'Bolivien', 'dje': 'Boolivi', 'dsb': 'Boliwiska', 'dyo': 'Boliivi', 'dz': 'བྷེ་ལི་བི་ཡ', 'ebu': 'Bolivia', 'ee': 'Bolivia nutome', 'el': 'Βολιβία', 'el-polyton': 'Βολιβία', 'en': 'Bolivia', 'en-Dsrt': '𐐒𐐬𐑊𐐮𐑂𐐨𐐲', 'eo': 'Bolivio', 'es': 'Bolivia', 'et': 'Boliivia', 'eu': 'Bolivia', 'ewo': 'Bolívia', 'fa': 'بولیوی', 'ff': 'Boliwii', 'ff-Adlm': '𞤄𞤮𞤤𞤭𞥅𞤾𞤭𞤴𞤢𞥄', 'ff-Latn': 'Boliwii', 'fi': 'Bolivia', 'fil': 'Bolivia', 'fo': 'Bolivia', 'fr': 'Bolivie', 'fur': 'Bolivie', 'fy': 'Bolivia', 'ga': 'an Bholaiv', 'gd': 'Boilibhia', 'gl': 'Bolivia', 'gsw': 'Boliivie', 'gu': 'બોલિવિયા', 'guz': 'Bolivia', 'ha': 'Bolibiya', 'he': 'בוליביה', 'hi': 'बोलीविया', 'hi-Latn': 'Bolivia', 'hr': 'Bolivija', 'hsb': 'Boliwiska', 'ht': 'Bolivie', 'hu': 'Bolívia', 'hy': 'Բոլիվիա', 'ia': 'Bolivia', 'id': 'Bolivia', 'ig': 'Bolivia', 'is': 'Bólivía', 'it': 'Bolivia', 'ja': 'ボリビア', 'jgo': 'Mbɔlivî', 'jmc': 'Bolivia', 'jv': 'Bolivia', 'ka': 'ბოლივია', 'kaa': 'Боливия', 'kaa-Cyrl': 'Боливия', 'kab': 'Bulivi', 'kam': 'Bolivia', 'kde': 'Bolivia', 'kea': 'Bolívia', 'kgp': 'Morivija', 'khq': 'Boolivi', 'ki': 'Bolivia', 'kk': 'Боливия', 'kk-Cyrl': 'Боливия', 'kln': 'Emetab Bolivia', 'km': 'បូលីវី', 'kn': 'ಬೊಲಿವಿಯಾ', 'ko': '볼리비아', 'kok': 'बोलिव्हिया', 'kok-Deva': 'बोलिव्हिया', 'ks': 'بولِوِیا', 'ks-Arab': 'بولِوِیا', 'ksb': 'Bolivia', 'ksf': 'bɔɔlíví', 'ksh': 'Bolliivije', 'ku': 'Bolîvya', 'kxv': 'boliviā', 'kxv-Deva': 'बोलीविया', 'kxv-Latn': 'boliviā', 'kxv-Orya': 'ବଲିୱିୟା', 'kxv-Telu': 'బొలివియా', 'ky': 'Боливия', 'lag': 'Bolívia', 'lb': 'Bolivien', 'lg': 'Boliviya', 'lld': 'Bolivia', 'ln': 'Bolivi', 'lo': 'ໂບລິເວຍ', 'lt': 'Bolivija', 'lu': 'Mbolivi', 'luo': 'Bolivia', 'luy': 'Bolivia', 'lv': 'Bolīvija', 'mai': 'बोलिबिया', 'mas': 'Bolivia', 'mer': 'Bolivia', 'mfe': 'Bolivi', 'mg': 'Bolivia', 'mi': 'Poriwia', 'mk': 'Боливија', 'ml': 'ബൊളീവിയ', 'mn': 'Боливи', 'mr': 'बोलिव्हिया', 'ms': 'Bolivia', 'mt': 'il-Bolivja', 'mua': 'boliviya', 'my': 'ဘိုလီးဗီးယား', 'mzn': 'بولیوی', 'naq': 'Boliviab', 'nb': 'Bolivia', 'nd': 'Bholiviya', 'ne': 'बोलिभिया', 'nl': 'Bolivia', 'nmg': 'Bolivia', 'nn': 'Bolivia', 'no': 'Bolivia', 'nqo': 'ߓߏߟߝ߭ߌ߫', 'nus': 'Bulibia', 'nyn': 'Boriiviya', 'om': 'Boliiviyaa', 'or': 'ବୋଲିଭିଆ', 'pa': 'ਬੋਲੀਵੀਆ', 'pa-Guru': 'ਬੋਲੀਵੀਆ', 'pcm': 'Bolívia', 'pl': 'Boliwia', 'ps': 'بولیویا', 'pt': 'Bolívia', 'qu': 'Bolivia', 'rm': 'Bolivia', 'rn': 'Boliviya', 'ro': 'Bolivia', 'rof': 'Bolivia', 'ru': 'Боливия', 'rwk': 'Bolivia', 'saq': 'Bolivia', 'sbp': 'Boliviya', 'sc': 'Bolìvia', 'sd': 'بوليويا', 'sd-Arab': 'بوليويا', 'se': 'Bolivia', 'seh': 'Bolívia', 'ses': 'Boolivi', 'sg': 'Bolivïi', 'shi': 'ⴱⵓⵍⵉⴼⵢⴰ', 'shi-Latn': 'bulibya', 'shi-Tfng': 'ⴱⵓⵍⵉⴼⵢⴰ', 'si': 'බොලීවියාව', 'sk': 'Bolívia', 'sl': 'Bolivija', 'smn': 'Bolivia', 'sn': 'Bolivia', 'so': 'Boliifiya', 'sq': 'Bolivi', 'sr': 'Боливија', 'sr-Cyrl': 'Боливија', 'sr-Latn': 'Bolivija', 'sv': 'Bolivia', 'sw': 'Bolivia', 'syr': 'ܒܘܠܝܒܝܐ', 'ta': 'பொலிவியா', 'te': 'బొలీవియా', 'teo': 'Bolivia', 'tg': 'Боливия', 'th': 'โบลิเวีย', 'ti': 'ቦሊቭያ', 'tk': 'Boliwiýa', 'to': 'Polīvia', 'tr': 'Bolivya', 'tt': 'Боливия', 'twq': 'Boolivi', 'tzm': 'Bulivya', 'ug': 'بولىۋىيە', 'uk': 'Болівія', 'ur': 'بولیویا', 'uz': 'Boliviya', 'uz-Cyrl': 'Боливия', 'uz-Latn': 'Boliviya', 'vai': 'ꕷꔷꔲꕩ', 'vai-Latn': 'Bholiviya', 'vai-Vaii': 'ꕷꔷꔲꕩ', 'vec': 'Bolivia', 'vi': 'Bolivia', 'vun': 'Bolivia', 'wae': 'Boliwie', 'wo': 'Boliwi', 'xh': 'EBolivia', 'xnr': 'बोलीविया', 'xog': 'Boliviya', 'yav': 'Polífia', 'yi': 'באליוויע', 'yo': 'Bọ̀lífíyà', 'yrl': 'Buríwia', 'yue': '玻利維亞', 'yue-Hans': '玻利维亚', 'yue-Hant': '玻利維亞', 'zgh': 'ⴱⵓⵍⵉⴱⵢⴰ', 'zh': '玻利维亚', 'zh-Hans': '玻利维亚', 'zh-Hant': '玻利維亞', 'zu': 'i-Bolivia'},
    'BQ': {'af': 'Karibiese Nederland', 'ak': 'Caribbean Netherlands', 'am': 'የካሪቢያን ኔዘርላንድስ', 'ar': 'هولندا الكاريبية', 'as': 'কেৰিবিয়ান নেদাৰলেণ্ডছ', 'ast': 'Caribe neerlandés', 'az': 'Karib Niderlandı', 'az-Latn': 'Karib Niderlandı', 'bal-Latn': 'Kerébi Nedarlánd', 'be': 'Карыбскія Нідэрланды', 'be-tarask': 'Карыбскія Нідэрланды', 'bg': 'Карибска Нидерландия', 'blo': 'Holanɖ kaKarayiib', 'bn': 'ক্যারিবিয়ান নেদারল্যান্ডস', 'br': 'Karib Nederlandat', 'brx': 'केरिबियान नेदारलेण्डस', 'bs': 'Karipska Holandija', 'bs-Cyrl': 'Карипска Холандија', 'bs-Latn': 'Karipska Holandija', 'ca': 'Carib Neerlandès', 'ca-ES-valencia': 'Carib Neerlandès', 'ccp': '𑄇𑄳𑄠𑄢𑄨𑄝𑄨𑄠𑄚𑄴 𑄚𑄬𑄘𑄢𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄧𑄥𑄴', 'ce': 'Бонэйр, Синт-Эстатиус а, Саба а', 'ceb': 'Caribbean Netherlands', 'chr': 'ᎧᎵᏈᎢᏂᎯ ᎾᏍᎩᏁᏛᎳᏂ', 'ckb': 'دوورگە کاریبیەکانی هۆڵەندا', 'cs': 'Karibské Nizozemsko', 'cv': 'Бонэйр, Синт-Эстатиус тата Саба', 'cy': 'Antilles yr Iseldiroedd', 'da': 'De tidligere Nederlandske Antiller', 'de': 'Karibische Niederlande', 'dsb': 'Karibiska Nižozemska', 'dz': 'ཀེ་རི་བི་ཡེན་ནེ་དར་ལནཌས྄', 'el': 'Ολλανδία Καραϊβικής', 'el-polyton': 'Ολλανδία Καραϊβικής', 'en': 'Caribbean Netherlands', 'es': 'Caribe neerlandés', 'et': 'Kariibi Madalmaad', 'eu': 'Karibeko Herbehereak', 'fa': 'جزایر کارائیب هلند', 'ff-Adlm': '𞤑𞤢𞤪𞤦𞤭𞤴𞤢𞥄 𞤖𞤮𞤤𞤢𞤲𞤣𞤭𞤴𞤢𞥄', 'fi': 'Karibian Alankomaat', 'fil': 'Caribbean Netherlands', 'fo': 'Niðurlonds Karibia', 'fr': 'Pays-Bas caribéens', 'fy': 'Karibysk Nederlân', 'ga': 'an Ísiltír Chairibeach', 'gd': 'Na Tìrean Ìsle Caraibeach', 'gl': 'Caribe Neerlandés', 'gu': 'કેરેબિયન નેધરલેન્ડ્ઝ', 'ha': 'Caribbean Netherlands', 'he': 'האיים הקריביים ההולנדיים', 'hi': 'कैरिबियन नीदरलैंड', 'hi-Latn': 'Caribbean Netherlands', 'hr': 'Karipski otoci Nizozemske', 'hsb': 'Karibiska Nižozemska', 'ht': 'Pays-Bas caribéens', 'hu': 'Holland Karib-térség', 'hy': 'Կարիբյան Նիդեռլանդներ', 'ia': 'Paises Basse caribe', 'id': 'Belanda Karibia', 'ig': 'Caribbean Netherlands', 'is': 'Karíbahafshluti Hollands', 'it': 'Caraibi Olandesi', 'ja': 'オランダ領カリブ', 'jv': 'Karibia Walanda', 'ka': 'კარიბის ნიდერლანდები', 'kaa': 'Бонэйр, Синт-Эстатиус ҳәм Саба', 'kaa-Cyrl': 'Бонэйр, Синт-Эстатиус ҳәм Саба', 'kea': 'Karaibas Olandezas', 'kgp': 'Pajisi Rur Karimejã', 'kk': 'Бонэйр, Синт-Эстатиус және Саба', 'kk-Cyrl': 'Бонэйр, Синт-Эстатиус және Саба', 'km': 'ហូឡង់ ការ៉ាប៊ីន', 'kn': 'ಕೆರೀಬಿಯನ್ ನೆದರ್\u200cಲ್ಯಾಂಡ್ಸ್', 'ko': '네덜란드령 카리브', 'kok': 'कॅरिबियन निदरलँड', 'kok-Deva': 'कॅरिबियन निदरलँड', 'ks': 'کیریبین نیدرلینڈس', 'ks-Arab': 'کیریبین نیدرلینڈس', 'ksh': 'de karribbesche Nederläng', 'ku': 'Holendaya Karayîbê', 'kxv': 'karibiyn nedrlends', 'kxv-Deva': 'केरिबियन नीदरलेंड', 'kxv-Latn': 'karibiyn nedrlends', 'kxv-Orya': 'କାରିବିୟନ ନେଦରଲ୍ୟାଣ୍ଡସ୍', 'kxv-Telu': 'కరీబియన్ నెదర్లాండ్స', 'ky': 'Кариб Нидерланддары', 'lb': 'Karibescht Holland', 'lld': 'Caraibi olandesc', 'lo': 'ຄາຣິບບຽນ ເນເທີແລນ', 'lt': 'Karibų Nyderlandai', 'lv': 'Nīderlandes Karību salas', 'mai': 'कैरिबियाई नीदरलैंड', 'mi': 'Karapīana Hōrana', 'mk': 'Карипска Холандија', 'ml': 'കരീബിയൻ നെതർലാൻഡ്സ്', 'mn': 'Карибын Нидерланд', 'mr': 'कॅरिबियन नेदरलँड्स', 'ms': 'Belanda Caribbean', 'mt': 'in-Netherlands tal-Karibew', 'my': 'ကာရစ်ဘီယံ နယ်သာလန်', 'mzn': 'هلند ِکاراییبی جزایر', 'nb': 'Karibisk Nederland', 'ne': 'क्यारिवियन नेदरल्याण्ड्स', 'nl': 'Caribisch Nederland', 'nn': 'Karibisk Nederland', 'no': 'Karibisk Nederland', 'nqo': 'ߤߏߟߊ߲ߘ ߞߊߙߌߓߋ߫', 'om': 'Neezerlaandota Kariibaan', 'or': 'କାରବିୟନ୍\u200c ନେଦରଲ୍ୟାଣ୍ଡ', 'pa': 'ਕੈਰੇਬੀਆਈ ਨੀਦਰਲੈਂਡ', 'pa-Guru': 'ਕੈਰੇਬੀਆਈ ਨੀਦਰਲੈਂਡ', 'pcm': 'Karíbián Nẹ́dalands', 'pl': 'Niderlandy Karaibskie', 'ps': 'کیریبین هالینډ', 'pt': 'Países Baixos Caribenhos', 'qu': 'Bonaire', 'rm': 'Antillas Ollandaisas', 'ro': 'Insulele Caraibe Olandeze', 'ru': 'Бонэйр, Синт-Эстатиус и Саба', 'sc': 'Caràibes olandesas', 'sd': 'ڪيريبين نيدرلينڊ', 'sd-Arab': 'ڪيريبين نيدرلينڊ', 'si': 'කැරිබියානු නෙදර්ලන්තය', 'sk': 'Karibské Holandsko', 'sl': 'Nizozemski Karibi', 'so': 'Karibiyaan Nadarlands', 'sq': 'Karaibet holandeze', 'sr': 'Карипска Холандија', 'sr-Cyrl': 'Карипска Холандија', 'sr-Latn': 'Karipska Holandija', 'sv': 'Karibiska Nederländerna', 'sw': 'Uholanzi ya Karibiani', 'syr': 'ܟܐܪܝܒܝܢ ܕܢܝܬܝܪܠܐܢܕܣ', 'ta': 'கரீபியன் நெதர்லாந்து', 'te': 'కరీబియన్ నెదర్లాండ్స్', 'tg': 'Кариби Нидерланд', 'th': 'เนเธอร์แลนด์แคริบเบียน', 'ti': 'ካሪብያን ኔዘርላንድ', 'tk': 'Karib Niderlandlary', 'to': 'Kalipiane fakahōlani', 'tr': 'Karayip Hollandası', 'tt': 'Кариб Нидерландлары', 'ug': 'كارىب دېڭىزى گوللاندىيە', 'uk': 'Карибські Нідерланди', 'ur': 'کریبیائی نیدرلینڈز', 'uz': 'Boneyr, Sint-Estatius va Saba', 'uz-Cyrl': 'Бонейр, Синт-Эстатиус ва Саба', 'uz-Latn': 'Boneyr, Sint-Estatius va Saba', 'vai': 'ꕪꔓꔬꘂꘋ ꖨꕮ ꗨꗳꗣ', 'vai-Vaii': 'ꕪꔓꔬꘂꘋ ꖨꕮ ꗨꗳꗣ', 'vec': 'Paezi Basi caraìbeghi', 'vi': 'Ca-ri-bê Hà Lan', 'wo': 'Pays-Bas bu Caraïbe', 'xh': 'ECaribbean Netherlands', 'xnr': 'कैरिबियन नीदरलैंड', 'yo': 'Kàríbíánì ti Nẹ́dálándì', 'yrl': 'Tetãma Iwiboí-ita Karíbi suí', 'yue': '荷蘭加勒比區', 'yue-Hans': '荷兰加勒比区', 'yue-Hant': '荷蘭加勒比區', 'zh': '荷属加勒比区', 'zh-Hans': '荷属加勒比区', 'zh-Hant': '荷蘭加勒比區', 'zu': 'i-Caribbean Netherlands'},
    'BR': {'af': 'Brasilië', 'agq': 'Bɨ̀làzîi', 'ak': 'Brazil', 'am': 'ብራዚል', 'ar': 'البرازيل', 'as': 'ব্ৰাজিল', 'asa': 'Brazili', 'ast': 'Brasil', 'az': 'Braziliya', 'az-Cyrl': 'Бразилија', 'az-Latn': 'Braziliya', 'bal-Latn': 'Brázil', 'bas': 'Bràsîl', 'be': 'Бразілія', 'be-tarask': 'Бразілія', 'bez': 'Hublazili', 'bg': 'Бразилия', 'bgn': 'برازیل', 'blo': 'Bresil', 'bm': 'Berezili', 'bn': 'ব্রাজিল', 'br': 'Brazil', 'brx': 'ब्राजील', 'bs': 'Brazil', 'bs-Cyrl': 'Бразил', 'bs-Latn': 'Brazil', 'ca': 'Brasil', 'ca-ES-valencia': 'Brasil', 'ccp': '𑄝𑄳𑄢𑄎𑄨𑄣𑄴', 'ce': 'Бразили', 'ceb': 'Brazil', 'cgg': 'Buraziiri', 'chr': 'ᏆᏏᎵ', 'ckb': 'برازیل', 'cs': 'Brazílie', 'cv': 'Бразили', 'cy': 'Brasil', 'da': 'Brasilien', 'dav': 'Brazili', 'de': 'Brasilien', 'dje': 'Breezil', 'doi': 'ब्राजील', 'dsb': 'Brazilska', 'dyo': 'Bresil', 'dz': 'བྲ་ཛིལ', 'ebu': 'Brazili', 'ee': 'Brazil nutome', 'el': 'Βραζιλία', 'el-polyton': 'Βραζιλία', 'en': 'Brazil', 'en-Dsrt': '𐐒𐑉𐐲𐑆𐐮𐑊', 'eo': 'Brazilo', 'es': 'Brasil', 'et': 'Brasiilia', 'eu': 'Brasil', 'ewo': 'Bəlazíl', 'fa': 'برزیل', 'ff': 'Beresiil', 'ff-Adlm': '𞤄𞤪𞤢𞤧𞤭𞤤', 'ff-Latn': 'Beresiil', 'fi': 'Brasilia', 'fil': 'Brazil', 'fo': 'Brasil', 'fr': 'Brésil', 'fur': 'Brasîl', 'fy': 'Brazilië', 'ga': 'an Bhrasaíl', 'gd': 'Braisil', 'gl': 'Brasil', 'gsw': 'Brasilie', 'gu': 'બ્રાઝિલ', 'guz': 'Brazili', 'ha': 'Birazil', 'he': 'ברזיל', 'hi': 'ब्राज़ील', 'hi-Latn': 'Brazil', 'hr': 'Brazil', 'hsb': 'Brazilska', 'ht': 'Brésil', 'hu': 'Brazília', 'hy': 'Բրազիլիա', 'ia': 'Brasil', 'id': 'Brasil', 'ig': 'Brazil', 'ii': 'ꀠꑭ', 'is': 'Brasilía', 'it': 'Brasile', 'ja': 'ブラジル', 'jgo': 'Mbɛlazîl', 'jmc': 'Brazili', 'jv': 'Brasil', 'ka': 'ბრაზილია', 'kaa': 'Бразилия', 'kaa-Cyrl': 'Бразилия', 'kab': 'Brizil', 'kam': 'Brazili', 'kde': 'Blazili', 'kea': 'Brazil', 'kgp': 'Mrasir', 'khq': 'Breezil', 'ki': 'Brazili', 'kk': 'Бразилия', 'kk-Cyrl': 'Бразилия', 'kln': 'Emetab Brazil', 'km': 'ប្រេស៊ីល', 'kn': 'ಬ್ರೆಜಿಲ್', 'ko': '브라질', 'kok': 'ब्राझील', 'kok-Deva': 'ब्राझील', 'ks': 'برازِل', 'ks-Arab': 'برازِل', 'ks-Deva': 'ब्राज़ील', 'ksb': 'Blazili', 'ksf': 'brɛsíl', 'ksh': 'Brasilije', 'ku': 'Brezîlya', 'kxv': 'brājil', 'kxv-Deva': 'ब्राज़ील', 'kxv-Latn': 'brājil', 'kxv-Orya': 'ବ୍ରାଜିଲ୍', 'kxv-Telu': 'బ్రాజిల్', 'ky': 'Бразилия', 'lag': 'Brasíili', 'lb': 'Brasilien', 'lg': 'Buraziiri', 'lld': 'Brasil', 'ln': 'Brezílɛ', 'lo': 'ບຣາຊິວ', 'lrc': 'بئرئزیل', 'lt': 'Brazilija', 'lu': 'Mnulezile', 'luo': 'Brazil', 'luy': 'Brazili', 'lv': 'Brazīlija', 'mai': 'ब्राज़ील', 'mas': 'Brazili', 'mer': 'Brazilu', 'mfe': 'Brezil', 'mg': 'Brezila', 'mi': 'Parīhi', 'mk': 'Бразил', 'ml': 'ബ്രസീൽ', 'mn': 'Бразил', 'mni': 'ব্রাজিল', 'mni-Beng': 'ব্রাজিল', 'mr': 'ब्राझिल', 'ms': 'Brazil', 'mt': 'Il-Brażil', 'mua': 'brazilya', 'my': 'ဘရာဇီး', 'mzn': 'برزیل', 'naq': 'Braziliab', 'nb': 'Brasil', 'nd': 'Brazili', 'ne': 'ब्राजिल', 'nl': 'Brazilië', 'nmg': 'Brésil', 'nn': 'Brasil', 'no': 'Brasil', 'nqo': 'ߓߙߋߖ߭ߌߟ', 'nus': 'Bäraadhiil', 'nyn': 'Buraziiri', 'om': 'Biraazil', 'or': 'ବ୍ରାଜିଲ୍', 'os': 'Бразили', 'pa': 'ਬ੍ਰਾਜ਼ੀਲ', 'pa-Guru': 'ਬ੍ਰਾਜ਼ੀਲ', 'pcm': 'Brázil', 'pl': 'Brazylia', 'ps': 'برازیل', 'pt': 'Brasil', 'qu': 'Brasil', 'rm': 'Brasilia', 'rn': 'Burezili', 'ro': 'Brazilia', 'rof': 'Brazili', 'ru': 'Бразилия', 'rwk': 'Brazili', 'sa': 'ब्राजील', 'sah': 'Бразилия', 'saq': 'Brazili', 'sat': 'ᱵᱨᱟᱡᱤᱞ', 'sat-Olck': 'ᱵᱨᱟᱡᱤᱞ', 'sbp': 'Bulasili', 'sc': 'Brasile', 'sd': 'برازيل', 'sd-Arab': 'برازيل', 'sd-Deva': 'ब्राज़ील', 'se': 'Brasil', 'seh': 'Brasil', 'ses': 'Breezil', 'sg': 'Brezîli', 'shi': 'ⴱⵔⴰⵣⵉⵍ', 'shi-Latn': 'brazil', 'shi-Tfng': 'ⴱⵔⴰⵣⵉⵍ', 'si': 'බ්\u200dරසීලය', 'sk': 'Brazília', 'sl': 'Brazilija', 'smn': 'Brasilia', 'sn': 'Brazil', 'so': 'Baraasiil', 'sq': 'Brazil', 'sr': 'Бразил', 'sr-Cyrl': 'Бразил', 'sr-Latn': 'Brazil', 'su': 'Brasil', 'su-Latn': 'Brasil', 'sv': 'Brasilien', 'sw': 'Brazil', 'syr': 'ܒܪܐܙܝܠ', 'ta': 'பிரேசில்', 'te': 'బ్రెజిల్', 'teo': 'Brazili', 'tg': 'Бразилия', 'th': 'บราซิล', 'ti': 'ብራዚል', 'tk': 'Braziliýa', 'to': 'Palāsili', 'tpi': 'Brasil', 'tr': 'Brezilya', 'tt': 'Бразилия', 'twq': 'Breezil', 'tzm': 'Bṛazil', 'ug': 'بىرازىلىيە', 'uk': 'Бразилія', 'ur': 'برازیل', 'uz': 'Braziliya', 'uz-Cyrl': 'Бразилия', 'uz-Latn': 'Braziliya', 'vai': 'ꖜꕟꔘꔀ', 'vai-Latn': 'Bhurazeli', 'vai-Vaii': 'ꖜꕟꔘꔀ', 'vec': 'Brazile', 'vi': 'Brazil', 'vun': 'Brazili', 'wae': 'Brasilie', 'wo': 'Beresil', 'xh': 'EBrazil', 'xnr': 'ब्राज़ील', 'xog': 'Buraziiri', 'yav': 'Pilesíl', 'yi': 'בראַזיל', 'yo': 'Bàràsílì', 'yrl': 'Brasiu', 'yue': '巴西', 'yue-Hans': '巴西', 'yue-Hant': '巴西', 'zgh': 'ⴱⵔⴰⵣⵉⵍ', 'zh': '巴西', 'zh-Hans': '巴西', 'zh-Hant': '巴西', 'zu': 'i-Brazil'},
    'BS': {'af': 'Bahamas', 'agq': 'Bàhamàs', 'ak': 'Bahama', 'am': 'ባሃማስ', 'ar': 'جزر البهاما', 'as': 'বাহামাছ', 'asa': 'Bahama', 'ast': 'Les Bahames', 'az': 'Baham adaları', 'az-Cyrl': 'Баһам адалары', 'az-Latn': 'Baham adaları', 'bal-Latn': 'Bahamas', 'bas': 'Bàhamàs', 'be': 'Багамскія астравы', 'be-tarask': 'Багамскія астравы', 'bez': 'Hubahama', 'bg': 'Бахамски острови', 'bgn': 'بهاماس', 'blo': 'Bahamaasɩ', 'bm': 'Bahamasi', 'bn': 'বাহামা দ্বীপপুঞ্জ', 'br': 'Bahamas', 'brx': 'बाहामास', 'bs': 'Bahami', 'bs-Cyrl': 'Бахами', 'bs-Latn': 'Bahami', 'ca': 'Bahames', 'ca-ES-valencia': 'Bahames', 'ccp': '𑄝𑄦𑄟 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Багаман гӀайренаш', 'ceb': 'Bahamas', 'cgg': 'Bahama', 'chr': 'ᎾᏍᎩ ᏆᎭᎹᏍ', 'ckb': 'بەھاما', 'cs': 'Bahamy', 'cv': 'Пахам утравӗсем', 'cy': 'Y Bahamas', 'da': 'Bahamas', 'dav': 'Bahama', 'de': 'Bahamas', 'dje': 'Bahamas', 'dsb': 'Bahamy', 'dyo': 'Bahama', 'dz': 'བྷ་ཧ་མས྄', 'ebu': 'Bahama', 'ee': 'Bahamas nutome', 'el': 'Μπαχάμες', 'el-polyton': 'Μπαχάμες', 'en': 'Bahamas', 'en-Dsrt': '𐐒𐐲𐐸𐐪𐑋𐐲𐑅', 'eo': 'Bahamoj', 'es': 'Bahamas', 'et': 'Bahama', 'eu': 'Bahamak', 'ewo': 'Bahámas', 'fa': 'باهاما', 'ff': 'Bahamaas', 'ff-Adlm': '𞤄𞤢𞤸𞤢𞤥𞤢𞥄𞤧', 'ff-Latn': 'Bahamaas', 'fi': 'Bahama', 'fil': 'Bahamas', 'fo': 'Bahamaoyggjar', 'fr': 'Bahamas', 'fur': 'Bahamas', 'fy': 'Bahama’s', 'ga': 'na Bahámaí', 'gd': 'Na h-Eileanan Bhathama', 'gl': 'Bahamas', 'gsw': 'Bahaamas', 'gu': 'બહામાસ', 'guz': 'Bahama', 'ha': 'Bahamas', 'he': 'איי בהאמה', 'hi': 'बहामास', 'hi-Latn': 'Bahamas', 'hr': 'Bahami', 'hsb': 'Bahamy', 'ht': 'Bahamas', 'hu': 'Bahama-szigetek', 'hy': 'Բահամյան կղզիներ', 'ia': 'Bahamas', 'id': 'Bahama', 'ig': 'Bahamas', 'is': 'Bahamaeyjar', 'it': 'Bahamas', 'ja': 'バハマ', 'jmc': 'Bahama', 'jv': 'Bahama', 'ka': 'ბაჰამის კუნძულები', 'kaa': 'Багама атаўлары', 'kaa-Cyrl': 'Багама атаўлары', 'kab': 'Bahamas', 'kam': 'Bahama', 'kde': 'Bahama', 'kea': 'Baamas', 'kgp': 'Mahámỹ', 'khq': 'Bahamas', 'ki': 'Bahama', 'kk': 'Багам аралдары', 'kk-Cyrl': 'Багам аралдары', 'kln': 'Emetab Bahamas', 'km': 'បាហាម៉ា', 'kn': 'ಬಹಾಮಾಸ್', 'ko': '바하마', 'kok': 'बहामास', 'kok-Deva': 'बहामास', 'ks': 'بَہامَس', 'ks-Arab': 'بَہامَس', 'ksb': 'Bahama', 'ksf': 'baamás', 'ksh': 'de Bahamas', 'ku': 'Bahama', 'kxv': 'bāhāmās', 'kxv-Deva': 'बहामास', 'kxv-Latn': 'bāhāmās', 'kxv-Orya': 'ବାହାମାସ୍', 'kxv-Telu': 'బహామాస్', 'ky': 'Багама аралдары', 'lag': 'Baháama', 'lb': 'Bahamas', 'lg': 'Bahamasi', 'lld': 'Bahamas', 'ln': 'Bahamasɛ', 'lo': 'ບາຮາມາສ', 'lt': 'Bahamos', 'lu': 'Bahamase', 'luo': 'Bahamas', 'luy': 'Bahama', 'lv': 'Bahamu salas', 'mai': 'बहामास', 'mas': 'Bahama', 'mer': 'Bahamasi', 'mfe': 'Bahamas', 'mg': 'Bahamas', 'mi': 'Pahama', 'mk': 'Бахами', 'ml': 'ബഹാമാസ്', 'mn': 'Багамын арлууд', 'mr': 'बहामाज', 'ms': 'Bahamas', 'mt': 'il-Bahamas', 'mua': 'bahamas', 'my': 'ဘဟားမား', 'mzn': 'باهاما', 'naq': 'Bahamas', 'nb': 'Bahamas', 'nd': 'Bhahamas', 'ne': 'बहामास', 'nl': 'Bahama’s', 'nmg': 'Bahamas', 'nn': 'Bahamas', 'no': 'Bahamas', 'nqo': 'ߓߤߊߡߊߛ', 'nus': 'Bämuɔth', 'nyn': 'Bahama', 'om': 'Bahaamas', 'or': 'ବାହାମାସ୍', 'pa': 'ਬਹਾਮਾਸ', 'pa-Guru': 'ਬਹਾਮਾਸ', 'pcm': 'Bahámas', 'pl': 'Bahamy', 'ps': 'باهماس', 'pt': 'Bahamas', 'qu': 'Bahamas', 'rm': 'Bahamas', 'rn': 'Bahamasi', 'ro': 'Bahamas', 'rof': 'Bahamasi', 'ru': 'Багамы', 'rwk': 'Bahama', 'saq': 'Bahama', 'sbp': 'Bahama', 'sc': 'Bahamas', 'sd': 'باهاماس', 'sd-Arab': 'باهاماس', 'se': 'Bahamas', 'seh': 'Bahamas', 'ses': 'Bahamas', 'sg': 'Bahâmasa', 'shi': 'ⴱⴰⵀⴰⵎⴰⵙ', 'shi-Latn': 'bahamas', 'shi-Tfng': 'ⴱⴰⵀⴰⵎⴰⵙ', 'si': 'බහමාස්', 'sk': 'Bahamy', 'sl': 'Bahami', 'smn': 'Bahama', 'sn': 'Bahama', 'so': 'Bahaamas', 'sq': 'Bahama', 'sr': 'Бахами', 'sr-Cyrl': 'Бахами', 'sr-Latn': 'Bahami', 'sv': 'Bahamas', 'sw': 'Bahama', 'syr': 'ܒܗܐܡܣ', 'ta': 'பஹாமாஸ்', 'te': 'బహామాస్', 'teo': 'Bahama', 'tg': 'Багам', 'th': 'บาฮามาส', 'ti': 'ባሃማስ', 'tk': 'Bagama adalary', 'to': 'Pahama', 'tr': 'Bahamalar', 'tt': 'Багам утраулары', 'twq': 'Bahamas', 'tzm': 'Bahamas', 'ug': 'باھاما', 'uk': 'Багамські Острови', 'ur': 'بہاماس', 'uz': 'Bagama orollari', 'uz-Cyrl': 'Багама ороллари', 'uz-Latn': 'Bagama orollari', 'vai': 'ꕑꕌꕮꔻ', 'vai-Latn': 'Bahámasi', 'vai-Vaii': 'ꕑꕌꕮꔻ', 'vec': 'Bahamas', 'vi': 'Bahamas', 'vun': 'Bahama', 'wae': 'Bahamas', 'wo': 'Bahamas', 'xh': 'EBahamas', 'xnr': 'बहामास', 'xog': 'Bahamasi', 'yav': 'Pahámas', 'yi': 'באַהאַמאַס', 'yo': 'Bàhámásì', 'yrl': 'Bayama', 'yue': '巴哈馬', 'yue-Hans': '巴哈马', 'yue-Hant': '巴哈馬', 'zgh': 'ⴱⴰⵀⴰⵎⴰⵙ', 'zh': '巴哈马', 'zh-Hans': '巴哈马', 'zh-Hant': '巴哈馬', 'zu': 'i-Bahamas'},
    'BT': {'af': 'Bhoetan', 'agq': 'Mbutàn', 'ak': 'Butan', 'am': 'ቡህታን', 'ar': 'بوتان', 'as': 'ভুটান', 'asa': 'Butani', 'ast': 'Bután', 'az': 'Butan', 'az-Cyrl': 'Бутан', 'az-Latn': 'Butan', 'bal-Latn': 'Buthán', 'bas': 'Bùtân', 'be': 'Бутан', 'be-tarask': 'Бутан', 'bez': 'Hubutani', 'bg': 'Бутан', 'bgn': 'بوتان', 'blo': 'Butan', 'bm': 'Butaŋ', 'bn': 'ভুটান', 'br': 'Bhoutan', 'brx': 'भूटान', 'bs': 'Butan', 'bs-Cyrl': 'Бутан', 'bs-Latn': 'Butan', 'ca': 'Bhutan', 'ca-ES-valencia': 'Bhutan', 'ccp': '𑄞𑄪𑄑𑄚𑄴', 'ce': 'Бутан', 'ceb': 'Bhutan', 'cgg': 'Butani', 'chr': 'ᏊᏔᏂ', 'ckb': 'بووتان', 'cs': 'Bhútán', 'cv': 'Бутан', 'cy': 'Bhutan', 'da': 'Bhutan', 'dav': 'Butani', 'de': 'Bhutan', 'dje': 'Buutaŋ', 'dsb': 'Bhutan', 'dyo': 'Butan', 'dz': 'འབྲུག', 'ebu': 'Butani', 'ee': 'Bhutan nutome', 'el': 'Μπουτάν', 'el-polyton': 'Μπουτάν', 'en': 'Bhutan', 'en-Dsrt': '𐐒𐐭𐐻𐐪𐑌', 'eo': 'Butano', 'es': 'Bután', 'et': 'Bhutan', 'eu': 'Bhutan', 'ewo': 'Butáŋ', 'fa': 'بوتان', 'ff': 'Butaan', 'ff-Adlm': '𞤄𞤵𞥅𞤼𞤢𞥄𞤲', 'ff-Latn': 'Butaan', 'fi': 'Bhutan', 'fil': 'Bhutan', 'fo': 'Butan', 'fr': 'Bhoutan', 'fur': 'Bhutan', 'fy': 'Bhutan', 'ga': 'an Bhútáin', 'gd': 'Butàn', 'gl': 'Bután', 'gsw': 'Bhutan', 'gu': 'ભૂટાન', 'guz': 'Butani', 'ha': 'Butan', 'he': 'בהוטן', 'hi': 'भूटान', 'hi-Latn': 'Bhutan', 'hr': 'Butan', 'hsb': 'Bhutan', 'ht': 'Bhoutan', 'hu': 'Bhután', 'hy': 'Բութան', 'ia': 'Bhutan', 'id': 'Bhutan', 'ig': 'Bhutan', 'is': 'Bútan', 'it': 'Bhutan', 'ja': 'ブータン', 'jmc': 'Butani', 'jv': 'Bhutan', 'ka': 'ბუტანი', 'kaa': 'Бутан', 'kaa-Cyrl': 'Бутан', 'kab': 'Bhutan', 'kam': 'Butani', 'kde': 'Butani', 'kea': 'Butan', 'kgp': 'Mutỹv', 'khq': 'Buutaŋ', 'ki': 'Butani', 'kk': 'Бутан', 'kk-Cyrl': 'Бутан', 'kln': 'Emetab Bhutan', 'km': 'ប៊ូតង់', 'kn': 'ಭೂತಾನ್', 'ko': '부탄', 'kok': 'भूतान', 'kok-Deva': 'भूतान', 'ks': 'بوٗٹان', 'ks-Arab': 'بوٗٹان', 'ksb': 'Butani', 'ksf': 'bután', 'ksh': 'Buttaan', 'ku': 'Bûtan', 'kxv': 'buṭān', 'kxv-Deva': 'बुटान', 'kxv-Latn': 'buṭān', 'kxv-Orya': 'ବୁଟାନ', 'kxv-Telu': 'బుటాన్', 'ky': 'Бутан', 'lag': 'Butáani', 'lb': 'Bhutan', 'lg': 'Butaani', 'lld': 'Bhutan', 'ln': 'Butáni', 'lo': 'ພູຖານ', 'lt': 'Butanas', 'lu': 'Butani', 'luo': 'Bhutan', 'luy': 'Butani', 'lv': 'Butāna', 'mai': 'भूटान', 'mas': 'Butani', 'mer': 'Butani', 'mfe': 'Boutan', 'mg': 'Bhotana', 'mi': 'Pūtana', 'mk': 'Бутан', 'ml': 'ഭൂട്ടാൻ', 'mn': 'Бутан', 'mr': 'भूतान', 'ms': 'Bhutan', 'mt': 'il-Bhutan', 'mua': 'butaŋ', 'my': 'ဘူတန်', 'mzn': 'بوتان', 'naq': 'Bhutans', 'nb': 'Bhutan', 'nd': 'Bhutani', 'ne': 'भुटान', 'nl': 'Bhutan', 'nmg': 'Butaŋ', 'nn': 'Bhutan', 'no': 'Bhutan', 'nqo': 'ߓߎߕߊ߲߫', 'nus': 'Buta̱n', 'nyn': 'Butani', 'om': 'Bihuutan', 'or': 'ଭୁଟାନ', 'pa': 'ਭੂਟਾਨ', 'pa-Guru': 'ਭੂਟਾਨ', 'pcm': 'Butan', 'pl': 'Bhutan', 'ps': 'بهوټان', 'pt': 'Butão', 'qu': 'Bután', 'rm': 'Bhutan', 'rn': 'Butani', 'ro': 'Bhutan', 'rof': 'Butani', 'ru': 'Бутан', 'rwk': 'Butani', 'saq': 'Butani', 'sbp': 'Butani', 'sc': 'Bhutàn', 'sd': 'ڀوٽان', 'sd-Arab': 'ڀوٽان', 'se': 'Bhutan', 'seh': 'Butão', 'ses': 'Buutaŋ', 'sg': 'Butäan', 'shi': 'ⴱⵀⵓⵜⴰⵏ', 'shi-Latn': 'bhutan', 'shi-Tfng': 'ⴱⵀⵓⵜⴰⵏ', 'si': 'භූතානය', 'sk': 'Bhután', 'sl': 'Butan', 'smn': 'Bhutan', 'sn': 'Bhutani', 'so': 'Buutan', 'sq': 'Butan', 'sr': 'Бутан', 'sr-Cyrl': 'Бутан', 'sr-Latn': 'Butan', 'sv': 'Bhutan', 'sw': 'Bhutan', 'syr': 'ܒܘܬܐܢ', 'ta': 'பூடான்', 'te': 'భూటాన్', 'teo': 'Butani', 'tg': 'Бутон', 'th': 'ภูฏาน', 'ti': 'ቡታን', 'tk': 'Butan', 'to': 'Pūtani', 'tr': 'Butan', 'tt': 'Бутан', 'twq': 'Buutaŋ', 'tzm': 'Buṭan', 'ug': 'بۇتان', 'uk': 'Бутан', 'ur': 'بھوٹان', 'uz': 'Butan', 'uz-Cyrl': 'Бутан', 'uz-Latn': 'Butan', 'vai': 'ꖜꕚꘋ', 'vai-Latn': 'Bhutaŋ', 'vai-Vaii': 'ꖜꕚꘋ', 'vec': 'Butan', 'vi': 'Bhutan', 'vun': 'Butani', 'wae': 'Bhutan', 'wo': 'Butaŋ', 'xh': 'EBhutan', 'xnr': 'भूटान', 'xog': 'Butaani', 'yav': 'Putaŋ', 'yi': 'בהוטאַן', 'yo': 'Bútánì', 'yrl': 'Butãu', 'yue': '不丹', 'yue-Hans': '不丹', 'yue-Hant': '不丹', 'zgh': 'ⴱⵀⵓⵜⴰⵏ', 'zh': '不丹', 'zh-Hans': '不丹', 'zh-Hant': '不丹', 'zu': 'i-Bhutan'},
    'BU': {'en': 'Burma'},
    'BV': {'af': 'Bouvet-eiland', 'ak': 'Bouvet Island', 'am': 'ቡቬት ደሴት', 'ar': 'جزيرة بوفيه', 'as': 'বুভে দ্বীপ', 'ast': 'Islla Bouvet', 'az': 'Buve adası', 'az-Cyrl': 'Буве адасы', 'az-Latn': 'Buve adası', 'bal-Latn': 'Bówet Islánd', 'be': 'Востраў Бувэ', 'be-tarask': 'Востраў Бувэ', 'bg': 'остров Буве', 'blo': 'Bufee kaAtukǝltǝna', 'bn': 'বোভেট দ্বীপ', 'br': 'Enez Bouvet', 'brx': 'बूभेट द्वीप', 'bs': 'Ostrvo Buve', 'bs-Cyrl': 'Острво Буве', 'bs-Latn': 'Ostrvo Buve', 'ca': 'Illa Bouvet', 'ca-ES-valencia': 'Illa Bouvet', 'ccp': '𑄝𑄮𑄞𑄬𑄑𑄴 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Бувен гӀайре', 'ceb': 'Bouvet Island', 'chr': 'ᏊᏪ ᎤᎦᏚᏛᎢ', 'ckb': 'دوورگەی بووڤێ', 'cs': 'Bouvetův ostrov', 'cv': 'Буве утравӗ', 'cy': 'Ynys Bouvet', 'da': 'Bouvetøen', 'de': 'Bouvetinsel', 'dsb': 'Bouvetowa kupa', 'dz': 'བོའུ་ཝེཊ་མཚོ་གླིང', 'ee': 'Bouvet ƒudomekpo nutome', 'el': 'Νήσος Μπουβέ', 'el-polyton': 'Νῆσος Μπουβέ', 'en': 'Bouvet Island', 'en-Dsrt': '𐐒𐐭𐑂𐐩 𐐌𐑊𐐲𐑌𐐼', 'es': 'Isla Bouvet', 'et': 'Bouvet’ saar', 'eu': 'Bouvet uhartea', 'fa': 'جزیرهٔ بووه', 'ff-Adlm': '𞤅𞤵𞤪𞤭𞥅𞤪𞤫 𞤄𞤵𞥅𞤾𞤫𞥅', 'fi': 'Bouvet’nsaari', 'fil': 'Bouvet Island', 'fo': 'Bouvetoyggj', 'fr': 'Île Bouvet', 'fur': 'Isule Bouvet', 'fy': 'Bouveteilân', 'ga': 'Oileán Bouvet', 'gd': 'Eilean Bouvet', 'gl': 'Illa Bouvet', 'gsw': 'Bouvet-Insle', 'gu': 'બૌવેત આઇલેન્ડ', 'ha': 'Tsibirin Bouvet', 'he': 'האי בובה', 'hi': 'बोवेत द्वीप', 'hi-Latn': 'Bouvet Island', 'hr': 'Otok Bouvet', 'hsb': 'Bouvetowa kupa', 'ht': 'Île Bouvet', 'hu': 'Bouvet-sziget', 'hy': 'Բուվե կղզի', 'ia': 'Insula de Bouvet', 'id': 'Pulau Bouvet', 'ig': 'Bouvet Island', 'is': 'Bouveteyja', 'it': 'Isola Bouvet', 'ja': 'ブーベ島', 'jv': 'Pulo Bovèt', 'ka': 'ბუვე', 'kaa': 'Буве атаўы', 'kaa-Cyrl': 'Буве атаўы', 'kea': 'Ilha Buvê', 'kgp': 'Muve Goj-vẽso', 'kk': 'Буве аралы', 'kk-Cyrl': 'Буве аралы', 'km': 'កោះ\u200bប៊ូវ៉េត', 'kn': 'ಬೋವೆಟ್ ದ್ವೀಪ', 'ko': '부베섬', 'kok': 'बोवट आयलँड', 'kok-Deva': 'बोवट आयलँड', 'ks': 'بووَٹ جٔزیٖرٕ', 'ks-Arab': 'بووَٹ جٔزیٖرٕ', 'ksh': 'de Buvee-Ensel', 'ku': 'Girava Bouvetê', 'ky': 'Буве аралы', 'lb': 'Bouvetinsel', 'lld': 'Isola Bouvet', 'lo': 'ເກາະບູເວດ', 'lt': 'Buvė Sala', 'lv': 'Buvē sala', 'mai': 'बोबेत द्वीप', 'mi': 'Motu Pūwei', 'mk': 'Остров Буве', 'ml': 'ബൗവെട്ട് ദ്വീപ്', 'mn': 'Буве арал', 'mr': 'बोउवेट बेट', 'ms': 'Pulau Bouvet', 'mt': 'Gżira Bouvet', 'my': 'ဘူဗက်ကျွန်း', 'mzn': 'بووت جزیره', 'nb': 'Bouvetøya', 'ne': 'बुभेट टापु', 'nl': 'Bouveteiland', 'nn': 'Bouvetøya', 'no': 'Bouvetøya', 'nqo': 'ߓߎߝ߭ߋ߫ ߕߌ߲', 'om': 'Odola Bowuvet', 'or': 'ବୌଭେଟ୍\u200c ଦ୍ୱୀପ', 'pa': 'ਬੌਵੇਟ ਟਾਪੂ', 'pa-Guru': 'ਬੌਵੇਟ ਟਾਪੂ', 'pcm': 'Buvẹ́ Aíland', 'pl': 'Wyspa Bouveta', 'ps': 'بوویټ ټاپو', 'pt': 'Ilha Bouvet', 'qu': 'Isla Bouvet', 'rm': 'Insla Bouvet', 'ro': 'Insula Bouvet', 'ru': 'о-в Буве', 'sc': 'Ìsula Bouvet', 'sd': 'بووٽ ٻيٽ', 'sd-Arab': 'بووٽ ٻيٽ', 'se': 'Bouvet-sullot', 'si': 'බුවට් දුපත්', 'sk': 'Bouvetov ostrov', 'sl': 'Bouvetov otok', 'smn': 'Bouvetsuálui', 'so': 'Buufet Island', 'sq': 'Ishulli Bove', 'sr': 'Острво Буве', 'sr-Cyrl': 'Острво Буве', 'sr-Latn': 'Ostrvo Buve', 'sv': 'Bouvetön', 'sw': 'Kisiwa cha Bouvet', 'syr': 'ܓܙܪܬܐ ܕܒܘܒܝܬ', 'ta': 'பொவேட் தீவு', 'te': 'బువై దీవి', 'tg': 'Ҷазираи Буве', 'th': 'เกาะบูเว', 'ti': 'ደሴት ቡቨት', 'tk': 'Buwe adasy', 'to': 'Motu Puveti', 'tr': 'Bouvet Adası', 'tt': 'Буве утравы', 'ug': 'بوۋېت ئارىلى', 'uk': 'Острів Буве', 'ur': 'بؤویٹ آئلینڈ', 'uz': 'Buve oroli', 'uz-Cyrl': 'Буве ороли', 'uz-Latn': 'Buve oroli', 'vai': 'ꖜꔍꔳ ꔳꘋꗣ', 'vai-Vaii': 'ꖜꔍꔳ ꔳꘋꗣ', 'vec': 'Ìzola Buvet', 'vi': 'Đảo Bouvet', 'wae': 'Bouvetinsla', 'wo': 'Dunu Buwet', 'xh': 'EBouvet Island', 'xnr': 'बोवेत द्वीप', 'yo': 'Erékùsù Bouvet', 'yrl': 'Kapuãma Buweti', 'yue': '布威島', 'yue-Hans': '布威岛', 'yue-Hant': '布威島', 'zh': '布韦岛', 'zh-Hans': '布韦岛', 'zh-Hant': '布威島', 'zu': 'i-Bouvet Island'},
    'BW': {'af': 'Botswana', 'agq': 'Bòtɨ̀swǎnà', 'ak': 'Bɔtswana', 'am': 'ቦትስዋና', 'ar': 'بوتسوانا', 'as': 'ব’টচোৱানা', 'asa': 'Botthwana', 'ast': 'Botsuana', 'az': 'Botsvana', 'az-Cyrl': 'Ботсвана', 'az-Latn': 'Botsvana', 'bal-Latn': 'Bostwáná', 'bas': 'Bòdsùanà', 'be': 'Батсвана', 'be-tarask': 'Батсвана', 'bez': 'Hubotiswana', 'bg': 'Ботсвана', 'blo': 'Bɔsʊwanaa', 'bm': 'Bɔtisiwana', 'bn': 'বতসোয়ানা', 'br': 'Botswana', 'brx': 'बत्स्वाना', 'bs': 'Bocvana', 'bs-Cyrl': 'Боцвана', 'bs-Latn': 'Bocvana', 'ca': 'Botswana', 'ca-ES-valencia': 'Botswana', 'ccp': '𑄝𑄧𑄖𑄴𑄥𑄮𑄠𑄚', 'ce': 'Ботсвана', 'ceb': 'Botswana', 'cgg': 'Botswana', 'chr': 'ᏆᏣᏩᎾ', 'ckb': 'بۆتسوانا', 'cs': 'Botswana', 'cv': 'Ботсвана', 'cy': 'Botswana', 'da': 'Botswana', 'dav': 'Botswana', 'de': 'Botsuana', 'dje': 'Botswaana', 'dsb': 'Botswana', 'dyo': 'Boswana', 'dz': 'བྷོཙ་ཝ་ན', 'ebu': 'Botswana', 'ee': 'Botswana nutome', 'el': 'Μποτσουάνα', 'el-polyton': 'Μποτσουάνα', 'en': 'Botswana', 'en-Dsrt': '𐐒𐐪𐐻𐑅𐐶𐐪𐑌𐐲', 'eo': 'Bocvano', 'es': 'Botsuana', 'et': 'Botswana', 'eu': 'Botswana', 'ewo': 'Botswaná', 'fa': 'بوتسوانا', 'ff': 'Botswaana', 'ff-Adlm': '𞤄𞤮𞤼𞤧𞤵𞤱𞤢𞥄𞤲𞤢', 'ff-Latn': 'Botswaana', 'fi': 'Botswana', 'fil': 'Botswana', 'fo': 'Botsvana', 'fr': 'Botswana', 'fur': 'Botswana', 'fy': 'Botswana', 'ga': 'an Bhotsuáin', 'gd': 'Botsuana', 'gl': 'Botswana', 'gsw': 'Botswana', 'gu': 'બોત્સ્વાના', 'guz': 'Botswana', 'ha': 'Baswana', 'he': 'בוטסואנה', 'hi': 'बोत्स्वाना', 'hi-Latn': 'Botswana', 'hr': 'Bocvana', 'hsb': 'Botswana', 'ht': 'Botswana', 'hu': 'Botswana', 'hy': 'Բոտսվանա', 'ia': 'Botswana', 'id': 'Botswana', 'ig': 'Botswana', 'is': 'Botsvana', 'it': 'Botswana', 'ja': 'ボツワナ', 'jgo': 'Mbɔtswána', 'jmc': 'Botswana', 'jv': 'Botswana', 'ka': 'ბოტსვანა', 'kaa': 'Ботсвана', 'kaa-Cyrl': 'Ботсвана', 'kab': 'Bustwana', 'kam': 'Botswana', 'kde': 'Botswana', 'kea': 'Botsuana', 'kgp': 'Monsuvỹnỹ', 'khq': 'Botswaana', 'ki': 'Botswana', 'kk': 'Ботсвана', 'kk-Cyrl': 'Ботсвана', 'kln': 'Emetab Botswana', 'km': 'បុតស្វាណា', 'kn': 'ಬೋಟ್ಸ್\u200cವಾನಾ', 'ko': '보츠와나', 'kok': 'बोत्सवाना', 'kok-Deva': 'बोत्सवाना', 'ks': 'بوتَسوانا', 'ks-Arab': 'بوتَسوانا', 'ksb': 'Botswana', 'ksf': 'botswana', 'ksh': 'Bozwaana', 'ku': 'Botswana', 'kxv': 'boṭsvana', 'kxv-Deva': 'बोत्स्वाना', 'kxv-Latn': 'boṭsvana', 'kxv-Orya': 'ବସ୍ଟୱାନା', 'kxv-Telu': 'బోట్స్వానా', 'ky': 'Ботсвана', 'lag': 'Botiswáana', 'lb': 'Botsuana', 'lg': 'Botiswana', 'lld': 'Botswana', 'ln': 'Botswana', 'lo': 'ບອດສະວານາ', 'lt': 'Botsvana', 'lu': 'Mbotswana', 'luo': 'Botswana', 'luy': 'Botswana', 'lv': 'Botsvāna', 'mai': 'बोत्सबाना', 'mas': 'Botswana', 'mer': 'Botswana', 'mfe': 'Botswana', 'mg': 'Botsoana', 'mi': 'Poriwana', 'mk': 'Боцвана', 'ml': 'ബോട്സ്വാന', 'mn': 'Ботсван', 'mr': 'बोट्सवाना', 'ms': 'Botswana', 'mt': 'il-Botswana', 'mua': 'botswana', 'my': 'ဘော့ဆွာနာ', 'mzn': 'بوتساوانا', 'naq': 'Botswanab', 'nb': 'Botswana', 'nd': 'Botswana', 'ne': 'बोट्स्वाना', 'nl': 'Botswana', 'nmg': 'Botswana', 'nn': 'Botswana', 'no': 'Botswana', 'nqo': 'ߓߐߛߎߥߣߊ߫', 'nus': 'Bothiwaana', 'nyn': 'Botswana', 'om': 'Botosowaanaa', 'or': 'ବୋଟସ୍ୱାନା', 'pa': 'ਬੋਤਸਵਾਨਾ', 'pa-Guru': 'ਬੋਤਸਵਾਨਾ', 'pcm': 'Botswána', 'pl': 'Botswana', 'ps': 'بوتسوانه', 'pt': 'Botsuana', 'qu': 'Botsuana', 'rm': 'Botswana', 'rn': 'Botswana', 'ro': 'Botswana', 'rof': 'Botswana', 'ru': 'Ботсвана', 'rwk': 'Botswana', 'saq': 'Botswana', 'sbp': 'Botiswana', 'sc': 'Botswana', 'sd': 'بوٽسوانا', 'sd-Arab': 'بوٽسوانا', 'se': 'Botswana', 'seh': 'Botsuana', 'ses': 'Botswaana', 'sg': 'Botswana', 'shi': 'ⴱⵓⵜⵙⵡⴰⵏⴰ', 'shi-Latn': 'butswana', 'shi-Tfng': 'ⴱⵓⵜⵙⵡⴰⵏⴰ', 'si': 'බොට්ස්වානා', 'sk': 'Botswana', 'sl': 'Bocvana', 'smn': 'Botswana', 'sn': 'Botswana', 'so': 'Botuswaana', 'sq': 'Botsvanë', 'sr': 'Боцвана', 'sr-Cyrl': 'Боцвана', 'sr-Latn': 'Bocvana', 'sv': 'Botswana', 'sw': 'Botswana', 'syr': 'ܒܘܛܣܘܐܢܐ', 'ta': 'போட்ஸ்வானா', 'te': 'బోట్స్వానా', 'teo': 'Botswana', 'tg': 'Ботсвана', 'th': 'บอตสวานา', 'ti': 'ቦትስዋና', 'tk': 'Botswana', 'tn': 'Botswana', 'to': 'Potisiuana', 'tr': 'Botsvana', 'tt': 'Ботсвана', 'twq': 'Botswaana', 'tzm': 'Butswana', 'ug': 'بوتسۋانا', 'uk': 'Ботсвана', 'ur': 'بوتسوانا', 'uz': 'Botsvana', 'uz-Cyrl': 'Ботсванна', 'uz-Latn': 'Botsvana', 'vai': 'ꕷꖬꕎꕯ', 'vai-Latn': 'Bhosuwana', 'vai-Vaii': 'ꕷꖬꕎꕯ', 'vec': 'Botzwana', 'vi': 'Botswana', 'vun': 'Botswana', 'wae': 'Botswana', 'wo': 'Botswana', 'xh': 'EBotswana', 'xnr': 'बोत्स्वाना', 'xog': 'Botiswana', 'yav': 'Posuána', 'yi': 'באצוואַנע', 'yo': 'Bọ̀tìsúwánà', 'yrl': 'Butisuwana', 'yue': '波札那', 'yue-Hans': '波札那', 'yue-Hant': '波札那', 'zgh': 'ⴱⵓⵜⵙⵡⴰⵏⴰ', 'zh': '博茨瓦纳', 'zh-Hans': '博茨瓦纳', 'zh-Hant': '波札那', 'zu': 'iBotswana'},
    'BY': {'af': 'Belarus', 'agq': 'Bɛlàlûs', 'ak': 'Bɛlarus', 'am': 'ቤላሩስ', 'ar': 'بيلاروس', 'as': 'বেলাৰুছ', 'asa': 'Belaruthi', 'ast': 'Bielorrusia', 'az': 'Belarus', 'az-Cyrl': 'Беларус', 'az-Latn': 'Belarus', 'bal-Latn': 'Bélárus', 'bas': 'Bèlarùs', 'be': 'Беларусь', 'be-tarask': 'Беларусь', 'bez': 'Hubelalusi', 'bg': 'Беларус', 'blo': 'Belaruus', 'bm': 'Belarusi', 'bn': 'বেলারুশ', 'br': 'Belarus', 'brx': 'बेलारूस', 'bs': 'Bjelorusija', 'bs-Cyrl': 'Бјелорусија', 'bs-Latn': 'Bjelorusija', 'ca': 'Belarús', 'ca-ES-valencia': 'Belarús', 'ccp': '𑄝𑄬𑄣𑄢𑄪𑄌𑄴', 'ce': 'Белорусси', 'ceb': 'Belarus', 'cgg': 'Bararusi', 'chr': 'ᏇᎳᎷᏍ', 'ckb': 'بیلاڕووس', 'cs': 'Bělorusko', 'cv': 'Беларуҫ', 'cy': 'Belarws', 'da': 'Belarus', 'dav': 'Belarusi', 'de': 'Belarus', 'dje': 'Biloriši', 'dsb': 'Běłoruska', 'dyo': 'Belarus', 'dz': 'བེལ་ཨ་རུ་སུ', 'ebu': 'Belarusi', 'ee': 'Belarus nutome', 'el': 'Λευκορωσία', 'el-polyton': 'Λευκορωσία', 'en': 'Belarus', 'en-Dsrt': '𐐒𐐯𐑊𐐲𐑉𐐭𐑅', 'eo': 'Belorusujo', 'es': 'Bielorrusia', 'et': 'Valgevene', 'eu': 'Bielorrusia', 'ewo': 'Bəlarús', 'fa': 'بلاروس', 'ff': 'Belaruus', 'ff-Adlm': '𞤄𞤫𞤤𞤢𞤪𞤵𞥅𞤧', 'ff-Latn': 'Belaruus', 'fi': 'Valko-Venäjä', 'fil': 'Belarus', 'fo': 'Hvítarussland', 'fr': 'Biélorussie', 'fur': 'Bielorussie', 'fy': 'Wit-Ruslân', 'ga': 'an Bhealarúis', 'gd': 'A’ Bhealaruis', 'gl': 'Belarús', 'gsw': 'Wiissrussland', 'gu': 'બેલારુસ', 'guz': 'Belarusi', 'ha': 'Belarus', 'he': 'בלארוס', 'hi': 'बेलारूस', 'hi-Latn': 'Belarus', 'hr': 'Bjelorusija', 'hsb': 'Běłoruska', 'ht': 'Biélorussie', 'hu': 'Belarusz', 'hy': 'Բելառուս', 'ia': 'Bielorussia', 'id': 'Belarus', 'ig': 'Belarus', 'is': 'Hvíta-Rússland', 'it': 'Bielorussia', 'ja': 'ベラルーシ', 'jmc': 'Belarusi', 'jv': 'Bélarus', 'ka': 'ბელარუსი', 'kaa': 'Беларусь', 'kaa-Cyrl': 'Беларусь', 'kab': 'Bilarus', 'kam': 'Belarusi', 'kde': 'Belalusi', 'kea': 'Belarus', 'kgp': 'Miero-Husija', 'khq': 'Biloriši', 'ki': 'Belarusi', 'kk': 'Беларусь', 'kk-Cyrl': 'Беларусь', 'kln': 'Emetab Belarus', 'km': 'បេឡារុស', 'kn': 'ಬೆಲಾರಸ್', 'ko': '벨라루스', 'kok': 'बेलारूस', 'kok-Deva': 'बेलारूस', 'ks': 'بیلاروٗس', 'ks-Arab': 'بیلاروٗس', 'ksb': 'Belalusi', 'ksf': 'bɛlaris', 'ksh': 'Wießrußland', 'ku': 'Belarûs', 'kxv': 'belarūs', 'kxv-Deva': 'बेलारूस', 'kxv-Latn': 'belarūs', 'kxv-Orya': 'ବେଲାରୂସ୍', 'kxv-Telu': 'బెలారస్', 'ky': 'Беларусь', 'lag': 'Belarúusi', 'lb': 'Wäissrussland', 'lg': 'Belarusi', 'lld': 'Belaruscia', 'ln': 'Byelorisi', 'lo': 'ເບລາຣຸສ', 'lt': 'Baltarusija', 'lu': 'Byelorisi', 'luo': 'Belarus', 'luy': 'Belarusi', 'lv': 'Baltkrievija', 'mai': 'बेलारूस', 'mas': 'Belarusi', 'mer': 'Belarusi', 'mfe': 'Belaris', 'mg': 'Belarosy', 'mi': 'Pērara', 'mk': 'Белорусија', 'ml': 'ബെലറൂസ്', 'mn': 'Беларусь', 'mr': 'बेलारूस', 'ms': 'Belarus', 'mt': 'il-Belarussja', 'mua': 'belarussiya', 'my': 'ဘီလာရုစ်', 'mzn': 'بلاروس', 'naq': 'Belarus', 'nb': 'Belarus', 'nd': 'Bhelarusi', 'ne': 'बेलारूस', 'nl': 'Belarus', 'nmg': 'Belarus', 'nn': 'Belarus', 'no': 'Belarus', 'nqo': 'ߓߌߟߏߙߌߛ', 'nus': 'Be̱lɛruth', 'nyn': 'Bararusi', 'om': 'Beelaarus', 'or': 'ବେଲାରୁଷ୍', 'pa': 'ਬੇਲਾਰੂਸ', 'pa-Guru': 'ਬੇਲਾਰੂਸ', 'pcm': 'Bẹ́larus', 'pl': 'Białoruś', 'ps': 'بیلاروس', 'pt': 'Bielorrússia', 'qu': 'Belarús', 'rm': 'Bielorussia', 'rn': 'Belausi', 'ro': 'Belarus', 'rof': 'Belarusi', 'ru': 'Беларусь', 'rwk': 'Belarusi', 'saq': 'Belarusi', 'sbp': 'Belalusi', 'sc': 'Bielorùssia', 'sd': 'بیلارس', 'sd-Arab': 'بیلارس', 'se': 'Vilges-Ruošša', 'seh': 'Belarus', 'ses': 'Biloriši', 'sg': 'Belarüsi', 'shi': 'ⴱⵉⵍⴰⵔⵓⵙⵢⴰ', 'shi-Latn': 'bilarusya', 'shi-Tfng': 'ⴱⵉⵍⴰⵔⵓⵙⵢⴰ', 'si': 'බෙලරුස්', 'sk': 'Bielorusko', 'sl': 'Belorusija', 'smn': 'Vielgis-Ruoššâ', 'sn': 'Belarusi', 'so': 'Belarus', 'sq': 'Bjellorusi', 'sr': 'Белорусија', 'sr-Cyrl': 'Белорусија', 'sr-Latn': 'Belorusija', 'sv': 'Belarus', 'sw': 'Belarus', 'syr': 'ܒܠܐܪܘܣ', 'ta': 'பெலாரஸ்', 'te': 'బెలారస్', 'teo': 'Belarusi', 'tg': 'Белорус', 'th': 'เบลารุส', 'ti': 'ቤላሩስ', 'tk': 'Belarus', 'to': 'Pelalusi', 'tr': 'Belarus', 'tt': 'Беларусь', 'twq': 'Biloriši', 'tzm': 'Bilarusya', 'ug': 'بېلارۇسىيە', 'uk': 'Білорусь', 'ur': 'بیلاروس', 'uz': 'Belarus', 'uz-Cyrl': 'Беларус', 'uz-Latn': 'Belarus', 'vai': 'ꗩꕞꖩꔻ', 'vai-Latn': 'Bhɛlarusi', 'vai-Vaii': 'ꗩꕞꖩꔻ', 'vec': 'Belorusia', 'vi': 'Belarus', 'vun': 'Belarusi', 'wae': 'Wísrussland', 'wo': 'Belaris', 'xh': 'EBelarus', 'xnr': 'बेलारूस', 'xog': 'Belarusi', 'yav': 'Pelalús', 'yi': 'בעלאַרוס', 'yo': 'Bélárúsì', 'yrl': 'Bieru-rúsiya', 'yue': '白俄羅斯', 'yue-Hans': '白俄罗斯', 'yue-Hant': '白俄羅斯', 'zgh': 'ⴱⵉⵍⴰⵔⵓⵙⵢⴰ', 'zh': '白俄罗斯', 'zh-Hans': '白俄罗斯', 'zh-Hant': '白俄羅斯', 'zu': 'i-Belarus'},
    'BZ': {'af': 'Belize', 'agq': 'Bɛ̀lezɨ̀', 'ak': 'Beliz', 'am': 'በሊዝ', 'ar': 'بليز', 'as': 'বেলিজ', 'asa': 'Belidhe', 'ast': 'Belize', 'az': 'Beliz', 'az-Cyrl': 'Белиз', 'az-Latn': 'Beliz', 'bal-Latn': 'Belize', 'bas': 'Bèlîs', 'be': 'Беліз', 'be-tarask': 'Беліз', 'bez': 'Hubelize', 'bg': 'Белиз', 'blo': 'Beliis', 'bm': 'Belizi', 'bn': 'বেলিজ', 'br': 'Belize', 'brx': 'बेलिज', 'bs': 'Belize', 'bs-Cyrl': 'Белизе', 'bs-Latn': 'Belize', 'ca': 'Belize', 'ca-ES-valencia': 'Belize', 'ccp': '𑄝𑄬𑄣𑄨𑄎𑄴', 'ce': 'Белиз', 'ceb': 'Belize', 'cgg': 'Berize', 'chr': 'ᏇᎵᏍ', 'ckb': 'بەلیز', 'cs': 'Belize', 'cv': 'Белиз', 'cy': 'Belize', 'da': 'Belize', 'dav': 'Belize', 'de': 'Belize', 'dje': 'Beliizi', 'dsb': 'Belize', 'dyo': 'Beliis', 'dz': 'བྷེ་ལིཛ', 'ebu': 'Belize', 'ee': 'Belize nutome', 'el': 'Μπελίζ', 'el-polyton': 'Μπελίζ', 'en': 'Belize', 'en-Dsrt': '𐐒𐐲𐑊𐐨𐑆', 'eo': 'Belizo', 'es': 'Belice', 'et': 'Belize', 'eu': 'Belize', 'ewo': 'Bəlís', 'fa': 'بلیز', 'ff': 'Beliise', 'ff-Adlm': '𞤄𞤫𞤤𞤭𞥅𞥁', 'ff-Latn': 'Beliise', 'fi': 'Belize', 'fil': 'Belize', 'fo': 'Belis', 'fr': 'Belize', 'fur': 'Belize', 'fy': 'Belize', 'ga': 'an Bheilís', 'gd': 'A’ Bheilìs', 'gl': 'Belize', 'gsw': 'Belize', 'gu': 'બેલીઝ', 'guz': 'Belize', 'ha': 'Beliz', 'he': 'בליז', 'hi': 'बेलीज़', 'hi-Latn': 'Belize', 'hr': 'Belize', 'hsb': 'Belize', 'ht': 'Belize', 'hu': 'Belize', 'hy': 'Բելիզ', 'ia': 'Belize', 'id': 'Belize', 'ig': 'Belize', 'is': 'Belís', 'it': 'Belize', 'ja': 'ベリーズ', 'jmc': 'Belize', 'jv': 'Bélisé', 'ka': 'ბელიზი', 'kaa': 'Белиз', 'kaa-Cyrl': 'Белиз', 'kab': 'Biliz', 'kam': 'Belize', 'kde': 'Belize', 'kea': 'Belizi', 'kgp': 'Merije', 'khq': 'Beliizi', 'ki': 'Belize', 'kk': 'Белиз', 'kk-Cyrl': 'Белиз', 'kln': 'Emetab Belize', 'km': 'បេលី', 'kn': 'ಬೆಲಿಜ್', 'ko': '벨리즈', 'kok': 'बेलिझ', 'kok-Deva': 'बेलिझ', 'ks': 'بیلز', 'ks-Arab': 'بیلز', 'ksb': 'Belize', 'ksf': 'bɛliz', 'ksh': 'Belize', 'ku': 'Belîze', 'kxv': 'belīj', 'kxv-Deva': 'बेलीज़', 'kxv-Latn': 'belīj', 'kxv-Orya': 'ବେଲିଜ୍', 'kxv-Telu': 'బెలిజ్', 'ky': 'Белиз', 'lag': 'Belíise', 'lb': 'Belize', 'lg': 'Belize', 'lld': 'Belize', 'ln': 'Belizɛ', 'lo': 'ເບລີຊ', 'lt': 'Belizas', 'lu': 'Belize', 'luo': 'Belize', 'luy': 'Belize', 'lv': 'Beliza', 'mai': 'बेलीज', 'mas': 'Belize', 'mer': 'Belize', 'mfe': 'Beliz', 'mg': 'Belize', 'mi': 'Perīhi', 'mk': 'Белизе', 'ml': 'ബെലീസ്', 'mn': 'Белизе', 'mr': 'बेलिझे', 'ms': 'Belize', 'mt': 'il-Belize', 'mua': 'beliziya', 'my': 'ဘလိဇ်', 'mzn': 'بلیز', 'naq': 'Belize', 'nb': 'Belize', 'nd': 'Bhelize', 'ne': 'बेलिज', 'nl': 'Belize', 'nmg': 'Bɛliz', 'nn': 'Belize', 'no': 'Belize', 'nqo': 'ߓߋߟߌߖ߭', 'nus': 'Bilidha', 'nyn': 'Berize', 'om': 'Belize', 'or': 'ବେଲିଜ୍', 'pa': 'ਬੇਲੀਜ਼', 'pa-Guru': 'ਬੇਲੀਜ਼', 'pcm': 'Bẹliz', 'pl': 'Belize', 'ps': 'بلیز', 'pt': 'Belize', 'qu': 'Belice', 'rm': 'Belize', 'rn': 'Belize', 'ro': 'Belize', 'rof': 'Belize', 'ru': 'Белиз', 'rwk': 'Belize', 'saq': 'Belize', 'sbp': 'Belise', 'sc': 'Belize', 'sd': 'بيليز', 'sd-Arab': 'بيليز', 'se': 'Belize', 'seh': 'Belize', 'ses': 'Beliizi', 'sg': 'Belîzi', 'shi': 'ⴱⵉⵍⵉⵣ', 'shi-Latn': 'biliz', 'shi-Tfng': 'ⴱⵉⵍⵉⵣ', 'si': 'බෙලීස්', 'sk': 'Belize', 'sl': 'Belize', 'smn': 'Belize', 'sn': 'Belize', 'so': 'Beliis', 'sq': 'Belizë', 'sr': 'Белизе', 'sr-Cyrl': 'Белизе', 'sr-Latn': 'Belize', 'sv': 'Belize', 'sw': 'Belize', 'syr': 'ܒܠܝܙ', 'ta': 'பெலிஸ்', 'te': 'బెలిజ్', 'teo': 'Belize', 'tg': 'Белиз', 'th': 'เบลีซ', 'ti': 'በሊዝ', 'tk': 'Beliz', 'to': 'Pelise', 'tr': 'Belize', 'tt': 'Белиз', 'twq': 'Beliizi', 'tzm': 'Biliz', 'ug': 'بېلىز', 'uk': 'Беліз', 'ur': 'بیلائز', 'uz': 'Beliz', 'uz-Cyrl': 'Белиз', 'uz-Latn': 'Beliz', 'vai': 'ꔆꔷꔘ', 'vai-Latn': 'Bheliz', 'vai-Vaii': 'ꔆꔷꔘ', 'vec': 'Belize', 'vi': 'Belize', 'vun': 'Belize', 'wae': 'Belize', 'wo': 'Belis', 'xh': 'EBelize', 'xnr': 'बेलीज़', 'xog': 'Belize', 'yav': 'Pelíse', 'yi': 'בעליז', 'yo': 'Bèlísẹ̀', 'yrl': 'Belisi', 'yue': '貝里斯', 'yue-Hans': '贝里斯', 'yue-Hant': '貝里斯', 'zgh': 'ⴱⵉⵍⵉⵣ', 'zh': '伯利兹', 'zh-Hans': '伯利兹', 'zh-Hant': '貝里斯', 'zu': 'i-Belize'},
    'Bali': {'ar': 'البالية', 'ast': 'balinés', 'bal-Latn': 'Báléni', 'bg': 'Балийски', 'bn': 'বালীয়', 'br': 'balinek', 'brx': 'बाली', 'bs': 'balijsko pismo', 'bs-Cyrl': 'балијско писмо', 'bs-Latn': 'balijsko pismo', 'ca': 'balinès', 'ca-ES-valencia': 'balinès', 'ccp': '𑄝𑄣𑄩𑄠𑄧', 'cs': 'balijské', 'da': 'balinesisk', 'de': 'Balinesisch', 'el': 'Μπαλινίζ', 'el-polyton': 'Μπαλινίζ', 'en': 'Balinese', 'en-Dsrt': '𐐒𐐪𐑊𐐲𐑌𐐨𐑆', 'es': 'balinés', 'eu': 'baliera', 'fa': 'بالیایی', 'ff-Adlm': '𞤄𞤢𞤤𞤭𞤲𞤭𞥅𞤧', 'fi': 'balilainen', 'fo': 'baliskt', 'fr': 'balinais', 'fur': 'balinês', 'fy': 'Balineesk', 'ga': 'Bailíoch', 'gsw': 'Balinesisch', 'gu': 'બાલીનીઝ', 'he': 'באלינזי', 'hi': 'बाली', 'hi-Latn': 'Baali', 'hr': 'balijsko pismo', 'ht': 'balinais', 'hu': 'Balinéz', 'ig': 'Balinese', 'is': 'balinesíska', 'it': 'balinese', 'ja': 'バリ文字', 'ka': 'ბალიური', 'kgp': 'marinẽj', 'kk': 'балий жазуы', 'kk-Cyrl': 'балий жазуы', 'km': 'បាលី', 'kn': 'ಬಾಲಿನೀಸ್', 'ko': '발리 문자', 'ks': 'بالَنیٖز', 'ks-Arab': 'بالَنیٖز', 'lb': 'Balinesesch', 'lo': 'ບາລີ', 'lt': 'Baliečių', 'lv': 'baliešu', 'mk': 'балиско', 'ml': 'ബാലിനീസ്', 'mn': 'Бали', 'mr': 'बाली', 'nb': 'balinesisk', 'ne': 'बाली', 'nl': 'Balinees', 'nn': 'balinesisk', 'no': 'balinesisk', 'nqo': 'ߓߊߟߌߣߍ', 'or': 'ବାଲିନୀଜ୍', 'pl': 'balijskie', 'pt': 'balinês', 'rm': 'balinais', 'ro': 'balineză', 'ru': 'балийская', 'sc': 'balinesu', 'sk': 'balijský', 'sl': 'balijski', 'so': 'Baliniis', 'sr': 'балијско писмо', 'sr-Cyrl': 'балијско писмо', 'sr-Latn': 'balijsko pismo', 'sv': 'balinesiska', 'ta': 'பாலினீஸ்', 'te': 'బాలినీస్', 'tg': 'Балинесӣ', 'th': 'บาหลี', 'to': 'tohinima fakapali', 'tr': 'Bali Dili', 'ug': 'بالى', 'uk': 'балійський', 'vec': 'balineze', 'vi': 'Chữ Bali', 'xnr': 'बाली', 'yrl': 'barineisi', 'yue': '峇里文', 'yue-Hans': '峇里文', 'yue-Hant': '峇里文', 'zh': '巴厘文', 'zh-Hans': '巴厘文', 'zh-Hant': '峇里文', 'zu': 'isi-Balinese'},
    'Bamu': {'ast': 'bamum', 'bal-Latn': 'Bámum', 'br': 'bamounek', 'bs': 'bamum pismo', 'bs-Cyrl': 'бамум писмо', 'bs-Latn': 'bamum pismo', 'ca': 'bamum', 'ca-ES-valencia': 'bamum', 'cs': 'bamumské', 'da': 'bamum', 'de': 'Bamun', 'en': 'Bamum', 'et': 'bamumi', 'eu': 'bamum', 'ff-Adlm': '𞤄𞤢𞤥𞤵', 'fi': 'bamum', 'fo': 'bamum', 'fy': 'Bamoun', 'ga': 'Bamum', 'gd': 'Bamum', 'he': 'במום', 'hi-Latn': 'Bamum', 'hr': 'bamum pismo', 'id': 'Bamum', 'ig': 'Bamum', 'it': 'bamum', 'ja': 'バムン文字', 'ka': 'ბამუმი', 'kgp': 'mamũm', 'kk': 'бамум жазуы', 'kk-Cyrl': 'бамум жазуы', 'km': 'បាមុន', 'ko': '바뭄 문자', 'lo': 'ບາມູມ', 'lt': 'Bamum', 'mk': 'бамумско', 'mn': 'Баму', 'nb': 'bamum', 'nl': 'Bamoun', 'nn': 'bamum', 'no': 'bamum', 'nqo': 'ߓߊߡߎ߲', 'pl': 'bamun', 'pt': 'bamum', 'ro': 'bamum', 'ru': 'бамум', 'sc': 'bamum', 'so': 'Bamum', 'sv': 'bamunska', 'tg': 'Бамумӣ', 'th': 'บามุม', 'to': 'tohinima fakapamumi', 'tr': 'Bamum', 'ug': 'بامۇم', 'uk': 'бамум', 'vec': 'bamun', 'vi': 'Chữ Bamum', 'yo': 'Bamumu', 'yrl': 'bamũ', 'yue': '巴姆穆文', 'yue-Hans': '巴姆穆文', 'yue-Hant': '巴姆穆文', 'zh': '巴姆穆文', 'zh-Hans': '巴姆穆文', 'zh-Hant': '巴姆穆文', 'zu': 'isi-Bamum'},
    'Bass': {'ast': 'bassa vah', 'bal-Latn': 'Bassa Wah', 'bs': 'bassa vah pismo', 'bs-Cyrl': 'баса вах писмо', 'bs-Latn': 'bassa vah pismo', 'ca': 'bassa vah', 'ca-ES-valencia': 'bassa vah', 'cs': 'bassa vah', 'da': 'bassa', 'de': 'Bassa', 'en': 'Bassa Vah', 'et': 'bassa', 'eu': 'bassa vah', 'ff-Adlm': '𞤄𞤢𞤧𞤢𞥄 𞤜𞤢𞥄', 'fi': 'bassa', 'fo': 'bassa vah', 'fy': 'Bassa Vah', 'ga': 'Bassa Vah', 'gd': 'Bassa Vah', 'he': 'באסה ואה', 'hi-Latn': 'Bassa Vah', 'hr': 'bassa vah pismo', 'id': 'Bassa Vah', 'ig': 'Bassa Vah', 'it': 'Bassa Vah', 'ja': 'バサ文字', 'ka': 'ბასა ვაჰი', 'kk': 'басса жазуы', 'kk-Cyrl': 'басса жазуы', 'km': 'បាសាវ៉ះ', 'ko': '바사바흐 문자', 'lo': 'ບັດຊາ', 'lt': 'Bassa Vah', 'mk': 'баса', 'mn': 'Басс', 'ms': 'Bassa Vah', 'nb': 'bassa vah', 'nl': 'Bassa Vah', 'nn': 'bassa vah', 'no': 'bassa vah', 'nqo': 'ߓߊ߯ߛߊ', 'ro': 'bassa vah', 'ru': 'басса (вах)', 'sc': 'bassa vah', 'so': 'Qoraalka Vah', 'sq': 'basavahisht', 'sv': 'bassaiska vah', 'tg': 'Басса Вахӣ', 'th': 'บัสซา', 'to': 'tohinima fakapasa-vā', 'tr': 'Bassa Vah', 'ug': 'باسسا', 'uk': 'басса', 'vec': 'vah', 'vi': 'Chữ Bassa Vah', 'yue': '巴薩文', 'yue-Hans': '巴萨文', 'yue-Hant': '巴薩文', 'zh': '巴萨文', 'zh-Hans': '巴萨文', 'zh-Hant': '巴薩文', 'zu': 'isi-Bassa Vah'},
    'Batk': {'ar': 'الباتاك', 'ast': 'batak', 'az': 'batak', 'az-Latn': 'batak', 'bal-Latn': 'Batak', 'bg': 'Батакска', 'bn': 'বাটাক', 'br': 'batak', 'brx': 'बटकी', 'bs': 'batak pismo', 'bs-Cyrl': 'батак писмо', 'bs-Latn': 'batak pismo', 'ca': 'batak', 'ca-ES-valencia': 'batak', 'ccp': '𑄖𑄑𑄇𑄴', 'cs': 'batacké', 'da': 'batak', 'de': 'Battakisch', 'el': 'Μπατάκ', 'el-polyton': 'Μπατάκ', 'en': 'Batak', 'en-Dsrt': '𐐒𐐲𐐻𐐪𐐿', 'es': 'batak', 'et': 'bataki', 'eu': 'batak', 'fa': 'باتاکی', 'ff-Adlm': '𞤄𞤢𞤼𞤢𞤳', 'fi': 'batakilainen', 'fo': 'batak', 'fr': 'batak', 'fy': 'Batak', 'ga': 'Batacach', 'gd': 'Batak', 'gsw': 'Battakisch', 'gu': 'બટાક', 'he': 'בטק', 'hi': 'बटकी', 'hi-Latn': 'Batak', 'hr': 'batak pismo', 'ht': 'batak', 'hu': 'Batak', 'id': 'Batak', 'ig': 'Batak', 'is': 'batakíska', 'it': 'batak', 'ja': 'バタク文字', 'ka': 'ბატაკი', 'kgp': 'mataki', 'kk': 'батак жазуы', 'kk-Cyrl': 'батак жазуы', 'km': 'បាតាក', 'kn': 'ಬಾಟಕ್', 'ko': '바타크 문자', 'ks': 'باتَک', 'ks-Arab': 'باتَک', 'lb': 'Battakesch', 'lo': 'ບາຕັກ', 'lt': 'batak', 'mk': 'батачко', 'ml': 'ബട്ടക്', 'mn': 'Батк', 'mr': 'बटाक', 'ms': 'Batak', 'nb': 'batak', 'ne': 'बाटक', 'nl': 'Batak', 'nn': 'batak', 'no': 'batak', 'nqo': 'ߓߕߊߞ', 'or': 'ବାଟାକ୍', 'pl': 'batak', 'pt': 'bataque', 'rm': 'batak', 'ro': 'batak', 'ru': 'батакская', 'sc': 'batak', 'sl': 'bataški', 'so': 'Batak', 'sq': 'batak', 'sr': 'батак писмо', 'sr-Cyrl': 'батак писмо', 'sr-Latn': 'batak pismo', 'sv': 'batak', 'ta': 'பாடாக்', 'te': 'బాటక్', 'tg': 'Батакӣ', 'th': 'บาตัก', 'to': 'tohinima fakapātaki', 'tr': 'Batak', 'ug': 'باتاك', 'uk': 'батак', 'vec': 'batak', 'vi': 'Chữ Batak', 'xnr': 'बटकी', 'yo': 'Bataki', 'yrl': 'bataki', 'yue': '巴塔克文', 'yue-Hans': '巴塔克文', 'yue-Hant': '巴塔克文', 'zh': '巴塔克文', 'zh-Hans': '巴塔克文', 'zh-Hant': '巴塔克文', 'zu': 'isi-Batak'},
    'Beng': {'af': 'Bengaals', 'ak': 'Bangala kasa', 'am': 'ቤንጋሊ', 'ar': 'البنغالية', 'as': 'বাংলা', 'ast': 'bengalín', 'az': 'benqal', 'az-Arab': 'بنگالجا', 'az-Latn': 'benqal', 'bal-Latn': 'Bang', 'be': 'бенгальскае', 'be-tarask': 'бенгальскае', 'bg': 'бенгалска', 'bgn': 'بنگالی', 'bn': 'বাংলা', 'br': 'bengali', 'brx': 'बांला', 'bs': 'bengalsko pismo', 'bs-Cyrl': 'бенгалско писмо', 'bs-Latn': 'bengalsko pismo', 'ca': 'bengalí', 'ca-ES-valencia': 'bengalí', 'ccp': '𑄝𑄁𑄣', 'ce': 'бенгалхойн', 'chr': 'ᏇᏂᎦᎠ', 'ckb': 'بەنگالی', 'cs': 'bengálské', 'cy': 'Bangla', 'da': 'bengali', 'de': 'Bengalisch', 'dsb': 'bengalski', 'dz': 'བངྒ་ལ་ཡིག་གུ', 'ee': 'bengaligbeŋɔŋlɔ', 'el': 'Μπενγκάλι', 'el-polyton': 'Μπενγκάλι', 'en': 'Bangla', 'en-Dsrt': '𐐒𐐯𐑌𐑀𐐪𐑊𐐨', 'es': 'bengalí', 'et': 'bengali', 'eu': 'bengalarra', 'fa': 'بنگالی', 'ff-Adlm': '𞤄𞤫𞤲𞤺𞤢𞤤𞤭', 'fi': 'bengalilainen', 'fil': 'Bangla', 'fo': 'bangla', 'fr': 'bengali', 'fur': 'bengalês', 'fy': 'Bengalees', 'ga': 'Beangálach', 'gd': 'Beangailis', 'gl': 'bengalí', 'gsw': 'Bengalisch', 'gu': 'બંગાળી', 'ha': 'Bangla', 'he': 'בנגלי', 'hi': 'बंगाली', 'hi-Latn': 'Bangla', 'hr': 'bengalsko pismo', 'hsb': 'bengalsce', 'ht': 'bengali', 'hu': 'Bengáli', 'hy': 'բենգալական', 'ia': 'bengalese', 'id': 'Bengali', 'ig': 'Mkpụrụ ọkwụ Bangla', 'is': 'bengalskt', 'it': 'bengalese', 'ja': 'ベンガル文字', 'jv': 'Bangla', 'ka': 'ბენგალური', 'kea': 'bengali', 'kgp': 'megari', 'kk': 'бенгал жазуы', 'kk-Cyrl': 'бенгал жазуы', 'km': 'បង់ក្លាដែស', 'kn': 'ಬೆಂಗಾಲಿ', 'ko': '벵골 문자', 'kok': 'बांग्ला', 'kok-Deva': 'बांग्ला', 'ks': 'بینگٲلؠ', 'ks-Arab': 'بینگٲلؠ', 'ksh': 'bängjaalesche Schreff', 'ku': 'bengalî', 'kxv': 'bangalī', 'kxv-Deva': 'बंगाली', 'kxv-Latn': 'bangalī', 'kxv-Orya': 'ବଙ୍ଗାଲୀ', 'kxv-Telu': 'బాంగ్లా', 'ky': 'Бенгал', 'lb': 'Bengalesch', 'lo': 'ເບັງກາ', 'lrc': 'بأنگالی', 'lt': 'bengalų', 'lv': 'bengāļu', 'mai': 'बंगाली', 'mi': 'Pāngara', 'mk': 'бенгалско писмо', 'ml': 'ബംഗാളി', 'mn': 'бенгал', 'mni': 'বাংলা', 'mni-Beng': 'বাংলা', 'mni-Mtei': 'ꯕꯪꯂꯥ', 'mr': 'बंगाली', 'ms': 'Benggala', 'my': 'ဘင်္ဂါလီ', 'mzn': 'بنگالی', 'nb': 'bengalsk', 'ne': 'बङ्गाली', 'nl': 'Bengaals', 'nn': 'bengalsk', 'no': 'bengalsk', 'nqo': 'ߓߍ߲ߜ߭ߊߟߌ', 'or': 'ବଙ୍ଗାଳୀ', 'pa': 'ਬੰਗਾਲੀ', 'pa-Guru': 'ਬੰਗਾਲੀ', 'pcm': 'Bángla', 'pl': 'bengalskie', 'ps': 'بنګله', 'pt': 'bengali', 'qu': 'Bangla Simi', 'rm': 'bengal', 'ro': 'bengaleză', 'ru': 'бенгальская', 'sc': 'bengalesu', 'sd': 'بنگلا', 'sd-Arab': 'بنگلا', 'si': 'බෙංගාලි', 'sk': 'bengálske', 'sl': 'bengalski', 'so': 'Baangla', 'sq': 'bengal', 'sr': 'бенгалско писмо', 'sr-Cyrl': 'бенгалско писмо', 'sr-Latn': 'bengalsko pismo', 'sv': 'bengaliska', 'sw': 'Kibengali', 'ta': 'வங்காளம்', 'te': 'బాంగ్లా', 'tg': 'Бангладешӣ', 'th': 'เบงกาลี', 'ti': 'ቋንቋ ቤንጋል', 'tk': 'Bengal elipbiýi', 'to': 'tohinima fakapāngilā', 'tr': 'Bengal', 'ug': 'بېنگال', 'uk': 'бенгальська', 'ur': 'بنگالی', 'uz': 'bengal', 'uz-Cyrl': 'Бенгали', 'uz-Latn': 'bengal', 'vec': 'bengaleze', 'vi': 'Chữ Bangla', 'wae': 'Bengališ', 'xnr': 'बंगाली', 'yo': 'Báńgílà', 'yrl': 'bẽgari', 'yue': '孟加拉文', 'yue-Hans': '孟加拉文', 'yue-Hant': '孟加拉文', 'zh': '孟加拉文', 'zh-Hans': '孟加拉文', 'zh-Hant': '孟加拉文', 'zu': 'isi-Bangla'},
    'Bhks': {'bal-Latn': 'Baykduki', 'bs': 'baiksuki pismo', 'bs-Cyrl': 'баиксуки писмо', 'bs-Latn': 'baiksuki pismo', 'ca': 'bhaiksuki', 'ca-ES-valencia': 'bhaiksuki', 'en': 'Bhaiksuki', 'eu': 'bhaiksuki', 'ff-Adlm': '𞤄𞤢𞤴𞤳𞤵𞤧𞤵𞤳𞤭', 'fi': 'sanskritin bhaiksuki-aakkosto', 'fo': 'bhaiksuki', 'ga': 'Bhaiksuki', 'gd': 'Bhaiksuki', 'he': 'בהיקסוקי', 'hi-Latn': 'Bhaiksuki', 'id': 'Bhaiksuki', 'ig': 'Bhaiksuki', 'kk': 'бхайксуки жазуы', 'kk-Cyrl': 'бхайксуки жазуы', 'km': 'ប៉ៃស៊ូគី', 'mn': 'Байксуки', 'ms': 'Bhaisuki', 'nl': 'Bhaiksuki', 'nqo': 'ߓߊߦߌߞߑߛߎߞߌ', 'ro': 'bhaiksuki', 'sc': 'bhaiksuki', 'so': 'Qoraalka Bhaiksuki', 'sq': 'baiksukisht', 'sv': 'bhaiksukiska', 'tg': 'Бхайксукӣ', 'vec': 'baiksuki', 'zh': '拜克舒克文', 'zh-Hans': '拜克舒克文', 'zh-Hant': '梵文', 'zu': 'isi-Bhaiksuki'},
    'Blis': {'ar': 'رموز بليس', 'ast': 'símbolos de Bliss', 'az': 'blissymbols', 'az-Latn': 'blissymbols', 'bg': 'Блис символи', 'bn': 'ব্লিসপ্রতীক', 'brx': 'ब्लीस चीन्ह', 'bs': 'blisimbolično pismo', 'bs-Cyrl': 'блисимболично писмо', 'bs-Latn': 'blisimbolično pismo', 'ca': 'símbols Bliss', 'ca-ES-valencia': 'símbols Bliss', 'ccp': '𑄝𑄳𑄣𑄨𑄌𑄴𑄛𑄳𑄢𑄧𑄖𑄩𑄇𑄴', 'cs': 'Blissovo písmo', 'da': 'blissymboler', 'de': 'Bliss-Symbole', 'el': 'Σύμβολα Bliss', 'el-polyton': 'Σύμβολα Bliss', 'en': 'Blissymbols', 'en-Dsrt': '𐐒𐑊𐐮𐑅𐐮𐑋𐐺𐐲𐑊𐑆', 'es': 'símbolos blis', 'et': 'Blissi sümbolid', 'fa': 'نمادهای بلیس', 'fi': 'bliss-symbolit', 'fr': 'symboles Bliss', 'fy': 'Blissymbolen', 'gd': 'Comharran Bliss', 'gsw': 'Bliss-Symbool', 'gu': 'બ્લિસિમ્બોલ્સ', 'hi': 'ब्लिसिम्बॉल्स', 'hi-Latn': 'Blissymbols', 'hr': 'blissymbols', 'ht': 'symboles Bliss', 'hu': 'Bliss jelképrendszer', 'id': 'Blissymbol', 'it': 'simboli bliss', 'ja': 'ブリスシンボル', 'ka': 'ბლისსიმბოლოები', 'kgp': 'sĩmuru mrij', 'kn': 'ಬ್ಲಿಸ್\u200cಸಿಂಬಲ್ಸ್', 'ko': '블리스기호 문자', 'ks': 'بِلِس سِمبلز', 'ks-Arab': 'بِلِس سِمبلز', 'lb': 'Bliss-Symboler', 'lo': 'ບລິກຊິມໂບລສ', 'lt': '„Bliss“ simboliai', 'mk': 'блиссимболи', 'ml': 'ബ്ലിസ് ചിത്ര ലിപി', 'mr': 'ब्लिसिम्बॉल्स', 'nb': 'blissymbol', 'ne': 'ब्लिजसिम्बोल्स', 'nl': 'Blissymbolen', 'nn': 'blissymbol', 'no': 'blissymbol', 'or': 'ବ୍ଲିସିମ୍ବଲସ୍', 'pl': 'symbole Blissa', 'pt': 'símbolos bliss', 'rm': 'simbols da Bliss', 'ru': 'блиссимволика', 'sl': 'znakovna pisava Bliss', 'sr': 'блисимболично писмо', 'sr-Cyrl': 'блисимболично писмо', 'sr-Latn': 'blisimbolično pismo', 'sv': 'blissymboler', 'ta': 'ப்லிஸ்ஸிமிபால்ஸ்', 'te': 'బ్లిస్సింబల్స్', 'th': 'บลิสซิมโบลส์', 'to': 'tohinima fakaʻilonga-pilisi', 'tr': 'Blis Sembolleri', 'ug': 'بىلىس بەلگىلىرى', 'uk': 'символи Блісса', 'vi': 'Chữ Blissymbols', 'yrl': 'rãgasaitá bliss', 'yue': '布列斯文', 'yue-Hans': '布列斯文', 'yue-Hant': '布列斯文', 'zh': '布列斯符号', 'zh-Hans': '布列斯符号', 'zh-Hant': '布列斯文'},
    'Bopo': {'af': 'Bopomofo', 'ak': 'Bopomofo kasa', 'am': 'ቦፖሞፎ', 'ar': 'البوبوموفو', 'as': 'বোপোমোফো', 'ast': 'bopomofo', 'az': 'bopomofo', 'az-Latn': 'bopomofo', 'bal-Latn': 'Bópó', 'be': 'бапамофа', 'be-tarask': 'бапамофа', 'bg': 'бопомофо', 'bgn': 'بوپوموفو', 'bn': 'বোপোমোফো', 'br': 'bopomofo', 'brx': 'बप’मफ’', 'bs': 'pismo bopomofo', 'bs-Cyrl': 'бопомофо писмо', 'bs-Latn': 'pismo bopomofo', 'ca': 'bopomofo', 'ca-ES-valencia': 'bopomofo', 'ccp': '𑄝𑄮𑄛𑄮𑄟𑄮𑄜𑄮', 'ce': 'бопомофо', 'chr': 'ᏆᏉᎼᏬ', 'ckb': 'بۆپۆمۆفۆ', 'cs': 'bopomofo', 'cy': 'Bopomofo', 'da': 'bopomofo', 'de': 'Bopomofo', 'dsb': 'bopomofo', 'dz': 'བོ་པོ་མོ་ཕཱོ་ཡིག་གུ', 'ee': 'bopomfogbeŋɔŋlɔ', 'el': 'Μποπομόφο', 'el-polyton': 'Μποπομόφο', 'en': 'Bopomofo', 'en-Dsrt': '𐐒𐐱𐐹𐐱𐑋𐐱𐑁𐐱', 'es': 'bopomofo', 'et': 'bopomofo', 'eu': 'bopomofoa', 'fa': 'بوپوموفو', 'ff-Adlm': '𞤄𞤮𞤨𞤮𞤥𞤮𞤬𞤮', 'fi': 'bopomofo', 'fil': 'Bopomofo', 'fo': 'bopomofo', 'fr': 'bopomofo', 'fy': 'Bopomofo', 'ga': 'Bopomofo', 'gd': 'Bopomofo', 'gl': 'bopomofo', 'gsw': 'Bopomofo', 'gu': 'બોપોમોફો', 'ha': 'Bopomofo', 'he': 'בופומופו', 'hi': 'बोपोमोफ़ो', 'hi-Latn': 'Bopomofo', 'hr': 'bopomofo pismo', 'hsb': 'bopomofo', 'ht': 'bopomofo', 'hu': 'Bopomofo', 'hy': 'բոպոմոֆո', 'ia': 'bopomofo', 'id': 'Bopomofo', 'ig': 'Mkpụrụ ọkwụ Bopomofo', 'is': 'bopomofo', 'it': 'bopomofo', 'ja': '注音字母', 'jv': 'Bopomofo', 'ka': 'ბოპომოფო', 'kea': 'bopomofo', 'kgp': 'mopomãfo', 'kk': 'бопомофо жазуы', 'kk-Cyrl': 'бопомофо жазуы', 'km': 'បូផូម៉ូហ្វូ', 'kn': 'ಬೋಪೊಮೋಫೋ', 'ko': '주음부호', 'kok': 'बोपोमोफो', 'kok-Deva': 'बोपोमोफो', 'ks': 'بوپوموفو', 'ks-Arab': 'بوپوموفو', 'ksh': 'schineeseche Ömschreff Bopomofo', 'ku': 'bopomofo', 'ky': 'Бопомофо', 'lb': 'Bopomofo', 'lo': 'ຈູ້ອິນ', 'lrc': 'بوٙپوٙ', 'lt': 'bopomofo', 'lv': 'bopomofo', 'mai': 'बोपोमोफो', 'mi': 'Papamawha', 'mk': 'бопомофо', 'ml': 'ബോപ്പോമോഫോ', 'mn': 'вопомофо', 'mr': 'बोपोमोफो', 'ms': 'Bopomofo', 'my': 'ဘိုပို', 'mzn': 'بوپوموفو', 'nb': 'bopomofo', 'ne': 'बोपोमोफो', 'nl': 'Bopomofo', 'nn': 'bopomofo', 'no': 'bopomofo', 'nqo': 'ߔߓߏߡߝߏ', 'or': 'ବୋପୋମୋଫୋ', 'pa': 'ਬੋਪੋਮੋਫੋ', 'pa-Guru': 'ਬੋਪੋਮੋਫੋ', 'pcm': 'Bopomófo', 'pl': 'bopomofo', 'ps': 'بوپوموفو', 'pt': 'bopomofo', 'qu': 'Bopomofo Simi', 'rm': 'bopomofo', 'ro': 'bopomofo', 'ru': 'бопомофо', 'sc': 'bopomofo', 'sd': 'بوپوموفو', 'sd-Arab': 'بوپوموفو', 'si': 'බොපොමොෆෝ', 'sk': 'bopomofo', 'sl': 'bopomofo', 'so': 'Bobomofo', 'sq': 'bopomof', 'sr': 'бопомофо писмо', 'sr-Cyrl': 'бопомофо писмо', 'sr-Latn': 'bopomofo pismo', 'sv': 'bopomofo', 'sw': 'Kibopomofo', 'ta': 'போபோமோஃபோ', 'te': 'బోపోమోఫో', 'tg': 'Бопомофоӣ', 'th': 'ปอพอมอฟอ', 'ti': 'ቦፖሞፎ', 'tk': 'Bopomofo elipbiýi', 'to': 'tohinima fakapopomofo', 'tr': 'Bopomofo', 'ug': 'خەنچە پىنيىن', 'uk': 'бопомофо', 'ur': 'بوپوموفو', 'uz': 'bopomofo', 'uz-Cyrl': 'Бопомофо', 'uz-Latn': 'bopomofo', 'vec': 'bopomofo', 'vi': 'Chữ Bopomofo', 'xnr': 'बोपोमोफ़ो', 'yo': 'Bopomófò', 'yrl': 'bupumufu', 'yue': '注音符號', 'yue-Hans': '注音符号', 'yue-Hant': '注音符號', 'zh': '注音符号', 'zh-Hans': '注音符号', 'zh-Hant': '注音符號', 'zu': 'isi-Bopomofo'},
    'Brah': {'ar': 'الهندوسية', 'ast': 'brahmi', 'az': 'brahmi', 'az-Latn': 'brahmi', 'bal-Latn': 'Brahmi', 'bg': 'Брахми', 'bn': 'ব্রাহ্মী', 'brx': 'ब्रह्मी', 'bs': 'bramansko pismo', 'bs-Cyrl': 'браманско писмо', 'bs-Latn': 'bramansko pismo', 'ca': 'brahmi', 'ca-ES-valencia': 'brahmi', 'ccp': '𑄝𑄳𑄢𑄟𑄴𑄦𑄴𑄟𑄩', 'cs': 'bráhmí', 'da': 'bramisk', 'de': 'Brahmi', 'el': 'Μπραχμί', 'el-polyton': 'Μπραχμί', 'en': 'Brahmi', 'en-Dsrt': '𐐒𐑉𐐪𐑋𐐨', 'es': 'brahmi', 'et': 'braahmi', 'eu': 'brahmiera', 'fa': 'براهمی', 'ff-Adlm': '𞤄𞤪𞤢𞤸𞤢𞤥𞤭', 'fi': 'brahmi', 'fo': 'brahmi', 'fr': 'brâhmî', 'fy': 'Brahmi', 'ga': 'Brámais', 'gd': 'Brahmi', 'gsw': 'Brahmi', 'gu': 'બ્રહ્મી', 'he': 'ברהמי', 'hi': 'ब्रह्मी', 'hi-Latn': 'Brahmi', 'hr': 'brahmi pismo', 'ht': 'brâhmî', 'hu': 'Brámi', 'id': 'Brahmi', 'ig': 'Brahmi', 'is': 'brahmíska', 'it': 'brahmi', 'ja': 'ブラーフミー文字', 'ka': 'ბრაჰმი', 'kgp': 'mramĩ', 'kk': 'брахми жазуы', 'kk-Cyrl': 'брахми жазуы', 'km': 'ប្រាមិ', 'kn': 'ಬ್ರಾಹ್ಮಿ', 'ko': '브라미', 'ks': 'براہمی', 'ks-Arab': 'براہمی', 'kxv': 'brahmi', 'kxv-Deva': 'ब्राह्मि', 'kxv-Latn': 'brahmi', 'kxv-Orya': 'ବ୍ରାହ୍ମି', 'kxv-Telu': 'బ్రాహ్మి', 'lb': 'Brahmi', 'lo': 'ພຮາຫມີ', 'lt': 'brahmi', 'lv': 'brahmi', 'mk': 'брами', 'ml': 'ബ്രാഹ്മി', 'mn': 'Брахми', 'mr': 'ब्रह्मी', 'ms': 'Brahmi', 'my': 'ဗွဟ်မမီ', 'nb': 'brahmi', 'ne': 'ब्राह्मी', 'nl': 'Brahmi', 'nn': 'brahmi', 'no': 'brahmi', 'nqo': 'ߓߑߙߊߤߡߌ', 'or': 'ବ୍ରାହ୍ମୀ', 'pl': 'brahmi', 'pt': 'brahmi', 'rm': 'brahmi', 'ro': 'brahmanică', 'ru': 'брахми', 'sc': 'brahmi', 'sl': 'bramanski', 'so': 'Dhirta Brahmi', 'sq': 'brahmisht', 'sr': 'браманско писмо', 'sr-Cyrl': 'браманско писмо', 'sr-Latn': 'bramansko pismo', 'sv': 'brami', 'ta': 'பிரம்மி', 'te': 'బ్రాహ్మి', 'tg': 'Брахмӣ', 'th': 'พราหมี', 'to': 'tohinima fakapalāmī', 'tr': 'Brahmi', 'ug': 'براخمى', 'uk': 'брахмі', 'vec': 'brahmi', 'vi': 'Chữ Brahmi', 'xnr': 'ब्रह्मी', 'yrl': 'brami', 'yue': '婆羅米文', 'yue-Hans': '婆罗米文', 'yue-Hant': '婆羅米文', 'zh': '婆罗米文字', 'zh-Hans': '婆罗米文字', 'zh-Hant': '婆羅米文', 'zu': 'isi-Brahmi'},
    'Brai': {'af': 'Braille', 'ak': 'Anifrafoɔ kasa', 'am': 'ብሬይል', 'ar': 'البرايل', 'as': 'ব্ৰেইল', 'ast': 'braille', 'az': 'brayl', 'az-Arab': 'بریل', 'az-Latn': 'brayl', 'bal-Latn': 'Brái', 'be': 'шрыфт Брайля', 'be-tarask': 'шрыфт Брайля', 'bg': 'брайлова', 'bn': 'ব্রেইল', 'br': 'Braille', 'brx': 'ब्रैल', 'bs': 'brajevo pismo', 'bs-Cyrl': 'брајево писмо', 'bs-Latn': 'brajevo pismo', 'ca': 'braille', 'ca-ES-valencia': 'braille', 'ccp': '𑄝𑄳𑄢𑄳𑄆𑄬𑄣𑄴', 'ce': 'брайля', 'chr': 'ᏗᏂᎨᏫ ᎤᏃᏪᎶᏙᏗ', 'ckb': 'برەیل', 'cs': 'Braillovo písmo', 'cy': 'Braille', 'da': 'punktskrift', 'de': 'Braille', 'dsb': 'braillowe pismo', 'dz': 'འབུར་ཡིག', 'ee': 'braillegbeŋɔŋlɔ', 'el': 'Μπράιγ', 'el-polyton': 'Μπράιγ', 'en': 'Braille', 'en-Dsrt': '𐐒𐑉𐐩𐑊', 'es': 'braille', 'et': 'punktkiri', 'eu': 'braillea', 'fa': 'بریل', 'ff-Adlm': '𞤄𞤢𞤪𞤢𞤭𞥅𞤤𞤵', 'fi': 'braille-pistekirjoitus', 'fil': 'Braille', 'fo': 'blindaskrift', 'fr': 'braille', 'fur': 'Braille', 'fy': 'Braille', 'ga': 'Braille', 'gd': 'Braille', 'gl': 'braille', 'gsw': 'Blindäschrift', 'gu': 'બ્રેલ', 'ha': 'Rubutun Makafi', 'he': 'ברייל', 'hi': 'ब्रेल', 'hi-Latn': 'Braille', 'hr': 'brajica', 'hsb': 'Braillowe pismo', 'ht': 'braille', 'hu': 'Vakírás', 'hy': 'Բրայլ', 'ia': 'braille', 'id': 'Braille', 'ig': 'Braille', 'is': 'blindraletur', 'it': 'braille', 'ja': 'ブライユ点字', 'jv': 'Braille', 'ka': 'ბრაილი', 'kea': 'braille', 'kgp': 'mrajiri', 'kk': 'Брайль жазуы', 'kk-Cyrl': 'Брайль жазуы', 'km': 'អក្សរ\u200bសម្រាប់មនុស្ស\u200bពិការ\u200bភ្នែក', 'kn': 'ಬ್ರೈಲ್', 'ko': '브라유 점자', 'kok': 'ब्रेल', 'kok-Deva': 'ब्रेल', 'ks': 'بریل', 'ks-Arab': 'بریل', 'ksh': 'Blindeschreff', 'ku': 'braille', 'ky': 'Брейл', 'lb': 'Blanneschrëft', 'lo': 'ເບຣວ', 'lrc': 'بئرئیل', 'lt': 'brailio', 'lv': 'Braila raksts', 'mai': 'ब्रेल', 'mi': 'Tuhi Matapō', 'mk': 'брајово писмо', 'ml': 'ബ്രെയ്\u200cലി', 'mn': 'брайл', 'mr': 'ब्रेल', 'ms': 'Braille', 'mt': 'Braille', 'my': 'ဘရေ', 'nb': 'punktskrift', 'ne': 'ब्रेल', 'nl': 'braille', 'nn': 'punktskrift', 'no': 'punktskrift', 'nqo': 'ߓߑߙߊߦߌ', 'or': 'ବ୍ରେଲ୍', 'pa': 'ਬਰੇਲ', 'pa-Guru': 'ਬਰੇਲ', 'pcm': 'Blaínd Pípol Raítín Sístẹm', 'pl': 'Braille’a', 'ps': 'بریلي', 'pt': 'braille', 'qu': 'Braile', 'rm': 'scrittira da Braille', 'ro': 'braille', 'ru': 'шрифт Брайля', 'sc': 'braille', 'sd': 'بريل', 'sd-Arab': 'بريل', 'si': 'බ්\u200dරේල්', 'sk': 'braillovo', 'sl': 'braillova pisava', 'so': 'Qoraalka Indhoolaha', 'sq': 'brailisht', 'sr': 'брајево писмо', 'sr-Cyrl': 'брајево писмо', 'sr-Latn': 'brajevo pismo', 'sv': 'punktskrift', 'sw': 'Nukta nundu', 'ta': 'பிரெயில்', 'te': 'బ్రెయిలీ', 'tg': 'Брайл', 'th': 'เบรลล์', 'ti': 'ብሬል', 'tk': 'Braýl elipbiýi', 'to': 'tohinima laukonga ki he kui', 'tr': 'Braille', 'ug': 'برائىل ئەمالار يېزىقى', 'uk': 'шрифт Брайля', 'ur': 'بریل', 'uz': 'brayl', 'uz-Cyrl': 'Браилле', 'uz-Latn': 'brayl', 'vec': 'braille', 'vi': 'Chữ nổi Braille', 'xnr': 'ब्रेल', 'yo': 'Bíráìlè', 'yrl': 'braire', 'yue': '盲人用點字', 'yue-Hans': '盲人用点字', 'yue-Hant': '盲人用點字', 'zh': '布莱叶盲文', 'zh-Hans': '布莱叶盲文', 'zh-Hant': '盲人用點字', 'zu': 'i-Braille'},
    'Bugi': {'ar': 'البجينيز', 'ast': 'lontara', 'az': 'buqin', 'az-Latn': 'buqin', 'bal-Latn': 'Bugini', 'bg': 'Бугинска', 'bn': 'বুগি', 'br': 'bougiek', 'brx': 'बुगीनी', 'bs': 'buginsko pismo', 'bs-Cyrl': 'бугинско писмо', 'bs-Latn': 'buginsko pismo', 'ca': 'buginès', 'ca-ES-valencia': 'buginès', 'ccp': '𑄝𑄪𑄉𑄨', 'cs': 'buginské', 'da': 'buginesisk', 'de': 'Buginesisch', 'el': 'Μπούγκις', 'el-polyton': 'Μπούγκις', 'en': 'Buginese', 'en-Dsrt': '𐐒𐐭𐑀𐐮𐑌𐐨𐑆', 'es': 'buginés', 'eu': 'buginera', 'fa': 'بوگیایی', 'ff-Adlm': '𞤄𞤵𞤺𞤭𞤲𞤭𞤴𞤢', 'fi': 'bugilainen', 'fo': 'buginskt', 'fr': 'bouguis', 'fur': 'buginês', 'fy': 'Bugineesk', 'ga': 'Buigineach', 'gd': 'Lontara', 'gsw': 'Buginesisch', 'gu': 'બગિનીસ', 'he': 'בוגינזי', 'hi': 'बगिनीस', 'hi-Latn': 'Buginese', 'hr': 'buginsko pismo', 'ht': 'bouguis', 'hu': 'Buginéz', 'id': 'Bugis', 'ig': 'Buginese', 'it': 'buginese', 'ja': 'ブギス文字', 'kgp': 'mugnẽj', 'kk': 'бугий жазуы', 'kk-Cyrl': 'бугий жазуы', 'km': 'ប៊ូគីនេ', 'kn': 'ಬಗಿನೀಸ್', 'ko': '부기 문자', 'ks': 'بُگِنیٖز', 'ks-Arab': 'بُگِنیٖز', 'lb': 'Buginesesch', 'lo': 'ບູກິສ', 'lt': 'buginezų', 'mk': 'бугиско', 'ml': 'ബുഗിനീസ്', 'mn': 'Бугинез', 'mr': 'बूगी', 'ms': 'Bugis', 'nb': 'buginesisk', 'ne': 'बुगिनिज', 'nl': 'Buginees', 'nn': 'buginesisk', 'no': 'buginesisk', 'nqo': 'ߓߎߜ߭ߌ', 'or': 'ବୁଗାନୀଜ୍', 'pl': 'bugińskie', 'pt': 'buginês', 'rm': 'buginais', 'ro': 'bugineză', 'ru': 'бугинизийская', 'sc': 'buginesu', 'sl': 'buginski', 'so': 'Luuqada Buginiiska', 'sq': 'buginisht', 'sr': 'бугинско писмо', 'sr-Cyrl': 'бугинско писмо', 'sr-Latn': 'buginsko pismo', 'sv': 'buginesiska', 'ta': 'புகினீஸ்', 'te': 'బ్యుగినీస్', 'tg': 'Бугинӣ', 'th': 'บูกิส', 'to': 'tohinima fakapukisi', 'tr': 'Bugis', 'ug': 'بۇگى', 'uk': 'бугійський', 'vec': 'bugineze', 'vi': 'Chữ Bugin', 'xnr': 'बगिनीस', 'yrl': 'buginei', 'yue': '布吉斯文', 'yue-Hans': '布吉斯文', 'yue-Hant': '布吉斯文', 'zh': '布吉文', 'zh-Hans': '布吉文', 'zh-Hant': '布吉斯文', 'zu': 'isi-Buginese'},
    'Buhd': {'am': 'ቡሂድ', 'ar': 'البهيدية', 'ast': 'buhid', 'az': 'buhid', 'az-Latn': 'buhid', 'bal-Latn': 'Buhid', 'bg': 'Бухид', 'bn': 'বুহিড', 'brx': 'बुहीद', 'bs': 'buhidsko pismo', 'bs-Cyrl': 'бухидско писмо', 'bs-Latn': 'buhidsko pismo', 'ca': 'buhid', 'ca-ES-valencia': 'buhid', 'ccp': '𑄝𑄪𑄦𑄨𑄓𑄴', 'cs': 'buhidské', 'da': 'buhid', 'de': 'Buhid', 'el': 'Μπουχίντ', 'el-polyton': 'Μπουχίντ', 'en': 'Buhid', 'en-Dsrt': '𐐒𐐭𐐸𐐮𐐼', 'es': 'buhid', 'et': 'buhidi', 'eu': 'buhid', 'fa': 'بوهید', 'ff-Adlm': '𞤄𞤵𞤸𞤭𞤣', 'fi': 'buhidilainen', 'fo': 'buhid', 'fr': 'bouhide', 'fy': 'Buhid', 'ga': 'Buthaideach', 'gd': 'Buhid', 'gsw': 'Buhid', 'gu': 'બુહિદ', 'he': 'בוהיד', 'hi': 'बुहिद', 'hi-Latn': 'Buhid', 'hr': 'buhid pismo', 'ht': 'bouhide', 'hu': 'Buhid', 'id': 'Buhid', 'ig': 'Buhid', 'it': 'buhid', 'ja': 'ブヒッド文字', 'ka': 'ბუჰიდი', 'kgp': 'muhin', 'kk': 'бухид жазуы', 'kk-Cyrl': 'бухид жазуы', 'km': 'ប៊ូហ៊ីដ', 'kn': 'ಬುಹಿದ್', 'ko': '부히드 문자', 'ks': 'بُہِد', 'ks-Arab': 'بُہِد', 'lb': 'Buhid', 'lo': 'ບູຮິດ', 'lt': 'buhid', 'mk': 'бухидско', 'ml': 'ബുഹിഡ്', 'mn': 'Бухид', 'mr': 'बुहिद', 'ms': 'Buhid', 'nb': 'buhid', 'ne': 'बुहिद', 'nl': 'Buhid', 'nn': 'buhid', 'no': 'buhid', 'nqo': 'ߓߎߤߌߘ', 'or': 'ବୁହିଦ୍', 'pl': 'buhid', 'pt': 'buhid', 'rm': 'buhid', 'ro': 'buhidă', 'ru': 'бухид', 'sc': 'buhid', 'sl': 'buhidski', 'so': 'Luuqada Buhid', 'sq': 'buhidisht', 'sr': 'бухидско писмо', 'sr-Cyrl': 'бухидско писмо', 'sr-Latn': 'buhidsko pismo', 'sv': 'buhid', 'ta': 'புகித்', 'te': 'బుహిడ్', 'tg': 'Бухидӣ', 'th': 'บูฮิด', 'to': 'tohinima fakapuhiti', 'tr': 'Buhid', 'ug': 'بۇخىت', 'uk': 'бухід', 'vec': 'buhid', 'vi': 'Chữ Buhid', 'xnr': 'बुहिद', 'yrl': 'buwidi', 'yue': '布希德文', 'yue-Hans': '布希德文', 'yue-Hant': '布希德文', 'zh': '布希德文', 'zh-Hans': '布希德文', 'zh-Hant': '布希德文', 'zu': 'isi-Buhid'},
    'CA': {'af': 'Kanada', 'agq': 'Kanadà', 'ak': 'Kanada', 'am': 'ካናዳ', 'ar': 'كندا', 'as': 'কানাডা', 'asa': 'Kanada', 'ast': 'Canadá', 'az': 'Kanada', 'az-Cyrl': 'Канада', 'az-Latn': 'Kanada', 'bal-Latn': 'Kaynadhá', 'bas': 'Kànadà', 'be': 'Канада', 'be-tarask': 'Канада', 'bez': 'Hukanada', 'bg': 'Канада', 'bgn': 'کاناڈا', 'blo': 'Kanaɖaa', 'bm': 'Kanada', 'bn': 'কানাডা', 'br': 'Kanada', 'brx': 'केनाडा', 'bs': 'Kanada', 'bs-Cyrl': 'Канада', 'bs-Latn': 'Kanada', 'ca': 'Canadà', 'ca-ES-valencia': 'Canadà', 'ccp': '𑄇𑄚𑄓', 'ce': 'Канада', 'ceb': 'Canada', 'cgg': 'Kanada', 'chr': 'ᎨᎾᏓ', 'ckb': 'کەنەدا', 'cs': 'Kanada', 'csw': 'ᑳᓇᑕ', 'cv': 'Канада', 'cy': 'Canada', 'da': 'Canada', 'dav': 'Kanada', 'de': 'Kanada', 'dje': 'Kanaada', 'dsb': 'Kanada', 'dyo': 'Kanada', 'dz': 'ཀེ་ན་ཌ', 'ebu': 'Kanada', 'ee': 'Canada nutome', 'el': 'Καναδάς', 'el-polyton': 'Καναδάς', 'en': 'Canada', 'en-Dsrt': '𐐗𐐰𐑌𐐲𐐼𐐲', 'eo': 'Kanado', 'es': 'Canadá', 'et': 'Kanada', 'eu': 'Kanada', 'ewo': 'kanadá', 'fa': 'کانادا', 'ff': 'Kanadaa', 'ff-Adlm': '𞤑𞤢𞤲𞤢𞤣𞤢𞥄', 'ff-Latn': 'Kanadaa', 'fi': 'Kanada', 'fil': 'Canada', 'fo': 'Kanada', 'fr': 'Canada', 'fur': 'Canade', 'fy': 'Canada', 'ga': 'Ceanada', 'gd': 'Canada', 'gl': 'Canadá', 'gsw': 'Kanada', 'gu': 'કેનેડા', 'guz': 'Kanada', 'ha': 'Kanada', 'haw': 'Kanakā', 'he': 'קנדה', 'hi': 'कनाडा', 'hi-Latn': 'Canada', 'hr': 'Kanada', 'hsb': 'Kanada', 'ht': 'Canada', 'hu': 'Kanada', 'hy': 'Կանադա', 'ia': 'Canada', 'id': 'Kanada', 'ig': 'Canada', 'is': 'Kanada', 'it': 'Canada', 'ja': 'カナダ', 'jgo': 'Kanadâ', 'jmc': 'Kanada', 'jv': 'Kanada', 'ka': 'კანადა', 'kaa': 'Канада', 'kaa-Cyrl': 'Канада', 'kab': 'Kanada', 'kam': 'Kananda', 'kde': 'Kanada', 'kea': 'Kanadá', 'kgp': 'Kanỹna', 'khq': 'Kanaada', 'ki': 'Kanada', 'kk': 'Канада', 'kk-Cyrl': 'Канада', 'kln': 'Emetab Canada', 'km': 'កាណាដា', 'kn': 'ಕೆನಡಾ', 'ko': '캐나다', 'kok': 'कॅनडा', 'kok-Deva': 'कॅनडा', 'ks': 'کینیڈا', 'ks-Arab': 'کینیڈا', 'ksb': 'Kanada', 'ksf': 'kanada', 'ksh': 'Kanada', 'ku': 'Kanada', 'kxv': 'kanaḍa', 'kxv-Deva': 'कानाडा', 'kxv-Latn': 'kanaḍa', 'kxv-Orya': 'କାନାଡା', 'kxv-Telu': 'కెనడా', 'ky': 'Канада', 'lag': 'Kánada', 'lb': 'Kanada', 'lg': 'Kanada', 'lkt': 'Uŋčíyapi Makȟóčhe', 'lld': 'Canada', 'ln': 'Kanada', 'lo': 'ແຄນາດາ', 'lt': 'Kanada', 'lu': 'Kanada', 'luo': 'Canada', 'luy': 'Kanada', 'lv': 'Kanāda', 'mai': 'कनाडा', 'mas': 'Kanada', 'mer': 'Kanada', 'mfe': 'Kanada', 'mg': 'Kanada', 'mgh': 'Ukanada', 'mi': 'Kānata', 'mk': 'Канада', 'ml': 'കാനഡ', 'mn': 'Канад', 'mr': 'कॅनडा', 'ms': 'Kanada', 'mt': 'il-Kanada', 'mua': 'kanada', 'my': 'ကနေဒါ', 'mzn': 'کانادا', 'naq': 'Kanadab', 'nb': 'Canada', 'nd': 'Khanada', 'ne': 'क्यानाडा', 'nl': 'Canada', 'nmg': 'Kanada', 'nn': 'Canada', 'no': 'Canada', 'nqo': 'ߞߣߊߘߊ߫', 'nus': 'Känɛda', 'nyn': 'Kanada', 'om': 'Kanaadaa', 'or': 'କାନାଡା', 'pa': 'ਕੈਨੇਡਾ', 'pa-Guru': 'ਕੈਨੇਡਾ', 'pcm': 'Kánáda', 'pl': 'Kanada', 'ps': 'کاناډا', 'pt': 'Canadá', 'qu': 'Canadá', 'rm': 'Canada', 'rn': 'Kanada', 'ro': 'Canada', 'rof': 'Kanada', 'ru': 'Канада', 'rwk': 'Kanada', 'sah': 'Канаада', 'saq': 'Kanada', 'sbp': 'Kanada', 'sc': 'Cànada', 'sd': 'ڪينيڊا', 'sd-Arab': 'ڪينيڊا', 'se': 'Kanáda', 'seh': 'Canadá', 'ses': 'Kanaada', 'sg': 'Kanadäa', 'shi': 'ⴽⴰⵏⴰⴷⴰ', 'shi-Latn': 'kanada', 'shi-Tfng': 'ⴽⴰⵏⴰⴷⴰ', 'si': 'කැනඩාව', 'sk': 'Kanada', 'sl': 'Kanada', 'smn': 'Kanada', 'sn': 'Kanada', 'so': 'Kanada', 'sq': 'Kanada', 'sr': 'Канада', 'sr-Cyrl': 'Канада', 'sr-Latn': 'Kanada', 'sv': 'Kanada', 'sw': 'Kanada', 'syr': 'ܟܢܕܐ', 'ta': 'கனடா', 'te': 'కెనడా', 'teo': 'Kanada', 'tg': 'Канада', 'th': 'แคนาดา', 'ti': 'ካናዳ', 'tk': 'Kanada', 'to': 'Kānata', 'tr': 'Kanada', 'tt': 'Канада', 'twq': 'Kanaada', 'tzm': 'Kanada', 'ug': 'كانادا', 'uk': 'Канада', 'ur': 'کینیڈا', 'uz': 'Kanada', 'uz-Cyrl': 'Канада', 'uz-Latn': 'Kanada', 'vai': 'ꕪꕯꕜ', 'vai-Latn': 'Kánáda', 'vai-Vaii': 'ꕪꕯꕜ', 'vec': 'Cànada', 'vi': 'Canada', 'vun': 'Kanada', 'wae': 'Kanada', 'wo': 'Kanadaa', 'xh': 'EKhanada', 'xnr': 'कनेडा', 'xog': 'Kanada', 'yav': 'Kánáta', 'yi': 'קאַנאַדע', 'yo': 'Kánádà', 'yrl': 'Kanadá', 'yue': '加拿大', 'yue-Hans': '加拿大', 'yue-Hant': '加拿大', 'zgh': 'ⴽⴰⵏⴰⴷⴰ', 'zh': '加拿大', 'zh-Hans': '加拿大', 'zh-Hant': '加拿大', 'zu': 'i-Canada'},
    'CC': {'af': 'Kokoseilande', 'ak': 'Kokoso Supɔ', 'am': 'ኮኮስ(ኬሊንግ) ደሴቶች', 'ar': 'جزر كوكوس (كيلينغ)', 'as': 'কোকোচ (কীলিং) দ্বীপপুঞ্জ', 'ast': 'Islles Cocos (Keeling)', 'az': 'Kokos (Kilinq) adaları', 'az-Cyrl': 'Кокос (Килинг) адалары', 'az-Latn': 'Kokos (Kilinq) adaları', 'bal-Latn': 'Kukus Islánd (Kileng)', 'be': 'Какосавыя (Кілінг) астравы', 'be-tarask': 'Какосавыя (Кілінг) астравы', 'bg': 'Кокосови острови (острови Кийлинг)', 'blo': 'Kokoos (Kiiliŋ) kaBʊtukǝltǝna', 'bn': 'কোকোস (কিলিং) দ্বীপপুঞ্জ', 'br': 'Inizi Kokoz', 'brx': 'कक’स द्वीपफोर', 'bs': 'Kokosova (Keelingova) ostrva', 'bs-Cyrl': 'Кокос (Келинг) Острва', 'bs-Latn': 'Kokosova (Keelingova) ostrva', 'ca': 'Illes Cocos (Keeling)', 'ca-ES-valencia': 'Illes Cocos (Keeling)', 'ccp': '𑄇𑄮𑄇𑄮𑄌𑄴 (𑄇𑄨𑄣𑄨𑄁) 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Кокосийн гӀайренаш', 'ceb': 'Cocos (Keeling) Islands', 'chr': 'ᎪᎪᏍ (ᎩᎵᏂ) ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی کیلینگ', 'cs': 'Kokosové ostrovy', 'cv': 'Кокос утравӗсем', 'cy': 'Ynysoedd Cocos (Keeling)', 'da': 'Cocosøerne', 'de': 'Kokosinseln', 'dsb': 'Kokosowe kupy', 'dz': 'ཀོ་ཀོས་གླིང་ཚོམ', 'ee': 'Kokos (Kiling) fudomekpo nutome', 'el': 'Νήσοι Κόκος (Κίλινγκ)', 'el-polyton': 'Νῆσοι Κόκος (Κήλινγκ)', 'en': 'Cocos (Keeling) Islands', 'en-Dsrt': '𐐗𐐬𐐿𐐬𐑆 𐐌𐑊𐐲𐑌𐐼𐑆', 'es': 'Islas Cocos', 'et': 'Kookossaared', 'eu': 'Cocos (Keeling) uharteak', 'fa': 'جزایر کوکوس', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤑𞤮𞤳𞤮𞥅𞤧 (𞤑𞤭𞥅𞤤𞤭𞤲𞤺)', 'fi': 'Kookossaaret (Keelingsaaret)', 'fil': 'Cocos (Keeling) Islands', 'fo': 'Kokosoyggjar', 'fr': 'Îles Cocos', 'fur': 'Isulis Cocos', 'fy': 'Kokosilanen', 'ga': 'Oileáin Cocos (Keeling)', 'gd': 'Na h-Eileanan Chocos (Keeling)', 'gl': 'Illas Cocos (Keeling)', 'gsw': 'Kokos-Insle', 'gu': 'કોકોઝ (કીલીંગ) આઇલેન્ડ્સ', 'ha': 'Tsibirai Cocos (Keeling)', 'he': 'איי קוקוס (קילינג)', 'hi': 'कोकोस (कीलिंग) द्वीपसमूह', 'hi-Latn': 'Cocos (Keeling) Islands', 'hr': 'Kokosovi (Keelingovi) Otoci', 'hsb': 'Kokosowe kupy', 'ht': 'Îles Cocos', 'hu': 'Kókusz (Keeling)-szigetek', 'hy': 'Կոկոսյան (Քիլինգ) կղզիներ', 'ia': 'Insulas Cocos (Keeling)', 'id': 'Kepulauan Cocos (Keeling)', 'ig': 'Agwaetiti Cocos (Keeling)', 'is': 'Kókoseyjar (Keeling)', 'it': 'Isole Cocos (Keeling)', 'ja': 'ココス(キーリング)諸島', 'jv': 'Kapuloan Cocos (Keeling)', 'ka': 'ქოქოსის (კილინგის) კუნძულები', 'kaa': 'Кокос атаўлары', 'kaa-Cyrl': 'Кокос атаўлары', 'kea': 'Ilhas Kokus (Keeling)', 'kgp': 'Kokonh Goj-vẽso (Killing)', 'kk': 'Кокос (Килинг) аралдары', 'kk-Cyrl': 'Кокос (Килинг) аралдары', 'km': 'កោះ\u200bកូកូស (គីលីង)', 'kn': 'ಕೊಕೊಸ್ (ಕೀಲಿಂಗ್) ದ್ವೀಪಗಳು', 'ko': '코코스 제도', 'kok': 'कोकोस (कीलिंग) आयलँड', 'kok-Deva': 'कोकोस (कीलिंग) आयलँड', 'ks': 'کوکَس (کیٖلِنگ) جٔزیٖرٕ', 'ks-Arab': 'کوکَس (کیٖلِنگ) جٔزیٖرٕ', 'ksh': 'de Kokkos-Enselle', 'ku': 'Giravên Kokosê (Keeling)', 'kxv': 'kokos keling dīp', 'kxv-Deva': 'कोकोस (कीलिंग) द्वीप', 'kxv-Latn': 'kokos keling dīp', 'kxv-Orya': 'କକସ୍ (କିଲିଂ) ଦିପ', 'kxv-Telu': 'కోకోస్ (కీలింగ్) దీవులు', 'ky': 'Кокос (Килинг) аралдары', 'lb': 'Kokosinselen', 'lld': 'Isoles Cocos (Keeling)', 'lo': 'ຫມູ່ເກາະໂກໂກສ', 'lt': 'Kokosų (Kilingo) Salos', 'lv': 'Kokosu (Kīlinga) salas', 'mai': 'कोकेज (कीलिंग) द्वीपसमूह', 'mi': 'Ngā Moutere Kokoko (Kirini)', 'mk': 'Кокосови (Килиншки) Острови', 'ml': 'കോക്കസ് (കീലിംഗ്) ദ്വീപുകൾ', 'mn': 'Кокос (Кийлинг) арлууд', 'mr': 'कोकोस (कीलिंग) बेटे', 'ms': 'Kepulauan Cocos (Keeling)', 'mt': 'Gżejjer Cocos (Keeling)', 'my': 'ကိုကိုးကျွန်း', 'mzn': 'کوک (کیلینگ) جزایر', 'nb': 'Kokosøyene', 'ne': 'कोकोस (किलिंग) टापुहरु', 'nl': 'Cocoseilanden', 'nn': 'Kokosøyane', 'no': 'Kokosøyene', 'nqo': 'ߞߏߞߏ߫ ߕߌ߲', 'om': 'Odoloota Kokos (Keeliing)', 'or': 'କୋକୋସ୍ (କୀଲିଂ) ଦ୍ଵୀପପୁଞ୍ଜ', 'pa': 'ਕੋਕੋਸ (ਕੀਲਿੰਗ) ਟਾਪੂ', 'pa-Guru': 'ਕੋਕੋਸ (ਕੀਲਿੰਗ) ਟਾਪੂ', 'pcm': 'Kókós Aílands', 'pl': 'Wyspy Kokosowe', 'ps': 'کوکوز (کيلنګ) ټاپوګان', 'pt': 'Ilhas Cocos (Keeling)', 'qu': 'Islas Cocos', 'rm': 'Inslas Cocos', 'ro': 'Insulele Cocos (Keeling)', 'ru': 'Кокосовые о-ва', 'sc': 'Ìsulas Cocos (Keeling)', 'sd': 'ڪوڪوس ٻيٽ', 'sd-Arab': 'ڪوڪوس ٻيٽ', 'se': 'Cocos-sullot', 'si': 'කොකෝස් දූපත්', 'sk': 'Kokosové ostrovy', 'sl': 'Kokosovi otoki', 'smn': 'Kookossuolluuh (Keelingsuolluuh)', 'so': 'Jasiiradda Kookoos', 'sq': 'Ishujt Kokos', 'sr': 'Кокосова (Килингова) Острва', 'sr-Cyrl': 'Кокосова (Килингова) Острва', 'sr-Latn': 'Kokosova (Kilingova) Ostrva', 'sv': 'Kokosöarna', 'sw': 'Visiwa vya Cocos (Keeling)', 'syr': 'ܓܙܝܖ̈ܐ ܕܟܘܟܘܣ', 'ta': 'கோகோஸ் (கீலிங்) தீவுகள்', 'te': 'కోకోస్ (కీలింగ్) దీవులు', 'tg': 'Ҷазираҳои Кокос (Килинг)', 'th': 'หมู่เกาะโคโคส (คีลิง)', 'ti': 'ደሴታት ኮኮስ', 'tk': 'Kokos (Kiling) adalary', 'to': 'ʻOtumotu Koko', 'tr': 'Cocos (Keeling) Adaları', 'tt': 'Кокос (Килинг) утраулары', 'ug': 'كوكوس (كىلىڭ) ئاراللىرى', 'uk': 'Кокосові (Кілінг) Острови', 'ur': 'کوکوس (کیلنگ) جزائر', 'uz': 'Kokos (Kiling) orollari', 'uz-Cyrl': 'Кокос (Килинг) ороллари', 'uz-Latn': 'Kokos (Kiling) orollari', 'vai': 'ꖏꖏꔻ (ꔞꔀꔷꘋ) ꔳꘋꗣ', 'vai-Vaii': 'ꖏꖏꔻ (ꔞꔀꔷꘋ) ꔳꘋꗣ', 'vec': 'Ìzole Cocos (Keeling)', 'vi': 'Quần đảo Cocos (Keeling)', 'wae': 'Kokosinslä', 'wo': 'Duni Koko (Kilin)', 'xh': 'ECocos (Keeling) Islands', 'xnr': 'कोकोस (कीलिंग) द्वीपां दा झुंड', 'yo': 'Erékùsù Cocos (Keeling)', 'yrl': 'Kapuãma-ita Kuku (Keering)', 'yue': '科科斯（基林）群島', 'yue-Hans': '科科斯（基林）群岛', 'yue-Hant': '科科斯（基林）群島', 'zh': '科科斯（基林）群岛', 'zh-Hans': '科科斯（基林）群岛', 'zh-Hant': '科克斯（基靈）群島', 'zu': 'i-Cocos (Keeling) Islands'},
    'CD': {'af': 'Demokratiese Republiek van die Kongo', 'agq': 'Dɛ̀mùkàlatì Lèkpubèlè è Kuŋgù', 'ak': 'Kongo Kinhyaahya', 'am': 'ኮንጎ-ኪንሻሳ', 'ar': 'الكونغو - كينشاسا', 'as': 'কঙ্গো - কিনচাছা', 'asa': 'Jamhuri ya Kidemokrathia ya Kongo', 'ast': 'Congu - Kinxasa', 'az': 'Konqo - Kinşasa', 'az-Cyrl': 'Конго-Киншаса', 'az-Latn': 'Konqo - Kinşasa', 'bal-Latn': 'Kángó-Kenshásá', 'bas': 'Kòŋgo ìkɛŋi', 'be': 'Конга (Кіншаса)', 'be-tarask': 'Конга (Кіншаса)', 'bez': 'Ijamhuri ya Hidemokrasi ya Hukongo', 'bg': 'Конго (Киншаса)', 'blo': 'Koŋgoo Kinshasaa', 'bm': 'Kongo ka republiki demɔkratiki', 'bn': 'কঙ্গো-কিনশাসা', 'br': 'Kongo - Kinshasa', 'brx': 'कंग’-किनशासा', 'bs': 'Demokratska Republika Kongo', 'bs-Cyrl': 'Демократска Република Конго', 'bs-Latn': 'Demokratska Republika Kongo', 'ca': 'Congo - Kinshasa', 'ca-ES-valencia': 'Congo - Kinshasa', 'ccp': '𑄇𑄧𑄋𑄴𑄉𑄮-𑄚𑄨𑄇𑄴𑄥𑄥', 'ce': 'Демократин Республика Конго', 'ceb': 'Congo - Kinshasa', 'cgg': 'Demokoratika Ripaaburika ya Kongo', 'chr': 'ᎧᏂᎪ - ᎨᏂᏝᏌ', 'ckb': 'کۆنگۆ کینشاسا', 'cs': 'Kongo – Kinshasa', 'cv': 'Конго - Киншаса', 'cy': 'Y Congo - Kinshasa', 'da': 'Congo-Kinshasa', 'dav': 'Jamhuri ya Kidemokrasia ya Kongo', 'de': 'Kongo-Kinshasa', 'dje': 'Kongoo demookaratiki laboo', 'dsb': 'Kongo-Kinshasa', 'dyo': 'Mofam demokratik mati Kongo', 'dz': 'ཀོང་གྷོ ཀིན་ཤ་ས', 'ebu': 'Jamhuri ya Kidemokrasia ya Kongo', 'ee': 'Kongo Kinshasa nutome', 'el': 'Κονγκό - Κινσάσα', 'el-polyton': 'Κονγκό, Λαϊκὴ Δημοκρατία τοῦ', 'en': 'Congo - Kinshasa', 'en-Dsrt': '𐐗𐐪𐑍𐑀𐐬 - 𐐗𐐲𐑌𐑇𐐪𐑅𐐲', 'es': 'República Democrática del Congo', 'et': 'Kongo DV', 'eu': 'Kongoko Errepublika Demokratikoa', 'ewo': 'ǹnam Kongó Demokəlatíg', 'fa': 'کنگو - کینشاسا', 'ff': 'Ndenndaandi Demokaraasiire Konngo', 'ff-Adlm': '𞤑𞤮𞤲𞤺𞤮 - 𞤑𞤭𞤲𞤧𞤢𞤧𞤢', 'ff-Latn': 'Ndenndaandi Demokaraasiire Konngo', 'fi': 'Kongon demokraattinen tasavalta', 'fil': 'Congo - Kinshasa', 'fo': 'Kongo, Dem. Lýðveldið', 'fr': 'Congo-Kinshasa', 'fur': 'Republiche Democratiche dal Congo', 'fy': 'Congo-Kinshasa', 'ga': 'Poblacht Dhaonlathach an Chongó', 'gd': 'Congo - Kinshasa', 'gl': 'República Democrática do Congo', 'gsw': 'Temokraatischi Republik Kongo', 'gu': 'કોંગો - કિંશાસા', 'guz': 'Jamhuri ya Kidemokrasia ya Kongo', 'ha': 'Jamhuriyar Dimokuraɗiyyar Kongo', 'he': 'קונגו - קינשאסה', 'hi': 'कांगो - किंशासा', 'hi-Latn': 'Congo - Kinshasa', 'hr': 'Kongo - Kinshasa', 'hsb': 'Kongo-Kinshasa', 'ht': 'Congo-Kinshasa', 'hu': 'Kongó – Kinshasa', 'hy': 'Կոնգո - Կինշասա', 'ia': 'Congo - Kinshasa', 'id': 'Kongo - Kinshasa', 'ig': 'Congo - Kinshasa', 'is': 'Kongó-Kinshasa', 'it': 'Congo - Kinshasa', 'ja': 'コンゴ民主共和国(キンシャサ)', 'jgo': 'Kɔ́ŋgɔ-Kinshása', 'jmc': 'Jamhuri ya Kidemokrasia ya Kongo', 'jv': 'Kongo - Kinshasa', 'ka': 'კონგო - კინშასა', 'kaa': 'Конго - Киншаса', 'kaa-Cyrl': 'Конго - Киншаса', 'kab': 'Tigduda Tagdudant n Kungu', 'kam': 'Jamhuri ya Kidemokrasia ya Kongo', 'kde': 'Jamuhuli ya Chidemoklasia ya kuKongo', 'kea': 'Kongu - Kinxasa', 'kgp': 'Kãgo - Kĩsaja', 'khq': 'Kongoo demookaratiki laboo', 'ki': 'Jamhuri ya Kidemokrasia ya Kongo', 'kk': 'Конго', 'kk-Cyrl': 'Конго', 'kln': 'Emetab Congo - Kinshasa', 'km': 'កុងហ្គោ- គីនស្ហាសា', 'kn': 'ಕಾಂಗೋ - ಕಿನ್ಶಾಸಾ', 'ko': '콩고-킨샤사', 'kok': 'कोंगो - किंशासा', 'kok-Deva': 'कोंगो - किंशासा', 'ks': 'کونگو کِنشاسا', 'ks-Arab': 'کونگو کِنشاسا', 'ksb': 'Jamhuli ya Kidemoklasia ya Kongo', 'ksf': 'kɔngó anyɔ́n', 'ksh': 'dä Konggo (Kinschasa)', 'ku': 'Kongo - Kînşasa', 'kxv': 'kongo - kinsasa', 'kxv-Deva': 'कांगो - किंसासा', 'kxv-Latn': 'kongo - kinsasa', 'kxv-Orya': 'କଙ୍ଗ-କିଂସାସା', 'kxv-Telu': 'కాంగో కిన్సాసా', 'ky': 'Конго-Киншаса', 'lag': 'Jamuhúuri ya Kɨdemokurasía ya Kóongo', 'lb': 'Kongo-Kinshasa', 'lg': 'Kongo - Zayire', 'lld': 'Congo - Kinshasa', 'ln': 'Republíki ya Kongó Demokratíki', 'lo': 'ຄອງໂກ - ຄິນຊາຊາ', 'lt': 'Kongas-Kinšasa', 'lu': 'Ditunga wa Kongu', 'luo': 'Democratic Republic of the Congo', 'luy': 'Jamhuri ya Kidemokrasia ya Kongo', 'lv': 'Kongo (Kinšasa)', 'mai': 'कांगो-किंशासा', 'mas': 'Jamhuri ya Kidemokrasia ya Kongo', 'mer': 'Nthĩ ya Kidemokrasĩ ya Kongo', 'mfe': 'Repiblik demokratik Kongo', 'mg': 'Repoblikan’i Kongo', 'mi': 'Kōngo - Kinihāha', 'mk': 'Конго - Киншаса', 'ml': 'കോംഗോ - കിൻഷാസ', 'mn': 'Конго-Киншаса', 'mr': 'काँगो - किंशासा', 'ms': 'Congo - Kinshasa', 'mt': 'ir-Repubblika Demokratika tal-Kongo', 'mua': 'Sǝr Kongo ma dii ne zair', 'my': 'ကွန်ဂို', 'mzn': 'کنگو کینشاسا', 'naq': 'Democratic Republic of the Congo', 'nb': 'Kongo', 'nd': 'Democratic Republic of the Congo', 'ne': 'कङ्गो - किन्शासा', 'nl': 'Congo-Kinshasa', 'nmg': 'Kongó Zaïre', 'nn': 'Kongo-Kinshasa', 'no': 'Kongo', 'nqo': 'ߞߏ߲߬ߜ߭ߏ߫-ߞߌ߲ߛߊߛߊ߫', 'nyn': 'Demokoratika Ripaaburika ya Kongo', 'om': 'Koongoo - Kinshaasaa', 'or': 'କଙ୍ଗୋ (ଡିଆରସି)', 'pa': 'ਕਾਂਗੋ - ਕਿੰਸ਼ਾਸਾ', 'pa-Guru': 'ਕਾਂਗੋ - ਕਿੰਸ਼ਾਸਾ', 'pcm': 'Kóngó – Kinshása', 'pl': 'Demokratyczna Republika Konga', 'ps': 'کانګو - کینشاسا', 'pt': 'Congo - Kinshasa', 'qu': 'Congo (RDC)', 'rm': 'Congo - Kinshasa', 'rn': 'Repubulika Iharanira Demokarasi ya Kongo', 'ro': 'Congo - Kinshasa', 'rof': 'Jamhuri ya Kidemokrasia ya Kongo', 'ru': 'Конго - Киншаса', 'rwk': 'Jamhuri ya Kidemokrasia ya Kongo', 'saq': 'Jamhuri ya Kidemokrasia ya Kongo', 'sbp': 'Jamuhuli ya Kidemokilasiya ya Kongo', 'sc': 'Congo - Kinshasa', 'sd': 'ڪانگو -ڪنشاسا', 'sd-Arab': 'ڪانگو -ڪنشاسا', 'se': 'Kongo-Kinshasa', 'seh': 'Congo-Kinshasa', 'ses': 'Kongoo demookaratiki laboo', 'sg': 'Ködörösêse tî Ngunuhalëzo tî kongö', 'shi': 'ⵜⴰⴳⴷⵓⴷⴰⵏⵜ ⵜⴰⴷⵉⵎⵓⵇⵔⴰⵜⵉⵜ ⵏ ⴽⵓⵏⴳⵓ', 'shi-Latn': 'tagdudant tadimukratit n Kongo', 'shi-Tfng': 'ⵜⴰⴳⴷⵓⴷⴰⵏⵜ ⵜⴰⴷⵉⵎⵓⵇⵔⴰⵜⵉⵜ ⵏ ⴽⵓⵏⴳⵓ', 'si': 'කොංගෝව (ඩීආර්සී)', 'sk': 'Konžská demokratická republika', 'sl': 'Demokratična republika Kongo', 'smn': 'Kongo demokraattisâš täsiväldi', 'sn': 'Democratic Republic of the Congo', 'so': 'Jamhuuriyadda Dimuquraadiga Kongo', 'sq': 'Kongo-Kinshasa', 'sr': 'Конго - Киншаса', 'sr-Cyrl': 'Конго - Киншаса', 'sr-Latn': 'Kongo - Kinšasa', 'sv': 'Kongo-Kinshasa', 'sw': 'Jamhuri ya Kidemokrasia ya Kongo', 'syr': 'ܟܘܢܓܘ - ܟܝܢܫܐܣܐ', 'ta': 'காங்கோ - கின்ஷாசா', 'te': 'కాంగో- కిన్షాసా', 'teo': 'Jamhuri ya Kidemokrasia ya Kongo', 'tg': 'Конго - Киншаса', 'th': 'คองโก - กินชาซา', 'ti': 'ደሞክራስያዊት ሪፓብሊክ ኮንጎ', 'tk': 'Kongo - Kinşasa', 'to': 'Kongo - Kinisasa', 'tr': 'Kongo - Kinşasa', 'tt': 'Конго (КДР)', 'twq': 'Kongoo demookaratiki laboo', 'tzm': 'Tagduda Tadimuqraṭit n Kungu', 'ug': 'كونگو - كىنشاسا', 'uk': 'Конго – Кіншаса', 'ur': 'کانگو - کنشاسا', 'uz': 'Kongo – Kinshasa', 'uz-Cyrl': 'Конго-Киншаса', 'uz-Latn': 'Kongo – Kinshasa', 'vai': 'ꖏꖐ ꗵꗞꖴꕟꔎ ꕸꖃꔀ', 'vai-Latn': 'Avorekoo', 'vai-Vaii': 'ꖏꖐ ꗵꗞꖴꕟꔎ ꕸꖃꔀ', 'vec': 'Repùblega Democràtega de’l Congo', 'vi': 'Congo - Kinshasa', 'vun': 'Jamhuri ya Kidemokrasia ya Kongo', 'wae': 'Kongo-Kinshasa', 'wo': 'Kongo (R K D)', 'xh': 'ECongo -Kinshasa', 'xnr': 'कांगो - किंशासा', 'xog': 'Kongo - Zayire', 'yav': 'kitɔŋ kí kongó', 'yi': 'קאנגא־קינשאַזע', 'yo': 'Kóńgò – Kinshasa', 'yrl': 'Kũgu - Kĩxasa', 'yue': '剛果（金夏沙）', 'yue-Hans': '刚果（金夏沙）', 'yue-Hant': '剛果（金夏沙）', 'zgh': 'ⵜⴰⴳⴷⵓⴷⴰⵏⵜ ⵜⴰⴷⵉⵎⵓⵇⵔⴰⵜⵉⵜ ⵏ ⴽⵓⵏⴳⵓ', 'zh': '刚果（金）', 'zh-Hans': '刚果（金）', 'zh-Hant': '剛果（金夏沙）', 'zu': 'i-Congo - Kinshasa'},
    'CF': {'af': 'Sentraal-Afrikaanse Republiek', 'agq': 'Sɛnta Afɨlekan Lèkpobèlè', 'ak': 'Afrika Finimfin Man', 'am': 'ማዕከላዊ አፍሪካ ሪፑብሊክ', 'ar': 'جمهورية أفريقيا الوسطى', 'as': 'মধ্য আফ্রিকান প্রজাতন্ত্র', 'asa': 'Jamhuri ya Afrika ya Kati', 'ast': 'República Centroafricana', 'az': 'Mərkəzi Afrika Respublikası', 'az-Cyrl': 'Мәркәзи Африка Республикасы', 'az-Latn': 'Mərkəzi Afrika Respublikası', 'bal-Latn': 'Myáni Apriki Eshtán', 'bas': 'Ŋ̀ɛm Afrīkà', 'be': 'Цэнтральна-Афрыканская Рэспубліка', 'be-tarask': 'Цэнтральна-Афрыканская Рэспубліка', 'bez': 'Ijamhuri ya Afrika ya Pagati', 'bg': 'Централноафриканска република', 'blo': 'Santrafrika', 'bm': 'Santarafiriki', 'bn': 'মধ্য আফ্রিকার প্রজাতন্ত্র', 'br': 'Republik Kreizafrikan', 'brx': 'मिरु आफ्रिका सुबुंखुंथाय', 'bs': 'Centralnoafrička Republika', 'bs-Cyrl': 'Централноафричка Република', 'bs-Latn': 'Centralnoafrička Republika', 'ca': 'República Centreafricana', 'ca-ES-valencia': 'República Centreafricana', 'ccp': '𑄟𑄧𑄖𑄴𑄙𑄳𑄠𑄧 𑄃𑄜𑄳𑄢𑄨𑄇𑄢𑄴𑄛𑄳𑄢𑄎𑄖𑄧𑄚𑄴𑄖𑄳𑄢𑄧', 'ce': 'Юккъерчу Африкин Республика', 'ceb': 'Central African Republic', 'cgg': 'Eihanga rya Rwagati ya Afirika', 'chr': 'ᎬᎿᎨᏍᏛ ᎠᏰᏟ ᏍᎦᏚᎩ', 'ckb': 'کۆماری ئەفریقای ناوەڕاست', 'cs': 'Středoafrická republika', 'cv': 'Тӗп Африка Республики', 'cy': 'Gweriniaeth Canolbarth Affrica', 'da': 'Den Centralafrikanske Republik', 'dav': 'Jamhuri ya Afrika ya Kati', 'de': 'Zentralafrikanische Republik', 'dje': 'Centraafriki koyra', 'dsb': 'Centralnoafriska republika', 'dz': 'སེན་ཊལ་ ཨཕ་རི་ཀཱན་ རི་པབ་ལིཀ', 'ebu': 'Jamhuri ya Afrika ya Kati', 'ee': 'Titina Afrika repɔblik nutome', 'el': 'Κεντροαφρικανική Δημοκρατία', 'el-polyton': 'Κεντροαφρικανικὴ Δημοκρατία', 'en': 'Central African Republic', 'en-Dsrt': '𐐝𐐯𐑌𐐻𐑉𐐲𐑊 𐐈𐑁𐑉𐐲𐐿𐐲𐑌 𐐡𐐨𐐹𐐲𐐺𐑊𐐮𐐿', 'eo': 'Centr-Afrika Respubliko', 'es': 'República Centroafricana', 'et': 'Kesk-Aafrika Vabariik', 'eu': 'Afrika Erdiko Errepublika', 'ewo': 'ǹnam Zǎŋ Afiriká', 'fa': 'جمهوری افریقای مرکزی', 'ff': 'Ndenndaandi Santarafrik', 'ff-Adlm': '𞤐𞤣𞤫𞤲𞤣𞤭 𞤚𞤵𞤥𞤦𞤮𞥅𞤪𞤭 𞤀𞤬𞤪𞤭𞤳𞤢𞥄', 'ff-Latn': 'Ndenndaandi Santarafrik', 'fi': 'Keski-Afrikan tasavalta', 'fil': 'Central African Republic', 'fo': 'Miðafrikalýðveldið', 'fr': 'République centrafricaine', 'fur': 'Republiche centri africane', 'fy': 'Sintraal-Afrikaanske Republyk', 'ga': 'Poblacht na hAfraice Láir', 'gd': 'Poblachd Meadhan Afraga', 'gl': 'República Centroafricana', 'gsw': 'Zentraalafrikaanischi Republik', 'gu': 'સેન્ટ્રલ આફ્રિકન રિપબ્લિક', 'guz': 'Jamhuri ya Afrika ya Kati', 'ha': 'Jamhuriyar Afirka Ta Tsakiya', 'he': 'הרפובליקה המרכז-אפריקאית', 'hi': 'मध्य अफ़्रीकी गणराज्य', 'hi-Latn': 'Central African Republic', 'hr': 'Srednjoafrička Republika', 'hsb': 'Centralnoafriska republika', 'ht': 'République centrafricaine', 'hu': 'Közép-afrikai Köztársaság', 'hy': 'Կենտրոնական Աֆրիկյան Հանրապետություն', 'ia': 'Republica African Central', 'id': 'Republik Afrika Tengah', 'ig': 'Central African Republik', 'is': 'Mið-Afríkulýðveldið', 'it': 'Repubblica Centrafricana', 'ja': '中央アフリカ共和国', 'jmc': 'Jamhuri ya Afrika ya Kati', 'jv': 'Républik Afrika Tengah', 'ka': 'ცენტრალური აფრიკის რესპუბლიკა', 'kaa': 'Орайлық Африка Республикасы', 'kaa-Cyrl': 'Орайлық Африка Республикасы', 'kab': 'Tigduda n Tefriqt Talemmast', 'kam': 'Jamhuri ya Afrika ya Kati', 'kde': 'Jamuhuli ya Afilika ya Paching’ati', 'kea': 'Republika Sentru-Afrikanu', 'kgp': 'Afrikanỹ-kuju Repumrika', 'khq': 'Centraafriki koyra', 'ki': 'Jamhuri ya Afrika ya Kati', 'kk': 'Орталық Африка Республикасы', 'kk-Cyrl': 'Орталық Африка Республикасы', 'kln': 'Emetab Afrika nebo Kwen', 'km': 'សាធារណរដ្ឋអាហ្វ្រិកកណ្ដាល', 'kn': 'ಮಧ್ಯ ಆಫ್ರಿಕಾ ಗಣರಾಜ್ಯ', 'ko': '중앙 아프리카 공화국', 'kok': 'मध्य अफ्रीकी लोकसत्तकराज्य', 'kok-Deva': 'मध्य अफ्रीकी लोकसत्तकराज्य', 'ks': 'مرکٔزی اَفریٖکی جموٗریَت', 'ks-Arab': 'مرکٔزی اَفریٖکی جموٗریَت', 'ksb': 'Jamhuli ya Afrika ya Gati', 'ksf': 'santrafrík', 'ksh': 'de Zäntraalaffrikaanesche Republik', 'ku': 'Komara Afrîkaya Navîn', 'kxv': 'madini aaprika republik', 'kxv-Deva': 'मध्य अप़्रीकी गणराज्य', 'kxv-Latn': 'madini aaprika republik', 'kxv-Orya': 'ମାଦିନି ଆପ୍ରିକା ରିପବ୍ଲିକ୍', 'kxv-Telu': 'సెంట్రల్ ఆప్రికన్ రిపబ్లిక్', 'ky': 'Борбордук Африка Республикасы', 'lag': 'Juhúuri ya Afɨrɨka ya katɨ katɨ', 'lb': 'Zentralafrikanesch Republik', 'lg': 'Lipubulika eya Senturafiriki', 'lld': 'Republica Zentrafricana', 'ln': 'Repibiki ya Afríka ya Káti', 'lo': 'ສາທາລະນະລັດອາຟຣິກາກາງ', 'lt': 'Centrinės Afrikos Respublika', 'lu': 'Ditunga dya Afrika wa munkatshi', 'luo': 'Central African Republic', 'luy': 'Jamhuri ya Afrika ya Kati', 'lv': 'Centrālāfrikas Republika', 'mai': 'मध्य अफ्रीकी रिपब्लिक', 'mas': 'Jamhuri ya Afrika ya Kati', 'mer': 'Nthĩ ya Afrika gatĩgatĩ', 'mfe': 'Repiblik Lafrik Santral', 'mg': 'Repoblika Ivon’Afrika', 'mi': 'Te Whenua Tūhake o Āwherika Waenga', 'mk': 'Централноафриканска Република', 'ml': 'സെൻട്രൽ ആഫ്രിക്കൻ റിപ്പബ്ലിക്ക്', 'mn': 'Төв Африкийн Бүгд Найрамдах Улс', 'mr': 'केंद्रीय अफ्रिकी प्रजासत्ताक', 'ms': 'Republik Afrika Tengah', 'mt': 'ir-Repubblika Ċentru-Afrikana', 'mua': 'centrafrika', 'my': 'ဗဟို အာဖရိက ပြည်ထောင်စု', 'mzn': 'مرکزی آفریقای جمهوری', 'naq': 'Central African Republiki', 'nb': 'Den sentralafrikanske republikk', 'nd': 'Central African Republic', 'ne': 'मध्य अफ्रिकी गणतन्त्र', 'nl': 'Centraal-Afrikaanse Republiek', 'nmg': 'Sentrafríka', 'nn': 'Den sentralafrikanske republikken', 'no': 'Den sentralafrikanske republikk', 'nqo': 'ߕߊ߲ߓߊ߲-ߝߘߊ߬ߝߌ߲߬ߠߊ߫ ߞߊ߲ߓߍ߲', 'nus': 'Cɛntrɔl aprika repuɔblic', 'nyn': 'Eihanga rya Rwagati ya Afirika', 'om': 'Rippaablika Afrikaa Gidduugaleessaa', 'or': 'ମଧ୍ୟ ଆଫ୍ରିକୀୟ ସାଧାରଣତନ୍ତ୍ର', 'pa': 'ਕੇਂਦਰੀ ਅਫ਼ਰੀਕੀ ਗਣਰਾਜ', 'pa-Guru': 'ਕੇਂਦਰੀ ਅਫ਼ਰੀਕੀ ਗਣਰਾਜ', 'pcm': 'Sẹ́ntrál Áfríkán Ripọ́blik', 'pl': 'Republika Środkowoafrykańska', 'ps': 'وسطي افريقا جمهور', 'pt': 'República Centro-Africana', 'qu': 'República Centroafricana', 'rm': 'Republica Centralafricana', 'rn': 'Repubulika ya Santarafurika', 'ro': 'Republica Centrafricană', 'rof': 'Jamhuri ya Afrika ya Kati', 'ru': 'Центрально-Африканская Республика', 'rwk': 'Jamhuri ya Afrika ya Kati', 'saq': 'Jamhuri ya Afrika ya Kati', 'sbp': 'Jamuhuli ya Afilika ya Pakhati', 'sc': 'Repùblica Tzentrafricana', 'sd': 'وچ آفريقي جمهوريه', 'sd-Arab': 'وچ آفريقي جمهوريه', 'se': 'Gaska-Afrihká dásseváldi', 'seh': 'República Centro-Africana', 'ses': 'Centraafriki koyra', 'sg': 'Ködörösêse tî Bêafrîka', 'shi': 'ⵜⴰⴳⴷⵓⴷⴰⵏⵜ ⵜⴰⵏⴰⵎⵎⴰⵙⵜ ⵏ ⵉⴼⵔⵉⵇⵢⴰ', 'shi-Latn': 'tagdudant tanammast n ifriqya', 'shi-Tfng': 'ⵜⴰⴳⴷⵓⴷⴰⵏⵜ ⵜⴰⵏⴰⵎⵎⴰⵙⵜ ⵏ ⵉⴼⵔⵉⵇⵢⴰ', 'si': 'මධ්\u200dයම අප්\u200dරිකානු ජනරජය', 'sk': 'Stredoafrická republika', 'sl': 'Srednjeafriška republika', 'smn': 'Koskâ-Afrika täsiväldi', 'sn': 'Central African Republic', 'so': 'Jamhuuriyadda Afrikada Dhexe', 'sq': 'Republika e Afrikës Qendrore', 'sr': 'Централноафричка Република', 'sr-Cyrl': 'Централноафричка Република', 'sr-Latn': 'Centralnoafrička Republika', 'sv': 'Centralafrikanska republiken', 'sw': 'Jamhuri ya Afrika ya Kati', 'syr': 'ܩܘܛܢܝܘܬܐ ܕܐܦܪܝܩܐ ܡܨܥܝܬܐ', 'ta': 'மத்திய ஆப்ரிக்கக் குடியரசு', 'te': 'సెంట్రల్ ఆఫ్రికన్ రిపబ్లిక్', 'teo': 'Jamhuri ya Afrika ya Kati', 'tg': 'Ҷумҳурии Африқои Марказӣ', 'th': 'สาธารณรัฐแอฟริกากลาง', 'ti': 'ሪፓብሊክ ማእከላይ ኣፍሪቃ', 'tk': 'Merkezi Afrika Respublikasy', 'to': 'Lepupelika ʻAfilika Lotoloto', 'tr': 'Orta Afrika Cumhuriyeti', 'tt': 'Үзәк Африка Республикасы', 'twq': 'Centraafriki koyra', 'tzm': 'Tagduda n Afrika Wammas', 'ug': 'ئوتتۇرا ئافرىقا جۇمھۇرىيىتى', 'uk': 'Центральноафриканська Республіка', 'ur': 'وسط افریقی جمہوریہ', 'uz': 'Markaziy Afrika Respublikasi', 'uz-Cyrl': 'Марказий Африка Республикаси', 'uz-Latn': 'Markaziy Afrika Respublikasi', 'vai': 'ꕉꔱꔸꕪ ꗳ ꗳ ꕸꖃꔀ', 'vai-Latn': 'Áfíríka Lumaã Tɛ Boloe', 'vai-Vaii': 'ꕉꔱꔸꕪ ꗳ ꗳ ꕸꖃꔀ', 'vec': 'Repùblega Sentrafregana', 'vi': 'Cộng hòa Trung Phi', 'vun': 'Jamhuri ya Afrika ya Kati', 'wae': 'Zentralafrikaniši Rebublik', 'wo': 'Repiblik Sàntar Afrik', 'xh': 'ECentral African Republic', 'xnr': 'बिचला अफ़्रीकी गणराज्य', 'xog': 'Lipabulika ya Senturafiriki', 'yav': 'Santalafilíik', 'yi': 'צענטראַל־אַפֿריקאַנישע רעפּובליק', 'yo': 'Àrin gùngun Áfíríkà', 'yrl': 'Repubirika Afirika-Piterapura', 'yue': '中非共和國', 'yue-Hans': '中非共和国', 'yue-Hant': '中非共和國', 'zgh': 'ⵜⴰⴳⴷⵓⴷⴰⵏⵜ ⵜⴰⵏⴰⵎⵎⴰⵙⵜ ⵏ ⵉⴼⵔⵉⵇⵢⴰ', 'zh': '中非共和国', 'zh-Hans': '中非共和国', 'zh-Hant': '中非共和國', 'zu': 'i-Central African Republic'},
    'CG': {'af': 'Kongo-Brazzaville', 'agq': 'Kuŋgù', 'ak': 'Kongo', 'am': 'ኮንጎ ብራዛቪል', 'ar': 'الكونغو - برازافيل', 'as': 'কঙ্গো - ব্রাজাভিল', 'asa': 'Kongo', 'ast': 'Congu - Brazzaville', 'az': 'Konqo - Brazzavil', 'az-Cyrl': 'Конго-Браззавил', 'az-Latn': 'Konqo - Brazzavil', 'bal-Latn': 'Kángó-Brázáwille', 'bas': 'Kòŋgo', 'be': 'Конга - Бразавіль', 'be-tarask': 'Конга - Бразавіль', 'bez': 'Hukongo', 'bg': 'Конго (Бразавил)', 'blo': 'Koŋgoo Brasafil', 'bm': 'Kongo', 'bn': 'কঙ্গো - ব্রাজাভিল', 'br': 'Kongo - Brazzaville', 'brx': 'कंग’- ब्राज्जाभील', 'bs': 'Kongo', 'bs-Cyrl': 'Конго', 'bs-Latn': 'Kongo', 'ca': 'Congo - Brazzaville', 'ca-ES-valencia': 'Congo - Brazzaville', 'ccp': '𑄇𑄧𑄋𑄴𑄉𑄮-𑄝𑄳𑄢𑄎𑄞𑄨𑄣𑄴', 'ce': 'Конго - Браззавиль', 'ceb': 'Congo - Brazzaville', 'cgg': 'Kongo', 'chr': 'ᎧᏂᎪ - ᏆᏌᏩᎵ', 'ckb': 'کۆنگۆ برازاڤیل', 'cs': 'Kongo – Brazzaville', 'cv': 'Конго - Браззавиль', 'cy': 'Y Congo - Brazzaville', 'da': 'Congo-Brazzaville', 'dav': 'Kongo', 'de': 'Kongo-Brazzaville', 'dje': 'Kongoo', 'dsb': 'Kongo-Brazzaville', 'dyo': 'Kongo', 'dz': 'ཀོང་གྷོ བྷྲ་ཛ་བིལ', 'ebu': 'Kongo', 'ee': 'Kongo Brazzaville nutome', 'el': 'Κονγκό - Μπραζαβίλ', 'el-polyton': 'Κονγκό - Μπραζαβίλ', 'en': 'Congo - Brazzaville', 'en-Dsrt': '𐐗𐐪𐑍𐑀𐐬 - 𐐒𐑉𐐪𐑆𐐲𐑂𐐮𐑊', 'eo': 'Kongo Brazavila', 'es': 'Congo', 'et': 'Kongo Vabariik', 'eu': 'Kongo', 'ewo': 'Kongó', 'fa': 'کنگو - برازویل', 'ff': 'Konngo', 'ff-Adlm': '𞤑𞤮𞤲𞤺𞤮 - 𞤄𞤪𞤢𞥁𞤢𞤾𞤭𞤤', 'ff-Latn': 'Konngo', 'fi': 'Kongon tasavalta', 'fil': 'Congo - Brazzaville', 'fo': 'Kongo', 'fr': 'Congo-Brazzaville', 'fur': 'Congo - Brazzaville', 'fy': 'Congo-Brazzaville', 'ga': 'Congó-Brazzaville', 'gd': 'A’ Chongo - Brazzaville', 'gl': 'República do Congo', 'gsw': 'Kongo', 'gu': 'કોંગો - બ્રાઝાવિલે', 'guz': 'Kongo', 'ha': 'Kongo', 'he': 'קונגו - ברזאויל', 'hi': 'कांगो – ब्राज़ाविल', 'hi-Latn': 'Congo - Brazzaville', 'hr': 'Kongo - Brazzaville', 'hsb': 'Kongo-Brazzaville', 'ht': 'Congo-Brazzaville', 'hu': 'Kongó – Brazzaville', 'hy': 'Կոնգո - Բրազավիլ', 'ia': 'Congo - Brazzaville', 'id': 'Kongo - Brazzaville', 'ig': 'Congo', 'is': 'Kongó-Brazzaville', 'it': 'Congo-Brazzaville', 'ja': 'コンゴ共和国(ブラザビル)', 'jgo': 'Kɔ́ŋgɔ-Mbɛlazavîl', 'jmc': 'Kongo', 'jv': 'Kongo - Brassaville', 'ka': 'კონგო - ბრაზავილი', 'kaa': 'Конго - Браззавиль', 'kaa-Cyrl': 'Конго - Браззавиль', 'kab': 'Kungu', 'kam': 'Kongo', 'kde': 'Kongo', 'kea': 'Kongu - Brazavili', 'kgp': 'Kãgo Repumrika', 'khq': 'Kongoo', 'ki': 'Kongo', 'kk': 'Конго-Браззавиль Республикасы', 'kk-Cyrl': 'Конго-Браззавиль Республикасы', 'kln': 'Emetab Congo - Brazzaville', 'km': 'កុងហ្គោ - ប្រាហ្សាវីល', 'kn': 'ಕಾಂಗೋ - ಬ್ರಾಜಾವಿಲ್ಲೇ', 'ko': '콩고-브라자빌', 'kok': 'कोंगो - ब्राझाविला', 'kok-Deva': 'कोंगो - ब्राझाविला', 'ks': 'کونگو بٔرزاوِلی', 'ks-Arab': 'کونگو بٔرزاوِلی', 'ksb': 'Kongo', 'ksf': 'kɔngó', 'ksh': 'dä Konggo (Brassavill)', 'ku': 'Kongo - Brazzaville', 'kxv': 'kongo - brajavil', 'kxv-Deva': 'कांगो – ब्राज़ाविल', 'kxv-Latn': 'kongo - brajavil', 'kxv-Orya': 'କଂଙ୍ଗ- ବ୍ରାଜାୱିଲି', 'kxv-Telu': 'కాంగో- బ్రాజావిల్లి', 'ky': 'Конго-Браззавил', 'lag': 'Kóongo', 'lb': 'Kongo-Brazzaville', 'lg': 'Kongo', 'lld': 'Congo - Brazzaville', 'ln': 'Kongo', 'lo': 'ຄອງໂກ - ບຣາຊາວິວ', 'lt': 'Kongas-Brazavilis', 'lu': 'Kongu', 'luo': 'Congo', 'luy': 'Kongo', 'lv': 'Kongo (Brazavila)', 'mai': 'कांगो-ब्राजाबिले', 'mas': 'Kongo', 'mer': 'Kongo', 'mfe': 'Kongo', 'mg': 'Kôngô', 'mgh': 'Ukongo', 'mi': 'Kōngo - Pārawhe', 'mk': 'Конго - Бразавил', 'ml': 'കോംഗോ - ബ്രാസവില്ലി', 'mn': 'Конго-Браззавиль', 'mr': 'काँगो - ब्राझाविले', 'ms': 'Congo - Brazzaville', 'mt': 'il-Kongo - Brazzaville', 'mua': 'kongo', 'my': 'ကွန်ဂို-ဘရာဇာဗီးလ်', 'mzn': 'کنگو برازاویل', 'naq': 'Congob', 'nb': 'Kongo-Brazzaville', 'nd': 'Khongo', 'ne': 'कङ्गो ब्राजाभिल', 'nl': 'Congo-Brazzaville', 'nmg': 'Kongo', 'nn': 'Kongo-Brazzaville', 'no': 'Kongo-Brazzaville', 'nqo': 'ߞߏ߲߬ߜ߭ߏ߫-ߓߙߊߖ߭ߊ߫', 'nus': 'Kɔŋgɔ', 'nyn': 'Kongo', 'om': 'Koongoo - Biraazaavil', 'or': 'କଙ୍ଗୋ-ବ୍ରାଜିଭିଲ୍ଲେ', 'pa': 'ਕਾਂਗੋ - ਬ੍ਰਾਜ਼ਾਵਿਲੇ', 'pa-Guru': 'ਕਾਂਗੋ - ਬ੍ਰਾਜ਼ਾਵਿਲੇ', 'pcm': 'Kóngo – Brázavil', 'pl': 'Kongo', 'ps': 'کانګو - بروزوییل', 'pt': 'República do Congo', 'qu': 'Congo', 'rm': 'Congo', 'rn': 'Kongo', 'ro': 'Congo - Brazzaville', 'rof': 'Kongo', 'ru': 'Конго - Браззавиль', 'rwk': 'Kongo', 'saq': 'Kongo', 'sbp': 'Kongo', 'sc': 'Congo - Bratzaville', 'sd': 'ڪانگو - برازاویل', 'sd-Arab': 'ڪانگو - برازاویل', 'se': 'Kongo-Brazzaville', 'seh': 'Congo', 'ses': 'Kongoo', 'sg': 'Kongö', 'shi': 'ⴽⵓⵏⴳⵓ', 'shi-Latn': 'kungu', 'shi-Tfng': 'ⴽⵓⵏⴳⵓ', 'si': 'කොංගො - බ්\u200dරසාවිල්', 'sk': 'Konžská republika', 'sl': 'Kongo - Brazzaville', 'smn': 'Kongo täsiväldi', 'sn': 'Kongo', 'so': 'Kongo', 'sq': 'Kongo-Brazavilë', 'sr': 'Конго - Бразавил', 'sr-Cyrl': 'Конго - Бразавил', 'sr-Latn': 'Kongo - Brazavil', 'sv': 'Kongo-Brazzaville', 'sw': 'Kongo - Brazzaville', 'syr': 'ܟܘܢܓܘ - ܒܪܐܙܐܒܝܠ', 'ta': 'காங்கோ - ப்ராஸாவில்லே', 'te': 'కాంగో- బ్రాజావిల్లి', 'teo': 'Kongo', 'tg': 'Конго - Браззавил', 'th': 'คองโก - บราซซาวิล', 'ti': 'ኮንጎ', 'tk': 'Kongo - Brazzawil', 'to': 'Kongo - Palasavila', 'tr': 'Kongo - Brazavil', 'tt': 'Конго - Браззавиль', 'twq': 'Kongoo', 'tzm': 'Kungu', 'ug': 'كونگو - بىراززاۋىل', 'uk': 'Конго – Браззавіль', 'ur': 'کانگو - برازاویلے', 'uz': 'Kongo – Brazzavil', 'uz-Cyrl': 'Конго Браззавиль', 'uz-Latn': 'Kongo – Brazzavil', 'vai': 'ꖏꖐ', 'vai-Latn': 'Kóngo', 'vai-Vaii': 'ꖏꖐ', 'vec': 'Repùblega de’l Congo', 'vi': 'Congo - Brazzaville', 'vun': 'Kongo', 'wae': 'Kongo Brazzaville', 'wo': 'Réewum Kongo', 'xh': 'ECongo - Brazzaville', 'xnr': 'कांगो – ब्राज़ाविल', 'xog': 'Kongo', 'yav': 'Kongó', 'yo': 'Kóńgò – Brazaville', 'yrl': 'Repubirika Kũguyara', 'yue': '剛果（布拉薩）', 'yue-Hans': '刚果（布拉萨）', 'yue-Hant': '剛果（布拉薩）', 'zgh': 'ⴽⵓⵏⴳⵓ', 'zh': '刚果（布）', 'zh-Hans': '刚果（布）', 'zh-Hant': '剛果（布拉薩）', 'zu': 'i-Congo - Brazzaville'},
    'CH': {'af': 'Switserland', 'agq': 'Suezàlân', 'ak': 'Swetzaland', 'am': 'ስዊዘርላንድ', 'ar': 'سويسرا', 'as': 'চুইজাৰলেণ্ড', 'asa': 'Uthwithi', 'ast': 'Suiza', 'az': 'İsveçrə', 'az-Cyrl': 'Исвечрә', 'az-Latn': 'İsveçrə', 'bal-Latn': 'Swizarlánd', 'bas': 'Sùwîs', 'be': 'Швейцарыя', 'be-tarask': 'Швейцарыя', 'bez': 'Huuswisi', 'bg': 'Швейцария', 'blo': 'Suwis', 'bm': 'Suwisi', 'bn': 'সুইজারল্যান্ড', 'br': 'Suis', 'brx': 'सुईजारलेण्ड', 'bs': 'Švicarska', 'bs-Cyrl': 'Швицарска', 'bs-Latn': 'Švicarska', 'ca': 'Suïssa', 'ca-ES-valencia': 'Suïssa', 'ccp': '𑄥𑄭𑄪𑄎𑄢𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴', 'ce': 'Швейцари', 'ceb': 'Switzerland', 'cgg': 'Swisi', 'chr': 'ᏍᏫᏍ', 'ckb': 'سویسڕا', 'cs': 'Švýcarsko', 'cv': 'Швейцари', 'cy': 'Y Swistir', 'da': 'Schweiz', 'dav': 'Uswisi', 'de': 'Schweiz', 'dje': 'Swisu', 'dsb': 'Šwicarska', 'dz': 'སུ་ཝིཊ་ཛར་ལེནཌ', 'ebu': 'Uswisi', 'ee': 'Switzerland nutome', 'el': 'Ελβετία', 'el-polyton': 'Ἑλβετία', 'en': 'Switzerland', 'en-Dsrt': '𐐝𐐶𐐮𐐻𐑅𐐲𐑉𐑊𐐲𐑌𐐼', 'eo': 'Svisujo', 'es': 'Suiza', 'et': 'Šveits', 'eu': 'Suitza', 'ewo': 'Suís', 'fa': 'سوئیس', 'ff': 'Suwiis', 'ff-Adlm': '𞤅𞤵𞤱𞤭𞤪𞤧𞤢𞥄', 'ff-Latn': 'Suwiis', 'fi': 'Sveitsi', 'fil': 'Switzerland', 'fo': 'Sveis', 'fr': 'Suisse', 'fur': 'Svuizare', 'fy': 'Switserlân', 'ga': 'an Eilvéis', 'gd': 'An Eilbheis', 'gl': 'Suíza', 'gsw': 'Schwiiz', 'gu': 'સ્વિટ્ઝર્લૅન્ડ', 'guz': 'Uswisi', 'ha': 'Suwizalan', 'he': 'שווייץ', 'hi': 'स्विट्ज़रलैंड', 'hi-Latn': 'Switzerland', 'hr': 'Švicarska', 'hsb': 'Šwicarska', 'ht': 'Suisse', 'hu': 'Svájc', 'hy': 'Շվեյցարիա', 'ia': 'Suissa', 'id': 'Swiss', 'ie': 'Svissia', 'ig': 'Switzerland', 'is': 'Sviss', 'it': 'Svizzera', 'ja': 'スイス', 'jgo': 'Sẅísɛ', 'jmc': 'Uswisi', 'jv': 'Switserlan', 'ka': 'შვეიცარია', 'kaa': 'Швейцария', 'kaa-Cyrl': 'Швейцария', 'kab': 'Swis', 'kam': 'Uswisi', 'kde': 'Uswisi', 'kea': 'Suisa', 'kgp': 'Suvisa', 'khq': 'Swisu', 'ki': 'Uswisi', 'kk': 'Швейцария', 'kk-Cyrl': 'Швейцария', 'kln': 'Emetab Switzerland', 'km': 'ស្វ៊ីស', 'kn': 'ಸ್ವಿಟ್ಜರ್ಲ್ಯಾಂಡ್', 'ko': '스위스', 'kok': 'स्विट्ज़रलैंड', 'kok-Deva': 'स्विट्ज़रलैंड', 'ks': 'سُوِزَرلینڑ', 'ks-Arab': 'سُوِزَرلینڑ', 'ksb': 'Uswisi', 'ksf': 'swís', 'ksh': 'de Schweiz', 'ku': 'Swîsre', 'kxv': 'svitjerlaṇḍ', 'kxv-Deva': 'स्विट्ज़रलेंड', 'kxv-Latn': 'svitjerlaṇḍ', 'kxv-Orya': 'ସ୍ୱୀଜଅର୍\u200dଲାଣ୍ଡ', 'kxv-Telu': 'స్విజర్లాండ్', 'ky': 'Швейцария', 'lag': 'Uswíisi', 'lb': 'Schwäiz', 'lg': 'Switizirandi', 'lld': 'Svizera', 'ln': 'Swisɛ', 'lo': 'ສະວິດເຊີແລນ', 'lt': 'Šveicarija', 'lu': 'Swise', 'luo': 'Switzerland', 'luy': 'Uswisi', 'lv': 'Šveice', 'mai': 'स्विट्जरलैंड', 'mas': 'Uswisi', 'mer': 'Swizilandi', 'mfe': 'Laswis', 'mg': 'Soisa', 'mgh': 'Uswisi', 'mi': 'Huiterangi', 'mk': 'Швајцарија', 'ml': 'സ്വിറ്റ്സർലാൻഡ്', 'mn': 'Швейцар', 'mr': 'स्वित्झर्लंड', 'ms': 'Switzerland', 'mt': 'l-Iżvizzera', 'mua': 'Sǝr Swiss', 'my': 'ဆွစ်ဇာလန်', 'mzn': 'سوییس', 'naq': 'Switzerlandi', 'nb': 'Sveits', 'nd': 'Switzerland', 'ne': 'स्विजरल्याण्ड', 'nl': 'Zwitserland', 'nmg': 'Switzɛrland', 'nn': 'Sveits', 'no': 'Sveits', 'nqo': 'ߛߎߥߌߛ', 'nyn': 'Swisi', 'om': 'Siwizerlaand', 'or': 'ସ୍ୱିଜରଲ୍ୟାଣ୍ଡ', 'pa': 'ਸਵਿਟਜ਼ਰਲੈਂਡ', 'pa-Guru': 'ਸਵਿਟਜ਼ਰਲੈਂਡ', 'pcm': 'Swítsaland', 'pl': 'Szwajcaria', 'ps': 'سویس', 'pt': 'Suíça', 'qu': 'Suiza', 'rm': 'Svizra', 'rn': 'Ubusuwisi', 'ro': 'Elveția', 'rof': 'Uswisi', 'ru': 'Швейцария', 'rwk': 'Uswisi', 'saq': 'Uswisi', 'sbp': 'Uswisi', 'sc': 'Isvìtzera', 'sd': 'سوئزرلينڊ', 'sd-Arab': 'سوئزرلينڊ', 'se': 'Šveica', 'seh': 'Suíça', 'ses': 'Swisu', 'sg': 'Sûîsi', 'shi': 'ⵙⵡⵉⵙⵔⴰ', 'shi-Latn': 'swisra', 'shi-Tfng': 'ⵙⵡⵉⵙⵔⴰ', 'si': 'ස්විස්ටර්ලන්තය', 'sk': 'Švajčiarsko', 'sl': 'Švica', 'smn': 'Sveitsi', 'sn': 'Switzerland', 'so': 'Swiiserlaand', 'sq': 'Zvicër', 'sr': 'Швајцарска', 'sr-Cyrl': 'Швајцарска', 'sr-Latn': 'Švajcarska', 'sv': 'Schweiz', 'sw': 'Uswisi', 'syr': 'ܣܘܝܣܪܐ', 'ta': 'ஸ்விட்சர்லாந்து', 'te': 'స్విట్జర్లాండ్', 'teo': 'Uswisi', 'tg': 'Швейтсария', 'th': 'สวิตเซอร์แลนด์', 'ti': 'ስዊዘርላንድ', 'tk': 'Şweýsariýa', 'to': 'Suisilani', 'tr': 'İsviçre', 'tt': 'Швейцария', 'twq': 'Swisu', 'tzm': 'Swisra', 'ug': 'شىۋېتسارىيە', 'uk': 'Швейцарія', 'ur': 'سوئٹزر لینڈ', 'uz': 'Shveytsariya', 'uz-Cyrl': 'Швейцария', 'uz-Latn': 'Shveytsariya', 'vai': 'ꖬꔃꕤ ꖨꕮꕊ', 'vai-Latn': 'Suweza Lumaã', 'vai-Vaii': 'ꖬꔃꕤ ꖨꕮꕊ', 'vec': 'Zvìsara', 'vi': 'Thụy Sĩ', 'vun': 'Uswisi', 'wae': 'Schwiz', 'wo': 'Siwis', 'xh': 'ESwitzerland', 'xnr': 'स्विट्ज़रलैंड', 'xog': 'Switizirandi', 'yav': 'suwíis', 'yi': 'שווייץ', 'yo': 'switiṣilandi', 'yrl': 'Suwisa', 'yue': '瑞士', 'yue-Hans': '瑞士', 'yue-Hant': '瑞士', 'zgh': 'ⵙⵡⵉⵙⵔⴰ', 'zh': '瑞士', 'zh-Hans': '瑞士', 'zh-Hant': '瑞士', 'zu': 'i-Switzerland'},
    'CI': {'af': 'Ivoorkus', 'agq': 'Ku Dɨ̀vûa', 'ak': 'Kodivuwa', 'am': 'ኮትዲቯር', 'ar': 'ساحل العاج', 'as': 'কোটে ডি আইভৰ', 'asa': 'Kodivaa', 'ast': 'Costa de Marfil', 'az': 'Kotd’ivuar', 'az-Cyrl': 'Котд’ивуар', 'az-Latn': 'Kotd’ivuar', 'bal-Latn': 'Kóté diáwóri', 'bas': 'Màŋ mi Njɔ̂k', 'be': 'Кот-д’Івуар', 'be-tarask': 'Кот-д’Івуар', 'bez': 'Hukodivaa', 'bg': 'Кот д’Ивоар', 'blo': 'Koɖifʊaa', 'bm': 'Kodiwari', 'bn': 'কোট ডি‘আইভোর', 'br': 'Aod an Olifant', 'brx': 'कट दिभवेर', 'bs': 'Obala Slonovače', 'bs-Cyrl': 'Обала Слоноваче (Кот д’Ивоар)', 'bs-Latn': 'Obala Slonovače', 'ca': 'Côte d’Ivoire', 'ca-ES-valencia': 'Côte d’Ivoire', 'ccp': '𑄃𑄭𑄞𑄧𑄢𑄨 𑄇𑄮𑄌𑄴𑄑𑄴', 'ce': 'Кот-Д’ивуар', 'ceb': 'Cote d’Ivoire', 'cgg': 'Aivore Kositi', 'chr': 'ᎢᏬᎵ ᎾᎿ ᎠᎹᏳᎶᏗ', 'ckb': 'کۆتدیڤوار', 'cs': 'Pobřeží slonoviny', 'cv': 'Кот-д’Ивуар', 'cy': 'Côte d’Ivoire', 'da': 'Elfenbenskysten', 'dav': 'Kodivaa', 'de': 'Côte d’Ivoire', 'dje': 'Kudwar', 'dsb': 'Côte d’Ivoire', 'dyo': 'Koddiwar', 'dz': 'ཀོ་ཊེ་ ཌི་ཨི་ཝོ་རེ', 'ebu': 'Kodivaa', 'ee': 'Kote d’Ivoire nutome', 'el': 'Ακτή Ελεφαντοστού', 'el-polyton': 'Ακτή Ελεφαντοστού', 'en': 'Côte d’Ivoire', 'en-Dsrt': '𐐌𐑂𐑉𐐨 𐐗𐐬𐑅𐐻', 'eo': 'Ebur-Bordo', 'es': 'Côte d’Ivoire', 'et': 'Elevandiluurannik', 'eu': 'Boli Kosta', 'ewo': 'Kód Divɔ́r', 'fa': 'ساحل عاج', 'ff': 'Kodduwaar', 'ff-Adlm': '𞤑𞤮𞤼𞤣𞤭𞤾𞤢𞥄𞤪', 'ff-Latn': 'Kodduwaar', 'fi': 'Norsunluurannikko', 'fil': 'Côte d’Ivoire', 'fo': 'Fílabeinsstrondin', 'fr': 'Côte d’Ivoire', 'fur': 'Cueste di Avoli', 'fy': 'Ivoorkust', 'ga': 'An Cósta Eabhair', 'gd': 'Côte d’Ivoire', 'gl': 'Côte d’Ivoire', 'gsw': 'Elfebäiküschte', 'gu': 'કોટ ડીઆઇવરી', 'guz': 'Kodivaa', 'ha': 'Aibari Kwas', 'he': 'חוף השנהב', 'hi': 'कोत दिवुआर', 'hi-Latn': 'Cote d’Ivoire', 'hr': 'Obala Bjelokosti', 'hsb': 'Côte d’Ivoire', 'ht': 'Côte d’Ivoire', 'hu': 'Elefántcsontpart', 'hy': 'Կոտ դ՚Իվուար', 'ia': 'Costa de Ebore', 'id': 'Côte d’Ivoire', 'ig': 'Côte d’Ivoire', 'is': 'Fílabeinsströndin', 'it': 'Costa d’Avorio', 'ja': 'コートジボワール', 'jgo': 'Kɔ́t Ndivwâ', 'jmc': 'Kodivaa', 'jv': 'Pasisir Gadhing', 'ka': 'კოტ-დივუარი', 'kaa': 'Кот-дИвуар', 'kaa-Cyrl': 'Кот-дИвуар', 'kab': 'Kuṭ Divwar', 'kam': 'Kodivaa', 'kde': 'Kodivaa', 'kea': 'Kosta di Marfin', 'kgp': 'Jãn-mág-kupri Fyr', 'khq': 'Kudwar', 'ki': 'Kodivaa', 'kk': 'Кот-д’Ивуар', 'kk-Cyrl': 'Кот-д’Ивуар', 'kln': 'Emetab Côte d’Ivoire', 'km': 'កូតឌីវ័រ', 'kn': 'ಕೋತ್ ದ್‘ಇವಾರ್', 'ko': '코트디부아르', 'kok': 'कोत द’ईवोआर', 'kok-Deva': 'कोत द’ईवोआर', 'ks': 'کوٹ ڈلوائر', 'ks-Arab': 'کوٹ ڈلوائر', 'ksb': 'Kodivaa', 'ksf': 'kɔtiwuár', 'ksh': 'Älfebeijn-Köß', 'ku': 'Côte d’Ivoire', 'kxv': 'koṭ ḍi vaa', 'kxv-Deva': 'कोट डी वोआ', 'kxv-Latn': 'koṭ ḍi vaa', 'kxv-Orya': 'କଟ ଡି ୱା', 'kxv-Telu': 'కోట్ డి ఐవోర్', 'ky': 'Кот-д’Ивуар', 'lag': 'Ivori Kositi', 'lb': 'Côte d’Ivoire', 'lg': 'Kote Divwa', 'lld': 'Costa d’Avore', 'ln': 'Kotídivualɛ', 'lo': 'ໂຄຕີ ວົວ', 'lt': 'Dramblio Kaulo Krantas', 'lu': 'Kotedivuale', 'luo': 'Côte d', 'luy': 'Kodivaa', 'lv': 'Kotdivuāra', 'mai': 'कोट डी आइवर', 'mas': 'Kodivaa', 'mer': 'Kodivaa', 'mfe': 'Côte d’Ivoire', 'mg': 'Côte d’Ivoire', 'mgh': 'Ukodiva', 'mi': 'Te Tai Rei', 'mk': 'Брегот на Слоновата Коска', 'ml': 'കോട്ട് ഡി വാർ', 'mn': 'Кот-д’Ивуар', 'mr': 'कोत द’ईवोआर', 'ms': 'Cote d’Ivoire', 'mt': 'il-Kosta tal-Avorju', 'mua': 'ser Ivoiriya', 'my': 'ကို့တ် ဒီဗွာ', 'mzn': 'عاج ِساحل', 'naq': 'Ivoorkusi', 'nb': 'Elfenbenskysten', 'nd': 'Ivory Coast', 'ne': 'कोट दिभोर', 'nl': 'Ivoorkust', 'nmg': 'Kote d´Ivoire', 'nn': 'Elfenbeinskysten', 'no': 'Elfenbenskysten', 'nqo': 'ߜߋ߲-ߞߐ߰ߖߌ߬ߘߊ', 'nus': 'Kodibo̱o̱', 'nyn': 'Aivore Kositi', 'om': 'Koti divoor', 'or': 'କୋତ୍ ଡି ଭ୍ଵାର୍', 'pa': 'ਕੋਟ ਡੀਵੋਆਰ', 'pa-Guru': 'ਕੋਟ ਡੀਵੋਆਰ', 'pcm': 'Aívri Kost', 'pl': 'Côte d’Ivoire', 'ps': 'د عاج ساحل', 'pt': 'Costa do Marfim', 'qu': 'Côte d’Ivoire', 'rm': 'Costa d’Ivur', 'rn': 'Kotedivuware', 'ro': 'Côte d’Ivoire', 'rof': 'Kodivaa', 'ru': 'Кот-д’Ивуар', 'rwk': 'Kodivaa', 'saq': 'Kodivaa', 'sbp': 'Kodivaya', 'sc': 'Costa de Avòriu', 'sd': 'ڪوٽي ويرا', 'sd-Arab': 'ڪوٽي ويرا', 'se': 'Elfenbenariddu', 'seh': 'Costa do Marfim', 'ses': 'Kudwar', 'sg': 'Kôdivüära', 'shi': 'ⴽⵓⵜ ⴷⵉⴼⵡⴰⵔ', 'shi-Latn': 'kut difwar', 'shi-Tfng': 'ⴽⵓⵜ ⴷⵉⴼⵡⴰⵔ', 'si': 'කෝට් දි අයිවරි', 'sk': 'Pobrežie Slonoviny', 'sl': 'Slonokoščena obala', 'smn': 'Côte d’Ivoire', 'sn': 'Ivory Coast', 'so': 'Ayfori Koost', 'sq': 'Côte d’Ivoire', 'sr': 'Обала Слоноваче (Кот д’Ивоар)', 'sr-Cyrl': 'Обала Слоноваче (Кот д’Ивоар)', 'sr-Latn': 'Obala Slonovače (Kot d’Ivoar)', 'sv': 'Elfenbenskusten', 'sw': 'Cote d’Ivoire', 'syr': 'ܩܘܛ ܕܝܒܘܐܪ', 'ta': 'கோட் தி’வாயர்', 'te': 'కోట్ డి ఐవోర్', 'teo': 'Kodivaa', 'tg': 'Кот-д’Ивуар', 'th': 'โกตดิวัวร์', 'ti': 'ኮት ዲቭዋር', 'tk': 'Kot-d’Iwuar', 'to': 'Matafonua ʻAivolī', 'tr': 'Côte d’Ivoire', 'tt': 'Кот-д’Ивуар', 'twq': 'Kudwar', 'tzm': 'Taɣazut n Uszer', 'ug': 'كوتې دې ئىۋوئىر', 'uk': 'Кот-дʼІвуар', 'ur': 'کوٹ ڈی آئیوری', 'uz': 'Kot-d’Ivuar', 'uz-Cyrl': 'Кот-д’Ивуар', 'uz-Latn': 'Kot-d’Ivuar', 'vai': 'ꖏꔳ ꕾꕎ', 'vai-Latn': 'Kódivówa', 'vai-Vaii': 'ꖏꔳ ꕾꕎ', 'vec': 'Costa d’Avorio', 'vi': 'Côte d’Ivoire', 'vun': 'Kodivaa', 'wae': 'Elfebeiküšta', 'wo': 'Kodiwaar', 'xh': 'ECôte d’Ivoire', 'xnr': 'कोट डी आइवर', 'xog': 'Kote Divwa', 'yav': 'Kótifualɛ', 'yi': 'העלפֿאַ נדביין בארטן', 'yo': 'Kóútè forà', 'yrl': 'Mafim Kupé', 'yue': '象牙海岸', 'yue-Hans': '象牙海岸', 'yue-Hant': '象牙海岸', 'zgh': 'ⴽⵓⵜ ⴷⵉⴼⵡⴰⵔ', 'zh': '科特迪瓦', 'zh-Hans': '科特迪瓦', 'zh-Hant': '象牙海岸', 'zu': 'i-Côte d’Ivoire'},
    'CK': {'af': 'Cookeilande', 'agq': 'Chwɨla ŋ̀ Kûʔ', 'ak': 'Kuk Nsupɔ', 'am': 'ኩክ ደሴቶች', 'ar': 'جزر كوك', 'as': 'কুক দ্বীপপুঞ্জ', 'asa': 'Vithiwa vya Cook', 'ast': 'Islles Cook', 'az': 'Kuk adaları', 'az-Cyrl': 'Кук адалары', 'az-Latn': 'Kuk adaları', 'bal-Latn': 'Kuk Islánd', 'bas': 'Bìòn bi Kook', 'be': 'Астравы Кука', 'be-tarask': 'Астравы Кука', 'bez': 'Ifisima fya Kook', 'bg': 'острови Кук', 'blo': 'Kʊkʊ kaBʊtukǝltǝna', 'bm': 'Kuki Gun', 'bn': 'কুক দ্বীপপুঞ্জ', 'br': 'Inizi Cook', 'brx': 'कुक द्वीप', 'bs': 'Kukova ostrva', 'bs-Cyrl': 'Кукова Острва', 'bs-Latn': 'Kukova ostrva', 'ca': 'Illes Cook', 'ca-ES-valencia': 'Illes Cook', 'ccp': '𑄇𑄪𑄇𑄪 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Кукан гӀайренаш', 'ceb': 'Cook Islands', 'cgg': 'Ebizinga bya Kuuku', 'chr': 'ᎠᏓᏍᏓᏴᎲᏍᎩ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی کوک', 'cs': 'Cookovy ostrovy', 'cv': 'Кук утравӗсем', 'cy': 'Ynysoedd Cook', 'da': 'Cookøerne', 'dav': 'Visiwa vya Cook', 'de': 'Cookinseln', 'dje': 'Kuuk gungey', 'dsb': 'Cookowe kupy', 'dz': 'ཀུག་གླིང་ཚོམ', 'ebu': 'Visiwa vya Cook', 'ee': 'Kook ƒudomekpo nutome', 'el': 'Νήσοι Κουκ', 'el-polyton': 'Νῆσοι Κούκ', 'en': 'Cook Islands', 'en-Dsrt': '𐐗𐐳𐐿 𐐌𐑊𐐲𐑌𐐼𐑆', 'eo': 'Kukinsuloj', 'es': 'Islas Cook', 'et': 'Cooki saared', 'eu': 'Cook uharteak', 'ewo': 'Minlán Mí kúg', 'fa': 'جزایر کوک', 'ff': 'Duuɗe Kuuk', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤑𞤵𞥅𞤳', 'ff-Latn': 'Duuɗe Kuuk', 'fi': 'Cookinsaaret', 'fil': 'Cook Islands', 'fo': 'Cooksoyggjar', 'fr': 'Îles Cook', 'fur': 'Isulis Cook', 'fy': 'Cookeilannen', 'ga': 'Oileáin Cook', 'gd': 'Eileanan Cook', 'gl': 'Illas Cook', 'gsw': 'Cook-Insle', 'gu': 'કુક આઇલેન્ડ્સ', 'guz': 'Visiwa vya Cook', 'ha': 'Tsibiran Cook', 'he': 'איי קוק', 'hi': 'कुक द्वीपसमूह', 'hi-Latn': 'Cook Islands', 'hr': 'Cookovi Otoci', 'hsb': 'Cookowe kupy', 'ht': 'Îles Cook', 'hu': 'Cook-szigetek', 'hy': 'Կուկի կղզիներ', 'ia': 'Insulas Cook', 'id': 'Kepulauan Cook', 'ig': 'Agwaetiti Cook', 'is': 'Cooks-eyjar', 'it': 'Isole Cook', 'ja': 'クック諸島', 'jmc': 'Visiwa vya Cook', 'jv': 'Kapuloan Cook', 'ka': 'კუკის კუნძულები', 'kaa': 'Кук атаўлары', 'kaa-Cyrl': 'Кук атаўлары', 'kab': 'Tigzirin n Kuk', 'kam': 'Isiwa sya Cook', 'kde': 'Chisiwa cha Cook', 'kea': 'Ilhas Kuk', 'kgp': 'Kuki Goj-vẽso', 'khq': 'Kuuk gungey', 'ki': 'Visiwa vya Cook', 'kk': 'Кук аралдары', 'kk-Cyrl': 'Кук аралдары', 'kln': 'Ikwembeyotab Cook', 'km': 'កោះ\u200bខូក', 'kn': 'ಕುಕ್ ದ್ವೀಪಗಳು', 'ko': '쿡 제도', 'kok': 'कुक आयलँड्स', 'kok-Deva': 'कुक आयलँड्स', 'ks': 'کُک جٔزیٖرٕ', 'ks-Arab': 'کُک جٔزیٖرٕ', 'ksb': 'Visiwa vya Cook', 'ksf': 'zɛ i kúk', 'ksh': 'de Kuuk-Enselle', 'ku': 'Giravên Cookê', 'kxv': 'kuk dīp', 'kxv-Deva': 'कुक द्वीप', 'kxv-Latn': 'kuk dīp', 'kxv-Orya': 'କୁକ ଦିପ', 'kxv-Telu': 'కుక్ దీపులు', 'ky': 'Кук аралдары', 'lag': 'Visíiwa vya Kúuku', 'lb': 'Cookinselen', 'lg': 'Bizinga bya Kkuki', 'lld': 'Isoles Cook', 'ln': 'Bisanga bya Kookɛ', 'lo': 'ໝູ່ເກາະຄຸກ', 'lt': 'Kuko Salos', 'lu': 'Lutanda lua Kookɛ', 'luo': 'Cook Islands', 'luy': 'Visiwa vya Cook', 'lv': 'Kuka salas', 'mai': 'कुक द्वीपसमूह', 'mas': 'Visiwa vya Cook', 'mer': 'Aĩrandi cia Cook', 'mfe': 'Zil Cook', 'mg': 'Nosy Kook', 'mi': 'Kuki Airani', 'mk': 'Кукови Острови', 'ml': 'കുക്ക് ദ്വീപുകൾ', 'mn': 'Күүкийн арлууд', 'mr': 'कुक बेटे', 'ms': 'Kepulauan Cook', 'mt': 'Gżejjer Cook', 'mua': 'kook ma laŋne', 'my': 'ကွတ် ကျွန်းစု', 'mzn': 'کوک جزایر', 'naq': 'Cook Islands', 'nb': 'Cookøyene', 'nd': 'Cook Islands', 'ne': 'कुक टापुहरु', 'nl': 'Cookeilanden', 'nmg': 'Maŋ́ má Kook', 'nn': 'Cookøyane', 'no': 'Cookøyene', 'nqo': 'ߞߎߞ ߕߌ߲', 'nus': 'Kuk ɣa̱ylɛn', 'nyn': 'Ebizinga bya Kuuku', 'om': 'Odoloota Kuuk', 'or': 'କୁକ୍\u200c ଦ୍ୱୀପପୁଞ୍ଜ', 'pa': 'ਕੁੱਕ ਟਾਪੂ', 'pa-Guru': 'ਕੁੱਕ ਟਾਪੂ', 'pcm': 'Kúk Aílands', 'pl': 'Wyspy Cooka', 'ps': 'کوک ټاپوګان', 'pt': 'Ilhas Cook', 'qu': 'Islas Cook', 'rm': 'Inslas Cook', 'rn': 'Izinga rya Kuku', 'ro': 'Insulele Cook', 'rof': 'Visiwa vya Cook', 'ru': 'о-ва Кука', 'rwk': 'Visiwa vya Cook', 'saq': 'Visiwa vya Cook', 'sbp': 'Figunguli fya Kooki', 'sc': 'Ìsulas Cook', 'sd': 'ڪوڪ ٻيٽ', 'sd-Arab': 'ڪوڪ ٻيٽ', 'se': 'Cook-sullot', 'seh': 'Ilhas Cook', 'ses': 'Kuuk gungey', 'sg': 'âzûâ Kûku', 'shi': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⴽⵓⴽ', 'shi-Latn': 'tigzirin n kuk', 'shi-Tfng': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⴽⵓⴽ', 'si': 'කුක් දූපත්', 'sk': 'Cookove ostrovy', 'sl': 'Cookovi otoki', 'smn': 'Cooksuolluuh', 'sn': 'Zvitsuwa zveCook', 'so': 'Jasiiradda Kook', 'sq': 'Ishujt Kuk', 'sr': 'Кукова Острва', 'sr-Cyrl': 'Кукова Острва', 'sr-Latn': 'Kukova Ostrva', 'sv': 'Cooköarna', 'sw': 'Visiwa vya Cook', 'syr': 'ܓܙܪܬܐ ܟܘܟ', 'ta': 'குக் தீவுகள்', 'te': 'కుక్ దీవులు', 'teo': 'Visiwa vya Cook', 'tg': 'Ҷазираҳои Кук', 'th': 'หมู่เกาะคุก', 'ti': 'ደሴታት ኩክ', 'tk': 'Kuk adalary', 'to': 'ʻOtumotu Kuki', 'tr': 'Cook Adaları', 'tt': 'Кук утраулары', 'twq': 'Kuuk gungey', 'tzm': 'Tigzirin n Kuk', 'ug': 'كۇك ئاراللىرى', 'uk': 'Острови Кука', 'ur': 'کک آئلینڈز', 'uz': 'Kuk orollari', 'uz-Cyrl': 'Кук ороллари', 'uz-Latn': 'Kuk orollari', 'vai': 'ꖏꕃ ꔳꘋꗣ', 'vai-Latn': 'Kóki Tiŋŋɛ', 'vai-Vaii': 'ꖏꕃ ꔳꘋꗣ', 'vec': 'Ìzole Cook', 'vi': 'Quần đảo Cook', 'vun': 'Visiwa vya Cook', 'wae': 'Cookinslä', 'wo': 'Duni Kuuk', 'xh': 'ECook Islands', 'xnr': 'कुक द्वीपां दा झुंड', 'xog': 'Ebizinga bya Kkuki', 'yav': 'Kúuke', 'yi': 'קוק אינזלען', 'yo': 'Etíokun Kùúkù', 'yrl': 'Kapuãma-ita Kooki', 'yue': '庫克群島', 'yue-Hans': '库克群岛', 'yue-Hant': '庫克群島', 'zgh': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⴽⵓⴽ', 'zh': '库克群岛', 'zh-Hans': '库克群岛', 'zh-Hant': '庫克群島', 'zu': 'i-Cook Islands'},
    'CL': {'af': 'Chili', 'agq': 'Chilè', 'ak': 'Kyili', 'am': 'ቺሊ', 'ar': 'تشيلي', 'as': 'চিলি', 'asa': 'Chile', 'ast': 'Chile', 'az': 'Çili', 'az-Cyrl': 'Чили', 'az-Latn': 'Çili', 'bal-Latn': 'Chilay', 'bas': 'Kìlî', 'be': 'Чылі', 'be-tarask': 'Чылі', 'bez': 'Huchile', 'bg': 'Чили', 'bgn': 'چیلی', 'blo': 'Shilii', 'bm': 'Sili', 'bn': 'চিলি', 'br': 'Chile', 'brx': 'चिले', 'bs': 'Čile', 'bs-Cyrl': 'Чиле', 'bs-Latn': 'Čile', 'ca': 'Xile', 'ca-ES-valencia': 'Xile', 'ccp': '𑄌𑄨𑄣𑄨', 'ce': 'Чили', 'ceb': 'Chile', 'cgg': 'Chile', 'chr': 'ᏥᎵ', 'ckb': 'چیلی', 'cs': 'Chile', 'cv': 'Чили', 'cy': 'Chile', 'da': 'Chile', 'dav': 'Chile', 'de': 'Chile', 'dje': 'Šiili', 'dsb': 'Chilska', 'dyo': 'Cili', 'dz': 'ཅི་ལི', 'ebu': 'Chile', 'ee': 'Tsile nutome', 'el': 'Χιλή', 'el-polyton': 'Χιλή', 'en': 'Chile', 'en-Dsrt': '𐐕𐐨𐑊𐐩', 'eo': 'Ĉilio', 'es': 'Chile', 'et': 'Tšiili', 'eu': 'Txile', 'ewo': 'Tsilí', 'fa': 'شیلی', 'ff': 'Cilii', 'ff-Adlm': '𞤕𞤭𞤤𞤫𞥊𞥅', 'ff-Latn': 'Cilii', 'fi': 'Chile', 'fil': 'Chile', 'fo': 'Kili', 'fr': 'Chili', 'fur': 'Cile', 'fy': 'Chili', 'ga': 'an tSile', 'gd': 'An t-Sile', 'gl': 'Chile', 'gsw': 'Tschile', 'gu': 'ચિલી', 'guz': 'Chile', 'ha': 'Chile', 'he': 'צ׳ילה', 'hi': 'चिली', 'hi-Latn': 'Chile', 'hr': 'Čile', 'hsb': 'Chilska', 'ht': 'Chili', 'hu': 'Chile', 'hy': 'Չիլի', 'ia': 'Chile', 'id': 'Cile', 'ig': 'Chile', 'is': 'Síle', 'it': 'Cile', 'ja': 'チリ', 'jgo': 'Cíllɛ', 'jmc': 'Chile', 'jv': 'Cilé', 'ka': 'ჩილე', 'kaa': 'Чили', 'kaa-Cyrl': 'Чили', 'kab': 'Cili', 'kam': 'Chile', 'kde': 'Chile', 'kea': 'Xili', 'kgp': 'Sire', 'khq': 'Šiili', 'ki': 'Chile', 'kk': 'Чили', 'kk-Cyrl': 'Чили', 'kln': 'Emetab Chile', 'km': 'ស៊ីលី', 'kn': 'ಚಿಲಿ', 'ko': '칠레', 'kok': 'चिली', 'kok-Deva': 'चिली', 'ks': 'چِلی', 'ks-Arab': 'چِلی', 'ksb': 'Chile', 'ksf': 'cíli', 'ksh': 'Schiile', 'ku': 'Şîle', 'kxv': 'cili', 'kxv-Deva': 'चिली', 'kxv-Latn': 'cili', 'kxv-Orya': 'ଚିଲି', 'kxv-Telu': 'చిలి', 'ky': 'Чили', 'lag': 'Chíile', 'lb': 'Chile', 'lg': 'Cile', 'lld': 'Cile', 'ln': 'Síli', 'lo': 'ຊິລີ', 'lt': 'Čilė', 'lu': 'Shili', 'luo': 'Chile', 'luy': 'Chile', 'lv': 'Čīle', 'mai': 'चिली', 'mas': 'Chile', 'mer': 'Chile', 'mfe': 'Shili', 'mg': 'Shili', 'mgh': 'Uchile', 'mi': 'Hiri', 'mk': 'Чиле', 'ml': 'ചിലി', 'mn': 'Чили', 'mr': 'चिली', 'ms': 'Chile', 'mt': 'iċ-Ċili', 'mua': 'syili', 'my': 'ချီလီ', 'mzn': 'شیلی', 'naq': 'Chilib', 'nb': 'Chile', 'nd': 'Chile', 'ne': 'चिली', 'nl': 'Chili', 'nmg': 'Tshili', 'nn': 'Chile', 'no': 'Chile', 'nqo': 'ߛ߭ߟߌ߫', 'nus': 'Cili̱', 'nyn': 'Chile', 'om': 'Chiilii', 'or': 'ଚିଲି', 'pa': 'ਚਿਲੀ', 'pa-Guru': 'ਚਿਲੀ', 'pcm': 'Chílẹ', 'pl': 'Chile', 'ps': 'چیلي', 'pt': 'Chile', 'qu': 'Chile', 'rm': 'Chile', 'rn': 'Shili', 'ro': 'Chile', 'rof': 'Chile', 'ru': 'Чили', 'rwk': 'Chile', 'sah': 'Чиили', 'saq': 'Chile', 'sbp': 'Shile', 'sc': 'Tzile', 'sd': 'چلي', 'sd-Arab': 'چلي', 'se': 'Čiile', 'seh': 'Chile', 'ses': 'Šiili', 'sg': 'Shilïi', 'shi': 'ⵛⵛⵉⵍⵉ', 'shi-Latn': 'ccili', 'shi-Tfng': 'ⵛⵛⵉⵍⵉ', 'si': 'චිලී', 'sk': 'Čile', 'sl': 'Čile', 'smn': 'Chile', 'sn': 'Chile', 'so': 'Jili', 'sq': 'Kili', 'sr': 'Чиле', 'sr-Cyrl': 'Чиле', 'sr-Latn': 'Čile', 'sv': 'Chile', 'sw': 'Chile', 'syr': 'ܬܫܝܠܝ', 'ta': 'சிலி', 'te': 'చిలీ', 'teo': 'Chile', 'tg': 'Чили', 'th': 'ชิลี', 'ti': 'ቺሌ', 'tk': 'Çili', 'to': 'Sili', 'tr': 'Şili', 'tt': 'Чили', 'twq': 'Šiili', 'tzm': 'Ccili', 'ug': 'چىلى', 'uk': 'Чилі', 'ur': 'چلی', 'uz': 'Chili', 'uz-Cyrl': 'Чили', 'uz-Latn': 'Chili', 'vai': 'ꔚꔷ', 'vai-Latn': 'Chéli', 'vai-Vaii': 'ꔚꔷ', 'vec': 'Cile', 'vi': 'Chile', 'vun': 'Chile', 'wae': 'Tšile', 'wo': 'Sili', 'xh': 'EChile', 'xnr': 'चिली', 'xog': 'Cile', 'yav': 'Silí', 'yi': 'טשילע', 'yo': 'Ṣílè', 'yrl': 'Xiri', 'yue': '智利', 'yue-Hans': '智利', 'yue-Hant': '智利', 'zgh': 'ⵛⵛⵉⵍⵉ', 'zh': '智利', 'zh-Hans': '智利', 'zh-Hant': '智利', 'zu': 'i-Chile'},
    'CM': {'af': 'Kameroen', 'agq': 'Kàmàlûŋ', 'ak': 'Kamɛrun', 'am': 'ካሜሩን', 'ar': 'الكاميرون', 'as': 'কেমেৰুণ', 'asa': 'Kameruni', 'ast': 'Camerún', 'az': 'Kamerun', 'az-Cyrl': 'Камерун', 'az-Latn': 'Kamerun', 'bal-Latn': 'Kaymrun', 'bas': 'Kàmɛ̀rûn', 'be': 'Камерун', 'be-tarask': 'Камерун', 'bez': 'Hukameruni', 'bg': 'Камерун', 'bgn': 'کامیرون', 'blo': 'Kamerun', 'bm': 'Kameruni', 'bn': 'ক্যামেরুন', 'br': 'Kameroun', 'brx': 'केमेरून', 'bs': 'Kamerun', 'bs-Cyrl': 'Камерун', 'bs-Latn': 'Kamerun', 'bss': 'Kamerûn', 'ca': 'Camerun', 'ca-ES-valencia': 'Camerun', 'ccp': '𑄇𑄳𑄠𑄟𑄬𑄢𑄪𑄚𑄴', 'ce': 'Камерун', 'ceb': 'Cameroon', 'cgg': 'Kameruuni', 'chr': 'ᎧᎹᎷᏂ', 'ckb': 'کامیرۆن', 'cs': 'Kamerun', 'cv': 'Камерун', 'cy': 'Camerŵn', 'da': 'Cameroun', 'dav': 'Kameruni', 'de': 'Kamerun', 'dje': 'Kameruun', 'dsb': 'Kamerun', 'dua': 'Cameroun', 'dyo': 'Kamerun', 'dz': 'ཀེ་མ་རུན', 'ebu': 'Kameruni', 'ee': 'Kamerun nutome', 'el': 'Καμερούν', 'el-polyton': 'Καμερούν', 'en': 'Cameroon', 'en-Dsrt': '𐐗𐐰𐑋𐐲𐑉𐐭𐑌', 'eo': 'Kameruno', 'es': 'Camerún', 'et': 'Kamerun', 'eu': 'Kamerun', 'ewo': 'Kamərún', 'fa': 'کامرون', 'ff': 'Kameruun', 'ff-Adlm': '𞤑𞤢𞤥𞤢𞤪𞤵𞥅𞤲', 'ff-Latn': 'Kameruun', 'fi': 'Kamerun', 'fil': 'Cameroon', 'fo': 'Kamerun', 'fr': 'Cameroun', 'fur': 'Camerun', 'fy': 'Kameroen', 'ga': 'Camarún', 'gd': 'Camarun', 'gl': 'Camerún', 'gsw': 'Kamerun', 'gu': 'કૅમરૂન', 'guz': 'Kameruni', 'ha': 'Kamaru', 'he': 'קמרון', 'hi': 'कैमरून', 'hi-Latn': 'Cameroon', 'hr': 'Kamerun', 'hsb': 'Kamerun', 'ht': 'Cameroun', 'hu': 'Kamerun', 'hy': 'Կամերուն', 'ia': 'Camerun', 'id': 'Kamerun', 'ig': 'Cameroon', 'is': 'Kamerún', 'it': 'Camerun', 'ja': 'カメルーン', 'jgo': 'Kamɛlûn', 'jmc': 'Kameruni', 'jv': 'Kamerun', 'ka': 'კამერუნი', 'kaa': 'Камерун', 'kaa-Cyrl': 'Камерун', 'kab': 'Kamirun', 'kam': 'Kameluni', 'kde': 'Kameluni', 'kea': 'Kamarons', 'kgp': 'Kamỹrãj', 'khq': 'Kameruun', 'ki': 'Kameruni', 'kk': 'Камерун', 'kk-Cyrl': 'Камерун', 'kkj': 'Kamɛrun', 'kln': 'Emetab Cameroon', 'km': 'កាមេរូន', 'kn': 'ಕ್ಯಾಮರೂನ್', 'ko': '카메룬', 'kok': 'कॅमेरून', 'kok-Deva': 'कॅमेरून', 'ks': 'کیمِروٗن', 'ks-Arab': 'کیمِروٗن', 'ksb': 'Kameluni', 'ksf': 'kamɛrún', 'ksh': 'Kammeruhn', 'ku': 'Kamerûn', 'kxv': 'kemarun', 'kxv-Deva': 'केमरून', 'kxv-Latn': 'kemarun', 'kxv-Orya': 'କାମେରନ୍', 'kxv-Telu': 'కామెరూన్', 'ky': 'Камерун', 'lag': 'Kamerúuni', 'lb': 'Kamerun', 'lg': 'Kameruuni', 'lld': 'Camerun', 'ln': 'Kamɛrune', 'lo': 'ຄາເມຣູນ', 'lt': 'Kamerūnas', 'lu': 'Kamerune', 'luo': 'Cameroon', 'luy': 'Kameruni', 'lv': 'Kamerūna', 'mai': 'कैमरून', 'mas': 'Kameruni', 'mer': 'Kameruni', 'mfe': 'Kamerounn', 'mg': 'Kamerona', 'mgo': 'Kamalun', 'mi': 'Kamarūna', 'mk': 'Камерун', 'ml': 'കാമറൂൺ', 'mn': 'Камерун', 'mr': 'कॅमेरून', 'ms': 'Cameroon', 'mt': 'il-Kamerun', 'mua': 'kameruŋ', 'my': 'ကင်မရွန်း', 'mzn': 'کامرون', 'naq': 'Cameroonni', 'nb': 'Kamerun', 'nd': 'Khameruni', 'ne': 'क्यामरून', 'nl': 'Kameroen', 'nmg': 'Kamerun', 'nn': 'Kamerun', 'nnh': 'Kàmalûm', 'no': 'Kamerun', 'nqo': 'ߞߊߡߋߙߎ߲', 'nus': 'Kɛmɛrun', 'nyn': 'Kameruuni', 'om': 'Kaameruun', 'or': 'କାମେରୁନ୍', 'pa': 'ਕੈਮਰੂਨ', 'pa-Guru': 'ਕੈਮਰੂਨ', 'pcm': 'Kamẹrun', 'pl': 'Kamerun', 'ps': 'کامرون', 'pt': 'Camarões', 'qu': 'Camerún', 'rm': 'Camerun', 'rn': 'Kameruni', 'ro': 'Camerun', 'rof': 'Kameruni', 'ru': 'Камерун', 'rwk': 'Kameruni', 'saq': 'Kameruni', 'sbp': 'Kameruni', 'sc': 'Camerùn', 'sd': 'ڪيمرون', 'sd-Arab': 'ڪيمرون', 'se': 'Kamerun', 'seh': 'República dos Camarões', 'ses': 'Kameruun', 'sg': 'Kamerûne', 'shi': 'ⴽⴰⵎⵉⵔⵓⵏ', 'shi-Latn': 'kamirun', 'shi-Tfng': 'ⴽⴰⵎⵉⵔⵓⵏ', 'si': 'කැමරූන්', 'sk': 'Kamerun', 'sl': 'Kamerun', 'smn': 'Kamerun', 'sn': 'Kameruni', 'so': 'Kaameruun', 'sq': 'Kamerun', 'sr': 'Камерун', 'sr-Cyrl': 'Камерун', 'sr-Latn': 'Kamerun', 'sv': 'Kamerun', 'sw': 'Kameruni', 'syr': 'ܟܐܡܪܘܢ', 'ta': 'கேமரூன்', 'te': 'కామెరూన్', 'teo': 'Kameruni', 'tg': 'Камерун', 'th': 'แคเมอรูน', 'ti': 'ካሜሩን', 'tk': 'Kamerun', 'to': 'Kameluni', 'tr': 'Kamerun', 'tt': 'Камерун', 'twq': 'Kameruun', 'tzm': 'Kamerun', 'ug': 'كامېرون', 'uk': 'Камерун', 'ur': 'کیمرون', 'uz': 'Kamerun', 'uz-Cyrl': 'Камерун', 'uz-Latn': 'Kamerun', 'vai': 'ꕪꔈꖩꘋ', 'vai-Latn': 'Kameruŋ', 'vai-Vaii': 'ꕪꔈꖩꘋ', 'vec': 'Càmerun', 'vi': 'Cameroon', 'vun': 'Kameruni', 'wae': 'Kamerun', 'wo': 'Kamerun', 'xh': 'ECameroon', 'xnr': 'कैमरून', 'xog': 'Kameruuni', 'yav': 'Kemelún', 'yi': 'קאַמערון', 'yo': 'Kamerúúnì', 'yrl': 'Puty-ita', 'yue': '喀麥隆', 'yue-Hans': '喀麦隆', 'yue-Hant': '喀麥隆', 'zgh': 'ⴽⴰⵎⵉⵔⵓⵏ', 'zh': '喀麦隆', 'zh-Hans': '喀麦隆', 'zh-Hant': '喀麥隆', 'zu': 'i-Cameroon'},
    'CN': {'af': 'China', 'agq': 'Chaenà', 'ak': 'Kyaena', 'am': 'ቻይና', 'ar': 'الصين', 'as': 'চীন', 'asa': 'China', 'ast': 'China', 'az': 'Çin', 'az-Cyrl': 'Чин', 'az-Latn': 'Çin', 'bal-Latn': 'Chin', 'bas': 'Kinà', 'be': 'Кітай', 'be-tarask': 'Кітай', 'bez': 'Huchina', 'bg': 'Китай', 'bgn': 'چین', 'blo': 'Caɩna', 'bm': 'Siniwajamana', 'bn': 'চীন', 'bo': 'རྒྱ་ནག', 'br': 'Sina', 'brx': 'चाइना', 'bs': 'Kina', 'bs-Cyrl': 'Кина', 'bs-Latn': 'Kina', 'ca': 'Xina', 'ca-ES-valencia': 'Xina', 'ccp': '𑄌𑄩𑄚𑄴', 'ce': 'Цийчоь', 'ceb': 'China', 'cgg': 'China', 'chr': 'ᏓᎶᏂᎨᏍᏛ', 'ckb': 'چین', 'cs': 'Čína', 'cv': 'Китай', 'cy': 'Tsieina', 'da': 'Kina', 'dav': 'China', 'de': 'China', 'dje': 'Šiin', 'doi': 'चीन', 'dsb': 'China', 'dyo': 'Siin', 'dz': 'རྒྱ་ནག', 'ebu': 'China', 'ee': 'Tsaina nutome', 'el': 'Κίνα', 'el-polyton': 'Κίνα', 'en': 'China', 'en-Dsrt': '𐐕𐐴𐑌𐐲', 'eo': 'Ĉinujo', 'es': 'China', 'et': 'Hiina', 'eu': 'Txina', 'ewo': 'Tsáina', 'fa': 'چین', 'ff': 'Siin', 'ff-Adlm': '𞤕𞤢𞤴𞤲𞤢', 'ff-Latn': 'Siin', 'fi': 'Kiina', 'fil': 'China', 'fo': 'Kina', 'fr': 'Chine', 'fur': 'Cine', 'fy': 'Sina', 'ga': 'an tSín', 'gd': 'An t-Sìn', 'gl': 'China', 'gsw': 'Chiina', 'gu': 'ચીન', 'guz': 'China', 'ha': 'Sin', 'haw': 'Kina', 'he': 'סין', 'hi': 'चीन', 'hi-Latn': 'China', 'hr': 'Kina', 'hsb': 'China', 'ht': 'Chine', 'hu': 'Kína', 'hy': 'Չինաստան', 'ia': 'China', 'id': 'Tiongkok', 'ig': 'China', 'ii': 'ꍏꇩ', 'is': 'Kína', 'it': 'Cina', 'ja': '中国', 'jgo': 'Shîn', 'jmc': 'China', 'jv': 'Tyongkok', 'ka': 'ჩინეთი', 'kaa': 'Қытай', 'kaa-Cyrl': 'Қытай', 'kab': 'Lacin', 'kam': 'Kyaina', 'kde': 'China', 'kea': 'Xina', 'kgp': 'Sĩnỹ', 'khq': 'Šiin', 'ki': 'Caina', 'kk': 'Қытай', 'kk-Cyrl': 'Қытай', 'kln': 'Emetab China', 'km': 'ចិន', 'kn': 'ಚೀನಾ', 'ko': '중국', 'kok': 'चीन', 'kok-Deva': 'चीन', 'kok-Latn': 'Chin', 'ks': 'چیٖن', 'ks-Arab': 'چیٖن', 'ks-Deva': 'चीन', 'ksb': 'China', 'ksf': 'cín', 'ksh': 'Schiina', 'ku': 'Çîn', 'kxv': 'caina', 'kxv-Deva': 'चीन', 'kxv-Latn': 'caina', 'kxv-Orya': 'ଚିନ୍', 'kxv-Telu': 'చినా', 'ky': 'Кытай', 'lag': 'Chíina', 'lb': 'China', 'lg': 'Cayina', 'lkt': 'Pȟečhókaŋhaŋska Makȟóčhe', 'lld': 'Cina', 'ln': 'Sinɛ', 'lo': 'ຈີນ', 'lrc': 'چین', 'lt': 'Kinija', 'lu': 'Shine', 'luo': 'China', 'luy': 'China', 'lv': 'Ķīna', 'mai': 'चीन', 'mas': 'China', 'mer': 'China', 'mfe': 'Lasinn', 'mg': 'Sina', 'mgh': 'Uchina', 'mi': 'Haina', 'mk': 'Кина', 'ml': 'ചൈന', 'mn': 'Хятад', 'mni': 'চিনা', 'mni-Beng': 'চিনা', 'mr': 'चीन', 'ms': 'China', 'mt': 'iċ-Ċina', 'mua': 'syiŋ', 'my': 'တရုတ်', 'mzn': 'چین', 'naq': 'Chinab', 'nb': 'Kina', 'nd': 'China', 'ne': 'चीन', 'nl': 'China', 'nmg': 'Shine', 'nn': 'Kina', 'no': 'Kina', 'nqo': 'ߛߌߣ', 'nus': 'Cayna', 'nyn': 'China', 'om': 'Chaayinaa', 'or': 'ଚୀନ୍\u200c', 'os': 'Китай', 'pa': 'ਚੀਨ', 'pa-Guru': 'ਚੀਨ', 'pcm': 'Chaína', 'pl': 'Chiny', 'ps': 'چین', 'pt': 'China', 'qu': 'China', 'rm': 'China', 'rn': 'Ubushinwa', 'ro': 'China', 'rof': 'China', 'ru': 'Китай', 'rwk': 'China', 'sa': 'चीन:', 'sah': 'Кытай', 'saq': 'China', 'sat': 'ᱪᱤᱱ', 'sat-Olck': 'ᱪᱤᱱ', 'sbp': 'Shina', 'sc': 'Tzina', 'sd': 'چين', 'sd-Arab': 'چين', 'sd-Deva': 'चीन', 'se': 'Kiinná', 'seh': 'China', 'ses': 'Šiin', 'sg': 'Shîna', 'shi': 'ⵛⵛⵉⵏⵡⴰ', 'shi-Latn': 'ccinwa', 'shi-Tfng': 'ⵛⵛⵉⵏⵡⴰ', 'si': 'චීනය', 'sk': 'Čína', 'sl': 'Kitajska', 'smn': 'Kiina', 'sn': 'China', 'so': 'Shiinaha', 'sq': 'Kinë', 'sr': 'Кина', 'sr-Cyrl': 'Кина', 'sr-Latn': 'Kina', 'su': 'Tiongkok', 'su-Latn': 'Tiongkok', 'sv': 'Kina', 'sw': 'Uchina', 'syr': 'ܨܝܢ', 'ta': 'சீனா', 'te': 'చైనా', 'teo': 'China', 'tg': 'Хитой', 'th': 'จีน', 'ti': 'ቻይና', 'tk': 'Hytaý', 'to': 'Siaina', 'tr': 'Çin', 'tt': 'Кытай', 'twq': 'Šiin', 'tzm': 'Ṣṣin', 'ug': 'جۇڭگو', 'uk': 'Китай', 'ur': 'چین', 'uz': 'Xitoy', 'uz-Cyrl': 'Хитой', 'uz-Latn': 'Xitoy', 'vai': 'ꕦꔤꕯ', 'vai-Latn': 'Cháína', 'vai-Vaii': 'ꕦꔤꕯ', 'vec': 'Sina', 'vi': 'Trung Quốc', 'vun': 'China', 'wae': 'China', 'wo': 'Siin', 'xh': 'ETshayina', 'xnr': 'चीन', 'xog': 'Cayina', 'yav': 'Síine', 'yi': 'כינע', 'yo': 'Ṣáínà', 'yrl': 'Xina', 'yue': '中國', 'yue-Hans': '中华人民共和国', 'yue-Hant': '中國', 'za': 'Cunghgoz', 'zgh': 'ⵛⵛⵉⵏⵡⴰ', 'zh': '中国', 'zh-Hans': '中国', 'zh-Hant': '中國', 'zu': 'i-China'},
    'CO': {'af': 'Colombië', 'agq': 'Kòlombìa', 'ak': 'Kolombia', 'am': 'ኮሎምቢያ', 'ar': 'كولومبيا', 'as': 'কলম্বিয়া', 'asa': 'Kolombia', 'ast': 'Colombia', 'az': 'Kolumbiya', 'az-Cyrl': 'Колумбија', 'az-Latn': 'Kolumbiya', 'bal-Latn': 'Kólambiá', 'bas': 'Kɔ̀lɔmbìà', 'be': 'Калумбія', 'be-tarask': 'Калумбія', 'bez': 'Hukolombia', 'bg': 'Колумбия', 'bgn': 'کولومبیا', 'blo': 'Kolɔmbii', 'bm': 'Kolombi', 'bn': 'কলম্বিয়া', 'br': 'Kolombia', 'brx': 'कलम्बिया', 'bs': 'Kolumbija', 'bs-Cyrl': 'Колумбија', 'bs-Latn': 'Kolumbija', 'ca': 'Colòmbia', 'ca-ES-valencia': 'Colòmbia', 'ccp': '𑄃𑄣𑄧𑄟𑄴𑄝𑄨𑄠', 'ce': 'Колумби', 'ceb': 'Colombia', 'cgg': 'Korombiya', 'chr': 'ᎪᎸᎻᏈᎢᎠ', 'ckb': 'کۆلۆمبیا', 'cs': 'Kolumbie', 'cv': 'Колумби', 'cy': 'Colombia', 'da': 'Colombia', 'dav': 'Kolombia', 'de': 'Kolumbien', 'dje': 'Kolombi', 'dsb': 'Kolumbiska', 'dyo': 'Kolombi', 'dz': 'ཀོ་ལོམ་བྷི་ཡ', 'ebu': 'Kolombia', 'ee': 'Kolombia nutome', 'el': 'Κολομβία', 'el-polyton': 'Κολομβία', 'en': 'Colombia', 'en-Dsrt': '𐐗𐐲𐑊𐐲𐑋𐐺𐐨𐐲', 'eo': 'Kolombio', 'es': 'Colombia', 'et': 'Colombia', 'eu': 'Kolonbia', 'ewo': 'Kolɔmbí', 'fa': 'کلمبیا', 'ff': 'Kolombiya', 'ff-Adlm': '𞤑𞤮𞤤𞤮𞤥𞤦𞤭𞤴𞤢𞥄', 'ff-Latn': 'Kolombiya', 'fi': 'Kolumbia', 'fil': 'Colombia', 'fo': 'Kolombia', 'fr': 'Colombie', 'fur': 'Colombie', 'fy': 'Kolombia', 'ga': 'an Cholóim', 'gd': 'Coloimbia', 'gl': 'Colombia', 'gsw': 'Kolumbie', 'gu': 'કોલમ્બિયા', 'guz': 'Kolombia', 'ha': 'Kolambiya', 'he': 'קולומביה', 'hi': 'कोलंबिया', 'hi-Latn': 'Colombia', 'hr': 'Kolumbija', 'hsb': 'Kolumbiska', 'ht': 'Colombie', 'hu': 'Kolumbia', 'hy': 'Կոլումբիա', 'ia': 'Colombia', 'id': 'Kolombia', 'ig': 'Colombia', 'is': 'Kólumbía', 'it': 'Colombia', 'ja': 'コロンビア', 'jgo': 'Kɔllɔmbî', 'jmc': 'Kolombia', 'jv': 'Kolombia', 'ka': 'კოლუმბია', 'kaa': 'Колумбия', 'kaa-Cyrl': 'Колумбия', 'kab': 'Kulumbi', 'kam': 'Kolombia', 'kde': 'Kolombia', 'kea': 'Kolómbia', 'kgp': 'Korãmija', 'khq': 'Kolombi', 'ki': 'Kolombia', 'kk': 'Колумбия', 'kk-Cyrl': 'Колумбия', 'kln': 'Emetab Colombia', 'km': 'កូឡុំប៊ី', 'kn': 'ಕೊಲಂಬಿಯಾ', 'ko': '콜롬비아', 'kok': 'कोलंबिया', 'kok-Deva': 'कोलंबिया', 'ks': 'کولَمبِیا', 'ks-Arab': 'کولَمبِیا', 'ksb': 'Kolombia', 'ksf': 'kolɔmbí', 'ksh': 'Kolumbije', 'ku': 'Kolombîya', 'kxv': 'kolmbiaa', 'kxv-Deva': 'कोलंबिया', 'kxv-Latn': 'kolmbiaa', 'kxv-Orya': 'କଲମ୍ବିଆ', 'kxv-Telu': 'కొలంబియా', 'ky': 'Колумбия', 'lag': 'Kolómbia', 'lb': 'Kolumbien', 'lg': 'Kolombya', 'lld': 'Colombia', 'ln': 'Kolombi', 'lo': 'ໂຄລົມເບຍ', 'lt': 'Kolumbija', 'lu': 'Kolombi', 'luo': 'Colombia', 'luy': 'Kolombia', 'lv': 'Kolumbija', 'mai': 'कोलंबिया', 'mas': 'Kolombia', 'mer': 'Kolombia', 'mfe': 'Kolonbi', 'mg': 'Kôlômbia', 'mgh': 'Ukolombia', 'mi': 'Koromōpia', 'mk': 'Колумбија', 'ml': 'കൊളംബിയ', 'mn': 'Колумби', 'mr': 'कोलम्बिया', 'ms': 'Colombia', 'mt': 'il-Kolombja', 'mua': 'kolombiya', 'my': 'ကိုလံဘီယာ', 'mzn': 'کلمبیا', 'naq': 'Colombiab', 'nb': 'Colombia', 'nd': 'Kholombiya', 'ne': 'कोलोम्बिया', 'nl': 'Colombia', 'nmg': 'Kɔlɔ́mbia', 'nn': 'Colombia', 'no': 'Colombia', 'nqo': 'ߞߏߟߐ߲ߓߌ߫', 'nus': 'Kolombia', 'nyn': 'Korombiya', 'om': 'Kolombiyaa', 'or': 'କଲମ୍ବିଆ', 'pa': 'ਕੋਲੰਬੀਆ', 'pa-Guru': 'ਕੋਲੰਬੀਆ', 'pcm': 'Kolómbia', 'pl': 'Kolumbia', 'ps': 'کولمبیا', 'pt': 'Colômbia', 'qu': 'Colombia', 'rm': 'Columbia', 'rn': 'Kolombiya', 'ro': 'Columbia', 'rof': 'Kolombia', 'ru': 'Колумбия', 'rwk': 'Kolombia', 'saq': 'Kolombia', 'sbp': 'Kolombiya', 'sc': 'Colòmbia', 'sd': 'ڪولمبيا', 'sd-Arab': 'ڪولمبيا', 'se': 'Kolombia', 'seh': 'Colômbia', 'ses': 'Kolombi', 'sg': 'Kolombïi', 'shi': 'ⴽⵓⵍⵓⵎⴱⵢⴰ', 'shi-Latn': 'culumbya', 'shi-Tfng': 'ⴽⵓⵍⵓⵎⴱⵢⴰ', 'si': 'කොළොම්බියාව', 'sk': 'Kolumbia', 'sl': 'Kolumbija', 'smn': 'Kolumbia', 'sn': 'Kolombia', 'so': 'Koloombiya', 'sq': 'Kolumbi', 'sr': 'Колумбија', 'sr-Cyrl': 'Колумбија', 'sr-Latn': 'Kolumbija', 'sv': 'Colombia', 'sw': 'Kolombia', 'syr': 'ܟܘܠܘܡܒܝܐ', 'ta': 'கொலம்பியா', 'te': 'కొలంబియా', 'teo': 'Kolombia', 'tg': 'Колумбия', 'th': 'โคลอมเบีย', 'ti': 'ኮሎምብያ', 'tk': 'Kolumbiýa', 'to': 'Kolomipia', 'tr': 'Kolombiya', 'tt': 'Колумбия', 'twq': 'Kolombi', 'tzm': 'Kulumbya', 'ug': 'كولومبىيە', 'uk': 'Колумбія', 'ur': 'کولمبیا', 'uz': 'Kolumbiya', 'uz-Cyrl': 'Колумбия', 'uz-Latn': 'Kolumbiya', 'vai': 'ꗛꗏꔭꕩ', 'vai-Latn': 'Kɔlɔmbiya', 'vai-Vaii': 'ꗛꗏꔭꕩ', 'vec': 'Colonbia', 'vi': 'Colombia', 'vun': 'Kolombia', 'wae': 'Kolumbie', 'wo': 'Kolombi', 'xh': 'EColombia', 'xnr': 'कोलंबिया', 'xog': 'Kolombya', 'yav': 'Kɔlɔ́mbía', 'yi': 'קאלאמביע', 'yo': 'Kòlómíbìa', 'yrl': 'Kurũbiya', 'yue': '哥倫比亞', 'yue-Hans': '哥伦比亚', 'yue-Hant': '哥倫比亞', 'zgh': 'ⴽⵓⵍⵓⵎⴱⵢⴰ', 'zh': '哥伦比亚', 'zh-Hans': '哥伦比亚', 'zh-Hant': '哥倫比亞', 'zu': 'i-Colombia'},
    'CP': {'af': 'Clippertoneiland', 'ak': 'Klepatin Aeland', 'am': 'ክሊፐርቶን ደሴት', 'ar': 'جزيرة كليبيرتون', 'as': 'ক্লিপাৰটোন দ্বীপ', 'ast': 'Islla Clipperton', 'az': 'Klipperton adası', 'az-Cyrl': 'Клиппертон адасы', 'az-Latn': 'Klipperton adası', 'bal-Latn': 'Klipperton Islánd', 'be': 'Востраў Кліпертан', 'be-tarask': 'Востраў Кліпертан', 'bg': 'остров Клипертон', 'blo': 'Klipɛɛtɔn kaAtukǝltǝna', 'bn': 'ক্লিপারটন দ্বীপপুঞ্জ', 'br': 'Enez Clipperton', 'brx': 'क्लिप्पार्टन द्वीप', 'bs': 'Ostrvo Clipperton', 'bs-Cyrl': 'Острво Клипертон', 'bs-Latn': 'Ostrvo Clipperton', 'ca': 'Illa Clipperton', 'ca-ES-valencia': 'Illa Clipperton', 'ccp': '𑄇𑄳𑄣𑄨𑄛𑄢𑄴𑄑𑄧𑄚𑄴 𑄃𑄭𑄣𑄳𑄠𑄚𑄳𑄓𑄴', 'ce': 'Клиппертон', 'ceb': 'Clipperton Island', 'chr': 'ᎦᏂᏴᏔᏅᎣᏓᎸ ᎤᎦᏚᏛᎢ', 'ckb': 'دوورگەی کلیپێرتۆن', 'cs': 'Clippertonův ostrov', 'cv': 'Клиппертон утравӗ', 'cy': 'Ynys Clipperton', 'da': 'Clippertonøen', 'de': 'Clipperton-Insel', 'dsb': 'Clippertonowa kupa', 'dz': 'ཀི་ལི་པེར་ཊོན་མཚོ་གླིང་', 'ee': 'Klipaton ƒudomekpo nutome', 'el': 'Νήσος Κλίπερτον', 'el-polyton': 'Νήσος Κλίπερτον', 'en': 'Clipperton Island', 'es': 'Isla Clipperton', 'et': 'Clippertoni saar', 'eu': 'Clipperton uhartea', 'fa': 'جزایر کلیپرتون', 'ff-Adlm': '𞤅𞤵𞤪𞤭𞥅𞤪𞤫 𞤑𞤭𞤤𞤭𞤨𞤫𞤪𞤼𞤮𞤲', 'fi': 'Clippertoninsaari', 'fil': 'Clipperton Island', 'fo': 'Clipperton', 'fr': 'Île Clipperton', 'fur': 'Isule Clipperton', 'fy': 'Clipperton', 'ga': 'Oileán Clipperton', 'gd': 'Eilean Clipperton', 'gl': 'Illa Clipperton', 'gu': 'ક્લિપરટન આઇલેન્ડ', 'ha': 'Tsibirin Clipperton', 'he': 'האי קליפרטון', 'hi': 'क्लिपर्टन द्वीप', 'hi-Latn': 'Clipperton Island', 'hr': 'Otok Clipperton', 'hsb': 'Clippertonowa kupa', 'ht': 'Île Clipperton', 'hu': 'Clipperton-sziget', 'hy': 'Քլիփերթոն կղզի', 'ia': 'Insula Clipperton', 'id': 'Pulau Clipperton', 'ig': 'Agwaetiti Clipperton', 'is': 'Clipperton-eyja', 'it': 'Isola di Clipperton', 'ja': 'クリッパートン島', 'jv': 'Pulo Clipperton', 'ka': 'კლიპერტონის კუნძული', 'kea': 'Ilha Kliperton', 'kgp': 'Kripertãn Goj-vẽso', 'kk': 'Клиппертон аралы', 'kk-Cyrl': 'Клиппертон аралы', 'km': 'កោះ\u200bឃ្លីភឺតុន', 'kn': 'ಕ್ಲಿಪ್ಪರ್\u200cಟಾನ್ ದ್ವೀಪ', 'ko': '클리퍼턴섬', 'kok': 'क्लिपरटॉन आयलँड', 'kok-Deva': 'क्लिपरटॉन आयलँड', 'ks': 'کلیپرٹن جزیرٕ', 'ks-Arab': 'کلیپرٹن جزیرٕ', 'ksh': 'de Klipperton-Ensel', 'ku': 'Girava Clippertonê', 'ky': 'Клиппертон аралы', 'lb': 'Clipperton-Insel', 'lld': 'Isola de Clipperton', 'lo': 'ເກາະຄລິບເປີຕັນ', 'lt': 'Klipertono sala', 'lv': 'Klipertona sala', 'mai': 'क्लिपर्टन द्वीप', 'mi': 'Te Moutere Kiripetone', 'mk': 'Остров Клипертон', 'ml': 'ക്ലിപ്പെർട്ടൻ ദ്വീപ്', 'mn': 'Клиппертон арал', 'mr': 'क्लिपरटोन बेट', 'ms': 'Pulau Clipperton', 'mt': 'il-Gżira Clipperton', 'my': 'ကလစ်ပါတန်ကျွန်း', 'mzn': 'کلیپرتون جزیره', 'nb': 'Clippertonøya', 'ne': 'क्लिप्पेर्टन टापु', 'nl': 'Clipperton', 'nn': 'Clippertonøya', 'no': 'Clippertonøya', 'nqo': 'ߞߟߌߔߍߙߑߕߐ߲߫ ߕߌ߲', 'om': 'Odola Kilippertoo', 'or': 'କ୍ଲିପରଟନ୍\u200c ଦ୍ୱୀପ', 'pa': 'ਕਲਿੱਪਰਟਨ ਟਾਪੂ', 'pa-Guru': 'ਕਲਿੱਪਰਟਨ ਟਾਪੂ', 'pcm': 'Klipatọ́n Aíland', 'pl': 'Wyspa Clippertona', 'ps': 'د کلپرټون ټاپو', 'pt': 'Ilha de Clipperton', 'qu': 'Isla Clipperton', 'rm': 'Insla da Clipperton', 'ro': 'Insula Clipperton', 'ru': 'о-в Клиппертон', 'sc': 'Ìsula de Clipperton', 'sd': 'ڪلپرٽن ٻيٽ', 'sd-Arab': 'ڪلپرٽن ٻيٽ', 'se': 'Clipperton-sullot', 'si': 'ක්ලීපර්ටන් දූපත', 'sk': 'Clipperton', 'sl': 'Otok Clipperton', 'smn': 'Clippertonsuálui', 'so': 'Jasiiradda Kilibarton', 'sq': 'Ishulli Klipërton', 'sr': 'Острво Клипертон', 'sr-Cyrl': 'Острво Клипертон', 'sr-Latn': 'Ostrvo Kliperton', 'sv': 'Clippertonön', 'sw': 'Kisiwa cha Clipperton', 'ta': 'கிலிப்பர்டன் தீவு', 'te': 'క్లిప్పర్టన్ దీవి', 'tg': 'Ҷазираи Клиппертон', 'th': 'เกาะคลิปเปอร์ตัน', 'ti': 'ደሴት ክሊፐርቶን', 'tk': 'Klipperton adasy', 'to': 'Motu Kilipatoni', 'tr': 'Clipperton Adası', 'tt': 'Клиппертон утравы', 'ug': 'كىلىپپېرتون ئاراللىرى', 'uk': 'Острів Кліппертон', 'ur': 'کلپرٹن آئلینڈ', 'uz': 'Klipperton oroli', 'uz-Cyrl': 'Клиппертон ороли', 'uz-Latn': 'Klipperton oroli', 'vai': 'ꕃꔒꕐꗋꘋ ꔳꘋꗣ', 'vai-Vaii': 'ꕃꔒꕐꗋꘋ ꔳꘋꗣ', 'vec': 'Ìzola de Clipperton', 'vi': 'Đảo Clipperton', 'wae': 'Clipperton Insla', 'wo': 'Ile Clipperton', 'xh': 'EClipperton Island', 'xnr': 'क्लिपर्टन द्वीप', 'yo': 'Erékùsù Clipperston', 'yrl': 'Kiripetũ Kapuãma', 'yue': '克里派頓島', 'yue-Hans': '克里派顿岛', 'yue-Hant': '克里派頓島', 'zh': '克利珀顿岛', 'zh-Hans': '克利珀顿岛', 'zh-Hant': '克里派頓島', 'zu': 'i-Clipperton Island'},
    'CQ': {'bal-Latn': 'Sárk', 'blo': 'Sark', 'en': 'Sark', 'et': 'Sark', 'fi': 'Sark', 'fo': 'Sark', 'gd': 'Sarc', 'hi-Latn': 'Sark', 'hu': 'Sark', 'ia': 'Sark', 'ig': 'Sark', 'ku': 'Sark', 'nb': 'Sark', 'nl': 'Sark', 'nn': 'Sark', 'no': 'Sark', 'sc': 'Sark', 'ti': 'ሳርክ', 'vec': 'Sarc'},
    'CR': {'af': 'Costa Rica', 'agq': 'Kòsɨ̀tà Lekà', 'ak': 'Kɔsta Rika', 'am': 'ኮስታሪካ', 'ar': 'كوستاريكا', 'as': 'কোষ্টা ৰিকা', 'asa': 'Kothtarika', 'ast': 'Costa Rica', 'az': 'Kosta Rika', 'az-Cyrl': 'Коста Рика', 'az-Latn': 'Kosta Rika', 'bal-Latn': 'Kóstá Riká', 'bas': 'Kòstà Rikà', 'be': 'Коста-Рыка', 'be-tarask': 'Коста-Рыка', 'bez': 'Hukostarika', 'bg': 'Коста Рика', 'blo': 'Kɔsta Rikaa', 'bm': 'Kɔsitarika', 'bn': 'কোস্টারিকা', 'br': 'Costa Rica', 'brx': 'कस्टा रिका', 'bs': 'Kostarika', 'bs-Cyrl': 'Костарика', 'bs-Latn': 'Kostarika', 'ca': 'Costa Rica', 'ca-ES-valencia': 'Costa Rica', 'ccp': '𑄇𑄮𑄥𑄳𑄑𑄢𑄨𑄇', 'ce': 'Коста-Рика', 'ceb': 'Costa Rica', 'cgg': 'Kositarika', 'chr': 'ᎪᏍᏓ ᎵᎧ', 'ckb': 'کۆستاریکا', 'cs': 'Kostarika', 'cv': 'Коста-Рика', 'cy': 'Costa Rica', 'da': 'Costa Rica', 'dav': 'Kostarika', 'de': 'Costa Rica', 'dje': 'Kosta rika', 'dsb': 'Kosta Rika', 'dyo': 'Kosta Rika', 'dz': 'ཀོས་ཊ་རི་ཀ', 'ebu': 'Kostarika', 'ee': 'Kosta Rika nutome', 'el': 'Κόστα Ρίκα', 'el-polyton': 'Κόστα Ρίκα', 'en': 'Costa Rica', 'en-Dsrt': '𐐗𐐱𐑅𐐻𐐲 𐐡𐐨𐐿𐐲', 'eo': 'Kostariko', 'es': 'Costa Rica', 'et': 'Costa Rica', 'eu': 'Costa Rica', 'ewo': 'Kosta Ríka', 'fa': 'کاستاریکا', 'ff': 'Kosta Rikaa', 'ff-Adlm': '𞤑𞤮𞤧𞤼𞤢 𞤈𞤭𞤳𞤢𞥄', 'ff-Latn': 'Kosta Rikaa', 'fi': 'Costa Rica', 'fil': 'Costa Rica', 'fo': 'Kosta Rika', 'fr': 'Costa Rica', 'fur': 'Costa Rica', 'fy': 'Costa Rica', 'ga': 'Cósta Ríce', 'gd': 'Costa Rìcea', 'gl': 'Costa Rica', 'gsw': 'Coschta Rica', 'gu': 'કોસ્ટા રિકા', 'guz': 'Kostarika', 'ha': 'Kwasta Rika', 'he': 'קוסטה ריקה', 'hi': 'कोस्टारिका', 'hi-Latn': 'Costa Rica', 'hr': 'Kostarika', 'hsb': 'Kosta Rika', 'ht': 'Costa Rica', 'hu': 'Costa Rica', 'hy': 'Կոստա Ռիկա', 'ia': 'Costa Rica', 'id': 'Kosta Rika', 'ig': 'Kosta Rika', 'is': 'Kostaríka', 'it': 'Costa Rica', 'ja': 'コスタリカ', 'jmc': 'Kostarika', 'jv': 'Kosta Rika', 'ka': 'კოსტა-რიკა', 'kaa': 'Коста-Рика', 'kaa-Cyrl': 'Коста-Рика', 'kab': 'Kusta Rika', 'kam': 'Kostarika', 'kde': 'Kostalika', 'kea': 'Kosta Rika', 'kgp': 'Konhta Rika', 'khq': 'Kosta rika', 'ki': 'Kostarika', 'kk': 'Коста-Рика', 'kk-Cyrl': 'Коста-Рика', 'kln': 'Emetab Costa Rica', 'km': 'កូស្តារីកា', 'kn': 'ಕೊಸ್ಟಾ ರಿಕಾ', 'ko': '코스타리카', 'kok': 'कोस्ता रिका', 'kok-Deva': 'कोस्ता रिका', 'ks': 'کوسٹا ریکا', 'ks-Arab': 'کوسٹا ریکا', 'ksb': 'Kostalika', 'ksf': 'kɔstaríka', 'ksh': 'Kostarika', 'ku': 'Kosta Rîka', 'kxv': 'kosta rika', 'kxv-Deva': 'कोस्टारिका', 'kxv-Latn': 'kosta rika', 'kxv-Orya': 'କସ୍ଟା ରିକା', 'kxv-Telu': 'కోస్టా రికా', 'ky': 'Коста-Рика', 'lag': 'Kósita Rɨ́ɨka', 'lb': 'Costa Rica', 'lg': 'Kosita Rika', 'lld': 'Costa Rica', 'ln': 'Kositarika', 'lo': 'ໂຄສຕາ ຣິກາ', 'lt': 'Kosta Rika', 'lu': 'Kositarika', 'luo': 'Costa Rica', 'luy': 'Kostarika', 'lv': 'Kostarika', 'mai': 'कोस्टारिका', 'mas': 'Kostarika', 'mer': 'Kostarika', 'mfe': 'Costa Rica', 'mg': 'Kosta Rikà', 'mi': 'Koto Rīka', 'mk': 'Костарика', 'ml': 'കോസ്റ്ററിക്ക', 'mn': 'Коста-Рика', 'mr': 'कोस्टा रिका', 'ms': 'Costa Rica', 'mt': 'il-Costa Rica', 'mua': 'kosta Rika', 'my': 'ကို့စ်တာရီကာ', 'mzn': 'کاستاریکا', 'naq': 'Costa Rica', 'nb': 'Costa Rica', 'nd': 'Khosta Rikha', 'ne': 'कोष्टारिका', 'nl': 'Costa Rica', 'nmg': 'Kosta Ríka', 'nn': 'Costa Rica', 'no': 'Costa Rica', 'nqo': 'ߞߐߛߑߕߊ߫ ߙߌߞߊ߫', 'nus': 'Kothtirika', 'nyn': 'Kositarika', 'om': 'Kostaa Rikaa', 'or': 'କୋଷ୍ଟା ରିକା', 'pa': 'ਕੋਸਟਾ ਰੀਕਾ', 'pa-Guru': 'ਕੋਸਟਾ ਰੀਕਾ', 'pcm': 'Kósta Ríka', 'pl': 'Kostaryka', 'ps': 'کوستاریکا', 'pt': 'Costa Rica', 'qu': 'Costa Rica', 'rm': 'Costa Rica', 'rn': 'Kositarika', 'ro': 'Costa Rica', 'rof': 'Kostarika', 'ru': 'Коста-Рика', 'rwk': 'Kostarika', 'saq': 'Kostarika', 'sbp': 'Kositalika', 'sc': 'Costa Rica', 'sd': 'ڪوسٽا ريڪا', 'sd-Arab': 'ڪوسٽا ريڪا', 'se': 'Costa Rica', 'seh': 'Costa Rica', 'ses': 'Kosta rika', 'sg': 'Kôsta Rîka', 'shi': 'ⴽⵓⵙⵜⴰ ⵔⵉⴽⴰ', 'shi-Latn': 'kusta rika', 'shi-Tfng': 'ⴽⵓⵙⵜⴰ ⵔⵉⴽⴰ', 'si': 'කොස්ටරිකාව', 'sk': 'Kostarika', 'sl': 'Kostarika', 'smn': 'Costa Rica', 'sn': 'Kostarika', 'so': 'Costa Rica', 'sq': 'Kosta-Rikë', 'sr': 'Костарика', 'sr-Cyrl': 'Костарика', 'sr-Latn': 'Kostarika', 'sv': 'Costa Rica', 'sw': 'Kostarika', 'syr': 'ܟܘܣܬܐ ܪܝܩܐ', 'ta': 'கோஸ்டாரிகா', 'te': 'కోస్టా రికా', 'teo': 'Kostarika', 'tg': 'Коста-Рика', 'th': 'คอสตาริกา', 'ti': 'ኮስታ ሪካ', 'tk': 'Kosta-Rika', 'to': 'Kosita Lika', 'tr': 'Kosta Rika', 'tt': 'Коста-Рика', 'twq': 'Kosta rika', 'tzm': 'Kusṭa Rika', 'ug': 'كوستارىكا', 'uk': 'Коста-Рика', 'ur': 'کوسٹا ریکا', 'uz': 'Kosta-Rika', 'uz-Cyrl': 'Коста-Рика', 'uz-Latn': 'Kosta-Rika', 'vai': 'ꖏꔻꕚ ꔸꕪ', 'vai-Latn': 'Kósíta Ríko', 'vai-Vaii': 'ꖏꔻꕚ ꔸꕪ', 'vec': 'Costa Rica', 'vi': 'Costa Rica', 'vun': 'Kostarika', 'wae': 'Costa Rica', 'wo': 'Kosta Rika', 'xh': 'ECosta Rica', 'xnr': 'कोस्टारिका', 'xog': 'Kosita Rika', 'yav': 'Kóstálíka', 'yi': 'קאסטאַ ריקאַ', 'yo': 'Kuusita Ríkà', 'yrl': 'Kupé Rika', 'yue': '哥斯大黎加', 'yue-Hans': '哥斯大黎加', 'yue-Hant': '哥斯大黎加', 'zgh': 'ⴽⵓⵙⵜⴰ ⵔⵉⴽⴰ', 'zh': '哥斯达黎加', 'zh-Hans': '哥斯达黎加', 'zh-Hant': '哥斯大黎加', 'zu': 'i-Costa Rica'},
    'CS': {'en': 'Serbia and Montenegro'},
    'CU': {'af': 'Kuba', 'agq': 'Kuuwbà', 'ak': 'Kuba', 'am': 'ኩባ', 'ar': 'كوبا', 'as': 'কিউবা', 'asa': 'Kuba', 'ast': 'Cuba', 'az': 'Kuba', 'az-Cyrl': 'Куба', 'az-Latn': 'Kuba', 'bal-Latn': 'Kyubá', 'bas': 'Kubà', 'be': 'Куба', 'be-tarask': 'Куба', 'bez': 'Hukuba', 'bg': 'Куба', 'bgn': 'کوبا', 'blo': 'Kubaa', 'bm': 'Kuba', 'bn': 'কিউবা', 'br': 'Kuba', 'brx': 'क्यूबा', 'bs': 'Kuba', 'bs-Cyrl': 'Куба', 'bs-Latn': 'Kuba', 'ca': 'Cuba', 'ca-ES-valencia': 'Cuba', 'ccp': '𑄇𑄨𑄃𑄪𑄝', 'ce': 'Куба', 'ceb': 'Cuba', 'cgg': 'Cuba', 'chr': 'ᎫᏆ', 'ckb': 'کووبا', 'cs': 'Kuba', 'cv': 'Куба', 'cy': 'Ciwba', 'da': 'Cuba', 'dav': 'Kuba', 'de': 'Kuba', 'dje': 'Kuuba', 'dsb': 'Kuba', 'dyo': 'Kuba', 'dz': 'ཀིའུ་བྷ', 'ebu': 'Kuba', 'ee': 'Kuba nutome', 'el': 'Κούβα', 'el-polyton': 'Κούβα', 'en': 'Cuba', 'en-Dsrt': '𐐗𐐷𐐭𐐺𐐲', 'eo': 'Kubo', 'es': 'Cuba', 'et': 'Kuuba', 'eu': 'Kuba', 'ewo': 'Kubá', 'fa': 'کوبا', 'ff': 'Kubaa', 'ff-Adlm': '𞤑𞤵𞥅𞤦𞤢𞥄', 'ff-Latn': 'Kubaa', 'fi': 'Kuuba', 'fil': 'Cuba', 'fo': 'Kuba', 'fr': 'Cuba', 'fur': 'Cuba', 'fy': 'Kuba', 'ga': 'Cúba', 'gd': 'Cùba', 'gl': 'Cuba', 'gsw': 'Kuba', 'gu': 'ક્યુબા', 'guz': 'Kuba', 'ha': 'Kyuba', 'he': 'קובה', 'hi': 'क्यूबा', 'hi-Latn': 'Cuba', 'hr': 'Kuba', 'hsb': 'Kuba', 'ht': 'Cuba', 'hu': 'Kuba', 'hy': 'Կուբա', 'ia': 'Cuba', 'id': 'Kuba', 'ig': 'Cuba', 'is': 'Kúba', 'it': 'Cuba', 'ja': 'キューバ', 'jgo': 'Kúba', 'jmc': 'Kuba', 'jv': 'Kuba', 'ka': 'კუბა', 'kaa': 'Куба', 'kaa-Cyrl': 'Куба', 'kab': 'Kuba', 'kam': 'Kuba', 'kde': 'Kuba', 'kea': 'Kuba', 'kgp': 'Kuma', 'khq': 'Kuuba', 'ki': 'Kiumba', 'kk': 'Куба', 'kk-Cyrl': 'Куба', 'kln': 'Emetab Cuba', 'km': 'គុយបា', 'kn': 'ಕ್ಯೂಬಾ', 'ko': '쿠바', 'kok': 'क्युबा', 'kok-Deva': 'क्युबा', 'ks': 'کیوٗبا', 'ks-Arab': 'کیوٗبا', 'ksb': 'Kuba', 'ksf': 'kuba', 'ksh': 'Kuhba', 'ku': 'Kuba', 'kxv': 'kyuba', 'kxv-Deva': 'क्यूबा', 'kxv-Latn': 'kyuba', 'kxv-Orya': 'କ୍ୟୁବା', 'kxv-Telu': 'క్యూబా', 'ky': 'Куба', 'lag': 'Kyúuba', 'lb': 'Kuba', 'lg': 'Cuba', 'lld': 'Cuba', 'ln': 'Kiba', 'lo': 'ຄິວບາ', 'lt': 'Kuba', 'lu': 'Kuba', 'luo': 'Cuba', 'luy': 'Kuba', 'lv': 'Kuba', 'mai': 'क्यूबा', 'mas': 'Kuba', 'mer': 'Kiuba', 'mfe': 'Cuba', 'mg': 'Kiobà', 'mgh': 'Ukuba', 'mi': 'Kiupa', 'mk': 'Куба', 'ml': 'ക്യൂബ', 'mn': 'Куба', 'mr': 'क्यूबा', 'ms': 'Cuba', 'mt': 'Kuba', 'mua': 'Kuba', 'my': 'ကျူးဘား', 'mzn': 'کوبا', 'naq': 'Cubab', 'nb': 'Cuba', 'nd': 'Cuba', 'ne': 'क्युबा', 'nl': 'Cuba', 'nmg': 'Kuba', 'nn': 'Cuba', 'no': 'Cuba', 'nqo': 'ߞߎ߳ߓߊ߫', 'nyn': 'Cuba', 'om': 'Kuubaa', 'or': 'କ\u200d୍ୟୁବା', 'pa': 'ਕਿਊਬਾ', 'pa-Guru': 'ਕਿਊਬਾ', 'pcm': 'Kiúbá', 'pl': 'Kuba', 'ps': 'کیوبا', 'pt': 'Cuba', 'qu': 'Cuba', 'rm': 'Cuba', 'rn': 'Kiba', 'ro': 'Cuba', 'rof': 'Kuba', 'ru': 'Куба', 'rwk': 'Kuba', 'sah': 'Кууба', 'saq': 'Kuba', 'sbp': 'Kuba', 'sc': 'Cuba', 'sd': 'ڪيوبا', 'sd-Arab': 'ڪيوبا', 'se': 'Kuba', 'seh': 'Cuba', 'ses': 'Kuuba', 'sg': 'Kubäa', 'shi': 'ⴽⵓⴱⴰ', 'shi-Latn': 'kuba', 'shi-Tfng': 'ⴽⵓⴱⴰ', 'si': 'කියුබාව', 'sk': 'Kuba', 'sl': 'Kuba', 'smn': 'Kuuba', 'sn': 'Cuba', 'so': 'Kuuba', 'sq': 'Kubë', 'sr': 'Куба', 'sr-Cyrl': 'Куба', 'sr-Latn': 'Kuba', 'sv': 'Kuba', 'sw': 'Kuba', 'syr': 'ܟܘܒܐ', 'ta': 'கியூபா', 'te': 'క్యూబా', 'teo': 'Kuba', 'tg': 'Куба', 'th': 'คิวบา', 'ti': 'ኩባ', 'tk': 'Kuba', 'to': 'Kiupa', 'tr': 'Küba', 'tt': 'Куба', 'twq': 'Kuuba', 'tzm': 'kuba', 'ug': 'كۇبا', 'uk': 'Куба', 'ur': 'کیوبا', 'uz': 'Kuba', 'uz-Cyrl': 'Куба', 'uz-Latn': 'Kuba', 'vai': 'ꕃꖳꕑ', 'vai-Latn': 'Kiyubha', 'vai-Vaii': 'ꕃꖳꕑ', 'vec': 'Cuba', 'vi': 'Cuba', 'vun': 'Kuba', 'wae': 'Kuba', 'wo': 'Kuba', 'xh': 'ECuba', 'xnr': 'क्यूबा', 'xog': 'Cuba', 'yav': 'kúpa', 'yi': 'קובאַ', 'yo': 'Kúbà', 'yrl': 'Kuba', 'yue': '古巴', 'yue-Hans': '古巴', 'yue-Hant': '古巴', 'zgh': 'ⴽⵓⴱⴰ', 'zh': '古巴', 'zh-Hans': '古巴', 'zh-Hant': '古巴', 'zu': 'i-Cuba'},
    'CV': {'af': 'Kaap Verde', 'agq': 'Chwɨla ŋ̀ Kɛ̀b Vɛ̂ɛ', 'ak': 'Kepvɛdfo Islands', 'am': 'ኬፕቨርዴ', 'ar': 'الرأس الأخضر', 'as': 'কেপ ভার্দে', 'asa': 'Kepuvede', 'ast': 'Cabu Verde', 'az': 'Kabo-Verde', 'az-Cyrl': 'Кабо-Верде', 'az-Latn': 'Kabo-Verde', 'bal-Latn': 'Kap Wardé', 'bas': 'Kabwɛ᷆r', 'be': 'Каба-Вердэ', 'be-tarask': 'Каба-Вердэ', 'bez': 'Hukepuvede', 'bg': 'Кабо Верде', 'blo': 'Kapfɛɛr', 'bm': 'Capivɛrdi', 'bn': 'কেপ ভার্দে', 'br': 'Kab-Glas', 'brx': 'केप भेर्दे', 'bs': 'Zelenortska Ostrva', 'bs-Cyrl': 'Зеленортска Острва', 'bs-Latn': 'Zelenortska Ostrva', 'ca': 'Cap Verd', 'ca-ES-valencia': 'Cap Verd', 'ccp': '𑄇𑄬𑄛𑄴𑄞𑄢𑄴𑄘𑄬', 'ce': 'Кабо-Верде', 'ceb': 'Cape Verde', 'cgg': 'Ebizinga bya Kepuvade', 'chr': 'ᎢᎬᎾᏕᎾ ᎢᏤᏳᏍᏗ', 'ckb': 'کەیپڤەرد', 'cs': 'Kapverdy', 'cv': 'Кабо-Верде', 'cy': 'Cabo Verde', 'da': 'Kap Verde', 'dav': 'Kepuvede', 'de': 'Cabo Verde', 'dje': 'Kapuver gungey', 'dsb': 'Kap Verde', 'dyo': 'Kap Ver', 'dz': 'ཀེཔ་བཱཌ', 'ebu': 'Kepuvede', 'ee': 'Kape Verde nutome', 'el': 'Πράσινο Ακρωτήριο', 'el-polyton': 'Πράσινο Ακρωτήριο', 'en': 'Cape Verde', 'en-Dsrt': '𐐗𐐩𐐹 𐐚𐐯𐑉𐐼𐐨', 'eo': 'Kaboverdo', 'es': 'Cabo Verde', 'et': 'Roheneemesaared', 'eu': 'Cabo Verde', 'ewo': 'Minlán Mí Káb Vɛr', 'fa': 'کیپ\u200cورد', 'ff': 'Duuɗe Kap Weer', 'ff-Adlm': '𞤑𞤢𞥄𞤦𞤮 𞤜𞤫𞤪𞤣𞤫', 'ff-Latn': 'Duuɗe Kap Weer', 'fi': 'Kap Verde', 'fil': 'Cape Verde', 'fo': 'Grønhøvdaoyggjar', 'fr': 'Cap-Vert', 'fur': 'Cjâf vert', 'fy': 'Kaapverdië', 'ga': 'Rinn Verde', 'gd': 'An Ceap Uaine', 'gl': 'Cabo Verde', 'gsw': 'Kap Verde', 'gu': 'કૅપ વર્ડે', 'guz': 'Kepuvede', 'ha': 'Tsibiran Cape Verde', 'he': 'כף ורדה', 'hi': 'केप वर्ड', 'hi-Latn': 'Cape Verde', 'hr': 'Zelenortska Republika', 'hsb': 'Kap Verde', 'ht': 'Cap-Vert', 'hu': 'Zöld-foki Köztársaság', 'hy': 'Կաբո Վերդե', 'ia': 'Capo Verde', 'id': 'Tanjung Verde', 'ig': 'Cape Verde', 'is': 'Grænhöfðaeyjar', 'it': 'Capo Verde', 'ja': 'カーボベルデ', 'jmc': 'Kepuvede', 'jv': 'Pongol Verdé', 'ka': 'კაბო-ვერდე', 'kaa': 'Кабо-Верде', 'kaa-Cyrl': 'Кабо-Верде', 'kab': 'Tigzirin n yixef azegzaw', 'kam': 'Kepuvede', 'kde': 'Kepuvede', 'kea': 'Kabu Verdi', 'kgp': 'Pu Tánh', 'khq': 'Kapuver gungey', 'ki': 'Kepuvede', 'kk': 'Кабо-Верде', 'kk-Cyrl': 'Кабо-Верде', 'kln': 'Ikwembeyotab Cape Verde', 'km': 'កាប់វែរ', 'kn': 'ಕೇಪ್ ವರ್ಡೆ', 'ko': '카보베르데', 'kok': 'केप वर्दी', 'kok-Deva': 'केप वर्दी', 'ks': 'کیپ ؤرڑی', 'ks-Arab': 'کیپ ؤرڑی', 'ksb': 'Kepuvede', 'ksf': 'kapvɛr', 'ksh': 'de kapvärdesche Enselle', 'ku': 'Kap Verde', 'kxv': 'kep vrḍi', 'kxv-Deva': 'केप वर्ड', 'kxv-Latn': 'kep vrḍi', 'kxv-Orya': 'କେପ୍ ୱଡ଼', 'kxv-Telu': 'కేప్ వడ్', 'ky': 'Кабо-Верде', 'lag': 'Kepuvéede', 'lb': 'Kap Verde', 'lg': 'Bizinga by’e Kepu Veredi', 'lld': 'Capo Verde', 'ln': 'Bisanga bya Kapevɛrɛ', 'lo': 'ເຄບ ເວີດ', 'lt': 'Žaliasis Kyšulys', 'lu': 'Lutanda lua Kapevele', 'luo': 'Cape Verde Islands', 'luy': 'Kepuvede', 'lv': 'Kaboverde', 'mai': 'कैप बर्डे', 'mas': 'Kepuvede', 'mer': 'Kepuvede', 'mfe': 'Kap-Ver', 'mg': 'Nosy Cap-Vert', 'mi': 'Te Kūrae Matomato', 'mk': 'Кабо Верде', 'ml': 'കേപ്പ് വേർഡ്', 'mn': 'Кабо-Верде', 'mr': 'केप व्हर्डे', 'ms': 'Cape Verde', 'mt': 'Cape Verde', 'mua': 'kap ma laŋne', 'my': 'ကိတ်ဗာဒီ', 'mzn': 'کیپ ورد', 'naq': 'Cape Verde Islands', 'nb': 'Kapp Verde', 'nd': 'Cape Verde Islands', 'ne': 'केप भर्डे', 'nl': 'Kaapverdië', 'nmg': 'Maŋ́ má Kapvɛr', 'nn': 'Kapp Verde', 'no': 'Kapp Verde', 'nqo': 'ߜߙߋߞߎ߲߫-ߝߙߌߛߌ߫', 'nus': 'Kɛp bedi ɣa̱ylɛn', 'nyn': 'Ebizinga bya Kepuvade', 'om': 'Keeppi Vaardee', 'or': 'କେପ୍ ଭର୍ଦେ', 'pa': 'ਕੇਪ ਵਰਡੇ', 'pa-Guru': 'ਕੇਪ ਵਰਡੇ', 'pcm': 'Kép Vẹ́d', 'pl': 'Republika Zielonego Przylądka', 'ps': 'کیپ ورد', 'pt': 'Cabo Verde', 'qu': 'Cabo Verde', 'rm': 'Cap Verd', 'rn': 'Ibirwa bya Kapuveri', 'ro': 'Capul Verde', 'rof': 'Kepuvede', 'ru': 'Кабо-Верде', 'rwk': 'Kepuvede', 'saq': 'Kepuvede', 'sbp': 'Kepuvede', 'sc': 'Cabu birde', 'sd': 'ڪيپ وردي', 'sd-Arab': 'ڪيپ وردي', 'se': 'Kap Verde', 'seh': 'Cabo Verde', 'ses': 'Kapuver gungey', 'sg': 'Azûâ tî Kâpo-Vêre', 'shi': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⴽⴰⴱⴱⵉⵔⴷⵉ', 'shi-Latn': 'tigzirin n kabbirdi', 'shi-Tfng': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⴽⴰⴱⴱⵉⵔⴷⵉ', 'si': 'කේප් වර්ඩ්', 'sk': 'Kapverdy', 'sl': 'Zelenortski otoki', 'smn': 'Cape Verde', 'sn': 'Zvitsuwa zveCape Verde', 'so': 'Jasiiradda Kayb Faarde', 'sq': 'Kepi i Gjelbër', 'sr': 'Зеленортска Острва', 'sr-Cyrl': 'Зеленортска Острва', 'sr-Latn': 'Zelenortska Ostrva', 'sv': 'Kap Verde', 'sw': 'Cape Verde', 'syr': 'ܟܐܦ ܒܝܪܕܝ (ܪܝܫܐ ܝܘܪܩܐ)', 'ta': 'கேப் வெர்டே', 'te': 'కేప్ వెర్డె', 'teo': 'Kepuvede', 'tg': 'Кабо-Верде', 'th': 'เคปเวิร์ด', 'ti': 'ኬፕ ቨርደ', 'tk': 'Kabo-Werde', 'to': 'Muiʻi Vēte', 'tr': 'Cabo Verde', 'tt': 'Кабо-Верде', 'twq': 'Kapuver gungey', 'tzm': 'Tigzirin n Iɣf Uzegzaw', 'ug': 'يېشىل تۇمشۇق', 'uk': 'Кабо-Верде', 'ur': 'کیپ ورڈی', 'uz': 'Kabo-Verde', 'uz-Cyrl': 'Кабо-Верде', 'uz-Latn': 'Kabo-Verde', 'vai': 'ꔞꔪ ꗲꔵ ꔳꘋꗣ', 'vai-Latn': 'Kepi Vɛdi Tiŋŋɛ', 'vai-Vaii': 'ꔞꔪ ꗲꔵ ꔳꘋꗣ', 'vec': 'Cao Verdo', 'vi': 'Cape Verde', 'vun': 'Kepuvede', 'wae': 'Kap Verde', 'wo': 'Kabo Werde', 'xh': 'ECape Verde', 'xnr': 'केप वर्ड', 'xog': 'Ebizinga bya Kepu Veredi', 'yav': 'Kápfɛl', 'yi': 'קאַפּווערדישע אינזלען', 'yo': 'Etíokun Kápé féndè', 'yrl': 'Kabu Suikiri', 'yue': '維德角', 'yue-Hans': '维德角', 'yue-Hant': '維德角', 'zgh': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⴽⴰⴱⴱⵉⵔⴷⵉ', 'zh': '佛得角', 'zh-Hans': '佛得角', 'zh-Hant': '維德角', 'zu': 'i-Cape Verde'},
    'CW': {'af': 'Curaçao', 'ak': 'Kurakaw', 'am': 'ኩራሳዎ', 'ar': 'كوراساو', 'as': 'কুৰাকাও', 'ast': 'Curaçao', 'az': 'Kurasao', 'az-Cyrl': 'Курасао', 'az-Latn': 'Kurasao', 'bal-Latn': 'Churácháó', 'be': 'Кюрасаа', 'be-tarask': 'Кюрасаа', 'bg': 'Кюрасао', 'blo': 'Kurasawuu', 'bn': 'কুরাসাও', 'br': 'Curaçao', 'brx': 'कुरासाव', 'bs': 'Kurasao', 'bs-Cyrl': 'Курасао', 'bs-Latn': 'Kurasao', 'ca': 'Curaçao', 'ca-ES-valencia': 'Curaçao', 'ccp': '𑄇𑄨𑄃𑄪𑄢𑄥𑄃𑄮', 'ce': 'Кюрасао', 'ceb': 'Curacao', 'chr': 'ᎫᎳᎨᎣ', 'ckb': 'کوراچاو', 'cs': 'Curaçao', 'cv': 'Кюрасао', 'cy': 'Curaçao', 'da': 'Curaçao', 'de': 'Curaçao', 'dsb': 'Curaçao', 'dz': 'ཀྱཱུར་ར་ཀོ', 'el': 'Κουρασάο', 'el-polyton': 'Κουρασάο', 'en': 'Curaçao', 'es': 'Curazao', 'et': 'Curaçao', 'eu': 'Curaçao', 'fa': 'کوراسائو', 'ff-Adlm': '𞤑𞤵𞥅𞤪𞤢𞤧𞤢𞤱𞤮', 'fi': 'Curaçao', 'fil': 'Curaçao', 'fo': 'Curaçao', 'fr': 'Curaçao', 'fy': 'Curaçao', 'ga': 'Cúrasó', 'gd': 'Curaçao', 'gl': 'Curaçao', 'gu': 'ક્યુરાસાઓ', 'ha': 'Ƙasar Curaçao', 'he': 'קוראסאו', 'hi': 'कुरासाओ', 'hi-Latn': 'Curacao', 'hr': 'Curaçao', 'hsb': 'Curaçao', 'ht': 'Curaçao', 'hu': 'Curaçao', 'hy': 'Կյուրասաո', 'ia': 'Curaçao', 'id': 'Curaçao', 'ig': 'Kurakao', 'is': 'Curacao', 'it': 'Curaçao', 'ja': 'キュラソー', 'jv': 'Kurasao', 'ka': 'კიურასაო', 'kaa': 'Кюрасао', 'kaa-Cyrl': 'Кюрасао', 'kea': 'Kurasau', 'kgp': 'Kurasavo', 'kk': 'Кюрасао', 'kk-Cyrl': 'Кюрасао', 'km': 'កូរ៉ាកៅ', 'kn': 'ಕುರಾಕಾವ್', 'ko': '퀴라소', 'kok': 'कुरसावो', 'kok-Deva': 'कुरसावो', 'ks': 'کیوراکو', 'ks-Arab': 'کیوراکو', 'ksh': 'Curaçao', 'ku': 'Curaçao', 'kxv': 'kyurasao', 'kxv-Deva': 'क्यूरासाओ', 'kxv-Latn': 'kyurasao', 'kxv-Orya': 'କ୍ୟୁରାସାଅ', 'kxv-Telu': 'క్వురసో', 'ky': 'Кюрасао', 'lb': 'Curaçao', 'lld': 'Curaçao', 'lo': 'ຄູຣາຊາວ', 'lt': 'Kiurasao', 'lv': 'Kirasao', 'mai': 'कुरासाओ', 'mi': 'Kurahao', 'mk': 'Курасао', 'ml': 'കുറാകാവോ', 'mn': 'Кюрасао', 'mr': 'क्युरासाओ', 'ms': 'Curacao', 'mt': 'Curaçao', 'my': 'ကျူရေးကိုးစ်', 'mzn': 'کوراسائو', 'nb': 'Curaçao', 'ne': 'कुराकाओ', 'nl': 'Curaçao', 'nn': 'Curaçao', 'no': 'Curaçao', 'nqo': 'ߞߎߙߛߊߏ߫', 'om': 'Kurakowaa', 'or': 'କୁରାକାଓ', 'pa': 'ਕੁਰਾਕਾਓ', 'pa-Guru': 'ਕੁਰਾਕਾਓ', 'pap': 'Kòrsou', 'pcm': 'Kiurásao', 'pl': 'Curaçao', 'ps': 'کوراکاو', 'pt': 'Curaçao', 'qu': 'Curazao', 'rm': 'Curaçao', 'ro': 'Curaçao', 'ru': 'Кюрасао', 'sc': 'Curaçao', 'sd': 'ڪيوراسائو', 'sd-Arab': 'ڪيوراسائو', 'se': 'Curaçao', 'si': 'කුරකාවෝ', 'sk': 'Curaçao', 'sl': 'Curaçao', 'smn': 'Curaçao', 'so': 'Kurakaaw', 'sq': 'Kurasao', 'sr': 'Курасао', 'sr-Cyrl': 'Курасао', 'sr-Latn': 'Kurasao', 'sv': 'Curaçao', 'sw': 'Curacao', 'syr': 'ܟܘܪܐܟܘ', 'ta': 'குராகவ்', 'te': 'క్యూరసో', 'tg': 'Кюрасао', 'th': 'คูราเซา', 'ti': 'ኩራሳው', 'tk': 'Kýurasao', 'to': 'Kulasao', 'tr': 'Curaçao', 'tt': 'Кюрасао', 'ug': 'كۇراچاۋ', 'uk': 'Кюрасао', 'ur': 'کیوراکاؤ', 'uz': 'Kyurasao', 'uz-Cyrl': 'Кюрасао', 'uz-Latn': 'Kyurasao', 'vai': 'ꖴꕟꖇꕱ', 'vai-Vaii': 'ꖴꕟꖇꕱ', 'vec': 'Curaçao', 'vi': 'Curaçao', 'wo': 'Kursawo', 'xh': 'ECuraçao', 'xnr': 'क्यूरासाओ', 'yi': 'קוראַסאַא', 'yo': 'Curaçao', 'yrl': 'Kurasau', 'yue': '庫拉索', 'yue-Hans': '库拉索', 'yue-Hant': '庫拉索', 'zh': '库拉索', 'zh-Hans': '库拉索', 'zh-Hant': '庫拉索', 'zu': 'i-Curaçao'},
    'CX': {'af': 'Kerseiland', 'ak': 'Buronya Supɔ', 'am': 'ክሪስማስ ደሴት', 'ar': 'جزيرة كريسماس', 'as': 'খ্ৰীষ্টমাছ দ্বীপ', 'ast': 'Islla Christmas', 'az': 'Milad adası', 'az-Cyrl': 'Милад адасы', 'az-Latn': 'Milad adası', 'bal-Latn': 'Kresmes Islánd', 'be': 'Востраў Каляд', 'be-tarask': 'Востраў Каляд', 'bg': 'остров Рождество', 'blo': 'Nowɛl kaAtukǝltǝna', 'bn': 'ক্রিসমাস দ্বীপ', 'br': 'Enez Christmas', 'brx': 'ख्रिसमास द्वीपफोर', 'bs': 'Božićno ostrvo', 'bs-Cyrl': 'Божићно острво', 'bs-Latn': 'Božićno ostrvo', 'ca': 'Illa Christmas', 'ca-ES-valencia': 'Illa Christmas', 'ccp': '𑄇𑄳𑄢𑄨𑄥𑄴𑄟𑄥𑄴 𑄞𑄨𑄘𑄳𑄠', 'ce': 'ГӀайре ӏиса пайхӏамар вина де', 'ceb': 'Christmas Island', 'chr': 'ᏓᏂᏍᏓᏲᎯᎲ ᎤᎦᏚᏛᎢ', 'ckb': 'دوورگەی کریسمس', 'cs': 'Vánoční ostrov', 'cv': 'Раштав утравӗ', 'cy': 'Ynys y Nadolig', 'da': 'Juleøen', 'de': 'Weihnachtsinsel', 'dsb': 'Gódowne kupy', 'dz': 'ཁི་རིསྟ་མེས་མཚོ་གླིང', 'ee': 'Kristmas ƒudomekpo nutome', 'el': 'Νήσος των Χριστουγέννων', 'el-polyton': 'Νῆσος Χριστουγέννων', 'en': 'Christmas Island', 'en-Dsrt': '𐐗𐑉𐐮𐑅𐑋𐐲𐑅 𐐌𐑊𐐲𐑌𐐼', 'es': 'Isla de Navidad', 'et': 'Jõulusaar', 'eu': 'Christmas uhartea', 'fa': 'جزیرهٔ کریسمس', 'ff-Adlm': '𞤅𞤵𞤪𞤭𞥅𞤪𞤫 𞤑𞤭𞤪𞤧𞤭𞤥𞤢𞥄𞤧', 'fi': 'Joulusaari', 'fil': 'Christmas Island', 'fo': 'Jólaoyggjin', 'fr': 'Île Christmas', 'fur': 'Isule Christmas', 'fy': 'Krysteilan', 'ga': 'Oileán na Nollag', 'gd': 'Eilean na Nollaig', 'gl': 'Illa Christmas', 'gsw': 'Wienachts-Insle', 'gu': 'ક્રિસમસ આઇલેન્ડ', 'ha': 'Tsibirin Kirsmati', 'he': 'אי חג המולד', 'hi': 'क्रिसमस द्वीप', 'hi-Latn': 'Christmas Island', 'hr': 'Božićni Otok', 'hsb': 'Hodowna kupa', 'ht': 'Île Christmas', 'hu': 'Karácsony-sziget', 'hy': 'Սուրբ Ծննդյան կղզի', 'ia': 'Insula de Natal', 'id': 'Pulau Natal', 'ig': 'Agwaetiti Christmas', 'is': 'Jólaey', 'it': 'Isola Christmas', 'ja': 'クリスマス島', 'jv': 'Pulo Natal', 'ka': 'შობის კუნძული', 'kea': 'Ilha di Natal', 'kgp': 'Krĩtimỹnh Goj-vẽso', 'kk': 'Рождество аралы', 'kk-Cyrl': 'Рождество аралы', 'km': 'កោះ\u200bគ្រីស្មាស', 'kn': 'ಕ್ರಿಸ್ಮಸ್ ದ್ವೀಪ', 'ko': '크리스마스섬', 'kok': 'क्रिसमस आयलँड', 'kok-Deva': 'क्रिसमस आयलँड', 'ks': 'کرِسمَس جٔزیٖرٕ', 'ks-Arab': 'کرِسمَس جٔزیٖرٕ', 'ksh': 'de Weihnaachs-Ensel', 'ku': 'Girava Christmasê', 'kxv': 'kristmas dīp', 'kxv-Deva': 'क्रिसमस द्वीप', 'kxv-Latn': 'kristmas dīp', 'kxv-Orya': 'କ୍ରିସ୍ଟମାସ୍ ଦିପ', 'kxv-Telu': 'క్రిస్ట మాస్ దీపుపు', 'ky': 'Рождество аралы', 'lb': 'Chrëschtdagsinsel', 'lld': 'Isola dl Nadé', 'lo': 'ເກາະຄຣິສມາດ', 'lt': 'Kalėdų Sala', 'lv': 'Ziemsvētku sala', 'mai': 'क्रिसमस द्वीप', 'mi': 'Te Moutere Kirihimete', 'mk': 'Божиќен Остров', 'ml': 'ക്രിസ്മസ് ദ്വീപ്', 'mn': 'Зул сарын арал', 'mr': 'ख्रिसमस बेट', 'ms': 'Pulau Krismas', 'mt': 'il-Gżira Christmas', 'my': 'ခရစ်စမတ် ကျွန်း', 'mzn': 'کریسمس جزیره', 'nb': 'Christmasøya', 'ne': 'क्रिष्टमस टापु', 'nl': 'Christmaseiland', 'nn': 'Christmasøya', 'no': 'Christmasøya', 'nqo': 'ߞߙߌߛߑߕߌߡߊ߫ ߕߌ߲', 'om': 'Odola Kirismaas', 'or': 'ଖ୍ରୀଷ୍ଟମାସ ଦ୍ୱୀପ', 'pa': 'ਕ੍ਰਿਸਮਿਸ ਟਾਪੂ', 'pa-Guru': 'ਕ੍ਰਿਸਮਿਸ ਟਾਪੂ', 'pcm': 'Krísmás Aíland', 'pl': 'Wyspa Bożego Narodzenia', 'ps': 'د کريسمس ټاپو', 'pt': 'Ilha Christmas', 'qu': 'Isla Christmas', 'rm': 'Insla da Nadal', 'ro': 'Insula Christmas', 'ru': 'о-в Рождества', 'sc': 'Ìsula de sa Natividade', 'sd': 'ڪرسمس ٻيٽ', 'sd-Arab': 'ڪرسمس ٻيٽ', 'se': 'Juovllat-sullot', 'si': 'ක්\u200dරිස්මස් දූපත', 'sk': 'Vianočný ostrov', 'sl': 'Božični otok', 'smn': 'Juovlâsuálui', 'so': 'Jasiiradda Kirismas', 'sq': 'Ishulli i Krishtlindjes', 'sr': 'Божићно Острво', 'sr-Cyrl': 'Божићно Острво', 'sr-Latn': 'Božićno Ostrvo', 'sv': 'Julön', 'sw': 'Kisiwa cha Krismasi', 'syr': 'ܓܙܪܬܐ ܕܟܪܝܣܬܡܣ', 'ta': 'கிறிஸ்துமஸ் தீவு', 'te': 'క్రిస్మస్ దీవి', 'tg': 'Ҷазираи Крисмас', 'th': 'เกาะคริสต์มาส', 'ti': 'ደሴት ክሪስማስ', 'tk': 'Roždestwo adasy', 'to': 'Motu Kilisimasi', 'tr': 'Christmas Adası', 'tt': 'Раштуа утравы', 'ug': 'مىلاد ئارىلى', 'uk': 'Острів Різдва', 'ur': 'جزیرہ کرسمس', 'uz': 'Rojdestvo oroli', 'uz-Cyrl': 'Рождество ороли', 'uz-Latn': 'Rojdestvo oroli', 'vai': 'ꔞꔻꕮꔻ ꔳꘋꗣ', 'vai-Vaii': 'ꔞꔻꕮꔻ ꔳꘋꗣ', 'vec': 'Ìzola de Nadale', 'vi': 'Đảo Giáng Sinh', 'wae': 'Wienäčtsinslä', 'wo': 'Dunu Kirismas', 'xh': 'EChristmas Island', 'xnr': 'क्रिसमस द्वीप', 'yo': 'Erékùsù Christmas', 'yrl': 'Kapuãma Kiritima', 'yue': '聖誕島', 'yue-Hans': '圣诞岛', 'yue-Hant': '聖誕島', 'zh': '圣诞岛', 'zh-Hans': '圣诞岛', 'zh-Hant': '聖誕島', 'zu': 'i-Christmas Island'},
    'CY': {'af': 'Siprus', 'agq': 'Saekpùlù', 'ak': 'Saeprɔso', 'am': 'ሳይፕረስ', 'ar': 'قبرص', 'as': 'চাইপ্ৰাছ', 'asa': 'Kuprothi', 'ast': 'Xipre', 'az': 'Kipr', 'az-Cyrl': 'Кипр', 'az-Latn': 'Kipr', 'bal-Latn': 'Sáipras', 'bas': 'Kiprò', 'be': 'Кіпр', 'be-tarask': 'Кіпр', 'bez': 'Hukuprosi', 'bg': 'Кипър', 'bgn': 'قبرس', 'blo': 'Ciprɔs', 'bm': 'Cipri', 'bn': 'সাইপ্রাস', 'br': 'Kiprenez', 'brx': 'साइप्रास', 'bs': 'Kipar', 'bs-Cyrl': 'Кипар', 'bs-Latn': 'Kipar', 'ca': 'Xipre', 'ca-ES-valencia': 'Xipre', 'ccp': '𑄥𑄭𑄛𑄳𑄢𑄥𑄴', 'ce': 'Кипр', 'ceb': 'Cyprus', 'cgg': 'Saipurasi', 'chr': 'ᏌᎢᏆᏍ', 'ckb': 'قیبرس', 'cs': 'Kypr', 'cv': 'Кипр', 'cy': 'Cyprus', 'da': 'Cypern', 'dav': 'Kuprosi', 'de': 'Zypern', 'dje': 'Šiipur', 'dsb': 'Cypriska', 'dyo': 'Siipr', 'dz': 'སཱའི་པྲས', 'ebu': 'Kuprosi', 'ee': 'Saiprus nutome', 'el': 'Κύπρος', 'el-polyton': 'Κύπρος', 'en': 'Cyprus', 'en-Dsrt': '𐐝𐐴𐐹𐑉𐐲𐑅', 'eo': 'Kipro', 'es': 'Chipre', 'et': 'Küpros', 'eu': 'Zipre', 'ewo': 'Sipəlús', 'fa': 'قبرس', 'ff': 'Siipar', 'ff-Adlm': '𞤑𞤵𞤦𞤪𞤵𞥅𞤧', 'ff-Latn': 'Siipar', 'fi': 'Kypros', 'fil': 'Cyprus', 'fo': 'Kýpros', 'fr': 'Chypre', 'fur': 'Cipri', 'fy': 'Syprus', 'ga': 'an Chipir', 'gd': 'Cìopras', 'gl': 'Chipre', 'gsw': 'Zypere', 'gu': 'સાયપ્રસ', 'guz': 'Kuprosi', 'ha': 'Saifurus', 'he': 'קפריסין', 'hi': 'साइप्रस', 'hi-Latn': 'Cyprus', 'hr': 'Cipar', 'hsb': 'Cypern', 'ht': 'Chypre', 'hu': 'Ciprus', 'hy': 'Կիպրոս', 'ia': 'Cypro', 'id': 'Siprus', 'ig': 'Cyprus', 'is': 'Kýpur', 'it': 'Cipro', 'ja': 'キプロス', 'jmc': 'Kuprosi', 'jv': 'Siprus', 'ka': 'კვიპროსი', 'kaa': 'Кипр', 'kaa-Cyrl': 'Кипр', 'kab': 'Cipr', 'kam': 'Kuprosi', 'kde': 'Kuplosi', 'kea': 'Xipri', 'kgp': 'Sipre', 'khq': 'Šiipur', 'ki': 'Kuprosi', 'kk': 'Кипр', 'kk-Cyrl': 'Кипр', 'kln': 'Emetab Cyprus', 'km': 'ស៊ីប', 'kn': 'ಸೈಪ್ರಸ್', 'ko': '키프로스', 'kok': 'सायप्रस', 'kok-Deva': 'सायप्रस', 'kok-Latn': 'Siprus', 'ks': 'سائپرس', 'ks-Arab': 'سائپرس', 'ksb': 'Kuplosi', 'ksf': 'cíprɛ', 'ksh': 'Züppere', 'ku': 'Qibris', 'kxv': 'saipras', 'kxv-Deva': 'साइप्रस', 'kxv-Latn': 'saipras', 'kxv-Orya': 'ସାଇପ୍ରସ୍', 'kxv-Telu': 'సైప్రస్', 'ky': 'Кипр', 'lag': 'Kupuróosi', 'lb': 'Zypern', 'lg': 'Sipuriya', 'lld': 'Ziper', 'ln': 'Sípɛlɛ', 'lo': 'ໄຊປຣັສ', 'lt': 'Kipras', 'lu': 'Shipele', 'luo': 'Cyprus', 'luy': 'Kuprosi', 'lv': 'Kipra', 'mai': 'साइप्रस', 'mas': 'Kuprosi', 'mer': 'Caipurasi', 'mfe': 'Cyprus', 'mg': 'Sypra', 'mgh': 'Ukuprosi', 'mi': 'Haipara', 'mk': 'Кипар', 'ml': 'സൈപ്രസ്', 'mn': 'Кипр', 'mr': 'सायप्रस', 'ms': 'Cyprus', 'mt': 'Ċipru', 'mua': 'Syipriya', 'my': 'ဆိုက်ပရပ်စ်', 'mzn': 'قبرس', 'naq': 'Cyprus', 'nb': 'Kypros', 'nd': 'Cyprus', 'ne': 'साइप्रस', 'nl': 'Cyprus', 'nmg': 'Sipria', 'nn': 'Kypros', 'no': 'Kypros', 'nqo': 'ߛߌߔߑߙߎ߫', 'nyn': 'Saipurasi', 'om': 'Qoophiroos', 'or': 'ସାଇପ୍ରସ୍', 'pa': 'ਸਾਇਪ੍ਰਸ', 'pa-Guru': 'ਸਾਇਪ੍ਰਸ', 'pcm': 'Saíprọs', 'pl': 'Cypr', 'ps': 'قبرس', 'pt': 'Chipre', 'qu': 'Chipre', 'rm': 'Cipra', 'rn': 'Izinga rya Shipure', 'ro': 'Cipru', 'rof': 'Kuprosi', 'ru': 'Кипр', 'rwk': 'Kuprosi', 'saq': 'Kuprosi', 'sbp': 'Kupilosi', 'sc': 'Tzipru', 'sd': 'سائپرس', 'sd-Arab': 'سائپرس', 'se': 'Kypros', 'seh': 'Chipre', 'ses': 'Šiipur', 'sg': 'Sîpri', 'shi': 'ⵇⵓⴱⵔⵓⵙ', 'shi-Latn': 'qubrus', 'shi-Tfng': 'ⵇⵓⴱⵔⵓⵙ', 'si': 'සයිප්\u200dරසය', 'sk': 'Cyprus', 'sl': 'Ciper', 'smn': 'Kypros', 'sn': 'Cyprus', 'so': 'Qubrus', 'sq': 'Qipro', 'sr': 'Кипар', 'sr-Cyrl': 'Кипар', 'sr-Latn': 'Kipar', 'sv': 'Cypern', 'sw': 'Saiprasi', 'syr': 'ܩܘܦܪܘܣ', 'ta': 'சைப்ரஸ்', 'te': 'సైప్రస్', 'teo': 'Kuprosi', 'tg': 'Кипр', 'th': 'ไซปรัส', 'ti': 'ቆጵሮስ', 'tk': 'Kipr', 'to': 'Saipalesi', 'tr': 'Kıbrıs', 'tt': 'Кипр', 'twq': 'Šiipur', 'tzm': 'Qubrus', 'ug': 'سىپرۇس', 'uk': 'Кіпр', 'ur': 'قبرص', 'uz': 'Kipr', 'uz-Cyrl': 'Кипр', 'uz-Latn': 'Kipr', 'vai': 'ꕢꗡꖛꗐꔻ', 'vai-Latn': 'Saɛpurɔ', 'vai-Vaii': 'ꕢꗡꖛꗐꔻ', 'vec': 'Sipro', 'vi': 'Síp', 'vun': 'Kuprosi', 'wae': 'Zypre', 'wo': 'Siipar', 'xh': 'ECyprus', 'xnr': 'साइप्रस', 'xog': 'Sipuriya', 'yav': 'síplɛ', 'yo': 'Kúrúsì', 'yrl': 'Xipiri', 'yue': '賽普勒斯', 'yue-Hans': '赛普勒斯', 'yue-Hant': '賽普勒斯', 'zgh': 'ⵇⵓⴱⵔⵓⵙ', 'zh': '塞浦路斯', 'zh-Hans': '塞浦路斯', 'zh-Hant': '賽普勒斯', 'zu': 'i-Cyprus'},
    'CZ': {'af': 'Tsjeggië', 'agq': 'Chɛ̂ Lèkpubèlè', 'ak': 'Kyɛk', 'am': 'ቼቺያ', 'ar': 'التشيك', 'as': 'চিজেচিয়া', 'asa': 'Jamhuri ya Cheki', 'ast': 'Chequia', 'az': 'Çexiya', 'az-Cyrl': 'Чехија', 'az-Latn': 'Çexiya', 'bal-Latn': 'Chéchiá', 'be': 'Чэхія', 'be-tarask': 'Чэхія', 'bez': 'Ijamhuri ya Cheki', 'bg': 'Чехия', 'blo': 'Cɛk', 'bm': 'Ceki republiki', 'bn': 'চেকিয়া', 'br': 'Tchekia', 'brx': 'चेखिया', 'bs': 'Češka', 'bs-Cyrl': 'Чешка', 'bs-Latn': 'Češka', 'ca': 'Txèquia', 'ca-ES-valencia': 'Txèquia', 'ccp': '𑄌𑄬𑄌𑄨𑄠', 'ce': 'Чехи', 'ceb': 'Czechia', 'cgg': 'Ripaaburika ya Zeeki', 'chr': 'ᏤᎩᎠ', 'ckb': 'کۆماری چیک', 'cs': 'Česko', 'cv': 'Чехи', 'cy': 'Tsiecia', 'da': 'Tjekkiet', 'dav': 'Jamhuri ya Cheki', 'de': 'Tschechien', 'dje': 'Cek labo', 'dsb': 'Česka republika', 'dyo': 'Mofam mati Cek', 'dz': 'ཅེཀ་ རི་པབ་ལིཀ', 'ebu': 'Jamhuri ya Cheki', 'ee': 'Tsɛk repɔblik nutome', 'el': 'Τσεχία', 'el-polyton': 'Τσεχία', 'en': 'Czechia', 'en-Dsrt': '𐐕𐐯𐐿 𐐡𐐨𐐹𐐲𐐺𐑊𐐮𐐿', 'eo': 'Ĉeĥujo', 'es': 'Chequia', 'et': 'Tšehhi', 'eu': 'Txekia', 'ewo': 'Ǹnam Tsɛ́g', 'fa': 'چک', 'ff': 'Ndenndaandi Cek', 'ff-Adlm': '𞤕𞤫𞥅𞤳𞤭𞤴𞤢𞥄', 'ff-Latn': 'Ndenndaandi Cek', 'fi': 'Tšekki', 'fil': 'Czechia', 'fo': 'Kekkia', 'fr': 'Tchéquie', 'fur': 'Republiche ceche', 'fy': 'Tsjechje', 'ga': 'an tSeicia', 'gd': 'An t-Seic', 'gl': 'Chequia', 'gsw': 'Tschechischi Republik', 'gu': 'ચેકીયા', 'guz': 'Jamhuri ya Cheki', 'ha': 'Czechia', 'he': 'צ׳כיה', 'hi': 'चेकिया', 'hi-Latn': 'Czechia', 'hr': 'Češka', 'hsb': 'Čěska republika', 'ht': 'Tchéquie', 'hu': 'Csehország', 'hy': 'Չեխիա', 'ia': 'Chechia', 'id': 'Ceko', 'ie': 'Tchekia', 'ig': 'Czechia', 'is': 'Tékkland', 'it': 'Cechia', 'ja': 'チェコ', 'jmc': 'Jamhuri ya Cheki', 'jv': 'Céko', 'ka': 'ჩეხეთი', 'kaa': 'Чехия', 'kaa-Cyrl': 'Чехия', 'kab': 'Čček', 'kam': 'Jamhuri ya Cheki', 'kde': 'Jamuhuli ya Chechi', 'kea': 'Txékia', 'kgp': 'Sékija', 'khq': 'Cek labo', 'ki': 'Jamhuri ya Cheki', 'kk': 'Чехия', 'kk-Cyrl': 'Чехия', 'kln': 'Emetab Czech Republic', 'km': 'ឆែក', 'kn': 'ಝೆಕಿಯಾ', 'ko': '체코', 'kok': 'चेकिया', 'kok-Deva': 'चेकिया', 'ks': 'چیکیا', 'ks-Arab': 'چیکیا', 'ksb': 'Jamhuli ya Cheki', 'ksf': 'cɛ́k', 'ksh': 'de Tschäschei', 'ku': 'Çekya', 'kxv': 'cekiya', 'kxv-Deva': 'चेकिया', 'kxv-Latn': 'cekiya', 'kxv-Orya': 'ଚେକିଆ', 'kxv-Telu': 'చెకియా', 'ky': 'Чехия', 'lag': 'Jamuhúuri ya Chéeki', 'lb': 'Tschechien', 'lg': 'Lipubulika ya Ceeka', 'lld': 'Cechia', 'ln': 'Shekia', 'lo': 'ເຊັກເຊຍ', 'lt': 'Čekija', 'lu': 'Ditunga dya Tsheka', 'luo': 'Czech Republic', 'luy': 'Jamhuri ya Cheki', 'lv': 'Čehija', 'mai': 'चेकिया', 'mas': 'Jamhuri ya Cheki', 'mer': 'Nthĩ ya Cheki', 'mfe': 'Repiblik Chek', 'mg': 'Repoblikan’i Tseky', 'mgh': 'Ucheki', 'mi': 'Tiekia', 'mk': 'Чешка', 'ml': 'ചെക്കിയ', 'mn': 'Чех', 'mr': 'झेकिया', 'ms': 'Czechia', 'mt': 'ir-Repubblika Ċeka', 'mua': 'Sǝr Syek', 'my': 'ချက်ကီယား', 'mzn': 'چک جمهوری', 'naq': 'Czech Republiki', 'nb': 'Tsjekkia', 'nd': 'Czech Republic', 'ne': 'चेकिया', 'nl': 'Tsjechië', 'nmg': 'Nlambɔ́ bó tschɛk', 'nn': 'Tsjekkia', 'no': 'Tsjekkia', 'nqo': 'ߗߍߞ', 'nyn': 'Ripaaburika ya Zeeki', 'om': 'Cheechiya', 'or': 'ଚେଚିଆ', 'pa': 'ਚੈਕੀਆ', 'pa-Guru': 'ਚੈਕੀਆ', 'pcm': 'Chẹ́kia', 'pl': 'Czechy', 'ps': 'چکیا', 'pt': 'Tchéquia', 'qu': 'Chequia', 'rm': 'Tschechia', 'rn': 'Repubulika ya Ceke', 'ro': 'Cehia', 'rof': 'Jamhuri ya Cheki', 'ru': 'Чехия', 'rwk': 'Jamhuri ya Cheki', 'saq': 'Jamhuri ya Cheki', 'sbp': 'Jamuhuli ya Sheki', 'sc': 'Tzèchia', 'sd': 'چيڪيا', 'sd-Arab': 'چيڪيا', 'se': 'Čeahkka', 'seh': 'República Tcheca', 'ses': 'Cek labo', 'sg': 'Ködörösêse tî Tyêki', 'shi': 'ⵜⴰⴳⴷⵓⴷⴰⵏⵜ ⵜⴰⵜⵛⵉⴽⵉⵜ', 'shi-Latn': 'tagdudant tatcikit', 'shi-Tfng': 'ⵜⴰⴳⴷⵓⴷⴰⵏⵜ ⵜⴰⵜⵛⵉⴽⵉⵜ', 'si': 'චෙචියාව', 'sk': 'Česko', 'sl': 'Češka', 'smn': 'Tšekki', 'sn': 'Czech Republic', 'so': 'Jekiya', 'sq': 'Çeki', 'sr': 'Чешка', 'sr-Cyrl': 'Чешка', 'sr-Latn': 'Češka', 'sv': 'Tjeckien', 'sw': 'Chechia', 'syr': 'ܬܫܝܟܝܐ', 'ta': 'செசியா', 'te': 'చెకియా', 'teo': 'Jamhuri ya Cheki', 'tg': 'Ҷумҳурии Чех', 'th': 'เช็ก', 'ti': 'ቸክያ', 'tk': 'Çehiýa', 'to': 'Sēkia', 'tr': 'Çekya', 'tt': 'Чехия Республикасы', 'twq': 'Cek labo', 'tzm': 'Tagduda n Čik', 'ug': 'چېخ جۇمھۇرىيىتى', 'uk': 'Чехія', 'ur': 'چیکیا', 'uz': 'Chexiya', 'uz-Cyrl': 'Чехия', 'uz-Latn': 'Chexiya', 'vai': 'ꗿꕃ ꕸꖃꔀ', 'vai-Latn': 'Chɛki Boloe', 'vai-Vaii': 'ꗿꕃ ꕸꖃꔀ', 'vec': 'Cechia', 'vi': 'Séc', 'vun': 'Jamhuri ya Cheki', 'wae': 'Tšečie', 'wo': 'Réewum Cek', 'xh': 'ECzechia', 'xnr': 'चेकिया', 'xog': 'Lipabulika ya Ceeka', 'yav': 'kitɔŋ kí cɛ́k', 'yi': 'טשעכיי', 'yo': 'Ṣẹ́ẹ́kì', 'yrl': 'Xekiya', 'yue': '捷克', 'yue-Hans': '捷克', 'yue-Hant': '捷克', 'zgh': 'ⵜⴰⴳⴷⵓⴷⴰⵏⵜ ⵜⴰⵜⵛⵉⴽⵉⵜ', 'zh': '捷克', 'zh-Hans': '捷克', 'zh-Hant': '捷克', 'zu': 'i-Czechia'},
    'Cakm': {'af': 'Chakma', 'ak': 'Kyakma kasa', 'am': 'ቻክማ', 'ar': 'شاكما', 'as': 'চাকমা', 'ast': 'chakma', 'az': 'çakma', 'az-Latn': 'çakma', 'bal-Latn': 'Chakma', 'be': 'чакма', 'be-tarask': 'чакма', 'bg': 'чакма', 'bn': 'চাকমা', 'br': 'chakmaek', 'brx': 'चाकमा', 'bs': 'čakmansko pismo', 'bs-Cyrl': 'чакманско писмо', 'bs-Latn': 'čakmansko pismo', 'ca': 'chakma', 'ca-ES-valencia': 'chakma', 'ccp': '𑄌𑄇𑄴𑄟', 'chr': 'ᏣᎧᎹ', 'cs': 'čakma', 'cy': 'Chakma', 'da': 'chakma', 'de': 'Chakma', 'dsb': 'chakmaske pismo', 'el': 'Τσάκμα', 'el-polyton': 'Τσάκμα', 'en': 'Chakma', 'en-Dsrt': '𐐕𐐪𐐿𐑋𐐲', 'es': 'chakma', 'et': 'tšaakma', 'eu': 'txakma', 'fa': 'چاکمایی', 'ff-Adlm': '𞤕𞤢𞤳𞤥𞤢', 'fi': 'chakmalainen', 'fil': 'Chakma', 'fo': 'chakma', 'fr': 'chakma', 'fy': 'Chakma', 'ga': 'Seácmais', 'gd': 'Chakma', 'gl': 'chakma', 'gu': 'ચકમા', 'ha': 'Chakma', 'he': 'צ׳אקמה', 'hi': 'चकमा', 'hi-Latn': 'Chakma', 'hr': 'čakmansko pismo', 'hsb': 'chakmaske pismo', 'ht': 'chakma', 'hu': 'Csakma', 'hy': 'չակմա', 'id': 'Chakma', 'ig': 'Chakma', 'is': 'chakma', 'it': 'chakma', 'ja': 'チャクマ文字', 'jv': 'Chakma', 'ka': 'ჩაკმა', 'kgp': 'kagme', 'kk': 'чакма жазуы', 'kk-Cyrl': 'чакма жазуы', 'km': 'ចាក់ម៉ា', 'kn': 'ಚಕ್ಮಾ', 'ko': '차크마 문자', 'kok': 'चाक्मा', 'kok-Deva': 'चाक्मा', 'ky': 'Чакма (жазуу)', 'lo': 'ຊາກມາ', 'lt': 'čakma', 'lv': 'čakmu', 'mai': 'चकमा', 'mi': 'Tiakamā', 'mk': 'чакманско', 'ml': 'ചക്മ', 'mn': 'Чакма бичиг', 'mr': 'चकमा', 'ms': 'Chakma', 'my': 'ချပ်ခ်မာ', 'nb': 'chakma', 'ne': 'काक्म्', 'nl': 'Chakma', 'nn': 'chakma', 'no': 'chakma', 'nqo': 'ߗߍߞߑߡߊ', 'or': 'ଚକମା', 'pa': 'ਚਕਮਾ', 'pa-Guru': 'ਚਕਮਾ', 'pcm': 'Chákmá Lángwej', 'pl': 'chakma', 'ps': 'چکما', 'qu': 'Chakma Simi', 'rm': 'chakma', 'ro': 'chakma', 'ru': 'чакмийская', 'sc': 'chakma', 'sd': 'چڪما', 'sd-Arab': 'چڪما', 'si': 'චක්මා', 'sk': 'čakma', 'sl': 'čakmajski', 'so': 'Jakma', 'sq': 'çakma', 'sr': 'чакма', 'sr-Cyrl': 'чакма', 'sr-Latn': 'čakma', 'sv': 'chakma', 'sw': 'Kichakma', 'ta': 'சக்மா', 'te': 'చక్మా', 'tg': 'Чакама', 'th': 'จักม่า', 'ti': 'ቻክማ', 'tk': 'çakma', 'to': 'tohinima fakasakimā', 'tr': 'Chakma', 'ug': 'چاكما', 'uk': 'чакма', 'ur': 'چکما', 'uz': 'chakma', 'uz-Latn': 'chakma', 'vec': 'ciakma', 'vi': 'Chữ Chakma', 'xnr': 'चकमा', 'yo': 'Kami', 'yrl': 'kakimi', 'yue': '查克馬文', 'yue-Hans': '查克马文', 'yue-Hant': '查克馬文', 'zh': '查克马文', 'zh-Hans': '查克马文', 'zh-Hant': '查克馬文', 'zu': 'isi-Chakma'},
    'Cans': {'af': 'Verenigde Kanadese Inheemse Lettergreepskrif', 'ak': 'Kanadafoɔ Kann Kasa a Wɔakeka Abom', 'am': 'የተዋሐዱ የካናዳ አቦሪጂኖች ፊደላት', 'ar': 'مقاطع كندية أصلية موحدة', 'as': 'ইউনিফাইড কানাডিয়ান এব’ৰিজিনেল ছিলেবিকছ', 'ast': 'silábicu unificáu de los nativos canadienses', 'az': 'birləşmiş kanada yerli yazısı', 'az-Latn': 'birləşmiş kanada yerli yazısı', 'bal-Latn': 'Kaynadhái Asligén Silábi', 'be': 'складавае пісьмо канадскіх абарыгенаў', 'be-tarask': 'складавае пісьмо канадскіх абарыгенаў', 'bg': 'унифицирани символи на канадски аборигени', 'bn': 'সংযুক্ত কানাডিয়ান অ্যাবোরিজিনাল সিলেবিক্স', 'br': 'silabennaoueg engenidik unvan Kanada', 'brx': 'जथाय कानाडीयान थागिबियारि रिंसार हानजा', 'bs': 'ujedinjeni kanadski aboridžinski slogovi', 'bs-Cyrl': 'уједињени канадски абориџински силабици', 'bs-Latn': 'ujedinjeni kanadski aboridžinski slogovi', 'ca': 'sil·labari aborigen canadenc unificat', 'ca-ES-valencia': 'sil·labari aborigen canadenc unificat', 'ccp': '𑄎𑄧𑄙 𑄇𑄚𑄓𑄨𑄠𑄚𑄴 𑄃𑄳𑄠𑄝𑄳𑄢𑄮𑄎𑄨𑄚𑄨𑄠𑄚𑄴 𑄥𑄨𑄣𑄬𑄝𑄨𑄇𑄴𑄥𑄧', 'chr': 'ᏌᏊ ᎨᎾᏓ ᎠᏂᏴᏫᏯ ᏗᎪᏪᎸ', 'cs': 'slabičné písmo kanadských domorodců', 'csw': 'ᓀᐦᐃᔭᐘᓯᓇᐦᐃᑳᑌᐤ', 'cy': 'Meysydd Llafur Cynfrodorol Unedig Canada', 'da': 'oprindelige canadiske symboler', 'de': 'Kanadische Aborigine-Silbenschrift', 'dsb': 'zjadnotnjone kanadiske aboriginske złožkowe pismo', 'el': 'Ενοποιημένοι Καναδεζικοί Συλλαβισμοί Ιθαγενών', 'el-polyton': 'Ενοποιημένοι Καναδεζικοί Συλλαβισμοί Ιθαγενών', 'en': 'Unified Canadian Aboriginal Syllabics', 'en-Dsrt': '𐐏𐐭𐑌𐐮𐑁𐐴𐐼 𐐗𐐲𐑌𐐩𐐼𐐨𐐲𐑌 𐐈𐐺𐐬𐑉𐐮𐐾𐐲𐑊𐐲𐑊 𐐝𐐮𐑊𐐰𐐺𐐮𐐿𐑅', 'es': 'silabarios aborígenes canadienses unificados', 'et': 'Kanada põlisrahvaste ühtlustatud silpkiri', 'eu': 'Kanadako aborigenen silabario bateratua', 'fa': 'زبان\u200cهای سیلابی بومی\u200cهای متحد کانادایی', 'ff-Adlm': '𞤑𞤢𞤱𞤪𞤢𞤤 𞤅𞤭𞤺𞤢𞤲𞤯𞤫 𞤚𞤢𞥄𞤳𞤢𞤲𞤶𞤫 𞤑𞤢𞤲𞤢𞤣𞤢𞥄', 'fi': 'kanadalaisten alkuperäiskansojen yhtenäistetty tavukirjoitus', 'fil': 'Unified Canadian Aboriginal Syllabics', 'fo': 'sameindir kanadiskir uppruna stavir', 'fr': 'syllabaire autochtone canadien unifié', 'fur': 'Silabari unificât aborigjens canadês', 'fy': 'Verenigde Canadese Aboriginal-symbolen', 'ga': 'Siollach Bundúchasach Ceanadach Aontaithe', 'gd': 'Sgrìobhadh Lideach Aonaichte nan Tùsanach Canadach', 'gl': 'silabario aborixe canadense unificado', 'gsw': 'UCAS', 'gu': 'યુનાઇટેડ કેનેડિયન એબોરિજનલ સિલેબિક્સ', 'ha': 'Haɗaɗɗun Gaɓoɓin harshe na Asali na Kanada', 'he': 'כתב הברתי קנדי ילידי מאוחד', 'hi': 'यूनिफ़ाइड कैनेडियन एबोरिजनल सिलेबिक्स', 'hi-Latn': 'Unified Canadian Aboriginal Syllabics', 'hr': 'unificirani kanadski aboriđinski slogovi', 'hsb': 'zjednotnjene kanadske aboriginske złóžkowe pismo', 'ht': 'syllabaire autochtone canadien unifié', 'hu': 'Egyesített kanadai őslakos jelek', 'hy': 'կանադական միասնական վանկագիր', 'id': 'Simbol Aborigin Kanada Kesatuan', 'ig': 'Unified Canadian Aboriginal Syllabics', 'is': 'samræmt kanadískt samstöfuletur frumbyggja', 'it': 'sillabario aborigeno canadese unificato', 'ja': '統合カナダ先住民音節文字', 'jv': 'Wanda Manunggal Aborigin Kanada', 'ka': 'გაერთიანებული კანადური სილაბური', 'kgp': 'sirama-vẽnhrá pir kanỹna tá kanhgág', 'kk': 'канадалық буын жүйелі жазу', 'kk-Cyrl': 'канадалық буын жүйелі жазу', 'km': 'ព្យាង្គអាបូរីជីន\u200bកាណាដារួម', 'kn': 'ಯುನಿಫೈಯ್ಡ್ ಕೆನೆಡಿಯನ್ ಅಬೊರಿಜಿನಲ್ ಸಿಲ್ಯಾಬಿಕ್ಸ್', 'ko': '통합 캐나다 원주민 음절문자', 'kok': 'संयुक्त कॅनडियन ॲबओरीजिनल सिलेबिक', 'kok-Deva': 'संयुक्त कॅनडियन ॲबओरीजिनल सिलेबिक', 'ks': 'یُنِفایِڑ کنیڑِیَن ایب آرجِنَل سِلیبِک', 'ks-Arab': 'یُنِفایِڑ کنیڑِیَن ایب آرجِنَل سِلیبِک', 'ky': 'канадалык муун жазуусу', 'lb': 'UCAS', 'lo': 'ສັນຍາລັກຊົນເຜົ່າພື້ນເມືອງແຄນນາດາ', 'lt': 'suvienodinti Kanados aborigenų silabiniai', 'lv': 'vienotā Kanādas aborigēnu zilbju rakstība', 'mai': 'यूनिफाइड कनाडियम एबोरिजनल सिलेबिक्स', 'mi': 'Ngā Kūoro o ngā Iwi Taketake o Kānata kua paiheretia', 'mk': 'канадско слоговно', 'ml': 'ഏകീകൃത കനേഡിയൻ ഗോത്രലിപി', 'mn': 'Канадын уугуул иргэдийн тэмдэгт бичиг', 'mr': 'यूनिफाइड कॅनेडियन अ\u200dॅबोरिजनल सिलॅबिक्स', 'ms': 'Suku Kata Orang Asli Kanada Bersatu', 'my': 'ကင်န်စ်', 'nb': 'felles kanadiske urspråksstavelser', 'ne': 'एकीकृत क्यानेडाली आदिवादीको सिलाबिक्स', 'nl': 'Verenigde Canadese Aboriginal-symbolen', 'nn': 'felles kanadiske urspråksstavingar', 'no': 'felles kanadiske urspråksstavelser', 'nqo': 'ߞߣߊߘߊ߫ ߘߎ߰ߙߋ߲߬ ߛߓߍ߬ߛߎ߲߬ ߞߝߊ ߟߊߘߍ߬ߣߍ߲', 'or': 'ୟୁନିଫାଏଡ୍ କାନାଡିଆନ୍ ଆବ୍ରୋଜିନାଲ୍ ସିଲାବିକସ୍', 'pa': 'ਯੂਨੀਫਾਈਡ ਕੈਨੇਡੀਅਨ ਐਬੋਰਿਜਿਨਲ ਸਿਲੇਬਿਕਸ', 'pa-Guru': 'ਯੂਨੀਫਾਈਡ ਕੈਨੇਡੀਅਨ ਐਬੋਰਿਜਿਨਲ ਸਿਲੇਬਿਕਸ', 'pcm': 'Nétív Kánádá Pípul Sílebul-dẹm Wé Dẹm Jọín Togẹ́da', 'pl': 'zunifikowane symbole kanadyjskich autochtonów', 'ps': 'متحد کاناډایی ابوریجینل سلیبکس', 'pt': 'escrita silábica unificada dos aborígenes canadenses', 'qu': 'Silabeo aborigen Simi (Canadiense unificado)', 'rm': 'simbols autoctons canadais unifitgads', 'ro': 'silabică aborigenă canadiană unificată', 'ru': 'канадское слоговое письмо', 'sc': 'sillabàriu aborìgenu canadesu unificadu', 'sd': 'يونيفائيڊ ڪينيڊيئن ابارجيني سليبڪس', 'sd-Arab': 'يونيفائيڊ ڪينيڊيئن ابارجيني سليبڪس', 'si': 'ඒකාබද්ධ කැනේඩියානු ආදිවාසී විෂය නිර්දේශ', 'sk': 'zjednotené kanadské domorodé slabiky', 'sl': 'poenotena zlogovna pisava kanadskih staroselcev', 'so': 'Qoraalka Luuqada Aborajiinka ee Kanada', 'sq': 'rrokje të unifikuara aborigjene kanadeze', 'sr': 'уједињени канадски абориџински силабици', 'sr-Cyrl': 'уједињени канадски абориџински силабици', 'sr-Latn': 'ujedinjeni kanadski aboridžinski silabici', 'sv': 'kanadensiska stavelsetecken', 'sw': 'Silabi Zilizounganishwa za Wakazi Asili wa Kanada', 'ta': 'ஒருங்கிணைக்கப்பட்ட கனடிய பழங்குடி எழுத்துகள்', 'te': 'యునిఫైడ్ కెనెడియన్ అబొరిజినల్ సిలబిక్స్', 'tg': 'Низоми ягонаи ҳиҷои аборигении каннадӣ', 'th': 'อักษรพยางค์ของชาวอะบอริจินในแคนาดา', 'ti': 'ውሁድ ካናዳዊ ኣቦርጅናል ሲላቢክስ', 'tk': 'Kanadanyň ýerlileriniň bogunlarynyň bitewileşdirilen ulgamy', 'to': 'tohinima fakatupuʻi-kānata-fakatahataha', 'tr': 'UCAS', 'ug': 'بىرلىككە كەلگەن كانادا يەرلىك بوغۇم جەدۋىلى', 'uk': 'уніфіковані символи канадських тубільців', 'ur': 'متحدہ کینیڈین ایبوریجنل سلیبکس', 'uz': 'kanada boʻgʻin yozuvi', 'uz-Latn': 'kanada boʻgʻin yozuvi', 'vec': 'silabaro aborìjeno canadeze unifegà', 'vi': 'Âm tiết Thổ dân Canada Hợp nhất', 'xnr': 'युनिफाइड कैनेडियन एबोरिजनल सिलेबिक्स', 'yo': 'Èdè Apapo Onile Onisilebu ti Kanada', 'yrl': 'yũpinimasá síraba irũ aburíjini kanadáwara suí', 'yue': '加拿大原住民通用字符', 'yue-Hans': '加拿大原住民通用字符', 'yue-Hant': '加拿大原住民通用字符', 'zh': '加拿大土著统一音节', 'zh-Hans': '加拿大土著统一音节', 'zh-Hant': '加拿大原住民通用字符', 'zu': 'i-Unified Canadian Aboriginal Syllabics'},
    'Cari': {'ar': 'الكارية', 'ast': 'cariu', 'az': 'kariyan', 'az-Latn': 'kariyan', 'bal-Latn': 'Charian', 'bg': 'Карийска', 'bn': 'ক্যারিয়ান', 'brx': 'कारियन', 'bs': 'karijsko pismo', 'bs-Cyrl': 'каријско писмо', 'bs-Latn': 'karijsko pismo', 'ca': 'carià', 'ca-ES-valencia': 'carià', 'ccp': '𑄇𑄳𑄠𑄢𑄨𑄠𑄚𑄴', 'cs': 'karijské', 'da': 'kariansk', 'de': 'Karisch', 'el': 'Καριάν', 'el-polyton': 'Καριάν', 'en': 'Carian', 'en-Dsrt': '𐐗𐐱𐑉𐐨𐐲𐑌', 'es': 'cario', 'et': 'kaaria', 'eu': 'kariera', 'fa': 'کاری', 'ff-Adlm': '𞤑𞤢𞤪𞤭𞤴𞤢𞤲', 'fi': 'kaarialainen', 'fo': 'carian', 'fr': 'carien', 'fy': 'Karysk', 'ga': 'Cló Cairiach', 'gd': 'Carian', 'gsw': 'Karisch', 'gu': 'કરૈન', 'he': 'קריאן', 'hi': 'करैन', 'hi-Latn': 'Carian', 'hr': 'karijsko pismo', 'ht': 'carien', 'hu': 'Kari', 'id': 'Karia', 'ig': 'Carian', 'it': 'carian', 'ja': 'カリア文字', 'ka': 'კარიული', 'kgp': 'karijỹnũ', 'kk': 'карий жазуы', 'kk-Cyrl': 'карий жазуы', 'km': 'ខារី', 'kn': 'ಕರೇನ್', 'ko': '카리 문자', 'ks': 'کاریَن', 'ks-Arab': 'کاریَن', 'lb': 'Karesch', 'lo': 'ຄາເຮຍ', 'lt': 'karių', 'mk': 'кариско', 'ml': 'ചരിയൻ', 'mn': 'Кари', 'mr': 'कॅरियन', 'ms': 'Carian', 'nb': 'karisk', 'ne': 'कारियन', 'nl': 'Carisch', 'nn': 'karisk', 'no': 'karisk', 'nqo': 'ߞߊ߯ߙߌ', 'or': 'କୈରନ୍', 'pl': 'karyjskie', 'pt': 'cariano', 'rm': 'carian', 'ro': 'cariană', 'ru': 'карийская', 'sc': 'carian', 'so': 'Luuqada kaariyaanka', 'sq': 'karianisht', 'sr': 'каријско писмо', 'sr-Cyrl': 'каријско писмо', 'sr-Latn': 'karijsko pismo', 'sv': 'kariska', 'ta': 'கரியன்', 'te': 'కారియన్', 'tg': 'Карианӣ', 'th': 'คาเรีย', 'to': 'tohinima fakakali', 'tr': 'Karya', 'ug': 'كارىيا', 'uk': 'каріанський', 'vec': 'cario', 'vi': 'Chữ Caria', 'xnr': 'करैन', 'yrl': 'karianu', 'yue': '卡里亞文', 'yue-Hans': '卡里亚文', 'yue-Hant': '卡里亞文', 'zh': '卡里亚文', 'zh-Hans': '卡里亚文', 'zh-Hant': '卡里亞文', 'zu': 'isi-Carian'},
    'Cham': {'ar': 'التشامية', 'az': 'çam', 'az-Latn': 'çam', 'bg': 'Хамитска', 'bn': 'চ্যাম', 'brx': 'कॅम', 'bs': 'čamsko pismo', 'bs-Cyrl': 'чамско писмо', 'bs-Latn': 'čamsko pismo', 'ccp': '𑄌𑄳𑄠𑄟𑄴', 'cs': 'čam', 'el': 'Τσαμ', 'el-polyton': 'Τσαμ', 'en-Dsrt': '𐐗𐐰𐑋', 'et': 'tšaami', 'eu': 'txamera', 'fa': 'چمی', 'ff-Adlm': '𞤕𞤢𞥄𞤥', 'fi': 'tšamilainen', 'he': 'צ׳אם', 'hi': 'चाम', 'hr': 'čamsko pismo', 'hu': 'Csám', 'ja': 'チャム文字', 'ka': 'ჩამი', 'kgp': 'sỹm', 'kk': 'чам жазуы', 'kk-Cyrl': 'чам жазуы', 'km': 'ចាម', 'kn': 'ಚಾಮ್', 'ko': '칸 고어', 'ks': 'چَم', 'ks-Arab': 'چَم', 'lo': 'ຈາມ', 'lt': 'čam', 'mk': 'чамско', 'ml': 'ഛം', 'mn': 'Чам', 'mr': 'चाम', 'ne': 'चाम', 'nqo': 'ߛߊ߲', 'or': 'ଛମ୍', 'pl': 'czamskie', 'ru': 'чамская', 'so': 'Jam', 'sq': 'çam', 'sr': 'чамско писмо', 'sr-Cyrl': 'чамско писмо', 'sr-Latn': 'čamsko pismo', 'ta': 'சாம்', 'te': 'చామ్', 'tg': 'Чамӣ', 'th': 'จาม', 'to': 'tohinima fakasami', 'ug': 'چام', 'uk': 'хамітський', 'vi': 'Chữ Chăm', 'xnr': 'चाम', 'yrl': 'xãmi', 'yue': '占文', 'yue-Hans': '占文', 'yue-Hant': '占文', 'zh': '占文', 'zh-Hans': '占文', 'zh-Hant': '占文', 'zu': 'isi-Cham', 'en': 'Cham'},
    'Cher': {'af': 'Cherokee', 'ak': 'Kɛroki', 'am': 'ቼሮኪ', 'ar': 'الشيروكي', 'as': 'চেৰ’কী', 'ast': 'cheroki', 'az': 'çiroki', 'az-Latn': 'çiroki', 'bal-Latn': 'Cheoki', 'be': 'чэрокі', 'be-tarask': 'чэрокі', 'bg': 'чероки', 'bn': 'চেরোকি', 'br': 'cherokee', 'brx': 'चेर’की', 'bs': 'čeroki pismo', 'bs-Cyrl': 'Чероки', 'bs-Latn': 'čeroki pismo', 'ca': 'cherokee', 'ca-ES-valencia': 'cherokee', 'ccp': '𑄌𑄬𑄇𑄮𑄇𑄨', 'chr': 'ᏣᎳᎩ', 'cs': 'čerokí', 'csw': 'ᒉᕑᐅᑫᕀ', 'cy': 'Cherokee', 'da': 'cherokee', 'de': 'Cherokee', 'dsb': 'cherokeeske pismo', 'el': 'Τσερόκι', 'el-polyton': 'Τσερόκι', 'en': 'Cherokee', 'en-Dsrt': '𐐕𐐯𐑉𐐬𐐿𐐨', 'es': 'cheroqui', 'et': 'tšerokii', 'eu': 'txerokiarra', 'fa': 'چروکی', 'ff-Adlm': '𞤕𞤫𞤪𞤮𞤳𞤭𞥅', 'fi': 'cherokeelainen', 'fil': 'Cherokee', 'fo': 'cherokee', 'fr': 'cherokee', 'fy': 'Cherokee', 'ga': 'Seiricíoch', 'gd': 'Cherokee', 'gl': 'cherokee', 'gsw': 'Cherokee', 'gu': 'ચેરોકી', 'ha': 'Cherokee', 'he': 'צ׳רוקי', 'hi': 'चेरोकी', 'hi-Latn': 'Cherokee', 'hr': 'čeroki pismo', 'hsb': 'cherokeeske pismo', 'ht': 'cherokee', 'hu': 'Cseroki', 'hy': 'չերոկի', 'id': 'Cherokee', 'ig': 'Cherokee', 'is': 'cherokí', 'it': 'cherokee', 'ja': 'チェロキー文字', 'jv': 'Sherokee', 'ka': 'ჩეროკი', 'kgp': 'seroki', 'kk': 'чероки жазуы', 'kk-Cyrl': 'чероки жазуы', 'km': 'ឆេរ៉ូគី', 'kn': 'ಚೆರೋಕೀ', 'ko': '체로키 문자', 'kok': 'चेरोकी', 'kok-Deva': 'चेरोकी', 'ks': 'چیروکی', 'ks-Arab': 'چیروکی', 'kxv': 'cerokī', 'kxv-Deva': 'चेरोकी', 'kxv-Latn': 'cerokī', 'kxv-Orya': 'ଚେରକୀ', 'kxv-Telu': 'చిరోకి', 'ky': 'чероки (жазуу)', 'lb': 'Cherokee', 'lo': 'ເຊໂຣກີ', 'lt': 'čerokių', 'lv': 'irokēzu', 'mai': 'चेरोकी', 'mi': 'Terokī', 'mk': 'черокиско', 'ml': 'ചെറോക്കി', 'mn': 'Чероки бичиг', 'mr': 'चेरोकी', 'ms': 'Cherokee', 'my': 'ချာရိုကီး', 'nb': 'cherokee', 'ne': 'चेरोकी', 'nl': 'Cherokee', 'nn': 'cherokee', 'no': 'cherokee', 'nqo': 'ߛߋߙߐߞ', 'or': 'ଚିରୁକୀ', 'pa': 'ਚੈਰੋਕੀ', 'pa-Guru': 'ਚੈਰੋਕੀ', 'pcm': 'Chíróki Lángwej', 'pl': 'czirokeskie', 'ps': 'چیروکي', 'pt': 'cherokee', 'qu': 'Cherokee Simi', 'rm': 'cherokee', 'ro': 'cherokee', 'ru': 'чероки', 'sc': 'cherokee', 'sd': 'چيروڪي', 'sd-Arab': 'چيروڪي', 'si': 'චෙරෝකී', 'sk': 'čerokézčina', 'sl': 'čerokeški', 'so': 'Jerokee', 'sq': 'çeroki', 'sr': 'чероки', 'sr-Cyrl': 'чероки', 'sr-Latn': 'čeroki', 'sv': 'cherokee', 'sw': 'Kicherokee', 'ta': 'செரோக்கி', 'te': 'చెరకీ', 'tg': 'Черокӣ', 'th': 'เชอโรกี', 'ti': 'ቼሪዮክ', 'tk': 'çeroki', 'to': 'tohinima fakaselokī', 'tr': 'Çeroki', 'ug': 'چېروكى', 'uk': 'черокі', 'ur': 'چیروکی', 'uz': 'cheroki', 'uz-Latn': 'cheroki', 'vec': 'cherokee', 'vi': 'Chữ Cherokee', 'xnr': 'चेरोकी', 'yo': 'Èdè Sheroki', 'yrl': 'xerokí', 'yue': '柴羅基文', 'yue-Hans': '柴罗基文', 'yue-Hant': '柴羅基文', 'zh': '切罗基文', 'zh-Hans': '切罗基文', 'zh-Hant': '柴羅基文', 'zu': 'isi-Cherokee'},
    'Chrs': {'bal-Latn': 'Chorasmi', 'bs': 'korasmijansko pismo', 'bs-Cyrl': 'чорамсианско писмо', 'bs-Latn': 'korasmijansko pismo', 'en': 'Chorasmian', 'et': 'horezmi', 'eu': 'korasmiera', 'ff-Adlm': '𞤑𞤮𞤪𞤢𞥄𞤧𞤥𞤭𞤴𞤢', 'fi': 'horemzi', 'fo': 'chorasmian', 'gd': 'Khwarazm', 'he': 'כורזמיאן', 'hi-Latn': 'Chorasmian', 'id': 'Chorasmian', 'ig': 'Chorasmian', 'kk': 'хорезм жазуы', 'kk-Cyrl': 'хорезм жазуы', 'km': 'ខូរ៉ាស្មី', 'mn': 'Корасми', 'nl': 'Chorasmisch', 'nqo': 'ߞߏߙߊߦߌߛߑߡߌ', 'ro': 'khorezmiană', 'sc': 'coràsmiu', 'so': 'Luuqada Korasmiyaanka', 'sq': 'korasmianisht', 'sv': 'khwarezmiska', 'tg': 'Хоразмиёнӣ', 'vec': 'corazmian', 'zh': '花拉子模文', 'zh-Hans': '花拉子模文', 'zu': 'isi-Chorasmian'},
    'Cirt': {'ar': 'السيرث', 'ast': 'cirth', 'az': 'sirt', 'az-Latn': 'sirt', 'bg': 'Кирт', 'bn': 'কির্ট', 'brx': 'सिर्थ', 'bs': 'cirt pismo', 'bs-Cyrl': 'цирт писмо', 'bs-Latn': 'cirt pismo', 'ca': 'cirth', 'ca-ES-valencia': 'cirth', 'ccp': '𑄇𑄨𑄢𑄴𑄑𑄧', 'cs': 'kirt', 'de': 'Cirth', 'el': 'Σερθ', 'el-polyton': 'Σερθ', 'en': 'Cirth', 'en-Dsrt': '𐐗𐐲𐑉𐑃', 'es': 'cirth', 'et': 'Cirthi', 'fa': 'کرت', 'fi': 'cirth', 'fr': 'cirth', 'fy': 'Cirth', 'gd': 'Cirth', 'gsw': 'Cirth', 'gu': 'સિર્થ', 'hi': 'किर्थ', 'hi-Latn': 'Cirth', 'hr': 'cirth pismo', 'ht': 'cirth', 'id': 'Cirth', 'it': 'cirth', 'ja': 'キアス文字', 'ka': 'კირთი', 'kgp': 'sirti', 'kn': 'ಸಿರ್ಥ್', 'ko': '키르쓰', 'ks': 'کِرتھ', 'ks-Arab': 'کِرتھ', 'lb': 'Cirth', 'lo': 'ເຊີຮ', 'lt': 'kirt', 'mk': 'кирт', 'ml': 'ചിർത്ത്', 'mr': 'किर्थ', 'nb': 'cirth', 'ne': 'किर्थ', 'nl': 'Cirth', 'nn': 'cirth', 'no': 'cirth', 'or': 'ସିର୍ଥ', 'pl': 'cirth', 'pt': 'cirth', 'rm': 'cirth', 'ru': 'кирт', 'sl': 'kirt', 'sr': 'цирт писмо', 'sr-Cyrl': 'цирт писмо', 'sr-Latn': 'cirt pismo', 'ta': 'கிர்த்', 'te': 'సిర్థ్', 'th': 'เซิร์ท', 'to': 'tohinima fakakīliti', 'tr': 'Cirth', 'ug': 'كىرسچە', 'uk': 'кирт', 'vi': 'Chữ Cirth', 'yrl': 'runikarana', 'yue': '色斯文', 'yue-Hans': '色斯文', 'yue-Hant': '色斯文', 'zh': '色斯文', 'zh-Hans': '色斯文', 'zh-Hant': '色斯文'},
    'Copt': {'am': 'ኮፕቲክ', 'ar': 'القبطية', 'ast': 'coptu', 'az': 'koptik', 'az-Latn': 'koptik', 'bal-Latn': 'koptek', 'bg': 'Коптска', 'bn': 'কোপ্টিক', 'br': 'koptek', 'brx': 'कॉप्ट', 'bs': 'koptičko pismo', 'bs-Cyrl': 'коптичко писмо', 'bs-Latn': 'koptičko pismo', 'ca': 'copte', 'ca-ES-valencia': 'copte', 'ccp': '𑄇𑄮𑄛𑄴𑄑𑄨𑄇𑄴', 'cs': 'koptské', 'da': 'koptisk', 'de': 'Koptisch', 'el': 'Κοπτικό', 'el-polyton': 'Κοπτικό', 'en': 'Coptic', 'en-Dsrt': '𐐗𐐱𐐹𐐻𐐮𐐿', 'es': 'copto', 'et': 'kopti', 'eu': 'koptikera', 'fa': 'قبطی', 'ff-Adlm': '𞤑𞤭𞤦𞤯𞤭𞤲𞤳𞤮', 'fi': 'koptilainen', 'fo': 'koptiskt', 'fr': 'copte', 'fy': 'Koptysk', 'ga': 'Coptach', 'gd': 'Coptais', 'gsw': 'Koptisch', 'gu': 'કોપ્ટિક', 'he': 'קופטי', 'hi': 'कॉप्टिक', 'hi-Latn': 'Coptic', 'hr': 'koptsko pismo', 'ht': 'copte', 'hu': 'Kopt', 'id': 'Koptik', 'ig': 'Coptic', 'is': 'koptíska', 'it': 'copto', 'ja': 'コプト文字', 'ka': 'კოპტური', 'kgp': 'kómtiko', 'kk': 'копт жазуы', 'kk-Cyrl': 'копт жазуы', 'km': 'ខូប្ទ', 'kn': 'ಕಾಪ್ಟಿಕ್', 'ko': '콥트 문자', 'ks': 'کاپٹِک', 'ks-Arab': 'کاپٹِک', 'ku': 'qiptî', 'lb': 'Koptesch', 'lo': 'ຄອບຕິກ', 'lt': 'koptų', 'lv': 'koptu', 'mk': 'коптско', 'ml': 'കോപ്റ്റിക്', 'mn': 'Коптик', 'mr': 'कॉप्टिक', 'ms': 'Coptic', 'nb': 'koptisk', 'ne': 'कप्टिक', 'nl': 'Koptisch', 'nn': 'koptisk', 'no': 'koptisk', 'nqo': 'ߞߐߔߑߕ', 'or': 'କପଟିକ୍', 'pl': 'koptyjskie', 'pt': 'cóptico', 'rm': 'coptic', 'ro': 'coptă', 'ru': 'коптская', 'sc': 'coptu', 'sl': 'koptski', 'so': 'Dadka Kotiga', 'sq': 'koptisht', 'sr': 'коптичко писмо', 'sr-Cyrl': 'коптичко писмо', 'sr-Latn': 'koptičko pismo', 'sv': 'koptiska', 'ta': 'காப்டிக்', 'te': 'కోప్టిక్', 'tg': 'Коптӣ', 'th': 'คอปติก', 'to': 'tohinima fakakopitika', 'tr': 'Kıpti', 'ug': 'كوپتىك', 'uk': 'коптський', 'vec': 'copto', 'vi': 'Chữ Coptic', 'xnr': 'कॉप्टिक', 'yrl': 'kupitiku', 'yue': '科普特文', 'yue-Hans': '科普特文', 'yue-Hant': '科普特文', 'zh': '克普特文', 'zh-Hans': '克普特文', 'zh-Hant': '科普特文', 'zu': 'isi-Coptic'},
    'Cpmn': {'bal-Latn': 'Sáepró-Minói', 'bs': 'ciprominojsko pismo', 'bs-Cyrl': 'кипро-миноанско писмо', 'bs-Latn': 'ciprominojsko pismo', 'en': 'Cypro-Minoan', 'et': 'Küprose minose', 'eu': 'zipro-minoera', 'ff-Adlm': '𞤅𞤭𞥅𞤨𞤪𞤮 𞤃𞤭𞤲𞤮𞤴𞤢', 'fi': 'kypro-minolainen', 'fo': 'cypro-minoan', 'gd': 'Mìneothais Chìopras', 'he': 'ציפרו-מינואן', 'hi-Latn': 'Cypro-Minoan', 'hu': 'Ciprusi-minószi', 'id': 'Cypro-Minoan', 'ig': 'Cypro-Minoan', 'kk': 'кипр-миной жазуы', 'kk-Cyrl': 'кипр-миной жазуы', 'km': 'ស៊ីប្រូមីណូ', 'mn': 'Кипр-Миноан', 'nl': 'Cypro-Minoïsch', 'nqo': 'ߛߌߔߙߏ-ߡߌߣߏߞߊ', 'ro': 'cipro-minoană', 'sc': 'tzipro-minòicu', 'sq': 'minoishte e Qipros', 'sv': 'cypro-minoisk skrift', 'tg': 'Кипро-Миноанӣ', 'vec': 'sipro minòego', 'zh': '塞浦路斯米诺斯文', 'zh-Hans': '塞浦路斯米诺斯文'},
    'Cprt': {'am': 'ሲፕሪኦት', 'ar': 'القبرصية', 'ast': 'xipriota', 'az': 'kipr', 'az-Latn': 'kipr', 'bal-Latn': 'Sáepriót', 'bg': 'Кипърска', 'bn': 'সাইপ্রোয়েট', 'br': 'silabennaoueg kipriek', 'brx': 'सीप्रीओट्', 'bs': 'kiparsko pismo', 'bs-Cyrl': 'кипарско писмо', 'bs-Latn': 'kiparsko pismo', 'ca': 'xipriota', 'ca-ES-valencia': 'xipriota', 'ccp': '𑄥𑄭𑄛𑄳𑄢𑄮𑄠𑄬𑄖𑄴', 'cs': 'kyperské', 'da': 'cypriotisk', 'de': 'Zypriotisch', 'el': 'Κυπριακό', 'el-polyton': 'Κυπριακό', 'en': 'Cypriot', 'en-Dsrt': '𐐝𐐮𐐹𐑉𐐨𐐲𐐻', 'es': 'chipriota', 'et': 'Küprose silpkiri', 'eu': 'ziprera', 'fa': 'قبرسی', 'ff-Adlm': '𞤑𞤵𞤦𞤭𞤪𞤧𞤵', 'fi': 'muinaiskyproslainen', 'fo': 'kýpriskt', 'fr': 'syllabaire chypriote', 'fur': 'cipriot', 'fy': 'Syprysk', 'ga': 'Cipireach', 'gd': 'Cìoprasais', 'gsw': 'Zypriotisch', 'gu': 'સિપ્રાયટ', 'he': 'קפריסאי', 'hi': 'काइप्रायट', 'hi-Latn': 'Cypriot', 'hr': 'cypriot pismo', 'ht': 'syllabaire chypriote', 'hu': 'Ciprusi', 'id': 'Siprus', 'ig': 'Cypriot', 'it': 'cipriota', 'ja': 'キプロス文字', 'ka': 'კვიპროსული', 'kgp': 'siprijota', 'kk': 'кипр жазуы', 'kk-Cyrl': 'кипр жазуы', 'km': 'ស៊ីប', 'kn': 'ಸಿಪ್ರಿಯಾಟ್', 'ko': '키프로스 문자', 'ks': 'کِپرایِٹ', 'ks-Arab': 'کِپرایِٹ', 'ku': 'qibrisî', 'lb': 'Zypriotesch', 'lo': 'ໄຊເປຍ', 'lt': 'kipro', 'mk': 'кипарско', 'ml': 'സൈപ്രിയോട്ട്', 'mn': 'Киприот', 'mr': 'सायप्रिऑट', 'ms': 'Cypriot', 'nb': 'kypriotisk', 'ne': 'कप्रियट', 'nl': 'Cyprisch', 'nn': 'kypriotisk', 'no': 'kypriotisk', 'nqo': 'ߛߔߙߌߐߕ ߜߋ߲߬ߞߝߊ', 'or': 'ସିପ୍ରଅଟ୍', 'pl': 'cypryjskie', 'pt': 'cipriota', 'rm': 'cipriot', 'ro': 'cipriotă', 'ru': 'кипрская', 'sc': 'tzipriotu', 'sl': 'ciprski', 'so': 'sibraas dhalad ah', 'sq': 'qipriotisht', 'sr': 'кипарско писмо', 'sr-Cyrl': 'кипарско писмо', 'sr-Latn': 'kiparsko pismo', 'sv': 'cypriotiska', 'ta': 'சைப்ரியாட்', 'te': 'సైప్రోట్', 'tg': 'Кипрӣ', 'th': 'ไซเปรียท', 'to': 'tohinima fakasaipalesi', 'tr': 'Kıbrıs', 'ug': 'سىپرۇس', 'uk': 'кіпрський', 'vec': 'siprioto', 'vi': 'Chứ Síp', 'xnr': 'काइप्रायट', 'yrl': 'sipiriuta', 'yue': '塞浦路斯文', 'yue-Hans': '塞浦路斯文', 'yue-Hant': '塞浦路斯文', 'zh': '塞浦路斯文', 'zh-Hans': '塞浦路斯文', 'zh-Hant': '塞浦路斯文', 'zu': 'isi-Cypriot'},
    'Cyrl': {'af': 'Cyrillies', 'ak': 'Kreleke', 'am': 'ሲይሪልክ', 'ar': 'السيريلية', 'as': 'চিৰিলিক', 'ast': 'cirílicu', 'az': 'kiril', 'az-Arab': 'سریلیکجه', 'az-Cyrl': 'Кирил', 'az-Latn': 'kiril', 'bal-Latn': 'Rusi', 'be': 'кірыліца', 'be-tarask': 'кірыліца', 'bg': 'кирилица', 'bgn': 'سیریلیک', 'blo': 'Siril kʊja', 'bn': 'সিরিলিক', 'br': 'kirillek', 'brx': 'सिरिल्लिक', 'bs': 'ćirilica', 'bs-Cyrl': 'ћирилица', 'bs-Latn': 'ćirilica', 'ca': 'ciríl·lic', 'ca-ES-valencia': 'ciríl·lic', 'ccp': '𑄥𑄨𑄢𑄨𑄣𑄨𑄇𑄴', 'ce': 'кириллица', 'ceb': 'Cyrillic', 'chr': 'ᏲᏂᎢ ᏗᎪᏪᎵ', 'ckb': 'سریلیک', 'cs': 'cyrilice', 'cv': 'кириллица', 'cy': 'Cyrilig', 'da': 'kyrillisk', 'de': 'Kyrillisch', 'doi': 'सिरिलिक', 'dsb': 'kyriliski', 'dz': 'སིརིལ་ལིཀ་ཡིག་གུ', 'ee': 'Cyrillicgbeŋɔŋlɔ', 'el': 'Κυριλλικό', 'el-polyton': 'Κυριλλικό', 'en': 'Cyrillic', 'en-Dsrt': '𐐝𐐲𐑉𐐮𐑊𐐮𐐿', 'es': 'cirílico', 'et': 'kirillitsa', 'eu': 'zirilikoa', 'fa': 'سیریلی', 'ff-Adlm': '𞤅𞤭𞤪𞤤𞤭𞤳', 'fi': 'kyrillinen', 'fil': 'Cyrillic', 'fo': 'kyrilliskt', 'fr': 'cyrillique', 'fur': 'cirilic', 'fy': 'Syrillysk', 'ga': 'Coireallach', 'gd': 'Cirilis', 'gl': 'cirílico', 'gsw': 'Kyrillisch', 'gu': 'સિરિલિક', 'ha': 'Cyrillic', 'he': 'קירילי', 'hi': 'सिरिलिक', 'hi-Latn': 'Cyrillic', 'hr': 'ćirilica', 'hsb': 'kyrilisce', 'ht': 'cyrillique', 'hu': 'Cirill', 'hy': 'կյուրեղագիր', 'ia': 'cyrillic', 'id': 'Sirilik', 'ie': 'cirillic', 'ig': 'Cyrillic', 'ii': 'ꀊꆨꌦꇁꃚꁱꂷ', 'is': 'kyrillískt', 'it': 'cirillico', 'ja': 'キリル文字', 'jv': 'Sirilik', 'ka': 'კირილიცა', 'kaa': 'кирилл жазыўы', 'kaa-Cyrl': 'кирилл жазыўы', 'kea': 'siríliku', 'kgp': 'siririko', 'kk': 'кирилл жазуы', 'kk-Cyrl': 'кирилл жазуы', 'km': 'ស៊ីរីលីក', 'kn': 'ಸಿರಿಲಿಕ್', 'ko': '키릴 문자', 'kok': 'सिरिलिक', 'kok-Deva': 'सिरिलिक', 'ks': 'سَیرِلِک', 'ks-Arab': 'سَیرِلِک', 'ks-Deva': 'सिरिलिक', 'ksh': 'kürrellesche Schreff', 'ku': 'kirîlî', 'kxv': 'sirilik', 'kxv-Deva': 'सिरिलिक', 'kxv-Latn': 'sirilik', 'kxv-Orya': 'ସିରିଲିକ୍', 'kxv-Telu': 'సిరిలిక్', 'ky': 'Кирилл', 'lb': 'Kyrillesch', 'lld': 'zirilich', 'lo': 'ຊີຣິວລິກ', 'lrc': 'سیریلیک', 'lt': 'kirilica', 'lv': 'kirilica', 'mai': 'सिरिलिक', 'mi': 'Hīririki', 'mk': 'кирилско писмо', 'ml': 'സിറിലിക്', 'mn': 'кирилл', 'mni': 'সিরিলিক', 'mni-Beng': 'সিরিলিক', 'mr': 'सीरिलिक', 'ms': 'Cyril', 'mt': 'Ċirilliku', 'my': 'စစ်ရိလစ်', 'mzn': 'سیریلیک', 'nb': 'kyrillisk', 'ne': 'सिरिलिक', 'nl': 'Cyrillisch', 'nn': 'kyrillisk', 'no': 'kyrillisk', 'nqo': 'ߛߙߌߟߌߞ', 'om': 'Saayiriilik', 'or': 'ସିରିଲିକ୍', 'os': 'Киррилицӕ', 'pa': 'ਸਿਰਿਲਿਕ', 'pa-Guru': 'ਸਿਰਿਲਿਕ', 'pcm': 'Sírílik', 'pl': 'cyrylica', 'ps': 'سیریلیک', 'pt': 'cirílico', 'qu': 'Cirilico', 'rm': 'cirillic', 'ro': 'chirilică', 'ru': 'кириллица', 'sa': 'सिरिलिक:', 'sah': 'Нууччалыы', 'sat': 'ᱥᱤᱨᱤᱞᱤᱠ', 'sat-Olck': 'ᱥᱤᱨᱤᱞᱤᱠ', 'sc': 'tzirìllicu', 'sd': 'سيريلي', 'sd-Arab': 'سيريلي', 'sd-Deva': 'सिरिलिक', 'se': 'kyrillalaš', 'si': 'සිරිලික්', 'sk': 'cyrilika', 'sl': 'cirilica', 'so': 'Siriylik', 'sq': 'cirilik', 'sr': 'ћирилица', 'sr-Cyrl': 'ћирилица', 'sr-Latn': 'ćirilica', 'su': 'Sirilik', 'su-Latn': 'Sirilik', 'sv': 'kyrilliska', 'sw': 'Kisiriliki', 'ta': 'சிரிலிக்', 'te': 'సిరిలిక్', 'tg': 'Кириллӣ', 'th': 'ซีริลลิก', 'ti': 'ቋንቋ ሲሪል', 'tk': 'Kiril elipbiýi', 'to': 'tohinima fakalūsia', 'tpi': 'Syrilik', 'tr': 'Kiril', 'tt': 'кирилл', 'ug': 'كىرىل', 'uk': 'кирилиця', 'ur': 'سیریلک', 'uz': 'kirill', 'uz-Cyrl': 'Кирил', 'uz-Latn': 'kirill', 'vec': 'sirìlego', 'vi': 'Chữ Kirin', 'wae': 'Kirilliš', 'wo': 'Sirilik', 'xh': 'IsiCyrillic', 'xnr': 'सिरिलिक', 'yi': 'ציריליש', 'yo': 'èdè ilẹ̀ Rọ́ṣíà', 'yrl': 'siríriku', 'yue': '斯拉夫文', 'yue-Hans': '斯拉夫文', 'yue-Hant': '斯拉夫文', 'zh': '西里尔文', 'zh-Hans': '西里尔文', 'zh-Hant': '西里爾文字', 'zu': 'isi-Cyrillic'},
    'Cyrs': {'ar': 'السيريلية السلافية الكنسية القديمة', 'ast': 'eslavónicu cirílicu eclesiásticu antiguu', 'az': 'qədimi kilsa kirili', 'az-Latn': 'qədimi kilsa kirili', 'bal-Latn': 'Ahdi Serelek', 'bn': 'প্রাচীন চার্চ স্লাভোনিক সিরিলিক', 'br': 'kirillek henslavonek', 'brx': 'पुरानी चर्च सिरिलिक्', 'bs': 'staroslovenska crkvena ćirilica', 'bs-Cyrl': 'старословенска црквена ћирилица', 'bs-Latn': 'staroslovenska crkvena ćirilica', 'ca': 'ciríl·lic de l’antic eslau eclesiàstic', 'ca-ES-valencia': 'ciríl·lic de l’antic eslau eclesiàstic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄨 𑄌𑄢𑄴𑄌𑄧 𑄥𑄳𑄣𑄞𑄮𑄚𑄨𑄇𑄴 𑄥𑄨𑄢𑄨𑄣𑄨𑄇𑄴', 'cs': 'cyrilce - staroslověnská', 'da': 'kyrillisk - oldkirkeslavisk variant', 'de': 'Altkirchenslawisch', 'el': 'Παλαιό Εκκλησιαστικό Σλαβικό Κυριλλικό', 'el-polyton': 'Παλαιό Εκκλησιαστικό Σλαβικό Κυριλλικό', 'en': 'Old Church Slavonic Cyrillic', 'en-Dsrt': '𐐄𐑊𐐼 𐐕𐐲𐑉𐐽 𐐝𐑊𐐲𐑂𐐱𐑌𐐮𐐿 𐐗𐐲𐑉𐐮𐑊𐐮𐐿', 'es': 'cirílico del antiguo eslavo eclesiástico', 'et': 'kürilliline kirikuslaavi', 'fi': 'kyrillinen muinaiskirkkoslaavimuunnelma', 'fr': 'cyrillique (variante slavonne)', 'fur': 'cirilic dal vieri slavonic de glesie', 'fy': 'Aldkerkslavysk Syrillysk', 'gd': 'Cirilis Seann-Slàbhais na h-Eaglaise', 'gsw': 'Altchileslawisch', 'gu': 'ઓલ્ડ ચર્ચ સ્લાવોનિક સિરિલિક', 'he': 'קירילי סלאבוני כנסייתי עתיק', 'hi': 'ओल्ड चर्च स्लावोनिक सिरिलिक', 'hi-Latn': 'Old Church Slavonic Cyrillic', 'hr': 'staroslavenska crkvena čirilica', 'ht': 'cyrillique (variante slavonne)', 'hu': 'Óegyházi szláv cirill', 'id': 'Gereja Slavonia Sirilik Lama', 'ig': 'Old Church Slavonic Cyrillic', 'it': 'cirillico antica chiesa slavonica', 'ja': '古代教会スラブ語キリル文字', 'ka': 'ძველი სლავური კირილიცა', 'kgp': 'siririko esiravo ekresijatiko', 'kn': 'ಪ್ರಾಚೀನ ಚರ್ಚ್ ಸ್ಲೋವಾನಿಕ್ ಸಿರಿಲಿಕ್', 'ko': '고대교회슬라브어 키릴문자', 'ks': 'پرون چٔرچسلیوونِک سَیرِلِک', 'ks-Arab': 'پرون چٔرچسلیوونِک سَیرِلِک', 'lb': 'Alkiercheslawesch', 'lo': 'ເຊຮັດສລາ ໂວນິກຊີຮິກລິກໂບຮານ', 'lt': 'senoji bažnytinė slavų kirilica', 'lv': 'senslāvu', 'mk': 'старословенска кирилица', 'ml': 'പുരാതന ചർച്ച് സ്ലവോണിക് സിറിലിക്', 'mr': 'पुरातन चर्च स्लाव्होनिक सिरिलिक', 'nb': 'kirkeslavisk kyrillisk', 'nl': 'Oudkerkslavisch Cyrillisch', 'nn': 'kyrillisk (kyrkjeslavisk variant)', 'no': 'kirkeslavisk kyrillisk', 'or': 'ଓଲ୍ଡ ଚର୍ଚ୍ଚ ସାଲଭୋନିକ୍ ସିରିଲିକ୍', 'pl': 'cyrylica staro-cerkiewno-słowiańska', 'pt': 'cirílico eslavo eclesiástico', 'rm': 'slav da baselgia vegl', 'ro': 'chirilică slavonă bisericească veche', 'ru': 'старославянская', 'sl': 'starocerkvenoslovanska cirilica', 'sr': 'Старословенска црквена ћирилица', 'sr-Cyrl': 'Старословенска црквена ћирилица', 'sr-Latn': 'Staroslovenska crkvena ćirilica', 'sv': 'fornkyrkoslavisk kyrilliska', 'ta': 'பழைய சர்ச் ஸ்லவோனிக் சிரிலிக்', 'te': 'ప్రాచీన చర్చ స్లావోనిక్ సిరిలిక్', 'th': 'เชอร์ชสลาโวนิกซีริลลิกโบราณ', 'to': 'tohinima fakalūsia-lotu-motuʻa', 'tr': 'Eski Kilise Slavcası Kiril', 'ug': 'قەدىمكى چىركاۋ سىلاۋيانچە كىرىل', 'uk': 'давньоцерковнословʼянський', 'vec': 'zlavo de cieza antigo', 'vi': 'Chữ Kirin Slavơ Nhà thờ cổ', 'xnr': 'ओल्ड चर्च स्लावोनिक सिरिलिक', 'yrl': 'siríriku isirawu ekeresiatiku', 'yue': '西里爾文（古教會斯拉夫文變體）', 'yue-Hans': '西里尔文（古教会斯拉夫文变体）', 'yue-Hant': '西里爾文（古教會斯拉夫文變體）', 'zh': '西里尔文字（古教会斯拉夫文的变体）', 'zh-Hans': '西里尔文字（古教会斯拉夫文的变体）', 'zh-Hant': '西里爾文（古教會斯拉夫文變體）'},
    'DD': {'en': 'German Democratic Republic'},
    'DE': {'af': 'Duitsland', 'agq': 'Dzamanè', 'ak': 'Gyaaman', 'am': 'ጀርመን', 'ar': 'ألمانيا', 'as': 'জাৰ্মানী', 'asa': 'Ujerumani', 'ast': 'Alemaña', 'az': 'Almaniya', 'az-Cyrl': 'Алманија', 'az-Latn': 'Almaniya', 'bal-Latn': 'Jarmani', 'bas': 'Jamân', 'be': 'Германія', 'be-tarask': 'Германія', 'bez': 'Huujerumani', 'bg': 'Германия', 'bgn': 'جرمنی', 'blo': 'Gajaamatǝna', 'bm': 'Alimaɲi', 'bn': 'জার্মানি', 'bo': 'འཇར་མན་', 'br': 'Alamagn', 'brx': 'जार्मानी', 'bs': 'Njemačka', 'bs-Cyrl': 'Њемачка', 'bs-Latn': 'Njemačka', 'ca': 'Alemanya', 'ca-ES-valencia': 'Alemanya', 'ccp': '𑄎𑄢𑄴𑄟𑄚𑄨', 'ce': 'Германи', 'ceb': 'Germany', 'cgg': 'Bugirimaani', 'chr': 'ᎠᏂᏛᏥ', 'ckb': 'ئەڵمانیا', 'cs': 'Německo', 'cv': 'Германи', 'cy': 'Yr Almaen', 'da': 'Tyskland', 'dav': 'Ujerumani', 'de': 'Deutschland', 'dje': 'Almaaɲe', 'doi': 'जर्मनी', 'dsb': 'Nimska', 'dyo': 'Almaañ', 'dz': 'ཇཱར་མ་ནི', 'ebu': 'Ujerumani', 'ee': 'Germania nutome', 'el': 'Γερμανία', 'el-polyton': 'Γερμανία', 'en': 'Germany', 'en-Dsrt': '𐐖𐐲𐑉𐑋𐐲𐑌𐐨', 'eo': 'Germanujo', 'es': 'Alemania', 'et': 'Saksamaa', 'eu': 'Alemania', 'ewo': 'Ndzáman', 'fa': 'آلمان', 'ff': 'Almaañ', 'ff-Adlm': '𞤔𞤫𞤪𞤥𞤢𞤲𞤭𞥅', 'ff-Latn': 'Almaañ', 'fi': 'Saksa', 'fil': 'Germany', 'fo': 'Týskland', 'fr': 'Allemagne', 'fur': 'Gjermanie', 'fy': 'Dútslân', 'ga': 'an Ghearmáin', 'gd': 'A’ Ghearmailt', 'gl': 'Alemaña', 'gsw': 'Tüütschland', 'gu': 'જર્મની', 'guz': 'Ujerumani', 'ha': 'Jamus', 'haw': 'Kelemānia', 'he': 'גרמניה', 'hi': 'जर्मनी', 'hi-Latn': 'Germany', 'hr': 'Njemačka', 'hsb': 'Němska', 'ht': 'Allemagne', 'hu': 'Németország', 'hy': 'Գերմանիա', 'ia': 'Germania', 'id': 'Jerman', 'ie': 'Germania', 'ig': 'Germany', 'ii': 'ꄓꇩ', 'is': 'Þýskaland', 'it': 'Germania', 'ja': 'ドイツ', 'jgo': 'Njáman', 'jmc': 'Ujerumani', 'jv': 'Jérman', 'ka': 'გერმანია', 'kaa': 'Германия', 'kaa-Cyrl': 'Германия', 'kab': 'Lalman', 'kam': 'Ujerumani', 'kde': 'Udyerumani', 'kea': 'Alimanha', 'kgp': 'Aremỹija', 'khq': 'Almaaɲe', 'ki': 'Njeremani', 'kk': 'Германия', 'kk-Cyrl': 'Германия', 'kln': 'Emetab Geruman', 'km': 'អាល្លឺម៉ង់', 'kn': 'ಜರ್ಮನಿ', 'ko': '독일', 'kok': 'जर्मनी', 'kok-Deva': 'जर्मनी', 'kok-Latn': 'Jermon', 'ks': 'جرمٔنی', 'ks-Arab': 'جرمٔنی', 'ks-Deva': 'जर्मन', 'ksb': 'Ujeumani', 'ksf': 'djɛrman', 'ksh': 'Doütschland', 'ku': 'Almanya', 'kxv': 'jermanī', 'kxv-Deva': 'जर्मनी', 'kxv-Latn': 'jermanī', 'kxv-Orya': 'ଜର୍ମାନି', 'kxv-Telu': 'జర్మనీ', 'ky': 'Германия', 'lag': 'Ʉjerumáani', 'lb': 'Däitschland', 'lg': 'Budaaki', 'lkt': 'Iyášiča Makȟóčhe', 'lld': 'Paisc Todësc', 'ln': 'Alemani', 'lo': 'ເຢຍລະມັນ', 'lrc': 'آلمان', 'lt': 'Vokietija', 'lu': 'Alemanu', 'luo': 'Germany', 'luy': 'Ujerumani', 'lv': 'Vācija', 'mai': 'जर्मनी', 'mas': 'Ujerumani', 'mer': 'Njamanĩ', 'mfe': 'Almagn', 'mg': 'Alemaina', 'mi': 'Tiamana', 'mk': 'Германија', 'ml': 'ജർമ്മനി', 'mn': 'Герман', 'mni': 'জর্মনি', 'mni-Beng': 'জর্মনি', 'mr': 'जर्मनी', 'ms': 'Jerman', 'mt': 'il-Ġermanja', 'mua': 'Germaniya', 'my': 'ဂျာမနီ', 'mzn': 'آلمان', 'naq': 'Duitslandi', 'nb': 'Tyskland', 'nd': 'Germany', 'nds': 'Düütschland', 'ne': 'जर्मनी', 'nl': 'Duitsland', 'nmg': 'Jaman', 'nn': 'Tyskland', 'no': 'Tyskland', 'nqo': 'ߊߟߑߡߊ߲ߘߎ߯', 'nyn': 'Bugirimaani', 'om': 'Jarmanii', 'or': 'ଜର୍ମାନୀ', 'os': 'Герман', 'pa': 'ਜਰਮਨੀ', 'pa-Guru': 'ਜਰਮਨੀ', 'pcm': 'Jámáni', 'pl': 'Niemcy', 'ps': 'المان', 'pt': 'Alemanha', 'qu': 'Alemania', 'rm': 'Germania', 'rn': 'Ubudage', 'ro': 'Germania', 'rof': 'Ujerumani', 'ru': 'Германия', 'rwk': 'Ujerumani', 'sa': 'जर्मनीदेश:', 'saq': 'Ujerumani', 'sat': 'ᱡᱟᱨᱢᱟᱱᱤ', 'sat-Olck': 'ᱡᱟᱨᱢᱟᱱᱤ', 'sbp': 'Wujelumani', 'sc': 'Germània', 'sd': 'جرمني', 'sd-Arab': 'جرمني', 'sd-Deva': 'जर्मनी', 'se': 'Duiska', 'seh': 'Alemanha', 'ses': 'Almaaɲe', 'sg': 'Zâmani', 'shi': 'ⴰⵍⵎⴰⵏⵢⴰ', 'shi-Latn': 'almanya', 'shi-Tfng': 'ⴰⵍⵎⴰⵏⵢⴰ', 'si': 'ජර්මනිය', 'sk': 'Nemecko', 'sl': 'Nemčija', 'smn': 'Saksa', 'sn': 'Germany', 'so': 'Jarmal', 'sq': 'Gjermani', 'sr': 'Немачка', 'sr-Cyrl': 'Немачка', 'sr-Latn': 'Nemačka', 'su': 'Jérman', 'su-Latn': 'Jérman', 'sv': 'Tyskland', 'sw': 'Ujerumani', 'syr': 'ܐܠܡܢܝܐ', 'ta': 'ஜெர்மனி', 'te': 'జర్మనీ', 'teo': 'Ujerumani', 'tg': 'Германия', 'th': 'เยอรมนี', 'ti': 'ጀርመን', 'tk': 'Germaniýa', 'to': 'Siamane', 'tpi': 'Jemani', 'tr': 'Almanya', 'tt': 'Германия', 'twq': 'Almaaɲe', 'tzm': 'Almanya', 'ug': 'گېرمانىيە', 'uk': 'Німеччина', 'ur': 'جرمنی', 'uz': 'Germaniya', 'uz-Cyrl': 'Германия', 'uz-Latn': 'Germaniya', 'vai': 'ꕧꕮꔧ', 'vai-Latn': 'Jamáĩ', 'vai-Vaii': 'ꕧꕮꔧ', 'vec': 'Jermania', 'vi': 'Đức', 'vun': 'Ujerumani', 'wae': 'Titšland', 'wo': 'Almaañ', 'xh': 'EJamani', 'xnr': 'जर्मनी', 'xog': 'Budaaki', 'yav': 'nsáman', 'yi': 'דייטשלאַנד', 'yo': 'Jámánì', 'yrl': 'Aremãya', 'yue': '德國', 'yue-Hans': '德国', 'yue-Hant': '德國', 'zgh': 'ⴰⵍⵎⴰⵏⵢⴰ', 'zh': '德国', 'zh-Hans': '德国', 'zh-Hant': '德國', 'zu': 'i-Germany'},
    'DG': {'af': 'Diego Garcia', 'ak': 'Diɛgo Gaasia', 'am': 'ዲዬጎ ጋርሲያ', 'ar': 'دييغو غارسيا', 'as': 'ডিয়েগো গাৰ্চিয়া', 'ast': 'Diego Garcia', 'az': 'Dieqo Qarsiya', 'az-Cyrl': 'Диего Гарсија', 'az-Latn': 'Dieqo Qarsiya', 'bal-Latn': 'Diégó Gárshiá', 'be': 'Востраў Дыега-Гарсія', 'be-tarask': 'Востраў Дыега-Гарсія', 'bg': 'Диего Гарсия', 'blo': 'Ɖiyego Garsiya', 'bn': 'দিয়েগো গার্সিয়া', 'br': 'Diego Garcia', 'brx': 'दियेग’ गार्सिया', 'bs': 'Dijego Garsija', 'bs-Cyrl': 'Дијего Гарсија', 'bs-Latn': 'Dijego Garsija', 'ca': 'Diego Garcia', 'ca-ES-valencia': 'Diego Garcia', 'ccp': '𑄘𑄨𑄠𑄬𑄉𑄮 𑄉𑄢𑄴𑄥𑄨𑄠', 'ce': 'Диего-Гарси', 'ceb': 'Diego Garcia', 'chr': 'ᏗᏰᎪ ᎦᏏᏯ', 'ckb': 'دیەگۆ گارسیا', 'cs': 'Diego García', 'cv': 'Диего-Гарсия', 'cy': 'Diego Garcia', 'da': 'Diego Garcia', 'de': 'Diego Garcia', 'dsb': 'Diego Garcia', 'dz': 'ཌི་ཡེ་གོ་གར་སིའོ', 'ee': 'Diego Garsia nutome', 'el': 'Ντιέγκο Γκαρσία', 'el-polyton': 'Ντιέγκο Γκαρσία', 'en': 'Diego Garcia', 'es': 'Diego García', 'et': 'Diego Garcia', 'eu': 'Diego García', 'fa': 'دیه\u200cگو گارسیا', 'ff-Adlm': '𞤔𞤮𞤺𞤮 𞤘𞤢𞥄𞤪𞤧𞤭𞤴𞤢', 'fi': 'Diego Garcia', 'fil': 'Diego Garcia', 'fo': 'Diego Garcia', 'fr': 'Diego Garcia', 'fur': 'Diego Garcia', 'fy': 'Diego Garcia', 'ga': 'Diego Garcia', 'gd': 'Diego Garcia', 'gl': 'Diego García', 'gu': 'ડિએગો ગારસિઆ', 'ha': 'Tsibirn Diego Garcia', 'he': 'דייגו גרסיה', 'hi': 'डिएगो गार्सिया', 'hi-Latn': 'Diego Garcia', 'hr': 'Diego Garcia', 'hsb': 'Diego Garcia', 'ht': 'Diego Garcia', 'hu': 'Diego Garcia', 'hy': 'Դիեգո Գարսիա', 'ia': 'Diego Garcia', 'id': 'Diego Garcia', 'ig': 'Diego Garcia', 'is': 'Diego Garcia', 'it': 'Diego Garcia', 'ja': 'ディエゴガルシア島', 'jv': 'Diego Garcia', 'ka': 'დიეგო-გარსია', 'kea': 'Diegu Garsia', 'kgp': 'Niego Garsija', 'kk': 'Диего-Гарсия', 'kk-Cyrl': 'Диего-Гарсия', 'km': 'ឌៀហ្គោហ្គាស៊ី', 'kn': 'ಡೈಗೋ ಗಾರ್ಸಿಯ', 'ko': '디에고 가르시아', 'kok': 'दिगो गार्सिया', 'kok-Deva': 'दिगो गार्सिया', 'ks': 'ڈیگو گریشیا', 'ks-Arab': 'ڈیگو گریشیا', 'ksh': 'de Diego-Garcia-Ensel', 'ku': 'Diego Garcia', 'kxv': 'ḍiego garsia', 'kxv-Deva': 'डिएगो गार्सिया', 'kxv-Latn': 'ḍiego garsia', 'kxv-Orya': 'ଡିଏଗ ଗାର୍ସିଆ', 'kxv-Telu': 'డియాగో గార్సియా', 'ky': 'Диего Гарсия', 'lb': 'Diego Garcia', 'lld': 'Diego Garcia', 'lo': 'ດິເອໂກ ກາເຊຍ', 'lt': 'Diego Garsija', 'lv': 'Djego Garsijas atols', 'mai': 'डियागो गार्सिया', 'mi': 'Tieko Kāhia', 'mk': 'Диего Гарсија', 'ml': 'ഡീഗോ ഗ്രാഷ്യ', 'mn': 'Диего Гарсиа', 'mr': 'दिएगो गार्सिया', 'ms': 'Diego Garcia', 'mt': 'Diego Garcia', 'my': 'ဒီအဲဂိုဂါစီရာ', 'mzn': 'دیگو گارسیا', 'nb': 'Diego Garcia', 'ne': 'डियगो गार्सिया', 'nl': 'Diego Garcia', 'nn': 'Diego Garcia', 'no': 'Diego Garcia', 'nqo': 'ߖߋߜ߭ߏ߫-ߜ߭ߊߙߑߛߌߦߊ߫', 'om': 'Diyeegoo Gaarshiyaa', 'or': 'ଡିଏଗୋ ଗାର୍ସିଆ', 'pa': 'ਡੀਇਗੋ ਗਾਰਸੀਆ', 'pa-Guru': 'ਡੀਇਗੋ ਗਾਰਸੀਆ', 'pcm': 'Diẹ́gó Garsia', 'pl': 'Diego Garcia', 'ps': 'ډایګو ګارسیا', 'pt': 'Diego Garcia', 'qu': 'Diego García', 'rm': 'Diego Garcia', 'ro': 'Diego Garcia', 'ru': 'Диего-Гарсия', 'sc': 'Diego Garcia', 'sd': 'ڊئيگو گارسيا', 'sd-Arab': 'ڊئيگو گارسيا', 'se': 'Diego Garcia', 'si': 'දියාගෝ ගාර්සියා', 'sk': 'Diego Garcia', 'sl': 'Diego Garcia', 'smn': 'Diego Garcia', 'so': 'Diyeego Karsiya', 'sq': 'Diego-Garsia', 'sr': 'Дијего Гарсија', 'sr-Cyrl': 'Дијего Гарсија', 'sr-Latn': 'Dijego Garsija', 'sv': 'Diego Garcia', 'sw': 'Diego Garcia', 'ta': 'டியகோ கார்ஷியா', 'te': 'డియాగో గార్సియా', 'tg': 'Диего-Гарсия', 'th': 'ดิเอโกการ์เซีย', 'ti': 'ድየጎ ጋርስያ', 'tk': 'Diýego-Garsiýa', 'to': 'Tieko Kāsia', 'tr': 'Diego Garcia', 'tt': 'Диего Гарсия', 'ug': 'دېگو-گارشىيا', 'uk': 'Дієго-Гарсія', 'ur': 'ڈائجو گارسیا', 'uz': 'Diyego-Garsiya', 'uz-Cyrl': 'Диего-Гарсия', 'uz-Latn': 'Diyego-Garsiya', 'vai': 'ꔵꔀꖑ ꔳꘋꗣ', 'vai-Vaii': 'ꔵꔀꖑ ꔳꘋꗣ', 'vec': 'Diego Garcia', 'vi': 'Diego Garcia', 'wae': 'Diego Garcia', 'wo': 'Garsiya', 'xh': 'EDiego Garcia', 'xnr': 'डिएगो गार्सिया', 'yo': 'Diego Gaṣia', 'yrl': 'Diyegu Garasiya', 'yue': '迪亞哥加西亞島', 'yue-Hans': '迪亚哥加西亚岛', 'yue-Hant': '迪亞哥加西亞島', 'zh': '迪戈加西亚岛', 'zh-Hans': '迪戈加西亚岛', 'zh-Hant': '迪亞哥加西亞島', 'zu': 'i-Diego Garcia'},
    'DJ': {'af': 'Djiboeti', 'agq': 'Dzìbuwtì', 'ak': 'Gyibuti', 'am': 'ጂቡቲ', 'ar': 'جيبوتي', 'as': 'জিবুটি', 'asa': 'Jibuti', 'ast': 'Xibuti', 'az': 'Cibuti', 'az-Cyrl': 'Ҹибути', 'az-Latn': 'Cibuti', 'bal-Latn': 'Djebuti', 'bas': 'Jìbutì', 'be': 'Джыбуці', 'be-tarask': 'Джыбуці', 'bez': 'Hujibuti', 'bg': 'Джибути', 'bgn': 'جیبوتی', 'blo': 'Jibutii', 'bm': 'Jibuti', 'bn': 'জিবুতি', 'br': 'Djibouti', 'brx': 'जिबौति', 'bs': 'Džibuti', 'bs-Cyrl': 'Џибути', 'bs-Latn': 'Džibuti', 'ca': 'Djibouti', 'ca-ES-valencia': 'Djibouti', 'ccp': '𑄎𑄨𑄝𑄪𑄖𑄨', 'ce': 'Джибути', 'ceb': 'Djibouti', 'cgg': 'Gyibuti', 'chr': 'ᏥᏊᏗ', 'ckb': 'جیبووتی', 'cs': 'Džibutsko', 'cv': 'Джибути', 'cy': 'Djibouti', 'da': 'Djibouti', 'dav': 'Jibuti', 'de': 'Dschibuti', 'dje': 'Jibuuti', 'dsb': 'Džibuti', 'dyo': 'Jibuti', 'dz': 'ཇི་བྷུ་ཊི', 'ebu': 'Jibuti', 'ee': 'Dzibuti nutome', 'el': 'Τζιμπουτί', 'el-polyton': 'Τζιμπουτί', 'en': 'Djibouti', 'en-Dsrt': '𐐖𐐲𐐺𐐭𐐼𐐨', 'eo': 'Ĝibutio', 'es': 'Yibuti', 'et': 'Djibouti', 'eu': 'Djibuti', 'ewo': 'Dzibutí', 'fa': 'جیبوتی', 'ff': 'Jibutii', 'ff-Adlm': '𞤔𞤭𞤦𞤵𞥅𞤼𞤭', 'ff-Latn': 'Jibutii', 'fi': 'Djibouti', 'fil': 'Djibouti', 'fo': 'Djibuti', 'fr': 'Djibouti', 'fur': 'Gibuti', 'fy': 'Djibouti', 'ga': 'Djibouti', 'gd': 'Diobùtaidh', 'gl': 'Djibuti', 'gsw': 'Tschibuuti', 'gu': 'જીબૌટી', 'guz': 'Jibuti', 'ha': 'Jibuti', 'he': 'ג׳יבוטי', 'hi': 'जिबूती', 'hi-Latn': 'Djibouti', 'hr': 'Džibuti', 'hsb': 'Dźibuti', 'ht': 'Djibouti', 'hu': 'Dzsibuti', 'hy': 'Ջիբութի', 'ia': 'Djibuti', 'id': 'Jibuti', 'ig': 'Djibouti', 'is': 'Djíbútí', 'it': 'Gibuti', 'ja': 'ジブチ', 'jgo': 'Njimbúti', 'jmc': 'Jibuti', 'jv': 'Jibuti', 'ka': 'ჯიბუტი', 'kaa': 'Джибути', 'kaa-Cyrl': 'Джибути', 'kab': 'Ǧibuti', 'kam': 'Jibuti', 'kde': 'Dyibuti', 'kea': 'Djibuti', 'kgp': 'Nhimuti', 'khq': 'Jibuuti', 'ki': 'Jibuti', 'kk': 'Джибути', 'kk-Cyrl': 'Джибути', 'kln': 'Emetab Djibouti', 'km': 'ជីប៊ូទី', 'kn': 'ಜಿಬೂಟಿ', 'ko': '지부티', 'kok': 'जिबूती', 'kok-Deva': 'जिबूती', 'ks': 'جِبوٗتی', 'ks-Arab': 'جِبوٗتی', 'ksb': 'Jibuti', 'ksf': 'dyibutí', 'ksh': 'Dschibutti', 'ku': 'Cîbûtî', 'kxv': 'jibutī', 'kxv-Deva': 'जिबूती', 'kxv-Latn': 'jibutī', 'kxv-Orya': 'ଜିବୁତି', 'kxv-Telu': 'జిబుతి', 'ky': 'Джибути', 'lag': 'Jibúuti', 'lb': 'Dschibuti', 'lg': 'Jjibuti', 'lld': 'Djibouti', 'ln': 'Dzibuti', 'lo': 'ຈິບູຕິ', 'lt': 'Džibutis', 'lu': 'Djibuti', 'luo': 'Djibouti', 'luy': 'Jibuti', 'lv': 'Džibutija', 'mai': 'जिबूति', 'mas': 'Jibuti', 'mer': 'Jibuti', 'mfe': 'Djibouti', 'mg': 'Djiboti', 'mgh': 'Ujibuti', 'mi': 'Tipūti', 'mk': 'Џибути', 'ml': 'ജിബൂത്തി', 'mn': 'Джибути', 'mr': 'जिबौटी', 'ms': 'Djibouti', 'mt': 'il-Djibouti', 'mua': 'Djibouti', 'my': 'ဂျီဘူတီ', 'mzn': 'جیبوتی', 'naq': 'Djibouti', 'nb': 'Djibouti', 'nd': 'Djibouti', 'ne': 'डिजिबुटी', 'nl': 'Djibouti', 'nmg': 'Jibúti', 'nn': 'Djibouti', 'no': 'Djibouti', 'nqo': 'ߖߌߓߎߕߌ߫', 'nyn': 'Gyibuti', 'om': 'Jibuutii', 'or': 'ଜିବୋଟି', 'pa': 'ਜਿਬੂਤੀ', 'pa-Guru': 'ਜਿਬੂਤੀ', 'pcm': 'Jibúti', 'pl': 'Dżibuti', 'ps': 'جبوتي', 'pt': 'Djibuti', 'qu': 'Yibuti', 'rm': 'Dschibuti', 'rn': 'Jibuti', 'ro': 'Djibouti', 'rof': 'Jibuti', 'ru': 'Джибути', 'rwk': 'Jibuti', 'saq': 'Jibuti', 'sbp': 'Jibuti', 'sc': 'Gibuti', 'sd': 'ڊجبيوتي', 'sd-Arab': 'ڊجبيوتي', 'se': 'Djibouti', 'seh': 'Djibuti', 'ses': 'Jibuuti', 'sg': 'Dibutùii', 'shi': 'ⴷⵊⵉⴱⵓⵜⵉ', 'shi-Latn': 'djibuti', 'shi-Tfng': 'ⴷⵊⵉⴱⵓⵜⵉ', 'si': 'ජිබුටි', 'sk': 'Džibutsko', 'sl': 'Džibuti', 'smn': 'Djibouti', 'sn': 'Djibouti', 'so': 'Jabuuti', 'sq': 'Xhibuti', 'sr': 'Џибути', 'sr-Cyrl': 'Џибути', 'sr-Latn': 'Džibuti', 'sv': 'Djibouti', 'sw': 'Jibuti', 'syr': 'ܓܝܒܘܛܝ', 'ta': 'ஜிபௌட்டி', 'te': 'జిబౌటి', 'teo': 'Jibuti', 'tg': 'Ҷибути', 'th': 'จิบูตี', 'ti': 'ጅቡቲ', 'tk': 'Jibuti', 'to': 'Siputi', 'tr': 'Cibuti', 'tt': 'Җибүти', 'twq': 'Jibuuti', 'tzm': 'Ǧibuti', 'ug': 'جىبۇتى', 'uk': 'Джибуті', 'ur': 'جبوتی', 'uz': 'Jibuti', 'uz-Cyrl': 'Жибути', 'uz-Latn': 'Jibuti', 'vai': 'ꕀꖜꔳ', 'vai-Latn': 'Jibhuti', 'vai-Vaii': 'ꕀꖜꔳ', 'vec': 'Jibuti', 'vi': 'Djibouti', 'vun': 'Jibuti', 'wae': 'Dšibuti', 'wo': 'Jibuti', 'xh': 'EDjibouti', 'xnr': 'जिबूती', 'xog': 'Jjibuti', 'yav': 'síputí', 'yi': 'דזשיבוטי', 'yo': 'Díbọ́ótì', 'yrl': 'Dijibuti', 'yue': '吉布地', 'yue-Hans': '吉布地', 'yue-Hant': '吉布地', 'zgh': 'ⴷⵊⵉⴱⵓⵜⵉ', 'zh': '吉布提', 'zh-Hans': '吉布提', 'zh-Hant': '吉布地', 'zu': 'i-Djibouti'},
    'DK': {'af': 'Denemarke', 'agq': 'Dɛnɨmà', 'ak': 'Dɛnmak', 'am': 'ዴንማርክ', 'ar': 'الدانمرك', 'as': 'ডেনমাৰ্ক', 'asa': 'Denmaki', 'ast': 'Dinamarca', 'az': 'Danimarka', 'az-Cyrl': 'Данимарка', 'az-Latn': 'Danimarka', 'bal-Latn': 'Denmárk', 'bas': 'Dànmârk', 'be': 'Данія', 'be-tarask': 'Данія', 'bez': 'Hudenmaki', 'bg': 'Дания', 'blo': 'Ɖanǝmark', 'bm': 'Danemarki', 'bn': 'ডেনমার্ক', 'br': 'Danmark', 'brx': 'डेनमार्क', 'bs': 'Danska', 'bs-Cyrl': 'Данска', 'bs-Latn': 'Danska', 'ca': 'Dinamarca', 'ca-ES-valencia': 'Dinamarca', 'ccp': '𑄓𑄬𑄚𑄴𑄟𑄢𑄴𑄇𑄧', 'ce': 'Дани', 'ceb': 'Denmark', 'cgg': 'Deenimaaka', 'chr': 'ᏗᏂᎹᎦ', 'ckb': 'دانمارک', 'cs': 'Dánsko', 'cv': 'Дани', 'cy': 'Denmarc', 'da': 'Danmark', 'dav': 'Denmaki', 'de': 'Dänemark', 'dje': 'Danemark', 'dsb': 'Dańska', 'dyo': 'Danmark', 'dz': 'ཌེན་མཱཀ', 'ebu': 'Denmaki', 'ee': 'Denmark nutome', 'el': 'Δανία', 'el-polyton': 'Δανία', 'en': 'Denmark', 'en-Dsrt': '𐐔𐐯𐑌𐑋𐐪𐑉𐐿', 'eo': 'Danujo', 'es': 'Dinamarca', 'et': 'Taani', 'eu': 'Danimarka', 'ewo': 'Danəmárəg', 'fa': 'دانمارک', 'ff': 'Danmark', 'ff-Adlm': '𞤁𞤢𞤲𞤵𞤥𞤢𞤪𞤳', 'ff-Latn': 'Danmark', 'fi': 'Tanska', 'fil': 'Denmark', 'fo': 'Danmark', 'fr': 'Danemark', 'fur': 'Danimarcje', 'fy': 'Denemarken', 'ga': 'an Danmhairg', 'gd': 'An Danmhairg', 'gl': 'Dinamarca', 'gsw': 'Tänemark', 'gu': 'ડેનમાર્ક', 'guz': 'Denmaki', 'ha': 'Danmark', 'haw': 'Kenemaka', 'he': 'דנמרק', 'hi': 'डेनमार्क', 'hi-Latn': 'Denmark', 'hr': 'Danska', 'hsb': 'Danska', 'ht': 'Danemark', 'hu': 'Dánia', 'hy': 'Դանիա', 'ia': 'Danmark', 'id': 'Denmark', 'ie': 'Dania', 'ig': 'Denmark', 'is': 'Danmörk', 'it': 'Danimarca', 'ja': 'デンマーク', 'jmc': 'Denmaki', 'jv': 'Dhènemarken', 'ka': 'დანია', 'kaa': 'Дания', 'kaa-Cyrl': 'Дания', 'kab': 'Denmark', 'kam': 'Denmaki', 'kde': 'Denmaki', 'kea': 'Dinamarka', 'kgp': 'Ninỹmỹrka', 'khq': 'Danemark', 'ki': 'Denmaki', 'kk': 'Дания', 'kk-Cyrl': 'Дания', 'kln': 'Emetab Denmark', 'km': 'ដាណឺម៉ាក', 'kn': 'ಡೆನ್ಮಾರ್ಕ್', 'ko': '덴마크', 'kok': 'डेनमार्क', 'kok-Deva': 'डेनमार्क', 'ks': 'ڈینمارٕک', 'ks-Arab': 'ڈینمارٕک', 'ksb': 'Denmaki', 'ksf': 'danmak', 'ksh': 'Dänemark', 'ku': 'Danîmarka', 'kxv': 'ḍenmark', 'kxv-Deva': 'डेनमार्क', 'kxv-Latn': 'ḍenmark', 'kxv-Orya': 'ଡେନମାର୍କ', 'kxv-Telu': 'డెన్మార్క', 'ky': 'Дания', 'lag': 'Denimaki', 'lb': 'Dänemark', 'lg': 'Denimaaka', 'lld': 'Danimarca', 'ln': 'Danɛmarike', 'lo': 'ເດນມາກ', 'lt': 'Danija', 'lu': 'Danemalaku', 'luo': 'Denmark', 'luy': 'Denmaki', 'lv': 'Dānija', 'mai': 'डेनमार्क', 'mas': 'Denmaki', 'mer': 'Denimaki', 'mfe': 'Dannmark', 'mg': 'Danmarka', 'mgh': 'Udenimaka', 'mi': 'Tenemāka', 'mk': 'Данска', 'ml': 'ഡെൻമാർക്ക്', 'mn': 'Дани', 'mr': 'डेन्मार्क', 'ms': 'Denmark', 'mt': 'id-Danimarka', 'mua': 'Daŋmark', 'my': 'ဒိန်းမတ်', 'mzn': 'دانمارک', 'naq': 'Denmarki', 'nb': 'Danmark', 'nd': 'Denmakhi', 'ne': 'डेनमार्क', 'nl': 'Denemarken', 'nmg': 'Danemark', 'nn': 'Danmark', 'no': 'Danmark', 'nqo': 'ߘߊߣߌߡߊߙߑߞ', 'nyn': 'Deenimaaka', 'om': 'Deenmaark', 'or': 'ଡେନମାର୍କ', 'pa': 'ਡੈਨਮਾਰਕ', 'pa-Guru': 'ਡੈਨਮਾਰਕ', 'pcm': 'Dẹ́nmak', 'pl': 'Dania', 'ps': 'ډنمارک', 'pt': 'Dinamarca', 'qu': 'Dinamarca', 'rm': 'Danemarc', 'rn': 'Danimariki', 'ro': 'Danemarca', 'rof': 'Denmaki', 'ru': 'Дания', 'rwk': 'Denmaki', 'saq': 'Denmaki', 'sbp': 'Denimaki', 'sc': 'Danimarca', 'sd': 'ڊينمارڪ', 'sd-Arab': 'ڊينمارڪ', 'se': 'Dánmárku', 'seh': 'Dinamarca', 'ses': 'Danemark', 'sg': 'Danemêrke', 'shi': 'ⴷⴰⵏⵎⴰⵔⴽ', 'shi-Latn': 'danmark', 'shi-Tfng': 'ⴷⴰⵏⵎⴰⵔⴽ', 'si': 'ඩෙන්මාර්කය', 'sk': 'Dánsko', 'sl': 'Danska', 'smn': 'Tanska', 'sn': 'Denmark', 'so': 'Denmark', 'sq': 'Danimarkë', 'sr': 'Данска', 'sr-Cyrl': 'Данска', 'sr-Latn': 'Danska', 'sv': 'Danmark', 'sw': 'Denmaki', 'syr': 'ܕܐܢܡܐܪܩ', 'ta': 'டென்மார்க்', 'te': 'డెన్మార్క్', 'teo': 'Denmaki', 'tg': 'Дания', 'th': 'เดนมาร์ก', 'ti': 'ደንማርክ', 'tk': 'Daniýa', 'to': 'Tenimaʻake', 'tr': 'Danimarka', 'tt': 'Дания', 'twq': 'Danemark', 'tzm': 'Danmark', 'ug': 'دانىيە', 'uk': 'Данія', 'ur': 'ڈنمارک', 'uz': 'Daniya', 'uz-Cyrl': 'Дания', 'uz-Latn': 'Daniya', 'vai': 'ꕜꕇꕮꕃ', 'vai-Latn': 'Danimaha', 'vai-Vaii': 'ꕜꕇꕮꕃ', 'vec': 'Danimarca', 'vi': 'Đan Mạch', 'vun': 'Denmaki', 'wae': 'Dänemark', 'wo': 'Danmàrk', 'xh': 'EDenmark', 'xnr': 'डेनमार्क', 'xog': 'Denimaaka', 'yav': 'tanemálk', 'yi': 'דענמאַרק', 'yo': 'Dẹ́mákì', 'yrl': 'Dinamaka', 'yue': '丹麥', 'yue-Hans': '丹麦', 'yue-Hant': '丹麥', 'zgh': 'ⴷⴰⵏⵎⴰⵔⴽ', 'zh': '丹麦', 'zh-Hans': '丹麦', 'zh-Hant': '丹麥', 'zu': 'i-Denmark'},
    'DM': {'af': 'Dominica', 'agq': 'Dòmenekà', 'ak': 'Dɔmeneka', 'am': 'ዶሚኒካ', 'ar': 'دومينيكا', 'as': 'ড’মিনিকা', 'asa': 'Dominika', 'ast': 'Dominica', 'az': 'Dominika', 'az-Cyrl': 'Доминика', 'az-Latn': 'Dominika', 'bal-Latn': 'Duminiká', 'bas': 'Dòmnîk', 'be': 'Дамініка', 'be-tarask': 'Дамініка', 'bez': 'Hudominika', 'bg': 'Доминика', 'blo': 'Ɖominikaa', 'bm': 'Dɔminiki', 'bn': 'ডোমিনিকা', 'br': 'Dominica', 'brx': 'दमिनिका', 'bs': 'Dominika', 'bs-Cyrl': 'Доминика', 'bs-Latn': 'Dominika', 'ca': 'Dominica', 'ca-ES-valencia': 'Dominica', 'ccp': '𑄓𑄮𑄟𑄨𑄚𑄨𑄇', 'ce': 'Доминика', 'ceb': 'Dominica', 'cgg': 'Dominika', 'chr': 'ᏙᎻᏂᎧ', 'ckb': 'دۆمینیکا', 'cs': 'Dominika', 'cv': 'Доминика', 'cy': 'Dominica', 'da': 'Dominica', 'dav': 'Dominika', 'de': 'Dominica', 'dsb': 'Dominika', 'dyo': 'Dominika', 'dz': 'ཌོ་མི་ནི་ཀ', 'ebu': 'Dominika', 'ee': 'Dominika nutome', 'el': 'Ντομίνικα', 'el-polyton': 'Ντομίνικα', 'en': 'Dominica', 'en-Dsrt': '𐐔𐐪𐑋𐐲𐑌𐐨𐐿𐐲', 'eo': 'Dominiko', 'es': 'Dominica', 'et': 'Dominica', 'eu': 'Dominika', 'ewo': 'Dómənika', 'fa': 'دومینیکا', 'ff': 'Dominika', 'ff-Adlm': '𞤁𞤮𞤥𞤭𞤲𞤭𞤳𞤢𞥄', 'ff-Latn': 'Dominika', 'fi': 'Dominica', 'fil': 'Dominica', 'fo': 'Dominika', 'fr': 'Dominique', 'fur': 'Dominiche', 'fy': 'Dominika', 'ga': 'Doiminice', 'gd': 'Doiminicea', 'gl': 'Dominica', 'gsw': 'Tominica', 'gu': 'ડોમિનિકા', 'guz': 'Dominika', 'ha': 'Dominika', 'he': 'דומיניקה', 'hi': 'डोमिनिका', 'hi-Latn': 'Dominica', 'hr': 'Dominika', 'hsb': 'Dominika', 'ht': 'Dominique', 'hu': 'Dominika', 'hy': 'Դոմինիկա', 'ia': 'Dominica', 'id': 'Dominika', 'ig': 'Dominica', 'is': 'Dóminíka', 'it': 'Dominica', 'ja': 'ドミニカ国', 'jmc': 'Dominika', 'jv': 'Dominika', 'ka': 'დომინიკა', 'kaa': 'Доминика', 'kaa-Cyrl': 'Доминика', 'kab': 'Duminik', 'kam': 'Dominika', 'kde': 'Dominika', 'kea': 'Dominika', 'kgp': 'Nomĩnĩka', 'khq': 'Doominiki', 'ki': 'Dominika', 'kk': 'Доминика', 'kk-Cyrl': 'Доминика', 'kln': 'Emetab Dominica', 'km': 'ដូមីនីក', 'kn': 'ಡೊಮಿನಿಕಾ', 'ko': '도미니카', 'kok': 'डोमिनीका', 'kok-Deva': 'डोमिनीका', 'ks': 'ڈومِنِکا', 'ks-Arab': 'ڈومِنِکا', 'ksb': 'Dominika', 'ksf': 'dɔminik', 'ksh': 'Dominnika', 'ku': 'Domînîka', 'kxv': 'dominika', 'kxv-Deva': 'डोमिनिका', 'kxv-Latn': 'dominika', 'kxv-Orya': 'ଡମିନିକା', 'kxv-Telu': 'డొమినికా', 'ky': 'Доминика', 'lag': 'Domínɨka', 'lb': 'Dominica', 'lg': 'Dominika', 'lld': 'Dominica', 'ln': 'Domínike', 'lo': 'ໂດມີນິຄາ', 'lt': 'Dominika', 'lu': 'Duminiku', 'luo': 'Dominica', 'luy': 'Dominika', 'lv': 'Dominika', 'mai': 'डोमिनिका', 'mas': 'Dominika', 'mer': 'Dominika', 'mfe': 'Dominik', 'mg': 'Dominika', 'mgh': 'Udominika', 'mi': 'Tominika', 'mk': 'Доминика', 'ml': 'ഡൊമിനിക്ക', 'mn': 'Доминика', 'mr': 'डोमिनिका', 'ms': 'Dominica', 'mt': 'Dominica', 'mua': 'Dominik', 'my': 'ဒိုမီနီကာ', 'mzn': 'دومنیکا', 'naq': 'Dominicab', 'nb': 'Dominica', 'nd': 'Dominikha', 'ne': 'डोमिनिका', 'nl': 'Dominica', 'nmg': 'Dominíka', 'nn': 'Dominica', 'no': 'Dominica', 'nqo': 'ߘߏߡߣߌߞ', 'nyn': 'Dominika', 'om': 'Dominiikaa', 'or': 'ଡୋମିନିକା', 'pa': 'ਡੋਮੀਨਿਕਾ', 'pa-Guru': 'ਡੋਮੀਨਿਕਾ', 'pcm': 'Dọmíníka', 'pl': 'Dominika', 'ps': 'دومینیکا', 'pt': 'Dominica', 'qu': 'Dominica', 'rm': 'Dominica', 'rn': 'Dominika', 'ro': 'Dominica', 'rof': 'Dominika', 'ru': 'Доминика', 'rwk': 'Dominika', 'saq': 'Dominika', 'sbp': 'Dominika', 'sc': 'Dominica', 'sd': 'ڊومينيڪا', 'sd-Arab': 'ڊومينيڪا', 'se': 'Dominica', 'seh': 'Dominica', 'ses': 'Doominiki', 'sg': 'Dömïnîka', 'shi': 'ⴷⵓⵎⵉⵏⵉⴽ', 'shi-Latn': 'duminik', 'shi-Tfng': 'ⴷⵓⵎⵉⵏⵉⴽ', 'si': 'ඩොමිනිකාව', 'sk': 'Dominika', 'sl': 'Dominika', 'smn': 'Dominica', 'sn': 'Dominica', 'so': 'Dominika', 'sq': 'Dominikë', 'sr': 'Доминика', 'sr-Cyrl': 'Доминика', 'sr-Latn': 'Dominika', 'sv': 'Dominica', 'sw': 'Dominika', 'syr': 'ܕܘܡܝܢܝܩܐ', 'ta': 'டொமினிகா', 'te': 'డొమినికా', 'teo': 'Dominika', 'tg': 'Доминика', 'th': 'โดมินิกา', 'ti': 'ዶሚኒካ', 'tk': 'Dominika', 'to': 'Tominika', 'tr': 'Dominika', 'tt': 'Доминика', 'twq': 'Doominiki', 'tzm': 'Ḍuminika', 'ug': 'دومىنىكا', 'uk': 'Домініка', 'ur': 'ڈومنیکا', 'uz': 'Dominika', 'uz-Cyrl': 'Доминика', 'uz-Latn': 'Dominika', 'vai': 'ꖁꕆꕇꕪ', 'vai-Latn': 'Domíiníka', 'vai-Vaii': 'ꖁꕆꕇꕪ', 'vec': 'Doménega', 'vi': 'Dominica', 'vun': 'Dominika', 'wae': 'Doninica', 'wo': 'Dominik', 'xh': 'EDominica', 'xnr': 'डोमिनिका', 'xog': 'Dominika', 'yav': 'túmúnéke', 'yi': 'דאמיניקע', 'yo': 'Dòmíníkà', 'yrl': 'Dominika', 'yue': '多米尼克', 'yue-Hans': '多米尼克', 'yue-Hant': '多米尼克', 'zgh': 'ⴷⵓⵎⵉⵏⵉⴽ', 'zh': '多米尼克', 'zh-Hans': '多米尼克', 'zh-Hant': '多米尼克', 'zu': 'i-Dominica'},
    'DO': {'af': 'Dominikaanse Republiek', 'agq': 'Dòmenekà Lèkpubèlè', 'ak': 'Dɔmeneka Man', 'am': 'ዶመኒካን ሪፑብሊክ', 'ar': 'جمهورية الدومينيكان', 'as': 'ড’মিনিকান ৰিপাব্লিক', 'asa': 'Jamhuri ya Dominika', 'ast': 'República Dominicana', 'az': 'Dominikan Respublikası', 'az-Cyrl': 'Доминикан Республикасы', 'az-Latn': 'Dominikan Respublikası', 'bal-Latn': 'Duminiki Están', 'bas': 'Dòmnikà', 'be': 'Дамініканская Рэспубліка', 'be-tarask': 'Дамініканская Рэспубліка', 'bez': 'Ijamhuri ya Hudominika', 'bg': 'Доминиканска република', 'blo': 'Ɖominikaa kaRepibliiki', 'bm': 'Dɔmimiki republiki', 'bn': 'ডোমেনিকান প্রজাতন্ত্র', 'br': 'Republik Dominikan', 'brx': 'दमिनिकान सुबुंखुंथाय', 'bs': 'Dominikanska Republika', 'bs-Cyrl': 'Доминиканска Република', 'bs-Latn': 'Dominikanska Republika', 'ca': 'República Dominicana', 'ca-ES-valencia': 'República Dominicana', 'ccp': '𑄓𑄮𑄟𑄨𑄚𑄨𑄇𑄚𑄴 𑄛𑄳𑄢𑄧𑄎𑄖𑄧𑄚𑄴𑄖𑄳𑄢𑄧', 'ce': 'Доминикан Республика', 'ceb': 'Dominican Republic', 'cgg': 'Ripaaburika ya Dominica', 'chr': 'ᏙᎻᏂᎧᏂ ᏍᎦᏚᎩ', 'ckb': 'کۆماری دۆمینیکا', 'cs': 'Dominikánská republika', 'cv': 'Доминикан Республики', 'cy': 'Gweriniaeth Dominica', 'da': 'Den Dominikanske Republik', 'dav': 'Jamhuri ya Dominika', 'de': 'Dominikanische Republik', 'dje': 'Doominiki laboo', 'dsb': 'Dominikańska republika', 'dyo': 'Mofam mati Dominik', 'dz': 'ཌོ་མི་ནི་ཀཱན་ རི་པབ་ལིཀ', 'ebu': 'Jamhuri ya Dominika', 'ee': 'Dominika repɔblik nutome', 'el': 'Δομινικανή Δημοκρατία', 'el-polyton': 'Δομινικανὴ Δημοκρατία', 'en': 'Dominican Republic', 'en-Dsrt': '𐐔𐐲𐑋𐐮𐑌𐐲𐐿𐐲𐑌 𐐡𐐨𐐹𐐲𐐺𐑊𐐮𐐿', 'eo': 'Domingo', 'es': 'República Dominicana', 'et': 'Dominikaani Vabariik', 'eu': 'Dominikar Errepublika', 'ewo': 'République dominicaine', 'fa': 'جمهوری دومینیکن', 'ff': 'Ndenndanndi Dominika', 'ff-Adlm': '𞤐𞤣𞤫𞤲𞤣𞤭 𞤁𞤮𞤥𞤭𞤲𞤭𞤳𞤢𞥄', 'ff-Latn': 'Ndenndanndi Dominika', 'fi': 'Dominikaaninen tasavalta', 'fil': 'Dominican Republic', 'fo': 'Dominikalýðveldið', 'fr': 'République dominicaine', 'fur': 'Republiche dominicane', 'fy': 'Dominikaanske Republyk', 'ga': 'an Phoblacht Dhoiminiceach', 'gd': 'A’ Phoblachd Dhoiminiceach', 'gl': 'República Dominicana', 'gsw': 'Tominikaanischi Republik', 'gu': 'ડોમિનિકન રિપબ્લિક', 'guz': 'Jamhuri ya Dominika', 'ha': 'Jamhuriyar Dominika', 'he': 'הרפובליקה הדומיניקנית', 'hi': 'डोमिनिकन गणराज्य', 'hi-Latn': 'Dominican Republic', 'hr': 'Dominikanska Republika', 'hsb': 'Dominikanska republika', 'ht': 'République dominicaine', 'hu': 'Dominikai Köztársaság', 'hy': 'Դոմինիկյան Հանրապետություն', 'ia': 'Republica Dominican', 'id': 'Republik Dominika', 'ig': 'Dominican Republik', 'is': 'Dóminíska lýðveldið', 'it': 'Repubblica Dominicana', 'ja': 'ドミニカ共和国', 'jmc': 'Jamhuri ya Dominika', 'jv': 'Républik Dominika', 'ka': 'დომინიკელთა რესპუბლიკა', 'kaa': 'Доминикана Республикасы', 'kaa-Cyrl': 'Доминикана Республикасы', 'kab': 'Tigduda Taduminikit', 'kam': 'Jamhuri ya Dominika', 'kde': 'Jamuhuli ya Dominika', 'kea': 'Repúblika Dominikana', 'kgp': 'Repumrika Nomĩnĩkỹnỹ', 'khq': 'Doominiki laboo', 'ki': 'Jamhuri ya Dominika', 'kk': 'Доминикан Республикасы', 'kk-Cyrl': 'Доминикан Республикасы', 'kln': 'Emetab Dominican Republic', 'km': 'សាធារណរដ្ឋ\u200bដូមីនីក', 'kn': 'ಡೊಮೆನಿಕನ್ ರಿಪಬ್ಲಿಕ್', 'ko': '도미니카 공화국', 'kok': 'डोमिनिकन प्रजासत्ताक', 'kok-Deva': 'डोमिनिकन प्रजासत्ताक', 'ks': 'ڈومِنِکَن جموٗرِیَت', 'ks-Arab': 'ڈومِنِکَن جموٗرِیَت', 'ksb': 'Jamhuli ya Dominika', 'ksf': 'dɔminik rɛpublík', 'ksh': 'de Dommenekaanesche Reppublik', 'ku': 'Komara Domînîkê', 'kxv': 'dominikan republik', 'kxv-Deva': 'डोमिनिकन गणराज्य', 'kxv-Latn': 'dominikan republik', 'kxv-Orya': 'ଡମିନିକା ତି ରିପବ୍ଲିକ୍', 'kxv-Telu': 'డొమినికాన్ రిపబ్లిక్', 'ky': 'Доминика Республикасы', 'lag': 'Jamuhúuri ya Dominɨka', 'lb': 'Dominikanesch Republik', 'lg': 'Lipubulika ya Dominika', 'lld': 'Republica Dominicana', 'ln': 'Repibiki ya Domínikɛ', 'lo': 'ສາທາລະນະລັດ ໂດມິນິກັນ', 'lt': 'Dominikos Respublika', 'lu': 'Ditunga wa Duminiku', 'luo': 'Dominican Republic', 'luy': 'Jamhuri ya Dominika', 'lv': 'Dominikāna', 'mai': 'डोमिनिकन रिपब्लिक', 'mas': 'Jamhuri ya Dominika', 'mer': 'Nthĩ ya Dominika', 'mfe': 'Repiblik dominikin', 'mg': 'Repoblika Dominikanina', 'mi': 'Te Whenua Tūhake o Tominika', 'mk': 'Доминиканска Република', 'ml': 'ഡൊമിനിക്കൻ റിപ്പബ്ലിക്ക്', 'mn': 'Бүгд Найрамдах Доминикан Улс', 'mr': 'डोमिनिकन प्रजासत्ताक', 'ms': 'Republik Dominica', 'mt': 'ir-Repubblika Dominicana', 'mua': 'Sǝr Dominik ma lii', 'my': 'ဒိုမီနီကန်', 'mzn': 'دومنیکن جمهوری', 'naq': 'Dominican Republic', 'nb': 'Den dominikanske republikk', 'nd': 'Dominican Republic', 'ne': 'डोमिनिकन गणतन्त्र', 'nl': 'Dominicaanse Republiek', 'nmg': 'Nlambɔ́ Dominíka', 'nn': 'Den dominikanske republikken', 'no': 'Den dominikanske republikk', 'nqo': 'ߘߏߡߣߌߞ ߞߊ߲ߓߍ߲', 'nyn': 'Ripaaburika ya Dominica', 'om': 'Dominikaa Rippaabilik', 'or': 'ଡୋମିନିକାନ୍\u200c ସାଧାରଣତନ୍ତ୍ର', 'pa': 'ਡੋਮੀਨਿਕਾਈ ਗਣਰਾਜ', 'pa-Guru': 'ਡੋਮੀਨਿਕਾਈ ਗਣਰਾਜ', 'pcm': 'Dọmíníka Ripọ́blik', 'pl': 'Dominikana', 'ps': 'جمهوريه ډومينيکن', 'pt': 'República Dominicana', 'qu': 'República Dominicana', 'rm': 'Republica Dominicana', 'rn': 'Repubulika ya Dominika', 'ro': 'Republica Dominicană', 'rof': 'Jamhuri ya Dominika', 'ru': 'Доминиканская Республика', 'rwk': 'Jamhuri ya Dominika', 'saq': 'Jamhuri ya Dominika', 'sbp': 'Jamuhuli ya Dominika', 'sc': 'Repùblica Dominicana', 'sd': 'ڊومينيڪن جمهوريه', 'sd-Arab': 'ڊومينيڪن جمهوريه', 'se': 'Dominikána dásseváldi', 'seh': 'República Dominicana', 'ses': 'Doominiki laboo', 'sg': 'Ködörösêse tî Dominîka', 'shi': 'ⵜⴰⴳⴷⵓⴷⴰⵏⵜ ⵜⴰⴷⵓⵎⵉⵏⵉⴽⵜ', 'shi-Latn': 'tagdudant taduminikt', 'shi-Tfng': 'ⵜⴰⴳⴷⵓⴷⴰⵏⵜ ⵜⴰⴷⵓⵎⵉⵏⵉⴽⵜ', 'si': 'ඩොමිනිකා ජනරජය', 'sk': 'Dominikánska republika', 'sl': 'Dominikanska republika', 'smn': 'Dominikaanisâš täsiväldi', 'sn': 'Dominican Republic', 'so': 'Jamhuuriyaddda Dominika', 'sq': 'Republika Dominikane', 'sr': 'Доминиканска Република', 'sr-Cyrl': 'Доминиканска Република', 'sr-Latn': 'Dominikanska Republika', 'sv': 'Dominikanska republiken', 'sw': 'Jamhuri ya Dominika', 'syr': 'ܩܘܛܢܝܘܬܐ ܕܘܡܝܢܝܩܐܢܝܬܐ', 'ta': 'டொமினிகன் குடியரசு', 'te': 'డొమినికన్ రిపబ్లిక్', 'teo': 'Jamhuri ya Dominika', 'tg': 'Ҷумҳурии Доминикан', 'th': 'สาธารณรัฐโดมินิกัน', 'ti': 'ዶሚኒካዊት ሪፓብሊክ', 'tk': 'Dominikan Respublikasy', 'to': 'Lepupelika Tominika', 'tr': 'Dominik Cumhuriyeti', 'tt': 'Доминикана Республикасы', 'twq': 'Doominiki laboo', 'tzm': 'Tagduda n Ḍuminikan', 'ug': 'دومىنىكا جۇمھۇرىيىتى', 'uk': 'Домініканська Республіка', 'ur': 'جمہوریہ ڈومينيکن', 'uz': 'Dominikan Respublikasi', 'uz-Cyrl': 'Доминикан Республикаси', 'uz-Latn': 'Dominikan Respublikasi', 'vai': 'ꖁꕆꕇꕪꘋ ꕸꕱꔀ', 'vai-Latn': 'Domíiníka Ɓoloe', 'vai-Vaii': 'ꖁꕆꕇꕪꘋ ꕸꕱꔀ', 'vec': 'Repùblega Domenegana', 'vi': 'Cộng hòa Dominica', 'vun': 'Jamhuri ya Dominika', 'wae': 'Dominikaniši Rebublik', 'wo': 'Repiblik Dominiken', 'xh': 'EDominican Republic', 'xnr': 'डोमिनिकन गणराज्य', 'xog': 'Lipabulika ya Dominika', 'yav': 'kitɔŋ kí tumunikɛ́ŋ', 'yi': 'דאמיניקאַנישע רעפּובליק', 'yo': 'Dòmíníkánì', 'yrl': 'Repubirika Duminikana', 'yue': '多明尼加共和國', 'yue-Hans': '多明尼加共和国', 'yue-Hant': '多明尼加共和國', 'zgh': 'ⵜⴰⴳⴷⵓⴷⴰⵏⵜ ⵜⴰⴷⵓⵎⵉⵏⵉⴽⵜ', 'zh': '多米尼加共和国', 'zh-Hans': '多米尼加共和国', 'zh-Hant': '多明尼加共和國', 'zu': 'i-Dominican Republic'},
    'DZ': {'af': 'Algerië', 'agq': 'Àadzɛlìa', 'ak': 'Ɔlgyeria', 'am': 'አልጄሪያ', 'ar': 'الجزائر', 'as': 'আলজেৰিয়া', 'asa': 'Aljeria', 'ast': 'Arxelia', 'az': 'Əlcəzair', 'az-Cyrl': 'Әлҹәзаир', 'az-Latn': 'Əlcəzair', 'bal-Latn': 'Aljiriá', 'bas': 'Àlgerìà', 'be': 'Алжыр', 'be-tarask': 'Алжыр', 'bez': 'Hualjelia', 'bg': 'Алжир', 'bgn': 'الجزایر', 'blo': 'Aljerii', 'bm': 'Alizeri', 'bn': 'আলজেরিয়া', 'br': 'Aljeria', 'brx': 'आल्जेरिया', 'bs': 'Alžir', 'bs-Cyrl': 'Алжир', 'bs-Latn': 'Alžir', 'ca': 'Algèria', 'ca-ES-valencia': 'Algèria', 'ccp': '𑄃𑄢𑄴𑄎𑄬𑄢𑄨𑄠', 'ce': 'Алжир', 'ceb': 'Algeria', 'cgg': 'Arigyeriya', 'chr': 'ᎠᎵᏥᎵᏯ', 'ckb': 'جەزایر', 'cs': 'Alžírsko', 'cv': 'Алжир', 'cy': 'Algeria', 'da': 'Algeriet', 'dav': 'Aljeria', 'de': 'Algerien', 'dje': 'Alžeeri', 'dsb': 'Algeriska', 'dyo': 'Alseri', 'dz': 'ཨཱལ་ཇི་རི་ཡ', 'ebu': 'Aljeria', 'ee': 'Algeria nutome', 'el': 'Αλγερία', 'el-polyton': 'Ἀλγερία', 'en': 'Algeria', 'en-Dsrt': '𐐈𐑊𐐾𐐮𐑉𐐨𐐲', 'eo': 'Alĝerio', 'es': 'Argelia', 'et': 'Alžeeria', 'eu': 'Aljeria', 'ewo': 'Aləyéria', 'fa': 'الجزایر', 'ff': 'Alaseri', 'ff-Adlm': '𞤀𞤤𞤶𞤢𞤪𞤭𞥅', 'ff-Latn': 'Alaseri', 'fi': 'Algeria', 'fil': 'Algeria', 'fo': 'Algeria', 'fr': 'Algérie', 'fur': 'Alzerie', 'fy': 'Algerije', 'ga': 'An Ailgéir', 'gd': 'Aildiria', 'gl': 'Alxeria', 'gsw': 'Algeerie', 'gu': 'અલ્જીરિયા', 'guz': 'Aljeria', 'ha': 'Aljeriya', 'he': 'אלג׳יריה', 'hi': 'अल्जीरिया', 'hi-Latn': 'Algeria', 'hr': 'Alžir', 'hsb': 'Algeriska', 'ht': 'Algérie', 'hu': 'Algéria', 'hy': 'Ալժիր', 'ia': 'Algeria', 'id': 'Aljazair', 'ig': 'Algeria', 'is': 'Alsír', 'it': 'Algeria', 'ja': 'アルジェリア', 'jgo': 'Aljɛlî', 'jmc': 'Aljeria', 'jv': 'Aljasair', 'ka': 'ალჟირი', 'kaa': 'Алжир', 'kaa-Cyrl': 'Алжир', 'kab': 'Lezzayer', 'kam': 'Aljeria', 'kde': 'Aljelia', 'kea': 'Arjélia', 'kgp': 'Arjérija', 'khq': 'Alžeeri', 'ki': 'Aljeria', 'kk': 'Алжир', 'kk-Cyrl': 'Алжир', 'kln': 'Emetab Algeria', 'km': 'អាល់ហ្សេរី', 'kn': 'ಅಲ್ಜೀರಿಯ', 'ko': '알제리', 'kok': 'अल्जेरिया', 'kok-Deva': 'अल्जेरिया', 'ks': 'اؠلجیرِیا', 'ks-Arab': 'اؠلجیرِیا', 'ksb': 'Aljelia', 'ksf': 'aljɛrí', 'ksh': 'Alljeerije', 'ku': 'Cezayîr', 'kxv': 'aljīriaa', 'kxv-Deva': 'अल्जीरिया', 'kxv-Latn': 'aljīriaa', 'kxv-Orya': 'ଅଲଜେରିଆ', 'kxv-Telu': 'ఆల్జిరియా', 'ky': 'Алжир', 'lag': 'Alijéria', 'lb': 'Algerien', 'lg': 'Aligerya', 'lld': 'Algeria', 'ln': 'Alizɛri', 'lo': 'ອັລຈິເຣຍ', 'lt': 'Alžyras', 'lu': 'Alijeri', 'luo': 'Algeria', 'luy': 'Aljeria', 'lv': 'Alžīrija', 'mai': 'अल्जीरिया', 'mas': 'Aljeria', 'mer': 'Angiria', 'mfe': 'Alzeri', 'mg': 'Alzeria', 'mgh': 'Alujeria', 'mi': 'Aratiria', 'mk': 'Алжир', 'ml': 'അൾജീരിയ', 'mn': 'Алжир', 'mr': 'अल्जीरिया', 'ms': 'Algeria', 'mt': 'l-Alġerija', 'mua': 'algeriya', 'my': 'အယ်လ်ဂျီးရီးယား', 'mzn': 'الجزیره', 'naq': 'Algeriab', 'nb': 'Algerie', 'nd': 'Aljeriya', 'ne': 'अल्जेरिया', 'nl': 'Algerije', 'nmg': 'Algeria', 'nn': 'Algerie', 'no': 'Algerie', 'nqo': 'ߊߟߑߖ߭ߋߙߌ߫', 'nus': 'Algeria', 'nyn': 'Arigyeriya', 'om': 'Aljeeriyaa', 'or': 'ଆଲଜେରିଆ', 'pa': 'ਅਲਜੀਰੀਆ', 'pa-Guru': 'ਅਲਜੀਰੀਆ', 'pcm': 'Aljíria', 'pl': 'Algieria', 'ps': 'الجزایر', 'pt': 'Argélia', 'qu': 'Argelia', 'rm': 'Algeria', 'rn': 'Alijeriya', 'ro': 'Algeria', 'rof': 'Aljeria', 'ru': 'Алжир', 'rwk': 'Aljeria', 'saq': 'Aljeria', 'sbp': 'Alijeliya', 'sc': 'Algeria', 'sd': 'الجيريا', 'sd-Arab': 'الجيريا', 'se': 'Algeria', 'seh': 'Argélia', 'ses': 'Alžeeri', 'sg': 'Alzerïi', 'shi': 'ⴷⵣⴰⵢⵔ', 'shi-Latn': 'dzayr', 'shi-Tfng': 'ⴷⵣⴰⵢⵔ', 'si': 'ඇල්ජීරියාව', 'sk': 'Alžírsko', 'sl': 'Alžirija', 'smn': 'Algeria', 'sn': 'Aljeria', 'so': 'Aljeeriya', 'sq': 'Algjeri', 'sr': 'Алжир', 'sr-Cyrl': 'Алжир', 'sr-Latn': 'Alžir', 'sv': 'Algeriet', 'sw': 'Aljeria', 'syr': 'ܓܙܐܪ', 'ta': 'அல்ஜீரியா', 'te': 'అల్జీరియా', 'teo': 'Aljeria', 'tg': 'Алҷазоир', 'th': 'แอลจีเรีย', 'ti': 'ኣልጀርያ', 'tk': 'Alžir', 'to': 'ʻAlisilia', 'tr': 'Cezayir', 'tt': 'Алжир', 'twq': 'Alžeeri', 'tzm': 'Dzayer', 'ug': 'ئالجىرىيە', 'uk': 'Алжир', 'ur': 'الجیریا', 'uz': 'Jazoir', 'uz-Cyrl': 'Жазоир', 'uz-Latn': 'Jazoir', 'vai': 'ꕉꔷꔠꔸꕩ', 'vai-Latn': 'Agiriya', 'vai-Vaii': 'ꕉꔷꔠꔸꕩ', 'vec': 'Aljerìa', 'vi': 'Algeria', 'vun': 'Aljeria', 'wae': 'Algerie', 'wo': 'Alseri', 'xh': 'E-Algeria', 'xnr': 'अल्जीरिया', 'xog': 'Aligerya', 'yav': 'Alselí', 'yo': 'Àlùgèríánì', 'yrl': 'Argeriya', 'yue': '阿爾及利亞', 'yue-Hans': '阿尔及利亚', 'yue-Hant': '阿爾及利亞', 'zgh': 'ⴷⵣⴰⵢⵔ', 'zh': '阿尔及利亚', 'zh-Hans': '阿尔及利亚', 'zh-Hant': '阿爾及利亞', 'zu': 'i-Algeria'},
    'Deva': {'af': 'Devanagari', 'ak': 'Dɛvanagari kasa', 'am': 'ደቫንጋሪ', 'ar': 'الديفاناجاري', 'as': 'দেৱনাগৰী', 'ast': 'devanagari', 'az': 'devanaqari', 'az-Latn': 'devanaqari', 'bal-Latn': 'Déwá', 'be': 'дэванагары', 'be-tarask': 'дэванагары', 'bg': 'деванагари', 'bgc': 'देवानागारी', 'bgn': 'دیوانگری', 'bho': 'देवानागारी', 'blo': 'defanagari kʊja', 'bn': 'দেবনগরি', 'br': 'devanagari', 'brx': 'देबनागिरि', 'bs': 'pismo devanagari', 'bs-Cyrl': 'деванагари', 'bs-Latn': 'pismo devanagari', 'ca': 'devanagari', 'ca-ES-valencia': 'devanagari', 'ccp': '𑄘𑄬𑄛𑄴𑄚𑄉𑄧𑄢𑄨', 'ce': 'деванагари', 'chr': 'ᏕᏫᎾᎦᎵ', 'ckb': 'دەڤەناگەری', 'cs': 'dévanágarí', 'cy': 'Devanagari', 'da': 'devanagari', 'de': 'Devanagari', 'doi': 'देवनागरी', 'dsb': 'devanagari', 'dz': 'དེ་ཝ་ན་ག་རི་ཡིག་གུ', 'ee': 'devanagarigbeŋɔŋlɔ', 'el': 'Ντεβαναγκάρι', 'el-polyton': 'Ντεβαναγκάρι', 'en': 'Devanagari', 'en-Dsrt': '𐐔𐐩𐑂𐐲𐑌𐐪𐑀𐐲𐑉𐐨', 'es': 'devanagari', 'et': 'devanaagari', 'eu': 'devanagaria', 'fa': 'دوناگری', 'ff-Adlm': '𞤁𞤫𞤾𞤢𞤲𞤢𞤺𞤢𞤪𞤭', 'fi': 'devanagari', 'fil': 'Devanagari', 'fo': 'devanagari', 'fr': 'dévanagari', 'fur': 'devanagari', 'fy': 'Devanagari', 'ga': 'Déiveanágrach', 'gd': 'Devanagari', 'gl': 'devanágari', 'gsw': 'Tövanagaari', 'gu': 'દેવનાગરી', 'ha': 'Devanagari', 'he': 'דוואנגרי', 'hi': 'देवनागरी', 'hi-Latn': 'Devanagari', 'hr': 'devangari pismo', 'hsb': 'devanagari', 'ht': 'dévanagari', 'hu': 'Devanagári', 'hy': 'դեւանագարի', 'ia': 'devanagari', 'id': 'Dewanagari', 'ig': 'Mkpụrụ ọkwụ Devangarị', 'is': 'devanagari', 'it': 'devanagari', 'ja': 'デーバナーガリー文字', 'jv': 'Devanagari', 'ka': 'დევანაგარი', 'kea': 'devanagari', 'kgp': 'nevanỹgari', 'kk': 'деванагари жазуы', 'kk-Cyrl': 'деванагари жазуы', 'km': 'ដាវ៉ាន់ណាការិ', 'kn': 'ದೇವನಾಗರಿ', 'ko': '데바나가리 문자', 'kok': 'देवनागरी', 'kok-Deva': 'देवनागरी', 'kok-Latn': 'Devanagari', 'ks': 'دیوناگری', 'ks-Arab': 'دیوناگری', 'ks-Deva': 'देवनागरी', 'ksh': 'indesche Devanajari-Schreff', 'ku': 'devanagarî', 'kxv': 'devnagrī', 'kxv-Deva': 'देवनागरी', 'kxv-Latn': 'devnagrī', 'kxv-Orya': 'ଦେୱନାଗରୀ', 'kxv-Telu': 'దేవనాగరి', 'ky': 'Деванагари', 'lb': 'Devanagari', 'lo': 'ດີວານາກາຣີ', 'lrc': 'دیڤانگأری', 'lt': 'devanagari', 'lv': 'dēvanāgari', 'mai': 'देवनागरी', 'mi': 'Tewhangāngari', 'mk': 'деванагари', 'ml': 'ദേവനാഗരി', 'mn': 'деванагари', 'mr': 'देवनागरी', 'ms': 'Devanagari', 'my': 'ဒီဗနာဂရီ', 'mzn': 'دیوانانگری', 'nb': 'devanagari', 'ne': 'देवानागरी', 'nl': 'Devanagari', 'nn': 'devanagari', 'no': 'devanagari', 'nqo': 'ߘߋߝ߭ߣߊߜ߭ߊߙߌ߫', 'or': 'ଦେବନାଗରୀ', 'pa': 'ਦੇਵਨਾਗਰੀ', 'pa-Guru': 'ਦੇਵਨਾਗਰੀ', 'pcm': 'Dẹvanágári', 'pl': 'dewanagari', 'ps': 'دیواناګري', 'pt': 'devanágari', 'qu': 'Devanagari', 'raj': 'देवनागरी', 'rm': 'devanagari', 'ro': 'devanagari', 'ru': 'деванагари', 'sat': 'ᱫᱮᱣᱟᱱᱟᱜᱟᱨᱤ', 'sat-Deva': 'देवानागारी', 'sat-Olck': 'ᱫᱮᱣᱟᱱᱟᱜᱟᱨᱤ', 'sc': 'devanagari', 'sd': 'ديوناگري', 'sd-Arab': 'ديوناگري', 'sd-Deva': 'देवनागिरी', 'si': 'දේවනාගරී', 'sk': 'dévanágarí', 'sl': 'devanagarščica', 'so': 'Dhefangaari', 'sq': 'devanagar', 'sr': 'деванагари', 'sr-Cyrl': 'деванагари', 'sr-Latn': 'devanagari', 'sv': 'devanagari', 'sw': 'Kidevanagari', 'ta': 'தேவநாகரி', 'te': 'దేవనాగరి', 'tg': 'Деванагарӣ', 'th': 'เทวนาครี', 'ti': 'ዴቫንጋሪ', 'tk': 'Dewanagari elipbiýi', 'to': 'tohinima fakaʻinitia-tevanākalī', 'tr': 'Devanagari', 'ug': 'دېۋاناگارى', 'uk': 'деванагарі', 'ur': 'دیوناگری', 'uz': 'devanagari', 'uz-Cyrl': 'Девангари', 'uz-Latn': 'devanagari', 'vec': 'devanàgari', 'vi': 'Chữ Devanagari', 'wae': 'Devanagari', 'xnr': 'देवनागरी', 'yi': 'דעוואַנאַגאַרי', 'yo': 'Dẹfanagárì', 'yrl': 'dewanagari', 'yue': '天城文', 'yue-Hans': '天城文', 'yue-Hant': '天城文', 'zh': '天城文', 'zh-Hans': '天城文', 'zh-Hant': '天城文', 'zu': 'isi-Devanagari'},
    'Diak': {'bal-Latn': 'Diwes Akuru', 'bs': 'dives akuru pismo', 'bs-Cyrl': 'дивес акуру писмо', 'bs-Latn': 'dives akuru pismo', 'en': 'Dives Akuru', 'et': 'divehi', 'eu': 'dives akuru', 'ff-Adlm': '𞤁𞤭𞤾𞤫𞥅𞤧 𞤀𞤳𞤵𞤪𞤵', 'fi': 'dives akuru', 'fo': 'dives akuru', 'gd': 'Dives Akuru', 'he': 'אותיות אי', 'hi-Latn': 'Dives Akuru', 'id': 'Dives Akuru', 'ig': 'Dives Akuru', 'kk': 'дивехи акуру жазуы', 'kk-Cyrl': 'дивехи акуру жазуы', 'mn': 'Диак', 'nl': 'Dives Akuru', 'nqo': 'ߊߞߎ߯ߙߎ', 'ro': 'dives akuru', 'sc': 'dives akuru', 'so': 'Luuqadaha Dives Akuru', 'sq': 'divesakuruisht', 'sv': 'dives akuru', 'tg': 'Дивс Акуру', 'vec': 'dives akuru', 'zh': '迪维西阿库鲁文', 'zh-Hans': '迪维西阿库鲁文', 'zu': 'isi-Dives Akuru'},
    'Dogr': {'bal-Latn': 'Dogra', 'bs': 'dogra pismo', 'bs-Cyrl': 'догра писмо', 'bs-Latn': 'dogra pismo', 'en': 'Dogra', 'eu': 'dogrera', 'ff-Adlm': '𞤁𞤮𞤺𞤪𞤢', 'fi': 'dogri', 'fo': 'dogra', 'gd': 'Dogra', 'he': 'דוגרה', 'hi-Latn': 'Dogra', 'id': 'Dogra', 'ig': 'Dogra', 'kk': 'догри жазуы', 'kk-Cyrl': 'догри жазуы', 'mn': 'Догра', 'ms': 'Dogra', 'nl': 'Dogra', 'nqo': 'ߘߏߜ߭ߑߙߌ', 'ro': 'dogra', 'sc': 'dogra', 'so': 'Dadka Dogra', 'sq': 'dograisht', 'sv': 'dogriska', 'tg': 'Догра', 'vec': 'dogri', 'zh': '多格拉文', 'zh-Hans': '多格拉文', 'zu': 'isi-Dogra'},
    'Dsrt': {'am': 'ዴዘረት', 'ar': 'الديسيريت', 'ast': 'alfabetu Deseret', 'az': 'deseret', 'az-Latn': 'deseret', 'bal-Latn': 'Deseret', 'bg': 'Дезерет', 'bn': 'দেসেরাত', 'brx': 'देसेरट्', 'bs': 'dezeret pismo', 'bs-Cyrl': 'Дезерет', 'bs-Latn': 'dezeret pismo', 'ca': 'deseret', 'ca-ES-valencia': 'deseret', 'ccp': '𑄘𑄬𑄥𑄬𑄢𑄖𑄴', 'cs': 'deseret', 'da': 'deseret', 'de': 'Deseret', 'el': 'Ντεσερέ', 'el-polyton': 'Ντεσερέ', 'en': 'Deseret', 'en-Dsrt': '𐐔𐐯𐑆𐐲𐑉𐐯𐐻', 'es': 'deseret', 'et': 'desereti', 'eu': 'deseret', 'fa': 'دیسرتی', 'ff-Adlm': '𞤁𞤫𞤧𞤫𞤪𞤫𞤼', 'fi': 'deseret', 'fo': 'deseret', 'fr': 'déséret', 'fy': 'Deseret', 'ga': 'Deseret', 'gd': 'Deseret', 'gsw': 'Teseret', 'gu': 'ડેસરેટ', 'he': 'דסרט', 'hi': 'डेसरेट', 'hi-Latn': 'Deseret', 'hr': 'deseret pismo', 'ht': 'déséret', 'hu': 'Deseret', 'id': 'Deseret', 'ig': 'Deseret', 'it': 'deseret', 'ja': 'デセレット文字', 'ka': 'დეზერეტის', 'kgp': 'nesereti', 'kk': 'дезерет жазуы', 'kk-Cyrl': 'дезерет жазуы', 'kn': 'ಡಸರ್ಟ್', 'ko': '디저렛 문자', 'ks': 'ڈیسٔریٹ', 'ks-Arab': 'ڈیسٔریٹ', 'lb': 'Deseret', 'lo': 'ເດເຊເຮ', 'lt': 'deseretas', 'mk': 'дезеретско', 'ml': 'ഡെസെർട്ട്', 'mn': 'Дезерет', 'mr': 'डेसर्ट', 'ms': 'Deseret', 'nb': 'deseret', 'ne': 'डेसेरेट', 'nl': 'Deseret', 'nn': 'deseret', 'no': 'deseret', 'nqo': 'ߘߖ߭ߋߙߋ', 'or': 'ଡେସର୍ଟ', 'pl': 'deseret', 'pt': 'deseret', 'rm': 'deseret', 'ro': 'mormonă', 'ru': 'дезерет', 'sc': 'deseret', 'sl': 'fonetska pisava deseret', 'so': 'Gobalka Deseret', 'sq': 'deseretisht', 'sr': 'Дезерет', 'sr-Cyrl': 'Дезерет', 'sr-Latn': 'Dezeret', 'sv': 'deseret', 'ta': 'டெசராட்', 'te': 'డేసెరెట్', 'tg': 'Дезерет', 'th': 'เดเซเรท', 'to': 'tohinima fakateseleti', 'tr': 'Deseret', 'ug': 'دېزېرېت', 'uk': 'дезерет', 'vec': 'deseret', 'vi': 'Chữ Deseret', 'xnr': 'डेसरेट', 'yrl': 'desereti', 'yue': '德瑟雷特文', 'yue-Hans': '德瑟雷特文', 'yue-Hant': '德瑟雷特文', 'zh': '德塞莱特文', 'zh-Hans': '德塞莱特文', 'zh-Hant': '德瑟雷特文', 'zu': 'isi-Deseret'},
    'Dupl': {'ast': 'taquigrafía Duployé', 'bal-Latn': 'Duplui Shárthand', 'br': 'berrskriverezh Duployé', 'bs': 'duploaje stenografija', 'bs-Cyrl': 'дуплојанska stenografija', 'bs-Latn': 'duploaje stenografija', 'ca': 'taquigrafia Duployé', 'ca-ES-valencia': 'taquigrafia Duployé', 'cs': 'Duployého těsnopis', 'da': 'Duploya-stenografi', 'de': 'Duployanisch', 'en': 'Duployan shorthand', 'et': 'Duployé kiirkiri', 'eu': 'duployiar takigrafia', 'ff-Adlm': '𞤁𞤵𞤨𞤤𞤮𞤴𞤢𞤲 𞤅𞤮𞥅𞤪𞤼𞤤𞤢𞤲𞤣', 'fi': 'Duployén pikakirjoitus', 'fy': 'Duployan snelschrift', 'ga': 'Gearrscríobh Duployan', 'gd': 'Gearr-sgrìobhadh Duployé', 'he': 'קיצור של דובילרן', 'hi-Latn': 'Duployan shorthand', 'id': 'Stenografi Duployan', 'ig': 'Duployan shorthand', 'it': 'stenografia duployan', 'ja': 'デュプロワエ式速記', 'ka': 'დუპლოის სტენოგრაფია', 'kk': 'дюплойе жазуы', 'kk-Cyrl': 'дюплойе жазуы', 'ko': '듀플로이안 문자', 'lo': 'ຊົວເລດັບໂລຍັນ', 'lt': 'Duplojė stenografija', 'mk': 'Диплојеево стенографско', 'mn': 'Дуплоян', 'ms': 'Trengkas Duployan', 'nb': 'duployan stenografi', 'nl': 'Duployan snelschrift', 'nn': 'duployan stenografi', 'no': 'duployan stenografi', 'nqo': 'ߛߑߕߋߣߏߜ߭ߙߊߝߌ ߘߎߔߑߟߏߦߋ', 'ro': 'stenografie duployană', 'ru': 'дуплоянская скоропись', 'sc': 'istenografia duployan', 'so': 'Qoraalka Duployan shorthand', 'sq': 'duplojanisht - formë e shkurtër', 'sv': 'Duployéstenografiska', 'tg': 'Стенографияи Дуплоянӣ', 'th': 'ชวเลขดัปโลยัน', 'to': 'tohinimanounou fakatupoloiē', 'tr': 'Duployé Stenografi', 'ug': 'دۇپلويان تېز خاتىرىلەش', 'vec': 'duployan stenografà', 'vi': 'Chữ tốc ký Duployan', 'yue': '杜普洛伊速記', 'yue-Hans': '杜普洛伊速记', 'yue-Hant': '杜普洛伊速記', 'zh': '杜普洛伊速记', 'zh-Hans': '杜普洛伊速记', 'zh-Hant': '杜普洛伊速記', 'zu': 'isi-Duployan shorthand'},
    'EA': {'af': 'Ceuta en Melilla', 'ak': 'Ceuta ne Melilla', 'am': 'ሴኡታና ሜሊላ', 'ar': 'سيوتا وميليلا', 'as': 'চেউটা আৰু মেলিলা', 'ast': 'Ceuta y Melilla', 'az': 'Seuta və Melilya', 'az-Cyrl': 'Сеута вә Мелилја', 'az-Latn': 'Seuta və Melilya', 'bal-Latn': 'Siótó o Melilá', 'be': 'Сеута і Мелілья', 'be-tarask': 'Сеута і Мелілья', 'bg': 'Сеута и Мелия', 'blo': 'Seyuta na Meliliya', 'bn': 'কুউটা এবং মেলিলা', 'br': 'Ceuta ha Melilla', 'brx': 'किउता आरो मेलिल्ला', 'bs': 'Seuta i Melilja', 'bs-Cyrl': 'Сеута и Мелиља', 'bs-Latn': 'Seuta i Melilja', 'ca': 'Ceuta i Melilla', 'ca-ES-valencia': 'Ceuta i Melilla', 'ccp': '𑄇𑄪𑄃𑄪𑄑 𑄃𑄳𑄃 𑄟𑄬𑄣𑄨𑄣', 'ce': 'Сеута а, Мелилья а', 'ceb': 'Ceuta & Melilla', 'chr': 'ᏑᏔ ᎠᎴ ᎺᎵᏯ', 'ckb': 'سێئووتا و مێلییا', 'cs': 'Ceuta a Melilla', 'cv': 'Сеута тата Мелилья', 'cy': 'Ceuta a Melilla', 'da': 'Ceuta og Melilla', 'de': 'Ceuta und Melilla', 'dsb': 'Ceuta a Melilla', 'dz': 'སེ་ཨུ་ཏ་ ཨེནཌ་ མེལ་ལི་ལ', 'ee': 'Keuta and Melilla nutome', 'el': 'Θέουτα και Μελίγια', 'el-polyton': 'Θέουτα και Μελίγια', 'en': 'Ceuta & Melilla', 'es': 'Ceuta y Melilla', 'et': 'Ceuta ja Melilla', 'eu': 'Ceuta eta Melilla', 'fa': 'سبته و ملیله', 'ff-Adlm': '𞤅𞤭𞤼𞥆𞤢 & 𞤃𞤫𞤤𞤭𞤤𞤢', 'fi': 'Ceuta ja Melilla', 'fil': 'Ceuta & Melilla', 'fo': 'Ceuta & Melilla', 'fr': 'Ceuta et Melilla', 'fur': 'Ceuta e Melilla', 'fy': 'Ceuta en Melilla', 'ga': 'Ceuta agus Melilla', 'gd': 'Ceuta agus Melilla', 'gl': 'Ceuta e Melilla', 'gu': 'સ્યુટા અને મેલિલા', 'ha': 'Ceuta da Melilla', 'he': 'סאוטה ומלייה', 'hi': 'सेउटा और मेलिला', 'hi-Latn': 'Ceuta & Melilla', 'hr': 'Ceuta i Melilla', 'hsb': 'Ceuta a Melilla', 'ht': 'Ceuta et Melilla', 'hu': 'Ceuta és Melilla', 'hy': 'Սեուտա և Մելիլյա', 'ia': 'Ceuta e Melilla', 'id': 'Ceuta dan Melilla', 'ig': 'Ceuta & Melilla', 'is': 'Ceuta og Melilla', 'it': 'Ceuta e Melilla', 'ja': 'セウタ・メリリャ', 'jv': 'Séuta lan Melila', 'ka': 'სეუტა და მელილა', 'kaa': 'Сеута ҳәм Мелилья', 'kaa-Cyrl': 'Сеута ҳәм Мелилья', 'kea': 'Seuta i Melilha', 'kgp': 'Sevuta kar Mẽrira', 'kk': 'Сеута және Мелилья', 'kk-Cyrl': 'Сеута және Мелилья', 'km': 'ជឺតា និង\u200bម៉េលីឡា', 'kn': 'ಸೆಯುಟಾ ಹಾಗೂ ಮೆಲಿಲ್ಲಾ', 'ko': '세우타 및 멜리야', 'kok': 'सिटा आनी मेलिल्ला', 'kok-Deva': 'सिटा आनी मेलिल्ला', 'ks': 'سیوٹا تٕہ مٔلیلا', 'ks-Arab': 'سیوٹا تٕہ مٔلیلا', 'ksh': 'Ze’uta un Meliija', 'ku': 'Ceuta û Melîla', 'kxv': 'seuṭa aḍe melila', 'kxv-Deva': 'सेउटा ऑड़े मेलिला', 'kxv-Latn': 'seuṭa aḍe melila', 'kxv-Orya': 'ସେଅଟା ଅଡ଼େ ମେଲିଲା', 'kxv-Telu': 'స్యూటా & మెలిల్లా', 'ky': 'Сеута жана Мелилла', 'lb': 'Ceuta a Melilla', 'lld': 'Ceuta y Melilla', 'lo': 'ເຊວຕາ & ເມລິນລາ', 'lt': 'Seuta ir Melila', 'lv': 'Seūta un Melilja', 'mai': 'सेउटा आ मेलिला', 'mi': 'Hūta me Merera', 'mk': 'Сеута и Мелиља', 'ml': 'സെയൂത്ത ആൻഡ് മെലിയ', 'mn': 'Сеута ба Мелилья', 'mr': 'स्यूटा आणि मेलिला', 'ms': 'Ceuta dan Melilla', 'mt': 'Ceuta u Melilla', 'my': 'ဆယ်ဥတာနှင့်မယ်လီလ်လာ', 'mzn': 'سوتا و ملیله', 'nb': 'Ceuta og Melilla', 'ne': 'सिउटा र मेलिला', 'nl': 'Ceuta en Melilla', 'nn': 'Ceuta og Melilla', 'no': 'Ceuta og Melilla', 'nqo': 'ߛߋߎߕߊ߫ ߣߌ߫ ߡߋߟߌߣߊ߫', 'om': 'Kewuta fi Mililaa', 'or': 'ସିଉଟା ଏବଂ ମେଲିଲା', 'pa': 'ਸਿਓਟਾ ਅਤੇ ਮੇਲਿੱਲਾ', 'pa-Guru': 'ਸਿਓਟਾ ਅਤੇ ਮੇਲਿੱਲਾ', 'pcm': 'Sẹúta & Mẹ́líla', 'pl': 'Ceuta i Melilla', 'ps': 'سيوتا او ماليلا', 'pt': 'Ceuta e Melilla', 'qu': 'Ceuta y Melilla', 'rm': 'Ceuta e Melilla', 'ro': 'Ceuta și Melilla', 'ru': 'Сеута и Мелилья', 'sc': 'Ceuta e Melilla', 'sd': 'سیوٽا ۽ میلیلا', 'sd-Arab': 'سیوٽا ۽ میلیلا', 'se': 'Ceuta ja Melilla', 'si': 'සෙයුටා සහ මෙලිල්ලා', 'sk': 'Ceuta a Melilla', 'sl': 'Ceuta in Melilla', 'smn': 'Ceuta já Melilla', 'so': 'Seyuta & Meliila', 'sq': 'Theuta e Melila', 'sr': 'Сеута и Мелиља', 'sr-Cyrl': 'Сеута и Мелиља', 'sr-Latn': 'Seuta i Melilja', 'sv': 'Ceuta och Melilla', 'sw': 'Ceuta na Melilla', 'syr': 'ܟܘܝܛܐ ܘܡܝܠܝܐ', 'ta': 'சியூடா & மெலில்லா', 'te': 'స్యూటా & మెలిల్లా', 'tg': 'Сеута ва Мелилла', 'th': 'เซวตาและเมลียา', 'ti': 'ሴውታን መሊላን', 'tk': 'Seuta we Melilýa', 'to': 'Siuta mo Melila', 'tr': 'Ceuta ve Melilla', 'tt': 'Сеута һәм Мелилья', 'ug': 'سېيتا ۋە مېلىلا', 'uk': 'Сеута і Мелілья', 'ur': 'سیئوٹا اور میلیلا', 'uz': 'Seuta va Melilya', 'uz-Cyrl': 'Сэута ва Мелилла', 'uz-Latn': 'Seuta va Melilya', 'vai': 'ꗻꕚ ꗪ ꔡꔷꕞ', 'vai-Vaii': 'ꗻꕚ ꗪ ꔡꔷꕞ', 'vec': 'Ceuta e Melila', 'vi': 'Ceuta và Melilla', 'wae': 'Ceuta und Melilla', 'wo': 'Ceuta & Melilla', 'xh': 'ECeuta & Melilla', 'xnr': 'सेउटा कने मेलिला', 'yo': 'Seuta àti Melilla', 'yrl': 'Seuta asuí Merira', 'yue': '休達與梅利利亞', 'yue-Hans': '休达与梅利利亚', 'yue-Hant': '休達與梅利利亞', 'zh': '休达及梅利利亚', 'zh-Hans': '休达及梅利利亚', 'zh-Hant': '休達與梅利利亞', 'zu': 'i-Cueta ne-Melilla'},
    'EC': {'af': 'Ecuador', 'agq': 'Ekwadò', 'ak': 'Yikuwedɔ', 'am': 'ኢኳዶር', 'ar': 'الإكوادور', 'as': 'ইকুৱেডৰ', 'asa': 'Ekwado', 'ast': 'Ecuador', 'az': 'Ekvador', 'az-Cyrl': 'Еквадор', 'az-Latn': 'Ekvador', 'bal-Latn': 'Ekwádór', 'bas': 'Èkwàtorìà', 'be': 'Эквадор', 'be-tarask': 'Эквадор', 'bez': 'Huekwado', 'bg': 'Еквадор', 'bgn': 'اکوادور', 'blo': 'Ekuwaɖɔɔr', 'bm': 'Ekwatɔr', 'bn': 'ইকুয়েডর', 'br': 'Ecuador', 'brx': 'एक्वाड’र', 'bs': 'Ekvador', 'bs-Cyrl': 'Еквадор', 'bs-Latn': 'Ekvador', 'ca': 'Equador', 'ca-ES-valencia': 'Equador', 'ccp': '𑄄𑄇𑄪𑄠𑄬𑄓𑄧𑄢𑄴', 'ce': 'Эквадор', 'ceb': 'Ecuador', 'cgg': 'Ikweda', 'chr': 'ᎡᏆᏙᎵ', 'ckb': 'ئیکوادۆر', 'cs': 'Ekvádor', 'cv': 'Эквадор', 'cy': 'Ecuador', 'da': 'Ecuador', 'dav': 'Ekwado', 'de': 'Ecuador', 'dje': 'Ekwateer', 'dsb': 'Ekwador', 'dyo': 'Ekuador', 'dz': 'ཨེ་ཁྭ་ཌོར', 'ebu': 'Ekwado', 'ee': 'Ekuadɔ nutome', 'el': 'Ισημερινός', 'el-polyton': 'Ἰσημερινός', 'en': 'Ecuador', 'en-Dsrt': '𐐇𐐿𐐶𐐲𐐼𐐱𐑉', 'eo': 'Ekvadoro', 'es': 'Ecuador', 'et': 'Ecuador', 'eu': 'Ekuador', 'ewo': 'Ekwatór', 'fa': 'اکوادور', 'ff': 'Ekuwatoor', 'ff-Adlm': '𞤉𞤳𞤵𞤱𞤢𞤣𞤮𞥅𞤪', 'ff-Latn': 'Ekuwatoor', 'fi': 'Ecuador', 'fil': 'Ecuador', 'fo': 'Ekvador', 'fr': 'Équateur', 'fur': 'Ecuador', 'fy': 'Ecuador', 'ga': 'Eacuadór', 'gd': 'Eacuador', 'gl': 'Ecuador', 'gsw': 'Ecuador', 'gu': 'એક્વાડોર', 'guz': 'Ekwado', 'ha': 'Ekwador', 'he': 'אקוודור', 'hi': 'इक्वाडोर', 'hi-Latn': 'Ecuador', 'hr': 'Ekvador', 'hsb': 'Ekwador', 'ht': 'Équateur', 'hu': 'Ecuador', 'hy': 'Էկվադոր', 'ia': 'Ecuador', 'id': 'Ekuador', 'ig': 'Ecuador', 'is': 'Ekvador', 'it': 'Ecuador', 'ja': 'エクアドル', 'jgo': 'Ɛkwandɔ̂', 'jmc': 'Ekwado', 'jv': 'Ékuadhor', 'ka': 'ეკვადორი', 'kaa': 'Эквадор', 'kaa-Cyrl': 'Эквадор', 'kab': 'Ikwaṭur', 'kam': 'Ekwado', 'kde': 'Ekwado', 'kea': 'Ekuador', 'kgp': 'Ekuvanor', 'khq': 'Ekwateer', 'ki': 'Ekwado', 'kk': 'Эквадор', 'kk-Cyrl': 'Эквадор', 'kln': 'Emetab Ecuador', 'km': 'អេក្វាទ័រ', 'kn': 'ಈಕ್ವೆಡಾರ್', 'ko': '에콰도르', 'kok': 'इक्वाडोर', 'kok-Deva': 'इक्वाडोर', 'ks': 'اِکواڑور', 'ks-Arab': 'اِکواڑور', 'ksb': 'Ekwado', 'ksf': 'ɛkwatɛǝ́', 'ksh': 'Äkwador', 'ku': 'Ekwador', 'kxv': 'ekvaḍor', 'kxv-Deva': 'इक्वाडोर', 'kxv-Latn': 'ekvaḍor', 'kxv-Orya': 'ଇକ୍ୱାଡର', 'kxv-Telu': 'ఈక్వడార్', 'ky': 'Эквадор', 'lag': 'Íkwado', 'lb': 'Ecuador', 'lg': 'Ekwado', 'lld': 'Ecuador', 'ln': 'Ekwatɛ́lɛ', 'lo': 'ເອກວາດໍ', 'lt': 'Ekvadoras', 'lu': 'Ekwatele', 'luo': 'Ecuador', 'luy': 'Ekwado', 'lv': 'Ekvadora', 'mai': 'इक्वाडोर', 'mas': 'Ekwado', 'mer': 'Ekwado', 'mfe': 'Ekwater', 'mg': 'Ekoatera', 'mi': 'Ekuatoa', 'mk': 'Еквадор', 'ml': 'ഇക്വഡോർ', 'mn': 'Эквадор', 'mr': 'इक्वाडोर', 'ms': 'Ecuador', 'mt': 'l-Ekwador', 'mua': 'Ekwatǝr', 'my': 'အီကွေဒေါ', 'mzn': 'اکوادر', 'naq': 'Ecuadori', 'nb': 'Ecuador', 'nd': 'Ecuador', 'ne': 'इक्वेडोर', 'nl': 'Ecuador', 'nmg': 'Ekuateur', 'nn': 'Ecuador', 'no': 'Ecuador', 'nqo': 'ߕߍߡߊߓߊ߲߮', 'nyn': 'Ikweda', 'om': 'Ekuwaador', 'or': 'ଇକ୍ୱେଡର୍\u200c', 'pa': 'ਇਕਵੇਡੋਰ', 'pa-Guru': 'ਇਕਵੇਡੋਰ', 'pcm': 'Ẹ́kwuádọ', 'pl': 'Ekwador', 'ps': 'اکوادور', 'pt': 'Equador', 'qu': 'Ecuador', 'rm': 'Ecuador', 'rn': 'Ekwateri', 'ro': 'Ecuador', 'rof': 'Ekwado', 'ru': 'Эквадор', 'rwk': 'Ekwado', 'saq': 'Ekwado', 'sbp': 'Ekwado', 'sc': 'Ècuador', 'sd': 'ايڪواڊور', 'sd-Arab': 'ايڪواڊور', 'se': 'Ecuador', 'seh': 'Equador', 'ses': 'Ekwateer', 'sg': 'Ekuatëre', 'shi': 'ⵉⴽⵡⴰⴷⵓⵔ', 'shi-Latn': 'ikwadur', 'shi-Tfng': 'ⵉⴽⵡⴰⴷⵓⵔ', 'si': 'ඉක්වදෝරය', 'sk': 'Ekvádor', 'sl': 'Ekvador', 'smn': 'Ecuador', 'sn': 'Ecuador', 'so': 'Ikuwadoor', 'sq': 'Ekuador', 'sr': 'Еквадор', 'sr-Cyrl': 'Еквадор', 'sr-Latn': 'Ekvador', 'sv': 'Ecuador', 'sw': 'Ecuador', 'syr': 'ܐܩܘܐܕܘܪ', 'ta': 'ஈக்வடார்', 'te': 'ఈక్వడార్', 'teo': 'Ekwado', 'tg': 'Эквадор', 'th': 'เอกวาดอร์', 'ti': 'ኤኳዶር', 'tk': 'Ekwador', 'to': 'ʻEkuetoa', 'tr': 'Ekvador', 'tt': 'Эквадор', 'twq': 'Ekwateer', 'tzm': 'Ikwaḍur', 'ug': 'ئېكۋاتور', 'uk': 'Еквадор', 'ur': 'ایکواڈور', 'uz': 'Ekvador', 'uz-Cyrl': 'Эквадор', 'uz-Latn': 'Ekvador', 'vai': 'ꗡꖴꔃꗍ', 'vai-Latn': 'Ɛ́kúwédɔ', 'vai-Vaii': 'ꗡꖴꔃꗍ', 'vec': 'Ècuador', 'vi': 'Ecuador', 'vun': 'Ekwado', 'wae': 'Ecuador', 'wo': 'Ekwaatër', 'xh': 'EEcuador', 'xnr': 'इक्वाडोर', 'xog': 'Ekwado', 'yav': 'ekuatɛ́l', 'yi': 'עקוואַדאר', 'yo': 'Ekuádò', 'yrl': 'Ekuadú', 'yue': '厄瓜多', 'yue-Hans': '厄瓜多', 'yue-Hant': '厄瓜多', 'zgh': 'ⵉⴽⵡⴰⴷⵓⵔ', 'zh': '厄瓜多尔', 'zh-Hans': '厄瓜多尔', 'zh-Hant': '厄瓜多', 'zu': 'i-Ecuador'},
    'EE': {'af': 'Estland', 'agq': 'Èsɨ̀tonyìa', 'ak': 'Ɛstonia', 'am': 'ኤስቶኒያ', 'ar': 'إستونيا', 'as': 'ইষ্টোনিয়া', 'asa': 'Ethtonia', 'ast': 'Estonia', 'az': 'Estoniya', 'az-Cyrl': 'Естонија', 'az-Latn': 'Estoniya', 'bal-Latn': 'Estóniá', 'bas': 'Èstonìà', 'be': 'Эстонія', 'be-tarask': 'Эстонія', 'bez': 'Huestonia', 'bg': 'Естония', 'blo': 'Ɛstoniya', 'bm': 'Esetoni', 'bn': 'এস্তোনিয়া', 'br': 'Estonia', 'brx': 'एस्ट’निया', 'bs': 'Estonija', 'bs-Cyrl': 'Естонија', 'bs-Latn': 'Estonija', 'ca': 'Estònia', 'ca-ES-valencia': 'Estònia', 'ccp': '𑄆𑄌𑄴𑄖𑄮𑄚𑄨𑄠', 'ce': 'Эстони', 'ceb': 'Estonia', 'cgg': 'Esitoniya', 'chr': 'ᎡᏍᏙᏂᏯ', 'ckb': 'ئیستۆنیا', 'cs': 'Estonsko', 'cv': 'Эстони', 'cy': 'Estonia', 'da': 'Estland', 'dav': 'Estonia', 'de': 'Estland', 'dje': 'Estooni', 'dsb': 'Estniska', 'dyo': 'Estoni', 'dz': 'ཨེས་ཊོ་ནི་ཡ', 'ebu': 'Estonia', 'ee': 'Estonia nutome', 'el': 'Εσθονία', 'el-polyton': 'Ἐσθονία', 'en': 'Estonia', 'en-Dsrt': '𐐇𐑅𐐻𐐬𐑌𐐨𐐲', 'eo': 'Estonujo', 'es': 'Estonia', 'et': 'Eesti', 'eu': 'Estonia', 'ewo': 'Esetoní', 'fa': 'استونی', 'ff': 'Estoni', 'ff-Adlm': '𞤉𞤧𞤼𞤮𞤲𞤭𞤴𞤢𞥄', 'ff-Latn': 'Estoni', 'fi': 'Viro', 'fil': 'Estonia', 'fo': 'Estland', 'fr': 'Estonie', 'fur': 'Estonie', 'fy': 'Estlân', 'ga': 'an Eastóin', 'gd': 'An Eastoin', 'gl': 'Estonia', 'gsw': 'Eestland', 'gu': 'એસ્ટોનિયા', 'guz': 'Estonia', 'ha': 'Estoniya', 'he': 'אסטוניה', 'hi': 'एस्टोनिया', 'hi-Latn': 'Estonia', 'hr': 'Estonija', 'hsb': 'Estiska', 'ht': 'Estonie', 'hu': 'Észtország', 'hy': 'Էստոնիա', 'ia': 'Estonia', 'id': 'Estonia', 'ie': 'Estonia', 'ig': 'Estonia', 'is': 'Eistland', 'it': 'Estonia', 'ja': 'エストニア', 'jmc': 'Estonia', 'jv': 'Éstonia', 'ka': 'ესტონეთი', 'kaa': 'Эстония', 'kaa-Cyrl': 'Эстония', 'kab': 'Istunya', 'kam': 'Estonia', 'kde': 'Estonia', 'kea': 'Stónia', 'kgp': 'Enhtonĩja', 'khq': 'Estooni', 'ki': 'Estonia', 'kk': 'Эстония', 'kk-Cyrl': 'Эстония', 'kln': 'Emetab Estonia', 'km': 'អេស្តូនី', 'kn': 'ಎಸ್ಟೋನಿಯಾ', 'ko': '에스토니아', 'kok': 'एस्टोनिया', 'kok-Deva': 'एस्टोनिया', 'ks': 'ایسٹونِیا', 'ks-Arab': 'ایسٹونِیا', 'ksb': 'Estonia', 'ksf': 'ɛstoní', 'ksh': 'Äßland', 'ku': 'Estonya', 'kxv': 'esṭoniya', 'kxv-Deva': 'एस्टोनिया', 'kxv-Latn': 'esṭoniya', 'kxv-Orya': 'ଏସ୍ଟନିଆ', 'kxv-Telu': 'ఎస్టోనియా', 'ky': 'Эстония', 'lag': 'Estonía', 'lb': 'Estland', 'lg': 'Esitoniya', 'lld': 'Estonia', 'ln': 'Esitoni', 'lo': 'ເອສໂຕເນຍ', 'lt': 'Estija', 'lu': 'Esitoni', 'luo': 'Estonia', 'luy': 'Estonia', 'lv': 'Igaunija', 'mai': 'एस्टोनिया', 'mas': 'Estonia', 'mer': 'Estonia', 'mfe': 'Estoni', 'mg': 'Estonia', 'mi': 'Etōnia', 'mk': 'Естонија', 'ml': 'എസ്റ്റോണിയ\u200d', 'mn': 'Эстони', 'mr': 'एस्टोनिया', 'ms': 'Estonia', 'mt': 'l-Estonja', 'mua': 'Estoniya', 'my': 'အက်စတိုးနီးယား', 'mzn': 'استونی', 'naq': 'Estoniab', 'nb': 'Estland', 'nd': 'Estonia', 'ne': 'इस्टोनिया', 'nl': 'Estland', 'nmg': 'Ɛstonia', 'nn': 'Estland', 'no': 'Estland', 'nqo': 'ߋߛߑߕߏߣߌ߫', 'nyn': 'Esitoniya', 'om': 'Istooniyaa', 'or': 'ଏସ୍ତୋନିଆ', 'pa': 'ਇਸਟੋਨੀਆ', 'pa-Guru': 'ਇਸਟੋਨੀਆ', 'pcm': 'Ẹstónia', 'pl': 'Estonia', 'ps': 'استونیا', 'pt': 'Estônia', 'qu': 'Estonia', 'rm': 'Estonia', 'rn': 'Esitoniya', 'ro': 'Estonia', 'rof': 'Estonia', 'ru': 'Эстония', 'rwk': 'Estonia', 'sah': 'Эстония', 'saq': 'Estonia', 'sbp': 'Esitoniya', 'sc': 'Estònia', 'sd': 'ايسٽونيا', 'sd-Arab': 'ايسٽونيا', 'se': 'Estlánda', 'seh': 'Estônia', 'ses': 'Estooni', 'sg': 'Estonïi', 'shi': 'ⵉⵙⵜⵓⵏⵢⴰ', 'shi-Latn': 'istunya', 'shi-Tfng': 'ⵉⵙⵜⵓⵏⵢⴰ', 'si': 'එස්තෝනියාව', 'sk': 'Estónsko', 'sl': 'Estonija', 'smn': 'Eestieennâm', 'sn': 'Estonia', 'so': 'Estooniya', 'sq': 'Estoni', 'sr': 'Естонија', 'sr-Cyrl': 'Естонија', 'sr-Latn': 'Estonija', 'sv': 'Estland', 'sw': 'Estonia', 'syr': 'ܐܣܛܘܢܝܐ', 'ta': 'எஸ்டோனியா', 'te': 'ఎస్టోనియా', 'teo': 'Estonia', 'tg': 'Эстония', 'th': 'เอสโตเนีย', 'ti': 'ኤስቶንያ', 'tk': 'Estoniýa', 'to': 'ʻEsitōnia', 'tr': 'Estonya', 'tt': 'Эстония', 'twq': 'Estooni', 'tzm': 'Isṭunya', 'ug': 'ئېستونىيە', 'uk': 'Естонія', 'ur': 'اسٹونیا', 'uz': 'Estoniya', 'uz-Cyrl': 'Эстония', 'uz-Latn': 'Estoniya', 'vai': 'ꗡꔻꕿꕇꕰ', 'vai-Latn': 'Ɛsitóninya', 'vai-Vaii': 'ꗡꔻꕿꕇꕰ', 'vec': 'Estonia', 'vi': 'Estonia', 'vun': 'Estonia', 'wae': 'Eštland', 'wo': 'Estoni', 'xh': 'E-Estonia', 'xnr': 'एस्टोनिया', 'xog': 'Esitoniya', 'yav': 'ɛstoni', 'yi': 'עסטלאַנד', 'yo': 'Esitonia', 'yrl': 'Isituniya', 'yue': '愛沙尼亞', 'yue-Hans': '爱沙尼亚', 'yue-Hant': '愛沙尼亞', 'zgh': 'ⵉⵙⵜⵓⵏⵢⴰ', 'zh': '爱沙尼亚', 'zh-Hans': '爱沙尼亚', 'zh-Hant': '愛沙尼亞', 'zu': 'i-Estonia'},
    'EG': {'af': 'Egipte', 'agq': 'Edzì', 'ak': 'Misrim', 'am': 'ግብጽ', 'ar': 'مصر', 'as': 'ইজিপ্ত', 'asa': 'Mithri', 'ast': 'Exiptu', 'az': 'Misir', 'az-Cyrl': 'Мисир', 'az-Latn': 'Misir', 'bal-Latn': 'Mesr', 'bas': 'Ègîptò', 'be': 'Егіпет', 'be-tarask': 'Егіпет', 'bez': 'Humisri', 'bg': 'Египет', 'bgn': 'مصر', 'blo': 'Ejipti', 'bm': 'Eziputi', 'bn': 'মিশর', 'br': 'Egipt', 'brx': 'ईजिप्त', 'bs': 'Egipat', 'bs-Cyrl': 'Египат', 'bs-Latn': 'Egipat', 'ca': 'Egipte', 'ca-ES-valencia': 'Egipte', 'ccp': '𑄟𑄨𑄥𑄧𑄢𑄴', 'ce': 'Мисар', 'ceb': 'Egypt', 'cgg': 'Misiri', 'chr': 'ᎢᏥᏈᎢ', 'ckb': 'میسر', 'cs': 'Egypt', 'cv': 'Египет', 'cy': 'Yr Aifft', 'da': 'Egypten', 'dav': 'Misri', 'de': 'Ägypten', 'dje': 'Misra', 'dsb': 'Egyptojska', 'dyo': 'Esípt', 'dz': 'ཨི་ཇིབཊ', 'ebu': 'Misri', 'ee': 'Egypte nutome', 'el': 'Αίγυπτος', 'el-polyton': 'Αἴγυπτος', 'en': 'Egypt', 'en-Dsrt': '𐐀𐐾𐐲𐐹𐐻', 'eo': 'Egiptujo', 'es': 'Egipto', 'et': 'Egiptus', 'eu': 'Egipto', 'ewo': 'Ehíbətɛn', 'fa': 'مصر', 'ff': 'Ejipt', 'ff-Adlm': '𞤃𞤭𞤧𞤭𞤪𞤢', 'ff-Latn': 'Ejipt', 'fi': 'Egypti', 'fil': 'Egypt', 'fo': 'Egyptaland', 'fr': 'Égypte', 'fur': 'Egjit', 'fy': 'Egypte', 'ga': 'An Éigipt', 'gd': 'An Èipheit', 'gl': 'Exipto', 'gsw': 'Ägüpte', 'gu': 'ઇજિપ્ત', 'guz': 'Misri', 'ha': 'Misira', 'he': 'מצרים', 'hi': 'मिस्र', 'hi-Latn': 'Egypt', 'hr': 'Egipat', 'hsb': 'Egyptowska', 'ht': 'Égypte', 'hu': 'Egyiptom', 'hy': 'Եգիպտոս', 'ia': 'Egypto', 'id': 'Mesir', 'ig': 'Egypt', 'is': 'Egyptaland', 'it': 'Egitto', 'ja': 'エジプト', 'jgo': 'Ɛjíptɛ', 'jmc': 'Misri', 'jv': 'Mesir', 'ka': 'ეგვიპტე', 'kaa': 'Мысыр', 'kaa-Cyrl': 'Мысыр', 'kab': 'Maṣr', 'kam': 'Misili', 'kde': 'Misli', 'kea': 'Ejitu', 'kgp': 'Ejito', 'khq': 'Misra', 'ki': 'Misri', 'kk': 'Мысыр', 'kk-Cyrl': 'Мысыр', 'kln': 'Emetab Misiri', 'km': 'អេហ្ស៊ីប', 'kn': 'ಈಜಿಪ್ಟ್', 'ko': '이집트', 'kok': 'ईजिप्त', 'kok-Deva': 'ईजिप्त', 'kok-Latn': 'Ejipt', 'ks': 'مصر', 'ks-Arab': 'مصر', 'ksb': 'Misli', 'ksf': 'ɛjípt', 'ksh': 'Äjüpte', 'ku': 'Misir', 'kxv': 'ījipṭ', 'kxv-Deva': 'मिस्र', 'kxv-Latn': 'ījipṭ', 'kxv-Orya': 'ଇଜିପ୍ଟ', 'kxv-Telu': 'ఈజిప్ట్', 'ky': 'Египет', 'lag': 'Mísiri', 'lb': 'Egypten', 'lg': 'Misiri', 'lld': 'Egit', 'ln': 'Ezípite', 'lo': 'ອີຢິບ', 'lt': 'Egiptas', 'lu': 'Mushidi', 'luo': 'Egypt', 'luy': 'Misri', 'lv': 'Ēģipte', 'mai': 'मिस्र', 'mas': 'Misri', 'mer': 'Misiri', 'mfe': 'Lezipt', 'mg': 'Ejypta', 'mgh': 'Umisiri', 'mi': 'Īhipa', 'mk': 'Египет', 'ml': 'ഈജിപ്ത്', 'mn': 'Египет', 'mr': 'इजिप्त', 'ms': 'Mesir', 'mt': 'l-Eġittu', 'mua': 'Sǝr Egypt', 'my': 'အီဂျစ်', 'mzn': 'مصر', 'naq': 'Egipteb', 'nb': 'Egypt', 'nd': 'Egypt', 'ne': 'इजिप्ट', 'nl': 'Egypte', 'nmg': 'Ägyptɛn', 'nn': 'Egypt', 'no': 'Egypt', 'nqo': 'ߋߖ߭ߌߔߑߕ', 'nyn': 'Misiri', 'om': 'Missir', 'or': 'ଇଜିପ୍ଟ', 'pa': 'ਮਿਸਰ', 'pa-Guru': 'ਮਿਸਰ', 'pcm': 'Íjipt', 'pl': 'Egipt', 'ps': 'مصر', 'pt': 'Egito', 'qu': 'Egipto', 'rm': 'Egipta', 'rn': 'Misiri', 'ro': 'Egipt', 'rof': 'Misri', 'ru': 'Египет', 'rwk': 'Misri', 'saq': 'Misri', 'sbp': 'Misili', 'sc': 'Egitu', 'sd': 'مصر', 'sd-Arab': 'مصر', 'se': 'Egypta', 'seh': 'Egito', 'ses': 'Misra', 'sg': 'Kâmitâ', 'shi': 'ⵎⵉⵚⵕ', 'shi-Latn': 'miṣṛ', 'shi-Tfng': 'ⵎⵉⵚⵕ', 'si': 'ඊජිප්තුව', 'sk': 'Egypt', 'sl': 'Egipt', 'smn': 'Egypti', 'sn': 'Egypt', 'so': 'Masar', 'sq': 'Egjipt', 'sr': 'Египат', 'sr-Cyrl': 'Египат', 'sr-Latn': 'Egipat', 'sv': 'Egypten', 'sw': 'Misri', 'syr': 'ܡܨܪܝܢ', 'ta': 'எகிப்து', 'te': 'ఈజిప్ట్', 'teo': 'Misri', 'tg': 'Миср', 'th': 'อียิปต์', 'ti': 'ግብጺ', 'tk': 'Müsür', 'to': 'ʻIsipite', 'tr': 'Mısır', 'tt': 'Мисыр', 'twq': 'Misra', 'tzm': 'Miṣr', 'ug': 'مىسىر', 'uk': 'Єгипет', 'ur': 'مصر', 'uz': 'Misr', 'uz-Cyrl': 'Миср', 'uz-Latn': 'Misr', 'vai': 'ꕆꔖꕞ', 'vai-Latn': 'Míséla', 'vai-Vaii': 'ꕆꔖꕞ', 'vec': 'Ejito', 'vi': 'Ai Cập', 'vun': 'Misri', 'wae': 'Egypte', 'wo': 'Esipt', 'xh': 'IYiputa', 'xnr': 'मिस्र', 'xog': 'Misiri', 'yav': 'isípit', 'yi': 'עגיפּטן', 'yo': 'Égípítì', 'yrl': 'Egitu', 'yue': '埃及', 'yue-Hans': '埃及', 'yue-Hant': '埃及', 'zgh': 'ⵎⵉⵚⵕ', 'zh': '埃及', 'zh-Hans': '埃及', 'zh-Hant': '埃及', 'zu': 'i-Egypt'},
    'EH': {'af': 'Wes-Sahara', 'ak': 'Sahara Atɔeɛ', 'am': 'ምዕራባዊ ሳህራ', 'ar': 'الصحراء الغربية', 'as': 'পশ্চিমীয় ছাহাৰা', 'ast': 'Sáḥara Occidental', 'az': 'Qərbi Saxara', 'az-Latn': 'Qərbi Saxara', 'bal-Latn': 'Róbarkati Sahárá', 'be': 'Заходняя Сахара', 'be-tarask': 'Заходняя Сахара', 'bg': 'Западна Сахара', 'bgn': 'روچ\u200cکپتین سحرا', 'blo': 'Sarawii', 'bn': 'পশ্চিম সাহারা', 'br': 'Sahara ar Cʼhornôg', 'brx': 'सोनाबारि साहारा', 'bs': 'Zapadna Sahara', 'bs-Cyrl': 'Западна Сахара', 'bs-Latn': 'Zapadna Sahara', 'ca': 'Sàhara Occidental', 'ca-ES-valencia': 'Sàhara Occidental', 'ccp': '𑄛𑄧𑄎𑄨𑄟𑄴 𑄥𑄦𑄢', 'ce': 'Малхбузен Саьхьара', 'ceb': 'Western Sahara', 'chr': 'ᏭᏕᎵᎬ ᏗᏜ ᏌᎮᎳ', 'ckb': 'سەحرای ڕۆژاوا', 'cs': 'Západní Sahara', 'cv': 'Анӑҫ Сахара', 'cy': 'Gorllewin Sahara', 'da': 'Vestsahara', 'de': 'Westsahara', 'dsb': 'Pódwjacorna Sahara', 'dz': 'ནུབ་ཕྱོགས་ ས་ཧཱ་ར', 'ee': 'Ɣetoɖoƒe Sahara nutome', 'el': 'Δυτική Σαχάρα', 'el-polyton': 'Δυτικὴ Σαχάρα', 'en': 'Western Sahara', 'en-Dsrt': '𐐎𐐯𐑅𐐻𐐲𐑉𐑌 𐐝𐐲𐐸𐐱𐑉𐐲', 'eo': 'Okcidenta Saharo', 'es': 'Sáhara Occidental', 'et': 'Lääne-Sahara', 'eu': 'Mendebaldeko Sahara', 'fa': 'صحرای غربی', 'ff-Adlm': '𞤅𞤢𞥄𞤸𞤢𞤪𞤢 𞤖𞤭𞥅𞤲𞤢𞥄𞤪𞤭', 'fi': 'Länsi-Sahara', 'fil': 'Kanlurang Sahara', 'fo': 'Vestursahara', 'fr': 'Sahara occidental', 'fur': 'Sahara ocidentâl', 'fy': 'Westelijke Sahara', 'ga': 'An Sahára Thiar', 'gd': 'Sathara an Iar', 'gl': 'Sáhara Occidental', 'gsw': 'Weschtsahara', 'gu': 'પશ્ચિમી સહારા', 'ha': 'Yammacin Sahara', 'he': 'סהרה המערבית', 'hi': 'पश्चिमी सहारा', 'hi-Latn': 'Western Sahara', 'hr': 'Zapadna Sahara', 'hsb': 'Zapadna Sahara', 'ht': 'Sahara occidental', 'hu': 'Nyugat-Szahara', 'hy': 'Արևմտյան Սահարա', 'ia': 'Sahara occidental', 'id': 'Sahara Barat', 'ig': 'Ọdịda Anyanwụ Sahara', 'is': 'Vestur-Sahara', 'it': 'Sahara Occidentale', 'ja': '西サハラ', 'jv': 'Sahara Kulon', 'ka': 'დასავლეთ საჰარა', 'kaa': 'Батыс Сахара', 'kaa-Cyrl': 'Батыс Сахара', 'kea': 'Sara Osidental', 'kgp': 'Sahara Rãpur', 'kk': 'Батыс Сахара', 'kk-Cyrl': 'Батыс Сахара', 'km': 'សាហារ៉ាខាងលិច', 'kn': 'ಪಶ್ಚಿಮ ಸಹಾರಾ', 'ko': '서사하라', 'kok': 'अस्तंत सहारा', 'kok-Deva': 'अस्तंत सहारा', 'ks': 'مشرِقی سَہارا', 'ks-Arab': 'مشرِقی سَہارا', 'ksh': 'Wäß-Sahara', 'ku': 'Sahraya Rojava', 'kxv': 'vedakuṇpu sahara', 'kxv-Deva': 'वेड़ा कुण्पु सहारा', 'kxv-Latn': 'vedakuṇpu sahara', 'kxv-Orya': 'ୱେଡ଼ାକୁଣ୍ପୁ ସାହାରା', 'kxv-Telu': 'పడమటి సహారా', 'ky': 'Батыш Сахара', 'lb': 'Westsahara', 'lld': 'Sahara ozidentala', 'lo': 'ຊາຮາຣາຕາເວັນຕົກ', 'lt': 'Vakarų Sachara', 'lv': 'Rietumsahāra', 'mai': 'पश्चिमी सहारा', 'mi': 'Hahāra ki te Tonga', 'mk': 'Западна Сахара', 'ml': 'പശ്ചിമ സഹാറ', 'mn': 'Баруун Сахар', 'mr': 'पश्चिम सहारा', 'ms': 'Sahara Barat', 'mt': 'is-Saħara tal-Punent', 'my': 'အနောက် ဆာဟာရ', 'mzn': 'غربی صحرا', 'nb': 'Vest-Sahara', 'ne': 'पश्चिमी साहारा', 'nl': 'Westelijke Sahara', 'nn': 'Vest-Sahara', 'no': 'Vest-Sahara', 'nqo': 'ߞߌ߲߬ߢߍ߬ߞߏ߲ߞߏ߫ ߕߟߋ߬ߓߋ', 'om': 'Sahaaraa Dhihaa', 'or': 'ପଶ୍ଚିମ ସାହାରା', 'pa': 'ਪੱਛਮੀ ਸਹਾਰਾ', 'pa-Guru': 'ਪੱਛਮੀ ਸਹਾਰਾ', 'pcm': 'Wẹ́stán Sahára', 'pl': 'Sahara Zachodnia', 'ps': 'لويديځ صحارا', 'pt': 'Saara Ocidental', 'qu': 'Sahara Occidental', 'rm': 'Sahara Occidentala', 'ro': 'Sahara Occidentală', 'ru': 'Западная Сахара', 'sc': 'Sahara otzidentale', 'sd': 'اولهه صحارا', 'sd-Arab': 'اولهه صحارا', 'se': 'Oarje-Sahára', 'si': 'බටහිර සහරාව', 'sk': 'Západná Sahara', 'sl': 'Zahodna Sahara', 'so': 'Saxaraha Galbeed', 'sq': 'Saharaja Perëndimore', 'sr': 'Западна Сахара', 'sr-Cyrl': 'Западна Сахара', 'sr-Latn': 'Zapadna Sahara', 'sv': 'Västsahara', 'sw': 'Sahara Magharibi', 'syr': 'ܨܚܪܐ ܡܥܪܒܝܬܐ', 'ta': 'மேற்கு சஹாரா', 'te': 'పడమటి సహారా', 'tg': 'Сахараи Ғарбӣ', 'th': 'ซาฮาราตะวันตก', 'ti': 'ምዕራባዊ ሰሃራ', 'tk': 'Günbatar Sahara', 'to': 'Sahala fakahihifo', 'tr': 'Batı Sahra', 'tt': 'Көнбатыш Сахара', 'ug': 'غەربىي ساخارا', 'uk': 'Західна Сахара', 'ur': 'مغربی صحارا', 'uz': 'G‘arbiy Sahroi Kabir', 'uz-Cyrl': 'Ғарбий Саҳрои Кабир', 'uz-Latn': 'G‘arbiy Sahroi Kabir', 'vai': 'ꕢꕌꕟ ꔎꔒ ꕀꔤ', 'vai-Vaii': 'ꕢꕌꕟ ꔎꔒ ꕀꔤ', 'vec': 'Sahara Osidentale', 'vi': 'Tây Sahara', 'wae': 'Weštsahara', 'wo': 'Sahara bu sowwu', 'xh': 'EWestern Sahara', 'xnr': 'पश्चिमी सहारा', 'yo': 'Ìwọ̀òòrùn Sàhárà', 'yrl': 'Saara Usidẽtawara', 'yue': '西撒哈拉', 'yue-Hans': '西撒哈拉', 'yue-Hant': '西撒哈拉', 'zh': '西撒哈拉', 'zh-Hans': '西撒哈拉', 'zh-Hant': '西撒哈拉', 'zu': 'i-Western Sahara'},
    'ER': {'af': 'Eritrea', 'agq': 'Èletɨ̀là', 'ak': 'Ɛritrea', 'am': 'ኤርትራ', 'ar': 'إريتريا', 'as': 'এৰিত্ৰিয়া', 'asa': 'Eritrea', 'ast': 'Eritrea', 'az': 'Eritreya', 'az-Cyrl': 'Еритреја', 'az-Latn': 'Eritreya', 'bal-Latn': 'Eritiriá', 'bas': 'Èrìtrěà', 'be': 'Эрытрэя', 'be-tarask': 'Эрытрэя', 'bez': 'Hueritrea', 'bg': 'Еритрея', 'bgn': 'اریتره', 'blo': 'Eritree', 'bm': 'Eritere', 'bn': 'ইরিত্রিয়া', 'br': 'Eritrea', 'brx': 'एरिट्रिया', 'bs': 'Eritreja', 'bs-Cyrl': 'Еритреја', 'bs-Latn': 'Eritreja', 'ca': 'Eritrea', 'ca-ES-valencia': 'Eritrea', 'ccp': '𑄄𑄢𑄨𑄖𑄳𑄢𑄨𑄠', 'ce': 'Эритрей', 'ceb': 'Eritrea', 'cgg': 'Eriteriya', 'chr': 'ᎡᎵᏟᏯ', 'ckb': 'ئەریتریا', 'cs': 'Eritrea', 'cv': 'Эритрей', 'cy': 'Eritrea', 'da': 'Eritrea', 'dav': 'Eritrea', 'de': 'Eritrea', 'dje': 'Eritree', 'dsb': 'Eritreja', 'dyo': 'Eritree', 'dz': 'ཨེ་རི་ཊྲེ་ཡ', 'ebu': 'Eritrea', 'ee': 'Eritrea nutome', 'el': 'Ερυθραία', 'el-polyton': 'Ἐρυθραία', 'en': 'Eritrea', 'en-Dsrt': '𐐇𐑉𐐮𐐻𐑉𐐨𐐲', 'eo': 'Eritreo', 'es': 'Eritrea', 'et': 'Eritrea', 'eu': 'Eritrea', 'ewo': 'Elitəlé', 'fa': 'اریتره', 'ff': 'Eriteree', 'ff-Adlm': '𞤉𞤪𞤭𞥅𞤼𞤫𞤪𞤫', 'ff-Latn': 'Eriteree', 'fi': 'Eritrea', 'fil': 'Eritrea', 'fo': 'Eritrea', 'fr': 'Érythrée', 'fur': 'Eritree', 'fy': 'Eritrea', 'ga': 'an Eiritré', 'gd': 'Eartra', 'gl': 'Eritrea', 'gsw': 'Äritreea', 'gu': 'એરિટ્રિયા', 'guz': 'Eritrea', 'ha': 'Eritireya', 'he': 'אריתריאה', 'hi': 'इरिट्रिया', 'hi-Latn': 'Eritrea', 'hr': 'Eritreja', 'hsb': 'Eritreja', 'ht': 'Érythrée', 'hu': 'Eritrea', 'hy': 'Էրիթրեա', 'ia': 'Eritrea', 'id': 'Eritrea', 'ie': 'Eritrea', 'ig': 'Eritrea', 'is': 'Erítrea', 'it': 'Eritrea', 'ja': 'エリトリア', 'jgo': 'Ɛlitɛlɛ́ya', 'jmc': 'Eritrea', 'jv': 'Éritréa', 'ka': 'ერიტრეა', 'kaa': 'Эритрея', 'kaa-Cyrl': 'Эритрея', 'kab': 'Iritiria', 'kam': 'Eritrea', 'kde': 'Elitilea', 'kea': 'Iritreia', 'kgp': 'Erytiréja', 'khq': 'Eritree', 'ki': 'Eritrea', 'kk': 'Эритрея', 'kk-Cyrl': 'Эритрея', 'kln': 'Emetab Eritrea', 'km': 'អេរីត្រេ', 'kn': 'ಎರಿಟ್ರಿಯಾ', 'ko': '에리트리아', 'kok': 'इरिट्रिया', 'kok-Deva': 'इरिट्रिया', 'ks': 'اِرٕٹِیا', 'ks-Arab': 'اِرٕٹِیا', 'ksb': 'Elitlea', 'ksf': 'ɛritrɛ́', 'ksh': 'Äritreja', 'ku': 'Erître', 'kxv': 'īriṭriaa', 'kxv-Deva': 'इरिट्रिया', 'kxv-Latn': 'īriṭriaa', 'kxv-Orya': 'ଇରିଟ୍ରିଆ', 'kxv-Telu': 'ఇరిట్రియా', 'ky': 'Эритрея', 'lag': 'Eriterea', 'lb': 'Eritrea', 'lg': 'Eritureya', 'lld': 'Eritrea', 'ln': 'Elitelɛ', 'lo': 'ເອຣິເທຣຍ', 'lt': 'Eritrėja', 'lu': 'Elitele', 'luo': 'Eritrea', 'luy': 'Eritrea', 'lv': 'Eritreja', 'mai': 'एरिट्रिया', 'mas': 'Eritrea', 'mer': 'Eritrea', 'mfe': 'Erythre', 'mg': 'Eritrea', 'mgh': 'Uriterea', 'mi': 'Eritēria', 'mk': 'Еритреја', 'ml': 'എറിത്രിയ', 'mn': 'Эритрей', 'mr': 'एरिट्रिया', 'ms': 'Eritrea', 'mt': 'l-Eritrea', 'mua': 'Sǝr Eritre', 'my': 'အီရီထရီးယား', 'mzn': 'اریتره', 'naq': 'Eritreab', 'nb': 'Eritrea', 'nd': 'Eritrea', 'ne': 'एरिट्रीया', 'nl': 'Eritrea', 'nmg': 'Erytrea', 'nn': 'Eritrea', 'no': 'Eritrea', 'nqo': 'ߋߙߕߌߙߋ߫', 'nyn': 'Eriteriya', 'om': 'Eertiraa', 'or': 'ଇରିଟ୍ରିୟା', 'pa': 'ਇਰੀਟ੍ਰਿਆ', 'pa-Guru': 'ਇਰੀਟ੍ਰਿਆ', 'pcm': 'Ẹritrẹ́a', 'pl': 'Erytrea', 'ps': 'اریتره', 'pt': 'Eritreia', 'qu': 'Eritrea', 'rm': 'Eritrea', 'rn': 'Elitereya', 'ro': 'Eritreea', 'rof': 'Eritrea', 'ru': 'Эритрея', 'rwk': 'Eritrea', 'saq': 'Eritrea', 'sbp': 'Elitileya', 'sc': 'Eritrea', 'sd': 'ايريٽيريا', 'sd-Arab': 'ايريٽيريا', 'se': 'Eritrea', 'seh': 'Eritréia', 'ses': 'Eritree', 'sg': 'Eritrëe', 'shi': 'ⵉⵔⵉⵜⵉⵔⵢⴰ', 'shi-Latn': 'iritirya', 'shi-Tfng': 'ⵉⵔⵉⵜⵉⵔⵢⴰ', 'si': 'එරිත්\u200dරියාව', 'sk': 'Eritrea', 'sl': 'Eritreja', 'smn': 'Eritrea', 'sn': 'Eritrea', 'so': 'Eritreeya', 'sq': 'Eritre', 'sr': 'Еритреја', 'sr-Cyrl': 'Еритреја', 'sr-Latn': 'Eritreja', 'sv': 'Eritrea', 'sw': 'Eritrea', 'syr': 'ܐܪܬܪܝܐ', 'ta': 'எரிட்ரியா', 'te': 'ఎరిట్రియా', 'teo': 'Eritrea', 'tg': 'Эритрея', 'th': 'เอริเทรีย', 'ti': 'ኤርትራ', 'tk': 'Eritreýa', 'to': 'ʻElitulia', 'tr': 'Eritre', 'tt': 'Эритрея', 'twq': 'Eritree', 'tzm': 'Iritrya', 'ug': 'ئېرىترىيە', 'uk': 'Еритрея', 'ur': 'اریٹیریا', 'uz': 'Eritreya', 'uz-Cyrl': 'Эритрея', 'uz-Latn': 'Eritreya', 'vai': 'ꔀꔸꔳꕟ', 'vai-Latn': 'Ɛritera', 'vai-Vaii': 'ꔀꔸꔳꕟ', 'vec': 'Eritrèa', 'vi': 'Eritrea', 'vun': 'Eritrea', 'wae': 'Eritrea', 'wo': 'Eritere', 'xh': 'E-Eritrea', 'xnr': 'इरिट्रिया', 'xog': 'Eritureya', 'yav': 'elitée', 'yi': 'עריטרעע', 'yo': 'Eritira', 'yrl': 'Eritireya', 'yue': '厄利垂亞', 'yue-Hans': '厄利垂亚', 'yue-Hant': '厄利垂亞', 'zgh': 'ⵉⵔⵉⵜⵉⵔⵢⴰ', 'zh': '厄立特里亚', 'zh-Hans': '厄立特里亚', 'zh-Hant': '厄利垂亞', 'zu': 'i-Eritrea'},
    'ES': {'af': 'Spanje', 'agq': 'Sɨ̀kpɛ̂n', 'ak': 'Spain', 'am': 'ስፔን', 'ar': 'إسبانيا', 'as': 'স্পেইন', 'asa': 'Hithpania', 'ast': 'España', 'az': 'İspaniya', 'az-Cyrl': 'Испанија', 'az-Latn': 'İspaniya', 'bal-Latn': 'Espin', 'bas': 'Pànya', 'be': 'Іспанія', 'be-tarask': 'Іспанія', 'bez': 'Huhispania', 'bg': 'Испания', 'bgn': 'هسپانیه', 'blo': 'Ɛspanyǝ', 'bm': 'Esipaɲi', 'bn': 'স্পেন', 'br': 'Spagn', 'brx': 'स्पैन', 'bs': 'Španija', 'bs-Cyrl': 'Шпанија', 'bs-Latn': 'Španija', 'ca': 'Espanya', 'ca-ES-valencia': 'Espanya', 'ccp': '𑄥𑄳𑄛𑄬𑄚𑄴', 'ce': 'Испани', 'ceb': 'Espanya', 'cgg': 'Sipeyini', 'chr': 'ᎠᏂᏍᏆᏂᏱ', 'ckb': 'ئیسپانیا', 'cs': 'Španělsko', 'cv': 'Испани', 'cy': 'Sbaen', 'da': 'Spanien', 'dav': 'Hispania', 'de': 'Spanien', 'dje': 'Espaaɲe', 'dsb': 'Špańska', 'dyo': 'Espaañ', 'dz': 'ཨིས་པེན', 'ebu': 'Hispania', 'ee': 'Spain nutome', 'el': 'Ισπανία', 'el-polyton': 'Ἱσπανία', 'en': 'Spain', 'en-Dsrt': '𐐝𐐹𐐩𐑌', 'eo': 'Hispanujo', 'es': 'España', 'et': 'Hispaania', 'eu': 'Espainia', 'ewo': 'Kpənyá', 'fa': 'اسپانیا', 'ff': 'Espaañ', 'ff-Adlm': '𞤉𞤧𞤨𞤢𞤻𞤢𞥄', 'ff-Latn': 'Espaañ', 'fi': 'Espanja', 'fil': 'Spain', 'fo': 'Spania', 'fr': 'Espagne', 'fur': 'Spagne', 'fy': 'Spanje', 'ga': 'an Spáinn', 'gd': 'An Spàinnt', 'gl': 'España', 'gsw': 'Schpanie', 'gu': 'સ્પેન', 'guz': 'Hispania', 'ha': 'Sipen', 'haw': 'Kepania', 'he': 'ספרד', 'hi': 'स्पेन', 'hi-Latn': 'Spain', 'hr': 'Španjolska', 'hsb': 'Španiska', 'ht': 'Espagne', 'hu': 'Spanyolország', 'hy': 'Իսպանիա', 'ia': 'Espania', 'id': 'Spanyol', 'ie': 'Hispania', 'ig': 'Spain', 'is': 'Spánn', 'it': 'Spagna', 'ja': 'スペイン', 'jgo': 'Ɛspániya', 'jmc': 'Hispania', 'jv': 'Sepanyol', 'ka': 'ესპანეთი', 'kaa': 'Испания', 'kaa-Cyrl': 'Испания', 'kab': 'Spanya', 'kam': 'Hispania', 'kde': 'Hispania', 'kea': 'Spanha', 'kgp': 'Enhpỹnija', 'khq': 'Espaaɲe', 'ki': 'Hispania', 'kk': 'Испания', 'kk-Cyrl': 'Испания', 'kln': 'Emetab Spain', 'km': 'អេស្ប៉ាញ', 'kn': 'ಸ್ಪೇನ್', 'ko': '스페인', 'kok': 'स्पेन', 'kok-Deva': 'स्पेन', 'kok-Latn': 'Ispania', 'ks': 'سٕپین', 'ks-Arab': 'سٕپین', 'ksb': 'Hispania', 'ksf': 'kpanyá', 'ksh': 'Schpaanije', 'ku': 'Spanya', 'kxv': 'spein', 'kxv-Deva': 'स्पेन', 'kxv-Latn': 'spein', 'kxv-Orya': 'ସ୍ପେନ୍', 'kxv-Telu': 'స్పెన్', 'ky': 'Испания', 'lag': 'Hisipánia', 'lb': 'Spanien', 'lg': 'Sipeyini', 'lkt': 'Spayólaȟče Makȟóčhe', 'lld': 'Spagna', 'ln': 'Esipanye', 'lo': 'ສະເປນ', 'lt': 'Ispanija', 'lu': 'Nsipani', 'luo': 'Spain', 'luy': 'Hispania', 'lv': 'Spānija', 'mai': 'स्पेन', 'mas': 'Hispania', 'mer': 'Spĩni', 'mfe': 'Lespagn', 'mg': 'Espaina', 'mgh': 'Uhispania', 'mi': 'Peina', 'mk': 'Шпанија', 'ml': 'സ്\u200cപെയിൻ', 'mn': 'Испани', 'mr': 'स्पेन', 'ms': 'Sepanyol', 'mt': 'Spanja', 'mua': 'Espaŋiya', 'my': 'စပိန်', 'mzn': 'ایسپانیا', 'naq': 'Spanieb', 'nb': 'Spania', 'nd': 'Spain', 'ne': 'स्पेन', 'nl': 'Spanje', 'nmg': 'Paŋá', 'nn': 'Spania', 'no': 'Spania', 'nqo': 'ߊߛߌߔߊ߲߫', 'nyn': 'Sipeyini', 'oc': 'Espanha', 'om': 'Ispeen', 'or': 'ସ୍ପେନ୍', 'pa': 'ਸਪੇਨ', 'pa-Guru': 'ਸਪੇਨ', 'pcm': 'Spen', 'pl': 'Hiszpania', 'ps': 'هسپانیه', 'pt': 'Espanha', 'qu': 'España', 'rm': 'Spagna', 'rn': 'Hisipaniya', 'ro': 'Spania', 'rof': 'Hispania', 'ru': 'Испания', 'rwk': 'Hispania', 'saq': 'Hispania', 'sbp': 'Hisipaniya', 'sc': 'Ispagna', 'sd': 'اسپين', 'sd-Arab': 'اسپين', 'se': 'Spánia', 'seh': 'Espanha', 'ses': 'Espaaɲe', 'sg': 'Espânye', 'shi': 'ⵙⴱⴰⵏⵢⴰ', 'shi-Latn': 'sbanya', 'shi-Tfng': 'ⵙⴱⴰⵏⵢⴰ', 'si': 'ස්පාඤ්ඤය', 'sk': 'Španielsko', 'sl': 'Španija', 'smn': 'Espanja', 'sn': 'Spain', 'so': 'Isbeyn', 'sq': 'Spanjë', 'sr': 'Шпанија', 'sr-Cyrl': 'Шпанија', 'sr-Latn': 'Španija', 'sv': 'Spanien', 'sw': 'Uhispania', 'syr': 'ܐܣܦܢܝܐ', 'ta': 'ஸ்பெயின்', 'te': 'స్పెయిన్', 'teo': 'Hispania', 'tg': 'Испания', 'th': 'สเปน', 'ti': 'ስጳኛ', 'tk': 'Ispaniýa', 'to': 'Sipeini', 'tr': 'İspanya', 'tt': 'Испания', 'twq': 'Espaaɲe', 'tzm': 'Sbanya', 'ug': 'ئىسپانىيە', 'uk': 'Іспанія', 'ur': 'ہسپانیہ', 'uz': 'Ispaniya', 'uz-Cyrl': 'Испания', 'uz-Latn': 'Ispaniya', 'vai': 'ꕐꘊꔧ', 'vai-Latn': 'Panyɛĩ', 'vai-Vaii': 'ꕐꘊꔧ', 'vec': 'Spagna', 'vi': 'Tây Ban Nha', 'vun': 'Hispania', 'wae': 'Schpanie', 'wo': 'Españ', 'xh': 'ESpain', 'xnr': 'स्पेन', 'xog': 'Sipeyini', 'yav': 'panyá', 'yi': 'שפּאַניע', 'yo': 'Sípéìnì', 'yrl': 'Isipãya', 'yue': '西班牙', 'yue-Hans': '西班牙', 'yue-Hant': '西班牙', 'zgh': 'ⵙⴱⴰⵏⵢⴰ', 'zh': '西班牙', 'zh-Hans': '西班牙', 'zh-Hant': '西班牙', 'zu': 'i-Spain'},
    'ET': {'af': 'Ethiopië', 'agq': 'Ètyǒpìa', 'ak': 'Ithiopia', 'am': 'ኢትዮጵያ', 'ar': 'إثيوبيا', 'as': 'ইথিঅ’পিয়া', 'asa': 'Uhabeshi', 'ast': 'Etiopía', 'az': 'Efiopiya', 'az-Cyrl': 'Ефиопија', 'az-Latn': 'Efiopiya', 'bal-Latn': 'Etupiá', 'bas': 'Ètìopìà', 'be': 'Эфіопія', 'be-tarask': 'Эфіопія', 'bez': 'Huuhabeshi', 'bg': 'Етиопия', 'bgn': 'ایتوپیه', 'blo': 'Etiyopii', 'bm': 'Etiopi', 'bn': 'ইথিওপিয়া', 'br': 'Etiopia', 'brx': 'इथिय’पिया', 'bs': 'Etiopija', 'bs-Cyrl': 'Етиопија', 'bs-Latn': 'Etiopija', 'ca': 'Etiòpia', 'ca-ES-valencia': 'Etiòpia', 'ccp': '𑄃𑄨𑄜𑄨𑄃𑄮𑄛𑄨𑄠', 'ce': 'Эфиопи', 'ceb': 'Ethiopia', 'cgg': 'Ethiyopiya', 'chr': 'ᎢᏗᎣᏈᎠ', 'ckb': 'ئەتیۆپیا', 'cs': 'Etiopie', 'cv': 'Эфиопи', 'cy': 'Ethiopia', 'da': 'Etiopien', 'dav': 'Uhabeshi', 'de': 'Äthiopien', 'dje': 'Ecioopi', 'dsb': 'Etiopiska', 'dyo': 'Ecoopi', 'dz': 'ཨི་ཐི་ཡོ་པི་ཡ', 'ebu': 'Uhabeshi', 'ee': 'Etiopia nutome', 'el': 'Αιθιοπία', 'el-polyton': 'Αἰθιοπία', 'en': 'Ethiopia', 'en-Dsrt': '𐐀𐑃𐐨𐐬𐐹𐐨𐐲', 'eo': 'Etiopujo', 'es': 'Etiopía', 'et': 'Etioopia', 'eu': 'Etiopia', 'ewo': 'Etiopí', 'fa': 'اتیوپی', 'ff': 'Ecoppi', 'ff-Adlm': '𞤀𞤦𞤢𞤧𞤭𞤲𞤭𞥅', 'ff-Latn': 'Ecoppi', 'fi': 'Etiopia', 'fil': 'Ethiopia', 'fo': 'Etiopia', 'fr': 'Éthiopie', 'fur': 'Etiopie', 'fy': 'Ethiopië', 'ga': 'an Aetóip', 'gd': 'An Itiop', 'gl': 'Etiopía', 'gsw': 'Äthiopie', 'gu': 'ઇથિઓપિયા', 'guz': 'Uhabeshi', 'ha': 'Habasha', 'he': 'אתיופיה', 'hi': 'इथियोपिया', 'hi-Latn': 'Ethiopia', 'hr': 'Etiopija', 'hsb': 'Etiopiska', 'ht': 'Éthiopie', 'hu': 'Etiópia', 'hy': 'Եթովպիա', 'ia': 'Ethiopia', 'id': 'Etiopia', 'ie': 'Etiopia', 'ig': 'Ethiopia', 'is': 'Eþíópía', 'it': 'Etiopia', 'ja': 'エチオピア', 'jgo': 'Ɛtiyɔpî', 'jmc': 'Uhabeshi', 'jv': 'Étiopia', 'ka': 'ეთიოპია', 'kaa': 'Эфиопия', 'kaa-Cyrl': 'Эфиопия', 'kab': 'Utyupi', 'kam': 'Uhabeshi', 'kde': 'Uhabeshi', 'kea': 'Etiópia', 'kgp': 'Etiópija', 'khq': 'Ecioopi', 'ki': 'Uhabeshi', 'kk': 'Эфиопия', 'kk-Cyrl': 'Эфиопия', 'kln': 'Emetab Ethiopia', 'km': 'អេត្យូពី', 'kn': 'ಇಥಿಯೋಪಿಯಾ', 'ko': '에티오피아', 'kok': 'इथियोपिया', 'kok-Deva': 'इथियोपिया', 'ks': 'اِتھوپِیا', 'ks-Arab': 'اِتھوپِیا', 'ksb': 'Uhabeshi', 'ksf': 'ɛtyɔpí', 'ksh': 'Ättijoopije', 'ku': 'Etîyopya', 'kxv': 'ītiyopiya', 'kxv-Deva': 'इतियोपिया', 'kxv-Latn': 'ītiyopiya', 'kxv-Orya': 'ଇତିୟପିଆ', 'kxv-Telu': 'ఇతియోపియా', 'ky': 'Эфиопия', 'lag': 'Ʉhabéeshi', 'lb': 'Ethiopien', 'lg': 'Esyopya', 'lld': 'Etiopia', 'ln': 'Etsíopi', 'lo': 'ອີທິໂອເປຍ', 'lt': 'Etiopija', 'lu': 'Etshiopi', 'luo': 'Ethiopia', 'luy': 'Uhabeshi', 'lv': 'Etiopija', 'mai': 'इथोपिया', 'mas': 'Uhabeshi', 'mer': 'Ithiopia', 'mfe': 'Letiopi', 'mg': 'Ethiopia', 'mgh': 'Uhabeshi', 'mi': 'Etiopia', 'mk': 'Етиопија', 'ml': 'എത്യോപ്യ', 'mn': 'Этиоп', 'mr': 'इथिओपिया', 'ms': 'Ethiopia', 'mt': 'l-Etjopja', 'mua': 'Etiopia', 'my': 'အီသီယိုးပီးယား', 'mzn': 'اتیوپی', 'naq': 'Ethiopiab', 'nb': 'Etiopia', 'nd': 'Ethiopia', 'ne': 'इथियोपिया', 'nl': 'Ethiopië', 'nmg': 'Ethiopiá', 'nn': 'Etiopia', 'no': 'Etiopia', 'nqo': 'ߋߗߏߔߌ߫', 'nyn': 'Ethiyopiya', 'om': 'Itoophiyaa', 'or': 'ଇଥିଓପିଆ', 'pa': 'ਇਥੋਪੀਆ', 'pa-Guru': 'ਇਥੋਪੀਆ', 'pcm': 'Ẹtiópia', 'pl': 'Etiopia', 'ps': 'حبشه', 'pt': 'Etiópia', 'qu': 'Etiopía', 'rm': 'Etiopia', 'rn': 'Etiyopiya', 'ro': 'Etiopia', 'rof': 'Uhabeshi', 'ru': 'Эфиопия', 'rwk': 'Uhabeshi', 'saq': 'Uhabeshi', 'sbp': 'Uhabeshi', 'sc': 'Etiòpia', 'sd': 'ايٿوپيا', 'sd-Arab': 'ايٿوپيا', 'se': 'Etiopia', 'seh': 'Etiópia', 'ses': 'Ecioopi', 'sg': 'Etiopïi', 'shi': 'ⵉⵜⵢⵓⴱⵢⴰ', 'shi-Latn': 'ityubya', 'shi-Tfng': 'ⵉⵜⵢⵓⴱⵢⴰ', 'si': 'ඉතියෝපියාව', 'sk': 'Etiópia', 'sl': 'Etiopija', 'smn': 'Etiopia', 'sn': 'Etiopia', 'so': 'Itoobiya', 'sq': 'Etiopi', 'sr': 'Етиопија', 'sr-Cyrl': 'Етиопија', 'sr-Latn': 'Etiopija', 'sv': 'Etiopien', 'sw': 'Ethiopia', 'syr': 'ܟܘܫ', 'ta': 'எத்தியோப்பியா', 'te': 'ఇథియోపియా', 'teo': 'Uhabeshi', 'tg': 'Эфиопия', 'th': 'เอธิโอเปีย', 'ti': 'ኢትዮጵያ', 'tk': 'Efiopiýa', 'to': 'ʻĪtiōpia', 'tr': 'Etiyopya', 'tt': 'Эфиопия', 'twq': 'Ecioopi', 'tzm': 'Ityupya', 'ug': 'ئېفىيوپىيە', 'uk': 'Ефіопія', 'ur': 'ایتھوپیا', 'uz': 'Efiopiya', 'uz-Cyrl': 'Эфиопия', 'uz-Latn': 'Efiopiya', 'vai': 'ꔤꔳꖎꔪꕩ', 'vai-Latn': 'Ítiyópiya', 'vai-Vaii': 'ꔤꔳꖎꔪꕩ', 'vec': 'Etiopia', 'vi': 'Ethiopia', 'vun': 'Uhabeshi', 'wae': 'Ethiopie', 'wo': 'Ecopi', 'xh': 'E-Ethiopia', 'xnr': 'इथियोपिया', 'xog': 'Esyopya', 'yav': 'etiopí', 'yi': 'עטיאפּיע', 'yo': 'Etopia', 'yrl': 'Etiupiya', 'yue': '衣索比亞', 'yue-Hans': '衣索比亚', 'yue-Hant': '衣索比亞', 'zgh': 'ⵉⵜⵢⵓⴱⵢⴰ', 'zh': '埃塞俄比亚', 'zh-Hans': '埃塞俄比亚', 'zh-Hant': '衣索比亞', 'zu': 'i-Ethiopia'},
    'EU': {'af': 'Europese Unie', 'ak': 'Yuropu Nkabomkuo', 'am': 'የአውሮፓ ህብረት', 'ar': 'الاتحاد الأوروبي', 'as': 'ইউৰোপীয় সংঘ', 'ast': 'Xunión Europea', 'az': 'Avropa Birliyi', 'az-Cyrl': 'Авропа Бирлији', 'az-Latn': 'Avropa Birliyi', 'bal-Latn': 'Yuropi Yunian', 'be': 'Еўрапейскі саюз', 'be-tarask': 'Еўрапейскі саюз', 'bg': 'Европейски съюз', 'bgn': 'اورورپایی یکویی', 'blo': 'Ganɔ gaɖɔŋkɔnɔ kaBʊtǝna Garɩfɔntǝna nɩ', 'bn': 'ইউরোপীয় ইউনিয়ন', 'br': 'Unaniezh Europa', 'brx': 'यूरोपनि जथाय', 'bs': 'Evropska unija', 'bs-Cyrl': 'Европска унија', 'bs-Latn': 'Evropska unija', 'ca': 'Unió Europea', 'ca-ES-valencia': 'Unió Europea', 'ccp': '𑄄𑄃𑄪𑄢𑄮𑄛𑄩𑄠𑄧 𑄄𑄃𑄪𑄚𑄨𑄠𑄧𑄚𑄴', 'ce': 'Евробарт', 'ceb': 'European Union', 'chr': 'ᏳᎳᏛ ᎠᏂᎤᎾᏓᏡᎬ', 'ckb': 'یەکێتیی ئەورووپا', 'cs': 'Evropská unie', 'csw': 'ᐊᑳᒪᐢᑮᔭ', 'cv': 'Европа пӗрлешӗвӗ', 'cy': 'Yr Undeb Ewropeaidd', 'da': 'Den Europæiske Union', 'de': 'Europäische Union', 'dsb': 'Europska unija', 'dz': 'ཡུ་རོབ་གཅིག་བསྡོམས་ཚོགས་པ', 'ee': 'Europa Wɔɖeka nutome', 'el': 'Ευρωπαϊκή Ένωση', 'el-polyton': 'Ευρωπαϊκή Ένωση', 'en': 'European Union', 'en-Dsrt': '𐐏𐐲𐑉𐐲𐐹𐐨𐐲𐑌 𐐏𐐭𐑌𐐷𐐲𐑌', 'es': 'Unión Europea', 'et': 'Euroopa Liit', 'eu': 'Europar Batasuna', 'fa': 'اتحادیهٔ اروپا', 'ff-Adlm': '𞤑𞤢𞤱𞤼𞤢𞤤 𞤓𞤪𞤨𞤢', 'fi': 'Euroopan unioni', 'fil': 'European Union', 'fo': 'Evropasamveldið', 'fr': 'Union européenne', 'fur': 'Union europeane', 'fy': 'Europeeske Unie', 'ga': 'an tAontas Eorpach', 'gd': 'An t-Aonadh Eòrpach', 'gl': 'Unión Europea', 'gsw': 'Europääischi Unioon', 'gu': 'યુરોપિયન સંઘ', 'ha': 'Tarayyar Turai', 'he': 'האיחוד האירופי', 'hi': 'यूरोपीय संघ', 'hi-Latn': 'European Union', 'hr': 'Europska unija', 'hsb': 'Europska unija', 'ht': 'Union européenne', 'hu': 'Európai Unió', 'hy': 'Եվրոպական Միություն', 'ia': 'Union Europee', 'id': 'Uni Eropa', 'ie': 'Union Europan', 'ig': 'Otu nzukọ mba Europe', 'is': 'Evrópusambandið', 'it': 'Unione europea', 'ja': '欧州連合', 'jv': 'Uni Éropah', 'ka': 'ევროკავშირი', 'kaa': 'Европа Аўқамы', 'kaa-Cyrl': 'Европа Аўқамы', 'kea': 'Union Europeia', 'kgp': 'Unĩjáv Oropéja', 'kk': 'Еуропалық Одақ', 'kk-Cyrl': 'Еуропалық Одақ', 'km': 'សហភាព\u200bអឺរ៉ុប', 'kn': 'ಯುರೋಪಿಯನ್ ಒಕ್ಕೂಟ', 'ko': '유럽 연합', 'kok': 'युरोपियन युनियन', 'kok-Deva': 'युरोपियन युनियन', 'ks': 'یوٗرپی یونین', 'ks-Arab': 'یوٗرپی یونین', 'ksh': 'de Europähjesche Unijon', 'ku': 'Yekîtîya Ewropayê', 'ky': 'Европа Биримдиги', 'lb': 'Europäesch Unioun', 'lld': 'Uniun Europeica', 'lo': 'ສະຫະພາບຢູໂຣບ', 'lt': 'Europos Sąjunga', 'lv': 'Eiropas Savienība', 'mai': 'यूरोपीय यूनियन', 'mi': 'Te Uniana o Ūropi', 'mk': 'Европска унија', 'ml': 'യൂറോപ്യൻ യൂണിയൻ', 'mn': 'Европын Холбоо', 'mr': 'युरोपीय संघ', 'ms': 'Kesatuan Eropah', 'mt': 'Unjoni Ewropea', 'my': 'ဥရောပသမဂ္ဂ', 'mzn': 'اروپا اتحادیه', 'nb': 'Den europeiske union', 'ne': 'युरोपियन युनियन', 'nl': 'Europese Unie', 'nn': 'Den europeiske unionen', 'no': 'Den europeiske union', 'nqo': 'ߋߙߐߔߎ߬ ߘߍ߭', 'om': 'Gamtaa Awurooppaa', 'or': 'ୟୁରୋପୀୟ ସଂଘ', 'pa': 'ਯੂਰਪੀ ਸੰਘ', 'pa-Guru': 'ਯੂਰਪੀ ਸੰਘ', 'pcm': 'Yurópián Yúniọ́n', 'pl': 'Unia Europejska', 'ps': 'اروپايي اتحاديه', 'pt': 'União Europeia', 'qu': 'Union Europea', 'rm': 'Uniun Europeica', 'ro': 'Uniunea Europeană', 'ru': 'Европейский союз', 'sc': 'Unione Europea', 'sd': 'يورپين يونين', 'sd-Arab': 'يورپين يونين', 'se': 'Eurohpa Uniovdna', 'si': 'යුරෝපා සංගමය', 'sk': 'Európska únia', 'sl': 'Evropska unija', 'so': 'Midowga Yurub', 'sq': 'Bashkimi Evropian', 'sr': 'Европска унија', 'sr-Cyrl': 'Европска унија', 'sr-Latn': 'Evropska unija', 'sv': 'Europeiska unionen', 'sw': 'Umoja wa Ulaya', 'syr': 'ܚܘܝܕܐ ܐܘܪܘܦܝܐ', 'ta': 'ஐரோப்பிய யூனியன்', 'te': 'యూరోపియన్ యూనియన్', 'tg': 'Иттиҳоди Аврупо', 'th': 'สหภาพยุโรป', 'ti': 'ኤውሮጳዊ ሕብረት', 'tk': 'Ýewropa Bileleşigi', 'to': 'ʻEulope fakatahataha', 'tr': 'Avrupa Birliği', 'tt': 'Европа Берлеге', 'ug': 'ياۋروپا ئىتتىپاقى', 'uk': 'Європейський Союз', 'ur': 'یوروپی یونین', 'uz': 'Yevropa Ittifoqi', 'uz-Cyrl': 'Европа Иттифоқи', 'uz-Latn': 'Yevropa Ittifoqi', 'vec': 'Union europèa', 'vi': 'Liên Minh Châu Âu', 'wae': 'Europäiši Unio', 'wo': 'EZ', 'xh': 'I-European Union', 'xnr': 'यूरोपीय संघ', 'yi': 'אייראפּעישער פֿאַרבאַנד', 'yo': 'Àpapọ̀ Yúróòpù', 'yrl': 'Eurupa Yepewasusawa', 'yue': '歐盟', 'yue-Hans': '欧盟', 'yue-Hant': '歐盟', 'zh': '欧盟', 'zh-Hans': '欧盟', 'zh-Hant': '歐盟', 'zu': 'i-European Union'},
    'EZ': {'af': 'Eurosone', 'ak': 'Yuropu Fam', 'am': 'የዩሮ ዞን', 'ar': 'منطقة اليورو', 'as': 'ইউৰোজ’ন', 'ast': 'Eurozona', 'az': 'Avrozona', 'az-Latn': 'Avrozona', 'bal-Latn': 'Yurop-damag', 'be': 'Еўразона', 'be-tarask': 'Еўразона', 'bg': 'еврозона', 'blo': 'Eroo kaBʊtǝna', 'bn': 'ইউরোজোন', 'br': 'takad an euro', 'brx': 'यूरो ओनसोल', 'bs': 'Eurozona', 'bs-Cyrl': 'Еурозона', 'bs-Latn': 'Eurozona', 'ca': 'zona euro', 'ca-ES-valencia': 'zona euro', 'ce': 'еврозона', 'ceb': 'Eurozone', 'chr': 'ᏳᎶᎠᏍᏓᏅᏅ', 'ckb': 'ناوچەی یۆرۆ', 'cs': 'eurozóna', 'csw': 'ᐊᑳᒪᐢᑮᕽ', 'cv': 'Еврозон', 'cy': 'Ardal yr Ewro', 'da': 'eurozonen', 'de': 'Eurozone', 'dsb': 'europasmo', 'el': 'Ευρωζώνη', 'el-polyton': 'Ευρωζώνη', 'en': 'Eurozone', 'es': 'zona del euro', 'et': 'euroala', 'eu': 'Eurogunea', 'fa': 'منطقهٔ یورو', 'ff-Adlm': '𞤊𞤭𞤪𞤤𞤢 𞤓𞤪𞤮𞥅', 'fi': 'euroalue', 'fil': 'Eurozone', 'fo': 'Evrasona', 'fr': 'zone euro', 'ga': 'Limistéar an euro', 'gd': 'Raon an Eòro', 'gl': 'Eurozona', 'gu': 'યુરોઝોન', 'ha': 'Sashin Turai', 'he': 'גוש האירו', 'hi': 'यूरोज़ोन', 'hi-Latn': 'Eurozone', 'hr': 'eurozona', 'hsb': 'europasmo', 'ht': 'zone euro', 'hu': 'Eurózóna', 'hy': 'Եվրագոտի', 'ia': 'Zona euro', 'id': 'Zona Euro', 'ie': 'Zone de euro', 'ig': 'Gburugburu Euro', 'is': 'Evrusvæðið', 'it': 'Eurozona', 'ja': 'ユーロ圏', 'jv': 'Zona Éuro', 'ka': 'ევროზონა', 'kaa': 'Еврозона', 'kaa-Cyrl': 'Еврозона', 'kea': 'Eurozona', 'kgp': 'Evoro Ga', 'kk': 'Еуроаймақ', 'kk-Cyrl': 'Еуроаймақ', 'km': 'តំបន់ចាយលុយអឺរ៉ូ', 'kn': 'ಯೂರೋಝೋನ್\u200c', 'ko': '유로존', 'kok': 'युरोझोन', 'kok-Deva': 'युरोझोन', 'ks': 'یوٗرو زون', 'ks-Arab': 'یوٗرو زون', 'ku': 'Herêma Ewroyê', 'ky': 'Еврозона', 'lld': 'Zona Euro', 'lo': 'ເຂດຢູໂຣບ', 'lt': 'euro zona', 'lv': 'Eirozona', 'mai': 'यूरो जोन', 'mi': 'Te Rohe o Ūropi', 'mk': 'Еврозона', 'ml': 'യൂറോസോൺ', 'mn': 'Евро бүс', 'mr': 'युरोझोन', 'ms': 'Zon Euro', 'my': 'ဥရောပဒေသ', 'nb': 'eurosonen', 'ne': 'युरोजोन', 'nl': 'eurozone', 'nn': 'eurosona', 'no': 'eurosonen', 'nqo': 'ߋߙߐߔߎ߬ ߞߣߍ', 'om': 'Zooniiyuuroo', 'or': 'ୟୁରୋକ୍ଷେତ୍ର', 'pa': 'ਯੂਰੋਜ਼ੋਨ', 'pa-Guru': 'ਯੂਰੋਜ਼ੋਨ', 'pcm': 'Yúróéria', 'pl': 'strefa euro', 'ps': 'اروپايي سيمه', 'pt': 'zona do euro', 'qu': 'Eurozona', 'rm': 'zona da l’euro', 'ro': 'Zona euro', 'ru': 'еврозона', 'sc': 'Eurozona', 'sd': 'يورو زون', 'sd-Arab': 'يورو زون', 'si': 'යුරෝ කලාපය', 'sk': 'eurozóna', 'sl': 'evroobmočje', 'so': 'Yurusoon', 'sq': 'Zona euro', 'sr': 'Еврозона', 'sr-Cyrl': 'Еврозона', 'sr-Latn': 'Evrozona', 'sv': 'euroområdet', 'sw': 'Jumuiya ya Ulaya', 'syr': 'ܩܠܝܡܐ ܕܐܘܪܘ', 'ta': 'யூரோஜோன்', 'te': 'యూరోజోన్', 'tg': 'Минтақаи Аврупо', 'th': 'ยูโรโซน', 'ti': 'ዞባ ዩሮ', 'tk': 'Ýewro sebiti', 'to': 'ʻEulope fekauʻaki-paʻanga', 'tr': 'Euro Bölgesi', 'tt': 'Еврозона', 'uk': 'Єврозона', 'ur': 'یوروزون', 'uz': 'Yevrozona', 'uz-Latn': 'Yevrozona', 'vec': 'Eurozona', 'vi': 'Khu vực đồng Euro', 'wo': 'Eurozone', 'xh': 'I-Eurozone', 'xnr': 'यूरोज़ोन', 'yo': 'Agbègbè Yúrò', 'yrl': 'euru suna', 'yue': '歐元區', 'yue-Hans': '欧元区', 'yue-Hant': '歐元區', 'zh': '欧元区', 'zh-Hans': '欧元区', 'zh-Hant': '歐元區', 'zu': 'I-Eurozone'},
    'Egyd': {'ar': 'الديموطيقية', 'ast': 'demóticu exipcianu', 'az': 'misir demotik', 'az-Latn': 'misir demotik', 'bg': 'Египетско демотично писмо', 'bn': 'মিশরীয় ডেমোটিক', 'brx': 'मीस्री डैमोटीक्', 'bs': 'egipatsko narodno pismo', 'bs-Cyrl': 'египатско народно писмо', 'bs-Latn': 'egipatsko narodno pismo', 'ca': 'demòtic egipci', 'ca-ES-valencia': 'demòtic egipci', 'ccp': '𑄟𑄨𑄥𑄧𑄢𑄩𑄠𑄧 𑄓𑄬𑄟𑄮𑄑𑄨𑄇𑄴', 'cs': 'egyptské démotické', 'da': 'egyptisk demotisk', 'de': 'Ägyptisch - Demotisch', 'el': 'Λαϊκό Αιγυπτιακό', 'el-polyton': 'Λαϊκό Αιγυπτιακό', 'en': 'Egyptian demotic', 'en-Dsrt': '𐐀𐐾𐐮𐐹𐐽𐐲𐑌 𐐼𐐲𐑋𐐱𐐻𐐮𐐿', 'es': 'egipcio demótico', 'et': 'egiptuse demootiline', 'fi': 'egyptiläinen demoottinen', 'fr': 'démotique égyptien', 'fur': 'demotic egjizian', 'fy': 'Egyptysk demotysk', 'ga': 'Éigipteach coiteann', 'gsw': 'Temozisch-Ägüptisch', 'gu': 'ઇજિપ્શિયન ડેમોટિક', 'hi': 'इजिप्शियन डेमोटिक', 'hi-Latn': 'Egyptian demotic', 'hr': 'egipatsko narodno pismo', 'ht': 'démotique égyptien', 'hu': 'Egyiptomi demotikus', 'id': 'Demotik Mesir', 'it': 'egiziano demotico', 'ja': 'エジプト民衆文字', 'ka': 'ეგვიპტური დემოტიკური', 'kgp': 'nemãtiko ejimso', 'kn': 'ಈಜಿಪ್ಟಿಯನ್ ಡೆಮೋಟಿಕ್', 'ko': '고대 이집트 민중문자', 'ks': 'اِجپشِیَن ڈِماٹِک', 'ks-Arab': 'اِجپشِیَن ڈِماٹِک', 'lb': 'Egyptesch-Demotesch', 'lo': 'ດີໂມຕິກອີຍິບ', 'lt': 'Egipto liaudies', 'lv': 'demotiskais raksts', 'mk': 'египетско демотско', 'ml': 'ഈജിപ്ഷ്യൻ ഡിമോട്ടിക്', 'mr': 'इजिप्शियन डेमोटिक', 'nb': 'egyptisk demotisk', 'ne': 'इजिप्टियन डेमोटिक', 'nl': 'Egyptisch demotisch', 'nn': 'egyptisk demotisk', 'no': 'egyptisk demotisk', 'or': 'ଇଜିପ୍ଟିଆନ୍ ଡେମୋଟିକ୍', 'pl': 'egipskie demotyczne', 'pt': 'demótico egípcio', 'rm': 'egipzian demotic', 'ro': 'demotică egipteană', 'ru': 'египетская демотическая', 'sl': 'demotska egipčanska pisava', 'sr': 'египатско народно писмо', 'sr-Cyrl': 'египатско народно писмо', 'sr-Latn': 'egipatsko narodno pismo', 'sv': 'demotiska', 'ta': 'எகிப்தியன் டெமோட்டிக்', 'te': 'ఇజిప్షియన్ డెమోటిక్', 'th': 'ดีโมติกอียิปต์', 'to': 'tohinima temotika-fakaʻisipite', 'tr': 'Demotik Mısır', 'ug': 'دېموتىكچە مىسىر', 'uk': 'єгипетський демотичний', 'vi': 'Chữ Ai Cập bình dân', 'yrl': 'demútiku egipisiu', 'yue': '古埃及世俗體', 'yue-Hans': '古埃及世俗体', 'yue-Hant': '古埃及世俗體', 'zh': '后期埃及文', 'zh-Hans': '后期埃及文', 'zh-Hant': '古埃及世俗體'},
    'Egyh': {'ar': 'الهيراطيقية', 'ast': 'hieráticu exipcianu', 'az': 'misir hiyeratik', 'az-Latn': 'misir hiyeratik', 'bg': 'Египетско йератично писмо', 'bn': 'মিশরীয় হায়রেটিক', 'brx': 'मीस्री हैरैटीक्', 'bs': 'egipatsko hijeratsko pismo', 'bs-Cyrl': 'египатско хијератско писмо', 'bs-Latn': 'egipatsko hijeratsko pismo', 'ca': 'hieràtic egipci', 'ca-ES-valencia': 'hieràtic egipci', 'ccp': '𑄟𑄨𑄥𑄧𑄢𑄩𑄠𑄧 𑄦𑄠𑄴𑄢𑄬𑄑𑄨𑄇𑄴', 'cs': 'egyptské hieratické', 'da': 'egyptisk hieratisk', 'de': 'Ägyptisch - Hieratisch', 'el': 'Ιερατικό Αιγυπτιακό', 'el-polyton': 'Ιερατικό Αιγυπτιακό', 'en': 'Egyptian hieratic', 'en-Dsrt': '𐐀𐐾𐐮𐐹𐐽𐐲𐑌 𐐸𐐴𐑉𐐰𐐻𐐮𐐿', 'es': 'egipcio hierático', 'et': 'egiptuse hieraatiline', 'fa': 'کاهنی مصری', 'fi': 'egyptiläinen hieraattinen', 'fr': 'hiératique égyptien', 'fur': 'jeratic egjizian', 'fy': 'Egyptysk hiëratysk', 'ga': 'Éigipteach cliarúil', 'gsw': 'Hiraazisch-Ägüptisch', 'gu': 'ઇજિપ્શિયન હાઇરેટિક', 'hi': 'इजिप्शियन हाइरेटिक', 'hi-Latn': 'Egyptian hieratic', 'hr': 'egipatsko hijeratsko pismo', 'ht': 'hiératique égyptien', 'hu': 'Egyiptomi hieratikus', 'id': 'Hieratik Mesir', 'it': 'ieratico egiziano', 'ja': 'エジプト神官文字', 'ka': 'ეგვიპტური იერატიკული', 'kgp': 'jeratiko ejimso', 'kn': 'ಈಜಿಪ್ಟಿಯನ್ ಹಯಾರಿಟಿಕ್', 'ko': '고대 이집트 신관문자', 'ks': 'اِجِپشَن ہَیریٹِک', 'ks-Arab': 'اِجِپشَن ہَیریٹِک', 'lb': 'Egyptesch-Hieratesch', 'lo': 'ເຮຍຮາຕິກອີຍິບ', 'lt': 'Egipto žynių', 'lv': 'hierātiskais raksts', 'mk': 'египетско хиератско', 'ml': 'ഈജിപ്ഷ്യൻ ഹിരാറ്റിക്', 'mr': 'इजिप्शियन हायरेटिक', 'nb': 'egyptisk hieratisk', 'ne': 'इजिप्टियन हाइरटिक', 'nl': 'Egyptisch hiëratisch', 'nn': 'egyptisk hieratisk', 'no': 'egyptisk hieratisk', 'or': 'ଇଜିପ୍ଟିଆନ୍ ହାଇଅରଟିକ୍', 'pl': 'egipskie hieratyczne', 'pt': 'hierático egípcio', 'rm': 'egipzian ieratic', 'ro': 'hieratică egipteană', 'ru': 'египетская иератическая', 'sl': 'hieratska egipčanska pisava', 'sr': 'египатско хијератско писмо', 'sr-Cyrl': 'египатско хијератско писмо', 'sr-Latn': 'egipatsko hijeratsko pismo', 'sv': 'hieratiska', 'ta': 'எகிப்தியன் ஹைரேட்டிக்', 'te': 'ఇజిప్షియన్ హైరాటిక్', 'th': 'เฮียราติกอียิปต์', 'to': 'tohinima hielatika-fakaʻisipite', 'tr': 'Hiyeratik Mısır', 'ug': 'خىيەراتىكچە مىسىر', 'uk': 'єгипетський ієратичний', 'vi': 'Chữ Ai Cập thày tu', 'yrl': 'ierátiku egipisiu', 'yue': '古埃及僧侶體', 'yue-Hans': '古埃及僧侣体', 'yue-Hant': '古埃及僧侶體', 'zh': '古埃及僧侣书写体', 'zh-Hans': '古埃及僧侣书写体', 'zh-Hant': '古埃及僧侶體'},
    'Egyp': {'ar': 'الهيروغليفية', 'ast': 'xeroglíficos exipcianos', 'az': 'misir hiyeroqlif', 'az-Latn': 'misir hiyeroqlif', 'bal-Latn': 'Mesri hirógleps', 'bg': 'Египетски йероглифи', 'bn': 'মিশরীয় হায়ারোগ্লিপ', 'br': 'hieroglifoù egiptek', 'brx': 'मीस्री हैरोग्लीफ़्', 'bs': 'egipatski hijeroglifi', 'bs-Cyrl': 'египатски хијероглифи', 'bs-Latn': 'egipatski hijeroglifi', 'ca': 'jeroglífic egipci', 'ca-ES-valencia': 'jeroglífic egipci', 'ccp': '𑄟𑄨𑄥𑄧𑄢𑄩𑄠𑄧 𑄦𑄠𑄢𑄮𑄉𑄳𑄣𑄨𑄛𑄴', 'cs': 'egyptské hieroglyfy', 'da': 'egyptiske hieroglyffer', 'de': 'Ägyptische Hieroglyphen', 'el': 'Αιγυπτιακά Ιερογλυφικά', 'el-polyton': 'Αιγυπτιακά Ιερογλυφικά', 'en': 'Egyptian hieroglyphs', 'en-Dsrt': '𐐀𐐾𐐮𐐹𐐽𐐲𐑌 𐐸𐐴𐑉𐐬𐑀𐑊𐐮𐑁𐐮𐐿𐑅', 'es': 'jeroglíficos egipcios', 'et': 'egiptuse hieroglüüfkiri', 'eu': 'egiptoar hieroglifikoak', 'fa': 'هیروگلیف مصری', 'ff-Adlm': '𞤖𞤭𞤪𞤮𞤺𞤭𞤪𞤬𞤵 𞤃𞤭𞤧𞤭𞤪𞤢', 'fi': 'egyptiläiset hieroglyfit', 'fo': 'egyptiskar hieroglyffur', 'fr': 'hiéroglyphes égyptiens', 'fur': 'jeroglifics egjizians', 'fy': 'Egyptyske hiërogliefen', 'ga': 'Iairiglifí Éigipteacha', 'gd': 'Sealbh-sgrìobhadh Èipheiteach', 'gsw': 'Ägüptischi Hiroglüüfe', 'gu': 'ઇજિપ્શિયન હાઇરોગ્લિફ્સ', 'he': 'כתב חרטומים', 'hi': 'इजिप्शियन हाइरोग्लिफ्स', 'hi-Latn': 'Egyptian hieroglyphs', 'hr': 'egipatski hijeroglifi', 'ht': 'hiéroglyphes égyptiens', 'hu': 'Egyiptomi hieroglifák', 'id': 'Hieroglip Mesir', 'ig': 'Egyptian hieroglyphs', 'it': 'geroglifici egiziani', 'ja': 'エジプト聖刻文字', 'ka': 'ეგვიპტური იეროგლიფური', 'kgp': 'jerógrifo ejimso', 'kk': 'мысыр жазуы', 'kk-Cyrl': 'мысыр жазуы', 'kn': 'ಈಜಿಪ್ಟಿಯನ್ ಹೀರೋಗ್ಲಿಫ್ಸ್', 'ko': '고대 이집트 신성문자', 'ks': 'اِجِپشَن ہَیروگلِپھس', 'ks-Arab': 'اِجِپشَن ہَیروگلِپھس', 'ku': 'hîyeroglîfên misirî', 'lb': 'Egyptesch Hieroglyphen', 'lo': 'ເຮຍໂຮກລິຟອີຍິບ', 'lt': 'egipto hieroglifai', 'lv': 'ēģiptiešu hieroglifi', 'mk': 'египетски хиероглифи', 'ml': 'ഈജിപ്ഷ്യൻ ചിത്രലിപി', 'mn': 'Египетийн дүрс бичиг', 'mr': 'इजिप्शियन हायरोग्लिफ्स', 'ms': 'Hiroglif Mesir', 'nb': 'egyptiske hieroglyfer', 'ne': 'इजिप्टियन हाइरोग्लिफ्स', 'nl': 'Egyptische hiërogliefen', 'nn': 'egyptiske hieroglyfar', 'no': 'egyptiske hieroglyfer', 'nqo': 'ߋߖ߭ߌߔߑߕ ߤߌߙߏߜ߭ߟߌߝߌ', 'or': 'ଇଜିପ୍ଟିଆନ୍ ହାଅରଗ୍ଲିପସ୍', 'pl': 'hieroglify egipskie', 'pt': 'hieróglifos egípcios', 'rm': 'ieroglifas egipzianas', 'ro': 'hieroglife egiptene', 'ru': 'египетская иероглифическая', 'sc': 'geroglìficos egitzianos', 'sk': 'egyptské hieroglyfy', 'sl': 'egipčanska slikovna pisava', 'so': 'Fartii hore ee Masaarida', 'sq': 'hieroglife egjiptiane', 'sr': 'египатски хијероглифи', 'sr-Cyrl': 'египатски хијероглифи', 'sr-Latn': 'egipatski hijeroglifi', 'sv': 'egyptiska hieroglyfer', 'ta': 'எகிப்தியன் ஹைரோகிளிப்ஸ்', 'te': 'ఇజిప్షియన్ హైరోగ్లైఫ్స్', 'tg': 'Иероглифҳои Мисрӣ', 'th': 'เฮียโรกลิฟส์อียิปต์', 'to': 'tohinima tongitapu-fakaʻisipite', 'tr': 'Mısır Hiyeroglifleri', 'ug': 'تەسۋىرىي يېزىق مىسىر', 'uk': 'єгипетський ієрогліфічний', 'vec': 'jeroglìfeghi ejisiani', 'vi': 'Chữ tượng hình Ai Cập', 'xnr': 'इजिप्शियन हाइरोग्लिफ्स', 'yrl': 'egipsiu-ita kuatiara kuxiímawara', 'yue': '古埃及象形文字', 'yue-Hans': '古埃及象形文字', 'yue-Hant': '古埃及象形文字', 'zh': '古埃及象形文', 'zh-Hans': '古埃及象形文', 'zh-Hant': '古埃及聖書體', 'zu': 'i-Egyptian hieroglyphs'},
    'Elba': {'ast': 'elbasan', 'bal-Latn': 'Elbási', 'bs': 'elbasansko pismo', 'bs-Cyrl': 'елбасан писмо', 'bs-Latn': 'elbasansko pismo', 'ca': 'elbasan', 'ca-ES-valencia': 'elbasan', 'cs': 'elbasanské', 'de': 'Elbasanisch', 'en': 'Elbasan', 'et': 'Elbasani', 'eu': 'elbasanera', 'ff-Adlm': '𞤉𞤤𞤦𞤢𞤧𞤢𞤲', 'fi': 'elbasanilainen', 'fo': 'elbasan', 'ga': 'Elbasan', 'gd': 'Elbasan', 'he': 'אלבסאן', 'hi-Latn': 'Elbasan', 'id': 'Elbasan', 'ig': 'Elbasan', 'ja': 'エルバサン文字', 'kk': 'эльбасан жазуы', 'kk-Cyrl': 'эльбасан жазуы', 'ko': '엘바산 문자', 'lt': 'Elbasano', 'mk': 'елбасанско', 'mn': 'Элбасан', 'ms': 'Elbasan', 'nb': 'elbasisk', 'nl': 'Elbasan', 'nn': 'elbasisk', 'no': 'elbasisk', 'nqo': 'ߍߟߑߓߊߛߊ߲', 'ro': 'elbasan', 'sc': 'elbasan', 'so': 'Magaalada Elbasan', 'sq': 'shkrim i Elbasanit', 'sv': 'elbasiska', 'tg': 'Элбасан', 'th': 'เอลบ์ซาน', 'to': 'tohinima fakaʻelepasani', 'tr': 'Elbasan', 'vec': 'elbasan', 'yue': '愛爾巴桑文', 'yue-Hans': '爱尔巴桑文', 'yue-Hant': '愛爾巴桑文', 'zh': '爱尔巴桑文', 'zh-Hans': '爱尔巴桑文', 'zh-Hant': '愛爾巴桑文', 'zu': 'isi-Elbasan'},
    'Elym': {'bal-Latn': 'Élimái', 'bs': 'elimaično pismo', 'bs-Cyrl': 'елимаик писмо', 'bs-Latn': 'elimaično pismo', 'en': 'Elymaic', 'et': 'elümi', 'eu': 'elimaikera', 'ff-Adlm': '𞤉𞤤𞤴𞤥𞤢𞤴𞤳', 'fi': 'elymealainen', 'fo': 'elymaic', 'fr': 'élymaïque', 'gd': 'Elymaidheach', 'he': 'אלימי', 'hi-Latn': 'Elymaic', 'ht': 'élymaïque', 'id': 'Elymaic', 'ig': 'Elymaic', 'kk': 'элимай жазуы', 'kk-Cyrl': 'элимай жазуы', 'mn': 'Элимайк', 'ms': 'Elymaic', 'nl': 'Elymaisch', 'nqo': 'ߋߟߌߡߊߦߌߞ', 'ro': 'elimaică', 'sc': 'elimàicu', 'so': 'Qoraalka Elymaic', 'sq': 'elimaisht', 'sv': 'elymaiska', 'tg': 'Элимайӣ', 'vec': 'elimàego', 'zh': '埃利迈文', 'zh-Hans': '埃利迈文', 'zu': 'isi-Elymaic'},
    'Ethi': {'af': 'Etiopies', 'ak': 'Yitiopia kasa', 'am': 'ኢትዮፒክ', 'ar': 'الأثيوبية', 'as': 'ইথিঅ’পিক', 'ast': 'etíope', 'az': 'efiop', 'az-Latn': 'efiop', 'bal-Latn': 'Etyupi', 'be': 'эфіопскае', 'be-tarask': 'эфіопскае', 'bg': 'етиопска', 'bgn': 'گعزی', 'bn': 'ইথিওপিয়', 'br': 'etiopek', 'brx': 'ईथियपिक', 'bs': 'etiopsko pismo', 'bs-Cyrl': 'етиопско писмо', 'bs-Latn': 'etiopsko pismo', 'ca': 'etiòpic', 'ca-ES-valencia': 'etiòpic', 'ccp': '𑄃𑄨𑄗𑄨𑄃𑄮𑄛𑄨𑄠𑄧', 'ce': 'эфиопин', 'chr': 'ᎢᏗᏯᏈᎩ', 'ckb': 'ئەتیۆپیک', 'cs': 'etiopské', 'cy': 'Ethiopig', 'da': 'etiopisk', 'de': 'Äthiopisch', 'dsb': 'etiopiski', 'dz': 'ཨི་ཐི་ཡོ་པིཀ྄་ཡིག་གུ', 'ee': 'ethiopiagbeŋɔŋlɔ', 'el': 'Αιθιοπικό', 'el-polyton': 'Αἰθιοπικό', 'en': 'Ethiopic', 'en-Dsrt': '𐐀𐑃𐐨𐐪𐐹𐐮𐐿', 'es': 'etiópico', 'et': 'etioopia', 'eu': 'etiopiarra', 'fa': 'اتیوپیایی', 'ff-Adlm': '𞤖𞤢𞤦𞤢𞤧𞤭𞤲𞤳𞤮', 'fi': 'etiopialainen', 'fil': 'Ethiopic', 'fo': 'etiopiskt', 'fr': 'éthiopique', 'fur': 'etiopic', 'fy': 'Ethiopysk', 'ga': 'Aetóipic', 'gd': 'Ge’ez', 'gl': 'etíope', 'gsw': 'Äzioopisch', 'gu': 'ઇથિયોપિક', 'ha': 'Ethiopic', 'he': 'אתיופי', 'hi': 'इथियोपियाई', 'hi-Latn': 'Ethiopic', 'hr': 'etiopsko pismo', 'hsb': 'etiopisce', 'ht': 'éthiopique', 'hu': 'Etióp', 'hy': 'եթովպական', 'ia': 'ethiope', 'id': 'Etiopia', 'ig': 'Mkpụrụ ọkwụ Etọpịa', 'is': 'eþíópískt', 'it': 'etiope', 'ja': 'エチオピア文字', 'jv': 'Ethiopik', 'ka': 'ეთიოპიური', 'kea': 'etiópiku', 'kgp': 'etijópiko', 'kk': 'эфиопиялық жазу', 'kk-Cyrl': 'эфиопиялық жазу', 'km': 'អេត្យូពី', 'kn': 'ಇಥಿಯೋಪಿಕ್', 'ko': '에티오피아 문자', 'kok': 'इथियोपीक', 'kok-Deva': 'इथियोपीक', 'ks': 'اِتھیوپِک', 'ks-Arab': 'اِتھیوپِک', 'ksh': 'ätejoopesche Schreff', 'ku': 'etîyopîk', 'ky': 'Эфиоп', 'lb': 'Ethiopesch', 'lo': 'ອີທິໂອປິກ', 'lrc': 'ئتوٙیوٙپیایی', 'lt': 'etiopų', 'lv': 'etiopiešu', 'mai': 'इथोपिक', 'mi': 'Etiopika', 'mk': 'етиопско писмо', 'ml': 'എത്യോപിക്', 'mn': 'этиоп', 'mr': 'ईथिओपिक', 'ms': 'Ethiopia', 'my': 'အီသီယိုးပီးယား', 'mzn': 'اتیوپیایی', 'nb': 'etiopisk', 'ne': 'इथियोपिक', 'nl': 'Ethiopisch', 'nn': 'etiopisk', 'no': 'etiopisk', 'nqo': 'ߋߗߏߔߌߞ', 'or': 'ଇଥିଓପିକ୍\u200c', 'pa': 'ਇਥੀਓਪਿਕ', 'pa-Guru': 'ਇਥੀਓਪਿਕ', 'pcm': 'Ẹtiópik', 'pl': 'etiopskie', 'ps': 'ایتوپي', 'pt': 'etiópico', 'qu': 'Etiope', 'rm': 'etiopic', 'ro': 'etiopiană', 'ru': 'эфиопская', 'sc': 'etìope', 'sd': 'ايٿوپيائي', 'sd-Arab': 'ايٿوپيائي', 'si': 'ඉතියෝපියානු', 'sk': 'etiópske', 'sl': 'etiopski', 'so': 'Itoobiya', 'sq': 'etiopik', 'sr': 'етиопско писмо', 'sr-Cyrl': 'етиопско писмо', 'sr-Latn': 'etiopsko pismo', 'sv': 'etiopiska', 'sw': 'Kiethiopia', 'ta': 'எத்தியோபிக்', 'te': 'ఇథియోపిక్', 'tg': 'Эфиопӣ', 'th': 'เอธิโอปิก', 'ti': 'እትዮጵያዊ', 'tk': 'Efiop elipbiýi', 'to': 'tohinima fakaʻītiōpia', 'tr': 'Etiyopya', 'ug': 'ئېفىيوپىيەچە', 'uk': 'ефіопська', 'ur': 'ایتھوپیائی', 'uz': 'habash', 'uz-Cyrl': 'Ҳабаш', 'uz-Latn': 'habash', 'vec': 'etiòpego', 'vi': 'Chữ Ethiopia', 'wae': 'Ethiopiš', 'xnr': 'इथियोपियाई', 'yo': 'Ẹtiópíìkì', 'yrl': 'etiúpiku', 'yue': '衣索比亞文', 'yue-Hans': '衣索比亚文', 'yue-Hant': '衣索比亞文', 'zh': '埃塞俄比亚文', 'zh-Hans': '埃塞俄比亚文', 'zh-Hant': '衣索比亞文', 'zu': 'isi-Ethiopic'},
    'FI': {'af': 'Finland', 'agq': 'Fɨnlàn', 'ak': 'Finland', 'am': 'ፊንላንድ', 'ar': 'فنلندا', 'as': 'ফিনলেণ্ড', 'asa': 'Ufini', 'ast': 'Finlandia', 'az': 'Finlandiya', 'az-Cyrl': 'Финландија', 'az-Latn': 'Finlandiya', 'bal-Latn': 'Fenlánd', 'bas': 'Fìnlând', 'be': 'Фінляндыя', 'be-tarask': 'Фінляндыя', 'bez': 'Huufini', 'bg': 'Финландия', 'blo': 'Fɛnlanɖ', 'bm': 'Finilandi', 'bn': 'ফিনল্যান্ড', 'br': 'Finland', 'brx': 'फिनलेण्ड', 'bs': 'Finska', 'bs-Cyrl': 'Финска', 'bs-Latn': 'Finska', 'ca': 'Finlàndia', 'ca-ES-valencia': 'Finlàndia', 'ccp': '𑄜𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴', 'ce': 'Финлянди', 'ceb': 'Finland', 'cgg': 'Bufini', 'chr': 'ᏫᏂᎦᏙᎯ', 'ckb': 'فینلاند', 'cs': 'Finsko', 'cv': 'Финлянди', 'cy': 'Y Ffindir', 'da': 'Finland', 'dav': 'Ufini', 'de': 'Finnland', 'dje': 'Finlandu', 'dsb': 'Finska', 'dyo': 'Finland', 'dz': 'ཕིན་ལེནཌ', 'ebu': 'Ufini', 'ee': 'Finland nutome', 'el': 'Φινλανδία', 'el-polyton': 'Φινλανδία', 'en': 'Finland', 'en-Dsrt': '𐐙𐐮𐑌𐑊𐐲𐑌𐐼', 'eo': 'Finnlando', 'es': 'Finlandia', 'et': 'Soome', 'eu': 'Finlandia', 'ewo': 'Finəlán', 'fa': 'فنلاند', 'ff': 'Fenland', 'ff-Adlm': '𞤊𞤭𞤲𞤤𞤢𞤲𞤣', 'ff-Latn': 'Fenland', 'fi': 'Suomi', 'fil': 'Finland', 'fo': 'Finnland', 'fr': 'Finlande', 'fur': 'Finlandie', 'fy': 'Finlân', 'ga': 'an Fhionlainn', 'gd': 'An Fhionnlann', 'gl': 'Finlandia', 'gsw': 'Finnland', 'gu': 'ફિનલેન્ડ', 'guz': 'Ufini', 'ha': 'Finlan', 'he': 'פינלנד', 'hi': 'फ़िनलैंड', 'hi-Latn': 'Finland', 'hr': 'Finska', 'hsb': 'Finska', 'ht': 'Finlande', 'hu': 'Finnország', 'hy': 'Ֆինլանդիա', 'ia': 'Finlandia', 'id': 'Finlandia', 'ie': 'Finland', 'ig': 'Finland', 'is': 'Finnland', 'it': 'Finlandia', 'ja': 'フィンランド', 'jmc': 'Ufini', 'jv': 'Finlan', 'ka': 'ფინეთი', 'kaa': 'Финляндия', 'kaa-Cyrl': 'Финляндия', 'kab': 'Finlund', 'kam': 'Ufini', 'kde': 'Ufini', 'kea': 'Finlándia', 'kgp': 'Fĩrỹnija', 'khq': 'Finlandu', 'ki': 'Ufini', 'kk': 'Финляндия', 'kk-Cyrl': 'Финляндия', 'kln': 'Emetab Finland', 'km': 'ហ្វាំងឡង់', 'kn': 'ಫಿನ್\u200cಲ್ಯಾಂಡ್', 'ko': '핀란드', 'kok': 'फिनलँड', 'kok-Deva': 'फिनलँड', 'ks': 'فِن لینڈ', 'ks-Arab': 'فِن لینڈ', 'ksb': 'Ufini', 'ksf': 'fínlan', 'ksh': 'Finnland', 'ku': 'Fînlenda', 'kxv': 'pinlaṇd', 'kxv-Deva': 'प़िनलेंड', 'kxv-Latn': 'pinlaṇd', 'kxv-Orya': 'ପିନ୍\u200dଲାଣ୍ଡ', 'kxv-Telu': 'పిన్లాండ్', 'ky': 'Финляндия', 'lag': 'Ufíini', 'lb': 'Finnland', 'lg': 'Finilandi', 'lld': 'Finlandia', 'ln': 'Filandɛ', 'lo': 'ຟິນແລນ', 'lt': 'Suomija', 'lu': 'Filande', 'luo': 'Finland', 'luy': 'Ufini', 'lv': 'Somija', 'mai': 'फिनलैंड', 'mas': 'Ufini', 'mer': 'Finilandi', 'mfe': 'Finland', 'mg': 'Finlandy', 'mgh': 'Ufini', 'mi': 'Whinarana', 'mk': 'Финска', 'ml': 'ഫിൻലാൻഡ്', 'mn': 'Финланд', 'mr': 'फिनलंड', 'ms': 'Finland', 'mt': 'il-Finlandja', 'mua': 'Sǝr Finland', 'my': 'ဖင်လန်', 'mzn': 'فنلاند', 'naq': 'Finlandi', 'nb': 'Finland', 'nd': 'Finland', 'ne': 'फिनल्याण्ड', 'nl': 'Finland', 'nmg': 'Finlande', 'nn': 'Finland', 'no': 'Finland', 'nqo': 'ߝߍ߲ߟߊ߲ߘ', 'nyn': 'Bufini', 'om': 'Fiinlaand', 'or': 'ଫିନଲ୍ୟାଣ୍ଡ', 'pa': 'ਫਿਨਲੈਂਡ', 'pa-Guru': 'ਫਿਨਲੈਂਡ', 'pcm': 'Fínland', 'pl': 'Finlandia', 'ps': 'فنلینډ', 'pt': 'Finlândia', 'qu': 'Finlandia', 'rm': 'Finlanda', 'rn': 'Finilandi', 'ro': 'Finlanda', 'rof': 'Ufini', 'ru': 'Финляндия', 'rwk': 'Ufini', 'sah': 'Финляндия', 'saq': 'Ufini', 'sbp': 'Wufini', 'sc': 'Finlàndia', 'sd': 'فن لينڊ', 'sd-Arab': 'فن لينڊ', 'se': 'Suopma', 'seh': 'Finlândia', 'ses': 'Finlandu', 'sg': 'Fëlânde', 'shi': 'ⴼⵉⵍⵍⴰⵏⴷⴰ', 'shi-Latn': 'fillanda', 'shi-Tfng': 'ⴼⵉⵍⵍⴰⵏⴷⴰ', 'si': 'ෆින්ලන්තය', 'sk': 'Fínsko', 'sl': 'Finska', 'smn': 'Suomâ', 'sn': 'Finland', 'so': 'Finland', 'sq': 'Finlandë', 'sr': 'Финска', 'sr-Cyrl': 'Финска', 'sr-Latn': 'Finska', 'sv': 'Finland', 'sw': 'Ufini', 'syr': 'ܦܝܢܠܢܕ', 'ta': 'பின்லாந்து', 'te': 'ఫిన్లాండ్', 'teo': 'Ufini', 'tg': 'Финляндия', 'th': 'ฟินแลนด์', 'ti': 'ፊንላንድ', 'tk': 'Finlýandiýa', 'to': 'Finilani', 'tr': 'Finlandiya', 'tt': 'Финляндия', 'twq': 'Finlandu', 'tzm': 'Finlanḍa', 'ug': 'فىنلاندىيە', 'uk': 'Фінляндія', 'ur': 'فن لینڈ', 'uz': 'Finlandiya', 'uz-Cyrl': 'Финляндия', 'uz-Latn': 'Finlandiya', 'vai': 'ꔱꘋ ꖨꕮꕊ', 'vai-Latn': 'Fiŋlɛŋ', 'vai-Vaii': 'ꔱꘋ ꖨꕮꕊ', 'vec': 'Finlanda', 'vi': 'Phần Lan', 'vun': 'Ufini', 'wae': 'Finnland', 'wo': 'Finlànd', 'xh': 'EFinland', 'xnr': 'फ़िनलैंड', 'xog': 'Finilandi', 'yav': 'fɛnlánd', 'yi': 'פֿינלאַנד', 'yo': 'Filandi', 'yrl': 'Firãdiya', 'yue': '芬蘭', 'yue-Hans': '芬兰', 'yue-Hant': '芬蘭', 'zgh': 'ⴼⵉⵍⵍⴰⵏⴷⴰ', 'zh': '芬兰', 'zh-Hans': '芬兰', 'zh-Hant': '芬蘭', 'zu': 'i-Finland'},
    'FJ': {'af': 'Fidji', 'agq': 'Fidzi', 'ak': 'Figyi', 'am': 'ፊጂ', 'ar': 'فيجي', 'as': 'ফিজি', 'asa': 'Fiji', 'ast': 'Islles Fixi', 'az': 'Fici', 'az-Cyrl': 'Фиҹи', 'az-Latn': 'Fici', 'bal-Latn': 'Fiji', 'bas': 'Fiji', 'be': 'Фіджы', 'be-tarask': 'Фіджы', 'bez': 'Hufiji', 'bg': 'Фиджи', 'bgn': 'فیجی', 'blo': 'Fiji', 'bm': 'Fiji', 'bn': 'ফিজি', 'br': 'Fidji', 'brx': 'फिजी', 'bs': 'Fidži', 'bs-Cyrl': 'Фиџи', 'bs-Latn': 'Fidži', 'ca': 'Fiji', 'ca-ES-valencia': 'Fiji', 'ccp': '𑄜𑄨𑄎𑄨', 'ce': 'Фиджи', 'ceb': 'Fiji', 'cgg': 'Figyi', 'chr': 'ᏫᏥ', 'ckb': 'فیجی', 'cs': 'Fidži', 'cv': 'Фиджи', 'cy': 'Fiji', 'da': 'Fiji', 'dav': 'Fiji', 'de': 'Fidschi', 'dje': 'Fiji', 'dsb': 'Fidži', 'dyo': 'Fiji', 'dz': 'ཕི་ཇི', 'ebu': 'Fiji', 'ee': 'Fidzi nutome', 'el': 'Φίτζι', 'el-polyton': 'Φίτζι', 'en': 'Fiji', 'en-Dsrt': '𐐙𐐨𐐾𐐨', 'eo': 'Fiĝoj', 'es': 'Fiyi', 'et': 'Fidži', 'eu': 'Fiji', 'ewo': 'Fidzí', 'fa': 'فیجی', 'ff': 'Fijji', 'ff-Adlm': '𞤊𞤭𞤶𞤭𞥅', 'ff-Latn': 'Fijji', 'fi': 'Fidži', 'fil': 'Fiji', 'fo': 'Fiji', 'fr': 'Fidji', 'fur': 'Fizi', 'fy': 'Fiji', 'ga': 'Fidsí', 'gd': 'Fìdi', 'gl': 'Fixi', 'gsw': 'Fitschi', 'gu': 'ફીજી', 'guz': 'Fiji', 'ha': 'Fiji', 'he': 'פיג׳י', 'hi': 'फ़िजी', 'hi-Latn': 'Fiji', 'hr': 'Fidži', 'hsb': 'Fidźi', 'ht': 'Fidji', 'hu': 'Fidzsi', 'hy': 'Ֆիջի', 'ia': 'Fiji', 'id': 'Fiji', 'ie': 'Fidji', 'ig': 'Fiji', 'is': 'Fídjíeyjar', 'it': 'Figi', 'ja': 'フィジー', 'jmc': 'Fiji', 'jv': 'Fiji', 'ka': 'ფიჯი', 'kaa': 'Фиджи', 'kaa-Cyrl': 'Фиджи', 'kab': 'Fiji', 'kam': 'Fiji', 'kde': 'Fiji', 'kea': 'Fidji', 'kgp': 'Fiji', 'khq': 'Fiji', 'ki': 'Fiji', 'kk': 'Фиджи', 'kk-Cyrl': 'Фиджи', 'kln': 'Emetab Fiji', 'km': 'ហ្វីជី', 'kn': 'ಫಿಜಿ', 'ko': '피지', 'kok': 'फिजी', 'kok-Deva': 'फिजी', 'ks': 'فِجی', 'ks-Arab': 'فِجی', 'ksb': 'Fiji', 'ksf': 'fíji', 'ksh': 'de Fidschi-Endelle', 'ku': 'Fîjî', 'kxv': 'piji', 'kxv-Deva': 'प़िजी', 'kxv-Latn': 'piji', 'kxv-Orya': 'ପିଜି', 'kxv-Telu': 'పిజీ', 'ky': 'Фиджи', 'lag': 'Fíiji', 'lb': 'Fidschi', 'lg': 'Fiji', 'lld': 'Fiji', 'ln': 'Fidzi', 'lo': 'ຟິຈິ', 'lt': 'Fidžis', 'lu': 'Fuji', 'luo': 'Fiji', 'luy': 'Fiji', 'lv': 'Fidži', 'mai': 'फिजी', 'mas': 'Fiji', 'mer': 'Fiji', 'mfe': 'Fidji', 'mg': 'Fidji', 'mgh': 'Ufiji', 'mi': 'Whītī', 'mk': 'Фиџи', 'ml': 'ഫിജി', 'mn': 'Фижи', 'mr': 'फिजी', 'ms': 'Fiji', 'mt': 'Fiġi', 'mua': 'Sǝr Fiji', 'my': 'ဖီဂျီ', 'mzn': 'فیجی', 'naq': 'Fijib', 'nb': 'Fiji', 'nd': 'Fiji', 'ne': 'फिजी', 'nl': 'Fiji', 'nmg': 'Fijiá', 'nn': 'Fiji', 'no': 'Fiji', 'nqo': 'ߝߖߌ߫', 'nyn': 'Figyi', 'om': 'Fiijii', 'or': 'ଫିଜି', 'pa': 'ਫ਼ਿਜੀ', 'pa-Guru': 'ਫ਼ਿਜੀ', 'pcm': 'Fíji', 'pl': 'Fidżi', 'ps': 'فجي', 'pt': 'Fiji', 'qu': 'Fiyi', 'rm': 'Fidschi', 'rn': 'Fiji', 'ro': 'Fiji', 'rof': 'Fiji', 'ru': 'Фиджи', 'rwk': 'Fiji', 'saq': 'Fiji', 'sbp': 'Fiji', 'sc': 'Fiji', 'sd': 'فجي', 'sd-Arab': 'فجي', 'se': 'Fijisullot', 'seh': 'Fiji', 'ses': 'Fiji', 'sg': 'Fidyïi', 'shi': 'ⴼⵉⴷⵊⵉ', 'shi-Latn': 'fidji', 'shi-Tfng': 'ⴼⵉⴷⵊⵉ', 'si': 'ෆීජී', 'sk': 'Fidži', 'sl': 'Fidži', 'smn': 'Fidži', 'sn': 'Fiji', 'so': 'Fiji', 'sq': 'Fixhi', 'sr': 'Фиџи', 'sr-Cyrl': 'Фиџи', 'sr-Latn': 'Fidži', 'sv': 'Fiji', 'sw': 'Fiji', 'syr': 'ܦܝܓܝ', 'ta': 'ஃபிஜி', 'te': 'ఫిజీ', 'teo': 'Fiji', 'tg': 'Фиҷи', 'th': 'ฟิจิ', 'ti': 'ፊጂ', 'tk': 'Fiji', 'to': 'Fisi', 'tr': 'Fiji', 'tt': 'Фиджи', 'twq': 'Fiji', 'tzm': 'Fiji', 'ug': 'فىجى', 'uk': 'Фіджі', 'ur': 'فجی', 'uz': 'Fiji', 'uz-Cyrl': 'Фижи', 'uz-Latn': 'Fiji', 'vai': 'ꔱꔤꕀ', 'vai-Latn': 'Fíji', 'vai-Vaii': 'ꔱꔤꕀ', 'vec': 'Fiji', 'vi': 'Fiji', 'vun': 'Fiji', 'wae': 'Fidši', 'wo': 'Fijji', 'xh': 'EFiji', 'xnr': 'फ़िजी', 'xog': 'Fiji', 'yav': 'físi', 'yi': 'פֿידזשי', 'yo': 'Fíjì', 'yrl': 'Fiyi', 'yue': '斐濟', 'yue-Hans': '斐济', 'yue-Hant': '斐濟', 'zgh': 'ⴼⵉⴷⵊⵉ', 'zh': '斐济', 'zh-Hans': '斐济', 'zh-Hant': '斐濟', 'zu': 'i-Fiji'},
    'FK': {'af': 'Falklandeilande', 'agq': 'Chwɨlà fɨ Fakɨlàn', 'ak': 'Fɔkman Aeland', 'am': 'የፎክላንድ ደሴቶች', 'ar': 'جزر فوكلاند', 'as': 'ফকলেণ্ড দ্বীপপুঞ্জ', 'asa': 'Vithiwa vya Falkland', 'ast': 'Falkland Islands', 'az': 'Folklend adaları', 'az-Cyrl': 'Фолкленд адалары', 'az-Latn': 'Folklend adaları', 'bal-Latn': 'Páklánd Islánd', 'bas': 'Bìòn bi Falkland', 'be': 'Фалклендскія астравы', 'be-tarask': 'Фалклендскія астравы', 'bez': 'Ifisima fya Falkland', 'bg': 'Фолкландски острови', 'blo': 'Fɔklanɖ kaBʊtukǝltǝna', 'bm': 'Maluwini Gun', 'bn': 'ফকল্যান্ড দ্বীপপুঞ্জ', 'br': 'Inizi Falkland', 'brx': 'फकलेण्ड द्वीपफोर', 'bs': 'Folklandska ostrva', 'bs-Cyrl': 'Фокландска Острва', 'bs-Latn': 'Folklandska ostrva', 'ca': 'Illes Falkland', 'ca-ES-valencia': 'Illes Falkland', 'ccp': '𑄜𑄧𑄇𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Фолклендан гӀайренаш', 'ceb': 'Falkland Islands', 'cgg': 'Ebizinga bya Faakilanda', 'chr': 'ᏩᎩ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی مالڤیناس (دوورگەکانی فاڵکلاند)', 'cs': 'Falklandské ostrovy', 'cv': 'Фолкленд утравӗсем', 'cy': 'Ynysoedd y Falkland/Malvinas', 'da': 'Falklandsøerne', 'dav': 'Visiwa vya Falkland', 'de': 'Falklandinseln', 'dje': 'Kalkan gungey', 'dsb': 'Falklandske kupy', 'dz': 'ཕལྐ་ལནྜ་གླིང་ཚོམ', 'ebu': 'Visiwa vya Falkland', 'ee': 'Falkland ƒudomekpowo nutome', 'el': 'Νήσοι Φόκλαντ', 'el-polyton': 'Νήσοι Φόκλαντ', 'en': 'Falkland Islands', 'en-Dsrt': '𐐙𐐪𐑊𐐿𐑊𐐲𐑌𐐼 𐐌𐑊𐐲𐑌𐐼𐑆', 'es': 'Islas Malvinas', 'et': 'Falklandi saared', 'eu': 'Falklandak', 'ewo': 'Minlán Mi Fóləkəlan', 'fa': 'جزایر فالکلند', 'ff': 'Duuɗe Falkland', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤊𞤢𞤤𞤳𞤵𞤤𞤢𞤲𞤣', 'ff-Latn': 'Duuɗe Falkland', 'fi': 'Falklandinsaaret', 'fil': 'Falkland Islands', 'fo': 'Falklandsoyggjar', 'fr': 'Îles Malouines', 'fur': 'Isulis Falkland', 'fy': 'Falklâneilannen', 'ga': 'Oileáin Fháclainne', 'gd': 'Na h-Eileanan Fàclannach', 'gl': 'Illas Malvinas', 'gsw': 'Falkland-Insle', 'gu': 'ફૉકલેન્ડ આઇલેન્ડ્સ', 'guz': 'Visiwa vya Falkland', 'ha': 'Tsibiran Falkilan', 'he': 'איי פוקלנד', 'hi': 'फ़ॉकलैंड द्वीपसमूह', 'hi-Latn': 'Falkland Islands', 'hr': 'Falklandski Otoci', 'hsb': 'Falklandske kupy', 'ht': 'Îles Malouines', 'hu': 'Falkland-szigetek', 'hy': 'Ֆոլքլենդյան կղզիներ', 'ia': 'Insulas Falkland', 'id': 'Kepulauan Falkland', 'ig': 'Falkland Islands', 'is': 'Falklandseyjar', 'it': 'Isole Falkland', 'ja': 'フォークランド諸島', 'jmc': 'Visiwa vya Falkland', 'jv': 'Kapuloan Falkland', 'ka': 'ფოლკლენდის კუნძულები', 'kaa': 'Фолкленд атаўлары', 'kaa-Cyrl': 'Фолкленд атаўлары', 'kab': 'Tigzirin n Falkland', 'kam': 'Visiwa vya Falkland', 'kde': 'Chisiwa cha Falkland', 'kea': 'Ilhas Malvinas', 'kgp': 'Mỹrvĩnỹ Goj-vẽso', 'khq': 'Kalkan gungey', 'ki': 'Visiwa vya Falkland', 'kk': 'Фолкленд аралдары', 'kk-Cyrl': 'Фолкленд аралдары', 'kln': 'Ikwembeyotab Falkland', 'km': 'កោះ\u200bហ្វក់ឡែន', 'kn': 'ಫಾಕ್\u200cಲ್ಯಾಂಡ್ ದ್ವೀಪಗಳು', 'ko': '포클랜드 제도', 'kok': 'फ़ॉकलैंड आइलैंड्स', 'kok-Deva': 'फ़ॉकलैंड आइलैंड्स', 'ks': 'فٕلاکلینڑ جٔزیٖرٕ', 'ks-Arab': 'فٕلاکلینڑ جٔزیٖرٕ', 'ksb': 'Visiwa vya Falkland', 'ksf': 'zǝ maalwín', 'ksh': 'de Falkland-Enselle', 'ku': 'Giravên Falklandê', 'kxv': 'paklaṇd dīp', 'kxv-Deva': 'प़ॉकलेंड द्वीप', 'kxv-Latn': 'paklaṇd dīp', 'kxv-Orya': 'ପାକ୍\u200dଲାଣ୍ଡ ଦିପ', 'kxv-Telu': 'ఫాక్\u200cల్యాండ్ దీవులు', 'ky': 'Фолкленд аралдары', 'lag': 'Visíiwa vya Fakulandi', 'lb': 'Falklandinselen', 'lg': 'Bizinga by’eFalikalandi', 'lld': 'Isoles Falkland', 'ln': 'Bisanga bya Maluni', 'lo': 'ຫມູ່ເກາະຟອກແລນ', 'lt': 'Folklando Salos', 'lu': 'Lutanda lua Maluni', 'luo': 'Chuia mar Falkland', 'luy': 'Visiwa vya Falkland', 'lv': 'Folklenda salas', 'mai': 'फाकलैंड द्वीपसमूह', 'mas': 'Visiwa vya Falkland', 'mer': 'Aĩrandi cia Falklandi', 'mfe': 'Zil malwinn', 'mg': 'Nosy Falkand', 'mi': 'Motu Whākarangi', 'mk': 'Фолкландски Острови', 'ml': 'ഫാക്ക്\u200cലാന്റ് ദ്വീപുകൾ', 'mn': 'Фолклендийн арлууд', 'mr': 'फॉकलंड बेटे', 'ms': 'Kepulauan Falkland', 'mt': 'il-Gżejjer Falkland', 'mua': 'Sǝr malouniya ma laŋne', 'my': 'ဖော့ကလန် ကျွန်းစု', 'mzn': 'فالکلند جزیره\u200cئون', 'naq': 'Falkland Islands', 'nb': 'Falklandsøyene', 'nd': 'Falkland Islands', 'ne': 'फकल्याण्ड टापुहरु', 'nl': 'Falklandeilanden', 'nmg': 'Maŋ má Falkland', 'nn': 'Falklandsøyane', 'no': 'Falklandsøyene', 'nqo': 'ߡߊߟߎ߲ߌ߲߫ ߕߌ߲', 'nyn': 'Ebizinga bya Faakilanda', 'om': 'Odoloota Faalklaand', 'or': 'ଫକ୍\u200cଲ୍ୟାଣ୍ଡ ଦ୍ଵୀପପୁଞ୍ଜ', 'pa': 'ਫ਼ਾਕਲੈਂਡ ਟਾਪੂ', 'pa-Guru': 'ਫ਼ਾਕਲੈਂਡ ਟਾਪੂ', 'pcm': 'Fọ́klánd Aílands', 'pl': 'Falklandy', 'ps': 'فاکلينډ ټاپوګان', 'pt': 'Ilhas Malvinas', 'qu': 'Islas Malvinas', 'rm': 'Inslas dal Falkland', 'rn': 'Izinga rya Filikilandi', 'ro': 'Insulele Falkland', 'rof': 'Visiwa vya Falkland', 'ru': 'Фолклендские о-ва', 'rwk': 'Visiwa vya Falkland', 'saq': 'Visiwa vya Falkland', 'sbp': 'Figunguli fya Fokolendi', 'sc': 'Ìsulas Falkland', 'sd': 'فاڪ لينڊ ٻيٽ', 'sd-Arab': 'فاڪ لينڊ ٻيٽ', 'se': 'Falklandsullot', 'seh': 'Ilhas Malvinas', 'ses': 'Kalkan gungey', 'sg': 'Âzûâ tî Mälüîni', 'shi': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵎⴰⵍⴰⵡⵉ', 'shi-Latn': 'tigzirin n malawi', 'shi-Tfng': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵎⴰⵍⴰⵡⵉ', 'si': 'ෆෝක්ලන්ත දූපත්', 'sk': 'Falklandy', 'sl': 'Falklandski otoki', 'smn': 'Falklandsuolluuh', 'sn': 'Zvitsuwa zveFalklands', 'so': 'Jaziiradaha Fooklaan', 'sq': 'Ishujt Falkland', 'sr': 'Фокландска Острва', 'sr-Cyrl': 'Фокландска Острва', 'sr-Latn': 'Foklandska Ostrva', 'sv': 'Falklandsöarna', 'sw': 'Visiwa vya Falkland', 'syr': 'ܓܙܪܬܐ ܕܦܠܟܠܢܕ', 'ta': 'ஃபாக்லாந்து தீவுகள்', 'te': 'ఫాక్\u200cల్యాండ్ దీవులు', 'teo': 'Visiwa vya Falkland', 'tg': 'Ҷазираҳои Фолкленд', 'th': 'หมู่เกาะฟอล์กแลนด์', 'ti': 'ደሴታት ፎክላንድ', 'tk': 'Folklend adalary', 'to': 'ʻOtumotu Fokulani', 'tr': 'Falkland Adaları', 'tt': 'Фолкленд утраулары', 'twq': 'Kalkan gungey', 'tzm': 'Tigzirin n Falkland', 'ug': 'فالكلاند ئاراللىرى', 'uk': 'Фолклендські Острови', 'ur': 'فاکلینڈ جزائر', 'uz': 'Folklend orollari', 'uz-Cyrl': 'Фолкленд ороллари', 'uz-Latn': 'Folklend orollari', 'vai': 'ꕘꔷꕃ ꖨꕮ ꔳꘋꗣ', 'vai-Latn': 'Fáháki Luma Tiŋŋɛ', 'vai-Vaii': 'ꕘꔷꕃ ꖨꕮ ꔳꘋꗣ', 'vec': 'Ìzole Malvine', 'vi': 'Quần đảo Falkland', 'vun': 'Visiwa vya Falkland', 'wae': 'Falklandinslä', 'wo': 'Duni Falkland', 'xh': 'EFalkland Islands', 'xnr': 'फ़ॉकलैंड द्वीपां दा झुंड', 'xog': 'Ebiizinga bya Falikalandi', 'yav': 'maluwín', 'yi': 'פֿאַלקלאַנד אינזלען', 'yo': 'Etikun Fakalandi', 'yrl': 'Kapuãma-ita Mawina', 'yue': '福克蘭群島', 'yue-Hans': '福克兰群岛', 'yue-Hant': '福克蘭群島', 'zgh': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵎⴰⵍⴰⵡⵉ', 'zh': '福克兰群岛', 'zh-Hans': '福克兰群岛', 'zh-Hant': '福克蘭群島', 'zu': 'i-Falkland Islands'},
    'FM': {'af': 'Mikronesië', 'agq': 'Maekòlòneshìa', 'ak': 'Maekronehyia', 'am': 'ማይክሮኔዢያ', 'ar': 'ميكرونيزيا', 'as': 'মাইক্ৰোনেচিয়া', 'asa': 'Mikronethia', 'ast': 'Micronesia', 'az': 'Mikroneziya', 'az-Cyrl': 'Микронезија', 'az-Latn': 'Mikroneziya', 'bal-Latn': 'Mikrónéshiá', 'bas': 'Mìkrònesìà', 'be': 'Мікранезія', 'be-tarask': 'Мікранезія', 'bez': 'Humikronesia', 'bg': 'Микронезия', 'blo': 'Mikronesiya', 'bm': 'Mikironesi', 'bn': 'মাইক্রোনেশিয়া', 'br': 'Mikronezia', 'brx': 'माइक्रनेशिया', 'bs': 'Mikronezija', 'bs-Cyrl': 'Микронезија', 'bs-Latn': 'Mikronezija', 'ca': 'Micronèsia', 'ca-ES-valencia': 'Micronèsia', 'ccp': '𑄟𑄭𑄇𑄳𑄢𑄮𑄚𑄬𑄥𑄨𑄠', 'ce': 'Микронезин Федеративни штаташ', 'ceb': 'Micronesia', 'cgg': 'Mikironesiya', 'chr': 'ᎹᎢᏉᏂᏏᏯ', 'ckb': 'مایکرۆنیزیا', 'cs': 'Mikronésie', 'cv': 'Микронези', 'cy': 'Micronesia', 'da': 'Mikronesien', 'dav': 'Mikronesia', 'de': 'Mikronesien', 'dje': 'Mikronezi', 'dsb': 'Mikroneziska', 'dz': 'མའི་ཀྲོ་ནི་ཤི་ཡ', 'ebu': 'Mikronesia', 'ee': 'Mikronesia nutome', 'el': 'Μικρονησία', 'el-polyton': 'Μικρονησία, Ὁμόσπονδες Πολιτεῖες τῆς', 'en': 'Micronesia', 'en-Dsrt': '𐐣𐐴𐐿𐑉𐐲𐑌𐐨𐑈𐐲', 'eo': 'Mikronezio', 'es': 'Micronesia', 'et': 'Mikroneesia', 'eu': 'Mikronesia', 'ewo': 'Mikoronésia', 'fa': 'میکرونزی', 'ff': 'Mikoronesii', 'ff-Adlm': '𞤃𞤭𞤳𞤪𞤮𞤲𞤫𞥅𞤧𞤭𞤴𞤢', 'ff-Latn': 'Mikoronesii', 'fi': 'Mikronesia', 'fil': 'Micronesia', 'fo': 'Mikronesiasamveldið', 'fr': 'Micronésie', 'fur': 'Micronesie', 'fy': 'Micronesië', 'ga': 'an Mhicrinéis', 'gd': 'Na Meanbh-eileanan', 'gl': 'Micronesia', 'gsw': 'Mikroneesie', 'gu': 'માઇક્રોનેશિયા', 'guz': 'Mikronesia', 'ha': 'Mikuronesiya', 'he': 'מיקרונזיה', 'hi': 'माइक्रोनेशिया', 'hi-Latn': 'Micronesia', 'hr': 'Mikronezija', 'hsb': 'Mikroneziska', 'ht': 'Micronésie', 'hu': 'Mikronézia', 'hy': 'Միկրոնեզիա', 'ia': 'Micronesia', 'id': 'Mikronesia', 'ig': 'Micronesia', 'is': 'Míkrónesía', 'it': 'Micronesia', 'ja': 'ミクロネシア連邦', 'jmc': 'Mikronesia', 'jv': 'Féderasi Mikronésia', 'ka': 'მიკრონეზია', 'kaa': 'Микронезия Федеративлик Штатлары', 'kaa-Cyrl': 'Микронезия Федеративлик Штатлары', 'kab': 'Mikrunizya', 'kam': 'Mikronesia', 'kde': 'Mikilonesia', 'kea': 'Mikronézia', 'kgp': 'Goj-vẽso-sĩ Kẽsir', 'khq': 'Mikronezi', 'ki': 'Mikronesia', 'kk': 'Микронезия', 'kk-Cyrl': 'Микронезия', 'kln': 'Emetab Micronesia', 'km': 'មីក្រូណេស៊ី', 'kn': 'ಮೈಕ್ರೋನೇಶಿಯಾ', 'ko': '미크로네시아', 'kok': 'मायक्रोनेशिया', 'kok-Deva': 'मायक्रोनेशिया', 'ks': 'مائیکرونیشیا', 'ks-Arab': 'مائیکرونیشیا', 'ksb': 'Miklonesia', 'ksf': 'mikronɛ́si', 'ksh': 'Mikroneesije', 'ku': 'Mîkronezya', 'kxv': 'micronesiaa', 'kxv-Deva': 'माइक्रोनेसिया', 'kxv-Latn': 'micronesiaa', 'kxv-Orya': 'ମାକ୍ରନେସିଆ', 'kxv-Telu': 'మైక్రోనేసియా', 'ky': 'Микронезия', 'lag': 'Mikironésia', 'lb': 'Mikronesien', 'lg': 'Mikuronezya', 'lld': 'Micronesia', 'ln': 'Mikronezi', 'lo': 'ໄມໂຄຣນີເຊຍ', 'lt': 'Mikronezija', 'lu': 'Mikronezi', 'luo': 'Micronesia', 'luy': 'Mikronesia', 'lv': 'Mikronēzija', 'mai': 'माइक्रोनेशिया', 'mas': 'Mikronesia', 'mer': 'Mikronesia', 'mfe': 'Mikronezi', 'mg': 'Mikrônezia', 'mi': 'Mekanēhia', 'mk': 'Микронезија', 'ml': 'മൈക്രോനേഷ്യ', 'mn': 'Микронези', 'mr': 'मायक्रोनेशिया', 'ms': 'Micronesia', 'mt': 'il-Mikroneżja', 'mua': 'Micronesiya', 'my': 'မိုင်ခရိုနီရှား', 'mzn': 'میکرونزی', 'naq': 'Micronesia', 'nb': 'Mikronesiaføderasjonen', 'nd': 'Micronesia', 'ne': 'माइक्रोनेसिया', 'nl': 'Micronesia', 'nmg': 'Mikronesia', 'nn': 'Mikronesiaføderasjonen', 'no': 'Mikronesiaføderasjonen', 'nqo': 'ߡߌߞߙߏߣߋߖ߭ߌ߫', 'nyn': 'Mikironesiya', 'om': 'Maayikirooneeshiyaa', 'or': 'ମାଇକ୍ରୋନେସିଆ', 'pa': 'ਮਾਇਕ੍ਰੋਨੇਸ਼ੀਆ', 'pa-Guru': 'ਮਾਇਕ੍ਰੋਨੇਸ਼ੀਆ', 'pcm': 'Maikroníshia', 'pl': 'Mikronezja', 'ps': 'میکرونیزیا', 'pt': 'Micronésia', 'qu': 'Micronesia', 'rm': 'Micronesia', 'rn': 'Mikoroniziya', 'ro': 'Micronezia', 'rof': 'Mikronesia', 'ru': 'Федеративные Штаты Микронезии', 'rwk': 'Mikronesia', 'saq': 'Mikronesia', 'sbp': 'Mikilonesiya', 'sc': 'Micronèsia', 'sd': 'مائڪرونيشيا', 'sd-Arab': 'مائڪرونيشيا', 'se': 'Mikronesia', 'seh': 'Micronésia', 'ses': 'Mikronezi', 'sg': 'Mikronezïi', 'shi': 'ⵎⵉⴽⵔⵓⵏⵉⵣⵢⴰ', 'shi-Latn': 'mikrunizya', 'shi-Tfng': 'ⵎⵉⴽⵔⵓⵏⵉⵣⵢⴰ', 'si': 'මයික්\u200dරොනීසියාව', 'sk': 'Mikronézia', 'sl': 'Mikronezija', 'smn': 'Mikronesia littoväldi', 'sn': 'Micronesia', 'so': 'Mikroneesiya', 'sq': 'Mikronezi', 'sr': 'Микронезија', 'sr-Cyrl': 'Микронезија', 'sr-Latn': 'Mikronezija', 'sv': 'Mikronesien', 'sw': 'Mikronesia', 'syr': 'ܐܬܪܘܬܐ ܦܕܪܠܝܐ ܕܡܝܩܪܘܢܝܣܝܐ', 'ta': 'மைக்ரோனேஷியா', 'te': 'మైక్రోనేషియా', 'teo': 'Mikronesia', 'tg': 'Штатҳои Федеративии Микронезия', 'th': 'ไมโครนีเซีย', 'ti': 'ማይክሮነዥያ', 'tk': 'Mikroneziýa', 'to': 'Mikolonīsia', 'tr': 'Mikronezya', 'tt': 'Микронезия', 'twq': 'Mikronezi', 'tzm': 'Mikrunizya', 'ug': 'مىكرونېزىيە', 'uk': 'Мікронезія', 'ur': 'مائکرونیشیا', 'uz': 'Mikroneziya', 'uz-Cyrl': 'Микронезия', 'uz-Latn': 'Mikroneziya', 'vai': 'ꕆꖏꕇꔻꕩ', 'vai-Latn': 'Mikonisiya', 'vai-Vaii': 'ꕆꖏꕇꔻꕩ', 'vec': 'Micronezia', 'vi': 'Micronesia', 'vun': 'Mikronesia', 'wae': 'Mikronesie', 'wo': 'Mikoronesi', 'xh': 'EMicronesia', 'xnr': 'माइक्रोनेशिया', 'xog': 'Mikuronezya', 'yav': 'mikolonesí', 'yi': 'מיקראנעזיע', 'yo': 'Makoronesia', 'yrl': 'Mikuruneziya', 'yue': '密克羅尼西亞群島', 'yue-Hans': '密克罗尼西亚群岛', 'yue-Hant': '密克羅尼西亞群島', 'zgh': 'ⵎⵉⴽⵔⵓⵏⵉⵣⵢⴰ', 'zh': '密克罗尼西亚', 'zh-Hans': '密克罗尼西亚', 'zh-Hant': '密克羅尼西亞', 'zu': 'i-Micronesia'},
    'FO': {'af': 'Faroëreilande', 'ak': 'Faro Aeland', 'am': 'የፋሮ ደሴቶች', 'ar': 'جزر فارو', 'as': 'ফাৰো দ্বীপপুঞ্জ', 'ast': 'Islles Feroe', 'az': 'Farer adaları', 'az-Cyrl': 'Фарер адалары', 'az-Latn': 'Farer adaları', 'bal-Latn': 'Faróé Islánd', 'be': 'Фарэрскія астравы', 'be-tarask': 'Фарэрскія астравы', 'bg': 'Фарьорски острови', 'blo': 'Faroi kaBʊtukǝltǝna', 'bn': 'ফ্যারো দ্বীপপুঞ্জ', 'br': 'Inizi Faero', 'brx': 'फेर’ द्वीपफोर', 'bs': 'Farska ostrva', 'bs-Cyrl': 'Фарска острва', 'bs-Latn': 'Farska ostrva', 'ca': 'Illes Fèroe', 'ca-ES-valencia': 'Illes Fèroe', 'ccp': '𑄜𑄳𑄠𑄢𑄧𑄃𑄮 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Фарерийн гӀайренаш', 'ceb': 'Faroe Islands', 'chr': 'ᏪᎶ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی فارەو', 'cs': 'Faerské ostrovy', 'cv': 'Фарер утравӗсем', 'cy': 'Ynysoedd Ffaro', 'da': 'Færøerne', 'de': 'Färöer', 'dsb': 'Färöje', 'dz': 'ཕཱའེ་རོ་གླིང་ཚོམ', 'ee': 'Faroe ƒudomekpowo nutome', 'el': 'Νήσοι Φερόες', 'el-polyton': 'Νῆσοι Φερόες', 'en': 'Faroe Islands', 'en-Dsrt': '𐐙𐐯𐑉𐐬 𐐌𐑊𐐲𐑌𐐼𐑆', 'eo': 'Ferooj', 'es': 'Islas Feroe', 'et': 'Fääri saared', 'eu': 'Faroe uharteak', 'fa': 'جزایر فارو', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤊𞤢𞤪𞤵𞥅𞤧', 'fi': 'Färsaaret', 'fil': 'Faroe Islands', 'fo': 'Føroyar', 'fr': 'Îles Féroé', 'fur': 'Isulis Faroe', 'fy': 'Faeröer', 'ga': 'Oileáin Fharó', 'gd': 'Na h-Eileanan Fàro', 'gl': 'Illas Feroe', 'gsw': 'Färöer', 'gu': 'ફેરો આઇલેન્ડ્સ', 'ha': 'Tsibirai na Faroe', 'he': 'איי פארו', 'hi': 'फ़ेरो द्वीपसमूह', 'hi-Latn': 'Faroe Islands', 'hr': 'Ovčji Otoci', 'hsb': 'Färöske kupy', 'ht': 'Îles Féroé', 'hu': 'Feröer szigetek', 'hy': 'Ֆարերյան կղզիներ', 'ia': 'Insulas Feroe', 'id': 'Kepulauan Faroe', 'ig': 'Faroe Islands', 'is': 'Færeyjar', 'it': 'Isole Fær Øer', 'ja': 'フェロー諸島', 'jv': 'Kapuloan Faro', 'ka': 'ფარერის კუნძულები', 'kaa': 'Фарер атаўлары', 'kaa-Cyrl': 'Фарер атаўлары', 'kea': 'Ilhas Faroe', 'kgp': 'Faróve Goj-vẽso', 'kk': 'Фарер аралдары', 'kk-Cyrl': 'Фарер аралдары', 'km': 'កោះ\u200bហ្វារ៉ូ', 'kn': 'ಫರೋ ದ್ವೀಪಗಳು', 'ko': '페로 제도', 'kok': 'फैरो आयलँड्स', 'kok-Deva': 'फैरो आयलँड्स', 'ks': 'فارو جزیرہ', 'ks-Arab': 'فارو جزیرہ', 'ksh': 'de Färrör-Enselle', 'ku': 'Giravên Faroeyê', 'kxv': 'pēro dīp', 'kxv-Deva': 'पेरो दीप', 'kxv-Latn': 'pēro dīp', 'kxv-Orya': 'ପେରୋ ଦୀପ', 'kxv-Telu': 'పెరొ దీప', 'ky': 'Фарер аралдары', 'lb': 'Färöer', 'lld': 'Isoles Faroer', 'lo': 'ຫມູ່ເກາະແຟໂຣ', 'lt': 'Farerų Salos', 'lv': 'Fēru salas', 'mai': 'फेरो द्वीपसमूह', 'mi': 'Motu Wharau', 'mk': 'Фарски Острови', 'ml': 'ഫറോ ദ്വീപുകൾ', 'mn': 'Фарерын арлууд', 'mr': 'फेरो बेटे', 'ms': 'Kepulauan Faroe', 'mt': 'il-Gżejjer Faeroe', 'my': 'ဖာရိုး ကျွန်းစုများ', 'mzn': 'فارو جزایر', 'nb': 'Færøyene', 'ne': 'फारो टापुहरू', 'nl': 'Faeröer', 'nn': 'Færøyane', 'no': 'Færøyene', 'nqo': 'ߝߋߙߏߦߋ߫ ߕߌ߲', 'om': 'Odoloota Fafo’ee', 'or': 'ଫାରୋଇ ଦ୍ୱୀପପୁଞ୍ଜ', 'pa': 'ਫੈਰੋ ਟਾਪੂ', 'pa-Guru': 'ਫੈਰੋ ਟਾਪੂ', 'pcm': 'Fáro Aílands', 'pl': 'Wyspy Owcze', 'ps': 'فارو ټاپو', 'pt': 'Ilhas Faroé', 'qu': 'Islas Feroe', 'rm': 'Inslas Feroe', 'ro': 'Insulele Feroe', 'ru': 'Фарерские о-ва', 'sc': 'Ìsulas Føroyar', 'sd': 'فارو ٻيٽ', 'sd-Arab': 'فارو ٻيٽ', 'se': 'Fearsullot', 'si': 'ෆැරෝ දූපත්', 'sk': 'Faerské ostrovy', 'sl': 'Ferski otoki', 'smn': 'Färsuolluuh', 'so': 'Jasiiradda Faroo', 'sq': 'Ishujt Faroe', 'sr': 'Фарска Острва', 'sr-Cyrl': 'Фарска Острва', 'sr-Latn': 'Farska Ostrva', 'sv': 'Färöarna', 'sw': 'Visiwa vya Faroe', 'syr': 'ܓܙܝܖ̈ܐ ܕܦܪܘ', 'ta': 'ஃபாரோ தீவுகள்', 'te': 'ఫారో దీవులు', 'tg': 'Ҷазираҳои Фарер', 'th': 'หมู่เกาะแฟโร', 'ti': 'ደሴታት ፋሮ', 'tk': 'Farer adalary', 'to': 'ʻOtumotu Faloe', 'tr': 'Faroe Adaları', 'tt': 'Фарер утраулары', 'ug': 'فارو ئاراللىرى', 'uk': 'Фарерські Острови', 'ur': 'جزائر فارو', 'uz': 'Farer orollari', 'uz-Cyrl': 'Фарер ороллари', 'uz-Latn': 'Farer orollari', 'vai': 'ꕘꖄ ꔳꘋꗣ', 'vai-Vaii': 'ꕘꖄ ꔳꘋꗣ', 'vec': 'Ìzole Fàroe', 'vi': 'Quần đảo Faroe', 'wae': 'Färöe', 'wo': 'Duni Faro', 'xh': 'EFaroe Islands', 'xnr': 'फ़ेरो द्वीपां दा झुंड', 'yi': 'פֿאַרא אינזלען', 'yo': 'Àwọn Erékùsù ti Faroe', 'yrl': 'Kapuãma-ita Faruwe', 'yue': '法羅群島', 'yue-Hans': '法罗群岛', 'yue-Hant': '法羅群島', 'zh': '法罗群岛', 'zh-Hans': '法罗群岛', 'zh-Hant': '法羅群島', 'zu': 'i-Faroe Islands'},
    'FR': {'af': 'Frankryk', 'agq': 'Fàlâŋnsì', 'ak': 'Franse', 'am': 'ፈረንሳይ', 'ar': 'فرنسا', 'as': 'ফ্ৰান্স', 'asa': 'Ufarantha', 'ast': 'Francia', 'az': 'Fransa', 'az-Cyrl': 'Франса', 'az-Latn': 'Fransa', 'bal-Latn': 'Paráns', 'bas': 'Pùlàsi / Fɛ̀lɛ̀nsi /', 'be': 'Францыя', 'be-tarask': 'Францыя', 'bez': 'Huufaransa', 'bg': 'Франция', 'bgn': 'فرانسه', 'blo': 'Gafɔntǝna', 'bm': 'Faransi', 'bn': 'ফ্রান্স', 'br': 'Frañs', 'brx': 'फ्रान्स', 'bs': 'Francuska', 'bs-Cyrl': 'Француска', 'bs-Latn': 'Francuska', 'ca': 'França', 'ca-ES-valencia': 'França', 'ccp': '𑄜𑄳𑄢𑄚𑄴𑄥𑄴', 'ce': 'Франци', 'ceb': 'France', 'cgg': 'Bufaransa', 'chr': 'ᎦᎸᏥᏱ', 'ckb': 'فەڕەنسا', 'cs': 'Francie', 'cv': 'Франци', 'cy': 'Ffrainc', 'da': 'Frankrig', 'dav': 'Ufaransa', 'de': 'Frankreich', 'dje': 'Faransi', 'doi': 'फ्रांस', 'dsb': 'Francojska', 'dyo': 'Frans', 'dz': 'ཕྲཱནས', 'ebu': 'Ufaransa', 'ee': 'France nutome', 'el': 'Γαλλία', 'el-polyton': 'Γαλλία', 'en': 'France', 'en-Dsrt': '𐐙𐑉𐐰𐑌𐑅', 'eo': 'Francujo', 'es': 'Francia', 'et': 'Prantsusmaa', 'eu': 'Frantzia', 'ewo': 'Fulɛnsí', 'fa': 'فرانسه', 'ff': 'Farayse', 'ff-Adlm': '𞤊𞤢𞤪𞤢𞤲𞤧𞤭', 'ff-Latn': 'Farayse', 'fi': 'Ranska', 'fil': 'France', 'fo': 'Frakland', 'fr': 'France', 'fur': 'France', 'fy': 'Frankrijk', 'ga': 'an Fhrainc', 'gd': 'An Fhraing', 'gl': 'Francia', 'gsw': 'Frankriich', 'gu': 'ફ્રાંસ', 'guz': 'Ufaransa', 'ha': 'Faransa', 'haw': 'Palani', 'he': 'צרפת', 'hi': 'फ़्रांस', 'hi-Latn': 'France', 'hr': 'Francuska', 'hsb': 'Francoska', 'ht': 'France', 'hu': 'Franciaország', 'hy': 'Ֆրանսիա', 'ia': 'Francia', 'id': 'Prancis', 'ie': 'Francia', 'ig': 'France', 'ii': 'ꃔꇩ', 'is': 'Frakkland', 'it': 'Francia', 'ja': 'フランス', 'jgo': 'Fɛlánci', 'jmc': 'Ufaransa', 'jv': 'Prancis', 'ka': 'საფრანგეთი', 'kaa': 'Франция', 'kaa-Cyrl': 'Франция', 'kab': 'Fransa', 'kam': 'Uvalanza', 'kde': 'Ufalansa', 'kea': 'Fransa', 'kgp': 'Frỹsa', 'khq': 'Faransi', 'ki': 'Ubaranja', 'kk': 'Франция', 'kk-Cyrl': 'Франция', 'kln': 'Emetab France', 'km': 'បារាំង', 'kn': 'ಫ್ರಾನ್ಸ್', 'ko': '프랑스', 'kok': 'फ्रान्स', 'kok-Deva': 'फ्रान्स', 'kok-Latn': 'Frans', 'ks': 'فرانس', 'ks-Arab': 'فرانس', 'ks-Deva': 'फ्रांस', 'ksb': 'Ufalansa', 'ksf': 'pɛrɛsǝ́', 'ksh': 'Frankrisch', 'ku': 'Fransa', 'kxv': 'prans', 'kxv-Deva': 'प़्रांस', 'kxv-Latn': 'prans', 'kxv-Orya': 'ପ୍ରାନ୍ସ', 'kxv-Telu': 'ప్రాన్స్', 'ky': 'Франция', 'lag': 'Ʉfaráansa', 'lb': 'Frankräich', 'lg': 'Bufalansa', 'lij': 'Fransa', 'lld': 'Francia', 'ln': 'Falánsɛ', 'lo': 'ຝຣັ່ງ', 'lrc': 'فأرانسە', 'lt': 'Prancūzija', 'lu': 'Nfalanse', 'luo': 'France', 'luy': 'Ufaransa', 'lv': 'Francija', 'mai': 'फ़्रांस', 'mas': 'Ufaransa', 'mer': 'Fransi', 'mfe': 'Lafrans', 'mg': 'Frantsa', 'mgh': 'Ufaransa', 'mi': 'Wīwī', 'mk': 'Франција', 'ml': 'ഫ്രാൻസ്', 'mn': 'Франц', 'mni': 'ফ্রান্স', 'mni-Beng': 'ফ্রান্স', 'mr': 'फ्रान्स', 'ms': 'Perancis', 'mt': 'Franza', 'mua': 'Franssǝ', 'my': 'ပြင်သစ်', 'mzn': 'فرانسه', 'naq': 'Frankreiki', 'nb': 'Frankrike', 'nd': 'Furansi', 'ne': 'फ्रान्स', 'nl': 'Frankrijk', 'nmg': 'Fala', 'nn': 'Frankrike', 'no': 'Frankrike', 'nqo': 'ߝߊ߬ߙߊ߲߬ߛߌ߫', 'nyn': 'Bufaransa', 'oc': 'França', 'om': 'Faransaay', 'or': 'ଫ୍ରାନ୍ସ', 'os': 'Франц', 'pa': 'ਫ਼ਰਾਂਸ', 'pa-Guru': 'ਫ਼ਰਾਂਸ', 'pcm': 'Frans', 'pl': 'Francja', 'ps': 'فرانسه', 'pt': 'França', 'qu': 'Francia', 'rm': 'Frantscha', 'rn': 'Ubufaransa', 'ro': 'Franța', 'rof': 'Ufaransa', 'ru': 'Франция', 'rwk': 'Ufaransa', 'sa': 'फ़्रांस:', 'saq': 'Ufaransa', 'sat': 'ᱯᱷᱨᱟᱱᱥ', 'sat-Olck': 'ᱯᱷᱨᱟᱱᱥ', 'sbp': 'Wufalansa', 'sc': 'Frantza', 'sd': 'فرانس', 'sd-Arab': 'فرانس', 'sd-Deva': 'फ़्रांस', 'se': 'Frankriika', 'seh': 'França', 'ses': 'Faransi', 'sg': 'Farânzi', 'shi': 'ⴼⵔⴰⵏⵙⴰ', 'shi-Latn': 'fransa', 'shi-Tfng': 'ⴼⵔⴰⵏⵙⴰ', 'si': 'ප්\u200dරංශය', 'sk': 'Francúzsko', 'sl': 'Francija', 'smn': 'Ranska', 'sn': 'France', 'so': 'Faransiis', 'sq': 'Francë', 'sr': 'Француска', 'sr-Cyrl': 'Француска', 'sr-Latn': 'Francuska', 'su': 'Prancis', 'su-Latn': 'Prancis', 'sv': 'Frankrike', 'sw': 'Ufaransa', 'syr': 'ܦܪܢܣܐ', 'ta': 'பிரான்ஸ்', 'te': 'ఫ్రాన్స్\u200c', 'teo': 'Ufaransa', 'tg': 'Фаронса', 'th': 'ฝรั่งเศส', 'ti': 'ፈረንሳ', 'tk': 'Fransiýa', 'to': 'Falanisē', 'tpi': 'Frans', 'tr': 'Fransa', 'tt': 'Франция', 'twq': 'Faransi', 'tzm': 'Fṛansa', 'ug': 'فىرانسىيە', 'uk': 'Франція', 'ur': 'فرانس', 'uz': 'Fransiya', 'uz-Cyrl': 'Франция', 'uz-Latn': 'Fransiya', 'vai': 'ꖢꕟꘋꔻ', 'vai-Latn': 'Fɛŋsi', 'vai-Vaii': 'ꖢꕟꘋꔻ', 'vec': 'Fransa', 'vi': 'Pháp', 'vun': 'Ufaransa', 'wae': 'Frankrič', 'wo': 'Faraans', 'xh': 'EFrance', 'xnr': 'फ़्रांस', 'xog': 'Bufalansa', 'yav': 'felensí', 'yi': 'פֿראַנקרייך', 'yo': 'Faranse', 'yrl': 'Frãsa', 'yue': '法國', 'yue-Hans': '法国', 'yue-Hant': '法國', 'zgh': 'ⴼⵔⴰⵏⵙⴰ', 'zh': '法国', 'zh-Hans': '法国', 'zh-Hant': '法國', 'zu': 'i-France'},
    'FX': {'en': 'Metropolitan France'},
    'GA': {'af': 'Gaboen', 'agq': 'Gàbûn', 'ak': 'Gabɔn', 'am': 'ጋቦን', 'ar': 'الغابون', 'as': 'গেবন', 'asa': 'Gaboni', 'ast': 'Gabón', 'az': 'Qabon', 'az-Cyrl': 'Габон', 'az-Latn': 'Qabon', 'bal-Latn': 'Gabon', 'bas': 'Gàbɔ̂ŋ', 'be': 'Габон', 'be-tarask': 'Габон', 'bez': 'Hugaboni', 'bg': 'Габон', 'bgn': 'گابون', 'blo': 'Gabɔŋ', 'bm': 'Gabɔŋ', 'bn': 'গ্যাবন', 'br': 'Gabon', 'brx': 'गाबन', 'bs': 'Gabon', 'bs-Cyrl': 'Габон', 'bs-Latn': 'Gabon', 'ca': 'Gabon', 'ca-ES-valencia': 'Gabon', 'ccp': '𑄉𑄳𑄠𑄝𑄧𑄚𑄴', 'ce': 'Габон', 'ceb': 'Gabon', 'cgg': 'Gabooni', 'chr': 'ᎦᏉᏂ', 'ckb': 'گابۆن', 'cs': 'Gabon', 'cv': 'Габон', 'cy': 'Gabon', 'da': 'Gabon', 'dav': 'Gaboni', 'de': 'Gabun', 'dje': 'Gaabon', 'dsb': 'Gabun', 'dyo': 'Gabon', 'dz': 'གྷ་བྷོན', 'ebu': 'Gaboni', 'ee': 'Gabɔn nutome', 'el': 'Γκαμπόν', 'el-polyton': 'Γκαμπόν', 'en': 'Gabon', 'en-Dsrt': '𐐘𐐲𐐺𐐪𐑌', 'eo': 'Gabono', 'es': 'Gabón', 'et': 'Gabon', 'eu': 'Gabon', 'ewo': 'Gabóŋ', 'fa': 'گابن', 'ff': 'Gaboo', 'ff-Adlm': '𞤘𞤢𞤦𞤮𞤲', 'ff-Latn': 'Gaboo', 'fi': 'Gabon', 'fil': 'Gabon', 'fo': 'Gabon', 'fr': 'Gabon', 'fur': 'Gabon', 'fy': 'Gabon', 'ga': 'an Ghabúin', 'gd': 'Gabon', 'gl': 'Gabón', 'gsw': 'Gabun', 'gu': 'ગેબન', 'guz': 'Gaboni', 'ha': 'Gabon', 'he': 'גבון', 'hi': 'गैबॉन', 'hi-Latn': 'Gabon', 'hr': 'Gabon', 'hsb': 'Gabun', 'ht': 'Gabon', 'hu': 'Gabon', 'hy': 'Գաբոն', 'ia': 'Gabon', 'id': 'Gabon', 'ig': 'Gabon', 'is': 'Gabon', 'it': 'Gabon', 'ja': 'ガボン', 'jgo': 'Ŋgabɔ̂ŋ', 'jmc': 'Gaboni', 'jv': 'Gabon', 'ka': 'გაბონი', 'kaa': 'Габон', 'kaa-Cyrl': 'Габон', 'kab': 'Gabun', 'kam': 'Gaboni', 'kde': 'Gaboni', 'kea': 'Gabon', 'kgp': 'Gabã', 'khq': 'Gaabon', 'ki': 'Gaboni', 'kk': 'Габон', 'kk-Cyrl': 'Габон', 'kln': 'Emetab Gabon', 'km': 'ហ្គាបុង', 'kn': 'ಗೆಬೊನ್', 'ko': '가봉', 'kok': 'गॅबोन', 'kok-Deva': 'गॅबोन', 'ks': 'گیبان', 'ks-Arab': 'گیبان', 'ksb': 'Gaboni', 'ksf': 'gabɔŋ', 'ksh': 'Jabuhn', 'ku': 'Gabon', 'kxv': 'gbon', 'kxv-Deva': 'ग्याबॉन', 'kxv-Latn': 'gbon', 'kxv-Orya': 'ଗାବନ', 'kxv-Telu': 'గాబన్', 'ky': 'Габон', 'lag': 'Gabóoni', 'lb': 'Gabun', 'lg': 'Gaboni', 'lld': 'Gabun', 'ln': 'Gabɔ', 'lo': 'ກາບອນ', 'lt': 'Gabonas', 'lu': 'Ngabu', 'luo': 'Gabon', 'luy': 'Gaboni', 'lv': 'Gabona', 'mai': 'गैबन', 'mas': 'Gaboni', 'mer': 'Gaboni', 'mfe': 'Gabon', 'mg': 'Gabon', 'mgh': 'Ugaboni', 'mi': 'Kāpona', 'mk': 'Габон', 'ml': 'ഗാബൺ', 'mn': 'Габон', 'mr': 'गॅबॉन', 'ms': 'Gabon', 'mt': 'il-Gabon', 'mua': 'Gaboŋ', 'my': 'ဂါဘွန်', 'mzn': 'گابون', 'naq': 'Gaboni', 'nb': 'Gabon', 'nd': 'Gabhoni', 'ne': 'गावोन', 'nl': 'Gabon', 'nmg': 'Gabɔŋ', 'nn': 'Gabon', 'no': 'Gabon', 'nqo': 'ߜ߭ߊߓߐ߲߫', 'nyn': 'Gabooni', 'om': 'Gaaboon', 'or': 'ଗାବୋନ୍', 'pa': 'ਗਬੋਨ', 'pa-Guru': 'ਗਬੋਨ', 'pcm': 'Gabọn', 'pl': 'Gabon', 'ps': 'ګابن', 'pt': 'Gabão', 'qu': 'Gabón', 'rm': 'Gabun', 'rn': 'Gabo', 'ro': 'Gabon', 'rof': 'Gaboni', 'ru': 'Габон', 'rwk': 'Gaboni', 'saq': 'Gaboni', 'sbp': 'Gaboni', 'sc': 'Gabòn', 'sd': 'گبون', 'sd-Arab': 'گبون', 'se': 'Gabon', 'seh': 'Gabão', 'ses': 'Gaabon', 'sg': 'Gaböon', 'shi': 'ⴳⴰⴱⵓⵏ', 'shi-Latn': 'gabun', 'shi-Tfng': 'ⴳⴰⴱⵓⵏ', 'si': 'ගැබොන්', 'sk': 'Gabon', 'sl': 'Gabon', 'smn': 'Gabon', 'sn': 'Gabon', 'so': 'Gaaboon', 'sq': 'Gabon', 'sr': 'Габон', 'sr-Cyrl': 'Габон', 'sr-Latn': 'Gabon', 'sv': 'Gabon', 'sw': 'Gabon', 'syr': 'ܓܒܘܢ', 'ta': 'கேபான்', 'te': 'గేబన్', 'teo': 'Gaboni', 'tg': 'Габон', 'th': 'กาบอง', 'ti': 'ጋቦን', 'tk': 'Gabon', 'to': 'Kaponi', 'tr': 'Gabon', 'tt': 'Габон', 'twq': 'Gaabon', 'tzm': 'Gabun', 'ug': 'گابون', 'uk': 'Габон', 'ur': 'گیبون', 'uz': 'Gabon', 'uz-Cyrl': 'Габон', 'uz-Latn': 'Gabon', 'vai': 'ꕭꕷꘋ', 'vai-Latn': 'Gabhɔŋ', 'vai-Vaii': 'ꕭꕷꘋ', 'vec': 'Gabon', 'vi': 'Gabon', 'vun': 'Gaboni', 'wae': 'Gabon', 'wo': 'Gaboŋ', 'xh': 'EGabon', 'xnr': 'गैबॉन', 'xog': 'Gaboni', 'yav': 'kapɔ́ŋ', 'yi': 'גאַבאן', 'yo': 'Gabon', 'yrl': 'Gabãu', 'yue': '加彭', 'yue-Hans': '加彭', 'yue-Hant': '加彭', 'zgh': 'ⴳⴰⴱⵓⵏ', 'zh': '加蓬', 'zh-Hans': '加蓬', 'zh-Hant': '加彭', 'zu': 'i-Gabon'},
    'GB': {'af': 'Verenigde Koninkryk', 'agq': 'Yùnaetɛ Kiŋdɔ̀m', 'ak': 'UK', 'am': 'ዩናይትድ ኪንግደም', 'ar': 'المملكة المتحدة', 'as': 'সংযুক্ত ৰাজ্য', 'asa': 'Uingeredha', 'ast': 'Reinu Xuníu', 'az': 'Birləşmiş Krallıq', 'az-Cyrl': 'Бирләшмиш Краллыг', 'az-Latn': 'Birləşmiş Krallıq', 'bal-Latn': 'Bartániá', 'bas': 'Àdnà i Lɔ̂ŋ', 'be': 'Вялікабрытанія', 'be-tarask': 'Вялікабрытанія', 'bez': 'Huuingereza', 'bg': 'Обединеното кралство', 'blo': 'Gagɛɛshɩtǝna', 'bm': 'Angilɛtɛri', 'bn': 'যুক্তরাজ্য', 'bo': 'དབྱིན་ཇི་', 'br': 'Rouantelezh-Unanet', 'brx': 'जथाय ब्रिटिश हादोर', 'bs': 'Ujedinjeno Kraljevstvo', 'bs-Cyrl': 'Уједињено Краљевство', 'bs-Latn': 'Ujedinjeno Kraljevstvo', 'ca': 'Regne Unit', 'ca-ES-valencia': 'Regne Unit', 'ccp': '𑄎𑄧𑄙𑄢𑄬𑄌𑄴𑄎𑄮', 'ce': 'Йоккха Британи', 'ceb': 'United Kingdom', 'cgg': 'Bungyereza', 'chr': 'ᎩᎵᏏᏲ', 'ckb': 'شانشینی یەکگرتوو', 'cs': 'Spojené království', 'cv': 'Аслӑ Британи', 'cy': 'Y Deyrnas Unedig', 'da': 'Storbritannien', 'dav': 'Uingereza', 'de': 'Vereinigtes Königreich', 'dje': 'Albaasalaama Marganta', 'doi': 'यूनाइटेड किंगडम', 'dsb': 'Zjadnośone kralejstwo', 'dz': 'ཡུ་ནཱའི་ཊེཌ་ ཀིང་ཌམ', 'ebu': 'Uingereza', 'ee': 'United Kingdom nutome', 'el': 'Ηνωμένο Βασίλειο', 'el-polyton': 'Ἡνωμένο Βασίλειο', 'en': 'United Kingdom', 'en-Dsrt': '𐐏𐐭𐑌𐐴𐐻𐐲𐐼 𐐗𐐨𐑍𐐼𐐲𐑋', 'eo': 'Unuiĝinta Reĝlando', 'es': 'Reino Unido', 'et': 'Ühendkuningriik', 'eu': 'Erresuma Batua', 'ewo': 'Ǹnam Engəlis', 'fa': 'بریتانیا', 'ff': 'Laamateeri Rentundi', 'ff-Adlm': '𞤁𞤫𞤲𞤼𞤢𞤤 𞤐𞤺𞤫𞤯𞤵𞥅𞤪𞤭', 'ff-Latn': 'Laamateeri Rentundi', 'fi': 'Iso-Britannia', 'fil': 'United Kingdom', 'fo': 'Stórabretland', 'fr': 'Royaume-Uni', 'fur': 'Ream unît', 'fy': 'Verenigd Koninkrijk', 'ga': 'an Ríocht Aontaithe', 'gd': 'An Rìoghachd Aonaichte', 'gl': 'Reino Unido', 'gsw': 'Veräinigts Chönigriich', 'gu': 'યુનાઇટેડ કિંગડમ', 'guz': 'Uingereza', 'gv': 'Rywvaneth Unys', 'ha': 'Biritaniya', 'haw': 'Aupuni Mōʻī Hui Pū ʻIa', 'he': 'בריטניה', 'hi': 'यूनाइटेड किंगडम', 'hi-Latn': 'United Kingdom', 'hr': 'Ujedinjeno Kraljevstvo', 'hsb': 'Zjednoćene kralestwo', 'ht': 'Royaume-Uni', 'hu': 'Egyesült Királyság', 'hy': 'Միացյալ Թագավորություն', 'ia': 'Regno Unite', 'id': 'Inggris Raya', 'ie': 'Unit Reyia', 'ig': 'United Kingdom', 'ii': 'ꑱꇩ', 'is': 'Bretland', 'it': 'Regno Unito', 'ja': 'イギリス', 'jmc': 'Uingereza', 'jv': 'Karajan Manunggal', 'ka': 'გაერთიანებული სამეფო', 'kaa': 'Уллы Британия', 'kaa-Cyrl': 'Уллы Британия', 'kab': 'Tagelda Yedduklen', 'kam': 'Uingereza', 'kde': 'Nngalesa', 'kea': 'Reinu Unidu', 'kgp': 'Rejnũ Unĩnu', 'khq': 'Albaasalaama Marganta', 'ki': 'Ngeretha', 'kk': 'Ұлыбритания', 'kk-Cyrl': 'Ұлыбритания', 'kln': 'Emetab Kibagenge nebo Uingereza', 'km': 'ចក្រភព\u200bអង់គ្លេស', 'kn': 'ಯುನೈಟೆಡ್ ಕಿಂಗ್\u200cಡಮ್', 'ko': '영국', 'kok': 'युनायटेड किंगडम', 'kok-Deva': 'युनायटेड किंगडम', 'ks': 'متحدہ مملِکت', 'ks-Arab': 'متحدہ مملِکت', 'ks-Deva': 'मुतहीद बादशाहत', 'ksb': 'Uingeeza', 'ksf': 'kǝlɔŋ kǝ kǝtáatáaŋzǝn', 'ksh': 'Jruußbrettannije', 'ku': 'Qiralîyeta Yekbûyî', 'kw': 'Rywvaneth Unys', 'kxv': 'uniṭeḍ kingḍom', 'kxv-Deva': 'यूनाइटेड किंगडम', 'kxv-Latn': 'uniṭeḍ kingḍom', 'kxv-Orya': 'ଇଉନାଇଟେଡ୍ କିଂଡମ୍', 'kxv-Telu': 'యునైటెడ్ కింగ్\u200cడమ్', 'ky': 'Улуу Британия', 'lag': 'Ʉɨngeréesa', 'lb': 'Groussbritannien', 'lg': 'Bungereza', 'lld': 'Rëgn Uní', 'ln': 'Angɛlɛtɛ́lɛ', 'lo': 'ສະຫະລາດຊະອະນາຈັກ', 'lrc': 'بیریتانیا گأپ', 'lt': 'Jungtinė Karalystė', 'lu': 'Angeletele', 'luo': 'United Kingdom', 'luy': 'Uingereza', 'lv': 'Apvienotā Karaliste', 'mai': 'यूनाइटेड किंगडम', 'mas': 'Uingereza', 'mer': 'Ngeretha', 'mfe': 'United Kingdom', 'mg': 'Angletera', 'mi': 'Te Hononga o Piritene', 'mk': 'Обединето Кралство', 'ml': 'യുണൈറ്റഡ് കിംഗ്ഡം', 'mn': 'Их Британи', 'mni': 'য়ুনাইটেদ কিংদম', 'mni-Beng': 'য়ুনাইটেদ কিংদম', 'mr': 'युनायटेड किंगडम', 'ms': 'United Kingdom', 'mt': 'ir-Renju Unit', 'mua': 'Sǝr Anglofoŋ', 'my': 'ယူနိုက်တက်ကင်းဒမ်း', 'mzn': 'بریتانیا', 'naq': 'United Kingdom', 'nb': 'Storbritannia', 'nd': 'United Kingdom', 'ne': 'संयुक्त अधिराज्य', 'nl': 'Verenigd Koninkrijk', 'nmg': 'Nlambɔ́ Ngɛlɛn', 'nn': 'Storbritannia', 'no': 'Storbritannia', 'nqo': 'ߡߊ߲߬ߛߊ߬ߟߊ߫ ߟߊߘߍ߬ߣߍ߲', 'nyn': 'Bungyereza', 'om': 'United Kingdom', 'or': 'ଯୁକ୍ତରାଜ୍ୟ', 'os': 'Стыр Британи', 'pa': 'ਯੂਨਾਈਟਡ ਕਿੰਗਡਮ', 'pa-Guru': 'ਯੂਨਾਈਟਡ ਕਿੰਗਡਮ', 'pcm': 'Yunáítẹ́d Kíndọm', 'pl': 'Wielka Brytania', 'ps': 'برتانیه', 'pt': 'Reino Unido', 'qu': 'Reino Unido', 'rm': 'Reginavel Unì', 'rn': 'Ubwongereza', 'ro': 'Regatul Unit', 'rof': 'Uingereza', 'ru': 'Великобритания', 'rwk': 'Uingereza', 'sa': 'संयुक्त राष्ट्र:', 'sah': 'Улуу Британия', 'saq': 'Uingereza', 'sat': 'ᱭᱩᱱᱤᱭᱴᱮᱰ ᱠᱤᱝᱰᱚᱢ', 'sat-Olck': 'ᱭᱩᱱᱤᱭᱴᱮᱰ ᱠᱤᱝᱰᱚᱢ', 'sbp': 'Uwingelesa', 'sc': 'Regnu Unidu', 'sd': 'برطانيہ', 'sd-Arab': 'برطانيہ', 'sd-Deva': 'बरतानी', 'se': 'Stuorra-Británnia', 'seh': 'Reino Unido', 'ses': 'Albaasalaama Marganta', 'sg': 'Ködörögbïä--Ôko', 'shi': 'ⵜⴰⴳⵍⴷⵉⵜ ⵉⵎⵓⵏⵏ', 'shi-Latn': 'tagldit imunn', 'shi-Tfng': 'ⵜⴰⴳⵍⴷⵉⵜ ⵉⵎⵓⵏⵏ', 'si': 'එක්සත් රාජධානිය', 'sk': 'Spojené kráľovstvo', 'sl': 'Združeno kraljestvo', 'smn': 'Ovtâstum Kunâgâskodde', 'sn': 'United Kingdom', 'so': 'Boqortooyada Midowday', 'sq': 'Mbretëria e Bashkuar', 'sr': 'Уједињено Краљевство', 'sr-Cyrl': 'Уједињено Краљевство', 'sr-Latn': 'Ujedinjeno Kraljevstvo', 'su': 'Britania Raya', 'su-Latn': 'Britania Raya', 'sv': 'Storbritannien', 'sw': 'Ufalme wa Muungano', 'syr': 'ܡܠܟܘܬܐ ܡܚܝܕܬܐ', 'ta': 'யுனைடெட் கிங்டம்', 'te': 'యునైటెడ్ కింగ్\u200cడమ్', 'teo': 'Uingereza', 'tg': 'Шоҳигарии Муттаҳида', 'th': 'สหราชอาณาจักร', 'ti': 'ብሪጣንያ', 'tk': 'Birleşen Patyşalyk', 'to': 'Pilitānia', 'tpi': 'Yunaited Kingdom', 'tr': 'Birleşik Krallık', 'tt': 'Берләшкән Корольлек', 'twq': 'Albaasalaama Marganta', 'tzm': 'Tagelda Taddukelt', 'ug': 'بىرلەشمە پادىشاھلىق', 'uk': 'Велика Британія', 'ur': 'سلطنت متحدہ', 'uz': 'Buyuk Britaniya', 'uz-Cyrl': 'Буюк Британия', 'uz-Latn': 'Buyuk Britaniya', 'vai': 'ꖕꕯꔤꗳ', 'vai-Latn': 'Yunaitɛ Kíŋdɔŋ', 'vai-Vaii': 'ꖕꕯꔤꗳ', 'vec': 'Regno Unìo', 'vi': 'Vương quốc Anh', 'vun': 'Uingereza', 'wae': 'England', 'wo': 'Ruwaayom Ini', 'xh': 'E-United Kingdom', 'xnr': 'यूनाइटेड किंगडम', 'xog': 'Bungereza', 'yav': 'ingilíís', 'yi': 'פֿאַראייניגטע קעניגרייך', 'yo': 'Gẹ̀ẹ́sì', 'yrl': 'Reyinu Yepewasú', 'yue': '英國', 'yue-Hans': '英国', 'yue-Hant': '英國', 'zgh': 'ⵜⴰⴳⵍⴷⵉⵜ ⵉⵎⵓⵏⵏ', 'zh': '英国', 'zh-Hans': '英国', 'zh-Hant': '英國', 'zu': 'i-United Kingdom'},
    'GD': {'af': 'Grenada', 'agq': 'Gɨ̀lɛnadà', 'ak': 'Grenada', 'am': 'ግሬናዳ', 'ar': 'غرينادا', 'as': 'গ্ৰেনাডা', 'asa': 'Grenada', 'ast': 'Granada', 'az': 'Qrenada', 'az-Cyrl': 'Гренада', 'az-Latn': 'Qrenada', 'bal-Latn': 'Gerená', 'bas': 'Grènadà', 'be': 'Грэнада', 'be-tarask': 'Грэнада', 'bez': 'Hugrenada', 'bg': 'Гренада', 'blo': 'Grenaɖaa', 'bm': 'Granadi', 'bn': 'গ্রেনাডা', 'br': 'Grenada', 'brx': 'ग्रेनादा', 'bs': 'Grenada', 'bs-Cyrl': 'Гренада', 'bs-Latn': 'Grenada', 'ca': 'Grenada', 'ca-ES-valencia': 'Grenada', 'ccp': '𑄉𑄳𑄢𑄬𑄚𑄓', 'ce': 'Гренада', 'ceb': 'Grenada', 'cgg': 'Gurenada', 'chr': 'ᏋᎾᏓ', 'ckb': 'گرینادا', 'cs': 'Grenada', 'cv': 'Гренада', 'cy': 'Grenada', 'da': 'Grenada', 'dav': 'Grenada', 'de': 'Grenada', 'dje': 'Grenaada', 'dsb': 'Grenada', 'dyo': 'Grenada', 'dz': 'གྲྀ་ན་ཌ', 'ebu': 'Grenada', 'ee': 'Grenada nutome', 'el': 'Γρενάδα', 'el-polyton': 'Γρενάδα', 'en': 'Grenada', 'en-Dsrt': '𐐘𐑉𐐲𐑌𐐩𐐼𐐲', 'eo': 'Grenado', 'es': 'Granada', 'et': 'Grenada', 'eu': 'Grenada', 'ewo': 'Gələnádə', 'fa': 'گرنادا', 'ff': 'Garnaad', 'ff-Adlm': '𞤘𞤢𞤪𞤲𞤢𞤣𞤢𞥄', 'ff-Latn': 'Garnaad', 'fi': 'Grenada', 'fil': 'Grenada', 'fo': 'Grenada', 'fr': 'Grenade', 'fur': 'Grenada', 'fy': 'Grenada', 'ga': 'Greanáda', 'gd': 'Greanàda', 'gl': 'Granada', 'gsw': 'Grenada', 'gu': 'ગ્રેનેડા', 'guz': 'Grenada', 'ha': 'Girnada', 'he': 'גרנדה', 'hi': 'ग्रेनाडा', 'hi-Latn': 'Grenada', 'hr': 'Grenada', 'hsb': 'Grenada', 'ht': 'Grenade', 'hu': 'Grenada', 'hy': 'Գրենադա', 'ia': 'Grenada', 'id': 'Grenada', 'ig': 'Grenada', 'is': 'Grenada', 'it': 'Grenada', 'ja': 'グレナダ', 'jmc': 'Grenada', 'jv': 'Grénada', 'ka': 'გრენადა', 'kaa': 'Гренада', 'kaa-Cyrl': 'Гренада', 'kab': 'Grunad', 'kam': 'Grenada', 'kde': 'Glenada', 'kea': 'Granada', 'kgp': 'Granỹna', 'khq': 'Grenaada', 'ki': 'Grenada', 'kk': 'Гренада', 'kk-Cyrl': 'Гренада', 'kln': 'Emetab Grenada', 'km': 'ហ្គ្រើណាដ', 'kn': 'ಗ್ರೆನೆಡಾ', 'ko': '그레나다', 'kok': 'ग्रेनॅडा', 'kok-Deva': 'ग्रेनॅडा', 'ks': 'گرینیڈا', 'ks-Arab': 'گرینیڈا', 'ksb': 'Glenada', 'ksf': 'grɛnadǝ', 'ksh': 'Jrenahda', 'ku': 'Grenada', 'kxv': 'graneida', 'kxv-Deva': 'ग्रेनाडा', 'kxv-Latn': 'graneida', 'kxv-Orya': 'ଗ୍ରେନାଡା', 'kxv-Telu': 'గ్రెనడా', 'ky': 'Гренада', 'lag': 'Girenáada', 'lb': 'Grenada', 'lg': 'Gurenada', 'lld': 'Grenada', 'ln': 'Gelenadɛ', 'lo': 'ເກຣເນດາ', 'lt': 'Grenada', 'lu': 'Ngelenade', 'luo': 'Grenada', 'luy': 'Grenada', 'lv': 'Grenāda', 'mai': 'ग्रेनाडा', 'mas': 'Grenada', 'mer': 'Grenada', 'mfe': 'Grenad', 'mg': 'Grenady', 'mgh': 'Ugrenada', 'mi': 'Kerenāta', 'mk': 'Гренада', 'ml': 'ഗ്രനേഡ', 'mn': 'Гренада', 'mr': 'ग्रेनेडा', 'ms': 'Grenada', 'mt': 'Grenada', 'mua': 'Grenadǝ', 'my': 'ဂရီနေဒါ', 'mzn': 'گرانادا', 'naq': 'Grenada', 'nb': 'Grenada', 'nd': 'Grenada', 'ne': 'ग्रेनाडा', 'nl': 'Grenada', 'nmg': 'Grenada', 'nn': 'Grenada', 'no': 'Grenada', 'nqo': 'ߜ߭ߙߋߣߊߘ', 'nyn': 'Gurenada', 'om': 'Girinaada', 'or': 'ଗ୍ରେନାଡା', 'pa': 'ਗ੍ਰੇਨਾਡਾ', 'pa-Guru': 'ਗ੍ਰੇਨਾਡਾ', 'pcm': 'Grẹnéda', 'pl': 'Grenada', 'ps': 'ګرنادا', 'pt': 'Granada', 'qu': 'Granada', 'rm': 'Grenada', 'rn': 'Gerenada', 'ro': 'Grenada', 'rof': 'Grenada', 'ru': 'Гренада', 'rwk': 'Grenada', 'saq': 'Grenada', 'sbp': 'Gilenada', 'sc': 'Grenada', 'sd': 'گريناڊا', 'sd-Arab': 'گريناڊا', 'se': 'Grenada', 'seh': 'Granada', 'ses': 'Grenaada', 'sg': 'Grenâda', 'shi': 'ⵖⵔⵏⴰⵟⴰ', 'shi-Latn': 'ɣrnaṭa', 'shi-Tfng': 'ⵖⵔⵏⴰⵟⴰ', 'si': 'ග්\u200dරැනඩාව', 'sk': 'Grenada', 'sl': 'Grenada', 'smn': 'Grenada', 'sn': 'Grenada', 'so': 'Giriinaada', 'sq': 'Granadë', 'sr': 'Гренада', 'sr-Cyrl': 'Гренада', 'sr-Latn': 'Grenada', 'sv': 'Grenada', 'sw': 'Grenada', 'syr': 'ܓܪܝܢܐܕܐ', 'ta': 'கிரனெடா', 'te': 'గ్రెనడా', 'teo': 'Grenada', 'tg': 'Гренада', 'th': 'เกรเนดา', 'ti': 'ግረናዳ', 'tk': 'Grenada', 'to': 'Kelenatā', 'tr': 'Grenada', 'tt': 'Гренада', 'twq': 'Grenaada', 'tzm': 'Grinada', 'ug': 'گىرېنادا', 'uk': 'Гренада', 'ur': 'گریناڈا', 'uz': 'Grenada', 'uz-Cyrl': 'Гренада', 'uz-Latn': 'Grenada', 'vai': 'ꖶꕟꕯꕜ', 'vai-Latn': 'Gurinéda', 'vai-Vaii': 'ꖶꕟꕯꕜ', 'vec': 'Granada', 'vi': 'Grenada', 'vun': 'Grenada', 'wae': 'Grenada', 'wo': 'Garanad', 'xh': 'EGrenada', 'xnr': 'ग्रेनेडा', 'xog': 'Gurenada', 'yav': 'kelenáat', 'yi': 'גרענאַדאַ', 'yo': 'Genada', 'yrl': 'Garanada', 'yue': '格瑞那達', 'yue-Hans': '格瑞那达', 'yue-Hant': '格瑞那達', 'zgh': 'ⵖⵔⵏⴰⵟⴰ', 'zh': '格林纳达', 'zh-Hans': '格林纳达', 'zh-Hant': '格瑞那達', 'zu': 'i-Grenada'},
    'GE': {'af': 'Georgië', 'agq': 'Dzɔɔdzìa', 'ak': 'Gyɔgyea', 'am': 'ጆርጂያ', 'ar': 'جورجيا', 'as': 'জৰ্জিয়া', 'asa': 'Jojia', 'ast': 'Xeorxa', 'az': 'Gürcüstan', 'az-Cyrl': 'Ҝүрҹүстан', 'az-Latn': 'Gürcüstan', 'bal-Latn': 'Járjiá', 'bas': 'Gèɔrgìà', 'be': 'Грузія', 'be-tarask': 'Грузія', 'bez': 'Hujojia', 'bg': 'Грузия', 'bgn': 'گرجستان', 'blo': 'Jɔrjiya', 'bm': 'Zeyɔrzi', 'bn': 'জর্জিয়া', 'br': 'Jorjia', 'brx': 'जर्जिया', 'bs': 'Gruzija', 'bs-Cyrl': 'Грузија', 'bs-Latn': 'Gruzija', 'ca': 'Geòrgia', 'ca-ES-valencia': 'Geòrgia', 'ccp': '𑄎𑄧𑄢𑄴𑄎𑄨𑄠', 'ce': 'Гуьржийчоь', 'ceb': 'Georgia', 'cgg': 'Gyogiya', 'chr': 'ᏣᎠᏥᎢ', 'ckb': 'گورجستان', 'cs': 'Gruzie', 'cv': 'Грузи', 'cy': 'Georgia', 'da': 'Georgien', 'dav': 'Jojia', 'de': 'Georgien', 'dje': 'Gorgi', 'dsb': 'Georgiska', 'dyo': 'Seorsi', 'dz': 'ཇཽར་ཇཱ', 'ebu': 'Jojia', 'ee': 'Georgia nutome', 'el': 'Γεωργία', 'el-polyton': 'Γεωργία', 'en': 'Georgia', 'en-Dsrt': '𐐖𐐱𐑉𐐾𐐲', 'eo': 'Kartvelujo', 'es': 'Georgia', 'et': 'Gruusia', 'eu': 'Georgia', 'ewo': 'Horə́yia', 'fa': 'گرجستان', 'ff': 'Jeorgii', 'ff-Adlm': '𞤔𞤮𞤪𞤶𞤭𞤴𞤢𞥄', 'ff-Latn': 'Jeorgii', 'fi': 'Georgia', 'fil': 'Georgia', 'fo': 'Georgia', 'fr': 'Géorgie', 'fur': 'Gjeorgjie', 'fy': 'Georgië', 'ga': 'an tSeoirsia', 'gd': 'A’ Chairtbheil', 'gl': 'Xeorxia', 'gsw': 'Geoorgie', 'gu': 'જ્યોર્જિયા', 'guz': 'Jojia', 'ha': 'Jojiya', 'he': 'גאורגיה', 'hi': 'जॉर्जिया', 'hi-Latn': 'Georgia', 'hr': 'Gruzija', 'hsb': 'Georgiska', 'ht': 'Géorgie', 'hu': 'Grúzia', 'hy': 'Վրաստան', 'ia': 'Georgia', 'id': 'Georgia', 'ig': 'Georgia', 'is': 'Georgía', 'it': 'Georgia', 'ja': 'ジョージア', 'jmc': 'Jojia', 'jv': 'Géorgia', 'ka': 'საქართველო', 'kaa': 'Грузия', 'kaa-Cyrl': 'Грузия', 'kab': 'Jiyurji', 'kam': 'Jojia', 'kde': 'Dyodya', 'kea': 'Jiórjia', 'kgp': 'Jiórja', 'khq': 'Gorgi', 'ki': 'Jojia', 'kk': 'Грузия', 'kk-Cyrl': 'Грузия', 'kln': 'Emetab Georgia', 'km': 'ហ្សកហ្ស៊ី', 'kn': 'ಜಾರ್ಜಿಯಾ', 'ko': '조지아', 'kok': 'जॉर्जिया', 'kok-Deva': 'जॉर्जिया', 'ks': 'جارجِیا', 'ks-Arab': 'جارجِیا', 'ksb': 'Jojia', 'ksf': 'jɔrjí', 'ksh': 'Jeorrjije', 'ku': 'Gurcistan', 'kxv': 'jeorjiaa', 'kxv-Deva': 'जॉर्जिया', 'kxv-Latn': 'jeorjiaa', 'kxv-Orya': 'ଜର୍ଜିଆ', 'kxv-Telu': 'జార్జియా', 'ky': 'Грузия', 'lag': 'Jójia', 'lb': 'Georgien', 'lg': 'Gyogya', 'lld': 'Georgia', 'ln': 'Zorzi', 'lo': 'ຈໍເຈຍ', 'lt': 'Gruzija', 'lu': 'Joriji', 'luo': 'Georgia', 'luy': 'Jojia', 'lv': 'Gruzija', 'mai': 'जार्जिया', 'mas': 'Jojia', 'mer': 'Jojia', 'mfe': 'Zeorzi', 'mg': 'Zeorzia', 'mgh': 'Ujojia', 'mi': 'Hōria', 'mk': 'Грузија', 'ml': 'ജോർജ്ജിയ', 'mn': 'Гүрж', 'mr': 'जॉर्जिया', 'ms': 'Georgia', 'mt': 'il-Georgia', 'mua': 'Georgiya', 'my': 'ဂျော်ဂျီယာ', 'mzn': 'گرجستون', 'naq': 'Georgiab', 'nb': 'Georgia', 'nd': 'Georgia', 'ne': 'जर्जिया', 'nl': 'Georgië', 'nmg': 'Jɔrgia', 'nn': 'Georgia', 'no': 'Georgia', 'nqo': 'ߖ߭ߋߐߙߑߖ߭ߌ߫', 'nyn': 'Gyogiya', 'om': 'Joorjiyaa', 'or': 'ଜର୍ଜିଆ', 'os': 'Гуырдзыстон', 'pa': 'ਜਾਰਜੀਆ', 'pa-Guru': 'ਜਾਰਜੀਆ', 'pcm': 'Jọ́jia', 'pl': 'Gruzja', 'ps': 'گورجستان', 'pt': 'Geórgia', 'qu': 'Georgia', 'rm': 'Georgia', 'rn': 'Jeworujiya', 'ro': 'Georgia', 'rof': 'Jojia', 'ru': 'Грузия', 'rwk': 'Jojia', 'saq': 'Jojia', 'sbp': 'Jojiya', 'sc': 'Geòrgia', 'sd': 'جارجيا', 'sd-Arab': 'جارجيا', 'se': 'Georgia', 'seh': 'Geórgia', 'ses': 'Gorgi', 'sg': 'Zorzïi', 'shi': 'ⵊⵓⵔⵊⵢⴰ', 'shi-Latn': 'jurjya', 'shi-Tfng': 'ⵊⵓⵔⵊⵢⴰ', 'si': 'ජෝර්ජියාව', 'sk': 'Gruzínsko', 'sl': 'Gruzija', 'smn': 'Georgia', 'sn': 'Georgia', 'so': 'Joorjiya', 'sq': 'Gjeorgji', 'sr': 'Грузија', 'sr-Cyrl': 'Грузија', 'sr-Latn': 'Gruzija', 'sv': 'Georgien', 'sw': 'Jojia', 'syr': 'ܓܘܪܓܝܐ', 'ta': 'ஜார்ஜியா', 'te': 'జార్జియా', 'teo': 'Jojia', 'tg': 'Гурҷистон', 'th': 'จอร์เจีย', 'ti': 'ጆርጅያ', 'tk': 'Gruziýa', 'to': 'Seōsia', 'tr': 'Gürcistan', 'tt': 'Грузия', 'twq': 'Gorgi', 'tzm': 'Jyurjya', 'ug': 'گىرۇزىيە', 'uk': 'Грузія', 'ur': 'جارجیا', 'uz': 'Gruziya', 'uz-Cyrl': 'Грузия', 'uz-Latn': 'Gruziya', 'vai': 'ꗘꖺꕀꕩ', 'vai-Latn': 'Jɔɔjiya', 'vai-Vaii': 'ꗘꖺꕀꕩ', 'vec': 'Jeorja', 'vi': 'Georgia', 'vun': 'Jojia', 'wae': 'Georgie', 'wo': 'Seworsi', 'xh': 'EGeorgia', 'xnr': 'जॉर्जिया', 'xog': 'Gyogya', 'yav': 'sɔlsíi', 'yi': 'גרוזיע', 'yo': 'Gọgia', 'yrl': 'Geugiya', 'yue': '格魯吉亞', 'yue-Hans': '乔治亚共和国', 'yue-Hant': '格魯吉亞', 'zgh': 'ⵊⵓⵔⵊⵢⴰ', 'zh': '格鲁吉亚', 'zh-Hans': '格鲁吉亚', 'zh-Hant': '喬治亞', 'zu': 'i-Georgia'},
    'GF': {'af': 'Frans-Guyana', 'agq': 'Gàyanà è Fàlâŋnsì', 'ak': 'Frɛnkye Gayana', 'am': 'የፈረንሳይ ጉዊአና', 'ar': 'غويانا الفرنسية', 'as': 'ফ্ৰান্স গয়ানা', 'asa': 'Gwiyana ya Ufarantha', 'ast': 'Guyana Francesa', 'az': 'Fransa Qvianası', 'az-Cyrl': 'Франса Гвианасы', 'az-Latn': 'Fransa Qvianası', 'bal-Latn': 'Pránsi Gwiáná', 'bas': 'Gùyanà Pùlàsi', 'be': 'Французская Гвіяна', 'be-tarask': 'Французская Гвіяна', 'bez': 'Hugwiyana ya Huufaransa', 'bg': 'Френска Гвиана', 'blo': 'Guyanaa Gafɔntǝna kaja', 'bm': 'Faransi ka gwiyani', 'bn': 'ফরাসী গায়ানা', 'br': 'Gwiana cʼhall', 'brx': 'फ्रेन्च गिआना', 'bs': 'Francuska Gvajana', 'bs-Cyrl': 'Француска Гвајана', 'bs-Latn': 'Francuska Gvajana', 'ca': 'Guaiana Francesa', 'ca-ES-valencia': 'Guaiana Francesa', 'ccp': '𑄜𑄧𑄢𑄥𑄩 𑄉𑄠𑄚', 'ce': 'Французийн Гвиана', 'ceb': 'French Guiana', 'cgg': 'Guyana ya Bufaransa', 'chr': 'ᎠᏂᎦᎸᏥ ᎩᎠ', 'ckb': 'گیانای فەرەنسا', 'cs': 'Francouzská Guyana', 'cv': 'Франци Гвиана', 'cy': 'Guyane Ffrengig', 'da': 'Fransk Guyana', 'dav': 'Gwiyana ya Ufaransa', 'de': 'Französisch-Guayana', 'dje': 'Faransi Guyaan', 'dsb': 'Francojska Guyana', 'dz': 'གུའི་ཡ་ན་ ཕྲནས྄་མངའ་ཁོངས', 'ebu': 'Gwiyana ya Ufaransa', 'ee': 'Frentsi Gayana nutome', 'el': 'Γαλλική Γουιάνα', 'el-polyton': 'Γαλλικὴ Γουιάνα', 'en': 'French Guiana', 'en-Dsrt': '𐐙𐑉𐐯𐑌𐐽 𐐘𐐨𐐪𐑌𐐲', 'eo': 'Franca Gviano', 'es': 'Guayana Francesa', 'et': 'Prantsuse Guajaana', 'eu': 'Guyana Frantsesa', 'ewo': 'Guyán yá Fulɛnsí', 'fa': 'گویان فرانسه', 'ff': 'Giyaan Farayse', 'ff-Adlm': '𞤘𞤵𞤴𞤢𞥄𞤲 𞤊𞤪𞤢𞤲𞤧𞤭𞤲𞤳𞤮', 'ff-Latn': 'Giyaan Farayse', 'fi': 'Ranskan Guayana', 'fil': 'French Guiana', 'fo': 'Franska Gujana', 'fr': 'Guyane française', 'fur': 'Guiana francês', 'fy': 'Frans-Guyana', 'ga': 'Guáin na Fraince', 'gd': 'Guidheàna na Frainge', 'gl': 'Güiana Francesa', 'gsw': 'Französisch-Guäjaana', 'gu': 'ફ્રેંચ ગયાના', 'guz': 'Gwiyana ya Ufaransa', 'ha': 'Gini Ta Faransa', 'he': 'גיאנה הצרפתית', 'hi': 'फ़्रेंच गुयाना', 'hi-Latn': 'French Guiana', 'hr': 'Francuska Gijana', 'hsb': 'Francoska Guyana', 'ht': 'Guyane française', 'hu': 'Francia Guyana', 'hy': 'Ֆրանսիական Գվիանա', 'ia': 'Guyana francese', 'id': 'Guyana Prancis', 'ig': 'French Guiana', 'is': 'Franska Gvæjana', 'it': 'Guyana Francese', 'ja': '仏領ギアナ', 'jmc': 'Gwiyana ya Ufaransa', 'jv': 'Guyana Prancis', 'ka': 'საფრანგეთის გვიანა', 'kaa': 'Француз Гвианасы', 'kaa-Cyrl': 'Француз Гвианасы', 'kab': 'Ɣana tafransist', 'kam': 'Gwiyana ya Ufaransa', 'kde': 'Gwiyana ya Ufalansa', 'kea': 'Giana Franseza', 'kgp': 'Frỹsa Gijanỹ', 'khq': 'Faransi Guyaan', 'ki': 'Gwiyana ya Ufaransa', 'kk': 'Француз Гвианасы', 'kk-Cyrl': 'Француз Гвианасы', 'kln': 'Emetab Guiana nebo Ufaransa', 'km': 'ហ្គីអាណា បារាំង', 'kn': 'ಫ್ರೆಂಚ್ ಗಯಾನಾ', 'ko': '프랑스령 기아나', 'kok': 'फ्रेन्च गयाना', 'kok-Deva': 'फ्रेन्च गयाना', 'ks': 'فرانسِسی گِانا', 'ks-Arab': 'فرانسِسی گِانا', 'ksb': 'Gwiyana ya Ufalansa', 'ksf': 'guyán i pɛrɛsǝ́', 'ksh': 'Franzüüsesch Jujaana', 'ku': 'Guyanaya Fransî', 'kxv': 'prenc guiyana', 'kxv-Deva': 'प़्रेंच गुयाना', 'kxv-Latn': 'prenc guiyana', 'kxv-Orya': 'ପ୍ରେଞ୍ଚ୍ ଗୁୟାନା', 'kxv-Telu': 'ప్రెంచ్ గుయానా', 'ky': 'Франция Гвианасы', 'lag': 'Gwiyáana yʉ Ʉfaráansa', 'lb': 'Guayane', 'lg': 'Guyana enfalansa', 'lld': 'Guiana franzeja', 'ln': 'Giyanɛ ya Falánsɛ', 'lo': 'ເຟຣນຊ໌ ກຸຍອານາ', 'lt': 'Prancūzijos Gviana', 'lu': 'Giyane wa Nfalanse', 'luo': 'French Guiana', 'luy': 'Gwiyana ya Ufaransa', 'lv': 'Francijas Gviāna', 'mai': 'फ्रेंच गुयाना', 'mas': 'Gwiyana ya Ufaransa', 'mer': 'Gwiyana ya Fransi', 'mfe': 'Gwiyann franse', 'mg': 'Guyana frantsay', 'mgh': 'Ufaransa yo Gwaya', 'mi': 'Kiāna Wīwī', 'mk': 'Француска Гвајана', 'ml': 'ഫ്രഞ്ച് ഗയാന', 'mn': 'Францын Гвиана', 'mr': 'फ्रेंच गयाना', 'ms': 'Guiana Perancis', 'mt': 'il-Guyana Franċiża', 'mua': 'Sǝr Guyana ma Franssǝ', 'my': 'ပြင်သစ် ဂီယာနာ', 'mzn': 'فرانسه\u200cی ِگویان', 'naq': 'French Guiana', 'nb': 'Fransk Guyana', 'nd': 'Gwiyana ye Furansi', 'ne': 'फ्रान्सेली गायना', 'nl': 'Frans-Guyana', 'nmg': 'Guyane Fala', 'nn': 'Fransk Guyana', 'no': 'Fransk Guyana', 'nqo': 'ߝߊ߲߬ߙߊ߲߬ߛߌ߫ ߜ߭ߎ߳ߦߊߣ', 'nyn': 'Guyana ya Bufaransa', 'om': 'Faransaay Guyiinaa', 'or': 'ଫ୍ରେଞ୍ଚ ଗୁଇନା', 'pa': 'ਫਰੈਂਚ ਗੁਇਆਨਾ', 'pa-Guru': 'ਫਰੈਂਚ ਗੁਇਆਨਾ', 'pcm': 'Frẹ́nch Giána', 'pl': 'Gujana Francuska', 'ps': 'فرانسوي ګانا', 'pt': 'Guiana Francesa', 'qu': 'Guayana Francesa', 'rm': 'Guyana Franzosa', 'rn': 'Gwayana y’Abafaransa', 'ro': 'Guyana Franceză', 'rof': 'Gwiyana ya Ufaransa', 'ru': 'Французская Гвиана', 'rwk': 'Gwiyana ya Ufaransa', 'saq': 'Gwiyana ya Ufaransa', 'sbp': 'Gwiyana ya Wufalansa', 'sc': 'Guiana frantzesa', 'sd': 'فرانسيسي گيانا', 'sd-Arab': 'فرانسيسي گيانا', 'se': 'Frankriikka Guayana', 'seh': 'Guiana Francesa', 'ses': 'Faransi Guyaan', 'sg': 'Güyâni tî farânzi', 'shi': 'ⴳⵡⵉⵢⴰⵏ ⵜⴰⴼⵔⴰⵏⵙⵉⵙⵜ', 'shi-Latn': 'gwiyan tafransist', 'shi-Tfng': 'ⴳⵡⵉⵢⴰⵏ ⵜⴰⴼⵔⴰⵏⵙⵉⵙⵜ', 'si': 'ප්\u200dරංශ ගයනාව', 'sk': 'Francúzska Guyana', 'sl': 'Francoska Gvajana', 'smn': 'Ranska Guyana', 'sn': 'French Guiana', 'so': 'Faransiis Gini', 'sq': 'Guajana Franceze', 'sr': 'Француска Гвајана', 'sr-Cyrl': 'Француска Гвајана', 'sr-Latn': 'Francuska Gvajana', 'sv': 'Franska Guyana', 'sw': 'Guiana ya Ufaransa', 'syr': 'ܓܘܝܐܢܐ ܦܪܢܣܝܬܐ', 'ta': 'பிரெஞ்சு கயானா', 'te': 'ఫ్రెంచ్ గియానా', 'teo': 'Gwiyana ya Ufaransa', 'tg': 'Гвианаи Фаронса', 'th': 'เฟรนช์เกียนา', 'ti': 'ፈረንሳዊት ጊያና', 'tk': 'Fransuz Gwianasy', 'to': 'Kuiana fakafalanisē', 'tr': 'Fransız Guyanası', 'tt': 'Француз Гвианасы', 'twq': 'Faransi Guyaan', 'tzm': 'Guyana Tafransist', 'ug': 'فىرانسىيەگە قاراشلىق گىۋىيانا', 'uk': 'Французька Гвіана', 'ur': 'فرینچ گیانا', 'uz': 'Fransuz Gvianasi', 'uz-Cyrl': 'Француз Гвианаси', 'uz-Latn': 'Fransuz Gvianasi', 'vai': 'ꗱꘋꔻ ꖶꕎꕯ', 'vai-Latn': 'Fɛŋsi Giwana', 'vai-Vaii': 'ꗱꘋꔻ ꖶꕎꕯ', 'vec': 'Guyana franseze', 'vi': 'Guiana thuộc Pháp', 'vun': 'Gwiyana ya Ufaransa', 'wae': 'Französiš Guiana', 'wo': 'Guyaan Farañse', 'xh': 'EFrench Guiana', 'xnr': 'फ़्रेंच गुयाना', 'xog': 'Guyana enfalansa', 'yav': 'kuyáan u felensí', 'yi': 'פֿראַנצויזישע גויאַנע', 'yo': 'Firenṣi Guana', 'yrl': 'Giyana Frãsa yara', 'yue': '法屬圭亞那', 'yue-Hans': '法属圭亚那', 'yue-Hant': '法屬圭亞那', 'zgh': 'ⴳⵡⵉⵢⴰⵏ ⵜⴰⴼⵔⴰⵏⵙⵉⵙⵜ', 'zh': '法属圭亚那', 'zh-Hans': '法属圭亚那', 'zh-Hant': '法屬圭亞那', 'zu': 'i-French Guiana'},
    'GG': {'af': 'Guernsey', 'ak': 'Guɛnse', 'am': 'ጉርነሲ', 'ar': 'غيرنزي', 'as': 'গোৰেনচি', 'ast': 'Guernsey', 'az': 'Gernsi', 'az-Cyrl': 'Ҝернси', 'az-Latn': 'Gernsi', 'bal-Latn': 'Gwernsay', 'be': 'Гернсі', 'be-tarask': 'Гернсі', 'bg': 'Гърнзи', 'blo': 'Gǝrǝnsɛɩ', 'bn': 'গার্নসি', 'br': 'Gwernenez', 'brx': 'गोर्नसि', 'bs': 'Guernsey', 'bs-Cyrl': 'Гернзи', 'bs-Latn': 'Guernsey', 'ca': 'Guernsey', 'ca-ES-valencia': 'Guernsey', 'ccp': '𑄉𑄳𑄢𑄚𑄴𑄏𑄨', 'ce': 'Гернси', 'ceb': 'Guernsey', 'chr': 'ᎬᏂᏏ', 'ckb': 'گێرنزی', 'cs': 'Guernsey', 'cv': 'Гернси', 'cy': 'Ynys y Garn', 'da': 'Guernsey', 'de': 'Guernsey', 'dsb': 'Guernsey', 'dz': 'གུ་ཨེརྣ་སི', 'ee': 'Guernse nutome', 'el': 'Γκέρνζι', 'el-polyton': 'Γκέρνζι', 'en': 'Guernsey', 'en-Dsrt': '𐐘𐐲𐑉𐑌𐑆𐐨', 'es': 'Guernesey', 'et': 'Guernsey', 'eu': 'Guernesey', 'fa': 'گرنزی', 'ff-Adlm': '𞤘𞤢𞤪𞤲𞤫𞤧𞤭𞥅', 'fi': 'Guernsey', 'fil': 'Guernsey', 'fo': 'Guernsey', 'fr': 'Guernesey', 'fur': 'Guernsey', 'fy': 'Guernsey', 'ga': 'Geansaí', 'gd': 'Geàrnsaidh', 'gl': 'Guernsey', 'gsw': 'Gäärnsi', 'gu': 'ગ્વેર્નસે', 'ha': 'Yankin Guernsey', 'he': 'גרנזי', 'hi': 'गर्नसी', 'hi-Latn': 'Guernsey', 'hr': 'Guernsey', 'hsb': 'Guernsey', 'ht': 'Guernesey', 'hu': 'Guernsey', 'hy': 'Գերնսի', 'ia': 'Guernsey', 'id': 'Guernsey', 'ig': 'Guernsey', 'is': 'Guernsey', 'it': 'Guernsey', 'ja': 'ガーンジー', 'jv': 'Guernsei', 'ka': 'გერნსი', 'kaa': 'Гернси', 'kaa-Cyrl': 'Гернси', 'kea': 'Gernzi', 'kgp': 'Gérnesej', 'kk': 'Гернси', 'kk-Cyrl': 'Гернси', 'km': 'ហ្គេនស៊ី', 'kn': 'ಗುರ್ನ್\u200cಸೆ', 'ko': '건지', 'kok': 'गर्नसी', 'kok-Deva': 'गर्नसी', 'ks': 'گورنسے', 'ks-Arab': 'گورنسے', 'ksh': 'Jöönsei', 'ku': 'Guernsey', 'kxv': 'grnse', 'kxv-Deva': 'गर्नसी', 'kxv-Latn': 'grnse', 'kxv-Orya': 'ଗର୍ନସି', 'kxv-Telu': 'గర్నసీ', 'ky': 'Гернси', 'lb': 'Guernsey', 'lld': 'Guernsey', 'ln': 'Guernesey', 'lo': 'ເກີນຊີ', 'lt': 'Gernsis', 'lv': 'Gērnsija', 'mai': 'गर्नसी', 'mi': 'Kōnihi', 'mk': 'Гернзи', 'ml': 'ഗേൺസി', 'mn': 'Гернси', 'mr': 'ग्वेर्नसे', 'ms': 'Guernsey', 'mt': 'Guernsey', 'my': 'ဂွန်းဇီ', 'mzn': 'گرنزی', 'nb': 'Guernsey', 'ne': 'ग्यूर्न्सी', 'nl': 'Guernsey', 'nn': 'Guernsey', 'no': 'Guernsey', 'nqo': 'ߜ߭ߋߙߑߣߋߖ߭ߌ߫', 'om': 'Guwernisey', 'or': 'ଗୁଏରନେସି', 'pa': 'ਗਰਨਜੀ', 'pa-Guru': 'ਗਰਨਜੀ', 'pcm': 'Guẹnzi', 'pl': 'Guernsey', 'ps': 'ګرنسي', 'pt': 'Guernsey', 'qu': 'Guernesey', 'rm': 'Guernsey', 'ro': 'Guernsey', 'ru': 'Гернси', 'sc': 'Guernsey', 'sd': 'گورنسي', 'sd-Arab': 'گورنسي', 'se': 'Guernsey', 'si': 'ගර්න්සිය', 'sk': 'Guernsey', 'sl': 'Guernsey', 'smn': 'Guernsey', 'so': 'Guurnsey', 'sq': 'Gernsej', 'sr': 'Гернзи', 'sr-Cyrl': 'Гернзи', 'sr-Latn': 'Gernzi', 'sv': 'Guernsey', 'sw': 'Guernsey', 'syr': 'ܓܘܪܢܙܝ', 'ta': 'கெர்ன்சி', 'te': 'గర్న్\u200cసీ', 'tg': 'Гернси', 'th': 'เกิร์นซีย์', 'ti': 'ገርንዚ', 'tk': 'Gernsi', 'to': 'Kuenisī', 'tr': 'Guernsey', 'tt': 'Гернси', 'ug': 'گۇرنسېي', 'uk': 'Гернсі', 'ur': 'گوئرنسی', 'uz': 'Gernsi', 'uz-Cyrl': 'Гернси', 'uz-Latn': 'Gernsi', 'vai': 'ꖶꗦꘋꔻ', 'vai-Vaii': 'ꖶꗦꘋꔻ', 'vec': 'Guernsey', 'vi': 'Guernsey', 'wae': 'Guernsey', 'wo': 'Gernase', 'xh': 'EGuernsey', 'xnr': 'गर्नसी', 'yi': 'גערנזי', 'yo': 'Guernsey', 'yrl': 'Guwẽnisei', 'yue': '根西島', 'yue-Hans': '根西岛', 'yue-Hant': '根西島', 'zh': '根西岛', 'zh-Hans': '根西岛', 'zh-Hant': '根息', 'zu': 'i-Guernsey'},
    'GH': {'af': 'Ghana', 'agq': 'Gaanà', 'ak': 'Gaana', 'am': 'ጋና', 'ar': 'غانا', 'as': 'ঘানা', 'asa': 'Ghana', 'ast': 'Ghana', 'az': 'Qana', 'az-Cyrl': 'Гана', 'az-Latn': 'Qana', 'bal-Latn': 'Gáná', 'bas': 'Ganà', 'be': 'Гана', 'be-tarask': 'Гана', 'bez': 'Hughana', 'bg': 'Гана', 'bgn': 'گانا', 'blo': 'Gana', 'bm': 'Gana', 'bn': 'ঘানা', 'br': 'Ghana', 'brx': 'घाना', 'bs': 'Gana', 'bs-Cyrl': 'Гана', 'bs-Latn': 'Gana', 'ca': 'Ghana', 'ca-ES-valencia': 'Ghana', 'ccp': '𑄊𑄚', 'ce': 'Гана', 'ceb': 'Ghana', 'cgg': 'Gana', 'chr': 'ᎦᎠᎾ', 'ckb': 'غەنا', 'cs': 'Ghana', 'cv': 'Гана', 'cy': 'Ghana', 'da': 'Ghana', 'dav': 'Ghana', 'de': 'Ghana', 'dje': 'Gaana', 'dsb': 'Ghana', 'dyo': 'Gaana', 'dz': 'གྷ་ན', 'ebu': 'Ghana', 'ee': 'Ghana nutome', 'el': 'Γκάνα', 'el-polyton': 'Γκάνα', 'en': 'Ghana', 'en-Dsrt': '𐐘𐐪𐑌𐐲', 'eo': 'Ganao', 'es': 'Ghana', 'et': 'Ghana', 'eu': 'Ghana', 'ewo': 'Ganá', 'fa': 'غنا', 'ff': 'Ganaa', 'ff-Adlm': '𞤘𞤢𞤲𞤢', 'ff-Latn': 'Ganaa', 'fi': 'Ghana', 'fil': 'Ghana', 'fo': 'Gana', 'fr': 'Ghana', 'fur': 'Ghana', 'fy': 'Ghana', 'ga': 'Gána', 'gaa': 'Ghana', 'gd': 'Gàna', 'gl': 'Ghana', 'gsw': 'Gaana', 'gu': 'ઘાના', 'guz': 'Ghana', 'ha': 'Gana', 'he': 'גאנה', 'hi': 'घाना', 'hi-Latn': 'Ghana', 'hr': 'Gana', 'hsb': 'Ghana', 'ht': 'Ghana', 'hu': 'Ghána', 'hy': 'Գանա', 'ia': 'Ghana', 'id': 'Ghana', 'ig': 'Ghana', 'is': 'Gana', 'it': 'Ghana', 'ja': 'ガーナ', 'jgo': 'Ŋgána', 'jmc': 'Ghana', 'jv': 'Ghana', 'ka': 'განა', 'kaa': 'Гана', 'kaa-Cyrl': 'Гана', 'kab': 'Ɣana', 'kam': 'Ghana', 'kde': 'Ghana', 'kea': 'Gana', 'kgp': 'Ganỹ', 'khq': 'Gaana', 'ki': 'Ngana', 'kk': 'Гана', 'kk-Cyrl': 'Гана', 'kln': 'Emetab Ghana', 'km': 'ហ្គាណា', 'kn': 'ಘಾನಾ', 'ko': '가나', 'kok': 'घाना', 'kok-Deva': 'घाना', 'ks': 'گانا', 'ks-Arab': 'گانا', 'ksb': 'Ghana', 'ksf': 'gána', 'ksh': 'Jaana', 'ku': 'Gana', 'kxv': 'gana', 'kxv-Deva': 'गाना', 'kxv-Latn': 'gana', 'kxv-Orya': 'ଗାନା', 'kxv-Telu': 'గనా', 'ky': 'Гана', 'lag': 'Gáana', 'lb': 'Ghana', 'lg': 'Gana', 'lld': 'Ghana', 'ln': 'Gana', 'lo': 'ການາ', 'lt': 'Gana', 'lu': 'Ngana', 'luo': 'Ghana', 'luy': 'Ghana', 'lv': 'Gana', 'mai': 'घाना', 'mas': 'Ghana', 'mer': 'Ghana', 'mfe': 'Ghana', 'mg': 'Ghana', 'mgh': 'Ugana', 'mi': 'Kāna', 'mk': 'Гана', 'ml': 'ഘാന', 'mn': 'Гана', 'mr': 'घाना', 'ms': 'Ghana', 'mt': 'il-Ghana', 'mua': 'Gana', 'my': 'ဂါနာ', 'mzn': 'غنا', 'naq': 'Ghanab', 'nb': 'Ghana', 'nd': 'Ghana', 'ne': 'घाना', 'nl': 'Ghana', 'nmg': 'Gána', 'nn': 'Ghana', 'no': 'Ghana', 'nqo': 'ߜ߭ߊ߯ߣߊ߫', 'nyn': 'Gana', 'om': 'Gaanaa', 'or': 'ଘାନା', 'pa': 'ਘਾਨਾ', 'pa-Guru': 'ਘਾਨਾ', 'pcm': 'Gána', 'pl': 'Ghana', 'ps': 'ګانا', 'pt': 'Gana', 'qu': 'Ghana', 'rm': 'Ghana', 'rn': 'Gana', 'ro': 'Ghana', 'rof': 'Ghana', 'ru': 'Гана', 'rwk': 'Ghana', 'saq': 'Ghana', 'sbp': 'Khana', 'sc': 'Ghana', 'sd': 'گهانا', 'sd-Arab': 'گهانا', 'se': 'Ghana', 'seh': 'Gana', 'ses': 'Gaana', 'sg': 'Ganäa', 'shi': 'ⵖⴰⵏⴰ', 'shi-Latn': 'ɣana', 'shi-Tfng': 'ⵖⴰⵏⴰ', 'si': 'ඝානාව', 'sk': 'Ghana', 'sl': 'Gana', 'smn': 'Ghana', 'sn': 'Ghana', 'so': 'Gaana', 'sq': 'Ganë', 'sr': 'Гана', 'sr-Cyrl': 'Гана', 'sr-Latn': 'Gana', 'sv': 'Ghana', 'sw': 'Ghana', 'syr': 'ܓܐܢܐ', 'ta': 'கானா', 'te': 'ఘనా', 'teo': 'Ghana', 'tg': 'Гана', 'th': 'กานา', 'ti': 'ጋና', 'tk': 'Gana', 'to': 'Kana', 'tr': 'Gana', 'tt': 'Гана', 'twq': 'Gaana', 'tzm': 'Ɣana', 'ug': 'گانا', 'uk': 'Гана', 'ur': 'گھانا', 'uz': 'Gana', 'uz-Cyrl': 'Гана', 'uz-Latn': 'Gana', 'vai': 'ꕭꕌꕯ', 'vai-Latn': 'Gana', 'vai-Vaii': 'ꕭꕌꕯ', 'vec': 'Gana', 'vi': 'Ghana', 'vun': 'Ghana', 'wae': 'Gana', 'wo': 'Gana', 'xh': 'EGhana', 'xnr': 'घाना', 'xog': 'Gana', 'yav': 'kaná', 'yi': 'גהאַנע', 'yo': 'Gana', 'yrl': 'Gana', 'yue': '迦納', 'yue-Hans': '迦纳', 'yue-Hant': '迦納', 'zgh': 'ⵖⴰⵏⴰ', 'zh': '加纳', 'zh-Hans': '加纳', 'zh-Hant': '迦納', 'zu': 'i-Ghana'},
    'GI': {'af': 'Gibraltar', 'agq': 'Dzibɨ̀latà', 'ak': 'Gyebralta', 'am': 'ጂብራልተር', 'ar': 'جبل طارق', 'as': 'জিব্ৰাল্টৰ', 'asa': 'Jibralta', 'ast': 'Xibraltar', 'az': 'Cəbəllütariq', 'az-Cyrl': 'Ҹәбәллүтариг', 'az-Latn': 'Cəbəllütariq', 'bal-Latn': 'Gibráltar', 'bas': 'Gìlbràtâr', 'be': 'Гібралтар', 'be-tarask': 'Гібралтар', 'bez': 'Hujiblalta', 'bg': 'Гибралтар', 'blo': 'Gibraltaa', 'bm': 'Zibralitari', 'bn': 'জিব্রাল্টার', 'br': 'Jibraltar', 'brx': 'जिब्राल्टार', 'bs': 'Gibraltar', 'bs-Cyrl': 'Гибралтар', 'bs-Latn': 'Gibraltar', 'ca': 'Gibraltar', 'ca-ES-valencia': 'Gibraltar', 'ccp': '𑄎𑄨𑄝𑄳𑄢𑄣𑄴𑄑𑄢𑄴', 'ce': 'Гибралтар', 'ceb': 'Gibraltar', 'cgg': 'Giburaata', 'chr': 'ᏥᏆᎵᏓ', 'ckb': 'گیبرالتار', 'cs': 'Gibraltar', 'cv': 'Гибралтар', 'cy': 'Gibraltar', 'da': 'Gibraltar', 'dav': 'Jibralta', 'de': 'Gibraltar', 'dje': 'Gibraltar', 'dsb': 'Gibraltar', 'dyo': 'Sipraltaar', 'dz': 'ཇིབ་རཱལ་ཊར', 'ebu': 'Jibralta', 'ee': 'Gibraltar nutome', 'el': 'Γιβραλτάρ', 'el-polyton': 'Γιβραλτάρ', 'en': 'Gibraltar', 'en-Dsrt': '𐐖𐐲𐐺𐑉𐐱𐑊𐐻𐐲𐑉', 'eo': 'Ĝibraltaro', 'es': 'Gibraltar', 'et': 'Gibraltar', 'eu': 'Gibraltar', 'ewo': 'Yiləbalatár', 'fa': 'جبل\u200cالطارق', 'ff': 'Jibraltaar', 'ff-Adlm': '𞤔𞤭𞤦𞤪𞤢𞤤𞤼𞤢𞥄', 'ff-Latn': 'Jibraltaar', 'fi': 'Gibraltar', 'fil': 'Gibraltar', 'fo': 'Gibraltar', 'fr': 'Gibraltar', 'fur': 'Gjibraltar', 'fy': 'Gibraltar', 'ga': 'Giobráltar', 'gd': 'Diobraltar', 'gl': 'Xibraltar', 'gsw': 'Gibraltar', 'gu': 'જીબ્રાલ્ટર', 'guz': 'Jibralta', 'ha': 'Jibaraltar', 'he': 'גיברלטר', 'hi': 'जिब्राल्टर', 'hi-Latn': 'Gibraltar', 'hr': 'Gibraltar', 'hsb': 'Gibraltar', 'ht': 'Gibraltar', 'hu': 'Gibraltár', 'hy': 'Ջիբրալթար', 'ia': 'Gibraltar', 'id': 'Gibraltar', 'ig': 'Gibraltar', 'is': 'Gíbraltar', 'it': 'Gibilterra', 'ja': 'ジブラルタル', 'jmc': 'Jibralta', 'jv': 'Gibraltar', 'ka': 'გიბრალტარი', 'kaa': 'Гибралтар', 'kaa-Cyrl': 'Гибралтар', 'kab': 'Jibraltar', 'kam': 'Jibralta', 'kde': 'Diblalta', 'kea': 'Jibraltar', 'kgp': 'Gimrar-tar', 'khq': 'Gibraltar', 'ki': 'Jibralta', 'kk': 'Гибралтар', 'kk-Cyrl': 'Гибралтар', 'kln': 'Emetab Gibraltar', 'km': 'ហ្ស៊ីប្រាល់តា', 'kn': 'ಗಿಬ್ರಾಲ್ಟರ್', 'ko': '지브롤터', 'kok': 'जिब्राल्टर', 'kok-Deva': 'जिब्राल्टर', 'ks': 'جِبرالٹَر', 'ks-Arab': 'جِبرالٹَر', 'ksb': 'Jiblalta', 'ksf': 'jibraltá', 'ksh': 'Jibralltaa', 'ku': 'Cebelîtariq', 'kxv': 'jbralṭr', 'kxv-Deva': 'जिब्राल्टर', 'kxv-Latn': 'jbralṭr', 'kxv-Orya': 'ଜିବ୍ରାଲଟର', 'kxv-Telu': 'జిబ్రాల్టర్', 'ky': 'Гибралтар', 'lag': 'Jiburálita', 'lb': 'Gibraltar', 'lg': 'Giburalita', 'lld': 'Gibraltar', 'ln': 'Zibatalɛ', 'lo': 'ຈິບບຣອນທາ', 'lt': 'Gibraltaras', 'lu': 'Jibeletale', 'luo': 'Gibraltar', 'luy': 'Jibralta', 'lv': 'Gibraltārs', 'mai': 'जिब्राल्टर', 'mas': 'Jibralta', 'mer': 'Ngĩbrata', 'mfe': 'Zibraltar', 'mg': 'Zibraltara', 'mgh': 'Ujibralta', 'mi': 'Kāmaka', 'mk': 'Гибралтар', 'ml': 'ജിബ്രാൾട്ടർ', 'mn': 'Гибралтар', 'mr': 'जिब्राल्टर', 'ms': 'Gibraltar', 'mt': 'Ġibiltà', 'mua': 'Sǝr Gibraltar', 'my': 'ဂျီဘရော်လ်တာ', 'mzn': 'جبل طارق', 'naq': 'Gibraltar', 'nb': 'Gibraltar', 'nd': 'Gibraltar', 'ne': 'जिब्राल्टार', 'nl': 'Gibraltar', 'nmg': 'Gilbratar', 'nn': 'Gibraltar', 'no': 'Gibraltar', 'nqo': 'ߜ߭ߌߓߙߊߟߑߕߊߙ', 'nyn': 'Giburaata', 'om': 'Gibraaltar', 'or': 'ଜିବ୍ରାଲ୍ଟର୍', 'pa': 'ਜਿਬਰਾਲਟਰ', 'pa-Guru': 'ਜਿਬਰਾਲਟਰ', 'pcm': 'Jibrọ́lta', 'pl': 'Gibraltar', 'ps': 'جبل الطارق', 'pt': 'Gibraltar', 'qu': 'Gibraltar', 'rm': 'Gibraltar', 'rn': 'Juburalitari', 'ro': 'Gibraltar', 'rof': 'Jibralta', 'ru': 'Гибралтар', 'rwk': 'Jibralta', 'saq': 'Jibralta', 'sbp': 'Jibulalita', 'sc': 'Gibilterra', 'sd': 'جبرالٽر', 'sd-Arab': 'جبرالٽر', 'se': 'Gibraltar', 'seh': 'Gibraltar', 'ses': 'Gibraltar', 'sg': 'Zibraltära, Zibaratära', 'shi': 'ⴰⴷⵔⴰⵔ ⵏ ⵟⴰⵕⵉⵇ', 'shi-Latn': 'adrar n ṭaṛiq', 'shi-Tfng': 'ⴰⴷⵔⴰⵔ ⵏ ⵟⴰⵕⵉⵇ', 'si': 'ජිබ්\u200dරෝල්ටාව', 'sk': 'Gibraltár', 'sl': 'Gibraltar', 'smn': 'Gibraltar', 'sn': 'Gibraltar', 'so': 'Gibraltar', 'sq': 'Gjibraltar', 'sr': 'Гибралтар', 'sr-Cyrl': 'Гибралтар', 'sr-Latn': 'Gibraltar', 'sv': 'Gibraltar', 'sw': 'Gibraltar', 'syr': 'ܓܒܪܠܛܪ', 'ta': 'ஜிப்ரால்டர்', 'te': 'జిబ్రాల్టర్', 'teo': 'Jibralta', 'tg': 'Гибралтар', 'th': 'ยิบรอลตาร์', 'ti': 'ጂብራልታር', 'tk': 'Gibraltar', 'to': 'Sipalālitā', 'tr': 'Cebelitarık', 'tt': 'Гибралтар', 'twq': 'Gibraltar', 'tzm': 'Jibralṭar', 'ug': 'جەبىلتارىق', 'uk': 'Гібралтар', 'ur': 'جبل الطارق', 'uz': 'Gibraltar', 'uz-Cyrl': 'Гибралтар', 'uz-Latn': 'Gibraltar', 'vai': 'ꕀꖜꕟꕚ', 'vai-Latn': 'Jibhurata', 'vai-Vaii': 'ꕀꖜꕟꕚ', 'vec': 'Gibiltera', 'vi': 'Gibraltar', 'vun': 'Jibralta', 'wae': 'Gibraltar', 'wo': 'Sibraltaar', 'xh': 'EGibraltar', 'xnr': 'जिब्राल्टर', 'xog': 'Giburalita', 'yav': 'sílpalatáal', 'yi': 'גיבראַלטאַר', 'yo': 'Gibaratara', 'yrl': 'Gibarautá', 'yue': '直布羅陀', 'yue-Hans': '直布罗陀', 'yue-Hant': '直布羅陀', 'zgh': 'ⴰⴷⵔⴰⵔ ⵏ ⵟⴰⵕⵉⵇ', 'zh': '直布罗陀', 'zh-Hans': '直布罗陀', 'zh-Hant': '直布羅陀', 'zu': 'i-Gibraltar'},
    'GL': {'af': 'Groenland', 'agq': 'Gɨ̀lenlân', 'ak': 'Greenman', 'am': 'ግሪንላንድ', 'ar': 'غرينلاند', 'as': 'গ্ৰীণলেণ্ড', 'asa': 'Grinlandi', 'ast': 'Groenlandia', 'az': 'Qrenlandiya', 'az-Cyrl': 'Гренландија', 'az-Latn': 'Qrenlandiya', 'bal-Latn': 'Grinlánd', 'bas': 'Grǐnlànd', 'be': 'Грэнландыя', 'be-tarask': 'Грэнландыя', 'bez': 'Hujinlandi', 'bg': 'Гренландия', 'bgn': 'گرینلاند', 'blo': 'Grinlanɖ', 'bm': 'Gɔrɔhenelandi', 'bn': 'গ্রীনল্যান্ড', 'br': 'Greunland', 'brx': 'ग्रीनलेण्ड', 'bs': 'Grenland', 'bs-Cyrl': 'Гренланд', 'bs-Latn': 'Grenland', 'ca': 'Groenlàndia', 'ca-ES-valencia': 'Groenlàndia', 'ccp': '𑄉𑄳𑄢𑄩𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴', 'ce': 'Гренланди', 'ceb': 'Greenland', 'cgg': 'Guriinirandi', 'chr': 'ᎢᏤᏍᏛᏱ', 'ckb': 'گرینلاند', 'cs': 'Grónsko', 'cv': 'Гренланди', 'cy': 'Yr Ynys Las', 'da': 'Grønland', 'dav': 'Grinlandi', 'de': 'Grönland', 'dje': 'Grinland', 'dsb': 'Grönlandska', 'dyo': 'Greenland', 'dz': 'གིརཱིན་ལནཌ྄', 'ebu': 'Grinlandi', 'ee': 'Grinland nutome', 'el': 'Γροιλανδία', 'el-polyton': 'Γροιλανδία', 'en': 'Greenland', 'en-Dsrt': '𐐘𐑉𐐨𐑌𐑊𐐲𐑌𐐼', 'eo': 'Gronlando', 'es': 'Groenlandia', 'et': 'Gröönimaa', 'eu': 'Groenlandia', 'ewo': 'Goelán', 'fa': 'گرینلند', 'ff': 'Gorwendland', 'ff-Adlm': '𞤘𞤭𞤪𞤤𞤢𞤲𞤣𞤭', 'ff-Latn': 'Gorwendland', 'fi': 'Grönlanti', 'fil': 'Greenland', 'fo': 'Grønland', 'fr': 'Groenland', 'fur': 'Groenlande', 'fy': 'Grienlân', 'ga': 'an Ghraonlainn', 'gd': 'A’ Ghraonlann', 'gl': 'Groenlandia', 'gsw': 'Gröönland', 'gu': 'ગ્રીનલેન્ડ', 'guz': 'Grinlandi', 'ha': 'Grinlan', 'he': 'גרינלנד', 'hi': 'ग्रीनलैंड', 'hi-Latn': 'Greenland', 'hr': 'Grenland', 'hsb': 'Grönlandska', 'ht': 'Groenland', 'hu': 'Grönland', 'hy': 'Գրենլանդիա', 'ia': 'Groenlandia', 'id': 'Greenland', 'ig': 'Greenland', 'is': 'Grænland', 'it': 'Groenlandia', 'ja': 'グリーンランド', 'jmc': 'Grinlandi', 'jv': 'Greenland', 'ka': 'გრენლანდია', 'kaa': 'Гренландия', 'kaa-Cyrl': 'Гренландия', 'kab': 'Grunland', 'kam': 'Grinlandi', 'kde': 'Glinlandi', 'kea': 'Gronelándia', 'kgp': 'Groẽrỹnija', 'khq': 'Grinland', 'ki': 'Grinlandi', 'kk': 'Гренландия', 'kk-Cyrl': 'Гренландия', 'kl': 'Kalaallit Nunaat', 'kln': 'Emetab Greenland', 'km': 'ហ្គ្រោអង់ឡង់', 'kn': 'ಗ್ರೀನ್\u200cಲ್ಯಾಂಡ್', 'ko': '그린란드', 'kok': 'ग्रीनलँड', 'kok-Deva': 'ग्रीनलँड', 'ks': 'گرین لینڈ', 'ks-Arab': 'گرین لینڈ', 'ksb': 'Glinlandi', 'ksf': 'grínlan', 'ksh': 'Jröhnland', 'ku': 'Grînlanda', 'kxv': 'grīnlanḍ', 'kxv-Deva': 'ग्रीनलेंड', 'kxv-Latn': 'grīnlanḍ', 'kxv-Orya': 'ଗ୍ରୀନ୍ ଲାଣ୍ଡ', 'kxv-Telu': 'గ్రీన్లండ్', 'ky': 'Гренландия', 'lag': 'Giriniláandi', 'lb': 'Grönland', 'lg': 'Gurenelandi', 'lld': 'Groenlandia', 'ln': 'Gowelande', 'lo': 'ກຣີນແລນ', 'lt': 'Grenlandija', 'lu': 'Ngowelande', 'luo': 'Greenland', 'luy': 'Grinlandi', 'lv': 'Grenlande', 'mai': 'ग्रीनलैंड', 'mas': 'Grinlandi', 'mer': 'Ngirinilandi', 'mfe': 'Greenland', 'mg': 'Groenland', 'mgh': 'Ugrinlandi', 'mi': 'Whenuakāriki', 'mk': 'Гренланд', 'ml': 'ഗ്രീൻലൻഡ്', 'mn': 'Гренланд', 'mr': 'ग्रीनलंड', 'ms': 'Greenland', 'mt': 'Greenland', 'mua': 'Sǝr Groenland', 'my': 'ဂရင်းလန်း', 'mzn': 'گرینلند', 'naq': 'Greenland', 'nb': 'Grønland', 'nd': 'Greenland', 'ne': 'ग्रिनल्याण्ड', 'nl': 'Groenland', 'nmg': 'Greenland', 'nn': 'Grønland', 'no': 'Grønland', 'nqo': 'ߜ߭ߎߙߎ߲ߟߊ߲ߘ', 'nyn': 'Guriinirandi', 'om': 'Giriinlaand', 'or': 'ଗ୍ରୀନଲ୍ୟାଣ୍ଡ', 'pa': 'ਗ੍ਰੀਨਲੈਂਡ', 'pa-Guru': 'ਗ੍ਰੀਨਲੈਂਡ', 'pcm': 'Grínland', 'pl': 'Grenlandia', 'ps': 'ګرینلینډ', 'pt': 'Groenlândia', 'qu': 'Groenlandia', 'rm': 'Grönlanda', 'rn': 'Gurunilandi', 'ro': 'Groenlanda', 'rof': 'Grinlandi', 'ru': 'Гренландия', 'rwk': 'Grinlandi', 'saq': 'Grinlandi', 'sbp': 'Gilinilandi', 'sc': 'Groenlàndia', 'sd': 'گرين لينڊ', 'sd-Arab': 'گرين لينڊ', 'se': 'Kalaallit Nunaat', 'seh': 'Groênlandia', 'ses': 'Grinland', 'sg': 'Gorolânde', 'shi': 'ⴳⵔⵉⵍⴰⵏⴷ', 'shi-Latn': 'griland', 'shi-Tfng': 'ⴳⵔⵉⵍⴰⵏⴷ', 'si': 'ග්\u200dරීන්ලන්තය', 'sk': 'Grónsko', 'sl': 'Grenlandija', 'smn': 'Grönland', 'sn': 'Greenland', 'so': 'Greenland', 'sq': 'Grënlandë', 'sr': 'Гренланд', 'sr-Cyrl': 'Гренланд', 'sr-Latn': 'Grenland', 'sv': 'Grönland', 'sw': 'Greenland', 'syr': 'ܓܪܝܢܠܢܕ', 'ta': 'கிரீன்லாந்து', 'te': 'గ్రీన్\u200cల్యాండ్', 'teo': 'Grinlandi', 'tg': 'Гренландия', 'th': 'กรีนแลนด์', 'ti': 'ግሪንላንድ', 'tk': 'Grenlandiýa', 'to': 'Kulinilani', 'tr': 'Grönland', 'tt': 'Гренландия', 'twq': 'Grinland', 'tzm': 'Grinlanḍa', 'ug': 'گىرېنلاندىيە', 'uk': 'Гренландія', 'ur': 'گرین لینڈ', 'uz': 'Grenlandiya', 'uz-Cyrl': 'Гренландия', 'uz-Latn': 'Grenlandiya', 'vai': 'ꕧꕓ ꖴꕎ ꖨꕮꕊ', 'vai-Latn': 'Jamba Kuwa Lumaã', 'vai-Vaii': 'ꕧꕓ ꖴꕎ ꖨꕮꕊ', 'vec': 'Groenlanda', 'vi': 'Greenland', 'vun': 'Grinlandi', 'wae': 'Grönland', 'wo': 'Girinlànd', 'xh': 'EGreenland', 'xnr': 'ग्रीनलैंड', 'xog': 'Gurenelandi', 'yav': 'kuluɛnlánd', 'yi': 'גרינלאַנד', 'yo': 'Gerelandi', 'yrl': 'Guruẽrãdiya', 'yue': '格陵蘭', 'yue-Hans': '格陵兰', 'yue-Hant': '格陵蘭', 'zgh': 'ⴳⵔⵉⵍⴰⵏⴷ', 'zh': '格陵兰', 'zh-Hans': '格陵兰', 'zh-Hant': '格陵蘭', 'zu': 'i-Greenland'},
    'GM': {'af': 'Gambië', 'agq': 'Gambìa', 'ak': 'Gambia', 'am': 'ጋምቢያ', 'ar': 'غامبيا', 'as': 'গাম্বিয়া', 'asa': 'Gambia', 'ast': 'Gambia', 'az': 'Qambiya', 'az-Cyrl': 'Гамбија', 'az-Latn': 'Qambiya', 'bal-Latn': 'Gambiá', 'bas': 'Gàmbià', 'be': 'Гамбія', 'be-tarask': 'Гамбія', 'bez': 'Hugambia', 'bg': 'Гамбия', 'bgn': 'گامبیا', 'blo': 'Gambii', 'bm': 'Ganbi', 'bn': 'গাম্বিয়া', 'br': 'Gambia', 'brx': 'गाम्बिया', 'bs': 'Gambija', 'bs-Cyrl': 'Гамбија', 'bs-Latn': 'Gambija', 'ca': 'Gàmbia', 'ca-ES-valencia': 'Gàmbia', 'ccp': '𑄉𑄟𑄴𑄝𑄨𑄠', 'ce': 'Гамби', 'ceb': 'Gambia', 'cgg': 'Gambiya', 'chr': 'ᎦᎹᏈᎢᎠ', 'ckb': 'گامبیا', 'cs': 'Gambie', 'cv': 'Гамби', 'cy': 'Gambia', 'da': 'Gambia', 'dav': 'Gambia', 'de': 'Gambia', 'dje': 'Gambi', 'dsb': 'Gambija', 'dyo': 'Gambi', 'dz': 'གྷེམ་བི་ཡ', 'ebu': 'Gambia', 'ee': 'Gambia nutome', 'el': 'Γκάμπια', 'el-polyton': 'Γκάμπια', 'en': 'Gambia', 'en-Dsrt': '𐐘𐐰𐑋𐐺𐐨𐐲', 'eo': 'Gambio', 'es': 'Gambia', 'et': 'Gambia', 'eu': 'Gambia', 'ewo': 'Gambí', 'fa': 'گامبیا', 'ff': 'Gammbi', 'ff-Adlm': '𞤘𞤢𞤥𞤦𞤭𞤴𞤢', 'ff-Latn': 'Gammbi', 'fi': 'Gambia', 'fil': 'Gambia', 'fo': 'Gambia', 'fr': 'Gambie', 'fur': 'Gambia', 'fy': 'Gambia', 'ga': 'An Ghaimbia', 'gd': 'A’ Ghaimbia', 'gl': 'Gambia', 'gsw': 'Gambia', 'gu': 'ગેમ્બિયા', 'guz': 'Gambia', 'ha': 'Gambiya', 'he': 'גמביה', 'hi': 'गाम्बिया', 'hi-Latn': 'Gambia', 'hr': 'Gambija', 'hsb': 'Gambija', 'ht': 'Gambie', 'hu': 'Gambia', 'hy': 'Գամբիա', 'ia': 'Gambia', 'id': 'Gambia', 'ig': 'Gambia', 'is': 'Gambía', 'it': 'Gambia', 'ja': 'ガンビア', 'jgo': 'Ŋgambî', 'jmc': 'Gambia', 'jv': 'Gambia', 'ka': 'გამბია', 'kaa': 'Гамбия', 'kaa-Cyrl': 'Гамбия', 'kab': 'Gambya', 'kam': 'Gambia', 'kde': 'Gambia', 'kea': 'Gámbia', 'kgp': 'Gỹmija', 'khq': 'Gambi', 'ki': 'Gambia', 'kk': 'Гамбия', 'kk-Cyrl': 'Гамбия', 'kln': 'Emetab Gambia', 'km': 'ហ្គំប៊ី', 'kn': 'ಗ್ಯಾಂಬಿಯಾ', 'ko': '감비아', 'kok': 'गॅम्बिया', 'kok-Deva': 'गॅम्बिया', 'ks': 'گَمبِیا', 'ks-Arab': 'گَمبِیا', 'ksb': 'Gambia', 'ksf': 'gambí', 'ksh': 'Jambija', 'ku': 'Gambîya', 'kxv': 'gambia', 'kxv-Deva': 'गाम्बिया', 'kxv-Latn': 'gambia', 'kxv-Orya': 'ଗାମ୍ବିଆ', 'kxv-Telu': 'గంబియా', 'ky': 'Гамбия', 'lag': 'Gámbia', 'lb': 'Gambia', 'lg': 'Gambya', 'lld': 'Gambia', 'ln': 'Gambi', 'lo': 'ສາທາລະນະລັດແກມເບຍ', 'lt': 'Gambija', 'lu': 'Gambi', 'luo': 'Gambia', 'luy': 'Gambia', 'lv': 'Gambija', 'mai': 'गाम्बिया', 'mas': 'Gambia', 'mer': 'Gambia', 'mfe': 'Gambi', 'mg': 'Gambia', 'mgh': 'Ugambia', 'mi': 'Kamopia', 'mk': 'Гамбија', 'ml': 'ഗാംബിയ', 'mn': 'Гамби', 'mr': 'गाम्बिया', 'ms': 'Gambia', 'mt': 'il-Gambja', 'mua': 'Gambiya', 'my': 'ဂမ်ဘီရာ', 'mzn': 'گامبیا', 'naq': 'Gambiab', 'nb': 'Gambia', 'nd': 'Gambiya', 'ne': 'गाम्विया', 'nl': 'Gambia', 'nmg': 'Gambia', 'nn': 'Gambia', 'no': 'Gambia', 'nqo': 'ߜ߭ߊ߲ߓߌ߫', 'nyn': 'Gambiya', 'om': 'Gaambiyaa', 'or': 'ଗାମ୍ବିଆ', 'pa': 'ਗੈਂਬੀਆ', 'pa-Guru': 'ਗੈਂਬੀਆ', 'pcm': 'Gámbia', 'pl': 'Gambia', 'ps': 'ګامبیا', 'pt': 'Gâmbia', 'qu': 'Gambia', 'rm': 'Gambia', 'rn': 'Gambiya', 'ro': 'Gambia', 'rof': 'Gambia', 'ru': 'Гамбия', 'rwk': 'Gambia', 'saq': 'Gambia', 'sbp': 'Gambiya', 'sc': 'Gàmbia', 'sd': 'گيمبيا', 'sd-Arab': 'گيمبيا', 'se': 'Gámbia', 'seh': 'Gâmbia', 'ses': 'Gambi', 'sg': 'Gambïi', 'shi': 'ⴳⴰⵎⴱⵢⴰ', 'shi-Latn': 'gambya', 'shi-Tfng': 'ⴳⴰⵎⴱⵢⴰ', 'si': 'ගැම්බියාව', 'sk': 'Gambia', 'sl': 'Gambija', 'smn': 'Gambia', 'sn': 'Gambia', 'so': 'Gambiya', 'sq': 'Gambi', 'sr': 'Гамбија', 'sr-Cyrl': 'Гамбија', 'sr-Latn': 'Gambija', 'sv': 'Gambia', 'sw': 'Gambia', 'syr': 'ܓܡܒܝܐ', 'ta': 'காம்பியா', 'te': 'గాంబియా', 'teo': 'Gambia', 'tg': 'Гамбия', 'th': 'แกมเบีย', 'ti': 'ጋምብያ', 'tk': 'Gambiýa', 'to': 'Kamipia', 'tr': 'Gambiya', 'tt': 'Гамбия', 'twq': 'Gambi', 'tzm': 'Gambya', 'ug': 'گامبىيە', 'uk': 'Гамбія', 'ur': 'گیمبیا', 'uz': 'Gambiya', 'uz-Cyrl': 'Гамбия', 'uz-Latn': 'Gambiya', 'vai': 'ꕭꔭꕩ', 'vai-Latn': 'Gambiya', 'vai-Vaii': 'ꕭꔭꕩ', 'vec': 'Ganbia', 'vi': 'Gambia', 'vun': 'Gambia', 'wae': 'Gambia', 'wo': 'Gàmbi', 'xh': 'EGambia', 'xnr': 'गाम्बिया', 'xog': 'Gambya', 'yav': 'kambíi', 'yi': 'גאַמביע', 'yo': 'Gambia', 'yrl': 'Gãbiya', 'yue': '甘比亞', 'yue-Hans': '甘比亚', 'yue-Hant': '甘比亞', 'zgh': 'ⴳⴰⵎⴱⵢⴰ', 'zh': '冈比亚', 'zh-Hans': '冈比亚', 'zh-Hant': '甘比亞', 'zu': 'i-Gambia'},
    'GN': {'af': 'Guinee', 'agq': 'Ginè', 'ak': 'Gini', 'am': 'ጊኒ', 'ar': 'غينيا', 'as': 'গিনি', 'asa': 'Gine', 'ast': 'Guinea', 'az': 'Qvineya', 'az-Cyrl': 'Гвинеја', 'az-Latn': 'Qvineya', 'bal-Latn': 'Giniá', 'bas': 'Gìnê', 'be': 'Гвінея', 'be-tarask': 'Гвінея', 'bez': 'Hujine', 'bg': 'Гвинея', 'bgn': 'گوینیا', 'blo': 'Ginee', 'bm': 'Gine', 'bn': 'গিনি', 'br': 'Ginea', 'brx': 'गिनीया', 'bs': 'Gvineja', 'bs-Cyrl': 'Гвинеја', 'bs-Latn': 'Gvineja', 'ca': 'Guinea', 'ca-ES-valencia': 'Guinea', 'ccp': '𑄉𑄨𑄚𑄨', 'ce': 'Гвиней', 'ceb': 'Guinea', 'cgg': 'Gine', 'chr': 'ᎩᎢᏂ', 'ckb': 'گینێ', 'cs': 'Guinea', 'cv': 'Гвиней', 'cy': 'Gini', 'da': 'Guinea', 'dav': 'Gine', 'de': 'Guinea', 'dje': 'Gine', 'dsb': 'Gineja', 'dyo': 'Giné', 'dz': 'གྷི་ནི', 'ebu': 'Gine', 'ee': 'Guini nutome', 'el': 'Γουινέα', 'el-polyton': 'Γουινέα', 'en': 'Guinea', 'en-Dsrt': '𐐘𐐮𐑌𐐨', 'eo': 'Gvineo', 'es': 'Guinea', 'et': 'Guinea', 'eu': 'Ginea', 'ewo': 'Giné', 'fa': 'گینه', 'ff': 'Gine', 'ff-Adlm': '𞤘𞤭𞤲𞤫', 'ff-Latn': 'Gine', 'fi': 'Guinea', 'fil': 'Guinea', 'fo': 'Guinea', 'fr': 'Guinée', 'fur': 'Guinee', 'fy': 'Guinee', 'ga': 'An Ghuine', 'gd': 'Gini', 'gl': 'Guinea', 'gsw': 'Gineea', 'gu': 'ગિની', 'guz': 'Gine', 'ha': 'Gini', 'he': 'גינאה', 'hi': 'गिनी', 'hi-Latn': 'Guinea', 'hr': 'Gvineja', 'hsb': 'Gineja', 'ht': 'Guinée', 'hu': 'Guinea', 'hy': 'Գվինեա', 'ia': 'Guinea', 'id': 'Guinea', 'ig': 'Guinea', 'is': 'Gínea', 'it': 'Guinea', 'ja': 'ギニア', 'jgo': 'Ŋginɛ̂', 'jmc': 'Gine', 'jv': 'Guinea', 'ka': 'გვინეა', 'kaa': 'Гвинея', 'kaa-Cyrl': 'Гвинея', 'kab': 'Ɣinya', 'kam': 'Gine', 'kde': 'Gine', 'kea': 'Gine', 'kgp': 'Ginẽ', 'khq': 'Gine', 'ki': 'Gine', 'kk': 'Гвинея', 'kk-Cyrl': 'Гвинея', 'kln': 'Emetab Guinea', 'km': 'ហ្គីណេ', 'kn': 'ಗಿನಿ', 'ko': '기니', 'kok': 'गुएनिया', 'kok-Deva': 'गुएनिया', 'ks': 'گِنی', 'ks-Arab': 'گِنی', 'ksb': 'Gine', 'ksf': 'ginɛ́', 'ksh': 'Jinneha', 'ku': 'Gîne', 'kxv': 'gini', 'kxv-Deva': 'गिनी', 'kxv-Latn': 'gini', 'kxv-Orya': 'ଗିନି', 'kxv-Telu': 'గినియా', 'ky': 'Гвинея', 'lag': 'Gíine', 'lb': 'Guinea', 'lg': 'Gini', 'lld': 'Guinea', 'ln': 'Ginɛ', 'lo': 'ກິນີ', 'lt': 'Gvinėja', 'lu': 'Ngine', 'luo': 'Guinea', 'luy': 'Gine', 'lv': 'Gvineja', 'mai': 'गिनी', 'mas': 'Gine', 'mer': 'Gine', 'mfe': 'Gine', 'mg': 'Ginea', 'mgh': 'Ugine', 'mi': 'Kini', 'mk': 'Гвинеја', 'ml': 'ഗിനിയ', 'mn': 'Гвиней', 'mr': 'गिनी', 'ms': 'Guinea', 'mt': 'il-Guinea', 'mua': 'Guine', 'my': 'ဂီနီ', 'mzn': 'گینه', 'naq': 'Guineab', 'nb': 'Guinea', 'nd': 'Guinea', 'ne': 'गिनी', 'nl': 'Guinee', 'nmg': 'Guine', 'nn': 'Guinea', 'no': 'Guinea', 'nqo': 'ߖߌ߬ߣߍ߫', 'nyn': 'Gine', 'om': 'Giinii', 'or': 'ଗୁଇନିଆ', 'pa': 'ਗਿਨੀ', 'pa-Guru': 'ਗਿਨੀ', 'pcm': 'Gíni', 'pl': 'Gwinea', 'ps': 'ګینه', 'pt': 'Guiné', 'qu': 'Guinea', 'rm': 'Guinea', 'rn': 'Guneya', 'ro': 'Guineea', 'rof': 'Gine', 'ru': 'Гвинея', 'rwk': 'Gine', 'saq': 'Gine', 'sbp': 'Gine', 'sc': 'Guinea', 'sd': 'گني', 'sd-Arab': 'گني', 'se': 'Guinea', 'seh': 'Guiné', 'ses': 'Gine', 'sg': 'Ginëe', 'shi': 'ⵖⵉⵏⵢⴰ', 'shi-Latn': 'ɣinya', 'shi-Tfng': 'ⵖⵉⵏⵢⴰ', 'si': 'ගිණියාව', 'sk': 'Guinea', 'sl': 'Gvineja', 'smn': 'Guinea', 'sn': 'Guinea', 'so': 'Gini', 'sq': 'Guine', 'sr': 'Гвинеја', 'sr-Cyrl': 'Гвинеја', 'sr-Latn': 'Gvineja', 'sv': 'Guinea', 'sw': 'Gine', 'syr': 'ܓܝܢܝܐ', 'ta': 'கினியா', 'te': 'గినియా', 'teo': 'Gine', 'tg': 'Гвинея', 'th': 'กินี', 'ti': 'ጊኒ', 'tk': 'Gwineýa', 'to': 'Kini', 'tr': 'Gine', 'tt': 'Гвинея', 'twq': 'Gine', 'tzm': 'Ɣinya', 'ug': 'گىۋىنىيە', 'uk': 'Гвінея', 'ur': 'گنی', 'uz': 'Gvineya', 'uz-Cyrl': 'Гвинея', 'uz-Latn': 'Gvineya', 'vai': 'ꕅꔤꕇ', 'vai-Latn': 'Gini', 'vai-Vaii': 'ꕅꔤꕇ', 'vec': 'Guinèa', 'vi': 'Guinea', 'vun': 'Gine', 'wae': 'Ginea', 'wo': 'Gine', 'xh': 'EGuinea', 'xnr': 'गिनी', 'xog': 'Gini', 'yav': 'kiiné', 'yi': 'גינע', 'yo': 'Gene', 'yrl': 'Giné', 'yue': '幾內亞', 'yue-Hans': '几内亚', 'yue-Hant': '幾內亞', 'zgh': 'ⵖⵉⵏⵢⴰ', 'zh': '几内亚', 'zh-Hans': '几内亚', 'zh-Hant': '幾內亞', 'zu': 'i-Guinea'},
    'GP': {'af': 'Guadeloupe', 'agq': 'Gwadalukpɛ̀', 'ak': 'Guwadelup', 'am': 'ጉዋደሉፕ', 'ar': 'غوادلوب', 'as': 'গুৱাডেলুপ', 'asa': 'Gwadelupe', 'ast': 'Guadalupe', 'az': 'Qvadelupa', 'az-Cyrl': 'Гваделупа', 'az-Latn': 'Qvadelupa', 'bal-Latn': 'Gwádelóp', 'bas': 'Gwàdèlûp', 'be': 'Гвадэлупа', 'be-tarask': 'Гвадэлупа', 'bez': 'Hugwadelupe', 'bg': 'Гваделупа', 'blo': 'Guwaɖeluupu', 'bm': 'Gwadelup', 'bn': 'গুয়াদেলৌপ', 'br': 'Gwadeloup', 'brx': 'ग्वादेल्यूप', 'bs': 'Gvadalupe', 'bs-Cyrl': 'Гваделупе', 'bs-Latn': 'Gvadalupe', 'ca': 'Guadalupe', 'ca-ES-valencia': 'Guadalupe', 'ccp': '𑄉𑄪𑄠𑄘𑄬𑄣𑄯𑄛𑄴', 'ce': 'Гваделупа', 'ceb': 'Guadeloupe', 'cgg': 'Gwaderupe', 'chr': 'ᏩᏓᎷᏇ', 'ckb': 'گوادێلۆپ', 'cs': 'Guadeloupe', 'cv': 'Гваделупа', 'cy': 'Guadeloupe', 'da': 'Guadeloupe', 'dav': 'Gwadelupe', 'de': 'Guadeloupe', 'dje': 'Gwadeluup', 'dsb': 'Guadeloupe', 'dyo': 'Guwadalup', 'dz': 'གོ་ཌེ་ལུ་པེ', 'ebu': 'Gwadelupe', 'ee': 'Guadelupe nutome', 'el': 'Γουαδελούπη', 'el-polyton': 'Γουαδελούπη', 'en': 'Guadeloupe', 'en-Dsrt': '𐐘𐐶𐐪𐐼𐐲𐑊𐐭𐐹', 'eo': 'Gvadelupo', 'es': 'Guadalupe', 'et': 'Guadeloupe', 'eu': 'Guadalupe', 'ewo': 'Guadəlúb', 'fa': 'گوادلوپ', 'ff': 'Gwaadalup', 'ff-Adlm': '𞤘𞤵𞤱𞤢𞤣𞤢𞤤𞤵𞤨𞤫𞥅', 'ff-Latn': 'Gwaadalup', 'fi': 'Guadeloupe', 'fil': 'Guadeloupe', 'fo': 'Guadeloupe', 'fr': 'Guadeloupe', 'fur': 'Guadalupe', 'fy': 'Guadeloupe', 'ga': 'Guadalúip', 'gd': 'Guadalup', 'gl': 'Guadalupe', 'gsw': 'Guadälup', 'gu': 'ગ્વાડેલોપ', 'guz': 'Gwadelupe', 'ha': 'Gwadaluf', 'he': 'גוואדלופ', 'hi': 'ग्वाडेलूप', 'hi-Latn': 'Guadeloupe', 'hr': 'Guadalupe', 'hsb': 'Guadeloupe', 'ht': 'Guadeloupe', 'hu': 'Guadeloupe', 'hy': 'Գվադելուպա', 'ia': 'Guadeloupe', 'id': 'Guadeloupe', 'ig': 'Guadeloupe', 'is': 'Gvadelúpeyjar', 'it': 'Guadalupa', 'ja': 'グアドループ', 'jmc': 'Gwadelupe', 'jv': 'Guadélup', 'ka': 'გვადელუპა', 'kaa': 'Гваделупа', 'kaa-Cyrl': 'Гваделупа', 'kab': 'Gwadalupi', 'kam': 'Gwadelupe', 'kde': 'Gwadelupe', 'kea': 'Guadalupi', 'kgp': 'Guvanarupe', 'khq': 'Gwadeluup', 'ki': 'Gwadelupe', 'kk': 'Гваделупа', 'kk-Cyrl': 'Гваделупа', 'kln': 'Emetab Guadeloupe', 'km': 'ហ្គោដឺឡុប', 'kn': 'ಗುಡೆಲೋಪ್', 'ko': '과들루프', 'kok': 'ग्वाडेलोप', 'kok-Deva': 'ग्वाडेलोप', 'ks': 'گواڈلوپ', 'ks-Arab': 'گواڈلوپ', 'ksb': 'Gwadelupe', 'ksf': 'gwadɛlúp', 'ksh': 'Juadelupp', 'ku': 'Guadeloupe', 'kxv': 'gvadelup', 'kxv-Deva': 'ग्वाडेलूप', 'kxv-Latn': 'gvadelup', 'kxv-Orya': 'ଗ୍ୱାଡେଲୁପ', 'kxv-Telu': 'గ్వడెలుప్', 'ky': 'Гваделупа', 'lag': 'Gwadelúupe', 'lb': 'Guadeloupe', 'lg': 'Gwadalupe', 'lld': 'Guadeloupe', 'ln': 'Gwadɛlupɛ', 'lo': 'ກົວດາລູບ', 'lt': 'Gvadelupa', 'lu': 'Ngwadelupe', 'luo': 'Guadeloupe', 'luy': 'Gwadelupe', 'lv': 'Gvadelupa', 'mai': 'ग्वाडेलूप', 'mas': 'Gwadelupe', 'mer': 'Gwadelupe', 'mfe': 'Guadloup', 'mg': 'Goadelopy', 'mgh': 'Ugwadelupe', 'mi': 'Kuatarupa', 'mk': 'Гвадалупе', 'ml': 'ഗ്വാഡലൂപ്പ്', 'mn': 'Гваделуп', 'mr': 'ग्वाडेलोउपे', 'ms': 'Guadeloupe', 'mt': 'Guadeloupe', 'mua': 'Sǝr Gwadeloupǝ', 'my': 'ဂွါဒီလု', 'mzn': 'گوادلوپ', 'naq': 'Guadeloupe', 'nb': 'Guadeloupe', 'nd': 'Guadeloupe', 'ne': 'ग्वाडेलुप', 'nl': 'Guadeloupe', 'nmg': 'Guadeloup', 'nn': 'Guadeloupe', 'no': 'Guadeloupe', 'nqo': 'ߜ߭ߎߥߊߘߋߟߎߔ', 'nyn': 'Gwaderupe', 'om': 'Gowadelowape', 'or': 'ଗୁଆଡେଲୋପ୍', 'pa': 'ਗੁਆਡੇਲੋਪ', 'pa-Guru': 'ਗੁਆਡੇਲੋਪ', 'pcm': 'Guadalúpẹ', 'pl': 'Gwadelupa', 'ps': 'ګوادلوپ', 'pt': 'Guadalupe', 'qu': 'Guadalupe', 'rm': 'Guadeloupe', 'rn': 'Gwadelupe', 'ro': 'Guadelupa', 'rof': 'Gwadelupe', 'ru': 'Гваделупа', 'rwk': 'Gwadelupe', 'saq': 'Gwadelupe', 'sbp': 'Gwadelupe', 'sc': 'Guadalupa', 'sd': 'گواڊیلوپ', 'sd-Arab': 'گواڊیلوپ', 'se': 'Guadeloupe', 'seh': 'Guadalupe', 'ses': 'Gwadeluup', 'sg': 'Guadelûpu', 'shi': 'ⴳⵡⴰⴷⴰⵍⵓⴱ', 'shi-Latn': 'gwadalub', 'shi-Tfng': 'ⴳⵡⴰⴷⴰⵍⵓⴱ', 'si': 'ග්වෝඩලෝප්', 'sk': 'Guadeloupe', 'sl': 'Guadeloupe', 'smn': 'Guadeloupe', 'sn': 'Guadeloupe', 'so': 'Guadeluub', 'sq': 'Guadelupë', 'sr': 'Гваделуп', 'sr-Cyrl': 'Гваделуп', 'sr-Latn': 'Gvadelup', 'sv': 'Guadeloupe', 'sw': 'Guadeloupe', 'syr': 'ܓܘܐܕܘܠܘܦܐܝ', 'ta': 'க்வாதேலோப்', 'te': 'గ్వాడెలోప్', 'teo': 'Gwadelupe', 'tg': 'Гваделупа', 'th': 'กวาเดอลูป', 'ti': 'ጓደሉፕ', 'tk': 'Gwadelupa', 'to': 'Kuatalupe', 'tr': 'Guadeloupe', 'tt': 'Гваделупа', 'twq': 'Gwadeluup', 'tzm': 'Gwadalup', 'ug': 'گىۋادېلۇپ', 'uk': 'Гваделупа', 'ur': 'گواڈیلوپ', 'uz': 'Gvadelupe', 'uz-Cyrl': 'Гваделупе', 'uz-Latn': 'Gvadelupe', 'vai': 'ꖶꕎꔐꖨꔅ', 'vai-Latn': 'Guwadelupe', 'vai-Vaii': 'ꖶꕎꔐꖨꔅ', 'vec': 'Guadalupa', 'vi': 'Guadeloupe', 'vun': 'Gwadelupe', 'wae': 'Guadeloupe', 'wo': 'Guwaadelup', 'xh': 'EGuadeloupe', 'xnr': 'ग्वाडेलूप', 'xog': 'Gwadalupe', 'yav': 'kuatelúup', 'yi': 'גוואַדעלופ', 'yo': 'Gadelope', 'yrl': 'Guadarupi', 'yue': '瓜地洛普', 'yue-Hans': '瓜地洛普', 'yue-Hant': '瓜地洛普', 'zgh': 'ⴳⵡⴰⴷⴰⵍⵓⴱ', 'zh': '瓜德罗普', 'zh-Hans': '瓜德罗普', 'zh-Hant': '瓜地洛普', 'zu': 'i-Guadeloupe'},
    'GQ': {'af': 'Ekwatoriaal-Guinee', 'agq': 'Èkwɛ̀tolia Ginè', 'ak': 'Gini Ikuweta', 'am': 'ኢኳቶሪያል ጊኒ', 'ar': 'غينيا الاستوائية', 'as': 'ইকুৱেটৰিয়েল গিনি', 'asa': 'Ginekweta', 'ast': 'Guinea Ecuatorial', 'az': 'Ekvatorial Qvineya', 'az-Cyrl': 'Екваториал Гвинеја', 'az-Latn': 'Ekvatorial Qvineya', 'bal-Latn': 'Ekwáturi Giniá', 'bas': 'Gìne Èkwàtorìà', 'be': 'Экватарыяльная Гвінея', 'be-tarask': 'Экватарыяльная Гвінея', 'bez': 'Huginekweta', 'bg': 'Екваториална Гвинея', 'blo': 'Ginee Malabo', 'bm': 'Gine ekwatɔri', 'bn': 'নিরক্ষীয় গিনি', 'br': 'Ginea ar Cʼheheder', 'brx': 'बिसुबारि गिनीया', 'bs': 'Ekvatorijalna Gvineja', 'bs-Cyrl': 'Екваторијална Гвинеја', 'bs-Latn': 'Ekvatorijalna Gvineja', 'ca': 'Guinea Equatorial', 'ca-ES-valencia': 'Guinea Equatorial', 'ccp': '𑄚𑄨𑄢𑄧𑄇𑄴𑄈𑄩𑄠𑄧 𑄉𑄨𑄚𑄨', 'ce': 'Экваторан Гвиней', 'ceb': 'Equatorial Guinea', 'cgg': 'Guni', 'chr': 'ᎡᏆᏙᎵᎠᎵ ᎩᎢᏂ', 'ckb': 'گینێی ئیستوایی', 'cs': 'Rovníková Guinea', 'cv': 'Экваториаллӑ Гвиней', 'cy': 'Gini Gyhydeddol', 'da': 'Ækvatorialguinea', 'dav': 'Ginekweta', 'de': 'Äquatorialguinea', 'dje': 'Ginee Ekwatorial', 'dsb': 'Ekwatorialna Gineja', 'dz': 'ཨེ་ཀུ་ཊོ་རེལ་ གི་ནི', 'ebu': 'Ginekweta', 'ee': 'Ekuatorial Guini nutome', 'el': 'Ισημερινή Γουινέα', 'el-polyton': 'Ἰσημερινὴ Γουινέα', 'en': 'Equatorial Guinea', 'en-Dsrt': '𐐇𐐿𐐶𐐲𐐻𐐱𐑉𐐨𐐲𐑊 𐐘𐐮𐑌𐐨', 'eo': 'Ekvatora Gvineo', 'es': 'Guinea Ecuatorial', 'et': 'Ekvatoriaal-Guinea', 'eu': 'Ekuatore Ginea', 'ewo': 'Giné Ekwató', 'fa': 'گینهٔ استوایی', 'ff': 'Ginee Ekuwaatoriyaal', 'ff-Adlm': '𞤘𞤭𞤲𞤫 𞤕𞤢𞤳𞤢𞤲𞤼𞤫𞥅𞤪𞤭', 'ff-Latn': 'Ginee Ekuwaatoriyaal', 'fi': 'Päiväntasaajan Guinea', 'fil': 'Equatorial Guinea', 'fo': 'Ekvatorguinea', 'fr': 'Guinée équatoriale', 'fur': 'Guinee ecuatoriâl', 'fy': 'Equatoriaal-Guinea', 'ga': 'an Ghuine Mheánchiorclach', 'gd': 'Gini Mheadhan-Chriosach', 'gl': 'Guinea Ecuatorial', 'gsw': 'Äquatoriaalgineea', 'gu': 'ઇક્વેટોરિયલ ગિની', 'guz': 'Ginekweta', 'ha': 'Ikwatoriyal Gini', 'he': 'גינאה המשוונית', 'hi': 'इक्वेटोरियल गिनी', 'hi-Latn': 'Equatorial Guinea', 'hr': 'Ekvatorska Gvineja', 'hsb': 'Ekwatorialna Gineja', 'ht': 'Guinée équatoriale', 'hu': 'Egyenlítői-Guinea', 'hy': 'Հասարակածային Գվինեա', 'ia': 'Guinea equatorial', 'id': 'Guinea Ekuatorial', 'ig': 'Equatorial Guinea', 'is': 'Miðbaugs-Gínea', 'it': 'Guinea Equatoriale', 'ja': '赤道ギニア', 'jgo': 'Ŋginɛ̂ Ɛkwatɔliyâl', 'jmc': 'Ginekweta', 'jv': 'Guinéa Katulistiwa', 'ka': 'ეკვატორული გვინეა', 'kaa': 'Экваториальлық Гвинея', 'kaa-Cyrl': 'Экваториальлық Гвинея', 'kab': 'Ɣinya Tasebgast', 'kam': 'Ginekweta', 'kde': 'Ginekweta', 'kea': 'Gine Ekuatorial', 'kgp': 'Ginẽ Ekuvatoriar', 'khq': 'Ginee Ekwatorial', 'ki': 'Ginekweta', 'kk': 'Экваторлық Гвинея', 'kk-Cyrl': 'Экваторлық Гвинея', 'kln': 'Emetab Equatorial Guinea', 'km': 'ហ្គីណេអេក្វាទ័រ', 'kn': 'ಈಕ್ವೆಟೋರಿಯಲ್ ಗಿನಿ', 'ko': '적도 기니', 'kok': 'इक्वेटोरियल गुएनिया', 'kok-Deva': 'इक्वेटोरियल गुएनिया', 'ks': 'اِکوِٹورِیَل گِنی', 'ks-Arab': 'اِکوِٹورِیَل گِنی', 'ksb': 'Ginekweta', 'ksf': 'ginɛ́ ɛkwatɔrial', 'ksh': 'Äquatorial Jineeja', 'ku': 'Gîneya Ekwadorê', 'kxv': 'ikveṭorial gini', 'kxv-Deva': 'इक्वेटोरियल गिनी', 'kxv-Latn': 'ikveṭorial gini', 'kxv-Orya': 'ଇକ୍ୟୁଏଟେରିୟଲ ଗିନି', 'kxv-Telu': 'ఈక్వటోరియల్ గినియా', 'ky': 'Экватордук Гвинея', 'lag': 'Gíine Ikwéeta', 'lb': 'Equatorialguinea', 'lg': 'Gayana ey’oku ekweta', 'lld': 'Guinea ecuatoriala', 'ln': 'Ginɛ́kwatɛ́lɛ', 'lo': 'ເອຄົວໂທຣຽວ ກີນີ', 'lt': 'Pusiaujo Gvinėja', 'lu': 'Gine Ekwatele', 'luo': 'Equatorial Guinea', 'luy': 'Ginekweta', 'lv': 'Ekvatoriālā Gvineja', 'mai': 'इक्वेटोरियल गिनी', 'mas': 'Ginekweta', 'mer': 'Gine ya Iquita', 'mfe': 'Gine ekwatoryal', 'mg': 'Guinea Ekoatera', 'mi': 'Kini Ekuatoria', 'mk': 'Екваторска Гвинеја', 'ml': 'ഇക്വറ്റോറിയൽ ഗിനിയ', 'mn': 'Экваторын Гвиней', 'mr': 'इक्वेटोरियल गिनी', 'ms': 'Guinea Khatulistiwa', 'mt': 'il-Guinea Ekwatorjali', 'mua': 'Sǝr Guine', 'my': 'အီကွေတာ ဂီနီ', 'mzn': 'استوایی گینه', 'naq': 'Equatorial Guineab', 'nb': 'Ekvatorial-Guinea', 'nd': 'Equatorial Guinea', 'ne': 'भू-मध्यीय गिनी', 'nl': 'Equatoriaal-Guinea', 'nmg': 'Guine Ekuatorial', 'nn': 'Ekvatorial-Guinea', 'no': 'Ekvatorial-Guinea', 'nqo': 'ߕߍߡߊߓߊ߲߮-ߖߌ߬ߣߍ߫', 'nyn': 'Guni', 'om': 'Ikkuwaatooriyaal Giinii', 'or': 'ଇକ୍ବାଟୋରିଆଲ୍ ଗୁଇନିଆ', 'pa': 'ਭੂ-ਖੰਡੀ ਗਿਨੀ', 'pa-Guru': 'ਭੂ-ਖੰਡੀ ਗਿਨੀ', 'pcm': 'Ikwétóriál Gíni', 'pl': 'Gwinea Równikowa', 'ps': 'استوایی ګیني', 'pt': 'Guiné Equatorial', 'qu': 'Guinea Ecuatorial', 'rm': 'Guinea Equatoriala', 'rn': 'Gineya Ekwatoriyali', 'ro': 'Guineea Ecuatorială', 'rof': 'Ginekweta', 'ru': 'Экваториальная Гвинея', 'rwk': 'Ginekweta', 'saq': 'Ginekweta', 'sbp': 'Ginekweta', 'sc': 'Guinea Ecuadoriale', 'sd': 'ايڪوٽوريل گائينا', 'sd-Arab': 'ايڪوٽوريل گائينا', 'se': 'Ekvatoriála Guinea', 'seh': 'Guiné Equatorial', 'ses': 'Ginee Ekwatorial', 'sg': 'Ginëe tî Ekuatëre', 'shi': 'ⵖⵉⵏⵢⴰ ⵏ ⵉⴽⵡⴰⴷⵓⵔ', 'shi-Latn': 'ɣinya n ikwadur', 'shi-Tfng': 'ⵖⵉⵏⵢⴰ ⵏ ⵉⴽⵡⴰⴷⵓⵔ', 'si': 'සමක ගිනියාව', 'sk': 'Rovníková Guinea', 'sl': 'Ekvatorialna Gvineja', 'smn': 'Peeivitäsideijee Guinea', 'sn': 'Equatorial Guinea', 'so': 'Ekuwatooriyal Gini', 'sq': 'Guineja Ekuatoriale', 'sr': 'Екваторијална Гвинеја', 'sr-Cyrl': 'Екваторијална Гвинеја', 'sr-Latn': 'Ekvatorijalna Gvineja', 'sv': 'Ekvatorialguinea', 'sw': 'Guinea ya Ikweta', 'syr': 'ܓܝܢܝܐ ܫܘܝܬܐ', 'ta': 'ஈக்வடோரியல் கினியா', 'te': 'ఈక్వటోరియల్ గినియా', 'teo': 'Ginekweta', 'tg': 'Гвинеяи Экваторӣ', 'th': 'อิเควทอเรียลกินี', 'ti': 'ኢኳቶርያል ጊኒ', 'tk': 'Ekwatorial Gwineýa', 'to': 'ʻEkueta Kini', 'tr': 'Ekvator Ginesi', 'tt': 'Экваториаль Гвинея', 'twq': 'Ginee Ekwatorial', 'tzm': 'Ɣinya Tikwaṭur it', 'ug': 'ئېكۋاتور گىۋىنىيەسى', 'uk': 'Екваторіальна Гвінея', 'ur': 'استوائی گیانا', 'uz': 'Ekvatorial Gvineya', 'uz-Cyrl': 'Экваториал Гвинея', 'uz-Latn': 'Ekvatorial Gvineya', 'vai': 'ꖦꕰꕊ ꗳ ꕅꔤꕇ', 'vai-Latn': 'Dúúnyá Tɛ Giini', 'vai-Vaii': 'ꖦꕰꕊ ꗳ ꕅꔤꕇ', 'vec': 'Guinèa Ecuatoriale', 'vi': 'Guinea Xích Đạo', 'vun': 'Ginekweta', 'wae': 'Equatorialginea', 'wo': 'Gine Ekuwatoriyal', 'xh': 'E-Equatorial Guinea', 'xnr': 'इक्वेटोरियल गिनी', 'xog': 'Gayana yaku ekweta', 'yav': 'kinéekuatolial', 'yi': 'עקוואַטארישע גינע', 'yo': 'Ekutoria Gini', 'yrl': 'Giné Ekuaturiyau', 'yue': '赤道幾內亞', 'yue-Hans': '赤道几内亚', 'yue-Hant': '赤道幾內亞', 'zgh': 'ⵖⵉⵏⵢⴰ ⵏ ⵉⴽⵡⴰⴷⵓⵔ', 'zh': '赤道几内亚', 'zh-Hans': '赤道几内亚', 'zh-Hant': '赤道幾內亞', 'zu': 'i-Equatorial Guinea'},
    'GR': {'af': 'Griekeland', 'agq': 'Gɨ̀lês', 'ak': 'Greekman', 'am': 'ግሪክ', 'ar': 'اليونان', 'as': 'গ্ৰীচ', 'asa': 'Ugiriki', 'ast': 'Grecia', 'az': 'Yunanıstan', 'az-Cyrl': 'Јунаныстан', 'az-Latn': 'Yunanıstan', 'bal-Latn': 'Yunán', 'bas': 'Grǐkyà', 'be': 'Грэцыя', 'be-tarask': 'Грэцыя', 'bez': 'Huugiliki', 'bg': 'Гърция', 'bgn': 'یونان', 'blo': 'Grɛs', 'bm': 'Gɛrɛsi', 'bn': 'গ্রীস', 'br': 'Gres', 'brx': 'ग्रीस', 'bs': 'Grčka', 'bs-Cyrl': 'Грчка', 'bs-Latn': 'Grčka', 'ca': 'Grècia', 'ca-ES-valencia': 'Grècia', 'ccp': '𑄉𑄳𑄢𑄨𑄌𑄴', 'ce': 'Греци', 'ceb': 'Greece', 'cgg': 'Guriisi', 'chr': 'ᎪᎢᎯ', 'ckb': 'یۆنان', 'cs': 'Řecko', 'cv': 'Греци', 'cy': 'Gwlad Groeg', 'da': 'Grækenland', 'dav': 'Ugiriki', 'de': 'Griechenland', 'dje': 'Greece', 'dsb': 'Grichiska', 'dyo': 'Gres', 'dz': 'གིརིས྄', 'ebu': 'Ugiriki', 'ee': 'Greece nutome', 'el': 'Ελλάδα', 'el-polyton': 'Ἑλλάδα', 'en': 'Greece', 'en-Dsrt': '𐐘𐑉𐐨𐑅', 'eo': 'Grekujo', 'es': 'Grecia', 'et': 'Kreeka', 'eu': 'Grezia', 'ewo': 'Gəlɛ́s', 'fa': 'یونان', 'ff': 'Gerees', 'ff-Adlm': '𞤀𞤤𞤴𞤵𞤲𞤢𞥄𞤲', 'ff-Latn': 'Gerees', 'fi': 'Kreikka', 'fil': 'Greece', 'fo': 'Grikkaland', 'fr': 'Grèce', 'fur': 'Grecie', 'fy': 'Grikelân', 'ga': 'an Ghréig', 'gd': 'A’ Ghreug', 'gl': 'Grecia', 'gsw': 'Griecheland', 'gu': 'ગ્રીસ', 'guz': 'Ugiriki', 'ha': 'Girka', 'haw': 'Helene', 'he': 'יוון', 'hi': 'यूनान', 'hi-Latn': 'Greece', 'hr': 'Grčka', 'hsb': 'Grjekska', 'ht': 'Grèce', 'hu': 'Görögország', 'hy': 'Հունաստան', 'ia': 'Grecia', 'id': 'Yunani', 'ie': 'Grecia', 'ig': 'Greece', 'is': 'Grikkland', 'it': 'Grecia', 'ja': 'ギリシャ', 'jgo': 'Ŋgɛlɛ̂k', 'jmc': 'Ugiriki', 'jv': 'Grikenlan', 'ka': 'საბერძნეთი', 'kaa': 'Греция', 'kaa-Cyrl': 'Греция', 'kab': 'Lagris', 'kam': 'Ugiriki', 'kde': 'Ugilichi', 'kea': 'Grésia', 'kgp': 'Grésa', 'khq': 'Greece', 'ki': 'Ugiriki', 'kk': 'Грекия', 'kk-Cyrl': 'Грекия', 'kln': 'Emetab Greece', 'km': 'ក្រិក', 'kn': 'ಗ್ರೀಸ್', 'ko': '그리스', 'kok': 'ग्रीस', 'kok-Deva': 'ग्रीस', 'kok-Latn': 'Gres', 'ks': 'گریٖس', 'ks-Arab': 'گریٖس', 'ksb': 'Ugiiki', 'ksf': 'grɛ́k', 'ksh': 'Jriescheland', 'ku': 'Yûnanistan', 'kxv': 'grīs', 'kxv-Deva': 'यूनान', 'kxv-Latn': 'grīs', 'kxv-Orya': 'ଗ୍ରିସ୍', 'kxv-Telu': 'గ్రీస్', 'ky': 'Греция', 'lag': 'Ugiríki', 'lb': 'Griicheland', 'lg': 'Bugereeki/Buyonaani', 'lld': 'Grecia', 'ln': 'Geleki', 'lo': 'ກຣີຊ', 'lt': 'Graikija', 'lu': 'Ngeleka', 'luo': 'Greece', 'luy': 'Ugiriki', 'lv': 'Grieķija', 'mai': 'यूनान', 'mas': 'Ugiriki', 'mer': 'Ngiriki', 'mfe': 'Gres', 'mg': 'Gresy', 'mi': 'Kirihi', 'mk': 'Грција', 'ml': 'ഗ്രീസ്', 'mn': 'Грек', 'mr': 'ग्रीस', 'ms': 'Greece', 'mt': 'il-Greċja', 'mua': 'Sǝr Grek', 'my': 'ဂရိ', 'mzn': 'یونان', 'naq': 'Xrikelandi', 'nb': 'Hellas', 'nd': 'Greece', 'ne': 'ग्रीस', 'nl': 'Griekenland', 'nmg': 'Grɛce', 'nn': 'Hellas', 'no': 'Hellas', 'nqo': 'ߜ߭ߙߍ߬ߞߌ߬', 'nyn': 'Guriisi', 'om': 'Giriik', 'or': 'ଗ୍ରୀସ୍', 'pa': 'ਗ੍ਰੀਸ', 'pa-Guru': 'ਗ੍ਰੀਸ', 'pcm': 'Gris', 'pl': 'Grecja', 'ps': 'یونان', 'pt': 'Grécia', 'qu': 'Grecia', 'rm': 'Grezia', 'rn': 'Ubugereki', 'ro': 'Grecia', 'rof': 'Ugiriki', 'ru': 'Греция', 'rwk': 'Ugiriki', 'saq': 'Ugiriki', 'sbp': 'Wugiliki', 'sc': 'Grètzia', 'sd': 'يونان', 'sd-Arab': 'يونان', 'se': 'Greika', 'seh': 'Grécia', 'ses': 'Greece', 'sg': 'Gerêsi', 'shi': 'ⵍⵢⵓⵏⴰⵏ', 'shi-Latn': 'lyunan', 'shi-Tfng': 'ⵍⵢⵓⵏⴰⵏ', 'si': 'ග්\u200dරීසිය', 'sk': 'Grécko', 'sl': 'Grčija', 'smn': 'Kreikka', 'sn': 'Greece', 'so': 'Giriig', 'sq': 'Greqi', 'sr': 'Грчка', 'sr-Cyrl': 'Грчка', 'sr-Latn': 'Grčka', 'sv': 'Grekland', 'sw': 'Ugiriki', 'syr': 'ܝܘܢ', 'ta': 'கிரீஸ்', 'te': 'గ్రీస్', 'teo': 'Ugiriki', 'tg': 'Юнон', 'th': 'กรีซ', 'ti': 'ግሪኽ', 'tk': 'Gresiýa', 'to': 'Kalisi', 'tr': 'Yunanistan', 'tt': 'Греция', 'twq': 'Greece', 'tzm': 'Yunan', 'ug': 'گىرېتسىيە', 'uk': 'Греція', 'ur': 'یونان', 'uz': 'Gretsiya', 'uz-Cyrl': 'Греция', 'uz-Latn': 'Gretsiya', 'vai': 'ꗥꗷꘋ', 'vai-Latn': 'Hɛlɛŋ', 'vai-Vaii': 'ꗥꗷꘋ', 'vec': 'Grecia', 'vi': 'Hy Lạp', 'vun': 'Ugiriki', 'wae': 'Gričeland', 'wo': 'Gerees', 'xh': 'EGreece', 'xnr': 'यूनान', 'xog': 'Buyonaani', 'yav': 'kilɛ́ɛk', 'yi': 'גריכנלאַנד', 'yo': 'Gíríìsì', 'yrl': 'Geresiya', 'yue': '希臘', 'yue-Hans': '希腊', 'yue-Hant': '希臘', 'zgh': 'ⵍⵢⵓⵏⴰⵏ', 'zh': '希腊', 'zh-Hans': '希腊', 'zh-Hant': '希臘', 'zu': 'i-Greece'},
    'GS': {'af': 'Suid-Georgië en die Suidelike Sandwicheilande', 'ak': 'Gyɔɔgyia Anaafoɔ ne Sandwich Aeland Anaafoɔ', 'am': 'ደቡብ ጆርጂያ እና የደቡብ ሳንድዊች ደሴቶች', 'ar': 'جورجيا الجنوبية وجزر ساندويتش الجنوبية', 'as': 'দক্ষিণ জৰ্জিয়া আৰু দক্ষিণ চেণ্ডৱিচ দ্বীপপুঞ্জ', 'ast': 'Islles Xeorxa del Sur y Sandwich del Sur', 'az': 'Cənubi Corciya və Cənubi Sendviç adaları', 'az-Cyrl': 'Ҹәнуби Ҹорҹија вә Ҹәнуби Сендвич адалары', 'az-Latn': 'Cənubi Corciya və Cənubi Sendviç adaları', 'bal-Latn': 'Zerbári Járjiá', 'be': 'Паўднёвая Георгія і Паўднёвыя Сандвічавы астравы', 'be-tarask': 'Паўднёвая Георгія і Паўднёвыя Сандвічавы астравы', 'bg': 'Южна Джорджия и Южни Сандвичеви острови', 'blo': 'Jɔrjiya gʊnyɩsonolaŋ kaja na Sanɖuush gʊnyɩsonolaŋ kaBʊtukǝltǝna', 'bn': 'দক্ষিণ জর্জিয়া ও দক্ষিণ স্যান্ডউইচ দ্বীপপুঞ্জ', 'br': 'Inizi Georgia ar Su hag Inizi Sandwich ar Su', 'brx': 'खोला जर्जिया आरो खोला सेण्डवीच द्वीपफोर', 'bs': 'Južna Džordžija i Južna Sendvič ostrva', 'bs-Cyrl': 'Јужна Џорџија и Јужна Сендвичка Острва', 'bs-Latn': 'Južna Džordžija i Južna Sendvič ostrva', 'ca': 'Illes Geòrgia del Sud i Sandwich del Sud', 'ca-ES-valencia': 'Illes Geòrgia del Sud i Sandwich del Sud', 'ccp': '𑄘𑄧𑄉𑄨𑄚𑄴 𑄎𑄧𑄢𑄴𑄎𑄨𑄠 𑄃𑄮 𑄘𑄧𑄉𑄨𑄚𑄴 𑄥𑄳𑄠𑄚𑄴𑄓𑄃𑄪𑄃𑄨𑄌𑄴 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Къилба Джорджи а, Къилба Гавайн гӀайренаш а', 'ceb': 'South Georgia & South Sandwich Islands', 'chr': 'ᏧᎦᏃᏮ ᏣᎠᏥᎢ ᎠᎴ ᎾᏍᎩ ᏧᎦᏃᏮ ᎠᏍᏛᎭᏟ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی جۆرجیا و ساندویچی باشوور', 'cs': 'Jižní Georgie a Jižní Sandwichovy ostrovy', 'cv': 'Кӑнтӑр Георги тата Сандвичев утравӗсем', 'cy': 'De Georgia ac Ynysoedd Sandwich y De', 'da': 'South Georgia og De Sydlige Sandwichøer', 'de': 'Südgeorgien und die Südlichen Sandwichinseln', 'dsb': 'Pódpołdnjowa Georgiska a Pódpołdnjowe Sandwichowe kupy', 'dz': 'སཱའུཐ་ཇཽར་ཇཱ་ དང་ སཱའུཐ་སེནཌ྄་ཝིཅ་གླིང་ཚོམ', 'ee': 'Anyiehe Georgia kple Anyiehe Sandwich ƒudomekpowo nutome', 'el': 'Νήσοι Νότια Γεωργία και Νότιες Σάντουιτς', 'el-polyton': 'Νήσοι Νότια Γεωργία και Νότιες Σάντουιτς', 'en': 'South Georgia & South Sandwich Islands', 'en-Dsrt': '𐐝𐐵𐑃 𐐖𐐱𐑉𐐾𐐲 𐐰𐑌𐐼 𐑄 𐐝𐐵𐑃 𐐝𐐰𐑌𐐼𐐶𐐮𐐽 𐐌𐑊𐐲𐑌𐐼𐑆', 'eo': 'Sud-Georgio kaj Sud-Sandviĉinsuloj', 'es': 'Islas Georgia del Sur y Sandwich del Sur', 'et': 'Lõuna-Georgia ja Lõuna-Sandwichi saared', 'eu': 'Hegoaldeko Georgia eta Hegoaldeko Sandwich uharteak', 'fa': 'جورجیای جنوبی و جزایر ساندویچ جنوبی', 'ff-Adlm': '𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫 𞤔𞤮𞤪𞤶𞤭𞤴𞤢 & 𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫 𞤅𞤢𞤲𞤣𞤵𞤱𞤭𞥅𞤷', 'fi': 'Etelä-Georgia ja Eteläiset Sandwichinsaaret', 'fil': 'South Georgia & South Sandwich Islands', 'fo': 'Suðurgeorgia og Suðursandwichoyggjar', 'fr': 'Géorgie du Sud-et-les Îles Sandwich du Sud', 'fur': 'Georgia dal Sud e Isulis Sandwich dal Sud', 'fy': 'Sûd-Georgia en Sûdlike Sandwicheilannen', 'ga': 'An tSeoirsia Theas agus Oileáin Sandwich Theas', 'gd': 'Seòirsea a Deas is na h-Eileanan Sandwich a Deas', 'gl': 'Illas Xeorxia do Sur e Sandwich do Sur', 'gsw': 'Süüdgeorgie und d’süüdlichi Sändwitsch-Insle', 'gu': 'દક્ષિણ જ્યોર્જિયા અને દક્ષિણ સેન્ડવિચ આઇલેન્ડ્સ', 'ha': 'Kudancin Geogia da Kudancin Tsibirin Sandiwic', 'he': 'ג׳ורג׳יה הדרומית ואיי סנדוויץ׳ הדרומיים', 'hi': 'दक्षिण जॉर्जिया और दक्षिण सैंडविच द्वीपसमूह', 'hi-Latn': 'South Georgia & South Sandwich Islands', 'hr': 'Južna Georgia i Otoci Južni Sandwich', 'hsb': 'Južna Georgiska a Južne Sandwichowe kupy', 'ht': 'Géorgie du Sud-et-les Îles Sandwich du Sud', 'hu': 'Déli-Georgia és Déli-Sandwich-szigetek', 'hy': 'Հարավային Ջորջիա և Հարավային Սենդվիչյան կղզիներ', 'ia': 'Georgia del Sud e Insulas Sandwich Austral', 'id': 'Georgia Selatan & Kep. Sandwich Selatan', 'ig': 'South Georgia & South Sandwich Islands', 'is': 'Suður-Georgía og Suður-Sandvíkureyjar', 'it': 'Georgia del Sud e Sandwich Australi', 'ja': 'サウスジョージア・サウスサンドウィッチ諸島', 'jv': 'Georgia Kidul lan Kapuloan Sandwich Kidul', 'ka': 'სამხრეთ ჯორჯია და სამხრეთ სენდვიჩის კუნძულები', 'kaa': 'Қубла Георгия ҳәм Қубла Сэндвич атаўлары', 'kaa-Cyrl': 'Қубла Георгия ҳәм Қубла Сэндвич атаўлары', 'kea': 'Ilhas Jeórjia di Sul i Sanduixi di Sul', 'kgp': 'Jiórja-Sur kar Sỹnvisi-Sur Goj-vẽso Ag', 'kk': 'Оңтүстік Георгия және Оңтүстік Сандвич аралдары', 'kk-Cyrl': 'Оңтүстік Георгия және Оңтүстік Сандвич аралдары', 'km': 'កោះ\u200bហ្សកហ្ស៊ី\u200bខាងត្បូង និង សង់វិច\u200bខាងត្បូង', 'kn': 'ದಕ್ಷಿಣ ಜಾರ್ಜಿಯಾ ಮತ್ತು ದಕ್ಷಿಣ ಸ್ಯಾಂಡ್\u200dವಿಚ್ ದ್ವೀಪಗಳು', 'ko': '사우스조지아 사우스샌드위치 제도', 'kok': 'दक्षिण जोर्जिया आनी दक्षिण सॅण्डविच आयलँड्स', 'kok-Deva': 'दक्षिण जोर्जिया आनी दक्षिण सॅण्डविच आयलँड्स', 'ks': 'جنوٗبی جارجِیا تہٕ جنوٗبی سینڑوٕچ جٔزیٖرٕ', 'ks-Arab': 'جنوٗبی جارجِیا تہٕ جنوٗبی سینڑوٕچ جٔزیٖرٕ', 'ksh': 'Söd-Jeorjie un de södlijje Botteramms-Enselle', 'ku': 'Giravên Georgîyaya Başûr û Sandwicha Başûr', 'kxv': 'dakiṇ jārjiā aḍe dakiṇ sandwich dīp', 'kxv-Deva': 'दकिण जॉर्जिया अड़े दकिण सैंडविच दीप', 'kxv-Latn': 'dakiṇ jārjiā aḍe dakiṇ sandwich dīp', 'kxv-Orya': 'ଦକିନ ଜର୍ଜିଆ ଅଡ଼େ ଦକିନ ସାଣ୍ଡୱିଚ୍ ଦୀପ', 'kxv-Telu': 'దకిన జర్జిఆ అదే దకిన సండవిచ్ దిప', 'ky': 'Түштүк Жоржия жана Түштүк Сэндвич аралдары', 'lb': 'Südgeorgien an déi Südlech Sandwichinselen', 'lld': 'Georgia dl Süd y Isoles Sandwich dl Süd', 'ln': 'Îles de Géorgie du Sud et Sandwich du Sud', 'lo': 'ໝູ່ເກາະ ຈໍເຈຍຕອນໃຕ້ ແລະ ແຊນວິດຕອນໃຕ້', 'lt': 'Pietų Džordžija ir Pietų Sandvičo salos', 'lv': 'Dienviddžordžija un Dienvidsendviču salas', 'mai': 'दक्षिण जार्जिया आ दक्षिण सैंडविच द्वीपसमूह', 'mi': 'Hōria ki te Tonga me ngā Motu Hanawiti ki te Tonga', 'mk': 'Јужна Џорџија и Јужни Сендвички Острови', 'ml': 'ദക്ഷിണ ജോർജ്ജിയയും ദക്ഷിണ സാൻഡ്\u200cവിച്ച് ദ്വീപുകളും', 'mn': 'Өмнөд Жоржиа ба Өмнөд Сэндвичийн арлууд', 'mr': 'दक्षिण जॉर्जिया आणि दक्षिण सँडविच बेटे', 'ms': 'Kepulauan Georgia Selatan & Sandwich Selatan', 'mt': 'il-Georgia tan-Nofsinhar u l-Gżejjer Sandwich tan-Nofsinhar', 'my': 'တောင် ဂျော်ဂျီယာ နှင့် တောင် ဆင်းဒဝစ်ဂျ် ကျွန်းစုများ', 'mzn': 'جنوبی جورجیا و جنوبی ساندویچ جزایر', 'nb': 'Sør-Georgia og Sør-Sandwichøyene', 'ne': 'दक्षिण जर्जिया र दक्षिण स्यान्डवीच टापुहरू', 'nl': 'Zuid-Georgia en Zuidelijke Sandwicheilanden', 'nn': 'Sør-Georgia og Sør-Sandwichøyane', 'no': 'Sør-Georgia og Sør-Sandwichøyene', 'nqo': 'ߖ߭ߐߙߑߖ߭ߌ߫ ߥߙߏ߬ߘߎ߮ ߣߌ߫ ߛߊ߲ߘߎߥߌߛ ߕߌ߲', 'om': 'Joorjikaa Kibba fi Odoloota Saanduwiich Kibbaa', 'or': 'ଦକ୍ଷିଣ ଜର୍ଜିଆ ଏବଂ ଦକ୍ଷିଣ ସାଣ୍ଡୱିଚ୍ ଦ୍ୱୀପପୁଞ୍ଜ', 'pa': 'ਦੱਖਣੀ ਜਾਰਜੀਆ ਅਤੇ ਦੱਖਣੀ ਸੈਂਡਵਿਚ ਟਾਪੂ', 'pa-Guru': 'ਦੱਖਣੀ ਜਾਰਜੀਆ ਅਤੇ ਦੱਖਣੀ ਸੈਂਡਵਿਚ ਟਾਪੂ', 'pcm': 'Saút Jọ́jia an Saút Sándwích Aílands', 'pl': 'Georgia Południowa i Sandwich Południowy', 'ps': 'سويلي جارجيا او سويلي سېنډوچ ټاپوګان', 'pt': 'Ilhas Geórgia do Sul e Sandwich do Sul', 'qu': 'Georgia del Sur e Islas Sandwich del Sur', 'rm': 'Georgia dal Sid e las Inslas Sandwich dal Sid', 'ro': 'Georgia de Sud și Insulele Sandwich de Sud', 'ru': 'Южная Георгия и Южные Сандвичевы о-ва', 'sc': 'Geòrgia de su Sud e Ìsulas Sandwich Australes', 'sd': 'ڏکڻ جارجيا ۽ ڏکڻ سينڊوچ ٻيٽ', 'sd-Arab': 'ڏکڻ جارجيا ۽ ڏکڻ سينڊوچ ٻيٽ', 'se': 'Lulli Georgia ja Lulli Sandwich-sullot', 'si': 'දකුණු ජෝර්ජියාව සහ දකුණු සැන්ඩ්විච් දූපත්', 'sk': 'Južná Georgia a Južné Sandwichove ostrovy', 'sl': 'Južna Georgia in Južni Sandwichevi otoki', 'smn': 'Maadâ-Georgia já Máddááh Sandwichsuolluuh', 'so': 'Jasiiradda Joorjiyada Koonfureed & Sandwij', 'sq': 'Xhorxha Jugore dhe Ishujt Senduiçë të Jugut', 'sr': 'Јужна Џорџија и Јужна Сендвичка Острва', 'sr-Cyrl': 'Јужна Џорџија и Јужна Сендвичка Острва', 'sr-Latn': 'Južna Džordžija i Južna Sendvička Ostrva', 'sv': 'Sydgeorgien och Sydsandwichöarna', 'sw': 'Visiwa vya Georgia Kusini na Sandwich Kusini', 'syr': 'ܓܙܝܖ̈ܐ ܕܓܘܪܓܝܐ ܘܣܐܢܕܘܝܟ ܬܝܡܢܝ̈ܐ', 'ta': 'தெற்கு ஜார்ஜியா மற்றும் தெற்கு சாண்ட்விச் தீவுகள்', 'te': 'దక్షిణ జార్జియా మరియు దక్షిణ శాండ్విచ్ దీవులు', 'tg': 'Ҷорҷияи Ҷанубӣ ва Ҷазираҳои Сандвич', 'th': 'เกาะเซาท์จอร์เจียและหมู่เกาะเซาท์แซนด์วิช', 'ti': 'ደሴታት ደቡብ ጆርጅያን ደቡብ ሳንድዊችን', 'tk': 'Günorta Georgiýa we Günorta Sendwiç adasy', 'to': 'ʻOtumotu Seōsia-tonga mo Saniuisi-tonga', 'tr': 'Güney Georgia ve Güney Sandwich Adaları', 'tt': 'Көньяк Георгия һәм Көньяк Сандвич утраулары', 'ug': 'جەنۇبىي جورجىيە ۋە جەنۇبىي ساندۋىچ ئاراللىرى', 'uk': 'Південна Джорджія та Південні Сандвічеві Острови', 'ur': 'جنوبی جارجیا اور جنوبی سینڈوچ جزائر', 'uz': 'Janubiy Georgiya va Janubiy Sendvich orollari', 'uz-Cyrl': 'Жанубий Георгия ва Жанубий Сендвич ороллари', 'uz-Latn': 'Janubiy Georgiya va Janubiy Sendvich orollari', 'vai': 'ꗘꖺꕀꕩ ꗛꔤ ꔒꘋꗣ ꗏ ꗪ ꗇꖢ ꔳꘋꗣ ꗛꔤ ꔒꘋꗣ ꗏ', 'vai-Vaii': 'ꗘꖺꕀꕩ ꗛꔤ ꔒꘋꗣ ꗏ ꗪ ꗇꖢ ꔳꘋꗣ ꗛꔤ ꔒꘋꗣ ꗏ', 'vec': 'Georgia de’l sud e Sandwich de’l sud', 'vi': 'Nam Georgia & Quần đảo Nam Sandwich', 'wae': 'Südgeorgie und d’südliče Senwičinslä', 'wo': 'Seworsi di Sid ak Duni Sàndwiis di Sid', 'xh': 'ESouth Georgia & South Sandwich Islands', 'xnr': 'दखण जॉर्जिया कने दखण सैंडविच द्वीपां दा झुंड', 'yo': 'Gúúsù Georgia àti Gúúsù Àwọn Erékùsù Sandwich', 'yrl': 'Kapuãma-ita Geugiya Su asuí Sãduwixi Su', 'yue': '南佐治亞島同南桑威奇群島', 'yue-Hans': '南佐治亚岛同南桑威奇群岛', 'yue-Hant': '南佐治亞島同南桑威奇群島', 'zh': '南乔治亚和南桑威奇群岛', 'zh-Hans': '南乔治亚和南桑威奇群岛', 'zh-Hant': '南喬治亞與南三明治群島', 'zu': 'i-South Georgia ne-South Sandwich Islands'},
    'GT': {'af': 'Guatemala', 'agq': 'Gwàtɨ̀malà', 'ak': 'Guwatemala', 'am': 'ጉዋቲማላ', 'ar': 'غواتيمالا', 'as': 'গুৱাটেমালা', 'asa': 'Gwatemala', 'ast': 'Guatemala', 'az': 'Qvatemala', 'az-Cyrl': 'Гватемала', 'az-Latn': 'Qvatemala', 'bal-Latn': 'Gwátémálá', 'bas': 'Gwàtèmalà', 'be': 'Гватэмала', 'be-tarask': 'Гватэмала', 'bez': 'Hugwatemala', 'bg': 'Гватемала', 'blo': 'Guwatemalaa', 'bm': 'Gwatemala', 'bn': 'গুয়াতেমালা', 'br': 'Guatemala', 'brx': 'गुवाटेमाला', 'bs': 'Gvatemala', 'bs-Cyrl': 'Гватемала', 'bs-Latn': 'Gvatemala', 'ca': 'Guatemala', 'ca-ES-valencia': 'Guatemala', 'ccp': '𑄉𑄪𑄠𑄖𑄬𑄟𑄣', 'ce': 'Гватемала', 'ceb': 'Guatemala', 'cgg': 'Gwatemara', 'chr': 'ᏩᏔᎹᎳ', 'ckb': 'گواتیمالا', 'cs': 'Guatemala', 'cv': 'Гватемала', 'cy': 'Guatemala', 'da': 'Guatemala', 'dav': 'Gwatemala', 'de': 'Guatemala', 'dje': 'Gwatemaala', 'dsb': 'Guatemala', 'dyo': 'Guatemala', 'dz': 'གྷོ་ཊ་མ་ལ', 'ebu': 'Gwatemala', 'ee': 'Guatemala nutome', 'el': 'Γουατεμάλα', 'el-polyton': 'Γουατεμάλα', 'en': 'Guatemala', 'en-Dsrt': '𐐘𐐶𐐪𐐼𐐲𐑋𐐪𐑊𐐲', 'eo': 'Gvatemalo', 'es': 'Guatemala', 'et': 'Guatemala', 'eu': 'Guatemala', 'ewo': 'Guatemalá', 'fa': 'گواتمالا', 'ff': 'Gwaatemalaa', 'ff-Adlm': '𞤘𞤵𞤱𞤢𞤼𞤫𞤥𞤢𞤤𞤢𞥄', 'ff-Latn': 'Gwaatemalaa', 'fi': 'Guatemala', 'fil': 'Guatemala', 'fo': 'Guatemala', 'fr': 'Guatemala', 'fur': 'Guatemala', 'fy': 'Guatemala', 'ga': 'Guatamala', 'gd': 'Guatamala', 'gl': 'Guatemala', 'gsw': 'Guatemaala', 'gu': 'ગ્વાટેમાલા', 'guz': 'Gwatemala', 'ha': 'Gwatamala', 'he': 'גואטמלה', 'hi': 'ग्वाटेमाला', 'hi-Latn': 'Guatemala', 'hr': 'Gvatemala', 'hsb': 'Guatemala', 'ht': 'Guatemala', 'hu': 'Guatemala', 'hy': 'Գվատեմալա', 'ia': 'Guatemala', 'id': 'Guatemala', 'ig': 'Guatemala', 'is': 'Gvatemala', 'it': 'Guatemala', 'ja': 'グアテマラ', 'jmc': 'Gwatemala', 'jv': 'Guatémala', 'ka': 'გვატემალა', 'kaa': 'Гватемала', 'kaa-Cyrl': 'Гватемала', 'kab': 'Gwatimala', 'kam': 'Gwatemala', 'kde': 'Gwatemala', 'kea': 'Guatimala', 'kgp': 'Guvatimỹra', 'khq': 'Gwatemaala', 'ki': 'Gwatemala', 'kk': 'Гватемала', 'kk-Cyrl': 'Гватемала', 'kln': 'Emetab Guatemala', 'km': 'ក្វាតេម៉ាឡា', 'kn': 'ಗ್ವಾಟೆಮಾಲಾ', 'ko': '과테말라', 'kok': 'ग्वाटेमाला', 'kok-Deva': 'ग्वाटेमाला', 'ks': 'گواٹمالا', 'ks-Arab': 'گواٹمالا', 'ksb': 'Gwatemala', 'ksf': 'gwátǝmala', 'ksh': 'Juwatemahla', 'ku': 'Guatemala', 'kxv': 'gvaṭemala', 'kxv-Deva': 'ग्वाटेमाला', 'kxv-Latn': 'gvaṭemala', 'kxv-Orya': 'ଗ୍ୱାଟେମାଲା', 'kxv-Telu': 'గ్వటెమాలా', 'ky': 'Гватемала', 'lag': 'Gwatemáala', 'lb': 'Guatemala', 'lg': 'Gwatemala', 'lld': 'Guatemala', 'ln': 'Gwatémala', 'lo': 'ກົວເທມາລາ', 'lt': 'Gvatemala', 'lu': 'Ngwatemala', 'luo': 'Guatemala', 'luy': 'Gwatemala', 'lv': 'Gvatemala', 'mai': 'ग्वाटेमाला', 'mas': 'Gwatemala', 'mer': 'Gwatemala', 'mfe': 'Guatemala', 'mg': 'Goatemalà', 'mgh': 'Ugwatemala', 'mi': 'Kuatamāra', 'mk': 'Гватемала', 'ml': 'ഗ്വാട്ടിമാല', 'mn': 'Гватемал', 'mr': 'ग्वाटेमाला', 'ms': 'Guatemala', 'mt': 'il-Gwatemala', 'mua': 'Gwatemala', 'my': 'ဂွါတီမာလာ', 'mzn': 'گواتمالا', 'naq': 'Guatemala', 'nb': 'Guatemala', 'nd': 'Guatemala', 'ne': 'ग्वाटेमाला', 'nl': 'Guatemala', 'nmg': 'Guatemala', 'nn': 'Guatemala', 'no': 'Guatemala', 'nqo': 'ߜ߭ߎߥߊߕߋߡߟߊ߫', 'nyn': 'Gwatemara', 'om': 'Guwaatimaalaa', 'or': 'ଗୁଏତମାଲା', 'pa': 'ਗੁਆਟੇਮਾਲਾ', 'pa-Guru': 'ਗੁਆਟੇਮਾਲਾ', 'pcm': 'Guátẹmála', 'pl': 'Gwatemala', 'ps': 'ګواتیمالا', 'pt': 'Guatemala', 'qu': 'Guatemala', 'rm': 'Guatemala', 'rn': 'Gwatemala', 'ro': 'Guatemala', 'rof': 'Gwatemala', 'ru': 'Гватемала', 'rwk': 'Gwatemala', 'saq': 'Gwatemala', 'sbp': 'Gwatemala', 'sc': 'Guatemala', 'sd': 'گوئٽي مالا', 'sd-Arab': 'گوئٽي مالا', 'se': 'Guatemala', 'seh': 'Guatemala', 'ses': 'Gwatemaala', 'sg': 'Guatêmälä', 'shi': 'ⴳⵡⴰⵜⵉⵎⴰⵍⴰ', 'shi-Latn': 'gwatimala', 'shi-Tfng': 'ⴳⵡⴰⵜⵉⵎⴰⵍⴰ', 'si': 'ගෝතමාලාව', 'sk': 'Guatemala', 'sl': 'Gvatemala', 'smn': 'Guatemala', 'sn': 'Guatemala', 'so': 'Guwaatamaala', 'sq': 'Guatemalë', 'sr': 'Гватемала', 'sr-Cyrl': 'Гватемала', 'sr-Latn': 'Gvatemala', 'sv': 'Guatemala', 'sw': 'Guatemala', 'syr': 'ܓܘܐܬܝܡܐܠܐ', 'ta': 'கவுதமாலா', 'te': 'గ్వాటిమాలా', 'teo': 'Gwatemala', 'tg': 'Гватемала', 'th': 'กัวเตมาลา', 'ti': 'ጓቲማላ', 'tk': 'Gwatemala', 'to': 'Kuatamala', 'tr': 'Guatemala', 'tt': 'Гватемала', 'twq': 'Gwatemaala', 'tzm': 'Gwatimala', 'ug': 'گىۋاتېمالا', 'uk': 'Гватемала', 'ur': 'گواٹے مالا', 'uz': 'Gvatemala', 'uz-Cyrl': 'Гватемала', 'uz-Latn': 'Gvatemala', 'vai': 'ꖶꕎꔎꕮꕞ', 'vai-Latn': 'Guwatɛmala', 'vai-Vaii': 'ꖶꕎꔎꕮꕞ', 'vec': 'Guatemala', 'vi': 'Guatemala', 'vun': 'Gwatemala', 'wae': 'Guatemala', 'wo': 'Guwatemala', 'xh': 'EGuatemala', 'xnr': 'ग्वाटेमाला', 'xog': 'Gwatemala', 'yav': 'kuatemalá', 'yi': 'גוואַטעמאַלע', 'yo': 'Guatemálà', 'yrl': 'Guatemara', 'yue': '瓜地馬拉', 'yue-Hans': '瓜地马拉', 'yue-Hant': '瓜地馬拉', 'zgh': 'ⴳⵡⴰⵜⵉⵎⴰⵍⴰ', 'zh': '危地马拉', 'zh-Hans': '危地马拉', 'zh-Hant': '瓜地馬拉', 'zu': 'i-Guatemala'},
    'GU': {'af': 'Guam', 'agq': 'Gwam', 'ak': 'Guam', 'am': 'ጉዋም', 'ar': 'غوام', 'as': 'গুৱাম', 'asa': 'Gwam', 'ast': 'Guam', 'az': 'Quam', 'az-Cyrl': 'Гуам', 'az-Latn': 'Quam', 'bal-Latn': 'Guám', 'bas': 'Gùâm', 'be': 'Гуам', 'be-tarask': 'Гуам', 'bez': 'Hugwam', 'bg': 'Гуам', 'blo': 'Guwam', 'bm': 'Gwam', 'bn': 'গুয়াম', 'br': 'Guam', 'brx': 'गुआम', 'bs': 'Guam', 'bs-Cyrl': 'Гуам', 'bs-Latn': 'Guam', 'ca': 'Guam', 'ca-ES-valencia': 'Guam', 'ccp': '𑄉𑄪𑄠𑄟𑄴', 'ce': 'Гуам', 'ceb': 'Guam', 'cgg': 'Gwamu', 'chr': 'ᏆᎻ', 'ckb': 'گوام', 'cs': 'Guam', 'cv': 'Гуам', 'cy': 'Guam', 'da': 'Guam', 'dav': 'Gwam', 'de': 'Guam', 'dje': 'Guam', 'dsb': 'Guam', 'dyo': 'Guam', 'dz': 'གུ་འམ་ མཚོ་གླིང', 'ebu': 'Gwam', 'ee': 'Guam nutome', 'el': 'Γκουάμ', 'el-polyton': 'Γκουάμ', 'en': 'Guam', 'en-Dsrt': '𐐘𐐶𐐪𐑋', 'eo': 'Gvamo', 'es': 'Guam', 'et': 'Guam', 'eu': 'Guam', 'ewo': 'Guám', 'fa': 'گوام', 'ff': 'Guwam', 'ff-Adlm': '𞤘𞤵𞤱𞤢𞥄𞤥', 'ff-Latn': 'Guwam', 'fi': 'Guam', 'fil': 'Guam', 'fo': 'Guam', 'fr': 'Guam', 'fur': 'Guam', 'fy': 'Guam', 'ga': 'Guam', 'gd': 'Guam', 'gl': 'Guam', 'gsw': 'Guam', 'gu': 'ગ્વામ', 'guz': 'Gwam', 'ha': 'Guam', 'he': 'גואם', 'hi': 'गुआम', 'hi-Latn': 'Guam', 'hr': 'Guam', 'hsb': 'Guam', 'ht': 'Guam', 'hu': 'Guam', 'hy': 'Գուամ', 'ia': 'Guam', 'id': 'Guam', 'ig': 'Guam', 'is': 'Gvam', 'it': 'Guam', 'ja': 'グアム', 'jmc': 'Gwam', 'jv': 'Guam', 'ka': 'გუამი', 'kaa': 'Гуам', 'kaa-Cyrl': 'Гуам', 'kab': 'Gwam', 'kam': 'Gwam', 'kde': 'Gwam', 'kea': 'Guam', 'kgp': 'Guvỹm', 'khq': 'Guam', 'ki': 'Gwam', 'kk': 'Гуам', 'kk-Cyrl': 'Гуам', 'kln': 'Emetab Guam', 'km': 'ហ្គាំ', 'kn': 'ಗುವಾಮ್', 'ko': '괌', 'kok': 'गुआम', 'kok-Deva': 'गुआम', 'ks': 'گُوام', 'ks-Arab': 'گُوام', 'ksb': 'Gwam', 'ksf': 'gwám', 'ksh': 'Juham', 'ku': 'Guam', 'kxv': 'guām', 'kxv-Deva': 'गुआम', 'kxv-Latn': 'guām', 'kxv-Orya': 'ଗୁଆମ୍', 'kxv-Telu': 'గ్వమ్', 'ky': 'Гуам', 'lag': 'Gwani', 'lb': 'Guam', 'lg': 'Gwamu', 'lld': 'Guam', 'ln': 'Gwamɛ', 'lo': 'ກວາມ', 'lt': 'Guamas', 'lu': 'Ngwame', 'luo': 'Guam', 'luy': 'Gwam', 'lv': 'Guama', 'mai': 'गुआम', 'mas': 'Gwam', 'mer': 'Gwam', 'mfe': 'Guam', 'mg': 'Guam', 'mgh': 'Ugwam', 'mi': 'Kuama', 'mk': 'Гуам', 'ml': 'ഗ്വാം', 'mn': 'Гуам', 'mr': 'गुआम', 'ms': 'Guam', 'mt': 'Guam', 'mua': 'Gwam', 'my': 'ဂူအမ်', 'mzn': 'گوئام', 'naq': 'Guam', 'nb': 'Guam', 'nd': 'Guam', 'ne': 'गुवाम', 'nl': 'Guam', 'nmg': 'Guam', 'nn': 'Guam', 'no': 'Guam', 'nqo': 'ߜ߭ߎߥߊߡ', 'nyn': 'Gwamu', 'om': 'Guwama', 'or': 'ଗୁଆମ୍', 'pa': 'ਗੁਆਮ', 'pa-Guru': 'ਗੁਆਮ', 'pcm': 'Guám', 'pl': 'Guam', 'ps': 'ګوام', 'pt': 'Guam', 'qu': 'Guam', 'rm': 'Guam', 'rn': 'Gwamu', 'ro': 'Guam', 'rof': 'Gwam', 'ru': 'Гуам', 'rwk': 'Gwam', 'saq': 'Gwam', 'sbp': 'Gwamu', 'sc': 'Guàm', 'sd': 'گوام', 'sd-Arab': 'گوام', 'se': 'Guam', 'seh': 'Guam', 'ses': 'Guam', 'sg': 'Guâm', 'shi': 'ⴳⵡⴰⵎ', 'shi-Latn': 'gwam', 'shi-Tfng': 'ⴳⵡⴰⵎ', 'si': 'ගුවාම්', 'sk': 'Guam', 'sl': 'Guam', 'smn': 'Guam', 'sn': 'Guam', 'so': 'Guaam', 'sq': 'Guam', 'sr': 'Гуам', 'sr-Cyrl': 'Гуам', 'sr-Latn': 'Guam', 'sv': 'Guam', 'sw': 'Guam', 'syr': 'ܓܘܐܡ', 'ta': 'குவாம்', 'te': 'గ్వామ్', 'teo': 'Gwam', 'tg': 'Гуам', 'th': 'กวม', 'ti': 'ጓም', 'tk': 'Guam', 'to': 'Kuamu', 'tr': 'Guam', 'tt': 'Гуам', 'twq': 'Guam', 'tzm': 'Gwam', 'ug': 'گۇئام', 'uk': 'Гуам', 'ur': 'گوام', 'uz': 'Guam', 'uz-Cyrl': 'Гуам', 'uz-Latn': 'Guam', 'vai': 'ꖶꕎꕆ', 'vai-Latn': 'Guwami', 'vai-Vaii': 'ꖶꕎꕆ', 'vec': 'Guam', 'vi': 'Guam', 'vun': 'Gwam', 'wae': 'Guam', 'wo': 'Guwam', 'xh': 'EGuam', 'xnr': 'गुआम', 'xog': 'Gwamu', 'yav': 'kuamiɛ', 'yi': 'גוואַם', 'yo': 'Guamu', 'yrl': 'Guwã', 'yue': '關島', 'yue-Hans': '关岛', 'yue-Hant': '關島', 'zgh': 'ⴳⵡⴰⵎ', 'zh': '关岛', 'zh-Hans': '关岛', 'zh-Hant': '關島', 'zu': 'i-Guam'},
    'GW': {'af': 'Guinee-Bissau', 'agq': 'Ginè Bìsawù', 'ak': 'Gini Bisaw', 'am': 'ጊኒ-ቢሳው', 'ar': 'غينيا بيساو', 'as': 'গিনি-বিছাও', 'asa': 'Ginebisau', 'ast': 'Guinea-Bisáu', 'az': 'Qvineya-Bisau', 'az-Cyrl': 'Гвинеја-Бисау', 'az-Latn': 'Qvineya-Bisau', 'bal-Latn': 'Giniá-Bissáu', 'bas': 'Gìne Bìsàô', 'be': 'Гвінея-Бісау', 'be-tarask': 'Гвінея-Бісау', 'bez': 'Huginebisau', 'bg': 'Гвинея-Бисау', 'blo': 'Ginee Bisoo', 'bm': 'Gine Bisawo', 'bn': 'গিনি-বিসাউ', 'br': 'Ginea-Bissau', 'brx': 'गिनीया- बिसाऊ', 'bs': 'Gvineja-Bisao', 'bs-Cyrl': 'Гвинеја-Бисау', 'bs-Latn': 'Gvineja-Bisao', 'ca': 'Guinea Bissau', 'ca-ES-valencia': 'Guinea Bissau', 'ccp': '𑄉𑄨𑄚𑄨-𑄝𑄨𑄥𑄃𑄪', 'ce': 'Гвиней-Бисау', 'ceb': 'Guinea-Bissau', 'cgg': 'Ginebisau', 'chr': 'ᎩᎢᏂ-ᏈᏌᎤᏫ', 'ckb': 'گینێ بیساو', 'cs': 'Guinea-Bissau', 'cv': 'Гвиней-Бисау', 'cy': 'Guiné-Bissau', 'da': 'Guinea-Bissau', 'dav': 'Ginebisau', 'de': 'Guinea-Bissau', 'dje': 'Gine-Bisso', 'dsb': 'Gineja-Bissau', 'dyo': 'Giné Bisaau', 'dz': 'གྷི་ནི་ བྷི་སཱའུ', 'ebu': 'Ginebisau', 'ee': 'Gini-Bisao nutome', 'el': 'Γουινέα Μπισάου', 'el-polyton': 'Γουινέα Μπισάου', 'en': 'Guinea-Bissau', 'en-Dsrt': '𐐘𐐮𐑌𐐨-𐐒𐐮𐑅𐐵', 'eo': 'Gvineo-Bisaŭo', 'es': 'Guinea-Bisáu', 'et': 'Guinea-Bissau', 'eu': 'Ginea Bissau', 'ewo': 'Giné Bisaó', 'fa': 'گینهٔ بیسائو', 'ff': 'Gine-Bisaawo', 'ff-Adlm': '𞤘𞤭𞤲𞤫-𞤄𞤭𞤧𞤢𞤱𞤮𞥅', 'ff-Latn': 'Gine-Bisaawo', 'fi': 'Guinea-Bissau', 'fil': 'Guinea-Bissau', 'fo': 'Guinea-Bissau', 'fr': 'Guinée-Bissau', 'fur': 'Guinea-Bissau', 'fy': 'Guinee-Bissau', 'ga': 'Guine Bissau', 'gd': 'Gini-Bioso', 'gl': 'Guinea Bissau', 'gsw': 'Gineea-Bissau', 'gu': 'ગિની-બિસાઉ', 'guz': 'Ginebisau', 'ha': 'Gini Bisau', 'he': 'גינאה-ביסאו', 'hi': 'गिनी-बिसाउ', 'hi-Latn': 'Guinea-Bissau', 'hr': 'Gvineja Bisau', 'hsb': 'Gineja-Bissau', 'ht': 'Guinée-Bissau', 'hu': 'Bissau-Guinea', 'hy': 'Գվինեա-Բիսաու', 'ia': 'Guinea-Bissau', 'id': 'Guinea-Bissau', 'ig': 'Guinea-Bissau', 'is': 'Gínea-Bissá', 'it': 'Guinea-Bissau', 'ja': 'ギニアビサウ', 'jgo': 'Ŋginɛ̂ Mbisáwu', 'jmc': 'Ginebisau', 'jv': 'Guinea-Bissau', 'ka': 'გვინეა-ბისაუ', 'kaa': 'Гвинея-Бисау', 'kaa-Cyrl': 'Гвинея-Бисау', 'kab': 'Ɣinya-Bisaw', 'kam': 'Ginebisau', 'kde': 'Ginebisau', 'kea': 'Gine-Bisau', 'kgp': 'Ginẽ-Misav', 'khq': 'Gine-Bisso', 'ki': 'Ginebisau', 'kk': 'Гвинея-Бисау', 'kk-Cyrl': 'Гвинея-Бисау', 'kln': 'Emetab Guinea-Bissau', 'km': 'ហ្គីណេប៊ីស្សូ', 'kn': 'ಗಿನಿ-ಬಿಸ್ಸಾವ್', 'ko': '기니비사우', 'kok': 'गुअनिया-बिसाउ', 'kok-Deva': 'गुअनिया-बिसाउ', 'ks': 'گیٖنی بِساو', 'ks-Arab': 'گیٖنی بِساو', 'ksb': 'Ginebisau', 'ksf': 'ginɛ́ bisɔ́', 'ksh': 'Jinneha_Bißau', 'ku': 'Gîne-Bissau', 'kxv': 'gini-bisau', 'kxv-Deva': 'गिनी-बिसाउ', 'kxv-Latn': 'gini-bisau', 'kxv-Orya': 'ଗିନି ବିସାଉ', 'kxv-Telu': 'గినియా-బిస్సావ్', 'ky': 'Гвинея-Бисау', 'lag': 'Gíine Bisáau', 'lb': 'Guinea-Bissau', 'lg': 'Gini-Bisawu', 'lld': 'Guinea-Bissau', 'ln': 'Ginɛbisau', 'lo': 'ກິນີ-ບິສເຊົາ', 'lt': 'Bisau Gvinėja', 'lu': 'Nginebisau', 'luo': 'Guinea-Bissau', 'luy': 'Ginebisau', 'lv': 'Gvineja-Bisava', 'mai': 'गिनी-बिसाऊ', 'mas': 'Ginebisau', 'mer': 'Ginebisau', 'mfe': 'Gine-Bisau', 'mg': 'Giné-Bisao', 'mgh': 'Uginebisau', 'mi': 'Kini-Pihao', 'mk': 'Гвинеја Бисао', 'ml': 'ഗിനിയ-ബിസൗ', 'mn': 'Гвиней-Бисау', 'mr': 'गिनी-बिसाउ', 'ms': 'Guinea Bissau', 'mt': 'il-Guinea-Bissau', 'mua': 'Guine ma Bissao', 'my': 'ဂီနီ-ဘီစော', 'mzn': 'گینه بیسائو', 'naq': 'Guinea-Bissau', 'nb': 'Guinea-Bissau', 'nd': 'Guinea-Bissau', 'ne': 'गिनी-बिसाउ', 'nl': 'Guinee-Bissau', 'nmg': 'Guine Bisso', 'nn': 'Guinea-Bissau', 'no': 'Guinea-Bissau', 'nqo': 'ߖߌ߬ߣߍ߫ ߓߌߛߊߥߏ߫', 'nyn': 'Ginebisau', 'om': 'Giinii-Bisaawoo', 'or': 'ଗୁଇନିଆ-ବିସାଉ', 'pa': 'ਗਿਨੀ-ਬਿਸਾਉ', 'pa-Guru': 'ਗਿਨੀ-ਬਿਸਾਉ', 'pcm': 'Gíní-Bisáu', 'pl': 'Gwinea Bissau', 'ps': 'ګینه بیسو', 'pt': 'Guiné-Bissau', 'qu': 'Guinea-Bisáu', 'rm': 'Guinea-Bissau', 'rn': 'Gineya Bisawu', 'ro': 'Guineea-Bissau', 'rof': 'Ginebisau', 'ru': 'Гвинея-Бисау', 'rwk': 'Ginebisau', 'saq': 'Ginebisau', 'sbp': 'Ginebisawu', 'sc': 'Guinea-Bissau', 'sd': 'گني بسائو', 'sd-Arab': 'گني بسائو', 'se': 'Guinea-Bissau', 'seh': 'Guiné Bissau', 'ses': 'Gine-Bisso', 'sg': 'Gninëe-Bisau', 'shi': 'ⵖⵉⵏⵢⴰ ⴱⵉⵙⴰⵡ', 'shi-Latn': 'ɣinya bisaw', 'shi-Tfng': 'ⵖⵉⵏⵢⴰ ⴱⵉⵙⴰⵡ', 'si': 'ගිනි බිසව්', 'sk': 'Guinea-Bissau', 'sl': 'Gvineja Bissau', 'smn': 'Guinea-Bissau', 'sn': 'Guinea-Bissau', 'so': 'Gini-Bisaaw', 'sq': 'Guine-Bisau', 'sr': 'Гвинеја-Бисао', 'sr-Cyrl': 'Гвинеја-Бисао', 'sr-Latn': 'Gvineja-Bisao', 'sv': 'Guinea-Bissau', 'sw': 'Ginebisau', 'syr': 'ܓܝܢܝܐ ܒܝܣܐܘ', 'ta': 'கினியா-பிஸ்ஸாவ்', 'te': 'గినియా-బిస్సావ్', 'teo': 'Ginebisau', 'tg': 'Гвинея-Бисау', 'th': 'กินี-บิสเซา', 'ti': 'ጊኒ-ቢሳው', 'tk': 'Gwineýa-Bisau', 'to': 'Kini-Pisau', 'tr': 'Gine-Bissau', 'tt': 'Гвинея-Бисау', 'twq': 'Gine-Bisso', 'tzm': 'Ɣinya-Bissaw', 'ug': 'گىۋىنىيە بىسسائۇ', 'uk': 'Гвінея-Бісау', 'ur': 'گنی بساؤ', 'uz': 'Gvineya-Bisau', 'uz-Cyrl': 'Гвинея-Бисау', 'uz-Latn': 'Gvineya-Bisau', 'vai': 'ꕅꔤꕇ ꔫꕢꕴ', 'vai-Latn': 'Gini Bhisawo', 'vai-Vaii': 'ꕅꔤꕇ ꔫꕢꕴ', 'vec': 'Guinèa Bisào', 'vi': 'Guinea-Bissau', 'vun': 'Ginebisau', 'wae': 'Ginea Bissau', 'wo': 'Gine-Bisaawóo', 'xh': 'EGuinea-Bissau', 'xnr': 'गिनी-बिसाउ', 'xog': 'Gini-Bisawu', 'yav': 'kiinépisaó', 'yi': 'גינע־ביסאַו', 'yo': 'Gene-Busau', 'yrl': 'Giné Bisau', 'yue': '幾內亞比索', 'yue-Hans': '几内亚比索', 'yue-Hant': '幾內亞比索', 'zgh': 'ⵖⵉⵏⵢⴰ ⴱⵉⵙⴰⵡ', 'zh': '几内亚比绍', 'zh-Hans': '几内亚比绍', 'zh-Hant': '幾內亞比索', 'zu': 'i-Guinea-Bissau'},
    'GY': {'af': 'Guyana', 'agq': 'Gùyanà', 'ak': 'Gayana', 'am': 'ጉያና', 'ar': 'غيانا', 'as': 'গায়ানা', 'asa': 'Guyana', 'ast': 'Guyana', 'az': 'Qayana', 'az-Cyrl': 'Гајана', 'az-Latn': 'Qayana', 'bal-Latn': 'Goyáná', 'bas': 'Gùyanà', 'be': 'Гаяна', 'be-tarask': 'Гаяна', 'bez': 'Huguyana', 'bg': 'Гаяна', 'bgn': 'گویانا', 'blo': 'Guyanaa', 'bm': 'Gwiyana', 'bn': 'গিয়ানা', 'br': 'Guyana', 'brx': 'गुयाना', 'bs': 'Gvajana', 'bs-Cyrl': 'Гвајана', 'bs-Latn': 'Gvajana', 'ca': 'Guyana', 'ca-ES-valencia': 'Guyana', 'ccp': '𑄉𑄨𑄠𑄚', 'ce': 'Гайана', 'ceb': 'Guyana', 'cgg': 'Guyana', 'chr': 'ᎦᏯᎾ', 'ckb': 'گویانا', 'cs': 'Guyana', 'cv': 'Гайана', 'cy': 'Guyana', 'da': 'Guyana', 'dav': 'Guyana', 'de': 'Guyana', 'dje': 'Guyaane', 'dsb': 'Guyana', 'dyo': 'Giyan', 'dz': 'གྷ་ཡ་ན', 'ebu': 'Guyana', 'ee': 'Guyanadu', 'el': 'Γουιάνα', 'el-polyton': 'Γουιάνα', 'en': 'Guyana', 'en-Dsrt': '𐐘𐐴𐐰𐑌𐐲', 'eo': 'Gujano', 'es': 'Guyana', 'et': 'Guyana', 'eu': 'Guyana', 'ewo': 'Guyán', 'fa': 'گویان', 'ff': 'Giyaan', 'ff-Adlm': '𞤘𞤢𞤴𞤢𞤲𞤢𞥄', 'ff-Latn': 'Giyaan', 'fi': 'Guyana', 'fil': 'Guyana', 'fo': 'Gujana', 'fr': 'Guyana', 'fur': 'Guyana', 'fy': 'Guyana', 'ga': 'An Ghuáin', 'gd': 'Guidheàna', 'gl': 'Güiana', 'gsw': 'Guäjaana', 'gu': 'ગયાના', 'guz': 'Guyana', 'ha': 'Guyana', 'he': 'גיאנה', 'hi': 'गुयाना', 'hi-Latn': 'Guyana', 'hr': 'Gvajana', 'hsb': 'Guyana', 'ht': 'Guyana', 'hu': 'Guyana', 'hy': 'Գայանա', 'ia': 'Guyana', 'id': 'Guyana', 'ie': 'Guyana', 'ig': 'Guyana', 'is': 'Gvæjana', 'it': 'Guyana', 'ja': 'ガイアナ', 'jmc': 'Guyana', 'jv': 'Guyana', 'ka': 'გაიანა', 'kaa': 'Гайана', 'kaa-Cyrl': 'Гайана', 'kab': 'Guwana', 'kam': 'Guyana', 'kde': 'Guyana', 'kea': 'Giana', 'kgp': 'Gijỹnỹ', 'khq': 'Guyaane', 'ki': 'Guyana', 'kk': 'Гайана', 'kk-Cyrl': 'Гайана', 'kln': 'Emetab Guyana', 'km': 'ហ្គីយ៉ាន', 'kn': 'ಗಯಾನಾ', 'ko': '가이아나', 'kok': 'गयाना', 'kok-Deva': 'गयाना', 'ks': 'گُیانا', 'ks-Arab': 'گُیانا', 'ksb': 'Guyana', 'ksf': 'guyán', 'ksh': 'Jujaana', 'ku': 'Guyana', 'kxv': 'guyana', 'kxv-Deva': 'गुयाना', 'kxv-Latn': 'guyana', 'kxv-Orya': 'ଗୁୟାନା', 'kxv-Telu': 'గుయానా', 'ky': 'Гайана', 'lag': 'Guyáana', 'lb': 'Guyana', 'lg': 'Gayana', 'lld': 'Guyana', 'ln': 'Giyane', 'lo': 'ກາຍຢານາ', 'lt': 'Gajana', 'lu': 'Ngiyane', 'luo': 'Guyana', 'luy': 'Guyana', 'lv': 'Gajāna', 'mai': 'गुयाना', 'mas': 'Guyana', 'mer': 'Guyana', 'mfe': 'Guyana', 'mg': 'Guyana', 'mgh': 'Uguyana', 'mi': 'Kaiana', 'mk': 'Гвајана', 'ml': 'ഗയാന', 'mn': 'Гайана', 'mr': 'गयाना', 'ms': 'Guyana', 'mt': 'il-Guyana', 'mua': 'Guyana', 'my': 'ဂိုင်ယာနာ', 'mzn': 'گویان', 'naq': 'Guyana', 'nb': 'Guyana', 'nd': 'Guyana', 'ne': 'गुयाना', 'nl': 'Guyana', 'nmg': 'Guyana', 'nn': 'Guyana', 'no': 'Guyana', 'nqo': 'ߜ߭ߎߦߊߣ', 'nyn': 'Guyana', 'om': 'Guyaanaa', 'or': 'ଗୁଇନା', 'pa': 'ਗੁਯਾਨਾ', 'pa-Guru': 'ਗੁਯਾਨਾ', 'pcm': 'Gayána', 'pl': 'Gujana', 'ps': 'ګیانا', 'pt': 'Guiana', 'qu': 'Guyana', 'rm': 'Guyana', 'rn': 'Guyane', 'ro': 'Guyana', 'rof': 'Guyana', 'ru': 'Гайана', 'rwk': 'Guyana', 'saq': 'Guyana', 'sbp': 'Guyana', 'sc': 'Guyana', 'sd': 'گيانا', 'sd-Arab': 'گيانا', 'se': 'Guyana', 'seh': 'Guiana', 'ses': 'Guyaane', 'sg': 'Gayâna', 'shi': 'ⴳⵡⵉⵢⴰⵏⴰ', 'shi-Latn': 'gwiyana', 'shi-Tfng': 'ⴳⵡⵉⵢⴰⵏⴰ', 'si': 'ගයනාව', 'sk': 'Guyana', 'sl': 'Gvajana', 'smn': 'Guyana', 'sn': 'Guyana', 'so': 'Guyana', 'sq': 'Guajanë', 'sr': 'Гвајана', 'sr-Cyrl': 'Гвајана', 'sr-Latn': 'Gvajana', 'sv': 'Guyana', 'sw': 'Guyana', 'syr': 'ܓܘܝܐܢܐ', 'ta': 'கயானா', 'te': 'గయానా', 'teo': 'Guyana', 'tg': 'Гайана', 'th': 'กายอานา', 'ti': 'ጉያና', 'tk': 'Gaýana', 'to': 'Kuiana', 'tr': 'Guyana', 'tt': 'Гайана', 'twq': 'Guyaane', 'tzm': 'Guyana', 'ug': 'گىۋىيانا', 'uk': 'Гаяна', 'ur': 'گیانا', 'uz': 'Gayana', 'uz-Cyrl': 'Гаяна', 'uz-Latn': 'Gayana', 'vai': 'ꖶꕩꕯ', 'vai-Latn': 'Guyana', 'vai-Vaii': 'ꖶꕩꕯ', 'vec': 'Guyana', 'vi': 'Guyana', 'vun': 'Guyana', 'wae': 'Guyana', 'wo': 'Giyaan', 'xh': 'EGuyana', 'xnr': 'गुयाना', 'xog': 'Gayana', 'yav': 'kuyáan', 'yi': 'גויאַנע', 'yo': 'Guyana', 'yrl': 'Giyana', 'yue': '蓋亞那', 'yue-Hans': '盖亚那', 'yue-Hant': '蓋亞那', 'zgh': 'ⴳⵡⵉⵢⴰⵏⴰ', 'zh': '圭亚那', 'zh-Hans': '圭亚那', 'zh-Hant': '蓋亞那', 'zu': 'i-Guyana'},
    'Gara': {'bal-Latn': 'Garay', 'en': 'Garay', 'et': 'garai', 'fo': 'garay', 'gd': 'Garay', 'hi-Latn': 'Garay', 'ig': 'Garay', 'vec': 'garay'},
    'Geok': {'ar': 'الأبجدية الجورجية - أسومتافرلي و نسخري', 'ast': 'khutsuri xeorxanu', 'az': 'gürcü xutsuri', 'az-Latn': 'gürcü xutsuri', 'bg': 'Грузинска хуцури', 'bn': 'জর্জিয় খুৎসুরি', 'brx': 'जोर्जीयन खुतसुरी', 'bs': 'gruzijsko khutsuri pismo', 'bs-Cyrl': 'грузијско кхутсури писмо', 'bs-Latn': 'gruzijsko khutsuri pismo', 'ca': 'georgià hucuri', 'ca-ES-valencia': 'georgià hucuri', 'ccp': '𑄎𑄧𑄢𑄴𑄎𑄨𑄠𑄧 𑄈𑄪𑄖𑄴𑄥𑄪𑄢𑄨', 'cs': 'gruzínské chutsuri', 'da': 'georgisk kutsuri', 'de': 'Khutsuri', 'el': 'Γεωργιανό Κχουτσούρι', 'el-polyton': 'Γεωργιανό Κχουτσούρι', 'en': 'Georgian Khutsuri', 'en-Dsrt': '𐐖𐐱𐑉𐐾𐐲𐑌 𐐗𐐳𐐻𐑅𐐭𐑉𐐨', 'es': 'georgiano eclesiástico', 'et': 'hutsuri', 'fa': 'گرجی خوتسوری', 'fi': 'muinaisgeorgialainen', 'fr': 'géorgien khoutsouri', 'fy': 'Georgysk Khutsuri', 'gsw': 'Ghutsuri', 'gu': 'જ્યોર્જિઅન ખુતસુરી', 'hi': 'जॉर्जियन खुतसुरी', 'hi-Latn': 'Georgian Khutsuri', 'hr': 'gruzijsko khutsuri pismo', 'ht': 'géorgien khoutsouri', 'hu': 'Grúz kucsuri', 'id': 'Georgian Khutsuri', 'it': 'kutsuri', 'ja': 'ジョージア文字(フツリ)', 'ka': 'ხუცური', 'kgp': 'kehunsuri geórgijỹnũ', 'kn': 'ಜಾರ್ಜಿಯನ್ ಖುಸ್ತುರಿ', 'ko': '그루지야 쿠츠리 문자', 'ks': 'جارجِیَن کھتسوری', 'ks-Arab': 'جارجِیَن کھتسوری', 'lb': 'Khutsuri', 'lo': 'ຄອດຊູຮີຈໍເຈຍ', 'lt': 'gruzinų kutsuri', 'mk': 'грузиски хуцури', 'ml': 'ജോർജ്ജിയൻ ഖുട്സുരി', 'mr': 'जॉर्जियन खुत्सुरी', 'nb': 'georgisk khutsuri', 'ne': 'ग्रुजियाली खुट्सुरी', 'nl': 'Georgisch Khutsuri', 'nn': 'khutsuri (asomtavruli og nuskhuri)', 'no': 'georgisk khutsuri', 'or': 'ଜର୍ଜିଆନ୍ ଖୁଟସୁରୀ', 'pl': 'gruzińskie chucuri', 'pt': 'khutsuri georgiano', 'rm': 'kutsuri', 'ro': 'georgiană bisericească', 'ru': 'грузинская хуцури', 'sl': 'cerkvenogruzijski', 'sr': 'грузијско кхутсури писмо', 'sr-Cyrl': 'грузијско кхутсури писмо', 'sr-Latn': 'gruzijsko khutsuri pismo', 'sv': 'kutsuri', 'ta': 'ஜியார்ஜியன் குட்சுரி', 'te': 'జార్జియన్ ఖట్సూరి', 'th': 'คัตซูรีจอร์เจีย', 'to': 'tohinima fakakutusuli-seōsia', 'tr': 'Hutsuri Gürcü', 'ug': 'خۇتسۇرى گىرۇزىنچە', 'uk': 'кхутсурі', 'vi': 'Chữ Khutsuri Georgia', 'yrl': 'kutisuri geurgianu', 'yue': '喬治亞語系（阿索他路里和努斯克胡里文）', 'yue-Hans': '乔治亚语系（阿索他路里和努斯克胡里文）', 'yue-Hant': '喬治亞語系（阿索他路里和努斯克胡里文）', 'zh': '格鲁吉亚文（教堂体）', 'zh-Hans': '格鲁吉亚文（教堂体）', 'zh-Hant': '喬治亞語系（阿索他路里和努斯克胡里文）'},
    'Geor': {'af': 'Georgies', 'ak': 'Dwɔɔgyia kasa', 'am': 'ጆርጂያኛ', 'ar': 'الجورجية', 'as': 'জৰ্জিয়ান', 'ast': 'xeorxanu', 'az': 'gürcü', 'az-Arab': 'گۆرجوجه', 'az-Latn': 'gürcü', 'bal-Latn': 'Járj', 'be': 'грузінскае', 'be-tarask': 'грузінскае', 'bg': 'грузинска', 'bgn': 'گورجی', 'bn': 'জর্জিয়ান', 'br': 'jorjianek', 'brx': 'जर्जियान', 'bs': 'gruzijsko pismo', 'bs-Cyrl': 'грузијско писмо', 'bs-Latn': 'gruzijsko pismo', 'ca': 'georgià', 'ca-ES-valencia': 'georgià', 'ccp': '𑄎𑄧𑄢𑄴𑄎𑄨𑄠𑄚𑄴', 'ce': 'гуьржийн', 'chr': 'ᏦᏥᎠᏂ', 'ckb': 'گورجی', 'cs': 'gruzínské', 'cy': 'Georgaidd', 'da': 'georgisk', 'de': 'Georgisch', 'dsb': 'georgiski', 'dz': 'ཇཽ་ཇི་ཡཱན་ཡིག་གུ', 'ee': 'gɔgiagbeŋɔŋlɔ', 'el': 'Γεωργιανό', 'el-polyton': 'Γεωργιανό', 'en': 'Georgian', 'en-Dsrt': '𐐖𐐬𐑉𐐾𐐲𐑌', 'es': 'georgiano', 'et': 'gruusia', 'eu': 'georgiarra', 'fa': 'گرجی', 'ff-Adlm': '𞤔𞤮𞤪𞤶𞤭𞤴𞤢𞤲', 'fi': 'georgialainen', 'fil': 'Georgian', 'fo': 'georgianskt', 'fr': 'géorgien', 'fur': 'georgjian', 'fy': 'Georgysk', 'ga': 'Seoirseach', 'gd': 'Cairtbheilis', 'gl': 'xeorxiano', 'gsw': 'Georgisch', 'gu': 'જ્યોર્જિઅન', 'ha': 'Georgian', 'he': 'גאורגי', 'hi': 'जॉर्जियन', 'hi-Latn': 'Georgian', 'hr': 'gruzijsko pismo', 'hsb': 'georgisce', 'ht': 'géorgien', 'hu': 'Grúz', 'hy': 'վրացական', 'ia': 'georgian', 'id': 'Georgia', 'ig': 'Mkpụrụ ọkwụ Geọjịan', 'is': 'georgískt', 'it': 'georgiano', 'ja': 'ジョージア文字', 'jv': 'Georgia', 'ka': 'ქართული', 'kea': 'jorjianu', 'kgp': 'geórgijỹnũ', 'kk': 'грузин жазуы', 'kk-Cyrl': 'грузин жазуы', 'km': 'ហ្សកហ្ស៊ី', 'kn': 'ಜಾರ್ಜಿಯನ್', 'ko': '조지아 문자', 'kok': 'जॉर्जियन', 'kok-Deva': 'जॉर्जियन', 'ks': 'جارجِیَن', 'ks-Arab': 'جارجِیَن', 'ksh': 'je’orrjesche Schreff', 'ku': 'gurcî', 'ky': 'Грузин', 'lb': 'Georgesch', 'lo': 'ຈໍຈຽນ', 'lrc': 'گورجی', 'lt': 'gruzinų', 'lv': 'gruzīnu', 'mai': 'जार्जियन', 'mi': 'Hōriana', 'mk': 'грузиско писмо', 'ml': 'ജോർജ്ജിയൻ', 'mn': 'гүрж', 'mr': 'जॉर्जियन', 'ms': 'Georgia', 'my': 'ဂျော်ဂျီယာ', 'mzn': 'گرجی', 'nb': 'georgisk', 'ne': 'जर्जियाली', 'nl': 'Georgisch', 'nn': 'georgisk', 'no': 'georgisk', 'nqo': 'ߖ߭ߐߙߑߖ߭ߌߞߊ', 'or': 'ଜର୍ଜିୟ', 'pa': 'ਜਾਰਜੀਆਈ', 'pa-Guru': 'ਜਾਰਜੀਆਈ', 'pcm': 'Jọ́jia', 'pl': 'gruzińskie', 'ps': 'ګرجستاني', 'pt': 'georgiano', 'qu': 'Georgiano', 'rm': 'georgian', 'ro': 'georgiană', 'ru': 'грузинская', 'sc': 'georgianu', 'sd': 'جيورجيائي', 'sd-Arab': 'جيورجيائي', 'si': 'ජෝර්ජියානු', 'sk': 'gruzínske', 'sl': 'gruzijski', 'so': 'Joorjiya', 'sq': 'gjeorgjian', 'sr': 'грузијско писмо', 'sr-Cyrl': 'грузијско писмо', 'sr-Latn': 'gruzijsko pismo', 'sv': 'georgiska', 'sw': 'Kijojia', 'syr': 'ܓܘܪܓܝܬ', 'ta': 'ஜார்ஜியன்', 'te': 'జార్జియన్', 'tg': 'Гурҷӣ', 'th': 'จอร์เจีย', 'ti': 'ናይ ጆርጅያ', 'tk': 'Gruzin elipbiýi', 'to': 'tohinima fakaseōsia', 'tr': 'Gürcü', 'ug': 'گىرۇزىنچە', 'uk': 'грузинська', 'ur': 'جارجیائی', 'uz': 'gruzin', 'uz-Cyrl': 'Грузин', 'uz-Latn': 'gruzin', 'vec': 'jeorjan', 'vi': 'Chữ Georgia', 'wae': 'Georgiš', 'xnr': 'जॉर्जियन', 'yo': 'Jọ́jíànù', 'yrl': 'geurgianu', 'yue': '喬治亞文', 'yue-Hans': '乔治亚文', 'yue-Hant': '喬治亞文', 'zh': '格鲁吉亚文', 'zh-Hans': '格鲁吉亚文', 'zh-Hant': '喬治亞文', 'zu': 'isi-Georgian'},
    'Glag': {'ar': 'الجلاجوليتيك', 'ast': 'glagolíticu', 'az': 'qlaqolitik', 'az-Latn': 'qlaqolitik', 'bal-Latn': 'Galgoliti', 'bg': 'Глаголическа', 'bn': 'গ্লাগোলিটিক', 'br': 'glagolitek', 'brx': 'ग्लैगोलिटीक', 'bs': 'glagoljica', 'bs-Cyrl': 'глагољица', 'bs-Latn': 'glagoljica', 'ca': 'glagolític', 'ca-ES-valencia': 'glagolític', 'ccp': '𑄉𑄳𑄣𑄉𑄮𑄣𑄨𑄑𑄨𑄇𑄴', 'cs': 'hlaholice', 'da': 'glagolitisk', 'de': 'Glagolitisch', 'el': 'Γκλαγκολιτικό', 'el-polyton': 'Γκλαγκολιτικό', 'en': 'Glagolitic', 'en-Dsrt': '𐐘𐑊𐐰𐑀𐐬𐑊𐐮𐐻𐐮𐐿', 'es': 'glagolítico', 'et': 'glagoolitsa', 'eu': 'glagolitikera', 'fa': 'گلاگولیتی', 'ff-Adlm': '𞤘𞤭𞤤𞤢𞤺𞤮𞤤𞤭𞤼𞤭𞤳', 'fi': 'glagoliittinen', 'fo': 'glagolitic', 'fr': 'glagolitique', 'fur': 'glagolitic', 'fy': 'Glagolitysk', 'ga': 'Glagalach', 'gd': 'Glagoliticeach', 'gsw': 'Glagolitisch', 'gu': 'ગ્લેગોલિટિક', 'he': 'גלגוליטי', 'hi': 'ग्लेगोलिटिक', 'hi-Latn': 'Glagolitic', 'hr': 'glagoljica', 'ht': 'glagolitique', 'hu': 'Glagolitikus', 'id': 'Glagolitic', 'ig': 'Glagolitic', 'it': 'glagolitico', 'ja': 'グラゴル文字', 'ka': 'გლაგოლიცა', 'kgp': 'gragoritiko', 'kk': 'глаголица', 'kk-Cyrl': 'глаголица', 'kn': 'ಗ್ಲಾಗೋಲಿಟಿಕ್', 'ko': '글라골 문자', 'ks': 'گلیگولِٹِک', 'ks-Arab': 'گلیگولِٹِک', 'lb': 'Glagolitesch', 'lo': 'ກລາໂກລິຕິກ', 'lt': 'glagolitik', 'mk': 'глаголица', 'ml': 'ഗ്ലഗോലിറ്റിക്', 'mn': 'Глаголитик', 'mr': 'ग्लॅगोलिटिक', 'ms': 'Glagolitik', 'nb': 'glagolittisk', 'ne': 'ग्लागोलिटिक', 'nl': 'Glagolitisch', 'nn': 'glagolittisk', 'no': 'glagolittisk', 'nqo': 'ߜ߭ߑߟߊߜ߭ߏߟߌߕߌߞ', 'or': 'ଗ୍ଲାଗ୍ଲୋଟିକ୍', 'pl': 'głagolica', 'pt': 'glagolítico', 'rm': 'glagolitic', 'ro': 'glagolitică', 'ru': 'глаголица', 'sc': 'glagolìticu', 'sk': 'hlaholika', 'sl': 'glagoliški', 'so': 'Qoraalka Glagolitic', 'sq': 'glagolitikisht', 'sr': 'глагољица', 'sr-Cyrl': 'глагољица', 'sr-Latn': 'glagoljica', 'sv': 'glagolitiska', 'ta': 'க்லாகோலிடிக்', 'te': 'గ్లాగో లిటిక్', 'tg': 'Глаголитикӣ', 'th': 'กลาโกลิติก', 'to': 'tohinima fakakalakoliti', 'tr': 'Glagolit', 'ug': 'گىلاگوتچە', 'uk': 'глаголичний', 'vec': 'glagolìtego', 'vi': 'Chữ Glagolitic', 'xnr': 'ग्लेगोलिटिक', 'yrl': 'garagurítiku', 'yue': '格拉哥里文', 'yue-Hans': '格拉哥里文', 'yue-Hant': '格拉哥里文', 'zh': '格拉哥里文', 'zh-Hans': '格拉哥里文', 'zh-Hant': '格拉哥里文', 'zu': 'isi-Glagolitic'},
    'Gong': {'bal-Latn': 'Gunjála Góndi', 'bs': 'gundžala gondi pismo', 'bs-Cyrl': 'гуњала гонди писмо', 'bs-Latn': 'gundžala gondi pismo', 'cs': 'gundžala gondí', 'en': 'Gunjala Gondi', 'et': 'Gūnjāla gondi', 'eu': 'gunjala gondi', 'ff-Adlm': '𞤘𞤵𞤲𞤶𞤢𞤤𞤢 𞤘𞤮𞤲𞤣𞤭', 'fi': 'gondin gunjala', 'fo': 'gunjala gondi', 'gd': 'Gunjala Gondi', 'he': 'ג’ונג’לה מבולבל', 'hi-Latn': 'Gunjala Gondi', 'id': 'Gunjala Gondi', 'ig': 'Gunjala Gondi', 'kk': 'гунджала гонди жазуы', 'kk-Cyrl': 'гунджала гонди жазуы', 'km': 'គុនចាឡាកុនឌិ', 'mn': 'Гонг', 'ms': 'Gunjala Gondi', 'nl': 'Gunjala Gondi', 'nqo': 'ߜ߭ߏ߲ߖߌ߫ ߜ߭ߎ߲ߖߋߟߊ ߘߐ߫', 'ro': 'gunjala gondi', 'sc': 'gunjala gondi', 'so': 'Gumjala Gondi', 'sv': 'gunjalgondiska', 'tg': 'Гунҷала Гондӣ', 'vec': 'gunjala', 'yo': 'Gunjala Gondi', 'zh': '贡贾拉贡德文', 'zh-Hans': '贡贾拉贡德文', 'zu': 'isi-Gunjala Gondi'},
    'Gonm': {'bal-Latn': 'Masáram Góndi', 'bs': 'masaram gondi pismo', 'bs-Cyrl': 'масарам гонди писмо', 'bs-Latn': 'masaram gondi pismo', 'en': 'Masaram Gondi', 'et': 'Masarami gondi', 'eu': 'masaram gondiera', 'ff-Adlm': '𞤃𞤢𞤧𞤢𞤪𞤢𞤲 𞤘𞤮𞤲𞤣𞤭', 'fi': 'masaram-gondi', 'fo': 'masaram gondi', 'ga': 'Masaram Gondi', 'gd': 'Masaram Gondi', 'he': 'מסארם גונדי', 'hi-Latn': 'Masaram Gondi', 'id': 'Masaram Gondi', 'ig': 'Masaram Gondi', 'kk': 'масарам гонди жазуы', 'kk-Cyrl': 'масарам гонди жазуы', 'mn': 'Масарам Гонди', 'ms': 'Masaram Gonti', 'nl': 'Masaram Gondi', 'nqo': 'ߡߛߊߙߊ߲߫-ߜ߭ߏ߲ߘߌ', 'ro': 'masaram gondi', 'sc': 'gondi de Masaram', 'so': 'Qoraalka Masaram Gondi', 'sq': 'masaramgondisht', 'sv': 'masaram-gondi', 'tg': 'Масарам Гондӣ', 'vec': 'gondi masaram', 'zh': '马萨拉姆冈德文', 'zh-Hans': '马萨拉姆冈德文', 'zh-Hant': '岡德文', 'zu': 'isi-Masaram Gondi'},
    'Goth': {'am': 'ጐቲክ', 'ar': 'القوطية', 'ast': 'góticu', 'az': 'qotik', 'az-Latn': 'qotik', 'bal-Latn': 'Góti', 'bg': 'Готическа', 'bn': 'গোথিক', 'br': 'gotek', 'brx': 'गौथीक', 'bs': 'gotika', 'bs-Cyrl': 'готика', 'bs-Latn': 'gotika', 'ca': 'gòtic', 'ca-ES-valencia': 'gòtic', 'ccp': '𑄉𑄮𑄗𑄨𑄇𑄴', 'cs': 'gotické', 'da': 'gotisk', 'de': 'Gotisch', 'el': 'Γοτθικό', 'el-polyton': 'Γοτθικό', 'en': 'Gothic', 'en-Dsrt': '𐐘𐐱𐑃𐐮𐐿', 'es': 'gótico', 'et': 'gooti', 'eu': 'gotikoa', 'fa': 'گوتی', 'ff-Adlm': '𞤘𞤵𞥅𞤼𞤭𞤲𞤳𞤮', 'fi': 'goottilainen', 'fo': 'gotiskt', 'fr': 'gotique', 'fur': 'gotic', 'fy': 'Gothysk', 'ga': 'Gotach', 'gd': 'Gotais', 'gsw': 'Gotisch', 'gu': 'ગોથિક', 'he': 'גותי', 'hi': 'गोथिक', 'hi-Latn': 'Gothic', 'hr': 'gotičko pismo', 'ht': 'gotique', 'hu': 'Gót', 'id': 'Gothic', 'ig': 'Gothic', 'it': 'gotico', 'ja': 'ゴート文字', 'ka': 'გოთური', 'kgp': 'gótiko', 'kk': 'гот жазуы', 'kk-Cyrl': 'гот жазуы', 'kn': 'ಗೋತಿಕ್', 'ko': '고트 문자', 'ks': 'گوتھِک', 'ks-Arab': 'گوتھِک', 'ku': 'gotîk', 'lb': 'Gotesch', 'lo': 'ໂກຮິກ', 'lt': 'gotų', 'lv': 'gotu', 'mk': 'готско', 'ml': 'ഗോഥിക്', 'mn': 'Готик', 'mr': 'गोथिक', 'ms': 'Gothic', 'nb': 'gotisk', 'ne': 'गोथिक', 'nl': 'Gothisch', 'nn': 'gotisk', 'no': 'gotisk', 'nqo': 'ߜ߭ߏߕߌߞ', 'or': 'ଗୋଥିକ୍', 'pl': 'gotyckie', 'pt': 'gótico', 'rm': 'gotic', 'ro': 'gotică', 'ru': 'готская', 'sc': 'gòticu', 'sk': 'gotický', 'sl': 'gotski', 'so': 'Dadka Gothic', 'sq': 'gotik', 'sr': 'Готика', 'sr-Cyrl': 'Готика', 'sr-Latn': 'Gotika', 'sv': 'gotiska', 'ta': 'கோதிக்', 'te': 'గోతిక్', 'tg': 'Готика', 'th': 'โกธิก', 'to': 'tohinima fakakotika', 'tr': 'Gotik', 'ug': 'گوتچە', 'uk': 'готичний', 'vec': 'gòtego', 'vi': 'Chữ Gô-tích', 'xnr': 'गोथिक', 'yrl': 'gútiku', 'yue': '歌德文', 'yue-Hans': '歌德文', 'yue-Hant': '歌德文', 'zh': '哥特文', 'zh-Hans': '哥特文', 'zh-Hant': '歌德文', 'zu': 'isi-Gothic'},
    'Gran': {'ast': 'grantha', 'bal-Latn': 'Grantá', 'bs': 'grantha pismo', 'bs-Cyrl': 'гранта писмо', 'bs-Latn': 'grantha pismo', 'ca': 'grantha', 'ca-ES-valencia': 'grantha', 'cs': 'grantha', 'da': 'grantha', 'de': 'Grantha', 'en': 'Grantha', 'et': 'grantha', 'eu': 'grantha', 'ff-Adlm': '𞤘𞤢𞤪𞤢𞤲𞤼𞤢', 'fi': 'grantha', 'fo': 'grantha', 'fy': 'Grantha', 'ga': 'Grantha', 'gd': 'Grantha', 'he': 'גרנתה', 'hi': 'ग्रन्थ', 'hi-Latn': 'Grantha', 'hr': 'grantha pismo', 'id': 'Grantha', 'ig': 'Grantha', 'it': 'grantha', 'ja': 'グランタ文字', 'ka': 'გრანთა', 'kk': 'грантха жазуы', 'kk-Cyrl': 'грантха жазуы', 'ko': '그란타 문자', 'lo': 'ເຄນທາ', 'lt': 'Granta', 'mk': 'гранта', 'mn': 'Гранта', 'ms': 'Grantha', 'nb': 'gammeltamilsk', 'nl': 'Grantha', 'nn': 'gammeltamilsk', 'no': 'gammeltamilsk', 'nqo': 'ߜ߭ߑߙߊ߲ߗߊ', 'ro': 'grantha', 'ru': 'грантха', 'sc': 'grantha', 'so': 'Qoraalka Grantha', 'sq': 'grantaisht', 'sv': 'gammaltamilska', 'tg': 'Гранта', 'th': 'คฤณห์', 'to': 'tohinima fakasilanitā', 'tr': 'Grantha', 'ug': 'گىرانتاچە', 'vec': 'granta', 'vi': 'Chữ Grantha', 'xnr': 'ग्रन्थ', 'yue': '格蘭他文字', 'yue-Hans': '格兰他文字', 'yue-Hant': '格蘭他文字', 'zh': '格兰塔文', 'zh-Hans': '格兰塔文', 'zh-Hant': '格蘭他文字', 'zu': 'isi-Grantha'},
    'Grek': {'af': 'Grieks', 'ak': 'Griiki kasa', 'am': 'ግሪክ', 'ar': 'اليونانية', 'as': 'গ্ৰীক', 'ast': 'griegu', 'az': 'yunan', 'az-Arab': 'یونانجا', 'az-Latn': 'yunan', 'bal-Latn': 'Yun', 'be': 'грэчаскае', 'be-tarask': 'грэчаскае', 'bg': 'гръцка', 'bgn': 'یونانی', 'bn': 'গ্রিক', 'br': 'gresianek', 'brx': 'ग्रीक', 'bs': 'grčko pismo', 'bs-Cyrl': 'грчко писмо', 'bs-Latn': 'grčko pismo', 'ca': 'grec', 'ca-ES-valencia': 'grec', 'ccp': '𑄉𑄳𑄢𑄨𑄇𑄴', 'ce': 'грекийн', 'chr': 'ᎪᎢ', 'ckb': 'یۆنانی', 'cs': 'řecké', 'cy': 'Groegaidd', 'da': 'græsk', 'de': 'Griechisch', 'dsb': 'grichiski', 'dz': 'གྲིཀ་ཡིག་གུ', 'ee': 'grisigbeŋɔŋlɔ', 'el': 'Ελληνικό', 'el-polyton': 'Ἑλληνικό', 'en': 'Greek', 'en-Dsrt': '𐐘𐑉𐐨𐐿', 'es': 'griego', 'et': 'kreeka', 'eu': 'grekoa', 'fa': 'یونانی', 'ff-Adlm': '𞤘𞤭𞤪𞤧𞤢', 'fi': 'kreikkalainen', 'fil': 'Greek', 'fo': 'grikskt', 'fr': 'grec', 'fur': 'grêc', 'fy': 'Grieks', 'ga': 'Gréagach', 'gd': 'Greugais', 'gl': 'grego', 'gsw': 'Griechisch', 'gu': 'ગ્રીક', 'ha': 'Girkanci', 'he': 'יווני', 'hi': 'यूनानी', 'hi-Latn': 'Greek', 'hr': 'grčko pismo', 'hsb': 'grjeksce', 'ht': 'grec', 'hu': 'Görög', 'hy': 'հունական', 'ia': 'grec', 'id': 'Yunani', 'ig': 'Mkpụrụ ọkwụ grịk', 'is': 'grískt', 'it': 'greco', 'ja': 'ギリシャ文字', 'jv': 'Yunani', 'ka': 'ბერძნული', 'kea': 'gregu', 'kgp': 'gregu', 'kk': 'грек жазуы', 'kk-Cyrl': 'грек жазуы', 'km': 'ក្រិច', 'kn': 'ಗ್ರೀಕ್', 'ko': '그리스 문자', 'kok': 'ग्रीक', 'kok-Deva': 'ग्रीक', 'ks': 'گرَنتھا', 'ks-Arab': 'گرَنتھا', 'ksh': 'jriischesche Schreff', 'ku': 'yûnanî', 'ky': 'Грек', 'lb': 'Griichesch', 'lo': 'ກຣີກ', 'lrc': 'یوٙنانی', 'lt': 'graikų', 'lv': 'grieķu', 'mai': 'यूनानी', 'mi': 'Kariki', 'mk': 'грчко писмо', 'ml': 'ഗ്രീക്ക്', 'mn': 'грек', 'mr': 'ग्रीक', 'ms': 'Greek', 'mt': 'Grieg', 'my': 'ဂရိ', 'mzn': 'یونانی', 'nb': 'gresk', 'ne': 'ग्रीक', 'nl': 'Grieks', 'nn': 'gresk', 'no': 'gresk', 'nqo': 'ߜ߭ߙߍ߬ߞߌ', 'or': 'ଗ୍ରୀକ୍', 'pa': 'ਯੂਨਾਨੀ', 'pa-Guru': 'ਯੂਨਾਨੀ', 'pcm': 'Grík', 'pl': 'greckie', 'ps': 'یوناني', 'pt': 'grego', 'qu': 'Griego Simi', 'rm': 'grec', 'ro': 'greacă', 'ru': 'греческая', 'sah': 'Гириэктии', 'sc': 'grecu', 'sd': 'يوناني', 'sd-Arab': 'يوناني', 'se': 'greikkalaš', 'si': 'ග්\u200dරීක', 'sk': 'grécke', 'sl': 'grški', 'so': 'Giriik', 'sr': 'грчко писмо', 'sr-Cyrl': 'грчко писмо', 'sr-Latn': 'grčko pismo', 'sv': 'grekiska', 'sw': 'Kigiriki', 'syr': 'ܝܘܢܝܬ', 'ta': 'கிரேக்கம்', 'te': 'గ్రీక్', 'tg': 'Юнонӣ', 'th': 'กรีก', 'ti': 'ግሪክ', 'tk': 'Grek elipbiýi', 'to': 'tohinima fakakalisi', 'tr': 'Yunan', 'ug': 'گىرېكچە', 'uk': 'грецька', 'ur': 'یونانی', 'uz-Cyrl': 'Юнон', 'vec': 'grego', 'vi': 'Chữ Hy Lạp', 'wae': 'Gričiš', 'xnr': 'यूनानी', 'yi': 'גריכיש', 'yo': 'Gíríkì', 'yrl': 'geregu', 'yue': '希臘文', 'yue-Hans': '希腊文', 'yue-Hant': '希臘文', 'zh': '希腊文', 'zh-Hans': '希腊文', 'zh-Hant': '希臘字母', 'zu': 'isi-Greek'},
    'Gujr': {'af': 'Goedjarati', 'ak': 'Gudwurati kasa', 'am': 'ጉጃራቲ', 'ar': 'التاغجراتية', 'as': 'গুজৰাটী', 'ast': 'guyarati', 'az': 'qucarat', 'az-Latn': 'qucarat', 'bal-Latn': 'Gojr', 'be': 'гуджараці', 'be-tarask': 'гуджараці', 'bg': 'гуджарати', 'bgn': 'گوجراتی', 'bn': 'গুজরাটি', 'br': 'gujarati', 'brx': 'गुजराती', 'bs': 'pismo gudžarati', 'bs-Cyrl': 'гуџарати', 'bs-Latn': 'pismo gudžarati', 'ca': 'gujarati', 'ca-ES-valencia': 'gujarati', 'ccp': '𑄉𑄪𑄎𑄴𑄢𑄑𑄨', 'ce': 'гуджарати', 'chr': 'ᎫᏣᎳᏘ', 'ckb': 'گوجەراتی', 'cs': 'gudžarátí', 'cy': 'Gwjarataidd', 'da': 'gujarati', 'de': 'Gujarati', 'dsb': 'gujarati', 'dz': 'གུ་ཇ་ར་ཏི་ཡིག་གུ', 'ee': 'gudzaratigbeŋɔŋlɔ', 'el': 'Γκουγιαράτι', 'el-polyton': 'Γκουγιαράτι', 'en': 'Gujarati', 'en-Dsrt': '𐐘𐐳𐐾𐐲𐑉𐐪𐐼𐐨', 'es': 'guyaratí', 'et': 'gudžarati', 'eu': 'gujaratarra', 'fa': 'گجراتی', 'ff-Adlm': '𞤘𞤵𞤶𞤢𞤪𞤢𞤼𞤭𞥅', 'fi': 'gudžaratilainen', 'fil': 'Gujarati', 'fo': 'gujarati', 'fr': 'goudjarâtî', 'fur': 'gujarati', 'fy': 'Gujarati', 'ga': 'Gúisearátach', 'gd': 'Gujarati', 'gl': 'guxarati', 'gsw': 'Guscharati', 'gu': 'ગુજરાતી', 'ha': 'Gujarati', 'he': 'גוג׳רטי', 'hi': 'गुजराती', 'hi-Latn': 'Gujarati', 'hr': 'gudžaratsko pismo', 'hsb': 'gujarati', 'ht': 'goudjarâtî', 'hu': 'Gudzsaráti', 'hy': 'գուջարաթի', 'ia': 'gujarati', 'id': 'Gujarat', 'ig': 'Mkpụrụ ọkwụ Gụjaratị', 'is': 'gújaratí', 'it': 'gujarati', 'ja': 'グジャラート文字', 'jv': 'Gujarati', 'ka': 'გუჯარათული', 'kea': 'gujarati', 'kgp': 'guserate', 'kk': 'гуджарати жазуы', 'kk-Cyrl': 'гуджарати жазуы', 'km': 'គូចារ៉ាទី', 'kn': 'ಗುಜರಾತಿ', 'ko': '구자라트 문자', 'kok': 'गुजराती', 'kok-Deva': 'गुजराती', 'ks': 'گریٖک', 'ks-Arab': 'گریٖک', 'ksh': 'jujaraatesche Schreff', 'ku': 'gujeratî', 'kxv': 'gujraṭī', 'kxv-Deva': 'गुजराती', 'kxv-Latn': 'gujraṭī', 'kxv-Orya': 'ଗୁଜୁରାଟୀ', 'kxv-Telu': 'గుజరాతి', 'ky': 'Гужарати', 'lb': 'Gujarati', 'lo': 'ຈູຈາຣາທີ', 'lrc': 'گوجأراتی', 'lt': 'gudžaratų', 'lv': 'gudžaratu', 'mai': 'गुजराती', 'mi': 'Kutarāti', 'mk': 'гуџарати', 'ml': 'ഗുജറാത്തി', 'mn': 'гужарати', 'mr': 'गुजराती', 'ms': 'Gujarat', 'my': 'ဂုဂျာရသီ', 'mzn': 'گجراتی', 'nb': 'gujarati', 'ne': 'गुजराती', 'nl': 'Gujarati', 'nn': 'gujarati', 'no': 'gujarati', 'nqo': 'ߜ߭ߎߖߙߊߕߌ', 'or': 'ଗୁଜରାଟୀ', 'pa': 'ਗੁਜਰਾਤੀ', 'pa-Guru': 'ਗੁਜਰਾਤੀ', 'pcm': 'Gujaráti', 'pl': 'gudżarati', 'ps': 'ګجراتي', 'pt': 'guzerate', 'qu': 'Gujarati Simi', 'rm': 'gujarati', 'ro': 'gujarati', 'ru': 'гуджарати', 'sc': 'gujarati', 'sd': 'گجراتي', 'sd-Arab': 'گجراتي', 'si': 'ගුජරාටි', 'sk': 'gudžarátí', 'sl': 'gudžaratski', 'so': 'Gujaraati', 'sq': 'guxharat', 'sr': 'гуџаратско писмо', 'sr-Cyrl': 'гуџаратско писмо', 'sr-Latn': 'gudžaratsko pismo', 'sv': 'gujarati', 'sw': 'Kigujarati', 'syr': 'ܓܘܓܐܪܝܬ', 'ta': 'குஜராத்தி', 'te': 'గుజరాతీ', 'tg': 'Гуҷаротӣ', 'th': 'คุชราต', 'ti': 'ጉጃርቲ', 'tk': 'Gujarati elipbiýi', 'to': 'tohinima fakaʻinitia-kutalati', 'tr': 'Gücerat', 'ug': 'گۇجاراتچە', 'uk': 'гуджараті', 'ur': 'گجراتی', 'uz': 'gujarot', 'uz-Cyrl': 'Гужарати', 'uz-Latn': 'gujarot', 'vec': 'gujarati', 'vi': 'Chữ Gujarati', 'wae': 'Gujarati', 'xnr': 'गुजराती', 'yo': 'Gujaráti', 'yrl': 'guserati', 'yue': '古吉拉特文', 'yue-Hans': '古吉拉特文', 'yue-Hant': '古吉拉特文', 'zh': '古吉拉特文', 'zh-Hans': '古吉拉特文', 'zh-Hant': '古吉拉特文', 'zu': 'isi-Gujarati'},
    'Gukh': {'bal-Latn': 'Gurong Khémá', 'en': 'Gurung Khema', 'fo': 'gurung khema', 'gd': 'Gurung Khema', 'hi-Latn': 'Gurung Khema', 'ig': 'Gurung Khema', 'vec': 'gurung khema'},
    'Guru': {'af': 'Gurmukhi', 'ak': 'Gurumuki kasa', 'am': 'ጉርሙኪ', 'ar': 'الجرمخي', 'as': 'গুৰুমুখী', 'ast': 'gurmukhi', 'az': 'qurmuxi', 'az-Latn': 'qurmuxi', 'bal-Latn': 'Gurukuki', 'be': 'гурмукхі', 'be-tarask': 'гурмукхі', 'bg': 'гурмукхи', 'bgn': 'گرمکهی', 'bn': 'গুরুমুখি', 'br': 'gurmukhi', 'brx': 'गुरमुखी', 'bs': 'pismo gurmuki', 'bs-Cyrl': 'гурмуки писмо', 'bs-Latn': 'pismo gurmuki', 'ca': 'gurmukhi', 'ca-ES-valencia': 'gurmukhi', 'ccp': '𑄉𑄪𑄢𑄪𑄟𑄪𑄈𑄨', 'ce': 'гурмукхи', 'chr': 'ᎬᎹᎩ', 'ckb': 'گورموکھی', 'cs': 'gurmukhi', 'cy': 'Gwrmwci', 'da': 'gurmukhi', 'de': 'Gurmukhi', 'dsb': 'gurmukhi', 'dz': 'གུ་རུ་མུ་ཁ་ཡིག་གུ', 'ee': 'gurmukhigbeŋɔŋlɔ', 'el': 'Γκουρμουκχί', 'el-polyton': 'Γκουρμουκχί', 'en': 'Gurmukhi', 'en-Dsrt': '𐐘𐐳𐑉𐑋𐐲𐐿𐐨', 'es': 'gurmuji', 'et': 'gurmukhi', 'eu': 'gurmukhia', 'fa': 'گورومخی', 'ff-Adlm': '𞤘𞤵𞤪𞤥𞤵𞤿𞤭', 'fi': 'gurmukhi', 'fil': 'Gurmukhi', 'fo': 'gurmukhi', 'fr': 'gourmoukhî', 'fy': 'Gurmukhi', 'ga': 'Gurmúcach', 'gd': 'Gurmukhi', 'gl': 'gurmukhi', 'gsw': 'Gurmukhi', 'gu': 'ગુરૂમુખી', 'ha': 'Gurmukhi', 'he': 'גורמוקי', 'hi': 'गुरमुखी', 'hi-Latn': 'Gurmukhi', 'hr': 'gurmukhi pismo', 'hsb': 'gurmukhi', 'ht': 'gourmoukhî', 'hu': 'Gurmuki', 'hy': 'գուրմուխի', 'ia': 'gurmukhi', 'id': 'Gurmukhi', 'ig': 'Mkpụrụ ọkwụ Gụrmụkị', 'is': 'gurmukhi', 'it': 'gurmukhi', 'ja': 'グルムキー文字', 'jv': 'Gurmukhi', 'ka': 'გურმუხი', 'kea': 'gurmuki', 'kgp': 'gurmũki', 'kk': 'гурмукхи жазуы', 'kk-Cyrl': 'гурмукхи жазуы', 'km': 'កុមុយឃី', 'kn': 'ಗುರ್ಮುಖಿ', 'ko': '구르무키 문자', 'kok': 'गुरूमुखी', 'kok-Deva': 'गुरूमुखी', 'ks': 'گُجرٲتؠ', 'ks-Arab': 'گُجرٲتؠ', 'ksh': 'indesche Gurmukhi-Schreff', 'kxv': 'gurmukī', 'kxv-Deva': 'गुरमुकी', 'kxv-Latn': 'gurmukī', 'kxv-Orya': 'ଗୁରମୁକୀ', 'kxv-Telu': 'గురుముకి', 'ky': 'Гурмухи', 'lb': 'Gurmukhi', 'lo': 'ກົວມູຄີ', 'lrc': 'گوٙروٙمخی', 'lt': 'gurmuki', 'lv': 'pandžabu', 'mai': 'गुरमुखी', 'mi': 'Kūmuki', 'mk': 'гурмуки', 'ml': 'ഗുരുമുഖി', 'mn': 'гүрмүх', 'mr': 'गुरुमुखी', 'ms': 'Gurmukhi', 'my': 'ဂူရူ', 'mzn': 'گورموخی', 'nb': 'gurmukhi', 'ne': 'गुरूमुखी', 'nl': 'Gurmukhi', 'nn': 'gurmukhi', 'no': 'gurmukhi', 'nqo': 'ߜ߭ߎߙߑߡߎߞߌ', 'or': 'ଗୁରମୁଖୀ', 'pa': 'ਗੁਰਮੁਖੀ', 'pa-Arab': 'گُرمُکھی', 'pa-Guru': 'ਗੁਰਮੁਖੀ', 'pcm': 'Gúrmukhi', 'pl': 'gurmukhi', 'ps': 'ګرومي', 'pt': 'gurmuqui', 'qu': 'Gurmukhi Simi', 'rm': 'gurmukhi', 'ro': 'gurmukhi', 'ru': 'гурмукхи', 'sc': 'gurmukhi', 'sd': 'گرمکي', 'sd-Arab': 'گرمکي', 'si': 'ගුර්මුඛි', 'sk': 'gurmukhi', 'sl': 'gurmuki', 'so': 'Luuqada gujarati', 'sq': 'gurmuk', 'sr': 'гурмуки писмо', 'sr-Cyrl': 'гурмуки писмо', 'sr-Latn': 'gurmuki pismo', 'sv': 'gurmukhiska', 'sw': 'Kigurmukhi', 'ta': 'குர்முகி', 'te': 'గుర్ముఖి', 'tg': 'Гумрухӣ', 'th': 'กูร์มูคี', 'ti': 'ጉርሙኪ', 'tk': 'Gurmuhi elipbiýi', 'to': 'tohinima fakakūmuki', 'tr': 'Gurmukhi', 'ug': 'گۇرمۇكىچە', 'uk': 'гурмухі', 'ur': 'گرمکھی', 'uz': 'gurmukxi', 'uz-Cyrl': 'Гурмухи', 'uz-Latn': 'gurmukxi', 'vec': 'gurmuki', 'vi': 'Chữ Gurmukhi', 'xnr': 'गुरमुखी', 'yo': 'Gurumúkhì', 'yrl': 'gumuki', 'yue': '古魯穆奇文', 'yue-Hans': '古鲁穆奇文', 'yue-Hant': '古魯穆奇文', 'zh': '果鲁穆奇文', 'zh-Hans': '果鲁穆奇文', 'zh-Hant': '古魯穆奇文', 'zu': 'isi-Gurmukhi'},
    'HK': {'af': 'Hongkong SAS China', 'ak': 'Hɔnkɔn Kyaena', 'am': 'ሆንግ ኮንግ ልዩ የአስተዳደር ክልል ቻይና', 'ar': 'هونغ كونغ الصينية (منطقة إدارية خاصة)', 'as': 'হং কং এছ. এ. আৰ. চীন', 'ast': 'ARE China de Ḥong Kong', 'az': 'Honq Konq Xüsusi İnzibati Rayonu Çin', 'az-Cyrl': 'Һонк Конг Хүсуси Инзибати Әрази Чин', 'az-Latn': 'Honq Konq Xüsusi İnzibati Rayonu Çin', 'bal-Latn': 'Háng Káng o SAR Chin', 'be': 'Ганконг, САР (Кітай)', 'be-tarask': 'Ганконг, САР (Кітай)', 'bg': 'Хонконг, САР на Китай', 'bgn': 'هانگ کانگ', 'blo': 'Hɔŋ Kɔŋ Caɩna kaja', 'bn': 'হংকং এসএআর চীনা', 'br': 'Hong Kong RMD Sina', 'brx': 'हंकं एस.ए.आर चाइना', 'bs': 'Hong Kong (SAR Kina)', 'bs-Cyrl': 'Хонг Конг С. А. Р.', 'bs-Latn': 'Hong Kong (SAR Kina)', 'ca': 'Hong Kong (RAE Xina)', 'ca-ES-valencia': 'Hong Kong (RAE Xina)', 'ccp': '𑄦𑄧𑄁𑄇𑄧𑄁 𑄆𑄌𑄴𑄃𑄬𑄃𑄢𑄴 𑄌𑄩𑄚', 'ce': 'Гонконг (ша-къаьстина кӀошт)', 'ceb': 'Hong Kong SAR China', 'chr': 'ᎰᏂᎩ ᎪᏂᎩ ᎤᏓᏤᎵᏓ ᏧᏂᎸᏫᏍᏓᏁᏗ ᎢᎬᎾᏕᎾ ᏓᎶᏂᎨᏍᏛ', 'ckb': 'هۆنگ کۆنگ', 'cs': 'Hongkong – ZAO Číny', 'cv': 'Гонконг (САР)', 'cy': 'Hong Kong SAR Tsieina', 'da': 'SAR Hongkong', 'de': 'Sonderverwaltungsregion Hongkong', 'dsb': 'Wósebna zastojnstwowa cona Hongkong', 'dz': 'ཧོང་ཀོང་ཅཱའི་ན', 'ee': 'Hɔng Kɔng SAR Tsaina nutome', 'el': 'Χονγκ Κονγκ ΕΔΠ Κίνας', 'el-polyton': 'Χὸνγκ Κόνγκ, Εἰδικὴ Διοικητικὴ Περιφέρεια τῆς Κίνας', 'en': 'Hong Kong SAR China', 'en-Dsrt': '𐐐𐐬𐑍 𐐗𐐬𐑍 𐐝𐐈𐐡 𐐕𐐴𐑌𐐲', 'es': 'RAE de Hong Kong (China)', 'et': 'Hongkongi erihalduspiirkond', 'eu': 'Hong Kong Txinako AEB', 'fa': 'هنگ\u200cکنگ، منطقهٔ ویژهٔ اداری چین', 'ff-Adlm': '𞤖𞤂𞤀 𞤕𞤢𞤴𞤲𞤢 𞤫 𞤖𞤮𞤲𞤺 𞤑𞤮𞤲𞤺', 'fi': 'Hongkong – Kiinan erityishallintoalue', 'fil': 'Hong Kong SAR China', 'fo': 'Hong Kong SAR Kina', 'fr': 'R.A.S. chinoise de Hong Kong', 'fur': 'Regjon aministrative speciâl de Cine di Hong Kong', 'fy': 'Hongkong SAR van Sina', 'ga': 'Sainréigiún Riaracháin Hong Cong, Daonphoblacht na Síne', 'gd': 'Hong Kong SAR na Sìne', 'gl': 'Hong Kong RAE da China', 'gsw': 'Sonderverwaltigszone Hongkong', 'gu': 'હોંગકોંગ SAR ચીન', 'ha': 'Babban Yankin Mulkin Hong Kong na Ƙasar Sin', 'he': 'הונג קונג (אזור מנהלי מיוחד של סין)', 'hi': 'हाँग काँग (चीन विशेष प्रशासनिक क्षेत्र)', 'hi-Latn': 'Hong Kong SAR China', 'hr': 'PUP Hong Kong Kina', 'hsb': 'Wosebita zarjadniska cona Hongkong', 'ht': 'R.A.S. chinoise de Hong Kong', 'hu': 'Hongkong KKT', 'hy': 'Հոնկոնգի ՀՎՇ', 'ia': 'Hongkong, R.A.S. de China', 'id': 'Hong Kong DAK Tiongkok', 'ig': 'Hong Kong SAR China', 'is': 'sérstjórnarsvæðið Hong Kong', 'it': 'RAS di Hong Kong', 'ja': '中華人民共和国香港特別行政区', 'jv': 'Laladan Administratif Astamiwa Hong Kong', 'ka': 'ჰონკონგის სპეციალური ადმინისტრაციული რეგიონი, ჩინეთი', 'kaa': 'Гонконг ААР', 'kaa-Cyrl': 'Гонконг ААР', 'kea': 'Hong Kong, Rejion Administrativu Spesial di Xina', 'kgp': 'Hãg Kãg, Sinỹ ERA', 'kk': 'Сянган АӘА', 'kk-Cyrl': 'Сянган АӘА', 'km': 'ហុងកុង តំបន់រដ្ឋបាលពិសេសចិន', 'kn': 'ಹಾಂಗ್ ಕಾಂಗ್ ಎಸ್ಎಆರ್ ಚೈನಾ', 'ko': '홍콩(중국 특별행정구)', 'kok': 'हाँग काँग एसएआर चीन', 'kok-Deva': 'हाँग काँग एसएआर चीन', 'ks': 'ہانگ کانگ ایس اے آر چیٖن', 'ks-Arab': 'ہانگ کانگ ایس اے آر چیٖن', 'ksh': 'Hongkong', 'ku': 'Hong Konga HîT ya Çînê', 'kxv': 'hong kong (sar) caina', 'kxv-Deva': 'हाँग काँग (एस ए आर चीन)', 'kxv-Latn': 'hong kong (sar) caina', 'kxv-Orya': 'ହଂଗ କଂଗ (ଏସଏଆର୍ ଚିନା)', 'kxv-Telu': 'హాంకాంగ్ ఎస్ఎఆర్ చినా', 'ky': 'Гонконг Кытай ААА', 'lb': 'Spezialverwaltungszon Hong Kong', 'lld': 'Regiun aministrativa speziala de Hong Kong', 'lo': 'ຮົງກົງ ເຂດປົກຄອງພິເສດ ຈີນ', 'lt': 'Ypatingasis Administracinis Kinijos Regionas Honkongas', 'lv': 'Ķīnas īpašās pārvaldes apgabals Honkonga', 'mai': 'हांग कांग (विशेष प्रशासनिक क्षेत्र चीन)', 'mi': 'Hongipua Haina', 'mk': 'Хонгконг САР Кина', 'ml': 'ഹോങ്കോങ് (SAR) ചൈന', 'mn': 'БНХАУ-ын Тусгай захиргааны бүс Хонг-Конг', 'mr': 'हाँगकाँग एसएआर चीन', 'ms': 'Hong Kong SAR China', 'mt': 'ir-Reġjun Amministrattiv Speċjali ta’ Hong Kong tar-Repubblika tal-Poplu taċ-Ċina', 'my': 'ဟောင်ကောင် (တရုတ်ပြည်)', 'mzn': 'هنگ کنگ', 'nb': 'Hongkong SAR Kina', 'ne': 'हङकङ चिनियाँ विशेष प्रशासनिक क्षेत्र', 'nl': 'Hongkong SAR van China', 'nn': 'Hongkong SAR Kina', 'no': 'Hongkong SAR Kina', 'nqo': 'ߛߌߣ ߕߌ߲߬ߞߎߘߎ߲߫ ߡߊߡߙߊ߬ߣߍ߲ ߤߐ߲ߞߐ߲߫', 'oc': 'Hong Kong', 'om': 'Hoong Koong SAR Chaayinaa', 'or': 'ହଂ କଂ ଏସଏଆର୍\u200c ଚାଇନା', 'pa': 'ਹਾਂਗ ਕਾਂਗ ਐਸਏਆਰ ਚੀਨ', 'pa-Guru': 'ਹਾਂਗ ਕਾਂਗ ਐਸਏਆਰ ਚੀਨ', 'pcm': 'Họng Kọng SAR', 'pl': 'SRA Hongkong (Chiny)', 'ps': 'هانګ کانګ SAR چین', 'pt': 'Hong Kong, RAE da China', 'qu': 'Hong Kong RAE China', 'rm': 'Regiun d’administraziun speziala da Hongkong, China', 'ro': 'R.A.S. Hong Kong, China', 'ru': 'Гонконг (САР)', 'sc': 'RAS tzinesa de Hong Kong', 'sd': 'هانگ ڪانگ SAR', 'sd-Arab': 'هانگ ڪانگ SAR', 'se': 'Hongkong', 'si': 'හොංකොං විශේෂ පරිපාලන කලාපය චීනය', 'sk': 'Hongkong – OAO Číny', 'sl': 'Posebno upravno območje Ljudske republike Kitajske Hongkong', 'smn': 'Hongkong – Kiina e.h.k.', 'so': 'Hong Kong', 'sq': 'RPA i Hong-Kongut', 'sr': 'САР Хонгконг (Кина)', 'sr-Cyrl': 'САР Хонгконг (Кина)', 'sr-Latn': 'SAR Hongkong (Kina)', 'sv': 'Hongkong SAR', 'sw': 'Hong Kong SAR China', 'syr': 'ܗܘܢܓ ܟܘܢܓ', 'ta': 'ஹாங்காங் எஸ்ஏஆர் சீனா', 'te': 'హాంకాంగ్ ఎస్ఏఆర్ చైనా', 'tg': 'Ҳонконг (МММ)', 'th': 'เขตปกครองพิเศษฮ่องกงแห่งสาธารณรัฐประชาชนจีน', 'ti': 'ፍሉይ ምምሕዳራዊ ዞባ ሆንግ ኮንግ (ቻይና)', 'tk': 'Gonkong AAS Hytaý', 'to': 'Hongi Kongi SAR Siaina', 'tr': 'Çin Hong Kong ÖİB', 'tt': 'Гонконг Махсус Идарәле Төбәге', 'ug': 'شياڭگاڭ ئالاھىدە مەمۇرىي رايونى (جۇڭگو)', 'uk': 'Гонконг, ОАР Китаю', 'ur': 'ہانگ کانگ SAR چین', 'uz': 'Gonkong (Xitoy MMH)', 'uz-Cyrl': 'Гонконг (Хитой ММҲ)', 'uz-Latn': 'Gonkong (Xitoy MMH)', 'vec': 'RAS de Hong Kong', 'vi': 'Đặc khu Hành chính Hồng Kông, Trung Quốc', 'wae': 'Sonderverwaltigszona Hongkong', 'wo': 'Ooŋ Koŋ', 'xh': 'EHong Kong SAR China', 'xnr': 'हाँग काँग (चीन खास प्रशासनिक लाका)', 'yo': 'Agbègbè Ìṣàkóso Ìṣúná Hong Kong Tí Ṣánà Ń Darí', 'yrl': 'Hũgi Kũgi, RAE Xina yara', 'yue': '中國香港特別行政區', 'yue-Hans': '中华人民共和国香港特别行政区', 'yue-Hant': '中國香港特別行政區', 'zh': '中国香港特别行政区', 'zh-Hans': '中国香港特别行政区', 'zh-Hant': '中國香港特別行政區', 'zu': 'i-Hong Kong SAR China'},
    'HM': {'af': 'Heardeiland en McDonaldeilande', 'ak': 'Heard ne McDonald Supɔ', 'am': 'ኽርድ ኣና ማክዶናልድ ደሴቶች', 'ar': 'جزيرة هيرد وجزر ماكدونالد', 'as': 'হাৰ্ড দ্বীপ আৰু মেকডোনাল্ড দ্বীপপুঞ্জ', 'ast': 'Islles Heard y McDonald', 'az': 'Herd və Makdonald adaları', 'az-Cyrl': 'Һерд вә Макдоналд адалары', 'az-Latn': 'Herd və Makdonald adaları', 'bal-Latn': 'Hard o Mekdónald Islánd', 'be': 'Астравы Херд і Макдональд', 'be-tarask': 'Астравы Херд і Макдональд', 'bg': 'острови Хърд и Макдоналд', 'blo': 'Hɛɛrɖ na Mɛkɖɔnalɖ kaBʊtukǝltǝna', 'bn': 'হার্ড এবং ম্যাকডোনাল্ড দ্বীপপুঞ্জ', 'br': 'Inizi Heard ha McDonald', 'brx': 'होर्द आरो मेकद’नाल्ड द्वीपफोर', 'bs': 'Ostrvo Heard i arhipelag McDonald', 'bs-Cyrl': 'Херд и Мекдоналд Острва', 'bs-Latn': 'Ostrvo Heard i arhipelag McDonald', 'ca': 'Illes Heard i McDonald', 'ca-ES-valencia': 'Illes Heard i McDonald', 'ccp': '𑄦𑄢𑄴𑄓𑄴 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠 𑄃𑄳𑄃 𑄟𑄳𑄠𑄇𑄴𑄓𑄮𑄚𑄴𑄓𑄴 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Херд гӀайре а, Макдональд гӀайренаш а', 'ceb': 'Heard & McDonald Islands', 'chr': 'ᎲᏗ ᎤᎦᏚᏛᎢ ᎠᎴ ᎺᎩᏓᎾᎵᏗ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی هێرد و مەکدانڵد', 'cs': 'Heardův ostrov a McDonaldovy ostrovy', 'cv': 'Херд тата Макдональд утравӗ', 'cy': 'Ynys Heard ac Ynysoedd McDonald', 'da': 'Heard Island og McDonald Islands', 'de': 'Heard und McDonaldinseln', 'dsb': 'Heardowa kupa a McDonaldowe kupy', 'dz': 'ཧཱརཌ་མཚོ་གླིང་ དང་ མེཀ་ཌོ་ནལཌ྄་གླིང་ཚོམ', 'ee': 'Heard kple Mcdonald ƒudomekpowo nutome', 'el': 'Νήσοι Χερντ και Μακντόναλντ', 'el-polyton': 'Νῆσοι Χὲρντ καὶ Μακντόναλντ', 'en': 'Heard & McDonald Islands', 'en-Dsrt': '𐐐𐐲𐑉𐐼 𐐌𐑊𐐲𐑌𐐼 𐐰𐑌𐐼 𐐣𐐿𐐔𐐱𐑌𐐲𐑊𐐼 𐐌𐑊𐐲𐑌𐐼𐑆', 'eo': 'Herda kaj Makdonaldaj Insuloj', 'es': 'Islas Heard y McDonald', 'et': 'Heardi ja McDonaldi saared', 'eu': 'Heard eta McDonald uharteak', 'fa': 'هرد و جزایر مک\u200cدونالد', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤖𞤭𞤪𞤣𞤭 & 𞤃𞤢𞤳𞤣𞤮𞤲𞤢𞤤', 'fi': 'Heard ja McDonaldinsaaret', 'fil': 'Heard & McDonald Islands', 'fo': 'Heard og McDonaldoyggjar', 'fr': 'Îles Heard-et-MacDonald', 'fur': 'Isule Heard e Isulis McDonald', 'fy': 'Heard- en McDonaldeilannen', 'ga': 'Oileán Heard agus Oileáin McDonald', 'gd': 'Eilean Heard is Eileanan MhicDhòmhnaill', 'gl': 'Illa Heard e Illas McDonald', 'gsw': 'Höörd- und MäcDonald-Insle', 'gu': 'હર્ડ અને મેકડોનાલ્ડ આઇલેન્ડ્સ', 'ha': 'Tsibirin Heard da McDonald', 'he': 'איי הרד ומקדונלד', 'hi': 'हर्ड द्वीप और मैकडोनॉल्ड द्वीपसमूह', 'hi-Latn': 'Heard & McDonald Islands', 'hr': 'Otoci Heard i McDonald', 'hsb': 'Heardowa kupa a McDonaldowe kupy', 'ht': 'Îles Heard-et-MacDonald', 'hu': 'Heard-sziget és McDonald-szigetek', 'hy': 'Հերդ կղզի և ՄակԴոնալդի կղզիներ', 'ia': 'Insulas Heard e McDonald', 'id': 'Pulau Heard dan Kepulauan McDonald', 'ig': 'Agwaetiti Heard na Agwaetiti McDonald', 'is': 'Heard og McDonaldseyjar', 'it': 'Isole Heard e McDonald', 'ja': 'ハード島・マクドナルド諸島', 'jv': 'Kapuloan Heard lan McDonald', 'ka': 'ჰერდი და მაკდონალდის კუნძულები', 'kea': 'Ilhas Heard i McDonald', 'kgp': 'Hárni kar Magtonarni Goj-vẽso Ag', 'kk': 'Херд аралы және Макдональд аралдары', 'kk-Cyrl': 'Херд аралы және Макдональд аралдары', 'km': 'កោះ\u200bហឺដនិង\u200bម៉ាក់ដូណាល់', 'kn': 'ಹರ್ಡ್ ಮತ್ತು ಮ್ಯಾಕ್\u200cಡೋನಾಲ್ಡ್ ದ್ವೀಪಗಳು', 'ko': '허드 맥도널드 제도', 'kok': 'हर्ड ऍंड मॅक्डोनाल्ड आयलँड्स', 'kok-Deva': 'हर्ड ऍंड मॅक्डोनाल्ड आयलँड्स', 'ks': 'ہَرٕڑ تہٕ مؠکڈونالڑٕ جٔزیٖرٕ', 'ks-Arab': 'ہَرٕڑ تہٕ مؠکڈونالڑٕ جٔزیٖرٕ', 'ksh': 'de Heart Ensel un de McDonald-Enselle', 'ku': 'Giravên Heard û MacDonaldê', 'ky': 'Херд жана Макдональд аралдары', 'lb': 'Heard- a McDonald-Inselen', 'lld': 'Isoles Heard y McDonald', 'ln': 'Ile Heard et Iles McDonald', 'lo': 'ໝູ່ເກາະເຮີດ & ແມັກໂດນອລ', 'lt': 'Herdo ir Makdonaldo Salos', 'lv': 'Hērda sala un Makdonalda salas', 'mai': 'हर्ड आ मैकडोनल्ड द्वीपसमूह', 'mi': 'Ngā Moutere Heriti me Makitānara', 'mk': 'Остров Херд и Острови Мекдоналд', 'ml': 'ഹേർഡ്, മക്ഡോണൾഡ് ദ്വീപുകൾ', 'mn': 'Херд ба Макдональдийн арлууд', 'mr': 'हर्ड आणि मॅक्डोनाल्ड बेटे', 'ms': 'Kepulauan Heard & McDonald', 'mt': 'il-Gżejjer Heard u l-Gżejjer McDonald', 'my': 'ဟတ်ကျွန်းနှင့်မက်ဒေါနယ်ကျွန်းစု', 'mzn': 'هارد و مک\u200cدونالد جزایر', 'nb': 'Heard- og McDonaldøyene', 'ne': 'हर्ड टापु र म्याकडोनाल्ड टापुहरु', 'nl': 'Heard en McDonaldeilanden', 'nn': 'Heardøya og McDonaldøyane', 'no': 'Heard- og McDonaldøyene', 'nqo': 'ߡߊߞߑߘߏߣߊߟߑߘ ߕߌ߲', 'om': 'Odoloota Herdii fi MaakDoonaald', 'or': 'ହାର୍ଡ୍\u200c ଏବଂ ମ୍ୟାକଡୋନାଲ୍ଡ ଦ୍ୱୀପପୁଞ୍ଜ', 'pa': 'ਹਰਡ ਤੇ ਮੈਕਡੋਨਾਲਡ ਟਾਪੂ', 'pa-Guru': 'ਹਰਡ ਤੇ ਮੈਕਡੋਨਾਲਡ ਟਾਪੂ', 'pcm': 'Hiád & MakDónáld Aílands', 'pl': 'Wyspy Heard i McDonalda', 'ps': 'هارډ او ميکډانلډ ټاپوګان', 'pt': 'Ilhas Heard e McDonald', 'qu': 'Islas Heard y McDonald', 'rm': 'Inslas da Heard e da McDonald', 'ro': 'Insula Heard și Insulele McDonald', 'ru': 'о-ва Херд и Макдональд', 'sc': 'Ìsulas Heard e McDonald', 'sd': 'هرڊ ۽ مڪڊونلڊ ٻيٽ', 'sd-Arab': 'هرڊ ۽ مڪڊونلڊ ٻيٽ', 'se': 'Heard- ja McDonald-sullot', 'si': 'හර්ඩ් දූපත සහ මැක්ඩොනල්ඩ් දූපත්', 'sk': 'Heardov ostrov a Macdonaldove ostrovy', 'sl': 'Heardov otok in McDonaldovi otoki', 'smn': 'Heard já McDonaldsuolluuh', 'so': 'Jasiiradda Haad & MakDonald', 'sq': 'Ishujt Hërd e Mekdonald', 'sr': 'Острво Херд и Мекдоналдова острва', 'sr-Cyrl': 'Острво Херд и Мекдоналдова острва', 'sr-Latn': 'Ostrvo Herd i Mekdonaldova ostrva', 'sv': 'Heardön och McDonaldöarna', 'sw': 'Kisiwa cha Heard na Visiwa vya McDonald', 'ta': 'ஹேர்ட் மற்றும் மெக்டொனால்டு தீவுகள்', 'te': 'హెర్డ్ దీవి మరియు మెక్\u200cడొనాల్డ్ దీవులు', 'tg': 'Ҷазираи Ҳерд ва Ҷазираҳои Макдоналд', 'th': 'เกาะเฮิร์ดและหมู่เกาะแมกดอนัลด์', 'ti': 'ደሴታት ሄርድን ማክዶናልድን', 'tk': 'Herd we Makdonald adalary', 'to': 'ʻOtumotu Heati mo Makitonali', 'tr': 'Heard Adası ve McDonald Adaları', 'tt': 'Херд утравы һәм Макдональд утраулары', 'ug': 'ھېرد ئارىلى ۋە ماكدونالد ئاراللىرى', 'uk': 'Острови Герд і Макдоналд', 'ur': 'ہیرڈ جزیرہ و میکڈولینڈ جزائر', 'uz': 'Xerd va Makdonald orollari', 'uz-Cyrl': 'Херд ва Макдоналд ороллари', 'uz-Latn': 'Xerd va Makdonald orollari', 'vai': 'ꗥꗡꔵ ꗪ ꕮꖁꕯ', 'vai-Vaii': 'ꗥꗡꔵ ꗪ ꕮꖁꕯ', 'vec': 'Ìzole Heard e McDonald', 'vi': 'Quần đảo Heard và McDonald', 'wae': 'Heard- und McDonald-Inslä', 'wo': 'Duni Hërd ak Duni MakDonald', 'xh': 'EHeard & McDonald Islands', 'xnr': 'हर्ड द्वीप कने मैकडोनॉल्ड द्वीपां दा झुंड', 'yo': 'Erékùsù Heard àti Erékùsù McDonald', 'yrl': 'Kapuãma-ita Heard asuí McDonald', 'yue': '赫德島同麥克唐納群島', 'yue-Hans': '赫德岛同麦克唐纳群岛', 'yue-Hant': '赫德島同麥克唐納群島', 'zh': '赫德岛和麦克唐纳群岛', 'zh-Hans': '赫德岛和麦克唐纳群岛', 'zh-Hant': '赫德島及麥唐納群島', 'zu': 'I-Heard & McDonald Island'},
    'HN': {'af': 'Honduras', 'agq': 'Hɔndulàs', 'ak': 'Hɔnduras', 'am': 'ሆንዱራስ', 'ar': 'هندوراس', 'as': 'হন্দুৰাছ', 'asa': 'Hondurathi', 'ast': 'Hondures', 'az': 'Honduras', 'az-Cyrl': 'Һондурас', 'az-Latn': 'Honduras', 'bal-Latn': 'Honduras', 'bas': 'Ɔ̀ŋduràs', 'be': 'Гандурас', 'be-tarask': 'Гандурас', 'bez': 'Huhondulasi', 'bg': 'Хондурас', 'blo': 'Hɔnɖuraasɩ', 'bm': 'Hɔndirasi', 'bn': 'হন্ডুরাস', 'br': 'Honduras', 'brx': 'हण्डूरास', 'bs': 'Honduras', 'bs-Cyrl': 'Хондурас', 'bs-Latn': 'Honduras', 'ca': 'Hondures', 'ca-ES-valencia': 'Hondures', 'ccp': '𑄦𑄪𑄚𑄴𑄓𑄪𑄢𑄥𑄴', 'ce': 'Гондурас', 'ceb': 'Honduras', 'cgg': 'Hondurasi', 'chr': 'ᎭᏂᏚᎳᏍ', 'ckb': 'ھۆندووراس', 'cs': 'Honduras', 'cv': 'Гондурас', 'cy': 'Honduras', 'da': 'Honduras', 'dav': 'Hondurasi', 'de': 'Honduras', 'dje': 'Honduras', 'dsb': 'Honduras', 'dyo': 'Onduras', 'dz': 'ཧཱན་ཌུ་རཱས྄', 'ebu': 'Hondurasi', 'ee': 'Hondurasdu', 'el': 'Ονδούρα', 'el-polyton': 'Ὁνδούρα', 'en': 'Honduras', 'en-Dsrt': '𐐐𐐪𐑌𐐼𐐭𐑉𐐲𐑅', 'eo': 'Honduro', 'es': 'Honduras', 'et': 'Honduras', 'eu': 'Honduras', 'ewo': 'Ondurás', 'fa': 'هندوراس', 'ff': 'Onnduraas', 'ff-Adlm': '𞤖𞤮𞤲𞤣𞤭𞤪𞤢𞥄𞤧', 'ff-Latn': 'Onnduraas', 'fi': 'Honduras', 'fil': 'Honduras', 'fo': 'Honduras', 'fr': 'Honduras', 'fur': 'Honduras', 'fy': 'Honduras', 'ga': 'Hondúras', 'gd': 'Hondùras', 'gl': 'Honduras', 'gsw': 'Honduras', 'gu': 'હોન્ડુરસ', 'guz': 'Hondurasi', 'ha': 'Yankin Honduras', 'he': 'הונדורס', 'hi': 'होंडूरास', 'hi-Latn': 'Honduras', 'hr': 'Honduras', 'hsb': 'Honduras', 'ht': 'Honduras', 'hu': 'Honduras', 'hy': 'Հոնդուրաս', 'ia': 'Honduras', 'id': 'Honduras', 'ig': 'Honduras', 'is': 'Hondúras', 'it': 'Honduras', 'ja': 'ホンジュラス', 'jmc': 'Hondurasi', 'jv': 'Honduras', 'ka': 'ჰონდურასი', 'kaa': 'Гондурас', 'kaa-Cyrl': 'Гондурас', 'kab': 'Hunduras', 'kam': 'Hondurasi', 'kde': 'Hondulasi', 'kea': 'Onduras', 'kgp': 'Hãnura', 'khq': 'Honduras', 'ki': 'Hondurasi', 'kk': 'Гондурас', 'kk-Cyrl': 'Гондурас', 'kln': 'Emetab Honduras', 'km': 'ហុងឌូរ៉ាស', 'kn': 'ಹೊಂಡುರಾಸ್', 'ko': '온두라스', 'kok': 'हॉनडुरस', 'kok-Deva': 'हॉनडुरस', 'ks': 'ہونڈورس', 'ks-Arab': 'ہونڈورس', 'ksb': 'Honduasi', 'ksf': 'ɔnduras', 'ksh': 'Honduras', 'ku': 'Hondûras', 'kxv': 'honḍuras', 'kxv-Deva': 'होंडूरास', 'kxv-Latn': 'honḍuras', 'kxv-Orya': 'ହଣ୍ଡୁରାସ', 'kxv-Telu': 'హండురాస్', 'ky': 'Гондурас', 'lag': 'Honduráasi', 'lb': 'Honduras', 'lg': 'Hundurasi', 'lld': 'Honduras', 'ln': 'Ondurasɛ', 'lo': 'ຮອນດູຣັສ', 'lt': 'Hondūras', 'lu': 'Ondurase', 'luo': 'Honduras', 'luy': 'Hondurasi', 'lv': 'Hondurasa', 'mai': 'होंडुरास', 'mas': 'Hondurasi', 'mer': 'Hondurasi', 'mfe': 'Honduras', 'mg': 'Hondiorasy', 'mgh': 'Uhondurasi', 'mi': 'Honotura', 'mk': 'Хондурас', 'ml': 'ഹോണ്ടുറാസ്', 'mn': 'Гондурас', 'mr': 'होंडुरास', 'ms': 'Honduras', 'mt': 'il-Honduras', 'mua': 'Sǝr Honduras', 'my': 'ဟွန်ဒူးရပ်စ်', 'mzn': 'هندوراس', 'naq': 'Honduras', 'nb': 'Honduras', 'nd': 'Honduras', 'ne': 'हन्डुरास', 'nl': 'Honduras', 'nmg': 'Ɔndúras', 'nn': 'Honduras', 'no': 'Honduras', 'nqo': 'ߤߎ߲ߘߎߙߊ߫', 'nyn': 'Hondurasi', 'om': 'Hondurus', 'or': 'ହୋଣ୍ଡୁରାସ୍\u200c', 'pa': 'ਹੋਂਡੁਰਸ', 'pa-Guru': 'ਹੋਂਡੁਰਸ', 'pcm': 'Họndúras', 'pl': 'Honduras', 'ps': 'هانډوراس', 'pt': 'Honduras', 'qu': 'Honduras', 'rm': 'Honduras', 'rn': 'Hondurasi', 'ro': 'Honduras', 'rof': 'Hondurasi', 'ru': 'Гондурас', 'rwk': 'Hondurasi', 'saq': 'Hondurasi', 'sbp': 'Hondulasi', 'sc': 'Honduras', 'sd': 'هنڊورس', 'sd-Arab': 'هنڊورس', 'se': 'Honduras', 'seh': 'Honduras', 'ses': 'Honduras', 'sg': 'Honduräsi', 'shi': 'ⵀⵓⵏⴷⵓⵔⴰⵙ', 'shi-Latn': 'hunduras', 'shi-Tfng': 'ⵀⵓⵏⴷⵓⵔⴰⵙ', 'si': 'හොන්ඩුරාස්', 'sk': 'Honduras', 'sl': 'Honduras', 'smn': 'Honduras', 'sn': 'Honduras', 'so': 'Honduras', 'sq': 'Honduras', 'sr': 'Хондурас', 'sr-Cyrl': 'Хондурас', 'sr-Latn': 'Honduras', 'sv': 'Honduras', 'sw': 'Honduras', 'syr': 'ܗܘܢܕܘܪܣ', 'ta': 'ஹோண்டூராஸ்', 'te': 'హోండురాస్', 'teo': 'Hondurasi', 'tg': 'Гондурас', 'th': 'ฮอนดูรัส', 'ti': 'ሆንዱራስ', 'tk': 'Gonduras', 'to': 'Honitulasi', 'tr': 'Honduras', 'tt': 'Гондурас', 'twq': 'Honduras', 'tzm': 'Hinduras', 'ug': 'ھوندۇراس', 'uk': 'Гондурас', 'ur': 'ہونڈاروس', 'uz': 'Gonduras', 'uz-Cyrl': 'Гондурас', 'uz-Latn': 'Gonduras', 'vai': 'ꖽꖫꕟ', 'vai-Latn': 'Hɔndura', 'vai-Vaii': 'ꖽꖫꕟ', 'vec': 'Honduras', 'vi': 'Honduras', 'vun': 'Hondurasi', 'wae': 'Honduras', 'wo': 'Onduraas', 'xh': 'EHonduras', 'xnr': 'होंडूरास', 'xog': 'Hundurasi', 'yav': 'ɔndúlas', 'yi': 'האנדוראַס', 'yo': 'Hondurasi', 'yrl': 'Ũdura', 'yue': '宏都拉斯', 'yue-Hans': '宏都拉斯', 'yue-Hant': '宏都拉斯', 'zgh': 'ⵀⵓⵏⴷⵓⵔⴰⵙ', 'zh': '洪都拉斯', 'zh-Hans': '洪都拉斯', 'zh-Hant': '宏都拉斯', 'zu': 'i-Honduras'},
    'HR': {'af': 'Kroasië', 'agq': 'Kòwɛshìa', 'ak': 'Krowehyia', 'am': 'ክሮኤሽያ', 'ar': 'كرواتيا', 'as': 'ক্ৰোৱেছিয়া', 'asa': 'Korathia', 'ast': 'Croacia', 'az': 'Xorvatiya', 'az-Cyrl': 'Хорватија', 'az-Latn': 'Xorvatiya', 'bal-Latn': 'Króshiá', 'bas': 'Kròasìà', 'be': 'Харватыя', 'be-tarask': 'Харватыя', 'bez': 'Hukorasia', 'bg': 'Хърватия', 'blo': 'Krowasii', 'bm': 'Kroasi', 'bn': 'ক্রোয়েশিয়া', 'br': 'Kroatia', 'brx': 'क्रवेशिया', 'bs': 'Hrvatska', 'bs-Cyrl': 'Хрватска', 'bs-Latn': 'Hrvatska', 'ca': 'Croàcia', 'ca-ES-valencia': 'Croàcia', 'ccp': '𑄇𑄳𑄢𑄮𑄠𑄬𑄥𑄨𑄠', 'ce': 'Хорвати', 'ceb': 'Croatia', 'cgg': 'Korasiya', 'chr': 'ᎧᎶᎡᏏᎠ', 'ckb': 'کرۆواتیا', 'cs': 'Chorvatsko', 'cv': 'Хорвати', 'cy': 'Croatia', 'da': 'Kroatien', 'dav': 'Korasia', 'de': 'Kroatien', 'dje': 'Krwaasi', 'dsb': 'Chorwatska', 'dyo': 'Kroasi', 'dz': 'ཀྲོ་ཨེ་ཤ', 'ebu': 'Korasia', 'ee': 'Kroatsia nutome', 'el': 'Κροατία', 'el-polyton': 'Κροατία', 'en': 'Croatia', 'en-Dsrt': '𐐗𐑉𐐬𐐩𐑇𐐲', 'eo': 'Kroatujo', 'es': 'Croacia', 'et': 'Horvaatia', 'eu': 'Kroazia', 'ewo': 'Kəlowásia', 'fa': 'کرواسی', 'ff': 'Korwasii', 'ff-Adlm': '𞤑𞤵𞤪𞤱𞤢𞥄𞤧𞤭𞤴𞤢', 'ff-Latn': 'Korwasii', 'fi': 'Kroatia', 'fil': 'Croatia', 'fo': 'Kroatia', 'fr': 'Croatie', 'fur': 'Cravuazie', 'fy': 'Kroatië', 'ga': 'an Chróit', 'gd': 'A’ Chròthais', 'gl': 'Croacia', 'gsw': 'Kroaazie', 'gu': 'ક્રોએશિયા', 'guz': 'Korasia', 'ha': 'Kurowaishiya', 'he': 'קרואטיה', 'hi': 'क्रोएशिया', 'hi-Latn': 'Croatia', 'hr': 'Hrvatska', 'hsb': 'Chorwatska', 'ht': 'Croatie', 'hu': 'Horvátország', 'hy': 'Խորվաթիա', 'ia': 'Croatia', 'id': 'Kroasia', 'ie': 'Croatia', 'ig': 'Croatia', 'is': 'Króatía', 'it': 'Croazia', 'ja': 'クロアチア', 'jmc': 'Korasia', 'jv': 'Kroasia', 'ka': 'ხორვატია', 'kaa': 'Хорватия', 'kaa-Cyrl': 'Хорватия', 'kab': 'Kerwasya', 'kam': 'Korasia', 'kde': 'Kolasia', 'kea': 'Kroásia', 'kgp': 'Kroasa', 'khq': 'Krwaasi', 'ki': 'Korasia', 'kk': 'Хорватия', 'kk-Cyrl': 'Хорватия', 'kln': 'Emetab Croatia', 'km': 'ក្រូអាស៊ី', 'kn': 'ಕ್ರೊಯೇಷಿಯಾ', 'ko': '크로아티아', 'kok': 'क्रोयेशीया', 'kok-Deva': 'क्रोयेशीया', 'ks': 'کروشِیا', 'ks-Arab': 'کروشِیا', 'ksb': 'Kolasia', 'ksf': 'krwasí', 'ksh': 'Krowazije', 'ku': 'Xirwatistan', 'kxv': 'kroesia', 'kxv-Deva': 'क्रोएसिया', 'kxv-Latn': 'kroesia', 'kxv-Orya': 'କ୍ରସିଆ', 'kxv-Telu': 'క్రొయేసియా', 'ky': 'Хорватия', 'lag': 'Koréshia', 'lb': 'Kroatien', 'lg': 'Kurowesya', 'lld': 'Croazia', 'ln': 'Krowasi', 'lo': 'ໂຄຣເອເທຍ', 'lt': 'Kroatija', 'lu': 'Krowasi', 'luo': 'Croatia', 'luy': 'Korasia', 'lv': 'Horvātija', 'mai': 'क्रोएशिया', 'mas': 'Korasia', 'mer': 'Koroashia', 'mfe': 'Kroasi', 'mg': 'Kroasia', 'mgh': 'Ukorasia', 'mi': 'Koroātia', 'mk': 'Хрватска', 'ml': 'ക്രൊയേഷ്യ', 'mn': 'Хорват', 'mr': 'क्रोएशिया', 'ms': 'Croatia', 'mt': 'il-Kroazja', 'mua': 'kroatiya', 'my': 'ခရိုအေးရှား', 'mzn': 'کرواسی', 'naq': 'Croatiab', 'nb': 'Kroatia', 'nd': 'Croatia', 'ne': 'क्रोएशिया', 'nl': 'Kroatië', 'nmg': 'Kroasia', 'nn': 'Kroatia', 'no': 'Kroatia', 'nqo': 'ߞߙߏߥߊߛߌ߫', 'nus': 'Korwaatia', 'nyn': 'Korasiya', 'om': 'Kirooshiyaa', 'or': 'କ୍ରୋଏସିଆ', 'pa': 'ਕਰੋਏਸ਼ੀਆ', 'pa-Guru': 'ਕਰੋਏਸ਼ੀਆ', 'pcm': 'Kroéshia', 'pl': 'Chorwacja', 'ps': 'کرواشيا', 'pt': 'Croácia', 'qu': 'Croacia', 'rm': 'Croazia', 'rn': 'Korowasiya', 'ro': 'Croația', 'rof': 'Korasia', 'ru': 'Хорватия', 'rwk': 'Korasia', 'saq': 'Korasia', 'sbp': 'Kolasiya', 'sc': 'Croàtzia', 'sd': 'ڪروئيشيا', 'sd-Arab': 'ڪروئيشيا', 'se': 'Kroátia', 'seh': 'Croácia', 'ses': 'Krwaasi', 'sg': 'Kroasïi', 'shi': 'ⴽⵔⵡⴰⵜⵢⴰ', 'shi-Latn': 'krwatya', 'shi-Tfng': 'ⴽⵔⵡⴰⵜⵢⴰ', 'si': 'ක්\u200dරොඒෂියාව', 'sk': 'Chorvátsko', 'sl': 'Hrvaška', 'smn': 'Kroatia', 'sn': 'Korasia', 'so': 'Korweeshiya', 'sq': 'Kroaci', 'sr': 'Хрватска', 'sr-Cyrl': 'Хрватска', 'sr-Latn': 'Hrvatska', 'sv': 'Kroatien', 'sw': 'Croatia', 'syr': 'ܩܪܘܐܛܝܐ', 'ta': 'குரோஷியா', 'te': 'క్రొయేషియా', 'teo': 'Korasia', 'tg': 'Хорватия', 'th': 'โครเอเชีย', 'ti': 'ክሮኤሽያ', 'tk': 'Horwatiýa', 'to': 'Kuloisia', 'tr': 'Hırvatistan', 'tt': 'Хорватия', 'twq': 'Krwaasi', 'tzm': 'Krwatya', 'ug': 'كىرودىيە', 'uk': 'Хорватія', 'ur': 'کروشیا', 'uz': 'Xorvatiya', 'uz-Cyrl': 'Хорватия', 'uz-Latn': 'Xorvatiya', 'vai': 'ꖏꔓꔻꕩ', 'vai-Latn': 'Koresiya', 'vai-Vaii': 'ꖏꔓꔻꕩ', 'vec': 'Croasia', 'vi': 'Croatia', 'vun': 'Korasia', 'wae': 'Kroatie', 'wo': 'Korowasi', 'xh': 'ECroatia', 'xnr': 'क्रोएशिया', 'xog': 'Kurowesya', 'yav': 'Kolowasíi', 'yi': 'קראאַטיע', 'yo': 'Kòróátíà', 'yrl': 'Kuruwasiya', 'yue': '克羅埃西亞', 'yue-Hans': '克罗埃西亚', 'yue-Hant': '克羅埃西亞', 'zgh': 'ⴽⵔⵡⴰⵜⵢⴰ', 'zh': '克罗地亚', 'zh-Hans': '克罗地亚', 'zh-Hant': '克羅埃西亞', 'zu': 'i-Croatia'},
    'HT': {'af': 'Haïti', 'agq': 'Hǎetì', 'ak': 'Heiti', 'am': 'ሀይቲ', 'ar': 'هايتي', 'as': 'হাইটি', 'asa': 'Haiti', 'ast': 'Haití', 'az': 'Haiti', 'az-Cyrl': 'Һаити', 'az-Latn': 'Haiti', 'bal-Latn': 'Hayti', 'bas': 'Àitì', 'be': 'Гаіці', 'be-tarask': 'Гаіці', 'bez': 'Huhaiti', 'bg': 'Хаити', 'blo': 'Hayitii', 'bm': 'Ayiti', 'bn': 'হাইতি', 'br': 'Haiti', 'brx': 'हाइती', 'bs': 'Haiti', 'bs-Cyrl': 'Хаити', 'bs-Latn': 'Haiti', 'ca': 'Haití', 'ca-ES-valencia': 'Haití', 'ccp': '𑄦𑄭𑄖𑄨', 'ce': 'Гаити', 'ceb': 'Haiti', 'cgg': 'Haiti', 'chr': 'ᎮᎢᏘ', 'ckb': 'ھایتی', 'cs': 'Haiti', 'cv': 'Гаити', 'cy': 'Haiti', 'da': 'Haiti', 'dav': 'Haiti', 'de': 'Haiti', 'dje': 'Haiti', 'dsb': 'Haiti', 'dyo': 'Ayti', 'dz': 'ཧེ་ཊི', 'ebu': 'Haiti', 'ee': 'Haiti nutome', 'el': 'Αϊτή', 'el-polyton': 'Ἁϊτή', 'en': 'Haiti', 'en-Dsrt': '𐐐𐐩𐐻𐐨', 'eo': 'Haitio', 'es': 'Haití', 'et': 'Haiti', 'eu': 'Haiti', 'ewo': 'Aití', 'fa': 'هائیتی', 'ff': 'Haytii', 'ff-Adlm': '𞤖𞤢𞤴𞤼𞤭𞥅', 'ff-Latn': 'Haytii', 'fi': 'Haiti', 'fil': 'Haiti', 'fo': 'Haiti', 'fr': 'Haïti', 'fur': 'Haiti', 'fy': 'Haïti', 'ga': 'Háítí', 'gd': 'Haidhti', 'gl': 'Haití', 'gsw': 'Haiti', 'gu': 'હૈતિ', 'guz': 'Haiti', 'ha': 'Haiti', 'he': 'האיטי', 'hi': 'हैती', 'hi-Latn': 'Haiti', 'hr': 'Haiti', 'hsb': 'Haiti', 'ht': 'Haïti', 'hu': 'Haiti', 'hy': 'Հայիթի', 'ia': 'Haiti', 'id': 'Haiti', 'ig': 'Haiti', 'is': 'Haítí', 'it': 'Haiti', 'ja': 'ハイチ', 'jmc': 'Haiti', 'jv': 'Haiti', 'ka': 'ჰაიტი', 'kaa': 'Гаити', 'kaa-Cyrl': 'Гаити', 'kab': 'Hayti', 'kam': 'Haiti', 'kde': 'Haiti', 'kea': 'Aití', 'kgp': 'Ajti', 'khq': 'Haiti', 'ki': 'Haiti', 'kk': 'Гаити', 'kk-Cyrl': 'Гаити', 'kln': 'Emetab Haiti', 'km': 'ហៃទី', 'kn': 'ಹೈಟಿ', 'ko': '아이티', 'kok': 'हैती', 'kok-Deva': 'हैती', 'ks': 'ہیتی', 'ks-Arab': 'ہیتی', 'ksb': 'Haiti', 'ksf': 'ayiti', 'ksh': 'Ha’itti', 'ku': 'Haîtî', 'kxv': 'haiti', 'kxv-Deva': 'हाइती', 'kxv-Latn': 'haiti', 'kxv-Orya': 'ହାଇତି', 'kxv-Telu': 'హైటి', 'ky': 'Гаити', 'lag': 'Haíiti', 'lb': 'Haiti', 'lg': 'Hayiti', 'lld': 'Haiti', 'ln': 'Ayiti', 'lo': 'ໄຮຕິ', 'lt': 'Haitis', 'lu': 'Ayiti', 'luo': 'Haiti', 'luy': 'Haiti', 'lv': 'Haiti', 'mai': 'हैती', 'mas': 'Haiti', 'mer': 'Haiti', 'mfe': 'Ayti', 'mg': 'Haiti', 'mgh': 'Uhaiti', 'mi': 'Haiti', 'mk': 'Хаити', 'ml': 'ഹെയ്തി', 'mn': 'Гаити', 'mr': 'हैती', 'ms': 'Haiti', 'mt': 'il-Haiti', 'mua': 'Sǝr Haiti', 'my': 'ဟေတီ', 'mzn': 'هاییتی', 'naq': 'Haiti', 'nb': 'Haiti', 'nd': 'Hayiti', 'ne': 'हैटी', 'nl': 'Haïti', 'nmg': 'Haïti', 'nn': 'Haiti', 'no': 'Haiti', 'nqo': 'ߤߊߦߕߌ߫', 'nyn': 'Haiti', 'om': 'Haayitii', 'or': 'ହାଇତି', 'pa': 'ਹੈਤੀ', 'pa-Guru': 'ਹੈਤੀ', 'pcm': 'Haíti', 'pl': 'Haiti', 'ps': 'هایټي', 'pt': 'Haiti', 'qu': 'Haití', 'rm': 'Haiti', 'rn': 'Hayiti', 'ro': 'Haiti', 'rof': 'Haiti', 'ru': 'Гаити', 'rwk': 'Haiti', 'saq': 'Haiti', 'sbp': 'Hayiti', 'sc': 'Haiti', 'sd': 'هيٽي', 'sd-Arab': 'هيٽي', 'se': 'Haiti', 'seh': 'Haiti', 'ses': 'Haiti', 'sg': 'Haitïi', 'shi': 'ⵀⴰⵢⵜⵉ', 'shi-Latn': 'hayti', 'shi-Tfng': 'ⵀⴰⵢⵜⵉ', 'si': 'හයිටි', 'sk': 'Haiti', 'sl': 'Haiti', 'smn': 'Haiti', 'sn': 'Haiti', 'so': 'Haiti', 'sq': 'Haiti', 'sr': 'Хаити', 'sr-Cyrl': 'Хаити', 'sr-Latn': 'Haiti', 'sv': 'Haiti', 'sw': 'Haiti', 'syr': 'ܗܐܝܬܝ', 'ta': 'ஹைட்டி', 'te': 'హైతీ', 'teo': 'Haiti', 'tg': 'Гаити', 'th': 'เฮติ', 'ti': 'ሃይቲ', 'tk': 'Gaiti', 'to': 'Haiti', 'tr': 'Haiti', 'tt': 'Гаити', 'twq': 'Haiti', 'tzm': 'Hayti', 'ug': 'ھايتى', 'uk': 'Гаїті', 'ur': 'ہیٹی', 'uz': 'Gaiti', 'uz-Cyrl': 'Гаити', 'uz-Latn': 'Gaiti', 'vai': 'ꕌꔤꔳ', 'vai-Latn': 'Háiti', 'vai-Vaii': 'ꕌꔤꔳ', 'vec': 'Haiti', 'vi': 'Haiti', 'vun': 'Haiti', 'wae': 'Haiti', 'wo': 'Ayti', 'xh': 'EHaiti', 'xnr': 'हैती', 'xog': 'Hayiti', 'yav': 'ayíti', 'yi': 'האַיטי', 'yo': 'Haati', 'yrl': 'Aití', 'yue': '海地', 'yue-Hans': '海地', 'yue-Hant': '海地', 'zgh': 'ⵀⴰⵢⵜⵉ', 'zh': '海地', 'zh-Hans': '海地', 'zh-Hant': '海地', 'zu': 'i-Haiti'},
    'HU': {'af': 'Hongarye', 'agq': 'Hɔŋgàlè', 'ak': 'Hangari', 'am': 'ሀንጋሪ', 'ar': 'هنغاريا', 'as': 'হাংগেৰী', 'asa': 'Hungaria', 'ast': 'Hungría', 'az': 'Macarıstan', 'az-Cyrl': 'Маҹарыстан', 'az-Latn': 'Macarıstan', 'bal-Latn': 'Hangari', 'bas': 'Ɔ̀ŋgriì', 'be': 'Венгрыя', 'be-tarask': 'Венгрыя', 'bez': 'Huhungalia', 'bg': 'Унгария', 'bgn': 'هنگری', 'blo': 'Ɔŋgrii', 'bm': 'Hɔngri', 'bn': 'হাঙ্গেরি', 'br': 'Hungaria', 'brx': 'हांगारी', 'bs': 'Mađarska', 'bs-Cyrl': 'Мађарска', 'bs-Latn': 'Mađarska', 'ca': 'Hongria', 'ca-ES-valencia': 'Hongria', 'ccp': '𑄦𑄧𑄋𑄴𑄉𑄬𑄢𑄨', 'ce': 'Венгри', 'ceb': 'Hungary', 'cgg': 'Hangare', 'chr': 'ᎲᏂᎦᎵ', 'ckb': 'هەنگاریا', 'cs': 'Maďarsko', 'cv': 'Венгри', 'cy': 'Hwngari', 'da': 'Ungarn', 'dav': 'Hungaria', 'de': 'Ungarn', 'dje': 'Hungaari', 'dsb': 'Hungorska', 'dyo': 'Oŋri', 'dz': 'ཧཱང་གྷ་རི', 'ebu': 'Hungaria', 'ee': 'Hungari nutome', 'el': 'Ουγγαρία', 'el-polyton': 'Οὑγγαρία', 'en': 'Hungary', 'en-Dsrt': '𐐐𐐲𐑍𐑀𐐲𐑉𐐨', 'eo': 'Hungarujo', 'es': 'Hungría', 'et': 'Ungari', 'eu': 'Hungaria', 'ewo': 'Ongirí', 'fa': 'مجارستان', 'ff': 'Onngiri', 'ff-Adlm': '𞤖𞤢𞤲𞤺𞤢𞤪𞤭𞤴𞤢𞥄', 'ff-Latn': 'Onngiri', 'fi': 'Unkari', 'fil': 'Hungary', 'fo': 'Ungarn', 'fr': 'Hongrie', 'fur': 'Ongjarie', 'fy': 'Hongarije', 'ga': 'an Ungáir', 'gd': 'An Ungair', 'gl': 'Hungría', 'gsw': 'Ungarn', 'gu': 'હંગેરી', 'guz': 'Hungaria', 'ha': 'Hungari', 'he': 'הונגריה', 'hi': 'हंगरी', 'hi-Latn': 'Hungary', 'hr': 'Mađarska', 'hsb': 'Madźarska', 'ht': 'Hongrie', 'hu': 'Magyarország', 'hy': 'Հունգարիա', 'ia': 'Hungaria', 'id': 'Hungaria', 'ie': 'Hungaria', 'ig': 'Hungary', 'is': 'Ungverjaland', 'it': 'Ungheria', 'ja': 'ハンガリー', 'jmc': 'Hungaria', 'jv': 'Honggari', 'ka': 'უნგრეთი', 'kaa': 'Венгрия', 'kaa-Cyrl': 'Венгрия', 'kab': 'Hungri', 'kam': 'Hungaria', 'kde': 'Hungalia', 'kea': 'Ungria', 'kgp': 'Ũgrija', 'khq': 'Hungaari', 'ki': 'Hungaria', 'kk': 'Венгрия', 'kk-Cyrl': 'Венгрия', 'kln': 'Emetab Hungary', 'km': 'ហុងគ្រី', 'kn': 'ಹಂಗೇರಿ', 'ko': '헝가리', 'kok': 'हंगेरी', 'kok-Deva': 'हंगेरी', 'ks': 'ہَنگری', 'ks-Arab': 'ہَنگری', 'ksb': 'Hungalia', 'ksf': 'ɔngrí', 'ksh': 'Unjann', 'ku': 'Macaristan', 'kxv': 'hungarī', 'kxv-Deva': 'हंगरी', 'kxv-Latn': 'hungarī', 'kxv-Orya': 'ହଙ୍ଗେରି', 'kxv-Telu': 'హంగేరీ', 'ky': 'Венгрия', 'lag': 'Hungária', 'lb': 'Ungarn', 'lg': 'Hangare', 'lld': 'Ungaria', 'ln': 'Ongili', 'lo': 'ຮັງກາຣີ', 'lt': 'Vengrija', 'lu': 'Ongili', 'luo': 'Hungary', 'luy': 'Hungaria', 'lv': 'Ungārija', 'mai': 'हंगरी', 'mas': 'Hungaria', 'mer': 'Hangarĩ', 'mfe': 'Ongri', 'mg': 'Hongria', 'mgh': 'Uhungaria', 'mi': 'Hanekari', 'mk': 'Унгарија', 'ml': 'ഹംഗറി', 'mn': 'Унгар', 'mr': 'हंगेरी', 'ms': 'Hungary', 'mt': 'l-Ungerija', 'mua': 'Hungriya', 'my': 'ဟန်ဂေရီ', 'mzn': 'مجارستون', 'naq': 'Hongareieb', 'nb': 'Ungarn', 'nd': 'Hungary', 'ne': 'हङ्गेरी', 'nl': 'Hongarije', 'nmg': 'Ɔngría', 'nn': 'Ungarn', 'no': 'Ungarn', 'nqo': 'ߤߐ߲ߜ߭ߙߌ߫', 'nyn': 'Hangare', 'om': 'Hangaarii', 'or': 'ହଙ୍ଗେରୀ', 'pa': 'ਹੰਗਰੀ', 'pa-Guru': 'ਹੰਗਰੀ', 'pcm': 'Họ́ngári', 'pl': 'Węgry', 'ps': 'مجارستان', 'pt': 'Hungria', 'qu': 'Hungría', 'rm': 'Ungaria', 'rn': 'Hungariya', 'ro': 'Ungaria', 'rof': 'Hungaria', 'ru': 'Венгрия', 'rwk': 'Hungaria', 'saq': 'Hungaria', 'sbp': 'Hungaliya', 'sc': 'Ungheria', 'sd': 'هنگري', 'sd-Arab': 'هنگري', 'se': 'Ungár', 'seh': 'Hungria', 'ses': 'Hungaari', 'sg': 'Hongirùii', 'shi': 'ⵀⵏⵖⴰⵔⵢⴰ', 'shi-Latn': 'hnɣarya', 'shi-Tfng': 'ⵀⵏⵖⴰⵔⵢⴰ', 'si': 'හන්ගේරියාව', 'sk': 'Maďarsko', 'sl': 'Madžarska', 'smn': 'Uŋgar', 'sn': 'Hungary', 'so': 'Hangari', 'sq': 'Hungari', 'sr': 'Мађарска', 'sr-Cyrl': 'Мађарска', 'sr-Latn': 'Mađarska', 'sv': 'Ungern', 'sw': 'Hungaria', 'syr': 'ܡܓܪ', 'ta': 'ஹங்கேரி', 'te': 'హంగేరీ', 'teo': 'Hungaria', 'tg': 'Маҷористон', 'th': 'ฮังการี', 'ti': 'ሃንጋሪ', 'tk': 'Wengriýa', 'to': 'Hungakalia', 'tr': 'Macaristan', 'tt': 'Венгрия', 'twq': 'Hungaari', 'tzm': 'Henɣarya', 'ug': 'ۋېنگىرىيە', 'uk': 'Угорщина', 'ur': 'ہنگری', 'uz': 'Vengriya', 'uz-Cyrl': 'Венгрия', 'uz-Latn': 'Vengriya', 'vai': 'ꖽꘋꕭꔓ', 'vai-Latn': 'Hɔ́ngare', 'vai-Vaii': 'ꖽꘋꕭꔓ', 'vec': 'Ongarìa', 'vi': 'Hungary', 'vun': 'Hungaria', 'wae': 'Ungare', 'wo': 'Ongari', 'xh': 'EHungary', 'xnr': 'हंगरी', 'xog': 'Hangare', 'yav': 'ɔngilí', 'yi': 'אונגערן', 'yo': 'Hungari', 'yrl': 'Ũgiriya', 'yue': '匈牙利', 'yue-Hans': '匈牙利', 'yue-Hant': '匈牙利', 'zgh': 'ⵀⵏⵖⴰⵔⵢⴰ', 'zh': '匈牙利', 'zh-Hans': '匈牙利', 'zh-Hant': '匈牙利', 'zu': 'i-Hungary'},
    'Hanb': {'af': 'Han met Bopomofo', 'ak': 'Hanse a Bopomofo kasa ka ho', 'am': 'ሃን ከቦፖሞፎ ጋር', 'ar': 'هانب', 'as': 'বোপোমোফোৰ সৈতে হান', 'bal-Latn': 'Hán gón Bópómópóá', 'be': 'хан з бапамофа', 'be-tarask': 'хан з бапамофа', 'bg': 'ханб', 'bgn': 'هانب', 'bn': 'হ্যান-বোপোমোফো', 'br': 'han gant bopomofo', 'brx': 'बप’मफ’ थानाय हान', 'bs': 'pismo hanb', 'bs-Cyrl': 'хан с бопомофо писмом', 'bs-Latn': 'pismo hanb', 'ca': 'han amb bopomofo', 'ca-ES-valencia': 'han amb bopomofo', 'ccp': '𑄦𑄳𑄠𑄚𑄴𑄝𑄨', 'ce': 'ханьб', 'chr': 'ᎭᏂ ᎾᎿ ᏆᏉᎼᏬ', 'ckb': 'هان لەگەڵ بۆپۆمۆفۆ', 'cy': 'Han gyda Bopomofo', 'da': 'han med bopomofo', 'de': 'Han mit Bopomofo', 'dsb': 'chinšćina z bopomofo', 'el': 'Χανμπ', 'el-polyton': 'Χανμπ', 'en': 'Han with Bopomofo', 'es': 'han con bopomofo', 'et': 'hanbi', 'eu': 'idazkera txinatarra bopomofoarekin', 'fa': 'هان با بوموپوفو', 'ff-Adlm': '𞤖𞤢𞥄𞤲 𞤫 𞤄𞤮𞤨𞤮𞤥𞤮𞤬𞤮', 'fi': 'kiinan han ja bopomofo', 'fil': 'Han na may Bopomofo', 'fr': 'han avec bopomofo', 'ga': 'Han agus Bopomofo', 'gd': 'Han le Bopomofo', 'gl': 'han con bopomofo', 'gu': 'હાન્બ', 'ha': 'Han tare da Bopomofo', 'he': 'האן עם בופומופו', 'hi': 'हांब', 'hi-Latn': 'Han with Bopomofo', 'hr': 'hanb pismo', 'hsb': 'chinšćina z bopomofo', 'ht': 'han avec bopomofo', 'hu': 'Han bopomofóval', 'hy': 'հանբ', 'ia': 'han con bopomofo', 'id': 'Han dengan Bopomofo', 'ig': 'Han na Bopomofo', 'it': 'han, bopomofo', 'ja': '漢語注音字母', 'jv': 'Han nganggo Bopomofo', 'ka': 'ჰანბი', 'kgp': 'hỹnme', 'kk': 'ханб жазуы', 'kk-Cyrl': 'ханб жазуы', 'km': 'ហានប៍', 'kn': 'ಬೋಪೋಮೊಫೋ ಜೊತೆಗೆ ಹಾನ್', 'ko': '주음 자모', 'kok': 'बोपोफोमो सयत हान', 'kok-Deva': 'बोपोफोमो सयत हान', 'ku': 'hanîya bi bopomofoyê', 'ky': 'Ханб', 'lo': 'ຮັນ', 'lt': 'hanbų', 'lv': 'haņu ar bopomofo', 'mai': 'बोपोमोफो कें संग हान', 'mi': 'Hana me te Papamawha', 'mk': 'ханб', 'ml': 'ഹൻബ്', 'mn': 'Бопомофотой ханз', 'mr': 'बोपोमोफोसह हान', 'ms': 'Han dengan Bopomofo', 'my': 'ဟန်ဘ်', 'nb': 'han-kinesisk med bopomofo', 'ne': 'हान्ब', 'nl': 'Han met Bopomofo', 'no': 'han-kinesisk med bopomofo', 'nqo': 'ߤߊ߲߫ ߓߔߏߡߝߏ ߘߐ߫', 'or': 'ବୋପୋମୋଫୋ ସହିତ ହାନ୍\u200c', 'pa': 'ਹਾਂਬ', 'pa-Guru': 'ਹਾਂਬ', 'pcm': 'Han Wit Bopomófo', 'pl': 'chińskie z bopomofo', 'ps': 'هن او بوپوفومو', 'qu': 'Han with Bopomofo', 'rm': 'han cun bopomofo', 'ru': 'ханьб', 'sc': 'han cun bopomofo', 'sd': 'بوپوموفو سان هين', 'sd-Arab': 'بوپوموفو سان هين', 'si': 'හැන්ඩ්බ්', 'sk': 'čínske a bopomofo', 'sl': 'han + bopomofo', 'so': 'Han iyo Bobomofo', 'sq': 'hanbik', 'sr': 'ханб', 'sr-Cyrl': 'ханб', 'sv': 'han med bopomofo', 'sw': 'Kihan chenye Bopomofo', 'ta': 'ஹன்ப்', 'te': 'హాన్బ్', 'tg': 'Хан бо Бопомофо', 'th': 'จีนกลาง', 'ti': 'ሃን ምስ ቦፖሞፎ', 'tk': 'Bopomofo han elipbiýi', 'to': 'tohinima fakahānipi', 'tr': 'Han - Bopomofo', 'uk': 'хань', 'ur': 'ہینب', 'vec': 'han co bopomofo', 'vi': 'Chữ Hán có chú âm', 'xnr': 'हांब', 'yo': 'Han pẹ̀lú Bopomófò', 'yrl': 'hãbi', 'yue': '漢語注音', 'yue-Hans': '汉语注音', 'yue-Hant': '漢語注音', 'zh': '注音汉字', 'zh-Hans': '注音汉字', 'zh-Hant': '標上注音符號的漢字', 'zu': 'isi-Hanb'},
    'Hang': {'af': 'Hangul', 'ak': 'Hangul kasa', 'am': 'ሐንጉል', 'ar': 'الهانغول', 'as': 'হেঙ্গুল', 'ast': 'hangul', 'az': 'hanqıl', 'az-Latn': 'hanqıl', 'bal-Latn': 'Hangul', 'be': 'хангыль', 'be-tarask': 'хангыль', 'bg': 'хангъл', 'bgn': 'هانگول', 'bn': 'হাঙ্গুল', 'br': 'hangeul', 'brx': 'हांगुल', 'bs': 'pismo hangul', 'bs-Cyrl': 'хангул', 'bs-Latn': 'pismo hangul', 'ca': 'hangul', 'ca-ES-valencia': 'hangul', 'ccp': '𑄦𑄋𑄴𑄉𑄪𑄣𑄴', 'ce': 'хангыль', 'chr': 'ᎭᏂᎫᎵ', 'ckb': 'ھانگول', 'cs': 'hangul', 'cy': 'Hangul', 'da': 'hangul', 'de': 'Hangul', 'dsb': 'hangul', 'dz': 'ཧཱན་གུལ་ཡིག་གུ', 'ee': 'hangulgbeŋɔŋlɔ', 'el': 'Χανγκούλ', 'el-polyton': 'Χανγκούλ', 'en': 'Hangul', 'en-Dsrt': '𐐐𐐪𐑌𐑀𐐲𐑊', 'es': 'hangul', 'et': 'korea', 'eu': 'hangula', 'fa': 'هانگول', 'ff-Adlm': '𞤖𞤢𞤲𞤺𞤵𞥅𞤤', 'fi': 'hangul', 'fil': 'Hangul', 'fo': 'hangul', 'fr': 'hangûl', 'fy': 'Hangul', 'ga': 'Hangalach', 'gd': 'Hangul', 'gl': 'hangul', 'gsw': 'Hangul', 'gu': 'હંગુલ', 'ha': 'Yaren Hangul', 'he': 'האנגול', 'hi': 'हंगुल', 'hi-Latn': 'Hangul', 'hr': 'hangul pismo', 'hsb': 'hangul', 'ht': 'hangûl', 'hu': 'Hangul', 'hy': 'հանգըլ', 'ia': 'hangul', 'id': 'Hangul', 'ig': 'Mkpụrụ ọkwụ Hangụl', 'is': 'hangul', 'it': 'hangul', 'ja': 'ハングル', 'jv': 'Hangul', 'ka': 'ჰანგული', 'kea': 'hangul', 'kgp': 'hỹngur', 'kk': 'хангыл жазуы', 'kk-Cyrl': 'хангыл жазуы', 'km': 'ហាំងកុល', 'kn': 'ಹ್ಯಾಂಗುಲ್', 'ko': '한글', 'kok': 'हंगूल', 'kok-Deva': 'हंगूल', 'ks': 'ہانگُل', 'ks-Arab': 'ہانگُل', 'ksh': 'korrejaanesche Schreff', 'ku': 'hangulî', 'ky': 'Хангул', 'lb': 'Hangul', 'lo': 'ຮັນກູນ', 'lrc': 'ھانگوٙل', 'lt': 'hangul', 'lv': 'hangils', 'mai': 'हंगुल', 'mi': 'Hāngū', 'mk': 'хангул', 'ml': 'ഹാംഗുൽ', 'mn': 'хангыль', 'mr': 'हंगुल', 'ms': 'Hangul', 'my': 'ဟန်ဂူးလ်', 'mzn': 'هانگول', 'nb': 'hangul', 'ne': 'हान्गुल', 'nl': 'Hangul', 'nn': 'hangul', 'no': 'hangul', 'nqo': 'ߤߊ߲ߜ߭ߎ', 'or': 'ହାଙ୍ଗୁଲ୍', 'pa': 'ਹੰਗੁਲ', 'pa-Guru': 'ਹੰਗੁਲ', 'pcm': 'Hángul', 'pl': 'hangul', 'ps': 'هنګولي', 'pt': 'hangul', 'qu': 'Hangul', 'rm': 'hangul', 'ro': 'hangul', 'ru': 'хангыль', 'sc': 'hangul', 'sd': 'هنگول', 'sd-Arab': 'هنگول', 'se': 'hangul', 'si': 'හැන්ගුල්', 'sk': 'hangul', 'sl': 'hangul', 'so': 'Hanguul', 'sq': 'hangul', 'sr': 'хангул', 'sr-Cyrl': 'хангул', 'sr-Latn': 'hangul', 'sv': 'hangul', 'sw': 'Kihangul', 'ta': 'ஹங்குல்', 'te': 'హంగుల్', 'tg': 'Ҳангул', 'th': 'ฮันกึล', 'ti': 'ሃንጉል', 'tk': 'Hangyl elipbiýi', 'to': 'tohinima fakakōlea-hāngūlu', 'tr': 'Hangıl', 'uk': 'хангиль', 'ur': 'ہنگول', 'uz': 'hangul', 'uz-Cyrl': 'Хангул', 'uz-Latn': 'hangul', 'vec': 'hangul', 'vi': 'Chữ Hàn', 'xnr': 'हंगुल', 'yo': 'Háńgùlù', 'yrl': 'hãgu', 'yue': '韓文字', 'yue-Hans': '韩文字', 'yue-Hant': '韓文字', 'zh': '谚文', 'zh-Hans': '谚文', 'zh-Hant': '諺文', 'zu': 'isi-Hangul'},
    'Hani': {'af': 'Han', 'ak': 'Han', 'am': 'ሃን', 'ar': 'الهان', 'as': 'হান', 'ast': 'han', 'az': 'han', 'az-Latn': 'han', 'bal-Latn': 'Han', 'be': 'хан', 'be-tarask': 'хан', 'bg': 'хан', 'bgn': 'هانی', 'bn': 'হ্যান', 'br': 'sinalunioù (han)', 'brx': 'हान', 'bs': 'pismo han', 'bs-Cyrl': 'хан', 'bs-Latn': 'pismo han', 'ca': 'han', 'ca-ES-valencia': 'han', 'ccp': '𑄦𑄳𑄠𑄚𑄴', 'ce': 'китайн', 'chr': 'ᎭᏂ', 'ckb': 'ھان', 'cs': 'han', 'cy': 'Han', 'da': 'han', 'de': 'Chinesisch', 'dsb': 'han', 'dz': 'རྒྱ་ནག་ཡིག་གུ', 'ee': 'hangbeŋɔŋlɔ', 'el': 'Χαν', 'el-polyton': 'Χαν', 'en': 'Han', 'en-Dsrt': '𐐐𐐪𐑌', 'es': 'han', 'eu': 'idazkera txinatarra', 'fa': 'هان', 'ff-Adlm': '𞤖𞤢𞥄𞤲', 'fi': 'kiinalainen han', 'fil': 'Han', 'fo': 'han', 'fr': 'sinogrammes', 'fur': 'han', 'fy': 'Han', 'ga': 'Han', 'gd': 'Han', 'gl': 'han', 'gsw': 'Chineesisch', 'gu': 'હાન', 'ha': 'Mutanen Han na ƙasar Sin', 'he': 'האן', 'hi': 'हान', 'hi-Latn': 'Han', 'hr': 'hansko pismo', 'hsb': 'chinsce', 'ht': 'sinogrammes', 'hu': 'Han', 'hy': 'չինական', 'ia': 'han', 'id': 'Han', 'ig': 'Mkpụrụ ọkwụ Han', 'is': 'kínverskt', 'it': 'han', 'ja': '漢字', 'jv': 'Han', 'ka': 'ჰანი', 'kea': 'han', 'kgp': 'hỹn', 'kk': 'қытай жазуы', 'kk-Cyrl': 'қытай жазуы', 'km': 'ហាន', 'kn': 'ಹಾನ್', 'ko': '한자', 'kok': 'हान', 'kok-Deva': 'हान', 'ks': 'ہان', 'ks-Arab': 'ہان', 'ksh': 'en schineesesche Schreff', 'ku': 'hanî', 'ky': 'Хань', 'lb': 'Chinesesch', 'lo': 'ຮານ', 'lrc': 'ھانی', 'lt': 'han', 'lv': 'haņu', 'mai': 'हान', 'mi': 'Hana', 'mk': 'ханско писмо', 'ml': 'ഹാൻ', 'mn': 'ханз', 'mr': 'हान', 'ms': 'Han', 'my': 'ဟန်', 'mzn': 'هان', 'nb': 'han', 'ne': 'हान', 'nl': 'Han', 'nn': 'han', 'no': 'han', 'nqo': 'ߛߌߣߏߜ߭ߙߊߡ', 'or': 'ହାନ୍', 'pa': 'ਹਾਨ', 'pa-Guru': 'ਹਾਨ', 'pcm': 'Chainiz', 'pl': 'chińskie', 'ps': 'هن', 'pt': 'han', 'qu': 'Han', 'rm': 'han', 'ro': 'han', 'ru': 'китайская', 'sc': 'han', 'sd': 'هين', 'sd-Arab': 'هين', 'se': 'kiinnaš', 'si': 'හන්', 'sk': 'čínske', 'sl': 'kanji', 'so': 'Han', 'sq': 'han', 'sr': 'хан', 'sr-Cyrl': 'хан', 'sr-Latn': 'han', 'sv': 'han', 'sw': 'Kihan', 'ta': 'ஹன்', 'te': 'హాన్', 'tg': 'Хан', 'th': 'ฮั่น', 'ti': 'ሃን', 'tk': 'Han elipbiýi', 'to': 'tohinima fakasiaina', 'tr': 'Han', 'ug': 'خەنچە', 'uk': 'китайська', 'ur': 'ہان', 'uz': 'xitoy', 'uz-Cyrl': 'Хан', 'uz-Latn': 'xitoy', 'vec': 'han', 'vi': 'Chữ Hán', 'xnr': 'हान', 'yo': 'Háànù', 'yrl': 'hã', 'yue': '漢語', 'yue-Hans': '汉语', 'yue-Hant': '漢語', 'zh': '汉字', 'zh-Hans': '汉字', 'zh-Hant': '漢字', 'zu': 'isi-Han'},
    'Hano': {'am': 'ሀኑኦ', 'ar': 'الهانونو', 'ast': 'hanunó’o', 'az': 'hanunu', 'az-Latn': 'hanunu', 'bal-Latn': 'Hanunu', 'bg': 'Хануну', 'bn': 'হ্যানুনু', 'brx': 'हानुनु', 'bs': 'hanuno pismo', 'bs-Cyrl': 'хануно', 'bs-Latn': 'hanuno pismo', 'ca': 'hanunoo', 'ca-ES-valencia': 'hanunoo', 'ccp': '𑄦𑄳𑄠𑄚𑄪𑄚𑄪', 'cs': 'hanunóo', 'da': 'hanunoo', 'de': 'Hanunoo', 'el': 'Χανούνου', 'el-polyton': 'Χανούνου', 'en': 'Hanunoo', 'en-Dsrt': '𐐐𐐲𐑌𐐭𐐲𐑌𐐭', 'es': 'hanunoo', 'et': 'hanunoo', 'eu': 'hanunuera', 'fa': 'هانونویی', 'ff-Adlm': '𞤖𞤢𞤲𞤵𞥅𞤲𞤮', 'fi': 'hanunoolainen', 'fo': 'hanunoo', 'fr': 'hanounóo', 'fy': 'Hanunoo', 'ga': 'Hananúis', 'gd': 'Hanunoo', 'gsw': 'Hanunoo', 'gu': 'હનુનૂ', 'he': 'האנונו', 'hi': 'हनुनू', 'hi-Latn': 'Hanunoo', 'hr': 'hanunoo pismo', 'ht': 'hanounóo', 'hu': 'Hanunoo', 'id': 'Hanunoo', 'ig': 'Hanunoo', 'it': 'hanunoo', 'ja': 'ハヌノオ文字', 'ka': 'ჰანუნოო', 'kgp': 'hỹnũnũ', 'kk': 'хануну жазуы', 'kk-Cyrl': 'хануну жазуы', 'kn': 'ಹನೂನೂ', 'ko': '하누누 문자', 'ks': 'ہانُنوٗ', 'ks-Arab': 'ہانُنوٗ', 'lb': 'Hanunoo', 'lo': 'ຮານູໂນໂອ', 'lt': 'hanuno', 'mk': 'хануноовско', 'ml': 'ഹനുനൂ', 'mn': 'Хануноо', 'mr': 'हनुनू', 'ms': 'Hanunoo', 'nb': 'hanunoo', 'ne': 'हानुनु', 'nl': 'Hanunoo', 'nn': 'hanunoo', 'no': 'hanunoo', 'nqo': 'ߤߊߣߎߣߏ', 'or': 'ହାନୁନ୍', 'pl': 'hanunoo', 'pt': 'hanunoo', 'rm': 'hanunoo', 'ro': 'hanunoo', 'ru': 'хануну', 'sc': 'hanunoo', 'sl': 'hanunski', 'so': 'Qoraalka Hanunoo', 'sq': 'hanunoisht', 'sr': 'хануно', 'sr-Cyrl': 'хануно', 'sr-Latn': 'hanuno', 'sv': 'hanunó’o', 'ta': 'ஹனுனூ', 'te': 'హనునూ', 'tg': 'Хануну', 'th': 'ฮานูโนโอ', 'to': 'tohinima fakahanunōʻo', 'tr': 'Hanunoo', 'ug': 'خانۇنۇچە', 'uk': 'хануну', 'vec': 'hanunoo', 'vi': 'Chữ Hanunoo', 'xnr': 'हनुनू', 'yrl': 'hanunu', 'yue': '哈努諾文', 'yue-Hans': '哈努诺文', 'yue-Hant': '哈努諾文', 'zh': '汉奴罗文', 'zh-Hans': '汉奴罗文', 'zh-Hant': '哈努諾文', 'zu': 'isi-Hanunoo'},
    'Hans': {'af': 'Vereenvoudig', 'ak': 'Kyaena Kasa Hanse', 'am': 'ቀለል ያለ', 'ar': 'المبسطة', 'as': 'সৰলীকৃত', 'ast': 'simplificáu', 'az': 'sadələşmiş', 'az-Arab': 'ساده\u200cلنمیش', 'az-Latn': 'sadələşmiş', 'bal-Latn': 'Hán (sádah kortagén)', 'be': 'спрошчанае кітайскае', 'be-tarask': 'спрошчанае кітайскае', 'bg': 'опростена', 'blo': 'aburoka', 'bn': 'সরলীকৃত', 'bo': 'རྒྱ་ཡིག་གསར་པ།', 'br': 'eeunaet', 'brx': 'गोरलै खालामना होनाय', 'bs': 'pojednostavljeno', 'bs-Cyrl': 'поједностављени', 'bs-Latn': 'pojednostavljeno', 'ca': 'simplificat', 'ca-ES-valencia': 'simplificat', 'ccp': '𑄅𑄪𑄎𑄪𑄅𑄪𑄏𑄫', 'ce': 'атта китайн', 'ceb': 'Pinasimple', 'chr': 'ᎠᎯᏗᎨ', 'ckb': 'ئاسانکراو', 'cs': 'zjednodušené', 'cv': 'ҫӑмӑллатнӑн китай', 'cy': 'Symledig', 'da': 'forenklet', 'de': 'Vereinfacht', 'doi': 'सरलीकृत', 'dsb': 'zjadnorjone', 'dz': 'རྒྱ་ཡིག་ ལུགས་གསར་', 'ee': 'Chinesegbeŋɔŋlɔ', 'el': 'Απλοποιημένο', 'el-polyton': 'Απλοποιημένο', 'en': 'Simplified', 'en-Dsrt': '𐐝𐐮𐑋𐐹𐑊𐐮𐑁𐐴𐐼 𐐐𐐪𐑌', 'es': 'simplificado', 'et': 'lihtsustatud', 'eu': 'sinplifikatua', 'fa': 'ساده\u200cشده', 'ff-Adlm': '𞤖𞤮𞤴𞤲𞤢𞥄𞤲𞤣𞤫', 'fi': 'yksinkertaistettu', 'fil': 'Pinasimple', 'fo': 'einkult', 'fr': 'simplifié', 'fur': 'Han semplificât', 'fy': 'Ferienfâldigd', 'ga': 'Simplithe', 'gd': 'Simplichte', 'gl': 'simplificado', 'gsw': 'Veräifachti Chineesischi Schrift', 'gu': 'સરળીકૃત', 'ha': 'Sauƙaƙaƙƙe', 'he': 'פשוט', 'hi': 'सरलीकृत', 'hi-Latn': 'Simplified', 'hr': 'pojednostavljeno pismo', 'hsb': 'zjednorjene', 'ht': 'simplifié', 'hu': 'Egyszerűsített', 'hy': 'պարզեցված', 'ia': 'simplificate', 'id': 'Sederhana', 'ie': 'simplificat', 'ig': 'Nke dị mfe', 'ii': 'ꈝꐮꁱꂷ', 'is': 'einfaldað', 'it': 'semplificato', 'ja': '簡体字', 'jv': 'Prasaja', 'ka': 'გამარტივებული', 'kaa': 'аңсатластырылған', 'kaa-Cyrl': 'аңсатластырылған', 'kea': 'simplifikadu', 'kgp': 'sĩpri-há', 'kk': 'жеңілдетілген жазу', 'kk-Cyrl': 'жеңілдетілген жазу', 'km': 'អក្សរ\u200bចិន\u200bកាត់', 'kn': 'ಸರಳೀಕೃತ', 'ko': '간체', 'kok': 'सोंपी', 'kok-Deva': 'सोंपी', 'ks': 'سَہل ﴿ترجمع اِشارٕ: یِم ورژن رَسم الخط ہُک ناؤ چھُ چیٖنی باپتھ زَبانٕ ناؤ کِس مجموعَس سٕتؠ اِستعمال یِوان کرنٕہ۔﴾', 'ks-Arab': 'سَہل ﴿ترجمع اِشارٕ: یِم ورژن رَسم الخط ہُک ناؤ چھُ چیٖنی باپتھ زَبانٕ ناؤ کِس مجموعَس سٕتؠ اِستعمال یِوان کرنٕہ۔﴾', 'ks-Deva': 'आसान (तरजुम इशार: स्क्रिप्ट नवुक यि वर्ज़न छु चीनी बापथ ज़बान नाव किस मुरकब कि इस्तिमल करान।)', 'ksh': 'eijfacher', 'ku': 'sadekirî', 'kxv': 'sahaj', 'kxv-Deva': 'साॅहाॅजाॅ', 'kxv-Latn': 'sahaj', 'kxv-Orya': 'ସହଜ', 'kxv-Telu': 'సరళీకృతం', 'ky': 'Жөнөкөйлөштүрүлгөн', 'lb': 'Vereinfacht', 'lld': 'scemplifiché', 'lo': 'ແບບຮຽບງ່າຍ', 'lrc': 'سادە بیە', 'lt': 'supaprastinti', 'lv': 'vienkāršotā', 'mai': 'सरलीकृत', 'mi': 'Māmā', 'mk': 'поедноставено', 'ml': 'ലളിതവൽക്കരിച്ചത്', 'mn': 'хялбаршуулсан', 'mni': 'লাইথোকহল্লবা', 'mni-Beng': 'লাইথোকহল্লবা', 'mr': 'सरलीकृत', 'ms': 'Ringkas', 'mt': 'Simplifikat', 'my': 'ရိုးရှင်း', 'mzn': 'ساده\u200cبَیی هان', 'nb': 'forenklet', 'ne': 'सरलिकृत चिनियाँ', 'nl': 'vereenvoudigd', 'nn': 'forenkla', 'no': 'forenklet', 'nqo': 'ߘߐߞߣߍߣߍ߲', 'om': 'Salphifame', 'or': 'ସରଳୀକୃତ', 'os': 'Ӕнцонгонд китайаг', 'pa': 'ਸਰਲ', 'pa-Guru': 'ਸਰਲ', 'pcm': 'Ízí Chainíz Lángwej', 'pl': 'uproszczone', 'ps': 'ساده شوی', 'pt': 'simplificado', 'qu': 'Simplificado', 'rm': 'simplifitgà', 'ro': 'simplificată', 'ru': 'упрощенная', 'sa': 'सरलीकृत', 'sat': 'ᱥᱤᱢᱯᱞᱤᱯᱟᱭᱤᱰ', 'sat-Olck': 'ᱥᱤᱢᱯᱞᱤᱯᱟᱭᱤᱰ', 'sc': 'semplificadu', 'sd': 'سادي', 'sd-Arab': 'سادي', 'sd-Deva': 'सादी थियल (तरजुमे जो द॒स : लिखत जे नाले जे हिन बयानु खे चीनीअ लाए भाषा जे नाले सां गद॒ मिलाए करे इस्तेमाल कयो वेंदो आहे)', 'se': 'álki', 'si': 'සුළුකළ', 'sk': 'zjednodušené', 'sl': 'poenostavljena pisava', 'so': 'La fududeeyay', 'sq': 'i thjeshtuar', 'sr': 'поједностављено кинеско писмо', 'sr-Cyrl': 'поједностављено кинеско писмо', 'sr-Latn': 'pojednostavljeno kinesko pismo', 'su': 'Sederhana', 'su-Latn': 'Sederhana', 'sv': 'förenklad', 'sw': 'Rahisi', 'ta': 'எளிதாக்கப்பட்டது', 'te': 'సరళీకృతం', 'tg': 'Осонфаҳм', 'th': 'ตัวย่อ', 'ti': 'ዝተቐለለ', 'tk': 'Ýönekeýleşdirilen', 'to': 'fakafaingofua', 'tpi': 'Han (isipela)', 'tr': 'Basitleştirilmiş', 'tt': 'гадиләштерелгән', 'ug': 'ئاددىي خەنچە', 'uk': 'спрощена', 'ur': 'آسان', 'uz': 'soddalashgan', 'uz-Cyrl': 'Соддалаштирилган', 'uz-Latn': 'soddalashgan', 'vec': 'senpio', 'vi': 'Giản thể', 'wae': 'Vereifačt', 'wo': 'Buñ woyofal', 'xh': 'IsiHans Esenziwe Lula', 'xnr': 'सादी', 'yo': 'tí wọ́n mú rọrùn.', 'yrl': 'iwasuĩma', 'yue': '簡體', 'yue-Hans': '简体', 'yue-Hant': '簡體', 'zh': '简体', 'zh-Hans': '简体', 'zh-Hant': '簡體', 'zu': 'enziwe lula'},
    'Hant': {'af': 'Tradisioneel', 'ak': 'Tete', 'am': 'ባህላዊ', 'ar': 'التقليدية', 'as': 'পৰম্পৰাগত', 'ast': 'tradicional', 'az': 'ənənəvi', 'az-Arab': 'سنتی', 'az-Latn': 'ənənəvi', 'bal-Latn': 'Hán (asligén)', 'be': 'традыцыйнае кітайскае', 'be-tarask': 'традыцыйнае кітайскае', 'bg': 'традиционна', 'blo': 'tututu', 'bn': 'ঐতিহ্যবাহী', 'bo': 'རྒྱ་ཡིག་རྙིང་པ།', 'br': 'hengounel', 'brx': 'आसार-खान्थिनि', 'bs': 'tradicionalno', 'bs-Cyrl': 'традиционални', 'bs-Latn': 'tradicionalno', 'ca': 'tradicional', 'ca-ES-valencia': 'tradicional', 'ccp': '𑄢𑄨𑄘𑄨𑄥𑄪𑄘𑄮𑄟𑄴', 'ce': 'ламастан китайн', 'ceb': 'Tradisyonal', 'chr': 'ᎤᏦᏍᏗ', 'ckb': 'دێرین', 'cs': 'tradiční', 'cv': 'традициллӗн китай', 'cy': 'Traddodiadol', 'da': 'traditionelt', 'de': 'Traditionell', 'doi': 'रवायती', 'dsb': 'tradionalne', 'dz': 'ལུགས་རྙིང་ རྒྱ་ཡིག', 'ee': 'Blema Chinesegbeŋɔŋlɔ', 'el': 'Παραδοσιακό', 'el-polyton': 'Παραδοσιακό', 'en': 'Traditional', 'en-Dsrt': '𐐓𐑉𐐲𐐼𐐮𐑇𐐲𐑌𐐲𐑊 𐐐𐐪𐑌', 'es': 'tradicional', 'et': 'traditsiooniline', 'eu': 'tradizionala', 'fa': 'سنتی', 'ff-Adlm': '𞤚𞤢𞤱𞤢𞥄𞤲𞤣𞤫', 'fi': 'perinteinen', 'fil': 'Tradisyonal', 'fo': 'vanligt', 'fr': 'traditionnel', 'fur': 'Han tradizionâl', 'fy': 'Traditjoneel', 'ga': 'Traidisiúnta', 'gd': 'Tradaiseanta', 'gl': 'tradicional', 'gsw': 'Tradizionelli Chineesischi Schrift', 'gu': 'પરંપરાગત', 'ha': 'Na gargajiya', 'he': 'מסורתי', 'hi': 'पारंपरिक', 'hi-Latn': 'Traditional', 'hr': 'tradicionalno pismo', 'hsb': 'tradicionalne', 'ht': 'traditionnel', 'hu': 'Hagyományos', 'hy': 'ավանդական', 'ia': 'traditional', 'id': 'Tradisional', 'ie': 'traditional', 'ig': 'Omenala', 'ii': 'ꀎꋏꁱꂷ', 'is': 'hefðbundið', 'it': 'tradizionale', 'ja': '繁体字', 'jv': 'Tradhisional', 'ka': 'ტრადიციული', 'kaa': 'дәстүрли', 'kaa-Cyrl': 'дәстүрли', 'kea': 'tradisional', 'kgp': 'si-pẽ', 'kk': 'дәстүрлі жазу', 'kk-Cyrl': 'дәстүрлі жазу', 'km': 'អក្សរ\u200bចិន\u200bពេញ', 'kn': 'ಸಾಂಪ್ರದಾಯಿಕ', 'ko': '번체', 'kok': 'पारंपारीक', 'kok-Deva': 'पारंपारीक', 'ks': 'رِوٲجی ﴿ترجمع اِشارٕ: یِم ورژن رَسم الخط ہُک ناؤ چھُ چیٖنی باپتھ زَبانٕ ناؤ کِس مجموعَس سٕتؠ اِستعمال یِوان کرنٕہ۔﴾', 'ks-Arab': 'رِوٲجی ﴿ترجمع اِشارٕ: یِم ورژن رَسم الخط ہُک ناؤ چھُ چیٖنی باپتھ زَبانٕ ناؤ کِس مجموعَس سٕتؠ اِستعمال یِوان کرنٕہ۔﴾', 'ks-Deva': 'रिवायाती (तरजुम इशार: स्क्रिप्ट नवुक यि वर्ज़न छु चीनी बापथ ज़बान नाव किस मुरकब कि इस्तिमल करान।)', 'ksh': 'tradizjonäll', 'ku': 'kevneşopî', 'kxv': 'hirudlu', 'kxv-Deva': 'हिरूदोल्लु', 'kxv-Latn': 'hirudlu', 'kxv-Orya': 'ହିରୁଦଲୁ', 'kxv-Telu': 'సాంప్రదాయక', 'ky': 'Салттуу', 'lb': 'Traditionell', 'lld': 'tradizional', 'lo': 'ແບບດັ້ງເດີມ', 'lrc': 'سونأتی', 'lt': 'tradiciniai', 'lv': 'tradicionālā', 'mai': 'पारंपरिक', 'mi': 'Tuku iho', 'mk': 'традиционално', 'ml': 'പരമ്പരാഗതം', 'mn': 'уламжлалт', 'mni': 'ত্রেদিস্নেল', 'mni-Beng': 'ত্রেদিস্নেল', 'mr': 'पारंपारिक', 'ms': 'Tradisional', 'mt': 'Tradizzjonali', 'my': 'ရိုးရာ', 'mzn': 'سنتی هانت', 'nb': 'tradisjonell', 'ne': 'परम्परागत', 'nl': 'traditioneel', 'nn': 'tradisjonell', 'no': 'tradisjonell', 'nqo': 'ߦߋߢߐ߲߯ߠߊ', 'om': 'Kan Durii', 'or': 'ପାରମ୍ପରିକ', 'os': 'Традицион китайаг', 'pa': 'ਰਵਾਇਤੀ', 'pa-Guru': 'ਰਵਾਇਤੀ', 'pcm': 'Nọ́mal', 'pl': 'tradycyjne', 'ps': 'دودیزه', 'pt': 'tradicional', 'qu': 'Tradicional', 'rm': 'tradiziunal', 'ro': 'tradițională', 'ru': 'традиционная', 'sa': 'परम्परागत', 'sat': 'ᱴᱨᱮᱰᱤᱥᱚᱱᱟᱞ', 'sat-Olck': 'ᱴᱨᱮᱰᱤᱥᱚᱱᱟᱞ', 'sc': 'traditzionale', 'sd': 'روايتي', 'sd-Arab': 'روايتي', 'sd-Deva': 'रवायती (तरजुमे जो द॒स : लिखत जे नाले जे हिन बयानु खे चीनीअ लाए भाषा जे नाले सां गद॒ मिलाए करे इस्तेमाल कयो वेंदो आहे)', 'se': 'árbevirolaš', 'si': 'සාම්ප්\u200dරදායික', 'sk': 'tradičné', 'sl': 'tradicionalna pisava', 'so': 'Hore', 'sq': 'tradicional', 'sr': 'традиционално кинеско писмо', 'sr-Cyrl': 'традиционално кинеско писмо', 'sr-Latn': 'tradicionalno kinesko pismo', 'su': 'Tradisional', 'su-Latn': 'Tradisional', 'sv': 'traditionell', 'sw': 'Cha jadi', 'ta': 'பாரம்பரியம்', 'te': 'సాంప్రదాయక', 'tg': 'Анъанавӣ', 'th': 'ตัวเต็ม', 'ti': 'ባህላዊ', 'tk': 'Adaty', 'to': 'tukufakaholo', 'tpi': 'Han (tredisinol)', 'tr': 'Geleneksel', 'tt': 'традицион', 'ug': 'مۇرەككەپ خەنچە', 'uk': 'традиційна', 'ur': 'روایتی', 'uz': 'anʼanaviy', 'uz-Cyrl': 'Анъанавий', 'uz-Latn': 'anʼanaviy', 'vec': 'tradisionale', 'vi': 'Phồn thể', 'wae': 'Traditionell', 'wo': 'Cosaan', 'xh': 'IsiHant Esiqhelekileyo', 'xnr': 'पारंपरिक', 'yo': 'Àbáláyé', 'yrl': 'katuwa', 'yue': '繁體', 'yue-Hans': '繁体', 'yue-Hant': '繁體', 'zh': '繁体', 'zh-Hans': '繁体', 'zh-Hant': '繁體', 'zu': 'okosiko'},
    'Hatr': {'ast': 'hatranu', 'bal-Latn': 'Hatran', 'bs': 'hatran pismo', 'bs-Cyrl': 'хатран писмо', 'bs-Latn': 'hatran pismo', 'en': 'Hatran', 'et': 'Hatra', 'eu': 'hatreoera', 'ff-Adlm': '𞤖𞤢𞤼𞤢𞤪𞤢𞤲', 'fi': 'hatralainen', 'fo': 'hatran', 'ga': 'Hatran', 'gd': 'Hatran', 'he': 'חטרן', 'hi-Latn': 'Hatran', 'id': 'Hatran', 'ig': 'Hatran', 'kk': 'хатра жазуы', 'kk-Cyrl': 'хатра жазуы', 'mn': 'Хатран', 'ms': 'Hatran', 'nb': 'hatransk armensk', 'nl': 'Hatran', 'nn': 'hatransk armensk', 'no': 'hatransk armensk', 'nqo': 'ߤߊߕߙߋߣߌ', 'ro': 'hatrană', 'sc': 'hatran', 'so': 'Qoraalka Hatran', 'sq': 'hatranisht', 'sv': 'hatran', 'tg': 'Хатран', 'vec': 'hatran', 'zh': '哈特兰文', 'zh-Hans': '哈特兰文', 'zh-Hant': '哈特拉文', 'zu': 'isi-Hatran'},
    'Hebr': {'af': 'Hebreeus', 'ak': 'Hibri kasa', 'am': 'እብራይስጥ', 'ar': 'العبرية', 'as': 'হিব্ৰু', 'ast': 'hebréu', 'az': 'ibrani', 'az-Arab': 'عبریجه', 'az-Latn': 'ibrani', 'bal-Latn': 'Ebráni', 'be': 'яўрэйскае', 'be-tarask': 'яўрэйскае', 'bg': 'иврит', 'bgn': 'عبرانی', 'bn': 'হিব্রু', 'br': 'hebraek', 'brx': 'हिब्रिऊ', 'bs': 'hebrejsko pismo', 'bs-Cyrl': 'хебрејско писмо', 'bs-Latn': 'hebrejsko pismo', 'ca': 'hebreu', 'ca-ES-valencia': 'hebreu', 'ccp': '𑄦𑄨𑄛𑄴𑄝𑄳𑄢𑄪', 'ce': 'жугтийн', 'chr': 'ᎠᏂᏈᎵ', 'ckb': 'عیبری', 'cs': 'hebrejské', 'cy': 'Hebreig', 'da': 'hebraisk', 'de': 'Hebräisch', 'dsb': 'hebrejski', 'dz': 'ཧེ་བྲུ་ཡིག་གུ', 'ee': 'hebrigbeŋɔŋlɔ', 'el': 'Εβραϊκό', 'el-polyton': 'Ἑβραϊκό', 'en': 'Hebrew', 'en-Dsrt': '𐐐𐐨𐐺𐑉𐐭', 'es': 'hebreo', 'et': 'heebrea', 'eu': 'hebrearra', 'fa': 'عبری', 'ff-Adlm': '𞤖𞤢𞤦𞤵𞤪𞤢', 'fi': 'heprealainen', 'fil': 'Hebrew', 'fo': 'hebraiskt', 'fr': 'hébreu', 'fur': 'ebreu', 'fy': 'Hebreeuwsk', 'ga': 'Eabhrach', 'gd': 'Eabhra', 'gl': 'hebreo', 'gsw': 'Hebräisch', 'gu': 'હીબ્રુ', 'ha': 'Ibrananci', 'he': 'עברי', 'hi': 'हिब्रू', 'hi-Latn': 'Hebrew', 'hr': 'hebrejsko pismo', 'hsb': 'hebrejsce', 'ht': 'hébreu', 'hu': 'Héber', 'hy': 'եբրայական', 'ia': 'hebraic', 'id': 'Ibrani', 'ig': 'Mkpụrụ ọkwụ Hebrew', 'is': 'hebreskt', 'it': 'ebraico', 'ja': 'ヘブライ文字', 'jv': 'Ibrani', 'ka': 'ებრაული', 'kea': 'ebraiku', 'kgp': 'hemraiko', 'kk': 'иврит жазуы', 'kk-Cyrl': 'иврит жазуы', 'km': 'អ៊ីស្រាអែល', 'kn': 'ಹೀಬ್ರೂ', 'ko': '히브리 문자', 'kok': 'हिब्रू', 'kok-Deva': 'हिब्रू', 'ks': 'ہِبرِو', 'ks-Arab': 'ہِبرِو', 'ksh': 'hebrääjesche Schreff', 'ku': 'îbranî', 'ky': 'Иврит', 'lb': 'Hebräesch', 'lo': 'ຮີບຣິວ', 'lrc': 'عئبری', 'lt': 'hebrajų', 'lv': 'ivrits', 'mai': 'हीब्रू', 'mi': 'Hīperu', 'mk': 'хебрејско писмо', 'ml': 'ഹീബ്രു', 'mn': 'еврей', 'mr': 'हिब्रू', 'ms': 'Ibrani', 'my': 'ဟီဗရူး', 'mzn': 'عبری', 'nb': 'hebraisk', 'ne': 'हिब्रु', 'nl': 'Hebreeuws', 'nn': 'hebraisk', 'no': 'hebraisk', 'nqo': 'ߋߓߙߋ', 'or': 'ହିବୃ', 'pa': 'ਹਿਬਰੂ', 'pa-Guru': 'ਹਿਬਰੂ', 'pcm': 'Híbrú', 'pl': 'hebrajskie', 'ps': 'عبراني', 'pt': 'hebraico', 'qu': 'Hebreo Simi', 'rm': 'ebraic', 'ro': 'ebraică', 'ru': 'еврейская', 'sc': 'ebràicu', 'sd': 'عبراني', 'sd-Arab': 'عبراني', 'si': 'හීබෲ', 'sk': 'hebrejské', 'sl': 'hebrejski', 'so': 'Cibraani', 'sq': 'hebraik', 'sr': 'хебрејско писмо', 'sr-Cyrl': 'хебрејско писмо', 'sr-Latn': 'hebrejsko pismo', 'sv': 'hebreiska', 'sw': 'Kiebrania', 'syr': 'ܥܒܪܝܬ', 'ta': 'ஹீப்ரு', 'te': 'హీబ్రూ', 'tg': 'Яҳудӣ', 'th': 'ฮีบรู', 'ti': 'ኢብራይስጥ', 'tk': 'Ýewreý elipbiýi', 'to': 'tohinima fakahepelū', 'tr': 'İbrani', 'ug': 'ئىبرانىچە', 'uk': 'іврит', 'ur': 'عبرانی', 'uz': 'ivrit', 'uz-Cyrl': 'Иброний', 'uz-Latn': 'ivrit', 'vec': 'ebràego', 'vi': 'Chữ Do Thái', 'wae': 'Hebräiš', 'xnr': 'हिब्रू', 'yi': 'העברעיש', 'yo': 'Hébérù', 'yrl': 'heburaiku', 'yue': '希伯來文', 'yue-Hans': '希伯来文', 'yue-Hant': '希伯來文', 'zh': '希伯来文', 'zh-Hans': '希伯来文', 'zh-Hant': '希伯來文', 'zu': 'isi-Hebrew'},
    'Hira': {'af': 'Hiragana', 'ak': 'Hiragana kasa', 'am': 'ሂራጋና', 'ar': 'الهيراجانا', 'as': 'হিৰাগানা', 'ast': 'ḥiragana', 'az': 'iraqana', 'az-Arab': 'هیراگانا', 'az-Latn': 'iraqana', 'bal-Latn': 'Hirágáná', 'be': 'хірагана', 'be-tarask': 'хірагана', 'bg': 'хирагана', 'bn': 'হিরাগানা', 'br': 'hiragana', 'brx': 'हीरागाना', 'bs': 'pismo hiragana', 'bs-Cyrl': 'хирагана', 'bs-Latn': 'pismo hiragana', 'ca': 'hiragana', 'ca-ES-valencia': 'hiragana', 'ccp': '𑄦𑄨𑄢𑄉𑄚', 'ce': 'хирагана', 'chr': 'ᎯᎳᎦᎾ', 'ckb': 'ھیراگانا', 'cs': 'hiragana', 'cy': 'Hiragana', 'da': 'hiragana', 'de': 'Hiragana', 'dsb': 'hiragana', 'dz': 'ཇ་པཱན་གྱི་ཧི་ར་ག་ན་ཡིག་གུ', 'ee': 'hiraganagbeŋɔŋlɔ', 'el': 'Χιραγκάνα', 'el-polyton': 'Χιραγκάνα', 'en': 'Hiragana', 'en-Dsrt': '𐐐𐐮𐑉𐐲𐑀𐐪𐑌𐐲', 'es': 'hiragana', 'et': 'hiragana', 'eu': 'hiragana', 'fa': 'هیراگانا', 'ff-Adlm': '𞤖𞤭𞤪𞤢𞤺𞤢𞤲𞤢', 'fi': 'hiragana', 'fil': 'Hiragana', 'fo': 'hiragana', 'fr': 'hiragana', 'fy': 'Hiragana', 'ga': 'Hireagánach', 'gd': 'Hiragana', 'gl': 'hiragana', 'gsw': 'Hiragana', 'gu': 'હિરાગાના', 'ha': 'Tsarin Rubutun Hiragana', 'he': 'הירגאנה', 'hi': 'हिरागाना', 'hi-Latn': 'Hiragana', 'hr': 'hiragana pismo', 'hsb': 'hiragana', 'ht': 'hiragana', 'hu': 'Hiragana', 'hy': 'հիրագանա', 'ia': 'hiragana', 'id': 'Hiragana', 'ig': 'Mkpụrụ okwụ Hịragana', 'is': 'hiragana', 'it': 'hiragana', 'ja': 'ひらがな', 'jv': 'Hiragana', 'ka': 'ჰირაგანა', 'kea': 'iragana', 'kgp': 'hiragỹnỹ', 'kk': 'хирагана жазуы', 'kk-Cyrl': 'хирагана жазуы', 'km': 'ហ៊ីរ៉ាកាណា', 'kn': 'ಹಿರಾಗನ', 'ko': '히라가나', 'kok': 'हिरागना', 'kok-Deva': 'हिरागना', 'ks': 'ہیٖراگانا', 'ks-Arab': 'ہیٖراگانا', 'ksh': 'japaanesche Hiddajaana-Schreff', 'ku': 'hîraganayî', 'ky': 'Хирагана', 'lb': 'Hiragana', 'lo': 'ຣິຣະງະນະ', 'lrc': 'ھیراگانا', 'lt': 'hiragana', 'lv': 'hiragana', 'mai': 'हीरागना', 'mi': 'Hirakana', 'mk': 'хирагана', 'ml': 'ഹിരഗാന', 'mn': 'хирагана', 'mr': 'हिरागाना', 'ms': 'Hiragana', 'my': 'ဟီရဂန', 'mzn': 'هیراگانا', 'nb': 'hiragana', 'ne': 'हिरागना', 'nl': 'Hiragana', 'nn': 'hiragana', 'no': 'hiragana', 'nqo': 'ߤߌߙߊߜ߭ߣߊ', 'or': 'ହିରାଗାନା', 'pa': 'ਹਿਰਾਗਾਨਾ', 'pa-Guru': 'ਹਿਰਾਗਾਨਾ', 'pcm': 'Hiagána', 'pl': 'hiragana', 'ps': 'هیراګانا', 'pt': 'hiragana', 'qu': 'Hiragana', 'rm': 'hiragana', 'ro': 'hiragana', 'ru': 'хирагана', 'sc': 'hiragana', 'sd': 'هراگنا', 'sd-Arab': 'هراگنا', 'se': 'hiragana', 'si': 'හිරඟනා', 'sk': 'hiragana', 'sl': 'hiragana', 'so': 'Hiragana', 'sq': 'hiragan', 'sr': 'хирагана', 'sr-Cyrl': 'хирагана', 'sr-Latn': 'hiragana', 'sv': 'hiragana', 'sw': 'Hiragana', 'ta': 'ஹிராகானா', 'te': 'హిరాగాన', 'tg': 'Хирагана', 'th': 'ฮิระงะนะ', 'ti': 'ሂራጋና', 'tk': 'Hiragana elipbiýi', 'to': 'tohinima fakasiapani-hilakana', 'tr': 'Hiragana', 'ug': 'خىراگانا', 'uk': 'хірагана', 'ur': 'ہیراگینا', 'uz': 'hiragana', 'uz-Cyrl': 'Хирагана', 'uz-Latn': 'hiragana', 'vec': 'hiragana', 'vi': 'Chữ Hiragana', 'xnr': 'हिरागाना', 'yo': 'Hiragánà', 'yrl': 'hiragana', 'yue': '平假名', 'yue-Hans': '平假名', 'yue-Hant': '平假名', 'zh': '平假名', 'zh-Hans': '平假名', 'zh-Hant': '平假名', 'zu': 'isi-Hiragana'},
    'Hluw': {'ast': 'xeroglíficos anatolios', 'bal-Latn': 'Anátólyái Hérógleps', 'br': 'hieroglifoù Anatolia', 'bs': 'anatolijski hijeroglifi', 'bs-Cyrl': 'анатолијски хијероглифи', 'bs-Latn': 'anatolijski hijeroglifi', 'ca': 'jeroglífic anatoli', 'ca-ES-valencia': 'jeroglífic anatoli', 'cs': 'anatolské hieroglyfy', 'da': 'anatolske hieroglyffer', 'de': 'Hieroglyphen-Luwisch', 'en': 'Anatolian Hieroglyphs', 'et': 'Anatoolia hieroglüüfkiri', 'eu': 'hieroglifiko anatoliarrak', 'fa': 'هیروگلیف آناتولی', 'ff-Adlm': '𞤖𞤭𞤪𞤮𞤺𞤭𞤪𞤬𞤵 𞤀𞤲𞤢𞤼𞤮𞤤𞤭𞤴𞤢', 'fi': 'anatolialaiset hieroglyfit', 'fo': 'anatoliskar hieroglyffur', 'fy': 'Anatolyske hiërogliefen', 'ga': 'Iairiglifí Anatólacha', 'gd': 'Dealbh-sgrìobhadh Anatolach', 'he': 'הירוגליפים האנטוליים', 'hi-Latn': 'Anatolian Hieroglyphs', 'hr': 'anatolijski hijeroglifi', 'hu': 'Anatóliai hieroglifák', 'id': 'Hieroglif Anatolia', 'ig': 'Anatolian Hieroglyphs', 'it': 'geroglifici anatolici', 'ja': 'アナトリア象形文字', 'ka': 'ანატოლიური იეროგლიფური', 'kk': 'Анадолы иероглифтері', 'kk-Cyrl': 'Анадолы иероглифтері', 'ko': '아나톨리아 상형문자', 'lo': 'ອັກລຮະອານາໂຕເລຍ', 'lt': 'Anatolijaus hieroglifai', 'mk': 'анадолски хиероглифи', 'mn': 'Анатолийн дүрс бичиг', 'ms': 'Hiroglif Anatoli', 'nb': 'anatoliske hieroglyfer', 'nl': 'Anatolische hiërogliefen', 'nn': 'anatoliske hieroglyfer', 'no': 'anatoliske hieroglyfer', 'nqo': 'ߤߕߌ-ߤߙߏߜ߭ߟߌߝߌ', 'ro': 'hieroglife anatoliene', 'ru': 'лувийские иероглифы', 'sc': 'geroglìficos anatòlicos', 'so': 'Qoraalka Anatolian Hieroglyphs', 'sq': 'hieroglife anatoliane', 'sv': 'hittitiska hieroglyfer', 'tg': 'Иероглифҳои Анатолӣ', 'th': 'อักขระอานาโตเลีย', 'to': 'tohinima tongitapu-fakaʻanatolia', 'tr': 'Anadolu Hiyeroglifleri', 'ug': 'ئاناتولىيە تەسۋىرىي يېزىق', 'vec': 'jeroglìfeghi anatòleghi', 'vi': 'Chữ tượng hình Anatolia', 'yue': '安那托利亞象形文字', 'yue-Hans': '安那托利亚象形文字', 'yue-Hant': '安那托利亞象形文字', 'zh': '安那托利亚象形文字', 'zh-Hans': '安那托利亚象形文字', 'zh-Hant': '安那托利亞象形文字', 'zu': 'isi-Anatolian Hieroglyphs'},
    'Hmng': {'ar': 'الباهوه همونج', 'ast': 'pahawh hmong', 'az': 'pahav monq', 'az-Latn': 'pahav monq', 'bal-Latn': 'Pahaw Hmong', 'bg': 'Пахау хмонг', 'bn': 'ফাহাও মঙ', 'brx': 'पाहवाह ह्मौंग', 'bs': 'pahawh hmong pismo', 'bs-Cyrl': 'пахав хмонг писмо', 'bs-Latn': 'pahawh hmong pismo', 'ca': 'pahawh hmong', 'ca-ES-valencia': 'pahawh hmong', 'ccp': '𑄜𑄦𑄃𑄮𑄟𑄧𑄋𑄴', 'cs': 'hmongské', 'da': 'pahawh hmong', 'de': 'Pahawh Hmong', 'el': 'Παχάχ Χμονγκ', 'el-polyton': 'Παχάχ Χμονγκ', 'en': 'Pahawh Hmong', 'es': 'pahawh hmong', 'et': 'phahau-hmongi kiri', 'eu': 'pahawh hmongera', 'ff-Adlm': '𞤆𞤢𞤸𞤢𞤱 𞤖𞤢𞤥𞤮𞤲𞤺', 'fi': 'pahawh hmong', 'fo': 'pahawh hmong', 'fr': 'pahawh hmong', 'fy': 'Pahawh Hmong', 'ga': 'Pahawh Hmong', 'gd': 'Pahawh Hmong', 'gsw': 'Pahawh Hmong', 'gu': 'પહાઉ મોન્ગ', 'he': 'פאהח המונג', 'hi': 'पाहो ह्मोन्ग', 'hi-Latn': 'Pahawh Hmong', 'hr': 'pahawh hmong pismo', 'ht': 'pahawh hmong', 'hu': 'Pahawh hmong', 'id': 'Pahawh Hmong', 'ig': 'Pahawh Hmong', 'it': 'pahawn hmong', 'ja': 'パハウ・フモン文字', 'ka': 'ფაჰაუ-მონი', 'kgp': 'pahav homãg', 'kk': 'пахау жазуы', 'kk-Cyrl': 'пахау жазуы', 'kn': 'ಪಹವ್ ಹ್ಮೋಂಗ್', 'ko': '파하우 몽 문자', 'ks': 'پَہاو مانگ', 'ks-Arab': 'پَہاو مانگ', 'lb': 'Pahawh Hmong', 'lo': 'ປາເຮາເມັງ', 'lt': 'pahav hmong', 'mk': 'пахауанско хмоншко', 'ml': 'പഹ്വാ ഹമോംഗ്', 'mn': 'Пахав Хмонг', 'mr': 'पहाउ मंग', 'ms': 'Pahawh Hmong', 'nb': 'pahawh hmong', 'ne': 'पहावह हमोङ्ग', 'nl': 'Pahawh Hmong', 'nn': 'pahawk hmong', 'no': 'pahawh hmong', 'nqo': 'ߔߤߊߥ ߡߐ߲ߜ߭', 'or': 'ପାହୋ ହୋଙ୍ଗ', 'pl': 'pahawh hmong', 'pt': 'pahawh hmong', 'rm': 'pahawn hmong', 'ro': 'pahawh hmong', 'ru': 'пахау хмонг', 'sc': 'pahawn hmong', 'sl': 'pahavhmonska zlogovna pisava', 'so': 'Hmonga pahawh', 'sq': 'pahauhmonisht', 'sr': 'пахав хмонг писмо', 'sr-Cyrl': 'пахав хмонг писмо', 'sr-Latn': 'pahav hmong pismo', 'sv': 'pahaw mong', 'ta': 'பஹாவ் மாங்க்', 'te': 'పాహవా హ్మోంగ్', 'tg': 'Пахах Хмонг', 'th': 'ปาเฮาห์ม้ง', 'to': 'tohinima fakapahaumongi', 'tr': 'Pahavh Hmong', 'ug': 'مۆڭچە', 'uk': 'пахау хмонг', 'vec': 'pahaw hmong', 'vi': 'Chữ Pahawh Hmong', 'xnr': 'पाहो ह्मोन्ग', 'yrl': 'parau himũgi', 'yue': '楊松錄苗文', 'yue-Hans': '杨松录苗文', 'yue-Hant': '楊松錄苗文', 'zh': '杨松录苗文', 'zh-Hans': '杨松录苗文', 'zh-Hant': '楊松錄苗文', 'zu': 'isi-Pahawh Hmong'},
    'Hmnp': {'bal-Latn': 'Niákeng Puáche Hamung', 'bs': 'nijakeng puaču hmong pismo', 'bs-Cyrl': 'нјакенг пауче хмонг писмо', 'bs-Latn': 'nijakeng puaču hmong pismo', 'cs': 'nyiakeng puachue hmong', 'en': 'Nyiakeng Puachue Hmong', 'eu': 'nyiakeng puachue hmong', 'ff-Adlm': '𞤙𞤭𞤢𞤳𞤫𞤲𞤺 𞤆𞤵𞤢𞤧𞤵𞥅 𞤖𞤥𞤮𞤲𞤺', 'fi': 'hmongin nyiakeng puachue', 'fo': 'nyiakeng puachue hmong', 'fr': 'nyiakeng puachue hmong', 'gd': 'Nyiakeng Puachue Hmong', 'he': 'ניאקנג פואצ׳ה המונג', 'hi-Latn': 'Nyiakeng Puachue Hmong', 'ht': 'nyiakeng puachue hmong', 'id': 'Nyiakeng Puachue Hmong', 'ig': 'Nyiakeng Puachue Hmong', 'kk': 'ньиакенг пуачуэ хмонг жазуы', 'kk-Cyrl': 'ньиакенг пуачуэ хмонг жазуы', 'km': 'នីយ៉ាកេងពួជឺម៉ុង', 'mn': 'Хмнп', 'ms': 'Nyiakeng Puachue Hmong', 'nl': 'Nyiakeng Puachue Hmong', 'nqo': 'ߢߊߞߋ߲ߜ߭ ߔߎߛߑߡߐ߲ߜ߭', 'ro': 'nyiakeng puachue hmong', 'sc': 'nyiakeng puachue hmong', 'so': 'Hmonga Nyiakeng Puachue', 'sq': 'niakeng puaçue hmong', 'sv': 'nyiakeng puachue hmong', 'tg': 'Някенг Пуачэ Хмонг', 'vec': 'niakeng puachu hmong', 'yo': 'Nyiakengi Puase Himongi', 'zh': '尼亚肯蒲丘苗文', 'zh-Hans': '尼亚肯蒲丘苗文', 'zu': 'i-Nyiakeng Puachue Hmong'},
    'Hrkt': {'af': 'Japannese lettergreepskrif', 'ak': 'Gyapanfoɔ selabolo kasa', 'am': 'ጃፓንኛ ፊደላት', 'ar': 'أبجدية مقطعية يابانية', 'as': 'জাপানী ছিলেবেৰিজ', 'ast': 'silabarios xaponeses', 'az': 'hecalı yapon əlifbası', 'az-Latn': 'hecalı yapon əlifbası', 'bal-Latn': 'Jápáni Selibari', 'be': 'японскія складовыя пісьмы', 'be-tarask': 'японскія складовыя пісьмы', 'bg': 'японска сричкова', 'bn': 'জাপানি অক্ষরমালা', 'br': 'silabennaouegoù japanek', 'brx': 'जापानीज रिंसारथिफोर/हांखोफोर', 'bs': 'katakana ili hiragana', 'bs-Cyrl': 'јапанско слоговно писмо', 'bs-Latn': 'katakana ili hiragana', 'ca': 'sil·labaris japonesos', 'ca-ES-valencia': 'sil·labaris japonesos', 'ccp': '𑄎𑄛𑄚𑄨 𑄦𑄧𑄢𑄧𑄇𑄴', 'ce': 'катакана я хирагана', 'chr': 'ᏣᏩᏂᏏ ᏧᏃᏴᎩ', 'ckb': 'ژاپۆنیی بڕگەیی', 'cs': 'japonské slabičné', 'cy': 'Syllwyddor Japaneaidd', 'da': 'japanske skrifttegn', 'de': 'Japanische Silbenschrift', 'dsb': 'japańske złožkowe pismo', 'el': 'Κατακάνα ή Χιραγκάνα', 'el-polyton': 'Κατακάνα ή Χιραγκάνα', 'en': 'Japanese syllabaries', 'en-Dsrt': '𐐗𐐪𐐻𐐲𐐿𐐪𐑌𐐲 𐐬𐑉 𐐐𐐮𐑉𐐲𐑀𐐪𐑌𐐲', 'es': 'silabarios japoneses', 'et': 'jaapani silpkirjad', 'eu': 'silabario japoniarrak', 'fa': 'هجانگاری ژاپنی', 'ff-Adlm': '𞤅𞤭𞤺𞤢𞤲𞤯𞤫 𞤐𞤭𞤨𞤮𞤲𞤶𞤫', 'fi': 'japanin tavumerkistöt', 'fil': 'Japanese syllabaries', 'fo': 'japanskir stavir', 'fr': 'katakana ou hiragana', 'fur': 'katakana o hiragana', 'fy': 'Katakana of Hiragana', 'ga': 'Siollabraí Seapánacha', 'gd': 'Katakana no Hiragana', 'gl': 'silabarios xaponeses', 'gsw': 'Katakana oder Hiragana', 'gu': 'જાપાનીઝ વર્ણમાળા', 'ha': 'kalaman Jafananci', 'he': 'הברתי יפני', 'hi': 'जापानी सिलेबरीज़', 'hi-Latn': 'Japanese syllabaries', 'hr': 'japansko slogovno pismo', 'hsb': 'japanske złóžkowe pismo', 'ht': 'katakana ou hiragana', 'hu': 'Katakana vagy hiragana', 'hy': 'ճապոնական վանկագիր', 'ia': 'syllabarios japonese', 'id': 'Katakana atau Hiragana', 'ig': 'mkpụrụ ọkwụ Japanịsị', 'is': 'japönsk samstöfuletur', 'it': 'katanaka o hiragana', 'ja': '仮名', 'jv': 'Silabaris Jepang', 'ka': 'იაპონური კანა', 'kea': 'silabáriu japunes', 'kgp': 'sirama-vẽnhrá japunẽj', 'kk': 'хирагана немесе катакана', 'kk-Cyrl': 'хирагана немесе катакана', 'km': 'សញ្ញាសំឡេងភាសាជប៉ុន', 'kn': 'ಜಪಾನೀಸ್ ಸಿಲಬರೀಸ್', 'ko': '가나', 'kok': 'जपानी अक्षरमाळ', 'kok-Deva': 'जपानी अक्षरमाळ', 'ks': 'کَٹاکانا یا ہِراگانا', 'ks-Arab': 'کَٹاکانا یا ہِراگانا', 'ku': 'nivîsên heceyî yên japonî', 'ky': 'Жапон силлабографиясы', 'lb': 'Katakana oder Hiragana', 'lo': 'ຕາຕາລາງພະຍາງພາສາຍີ່ປຸ່ນ', 'lt': 'katakana / hiragana', 'lv': 'japāņu zilbju alfabēts', 'mai': 'जापानी शब्दांश', 'mi': 'Kūoro Hapanihi', 'mk': 'јапонско слоговно', 'ml': 'ജാപ്പനീസ് സില്ലബറീസ്', 'mn': 'япон хэлний үеийн цагаан толгой', 'mr': 'जापानी स्वरलिपी', 'ms': 'Ejaan sukuan Jepun', 'my': 'ဂျပန် အက္ခရာ', 'nb': 'japanske stavelsesskrifter', 'ne': 'काताकाना वा हिरागाना', 'nl': 'Katakana of Hiragana', 'nn': 'japanske stavingsskrifter', 'no': 'japanske stavelsesskrifter', 'nqo': 'ߞߊߕߊ߲ߞߣߊ', 'or': 'ଜାପାନୀ ସିଲାବରିଜ୍\u200c', 'pa': 'ਜਾਪਾਨੀ ਸਿਲੇਬਰੀਜ਼', 'pa-Guru': 'ਜਾਪਾਨੀ ਸਿਲੇਬਰੀਜ਼', 'pcm': 'Pát ọf Japán Raítín Sístẹm', 'pl': 'sylabariusze japońskie', 'ps': 'د جاپاني سیلابري', 'pt': 'silabários japoneses', 'qu': 'Japones silabico sananpakuna', 'rm': 'scrittira da silbas giapunaisa', 'ro': 'silabică japoneză', 'ru': 'катакана и хирагана', 'sc': 'sillabàrios giaponesos', 'sd': 'جاپاني لکت', 'sd-Arab': 'جاپاني لکت', 'si': 'ජෑපනීස් සිලබරීස්', 'sk': 'kana', 'sl': 'japonska zlogovnica', 'so': 'Qoraalka Xuruufta Jabaaniiska', 'sq': 'alfabet rrokjesor japonez', 'sr': 'јапанска слоговна писма', 'sr-Cyrl': 'јапанска слоговна писма', 'sr-Latn': 'japanska slogovna pisma', 'sv': 'katakana/hiragana', 'sw': 'Hati za Kijapani', 'ta': 'ஜப்பானிய எழுத்துருக்கள்', 'te': 'జపనీస్ సిలబెరీస్', 'tg': 'Ҳиҷоҳои ҷопонӣ', 'th': 'คะตะกะนะหรือฮิระงะนะ', 'ti': 'ጃፓናዊ ሲለባሪታት', 'tk': 'Ýapon bogun elipbiýleri', 'to': 'tohinima fakasilapa-siapani', 'tr': 'Japon hece alfabeleri', 'ug': 'ياپونچە خىراگانا ياكى كاتاكانا', 'uk': 'японські силабарії', 'ur': 'جاپانی سیلابریز', 'uz': 'katakana yoki hiragana', 'uz-Latn': 'katakana yoki hiragana', 'vec': 'silabari japonezi', 'vi': 'Bảng ký hiệu âm tiết Tiếng Nhật', 'xnr': 'जापानी सिलेबरीज़', 'yo': 'ìlànà àfọwọ́kọ ará Jàpánù', 'yrl': 'yapunei síraba irũ', 'yue': '片假名或平假名', 'yue-Hans': '片假名或平假名', 'yue-Hant': '片假名或平假名', 'zh': '假名', 'zh-Hans': '假名', 'zh-Hant': '片假名或平假名', 'zu': 'i-Japanese syllabaries'},
    'Hung': {'ar': 'المجرية القديمة', 'ast': 'húngaru antiguu', 'az': 'qədimi macar', 'az-Latn': 'qədimi macar', 'bal-Latn': 'Ahdi Hangari', 'bg': 'Староунгарска', 'bn': 'পুরোনো হাঙ্গেরীয়', 'br': 'henhungarek', 'brx': 'पुरानी हंगैरीयन', 'bs': 'staromađarsko pismo', 'bs-Cyrl': 'старомађарско писмо', 'bs-Latn': 'staromađarsko pismo', 'ca': 'hongarès antic', 'ca-ES-valencia': 'hongarès antic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄴 𑄦𑄋𑄴𑄉𑄬𑄢𑄨𑄠𑄧', 'cs': 'staromaďarské', 'da': 'oldungarsk', 'de': 'Altungarisch', 'el': 'Παλαιό Ουγγρικό', 'el-polyton': 'Παλαιό Ουγγρικό', 'en': 'Old Hungarian', 'en-Dsrt': '𐐄𐑊𐐼 𐐐𐐲𐑍𐐩𐑉𐐨𐐲𐑌', 'es': 'húngaro antiguo', 'et': 'vanaungari', 'eu': 'hungariera zaharra', 'fa': 'مجاری باستان', 'ff-Adlm': '𞤑𞤭𞤯𞥆𞤭 𞤖𞤢𞤲𞤺𞤢𞤪𞤭𞤴𞤢𞥄', 'fi': 'muinaisunkarilainen', 'fo': 'gamalt ungarskt', 'fr': 'ancien hongrois', 'fur': 'vieri ongjarês', 'fy': 'Aldhongaars', 'ga': 'Sean-Ungárach', 'gd': 'Seann-Ungarais', 'gsw': 'Altungarisch', 'gu': 'ઓલ્ડ હંગેરિયન', 'he': 'הונגרי עתיק', 'hi': 'ऑल्ड हंगेरियन', 'hi-Latn': 'Old Hungarian', 'hr': 'staro mađarsko pismo', 'ht': 'ancien hongrois', 'hu': 'Ómagyar', 'id': 'Hungaria Kuno', 'ig': 'Old Hungarian', 'it': 'antico ungherese', 'ja': '古代ハンガリー文字', 'ka': 'ძველი უნგრული', 'kgp': 'hũgaru si', 'kk': 'мажар рунасы', 'kk-Cyrl': 'мажар рунасы', 'kn': 'ಪ್ರಾಚೀನ ಹಂಗೇರಿಯನ್', 'ko': '고대 헝가리 문자', 'ks': 'پرون ہَنگیریَن', 'ks-Arab': 'پرون ہَنگیریَن', 'lb': 'Alungaresch', 'lo': 'ຮັງກາຮີໂບຮານ', 'lt': 'senasis vengrų', 'lv': 'senungāru', 'mk': 'староунгарско', 'ml': 'പുരാതന ഹംഗേറിയൻ', 'mn': 'Хуучин Унгар', 'mr': 'पुरातन हंगेरियन', 'ms': 'Hungary Lama', 'nb': 'gammelungarsk', 'ne': 'पुरानो हङ्गेरियाली', 'nl': 'Oudhongaars', 'nn': 'gammalungarsk', 'no': 'gammelungarsk', 'nqo': 'ߤߐ߲ߜ߭ߙߌ߫ ߞߘߐ߬ߡߊ߲', 'or': 'ପୁରୁଣା ହଙ୍ଗେରିଆନ୍', 'pl': 'starowęgierskie', 'pt': 'húngaro antigo', 'rm': 'ungarais vegl', 'ro': 'maghiară veche', 'ru': 'старовенгерская', 'sc': 'ungheresu antigu', 'sl': 'staroogrski', 'so': 'Hangariyaankii Hore', 'sq': 'hungarishte e vjetër', 'sr': 'старомађарско писмо', 'sr-Cyrl': 'старомађарско писмо', 'sr-Latn': 'staromađarsko pismo', 'sv': 'fornungerska', 'ta': 'பழைய ஹங்கேரியன்', 'te': 'ప్రాచీన హంగేరియన్', 'tg': 'Венгерии Куҳна', 'th': 'ฮังการีโบราณ', 'to': 'tohinima fakahungakalia-motuʻa', 'tr': 'Eski Macar', 'ug': 'قەدىمكى ماجارچە', 'uk': 'давньоугорський', 'vec': 'ongareze antigo', 'vi': 'Chữ Hungary cổ', 'xnr': 'ऑल्ड हंगेरियन', 'yrl': 'ũgaru kuxiímawara', 'yue': '古匈牙利文', 'yue-Hans': '古匈牙利文', 'yue-Hant': '古匈牙利文', 'zh': '古匈牙利文', 'zh-Hans': '古匈牙利文', 'zh-Hant': '古匈牙利文', 'zu': 'isi-Old Hungarian'},
    'IC': {'af': 'Kanariese Eilande', 'ak': 'Canary Islands', 'am': 'የካናሪ ደሴቶች', 'ar': 'جزر الكناري', 'as': 'কেনেৰী দ্বীপপুঞ্জ', 'ast': 'Islles Canaries', 'az': 'Kanar adaları', 'az-Cyrl': 'Канар адалары', 'az-Latn': 'Kanar adaları', 'bal-Latn': 'Kanaray Islánd', 'be': 'Канарскія астравы', 'be-tarask': 'Канарскія астравы', 'bg': 'Канарски острови', 'blo': 'Kanarii kaBʊtukǝltǝna', 'bn': 'ক্যানারি দ্বীপপুঞ্জ', 'br': 'Inizi Kanariez', 'brx': 'कानारि द्वीपफोर', 'bs': 'Kanarska ostrva', 'bs-Cyrl': 'Канарска острва', 'bs-Latn': 'Kanarska ostrva', 'ca': 'Illes Canàries', 'ca-ES-valencia': 'Illes Canàries', 'ccp': '𑄇𑄳𑄠𑄚𑄢𑄨 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Канаран гӀайренаш', 'ceb': 'Canary Islands', 'chr': 'ᏥᏍᏆ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی کەناری', 'cs': 'Kanárské ostrovy', 'cv': 'Канар утравӗсем', 'cy': 'Yr Ynysoedd Dedwydd', 'da': 'Kanariske øer', 'de': 'Kanarische Inseln', 'dsb': 'Kanariske kupy', 'dz': 'ཀ་ནེ་རི་གླིང་ཚོམ', 'ee': 'Kanari ƒudomekpowo nutome', 'el': 'Κανάριοι Νήσοι', 'el-polyton': 'Κανάριοι Νήσοι', 'en': 'Canary Islands', 'es': 'Canarias', 'et': 'Kanaari saared', 'eu': 'Kanariak', 'fa': 'جزایر قناری', 'ff-Adlm': '𞤅𞤵𞤪𞤭𞥅𞤪𞤫-𞤑𞤢𞤲𞤢𞤪𞤭𞥅', 'fi': 'Kanariansaaret', 'fil': 'Canary Islands', 'fo': 'Kanariuoyggjar', 'fr': 'Îles Canaries', 'fur': 'Isulis Canariis', 'fy': 'Kanaryske Eilânnen', 'ga': 'Na hOileáin Chanáracha', 'gd': 'Na h-Eileanan Canàrach', 'gl': 'Illas Canarias', 'gu': 'કૅનેરી આઇલેન્ડ્સ', 'ha': 'Tsibiran Canary', 'he': 'האיים הקנריים', 'hi': 'कैनेरी द्वीपसमूह', 'hi-Latn': 'Canary Islands', 'hr': 'Kanari', 'hsb': 'Kanariske kupy', 'ht': 'Îles Canaries', 'hu': 'Kanári-szigetek', 'hy': 'Կանարյան կղզիներ', 'ia': 'Insulas Canari', 'id': 'Kepulauan Canaria', 'ig': 'Canary Islands', 'is': 'Kanaríeyjar', 'it': 'Isole Canarie', 'ja': 'カナリア諸島', 'jv': 'Kapuloan Kanari', 'ka': 'კანარის კუნძულები', 'kaa': 'Канар атаўлары', 'kaa-Cyrl': 'Канар атаўлары', 'kea': 'Kanárias', 'kgp': 'Kanỹrija Goj-vẽso', 'kk': 'Канар аралдары', 'kk-Cyrl': 'Канар аралдары', 'km': 'កោះ\u200bកាណារី', 'kn': 'ಕ್ಯಾನರಿ ದ್ವೀಪಗಳು', 'ko': '카나리아 제도', 'kok': 'कॅनरी आयलैंड्स', 'kok-Deva': 'कॅनरी आयलैंड्स', 'ks': 'کینری جزیرٕ', 'ks-Arab': 'کینری جزیرٕ', 'ksh': 'de Kannaresche Enselle', 'ku': 'Giravên Kanaryayê', 'kxv': 'kanari dīp', 'kxv-Deva': 'केनेरी द्वीप', 'kxv-Latn': 'kanari dīp', 'kxv-Orya': 'କେନାରି ଦିପ', 'kxv-Telu': 'కేనరీ దీపులు', 'ky': 'Канар аралдары', 'lb': 'Kanaresch Inselen', 'lld': 'Isoles Canaries', 'lo': 'ໝູ່ເກາະຄານາຣີ', 'lt': 'Kanarų salos', 'lv': 'Kanāriju salas', 'mai': 'कैनरी द्वीपसमूह', 'mi': 'Motu Kanēre', 'mk': 'Канарски Острови', 'ml': 'കാനറി ദ്വീപുകൾ', 'mn': 'Канарын арлууд', 'mr': 'कॅनरी बेटे', 'ms': 'Kepulauan Canary', 'mt': 'il-Gżejjer Canary', 'my': 'ကနေရီ ကျွန်းစု', 'mzn': 'قناری جزایر', 'nb': 'Kanariøyene', 'ne': 'क्यानारी टापुहरू', 'nl': 'Canarische Eilanden', 'nn': 'Kanariøyane', 'no': 'Kanariøyene', 'nqo': 'ߞߣߊߙߌ߫ ߕߌ߲', 'om': 'Odoloota Kanaarii', 'or': 'କେନେରୀ ଦ୍ୱୀପପୁଞ୍ଜ', 'pa': 'ਕੇਨਾਰੀ ਟਾਪੂ', 'pa-Guru': 'ਕੇਨਾਰੀ ਟਾਪੂ', 'pcm': 'Kenerí Aílands', 'pl': 'Wyspy Kanaryjskie', 'ps': 'د کناري ټاپوګان', 'pt': 'Ilhas Canárias', 'qu': 'Islas Canarias', 'rm': 'Inslas Canarias', 'ro': 'Insulele Canare', 'ru': 'Канарские о-ва', 'sc': 'Ìsulas Canàrias', 'sd': 'ڪينري ٻيٽ', 'sd-Arab': 'ڪينري ٻيٽ', 'se': 'Kanáriasullot', 'si': 'කැනරි සූපත්', 'sk': 'Kanárske ostrovy', 'sl': 'Kanarski otoki', 'smn': 'Kanariasuolluuh', 'so': 'Jasiiradda Kanari', 'sq': 'Ishujt Kanarie', 'sr': 'Канарска Острва', 'sr-Cyrl': 'Канарска Острва', 'sr-Latn': 'Kanarska Ostrva', 'sv': 'Kanarieöarna', 'sw': 'Visiwa vya Kanari', 'syr': 'ܓܙܝܖ̈ܐ ܕܟܐܢܪܝ', 'ta': 'கேனரி தீவுகள்', 'te': 'కేనరీ దీవులు', 'tg': 'Ҷазираҳои Канария', 'th': 'หมู่เกาะคานารี', 'ti': 'ደሴታት ካናሪ', 'tk': 'Kanar adalary', 'to': 'ʻOtumotu Kaneli', 'tr': 'Kanarya Adaları', 'tt': 'Канар утраулары', 'ug': 'كانارى ئاراللىرى', 'uk': 'Канарські Острови', 'ur': 'کینری آئلینڈز', 'uz': 'Kanar orollari', 'uz-Cyrl': 'Канар ороллари', 'uz-Latn': 'Kanar orollari', 'vai': 'ꗛꖺꔻꕩ ꔳꘋꗣ', 'vai-Vaii': 'ꗛꖺꔻꕩ ꔳꘋꗣ', 'vec': 'Ìzole Canarie', 'vi': 'Quần đảo Canary', 'wae': 'Kanariše Inslä', 'wo': 'Ile Canary', 'xh': 'ECanary Islands', 'xnr': 'कैनेरी द्वीपां दा झुंड', 'yi': 'קאַנאַרישע אינזלען', 'yo': 'Ẹrékùsù Kánárì', 'yrl': 'Kapuãma-ita Kanariya', 'yue': '加那利群島', 'yue-Hans': '加那利群岛', 'yue-Hant': '加那利群島', 'zh': '加纳利群岛', 'zh-Hans': '加纳利群岛', 'zh-Hant': '加那利群島', 'zu': 'i-Canary Islands'},
    'ID': {'af': 'Indonesië', 'agq': 'Èndòneshìa', 'ak': 'Indɔnehyia', 'am': 'ኢንዶኔዢያ', 'ar': 'إندونيسيا', 'as': 'ইণ্ডোনেচিয়া', 'asa': 'Indonethia', 'ast': 'Indonesia', 'az': 'İndoneziya', 'az-Cyrl': 'Индонезија', 'az-Latn': 'İndoneziya', 'bal-Latn': 'Endhonéshiá', 'bas': 'Indònèsià', 'be': 'Інданезія', 'be-tarask': 'Інданезія', 'bez': 'Huindonesia', 'bg': 'Индонезия', 'bgn': 'ایندونیزیا', 'blo': 'Ɛnɖonosii', 'bm': 'Ɛndonezi', 'bn': 'ইন্দোনেশিয়া', 'br': 'Indonezia', 'brx': 'इण्ड’नेशिया', 'bs': 'Indonezija', 'bs-Cyrl': 'Индонезија', 'bs-Latn': 'Indonezija', 'ca': 'Indonèsia', 'ca-ES-valencia': 'Indonèsia', 'ccp': '𑄄𑄚𑄴𑄘𑄮𑄚𑄬𑄥𑄨𑄠', 'ce': 'Индонези', 'ceb': 'Indonesia', 'cgg': 'Indoneeziya', 'chr': 'ᎢᏂᏙᏂᏍᏯ', 'ckb': 'ئیندۆنیزیا', 'cs': 'Indonésie', 'cv': 'Индонези', 'cy': 'Indonesia', 'da': 'Indonesien', 'dav': 'Indonesia', 'de': 'Indonesien', 'dje': 'Indoneezi', 'dsb': 'Indoneziska', 'dyo': 'Endonesi', 'dz': 'ཨིན་ཌོ་ནེ་ཤི་ཡ', 'ebu': 'Indonesia', 'ee': 'Indonesia nutome', 'el': 'Ινδονησία', 'el-polyton': 'Ἰνδονησία', 'en': 'Indonesia', 'en-Dsrt': '𐐆𐑌𐐼𐐲𐑌𐐨𐑈𐐲', 'eo': 'Indonezio', 'es': 'Indonesia', 'et': 'Indoneesia', 'eu': 'Indonesia', 'ewo': 'ɛndonésia', 'fa': 'اندونزی', 'ff': 'Enndonesii', 'ff-Adlm': '𞤋𞤲𞤣𞤮𞤲𞤭𞥅𞤧𞤴𞤢', 'ff-Latn': 'Enndonesii', 'fi': 'Indonesia', 'fil': 'Indonesia', 'fo': 'Indonesia', 'fr': 'Indonésie', 'fur': 'Indonesie', 'fy': 'Yndonesië', 'ga': 'an Indinéis', 'gd': 'Na h-Innd-innse', 'gl': 'Indonesia', 'gsw': 'Indoneesie', 'gu': 'ઇન્ડોનેશિયા', 'guz': 'Indonesia', 'ha': 'Indunusiya', 'he': 'אינדונזיה', 'hi': 'इंडोनेशिया', 'hi-Latn': 'Indonesia', 'hr': 'Indonezija', 'hsb': 'Indoneska', 'ht': 'Indonésie', 'hu': 'Indonézia', 'hy': 'Ինդոնեզիա', 'ia': 'Indonesia', 'id': 'Indonesia', 'ie': 'Indonesia', 'ig': 'Indonesia', 'is': 'Indónesía', 'it': 'Indonesia', 'ja': 'インドネシア', 'jmc': 'Indonesia', 'jv': 'Indonésia', 'ka': 'ინდონეზია', 'kaa': 'Индонезия', 'kaa-Cyrl': 'Индонезия', 'kab': 'Indunizi', 'kam': 'Indonesia', 'kde': 'Indonesia', 'kea': 'Indonézia', 'kgp': 'Ĩnonẽja', 'khq': 'Indoneezi', 'ki': 'Indonesia', 'kk': 'Индонезия', 'kk-Cyrl': 'Индонезия', 'kln': 'Emetab Indonesia', 'km': 'ឥណ្ឌូណេស៊ី', 'kn': 'ಇಂಡೋನೇಶಿಯಾ', 'ko': '인도네시아', 'kok': 'इंडोनेशीया', 'kok-Deva': 'इंडोनेशीया', 'ks': 'انڈونیشیا', 'ks-Arab': 'انڈونیشیا', 'ksb': 'Indonesia', 'ksf': 'indonɛsí', 'ksh': 'Indoneesije', 'ku': 'Endonezya', 'kxv': 'inḍonesiya', 'kxv-Deva': 'इंडोनेसिया', 'kxv-Latn': 'inḍonesiya', 'kxv-Orya': 'ଇଣ୍ଡନେସିୟା', 'kxv-Telu': 'ఇండోనేసియా', 'ky': 'Индонезия', 'lag': 'Indonésia', 'lb': 'Indonesien', 'lg': 'Yindonezya', 'lld': 'Indonesia', 'ln': 'Indonezi', 'lo': 'ອິນໂດເນເຊຍ', 'lt': 'Indonezija', 'lu': 'Indonezi', 'luo': 'Indonesia', 'luy': 'Indonesia', 'lv': 'Indonēzija', 'mai': 'इंडोनेशिया', 'mas': 'Indonesia', 'mer': 'Indonesia', 'mfe': 'Indonezi', 'mg': 'Indonezia', 'mgh': 'Undonesia', 'mi': 'Initonīhia', 'mk': 'Индонезија', 'ml': 'ഇന്തോനേഷ്യ', 'mn': 'Индонез', 'mr': 'इंडोनेशिया', 'ms': 'Indonesia', 'mt': 'l-Indoneżja', 'mua': 'Indonesiya', 'my': 'အင်ဒိုနီးရှား', 'mzn': 'اندونزی', 'naq': 'Indonesiab', 'nb': 'Indonesia', 'nd': 'Indonesiya', 'ne': 'इन्डोनेशिया', 'nl': 'Indonesië', 'nmg': 'Indonesia', 'nn': 'Indonesia', 'no': 'Indonesia', 'nqo': 'ߍ߲ߘߣߏߖ߭ߌ߫', 'nyn': 'Indoneeziya', 'om': 'Indooneeshiyaa', 'or': 'ଇଣ୍ଡୋନେସିଆ', 'pa': 'ਇੰਡੋਨੇਸ਼ੀਆ', 'pa-Guru': 'ਇੰਡੋਨੇਸ਼ੀਆ', 'pcm': 'Indoníshia', 'pl': 'Indonezja', 'ps': 'اندونیزیا', 'pt': 'Indonésia', 'qu': 'Indonesia', 'rm': 'Indonesia', 'rn': 'Indoneziya', 'ro': 'Indonezia', 'rof': 'Indonesia', 'ru': 'Индонезия', 'rwk': 'Indonesia', 'saq': 'Indonesia', 'sbp': 'Indonesiya', 'sc': 'Indonèsia', 'sd': 'انڊونيشيا', 'sd-Arab': 'انڊونيشيا', 'se': 'Indonesia', 'seh': 'Indonésia', 'ses': 'Indoneezi', 'sg': 'Ênndonezïi', 'shi': 'ⴰⵏⴷⵓⵏⵉⵙⵢⴰ', 'shi-Latn': 'andunisya', 'shi-Tfng': 'ⴰⵏⴷⵓⵏⵉⵙⵢⴰ', 'si': 'ඉන්දුනීසියාව', 'sk': 'Indonézia', 'sl': 'Indonezija', 'smn': 'Indonesia', 'sn': 'Indonesia', 'so': 'Indoneesiya', 'sq': 'Indonezi', 'sr': 'Индонезија', 'sr-Cyrl': 'Индонезија', 'sr-Latn': 'Indonezija', 'su': 'Indonesia', 'su-Latn': 'Indonesia', 'sv': 'Indonesien', 'sw': 'Indonesia', 'syr': 'ܐܝܢܕܘܢܝܣܝܐ', 'ta': 'இந்தோனேசியா', 'te': 'ఇండోనేషియా', 'teo': 'Indonesia', 'tg': 'Индонезия', 'th': 'อินโดนีเซีย', 'ti': 'ኢንዶነዥያ', 'tk': 'Indoneziýa', 'to': 'ʻInitonēsia', 'tr': 'Endonezya', 'tt': 'Индонезия', 'twq': 'Indoneezi', 'tzm': 'Indunizya', 'ug': 'ھىندونېزىيە', 'uk': 'Індонезія', 'ur': 'انڈونیشیا', 'uz': 'Indoneziya', 'uz-Cyrl': 'Индонезия', 'uz-Latn': 'Indoneziya', 'vai': 'ꔤꖆꕇꔻꕩ', 'vai-Latn': 'Índonisiya', 'vai-Vaii': 'ꔤꖆꕇꔻꕩ', 'vec': 'Indonezia', 'vi': 'Indonesia', 'vun': 'Indonesia', 'wae': 'Indonesie', 'wo': 'Indonesi', 'xh': 'E-Indonesia', 'xnr': 'इंडोनेशिया', 'xog': 'Yindonezya', 'yav': 'ɛndonesí', 'yi': 'אינדאנעזיע', 'yo': 'Indonéṣíà', 'yrl': 'Ĩdunesiya', 'yue': '印尼', 'yue-Hans': '印尼', 'yue-Hant': '印尼', 'zgh': 'ⴰⵏⴷⵓⵏⵉⵙⵢⴰ', 'zh': '印度尼西亚', 'zh-Hans': '印度尼西亚', 'zh-Hant': '印尼', 'zu': 'i-Indonesia'},
    'IE': {'af': 'Ierland', 'agq': 'Aelɨ̀lân', 'ak': 'Aereland', 'am': 'አየርላንድ', 'ar': 'أيرلندا', 'as': 'আয়াৰলেণ্ড', 'asa': 'Ayalandi', 'ast': 'Irlanda', 'az': 'İrlandiya', 'az-Cyrl': 'Ирландија', 'az-Latn': 'İrlandiya', 'bal-Latn': 'Áerlánd', 'bas': 'Ìrlând', 'be': 'Ірландыя', 'be-tarask': 'Ірландыя', 'bez': 'Huayalandi', 'bg': 'Ирландия', 'blo': 'Irlanɖ', 'bm': 'Irilandi', 'bn': 'আয়ারল্যান্ড', 'br': 'Iwerzhon', 'brx': 'आयारलेण्ड', 'bs': 'Irska', 'bs-Cyrl': 'Ирска', 'bs-Latn': 'Irska', 'ca': 'Irlanda', 'ca-ES-valencia': 'Irlanda', 'ccp': '𑄃𑄠𑄢𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴', 'ce': 'Ирланди', 'ceb': 'Ireland', 'cgg': 'Irerandi', 'chr': 'ᎠᏲᎳᏂ', 'ckb': 'ئیرلەند', 'cs': 'Irsko', 'cv': 'Ирланди', 'cy': 'Iwerddon', 'da': 'Irland', 'dav': 'Ayalandi', 'de': 'Irland', 'dje': 'Irlandu', 'dsb': 'Irska', 'dyo': 'Irland', 'dz': 'ཨཱ་ཡ་ལེནཌ', 'ebu': 'Ayalandi', 'ee': 'Ireland nutome', 'el': 'Ιρλανδία', 'el-polyton': 'Ἰρλανδία', 'en': 'Ireland', 'en-Dsrt': '𐐌𐑉𐑊𐐲𐑌𐐼', 'eo': 'Irlando', 'es': 'Irlanda', 'et': 'Iirimaa', 'eu': 'Irlanda', 'ewo': 'Irəlándə', 'fa': 'ایرلند', 'ff': 'Irlannda', 'ff-Adlm': '𞤋𞤪𞤤𞤢𞤲𞤣', 'ff-Latn': 'Irlannda', 'fi': 'Irlanti', 'fil': 'Ireland', 'fo': 'Írland', 'fr': 'Irlande', 'fur': 'Irlande', 'fy': 'Ierlân', 'ga': 'Éire', 'gd': 'Èirinn', 'gl': 'Irlanda', 'gsw': 'Irland', 'gu': 'આયર્લેન્ડ', 'guz': 'Ayalandi', 'ha': 'Ayalan', 'haw': 'ʻIlelani', 'he': 'אירלנד', 'hi': 'आयरलैंड', 'hi-Latn': 'Ireland', 'hr': 'Irska', 'hsb': 'Irska', 'ht': 'Irlande', 'hu': 'Írország', 'hy': 'Իռլանդիա', 'ia': 'Irlanda', 'id': 'Irlandia', 'ie': 'Irland', 'ig': 'Ireland', 'is': 'Írland', 'it': 'Irlanda', 'ja': 'アイルランド', 'jmc': 'Ayalandi', 'jv': 'Républik Irlan', 'ka': 'ირლანდია', 'kaa': 'Ирландия', 'kaa-Cyrl': 'Ирландия', 'kab': 'Lirlund', 'kam': 'Ayalandi', 'kde': 'Ayalandi', 'kea': 'Irlanda', 'kgp': 'Irỹna', 'khq': 'Irlandu', 'ki': 'Ayalandi', 'kk': 'Ирландия', 'kk-Cyrl': 'Ирландия', 'kln': 'Emetab Ireland', 'km': 'អៀរឡង់', 'kn': 'ಐರ್ಲೆಂಡ್', 'ko': '아일랜드', 'kok': 'आयरलँड', 'kok-Deva': 'आयरलँड', 'ks': 'اَیَرلینڑ', 'ks-Arab': 'اَیَرلینڑ', 'ksb': 'Ayalandi', 'ksf': 'ilán', 'ksh': 'Irrland', 'ku': 'Îrlanda', 'kxv': 'aayarlanḍ', 'kxv-Deva': 'आयरलेंड', 'kxv-Latn': 'aayarlanḍ', 'kxv-Orya': 'ଅଇରଲାଣ୍ଡ', 'kxv-Telu': 'ఐర్లాండ్', 'ky': 'Ирландия', 'lag': 'Ayaláandi', 'lb': 'Irland', 'lg': 'Ayalandi', 'lld': 'Irlanda', 'ln': 'Irelandɛ', 'lo': 'ໄອແລນ', 'lt': 'Airija', 'lu': 'Irelande', 'luo': 'Ireland', 'luy': 'Ayalandi', 'lv': 'Īrija', 'mai': 'आयरलैंड', 'mas': 'Ayalandi', 'mer': 'Aelandi', 'mfe': 'Irland', 'mg': 'Irlandy', 'mgh': 'Uayalandi', 'mi': 'Airani', 'mk': 'Ирска', 'ml': 'അയർലൻഡ്', 'mn': 'Ирланд', 'mr': 'आयर्लंड', 'ms': 'Ireland', 'mt': 'l-Irlanda', 'mua': 'Sǝr Ireland', 'my': 'အိုင်ယာလန်', 'mzn': 'ایرلند', 'naq': 'Irlandi', 'nb': 'Irland', 'nd': 'Ireland', 'ne': 'आयरल्याण्ड', 'nl': 'Ierland', 'nmg': 'Irland', 'nn': 'Irland', 'no': 'Irland', 'nqo': 'ߌߙߑߟߊ߲ߘ', 'nyn': 'Irerandi', 'om': 'Ayeerlaand', 'or': 'ଆୟରଲ୍ୟାଣ୍ଡ', 'pa': 'ਆਇਰਲੈਂਡ', 'pa-Guru': 'ਆਇਰਲੈਂਡ', 'pcm': 'Ayaland', 'pl': 'Irlandia', 'ps': 'آيرلېنډ', 'pt': 'Irlanda', 'qu': 'Irlanda', 'rm': 'Irlanda', 'rn': 'Irilandi', 'ro': 'Irlanda', 'rof': 'Ayalandi', 'ru': 'Ирландия', 'rwk': 'Ayalandi', 'sah': 'Ирландия', 'saq': 'Ayalandi', 'sbp': 'Ayalandi', 'sc': 'Irlanda', 'sd': 'آئرلينڊ', 'sd-Arab': 'آئرلينڊ', 'se': 'Irlánda', 'seh': 'Irlanda', 'ses': 'Irlandu', 'sg': 'Irlânde', 'shi': 'ⵉⵔⵍⴰⵏⴷⴰ', 'shi-Latn': 'irlanda', 'shi-Tfng': 'ⵉⵔⵍⴰⵏⴷⴰ', 'si': 'අයර්ලන්තය', 'sk': 'Írsko', 'sl': 'Irska', 'smn': 'Irland', 'sn': 'Ireland', 'so': 'Ayrlaand', 'sq': 'Irlandë', 'sr': 'Ирска', 'sr-Cyrl': 'Ирска', 'sr-Latn': 'Irska', 'sv': 'Irland', 'sw': 'Ayalandi', 'syr': 'ܐܝܪܠܢܕ', 'ta': 'அயர்லாந்து', 'te': 'ఐర్లాండ్', 'teo': 'Ayalandi', 'tg': 'Ирландия', 'th': 'ไอร์แลนด์', 'ti': 'ኣየርላንድ', 'tk': 'Irlandiýa', 'to': 'ʻAealani', 'tr': 'İrlanda', 'tt': 'Ирландия', 'twq': 'Irlandu', 'tzm': 'Irlanḍa', 'ug': 'ئىرېلاندىيە', 'uk': 'Ірландія', 'ur': 'آئرلینڈ', 'uz': 'Irlandiya', 'uz-Cyrl': 'Ирландия', 'uz-Latn': 'Irlandiya', 'vai': 'ꕉꔓ ꖨꕮꕊ', 'vai-Latn': 'Áre Lumaã', 'vai-Vaii': 'ꕉꔓ ꖨꕮꕊ', 'vec': 'Irlanda', 'vi': 'Ireland', 'vun': 'Ayalandi', 'wae': 'Irland', 'wo': 'Irlànd', 'xh': 'E-Ireland', 'xnr': 'आयरलैंड', 'xog': 'Ayalandi', 'yav': 'ililánd', 'yi': 'אירלאַנד', 'yo': 'Ailandi', 'yrl': 'Irãda', 'yue': '愛爾蘭', 'yue-Hans': '爱尔兰', 'yue-Hant': '愛爾蘭', 'zgh': 'ⵉⵔⵍⴰⵏⴷⴰ', 'zh': '爱尔兰', 'zh-Hans': '爱尔兰', 'zh-Hant': '愛爾蘭', 'zu': 'i-Ireland'},
    'IL': {'af': 'Israel', 'agq': 'Ezɨ̀lɛ̂', 'ak': 'Israe', 'am': 'እስራኤል', 'ar': 'إسرائيل', 'as': 'ইজৰাইল', 'asa': 'Ithraeli', 'ast': 'Israel', 'az': 'İsrail', 'az-Cyrl': 'Исраил', 'az-Latn': 'İsrail', 'bal-Latn': 'Esráil', 'bas': 'Isràɛ̂l', 'be': 'Ізраіль', 'be-tarask': 'Ізраіль', 'bez': 'Huislaheli', 'bg': 'Израел', 'bgn': 'اسرائیل', 'blo': 'Yishraɛl', 'bm': 'Isirayeli', 'bn': 'ইজরায়েল', 'br': 'Israel', 'brx': 'इज्राईल', 'bs': 'Izrael', 'bs-Cyrl': 'Израел', 'bs-Latn': 'Izrael', 'ca': 'Israel', 'ca-ES-valencia': 'Israel', 'ccp': '𑄄𑄎𑄴𑄢𑄠𑄬𑄣𑄴', 'ce': 'Израиль', 'ceb': 'Israel', 'cgg': 'Isirairi', 'chr': 'ᎢᏏᎵᏱ', 'ckb': 'ئیسرائیل', 'cs': 'Izrael', 'cv': 'Израиль', 'cy': 'Israel', 'da': 'Israel', 'dav': 'Israeli', 'de': 'Israel', 'dje': 'Israyel', 'dsb': 'Israel', 'dyo': 'Israel', 'dz': 'ཨིས་ར་ཡེལ', 'ebu': 'Israeli', 'ee': 'Israel nutome', 'el': 'Ισραήλ', 'el-polyton': 'Ἰσραήλ', 'en': 'Israel', 'en-Dsrt': '𐐆𐑆𐑉𐐨𐐲𐑊', 'eo': 'Israelo', 'es': 'Israel', 'et': 'Iisrael', 'eu': 'Israel', 'ewo': 'Isəraɛ́l', 'fa': 'اسرائیل', 'ff': 'Israa’iila', 'ff-Adlm': '𞤋𞤧𞤪𞤢𞥄𞤴𞤭𞥅𞤤', 'ff-Latn': 'Israa’iila', 'fi': 'Israel', 'fil': 'Israel', 'fo': 'Ísrael', 'fr': 'Israël', 'fur': 'Israêl', 'fy': 'Israël', 'ga': 'Iosrael', 'gd': 'Iosrael', 'gl': 'Israel', 'gsw': 'Israel', 'gu': 'ઇઝરાઇલ', 'guz': 'Israeli', 'ha': 'Israʼila', 'haw': 'ʻIseraʻela', 'he': 'ישראל', 'hi': 'इज़राइल', 'hi-Latn': 'Israel', 'hr': 'Izrael', 'hsb': 'Israel', 'ht': 'Israël', 'hu': 'Izrael', 'hy': 'Իսրայել', 'ia': 'Israel', 'id': 'Israel', 'ig': 'Israel', 'is': 'Ísrael', 'it': 'Israele', 'ja': 'イスラエル', 'jgo': 'Islayɛ̂l', 'jmc': 'Israeli', 'jv': 'Israèl', 'ka': 'ისრაელი', 'kaa': 'Израиль', 'kaa-Cyrl': 'Израиль', 'kab': 'Izrayil', 'kam': 'Israeli', 'kde': 'Islaeli', 'kea': 'Israel', 'kgp': 'Isihaé', 'khq': 'Israyel', 'ki': 'Israeli', 'kk': 'Израиль', 'kk-Cyrl': 'Израиль', 'kln': 'Emetab Israel', 'km': 'អ៊ីស្រាអែល', 'kn': 'ಇಸ್ರೇಲ್', 'ko': '이스라엘', 'kok': 'इस्त्राइल', 'kok-Deva': 'इस्त्राइल', 'ks': 'اسرا ییل', 'ks-Arab': 'اسرا ییل', 'ksb': 'Islaeli', 'ksf': 'israɛ́l', 'ksh': 'Ißraäl', 'ku': 'Îsraîl', 'kxv': 'israel', 'kxv-Deva': 'इज़राइल', 'kxv-Latn': 'israel', 'kxv-Orya': 'ଇସରାଇଲ', 'kxv-Telu': 'ఇజ్రాయెల్', 'ky': 'Израиль', 'lag': 'Isiraéeli', 'lb': 'Israel', 'lg': 'Yisirayeri', 'lld': 'Israel', 'ln': 'Isirayelɛ', 'lo': 'ອິສຣາເອວ', 'lt': 'Izraelis', 'lu': 'Isirayele', 'luo': 'Israel', 'luy': 'Israeli', 'lv': 'Izraēla', 'mai': 'इजरायल', 'mas': 'Israeli', 'mer': 'Isiraeli', 'mfe': 'Izrael', 'mg': 'Israely', 'mgh': 'Uisraeli', 'mi': 'Iharaira', 'mk': 'Израел', 'ml': 'ഇസ്രായേൽ', 'mn': 'Израил', 'mr': 'इस्त्राइल', 'ms': 'Israel', 'mt': 'Iżrael', 'mua': 'Sǝr Israel', 'my': 'အစ္စရေး', 'mzn': 'ایسراییل', 'naq': 'Israeli', 'nb': 'Israel', 'nd': 'Isuraeli', 'ne': 'इजरायल', 'nl': 'Israël', 'nmg': 'Äsrɛl', 'nn': 'Israel', 'no': 'Israel', 'nqo': 'ߌߛߑߙߊߍߟ', 'nyn': 'Isirairi', 'om': 'Israa’eel', 'or': 'ଇସ୍ରାଏଲ୍', 'pa': 'ਇਜ਼ਰਾਈਲ', 'pa-Guru': 'ਇਜ਼ਰਾਈਲ', 'pcm': 'Ízrẹl', 'pl': 'Izrael', 'ps': 'اسراييل', 'pt': 'Israel', 'qu': 'Israel', 'rm': 'Israel', 'rn': 'Isiraheli', 'ro': 'Israel', 'rof': 'Israeli', 'ru': 'Израиль', 'rwk': 'Israeli', 'saq': 'Israeli', 'sbp': 'Isilaeli', 'sc': 'Israele', 'sd': 'اسرائيل', 'sd-Arab': 'اسرائيل', 'se': 'Israel', 'seh': 'Israel', 'ses': 'Israyel', 'sg': 'Israëli', 'shi': 'ⵉⵙⵔⴰⵢⵉⵍ', 'shi-Latn': 'israyil', 'shi-Tfng': 'ⵉⵙⵔⴰⵢⵉⵍ', 'si': 'ඊශ්\u200dරායලය', 'sk': 'Izrael', 'sl': 'Izrael', 'smn': 'Israel', 'sn': 'Izuraeri', 'so': 'Israaʼiil', 'sq': 'Izrael', 'sr': 'Израел', 'sr-Cyrl': 'Израел', 'sr-Latn': 'Izrael', 'sv': 'Israel', 'sw': 'Israeli', 'syr': 'ܐܝܣܪܐܝܠ', 'ta': 'இஸ்ரேல்', 'te': 'ఇజ్రాయెల్', 'teo': 'Israeli', 'tg': 'Исроил', 'th': 'อิสราเอล', 'ti': 'እስራኤል', 'tk': 'Ysraýyl', 'to': 'ʻIsileli', 'tr': 'İsrail', 'tt': 'Израиль', 'twq': 'Israyel', 'tzm': 'Israeil', 'ug': 'ئىسرائىلىيە', 'uk': 'Ізраїль', 'ur': 'اسرائیل', 'uz': 'Isroil', 'uz-Cyrl': 'Исроил', 'uz-Latn': 'Isroil', 'vai': 'ꕑꕇꔻꕞꔤꕞ', 'vai-Latn': 'Bhanísiláila', 'vai-Vaii': 'ꕑꕇꔻꕞꔤꕞ', 'vec': 'Izraele', 'vi': 'Israel', 'vun': 'Israeli', 'wae': 'Israel', 'wo': 'Israyel', 'xh': 'E-Israel', 'xnr': 'इज़राइल', 'xog': 'Yisirayeri', 'yav': 'ísilayɛ́l', 'yi': 'ישראל', 'yo': 'Iserẹli', 'yrl': 'Isirayeu', 'yue': '以色列', 'yue-Hans': '以色列', 'yue-Hant': '以色列', 'zgh': 'ⵉⵙⵔⴰⵢⵉⵍ', 'zh': '以色列', 'zh-Hans': '以色列', 'zh-Hant': '以色列', 'zu': 'kwa-Israel'},
    'IM': {'af': 'Eiland Man', 'ak': 'Isle of Man', 'am': 'አይል ኦፍ ማን', 'ar': 'جزيرة مان', 'as': 'আইল অফ মেন', 'ast': 'Islla de Man', 'az': 'Men adası', 'az-Cyrl': 'Мен адасы', 'az-Latn': 'Men adası', 'bal-Latn': 'Áisale Mardom', 'be': 'Востраў Мэн', 'be-tarask': 'Востраў Мэн', 'bg': 'остров Ман', 'blo': 'Man kaAtukǝltǝna', 'bn': 'আইল অফ ম্যান', 'br': 'Enez Vanav', 'brx': 'आयोल अफ मेन', 'bs': 'Ostrvo Man', 'bs-Cyrl': 'Острво Мен', 'bs-Latn': 'Ostrvo Man', 'ca': 'Illa de Man', 'ca-ES-valencia': 'Illa de Man', 'ccp': '𑄃𑄭𑄣𑄴 𑄃𑄧𑄜𑄴 𑄟𑄳𑄠𑄚𑄴', 'ce': 'Мэн гӀайре', 'ceb': 'Isle of Man', 'chr': 'ᎤᏍᏗ ᎤᎦᏚᏛᎢ ᎾᎿ ᎠᏍᎦᏯ', 'ckb': 'دوورگەی مان', 'cs': 'Ostrov Man', 'cv': 'Мэн утравӗ', 'cy': 'Ynys Manaw', 'da': 'Isle of Man', 'de': 'Isle of Man', 'dsb': 'Man', 'dz': 'ཨ་ཡུལ་ ཨོཕ་ མཱན', 'ee': 'Aisle of Man nutome', 'el': 'Νήσος του Μαν', 'el-polyton': 'Νήσος του Μαν', 'en': 'Isle of Man', 'en-Dsrt': '𐐌𐐲𐑊 𐐲𐑁 𐐣𐐰𐑌', 'es': 'Isla de Man', 'et': 'Mani saar', 'eu': 'Man uhartea', 'fa': 'جزیرهٔ من', 'ff-Adlm': '𞤅𞤵𞤪𞤭𞥅𞤪𞤫 𞤃𞤫𞥅𞤲', 'fi': 'Mansaari', 'fil': 'Isle of Man', 'fo': 'Isle of Man', 'fr': 'Île de Man', 'fur': 'Isule di Man', 'fy': 'Isle of Man', 'ga': 'Oileán Mhanann', 'gd': 'Eilean Mhanainn', 'gl': 'Illa de Man', 'gsw': 'Insle vo Män', 'gu': 'આઇલ ઑફ મેન', 'gv': 'Ellan Vannin', 'ha': 'Isle of Man', 'he': 'האי מאן', 'hi': 'आइल ऑफ़ मैन', 'hi-Latn': 'Isle of Man', 'hr': 'Otok Man', 'hsb': 'Man', 'ht': 'Île de Man', 'hu': 'Man-sziget', 'hy': 'Մեն կղզի', 'ia': 'Insula de Man', 'id': 'Pulau Man', 'ig': 'Isle of Man', 'is': 'Mön', 'it': 'Isola di Man', 'ja': 'マン島', 'jv': 'Pulo Man', 'ka': 'მენის კუნძული', 'kaa': 'Мэн атаўы', 'kaa-Cyrl': 'Мэн атаўы', 'kea': 'Ilha di Man', 'kgp': 'Mỹn Goj-vẽso', 'kk': 'Мэн аралы', 'kk-Cyrl': 'Мэн аралы', 'km': 'អែលអុហ្វមែន', 'kn': 'ಐಲ್ ಆಫ್ ಮ್ಯಾನ್', 'ko': '맨섬', 'kok': 'इसले ऑफ मॅन', 'kok-Deva': 'इसले ऑफ मॅन', 'ks': 'آیِل آف مین', 'ks-Arab': 'آیِل آف مین', 'ksh': 'de Ensel Män', 'ku': 'Girava Manê', 'kxv': 'aail āp man', 'kxv-Deva': 'आइल ऑप़ मेन', 'kxv-Latn': 'aail āp man', 'kxv-Orya': 'ମାଣିସି ତା ଦିପ', 'kxv-Telu': 'ఐల్ ఆప్ మాన్', 'ky': 'Мэн аралы', 'lb': 'Isle of Man', 'lld': 'Isola de Man', 'lo': 'ເອວ ອອບ ແມນ', 'lt': 'Meno Sala', 'lv': 'Menas sala', 'mai': 'आइल ऑफ मैन', 'mi': 'Te Moutere Mana', 'mk': 'Остров Ман', 'ml': 'ഐൽ ഓഫ് മാൻ', 'mn': 'Мэн Арал', 'mr': 'आयल ऑफ मॅन', 'ms': 'Isle of Man', 'mt': 'Isle of Man', 'my': 'မန်ကျွန်း', 'mzn': 'من ِجزیره', 'nb': 'Man', 'ne': 'आइल अफ म्यान', 'nl': 'Isle of Man', 'nn': 'Man', 'no': 'Man', 'nqo': 'ߡߊ߲߯ ߕߌ߲', 'om': 'Islee oof Maan', 'or': 'ଆଇଲ୍\u200c ଅଫ୍\u200c ମ୍ୟାନ୍\u200c', 'pa': 'ਆਇਲ ਆਫ ਮੈਨ', 'pa-Guru': 'ਆਇਲ ਆਫ ਮੈਨ', 'pcm': 'Aíl ọf Man', 'pl': 'Wyspa Man', 'ps': 'د آئل آف مین', 'pt': 'Ilha de Man', 'qu': 'Isla de Man', 'rm': 'Insla da Man', 'ro': 'Insula Man', 'ru': 'о-в Мэн', 'sah': 'Мэн арыы', 'sc': 'Ìsula de Man', 'sd': 'انسانن جو ٻيٽ', 'sd-Arab': 'انسانن جو ٻيٽ', 'se': 'Mann-sullot', 'si': 'අයිල් ඔෆ් මෑන්', 'sk': 'Ostrov Man', 'sl': 'Otok Man', 'smn': 'Mansuálui', 'so': 'Jasiiradda Isle of Man', 'sq': 'Ishulli i Manit', 'sr': 'Острво Ман', 'sr-Cyrl': 'Острво Ман', 'sr-Latn': 'Ostrvo Man', 'sv': 'Isle of Man', 'sw': 'Kisiwa cha Man', 'syr': 'ܓܙܪܬܐ ܕܡܐܢ', 'ta': 'ஐல் ஆஃப் மேன்', 'te': 'ఐల్ ఆఫ్ మాన్', 'tg': 'Ҷазираи Мэн', 'th': 'เกาะแมน', 'ti': 'ኣይል ኦፍ ማን', 'tk': 'Men adasy', 'to': 'Motu Mani', 'tr': 'Man Adası', 'tt': 'Мэн утравы', 'ug': 'مان ئارىلى', 'uk': 'Острів Мен', 'ur': 'آئل آف مین', 'uz': 'Men oroli', 'uz-Cyrl': 'Мэн ороли', 'uz-Latn': 'Men oroli', 'vai': 'ꕮꘋ ꔳꘋꗣ', 'vai-Vaii': 'ꕮꘋ ꔳꘋꗣ', 'vec': 'Ìzola de Man', 'vi': 'Đảo Man', 'wae': 'Isle of Man', 'wo': 'Dunu Maan', 'xh': 'E-Isle of Man', 'xnr': 'आइल ऑफ़ मैन', 'yo': 'Erékùṣù ilẹ̀ Man', 'yrl': 'Mã Kapuãma', 'yue': '曼島', 'yue-Hans': '曼岛', 'yue-Hant': '曼島', 'zh': '马恩岛', 'zh-Hans': '马恩岛', 'zh-Hant': '曼島', 'zu': 'i-Isle of Man'},
    'IN': {'af': 'Indië', 'agq': 'Endìa', 'ak': 'India', 'am': 'ህንድ', 'ar': 'الهند', 'as': 'ভাৰত', 'asa': 'India', 'ast': 'India', 'az': 'Hindistan', 'az-Cyrl': 'Һиндистан', 'az-Latn': 'Hindistan', 'bal-Latn': 'Hendostán', 'bas': 'Indìà', 'be': 'Індыя', 'be-tarask': 'Індыя', 'bez': 'Huindia', 'bg': 'Индия', 'bgc': 'भारत', 'bho': 'भारत', 'blo': 'Inɖiya', 'bm': 'Ɛndujamana', 'bn': 'ভারত', 'bo': 'རྒྱ་གར་', 'br': 'India', 'brx': 'भारत', 'bs': 'Indija', 'bs-Cyrl': 'Индија', 'bs-Latn': 'Indija', 'ca': 'Índia', 'ca-ES-valencia': 'Índia', 'ccp': '𑄞𑄢𑄧𑄖𑄴', 'ce': 'ХӀинди', 'ceb': 'India', 'cgg': 'Indiya', 'chr': 'ᎢᏅᏗᎾ', 'ckb': 'ھیندستان', 'cs': 'Indie', 'cv': 'Инди', 'cy': 'India', 'da': 'Indien', 'dav': 'India', 'de': 'Indien', 'dje': 'Indu laboo', 'doi': 'भारत', 'dsb': 'Indiska', 'dyo': 'End', 'dz': 'རྒྱ་གར', 'ebu': 'India', 'ee': 'India nutome', 'el': 'Ινδία', 'el-polyton': 'Ἰνδία', 'en': 'India', 'en-Dsrt': '𐐆𐑌𐐼𐐨𐐲', 'eo': 'Hindujo', 'es': 'India', 'et': 'India', 'eu': 'India', 'ewo': 'ɛ́ndə', 'fa': 'هند', 'ff': 'Enndo', 'ff-Adlm': '𞤋𞤲𞤣𞤭𞤴𞤢', 'ff-Latn': 'Enndo', 'fi': 'Intia', 'fil': 'India', 'fo': 'India', 'fr': 'Inde', 'fur': 'India', 'fy': 'India', 'ga': 'an India', 'gd': 'Na h-Innseachan', 'gl': 'India', 'gsw': 'Indie', 'gu': 'ભારત', 'guz': 'India', 'ha': 'Indiya', 'haw': 'ʻĪnia', 'he': 'הודו', 'hi': 'भारत', 'hi-Latn': 'Bharat', 'hr': 'Indija', 'hsb': 'Indiska', 'ht': 'Inde', 'hu': 'India', 'hy': 'Հնդկաստան', 'ia': 'India', 'id': 'India', 'ie': 'India', 'ig': 'India', 'ii': 'ꑴꄗ', 'is': 'Indland', 'it': 'India', 'ja': 'インド', 'jgo': 'Ándɛ', 'jmc': 'India', 'jv': 'Indhia', 'ka': 'ინდოეთი', 'kaa': 'Ҳиндстан', 'kaa-Cyrl': 'Ҳиндстан', 'kab': 'Lhend', 'kam': 'India', 'kde': 'India', 'kea': 'Índia', 'kgp': 'Ĩnija', 'khq': 'Indu laboo', 'ki': 'India', 'kk': 'Үндістан', 'kk-Cyrl': 'Үндістан', 'kln': 'Emetab India', 'km': 'ឥណ្ឌា', 'kn': 'ಭಾರತ', 'ko': '인도', 'kok': 'भारत', 'kok-Deva': 'भारत', 'kok-Latn': 'Bharot', 'ks': 'ہِندوستان', 'ks-Arab': 'ہِندوستان', 'ks-Deva': 'हिंदोस्तान', 'ksb': 'India', 'ksf': 'indí', 'ksh': 'Indije', 'ku': 'Hindistan', 'kxv': 'inḍiā', 'kxv-Deva': 'बारत', 'kxv-Latn': 'inḍiā', 'kxv-Orya': 'ବାରତ', 'kxv-Telu': 'బారతదెసాం', 'ky': 'Индия', 'lag': 'Índia', 'lb': 'Indien', 'lg': 'Buyindi', 'lld': 'India', 'ln': 'Índɛ', 'lo': 'ອິນເດຍ', 'lrc': 'ھئن', 'lt': 'Indija', 'lu': 'Inde', 'luo': 'India', 'luy': 'India', 'lv': 'Indija', 'mai': 'भारत', 'mas': 'India', 'mer': 'India', 'mfe': 'Lenn', 'mg': 'Indy', 'mgh': 'Uhindini', 'mi': 'Inia', 'mk': 'Индија', 'ml': 'ഇന്ത്യ', 'mn': 'Энэтхэг', 'mni': 'ইন্দিয়া', 'mni-Beng': 'ইন্দিয়া', 'mr': 'भारत', 'ms': 'India', 'mt': 'l-Indja', 'mua': 'Sǝr Indǝ', 'my': 'အိန္ဒိယ', 'mzn': 'هند', 'naq': 'Indiab', 'nb': 'India', 'nd': 'Indiya', 'ne': 'भारत', 'nl': 'India', 'nmg': 'India', 'nn': 'India', 'no': 'India', 'nqo': 'ߌߘߎ߬', 'nyn': 'Indiya', 'om': 'Hindii', 'or': 'ଭାରତ', 'os': 'Инди', 'pa': 'ਭਾਰਤ', 'pa-Guru': 'ਭਾਰਤ', 'pcm': 'Índia', 'pl': 'Indie', 'ps': 'هند', 'pt': 'Índia', 'qu': 'India', 'raj': 'भारत', 'rm': 'India', 'rn': 'Ubuhindi', 'ro': 'India', 'rof': 'India', 'ru': 'Индия', 'rwk': 'India', 'sa': 'भारतः', 'saq': 'India', 'sat': 'ᱤᱱᱰᱤᱭᱟ', 'sat-Olck': 'ᱤᱱᱰᱤᱭᱟ', 'sbp': 'Indiya', 'sc': 'Ìndia', 'sd': 'ڀارت', 'sd-Arab': 'ڀارت', 'sd-Deva': 'भारत', 'se': 'India', 'seh': 'Índia', 'ses': 'Indu laboo', 'sg': 'Ênnde', 'shi': 'ⵍⵀⵉⵏⴷ', 'shi-Latn': 'lhind', 'shi-Tfng': 'ⵍⵀⵉⵏⴷ', 'si': 'ඉන්දියාව', 'sk': 'India', 'sl': 'Indija', 'smn': 'India', 'sn': 'India', 'so': 'Hindiya', 'sq': 'Indi', 'sr': 'Индија', 'sr-Cyrl': 'Индија', 'sr-Latn': 'Indija', 'su': 'India', 'su-Latn': 'India', 'sv': 'Indien', 'sw': 'India', 'syr': 'ܗܢܕܘ', 'ta': 'இந்தியா', 'te': 'భారతదేశం', 'teo': 'India', 'tg': 'Ҳиндустон', 'th': 'อินเดีย', 'ti': 'ህንዲ', 'tk': 'Hindistan', 'to': 'ʻInitia', 'tr': 'Hindistan', 'tt': 'Индия', 'twq': 'Indu laboo', 'tzm': 'Hind', 'ug': 'ھىندىستان', 'uk': 'Індія', 'ur': 'بھارت', 'uz': 'Hindiston', 'uz-Cyrl': 'Ҳиндистон', 'uz-Latn': 'Hindiston', 'vai': 'ꔤꔺꕩ', 'vai-Latn': 'Índiya', 'vai-Vaii': 'ꔤꔺꕩ', 'vec': 'India', 'vi': 'Ấn Độ', 'vun': 'India', 'wae': 'Indie', 'wo': 'End', 'xh': 'E-Indiya', 'xnr': 'भारत', 'xog': 'Buyindi', 'yav': 'ɛ́ɛnd', 'yi': 'אינדיע', 'yo': 'Íńdíà', 'yrl': 'Ĩdiya', 'yue': '印度', 'yue-Hans': '印度', 'yue-Hant': '印度', 'zgh': 'ⵍⵀⵉⵏⴷ', 'zh': '印度', 'zh-Hans': '印度', 'zh-Hant': '印度', 'zu': 'i-India'},
    'IO': {'af': 'Brits-Indiese Oseaangebied', 'ak': 'Britenfo Man Wɔ India Po No Mu', 'am': 'የብሪታኒያ ህንድ ውቂያኖስ ግዛት', 'ar': 'الإقليم البريطاني في المحيط الهندي', 'as': 'ব্ৰিটিছ ইণ্ডিয়ান অ’চন টেৰিট’ৰি', 'az': 'Britaniyanın Hind Okeanı Ərazisi', 'az-Latn': 'Britaniyanın Hind Okeanı Ərazisi', 'bal-Latn': 'Bartáni Hendi Zerdamag', 'be': 'Брытанская тэрыторыя ў Індыйскім акіяне', 'be-tarask': 'Брытанская тэрыторыя ў Індыйскім акіяне', 'bg': 'Британска територия в Индийския океан', 'blo': 'Gɛɛshɩ kaAtǝna Inɖiya kaTeŋku nɩ', 'bn': 'ব্রিটিশ ভারত মহাসাগরীয় অঞ্চল', 'brx': 'ब्रिटिशनि भारतारि लैथो बाहागो', 'bs': 'Britanska Teritorija u Indijskom Okeanu', 'bs-Cyrl': 'Британска територија у Индијском океану', 'bs-Latn': 'Britanska Teritorija u Indijskom Okeanu', 'ca': 'Territori Britànic de l’Oceà Índic', 'ca-ES-valencia': 'Territori Britànic de l’Oceà Índic', 'ceb': 'Teritoryo sa British Indian Ocean', 'chr': 'ᏈᏗᏏ ᏴᏫᏯ ᎠᎺᏉ ᎢᎬᎾᏕᏅ', 'cs': 'Britské indickooceánské území', 'cy': 'Tiriogaeth Brydeinig Cefnfor India', 'da': 'Det Britiske Territorium i Det Indiske Ocean', 'de': 'Britisches Territorium im Indischen Ozean', 'dsb': 'Britiski indiskooceaniski teritorium', 'el': 'Βρετανικά Εδάφη Ινδικού Ωκεανού', 'el-polyton': 'Βρετανικά Εδάφη Ινδικού Ωκεανού', 'en': 'British Indian Ocean Territory', 'es': 'Territorio Británico del Océano Índico', 'et': 'Briti India ookeani ala', 'eu': 'Indiako Ozeanoko lurralde britainiarra', 'fa': 'قلمرو بریتانیا در اقیانوس هند', 'fi': 'Brittiläinen Intian valtameren alue', 'fil': 'British Indian Ocean Territory', 'fo': 'Stóra Bretlands Indiahavoyggjar', 'fr': 'Territoire britannique de l’océan Indien', 'ga': 'Críoch Aigéan Indiach na Breataine', 'gd': 'Ranntair Breatannach Cuan nan Innseachan', 'gl': 'Territorio Británico do Océano Índico', 'gu': 'બ્રિટિશ ઇન્ડિયન ઓશન ટેરિટરી', 'ha': 'Yankin Birtaniya Na Tekun Indiya', 'he': 'הטריטוריה הבריטית באוקיינוס ההודי', 'hi': 'ब्रिटिश हिंद महासागरीय क्षेत्र', 'hi-Latn': 'British Indian Ocean Territory', 'hr': 'Britanski Indijskooceanski Teritorij', 'hsb': 'Britiski teritorij w Indiskim oceanje', 'ht': 'Territoire britannique de l’océan Indien', 'hu': 'Brit Indiai-óceáni Terület', 'hy': 'Բրիտանական տարածք Հնդկական Օվկիանոսում', 'ia': 'Territorio oceanic britanno-indian', 'id': 'Wilayah Inggris di Samudra Hindia', 'ig': 'British Indian Ocean Territory', 'is': 'Bresku Indlandshafseyjar', 'it': 'Territorio Britannico dell’Oceano Indiano', 'ja': '英領インド洋地域', 'jv': 'Wilayah Inggris ing Segara Hindia', 'ka': 'ბრიტანეთის ტერიტორია ინდოეთის ოკეანეში', 'kaa': 'Британияныӊ Ҳинд океанындағы территориясы', 'kaa-Cyrl': 'Британияныӊ Ҳинд океанындағы территориясы', 'kk': 'Үнді мұхитындағы Британ аймағы', 'kk-Cyrl': 'Үнді мұхитындағы Британ аймағы', 'km': 'ដែនដី\u200bអង់គ្លេស\u200bនៅ\u200bមហា\u200bសមុទ្រ\u200bឥណ្ឌា', 'kn': 'ಬ್ರಿಟೀಷ್ ಹಿಂದೂ ಮಹಾಸಾಗರದ ಪ್ರದೇಶ', 'ko': '영국령 인도양 지역', 'kok': 'ब्रिटिश हिंद महासागरीय क्षेत्र', 'kok-Deva': 'ब्रिटिश हिंद महासागरीय क्षेत्र', 'ku': 'Herêma Okyanûsa Hindî ya Brîtanyayê', 'kxv': 'briṭis inḍiāti samudra handi', 'kxv-Deva': 'ब्रिटिस हिंद सामुद्रि हांडि', 'kxv-Latn': 'briṭis inḍiāti samudra handi', 'kxv-Orya': 'ବ୍ରିଟିସ୍ ହିନ୍ଦ ସାମୁଦ୍ରି ହାନ୍ଦି', 'kxv-Telu': 'బ్రిటిస్ హీందూ మహాసముద్ర ప్రాంతం', 'ky': 'Инди океанындагы Британ территориясы', 'lld': 'Teritore britanich dl Ozean Indian', 'lo': 'ເຂດແດນອັງກິດໃນມະຫາສະໝຸດອິນເດຍ', 'lt': 'Indijos Vandenyno Britų Sritis', 'lv': 'Indijas okeāna Britu teritorija', 'mai': 'ब्रिटिश हिंद महासागरीय क्षेत्र', 'mi': 'Te Rohe o te Moana Īniana Piritihi', 'mk': 'Британска Индоокеанска Територија', 'ml': 'ബ്രിട്ടീഷ് ഇന്ത്യൻ ഓഷ്യൻ ടെറിട്ടറി', 'mn': 'Британийн харьяа Энэтхэгийн далай дахь нутаг дэвсгэр', 'mr': 'ब्रिटिश हिंद महासागर प्रदेश', 'ms': 'Wilayah Lautan Hindi British', 'my': 'ဗြိတိသျှပိုင် အိန္ဒိယသမုဒ္ဒရာကျွန်းများ', 'nb': 'Det britiske territoriet i Indiahavet', 'ne': 'बेलायती हिन्द महासागर क्षेत्र', 'nl': 'Brits Indische Oceaanterritorium', 'nn': 'Det britiske territoriet i Indiahavet', 'no': 'Det britiske territoriet i Indiahavet', 'nqo': 'ߓߙߌߕߊ߲ߓߊ߫ ߟߊ߫ ߌ߲ߘߎ߫ ߟߌ߲ߓߊ߲ ߞߣߍ', 'om': 'Daangaa Galaana Hindii Biritish', 'or': 'ବ୍ରିଟିଶ୍\u200c ଭାରତୀୟ ମହାସାଗର କ୍ଷେତ୍ର', 'pa': 'ਬਰਤਾਨਵੀ ਹਿੰਦ ਮਹਾਂਸਾਗਰ ਖਿੱਤਾ', 'pa-Guru': 'ਬਰਤਾਨਵੀ ਹਿੰਦ ਮਹਾਂਸਾਗਰ ਖਿੱਤਾ', 'pcm': 'Brítísh Índián Óshen Tẹ́rẹ́tri', 'pl': 'Brytyjskie Terytorium Oceanu Indyjskiego', 'ps': 'د برتانوي هند سمندري سيمه', 'pt': 'Território Britânico do Oceano Índico', 'qu': 'Territorio Británico del Océano Índico', 'ro': 'Teritoriul Britanic din Oceanul Indian', 'ru': 'Британская территория в Индийском океане', 'sc': 'Territòriu Britànnicu de s’Otzèanu Indianu', 'sd': 'برطانوي هندي سمنڊ خطو', 'sd-Arab': 'برطانوي هندي سمنڊ خطو', 'si': 'බ්\u200dරිතාන්\u200dය ඉන්දීය සාගර බල ප්\u200dරදේශය', 'sk': 'Britské indickooceánske územie', 'sl': 'Britansko ozemlje v Indijskem oceanu', 'so': 'Dhul xadeedka Badweynta Hindiya ee Ingiriiska', 'sq': 'Territori Britanik i Oqeanit Indian', 'sr': 'Британска територија Индијског океана', 'sr-Cyrl': 'Британска територија Индијског океана', 'sr-Latn': 'Britanska teritorija Indijskog okeana', 'sv': 'Brittiska territoriet i Indiska oceanen', 'sw': 'Eneo la Uingereza katika Bahari Hindi', 'syr': 'ܩܠܝܡܐ ܕܒܪܝܛܢܝܐ ܓܘ ܐܘܩܝܢܘܣ ܗܢܕܘܝܐ', 'ta': 'பிரிட்டிஷ் இந்தியப் பெருங்கடல் பிரதேசம்', 'te': 'బ్రిటిష్ హిందూ మహాసముద్ర ప్రాంతం', 'tg': 'Қаламрави Британия дар уқёнуси Ҳинд', 'th': 'บริติชอินเดียนโอเชียนเทร์ริทอรี', 'ti': 'ብሪጣንያዊ ህንዳዊ ውቅያኖስ ግዝኣት', 'tk': 'Britaniýanyň Hindi okeanyndaky territoriýalary', 'to': 'Potu fonua moana ʻInitia fakapilitānia', 'tr': 'Britanya Hint Okyanusu Toprakları', 'tt': 'Британиянең Һинд Океанындагы Территориясе', 'uk': 'Британська територія в Індійському океані', 'ur': 'برطانوی بحر ہند کا علاقہ', 'uz': 'Britaniyaning Hind okeanidagi hududi', 'uz-Latn': 'Britaniyaning Hind okeanidagi hududi', 'vec': 'Teritorio britànego de l’Osèano Indian', 'vi': 'Lãnh thổ Ấn Độ Dương thuộc Anh', 'wo': 'Terituwaaru Brëtaañ ci Oseyaa Enjeŋ', 'xh': 'EBritish Indian Ocean Territory', 'xnr': 'ब्रिटिश हिंद महासागरीय लाका', 'yo': 'Etíkun Índíánì ti Ìlú Bírítísì', 'yue': '英屬印度洋領地', 'yue-Hans': '英属印度洋领地', 'yue-Hant': '英屬印度洋領地', 'zh': '英属印度洋领地', 'zh-Hans': '英属印度洋领地', 'zh-Hant': '英屬印度洋領地', 'zu': 'i-British Indian Ocean Territory'},
    'IQ': {'af': 'Irak', 'agq': 'Èlâkɨ̀', 'ak': 'Irak', 'am': 'ኢራቅ', 'ar': 'العراق', 'as': 'ইৰাক', 'asa': 'Iraki', 'ast': 'Iraq', 'az': 'İraq', 'az-Cyrl': 'Ираг', 'az-Latn': 'İraq', 'bal-Latn': 'Erák', 'bas': 'Ìrâk', 'be': 'Ірак', 'be-tarask': 'Ірак', 'bez': 'Huilaki', 'bg': 'Ирак', 'bgn': 'عراق', 'blo': 'Ɩraakɩ', 'bm': 'Iraki', 'bn': 'ইরাক', 'br': 'Iraq', 'brx': 'ईराक', 'bs': 'Irak', 'bs-Cyrl': 'Ирак', 'bs-Latn': 'Irak', 'ca': 'Iraq', 'ca-ES-valencia': 'Iraq', 'ccp': '𑄄𑄢𑄇𑄴', 'ce': 'Ӏиракъ', 'ceb': 'Iraq', 'cgg': 'Iraaka', 'chr': 'ᎢᎳᎩ', 'ckb': 'عێراق', 'cs': 'Irák', 'cv': 'Ирак', 'cy': 'Irac', 'da': 'Irak', 'dav': 'Iraki', 'de': 'Irak', 'dje': 'Iraak', 'dsb': 'Irak', 'dyo': 'Irak', 'dz': 'ཨི་རཱཀ', 'ebu': 'Iraki', 'ee': 'iraqdukɔ', 'el': 'Ιράκ', 'el-polyton': 'Ἰράκ', 'en': 'Iraq', 'en-Dsrt': '𐐆𐑉𐐰𐐿', 'eo': 'Irako', 'es': 'Irak', 'et': 'Iraak', 'eu': 'Irak', 'ewo': 'Irág', 'fa': 'عراق', 'ff': 'Iraak', 'ff-Adlm': '𞤋𞤪𞤢𞥄𞤳', 'ff-Latn': 'Iraak', 'fi': 'Irak', 'fil': 'Iraq', 'fo': 'Irak', 'fr': 'Irak', 'fur': 'Iraq', 'fy': 'Irak', 'ga': 'an Iaráic', 'gd': 'Ioràc', 'gl': 'Iraq', 'gsw': 'Iraak', 'gu': 'ઇરાક', 'guz': 'Iraki', 'ha': 'Iraƙi', 'he': 'עיראק', 'hi': 'इराक', 'hi-Latn': 'Iraq', 'hr': 'Irak', 'hsb': 'Irak', 'ht': 'Irak', 'hu': 'Irak', 'hy': 'Իրաք', 'ia': 'Irak', 'id': 'Irak', 'ig': 'Iraq', 'is': 'Írak', 'it': 'Iraq', 'ja': 'イラク', 'jgo': 'Ilâk', 'jmc': 'Iraki', 'jv': 'Irak', 'ka': 'ერაყი', 'kaa': 'Ирак', 'kaa-Cyrl': 'Ирак', 'kab': 'Lɛiraq', 'kam': 'Iraki', 'kde': 'Ilaki', 'kea': 'Iraki', 'kgp': 'Iraki', 'khq': 'Iraak', 'ki': 'Iraki', 'kk': 'Ирак', 'kk-Cyrl': 'Ирак', 'kln': 'Emetab Iraq', 'km': 'អ៊ីរ៉ាក់', 'kn': 'ಇರಾಕ್', 'ko': '이라크', 'kok': 'इराक', 'kok-Deva': 'इराक', 'ks': 'ایٖراق', 'ks-Arab': 'ایٖراق', 'ksb': 'Ilaki', 'ksf': 'irák', 'ksh': 'Irak', 'ku': 'Îraq', 'kxv': 'irak', 'kxv-Deva': 'इराक', 'kxv-Latn': 'irak', 'kxv-Orya': 'ଇରାକ', 'kxv-Telu': 'ఇరాక్', 'ky': 'Ирак', 'lag': 'Iráaki', 'lb': 'Irak', 'lg': 'Yiraaka', 'lld': 'Irak', 'ln': 'Iraki', 'lo': 'ອີຣັກ', 'lt': 'Irakas', 'lu': 'Iraki', 'luo': 'Iraq', 'luy': 'Iraki', 'lv': 'Irāka', 'mai': 'इराक', 'mas': 'Iraki', 'mer': 'Iraki', 'mfe': 'Irak', 'mg': 'Irak', 'mgh': 'Wiraki', 'mi': 'Irāka', 'mk': 'Ирак', 'ml': 'ഇറാഖ്', 'mn': 'Ирак', 'mr': 'इराक', 'ms': 'Iraq', 'mt': 'l-Iraq', 'mua': 'Irak', 'my': 'အီရတ်', 'mzn': 'عراق', 'naq': 'Iraqi', 'nb': 'Irak', 'nd': 'Iraki', 'ne': 'इराक', 'nl': 'Irak', 'nmg': 'Irak', 'nn': 'Irak', 'no': 'Irak', 'nqo': 'ߌߙߊߞߌ߬', 'nyn': 'Iraaka', 'om': 'Iraaq', 'or': 'ଇରାକ୍', 'pa': 'ਇਰਾਕ', 'pa-Guru': 'ਇਰਾਕ', 'pcm': 'Irak', 'pl': 'Irak', 'ps': 'عراق', 'pt': 'Iraque', 'qu': 'Irak', 'rm': 'Irac', 'rn': 'Iraki', 'ro': 'Irak', 'rof': 'Iraki', 'ru': 'Ирак', 'rwk': 'Iraki', 'saq': 'Iraki', 'sbp': 'Ilaki', 'sc': 'Iraq', 'sd': 'عراق', 'sd-Arab': 'عراق', 'se': 'Irak', 'seh': 'Iraque', 'ses': 'Iraak', 'sg': 'Irâki', 'shi': 'ⵍⵄⵉⵔⴰⵇ', 'shi-Latn': 'lɛiraq', 'shi-Tfng': 'ⵍⵄⵉⵔⴰⵇ', 'si': 'ඉරාකය', 'sk': 'Irak', 'sl': 'Irak', 'smn': 'Irak', 'sn': 'Iraq', 'so': 'Ciraaq', 'sq': 'Irak', 'sr': 'Ирак', 'sr-Cyrl': 'Ирак', 'sr-Latn': 'Irak', 'sv': 'Irak', 'sw': 'Iraki', 'syr': 'ܥܝܪܩ', 'ta': 'ஈராக்', 'te': 'ఇరాక్', 'teo': 'Iraki', 'tg': 'Ироқ', 'th': 'อิรัก', 'ti': 'ዒራቕ', 'tk': 'Yrak', 'to': 'ʻIlaaki', 'tr': 'Irak', 'tt': 'Гыйрак', 'twq': 'Iraak', 'tzm': 'Ɛiraq', 'ug': 'ئىراق', 'uk': 'Ірак', 'ur': 'عراق', 'uz': 'Iroq', 'uz-Cyrl': 'Ироқ', 'uz-Latn': 'Iroq', 'vai': 'ꔤꕟꕃ', 'vai-Latn': 'Iraki', 'vai-Vaii': 'ꔤꕟꕃ', 'vec': 'Irac', 'vi': 'Iraq', 'vun': 'Iraki', 'wae': 'Irak', 'wo': 'Irag', 'xh': 'E-Iraq', 'xnr': 'इराक', 'xog': 'Yiraaka', 'yav': 'ilák', 'yo': 'Iraki', 'yrl': 'Iraki', 'yue': '伊拉克', 'yue-Hans': '伊拉克', 'yue-Hant': '伊拉克', 'zgh': 'ⵍⵄⵉⵔⴰⵇ', 'zh': '伊拉克', 'zh-Hans': '伊拉克', 'zh-Hant': '伊拉克', 'zu': 'i-Iraq'},
    'IR': {'af': 'Iran', 'agq': 'Èlân', 'ak': 'Iran', 'am': 'ኢራን', 'ar': 'إيران', 'as': 'ইৰান', 'asa': 'Uajemi', 'ast': 'Irán', 'az': 'İran', 'az-Cyrl': 'Иран', 'az-Latn': 'İran', 'bal-Latn': 'Érán', 'bas': 'Ìrâŋ', 'be': 'Іран', 'be-tarask': 'Іран', 'bez': 'Huuajemi', 'bg': 'Иран', 'blo': 'Iraŋ', 'bm': 'Iraŋ', 'bn': 'ইরান', 'br': 'Iran', 'brx': 'ईरान', 'bs': 'Iran', 'bs-Cyrl': 'Иран', 'bs-Latn': 'Iran', 'ca': 'Iran', 'ca-ES-valencia': 'Iran', 'ccp': '𑄄𑄢𑄚𑄴', 'ce': 'ГӀажарийчоь', 'ceb': 'Iran', 'cgg': 'Iraani', 'chr': 'ᎢᎴᏂ', 'ckb': 'ئێران', 'cs': 'Írán', 'cv': 'Иран', 'cy': 'Iran', 'da': 'Iran', 'dav': 'Uajemi', 'de': 'Iran', 'dje': 'Iraan', 'dsb': 'Iran', 'dyo': 'Iran', 'dz': 'ཨི་རཱན', 'ebu': 'Uajemi', 'ee': 'Iran nutome', 'el': 'Ιράν', 'el-polyton': 'Ἰράν, Ἰσλαμικὴ Δημοκρατία τοῦ', 'en': 'Iran', 'en-Dsrt': '𐐆𐑉𐐪𐑌', 'eo': 'Irano', 'es': 'Irán', 'et': 'Iraan', 'eu': 'Iran', 'ewo': 'Irán', 'fa': 'ایران', 'ff': 'Iraan', 'ff-Adlm': '𞤋𞤪𞤢𞥄𞤲', 'ff-Latn': 'Iraan', 'fi': 'Iran', 'fil': 'Iran', 'fo': 'Iran', 'fr': 'Iran', 'fur': 'Iran', 'fy': 'Iran', 'ga': 'an Iaráin', 'gd': 'Ioràn', 'gl': 'Irán', 'gsw': 'Iraan', 'gu': 'ઈરાન', 'guz': 'Uajemi', 'ha': 'Iran', 'he': 'איראן', 'hi': 'ईरान', 'hi-Latn': 'Iran', 'hr': 'Iran', 'hsb': 'Iran', 'ht': 'Iran', 'hu': 'Irán', 'hy': 'Իրան', 'ia': 'Iran', 'id': 'Iran', 'ie': 'Iran', 'ig': 'Iran', 'is': 'Íran', 'it': 'Iran', 'ja': 'イラン', 'jmc': 'Uajemi', 'jv': 'Iran', 'ka': 'ირანი', 'kaa': 'Иран', 'kaa-Cyrl': 'Иран', 'kab': 'Iran', 'kam': 'Uajemi', 'kde': 'Uadyemi', 'kea': 'Irãu', 'kgp': 'Irỹ', 'khq': 'Iraan', 'ki': 'Uajemi', 'kk': 'Иран', 'kk-Cyrl': 'Иран', 'kln': 'Emetab Iran', 'km': 'អ៊ីរ៉ង់', 'kn': 'ಇರಾನ್', 'ko': '이란', 'kok': 'इरान', 'kok-Deva': 'इरान', 'ks': 'ایٖران', 'ks-Arab': 'ایٖران', 'ksb': 'Uajemi', 'ksf': 'iráŋ', 'ksh': 'Persije', 'ku': 'Îran', 'kxv': 'iran', 'kxv-Deva': 'ईरान', 'kxv-Latn': 'iran', 'kxv-Orya': 'ଇରାନ', 'kxv-Telu': 'ఇరాన్', 'ky': 'Иран', 'lag': 'Ʉajéemi', 'lb': 'Iran', 'lg': 'Yiraani', 'lld': 'Iran', 'ln': 'Irâ', 'lo': 'ອີຣານ', 'lt': 'Iranas', 'lu': 'Ira', 'luo': 'Iran', 'luy': 'Uajemi', 'lv': 'Irāna', 'mai': 'ईरान', 'mas': 'Uajemi', 'mer': 'Irani', 'mfe': 'Iran', 'mg': 'Iran', 'mi': 'Irāna', 'mk': 'Иран', 'ml': 'ഇറാൻ', 'mn': 'Иран', 'mr': 'इराण', 'ms': 'Iran', 'mt': 'l-Iran', 'mua': 'Iraŋ', 'my': 'အီရန်', 'mzn': 'ایران', 'naq': 'Iranni', 'nb': 'Iran', 'nd': 'Iran', 'ne': 'इरान', 'nl': 'Iran', 'nmg': 'Iran', 'nn': 'Iran', 'no': 'Iran', 'nqo': 'ߌߙߊ߲߫', 'nyn': 'Iraani', 'om': 'Iraan', 'or': 'ଇରାନ', 'pa': 'ਈਰਾਨ', 'pa-Guru': 'ਈਰਾਨ', 'pcm': 'Irán', 'pl': 'Iran', 'ps': 'ايران', 'pt': 'Irã', 'qu': 'Irán', 'rm': 'Iran', 'rn': 'Irani', 'ro': 'Iran', 'rof': 'Uajemi', 'ru': 'Иран', 'rwk': 'Uajemi', 'saq': 'Uajemi', 'sbp': 'Uwajemi', 'sc': 'Iràn', 'sd': 'ايران', 'sd-Arab': 'ايران', 'se': 'Iran', 'seh': 'Irã', 'ses': 'Iraan', 'sg': 'Iräan', 'shi': 'ⵉⵔⴰⵏ', 'shi-Latn': 'iran', 'shi-Tfng': 'ⵉⵔⴰⵏ', 'si': 'ඉරානය', 'sk': 'Irán', 'sl': 'Iran', 'smn': 'Iran', 'sn': 'Iran', 'so': 'Iiraan', 'sq': 'Iran', 'sr': 'Иран', 'sr-Cyrl': 'Иран', 'sr-Latn': 'Iran', 'sv': 'Iran', 'sw': 'Iran', 'syr': 'ܐܝܪܐܢ', 'ta': 'ஈரான்', 'te': 'ఇరాన్', 'teo': 'Uajemi', 'tg': 'Эрон', 'th': 'อิหร่าน', 'ti': 'ኢራን', 'tk': 'Eýran', 'to': 'ʻIlaani', 'tr': 'İran', 'tt': 'Иран', 'twq': 'Iraan', 'tzm': 'Iran', 'ug': 'ئىران', 'uk': 'Іран', 'ur': 'ایران', 'uz': 'Eron', 'uz-Cyrl': 'Эрон', 'uz-Latn': 'Eron', 'vai': 'ꔤꕟꘋ', 'vai-Latn': 'Iraŋ', 'vai-Vaii': 'ꔤꕟꘋ', 'vec': 'Iran', 'vi': 'Iran', 'vun': 'Uajemi', 'wae': 'Iran', 'wo': 'Iraŋ', 'xh': 'E-Iran', 'xnr': 'ईरान', 'xog': 'Yiraani', 'yav': 'iláŋ', 'yi': 'איראַן', 'yo': 'Irani', 'yrl': 'Irã', 'yue': '伊朗', 'yue-Hans': '伊朗', 'yue-Hant': '伊朗', 'zgh': 'ⵉⵔⴰⵏ', 'zh': '伊朗', 'zh-Hans': '伊朗', 'zh-Hant': '伊朗', 'zu': 'i-Iran'},
    'IS': {'af': 'Ysland', 'agq': 'Aesɨ̀lân', 'ak': 'Aesland', 'am': 'አይስላንድ', 'ar': 'آيسلندا', 'as': 'আইচলেণ্ড', 'asa': 'Aithlandi', 'ast': 'Islandia', 'az': 'İslandiya', 'az-Cyrl': 'Исландија', 'az-Latn': 'İslandiya', 'bal-Latn': 'Áeslánd', 'bas': 'Ìslandìà', 'be': 'Ісландыя', 'be-tarask': 'Ісландыя', 'bez': 'Huaislandi', 'bg': 'Исландия', 'blo': 'Islanɖ', 'bm': 'Isilandi', 'bn': 'আইসল্যান্ড', 'br': 'Island', 'brx': 'आइसलेण्ड', 'bs': 'Island', 'bs-Cyrl': 'Исланд', 'bs-Latn': 'Island', 'ca': 'Islàndia', 'ca-ES-valencia': 'Islàndia', 'ccp': '𑄃𑄭𑄥𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴', 'ce': 'Исланди', 'ceb': 'Iceland', 'cgg': 'Aisilandi', 'chr': 'ᏧᏁᏍᏓᎸᎯ', 'ckb': 'ئایسلەند', 'cs': 'Island', 'cv': 'Исланди', 'cy': 'Gwlad yr Iâ', 'da': 'Island', 'dav': 'Aislandi', 'de': 'Island', 'dje': 'Ayseland', 'dsb': 'Islandska', 'dyo': 'Iisland', 'dz': 'ཨཱའིས་ལེནཌ', 'ebu': 'Aislandi', 'ee': 'Aiseland nutome', 'el': 'Ισλανδία', 'el-polyton': 'Ἰσλανδία', 'en': 'Iceland', 'en-Dsrt': '𐐌𐑅𐑊𐐲𐑌𐐼', 'eo': 'Islando', 'es': 'Islandia', 'et': 'Island', 'eu': 'Islandia', 'ewo': 'Isəlándə', 'fa': 'ایسلند', 'ff': 'Islannda', 'ff-Adlm': '𞤀𞤴𞤧𞤵𞤤𞤢𞤲𞤣', 'ff-Latn': 'Islannda', 'fi': 'Islanti', 'fil': 'Iceland', 'fo': 'Ísland', 'fr': 'Islande', 'fur': 'Islande', 'fy': 'Yslân', 'ga': 'an Íoslainn', 'gd': 'Innis Tìle', 'gl': 'Islandia', 'gsw': 'Iisland', 'gu': 'આઇસલેન્ડ', 'guz': 'Aislandi', 'ha': 'Aisalan', 'he': 'איסלנד', 'hi': 'आइसलैंड', 'hi-Latn': 'Iceland', 'hr': 'Island', 'hsb': 'Islandska', 'ht': 'Islande', 'hu': 'Izland', 'hy': 'Իսլանդիա', 'ia': 'Islanda', 'id': 'Islandia', 'ie': 'Island', 'ig': 'Iceland', 'is': 'Ísland', 'it': 'Islanda', 'ja': 'アイスランド', 'jmc': 'Aislandi', 'jv': 'Èslan', 'ka': 'ისლანდია', 'kaa': 'Исландия', 'kaa-Cyrl': 'Исландия', 'kab': 'Island', 'kam': 'Aislandi', 'kde': 'Aislandi', 'kea': 'Islándia', 'kgp': 'Inhrỹnija', 'khq': 'Ayceland', 'ki': 'Aislandi', 'kk': 'Исландия', 'kk-Cyrl': 'Исландия', 'kln': 'Emetab Iceland', 'km': 'អ៊ីស្លង់', 'kn': 'ಐಸ್\u200cಲ್ಯಾಂಡ್', 'ko': '아이슬란드', 'kok': 'आइसलैंड', 'kok-Deva': 'आइसलैंड', 'ks': 'اَیِسلینڑ', 'ks-Arab': 'اَیِسلینڑ', 'ksb': 'Aislandi', 'ksf': 'zǝ i glás', 'ksh': 'Ißland', 'ku': 'Îslanda', 'kxv': 'aislanḍ', 'kxv-Deva': 'आइसलेंड', 'kxv-Latn': 'aislanḍ', 'kxv-Orya': 'ଆଇସଲାଣ୍ଡ', 'kxv-Telu': 'ఐస్లాండ్', 'ky': 'Исландия', 'lag': 'Aisiláandi', 'lb': 'Island', 'lg': 'Ayisirandi', 'lld': 'Islanda', 'ln': 'Isilandɛ', 'lo': 'ໄອສແລນ', 'lt': 'Islandija', 'lu': 'Isilande', 'luo': 'Iceland', 'luy': 'Aislandi', 'lv': 'Islande', 'mai': 'आइसलैंड', 'mas': 'Aislandi', 'mer': 'Aisilandi', 'mfe': 'Island', 'mg': 'Islandy', 'mi': 'Tiorangi', 'mk': 'Исланд', 'ml': 'ഐസ്\u200cലാന്റ്', 'mn': 'Исланд', 'mr': 'आइसलँड', 'ms': 'Iceland', 'mt': 'l-Iżlanda', 'mua': 'Sǝr Island', 'my': 'အိုက်စလန်', 'mzn': 'ایسلند', 'naq': 'Iceland', 'nb': 'Island', 'nd': 'Iceland', 'ne': 'आइस्ल्याण्ड', 'nl': 'IJsland', 'nmg': 'Island', 'nn': 'Island', 'no': 'Island', 'nqo': 'ߌߛߑߟߊߘ', 'nyn': 'Aisilandi', 'om': 'Ayeslaand', 'or': 'ଆଇସଲ୍ୟାଣ୍ଡ', 'pa': 'ਆਈਸਲੈਂਡ', 'pa-Guru': 'ਆਈਸਲੈਂਡ', 'pcm': 'Aísland', 'pl': 'Islandia', 'ps': 'آیسلینډ', 'pt': 'Islândia', 'qu': 'Islandia', 'rm': 'Islanda', 'rn': 'Ayisilandi', 'ro': 'Islanda', 'rof': 'Aislandi', 'ru': 'Исландия', 'rwk': 'Aislandi', 'sah': 'Исландия', 'saq': 'Aislandi', 'sbp': 'Ayisilendi', 'sc': 'Islanda', 'sd': 'آئس لينڊ', 'sd-Arab': 'آئس لينڊ', 'se': 'Islánda', 'seh': 'Islândia', 'ses': 'Ayceland', 'sg': 'Islânde', 'shi': 'ⵉⵙⵍⴰⵏⴷ', 'shi-Latn': 'island', 'shi-Tfng': 'ⵉⵙⵍⴰⵏⴷ', 'si': 'අයිස්ලන්තය', 'sk': 'Island', 'sl': 'Islandija', 'smn': 'Island', 'sn': 'Iceland', 'so': 'Ayslaand', 'sq': 'Islandë', 'sr': 'Исланд', 'sr-Cyrl': 'Исланд', 'sr-Latn': 'Island', 'sv': 'Island', 'sw': 'Aisilandi', 'syr': 'ܐܝܣܠܢܕ', 'ta': 'ஐஸ்லாந்து', 'te': 'ఐస్లాండ్', 'teo': 'Aislandi', 'tg': 'Исландия', 'th': 'ไอซ์แลนด์', 'ti': 'ኣይስላንድ', 'tk': 'Islandiýa', 'to': 'ʻAisilani', 'tr': 'İzlanda', 'tt': 'Исландия', 'twq': 'Ayseland', 'tzm': 'Islanḍa', 'ug': 'ئىسلاندىيە', 'uk': 'Ісландія', 'ur': 'آئس لینڈ', 'uz': 'Islandiya', 'uz-Cyrl': 'Исландия', 'uz-Latn': 'Islandiya', 'vai': 'ꕉꔤꔻ ꖨꕮꕊ', 'vai-Latn': 'Áisi Lumaã', 'vai-Vaii': 'ꕉꔤꔻ ꖨꕮꕊ', 'vec': 'Izlanda', 'vi': 'Iceland', 'vun': 'Aislandi', 'wae': 'Island', 'wo': 'Islànd', 'xh': 'E-Iceland', 'xnr': 'आइसलैंड', 'xog': 'Ayisirandi', 'yav': 'isláand', 'yi': 'איסלאַנד', 'yo': 'Aṣilandi', 'yrl': 'Isirãdiya', 'yue': '冰島', 'yue-Hans': '冰岛', 'yue-Hant': '冰島', 'zgh': 'ⵉⵙⵍⴰⵏⴷ', 'zh': '冰岛', 'zh-Hans': '冰岛', 'zh-Hant': '冰島', 'zu': 'i-Iceland'},
    'IT': {'af': 'Italië', 'agq': 'Etalè', 'ak': 'Itali', 'am': 'ጣሊያን', 'ar': 'إيطاليا', 'as': 'ইটালি', 'asa': 'Italia', 'ast': 'Italia', 'az': 'İtaliya', 'az-Cyrl': 'Италија', 'az-Latn': 'İtaliya', 'bal-Latn': 'Itáliá', 'bas': 'Ìtalìà', 'be': 'Італія', 'be-tarask': 'Італія', 'bez': 'Huitalia', 'bg': 'Италия', 'bgn': 'ایتالیه', 'blo': 'Italii', 'bm': 'Itali', 'bn': 'ইতালি', 'bo': 'ཨི་ཀྲར་ལི་', 'br': 'Italia', 'brx': 'इटाली', 'bs': 'Italija', 'bs-Cyrl': 'Италија', 'bs-Latn': 'Italija', 'ca': 'Itàlia', 'ca-ES-valencia': 'Itàlia', 'ccp': '𑄄𑄖𑄣𑄨', 'ce': 'Итали', 'ceb': 'Italya', 'cgg': 'Itare', 'chr': 'ᎢᏔᎵ', 'ckb': 'ئیتالیا', 'cs': 'Itálie', 'cv': 'Итали', 'cy': 'Yr Eidal', 'da': 'Italien', 'dav': 'Italia', 'de': 'Italien', 'dje': 'Itaali', 'doi': 'इटली', 'dsb': 'Italska', 'dyo': 'Itali', 'dz': 'ཨི་ཊ་ལི', 'ebu': 'Italia', 'ee': 'Italia nutome', 'el': 'Ιταλία', 'el-polyton': 'Ἰταλία', 'en': 'Italy', 'en-Dsrt': '𐐆𐐻𐐲𐑊𐐨', 'eo': 'Italujo', 'es': 'Italia', 'et': 'Itaalia', 'eu': 'Italia', 'ewo': 'Itáliɛn', 'fa': 'ایتالیا', 'ff': 'Itali', 'ff-Adlm': '𞤋𞤼𞤢𞤤𞤭𞥅', 'ff-Latn': 'Itali', 'fi': 'Italia', 'fil': 'Italy', 'fo': 'Italia', 'fr': 'Italie', 'fur': 'Italie', 'fy': 'Italië', 'ga': 'an Iodáil', 'gd': 'An Eadailt', 'gl': 'Italia', 'gsw': 'Itaalie', 'gu': 'ઇટાલી', 'guz': 'Italia', 'ha': 'Italiya', 'haw': 'ʻĪkālia', 'he': 'איטליה', 'hi': 'इटली', 'hi-Latn': 'Italy', 'hr': 'Italija', 'hsb': 'Italska', 'ht': 'Italie', 'hu': 'Olaszország', 'hy': 'Իտալիա', 'ia': 'Italia', 'id': 'Italia', 'ie': 'Italia', 'ig': 'Italy', 'ii': 'ꑴꄊꆺ', 'is': 'Ítalía', 'it': 'Italia', 'ja': 'イタリア', 'jgo': 'Italî', 'jmc': 'Italia', 'jv': 'Itali', 'ka': 'იტალია', 'kaa': 'Италия', 'kaa-Cyrl': 'Италия', 'kab': 'Ṭelyan', 'kam': 'Italia', 'kde': 'Italia', 'kea': 'Itália', 'kgp': 'Itarija', 'khq': 'Itaali', 'ki': 'Italia', 'kk': 'Италия', 'kk-Cyrl': 'Италия', 'kln': 'Emetab Italy', 'km': 'អ៊ីតាលី', 'kn': 'ಇಟಲಿ', 'ko': '이탈리아', 'kok': 'इटली', 'kok-Deva': 'इटली', 'kok-Latn': 'Italia', 'ks': 'اِٹلی', 'ks-Arab': 'اِٹلی', 'ks-Deva': 'इटली', 'ksb': 'Italia', 'ksf': 'italí', 'ksh': 'Itaalije', 'ku': 'Îtalya', 'kxv': 'iṭalī', 'kxv-Deva': 'इटली', 'kxv-Latn': 'iṭalī', 'kxv-Orya': 'ଇଟାଲି', 'kxv-Telu': 'ఇటలి', 'ky': 'Италия', 'lag': 'Itália', 'lb': 'Italien', 'lg': 'Yitale', 'lij': 'Italia', 'lld': 'Talia', 'lmo': 'Italia', 'ln': 'Itali', 'lo': 'ອິຕາລີ', 'lrc': 'ئیتالیا', 'lt': 'Italija', 'lu': 'Itali', 'luo': 'Italy', 'luy': 'Italia', 'lv': 'Itālija', 'mai': 'इटली', 'mas': 'Italia', 'mer': 'Italĩ', 'mfe': 'Itali', 'mg': 'Italia', 'mgh': 'Italia', 'mi': 'Itāria', 'mk': 'Италија', 'ml': 'ഇറ്റലി', 'mn': 'Итали', 'mni': 'ইটালি', 'mni-Beng': 'ইটালি', 'mr': 'इटली', 'ms': 'Itali', 'mt': 'l-Italja', 'mua': 'Italiya', 'my': 'အီတလီ', 'mzn': 'ایتالیا', 'naq': 'Italiab', 'nb': 'Italia', 'nd': 'Itali', 'ne': 'इटली', 'nl': 'Italië', 'nmg': 'Italia', 'nn': 'Italia', 'no': 'Italia', 'nqo': 'ߌߕߊߟߌ߫', 'nyn': 'Itare', 'om': 'Xaaliyaan', 'or': 'ଇଟାଲୀ', 'os': 'Итали', 'pa': 'ਇਟਲੀ', 'pa-Guru': 'ਇਟਲੀ', 'pcm': 'Ítáli', 'pl': 'Włochy', 'ps': 'ایټالیه', 'pt': 'Itália', 'qu': 'Italia', 'rm': 'Italia', 'rn': 'Ubutaliyani', 'ro': 'Italia', 'rof': 'Italia', 'ru': 'Италия', 'rwk': 'Italia', 'sa': 'इटली:', 'saq': 'Italia', 'sat': 'ᱤᱴᱞᱤ', 'sat-Olck': 'ᱤᱴᱞᱤ', 'sbp': 'Italiya', 'sc': 'Itàlia', 'sd': 'اٽلي', 'sd-Arab': 'اٽلي', 'sd-Deva': 'इटली', 'se': 'Itália', 'seh': 'Itália', 'ses': 'Itaali', 'sg': 'Italùii', 'shi': 'ⵉⵟⴰⵍⵢⴰ', 'shi-Latn': 'iṭalya', 'shi-Tfng': 'ⵉⵟⴰⵍⵢⴰ', 'si': 'ඉතාලිය', 'sk': 'Taliansko', 'sl': 'Italija', 'smn': 'Italia', 'sn': 'Italy', 'so': 'Talyaani', 'sq': 'Itali', 'sr': 'Италија', 'sr-Cyrl': 'Италија', 'sr-Latn': 'Italija', 'su': 'Italia', 'su-Latn': 'Italia', 'sv': 'Italien', 'sw': 'Italia', 'syr': 'ܐܝܛܠܝܐ', 'ta': 'இத்தாலி', 'te': 'ఇటలీ', 'teo': 'Italia', 'tg': 'Италия', 'th': 'อิตาลี', 'ti': 'ኢጣልያ', 'tk': 'Italiýa', 'to': 'ʻĪtali', 'tpi': 'Itali', 'tr': 'İtalya', 'tt': 'Италия', 'twq': 'Itaali', 'tzm': 'Iṭalya', 'ug': 'ئىتالىيە', 'uk': 'Італія', 'ur': 'اٹلی', 'uz': 'Italiya', 'uz-Cyrl': 'Италия', 'uz-Latn': 'Italiya', 'vai': 'ꔤꕚꔷ', 'vai-Latn': 'Ítali', 'vai-Vaii': 'ꔤꕚꔷ', 'vec': 'Italia', 'vi': 'Italy', 'vun': 'Italia', 'wae': 'Italie', 'wo': 'Itali', 'xh': 'E-Italy', 'xnr': 'इटली', 'xog': 'Yitale', 'yav': 'italí', 'yi': 'איטאַליע', 'yo': 'Itáli', 'yrl': 'Itariya', 'yue': '意大利', 'yue-Hans': '义大利', 'yue-Hant': '意大利', 'zgh': 'ⵉⵟⴰⵍⵢⴰ', 'zh': '意大利', 'zh-Hans': '意大利', 'zh-Hant': '義大利', 'zu': 'i-Italy'},
    'Inds': {'ar': 'اندس - هارابان', 'ast': 'indus', 'az': 'hindistan', 'az-Latn': 'hindistan', 'bg': 'Харапска', 'bn': 'সিন্ধু', 'brx': 'सिन्धु', 'bs': 'induško pismo', 'bs-Cyrl': 'индушко писмо', 'bs-Latn': 'induško pismo', 'ca': 'escriptura de la vall de l’Indus', 'ca-ES-valencia': 'escriptura de la vall de l’Indus', 'ccp': '𑄥𑄨𑄚𑄴𑄙𑄪', 'cs': 'harappské', 'da': 'indus', 'de': 'Indus-Schrift', 'el': 'Ίνδους', 'el-polyton': 'Ίνδους', 'en': 'Indus', 'en-Dsrt': '𐐆𐑌𐐼𐐲𐑅', 'es': 'Indio (harappan)', 'et': 'Induse', 'fa': 'ایندوس', 'fi': 'induslainen', 'fr': 'indus', 'fy': 'Indus', 'gsw': 'Indus-Schrift', 'gu': 'સિન્ધુ', 'he': 'אינדוס', 'hi': 'सिन्धु', 'hi-Latn': 'Sindhu', 'hr': 'indijsko pismo', 'ht': 'indus', 'hu': 'Indus', 'id': 'Indus', 'it': 'indu', 'ja': 'インダス文字', 'kgp': 'ĩnu', 'kn': 'ಸಿಂಧೂ', 'ko': '인더스 문자', 'ks': 'اِنڈَس', 'ks-Arab': 'اِنڈَس', 'lb': 'Indus-Schrëft', 'lo': 'ອິນດັດ', 'lt': 'indus', 'mk': 'харапско', 'ml': 'സിന്ധു', 'mr': 'सिन्धु', 'nb': 'indus', 'ne': 'इन्दुस', 'nl': 'Indus', 'nn': 'indus', 'no': 'indus', 'or': 'ସିନ୍ଧୁ', 'pl': 'indus', 'pt': 'indo', 'rm': 'indus', 'ro': 'indus', 'ru': 'хараппская (письменность долины Инда)', 'sl': 'induški', 'sr': 'индушко писмо', 'sr-Cyrl': 'индушко писмо', 'sr-Latn': 'induško pismo', 'sv': 'indus', 'ta': 'சிந்து', 'te': 'సింధు', 'th': 'อินดัส', 'to': 'tohinima fakaʻinitusi', 'tr': 'Indus', 'ug': 'ئىندۇسچە', 'uk': 'харапський', 'vi': 'Chữ Indus', 'yrl': 'ĩdu', 'yue': '印度河流域（哈拉帕文）', 'yue-Hans': '印度河流域（哈拉帕文）', 'yue-Hant': '印度河流域（哈拉帕文）', 'zh': '印度河文字', 'zh-Hans': '印度河文字', 'zh-Hant': '印度河流域（哈拉帕文）'},
    'Ital': {'ar': 'الإيطالية القديمة', 'ast': 'itálicu antiguu', 'az': 'qədimi italyalı', 'az-Latn': 'qədimi italyalı', 'bal-Latn': 'Ahdi Itáliái', 'bg': 'Древно италийска', 'bn': 'প্রাচীন ইতালি', 'br': 'henitalek', 'brx': 'पुरानी इटैलियन', 'bs': 'staro italsko pismo', 'bs-Cyrl': 'стари италик', 'bs-Latn': 'staro italsko pismo', 'ca': 'cursiva antiga', 'ca-ES-valencia': 'cursiva antiga', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄩 𑄃𑄨𑄖𑄣𑄨', 'cs': 'etruské', 'da': 'Olditalisk', 'de': 'Altitalisch', 'el': 'Παλαιό Ιταλικό', 'el-polyton': 'Παλαιό Ιταλικό', 'en': 'Old Italic', 'en-Dsrt': '𐐄𐑊𐐼 𐐆𐐻𐐰𐑊𐐮𐐿', 'es': 'antigua bastardilla', 'et': 'vanaitali', 'eu': 'italiera zaharra', 'fa': 'ایتالی باستان', 'ff-Adlm': '𞤑𞤭𞤯𞤭 𞤋𞤼𞤢𞤤𞤭𞤳', 'fi': 'muinaisitalialainen', 'fo': 'gamalt italienskt', 'fr': 'ancien italique', 'fur': 'vieri italic', 'fy': 'Ald-italysk', 'ga': 'Sean-Iodáilic', 'gd': 'Seann-Eadailtis', 'gsw': 'Altitalisch', 'gu': 'જૂનુ ઇટાલિક', 'he': 'איטלקי עתיק', 'hi': 'पुरानी इटली', 'hi-Latn': 'Old Italic', 'hr': 'staro talijansko pismo', 'ht': 'ancien italique', 'hu': 'Régi olasz', 'id': 'Italia Lama', 'ig': 'Old Italic', 'it': 'italico antico', 'ja': '古イタリア文字', 'kgp': 'itariko si', 'kk': 'ескі италия жазуы', 'kk-Cyrl': 'ескі италия жазуы', 'kn': 'ಪ್ರಾಚೀನ್ ಇಟಾಲಿಕ್', 'ko': '고대 이탈리아 문자', 'ks': 'اولڈ اِٹیلِک', 'ks-Arab': 'اولڈ اِٹیلِک', 'lb': 'Alitalesch', 'lo': 'ອີຕາລີໂບຮານ', 'lt': 'senasis italų', 'lv': 'vecitāļu', 'mk': 'староиталско', 'ml': 'പഴയ ഇറ്റാലിയൻ', 'mn': 'Хуучин итали', 'mr': 'जुनी इटालिक', 'ms': 'Italik Lama', 'nb': 'gammelitalisk', 'ne': 'पुरानो इटालिक', 'nl': 'Oud-italisch', 'nn': 'gammalitalisk', 'no': 'gammelitalisk', 'nqo': 'ߌߕߊߟߌߞ ߞߘߐ߬ߡߊ߲', 'or': 'ପୁରୁଣା ଇଟାଲୀ', 'pl': 'starowłoskie', 'pt': 'itálico antigo', 'rm': 'italic vegl', 'ro': 'italică veche', 'ru': 'староитальянская', 'sc': 'itàlicu antigu', 'sd': 'قديم اطالوي', 'sd-Arab': 'قديم اطالوي', 'sl': 'staroitalski', 'so': 'Itaaliggii Hore', 'sq': 'italishte e vjetër', 'sr': 'стари италик', 'sr-Cyrl': 'стари италик', 'sr-Latn': 'stari italik', 'sv': 'fornitaliska', 'ta': 'பழைய இத்தாலி', 'te': 'ప్రాచిన ఐటాలిక్', 'tg': 'Курсиви Куҳна', 'th': 'อิตาลีโบราณ', 'to': 'tohinima fakaʻītali-motuʻa', 'tr': 'Eski İtalyan', 'ug': 'قەدىمكى ئىتاليانچە', 'uk': 'давньоіталійський', 'vec': 'itàlego antigo', 'vi': 'Chữ Italic cổ', 'xnr': 'पराणी इटली', 'yrl': 'itáriku kuxiímawara', 'yue': '古意大利文', 'yue-Hans': '古意大利文', 'yue-Hant': '古意大利文', 'zh': '古意大利文', 'zh-Hans': '古意大利文', 'zh-Hant': '古意大利文', 'zu': 'i-Old Italic'},
    'JE': {'af': 'Jersey', 'ak': 'Gyɛsi', 'am': 'ጀርዚ', 'ar': 'جيرسي', 'as': 'জাৰ্চি', 'ast': 'Jersey', 'az': 'Cersi', 'az-Cyrl': 'Ҹерси', 'az-Latn': 'Cersi', 'bal-Latn': 'Jersé', 'be': 'Джэрсі', 'be-tarask': 'Джэрсі', 'bg': 'Джърси', 'blo': 'Jersei', 'bn': 'জার্সি', 'br': 'Jerzenez', 'brx': 'जोर्सि', 'bs': 'Jersey', 'bs-Cyrl': 'Џерзи', 'bs-Latn': 'Jersey', 'ca': 'Jersey', 'ca-ES-valencia': 'Jersey', 'ccp': '𑄎𑄢𑄴𑄥𑄨', 'ce': 'Джерси', 'ceb': 'Jersey', 'chr': 'ᏨᎵᏏ', 'ckb': 'جێرسی', 'cs': 'Jersey', 'cv': 'Джерси', 'cy': 'Jersey', 'da': 'Jersey', 'de': 'Jersey', 'dsb': 'Jersey', 'dz': 'ཇེར་སི', 'ee': 'Dzɛse nutome', 'el': 'Τζέρζι', 'el-polyton': 'Τζέρζι', 'en': 'Jersey', 'en-Dsrt': '𐐖𐐲𐑉𐑆𐐨', 'es': 'Jersey', 'et': 'Jersey', 'eu': 'Jersey', 'fa': 'جرزی', 'ff-Adlm': '𞤔𞤫𞤪𞤧𞤭𞥅', 'fi': 'Jersey', 'fil': 'Jersey', 'fo': 'Jersey', 'fr': 'Jersey', 'fur': 'Jersey', 'fy': 'Jersey', 'ga': 'Geirsí', 'gd': 'Deàrsaidh', 'gl': 'Jersey', 'gsw': 'Dschörsi', 'gu': 'જર્સી', 'ha': 'Kasar Jersey', 'he': 'ג׳רזי', 'hi': 'जर्सी', 'hi-Latn': 'Jersey', 'hr': 'Jersey', 'hsb': 'Jersey', 'ht': 'Jersey', 'hu': 'Jersey', 'hy': 'Ջերսի', 'ia': 'Jersey', 'id': 'Jersey', 'ig': 'Jersey', 'is': 'Jersey', 'it': 'Jersey', 'ja': 'ジャージー', 'jv': 'Jersey', 'ka': 'ჯერსი', 'kaa': 'Джерси', 'kaa-Cyrl': 'Джерси', 'kea': 'Jersi', 'kgp': 'Jérsej', 'kk': 'Джерси', 'kk-Cyrl': 'Джерси', 'km': 'ជើស៊ី', 'kn': 'ಜೆರ್ಸಿ', 'ko': '저지', 'kok': 'जर्सी', 'kok-Deva': 'जर्सी', 'ks': 'جٔرسی', 'ks-Arab': 'جٔرسی', 'ksh': 'Jööseh', 'ku': 'Jersey', 'kxv': 'jersī', 'kxv-Deva': 'जर्सी', 'kxv-Latn': 'jersī', 'kxv-Orya': 'ଜରସି', 'kxv-Telu': 'జెర్సీ', 'ky': 'Жерси', 'lb': 'Jersey', 'lld': 'Jersey', 'lo': 'ເຈີຊີ', 'lt': 'Džersis', 'lv': 'Džērsija', 'mai': 'जरसी', 'mi': 'Tōrehe', 'mk': 'Џерси', 'ml': 'ജേഴ്സി', 'mn': 'Жерси', 'mr': 'जर्सी', 'ms': 'Jersey', 'mt': 'Jersey', 'my': 'ဂျာစီ', 'mzn': 'جرسی', 'nb': 'Jersey', 'ne': 'जर्सी', 'nl': 'Jersey', 'nn': 'Jersey', 'no': 'Jersey', 'nqo': 'ߖߋߙߑߖ߭ߌ߫', 'om': 'Jeersii', 'or': 'ଜର୍ସି', 'pa': 'ਜਰਸੀ', 'pa-Guru': 'ਜਰਸੀ', 'pcm': 'Jẹ́si', 'pl': 'Jersey', 'ps': 'جرسی', 'pt': 'Jersey', 'qu': 'Jersey', 'rm': 'Jersey', 'ro': 'Jersey', 'ru': 'Джерси', 'sc': 'Jersey', 'sd': 'جرسي', 'sd-Arab': 'جرسي', 'se': 'Jersey', 'si': 'ජර්සි', 'sk': 'Jersey', 'sl': 'Jersey', 'smn': 'Jersey', 'so': 'Jaarsey', 'sq': 'Xhersej', 'sr': 'Џерзи', 'sr-Cyrl': 'Џерзи', 'sr-Latn': 'Džerzi', 'sv': 'Jersey', 'sw': 'Jersey', 'syr': 'ܓܝܪܙܝ', 'ta': 'ஜெர்சி', 'te': 'జెర్సీ', 'tg': 'Ҷерси', 'th': 'เจอร์ซีย์', 'ti': 'ጀርዚ', 'tk': 'Jersi', 'to': 'Selusī', 'tr': 'Jersey', 'tt': 'Джерси', 'ug': 'جېرسېي', 'uk': 'Джерсі', 'ur': 'جرسی', 'uz': 'Jersi', 'uz-Cyrl': 'Жерси', 'uz-Latn': 'Jersi', 'vai': 'ꘀꗡꔘ', 'vai-Vaii': 'ꘀꗡꔘ', 'vec': 'Jersey', 'vi': 'Jersey', 'wae': 'Jersey', 'wo': 'Serse', 'xh': 'EJersey', 'xnr': 'जर्सी', 'yi': 'דזשערזי', 'yo': 'Jẹsì', 'yrl': 'Yesei', 'yue': '澤西島', 'yue-Hans': '泽西岛', 'yue-Hant': '澤西島', 'zh': '泽西岛', 'zh-Hans': '泽西岛', 'zh-Hant': '澤西島', 'zu': 'i-Jersey'},
    'JM': {'af': 'Jamaika', 'agq': 'Dzàmɛkà', 'ak': 'Gyameka', 'am': 'ጃማይካ', 'ar': 'جامايكا', 'as': 'জামাইকা', 'asa': 'Jamaika', 'ast': 'Xamaica', 'az': 'Yamayka', 'az-Cyrl': 'Јамајка', 'az-Latn': 'Yamayka', 'bal-Latn': 'Jamáeká', 'bas': 'Jàmàikà', 'be': 'Ямайка', 'be-tarask': 'Ямайка', 'bez': 'Hujamaika', 'bg': 'Ямайка', 'blo': 'Jamaɩka', 'bm': 'Zamayiki', 'bn': 'জামাইকা', 'br': 'Jamaika', 'brx': 'जामाइका', 'bs': 'Jamajka', 'bs-Cyrl': 'Јамајка', 'bs-Latn': 'Jamajka', 'ca': 'Jamaica', 'ca-ES-valencia': 'Jamaica', 'ccp': '𑄎𑄟𑄭𑄇', 'ce': 'Ямайка', 'ceb': 'Jamaica', 'cgg': 'Gyamaika', 'chr': 'ᏣᎺᎢᎧ', 'ckb': 'جامایکا', 'cs': 'Jamajka', 'cv': 'Ямайка', 'cy': 'Jamaica', 'da': 'Jamaica', 'dav': 'Jamaika', 'de': 'Jamaika', 'dje': 'Jamaayik', 'dsb': 'Jamaika', 'dyo': 'Samaik', 'dz': 'ཇཱ་མཻ་ཀ', 'ebu': 'Jamaika', 'ee': 'Dzamaika nutome', 'el': 'Τζαμάικα', 'el-polyton': 'Τζαμάικα', 'en': 'Jamaica', 'en-Dsrt': '𐐖𐐲𐑋𐐩𐐿𐐲', 'eo': 'Jamajko', 'es': 'Jamaica', 'et': 'Jamaica', 'eu': 'Jamaika', 'ewo': 'Hamaíka', 'fa': 'جامائیکا', 'ff': 'Jamayka', 'ff-Adlm': '𞤔𞤢𞤥𞤢𞤴𞤳𞤢𞥄', 'ff-Latn': 'Jamayka', 'fi': 'Jamaika', 'fil': 'Jamaica', 'fo': 'Jamaika', 'fr': 'Jamaïque', 'fur': 'Gjamaiche', 'fy': 'Jamaica', 'ga': 'Iamáice', 'gd': 'Diameuga', 'gl': 'Xamaica', 'gsw': 'Dschamäika', 'gu': 'જમૈકા', 'guz': 'Jamaika', 'ha': 'Jamaika', 'he': 'ג׳מייקה', 'hi': 'जमैका', 'hi-Latn': 'Jamaica', 'hr': 'Jamajka', 'hsb': 'Jamaika', 'ht': 'Jamaïque', 'hu': 'Jamaica', 'hy': 'Ճամայկա', 'ia': 'Jamaica', 'id': 'Jamaika', 'ig': 'Jamaika', 'is': 'Jamaíka', 'it': 'Giamaica', 'ja': 'ジャマイカ', 'jmc': 'Jamaika', 'jv': 'Jamaika', 'ka': 'იამაიკა', 'kaa': 'Ямайка', 'kaa-Cyrl': 'Ямайка', 'kab': 'Jamyika', 'kam': 'Jamaika', 'kde': 'Dyamaika', 'kea': 'Jamaika', 'kgp': 'Jamỹjka', 'khq': 'Jamaayik', 'ki': 'Jamaika', 'kk': 'Ямайка', 'kk-Cyrl': 'Ямайка', 'kln': 'Emetab Jamaica', 'km': 'ហ្សាម៉ាអ៊ីក', 'kn': 'ಜಮೈಕಾ', 'ko': '자메이카', 'kok': 'जमैका', 'kok-Deva': 'जमैका', 'ks': 'جَمایکا', 'ks-Arab': 'جَمایکا', 'ksb': 'Jamaika', 'ksf': 'jamaík', 'ksh': 'Jammaika', 'ku': 'Jamaîka', 'kxv': 'jamaika', 'kxv-Deva': 'जमेका', 'kxv-Latn': 'jamaika', 'kxv-Orya': 'ଜାମାଇକା', 'kxv-Telu': 'జమైకా', 'ky': 'Ямайка', 'lag': 'Jamáika', 'lb': 'Jamaika', 'lg': 'Jamayika', 'lld': 'Giamaica', 'ln': 'Zamaiki', 'lo': 'ຈາໄມຄາ', 'lt': 'Jamaika', 'lu': 'Jamaiki', 'luo': 'Jamaica', 'luy': 'Jamaika', 'lv': 'Jamaika', 'mai': 'जमैका', 'mas': 'Jamaika', 'mer': 'Jamaika', 'mfe': 'Zamaik', 'mg': 'Jamaïka', 'mgh': 'Ujamaika', 'mi': 'Hemeika', 'mk': 'Јамајка', 'ml': 'ജമൈക്ക', 'mn': 'Ямайка', 'mr': 'जमैका', 'ms': 'Jamaica', 'mt': 'il-Ġamajka', 'mua': 'Jamaika', 'my': 'ဂျမေကာ', 'mzn': 'جاماییکا', 'naq': 'Jamaicab', 'nb': 'Jamaica', 'nd': 'Jamaica', 'ne': 'जमैका', 'nl': 'Jamaica', 'nmg': 'Jamaika', 'nn': 'Jamaica', 'no': 'Jamaica', 'nqo': 'ߖ߭ߡߊߦߌߞ', 'nyn': 'Gyamaika', 'om': 'Jamaayikaa', 'or': 'ଜାମାଇକା', 'pa': 'ਜਮਾਇਕਾ', 'pa-Guru': 'ਜਮਾਇਕਾ', 'pcm': 'Jamaíka', 'pl': 'Jamajka', 'ps': 'جمیکا', 'pt': 'Jamaica', 'qu': 'Jamaica', 'rm': 'Giamaica', 'rn': 'Jamayika', 'ro': 'Jamaica', 'rof': 'Jamaika', 'ru': 'Ямайка', 'rwk': 'Jamaika', 'sah': 'Дьамаайка', 'saq': 'Jamaika', 'sbp': 'Jamaika', 'sc': 'Giamàica', 'sd': 'جميڪا', 'sd-Arab': 'جميڪا', 'se': 'Jamaica', 'seh': 'Jamaica', 'ses': 'Jamaayik', 'sg': 'Zamaîka', 'shi': 'ⵊⴰⵎⴰⵢⴽⴰ', 'shi-Latn': 'jamayka', 'shi-Tfng': 'ⵊⴰⵎⴰⵢⴽⴰ', 'si': 'ජැමෙයිකාව', 'sk': 'Jamajka', 'sl': 'Jamajka', 'smn': 'Jamaika', 'sn': 'Jamaica', 'so': 'Jamaaika', 'sq': 'Xhamajkë', 'sr': 'Јамајка', 'sr-Cyrl': 'Јамајка', 'sr-Latn': 'Jamajka', 'sv': 'Jamaica', 'sw': 'Jamaika', 'syr': 'ܓܡܝܟܐ', 'ta': 'ஜமைகா', 'te': 'జమైకా', 'teo': 'Jamaika', 'tg': 'Ямайка', 'th': 'จาเมกา', 'ti': 'ጃማይካ', 'tk': 'Ýamaýka', 'to': 'Samaika', 'tr': 'Jamaika', 'tt': 'Ямайка', 'twq': 'Jamaayik', 'tzm': 'Jamayka', 'ug': 'يامايكا', 'uk': 'Ямайка', 'ur': 'جمائیکا', 'uz': 'Yamayka', 'uz-Cyrl': 'Ямайка', 'uz-Latn': 'Yamayka', 'vai': 'ꕧꕮꔧꕪ', 'vai-Latn': 'Jamaika', 'vai-Vaii': 'ꕧꕮꔧꕪ', 'vec': 'Jamàega', 'vi': 'Jamaica', 'vun': 'Jamaika', 'wae': 'Jamaika', 'wo': 'Samayig', 'xh': 'EJamaica', 'xnr': 'जमैका', 'xog': 'Jamayika', 'yav': 'samayíik', 'yi': 'דזשאַמייקע', 'yo': 'Jamaika', 'yrl': 'Yamaika', 'yue': '牙買加', 'yue-Hans': '牙买加', 'yue-Hant': '牙買加', 'zgh': 'ⵊⴰⵎⴰⵢⴽⴰ', 'zh': '牙买加', 'zh-Hans': '牙买加', 'zh-Hant': '牙買加', 'zu': 'i-Jamaica'},
    'JO': {'af': 'Jordanië', 'agq': 'Dzodàn', 'ak': 'Gyɔdan', 'am': 'ጆርዳን', 'ar': 'الأردن', 'as': 'জৰ্ডান', 'asa': 'Yordani', 'ast': 'Xordania', 'az': 'İordaniya', 'az-Cyrl': 'Иорданија', 'az-Latn': 'İordaniya', 'bal-Latn': 'Ordon', 'bas': 'Yɔ̀rdanià', 'be': 'Іарданія', 'be-tarask': 'Іарданія', 'bez': 'Huyolodani', 'bg': 'Йордания', 'bgn': 'اردن', 'blo': 'Jɔrɖanii', 'bm': 'Zɔrdani', 'bn': 'জর্ডন', 'br': 'Jordania', 'brx': 'जर्डान', 'bs': 'Jordan', 'bs-Cyrl': 'Јордан', 'bs-Latn': 'Jordan', 'ca': 'Jordània', 'ca-ES-valencia': 'Jordània', 'ccp': '𑄎𑄧𑄢𑄴𑄓𑄧𑄚𑄴', 'ce': 'Урдан', 'ceb': 'Jordan', 'cgg': 'Yorudaani', 'chr': 'ᏦᏓᏂ', 'ckb': 'ئوردن', 'cs': 'Jordánsko', 'cv': 'Иордани', 'cy': 'Gwlad Iorddonen', 'da': 'Jordan', 'dav': 'Yordani', 'de': 'Jordanien', 'dje': 'Urdun', 'dsb': 'Jordaniska', 'dz': 'ཇོར་ཌན', 'ebu': 'Yordani', 'ee': 'Yordan nutome', 'el': 'Ιορδανία', 'el-polyton': 'Ἰορδανία', 'en': 'Jordan', 'en-Dsrt': '𐐖𐐱𐑉𐐼𐐲𐑌', 'eo': 'Jordanio', 'es': 'Jordania', 'et': 'Jordaania', 'eu': 'Jordania', 'ewo': 'Horədaní', 'fa': 'اردن', 'ff': 'Jordani', 'ff-Adlm': '𞤔𞤮𞤪𞤣𞤢𞥄𞤲', 'ff-Latn': 'Jordani', 'fi': 'Jordania', 'fil': 'Jordan', 'fo': 'Jordan', 'fr': 'Jordanie', 'fur': 'Jordanie', 'fy': 'Jordanië', 'ga': 'an Iordáin', 'gd': 'Iòrdan', 'gl': 'Xordania', 'gsw': 'Jordaanie', 'gu': 'જોર્ડન', 'guz': 'Yordani', 'ha': 'Jordan', 'he': 'ירדן', 'hi': 'जॉर्डन', 'hi-Latn': 'Jordan', 'hr': 'Jordan', 'hsb': 'Jordaniska', 'ht': 'Jordanie', 'hu': 'Jordánia', 'hy': 'Հորդանան', 'ia': 'Jordania', 'id': 'Yordania', 'ig': 'Jordan', 'is': 'Jórdanía', 'it': 'Giordania', 'ja': 'ヨルダン', 'jmc': 'Yordani', 'jv': 'Yordania', 'ka': 'იორდანია', 'kaa': 'Иордания', 'kaa-Cyrl': 'Иордания', 'kab': 'Lajurdani', 'kam': 'Yordani', 'kde': 'Yordani', 'kea': 'Jordánia', 'kgp': 'Jornánĩja', 'khq': 'Urdun', 'ki': 'Njorondani', 'kk': 'Иордания', 'kk-Cyrl': 'Иордания', 'kln': 'Emetab Jordan', 'km': 'ហ៊្សកដានី', 'kn': 'ಜೋರ್ಡಾನ್', 'ko': '요르단', 'kok': 'जॉर्डन', 'kok-Deva': 'जॉर्डन', 'ks': 'جورڈن', 'ks-Arab': 'جورڈن', 'ksb': 'Yoldani', 'ksf': 'jɔrdán', 'ksh': 'Jordaanije', 'ku': 'Urdun', 'kxv': 'jorḍan', 'kxv-Deva': 'जॉर्डन', 'kxv-Latn': 'jorḍan', 'kxv-Orya': 'ଜର୍ଡାନ', 'kxv-Telu': 'జోర్డాన్', 'ky': 'Иордания', 'lag': 'Jódani', 'lb': 'Jordanien', 'lg': 'Yorodani', 'lld': 'Iordania', 'ln': 'Zɔdani', 'lo': 'ຈໍແດນ', 'lt': 'Jordanija', 'lu': 'Jodani', 'luo': 'Jordan', 'luy': 'Yordani', 'lv': 'Jordānija', 'mai': 'जार्डन', 'mas': 'Yordani', 'mer': 'Jorondani', 'mfe': 'Zordani', 'mg': 'Jordania', 'mgh': 'Uyordani', 'mi': 'Hōrano', 'mk': 'Јордан', 'ml': 'ജോർദ്ദാൻ', 'mn': 'Йордан', 'mr': 'जॉर्डन', 'ms': 'Jordan', 'mt': 'il-Ġordan', 'mua': 'Jordaniya', 'my': 'ဂျော်ဒန်', 'mzn': 'اردن', 'naq': 'Jordanni', 'nb': 'Jordan', 'nd': 'Jodani', 'ne': 'जोर्डन', 'nl': 'Jordanië', 'nmg': 'Jɔrdania', 'nn': 'Jordan', 'no': 'Jordan', 'nqo': 'ߖߐߙߑߘߊߣߌ߫', 'nyn': 'Yorudaani', 'om': 'Jirdaan', 'or': 'ଜୋର୍ଡାନ୍', 'pa': 'ਜਾਰਡਨ', 'pa-Guru': 'ਜਾਰਡਨ', 'pcm': 'Jọ́dan', 'pl': 'Jordania', 'ps': 'اردن', 'pt': 'Jordânia', 'qu': 'Jordania', 'rm': 'Jordania', 'rn': 'Yorudaniya', 'ro': 'Iordania', 'rof': 'Yordani', 'ru': 'Иордания', 'rwk': 'Yordani', 'saq': 'Yordani', 'sbp': 'Yolodani', 'sc': 'Giordània', 'sd': 'اردن', 'sd-Arab': 'اردن', 'se': 'Jordánia', 'seh': 'Jordânia', 'ses': 'Urdun', 'sg': 'Zordanïi', 'shi': 'ⵍⵓⵔⴷⵓⵏ', 'shi-Latn': 'lurdun', 'shi-Tfng': 'ⵍⵓⵔⴷⵓⵏ', 'si': 'ජෝර්දානය', 'sk': 'Jordánsko', 'sl': 'Jordanija', 'smn': 'Jordan', 'sn': 'Jordan', 'so': 'Urdun', 'sq': 'Jordani', 'sr': 'Јордан', 'sr-Cyrl': 'Јордан', 'sr-Latn': 'Jordan', 'sv': 'Jordanien', 'sw': 'Jordan', 'syr': 'ܐܘܪܕܘܢ', 'ta': 'ஜோர்டான்', 'te': 'జోర్డాన్', 'teo': 'Yordani', 'tg': 'Урдун', 'th': 'จอร์แดน', 'ti': 'ዮርዳኖስ', 'tk': 'Iordaniýa', 'to': 'Soatane', 'tr': 'Ürdün', 'tt': 'Иордания', 'twq': 'Urdun', 'tzm': 'Urḍun', 'ug': 'ئىيوردانىيە', 'uk': 'Йорданія', 'ur': 'اردن', 'uz': 'Iordaniya', 'uz-Cyrl': 'Иордания', 'uz-Latn': 'Iordaniya', 'vai': 'ꗘꖺꗵꘋ', 'vai-Latn': 'Jɔɔdaŋ', 'vai-Vaii': 'ꗘꖺꗵꘋ', 'vec': 'Jordania', 'vi': 'Jordan', 'vun': 'Yordani', 'wae': 'Jordanie', 'wo': 'Sordani', 'xh': 'EJordan', 'xnr': 'जॉर्डन', 'xog': 'Yorodani', 'yav': 'sɔltaní', 'yo': 'Jọdani', 'yrl': 'Yudãniya', 'yue': '約旦', 'yue-Hans': '约旦', 'yue-Hant': '約旦', 'zgh': 'ⵍⵓⵔⴷⵓⵏ', 'zh': '约旦', 'zh-Hans': '约旦', 'zh-Hant': '約旦', 'zu': 'i-Jordan'},
    'JP': {'af': 'Japan', 'agq': 'Dzàkpân', 'ak': 'Gyapan', 'am': 'ጃፓን', 'ar': 'اليابان', 'as': 'জাপান', 'asa': 'Japani', 'ast': 'Xapón', 'az': 'Yaponiya', 'az-Cyrl': 'Јапонија', 'az-Latn': 'Yaponiya', 'bal-Latn': 'Jápán', 'be': 'Японія', 'be-tarask': 'Японія', 'bez': 'Hujapani', 'bg': 'Япония', 'blo': 'Japaŋ', 'bm': 'Zapɔn', 'bn': 'জাপান', 'bo': 'ཉི་ཧོང་', 'br': 'Japan', 'brx': 'जापान', 'bs': 'Japan', 'bs-Cyrl': 'Јапан', 'bs-Latn': 'Japan', 'ca': 'Japó', 'ca-ES-valencia': 'Japó', 'ccp': '𑄎𑄛𑄚𑄴', 'ce': 'Япони', 'ceb': 'Japan', 'cgg': 'Gyapaani', 'chr': 'ᏣᏩᏂᏏ', 'ckb': 'ژاپۆن', 'cs': 'Japonsko', 'cv': 'Япони', 'cy': 'Japan', 'da': 'Japan', 'dav': 'Japani', 'de': 'Japan', 'dje': 'Jaapoŋ', 'doi': 'जापान', 'dsb': 'Japańska', 'dyo': 'Sapoŋ', 'dz': 'ཇ་པཱན', 'ebu': 'Japani', 'ee': 'Dzapan nutome', 'el': 'Ιαπωνία', 'el-polyton': 'Ἰαπωνία', 'en': 'Japan', 'en-Dsrt': '𐐖𐐲𐐹𐐰𐑌', 'eo': 'Japanujo', 'es': 'Japón', 'et': 'Jaapan', 'eu': 'Japonia', 'ewo': 'Hapɔ́n', 'fa': 'ژاپن', 'ff': 'Sapoo', 'ff-Adlm': '𞤐𞤭𞤨𞥆𞤮𞤲', 'ff-Latn': 'Sapoo', 'fi': 'Japani', 'fil': 'Japan', 'fo': 'Japan', 'fr': 'Japon', 'fur': 'Gjapon', 'fy': 'Japan', 'ga': 'an tSeapáin', 'gd': 'An t-Seapan', 'gl': 'Xapón', 'gsw': 'Japan', 'gu': 'જાપાન', 'guz': 'Japani', 'ha': 'Japan', 'haw': 'Iāpana', 'he': 'יפן', 'hi': 'जापान', 'hi-Latn': 'Japan', 'hr': 'Japan', 'hsb': 'Japanska', 'ht': 'Japon', 'hu': 'Japán', 'hy': 'Ճապոնիա', 'ia': 'Japon', 'id': 'Jepang', 'ig': 'Japan', 'ii': 'ꏝꀪ', 'is': 'Japan', 'it': 'Giappone', 'ja': '日本', 'jgo': 'Japɔ̂n', 'jmc': 'Japani', 'jv': 'Jepang', 'ka': 'იაპონია', 'kaa': 'Япония', 'kaa-Cyrl': 'Япония', 'kab': 'Jappu', 'kam': 'Japani', 'kde': 'Dyapani', 'kea': 'Japon', 'kgp': 'Japã', 'khq': 'Jaapoŋ', 'ki': 'Njabani', 'kk': 'Жапония', 'kk-Cyrl': 'Жапония', 'kln': 'Emetab Japan', 'km': 'ជប៉ុន', 'kn': 'ಜಪಾನ್', 'ko': '일본', 'kok': 'जपान', 'kok-Deva': 'जपान', 'ks': 'جاپان', 'ks-Arab': 'جاپان', 'ks-Deva': 'जापान', 'ksb': 'Japani', 'ksf': 'japɔ́ŋ', 'ksh': 'Japan', 'ku': 'Japonya', 'kxv': 'japan', 'kxv-Deva': 'जापान', 'kxv-Latn': 'japan', 'kxv-Orya': 'ଜାପାନ', 'kxv-Telu': 'జపాన్', 'ky': 'Жапония', 'lag': 'Japáani', 'lb': 'Japan', 'lg': 'Japani', 'lkt': 'Kisúŋla Makȟóčhe', 'lld': 'Iapan', 'ln': 'Zapɔ', 'lo': 'ຍີ່ປຸ່ນ', 'lrc': 'جاپوٙن', 'lt': 'Japonija', 'lu': 'Japu', 'luo': 'Japan', 'luy': 'Japani', 'lv': 'Japāna', 'mai': 'जापान', 'mas': 'Japani', 'mer': 'Japani', 'mfe': 'Zapon', 'mg': 'Japana', 'mgh': 'Ujapani', 'mi': 'Hapani', 'mk': 'Јапонија', 'ml': 'ജപ്പാൻ', 'mn': 'Япон', 'mni': 'জাপান', 'mni-Beng': 'জাপান', 'mr': 'जपान', 'ms': 'Jepun', 'mt': 'il-Ġappun', 'mua': 'Japaŋ', 'my': 'ဂျပန်', 'mzn': 'جاپون', 'naq': 'Japanni', 'nb': 'Japan', 'nd': 'Japan', 'ne': 'जापान', 'nl': 'Japan', 'nmg': 'Japɔn', 'nn': 'Japan', 'no': 'Japan', 'nqo': 'ߖ߭ߊߔߐ߲߫', 'nyn': 'Gyapaani', 'om': 'Jaappaan', 'or': 'ଜାପାନ', 'os': 'Япон', 'pa': 'ਜਪਾਨ', 'pa-Guru': 'ਜਪਾਨ', 'pcm': 'Japán', 'pl': 'Japonia', 'ps': 'جاپان', 'pt': 'Japão', 'qu': 'Japón', 'rm': 'Giapun', 'rn': 'Ubuyapani', 'ro': 'Japonia', 'rof': 'Japani', 'ru': 'Япония', 'rwk': 'Japani', 'sa': 'जापन:', 'saq': 'Japani', 'sat': 'ᱡᱟᱯᱟᱱ', 'sat-Olck': 'ᱡᱟᱯᱟᱱ', 'sbp': 'Japani', 'sc': 'Giapone', 'sd': 'جاپان', 'sd-Arab': 'جاپان', 'sd-Deva': 'जापान', 'se': 'Japána', 'seh': 'Japão', 'ses': 'Jaapoŋ', 'sg': 'Zapöon', 'shi': 'ⵍⵢⴰⴱⴰⵏ', 'shi-Latn': 'lyaban', 'shi-Tfng': 'ⵍⵢⴰⴱⴰⵏ', 'si': 'ජපානය', 'sk': 'Japonsko', 'sl': 'Japonska', 'smn': 'Jaapaan', 'sn': 'Japan', 'so': 'Jabaan', 'sq': 'Japoni', 'sr': 'Јапан', 'sr-Cyrl': 'Јапан', 'sr-Latn': 'Japan', 'su': 'Jepang', 'su-Latn': 'Jepang', 'sv': 'Japan', 'sw': 'Japani', 'syr': 'ܝܦܢ', 'ta': 'ஜப்பான்', 'te': 'జపాన్', 'teo': 'Japani', 'tg': 'Япония', 'th': 'ญี่ปุ่น', 'ti': 'ጃፓን', 'tk': 'Ýaponiýa', 'to': 'Siapani', 'tr': 'Japonya', 'tt': 'Япония', 'twq': 'Jaapoŋ', 'tzm': 'Jjappun', 'ug': 'ياپونىيە', 'uk': 'Японія', 'ur': 'جاپان', 'uz': 'Yaponiya', 'uz-Cyrl': 'Япония', 'uz-Latn': 'Yaponiya', 'vai': 'ꔛꗨꗢ', 'vai-Latn': 'Japaŋ', 'vai-Vaii': 'ꔛꗨꗢ', 'vec': 'Japon', 'vi': 'Nhật Bản', 'vun': 'Japani', 'wae': 'Japan', 'wo': 'Sàppoŋ', 'xh': 'EJapan', 'xnr': 'जापान', 'xog': 'Japani', 'yav': 'sapɔ́ɔŋ', 'yi': 'יאַפּאַן', 'yo': 'Japani', 'yrl': 'Nipõ', 'yue': '日本', 'yue-Hans': '日本', 'yue-Hant': '日本', 'zgh': 'ⵍⵢⴰⴱⴰⵏ', 'zh': '日本', 'zh-Hans': '日本', 'zh-Hant': '日本', 'zu': 'i-Japan'},
    'Jamo': {'ak': 'Gyamo kasa', 'am': 'ጃሞ', 'ar': 'جامو', 'as': 'জামো', 'bal-Latn': 'Jamó', 'be': 'чамо', 'be-tarask': 'чамо', 'bg': 'джамо', 'bn': 'জ্যামো', 'brx': 'जाम’', 'bs': 'pismo jamo', 'bs-Cyrl': 'џамо', 'bs-Latn': 'pismo jamo', 'ccp': '𑄎𑄳𑄠𑄟𑄮', 'ce': 'джамо', 'chr': 'ᏣᎼ', 'ckb': 'جامۆ', 'el': 'Τζάμο', 'el-polyton': 'Τζάμο', 'eu': 'jamoa', 'fa': 'جامو', 'ff-Adlm': '𞤔𞤢𞤥𞤮', 'fi': 'korean hangulin jamo-elementit', 'ga': 'Seamó', 'gu': 'જેમો', 'he': 'ג׳אמו', 'hi': 'जामो', 'hr': 'jamo pismo', 'hy': 'ջամո', 'ja': '字母', 'ka': 'ჯამო', 'kgp': 'jỹmo', 'kk': 'джамо жазуы', 'kk-Cyrl': 'джамо жазуы', 'km': 'ចាម៉ូ', 'kn': 'ಜಮೋ', 'ko': '자모', 'kok': 'जामो', 'kok-Deva': 'जामो', 'ku': 'jamoyî', 'ky': 'Джамо', 'lo': 'ຈາໂມ', 'lt': 'Jamo simboliai', 'mai': 'जमो', 'mi': 'Hamo', 'mk': 'џамо', 'ml': 'ജാമോ', 'mn': 'жамо', 'mr': 'जामो', 'my': 'ဂျမို', 'ne': 'जामो', 'nqo': 'ߖߡߏ', 'or': 'ଜାମୋ', 'pa': 'ਜਾਮੋ', 'pa-Guru': 'ਜਾਮੋ', 'pcm': 'Jámo', 'ps': 'جامو', 'ru': 'джамо', 'sd': 'جامو', 'sd-Arab': 'جامو', 'si': 'ජාමො', 'so': 'Jaamo', 'sq': 'jamosisht', 'sr': 'џамо', 'sr-Cyrl': 'џамо', 'sr-Latn': 'džamo', 'ta': 'ஜமோ', 'te': 'జమో', 'tg': 'Ҷамо', 'th': 'จาโม', 'ti': 'ጃሞ', 'tk': 'Jamo elipbiýi', 'to': 'tohinima fakasamo', 'uk': 'чамо', 'ur': 'جامو', 'vi': 'Chữ Jamo', 'xnr': 'जामो', 'yrl': 'yamu', 'yue': '韓文字母', 'yue-Hans': '韩文字母', 'yue-Hant': '韓文字母', 'zh': '韩文字母', 'zh-Hans': '韩文字母', 'zh-Hant': '韓文字母', 'zu': 'isi-Jamo', 'en': 'Jamo (alias for Jamo subset of Hangul)'},
    'Java': {'ar': 'الجاوية', 'ast': 'xavanés', 'az': 'cava', 'az-Latn': 'cava', 'bal-Latn': 'Jáwá', 'bg': 'Яванска', 'bn': 'জাভানিজ', 'br': 'javanek', 'brx': 'जावानीस', 'bs': 'javansko pismo', 'bs-Cyrl': 'јаванско писмо', 'bs-Latn': 'javansko pismo', 'ca': 'javanès', 'ca-ES-valencia': 'javanès', 'ccp': '𑄎𑄞𑄚𑄨𑄎𑄴', 'cs': 'javánské', 'da': 'javanesisk', 'de': 'Javanesisch', 'el': 'Ιαβανεζικό', 'el-polyton': 'Ιαβανεζικό', 'en': 'Javanese', 'en-Dsrt': '𐐖𐐪𐑂𐐲𐑌𐐨𐑆', 'es': 'javanés', 'et': 'jaava', 'eu': 'javaniera', 'fa': 'جاوه\u200cای', 'ff-Adlm': '𞤟𞤢𞤾𞤢𞥄', 'fi': 'jaavalainen', 'fo': 'javanskt', 'fr': 'javanais', 'fur': 'gjavanês', 'fy': 'Javaansk', 'ga': 'Iávach', 'gd': 'Deàbhanais', 'gsw': 'Javanesisch', 'gu': 'જાવાનીસ', 'he': 'ג׳אוונזי', 'hi': 'जावानीस', 'hi-Latn': 'Javanese', 'hr': 'javansko pismo', 'ht': 'javanais', 'hu': 'Jávai', 'id': 'Jawa', 'ig': 'Javanese', 'is': 'javanesíska', 'it': 'javanese', 'ja': 'ジャワ文字', 'ka': 'იავური', 'kgp': 'javanẽj', 'kk': 'ява жазуы', 'kk-Cyrl': 'ява жазуы', 'km': 'ជ្វា', 'kn': 'ಜಾವನೀಸ್', 'ko': '자바 문자', 'ks': 'جاوَنیٖز', 'ks-Arab': 'جاوَنیٖز', 'lb': 'Javanesesch', 'lo': 'ຈາວາ', 'lt': 'javiečių', 'lv': 'javiešu', 'mk': 'јаванско', 'ml': 'ജാവനീസ്', 'mn': 'Жава', 'mr': 'जावानीस', 'ms': 'Jawa', 'my': 'ဂျာဗားနီးစ်', 'nb': 'javanesisk', 'ne': 'जाभानी', 'nl': 'Javaans', 'nn': 'javanesisk', 'no': 'javanesisk', 'nqo': 'ߖ߭ߝ߭ߊߣߍߛ', 'or': 'ଜାଭାନୀଜ୍', 'pl': 'jawajskie', 'pt': 'javanês', 'rm': 'javanais', 'ro': 'javaneză', 'ru': 'яванская', 'sc': 'giavanesu', 'sd': 'جاوانيز', 'sd-Arab': 'جاوانيز', 'sl': 'javanski', 'so': 'Jafaniis', 'sr': 'јаванско писмо', 'sr-Cyrl': 'јаванско писмо', 'sr-Latn': 'javansko pismo', 'sv': 'javanska', 'ta': 'ஜாவனீஸ்', 'te': 'జావనీస్', 'tg': 'Ҷаванесӣ', 'th': 'ชวา', 'to': 'tohinima fakasava', 'tr': 'Cava Dili', 'ug': 'ياۋاچە', 'uk': 'яванський', 'vec': 'javaneze', 'vi': 'Chữ Java', 'xnr': 'जावानीस', 'yo': 'Èdè Jafaniisi', 'yrl': 'yawanei', 'yue': '爪哇文', 'yue-Hans': '爪哇文', 'yue-Hant': '爪哇文', 'zh': '爪哇文', 'zh-Hans': '爪哇文', 'zh-Hant': '爪哇文', 'zu': 'isi-Javanese'},
    'Jpan': {'af': 'Japannees', 'ak': 'Gyapanfoɔ kasa', 'am': 'ጃፓንኛ', 'ar': 'اليابانية', 'as': 'জাপানী', 'ast': 'xaponés', 'az': 'yapon', 'az-Arab': 'ژاپونجا', 'az-Latn': 'yapon', 'bal-Latn': 'Jápáni', 'be': 'японскае', 'be-tarask': 'японскае', 'bg': 'японска', 'bgn': 'جاپانی', 'blo': 'Japaŋ kʊja', 'bn': 'জাপানী', 'br': 'japanek', 'brx': 'जापानीज', 'bs': 'japansko pismo', 'bs-Cyrl': 'јапанско писмо', 'bs-Latn': 'japansko pismo', 'ca': 'japonès', 'ca-ES-valencia': 'japonès', 'ccp': '𑄎𑄛𑄚𑄩', 'ce': 'японийн', 'ceb': 'Japanese', 'chr': 'ᏣᏆᏂᏏ', 'ckb': 'ژاپۆنی', 'cs': 'japonské', 'cv': 'япони', 'cy': 'Japaneaidd', 'da': 'japansk', 'de': 'Japanisch', 'dsb': 'japański', 'dz': 'ཇ་པཱན་ཡིག་གུ', 'ee': 'Japanesegbeŋɔŋlɔ', 'el': 'Ιαπωνικό', 'el-polyton': 'Ιαπωνικό', 'en': 'Japanese', 'en-Dsrt': '𐐖𐐪𐐹𐐲𐑌𐐨𐑆', 'es': 'japonés', 'et': 'jaapani', 'eu': 'japoniarra', 'fa': 'ژاپنی', 'ff-Adlm': '𞤐𞤭𞤨𞤮𞤲', 'fi': 'japanilainen', 'fil': 'Japanese', 'fo': 'japanskt', 'fr': 'japonais', 'fur': 'gjaponês', 'fy': 'Japans', 'ga': 'Seapánach', 'gd': 'Seapanais', 'gl': 'xaponés', 'gsw': 'Japanisch', 'gu': 'જાપાની', 'ha': 'Jafanis', 'he': 'יפני', 'hi': 'जापानी', 'hi-Latn': 'Japanese', 'hr': 'japansko pismo', 'hsb': 'japansce', 'ht': 'japonais', 'hu': 'Japán', 'hy': 'ճապոնական', 'ia': 'japonese', 'id': 'Jepang', 'ie': 'japanesi', 'ig': 'Japanese', 'is': 'japanskt', 'it': 'giapponese', 'ja': '日本語の文字', 'jv': 'Jepang', 'ka': 'იაპონური', 'kaa': 'япон жазыўы', 'kaa-Cyrl': 'япон жазыўы', 'kea': 'japones', 'kgp': 'japunẽj', 'kk': 'жапон жазуы', 'kk-Cyrl': 'жапон жазуы', 'km': 'ជប៉ុន', 'kn': 'ಜಾಪನೀಸ್', 'ko': '일본 문자', 'kok': 'जपानी', 'kok-Deva': 'जपानी', 'ks': 'جاپٲنی', 'ks-Arab': 'جاپٲنی', 'ksh': 'en japaanesche Schreff', 'ku': 'japonî', 'ky': 'Жапан', 'lb': 'Japanesch', 'lld': 'iapanesc', 'lo': 'ຍີ່ປຸ່ນ', 'lrc': 'جاپوٙنی', 'lt': 'japonų', 'lv': 'japāņu', 'mai': 'जापानी', 'mi': 'Hapanihi', 'mk': 'јапонско писмо', 'ml': 'ജാപ്പനീസ്', 'mn': 'япон', 'mni': 'জপানিজ', 'mni-Beng': 'জপানিজ', 'mr': 'जपानी', 'ms': 'Jepun', 'mt': 'Ġappuniż', 'my': 'ဂျပန်', 'mzn': 'جاپونی', 'nb': 'japansk', 'ne': 'जापानी', 'nl': 'Japans', 'nn': 'japansk', 'no': 'japansk', 'nqo': 'ߣߌߔߐ߲߫', 'om': 'Afaan Jaappaan', 'or': 'ଜାପାନୀଜ୍', 'pa': 'ਜਪਾਨੀ', 'pa-Guru': 'ਜਪਾਨੀ', 'pcm': 'Japan', 'pl': 'japońskie', 'ps': 'جاپاني', 'pt': 'japonês', 'qu': 'Japones Simi', 'rm': 'giapunais', 'ro': 'japoneză', 'ru': 'японская', 'sah': 'Дьоппуоннуу', 'sc': 'giaponesu', 'sd': 'جاپاني', 'sd-Arab': 'جاپاني', 'si': 'ජපන්', 'sk': 'japonské', 'sl': 'japonski', 'so': 'Jabaaniis', 'sq': 'japonez', 'sr': 'јапанско писмо', 'sr-Cyrl': 'јапанско писмо', 'sr-Latn': 'japansko pismo', 'sv': 'japanska', 'sw': 'Kijapani', 'ta': 'ஜப்பானியம்', 'te': 'జపనీస్', 'tg': 'Ҷопонӣ', 'th': 'ญี่ปุ่น', 'ti': 'ጃፓናዊ', 'tk': 'Ýapon elipbiýi', 'to': 'tohinima fakasiapani', 'tok': 'sitelen Nijon', 'tr': 'Japon', 'tt': 'япон', 'ug': 'ياپونچە', 'uk': 'японська', 'ur': 'جاپانی', 'uz': 'yapon', 'uz-Cyrl': 'Япон', 'uz-Latn': 'yapon', 'vec': 'japoneze', 'vi': 'Chữ Nhật Bản', 'wae': 'Japaniš', 'wo': 'Nihon no', 'xh': 'IsiJapanese', 'xnr': 'जापानी', 'yo': 'èdè jàpáànù', 'yrl': 'yapunei', 'yue': '日文', 'yue-Hans': '日文', 'yue-Hant': '日文', 'zh': '日文', 'zh-Hans': '日文', 'zh-Hant': '日文', 'zu': 'isi-Japanese'},
    'Jurc': {'ast': 'jurchen', 'bs': 'jurchen pismo', 'bs-Latn': 'jurchen pismo', 'ca': 'jürchen', 'ca-ES-valencia': 'jürchen', 'cs': 'džürčenské', 'da': 'jurchen', 'de': 'Jurchen', 'en': 'Jurchen', 'et': 'tšurtšeni', 'fi': 'džurtšen', 'fy': 'Jurchen', 'gd': 'Jurchen', 'hi-Latn': 'Jurchen', 'hr': 'jurchen pismo', 'id': 'Jurchen', 'it': 'jurchen', 'ja': '女真文字', 'ka': 'ჯურჯენული', 'ko': '줄첸 문자', 'lo': 'ຈູຮເຊັນ', 'lt': 'Jurchen', 'mk': 'џурченско', 'nb': 'jurchen', 'nl': 'Jurchen', 'nn': 'jurchen', 'no': 'jurchen', 'ru': 'чжурчжэньская', 'sv': 'jurchenska', 'th': 'จูร์เชน', 'to': 'tohinima fakaiūkeni', 'tr': 'Jurchen', 'ug': 'جۇرچېنچە', 'vi': 'Chữ Jurchen', 'yue': '女真文字', 'yue-Hans': '女真文字', 'yue-Hant': '女真文字', 'zh': '女真文', 'zh-Hans': '女真文', 'zh-Hant': '女真文字'},
    'KE': {'af': 'Kenia', 'agq': 'Kɨnyà', 'ak': 'Kenya', 'am': 'ኬንያ', 'ar': 'كينيا', 'as': 'কেনিয়া', 'asa': 'Kenya', 'ast': 'Kenia', 'az': 'Keniya', 'az-Cyrl': 'Кенија', 'az-Latn': 'Keniya', 'bal-Latn': 'Kiniá', 'bas': 'Kenìà', 'be': 'Кенія', 'be-tarask': 'Кенія', 'bez': 'Hukenya', 'bg': 'Кения', 'bgn': 'کینیا', 'blo': 'Keniya', 'bm': 'Keniya', 'bn': 'কেনিয়া', 'br': 'Kenya', 'brx': 'केनिया', 'bs': 'Kenija', 'bs-Cyrl': 'Кенија', 'bs-Latn': 'Kenija', 'ca': 'Kenya', 'ca-ES-valencia': 'Kenya', 'ccp': '𑄇𑄬𑄚𑄨𑄠', 'ce': 'Кени', 'ceb': 'Kenya', 'cgg': 'Kenya', 'chr': 'ᎨᏂᏯ', 'ckb': 'کینیا', 'cs': 'Keňa', 'cv': 'Кени', 'cy': 'Kenya', 'da': 'Kenya', 'dav': 'Kenya', 'de': 'Kenia', 'dje': 'Keeniya', 'dsb': 'Kenia', 'dyo': 'Keniya', 'dz': 'ཀེན་ཡ', 'ebu': 'Kenya', 'ee': 'Kenya nutome', 'el': 'Κένυα', 'el-polyton': 'Κένυα', 'en': 'Kenya', 'en-Dsrt': '𐐗𐐯𐑌𐐷𐐲', 'eo': 'Kenjo', 'es': 'Kenia', 'et': 'Keenia', 'eu': 'Kenya', 'ewo': 'Keniá', 'fa': 'کنیا', 'ff': 'Keñaa', 'ff-Adlm': '𞤑𞤫𞤲𞤭𞤴𞤢𞥄', 'ff-Latn': 'Keñaa', 'fi': 'Kenia', 'fil': 'Kenya', 'fo': 'Kenja', 'fr': 'Kenya', 'fur': 'Kenya', 'fy': 'Kenia', 'ga': 'an Chéinia', 'gd': 'Ceinia', 'gl': 'Kenya', 'gsw': 'Keenia', 'gu': 'કેન્યા', 'guz': 'Kenya', 'ha': 'Kenya', 'he': 'קניה', 'hi': 'केन्या', 'hi-Latn': 'Kenya', 'hr': 'Kenija', 'hsb': 'Kenija', 'ht': 'Kenya', 'hu': 'Kenya', 'hy': 'Քենիա', 'ia': 'Kenya', 'id': 'Kenya', 'ig': 'Kenya', 'is': 'Kenía', 'it': 'Kenya', 'ja': 'ケニア', 'jgo': 'Kɛ́nya', 'jmc': 'Kenya', 'jv': 'Kénya', 'ka': 'კენია', 'kaa': 'Кения', 'kaa-Cyrl': 'Кения', 'kab': 'Kinya', 'kam': 'Kenya', 'kde': 'Kenya', 'kea': 'Kénia', 'kgp': 'Kenĩja', 'khq': 'Keeniya', 'ki': 'Kenya', 'kk': 'Кения', 'kk-Cyrl': 'Кения', 'kln': 'Emetab Kenya', 'km': 'កេនយ៉ា', 'kn': 'ಕೀನ್ಯಾ', 'ko': '케냐', 'kok': 'केनया', 'kok-Deva': 'केनया', 'ks': 'کِنیا', 'ks-Arab': 'کِنیا', 'ksb': 'Kenya', 'ksf': 'kɛnya', 'ksh': 'Keenija', 'ku': 'Kenya', 'kxv': 'kenya', 'kxv-Deva': 'केन्या', 'kxv-Latn': 'kenya', 'kxv-Orya': 'କେନ୍ୟା', 'kxv-Telu': 'కెన్యా', 'ky': 'Кения', 'lag': 'Kéenya', 'lb': 'Kenia', 'lg': 'Kenya', 'lld': 'Kenia', 'ln': 'Kenya', 'lo': 'ເຄນຢາ', 'lt': 'Kenija', 'lu': 'Kenya', 'luo': 'Kenya', 'luy': 'Kenya', 'lv': 'Kenija', 'mai': 'केन्या', 'mas': 'Kenya', 'mer': 'Kenya', 'mfe': 'Kenya', 'mg': 'Kenya', 'mgh': 'Ukenya', 'mi': 'Kenia', 'mk': 'Кенија', 'ml': 'കെനിയ', 'mn': 'Кени', 'mr': 'केनिया', 'ms': 'Kenya', 'mt': 'il-Kenja', 'mua': 'Sǝr Kenya', 'my': 'ကင်ညာ', 'mzn': 'کنیا', 'naq': 'Kenyab', 'nb': 'Kenya', 'nd': 'Khenya', 'ne': 'केन्या', 'nl': 'Kenia', 'nmg': 'Kɛnya', 'nn': 'Kenya', 'no': 'Kenya', 'nqo': 'ߞߋߣߌߦߊ߫', 'nyn': 'Kenya', 'om': 'Keeniyaa', 'or': 'କେନିୟା', 'pa': 'ਕੀਨੀਆ', 'pa-Guru': 'ਕੀਨੀਆ', 'pcm': 'Kẹ́nya', 'pl': 'Kenia', 'ps': 'کینیا', 'pt': 'Quênia', 'qu': 'Kenia', 'rm': 'Kenia', 'rn': 'Kenya', 'ro': 'Kenya', 'rof': 'Kenya', 'ru': 'Кения', 'rwk': 'Kenya', 'saq': 'Kenya', 'sbp': 'Kenya', 'sc': 'Kènya', 'sd': 'ڪينيا', 'sd-Arab': 'ڪينيا', 'se': 'Kenia', 'seh': 'Quênia', 'ses': 'Keeniya', 'sg': 'Kenyäa', 'shi': 'ⴽⵉⵏⵢⴰ', 'shi-Latn': 'kinya', 'shi-Tfng': 'ⴽⵉⵏⵢⴰ', 'si': 'කෙන්යාව', 'sk': 'Keňa', 'sl': 'Kenija', 'smn': 'Kenia', 'sn': 'Kenya', 'so': 'Kenya', 'sq': 'Kenia', 'sr': 'Кенија', 'sr-Cyrl': 'Кенија', 'sr-Latn': 'Kenija', 'sv': 'Kenya', 'sw': 'Kenya', 'syr': 'ܩܝܢܝܐ', 'ta': 'கென்யா', 'te': 'కెన్యా', 'teo': 'Kenia', 'tg': 'Кения', 'th': 'เคนยา', 'ti': 'ኬንያ', 'tk': 'Keniýa', 'to': 'Keniā', 'tr': 'Kenya', 'tt': 'Кения', 'twq': 'Keeniya', 'tzm': 'Kinya', 'ug': 'كېنىيە', 'uk': 'Кенія', 'ur': 'کینیا', 'uz': 'Keniya', 'uz-Cyrl': 'Кения', 'uz-Latn': 'Keniya', 'vai': 'ꔞꕰ', 'vai-Latn': 'Kénya', 'vai-Vaii': 'ꔞꕰ', 'vec': 'Kenya', 'vi': 'Kenya', 'vun': 'Kenya', 'wae': 'Kenya', 'wo': 'Keeña', 'xh': 'EKenya', 'xnr': 'केन्या', 'xog': 'Kenya', 'yav': 'kénia', 'yi': 'קעניע', 'yo': 'Kenya', 'yrl': 'Kẽniya', 'yue': '肯亞', 'yue-Hans': '肯亚', 'yue-Hant': '肯亞', 'zgh': 'ⴽⵉⵏⵢⴰ', 'zh': '肯尼亚', 'zh-Hans': '肯尼亚', 'zh-Hant': '肯亞', 'zu': 'i-Kenya'},
    'KG': {'af': 'Kirgistan', 'agq': 'Kìdzisɨ̀tân', 'ak': 'Kɛɛgestan', 'am': 'ኪርጊስታን', 'ar': 'قيرغيزستان', 'as': 'কিৰ্গিজস্তান', 'asa': 'Kirigizithtani', 'ast': 'Kirguistán', 'az': 'Qırğızıstan', 'az-Cyrl': 'Гырғызыстан', 'az-Latn': 'Qırğızıstan', 'bal-Latn': 'Karghazestán', 'bas': 'Kìrgìzìstàŋ', 'be': 'Кыргызстан', 'be-tarask': 'Кыргызстан', 'bez': 'Hukiligizistani', 'bg': 'Киргизстан', 'bgn': 'قیرغیزستان', 'blo': 'Kirgistan', 'bm': 'Kirigizisitaŋ', 'bn': 'কিরগিজিস্তান', 'br': 'Kyrgyzstan', 'brx': 'किर्गीस्तान', 'bs': 'Kirgistan', 'bs-Cyrl': 'Киргизстан', 'bs-Latn': 'Kirgistan', 'ca': 'Kirguizstan', 'ca-ES-valencia': 'Kirguizstan', 'ccp': '𑄇𑄨𑄢𑄴𑄉𑄨𑄎𑄨𑄌𑄴𑄖𑄚𑄴', 'ce': 'Киргизи', 'ceb': 'Kyrgyzstan', 'cgg': 'Kirigizistani', 'chr': 'ᎩᎵᏣᎢᏍ', 'ckb': 'کرگیزستان', 'cs': 'Kyrgyzstán', 'cv': 'Киргизи', 'cy': 'Kyrgyzstan', 'da': 'Kirgisistan', 'dav': 'Kirigizistani', 'de': 'Kirgisistan', 'dje': 'Kyrgyzstan', 'dsb': 'Kirgizistan', 'dz': 'ཀིར་གིས་སཏཱན', 'ebu': 'Kirigizistani', 'ee': 'Kirgizstan nutome', 'el': 'Κιργιστάν', 'el-polyton': 'Κιργιστάν', 'en': 'Kyrgyzstan', 'en-Dsrt': '𐐗𐐮𐑉𐑀𐐲𐑅𐐻𐐰𐑌', 'eo': 'Kirgizujo', 'es': 'Kirguistán', 'et': 'Kõrgõzstan', 'eu': 'Kirgizistan', 'ewo': 'Kirigisətán', 'fa': 'قرقیزستان', 'ff': 'Kirgistaan', 'ff-Adlm': '𞤑𞤭𞤪𞤶𞤭𞤧𞤼𞤢𞥄𞤲', 'ff-Latn': 'Kirgistaan', 'fi': 'Kirgisia', 'fil': 'Kyrgyzstan', 'fo': 'Kirgisia', 'fr': 'Kirghizstan', 'fur': 'Kirghizstan', 'fy': 'Kirgizië', 'ga': 'an Chirgeastáin', 'gd': 'Cìorgastan', 'gl': 'Kirguizistán', 'gsw': 'Kirgiisischtan', 'gu': 'કિર્ગિઝ્સ્તાન', 'guz': 'Kirigizistani', 'ha': 'Kirgizistan', 'he': 'קירגיזסטן', 'hi': 'किर्गिज़स्तान', 'hi-Latn': 'Kyrgyzstan', 'hr': 'Kirgistan', 'hsb': 'Kirgizistan', 'ht': 'Kirghizstan', 'hu': 'Kirgizisztán', 'hy': 'Ղրղզստան', 'ia': 'Kirghizistan', 'id': 'Kirgizstan', 'ig': 'Kyrgyzstan', 'is': 'Kirgistan', 'it': 'Kirghizistan', 'ja': 'キルギス', 'jmc': 'Kirigizistani', 'jv': 'Kirgistan', 'ka': 'ყირგიზეთი', 'kaa': 'Қырғызстан', 'kaa-Cyrl': 'Қырғызстан', 'kab': 'Kirigistan', 'kam': 'Kirigizistani', 'kde': 'Kiligizistani', 'kea': 'Kirgistan', 'kgp': 'Kirginhtỹv', 'khq': 'Kyrgyzstan', 'ki': 'Kirigizistani', 'kk': 'Қырғызстан', 'kk-Cyrl': 'Қырғызстан', 'kln': 'Emetab Kyrgyzstan', 'km': 'កៀហ្ស៊ីស៊ីស្ថាន', 'kn': 'ಕಿರ್ಗಿಸ್ಥಾನ್', 'ko': '키르기스스탄', 'kok': 'किर्गिझस्तान', 'kok-Deva': 'किर्गिझस्तान', 'ks': 'کرغزستان', 'ks-Arab': 'کرغزستان', 'ksb': 'Kiigizistani', 'ksf': 'kigistáŋ', 'ksh': 'Kirrjiisije', 'ku': 'Qirgizistan', 'kxv': 'kirgistan', 'kxv-Deva': 'किर्गिज़स्तान', 'kxv-Latn': 'kirgistan', 'kxv-Orya': 'କିରଗିସ୍ତାନ', 'kxv-Telu': 'కిర్గజిస్తాన్', 'ky': 'Кыргызстан', 'lag': 'Kirigisitáani', 'lb': 'Kirgisistan', 'lg': 'Kirigizisitaani', 'lld': 'Kyrgystan', 'ln': 'Kigizisitá', 'lo': 'ຄຽກກິດສະຖານ', 'lt': 'Kirgizija', 'lu': 'Kigizisita', 'luo': 'Kyrgyzstan', 'luy': 'Kirigizistani', 'lv': 'Kirgizstāna', 'mai': 'किर्गिजस्तान', 'mas': 'Kirigizistani', 'mer': 'Kirigizistani', 'mfe': 'Kirghizistan', 'mg': 'Kiordistan', 'mi': 'Kikitānga', 'mk': 'Киргистан', 'ml': 'കിർഗിസ്ഥാൻ', 'mn': 'Киргиз', 'mr': 'किरगिझस्तान', 'ms': 'Kyrgyzstan', 'mt': 'il-Kirgiżistan', 'mua': 'Kirgizstaŋ', 'my': 'ကာဂျစ္စတန်', 'mzn': 'قرقیزستون', 'naq': 'Kyrgyzstanni', 'nb': 'Kirgisistan', 'nd': 'Kyrgyzstan', 'ne': 'किर्गिस्तान', 'nl': 'Kirgizië', 'nmg': 'Kyrgystaŋ', 'nn': 'Kirgisistan', 'no': 'Kirgisistan', 'nqo': 'ߞߌߙߑߜ߭ߌߛߑߕߊ߲߫', 'nyn': 'Kirigizistani', 'om': 'Kiyirigiyizistan', 'or': 'କିର୍ଗିଜିସ୍ତାନ', 'pa': 'ਕਿਰਗਿਜ਼ਸਤਾਨ', 'pa-Guru': 'ਕਿਰਗਿਜ਼ਸਤਾਨ', 'pcm': 'Kẹjístan', 'pl': 'Kirgistan', 'ps': 'قرغزستان', 'pt': 'Quirguistão', 'qu': 'Kirguistán', 'rm': 'Kirgisistan', 'rn': 'Kirigisitani', 'ro': 'Kârgâzstan', 'rof': 'Kirigizistani', 'ru': 'Киргизия', 'rwk': 'Kirigizistani', 'saq': 'Kirigizistani', 'sbp': 'Kiligisisitani', 'sc': 'Kirghìzistan', 'sd': 'ڪرغستان', 'sd-Arab': 'ڪرغستان', 'se': 'Kirgisistan', 'seh': 'Quirguistão', 'ses': 'Kyrgyzstan', 'sg': 'Kirigizitùaan', 'shi': 'ⴽⵉⵔⵖⵉⵣⵉⵙⵜⴰⵏ', 'shi-Latn': 'kirɣizistan', 'shi-Tfng': 'ⴽⵉⵔⵖⵉⵣⵉⵙⵜⴰⵏ', 'si': 'කිර්ගිස්තානය', 'sk': 'Kirgizsko', 'sl': 'Kirgizistan', 'smn': 'Kirgisia', 'sn': 'Kyrgyzstan', 'so': 'Kirgistaan', 'sq': 'Kirgizi', 'sr': 'Киргистан', 'sr-Cyrl': 'Киргистан', 'sr-Latn': 'Kirgistan', 'sv': 'Kirgizistan', 'sw': 'Kirigizistani', 'syr': 'ܩܝܪܓܝܙܣܬܐܢ', 'ta': 'கிர்கிஸ்தான்', 'te': 'కిర్గిజిస్తాన్', 'teo': 'Kirigizistani', 'tg': 'Қирғизистон', 'th': 'คีร์กีซสถาน', 'ti': 'ኪርጊዝስታን', 'tk': 'Gyrgyzystan', 'to': 'Kīkisitani', 'tr': 'Kırgızistan', 'tt': 'Кыргызстан', 'twq': 'Kyrgyzstan', 'tzm': 'Kirɣistan', 'ug': 'قىرغىزىستان', 'uk': 'Киргизстан', 'ur': 'کرغزستان', 'uz': 'Qirgʻiziston', 'uz-Cyrl': 'Қирғизистон', 'uz-Latn': 'Qirgʻiziston', 'vai': 'ꕃꕅꔻꕚꘋ', 'vai-Latn': 'Kigisitaŋ', 'vai-Vaii': 'ꕃꕅꔻꕚꘋ', 'vec': 'Kirghizistan', 'vi': 'Kyrgyzstan', 'vun': 'Kirigizistani', 'wae': 'Kirgištan', 'wo': 'Kirgistaŋ', 'xh': 'EKyrgyzstan', 'xnr': 'किर्गिज़स्तान', 'xog': 'Kirigizisitaani', 'yav': 'kilikisistáŋ', 'yo': 'Kuriṣisitani', 'yrl': 'Kirigiretãma', 'yue': '吉爾吉斯', 'yue-Hans': '吉尔吉斯', 'yue-Hant': '吉爾吉斯', 'zgh': 'ⴽⵉⵔⵖⵉⵣⵉⵙⵜⴰⵏ', 'zh': '吉尔吉斯斯坦', 'zh-Hans': '吉尔吉斯斯坦', 'zh-Hant': '吉爾吉斯', 'zu': 'i-Kyrgyzstan'},
    'KH': {'af': 'Kambodja', 'agq': 'Kàmbodìa', 'ak': 'Kambodia', 'am': 'ካምቦዲያ', 'ar': 'كمبوديا', 'as': 'কম্বোডিয়া', 'asa': 'Kambodia', 'ast': 'Camboya', 'az': 'Kamboca', 'az-Cyrl': 'Камбоҹа', 'az-Latn': 'Kamboca', 'bal-Latn': 'Kambódhiá', 'bas': 'Kàmbodìà', 'be': 'Камбоджа', 'be-tarask': 'Камбоджа', 'bez': 'Hukambodia', 'bg': 'Камбоджа', 'bgn': 'کمبودیا', 'blo': 'Kamboɖiya', 'bm': 'Kamboji', 'bn': 'কম্বোডিয়া', 'br': 'Kambodja', 'brx': 'कम्ब’डिया', 'bs': 'Kambodža', 'bs-Cyrl': 'Камбоџа', 'bs-Latn': 'Kambodža', 'ca': 'Cambodja', 'ca-ES-valencia': 'Cambodja', 'ccp': '𑄇𑄧𑄟𑄴𑄝𑄮𑄓𑄨𑄠', 'ce': 'Камбоджа', 'ceb': 'Cambodia', 'cgg': 'Kambodiya', 'chr': 'ᎧᎹᏉᏗᎠᏂ', 'ckb': 'کەمبۆدیا', 'cs': 'Kambodža', 'cv': 'Камбоджа', 'cy': 'Cambodia', 'da': 'Cambodja', 'dav': 'Kambodia', 'de': 'Kambodscha', 'dje': 'kamboogi', 'dsb': 'Kambodža', 'dyo': 'Kamboj', 'dz': 'ཀམ་བྷོ་ཌི་ཡ', 'ebu': 'Kambodia', 'ee': 'Kambodia nutome', 'el': 'Καμπότζη', 'el-polyton': 'Καμπότζη', 'en': 'Cambodia', 'en-Dsrt': '𐐗𐐰𐑋𐐺𐐬𐐼𐐨𐐲', 'eo': 'Kamboĝo', 'es': 'Camboya', 'et': 'Kambodža', 'eu': 'Kanbodia', 'ewo': 'kambodía', 'fa': 'کامبوج', 'ff': 'Kambodso', 'ff-Adlm': '𞤑𞤢𞤥𞤦𞤮𞥅𞤣𞤭𞤴𞤢', 'ff-Latn': 'Kambodso', 'fi': 'Kambodža', 'fil': 'Cambodia', 'fo': 'Kambodja', 'fr': 'Cambodge', 'fur': 'Camboze', 'fy': 'Cambodja', 'ga': 'an Chambóid', 'gd': 'Cambuidea', 'gl': 'Camboxa', 'gsw': 'Kambodscha', 'gu': 'કંબોડિયા', 'guz': 'Kambodia', 'ha': 'Kambodiya', 'he': 'קמבודיה', 'hi': 'कंबोडिया', 'hi-Latn': 'Cambodia', 'hr': 'Kambodža', 'hsb': 'Kambodźa', 'ht': 'Cambodge', 'hu': 'Kambodzsa', 'hy': 'Կամբոջա', 'ia': 'Cambodgia', 'id': 'Kamboja', 'ie': 'Cambodja', 'ig': 'Cambodia', 'is': 'Kambódía', 'it': 'Cambogia', 'ja': 'カンボジア', 'jmc': 'Kambodia', 'jv': 'Kamboja', 'ka': 'კამბოჯა', 'kaa': 'Камбоджа', 'kaa-Cyrl': 'Камбоджа', 'kab': 'Cambudya', 'kam': 'Kambodia', 'kde': 'Kambodia', 'kea': 'Kambodja', 'kgp': 'Kỹmója', 'khq': 'Kamboogi', 'ki': 'Kambodia', 'kk': 'Камбоджа', 'kk-Cyrl': 'Камбоджа', 'kln': 'Emetab Cambodia', 'km': 'កម្ពុជា', 'kn': 'ಕಾಂಬೋಡಿಯಾ', 'ko': '캄보디아', 'kok': 'कंबोडिया', 'kok-Deva': 'कंबोडिया', 'ks': 'کَمبوڑِیا', 'ks-Arab': 'کَمبوڑِیا', 'ksb': 'Kambodia', 'ksf': 'kambodj', 'ksh': 'Kambodscha', 'ku': 'Kamboçya', 'kxv': 'kamboḍia', 'kxv-Deva': 'कंबोडिया', 'kxv-Latn': 'kamboḍia', 'kxv-Orya': 'କମ୍ବଡିଆ', 'kxv-Telu': 'కంబోడియా', 'ky': 'Камбоджа', 'lag': 'Kambódia', 'lb': 'Kambodscha', 'lg': 'Kambodya', 'lld': 'Cambogia', 'ln': 'Kambodza', 'lo': 'ກຳປູເຈຍ', 'lt': 'Kambodža', 'lu': 'Kambodza', 'luo': 'Cambodia', 'luy': 'Kambodia', 'lv': 'Kambodža', 'mai': 'कंबोडिया', 'mas': 'Kambodia', 'mer': 'Kambodia', 'mfe': 'Kambodj', 'mg': 'Kambôdja', 'mgh': 'Ukambodia', 'mi': 'Kamapōtia', 'mk': 'Камбоџа', 'ml': 'കംബോഡിയ', 'mn': 'Камбож', 'mr': 'कंबोडिया', 'ms': 'Kemboja', 'mt': 'il-Kambodja', 'mua': 'kambodiya', 'my': 'ကမ္ဘောဒီးယား', 'mzn': 'کامبوج', 'naq': 'Cambodiab', 'nb': 'Kambodsja', 'nd': 'Cambodia', 'ne': 'कम्बोडिया', 'nl': 'Cambodja', 'nmg': 'Kambodia', 'nn': 'Kambodsja', 'no': 'Kambodsja', 'nqo': 'ߞߊ߲ߓߐߘߑߖ', 'nus': 'Kombodia', 'nyn': 'Kambodiya', 'om': 'Kamboodiyaa', 'or': 'କାମ୍ବୋଡିଆ', 'pa': 'ਕੰਬੋਡੀਆ', 'pa-Guru': 'ਕੰਬੋਡੀਆ', 'pcm': 'Kambódia', 'pl': 'Kambodża', 'ps': 'کمبودیا', 'pt': 'Camboja', 'qu': 'Camboya', 'rm': 'Cambodscha', 'rn': 'Kamboje', 'ro': 'Cambodgia', 'rof': 'Kambodia', 'ru': 'Камбоджа', 'rwk': 'Kambodia', 'saq': 'Kambodia', 'sbp': 'Kambodiya', 'sc': 'Cambòdia', 'sd': 'ڪمبوڊيا', 'sd-Arab': 'ڪمبوڊيا', 'se': 'Kambodža', 'seh': 'Camboja', 'ses': 'Kamboogi', 'sg': 'Kämbôzi', 'shi': 'ⴽⴰⵎⴱⵓⴷⵢⴰ', 'shi-Latn': 'kambudya', 'shi-Tfng': 'ⴽⴰⵎⴱⵓⴷⵢⴰ', 'si': 'කාම්බෝජය', 'sk': 'Kambodža', 'sl': 'Kambodža', 'smn': 'Kambodža', 'sn': 'Kambodia', 'so': 'Kamboodiya', 'sq': 'Kamboxhia', 'sr': 'Камбоџа', 'sr-Cyrl': 'Камбоџа', 'sr-Latn': 'Kambodža', 'sv': 'Kambodja', 'sw': 'Kambodia', 'syr': 'ܟܡܒܘܕܝܐ', 'ta': 'கம்போடியா', 'te': 'కంబోడియా', 'teo': 'Kambodia', 'tg': 'Камбоҷа', 'th': 'กัมพูชา', 'ti': 'ካምቦድያ', 'tk': 'Kamboja', 'to': 'Kamipōtia', 'tr': 'Kamboçya', 'tt': 'Камбоджа', 'twq': 'kamboogi', 'tzm': 'Kambudj', 'ug': 'كامبودژا', 'uk': 'Камбоджа', 'ur': 'کمبوڈیا', 'uz': 'Kambodja', 'uz-Cyrl': 'Камбоджа', 'uz-Latn': 'Kambodja', 'vai': 'ꕪꕹꔵꕩ', 'vai-Latn': 'Kaŋbhodiya', 'vai-Vaii': 'ꕪꕹꔵꕩ', 'vec': 'Canboja', 'vi': 'Campuchia', 'vun': 'Kambodia', 'wae': 'Kambodša', 'wo': 'Kàmboj', 'xh': 'ECambodia', 'xnr': 'कंबोडिया', 'xog': 'Kambodya', 'yav': 'Kámbóse', 'yi': 'קאַמבאדיע', 'yo': 'Kàmùbódíà', 'yrl': 'Kãbuya', 'yue': '柬埔寨', 'yue-Hans': '柬埔寨', 'yue-Hant': '柬埔寨', 'zgh': 'ⴽⴰⵎⴱⵓⴷⵢⴰ', 'zh': '柬埔寨', 'zh-Hans': '柬埔寨', 'zh-Hant': '柬埔寨', 'zu': 'i-Cambodia'},
    'KI': {'af': 'Kiribati', 'agq': 'Kèlèbati', 'ak': 'Kiribati', 'am': 'ኪሪባቲ', 'ar': 'كيريباتي', 'as': 'কিৰিবাটি', 'asa': 'Kiribati', 'ast': 'Kiribati', 'az': 'Kiribati', 'az-Cyrl': 'Кирибати', 'az-Latn': 'Kiribati', 'bal-Latn': 'Kiribáti', 'bas': 'Kìrìbatì', 'be': 'Кірыбаці', 'be-tarask': 'Кірыбаці', 'bez': 'Hukilibati', 'bg': 'Кирибати', 'blo': 'Kiribatii', 'bm': 'Kiribati', 'bn': 'কিরিবাতি', 'br': 'Kiribati', 'brx': 'किरिबाती', 'bs': 'Kiribati', 'bs-Cyrl': 'Кирибати', 'bs-Latn': 'Kiribati', 'ca': 'Kiribati', 'ca-ES-valencia': 'Kiribati', 'ccp': '𑄇𑄨𑄢𑄨𑄝𑄖𑄨', 'ce': 'Кирибати', 'ceb': 'Kiribati', 'cgg': 'Kiribati', 'chr': 'ᎧᎵᏆᏘ', 'ckb': 'کیریباس', 'cs': 'Kiribati', 'cv': 'Кирибати', 'cy': 'Kiribati', 'da': 'Kiribati', 'dav': 'Kiribati', 'de': 'Kiribati', 'dje': 'Kiribaati', 'dsb': 'Kiribati', 'dz': 'ཀི་རི་བ་ཏི་མཚོ་གླིང', 'ebu': 'Kiribati', 'ee': 'Kiribati nutome', 'el': 'Κιριμπάτι', 'el-polyton': 'Κιριμπάτι', 'en': 'Kiribati', 'en-Dsrt': '𐐗𐐮𐑉𐐲𐐺𐐪𐐻𐐨', 'eo': 'Kiribato', 'es': 'Kiribati', 'et': 'Kiribati', 'eu': 'Kiribati', 'ewo': 'Kiribatí', 'fa': 'کیریباتی', 'ff': 'Kiribari', 'ff-Adlm': '𞤑𞤭𞤪𞤦𞤢𞤼𞤭𞥅', 'ff-Latn': 'Kiribari', 'fi': 'Kiribati', 'fil': 'Kiribati', 'fo': 'Kiribati', 'fr': 'Kiribati', 'fur': 'Kiribati', 'fy': 'Kiribati', 'ga': 'Ciribeas', 'gd': 'Ciribeas', 'gl': 'Kiribati', 'gsw': 'Kiribaati', 'gu': 'કિરિબાટી', 'guz': 'Kiribati', 'ha': 'Kiribati', 'he': 'קיריבאטי', 'hi': 'किरिबाती', 'hi-Latn': 'Kiribati', 'hr': 'Kiribati', 'hsb': 'Kiribati', 'ht': 'Kiribati', 'hu': 'Kiribati', 'hy': 'Կիրիբատի', 'ia': 'Kiribati', 'id': 'Kiribati', 'ig': 'Kiribati', 'is': 'Kíribatí', 'it': 'Kiribati', 'ja': 'キリバス', 'jmc': 'Kiribati', 'jv': 'Kiribati', 'ka': 'კირიბატი', 'kaa': 'Кирибати', 'kaa-Cyrl': 'Кирибати', 'kab': 'Kiribati', 'kam': 'Kiribati', 'kde': 'Kilibati', 'kea': 'Kiribati', 'kgp': 'Kirimati', 'khq': 'Kiribaati', 'ki': 'Kiribati', 'kk': 'Кирибати', 'kk-Cyrl': 'Кирибати', 'kln': 'Emetab Kiribati', 'km': 'គិរីបាទី', 'kn': 'ಕಿರಿಬಾಟಿ', 'ko': '키리바시', 'kok': 'किरिबाती', 'kok-Deva': 'किरिबाती', 'ks': 'کِرٕباتی', 'ks-Arab': 'کِرٕباتی', 'ksb': 'Kiibati', 'ksf': 'kiribáti', 'ksh': 'Kiribati', 'ku': 'Kirîbatî', 'kxv': 'kiribaṭi', 'kxv-Deva': 'किरिबाती', 'kxv-Latn': 'kiribaṭi', 'kxv-Orya': 'କିରିବାତି', 'kxv-Telu': 'కీరిబాటి', 'ky': 'Кирибати', 'lag': 'Kiribáati', 'lb': 'Kiribati', 'lg': 'Kiribati', 'lld': 'Kiribati', 'ln': 'Kiribati', 'lo': 'ຄິຣິບາທິ', 'lt': 'Kiribatis', 'lu': 'Kiribati', 'luo': 'Kiribati', 'luy': 'Kiribati', 'lv': 'Kiribati', 'mai': 'किरिबाती', 'mas': 'Kiribati', 'mer': 'Kiribati', 'mfe': 'Kiribati', 'mg': 'Kiribati', 'mi': 'Kiripati', 'mk': 'Кирибати', 'ml': 'കിരിബാട്ടി', 'mn': 'Кирибати', 'mr': 'किरीबाटी', 'ms': 'Kiribati', 'mt': 'Kiribati', 'mua': 'Sǝr Kiribati', 'my': 'ခီရီဘာတီ', 'mzn': 'کیریباتی', 'naq': 'Kiribati', 'nb': 'Kiribati', 'nd': 'Khiribati', 'ne': 'किरिबाटी', 'nl': 'Kiribati', 'nmg': 'Kiribati', 'nn': 'Kiribati', 'no': 'Kiribati', 'nqo': 'ߞߙߌߓߊߕߌ߫', 'nyn': 'Kiribati', 'om': 'Kiribaatii', 'or': 'କିରିବାଟି', 'pa': 'ਕਿਰਬਾਤੀ', 'pa-Guru': 'ਕਿਰਬਾਤੀ', 'pcm': 'Kiribáti', 'pl': 'Kiribati', 'ps': 'کیري باتي', 'pt': 'Quiribati', 'qu': 'Kiribati', 'rm': 'Kiribati', 'rn': 'Kiribati', 'ro': 'Kiribati', 'rof': 'Kiribati', 'ru': 'Кирибати', 'rwk': 'Kiribati', 'saq': 'Kiribati', 'sbp': 'Kilibati', 'sc': 'Kiribati', 'sd': 'ڪرباتي', 'sd-Arab': 'ڪرباتي', 'se': 'Kiribati', 'seh': 'Quiribati', 'ses': 'Kiribaati', 'sg': 'Kiribati', 'shi': 'ⴽⵉⵔⵉⴱⴰⵜⵉ', 'shi-Latn': 'kiribati', 'shi-Tfng': 'ⴽⵉⵔⵉⴱⴰⵜⵉ', 'si': 'කිරිබති', 'sk': 'Kiribati', 'sl': 'Kiribati', 'smn': 'Kiribati', 'sn': 'Kiribati', 'so': 'Kiribati', 'sq': 'Kiribati', 'sr': 'Кирибати', 'sr-Cyrl': 'Кирибати', 'sr-Latn': 'Kiribati', 'sv': 'Kiribati', 'sw': 'Kiribati', 'syr': 'ܟܝܪܝܒܬܝ', 'ta': 'கிரிபாட்டி', 'te': 'కిరిబాటి', 'teo': 'Kiribati', 'tg': 'Кирибати', 'th': 'คิริบาส', 'ti': 'ኪሪባቲ', 'tk': 'Kiribati', 'to': 'Kilipasi', 'tr': 'Kiribati', 'tt': 'Кирибати', 'twq': 'Kiribaati', 'tzm': 'Kiribati', 'ug': 'كىرىباتى', 'uk': 'Кірибаті', 'ur': 'کریباتی', 'uz': 'Kiribati', 'uz-Cyrl': 'Кирибати', 'uz-Latn': 'Kiribati', 'vai': 'ꕃꔸꕑꔳ', 'vai-Latn': 'Kiribhati', 'vai-Vaii': 'ꕃꔸꕑꔳ', 'vec': 'Kiribati', 'vi': 'Kiribati', 'vun': 'Kiribati', 'wae': 'Kiribati', 'wo': 'Kiribati', 'xh': 'EKiribati', 'xnr': 'किरिबाती', 'xog': 'Kiribati', 'yav': 'kilipatí', 'yi': 'קיריבאַטי', 'yo': 'Kiribati', 'yrl': 'Kiribati', 'yue': '吉里巴斯', 'yue-Hans': '吉里巴斯', 'yue-Hant': '吉里巴斯', 'zgh': 'ⴽⵉⵔⵉⴱⴰⵜⵉ', 'zh': '基里巴斯', 'zh-Hans': '基里巴斯', 'zh-Hant': '吉里巴斯', 'zu': 'i-Kiribati'},
    'KM': {'af': 'Comore', 'agq': 'Komolòs', 'ak': 'Kɔmɔrɔs', 'am': 'ኮሞሮስ', 'ar': 'جزر القمر', 'as': 'কোমোৰোজ', 'asa': 'Komoro', 'ast': 'Les Comores', 'az': 'Komor adaları', 'az-Cyrl': 'Комор адалары', 'az-Latn': 'Komor adaları', 'bal-Latn': 'Komórós', 'bas': 'Kɔ̀mɔ̂r', 'be': 'Каморскія астравы', 'be-tarask': 'Каморскія астравы', 'bez': 'Hukomoro', 'bg': 'Коморски острови', 'bgn': 'کومورس', 'blo': 'Komɔɔr', 'bm': 'Komɔri', 'bn': 'কমোরোস', 'br': 'Komorez', 'brx': 'कम’र’ज', 'bs': 'Komori', 'bs-Cyrl': 'Комори', 'bs-Latn': 'Komori', 'ca': 'Comores', 'ca-ES-valencia': 'Comores', 'ccp': '𑄇𑄧𑄟𑄮𑄢𑄮𑄌𑄴', 'ce': 'Комораш', 'ceb': 'Comoros', 'cgg': 'Koromo', 'chr': 'ᎪᎼᎳᏍ', 'ckb': 'دوورگەکانی کۆمۆر', 'cs': 'Komory', 'cv': 'Комор утравӗсем', 'cy': 'Comoros', 'da': 'Comorerne', 'dav': 'Komoro', 'de': 'Komoren', 'dje': 'Komoor', 'dsb': 'Komory', 'dyo': 'Komor', 'dz': 'ཀོ་མོ་རོས', 'ebu': 'Komoro', 'ee': 'Komoros nutome', 'el': 'Κομόρες', 'el-polyton': 'Κομόρες', 'en': 'Comoros', 'en-Dsrt': '𐐗𐐪𐑋𐐲𐑉𐐬𐑆', 'eo': 'Komoroj', 'es': 'Comoras', 'et': 'Komoorid', 'eu': 'Komoreak', 'ewo': 'Komɔ́r', 'fa': 'کومور', 'ff': 'Komoor', 'ff-Adlm': '𞤑𞤮𞤥𞤮𞥅𞤪𞤮', 'ff-Latn': 'Komoor', 'fi': 'Komorit', 'fil': 'Comoros', 'fo': 'Komoroyggjar', 'fr': 'Comores', 'fur': 'Comoris', 'fy': 'Comoren', 'ga': 'Oileáin Chomóra', 'gd': 'Comoros', 'gl': 'Comores', 'gsw': 'Komoore', 'gu': 'કોમોરસ', 'guz': 'Komoro', 'ha': 'Kwamoras', 'he': 'קומורו', 'hi': 'कोमोरोस', 'hi-Latn': 'Comoros', 'hr': 'Komori', 'hsb': 'Komory', 'ht': 'Comores', 'hu': 'Comore-szigetek', 'hy': 'Կոմորյան կղզիներ', 'ia': 'Comoros', 'id': 'Komoro', 'ig': 'Comoros', 'is': 'Kómoreyjar', 'it': 'Comore', 'ja': 'コモロ', 'jgo': 'Kɔmɔ́lɔshi', 'jmc': 'Komoro', 'jv': 'Komoro', 'ka': 'კომორის კუნძულები', 'kaa': 'Комор атаўлары', 'kaa-Cyrl': 'Комор атаўлары', 'kab': 'Kumur', 'kam': 'Komoro', 'kde': 'Komolo', 'kea': 'Kamoris', 'kgp': 'Komãre', 'khq': 'Komoor', 'ki': 'Komoro', 'kk': 'Комор аралдары', 'kk-Cyrl': 'Комор аралдары', 'kln': 'Emetab Comoros', 'km': 'កូម័រ', 'kn': 'ಕೊಮೊರೊಸ್', 'ko': '코모로', 'kok': 'कोमोरोस', 'kok-Deva': 'कोमोरोस', 'ks': 'کَمورَس', 'ks-Arab': 'کَمورَس', 'ksb': 'Komolo', 'ksf': 'komɔr', 'ksh': 'de Kommoore', 'ku': 'Komor', 'kxv': 'komoros', 'kxv-Deva': 'कोमोरोस', 'kxv-Latn': 'komoros', 'kxv-Orya': 'କମରସ', 'kxv-Telu': 'కొమొరోస్', 'ky': 'Коморос', 'lag': 'Komóoro', 'lb': 'Komoren', 'lg': 'Bizinga by’eKomoro', 'lld': 'Comores', 'ln': 'Komorɛ', 'lo': 'ໂຄໂມໂຣສ', 'lt': 'Komorai', 'lu': 'Komoru', 'luo': 'Comoros', 'luy': 'Komoro', 'lv': 'Komoru salas', 'mai': 'कोमोरोस', 'mas': 'Komoro', 'mer': 'Komoro', 'mfe': 'Komor', 'mg': 'Kômaoro', 'mgh': 'Ukomoro', 'mi': 'Komoro', 'mk': 'Коморски Острови', 'ml': 'കോമൊറോസ്', 'mn': 'Коморын арлууд', 'mr': 'कोमोरोज', 'ms': 'Comoros', 'mt': 'Comoros', 'mua': 'komora', 'my': 'ကိုမိုရိုစ်', 'mzn': 'کومور', 'naq': 'Comoros', 'nb': 'Komorene', 'nd': 'Khomoro', 'ne': 'कोमोरोस', 'nl': 'Comoren', 'nmg': 'Kɔmɔr', 'nn': 'Komorane', 'no': 'Komorene', 'nqo': 'ߞߡߐߙ', 'nus': 'Komruth', 'nyn': 'Koromo', 'om': 'Komoroos', 'or': 'କୋମୋରସ୍', 'pa': 'ਕੋਮੋਰੋਸ', 'pa-Guru': 'ਕੋਮੋਰੋਸ', 'pcm': 'Kọ́mọ́ros', 'pl': 'Komory', 'ps': 'کوموروس', 'pt': 'Comores', 'qu': 'Comoras', 'rm': 'Comoras', 'rn': 'Izinga rya Komore', 'ro': 'Comore', 'rof': 'Komoro', 'ru': 'Коморы', 'rwk': 'Komoro', 'saq': 'Komoro', 'sbp': 'Komolo', 'sc': 'Comoras', 'sd': 'ڪوموروس', 'sd-Arab': 'ڪوموروس', 'se': 'Komoros', 'seh': 'Comores', 'ses': 'Komoor', 'sg': 'Kömôro', 'shi': 'ⵇⵓⵎⵓⵔ', 'shi-Latn': 'cumur', 'shi-Tfng': 'ⵇⵓⵎⵓⵔ', 'si': 'කොමොරෝස්', 'sk': 'Komory', 'sl': 'Komori', 'smn': 'Komoreh', 'sn': 'Komoro', 'so': 'Komooros', 'sq': 'Komore', 'sr': 'Коморска Острва', 'sr-Cyrl': 'Коморска Острва', 'sr-Latn': 'Komorska Ostrva', 'sv': 'Komorerna', 'sw': 'Komoro', 'syr': 'ܓܙܪܬܐ ܕܩܡܪ', 'ta': 'கோமரோஸ்', 'te': 'కొమొరోస్', 'teo': 'Komoro', 'tg': 'Комор', 'th': 'คอโมโรส', 'ti': 'ኮሞሮስ', 'tk': 'Komor adalary', 'to': 'Komolosi', 'tr': 'Komorlar', 'tt': 'Комор утраулары', 'twq': 'Komoor', 'tzm': 'Qumur', 'ug': 'كومورو', 'uk': 'Комори', 'ur': 'کوموروس', 'uz': 'Komor orollari', 'uz-Cyrl': 'Комор ороллари', 'uz-Latn': 'Komor orollari', 'vai': 'ꖏꕹꖄꔻ', 'vai-Latn': 'Komorosi', 'vai-Vaii': 'ꖏꕹꖄꔻ', 'vec': 'Ìzole Comore', 'vi': 'Comoros', 'vun': 'Komoro', 'wae': 'Komore', 'wo': 'Komoor', 'xh': 'EComoros', 'xnr': 'कोमोरोस', 'xog': 'Ebizinga bya Komoro', 'yav': 'Kɔmɔ́ɔl', 'yi': 'קאמאראס', 'yo': 'Kòmòrósì', 'yrl': 'Kumure-ita', 'yue': '葛摩', 'yue-Hans': '葛摩', 'yue-Hant': '葛摩', 'zgh': 'ⵇⵓⵎⵓⵔ', 'zh': '科摩罗', 'zh-Hans': '科摩罗', 'zh-Hant': '葛摩', 'zu': 'i-Comoros'},
    'KN': {'af': 'Sint Kitts en Nevis', 'agq': 'Sɛ̀n Kî à Nevì', 'ak': 'Saint Kitts ne Nɛves', 'am': 'ቅዱስ ኪትስ እና ኔቪስ', 'ar': 'سانت كيتس ونيفيس', 'as': 'ছেইণ্ট কিটছ আৰু নেভিছ', 'asa': 'Thantakitdhi na Nevith', 'ast': 'Saint Kitts y Nevis', 'az': 'Sent-Kits və Nevis', 'az-Cyrl': 'Сент-Китс вә Невис', 'az-Latn': 'Sent-Kits və Nevis', 'bal-Latn': 'St. Kitts o Newis', 'bas': 'Nûmpubi Kîts nì Nevìs', 'be': 'Сент-Кітс і Невіс', 'be-tarask': 'Сент-Кітс і Невіс', 'bez': 'Husantakitzi na Hunevis', 'bg': 'Сейнт Китс и Невис', 'blo': 'Sɛŋ Kits na Nefis', 'bm': 'Kristɔfo-Senu-ni-Ɲevɛs', 'bn': 'সেন্ট কিটস ও নেভিস', 'br': 'Saint Kitts ha Nevis', 'brx': 'सैन्ट कीत्स आरो नेभिस', 'bs': 'Sveti Kits i Nevis', 'bs-Cyrl': 'Свети Китс и Невис', 'bs-Latn': 'Sveti Kits i Nevis', 'ca': 'Saint Kitts i Nevis', 'ca-ES-valencia': 'Saint Kitts i Nevis', 'ccp': '𑄥𑄬𑄚𑄴𑄑𑄴 𑄇𑄨𑄑𑄴𑄥𑄴 𑄃𑄮 𑄚𑄬𑄞𑄨𑄌𑄴', 'ce': 'Сент-Китс а, Невис а', 'ceb': 'St. Kitts & Nevis', 'cgg': 'Senti Kittis na Nevisi', 'chr': 'ᎤᏓᏅᏘ ᎨᏘᏏ ᎠᎴ ᏁᏪᏏ', 'ckb': 'سەن کیتس و نیڤیس', 'cs': 'Svatý Kryštof a Nevis', 'cv': 'Сент-Китс тата Невис', 'cy': 'Saint Kitts a Nevis', 'da': 'Saint Kitts og Nevis', 'dav': 'Santakitzi na Nevis', 'de': 'St. Kitts und Nevis', 'dje': 'Seŋ Kitts nda Nevis', 'dsb': 'St. Kitts a Nevis', 'dz': 'སེནཊ་ ཀིཊས་ དང་ ནེ་བིས', 'ebu': 'Santakitzi na Nevis', 'ee': 'Saint Kitis kple Nevis nutome', 'el': 'Σεν Κιτς και Νέβις', 'el-polyton': 'Σαὶντ Κὶτς καὶ Νέβις', 'en': 'St. Kitts & Nevis', 'en-Dsrt': '𐐝𐐩𐑌𐐻 𐐗𐐮𐐻𐑅 𐐰𐑌𐐼 𐐤𐐨𐑂𐐮𐑅', 'eo': 'Sankta Kristoforo kaj Neviso', 'es': 'San Cristóbal y Nieves', 'et': 'Saint Kitts ja Nevis', 'eu': 'Saint Kitts eta Nevis', 'ewo': 'Ǹfúfúb-Kilisətóv-ai-Nevis', 'fa': 'سنت کیتس و نویس', 'ff': 'Sent Kits e Newis', 'ff-Adlm': '𞤅𞤼. 𞤑𞤪𞤭𞤧𞤼𞤮𞤦𞤢𞤤 & 𞤐𞤫𞥅𞤾𞤭𞤧', 'ff-Latn': 'Sent Kits e Newis', 'fi': 'Saint Kitts ja Nevis', 'fil': 'St. Kitts & Nevis', 'fo': 'St. Kitts & Nevis', 'fr': 'Saint-Christophe-et-Niévès', 'fur': 'San Kitts e Nevis', 'fy': 'Saint Kitts en Nevis', 'ga': 'San Críostóir-Nimheas', 'gd': 'Naomh Crìstean is Nibheis', 'gl': 'Saint Kitts e Nevis', 'gsw': 'St. Kitts und Niuwis', 'gu': 'સેંટ કિટ્સ અને નેવિસ', 'guz': 'Santakitzi na Nevis', 'ha': 'San Kiti Da Nebis', 'he': 'סנט קיטס ונוויס', 'hi': 'सेंट किट्स और नेविस', 'hi-Latn': 'St. Kitts & Nevis', 'hr': 'Sveti Kristofor i Nevis', 'hsb': 'St. Kitts a Nevis', 'ht': 'Saint-Christophe-et-Niévès', 'hu': 'Saint Kitts és Nevis', 'hy': 'Սենթ Քիտս և Նևիս', 'ia': 'Sancte Christophoro e Nevis', 'id': 'Saint Kitts dan Nevis', 'ig': 'St. Kitts & Nevis', 'is': 'Sankti Kitts og Nevis', 'it': 'Saint Kitts e Nevis', 'ja': 'セントクリストファー・ネーヴィス', 'jmc': 'Santakitzi na Nevis', 'jv': 'Saint Kits lan Nèvis', 'ka': 'სენტ-კიტსი და ნევისი', 'kaa': 'Сент-Китс ҳәм Невис', 'kaa-Cyrl': 'Сент-Китс ҳәм Невис', 'kab': 'San Kits d Nivis', 'kam': 'Santakitzi na Nevis', 'kde': 'Santakitzi na Nevis', 'kea': 'San Kristovan i Nevis', 'kgp': 'Sỹ Krinhtóvỹv kar Nẽvinh', 'khq': 'Seŋ Kitts nda Nevis', 'ki': 'Santakitzi na Nevis', 'kk': 'Сент-Китс және Невис', 'kk-Cyrl': 'Сент-Китс және Невис', 'kln': 'Emetab Saint Kitts ak Nevis', 'km': 'សាំង\u200bគីត និង ណេវីស', 'kn': 'ಸೇಂಟ್ ಕಿಟ್ಸ್ ಮತ್ತು ನೆವಿಸ್', 'ko': '세인트키츠 네비스', 'kok': 'सेंट किट्स आनी नेविस', 'kok-Deva': 'सेंट किट्स आनी नेविस', 'ks': 'سینٹ کِٹَس تہٕ نیوِس', 'ks-Arab': 'سینٹ کِٹَس تہٕ نیوِس', 'ksb': 'Santakitzi na Nevis', 'ksf': 'sɛnkrǝstɔ́f ri nyɛ́vǝ', 'ksh': 'Zint Kitts un Nevis', 'ku': 'Saint Kitts û Nevîs', 'kxv': 'seint kiṭs āḍe nebis', 'kxv-Deva': 'सेंट किट्स ऑड़े नेविस', 'kxv-Latn': 'seint kiṭs āḍe nebis', 'kxv-Orya': 'ସେଣ୍ଟ. କିଟ୍ସ ଅଡ଼େ ନେବିସ', 'kxv-Telu': 'సెయింట్ కిట్స్ మరియు నెవిస్', 'ky': 'Сент-Китс жана Невис', 'lag': 'Mʉtakatíifu kitisi na Nevíisi', 'lb': 'St. Kitts an Nevis', 'lg': 'Senti Kitisi ne Nevisi', 'lld': 'St. Kitts y Nevis', 'ln': 'Sántu krístofe mpé Nevɛ̀s', 'lo': 'ເຊນ ຄິດ ແລະ ເນວິສ', 'lt': 'Sent Kitsas ir Nevis', 'lu': 'Santu krístofe ne Neves', 'luo': 'Saint Kitts gi Nevis', 'luy': 'Santakitzi na Nevis', 'lv': 'Sentkitsa un Nevisa', 'mai': 'सेंट किट्स आ नेविस', 'mas': 'Santakitzi na Nevis', 'mer': 'Santakitzi na Nevis', 'mfe': 'Saint-Christophe-ek-Niévès', 'mg': 'Saint-Christophe-et-Niévès', 'mgh': 'Usantakitzi na Nevis', 'mi': 'Hato Kiti me Newhi', 'mk': 'Свети Китс и Невис', 'ml': 'സെന്റ് കിറ്റ്\u200cസും നെവിസും', 'mn': 'Сент-Киттс ба Невис', 'mr': 'सेंट किट्स आणि नेव्हिस', 'ms': 'Saint Kitts dan Nevis', 'mt': 'Saint Kitts u Nevis', 'mua': 'Sǝr Kristof ne Nievǝ', 'my': 'စိန့်ကစ်နှင့်နီဗီစ်', 'mzn': 'سنت کیتس و نویس', 'naq': 'Saint Kitts and Nevis', 'nb': 'Saint Kitts og Nevis', 'nd': 'Saint Kitts and Nevis', 'ne': 'सेन्ट किट्स र नेभिस', 'nl': 'Saint Kitts en Nevis', 'nmg': 'Saint Kitts na Nevis', 'nn': 'Saint Kitts og Nevis', 'no': 'Saint Kitts og Nevis', 'nqo': 'ߛߍ߲ߕ-ߞߙߌߛߑߕߐߝ ߣߌ߫ ߢߝ߭ߋ߫', 'nyn': 'Senti Kittis na Nevisi', 'om': 'St. Kiitis fi Neevis', 'or': 'ସେଣ୍ଟ କିଟସ୍\u200c ଏବଂ ନେଭିସ୍\u200c', 'pa': 'ਸੇਂਟ ਕਿਟਸ ਐਂਡ ਨੇਵਿਸ', 'pa-Guru': 'ਸੇਂਟ ਕਿਟਸ ਐਂਡ ਨੇਵਿਸ', 'pcm': 'Sent Kits & Nẹ́vis', 'pl': 'Saint Kitts i Nevis', 'ps': 'سینټ کټس او نیویس', 'pt': 'São Cristóvão e Névis', 'qu': 'San Cristóbal y Nieves', 'rm': 'Saint Kitts e Nevis', 'rn': 'Sekitsi na Nevisi', 'ro': 'Saint Kitts și Nevis', 'rof': 'Santakitzi na Nevis', 'ru': 'Сент-Китс и Невис', 'rwk': 'Santakitzi na Nevis', 'saq': 'Santakitzi na Nevis', 'sbp': 'Santakitisi ni Nevisi', 'sc': 'Santu Cristolu e Nevis', 'sd': 'سينٽ ڪٽس و نيوس', 'sd-Arab': 'سينٽ ڪٽس و نيوس', 'se': 'Saint Kitts ja Nevis', 'seh': 'São Cristovão e Nevis', 'ses': 'Seŋ Kitts nda Nevis', 'sg': 'Sên-Krïstôfo-na-Nevîsi', 'shi': 'ⵙⴰⵏⴽⵔⵉⵙ ⴷ ⵏⵉⴼⵉⵙ', 'shi-Latn': 'sankris d nifis', 'shi-Tfng': 'ⵙⴰⵏⴽⵔⵉⵙ ⴷ ⵏⵉⴼⵉⵙ', 'si': 'ශාන්ත කිට්ස් සහ නේවිස්', 'sk': 'Svätý Krištof a Nevis', 'sl': 'Saint Kitts in Nevis', 'smn': 'St. Kitts já Nevis', 'sn': 'Saint Kitts and Nevis', 'so': 'St. Kitts iyo Nevis', 'sq': 'Shën-Kits dhe Nevis', 'sr': 'Сент Китс и Невис', 'sr-Cyrl': 'Сент Китс и Невис', 'sr-Latn': 'Sent Kits i Nevis', 'sv': 'S:t Kitts och Nevis', 'sw': 'St. Kitts na Nevis', 'syr': 'ܣܐܢܬ ܟܝܬܣ ܘܢܝܒܝܣ', 'ta': 'செயின்ட் கிட்ஸ் & நெவிஸ்', 'te': 'సెయింట్ కిట్స్ మరియు నెవిస్', 'teo': 'Santakitzi na Nevis', 'tg': 'Сент-Китс ва Невис', 'th': 'เซนต์คิตส์และเนวิส', 'ti': 'ቅዱስ ኪትስን ኔቪስን', 'tk': 'Sent-Kits we Newis', 'to': 'Sā Kitisi mo Nevisi', 'tr': 'Saint Kitts ve Nevis', 'tt': 'Сент-Китс һәм Невис', 'twq': 'Seŋ Kitts nda Nevis', 'tzm': 'Santekits d Nivis', 'ug': 'ساينت كىتىس ۋە نېۋىس', 'uk': 'Сент-Кітс і Невіс', 'ur': 'سینٹ کٹس اور نیویس', 'uz': 'Sent-Kits va Nevis', 'uz-Cyrl': 'Сент-Китс ва Невис', 'uz-Latn': 'Sent-Kits va Nevis', 'vai': 'ꔻꘋ ꕃꔳꔻ ꗪ ꔕꔲꔻ', 'vai-Latn': 'Siŋ Kisi ɓɛ́ Nevisi', 'vai-Vaii': 'ꔻꘋ ꕃꔳꔻ ꗪ ꔕꔲꔻ', 'vec': 'S. Cristofer e Nevis', 'vi': 'St. Kitts và Nevis', 'vun': 'Santakitzi na Nevis', 'wae': 'St. Kitts und Nevis', 'wo': 'Saŋ Kits ak Newis', 'xh': 'ESt. Kitts & Nevis', 'xnr': 'सेंट किट्स कने नेविस', 'xog': 'Senti Kitisi ne Nevisi', 'yav': 'sɛ́ŋkilistɔ́f eniɛ́f', 'yo': 'Kiiti ati Neefi', 'yrl': 'Sã Kirituwãu suí Newi', 'yue': '聖基茨同尼維斯', 'yue-Hans': '圣基茨同尼维斯', 'yue-Hant': '聖基茨同尼維斯', 'zgh': 'ⵙⴰⵏⴽⵔⵉⵙ ⴷ ⵏⵉⴼⵉⵙ', 'zh': '圣基茨和尼维斯', 'zh-Hans': '圣基茨和尼维斯', 'zh-Hant': '聖克里斯多福及尼維斯', 'zu': 'i-Saint Kitts ne-Nevis'},
    'KP': {'af': 'Noord-Korea', 'agq': 'Kùulîa, Ekùw', 'ak': 'Korea Atifi', 'am': 'ሰሜን ኮሪያ', 'ar': 'كوريا الشمالية', 'as': 'উত্তৰ কোৰিয়া', 'asa': 'Korea Kathkazini', 'ast': 'Corea del Norte', 'az': 'Şimali Koreya', 'az-Cyrl': 'Шимали Кореја', 'az-Latn': 'Şimali Koreya', 'bal-Latn': 'Shamáli Kóriá', 'bas': 'Kɔ̀re ì Ŋ̀ɔmbɔk', 'be': 'Паўночная Карэя', 'be-tarask': 'Паўночная Карэя', 'bez': 'Hukolea Kaskazini', 'bg': 'Северна Корея', 'blo': 'Koree gʊnyɩpɛnɛlaŋ', 'bm': 'Kɛɲɛka Kore', 'bn': 'উত্তর কোরিয়া', 'br': 'Korea an Norzh', 'brx': 'साहा करिया', 'bs': 'Sjeverna Koreja', 'bs-Cyrl': 'Сјеверна Кореја', 'bs-Latn': 'Sjeverna Koreja', 'ca': 'Corea del Nord', 'ca-ES-valencia': 'Corea del Nord', 'ccp': '𑄅𑄪𑄖𑄴𑄖𑄮𑄢𑄴 𑄇𑄮𑄢𑄨𑄠', 'ce': 'Къилбаседа Корей', 'ceb': 'North Korea', 'cgg': 'Koreya Amatemba', 'chr': 'ᏧᏴᏢ ᎪᎵᎠ', 'ckb': 'کۆریای باکوور', 'cs': 'Severní Korea', 'cv': 'КХДР', 'cy': 'Gogledd Corea', 'da': 'Nordkorea', 'dav': 'Korea Kaskazini', 'de': 'Nordkorea', 'dje': 'Gurma Kooree', 'dsb': 'Pódpołnocna Koreja', 'dz': 'བྱང་ ཀོ་རི་ཡ', 'ebu': 'Korea Kaskazini', 'ee': 'Dziehe Korea nutome', 'el': 'Βόρεια Κορέα', 'el-polyton': 'Βόρεια Κορέα', 'en': 'North Korea', 'en-Dsrt': '𐐤𐐱𐑉𐑃 𐐗𐐲𐑉𐐨𐐲', 'eo': 'Nord-Koreo', 'es': 'Corea del Norte', 'et': 'Põhja-Korea', 'eu': 'Ipar Korea', 'ewo': 'Koré yá Nór', 'fa': 'کرهٔ شمالی', 'ff': 'Koree Rewo', 'ff-Adlm': '𞤑𞤮𞥅𞤪𞤫𞤴𞤢𞥄 𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫', 'ff-Latn': 'Koree Rewo', 'fi': 'Pohjois-Korea', 'fil': 'Hilagang Korea', 'fo': 'Norðurkorea', 'fr': 'Corée du Nord', 'fur': 'Coree dal nord', 'fy': 'Noard-Korea', 'ga': 'an Chóiré Thuaidh', 'gd': 'Coirèa a Tuath', 'gl': 'Corea do Norte', 'gsw': 'Demokraatischi Volksrepublik Koreea', 'gu': 'ઉત્તર કોરિયા', 'guz': 'Korea Kaskazini', 'ha': 'Koriya Ta Arewa', 'he': 'קוריאה הצפונית', 'hi': 'उत्तर कोरिया', 'hi-Latn': 'North Korea', 'hr': 'Sjeverna Koreja', 'hsb': 'Sewjerna Koreja', 'ht': 'Corée du Nord', 'hu': 'Észak-Korea', 'hy': 'Հյուսիսային Կորեա', 'ia': 'Corea del Nord', 'id': 'Korea Utara', 'ig': 'North Korea', 'is': 'Norður-Kórea', 'it': 'Corea del Nord', 'ja': '北朝鮮', 'jmc': 'Korea Kaskazini', 'jv': 'Korea Lor', 'ka': 'ჩრდილოეთ კორეა', 'kaa': 'Арқа Корея', 'kaa-Cyrl': 'Арқа Корея', 'kab': 'Kurya, Ufella', 'kam': 'Korea Kaskazini', 'kde': 'Kolea Kasikazini', 'kea': 'Koreia di Norti', 'kgp': 'Nãrti-Koréja', 'khq': 'Kooree, Gurma', 'ki': 'Korea Kaskazini', 'kk': 'Солтүстік Корея', 'kk-Cyrl': 'Солтүстік Корея', 'kln': 'Emetab Korea nebo murot katam', 'km': 'កូរ៉េ\u200bខាង\u200bជើង', 'kn': 'ಉತ್ತರ ಕೊರಿಯಾ', 'ko': '북한', 'kok': 'उत्तर कोरिया', 'kok-Deva': 'उत्तर कोरिया', 'ks': 'شُمٲلی کورِیا', 'ks-Arab': 'شُمٲلی کورِیا', 'ksb': 'Kolea Kaskazini', 'ksf': 'korɛanɔ́r', 'ksh': 'Nood-Koreja', 'ku': 'Koreya Bakur', 'kxv': 'uttar koriya', 'kxv-Deva': 'उतर कोरिया', 'kxv-Latn': 'uttar koriya', 'kxv-Orya': 'ଉତର କରିୟା', 'kxv-Telu': 'ఉత్తర కొరియా', 'ky': 'Түндүк Корея', 'lag': 'Koréa yʉ ʉtʉrʉko', 'lb': 'Nordkorea', 'lg': 'Koreya ey’omumambuka', 'lld': 'Corea dl Nord', 'ln': 'Korɛ ya nɔ́rdi', 'lo': 'ເກົາຫລີເໜືອ', 'lt': 'Šiaurės Korėja', 'lu': 'Kore wa muulu', 'luo': 'Korea Masawa', 'luy': 'Korea Kaskazini', 'lv': 'Ziemeļkoreja', 'mai': 'उत्तर कोरिया', 'mas': 'Korea Kaskazini', 'mer': 'Korea Nothi', 'mfe': 'Lakore-dinor', 'mg': 'Korea Avaratra', 'mgh': 'Ukorea Kaskazini', 'mi': 'Kōrea ki te Raki', 'mk': 'Северна Кореја', 'ml': 'ഉത്തരകൊറിയ', 'mn': 'Хойд Солонгос', 'mr': 'उत्तर कोरिया', 'ms': 'Korea Utara', 'mt': 'il-Korea ta’ Fuq', 'mua': 'Sǝr Kore fah sǝŋ', 'my': 'မြောက်ကိုရီးယား', 'mzn': 'شمالی کُره', 'naq': 'Koreab, Noord', 'nb': 'Nord-Korea', 'nd': 'North Korea', 'ne': 'उत्तर कोरिया', 'nl': 'Noord-Korea', 'nmg': 'Koré yí bvuɔ', 'nn': 'Nord-Korea', 'no': 'Nord-Korea', 'nqo': 'ߞߐ߬ߘߎ߮ ߞߏ߯ߙߋ߫', 'nyn': 'Koreya Amatemba', 'om': 'Kooriyaa Kaaba', 'or': 'ଉତ୍ତର କୋରିଆ', 'pa': 'ਉੱਤਰ ਕੋਰੀਆ', 'pa-Guru': 'ਉੱਤਰ ਕੋਰੀਆ', 'pcm': 'Nọ́t Koria', 'pl': 'Korea Północna', 'ps': 'شمالی کوریا', 'pt': 'Coreia do Norte', 'qu': 'Corea del Norte', 'rm': 'Corea dal Nord', 'rn': 'Koreya y’amajaruguru', 'ro': 'Coreea de Nord', 'rof': 'Korea Kaskazini', 'ru': 'КНДР', 'rwk': 'Korea Kaskazini', 'saq': 'Korea Kaskazini', 'sbp': 'Koleya ya luvala lwa Kunyamande', 'sc': 'Corea de su Nord', 'sd': 'اتر ڪوريا', 'sd-Arab': 'اتر ڪوريا', 'se': 'Davvi-Korea', 'seh': 'Coréia do Norte', 'ses': 'Kooree, Gurma', 'sg': 'Korëe tî Banga', 'shi': 'ⴽⵓⵔⵢⴰ ⵏ ⵉⵥⵥⵍⵎⴹ', 'shi-Latn': 'kurya n iẓẓlmḍ', 'shi-Tfng': 'ⴽⵓⵔⵢⴰ ⵏ ⵉⵥⵥⵍⵎⴹ', 'si': 'උතුරු කොරියාව', 'sk': 'Severná Kórea', 'sl': 'Severna Koreja', 'smn': 'Tave-Korea', 'sn': 'Korea, North', 'so': 'Kuuriyada Waqooyi', 'sq': 'Kore e Veriut', 'sr': 'Северна Кореја', 'sr-Cyrl': 'Северна Кореја', 'sr-Latn': 'Severna Koreja', 'sv': 'Nordkorea', 'sw': 'Korea Kaskazini', 'syr': 'ܟܘܪܝܐ ܕܓܪܒܝܐ', 'ta': 'வட கொரியா', 'te': 'ఉత్తర కొరియా', 'teo': 'Korea Kaskazini', 'tg': 'Кореяи Шимолӣ', 'th': 'เกาหลีเหนือ', 'ti': 'ሰሜን ኮርያ', 'tk': 'Demirgazyk Koreýa', 'to': 'Kōlea tokelau', 'tr': 'Kuzey Kore', 'tt': 'Төньяк Корея', 'twq': 'Kooree, Gurma', 'tzm': 'Kurya Tugafat', 'ug': 'چاۋشيەن', 'uk': 'Північна Корея', 'ur': 'شمالی کوریا', 'uz': 'Shimoliy Koreya', 'uz-Cyrl': 'Шимолий Корея', 'uz-Latn': 'Shimoliy Koreya', 'vai': 'ꖏꔸꕩ ꗛꔤ ꕪꘋꗒ', 'vai-Latn': 'Koriya Kɔi Kaŋndɔ', 'vai-Vaii': 'ꖏꔸꕩ ꗛꔤ ꕪꘋꗒ', 'vec': 'Corèa de’l nord', 'vi': 'Triều Tiên', 'vun': 'Korea Kaskazini', 'wae': 'Nordkorea', 'wo': 'Kore Noor', 'xh': 'EMntla Korea', 'xnr': 'उत्तर कोरिया', 'xog': 'Koreya eya mumambuka', 'yav': 'kɔlé u muɛnɛ́', 'yo': 'Guusu Kọria', 'yrl': 'Kureya Nuti suí', 'yue': '北韓', 'yue-Hans': '北韩', 'yue-Hant': '北韓', 'zgh': 'ⴽⵓⵔⵢⴰ ⵏ ⵉⵥⵥⵍⵎⴹ', 'zh': '朝鲜', 'zh-Hans': '朝鲜', 'zh-Hant': '北韓', 'zu': 'i-North Korea'},
    'KR': {'af': 'Suid-Korea', 'agq': 'Kùulîa, Emàm', 'ak': 'Korea Anaafoɔ', 'am': 'ደቡብ ኮሪያ', 'ar': 'كوريا الجنوبية', 'as': 'দক্ষিণ কোৰিয়া', 'asa': 'Korea Kuthini', 'ast': 'Corea del Sur', 'az': 'Cənubi Koreya', 'az-Cyrl': 'Ҹәнуби Кореја', 'az-Latn': 'Cənubi Koreya', 'bal-Latn': 'Zerbári Kóriá', 'bas': 'Kɔ̀re ì Ŋ̀wɛ̀lmbɔk', 'be': 'Паўднёвая Карэя', 'be-tarask': 'Паўднёвая Карэя', 'bez': 'Hukolea Kusini', 'bg': 'Южна Корея', 'blo': 'Koree gʊnyɩsonolaŋ', 'bm': 'Worodugu Kore', 'bn': 'দক্ষিণ কোরিয়া', 'bo': 'ལྷོ་ཀོ་རི་ཡ།', 'br': 'Korea ar Su', 'brx': 'खोला करिया', 'bs': 'Južna Koreja', 'bs-Cyrl': 'Јужна Кореја', 'bs-Latn': 'Južna Koreja', 'ca': 'Corea del Sud', 'ca-ES-valencia': 'Corea del Sud', 'ccp': '𑄘𑄧𑄉𑄨𑄚𑄴 𑄇𑄮𑄢𑄨𑄠', 'ce': 'Къилба Корей', 'ceb': 'South Korea', 'cgg': 'Koreya Amashuuma', 'chr': 'ᏧᎦᏃᏮ ᎪᎵᎠ', 'ckb': 'کۆریای باشوور', 'cs': 'Jižní Korea', 'cv': 'Корей Республики', 'cy': 'De Corea', 'da': 'Sydkorea', 'dav': 'Korea Kusini', 'de': 'Südkorea', 'dje': 'Hawsa Kooree', 'dsb': 'Pódpołdnjowa Koreja', 'dz': 'ལྷོ་ ཀོ་རི་ཡ', 'ebu': 'Korea Kusini', 'ee': 'Anyiehe Korea nutome', 'el': 'Νότια Κορέα', 'el-polyton': 'Νότια Κορέα', 'en': 'South Korea', 'en-Dsrt': '𐐝𐐵𐑃 𐐗𐐲𐑉𐐨𐐲', 'eo': 'Sud-Koreo', 'es': 'Corea del Sur', 'et': 'Lõuna-Korea', 'eu': 'Hego Korea', 'ewo': 'Koré yá Súd', 'fa': 'کرهٔ جنوبی', 'ff': 'Koree Worgo', 'ff-Adlm': '𞤑𞤮𞥅𞤪𞤫𞤴𞤢𞥄 𞤙𞤢𞤥𞤲𞤢𞥄𞤲𞤺𞤫', 'ff-Latn': 'Koree Worgo', 'fi': 'Etelä-Korea', 'fil': 'Timog Korea', 'fo': 'Suðurkorea', 'fr': 'Corée du Sud', 'fur': 'Coree dal sud', 'fy': 'Sûd-Korea', 'ga': 'an Chóiré Theas', 'gd': 'Coirèa', 'gl': 'Corea do Sur', 'gsw': 'Republik Koreea', 'gu': 'દક્ષિણ કોરિયા', 'guz': 'Korea Kusini', 'ha': 'Koriya Ta Kudu', 'he': 'קוריאה הדרומית', 'hi': 'दक्षिण कोरिया', 'hi-Latn': 'South Korea', 'hr': 'Južna Koreja', 'hsb': 'Južna Koreja', 'ht': 'Corée du Sud', 'hu': 'Dél-Korea', 'hy': 'Հարավային Կորեա', 'ia': 'Corea del Sud', 'id': 'Korea Selatan', 'ig': 'South Korea', 'is': 'Suður-Kórea', 'it': 'Corea del Sud', 'ja': '韓国', 'jmc': 'Korea Kusini', 'jv': 'Koréa Kidul', 'ka': 'სამხრეთ კორეა', 'kaa': 'Қубла Корея', 'kaa-Cyrl': 'Қубла Корея', 'kab': 'Kurya, Wadda', 'kam': 'Korea Kusini', 'kde': 'Kolea Kusini', 'kea': 'Koreia di Sul', 'kgp': 'Sur-Koréja', 'khq': 'Kooree, Hawsa', 'ki': 'Korea Kusini', 'kk': 'Оңтүстік Корея', 'kk-Cyrl': 'Оңтүстік Корея', 'kln': 'Emetab korea nebo murot tai', 'km': 'កូរ៉េ\u200bខាង\u200bត្បូង', 'kn': 'ದಕ್ಷಿಣ ಕೊರಿಯಾ', 'ko': '대한민국', 'kok': 'दक्षिण कोरिया', 'kok-Deva': 'दक्षिण कोरिया', 'ks': 'جنوٗبی کورِیا', 'ks-Arab': 'جنوٗبی کورِیا', 'ksb': 'Kolea Kusini', 'ksf': 'korɛasud', 'ksh': 'Söd-Koreja', 'ku': 'Koreya Başûr', 'kxv': 'dkiṇ koriyaa', 'kxv-Deva': 'दॉकिण कोरिया', 'kxv-Latn': 'dkiṇ koriyaa', 'kxv-Orya': 'ଦକିଣ କରିୟା', 'kxv-Telu': 'దకిణ కొరియా', 'ky': 'Түштүк Корея', 'lag': 'Koréa ya Saame', 'lb': 'Südkorea', 'lg': 'Koreya ey’omumaserengeta', 'lld': 'Corea dl Süd', 'ln': 'Korɛ ya súdi', 'lo': 'ເກົາຫລີໃຕ້', 'lt': 'Pietų Korėja', 'lu': 'Kore wa mwinshi', 'luo': 'Korea Milambo', 'luy': 'Korea Kusini', 'lv': 'Dienvidkoreja', 'mai': 'दक्षिण कोरिया', 'mas': 'Korea Kusini', 'mer': 'Korea Saũthi', 'mfe': 'Lakore-disid', 'mg': 'Korea Atsimo', 'mgh': 'Ukorea Kusini', 'mi': 'Kōrea ki te Tonga', 'mk': 'Јужна Кореја', 'ml': 'ദക്ഷിണകൊറിയ', 'mn': 'Өмнөд Солонгос', 'mr': 'दक्षिण कोरिया', 'ms': 'Korea Selatan', 'mt': 'il-Korea t’Isfel', 'mua': 'Sǝr Kore nekǝsǝŋ', 'my': 'တောင်ကိုရီးယား', 'mzn': 'جنوبی کُره', 'naq': 'Koreab, Suid', 'nb': 'Sør-Korea', 'nd': 'South Korea', 'ne': 'दक्षिण कोरिया', 'nl': 'Zuid-Korea', 'nmg': 'Koré yí sí', 'nn': 'Sør-Korea', 'no': 'Sør-Korea', 'nqo': 'ߕߟߋ߬ߓߋ ߞߏ߯ߙߋ߫', 'nyn': 'Koreya Amashuuma', 'om': 'Kooriyaa Kibbaa', 'or': 'ଦକ୍ଷିଣ କୋରିଆ', 'pa': 'ਦੱਖਣ ਕੋਰੀਆ', 'pa-Guru': 'ਦੱਖਣ ਕੋਰੀਆ', 'pcm': 'Saút Koria', 'pl': 'Korea Południowa', 'ps': 'سویلي کوریا', 'pt': 'Coreia do Sul', 'qu': 'Corea del Sur', 'rm': 'Corea dal Sid', 'rn': 'Koreya y’amajepfo', 'ro': 'Coreea de Sud', 'rof': 'Korea Kusini', 'ru': 'Республика Корея', 'rwk': 'Korea Kusini', 'saq': 'Korea Kusini', 'sbp': 'Koleya ya Kusini', 'sc': 'Corea de su Sud', 'sd': 'ڏکڻ ڪوريا', 'sd-Arab': 'ڏکڻ ڪوريا', 'se': 'Mátta-Korea', 'seh': 'Coréia do Sul', 'ses': 'Kooree, Hawsa', 'sg': 'Korëe tî Mbongo', 'shi': 'ⴽⵓⵔⵢⴰ ⵏ ⵉⴼⴼⵓⵙ', 'shi-Latn': 'kurya n iffus', 'shi-Tfng': 'ⴽⵓⵔⵢⴰ ⵏ ⵉⴼⴼⵓⵙ', 'si': 'දකුණු කොරියාව', 'sk': 'Južná Kórea', 'sl': 'Južna Koreja', 'smn': 'Maadâ-Korea', 'sn': 'Korea, South', 'so': 'Kuuriyada Koonfureed', 'sq': 'Kore e Jugut', 'sr': 'Јужна Кореја', 'sr-Cyrl': 'Јужна Кореја', 'sr-Latn': 'Južna Koreja', 'sv': 'Sydkorea', 'sw': 'Korea Kusini', 'syr': 'ܟܘܪܝܐ ܕܬܝܡܢܝܐ', 'ta': 'தென் கொரியா', 'te': 'దక్షిణ కొరియా', 'teo': 'Korea Kusini', 'tg': 'Кореяи Ҷанубӣ', 'th': 'เกาหลีใต้', 'ti': 'ደቡብ ኮርያ', 'tk': 'Günorta Koreýa', 'to': 'Kōlea tonga', 'tr': 'Güney Kore', 'tt': 'Көньяк Корея', 'twq': 'Kooree, Hawsa', 'tzm': 'Kurya Tunẓult', 'ug': 'كورېيە', 'uk': 'Південна Корея', 'ur': 'جنوبی کوریا', 'uz': 'Janubiy Koreya', 'uz-Cyrl': 'Жанубий Корея', 'uz-Latn': 'Janubiy Koreya', 'vai': 'ꖏꔸꕩ ꗛꔤ ꔒꘋꗣ ꗏ', 'vai-Latn': 'Koriya Kɔi Leŋŋɛ Lɔ', 'vai-Vaii': 'ꖏꔸꕩ ꗛꔤ ꔒꘋꗣ ꗏ', 'vec': 'Corèa de’l sud', 'vi': 'Hàn Quốc', 'vun': 'Korea Kusini', 'wae': 'Südkorea', 'wo': 'Corée du Sud', 'xh': 'EMzantsi Korea', 'xnr': 'दखण कोरिया', 'xog': 'Koreya eya mumaserengeta', 'yav': 'kɔlé wu mbát', 'yo': 'Ariwa Kọria', 'yrl': 'Kureya Su suí', 'yue': '南韓', 'yue-Hans': '南韩', 'yue-Hant': '南韓', 'zgh': 'ⴽⵓⵔⵢⴰ ⵏ ⵉⴼⴼⵓⵙ', 'zh': '韩国', 'zh-Hans': '韩国', 'zh-Hant': '南韓', 'zu': 'i-South Korea'},
    'KW': {'af': 'Koeweit', 'agq': 'Kùwɛ̂', 'ak': 'Kuweti', 'am': 'ኩዌት', 'ar': 'الكويت', 'as': 'কুৱেইট', 'asa': 'Kuwaiti', 'ast': 'Kuwait', 'az': 'Küveyt', 'az-Cyrl': 'Күвејт', 'az-Latn': 'Küveyt', 'bal-Latn': 'Kwayt', 'bas': 'Kòwêt', 'be': 'Кувейт', 'be-tarask': 'Кувейт', 'bez': 'Hukuwaiti', 'bg': 'Кувейт', 'bgn': 'کویٹ', 'blo': 'Koweeti', 'bm': 'Kowɛti', 'bn': 'কুয়েত', 'br': 'Koweit', 'brx': 'कुवैत', 'bs': 'Kuvajt', 'bs-Cyrl': 'Кувајт', 'bs-Latn': 'Kuvajt', 'ca': 'Kuwait', 'ca-ES-valencia': 'Kuwait', 'ccp': '𑄇𑄪𑄠𑄬𑄖𑄴', 'ce': 'Кувейт', 'ceb': 'Kuwait', 'cgg': 'Kuweiti', 'chr': 'ᎫᏪᎢᏘ', 'ckb': 'کوەیت', 'cs': 'Kuvajt', 'cv': 'Кувейт', 'cy': 'Kuwait', 'da': 'Kuwait', 'dav': 'Kuwaiti', 'de': 'Kuwait', 'dje': 'Kuweet', 'dsb': 'Kuwait', 'dz': 'ཀུ་ཝེཊ', 'ebu': 'Kuwaiti', 'ee': 'Kuwait nutome', 'el': 'Κουβέιτ', 'el-polyton': 'Κουβέιτ', 'en': 'Kuwait', 'en-Dsrt': '𐐗𐐲𐐶𐐩𐐻', 'eo': 'Kuvajto', 'es': 'Kuwait', 'et': 'Kuveit', 'eu': 'Kuwait', 'ewo': 'Kowɛ́d', 'fa': 'کویت', 'ff': 'Kuweyti', 'ff-Adlm': '𞤑𞤵𞤱𞤢𞤴𞤼𞤵', 'ff-Latn': 'Kuweyti', 'fi': 'Kuwait', 'fil': 'Kuwait', 'fo': 'Kuvait', 'fr': 'Koweït', 'fur': 'Kuwait', 'fy': 'Koeweit', 'ga': 'Cuáit', 'gd': 'Cuibhèit', 'gl': 'Kuwait', 'gsw': 'Kuwäit', 'gu': 'કુવૈત', 'guz': 'Kuwaiti', 'ha': 'Kuwet', 'he': 'כווית', 'hi': 'कुवैत', 'hi-Latn': 'Kuwait', 'hr': 'Kuvajt', 'hsb': 'Kuwait', 'ht': 'Koweït', 'hu': 'Kuvait', 'hy': 'Քուվեյթ', 'ia': 'Kuwait', 'id': 'Kuwait', 'ig': 'Kuwait', 'is': 'Kúveit', 'it': 'Kuwait', 'ja': 'クウェート', 'jmc': 'Kuwaiti', 'jv': 'Kuwait', 'ka': 'ქუვეითი', 'kaa': 'Кувейт', 'kaa-Cyrl': 'Кувейт', 'kab': 'Kuwayt', 'kam': 'Kuwaiti', 'kde': 'Kuwaiti', 'kea': 'Kueiti', 'kgp': 'Kuvajti', 'khq': 'Kuweet', 'ki': 'Kuwaiti', 'kk': 'Кувейт', 'kk-Cyrl': 'Кувейт', 'kln': 'Emetab Kuwait', 'km': 'កូវ៉ែត', 'kn': 'ಕುವೈತ್', 'ko': '쿠웨이트', 'kok': 'कुवेत', 'kok-Deva': 'कुवेत', 'ks': 'کُویت', 'ks-Arab': 'کُویت', 'ksb': 'Kuwaiti', 'ksf': 'kuwɛit', 'ksh': 'Kuweit', 'ku': 'Kuweyt', 'kxv': 'kuvet', 'kxv-Deva': 'कुवेत', 'kxv-Latn': 'kuvet', 'kxv-Orya': 'କୁୱେତ୍', 'kxv-Telu': 'కువైట్', 'ky': 'Кувейт', 'lag': 'Kʉwáiti', 'lb': 'Kuwait', 'lg': 'Kuweti', 'lld': 'Kuwait', 'ln': 'Koweti', 'lo': 'ກູເວດ', 'lt': 'Kuveitas', 'lu': 'Koweti', 'luo': 'Kuwait', 'luy': 'Kuwaiti', 'lv': 'Kuveita', 'mai': 'कुवैत', 'mas': 'Kuwaiti', 'mer': 'Kuwĩ ti', 'mfe': 'Koweit', 'mg': 'Kôeity', 'mi': 'Kūweiti', 'mk': 'Кувајт', 'ml': 'കുവൈറ്റ്', 'mn': 'Кувейт', 'mr': 'कुवेत', 'ms': 'Kuwait', 'mt': 'il-Kuwajt', 'mua': 'Sǝr Kowait', 'my': 'ကူဝိတ်', 'mzn': 'کویت', 'naq': 'Kuwaiti', 'nb': 'Kuwait', 'nd': 'Khuweiti', 'ne': 'कुवेत', 'nl': 'Koeweit', 'nmg': 'Kowɛit', 'nn': 'Kuwait', 'no': 'Kuwait', 'nqo': 'ߞߎ߬ߥߊ߬ߕ', 'nyn': 'Kuweiti', 'om': 'Kuweet', 'or': 'କୁଏତ୍', 'pa': 'ਕੁਵੈਤ', 'pa-Guru': 'ਕੁਵੈਤ', 'pcm': 'Kuwét', 'pl': 'Kuwejt', 'ps': 'کويت', 'pt': 'Kuwait', 'qu': 'Kuwait', 'rm': 'Kuwait', 'rn': 'Koweti', 'ro': 'Kuweit', 'rof': 'Kuwaiti', 'ru': 'Кувейт', 'rwk': 'Kuwaiti', 'saq': 'Kuwaiti', 'sbp': 'Kuwaiti', 'sc': 'Kuwait', 'sd': 'ڪويت', 'sd-Arab': 'ڪويت', 'se': 'Kuwait', 'seh': 'Kuwait', 'ses': 'Kuweet', 'sg': 'Köwêti', 'shi': 'ⵍⴽⵡⵉⵜ', 'shi-Latn': 'lkwit', 'shi-Tfng': 'ⵍⴽⵡⵉⵜ', 'si': 'කුවේටය', 'sk': 'Kuvajt', 'sl': 'Kuvajt', 'smn': 'Kuwait', 'sn': 'Kuwait', 'so': 'Kuwayt', 'sq': 'Kuvajt', 'sr': 'Кувајт', 'sr-Cyrl': 'Кувајт', 'sr-Latn': 'Kuvajt', 'sv': 'Kuwait', 'sw': 'Kuwait', 'syr': 'ܟܘܝܬ', 'ta': 'குவைத்', 'te': 'కువైట్', 'teo': 'Kuwaiti', 'tg': 'Қувайт', 'th': 'คูเวต', 'ti': 'ኩዌት', 'tk': 'Kuweýt', 'to': 'Kueiti', 'tr': 'Kuveyt', 'tt': 'Күвәйт', 'twq': 'Kuweet', 'tzm': 'Kuwwayt', 'ug': 'كۇۋەيت', 'uk': 'Кувейт', 'ur': 'کویت', 'uz': 'Quvayt', 'uz-Cyrl': 'Қувайт', 'uz-Latn': 'Quvayt', 'vai': 'ꖴꔃꔳ', 'vai-Latn': 'Kuweti', 'vai-Vaii': 'ꖴꔃꔳ', 'vec': 'Kuwait', 'vi': 'Kuwait', 'vun': 'Kuwaiti', 'wae': 'Kuweit', 'wo': 'Kowet', 'xh': 'EKuwait', 'xnr': 'कुवैत', 'xog': 'Kuweti', 'yav': 'kowéet', 'yo': 'Kuweti', 'yrl': 'Kuwaiti', 'yue': '科威特', 'yue-Hans': '科威特', 'yue-Hant': '科威特', 'zgh': 'ⵍⴽⵡⵉⵜ', 'zh': '科威特', 'zh-Hans': '科威特', 'zh-Hant': '科威特', 'zu': 'i-Kuwait'},
    'KY': {'af': 'Kaaimanseilande', 'agq': 'Chwɨlà ŋ̀ Kaemàn', 'ak': 'Kemanfo Islands', 'am': 'ካይማን ደሴቶች', 'ar': 'جزر كايمان', 'as': 'কেইমেন দ্বীপপুঞ্জ', 'asa': 'Vithiwa vya Kayman', 'ast': 'Islles Caimán', 'az': 'Kayman adaları', 'az-Cyrl': 'Кајман адалары', 'az-Latn': 'Kayman adaları', 'bal-Latn': 'Kaymi Islánd', 'bas': 'Bìòn bi Kaymàn', 'be': 'Кайманавы астравы', 'be-tarask': 'Кайманавы астравы', 'bez': 'Ifisima fya Kayman', 'bg': 'Кайманови острови', 'blo': 'Kayimaan kaBʊtukǝltǝna', 'bm': 'Bama Gun', 'bn': 'কেম্যান দ্বীপপুঞ্জ', 'br': 'Inizi Cayman', 'brx': 'कैमेन द्वीपफोर', 'bs': 'Kajmanska ostrva', 'bs-Cyrl': 'Кајманска острва', 'bs-Latn': 'Kajmanska ostrva', 'ca': 'Illes Caiman', 'ca-ES-valencia': 'Illes Caiman', 'ccp': '𑄇𑄬𑄟𑄳𑄠𑄚𑄴 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Кайман гӀайренаш', 'ceb': 'Cayman Islands', 'cgg': 'Ebizinga bya Kayimani', 'chr': 'ᎨᎢᎹᏂ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی کایمان', 'cs': 'Kajmanské ostrovy', 'cv': 'Кайман утравӗсем', 'cy': 'Ynysoedd Cayman', 'da': 'Caymanøerne', 'dav': 'Visiwa vya Kayman', 'de': 'Kaimaninseln', 'dje': 'Kayman gungey', 'dsb': 'Kajmaniske kupy', 'dz': 'ཁེ་མེན་གླིང་ཚོམ', 'ebu': 'Visiwa vya Kayman', 'ee': 'Kayman ƒudomekpowo nutome', 'el': 'Νήσοι Κέιμαν', 'el-polyton': 'Νῆσοι Κέιμαν', 'en': 'Cayman Islands', 'en-Dsrt': '𐐗𐐩𐑋𐐲𐑌 𐐌𐑊𐐲𐑌𐐼𐑆', 'eo': 'Kejmanoj', 'es': 'Islas Caimán', 'et': 'Kaimanisaared', 'eu': 'Kaiman uharteak', 'ewo': 'Minlán Mí Kalimáŋ', 'fa': 'جزایر کِیمن', 'ff': 'Duuɗe Kaymaa', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤑𞤢𞤴𞤥𞤢𞥄𞤲', 'ff-Latn': 'Duuɗe Kaymaa', 'fi': 'Caymansaaret', 'fil': 'Cayman Islands', 'fo': 'Caymanoyggjar', 'fr': 'Îles Caïmans', 'fur': 'Isulis Cayman', 'fy': 'Caymaneilannen', 'ga': 'Oileáin Cayman', 'gd': 'Na h-Eileanan Caimean', 'gl': 'Illas Caimán', 'gsw': 'Käimän-Insle', 'gu': 'કેમેન આઇલેન્ડ્સ', 'guz': 'Visiwa vya Kayman', 'ha': 'Tsibiran Kaiman', 'he': 'איי קיימן', 'hi': 'कैमेन द्वीपसमूह', 'hi-Latn': 'Cayman Islands', 'hr': 'Kajmanski Otoci', 'hsb': 'Kajmanske kupy', 'ht': 'Îles Caïmans', 'hu': 'Kajmán-szigetek', 'hy': 'Կայմանյան կղզիներ', 'ia': 'Insulas de Caiman', 'id': 'Kepulauan Cayman', 'ig': 'Cayman Islands', 'is': 'Caymaneyjar', 'it': 'Isole Cayman', 'ja': 'ケイマン諸島', 'jmc': 'Visiwa vya Kayman', 'jv': 'Kapuloan Kéman', 'ka': 'კაიმანის კუნძულები', 'kaa': 'Кайман атаўлары', 'kaa-Cyrl': 'Кайман атаўлары', 'kab': 'Tigzirin n Kamyan', 'kam': 'Isiwa sya Kayman', 'kde': 'Chisiwa cha Kemen', 'kea': 'Ilhas Kaimon', 'kgp': 'Kajmỹm Goj-vẽso', 'khq': 'Kayman gungey', 'ki': 'Visiwa vya Kayman', 'kk': 'Кайман аралдары', 'kk-Cyrl': 'Кайман аралдары', 'kln': 'Ikwembeyotab Cayman', 'km': 'កោះ\u200bកៃម៉ង់', 'kn': 'ಕೇಮನ್ ದ್ವೀಪಗಳು', 'ko': '케이맨 제도', 'kok': 'कैमेन आइलैंड्स', 'kok-Deva': 'कैमेन आइलैंड्स', 'ks': 'کیمَن جٔزیٖرٕ', 'ks-Arab': 'کیمَن جٔزیٖرٕ', 'ksb': 'Visiwa vya Kayman', 'ksf': 'zǝ i gan', 'ksh': 'de Kaiman-Enselle', 'ku': 'Giravên Kaymanê', 'kxv': 'keimen dīp', 'kxv-Deva': 'केमेन द्वीप', 'kxv-Latn': 'keimen dīp', 'kxv-Orya': 'କାଇମାନ ଦିପ', 'kxv-Telu': 'కేమాన్ దీపులు', 'ky': 'Кайман аралдары', 'lag': 'Visíiwa vya Kayimani', 'lb': 'Kaimaninselen', 'lg': 'Bizinga ebya Kayimaani', 'lld': 'Isoles Cayman', 'ln': 'Bisanga bya Kayíma', 'lo': 'ໝູ່ເກາະ ເຄແມນ', 'lt': 'Kaimanų Salos', 'lu': 'Lutanda lua Kayima', 'luo': 'Cayman Islands', 'luy': 'Visiwa vya Kayman', 'lv': 'Kaimanu salas', 'mai': 'कैमेन द्वीपसमूह', 'mas': 'Visiwa vya Kayman', 'mer': 'Aĩrandi cia Kayman', 'mfe': 'Zil Kayman', 'mg': 'Nosy Kayman', 'mi': 'Ngā Motu Keimana', 'mk': 'Кајмански Острови', 'ml': 'കേയ്മാൻ ദ്വീപുകൾ', 'mn': 'Кайманы арлууд', 'mr': 'केमन बेटे', 'ms': 'Kepulauan Cayman', 'mt': 'il-Gżejjer Cayman', 'mua': 'kayman ma laŋne', 'my': 'ကေမန် ကျွန်းစု', 'mzn': 'کیمن جزیره\u200cئون', 'naq': 'Cayman Islands', 'nb': 'Caymanøyene', 'nd': 'Cayman Islands', 'ne': 'केयमान टापु', 'nl': 'Kaaimaneilanden', 'nmg': 'Maŋ́ má kumbi', 'nn': 'Caymanøyane', 'no': 'Caymanøyene', 'nqo': 'ߓߊ߲߬ߓߊ߬-ߕߌ߲', 'nus': 'Kaymɛn ɣa̱ylɛn', 'nyn': 'Ebizinga bya Kayimani', 'om': 'Odoloota Saaymaan', 'or': 'କେମ୍ୟାନ୍\u200c ଦ୍ୱୀପପୁଞ୍ଜ', 'pa': 'ਕੇਮੈਨ ਟਾਪੂ', 'pa-Guru': 'ਕੇਮੈਨ ਟਾਪੂ', 'pcm': 'Kéman Aílands', 'pl': 'Kajmany', 'ps': 'کیمان ټاپوګان', 'pt': 'Ilhas Cayman', 'qu': 'Islas Caimán', 'rm': 'Inslas Cayman', 'rn': 'Ibirwa bya Keyimani', 'ro': 'Insulele Cayman', 'rof': 'Visiwa vya Kaimai', 'ru': 'о-ва Кайман', 'rwk': 'Visiwa vya Kayman', 'saq': 'Visiwa vya Kayman', 'sbp': 'Figunguli ifya Kayimayi', 'sc': 'Ìsulas Cayman', 'sd': 'ڪي مين ٻيٽ', 'sd-Arab': 'ڪي مين ٻيٽ', 'se': 'Cayman-sullot', 'seh': 'Ilhas Caiman', 'ses': 'Kayman gungey', 'sg': 'Âzûâ Ngundë, Kaimäni', 'shi': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⴽⴰⵢⵎⴰⵏ', 'shi-Latn': 'tigzirin n kayman', 'shi-Tfng': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⴽⴰⵢⵎⴰⵏ', 'si': 'කේමන් දූපත්', 'sk': 'Kajmanie ostrovy', 'sl': 'Kajmanski otoki', 'smn': 'Caymansuolluuh', 'sn': 'Zvitsuwa zveCayman', 'so': 'Cayman Islands', 'sq': 'Ishujt Kajman', 'sr': 'Кајманска Острва', 'sr-Cyrl': 'Кајманска Острва', 'sr-Latn': 'Kajmanska Ostrva', 'sv': 'Caymanöarna', 'sw': 'Visiwa vya Cayman', 'syr': 'ܓܙܝܖ̈ܐ ܕܟܐܝܡܐܢ', 'ta': 'கெய்மென் தீவுகள்', 'te': 'కేమాన్ దీవులు', 'teo': 'Visiwa vya Kayman', 'tg': 'Ҷазираҳои Кайман', 'th': 'หมู่เกาะเคย์แมน', 'ti': 'ደሴታት ካይማን', 'tk': 'Kaýman adalary', 'to': 'ʻOtumotu Keimeni', 'tr': 'Cayman Adaları', 'tt': 'Кайман утраулары', 'twq': 'Kayman gungey', 'tzm': 'Tigzirin n Kayman', 'ug': 'كايمان ئاراللىرى', 'uk': 'Кайманові Острови', 'ur': 'کیمین آئلینڈز', 'uz': 'Kayman orollari', 'uz-Cyrl': 'Кайман ороллари', 'uz-Latn': 'Kayman orollari', 'vai': 'ꔞꔀꕮꘋ ꔳꘋꗣ', 'vai-Latn': 'Keemaŋ Tiŋŋɛ', 'vai-Vaii': 'ꔞꔀꕮꘋ ꔳꘋꗣ', 'vec': 'Ìzole Caiman', 'vi': 'Quần đảo Cayman', 'vun': 'Visiwa vya Kayman', 'wae': 'Kaimaninslä', 'wo': 'Duni Kaymaŋ', 'xh': 'ECayman Islands', 'xnr': 'कैमेन द्वीपसमूह', 'xog': 'Ebizinga bya Kayimaani', 'yav': 'Káyímanɛ', 'yi': 'קיימאַן אינזלען', 'yo': 'Etíokun Kámánì', 'yrl': 'Kapuãma-ita Kaimã', 'yue': '開曼群島', 'yue-Hans': '开曼群岛', 'yue-Hant': '開曼群島', 'zgh': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⴽⴰⵢⵎⴰⵏ', 'zh': '开曼群岛', 'zh-Hans': '开曼群岛', 'zh-Hant': '開曼群島', 'zu': 'i-Cayman Islands'},
    'KZ': {'af': 'Kazakstan', 'agq': 'Kàzasɨ̀tân', 'ak': 'Kazakstan', 'am': 'ካዛኪስታን', 'ar': 'كازاخستان', 'as': 'কাজাখাস্তান', 'asa': 'Kazakithtani', 'ast': 'Kazakstán', 'az': 'Qazaxıstan', 'az-Cyrl': 'Газахыстан', 'az-Latn': 'Qazaxıstan', 'bal-Latn': 'Kázakhestán', 'bas': 'Kàzàkstâŋ', 'be': 'Казахстан', 'be-tarask': 'Казахстан', 'bez': 'Hukazakistani', 'bg': 'Казахстан', 'bgn': 'قزاقستان', 'blo': 'Kasastan', 'bm': 'Kazakistaŋ', 'bn': 'কাজাখস্তান', 'br': 'Kazakstan', 'brx': 'काजाखस्तान', 'bs': 'Kazahstan', 'bs-Cyrl': 'Казахстан', 'bs-Latn': 'Kazahstan', 'ca': 'Kazakhstan', 'ca-ES-valencia': 'Kazakhstan', 'ccp': '𑄇𑄎𑄈𑄌𑄴𑄖𑄚𑄴', 'ce': 'Кхазакхстан', 'ceb': 'Kazakhstan', 'cgg': 'Kazakisitani', 'chr': 'ᎧᏎᎧᏍᏕᏂ', 'ckb': 'کازاخستان', 'cs': 'Kazachstán', 'cv': 'Казахстан', 'cy': 'Kazakhstan', 'da': 'Kasakhstan', 'dav': 'Kazakistani', 'de': 'Kasachstan', 'dje': 'Kaazakstan', 'dsb': 'Kazachstan', 'dz': 'ཀ་ཛགས་སཏཱན', 'ebu': 'Kazakistani', 'ee': 'Kazakstan nutome', 'el': 'Καζακστάν', 'el-polyton': 'Καζακστάν', 'en': 'Kazakhstan', 'en-Dsrt': '𐐗𐐲𐑆𐐪𐐿𐑅𐐻𐐪𐑌', 'eo': 'Kazaĥujo', 'es': 'Kazajistán', 'et': 'Kasahstan', 'eu': 'Kazakhstan', 'ewo': 'Kazakətáŋ', 'fa': 'قزاقستان', 'ff': 'Kasakstaan', 'ff-Adlm': '𞤑𞤢𞥁𞤢𞤧𞤼𞤢𞥄𞤲', 'ff-Latn': 'Kasakstaan', 'fi': 'Kazakstan', 'fil': 'Kazakhstan', 'fo': 'Kasakstan', 'fr': 'Kazakhstan', 'fur': 'Kazachistan', 'fy': 'Kazachstan', 'ga': 'an Chasacstáin', 'gd': 'Casachstàn', 'gl': 'Kazakistán', 'gsw': 'Kasachschtan', 'gu': 'કઝાકિસ્તાન', 'guz': 'Kazakistani', 'ha': 'Kazakistan', 'he': 'קזחסטן', 'hi': 'कज़ाखस्तान', 'hi-Latn': 'Kazakhstan', 'hr': 'Kazahstan', 'hsb': 'Kazachstan', 'ht': 'Kazakhstan', 'hu': 'Kazahsztán', 'hy': 'Ղազախստան', 'ia': 'Kazakhstan', 'id': 'Kazakhstan', 'ig': 'Kazakhstan', 'is': 'Kasakstan', 'it': 'Kazakistan', 'ja': 'カザフスタン', 'jmc': 'Kazakistani', 'jv': 'Kasakstan', 'ka': 'ყაზახეთი', 'kaa': 'Қазақстан', 'kaa-Cyrl': 'Қазақстан', 'kab': 'Kazaxistan', 'kam': 'Kazakistani', 'kde': 'Kazachistani', 'kea': 'Kazakistan', 'kgp': 'Kajakinhtỹv', 'khq': 'Kaazakstan', 'ki': 'Kazakistani', 'kk': 'Қазақстан', 'kk-Cyrl': 'Қазақстан', 'kln': 'Emetab Kazakhstan', 'km': 'កាហ្សាក់ស្ថាន', 'kn': 'ಕಝಾಕಿಸ್ಥಾನ್', 'ko': '카자흐스탄', 'kok': 'कझाकस्तान', 'kok-Deva': 'कझाकस्तान', 'ks': 'قازقستان', 'ks-Arab': 'قازقستان', 'ksb': 'Kazakistani', 'ksf': 'kazakstáŋ', 'ksh': 'Kassakstahn', 'ku': 'Qazaxistan', 'kxv': 'kajakstan', 'kxv-Deva': 'कज़ाकस्तान', 'kxv-Latn': 'kajakstan', 'kxv-Orya': 'କାଜାକସ୍ତାନ', 'kxv-Telu': 'కజకిస్తాన్', 'ky': 'Казакстан', 'lag': 'Kazakasitáani', 'lb': 'Kasachstan', 'lg': 'Kazakisitaani', 'lld': 'Kazakhstan', 'ln': 'Kazakisitá', 'lo': 'ຄາຊັກສະຖານ', 'lt': 'Kazachstanas', 'lu': 'Kazakusita', 'luo': 'Kazakhstan', 'luy': 'Kazakistani', 'lv': 'Kazahstāna', 'mai': 'कजाखस्तान', 'mas': 'Kazakistani', 'mer': 'Kazakistani', 'mfe': 'Kazakstan', 'mg': 'Kazakhstan', 'mgh': 'Ukazakistani', 'mi': 'Katatānga', 'mk': 'Казахстан', 'ml': 'കസാഖിസ്ഥാൻ', 'mn': 'Казахстан', 'mr': 'कझाकस्तान', 'ms': 'Kazakhstan', 'mt': 'il-Każakistan', 'mua': 'Kazakstaŋ', 'my': 'ကာဇက်စတန်', 'mzn': 'قزاقستون', 'naq': 'Kazakhstanni', 'nb': 'Kasakhstan', 'nd': 'Kazakhstan', 'ne': 'काजाकस्तान', 'nl': 'Kazachstan', 'nmg': 'Kazakstaŋ', 'nn': 'Kasakhstan', 'no': 'Kasakhstan', 'nqo': 'ߞߖ߭ߊߞߌߛߑߕߊ߲߫', 'nyn': 'Kazakisitani', 'om': 'Kazakistaan', 'or': 'କାଜାଖସ୍ତାନ୍\u200c', 'pa': 'ਕਜ਼ਾਖਸਤਾਨ', 'pa-Guru': 'ਕਜ਼ਾਖਸਤਾਨ', 'pcm': 'Kazakstan', 'pl': 'Kazachstan', 'ps': 'قزاقستان', 'pt': 'Cazaquistão', 'qu': 'Kazajistán', 'rm': 'Kasachstan', 'rn': 'Kazakisitani', 'ro': 'Kazahstan', 'rof': 'Kazakistani', 'ru': 'Казахстан', 'rwk': 'Kazakistani', 'saq': 'Kazakistani', 'sbp': 'Kasakisitani', 'sc': 'Kazàkistan', 'sd': 'قازقستان', 'sd-Arab': 'قازقستان', 'se': 'Kasakstan', 'seh': 'Casaquistão', 'ses': 'Kaazakstan', 'sg': 'Kazakisitäan', 'shi': 'ⴽⴰⵣⴰⵅⵙⵜⴰⵏ', 'shi-Latn': 'kazaxstan', 'shi-Tfng': 'ⴽⴰⵣⴰⵅⵙⵜⴰⵏ', 'si': 'කසකස්තානය', 'sk': 'Kazachstan', 'sl': 'Kazahstan', 'smn': 'Kazakstan', 'sn': 'Kazakhstan', 'so': 'Kasaakhistaan', 'sq': 'Kazakistan', 'sr': 'Казахстан', 'sr-Cyrl': 'Казахстан', 'sr-Latn': 'Kazahstan', 'sv': 'Kazakstan', 'sw': 'Kazakistani', 'syr': 'ܟܙܩܣܬܐܢ', 'ta': 'கஸகஸ்தான்', 'te': 'కజకిస్తాన్', 'teo': 'Kazakistani', 'tg': 'Қазоқистон', 'th': 'คาซัคสถาน', 'ti': 'ካዛኪስታን', 'tk': 'Gazagystan', 'to': 'Kasakitani', 'tr': 'Kazakistan', 'tt': 'Казахстан', 'twq': 'Kaazakstan', 'tzm': 'Kazaxistan', 'ug': 'قازاقىستان', 'uk': 'Казахстан', 'ur': 'قزاخستان', 'uz': 'Qozogʻiston', 'uz-Cyrl': 'Қозоғистон', 'uz-Latn': 'Qozogʻiston', 'vai': 'ꕪꕤꔻꕚꘋ', 'vai-Latn': 'Kazasitaŋ', 'vai-Vaii': 'ꕪꕤꔻꕚꘋ', 'vec': 'Kazakistan', 'vi': 'Kazakhstan', 'vun': 'Kazakistani', 'wae': 'Kasačstan', 'wo': 'Kasaxstaŋ', 'xh': 'EKazakhstan', 'xnr': 'कज़ाखस्तान', 'xog': 'Kazakisitaani', 'yav': 'kasaksitáŋ', 'yo': 'Kaṣaṣatani', 'yrl': 'Kasakiretãma', 'yue': '哈薩克', 'yue-Hans': '哈萨克', 'yue-Hant': '哈薩克', 'zgh': 'ⴽⴰⵣⴰⵅⵙⵜⴰⵏ', 'zh': '哈萨克斯坦', 'zh-Hans': '哈萨克斯坦', 'zh-Hant': '哈薩克', 'zu': 'i-Kazakhstan'},
    'Kali': {'ar': 'الكياه لى', 'ast': 'kayah li', 'az': 'kayax li', 'az-Latn': 'kayax li', 'bal-Latn': 'Káyah Li', 'bg': 'Кая Ли', 'bn': 'কায়াহ লি', 'brx': 'कायाह ली', 'bs': 'kajah li pismo', 'bs-Cyrl': 'кајах-ли писмо', 'bs-Latn': 'kajah li pismo', 'ca': 'kayah li', 'ca-ES-valencia': 'kayah li', 'ccp': '𑄇𑄠𑄦𑄧𑄣𑄨', 'cs': 'kayah li', 'da': 'kaya li', 'de': 'Kayah Li', 'el': 'Καγιάχ Λι', 'el-polyton': 'Καγιάχ Λι', 'en': 'Kayah Li', 'en-Dsrt': '𐐗𐐪𐐷𐐪 𐐢𐐨', 'es': 'kayah li', 'et': 'kaja-lii', 'eu': 'kayah li', 'fa': 'کایالی', 'ff-Adlm': '𞤑𞤢𞤴𞤢 𞤂𞤭', 'fi': 'kayah li', 'fo': 'kayah li', 'fr': 'kayah li', 'fy': 'Kayah Li', 'ga': 'Kayah Li', 'gd': 'Kayah Li', 'gsw': 'Kayah Li', 'gu': 'કાયાહ લી', 'he': 'קאי לי', 'hi': 'कायाह ली', 'hi-Latn': 'Kayah Li', 'hr': 'kayah li pismo', 'ht': 'kayah li', 'hu': 'Kajah li', 'id': 'Kayah Li', 'ig': 'Kayah Li', 'it': 'kayah li', 'ja': 'カヤー文字', 'ka': 'კაიაჰ-ლი', 'kgp': 'kaja-ri', 'kk': 'кая-ли жазуы', 'kk-Cyrl': 'кая-ли жазуы', 'km': 'កាយ៉ាលី', 'kn': 'ಕೆಯಾ ಲಿ', 'ko': '카야 리 문자', 'ks': 'کایا لی', 'ks-Arab': 'کایا لی', 'lb': 'Kayah Li', 'lo': 'ຄຍາ', 'lt': 'kajah li', 'mk': 'каја ли', 'ml': 'കയാ ലി', 'mn': 'Кали', 'mr': 'कायाह ली', 'ms': 'Kayah Li', 'my': 'ကယားလီ', 'nb': 'kayah li', 'ne': 'कायाहली', 'nl': 'Kayah Li', 'nn': 'kayah li', 'no': 'kayah li', 'nqo': 'ߞߦߊߟߌ߫', 'or': 'କାୟାହା ଲୀ', 'pl': 'kayah li', 'pt': 'kayah li', 'rm': 'kayah li', 'ro': 'kayah li', 'ru': 'кайа', 'sc': 'kayah li', 'sl': 'karenski', 'so': 'Kayah LI', 'sq': 'kajali', 'sr': 'кајах-ли писмо', 'sr-Cyrl': 'кајах-ли писмо', 'sr-Latn': 'kajah-li pismo', 'sv': 'kaya li', 'ta': 'கயாஹ் லீ', 'te': 'కాయాహ్ లి', 'tg': 'Кайя Ли', 'th': 'คยาห์', 'to': 'tohinima fakakaialī', 'tr': 'Kayah Li', 'ug': 'كاياھچە', 'uk': 'кая лі', 'vec': 'kayah li', 'vi': 'Chữ Kayah Li', 'xnr': 'कायाह ली', 'yo': 'Èdè Kaya Li', 'yrl': 'kaya ri', 'yue': '克耶李文', 'yue-Hans': '克耶李文', 'yue-Hant': '克耶李文', 'zh': '克耶李文字', 'zh-Hans': '克耶李文字', 'zh-Hant': '克耶李文', 'zu': 'isi-Kayah Li'},
    'Kana': {'af': 'Katakana', 'ak': 'Katakana kasa', 'am': 'ካታካና', 'ar': 'الكتكانا', 'as': 'কাটাকানা', 'ast': 'katakana', 'az': 'katakana', 'az-Arab': 'کاتاکانا', 'az-Latn': 'katakana', 'bal-Latn': 'Katakaná', 'be': 'катакана', 'be-tarask': 'катакана', 'bg': 'катакана', 'bn': 'কাটাকানা', 'br': 'katakana', 'brx': 'काताकाना', 'bs': 'pismo katakana', 'bs-Cyrl': 'катакана', 'bs-Latn': 'pismo katakana', 'ca': 'katakana', 'ca-ES-valencia': 'katakana', 'ccp': '𑄇𑄑𑄇𑄚', 'ce': 'катакана', 'chr': 'ᎧᏔᎧᎾ', 'ckb': 'کاتاکانا', 'cs': 'katakana', 'cy': 'Catacana', 'da': 'katakana', 'de': 'Katakana', 'dsb': 'katakana', 'dz': 'ཇ་པཱན་གྱི་ཀ་ཏ་ཀ་ན་ཡིག་གུ', 'ee': 'katakanagbeŋɔŋlɔ', 'el': 'Κατακάνα', 'el-polyton': 'Κατακάνα', 'en': 'Katakana', 'en-Dsrt': '𐐗𐐪𐐻𐐲𐐿𐐪𐑌𐐲', 'es': 'katakana', 'et': 'katakana', 'eu': 'katakana', 'fa': 'کاتاکانا', 'ff-Adlm': '𞤑𞤢𞤼𞤢𞤳𞤢𞤲𞤢', 'fi': 'katakana', 'fil': 'Katakana', 'fo': 'katakana', 'fr': 'katakana', 'fy': 'Katakana', 'ga': 'Catacánach', 'gd': 'Katakana', 'gl': 'katakana', 'gsw': 'Katakana', 'gu': 'કટાકાના', 'ha': 'Tsarin Rubutun Katakana', 'he': 'קטקאנה', 'hi': 'काताकाना', 'hi-Latn': 'Katakana', 'hr': 'katakana pismo', 'hsb': 'katakana', 'ht': 'katakana', 'hu': 'Katakana', 'hy': 'կատականա', 'ia': 'katakana', 'id': 'Katakana', 'ig': 'Katakana', 'is': 'katakana', 'it': 'katakana', 'ja': 'カタカナ', 'jv': 'Katakana', 'ka': 'კატაკანა', 'kea': 'katakana', 'kgp': 'katakỹnỹ', 'kk': 'катакана жазуы', 'kk-Cyrl': 'катакана жазуы', 'km': 'កាតាកាណា', 'kn': 'ಕಟಕಾನಾ', 'ko': '가타카나', 'kok': 'कटाकाना', 'kok-Deva': 'कटाकाना', 'ks': 'کَتاکانا', 'ks-Arab': 'کَتاکانا', 'ksh': 'japaanesche Kattakaana-Schreff', 'ku': 'katakanayî', 'ky': 'Катакана', 'lb': 'Katakana', 'lo': 'ຄະຕະກະນະ', 'lrc': 'کاتانگا', 'lt': 'katakana', 'lv': 'katakana', 'mai': 'कटाकाना', 'mi': 'Katakana', 'mk': 'катакана', 'ml': 'കറ്റക്കാന', 'mn': 'катакана', 'mr': 'कॅटाकाना', 'ms': 'Katakana', 'my': 'ခတခန', 'mzn': 'کاتاکانا', 'nb': 'katakana', 'ne': 'काताकाना', 'nl': 'Katakana', 'nn': 'katakana', 'no': 'katakana', 'nqo': 'ߞߕߊߞߣߊ', 'or': 'କାତାକାନା', 'pa': 'ਕਾਟਾਕਾਨਾ', 'pa-Guru': 'ਕਾਟਾਕਾਨਾ', 'pcm': 'Katákána', 'pl': 'katakana', 'ps': 'کاتاکانا', 'pt': 'katakana', 'qu': 'Katakana', 'rm': 'katakana', 'ro': 'katakana', 'ru': 'катакана', 'sc': 'katakana', 'sd': 'ڪٽاڪانا', 'sd-Arab': 'ڪٽاڪانا', 'se': 'katakana', 'si': 'කතකනා', 'sk': 'katakana', 'sl': 'katakana', 'so': 'Katakaana', 'sq': 'katakan', 'sr': 'катакана', 'sr-Cyrl': 'катакана', 'sr-Latn': 'katakana', 'sv': 'katakana', 'sw': 'Kikatakana', 'ta': 'கதகானா', 'te': 'కాటాకాన', 'tg': 'Катакана', 'th': 'คะตะกะนะ', 'ti': 'ካታካና', 'tk': 'Katakana elipbiýi', 'to': 'tohinima fakasiapani-katakana', 'tr': 'Katakana', 'ug': 'كاتاكانا', 'uk': 'катакана', 'ur': 'کٹاکانا', 'uz': 'katakana', 'uz-Cyrl': 'Катакана', 'uz-Latn': 'katakana', 'vec': 'katanaka', 'vi': 'Chữ Katakana', 'xnr': 'काताकाना', 'yo': 'Katakánà', 'yrl': 'katakaná', 'yue': '片假名', 'yue-Hans': '片假名', 'yue-Hant': '片假名', 'zh': '片假名', 'zh-Hans': '片假名', 'zh-Hant': '片假名', 'zu': 'isi-Katakana'},
    'Kawi': {'bal-Latn': 'Káwi', 'bs': 'kavi pismo', 'bs-Cyrl': 'кави писмо', 'bs-Latn': 'kavi pismo', 'et': 'kaavi', 'ff-Adlm': '𞤑𞤢𞤱𞤭', 'kk': 'кави жазуы', 'kk-Cyrl': 'кави жазуы', 'mn': 'Кави', 'nl': 'Kawi-taal', 'sq': 'kavi', 'tg': 'Кавӣ', 'zh': '卡维文', 'zh-Hans': '卡维文'},
    'Khar': {'ar': 'الخاروشتى', 'ast': 'kharoshthi', 'az': 'xaroşti', 'az-Latn': 'xaroşti', 'bal-Latn': 'Karóshti', 'bg': 'Кхароштхи', 'bn': 'খরোষ্ঠী', 'brx': 'खरोष्टी', 'bs': 'karošti pismo', 'bs-Cyrl': 'карошти писмо', 'bs-Latn': 'karošti pismo', 'ca': 'kharosthi', 'ca-ES-valencia': 'kharosthi', 'ccp': '𑄈𑄢𑄮𑄌𑄴𑄒𑄩', 'cs': 'kháróšthí', 'da': 'kharoshti', 'de': 'Kharoshthi', 'el': 'Καρόσθι', 'el-polyton': 'Καρόσθι', 'en': 'Kharoshthi', 'en-Dsrt': '𐐗𐐲𐑉𐐬𐑇𐑃𐐨', 'es': 'kharosthi', 'et': 'kharoshthi', 'eu': 'kharoshthi', 'ff-Adlm': '𞤑𞤢𞤪𞤮𞥃𞤢𞤼𞤭', 'fi': 'kharosthi', 'fo': 'kharoshthi', 'fr': 'kharochthî', 'fy': 'Kharoshthi', 'ga': 'Kharoshthi', 'gd': 'Kharoshthi', 'gsw': 'Kharoshthi', 'gu': 'ખારોશ્થી', 'he': 'חרושתי', 'hi': 'खारोशथी', 'hi-Latn': 'Kharoshthi', 'hr': 'kharoshthi pismo', 'ht': 'kharochthî', 'hu': 'Kharoshthi', 'id': 'Kharoshthi', 'ig': 'Kharoshthi', 'it': 'kharoshthi', 'ja': 'カローシュティー文字', 'ka': 'ქაროშთი', 'kgp': 'karositi', 'kk': 'кхароштхи жазуы', 'kk-Cyrl': 'кхароштхи жазуы', 'kn': 'ಖರೋಶ್ತಿ', 'ko': '카로슈티 문자', 'ks': 'خَروشتھی', 'ks-Arab': 'خَروشتھی', 'lb': 'Kharoshthi', 'lo': 'ຂໍໂຮກສີ', 'lt': 'karošti', 'mk': 'карошти', 'ml': 'ഖരോഷ്ടി', 'mn': 'Карошти', 'mr': 'खारोश्थी', 'ms': 'Kharoshthi', 'nb': 'kharoshthi', 'ne': 'खारोस्थिति', 'nl': 'Kharoshthi', 'nn': 'kharoshthi', 'no': 'kharoshthi', 'nqo': 'ߞߊߙߏߗߌ߮', 'or': 'ଖାରୋସ୍ଥି', 'pl': 'charosti', 'pt': 'kharoshthi', 'rm': 'kharoshthi', 'ro': 'kharosthi', 'ru': 'кхароштхи', 'sc': 'kharoshthi', 'sl': 'gandarski', 'so': 'Koraalka kharooshi', 'sq': 'karoshtisht', 'sr': 'карошти писмо', 'sr-Cyrl': 'карошти писмо', 'sr-Latn': 'karošti pismo', 'sv': 'kharoshti', 'ta': 'கரோஷ்டி', 'te': 'ఖరోషథి', 'tg': 'Хароштӣ', 'th': 'ขโรษฐี', 'to': 'tohinima fakakalositī', 'tr': 'Kharoshthi', 'ug': 'كاروشتىچە', 'uk': 'кхароштхі', 'vec': 'karosti', 'vi': 'Chữ Kharoshthi', 'xnr': 'खारोशथी', 'yrl': 'karuxiti', 'yue': '卡羅須提文', 'yue-Hans': '卡罗须提文', 'yue-Hant': '卡羅須提文', 'zh': '卡罗须提文', 'zh-Hans': '卡罗须提文', 'zh-Hant': '卡羅須提文', 'zu': 'isi-Kharoshthi'},
    'Khmr': {'af': 'Khmer', 'ak': 'Kɛma kasa', 'am': 'ክህመር', 'ar': 'الخميرية', 'as': 'খমেৰ', 'ast': 'ḥemer', 'az': 'kxmer', 'az-Arab': 'خمئرجه', 'az-Latn': 'kxmer', 'bal-Latn': 'Khmér', 'be': 'кхмерскае', 'be-tarask': 'кхмерскае', 'bg': 'кхмерска', 'bn': 'খমের', 'br': 'khmer', 'brx': 'ख्मेर', 'bs': 'kmersko pismo', 'bs-Cyrl': 'кмерско писмо', 'bs-Latn': 'kmersko pismo', 'ca': 'khmer', 'ca-ES-valencia': 'khmer', 'ccp': '𑄈𑄬𑄟𑄬𑄢𑄴', 'ce': 'кхмерийн', 'chr': 'ᎩᎻᎷ', 'ckb': 'خمێر', 'cs': 'khmerské', 'cy': 'Chmeraidd', 'da': 'khmer', 'de': 'Khmer', 'dsb': 'khmer', 'dz': 'ཁེ་མེར་ཡིག་གུ', 'ee': 'khmergbeŋɔŋlɔ', 'el': 'Χμερ', 'el-polyton': 'Χμερ', 'en': 'Khmer', 'en-Dsrt': '𐐗𐐲𐑋𐐯𐑉', 'es': 'jemer', 'et': 'khmeeri', 'eu': 'khmertarra', 'fa': 'خمری', 'ff-Adlm': '𞤑𞤵𞤥𞤫𞥅𞤪', 'fi': 'khmeriläinen', 'fil': 'Khmer', 'fo': 'khmer', 'fr': 'khmer', 'fur': 'khmer', 'fy': 'Khmer', 'ga': 'Ciméarach', 'gd': 'Cmèar', 'gl': 'khmer', 'gsw': 'Khmer', 'gu': 'ખ્મેર', 'ha': 'Yaren Khmer', 'he': 'חמרי', 'hi': 'खमेर', 'hi-Latn': 'Khmer', 'hr': 'kmersko pismo', 'hsb': 'khmersce', 'ht': 'khmer', 'hu': 'Khmer', 'hy': 'քմերական', 'ia': 'khmer', 'id': 'Khmer', 'ig': 'Khmer', 'is': 'kmer', 'it': 'khmer', 'ja': 'クメール文字', 'jv': 'Khmer', 'ka': 'ქმერული', 'kea': 'kmer', 'kgp': 'kymẽr', 'kk': 'кхмер жазуы', 'kk-Cyrl': 'кхмер жазуы', 'km': 'ខ្មែរ', 'kn': 'ಖಮೇರ್', 'ko': '크메르 문자', 'kok': 'कंबोडियन', 'kok-Deva': 'कंबोडियन', 'ks': 'کھٕمیر', 'ks-Arab': 'کھٕمیر', 'ksh': 'Khmer-Schreff', 'ku': 'ximêrî', 'ky': 'Кмер', 'lb': 'Khmer', 'lo': 'ຂະແມ', 'lrc': 'خئمئر', 'lt': 'khmerų', 'lv': 'khmeru', 'mai': 'कुमेर', 'mi': 'Kimēra', 'mk': 'кмерско писмо', 'ml': 'ഖമെർ', 'mn': 'кхмер', 'mr': 'ख्मेर', 'ms': 'Khmer', 'my': 'ခမာ', 'nb': 'khmer', 'ne': 'खमेर', 'nl': 'Khmer', 'nn': 'khmer', 'no': 'khmer', 'nqo': 'ߞߑߡߍߙ', 'or': 'ଖମେର୍', 'pa': 'ਖਮੇਰ', 'pa-Guru': 'ਖਮੇਰ', 'pcm': 'Kemẹẹ', 'pl': 'khmerskie', 'ps': 'خمر', 'pt': 'khmer', 'qu': 'Khmer', 'rm': 'khmer/cambodschan', 'ro': 'khmeră', 'ru': 'кхмерская', 'sc': 'khmer', 'sd': 'خمر', 'sd-Arab': 'خمر', 'si': 'කමර්', 'sk': 'khmérske', 'sl': 'kmerski', 'so': 'Khamer', 'sq': 'kmer', 'sr': 'кмерско писмо', 'sr-Cyrl': 'кмерско писмо', 'sr-Latn': 'kmersko pismo', 'sv': 'khmeriska', 'sw': 'Kikambodia', 'ta': 'கமெர்', 'te': 'ఖ్మేర్', 'tg': 'Хмерӣ', 'th': 'เขมร', 'ti': 'ክመር', 'tk': 'Khmer elipbiýi', 'to': 'tohinima fakakamipōtia', 'tr': 'Kmer', 'ug': 'كېخمېرچە', 'uk': 'кхмерська', 'ur': 'خمیر', 'uz': 'kxmer', 'uz-Cyrl': 'Хмер', 'uz-Latn': 'kxmer', 'vec': 'kmer', 'vi': 'Chữ Khơ-me', 'wae': 'Khmer', 'xnr': 'खमेर', 'yo': 'Kẹmẹ̀', 'yrl': 'kimé', 'yue': '高棉文', 'yue-Hans': '高棉文', 'yue-Hant': '高棉文', 'zh': '高棉文', 'zh-Hans': '高棉文', 'zh-Hant': '高棉文', 'zu': 'isi-Khmer'},
    'Khoj': {'ast': 'khojki', 'bal-Latn': 'Kójki', 'bs': 'khojki pismo', 'bs-Cyrl': 'којки писмо', 'bs-Latn': 'khojki pismo', 'ca': 'khoja', 'ca-ES-valencia': 'khoja', 'cs': 'chodžiki', 'da': 'khojki', 'de': 'Khojki', 'en': 'Khojki', 'et': 'hodžki', 'eu': 'khojkiera', 'fa': 'خواجکی', 'ff-Adlm': '𞤑𞤮𞤶𞤳𞤭', 'fi': 'khojki', 'fo': 'khojki', 'fy': 'Khojki', 'ga': 'Khojki', 'gd': 'Khojki', 'he': 'חוג׳קי', 'hi-Latn': 'Khojki', 'hr': 'khojki pismo', 'id': 'Khojki', 'ig': 'Khojki', 'it': 'khojki', 'ja': 'ホジャ文字', 'ka': 'ქოჯკი', 'kk': 'ходжики жазуы', 'kk-Cyrl': 'ходжики жазуы', 'ko': '코즈키 문자', 'lo': 'ຄໍຈຄີ', 'lt': 'Khojki', 'mk': 'хоџки', 'mn': 'Кожки', 'ms': 'Khojki', 'nb': 'khojki', 'nl': 'Khojki', 'nn': 'khojki', 'no': 'khojki', 'nqo': 'ߞߏߖߞߌ', 'ro': 'khojki', 'ru': 'ходжики', 'sc': 'khojki', 'so': 'Qoraalka Khojki', 'sq': 'koxhkisht', 'sv': 'khojkiska', 'tg': 'Хочки', 'th': 'คอจคี', 'to': 'tohinima fakakosikī', 'tr': 'Khojki', 'ug': 'خوجكىچە', 'vec': 'kojiki', 'vi': 'Chữ Khojki', 'yue': '克吉奇文字', 'yue-Hans': '克吉奇文字', 'yue-Hant': '克吉奇文字', 'zh': '克吉奇文字', 'zh-Hans': '克吉奇文字', 'zh-Hant': '克吉奇文字', 'zu': 'isi-Khojki'},
    'Kitl': {'en': 'Khitan large script'},
    'Kits': {'bal-Latn': 'Kitan gwandhén syáhag', 'bs': 'kitansko pismo malim slovima', 'bs-Cyrl': 'китан мала слова', 'bs-Latn': 'kitansko pismo malim slovima', 'en': 'Khitan small script', 'et': 'kitani väike kiri', 'eu': 'khitanerako script txikiak', 'ff-Adlm': '𞤄𞤭𞤲𞤳𞤮𞤴 𞤑𞤭𞤼𞤢𞤲', 'fi': 'kitaanin pieni merkistö', 'gd': 'Litrichean beaga na Khitan', 'he': 'כתב קטן של חיטן', 'hi-Latn': 'Khitan small script', 'id': 'Skrip kecil Khitan', 'ig': 'Khitan small script', 'kk': 'шағын кидан жазуы', 'kk-Cyrl': 'шағын кидан жазуы', 'mn': 'Китан бага бичвэр', 'nl': 'Kitaans kleinschrift', 'nqo': 'ߞߌߕߊ߲߫ ߛߓߍߟߌߣߍ߲', 'ro': 'litere mici khitane', 'sc': 'iscritura khitan minore', 'so': 'Qoraalka yar ee Khitan', 'sq': 'shkrim i vogël kitan', 'sv': 'khitanska', 'tg': 'Хатти хурди Китонӣ', 'vec': 'kitan celo', 'zh': '契丹小字', 'zh-Hans': '契丹小字', 'zu': 'i-Khitan small script'},
    'Knda': {'af': 'Kannada', 'ak': 'Kanada kasa', 'am': 'ካናዳ', 'ar': 'الكانادا', 'as': 'কানাড়া', 'ast': 'canarés', 'az': 'kannada', 'az-Latn': 'kannada', 'bal-Latn': 'Kannadá', 'be': 'канада', 'be-tarask': 'канада', 'bg': 'каннада', 'bn': 'কানাড়া', 'br': 'kannada', 'brx': 'कण्णडा', 'bs': 'pismo kanada', 'bs-Cyrl': 'канада писмо', 'bs-Latn': 'pismo kanada', 'ca': 'kannada', 'ca-ES-valencia': 'kannada', 'ccp': '𑄇𑄚𑄢', 'ce': 'каннада', 'chr': 'ᎧᎾᏓ', 'ckb': 'کاننادا', 'cs': 'kannadské', 'cy': 'Canaraidd', 'da': 'kannada', 'de': 'Kannada', 'dsb': 'kannada', 'dz': 'ཀ་ན་ཌ་ཡིག་གུ', 'ee': 'kannadagbeŋɔŋlɔ', 'el': 'Κανάντα', 'el-polyton': 'Κανάντα', 'en': 'Kannada', 'en-Dsrt': '𐐗𐐪𐑌𐐲𐐼𐐲', 'es': 'canarés', 'et': 'kannada', 'eu': 'kanadarra', 'fa': 'کانارا', 'ff-Adlm': '𞤑𞤢𞤲𞥆𞤢𞤣𞤢', 'fi': 'kannadalainen', 'fil': 'Kannada', 'fo': 'kannada', 'fr': 'kannara', 'fur': 'kannada', 'fy': 'Kannada', 'ga': 'Cannadach', 'gd': 'Kannada', 'gl': 'kannará', 'gsw': 'Kannada', 'gu': 'કન્નડા', 'ha': 'Yaren Kannada', 'he': 'קאנאדה', 'hi': 'कन्नड़', 'hi-Latn': 'Kannada', 'hr': 'kannada pismo', 'hsb': 'kannadsce', 'ht': 'kannara', 'hu': 'Kannada', 'hy': 'կաննադա', 'ia': 'kannada', 'id': 'Kannada', 'ig': 'Kannada', 'is': 'kannada', 'it': 'kannada', 'ja': 'カンナダ文字', 'jv': 'Kannada', 'ka': 'კანადა', 'kea': 'kanares', 'kgp': 'kan-nỹna', 'kk': 'каннада жазуы', 'kk-Cyrl': 'каннада жазуы', 'km': 'ខាណាដា', 'kn': 'ಕನ್ನಡ', 'ko': '칸나다 문자', 'kok': 'कन्नड', 'kok-Deva': 'कन्नड', 'ks': 'کَنَڑا', 'ks-Arab': 'کَنَڑا', 'ksh': 'indesche Kannada-Schreff', 'ku': 'kannadayî', 'kxv': 'knnaḍa', 'kxv-Deva': 'कन्नड़', 'kxv-Latn': 'knnaḍa', 'kxv-Orya': 'କନ୍ନଡ଼', 'kxv-Telu': 'కన్నడ', 'ky': 'Каннада', 'lb': 'Kannada', 'lo': 'ຄັນນາດາ', 'lrc': 'کانادا', 'lt': 'kanadų', 'lv': 'kannadu', 'mai': 'कन्नड़', 'mi': 'Kanāra', 'mk': 'каннада', 'ml': 'കന്നഡ', 'mn': 'каннада', 'mr': 'कन्नड', 'ms': 'Kannada', 'my': 'ခန်နာဒါ', 'nb': 'kannada', 'ne': 'कान्नाडा', 'nl': 'Kannada', 'nn': 'kannada', 'no': 'kannada', 'nqo': 'ߞߌߡߍߙ', 'or': 'କନ୍ନଡ଼', 'pa': 'ਕੰਨੜ', 'pa-Guru': 'ਕੰਨੜ', 'pcm': 'Kánnad Raítín Sístẹm', 'pl': 'kannada', 'ps': 'کناډا', 'pt': 'kannada', 'qu': 'Kannada Simi', 'rm': 'kannada', 'ro': 'kannada', 'ru': 'каннада', 'sc': 'kannada', 'sd': 'ڪناڊا', 'sd-Arab': 'ڪناڊا', 'si': 'කණ්ණඩ', 'sk': 'kannadské', 'sl': 'kanadski', 'so': 'Kanada', 'sq': 'kanad', 'sr': 'канада писмо', 'sr-Cyrl': 'канада писмо', 'sr-Latn': 'kanada pismo', 'sv': 'kanaresiska', 'sw': 'Kikannada', 'ta': 'கன்னடம்', 'te': 'కన్నడ', 'tg': 'Каннада', 'th': 'กันนาดา', 'ti': 'ካናዳ', 'tk': 'Kannada elipbiýi', 'to': 'tohinima fakaʻinitia-kanata', 'tr': 'Kannada', 'ug': 'كانناداچە', 'uk': 'каннада', 'ur': 'کنڑ', 'uz': 'kannada', 'uz-Cyrl': 'Каннада', 'uz-Latn': 'kannada', 'vec': 'kanada', 'vi': 'Chữ Kannada', 'wae': 'Kannada', 'xnr': 'कन्नड़', 'yo': 'Kanada', 'yrl': 'kãnará', 'yue': '坎那達文', 'yue-Hans': '坎那达文', 'yue-Hant': '坎那達文', 'zh': '卡纳达文', 'zh-Hans': '卡纳达文', 'zh-Hant': '坎那達文', 'zu': 'isi-Kannada'},
    'Kore': {'af': 'Koreaans', 'ak': 'Korea kasa', 'am': 'ኮሪያኛ', 'ar': 'الكورية', 'as': 'কোৰিয়ান', 'ast': 'coreanu', 'az': 'koreya', 'az-Arab': 'کوْره\u200cجه', 'az-Latn': 'koreya', 'bal-Latn': 'Kóriái', 'be': 'карэйскае', 'be-tarask': 'карэйскае', 'bg': 'корейска', 'bgn': 'کوریایی', 'blo': 'Koree kʊja', 'bn': 'কোরিয়ান', 'br': 'koreanek', 'brx': 'क’रीयान', 'bs': 'korejsko pismo', 'bs-Cyrl': 'корејско писмо', 'bs-Latn': 'korejsko pismo', 'ca': 'coreà', 'ca-ES-valencia': 'coreà', 'ccp': '𑄇𑄮𑄢𑄨𑄠𑄚𑄴', 'ce': 'корейн', 'ceb': 'Korean', 'chr': 'ᎪᎵᎠᏂ', 'ckb': 'کۆری', 'cs': 'korejské', 'cv': 'корей', 'cy': 'Coreaidd', 'da': 'koreansk', 'de': 'Koreanisch', 'dsb': 'korejski', 'dz': 'ཀོ་རི་ཡཱན་ཡིག་གུ', 'ee': 'Koreagbeŋɔŋlɔ', 'el': 'Κορεατικό', 'el-polyton': 'Κορεατικό', 'en': 'Korean', 'en-Dsrt': '𐐗𐐬𐑉𐐨𐐲𐑌', 'es': 'coreano', 'et': 'korea segakiri', 'eu': 'korearra', 'fa': 'کره\u200cای', 'ff-Adlm': '𞤑𞤮𞥅𞤪𞤫𞤴𞤢𞤲', 'fi': 'korealainen', 'fil': 'Korean', 'fo': 'koreanskt', 'fr': 'coréen', 'fur': 'corean', 'fy': 'Koreaansk', 'ga': 'Cóiréach', 'gd': 'Coirèanais', 'gl': 'coreano', 'gsw': 'Koreanisch', 'gu': 'કોરિયન', 'ha': 'Rubutun Koriya', 'he': 'קוריאני', 'hi': 'कोरियाई', 'hi-Latn': 'Korean', 'hr': 'korejsko pismo', 'hsb': 'korejsce', 'ht': 'coréen', 'hu': 'Koreai', 'hy': 'կորեական', 'ia': 'corean', 'id': 'Korea', 'ie': 'korean', 'ig': 'Korean', 'is': 'kóreskt', 'it': 'coreano', 'ja': '韓国語の文字', 'jv': 'Korea', 'ka': 'კორეული', 'kaa': 'корейс жазыўы', 'kaa-Cyrl': 'корейс жазыўы', 'kea': 'korianu', 'kgp': 'korejỹnũ', 'kk': 'корей жазуы', 'kk-Cyrl': 'корей жазуы', 'km': 'កូរ៉េ', 'kn': 'ಕೊರಿಯನ್', 'ko': '한국 문자', 'kok': 'कोरियन', 'kok-Deva': 'कोरियन', 'ks': 'کوریَن', 'ks-Arab': 'کوریَن', 'ksh': 'korrejaanesche Schreff udder en schineesesche Schreff', 'ku': 'koreyî', 'ky': 'Корей', 'lb': 'Koreanesch', 'lld': 'corean', 'lo': 'ເກົາຫຼີ', 'lrc': 'کورئ یی', 'lt': 'korėjiečių', 'lv': 'korejiešu', 'mai': 'कोरियाई', 'mi': 'Kōreana', 'mk': 'корејско писмо', 'ml': 'കൊറിയൻ', 'mn': 'солонгос', 'mni': 'কোরিয়ন', 'mni-Beng': 'কোরিয়ন', 'mr': 'कोरियन', 'ms': 'Korea', 'my': 'ကိုရီးယား', 'nb': 'koreansk', 'ne': 'कोरियन', 'nl': 'Koreaans', 'nn': 'koreansk', 'no': 'koreansk', 'nqo': 'ߞߏߙߋ߫', 'om': 'Afaan Kooriyaa', 'or': 'କୋରିଆନ୍', 'pa': 'ਕੋਰੀਆਈ', 'pa-Guru': 'ਕੋਰੀਆਈ', 'pcm': 'Koria', 'pl': 'koreańskie', 'ps': 'کوریایی', 'pt': 'coreano', 'qu': 'Coreano Simi', 'rm': 'corean', 'ro': 'coreeană', 'ru': 'корейская', 'sah': 'Кэриэйдии', 'sc': 'coreanu', 'sd': 'ڪوريائي', 'sd-Arab': 'ڪوريائي', 'si': 'කොරියානු', 'sk': 'kórejské', 'sl': 'korejski', 'so': 'Kuuriyaan', 'sq': 'korean', 'sr': 'корејско писмо', 'sr-Cyrl': 'корејско писмо', 'sr-Latn': 'korejsko pismo', 'sv': 'koreanska', 'sw': 'Kikorea', 'ta': 'கொரியன்', 'te': 'కొరియన్', 'tg': 'Кореягӣ', 'th': 'เกาหลี', 'ti': 'ኮርያዊ', 'tk': 'Koreý elipbiýi', 'to': 'tohinima fakakōlea', 'tr': 'Korece', 'tt': 'корея', 'ug': 'كورېيەچە', 'uk': 'корейська', 'ur': 'کوریائی', 'uz': 'koreys', 'uz-Cyrl': 'Корейс', 'uz-Latn': 'koreys', 'vec': 'corean', 'vi': 'Chữ Hàn Quốc', 'wae': 'Korianiš', 'wo': 'hangug-ui', 'xh': 'IsiKorean', 'xnr': 'कोरियाई', 'yo': 'Kóríà', 'yrl': 'kureanu', 'yue': '韓文', 'yue-Hans': '韩文', 'yue-Hant': '韓文', 'zh': '韩文', 'zh-Hans': '韩文', 'zh-Hant': '韓文', 'zu': 'isi-Korean'},
    'Kpel': {'ast': 'kpelle', 'bs': 'kpelle pismo', 'bs-Latn': 'kpelle pismo', 'ca': 'kpelle', 'ca-ES-valencia': 'kpelle', 'cs': 'kpelle', 'da': 'kpelle', 'de': 'Kpelle', 'en': 'Kpelle', 'et': 'kpelle', 'fi': 'kpelle', 'fy': 'Kpelle', 'gd': 'Kpelle', 'hi-Latn': 'Kpelle', 'hr': 'kpelle pismo', 'id': 'Kpelle', 'it': 'Kpelle', 'ja': 'クペレ文字', 'ka': 'კპელე', 'ko': '크펠레 문자', 'lo': 'ເປລເລ', 'lt': 'Kpelų', 'mk': 'кпелско', 'nb': 'kpelle', 'nl': 'Kpelle', 'nn': 'kpelle', 'no': 'kpelle', 'ru': 'кпелле', 'sv': 'kpellé', 'th': 'เปลเล', 'to': 'tohinima fakakepele', 'tr': 'Kpelle', 'ug': 'كپېللېچە', 'vi': 'Chữ Kpelle', 'yue': '克培列文', 'yue-Hans': '克培列文', 'yue-Hant': '克培列文', 'zh': '克佩列文', 'zh-Hans': '克佩列文', 'zh-Hant': '克培列文'},
    'Krai': {'bal-Latn': 'Kirat Rai', 'en': 'Kirat Rai', 'gd': 'Kirat Rai', 'hi-Latn': 'Kirat Rai', 'ig': 'Kirat Rai', 'vec': 'kirat rai'},
    'Kthi': {'ast': 'kaithi', 'az': 'kti', 'az-Latn': 'kti', 'bal-Latn': 'Kaiti', 'bg': 'Кайтхи', 'bn': 'কাইথি', 'bs': 'kaićansko pismo', 'bs-Cyrl': 'каити', 'bs-Latn': 'kaićansko pismo', 'ca': 'kaithi', 'ca-ES-valencia': 'kaithi', 'ccp': '𑄇𑄭𑄗𑄨', 'cs': 'kaithi', 'de': 'Kaithi', 'el': 'Καϊθί', 'el-polyton': 'Καϊθί', 'en': 'Kaithi', 'en-Dsrt': '𐐗𐐴𐐮𐐻𐐨', 'et': 'kaithi', 'eu': 'kaithiera', 'fa': 'کثی', 'ff-Adlm': '𞤑𞤢𞤭𞤼𞤭', 'fi': 'kaithi', 'fr': 'kaithî', 'fy': 'Kaithi', 'ga': 'Kaithi', 'gd': 'Kaithi', 'gu': 'કૈથી', 'he': 'קאיתי', 'hi': 'कैथी', 'hi-Latn': 'Kaithi', 'hr': 'kaithi pismo', 'ht': 'kaithî', 'hu': 'Kaithi', 'id': 'Kaithi', 'ig': 'Kaithi', 'it': 'kaithi', 'ja': 'カイティ文字', 'ka': 'კაითი', 'kgp': 'kanhi', 'kk': 'кайтхи жазуы', 'kk-Cyrl': 'кайтхи жазуы', 'kn': 'ಕೈಥಿ', 'ko': '카이시 문자', 'lo': 'ກາຍຕິ', 'lt': 'kaithi', 'mk': 'кајти', 'ml': 'ക്തി', 'mn': 'Кайти', 'mr': 'काइथी', 'ms': 'Kaithi', 'nb': 'kaithisk', 'ne': 'क्थी', 'nl': 'Kaithi', 'nn': 'kaithisk', 'no': 'kaithisk', 'nqo': 'ߞߊߦߌߗߌ߮', 'or': 'କୈଥି', 'pl': 'kaithi', 'rm': 'kaithi', 'ro': 'kaithi', 'ru': 'кайтхи', 'sc': 'kaithi', 'sl': 'kajatski', 'so': 'kaithi', 'sq': 'kaitisht', 'sr': 'каити', 'sr-Cyrl': 'каити', 'sr-Latn': 'kaiti', 'sv': 'kaithiska', 'ta': 'காய்தி', 'te': 'కైథి', 'tg': 'Кайтӣ', 'th': 'กายติ', 'to': 'tohinima fakakaiatī', 'tr': 'Kaithi', 'ug': 'كاياتىچە', 'uk': 'каїті', 'vec': 'kaiti', 'vi': 'Chữ Kaithi', 'xnr': 'कैथी', 'yrl': 'kiti', 'yue': '凱提文', 'yue-Hans': '凯提文', 'yue-Hant': '凱提文', 'zh': '凯提文', 'zh-Hans': '凯提文', 'zh-Hant': '凱提文', 'zu': 'isi-Kaithi'},
    'LA': {'af': 'Laos', 'agq': 'Làwos', 'ak': 'Laos', 'am': 'ላኦስ', 'ar': 'لاوس', 'as': 'লাওচ', 'asa': 'Laothi', 'ast': 'Laos', 'az': 'Laos', 'az-Cyrl': 'Лаос', 'az-Latn': 'Laos', 'bal-Latn': 'Láus', 'bas': 'Làôs', 'be': 'Лаос', 'be-tarask': 'Лаос', 'bez': 'Hulaosi', 'bg': 'Лаос', 'bgn': 'لاوس', 'blo': 'Lawɔs', 'bm': 'Layosi', 'bn': 'লাওস', 'br': 'Laos', 'brx': 'लाओस', 'bs': 'Laos', 'bs-Cyrl': 'Лаос', 'bs-Latn': 'Laos', 'ca': 'Lao', 'ca-ES-valencia': 'Lao', 'ccp': '𑄣𑄃𑄮𑄌𑄴', 'ce': 'Лаос', 'ceb': 'Laos', 'cgg': 'Layosi', 'chr': 'ᎴᎣᏍ', 'ckb': 'لاوس', 'cs': 'Laos', 'cv': 'Лаос', 'cy': 'Laos', 'da': 'Laos', 'dav': 'Laosi', 'de': 'Laos', 'dje': 'Laawos', 'dsb': 'Laos', 'dz': 'ལཱ་ཝོས', 'ebu': 'Laosi', 'ee': 'Laos nutome', 'el': 'Λάος', 'el-polyton': 'Λατινικὴ Ἀμερική', 'en': 'Laos', 'en-Dsrt': '𐐢𐐪𐐬𐑅', 'eo': 'Laoso', 'es': 'Laos', 'et': 'Laos', 'eu': 'Laos', 'ewo': 'Laós', 'fa': 'لائوس', 'ff': 'Lawoos', 'ff-Adlm': '𞤂𞤢𞤱𞤮𞥅𞤧', 'ff-Latn': 'Lawoos', 'fi': 'Laos', 'fil': 'Laos', 'fo': 'Laos', 'fr': 'Laos', 'fur': 'Laos', 'fy': 'Laos', 'ga': 'Laos', 'gd': 'Làthos', 'gl': 'Laos', 'gsw': 'Laaos', 'gu': 'લાઓસ', 'guz': 'Laosi', 'ha': 'Lawos', 'he': 'לאוס', 'hi': 'लाओस', 'hi-Latn': 'Laos', 'hr': 'Laos', 'hsb': 'Laos', 'ht': 'Laos', 'hu': 'Laosz', 'hy': 'Լաոս', 'ia': 'Laos', 'id': 'Laos', 'ig': 'Laos', 'is': 'Laos', 'it': 'Laos', 'ja': 'ラオス', 'jmc': 'Laosi', 'jv': 'Laos', 'ka': 'ლაოსი', 'kaa': 'Лаос', 'kaa-Cyrl': 'Лаос', 'kab': 'Laws', 'kam': 'Laosi', 'kde': 'Laosi', 'kea': 'Laus', 'kgp': 'Raosi', 'khq': 'Laawos', 'ki': 'Laosi', 'kk': 'Лаос', 'kk-Cyrl': 'Лаос', 'kln': 'Emetab Laos', 'km': 'ឡាវ', 'kn': 'ಲಾವೋಸ್', 'ko': '라오스', 'kok': 'लाओस', 'kok-Deva': 'लाओस', 'ks': 'لاس', 'ks-Arab': 'لاس', 'ksb': 'Laosi', 'ksf': 'laɔs', 'ksh': 'La’os', 'ku': 'Laos', 'kxv': 'laos', 'kxv-Deva': 'लाओस', 'kxv-Latn': 'laos', 'kxv-Orya': 'ଲାଅସ୍', 'kxv-Telu': 'లావోస్', 'ky': 'Лаос', 'lag': 'Laóosi', 'lb': 'Laos', 'lg': 'Lawosi', 'lld': 'Laos', 'ln': 'Lawosi', 'lo': 'ລາວ', 'lt': 'Laosas', 'lu': 'Lawosi', 'luo': 'Laos', 'luy': 'Laosi', 'lv': 'Laosa', 'mai': 'लाओस', 'mas': 'Laosi', 'mer': 'Laosi', 'mfe': 'Laos', 'mg': 'Laôs', 'mi': 'Rāoho', 'mk': 'Лаос', 'ml': 'ലാവോസ്', 'mn': 'Лаос', 'mr': 'लाओस', 'ms': 'Laos', 'mt': 'il-Laos', 'mua': 'Sǝr Laos', 'my': 'လာအို', 'mzn': 'لائوس', 'naq': 'Laos', 'nb': 'Laos', 'nd': 'Laos', 'ne': 'लाओस', 'nl': 'Laos', 'nmg': 'Laos', 'nn': 'Laos', 'no': 'Laos', 'nqo': 'ߟߊߐߛ', 'nyn': 'Layosi', 'om': 'Laa’oos', 'or': 'ଲାଓସ୍', 'pa': 'ਲਾਓਸ', 'pa-Guru': 'ਲਾਓਸ', 'pcm': 'Láos', 'pl': 'Laos', 'ps': 'لاوس', 'pt': 'Laos', 'qu': 'Laos', 'rm': 'Laos', 'rn': 'Layosi', 'ro': 'Laos', 'rof': 'Laosi', 'ru': 'Лаос', 'rwk': 'Laosi', 'saq': 'Laosi', 'sbp': 'Layosi', 'sc': 'Laos', 'sd': 'لائوس', 'sd-Arab': 'لائوس', 'se': 'Laos', 'seh': 'Laos', 'ses': 'Laawos', 'sg': 'Lùaôsi', 'shi': 'ⵍⴰⵡⵙ', 'shi-Latn': 'laws', 'shi-Tfng': 'ⵍⴰⵡⵙ', 'si': 'ලාඕසය', 'sk': 'Laos', 'sl': 'Laos', 'smn': 'Laos', 'sn': 'Laos', 'so': 'Laos', 'sq': 'Laos', 'sr': 'Лаос', 'sr-Cyrl': 'Лаос', 'sr-Latn': 'Laos', 'sv': 'Laos', 'sw': 'Laos', 'syr': 'ܠܐܘܣ', 'ta': 'லாவோஸ்', 'te': 'లావోస్', 'teo': 'Laosi', 'tg': 'Лаос', 'th': 'ลาว', 'ti': 'ላኦስ', 'tk': 'Laos', 'to': 'Lau', 'tr': 'Laos', 'tt': 'Лаос', 'twq': 'Laawos', 'tzm': 'Laws', 'ug': 'لائوس', 'uk': 'Лаос', 'ur': 'لاؤس', 'uz': 'Laos', 'uz-Cyrl': 'Лаос', 'uz-Latn': 'Laos', 'vai': 'ꕞꕴꔻ', 'vai-Latn': 'Lawosi', 'vai-Vaii': 'ꕞꕴꔻ', 'vec': 'Laos', 'vi': 'Lào', 'vun': 'Laosi', 'wae': 'Laos', 'wo': 'Lawos', 'xh': 'ELaos', 'xnr': 'लाओस', 'xog': 'Lawosi', 'yav': 'lawós', 'yi': 'לאַאס', 'yo': 'Laosi', 'yrl': 'Rawo', 'yue': '老撾', 'yue-Hans': '寮国', 'yue-Hant': '老撾', 'zgh': 'ⵍⴰⵡⵙ', 'zh': '老挝', 'zh-Hans': '老挝', 'zh-Hant': '寮國', 'zu': 'i-Laos'},
    'LB': {'af': 'Libanon', 'agq': 'Lɛbanè', 'ak': 'Lɛbanɔn', 'am': 'ሊባኖስ', 'ar': 'لبنان', 'as': 'লেবানন', 'asa': 'Lebanoni', 'ast': 'Líbanu', 'az': 'Livan', 'az-Cyrl': 'Ливан', 'az-Latn': 'Livan', 'bal-Latn': 'Lebnán', 'bas': 'Lèbanòn', 'be': 'Ліван', 'be-tarask': 'Ліван', 'bez': 'Hulebanoni', 'bg': 'Ливан', 'bgn': 'لیبنان', 'blo': 'Liibaaŋ', 'bm': 'Libaŋ', 'bn': 'লেবানন', 'br': 'Liban', 'brx': 'लेबान’न', 'bs': 'Liban', 'bs-Cyrl': 'Либан', 'bs-Latn': 'Liban', 'ca': 'Líban', 'ca-ES-valencia': 'Líban', 'ccp': '𑄣𑄬𑄝𑄚𑄧𑄚𑄴', 'ce': 'Ливан', 'ceb': 'Lebanon', 'cgg': 'Lebanoni', 'chr': 'ᎴᏆᎾᏂ', 'ckb': 'لوبنان', 'cs': 'Libanon', 'cv': 'Ливан', 'cy': 'Libanus', 'da': 'Libanon', 'dav': 'Lebanoni', 'de': 'Libanon', 'dje': 'Lubnaan', 'dsb': 'Libanon', 'dz': 'ལེ་བ་ནོན', 'ebu': 'Lebanoni', 'ee': 'Lebanɔn nutome', 'el': 'Λίβανος', 'el-polyton': 'Λίβανος', 'en': 'Lebanon', 'en-Dsrt': '𐐢𐐯𐐺𐐲𐑌𐐪𐑌', 'eo': 'Libano', 'es': 'Líbano', 'et': 'Liibanon', 'eu': 'Libano', 'ewo': 'Libáŋ', 'fa': 'لبنان', 'ff': 'Libaa', 'ff-Adlm': '𞤂𞤭𞤦𞤢𞤲𞤮𞥅𞤲', 'ff-Latn': 'Libaa', 'fi': 'Libanon', 'fil': 'Lebanon', 'fo': 'Libanon', 'fr': 'Liban', 'fur': 'Liban', 'fy': 'Libanon', 'ga': 'an Liobáin', 'gd': 'Leabanon', 'gl': 'Líbano', 'gsw': 'Libanon', 'gu': 'લેબનોન', 'guz': 'Lebanoni', 'ha': 'Labanan', 'he': 'לבנון', 'hi': 'लेबनान', 'hi-Latn': 'Lebanon', 'hr': 'Libanon', 'hsb': 'Libanon', 'ht': 'Liban', 'hu': 'Libanon', 'hy': 'Լիբանան', 'ia': 'Libano', 'id': 'Lebanon', 'ig': 'Lebanon', 'is': 'Líbanon', 'it': 'Libano', 'ja': 'レバノン', 'jmc': 'Lebanoni', 'jv': 'Libanon', 'ka': 'ლიბანი', 'kaa': 'Ливан', 'kaa-Cyrl': 'Ливан', 'kab': 'Lubnan', 'kam': 'Lebanoni', 'kde': 'Lebanoni', 'kea': 'Líbanu', 'kgp': 'Rimanã', 'khq': 'Lubnaan', 'ki': 'Lebanoni', 'kk': 'Ливан', 'kk-Cyrl': 'Ливан', 'kln': 'Emetab Lebanon', 'km': 'លីបង់', 'kn': 'ಲೆಬನಾನ್', 'ko': '레바논', 'kok': 'लेबनान', 'kok-Deva': 'लेबनान', 'ks': 'لؠبنان', 'ks-Arab': 'لؠبنان', 'ksb': 'Lebanoni', 'ksf': 'libáŋ', 'ksh': 'Lebbannon', 'ku': 'Libnan', 'kxv': 'lebanon', 'kxv-Deva': 'लेबनान', 'kxv-Latn': 'lebanon', 'kxv-Orya': 'ଲେବନାନ', 'kxv-Telu': 'లెబనాన్', 'ky': 'Ливан', 'lag': 'Lebanóoni', 'lb': 'Libanon', 'lg': 'Lebanoni', 'lld': 'Libanon', 'ln': 'Libá', 'lo': 'ເລບານອນ', 'lt': 'Libanas', 'lu': 'Liba', 'luo': 'Lebanon', 'luy': 'Lebanoni', 'lv': 'Libāna', 'mai': 'लेबनान', 'mas': 'Lebanoni', 'mer': 'Lebanoni', 'mfe': 'Liban', 'mg': 'Libana', 'mgh': 'Ulebanoni', 'mi': 'Repanona', 'mk': 'Либан', 'ml': 'ലെബനൻ', 'mn': 'Ливан', 'mr': 'लेबनॉन', 'ms': 'Lubnan', 'mt': 'il-Libanu', 'mua': 'Libaŋ', 'my': 'လက်ဘနွန်', 'mzn': 'لبنان', 'naq': 'Lebanonni', 'nb': 'Libanon', 'nd': 'Lebhanoni', 'ne': 'लेबनन', 'nl': 'Libanon', 'nmg': 'Libaŋ', 'nn': 'Libanon', 'no': 'Libanon', 'nqo': 'ߟߌߓߊ߲߫', 'nyn': 'Lebanoni', 'om': 'Libaanoon', 'or': 'ଲେବାନନ୍', 'pa': 'ਲੈਬਨਾਨ', 'pa-Guru': 'ਲੈਬਨਾਨ', 'pcm': 'Lẹ́bánọn', 'pl': 'Liban', 'ps': 'لبنان', 'pt': 'Líbano', 'qu': 'Líbano', 'rm': 'Libanon', 'rn': 'Libani', 'ro': 'Liban', 'rof': 'Lebanoni', 'ru': 'Ливан', 'rwk': 'Lebanoni', 'saq': 'Lebanoni', 'sbp': 'Lebanoni', 'sc': 'Lèbanu', 'sd': 'لبنان', 'sd-Arab': 'لبنان', 'se': 'Libanon', 'seh': 'Líbano', 'ses': 'Lubnaan', 'sg': 'Libùaan', 'shi': 'ⵍⵓⴱⵏⴰⵏ', 'shi-Latn': 'lubnan', 'shi-Tfng': 'ⵍⵓⴱⵏⴰⵏ', 'si': 'ලෙබනනය', 'sk': 'Libanon', 'sl': 'Libanon', 'smn': 'Libanon', 'sn': 'Lebanon', 'so': 'Lubnaan', 'sq': 'Liban', 'sr': 'Либан', 'sr-Cyrl': 'Либан', 'sr-Latn': 'Liban', 'sv': 'Libanon', 'sw': 'Lebanon', 'syr': 'ܠܒܢܢ', 'ta': 'லெபனான்', 'te': 'లెబనాన్', 'teo': 'Lebanoni', 'tg': 'Лубнон', 'th': 'เลบานอน', 'ti': 'ሊባኖስ', 'tk': 'Liwan', 'to': 'Lepanoni', 'tr': 'Lübnan', 'tt': 'Ливан', 'twq': 'Lubnaan', 'tzm': 'Lubnan', 'ug': 'لىۋان', 'uk': 'Ліван', 'ur': 'لبنان', 'uz': 'Livan', 'uz-Cyrl': 'Ливан', 'uz-Latn': 'Livan', 'vai': 'ꔒꕑꗟꘋ', 'vai-Latn': 'Lebhanɔ', 'vai-Vaii': 'ꔒꕑꗟꘋ', 'vec': 'Lìbano', 'vi': 'Li-băng', 'vun': 'Lebanoni', 'wae': 'Libanon', 'wo': 'Libaa', 'xh': 'ELebanon', 'xnr': 'लेबनान', 'xog': 'Lebanoni', 'yav': 'lipáŋ', 'yi': 'לבנון', 'yo': 'Lebanoni', 'yrl': 'Ribanu', 'yue': '黎巴嫩', 'yue-Hans': '黎巴嫩', 'yue-Hant': '黎巴嫩', 'zgh': 'ⵍⵓⴱⵏⴰⵏ', 'zh': '黎巴嫩', 'zh-Hans': '黎巴嫩', 'zh-Hant': '黎巴嫩', 'zu': 'i-Lebanon'},
    'LC': {'af': 'Sint Lucia', 'agq': 'Sɛ̀n Lushìa', 'ak': 'Saint Lucia', 'am': 'ሴንት ሉቺያ', 'ar': 'سانت لوسيا', 'as': 'ছেইণ্ট লুচিয়া', 'asa': 'Thantaluthia', 'ast': 'Santa Llucía', 'az': 'Sent-Lusiya', 'az-Cyrl': 'Сент-Лусија', 'az-Latn': 'Sent-Lusiya', 'bal-Latn': 'St. Lusiá', 'bas': 'Nûmpubi Lusì', 'be': 'Сент-Люсія', 'be-tarask': 'Сент-Люсія', 'bez': 'Husantalusia', 'bg': 'Сейнт Лусия', 'blo': 'Sɛŋ Lusiya', 'bm': 'Lusi-Senu', 'bn': 'সেন্ট লুসিয়া', 'br': 'Saint Lucia', 'brx': 'सैन्ट लुसिया', 'bs': 'Sveta Lucija', 'bs-Cyrl': 'Света Луција', 'bs-Latn': 'Sveta Lucija', 'ca': 'Saint Lucia', 'ca-ES-valencia': 'Saint Lucia', 'ccp': '𑄥𑄬𑄚𑄴𑄑𑄴 𑄣𑄪𑄥𑄨𑄠', 'ce': 'Сент-Люси', 'ceb': 'St. Lucia', 'cgg': 'Senti Rusiya', 'chr': 'ᎤᏓᏅᏘ ᎷᏏᏯ', 'ckb': 'سەن لووسیا', 'cs': 'Svatá Lucie', 'cv': 'Сент-Люсия', 'cy': 'Saint Lucia', 'da': 'Saint Lucia', 'dav': 'Santalusia', 'de': 'St. Lucia', 'dje': 'Seŋ Lussia', 'dsb': 'St. Lucia', 'dyo': 'Saŋ Lusia', 'dz': 'སེནཊ་ ལུ་སི་ཡ', 'ebu': 'Santalusia', 'ee': 'Saint Lusia nutome', 'el': 'Αγία Λουκία', 'el-polyton': 'Ἁγία Λουκία', 'en': 'St. Lucia', 'eo': 'Sankta Lucio', 'es': 'Santa Lucía', 'et': 'Saint Lucia', 'eu': 'Santa Luzia', 'ewo': 'Ǹfúfúb-Lúsia', 'fa': 'سنت لوسیا', 'ff': 'Sent Lusiyaa', 'ff-Adlm': '𞤅𞤼. 𞤂𞤵𞥅𞤧𞤭𞤴𞤢', 'ff-Latn': 'Sent Lusiyaa', 'fi': 'Saint Lucia', 'fil': 'Saint Lucia', 'fo': 'St. Lusia', 'fr': 'Sainte-Lucie', 'fur': 'Sante Lusie', 'fy': 'Saint Lucia', 'ga': 'Saint Lucia', 'gd': 'Naomh Lùisea', 'gl': 'Santa Lucía', 'gsw': 'St. Lutschiia', 'gu': 'સેંટ લુસિયા', 'guz': 'Santalusia', 'ha': 'San Lusiya', 'he': 'סנט לוסיה', 'hi': 'सेंट लूसिया', 'hi-Latn': 'St. Lucia', 'hr': 'Sveta Lucija', 'hsb': 'St. Lucia', 'ht': 'Sainte-Lucie', 'hu': 'Saint Lucia', 'hy': 'Սենթ Լյուսիա', 'ia': 'Sancte Lucia', 'id': 'Saint Lucia', 'ig': 'St. Lucia', 'is': 'Sankti Lúsía', 'it': 'Saint Lucia', 'ja': 'セントルシア', 'jmc': 'Santalusia', 'jv': 'Santa Lusia', 'ka': 'სენტ-ლუსია', 'kaa': 'Сент-Люсия', 'kaa-Cyrl': 'Сент-Люсия', 'kab': 'San Lučya', 'kam': 'Santalusia', 'kde': 'Santalusia', 'kea': 'Santa Lúsia', 'kgp': 'Sỹta Rusija', 'khq': 'Seŋ Lussia', 'ki': 'Santalusia', 'kk': 'Сент-Люсия', 'kk-Cyrl': 'Сент-Люсия', 'kln': 'Emetab Lucia Ne', 'km': 'សាំងលូស៊ី', 'kn': 'ಸೇಂಟ್ ಲೂಸಿಯಾ', 'ko': '세인트루시아', 'kok': 'सँट लुसिया', 'kok-Deva': 'सँट लुसिया', 'ks': 'سینٹ لوٗسِیا', 'ks-Arab': 'سینٹ لوٗسِیا', 'ksb': 'Santalusia', 'ksf': 'sɛntlísí', 'ksh': 'de Ensel Zint-Lutscha', 'ku': 'Saint Lucia', 'kxv': 'seint lusiya', 'kxv-Deva': 'सेंट लूसिया', 'kxv-Latn': 'seint lusiya', 'kxv-Orya': 'ସେଣ୍ଟ. ଲୁସିଆ', 'kxv-Telu': 'సెయింట్ లూసియా', 'ky': 'Сент-Люсия', 'lag': 'Mʉtakatíifu Lusíia', 'lb': 'St. Lucia', 'lg': 'Senti Luciya', 'lld': 'St. Lucia', 'ln': 'Sántu lisi', 'lo': 'ເຊນ ລູເຊຍ', 'lt': 'Sent Lusija', 'lu': 'Santu lisi', 'luo': 'Saint Lucia', 'luy': 'Santalusia', 'lv': 'Sentlūsija', 'mai': 'सेंट लूसिया', 'mas': 'Santalusia', 'mer': 'Santalusia', 'mfe': 'Sainte-Lucie', 'mg': 'Sainte-Lucie', 'mgh': 'Usantalusia', 'mi': 'Hato Ruhia', 'mk': 'Сент Лусија', 'ml': 'സെന്റ് ലൂസിയ', 'mn': 'Сент Люсиа', 'mr': 'सेंट ल्यूसिया', 'ms': 'Saint Lucia', 'mt': 'Saint Lucia', 'mua': 'Sǝr Lucia', 'my': 'စိန့်လူစီယာ', 'mzn': 'سنت لوسیا', 'naq': 'Saint Lucia', 'nb': 'St. Lucia', 'nd': 'Saint Lucia', 'ne': 'सेन्ट लुसिया', 'nl': 'Saint Lucia', 'nmg': 'Saint Lucia', 'nn': 'St. Lucia', 'no': 'St. Lucia', 'nqo': 'ߛߍ߲ߕ-ߟߎ߳ߛߌ߫', 'nyn': 'Senti Rusiya', 'om': 'St. Suusiyaa', 'or': 'ସେଣ୍ଟ ଲୁସିଆ', 'pa': 'ਸੇਂਟ ਲੂਸੀਆ', 'pa-Guru': 'ਸੇਂਟ ਲੂਸੀਆ', 'pcm': 'Sent Lúshia', 'pl': 'Saint Lucia', 'ps': 'سینټ لوسیا', 'pt': 'Santa Lúcia', 'qu': 'Santa Lucia', 'rm': 'Saint Lucia', 'rn': 'Selusiya', 'ro': 'Sfânta Lucia', 'rof': 'Santalusia', 'ru': 'Сент-Люсия', 'rwk': 'Santalusia', 'saq': 'Santalusia', 'sbp': 'Santalusiya', 'sc': 'Santa Lughia', 'sd': 'سينٽ لوسيا', 'sd-Arab': 'سينٽ لوسيا', 'se': 'Saint Lucia', 'seh': 'Santa Lúcia', 'ses': 'Seŋ Lussia', 'sg': 'Sênt-Lisïi', 'shi': 'ⵙⴰⵏⵜⵍⵓⵙⵉ', 'shi-Latn': 'santlusi', 'shi-Tfng': 'ⵙⴰⵏⵜⵍⵓⵙⵉ', 'si': 'ශාන්ත ලුසියා', 'sk': 'Svätá Lucia', 'sl': 'Saint Lucia', 'smn': 'St. Lucia', 'sn': 'Saint Lucia', 'so': 'St. Lusia', 'sq': 'Shën-Luçia', 'sr': 'Света Луција', 'sr-Cyrl': 'Света Луција', 'sr-Latn': 'Sveta Lucija', 'sv': 'S:t Lucia', 'sw': 'St. Lucia', 'syr': 'ܡܪܬܝ ܠܘܫܐ', 'ta': 'செயின்ட் லூசியா', 'te': 'సెయింట్ లూసియా', 'teo': 'Santalusia', 'tg': 'Сент-Люсия', 'th': 'เซนต์ลูเซีย', 'ti': 'ቅድስቲ ሉስያ', 'tk': 'Sent-Lýusiýa', 'to': 'Sā Lūsia', 'tr': 'Saint Lucia', 'tt': 'Сент-Люсия', 'twq': 'Seŋ Lussia', 'tzm': 'Santelusya', 'ug': 'ساينت لۇسىيە', 'uk': 'Сент-Люсія', 'ur': 'سینٹ لوسیا', 'uz': 'Sent-Lyusiya', 'uz-Cyrl': 'Сент-Люсия', 'uz-Latn': 'Sent-Lyusiya', 'vai': 'ꔻꘋ ꖨꔻꕩ', 'vai-Latn': 'Siŋ Lusiya', 'vai-Vaii': 'ꔻꘋ ꖨꔻꕩ', 'vec': 'S. Lusìa', 'vi': 'St. Lucia', 'vun': 'Santalusia', 'wae': 'St. Lucia', 'wo': 'Saŋ Lusi', 'xh': 'ESt. Lucia', 'xnr': 'सेंट लूसिया', 'xog': 'Senti Luciya', 'yav': 'sɛ́ŋtɛ́lusí', 'yo': 'Luṣia', 'yrl': 'Sãta Lusiya', 'yue': '聖露西亞', 'yue-Hans': '圣露西亚', 'yue-Hant': '聖露西亞', 'zgh': 'ⵙⴰⵏⵜⵍⵓⵙⵉ', 'zh': '圣卢西亚', 'zh-Hans': '圣卢西亚', 'zh-Hant': '聖露西亞', 'zu': 'i-Saint Lucia'},
    'LI': {'af': 'Liechtenstein', 'agq': 'Letɨnshɨ̀n', 'ak': 'Lektenstaen', 'am': 'ሊችተንስታይን', 'ar': 'ليختنشتاين', 'as': 'লিচটেনষ্টেইন', 'asa': 'Lishenteni', 'ast': 'Liechtenstein', 'az': 'Lixtenşteyn', 'az-Cyrl': 'Лихтенштејн', 'az-Latn': 'Lixtenşteyn', 'bal-Latn': 'Lichtenstén', 'bas': 'Ligstɛntàn', 'be': 'Ліхтэнштэйн', 'be-tarask': 'Ліхтэнштэйн', 'bez': 'Hulishenteni', 'bg': 'Лихтенщайн', 'blo': 'Liishtɛntaɩn', 'bm': 'Lisɛnsitayini', 'bn': 'লিচেনস্টেইন', 'br': 'Liechtenstein', 'brx': 'लिक्तोनस्ताईन', 'bs': 'Lihtenštajn', 'bs-Cyrl': 'Лихтенштајн', 'bs-Latn': 'Lihtenštajn', 'ca': 'Liechtenstein', 'ca-ES-valencia': 'Liechtenstein', 'ccp': '𑄣𑄨𑄌𑄬𑄚𑄴𑄥𑄳𑄑𑄬𑄃𑄨𑄚𑄴', 'ce': 'Лихтенштейн', 'ceb': 'Liechtenstein', 'cgg': 'Lishenteni', 'chr': 'ᎵᎦᏗᏂᏍᏓᏂ', 'ckb': 'لیختنشتاین', 'cs': 'Lichtenštejnsko', 'cv': 'Лихтенштейн', 'cy': 'Liechtenstein', 'da': 'Liechtenstein', 'dav': 'Lishenteni', 'de': 'Liechtenstein', 'dje': 'Liechtenstein', 'dsb': 'Liechtenstein', 'dz': 'ལིཀ་ཏནས་ཏ་ཡིན', 'ebu': 'Lishenteni', 'ee': 'Litsenstein nutome', 'el': 'Λιχτενστάιν', 'el-polyton': 'Λιχτενστάιν', 'en': 'Liechtenstein', 'en-Dsrt': '𐐢𐐮𐐿𐐻𐐲𐑌𐑅𐐻𐐴𐑌', 'eo': 'Liĥtenŝtejno', 'es': 'Liechtenstein', 'et': 'Liechtenstein', 'eu': 'Liechtenstein', 'ewo': 'Lísə́sə́táin', 'fa': 'لیختن\u200cاشتاین', 'ff': 'Lincenstayn', 'ff-Adlm': '𞤂𞤭𞤧𞤼𞤫𞤲𞤧𞤭𞤼𞤫𞥅𞤲', 'ff-Latn': 'Lincenstayn', 'fi': 'Liechtenstein', 'fil': 'Liechtenstein', 'fo': 'Liktinstein', 'fr': 'Liechtenstein', 'fur': 'Liechtenstein', 'fy': 'Liechtenstein', 'ga': 'Lichtinstéin', 'gd': 'Lichtenstein', 'gl': 'Liechtenstein', 'gsw': 'Liächteschtäi', 'gu': 'લૈચટેંસ્ટેઇન', 'guz': 'Lishenteni', 'ha': 'Licansitan', 'he': 'ליכטנשטיין', 'hi': 'लिचेंस्टीन', 'hi-Latn': 'Liechtenstein', 'hr': 'Lihtenštajn', 'hsb': 'Liechtenstein', 'ht': 'Liechtenstein', 'hu': 'Liechtenstein', 'hy': 'Լիխտենշտեյն', 'ia': 'Liechtenstein', 'id': 'Liechtenstein', 'ig': 'Liechtenstein', 'is': 'Liechtenstein', 'it': 'Liechtenstein', 'ja': 'リヒテンシュタイン', 'jmc': 'Lishenteni', 'jv': 'Liktenstén', 'ka': 'ლიხტენშტაინი', 'kaa': 'Лихтенштейн', 'kaa-Cyrl': 'Лихтенштейн', 'kab': 'Layctenstan', 'kam': 'Lishenteni', 'kde': 'Lishenteni', 'kea': 'Lixenstain', 'kgp': 'Rinhsiténh-tajin', 'khq': 'Liechtenstein', 'ki': 'Lishenteni', 'kk': 'Лихтенштейн', 'kk-Cyrl': 'Лихтенштейн', 'kln': 'Emetab Liechtenstein', 'km': 'លិចតិនស្ដាញ', 'kn': 'ಲಿಚೆನ್\u200cಸ್ಟೈನ್', 'ko': '리히텐슈타인', 'kok': 'लिचेंस्टीन', 'kok-Deva': 'लिचेंस्टीन', 'ks': 'لِکٹیسٹیٖن', 'ks-Arab': 'لِکٹیسٹیٖن', 'ksb': 'Lishenteni', 'ksf': 'lictɛnstɛ́n', 'ksh': 'Lischteschtein', 'ku': 'Liechtenstein', 'kxv': 'likṭensṭein', 'kxv-Deva': 'लिक्टेन्स्टीन', 'kxv-Latn': 'likṭensṭein', 'kxv-Orya': 'ଲିକ୍\u200dଟନ୍\u200dସ୍ଟାଇନ୍', 'kxv-Telu': 'లిక్టెన్\u200cస్టెయిన్', 'ky': 'Лихтенштейн', 'lag': 'Lishentéeni', 'lb': 'Liechtenstein', 'lg': 'Licitensitayini', 'lld': 'Liechtenstein', 'ln': 'Lishɛteni', 'lo': 'ລິດເທນສະຕາຍ', 'lt': 'Lichtenšteinas', 'lu': 'Lishuteni', 'luo': 'Liechtenstein', 'luy': 'Lishenteni', 'lv': 'Lihtenšteina', 'mai': 'लिख्तेंस्ताइन', 'mas': 'Lishenteni', 'mer': 'Lishenteni', 'mfe': 'Liechtenstein', 'mg': 'Listenstein', 'mgh': 'Ushenteni', 'mi': 'Rīkenetaina', 'mk': 'Лихтенштајн', 'ml': 'ലിച്ചൺസ്റ്റൈൻ', 'mn': 'Лихтенштайн', 'mr': 'लिक्टेनस्टाइन', 'ms': 'Liechtenstein', 'mt': 'il-Liechtenstein', 'mua': 'Lichtǝnsteiŋ', 'my': 'လစ်တန်စတိန်း', 'mzn': 'لیختن اشتاین', 'naq': 'Liechtensteinni', 'nb': 'Liechtenstein', 'nd': 'Liechtenstein', 'ne': 'लिकटेन्सटाइन', 'nl': 'Liechtenstein', 'nmg': 'Lishenstein', 'nn': 'Liechtenstein', 'no': 'Liechtenstein', 'nqo': 'ߟߎߛߑߕߊ߲ߛߑߕߍ߲߫', 'nyn': 'Lishenteni', 'om': 'Lichistensteyin', 'or': 'ଲିକ୍ଟନ୍\u200cଷ୍ଟାଇନ୍\u200c', 'pa': 'ਲਿਚੇਂਸਟਾਇਨ', 'pa-Guru': 'ਲਿਚੇਂਸਟਾਇਨ', 'pcm': 'Líktẹ́nstain', 'pl': 'Liechtenstein', 'ps': 'لیختن اشتاین', 'pt': 'Liechtenstein', 'qu': 'Liechtenstein', 'rm': 'Liechtenstein', 'rn': 'Lishyitenshitayini', 'ro': 'Liechtenstein', 'rof': 'Lishenteni', 'ru': 'Лихтенштейн', 'rwk': 'Lishenteni', 'saq': 'Lishenteni', 'sbp': 'Lisheniteni', 'sc': 'Liechtenstein', 'sd': 'لچي ٽينسٽين', 'sd-Arab': 'لچي ٽينسٽين', 'se': 'Liechtenstein', 'seh': 'Liechtenstein', 'ses': 'Liechtenstein', 'sg': 'Liechtenstein,', 'shi': 'ⵍⵉⴽⵉⵏⵛⵜⴰⵢⵏ', 'shi-Latn': 'likinctayn', 'shi-Tfng': 'ⵍⵉⴽⵉⵏⵛⵜⴰⵢⵏ', 'si': 'ලික්ටන්ස්ටයින්', 'sk': 'Lichtenštajnsko', 'sl': 'Lihtenštajn', 'smn': 'Liechtenstein', 'sn': 'Liechtenstein', 'so': 'Liyjtensteyn', 'sq': 'Lihtenshtajn', 'sr': 'Лихтенштајн', 'sr-Cyrl': 'Лихтенштајн', 'sr-Latn': 'Lihtenštajn', 'sv': 'Liechtenstein', 'sw': 'Liechtenstein', 'syr': 'ܠܝܟܛܢܫܛܝܢ', 'ta': 'லிச்செண்ஸ்டெய்ன்', 'te': 'లిక్టెన్\u200cస్టెయిన్', 'teo': 'Lishenteni', 'tg': 'Лихтенштейн', 'th': 'ลิกเตนสไตน์', 'ti': 'ሊኽተንሽታይን', 'tk': 'Lihtenşteýn', 'to': 'Likitenisiteini', 'tr': 'Liechtenstein', 'tt': 'Лихтенштейн', 'twq': 'Liechtenstein', 'tzm': 'Lictencṭayn', 'ug': 'لىكتېنستېين', 'uk': 'Ліхтенштейн', 'ur': 'لیشٹنسٹائن', 'uz': 'Lixtenshteyn', 'uz-Cyrl': 'Лихтенштейн', 'uz-Latn': 'Lixtenshteyn', 'vai': 'ꔷꗿꘋꔻꗳꘋ', 'vai-Vaii': 'ꔷꗿꘋꔻꗳꘋ', 'vec': 'Liechtenstein', 'vi': 'Liechtenstein', 'vun': 'Lishenteni', 'wae': 'Liečteštei', 'wo': 'Liktensteyin', 'xh': 'ELiechtenstein', 'xnr': 'लिचेंस्टीन', 'xog': 'Licitensitayini', 'yav': 'lístɛ́nsitáyin', 'yi': 'ליכטנשטיין', 'yo': 'Lẹṣitẹnisiteni', 'yrl': 'Rixitẽxitaĩ', 'yue': '列支敦斯登', 'yue-Hans': '列支敦斯登', 'yue-Hant': '列支敦斯登', 'zgh': 'ⵍⵉⴽⵉⵏⵛⵜⴰⵢⵏ', 'zh': '列支敦士登', 'zh-Hans': '列支敦士登', 'zh-Hant': '列支敦斯登', 'zu': 'i-Liechtenstein'},
    'LK': {'af': 'Sri Lanka', 'agq': 'Sɨ̀le Laŋkà', 'ak': 'Sri Lanka', 'am': 'ሲሪላንካ', 'ar': 'سريلانكا', 'as': 'শ্রীলংকা', 'asa': 'Thirilanka', 'ast': 'Sri Lanka', 'az': 'Şri-Lanka', 'az-Cyrl': 'Шри-Ланка', 'az-Latn': 'Şri-Lanka', 'bal-Latn': 'Sari Lanká', 'bas': 'Srìlaŋkà', 'be': 'Шры-Ланка', 'be-tarask': 'Шры-Ланка', 'bez': 'Husirilanka', 'bg': 'Шри Ланка', 'blo': 'Siri Laŋkaa', 'bm': 'Sirilanka', 'bn': 'শ্রীলঙ্কা', 'br': 'Sri Lanka', 'brx': 'श्रीलंका', 'bs': 'Šri Lanka', 'bs-Cyrl': 'Шри Ланка', 'bs-Latn': 'Šri Lanka', 'ca': 'Sri Lanka', 'ca-ES-valencia': 'Sri Lanka', 'ccp': '𑄥𑄳𑄢𑄨𑄣𑄧𑄁𑄇', 'ce': 'Шри-Ланка', 'ceb': 'Sri Lanka', 'cgg': 'Siriranka', 'chr': 'ᏍᎵ ᎳᏂᎧ', 'ckb': 'سریلانکا', 'cs': 'Srí Lanka', 'cv': 'Шри-Ланка', 'cy': 'Sri Lanka', 'da': 'Sri Lanka', 'dav': 'Sirilanka', 'de': 'Sri Lanka', 'dje': 'Srilanka', 'dsb': 'Sri Lanka', 'dyo': 'Siri Lanka', 'dz': 'ཤྲཱི་ལང་ཀ', 'ebu': 'Sirilanka', 'ee': 'Sri Lanka nutome', 'el': 'Σρι Λάνκα', 'el-polyton': 'Σρὶ Λάνκα', 'en': 'Sri Lanka', 'en-Dsrt': '𐐟𐑉𐐨 𐐢𐐰𐑍𐐿𐐲', 'eo': 'Srilanko', 'es': 'Sri Lanka', 'et': 'Sri Lanka', 'eu': 'Sri Lanka', 'ewo': 'Səri Laŋká', 'fa': 'سری\u200cلانکا', 'ff': 'Siri Lanka', 'ff-Adlm': '𞤅𞤭𞤪 𞤂𞤢𞤲𞤳𞤢𞥄', 'ff-Latn': 'Siri Lanka', 'fi': 'Sri Lanka', 'fil': 'Sri Lanka', 'fo': 'Sri Lanka', 'fr': 'Sri Lanka', 'fur': 'Sri Lanka', 'fy': 'Sri Lanka', 'ga': 'Srí Lanca', 'gd': 'Sri Lanca', 'gl': 'Sri Lanka', 'gsw': 'Schri Lanka', 'gu': 'શ્રીલંકા', 'guz': 'Sirilanka', 'ha': 'Siri Lanka', 'he': 'סרי לנקה', 'hi': 'श्रीलंका', 'hi-Latn': 'Sri Lanka', 'hr': 'Šri Lanka', 'hsb': 'Sri Lanka', 'ht': 'Sri Lanka', 'hu': 'Srí Lanka', 'hy': 'Շրի Լանկա', 'ia': 'Sri Lanka', 'id': 'Sri Lanka', 'ie': 'Sri-Lanka', 'ig': 'Sri Lanka', 'is': 'Srí Lanka', 'it': 'Sri Lanka', 'ja': 'スリランカ', 'jmc': 'Sirilanka', 'jv': 'Sri Lanka', 'ka': 'შრი-ლანკა', 'kaa': 'Шри-Ланка', 'kaa-Cyrl': 'Шри-Ланка', 'kab': 'Sri Lanka', 'kam': 'Sirilanka', 'kde': 'Sililanka', 'kea': 'Sri Lanka', 'kgp': 'Siri Rỹnka', 'khq': 'Srilanka', 'ki': 'Sirilanka', 'kk': 'Шри-Ланка', 'kk-Cyrl': 'Шри-Ланка', 'kln': 'Emetab Sri Lanka', 'km': 'ស្រីលង្កា', 'kn': 'ಶ್ರೀಲಂಕಾ', 'ko': '스리랑카', 'kok': 'श्री लंका', 'kok-Deva': 'श्री लंका', 'ks': 'سری لنکا', 'ks-Arab': 'سری لنکا', 'ksb': 'Sililanka', 'ksf': 'srílaŋka', 'ksh': 'Sri Lanka', 'ku': 'Srî Lanka', 'kxv': 'sri lanka', 'kxv-Deva': 'स्रीलंका', 'kxv-Latn': 'sri lanka', 'kxv-Orya': 'ସ୍ରି ଲଙ୍କା', 'kxv-Telu': 'స్రీ లంక', 'ky': 'Шри-Ланка', 'lag': 'Siriláanka', 'lb': 'Sri Lanka', 'lg': 'Sirilanka', 'lld': 'Sri Lanka', 'ln': 'Sirilanka', 'lo': 'ສີລັງກາ', 'lt': 'Šri Lanka', 'lu': 'Sirilanka', 'luo': 'Sri Lanka', 'luy': 'Sirilanka', 'lv': 'Šrilanka', 'mai': 'श्रीलंका', 'mas': 'Sirilanka', 'mer': 'Sirilanka', 'mfe': 'Sri Lanka', 'mg': 'Sri Lanka', 'mgh': 'Usirilanka', 'mi': 'Hiri Rānaka', 'mk': 'Шри Ланка', 'ml': 'ശ്രീലങ്ക', 'mn': 'Шри-Ланка', 'mr': 'श्रीलंका', 'ms': 'Sri Lanka', 'mt': 'is-Sri Lanka', 'mua': 'Sǝr Lanka', 'my': 'သီရိလင်္ကာ', 'mzn': 'سریلانکا', 'naq': 'Sri Lankab', 'nb': 'Sri Lanka', 'nd': 'Sri Lanka', 'ne': 'श्रीलङ्का', 'nl': 'Sri Lanka', 'nmg': 'Sri Lanka', 'nn': 'Sri Lanka', 'no': 'Sri Lanka', 'nqo': 'ߛߙߌߟߊ߲ߞߊ߫', 'nyn': 'Siriranka', 'om': 'Siri Laankaa', 'or': 'ଶ୍ରୀଲଙ୍କା', 'pa': 'ਸ੍ਰੀ ਲੰਕਾ', 'pa-Guru': 'ਸ੍ਰੀ ਲੰਕਾ', 'pcm': 'Sri Lánka', 'pl': 'Sri Lanka', 'ps': 'سريلنکا', 'pt': 'Sri Lanka', 'qu': 'Sri Lanka', 'rm': 'Sri Lanka', 'rn': 'Sirilanka', 'ro': 'Sri Lanka', 'rof': 'Sirilanka', 'ru': 'Шри-Ланка', 'rwk': 'Sirilanka', 'saq': 'Sirilanka', 'sbp': 'Sililanka', 'sc': 'Sri Lanka', 'sd': 'سري لنڪا', 'sd-Arab': 'سري لنڪا', 'se': 'Sri Lanka', 'seh': 'Sri Lanka', 'ses': 'Srilanka', 'sg': 'Sirî-Lanka', 'shi': 'ⵙⵔⵉⵍⴰⵏⴽⴰ', 'shi-Latn': 'srilanka', 'shi-Tfng': 'ⵙⵔⵉⵍⴰⵏⴽⴰ', 'si': 'ශ්\u200dරී ලංකාව', 'sk': 'Srí Lanka', 'sl': 'Šrilanka', 'smn': 'Sri Lanka', 'sn': 'Sri Lanka', 'so': 'Sirilaanka', 'sq': 'Sri-Lankë', 'sr': 'Шри Ланка', 'sr-Cyrl': 'Шри Ланка', 'sr-Latn': 'Šri Lanka', 'sv': 'Sri Lanka', 'sw': 'Sri Lanka', 'syr': 'ܫܪܝ ܠܐܢܟܐ', 'ta': 'இலங்கை', 'te': 'శ్రీలంక', 'teo': 'Sirilanka', 'tg': 'Шри-Ланка', 'th': 'ศรีลังกา', 'ti': 'ስሪ ላንካ', 'tk': 'Şri-Lanka', 'to': 'Sīlangikā', 'tr': 'Sri Lanka', 'tt': 'Шри-Ланка', 'twq': 'Srilanka', 'tzm': 'Srilanka', 'ug': 'سىرىلانكا', 'uk': 'Шрі-Ланка', 'ur': 'سری لنکا', 'uz': 'Shri-Lanka', 'uz-Cyrl': 'Шри-Ланка', 'uz-Latn': 'Shri-Lanka', 'vai': 'ꖬꔸ ꕞꘋꕪ', 'vai-Latn': 'Suri Laŋka', 'vai-Vaii': 'ꖬꔸ ꕞꘋꕪ', 'vec': 'Sri Lanka', 'vi': 'Sri Lanka', 'vun': 'Sirilanka', 'wae': 'Sri Lanka', 'wo': 'Siri Lànka', 'xh': 'ESri Lanka', 'xnr': 'श्रीलंका', 'xog': 'Sirilanka', 'yav': 'silíláŋka', 'yi': 'סרי־לאַנקאַ', 'yo': 'Siri Lanka', 'yrl': 'Siri Rãka', 'yue': '斯里蘭卡', 'yue-Hans': '斯里兰卡', 'yue-Hant': '斯里蘭卡', 'zgh': 'ⵙⵔⵉⵍⴰⵏⴽⴰ', 'zh': '斯里兰卡', 'zh-Hans': '斯里兰卡', 'zh-Hant': '斯里蘭卡', 'zu': 'i-Sri Lanka'},
    'LR': {'af': 'Liberië', 'agq': 'Làebɛlìa', 'ak': 'Laeberia', 'am': 'ላይቤሪያ', 'ar': 'ليبيريا', 'as': 'লিবেৰিয়া', 'asa': 'Liberia', 'ast': 'Liberia', 'az': 'Liberiya', 'az-Cyrl': 'Либерија', 'az-Latn': 'Liberiya', 'bal-Latn': 'Láibériá', 'bas': 'Lìberìà', 'be': 'Ліберыя', 'be-tarask': 'Ліберыя', 'bez': 'Hulibelia', 'bg': 'Либерия', 'blo': 'Liberiya', 'bm': 'Liberiya', 'bn': 'লাইবেরিয়া', 'br': 'Liberia', 'brx': 'लाइबेरिया', 'bs': 'Liberija', 'bs-Cyrl': 'Либерија', 'bs-Latn': 'Liberija', 'ca': 'Libèria', 'ca-ES-valencia': 'Libèria', 'ccp': '𑄃𑄭𑄝𑄬𑄢𑄨𑄠', 'ce': 'Либери', 'ceb': 'Liberia', 'cgg': 'Liberiya', 'chr': 'ᎳᏈᎵᏯ', 'ckb': 'لیبەریا', 'cs': 'Libérie', 'cv': 'Либери', 'cy': 'Liberia', 'da': 'Liberia', 'dav': 'Liberia', 'de': 'Liberia', 'dje': 'Liberia', 'dsb': 'Liberija', 'dyo': 'Liberia', 'dz': 'ལཱའི་བེ་རི་ཡ', 'ebu': 'Liberia', 'ee': 'Liberia nutome', 'el': 'Λιβερία', 'el-polyton': 'Λιβερία', 'en': 'Liberia', 'en-Dsrt': '𐐢𐐴𐐺𐐮𐑉𐐨𐐲', 'eo': 'Liberio', 'es': 'Liberia', 'et': 'Libeeria', 'eu': 'Liberia', 'ewo': 'Libéria', 'fa': 'لیبریا', 'ff': 'Liberiyaa', 'ff-Adlm': '𞤂𞤢𞤦𞤭𞤪𞤭𞤴𞤢𞥄', 'ff-Latn': 'Liberiyaa', 'fi': 'Liberia', 'fil': 'Liberia', 'fo': 'Liberia', 'fr': 'Liberia', 'fur': 'Liberie', 'fy': 'Liberia', 'ga': 'An Libéir', 'gd': 'Libèir', 'gl': 'Liberia', 'gsw': 'Libeeria', 'gu': 'લાઇબેરિયા', 'guz': 'Liberia', 'ha': 'Laberiya', 'he': 'ליבריה', 'hi': 'लाइबेरिया', 'hi-Latn': 'Liberia', 'hr': 'Liberija', 'hsb': 'Liberija', 'ht': 'Liberia', 'hu': 'Libéria', 'hy': 'Լիբերիա', 'ia': 'Liberia', 'id': 'Liberia', 'ig': 'Liberia', 'is': 'Líbería', 'it': 'Liberia', 'ja': 'リベリア', 'jgo': 'Libɛrî', 'jmc': 'Liberia', 'jv': 'Libèria', 'ka': 'ლიბერია', 'kaa': 'Либерия', 'kaa-Cyrl': 'Либерия', 'kab': 'Libirya', 'kam': 'Liberia', 'kde': 'Libelia', 'kea': 'Libéria', 'kgp': 'Rimérija', 'khq': 'Liberia', 'ki': 'Liberia', 'kk': 'Либерия', 'kk-Cyrl': 'Либерия', 'kln': 'Emetab Liberia', 'km': 'លីបេរីយ៉ា', 'kn': 'ಲಿಬೇರಿಯಾ', 'ko': '라이베리아', 'kok': 'लायबेरीया', 'kok-Deva': 'लायबेरीया', 'ks': 'لایبیرِیا', 'ks-Arab': 'لایبیرِیا', 'ksb': 'Libelia', 'ksf': 'libɛrya', 'ksh': 'Libeerija', 'ku': 'Lîberya', 'kxv': 'laiberīya', 'kxv-Deva': 'लाइबेरिया', 'kxv-Latn': 'laiberīya', 'kxv-Orya': 'ଲାଇବେରିଆ', 'kxv-Telu': 'లైబీరియా', 'ky': 'Либерия', 'lag': 'Liibéria', 'lb': 'Liberia', 'lg': 'Liberya', 'lld': 'Liberia', 'ln': 'Libériya', 'lo': 'ລິເບີເຣຍ', 'lt': 'Liberija', 'lu': 'Liberiya', 'luo': 'Liberia', 'luy': 'Liberia', 'lv': 'Libērija', 'mai': 'लाइबेरिया', 'mas': 'Liberia', 'mer': 'Liberia', 'mfe': 'Liberia', 'mg': 'Liberia', 'mgh': 'Uliberia', 'mi': 'Raipiria', 'mk': 'Либерија', 'ml': 'ലൈബീരിയ', 'mn': 'Либери', 'mr': 'लायबेरिया', 'ms': 'Liberia', 'mt': 'il-Liberja', 'mua': 'Liberiya', 'my': 'လိုက်ဘေးရီးယား', 'mzn': 'لیبریا', 'naq': 'Liberiab', 'nb': 'Liberia', 'nd': 'Libheriya', 'ne': 'लाइबेरिया', 'nl': 'Liberia', 'nmg': 'Liberia', 'nn': 'Liberia', 'no': 'Liberia', 'nqo': 'ߟߌߓߋߙߌߦߊ߫', 'nyn': 'Liberiya', 'om': 'Laayibeeriyaa', 'or': 'ଲାଇବେରିଆ', 'pa': 'ਲਾਈਬੀਰੀਆ', 'pa-Guru': 'ਲਾਈਬੀਰੀਆ', 'pcm': 'Laibẹ́ria', 'pl': 'Liberia', 'ps': 'لايبيريا', 'pt': 'Libéria', 'qu': 'Liberia', 'rm': 'Liberia', 'rn': 'Liberiya', 'ro': 'Liberia', 'rof': 'Liberia', 'ru': 'Либерия', 'rwk': 'Liberia', 'saq': 'Liberia', 'sbp': 'Libeliya', 'sc': 'Libèria', 'sd': 'لائبیریا', 'sd-Arab': 'لائبیریا', 'se': 'Liberia', 'seh': 'Libéria', 'ses': 'Liberia', 'sg': 'Liberïa', 'shi': 'ⵍⵉⴱⵉⵔⵢⴰ', 'shi-Latn': 'libirya', 'shi-Tfng': 'ⵍⵉⴱⵉⵔⵢⴰ', 'si': 'ලයිබීරියාව', 'sk': 'Libéria', 'sl': 'Liberija', 'smn': 'Liberia', 'sn': 'Liberia', 'so': 'Laybeeriya', 'sq': 'Liberi', 'sr': 'Либерија', 'sr-Cyrl': 'Либерија', 'sr-Latn': 'Liberija', 'sv': 'Liberia', 'sw': 'Liberia', 'syr': 'ܠܝܒܝܪܝܐ', 'ta': 'லைபீரியா', 'te': 'లైబీరియా', 'teo': 'Liberia', 'tg': 'Либерия', 'th': 'ไลบีเรีย', 'ti': 'ላይበርያ', 'tk': 'Liberiýa', 'to': 'Laipelia', 'tr': 'Liberya', 'tt': 'Либерия', 'twq': 'Liberia', 'tzm': 'Libirya', 'ug': 'لىبېرىيە', 'uk': 'Ліберія', 'ur': 'لائبیریا', 'uz': 'Liberiya', 'uz-Cyrl': 'Либерия', 'uz-Latn': 'Liberiya', 'vai': 'ꕞꔤꔫꕩ', 'vai-Latn': 'Laibhiya', 'vai-Vaii': 'ꕞꔤꔫꕩ', 'vec': 'Liberia', 'vi': 'Liberia', 'vun': 'Liberia', 'wae': 'Liberia', 'wo': 'Liberiya', 'xh': 'ELiberia', 'xnr': 'लाइबेरिया', 'xog': 'Liberya', 'yav': 'lipélia', 'yi': 'ליבעריע', 'yo': 'Laberia', 'yrl': 'Ribériya', 'yue': '賴比瑞亞', 'yue-Hans': '赖比瑞亚', 'yue-Hant': '賴比瑞亞', 'zgh': 'ⵍⵉⴱⵉⵔⵢⴰ', 'zh': '利比里亚', 'zh-Hans': '利比里亚', 'zh-Hant': '賴比瑞亞', 'zu': 'i-Liberia'},
    'LS': {'af': 'Lesotho', 'agq': 'Lɛ̀sotù', 'ak': 'Lesoto', 'am': 'ሌሶቶ', 'ar': 'ليسوتو', 'as': 'লেছ’থ’', 'asa': 'Lethoto', 'ast': 'Lesothu', 'az': 'Lesoto', 'az-Cyrl': 'Лесото', 'az-Latn': 'Lesoto', 'bal-Latn': 'Lesótó', 'bas': 'Lesòtò', 'be': 'Лесота', 'be-tarask': 'Лесота', 'bez': 'Hulesoto', 'bg': 'Лесото', 'blo': 'Lesotoo', 'bm': 'Lesoto', 'bn': 'লেসোথো', 'br': 'Lesotho', 'brx': 'लेस’थ’', 'bs': 'Lesoto', 'bs-Cyrl': 'Лесото', 'bs-Latn': 'Lesoto', 'ca': 'Lesotho', 'ca-ES-valencia': 'Lesotho', 'ccp': '𑄣𑄬𑄥𑄮𑄗𑄮', 'ce': 'Лесото', 'ceb': 'Lesotho', 'cgg': 'Lesotho', 'chr': 'ᎴᏐᏙ', 'ckb': 'لەسۆتۆ', 'cs': 'Lesotho', 'cv': 'Лесото', 'cy': 'Lesotho', 'da': 'Lesotho', 'dav': 'Lesoto', 'de': 'Lesotho', 'dje': 'Leesoto', 'dsb': 'Lesotho', 'dz': 'ལཻ་སོ་ཐོ', 'ebu': 'Lesoto', 'ee': 'Lɛsoto nutome', 'el': 'Λεσότο', 'el-polyton': 'Λεσότο', 'en': 'Lesotho', 'en-Dsrt': '𐐢𐐲𐑅𐐬𐑃𐐬', 'eo': 'Lesoto', 'es': 'Lesoto', 'et': 'Lesotho', 'eu': 'Lesotho', 'ewo': 'Ləsotó', 'fa': 'لسوتو', 'ff': 'Lesoto', 'ff-Adlm': '𞤂𞤫𞤧𞤮𞤼𞤮𞥅', 'ff-Latn': 'Lesoto', 'fi': 'Lesotho', 'fil': 'Lesotho', 'fo': 'Lesoto', 'fr': 'Lesotho', 'fur': 'Lesotho', 'fy': 'Lesotho', 'ga': 'Leosóta', 'gd': 'Leasoto', 'gl': 'Lesotho', 'gsw': 'Lesooto', 'gu': 'લેસોથો', 'guz': 'Lesoto', 'ha': 'Lesoto', 'he': 'לסוטו', 'hi': 'लेसोथो', 'hi-Latn': 'Lesotho', 'hr': 'Lesoto', 'hsb': 'Lesotho', 'ht': 'Lesotho', 'hu': 'Lesotho', 'hy': 'Լեսոտո', 'ia': 'Lesotho', 'id': 'Lesotho', 'ig': 'Lesotho', 'is': 'Lesótó', 'it': 'Lesotho', 'ja': 'レソト', 'jgo': 'Lɛsɔ́tɔ', 'jmc': 'Lesoto', 'jv': 'Lésotho', 'ka': 'ლესოთო', 'kaa': 'Лесото', 'kaa-Cyrl': 'Лесото', 'kab': 'Lizuṭu', 'kam': 'Lesoto', 'kde': 'Lesoto', 'kea': 'Lezotu', 'kgp': 'Resotu', 'khq': 'Leesoto', 'ki': 'Lesoto', 'kk': 'Лесото', 'kk-Cyrl': 'Лесото', 'kln': 'Emetab Lesotho', 'km': 'ឡេសូតូ', 'kn': 'ಲೆಸೊಥೊ', 'ko': '레소토', 'kok': 'लिसोथो', 'kok-Deva': 'लिसोथो', 'ks': 'لیسوتھو', 'ks-Arab': 'لیسوتھو', 'ksb': 'Lesoto', 'ksf': 'lǝsóto', 'ksh': 'Lesooto', 'ku': 'Lesoto', 'kxv': 'lesotho', 'kxv-Deva': 'लेसोतो', 'kxv-Latn': 'lesotho', 'kxv-Orya': 'ଲେସତ', 'kxv-Telu': 'లెసోతో', 'ky': 'Лесото', 'lag': 'Lesóoto', 'lb': 'Lesotho', 'lg': 'Lesoso', 'lld': 'Lesotho', 'ln': 'Lesoto', 'lo': 'ເລໂຊໂທ', 'lt': 'Lesotas', 'lu': 'Lesoto', 'luo': 'Lesotho', 'luy': 'Lesoto', 'lv': 'Lesoto', 'mai': 'लेसोथो', 'mas': 'Lesoto', 'mer': 'Lesotho', 'mfe': 'Lezoto', 'mg': 'Lesotho', 'mgh': 'Ulesoto', 'mi': 'Teroto', 'mk': 'Лесото', 'ml': 'ലെസോതോ', 'mn': 'Лесото', 'mr': 'लेसोथो', 'ms': 'Lesotho', 'mt': 'il-Lesoto', 'mua': 'Sǝr Lesotho', 'my': 'လီဆိုသို', 'mzn': 'لسوتو', 'naq': 'Lesothob', 'nb': 'Lesotho', 'nd': 'Lesotho', 'ne': 'लेसोथो', 'nl': 'Lesotho', 'nmg': 'Lesoto', 'nn': 'Lesotho', 'no': 'Lesotho', 'nqo': 'ߟߋߛߕߏ߫', 'nyn': 'Lesotho', 'om': 'Leseettoo', 'or': 'ଲେସୋଥୋ', 'pa': 'ਲੇਸੋਥੋ', 'pa-Guru': 'ਲੇਸੋਥੋ', 'pcm': 'Lẹsóto', 'pl': 'Lesotho', 'ps': 'لسوتو', 'pt': 'Lesoto', 'qu': 'Lesoto', 'rm': 'Lesotho', 'rn': 'Lesoto', 'ro': 'Lesotho', 'rof': 'Lesoto', 'ru': 'Лесото', 'rwk': 'Lesoto', 'saq': 'Lesoto', 'sbp': 'Lesoto', 'sc': 'Lesotho', 'sd': 'ليسوٿو', 'sd-Arab': 'ليسوٿو', 'se': 'Lesotho', 'seh': 'Lesoto', 'ses': 'Leesoto', 'sg': 'Lesôtho', 'shi': 'ⵍⵉⵚⵓⵟⵓ', 'shi-Latn': 'liṣuṭu', 'shi-Tfng': 'ⵍⵉⵚⵓⵟⵓ', 'si': 'ලෙසතෝ', 'sk': 'Lesotho', 'sl': 'Lesoto', 'smn': 'Lesotho', 'sn': 'Lesotho', 'so': 'Losooto', 'sq': 'Lesoto', 'sr': 'Лесото', 'sr-Cyrl': 'Лесото', 'sr-Latn': 'Lesoto', 'st': 'Lesotho', 'sv': 'Lesotho', 'sw': 'Lesoto', 'syr': 'ܠܣܘܛܘ', 'ta': 'லெசோதோ', 'te': 'లెసోతో', 'teo': 'Lesoto', 'tg': 'Лесото', 'th': 'เลโซโท', 'ti': 'ሌሶቶ', 'tk': 'Lesoto', 'to': 'Lesoto', 'tr': 'Lesotho', 'tt': 'Лесото', 'twq': 'Leesoto', 'tzm': 'Lisuṭu', 'ug': 'لېسوتو', 'uk': 'Лесото', 'ur': 'لیسوتھو', 'uz': 'Lesoto', 'uz-Cyrl': 'Лесото', 'uz-Latn': 'Lesoto', 'vai': 'ꔷꖇꕿ', 'vai-Latn': 'Lisóto', 'vai-Vaii': 'ꔷꖇꕿ', 'vec': 'Lezoto', 'vi': 'Lesotho', 'vun': 'Lesoto', 'wae': 'Lesotho', 'wo': 'Lesoto', 'xh': 'ELesotho', 'xnr': 'लेसोथो', 'xog': 'Lesoso', 'yav': 'lesotó', 'yi': 'לעסאטא', 'yo': 'Lesoto', 'yrl': 'Resutu', 'yue': '賴索托', 'yue-Hans': '赖索托', 'yue-Hant': '賴索托', 'zgh': 'ⵍⵉⵚⵓⵟⵓ', 'zh': '莱索托', 'zh-Hans': '莱索托', 'zh-Hant': '賴索托', 'zu': 'iLesotho'},
    'LT': {'af': 'Litaue', 'agq': 'Lètwǎnyìa', 'ak': 'Lituwenia', 'am': 'ሊቱዌኒያ', 'ar': 'ليتوانيا', 'as': 'লিথুৱানিয়া', 'asa': 'Litwania', 'ast': 'Lituania', 'az': 'Litva', 'az-Cyrl': 'Литва', 'az-Latn': 'Litva', 'bal-Latn': 'Lituániá', 'bas': 'Lìtùanìà', 'be': 'Літва', 'be-tarask': 'Літва', 'bez': 'Hulitwania', 'bg': 'Литва', 'blo': 'Lituwaniya', 'bm': 'Lituyani', 'bn': 'লিথুয়ানিয়া', 'br': 'Lituania', 'brx': 'लिथुआनिया', 'bs': 'Litvanija', 'bs-Cyrl': 'Литванија', 'bs-Latn': 'Litvanija', 'ca': 'Lituània', 'ca-ES-valencia': 'Lituània', 'ccp': '𑄣𑄨𑄗𑄪𑄠𑄚𑄨𑄠', 'ce': 'Литва', 'ceb': 'Lithuania', 'cgg': 'Lithuania', 'chr': 'ᎵᏗᏪᏂᎠ', 'ckb': 'لیتوانایا', 'cs': 'Litva', 'cv': 'Литва', 'cy': 'Lithwania', 'da': 'Litauen', 'dav': 'Litwania', 'de': 'Litauen', 'dje': 'Lituaani', 'dsb': 'Litawska', 'dz': 'ལི་ཐུ་ཝེ་ནི་ཡ', 'ebu': 'Litwania', 'ee': 'Lituania nutome', 'el': 'Λιθουανία', 'el-polyton': 'Λιθουανία', 'en': 'Lithuania', 'en-Dsrt': '𐐢𐐮𐑃𐐲𐐶𐐩𐑌𐐨𐐲', 'eo': 'Litovujo', 'es': 'Lituania', 'et': 'Leedu', 'eu': 'Lituania', 'ewo': 'Lituaní', 'fa': 'لیتوانی', 'ff': 'Lituaanii', 'ff-Adlm': '𞤂𞤭𞤼𞤵𞤾𞤢', 'ff-Latn': 'Lituaanii', 'fi': 'Liettua', 'fil': 'Lithuania', 'fo': 'Litava', 'fr': 'Lituanie', 'fur': 'Lituanie', 'fy': 'Litouwen', 'ga': 'an Liotuáin', 'gd': 'An Liotuain', 'gl': 'Lituania', 'gsw': 'Littaue', 'gu': 'લિથુઆનિયા', 'guz': 'Litwania', 'ha': 'Lituweniya', 'he': 'ליטא', 'hi': 'लिथुआनिया', 'hi-Latn': 'Lithuania', 'hr': 'Litva', 'hsb': 'Litawska', 'ht': 'Lituanie', 'hu': 'Litvánia', 'hy': 'Լիտվա', 'ia': 'Lituania', 'id': 'Lituania', 'ig': 'Lithuania', 'is': 'Litháen', 'it': 'Lituania', 'ja': 'リトアニア', 'jmc': 'Litwania', 'jv': 'Litowen', 'ka': 'ლიეტუვა', 'kaa': 'Литва', 'kaa-Cyrl': 'Литва', 'kab': 'Liṭwanya', 'kam': 'Litwania', 'kde': 'Litwania', 'kea': 'Lituánia', 'kgp': 'Rituỹnĩja', 'khq': 'Lituaani', 'ki': 'Litwania', 'kk': 'Литва', 'kk-Cyrl': 'Литва', 'kln': 'Emetab Lithuania', 'km': 'លីទុយអានី', 'kn': 'ಲಿಥುವೇನಿಯಾ', 'ko': '리투아니아', 'kok': 'लिथुआनिया', 'kok-Deva': 'लिथुआनिया', 'ks': 'لِتھُوانِیا', 'ks-Arab': 'لِتھُوانِیا', 'ksb': 'Litwania', 'ksf': 'litwaní', 'ksh': 'Littaue', 'ku': 'Lîtvanya', 'kxv': 'liṭuaania', 'kxv-Deva': 'लितुआनिया', 'kxv-Latn': 'liṭuaania', 'kxv-Orya': 'ଲିତୁଆନିଆ', 'kxv-Telu': 'లీతువేనియ', 'ky': 'Литва', 'lag': 'Lisuánia', 'lb': 'Litauen', 'lg': 'Lisuwenya', 'lld': 'Lituania', 'ln': 'Litwani', 'lo': 'ລິທົວເນຍ', 'lt': 'Lietuva', 'lu': 'Litwani', 'luo': 'Lithuania', 'luy': 'Litwania', 'lv': 'Lietuva', 'mai': 'लिथुआनिया', 'mas': 'Litwania', 'mer': 'Lithuania', 'mfe': 'Lituani', 'mg': 'Litoania', 'mgh': 'Utwania', 'mi': 'Rituānia', 'mk': 'Литванија', 'ml': 'ലിത്വാനിയ', 'mn': 'Литва', 'mr': 'लिथुआनिया', 'ms': 'Lithuania', 'mt': 'il-Litwanja', 'mua': 'Lituaniya', 'my': 'လစ်သူဝေးနီးယား', 'mzn': 'لتونی', 'naq': 'Lithuaniab', 'nb': 'Litauen', 'nd': 'Lithuania', 'ne': 'लिथुएनिया', 'nl': 'Litouwen', 'nmg': 'Lituaniá', 'nn': 'Litauen', 'no': 'Litauen', 'nqo': 'ߟߎߕߎ߳ߦߊߣߌ߫', 'nyn': 'Lithuania', 'om': 'Lutaaniyaa', 'or': 'ଲିଥୁଆନିଆ', 'pa': 'ਲਿਥੁਆਨੀਆ', 'pa-Guru': 'ਲਿਥੁਆਨੀਆ', 'pcm': 'Lituénia', 'pl': 'Litwa', 'ps': 'لیتوانیا', 'pt': 'Lituânia', 'qu': 'Lituania', 'rm': 'Lituania', 'rn': 'Lituwaniya', 'ro': 'Lituania', 'rof': 'Litwania', 'ru': 'Литва', 'rwk': 'Litwania', 'sah': 'Литва', 'saq': 'Litwania', 'sbp': 'Litwaniya', 'sc': 'Lituània', 'sd': 'لٿونيا', 'sd-Arab': 'لٿونيا', 'se': 'Lietuva', 'seh': 'Lituânia', 'ses': 'Lituaani', 'sg': 'Lituanïi', 'shi': 'ⵍⵉⵜⵡⴰⵏⵢⴰ', 'shi-Latn': 'litwanya', 'shi-Tfng': 'ⵍⵉⵜⵡⴰⵏⵢⴰ', 'si': 'ලිතුවේනියාව', 'sk': 'Litva', 'sl': 'Litva', 'smn': 'Liettua', 'sn': 'Lithuania', 'so': 'Lituweeniya', 'sq': 'Lituani', 'sr': 'Литванија', 'sr-Cyrl': 'Литванија', 'sr-Latn': 'Litvanija', 'sv': 'Litauen', 'sw': 'Lithuania', 'syr': 'ܠܬܘܢܝܐ', 'ta': 'லிதுவேனியா', 'te': 'లిథువేనియా', 'teo': 'Litwania', 'tg': 'Литва', 'th': 'ลิทัวเนีย', 'ti': 'ሊትዌንያ', 'tk': 'Litwa', 'to': 'Lituania', 'tr': 'Litvanya', 'tt': 'Литва', 'twq': 'Lituaani', 'tzm': 'Litwanya', 'ug': 'لىتۋانىيە', 'uk': 'Литва', 'ur': 'لیتھونیا', 'uz': 'Litva', 'uz-Cyrl': 'Литва', 'uz-Latn': 'Litva', 'vai': 'ꔷꖤꔃꕇꕰ', 'vai-Latn': 'Lituweninya', 'vai-Vaii': 'ꔷꖤꔃꕇꕰ', 'vec': 'Lituania', 'vi': 'Litva', 'vun': 'Litwania', 'wae': 'Litaue', 'wo': 'Litiyani', 'xh': 'ELithuania', 'xnr': 'लिथुआनिया', 'xog': 'Lisuwenya', 'yav': 'litiyaní', 'yi': 'ליטע', 'yo': 'Lituania', 'yrl': 'Rituwãniya', 'yue': '立陶宛', 'yue-Hans': '立陶宛', 'yue-Hant': '立陶宛', 'zgh': 'ⵍⵉⵜⵡⴰⵏⵢⴰ', 'zh': '立陶宛', 'zh-Hans': '立陶宛', 'zh-Hant': '立陶宛', 'zu': 'i-Lithuania'},
    'LU': {'af': 'Luxemburg', 'agq': 'Luzɨmbùʔ', 'ak': 'Lusimbɛg', 'am': 'ሉክሰምበርግ', 'ar': 'لوكسمبورغ', 'as': 'লাক্সেমবাৰ্গ', 'asa': 'Lathembagi', 'ast': 'Luxemburgu', 'az': 'Lüksemburq', 'az-Cyrl': 'Лүксембург', 'az-Latn': 'Lüksemburq', 'bal-Latn': 'Loksembórg', 'bas': 'Lùgsàmbûr', 'be': 'Люксембург', 'be-tarask': 'Люксембург', 'bez': 'Hulasembagi', 'bg': 'Люксембург', 'blo': 'Lusɛmbuur', 'bm': 'Likisanburu', 'bn': 'লাক্সেমবার্গ', 'br': 'Luksembourg', 'brx': 'लाक्जेमबुर्ग', 'bs': 'Luksemburg', 'bs-Cyrl': 'Луксембург', 'bs-Latn': 'Luksemburg', 'ca': 'Luxemburg', 'ca-ES-valencia': 'Luxemburg', 'ccp': '𑄣𑄪𑄇𑄴𑄥𑄬𑄟𑄴𑄝𑄢𑄴𑄉𑄧', 'ce': 'Люксембург', 'ceb': 'Luxembourg', 'cgg': 'Lakizembaaga', 'chr': 'ᎸᎧᏎᏋᎩ', 'ckb': 'لوکسەمبورگ', 'cs': 'Lucembursko', 'cv': 'Люксембург', 'cy': 'Lwcsembwrg', 'da': 'Luxembourg', 'dav': 'Lasembagi', 'de': 'Luxemburg', 'dje': 'Luxembourg', 'dsb': 'Luxemburgska', 'dz': 'ལག་ཛམ་བོརྒ', 'ebu': 'Lasembagi', 'ee': 'Lazembɔg nutome', 'el': 'Λουξεμβούργο', 'el-polyton': 'Λουξεμβοῦργο', 'en': 'Luxembourg', 'en-Dsrt': '𐐢𐐲𐐿𐑅𐐲𐑋𐐺𐐲𐑉𐑀', 'eo': 'Luksemburgo', 'es': 'Luxemburgo', 'et': 'Luksemburg', 'eu': 'Luxenburgo', 'ewo': 'Lukəzambúd', 'fa': 'لوکزامبورگ', 'ff': 'Liksembuur', 'ff-Adlm': '𞤂𞤵𞤳𞤧𞤢𞤲𞤦𞤵𞥅𞤺', 'ff-Latn': 'Liksembuur', 'fi': 'Luxemburg', 'fil': 'Luxembourg', 'fo': 'Luksemborg', 'fr': 'Luxembourg', 'fur': 'Lussemburc', 'fy': 'Luxemburg', 'ga': 'Lucsamburg', 'gd': 'Lugsamburg', 'gl': 'Luxemburgo', 'gsw': 'Luxemburg', 'gu': 'લક્ઝમબર્ગ', 'guz': 'Lasembagi', 'ha': 'Lukusambur', 'he': 'לוקסמבורג', 'hi': 'लग्ज़मबर्ग', 'hi-Latn': 'Luxembourg', 'hr': 'Luksemburg', 'hsb': 'Luxemburgska', 'ht': 'Luxembourg', 'hu': 'Luxemburg', 'hy': 'Լյուքսեմբուրգ', 'ia': 'Luxemburg', 'id': 'Luksemburg', 'ie': 'Luxemburg', 'ig': 'Luxembourg', 'is': 'Lúxemborg', 'it': 'Lussemburgo', 'ja': 'ルクセンブルク', 'jmc': 'Lasembagi', 'jv': 'Luksemburg', 'ka': 'ლუქსემბურგი', 'kaa': 'Люксембург', 'kaa-Cyrl': 'Люксембург', 'kab': 'Luksamburg', 'kam': 'Lasembagi', 'kde': 'Lasembagi', 'kea': 'Luxemburgu', 'kgp': 'Rusẽmurgu', 'khq': 'Luxembourg', 'ki': 'Lasembagi', 'kk': 'Люксембург', 'kk-Cyrl': 'Люксембург', 'kln': 'Emetab Luxembourg', 'km': 'លុចសំបួ', 'kn': 'ಲಕ್ಸೆಂಬರ್ಗ್', 'ko': '룩셈부르크', 'kok': 'लक्सेमबर्ग', 'kok-Deva': 'लक्सेमबर्ग', 'ks': 'لَکسَمبٔرٕگ', 'ks-Arab': 'لَکسَمبٔرٕگ', 'ksb': 'Lasembagi', 'ksf': 'luksɛmbúr', 'ksh': 'Luxembursch', 'ku': 'Luksembûrg', 'kxv': 'lksemborg', 'kxv-Deva': 'लग्ज़मबर्ग', 'kxv-Latn': 'lksemborg', 'kxv-Orya': 'ଲଗ୍ଜମବର୍ଗ', 'kxv-Telu': 'లక్సెంబర్గ్', 'ky': 'Люксембург', 'lag': 'Lasembáagi', 'lb': 'Lëtzebuerg', 'lg': 'Lukisembaaga', 'lld': 'Lussemburgh', 'ln': 'Likisambulu', 'lo': 'ລຸກແຊມເບີກ', 'lt': 'Liuksemburgas', 'lu': 'Likisambulu', 'luo': 'Luxembourg', 'luy': 'Lasembagi', 'lv': 'Luksemburga', 'mai': 'लग्जमबर्ग', 'mas': 'Lasembagi', 'mer': 'Luxembogu', 'mfe': 'Luxembourg', 'mg': 'Lioksamboro', 'mgh': 'Usembaji', 'mi': 'Rakapuō', 'mk': 'Луксембург', 'ml': 'ലക്സംബർഗ്', 'mn': 'Люксембург', 'mr': 'लक्झेंबर्ग', 'ms': 'Luxembourg', 'mt': 'il-Lussemburgu', 'mua': 'Sǝr Luxemburg', 'my': 'လူဇင်ဘတ်', 'mzn': 'لوکزامبورگ', 'naq': 'Luxembourgi', 'nb': 'Luxemburg', 'nd': 'Luxembourg', 'ne': 'लक्जेमबर्ग', 'nl': 'Luxemburg', 'nmg': 'Luxembourg', 'nn': 'Luxembourg', 'no': 'Luxemburg', 'nqo': 'ߟߎߜ߭ߑߛߊ߲ߓߎ߯ߙ', 'nyn': 'Lakizembaaga', 'om': 'Luksembarg', 'or': 'ଲକ୍ସେମବର୍ଗ', 'pa': 'ਲਕਜ਼ਮਬਰਗ', 'pa-Guru': 'ਲਕਜ਼ਮਬਰਗ', 'pcm': 'Lọ́ksẹ́mbọg', 'pl': 'Luksemburg', 'ps': 'لوګزامبورګ', 'pt': 'Luxemburgo', 'qu': 'Luxemburgo', 'rm': 'Luxemburg', 'rn': 'Lukusamburu', 'ro': 'Luxemburg', 'rof': 'Lasembagi', 'ru': 'Люксембург', 'rwk': 'Lasembagi', 'saq': 'Lasembagi', 'sbp': 'Lasembagi', 'sc': 'Lussemburgu', 'sd': 'لگزمبرگ', 'sd-Arab': 'لگزمبرگ', 'se': 'Luxembourg', 'seh': 'Luxemburgo', 'ses': 'Luxembourg', 'sg': 'Lugzambûru', 'shi': 'ⵍⵓⴽⵙⴰⵏⴱⵓⵔⴳ', 'shi-Latn': 'luksanburg', 'shi-Tfng': 'ⵍⵓⴽⵙⴰⵏⴱⵓⵔⴳ', 'si': 'ලක්ශම්බර්ග්', 'sk': 'Luxembursko', 'sl': 'Luksemburg', 'smn': 'Luxemburg', 'sn': 'Luxembourg', 'so': 'Luksemboorg', 'sq': 'Luksemburg', 'sr': 'Луксембург', 'sr-Cyrl': 'Луксембург', 'sr-Latn': 'Luksemburg', 'sv': 'Luxemburg', 'sw': 'Luxembourg', 'syr': 'ܠܘܟܣܡܒܘܪܓ', 'ta': 'லக்ஸ்சம்பர்க்', 'te': 'లక్సెంబర్గ్', 'teo': 'Lasembagi', 'tg': 'Люксембург', 'th': 'ลักเซมเบิร์ก', 'ti': 'ሉክሰምበርግ', 'tk': 'Lýuksemburg', 'to': 'Lakisimipeki', 'tr': 'Lüksemburg', 'tt': 'Люксембург', 'twq': 'Luxembourg', 'tzm': 'Liksumburg', 'ug': 'لىيۇكسېمبۇرگ', 'uk': 'Люксембург', 'ur': 'لکسمبرگ', 'uz': 'Lyuksemburg', 'uz-Cyrl': 'Люксембург', 'uz-Latn': 'Lyuksemburg', 'vai': 'ꗏꔻꘋꗂꖺ', 'vai-Latn': 'Lusimbɔ', 'vai-Vaii': 'ꗏꔻꘋꗂꖺ', 'vec': 'Lusenburgo', 'vi': 'Luxembourg', 'vun': 'Lasembagi', 'wae': 'Luxeburg', 'wo': 'Liksàmbur', 'xh': 'ELuxembourg', 'xnr': 'लग्ज़मबर्ग', 'xog': 'Lukisembaaga', 'yav': 'liksambúul', 'yi': 'לוקסעמבורג', 'yo': 'Lusemogi', 'yrl': 'Ruxẽbugu', 'yue': '盧森堡', 'yue-Hans': '卢森堡', 'yue-Hant': '盧森堡', 'zgh': 'ⵍⵓⴽⵙⴰⵏⴱⵓⵔⴳ', 'zh': '卢森堡', 'zh-Hans': '卢森堡', 'zh-Hant': '盧森堡', 'zu': 'i-Luxembourg'},
    'LV': {'af': 'Letland', 'agq': 'Làtɨva', 'ak': 'Latvia', 'am': 'ላትቪያ', 'ar': 'لاتفيا', 'as': 'লাটভিয়া', 'asa': 'Lativia', 'ast': 'Letonia', 'az': 'Latviya', 'az-Cyrl': 'Латвија', 'az-Latn': 'Latviya', 'bal-Latn': 'Latwiá', 'bas': 'Làdvià', 'be': 'Латвія', 'be-tarask': 'Латвія', 'bez': 'Hulativia', 'bg': 'Латвия', 'blo': 'Lɛtfiya', 'bm': 'Letoni', 'bn': 'লাটভিয়া', 'br': 'Latvia', 'brx': 'लाटभीया', 'bs': 'Latvija', 'bs-Cyrl': 'Латвија', 'bs-Latn': 'Latvija', 'ca': 'Letònia', 'ca-ES-valencia': 'Letònia', 'ccp': '𑄣𑄖𑄴𑄞𑄨𑄠', 'ce': 'Латви', 'ceb': 'Latvia', 'cgg': 'Latviya', 'chr': 'ᎳᏘᏫᎠ', 'ckb': 'لاتڤیا', 'cs': 'Lotyšsko', 'cv': 'Латви', 'cy': 'Latfia', 'da': 'Letland', 'dav': 'Lativia', 'de': 'Lettland', 'dje': 'Letooni', 'dsb': 'Letiska', 'dz': 'ལཊ་བི་ཡ', 'ebu': 'Lativia', 'ee': 'Latvia nutome', 'el': 'Λετονία', 'el-polyton': 'Λετονία', 'en': 'Latvia', 'en-Dsrt': '𐐢𐐰𐐻𐑂𐐨𐐲', 'eo': 'Latvujo', 'es': 'Letonia', 'et': 'Läti', 'eu': 'Letonia', 'ewo': 'Lətoní', 'fa': 'لتونی', 'ff': 'Letonii', 'ff-Adlm': '𞤂𞤢𞤼𞤾𞤭𞤴𞤢', 'ff-Latn': 'Letonii', 'fi': 'Latvia', 'fil': 'Latvia', 'fo': 'Lettland', 'fr': 'Lettonie', 'fur': 'Letonie', 'fy': 'Letlân', 'ga': 'an Laitvia', 'gd': 'An Laitbhe', 'gl': 'Letonia', 'gsw': 'Lettland', 'gu': 'લાત્વિયા', 'guz': 'Lativia', 'ha': 'Litibiya', 'he': 'לטביה', 'hi': 'लातविया', 'hi-Latn': 'Latvia', 'hr': 'Latvija', 'hsb': 'Letiska', 'ht': 'Lettonie', 'hu': 'Lettország', 'hy': 'Լատվիա', 'ia': 'Lettonia', 'id': 'Latvia', 'ig': 'Latvia', 'is': 'Lettland', 'it': 'Lettonia', 'ja': 'ラトビア', 'jmc': 'Lativia', 'jv': 'Latvia', 'ka': 'ლატვია', 'kaa': 'Латвия', 'kaa-Cyrl': 'Латвия', 'kab': 'Latvia', 'kam': 'Lativia', 'kde': 'Lativia', 'kea': 'Letónia', 'kgp': 'Retãnĩja', 'khq': 'Letooni', 'ki': 'Lativia', 'kk': 'Латвия', 'kk-Cyrl': 'Латвия', 'kln': 'Emetab Latvia', 'km': 'ឡេតូនី', 'kn': 'ಲಾಟ್ವಿಯಾ', 'ko': '라트비아', 'kok': 'लॅटविया', 'kok-Deva': 'लॅटविया', 'ks': 'لیٹوِیا', 'ks-Arab': 'لیٹوِیا', 'ksb': 'Lativia', 'ksf': 'lɛtoní', 'ksh': 'Lätland', 'ku': 'Letonya', 'kxv': 'laṭviya', 'kxv-Deva': 'लातविया', 'kxv-Latn': 'laṭviya', 'kxv-Orya': 'ଲାତ୍\u200dୱିୟା', 'kxv-Telu': 'లాత్వియా', 'ky': 'Латвия', 'lag': 'Lativia', 'lb': 'Lettland', 'lg': 'Lativya', 'lld': 'Letonia', 'ln': 'Letoni', 'lo': 'ລັດເວຍ', 'lt': 'Latvija', 'lu': 'Letoni', 'luo': 'Latvia', 'luy': 'Lativia', 'lv': 'Latvija', 'mai': 'लातविया', 'mas': 'Lativia', 'mer': 'Lativia', 'mfe': 'Letoni', 'mg': 'Letonia', 'mgh': 'Ulativia', 'mi': 'Rāwhia', 'mk': 'Латвија', 'ml': 'ലാറ്റ്വിയ', 'mn': 'Латви', 'mr': 'लाटव्हिया', 'ms': 'Latvia', 'mt': 'il-Latvja', 'mua': 'Letoniya', 'my': 'လတ်ဗီးယား', 'mzn': 'لاتویا', 'naq': 'Latvia', 'nb': 'Latvia', 'nd': 'Latvia', 'ne': 'लाट्भिया', 'nl': 'Letland', 'nmg': 'Latvia', 'nn': 'Latvia', 'no': 'Latvia', 'nqo': 'ߟߋߕߏߣߌ߫', 'nyn': 'Latviya', 'om': 'Lativiyaa', 'or': 'ଲାଟଭିଆ', 'pa': 'ਲਾਤਵੀਆ', 'pa-Guru': 'ਲਾਤਵੀਆ', 'pcm': 'Látvia', 'pl': 'Łotwa', 'ps': 'ليتهويا', 'pt': 'Letônia', 'qu': 'Letonia', 'rm': 'Lettonia', 'rn': 'Lativa', 'ro': 'Letonia', 'rof': 'Lativia', 'ru': 'Латвия', 'rwk': 'Lativia', 'sah': 'Латвия', 'saq': 'Lativia', 'sbp': 'Lativiya', 'sc': 'Letònia', 'sd': 'لاتويا', 'sd-Arab': 'لاتويا', 'se': 'Látvia', 'seh': 'Letônia', 'ses': 'Letooni', 'sg': 'Letonùii', 'shi': 'ⵍⴰⵜⴼⵢⴰ', 'shi-Latn': 'latfya', 'shi-Tfng': 'ⵍⴰⵜⴼⵢⴰ', 'si': 'ලැට්වියාව', 'sk': 'Lotyšsko', 'sl': 'Latvija', 'smn': 'Latvia', 'sn': 'Latvia', 'so': 'Latfiya', 'sq': 'Letoni', 'sr': 'Летонија', 'sr-Cyrl': 'Летонија', 'sr-Latn': 'Letonija', 'sv': 'Lettland', 'sw': 'Latvia', 'syr': 'ܠܐܛܒܝܐ', 'ta': 'லாட்வியா', 'te': 'లాత్వియా', 'teo': 'Lativia', 'tg': 'Латвия', 'th': 'ลัตเวีย', 'ti': 'ላትቭያ', 'tk': 'Latwiýa', 'to': 'Lativia', 'tr': 'Letonya', 'tt': 'Латвия', 'twq': 'Letooni', 'tzm': 'Liṭṭunya', 'ug': 'لاتۋىيە', 'uk': 'Латвія', 'ur': 'لٹویا', 'uz': 'Latviya', 'uz-Cyrl': 'Латвия', 'uz-Latn': 'Latviya', 'vai': 'ꕞꔳꔲꕩ', 'vai-Latn': 'Lativiya', 'vai-Vaii': 'ꕞꔳꔲꕩ', 'vec': 'Letonia', 'vi': 'Latvia', 'vun': 'Lativia', 'wae': 'Lettland', 'wo': 'Letoni', 'xh': 'ELatvia', 'xnr': 'लातविया', 'xog': 'Lativya', 'yav': 'letoní', 'yi': 'לעטלאַנד', 'yo': 'Latifia', 'yrl': 'Retuniya', 'yue': '拉脫維亞', 'yue-Hans': '拉脱维亚', 'yue-Hant': '拉脫維亞', 'zgh': 'ⵍⴰⵜⴼⵢⴰ', 'zh': '拉脱维亚', 'zh-Hans': '拉脱维亚', 'zh-Hant': '拉脫維亞', 'zu': 'i-Latvia'},
    'LY': {'af': 'Libië', 'agq': 'Lebìa', 'ak': 'Libya', 'am': 'ሊቢያ', 'ar': 'ليبيا', 'as': 'লিবিয়া', 'asa': 'Libya', 'ast': 'Libia', 'az': 'Liviya', 'az-Cyrl': 'Ливија', 'az-Latn': 'Liviya', 'bal-Latn': 'Libyá', 'bas': 'Libìà', 'be': 'Лівія', 'be-tarask': 'Лівія', 'bez': 'Hulibiya', 'bg': 'Либия', 'bgn': 'لیبیا', 'blo': 'Libii', 'bm': 'Libi', 'bn': 'লিবিয়া', 'br': 'Libia', 'brx': 'लीबिया', 'bs': 'Libija', 'bs-Cyrl': 'Либија', 'bs-Latn': 'Libija', 'ca': 'Líbia', 'ca-ES-valencia': 'Líbia', 'ccp': '𑄣𑄨𑄝𑄨𑄠', 'ce': 'Ливи', 'ceb': 'Libya', 'cgg': 'Libya', 'chr': 'ᎵᏈᏯ', 'ckb': 'لیبیا', 'cs': 'Libye', 'cv': 'Ливи', 'cy': 'Libya', 'da': 'Libyen', 'dav': 'Libya', 'de': 'Libyen', 'dje': 'Liibi', 'dsb': 'Libyska', 'dz': 'ལི་བི་ཡ', 'ebu': 'Libya', 'ee': 'Libya nutome', 'el': 'Λιβύη', 'el-polyton': 'Λιβύη', 'en': 'Libya', 'en-Dsrt': '𐐢𐐮𐐺𐐨𐐲', 'eo': 'Libio', 'es': 'Libia', 'et': 'Liibüa', 'eu': 'Libia', 'ewo': 'Libí', 'fa': 'لیبی', 'ff': 'Libi', 'ff-Adlm': '𞤂𞤭𞤦𞤭𞤴𞤢𞥄', 'ff-Latn': 'Libi', 'fi': 'Libya', 'fil': 'Libya', 'fo': 'Libya', 'fr': 'Libye', 'fur': 'Libie', 'fy': 'Libië', 'ga': 'An Libia', 'gd': 'Libia', 'gl': 'Libia', 'gsw': 'Lüübie', 'gu': 'લિબિયા', 'guz': 'Libya', 'ha': 'Libiya', 'he': 'לוב', 'hi': 'लीबिया', 'hi-Latn': 'Libya', 'hr': 'Libija', 'hsb': 'Libyska', 'ht': 'Libye', 'hu': 'Líbia', 'hy': 'Լիբիա', 'ia': 'Libya', 'id': 'Libya', 'ig': 'Libia', 'is': 'Líbía', 'it': 'Libia', 'ja': 'リビア', 'jgo': 'Libî', 'jmc': 'Libya', 'jv': 'Libya', 'ka': 'ლიბია', 'kaa': 'Ливия', 'kaa-Cyrl': 'Ливия', 'kab': 'Libya', 'kam': 'Libya', 'kde': 'Libya', 'kea': 'Líbia', 'kgp': 'Rimija', 'khq': 'Liibi', 'ki': 'Libya', 'kk': 'Ливия', 'kk-Cyrl': 'Ливия', 'kln': 'Emetab Libya', 'km': 'លីប៊ី', 'kn': 'ಲಿಬಿಯಾ', 'ko': '리비아', 'kok': 'लीबिया', 'kok-Deva': 'लीबिया', 'kok-Latn': 'Libia', 'ks': 'لِبیا', 'ks-Arab': 'لِبیا', 'ksb': 'Libya', 'ksf': 'libí', 'ksh': 'Libbije', 'ku': 'Lîbya', 'kxv': 'libya', 'kxv-Deva': 'लीबिया', 'kxv-Latn': 'libya', 'kxv-Orya': 'ଲିବିୟା', 'kxv-Telu': 'లిబియా', 'ky': 'Ливия', 'lag': 'Líbia', 'lb': 'Libyen', 'lg': 'Libya', 'lld': 'Libia', 'ln': 'Libí', 'lo': 'ລິເບຍ', 'lt': 'Libija', 'lu': 'Libi', 'luo': 'Libya', 'luy': 'Libya', 'lv': 'Lībija', 'mai': 'लीबिया', 'mas': 'Libya', 'mer': 'Lĩbia', 'mfe': 'Libi', 'mg': 'Libya', 'mgh': 'Ulibya', 'mi': 'Ripia', 'mk': 'Либија', 'ml': 'ലിബിയ', 'mn': 'Ливи', 'mr': 'लिबिया', 'ms': 'Libya', 'mt': 'il-Libja', 'mua': 'Libiya', 'my': 'လစ်ဗျား', 'mzn': 'لیبی', 'naq': 'Libyab', 'nb': 'Libya', 'nd': 'Libhiya', 'ne': 'लिबिया', 'nl': 'Libië', 'nmg': 'Libya', 'nn': 'Libya', 'no': 'Libya', 'nqo': 'ߟߓߌ߫', 'nyn': 'Libya', 'om': 'Liibiyaa', 'or': 'ଲିବ୍ୟା', 'pa': 'ਲੀਬੀਆ', 'pa-Guru': 'ਲੀਬੀਆ', 'pcm': 'Líbia', 'pl': 'Libia', 'ps': 'لیبیا', 'pt': 'Líbia', 'qu': 'Libia', 'rm': 'Libia', 'rn': 'Libiya', 'ro': 'Libia', 'rof': 'Libya', 'ru': 'Ливия', 'rwk': 'Libya', 'sah': 'Лиибийэ', 'saq': 'Libya', 'sbp': 'Libiya', 'sc': 'Lìbia', 'sd': 'لبيا', 'sd-Arab': 'لبيا', 'se': 'Libya', 'seh': 'Líbia', 'ses': 'Liibi', 'sg': 'Libïi', 'shi': 'ⵍⵉⴱⵢⴰ', 'shi-Latn': 'libya', 'shi-Tfng': 'ⵍⵉⴱⵢⴰ', 'si': 'ලිබියාව', 'sk': 'Líbya', 'sl': 'Libija', 'smn': 'Libya', 'sn': 'Libya', 'so': 'Liibya', 'sq': 'Libi', 'sr': 'Либија', 'sr-Cyrl': 'Либија', 'sr-Latn': 'Libija', 'sv': 'Libyen', 'sw': 'Libya', 'syr': 'ܠܘܒܐ', 'ta': 'லிபியா', 'te': 'లిబియా', 'teo': 'Libya', 'tg': 'Либия', 'th': 'ลิเบีย', 'ti': 'ሊብያ', 'tk': 'Liwiýa', 'to': 'Līpia', 'tr': 'Libya', 'tt': 'Ливия', 'twq': 'Liibi', 'tzm': 'Libya', 'ug': 'لىۋىيە', 'uk': 'Лівія', 'ur': 'لیبیا', 'uz': 'Liviya', 'uz-Cyrl': 'Ливия', 'uz-Latn': 'Liviya', 'vai': 'ꔒꔫꕩ', 'vai-Latn': 'Lebhiya', 'vai-Vaii': 'ꔒꔫꕩ', 'vec': 'Libia', 'vi': 'Libya', 'vun': 'Libya', 'wae': 'Lübie', 'wo': 'Libi', 'xh': 'ELibya', 'xnr': 'लीबिया', 'xog': 'Libya', 'yav': 'lipíi', 'yi': 'ליביע', 'yo': 'Libiya', 'yrl': 'Ribiya', 'yue': '利比亞', 'yue-Hans': '利比亚', 'yue-Hant': '利比亞', 'zgh': 'ⵍⵉⴱⵢⴰ', 'zh': '利比亚', 'zh-Hans': '利比亚', 'zh-Hant': '利比亞', 'zu': 'i-Libya'},
    'Lana': {'ar': 'الانا', 'ast': 'lanna', 'az': 'lanna', 'az-Latn': 'lanna', 'bal-Latn': 'Lanná', 'bg': 'Ланна', 'bn': 'লান্না', 'brx': 'लाना', 'bs': 'lanna pismo', 'bs-Cyrl': 'ланна писмо', 'bs-Latn': 'lanna pismo', 'ca': 'lanna', 'ca-ES-valencia': 'lanna', 'ccp': '𑄣𑄚𑄳𑄦', 'cs': 'lanna', 'da': 'lanna', 'de': 'Lanna', 'el': 'Λάννα', 'el-polyton': 'Λάννα', 'en': 'Lanna', 'en-Dsrt': '𐐢𐐪𐑌𐐲', 'es': 'lanna', 'et': 'tai-thami', 'eu': 'lannera', 'fa': 'لانایی', 'ff-Adlm': '𞤂𞤢𞤲𞥆𞤢', 'fi': 'lanna', 'fo': 'lanna', 'fr': 'lanna', 'fy': 'Lanna', 'ga': 'Lanna', 'gd': 'Lanna', 'gsw': 'Lanna', 'gu': 'લાના', 'he': 'לאנה', 'hi': 'लाना', 'hi-Latn': 'Lanna', 'hr': 'lanna pismo', 'ht': 'lanna', 'hu': 'Lanna', 'id': 'Lanna', 'ig': 'Lanna', 'it': 'lanna', 'ja': 'ラーンナー文字', 'kgp': 'rỹnỹ', 'kk': 'ланна жазуы', 'kk-Cyrl': 'ланна жазуы', 'km': 'ឡាណា', 'kn': 'ಲಾನಾ', 'ko': '란나 문자', 'ks': 'لانا', 'ks-Arab': 'لانا', 'lb': 'Lanna', 'lo': 'ລ້ານນາ', 'mk': 'ланско', 'ml': 'ലന്ന', 'mn': 'Лана', 'mr': 'लाना', 'ms': 'Lanna', 'nb': 'lanna', 'ne': 'लान्ना', 'nl': 'Lanna', 'nn': 'lanna', 'no': 'lanna', 'nqo': 'ߟߣߊ', 'or': 'ଲାନା', 'pl': 'lanna', 'pt': 'lanna', 'rm': 'lanna', 'ro': 'lanna', 'ru': 'ланна', 'sc': 'lanna', 'so': 'Lanna', 'sr': 'ланна писмо', 'sr-Cyrl': 'ланна писмо', 'sr-Latn': 'lanna pismo', 'sv': 'lanna', 'ta': 'லன்னா', 'te': 'లన్నా', 'tg': 'Ланна', 'th': 'ล้านนา', 'to': 'tohinima fakalana', 'tr': 'Lanna', 'ug': 'لانناچە', 'uk': 'ланна', 'vi': 'Chữ Lanna', 'xnr': 'लाना', 'yo': 'Èdè Lana', 'yrl': 'rana', 'yue': '藍拿文', 'yue-Hans': '蓝拿文', 'yue-Hant': '藍拿文', 'zh': '兰拿文', 'zh-Hans': '兰拿文', 'zh-Hant': '藍拿文', 'zu': 'isi-Lanna'},
    'Laoo': {'af': 'Lao', 'ak': 'Lawo kasa', 'am': 'ላኦ', 'ar': 'اللاو', 'as': 'লাও', 'ast': 'laosianu', 'az': 'lao', 'az-Latn': 'lao', 'bal-Latn': 'Láu', 'be': 'лаоскае', 'be-tarask': 'лаоскае', 'bg': 'лаоска', 'bn': 'লাও', 'br': 'laosek', 'brx': 'लाओ', 'bs': 'laosko pismo', 'bs-Cyrl': 'лаоско писмо', 'bs-Latn': 'laosko pismo', 'ca': 'lao', 'ca-ES-valencia': 'lao', 'ccp': '𑄣𑄃𑄮', 'ce': 'лаоссийн', 'chr': 'ᎳᎣ', 'ckb': 'لائۆ', 'cs': 'laoské', 'cy': 'Laoaidd', 'da': 'lao', 'de': 'Laotisch', 'dsb': 'laoski', 'dz': 'ལའོ་ཡིག་གུ', 'ee': 'laogbeŋɔŋlɔ', 'el': 'Λαοτινό', 'el-polyton': 'Λαοτινό', 'en': 'Lao', 'en-Dsrt': '𐐢𐐵', 'es': 'laosiano', 'et': 'lao', 'eu': 'laostarra', 'fa': 'لائوسی', 'ff-Adlm': '𞤂𞤢𞤱𞤮𞥅', 'fi': 'laolainen', 'fil': 'Lao', 'fo': 'lao', 'fr': 'lao', 'fur': 'lao', 'fy': 'Lao', 'ga': 'Laosach', 'gd': 'Làtho', 'gl': 'laosiano', 'gsw': 'Laotisch', 'gu': 'લાઓ', 'ha': 'Yan lao', 'he': 'לאי', 'hi': 'लाओ', 'hi-Latn': 'Lao', 'hr': 'laosko pismo', 'hsb': 'laosce', 'ht': 'lao', 'hu': 'Lao', 'hy': 'լաոսական', 'ia': 'lao', 'id': 'Laos', 'ig': 'Lao', 'is': 'lao', 'it': 'lao', 'ja': 'ラオ文字', 'jv': 'Lao', 'ka': 'ლაოსური', 'kea': 'lausianu', 'kgp': 'ra’o', 'kk': 'лаос жазуы', 'kk-Cyrl': 'лаос жазуы', 'km': 'ឡាវ', 'kn': 'ಲಾವೋ', 'ko': '라오 문자', 'kok': 'लाओ', 'kok-Deva': 'लाओ', 'ks': 'لاو', 'ks-Arab': 'لاو', 'ksh': 'lahootesche Schreff', 'ku': 'laoyî', 'ky': 'Лао', 'lb': 'Laotesch', 'lo': 'ລາວ', 'lrc': 'لائو', 'lt': 'laosiečių', 'lv': 'laosiešu', 'mai': 'लाओ', 'mi': 'Rao', 'mk': 'лаошко писмо', 'ml': 'ലാവോ', 'mn': 'лаос', 'mr': 'लाओ', 'ms': 'Lao', 'my': 'လာအို', 'nb': 'laotisk', 'ne': 'लाओ', 'nl': 'Laotiaans', 'nn': 'laotisk', 'no': 'laotisk', 'nqo': 'ߟߊߏ', 'or': 'ଲାଓ', 'pa': 'ਲਾਓ', 'pa-Guru': 'ਲਾਓ', 'pcm': 'Láo', 'pl': 'laotańskie', 'ps': 'لاوو', 'pt': 'lao', 'qu': 'Lao Simi', 'rm': 'laot', 'ro': 'laoțiană', 'ru': 'лаосская', 'sc': 'laotianu', 'sd': 'لائو', 'sd-Arab': 'لائو', 'si': 'ලාඕ', 'sk': 'laoské', 'sl': 'laoški', 'so': 'Dalka Lao', 'sq': 'laosisht', 'sr': 'лаошко писмо', 'sr-Cyrl': 'лаошко писмо', 'sr-Latn': 'laoško pismo', 'sv': 'laotiska', 'sw': 'Kilaosi', 'ta': 'லாவோ', 'te': 'లావో', 'tg': 'Лао', 'th': 'ลาว', 'ti': 'ሌኦ', 'tk': 'Laos elipbiýi', 'to': 'tohinima fakalau', 'tr': 'Lao', 'ug': 'لائوسچە', 'uk': 'лаоська', 'ur': 'لاؤ', 'uz': 'laos', 'uz-Cyrl': 'Лао', 'uz-Latn': 'laos', 'vec': 'lao', 'vi': 'Chữ Lào', 'wae': 'Laotiš', 'xnr': 'लाओ', 'yo': 'Láò', 'yrl': 'rau', 'yue': '寮國文', 'yue-Hans': '寮国文', 'yue-Hant': '寮國文', 'zh': '老挝文', 'zh-Hans': '老挝文', 'zh-Hant': '寮文', 'zu': 'isi-Lao'},
    'Latf': {'ar': 'اللاتينية - متغير فراكتر', 'ast': 'fraktur llatín', 'az': 'fraktur latını', 'az-Latn': 'fraktur latını', 'bal-Latn': 'Praktur Látin', 'bg': 'Латинска фрактура', 'bn': 'ফ্রাক্টুর ল্যাটিন', 'brx': 'फ्रैक्तुर लैटिन', 'bs': 'latinica (fraktur varijanta)', 'bs-Cyrl': 'латиница (фрактур варијанта)', 'bs-Latn': 'latinica (fraktur varijanta)', 'ca': 'llatí fraktur', 'ca-ES-valencia': 'llatí fraktur', 'ccp': '𑄜𑄳𑄢𑄇𑄴𑄑𑄪𑄢𑄴 𑄣𑄳𑄠𑄑𑄨𑄚𑄴', 'cs': 'latinka - lomená', 'da': 'latinsk - frakturvariant', 'de': 'Lateinisch - Fraktur-Variante', 'el': 'Φράκτουρ Λατινικό', 'el-polyton': 'Φράκτουρ Λατινικό', 'en': 'Fraktur Latin', 'en-Dsrt': '𐐙𐑉𐐰𐐿𐐻𐐲𐑉 𐐢𐐰𐐻𐐮𐑌', 'es': 'latino fraktur', 'et': 'ladina fraktuurkiri', 'fa': 'لاتینی فراکتور', 'fi': 'latinalainen fraktuuramuunnelma', 'fr': 'latin (variante brisée)', 'fur': 'latin Fraktur', 'fy': 'Gotysk Latyn', 'gd': 'Laideann fraktur', 'gsw': 'Latiinisch - Fraktur-Variante', 'gu': 'ફ્રેકતુર લેટિન', 'hi': 'फ़्रैक्टुर लातिनी', 'hi-Latn': 'Fraktur Latin', 'hr': 'fraktur latinica', 'ht': 'latin (variante brisée)', 'hu': 'Fraktur latin', 'id': 'Latin Fraktur', 'ig': 'Fraktur Latin', 'it': 'variante fraktur del latino', 'ja': 'ラテン文字(ドイツ文字)', 'kgp': 'ratĩnh fragtur', 'kn': 'ಫ್ರಾಕ್ತರ್ ಲ್ಯಾಟಿನ್', 'ko': '독일식 로마자', 'ks': 'فرکتُر لیٹِن', 'ks-Arab': 'فرکتُر لیٹِن', 'lb': 'Laténgesch-Fraktur-Variant', 'lo': 'ລາຕິນ-ຟຮັ່ງເຕຣ', 'lt': 'fraktur lotynų', 'mk': 'фрактурна латиница', 'ml': 'ഫ്രാക്ടുർ ലാറ്റിൻ', 'mr': 'फ्रॅक्तुर लॅटिन', 'nb': 'frakturlatinsk', 'ne': 'फ्राक्टुर ल्याटिन', 'nl': 'Gotisch Latijns', 'nn': 'latinsk (frakturvariant)', 'no': 'frakturlatinsk', 'or': 'ଫ୍ରାକଥୁର୍ ଲାଟିନ୍', 'pl': 'łaciński - fraktura', 'pt': 'latim fraktur', 'rm': 'latin (scrittira gotica)', 'ro': 'latină Fraktur', 'ru': 'латинская фрактура', 'sl': 'fraktura', 'sr': 'латиница (фрактур варијанта)', 'sr-Cyrl': 'латиница (фрактур варијанта)', 'sr-Latn': 'latinica (fraktur varijanta)', 'sv': 'frakturlatin', 'ta': 'ஃப்ரக்டூர் லெத்தின்', 'te': 'ఫ్రాక్టూర్ లాటిన్', 'th': 'ลาติน - ฟรังเตอร์', 'to': 'tohinima fakalatina-falakituli', 'tr': 'Fraktur Latin', 'ug': 'فىراكتۇر لاتىنچە', 'uk': 'латинський фрактурний', 'vec': 'latin fraktur', 'vi': 'Chữ La-tinh Fraktur', 'xnr': 'फ़्रैक्टुर लातिनी', 'yrl': 'ratĩ farakitú', 'yue': '拉丁文（尖角體活字變體）', 'yue-Hans': '拉丁文（尖角体活字变体）', 'yue-Hant': '拉丁文（尖角體活字變體）', 'zh': '拉丁文（哥特式字体变体）', 'zh-Hans': '拉丁文（哥特式字体变体）', 'zh-Hant': '拉丁文（尖角體活字變體）'},
    'Latg': {'ar': 'اللاتينية - متغير غيلى', 'ast': 'gaélicu llatín', 'az': 'gael latını', 'az-Latn': 'gael latını', 'bal-Latn': 'Géli Látin', 'bg': 'Галска латинска', 'bn': 'গ্যালিক ল্যাটিন', 'br': 'latin gouezelek', 'brx': 'गैलीक लैटिन', 'bs': 'galska latinica', 'bs-Cyrl': 'галска латиница', 'bs-Latn': 'galska latinica', 'ca': 'llatí gaèlic', 'ca-ES-valencia': 'llatí gaèlic', 'ccp': '𑄉𑄳𑄠𑄣𑄨𑄇𑄴 𑄣𑄳𑄠𑄑𑄨𑄚𑄴', 'cs': 'latinka - galská', 'da': 'latinsk - gælisk variant', 'de': 'Lateinisch - Gälische Variante', 'el': 'Γαελικό Λατινικό', 'el-polyton': 'Γαελικό Λατινικό', 'en': 'Gaelic Latin', 'en-Dsrt': '𐐘𐐩𐑊𐐮𐐿 𐐢𐐰𐐻𐐮𐑌', 'es': 'latino gaélico', 'et': 'ladina gaeli', 'fa': 'لاتینی گیلی', 'fi': 'latinalainen gaelimuunnelma', 'fr': 'latin (variante gaélique)', 'fur': 'latin gaelic', 'fy': 'Gaelysk Latyn', 'ga': 'Cló Gaelach', 'gd': 'Laideann Ghàidhealach', 'gsw': 'Latiinisch - Gäälischi Variante', 'gu': 'ગૈલિક લેટિન', 'he': 'לטיני גאלי', 'hi': 'गेली लातिनी', 'hi-Latn': 'Gaelic Latin', 'hr': 'keltska latinica', 'ht': 'latin (variante gaélique)', 'hu': 'Gael latin', 'id': 'Latin Gaelik', 'ig': 'Gaelic Latin', 'it': 'variante gaelica del latino', 'ja': 'ラテン文字 (ゲール文字)', 'ka': 'გელური ლათინური', 'kgp': 'ratĩnh gajériko', 'kn': 'ಗೇಲಿಕ್ ಲ್ಯಾಟಿನ್', 'ko': '아일랜드식 로마자', 'ks': 'گیلِک لیٹَن', 'ks-Arab': 'گیلِک لیٹَن', 'lb': 'Laténgesch-Gällesch Variant', 'lo': 'ລາຕິນ-ແກລິກ', 'lt': 'gėlų lotynų', 'mk': 'гелска латиница', 'ml': 'ഗെയ്\u200cലിക് ലാറ്റിൻ', 'mr': 'गाएलिक लेटिन', 'nb': 'gælisk latinsk', 'ne': 'ग्यालिक ल्याटिन', 'nl': 'Gaelisch Latijns', 'nn': 'latinsk (gælisk variant)', 'no': 'gælisk latinsk', 'or': 'ଗାଏଲିକ୍ ଲାଟିନ୍', 'pl': 'łaciński - odmiana gaelicka', 'pt': 'latim gaélico', 'rm': 'latin (scrittira gaelica)', 'ro': 'latină gaelică', 'ru': 'гэльская латинская', 'sl': 'gelski latinični', 'sr': 'галска латиница', 'sr-Cyrl': 'галска латиница', 'sr-Latn': 'galska latinica', 'sv': 'gaeliskt latin', 'ta': 'கேலிக் லெத்தின்', 'te': 'గేలిక్ లాటిన్', 'th': 'ลาติน - แกลิก', 'to': 'tohinima fakalatina-kaeliki', 'tr': 'Gael Latin', 'ug': 'سىكوت لاتىنچە', 'uk': 'латинський гельський', 'vec': 'latin gaèlego', 'vi': 'Chữ La-tinh Xcốt-len', 'xnr': 'गेली लातिनी', 'yrl': 'ratĩ gaériku', 'yue': '拉丁文（蓋爾語變體）', 'yue-Hans': '拉丁文（盖尔语变体）', 'yue-Hant': '拉丁文（蓋爾語變體）', 'zh': '拉丁文（盖尔文变体）', 'zh-Hans': '拉丁文（盖尔文变体）', 'zh-Hant': '拉丁文（蓋爾語變體）'},
    'Latn': {'af': 'Latyn', 'ak': 'Laatin', 'am': 'ላቲን', 'ar': 'اللاتينية', 'as': 'লেটিন', 'ast': 'llatín', 'az': 'latın', 'az-Arab': 'لاتینجه', 'az-Latn': 'latın', 'bal': 'لاتینی', 'bal-Arab': 'لاتینی', 'bal-Latn': 'Látin Engrézi (Látin) Engrézi (Látin, Amrikáay Tepákén Están) syáhag: Látini', 'be': 'лацініца', 'be-tarask': 'лацініца', 'bg': 'латиница', 'bgn': 'لاتین', 'blo': 'latɛŋ kʊja', 'bn': 'ল্যাটিন', 'br': 'latin', 'brx': 'लेटीन', 'bs': 'latinica', 'bs-Cyrl': 'латиница', 'bs-Latn': 'latinica', 'bss': 'Akoose', 'ca': 'llatí', 'ca-ES-valencia': 'llatí', 'ccp': '𑄣𑄳𑄠𑄑𑄨𑄚𑄴', 'ce': 'латинан', 'ceb': 'Latin', 'cho': 'Latin', 'chr': 'ᎳᏘᏂ', 'ckb': 'لاتینی', 'cs': 'latinka', 'csw': 'ᐋᑲᓈᓰᒨᐏᐣ', 'cv': 'латин', 'cy': 'Lladin', 'da': 'latinsk', 'de': 'Lateinisch', 'doi': 'लैटिन', 'dsb': 'łatyński', 'dz': 'ལེ་ཊིན་ཡིག་གུ', 'ee': 'latingbeŋɔŋlɔ', 'el': 'Λατινικό', 'el-polyton': 'Λατινικό', 'en': 'Latin', 'en-Dsrt': '𐐢𐐰𐐻𐐮𐑌', 'eo': 'latina', 'es': 'latino', 'et': 'ladina', 'eu': 'latinoa', 'fa': 'لاتین', 'ff-Adlm': '𞤂𞤢𞤼𞤫𞤲', 'fi': 'latinalainen', 'fil': 'Latin', 'fo': 'latínskt', 'fr': 'latin', 'fur': 'latin', 'fy': 'Latyn', 'ga': 'Laidineach', 'gaa': 'Niŋmaa Ni Asharaa Yiteŋ', 'gd': 'Laideann', 'gl': 'latino', 'gsw': 'Latiinisch', 'gu': 'લેટિન', 'ha': 'Latin', 'he': 'לטיני', 'hi': 'लैटिन', 'hi-Latn': 'Latin', 'hr': 'latinica', 'hsb': 'łaćonsce', 'ht': 'latin', 'hu': 'Latin', 'hy': 'լատինական', 'ia': 'latin', 'id': 'Latin', 'ie': 'latin', 'ig': 'Latin', 'ii': 'ꇁꄂꁱꂷ', 'is': 'latneskt', 'it': 'latino', 'ja': 'ラテン文字', 'jgo': 'mík -ŋwaꞌnɛ yi ɛ́ líŋɛ́nɛ Latɛ̂ŋ', 'jv': 'Latin', 'ka': 'ლათინური', 'kaa': 'латын жазыўы', 'kaa-Cyrl': 'латын жазыўы', 'kea': 'latinu', 'kgp': 'ratĩnh', 'kk': 'латын жазуы', 'kk-Cyrl': 'латын жазуы', 'km': 'ឡាតាំង', 'kn': 'ಲ್ಯಾಟಿನ್', 'ko': '로마자', 'kok': 'लॅटीन', 'kok-Deva': 'लॅटीन', 'kok-Latn': 'Romi', 'ks': 'لاطیٖنی', 'ks-Arab': 'لاطیٖنی', 'ks-Deva': 'लातिनी', 'ksh': 'lateinesche Schreff', 'ku': 'latînî', 'kxv': 'laṭin', 'kxv-Deva': 'लातिन', 'kxv-Latn': 'laṭin', 'kxv-Orya': 'ଲାଟିନ୍', 'kxv-Telu': 'లాటిన్', 'ky': 'Латын', 'lb': 'Laténgesch', 'lij': 'latin', 'lld': 'latin', 'lmo': 'Latin', 'lo': 'ລາຕິນ', 'lrc': 'لاتین', 'lt': 'lotynų', 'lv': 'latīņu', 'mai': 'लातिन', 'mgo': 'ngam ŋwaʼri', 'mi': 'Rātini', 'mk': 'латинично писмо', 'ml': 'ലാറ്റിൻ', 'mn': 'латин', 'mni': 'লেটিন', 'mni-Beng': 'লেটিন', 'mr': 'लॅटिन', 'ms': 'Latin', 'mt': 'Latin', 'my': 'လက်တင်', 'nb': 'latinsk', 'nds': 'Latiensch', 'ne': 'ल्याटिन', 'nl': 'Latijns', 'nn': 'latinsk', 'no': 'latinsk', 'nqo': 'ߟߊ߬ߕߍ߲', 'nso': 'SeLatina', 'oc': 'latin', 'om': 'Laatinii', 'or': 'ଲାଟିନ୍', 'os': 'Латинаг', 'pa': 'ਲਾਤੀਨੀ', 'pa-Guru': 'ਲਾਤੀਨੀ', 'pcm': 'Látin', 'pis': 'Latin', 'pl': 'łacińskie', 'ps': 'لاتين/لاتيني', 'pt': 'latim', 'qu': 'Latin Simi', 'rm': 'latin', 'ro': 'latină', 'ru': 'латиница', 'rw': 'Latin', 'sa': 'लैटिन:', 'sah': 'Латыынныы', 'sat': 'ᱞᱮᱴᱤᱱ', 'sat-Olck': 'ᱞᱮᱴᱤᱱ', 'sc': 'latinu', 'sd': 'لاطيني', 'sd-Arab': 'لاطيني', 'sd-Deva': 'लैटिन', 'se': 'láhtenaš', 'si': 'ලතින්', 'sk': 'latinka', 'sl': 'latinica', 'so': 'Laatiin', 'sq': 'latin', 'sr': 'латиница', 'sr-Cyrl': 'латиница', 'sr-Latn': 'latinica', 'st': 'Selatine', 'su': 'Latin', 'su-Latn': 'Latin', 'sv': 'latinska', 'sw': 'Kilatini', 'szl': 'łacińske', 'ta': 'லத்தின்', 'te': 'లాటిన్', 'tg': 'Лотинӣ', 'th': 'ละติน', 'ti': 'ላቲን', 'tk': 'Latyn elipbiýi', 'tn': 'Selatine', 'to': 'tohinima fakalatina', 'tok': 'sitelen Lasina', 'tpi': 'Latin', 'tr': 'Latin', 'tt': 'латин', 'ug': 'لاتىنچە', 'uk': 'латиниця', 'ur': 'لاطینی', 'uz': 'lotin', 'uz-Cyrl': 'Лотин', 'uz-Latn': 'lotin', 'vec': 'latin', 'vi': 'Chữ La tinh', 'vmw': 'latim', 'wae': 'Latiniš', 'wo': 'Latin', 'xh': 'IsiLatin', 'xnr': 'लैटिन', 'yi': 'גַלחיש', 'yo': 'Èdè Látìn', 'yrl': 'ratĩ', 'yue': '拉丁文', 'yue-Hans': '拉丁文', 'yue-Hant': '拉丁文', 'za': 'Lahdinghvwnz', 'zh': '拉丁文', 'zh-Hans': '拉丁文', 'zh-Hant': '拉丁字母', 'zu': 'isi-Latin'},
    'Leke': {'en': 'Leke'},
    'Lepc': {'ar': 'الليبتشا - رونج', 'ast': 'lepcha', 'az': 'lepçə', 'az-Latn': 'lepçə', 'bal-Latn': 'Lepchá', 'bg': 'Лепча', 'bn': 'লেপ্চা', 'brx': 'लेपचा', 'bs': 'lepča pismo', 'bs-Cyrl': 'лепча писмо', 'bs-Latn': 'lepča pismo', 'ca': 'lepcha', 'ca-ES-valencia': 'lepcha', 'ccp': '𑄣𑄬𑄛𑄴𑄌', 'cs': 'lepčské', 'da': 'lepcha', 'de': 'Lepcha', 'el': 'Λέπτσα', 'el-polyton': 'Λέπτσα', 'en': 'Lepcha', 'en-Dsrt': '𐐢𐐯𐐹𐐽𐐲', 'es': 'lepcha', 'et': 'leptša', 'eu': 'leptxa', 'ff-Adlm': '𞤂𞤫𞤨𞤷𞤢', 'fi': 'lepchalainen', 'fo': 'lepcha', 'fr': 'lepcha', 'fy': 'Lepcha', 'ga': 'Lepcha', 'gd': 'Lepcha', 'gsw': 'Lepcha', 'gu': 'લેપચા', 'he': 'לפחה', 'hi': 'लेपचा', 'hi-Latn': 'Lepcha', 'hr': 'lepcha pismo', 'ht': 'lepcha', 'hu': 'Lepcha', 'id': 'Lepcha', 'ig': 'Lepcha', 'it': 'lepcha', 'ja': 'レプチャ文字', 'kgp': 'rémsa', 'kk': 'лепча жазуы', 'kk-Cyrl': 'лепча жазуы', 'km': 'ឡេបចា', 'kn': 'ಲೆಪ್ಚಾ', 'ko': '렙차 문자', 'ks': 'لیپکا', 'ks-Arab': 'لیپکا', 'lb': 'Lepcha', 'lo': 'ເລຊາ', 'lt': 'lepča', 'mk': 'лепчанско', 'ml': 'ലെപ്ച', 'mn': 'Лепк', 'mr': 'लेपचा', 'ms': 'Lepcha', 'nb': 'lepcha', 'ne': 'लेप्चा', 'nl': 'Lepcha', 'nn': 'lepcha', 'no': 'lepcha', 'nqo': 'ߟߋߔߑߛߊ', 'or': 'ଲେପଚା', 'pl': 'lepcha', 'pt': 'lepcha', 'rm': 'lepcha', 'ro': 'lepcha', 'ru': 'лепха', 'sc': 'lepcha', 'sl': 'lepški', 'so': 'Lebja', 'sq': 'lepça', 'sr': 'лепча писмо', 'sr-Cyrl': 'лепча писмо', 'sr-Latn': 'lepča pismo', 'sv': 'rong', 'ta': 'லெப்சா', 'te': 'లేప్చా', 'tg': 'Лепча', 'th': 'เลปชา', 'to': 'tohinima fakalepasā', 'tr': 'Lepcha', 'ug': 'لەپچاچە', 'uk': 'лепча', 'vec': 'lepcha', 'vi': 'Chữ Lepcha', 'xnr': 'लेपचा', 'yo': 'Èdè Lepika', 'yrl': 'repixa', 'yue': '雷布查文', 'yue-Hans': '雷布查文', 'yue-Hant': '雷布查文', 'zh': '雷布查文', 'zh-Hans': '雷布查文', 'zh-Hant': '雷布查文', 'zu': 'isi-Lepcha'},
    'Limb': {'am': 'ሊምቡ', 'ar': 'الليمبو', 'ast': 'limbu', 'az': 'limbu', 'az-Latn': 'limbu', 'bal-Latn': 'Limbu', 'bg': 'Лимбу', 'bn': 'লিম্বু', 'brx': 'लिम्बु', 'bs': 'limbu pismo', 'bs-Cyrl': 'лимбу писмо', 'bs-Latn': 'limbu pismo', 'ca': 'limbu', 'ca-ES-valencia': 'limbu', 'ccp': '𑄣𑄨𑄟𑄴𑄝𑄪', 'cs': 'limbu', 'da': 'limbu', 'de': 'Limbu', 'el': 'Λιμπού', 'el-polyton': 'Λιμπού', 'en': 'Limbu', 'en-Dsrt': '𐐢𐐮𐑋𐐺𐐭', 'es': 'limbu', 'et': 'limbu', 'eu': 'linbuera', 'fa': 'لیمبایی', 'ff-Adlm': '𞤂𞤭𞤥𞤦𞤵', 'fi': 'limbulainen', 'fo': 'limbu', 'fr': 'limbou', 'fy': 'Limbu', 'ga': 'Liombúch', 'gd': 'Limbu', 'gsw': 'Limbu', 'gu': 'લિમ્બૂ', 'he': 'לימבו', 'hi': 'लिम्बू', 'hi-Latn': 'Limbu', 'hr': 'limbu pismo', 'ht': 'limbou', 'hu': 'Limbu', 'id': 'Limbu', 'ig': 'Limbu', 'it': 'limbu', 'ja': 'リンブ文字', 'ka': 'ლიმბუ', 'kgp': 'rĩmu', 'kk': 'лимбу жазуы', 'kk-Cyrl': 'лимбу жазуы', 'km': 'លីបប៊ូ', 'kn': 'ಲಿಂಬು', 'ko': '림부 문자', 'ks': 'لِمبوٗ', 'ks-Arab': 'لِمبوٗ', 'lb': 'Limbu', 'lo': 'ລິມບູ', 'lt': 'limbu', 'mk': 'лимбу', 'ml': 'ലിംബു', 'mn': 'Лимб', 'mr': 'लिम्बू', 'ms': 'Limbu', 'nb': 'limbu', 'ne': 'लिम्बु', 'nl': 'Limbu', 'nn': 'lumbu', 'no': 'limbu', 'nqo': 'ߟߌ߲ߓߎ', 'or': 'ଲିମ୍ବୁ', 'pl': 'limbu', 'pt': 'limbu', 'rm': 'limbu', 'ro': 'limbu', 'ru': 'лимбу', 'sc': 'limbu', 'sl': 'limbuški', 'so': 'Limbu', 'sq': 'limbu', 'sr': 'лимбу писмо', 'sr-Cyrl': 'лимбу писмо', 'sr-Latn': 'limbu pismo', 'sv': 'limbu', 'ta': 'லிம்பு', 'te': 'లింబు', 'tg': 'Лимбу', 'th': 'ลิมบู', 'to': 'tohinima fakalimipū', 'tr': 'Limbu', 'ug': 'لىمبۇچە', 'uk': 'лімбу', 'vec': 'linbu', 'vi': 'Chữ Limbu', 'xnr': 'लिम्बू', 'yo': 'Èdè Limbu', 'yrl': 'rĩbu', 'yue': '林佈文', 'yue-Hans': '林布文', 'yue-Hant': '林佈文', 'zh': '林布文', 'zh-Hans': '林布文', 'zh-Hant': '林佈文', 'zu': 'isi-Limbu'},
    'Lina': {'am': 'ሊኒያር ኤ', 'ar': 'الخطية أ', 'ast': 'llinial A', 'bal-Latn': 'Layni A', 'bg': 'Линейна А', 'bn': 'লিনিয়ার এ', 'brx': 'लीनीयर ए', 'bs': 'linearno A pismo', 'bs-Cyrl': 'линеарно А писмо', 'bs-Latn': 'linearno A pismo', 'ca': 'lineal A', 'ca-ES-valencia': 'lineal A', 'ccp': '𑄣𑄨𑄚𑄨𑄠𑄢𑄴 𑄆', 'cs': 'lineární A', 'da': 'lineær A', 'de': 'Linear A', 'el': 'Γραμμικό Α', 'el-polyton': 'Γραμμικό Α', 'en': 'Linear A', 'en-Dsrt': '𐐢𐐮𐑌𐐨𐐲𐑉 𐐁', 'es': 'lineal A', 'et': 'lineaarkiri A', 'eu': 'A linearra', 'fa': 'خطی الف', 'ff-Adlm': '𞤊𞤮𞤷𞥆𞤭𞥅𞤲𞤺𞤮𞤤 𞤀', 'fi': 'lineaari-A', 'fr': 'linéaire A', 'fur': 'lineâr A', 'fy': 'Lineair A', 'ga': 'Líneach A', 'gd': 'Linear A', 'gsw': 'Linear A', 'gu': 'લીનિયર અ', 'he': 'ליניארי א', 'hi': 'लीनियर A', 'hi-Latn': 'Linear A', 'hr': 'linear A pismo', 'ht': 'linéaire A', 'hu': 'Lineáris A', 'id': 'Linear A', 'ig': 'Linear A', 'it': 'lineare A', 'ja': '線文字A', 'ka': 'A-ხაზოვანი', 'kgp': 'rĩnẽjar A', 'kk': 'A линиялық жазуы', 'kk-Cyrl': 'A линиялық жазуы', 'kn': 'ಲೀನಯರ್ ಎ', 'ko': '선형 문자(A)', 'ks': 'لیٖنیَر اے', 'ks-Arab': 'لیٖنیَر اے', 'lb': 'Linear A', 'lo': 'ລີເນຍ', 'lt': 'linijiniai A', 'lv': 'lineārā A', 'mk': 'линеарно А', 'ml': 'സമരേഖയിലുള്ള എ', 'mn': 'Зураасан A', 'mr': 'लीनियार अ', 'ms': 'Linear A', 'nb': 'lineær A', 'nl': 'Lineair A', 'nn': 'lineær A', 'no': 'lineær A', 'nqo': 'ߟߌߣߍ߯ߙ ߊ', 'or': 'ଲିନିୟର୍', 'pl': 'linearne A', 'pt': 'linear A', 'rm': 'linear A', 'ro': 'lineară A', 'ru': 'линейное письмо А', 'sc': 'lineare A', 'sk': 'lineárna A', 'sl': 'linearna pisava A', 'so': 'Nidaamka qoraalka Linear A', 'sq': 'Linear A', 'sr': 'линеарно А писмо', 'sr-Cyrl': 'линеарно А писмо', 'sr-Latn': 'linearno A pismo', 'sv': 'linjär A', 'ta': 'லினியர் ஏ', 'te': 'లినియర్ ఎ', 'tg': 'Хати А', 'th': 'ลีเนียร์เอ', 'to': 'tohinima fakalinea-A', 'tr': 'Lineer A', 'ug': 'سىزىقلىق A', 'uk': 'лінійний А', 'vec': 'linear A', 'vi': 'Chữ Linear A', 'xnr': 'रेखा आली ए', 'yrl': 'satãbika A', 'yue': '線性文字（A）', 'yue-Hans': '线性文字（A）', 'yue-Hant': '線性文字（A）', 'zh': '线形文字（A）', 'zh-Hans': '线形文字（A）', 'zh-Hant': '線性文字（A）', 'zu': 'i-Linear A'},
    'Linb': {'am': 'ሊኒያር ቢ', 'ar': 'الخطية ب', 'ast': 'llinial B', 'bal-Latn': 'Layni B', 'bg': 'Линейна Б', 'bn': 'লিনিয়ার বি', 'brx': 'लीनीयर बी', 'bs': 'linearno B pismo', 'bs-Cyrl': 'линеарно Б писмо', 'bs-Latn': 'linearno B pismo', 'ca': 'lineal B', 'ca-ES-valencia': 'lineal B', 'ccp': '𑄣𑄨𑄚𑄨𑄠𑄢𑄴 𑄝𑄨', 'cs': 'lineární B', 'da': 'lineær B', 'de': 'Linear B', 'el': 'Γραμμικό Β', 'el-polyton': 'Γραμμικό Β', 'en': 'Linear B', 'en-Dsrt': '𐐢𐐮𐑌𐐨𐐲𐑉 𐐒', 'es': 'lineal B', 'et': 'lineaarkiri B', 'eu': 'B linearra', 'fa': 'خطی ب', 'ff-Adlm': '𞤊𞤮𞤷𞥆𞤭𞥅𞤲𞤺𞤮𞤤 𞤄', 'fi': 'lineaari-B', 'fr': 'linéaire B', 'fur': 'lineâr B', 'fy': 'Lineair B', 'ga': 'Líneach B', 'gd': 'Linear B', 'gsw': 'Linear B', 'gu': 'લીનિયર બી', 'he': 'ליניארי ב', 'hi': 'लीनियर बी', 'hi-Latn': 'Linear B', 'hr': 'linear B pismo', 'ht': 'linéaire B', 'hu': 'Lineáris B', 'id': 'Linear B', 'ig': 'Linear B', 'it': 'lineare B', 'ja': '線文字B', 'ka': 'B-ხაზოვანი', 'kgp': 'rĩnẽjar B', 'kk': 'B линиялық жазуы', 'kk-Cyrl': 'B линиялық жазуы', 'kn': 'ಲೀನಯರ್ ಬಿ', 'ko': '선형 문자(B)', 'ks': 'لیٖنیَر بی', 'ks-Arab': 'لیٖنیَر بی', 'lb': 'Linear B', 'lo': 'ລີເນຍຣ', 'lt': 'linijiniai B', 'lv': 'lineārā B', 'mk': 'линеарно Б', 'ml': 'ലീനിയർ ബി', 'mn': 'Зураасан B', 'mr': 'लीनियर बी', 'ms': 'Linear B', 'nb': 'lineær B', 'nl': 'Lineair B', 'nn': 'lineær B', 'no': 'lineær B', 'nqo': 'ߟߌߣߍ߯ߙ ߓ', 'or': 'ଲିନିୟର୍ ବି', 'pl': 'linearne B', 'pt': 'linear B', 'rm': 'linear B', 'ro': 'lineară B', 'ru': 'линейное письмо Б', 'sc': 'lineare B', 'sk': 'lineárna B', 'sl': 'linearna pisava B', 'so': 'Nidaamka qoraalka Linear B', 'sq': 'Linear B', 'sr': 'линеарно Б писмо', 'sr-Cyrl': 'линеарно Б писмо', 'sr-Latn': 'linearno B pismo', 'sv': 'linjär B', 'ta': 'லினியர் பி', 'te': 'లినియర్ బి', 'tg': 'Хати Б', 'th': 'ลีเนียร์บี', 'to': 'tohinima fakalinea-P', 'tr': 'Lineer B', 'ug': 'سىزىقلىق B', 'uk': 'лінійний В', 'vec': 'linear B', 'vi': 'Chữ Linear B', 'xnr': 'रेखा आली बी', 'yrl': 'satãbika B', 'yue': '線性文字（B）', 'yue-Hans': '线性文字（B）', 'yue-Hant': '線性文字（B）', 'zh': '线形文字（B）', 'zh-Hans': '线形文字（B）', 'zh-Hant': '線性文字（B）', 'zu': 'i-Linear B'},
    'Lisu': {'ast': 'alfabetu de Fraser', 'bal-Latn': 'Prásar', 'bs': 'fraser pismo', 'bs-Cyrl': 'фрасер писмо', 'bs-Latn': 'fraser pismo', 'cs': 'Fraserovo', 'de': 'Fraser', 'en': 'Fraser', 'eu': 'fraserera', 'ff-Adlm': '𞤂𞤭𞤧𞤵', 'fi': 'Fraserin aakkoset', 'fo': 'fraser', 'fy': 'Fraser', 'ga': 'Fraser', 'gd': 'Fraser', 'he': 'פרייזר', 'hi-Latn': 'Fraser', 'hr': 'fraser pismo', 'id': 'Fraser', 'ig': 'Fraser', 'ja': 'フレイザー文字', 'kgp': 'risu', 'kk': 'фрейзер жазуы', 'kk-Cyrl': 'фрейзер жазуы', 'km': 'ហ្វ្រាសឺ', 'ko': '프레이저 문자', 'lo': 'ເຟຣເຊຮ', 'lt': 'Fraser', 'mk': 'Фрејзерово', 'mn': 'Лису', 'ms': 'Fraser', 'nb': 'fraser', 'nl': 'Fraser', 'nn': 'fraser', 'no': 'fraser', 'nqo': 'ߟߌߛߎ', 'ro': 'fraser', 'ru': 'лису', 'so': 'Wabiga Fraser', 'sq': 'fraser', 'sv': 'Fraser', 'tg': 'Фрейзер', 'th': 'เฟรเซอร์', 'to': 'tohinima fakafalāse', 'tr': 'Fraser', 'ug': 'فراسېرچە', 'uk': 'абетка Фрейзера', 'vec': 'fraser', 'vi': 'Chữ Fraser', 'yo': 'Furasa', 'yrl': 'risu', 'yue': '栗僳文', 'yue-Hans': '栗僳文', 'yue-Hant': '栗僳文', 'zh': '傈僳文', 'zh-Hans': '傈僳文', 'zh-Hant': '栗僳文', 'zu': 'isi-Fraser'},
    'Loma': {'bs': 'loma pismo', 'bs-Latn': 'loma pismo', 'hr': 'loma pismo', 'ja': 'ロマ文字', 'ka': 'ლომა', 'ko': '로마 문자', 'lo': 'ໂລມາ', 'mk': 'ломско', 'ru': 'лома', 'th': 'โลมา', 'to': 'tohinima fakaloma', 'ug': 'لوماچە', 'uk': 'лома', 'vi': 'Chữ Loma', 'yue': '洛馬文', 'yue-Hans': '洛马文', 'yue-Hant': '洛馬文', 'zh': '洛马文', 'zh-Hans': '洛马文', 'zh-Hant': '洛馬文', 'en': 'Loma'},
    'Lyci': {'ar': 'الليسية', 'ast': 'liciu', 'az': 'lusian', 'az-Latn': 'lusian', 'bal-Latn': 'Lisiái', 'bg': 'Лицийска', 'bn': 'লাইসিয়ান', 'br': 'likiek', 'brx': 'लीसीयन', 'bs': 'lisijsko pismo', 'bs-Cyrl': 'лисијско писмо', 'bs-Latn': 'lisijsko pismo', 'ca': 'lici', 'ca-ES-valencia': 'lici', 'ccp': '𑄣𑄭𑄥𑄨𑄠𑄚𑄴', 'cs': 'lýkijské', 'da': 'lykisk', 'de': 'Lykisch', 'el': 'Λυκιανικό', 'el-polyton': 'Λυκιανικό', 'en': 'Lycian', 'en-Dsrt': '𐐢𐐮𐑇𐐲𐑌', 'es': 'licio', 'et': 'lüükia', 'eu': 'liziera', 'fa': 'لسیایی', 'ff-Adlm': '𞤂𞤭𞥅𞤧𞤭𞤴𞤢𞤲', 'fi': 'lyykialainen', 'fr': 'lycien', 'fy': 'Lycysk', 'ga': 'Liciach', 'gd': 'Lycian', 'gsw': 'Lykisch', 'gu': 'લિશિયન', 'he': 'ליקי', 'hi': 'लिशियन', 'hi-Latn': 'Lycian', 'hr': 'likijsko pismo', 'ht': 'lycien', 'hu': 'Líciai', 'id': 'Lycia', 'ig': 'Lycian', 'ja': 'リキア文字', 'ka': 'ლიკიური', 'kgp': 'risijo', 'kk': 'ликий жазуы', 'kk-Cyrl': 'ликий жазуы', 'kn': 'ಲೈಸಿಯನ್', 'ko': '리키아 문자', 'ks': 'لیسِیَن', 'ks-Arab': 'لیسِیَن', 'lb': 'Lykesch', 'lo': 'ໄລເຊຍ', 'lt': 'lician', 'mk': 'ликиско', 'ml': 'ലൈസിൻ', 'mn': 'Лиси', 'mr': 'लायशियान', 'ms': 'Lycia', 'nb': 'lykisk', 'ne': 'लाइसियन', 'nl': 'Lycisch', 'nn': 'lykisk', 'no': 'lykisk', 'nqo': 'ߟߛߌ', 'or': 'ଲିଶିୟନ୍', 'pl': 'likijskie', 'pt': 'lício', 'rm': 'lichic', 'ro': 'liciană', 'ru': 'лициан', 'sc': 'lìtziu', 'sl': 'licijski', 'so': 'Lyciantii Hore', 'sq': 'licianisht', 'sr': 'лисијско писмо', 'sr-Cyrl': 'лисијско писмо', 'sr-Latn': 'lisijsko pismo', 'sv': 'lykiska', 'ta': 'லிசியன்', 'te': 'లిసియన్', 'tg': 'Ликия', 'th': 'ไลเซีย', 'to': 'tohinima fakalīsia', 'tr': 'Likya', 'ug': 'لىسىيانچە', 'uk': 'лікійський', 'vec': 'lisio', 'vi': 'Chữ Lycia', 'xnr': 'लिशियन', 'yrl': 'rísiu', 'yue': '呂西亞語', 'yue-Hans': '吕西亚语', 'yue-Hant': '呂西亞語', 'zh': '利西亚文', 'zh-Hans': '利西亚文', 'zh-Hant': '呂西亞語', 'zu': 'i-Lycian'},
    'Lydi': {'ar': 'الليدية', 'ast': 'lidiu', 'az': 'ludian', 'az-Latn': 'ludian', 'bal-Latn': 'Lidiái', 'bg': 'Лидийска', 'bn': 'লাইডিয়ান', 'br': 'lidiek', 'brx': 'लीडीयन', 'bs': 'lidijsko pismo', 'bs-Cyrl': 'лидијско писмо', 'bs-Latn': 'lidijsko pismo', 'ca': 'lidi', 'ca-ES-valencia': 'lidi', 'ccp': '𑄣𑄭𑄓𑄨𑄠𑄚𑄴', 'cs': 'lýdské', 'da': 'lydisk', 'de': 'Lydisch', 'el': 'Λυδιανικό', 'el-polyton': 'Λυδιανικό', 'en': 'Lydian', 'en-Dsrt': '𐐢𐐮𐐼𐐨𐐲𐑌', 'es': 'lidio', 'et': 'lüüdia', 'eu': 'lidiera', 'fa': 'لدیایی', 'ff-Adlm': '𞤂𞤭𞤣𞤭𞤴𞤢𞤲', 'fi': 'lyydialainen', 'fr': 'lydien', 'fy': 'Lydysk', 'ga': 'Lidiach', 'gd': 'Lydian', 'gsw': 'Lydisch', 'gu': 'લિડિયન', 'he': 'לידי', 'hi': 'लिडियन', 'hi-Latn': 'Lydian', 'hr': 'lidijsko pismo', 'ht': 'lydien', 'hu': 'Lídiai', 'id': 'Lydia', 'ig': 'Lydian', 'ja': 'リディア文字', 'ka': 'ლიდიური', 'kgp': 'rinh-jo', 'kk': 'лидий жазуы', 'kk-Cyrl': 'лидий жазуы', 'kn': 'ಲಿಡಿಯನ್', 'ko': '리디아 문자', 'ks': 'لیدِیَن', 'ks-Arab': 'لیدِیَن', 'lb': 'Lydesch', 'lo': 'ລີເດຍ', 'lt': 'lidian', 'lv': 'līdiešu', 'mk': 'лидиско', 'ml': 'ലൈഡിയൻ', 'mn': 'Лиди', 'mr': 'लायडियान', 'ms': 'Lydia', 'nb': 'lydisk', 'ne': 'लाइडियन', 'nl': 'Lydisch', 'nn': 'lydisk', 'no': 'lydisk', 'nqo': 'ߟߘߌ', 'or': 'ଲିଡିୟନ୍', 'pl': 'lidyjskie', 'pt': 'lídio', 'rm': 'lidic', 'ro': 'lidiană', 'ru': 'лидийская', 'sc': 'lìdiu', 'sl': 'lidijski', 'so': 'Lydian', 'sq': 'lidianisht', 'sr': 'лидијско писмо', 'sr-Cyrl': 'лидијско писмо', 'sr-Latn': 'lidijsko pismo', 'sv': 'lydiska', 'ta': 'லிடியன்', 'te': 'లిడియన్', 'tg': 'Лидия', 'th': 'ลีเดีย', 'to': 'tohinima fakalītia', 'tr': 'Lidya', 'ug': 'لىدىيەچە', 'uk': 'лідійський', 'vec': 'lidio', 'vi': 'Chữ Lydia', 'xnr': 'लिडियन', 'yrl': 'rídiu', 'yue': '里底亞語', 'yue-Hans': '里底亚语', 'yue-Hant': '里底亞語', 'zh': '吕底亚文', 'zh-Hans': '吕底亚文', 'zh-Hant': '里底亞語', 'zu': 'i-Lydian'},
    'MA': {'af': 'Marokko', 'agq': 'Mòlokò', 'ak': 'Moroko', 'am': 'ሞሮኮ', 'ar': 'المغرب', 'as': 'মৰক্কো', 'asa': 'Moroko', 'ast': 'Marruecos', 'az': 'Mərakeş', 'az-Cyrl': 'Мәракеш', 'az-Latn': 'Mərakeş', 'bal-Latn': 'Morókó', 'bas': 'Màrokò', 'be': 'Марока', 'be-tarask': 'Марока', 'bez': 'Humoloko', 'bg': 'Мароко', 'bgn': 'مراکو', 'blo': 'Morooko', 'bm': 'Marɔku', 'bn': 'মোরক্কো', 'br': 'Maroko', 'brx': 'मरक्क’', 'bs': 'Maroko', 'bs-Cyrl': 'Мароко', 'bs-Latn': 'Maroko', 'ca': 'Marroc', 'ca-ES-valencia': 'Marroc', 'ccp': '𑄟𑄮𑄢𑄧𑄇𑄴𑄇𑄮', 'ce': 'Марокко', 'ceb': 'Morocco', 'cgg': 'Morocco', 'chr': 'ᎼᎶᎪ', 'ckb': 'مەغریب', 'cs': 'Maroko', 'cv': 'Марокко', 'cy': 'Moroco', 'da': 'Marokko', 'dav': 'Moroko', 'de': 'Marokko', 'dje': 'Maarok', 'dsb': 'Marokko', 'dz': 'མོ་རོ་ཀོ', 'ebu': 'Moroko', 'ee': 'Moroko nutome', 'el': 'Μαρόκο', 'el-polyton': 'Μαρόκο', 'en': 'Morocco', 'en-Dsrt': '𐐣𐐲𐑉𐐪𐐿𐐬', 'eo': 'Maroko', 'es': 'Marruecos', 'et': 'Maroko', 'eu': 'Maroko', 'ewo': 'Marɔ́g', 'fa': 'مراکش', 'ff': 'Maruk', 'ff-Adlm': '𞤃𞤢𞤪𞤮𞥅𞤳', 'ff-Latn': 'Maruk', 'fi': 'Marokko', 'fil': 'Morocco', 'fo': 'Marokko', 'fr': 'Maroc', 'fur': 'Maroc', 'fy': 'Marokko', 'ga': 'Maracó', 'gd': 'Moroco', 'gl': 'Marrocos', 'gsw': 'Marokko', 'gu': 'મોરોક્કો', 'guz': 'Moroko', 'ha': 'Maroko', 'he': 'מרוקו', 'hi': 'मोरक्को', 'hi-Latn': 'Morocco', 'hr': 'Maroko', 'hsb': 'Marokko', 'ht': 'Maroc', 'hu': 'Marokkó', 'hy': 'Մարոկկո', 'ia': 'Marocco', 'id': 'Maroko', 'ig': 'Morocco', 'is': 'Marokkó', 'it': 'Marocco', 'ja': 'モロッコ', 'jgo': 'Mɔlɔ̂k', 'jmc': 'Moroko', 'jv': 'Maroko', 'ka': 'მაროკო', 'kaa': 'Марокко', 'kaa-Cyrl': 'Марокко', 'kab': 'Lmerruk', 'kam': 'Moroko', 'kde': 'Moloko', 'kea': 'Marokus', 'kgp': 'Mỹhókonh', 'khq': 'Maarok', 'ki': 'Moroko', 'kk': 'Марокко', 'kk-Cyrl': 'Марокко', 'kln': 'Emetab Morocco', 'km': 'ម៉ារ៉ុក', 'kn': 'ಮೊರಾಕ್ಕೊ', 'ko': '모로코', 'kok': 'मोरोक्को', 'kok-Deva': 'मोरोक्को', 'ks': 'موروکو', 'ks-Arab': 'موروکو', 'ksb': 'Moloko', 'ksf': 'marɔk', 'ksh': 'Marokko', 'ku': 'Fas', 'kxv': 'morkko', 'kxv-Deva': 'मोरक्को', 'kxv-Latn': 'morkko', 'kxv-Orya': 'ମରକ', 'kxv-Telu': 'మొరాకో', 'ky': 'Марокко', 'lag': 'Moróoko', 'lb': 'Marokko', 'lg': 'Moroko', 'lld': 'Maroco', 'ln': 'Marokɛ', 'lo': 'ໂມຣັອກໂຄ', 'lt': 'Marokas', 'lu': 'Maroke', 'luo': 'Morocco', 'luy': 'Moroko', 'lv': 'Maroka', 'mai': 'मोरोक्को', 'mas': 'Moroko', 'mer': 'Moroko', 'mfe': 'Marok', 'mg': 'Marôka', 'mi': 'Moroko', 'mk': 'Мароко', 'ml': 'മൊറോക്കൊ', 'mn': 'Морокко', 'mr': 'मोरोक्को', 'ms': 'Maghribi', 'mt': 'il-Marokk', 'mua': 'Marok', 'my': 'မော်ရိုကို', 'mzn': 'مراکش', 'naq': 'Morocco', 'nb': 'Marokko', 'nd': 'Morokho', 'ne': 'मोरोक्को', 'nl': 'Marokko', 'nmg': 'Marɔk', 'nn': 'Marokko', 'no': 'Marokko', 'nqo': 'ߡߊ߬ߙߐߞߎ߬', 'nyn': 'Morocco', 'om': 'Morookoo', 'or': 'ମୋରୋକ୍କୋ', 'pa': 'ਮੋਰੱਕੋ', 'pa-Guru': 'ਮੋਰੱਕੋ', 'pcm': 'Morọko', 'pl': 'Maroko', 'ps': 'مراکش', 'pt': 'Marrocos', 'qu': 'Marruecos', 'rm': 'Maroc', 'rn': 'Maroke', 'ro': 'Maroc', 'rof': 'Moroko', 'ru': 'Марокко', 'rwk': 'Moroko', 'saq': 'Moroko', 'sbp': 'Moloko', 'sc': 'Marocu', 'sd': 'مراڪش', 'sd-Arab': 'مراڪش', 'se': 'Marokko', 'seh': 'Marrocos', 'ses': 'Maarok', 'sg': 'Marôko', 'shi': 'ⵍⵎⵖⵔⵉⴱ', 'shi-Latn': 'lmɣrib', 'shi-Tfng': 'ⵍⵎⵖⵔⵉⴱ', 'si': 'මොරොක්කෝව', 'sk': 'Maroko', 'sl': 'Maroko', 'smn': 'Marokko', 'sn': 'Morocco', 'so': 'Morooko', 'sq': 'Marok', 'sr': 'Мароко', 'sr-Cyrl': 'Мароко', 'sr-Latn': 'Maroko', 'sv': 'Marocko', 'sw': 'Morocco', 'syr': 'ܡܓܪܒ', 'ta': 'மொராக்கோ', 'te': 'మొరాకో', 'teo': 'Moroko', 'tg': 'Марокаш', 'th': 'โมร็อกโก', 'ti': 'ሞሮኮ', 'tk': 'Marokko', 'to': 'Moloko', 'tr': 'Fas', 'tt': 'Марокко', 'twq': 'Maarok', 'tzm': 'Meṛṛuk', 'ug': 'ماراكەش', 'uk': 'Марокко', 'ur': 'مراکش', 'uz': 'Marokash', 'uz-Cyrl': 'Марокаш', 'uz-Latn': 'Marokash', 'vai': 'ꗞꕟꖏ', 'vai-Latn': 'Mɔroko', 'vai-Vaii': 'ꗞꕟꖏ', 'vec': 'Maroco', 'vi': 'Ma-rốc', 'vun': 'Moroko', 'wae': 'Maroko', 'wo': 'Marog', 'xh': 'EMorocco', 'xnr': 'मोरक्को', 'xog': 'Moroko', 'yav': 'malóok', 'yi': 'מאַראקא', 'yo': 'Moroko', 'yrl': 'Maruku', 'yue': '摩洛哥', 'yue-Hans': '摩洛哥', 'yue-Hant': '摩洛哥', 'zgh': 'ⵍⵎⵖⵔⵉⴱ', 'zh': '摩洛哥', 'zh-Hans': '摩洛哥', 'zh-Hant': '摩洛哥', 'zu': 'i-Morocco'},
    'MC': {'af': 'Monaco', 'agq': 'Mùnaku', 'ak': 'Monako', 'am': 'ሞናኮ', 'ar': 'موناكو', 'as': 'মোনাকো', 'asa': 'Monako', 'ast': 'Mónacu', 'az': 'Monako', 'az-Cyrl': 'Монако', 'az-Latn': 'Monako', 'bal-Latn': 'Monákó', 'bas': 'Mònakò', 'be': 'Манака', 'be-tarask': 'Манака', 'bez': 'Humonako', 'bg': 'Монако', 'blo': 'Monakoo', 'bm': 'Monako', 'bn': 'মোনাকো', 'br': 'Monaco', 'brx': 'मनाक’', 'bs': 'Monako', 'bs-Cyrl': 'Монако', 'bs-Latn': 'Monako', 'ca': 'Mònaco', 'ca-ES-valencia': 'Mònaco', 'ccp': '𑄟𑄮𑄚𑄇𑄮', 'ce': 'Монако', 'ceb': 'Monaco', 'cgg': 'Monaco', 'chr': 'ᎹᎾᎪ', 'ckb': 'مۆناکۆ', 'cs': 'Monako', 'cv': 'Монако', 'cy': 'Monaco', 'da': 'Monaco', 'dav': 'Monako', 'de': 'Monaco', 'dje': 'Monako', 'dsb': 'Monaco', 'dz': 'མོ་ན་ཀོ', 'ebu': 'Monako', 'ee': 'Monako nutome', 'el': 'Μονακό', 'el-polyton': 'Μονακό', 'en': 'Monaco', 'en-Dsrt': '𐐣𐐪𐑌𐐲𐐿𐐬', 'eo': 'Monako', 'es': 'Mónaco', 'et': 'Monaco', 'eu': 'Monako', 'ewo': 'Mɔnakó', 'fa': 'موناکو', 'ff': 'Monaakoo', 'ff-Adlm': '𞤃𞤮𞤲𞤢𞤳𞤮𞥅', 'ff-Latn': 'Monaakoo', 'fi': 'Monaco', 'fil': 'Monaco', 'fo': 'Monako', 'fr': 'Monaco', 'fur': 'Monaco', 'fy': 'Monaco', 'ga': 'Monacó', 'gd': 'Monaco', 'gl': 'Mónaco', 'gsw': 'Monaco', 'gu': 'મોનાકો', 'guz': 'Monako', 'ha': 'Monako', 'he': 'מונקו', 'hi': 'मोनाको', 'hi-Latn': 'Monaco', 'hr': 'Monako', 'hsb': 'Monaco', 'ht': 'Monaco', 'hu': 'Monaco', 'hy': 'Մոնակո', 'ia': 'Monaco', 'id': 'Monako', 'ie': 'Mónaco', 'ig': 'Monaco', 'is': 'Mónakó', 'it': 'Monaco', 'ja': 'モナコ', 'jmc': 'Monako', 'jv': 'Monako', 'ka': 'მონაკო', 'kaa': 'Монако', 'kaa-Cyrl': 'Монако', 'kab': 'Munaku', 'kam': 'Monako', 'kde': 'Monako', 'kea': 'Mónaku', 'kgp': 'Mãnỹko', 'khq': 'Monako', 'ki': 'Monako', 'kk': 'Монако', 'kk-Cyrl': 'Монако', 'kln': 'Emetab Monaco', 'km': 'ម៉ូណាកូ', 'kn': 'ಮೊನಾಕೊ', 'ko': '모나코', 'kok': 'मोनॅको', 'kok-Deva': 'मोनॅको', 'ks': 'مونیکو', 'ks-Arab': 'مونیکو', 'ksb': 'Monako', 'ksf': 'monako', 'ksh': 'Monakko', 'ku': 'Monako', 'kxv': 'monako', 'kxv-Deva': 'मोनाको', 'kxv-Latn': 'monako', 'kxv-Orya': 'ମନାକ', 'kxv-Telu': 'మొనాకో', 'ky': 'Монако', 'lag': 'Monáako', 'lb': 'Monaco', 'lg': 'Monako', 'lij': 'Monego', 'lld': 'Monaco', 'ln': 'Monako', 'lo': 'ໂມນາໂຄ', 'lt': 'Monakas', 'lu': 'Monaku', 'luo': 'Monaco', 'luy': 'Monako', 'lv': 'Monako', 'mai': 'मोनाको', 'mas': 'Monako', 'mer': 'Monako', 'mfe': 'Monako', 'mg': 'Mônakô', 'mi': 'Monāko', 'mk': 'Монако', 'ml': 'മൊണാക്കോ', 'mn': 'Монако', 'mr': 'मोनॅको', 'ms': 'Monaco', 'mt': 'Monaco', 'mua': 'Monako', 'my': 'မိုနာကို', 'mzn': 'موناکو', 'naq': 'Monaco', 'nb': 'Monaco', 'nd': 'Monakho', 'ne': 'मोनाको', 'nl': 'Monaco', 'nmg': 'Monako', 'nn': 'Monaco', 'no': 'Monaco', 'nqo': 'ߡߏߣߊߞߏ߫', 'nyn': 'Monaco', 'om': 'Moonaakoo', 'or': 'ମୋନାକୋ', 'pa': 'ਮੋਨਾਕੋ', 'pa-Guru': 'ਮੋਨਾਕੋ', 'pcm': 'Mọ́náko', 'pl': 'Monako', 'ps': 'موناکو', 'pt': 'Mônaco', 'qu': 'Mónaco', 'rm': 'Monaco', 'rn': 'Monako', 'ro': 'Monaco', 'rof': 'Monako', 'ru': 'Монако', 'rwk': 'Monako', 'saq': 'Monako', 'sbp': 'Monako', 'sc': 'Mònacu', 'sd': 'موناڪو', 'sd-Arab': 'موناڪو', 'se': 'Monaco', 'seh': 'Mônaco', 'ses': 'Monako', 'sg': 'Monaköo', 'shi': 'ⵎⵓⵏⴰⴽⵓ', 'shi-Latn': 'munaku', 'shi-Tfng': 'ⵎⵓⵏⴰⴽⵓ', 'si': 'මොනාකෝව', 'sk': 'Monako', 'sl': 'Monako', 'smn': 'Monaco', 'sn': 'Monaco', 'so': 'Moonako', 'sq': 'Monako', 'sr': 'Монако', 'sr-Cyrl': 'Монако', 'sr-Latn': 'Monako', 'sv': 'Monaco', 'sw': 'Monaco', 'syr': 'ܡܘܢܐܩܘ', 'ta': 'மொனாக்கோ', 'te': 'మొనాకో', 'teo': 'Monako', 'tg': 'Монако', 'th': 'โมนาโก', 'ti': 'ሞናኮ', 'tk': 'Monako', 'to': 'Monako', 'tr': 'Monako', 'tt': 'Монако', 'twq': 'Monako', 'tzm': 'Munaku', 'ug': 'موناكو', 'uk': 'Монако', 'ur': 'موناکو', 'uz': 'Monako', 'uz-Cyrl': 'Монако', 'uz-Latn': 'Monako', 'vai': 'ꗞꕯꖏ', 'vai-Latn': 'Mɔnako', 'vai-Vaii': 'ꗞꕯꖏ', 'vec': 'Mònaco', 'vi': 'Monaco', 'vun': 'Monako', 'wae': 'Monago', 'wo': 'Monako', 'xh': 'EMonaco', 'xnr': 'मोनाको', 'xog': 'Monako', 'yav': 'monakó', 'yi': 'מאנאַקא', 'yo': 'Monako', 'yrl': 'Mũnaku', 'yue': '摩納哥', 'yue-Hans': '摩纳哥', 'yue-Hant': '摩納哥', 'zgh': 'ⵎⵓⵏⴰⴽⵓ', 'zh': '摩纳哥', 'zh-Hans': '摩纳哥', 'zh-Hant': '摩納哥', 'zu': 'i-Monaco'},
    'MD': {'af': 'Moldowa', 'agq': 'Mòodovà', 'ak': 'Mɔldova', 'am': 'ሞልዶቫ', 'ar': 'مولدوفا', 'as': 'মোলডোভা', 'asa': 'Moldova', 'ast': 'Moldavia', 'az': 'Moldova', 'az-Cyrl': 'Молдова', 'az-Latn': 'Moldova', 'bal-Latn': 'Moldowá', 'bas': 'Moldavìà', 'be': 'Малдова', 'be-tarask': 'Малдова', 'bez': 'Humoldova', 'bg': 'Молдова', 'bgn': 'مالداویا', 'blo': 'Mɔlɖafiya', 'bm': 'Molidavi', 'bn': 'মলডোভা', 'br': 'Moldova', 'brx': 'मल्ड’भा', 'bs': 'Moldavija', 'bs-Cyrl': 'Молдавија', 'bs-Latn': 'Moldavija', 'ca': 'Moldàvia', 'ca-ES-valencia': 'Moldàvia', 'ccp': '𑄟𑄮𑄣𑄴𑄘𑄞𑄨𑄠', 'ce': 'Молдави', 'ceb': 'Moldova', 'cgg': 'Moridova', 'chr': 'ᎹᎵᏙᏫᎠ', 'ckb': 'مۆلدۆڤا', 'cs': 'Moldavsko', 'cv': 'Молдова', 'cy': 'Moldofa', 'da': 'Moldova', 'dav': 'Moldova', 'de': 'Republik Moldau', 'dje': 'Moldovi', 'dsb': 'Moldawska', 'dz': 'མོལ་དོ་བཱ', 'ebu': 'Moldova', 'ee': 'Moldova nutome', 'el': 'Μολδαβία', 'el-polyton': 'Μολδαβία, Δημοκρατία τῆς', 'en': 'Moldova', 'en-Dsrt': '𐐣𐐱𐑊𐐼𐐬𐑂𐐲', 'eo': 'Moldavujo', 'es': 'Moldavia', 'et': 'Moldova', 'eu': 'Moldavia', 'ewo': 'Molədaví', 'fa': 'مولداوی', 'ff': 'Moldawii', 'ff-Adlm': '𞤃𞤮𞤤𞤣𞤮𞤾𞤢𞥄', 'ff-Latn': 'Moldawii', 'fi': 'Moldova', 'fil': 'Moldova', 'fo': 'Moldova', 'fr': 'Moldavie', 'fur': 'Moldavie', 'fy': 'Moldavië', 'ga': 'an Mholdóiv', 'gd': 'A’ Mholdobha', 'gl': 'República Moldova', 'gsw': 'Republik Moldau', 'gu': 'મોલડોવા', 'guz': 'Moldova', 'ha': 'Maldoba', 'he': 'מולדובה', 'hi': 'मॉल्डोवा', 'hi-Latn': 'Moldova', 'hr': 'Moldavija', 'hsb': 'Moldawska', 'ht': 'Moldavie', 'hu': 'Moldova', 'hy': 'Մոլդովա', 'ia': 'Moldavia', 'id': 'Moldova', 'ig': 'Moldova', 'is': 'Moldóva', 'it': 'Moldavia', 'ja': 'モルドバ', 'jmc': 'Moldova', 'jv': 'Moldova', 'ka': 'მოლდოვა', 'kaa': 'Молдова', 'kaa-Cyrl': 'Молдова', 'kab': 'Muldabi', 'kam': 'Moldova', 'kde': 'Moldova', 'kea': 'Moldávia', 'kgp': 'Mãrnova', 'khq': 'Moldovi', 'ki': 'Moldova', 'kk': 'Молдова', 'kk-Cyrl': 'Молдова', 'kln': 'Emetab Moldova', 'km': 'ម៉ុលដាវី', 'kn': 'ಮೊಲ್ಡೋವಾ', 'ko': '몰도바', 'kok': 'माल्डोवा', 'kok-Deva': 'माल्डोवा', 'ks': 'مولڈووا', 'ks-Arab': 'مولڈووا', 'ksb': 'Moldova', 'ksf': 'mɔldaví', 'ksh': 'Moldaavije', 'ku': 'Moldova', 'kxv': 'molḍovaa', 'kxv-Deva': 'मॉल्डोवा', 'kxv-Latn': 'molḍovaa', 'kxv-Orya': 'ମଲ୍\u200dଡୱା', 'kxv-Telu': 'మోల్డొవా', 'ky': 'Молдова', 'lag': 'Molidóova', 'lb': 'Moldawien', 'lg': 'Molodova', 'lld': 'Moldavia', 'ln': 'Molidavi', 'lo': 'ໂມນໂດວາ', 'lt': 'Moldova', 'lu': 'Molidavi', 'luo': 'Moldova', 'luy': 'Moldova', 'lv': 'Moldova', 'mai': 'मोलडोवा', 'mas': 'Moldova', 'mer': 'Moldova', 'mfe': 'Moldavi', 'mg': 'Môldavia', 'mi': 'Morotawa', 'mk': 'Молдавија', 'ml': 'മൾഡോവ', 'mn': 'Молдова', 'mr': 'मोल्डोव्हा', 'ms': 'Moldova', 'mt': 'il-Moldova', 'mua': 'Moldoviya', 'my': 'မောလ်ဒိုဗာ', 'mzn': 'مولداوی', 'naq': 'Moldova', 'nb': 'Moldova', 'nd': 'Moldova', 'ne': 'माल्डोभा', 'nl': 'Moldavië', 'nmg': 'Mɔldavia', 'nn': 'Moldova', 'no': 'Moldova', 'nqo': 'ߡߐߟߑߘߊߝ߭ߌ߫', 'nyn': 'Moridova', 'om': 'Moldoovaa', 'or': 'ମୋଲଡୋଭା', 'pa': 'ਮੋਲਡੋਵਾ', 'pa-Guru': 'ਮੋਲਡੋਵਾ', 'pcm': 'Mọldóva', 'pl': 'Mołdawia', 'ps': 'مولدوا', 'pt': 'Moldávia', 'qu': 'Moldova', 'rm': 'Moldavia', 'rn': 'Moludavi', 'ro': 'Republica Moldova', 'rof': 'Moldova', 'ru': 'Молдова', 'rwk': 'Moldova', 'saq': 'Moldova', 'sbp': 'Molidova', 'sc': 'Moldàvia', 'sd': 'مالدووا', 'sd-Arab': 'مالدووا', 'se': 'Moldávia', 'seh': 'Moldávia', 'ses': 'Moldovi', 'sg': 'Moldavùii', 'shi': 'ⵎⵓⵍⴷⵓⴼⵢⴰ', 'shi-Latn': 'muldufya', 'shi-Tfng': 'ⵎⵓⵍⴷⵓⴼⵢⴰ', 'si': 'මොල්ඩෝවාව', 'sk': 'Moldavsko', 'sl': 'Moldavija', 'smn': 'Moldova', 'sn': 'Moldova', 'so': 'Moldofa', 'sq': 'Moldavi', 'sr': 'Молдавија', 'sr-Cyrl': 'Молдавија', 'sr-Latn': 'Moldavija', 'sv': 'Moldavien', 'sw': 'Moldova', 'syr': 'ܡܘܠܕܘܒܐ', 'ta': 'மால்டோவா', 'te': 'మోల్డోవా', 'teo': 'Moldova', 'tg': 'Молдова', 'th': 'มอลโดวา', 'ti': 'ሞልዶቫ', 'tk': 'Moldowa', 'to': 'Molotova', 'tr': 'Moldova', 'tt': 'Молдова', 'twq': 'Moldovi', 'tzm': 'Mulḍavya', 'ug': 'مولدوۋا', 'uk': 'Молдова', 'ur': 'مالدووا', 'uz': 'Moldova', 'uz-Cyrl': 'Молдова', 'uz-Latn': 'Moldova', 'vai': 'ꖒꔷꖁꕙ', 'vai-Latn': 'Mɔlidova', 'vai-Vaii': 'ꖒꔷꖁꕙ', 'vec': 'Moldavia', 'vi': 'Moldova', 'vun': 'Moldova', 'wae': 'Moldau', 'wo': 'Moldawi', 'xh': 'EMoldova', 'xnr': 'मॉल्डोवा', 'xog': 'Molodova', 'yav': 'moltafí', 'yi': 'מאלדאווע', 'yo': 'Modofia', 'yrl': 'Mũduwa', 'yue': '摩爾多瓦', 'yue-Hans': '摩尔多瓦', 'yue-Hant': '摩爾多瓦', 'zgh': 'ⵎⵓⵍⴷⵓⴼⵢⴰ', 'zh': '摩尔多瓦', 'zh-Hans': '摩尔多瓦', 'zh-Hant': '摩爾多瓦', 'zu': 'i-Moldova'},
    'ME': {'af': 'Montenegro', 'ak': 'Mɔntenegro', 'am': 'ሞንተኔግሮ', 'ar': 'الجبل الأسود', 'as': 'মণ্টেনেগ্ৰু', 'ast': 'Montenegru', 'az': 'Monteneqro', 'az-Cyrl': 'Монтенегро', 'az-Latn': 'Monteneqro', 'bal-Latn': 'Montenegró', 'be': 'Чарнагорыя', 'be-tarask': 'Чарнагорыя', 'bg': 'Черна гора', 'blo': 'Mɔntenegroo', 'bn': 'মন্টিনিগ্রো', 'br': 'Montenegro', 'brx': 'मन्टेनेग्र’', 'bs': 'Crna Gora', 'bs-Cyrl': 'Црна Гора', 'bs-Latn': 'Crna Gora', 'ca': 'Montenegro', 'ca-ES-valencia': 'Montenegro', 'ccp': '𑄟𑄧𑄚𑄴𑄑𑄨𑄚𑄨𑄉𑄳𑄢𑄮', 'ce': 'Ӏаьржаламанчоь', 'ceb': 'Montenegro', 'chr': 'ᎼᏂᏔᏁᎦᎶ', 'ckb': 'مۆنتینیگرۆ', 'cs': 'Černá Hora', 'cv': 'Черногори', 'cy': 'Montenegro', 'da': 'Montenegro', 'de': 'Montenegro', 'dsb': 'Carna Góra', 'dz': 'མོན་ཊི་ནེག་རོ', 'ee': 'Montenegro nutome', 'el': 'Μαυροβούνιο', 'el-polyton': 'Μαυροβούνιο', 'en': 'Montenegro', 'en-Dsrt': '𐐣𐐪𐑌𐐲𐑌𐐨𐑀𐑉𐐬', 'es': 'Montenegro', 'et': 'Montenegro', 'eu': 'Montenegro', 'fa': 'مونته\u200cنگرو', 'ff-Adlm': '𞤃𞤮𞤲𞤼𞤫𞤲𞤫𞥅𞤺𞤮𞤪𞤮', 'fi': 'Montenegro', 'fil': 'Montenegro', 'fo': 'Montenegro', 'fr': 'Monténégro', 'fur': 'Montenegro', 'fy': 'Montenegro', 'ga': 'Montainéagró', 'gd': 'Am Monadh Neagrach', 'gl': 'Montenegro', 'gsw': 'Monteneegro', 'gu': 'મૉન્ટેનેગ્રો', 'ha': 'Manteneguro', 'he': 'מונטנגרו', 'hi': 'मोंटेनेग्रो', 'hi-Latn': 'Montenegro', 'hr': 'Crna Gora', 'hsb': 'Montenegro', 'ht': 'Monténégro', 'hu': 'Montenegró', 'hy': 'Չեռնոգորիա', 'ia': 'Montenegro', 'id': 'Montenegro', 'ie': 'Montenegro', 'ig': 'Montenegro', 'is': 'Svartfjallaland', 'it': 'Montenegro', 'ja': 'モンテネグロ', 'jv': 'Montenégro', 'ka': 'მონტენეგრო', 'kaa': 'Черногория', 'kaa-Cyrl': 'Черногория', 'kea': 'Montenegru', 'kgp': 'Krĩsá', 'kk': 'Черногория', 'kk-Cyrl': 'Черногория', 'km': 'ម៉ុងតេណេហ្គ្រោ', 'kn': 'ಮೊಂಟೆನೆಗ್ರೋ', 'ko': '몬테네그로', 'kok': 'मॉन्टॅनग्रो', 'kok-Deva': 'मॉन्टॅनग्रो', 'ks': 'موٹونیگِریو', 'ks-Arab': 'موٹونیگِریو', 'ksh': 'et Monteneejro', 'ku': 'Montenegro', 'kxv': 'monṭenegro', 'kxv-Deva': 'मोंटेनेग्रो', 'kxv-Latn': 'monṭenegro', 'kxv-Orya': 'ମଣ୍ଟେନେଗ୍ର', 'kxv-Telu': 'మాంటెనెగ్రో', 'ky': 'Черногория', 'lb': 'Montenegro', 'lld': 'Montenegro', 'ln': 'Monténégro', 'lo': 'ມອນເຕເນໂກຣ', 'lt': 'Juodkalnija', 'lv': 'Melnkalne', 'mai': 'मोंटेनीग्रो', 'mgh': 'Umantegro', 'mi': 'Maungakororiko', 'mk': 'Црна Гора', 'ml': 'മോണ്ടെനെഗ്രോ', 'mn': 'Монтенегро', 'mr': 'मोंटेनेग्रो', 'ms': 'Montenegro', 'mt': 'il-Montenegro', 'my': 'မွန်တီနိဂရိုး', 'mzn': 'مونته\u200cنگرو', 'nb': 'Montenegro', 'ne': 'मोन्टेनिग्रो', 'nl': 'Montenegro', 'nn': 'Montenegro', 'no': 'Montenegro', 'nqo': 'ߡߐ߲ߕߣߋߜ߭ߙߏ߫', 'om': 'Montenegiroo', 'or': 'ମଣ୍ଟେନିଗ୍ରୋ', 'pa': 'ਮੋਂਟੇਨੇਗਰੋ', 'pa-Guru': 'ਮੋਂਟੇਨੇਗਰੋ', 'pcm': 'Mọntinígro', 'pl': 'Czarnogóra', 'ps': 'مونټینیګرو', 'pt': 'Montenegro', 'qu': 'Montenegro', 'rm': 'Montenegro', 'ro': 'Muntenegru', 'ru': 'Черногория', 'sc': 'Montenegro', 'sd': 'مونٽي نيگرو', 'sd-Arab': 'مونٽي نيگرو', 'se': 'Montenegro', 'si': 'මොන්ටෙනීග්\u200dරෝ', 'sk': 'Čierna Hora', 'sl': 'Črna gora', 'smn': 'Montenegro', 'so': 'Moontenegro', 'sq': 'Mal i Zi', 'sr': 'Црна Гора', 'sr-Cyrl': 'Црна Гора', 'sr-Latn': 'Crna Gora', 'sv': 'Montenegro', 'sw': 'Montenegro', 'syr': 'ܡܘܢܛܝܢܝܓܪܘ', 'ta': 'மான்டேனெக்ரோ', 'te': 'మాంటెనెగ్రో', 'tg': 'Черногория', 'th': 'มอนเตเนโกร', 'ti': 'ሞንተኔግሮ', 'tk': 'Çernogoriýa', 'to': 'Monitenikalo', 'tr': 'Karadağ', 'tt': 'Черногория', 'ug': 'قارا تاغ', 'uk': 'Чорногорія', 'ur': 'مونٹے نیگرو', 'uz': 'Chernogoriya', 'uz-Cyrl': 'Черногория', 'uz-Latn': 'Chernogoriya', 'vai': 'ꗞꔳꕇꖶꖄ', 'vai-Vaii': 'ꗞꔳꕇꖶꖄ', 'vec': 'Montenegro', 'vi': 'Montenegro', 'wae': 'Montenegro', 'wo': 'Montenegoro', 'xh': 'EMontenegro', 'xnr': 'मोंटेनेग्रो', 'yi': 'מאנטענעגרא', 'yo': 'Montenégrò', 'yrl': 'Mũteneguru', 'yue': '蒙特內哥羅', 'yue-Hans': '蒙特内哥罗', 'yue-Hant': '蒙特內哥羅', 'zgh': 'ⵎⵓⵏⵜⵉⵏⵉⴳⵔⵓ', 'zh': '黑山', 'zh-Hans': '黑山', 'zh-Hant': '蒙特內哥羅', 'zu': 'i-Montenegro'},
    'MF': {'af': 'Sint Martin', 'ak': 'St. Maatin', 'am': 'ሴንት ማርቲን', 'ar': 'سان مارتن', 'as': 'ছেইণ্ট মাৰ্টিন', 'ast': 'Saint Martin', 'az': 'Sent Martin', 'az-Cyrl': 'Сент Мартин', 'az-Latn': 'Sent Martin', 'bal-Latn': 'St. Mártin', 'be': 'Сен-Мартэн', 'be-tarask': 'Сен-Мартэн', 'bg': 'Сен Мартен', 'blo': 'Sɛŋ Martɛɛŋ', 'bn': 'সেন্ট মার্টিন', 'br': 'Saint Martin', 'brx': 'सैन्ट मार्तिन', 'bs': 'Sveti Martin', 'bs-Cyrl': 'Свети Мартин', 'bs-Latn': 'Sveti Martin', 'ca': 'Saint Martin', 'ca-ES-valencia': 'Saint Martin', 'ccp': '𑄥𑄬𑄚𑄴𑄑𑄴 𑄟𑄢𑄴𑄑𑄨𑄚𑄴', 'ce': 'Сен-Мартен', 'ceb': 'St. Martin', 'chr': 'ᎤᏓᏅᏘ ᏡᏡ', 'ckb': 'سەن مارتین', 'cs': 'Svatý Martin (Francie)', 'cv': 'Сен-Мартен', 'cy': 'Saint Martin', 'da': 'Saint Martin', 'de': 'St. Martin', 'dsb': 'St. Martin', 'dz': 'སེནཊ་ མཱར་ཊིན', 'ee': 'Saint Martin nutome', 'el': 'Άγιος Μαρτίνος (Γαλλικό τμήμα)', 'el-polyton': 'Άγιος Μαρτίνος (Γαλλικό τμήμα)', 'en': 'St. Martin', 'en-Dsrt': '𐐝𐐩𐑌𐐻 𐐣𐐪𐑉𐐻𐑌', 'es': 'San Martín', 'et': 'Saint-Martin', 'eu': 'San Martin', 'fa': 'سنت مارتین', 'ff-Adlm': '𞤅𞤼. 𞤃𞤢𞤪𞤼𞤫𞤲', 'fi': 'Saint-Martin', 'fil': 'Saint Martin', 'fo': 'St-Martin', 'fr': 'Saint-Martin', 'fur': 'Sant Martin', 'fy': 'Saint-Martin', 'ga': 'Saint-Martin', 'gd': 'Naomh Màrtainn', 'gl': 'Saint Martin', 'gsw': 'St. Martin', 'gu': 'સેંટ માર્ટિન', 'ha': 'San Martin', 'he': 'סן מרטן', 'hi': 'सेंट मार्टिन', 'hi-Latn': 'St. Martin', 'hr': 'Saint Martin', 'hsb': 'St. Martin', 'ht': 'Saint-Martin', 'hu': 'Saint Martin', 'hy': 'Սեն Մարտեն', 'ia': 'Sancte Martino francese', 'id': 'Saint Martin', 'ig': 'St. Martin', 'is': 'Saint-Martin', 'it': 'Saint Martin', 'ja': 'サン・マルタン', 'jv': 'Santa Martin', 'ka': 'სენ-მარტენი', 'kaa': 'Сен-Мартен', 'kaa-Cyrl': 'Сен-Мартен', 'kea': 'San Martinhu (Fransa)', 'kgp': 'Sỹ Mỹrtĩjũ', 'kk': 'Сен-Мартен', 'kk-Cyrl': 'Сен-Мартен', 'km': 'សាំង\u200bម៉ាទីន', 'kn': 'ಸೇಂಟ್ ಮಾರ್ಟಿನ್', 'ko': '생마르탱', 'kok': 'सॅंट मार्टिन', 'kok-Deva': 'सॅंट मार्टिन', 'ks': 'سینٹ مارٹِن', 'ks-Arab': 'سینٹ مارٹِن', 'ksh': 'de Zint-Määtes-Ensel', 'ku': 'Saint Martin', 'kxv': 'seint martin', 'kxv-Deva': 'सेंट मार्टिन', 'kxv-Latn': 'seint martin', 'kxv-Orya': 'ସେଣ୍ଟ ମାର୍ଟିନ', 'kxv-Telu': 'సెయింట్ మార్టిన్', 'ky': 'Сент-Мартин', 'lb': 'St. Martin', 'lld': 'St. Martin', 'lo': 'ເຊນ ມາທິນ', 'lt': 'Sen Martenas', 'lv': 'Senmartēna', 'mai': 'सेंट मार्टिन', 'mi': 'Hato Mātene', 'mk': 'Сент Мартин', 'ml': 'സെന്റ് മാർട്ടിൻ', 'mn': 'Сент-Мартин', 'mr': 'सेंट मार्टिन', 'ms': 'Saint Martin', 'mt': 'Saint Martin', 'my': 'စိန့်မာတင်', 'mzn': 'سنت مارتین', 'nb': 'Saint-Martin', 'ne': 'सेन्ट मार्टिन', 'nl': 'Saint-Martin', 'nn': 'Saint-Martin', 'no': 'Saint-Martin', 'nqo': 'ߛߍ߲ߕ-ߡߊߙߑߕߍ߲߫', 'om': 'St. Martiin', 'or': 'ସେଣ୍ଟ ମାର୍ଟିନ୍', 'pa': 'ਸੇਂਟ ਮਾਰਟਿਨ', 'pa-Guru': 'ਸੇਂਟ ਮਾਰਟਿਨ', 'pcm': 'Sent Mátin', 'pl': 'Saint-Martin', 'ps': 'سینټ مارټن', 'pt': 'São Martinho', 'qu': 'San Martín', 'rm': 'Saint Martin', 'ro': 'Sfântul Martin', 'ru': 'Сен-Мартен', 'sc': 'Santu Martine', 'sd': 'سينٽ مارٽن', 'sd-Arab': 'سينٽ مارٽن', 'se': 'Frankriikka Saint Martin', 'si': 'ශාන්ත මාර්ටින්', 'sk': 'Svätý Martin (fr.)', 'sl': 'Saint Martin', 'smn': 'St. Martin', 'so': 'St. Maartin', 'sq': 'Sen-Marten', 'sr': 'Свети Мартин (Француска)', 'sr-Cyrl': 'Свети Мартин (Француска)', 'sr-Latn': 'Sveti Martin (Francuska)', 'sv': 'Saint-Martin', 'sw': 'St. Martin', 'syr': 'ܣܐܢܬ ܡܐܪܬܝܢ', 'ta': 'செயின்ட் மார்ட்டீன்', 'te': 'సెయింట్ మార్టిన్', 'tg': 'Ҷазираи Сент-Мартин', 'th': 'เซนต์มาร์ติน', 'ti': 'ቅዱስ ማርቲን', 'tk': 'Sen-Marten', 'to': 'Sā Mātini (fakafalanisē)', 'tr': 'Saint Martin', 'tt': 'Сент-Мартин', 'ug': 'ساينت مارتىن', 'uk': 'Сен-Мартен', 'ur': 'سینٹ مارٹن', 'uz': 'Sent-Martin', 'uz-Cyrl': 'Сент-Мартин', 'uz-Latn': 'Sent-Martin', 'vai': 'ꕪꘋꕓ ꗞꗢ ꕮꕊꔳꘋ', 'vai-Vaii': 'ꕪꘋꕓ ꗞꗢ ꕮꕊꔳꘋ', 'vec': 'S. Martin (Fransa)', 'vi': 'St. Martin', 'wae': 'St. Martin', 'wo': 'Saŋ Marteŋ', 'xh': 'ESt. Martin', 'xnr': 'सेंट मार्टिन', 'yo': 'Ìlú Màtìnì', 'yrl': 'Sã Matiyũ', 'yue': '法屬聖馬丁', 'yue-Hans': '法属圣马丁', 'yue-Hant': '法屬聖馬丁', 'zh': '法属圣马丁', 'zh-Hans': '法属圣马丁', 'zh-Hant': '法屬聖馬丁', 'zu': 'i-Saint Martin'},
    'MG': {'af': 'Madagaskar', 'agq': 'Màdàgasɨkà', 'ak': 'Madagaska', 'am': 'ማዳጋስካር', 'ar': 'مدغشقر', 'as': 'মাদাগাস্কাৰ', 'asa': 'Bukini', 'ast': 'Madagascar', 'az': 'Madaqaskar', 'az-Cyrl': 'Мадагаскар', 'az-Latn': 'Madaqaskar', 'bal-Latn': 'Madagáskar', 'bas': 'Màdàgàskâr', 'be': 'Мадагаскар', 'be-tarask': 'Мадагаскар', 'bez': 'Hubukini', 'bg': 'Мадагаскар', 'bgn': 'ماداگاسکار', 'blo': 'Maɖagaskaa', 'bm': 'Madagasikari', 'bn': 'মাদাগাস্কার', 'br': 'Madagaskar', 'brx': 'मादागास्कार', 'bs': 'Madagaskar', 'bs-Cyrl': 'Мадагаскар', 'bs-Latn': 'Madagaskar', 'ca': 'Madagascar', 'ca-ES-valencia': 'Madagascar', 'ccp': '𑄟𑄘𑄉𑄌𑄴𑄇𑄢𑄴', 'ce': 'Мадагаскар', 'ceb': 'Madagascar', 'cgg': 'Madagasika', 'chr': 'ᎹᏓᎦᏍᎧᎵ', 'ckb': 'ماداگاسکار', 'cs': 'Madagaskar', 'cv': 'Мадагаскар', 'cy': 'Madagascar', 'da': 'Madagaskar', 'dav': 'Bukini', 'de': 'Madagaskar', 'dje': 'Madagascar', 'dsb': 'Madagaskar', 'dyo': 'Madagaskaar', 'dz': 'མ་དཱ་གེས་ཀར', 'ebu': 'Bukini', 'ee': 'Madagaska nutome', 'el': 'Μαδαγασκάρη', 'el-polyton': 'Μαδαγασκάρη', 'en': 'Madagascar', 'en-Dsrt': '𐐣𐐰𐐼𐐲𐑀𐐰𐑅𐐿𐐲𐑉', 'eo': 'Madagaskaro', 'es': 'Madagascar', 'et': 'Madagaskar', 'eu': 'Madagaskar', 'ewo': 'Madagasəkárə', 'fa': 'ماداگاسکار', 'ff': 'Madagaskaar', 'ff-Adlm': '𞤃𞤢𞤣𞤢𞤺𞤢𞤧𞤳𞤢𞥄𞤪', 'ff-Latn': 'Madagaskaar', 'fi': 'Madagaskar', 'fil': 'Madagascar', 'fo': 'Madagaskar', 'fr': 'Madagascar', 'fur': 'Madagascar', 'fy': 'Madeiaskar', 'ga': 'Madagascar', 'gd': 'Madagasgar', 'gl': 'Madagascar', 'gsw': 'Madagaschkar', 'gu': 'મેડાગાસ્કર', 'guz': 'Bukini', 'ha': 'Madagaskar', 'he': 'מדגסקר', 'hi': 'मेडागास्कर', 'hi-Latn': 'Madagascar', 'hr': 'Madagaskar', 'hsb': 'Madagaskar', 'ht': 'Madagascar', 'hu': 'Madagaszkár', 'hy': 'Մադագասկար', 'ia': 'Madagascar', 'id': 'Madagaskar', 'ig': 'Madagascar', 'is': 'Madagaskar', 'it': 'Madagascar', 'ja': 'マダガスカル', 'jgo': 'Mándaŋgasɛkâ', 'jmc': 'Bukini', 'jv': 'Madagaskar', 'ka': 'მადაგასკარი', 'kaa': 'Мадагаскар', 'kaa-Cyrl': 'Мадагаскар', 'kab': 'Madaɣecqer', 'kam': 'Bukini', 'kde': 'Bukini', 'kea': 'Madagaskar', 'kgp': 'Mỹnaganhtar', 'khq': 'Madagascar', 'ki': 'Bukini', 'kk': 'Мадагаскар', 'kk-Cyrl': 'Мадагаскар', 'kln': 'Emetab Madagascar', 'km': 'ម៉ាដាហ្គាស្កា', 'kn': 'ಮಡಗಾಸ್ಕರ್', 'ko': '마다가스카르', 'kok': 'माडागास्कर', 'kok-Deva': 'माडागास्कर', 'ks': 'میڈاگاسکار', 'ks-Arab': 'میڈاگاسکار', 'ksb': 'Bukini', 'ksf': 'madagaska', 'ksh': 'Maddajaska', 'ku': 'Madagaskar', 'kxv': 'madagascar', 'kxv-Deva': 'मेडागास्कर', 'kxv-Latn': 'madagascar', 'kxv-Orya': 'ମାଡ଼ାଗାସ୍କାର', 'kxv-Telu': 'మడగాస్కర్', 'ky': 'Мадагаскар', 'lag': 'Bukíini', 'lb': 'Madagaskar', 'lg': 'Madagasika', 'lld': 'Madagascar', 'ln': 'Madagasikari', 'lo': 'ມາດາກາສະກາ', 'lt': 'Madagaskaras', 'lu': 'Madagasikari', 'luo': 'Madagascar', 'luy': 'Bukini', 'lv': 'Madagaskara', 'mai': 'मेडागास्कर', 'mas': 'Bukini', 'mer': 'Madagasika', 'mfe': 'Madagaskar', 'mg': 'Madagasikara', 'mgh': 'Ubukini', 'mi': 'Matakāhika', 'mk': 'Мадагаскар', 'ml': 'മഡഗാസ്കർ', 'mn': 'Мадагаскар', 'mr': 'मादागास्कर', 'ms': 'Madagaskar', 'mt': 'Madagascar', 'mua': 'Madagaskar', 'my': 'မဒါဂတ်စကား', 'mzn': 'ماداگاسکار', 'naq': 'Madagascari', 'nb': 'Madagaskar', 'nd': 'Madagaska', 'ne': 'माडागास्कर', 'nl': 'Madagaskar', 'nmg': 'Madagaskar', 'nn': 'Madagaskar', 'no': 'Madagaskar', 'nqo': 'ߡߘߊߜ߭ߛߑߞߊ߯ߙ', 'nyn': 'Madagasika', 'om': 'Madagaaskaar', 'or': 'ମାଡାଗାସ୍କର୍', 'pa': 'ਮੈਡਾਗਾਸਕਰ', 'pa-Guru': 'ਮੈਡਾਗਾਸਕਰ', 'pcm': 'Madagáska', 'pl': 'Madagaskar', 'ps': 'مدغاسکر', 'pt': 'Madagascar', 'qu': 'Madagascar', 'rm': 'Madagascar', 'rn': 'Madagasikari', 'ro': 'Madagascar', 'rof': 'Bukini', 'ru': 'Мадагаскар', 'rwk': 'Bukini', 'saq': 'Bukini', 'sbp': 'Bukini', 'sc': 'Madagascàr', 'sd': 'مدگاسڪر', 'sd-Arab': 'مدگاسڪر', 'se': 'Madagaskar', 'seh': 'Madagascar', 'ses': 'Madagascar', 'sg': 'Madagaskära', 'shi': 'ⵎⴰⴷⴰⵖⴰⵛⵇⴰⵔ', 'shi-Latn': 'madaɣacqar', 'shi-Tfng': 'ⵎⴰⴷⴰⵖⴰⵛⵇⴰⵔ', 'si': 'මැඩගස්කරය', 'sk': 'Madagaskar', 'sl': 'Madagaskar', 'smn': 'Madagaskar', 'sn': 'Madagascar', 'so': 'Madagaskar', 'sq': 'Madagaskar', 'sr': 'Мадагаскар', 'sr-Cyrl': 'Мадагаскар', 'sr-Latn': 'Madagaskar', 'sv': 'Madagaskar', 'sw': 'Madagaska', 'syr': 'ܡܕܓܣܩܪ', 'ta': 'மடகாஸ்கர்', 'te': 'మడగాస్కర్', 'teo': 'Bukini', 'tg': 'Мадагаскар', 'th': 'มาดากัสการ์', 'ti': 'ማዳጋስካር', 'tk': 'Madagaskar', 'to': 'Matakasika', 'tr': 'Madagaskar', 'tt': 'Мадагаскар', 'twq': 'Madagascar', 'tzm': 'Madaɣacqar', 'ug': 'ماداغاسقار', 'uk': 'Мадагаскар', 'ur': 'مڈغاسکر', 'uz': 'Madagaskar', 'uz-Cyrl': 'Мадагаскар', 'uz-Latn': 'Madagaskar', 'vai': 'ꕮꕜꕭꔻꕪ', 'vai-Latn': 'Madagasita', 'vai-Vaii': 'ꕮꕜꕭꔻꕪ', 'vec': 'Madagascar', 'vi': 'Madagascar', 'vun': 'Bukini', 'wae': 'Madagaskar', 'wo': 'Madagaskaar', 'xh': 'EMadagascar', 'xnr': 'मेडागास्कर', 'xog': 'Madagasika', 'yav': 'matakaskáal', 'yi': 'מאַדאַגאַסקאַר', 'yo': 'Madasika', 'yrl': 'Madagasiká', 'yue': '馬達加斯加', 'yue-Hans': '马达加斯加', 'yue-Hant': '馬達加斯加', 'zgh': 'ⵎⴰⴷⴰⵖⴰⵛⵇⴰⵔ', 'zh': '马达加斯加', 'zh-Hans': '马达加斯加', 'zh-Hant': '馬達加斯加', 'zu': 'i-Madagascar'},
    'MH': {'af': 'Marshalleilande', 'agq': 'Chwɨlà fɨ Mashà', 'ak': 'Mahyaa Aeland', 'am': 'ማርሻል ደሴቶች', 'ar': 'جزر مارشال', 'as': 'মাৰ্শ্বাল দ্বীপপুঞ্জ', 'asa': 'Vithiwa vya Marshal', 'ast': 'Islles Marshall', 'az': 'Marşal adaları', 'az-Cyrl': 'Маршал адалары', 'az-Latn': 'Marşal adaları', 'bal-Latn': 'Marshall Islánd', 'bas': 'Bìòn bi Marcàl', 'be': 'Маршалавы астравы', 'be-tarask': 'Маршалавы астравы', 'bez': 'Ifisima fya Marshal', 'bg': 'Маршалови острови', 'blo': 'Marshal kaBʊtukǝltǝna', 'bm': 'Marisali Gun', 'bn': 'মার্শাল দ্বীপপুঞ্জ', 'br': 'Inizi Marshall', 'brx': 'मार्शेल द्वीप', 'bs': 'Maršalova ostrva', 'bs-Cyrl': 'Маршалска Острва', 'bs-Latn': 'Maršalova ostrva', 'ca': 'Illes Marshall', 'ca-ES-valencia': 'Illes Marshall', 'ccp': '𑄟𑄢𑄴𑄥𑄣𑄴 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Маршаллан гӀайренаш', 'ceb': 'Marshall Islands', 'cgg': 'Ebizinga bya Marshaa', 'chr': 'ᎹᏌᎵ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی مارشاڵ', 'cs': 'Marshallovy ostrovy', 'cv': 'Маршаллов утравӗсем', 'cy': 'Ynysoedd Marshall', 'da': 'Marshalløerne', 'dav': 'Visiwa vya Marshal', 'de': 'Marshallinseln', 'dje': 'Maršal gungey', 'dsb': 'Marshallowe kupy', 'dz': 'མར་ཤེལ་གླིང་ཚོམ', 'ebu': 'Visiwa vya Marshal', 'ee': 'Marshal ƒudomekpowo nutome', 'el': 'Νήσοι Μάρσαλ', 'el-polyton': 'Νῆσοι Μάρσαλ', 'en': 'Marshall Islands', 'en-Dsrt': '𐐣𐐪𐑉𐑇𐐲𐑊 𐐌𐑊𐐲𐑌𐐼𐑆', 'eo': 'Marŝaloj', 'es': 'Islas Marshall', 'et': 'Marshalli Saared', 'eu': 'Marshall Uharteak', 'ewo': 'Minlán Mí Maresál', 'fa': 'جزایر مارشال', 'ff': 'Duuɗe Marsaal', 'ff-Adlm': '𞤅𞤵𞤪𞤭𞥅𞤪𞤫 𞤃𞤢𞤪𞥃𞤢𞤤', 'ff-Latn': 'Duuɗe Marsaal', 'fi': 'Marshallinsaaret', 'fil': 'Marshall Islands', 'fo': 'Marshalloyggjar', 'fr': 'Îles Marshall', 'fur': 'Isulis Marshall', 'fy': 'Marshalleilannen', 'ga': 'Oileáin Marshall', 'gd': 'Eileanan Mharshall', 'gl': 'Illas Marshall', 'gsw': 'Marshallinsle', 'gu': 'માર્શલ આઇલેન્ડ્સ', 'guz': 'Visiwa vya Marshal', 'ha': 'Tsibiran Marshal', 'he': 'איי מרשל', 'hi': 'मार्शल द्वीपसमूह', 'hi-Latn': 'Marshall Islands', 'hr': 'Maršalovi Otoci', 'hsb': 'Marshallowe kupy', 'ht': 'Îles Marshall', 'hu': 'Marshall-szigetek', 'hy': 'Մարշալյան կղզիներ', 'ia': 'Insulas Marshall', 'id': 'Kepulauan Marshall', 'ig': 'Agwaetiti Marshall', 'is': 'Marshalleyjar', 'it': 'Isole Marshall', 'ja': 'マーシャル諸島', 'jmc': 'Visiwa vya Marshal', 'jv': 'Kapuloan Marshall', 'ka': 'მარშალის კუნძულები', 'kaa': 'Маршалл атаўлары', 'kaa-Cyrl': 'Маршалл атаўлары', 'kab': 'Tigzirin n Marcal', 'kam': 'Visiwa vya Marshal', 'kde': 'Chisiwa cha Malushal', 'kea': 'Ilhas Marxal', 'kgp': 'MỹrSar Goj-vẽso', 'khq': 'Maršal gungey', 'ki': 'Visiwa vya Marshal', 'kk': 'Маршалл аралдары', 'kk-Cyrl': 'Маршалл аралдары', 'kln': 'Ikwembeiyotab Marshall', 'km': 'កោះ\u200bម៉ាស់សល', 'kn': 'ಮಾರ್ಷಲ್ ದ್ವೀಪಗಳು', 'ko': '마셜 제도', 'kok': 'मार्शल आयलँड्स', 'kok-Deva': 'मार्शल आयलँड्स', 'ks': 'مارشَل جٔزیٖرٕ', 'ks-Arab': 'مارشَل جٔزیٖرٕ', 'ksb': 'Visiwa vya Mashal', 'ksf': 'zǝ i marcál', 'ksh': 'de Machschall-Enselle', 'ku': 'Giravên Marşalê', 'kxv': 'marsall dīp', 'kxv-Deva': 'मार्सल द्वीप', 'kxv-Latn': 'marsall dīp', 'kxv-Orya': 'ମାର୍ସାଲ ଦିପ', 'kxv-Telu': 'మార్సాల్ దీపులు', 'ky': 'Маршалл аралдары', 'lag': 'Visíiwa vya Marisháali', 'lb': 'Marshallinselen', 'lg': 'Bizinga bya Mariso', 'lld': 'Isoles Marshall', 'ln': 'Bisanga bya Marishalɛ', 'lo': 'ຫມູ່ເກາະມາແຊວ', 'lt': 'Maršalo Salos', 'lu': 'Lutanda lua Marishale', 'luo': 'Chuia mar Marshall', 'luy': 'Visiwa vya Marshal', 'lv': 'Māršala salas', 'mai': 'मार्शल द्वीपसमूह', 'mas': 'Visiwa vya Marshal', 'mer': 'Aĩrandi cia Marshal', 'mfe': 'Zil Marshall', 'mg': 'Nosy Marshall', 'mi': 'Ngā Motu Māhara', 'mk': 'Маршалски Острови', 'ml': 'മാർഷൽ ദ്വീപുകൾ', 'mn': 'Маршаллын арлууд', 'mr': 'मार्शल बेटे', 'ms': 'Kepulauan Marshall', 'mt': 'Gżejjer Marshall', 'mua': 'Sǝr Marshall ma laŋne', 'my': 'မာရှယ် ကျွန်းစု', 'mzn': 'مارشال جزایر', 'naq': 'Marshall Islands', 'nb': 'Marshalløyene', 'nd': 'Marshall Islands', 'ne': 'मार्शल टापुहरु', 'nl': 'Marshalleilanden', 'nmg': 'Maŋ́ má Marshall', 'nn': 'Marshalløyane', 'no': 'Marshalløyene', 'nqo': 'ߡߊߙߑߛߊߟ ߕߌ߲', 'nyn': 'Ebizinga bya Marshaa', 'om': 'Odoloota Maarshaal', 'or': 'ମାର୍ଶାଲ୍ ଦ୍ୱୀପପୁଞ୍ଜ', 'pa': 'ਮਾਰਸ਼ਲ ਟਾਪੂ', 'pa-Guru': 'ਮਾਰਸ਼ਲ ਟਾਪੂ', 'pcm': 'Máshál Aílands', 'pl': 'Wyspy Marshalla', 'ps': 'مارشل ټاپوګان', 'pt': 'Ilhas Marshall', 'qu': 'Islas Marshall', 'rm': 'Inslas da Marshall', 'rn': 'Izinga rya Marishari', 'ro': 'Insulele Marshall', 'rof': 'Visiwa vya Marshal', 'ru': 'Маршалловы о-ва', 'rwk': 'Visiwa vya Marshal', 'saq': 'Visiwa vya Marshal', 'sbp': 'Figunguli ifya Malishali', 'sc': 'Ìsulas Marshall', 'sd': 'مارشل ٻيٽ', 'sd-Arab': 'مارشل ٻيٽ', 'se': 'Marshallsullot', 'seh': 'Ilhas Marshall', 'ses': 'Maršal gungey', 'sg': 'Âzûâ Märshâl', 'shi': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵎⴰⵔⵛⴰⵍ', 'shi-Latn': 'tigzirin n marcal', 'shi-Tfng': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵎⴰⵔⵛⴰⵍ', 'si': 'මාෂල් දූපත්', 'sk': 'Marshallove ostrovy', 'sl': 'Marshallovi otoki', 'smn': 'Marshallsuolluuh', 'sn': 'Zvitsuwa zveMarshall', 'so': 'Jasiiradda Maarshal', 'sq': 'Ishujt Marshall', 'sr': 'Маршалска Острва', 'sr-Cyrl': 'Маршалска Острва', 'sr-Latn': 'Maršalska Ostrva', 'sv': 'Marshallöarna', 'sw': 'Visiwa vya Marshall', 'syr': 'ܓܙܪܬܐ ܡܐܪܫܐܠ', 'ta': 'மார்ஷல் தீவுகள்', 'te': 'మార్షల్ దీవులు', 'teo': 'Visiwa vya Marshal', 'tg': 'Ҷазираҳои Маршалл', 'th': 'หมู่เกาะมาร์แชลล์', 'ti': 'ደሴታት ማርሻል', 'tk': 'Marşall adalary', 'to': 'ʻOtumotu Māsolo', 'tr': 'Marshall Adaları', 'tt': 'Маршалл утраулары', 'twq': 'Maršal gungey', 'tzm': 'Tigzirin n Marcal', 'ug': 'مارشال ئاراللىرى', 'uk': 'Маршаллові Острови', 'ur': 'مارشل آئلینڈز', 'uz': 'Marshall orollari', 'uz-Cyrl': 'Маршал ороллари', 'uz-Latn': 'Marshall orollari', 'vai': 'ꕮꕊꕣ ꔳꘋꗣ', 'vai-Latn': 'Masha Tiŋŋɛ', 'vai-Vaii': 'ꕮꕊꕣ ꔳꘋꗣ', 'vec': 'Ìzole Marshall', 'vi': 'Quần đảo Marshall', 'vun': 'Visiwa vya Marshal', 'wae': 'Maršalinslä', 'wo': 'Duni Marsaal', 'xh': 'EMarshall Islands', 'xnr': 'मार्शल द्वीपां दा झुंड', 'xog': 'Bizinga bya Mariso', 'yav': 'ílmalasáal', 'yi': 'מאַרשאַל אינזלען', 'yo': 'Etikun Máṣali', 'yrl': 'Kapuãma-ita Marshall', 'yue': '馬紹爾群島', 'yue-Hans': '马绍尔群岛', 'yue-Hant': '馬紹爾群島', 'zgh': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵎⴰⵔⵛⴰⵍ', 'zh': '马绍尔群岛', 'zh-Hans': '马绍尔群岛', 'zh-Hant': '馬紹爾群島', 'zu': 'i-Marshall Islands'},
    'MK': {'af': 'Noord-Macedonië', 'ak': 'Mesidonia Atifi', 'am': 'ሰሜን መቄዶንያ', 'ar': 'مقدونيا الشمالية', 'as': 'উত্তৰ মেচিডোনীয়া', 'az': 'Şimali Makedoniya', 'az-Latn': 'Şimali Makedoniya', 'bal-Latn': 'Shemáli Makduniá', 'be': 'Паўночная Македонія', 'be-tarask': 'Паўночная Македонія', 'bg': 'Северна Македония', 'blo': 'Maseɖoniya gʊnyɩpɛnɛlaŋ kaja', 'bn': 'উত্তর ম্যাসেডোনিয়া', 'br': 'Makedonia an Norzh', 'brx': 'साहा मेसेड’निया', 'bs': 'Sjeverna Makedonija', 'bs-Cyrl': 'Сјеверна Македонија', 'bs-Latn': 'Sjeverna Makedonija', 'ca': 'Macedònia del Nord', 'ca-ES-valencia': 'Macedònia del Nord', 'ccp': '𑄟𑄳𑄠𑄥𑄓𑄮𑄚𑄨𑄠', 'ceb': 'North Macedonia', 'chr': 'ᏧᏴᏜ ᎹᏎᏙᏂᏯ', 'ckb': 'مەکدۆنیای باکوور', 'cs': 'Severní Makedonie', 'cv': 'Ҫурҫӗр Македони', 'cy': 'Gogledd Macedonia', 'da': 'Nordmakedonien', 'de': 'Nordmazedonien', 'dsb': 'Pódpołnocna Makedańska', 'el': 'Βόρεια Μακεδονία', 'el-polyton': 'Βόρεια Μακεδονία', 'en': 'North Macedonia', 'en-Dsrt': '𐐣𐐰𐑅𐐲𐐼𐐬𐑌𐐨𐐲', 'es': 'Macedonia del Norte', 'et': 'Põhja-Makedoonia', 'eu': 'Ipar Mazedonia', 'fa': 'مقدونیهٔ شمالی', 'ff-Adlm': '𞤃𞤢𞤳𞤫𞤣𞤮𞤲𞤭𞤴𞤢𞥄 𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫', 'fi': 'Pohjois-Makedonia', 'fil': 'North Macedonia', 'fo': 'Norður Makedónia', 'fr': 'Macédoine du Nord', 'ga': 'an Mhacadóin Thuaidh', 'gd': 'A’ Mhasadon a Tuath', 'gl': 'Macedonia do Norte', 'gu': 'ઉત્તર મેસેડોનિયા', 'ha': 'Macedonia ta Arewa', 'he': 'מקדוניה הצפונית', 'hi': 'उत्तरी मकदूनिया', 'hi-Latn': 'North Macedonia', 'hr': 'Sjeverna Makedonija', 'hsb': 'Serwjerna Makedonska', 'ht': 'Macédoine du Nord', 'hu': 'Észak-Macedónia', 'hy': 'Հյուսիսային Մակեդոնիա', 'ia': 'Macedonia del Nord', 'id': 'Makedonia Utara', 'ie': 'Nord-Macedonia', 'ig': 'North Macedonia', 'is': 'Norður-Makedónía', 'it': 'Macedonia del Nord', 'ja': '北マケドニア', 'jv': 'Républik Makédonia Lor', 'ka': 'ჩრდილოეთ მაკედონია', 'kaa': 'Арқа Македония', 'kaa-Cyrl': 'Арқа Македония', 'kea': 'Masidónia di Norti', 'kgp': 'Nãrti-Mỹsenonĩja', 'kk': 'Солтүстік Македония', 'kk-Cyrl': 'Солтүстік Македония', 'km': 'ម៉ាសេដ្វានខាងជើង', 'kn': 'ಉತ್ತರ ಮ್ಯಾಸಿಡೋನಿಯಾ', 'ko': '북마케도니아', 'kok': 'उत्तर मॅसिडोनिया', 'kok-Deva': 'उत्तर मॅसिडोनिया', 'kok-Latn': 'Ut’tor Masedonia', 'ks': 'شُمالی میسڈونیا', 'ks-Arab': 'شُمالی میسڈونیا', 'ku': 'Makendonyaya Bakur', 'kxv': 'uttar mesiḍoniya', 'kxv-Deva': 'उतॉरॉ मकदूनिया', 'kxv-Latn': 'uttar mesiḍoniya', 'kxv-Orya': 'ଉତର ମାସାଡନିୟା', 'kxv-Telu': 'ఉత్తర మాసిడోనియా', 'ky': 'Түндүк Македония', 'lb': 'Nordmazedonien', 'lld': 'Macedonia dl Nord', 'lo': 'ແມຊິໂດເນຍເໜືອ', 'lt': 'Šiaurės Makedonija', 'lv': 'Ziemeļmaķedonija', 'mai': 'उत्तरी मकदूनिया', 'mi': 'Makerōnia ki te Raki', 'mk': 'Северна Македонија', 'ml': 'നോർത്ത് മാസിഡോണിയ', 'mn': 'Хойд Македон', 'mr': 'उत्तर मॅसेडोनिया', 'ms': 'Macedonia Utara', 'mt': 'il-Maċedonja ta’ Fuq', 'my': 'မြောက် မက်ဆီဒိုးနီးယား', 'nb': 'Nord-Makedonia', 'ne': 'उत्तर म्यासेडोनिया', 'nl': 'Noord-Macedonië', 'nn': 'Nord-Makedonia', 'no': 'Nord-Makedonia', 'nqo': 'ߞߐ߬ߘߎ߮ ߡߊߛߋߘߏߣߌ߫', 'om': 'Maqdooniyaa Kaabaa', 'or': 'ଉତ୍ତର ମାସେଡୋନିଆ', 'pa': 'ਉੱਤਰੀ ਮੈਕਡੋਨੀਆ', 'pa-Guru': 'ਉੱਤਰੀ ਮੈਕਡੋਨੀਆ', 'pcm': 'Nọ́t Masidónia', 'pl': 'Macedonia Północna', 'ps': 'شمالي مقدونيه', 'pt': 'Macedônia do Norte', 'qu': 'Macedonia del Norte', 'quc': 'Macedonia del Norte', 'rm': 'Macedonia dal Nord', 'ro': 'Macedonia de Nord', 'ru': 'Северная Македония', 'rw': 'Masedoniya y’Amajyaruguru', 'sc': 'Matzedònia de su Nord', 'sd': 'اتر مقدونيا', 'sd-Arab': 'اتر مقدونيا', 'si': 'උතුරු මැසිඩෝනියාව', 'sk': 'Severné Macedónsko', 'sl': 'Severna Makedonija', 'so': 'Masedooniya Waqooyi', 'sq': 'Maqedonia e Veriut', 'sr': 'Северна Македонија', 'sr-Cyrl': 'Северна Македонија', 'sr-Latn': 'Severna Makedonija', 'sv': 'Nordmakedonien', 'sw': 'Masedonia ya Kaskazini', 'syr': 'ܓܪܒܝ ܡܩܕܘܢܝܐ', 'ta': 'வடக்கு மாசிடோனியா', 'te': 'ఉత్తర మాసిడోనియా', 'tg': 'Македонияи Шимолӣ', 'th': 'มาซิโดเนียเหนือ', 'ti': 'ሰሜን መቄዶንያ', 'tk': 'Demirgazyk Makedoniýa', 'to': 'Masetōnia fakatokelau', 'tr': 'Kuzey Makedonya', 'tt': 'Төньяк Македония', 'ug': 'شىمالىي ماكېدونىيە', 'uk': 'Північна Македонія', 'ur': 'شمالی مقدونیہ', 'uz': 'Shimoliy Makedoniya', 'uz-Cyrl': 'Шимолий Македония', 'uz-Latn': 'Shimoliy Makedoniya', 'vec': 'Masedonia de’l nord', 'vi': 'Bắc Macedonia', 'wo': 'Maseduwaan bëj Gànnaar', 'xh': 'EMntla Macedonia', 'xnr': 'उत्तरी मकदूनिया', 'yo': 'Àríwá Macedonia', 'yrl': 'Masedũniya', 'yue': '北馬其頓', 'yue-Hans': '马其顿', 'yue-Hant': '北馬其頓', 'zh': '北马其顿', 'zh-Hans': '北马其顿', 'zh-Hant': '北馬其頓', 'zu': 'i-North Macedonia'},
    'ML': {'af': 'Mali', 'agq': 'Malè', 'ak': 'Mali', 'am': 'ማሊ', 'ar': 'مالي', 'as': 'মালি', 'asa': 'Mali', 'ast': 'Malí', 'az': 'Mali', 'az-Cyrl': 'Мали', 'az-Latn': 'Mali', 'bal-Latn': 'Máli', 'bas': 'Màli', 'be': 'Малі', 'be-tarask': 'Малі', 'bez': 'Humali', 'bg': 'Мали', 'bgn': 'مالی', 'blo': 'Malii', 'bm': 'Mali', 'bn': 'মালি', 'br': 'Mali', 'brx': 'माली', 'bs': 'Mali', 'bs-Cyrl': 'Мали', 'bs-Latn': 'Mali', 'ca': 'Mali', 'ca-ES-valencia': 'Mali', 'ccp': '𑄟𑄣𑄨', 'ce': 'Мали', 'ceb': 'Mali', 'cgg': 'Mari', 'chr': 'ᎹᎵ', 'ckb': 'مالی', 'cs': 'Mali', 'cv': 'Мали', 'cy': 'Mali', 'da': 'Mali', 'dav': 'Mali', 'de': 'Mali', 'dje': 'Maali', 'dsb': 'Mali', 'dyo': 'Mali', 'dz': 'མཱ་ལི', 'ebu': 'Mali', 'ee': 'Mali nutome', 'el': 'Μάλι', 'el-polyton': 'Μαλί', 'en': 'Mali', 'en-Dsrt': '𐐣𐐪𐑊𐐨', 'eo': 'Malio', 'es': 'Mali', 'et': 'Mali', 'eu': 'Mali', 'ewo': 'Malí', 'fa': 'مالی', 'ff': 'Maali', 'ff-Adlm': '𞤃𞤢𞥄𞤤𞤭', 'ff-Latn': 'Maali', 'fi': 'Mali', 'fil': 'Mali', 'fo': 'Mali', 'fr': 'Mali', 'fur': 'Mali', 'fy': 'Mali', 'ga': 'Mailí', 'gd': 'Màili', 'gl': 'Malí', 'gsw': 'Maali', 'gu': 'માલી', 'guz': 'Mali', 'ha': 'Mali', 'he': 'מאלי', 'hi': 'माली', 'hi-Latn': 'Mali', 'hr': 'Mali', 'hsb': 'Mali', 'ht': 'Mali', 'hu': 'Mali', 'hy': 'Մալի', 'ia': 'Mali', 'id': 'Mali', 'ig': 'Mali', 'is': 'Malí', 'it': 'Mali', 'ja': 'マリ', 'jgo': 'Malî', 'jmc': 'Mali', 'jv': 'Mali', 'ka': 'მალი', 'kaa': 'Мали', 'kaa-Cyrl': 'Мали', 'kab': 'Mali', 'kam': 'Mali', 'kde': 'Mali', 'kea': 'Mali', 'kgp': 'Mỹri', 'khq': 'Maali', 'ki': 'Mali', 'kk': 'Мали', 'kk-Cyrl': 'Мали', 'kln': 'Emetab Mali', 'km': 'ម៉ាលី', 'kn': 'ಮಾಲಿ', 'ko': '말리', 'kok': 'माली', 'kok-Deva': 'माली', 'ks': 'مالی', 'ks-Arab': 'مالی', 'ksb': 'Mali', 'ksf': 'mali', 'ksh': 'Maali', 'ku': 'Malî', 'kxv': 'mali', 'kxv-Deva': 'माली', 'kxv-Latn': 'mali', 'kxv-Orya': 'ମାଲି', 'kxv-Telu': 'మాలి', 'ky': 'Мали', 'lag': 'Máali', 'lb': 'Mali', 'lg': 'Mali', 'lld': 'Mali', 'ln': 'Malí', 'lo': 'ມາລີ', 'lt': 'Malis', 'lu': 'Mali', 'luo': 'Mali', 'luy': 'Mali', 'lv': 'Mali', 'mai': 'माली', 'mas': 'Mali', 'mer': 'Mali', 'mfe': 'Mali', 'mg': 'Mali', 'mi': 'Māri', 'mk': 'Мали', 'ml': 'മാലി', 'mn': 'Мали', 'mr': 'माली', 'ms': 'Mali', 'mt': 'il-Mali', 'mua': 'Sǝr Mali', 'my': 'မာလီ', 'mzn': 'مالی', 'naq': 'Malib', 'nb': 'Mali', 'nd': 'Mali', 'ne': 'माली', 'nl': 'Mali', 'nmg': 'Mali', 'nn': 'Mali', 'no': 'Mali', 'nqo': 'ߡߊ߬ߟߌ߬', 'nyn': 'Mari', 'om': 'Maalii', 'or': 'ମାଲି', 'pa': 'ਮਾਲੀ', 'pa-Guru': 'ਮਾਲੀ', 'pcm': 'Máli', 'pl': 'Mali', 'ps': 'مالي', 'pt': 'Mali', 'qu': 'Malí', 'rm': 'Mali', 'rn': 'Mali', 'ro': 'Mali', 'rof': 'Mali', 'ru': 'Мали', 'rwk': 'Mali', 'saq': 'Mali', 'sbp': 'Mali', 'sc': 'Mali', 'sd': 'مالي', 'sd-Arab': 'مالي', 'se': 'Mali', 'seh': 'Mali', 'ses': 'Maali', 'sg': 'Malïi', 'shi': 'ⵎⴰⵍⵉ', 'shi-Latn': 'mali', 'shi-Tfng': 'ⵎⴰⵍⵉ', 'si': 'මාලි', 'sk': 'Mali', 'sl': 'Mali', 'smn': 'Mali', 'sn': 'Mali', 'so': 'Maali', 'sq': 'Mali', 'sr': 'Мали', 'sr-Cyrl': 'Мали', 'sr-Latn': 'Mali', 'sv': 'Mali', 'sw': 'Mali', 'syr': 'ܡܐܠܝ', 'ta': 'மாலி', 'te': 'మాలి', 'teo': 'Mali', 'tg': 'Мали', 'th': 'มาลี', 'ti': 'ማሊ', 'tk': 'Mali', 'to': 'Māli', 'tr': 'Mali', 'tt': 'Мали', 'twq': 'Maali', 'tzm': 'Mali', 'ug': 'مالى', 'uk': 'Малі', 'ur': 'مالی', 'uz': 'Mali', 'uz-Cyrl': 'Мали', 'uz-Latn': 'Mali', 'vai': 'ꕮꔷ', 'vai-Latn': 'Mali', 'vai-Vaii': 'ꕮꔷ', 'vec': 'Mali', 'vi': 'Mali', 'vun': 'Mali', 'wae': 'Mali', 'wo': 'Mali', 'xh': 'EMali', 'xnr': 'माली', 'xog': 'Mali', 'yav': 'malí', 'yi': 'מאַלי', 'yo': 'Mali', 'yrl': 'Mari', 'yue': '馬利', 'yue-Hans': '马利', 'yue-Hant': '馬利', 'zgh': 'ⵎⴰⵍⵉ', 'zh': '马里', 'zh-Hans': '马里', 'zh-Hant': '馬利', 'zu': 'iMali'},
    'MM': {'af': 'Mianmar (Birma)', 'agq': 'Mǐanmà', 'ak': 'Mayaama (Bɛɛma)', 'am': 'ማይናማር(በርማ)', 'ar': 'ميانمار (بورما)', 'as': 'ম্যানমাৰ (বাৰ্মা)', 'asa': 'Myama', 'ast': 'Myanmar (Birmania)', 'az': 'Myanma', 'az-Cyrl': 'Мјанма', 'az-Latn': 'Myanma', 'bal-Latn': 'Myanmár (Barmá)', 'bas': 'Myànmâr', 'be': 'М’янма (Бірма)', 'be-tarask': 'М’янма (Бірма)', 'bez': 'Humyama', 'bg': 'Мианмар (Бирма)', 'blo': 'Miyanmaa (Birmanii)', 'bm': 'Myanimari', 'bn': 'মায়ানমার (বার্মা)', 'br': 'Myanmar (Birmania)', 'brx': 'म्यानमार', 'bs': 'Mijanmar', 'bs-Cyrl': 'Мјанмар', 'bs-Latn': 'Mijanmar', 'ca': 'Myanmar (Birmània)', 'ca-ES-valencia': 'Myanmar (Birmània)', 'ccp': '𑄟𑄠𑄚𑄴𑄟𑄢𑄴 (𑄝𑄢𑄴𑄟)', 'ce': 'Мьянма (Бирма)', 'ceb': 'Myanmar (Burma)', 'cgg': 'Myanamar', 'chr': 'ᎹᏯᎹᎵ (ᏇᎵᎹ)', 'ckb': 'میانمار', 'cs': 'Myanmar (Barma)', 'cv': 'Мьянма (Бирма)', 'cy': 'Myanmar (Burma)', 'da': 'Myanmar (Burma)', 'dav': 'Myama', 'de': 'Myanmar', 'dje': 'Maynamar', 'dsb': 'Myanmar', 'dz': 'མི་ཡཱན་མར་ (བྷར་མ)', 'ebu': 'Myama', 'ee': 'Myanmar (Burma) nutome', 'el': 'Μιανμάρ (Βιρμανία)', 'el-polyton': 'Μιανμάρ (Βιρμανία)', 'en': 'Myanmar (Burma)', 'en-Dsrt': '𐐣𐐨𐐲𐑌𐑋𐐪𐑉', 'eo': 'Birmo', 'es': 'Myanmar (Birmania)', 'et': 'Myanmar (Birma)', 'eu': 'Myanmar (Birmania)', 'ewo': 'Mianəmár', 'fa': 'میانمار (برمه)', 'ff': 'Miyamaar', 'ff-Adlm': '𞤃𞤭𞤴𞤢𞤥𞤢𞥄𞤪 (𞤄𞤵𞥅𞤪𞤥𞤢)', 'ff-Latn': 'Miyamaar', 'fi': 'Myanmar (Burma)', 'fil': 'Myanmar (Burma)', 'fo': 'Myanmar (Burma)', 'fr': 'Myanmar (Birmanie)', 'fur': 'Birmanie', 'fy': 'Myanmar (Birma)', 'ga': 'Maenmar (Burma)', 'gd': 'Miànmar', 'gl': 'Myanmar (Birmania)', 'gsw': 'Myanmar (Burma)', 'gu': 'મ્યાંમાર (બર્મા)', 'guz': 'Myama', 'ha': 'Burma, Miyamar', 'he': 'מיאנמר (בורמה)', 'hi': 'म्यांमार (बर्मा)', 'hi-Latn': 'Myanmar (Burma)', 'hr': 'Mjanmar (Burma)', 'hsb': 'Myanmar', 'ht': 'Myanmar (Birmanie)', 'hu': 'Mianmar', 'hy': 'Մյանմա (Բիրմա)', 'ia': 'Myanmar (Birmania)', 'id': 'Myanmar (Burma)', 'ig': 'Myanmar (Burma)', 'is': 'Mjanmar (Búrma)', 'it': 'Myanmar (Birmania)', 'ja': 'ミャンマー (ビルマ)', 'jmc': 'Myama', 'jv': 'Myanmar (Burma)', 'ka': 'მიანმარი (ბირმა)', 'kaa': 'Мьянма (Бирма)', 'kaa-Cyrl': 'Мьянма (Бирма)', 'kab': 'Myanmar', 'kam': 'Myama', 'kde': 'Myama', 'kea': 'Mianmar (Birmánia)', 'kgp': 'Mĩjỹmỹr', 'khq': 'Maynamar', 'ki': 'Myama', 'kk': 'Мьянма (Бирма)', 'kk-Cyrl': 'Мьянма (Бирма)', 'kln': 'Emetab Myanmar', 'km': 'មីយ៉ាន់ម៉ា (ភូមា)', 'kn': 'ಮಯನ್ಮಾರ್ (ಬರ್ಮಾ)', 'ko': '미얀마', 'kok': 'म्यानमार (बर्मा)', 'kok-Deva': 'म्यानमार (बर्मा)', 'ks': 'میانمار (برما)', 'ks-Arab': 'میانمار (برما)', 'ksb': 'Myama', 'ksf': 'myanmár', 'ksh': 'Birma', 'ku': 'Myanmar (Bûrma)', 'kxv': 'myanmar (brma)', 'kxv-Deva': 'म्यांमार (बर्मा)', 'kxv-Latn': 'myanmar (brma)', 'kxv-Orya': 'ମିୟାଁମାର (ବର୍ମା)', 'kxv-Telu': 'మయన్మార్ (బర్మా)', 'ky': 'Мьянма (Бирма)', 'lag': 'Miáama', 'lb': 'Myanmar', 'lg': 'Myanima', 'lld': 'Myanmar (Burma)', 'ln': 'Birmanie', 'lo': 'ມຽນມາ (ເບີມາ)', 'lt': 'Mianmaras (Birma)', 'lu': 'Myamare', 'luo': 'Myanmar', 'luy': 'Myama', 'lv': 'Mjanma (Birma)', 'mai': 'मयनमार (बर्मा)', 'mas': 'Myama', 'mer': 'Myanima', 'mfe': 'Myanmar', 'mg': 'Myanmar', 'mi': 'Pēma', 'mk': 'Мјанмар (Бурма)', 'ml': 'മ്യാൻമാർ (ബർമ്മ)', 'mn': 'Мьянмар', 'mr': 'म्यानमार (बर्मा)', 'ms': 'Myanmar (Burma)', 'mt': 'il-Myanmar/Burma', 'mua': 'Sǝr Myanmar', 'my': 'မြန်မာ', 'mzn': 'میانمار', 'naq': 'Myanmar', 'nb': 'Myanmar (Burma)', 'nd': 'Myanmar', 'ne': 'म्यान्मार (बर्मा)', 'nl': 'Myanmar (Birma)', 'nmg': 'Myanmar', 'nn': 'Myanmar (Burma)', 'no': 'Myanmar (Burma)', 'nqo': 'ߡߌߦߊߡߊ߯ߙ ( ߓߙߌߡߊߣߌ߫ )', 'nyn': 'Myanamar', 'om': 'Maayinaamar (Burma)', 'or': 'ମିଆଁମାର', 'pa': 'ਮਿਆਂਮਾਰ (ਬਰਮਾ)', 'pa-Guru': 'ਮਿਆਂਮਾਰ (ਬਰਮਾ)', 'pcm': 'Miánma (Bọ́ma)', 'pl': 'Mjanma (Birma)', 'ps': 'ميانمار (برما)', 'pt': 'Mianmar (Birmânia)', 'qu': 'Myanmar', 'rm': 'Myanmar (Burma)', 'rn': 'Birimaniya', 'ro': 'Myanmar (Birmania)', 'rof': 'Myama', 'ru': 'Мьянма (Бирма)', 'rwk': 'Myama', 'saq': 'Myama', 'sbp': 'Muyama', 'sc': 'Myanmàr (Birmània)', 'sd': 'ميانمار (برما)', 'sd-Arab': 'ميانمار (برما)', 'se': 'Burma', 'seh': 'Mianmar', 'ses': 'Maynamar', 'sg': 'Myämâra', 'shi': 'ⵎⵢⴰⵏⵎⴰⵔ', 'shi-Latn': 'myanmar', 'shi-Tfng': 'ⵎⵢⴰⵏⵎⴰⵔ', 'shn': 'မျၢၼ်ႇမႃႇ (မိူင်းမၢၼ်ႈ)', 'si': 'මියන්මාරය (බුරුමය)', 'sk': 'Mjanmarsko', 'sl': 'Mjanmar (Burma)', 'smn': 'Myanmar (Burma)', 'sn': 'Myanmar', 'so': 'Mayanmar', 'sq': 'Mianmar (Burmë)', 'sr': 'Мијанмар (Бурма)', 'sr-Cyrl': 'Мијанмар (Бурма)', 'sr-Latn': 'Mijanmar (Burma)', 'sv': 'Myanmar (Burma)', 'sw': 'Myanmar (Burma)', 'syr': 'ܡܝܐܢܡܐܪ (ܒܘܪܡܐ)', 'ta': 'மியான்மார் (பர்மா)', 'te': 'మయన్మార్', 'teo': 'Myama', 'tg': 'Мянма', 'th': 'เมียนมา (พม่า)', 'ti': 'ሚያንማር (በርማ)', 'tk': 'Mýanma (Birma)', 'to': 'Mianimā (Pema)', 'tr': 'Myanmar (Burma)', 'tt': 'Мьянма (Бирма)', 'twq': 'Maynamar', 'tzm': 'Myanmar', 'ug': 'بىرما', 'uk': 'Мʼянма (Бірма)', 'ur': 'میانمار (برما)', 'uz': 'Myanma (Birma)', 'uz-Cyrl': 'Мьянма (Бирма)', 'uz-Latn': 'Myanma (Birma)', 'vai': 'ꕆꕩꘋꕮ', 'vai-Latn': 'Miyamaha', 'vai-Vaii': 'ꕆꕩꘋꕮ', 'vec': 'Myanmar (Birmania)', 'vi': 'Myanmar (Miến Điện)', 'vun': 'Myama', 'wae': 'Burma', 'wo': 'Miyanmaar', 'xh': 'EMyanmar (Burma)', 'xnr': 'म्यांमार (बर्मा)', 'xog': 'Myanima', 'yav': 'miaŋmáal', 'yi': 'מיאַנמאַר', 'yo': 'Manamari', 'yrl': 'Miyamá (Bimãniya)', 'yue': '緬甸', 'yue-Hans': '缅甸', 'yue-Hant': '緬甸', 'zgh': 'ⵎⵢⴰⵏⵎⴰⵔ', 'zh': '缅甸', 'zh-Hans': '缅甸', 'zh-Hant': '緬甸', 'zu': 'i-Myanmar (Burma)'},
    'MN': {'af': 'Mongolië', 'agq': 'Mùŋgolìa', 'ak': 'Mɔngolia', 'am': 'ሞንጎሊያ', 'ar': 'منغوليا', 'as': 'মঙ্গোলিয়া', 'asa': 'Mongolia', 'ast': 'Mongolia', 'az': 'Monqolustan', 'az-Cyrl': 'Монголустан', 'az-Latn': 'Monqolustan', 'bal-Latn': 'Mangóliá', 'bas': 'Mòŋgolìà', 'be': 'Манголія', 'be-tarask': 'Манголія', 'bez': 'Humongolia', 'bg': 'Монголия', 'blo': 'Mɔŋgolii', 'bm': 'Moŋoli', 'bn': 'মঙ্গোলিয়া', 'br': 'Mongolia', 'brx': 'मंगलिया', 'bs': 'Mongolija', 'bs-Cyrl': 'Монголија', 'bs-Latn': 'Mongolija', 'ca': 'Mongòlia', 'ca-ES-valencia': 'Mongòlia', 'ccp': '𑄟𑄧𑄋𑄴𑄉𑄮𑄣𑄨𑄠', 'ce': 'Монголи', 'ceb': 'Mongolia', 'cgg': 'Mongoria', 'chr': 'ᎹᏂᎪᎵᎠ', 'ckb': 'مەنگۆلیا', 'cs': 'Mongolsko', 'cv': 'Монголи', 'cy': 'Mongolia', 'da': 'Mongoliet', 'dav': 'Mongolia', 'de': 'Mongolei', 'dje': 'Mongooli', 'dsb': 'Mongolska', 'dz': 'སོག་པོ་ཡུལ', 'ebu': 'Mongolia', 'ee': 'Mongolia nutome', 'el': 'Μογγολία', 'el-polyton': 'Μογγολία', 'en': 'Mongolia', 'en-Dsrt': '𐐣𐐪𐑍𐑀𐐬𐑊𐐨𐐲', 'eo': 'Mongolujo', 'es': 'Mongolia', 'et': 'Mongoolia', 'eu': 'Mongolia', 'ewo': 'Mɔngɔ́lia', 'fa': 'مغولستان', 'ff': 'Monngolii', 'ff-Adlm': '𞤃𞤮𞤲𞤺𞤮𞤤𞤭𞤴𞤢', 'ff-Latn': 'Monngolii', 'fi': 'Mongolia', 'fil': 'Mongolia', 'fo': 'Mongolia', 'fr': 'Mongolie', 'fur': 'Mongolie', 'fy': 'Mongolië', 'ga': 'an Mhongóil', 'gd': 'Dùthaich nam Mongol', 'gl': 'Mongolia', 'gsw': 'Mongolei', 'gu': 'મંગોલિયા', 'guz': 'Mongolia', 'ha': 'Mangoliya', 'he': 'מונגוליה', 'hi': 'मंगोलिया', 'hi-Latn': 'Mongolia', 'hr': 'Mongolija', 'hsb': 'Mongolska', 'ht': 'Mongolie', 'hu': 'Mongólia', 'hy': 'Մոնղոլիա', 'ia': 'Mongolia', 'id': 'Mongolia', 'ig': 'Mongolia', 'is': 'Mongólía', 'it': 'Mongolia', 'ja': 'モンゴル', 'jmc': 'Mongolia', 'jv': 'Mongolia', 'ka': 'მონღოლეთი', 'kaa': 'Монголия', 'kaa-Cyrl': 'Монголия', 'kab': 'Mungulya', 'kam': 'Mongolia', 'kde': 'Mongolia', 'kea': 'Mongólia', 'kgp': 'Mãngórija', 'khq': 'Mongooli', 'ki': 'Mongolia', 'kk': 'Моңғолия', 'kk-Cyrl': 'Моңғолия', 'kln': 'Emetab Mongolia', 'km': 'ម៉ុងហ្គោលី', 'kn': 'ಮಂಗೋಲಿಯಾ', 'ko': '몽골', 'kok': 'मंगोलिया', 'kok-Deva': 'मंगोलिया', 'ks': 'مَنگولِیا', 'ks-Arab': 'مَنگولِیا', 'ksb': 'Mongolia', 'ksf': 'mɔŋolí', 'ksh': 'Mongjolei', 'ku': 'Moxolistan', 'kxv': 'mongolia', 'kxv-Deva': 'मंगोलिया', 'kxv-Latn': 'mongolia', 'kxv-Orya': 'ମଙ୍ଗଲିୟା', 'kxv-Telu': 'మంగోలియా', 'ky': 'Монголия', 'lag': 'Mongólia', 'lb': 'Mongolei', 'lg': 'Mongoliya', 'lld': 'Mongolia', 'ln': 'Mongolí', 'lo': 'ມອງໂກເລຍ', 'lt': 'Mongolija', 'lu': 'Mongoli', 'luo': 'Mongolia', 'luy': 'Mongolia', 'lv': 'Mongolija', 'mai': 'मंगोलिया', 'mas': 'Mongolia', 'mer': 'Mongolia', 'mfe': 'Mongoli', 'mg': 'Môngôlia', 'mi': 'Mongōria', 'mk': 'Монголија', 'ml': 'മംഗോളിയ', 'mn': 'Монгол', 'mr': 'मंगोलिया', 'ms': 'Mongolia', 'mt': 'il-Mongolja', 'mua': 'Mongoliya', 'my': 'မွန်ဂိုးလီးယား', 'mzn': 'مغولستون', 'naq': 'Mongolia', 'nb': 'Mongolia', 'nd': 'Mongolia', 'ne': 'मङ्गोलिया', 'nl': 'Mongolië', 'nmg': 'Mɔngolia', 'nn': 'Mongolia', 'no': 'Mongolia', 'nqo': 'ߡߐ߲ߜ߭ߐߟߌ߫', 'nyn': 'Mongoria', 'om': 'Mongoliyaa', 'or': 'ମଙ୍ଗୋଲିଆ', 'pa': 'ਮੰਗੋਲੀਆ', 'pa-Guru': 'ਮੰਗੋਲੀਆ', 'pcm': 'Mọngólia', 'pl': 'Mongolia', 'ps': 'منګوليا', 'pt': 'Mongólia', 'qu': 'Mongolia', 'rm': 'Mongolia', 'rn': 'Mongoliya', 'ro': 'Mongolia', 'rof': 'Mongolia', 'ru': 'Монголия', 'rwk': 'Mongolia', 'saq': 'Mongolia', 'sbp': 'Mongoliya', 'sc': 'Mongòlia', 'sd': 'منگوليا', 'sd-Arab': 'منگوليا', 'se': 'Mongolia', 'seh': 'Mongólia', 'ses': 'Mongooli', 'sg': 'Mongolïi', 'shi': 'ⵎⵏⵖⵓⵍⵢⴰ', 'shi-Latn': 'mnɣulya', 'shi-Tfng': 'ⵎⵏⵖⵓⵍⵢⴰ', 'si': 'මොන්ගෝලියාව', 'sk': 'Mongolsko', 'sl': 'Mongolija', 'smn': 'Mongolia', 'sn': 'Mongolia', 'so': 'Mongooliya', 'sq': 'Mongoli', 'sr': 'Монголија', 'sr-Cyrl': 'Монголија', 'sr-Latn': 'Mongolija', 'sv': 'Mongoliet', 'sw': 'Mongolia', 'syr': 'ܡܘܢܓܘܠܝܐ', 'ta': 'மங்கோலியா', 'te': 'మంగోలియా', 'teo': 'Mongolia', 'tg': 'Муғулистон', 'th': 'มองโกเลีย', 'ti': 'ሞንጎልያ', 'tk': 'Mongoliýa', 'to': 'Mongokōlia', 'tr': 'Moğolistan', 'tt': 'Монголия', 'twq': 'Mongooli', 'tzm': 'Manɣulya', 'ug': 'موڭغۇلىيە', 'uk': 'Монголія', 'ur': 'منگولیا', 'uz': 'Mongoliya', 'uz-Cyrl': 'Монголия', 'uz-Latn': 'Mongoliya', 'vai': 'ꗞꖐꔷꕩ', 'vai-Latn': 'Mɔngoliya', 'vai-Vaii': 'ꗞꖐꔷꕩ', 'vec': 'Mongolia', 'vi': 'Mông Cổ', 'vun': 'Mongolia', 'wae': 'Mongolei', 'wo': 'Mongoli', 'xh': 'EMongolia', 'xnr': 'मंगोलिया', 'xog': 'Mongoliya', 'yav': 'mongolí', 'yi': 'מאנגאליי', 'yo': 'Mogolia', 'yrl': 'Mũguriya', 'yue': '蒙古', 'yue-Hans': '蒙古', 'yue-Hant': '蒙古', 'zgh': 'ⵎⵏⵖⵓⵍⵢⴰ', 'zh': '蒙古', 'zh-Hans': '蒙古', 'zh-Hant': '蒙古', 'zu': 'i-Mongolia'},
    'MO': {'af': 'Macau SAS China', 'ak': 'Makaw Kyaena', 'am': 'ማካኦ ልዩ የአስተዳደር ክልል ቻይና', 'ar': 'منطقة ماكاو الإدارية الخاصة', 'as': 'মাকাও এছ. এ. আৰ. চীন', 'ast': 'ARE China de Macáu', 'az': 'Makao XİR Çin', 'az-Cyrl': 'Макао Хүсуси Инзибати Әрази Чин', 'az-Latn': 'Makao XİR Çin', 'bal-Latn': 'Makaó SAR Chin', 'be': 'Макаа, САР (Кітай)', 'be-tarask': 'Макаа, САР (Кітай)', 'bg': 'Макао, САР на Китай', 'blo': 'Makawoo Caɩna kaja', 'bn': 'ম্যাকাও এসএআর চীন', 'br': 'Macau RMD Sina', 'brx': 'माकाउ एस.ए.आर चाइना', 'bs': 'Makao (SAR Kina)', 'bs-Cyrl': 'Макао С. А. Р.', 'bs-Latn': 'Makao (SAR Kina)', 'ca': 'Macau (RAE Xina)', 'ca-ES-valencia': 'Macau (RAE Xina)', 'ccp': '𑄟𑄳𑄠𑄇𑄃𑄮 𑄆𑄌𑄴𑄃𑄬𑄃𑄢𑄴 𑄌𑄩𑄚', 'ce': 'Макао (ша-къаьстина кӀошт)', 'ceb': 'Macao SAR China', 'chr': 'ᎹᎧᎣ (ᎤᏓᏤᎵᏓ ᏧᏂᎸᏫᏍᏓᏁᏗ ᎢᎬᎾᏕᎾ) ᏣᎢ', 'ckb': 'ماکائۆ', 'cs': 'Macao – ZAO Číny', 'cv': 'Макао (САР)', 'cy': 'Macau SAR Tsieina', 'da': 'SAR Macao', 'de': 'Sonderverwaltungsregion Macau', 'dsb': 'Wósebna zastojnstwowa cona Macao', 'dz': 'མཀ་ཨའུ་ཅཱའི་ན', 'ee': 'Macau SAR Tsaina nutome', 'el': 'Μακάο ΕΔΠ Κίνας', 'el-polyton': 'Μακάο, Εἰδικὴ Διοικητικὴ Περιφέρεια τῆς Κίνας', 'en': 'Macao SAR China', 'en-Dsrt': '𐐣𐐲𐐿𐐵 𐐝𐐈𐐡 𐐕𐐴𐑌𐐲', 'es': 'RAE de Macao (China)', 'et': 'Macau erihalduspiirkond', 'eu': 'Macau Txinako AEB', 'fa': 'ماکائو، منطقهٔ ویژهٔ اداری چین', 'ff-Adlm': '𞤖𞤂𞤀 𞤕𞤢𞤴𞤲𞤢 𞤫 𞤃𞤢𞤳𞤢𞤱𞤮𞥅', 'fi': 'Macao – Kiinan erityishallintoalue', 'fil': 'Macau SAR China', 'fo': 'Makao SAR Kina', 'fr': 'R.A.S. chinoise de Macao', 'fur': 'Regjon aministrative speciâl de Cine di Macao', 'fy': 'Macao SAR van Sina', 'ga': 'Sainréigiún Riaracháin Macao, Daonphoblacht na Síne', 'gd': 'Macàthu SAR na Sìne', 'gl': 'Macau RAE da China', 'gsw': 'Sonderverwaltigszone Macao', 'gu': 'મકાઉ SAR ચીન', 'ha': 'Babban Yankin Mulkin Macao na Ƙasar Sin', 'he': 'מקאו (אזור מנהלי מיוחד של סין)', 'hi': 'मकाऊ (विशेष प्रशासनिक क्षेत्र चीन)', 'hi-Latn': 'Macao SAR China', 'hr': 'PUP Makao Kina', 'hsb': 'Wosebita zarjadniska cona Macao', 'ht': 'R.A.S. chinoise de Macao', 'hu': 'Makaó KKT', 'hy': 'Չինաստանի Մակաո ՀՎՇ', 'ia': 'Macao, R.A.S. de China', 'id': 'Makau DAK Tiongkok', 'ig': 'Macao SAR China', 'is': 'sérstjórnarsvæðið Makaó', 'it': 'RAS di Macao', 'ja': '中華人民共和国マカオ特別行政区', 'jv': 'Laladan Administratif Astamiwa Makau', 'ka': 'მაკაოს სპეციალური ადმინისტრაციული რეგიონი, ჩინეთი', 'kaa': 'Макао ААР', 'kaa-Cyrl': 'Макао ААР', 'kea': 'Makau, Rejion Administrativu Spesial di Xina', 'kgp': 'Mỹkav, Sĩnỹ ERA', 'kk': 'Макао АӘА', 'kk-Cyrl': 'Макао АӘА', 'km': 'ម៉ាកាវ តំបន់រដ្ឋបាលពិសេសចិន', 'kn': 'ಮಕಾವು ಎಸ್ಎಆರ್ ಚೈನಾ', 'ko': '마카오(중국 특별행정구)', 'kok': 'मकाव एसएआर चीन', 'kok-Deva': 'मकाव एसएआर चीन', 'ks': 'مَکاوو ایس اے آر چیٖن', 'ks-Arab': 'مَکاوو ایس اے آر چیٖن', 'ksh': 'Makau', 'ku': 'Makaoya Hît ya Çînê', 'kxv': 'makao sar cina', 'kxv-Deva': 'मकाऊ (एस ए आर चीन)', 'kxv-Latn': 'makao sar cina', 'kxv-Orya': 'ମକାଉ ଏସଏଆର ଚିନା', 'kxv-Telu': 'మకావ్ ఎస్ఏఆర్ చైనా', 'ky': 'Макао Кытай ААА', 'lb': 'Spezialverwaltungszon Macau', 'lld': 'Regiun aministrativa speziala de Macao', 'lo': 'ມາກາວ ເຂດປົກຄອງພິເສດ ຈີນ', 'lt': 'Ypatingasis Administracinis Kinijos Regionas Makao', 'lv': 'ĶTR īpašais administratīvais reģions Makao', 'mai': 'मकाउ (विशेष प्रशासनिक क्षेत्र चीन)', 'mi': 'Makau Haina', 'mk': 'Макао САР', 'ml': 'മക്കാവു എസ്.എ.ആർ. ചൈന', 'mn': 'БНХАУ-ын Тусгай захиргааны бүс Макао', 'mr': 'मकाओ एसएआर चीन', 'ms': 'Macau SAR China', 'mt': 'ir-Reġjun Amministrattiv Speċjali tal-Macao tar-Repubblika tal-Poplu taċ-Ċina', 'my': 'မကာအို (တရုတ်ပြည်)', 'mzn': 'ماکائو (چین دله)', 'nb': 'Macao SAR Kina', 'ne': 'मकाउ चिनियाँ विशेष प्रशासनिक क्षेत्र', 'nl': 'Macau SAR van China', 'nn': 'Macao SAR Kina', 'no': 'Macao SAR Kina', 'nqo': 'ߛߌߣ ߕߌ߲߬ߞߎߘߎ߲߫ ߡߊߡߙߊ߬ߣߍ߲ ߡߞߊߥߏ߫', 'om': 'Maka’oo SAR Chaayinaa', 'or': 'ମାକାଉ ଏସଏଆର୍\u200c ଚାଇନା', 'pa': 'ਮਕਾਉ ਐਸਏਆਰ ਚੀਨ', 'pa-Guru': 'ਮਕਾਉ ਐਸਏਆਰ ਚੀਨ', 'pcm': 'Makáo SAR Chaína', 'pl': 'SRA Makau (Chiny)', 'ps': 'مکاو SAR چین', 'pt': 'Macau, RAE da China', 'qu': 'Macao RAE China', 'rm': 'Regiun d’administraziun speziala Macao, China', 'ro': 'R.A.S. Macao, China', 'ru': 'Макао (САР)', 'sc': 'RAS tzinesa de Macao', 'sd': 'مڪائو SAR چين', 'sd-Arab': 'مڪائو SAR چين', 'se': 'Makáo', 'si': 'මැකාවු විශේෂ පරිපාලන කලාපය චීනය', 'sk': 'Macao – OAO Číny', 'sl': 'Posebno upravno območje Ljudske republike Kitajske Macao', 'smn': 'Macao - – Kiina e.h.k.', 'so': 'Makaaw', 'sq': 'RPA i Makaos', 'sr': 'САР Макао (Кина)', 'sr-Cyrl': 'САР Макао (Кина)', 'sr-Latn': 'SAR Makao (Kina)', 'sv': 'Macao SAR', 'sw': 'Makau SAR China', 'ta': 'மகாவ் எஸ்ஏஆர் சீனா', 'te': 'మకావ్ ఎస్ఏఆర్ చైనా', 'tg': 'Макао (МММ)', 'th': 'เขตปกครองพิเศษมาเก๊าแห่งสาธารณรัฐประชาชนจีน', 'ti': 'ፍሉይ ምምሕዳራዊ ዞባ ማካው (ቻይና)', 'tk': 'Makao AAS Hytaý', 'to': 'Makau SAR Siaina', 'tr': 'Çin Makao ÖİB', 'tt': 'Макао Махсус Идарәле Төбәге', 'ug': 'ئاۋمېن ئالاھىدە مەمۇرىي رايونى', 'uk': 'Макао, ОАР Китаю', 'ur': 'مکاؤ SAR چین', 'uz': 'Makao (Xitoy MMH)', 'uz-Cyrl': 'Макао (Хитой ММҲ)', 'uz-Latn': 'Makao (Xitoy MMH)', 'vai': 'ꕮꗛꖺ', 'vai-Vaii': 'ꕮꗛꖺ', 'vec': 'RAS de Macào', 'vi': 'Đặc khu Hành chính Macao, Trung Quốc', 'wae': 'Sonderverwaltigszona Makau', 'wo': 'Makaawo', 'xh': 'EMacao SAR China', 'xnr': 'मकाऊ (खास प्रशासनिक लाका चीन)', 'yo': 'Agbègbè Ìṣàkóso Pàtàkì Macao', 'yrl': 'Makau, RAE Xina yara', 'yue': '中國澳門特別行政區', 'yue-Hans': '中华人民共和国澳门特别行政区', 'yue-Hant': '中國澳門特別行政區', 'zh': '中国澳门特别行政区', 'zh-Hans': '中国澳门特别行政区', 'zh-Hant': '中國澳門特別行政區', 'zu': 'i-Macau SAR China'},
    'MP': {'af': 'Noord-Mariane-eilande', 'agq': 'Chwɨlà m̀ Màlǐanà mɨ̀ Ekùw mò', 'ak': 'Mariana Atifi Fam Aeland', 'am': 'የሰሜናዊ ማሪያና ደሴቶች', 'ar': 'جزر ماريانا الشمالية', 'as': 'উত্তৰ মাৰিয়ানা দ্বীপপুঞ্জ', 'asa': 'Vithiwa vya Mariana vya Kathkazini', 'ast': 'Islles Marianes del Norte', 'az': 'Şimali Marian adaları', 'az-Cyrl': 'Шимали Мариан адалары', 'az-Latn': 'Şimali Marian adaları', 'bal-Latn': 'Shemáli Máriáná Islánd', 'bas': 'Bìòn bi Marìanà ŋ̀ɔmbɔk', 'be': 'Паўночныя Марыянскія астравы', 'be-tarask': 'Паўночныя Марыянскія астравы', 'bez': 'Ifisima fya Mariana fya Hukaskazini', 'bg': 'Северни Мариански острови', 'blo': 'Mariyan kaBʊtukǝltǝna gʊnyɩpɛnɛlaŋ', 'bm': 'Kɛɲɛka Mariyani Gun', 'bn': 'উত্তরাঞ্চলীয় মারিয়ানা দ্বীপপুঞ্জ', 'br': 'Inizi Mariana an Norzh', 'brx': 'साहायारि मारियाना द्वीप', 'bs': 'Sjeverna Marijanska ostrva', 'bs-Cyrl': 'Сјеверна Маријанска Острва', 'bs-Latn': 'Sjeverna Marijanska ostrva', 'ca': 'Illes Marianes del Nord', 'ca-ES-valencia': 'Illes Marianes del Nord', 'ccp': '𑄅𑄪𑄖𑄴𑄖𑄮𑄉𑄎𑄢𑄴 𑄟𑄢𑄨𑄠𑄚 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Къилбаседа Марианан гӀайренаш', 'ceb': 'Northern Mariana Islands', 'cgg': 'Ebizinga by’amatemba ga Mariana', 'chr': 'ᏧᏴᏢ ᏗᏜ ᎹᎵᎠᎾ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی ماریانای باکوور', 'cs': 'Severní Mariany', 'cv': 'Ҫурҫӗр Мариан утравӗсем', 'cy': 'Ynysoedd Gogledd Mariana', 'da': 'Nordmarianerne', 'dav': 'Visiwa vya Mariana vya Kaskazini', 'de': 'Nördliche Marianen', 'dje': 'Mariana Gurma Gungey', 'dsb': 'Pódpołnocne Mariany', 'dz': 'བྱང་ཕྱོགས་ཀྱི་མ་ར་ཡ་ན་གླིང་ཚོམ', 'ebu': 'Visiwa vya Mariana vya Kaskazini', 'ee': 'Dziehe Marina ƒudomekpowo nutome', 'el': 'Νήσοι Βόρειες Μαριάνες', 'el-polyton': 'Νῆσοι Βόρειες Μαριάνες', 'en': 'Northern Mariana Islands', 'en-Dsrt': '𐐤𐐱𐑉𐑄𐐲𐑉𐑌 𐐣𐐰𐑉𐐨𐐱𐑌𐐲 𐐌𐑊𐐲𐑌𐐼𐑆', 'eo': 'Nord-Marianoj', 'es': 'Islas Marianas del Norte', 'et': 'Põhja-Mariaanid', 'eu': 'Ipar Mariana uharteak', 'ewo': 'Minlán Mi Marián yá Nór', 'fa': 'جزایر ماریانای شمالی', 'ff': 'Duuɗe Mariyaana Rewo', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤃𞤢𞤪𞤭𞤴𞤢𞥄𞤲𞤢 𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤭', 'ff-Latn': 'Duuɗe Mariyaana Rewo', 'fi': 'Pohjois-Mariaanit', 'fil': 'Northern Mariana Islands', 'fo': 'Norðaru Mariuoyggjar', 'fr': 'Îles Mariannes du Nord', 'fur': 'Isulis Mariana dal Nord', 'fy': 'Noardlike Marianeneilannen', 'ga': 'Na hOileáin Mháirianacha Thuaidh', 'gd': 'Na h-Eileanan Mairianach a Tuath', 'gl': 'Illas Marianas do Norte', 'gsw': 'Nördlichi Mariaane', 'gu': 'ઉત્તરી મારિયાના આઇલેન્ડ્સ', 'guz': 'Visiwa vya Mariana vya Kaskazini', 'ha': 'Tsibiran Mariyana Na Arewa', 'he': 'איי מריאנה הצפוניים', 'hi': 'उत्तरी मारियाना द्वीपसमूह', 'hi-Latn': 'Northern Mariana Islands', 'hr': 'Sjevernomarijanski Otoci', 'hsb': 'Sewjerne Mariany', 'ht': 'Îles Mariannes du Nord', 'hu': 'Északi Mariana-szigetek', 'hy': 'Հյուսիսային Մարիանյան կղզիներ', 'ia': 'Insulas Marianna del Nord', 'id': 'Kepulauan Mariana Utara', 'ig': 'Agwaetiti Northern Mariana', 'is': 'Norður-Maríanaeyjar', 'it': 'Isole Marianne Settentrionali', 'ja': '北マリアナ諸島', 'jmc': 'Visiwa vya Mariana vya Kaskazini', 'jv': 'Kapuloan Mariana Lor', 'ka': 'ჩრდილოეთ მარიანას კუნძულები', 'kab': 'Tigzirin n Maryan Ufella', 'kam': 'Visiwa vya Mariana vya Kaskazini', 'kde': 'Chisiwa cha Marian cha Kasikazini', 'kea': 'Ilhas Marianas di Norti', 'kgp': 'Nãrti-Mỹrijỹnỹ Goj-vẽso', 'khq': 'Mariana Gurma Gungey', 'ki': 'Visiwa vya Mariana vya Kaskazini', 'kk': 'Солтүстік Мариана аралдары', 'kk-Cyrl': 'Солтүстік Мариана аралдары', 'kln': 'Ikwembeiyotab Mariana nebo murot katam', 'km': 'កោះ\u200bម៉ារីណា\u200bខាង\u200bជើង', 'kn': 'ಉತ್ತರ ಮರಿಯಾನಾ ದ್ವೀಪಗಳು', 'ko': '북마리아나제도', 'kok': 'उत्तरी मरिना आयसलैण्ड', 'kok-Deva': 'उत्तरी मरिना आयसलैण्ड', 'ks': 'شُمٲلی مارِیانا جٔزیٖرٕ', 'ks-Arab': 'شُمٲلی مارِیانا جٔزیٖرٕ', 'ksb': 'Visiwa vya Maliana vya Kaskazini', 'ksf': 'zǝ maryánnɔ́r', 'ksh': 'de nöödlijje Marijanne-Enselle', 'ku': 'Giravên Bakurê Marianan', 'kxv': 'uttar mariyana dīp', 'kxv-Deva': 'उतॉरॉ मारियाना द्वीप', 'kxv-Latn': 'uttar mariyana dīp', 'kxv-Orya': 'ଉତର ମାରିୟାନ ଦିପ', 'kxv-Telu': 'ఉత్తర మరియానా దీవులు', 'ky': 'Түндүк Мариана аралдары', 'lag': 'Visiwa vya Mariana vya Kaskazini', 'lb': 'Nërdlech Marianen', 'lg': 'Bizinga bya Mariyana eby’omumambuka', 'lld': 'Isoles Marianes dl Nord', 'ln': 'Bisanga bya Marianɛ ya nɔ́rdi', 'lo': 'ໝູ່ເກາະມາເຣຍນາຕອນເໜືອ', 'lt': 'Marianos Šiaurinės Salos', 'lu': 'Lutanda lua Mariane wa muulu', 'luo': 'Northern Mariana Islands', 'luy': 'Visiwa vya Mariana vya Kaskazini', 'lv': 'Ziemeļu Marianas salas', 'mai': 'उत्तर मरीना द्वीपसमूह', 'mas': 'Visiwa vya Mariana vya Kaskazini', 'mer': 'Aĩrandi cia Mariana ya nothi', 'mfe': 'Zil Maryann dinor', 'mg': 'Nosy Mariana Atsinanana', 'mi': 'Ngā Motu Mariana ki te Raki', 'mk': 'Северни Маријански Острови', 'ml': 'ഉത്തര മറിയാനാ ദ്വീപുകൾ', 'mn': 'Хойд Марианы арлууд', 'mr': 'उत्तरी मारियाना बेटे', 'ms': 'Kepulauan Mariana Utara', 'mt': 'Ġżejjer Mariana tat-Tramuntana', 'mua': 'Sǝr Maria ma laŋne', 'my': 'မြောက်ပိုင်းမာရီအာနာကျွန်းစု', 'mzn': 'شمالی ماریانا جزایر', 'naq': 'Northern Mariana Islands', 'nb': 'Nord-Marianene', 'nd': 'Northern Mariana Islands', 'ne': 'उत्तरी मारिआना टापु', 'nl': 'Noordelijke Marianen', 'nmg': 'Maŋ́ Mariá', 'nn': 'Nord-Marianane', 'no': 'Nord-Marianene', 'nqo': 'ߡߊߙߌߦߊߣ ߞߐ߬ߘߎ߮ ߕߌ߲', 'nyn': 'Ebizinga by’amatemba ga Mariana', 'om': 'Odola Maariyaanaa Kaabaa', 'or': 'ଉତ୍ତର ମାରିଆନା ଦ୍ୱୀପପୁଞ୍ଜ', 'pa': 'ਉੱਤਰੀ ਮਾਰੀਆਨਾ ਟਾਪੂ', 'pa-Guru': 'ਉੱਤਰੀ ਮਾਰੀਆਨਾ ਟਾਪੂ', 'pcm': 'Nọ́tán Mariána Aílands', 'pl': 'Mariany Północne', 'ps': 'شمالي ماريانا ټاپوګان', 'pt': 'Ilhas Marianas do Norte', 'qu': 'Islas Marianas del Norte', 'rm': 'Inslas Mariannas dal Nord', 'rn': 'Amazinga ya Mariyana ryo mu majaruguru', 'ro': 'Insulele Mariane de Nord', 'rof': 'Visiwa vya Mariana vya Kaskazini', 'ru': 'Северные Марианские о-ва', 'rwk': 'Visiwa vya Mariana vya Kaskazini', 'saq': 'Visiwa vya Mariana vya Kaskazini', 'sbp': 'Figunguli fya Maliyana ifya luvala lwa Kunyamande', 'sc': 'Ìsulas Mariannas setentrionales', 'sd': 'اتريان ماريانا ٻيٽ', 'sd-Arab': 'اتريان ماريانا ٻيٽ', 'se': 'Davvi-Mariánat', 'seh': 'Ilhas Marianas do Norte', 'ses': 'Mariana Gurma Gungey', 'sg': 'Âzûâ Märïâni tî Banga', 'shi': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵎⴰⵔⵢⴰⵏ ⵏ ⵉⵥⵥⵍⵎⴹ', 'shi-Latn': 'tigzirin n maryan n iẓẓlmḍ', 'shi-Tfng': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵎⴰⵔⵢⴰⵏ ⵏ ⵉⵥⵥⵍⵎⴹ', 'si': 'උතුරු මරියානා දූපත්', 'sk': 'Severné Mariány', 'sl': 'Severni Marianski otoki', 'smn': 'Tave-Marianeh', 'sn': 'Zvitsuwa zvekumaodzanyemba eMariana', 'so': 'Jasiiradda Waqooyiga Mariaana', 'sq': 'Ishujt e Marianës Veriore', 'sr': 'Северна Маријанска Острва', 'sr-Cyrl': 'Северна Маријанска Острва', 'sr-Latn': 'Severna Marijanska Ostrva', 'sv': 'Nordmarianerna', 'sw': 'Visiwa vya Mariana vya Kaskazini', 'syr': 'ܓܙܝܖ̈ܐ ܕܡܪܝܢܐ ܓܪܒܝܐ', 'ta': 'வடக்கு மரியானா தீவுகள்', 'te': 'ఉత్తర మరియానా దీవులు', 'teo': 'Visiwa vya Mariana vya Kaskazini', 'tg': 'Ҷазираҳои Марианаи Шимолӣ', 'th': 'หมู่เกาะนอร์เทิร์นมาเรียนา', 'ti': 'ሰሜናዊ ደሴታት ማርያና', 'tk': 'Demirgazyk Mariana adalary', 'to': 'ʻOtumotu Maliana tokelau', 'tr': 'Kuzey Mariana Adaları', 'tt': 'Төньяк Мариана утраулары', 'twq': 'Mariana Gurma Gungey', 'tzm': 'Tigzirin n Maryana Tugafat', 'ug': 'شىمالىي مارىيانا ئاراللىرى', 'uk': 'Північні Маріанські Острови', 'ur': 'شمالی ماریانا آئلینڈز', 'uz': 'Shimoliy Mariana orollari', 'uz-Cyrl': 'Шимолий Марианна ороллари', 'uz-Latn': 'Shimoliy Mariana orollari', 'vai': 'ꗛꔤ ꕪꘋꗒ ꕮꔸꕩꕯ ꔳꘋꗣ', 'vai-Latn': 'Kɔi Kaŋndɔ Mariyana Tiŋŋɛ', 'vai-Vaii': 'ꗛꔤ ꕪꘋꗒ ꕮꔸꕩꕯ ꔳꘋꗣ', 'vec': 'Ìzole Mariane setentrionali', 'vi': 'Quần đảo Bắc Mariana', 'vun': 'Visiwa vya Mariana vya Kaskazini', 'wae': 'Nördliči Mariane', 'wo': 'Duni Mariyaan Noor', 'xh': 'ENorthern Mariana Islands', 'xnr': 'उत्तरी मारियाना द्वीपां दा झुंड', 'xog': 'Bizinga bya Mariyana ebyamumambuka', 'yav': 'il maliyanɛ u muɛnɛ́', 'yo': 'Etikun Guusu Mariana', 'yrl': 'Kapuãma-ita Mariyãna Nuti suí', 'yue': '北馬里亞納群島', 'yue-Hans': '北马里亚纳群岛', 'yue-Hant': '北馬里亞納群島', 'zgh': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵎⴰⵔⵢⴰⵏ ⵏ ⵉⵥⵥⵍⵎⴹ', 'zh': '北马里亚纳群岛', 'zh-Hans': '北马里亚纳群岛', 'zh-Hant': '北馬利安納群島', 'zu': 'i-Northern Mariana Islands'},
    'MQ': {'af': 'Martinique', 'agq': 'Màtìnekì', 'ak': 'Matinik', 'am': 'ማርቲኒክ', 'ar': 'جزر المارتينيك', 'as': 'মাৰ্টিনিক', 'asa': 'Martiniki', 'ast': 'La Martinica', 'az': 'Martinik', 'az-Cyrl': 'Мартиник', 'az-Latn': 'Martinik', 'bal-Latn': 'Mártinik', 'bas': 'Màrtìnîk', 'be': 'Марцініка', 'be-tarask': 'Марцініка', 'bez': 'Humartiniki', 'bg': 'Мартиника', 'blo': 'Martiniiki', 'bm': 'Maritiniki', 'bn': 'মার্টিনিক', 'br': 'Martinik', 'brx': 'मार्तिनिक', 'bs': 'Martinik', 'bs-Cyrl': 'Мартиник', 'bs-Latn': 'Martinik', 'ca': 'Martinica', 'ca-ES-valencia': 'Martinica', 'ccp': '𑄟𑄢𑄴𑄑𑄨𑄚𑄨𑄇𑄴', 'ce': 'Мартиника', 'ceb': 'Martinique', 'cgg': 'Martinique', 'chr': 'ᎹᏘᏂᎨ', 'ckb': 'مارتینیک', 'cs': 'Martinik', 'cv': 'Мартиника', 'cy': 'Martinique', 'da': 'Martinique', 'dav': 'Martiniki', 'de': 'Martinique', 'dje': 'Martiniiki', 'dsb': 'Martinique', 'dz': 'མཱར་ཊི་ནིཀ', 'ebu': 'Martiniki', 'ee': 'Martiniki nutome', 'el': 'Μαρτινίκα', 'el-polyton': 'Μαρτινίκα', 'en': 'Martinique', 'en-Dsrt': '𐐣𐐪𐑉𐐻𐑌𐐨𐐿', 'eo': 'Martiniko', 'es': 'Martinica', 'et': 'Martinique', 'eu': 'Martinika', 'ewo': 'Marətiníg', 'fa': 'مارتینیک', 'ff': 'Martinik', 'ff-Adlm': '𞤃𞤢𞤪𞤼𞤭𞤲𞤭𞤳𞤢𞥄', 'ff-Latn': 'Martinik', 'fi': 'Martinique', 'fil': 'Martinique', 'fo': 'Martinique', 'fr': 'Martinique', 'fur': 'Martiniche', 'fy': 'Martinique', 'ga': 'Martinique', 'gd': 'Mairtinic', 'gl': 'Martinica', 'gsw': 'Martinigg', 'gu': 'માર્ટીનીક', 'guz': 'Martiniki', 'ha': 'Martinik', 'he': 'מרטיניק', 'hi': 'मार्टीनिक', 'hi-Latn': 'Martinique', 'hr': 'Martinik', 'hsb': 'Martinique', 'ht': 'Martinique', 'hu': 'Martinique', 'hy': 'Մարտինիկա', 'ia': 'Martinica', 'id': 'Martinik', 'ie': 'Martinica', 'ig': 'Martinique', 'is': 'Martiník', 'it': 'Martinica', 'ja': 'マルティニーク', 'jmc': 'Martiniki', 'jv': 'Martinik', 'ka': 'მარტინიკა', 'kaa': 'Мартиника', 'kaa-Cyrl': 'Мартиника', 'kab': 'Martinik', 'kam': 'Martiniki', 'kde': 'Malitiniki', 'kea': 'Martinika', 'kgp': 'Mỹrtinĩka', 'khq': 'Martiniiki', 'ki': 'Martiniki', 'kk': 'Мартиника', 'kk-Cyrl': 'Мартиника', 'kln': 'Emetab Martinique', 'km': 'ម៉ាទីនីក', 'kn': 'ಮಾರ್ಟಿನಿಕ್', 'ko': '마르티니크', 'kok': 'मार्टीनिक', 'kok-Deva': 'मार्टीनिक', 'ks': 'مارٹِنِک', 'ks-Arab': 'مارٹِنِک', 'ksb': 'Maltiniki', 'ksf': 'matiník', 'ksh': 'Machtinik', 'ku': 'Martînîk', 'kxv': 'marṭinik', 'kxv-Deva': 'मार्टीनिक', 'kxv-Latn': 'marṭinik', 'kxv-Orya': 'ମାର୍ଟିନିକ୍', 'kxv-Telu': 'మార్ర్టనీక్', 'ky': 'Мартиника', 'lag': 'Maritiníiki', 'lb': 'Martinique', 'lg': 'Maritiniiki', 'lld': 'Martinica', 'ln': 'Martiniki', 'lo': 'ມາຕິນີກ', 'lt': 'Martinika', 'lu': 'Martiniki', 'luo': 'Martinique', 'luy': 'Martiniki', 'lv': 'Martinika', 'mai': 'मार्टिनिक', 'mas': 'Martiniki', 'mer': 'Martiniki', 'mfe': 'Martinik', 'mg': 'Martinika', 'mi': 'Mātiniki', 'mk': 'Мартиник', 'ml': 'മാർട്ടിനിക്ക്', 'mn': 'Мартиник', 'mr': 'मार्टिनिक', 'ms': 'Martinique', 'mt': 'Martinique', 'mua': 'Martinika', 'my': 'မာတီနိခ်', 'mzn': 'مارتینیک جزیره\u200cئون', 'naq': 'Martinique', 'nb': 'Martinique', 'nd': 'Martinique', 'ne': 'मार्टिनिक', 'nl': 'Martinique', 'nmg': 'Martinika', 'nn': 'Martinique', 'no': 'Martinique', 'nqo': 'ߡߊߙߑߕߣߌߞ', 'nyn': 'Martinique', 'om': 'Martinikuwee', 'or': 'ମାର୍ଟିନିକ୍ୟୁ', 'pa': 'ਮਾਰਟੀਨਿਕ', 'pa-Guru': 'ਮਾਰਟੀਨਿਕ', 'pcm': 'Matínik', 'pl': 'Martynika', 'ps': 'مارټینیک', 'pt': 'Martinica', 'qu': 'Martinica', 'rm': 'Martinique', 'rn': 'Maritiniki', 'ro': 'Martinica', 'rof': 'Martiniki', 'ru': 'Мартиника', 'rwk': 'Martiniki', 'saq': 'Martiniki', 'sbp': 'Malitiniki', 'sc': 'Martinica', 'sd': 'مارتينڪ', 'sd-Arab': 'مارتينڪ', 'se': 'Martinique', 'seh': 'Martinica', 'ses': 'Martiniiki', 'sg': 'Märtïnîki', 'shi': 'ⵎⴰⵔⵜⵉⵏⵉⴽ', 'shi-Latn': 'martinik', 'shi-Tfng': 'ⵎⴰⵔⵜⵉⵏⵉⴽ', 'si': 'මර්ටිනික්', 'sk': 'Martinik', 'sl': 'Martinik', 'smn': 'Martinique', 'sn': 'Martinique', 'so': 'Maartinik', 'sq': 'Martinikë', 'sr': 'Мартиник', 'sr-Cyrl': 'Мартиник', 'sr-Latn': 'Martinik', 'sv': 'Martinique', 'sw': 'Martinique', 'syr': 'ܡܐܪܬܝܢܝܩ', 'ta': 'மார்டினிக்', 'te': 'మార్టినీక్', 'teo': 'Martiniki', 'tg': 'Мартиника', 'th': 'มาร์ตินีก', 'ti': 'ማርቲኒክ', 'tk': 'Martinika', 'to': 'Mātiniki', 'tr': 'Martinik', 'tt': 'Мартиника', 'twq': 'Martiniiki', 'tzm': 'Martinik', 'ug': 'مارتىنىكا', 'uk': 'Мартиніка', 'ur': 'مارٹینک', 'uz': 'Martinika', 'uz-Cyrl': 'Мартиника', 'uz-Latn': 'Martinika', 'vai': 'ꕮꔳꕇꕃ', 'vai-Latn': 'Matiniki', 'vai-Vaii': 'ꕮꔳꕇꕃ', 'vec': 'Martiniga', 'vi': 'Martinique', 'vun': 'Martiniki', 'wae': 'Martinique', 'wo': 'Martinik', 'xh': 'EMartinique', 'xnr': 'मार्टीनिक', 'xog': 'Maritiniiki', 'yav': 'maltiníik', 'yi': 'מאַרטיניק', 'yo': 'Matinikuwi', 'yrl': 'Matinika', 'yue': '馬丁尼克島', 'yue-Hans': '马丁尼克岛', 'yue-Hant': '馬丁尼克島', 'zgh': 'ⵎⴰⵔⵜⵉⵏⵉⴽ', 'zh': '马提尼克', 'zh-Hans': '马提尼克', 'zh-Hant': '馬丁尼克', 'zu': 'i-Martinique'},
    'MR': {'af': 'Mauritanië', 'agq': 'Mùlètanyìa', 'ak': 'Mɔretenia', 'am': 'ሞሪቴኒያ', 'ar': 'موريتانيا', 'as': 'মাউৰিটানিয়া', 'asa': 'Moritania', 'ast': 'Mauritania', 'az': 'Mavritaniya', 'az-Cyrl': 'Мавританија', 'az-Latn': 'Mavritaniya', 'bal-Latn': 'Muritániá', 'bas': 'Mòrìtanìà', 'be': 'Маўрытанія', 'be-tarask': 'Маўрытанія', 'bez': 'Humolitania', 'bg': 'Мавритания', 'blo': 'Moritanii', 'bm': 'Mɔritani', 'bn': 'মরিতানিয়া', 'br': 'Maouritania', 'brx': 'माउरिटानिया', 'bs': 'Mauritanija', 'bs-Cyrl': 'Мауританија', 'bs-Latn': 'Mauritanija', 'ca': 'Mauritània', 'ca-ES-valencia': 'Mauritània', 'ccp': '𑄟𑄧𑄢𑄨𑄖𑄚𑄨𑄠', 'ce': 'Мавритани', 'ceb': 'Mauritania', 'cgg': 'Mauriteeniya', 'chr': 'ᎹᏘᎢᏯ', 'ckb': 'مۆریتانیا', 'cs': 'Mauritánie', 'cv': 'Мавритани', 'cy': 'Mauritania', 'da': 'Mauretanien', 'dav': 'Moritania', 'de': 'Mauretanien', 'dje': 'Mooritaani', 'dsb': 'Mawretańska', 'dz': 'མོ་རི་ཊེ་ནི་ཡ', 'ebu': 'Moritania', 'ee': 'Mauritania nutome', 'el': 'Μαυριτανία', 'el-polyton': 'Μαυριτανία', 'en': 'Mauritania', 'en-Dsrt': '𐐣𐐱𐑉𐐲𐐻𐐩𐑌𐐨𐐲', 'eo': 'Maŭritanujo', 'es': 'Mauritania', 'et': 'Mauritaania', 'eu': 'Mauritania', 'ewo': 'Moritaní', 'fa': 'موریتانی', 'ff': 'Muritani', 'ff-Adlm': '𞤃𞤮𞤪𞤼𞤢𞤲𞤭𞥅', 'ff-Latn': 'Muritani', 'fi': 'Mauritania', 'fil': 'Mauritania', 'fo': 'Móritania', 'fr': 'Mauritanie', 'fur': 'Mauritanie', 'fy': 'Mauritanië', 'ga': 'An Mháratái', 'gd': 'Moratàinea', 'gl': 'Mauritania', 'gsw': 'Mauretaanie', 'gu': 'મૌરિટાનિયા', 'guz': 'Moritania', 'ha': 'Moritaniya', 'he': 'מאוריטניה', 'hi': 'मॉरिटानिया', 'hi-Latn': 'Mauritania', 'hr': 'Mauretanija', 'hsb': 'Mawretanska', 'ht': 'Mauritanie', 'hu': 'Mauritánia', 'hy': 'Մավրիտանիա', 'ia': 'Mauritania', 'id': 'Mauritania', 'ig': 'Mauritania', 'is': 'Máritanía', 'it': 'Mauritania', 'ja': 'モーリタニア', 'jgo': 'Mɔlitanî', 'jmc': 'Moritania', 'jv': 'Mauritania', 'ka': 'მავრიტანია', 'kaa': 'Мавритания', 'kaa-Cyrl': 'Мавритания', 'kab': 'Muriṭanya', 'kam': 'Moritania', 'kde': 'Molitania', 'kea': 'Mauritánia', 'kgp': 'Mãritỹnĩja', 'khq': 'Mooritaani', 'ki': 'Moritania', 'kk': 'Мавритания', 'kk-Cyrl': 'Мавритания', 'kln': 'Emetab Mauritania', 'km': 'ម៉ូរីតានី', 'kn': 'ಮಾರಿಟೇನಿಯಾ', 'ko': '모리타니', 'kok': 'मॉरिटानिया', 'kok-Deva': 'मॉरिटानिया', 'ks': 'مارٕٹانِیا', 'ks-Arab': 'مارٕٹانِیا', 'ksb': 'Maulitania', 'ksf': 'mwaritaní', 'ksh': 'Mautitaanije', 'ku': 'Morîtanya', 'kxv': 'mauriṭaniya', 'kxv-Deva': 'मॉरिटानिया', 'kxv-Latn': 'mauriṭaniya', 'kxv-Orya': 'ମଉରିଟାନିୟା', 'kxv-Telu': 'మౌరిటేనియా', 'ky': 'Мавритания', 'lag': 'Moritánia', 'lb': 'Mauretanien', 'lg': 'Mawulitenya', 'lld': 'Mauritania', 'ln': 'Moritani', 'lo': 'ມົວຣິເທເນຍ', 'lt': 'Mauritanija', 'lu': 'Moritani', 'luo': 'Mauritania', 'luy': 'Moritania', 'lv': 'Mauritānija', 'mai': 'मारिटानिया', 'mas': 'Moritania', 'mer': 'Mauritania', 'mfe': 'Moritani', 'mg': 'Maoritania', 'mi': 'Mauritānia', 'mk': 'Мавританија', 'ml': 'മൗറിറ്റാനിയ', 'mn': 'Мавритани', 'mr': 'मॉरिटानिया', 'ms': 'Mauritania', 'mt': 'il-Mauritania', 'mua': 'Mauritaniya', 'my': 'မော်ရီတေးနီးယား', 'mzn': 'موریتانی', 'naq': 'Mauritania', 'nb': 'Mauritania', 'nd': 'Mauritania', 'ne': 'माउरिटानिया', 'nl': 'Mauritanië', 'nmg': 'Moritania', 'nn': 'Mauritania', 'no': 'Mauritania', 'nqo': 'ߡߏߙߌߕߊߣߌ߫', 'nyn': 'Mauriteeniya', 'om': 'Mawuritaaniyaa', 'or': 'ମୌରିଟାନିଆ', 'pa': 'ਮੋਰਿਟਾਨੀਆ', 'pa-Guru': 'ਮੋਰਿਟਾਨੀਆ', 'pcm': 'Mọriténia', 'pl': 'Mauretania', 'ps': 'موریتانیا', 'pt': 'Mauritânia', 'qu': 'Mauritania', 'rm': 'Mauretania', 'rn': 'Moritaniya', 'ro': 'Mauritania', 'rof': 'Moritania', 'ru': 'Мавритания', 'rwk': 'Moritania', 'saq': 'Moritania', 'sbp': 'Molitaniya', 'sc': 'Mauritània', 'sd': 'موريتانيا', 'sd-Arab': 'موريتانيا', 'se': 'Mauretánia', 'seh': 'Mauritânia', 'ses': 'Mooritaani', 'sg': 'Moritanïi', 'shi': 'ⵎⵓⵕⵉⵟⴰⵏⵢⴰ', 'shi-Latn': 'muṛiṭanya', 'shi-Tfng': 'ⵎⵓⵕⵉⵟⴰⵏⵢⴰ', 'si': 'මොරිටේනියාව', 'sk': 'Mauritánia', 'sl': 'Mavretanija', 'smn': 'Mauritania', 'sn': 'Mauritania', 'so': 'Muritaaniya', 'sq': 'Mauritani', 'sr': 'Мауританија', 'sr-Cyrl': 'Мауританија', 'sr-Latn': 'Mauritanija', 'sv': 'Mauretanien', 'sw': 'Moritania', 'syr': 'ܡܘܪܝܛܢܝܐ', 'ta': 'மௌரிடானியா', 'te': 'మారిటేనియా', 'teo': 'Moritania', 'tg': 'Мавритания', 'th': 'มอริเตเนีย', 'ti': 'ማውሪታንያ', 'tk': 'Mawritaniýa', 'to': 'Maulitenia', 'tr': 'Moritanya', 'tt': 'Мавритания', 'twq': 'Mooritaani', 'tzm': 'Muritanya', 'ug': 'ماۋرىتانىيە', 'uk': 'Мавританія', 'ur': 'موریطانیہ', 'uz': 'Mavritaniya', 'uz-Cyrl': 'Мавритания', 'uz-Latn': 'Mavritaniya', 'vai': 'ꗞꔓꔎꕇꕰ', 'vai-Latn': 'Mɔretaninya', 'vai-Vaii': 'ꗞꔓꔎꕇꕰ', 'vec': 'Mauritania', 'vi': 'Mauritania', 'vun': 'Moritania', 'wae': 'Mauretanie', 'wo': 'Mooritani', 'xh': 'EMauritania', 'xnr': 'मॉरिटानिया', 'xog': 'Mawulitenya', 'yav': 'molitaní', 'yi': 'מאַריטאַניע', 'yo': 'Maritania', 'yrl': 'Mauritaniya', 'yue': '茅利塔尼亞', 'yue-Hans': '茅利塔尼亚', 'yue-Hant': '茅利塔尼亞', 'zgh': 'ⵎⵓⵕⵉⵟⴰⵏⵢⴰ', 'zh': '毛里塔尼亚', 'zh-Hans': '毛里塔尼亚', 'zh-Hant': '茅利塔尼亞', 'zu': 'i-Mauritania'},
    'MS': {'af': 'Montserrat', 'agq': 'Mùŋtselà', 'ak': 'Mantserat', 'am': 'ሞንትሴራት', 'ar': 'مونتسرات', 'as': 'ম’ণ্টছেৰাট', 'asa': 'Monttherrati', 'ast': 'Montserrat', 'az': 'Monserat', 'az-Cyrl': 'Монсерат', 'az-Latn': 'Monserat', 'bal-Latn': 'Montserrat', 'bas': 'Mɔ̀ŋseràt', 'be': 'Мантсерат', 'be-tarask': 'Мантсерат', 'bez': 'Humontserrati', 'bg': 'Монтсерат', 'blo': 'Mɔnsɛraatɩ', 'bm': 'Moŋsera', 'bn': 'মন্টসেরাট', 'br': 'Montserrat', 'brx': 'मनत्सेरात', 'bs': 'Monserat', 'bs-Cyrl': 'Монсерат', 'bs-Latn': 'Monserat', 'ca': 'Montserrat', 'ca-ES-valencia': 'Montserrat', 'ccp': '𑄟𑄧𑄚𑄴𑄑𑄴𑄥𑄬𑄢𑄑𑄴', 'ce': 'Монтсеррат', 'ceb': 'Montserrat', 'cgg': 'Montserrati', 'chr': 'ᎹᏂᏘᏌᎳᏗ', 'ckb': 'مۆنتسێرات', 'cs': 'Montserrat', 'cv': 'Монтсеррат', 'cy': 'Montserrat', 'da': 'Montserrat', 'dav': 'Montserrati', 'de': 'Montserrat', 'dje': 'Montserrat', 'dsb': 'Montserrat', 'dz': 'མོན་ས་རཊ', 'ebu': 'Montserrati', 'ee': 'Montserrat nutome', 'el': 'Μονσεράτ', 'el-polyton': 'Μονσεράτ', 'en': 'Montserrat', 'en-Dsrt': '𐐣𐐪𐑌𐐻𐑅𐐲𐑉𐐪𐐻', 'es': 'Montserrat', 'et': 'Montserrat', 'eu': 'Montserrat', 'ewo': 'Mɔ́ntserád', 'fa': 'مونت\u200cسرات', 'ff': 'Monseraat', 'ff-Adlm': '𞤃𞤮𞤲𞤧𞤭𞤪𞤢𞥄𞤼', 'ff-Latn': 'Monseraat', 'fi': 'Montserrat', 'fil': 'Montserrat', 'fo': 'Montserrat', 'fr': 'Montserrat', 'fur': 'Montserrat', 'fy': 'Montserrat', 'ga': 'Montsarat', 'gd': 'Montsarat', 'gl': 'Montserrat', 'gsw': 'Moosörrat', 'gu': 'મોંટસેરાત', 'guz': 'Montserrati', 'ha': 'Manserati', 'he': 'מונסראט', 'hi': 'मोंटसेरात', 'hi-Latn': 'Montserrat', 'hr': 'Montserrat', 'hsb': 'Montserrat', 'ht': 'Montserrat', 'hu': 'Montserrat', 'hy': 'Մոնսեռատ', 'ia': 'Montserrat', 'id': 'Montserrat', 'ig': 'Montserrat', 'is': 'Montserrat', 'it': 'Montserrat', 'ja': 'モントセラト', 'jmc': 'Montserrati', 'jv': 'Monsérat', 'ka': 'მონსერატი', 'kaa': 'Монтсеррат', 'kaa-Cyrl': 'Монтсеррат', 'kab': 'Munsirat', 'kam': 'Montserrati', 'kde': 'Monselati', 'kea': 'Monserat', 'kgp': 'Mãtisehati', 'khq': 'Montserrat', 'ki': 'Montserrati', 'kk': 'Монтсеррат', 'kk-Cyrl': 'Монтсеррат', 'kln': 'Emetab Montserrat', 'km': 'ម៉ុងស៊ែរ៉ា', 'kn': 'ಮಾಂಟ್\u200cಸೆರಟ್', 'ko': '몬트세라트', 'kok': 'मॉन्टसेराट', 'kok-Deva': 'मॉन्टसेराट', 'ks': 'مانٹسیراٹ', 'ks-Arab': 'مانٹسیراٹ', 'ksb': 'Montselati', 'ksf': 'mɔnsɛrat', 'ksh': 'Mongßerrat', 'ku': 'Montserat', 'kxv': 'monṭserrarṭ', 'kxv-Deva': 'मोंटसेरात', 'kxv-Latn': 'monṭserrarṭ', 'kxv-Orya': 'ମଣ୍ଟସେରେଟ', 'kxv-Telu': 'మాంట్సెరాట్', 'ky': 'Монтсеррат', 'lag': 'Monteráati', 'lb': 'Montserrat', 'lg': 'Monteseraati', 'lld': 'Montserrat', 'ln': 'Mɔsera', 'lo': 'ມອນເຊີຣາດ', 'lt': 'Montseratas', 'lu': 'Musera', 'luo': 'Montserrat', 'luy': 'Montserrati', 'lv': 'Montserrata', 'mai': 'मोंटसेरात', 'mas': 'Montserrati', 'mer': 'Montserrati', 'mfe': 'Montsera', 'mg': 'Montserrat', 'mi': 'Monoterā', 'mk': 'Монсерат', 'ml': 'മൊണ്ടെസരത്ത്', 'mn': 'Монтсеррат', 'mr': 'मॉन्ट्सेराट', 'ms': 'Montserrat', 'mt': 'Montserrat', 'mua': 'Sǝr Montserrat', 'my': 'မောင့်စဲရက်', 'mzn': 'مونتسرات', 'naq': 'Montserrat', 'nb': 'Montserrat', 'nd': 'Montserrat', 'ne': 'मोन्टसेर्राट', 'nl': 'Montserrat', 'nmg': 'Mɔnserrat', 'nn': 'Montserrat', 'no': 'Montserrat', 'nqo': 'ߡߐ߲ߗߋߙߊ߫', 'nyn': 'Montserrati', 'om': 'Montiseerat', 'or': 'ମଣ୍ଟେସେରାଟ୍', 'pa': 'ਮੋਂਟਸੇਰਾਤ', 'pa-Guru': 'ਮੋਂਟਸੇਰਾਤ', 'pcm': 'Mọntsẹrat', 'pl': 'Montserrat', 'ps': 'مانټیسیرت', 'pt': 'Montserrat', 'qu': 'Montserrat', 'rm': 'Montserrat', 'rn': 'Monteserati', 'ro': 'Montserrat', 'rof': 'Montserrati', 'ru': 'Монтсеррат', 'rwk': 'Montserrati', 'saq': 'Montserrati', 'sbp': 'Monitiselati', 'sc': 'Montserrat', 'sd': 'مونٽسراٽ', 'sd-Arab': 'مونٽسراٽ', 'se': 'Montserrat', 'seh': 'Montserrat', 'ses': 'Montserrat', 'sg': 'Monserâte', 'shi': 'ⵎⵓⵏⵙⵉⵔⴰⵜ', 'shi-Latn': 'munsirat', 'shi-Tfng': 'ⵎⵓⵏⵙⵉⵔⴰⵜ', 'si': 'මොන්සෙරාට්', 'sk': 'Montserrat', 'sl': 'Montserrat', 'smn': 'Montserrat', 'sn': 'Montserrat', 'so': 'Montserrat', 'sq': 'Montserat', 'sr': 'Монсерат', 'sr-Cyrl': 'Монсерат', 'sr-Latn': 'Monserat', 'sv': 'Montserrat', 'sw': 'Montserrat', 'syr': 'ܡܘܢܣܝܪܐܬ', 'ta': 'மாண்ட்செராட்', 'te': 'మాంట్సెరాట్', 'teo': 'Montserrati', 'tg': 'Монтсеррат', 'th': 'มอนต์เซอร์รัต', 'ti': 'ሞንትሰራት', 'tk': 'Monserrat', 'to': 'Moʻungaselati', 'tr': 'Montserrat', 'tt': 'Монтсеррат', 'twq': 'Montserrat', 'tzm': 'Muntsirra', 'ug': 'مونتسېررات', 'uk': 'Монтсеррат', 'ur': 'مونٹسیراٹ', 'uz': 'Montserrat', 'uz-Cyrl': 'Монтсеррат', 'uz-Latn': 'Montserrat', 'vai': 'ꗞꘋꔖꕟꔳ', 'vai-Latn': 'Mɔserati', 'vai-Vaii': 'ꗞꘋꔖꕟꔳ', 'vec': 'Montserat', 'vi': 'Montserrat', 'vun': 'Montserrati', 'wae': 'Monserrat', 'wo': 'Mooseraa', 'xh': 'EMontserrat', 'xnr': 'मोंटसेरात', 'xog': 'Monteseraati', 'yav': 'mɔŋsilá', 'yi': 'מאנטסעראַט', 'yo': 'Motserati', 'yrl': 'Mũtiserati', 'yue': '蒙哲臘', 'yue-Hans': '蒙哲腊', 'yue-Hant': '蒙哲臘', 'zgh': 'ⵎⵓⵏⵙⵉⵔⴰⵜ', 'zh': '蒙特塞拉特', 'zh-Hans': '蒙特塞拉特', 'zh-Hant': '蒙哲臘', 'zu': 'i-Montserrat'},
    'MT': {'af': 'Malta', 'agq': 'Maatà', 'ak': 'Mɔlta', 'am': 'ማልታ', 'ar': 'مالطا', 'as': 'মাল্টা', 'asa': 'Malta', 'ast': 'Malta', 'az': 'Malta', 'az-Cyrl': 'Малта', 'az-Latn': 'Malta', 'bal-Latn': 'Máltá', 'bas': 'Maltà', 'be': 'Мальта', 'be-tarask': 'Мальта', 'bez': 'Humalta', 'bg': 'Малта', 'bgn': 'مالته', 'blo': 'Malta', 'bm': 'Malti', 'bn': 'মাল্টা', 'br': 'Malta', 'brx': 'माल्टा', 'bs': 'Malta', 'bs-Cyrl': 'Малта', 'bs-Latn': 'Malta', 'ca': 'Malta', 'ca-ES-valencia': 'Malta', 'ccp': '𑄟𑄣𑄴𑄑', 'ce': 'Мальта', 'ceb': 'Malta', 'cgg': 'Marita', 'chr': 'ᎹᎵᏔ', 'ckb': 'ماڵتا', 'cs': 'Malta', 'cv': 'Мальта', 'cy': 'Malta', 'da': 'Malta', 'dav': 'Malta', 'de': 'Malta', 'dje': 'Malta', 'dsb': 'Malta', 'dz': 'མཱལ་ཊ', 'ebu': 'Malta', 'ee': 'Malta nutome', 'el': 'Μάλτα', 'el-polyton': 'Μάλτα', 'en': 'Malta', 'en-Dsrt': '𐐣𐐱𐑊𐐻𐐲', 'eo': 'Malto', 'es': 'Malta', 'et': 'Malta', 'eu': 'Malta', 'ewo': 'Málətə', 'fa': 'مالت', 'ff': 'Malte', 'ff-Adlm': '𞤃𞤢𞤤𞤼𞤢', 'ff-Latn': 'Malte', 'fi': 'Malta', 'fil': 'Malta', 'fo': 'Malta', 'fr': 'Malte', 'fur': 'Malta', 'fy': 'Malta', 'ga': 'Málta', 'gd': 'Malta', 'gl': 'Malta', 'gsw': 'Malta', 'gu': 'માલ્ટા', 'guz': 'Malta', 'ha': 'Malta', 'he': 'מלטה', 'hi': 'माल्टा', 'hi-Latn': 'Malta', 'hr': 'Malta', 'hsb': 'Malta', 'ht': 'Malte', 'hu': 'Málta', 'hy': 'Մալթա', 'ia': 'Malta', 'id': 'Malta', 'ie': 'Malta', 'ig': 'Malta', 'is': 'Malta', 'it': 'Malta', 'ja': 'マルタ', 'jmc': 'Malta', 'jv': 'Malta', 'ka': 'მალტა', 'kaa': 'Мальта', 'kaa-Cyrl': 'Мальта', 'kab': 'Malṭ', 'kam': 'Malta', 'kde': 'Malta', 'kea': 'Malta', 'kgp': 'Mỹrta', 'khq': 'Malta', 'ki': 'Malta', 'kk': 'Мальта', 'kk-Cyrl': 'Мальта', 'kln': 'Emetab Malta', 'km': 'ម៉ាល់ត៍', 'kn': 'ಮಾಲ್ಟಾ', 'ko': '몰타', 'kok': 'माल्टा', 'kok-Deva': 'माल्टा', 'ks': 'مالٹا', 'ks-Arab': 'مالٹا', 'ksb': 'Malta', 'ksf': 'maltǝ', 'ksh': 'Malta', 'ku': 'Malta', 'kxv': 'malṭaa', 'kxv-Deva': 'माल्टा', 'kxv-Latn': 'malṭaa', 'kxv-Orya': 'ମାଲ୍ଟା', 'kxv-Telu': 'మాల్డా', 'ky': 'Мальта', 'lag': 'Málita', 'lb': 'Malta', 'lg': 'Malita', 'lld': 'Malta', 'ln': 'Malitɛ', 'lo': 'ມອນທາ', 'lt': 'Malta', 'lu': 'Malite', 'luo': 'Malta', 'luy': 'Malta', 'lv': 'Malta', 'mai': 'माल्टा', 'mas': 'Malta', 'mer': 'Malta', 'mfe': 'Malt', 'mg': 'Malta', 'mi': 'Mārata', 'mk': 'Малта', 'ml': 'മാൾട്ട', 'mn': 'Мальта', 'mr': 'माल्टा', 'ms': 'Malta', 'mt': 'Malta', 'mua': 'Sǝr Malta', 'my': 'မောလ်တာ', 'mzn': 'مالت', 'naq': 'Malta', 'nb': 'Malta', 'nd': 'Malta', 'ne': 'माल्टा', 'nl': 'Malta', 'nmg': 'Malta', 'nn': 'Malta', 'no': 'Malta', 'nqo': 'ߡߊߟߑߕ', 'nyn': 'Marita', 'om': 'Maaltaa', 'or': 'ମାଲ୍ଟା', 'pa': 'ਮਾਲਟਾ', 'pa-Guru': 'ਮਾਲਟਾ', 'pcm': 'Mọ́lta', 'pl': 'Malta', 'ps': 'مالټا', 'pt': 'Malta', 'qu': 'Malta', 'rm': 'Malta', 'rn': 'Malita', 'ro': 'Malta', 'rof': 'Malta', 'ru': 'Мальта', 'rwk': 'Malta', 'saq': 'Malta', 'sbp': 'Malita', 'sc': 'Malta', 'sd': 'مالٽا', 'sd-Arab': 'مالٽا', 'se': 'Málta', 'seh': 'Malta', 'ses': 'Malta', 'sg': 'Mâlta', 'shi': 'ⵎⴰⵍⵟⴰ', 'shi-Latn': 'malṭa', 'shi-Tfng': 'ⵎⴰⵍⵟⴰ', 'si': 'මෝල්ටාව', 'sk': 'Malta', 'sl': 'Malta', 'smn': 'Malta', 'sn': 'Malta', 'so': 'Maalta', 'sq': 'Maltë', 'sr': 'Малта', 'sr-Cyrl': 'Малта', 'sr-Latn': 'Malta', 'sv': 'Malta', 'sw': 'Malta', 'syr': 'ܡܝܠܛܐ', 'ta': 'மால்டா', 'te': 'మాల్టా', 'teo': 'Malta', 'tg': 'Малта', 'th': 'มอลตา', 'ti': 'ማልታ', 'tk': 'Malta', 'to': 'Malita', 'tr': 'Malta', 'tt': 'Мальта', 'twq': 'Malta', 'tzm': 'Malṭa', 'ug': 'مالتا', 'uk': 'Мальта', 'ur': 'مالٹا', 'uz': 'Malta', 'uz-Cyrl': 'Мальта', 'uz-Latn': 'Malta', 'vai': 'ꕮꕊꕚ', 'vai-Latn': 'Malita', 'vai-Vaii': 'ꕮꕊꕚ', 'vec': 'Malta', 'vi': 'Malta', 'vun': 'Malta', 'wae': 'Malta', 'wo': 'Malt', 'xh': 'EMalta', 'xnr': 'माल्टा', 'xog': 'Malita', 'yav': 'málɛ́t', 'yi': 'מאַלטאַ', 'yo': 'Malata', 'yrl': 'Mauta', 'yue': '馬爾他', 'yue-Hans': '马尔他', 'yue-Hant': '馬爾他', 'zgh': 'ⵎⴰⵍⵟⴰ', 'zh': '马耳他', 'zh-Hans': '马耳他', 'zh-Hant': '馬爾他', 'zu': 'i-Malta'},
    'MU': {'af': 'Mauritius', 'agq': 'Mùleshwɨ̀s', 'ak': 'Mɔrehyeɔs', 'am': 'ሞሪሸስ', 'ar': 'موريشيوس', 'as': 'মৰিছাছ', 'asa': 'Morithi', 'ast': 'Mauriciu', 'az': 'Mavriki', 'az-Cyrl': 'Маврики', 'az-Latn': 'Mavriki', 'bal-Latn': 'Murishias', 'bas': 'Mòrîs', 'be': 'Маўрыкій', 'be-tarask': 'Маўрыкій', 'bez': 'Humolisi', 'bg': 'Мавриций', 'bgn': 'موریتانیا', 'blo': 'Imoris', 'bm': 'Morisi', 'bn': 'মরিশাস', 'br': 'Moris', 'brx': 'मरिसियास', 'bs': 'Mauricijus', 'bs-Cyrl': 'Маурицијус', 'bs-Latn': 'Mauricijus', 'ca': 'Maurici', 'ca-ES-valencia': 'Maurici', 'ccp': '𑄟𑄧𑄢𑄨𑄥𑄥𑄴', 'ce': 'Маврики', 'ceb': 'Mauritius', 'cgg': 'Maurishiasi', 'chr': 'ᎼᎵᏏᎥᏍ', 'ckb': 'مووریتیووس', 'cs': 'Mauricius', 'cv': 'Маврики', 'cy': 'Mauritius', 'da': 'Mauritius', 'dav': 'Morisi', 'de': 'Mauritius', 'dje': 'Mooris gungey', 'dsb': 'Mauritius', 'dz': 'མོ་རི་ཤཱས', 'ebu': 'Morisi', 'ee': 'mauritiusdukɔ', 'el': 'Μαυρίκιος', 'el-polyton': 'Μαυρίκιος', 'en': 'Mauritius', 'en-Dsrt': '𐐣𐐱𐑉𐐮𐑇𐐲𐑅', 'eo': 'Maŭricio', 'es': 'Mauricio', 'et': 'Mauritius', 'eu': 'Maurizio', 'ewo': 'Morís', 'fa': 'موریس', 'ff': 'Moriis', 'ff-Adlm': '𞤃𞤮𞤪𞤭𞥅𞤧𞤭', 'ff-Latn': 'Moriis', 'fi': 'Mauritius', 'fil': 'Mauritius', 'fo': 'Móritius', 'fr': 'Maurice', 'fur': 'Maurizi', 'fy': 'Mauritius', 'ga': 'Oileán Mhuirís', 'gd': 'Na h-Eileanan Mhoiriseas', 'gl': 'Mauricio', 'gsw': 'Mauriizius', 'gu': 'મોરિશિયસ', 'guz': 'Morisi', 'ha': 'Moritus', 'he': 'מאוריציוס', 'hi': 'मॉरीशस', 'hi-Latn': 'Mauritius', 'hr': 'Mauricijus', 'hsb': 'Mauritius', 'ht': 'Maurice', 'hu': 'Mauritius', 'hy': 'Մավրիկիոս', 'ia': 'Mauritio', 'id': 'Mauritius', 'ie': 'Mauricio', 'ig': 'Mauritius', 'is': 'Máritíus', 'it': 'Mauritius', 'ja': 'モーリシャス', 'jmc': 'Morisi', 'jv': 'Mauritius', 'ka': 'მავრიკი', 'kaa': 'Маврикий', 'kaa-Cyrl': 'Маврикий', 'kab': 'Muris', 'kam': 'Morisi', 'kde': 'Molisi', 'kea': 'Maurísia', 'kgp': 'Mãriso', 'khq': 'Mooris gungey', 'ki': 'Morisi', 'kk': 'Маврикий', 'kk-Cyrl': 'Маврикий', 'kln': 'Emetab Mauritius', 'km': 'ម៉ូរីស', 'kn': 'ಮಾರಿಷಸ್', 'ko': '모리셔스', 'kok': 'मॉरिशस', 'kok-Deva': 'मॉरिशस', 'ks': 'مورِشَس', 'ks-Arab': 'مورِشَس', 'ksb': 'Molisi', 'ksf': 'mwarís', 'ksh': 'Maurizijus', 'ku': 'Mauritius', 'kxv': 'mauriss', 'kxv-Deva': 'मॉरीसस', 'kxv-Latn': 'mauriss', 'kxv-Orya': 'ମରିସସ୍', 'kxv-Telu': 'మారిసస్', 'ky': 'Маврикий', 'lag': 'Moríisi', 'lb': 'Mauritius', 'lg': 'Mawulisyasi', 'lld': 'Mauritius', 'ln': 'Morisɛ', 'lo': 'ມົວຣິຊຽສ', 'lt': 'Mauricijus', 'lu': 'Morise', 'luo': 'Mauritius', 'luy': 'Morisi', 'lv': 'Maurīcija', 'mai': 'मारीशस', 'mas': 'Morisi', 'mer': 'Maurĩtiasi', 'mfe': 'Moris', 'mg': 'Maorisy', 'mi': 'Marihi', 'mk': 'Маврициус', 'ml': 'മൗറീഷ്യസ്', 'mn': 'Маврикий', 'mr': 'मॉरिशस', 'ms': 'Mauritius', 'mt': 'Mauritius', 'mua': 'Sǝr Mauricǝ', 'my': 'မောရစ်ရှ', 'mzn': 'مورى تيوس', 'naq': 'Mauritius', 'nb': 'Mauritius', 'nd': 'Mauritius', 'ne': 'मौरिसियस', 'nl': 'Mauritius', 'nmg': 'Morisse', 'nn': 'Mauritius', 'no': 'Mauritius', 'nqo': 'ߡߏߙߌߛ', 'nyn': 'Maurishiasi', 'om': 'Moorishiyees', 'or': 'ମରିସସ', 'pa': 'ਮੌਰੀਸ਼ਸ', 'pa-Guru': 'ਮੌਰੀਸ਼ਸ', 'pcm': 'Mọríshọs', 'pl': 'Mauritius', 'ps': 'موریشیس', 'pt': 'Maurício', 'qu': 'Mauricio', 'rm': 'Mauritius', 'rn': 'Izinga rya Morise', 'ro': 'Mauritius', 'rof': 'Morisi', 'ru': 'Маврикий', 'rwk': 'Morisi', 'saq': 'Morisi', 'sbp': 'Molisi', 'sc': 'Maurìtzius', 'sd': 'موريشس', 'sd-Arab': 'موريشس', 'se': 'Mauritius', 'seh': 'Maurício', 'ses': 'Mooris gungey', 'sg': 'Mörîsi', 'shi': 'ⵎⵓⵔⵉⵙ', 'shi-Latn': 'muris', 'shi-Tfng': 'ⵎⵓⵔⵉⵙ', 'si': 'මුරුසිය', 'sk': 'Maurícius', 'sl': 'Mauritius', 'smn': 'Mauritius', 'sn': 'Mauritius', 'so': 'Mawrishiyaas', 'sq': 'Mauritius', 'sr': 'Маурицијус', 'sr-Cyrl': 'Маурицијус', 'sr-Latn': 'Mauricijus', 'sv': 'Mauritius', 'sw': 'Morisi', 'syr': 'ܡܘܪܝܛܝܘܣ', 'ta': 'மொரிசியஸ்', 'te': 'మారిషస్', 'teo': 'Morisi', 'tg': 'Маврикий', 'th': 'มอริเชียส', 'ti': 'ማውሪሸስ', 'tk': 'Mawrikiý', 'to': 'Maulitiusi', 'tr': 'Mauritius', 'tt': 'Маврикий', 'twq': 'Mooris gungey', 'tzm': 'Muris', 'ug': 'ماۋرىتىيۇس', 'uk': 'Маврикій', 'ur': 'ماریشس', 'uz': 'Mavrikiy', 'uz-Cyrl': 'Маврикий', 'uz-Latn': 'Mavrikiy', 'vai': 'ꗞꔓꗔ', 'vai-Latn': 'Mɔreshɔ', 'vai-Vaii': 'ꗞꔓꗔ', 'vec': 'Ìzole Maurisio', 'vi': 'Mauritius', 'vun': 'Morisi', 'wae': 'Mauritius', 'wo': 'Moriis', 'xh': 'EMauritius', 'xnr': 'मॉरीशस', 'xog': 'Mawulisyasi', 'yav': 'molís', 'yi': 'מאריציוס', 'yo': 'Maritiusi', 'yrl': 'Maurisiyu', 'yue': '模里西斯', 'yue-Hans': '模里西斯', 'yue-Hant': '模里西斯', 'zgh': 'ⵎⵓⵔⵉⵙ', 'zh': '毛里求斯', 'zh-Hans': '毛里求斯', 'zh-Hant': '模里西斯', 'zu': 'i-Mauritius'},
    'MV': {'af': 'Maledive', 'agq': 'Màdivè', 'ak': 'Maldives', 'am': 'ማልዲቭስ', 'ar': 'جزر المالديف', 'as': 'মালদ্বীপ', 'asa': 'Modivu', 'ast': 'Les Maldives', 'az': 'Maldiv adaları', 'az-Cyrl': 'Малдив адалары', 'az-Latn': 'Maldiv adaları', 'bal-Latn': 'Máldip', 'bas': 'Màldîf', 'be': 'Мальдывы', 'be-tarask': 'Мальдывы', 'bez': 'Humodivu', 'bg': 'Малдиви', 'blo': 'Malɖiifu', 'bm': 'Maldivi', 'bn': 'মালদ্বীপ', 'br': 'Maldivez', 'brx': 'मालदीव', 'bs': 'Maldivi', 'bs-Cyrl': 'Малдиви', 'bs-Latn': 'Maldivi', 'ca': 'Maldives', 'ca-ES-valencia': 'Maldives', 'ccp': '𑄟𑄣𑄴𑄘𑄨𑄛𑄴', 'ce': 'Мальдиваш', 'ceb': 'Maldives', 'cgg': 'Maridives', 'chr': 'ᎹᎵᏗᏫᏍ', 'ckb': 'مالدیڤ', 'cs': 'Maledivy', 'cv': 'Мальдивсем', 'cy': 'Y Maldives', 'da': 'Maldiverne', 'dav': 'Modivu', 'de': 'Malediven', 'dje': 'Maldiivu', 'dsb': 'Malediwy', 'dz': 'མཱལ་དིབས', 'ebu': 'Modivu', 'ee': 'maldivesdukɔ', 'el': 'Μαλδίβες', 'el-polyton': 'Μαλδίβες', 'en': 'Maldives', 'en-Dsrt': '𐐣𐐪𐑊𐐼𐐨𐑂𐑆', 'eo': 'Maldivoj', 'es': 'Maldivas', 'et': 'Maldiivid', 'eu': 'Maldivak', 'ewo': 'Malədívə', 'fa': 'مالدیو', 'ff': 'Maldiiwe', 'ff-Adlm': '𞤃𞤢𞤤𞤣𞤭𞥅𞤬', 'ff-Latn': 'Maldiiwe', 'fi': 'Malediivit', 'fil': 'Maldives', 'fo': 'Maldivoyggjar', 'fr': 'Maldives', 'fur': 'Maldivis', 'fy': 'Maldiven', 'ga': 'Oileáin Mhaildíve', 'gd': 'Na h-Eileanan Mhaladaibh', 'gl': 'Maldivas', 'gsw': 'Malediiwe', 'gu': 'માલદિવ્સ', 'guz': 'Modivu', 'ha': 'Maldibi', 'he': 'האיים המלדיביים', 'hi': 'मालदीव', 'hi-Latn': 'Maldives', 'hr': 'Maldivi', 'hsb': 'Malediwy', 'ht': 'Maldives', 'hu': 'Maldív-szigetek', 'hy': 'Մալդիվներ', 'ia': 'Maldivas', 'id': 'Maladewa', 'ie': 'Maldivas', 'ig': 'Maldivesa', 'is': 'Maldíveyjar', 'it': 'Maldive', 'ja': 'モルディブ', 'jmc': 'Modivu', 'jv': 'Maladéwa', 'ka': 'მალდივები', 'kaa': 'Мальдив', 'kaa-Cyrl': 'Мальдив', 'kab': 'Maldib', 'kam': 'Modivu', 'kde': 'Modivu', 'kea': 'Maldivas', 'kgp': 'Mỹrniva', 'khq': 'Maldiivu', 'ki': 'Modivu', 'kk': 'Мальдив аралдары', 'kk-Cyrl': 'Мальдив аралдары', 'kln': 'Emetab Maldives', 'km': 'ម៉ាល់ឌីវ', 'kn': 'ಮಾಲ್ಡೀವ್ಸ್', 'ko': '몰디브', 'kok': 'मालदीव', 'kok-Deva': 'मालदीव', 'ks': 'مالدیٖو', 'ks-Arab': 'مالدیٖو', 'ksb': 'Modivu', 'ksf': 'maldivǝ', 'ksh': 'Mallediive', 'ku': 'Maldîva', 'kxv': 'māldīp', 'kxv-Deva': 'मालदीप', 'kxv-Latn': 'māldīp', 'kxv-Orya': 'ମାଲଦୀପି', 'kxv-Telu': 'మాలదిపి', 'ky': 'Мальдив', 'lag': 'Modíivu', 'lb': 'Maldiven', 'lg': 'Bizinga by’eMalidive', 'lld': 'Maldives', 'ln': 'Madívɛ', 'lo': 'ມັນດິຟ', 'lt': 'Maldyvai', 'lu': 'Madive', 'luo': 'Maldives', 'luy': 'Modivu', 'lv': 'Maldīvija', 'mai': 'मालदीब', 'mas': 'Modivu', 'mer': 'Modivu', 'mfe': 'Maldiv', 'mg': 'Maldiva', 'mi': 'Māratiri', 'mk': 'Малдиви', 'ml': 'മാലിദ്വീപ്', 'mn': 'Мальдив', 'mr': 'मालदीव', 'ms': 'Maldives', 'mt': 'il-Maldivi', 'mua': 'Maldivǝ', 'my': 'မော်လ်ဒိုက်', 'mzn': 'مالدیو', 'naq': 'Maldives', 'nb': 'Maldivene', 'nd': 'Maldives', 'ne': 'माल्दिभ्स', 'nl': 'Maldiven', 'nmg': 'Maldivia', 'nn': 'Maldivane', 'no': 'Maldivene', 'nqo': 'ߡߊߟߑߘߌߝ߭', 'nyn': 'Maridives', 'om': 'Maaldiivs', 'or': 'ମାଲଦିଭସ୍\u200c', 'pa': 'ਮਾਲਦੀਵ', 'pa-Guru': 'ਮਾਲਦੀਵ', 'pcm': 'Mọ́ldivs', 'pl': 'Malediwy', 'ps': 'مالديپ', 'pt': 'Maldivas', 'qu': 'Maldivas', 'rm': 'Maldivas', 'rn': 'Moludave', 'ro': 'Maldive', 'rof': 'Modivu', 'ru': 'Мальдивы', 'rwk': 'Modivu', 'saq': 'Modivu', 'sbp': 'Modivu', 'sc': 'Maldivas', 'sd': 'مالديپ', 'sd-Arab': 'مالديپ', 'se': 'Malediivvat', 'seh': 'Maldivas', 'ses': 'Maldiivu', 'sg': 'Maldîva', 'shi': 'ⵎⴰⵍⴷⵉⴼ', 'shi-Latn': 'maldif', 'shi-Tfng': 'ⵎⴰⵍⴷⵉⴼ', 'si': 'මාල දිවයින', 'sk': 'Maldivy', 'sl': 'Maldivi', 'smn': 'Malediveh', 'sn': 'Maldives', 'so': 'Maaldiifis', 'sq': 'Maldive', 'sr': 'Малдиви', 'sr-Cyrl': 'Малдиви', 'sr-Latn': 'Maldivi', 'sv': 'Maldiverna', 'sw': 'Maldivi', 'syr': 'ܓܙܪܬܐ ܡܐܠܕܝܒܝܬܐ', 'ta': 'மாலத்தீவு', 'te': 'మాల్దీవులు', 'teo': 'Modivu', 'tg': 'Малдив', 'th': 'มัลดีฟส์', 'ti': 'ማልዲቭስ', 'tk': 'Maldiwler', 'to': 'Malativisi', 'tr': 'Maldivler', 'tt': 'Мальдив утраулары', 'twq': 'Maldiivu', 'tzm': 'Maldiv', 'ug': 'مالدىۋې', 'uk': 'Мальдіви', 'ur': 'مالدیپ', 'uz': 'Maldiv orollari', 'uz-Cyrl': 'Мальдив ороллари', 'uz-Latn': 'Maldiv orollari', 'vai': 'ꕮꔷꕜꔍ', 'vai-Latn': 'Malidavi', 'vai-Vaii': 'ꕮꔷꕜꔍ', 'vec': 'Maldive', 'vi': 'Maldives', 'vun': 'Modivu', 'wae': 'Malediwe', 'wo': 'Maldiiw', 'xh': 'EMaldives', 'xnr': 'मालदीव', 'xog': 'Ebizinga bya Malidive', 'yav': 'maletíif', 'yi': 'מאַלדיוון', 'yo': 'Maladifi', 'yrl': 'Maudiwa-ita', 'yue': '馬爾地夫', 'yue-Hans': '马尔地夫', 'yue-Hant': '馬爾地夫', 'zgh': 'ⵎⴰⵍⴷⵉⴼ', 'zh': '马尔代夫', 'zh-Hans': '马尔代夫', 'zh-Hant': '馬爾地夫', 'zu': 'i-Maldives'},
    'MW': {'af': 'Malawi', 'agq': 'Màlawì', 'ak': 'Malawi', 'am': 'ማላዊ', 'ar': 'ملاوي', 'as': 'মালাৱি', 'asa': 'Malawi', 'ast': 'Malaui', 'az': 'Malavi', 'az-Cyrl': 'Малави', 'az-Latn': 'Malavi', 'bal-Latn': 'Maláwi', 'bas': 'Màlàwi', 'be': 'Малаві', 'be-tarask': 'Малаві', 'bez': 'Humalawi', 'bg': 'Малави', 'blo': 'Malawii', 'bm': 'Malawi', 'bn': 'মালাউই', 'br': 'Malawi', 'brx': 'मालावी', 'bs': 'Malavi', 'bs-Cyrl': 'Малави', 'bs-Latn': 'Malavi', 'ca': 'Malawi', 'ca-ES-valencia': 'Malawi', 'ccp': '𑄟𑄣𑄃𑄪𑄃𑄨', 'ce': 'Малави', 'ceb': 'Malawi', 'cgg': 'Marawi', 'chr': 'ᎹᎳᏫ', 'ckb': 'مالاوی', 'cs': 'Malawi', 'cv': 'Малави', 'cy': 'Malawi', 'da': 'Malawi', 'dav': 'Malawi', 'de': 'Malawi', 'dje': 'Malaawi', 'dsb': 'Malawi', 'dz': 'མ་ལ་ཝི', 'ebu': 'Malawi', 'ee': 'Malawi nutome', 'el': 'Μαλάουι', 'el-polyton': 'Μαλάουι', 'en': 'Malawi', 'en-Dsrt': '𐐣𐐲𐑊𐐪𐐶𐐨', 'eo': 'Malavio', 'es': 'Malaui', 'et': 'Malawi', 'eu': 'Malawi', 'ewo': 'Malawí', 'fa': 'مالاوی', 'ff': 'Malaawi', 'ff-Adlm': '𞤃𞤢𞤤𞤢𞤱𞤭𞥅', 'ff-Latn': 'Malaawi', 'fi': 'Malawi', 'fil': 'Malawi', 'fo': 'Malavi', 'fr': 'Malawi', 'fur': 'Malawi', 'fy': 'Malawi', 'ga': 'an Mhaláiv', 'gd': 'Malabhaidh', 'gl': 'Malawi', 'gsw': 'Malaawi', 'gu': 'માલાવી', 'guz': 'Malawi', 'ha': 'Malawi', 'he': 'מלאווי', 'hi': 'मलावी', 'hi-Latn': 'Malawi', 'hr': 'Malavi', 'hsb': 'Malawi', 'ht': 'Malawi', 'hu': 'Malawi', 'hy': 'Մալավի', 'ia': 'Malawi', 'id': 'Malawi', 'ig': 'Malawi', 'is': 'Malaví', 'it': 'Malawi', 'ja': 'マラウイ', 'jgo': 'Maláwi', 'jmc': 'Malawi', 'jv': 'Malawi', 'ka': 'მალავი', 'kaa': 'Малави', 'kaa-Cyrl': 'Малави', 'kab': 'Malawi', 'kam': 'Malawi', 'kde': 'Malawi', 'kea': 'Malaui', 'kgp': 'Mỹravi', 'khq': 'Malaawi', 'ki': 'Malawi', 'kk': 'Малави', 'kk-Cyrl': 'Малави', 'kln': 'Emetab Malawi', 'km': 'ម៉ាឡាវី', 'kn': 'ಮಲಾವಿ', 'ko': '말라위', 'kok': 'मलावी', 'kok-Deva': 'मलावी', 'ks': 'ملاوی', 'ks-Arab': 'ملاوی', 'ksb': 'Malawi', 'ksf': 'malawi', 'ksh': 'Malawi', 'ku': 'Malawî', 'kxv': 'malavī', 'kxv-Deva': 'मलावी', 'kxv-Latn': 'malavī', 'kxv-Orya': 'ମଲାୱି', 'kxv-Telu': 'మలావీ', 'ky': 'Малави', 'lag': 'Maláawi', 'lb': 'Malawi', 'lg': 'Malawi', 'lld': 'Malawi', 'ln': 'Malawi', 'lo': 'ມາລາວີ', 'lt': 'Malavis', 'lu': 'Malawi', 'luo': 'Malawi', 'luy': 'Malawi', 'lv': 'Malāvija', 'mai': 'मलाबी', 'mas': 'Malawi', 'mer': 'Malawi', 'mfe': 'Malawi', 'mg': 'Malaoì', 'mgh': 'Umalawi', 'mi': 'Marāwi', 'mk': 'Малави', 'ml': 'മലാവി', 'mn': 'Малави', 'mr': 'मलावी', 'ms': 'Malawi', 'mt': 'il-Malawi', 'mua': 'Sǝr Malawi', 'my': 'မာလာဝီ', 'mzn': 'مالاوی', 'naq': 'Malawib', 'nb': 'Malawi', 'nd': 'Malawi', 'ne': 'मालावी', 'nl': 'Malawi', 'nmg': 'Malawi', 'nn': 'Malawi', 'no': 'Malawi', 'nqo': 'ߡߟߊߥߌ߫', 'nyn': 'Marawi', 'om': 'Maalaawwii', 'or': 'ମାଲୱି', 'pa': 'ਮਲਾਵੀ', 'pa-Guru': 'ਮਲਾਵੀ', 'pcm': 'Maláwi', 'pl': 'Malawi', 'ps': 'مالاوي', 'pt': 'Malaui', 'qu': 'Malawi', 'rm': 'Malawi', 'rn': 'Malawi', 'ro': 'Malawi', 'rof': 'Malawi', 'ru': 'Малави', 'rwk': 'Malawi', 'saq': 'Malawi', 'sbp': 'Malawi', 'sc': 'Malawi', 'sd': 'مالاوي', 'sd-Arab': 'مالاوي', 'se': 'Malawi', 'seh': 'Malawi', 'ses': 'Malaawi', 'sg': 'Malawïi', 'shi': 'ⵎⴰⵍⴰⵡⵉ', 'shi-Latn': 'malawi', 'shi-Tfng': 'ⵎⴰⵍⴰⵡⵉ', 'si': 'මලාවි', 'sk': 'Malawi', 'sl': 'Malavi', 'smn': 'Malawi', 'sn': 'Malawi', 'so': 'Malaawi', 'sq': 'Malavi', 'sr': 'Малави', 'sr-Cyrl': 'Малави', 'sr-Latn': 'Malavi', 'sv': 'Malawi', 'sw': 'Malawi', 'syr': 'ܡܠܐܘܝ', 'ta': 'மலாவி', 'te': 'మలావీ', 'teo': 'Malawi', 'tg': 'Малави', 'th': 'มาลาวี', 'ti': 'ማላዊ', 'tk': 'Malawi', 'to': 'Malaui', 'tr': 'Malavi', 'tt': 'Малави', 'twq': 'Malaawi', 'tzm': 'Malawi', 'ug': 'مالاۋى', 'uk': 'Малаві', 'ur': 'ملاوی', 'uz': 'Malavi', 'uz-Cyrl': 'Малави', 'uz-Latn': 'Malavi', 'vai': 'ꕮꕞꕌꔨ', 'vai-Latn': 'Malawi', 'vai-Vaii': 'ꕮꕞꕌꔨ', 'vec': 'Malawi', 'vi': 'Malawi', 'vun': 'Malawi', 'wae': 'Malawi', 'wo': 'Malawi', 'xh': 'EMalawi', 'xnr': 'मलावी', 'xog': 'Malawi', 'yav': 'malawí', 'yi': 'מאַלאַווי', 'yo': 'Malawi', 'yrl': 'Marawi', 'yue': '馬拉威', 'yue-Hans': '马拉威', 'yue-Hant': '馬拉威', 'zgh': 'ⵎⴰⵍⴰⵡⵉ', 'zh': '马拉维', 'zh-Hans': '马拉维', 'zh-Hant': '馬拉威', 'zu': 'iMalawi'},
    'MX': {'af': 'Meksiko', 'agq': 'Mɛkɨzikù', 'ak': 'Mɛksiko', 'am': 'ሜክሲኮ', 'ar': 'المكسيك', 'as': 'মেক্সিকো', 'asa': 'Mekthiko', 'ast': 'Méxicu', 'az': 'Meksika', 'az-Cyrl': 'Мексика', 'az-Latn': 'Meksika', 'bal-Latn': 'Meksikó', 'bas': 'Mɛ̀gsîk', 'be': 'Мексіка', 'be-tarask': 'Мексіка', 'bez': 'Humeksiko', 'bg': 'Мексико', 'bgn': 'مکسیکو', 'blo': 'Mɛsik', 'bm': 'Meksiki', 'bn': 'মেক্সিকো', 'br': 'Mecʼhiko', 'brx': 'मेक्सिक’', 'bs': 'Meksiko', 'bs-Cyrl': 'Мексико', 'bs-Latn': 'Meksiko', 'ca': 'Mèxic', 'ca-ES-valencia': 'Mèxic', 'ccp': '𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄮', 'ce': 'Мексика', 'ceb': 'Mexico', 'cgg': 'Mexico', 'chr': 'ᎠᏂᏍᏆᏂ', 'ckb': 'مەکسیک', 'cs': 'Mexiko', 'cv': 'Мексика', 'cy': 'Mecsico', 'da': 'Mexico', 'dav': 'Meksiko', 'de': 'Mexiko', 'dje': 'Mexiki', 'dsb': 'Mexiko', 'dz': 'མེཀ་སི་ཀོ', 'ebu': 'Meksiko', 'ee': 'Mexico nutome', 'el': 'Μεξικό', 'el-polyton': 'Μεξικό', 'en': 'Mexico', 'en-Dsrt': '𐐣𐐯𐐿𐑅𐐲𐐿𐐬', 'eo': 'Meksiko', 'es': 'México', 'et': 'Mehhiko', 'eu': 'Mexiko', 'ewo': 'Mɛkəsíg', 'fa': 'مکزیک', 'ff': 'Meksik', 'ff-Adlm': '𞤃𞤫𞤳𞤧𞤭𞤳𞤮𞥅', 'ff-Latn': 'Meksik', 'fi': 'Meksiko', 'fil': 'Mexico', 'fo': 'Meksiko', 'fr': 'Mexique', 'fur': 'Messic', 'fy': 'Mexico', 'ga': 'Meicsiceo', 'gd': 'Meagsago', 'gl': 'México', 'gsw': 'Mexiko', 'gu': 'મેક્સિકો', 'guz': 'Meksiko', 'ha': 'Mesiko', 'haw': 'Mekiko', 'he': 'מקסיקו', 'hi': 'मैक्सिको', 'hi-Latn': 'Mexico', 'hr': 'Meksiko', 'hsb': 'Mexiko', 'ht': 'Mexique', 'hu': 'Mexikó', 'hy': 'Մեքսիկա', 'ia': 'Mexico', 'id': 'Meksiko', 'ig': 'Mexico', 'ii': 'ꃀꑭꇬ', 'is': 'Mexíkó', 'it': 'Messico', 'ja': 'メキシコ', 'jgo': 'Mɛksîk', 'jmc': 'Meksiko', 'jv': 'Mèksiko', 'ka': 'მექსიკა', 'kaa': 'Мексика', 'kaa-Cyrl': 'Мексика', 'kab': 'Meksik', 'kam': 'Meksiko', 'kde': 'Meksiko', 'kea': 'Méxiku', 'kgp': 'Mẽsiku', 'khq': 'Mexiki', 'ki': 'Meksiko', 'kk': 'Мексика', 'kk-Cyrl': 'Мексика', 'kln': 'Emetab Mexico', 'km': 'ម៉ិកស៊ិក', 'kn': 'ಮೆಕ್ಸಿಕೊ', 'ko': '멕시코', 'kok': 'मेक्सिको', 'kok-Deva': 'मेक्सिको', 'ks': 'مؠکسِکو', 'ks-Arab': 'مؠکسِکو', 'ksb': 'Meksiko', 'ksf': 'mɛksík', 'ksh': 'Mäxiko', 'ku': 'Meksîka', 'kxv': 'meksīko', 'kxv-Deva': 'मेक्सिको', 'kxv-Latn': 'meksīko', 'kxv-Orya': 'ମେକ୍ସିକ', 'kxv-Telu': 'మెక్సికో', 'ky': 'Мексика', 'lag': 'Mekisiko', 'lb': 'Mexiko', 'lg': 'Mekisiko', 'lkt': 'Spayóla Makȟóčhe', 'lld': 'Messich', 'ln': 'Meksike', 'lo': 'ເມັກຊິໂກ', 'lt': 'Meksika', 'lu': 'Meksike', 'luo': 'Mexico', 'luy': 'Meksiko', 'lv': 'Meksika', 'mai': 'मैक्सिको', 'mas': 'Meksiko', 'mer': 'Mexiko', 'mfe': 'Mexik', 'mg': 'Meksika', 'mi': 'Mēhiko', 'mk': 'Мексико', 'ml': 'മെക്സിക്കോ', 'mn': 'Мексик', 'mr': 'मेक्सिको', 'ms': 'Mexico', 'mt': 'il-Messiku', 'mua': 'Mexiko', 'my': 'မက်ကဆီကို', 'mzn': 'مکزیک', 'naq': 'Mexicob', 'nb': 'Mexico', 'nd': 'Meksikho', 'ne': 'मेक्सिको', 'nl': 'Mexico', 'nmg': 'Mɛxik', 'nn': 'Mexico', 'no': 'Mexico', 'nqo': 'ߡߍߞߑߛߌߞ', 'nyn': 'Mexico', 'om': 'Meeksiikoo', 'or': 'ମେକ୍ସିକୋ', 'pa': 'ਮੈਕਸੀਕੋ', 'pa-Guru': 'ਮੈਕਸੀਕੋ', 'pcm': 'Mẹ́ksíko', 'pl': 'Meksyk', 'ps': 'میکسیکو', 'pt': 'México', 'qu': 'México', 'rm': 'Mexico', 'rn': 'Migizike', 'ro': 'Mexic', 'rof': 'Meksiko', 'ru': 'Мексика', 'rwk': 'Meksiko', 'sah': 'Миэксикэ', 'saq': 'Meksiko', 'sbp': 'Mekisiko', 'sc': 'Mèssicu', 'sd': 'ميڪسيڪو', 'sd-Arab': 'ميڪسيڪو', 'se': 'Meksiko', 'seh': 'México', 'ses': 'Mexiki', 'sg': 'Mekisîki', 'shi': 'ⵎⵉⴽⵙⵉⴽ', 'shi-Latn': 'miksik', 'shi-Tfng': 'ⵎⵉⴽⵙⵉⴽ', 'si': 'මෙක්සිකෝව', 'sk': 'Mexiko', 'sl': 'Mehika', 'smn': 'Meksiko', 'sn': 'Mexico', 'so': 'Meksiko', 'sq': 'Meksikë', 'sr': 'Мексико', 'sr-Cyrl': 'Мексико', 'sr-Latn': 'Meksiko', 'sv': 'Mexiko', 'sw': 'Meksiko', 'syr': 'ܡܟܣܝܟܘ', 'ta': 'மெக்சிகோ', 'te': 'మెక్సికో', 'teo': 'Meksiko', 'tg': 'Мексика', 'th': 'เม็กซิโก', 'ti': 'ሜክሲኮ', 'tk': 'Meksika', 'to': 'Mekisikou', 'tr': 'Meksika', 'tt': 'Мексика', 'twq': 'Mexiki', 'tzm': 'Miksik', 'ug': 'مېكسىكا', 'uk': 'Мексика', 'ur': 'میکسیکو', 'uz': 'Meksika', 'uz-Cyrl': 'Мексика', 'uz-Latn': 'Meksika', 'vai': 'ꘈꔻꖏ', 'vai-Latn': 'Mɛsíko', 'vai-Vaii': 'ꘈꔻꖏ', 'vec': 'Mèsego', 'vi': 'Mexico', 'vun': 'Meksiko', 'wae': 'Mexiko', 'wo': 'Meksiko', 'xh': 'EMexico', 'xnr': 'मैक्सिको', 'xog': 'Mekisiko', 'yav': 'mɛksíik', 'yi': 'מעקסיקע', 'yo': 'Mesiko', 'yrl': 'Mẽsiku', 'yue': '墨西哥', 'yue-Hans': '墨西哥', 'yue-Hant': '墨西哥', 'zgh': 'ⵎⵉⴽⵙⵉⴽ', 'zh': '墨西哥', 'zh-Hans': '墨西哥', 'zh-Hant': '墨西哥', 'zu': 'i-Mexico'},
    'MY': {'af': 'Maleisië', 'agq': 'Màlɛshìa', 'ak': 'Malehyia', 'am': 'ማሌዢያ', 'ar': 'ماليزيا', 'as': 'মালয়েচিয়া', 'asa': 'Malethia', 'ast': 'Malasia', 'az': 'Malayziya', 'az-Cyrl': 'Малајзија', 'az-Latn': 'Malayziya', 'bal-Latn': 'Maléshiá', 'bas': 'Màlɛ̀sìà', 'be': 'Малайзія', 'be-tarask': 'Малайзія', 'bez': 'Humalesia', 'bg': 'Малайзия', 'bgn': 'مالیزیا', 'blo': 'Malɛsii', 'bm': 'Malɛzi', 'bn': 'মালয়েশিয়া', 'br': 'Malaysia', 'brx': 'मालेशिया', 'bs': 'Malezija', 'bs-Cyrl': 'Малезија', 'bs-Latn': 'Malezija', 'ca': 'Malàisia', 'ca-ES-valencia': 'Malàisia', 'ccp': '𑄟𑄣𑄴𑄠𑄬𑄥𑄨𑄠', 'ce': 'Малайзи', 'ceb': 'Malaysia', 'cgg': 'marayizia', 'chr': 'ᎹᎴᏏᎢᎠ', 'ckb': 'مالیزیا', 'cs': 'Malajsie', 'cv': 'Малайзи', 'cy': 'Malaysia', 'da': 'Malaysia', 'dav': 'Malesia', 'de': 'Malaysia', 'dje': 'Maleezi', 'dsb': 'Malajzija', 'dz': 'མ་ལེ་ཤི་ཡ', 'ebu': 'Malesia', 'ee': 'Malaysia nutome', 'el': 'Μαλαισία', 'el-polyton': 'Μαλαισία', 'en': 'Malaysia', 'en-Dsrt': '𐐣𐐲𐑊𐐩𐑈𐐲', 'eo': 'Malajzio', 'es': 'Malasia', 'et': 'Malaisia', 'eu': 'Malaysia', 'ewo': 'Malɛ́zia', 'fa': 'مالزی', 'ff': 'Malesii', 'ff-Adlm': '𞤃𞤢𞤤𞤫𞥅𞤧𞤭𞤴𞤢', 'ff-Latn': 'Malesii', 'fi': 'Malesia', 'fil': 'Malaysia', 'fo': 'Malaisia', 'fr': 'Malaisie', 'fur': 'Malaysia', 'fy': 'Maleisië', 'ga': 'an Mhalaeisia', 'gd': 'Malaidhsea', 'gl': 'Malaisia', 'gsw': 'Maläisia', 'gu': 'મલેશિયા', 'guz': 'Malesia', 'ha': 'Malesiya', 'he': 'מלזיה', 'hi': 'मलेशिया', 'hi-Latn': 'Malaysia', 'hr': 'Malezija', 'hsb': 'Malajzija', 'ht': 'Malaisie', 'hu': 'Malajzia', 'hy': 'Մալայզիա', 'ia': 'Malaysia', 'id': 'Malaysia', 'ig': 'Malaysia', 'is': 'Malasía', 'it': 'Malaysia', 'ja': 'マレーシア', 'jmc': 'Malesia', 'jv': 'Malaysia', 'ka': 'მალაიზია', 'kaa': 'Малайзия', 'kaa-Cyrl': 'Малайзия', 'kab': 'Malizya', 'kam': 'Malesia', 'kde': 'Malesia', 'kea': 'Malázia', 'kgp': 'Mỹraja', 'khq': 'Maleezi', 'ki': 'Malesia', 'kk': 'Малайзия', 'kk-Cyrl': 'Малайзия', 'kln': 'Emetab Malaysia', 'km': 'ម៉ាឡេស៊ី', 'kn': 'ಮಲೇಶಿಯಾ', 'ko': '말레이시아', 'kok': 'मलेशिया', 'kok-Deva': 'मलेशिया', 'ks': 'مَلیشِیا', 'ks-Arab': 'مَلیشِیا', 'ksb': 'Malesia', 'ksf': 'malɛsí', 'ksh': 'Malaisije', 'ku': 'Malezya', 'kxv': 'maleseāā', 'kxv-Deva': 'मलेसिया', 'kxv-Latn': 'maleseāā', 'kxv-Orya': 'ମାଲେସିଆ', 'kxv-Telu': 'మలేసియా', 'ky': 'Малайзия', 'lag': 'Maleísia', 'lb': 'Malaysia', 'lg': 'Malezya', 'lld': 'Malesia', 'ln': 'Malezi', 'lo': 'ມາເລເຊຍ', 'lt': 'Malaizija', 'lu': 'Malezi', 'luo': 'Malaysia', 'luy': 'Malesia', 'lv': 'Malaizija', 'mai': 'मलेशिया', 'mas': 'Malesia', 'mer': 'Malĩsia', 'mfe': 'Malezi', 'mg': 'Malaizia', 'mi': 'Mareia', 'mk': 'Малезија', 'ml': 'മലേഷ്യ', 'mn': 'Малайз', 'mr': 'मलेशिया', 'ms': 'Malaysia', 'mt': 'il-Malasja', 'mua': 'Malaysiya', 'my': 'မလေးရှား', 'mzn': 'مالزی', 'naq': 'Malaysiab', 'nb': 'Malaysia', 'nd': 'Malezhiya', 'ne': 'मलेसिया', 'nl': 'Maleisië', 'nmg': 'Malaysia', 'nn': 'Malaysia', 'no': 'Malaysia', 'nqo': 'ߡߊߟߍߘߎ߯', 'nyn': 'marayizia', 'om': 'Maleeshiyaa', 'or': 'ମାଲେସିଆ', 'pa': 'ਮਲੇਸ਼ੀਆ', 'pa-Guru': 'ਮਲੇਸ਼ੀਆ', 'pcm': 'Maléshia', 'pl': 'Malezja', 'ps': 'مالیزیا', 'pt': 'Malásia', 'qu': 'Malasia', 'rm': 'Malaisia', 'rn': 'Maleziya', 'ro': 'Malaysia', 'rof': 'Malesia', 'ru': 'Малайзия', 'rwk': 'Malesia', 'saq': 'Malesia', 'sbp': 'Malesiya', 'sc': 'Malèsia', 'sd': 'ملائيشيا', 'sd-Arab': 'ملائيشيا', 'se': 'Malesia', 'seh': 'Malásia', 'ses': 'Maleezi', 'sg': 'Malezïi', 'shi': 'ⵎⴰⵍⵉⵣⵢⴰ', 'shi-Latn': 'malizya', 'shi-Tfng': 'ⵎⴰⵍⵉⵣⵢⴰ', 'si': 'මැලේසියාව', 'sk': 'Malajzia', 'sl': 'Malezija', 'smn': 'Malaysia', 'sn': 'Malaysia', 'so': 'Malaysiya', 'sq': 'Malajzi', 'sr': 'Малезија', 'sr-Cyrl': 'Малезија', 'sr-Latn': 'Malezija', 'sv': 'Malaysia', 'sw': 'Malesia', 'syr': 'ܡܠܝܙܝܐ', 'ta': 'மலேசியா', 'te': 'మలేషియా', 'teo': 'Malesia', 'tg': 'Малайзия', 'th': 'มาเลเซีย', 'ti': 'ማለዥያ', 'tk': 'Malaýziýa', 'to': 'Malēsia', 'tr': 'Malezya', 'tt': 'Малайзия', 'twq': 'Maleezi', 'tzm': 'Malizya', 'ug': 'مالايسىيا', 'uk': 'Малайзія', 'ur': 'ملائشیا', 'uz': 'Malayziya', 'uz-Cyrl': 'Малайзия', 'uz-Latn': 'Malayziya', 'vai': 'ꕮꔒꔻꕩ', 'vai-Latn': 'Malesiya', 'vai-Vaii': 'ꕮꔒꔻꕩ', 'vec': 'Malèizia', 'vi': 'Malaysia', 'vun': 'Malesia', 'wae': 'Malaysia', 'wo': 'Malesi', 'xh': 'EMalaysia', 'xnr': 'मलेशिया', 'xog': 'Malezya', 'yav': 'malesí', 'yi': 'מאַלייזיע', 'yo': 'Malasia', 'yrl': 'Malasiya', 'yue': '馬來西亞', 'yue-Hans': '马来西亚', 'yue-Hant': '馬來西亞', 'zgh': 'ⵎⴰⵍⵉⵣⵢⴰ', 'zh': '马来西亚', 'zh-Hans': '马来西亚', 'zh-Hant': '馬來西亞', 'zu': 'i-Malaysia'},
    'MZ': {'af': 'Mosambiek', 'agq': 'Mùzàmbî', 'ak': 'Mozambik', 'am': 'ሞዛምቢክ', 'ar': 'موزمبيق', 'as': 'ম’জামবিক', 'asa': 'Mthumbiji', 'ast': 'Mozambique', 'az': 'Mozambik', 'az-Cyrl': 'Мозамбик', 'az-Latn': 'Mozambik', 'bal-Latn': 'Mózambik', 'bas': 'Mòsàmbîk', 'be': 'Мазамбік', 'be-tarask': 'Мазамбік', 'bez': 'Humusumbiji', 'bg': 'Мозамбик', 'blo': 'Mosambii', 'bm': 'Mozanbiki', 'bn': 'মোজাম্বিক', 'br': 'Mozambik', 'brx': 'मजाम्बिक', 'bs': 'Mozambik', 'bs-Cyrl': 'Мозамбик', 'bs-Latn': 'Mozambik', 'ca': 'Moçambic', 'ca-ES-valencia': 'Moçambic', 'ccp': '𑄟𑄮𑄎𑄟𑄴𑄝𑄨𑄇𑄴', 'ce': 'Мозамбик', 'ceb': 'Mozambique', 'cgg': 'Mozambique', 'chr': 'ᎼᏎᎻᏇᎩ', 'ckb': 'مۆزامبیک', 'cs': 'Mosambik', 'cv': 'Мозамбик', 'cy': 'Mozambique', 'da': 'Mozambique', 'dav': 'Msumbiji', 'de': 'Mosambik', 'dje': 'Mozambik', 'dsb': 'Mosambik', 'dz': 'མོ་ཛམ་བྷིཀ', 'ebu': 'Msumbiji', 'ee': 'Mozambiki nutome', 'el': 'Μοζαμβίκη', 'el-polyton': 'Μοζαμβίκη', 'en': 'Mozambique', 'en-Dsrt': '𐐣𐐬𐑆𐐰𐑋𐐺𐐨𐐿', 'eo': 'Mozambiko', 'es': 'Mozambique', 'et': 'Mosambiik', 'eu': 'Mozambike', 'ewo': 'Mozambíg', 'fa': 'موزامبیک', 'ff': 'Mosammbik', 'ff-Adlm': '𞤃𞤮𞤧𞤢𞤥𞤦𞤭𞥅𞤳', 'ff-Latn': 'Mosammbik', 'fi': 'Mosambik', 'fil': 'Mozambique', 'fo': 'Mosambik', 'fr': 'Mozambique', 'fur': 'Mozambic', 'fy': 'Mozambique', 'ga': 'Mósaimbíc', 'gd': 'Mòsaimbic', 'gl': 'Mozambique', 'gsw': 'Mosambik', 'gu': 'મોઝામ્બિક', 'guz': 'Msumbiji', 'ha': 'Mozambik', 'he': 'מוזמביק', 'hi': 'मोज़ांबिक', 'hi-Latn': 'Mozambique', 'hr': 'Mozambik', 'hsb': 'Mosambik', 'ht': 'Mozambique', 'hu': 'Mozambik', 'hy': 'Մոզամբիկ', 'ia': 'Mozambique', 'id': 'Mozambik', 'ig': 'Mozambique', 'is': 'Mósambík', 'it': 'Mozambico', 'ja': 'モザンビーク', 'jgo': 'Mɔzambîk', 'jmc': 'Msumbiji', 'jv': 'Mosambik', 'ka': 'მოზამბიკი', 'kaa': 'Мозамбик', 'kaa-Cyrl': 'Мозамбик', 'kab': 'Muzembiq', 'kam': 'Msumbiji', 'kde': 'Msumbiji', 'kea': 'Musambiki', 'kgp': 'Mãsỹmiki', 'khq': 'Mozambik', 'ki': 'Msumbiji', 'kk': 'Мозамбик', 'kk-Cyrl': 'Мозамбик', 'kln': 'Emetab Mozambique', 'km': 'ម៉ូសំប៊ិក', 'kn': 'ಮೊಜಾಂಬಿಕ್', 'ko': '모잠비크', 'kok': 'मॉझांबीक', 'kok-Deva': 'मॉझांबीक', 'ks': 'موزَمبِک', 'ks-Arab': 'موزَمبِک', 'ksb': 'Msumbiji', 'ksf': 'mosambík', 'ksh': 'Mosambik', 'ku': 'Mozambîk', 'kxv': 'mojambik', 'kxv-Deva': 'मोज़ांबिक', 'kxv-Latn': 'mojambik', 'kxv-Orya': 'ମଜାମ୍ବିକ', 'kxv-Telu': 'మొజాంబిక్', 'ky': 'Мозамбик', 'lag': 'Musumbíiji', 'lb': 'Mosambik', 'lg': 'Mozambiiki', 'lld': 'Mozambich', 'ln': 'Mozambíki', 'lo': 'ໂມແຊມບິກ', 'lt': 'Mozambikas', 'lu': 'Mozambiki', 'luo': 'Mozambique', 'luy': 'Msumbiji', 'lv': 'Mozambika', 'mai': 'मोजाम्बिक', 'mas': 'Msumbiji', 'mer': 'Mozambiki', 'mfe': 'Mozambik', 'mg': 'Mozambika', 'mgh': 'Umozambiki', 'mi': 'Mohapiki', 'mk': 'Мозамбик', 'ml': 'മൊസാംബിക്ക്', 'mn': 'Мозамбик', 'mr': 'मोझाम्बिक', 'ms': 'Mozambique', 'mt': 'il-Mozambique', 'mua': 'Mozambika', 'my': 'မိုဇမ်ဘစ်', 'mzn': 'موزامبیک', 'naq': 'Mozambiki', 'nb': 'Mosambik', 'nd': 'Mozambique', 'ne': 'मोजाम्बिक', 'nl': 'Mozambique', 'nmg': 'Mozambik', 'nn': 'Mosambik', 'no': 'Mosambik', 'nqo': 'ߡߏߖ߭ߊ߲ߓߌߞ', 'nyn': 'Mozambique', 'om': 'Moozaambik', 'or': 'ମୋଜାମ୍ବିକ୍\u200c', 'pa': 'ਮੋਜ਼ਾਮਬੀਕ', 'pa-Guru': 'ਮੋਜ਼ਾਮਬੀਕ', 'pcm': 'Mozámbik', 'pl': 'Mozambik', 'ps': 'موزمبيق', 'pt': 'Moçambique', 'qu': 'Mozambique', 'rm': 'Mosambic', 'rn': 'Mozambiki', 'ro': 'Mozambic', 'rof': 'Msumbiji', 'ru': 'Мозамбик', 'rwk': 'Msumbiji', 'saq': 'Msumbiji', 'sbp': 'Musumbiji', 'sc': 'Mozambicu', 'sd': 'موزمبیق', 'sd-Arab': 'موزمبیق', 'se': 'Mosambik', 'seh': 'Moçambique', 'ses': 'Mozambik', 'sg': 'Mözämbîka', 'shi': 'ⵎⵓⵣⵏⴱⵉⵇ', 'shi-Latn': 'muznbiq', 'shi-Tfng': 'ⵎⵓⵣⵏⴱⵉⵇ', 'si': 'මොසැම්බික්', 'sk': 'Mozambik', 'sl': 'Mozambik', 'smn': 'Mosambik', 'sn': 'Mozambique', 'so': 'Musambiik', 'sq': 'Mozambik', 'sr': 'Мозамбик', 'sr-Cyrl': 'Мозамбик', 'sr-Latn': 'Mozambik', 'sv': 'Moçambique', 'sw': 'Msumbiji', 'syr': 'ܡܘܙܡܒܝܩ', 'ta': 'மொசாம்பிக்', 'te': 'మొజాంబిక్', 'teo': 'Msumbiji', 'tg': 'Мозамбик', 'th': 'โมซัมบิก', 'ti': 'ሞዛምቢክ', 'tk': 'Mozambik', 'to': 'Mosēmipiki', 'tr': 'Mozambik', 'tt': 'Мозамбик', 'twq': 'Mozambik', 'tzm': 'Muzambiq', 'ug': 'موزامبىك', 'uk': 'Мозамбік', 'ur': 'موزمبیق', 'uz': 'Mozambik', 'uz-Cyrl': 'Мозамбик', 'uz-Latn': 'Mozambik', 'vai': 'ꕹꕤꔭꕃ', 'vai-Latn': 'Mozambiki', 'vai-Vaii': 'ꕹꕤꔭꕃ', 'vec': 'Mozanbigo', 'vi': 'Mozambique', 'vmw': 'oMosambikhi', 'vun': 'Msumbiji', 'wae': 'Mosambik', 'wo': 'Mosàmbig', 'xh': 'EMozambique', 'xnr': 'मोज़ांबिक', 'xog': 'Mozambiiki', 'yav': 'mosambík', 'yi': 'מאזאַמביק', 'yo': 'Moṣamibiku', 'yrl': 'Musãbiki', 'yue': '莫三比克', 'yue-Hans': '莫三比克', 'yue-Hant': '莫三比克', 'zgh': 'ⵎⵓⵣⵏⴱⵉⵇ', 'zh': '莫桑比克', 'zh-Hans': '莫桑比克', 'zh-Hant': '莫三比克', 'zu': 'i-Mozambique'},
    'Mahj': {'ast': 'mahajani', 'bal-Latn': 'Mahájani', 'bs': 'mahadžani pismo', 'bs-Cyrl': 'махајани писмо', 'bs-Latn': 'mahadžani pismo', 'ca': 'mahajani', 'ca-ES-valencia': 'mahajani', 'cs': 'mahádžaní', 'de': 'Mahajani', 'en': 'Mahajani', 'et': 'mahaadžani', 'eu': 'mahajaniera', 'ff-Adlm': '𞤃𞤢𞤸𞤢𞤶𞤢𞤲𞤭𞥅', 'fi': 'mahajanilainen', 'ga': 'Mahasánach', 'gd': 'Mahajani', 'he': 'מהג׳אני', 'hi-Latn': 'Mahajani', 'id': 'Mahajani', 'ig': 'Mahajani', 'ja': 'マハージャニー文字', 'kk': 'махаджани жазуы', 'kk-Cyrl': 'махаджани жазуы', 'ko': '마하자니 문자', 'lt': 'Mahadžani', 'mk': 'махаџани', 'mn': 'Махажани', 'ms': 'Mahajani', 'nb': 'mahajani', 'nl': 'Mahajani', 'nn': 'mahajani', 'no': 'mahajani', 'nqo': 'ߡߤߊߖߊߣߌ', 'ro': 'mahajani', 'sc': 'mahajani', 'so': 'Mahajani', 'sq': 'mahaxhanisht', 'sv': 'mahajaniska', 'tg': 'Махаҷанӣ', 'th': 'มหาชนี', 'to': 'tohinima fakamahasani', 'tr': 'Mahajani', 'vec': 'mahajan', 'zh': '默哈金文', 'zh-Hans': '默哈金文', 'zh-Hant': '印地文', 'zu': 'i-Mahajani'},
    'Maka': {'bal-Latn': 'Makasar', 'bs': 'makasar pismo', 'bs-Cyrl': 'макасар писмо', 'bs-Latn': 'makasar pismo', 'en': 'Makasar', 'et': 'makassari', 'eu': 'makasarrera', 'ff-Adlm': '𞤃𞤢𞤳𞤢𞤧𞤢𞤪', 'fi': 'makassar', 'gd': 'Makasar', 'he': 'מקאסאר', 'hi-Latn': 'Makasar', 'id': 'Makassar', 'ig': 'Makasar', 'kk': 'макасар жазуы', 'kk-Cyrl': 'макасар жазуы', 'mn': 'Макасар', 'ms': 'Makasar', 'nl': 'Makasar', 'nqo': 'ߡߞߊߛߊ߯ߙ', 'ro': 'makasar', 'sc': 'makasar', 'so': 'Makasar', 'sq': 'makasarisht', 'sv': 'makasariska', 'tg': 'Макасарӣ', 'vec': 'makasar', 'zh': '望加锡文', 'zh-Hans': '望加锡文', 'zu': 'i-Makasar'},
    'Mand': {'ar': 'المانداينية', 'ast': 'mandéu', 'az': 'mandayen', 'az-Latn': 'mandayen', 'bal-Latn': 'Mandáin', 'bg': 'Мандаринска', 'bn': 'ম্যান্ডায়ীন', 'brx': 'मांडे', 'bs': 'mandeansko pismo', 'bs-Cyrl': 'мандеанско писмо', 'bs-Latn': 'mandeansko pismo', 'ca': 'mandaic', 'ca-ES-valencia': 'mandaic', 'ccp': '𑄟𑄳𑄠𑄚𑄴𑄓𑄠𑄩𑄚𑄴', 'cs': 'mandejské', 'da': 'mandaisk', 'de': 'Mandäisch', 'el': 'Μανδαϊκό', 'el-polyton': 'Μανδαϊκό', 'en': 'Mandaean', 'en-Dsrt': '𐐣𐐰𐑌𐐼𐐨𐐲𐑌', 'es': 'mandeo', 'et': 'mandea', 'eu': 'mandaera', 'fa': 'منده\u200cای', 'ff-Adlm': '𞤃𞤢𞤲𞤣𞤫𞥅𞤲', 'fi': 'mandealainen', 'fo': 'mandaean', 'fr': 'mandéen', 'fy': 'Mandaeans', 'ga': 'Mandaean', 'gd': 'Mandaean', 'gsw': 'Mandäisch', 'gu': 'માન્ડાયીન', 'he': 'מנדאית', 'hi': 'मनडेन', 'hi-Latn': 'Mandaean', 'hr': 'mandai pismo', 'ht': 'mandéen', 'hu': 'Mandai', 'id': 'Mandae', 'ig': 'Mandaean', 'is': 'mandaíska', 'it': 'mandaico', 'ja': 'マンダ文字', 'ka': 'მანდეური', 'kgp': 'mỹnajku', 'kk': 'мандей жазуы', 'kk-Cyrl': 'мандей жазуы', 'km': 'ម៉ានដា', 'kn': 'ಮಂಡೇಯನ್', 'ko': '만다이아 문자', 'ks': 'مَندییَن', 'ks-Arab': 'مَندییَن', 'lb': 'Mandäesch', 'lo': 'ແມນດຽນ', 'lt': 'mandėjų', 'mk': 'мандејско', 'ml': 'മൻഡേയൻ', 'mn': 'Манд', 'mr': 'मान्डायीन', 'ms': 'Mandaean', 'nb': 'mandaisk', 'ne': 'मान्डाएन', 'nl': 'Mandaeans', 'nn': 'mandaisk', 'no': 'mandaisk', 'nqo': 'ߡߊ߲߬ߘߋ߲߬ߞߊ ߡߋߛߔߏߕߊߡߌ߫', 'or': 'ମାନଡେନ୍', 'pl': 'mandejskie', 'pt': 'mandaico', 'rm': 'mandaic', 'ro': 'mandeană', 'ru': 'мандейская', 'sc': 'mandàicu', 'sl': 'mandanski', 'so': 'Luuqada Mandaean', 'sq': 'mande', 'sr': 'мандеанско писмо', 'sr-Cyrl': 'мандеанско писмо', 'sr-Latn': 'mandeansko pismo', 'sv': 'mandaéiska', 'ta': 'மேன்டியன்', 'te': 'మాన్డియన్', 'tg': 'Мандаеан', 'th': 'แมนเดียน', 'to': 'tohinima fakamanitaea', 'tr': 'Manden', 'ug': 'ماندائىكچە', 'uk': 'мандейський', 'vec': 'mandàego', 'vi': 'Chữ Mandaean', 'xnr': 'मनडेन', 'yo': 'Èdè Mandaiani', 'yrl': 'mãdaiku', 'yue': '曼底安文', 'yue-Hans': '曼底安文', 'yue-Hant': '曼底安文', 'zh': '阿拉米文', 'zh-Hans': '阿拉米文', 'zh-Hant': '曼底安文', 'zu': 'isi-Mandaean'},
    'Mani': {'ast': 'maniquéu', 'az': 'maniçayen', 'az-Latn': 'maniçayen', 'bal-Latn': 'Manichái', 'bg': 'Манихейска', 'bn': 'ম্যানিচাইন', 'br': 'manikeek', 'brx': 'मानीकी', 'bs': 'manihejsko pismo', 'bs-Cyrl': 'манихејско писмо', 'bs-Latn': 'manihejsko pismo', 'ca': 'maniqueu', 'ca-ES-valencia': 'maniqueu', 'ccp': '𑄟𑄳𑄠𑄚𑄨𑄌𑄭𑄚𑄴', 'cs': 'manichejské', 'da': 'manikæisk', 'de': 'Manichäisch', 'el': 'Μανιχαϊκό', 'el-polyton': 'Μανιχαϊκό', 'en': 'Manichaean', 'en-Dsrt': '𐐣𐐰𐑌𐐲𐐿𐐨𐐲𐑌', 'eu': 'manikeoa', 'fa': 'مانوی', 'ff-Adlm': '𞤃𞤢𞤲𞤭𞤳𞤭𞤴𞤫𞤲', 'fi': 'manikealainen', 'fr': 'manichéen', 'fy': 'Manicheaansk', 'ga': 'Mainicéasach', 'gd': 'Manichaean', 'gsw': 'Manichäisch', 'gu': 'માનીચાયીન', 'he': 'מניצ׳י', 'hi': 'मनीशीन', 'hi-Latn': 'Manichaean', 'hr': 'manihejsko pismo', 'ht': 'manichéen', 'hu': 'Manicheus', 'id': 'Manikhei', 'ig': 'Manichaean', 'it': 'manicheo', 'ja': 'マニ文字', 'ka': 'მანიქეური', 'kgp': 'mỹnĩkejỹnũ', 'kk': 'манихей жазуы', 'kk-Cyrl': 'манихей жазуы', 'kn': 'ಮನಿಚೈಯನ್', 'ko': '마니교 문자', 'ks': 'مانیشییَن', 'ks-Arab': 'مانیشییَن', 'lb': 'Manichäesch', 'lo': 'ມານິແຊນ', 'lt': 'maničų', 'mk': 'манихејско', 'ml': 'മണിചേയൻ', 'mn': 'Манич', 'mr': 'मानीचायीन', 'ms': 'Manichaean', 'nb': 'manikeisk', 'ne': 'मानिकाएन', 'nl': 'Manicheaans', 'nn': 'manikeisk', 'no': 'manikeisk', 'nqo': 'ߡߊߣߌߛߋ', 'or': 'ମନଶୀନ୍', 'pl': 'manichejskie', 'pt': 'maniqueano', 'rm': 'manicheic', 'ro': 'maniheeană', 'ru': 'манихейская', 'sc': 'manicheu', 'sl': 'manihejski', 'so': 'Manichaean', 'sq': 'manikeanisht', 'sr': 'манихејско писмо', 'sr-Cyrl': 'манихејско писмо', 'sr-Latn': 'manihejsko pismo', 'sv': 'manikeanska', 'ta': 'மனிசெய்ன்', 'te': 'మానిచేన్', 'tg': 'Манихейӣ', 'th': 'มานิแชน', 'to': 'tohinima fakamanikaea', 'tr': 'Maniheist', 'ug': 'مانەكېزەمچە', 'uk': 'маніхейський', 'vec': 'manighèo', 'vi': 'Chữ Manichaean', 'xnr': 'मनीशीन', 'yrl': 'manikeanu', 'yue': '摩尼教文', 'yue-Hans': '摩尼教文', 'yue-Hant': '摩尼教文', 'zh': '摩尼教文', 'zh-Hans': '摩尼教文', 'zh-Hant': '摩尼教文', 'zu': 'i-Manichaean'},
    'Marc': {'bal-Latn': 'Marchen', 'bs': 'marčensko pismo', 'bs-Cyrl': 'марчен писмо', 'bs-Latn': 'marčensko pismo', 'en': 'Marchen', 'eu': 'martxenera', 'ff-Adlm': '𞤃𞤢𞤪𞤷𞤫𞤲', 'fi': 'tiibetiläinen marchan-kirjoitus', 'ga': 'Marchen', 'gd': 'Marchen', 'he': 'מרצ׳ן', 'hi-Latn': 'Marchen', 'id': 'Marchen', 'ig': 'Marchen', 'kk': 'марчен жазуы', 'kk-Cyrl': 'марчен жазуы', 'mn': 'Марк', 'ms': 'Marchen', 'nl': 'Marchen', 'nqo': 'ߡߊߙߑߛߋ', 'ro': 'marchen', 'sc': 'marchen', 'so': 'Marchen', 'sq': 'markenisht', 'sv': 'marchenska', 'tg': 'Маршенӣ', 'vec': 'marchen', 'zh': '大玛尔文', 'zh-Hans': '大玛尔文', 'zh-Hant': '瑪欽文', 'zu': 'i-Marchen'},
    'Maya': {'ar': 'المايا الهيروغليفية', 'ast': 'xeroglíficos mayes', 'az': 'maya hiyeroqlifi', 'az-Latn': 'maya hiyeroqlifi', 'bg': 'Йероглифи на Маите', 'bn': 'মায়ান হায়ারোগ্লিপ', 'br': 'hieroglifoù mayaek', 'brx': 'माया हीरोग्लीफ्', 'bs': 'majanski hijeroglifi', 'bs-Cyrl': 'мајански хијероглифи', 'bs-Latn': 'majanski hijeroglifi', 'ca': 'jeroglífics maies', 'ca-ES-valencia': 'jeroglífics maies', 'ccp': '𑄟𑄠𑄚𑄴 𑄦𑄠𑄢𑄮𑄉𑄳𑄣𑄨𑄛𑄴', 'cs': 'mayské hieroglyfy', 'da': 'mayahieroglyffer', 'de': 'Maya-Hieroglyphen', 'el': 'Ιερογλυφικά Μάγια', 'el-polyton': 'Ιερογλυφικά Μάγια', 'en': 'Mayan hieroglyphs', 'en-Dsrt': '𐐣𐐴𐐲𐑌 𐐸𐐴𐑉𐐬𐑀𐑊𐐮𐑁𐐮𐐿', 'es': 'jeroglíficos mayas', 'et': 'maaja hieroglüüfkiri', 'fa': 'هیروگلیف مایایی', 'fi': 'maya-hieroglyfit', 'fr': 'hiéroglyphes mayas', 'fur': 'jeroglifics Maya', 'fy': 'Mayahiërogliefen', 'ga': 'Iairiglifí Máigheacha', 'gd': 'Dealbh-sgrìobhadh Mayach', 'gsw': 'Maya-Hieroglyphä', 'gu': 'મયાન હાઇરોગ્લિફ્સ', 'he': 'מאיה', 'hi': 'मयान हाइरोग्लिफ्स', 'hi-Latn': 'Mayan hieroglyphs', 'hr': 'majanski hijeroglifi', 'ht': 'hiéroglyphes mayas', 'hu': 'Maja hieroglifák', 'id': 'Hieroglip Maya', 'it': 'geroglifici maya', 'ja': 'マヤ象形文字', 'ka': 'მაიას იეროგლიფები', 'kgp': 'hijerógrifu maja', 'kn': 'ಮಯಾನ್ ಹೀರೋಗ್ಲಿಫ್ಸ್', 'ko': '마야 상형 문자', 'ks': 'مایَن ہیٖروگلِپھ', 'ks-Arab': 'مایَن ہیٖروگلِپھ', 'lb': 'Maya-Hieroglyphen', 'lo': 'ມາຍາໄຮໂຮກລິບ', 'lt': 'malų hieroglifai', 'lv': 'maiju', 'mk': 'мајански хиероглифи', 'ml': 'മായൻ ചിത്രലിപി', 'mr': 'मायान हाइरोग्लिफ्स', 'nb': 'maya-hieroglyfer', 'ne': 'माया हाइरोग्लिफ्स', 'nl': 'Mayahiërogliefen', 'nn': 'maya-hieroglyfar', 'no': 'maya-hieroglyfer', 'or': 'ମୟାନ୍ ହାୟରଲଜିକସ୍', 'pl': 'hieroglify Majów', 'pt': 'hieróglifos maias', 'rm': 'ieroglifas maya', 'ro': 'hieroglife maya', 'ru': 'майя', 'sk': 'mayské hieroglyfy', 'sl': 'majevska slikovna pisava', 'sr': 'мајански хијероглифи', 'sr-Cyrl': 'мајански хијероглифи', 'sr-Latn': 'majanski hijeroglifi', 'sv': 'mayahieroglyfer', 'ta': 'மயான் ஹைரோகிளிப்', 'te': 'మాయన్ హైరోగ్లైఫ్స్', 'th': 'มายาไฮโรกลิฟส์', 'to': 'tohinima tongitapu fakamaia', 'tr': 'Maya Hiyeroglifleri', 'ug': 'ماياچە تەسۋىرىي يېزىق', 'uk': 'майя ієрогліфічний', 'vi': 'Chữ tượng hình Maya', 'yrl': 'maya-ita kuatiara kuxiímawara', 'yue': '瑪雅象形文字', 'yue-Hans': '玛雅象形文字', 'yue-Hant': '瑪雅象形文字', 'zh': '玛雅圣符文', 'zh-Hans': '玛雅圣符文', 'zh-Hant': '瑪雅象形文字'},
    'Medf': {'bal-Latn': 'Medepáidrin', 'bs': 'medefaidrinsko pismo', 'bs-Cyrl': 'медефаидрин писмо', 'bs-Latn': 'medefaidrinsko pismo', 'en': 'Medefaidrin', 'eu': 'medefaidrinera', 'ff-Adlm': '𞤃𞤢𞤣𞤬𞤫𞤣𞤭𞤪𞤭𞥅𞤲', 'fi': 'medefaidrin', 'gd': 'Medefaidrin', 'he': 'מדפיידרין', 'hi-Latn': 'Medefaidrin', 'id': 'Medefaidrin', 'ig': 'Medefaidrin', 'kk': 'обэри окаимэ жазуы', 'kk-Cyrl': 'обэри окаимэ жазуы', 'mn': 'Медефайдрин', 'ms': 'Medefaidrin', 'nl': 'Medefaidrin', 'nqo': 'ߡߋߘߑߝߘߌߙߌ߲', 'ro': 'medefaidrin', 'sc': 'medefaidrin', 'so': 'Madefaidrin', 'sq': 'medefaidrinisht', 'sv': 'medefaidrin', 'tg': 'Медефаидринӣ', 'vec': 'medefaidrin', 'zh': '梅德法伊德林文', 'zh-Hans': '梅德法伊德林文', 'zu': 'i-Medefaidrin'},
    'Mend': {'ast': 'mende', 'bal-Latn': 'Mende', 'bs': 'mende pismo', 'bs-Cyrl': 'менде писмо', 'bs-Latn': 'mende pismo', 'ca': 'mende', 'ca-ES-valencia': 'mende', 'cs': 'mendské', 'da': 'mende', 'de': 'Mende', 'en': 'Mende', 'et': 'mende', 'eu': 'mende', 'ff-Adlm': '𞤃𞤫𞤲𞤣𞤫', 'fi': 'mende', 'fo': 'mende', 'fy': 'Mende', 'ga': 'Meindeach', 'gd': 'Mende', 'he': 'מנדה', 'hi-Latn': 'Mende', 'hr': 'mende pismo', 'id': 'Mende', 'ig': 'Mende', 'it': 'mende', 'ja': 'メンデ文字', 'ka': 'მენდე', 'kk': 'кикакуи жазуы', 'kk-Cyrl': 'кикакуи жазуы', 'ko': '멘데 문자', 'lo': 'ເມນເດ', 'lt': 'Mende', 'mk': 'мендско', 'mn': 'Менди', 'ms': 'Mende', 'nb': 'mende', 'nl': 'Mende', 'nn': 'mende', 'no': 'mende', 'nqo': 'ߡߍ߲߬ߘߋ', 'ro': 'mende', 'ru': 'менде', 'sc': 'mende', 'so': 'Mende', 'sq': 'mendeisht', 'sv': 'mende', 'tg': 'Менде', 'th': 'เมนเด', 'to': 'tohinima fakamēniti', 'tr': 'Mende', 'ug': 'مېندېچە', 'vec': 'mende', 'vi': 'Chữ Mende', 'yue': '門德文', 'yue-Hans': '门德文', 'yue-Hant': '門德文', 'zh': '门迪文', 'zh-Hans': '门迪文', 'zh-Hant': '門德文', 'zu': 'i-Mende'},
    'Merc': {'ast': 'meroíticu en cursiva', 'bal-Latn': 'Meriuti Karsi', 'bs': 'meroitski kurziv', 'bs-Cyrl': 'меоитиц курзив', 'bs-Latn': 'meroitski kurziv', 'ca': 'cursiva meroítica', 'ca-ES-valencia': 'cursiva meroítica', 'cs': 'meroitické psací', 'da': 'metroitisk sammenhængende', 'de': 'Meroitisch kursiv', 'en': 'Meroitic Cursive', 'et': 'meroe kursiivkiri', 'eu': 'meroitiar etzana', 'ff-Adlm': '𞤃𞤫𞤪𞤱𞤫𞤼𞤭𞤳 𞤅𞤢𞤪𞤰𞤵𞤯𞤭', 'fi': 'meroiittinen kursiivikirjoitus', 'fy': 'Meroitysk cursief', 'ga': 'Meroitic Cursive', 'gd': 'Meroiticeach ceangailte', 'he': 'קורסית מרואיטית', 'hi-Latn': 'Meroitic Cursive', 'hr': 'meroitski kurziv', 'id': 'Kursif Meroitik', 'ig': 'Meroitic Cursive', 'it': 'corsivo meroitico', 'ja': 'メロエ文字草書体', 'ka': 'მეროიტული კურსივი', 'kgp': 'mẽrojitiku nĩgé-tỹ', 'kk': 'мероит курсив жазуы', 'kk-Cyrl': 'мероит курсив жазуы', 'ko': '메로에 필기체', 'lo': 'ເຄເລີຊີເມໂຮອິຕິກ', 'lt': 'Merojitų rankraštinis', 'mk': 'мероитско ракописно', 'mn': 'Меройтик Курсив', 'ms': 'Kursif Meroitic', 'nb': 'meroitisk kursiv', 'nl': 'Meroitisch cursief', 'nn': 'meroitisk kursiv', 'no': 'meroitisk kursiv', 'nqo': 'ߡߋߞߙߏߕߌߞ ߞߎ߳ߙߑߛߌߝ', 'pt': 'meroítico cursivo', 'ro': 'meroitică cursivă', 'ru': 'мероитская курсивная', 'sc': 'corsivu meroìticu', 'so': 'Meroitic Curve', 'sq': 'meroitik kursiv', 'sv': 'kursiv-meroitiska', 'tg': 'Курсиви Мероитӣ', 'th': 'เคอร์ซีฟ-เมโรอิติก', 'to': 'tohinima fakameloue-heihei', 'tr': 'Meroitik El Yazısı', 'ug': 'مېتروئىت يازمىچە', 'vec': 'meroìtego corsivo', 'vi': 'Chữ Meroitic Nét thảo', 'yrl': 'meruítiku kusiwu', 'yue': '麥羅埃文（曲線字體）', 'yue-Hans': '麦罗埃文（曲线字体）', 'yue-Hant': '麥羅埃文（曲線字體）', 'zh': '麦罗埃草书', 'zh-Hans': '麦罗埃草书', 'zh-Hant': '麥羅埃文（曲線字體）', 'zu': 'i-Meroitic Cursive'},
    'Mero': {'ar': 'الميرويتيك', 'ast': 'meroíticu', 'az': 'meroytik', 'az-Latn': 'meroytik', 'bal-Latn': 'Meriuti', 'bg': 'Мероитска', 'bn': 'মেরোইটিক', 'brx': 'मेरोईटीक्', 'bs': 'meroitik pismo', 'bs-Cyrl': 'мероитик писмо', 'bs-Latn': 'meroitik pismo', 'ca': 'meroític', 'ca-ES-valencia': 'meroític', 'ccp': '𑄟𑄬𑄢𑄮𑄃𑄨𑄑𑄨𑄇𑄴', 'cs': 'meroitické', 'da': 'meroitisk', 'de': 'Meroitisch', 'el': 'Μεροϊτικό', 'el-polyton': 'Μεροϊτικό', 'en': 'Meroitic', 'en-Dsrt': '𐐣𐐯𐑉𐐬𐐮𐐻𐐮𐐿', 'es': 'meroítico', 'et': 'meroe', 'eu': 'meroitirra', 'fa': 'مروییتی', 'ff-Adlm': '𞤃𞤫𞤪𞤱𞤫𞤼𞤭𞤳', 'fi': 'meroiittinen', 'fr': 'méroïtique', 'fy': 'Meroïtysk', 'ga': 'Meroitic', 'gd': 'Meroiticeach', 'gsw': 'Meroitisch', 'gu': 'મેરોઇટિક', 'he': 'מרואיטית', 'hi': 'मेरोइटिक', 'hi-Latn': 'Meroitic', 'hr': 'meroitic pismo', 'ht': 'méroïtique', 'hu': 'Meroitikus', 'id': 'Meroitik', 'ig': 'Meroitic', 'it': 'meroitico', 'ja': 'メロエ文字', 'ka': 'მეროიტული', 'kgp': 'mẽrojitiku', 'kk': 'мероит жазуы', 'kk-Cyrl': 'мероит жазуы', 'kn': 'ಮೆರೊಯಿಟಿಕ್', 'ko': '메로에 문자', 'ks': 'مِرایٹِک', 'ks-Arab': 'مِرایٹِک', 'lb': 'Meroitesch', 'lo': 'ເມໂຮຕິກ', 'lt': 'meroitik', 'mk': 'мероитско', 'ml': 'മെറോയിറ്റിക്', 'mn': 'Меройтик', 'mr': 'मेरोइटिक', 'ms': 'Meroitic', 'nb': 'meroitisk', 'ne': 'मेरियोटिक', 'nl': 'Meroïtisch', 'nn': 'meroitisk', 'no': 'meroitisk', 'nqo': 'ߡߋߙߏߦߕߌߞ', 'or': 'ମେରୋଇଟିକ୍', 'pl': 'meroickie', 'pt': 'meroítico', 'rm': 'meroitic', 'ro': 'meroitică', 'ru': 'мероитская', 'sc': 'meroìticu', 'sl': 'meroitski', 'so': 'Meroitic', 'sq': 'meroitik', 'sr': 'мероитик писмо', 'sr-Cyrl': 'мероитик писмо', 'sr-Latn': 'meroitik pismo', 'sv': 'meroitiska', 'ta': 'மெராய்டிக்', 'te': 'మెరోఇటిక్', 'tg': 'Мероитӣ', 'th': 'เมโรติก', 'to': 'tohinima fakameloue', 'tr': 'Meroitik', 'ug': 'مېتروئىت', 'uk': 'мероїтський', 'vec': 'meroìtego', 'vi': 'Chữ Meroitic', 'xnr': 'मेरोइटिक', 'yrl': 'meruítiku', 'yue': '麥羅埃文', 'yue-Hans': '麦罗埃文', 'yue-Hant': '麥羅埃文', 'zh': '麦若提克文', 'zh-Hans': '麦若提克文', 'zh-Hant': '麥羅埃文', 'zu': 'i-Meroitic'},
    'Mlym': {'af': 'Malabaars', 'ak': 'Malayalam kasa', 'am': 'ማላይላም', 'ar': 'الماليالام', 'as': 'মালায়ালম', 'ast': 'malayalam', 'az': 'malayalam', 'az-Latn': 'malayalam', 'bal-Latn': 'Malyálam', 'be': 'малаялам', 'be-tarask': 'малаялам', 'bg': 'малаялам', 'bn': 'মালায়ালাম', 'br': 'malayalam', 'brx': 'मलयालम', 'bs': 'malajalamsko pismo', 'bs-Cyrl': 'малајалам писмо', 'bs-Latn': 'malajalamsko pismo', 'ca': 'malaiàlam', 'ca-ES-valencia': 'malaiàlam', 'ccp': '𑄟𑄣𑄠𑄣𑄟𑄴', 'ce': 'малаялийн', 'chr': 'ᎹᎳᏯᎳᎻ', 'ckb': 'مالایالام', 'cs': 'malajlámské', 'cy': 'Malayalamaidd', 'da': 'malayalam', 'de': 'Malayalam', 'dsb': 'malayalamski', 'dz': 'མ་ལ་ཡ་ལམ་ཡིག་གུ', 'ee': 'malayagbeŋɔŋlɔ', 'el': 'Μαλαγιάλαμ', 'el-polyton': 'Μαλαγιάλαμ', 'en': 'Malayalam', 'en-Dsrt': '𐐣𐐲𐑊𐐩𐐲𐑊𐐪𐑋', 'es': 'malayálam', 'et': 'malajalami', 'eu': 'malabartarra', 'fa': 'مالایالامی', 'ff-Adlm': '𞤃𞤢𞤤𞤢𞤴𞤢𞤤𞤢𞤥', 'fi': 'malajalamilainen', 'fil': 'Malayalam', 'fo': 'malayalam', 'fr': 'malayalam', 'fur': 'malayalam', 'fy': 'Malayalam', 'ga': 'Mailéalamach', 'gd': 'Malayalam', 'gl': 'malabar', 'gsw': 'Malaysisch', 'gu': 'મલયાલમ', 'ha': 'Yaren Malayalam', 'he': 'מליאלאם', 'hi': 'मलयालम', 'hi-Latn': 'Malayalam', 'hr': 'malajalamsko pismo', 'hsb': 'malayalamsce', 'ht': 'malayalam', 'hu': 'Malajálam', 'hy': 'մալայալամ', 'ia': 'malayalam', 'id': 'Malayalam', 'ig': 'Malayalam', 'is': 'malalajam', 'it': 'malayalam', 'ja': 'マラヤーラム文字', 'jv': 'Malayalam', 'ka': 'მალაიალამური', 'kea': 'malaialam', 'kgp': 'marajara', 'kk': 'малаялам жазуы', 'kk-Cyrl': 'малаялам жазуы', 'km': 'ម៉ាឡាយ៉ាឡាម', 'kn': 'ಮಲಯಾಳಂ', 'ko': '말라얄람 문자', 'kok': 'मल्याळम', 'kok-Deva': 'मल्याळम', 'ks': 'مَلیالَم', 'ks-Arab': 'مَلیالَم', 'ksh': 'indesche Malajalam-Schreff', 'ku': 'malayamî', 'kxv': 'malayalam', 'kxv-Deva': 'मलयालम', 'kxv-Latn': 'malayalam', 'kxv-Orya': 'ମାଲାୟଲମ୍', 'kxv-Telu': 'మలయాళం', 'ky': 'Малайалам', 'lb': 'Malaysesch', 'lo': 'ມາເລຢາລາມ', 'lrc': 'مالایام', 'lt': 'malajalių', 'lv': 'malajalu', 'mai': 'मलयालम', 'mi': 'Maraiārama', 'mk': 'малајалско писмо', 'ml': 'മലയാളം', 'mn': 'малаялам', 'mr': 'मल्याळम', 'ms': 'Malayalam', 'my': 'မလေယာလမ်', 'nb': 'malayalam', 'ne': 'मलायालम', 'nl': 'Malayalam', 'nn': 'malayalam', 'no': 'malayalam', 'nqo': 'ߡߟߊߦߊߟߊ߲', 'or': 'ମାଲାୟାଲମ୍', 'pa': 'ਮਲਿਆਲਮ', 'pa-Guru': 'ਮਲਿਆਲਮ', 'pcm': 'Maléyálam', 'pl': 'malajalam', 'ps': 'مالایالم', 'pt': 'malaiala', 'qu': 'Malayalam Simi', 'rm': 'malaisian', 'ro': 'malayalam', 'ru': 'малаялам', 'sc': 'malayalam', 'sd': 'ملايالم', 'sd-Arab': 'ملايالم', 'si': 'මලයාලම්', 'sk': 'malajálamske', 'sl': 'malajalamski', 'so': 'Maalayalam', 'sq': 'malajalam', 'sr': 'малајаламско писмо', 'sr-Cyrl': 'малајаламско писмо', 'sr-Latn': 'malajalamsko pismo', 'sv': 'malayalam', 'sw': 'Kimalayalam', 'ta': 'மலையாளம்', 'te': 'మలయాళం', 'tg': 'Малаяламӣ', 'th': 'มาลายาลัม', 'ti': 'ማላያላም', 'tk': 'Malaýalam elipbiýi', 'to': 'tohinima fakaʻinitia-malāialami', 'tr': 'Malayalam', 'ug': 'مالايامچە', 'uk': 'малаяламська', 'ur': 'ملیالم', 'uz': 'malayalam', 'uz-Cyrl': 'Малайалам', 'uz-Latn': 'malayalam', 'vec': 'malayàlam', 'vi': 'Chữ Malayalam', 'wae': 'Malaisiš', 'xnr': 'मलयालम', 'yo': 'Málàyálámù', 'yrl': 'marayara', 'yue': '馬來亞拉姆文', 'yue-Hans': '马来亚拉姆文', 'yue-Hant': '馬來亞拉姆文', 'zh': '马拉雅拉姆文', 'zh-Hans': '马拉雅拉姆文', 'zh-Hant': '馬來亞拉姆文', 'zu': 'isi-Malayalam'},
    'Modi': {'bal-Latn': 'Módi', 'bs': 'modi pismo', 'bs-Cyrl': 'моди писмо', 'bs-Latn': 'modi pismo', 'cs': 'modí', 'eu': 'modiera', 'ff-Adlm': '𞤃𞤮𞤣𞤭', 'fi': 'modi-aakkoset', 'he': 'מודי', 'ja': 'モーディー文字', 'kk': 'моди жазуы', 'kk-Cyrl': 'моди жазуы', 'mk': 'моди', 'mn': 'Моди', 'nqo': 'ߡߏߘߌ', 'so': 'Moodi', 'sq': 'modisht', 'sv': 'modiska', 'tg': 'Модӣ', 'th': 'โมฑี', 'to': 'tohinima fakamotī', 'zh': '莫迪文', 'zh-Hans': '莫迪文', 'zh-Hant': '馬拉地文', 'zu': 'i-Modi', 'en': 'Modi'},
    'Mong': {'af': 'Mongools', 'ak': 'Mongoliafoɔ kasa', 'am': 'ሞንጎሊያኛ', 'ar': 'المغولية', 'as': 'মঙ্গোলিয়', 'ast': 'mongol', 'az': 'monqol', 'az-Arab': 'موْغولجا', 'az-Latn': 'monqol', 'bal-Latn': 'Mongóli', 'be': 'старамангольскае', 'be-tarask': 'старамангольскае', 'bg': 'монголска', 'bn': 'মোঙ্গোলীয়', 'br': 'mongolek', 'brx': 'मंगलियान', 'bs': 'mongolsko pismo', 'bs-Cyrl': 'монголско писмо', 'bs-Latn': 'mongolsko pismo', 'ca': 'mongol', 'ca-ES-valencia': 'mongol', 'ccp': '𑄟𑄮𑄋𑄴𑄉𑄮𑄣𑄩𑄠𑄧', 'ce': 'монголийн', 'chr': 'ᎹᏂᎪᎵᎠᏂ', 'ckb': 'مەنگۆلی', 'cs': 'mongolské', 'cy': 'Mongolaidd', 'da': 'mongolsk', 'de': 'Mongolisch', 'dsb': 'mongolski', 'dz': 'སོག་པོའི་ཡིག་གུ', 'ee': 'mongoliagbeŋɔŋlɔ', 'el': 'Μογγολικό', 'el-polyton': 'Μογγολικό', 'en': 'Mongolian', 'en-Dsrt': '𐐣𐐱𐑍𐐬𐑊𐐨𐐲𐑌', 'es': 'mongol', 'et': 'mongoli', 'eu': 'mongoliarra', 'fa': 'مغولی', 'ff-Adlm': '𞤃𞤮𞤲𞤺𞤮𞤤𞤭𞤴𞤢𞤲', 'fi': 'mongolilainen', 'fil': 'Mongolian', 'fo': 'mongolsk', 'fr': 'mongol', 'fur': 'mongul', 'fy': 'Mongools', 'ga': 'Mongólach', 'gd': 'Mongolais', 'gl': 'mongol', 'gsw': 'Mongolisch', 'gu': 'મોંગોલિયન', 'ha': 'Na kasar Mongolia', 'he': 'מונגולי', 'hi': 'मंगोलियाई', 'hi-Latn': 'Mongolian', 'hr': 'mongolsko pismo', 'hsb': 'mongolsce', 'ht': 'mongol', 'hu': 'Mongol', 'hy': 'մոնղոլական', 'ia': 'mongol', 'id': 'Mongolia', 'ig': 'Mongolian', 'is': 'mongólskt', 'it': 'mongolo', 'ja': 'モンゴル文字', 'jv': 'Mongolia', 'ka': 'მონღოლური', 'kea': 'mongol', 'kgp': 'mãgór', 'kk': 'моңғол жазуы', 'kk-Cyrl': 'моңғол жазуы', 'km': 'ម៉ុងហ្គោលី', 'kn': 'ಮಂಗೋಲಿಯನ್', 'ko': '몽골 문자', 'kok': 'मोंगोलियन', 'kok-Deva': 'मोंगोलियन', 'ks': 'مَنگولیَن', 'ks-Arab': 'مَنگولیَن', 'ksh': 'mongjoolesche Schreff', 'ku': 'moxolî', 'ky': 'Монгол', 'lb': 'Mongolesch', 'lo': 'ມົງໂກນ', 'lrc': 'موغولی', 'lt': 'mongolų', 'lv': 'mongoļu', 'mai': 'मंंगोलियन', 'mi': 'Mongōria', 'mk': 'монголско писмо', 'ml': 'മംഗോളിയൻ', 'mn': 'монгол бичиг', 'mr': 'मंगोलियन', 'ms': 'Mongolia', 'my': 'မွန်ဂိုလီးယား', 'nb': 'mongolsk', 'ne': 'मङ्गोल', 'nl': 'Mongools', 'nn': 'mongolsk', 'no': 'mongolsk', 'nqo': 'ߡߐ߲ߜ߭ߐߟ', 'or': 'ମଙ୍ଗୋଲିଆନ୍', 'pa': 'ਮੰਗੋਲੀਅਨ', 'pa-Guru': 'ਮੰਗੋਲੀਅਨ', 'pcm': 'Mọngólia', 'pl': 'mongolskie', 'ps': 'منګولیایي', 'pt': 'mongol', 'qu': 'Mongol Simi', 'rm': 'mongolic', 'ro': 'mongolă', 'ru': 'монгольская', 'sah': 'Моҕуоллуу', 'sc': 'mòngolu', 'sd': 'منگولي', 'sd-Arab': 'منگولي', 'si': 'මොන්ගෝලියානු', 'sk': 'mongolské', 'sl': 'mongolska', 'so': 'Mongooliyaan', 'sq': 'mongolisht', 'sr': 'монголско писмо', 'sr-Cyrl': 'монголско писмо', 'sr-Latn': 'mongolsko pismo', 'sv': 'mongoliska', 'sw': 'Kimongolia', 'ta': 'மங்கோலியன்', 'te': 'మంగోలియన్', 'tg': 'Муғулӣ', 'th': 'มองโกเลีย', 'ti': 'ማኦንጎላዊ', 'tk': 'Mongol elipbiýi', 'to': 'tohinima fakamongokōlia', 'tr': 'Moğol', 'ug': 'موڭغۇلچە', 'uk': 'монгольська', 'ur': 'منگولیائی', 'uz': 'mongol', 'uz-Cyrl': 'Мўғулча', 'uz-Latn': 'mongol', 'vec': 'móngolo', 'vi': 'Chữ Mông Cổ', 'xnr': 'मंगोलियाई', 'yo': 'Èdè Mòngólíà', 'yrl': 'mũgú', 'yue': '蒙古文', 'yue-Hans': '蒙古文', 'yue-Hant': '蒙古文', 'zh': '蒙古文', 'zh-Hans': '蒙古文', 'zh-Hant': '蒙古文', 'zu': 'isi-Mongolian'},
    'Moon': {'ar': 'مون', 'ast': 'tipos Moon', 'az': 'mun', 'az-Latn': 'mun', 'bg': 'Мун', 'bn': 'মুন', 'brx': 'मुन्', 'bs': 'munova azbuka', 'bs-Cyrl': 'месечево писмо', 'bs-Latn': 'munova azbuka', 'ccp': '𑄟𑄪𑄚𑄴', 'cs': 'Moonovo písmo', 'el': 'Μουν', 'el-polyton': 'Μουν', 'en-Dsrt': '𐐣𐐭𐑌', 'et': 'Mooni', 'fa': 'مونی', 'fi': 'moon-kohokirjoitus', 'gu': 'મૂન', 'hi': 'मून', 'hr': 'moon pismo', 'ja': 'ムーン文字', 'kgp': 'mũn', 'kn': 'ಮೂನ್', 'ko': '문 문자', 'ks': 'موٗن', 'ks-Arab': 'موٗن', 'lo': 'ມູນ', 'lt': 'mūn', 'lv': 'Mūna raksts', 'mk': 'Муново', 'ml': 'മൂൺ', 'mr': 'मून', 'ne': 'जून', 'or': 'ଚନ୍ଦ୍ର', 'pl': 'Moon’a', 'ru': 'азбука муна', 'sl': 'Moonova pisava za slepe', 'sr': 'месечево писмо', 'sr-Cyrl': 'месечево писмо', 'sr-Latn': 'mesečevo pismo', 'ta': 'மூன்', 'te': 'మూన్', 'th': 'มูน', 'to': 'tohinima laukonga ki he kui-māhina', 'ug': 'كورىيە', 'uk': 'мун', 'vi': 'Chữ nổi Moon', 'yrl': 'Moon kuatiara', 'yue': '蒙氏點字', 'yue-Hans': '蒙氏点字', 'yue-Hant': '蒙氏點字', 'zh': '穆恩字母', 'zh-Hans': '穆恩字母', 'zh-Hant': '蒙氏點字', 'en': 'Moon'},
    'Mroo': {'ast': 'mro', 'bal-Latn': 'Mró', 'bs': 'mro pismo', 'bs-Cyrl': 'мро писмо', 'bs-Latn': 'mro pismo', 'ca': 'mro', 'ca-ES-valencia': 'mro', 'cs': 'mro', 'de': 'Mro', 'en': 'Mro', 'et': 'mruu', 'eu': 'mroera', 'ff-Adlm': '𞤃𞤮𞤪𞤮𞥅', 'fi': 'mro', 'fy': 'Mro', 'ga': 'Mro', 'gd': 'Mro', 'he': 'מרו', 'hi-Latn': 'Mro', 'hr': 'mro pismo', 'id': 'Mro', 'ig': 'Mro', 'it': 'mro', 'ja': 'ムロ文字', 'ka': 'მრო', 'kk': 'мро жазуы', 'kk-Cyrl': 'мро жазуы', 'ko': '므로 문자', 'lo': 'ເມໂຮ', 'lt': 'Mro', 'mk': 'мро', 'mn': 'Мру', 'ms': 'Mro', 'nb': 'mro', 'nl': 'Mro', 'nn': 'mro', 'no': 'mro', 'nqo': 'ߡߑߙߏ', 'ro': 'mro', 'ru': 'мро', 'sc': 'mro', 'so': 'Mro', 'sq': 'mroisht', 'sv': 'mru', 'tg': 'Мро', 'th': 'มโร', 'to': 'tohinima fakamolō', 'tr': 'Mro', 'ug': 'مروچە', 'vec': 'mro', 'vi': 'Chữ Mro', 'yue': '謬文', 'yue-Hans': '谬文', 'yue-Hant': '謬文', 'zh': '谬文', 'zh-Hans': '谬文', 'zh-Hant': '謬文', 'zu': 'i-Mro'},
    'Mtei': {'af': 'Meitei-Majek', 'ak': 'Meeti Mayɛke kasa', 'am': 'ሜቴ ማይክ', 'ar': 'ميتي ماييك', 'as': 'মেইতেই মায়েক', 'ast': 'meitei mayek', 'az': 'meytey mayek', 'az-Latn': 'meytey mayek', 'bal-Latn': 'Miyeti Máyek', 'be': 'маніпуры', 'be-tarask': 'маніпуры', 'bg': 'манипури', 'bn': 'মেইটেই মায়েক', 'brx': 'मैतेई मायेक', 'bs': 'meitei majek pismo', 'bs-Cyrl': 'меитеи мајек писмо', 'bs-Latn': 'meitei majek pismo', 'ca': 'manipuri', 'ca-ES-valencia': 'manipuri', 'ccp': '𑄟𑄳𑄆𑄬𑄑𑄳𑄆𑄬 𑄟𑄠𑄬𑄇𑄴', 'chr': 'ᎺᏖ ᎹᏰᎩ', 'cs': 'mejtej majek (manipurské)', 'cy': 'Meitei Mayek', 'da': 'meitei-mayek', 'de': 'Meitei-Mayek', 'dsb': 'meitei-mayekse pismo', 'el': 'Μεϊτέι Μάγεκ', 'el-polyton': 'Μεϊτέι Μάγεκ', 'en': 'Meitei Mayek', 'en-Dsrt': '𐐣𐐩𐐻𐐩 𐐣𐐴𐐯𐐿', 'es': 'meitei', 'et': 'meitei', 'eu': 'meiteiarra', 'fa': 'مایک میتی', 'ff-Adlm': '𞤃𞤫𞤼𞤭 𞤃𞤢𞤴𞤫𞤳', 'fi': 'meitei', 'fil': 'Meitei Mayek', 'fo': 'meitei mayek', 'fr': 'meitei mayek', 'fy': 'Meitei', 'ga': 'Meitei Mayek', 'gd': 'Meitei Mayek', 'gl': 'meitei mayek', 'gsw': 'Meitei Mayek', 'gu': 'મેઇતેઇ માયેક', 'ha': 'Meitei Mayek', 'he': 'מאיטי מאייק', 'hi': 'मेइतेइ मायेक', 'hi-Latn': 'Meitei Mayek', 'hr': 'meitei mayek pismo', 'hsb': 'meitei-mayekse pismo', 'ht': 'meitei mayek', 'hu': 'Meitei mayek', 'hy': 'մանիպուրի', 'id': 'Meitei Mayek', 'ig': 'Meitei Mayek', 'is': 'meitei mayek', 'it': 'meetei mayek', 'ja': 'メイテイ文字', 'jv': 'Meitei Mayek', 'ka': 'მანიპური', 'kgp': 'mẽjtej mỹjéki', 'kk': 'мейтей жазуы', 'kk-Cyrl': 'мейтей жазуы', 'km': 'ម៉ីតីម៉ាយ៉ែក', 'kn': 'ಮೈತಿ ಮಯೆಕ್', 'ko': '메이테이 마옉 문자', 'kok': 'मेयतेई मायेक', 'kok-Deva': 'मेयतेई मायेक', 'ks': 'میتی مایَک', 'ks-Arab': 'میتی مایَک', 'ky': 'мейтей-маек (жазуу)', 'lb': 'Meitei Mayek', 'lo': 'ເມເທມາເຍກ', 'lt': 'meitei majek', 'lv': 'meitei-majeku', 'mai': 'मैती मायेक', 'mi': 'Meitei Maeke', 'mk': 'мејтејско', 'ml': 'മേറ്റി മായക്', 'mn': 'Мейтей маек бичиг', 'mni': 'মীতৈ ময়েক', 'mni-Beng': 'মীতৈ ময়েক', 'mni-Mtei': 'ꯃꯤꯇꯩ ꯃꯌꯦꯛ', 'mr': 'मेइतेइ मायेक', 'ms': 'Meitei Mayek', 'my': 'မေတဲမာယက်', 'nb': 'meitei-mayek', 'ne': 'माइटेइ मायेक', 'nl': 'Meitei', 'nn': 'meitei-mayek', 'no': 'meitei-mayek', 'nqo': 'ߡߋߦߌߕߋ ߡߊߦߌߞ', 'or': 'ମାଏତି ମାୟେକ୍', 'pa': 'ਮਿਤੇਈ ਮਾਏਕ', 'pa-Guru': 'ਮਿਤੇਈ ਮਾਏਕ', 'pcm': 'Mẹitẹí Mayẹk Lángwej', 'pl': 'meitei mayek', 'ps': 'میټي مایک', 'pt': 'meitei mayek', 'qu': 'Meitei Mayek Simi', 'rm': 'meetei mayek', 'ro': 'meitei mayek', 'ru': 'манипури', 'sc': 'meitei mayek', 'sd': 'ميئيٽي مائيڪ', 'sd-Arab': 'ميئيٽي مائيڪ', 'si': 'මෙයිටෙයි මයක්', 'sk': 'mejtej majek (manipurské)', 'sl': 'meiteiski', 'so': 'Qoraalka Luuqada Meitei', 'sq': 'mitei-majek', 'sr': 'меитеи мајек', 'sr-Cyrl': 'меитеи мајек', 'sr-Latn': 'meitei majek', 'sv': 'meitei-mayek', 'sw': 'Meitei Mayek', 'ta': 'மெய்தெய் மயக்', 'te': 'మీటి మయెక్', 'tg': 'Мейтеи Майек', 'th': 'เมเทมาเยก', 'ti': 'መይተይ ማየክ', 'tk': 'meýteý-maýek', 'to': 'tohinima fakametei-maieki', 'tr': 'Meitei Mayek', 'ug': 'مانىپۇرىچە', 'uk': 'мейтей майєк', 'ur': 'میتی مئیک', 'uz': 'manipuri', 'uz-Latn': 'manipuri', 'vec': 'meitéi', 'vi': 'Chữ Meitei Mayek', 'xnr': 'मेइतेइ मायेक', 'yo': 'Èdè Meitei Mayeki', 'yrl': 'manipuri kuatiara', 'yue': '曼尼普爾文', 'yue-Hans': '曼尼普尔文', 'yue-Hant': '曼尼普爾文', 'zh': '曼尼普尔文', 'zh-Hans': '曼尼普尔文', 'zh-Hant': '曼尼普爾文', 'zu': 'isi-Meitei Mayek'},
    'Mult': {'ast': 'multani', 'bal-Latn': 'Moltáni', 'bs': 'multani pismo', 'bs-Cyrl': 'мултани писмо', 'bs-Latn': 'multani pismo', 'ca': 'multani', 'ca-ES-valencia': 'multani', 'en': 'Multani', 'et': 'Multani', 'eu': 'multaniera', 'ff-Adlm': '𞤃𞤵𞤤𞤼𞤢𞤲𞤭', 'fi': 'multanilainen', 'ga': 'Multani', 'gd': 'Multani', 'he': 'מולטיני', 'hi-Latn': 'Multani', 'id': 'Multani', 'ig': 'Multani', 'kk': 'мултани жазуы', 'kk-Cyrl': 'мултани жазуы', 'mn': 'Мултани', 'ms': 'Multani', 'nb': 'multani', 'nl': 'Multani', 'nn': 'multani', 'no': 'multani', 'nqo': 'ߡߎߟߑߕߊߣߌ', 'ro': 'multani', 'sc': 'multani', 'sd': 'ملتاني', 'sd-Arab': 'ملتاني', 'so': 'Multani', 'sq': 'multanisht', 'sv': 'multaniska', 'tg': 'Мултанӣ', 'vec': 'multani', 'zh': '穆尔坦文', 'zh-Hans': '穆尔坦文', 'zh-Hant': '木爾坦文', 'zu': 'i-Multani'},
    'Mymr': {'af': 'Mianmar', 'ak': 'Mayama kasa', 'am': 'ምያንማር', 'ar': 'الميانمار', 'as': 'ম্যানমাৰ', 'ast': 'birmanu', 'az': 'myanmar', 'az-Latn': 'myanmar', 'bal-Latn': 'Myanmár', 'be': 'бірманскае', 'be-tarask': 'бірманскае', 'bg': 'бирманска', 'bn': 'মায়ানমার', 'br': 'myanmar', 'brx': 'म्यानमार', 'bs': 'mijanmarsko pismo', 'bs-Cyrl': 'мијанмарско писмо', 'bs-Latn': 'mijanmarsko pismo', 'ca': 'birmà', 'ca-ES-valencia': 'birmà', 'ccp': '𑄟𑄠𑄚𑄴𑄟𑄢𑄴', 'ce': 'мьянманийн', 'chr': 'ᎹᎡᏂᎹᎳ', 'ckb': 'میانمار', 'cs': 'myanmarské', 'cy': 'Myanmaraidd', 'da': 'burmesisk', 'de': 'Birmanisch', 'dsb': 'burmaski', 'dz': 'བར་མིས་ཡིག་གུ', 'ee': 'myanmargbeŋɔŋlɔ', 'el': 'Μιανμάρ', 'el-polyton': 'Μιανμάρ', 'en': 'Myanmar', 'en-Dsrt': '𐐣𐐨𐐲𐑌𐑋𐐪𐑉', 'es': 'birmano', 'et': 'birma', 'eu': 'birmaniarra', 'fa': 'میانمار', 'ff-Adlm': '𞤃𞤭𞤴𞤢𞤥𞤢𞥄𞤪', 'fi': 'burmalainen', 'fil': 'Myanmar', 'fo': 'myanmarskt', 'fr': 'birman', 'fur': 'myanmar', 'fy': 'Myanmar', 'ga': 'Maenmarach', 'gd': 'Miànmar', 'gl': 'birmano', 'gsw': 'Burmesisch', 'gu': 'મ્યાંમાર', 'ha': 'Ƙasar Myanmar', 'he': 'מיאנמר', 'hi': 'म्यांमार', 'hi-Latn': 'Burmese', 'hr': 'mjanmarsko pismo', 'hsb': 'burmasce', 'ht': 'birman', 'hu': 'Burmai', 'hy': 'մյանմարական', 'ia': 'birman', 'id': 'Myanmar', 'ig': 'Myanmar', 'is': 'mjanmarskt', 'it': 'birmano', 'ja': 'ミャンマー文字', 'jv': 'Myanmar', 'ka': 'მიანმური', 'kea': 'birmanes', 'kgp': 'mirmỹnẽj', 'kk': 'мьянма жазуы', 'kk-Cyrl': 'мьянма жазуы', 'km': 'ភូមា', 'kn': 'ಮ್ಯಾನ್ಮಾರ್', 'ko': '미얀마 문자', 'kok': 'म्यानमार', 'kok-Deva': 'म्यानमार', 'ks': 'مَیَنمار', 'ks-Arab': 'مَیَنمار', 'ksh': 'birmahnesche Schreff', 'ku': 'myanmarî', 'ky': 'Мйанмар', 'lb': 'Birmanesch', 'lo': 'ມຽນມາ', 'lrc': 'میانمار', 'lt': 'birmiečių', 'lv': 'birmiešu', 'mai': 'मयनमार', 'mi': 'Mienemā', 'mk': 'мјанмарско писмо', 'ml': 'മ്യാൻമാർ', 'mn': 'мьянмар', 'mr': 'म्यानमार', 'ms': 'Myammar', 'my': 'မြန်မာ', 'nb': 'burmesisk', 'ne': 'म्यान्मार', 'nl': 'Birmaans', 'nn': 'burmesisk', 'no': 'burmesisk', 'nqo': 'ߓߙߌߡߊ߲', 'or': 'ମ୍ୟାନମାର୍', 'pa': 'ਮਿਆਂਮਾਰ', 'pa-Guru': 'ਮਿਆਂਮਾਰ', 'pcm': 'Miánmar', 'pl': 'birmańskie', 'ps': 'میانمار', 'pt': 'birmanês', 'qu': 'Myanmar', 'rm': 'burmais', 'ro': 'birmană', 'ru': 'мьянманская', 'sc': 'birmanu', 'sd': 'ميانمر', 'sd-Arab': 'ميانمر', 'shn': 'မျၢၼ်ႇမႃႇ', 'si': 'මියන්මාර', 'sk': 'barmské', 'sl': 'mjanmarski', 'so': 'Mayanmaar', 'sq': 'birman', 'sr': 'мијанмарско писмо', 'sr-Cyrl': 'мијанмарско писмо', 'sr-Latn': 'mijanmarsko pismo', 'sv': 'burmesiska', 'sw': 'Myama', 'ta': 'மியான்மர்', 'te': 'మయన్మార్', 'tg': 'Мянмар', 'th': 'พม่า', 'ti': 'ማይንማር', 'tk': 'Mýanma elipbiýi', 'to': 'tohinima fakapema', 'tr': 'Burma', 'ug': 'بىرماچە', 'uk': 'мʼянмська', 'ur': 'میانمار', 'uz': 'myanma', 'uz-Cyrl': 'Мьянма', 'uz-Latn': 'myanma', 'vec': 'birman', 'vi': 'Chữ Myanmar', 'wae': 'Burmesiš', 'xnr': 'म्यांमार', 'yo': 'Myánmarà', 'yrl': 'birimanei', 'yue': '緬甸文', 'yue-Hans': '缅甸文', 'yue-Hant': '緬甸文', 'zh': '缅甸文', 'zh-Hans': '缅甸文', 'zh-Hant': '緬甸文', 'zu': 'isi-Myanmar'},
    'NA': {'af': 'Namibië', 'agq': 'Nàmibìa', 'ak': 'Namibia', 'am': 'ናሚቢያ', 'ar': 'ناميبيا', 'as': 'নামিবিয়া', 'asa': 'Namibia', 'ast': 'Namibia', 'az': 'Namibiya', 'az-Cyrl': 'Намибија', 'az-Latn': 'Namibiya', 'bal-Latn': 'Namibiá', 'bas': 'Nàmibìà', 'be': 'Намібія', 'be-tarask': 'Намібія', 'bez': 'Hunamibia', 'bg': 'Намибия', 'blo': 'Namibii', 'bm': 'Namibi', 'bn': 'নামিবিয়া', 'br': 'Namibia', 'brx': 'नामीबिया', 'bs': 'Namibija', 'bs-Cyrl': 'Намибија', 'bs-Latn': 'Namibija', 'ca': 'Namíbia', 'ca-ES-valencia': 'Namíbia', 'ccp': '𑄚𑄟𑄨𑄝𑄨𑄠', 'ce': 'Намиби', 'ceb': 'Namibia', 'cgg': 'Namibiya', 'chr': 'ᎾᎻᏈᎢᏯ', 'ckb': 'نامیبیا', 'cs': 'Namibie', 'cv': 'Намиби', 'cy': 'Namibia', 'da': 'Namibia', 'dav': 'Namibia', 'de': 'Namibia', 'dje': 'Naamibi', 'dsb': 'Namibija', 'dz': 'ན་མི་བི་ཡ', 'ebu': 'Namibia', 'ee': 'Namibia nutome', 'el': 'Ναμίμπια', 'el-polyton': 'Ναμίμπια', 'en': 'Namibia', 'en-Dsrt': '𐐤𐐲𐑋𐐮𐐺𐐨𐐲', 'eo': 'Namibio', 'es': 'Namibia', 'et': 'Namiibia', 'eu': 'Namibia', 'ewo': 'Namibí', 'fa': 'نامیبیا', 'ff': 'Namibii', 'ff-Adlm': '𞤐𞤢𞤥𞤭𞥅𞤦𞤭𞤴𞤢𞥄', 'ff-Latn': 'Namibii', 'fi': 'Namibia', 'fil': 'Namibia', 'fo': 'Namibia', 'fr': 'Namibie', 'fur': 'Namibie', 'fy': 'Namibië', 'ga': 'an Namaib', 'gd': 'An Namaib', 'gl': 'Namibia', 'gsw': 'Namiibia', 'gu': 'નામિબિયા', 'guz': 'Namibia', 'ha': 'Namibiya', 'he': 'נמיביה', 'hi': 'नामीबिया', 'hi-Latn': 'Namibia', 'hr': 'Namibija', 'hsb': 'Namibija', 'ht': 'Namibie', 'hu': 'Namíbia', 'hy': 'Նամիբիա', 'ia': 'Namibia', 'id': 'Namibia', 'ig': 'Namibia', 'is': 'Namibía', 'it': 'Namibia', 'ja': 'ナミビア', 'jgo': 'Namimbî', 'jmc': 'Namibia', 'jv': 'Namibia', 'ka': 'ნამიბია', 'kaa': 'Намибия', 'kaa-Cyrl': 'Намибия', 'kab': 'Namibya', 'kam': 'Namibia', 'kde': 'Namibia', 'kea': 'Namíbia', 'kgp': 'Nỹmĩmija', 'khq': 'Naamibi', 'ki': 'Namimbia', 'kk': 'Намибия', 'kk-Cyrl': 'Намибия', 'kln': 'Emetab Namibia', 'km': 'ណាមីប៊ី', 'kn': 'ನಮೀಬಿಯಾ', 'ko': '나미비아', 'kok': 'नामीबिया', 'kok-Deva': 'नामीबिया', 'ks': 'نامِبِیا', 'ks-Arab': 'نامِبِیا', 'ksb': 'Namibia', 'ksf': 'namibí', 'ksh': 'Namiibije', 'ku': 'Namîbya', 'kxv': 'namibia', 'kxv-Deva': 'नामीबिया', 'kxv-Latn': 'namibia', 'kxv-Orya': 'ନାମିବିୟା', 'kxv-Telu': 'నమీబియా', 'ky': 'Намибия', 'lag': 'Namíbia', 'lb': 'Namibia', 'lg': 'Namibiya', 'lld': 'Namibia', 'ln': 'Namibi', 'lo': 'ນາມີເບຍ', 'lt': 'Namibija', 'lu': 'Namibi', 'luo': 'Namibia', 'luy': 'Namibia', 'lv': 'Namībija', 'mai': 'नामीबिया', 'mas': 'Namibia', 'mer': 'Namibia', 'mfe': 'Namibi', 'mg': 'Namibia', 'mi': 'Namipia', 'mk': 'Намибија', 'ml': 'നമീബിയ', 'mn': 'Намиби', 'mr': 'नामिबिया', 'ms': 'Namibia', 'mt': 'in-Namibja', 'mua': 'Namibiya', 'my': 'နမီးဘီးယား', 'mzn': 'نامبیا', 'naq': 'Namibiab', 'nb': 'Namibia', 'nd': 'Namibhiya', 'ne': 'नामिबिया', 'nl': 'Namibië', 'nmg': 'Namibia', 'nn': 'Namibia', 'no': 'Namibia', 'nqo': 'ߣߊߡߌ߲ߓߌ߫', 'nyn': 'Namibiya', 'om': 'Namiibiyaa', 'or': 'ନାମିବିଆ', 'pa': 'ਨਾਮੀਬੀਆ', 'pa-Guru': 'ਨਾਮੀਬੀਆ', 'pcm': 'Namíbia', 'pl': 'Namibia', 'ps': 'نیمبیا', 'pt': 'Namíbia', 'qu': 'Namibia', 'rm': 'Namibia', 'rn': 'Namibiya', 'ro': 'Namibia', 'rof': 'Namibia', 'ru': 'Намибия', 'rwk': 'Namibia', 'saq': 'Namibia', 'sbp': 'Namibiya', 'sc': 'Namìbia', 'sd': 'نيميبيا', 'sd-Arab': 'نيميبيا', 'se': 'Namibia', 'seh': 'Namíbia', 'ses': 'Naamibi', 'sg': 'Namibùii', 'shi': 'ⵏⴰⵎⵉⴱⵢⴰ', 'shi-Latn': 'namibya', 'shi-Tfng': 'ⵏⴰⵎⵉⴱⵢⴰ', 'si': 'නැමීබියාව', 'sk': 'Namíbia', 'sl': 'Namibija', 'smn': 'Namibia', 'sn': 'Namibia', 'so': 'Namiibiya', 'sq': 'Namibi', 'sr': 'Намибија', 'sr-Cyrl': 'Намибија', 'sr-Latn': 'Namibija', 'sv': 'Namibia', 'sw': 'Namibia', 'syr': 'ܢܡܝܒܝܐ', 'ta': 'நமீபியா', 'te': 'నమీబియా', 'teo': 'Namibia', 'tg': 'Намибия', 'th': 'นามิเบีย', 'ti': 'ናሚብያ', 'tk': 'Namibiýa', 'to': 'Namipia', 'tr': 'Namibya', 'tt': 'Намибия', 'twq': 'Naamibi', 'tzm': 'Namibya', 'ug': 'نامىبىيە', 'uk': 'Намібія', 'ur': 'نامیبیا', 'uz': 'Namibiya', 'uz-Cyrl': 'Намибия', 'uz-Latn': 'Namibiya', 'vai': 'ꕯꕆꔫꕩ', 'vai-Latn': 'Namibiya', 'vai-Vaii': 'ꕯꕆꔫꕩ', 'vec': 'Namibia', 'vi': 'Namibia', 'vun': 'Namibia', 'wae': 'Namibia', 'wo': 'Namibi', 'xh': 'ENamibia', 'xnr': 'नामीबिया', 'xog': 'Namibiya', 'yav': 'namipí', 'yi': 'נאַמיביע', 'yo': 'Namibia', 'yrl': 'Namíbiya', 'yue': '納米比亞', 'yue-Hans': '纳米比亚', 'yue-Hant': '納米比亞', 'zgh': 'ⵏⴰⵎⵉⴱⵢⴰ', 'zh': '纳米比亚', 'zh-Hans': '纳米比亚', 'zh-Hant': '納米比亞', 'zu': 'i-Namibia'},
    'NC': {'af': 'Nieu-Kaledonië', 'agq': 'Kàlèdonyìa È fūghū', 'ak': 'Kaledonia Foforo', 'am': 'ኒው ካሌዶኒያ', 'ar': 'كاليدونيا الجديدة', 'as': 'নিউ কেলিডোনিয়া', 'asa': 'Nyukaledonia', 'ast': 'Nueva Caledonia', 'az': 'Yeni Kaledoniya', 'az-Cyrl': 'Јени Каледонија', 'az-Latn': 'Yeni Kaledoniya', 'bal-Latn': 'Niu Káledóniá', 'bas': 'Kàlèdonìà Yɔ̀ndɔ', 'be': 'Новая Каледонія', 'be-tarask': 'Новая Каледонія', 'bez': 'Hunyukaledonia', 'bg': 'Нова Каледония', 'blo': 'Kaleɖonii afɔlɩ', 'bm': 'Kaledoni Koura', 'bn': 'নিউ ক্যালেডোনিয়া', 'br': 'Kaledonia Nevez', 'brx': 'निउ केलेडनिया', 'bs': 'Nova Kaledonija', 'bs-Cyrl': 'Нова Каледонија', 'bs-Latn': 'Nova Kaledonija', 'ca': 'Nova Caledònia', 'ca-ES-valencia': 'Nova Caledònia', 'ccp': '𑄚𑄱 𑄇𑄳𑄠𑄣𑄬𑄓𑄮𑄚𑄨𑄠', 'ce': 'Керла Каледони', 'ceb': 'New Caledonia', 'cgg': 'Niukaredonia', 'chr': 'ᎢᏤ ᎧᎵᏙᏂᎠᏂ', 'ckb': 'نیووکالێدۆنیا', 'cs': 'Nová Kaledonie', 'cv': 'Ҫӗнӗ Каледони', 'cy': 'Caledonia Newydd', 'da': 'Ny Kaledonien', 'dav': 'Nyukaledonia', 'de': 'Neukaledonien', 'dje': 'Kaaledooni Taagaa', 'dsb': 'Nowa Kaledoniska', 'dz': 'ནིའུ་ཀ་ལི་དོ་ནི་ཡ', 'ebu': 'Nyukaledonia', 'ee': 'New Kaledonia nutome', 'el': 'Νέα Καληδονία', 'el-polyton': 'Νέα Καληδονία', 'en': 'New Caledonia', 'en-Dsrt': '𐐤𐐭 𐐗𐐰𐑊𐐲𐐼𐐬𐑌𐐷𐐲', 'eo': 'Nov-Kaledonio', 'es': 'Nueva Caledonia', 'et': 'Uus-Kaledoonia', 'eu': 'Kaledonia Berria', 'ewo': 'Ǹkpámɛn Kaledónia', 'fa': 'کالدونیای جدید', 'ff': 'Nuwel Kaledonii', 'ff-Adlm': '𞤑𞤢𞤤𞤭𞤣𞤮𞤲𞤭𞤴𞤢𞥄 𞤖𞤫𞤧𞤮', 'ff-Latn': 'Nuwel Kaledonii', 'fi': 'Uusi-Kaledonia', 'fil': 'New Caledonia', 'fo': 'Nýkaledónia', 'fr': 'Nouvelle-Calédonie', 'fur': 'Gnove Caledonie', 'fy': 'Nij-Caledonië', 'ga': 'an Nua-Chaladóin', 'gd': 'Cailleann Nuadh', 'gl': 'Nova Caledonia', 'gsw': 'Nöikaledoonie', 'gu': 'ન્યુ સેલેડોનિયા', 'guz': 'Nyukaledonia', 'ha': 'Kaledoniya Sabuwa', 'he': 'קלדוניה החדשה', 'hi': 'न्यू कैलेडोनिया', 'hi-Latn': 'New Caledonia', 'hr': 'Nova Kaledonija', 'hsb': 'Nowa Kaledoniska', 'ht': 'Nouvelle-Calédonie', 'hu': 'Új-Kaledónia', 'hy': 'Նոր Կալեդոնիա', 'ia': 'Nove Caledonia', 'id': 'Kaledonia Baru', 'ig': 'New Caledonia', 'is': 'Nýja-Kaledónía', 'it': 'Nuova Caledonia', 'ja': 'ニューカレドニア', 'jmc': 'Nyukaledonia', 'jv': 'Kalédonia Anyar', 'ka': 'ახალი კალედონია', 'kaa': 'Жаӊа Каледония', 'kaa-Cyrl': 'Жаӊа Каледония', 'kab': 'Kalidunya Tamaynut', 'kam': 'Nyukaledonia', 'kde': 'Nyukaledonia', 'kea': 'Nova Kalidónia', 'kgp': 'Karenonĩja Tãg', 'khq': 'Kaaledooni Taagaa', 'ki': 'Nyukaledonia', 'kk': 'Жаңа Каледония', 'kk-Cyrl': 'Жаңа Каледония', 'kln': 'Emetab New Caledonia', 'km': 'នូវែល\u200bកាឡេដូនី', 'kn': 'ನ್ಯೂ ಕ್ಯಾಲಿಡೋನಿಯಾ', 'ko': '뉴칼레도니아', 'kok': 'न्यू कॅलिडोनिया', 'kok-Deva': 'न्यू कॅलिडोनिया', 'ks': 'نِو کیلِڑونِیا', 'ks-Arab': 'نِو کیلِڑونِیا', 'ksb': 'Nyukaledonia', 'ksf': 'kalɛdoní anyɔ́n', 'ksh': 'Neuschottland', 'ku': 'Kaledonyaya Nû', 'kxv': 'nyu keleḍoniya', 'kxv-Deva': 'न्यू केलेडोनिया', 'kxv-Latn': 'nyu keleḍoniya', 'kxv-Orya': 'ନ୍ୟୁ କେଲେଡନିୟା', 'kxv-Telu': 'క్రొత్త కెలెడోనియా', 'ky': 'Жаңы Каледония', 'lag': 'Kaledónia Ifya', 'lb': 'Neikaledonien', 'lg': 'Kaledonya mupya', 'lld': 'Nöia Caledonia', 'ln': 'Kaledoni ya sika', 'lo': 'ນິວ ຄາເລໂດເນຍ', 'lt': 'Naujoji Kaledonija', 'lu': 'Kaledoni wa mumu', 'luo': 'New Caledonia', 'luy': 'Nyukaledonia', 'lv': 'Jaunkaledonija', 'mai': 'न्यू कैलेडोनिया', 'mas': 'Nyukaledonia', 'mer': 'Kalendoia Ĩnjeru', 'mfe': 'Nouvel-Kaledoni', 'mg': 'Nouvelle-Calédonie', 'mi': 'Whenua Kanaki', 'mk': 'Нова Каледонија', 'ml': 'ന്യൂ കാലിഡോണിയ', 'mn': 'Шинэ Каледони', 'mr': 'न्यू कॅलेडोनिया', 'ms': 'New Caledonia', 'mt': 'New Caledonia', 'mua': 'Kaledoniya mafuu', 'my': 'နယူး ကယ်လီဒိုနီးယား', 'mzn': 'نیو کالیدونیا', 'naq': 'New Caledonia', 'nb': 'Ny-Caledonia', 'nd': 'New Caledonia', 'ne': 'न्यु क्यालेडोनिया', 'nl': 'Nieuw-Caledonië', 'nmg': 'Kaledoni nwanah', 'nn': 'Ny-Caledonia', 'no': 'Ny-Caledonia', 'nqo': 'ߞߊߟߋߘߏߣߌ߫-ߞߎߘߊ߫', 'nyn': 'Niukaredonia', 'om': 'Neewu Kaaleedoniyaa', 'or': 'ନୂତନ କାଲେଡୋନିଆ', 'pa': 'ਨਿਊ ਕੈਲੇਡੋਨੀਆ', 'pa-Guru': 'ਨਿਊ ਕੈਲੇਡੋਨੀਆ', 'pcm': 'Niú Kalẹdónia', 'pl': 'Nowa Kaledonia', 'ps': 'نوی کالیډونیا', 'pt': 'Nova Caledônia', 'qu': 'Nueva Caledonia', 'rm': 'Nova Caledonia', 'rn': 'Niyukaledoniya', 'ro': 'Noua Caledonie', 'rof': 'Nyukaledonia', 'ru': 'Новая Каледония', 'rwk': 'Nyukaledonia', 'saq': 'Nyukaledonia', 'sbp': 'Nyukaledoniya', 'sc': 'Caledònia Noa', 'sd': 'نیو ڪالیڊونیا', 'sd-Arab': 'نیو ڪالیڊونیا', 'se': 'Ođđa-Kaledonia', 'seh': 'Nova Caledônia', 'ses': 'Kaaledooni Taagaa', 'sg': 'Finî Kaledonïi', 'shi': 'ⴽⴰⵍⵉⴷⵓⵏⵢⴰ ⵜⴰⵎⴰⵢⵏⵓⵜ', 'shi-Latn': 'kalidunya tamaynut', 'shi-Tfng': 'ⴽⴰⵍⵉⴷⵓⵏⵢⴰ ⵜⴰⵎⴰⵢⵏⵓⵜ', 'si': 'නව කැලිඩෝනියාව', 'sk': 'Nová Kaledónia', 'sl': 'Nova Kaledonija', 'smn': 'Uđđâ-Kaledonia', 'sn': 'New Caledonia', 'so': 'Jasiiradda Niyuu Kaledooniya', 'sq': 'Kaledoni e Re', 'sr': 'Нова Каледонија', 'sr-Cyrl': 'Нова Каледонија', 'sr-Latn': 'Nova Kaledonija', 'sv': 'Nya Kaledonien', 'sw': 'New Caledonia', 'ta': 'நியூ கேலிடோனியா', 'te': 'క్రొత్త కాలెడోనియా', 'teo': 'Nyukaledonia', 'tg': 'Каледонияи Нав', 'th': 'นิวแคลิโดเนีย', 'ti': 'ኒው ካለዶንያ', 'tk': 'Täze Kaledoniýa', 'to': 'Niu Kaletōnia', 'tr': 'Yeni Kaledonya', 'tt': 'Яңа Каледония', 'twq': 'Kaaledooni Taagaa', 'tzm': 'kalidunya Tamaynut', 'ug': 'يېڭى كالېدونىيە', 'uk': 'Нова Каледонія', 'ur': 'نیو کلیڈونیا', 'uz': 'Yangi Kaledoniya', 'uz-Cyrl': 'Янги Каледония', 'uz-Latn': 'Yangi Kaledoniya', 'vai': 'ꕪꔷꖁꕇꕰ ꕯꕮꕊ', 'vai-Latn': 'Kalidoninya Námaá', 'vai-Vaii': 'ꕪꔷꖁꕇꕰ ꕯꕮꕊ', 'vec': 'Nova Caledonia', 'vi': 'New Caledonia', 'vun': 'Nyukaledonia', 'wae': 'Niwkaledonie', 'wo': 'Nuwel Kaledoni', 'xh': 'ENew Caledonia', 'xnr': 'न्यू कैलेडोनिया', 'xog': 'Kaledonya mupya', 'yav': 'nufɛ́l kaletoní', 'yi': 'נײַ קאַלעדאניע', 'yo': 'Kaledonia Titun', 'yrl': 'Karedũniya Pisasú', 'yue': '新喀里多尼亞', 'yue-Hans': '新喀里多尼亚', 'yue-Hant': '新喀里多尼亞', 'zgh': 'ⴽⴰⵍⵉⴷⵓⵏⵢⴰ ⵜⴰⵎⴰⵢⵏⵓⵜ', 'zh': '新喀里多尼亚', 'zh-Hans': '新喀里多尼亚', 'zh-Hant': '新喀里多尼亞', 'zu': 'i-New Caledonia'},
    'NE': {'af': 'Niger', 'agq': 'Naedzà', 'ak': 'Nigyɛɛ', 'am': 'ኒጀር', 'ar': 'النيجر', 'as': 'নাইজাৰ', 'asa': 'Nijeri', 'ast': 'El Níxer', 'az': 'Niger', 'az-Cyrl': 'Ниҝер', 'az-Latn': 'Niger', 'bal-Latn': 'Náiger', 'bas': 'Nìjɛ̂r', 'be': 'Нігер', 'be-tarask': 'Нігер', 'bez': 'Hunijeli', 'bg': 'Нигер', 'bgn': 'نیجیر', 'blo': 'Nijɛr', 'bm': 'Nizɛri', 'bn': 'নাইজার', 'br': 'Niger', 'brx': 'नाइजेर', 'bs': 'Niger', 'bs-Cyrl': 'Нигер', 'bs-Latn': 'Niger', 'ca': 'Níger', 'ca-ES-valencia': 'Níger', 'ccp': '𑄚𑄭𑄎𑄢𑄴', 'ce': 'Нигер', 'ceb': 'Niger', 'cgg': 'Naigya', 'chr': 'ᎾᎢᏨ', 'ckb': 'نیجەر', 'cs': 'Niger', 'cv': 'Нигер', 'cy': 'Niger', 'da': 'Niger', 'dav': 'Nijeri', 'de': 'Niger', 'dje': 'Nižer', 'dsb': 'Niger', 'dz': 'ནཱའི་ཇཱ', 'ebu': 'Nijeri', 'ee': 'Niger nutome', 'el': 'Νίγηρας', 'el-polyton': 'Νίγηρας', 'en': 'Niger', 'en-Dsrt': '𐐤𐐴𐐾𐐲𐑉', 'eo': 'Niĝero', 'es': 'Níger', 'et': 'Niger', 'eu': 'Niger', 'ewo': 'Nihɛ́r', 'fa': 'نیجر', 'ff': 'Nijeer', 'ff-Adlm': '𞤐𞤭𞥅𞤶𞤫𞤪', 'ff-Latn': 'Nijeer', 'fi': 'Niger', 'fil': 'Niger', 'fo': 'Niger', 'fr': 'Niger', 'fur': 'Niger', 'fy': 'Niger', 'ga': 'An Nígir', 'gd': 'Nìgeir', 'gl': 'Níxer', 'gsw': 'Niger', 'gu': 'નાઇજર', 'guz': 'Nijeri', 'ha': 'Nijar', 'he': 'ניז׳ר', 'hi': 'नाइजर', 'hi-Latn': 'Niger', 'hr': 'Niger', 'hsb': 'Niger', 'ht': 'Niger', 'hu': 'Niger', 'hy': 'Նիգեր', 'ia': 'Niger', 'id': 'Niger', 'ig': 'Niger', 'is': 'Níger', 'it': 'Niger', 'ja': 'ニジェール', 'jgo': 'Nijɛ̂', 'jmc': 'Nijeri', 'jv': 'Nigér', 'ka': 'ნიგერი', 'kaa': 'Нигер', 'kaa-Cyrl': 'Нигер', 'kab': 'Nijer', 'kam': 'Nijeri', 'kde': 'Nidyeli', 'kea': 'Nijer', 'kgp': 'Nĩjer', 'khq': 'Nižer', 'ki': 'Nijeri', 'kk': 'Нигер', 'kk-Cyrl': 'Нигер', 'kln': 'Emetab niger', 'km': 'នីហ្សេ', 'kn': 'ನೈಜರ್', 'ko': '니제르', 'kok': 'नायजर', 'kok-Deva': 'नायजर', 'ks': 'نایجَر', 'ks-Arab': 'نایجَر', 'ksb': 'Naija', 'ksf': 'nijɛ́r', 'ksh': 'Nijer', 'ku': 'Nîjer', 'kxv': 'naījr', 'kxv-Deva': 'नाइजर', 'kxv-Latn': 'naījr', 'kxv-Orya': 'ନାଇଜର', 'kxv-Telu': 'నఇజర్', 'ky': 'Нигер', 'lag': 'Níija', 'lb': 'Niger', 'lg': 'Nije', 'lld': 'Niger', 'ln': 'Nizɛrɛ', 'lo': 'ນິເຈີ', 'lt': 'Nigeris', 'lu': 'Nijere', 'luo': 'Niger', 'luy': 'Nijeri', 'lv': 'Nigēra', 'mai': 'नाइजर', 'mas': 'Nijeri', 'mer': 'Nija', 'mfe': 'Nizer', 'mg': 'Niger', 'mgh': 'Unijeri', 'mi': 'Ngāika', 'mk': 'Нигер', 'ml': 'നൈജർ', 'mn': 'Нигер', 'mr': 'नाइजर', 'ms': 'Niger', 'mt': 'in-Niġer', 'mua': 'Sǝr Niger', 'my': 'နိုင်ဂျာ', 'mzn': 'نیجر', 'naq': 'Nigeri', 'nb': 'Niger', 'nd': 'Niger', 'ne': 'नाइजर', 'nl': 'Niger', 'nmg': 'Niger', 'nn': 'Niger', 'no': 'Niger', 'nqo': 'ߖߋ߬ߟߌ߬ߓߊߘߎ߯', 'nyn': 'Naigya', 'om': 'Niijer', 'or': 'ନାଇଜର', 'pa': 'ਨਾਈਜਰ', 'pa-Guru': 'ਨਾਈਜਰ', 'pcm': 'Nizhẹr', 'pl': 'Niger', 'ps': 'نايجير', 'pt': 'Níger', 'qu': 'Níger', 'rm': 'Niger', 'rn': 'Nijeri', 'ro': 'Niger', 'rof': 'Nijeri', 'ru': 'Нигер', 'rwk': 'Nijeri', 'saq': 'Nijeri', 'sbp': 'Nijeli', 'sc': 'Niger', 'sd': 'نائيجر', 'sd-Arab': 'نائيجر', 'se': 'Niger', 'seh': 'Níger', 'ses': 'Nižer', 'sg': 'Nizëre', 'shi': 'ⵏⵏⵉⵊⵉⵔ', 'shi-Latn': 'nnijir', 'shi-Tfng': 'ⵏⵏⵉⵊⵉⵔ', 'si': 'නයිජර්', 'sk': 'Niger', 'sl': 'Niger', 'smn': 'Niger', 'sn': 'Niger', 'so': 'Nayjer', 'sq': 'Niger', 'sr': 'Нигер', 'sr-Cyrl': 'Нигер', 'sr-Latn': 'Niger', 'sv': 'Niger', 'sw': 'Niger', 'syr': 'ܢܝܓܪ', 'ta': 'நைஜர்', 'te': 'నైజర్', 'teo': 'Nijeri', 'tg': 'Нигер', 'th': 'ไนเจอร์', 'ti': 'ኒጀር', 'tk': 'Niger', 'to': 'Nisia', 'tr': 'Nijer', 'tt': 'Нигер', 'twq': 'Nižer', 'tzm': 'Nnijer', 'ug': 'نىگېر', 'uk': 'Нігер', 'ur': 'نائجر', 'uz': 'Niger', 'uz-Cyrl': 'Нигер', 'uz-Latn': 'Niger', 'vai': 'ꕯꔤꕧ', 'vai-Latn': 'Naĩja', 'vai-Vaii': 'ꕯꔤꕧ', 'vec': 'Niger', 'vi': 'Niger', 'vun': 'Nijeri', 'wae': 'Niger', 'wo': 'Niiseer', 'xh': 'ENiger', 'xnr': 'नाइजर', 'xog': 'Nije', 'yav': 'nisɛ́ɛl', 'yi': 'ניזשער', 'yo': 'Nàìjá', 'yrl': 'Nige', 'yue': '尼日', 'yue-Hans': '尼日', 'yue-Hant': '尼日', 'zgh': 'ⵏⵏⵉⵊⵉⵔ', 'zh': '尼日尔', 'zh-Hans': '尼日尔', 'zh-Hant': '尼日', 'zu': 'i-Niger'},
    'NF': {'af': 'Norfolkeiland', 'agq': 'Chwɨlà fɨ Nufòʔ', 'ak': 'Norfold Supɔ', 'am': 'ኖርፎልክ ደሴት', 'ar': 'جزيرة نورفولك', 'as': 'ন’ৰফ’ক দ্বীপ', 'asa': 'Kithiwa cha Norfok', 'ast': 'Islla Norfolk', 'az': 'Norfolk adası', 'az-Cyrl': 'Норфолк адасы', 'az-Latn': 'Norfolk adası', 'bal-Latn': 'Nórfolk Islánd', 'bas': 'Òn i Nɔrfɔ̂k', 'be': 'Востраў Норфалк', 'be-tarask': 'Востраў Норфалк', 'bez': 'Ihisima sha Norfok', 'bg': 'остров Норфолк', 'blo': 'Nɔrfook kaAtukǝltǝna', 'bm': 'Nɔrofoliki Gun', 'bn': 'নরফোক দ্বীপ', 'br': 'Enez Norfolk', 'brx': 'नर्फ’क द्वीप', 'bs': 'Ostrvo Norfolk', 'bs-Cyrl': 'Острво Норфолк', 'bs-Latn': 'Ostrvo Norfolk', 'ca': 'Illa Norfolk', 'ca-ES-valencia': 'Illa Norfolk', 'ccp': '𑄚𑄨𑄢𑄴𑄜𑄮𑄇𑄴 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Норфолк гӀайре', 'ceb': 'Norfolk Island', 'cgg': 'Ekizinga Norifoko', 'chr': 'ᏃᎵᏬᎵᎩ ᎤᎦᏚᏛᎢ', 'ckb': 'دوورگەی نۆرفۆڵک', 'cs': 'Norfolk', 'cv': 'Норфолк утравӗ', 'cy': 'Ynys Norfolk', 'da': 'Norfolk Island', 'dav': 'Kisiwa cha Norfok', 'de': 'Norfolkinsel', 'dje': 'Norfolk Gungoo', 'dsb': 'Norfolkowa kupa', 'dyo': 'Ecinkey yati Noorfok', 'dz': 'ནོར་ཕོལཀ་མཚོ་གླིང༌', 'ebu': 'Kisiwa cha Norfok', 'ee': 'Norfolk ƒudomekpo nutome', 'el': 'Νήσος Νόρφολκ', 'el-polyton': 'Νῆσος Νόρφολκ', 'en': 'Norfolk Island', 'en-Dsrt': '𐐤𐐱𐑉𐑁𐐲𐐿 𐐌𐑊𐐲𐑌𐐼', 'eo': 'Norfolkinsulo', 'es': 'Isla Norfolk', 'et': 'Norfolk', 'eu': 'Norfolk uhartea', 'ewo': 'Minlán Nɔrəfɔ́ləkə', 'fa': 'جزیرهٔ نورفولک', 'ff': 'Duuɗe Norfolk', 'ff-Adlm': '𞤅𞤵𞤪𞤭𞥅𞤪𞤫 𞤐𞤮𞤪𞤬𞤮𞤤𞤳𞤵', 'ff-Latn': 'Duuɗe Norfolk', 'fi': 'Norfolkinsaari', 'fil': 'Norfolk Island', 'fo': 'Norfolksoyggj', 'fr': 'Île Norfolk', 'fur': 'Isole Norfolk', 'fy': 'Norfolkeilân', 'ga': 'Oileán Norfolk', 'gd': 'Eilean Norfolk', 'gl': 'Illa Norfolk', 'gsw': 'Norfolk-Insle', 'gu': 'નોરફોક આઇલેન્ડ્સ', 'guz': 'Kisiwa cha Norfok', 'ha': 'Tsibirin Narfalk', 'he': 'האי נורפוק', 'hi': 'नॉरफ़ॉक द्वीप', 'hi-Latn': 'Norfolk Island', 'hr': 'Otok Norfolk', 'hsb': 'Norfolkowa kupa', 'ht': 'Île Norfolk', 'hu': 'Norfolk-sziget', 'hy': 'Նորֆոլկ կղզի', 'ia': 'Insula Norfolk', 'id': 'Kepulauan Norfolk', 'ie': 'Insul Norfolk', 'ig': 'Agwaetiti Norfolk', 'is': 'Norfolkeyja', 'it': 'Isola Norfolk', 'ja': 'ノーフォーク島', 'jmc': 'Kisiwa cha Norfok', 'jv': 'Pulo Norfolk', 'ka': 'ნორფოლკის კუნძული', 'kab': 'Tigzirin Tinawfukin', 'kam': 'Kisiwa cha Norfok', 'kde': 'Chisiwa cha Nolufok', 'kea': 'Ilhas Norfolk', 'kgp': 'Nãrforki Goj-vẽso', 'khq': 'Norfolk Gungoo', 'ki': 'Kisiwa cha Norfok', 'kk': 'Норфолк аралы', 'kk-Cyrl': 'Норфолк аралы', 'kln': 'Ikwembeiyotab Norfork', 'km': 'កោះ\u200bណ័រហ្វក់', 'kn': 'ನಾರ್ಫೋಕ್ ದ್ವೀಪ', 'ko': '노퍽섬', 'kok': 'नॉरफॉक आयलँड', 'kok-Deva': 'नॉरफॉक आयलँड', 'ks': 'نارفاک جٔزیٖرٕ', 'ks-Arab': 'نارفاک جٔزیٖرٕ', 'ksb': 'Kisiwa cha Nolfok', 'ksf': 'zɛ nɔ́fɔlk', 'ksh': 'de Noofok-Ensel', 'ku': 'Girava Norfolkê', 'kxv': 'norpok dīp', 'kxv-Deva': 'नॉरप़ॉक द्वीप', 'kxv-Latn': 'norpok dīp', 'kxv-Orya': 'ନର୍\u200dପକ୍ ଦିପ', 'kxv-Telu': 'నార్పోక్ దీవ', 'ky': 'Норфолк аралы', 'lag': 'Kisíiwa cha Nofifóoki', 'lb': 'Norfolkinsel', 'lg': 'Kizinga ky’eNorofoko', 'lld': 'Isola Norfolk', 'ln': 'Esanga Norfokɛ', 'lo': 'ເກາະນໍໂຟກ', 'lt': 'Norfolko sala', 'lu': 'Lutanda lua Norfok', 'luo': 'Chuia mar Norfolk', 'luy': 'Kisiwa cha Norfok', 'lv': 'Norfolkas sala', 'mai': 'नौरफोक द्वीप', 'mas': 'Kisiwa cha Norfok', 'mer': 'Aĩrandi cia Norfok', 'mfe': 'Lil Norfolk', 'mg': 'Nosy Norfolk', 'mi': 'Te Moutere Nōpoke', 'mk': 'Норфолшки Остров', 'ml': 'നോർഫോക് ദ്വീപ്', 'mn': 'Норфолк арал', 'mr': 'नॉरफॉक बेट', 'ms': 'Pulau Norfolk', 'mt': 'Gżira Norfolk', 'mua': 'Norfolk ma laŋne', 'my': 'နောဖုတ်ကျွန်း', 'mzn': 'نورفولک جزیره', 'naq': 'Norfolk Island', 'nb': 'Norfolkøya', 'nd': 'Norfolk Island', 'ne': 'नोरफोल्क टापु', 'nl': 'Norfolk', 'nmg': 'Maŋ́ má Nɔrfɔrk', 'nn': 'Norfolkøya', 'no': 'Norfolkøya', 'nqo': 'ߣߐߙߑߝߐߟߑߞ', 'nyn': 'Ekizinga Norifoko', 'om': 'Odola Noorfoolk', 'or': 'ନର୍ଫକ୍ ଦ୍ଵୀପ', 'pa': 'ਨੋਰਫੌਕ ਟਾਪੂ', 'pa-Guru': 'ਨੋਰਫੌਕ ਟਾਪੂ', 'pcm': 'Nọ́fọlk Aíland', 'pl': 'Norfolk', 'ps': 'نارفولک ټاپوګان', 'pt': 'Ilha Norfolk', 'qu': 'Isla Norfolk', 'rm': 'Insla Norfolk', 'rn': 'izinga rya Norufoluke', 'ro': 'Insula Norfolk', 'rof': 'Kisiwa cha Norfok', 'ru': 'о-в Норфолк', 'rwk': 'Kisiwa cha Norfok', 'saq': 'Kisiwa cha Norfok', 'sbp': 'Shigunguli sha Nolifoki', 'sc': 'Ìsula Norfolk', 'sd': 'نورفوڪ ٻيٽ', 'sd-Arab': 'نورفوڪ ٻيٽ', 'se': 'Norfolksullot', 'seh': 'Ilhas Norfolk', 'ses': 'Norfolk Gungoo', 'sg': 'Zûâ Nôrfôlko', 'shi': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵏⵓⵔⴼⵓⵍⴽ', 'shi-Latn': 'tigzirin n nurfulk', 'shi-Tfng': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵏⵓⵔⴼⵓⵍⴽ', 'si': 'නෝෆෝක් දූපත', 'sk': 'Norfolk', 'sl': 'Norfolški otok', 'smn': 'Norfolksuálui', 'sn': 'Chitsuwa cheNorfolk', 'so': 'Jasiiradda Noorfolk', 'sq': 'Ishulli Norfolk', 'sr': 'Острво Норфок', 'sr-Cyrl': 'Острво Норфок', 'sr-Latn': 'Ostrvo Norfok', 'sv': 'Norfolkön', 'sw': 'Kisiwa cha Norfolk', 'ta': 'நார்ஃபோக் தீவு', 'te': 'నార్ఫోక్ దీవి', 'teo': 'Kisiwa cha Norfok', 'tg': 'Ҷазираи Норфолк', 'th': 'เกาะนอร์ฟอล์ก', 'ti': 'ደሴት ኖርፎልክ', 'tk': 'Norfolk adasy', 'to': 'Motu Nōfoliki', 'tr': 'Norfolk Adası', 'tt': 'Норфолк утравы', 'twq': 'Norfolk Gungoo', 'tzm': 'Tigzirt Nurfulk', 'ug': 'نورفولك ئارىلى', 'uk': 'Острів Норфолк', 'ur': 'نارفوک آئلینڈ', 'uz': 'Norfolk oroli', 'uz-Cyrl': 'Норфолк ороллари', 'uz-Latn': 'Norfolk oroli', 'vai': 'ꗟꖺꗉ ꔳꘋꗣ', 'vai-Latn': 'Nɔfɔ Tiŋŋɛ', 'vai-Vaii': 'ꗟꖺꗉ ꔳꘋꗣ', 'vec': 'Ìzola Norfolk', 'vi': 'Đảo Norfolk', 'vun': 'Kisiwa cha Norfok', 'wae': 'Norfolkinsla', 'wo': 'Dunu Norfolk', 'xh': 'ENorfolk Island', 'xnr': 'नॉरफ़ॉक द्वीप', 'xog': 'Kizinga ky’eNorofoko', 'yav': 'il nɔ́lfɔ́lɔk', 'yi': 'נארפֿאלק אינזל', 'yo': 'Erékùsù Nọ́úfókì', 'yrl': 'Kapuãma Norfolk', 'yue': '諾福克島', 'yue-Hans': '诺福克岛', 'yue-Hant': '諾福克島', 'zgh': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵏⵓⵔⴼⵓⵍⴽ', 'zh': '诺福克岛', 'zh-Hans': '诺福克岛', 'zh-Hant': '諾福克島', 'zu': 'i-Norfolk Island'},
    'NG': {'af': 'Nigerië', 'agq': 'Gɨ̀anyɨ', 'ak': 'Naegyeria', 'am': 'ናይጄሪያ', 'ar': 'نيجيريا', 'as': 'নাইজেৰিয়া', 'asa': 'Nijeria', 'ast': 'Nixeria', 'az': 'Nigeriya', 'az-Cyrl': 'Ниҝерија', 'az-Latn': 'Nigeriya', 'bal-Latn': 'Náijériá', 'bas': 'Nìgerìà', 'be': 'Нігерыя', 'be-tarask': 'Нігерыя', 'bez': 'Hunijelia', 'bg': 'Нигерия', 'bgn': 'نایجیریا', 'blo': 'Nanjiiriya', 'bm': 'Nizeriya', 'bn': 'নাইজেরিয়া', 'br': 'Nigeria', 'brx': 'नाइजेरिया', 'bs': 'Nigerija', 'bs-Cyrl': 'Нигерија', 'bs-Latn': 'Nigerija', 'ca': 'Nigèria', 'ca-ES-valencia': 'Nigèria', 'ccp': '𑄚𑄭𑄎𑄬𑄢𑄨𑄠', 'ce': 'Нигери', 'ceb': 'Nigeria', 'cgg': 'Naigyeriya', 'chr': 'ᏂᏥᎵᏯ', 'ckb': 'نیجریا', 'cs': 'Nigérie', 'cv': 'Нигери', 'cy': 'Nigeria', 'da': 'Nigeria', 'dav': 'Nijeria', 'de': 'Nigeria', 'dje': 'Naajiriia', 'dsb': 'Nigerija', 'dz': 'ནཱའི་ཇི་རི་ཡ', 'ebu': 'Nijeria', 'ee': 'Nigeria nutome', 'el': 'Νιγηρία', 'el-polyton': 'Νιγηρία', 'en': 'Nigeria', 'en-Dsrt': '𐐤𐐴𐐾𐐮𐑉𐐨𐐲', 'eo': 'Niĝerio', 'es': 'Nigeria', 'et': 'Nigeeria', 'eu': 'Nigeria', 'ewo': 'Nihéria', 'fa': 'نیجریه', 'ff': 'Nijeriyaa', 'ff-Adlm': '𞤐𞤢𞤶𞤫𞤪𞤭𞤴𞤢𞥄', 'ff-Latn': 'Nijeriyaa', 'fi': 'Nigeria', 'fil': 'Nigeria', 'fo': 'Nigeria', 'fr': 'Nigeria', 'fur': 'Nigerie', 'fy': 'Nigeria', 'ga': 'An Nigéir', 'gd': 'Nigèiria', 'gl': 'Nixeria', 'gsw': 'Nigeeria', 'gu': 'નાઇજેરિયા', 'guz': 'Nijeria', 'ha': 'Nijeriya', 'he': 'ניגריה', 'hi': 'नाइजीरिया', 'hi-Latn': 'Nigeria', 'hr': 'Nigerija', 'hsb': 'Nigerija', 'ht': 'Nigeria', 'hu': 'Nigéria', 'hy': 'Նիգերիա', 'ia': 'Nigeria', 'id': 'Nigeria', 'ig': 'Naịjịrịa', 'is': 'Nígería', 'it': 'Nigeria', 'ja': 'ナイジェリア', 'jgo': 'Ninjɛliyâ', 'jmc': 'Nijeria', 'jv': 'Nigéria', 'ka': 'ნიგერია', 'kaa': 'Нигерия', 'kaa-Cyrl': 'Нигерия', 'kab': 'Nijirya', 'kam': 'Nijeria', 'kde': 'Nidyelia', 'kea': 'Nijéria', 'kgp': 'Nĩjérija', 'khq': 'Naajiriia', 'ki': 'Nainjeria', 'kk': 'Нигерия', 'kk-Cyrl': 'Нигерия', 'kln': 'Emetab Nigeria', 'km': 'នីហ្សេរីយ៉ា', 'kn': 'ನೈಜೀರಿಯಾ', 'ko': '나이지리아', 'kok': 'नायजेरिया', 'kok-Deva': 'नायजेरिया', 'ks': 'نایجیرِیا', 'ks-Arab': 'نایجیرِیا', 'ksb': 'Naijelia', 'ksf': 'nijɛ́rya', 'ksh': 'Nikaraagua', 'ku': 'Nîjerya', 'kxv': 'naigeria', 'kxv-Deva': 'नाइजीरिया', 'kxv-Latn': 'naigeria', 'kxv-Orya': 'ନାଇଜେରିଆ', 'kxv-Telu': 'నౌజీరియా', 'ky': 'Нигерия', 'lag': 'Niijéria', 'lb': 'Nigeria', 'lg': 'Nayijerya', 'lld': 'Nigeria', 'ln': 'Nizerya', 'lo': 'ໄນຈີເຣຍ', 'lt': 'Nigerija', 'lu': 'Nijerya', 'luo': 'Nigeria', 'luy': 'Nijeria', 'lv': 'Nigērija', 'mai': 'नाइजीरिया', 'mas': 'Nijeria', 'mer': 'Nijeria', 'mfe': 'Nizeria', 'mg': 'Nizeria', 'mgh': 'Unijeria', 'mi': 'Ngāitiria', 'mk': 'Нигерија', 'ml': 'നൈജീരിയ', 'mn': 'Нигери', 'mr': 'नायजेरिया', 'ms': 'Nigeria', 'mt': 'in-Niġerja', 'mua': 'Nigeriya', 'my': 'နိုင်ဂျီးရီးယား', 'mzn': 'نیجریه', 'naq': 'Nigerieb', 'nb': 'Nigeria', 'nd': 'Nigeriya', 'ne': 'नाइजेरिया', 'nl': 'Nigeria', 'nmg': 'Nigeria', 'nn': 'Nigeria', 'no': 'Nigeria', 'nqo': 'ߖߋ߬ߟߌ߬ߓߊ߬ߟߊ߫', 'nyn': 'Naigyeriya', 'om': 'Naayijeeriyaa', 'or': 'ନାଇଜେରିଆ', 'pa': 'ਨਾਈਜੀਰੀਆ', 'pa-Guru': 'ਨਾਈਜੀਰੀਆ', 'pcm': 'Naijíria', 'pl': 'Nigeria', 'ps': 'نایجیریا', 'pt': 'Nigéria', 'qu': 'Nigeria', 'rm': 'Nigeria', 'rn': 'Nijeriya', 'ro': 'Nigeria', 'rof': 'Nijeria', 'ru': 'Нигерия', 'rwk': 'Nijeria', 'saq': 'Nijeria', 'sbp': 'Nijeliya', 'sc': 'Nigèria', 'sd': 'نائيجيريا', 'sd-Arab': 'نائيجيريا', 'se': 'Nigeria', 'seh': 'Nigéria', 'ses': 'Naajiriia', 'sg': 'Nizerïa', 'shi': 'ⵏⵉⵊⵉⵔⵢⴰ', 'shi-Latn': 'nijirya', 'shi-Tfng': 'ⵏⵉⵊⵉⵔⵢⴰ', 'si': 'නයිජීරියාව', 'sk': 'Nigéria', 'sl': 'Nigerija', 'smn': 'Nigeria', 'sn': 'Nigeria', 'so': 'Nayjeeriya', 'sq': 'Nigeri', 'sr': 'Нигерија', 'sr-Cyrl': 'Нигерија', 'sr-Latn': 'Nigerija', 'sv': 'Nigeria', 'sw': 'Nigeria', 'syr': 'ܢܝܓܝܪܝܐ', 'ta': 'நைஜீரியா', 'te': 'నైజీరియా', 'teo': 'Nijeria', 'tg': 'Нигерия', 'th': 'ไนจีเรีย', 'ti': 'ናይጀርያ', 'tk': 'Nigeriýa', 'to': 'Naisilia', 'tr': 'Nijerya', 'tt': 'Нигерия', 'twq': 'Naajiriia', 'tzm': 'Nijiria', 'ug': 'نىگېرىيە', 'uk': 'Нігерія', 'ur': 'نائجیریا', 'uz': 'Nigeriya', 'uz-Cyrl': 'Нигерия', 'uz-Latn': 'Nigeriya', 'vai': 'ꕯꔤꕀꔸꕩ', 'vai-Latn': 'Naĩjiriya', 'vai-Vaii': 'ꕯꔤꕀꔸꕩ', 'vec': 'Nigeria', 'vi': 'Nigeria', 'vun': 'Nijeria', 'wae': 'Nigeria', 'wo': 'Niseriya', 'xh': 'ENigeria', 'xnr': 'नाइजीरिया', 'xog': 'Nayijerya', 'yav': 'nisélia', 'yi': 'ניגעריע', 'yo': 'Nàìjíríà', 'yrl': 'Nigeriya', 'yue': '奈及利亞', 'yue-Hans': '奈及利亚', 'yue-Hant': '奈及利亞', 'zgh': 'ⵏⵉⵊⵉⵔⵢⴰ', 'zh': '尼日利亚', 'zh-Hans': '尼日利亚', 'zh-Hant': '奈及利亞', 'zu': 'i-Nigeria'},
    'NI': {'af': 'Nicaragua', 'agq': 'Nikàlagwà', 'ak': 'Nekaraguwa', 'am': 'ኒካራጓ', 'ar': 'نيكاراغوا', 'as': 'নিকাৰাগুৱা', 'asa': 'Nikaragwa', 'ast': 'Nicaragua', 'az': 'Nikaraqua', 'az-Cyrl': 'Никарагуа', 'az-Latn': 'Nikaraqua', 'bal-Latn': 'Nekárágóá', 'bas': 'Nìkàragwà', 'be': 'Нікарагуа', 'be-tarask': 'Нікарагуа', 'bez': 'Hunikaragwa', 'bg': 'Никарагуа', 'blo': 'Nikaraguwaa', 'bm': 'Nikaragwa', 'bn': 'নিকারাগুয়া', 'br': 'Nicaragua', 'brx': 'निकारागुआ', 'bs': 'Nikaragva', 'bs-Cyrl': 'Никарагва', 'bs-Latn': 'Nikaragva', 'ca': 'Nicaragua', 'ca-ES-valencia': 'Nicaragua', 'ccp': '𑄚𑄨𑄇𑄢𑄉𑄪𑄠', 'ce': 'Никарагуа', 'ceb': 'Nicaragua', 'cgg': 'Nikaragwa', 'chr': 'ᏂᎧᎳᏆ', 'ckb': 'نیکاراگوا', 'cs': 'Nikaragua', 'cv': 'Никарагуа', 'cy': 'Nicaragua', 'da': 'Nicaragua', 'dav': 'Nikaragwa', 'de': 'Nicaragua', 'dje': 'Nikaragwa', 'dsb': 'Nikaragua', 'dz': 'ནི་ཀྲ་ཝ་ག', 'ebu': 'Nikaragwa', 'ee': 'Nicaraguadukɔ', 'el': 'Νικαράγουα', 'el-polyton': 'Νικαράγουα', 'en': 'Nicaragua', 'en-Dsrt': '𐐤𐐮𐐿𐐲𐑉𐐪𐑀𐐶𐐲', 'eo': 'Nikaragvo', 'es': 'Nicaragua', 'et': 'Nicaragua', 'eu': 'Nikaragua', 'ewo': 'Nikarágua', 'fa': 'نیکاراگوئه', 'ff': 'Nikaraguwaa', 'ff-Adlm': '𞤐𞤭𞤳𞤢𞤪𞤢𞤺𞤵𞤱𞤢𞥄', 'ff-Latn': 'Nikaraguwaa', 'fi': 'Nicaragua', 'fil': 'Nicaragua', 'fo': 'Nikaragua', 'fr': 'Nicaragua', 'fur': 'Nicaragua', 'fy': 'Nicaragua', 'ga': 'Nicearagua', 'gd': 'Niocaragua', 'gl': 'Nicaragua', 'gsw': 'Nicaraagua', 'gu': 'નિકારાગુઆ', 'guz': 'Nikaragwa', 'ha': 'Nikaraguwa', 'he': 'ניקרגואה', 'hi': 'निकारागुआ', 'hi-Latn': 'Nicaragua', 'hr': 'Nikaragva', 'hsb': 'Nikaragua', 'ht': 'Nicaragua', 'hu': 'Nicaragua', 'hy': 'Նիկարագուա', 'ia': 'Nicaragua', 'id': 'Nikaragua', 'ig': 'Nicaragua', 'is': 'Níkaragva', 'it': 'Nicaragua', 'ja': 'ニカラグア', 'jmc': 'Nikaragwa', 'jv': 'Nikaragua', 'ka': 'ნიკარაგუა', 'kaa': 'Никарагуа', 'kaa-Cyrl': 'Никарагуа', 'kab': 'Nikaragwa', 'kam': 'Nikaragwa', 'kde': 'Nikalagwa', 'kea': 'Nikarágua', 'kgp': 'Nĩkaragva', 'khq': 'Nikaragwa', 'ki': 'Nikaragwa', 'kk': 'Никарагуа', 'kk-Cyrl': 'Никарагуа', 'kln': 'Emetab Nicaragua', 'km': 'នីការ៉ាហ្គា', 'kn': 'ನಿಕಾರಾಗುವಾ', 'ko': '니카라과', 'kok': 'निकारगुवा', 'kok-Deva': 'निकारगुवा', 'ks': 'ناکاراگُوا', 'ks-Arab': 'ناکاراگُوا', 'ksb': 'Nikalagwa', 'ksf': 'níkarágwa', 'ksh': 'Nikarahguwa', 'ku': 'Nîkaragua', 'kxv': 'nikaraguaa', 'kxv-Deva': 'निकारागुआ', 'kxv-Latn': 'nikaraguaa', 'kxv-Orya': 'ନିକାରଗୁଆ', 'kxv-Telu': 'నికరాగువా', 'ky': 'Никарагуа', 'lag': 'Nikarágʉa', 'lb': 'Nicaragua', 'lg': 'Nikaraguwa', 'lld': 'Nicaragua', 'ln': 'Nikaragwa', 'lo': 'ນິກຄາຣາກົວ', 'lt': 'Nikaragva', 'lu': 'Nikaragwa', 'luo': 'Nicaragua', 'luy': 'Nikaragwa', 'lv': 'Nikaragva', 'mai': 'निकारागुआ', 'mas': 'Nikaragwa', 'mer': 'Nikaragwa', 'mfe': 'Nicaragua', 'mg': 'Nikaragoà', 'mi': 'Nikarāhua', 'mk': 'Никарагва', 'ml': 'നിക്കരാഗ്വ', 'mn': 'Никарагуа', 'mr': 'निकाराग्वा', 'ms': 'Nicaragua', 'mt': 'in-Nikaragwa', 'mua': 'Nikaragwa', 'my': 'နီကာရာဂွါ', 'mzn': 'نیکاراگوئه', 'naq': 'Nicaraguab', 'nb': 'Nicaragua', 'nd': 'Nicaragua', 'ne': 'निकारागुवा', 'nl': 'Nicaragua', 'nmg': 'Nikaragua', 'nn': 'Nicaragua', 'no': 'Nicaragua', 'nqo': 'ߣߌߞߙߊߜ߭ߎߥߊ߫', 'nyn': 'Nikaragwa', 'om': 'Nikaraguwaa', 'or': 'ନିକାରାଗୁଆ', 'pa': 'ਨਿਕਾਰਾਗੁਆ', 'pa-Guru': 'ਨਿਕਾਰਾਗੁਆ', 'pcm': 'Nikarágwua', 'pl': 'Nikaragua', 'ps': 'نکاراګوا', 'pt': 'Nicarágua', 'qu': 'Nicaragua', 'rm': 'Nicaragua', 'rn': 'Nikaragwa', 'ro': 'Nicaragua', 'rof': 'Nikaragwa', 'ru': 'Никарагуа', 'rwk': 'Nikaragwa', 'saq': 'Nikaragwa', 'sbp': 'Nikalagwa', 'sc': 'Nicaràgua', 'sd': 'نڪراگوا', 'sd-Arab': 'نڪراگوا', 'se': 'Nicaragua', 'seh': 'Nicarágua', 'ses': 'Nikaragwa', 'sg': 'Nikaragua', 'shi': 'ⵏⵉⴽⴰⵔⴰⴳⵡⴰ', 'shi-Latn': 'nikaragwa', 'shi-Tfng': 'ⵏⵉⴽⴰⵔⴰⴳⵡⴰ', 'si': 'නිකරගුවාව', 'sk': 'Nikaragua', 'sl': 'Nikaragva', 'smn': 'Nicaragua', 'sn': 'Nicaragua', 'so': 'Nikaraaguwa', 'sq': 'Nikaragua', 'sr': 'Никарагва', 'sr-Cyrl': 'Никарагва', 'sr-Latn': 'Nikaragva', 'sv': 'Nicaragua', 'sw': 'Nikaragwa', 'syr': 'ܢܝܟܪܐܓܘܐ', 'ta': 'நிகரகுவா', 'te': 'నికరాగువా', 'teo': 'Nikaragwa', 'tg': 'Никарагуа', 'th': 'นิการากัว', 'ti': 'ኒካራጓ', 'tk': 'Nikaragua', 'to': 'Nikalakua', 'tr': 'Nikaragua', 'tt': 'Никарагуа', 'twq': 'Nikaragwa', 'tzm': 'Nikaragwa', 'ug': 'نىكاراگۇئا', 'uk': 'Нікарагуа', 'ur': 'نکاراگووا', 'uz': 'Nikaragua', 'uz-Cyrl': 'Никарагуа', 'uz-Latn': 'Nikaragua', 'vai': 'ꕇꕪꕟꖶꕎ', 'vai-Latn': 'Nikaraguwa', 'vai-Vaii': 'ꕇꕪꕟꖶꕎ', 'vec': 'Nicaragua', 'vi': 'Nicaragua', 'vun': 'Nikaragwa', 'wae': 'Nicaragua', 'wo': 'Nikaraguwa', 'xh': 'ENicaragua', 'xnr': 'निकारागुआ', 'xog': 'Nikaraguwa', 'yav': 'nikalaká', 'yi': 'ניקאַראַגוע', 'yo': 'Nikaragua', 'yrl': 'Nicaraguwa', 'yue': '尼加拉瓜', 'yue-Hans': '尼加拉瓜', 'yue-Hant': '尼加拉瓜', 'zgh': 'ⵏⵉⴽⴰⵔⴰⴳⵡⴰ', 'zh': '尼加拉瓜', 'zh-Hans': '尼加拉瓜', 'zh-Hant': '尼加拉瓜', 'zu': 'i-Nicaragua'},
    'NL': {'af': 'Nederland', 'agq': 'Nedàlân', 'ak': 'Nɛdɛland', 'am': 'ኔዘርላንድ', 'ar': 'هولندا', 'as': 'নেডাৰলেণ্ড', 'asa': 'Uholandhi', 'ast': 'Países Baxos', 'az': 'Niderland', 'az-Cyrl': 'Нидерланд', 'az-Latn': 'Niderland', 'bal-Latn': 'Nedarlánd', 'bas': 'Ǹlɛndi', 'be': 'Нідэрланды', 'be-tarask': 'Нідэрланды', 'bez': 'Huuholanzi', 'bg': 'Нидерландия', 'blo': 'Holanɖ', 'bm': 'Peyiba', 'bn': 'নেদারল্যান্ডস', 'br': 'Izelvroioù', 'brx': 'नेदारलेण्ड', 'bs': 'Nizozemska', 'bs-Cyrl': 'Холандија', 'bs-Latn': 'Nizozemska', 'ca': 'Països Baixos', 'ca-ES-valencia': 'Països Baixos', 'ccp': '𑄚𑄬𑄘𑄢𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴𑄥𑄴', 'ce': 'Нидерландаш', 'ceb': 'Netherlands', 'cgg': 'Hoorandi', 'chr': 'ᏁᏛᎳᏂ', 'ckb': 'ھۆڵەندا', 'cs': 'Nizozemsko', 'cv': 'Нидерланд', 'cy': 'Yr Iseldiroedd', 'da': 'Nederlandene', 'dav': 'Uholanzi', 'de': 'Niederlande', 'dje': 'Hollandu', 'dsb': 'Nižozemska', 'dz': 'ནེ་དར་ལནཌས྄', 'ebu': 'Uholanzi', 'ee': 'Netherlands nutome', 'el': 'Κάτω Χώρες', 'el-polyton': 'Ὁλλανδία', 'en': 'Netherlands', 'en-Dsrt': '𐐤𐐯𐑄𐐲𐑉𐑊𐐲𐑌𐐼𐑆', 'eo': 'Nederlando', 'es': 'Países Bajos', 'et': 'Holland', 'eu': 'Herbehereak', 'ewo': 'Pɛíbá', 'fa': 'هلند', 'ff': 'Nederlannda', 'ff-Adlm': '𞤖𞤮𞤤𞤢𞤲𞤣𞤭𞤴𞤢𞥄', 'ff-Latn': 'Nederlannda', 'fi': 'Alankomaat', 'fil': 'Netherlands', 'fo': 'Niðurlond', 'fr': 'Pays-Bas', 'fur': 'Paîs bas', 'fy': 'Nederlân', 'ga': 'an Ísiltír', 'gd': 'Na Tìrean Ìsle', 'gl': 'Países Baixos', 'gsw': 'Holland', 'gu': 'નેધરલેન્ડ્સ', 'guz': 'Uholanzi', 'ha': 'Holan', 'haw': 'Hōlani', 'he': 'הולנד', 'hi': 'नीदरलैंड', 'hi-Latn': 'Netherlands', 'hr': 'Nizozemska', 'hsb': 'Nižozemska', 'ht': 'Pays-Bas', 'hu': 'Hollandia', 'hy': 'Նիդեռլանդներ', 'ia': 'Nederlandia', 'id': 'Belanda', 'ig': 'Netherlands', 'is': 'Holland', 'it': 'Paesi Bassi', 'ja': 'オランダ', 'jmc': 'Uholanzi', 'jv': 'Walanda', 'ka': 'ნიდერლანდები', 'kaa': 'Нидерландия', 'kaa-Cyrl': 'Нидерландия', 'kab': 'Timura-Yessakesren', 'kam': 'Uholanzi', 'kde': 'Uholanzi', 'kea': 'Olanda', 'kgp': 'Pajisi Rur', 'khq': 'Hollandu', 'ki': 'Uholanzi', 'kk': 'Нидерланд', 'kk-Cyrl': 'Нидерланд', 'kln': 'Emetab Holand', 'km': 'ហូឡង់', 'kn': 'ನೆದರ್\u200cಲ್ಯಾಂಡ್ಸ್', 'ko': '네덜란드', 'kok': 'नॅदरलँड', 'kok-Deva': 'नॅदरलँड', 'ks': 'نیٖدَرلینڑ', 'ks-Arab': 'نیٖدَرلینڑ', 'ksb': 'Uholanzi', 'ksf': 'kǝlɔŋ kǝ ázǝ', 'ksh': 'de Nederläng', 'ku': 'Holanda', 'kxv': 'nederlanḍs', 'kxv-Deva': 'नीदरलेंड', 'kxv-Latn': 'nederlanḍs', 'kxv-Orya': 'ନେଦରଲାଣ୍ଡ', 'kxv-Telu': 'నెదర్లాండ్స్', 'ky': 'Нидерланд', 'lag': 'Ʉholáanzi', 'lb': 'Holland', 'lg': 'Holandi', 'lld': 'Paisc Basc', 'ln': 'Olandɛ', 'lo': 'ເນເທີແລນ', 'lt': 'Nyderlandai', 'lu': 'Olandɛ', 'luo': 'Netherlands', 'luy': 'Uholanzi', 'lv': 'Nīderlande', 'mai': 'नीदरलैंड', 'mas': 'Uholanzi', 'mer': 'Holandi', 'mfe': 'Oland', 'mg': 'Holanda', 'mi': 'Hōrana', 'mk': 'Холандија', 'ml': 'നെതർലാൻഡ്\u200cസ്', 'mn': 'Нидерланд', 'mr': 'नेदरलँड', 'ms': 'Belanda', 'mt': 'in-Netherlands', 'mua': 'Sǝr ma kasǝŋ', 'my': 'နယ်သာလန်', 'mzn': 'هلند', 'naq': 'Netherlands', 'nb': 'Nederland', 'nd': 'Netherlands', 'nds': 'Nedderlannen', 'ne': 'नेदरल्याण्ड', 'nl': 'Nederland', 'nmg': 'Nedɛrland', 'nn': 'Nederland', 'no': 'Nederland', 'nqo': 'ߤߏߟߊ߲ߘ', 'nyn': 'Hoorandi', 'om': 'Neezerlaand', 'or': 'ନେଦରଲ୍ୟାଣ୍ଡ', 'pa': 'ਨੀਦਰਲੈਂਡ', 'pa-Guru': 'ਨੀਦਰਲੈਂਡ', 'pcm': 'Nẹ́dalands', 'pl': 'Holandia', 'ps': 'هالېنډ', 'pt': 'Países Baixos', 'qu': 'Países Bajos', 'rm': 'Pajais Bass', 'rn': 'Ubuholandi', 'ro': 'Țările de Jos', 'rof': 'Uholanzi', 'ru': 'Нидерланды', 'rwk': 'Uholanzi', 'saq': 'Uholanzi', 'sbp': 'Wuholansi', 'sc': 'Paisos Bassos', 'sd': 'نيدرلينڊ', 'sd-Arab': 'نيدرلينڊ', 'se': 'Vuolleeatnamat', 'seh': 'Holanda', 'ses': 'Hollandu', 'sg': 'Holände', 'shi': 'ⵀⵓⵍⴰⵏⴷⴰ', 'shi-Latn': 'hulanda', 'shi-Tfng': 'ⵀⵓⵍⴰⵏⴷⴰ', 'si': 'නෙදර්ලන්තය', 'sk': 'Holandsko', 'sl': 'Nizozemska', 'smn': 'Vuáládâhenâmeh', 'sn': 'Netherlands', 'so': 'Nederlaands', 'sq': 'Holandë', 'sr': 'Холандија', 'sr-Cyrl': 'Холандија', 'sr-Latn': 'Holandija', 'sv': 'Nederländerna', 'sw': 'Uholanzi', 'syr': 'ܗܘܠܢܕܐ', 'ta': 'நெதர்லாந்து', 'te': 'నెదర్లాండ్స్', 'teo': 'Uholanzi', 'tg': 'Нидерландия', 'th': 'เนเธอร์แลนด์', 'ti': 'ኔዘርላንድ', 'tk': 'Niderlandlar', 'to': 'Hōlani', 'tr': 'Hollanda', 'tt': 'Нидерланд', 'twq': 'Hollandu', 'tzm': 'Hulanḍa', 'ug': 'گوللاندىيە', 'uk': 'Нідерланди', 'ur': 'نیدر لینڈز', 'uz': 'Niderlandiya', 'uz-Cyrl': 'Нидерландия', 'uz-Latn': 'Niderlandiya', 'vai': 'ꘉꕜ ꖨꕮꕊ', 'vai-Latn': 'Nidɔlɛŋ', 'vai-Vaii': 'ꘉꕜ ꖨꕮꕊ', 'vec': 'Paezi Basi', 'vi': 'Hà Lan', 'vun': 'Uholanzi', 'wae': 'Holand', 'wo': 'Peyi Baa', 'xh': 'ENetherlands', 'xnr': 'नीदरलैंड', 'xog': 'Holandi', 'yav': 'nitililáand', 'yi': 'האלאַנד', 'yo': 'Nedalandi', 'yrl': 'Tetãma Iwiboí-ita', 'yue': '荷蘭', 'yue-Hans': '荷兰', 'yue-Hant': '荷蘭', 'zgh': 'ⵀⵓⵍⴰⵏⴷⴰ', 'zh': '荷兰', 'zh-Hans': '荷兰', 'zh-Hant': '荷蘭', 'zu': 'i-Netherlands'},
    'NO': {'af': 'Noorweë', 'agq': 'Noowɛ̂ɛ', 'ak': 'Nɔɔwe', 'am': 'ኖርዌይ', 'ar': 'النرويج', 'as': 'নৰৱে', 'asa': 'Norwe', 'ast': 'Noruega', 'az': 'Norveç', 'az-Cyrl': 'Норвеч', 'az-Latn': 'Norveç', 'bal-Latn': 'Nárway', 'bas': 'Nɔ̀rvegìà', 'be': 'Нарвегія', 'be-tarask': 'Нарвегія', 'bez': 'Hunolwe', 'bg': 'Норвегия', 'blo': 'Nɔrfɛsh', 'bm': 'Nɔriwɛzi', 'bn': 'নরওয়ে', 'br': 'Norvegia', 'brx': 'नरवै', 'bs': 'Norveška', 'bs-Cyrl': 'Норвешка', 'bs-Latn': 'Norveška', 'ca': 'Noruega', 'ca-ES-valencia': 'Noruega', 'ccp': '𑄚𑄧𑄢𑄴𑄃𑄮𑄠𑄬', 'ce': 'Норвеги', 'ceb': 'Norway', 'cgg': 'Noorwe', 'chr': 'ᏃᏪ', 'ckb': 'نۆرویژ', 'cs': 'Norsko', 'cv': 'Норвеги', 'cy': 'Norwy', 'da': 'Norge', 'dav': 'Norwe', 'de': 'Norwegen', 'dje': 'Norveej', 'dsb': 'Norwegska', 'dz': 'ནོར་ཝེ', 'ebu': 'Norwe', 'ee': 'Norway nutome', 'el': 'Νορβηγία', 'el-polyton': 'Νορβηγία', 'en': 'Norway', 'en-Dsrt': '𐐤𐐱𐑉𐐶𐐩', 'eo': 'Norvegujo', 'es': 'Noruega', 'et': 'Norra', 'eu': 'Norvegia', 'ewo': 'Nɔrəvɛ́s', 'fa': 'نروژ', 'ff': 'Norwees', 'ff-Adlm': '𞤐𞤮𞤪𞤺𞤫𞤴𞤢𞥄', 'ff-Latn': 'Norwees', 'fi': 'Norja', 'fil': 'Norway', 'fo': 'Noreg', 'fr': 'Norvège', 'fur': 'Norvegje', 'fy': 'Noarwegen', 'ga': 'an Iorua', 'gd': 'Nirribhidh', 'gl': 'Noruega', 'gsw': 'Norweege', 'gu': 'નૉર્વે', 'guz': 'Norwe', 'ha': 'Norwe', 'he': 'נורווגיה', 'hi': 'नॉर्वे', 'hi-Latn': 'Norway', 'hr': 'Norveška', 'hsb': 'Norwegska', 'ht': 'Norvège', 'hu': 'Norvégia', 'hy': 'Նորվեգիա', 'ia': 'Norvegia', 'id': 'Norwegia', 'ig': 'Norway', 'is': 'Noregur', 'it': 'Norvegia', 'ja': 'ノルウェー', 'jgo': 'Nɔlɛvɛ́jɛ', 'jmc': 'Norwe', 'jv': 'Nurwègen', 'ka': 'ნორვეგია', 'kaa': 'Норвегия', 'kaa-Cyrl': 'Норвегия', 'kab': 'Nurvij', 'kam': 'Norwe', 'kde': 'Norwe', 'kea': 'Noruega', 'kgp': 'Nãrovéga', 'khq': 'Norveej', 'ki': 'Norwe', 'kk': 'Норвегия', 'kk-Cyrl': 'Норвегия', 'kln': 'Emetab Norway', 'km': 'ន័រវែស', 'kn': 'ನಾರ್ವೆ', 'ko': '노르웨이', 'kok': 'नॉर्वे', 'kok-Deva': 'नॉर्वे', 'ks': 'ناروے', 'ks-Arab': 'ناروے', 'ksb': 'Nolwei', 'ksf': 'nɔrvɛjǝ', 'ksh': 'Norrweeje', 'ku': 'Norwêc', 'kxv': 'norvay', 'kxv-Deva': 'नॉर्वे', 'kxv-Latn': 'norvay', 'kxv-Orya': 'ନରୱେ', 'kxv-Telu': 'నార్వే', 'ky': 'Норвегия', 'lag': 'Norwe', 'lb': 'Norwegen', 'lg': 'Nowe', 'lld': 'Norvegia', 'ln': 'Norivezɛ', 'lo': 'ນໍເວ', 'lt': 'Norvegija', 'lu': 'Noriveje', 'luo': 'Norway', 'luy': 'Norwe', 'lv': 'Norvēģija', 'mai': 'नार्वे', 'mas': 'Norwe', 'mer': 'Norwi', 'mfe': 'Norvez', 'mg': 'Nôrvezy', 'mgh': 'Unorwe', 'mi': 'Nōwei', 'mk': 'Норвешка', 'ml': 'നോർവെ', 'mn': 'Норвег', 'mr': 'नॉर्वे', 'ms': 'Norway', 'mt': 'in-Norveġja', 'mua': 'Norvegǝ', 'my': 'နော်ဝေ', 'mzn': 'نروژ', 'naq': 'Noorweeb', 'nb': 'Norge', 'nd': 'Noweyi', 'ne': 'नर्वे', 'nl': 'Noorwegen', 'nmg': 'Nɔrvɛg', 'nn': 'Noreg', 'no': 'Norge', 'nqo': 'ߣߐߙߑߝ߭ߍߖ', 'nyn': 'Noorwe', 'om': 'Noorwey', 'or': 'ନରୱେ', 'pa': 'ਨਾਰਵੇ', 'pa-Guru': 'ਨਾਰਵੇ', 'pcm': 'Nọ́we', 'pl': 'Norwegia', 'ps': 'ناروۍ', 'pt': 'Noruega', 'qu': 'Noruega', 'rm': 'Norvegia', 'rn': 'Noruveji', 'ro': 'Norvegia', 'rof': 'Norwe', 'ru': 'Норвегия', 'rwk': 'Norwe', 'sah': 'Норвегия', 'saq': 'Norwe', 'sbp': 'Nolwe', 'sc': 'Norvègia', 'sd': 'ناروي', 'sd-Arab': 'ناروي', 'se': 'Norga', 'seh': 'Noruega', 'ses': 'Norveej', 'sg': 'Nörvêzi', 'shi': 'ⵏⵏⵔⵡⵉⵊ', 'shi-Latn': 'nnrwij', 'shi-Tfng': 'ⵏⵏⵔⵡⵉⵊ', 'si': 'නෝර්වේ', 'sk': 'Nórsko', 'sl': 'Norveška', 'smn': 'Taažâ', 'sn': 'Norway', 'so': 'Noorweey', 'sq': 'Norvegji', 'sr': 'Норвешка', 'sr-Cyrl': 'Норвешка', 'sr-Latn': 'Norveška', 'sv': 'Norge', 'sw': 'Norway', 'syr': 'ܢܘܪܒܝܓ', 'ta': 'நார்வே', 'te': 'నార్వే', 'teo': 'Norwe', 'tg': 'Норвегия', 'th': 'นอร์เวย์', 'ti': 'ኖርወይ', 'tk': 'Norwegiýa', 'to': 'Noauē', 'tr': 'Norveç', 'tt': 'Норвегия', 'twq': 'Norveej', 'tzm': 'Nnurwij', 'ug': 'نورۋېگىيە', 'uk': 'Норвегія', 'ur': 'ناروے', 'uz': 'Norvegiya', 'uz-Cyrl': 'Норвегия', 'uz-Latn': 'Norvegiya', 'vai': 'ꗟꖺꔃ', 'vai-Latn': 'Nɔɔwe', 'vai-Vaii': 'ꗟꖺꔃ', 'vec': 'Norveja', 'vi': 'Na Uy', 'vun': 'Norwe', 'wae': 'Norwäge', 'wo': 'Norwees', 'xh': 'ENorway', 'xnr': 'नॉर्वे', 'xog': 'Nowe', 'yav': 'nɔlfɛ́ɛs', 'yi': 'נארוועגיע', 'yo': 'Nọọwii', 'yrl': 'Nuruwega', 'yue': '挪威', 'yue-Hans': '挪威', 'yue-Hant': '挪威', 'zgh': 'ⵏⵏⵔⵡⵉⵊ', 'zh': '挪威', 'zh-Hans': '挪威', 'zh-Hant': '挪威', 'zu': 'i-Norway'},
    'NP': {'af': 'Nepal', 'agq': 'Nɛkpâa', 'ak': 'Nɛpal', 'am': 'ኔፓል', 'ar': 'نيبال', 'as': 'নেপাল', 'asa': 'Nepali', 'ast': 'Nepal', 'az': 'Nepal', 'az-Cyrl': 'Непал', 'az-Latn': 'Nepal', 'bal-Latn': 'Népál', 'bas': 'Nèpâl', 'be': 'Непал', 'be-tarask': 'Непал', 'bez': 'Hunepali', 'bg': 'Непал', 'blo': 'Neepal', 'bm': 'Nepali', 'bn': 'নেপাল', 'bo': 'བལ་ཡུལ་', 'br': 'Nepal', 'brx': 'नेपाल', 'bs': 'Nepal', 'bs-Cyrl': 'Непал', 'bs-Latn': 'Nepal', 'ca': 'Nepal', 'ca-ES-valencia': 'Nepal', 'ccp': '𑄚𑄬𑄛𑄣𑄴', 'ce': 'Непал', 'ceb': 'Nepal', 'cgg': 'Nepo', 'chr': 'ᏁᏆᎵ', 'ckb': 'نیپال', 'cs': 'Nepál', 'cv': 'Непал', 'cy': 'Nepal', 'da': 'Nepal', 'dav': 'Nepali', 'de': 'Nepal', 'dje': 'Neepal', 'dsb': 'Nepal', 'dz': 'བལ་ཡུལ', 'ebu': 'Nepali', 'ee': 'Nepal nutome', 'el': 'Νεπάλ', 'el-polyton': 'Νεπάλ', 'en': 'Nepal', 'en-Dsrt': '𐐤𐐩𐐹𐐪𐑊', 'eo': 'Nepalo', 'es': 'Nepal', 'et': 'Nepal', 'eu': 'Nepal', 'ewo': 'Nepál', 'fa': 'نپال', 'ff': 'Nepaal', 'ff-Adlm': '𞤐𞤭𞤨𞤢𞥄𞤤', 'ff-Latn': 'Nepaal', 'fi': 'Nepal', 'fil': 'Nepal', 'fo': 'Nepal', 'fr': 'Népal', 'fur': 'Nepal', 'fy': 'Nepal', 'ga': 'Neipeal', 'gd': 'Neapàl', 'gl': 'Nepal', 'gsw': 'Neepal', 'gu': 'નેપાળ', 'guz': 'Nepali', 'ha': 'Nefal', 'he': 'נפאל', 'hi': 'नेपाल', 'hi-Latn': 'Nepal', 'hr': 'Nepal', 'hsb': 'Nepal', 'ht': 'Népal', 'hu': 'Nepál', 'hy': 'Նեպալ', 'ia': 'Nepal', 'id': 'Nepal', 'ig': 'Nepal', 'is': 'Nepal', 'it': 'Nepal', 'ja': 'ネパール', 'jmc': 'Nepali', 'jv': 'Népal', 'ka': 'ნეპალი', 'kaa': 'Непал', 'kaa-Cyrl': 'Непал', 'kab': 'Nipal', 'kam': 'Nepali', 'kde': 'Nepali', 'kea': 'Nepal', 'kgp': 'Nẽpar', 'khq': 'Neepal', 'ki': 'Nepali', 'kk': 'Непал', 'kk-Cyrl': 'Непал', 'kln': 'Emetab Nepal', 'km': 'នេប៉ាល់', 'kn': 'ನೇಪಾಳ', 'ko': '네팔', 'kok': 'नेपाळ', 'kok-Deva': 'नेपाळ', 'ks': 'نیپال', 'ks-Arab': 'نیپال', 'ksb': 'Nepali', 'ksf': 'nɛpal', 'ksh': 'Nepall', 'ku': 'Nepal', 'kxv': 'nepal', 'kxv-Deva': 'नेपाल', 'kxv-Latn': 'nepal', 'kxv-Orya': 'ନେପାଲ', 'kxv-Telu': 'నేపాల్', 'ky': 'Непал', 'lag': 'Nepáali', 'lb': 'Nepal', 'lg': 'Nepalo', 'lld': 'Nepal', 'ln': 'Nepálɛ', 'lo': 'ເນປານ', 'lt': 'Nepalas', 'lu': 'Nepálɛ', 'luo': 'Nepal', 'luy': 'Nepali', 'lv': 'Nepāla', 'mai': 'नेपाल', 'mas': 'Nepali', 'mer': 'Nepali', 'mfe': 'Nepal', 'mg': 'Nepala', 'mi': 'Nepōra', 'mk': 'Непал', 'ml': 'നേപ്പാൾ', 'mn': 'Балба', 'mr': 'नेपाळ', 'ms': 'Nepal', 'mt': 'in-Nepal', 'mua': 'Sǝr Nepal', 'my': 'နီပေါ', 'mzn': 'نپال', 'naq': 'Nepali', 'nb': 'Nepal', 'nd': 'Nephali', 'ne': 'नेपाल', 'nl': 'Nepal', 'nmg': 'Nepal', 'nn': 'Nepal', 'no': 'Nepal', 'nqo': 'ߣߋߔߊߟ', 'nyn': 'Nepo', 'om': 'Neeppal', 'or': 'ନେପାଳ', 'pa': 'ਨੇਪਾਲ', 'pa-Guru': 'ਨੇਪਾਲ', 'pcm': 'Nẹ́pal', 'pl': 'Nepal', 'ps': 'نیپال', 'pt': 'Nepal', 'qu': 'Nepal', 'rm': 'Nepal', 'rn': 'Nepali', 'ro': 'Nepal', 'rof': 'Nepali', 'ru': 'Непал', 'rwk': 'Nepali', 'saq': 'Nepali', 'sbp': 'Nepali', 'sc': 'Nèpal', 'sd': 'نيپال', 'sd-Arab': 'نيپال', 'se': 'Nepal', 'seh': 'Nepal', 'ses': 'Neepal', 'sg': 'Nëpâli', 'shi': 'ⵏⵉⴱⴰⵍ', 'shi-Latn': 'nibal', 'shi-Tfng': 'ⵏⵉⴱⴰⵍ', 'si': 'නේපාලය', 'sk': 'Nepál', 'sl': 'Nepal', 'smn': 'Nepal', 'sn': 'Nepal', 'so': 'Nebaal', 'sq': 'Nepal', 'sr': 'Непал', 'sr-Cyrl': 'Непал', 'sr-Latn': 'Nepal', 'sv': 'Nepal', 'sw': 'Nepal', 'syr': 'ܢܝܦܐܠ', 'ta': 'நேபாளம்', 'te': 'నేపాల్', 'teo': 'Nepali', 'tg': 'Непал', 'th': 'เนปาล', 'ti': 'ኔፓል', 'tk': 'Nepal', 'to': 'Nepali', 'tr': 'Nepal', 'tt': 'Непал', 'twq': 'Neepal', 'tzm': 'Nippal', 'ug': 'نېپال', 'uk': 'Непал', 'ur': 'نیپال', 'uz': 'Nepal', 'uz-Cyrl': 'Непал', 'uz-Latn': 'Nepal', 'vai': 'ꕇꕐꔷ', 'vai-Latn': 'Nepa', 'vai-Vaii': 'ꕇꕐꔷ', 'vec': 'Nepal', 'vi': 'Nepal', 'vun': 'Nepali', 'wae': 'Nepal', 'wo': 'Nepaal', 'xh': 'ENepal', 'xnr': 'नेपाल', 'xog': 'Nepalo', 'yav': 'nepáal', 'yi': 'נעפּאַל', 'yo': 'Nepa', 'yrl': 'Nepau', 'yue': '尼泊爾', 'yue-Hans': '尼泊尔', 'yue-Hant': '尼泊爾', 'zgh': 'ⵏⵉⴱⴰⵍ', 'zh': '尼泊尔', 'zh-Hans': '尼泊尔', 'zh-Hant': '尼泊爾', 'zu': 'i-Nepal'},
    'NR': {'af': 'Nauru', 'agq': 'Nàwulù', 'ak': 'Naworu', 'am': 'ናኡሩ', 'ar': 'ناورو', 'as': 'নাউৰু', 'asa': 'Nauru', 'ast': 'Nauru', 'az': 'Nauru', 'az-Cyrl': 'Науру', 'az-Latn': 'Nauru', 'bal-Latn': 'Náuru', 'bas': 'Nerù', 'be': 'Науру', 'be-tarask': 'Науру', 'bez': 'Hunauru', 'bg': 'Науру', 'blo': 'Nawuru', 'bm': 'Nawuru', 'bn': 'নাউরু', 'br': 'Nauru', 'brx': 'नाउरू', 'bs': 'Nauru', 'bs-Cyrl': 'Науру', 'bs-Latn': 'Nauru', 'ca': 'Nauru', 'ca-ES-valencia': 'Nauru', 'ccp': '𑄚𑄃𑄪𑄢𑄪', 'ce': 'Науру', 'ceb': 'Nauru', 'cgg': 'Nauru', 'chr': 'ᏃᎤᎷ', 'ckb': 'نائوروو', 'cs': 'Nauru', 'cv': 'Науру', 'cy': 'Nauru', 'da': 'Nauru', 'dav': 'Nauru', 'de': 'Nauru', 'dje': 'Nauru', 'dsb': 'Nauru', 'dz': 'ནའུ་རུ་', 'ebu': 'Nauru', 'ee': 'Nauru nutome', 'el': 'Ναουρού', 'el-polyton': 'Ναουρού', 'en': 'Nauru', 'en-Dsrt': '𐐤𐐪𐐭𐑉𐐭', 'eo': 'Nauro', 'es': 'Nauru', 'et': 'Nauru', 'eu': 'Nauru', 'ewo': 'Naurú', 'fa': 'نائورو', 'ff': 'Nawuru', 'ff-Adlm': '𞤐𞤢𞤱𞤪𞤵', 'ff-Latn': 'Nawuru', 'fi': 'Nauru', 'fil': 'Nauru', 'fo': 'Nauru', 'fr': 'Nauru', 'fur': 'Nauru', 'fy': 'Nauru', 'ga': 'Nárú', 'gd': 'Nabhru', 'gl': 'Nauru', 'gsw': 'Nauru', 'gu': 'નૌરુ', 'guz': 'Nauru', 'ha': 'Nauru', 'he': 'נאורו', 'hi': 'नाउरु', 'hi-Latn': 'Nauru', 'hr': 'Nauru', 'hsb': 'Nauru', 'ht': 'Nauru', 'hu': 'Nauru', 'hy': 'Նաուրու', 'ia': 'Nauru', 'id': 'Nauru', 'ie': 'Nauru', 'ig': 'Nauru', 'is': 'Nárú', 'it': 'Nauru', 'ja': 'ナウル', 'jmc': 'Nauru', 'jv': 'Nauru', 'ka': 'ნაურუ', 'kaa': 'Науру', 'kaa-Cyrl': 'Науру', 'kab': 'Nuru', 'kam': 'Nauru', 'kde': 'Naulu', 'kea': 'Nauru', 'kgp': 'Nỹuru', 'khq': 'Nauru', 'ki': 'Nauru', 'kk': 'Науру', 'kk-Cyrl': 'Науру', 'kln': 'Emetab Nauru', 'km': 'ណូរូ', 'kn': 'ನೌರು', 'ko': '나우루', 'kok': 'नावरू', 'kok-Deva': 'नावरू', 'ks': 'نارووٗ', 'ks-Arab': 'نارووٗ', 'ksb': 'Naulu', 'ksf': 'nwarú', 'ksh': 'Nauru', 'ku': 'Naûrû', 'kxv': 'nauru', 'kxv-Deva': 'नाउरु', 'kxv-Latn': 'nauru', 'kxv-Orya': 'ନାଉରୁ', 'kxv-Telu': 'నౌరు', 'ky': 'Науру', 'lag': 'Naúuru', 'lb': 'Nauru', 'lg': 'Nawuru', 'lld': 'Nauru', 'ln': 'Nauru', 'lo': 'ນາອູຣູ', 'lt': 'Nauru', 'lu': 'Nauru', 'luo': 'Nauru', 'luy': 'Nauru', 'lv': 'Nauru', 'mai': 'नौरू', 'mas': 'Nauru', 'mer': 'Nauru', 'mfe': 'Nauru', 'mg': 'Naorò', 'mi': 'Nauru', 'mk': 'Науру', 'ml': 'നൗറു', 'mn': 'Науру', 'mr': 'नाउरू', 'ms': 'Nauru', 'mt': 'Nauru', 'mua': 'Sǝr Nauru', 'my': 'နော်ရူး', 'mzn': 'نائورو', 'naq': 'Nauru', 'nb': 'Nauru', 'nd': 'Nauru', 'ne': 'नाउरू', 'nl': 'Nauru', 'nmg': 'Noru', 'nn': 'Nauru', 'no': 'Nauru', 'nqo': 'ߣߏ߯ߙߎ߫', 'nyn': 'Nauru', 'om': 'Naawuruu', 'or': 'ନାଉରୁ', 'pa': 'ਨਾਉਰੂ', 'pa-Guru': 'ਨਾਉਰੂ', 'pcm': 'Náuru', 'pl': 'Nauru', 'ps': 'نایرو', 'pt': 'Nauru', 'qu': 'Nauru', 'rm': 'Nauru', 'rn': 'Nawuru', 'ro': 'Nauru', 'rof': 'Nauru', 'ru': 'Науру', 'rwk': 'Nauru', 'saq': 'Nauru', 'sbp': 'Nawulu', 'sc': 'Nauru', 'sd': 'نائورو', 'sd-Arab': 'نائورو', 'se': 'Nauru', 'seh': 'Nauru', 'ses': 'Nauru', 'sg': 'Nauru', 'shi': 'ⵏⴰⵡⵔⵓ', 'shi-Latn': 'nawru', 'shi-Tfng': 'ⵏⴰⵡⵔⵓ', 'si': 'නාවුරු', 'sk': 'Nauru', 'sl': 'Nauru', 'smn': 'Nauru', 'sn': 'Nauru', 'so': 'Nauru', 'sq': 'Nauru', 'sr': 'Науру', 'sr-Cyrl': 'Науру', 'sr-Latn': 'Nauru', 'sv': 'Nauru', 'sw': 'Nauru', 'syr': 'ܢܐܘܪܘ', 'ta': 'நௌரு', 'te': 'నౌరు', 'teo': 'Nauru', 'tg': 'Науру', 'th': 'นาอูรู', 'ti': 'ናውሩ', 'tk': 'Nauru', 'to': 'Naulu', 'tr': 'Nauru', 'tt': 'Науру', 'twq': 'Nauru', 'tzm': 'Nawru', 'ug': 'ناۋرۇ', 'uk': 'Науру', 'ur': 'نؤرو', 'uz': 'Nauru', 'uz-Cyrl': 'Науру', 'uz-Latn': 'Nauru', 'vai': 'ꖆꖩ', 'vai-Latn': 'Noru', 'vai-Vaii': 'ꖆꖩ', 'vec': 'Nauru', 'vi': 'Nauru', 'vun': 'Nauru', 'wae': 'Nauru', 'wo': 'Nawru', 'xh': 'ENauru', 'xnr': 'नाउरु', 'xog': 'Nawuru', 'yav': 'nawulú', 'yo': 'Nauru', 'yrl': 'Nauru', 'yue': '諾魯', 'yue-Hans': '诺鲁', 'yue-Hant': '諾魯', 'zgh': 'ⵏⴰⵡⵔⵓ', 'zh': '瑙鲁', 'zh-Hans': '瑙鲁', 'zh-Hant': '諾魯', 'zu': 'i-Nauru'},
    'NT': {'en': 'Neutral Zone'},
    'NU': {'af': 'Niue', 'agq': 'Niyu', 'ak': 'Niyu', 'am': 'ኒዌ', 'ar': 'نيوي', 'as': 'নিউ', 'asa': 'Niue', 'ast': 'Niue', 'az': 'Niue', 'az-Cyrl': 'Ниуе', 'az-Latn': 'Niue', 'bal-Latn': 'Niué', 'bas': 'Nìuɛ̀', 'be': 'Ніуэ', 'be-tarask': 'Ніуэ', 'bez': 'Huniue', 'bg': 'Ниуе', 'blo': 'Niwuye', 'bm': 'Nyuwe', 'bn': 'নিউয়ে', 'br': 'Niue', 'brx': 'नीयूए', 'bs': 'Niue', 'bs-Cyrl': 'Ниуе', 'bs-Latn': 'Niue', 'ca': 'Niue', 'ca-ES-valencia': 'Niue', 'ccp': '𑄚𑄨𑄃𑄪𑄠𑄬', 'ce': 'Ниуэ', 'ceb': 'Niue', 'cgg': 'Niue', 'chr': 'ᏂᏳ', 'ckb': 'نیووئی', 'cs': 'Niue', 'cv': 'Ниуэ', 'cy': 'Niue', 'da': 'Niue', 'dav': 'Niue', 'de': 'Niue', 'dje': 'Niue', 'dsb': 'Niue', 'dz': 'ནི་ཨུ་ཨཻ', 'ebu': 'Niue', 'ee': 'Niue nutome', 'el': 'Νιούε', 'el-polyton': 'Νιούε', 'en': 'Niue', 'en-Dsrt': '𐐤𐐷𐐭𐐩', 'eo': 'Niuo', 'es': 'Niue', 'et': 'Niue', 'eu': 'Niue', 'ewo': 'Niué', 'fa': 'نیوئه', 'ff': 'Niuwe', 'ff-Adlm': '𞤐𞤵𞥅𞤱𞤭', 'ff-Latn': 'Niuwe', 'fi': 'Niue', 'fil': 'Niue', 'fo': 'Niue', 'fr': 'Niue', 'fur': 'Niue', 'fy': 'Niue', 'ga': 'Niue', 'gd': 'Niue', 'gl': 'Niue', 'gsw': 'Niue', 'gu': 'નીયુ', 'guz': 'Niue', 'ha': 'Niue', 'he': 'ניווה', 'hi': 'नीयू', 'hi-Latn': 'Niue', 'hr': 'Niue', 'hsb': 'Niue', 'ht': 'Niue', 'hu': 'Niue', 'hy': 'Նիուե', 'ia': 'Niue', 'id': 'Niue', 'ig': 'Niue', 'is': 'Niue', 'it': 'Niue', 'ja': 'ニウエ', 'jmc': 'Niue', 'jv': 'Niue', 'ka': 'ნიუე', 'kab': 'Niwi', 'kam': 'Niue', 'kde': 'Niue', 'kea': 'Niue', 'kgp': 'Nĩvue', 'khq': 'Niue', 'ki': 'Niue', 'kk': 'Ниуэ', 'kk-Cyrl': 'Ниуэ', 'kln': 'Emetab Niue', 'km': 'ណៀ', 'kn': 'ನಿಯು', 'ko': '니우에', 'kok': 'नीयू', 'kok-Deva': 'नीयू', 'ks': 'نیوٗ', 'ks-Arab': 'نیوٗ', 'ksb': 'Niue', 'ksf': 'niwɛ́', 'ksh': 'Niue', 'ku': 'Niûe', 'kxv': 'niyu', 'kxv-Deva': 'नीयू', 'kxv-Latn': 'niyu', 'kxv-Orya': 'ନିୟୁ', 'kxv-Telu': 'నియూ', 'ky': 'Ниуэ', 'lag': 'Niúue', 'lb': 'Niue', 'lg': 'Niyuwe', 'lld': 'Niue', 'ln': 'Nyué', 'lo': 'ນີອູເອ', 'lt': 'Niujė', 'lu': 'Nyue', 'luo': 'Niue', 'luy': 'Niue', 'lv': 'Niue', 'mai': 'नीयू', 'mas': 'Niue', 'mer': 'Niue', 'mfe': 'Niowe', 'mg': 'Nioé', 'mi': 'Niue', 'mk': 'Ниује', 'ml': 'ന്യൂയി', 'mn': 'Ниуэ', 'mr': 'नीयू', 'ms': 'Niue', 'mt': 'Niue', 'mua': 'Niwe', 'my': 'နီဥူအေ', 'mzn': 'نیئو', 'naq': 'Niue', 'nb': 'Niue', 'nd': 'Niue', 'ne': 'नियुइ', 'nl': 'Niue', 'nmg': 'Niuɛ', 'nn': 'Niue', 'no': 'Niue', 'nqo': 'ߣߎ߳ߋ߫', 'nyn': 'Niue', 'om': 'Niwu’e', 'or': 'ନିଉ', 'pa': 'ਨਿਯੂ', 'pa-Guru': 'ਨਿਯੂ', 'pcm': 'Niúẹ', 'pl': 'Niue', 'ps': 'نیوو', 'pt': 'Niue', 'qu': 'Niue', 'rm': 'Niue', 'rn': 'Niyuwe', 'ro': 'Niue', 'rof': 'Niue', 'ru': 'Ниуэ', 'rwk': 'Niue', 'saq': 'Niue', 'sbp': 'Niwue', 'sc': 'Niue', 'sd': 'نووي', 'sd-Arab': 'نووي', 'se': 'Niue', 'seh': 'Niue', 'ses': 'Niue', 'sg': 'Niue', 'shi': 'ⵏⵉⵡⵉ', 'shi-Latn': 'niwi', 'shi-Tfng': 'ⵏⵉⵡⵉ', 'si': 'නියූ', 'sk': 'Niue', 'sl': 'Niue', 'smn': 'Niue', 'sn': 'Niue', 'so': 'Niue', 'sq': 'Niue', 'sr': 'Ниуе', 'sr-Cyrl': 'Ниуе', 'sr-Latn': 'Niue', 'sv': 'Niue', 'sw': 'Niue', 'ta': 'நியுவே', 'te': 'నియూ', 'teo': 'Niue', 'tg': 'Ниуэ', 'th': 'นีอูเอ', 'ti': 'ኒዩ', 'tk': 'Niue', 'to': 'Niuē', 'tr': 'Niue', 'tt': 'Ниуэ', 'twq': 'Niue', 'tzm': 'Niwi', 'ug': 'نيۇئې', 'uk': 'Ніуе', 'ur': 'نیئو', 'uz': 'Niue', 'uz-Cyrl': 'Ниуэ', 'uz-Latn': 'Niue', 'vai': 'ꖸꔃꔤ', 'vai-Latn': 'Niwe', 'vai-Vaii': 'ꖸꔃꔤ', 'vec': 'Niue', 'vi': 'Niue', 'vun': 'Niue', 'wae': 'Niue', 'wo': 'Niw', 'xh': 'ENiue', 'xnr': 'नीयू', 'xog': 'Niyuwe', 'yav': 'niyuwé', 'yo': 'Niue', 'yrl': 'Niwe', 'yue': '紐埃島', 'yue-Hans': '纽埃岛', 'yue-Hant': '紐埃島', 'zgh': 'ⵏⵉⵡⵉ', 'zh': '纽埃', 'zh-Hans': '纽埃', 'zh-Hant': '紐埃島', 'zu': 'i-Niue'},
    'NZ': {'af': 'Nieu-Seeland', 'agq': 'Zìlân È fūghū', 'ak': 'Ziland Foforo', 'am': 'ኒው ዚላንድ', 'ar': 'نيوزيلندا', 'as': 'নিউজিলেণ্ড', 'asa': 'Nyudhilandi', 'ast': 'Nueva Zelanda', 'az': 'Yeni Zelandiya', 'az-Cyrl': 'Јени Зеландија', 'az-Latn': 'Yeni Zelandiya', 'bal-Latn': 'Nyu Zilánd', 'bas': 'Sìlând Yɔ̀ndɔ', 'be': 'Новая Зеландыя', 'be-tarask': 'Новая Зеландыя', 'bez': 'Hunyuzilandi', 'bg': 'Нова Зеландия', 'bgn': 'نیوزلنڈ', 'blo': 'Selanɖ afɔlɩ', 'bm': 'Zelandi Koura', 'bn': 'নিউজিল্যান্ড', 'br': 'Zeland-Nevez', 'brx': 'न्यूजीलेण्ड', 'bs': 'Novi Zeland', 'bs-Cyrl': 'Нови Зеланд', 'bs-Latn': 'Novi Zeland', 'ca': 'Nova Zelanda', 'ca-ES-valencia': 'Nova Zelanda', 'ccp': '𑄚𑄨𑄃𑄪𑄎𑄨𑄣𑄳𑄠𑄚𑄴𑄓𑄴', 'ce': 'Керла Зеланди', 'ceb': 'New Zealand', 'cgg': 'Niuzirandi', 'chr': 'ᎢᏤ ᏏᎢᎴᏂᏗ', 'ckb': 'نیوزیلاند', 'cs': 'Nový Zéland', 'cv': 'Ҫӗнӗ Зеланди', 'cy': 'Seland Newydd', 'da': 'New Zealand', 'dav': 'Nyuzilandi', 'de': 'Neuseeland', 'dje': 'Zeelandu Taaga', 'dsb': 'Nowoseelandska', 'dz': 'ནིའུ་ཛི་ལེནཌ', 'ebu': 'Nyuzilandi', 'ee': 'New Zealand nutome', 'el': 'Νέα Ζηλανδία', 'el-polyton': 'Νέα Ζηλανδία', 'en': 'New Zealand', 'en-Dsrt': '𐐤𐐭 𐐞𐐨𐑊𐐲𐑌𐐼', 'eo': 'Nov-Zelando', 'es': 'Nueva Zelanda', 'et': 'Uus-Meremaa', 'eu': 'Zeelanda Berria', 'ewo': 'Ǹkpámɛn Zeláŋ', 'fa': 'نیوزیلند', 'ff': 'Nuwel Selannda', 'ff-Adlm': '𞤐𞤫𞤱 𞤟𞤫𞤤𞤢𞤲𞤣', 'ff-Latn': 'Nuwel Selannda', 'fi': 'Uusi-Seelanti', 'fil': 'New Zealand', 'fo': 'Nýsæland', 'fr': 'Nouvelle-Zélande', 'fur': 'Gnove Zelande', 'fy': 'Nij-Seelân', 'ga': 'an Nua-Shéalainn', 'gd': 'Sealainn Nuadh', 'gl': 'Nova Zelandia', 'gsw': 'Nöiseeland', 'gu': 'ન્યુઝીલેન્ડ', 'guz': 'Nyuzilandi', 'ha': 'Nuzilan', 'haw': 'Aotearoa', 'he': 'ניו זילנד', 'hi': 'न्यूज़ीलैंड', 'hi-Latn': 'New Zealand', 'hr': 'Novi Zeland', 'hsb': 'Nowoseelandska', 'ht': 'Nouvelle-Zélande', 'hu': 'Új-Zéland', 'hy': 'Նոր Զելանդիա', 'ia': 'Nove Zelanda', 'id': 'Selandia Baru', 'ie': 'Nov-Zeland', 'ig': 'New Zealand', 'is': 'Nýja-Sjáland', 'it': 'Nuova Zelanda', 'ja': 'ニュージーランド', 'jmc': 'Nyuzilandi', 'jv': 'Selandia Anyar', 'ka': 'ახალი ზელანდია', 'kaa': 'Жаӊа Зеландия', 'kaa-Cyrl': 'Жаӊа Зеландия', 'kab': 'Ziland Tamaynut', 'kam': 'Nyuzilandi', 'kde': 'Nyuzilandi', 'kea': 'Nova Zilándia', 'kgp': 'Jerỹnija Tãg', 'khq': 'Zeelandu Taaga', 'ki': 'Nyuzilandi', 'kk': 'Жаңа Зеландия', 'kk-Cyrl': 'Жаңа Зеландия', 'kln': 'Emetab New Zealand', 'km': 'នូវែល\u200bសេឡង់', 'kn': 'ನ್ಯೂಜಿಲೆಂಡ್', 'ko': '뉴질랜드', 'kok': 'न्युझीलॅन्ड', 'kok-Deva': 'न्युझीलॅन्ड', 'ks': 'نیوزی لینڈ', 'ks-Arab': 'نیوزی لینڈ', 'ksb': 'Nyuzilandi', 'ksf': 'zɛlan anyɔ́n', 'ksh': 'Neuseeland', 'ku': 'Zelandaya Nû', 'kxv': 'nyu jīlanḍ', 'kxv-Deva': 'न्यूज़ीलेंड', 'kxv-Latn': 'nyu jīlanḍ', 'kxv-Orya': 'ନିୟୁଜିଲାଣ୍ଡ', 'kxv-Telu': 'న్యూజిలాండ్', 'ky': 'Жаңы Зеландия', 'lag': 'Nyuzílandi', 'lb': 'Neiséiland', 'lg': 'Niyuziirandi', 'lld': 'Nöia Zelanda', 'ln': 'Zelandɛ ya sika', 'lo': 'ນິວຊີແລນ', 'lt': 'Naujoji Zelandija', 'lu': 'Zelanda wa mumu', 'luo': 'New Zealand', 'luy': 'Nyuzilandi', 'lv': 'Jaunzēlande', 'mai': 'न्यूजीलैंड', 'mas': 'Nyuzilandi', 'mer': 'Niuzilandi', 'mfe': 'Nouvel Zeland', 'mg': 'Nouvelle-Zélande', 'mi': 'Aotearoa', 'mk': 'Нов Зеланд', 'ml': 'ന്യൂസിലൻഡ്', 'mn': 'Шинэ Зеланд', 'mr': 'न्यूझीलंड', 'ms': 'New Zealand', 'mt': 'New Zealand', 'mua': 'Zeland mafuu', 'my': 'နယူးဇီလန်', 'mzn': 'نیوزلند', 'naq': 'New Zealandi', 'nb': 'New Zealand', 'nd': 'New Zealand', 'ne': 'न्युजिल्याण्ड', 'nl': 'Nieuw-Zeeland', 'nmg': 'Zeland nwanah', 'nn': 'New Zealand', 'no': 'New Zealand', 'nqo': 'ߖ߭ߋߟߊ߲ߘߌ߫-ߞߎߘߊ߫', 'nyn': 'Niuzirandi', 'om': 'Neewu Zilaand', 'or': 'ନ୍ୟୁଜିଲାଣ୍ଡ', 'pa': 'ਨਿਊਜ਼ੀਲੈਂਡ', 'pa-Guru': 'ਨਿਊਜ਼ੀਲੈਂਡ', 'pcm': 'Niú Zíland', 'pl': 'Nowa Zelandia', 'ps': 'نیوزیلنډ', 'pt': 'Nova Zelândia', 'qu': 'Nueva Zelanda', 'rm': 'Nova Zelanda', 'rn': 'Nuvelizelandi', 'ro': 'Noua Zeelandă', 'rof': 'Nyuzilandi', 'ru': 'Новая Зеландия', 'rwk': 'Nyuzilandi', 'saq': 'Nyuzilandi', 'sbp': 'Nyusilendi', 'sc': 'Zelanda Noa', 'sd': 'نيو زيلينڊ', 'sd-Arab': 'نيو زيلينڊ', 'se': 'Ođđa-Selánda', 'seh': 'Nova Zelândia', 'ses': 'Zeelandu Taaga', 'sg': 'Finî Zelânde', 'shi': 'ⵏⵢⵓⵣⵉⵍⴰⵏⴷⴰ', 'shi-Latn': 'nyuzilanda', 'shi-Tfng': 'ⵏⵢⵓⵣⵉⵍⴰⵏⴷⴰ', 'si': 'නවසීලන්තය', 'sk': 'Nový Zéland', 'sl': 'Nova Zelandija', 'smn': 'Uđđâ-Seeland', 'sn': 'New Zealand', 'so': 'Niyuusiilaand', 'sq': 'Zelandë e Re', 'sr': 'Нови Зеланд', 'sr-Cyrl': 'Нови Зеланд', 'sr-Latn': 'Novi Zeland', 'sv': 'Nya Zeeland', 'sw': 'Nyuzilandi', 'syr': 'ܢܝܘ ܙܝܠܢܕ', 'ta': 'நியூசிலாந்து', 'te': 'న్యూజిలాండ్', 'teo': 'Nyuzilandi', 'tg': 'Зеландияи Нав', 'th': 'นิวซีแลนด์', 'ti': 'ኒው ዚላንድ', 'tk': 'Täze Zelandiýa', 'to': 'Nuʻusila', 'tr': 'Yeni Zelanda', 'tt': 'Яңа Зеландия', 'twq': 'Zeelandu Taaga', 'tzm': 'Zilanḍa Tamaynut', 'ug': 'يېڭى زېلاندىيە', 'uk': 'Нова Зеландія', 'ur': 'نیوزی لینڈ', 'uz': 'Yangi Zelandiya', 'uz-Cyrl': 'Янги Зеландия', 'uz-Latn': 'Yangi Zelandiya', 'vai': 'ꔽꔤ ꖨꕮ ꕯꕮꕊ', 'vai-Latn': 'Zilɛŋ Námaá', 'vai-Vaii': 'ꔽꔤ ꖨꕮ ꕯꕮꕊ', 'vec': 'Nova Zelanda', 'vi': 'New Zealand', 'vun': 'Nyuzilandi', 'wae': 'Niwséland', 'wo': 'Nuwel Selànd', 'xh': 'ENew Zealand', 'xnr': 'न्यूज़ीलैंड', 'xog': 'Niyuziirandi', 'yav': 'nufɛ́l seláand', 'yi': 'ניו זילאַנד', 'yo': 'Ṣilandi Titun', 'yrl': 'Serãdiya Pisasú', 'yue': '新西蘭', 'yue-Hans': '纽西兰', 'yue-Hant': '新西蘭', 'zgh': 'ⵏⵢⵓⵣⵉⵍⴰⵏⴷⴰ', 'zh': '新西兰', 'zh-Hans': '新西兰', 'zh-Hant': '紐西蘭', 'zu': 'i-New Zealand'},
    'Nagm': {'bal-Latn': 'Nag Mundari', 'bs': 'nag mundari pismo', 'bs-Cyrl': 'наг мундари писмо', 'bs-Latn': 'nag mundari pismo', 'en': 'Nag Mundari', 'et': 'Nagi mundari', 'eu': 'nag mundariera', 'ff-Adlm': '𞤐𞤢𞤺 𞤃𞤵𞤲𞤣𞤢𞤪𞤭', 'fi': 'Nagin mundariaakkoset', 'gd': 'Nag Mundari', 'hi-Latn': 'Nag Mundari', 'ig': 'Nag Mundari', 'kk': 'наг мундари жазуы', 'kk-Cyrl': 'наг мундари жазуы', 'mn': 'Наг Мундари', 'nl': 'Nag Mundari', 'ro': 'nag mundari', 'sc': 'nag mundari', 'sq': 'nag mundari', 'tg': 'Наг Мундарӣ', 'vec': 'nag mundari'},
    'Nand': {'bal-Latn': 'Nandinagari', 'bs': 'nandinagari pismo', 'bs-Cyrl': 'нандинагари писмо', 'bs-Latn': 'nandinagari pismo', 'en': 'Nandinagari', 'eu': 'nandinagariera', 'ff-Adlm': '𞤐𞤢𞤲𞤣𞤭𞤲𞤢𞤺𞤢𞤪𞤭', 'fi': 'nandinagari', 'fr': 'nandinagari', 'gd': 'Nandinagari', 'he': 'ננדינאגרי', 'hi-Latn': 'Nandinagari', 'ht': 'nandinagari', 'id': 'Nandinagari', 'ig': 'Nandinagari', 'kk': 'нандинагари жазуы', 'kk-Cyrl': 'нандинагари жазуы', 'mn': 'Нандинагари', 'ms': 'Nandinagari', 'nl': 'Nandinagari', 'nqo': 'ߣߊ߲ߘߌߣߜ߭ߊߙߌ', 'ro': 'nandinagari', 'sc': 'nandinagari', 'so': 'Nandinagari', 'sq': 'nandigarisht', 'sv': 'nandinagari', 'tg': 'Нандинагарӣ', 'vec': 'nandinagari', 'zh': '楠迪梵文', 'zh-Hans': '楠迪梵文', 'zu': 'i-Nandinagari'},
    'Narb': {'ar': 'العربية الشمالية القديمة', 'ast': 'árabe del norte antiguu', 'bal-Latn': 'Ahdi Shemáli Arabi', 'br': 'henarabek an Norzh', 'bs': 'staro sjevernoarapsko pismo', 'bs-Cyrl': 'старо сјеверно арапско писмо', 'bs-Latn': 'staro sjevernoarapsko pismo', 'ca': 'antic nord-aràbic', 'ca-ES-valencia': 'antic nord-aràbic', 'cs': 'staroseveroarabské', 'da': 'gammelt nordarabisk', 'de': 'Altnordarabisch', 'en': 'Old North Arabian', 'et': 'Põhja-Araabia', 'eu': 'iparraldeko arabiera zaharra', 'fa': 'عربی شمالی باستان', 'ff-Adlm': '𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫𞤲𞤳𞤮 𞤀𞥄𞤪𞤢𞤦𞤵 𞤑𞤭𞤯𞤭', 'fi': 'muinaispohjoisarabialainen', 'fy': 'Ald Noard-Arabysk', 'ga': 'Sean-Arabach Thuaidh', 'gd': 'Seann-Arabach Thuathach', 'he': 'ערב צפון עתיק', 'hi-Latn': 'Old North Arabian', 'hr': 'staro sjevernoarapsko pismo', 'id': 'Arab Utara Kuno', 'ig': 'Old North Arabian', 'it': 'arabo settentrionale antico', 'ja': '古代北アラビア文字', 'ka': 'ძველი ჩრდილოეთ-არაბული', 'kk': 'көне солтүстік араб жазуы', 'kk-Cyrl': 'көне солтүстік араб жазуы', 'ko': '옛 북부 아라비아 문자', 'lo': 'ອາຮະເບຍເໜືອໂບຮານ', 'lt': 'Senasis šiaurės arabų', 'mk': 'старосеверноарапско', 'mn': 'Хуучин Хойд Араб бичиг', 'ms': 'Arab Utara Lama', 'nb': 'gammelnordarabisk', 'nl': 'Oud Noord-Arabisch', 'nn': 'gammelnordarabisk', 'no': 'gammelnordarabisk', 'nqo': 'ߕߟߋ߬ߓߐ-ߊߙߊߓߎ', 'rm': 'arab vegl dal nord', 'ro': 'arabă veche din nord', 'ru': 'северноаравийское', 'sc': 'àrabu setentrionale antigu', 'so': 'Carabiyadii Hore ee Wuqooye', 'sq': 'arabishte veriore e vjetër', 'sv': 'fornnordarabiska', 'tg': 'Арабии Шимолии Куҳна', 'th': 'อาระเบียเหนือโบราณ', 'to': 'tohinima fakaʻalepea-tokelau-motuʻa', 'tr': 'Eski Kuzey Arap', 'ug': 'قەدىمكى شىمالىي ئەرەبچە', 'vec': 'àrabo antigo de’l nord', 'vi': 'Chữ Bắc Ả Rập cổ', 'yue': '古北阿拉伯文', 'yue-Hans': '古北阿拉伯文', 'yue-Hant': '古北阿拉伯文', 'zh': '古北方阿拉伯文', 'zh-Hans': '古北方阿拉伯文', 'zh-Hant': '古北阿拉伯文', 'zu': 'i-Old North Arabian'},
    'Nbat': {'ast': 'nabatéu', 'bal-Latn': 'Nabatái', 'bs': 'nabatejsko pismo', 'bs-Cyrl': 'набатаен писмо', 'bs-Latn': 'nabatejsko pismo', 'ca': 'nabateu', 'ca-ES-valencia': 'nabateu', 'cs': 'nabatejské', 'da': 'nabateisk', 'de': 'Nabatäisch', 'en': 'Nabataean', 'et': 'Nabatea', 'eu': 'nabatera', 'fa': 'نبطی', 'ff-Adlm': '𞤐𞤢𞤦𞤢𞤼𞤭𞤴𞤢𞤲', 'fi': 'nabatealainen', 'fy': 'Nabateaansk', 'ga': 'Nabataean', 'gd': 'Nabataean', 'he': 'נבטי', 'hi-Latn': 'Nabataean', 'hr': 'nabatejsko pismo', 'hu': 'Nabateus', 'id': 'Nabataea', 'ig': 'Nabataean', 'it': 'nabateo', 'ja': 'ナバテア文字', 'ka': 'ნაბატეური', 'kk': 'набатей жазуы', 'kk-Cyrl': 'набатей жазуы', 'ko': '나바테아 문자', 'lo': 'ນາບາທາທຽນ', 'lt': 'Nabatėjų', 'mk': 'набатејско', 'mn': 'Набата', 'ms': 'Nabataean', 'nb': 'nabataeansk', 'nl': 'Nabateaans', 'nn': 'nabataeansk', 'no': 'nabataeansk', 'nqo': 'ߣߓߊߕߋ', 'ro': 'nabateeană', 'ru': 'набатейская', 'sc': 'nabateu', 'so': 'Nabataean', 'sq': 'nabateanisht', 'sv': 'nabateiska', 'tg': 'Набатаинӣ', 'th': 'นาบาทาเอียน', 'to': 'tohinima fakanapatea', 'tr': 'Nebati', 'ug': 'ئانباتچە', 'vec': 'nabatèo', 'vi': 'Chữ Nabataean', 'yue': '納巴泰文字', 'yue-Hans': '纳巴泰文字', 'yue-Hant': '納巴泰文字', 'zh': '纳巴泰文', 'zh-Hans': '纳巴泰文', 'zh-Hant': '納巴泰文字', 'zu': 'i-Nabataean'},
    'Newa': {'bal-Latn': 'Newá', 'bs': 'neva pismo', 'bs-Cyrl': 'нева писмо', 'bs-Latn': 'neva pismo', 'ca': 'newar', 'ca-ES-valencia': 'newar', 'et': 'nevari', 'eu': 'newaera', 'ff-Adlm': '𞤐𞤭𞤱𞤢', 'fi': 'newarin newa-tavukirjoitus', 'he': 'נווארית', 'kk': 'ньюа жазуы', 'kk-Cyrl': 'ньюа жазуы', 'km': 'ណេវ៉ា', 'mn': 'Нева', 'nl': 'Newari', 'nqo': 'ߣߋߥߊ', 'sq': 'neva', 'sv': 'newariska', 'tg': 'Нева', 'uk': 'нева', 'yo': 'Èdè Newa', 'zh': '尼瓦文', 'zh-Hans': '尼瓦文', 'zh-Hant': 'Vote 尼瓦爾文', 'en': 'Newa'},
    'Nkdb': {'en': 'Naxi Dongba'},
    'Nkgb': {'ast': 'geba del naxi', 'bs': 'naxi geba pismo', 'bs-Latn': 'naxi geba pismo', 'ca': 'geba', 'ca-ES-valencia': 'geba', 'cs': 'naxi geba', 'da': 'nakhi geba', 'de': 'Geba', 'en': 'Naxi Geba', 'et': 'nasi', 'fi': 'naxi geba', 'fy': 'Naxi Geba', 'gd': 'Naxi Geba', 'hi-Latn': 'Naxi Geba', 'hr': 'naxi geba pismo', 'id': 'Naxi Geba', 'it': 'geba naxi', 'ja': 'ナシ族ゲバ文字', 'ko': '나시 게바 문자', 'lo': 'ກີບາ-ນາຊີ', 'lt': 'Naxi Geba', 'mk': 'насиска геба', 'nb': 'naxi geba', 'nl': 'Naxi Geba', 'nn': 'naxi geba', 'no': 'naxi geba', 'ru': 'наси геба', 'sv': 'naxi geba', 'th': 'กีบา-นาซี', 'to': 'tohinima fakanati-sepa', 'tr': 'Naksi Geba', 'ug': 'ناشىچە', 'vi': 'Chữ Naxi Geba', 'yue': '納西格巴文', 'yue-Hans': '纳西格巴文', 'yue-Hant': '納西格巴文', 'zh': '纳西格巴文', 'zh-Hans': '纳西格巴文', 'zh-Hant': '納西格巴文'},
    'Nkoo': {'af': 'N’Ko', 'ak': 'Nko kasa', 'am': 'ንኮ', 'ar': 'أنكو', 'as': 'এন্\u200dক’', 'ast': 'n’ko', 'az': 'nko', 'az-Latn': 'nko', 'bal-Latn': 'Nekó', 'be': 'нко', 'be-tarask': 'нко', 'bg': 'Н’Ко', 'bn': 'এন’কো', 'brx': 'नंक’', 'bs': 'n’ko pismo', 'bs-Cyrl': 'н’ко писмо', 'bs-Latn': 'n’ko pismo', 'ca': 'n’Ko', 'ca-ES-valencia': 'n’Ko', 'ccp': '𑄃𑄬𑄚𑄴𑄇𑄮', 'chr': 'ᎾᎪ', 'cs': 'n’ko', 'cy': 'N’Ko', 'da': 'n’ko', 'de': 'N’Ko', 'dsb': 'n’Koske pismo', 'el': 'Ν’Κο', 'el-polyton': 'Ν’Κο', 'en': 'N’Ko', 'en-Dsrt': '𐐤’𐐗𐐬', 'es': 'n’ko', 'eu': 'n’koa', 'fa': 'اِن کو', 'ff-Adlm': '𞤐𞤳𞤮𞥅', 'fi': 'n’ko', 'fil': 'N’Ko', 'fo': 'n’ko', 'fr': 'n’ko', 'fy': 'N’Ko', 'ga': 'N-cóis', 'gd': 'N’ko', 'gl': 'n’ko', 'gsw': 'N’Ko', 'gu': 'એન’ કો', 'ha': 'N’Ko', 'he': 'נ׳קו', 'hi': 'एन्‘को', 'hi-Latn': 'N’Ko', 'hr': 'n’ko pismo', 'hsb': 'n’Koske pismo', 'ht': 'n’ko', 'hu': 'N’ko', 'hy': 'նկո', 'id': 'N’Ko', 'ig': 'N’Ko', 'is': 'n-kó', 'it': 'n’ko', 'ja': 'ンコ文字', 'jv': 'N’Ko', 'ka': 'ნკო', 'kgp': 'nyko', 'kk': 'нко жазуы', 'kk-Cyrl': 'нко жазуы', 'km': 'នកូ', 'kn': 'ಎನ್\u200dಕೋ', 'ko': '응코 문자', 'kok': 'नको', 'kok-Deva': 'नको', 'ks': 'این کو', 'ks-Arab': 'این کو', 'ky': 'нко (жазуу)', 'lb': 'N’Ko', 'lo': 'ເອັນໂກ', 'lt': 'enko', 'lv': 'nko', 'mai': 'एनको', 'mi': 'Unukō', 'mk': 'нко', 'ml': 'എൻകോ', 'mn': 'Нко бичиг', 'mr': 'एन्‘को', 'ms': 'N’ko', 'my': 'အွန်ကို', 'nb': 'n’ko', 'ne': 'एन्को', 'nl': 'N’Ko', 'nn': 'n’ko', 'no': 'n’ko', 'nqo': 'ߒߞߏ', 'or': 'ଏନ୍ କୋ', 'pa': 'ਐਨ’ਕੋ', 'pa-Guru': 'ਐਨ’ਕੋ', 'pcm': 'N’Ko Lángwej', 'pl': 'n’ko', 'ps': 'ناکو', 'pt': 'n’ko', 'qu': 'N’Ko Simi', 'rm': 'n’ko', 'ro': 'n’ko', 'ru': 'нко', 'sc': 'n’ko', 'sd': 'نڪو', 'sd-Arab': 'نڪو', 'si': 'එන්‘කෝ', 'sk': 'bambarčina', 'sl': 'nkojski', 'so': 'N’Ko', 'sq': 'nko', 'sr': 'нко', 'sr-Cyrl': 'нко', 'sr-Latn': 'nko', 'sv': 'n-kå', 'sw': 'N’Ko', 'ta': 'என்‘கோ', 'te': 'న్కో', 'tg': 'Н’Ко', 'th': 'เอ็นโก', 'ti': 'ንኮ', 'tk': 'nko', 'to': 'tohinima fakanikō', 'tr': 'N’Ko', 'ug': 'نىكوچە', 'uk': 'нко', 'ur': 'نکو', 'uz': 'nko', 'uz-Latn': 'nko', 'vec': 'n’ko', 'vi': 'Chữ N’Ko', 'xnr': 'एन्‘को', 'yo': 'Èdè Nkoo', 'yrl': 'ĩku', 'yue': '西非書面語言 (N’Ko)', 'yue-Hans': '西非书面语言 (N’Ko)', 'yue-Hant': '西非書面語言 (N’Ko)', 'zh': '西非书面文字（N’Ko）', 'zh-Hans': '西非书面文字（N’Ko）', 'zh-Hant': '西非書面語言 (N’Ko)', 'zu': 'isi-N’Ko'},
    'Nshu': {'ast': 'nüshu', 'bal-Latn': 'Nushu', 'bs': 'nushu pismo', 'bs-Cyrl': 'нушу писмо', 'bs-Latn': 'nushu pismo', 'ca': 'nü shu', 'ca-ES-valencia': 'nü shu', 'cs': 'nü-šu', 'da': 'nüshu', 'de': 'Frauenschrift', 'en': 'Nüshu', 'et': 'nüšu', 'eu': 'nushuera', 'ff-Adlm': '𞤐𞤵𞥅𞤧𞤵', 'fi': 'nüshu', 'fo': 'nüshu', 'fy': 'Nüshu', 'ga': 'Nüshu', 'gd': 'Nüshu', 'he': 'נושו', 'hi-Latn': 'Nüshu', 'hr': 'nushu pismo', 'id': 'Nushu', 'ig': 'Nüshu', 'it': 'nushu', 'ja': '女書', 'ka': 'ნუშუ', 'kk': 'нюй-шу жазуы', 'kk-Cyrl': 'нюй-шу жазуы', 'ko': '누슈 문자', 'lo': 'ນຸຊຸ', 'lt': 'Nüshu', 'mk': 'нишу', 'mn': 'Нүшү', 'ms': 'Nushu', 'nb': 'nüshu', 'nl': 'Nüshu', 'nn': 'nüshu', 'no': 'nüshu', 'nqo': 'ߣߎ߳ߛߎ', 'ro': 'nüshu', 'ru': 'нюй-шу', 'sc': 'nüshu', 'so': 'Nüshu', 'sq': 'nyshuisht', 'sv': 'nüshu', 'tg': 'Нушу', 'th': 'นุซุ', 'to': 'tohinima fakanasiū', 'tr': 'Nüshu', 'ug': 'نۈشۇچە', 'vec': 'nushu', 'vi': 'Chữ Nüshu', 'yue': '女書文字', 'yue-Hans': '女书文字', 'yue-Hant': '女書文字', 'zh': '女书', 'zh-Hans': '女书', 'zh-Hant': '女書', 'zu': 'i-Nüshu'},
    'OM': {'af': 'Oman', 'agq': 'Umàn', 'ak': 'Oman', 'am': 'ኦማን', 'ar': 'عُمان', 'as': 'ওমান', 'asa': 'Omani', 'ast': 'Omán', 'az': 'Oman', 'az-Cyrl': 'Оман', 'az-Latn': 'Oman', 'bal-Latn': 'Omán', 'bas': 'Òmân', 'be': 'Аман', 'be-tarask': 'Аман', 'bez': 'Huomani', 'bg': 'Оман', 'bgn': 'ئومان', 'blo': 'Oman', 'bm': 'Omaŋ', 'bn': 'ওমান', 'br': 'Oman', 'brx': 'ओमान', 'bs': 'Oman', 'bs-Cyrl': 'Оман', 'bs-Latn': 'Oman', 'ca': 'Oman', 'ca-ES-valencia': 'Oman', 'ccp': '𑄃𑄮𑄟𑄚𑄴', 'ce': 'Ӏоман', 'ceb': 'Oman', 'cgg': 'Omaani', 'chr': 'ᎣᎺᏂ', 'ckb': 'عومان', 'cs': 'Omán', 'cv': 'Оман', 'cy': 'Oman', 'da': 'Oman', 'dav': 'Omani', 'de': 'Oman', 'dje': 'Omaan', 'dsb': 'Oman', 'dz': 'ཨོ་མཱན', 'ebu': 'Omani', 'ee': 'Oman nutome', 'el': 'Ομάν', 'el-polyton': 'Ὀμάν', 'en': 'Oman', 'en-Dsrt': '𐐄𐑋𐐲𐑌', 'eo': 'Omano', 'es': 'Omán', 'et': 'Omaan', 'eu': 'Oman', 'ewo': 'Omán', 'fa': 'عمان', 'ff': 'Omaan', 'ff-Adlm': '𞤌𞥅𞤥𞤢𞥄𞤲', 'ff-Latn': 'Omaan', 'fi': 'Oman', 'fil': 'Oman', 'fo': 'Oman', 'fr': 'Oman', 'fur': 'Oman', 'fy': 'Oman', 'ga': 'Óman', 'gd': 'Omàn', 'gl': 'Omán', 'gsw': 'Omaan', 'gu': 'ઓમાન', 'guz': 'Omani', 'ha': 'Oman', 'he': 'עומאן', 'hi': 'ओमान', 'hi-Latn': 'Oman', 'hr': 'Oman', 'hsb': 'Oman', 'ht': 'Oman', 'hu': 'Omán', 'hy': 'Օման', 'ia': 'Oman', 'id': 'Oman', 'ig': 'Oman', 'is': 'Óman', 'it': 'Oman', 'ja': 'オマーン', 'jmc': 'Omani', 'jv': 'Oman', 'ka': 'ომანი', 'kaa': 'Оман', 'kaa-Cyrl': 'Оман', 'kab': 'Ɛuman', 'kam': 'Omani', 'kde': 'Omani', 'kea': 'Oman', 'kgp': 'Omỹ', 'khq': 'Omaan', 'ki': 'Omani', 'kk': 'Оман', 'kk-Cyrl': 'Оман', 'kln': 'Emetab Oman', 'km': 'អូម៉ង់', 'kn': 'ಓಮನ್', 'ko': '오만', 'kok': 'ओमान', 'kok-Deva': 'ओमान', 'ks': 'اومان', 'ks-Arab': 'اومان', 'ksb': 'Omani', 'ksf': 'oman', 'ksh': 'Oman', 'ku': 'Oman', 'kxv': 'oman', 'kxv-Deva': 'ओमान', 'kxv-Latn': 'oman', 'kxv-Orya': 'ଅମାନ', 'kxv-Telu': 'ఓమన్', 'ky': 'Оман', 'lag': 'Ómani', 'lb': 'Oman', 'lg': 'Omaani', 'lld': 'Oman', 'ln': 'Ománɛ', 'lo': 'ໂອມານ', 'lt': 'Omanas', 'lu': 'Omane', 'luo': 'Oman', 'luy': 'Omani', 'lv': 'Omāna', 'mai': 'ओमान', 'mas': 'Omani', 'mer': 'Omani', 'mfe': 'Oman', 'mg': 'Oman', 'mgh': 'Uomani', 'mi': 'Ōmana', 'mk': 'Оман', 'ml': 'ഒമാൻ', 'mn': 'Оман', 'mr': 'ओमान', 'ms': 'Oman', 'mt': 'l-Oman', 'mua': 'Omaŋ', 'my': 'အိုမန်', 'mzn': 'عمان', 'naq': 'Oman', 'nb': 'Oman', 'nd': 'Omani', 'ne': 'ओमन', 'nl': 'Oman', 'nmg': 'Oman', 'nn': 'Oman', 'no': 'Oman', 'nqo': 'ߏߡߊ߲߫', 'nyn': 'Omaani', 'om': 'Omaan', 'or': 'ଓମାନ୍', 'pa': 'ਓਮਾਨ', 'pa-Guru': 'ਓਮਾਨ', 'pcm': 'Omán', 'pl': 'Oman', 'ps': 'عمان', 'pt': 'Omã', 'qu': 'Omán', 'rm': 'Oman', 'rn': 'Omani', 'ro': 'Oman', 'rof': 'Omani', 'ru': 'Оман', 'rwk': 'Omani', 'saq': 'Omani', 'sbp': 'Omani', 'sc': 'Omàn', 'sd': 'عمان', 'sd-Arab': 'عمان', 'se': 'Oman', 'seh': 'Omã', 'ses': 'Omaan', 'sg': 'Omâni', 'shi': 'ⵄⵓⵎⴰⵏ', 'shi-Latn': 'ɛuman', 'shi-Tfng': 'ⵄⵓⵎⴰⵏ', 'si': 'ඕමානය', 'sk': 'Omán', 'sl': 'Oman', 'smn': 'Oman', 'sn': 'Oman', 'so': 'Cumaan', 'sq': 'Oman', 'sr': 'Оман', 'sr-Cyrl': 'Оман', 'sr-Latn': 'Oman', 'sv': 'Oman', 'sw': 'Oman', 'syr': 'ܥܘܡܐܢ', 'ta': 'ஓமன்', 'te': 'ఓమన్', 'teo': 'Omani', 'tg': 'Умон', 'th': 'โอมาน', 'ti': 'ዖማን', 'tk': 'Oman', 'to': 'ʻOmani', 'tr': 'Umman', 'tt': 'Оман', 'twq': 'Omaan', 'tzm': 'Ɛumman', 'ug': 'ئومان', 'uk': 'Оман', 'ur': 'عمان', 'uz': 'Ummon', 'uz-Cyrl': 'Уммон', 'uz-Latn': 'Ummon', 'vai': 'ꕱꕮꘋ', 'vai-Latn': 'Omaŋ', 'vai-Vaii': 'ꕱꕮꘋ', 'vec': 'Oman', 'vi': 'Oman', 'vun': 'Omani', 'wae': 'Oman', 'wo': 'Omaan', 'xh': 'E-Oman', 'xnr': 'ओमान', 'xog': 'Omaani', 'yav': 'omáŋ', 'yo': 'Ọọma', 'yrl': 'Umã', 'yue': '阿曼王國', 'yue-Hans': '阿曼王国', 'yue-Hant': '阿曼王國', 'zgh': 'ⵄⵓⵎⴰⵏ', 'zh': '阿曼', 'zh-Hans': '阿曼', 'zh-Hant': '阿曼', 'zu': 'i-Oman'},
    'Ogam': {'am': 'ኦግሀም', 'ar': 'الأوجهام', 'ast': 'ogham', 'az': 'oğam', 'az-Latn': 'oğam', 'bal-Latn': 'Oghám', 'bg': 'Огамическа', 'bn': 'ওঘাম', 'brx': 'ओगहैम', 'bs': 'ogham pismo', 'bs-Cyrl': 'огамско писмо', 'bs-Latn': 'ogham pismo', 'ca': 'ogham', 'ca-ES-valencia': 'ogham', 'ccp': '𑄃𑄮𑄊𑄟𑄴', 'cs': 'ogamské', 'da': 'ogham', 'de': 'Ogham', 'el': 'Όγκχαμ', 'el-polyton': 'Όγκχαμ', 'en': 'Ogham', 'en-Dsrt': '𐐄𐐲𐑋', 'es': 'ogham', 'eu': 'oghamera', 'fa': 'اوگامی', 'ff-Adlm': '𞤌𞥅𞤺𞤢𞤥𞤵', 'fo': 'ogham', 'fy': 'Ogham', 'ga': 'Ogham', 'gd': 'Ogham-chraobh', 'gsw': 'Ogham', 'gu': 'ઓઘામ', 'he': 'אוהם', 'hi': 'ओगम', 'hi-Latn': 'Ogham', 'hr': 'ogham pismo', 'hu': 'Ogham', 'id': 'Ogham', 'ig': 'Ogham', 'it': 'ogham', 'ja': 'オガム文字', 'ka': 'ოღამი', 'kgp': 'ogỹmiku', 'kk': 'огам жазуы', 'kk-Cyrl': 'огам жазуы', 'kn': 'ಓಘಮ್', 'ko': '오검 문자', 'ks': 'اوگہام', 'ks-Arab': 'اوگہام', 'lb': 'Ogham', 'lo': 'ອອກຄອນ', 'lt': 'ogham', 'lv': 'ogamiskais raksts', 'mk': 'огам', 'ml': 'ഒഖാം', 'mn': 'Огам', 'mr': 'ओघाम', 'ms': 'Ogham', 'nb': 'ogham', 'ne': 'ओघाम', 'nl': 'Ogham', 'nn': 'ogham', 'no': 'ogham', 'nqo': 'ߏߜ߭ߊ߲', 'or': 'ଓଘାମା', 'pl': 'ogham', 'pt': 'ogâmico', 'rm': 'ogham', 'ro': 'ogham', 'ru': 'огамическая', 'sc': 'ogham', 'sl': 'ogamski', 'so': 'Ogham', 'sq': 'ogamisht', 'sr': 'огамско писмо', 'sr-Cyrl': 'огамско писмо', 'sr-Latn': 'ogamsko pismo', 'sv': 'ogham', 'ta': 'ஒகாம்', 'te': 'ఒఘమ్', 'tg': 'Огам', 'th': 'โอคัม', 'to': 'tohinima fakaʻokami', 'tr': 'Ogham', 'ug': 'ئوگەمچە', 'uk': 'огамічний', 'vi': 'Chữ Ogham', 'xnr': 'ओगम', 'yrl': 'ugãmiku', 'yue': '歐甘文', 'yue-Hans': '欧甘文', 'yue-Hant': '歐甘文', 'zh': '欧甘文', 'zh-Hans': '欧甘文', 'zh-Hant': '歐甘文', 'zu': 'i-Ogham'},
    'Olck': {'af': 'Ol Chiki', 'ak': 'Ol Kyiki kasa', 'am': 'ኦይ ቺኪ', 'ar': 'أول تشيكي', 'as': 'অ’ল চিকি', 'ast': 'ol chiki', 'az': 'ol çiki', 'az-Latn': 'ol çiki', 'bal-Latn': 'Ol cheki', 'be': 'ол-чыкі', 'be-tarask': 'ол-чыкі', 'bg': 'Ол Чики', 'bn': 'ওল চিকি', 'brx': 'अल चीकी', 'bs': 'ol čiki pismo', 'bs-Cyrl': 'ол чики писмо', 'bs-Latn': 'ol čiki pismo', 'ca': 'santali', 'ca-ES-valencia': 'santali', 'ccp': '𑄃𑄮𑄣𑄴𑄌𑄨𑄇𑄨', 'chr': 'ᎣᎵ ᏥᎩ', 'cs': 'santálské (ol chiki)', 'cy': 'Ol Chiki', 'da': 'ol-chiki', 'de': 'Ol Chiki', 'dsb': 'ol-chikiske pismo', 'el': 'Ολ Τσίκι', 'el-polyton': 'Ολ Τσίκι', 'en': 'Ol Chiki', 'en-Dsrt': '𐐄𐑊 𐐕𐐨𐐿𐐨', 'es': 'ol chiki', 'et': 'santali', 'eu': 'ol chikia', 'fa': 'اول چیکی', 'ff-Adlm': '𞤌𞤭-𞤕𞤭𞤳𞤭', 'fi': 'ol chiki', 'fil': 'Ol Chiki', 'fo': 'ol chiki', 'fr': 'ol-chiki', 'fy': 'Ol Chiki', 'ga': 'Ol Chiki', 'gd': 'Ol Chiki', 'gl': 'ol chiki', 'gsw': 'Ol Chiki', 'gu': 'ઓલ ચિકી', 'ha': 'Ol Chiki', 'he': 'אול צ׳יקי', 'hi': 'ओल चिकी', 'hi-Latn': 'Ol Chiki', 'hr': 'ol chiki pismo', 'hsb': 'ol-chikiske pismo', 'ht': 'ol-chiki', 'hu': 'Ol chiki', 'hy': 'օլ չիկի', 'id': 'Chiki Lama', 'ig': 'Ol Chiki', 'is': 'ol chiki', 'it': 'ol chiki', 'ja': 'オルチキ文字', 'jv': 'Ol Chiki', 'ka': 'ოლ-ჩიკი', 'kgp': 'or siki', 'kk': 'ол-чики жазуы', 'kk-Cyrl': 'ол-чики жазуы', 'km': 'អូលឈិគិ', 'kn': 'ಓಲ್ ಚಿಕಿ', 'ko': '올 치키 문자', 'kok': 'ओल चिकी', 'kok-Deva': 'ओल चिकी', 'ks': 'اول چِکی', 'ks-Arab': 'اول چِکی', 'ky': 'Ол-чики (жазуу)', 'lb': 'Ol Chiki', 'lo': 'ໂອຊິກິ', 'lt': 'ol čiki', 'lv': 'olčiki', 'mai': 'ओल चिकी', 'mi': 'Ōtiki', 'mk': 'ол чики', 'ml': 'ഒൽ ചിക്കി', 'mn': 'Ол чики бичиг', 'mr': 'ओल चिकि', 'ms': 'Ol Chiki', 'my': 'အိုလ်ချီကီ', 'nb': 'ol-chiki', 'ne': 'ओलचिकी', 'nl': 'Ol Chiki', 'nn': 'ol-chiki', 'no': 'ol-chiki', 'nqo': 'ߏߟߑߗߞߌ', 'or': 'ଅଲ୍ ଚିକି', 'pa': 'ਓਲ ਚੀਕੀ', 'pa-Guru': 'ਓਲ ਚੀਕੀ', 'pcm': 'Ol Chíkí', 'pl': 'ol ciki', 'ps': 'اول چکی۔', 'pt': 'ol chiki', 'qu': 'Ol Chiki Simi', 'rm': 'ol chiki', 'ro': 'ol chiki', 'ru': 'ол чики', 'sat': 'ᱚᱞ ᱪᱤᱠᱤ', 'sat-Deva': 'अल चीकी', 'sat-Olck': 'ᱚᱞ ᱪᱤᱠᱤ', 'sc': 'ol chiki', 'sd': 'اول چڪي', 'sd-Arab': 'اول چڪي', 'si': 'ඔල් චිකි', 'sk': 'santálske (ol chiki)', 'sl': 'santalski', 'so': 'Ol Jiki', 'sq': 'ol çiki', 'sr': 'ол чики', 'sr-Cyrl': 'ол чики', 'sr-Latn': 'ol čiki', 'sv': 'ol-chiki', 'sw': 'Kiol Chiki', 'ta': 'ஒல் சிக்கி', 'te': 'ఓల్ చికి', 'tg': 'Ол Чикӣ', 'th': 'โอลชิกิ', 'ti': 'ኦል ቺኪ', 'tk': 'ol-çiki', 'to': 'tohinima fakaʻolisiki', 'tr': 'Ol Chiki', 'ug': 'ئول-چىكىچە', 'uk': 'ол-чикі', 'ur': 'اول چکی', 'uz': 'ol chiki', 'uz-Latn': 'ol chiki', 'vec': 'ol ciki', 'vi': 'Chữ Ol Chiki', 'xnr': 'ऑल चिकी', 'yo': 'Èdè Ol Siki', 'yrl': 'uxiki', 'yue': '桑塔利文', 'yue-Hans': '桑塔利文', 'yue-Hant': '桑塔利文', 'zh': '桑塔利文', 'zh-Hans': '桑塔利文', 'zh-Hant': '桑塔利文', 'zu': 'isi-Ol Chiki'},
    'Onao': {'bal-Latn': 'Ol Onal', 'en': 'Ol Onal', 'fo': 'ol onal', 'gd': 'Ol Onal', 'hi-Latn': 'Ol Onal', 'ig': 'Ol Onal', 'vec': 'ol onal'},
    'Orkh': {'ar': 'الأورخون', 'ast': 'orkhon', 'az': 'orxon', 'az-Latn': 'orxon', 'bal-Latn': 'Orkhon', 'bg': 'Орхоно-енисейска', 'bn': 'অর্খোন', 'brx': 'ओरखोन', 'bs': 'orkhon pismo', 'bs-Cyrl': 'орконско писмо', 'bs-Latn': 'orkhon pismo', 'ca': 'orkhon', 'ca-ES-valencia': 'orkhon', 'ccp': '𑄃𑄧𑄢𑄴𑄈𑄮𑄚𑄴', 'cs': 'orchonské', 'da': 'orkhon', 'de': 'Orchon-Runen', 'el': 'Όρκχον', 'el-polyton': 'Όρκχον', 'en': 'Orkhon', 'en-Dsrt': '𐐄𐑉𐐿𐐱𐑌', 'es': 'orkhon', 'et': 'Orhoni', 'eu': 'orkhonera', 'fa': 'اورخونی', 'ff-Adlm': '𞤌𞤪𞤳𞤮𞥅𞤲', 'fi': 'orkhon', 'fo': 'orkhon', 'fr': 'orkhon', 'fy': 'Orkhon', 'ga': 'Orkhon', 'gd': 'Orkhon', 'gsw': 'Orchon-Runä', 'gu': 'ઓરખોન', 'he': 'אורחון', 'hi': 'ओरखोन', 'hi-Latn': 'Orkhon', 'hr': 'orkhon pismo', 'ht': 'orkhon', 'hu': 'Orhon', 'id': 'Orkhon', 'ig': 'Orkhon', 'it': 'orkhon', 'ja': 'オルホン文字', 'ka': 'ორხონული', 'kgp': 'orkihãn', 'kk': 'көне түркі жазбалары', 'kk-Cyrl': 'көне түркі жазбалары', 'kn': 'ಓರ್ಖೋನ್', 'ko': '오르혼어', 'ks': 'اورکھون', 'ks-Arab': 'اورکھون', 'lb': 'Orchon-Runen', 'lo': 'ອອກສມັນຍາ', 'lt': 'orkon', 'mk': 'старотурско', 'ml': 'ഒർഖോൺ', 'mn': 'Орхон', 'mr': 'ओर्खोन', 'ms': 'Orkhon', 'nb': 'orkhon', 'ne': 'ओर्खोन', 'nl': 'Orkhon', 'nn': 'orkhon', 'no': 'orkhon', 'nqo': 'ߐߙߑߞߐ߲', 'or': 'ଓରୋଖନ୍', 'pl': 'orchońskie', 'pt': 'orkhon', 'rm': 'orkhon', 'ro': 'orhon', 'ru': 'орхоно-енисейская', 'sc': 'orkhon', 'sl': 'orkonski', 'so': 'Orkhon', 'sq': 'orkonisht', 'sr': 'орконско писмо', 'sr-Cyrl': 'орконско писмо', 'sr-Latn': 'orkonsko pismo', 'sv': 'orkon', 'ta': 'ஆர்கான்', 'te': 'ఓర్ఖోన్', 'tg': 'Оркон', 'th': 'ออร์คอน', 'to': 'tohinima fakaʻolikoni', 'tr': 'Orhun', 'ug': 'ئورخۇنچە', 'uk': 'орхонський', 'vec': 'orcon', 'vi': 'Chữ Orkhon', 'xnr': 'ओरखोन', 'yrl': 'urikũ', 'yue': '鄂爾渾文', 'yue-Hans': '鄂尔浑文', 'yue-Hant': '鄂爾渾文', 'zh': '鄂尔浑文', 'zh-Hans': '鄂尔浑文', 'zh-Hant': '鄂爾渾文', 'zu': 'i-Orkhon'},
    'Orya': {'af': 'Oriya', 'ak': 'Odia kasa', 'am': 'ኦዲያ', 'ar': 'الأوريا', 'as': 'ওড়িয়া', 'ast': 'oriya', 'az': 'oriya', 'az-Latn': 'oriya', 'bal-Latn': 'Ódiá', 'be': 'орыя', 'be-tarask': 'орыя', 'bg': 'ория', 'bn': 'ওড়িয়া', 'br': 'oriya', 'brx': 'उड़िया', 'bs': 'pismo orija', 'bs-Cyrl': 'одија писмо', 'bs-Latn': 'pismo orija', 'ca': 'oriya', 'ca-ES-valencia': 'oriya', 'ccp': '𑄃𑄮𑄢𑄨𑄠', 'ce': 'ори', 'chr': 'ᎣᏗᎠ', 'ckb': 'ئۆدیا', 'cs': 'urijské', 'cy': 'Orïaidd', 'da': 'oriya', 'de': 'Oriya', 'dsb': 'oriya', 'dz': 'ཨོ་རི་ཡ་ཡིག་གུ', 'ee': 'oriyagbeŋɔŋlɔ', 'el': 'Όντια', 'el-polyton': 'Όντια', 'en': 'Odia', 'en-Dsrt': '𐐉𐑉𐐨𐐲', 'es': 'oriya', 'et': 'oria', 'eu': 'oriyarra', 'fa': 'اودیه', 'ff-Adlm': '𞤌𞤪𞤭𞤴𞤢', 'fi': 'orijalainen', 'fil': 'Odia', 'fo': 'odia', 'fr': 'odia', 'fur': 'oriya', 'fy': 'Odia', 'ga': 'Oiríseach', 'gd': 'Oriya', 'gl': 'odiá', 'gsw': 'Oriya', 'gu': 'ઉડિયા', 'ha': 'Yaren Odia', 'he': 'אודייה', 'hi': 'ओड़िया', 'hi-Latn': 'Odia', 'hr': 'orijsko pismo', 'hsb': 'oriya', 'ht': 'odia', 'hu': 'Oriya', 'hy': 'օրիյա', 'id': 'Oriya', 'ig': 'Odia', 'is': 'oriya', 'it': 'oriya', 'ja': 'オディア文字', 'jv': 'Odia', 'ka': 'ორია', 'kea': 'oriya', 'kgp': 'orija', 'kk': 'ория жазуы', 'kk-Cyrl': 'ория жазуы', 'km': 'អូឌៀ', 'kn': 'ಒಡಿಯಾ', 'ko': '오리야 문자', 'kok': 'ओडिया', 'kok-Deva': 'ओडिया', 'ks': 'اورِیا', 'ks-Arab': 'اورِیا', 'ksh': 'indesche Orija-Schreff', 'kxv': 'oḍiaa', 'kxv-Deva': 'ऑड़िया', 'kxv-Latn': 'oḍiaa', 'kxv-Orya': 'ଅଡ଼ିଆ', 'kxv-Telu': 'ఒడియా', 'ky': 'Орийа', 'lb': 'Oriya', 'lo': 'ໂອເດຍ', 'lrc': 'ئوریا', 'lt': 'orijų', 'lv': 'oriju', 'mai': 'उड़िया', 'mi': 'Otia', 'mk': 'оријанско писмо', 'ml': 'ഒഡിയ', 'mn': 'ория', 'mr': 'उडिया', 'ms': 'Oriya', 'my': 'အိုရာ', 'nb': 'odia', 'ne': 'ओडिया', 'nl': 'Odia', 'nn': 'odia', 'no': 'odia', 'nqo': 'ߏߖߊ', 'or': 'ଓଡ଼ିଆ', 'pa': 'ਉੜੀਆ', 'pa-Guru': 'ਉੜੀਆ', 'pcm': 'Ódia', 'pl': 'orija', 'ps': 'اوديا', 'pt': 'oriá', 'qu': 'Odia Simi', 'rm': 'oriya', 'ro': 'oriya', 'ru': 'ория', 'sc': 'odia', 'sd': 'اوڊيا', 'sd-Arab': 'اوڊيا', 'si': 'ඔරියා', 'sk': 'uríjske', 'sl': 'orijski', 'so': 'Oodhiya', 'sq': 'orija', 'sr': 'оријанско писмо', 'sr-Cyrl': 'оријанско писмо', 'sr-Latn': 'orijansko pismo', 'sv': 'oriya', 'sw': 'Kioriya', 'ta': 'ஒடியா', 'te': 'ఒడియా', 'tg': 'Одия', 'th': 'โอริยา', 'ti': 'ኦዲያ', 'tk': 'Oriýa elipbiýi', 'to': 'tohinima fakaʻotia', 'tr': 'Oriya', 'ug': 'ئورىياچە', 'uk': 'орія', 'ur': 'اڑیہ', 'uz': 'oriya', 'uz-Cyrl': 'Ория', 'uz-Latn': 'oriya', 'vec': 'orija', 'vi': 'Chữ Odia', 'wae': 'Oriya', 'xnr': 'उड़िया', 'yo': 'Òdíà', 'yrl': 'uriá', 'yue': '歐利亞文', 'yue-Hans': '欧利亚文', 'yue-Hant': '歐利亞文', 'zh': '奥里亚文', 'zh-Hans': '奥里亚文', 'zh-Hant': '歐迪亞文', 'zu': 'isi-Odia'},
    'Osge': {'bal-Latn': 'Oséj', 'bs': 'osage pismo', 'bs-Cyrl': 'осаге писмо', 'bs-Latn': 'osage pismo', 'ca': 'osage', 'ca-ES-valencia': 'osage', 'en': 'Osage', 'et': 'oseidži', 'eu': 'osagera', 'ff-Adlm': '𞤌𞤧𞤢𞥄𞤶', 'fi': 'osagen aakkosto', 'fo': 'osage', 'ga': 'Ósáis', 'gd': 'Osage', 'he': 'אוסייג׳', 'hi-Latn': 'Osage', 'id': 'Osage', 'ig': 'Osage', 'kk': 'осейдж жазуы', 'kk-Cyrl': 'осейдж жазуы', 'km': 'អូស្គ', 'mn': 'Осге', 'ms': 'Osage', 'nl': 'Osage', 'nqo': 'ߏߛߊߖ߭', 'ro': 'osage', 'sc': 'osage', 'so': 'Osage', 'sq': 'osage', 'sv': 'osage', 'tg': 'Осейҷӣ', 'uk': 'осейджиська', 'vec': 'osage', 'yo': 'Èdè Osage', 'zh': '欧塞奇文', 'zh-Hans': '欧塞奇文', 'zh-Hant': '歐塞奇文', 'zu': 'isi-Osage'},
    'Osma': {'am': 'ኦስማኒያ', 'ar': 'الأوسمانيا', 'ast': 'osmanya', 'az': 'osmanya', 'az-Latn': 'osmanya', 'bal-Latn': 'Osmánia', 'bg': 'Османска', 'bn': 'ওসমানিয়', 'brx': 'ओस्मानिया', 'bs': 'osmanja pismo', 'bs-Cyrl': 'осмањанско писмо', 'bs-Latn': 'osmanja pismo', 'ca': 'osmanya', 'ca-ES-valencia': 'osmanya', 'ccp': '𑄃𑄮𑄥𑄟𑄚𑄨𑄠𑄧', 'cs': 'osmanské', 'da': 'osmannisk', 'de': 'Osmanisch', 'el': 'Οσμάνγια', 'el-polyton': 'Οσμάνγια', 'en': 'Osmanya', 'en-Dsrt': '𐐉𐑅𐑋𐐪𐑌𐐷𐐪', 'es': 'osmaniya', 'et': 'osmani', 'eu': 'osmaiera', 'ff-Adlm': '𞤌𞤧𞤥𞤢𞤲𞤭𞤴𞤢', 'fi': 'osmanjalainen', 'fo': 'osmanya', 'fr': 'osmanais', 'fy': 'Osmanya', 'ga': 'Osmanya', 'gd': 'Osmanya', 'gsw': 'Osmanisch', 'gu': 'ઓસ્માન્યા', 'he': 'אוסמניה', 'hi': 'ओस्मान्या', 'hi-Latn': 'Osmanya', 'hr': 'osmanya pismo', 'ht': 'osmanais', 'hu': 'Oszmán', 'id': 'Osmanya', 'ig': 'Osmanya', 'it': 'osmanya', 'ja': 'オスマニア文字', 'ka': 'ოსმანია', 'kgp': 'ojmỹnja', 'kk': 'исмания жазуы', 'kk-Cyrl': 'исмания жазуы', 'kn': 'ಓಸ್ಮಾನ್ಯಾ', 'ko': '오스마니아 문자', 'ks': 'اوسمانیا', 'ks-Arab': 'اوسمانیا', 'lb': 'Osmanesch', 'lt': 'osmanų', 'lv': 'osmaņu turku', 'mk': 'сомалиско', 'ml': 'ഒസ്\u200cമാനിയ', 'mn': 'Османи', 'mr': 'उस्मानिया', 'ms': 'Osmanya', 'nb': 'osmanya', 'ne': 'ओस्मान्या', 'nl': 'Osmanya', 'nn': 'osmanya', 'no': 'osmanya', 'nqo': 'ߎߛߑߡߊߣߌ', 'or': 'ଓସୋମାନିୟା', 'pl': 'osmanya', 'pt': 'osmania', 'rm': 'osman', 'ro': 'osmanya', 'ru': 'османская', 'sc': 'osmanya', 'sk': 'osmanský', 'sl': 'osmanski', 'so': 'Osmanya', 'sq': 'osmaniaisht', 'sr': 'осмањанско писмо', 'sr-Cyrl': 'осмањанско писмо', 'sr-Latn': 'osmanjansko pismo', 'sv': 'osmanja', 'ta': 'ஒஸ்மான்யா', 'te': 'ఓసమాన్య', 'tg': 'Османияӣ', 'th': 'ออสมันยา', 'to': 'tohinima fakaʻosimānia', 'tr': 'Osmanya', 'ug': 'ئوسمانيە', 'uk': 'османський', 'vec': 'ozmania', 'vi': 'Chữ Osmanya', 'xnr': 'ओस्मान्या', 'yrl': 'usmania', 'yue': '歐斯曼亞文', 'yue-Hans': '欧斯曼亚文', 'yue-Hant': '歐斯曼亞文', 'zh': '奥斯曼亚文', 'zh-Hans': '奥斯曼亚文', 'zh-Hant': '歐斯曼亞文', 'zu': 'i-Osmanya'},
    'Ougr': {'bal-Latn': 'Ahdi Yógher', 'bs': 'starougursko pismo', 'bs-Cyrl': 'старо ујгур писмо', 'bs-Latn': 'starougursko pismo', 'en': 'Old Uyghur', 'et': 'vanauiguuri', 'eu': 'uigurrera zaharra', 'ff-Adlm': '𞤏𞤭𞤺𞤵𞥅𞤪 𞤑𞤭𞤯𞥆𞤵𞤲', 'fi': 'vanha uiguurilainen', 'fo': 'gamalt uyghur', 'gd': 'Seann-Ùigiurais', 'he': 'אוגורית ישנה', 'hi-Latn': 'Old Uyghur', 'id': 'Uyghur Lama', 'ig': 'Old Uyghur', 'kk': 'көне ұйғыр жазуы', 'kk-Cyrl': 'көне ұйғыр жазуы', 'mn': 'Хуучин уйгар', 'nl': 'Oud Oeigoers', 'nqo': 'ߎߜ߭ߎߙ ߞߘߐ߬ߡߊ߲', 'ro': 'uigură veche', 'sc': 'uiguru antigu', 'sq': 'ujgurishte e vjetër', 'tg': 'Уйғури Куҳна', 'vec': 'uigur antigo', 'zh': '回鹘文', 'zh-Hans': '回鹘文'},
    'PA': {'af': 'Panama', 'agq': 'Kpanàma', 'ak': 'Panama', 'am': 'ፓናማ', 'ar': 'بنما', 'as': 'পানামা', 'asa': 'Panama', 'ast': 'Panamá', 'az': 'Panama', 'az-Cyrl': 'Панама', 'az-Latn': 'Panama', 'bal-Latn': 'Pánámá', 'bas': 'Pànàma', 'be': 'Панама', 'be-tarask': 'Панама', 'bez': 'Hupanama', 'bg': 'Панама', 'bgn': 'پانامه', 'blo': 'Panamaa', 'bm': 'Panama', 'bn': 'পানামা', 'br': 'Panamá', 'brx': 'पानामा', 'bs': 'Panama', 'bs-Cyrl': 'Панама', 'bs-Latn': 'Panama', 'ca': 'Panamà', 'ca-ES-valencia': 'Panamà', 'ccp': '𑄛𑄚𑄟', 'ce': 'Панама', 'ceb': 'Panama', 'cgg': 'Panama', 'chr': 'ᏆᎾᎹ', 'ckb': 'پاناما', 'cs': 'Panama', 'cv': 'Панама', 'cy': 'Panama', 'da': 'Panama', 'dav': 'Panama', 'de': 'Panama', 'dje': 'Panama', 'dsb': 'Panama', 'dz': 'པ་ན་མ', 'ebu': 'Panama', 'ee': 'Panama nutome', 'el': 'Παναμάς', 'el-polyton': 'Παναμάς', 'en': 'Panama', 'en-Dsrt': '𐐑𐐰𐑌𐐲𐑋𐐪', 'eo': 'Panamo', 'es': 'Panamá', 'et': 'Panama', 'eu': 'Panama', 'ewo': 'Panamá', 'fa': 'پاناما', 'ff': 'Panamaa', 'ff-Adlm': '𞤆𞤢𞤲𞤢𞤥𞤢', 'ff-Latn': 'Panamaa', 'fi': 'Panama', 'fil': 'Panama', 'fo': 'Panama', 'fr': 'Panama', 'fur': 'Panamà', 'fy': 'Panama', 'ga': 'Panama', 'gd': 'Panama', 'gl': 'Panamá', 'gsw': 'Panama', 'gu': 'પનામા', 'guz': 'Panama', 'ha': 'Panama', 'he': 'פנמה', 'hi': 'पनामा', 'hi-Latn': 'Panama', 'hr': 'Panama', 'hsb': 'Panama', 'ht': 'Panama', 'hu': 'Panama', 'hy': 'Պանամա', 'ia': 'Panama', 'id': 'Panama', 'ig': 'Panama', 'is': 'Panama', 'it': 'Panama', 'ja': 'パナマ', 'jmc': 'Panama', 'jv': 'Panama', 'ka': 'პანამა', 'kaa': 'Панама', 'kaa-Cyrl': 'Панама', 'kab': 'Panam', 'kam': 'Panama', 'kde': 'Panama', 'kea': 'Panamá', 'kgp': 'Panỹmỹ', 'khq': 'Panama', 'ki': 'Panama', 'kk': 'Панама', 'kk-Cyrl': 'Панама', 'kln': 'Emetab Panama', 'km': 'ប៉ាណាម៉ា', 'kn': 'ಪನಾಮಾ', 'ko': '파나마', 'kok': 'पनामा', 'kok-Deva': 'पनामा', 'ks': 'پَناما', 'ks-Arab': 'پَناما', 'ksb': 'Panama', 'ksf': 'panama', 'ksh': 'Pannama', 'ku': 'Panama', 'kxv': 'panema', 'kxv-Deva': 'पनामा', 'kxv-Latn': 'panema', 'kxv-Orya': 'ପନମା', 'kxv-Telu': 'పనామా', 'ky': 'Панама', 'lag': 'Panáama', 'lb': 'Panama', 'lg': 'Panama', 'lld': 'Panama', 'ln': 'Panama', 'lo': 'ພານາມາ', 'lt': 'Panama', 'lu': 'Panama', 'luo': 'Panama', 'luy': 'Panama', 'lv': 'Panama', 'mai': 'पनामा', 'mas': 'Panama', 'mer': 'Panama', 'mfe': 'Panama', 'mg': 'Panama', 'mgh': 'Upanama', 'mi': 'Panama', 'mk': 'Панама', 'ml': 'പനാമ', 'mn': 'Панам', 'mr': 'पनामा', 'ms': 'Panama', 'mt': 'il-Panama', 'mua': 'Sǝr Panama', 'my': 'ပနားမား', 'mzn': 'پاناما', 'naq': 'Panama', 'nb': 'Panama', 'nd': 'Panama', 'ne': 'प्यानामा', 'nl': 'Panama', 'nmg': 'Panama', 'nn': 'Panama', 'no': 'Panama', 'nqo': 'ߔߣߊߡߊ߫', 'nyn': 'Panama', 'om': 'Paanamaa', 'or': 'ପାନାମା', 'pa': 'ਪਨਾਮਾ', 'pa-Guru': 'ਪਨਾਮਾ', 'pcm': 'Pánáma', 'pl': 'Panama', 'ps': 'پاناما', 'pt': 'Panamá', 'qu': 'Panamá', 'rm': 'Panama', 'rn': 'Panama', 'ro': 'Panama', 'rof': 'Panama', 'ru': 'Панама', 'rwk': 'Panama', 'saq': 'Panama', 'sbp': 'Panama', 'sc': 'Pànama', 'sd': 'پناما', 'sd-Arab': 'پناما', 'se': 'Panama', 'seh': 'Panamá', 'ses': 'Panama', 'sg': 'Panama', 'shi': 'ⴱⴰⵏⴰⵎⴰ', 'shi-Latn': 'banama', 'shi-Tfng': 'ⴱⴰⵏⴰⵎⴰ', 'si': 'පැනමාව', 'sk': 'Panama', 'sl': 'Panama', 'smn': 'Panama', 'sn': 'Panama', 'so': 'Baanama', 'sq': 'Panama', 'sr': 'Панама', 'sr-Cyrl': 'Панама', 'sr-Latn': 'Panama', 'sv': 'Panama', 'sw': 'Panama', 'syr': 'ܦܢܡܐ', 'ta': 'பனாமா', 'te': 'పనామా', 'teo': 'Panama', 'tg': 'Панама', 'th': 'ปานามา', 'ti': 'ፓናማ', 'tk': 'Panama', 'to': 'Panamā', 'tr': 'Panama', 'tt': 'Панама', 'twq': 'Panama', 'tzm': 'Panama', 'ug': 'پاناما', 'uk': 'Панама', 'ur': 'پانامہ', 'uz': 'Panama', 'uz-Cyrl': 'Панама', 'uz-Latn': 'Panama', 'vai': 'ꕐꕯꕮ', 'vai-Latn': 'Panama', 'vai-Vaii': 'ꕐꕯꕮ', 'vec': 'Pànama', 'vi': 'Panama', 'vun': 'Panama', 'wae': 'Panama', 'wo': 'Panama', 'xh': 'EPanama', 'xnr': 'पनामा', 'xog': 'Panama', 'yav': 'panamá', 'yi': 'פּאַנאַמאַ', 'yo': 'Paramá', 'yrl': 'Panamã', 'yue': '巴拿馬', 'yue-Hans': '巴拿马', 'yue-Hant': '巴拿馬', 'zgh': 'ⴱⴰⵏⴰⵎⴰ', 'zh': '巴拿马', 'zh-Hans': '巴拿马', 'zh-Hant': '巴拿馬', 'zu': 'i-Panama'},
    'PE': {'af': 'Peru', 'agq': 'Kpɛlû', 'ak': 'Peru', 'am': 'ፔሩ', 'ar': 'بيرو', 'as': 'পেৰু', 'asa': 'Peru', 'ast': 'Perú', 'az': 'Peru', 'az-Cyrl': 'Перу', 'az-Latn': 'Peru', 'bal-Latn': 'Péru', 'bas': 'Pèrû', 'be': 'Перу', 'be-tarask': 'Перу', 'bez': 'Hupelu', 'bg': 'Перу', 'bgn': 'پیرو', 'blo': 'Peruu', 'bm': 'Peru', 'bn': 'পেরু', 'br': 'Perou', 'brx': 'पेरू', 'bs': 'Peru', 'bs-Cyrl': 'Перу', 'bs-Latn': 'Peru', 'ca': 'Perú', 'ca-ES-valencia': 'Perú', 'ccp': '𑄛𑄬𑄢𑄪', 'ce': 'Перу', 'ceb': 'Peru', 'cgg': 'Peru', 'chr': 'ᏇᎷ', 'ckb': 'پێروو', 'cs': 'Peru', 'cv': 'Перу', 'cy': 'Periw', 'da': 'Peru', 'dav': 'Peru', 'de': 'Peru', 'dje': 'Peeru', 'dsb': 'Peru', 'dz': 'པེ་རུ', 'ebu': 'Peru', 'ee': 'Peru nutome', 'el': 'Περού', 'el-polyton': 'Περού', 'en': 'Peru', 'en-Dsrt': '𐐑𐐲𐑉𐐭', 'eo': 'Peruo', 'es': 'Perú', 'et': 'Peruu', 'eu': 'Peru', 'ewo': 'Perú', 'fa': 'پرو', 'ff': 'Peru', 'ff-Adlm': '𞤆𞤫𞤪𞤵𞥅', 'ff-Latn': 'Peru', 'fi': 'Peru', 'fil': 'Peru', 'fo': 'Peru', 'fr': 'Pérou', 'fur': 'Perù', 'fy': 'Peru', 'ga': 'Peiriú', 'gd': 'Pearù', 'gl': 'Perú', 'gsw': 'Peru', 'gu': 'પેરુ', 'guz': 'Peru', 'ha': 'Feru', 'he': 'פרו', 'hi': 'पेरू', 'hi-Latn': 'Peru', 'hr': 'Peru', 'hsb': 'Peru', 'ht': 'Pérou', 'hu': 'Peru', 'hy': 'Պերու', 'ia': 'Peru', 'id': 'Peru', 'ie': 'Perú', 'ig': 'Peru', 'is': 'Perú', 'it': 'Perù', 'ja': 'ペルー', 'jgo': 'Pɛlû', 'jmc': 'Peru', 'jv': 'Peru', 'ka': 'პერუ', 'kaa': 'Перу', 'kaa-Cyrl': 'Перу', 'kab': 'Piru', 'kam': 'Peru', 'kde': 'Pelu', 'kea': 'Peru', 'kgp': 'Piru', 'khq': 'Peeru', 'ki': 'Peru', 'kk': 'Перу', 'kk-Cyrl': 'Перу', 'kln': 'Emetab Peru', 'km': 'ប៉េរូ', 'kn': 'ಪೆರು', 'ko': '페루', 'kok': 'पेरू', 'kok-Deva': 'पेरू', 'ks': 'پیٖروٗ', 'ks-Arab': 'پیٖروٗ', 'ksb': 'Pelu', 'ksf': 'pɛrú', 'ksh': 'Perruh', 'ku': 'Perû', 'kxv': 'peru', 'kxv-Deva': 'पेरू', 'kxv-Latn': 'peru', 'kxv-Orya': 'ପେରୁ', 'kxv-Telu': 'పెరూ', 'ky': 'Перу', 'lag': 'Péeru', 'lb': 'Peru', 'lg': 'Peru', 'lld': 'Perú', 'ln': 'Péru', 'lo': 'ເປຣູ', 'lt': 'Peru', 'lu': 'Peru', 'luo': 'Peru', 'luy': 'Peru', 'lv': 'Peru', 'mai': 'पेरू', 'mas': 'Peru', 'mer': 'Peru', 'mfe': 'Perou', 'mg': 'Peroa', 'mgh': 'Uperuu', 'mi': 'Peru', 'mk': 'Перу', 'ml': 'പെറു', 'mn': 'Перу', 'mr': 'पेरू', 'ms': 'Peru', 'mt': 'il-Perù', 'mua': 'Peru', 'my': 'ပီရူး', 'mzn': 'پرو', 'naq': 'Perub', 'nb': 'Peru', 'nd': 'Pheru', 'ne': 'पेरू', 'nl': 'Peru', 'nmg': 'Peru', 'nn': 'Peru', 'no': 'Peru', 'nqo': 'ߔߋߙߎ߫', 'nyn': 'Peru', 'om': 'Peeruu', 'or': 'ପେରୁ', 'pa': 'ਪੇਰੂ', 'pa-Guru': 'ਪੇਰੂ', 'pcm': 'Pẹ́ru', 'pl': 'Peru', 'ps': 'پیرو', 'pt': 'Peru', 'qu': 'Perú', 'rm': 'Peru', 'rn': 'Peru', 'ro': 'Peru', 'rof': 'Peru', 'ru': 'Перу', 'rwk': 'Peru', 'saq': 'Peru', 'sbp': 'Pelu', 'sc': 'Perù', 'sd': 'پيرو', 'sd-Arab': 'پيرو', 'se': 'Peru', 'seh': 'Peru', 'ses': 'Peeru', 'sg': 'Perüu', 'shi': 'ⴱⵉⵔⵓ', 'shi-Latn': 'biru', 'shi-Tfng': 'ⴱⵉⵔⵓ', 'si': 'පේරු', 'sk': 'Peru', 'sl': 'Peru', 'smn': 'Peru', 'sn': 'Peru', 'so': 'Beeru', 'sq': 'Peru', 'sr': 'Перу', 'sr-Cyrl': 'Перу', 'sr-Latn': 'Peru', 'sv': 'Peru', 'sw': 'Peru', 'syr': 'ܦܝܪܘ', 'ta': 'பெரு', 'te': 'పెరూ', 'teo': 'Peru', 'tg': 'Перу', 'th': 'เปรู', 'ti': 'ፔሩ', 'tk': 'Peru', 'to': 'Pelū', 'tr': 'Peru', 'tt': 'Перу', 'twq': 'Peeru', 'tzm': 'Piru', 'ug': 'پېرۇ', 'uk': 'Перу', 'ur': 'پیرو', 'uz': 'Peru', 'uz-Cyrl': 'Перу', 'uz-Latn': 'Peru', 'vai': 'ꗨꗡꖩ', 'vai-Latn': 'Pɛru', 'vai-Vaii': 'ꗨꗡꖩ', 'vec': 'Perù', 'vi': 'Peru', 'vun': 'Peru', 'wae': 'Peru', 'wo': 'Peru', 'xh': 'EPeru', 'xnr': 'पेरू', 'xog': 'Peru', 'yav': 'pelú', 'yi': 'פּערו', 'yo': 'Pèérù', 'yrl': 'Peru', 'yue': '秘魯', 'yue-Hans': '秘鲁', 'yue-Hant': '秘魯', 'zgh': 'ⴱⵉⵔⵓ', 'zh': '秘鲁', 'zh-Hans': '秘鲁', 'zh-Hant': '秘魯', 'zu': 'i-Peru'},
    'PF': {'af': 'Frans-Polinesië', 'agq': 'Kpoleneshìa è Fàlâŋnsì', 'ak': 'Frɛnkye Pɔlenehyia', 'am': 'የፈረንሳይ ፖሊኔዢያ', 'ar': 'بولينيزيا الفرنسية', 'as': 'ফ্ৰান্স পোলেনচিয়া', 'asa': 'Polinesia ya Ufarantha', 'ast': 'Polinesia Francesa', 'az': 'Fransa Polineziyası', 'az-Cyrl': 'Франса Полинезијасы', 'az-Latn': 'Fransa Polineziyası', 'bal-Latn': 'Paránsi Pulinishiá', 'bas': 'Pòlìnesìà Pùlàsi', 'be': 'Французская Палінезія', 'be-tarask': 'Французская Палінезія', 'bez': 'Hupolinesia ya Huufaransa', 'bg': 'Френска Полинезия', 'blo': 'Polinesiya Gafɔntǝna kaja', 'bm': 'Faransi ka polinezi', 'bn': 'ফরাসী পলিনেশিয়া', 'br': 'Polinezia Cʼhall', 'brx': 'फ्रेन्च पलिनेशिया', 'bs': 'Francuska Polinezija', 'bs-Cyrl': 'Француска Полинезија', 'bs-Latn': 'Francuska Polinezija', 'ca': 'Polinèsia Francesa', 'ca-ES-valencia': 'Polinèsia Francesa', 'ccp': '𑄜𑄧𑄢𑄥𑄩 𑄛𑄧𑄣𑄨𑄚𑄬𑄥𑄨𑄠', 'ce': 'Французийн Полинези', 'ceb': 'French Polynesia', 'cgg': 'Polinesia ya Bufaransa', 'chr': 'ᎠᏂᎦᎸᏥ ᏆᎵᏂᏏᎠ', 'ckb': 'پۆلینیسیای فەرەنسا', 'cs': 'Francouzská Polynésie', 'cv': 'Франци Полинези', 'cy': 'Polynesia Ffrengig', 'da': 'Fransk Polynesien', 'dav': 'Polinesia ya Ufaransa', 'de': 'Französisch-Polynesien', 'dje': 'Faransi Polineezi', 'dsb': 'Francojska Polyneziska', 'dz': 'ཕྲཱནས྄་ཀྱི་པོ་ལི་ནི་ཤི་ཡ', 'ebu': 'Polinesia ya Ufaransa', 'ee': 'Frentsi Pɔlinesia nutome', 'el': 'Γαλλική Πολυνησία', 'el-polyton': 'Γαλλικὴ Πολυνησία', 'en': 'French Polynesia', 'en-Dsrt': '𐐙𐑉𐐯𐑌𐐽 𐐑𐐪𐑊𐐲𐑌𐐨𐑈𐐲', 'eo': 'Franca Polinezio', 'es': 'Polinesia Francesa', 'et': 'Prantsuse Polüneesia', 'eu': 'Polinesia Frantsesa', 'ewo': 'Polinesí yá Fulɛnsí', 'fa': 'پلی\u200cنزی فرانسه', 'ff': 'Polinesii Farayse', 'ff-Adlm': '𞤆𞤮𞤤𞤭𞤲𞤫𞥅𞤧𞤭𞤴𞤢 𞤊𞤪𞤢𞤲𞤧𞤭𞤲𞤳𞤮', 'ff-Latn': 'Polinesii Farayse', 'fi': 'Ranskan Polynesia', 'fil': 'French Polynesia', 'fo': 'Franska Polynesia', 'fr': 'Polynésie française', 'fur': 'Polinesie francês', 'fy': 'Frans-Polynesië', 'ga': 'Polainéis na Fraince', 'gd': 'Poilinèis na Frainge', 'gl': 'Polinesia Francesa', 'gsw': 'Französisch-Polineesie', 'gu': 'ફ્રેંચ પોલિનેશિયા', 'guz': 'Polinesia ya Ufaransa', 'ha': 'Folinesiya Ta Faransa', 'he': 'פולינזיה הצרפתית', 'hi': 'फ़्रेंच पोलिनेशिया', 'hi-Latn': 'French Polynesia', 'hr': 'Francuska Polinezija', 'hsb': 'Francoska Polyneziska', 'ht': 'Polynésie française', 'hu': 'Francia Polinézia', 'hy': 'Ֆրանսիական Պոլինեզիա', 'ia': 'Polynesia francese', 'id': 'Polinesia Prancis', 'ig': 'French Polynesia', 'is': 'Franska Pólýnesía', 'it': 'Polinesia Francese', 'ja': '仏領ポリネシア', 'jmc': 'Polinesia ya Ufaransa', 'jv': 'Polinesia Prancis', 'ka': 'საფრანგეთის პოლინეზია', 'kab': 'Pulunizi tafransist', 'kam': 'Polinesia ya Ufaransa', 'kde': 'Polinesia ya Ufalansa', 'kea': 'Polinézia Franseza', 'kgp': 'Frỹsa Porinẽja', 'khq': 'Faransi Polineezi', 'ki': 'Polinesia ya Ufaransa', 'kk': 'Француз Полинезиясы', 'kk-Cyrl': 'Француз Полинезиясы', 'kln': 'Emetab Polynesia nebo ufaransa', 'km': 'ប៉ូលី\u200bណេស៊ី\u200bបារាំង', 'kn': 'ಫ್ರೆಂಚ್ ಪಾಲಿನೇಷ್ಯಾ', 'ko': '프랑스령 폴리네시아', 'kok': 'फ्रेन्च पोलिनेसिया', 'kok-Deva': 'फ्रेन्च पोलिनेसिया', 'ks': 'فرانسی پولِنیشِیا', 'ks-Arab': 'فرانسی پولِنیشِیا', 'ksb': 'Polinesia ya Ufalansa', 'ksf': 'pɔlinɛsí a pɛrɛsǝ́', 'ksh': 'Franzüüsesch Pollineesije', 'ku': 'Polînezyaya Fransizî', 'kxv': 'prenc', 'kxv-Deva': 'प़्रेंच पोलिनेसिया', 'kxv-Latn': 'prenc', 'kxv-Orya': 'ପ୍ରେଞ୍ଚ୍ ପଲିନେସିୟା', 'kxv-Telu': 'ప్రెంచ్ పోలినిసియా', 'ky': 'Полинезия (франциялык)', 'lag': 'Polinésia yʉ Ʉfaráansa', 'lb': 'Franséisch-Polynesien', 'lg': 'Polinesiya enfalansa', 'lld': 'Polinesia franzeja', 'ln': 'Polinezi ya Falánsɛ', 'lo': 'ເຟຣນຊ໌ ໂພລິນີເຊຍ', 'lt': 'Prancūzijos Polinezija', 'lu': 'Polinezi wa Nfalanse', 'luo': 'French Polynesia', 'luy': 'Polinesia ya Ufaransa', 'lv': 'Francijas Polinēzija', 'mai': 'फ्रेंच पोलिनेशिया', 'mas': 'Polinesia ya Ufaransa', 'mer': 'Polinesia ya Fransi', 'mfe': 'Polinezi franse', 'mg': 'Polynezia frantsay', 'mgh': 'Ufaransa yo Potina', 'mi': 'Poronēhia Wīwī', 'mk': 'Француска Полинезија', 'ml': 'ഫ്രഞ്ച് പോളിനേഷ്യ', 'mn': 'Францын Полинез', 'mr': 'फ्रेंच पॉलिनेशिया', 'ms': 'Polinesia Perancis', 'mt': 'Polineżja Franċiża', 'mua': 'Sǝr Polynesiya ma Franssǝ', 'my': 'ပြင်သစ် ပေါ်လီနီးရှား', 'mzn': 'فرانسه\u200cی پولی\u200cنزی', 'naq': 'French Polynesia', 'nb': 'Fransk Polynesia', 'nd': 'Pholinesiya ye Fulansi', 'ne': 'फ्रान्सेली पोलिनेसिया', 'nl': 'Frans-Polynesië', 'nmg': 'Polynesia Fala', 'nn': 'Fransk Polynesia', 'no': 'Fransk Polynesia', 'nqo': 'ߝߊ߬ߙߊ߲߬ߛߌ߫ ߔߏߟߌߣߋߖ߭ߌ߫', 'nyn': 'Polinesia ya Bufaransa', 'om': 'Polineeshiyaa Faransaay', 'or': 'ଫ୍ରେଞ୍ଚ ପଲିନେସିଆ', 'pa': 'ਫਰੈਂਚ ਪੋਲੀਨੇਸ਼ੀਆ', 'pa-Guru': 'ਫਰੈਂਚ ਪੋਲੀਨੇਸ਼ੀਆ', 'pcm': 'Frẹ́nch Poliníshia', 'pl': 'Polinezja Francuska', 'ps': 'فرانسوي پولينيسيا', 'pt': 'Polinésia Francesa', 'qu': 'Polinesia Francesa', 'rm': 'Polinesia Franzosa', 'rn': 'Polineziya y’Abafaransa', 'ro': 'Polinezia Franceză', 'rof': 'Polinesia ya Ufaransa', 'ru': 'Французская Полинезия', 'rwk': 'Polinesia ya Ufaransa', 'saq': 'Polinesia ya Ufaransa', 'sbp': 'Polinesiya ya Wufalansa', 'sc': 'Polinèsia frantzesa', 'sd': 'فرانسيسي پولينيشيا', 'sd-Arab': 'فرانسيسي پولينيشيا', 'se': 'Frankriikka Polynesia', 'seh': 'Polinésia Francesa', 'ses': 'Faransi Polineezi', 'sg': 'Polinezïi tî farânzi', 'shi': 'ⴱⵓⵍⵉⵏⵉⵣⵢⴰ ⵜⴰⴼⵔⴰⵏⵙⵉⵙⵜ', 'shi-Latn': 'bulinizya tafransist', 'shi-Tfng': 'ⴱⵓⵍⵉⵏⵉⵣⵢⴰ ⵜⴰⴼⵔⴰⵏⵙⵉⵙⵜ', 'si': 'ප්\u200dරංශ පොලිනීසියාව', 'sk': 'Francúzska Polynézia', 'sl': 'Francoska Polinezija', 'smn': 'Ranska Polynesia', 'sn': 'French Polynesia', 'so': 'Booliyneesiya Faransiiska', 'sq': 'Polinezia Franceze', 'sr': 'Француска Полинезија', 'sr-Cyrl': 'Француска Полинезија', 'sr-Latn': 'Francuska Polinezija', 'sv': 'Franska Polynesien', 'sw': 'Polynesia ya Ufaransa', 'ta': 'பிரெஞ்சு பாலினேஷியா', 'te': 'ఫ్రెంచ్ పోలినీషియా', 'teo': 'Polinesia ya Ufaransa', 'tg': 'Полинезияи Фаронса', 'th': 'เฟรนช์โปลินีเซีย', 'ti': 'ፈረንሳዊት ፖሊነዥያ', 'tk': 'Fransuz Polineziýasy', 'to': 'Polinisia fakafalanisē', 'tr': 'Fransız Polinezyası', 'tt': 'Француз Полинезиясе', 'twq': 'Faransi Polineezi', 'tzm': 'Pulinizya Tafransist', 'ug': 'فىرانسىيەگە قاراشلىق پولىنېزىيە', 'uk': 'Французька Полінезія', 'ur': 'فرانسیسی پولینیشیا', 'uz': 'Fransuz Polineziyasi', 'uz-Cyrl': 'Француз Полинезияси', 'uz-Latn': 'Fransuz Polineziyasi', 'vai': 'ꗱꘋꔻ ꕶꔷꕇꔻꕩ', 'vai-Latn': 'Fɛŋsi Polinísiya', 'vai-Vaii': 'ꗱꘋꔻ ꕶꔷꕇꔻꕩ', 'vec': 'Polinezia franseze', 'vi': 'Polynesia thuộc Pháp', 'vun': 'Polinesia ya Ufaransa', 'wae': 'Französiš Polinesie', 'wo': 'Polinesi Farañse', 'xh': 'EFrench Polynesia', 'xnr': 'फ़्रेंच पोलिनेशिया', 'xog': 'Polinesiya enfalansa', 'yav': 'polinesí u felensí', 'yi': 'פֿראַנצויזישע פּאלינעזיע', 'yo': 'Firenṣi Polinesia', 'yrl': 'Pulinesiya Frãsa yara', 'yue': '法屬玻里尼西亞', 'yue-Hans': '法属玻里尼西亚', 'yue-Hant': '法屬玻里尼西亞', 'zgh': 'ⴱⵓⵍⵉⵏⵉⵣⵢⴰ ⵜⴰⴼⵔⴰⵏⵙⵉⵙⵜ', 'zh': '法属波利尼西亚', 'zh-Hans': '法属波利尼西亚', 'zh-Hant': '法屬玻里尼西亞', 'zu': 'i-French Polynesia'},
    'PG': {'af': 'Papoea-Nieu-Guinee', 'agq': 'Kpakpua Ginè È fūghū', 'ak': 'Papua Gini Foforɔ', 'am': 'ፓፑዋ ኒው ጊኒ', 'ar': 'بابوا غينيا الجديدة', 'as': 'পাপুৱা নিউ গিনি', 'asa': 'Papua', 'ast': 'Papúa Nueva Guinea', 'az': 'Papua-Yeni Qvineya', 'az-Cyrl': 'Папуа-Јени Гвинеја', 'az-Latn': 'Papua-Yeni Qvineya', 'bal-Latn': 'Pápuá Niu Giniá', 'bas': 'Gìne ì Pàpu', 'be': 'Папуа-Новая Гвінея', 'be-tarask': 'Папуа-Новая Гвінея', 'bez': 'Hupapua', 'bg': 'Папуа-Нова Гвинея', 'blo': 'Papuasii Ginee afɔlɩ', 'bm': 'Papuwasi-Gine-Koura', 'bn': 'পাপুয়া নিউ গিনি', 'br': 'Papoua Ginea-Nevez', 'brx': 'पापुआ निउ गिनी', 'bs': 'Papua Nova Gvineja', 'bs-Cyrl': 'Папуа Нова Гвинеја', 'bs-Latn': 'Papua Nova Gvineja', 'ca': 'Papua Nova Guinea', 'ca-ES-valencia': 'Papua Nova Guinea', 'ccp': '𑄛𑄛𑄪𑄠 𑄚𑄨𑄃𑄪 𑄉𑄨𑄚𑄨', 'ce': 'Папуа — Керла Гвиней', 'ceb': 'Papua New Guinea', 'cgg': 'Papua', 'chr': 'ᏆᏇ ᎢᏤ ᎩᎢᏂ', 'ckb': 'پاپوا گینێی نوێ', 'cs': 'Papua-Nová Guinea', 'cv': 'Папуа — Ҫӗнӗ Гвиней', 'cy': 'Papua Guinea Newydd', 'da': 'Papua Ny Guinea', 'dav': 'Papua', 'de': 'Papua-Neuguinea', 'dje': 'Papua Ginee Taaga', 'dsb': 'Papua-Neuguinea', 'dz': 'པ་པུ་ ནིའུ་གི་ནི', 'ebu': 'Papua', 'ee': 'Papua New Gini nutome', 'el': 'Παπούα Νέα Γουινέα', 'el-polyton': 'Παπούα Νέα Γουινέα', 'en': 'Papua New Guinea', 'en-Dsrt': '𐐑𐐰𐐹𐐷𐐳𐐲 𐐤𐐭 𐐘𐐮𐑌𐐨', 'eo': 'Papuo-Nov-Gvineo', 'es': 'Papúa Nueva Guinea', 'et': 'Paapua Uus-Guinea', 'eu': 'Papua Ginea Berria', 'ewo': 'Papwazi yá Ǹkpámɛ́n Giné', 'fa': 'پاپوا گینهٔ نو', 'ff': 'Papuwaa Nuwel Gine', 'ff-Adlm': '𞤆𞤢𞤨𞤵𞤱𞤢 𞤘𞤭𞤲𞤫 𞤖𞤫𞤧𞤮', 'ff-Latn': 'Papuwaa Nuwel Gine', 'fi': 'Papua-Uusi-Guinea', 'fil': 'Papua New Guinea', 'fo': 'Papua Nýguinea', 'fr': 'Papouasie-Nouvelle-Guinée', 'fur': 'Papue Gnove Guinee', 'fy': 'Papoea-Nij-Guinea', 'ga': 'Nua-Ghuine Phapua', 'gd': 'Gini Nuadh Phaputhach', 'gl': 'Papúa-Nova Guinea', 'gsw': 'Papua-Neuguinea', 'gu': 'પાપુઆ ન્યૂ ગિની', 'guz': 'Papua', 'ha': 'Papuwa Nugini', 'he': 'פפואה גינאה החדשה', 'hi': 'पापुआ न्यू गिनी', 'hi-Latn': 'Papua New Guinea', 'hr': 'Papua Nova Gvineja', 'hsb': 'Papuwa-Nowa Gineja', 'ht': 'Papouasie-Nouvelle-Guinée', 'hu': 'Pápua Új-Guinea', 'hy': 'Պապուա Նոր Գվինեա', 'ia': 'Papua Nove Guinea', 'id': 'Papua Nugini', 'ig': 'Papua New Guinea', 'is': 'Papúa Nýja-Gínea', 'it': 'Papua Nuova Guinea', 'ja': 'パプアニューギニア', 'jmc': 'Papua', 'jv': 'Papua Nugini', 'ka': 'პაპუა-ახალი გვინეა', 'kaa': 'Папуа — Жаӊа Гвинея', 'kaa-Cyrl': 'Папуа — Жаӊа Гвинея', 'kab': 'Ɣinya Tamaynut Tapaput', 'kam': 'Papua', 'kde': 'Papua', 'kea': 'Papua-Nova Gine', 'kgp': 'Papuva-Ginẽ Tãg', 'khq': 'Papua Ginee Taaga', 'ki': 'Papua', 'kk': 'Папуа — Жаңа Гвинея', 'kk-Cyrl': 'Папуа — Жаңа Гвинея', 'kln': 'Emetab Papua New Guinea', 'km': 'ប៉ាពូអាស៊ី\u200bនូវែលហ្គីណេ', 'kn': 'ಪಪುವಾ ನ್ಯೂಗಿನಿಯಾ', 'ko': '파푸아뉴기니', 'kok': 'पापुआ न्यु गिनी', 'kok-Deva': 'पापुआ न्यु गिनी', 'ks': 'پاپُوا نیوٗ گیٖنی', 'ks-Arab': 'پاپُوا نیوٗ گیٖنی', 'ksb': 'Papua', 'ksf': 'papwazí ginɛ́ anyɔ́n', 'ksh': 'Papuwa Neujineeja', 'ku': 'Papua Gîneya Nû', 'kxv': 'papua nyu gini', 'kxv-Deva': 'पापुआ न्यू गिनी', 'kxv-Latn': 'papua nyu gini', 'kxv-Orya': 'ପପୁଆ ନିୟୁ ଗିନି', 'kxv-Telu': 'పాపువా న్యూ గనియా', 'ky': 'Папуа-Жаңы Гвинея', 'lag': 'Papúua', 'lb': 'Papua-Neiguinea', 'lg': 'Papwa Nyugini', 'lld': 'Papua Nöia Guinea', 'ln': 'Papwazi Ginɛ ya sika', 'lo': 'ປາປົວນິວກີນີ', 'lt': 'Papua Naujoji Gvinėja', 'lu': 'Papwazi wa Nginɛ wa mumu', 'luo': 'Papua New Guinea', 'luy': 'Papua', 'lv': 'Papua-Jaungvineja', 'mai': 'पापुआ न्यू गिनी', 'mas': 'Papua', 'mer': 'Papua', 'mfe': 'Papouazi-Nouvel-Gine', 'mg': 'Papouasie-Nouvelle-Guinée', 'mgh': 'Upapua', 'mi': 'Papua Nūkini', 'mk': 'Папуа Нова Гвинеја', 'ml': 'പാപ്പുവ ന്യൂ ഗിനിയ', 'mn': 'Папуа Шинэ Гвиней', 'mr': 'पापुआ न्यू गिनी', 'ms': 'Papua New Guinea', 'mt': 'Papua New Guinea', 'mua': 'Papuasiya Guine mafuu', 'my': 'ပါပူအာ နယူးဂီနီ', 'mzn': 'پاپوا نو گینه', 'naq': 'Papua New Guineab', 'nb': 'Papua Ny-Guinea', 'nd': 'Papua New Guinea', 'ne': 'पपुआ न्यू गाइनिया', 'nl': 'Papoea-Nieuw-Guinea', 'nmg': 'Guine Papuasi', 'nn': 'Papua Ny-Guinea', 'no': 'Papua Ny-Guinea', 'nqo': 'ߡߊߡߎߥߊߖ߭ߌ߫ ߖߌ߬ߣߍ߬-ߞߎߘߊ߫', 'nyn': 'Papua', 'om': 'Papuwa Neawu Giinii', 'or': 'ପପୁଆ ନ୍ୟୁ ଗିନି', 'pa': 'ਪਾਪੂਆ ਨਿਊ ਗਿਨੀ', 'pa-Guru': 'ਪਾਪੂਆ ਨਿਊ ਗਿਨੀ', 'pcm': 'Pápuá Niú Gíni', 'pl': 'Papua-Nowa Gwinea', 'ps': 'پاپوا نيو ګيني', 'pt': 'Papua-Nova Guiné', 'qu': 'Papúa Nueva Guinea', 'rm': 'Papua Nova Guinea', 'rn': 'Papuwa Niyugineya', 'ro': 'Papua-Noua Guinee', 'rof': 'Papua', 'ru': 'Папуа — Новая Гвинея', 'rwk': 'Papua', 'saq': 'Papua', 'sbp': 'Papuwa', 'sc': 'Pàpua Guinea Noa', 'sd': 'پاپوا نیو گني', 'sd-Arab': 'پاپوا نیو گني', 'se': 'Papua-Ođđa-Guinea', 'seh': 'Papua-Nova Guiné', 'ses': 'Papua Ginee Taaga', 'sg': 'Papû Finî Ginëe, Papuazïi', 'shi': 'ⴱⴰⴱⵡⴰ ⵖⵉⵏⵢⴰ ⵜⴰⵎⴰⵢⵏⵓⵜ', 'shi-Latn': 'babwa ɣinya tamaynut', 'shi-Tfng': 'ⴱⴰⴱⵡⴰ ⵖⵉⵏⵢⴰ ⵜⴰⵎⴰⵢⵏⵓⵜ', 'si': 'පැපුවා නිව් ගිනියාව', 'sk': 'Papua-Nová Guinea', 'sl': 'Papua Nova Gvineja', 'smn': 'Papua-Uđđâ-Guinea', 'sn': 'Papua New Guinea', 'so': 'Babwa Niyuu Gini', 'sq': 'Guineja e Re-Papua', 'sr': 'Папуа Нова Гвинеја', 'sr-Cyrl': 'Папуа Нова Гвинеја', 'sr-Latn': 'Papua Nova Gvineja', 'sv': 'Papua Nya Guinea', 'sw': 'Papua New Guinea', 'syr': 'ܦܐܦܘܐ ܓܝܢܝܐ ܚܕܬܐ', 'ta': 'பப்புவா நியூ கினியா', 'te': 'పాపువా న్యూ గినియా', 'teo': 'Papua', 'tg': 'Папуа Гвинеяи Нав', 'th': 'ปาปัวนิวกินี', 'ti': 'ፓፕዋ ኒው ጊኒ', 'tk': 'Papua - Täze Gwineýa', 'to': 'Papuaniukini', 'tpi': 'Papua Niugini', 'tr': 'Papua Yeni Gine', 'tt': 'Папуа - Яңа Гвинея', 'twq': 'Papua Ginee Taaga', 'tzm': 'Papwa Ɣinya Tamaynut', 'ug': 'پاپۇئا يېڭى گىۋىنىيەسى', 'uk': 'Папуа-Нова Гвінея', 'ur': 'پاپوآ نیو گنی', 'uz': 'Papua – Yangi Gvineya', 'uz-Cyrl': 'Папуа - Янги Гвинея', 'uz-Latn': 'Papua – Yangi Gvineya', 'vai': 'ꕐꖛꕎ ꕅꔤꕇ ꕯꕮꕊ', 'vai-Latn': 'Papuwa Gini Námaá', 'vai-Vaii': 'ꕐꖛꕎ ꕅꔤꕇ ꕯꕮꕊ', 'vec': 'Papua Nova Guinèa', 'vi': 'Papua New Guinea', 'vun': 'Papua', 'wae': 'Papua Niwginea', 'wo': 'Papuwasi Gine Gu Bees', 'xh': 'EPapua New Guinea', 'xnr': 'पापुआ न्यू गिनी', 'xog': 'Papwa Nyugini', 'yav': 'papuasí nufɛ́l kiiné', 'yi': 'פּאַפּואַ נײַ גינע', 'yo': 'Paapu ti Giini', 'yrl': 'Papuwa-Giné Pisasú', 'yue': '巴布亞紐幾內亞', 'yue-Hans': '巴布亚纽几内亚', 'yue-Hant': '巴布亞紐幾內亞', 'zgh': 'ⴱⴰⴱⵡⴰ ⵖⵉⵏⵢⴰ ⵜⴰⵎⴰⵢⵏⵓⵜ', 'zh': '巴布亚新几内亚', 'zh-Hans': '巴布亚新几内亚', 'zh-Hant': '巴布亞紐幾內亞', 'zu': 'i-Papua New Guinea'},
    'PH': {'af': 'Filippyne', 'agq': 'Felèkpî', 'ak': 'Filipin', 'am': 'ፊሊፒንስ', 'ar': 'الفلبين', 'as': 'ফিলিপাইনছ', 'asa': 'Filipino', 'ast': 'Filipines', 'az': 'Filippin', 'az-Cyrl': 'Филиппин', 'az-Latn': 'Filippin', 'bal-Latn': 'Pelpin', 'bas': 'Fìlìpîn', 'be': 'Філіпіны', 'be-tarask': 'Філіпіны', 'bez': 'Hufilipino', 'bg': 'Филипини', 'bgn': 'فلیپین', 'blo': 'Filipiin', 'bm': 'Filipini', 'bn': 'ফিলিপাইন', 'br': 'Filipinez', 'brx': 'फिलीपिन्स', 'bs': 'Filipini', 'bs-Cyrl': 'Филипини', 'bs-Latn': 'Filipini', 'ca': 'Filipines', 'ca-ES-valencia': 'Filipines', 'ccp': '𑄜𑄨𑄣𑄨𑄛𑄭𑄚𑄴', 'ce': 'Филиппинаш', 'ceb': 'Pilipinas', 'cgg': 'Firipino', 'chr': 'ᎠᏂᏈᎵᎩᏃ', 'ckb': 'فلیپین', 'cs': 'Filipíny', 'cv': 'Филиппинсем', 'cy': 'Y Philipinau', 'da': 'Filippinerne', 'dav': 'Filipino', 'de': 'Philippinen', 'dje': 'Filipine', 'dsb': 'Filipiny', 'dz': 'ཕི་ལི་པིནས', 'ebu': 'Filipino', 'ee': 'Filipini nutome', 'el': 'Φιλιππίνες', 'el-polyton': 'Φιλιππίνες', 'en': 'Philippines', 'en-Dsrt': '𐐙𐐮𐑊𐐲𐐹𐐨𐑌𐑆', 'eo': 'Filipinoj', 'es': 'Filipinas', 'et': 'Filipiinid', 'eu': 'Filipinak', 'ewo': 'Filipín', 'fa': 'فیلیپین', 'ff': 'Filipiin', 'ff-Adlm': '𞤊𞤭𞤤𞤭𞤨𞤭𞥅𞤲', 'ff-Latn': 'Filipiin', 'fi': 'Filippiinit', 'fil': 'Pilipinas', 'fo': 'Filipsoyggjar', 'fr': 'Philippines', 'fur': 'Filipinis', 'fy': 'Filipijnen', 'ga': 'Na hOileáin Fhilipíneacha', 'gd': 'Na h-Eileanan Filipineach', 'gl': 'Filipinas', 'gsw': 'Philippiine', 'gu': 'ફિલિપિન્સ', 'guz': 'Filipino', 'ha': 'Filipin', 'haw': 'ʻĀina Pilipino', 'he': 'הפיליפינים', 'hi': 'फ़िलिपींस', 'hi-Latn': 'Philippines', 'hr': 'Filipini', 'hsb': 'Filipiny', 'ht': 'Philippines', 'hu': 'Fülöp-szigetek', 'hy': 'Ֆիլիպիններ', 'ia': 'Philippinas', 'id': 'Filipina', 'ie': 'Filipines', 'ig': 'Philippines', 'is': 'Filippseyjar', 'it': 'Filippine', 'ja': 'フィリピン', 'jmc': 'Filipino', 'jv': 'Pilipina', 'ka': 'ფილიპინები', 'kaa': 'Филиппин', 'kaa-Cyrl': 'Филиппин', 'kab': 'Filipin', 'kam': 'Filipino', 'kde': 'Filipino', 'kea': 'Filipinas', 'kgp': 'Firipinỹ', 'khq': 'Filipine', 'ki': 'Filipino', 'kk': 'Филиппин аралдары', 'kk-Cyrl': 'Филиппин аралдары', 'kln': 'Emetab Philippines', 'km': 'ហ្វ៊ីលីពីន', 'kn': 'ಫಿಲಿಫೈನ್ಸ್', 'ko': '필리핀', 'kok': 'फिलीपिन्झ', 'kok-Deva': 'फिलीपिन्झ', 'ks': 'فلپائن', 'ks-Arab': 'فلپائن', 'ksb': 'Filipino', 'ksf': 'filipǝ́n', 'ksh': 'Fillipiine', 'ku': 'Fîlîpîn', 'kxv': 'pilippines', 'kxv-Deva': 'प़िलिपींस', 'kxv-Latn': 'pilippines', 'kxv-Orya': 'ପିଲିପାଇନ୍ସ', 'kxv-Telu': 'పిలిప్పైన్స్', 'ky': 'Филиппин', 'lag': 'Ufilipíino', 'lb': 'Philippinnen', 'lg': 'Bizinga bya Firipino', 'lld': 'Filipines', 'ln': 'Filipinɛ', 'lo': 'ຟິລິບປິນ', 'lt': 'Filipinai', 'lu': 'Nfilipi', 'luo': 'Philippines', 'luy': 'Filipino', 'lv': 'Filipīnas', 'mai': 'फिलिपींस', 'mas': 'Filipino', 'mer': 'Filipino', 'mfe': 'Filipinn', 'mg': 'Filipina', 'mgh': 'Ufilipino', 'mi': 'Piripīni', 'mk': 'Филипини', 'ml': 'ഫിലിപ്പീൻസ്', 'mn': 'Филиппин', 'mr': 'फिलिपिन्स', 'ms': 'Filipina', 'mt': 'il-Filippini', 'mua': 'Filipiŋ', 'my': 'ဖိလစ်ပိုင်', 'mzn': 'فیلیپین', 'naq': 'Philippinni', 'nb': 'Filippinene', 'nd': 'Philippines', 'ne': 'फिलिपिन्स', 'nl': 'Filipijnen', 'nmg': 'Filipin', 'nn': 'Filippinane', 'no': 'Filippinene', 'nqo': 'ߝߟߌߔߌ߲ߣ', 'nyn': 'Firipino', 'om': 'Filippiins', 'or': 'ଫିଲିପାଇନସ୍', 'pa': 'ਫਿਲੀਪੀਨਜ', 'pa-Guru': 'ਫਿਲੀਪੀਨਜ', 'pcm': 'Fílípins', 'pl': 'Filipiny', 'ps': 'فلپين', 'pt': 'Filipinas', 'qu': 'Filipinas', 'rm': 'Filippinas', 'rn': 'Amazinga ya Filipine', 'ro': 'Filipine', 'rof': 'Filipino', 'ru': 'Филиппины', 'rwk': 'Filipino', 'saq': 'Filipino', 'sbp': 'Filipino', 'sc': 'Filipinas', 'sd': 'فلپائن', 'sd-Arab': 'فلپائن', 'se': 'Filippiinnat', 'seh': 'Filipinas', 'ses': 'Filipine', 'sg': 'Filipîni', 'shi': 'ⴼⵉⵍⵉⴱⴱⵉⵏ', 'shi-Latn': 'filibbin', 'shi-Tfng': 'ⴼⵉⵍⵉⴱⴱⵉⵏ', 'si': 'පිලිපීනය', 'sk': 'Filipíny', 'sl': 'Filipini', 'smn': 'Filipineh', 'sn': 'Philippines', 'so': 'Filibiin', 'sq': 'Filipine', 'sr': 'Филипини', 'sr-Cyrl': 'Филипини', 'sr-Latn': 'Filipini', 'sv': 'Filippinerna', 'sw': 'Ufilipino', 'syr': 'ܦܝܠܝܦܝܢܝܐ', 'ta': 'பிலிப்பைன்ஸ்', 'te': 'ఫిలిప్పైన్స్', 'teo': 'Filipino', 'tg': 'Филиппин', 'th': 'ฟิลิปปินส์', 'ti': 'ፊሊፒንስ', 'tk': 'Filippinler', 'to': 'Filipaini', 'tr': 'Filipinler', 'tt': 'Филиппин', 'twq': 'Filipine', 'tzm': 'Filippin', 'ug': 'فىلىپپىن', 'uk': 'Філіппіни', 'ur': 'فلپائن', 'uz': 'Filippin', 'uz-Cyrl': 'Филиппин', 'uz-Latn': 'Filippin', 'vai': 'ꔱꔒꔪꘋ', 'vai-Latn': 'Félepiŋ', 'vai-Vaii': 'ꔱꔒꔪꘋ', 'vec': 'Filipine', 'vi': 'Philippines', 'vun': 'Filipino', 'wae': 'Philippine', 'wo': 'Filipin', 'xh': 'EPhilippines', 'xnr': 'फ़िलिपींस', 'xog': 'Ebizinga bya Firipino', 'yav': 'filipíin', 'yi': 'פֿיליפּינען', 'yo': 'Filipini', 'yrl': 'Firipina', 'yue': '菲律賓', 'yue-Hans': '菲律宾', 'yue-Hant': '菲律賓', 'zgh': 'ⴼⵉⵍⵉⴱⴱⵉⵏ', 'zh': '菲律宾', 'zh-Hans': '菲律宾', 'zh-Hant': '菲律賓', 'zu': 'i-Philippines'},
    'PK': {'af': 'Pakistan', 'agq': 'Kpakìsɨ̀tân', 'ak': 'Pakistan', 'am': 'ፓኪስታን', 'ar': 'باكستان', 'as': 'পাকিস্তান', 'asa': 'Pakithtani', 'ast': 'Paquistán', 'az': 'Pakistan', 'az-Cyrl': 'Пакистан', 'az-Latn': 'Pakistan', 'bal-Latn': 'Pákestán', 'bas': 'Pàkìstân', 'be': 'Пакістан', 'be-tarask': 'Пакістан', 'bez': 'Hupakistani', 'bg': 'Пакистан', 'blo': 'Pakistan', 'bm': 'Pakisitaŋ', 'bn': 'পাকিস্তান', 'br': 'Pakistan', 'brx': 'पाकिस्तान', 'bs': 'Pakistan', 'bs-Cyrl': 'Пакистан', 'bs-Latn': 'Pakistan', 'ca': 'Pakistan', 'ca-ES-valencia': 'Pakistan', 'ccp': '𑄛𑄇𑄨𑄌𑄴𑄖𑄚𑄴', 'ce': 'Пакистан', 'ceb': 'Pakistan', 'cgg': 'Pakisitaani', 'chr': 'ᏆᎩᏍᏖᏂ', 'ckb': 'پاکستان', 'cs': 'Pákistán', 'cv': 'Пакистан', 'cy': 'Pakistan', 'da': 'Pakistan', 'dav': 'Pakistani', 'de': 'Pakistan', 'dje': 'Paakistan', 'dsb': 'Pakistan', 'dz': 'པ་ཀི་སཏཱན', 'ebu': 'Pakistani', 'ee': 'Pakistan nutome', 'el': 'Πακιστάν', 'el-polyton': 'Πακιστάν', 'en': 'Pakistan', 'en-Dsrt': '𐐑𐐰𐐿𐐲𐑅𐐻𐐰𐑌', 'eo': 'Pakistano', 'es': 'Pakistán', 'et': 'Pakistan', 'eu': 'Pakistan', 'ewo': 'Pakisətán', 'fa': 'پاکستان', 'ff': 'Pakistaan', 'ff-Adlm': '𞤆𞤢𞤳𞤭𞤧𞤼𞤢𞥄𞤲', 'ff-Latn': 'Pakistaan', 'fi': 'Pakistan', 'fil': 'Pakistan', 'fo': 'Pakistan', 'fr': 'Pakistan', 'fur': 'Pakistan', 'fy': 'Pakistan', 'ga': 'an Phacastáin', 'gd': 'Pagastàn', 'gl': 'Paquistán', 'gsw': 'Pakischtan', 'gu': 'પાકિસ્તાન', 'guz': 'Pakistani', 'ha': 'Pakistan', 'he': 'פקיסטן', 'hi': 'पाकिस्तान', 'hi-Latn': 'Pakistan', 'hr': 'Pakistan', 'hsb': 'Pakistan', 'ht': 'Pakistan', 'hu': 'Pakisztán', 'hy': 'Պակիստան', 'ia': 'Pakistan', 'id': 'Pakistan', 'ie': 'Pakistan', 'ig': 'Pakistan', 'is': 'Pakistan', 'it': 'Pakistan', 'ja': 'パキスタン', 'jmc': 'Pakistani', 'jv': 'Pakistan', 'ka': 'პაკისტანი', 'kaa': 'Пакистан', 'kaa-Cyrl': 'Пакистан', 'kab': 'Pakistan', 'kam': 'Pakistani', 'kde': 'Pakistani', 'kea': 'Pakistan', 'kgp': 'Pakinhtỹv', 'khq': 'Paakistan', 'ki': 'Pakistani', 'kk': 'Пәкістан', 'kk-Cyrl': 'Пәкістан', 'kln': 'Emetab Pakistan', 'km': 'ប៉ាគីស្ថាន', 'kn': 'ಪಾಕಿಸ್ತಾನ', 'ko': '파키스탄', 'kok': 'पाकिस्तान', 'kok-Deva': 'पाकिस्तान', 'ks': 'پاکِستان', 'ks-Arab': 'پاکِستان', 'ksb': 'Pakistani', 'ksf': 'pakistáŋ', 'ksh': 'Pakistahn', 'ku': 'Pakistan', 'kxv': 'pakistan', 'kxv-Deva': 'पाकिस्तान', 'kxv-Latn': 'pakistan', 'kxv-Orya': 'ପାକିସ୍ତାନ', 'kxv-Telu': 'పాకిస్తాన్', 'ky': 'Пакистан', 'lag': 'Pakisitáani', 'lb': 'Pakistan', 'lg': 'Pakisitaani', 'lld': 'Pakistan', 'ln': 'Pakisitá', 'lo': 'ປາກິດສະຖານ', 'lt': 'Pakistanas', 'lu': 'Pakisita', 'luo': 'Pakistan', 'luy': 'Pakistani', 'lv': 'Pakistāna', 'mai': 'पाकिस्तान', 'mas': 'Pakistani', 'mer': 'Pakistani', 'mfe': 'Pakistan', 'mg': 'Pakistan', 'mgh': 'Upakistani', 'mi': 'Pakitāne', 'mk': 'Пакистан', 'ml': 'പാക്കിസ്ഥാൻ', 'mn': 'Пакистан', 'mr': 'पाकिस्तान', 'ms': 'Pakistan', 'mt': 'il-Pakistan', 'mua': 'Pakistaŋ', 'my': 'ပါကစ္စတန်', 'mzn': 'پاکستون', 'naq': 'Pakistanni', 'nb': 'Pakistan', 'nd': 'Phakistani', 'ne': 'पाकिस्तान', 'nl': 'Pakistan', 'nmg': 'Pakistan', 'nn': 'Pakistan', 'no': 'Pakistan', 'nqo': 'ߔߊߞߌߛߑߕߊ߲߫', 'nyn': 'Pakisitaani', 'om': 'Paakistaan', 'or': 'ପାକିସ୍ତାନ', 'pa': 'ਪਾਕਿਸਤਾਨ', 'pa-Arab': 'پاکستان', 'pa-Guru': 'ਪਾਕਿਸਤਾਨ', 'pcm': 'Pakístan', 'pl': 'Pakistan', 'ps': 'پاکستان', 'pt': 'Paquistão', 'qu': 'Pakistán', 'rm': 'Pakistan', 'rn': 'Pakisitani', 'ro': 'Pakistan', 'rof': 'Pakistani', 'ru': 'Пакистан', 'rwk': 'Pakistani', 'saq': 'Pakistani', 'sbp': 'Pakisitani', 'sc': 'Pàkistan', 'sd': 'پاڪستان', 'sd-Arab': 'پاڪستان', 'sd-Deva': 'पाकिस्तान', 'se': 'Pakistan', 'seh': 'Paquistão', 'ses': 'Paakistan', 'sg': 'Pakistäan', 'shi': 'ⴱⴰⴽⵉⵙⵜⴰⵏ', 'shi-Latn': 'bakistan', 'shi-Tfng': 'ⴱⴰⴽⵉⵙⵜⴰⵏ', 'si': 'පාකිස්තානය', 'sk': 'Pakistan', 'sl': 'Pakistan', 'smn': 'Pakistan', 'sn': 'Pakistan', 'so': 'Bakistaan', 'sq': 'Pakistan', 'sr': 'Пакистан', 'sr-Cyrl': 'Пакистан', 'sr-Latn': 'Pakistan', 'sv': 'Pakistan', 'sw': 'Pakistani', 'syr': 'ܦܐܟܣܬܐܢ', 'ta': 'பாகிஸ்தான்', 'te': 'పాకిస్తాన్', 'teo': 'Pakistani', 'tg': 'Покистон', 'th': 'ปากีสถาน', 'ti': 'ፓኪስታን', 'tk': 'Pakistan', 'to': 'Pākisitani', 'tr': 'Pakistan', 'tt': 'Пакистан', 'twq': 'Paakistan', 'tzm': 'Pakistan', 'ug': 'پاكىستان', 'uk': 'Пакистан', 'ur': 'پاکستان', 'uz': 'Pokiston', 'uz-Cyrl': 'Покистон', 'uz-Latn': 'Pokiston', 'vai': 'ꕐꕃꔻꕚꘋ', 'vai-Latn': 'Pakisitaŋ', 'vai-Vaii': 'ꕐꕃꔻꕚꘋ', 'vec': 'Pakistan', 'vi': 'Pakistan', 'vun': 'Pakistani', 'wae': 'Pakištan', 'wo': 'Pakistaŋ', 'xh': 'EPakistan', 'xnr': 'पाकिस्तान', 'xog': 'Pakisitaani', 'yav': 'pakistáŋ', 'yi': 'פּאַקיסטאַן', 'yo': 'Pakisitan', 'yrl': 'Pakiretãma', 'yue': '巴基斯坦', 'yue-Hans': '巴基斯坦', 'yue-Hant': '巴基斯坦', 'zgh': 'ⴱⴰⴽⵉⵙⵜⴰⵏ', 'zh': '巴基斯坦', 'zh-Hans': '巴基斯坦', 'zh-Hant': '巴基斯坦', 'zu': 'i-Pakistan'},
    'PL': {'af': 'Pole', 'agq': 'Kpulàn', 'ak': 'Pɔland', 'am': 'ፖላንድ', 'ar': 'بولندا', 'as': 'পোলেণ্ড', 'asa': 'Polandi', 'ast': 'Polonia', 'az': 'Polşa', 'az-Cyrl': 'Полша', 'az-Latn': 'Polşa', 'bal-Latn': 'Pólánd', 'bas': 'Pòlànd', 'be': 'Польшча', 'be-tarask': 'Польшча', 'bez': 'Hupolandi', 'bg': 'Полша', 'blo': 'Polanɖ', 'bm': 'Poloɲi', 'bn': 'পোল্যান্ড', 'br': 'Polonia', 'brx': 'पलेण्ड', 'bs': 'Poljska', 'bs-Cyrl': 'Пољска', 'bs-Latn': 'Poljska', 'ca': 'Polònia', 'ca-ES-valencia': 'Polònia', 'ccp': '𑄛𑄮𑄣𑄳𑄠𑄚𑄴𑄓𑄴', 'ce': 'Польша', 'ceb': 'Poland', 'cgg': 'Poorandi', 'chr': 'ᏉᎳᏂ', 'ckb': 'پۆڵەندا', 'cs': 'Polsko', 'cv': 'Польша', 'cy': 'Gwlad Pwyl', 'da': 'Polen', 'dav': 'Polandi', 'de': 'Polen', 'dje': 'Poloɲe', 'dsb': 'Pólska', 'dz': 'པོ་ལེནཌ', 'ebu': 'Polandi', 'ee': 'Poland nutome', 'el': 'Πολωνία', 'el-polyton': 'Πολωνία', 'en': 'Poland', 'en-Dsrt': '𐐑𐐬𐑊𐐲𐑌𐐼', 'eo': 'Pollando', 'es': 'Polonia', 'et': 'Poola', 'eu': 'Polonia', 'ewo': 'fólis', 'fa': 'لهستان', 'ff': 'Poloñ', 'ff-Adlm': '𞤆𞤮𞤤𞤢𞤲𞤣', 'ff-Latn': 'Poloñ', 'fi': 'Puola', 'fil': 'Poland', 'fo': 'Pólland', 'fr': 'Pologne', 'fur': 'Polonie', 'fy': 'Polen', 'ga': 'an Pholainn', 'gd': 'A’ Phòlainn', 'gl': 'Polonia', 'gsw': 'Poole', 'gu': 'પોલેંડ', 'guz': 'Polandi', 'ha': 'Polan', 'he': 'פולין', 'hi': 'पोलैंड', 'hi-Latn': 'Poland', 'hr': 'Poljska', 'hsb': 'Pólska', 'ht': 'Pologne', 'hu': 'Lengyelország', 'hy': 'Լեհաստան', 'ia': 'Polonia', 'id': 'Polandia', 'ie': 'Polonia', 'ig': 'Poland', 'is': 'Pólland', 'it': 'Polonia', 'ja': 'ポーランド', 'jmc': 'Polandi', 'jv': 'Polen', 'ka': 'პოლონეთი', 'kaa': 'Польша', 'kaa-Cyrl': 'Польша', 'kab': 'Pulund', 'kam': 'Polandi', 'kde': 'Polandi', 'kea': 'Pulónia', 'kgp': 'Porãnija', 'khq': 'Poloɲe', 'ki': 'Polandi', 'kk': 'Польша', 'kk-Cyrl': 'Польша', 'kln': 'Emetab Poland', 'km': 'ប៉ូឡូញ', 'kn': 'ಪೋಲ್ಯಾಂಡ್', 'ko': '폴란드', 'kok': 'पोलंड', 'kok-Deva': 'पोलंड', 'ks': 'پولینڈ', 'ks-Arab': 'پولینڈ', 'ksb': 'Polandi', 'ksf': 'polɔ́n', 'ksh': 'Pole', 'ku': 'Polonya', 'kxv': 'polanḍ', 'kxv-Deva': 'पोलेंड', 'kxv-Latn': 'polanḍ', 'kxv-Orya': 'ପଲାଣ୍ଡ', 'kxv-Telu': 'పోలాండ్', 'ky': 'Польша', 'lag': 'Pólandi', 'lb': 'Polen', 'lg': 'Polandi', 'lld': 'Polonia', 'ln': 'Poloni', 'lo': 'ໂປແລນ', 'lt': 'Lenkija', 'lu': 'Mpoloni', 'luo': 'Poland', 'luy': 'Polandi', 'lv': 'Polija', 'mai': 'पोलैंड', 'mas': 'Polandi', 'mer': 'Polandi', 'mfe': 'Pologn', 'mg': 'Pôlôna', 'mgh': 'Upolandi', 'mi': 'Pōrana', 'mk': 'Полска', 'ml': 'പോളണ്ട്', 'mn': 'Польш', 'mr': 'पोलंड', 'ms': 'Poland', 'mt': 'il-Polonja', 'mua': 'Pologŋ', 'my': 'ပိုလန်', 'mzn': 'لهستون', 'naq': 'Polandi', 'nb': 'Polen', 'nd': 'Pholandi', 'ne': 'पोल्याण्ड', 'nl': 'Polen', 'nmg': 'Pɔlɔŋ', 'nn': 'Polen', 'no': 'Polen', 'nqo': 'ߔߏߟߐߢ', 'nyn': 'Poorandi', 'om': 'Poolaand', 'or': 'ପୋଲାଣ୍ଡ', 'pa': 'ਪੋਲੈਂਡ', 'pa-Guru': 'ਪੋਲੈਂਡ', 'pcm': 'Póland', 'pl': 'Polska', 'prg': 'Pōli', 'ps': 'پولنډ', 'pt': 'Polônia', 'qu': 'Polonia', 'rm': 'Pologna', 'rn': 'Polonye', 'ro': 'Polonia', 'rof': 'Polandi', 'ru': 'Польша', 'rwk': 'Polandi', 'saq': 'Polandi', 'sbp': 'Polandi', 'sc': 'Polònia', 'sd': 'پولينڊ', 'sd-Arab': 'پولينڊ', 'se': 'Polen', 'seh': 'Polônia', 'ses': 'Poloɲe', 'sg': 'Pölôni', 'shi': 'ⴱⵓⵍⵓⵏⵢⴰ', 'shi-Latn': 'bulunya', 'shi-Tfng': 'ⴱⵓⵍⵓⵏⵢⴰ', 'si': 'පෝලන්තය', 'sk': 'Poľsko', 'sl': 'Poljska', 'smn': 'Puola', 'sn': 'Poland', 'so': 'Booland', 'sq': 'Poloni', 'sr': 'Пољска', 'sr-Cyrl': 'Пољска', 'sr-Latn': 'Poljska', 'sv': 'Polen', 'sw': 'Poland', 'syr': 'ܦܘܠܢܕ', 'szl': 'Polska', 'ta': 'போலந்து', 'te': 'పోలాండ్', 'teo': 'Polandi', 'tg': 'Лаҳистон', 'th': 'โปแลนด์', 'ti': 'ፖላንድ', 'tk': 'Polşa', 'to': 'Polani', 'tr': 'Polonya', 'tt': 'Польша', 'twq': 'Poloɲe', 'tzm': 'Pulunya', 'ug': 'پولشا', 'uk': 'Польща', 'ur': 'پولینڈ', 'uz': 'Polsha', 'uz-Cyrl': 'Польша', 'uz-Latn': 'Polsha', 'vai': 'ꕶꗷꘋ', 'vai-Latn': 'Pólɛŋ', 'vai-Vaii': 'ꕶꗷꘋ', 'vec': 'Polonia', 'vi': 'Ba Lan', 'vun': 'Polandi', 'wae': 'Pole', 'wo': 'Poloñ', 'xh': 'EPoland', 'xnr': 'पोलैंड', 'xog': 'Polandi', 'yav': 'pɔlɔ́ɔny', 'yi': 'פּוילן', 'yo': 'Polandi', 'yrl': 'Puluniya', 'yue': '波蘭', 'yue-Hans': '波兰', 'yue-Hant': '波蘭', 'zgh': 'ⴱⵓⵍⵓⵏⵢⴰ', 'zh': '波兰', 'zh-Hans': '波兰', 'zh-Hant': '波蘭', 'zu': 'i-Poland'},
    'PM': {'af': 'Sint Pierre en Miquelon', 'agq': 'Sɛ̀n Kpiyɛ̀ à Mikelɔŋ', 'ak': 'Saint Pierre ne Miquelon', 'am': 'ሴንት ፒዬር እና ሚኩኤሎን', 'ar': 'سان بيير ومكويلون', 'as': 'ছেইণ্ট পিয়েৰে আৰু মিকিউৱেলন', 'asa': 'Thantapieri na Mikeloni', 'ast': 'Saint Pierre y Miquelon', 'az': 'Müqəddəs Pyer və Mikelon', 'az-Cyrl': 'Мүгәддәс Пјер вә Микелон', 'az-Latn': 'Müqəddəs Pyer və Mikelon', 'bal-Latn': 'St. Péri o Mikwélin', 'bas': 'Nûmpubi Petrò nì Mikèlôn', 'be': 'Сен-П’ер і Мікелон', 'be-tarask': 'Сен-П’ер і Мікелон', 'bez': 'Husantapieri na Humikeloni', 'bg': 'Сен Пиер и Микелон', 'blo': 'Sɛŋ-Petrɔs na Mikelɔŋ', 'bm': 'Piyɛri-Senu-ni-Mikelɔŋ', 'bn': 'সেন্ট পিয়ের ও মিকুয়েলন', 'br': 'Sant-Pêr-ha-Mikelon', 'brx': 'सैन्ट. पिएर आरो मिकैलन', 'bs': 'Sveti Petar i Mikelon', 'bs-Cyrl': 'Сен Пјер и Микелон', 'bs-Latn': 'Sveti Petar i Mikelon', 'ca': 'Saint-Pierre-et-Miquelon', 'ca-ES-valencia': 'Saint-Pierre-et-Miquelon', 'ccp': '𑄥𑄬𑄚𑄴𑄑𑄴 𑄛𑄨𑄠𑄬𑄢𑄴 𑄃𑄮 𑄟𑄨𑄢𑄪𑄠𑄬𑄣𑄧𑄚𑄴', 'ce': 'Сен-Пьер а, Микелон а', 'ceb': 'St. Pierre & Miquelon', 'cgg': 'Senti Piyerre na Mikweron', 'chr': 'ᎤᏓᏅᏘ ᏈᏰ ᎠᎴ ᎻᏇᎶᏂ', 'ckb': 'سەن پیێر و میکێلۆن', 'cs': 'Saint-Pierre a Miquelon', 'cv': 'Сен-Пьер & Микелон', 'cy': 'Saint-Pierre-et-Miquelon', 'da': 'Saint Pierre og Miquelon', 'dav': 'Santapieri na Mikeloni', 'de': 'St. Pierre und Miquelon', 'dje': 'Seŋ Piyer nda Mikelon', 'dsb': 'St. Pierre a Miquelon', 'dz': 'སིནཊ་པི་ཡེར་ ཨེནཌ་ མིཀོ་ལེན', 'ebu': 'Santapieri na Mikeloni', 'ee': 'Saint Pierre kple Mikelɔn nutome', 'el': 'Σεν Πιερ και Μικελόν', 'el-polyton': 'Σαὶντ Πιὲρ καὶ Μικελόν', 'en': 'St. Pierre & Miquelon', 'en-Dsrt': '𐐝𐐩𐑌𐐻 𐐑𐐨𐐯𐑉 𐐰𐑌𐐼 𐐣𐐨𐐿𐐲𐑊𐐪𐑌', 'eo': 'Sankta Piero kaj Mikelono', 'es': 'San Pedro y Miquelón', 'et': 'Saint-Pierre ja Miquelon', 'eu': 'Saint-Pierre eta Mikelune', 'ewo': 'Ǹfúfúb-Píɛr-ai-Mikəlɔ́ŋ', 'fa': 'سن پیر و میکلن', 'ff': 'See Piyeer e Mikeloo', 'ff-Adlm': '𞤅𞤼. 𞤆𞤭𞤴𞤫𞥅𞤪 & 𞤃𞤭𞤳𞤫𞤤𞤮𞤲', 'ff-Latn': 'See Piyeer e Mikeloo', 'fi': 'Saint-Pierre ja Miquelon', 'fil': 'St. Pierre & Miquelon', 'fo': 'Saint Pierre & Miquelon', 'fr': 'Saint-Pierre-et-Miquelon', 'fur': 'San Pierre e Miquelon', 'fy': 'Saint-Pierre en Miquelon', 'ga': 'San Pierre agus Miquelon', 'gd': 'Saint Pierre agus Miquelon', 'gl': 'Saint Pierre et Miquelon', 'gsw': 'St. Pierr und Miggelo', 'gu': 'સેંટ પીએરી અને મિક્યુલોન', 'guz': 'Santapieri na Mikeloni', 'ha': 'San Piyar da Mikelan', 'he': 'סנט פייר ומיקלון', 'hi': 'सेंट पिएरे और मिक्वेलान', 'hi-Latn': 'St. Pierre & Miquelon', 'hr': 'Sveti Petar i Mikelon', 'hsb': 'St. Pierre a Miquelon', 'ht': 'Saint-Pierre-et-Miquelon', 'hu': 'Saint-Pierre és Miquelon', 'hy': 'Սեն Պիեռ և Միքելոն', 'ia': 'St. Pierre e Miquelon', 'id': 'Saint Pierre dan Miquelon', 'ig': 'St. Pierre & Miquelon', 'is': 'Sankti Pierre og Miquelon', 'it': 'Saint-Pierre e Miquelon', 'ja': 'サンピエール島・ミクロン島', 'jmc': 'Santapieri na Mikeloni', 'jv': 'Saint Pièr lan Mikuélon', 'ka': 'სენ-პიერი და მიკელონი', 'kaa': 'Сен-Пьер ҳәм Микелон', 'kaa-Cyrl': 'Сен-Пьер ҳәм Микелон', 'kab': 'San Pyar d Miklun', 'kam': 'Santapieri na Mikeloni', 'kde': 'Santapieli na Mikeloni', 'kea': 'San Piere i Mikelon', 'kgp': 'Sỹ Penru kar Mĩkerỹv', 'khq': 'Seŋ Piyer nda Mikelon', 'ki': 'Santapieri na Mikeloni', 'kk': 'Сен-Пьер және Микелон', 'kk-Cyrl': 'Сен-Пьер және Микелон', 'kln': 'Emetab Peter Ne titil ak Miquelon', 'km': 'សង់ព្យែរ និង\u200bមីគីឡុង', 'kn': 'ಸೇಂಟ್ ಪಿಯರ್ ಮತ್ತು ಮಿಕ್ವೆಲನ್', 'ko': '생피에르 미클롱', 'kok': 'सँ. पायरे आनी मिकेलन', 'kok-Deva': 'सँ. पायरे आनी मिकेलन', 'ks': 'سینٹ پیٖری تہٕ موکیلِیَن', 'ks-Arab': 'سینٹ پیٖری تہٕ موکیلِیَن', 'ksb': 'Santapieli na Mikeloni', 'ksf': 'sɛnpyɛr ri mikɛlɔŋ', 'ksh': 'Zint Pjäär un Mikelong', 'ku': 'Saint-Pierre û Miquelon', 'kxv': 'seint pierri ande mikelon', 'kxv-Deva': 'सेंट पिएरे ऑड़े मिक्वेलान', 'kxv-Latn': 'seint pierri ande mikelon', 'kxv-Orya': 'ସେଣ୍ଟ ପିଏରେ ଅଡ଼େ ମିକ୍ୱେଲାନ', 'kxv-Telu': 'సెయింట్ పియెర్ మరియు మికెలాన్', 'ky': 'Сен-Пьер жана Микелон', 'lag': 'Mʉtakatíifu Peéteri na Mɨkaéeli', 'lb': 'St. Pierre a Miquelon', 'lg': 'Senti Piyere ne Mikeloni', 'lld': 'St. Pierre y Miquelon', 'ln': 'Sántu pététo mpé Mikelɔ', 'lo': 'ເຊນ ປີແອ ມິເກວລອນ', 'lt': 'Sen Pjeras ir Mikelonas', 'lu': 'Santu pététo ne Mikelu', 'luo': 'Saint Pierre gi Miquelon', 'luy': 'Santapieri na Mikeloni', 'lv': 'Senpjēra un Mikelona', 'mai': 'सेंट पियरे आ मिक्वेलान', 'mas': 'Santapieri na Mikeloni', 'mer': 'Santapieri na Mikeloni', 'mfe': 'Saint-Pierre-ek-Miquelon', 'mg': 'Saint-Pierre-et-Miquelon', 'mgh': 'Usantapieri na Mikeloni', 'mi': 'Hato Piere & Mikerona', 'mk': 'Сент Пјер и Микелан', 'ml': 'സെന്റ് പിയറി ആൻഡ് മിക്വലൻ', 'mn': 'Сент-Пьер ба Микело', 'mr': 'सेंट पियरे आणि मिक्वेलोन', 'ms': 'Saint Pierre dan Miquelon', 'mt': 'Saint Pierre u Miquelon', 'mua': 'Sǝr Pǝtar ne Mikǝlon', 'my': 'စိန့်ပီအဲရ်နှင့် မီကွီလွန်', 'mzn': 'سن پییر و میکلن', 'naq': 'Saint Pierre and Miquelon', 'nb': 'Saint-Pierre-et-Miquelon', 'nd': 'Saint Pierre and Miquelon', 'ne': 'सेन्ट पिर्रे र मिक्केलोन', 'nl': 'Saint-Pierre en Miquelon', 'nmg': 'Saint Peter ba Mikelɔn', 'nn': 'Saint-Pierre-et-Miquelon', 'no': 'Saint-Pierre-et-Miquelon', 'nqo': 'ߛߍ߲ߕ-ߔߍ߯ߙ ߣߌ߫ ߡߌ߲ߞߋߟߐ߲߫', 'nyn': 'Senti Piyerre na Mikweron', 'om': 'Ql. Piyeeree fi Mikuyelon', 'or': 'ସେଣ୍ଟ ପିଏରେ ଏବଂ ମିକ୍ୱେଲନ୍\u200c', 'pa': 'ਸੇਂਟ ਪੀਅਰੇ ਐਂਡ ਮਿਕੇਲਨ', 'pa-Guru': 'ਸੇਂਟ ਪੀਅਰੇ ਐਂਡ ਮਿਕੇਲਨ', 'pcm': 'Sent Piẹr & Míkẹlọn', 'pl': 'Saint-Pierre i Miquelon', 'ps': 'سینټ پییر او میکولون', 'pt': 'São Pedro e Miquelão', 'qu': 'San Pedro y Miquelón', 'rm': 'Saint Pierre e Miquelon', 'rn': 'Sempiyeri na Mikeloni', 'ro': 'Saint-Pierre și Miquelon', 'rof': 'Santapieri na Mikeloni', 'ru': 'Сен-Пьер и Микелон', 'rwk': 'Santapieri na Mikeloni', 'saq': 'Santapieri na Mikeloni', 'sbp': 'Santapieli ni Mikeloni', 'sc': 'Santu Predu e Miquelon', 'sd': 'سینٽ پیئر و میڪوئیلون', 'sd-Arab': 'سینٽ پیئر و میڪوئیلون', 'se': 'Saint Pierre ja Miquelon', 'seh': 'Saint Pierre e Miquelon', 'ses': 'Seŋ Piyer nda Mikelon', 'sg': 'Sên-Pyêre na Mikelöon', 'shi': 'ⵙⴰⵏⴱⵢⵉⵔ ⴷ ⵎⵉⴽⵍⵓⵏ', 'shi-Latn': 'sanbyir d miklun', 'shi-Tfng': 'ⵙⴰⵏⴱⵢⵉⵔ ⴷ ⵎⵉⴽⵍⵓⵏ', 'si': 'ශාන්ත පියරේ සහ මැකෝලන්', 'sk': 'Saint Pierre a Miquelon', 'sl': 'Saint Pierre in Miquelon', 'smn': 'St. Pierre já Miquelon', 'sn': 'Saint Pierre and Miquelon', 'so': 'St. Pierre iyo Miquelon', 'sq': 'Shën-Pier dhe Mikelon', 'sr': 'Сен Пјер и Микелон', 'sr-Cyrl': 'Сен Пјер и Микелон', 'sr-Latn': 'Sen Pjer i Mikelon', 'sv': 'S:t Pierre och Miquelon', 'sw': 'Santapierre na Miquelon', 'syr': 'ܣܐܢܬ ܦܝܥܪ ܘܡܩܘܠܘܢ', 'ta': 'செயின்ட் பியர் & மிக்வேலான்', 'te': 'సెయింట్ పియెర్ మరియు మికెలాన్', 'teo': 'Santapieri na Mikeloni', 'tg': 'Сент-Пер ва Микелон', 'th': 'แซงปีแยร์และมีเกอลง', 'ti': 'ቅዱስ ፕየርን ሚከሎንን', 'tk': 'Sen-Pýer we Mikelon', 'to': 'Sā Piea mo Mikeloni', 'tr': 'Saint Pierre ve Miquelon', 'tt': 'Сен-Пьер һәм Микелон', 'twq': 'Seŋ Piyer nda Mikelon', 'tzm': 'Santepyir d Mikelun', 'ug': 'ساينت پىيېر ۋە مىكېلون ئاراللىرى', 'uk': 'Сен-Пʼєр і Мікелон', 'ur': 'سینٹ پیئر اور میکلیئون', 'uz': 'Sen-Pyer va Mikelon', 'uz-Cyrl': 'Сент-Пьер ва Микелон', 'uz-Latn': 'Sen-Pyer va Mikelon', 'vai': 'ꔻꘋ ꔪꘂ ꗪ ꕆꔞꗏꘋ', 'vai-Latn': 'Siŋ Piiyɛ ɓɛ́ Mikelɔŋ', 'vai-Vaii': 'ꔻꘋ ꔪꘂ ꗪ ꕆꔞꗏꘋ', 'vec': 'S. Piero e Michelon', 'vi': 'Saint Pierre và Miquelon', 'vun': 'Santapieri na Mikeloni', 'wae': 'St. Pierre und Miquelon', 'wo': 'Saŋ Peer ak Mikeloŋ', 'xh': 'ESt. Pierre & Miquelon', 'xnr': 'सेंट पिएरे कने मिक्वेलान', 'xog': 'Senti Piyere ni Mikeloni', 'yav': 'sɛ́ŋpiɛ́l e mikelɔ́ŋ', 'yo': 'Pẹẹri ati mikuloni', 'yrl': 'Sã Peduru asuí Mikelãu', 'yue': '聖皮埃爾同密克隆群島', 'yue-Hans': '圣皮埃尔同密克隆群岛', 'yue-Hant': '聖皮埃爾同密克隆群島', 'zgh': 'ⵙⴰⵏⴱⵢⵉⵔ ⴷ ⵎⵉⴽⵍⵓⵏ', 'zh': '圣皮埃尔和密克隆群岛', 'zh-Hans': '圣皮埃尔和密克隆群岛', 'zh-Hant': '聖皮埃與密克隆群島', 'zu': 'i-Saint Pierre kanye ne-Miquelon'},
    'PN': {'af': 'Pitcairneilande', 'agq': 'Kpitɨ̀kalè', 'ak': 'Pitkaan Nsupɔ', 'am': 'ፒትካኢርን ደሴቶች', 'ar': 'جزر بيتكيرن', 'as': 'পিটকেইৰ্ণ দ্বীপপুঞ্জ', 'asa': 'Pitkairni', 'ast': 'Islles Pitcairn', 'az': 'Pitkern adaları', 'az-Cyrl': 'Питкерн адалары', 'az-Latn': 'Pitkern adaları', 'bal-Latn': 'Pitkarén Islánd', 'bas': 'Pìdkaìrn', 'be': 'Астравы Піткэрн', 'be-tarask': 'Астравы Піткэрн', 'bez': 'Hupitkaini', 'bg': 'Острови Питкерн', 'blo': 'Pɩtkɛɛn kaBʊtukǝltǝna', 'bm': 'Pitikarini', 'bn': 'পিটকেয়ার্ন দ্বীপপুঞ্জ', 'br': 'Enez Pitcairn', 'brx': 'पिटकाईर्न द्वीप', 'bs': 'Pitkernska Ostrva', 'bs-Cyrl': 'Питкерн', 'bs-Latn': 'Pitkernska Ostrva', 'ca': 'Illes Pitcairn', 'ca-ES-valencia': 'Illes Pitcairn', 'ccp': '𑄛𑄨𑄇𑄴𑄇𑄬𑄠𑄢𑄴𑄚𑄴 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Питкэрн гӀайренаш', 'ceb': 'Pitcairn Islands', 'cgg': 'Pitkaini', 'chr': 'ᏈᎧᎵᏂ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی پیتکەرن', 'cs': 'Pitcairnovy ostrovy', 'cv': 'Питкэрн утравӗсем', 'cy': 'Ynysoedd Pitcairn', 'da': 'Pitcairn', 'dav': 'Pitkairni', 'de': 'Pitcairninseln', 'dje': 'Pitikarin', 'dsb': 'Pitcairnowe kupy', 'dz': 'པིཊ་ཀེ་ཡེརན་གླིང་ཚོམ', 'ebu': 'Pitkairni', 'ee': 'Pitkairn ƒudomekpo nutome', 'el': 'Νήσοι Πίτκερν', 'el-polyton': 'Νήσοι Πίτκερν', 'en': 'Pitcairn Islands', 'en-Dsrt': '𐐑𐐮𐐻𐐿𐐯𐑉𐑌', 'eo': 'Pitkarna Insulo', 'es': 'Islas Pitcairn', 'et': 'Pitcairni saared', 'eu': 'Pitcairn uharteak', 'ewo': 'Pítə́kɛ́rɛnə', 'fa': 'جزایر پیت\u200cکرن', 'ff': 'Pitkern', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤆𞤭𞤼𞤳𞤭𞥅𞤪𞤲𞤵', 'ff-Latn': 'Pitkern', 'fi': 'Pitcairn', 'fil': 'Pitcairn Islands', 'fo': 'Pitcairnoyggjar', 'fr': 'Îles Pitcairn', 'fur': 'Pitcairn', 'fy': 'Pitcairneilannen', 'ga': 'Oileáin Pitcairn', 'gd': 'Eileanan Pheit a’ Chàirn', 'gl': 'Illas Pitcairn', 'gsw': 'Pitggäärn', 'gu': 'પીટકૈર્ન આઇલેન્ડ્સ', 'guz': 'Pitkairni', 'ha': 'Tsibiran Pitcairn', 'he': 'איי פיטקרן', 'hi': 'पिटकैर्न द्वीपसमूह', 'hi-Latn': 'Pitcairn Islands', 'hr': 'Pitcairnovi Otoci', 'hsb': 'Pitcairnowe kupy', 'ht': 'Îles Pitcairn', 'hu': 'Pitcairn-szigetek', 'hy': 'Պիտկեռն կղզիներ', 'ia': 'Insulas Pitcairn', 'id': 'Kepulauan Pitcairn', 'ig': 'Agwaetiti Pitcairn', 'is': 'Pitcairn-eyjar', 'it': 'Isole Pitcairn', 'ja': 'ピトケアン諸島', 'jmc': 'Pitkairni', 'jv': 'Kapuloan Pitcairn', 'ka': 'პიტკერნის კუნძულები', 'kaa': 'Питкэрн атаўлары', 'kaa-Cyrl': 'Питкэрн атаўлары', 'kab': 'Pitkarin', 'kam': 'Pitkairni', 'kde': 'Pitikeluni', 'kea': 'Ilhas Pitkairn', 'kgp': 'Pinkajir Goj-vẽso', 'khq': 'Pitikarin', 'ki': 'Pitkairni', 'kk': 'Питкэрн аралдары', 'kk-Cyrl': 'Питкэрн аралдары', 'kln': 'Emetab Pitcairn', 'km': 'កោះ\u200bភីតកាន', 'kn': 'ಪಿಟ್\u200cಕೈರ್ನ್ ದ್ವೀಪಗಳು', 'ko': '핏케언 제도', 'kok': 'पिटकॅरन आयलँड्स', 'kok-Deva': 'पिटकॅरन आयलँड्स', 'ks': 'پِٹکیرٕنؠ جٔزیٖرٕ', 'ks-Arab': 'پِٹکیرٕنؠ جٔزیٖرٕ', 'ksb': 'Pitkailni', 'ksf': 'pitkɛ́n', 'ksh': 'de Pitkärn-Ensel', 'ku': 'Giravên Pitcairnê', 'kxv': 'piṭkarn dīp', 'kxv-Deva': 'पिटकेर्न द्वीप', 'kxv-Latn': 'piṭkarn dīp', 'kxv-Orya': 'ପିଟ୍\u200dକଇର୍ନ୍ ଦିପ', 'kxv-Telu': 'పిట్\u200cకెయిర్న్ దీవులు', 'ky': 'Питкэрн аралдары', 'lag': 'Patikaírini', 'lb': 'Pitcairninselen', 'lg': 'Pitikeeni', 'lld': 'Isoles Pitcairn', 'ln': 'Pikairni', 'lo': 'ໝູ່ເກາະພິດແຄນ', 'lt': 'Pitkerno salos', 'lu': 'Pikairni', 'luo': 'Pitcairn', 'luy': 'Pitkairni', 'lv': 'Pitkērnas salas', 'mai': 'पिटकेयर्न द्वीपसमूह', 'mas': 'Pitkairni', 'mer': 'Pitkairni', 'mfe': 'Pitcairn', 'mg': 'Pitkairn', 'mgh': 'Upitkairni', 'mi': 'Pitikeina', 'mk': 'Питкернски Острови', 'ml': 'പിറ്റ്\u200cകെയ്\u200cൻ ദ്വീപുകൾ', 'mn': 'Питкэрн арлууд', 'mr': 'पिटकैर्न बेटे', 'ms': 'Kepulauan Pitcairn', 'mt': 'Gżejjer Pitcairn', 'mua': 'Pitkairn', 'my': 'ပစ်တ်ကိန်းကျွန်းစု', 'mzn': 'پیتکارین جزایر', 'naq': 'Pitcairn', 'nb': 'Pitcairnøyene', 'nd': 'Pitcairn', 'ne': 'पिटकाइर्न टापुहरु', 'nl': 'Pitcairneilanden', 'nmg': 'Pitkairn', 'nn': 'Pitcairn', 'no': 'Pitcairnøyene', 'nqo': 'ߔߌߕߑߞߍ߲ ߕߌ߲', 'nyn': 'Pitkaini', 'om': 'Odoloota Pitikaayirin', 'or': 'ପିଟକାଇରିନ୍\u200c ଦ୍ୱୀପପୁଞ୍ଜ', 'pa': 'ਪਿਟਕੇਰਨ ਟਾਪੂ', 'pa-Guru': 'ਪਿਟਕੇਰਨ ਟਾਪੂ', 'pcm': 'Pítkén Aílands', 'pl': 'Pitcairn', 'ps': 'پيټکيرن ټاپوګان', 'pt': 'Ilhas Pitcairn', 'qu': 'Islas Pitcairn', 'rm': 'Pitcairn', 'rn': 'Pitikeyirini', 'ro': 'Insulele Pitcairn', 'rof': 'Pitkairni', 'ru': 'о-ва Питкэрн', 'rwk': 'Pitkairni', 'saq': 'Pitkairni', 'sbp': 'Pitikailini', 'sc': 'Ìsulas Pìtcairn', 'sd': 'پٽڪئرن ٻيٽ', 'sd-Arab': 'پٽڪئرن ٻيٽ', 'se': 'Pitcairn', 'seh': 'Pitcairn', 'ses': 'Pitikarin', 'sg': 'Pitikêrni', 'shi': 'ⴱⵉⵜⴽⴰⵢⵔⵏ', 'shi-Latn': 'bitkayrn', 'shi-Tfng': 'ⴱⵉⵜⴽⴰⵢⵔⵏ', 'si': 'පිට්කෙය්න් දූපත්', 'sk': 'Pitcairnove ostrovy', 'sl': 'Pitcairn', 'smn': 'Pitcairn', 'sn': 'Pitcairn', 'so': 'Bitkairn', 'sq': 'Ishujt Pitkern', 'sr': 'Питкерн', 'sr-Cyrl': 'Питкерн', 'sr-Latn': 'Pitkern', 'sv': 'Pitcairnöarna', 'sw': 'Visiwa vya Pitcairn', 'ta': 'பிட்கெய்ர்ன் தீவுகள்', 'te': 'పిట్\u200cకెయిర్న్ దీవులు', 'teo': 'Pitkairni', 'tg': 'Ҷазираҳои Питкейрн', 'th': 'หมู่เกาะพิตแคร์น', 'ti': 'ደሴታት ፒትካርን', 'tk': 'Pitkern adalary', 'to': 'ʻOtumotu Pitikeni', 'tr': 'Pitcairn Adaları', 'tt': 'Питкэрн утраулары', 'twq': 'Pitikarin', 'tzm': 'Pitkirn', 'ug': 'پىتكايرن ئاراللىرى', 'uk': 'Острови Піткерн', 'ur': 'پٹکائرن جزائر', 'uz': 'Pitkern orollari', 'uz-Cyrl': 'Питкэрн ороллари', 'uz-Latn': 'Pitkern orollari', 'vai': 'ꔪꔳꕪꕆ', 'vai-Latn': 'Pitikɛŋ', 'vai-Vaii': 'ꔪꔳꕪꕆ', 'vec': 'Ìzole Pitcairn', 'vi': 'Quần đảo Pitcairn', 'vun': 'Pitkairni', 'wae': 'Pitcairn', 'wo': 'Duni Pitkayirn', 'xh': 'EPitcairn Islands', 'xnr': 'पिटकैर्न द्वीपसमूह', 'xog': 'Pitikeeni', 'yav': 'pitikɛ́ɛlínɛ́', 'yi': 'פּיטקערן אינזלען', 'yo': 'Pikarini', 'yrl': 'Kapuãma-ita Pitcairn', 'yue': '皮特肯群島', 'yue-Hans': '皮特肯群岛', 'yue-Hant': '皮特肯群島', 'zgh': 'ⴱⵉⵜⴽⴰⵢⵔⵏ', 'zh': '皮特凯恩群岛', 'zh-Hans': '皮特凯恩群岛', 'zh-Hant': '皮特肯群島', 'zu': 'i-Pitcairn Islands'},
    'PR': {'af': 'Puerto Rico', 'agq': 'Kpǒto Leko', 'ak': 'Puɛto Riko', 'am': 'ፑዌርቶ ሪኮ', 'ar': 'بورتوريكو', 'as': 'পুৱেৰ্টো ৰিকো', 'asa': 'Pwetoriko', 'ast': 'Puertu Ricu', 'az': 'Puerto Riko', 'az-Cyrl': 'Пуерто Рико', 'az-Latn': 'Puerto Riko', 'bal-Latn': 'Piuró Rikó', 'bas': 'Pɔ̀rtò Rikò', 'be': 'Пуэрта-Рыка', 'be-tarask': 'Пуэрта-Рыка', 'bez': 'Hupwetoriko', 'bg': 'Пуерто Рико', 'blo': 'Pɔrto Rikoo', 'bm': 'Pɔrotoriko', 'bn': 'পুয়ের্তো রিকো', 'br': 'Puerto Rico', 'brx': 'पुवेर्ट’ रीक’', 'bs': 'Porto Riko', 'bs-Cyrl': 'Порторико', 'bs-Latn': 'Porto Riko', 'ca': 'Puerto Rico', 'ca-ES-valencia': 'Puerto Rico', 'ccp': '𑄛𑄪𑄠𑄬𑄢𑄴𑄖𑄮 𑄢𑄨𑄇𑄮', 'ce': 'Пуэрто-Рико', 'ceb': 'Puerto Rico', 'cgg': 'Pwetoriko', 'chr': 'ᏇᎡᏙ ᎵᎢᎪ', 'ckb': 'پۆرتۆڕیکۆ', 'cs': 'Portoriko', 'cv': 'Пуэрто-Рико', 'cy': 'Puerto Rico', 'da': 'Puerto Rico', 'dav': 'Pwetoriko', 'de': 'Puerto Rico', 'dje': 'Porto Riko', 'dsb': 'Puerto Rico', 'dz': 'པུ་འེར་ཊོ་རི་ཁོ', 'ebu': 'Pwetoriko', 'ee': 'Puerto Riko nutome', 'el': 'Πουέρτο Ρίκο', 'el-polyton': 'Πουέρτο Ρίκο', 'en': 'Puerto Rico', 'en-Dsrt': '𐐑𐐶𐐯𐑉𐐻𐐬 𐐡𐐨𐐿𐐬', 'eo': 'Puertoriko', 'es': 'Puerto Rico', 'et': 'Puerto Rico', 'eu': 'Puerto Rico', 'ewo': 'Pwɛrəto Ríko', 'fa': 'پورتوریکو', 'ff': 'Porto Rikoo', 'ff-Adlm': '𞤆𞤮𞤪𞤼𞤮 𞤈𞤭𞤳𞤮𞥅', 'ff-Latn': 'Porto Rikoo', 'fi': 'Puerto Rico', 'fil': 'Puerto Rico', 'fo': 'Puerto Riko', 'fr': 'Porto Rico', 'fur': 'Porto Rico', 'fy': 'Puerto Rico', 'ga': 'Pórtó Ríce', 'gd': 'Porto Rìceo', 'gl': 'Porto Rico', 'gsw': 'Puerto Riggo', 'gu': 'પ્યુઅર્ટો રિકો', 'guz': 'Pwetoriko', 'ha': 'Porto Riko', 'he': 'פוארטו ריקו', 'hi': 'पोर्टो रिको', 'hi-Latn': 'Puerto Rico', 'hr': 'Portoriko', 'hsb': 'Puerto Rico', 'ht': 'Porto Rico', 'hu': 'Puerto Rico', 'hy': 'Պուերտո Ռիկո', 'ia': 'Porto Rico', 'id': 'Puerto Riko', 'ie': 'Porto-Rico', 'ig': 'Puerto Rico', 'is': 'Púertó Ríkó', 'it': 'Portorico', 'ja': 'プエルトリコ', 'jmc': 'Pwetoriko', 'jv': 'Puèrto Riko', 'ka': 'პუერტო-რიკო', 'kaa': 'Пуэрто-Рико', 'kaa-Cyrl': 'Пуэрто-Рико', 'kab': 'Purtu Riku', 'kam': 'Pwetoriko', 'kde': 'Pwetoliko', 'kea': 'Portu Riku', 'kgp': 'Portu Hiku', 'khq': 'Porto Riko', 'ki': 'Pwetoriko', 'kk': 'Пуэрто-Рико', 'kk-Cyrl': 'Пуэрто-Рико', 'kln': 'Emetab Puerto Rico', 'km': 'ព័រតូរីកូ', 'kn': 'ಪ್ಯೂರ್ಟೋ ರಿಕೊ', 'ko': '푸에르토리코', 'kok': 'प्युएर्तो रिको', 'kok-Deva': 'प्युएर्तो रिको', 'ks': 'پٔرٹو رِکو', 'ks-Arab': 'پٔرٹو رِکو', 'ksb': 'Pwetoliko', 'ksf': 'pɔtoríko', 'ksh': 'Pochtoriko', 'ku': 'Porto Rîko', 'kxv': 'puerto rico', 'kxv-Deva': 'पोर्टो रिको', 'kxv-Latn': 'puerto rico', 'kxv-Orya': 'ପର୍ଟ ରିକ', 'kxv-Telu': 'ప్యూర్టో రికో', 'ky': 'Пуэрто-Рико', 'lag': 'Pwetorɨ́ɨko', 'lb': 'Puerto Rico', 'lg': 'Potoriko', 'lld': 'Puerto Rico', 'ln': 'Pɔtoriko', 'lo': 'ເພືອໂຕ ຣິໂກ', 'lt': 'Puerto Rikas', 'lu': 'Mpotoriku', 'luo': 'Puerto Rico', 'luy': 'Pwetoriko', 'lv': 'Puertoriko', 'mai': 'प्योर्टो रिको', 'mas': 'Pwetoriko', 'mer': 'Pwetoriko', 'mfe': 'Porto Rico', 'mg': 'Pôrtô Rikô', 'mgh': 'Upwetoriko', 'mi': 'Peta Riko', 'mk': 'Порторико', 'ml': 'പോർട്ടോ റിക്കോ', 'mn': 'Пуэрто-Рико', 'mr': 'प्युएर्तो रिको', 'ms': 'Puerto Rico', 'mt': 'Puerto Rico', 'mua': 'Porto Riko', 'my': 'ပေါ်တိုရီကို', 'mzn': 'پورتوریکو', 'naq': 'Puerto Rico', 'nb': 'Puerto Rico', 'nd': 'Puerto Rico', 'ne': 'पुएर्टो रिको', 'nl': 'Puerto Rico', 'nmg': 'Puɛrto Riko', 'nn': 'Puerto Rico', 'no': 'Puerto Rico', 'nqo': 'ߔߐߙߑߕߏ߫-ߙߌߞߏ߫', 'nyn': 'Pwetoriko', 'om': 'Poortaar Riikoo', 'or': 'ପୁଏର୍ତ୍ତୋ ରିକୋ', 'pa': 'ਪਿਊਰਟੋ ਰਿਕੋ', 'pa-Guru': 'ਪਿਊਰਟੋ ਰਿਕੋ', 'pcm': 'Puẹ́rto Ríkọ', 'pl': 'Portoryko', 'ps': 'پورتو ریکو', 'pt': 'Porto Rico', 'qu': 'Puerto Rico', 'rm': 'Puerto Rico', 'rn': 'Puwetoriko', 'ro': 'Puerto Rico', 'rof': 'Pwetoriko', 'ru': 'Пуэрто-Рико', 'rwk': 'Pwetoriko', 'saq': 'Pwetoriko', 'sbp': 'Pwetoliko', 'sc': 'Puerto Rico', 'sd': 'پيوئرٽو ريڪو', 'sd-Arab': 'پيوئرٽو ريڪو', 'se': 'Puerto Rico', 'seh': 'Porto Rico', 'ses': 'Porto Riko', 'sg': 'Porto Rîko', 'shi': 'ⴱⵓⵔⵜⵓ ⵔⵉⴽⵓ', 'shi-Latn': 'burtu riku', 'shi-Tfng': 'ⴱⵓⵔⵜⵓ ⵔⵉⴽⵓ', 'si': 'පුවර්ටෝ රිකෝ', 'sk': 'Portoriko', 'sl': 'Portoriko', 'smn': 'Puerto Rico', 'sn': 'Puerto Rico', 'so': 'Bueerto Riiko', 'sq': 'Porto-Riko', 'sr': 'Порторико', 'sr-Cyrl': 'Порторико', 'sr-Latn': 'Portoriko', 'sv': 'Puerto Rico', 'sw': 'Puerto Rico', 'syr': 'ܦܘܐܪܛܘ ܪܝܩܘ', 'ta': 'பியூர்டோ ரிகோ', 'te': 'ప్యూర్టో రికో', 'teo': 'Pwetoriko', 'tg': 'Пуэрто-Рико', 'th': 'เปอร์โตริโก', 'ti': 'ፖርቶ ሪኮ', 'tk': 'Puerto-Riko', 'to': 'Puēto Liko', 'tr': 'Porto Riko', 'tt': 'Пуэрто-Рико', 'twq': 'Porto Riko', 'tzm': 'Purturiku', 'ug': 'پۇئېرتو رىكو', 'uk': 'Пуерто-Рико', 'ur': 'پیورٹو ریکو', 'uz': 'Puerto-Riko', 'uz-Cyrl': 'Пуэрто-Рико', 'uz-Latn': 'Puerto-Riko', 'vai': 'ꔪꖳꕿ ꔸꖏ', 'vai-Latn': 'Piyuto Riko', 'vai-Vaii': 'ꔪꖳꕿ ꔸꖏ', 'vec': 'Portorico', 'vi': 'Puerto Rico', 'vun': 'Pwetoriko', 'wae': 'Puerto Rico', 'wo': 'Porto Riko', 'xh': 'EPuerto Rico', 'xnr': 'पोर्टो रिको', 'xog': 'Potoriko', 'yav': 'pólótolíko', 'yi': 'פּארטא־ריקא', 'yo': 'Pọto Riko', 'yrl': 'Igarapawa Riku', 'yue': '波多黎各', 'yue-Hans': '波多黎各', 'yue-Hant': '波多黎各', 'zgh': 'ⴱⵓⵔⵜⵓ ⵔⵉⴽⵓ', 'zh': '波多黎各', 'zh-Hans': '波多黎各', 'zh-Hant': '波多黎各', 'zu': 'i-Puerto Rico'},
    'PS': {'af': 'Palestynse Grondgebiede', 'agq': 'Adzɨmā kɨ ŋgùŋ kɨ Palɛsɨtɨnyia à kɨ Gazà kò', 'ak': 'Palestaen West Bank ne Gaza', 'am': 'የፍልስጤም ግዛት', 'ar': 'الأراضي الفلسطينية', 'as': 'ফিলিস্তিন অঞ্চল', 'asa': 'Palestina', 'ast': 'Territorios Palestinos', 'az': 'Fələstin Əraziləri', 'az-Latn': 'Fələstin Əraziləri', 'bal-Latn': 'Palastinay Damag', 'bas': 'Pàlɛ̀htinà Hyɔ̀ŋg nì Gazà', 'be': 'Палесцінскія Тэрыторыі', 'be-tarask': 'Палесцінскія Тэрыторыі', 'bez': 'Ulubali lwa Magharibi nu Gaza wa Hupalestina', 'bg': 'Палестински територии', 'blo': 'Palɛstiin kAsàʊ', 'bm': 'Palesitini', 'bn': 'প্যালেস্টাইন ভূখণ্ড', 'br': 'Tiriadoù Palestina', 'brx': 'पेलेस्ताइननि ओनसोलफोर', 'bs': 'Palestinska Teritorija', 'bs-Cyrl': 'Палестинске територије', 'bs-Latn': 'Palestinska Teritorija', 'ca': 'Territoris palestins', 'ca-ES-valencia': 'Territoris palestins', 'ccp': '𑄜𑄨𑄣𑄨𑄌𑄴𑄖𑄨𑄚𑄴 𑄎𑄉𑄊𑄚𑄨', 'ce': 'ПалестӀинан латтанаш', 'ceb': 'Palestinian Territories', 'chr': 'ᏆᎴᏍᏗᏂᎠᏂ ᏄᎬᏫᏳᏌᏕᎩ', 'ckb': 'ناوچە فەلەستینیەکان', 'cs': 'Palestinská území', 'cv': 'Палестинӑн территорийӗсем', 'cy': 'Tiriogaethau Palesteinaidd', 'da': 'De palæstinensiske områder', 'dav': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'de': 'Palästinensische Autonomiegebiete', 'dje': 'Palestine Dangay nda Gaaza', 'dsb': 'Palestinski awtonomny teritorium', 'dz': 'པེ་ལིསི་ཊི་ནི་ཡན་ཊེ་རི་ཐོ་རི', 'ebu': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'ee': 'Palestinia nutome', 'el': 'Παλαιστινιακά Εδάφη', 'el-polyton': 'Παλαιστινιακὰ Ἐδάφη', 'en': 'Palestinian Territories', 'en-Dsrt': '𐐑𐐰𐑊𐐲𐑅𐐻𐐮𐑌𐐨𐐲𐑌 𐐓𐐯𐑉𐐲𐐻𐐱𐑉𐐨', 'es': 'Territorios Palestinos', 'et': 'Palestiina alad', 'eu': 'Palestinar Lurralde Okupatuak', 'ewo': 'Ǹnam Palɛsətín', 'fa': 'سرزمین\u200cهای فلسطینی', 'ff': 'Palestiin Sisjordani e Gaasaa', 'ff-Adlm': '𞤂𞤫𞤧𞤣𞤭𞥅𞤶𞤭 𞤊𞤢𞤤𞤫𞤧𞤼𞤭𞥅𞤲', 'ff-Latn': 'Palestiin Sisjordani e Gaasaa', 'fi': 'Palestiinalaisalue', 'fil': 'Palestinian Territories', 'fo': 'Palestinskt landøki', 'fr': 'Territoires palestiniens', 'fur': 'Teritoris palestinês', 'fy': 'Palestynske gebieten', 'ga': 'na Críocha Palaistíneacha', 'gd': 'Ùghdarras nam Palastaineach', 'gl': 'Territorios Palestinos', 'gsw': 'Paläschtinänsischi Gebiet', 'gu': 'પેલેસ્ટિનિયન ટેરિટરી', 'guz': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'ha': 'Yankunan Palasɗinu', 'he': 'השטחים הפלסטיניים', 'hi': 'फ़िलिस्तीनी क्षेत्र', 'hi-Latn': 'Palestinian Territories', 'hr': 'Palestinsko područje', 'hsb': 'Palestinski awtonomny teritorij', 'ht': 'Territoires palestiniens', 'hu': 'Palesztin Autonómia', 'hy': 'Պաղեստինյան տարածքներ', 'ia': 'Territorios palestin', 'id': 'Wilayah Palestina', 'ig': 'Mpaghara ndị Palestine', 'is': 'Heimastjórnarsvæði Palestínumanna', 'it': 'Territori Palestinesi', 'ja': 'パレスチナ自治区', 'jmc': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'jv': 'Tlatah Palèstina', 'ka': 'პალესტინის ტერიტორიები', 'kaa': 'Палестина территориялары', 'kaa-Cyrl': 'Палестина территориялары', 'kab': 'Falisṭin d Ɣezza', 'kam': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'kde': 'Nchingu wa Magalibi wa Mpanda wa kuGaza wa kuPales', 'kea': 'Tiritóriu palistinianu', 'kgp': 'Tehitórijo Parenhtinũ', 'khq': 'Palestine Dangay nda Gaaza', 'ki': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'kk': 'Палестина аймақтары', 'kk-Cyrl': 'Палестина аймақтары', 'kln': 'Emetab Palestine', 'km': 'ដែនដីប៉ាឡេស្ទីន', 'kn': 'ಪ್ಯಾಲೇಸ್ಟೇನಿಯನ್ ಪ್ರದೇಶಗಳು', 'ko': '팔레스타인 지구', 'kok': 'पेलेस्टीनियन प्रांत', 'kok-Deva': 'पेलेस्टीनियन प्रांत', 'ks': 'فلسطینی علاقٕہ', 'ks-Arab': 'فلسطینی علاقٕہ', 'ksb': 'Ukingo wa Maghalibi na Ukanda wa Gaza wa Palestina', 'ksf': 'zǝ palɛstínǝ', 'ksh': 'Palästina', 'ku': 'Herêmên Filîstînî', 'kxv': 'palesṭiati handi', 'kxv-Deva': 'प़िलिस्तीनी क्षेत्र', 'kxv-Latn': 'palesṭiati handi', 'kxv-Orya': 'ପାଲେସ୍ଟିଆତି ହାନ୍ଦି', 'kxv-Telu': 'పాలస్తీనియన్ ప్రాంతాలు', 'ky': 'Палестина аймактары', 'lag': 'Mweemberera wa kʉmweeri wa Gáaza', 'lb': 'Palestinensesch Autonomiegebidder', 'lg': 'Palesitayini', 'lld': 'Teritori palestinesc', 'ln': 'Palɛsine', 'lo': 'ດິນແດນປາເລດສະຕິນ', 'lt': 'Palestinos teritorija', 'lu': 'Palesine', 'luo': 'Palestinian West Bank gi Gaza', 'luy': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'lv': 'Palestīnas teritorijas', 'mai': 'फिलिस्तीनी क्षेत्र', 'mas': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'mer': 'Rũtere rwa Westi banki na Gaza cia Palestina', 'mfe': 'Teritwar Palestinn', 'mg': 'Palestina', 'mi': 'Ngā Rohe o Parihitini', 'mk': 'Палестински Територии', 'ml': 'പാലസ്\u200cതീൻ പ്രദേശങ്ങൾ', 'mn': 'Палестины нутаг дэвсгэр', 'mr': 'पॅलेस्टिनियन प्रदेश', 'ms': 'Wilayah Palestin', 'mt': 'it-Territorji Palestinjani', 'mua': 'Sǝr Palestiniya', 'my': 'ပါလက်စတိုင်း ပိုင်နက်', 'mzn': 'فلسطین ِسرزمین', 'naq': 'Palestinian West Bank and Gaza', 'nb': 'Det palestinske området', 'nd': 'Palestinian West Bank and Gaza', 'ne': 'प्यालेस्टिनी भू-भागहरू', 'nl': 'Palestijnse gebieden', 'nmg': 'Palɛstin', 'nn': 'Det palestinske området', 'no': 'Det palestinske området', 'nqo': 'ߔߊߟߍߛߑߕߌߣ ߞߣߍ', 'nyn': 'Parestiina na Gaza', 'om': 'Daangaawwan Paalestaayin', 'or': 'ପାଲେଷ୍ଟେନିୟ ଭୂଭାଗ', 'pa': 'ਫਿਲੀਸਤੀਨੀ ਇਲਾਕਾ', 'pa-Guru': 'ਫਿਲੀਸਤੀਨੀ ਇਲਾਕਾ', 'pcm': 'Pálẹ́staín Éria-dẹm', 'pl': 'Terytoria Palestyńskie', 'ps': 'فلسطیني سيمې', 'pt': 'Territórios palestinos', 'qu': 'Palestina Kamachikuq', 'rm': 'Territori Palestinais', 'rn': 'Palesitina Wesitibanka na Gaza', 'ro': 'Teritoriile Palestiniene', 'rof': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'ru': 'Палестинские территории', 'rwk': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'saq': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'sbp': 'Munjema gwa Kusikha nu Luvala lwa Gasa lwa Palesit', 'sc': 'Territòrios palestinesos', 'sd': 'فلسطيني علائقا', 'sd-Arab': 'فلسطيني علائقا', 'se': 'Palestina', 'seh': 'Território da Palestina', 'ses': 'Palestine Dangay nda Gaaza', 'sg': 'Sêse tî Palestîni', 'shi': 'ⴰⴳⵎⵎⴰⴹ ⵏ ⵜⴰⴳⵓⵜ ⴷ ⵖⵣⵣⴰ', 'shi-Latn': 'agmmaḍ n tagut d ɣzza', 'shi-Tfng': 'ⴰⴳⵎⵎⴰⴹ ⵏ ⵜⴰⴳⵓⵜ ⴷ ⵖⵣⵣⴰ', 'si': 'පලස්තීන රාජ්\u200dයය', 'sk': 'Palestínske územia', 'sl': 'Palestinsko ozemlje', 'so': 'Dhulka Falastiiniyiinta daanta galbeed iyo marinka qasa', 'sq': 'Territoret Palestineze', 'sr': 'Палестинске територије', 'sr-Cyrl': 'Палестинске територије', 'sr-Latn': 'Palestinske teritorije', 'sv': 'Palestinska territorierna', 'sw': 'Maeneo ya Palestina', 'syr': 'ܐܬܖ̈ܘܬܐ ܕܦܠܣܛܝܢ', 'ta': 'பாலஸ்தீனிய பிரதேசங்கள்', 'te': 'పాలస్తీనియన్ ప్రాంతాలు', 'teo': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'tg': 'Қаламравҳои Фаластинӣ', 'th': 'ดินแดนปาเลสไตน์', 'ti': 'ግዝኣታት ፍልስጤም', 'tk': 'Palestina territoriýasy', 'to': 'Potu Palesitaine', 'tr': 'Filistin Bölgeleri', 'tt': 'Фәләстин территорияләре', 'twq': 'Palestine Dangay nda Gaaza', 'tzm': 'Agemmaḍ Ugut d Ɣazza Ifilisṭiniyen', 'ug': 'پەلەستىن زېمىنى', 'uk': 'Палестинські території', 'ur': 'فلسطینی خطے', 'uz': 'Falastin hududlari', 'uz-Cyrl': 'Фаластин ҳудуди', 'uz-Latn': 'Falastin hududlari', 'vai': 'ꕐꔒꔻꔳꕯ ꔎꔒ ꕀꔤ ꗛꔤ ꕞ ꗱ ꗪ ꕭꕌꕤ', 'vai-Latn': 'Palesitininya Tele Jii Kɔiyɛ lá hĩ Gaza', 'vai-Vaii': 'ꕐꔒꔻꔳꕯ ꔎꔒ ꕀꔤ ꗛꔤ ꕞ ꗱ ꗪ ꕭꕌꕤ', 'vec': 'Teritori palestinezi', 'vi': 'Lãnh thổ Palestine', 'vun': 'Ukingo wa Magharibi na Ukanda wa Gaza wa Palestina', 'wae': 'Paleština', 'wo': 'réew yu Palestine', 'xh': 'IPalestinian Territories', 'xnr': 'फ़िलिस्तीनी लाका', 'xog': 'Palesitayini ni Gaza', 'yav': 'kitɔŋ ki palɛstíin', 'yo': 'Agbègbè ara Palẹsítínì', 'yrl': 'Tetãma Paretinu-ita yara', 'yue': '巴勒斯坦自治區', 'yue-Hans': '巴勒斯坦自治区', 'yue-Hant': '巴勒斯坦自治區', 'zgh': 'ⴰⴳⵎⵎⴰⴹ ⵏ ⵜⴰⴳⵓⵜ ⴷ ⵖⵣⵣⴰ', 'zh': '巴勒斯坦领土', 'zh-Hans': '巴勒斯坦领土', 'zh-Hant': '巴勒斯坦自治區', 'zu': 'i-Palestinian Territories'},
    'PT': {'af': 'Portugal', 'agq': 'Kputuwgà', 'ak': 'Pɔtugal', 'am': 'ፖርቱጋል', 'ar': 'البرتغال', 'as': 'পৰ্তুগাল', 'asa': 'Ureno', 'ast': 'Portugal', 'az': 'Portuqaliya', 'az-Cyrl': 'Португалија', 'az-Latn': 'Portuqaliya', 'bal-Latn': 'Portogál', 'bas': 'Pɔ̀tɔkì', 'be': 'Партугалія', 'be-tarask': 'Партугалія', 'bez': 'Huuleno', 'bg': 'Португалия', 'bgn': 'پورتگال', 'blo': 'Pɔrtigal', 'bm': 'Pɔritigali', 'bn': 'পর্তুগাল', 'br': 'Portugal', 'brx': 'पर्तुगाल', 'bs': 'Portugal', 'bs-Cyrl': 'Португал', 'bs-Latn': 'Portugal', 'ca': 'Portugal', 'ca-ES-valencia': 'Portugal', 'ccp': '𑄛𑄧𑄢𑄴𑄖𑄪𑄉𑄣𑄴', 'ce': 'Португали', 'ceb': 'Portugal', 'cgg': 'Pocugo', 'chr': 'ᏉᏥᎦᎳ', 'ckb': 'پورتوگال', 'cs': 'Portugalsko', 'cv': 'Португали', 'cy': 'Portiwgal', 'da': 'Portugal', 'dav': 'Ureno', 'de': 'Portugal', 'dje': 'Portugaal', 'dsb': 'Portugalska', 'dz': 'པོར་ཅུ་གཱལ', 'ebu': 'Ureno', 'ee': 'Portugal nutome', 'el': 'Πορτογαλία', 'el-polyton': 'Πορτογαλία', 'en': 'Portugal', 'en-Dsrt': '𐐑𐐱𐑉𐐽𐐲𐑀𐐲𐑊', 'eo': 'Portugalujo', 'es': 'Portugal', 'et': 'Portugal', 'eu': 'Portugal', 'ewo': 'fɔrətugɛ́s', 'fa': 'پرتغال', 'ff': 'Purtugaal', 'ff-Adlm': '𞤆𞤮𞥅𞤪𞤼𞤵𞤺𞤢𞥄𞤤', 'ff-Latn': 'Purtugaal', 'fi': 'Portugali', 'fil': 'Portugal', 'fo': 'Portugal', 'fr': 'Portugal', 'fur': 'Portugal', 'fy': 'Portugal', 'ga': 'an Phortaingéil', 'gd': 'A’ Phortagail', 'gl': 'Portugal', 'gsw': 'Portugal', 'gu': 'પોર્ટુગલ', 'guz': 'Ureno', 'ha': 'Portugal', 'he': 'פורטוגל', 'hi': 'पुर्तगाल', 'hi-Latn': 'Portugal', 'hr': 'Portugal', 'hsb': 'Portugalska', 'ht': 'Portugal', 'hu': 'Portugália', 'hy': 'Պորտուգալիա', 'ia': 'Portugal', 'id': 'Portugal', 'ie': 'Portugal', 'ig': 'Portugal', 'is': 'Portúgal', 'it': 'Portogallo', 'ja': 'ポルトガル', 'jmc': 'Ureno', 'jv': 'Portugal', 'ka': 'პორტუგალია', 'kaa': 'Португалия', 'kaa-Cyrl': 'Португалия', 'kab': 'Purtugal', 'kam': 'Ureno', 'kde': 'Uleno', 'kea': 'Purtugal', 'kgp': 'Portugar', 'khq': 'Portugaal', 'ki': 'Ureno', 'kk': 'Португалия', 'kk-Cyrl': 'Португалия', 'kln': 'Emetab Portugal', 'km': 'ព័រទុយហ្កាល់', 'kn': 'ಪೋರ್ಚುಗಲ್', 'ko': '포르투갈', 'kok': 'पुर्तगाल', 'kok-Deva': 'पुर्तगाल', 'ks': 'پُرتِگال', 'ks-Arab': 'پُرتِگال', 'ksb': 'Uleno', 'ksf': 'portugál', 'ksh': 'Pochtojall', 'ku': 'Portûgal', 'kxv': 'portugal', 'kxv-Deva': 'पुर्तगाल', 'kxv-Latn': 'portugal', 'kxv-Orya': 'ପର୍ତୁଗାଲ', 'kxv-Telu': 'పోర్చుగల్', 'ky': 'Португалия', 'lag': 'Ʉréeno', 'lb': 'Portugal', 'lg': 'Potugaali', 'lld': 'Portugal', 'ln': 'Putúlugɛsi', 'lo': 'ພອລທູໂກ', 'lt': 'Portugalija', 'lu': 'Mputulugeshi', 'luo': 'Portugal', 'luy': 'Ureno', 'lv': 'Portugāle', 'mai': 'पुर्तगाल', 'mas': 'Ureno', 'mer': 'Potogo', 'mfe': 'Portigal', 'mg': 'Pôrtiogala', 'mi': 'Potukara', 'mk': 'Португалија', 'ml': 'പോർച്ചുഗൽ', 'mn': 'Португал', 'mr': 'पोर्तुगाल', 'ms': 'Portugal', 'mt': 'il-Portugall', 'mua': 'Sǝr Portugal', 'my': 'ပေါ်တူဂီ', 'mzn': 'پرتغال', 'naq': 'Portugali', 'nb': 'Portugal', 'nd': 'Portugal', 'ne': 'पोर्चुगल', 'nl': 'Portugal', 'nmg': 'Pɔrtugal', 'nn': 'Portugal', 'no': 'Portugal', 'nqo': 'ߔߐߙߑߕߎߜ߭ߊߟ', 'nyn': 'Pocugo', 'om': 'Poorchugaal', 'or': 'ପର୍ତ୍ତୁଗାଲ୍', 'pa': 'ਪੁਰਤਗਾਲ', 'pa-Guru': 'ਪੁਰਤਗਾਲ', 'pcm': 'Pọ́túgal', 'pl': 'Portugalia', 'ps': 'پورتګال', 'pt': 'Portugal', 'qu': 'Portugal', 'rm': 'Portugal', 'rn': 'Porutugali', 'ro': 'Portugalia', 'rof': 'Ureno', 'ru': 'Португалия', 'rwk': 'Ureno', 'saq': 'Ureno', 'sbp': 'Wuleno', 'sc': 'Portogallu', 'sd': 'پرتگال', 'sd-Arab': 'پرتگال', 'se': 'Portugála', 'seh': 'Portugal', 'ses': 'Portugaal', 'sg': 'Pörtugäle, Ködörö Pûra', 'shi': 'ⴱⵕⵟⵇⵉⵣ', 'shi-Latn': 'bṛṭqiz', 'shi-Tfng': 'ⴱⵕⵟⵇⵉⵣ', 'si': 'පෘතුගාලය', 'sk': 'Portugalsko', 'sl': 'Portugalska', 'smn': 'Portugal', 'sn': 'Portugal', 'so': 'Bortugaal', 'sq': 'Portugali', 'sr': 'Португалија', 'sr-Cyrl': 'Португалија', 'sr-Latn': 'Portugalija', 'sv': 'Portugal', 'sw': 'Ureno', 'syr': 'ܦܘܪܛܘܓܠ', 'ta': 'போர்ச்சுக்கல்', 'te': 'పోర్చుగల్', 'teo': 'Ureno', 'tg': 'Португалия', 'th': 'โปรตุเกส', 'ti': 'ፖርቱጋል', 'tk': 'Portugaliýa', 'to': 'Potukali', 'tr': 'Portekiz', 'tt': 'Португалия', 'twq': 'Portugaal', 'tzm': 'Purtuɣal', 'ug': 'پورتۇگالىيە', 'uk': 'Португалія', 'ur': 'پرتگال', 'uz': 'Portugaliya', 'uz-Cyrl': 'Португалия', 'uz-Latn': 'Portugaliya', 'vai': 'ꕶꕿꕃꔤ ꕸꖃꔀ', 'vai-Latn': 'Potokíi', 'vai-Vaii': 'ꕶꕿꕃꔤ ꕸꖃꔀ', 'vec': 'Portogalo', 'vi': 'Bồ Đào Nha', 'vun': 'Ureno', 'wae': 'Portugal', 'wo': 'Portigaal', 'xh': 'EPortugal', 'xnr': 'पुर्तगाल', 'xog': 'Potugaali', 'yav': 'pɔltukáal', 'yi': 'פּארטוגאַל', 'yo': 'Pọ́túgà', 'yrl': 'Putugau', 'yue': '葡萄牙', 'yue-Hans': '葡萄牙', 'yue-Hant': '葡萄牙', 'zgh': 'ⴱⵕⵟⵇⵉⵣ', 'zh': '葡萄牙', 'zh-Hans': '葡萄牙', 'zh-Hant': '葡萄牙', 'zu': 'i-Portugal'},
    'PW': {'af': 'Palau', 'agq': 'Kpàlawù', 'ak': 'Palau', 'am': 'ፓላው', 'ar': 'بالاو', 'as': 'পালাউ', 'asa': 'Palau', 'ast': 'Paláu', 'az': 'Palau', 'az-Cyrl': 'Палау', 'az-Latn': 'Palau', 'bal-Latn': 'Paláu', 'bas': 'Pàlaù', 'be': 'Палау', 'be-tarask': 'Палау', 'bez': 'Hupalau', 'bg': 'Палау', 'blo': 'Palawoo', 'bm': 'Palawu', 'bn': 'পালাউ', 'br': 'Palau', 'brx': 'पालाऊ', 'bs': 'Palau', 'bs-Cyrl': 'Палау', 'bs-Latn': 'Palau', 'ca': 'Palau', 'ca-ES-valencia': 'Palau', 'ccp': '𑄛𑄣𑄃𑄪', 'ce': 'Палау', 'ceb': 'Palau', 'cgg': 'Palaawu', 'chr': 'ᏆᎴᎠᏫ', 'ckb': 'پالاو', 'cs': 'Palau', 'cv': 'Палау', 'cy': 'Palau', 'da': 'Palau', 'dav': 'Palau', 'de': 'Palau', 'dje': 'Palu', 'dsb': 'Palau', 'dz': 'པ་ལའུ', 'ebu': 'Palau', 'ee': 'Palau nutome', 'el': 'Παλάου', 'el-polyton': 'Παλάου', 'en': 'Palau', 'en-Dsrt': '𐐑𐐲𐑊𐐵', 'eo': 'Palaŭo', 'es': 'Palaos', 'et': 'Belau', 'eu': 'Palau', 'ewo': 'Palau', 'fa': 'پالائو', 'ff': 'Palawu', 'ff-Adlm': '𞤆𞤢𞤤𞤢𞤱', 'ff-Latn': 'Palawu', 'fi': 'Palau', 'fil': 'Palau', 'fo': 'Palau', 'fr': 'Palaos', 'fur': 'Palau', 'fy': 'Palau', 'ga': 'Oileáin Palau', 'gd': 'Palabh', 'gl': 'Palau', 'gsw': 'Palau', 'gu': 'પલાઉ', 'guz': 'Palau', 'ha': 'Palau', 'he': 'פלאו', 'hi': 'पलाऊ', 'hi-Latn': 'Palau', 'hr': 'Palau', 'hsb': 'Palau', 'ht': 'Palaos', 'hu': 'Palau', 'hy': 'Պալաու', 'ia': 'Palau', 'id': 'Palau', 'ie': 'Palau', 'ig': 'Palau', 'is': 'Palá', 'it': 'Palau', 'ja': 'パラオ', 'jmc': 'Palau', 'jv': 'Palau', 'ka': 'პალაუ', 'kaa': 'Палау', 'kaa-Cyrl': 'Палау', 'kab': 'Palu', 'kam': 'Palau', 'kde': 'Palau', 'kea': 'Palau', 'kgp': 'Paravu', 'khq': 'Palu', 'ki': 'Palau', 'kk': 'Палау', 'kk-Cyrl': 'Палау', 'kln': 'Emetab Palau', 'km': 'ផៅឡូ', 'kn': 'ಪಲಾವು', 'ko': '팔라우', 'kok': 'पलाऊ', 'kok-Deva': 'पलाऊ', 'ks': 'پَلاو', 'ks-Arab': 'پَلاو', 'ksb': 'Palau', 'ksf': 'palwa', 'ksh': 'Pallau', 'ku': 'Palau', 'kxv': 'palau', 'kxv-Deva': 'पलाऊ', 'kxv-Latn': 'palau', 'kxv-Orya': 'ପଲାଉ', 'kxv-Telu': 'పాలావ్', 'ky': 'Палау', 'lag': 'Paláau', 'lb': 'Palau', 'lg': 'Palawu', 'lld': 'Palau', 'ln': 'Palau', 'lo': 'ປາລາວ', 'lt': 'Palau', 'lu': 'Palau', 'luo': 'Palau', 'luy': 'Palau', 'lv': 'Palau', 'mai': 'पलाऊ', 'mas': 'Palau', 'mer': 'Palau', 'mfe': 'Palau', 'mg': 'Palao', 'mi': 'Pārau', 'mk': 'Палау', 'ml': 'പലാവു', 'mn': 'Палау', 'mr': 'पलाऊ', 'ms': 'Palau', 'mt': 'Palau', 'mua': 'Sǝr Palau', 'my': 'ပလာအို', 'mzn': 'پالائو', 'naq': 'Palau', 'nb': 'Palau', 'nd': 'Palau', 'ne': 'पलाउ', 'nl': 'Palau', 'nmg': 'Palo', 'nn': 'Palau', 'no': 'Palau', 'nqo': 'ߔߟߊߐߛ', 'nyn': 'Palaawu', 'om': 'Palaawu', 'or': 'ପାଲାଉ', 'pa': 'ਪਲਾਉ', 'pa-Guru': 'ਪਲਾਉ', 'pcm': 'Paláu', 'pl': 'Palau', 'ps': 'پلاؤ', 'pt': 'Palau', 'qu': 'Palaos', 'rm': 'Palau', 'rn': 'Palawu', 'ro': 'Palau', 'rof': 'Palau', 'ru': 'Палау', 'rwk': 'Palau', 'saq': 'Palau', 'sbp': 'Palawu', 'sc': 'Palau', 'sd': 'پلائو', 'sd-Arab': 'پلائو', 'se': 'Palau', 'seh': 'Palau', 'ses': 'Palu', 'sg': 'Palau', 'shi': 'ⴱⴰⵍⴰⵡ', 'shi-Latn': 'balaw', 'shi-Tfng': 'ⴱⴰⵍⴰⵡ', 'si': 'පලාවු', 'sk': 'Palau', 'sl': 'Palau', 'smn': 'Palau', 'sn': 'Palau', 'so': 'Balaaw', 'sq': 'Palau', 'sr': 'Палау', 'sr-Cyrl': 'Палау', 'sr-Latn': 'Palau', 'sv': 'Palau', 'sw': 'Palau', 'syr': 'ܦܠܐܘ', 'ta': 'பாலோ', 'te': 'పాలావ్', 'teo': 'Palau', 'tg': 'Палау', 'th': 'ปาเลา', 'ti': 'ፓላው', 'tk': 'Palau', 'to': 'Palau', 'tr': 'Palau', 'tt': 'Палау', 'twq': 'Palu', 'tzm': 'Palu', 'ug': 'پالائۇ', 'uk': 'Палау', 'ur': 'پلاؤ', 'uz': 'Palau', 'uz-Cyrl': 'Палау', 'uz-Latn': 'Palau', 'vai': 'ꕐꖃ', 'vai-Latn': 'Palo', 'vai-Vaii': 'ꕐꖃ', 'vec': 'Palàu', 'vi': 'Palau', 'vun': 'Palau', 'wae': 'Palau', 'wo': 'Palaw', 'xh': 'EPalau', 'xnr': 'पलाऊ', 'xog': 'Palawu', 'yav': 'palawú', 'yo': 'Paalu', 'yrl': 'Parau', 'yue': '帛琉', 'yue-Hans': '帛琉', 'yue-Hant': '帛琉', 'zgh': 'ⴱⴰⵍⴰⵡ', 'zh': '帕劳', 'zh-Hans': '帕劳', 'zh-Hant': '帛琉', 'zu': 'i-Palau'},
    'PY': {'af': 'Paraguay', 'agq': 'Kpalàgwɛ̂', 'ak': 'Paraguae', 'am': 'ፓራጓይ', 'ar': 'باراغواي', 'as': 'পাৰাগুৱে', 'asa': 'Paragwai', 'ast': 'Paraguái', 'az': 'Paraqvay', 'az-Cyrl': 'Парагвај', 'az-Latn': 'Paraqvay', 'bal-Latn': 'Parágóay', 'bas': 'Pàràgwê', 'be': 'Парагвай', 'be-tarask': 'Парагвай', 'bez': 'Hupalagwai', 'bg': 'Парагвай', 'bgn': 'پاراگوی', 'blo': 'Paraguwee', 'bm': 'Paraguwayi', 'bn': 'প্যারাগুয়ে', 'br': 'Paraguay', 'brx': 'पारागुए', 'bs': 'Paragvaj', 'bs-Cyrl': 'Парагвај', 'bs-Latn': 'Paragvaj', 'ca': 'Paraguai', 'ca-ES-valencia': 'Paraguai', 'ccp': '𑄛𑄳𑄠𑄢𑄉𑄪𑄠𑄬', 'ce': 'Парагвай', 'ceb': 'Paraguay', 'cgg': 'Paragwai', 'chr': 'ᏆᎳᏇᎢᏯ', 'ckb': 'پاراگوای', 'cs': 'Paraguay', 'cv': 'Парагвай', 'cy': 'Paraguay', 'da': 'Paraguay', 'dav': 'Paragwai', 'de': 'Paraguay', 'dje': 'Paraguwey', 'dsb': 'Paraguay', 'dz': 'པ་ར་གུ་ཝའི', 'ebu': 'Paragwai', 'ee': 'Paragua nutome', 'el': 'Παραγουάη', 'el-polyton': 'Παραγουάη', 'en': 'Paraguay', 'en-Dsrt': '𐐑𐐯𐑉𐐲𐑀𐐶𐐴', 'eo': 'Paragvajo', 'es': 'Paraguay', 'et': 'Paraguay', 'eu': 'Paraguai', 'ewo': 'Paragué', 'fa': 'پاراگوئه', 'ff': 'Paraguwaay', 'ff-Adlm': '𞤆𞤢𞥄𞤪𞤢𞤺𞤵𞤱𞤢𞥄𞤴', 'ff-Latn': 'Paraguwaay', 'fi': 'Paraguay', 'fil': 'Paraguay', 'fo': 'Paraguai', 'fr': 'Paraguay', 'fur': 'Paraguay', 'fy': 'Paraguay', 'ga': 'Paragua', 'gd': 'Paraguaidh', 'gl': 'Paraguai', 'gsw': 'Paraguai', 'gu': 'પેરાગ્વે', 'guz': 'Paragwai', 'ha': 'Faragwai', 'he': 'פרגוואי', 'hi': 'पराग्वे', 'hi-Latn': 'Paraguay', 'hr': 'Paragvaj', 'hsb': 'Paraguay', 'ht': 'Paraguay', 'hu': 'Paraguay', 'hy': 'Պարագվայ', 'ia': 'Paraguay', 'id': 'Paraguay', 'ig': 'Paraguay', 'is': 'Paragvæ', 'it': 'Paraguay', 'ja': 'パラグアイ', 'jmc': 'Paragwai', 'jv': 'Paraguay', 'ka': 'პარაგვაი', 'kaa': 'Парагвай', 'kaa-Cyrl': 'Парагвай', 'kab': 'Paragway', 'kam': 'Paragwai', 'kde': 'Palagwai', 'kea': 'Paraguai', 'kgp': 'Paraguvaj', 'khq': 'Paraguwey', 'ki': 'Paragwai', 'kk': 'Парагвай', 'kk-Cyrl': 'Парагвай', 'kln': 'Emetab Paraguay', 'km': 'ប៉ារ៉ាហ្គាយ', 'kn': 'ಪರಾಗ್ವೇ', 'ko': '파라과이', 'kok': 'पैराग्वे', 'kok-Deva': 'पैराग्वे', 'ks': 'پَراگُے', 'ks-Arab': 'پَراگُے', 'ksb': 'Palagwai', 'ksf': 'paragwɛ́', 'ksh': 'Parraguwai', 'ku': 'Paragûay', 'kxv': 'pareguvai', 'kxv-Deva': 'पराग्वे', 'kxv-Latn': 'pareguvai', 'kxv-Orya': 'ପରଗ୍ୱେ', 'kxv-Telu': 'పరాగ్వే', 'ky': 'Парагвай', 'lag': 'Paraguáai', 'lb': 'Paraguay', 'lg': 'Paragwayi', 'lld': 'Paraguay', 'ln': 'Palagwei', 'lo': 'ພາຣາກວຍ', 'lt': 'Paragvajus', 'lu': 'Palagwei', 'luo': 'Paraguay', 'luy': 'Paragwai', 'lv': 'Paragvaja', 'mai': 'पराग्वे', 'mas': 'Paragwai', 'mer': 'Paragwai', 'mfe': 'Paraguay', 'mg': 'Paragoay', 'mgh': 'Paragwai', 'mi': 'Parakai', 'mk': 'Парагвај', 'ml': 'പരാഗ്വേ', 'mn': 'Парагвай', 'mr': 'पराग्वे', 'ms': 'Paraguay', 'mt': 'il-Paragwaj', 'mua': 'Paragwai', 'my': 'ပါရာဂွေး', 'mzn': 'پاراگوئه', 'naq': 'Paraguaib', 'nb': 'Paraguay', 'nd': 'Paraguay', 'ne': 'प्याराग्वे', 'nl': 'Paraguay', 'nmg': 'Paraguay', 'nn': 'Paraguay', 'no': 'Paraguay', 'nqo': 'ߔߙߊߜ߭ߏߦߋ߫', 'nyn': 'Paragwai', 'om': 'Paaraguwaay', 'or': 'ପାରାଗୁଏ', 'pa': 'ਪੈਰਾਗਵੇ', 'pa-Guru': 'ਪੈਰਾਗਵੇ', 'pcm': 'Párágwue', 'pl': 'Paragwaj', 'ps': 'پاراګوی', 'pt': 'Paraguai', 'qu': 'Paraguay', 'rm': 'Paraguai', 'rn': 'Paragwe', 'ro': 'Paraguay', 'rof': 'Paragwai', 'ru': 'Парагвай', 'rwk': 'Paragwai', 'saq': 'Paragwai', 'sbp': 'Palagwayi', 'sc': 'Paraguày', 'sd': 'پيراگوءِ', 'sd-Arab': 'پيراگوءِ', 'se': 'Paraguay', 'seh': 'Paraguai', 'ses': 'Paraguwey', 'sg': 'Paraguëe', 'shi': 'ⴱⴰⵔⴰⴳⵡⴰⵢ', 'shi-Latn': 'baragway', 'shi-Tfng': 'ⴱⴰⵔⴰⴳⵡⴰⵢ', 'si': 'පැරගුවේ', 'sk': 'Paraguaj', 'sl': 'Paragvaj', 'smn': 'Paraguay', 'sn': 'Paraguay', 'so': 'Baraguaay', 'sq': 'Paraguai', 'sr': 'Парагвај', 'sr-Cyrl': 'Парагвај', 'sr-Latn': 'Paragvaj', 'sv': 'Paraguay', 'sw': 'Paraguay', 'syr': 'ܦܪܓܘܐܝ', 'ta': 'பராகுவே', 'te': 'పరాగ్వే', 'teo': 'Paragwai', 'tg': 'Парагвай', 'th': 'ปารากวัย', 'ti': 'ፓራጓይ', 'tk': 'Paragwaý', 'to': 'Palakuai', 'tr': 'Paraguay', 'tt': 'Парагвай', 'twq': 'Paraguwey', 'tzm': 'Paragway', 'ug': 'پاراگۋاي', 'uk': 'Парагвай', 'ur': 'پیراگوئے', 'uz': 'Paragvay', 'uz-Cyrl': 'Парагвай', 'uz-Latn': 'Paragvay', 'vai': 'ꕐꕟꗝꔀ', 'vai-Latn': 'Paragɔe', 'vai-Vaii': 'ꕐꕟꗝꔀ', 'vec': 'Paraguài', 'vi': 'Paraguay', 'vun': 'Paragwai', 'wae': 'Paraguai', 'wo': 'Paraguwe', 'xh': 'EParaguay', 'xnr': 'पराग्वे', 'xog': 'Paragwayi', 'yav': 'palakúé', 'yi': 'פּאַראַגווײַ', 'yo': 'Paraguye', 'yrl': 'Paraguwai', 'yue': '巴拉圭', 'yue-Hans': '巴拉圭', 'yue-Hant': '巴拉圭', 'zgh': 'ⴱⴰⵔⴰⴳⵡⴰⵢ', 'zh': '巴拉圭', 'zh-Hans': '巴拉圭', 'zh-Hant': '巴拉圭', 'zu': 'i-Paraguay'},
    'Palm': {'ast': 'palmirenu', 'bal-Latn': 'Palmrén', 'bs': 'palmyrene pismo', 'bs-Cyrl': 'палмѕрене писмо', 'bs-Latn': 'palmyrene pismo', 'ca': 'palmirè', 'ca-ES-valencia': 'palmirè', 'cs': 'palmýrské', 'da': 'palmyrensk', 'de': 'Palmyrenisch', 'en': 'Palmyrene', 'et': 'Palmyra', 'eu': 'palmiera', 'fa': 'پالمیرایی', 'ff-Adlm': '𞤆𞤢𞤤𞤥𞤫𞤪𞤫𞥅𞤲', 'fi': 'palmyralainen', 'fo': 'palmyrene', 'fy': 'Palmyreens', 'ga': 'Palmyrene', 'gd': 'Palmyrene', 'he': 'פלמירן', 'hi-Latn': 'Palmyrene', 'hr': 'palmyrene pismo', 'id': 'Palmira', 'ig': 'Palmyrene', 'it': 'palmireno', 'ja': 'パルミラ文字', 'ka': 'პალმირული', 'kk': 'палмир жазуы', 'kk-Cyrl': 'палмир жазуы', 'ko': '팔미라 문자', 'lo': 'ພາລໄມຮິນ', 'lt': 'Palmiros', 'mk': 'палмирско', 'mn': 'Палмирен', 'ms': 'Palmyrene', 'nb': 'palmyrensk', 'nl': 'Palmyreens', 'nn': 'palmyrensk', 'no': 'palmyrensk', 'nqo': 'ߔߊߟߑߡߙߋߣߌ', 'ro': 'palmirenă', 'ru': 'пальмиры', 'sc': 'palmirenu', 'so': 'Palmyrene', 'sq': 'palmirenisht', 'sv': 'palmyreniska', 'tg': 'Палмирена', 'th': 'พาลไมรีน', 'to': 'tohinima fakapalamilene', 'tr': 'Palmira', 'ug': 'پالمىراچە', 'vec': 'palmiren', 'vi': 'Chữ Palmyrene', 'yue': '帕米瑞拉文字', 'yue-Hans': '帕米瑞拉文字', 'yue-Hant': '帕米瑞拉文字', 'zh': '帕尔迈拉文', 'zh-Hans': '帕尔迈拉文', 'zh-Hant': '帕米瑞拉文字', 'zu': 'i-Palmyrene'},
    'Pauc': {'ast': 'pau cin hau', 'bal-Latn': 'Páu Chen Háu', 'bs': 'pau cin hau pismo', 'bs-Cyrl': 'пау цин хау писмо', 'bs-Latn': 'pau cin hau pismo', 'ca': 'Pau Cin Hau', 'ca-ES-valencia': 'Pau Cin Hau', 'cs': 'pau cin hau', 'de': 'Pau Cin Hau', 'en': 'Pau Cin Hau', 'eu': 'pau cin hau', 'ff-Adlm': '𞤆𞤢𞤱 𞤅𞤭𞥅𞤲 𞤖𞤢𞥄𞤱', 'fi': 'zotuallai', 'fo': 'pau cin hau', 'ga': 'Pau Cin Hau', 'gd': 'Pau Cin Hau', 'he': 'פאו צ׳ין האו', 'hi-Latn': 'Pau Cin Hau', 'id': 'Pau Cin Hau', 'ig': 'Pau Cin Hau', 'ja': 'パウ・チン・ハウ文字', 'kk': 'пау син хау жазуы', 'kk-Cyrl': 'пау син хау жазуы', 'lt': 'Pau Cin Hau', 'mk': 'Паучинхауово', 'mn': 'Пай Чин Хау', 'ms': 'Pau Cin Hau', 'nb': 'pau cin hau', 'nl': 'Pau Cin Hau', 'nn': 'pau cin hau', 'no': 'pau cin hau', 'nqo': 'ߔߊߎ-ߛߌ߲ߤߊߥߎ', 'ro': 'pau cin hau', 'sc': 'pau cin hau', 'so': 'Baaw Sin Haaw', 'sq': 'pausinhauisht', 'sv': 'Pau Cin Hau-skrift', 'tg': 'Пау Син Хау', 'th': 'ป่อจิ้งฮอ', 'to': 'tohinima fakapausinihau', 'tr': 'Pau Cin Hau', 'vec': 'pau cin hau', 'zh': '包金豪文', 'zh-Hans': '包金豪文', 'zh-Hant': '鮑欽豪文', 'zu': 'i-Pau Cin Hau'},
    'Pcun': {'en': 'Proto-Cuneiform'},
    'Pelm': {'en': 'Proto-Elamite'},
    'Perm': {'ar': 'البيرميكية القديمة', 'ast': 'pérmicu antiguu', 'az': 'qədimi permik', 'az-Latn': 'qədimi permik', 'bal-Latn': 'Ahdi Permi', 'bg': 'Древно пермска', 'bn': 'প্রাচীন পার্মিক', 'brx': 'पुरानी पर्मीक्', 'bs': 'staro permiksko pismo', 'bs-Cyrl': 'старо пермикско писмо', 'bs-Latn': 'staro permiksko pismo', 'ca': 'antic pèrmic', 'ca-ES-valencia': 'antic pèrmic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄴 𑄛𑄢𑄴𑄟𑄨𑄇𑄴', 'cs': 'staropermské', 'da': 'oldpermisk', 'de': 'Altpermisch', 'el': 'Παλαιό Περμικό', 'el-polyton': 'Παλαιό Περμικό', 'en': 'Old Permic', 'en-Dsrt': '𐐄𐑊𐐼 𐐑𐐯𐑉𐑋𐐮𐐿', 'es': 'permiano antiguo', 'et': 'vanapermi', 'eu': 'permiera zaharra', 'fa': 'پرمی باستان', 'ff-Adlm': '𞤆𞤫𞤪𞤥𞤭𞤳 𞤑𞤭𞤯𞥆𞤵𞤲', 'fi': 'muinaispermiläinen', 'fo': 'gamalt permic', 'fr': 'ancien permien', 'fy': 'Aldpermysk', 'ga': 'Sean-Pheirmeach', 'gd': 'Seann-Phermic', 'gsw': 'Altpermisch', 'gu': 'ઓલ્ડ પરમિક', 'he': 'פרמית ישנה', 'hi': 'ओल्ड परमिक', 'hi-Latn': 'Old Permic', 'hr': 'staro permic pismo', 'ht': 'ancien permien', 'hu': 'Ópermikus', 'id': 'Permik Kuno', 'ig': 'Old Permic', 'it': 'permico antico', 'ja': '古ぺルム文字', 'ka': 'ძველი პერმული', 'kgp': 'pérmĩku si', 'kk': 'көне перм жазуы', 'kk-Cyrl': 'көне перм жазуы', 'kn': 'ಪ್ರಾಚೀನ ಪೆರ್ಮಿಕ್', 'ko': '고대 페름 문자', 'ks': 'اولڈ پٔرمِک', 'ks-Arab': 'اولڈ پٔرمِک', 'lb': 'Alpermesch', 'lo': 'ເພີມີໂບຮານ', 'lt': 'senieji permės', 'mk': 'старопермско', 'ml': 'പുരാതന പെർമിക്', 'mn': 'Хуучин Пермик', 'mr': 'पुरातन पर्मिक', 'ms': 'Permic Lama', 'nb': 'gammelpermisk', 'ne': 'पुरानो पर्मिक', 'nl': 'Oudpermisch', 'nn': 'gammalpermisk', 'no': 'gammelpermisk', 'nqo': 'ߡߙߋߡߌ߫ ߞߘߐ߬ߡߊ߲', 'or': 'ଓଲ୍ଡ ପରମିକ୍', 'pl': 'staropermskie', 'pt': 'pérmico antigo', 'rm': 'permic vegl', 'ro': 'permică veche', 'ru': 'древнепермская', 'sc': 'pèrmicu antigu', 'sl': 'staropermijski', 'so': 'Permic gii hore', 'sq': 'permike e vjetër', 'sr': 'старо пермикско писмо', 'sr-Cyrl': 'старо пермикско писмо', 'sr-Latn': 'staro permiksko pismo', 'sv': 'fornpermiska', 'ta': 'பழைய பெர்மிக்', 'te': 'ప్రాచీన పెర్మిక్', 'tg': 'Пермикии Куҳна', 'th': 'เปอร์มิกโบราณ', 'to': 'tohinima fakapēmi-motuʻa', 'tr': 'Eski Permik', 'ug': 'قەدىمكى پېرمىكچە', 'uk': 'давньопермський', 'vec': 'pèrmego antigo', 'vi': 'Chữ Permic cổ', 'xnr': 'ओल्ड परमिक', 'yrl': 'périmiku kuxiímawara', 'yue': '古彼爾姆諸文', 'yue-Hans': '古彼尔姆诸文', 'yue-Hant': '古彼爾姆諸文', 'zh': '古彼尔姆文', 'zh-Hans': '古彼尔姆文', 'zh-Hant': '古彼爾姆諸文', 'zu': 'i-Old Permic'},
    'Phag': {'ar': 'الفاجسبا', 'ast': 'escritura ‘Phags-pa', 'az': 'faqs-pa', 'az-Latn': 'faqs-pa', 'bal-Latn': 'Pags-pa', 'bg': 'Фагс-па', 'bn': 'ফাগ্স-পা', 'brx': 'फाग्स पा', 'bs': 'phags-pa pismo', 'bs-Cyrl': 'пагс-па писмо', 'bs-Latn': 'phags-pa pismo', 'ca': 'phagspa', 'ca-ES-valencia': 'phagspa', 'ccp': '𑄜𑄧𑄉𑄴𑄥𑄧-𑄛', 'cs': 'phags-pa', 'da': 'phags-pa', 'de': 'Phags-pa', 'el': 'Παγκς-πα', 'el-polyton': 'Παγκς-πα', 'en': 'Phags-pa', 'es': 'phags-pa', 'et': 'phakpa', 'eu': 'phags-pa', 'ff-Adlm': '𞤊𞤢𞤺𞤧 𞤆𞤢', 'fi': 'phags-pa', 'fo': 'phags-pa', 'fr': 'phags pa', 'fy': 'Phags-pa', 'ga': 'Phags-pa', 'gd': 'Phags-pa', 'gsw': 'Phags-pa', 'gu': 'ફાગ્સ-પા', 'he': 'פאגס', 'hi': 'फाग्स-पा', 'hi-Latn': 'Phags-pa', 'hr': 'phags-pa pismo', 'ht': 'phags pa', 'hu': 'Phags-pa', 'id': 'Phags-pa', 'ig': 'Phags-pa', 'it': 'phags-pa', 'ja': 'パスパ文字', 'ka': 'ფაგსპა', 'kgp': 'fagpa', 'kk': 'тибет-моңғол жазуы', 'kk-Cyrl': 'тибет-моңғол жазуы', 'kn': 'ಫಾಗ್ಸ್-ಪಾ', 'ko': '파스파 문자', 'ks': 'پھاگس پا', 'ks-Arab': 'پھاگس پا', 'lb': 'Phags-pa', 'lo': 'ຟາກສ-ປາ', 'lt': 'pagsa pa', 'mk': 'пагспа', 'ml': 'ഫഗസ് പ', 'mn': 'Фагс-па', 'mr': 'फाग्स-पा', 'ms': 'Phags-pa', 'nb': 'phags-pa', 'ne': 'फाग्स-पा', 'nl': 'Phags-pa', 'nn': 'phags-pa', 'no': 'phags-pa', 'nqo': 'ߝߊߜ߭-ߔߊ', 'or': 'ଫାଗସ୍-ପା', 'pl': 'phags-pa', 'pt': 'phags-pa', 'rm': 'phags-pa', 'ro': 'phags-pa', 'ru': 'пагспа', 'sc': 'phags-pa', 'sl': 'pagpajski', 'so': 'Qoraalka Phags-pa', 'sq': 'fagspaisht', 'sr': 'пагс-па писмо', 'sr-Cyrl': 'пагс-па писмо', 'sr-Latn': 'pags-pa pismo', 'sv': 'phags-pa', 'ta': 'பக்ஸ்-பா', 'te': 'ఫాగ్స్-పా', 'tg': 'Фагс-па', 'th': 'ฟากส์-ปา', 'to': 'tohinima fakapākisipā', 'tr': 'Phags-pa', 'ug': 'پاسپاچە', 'uk': 'пхагс-па', 'vec': 'pagspa', 'vi': 'Chữ Phags-pa', 'xnr': 'फाग्स-पा', 'yrl': 'phagipa', 'yue': '八思巴文', 'yue-Hans': '八思巴文', 'yue-Hant': '八思巴文', 'zh': '八思巴文', 'zh-Hans': '八思巴文', 'zh-Hant': '八思巴文', 'zu': 'i-Phags-pa'},
    'Phli': {'ast': 'pahlavi d’inscripciones', 'az': 'fli', 'az-Latn': 'fli', 'bal-Latn': 'Kondahi Pahlawi', 'bn': 'খদিত পাহলভি', 'bs': 'pisani pahlavi', 'bs-Cyrl': 'писани пахлави', 'bs-Latn': 'pisani pahlavi', 'ca': 'pahlavi inscripcional', 'ca-ES-valencia': 'pahlavi inscripcional', 'ccp': '𑄈𑄧𑄘𑄨𑄖𑄧 𑄛𑄳𑄦𑄣𑄧𑄞𑄩', 'cs': 'pahlavské klínové', 'de': 'Buch-Pahlavi', 'el': 'Επιγραφικό Παχλάβι', 'el-polyton': 'Επιγραφικό Παχλάβι', 'en': 'Inscriptional Pahlavi', 'en-Dsrt': '𐐆𐑌𐑅𐐿𐑉𐐮𐐹𐑇𐐲𐑌𐐲𐑊 𐐑𐐪𐑊𐐲𐑂𐐨', 'et': 'pahlavi raidkiri', 'eu': 'pahlavi inskripzioak', 'fa': 'پهلوی کتیبه\u200cای', 'ff-Adlm': '𞤄𞤭𞤲𞤣𞤭 𞤆𞤢𞤤𞤢𞤾𞤭', 'fi': 'piirtokirjoituspahlavilainen', 'fr': 'pehlevi des inscriptions', 'fy': 'Inscriptioneel Pahlavi', 'ga': 'Pachlavais Inscríbhinne', 'gd': 'Pahlavi nan snaidh-sgrìobhaidhean', 'gu': 'ઇન્સ્ક્રિપ્શનલ પહલવી', 'he': 'פהלווי כתובתי', 'hi': 'इंस्क्रिपश्नल पाहलवी', 'hi-Latn': 'Inscriptional Pahlavi', 'hr': 'pisani pahlavi', 'ht': 'pehlevi des inscriptions', 'hu': 'Felriatos pahlavi', 'id': 'Pahlevi', 'ig': 'Inscriptional Pahlavi', 'it': 'pahlavi delle iscrizioni', 'ja': '碑文パフラヴィー文字', 'ka': 'მონუმენტური ფალაური', 'kgp': 'pahir', 'kk': 'жазба пехлеви', 'kk-Cyrl': 'жазба пехлеви', 'kn': 'ಇನ್ಸ್\u200cಕ್ರಿಪ್ಶನಲ್ ಪಾಹ್ಲವಿ', 'ko': '명문 팔라비 문자', 'lo': 'ປະຫລາວີອິນສຄິບຊັນແນລ', 'lt': 'rašytiniai pahlavi', 'mk': 'натписно средноперсиско', 'ml': 'എഴുത്തു പഹൽവി', 'mn': 'Пахлави', 'mr': 'इन्स्क्रिप्शनल पाहलवी', 'ms': 'Inskripsi Pahlavi', 'nb': 'inskripsjonspahlavi', 'ne': 'फ्लि', 'nl': 'Inscriptioneel Pahlavi', 'nn': 'inskripsjonspahlavi', 'no': 'inskripsjonspahlavi', 'nqo': 'ߝߋ߲ߟߋߥߌ߫ ߛߓߍߟߌ ߟߎ߬', 'or': 'ଇନସ୍କ୍ରୀପସାନଲ୍ ପାହାଲାୱୀ', 'pl': 'inskrypcyjne pahlawi', 'rm': 'pahlavi dad inscripziuns', 'ro': 'pahlavi pentru inscripții', 'ru': 'пехлевийская', 'sc': 'pahlavi de sas iscritziones', 'sl': 'vrezani napisi pahlavi', 'so': 'Qoraaladii hore ee Pahlavi', 'sq': 'palavishte mbishkrimesh', 'sr': 'писани пахлави', 'sr-Cyrl': 'писани пахлави', 'sr-Latn': 'pisani pahlavi', 'sv': 'tidig pahlavi', 'ta': 'இன்ஸ்கிரிப்ஷனல் பஹலவி', 'te': 'ఇంస్క్రిప్షనాల్ పహ్లావి', 'tg': 'Паҳлавии Хаттӣ', 'th': 'ปะห์ลาวีอินสคริปชันแนล', 'to': 'tohinima fakapālavi-tongi', 'tr': 'Pehlevi Kitabe Dili', 'ug': 'پەھلىۋىچە ئويما خەت', 'uk': 'пехлеві написів', 'vec': 'palavo de le iscrision', 'vi': 'Chữ Pahlavi Văn bia', 'xnr': 'इंस्क्रिपश्नल पाहलवी', 'yue': '巴列維文（碑銘體）', 'yue-Hans': '巴列维文（碑铭体）', 'yue-Hant': '巴列維文（碑銘體）', 'zh': '巴列维文碑铭体', 'zh-Hans': '巴列维文碑铭体', 'zh-Hant': '巴列維文（碑銘體）', 'zu': 'i-Inscriptional Pahlavi'},
    'Phlp': {'ast': 'pahlavi de salteriu', 'az': 'flp', 'az-Latn': 'flp', 'bal-Latn': 'Psáltar Pahlawi', 'bn': 'সল্টার পাহলভি', 'bs': 'psalter pahlavi pismo', 'bs-Cyrl': 'псалтер пахлави', 'bs-Latn': 'psalter pahlavi pismo', 'ca': 'psalter pahlavi', 'ca-ES-valencia': 'psalter pahlavi', 'ccp': '𑄥𑄧𑄣𑄴𑄑𑄢𑄴 𑄛𑄳𑄦𑄣𑄧𑄞𑄩', 'cs': 'pahlavské žalmové', 'de': 'Psalter-Pahlavi', 'el': 'Ψάλτερ Παχλάβι', 'el-polyton': 'Ψάλτερ Παχλάβι', 'en': 'Psalter Pahlavi', 'en-Dsrt': '𐐝𐐱𐑊𐐻𐐲𐑉 𐐑𐐪𐑊𐐲𐑂𐐨', 'et': 'pahlavi psalmikiri', 'eu': 'Pahlavi salmo-liburua', 'fa': 'پهلوی زبوری', 'ff-Adlm': '𞤅𞤮𞤤𞤼𞤮𞥅 𞤆𞤢𞤤𞤢𞤾𞤭', 'fi': 'psalttaripahlavilainen', 'fr': 'pehlevi des psautiers', 'fy': 'Psalmen Pahlavi', 'ga': 'Pachlavais Saltrach', 'gd': 'Pahlavi nan saltair', 'gu': 'સાલટર પહલવી', 'he': 'מזמור פהלווי', 'hi': 'सॉल्टर पाहलवी', 'hi-Latn': 'Psalter Pahlavi', 'hr': 'psalter pahlavi', 'ht': 'pehlevi des psautiers', 'hu': 'Psalter pahlavi', 'id': 'Mazmur Pahlevi', 'ig': 'Psalter Pahlavi', 'it': 'pahlavi psalter', 'ja': '詩編用パフラヴィー文字', 'ka': 'ფსალმუნური ფალაური', 'kgp': 'pahin', 'kk': 'забур пехлеви жазуы', 'kk-Cyrl': 'забур пехлеви жазуы', 'kn': 'ಸಾಲ್ಟರ್ ಪಾಹ್ಲವಿ', 'ko': '솔터 팔라비 문자', 'lo': 'ປະຫລາວີຊອດເຕຮ໌', 'lt': 'pselter pahlavi', 'mk': 'псалтирско средноперсиско', 'ml': 'സാൾട്ടർ പഹൽവി', 'mn': 'Псалтер Пахлави', 'mr': 'सॉल्टर पाहलवी', 'ms': 'Pslater Pahlavi', 'nb': 'psalter pahlavi', 'ne': 'फ्ल्प', 'nl': 'Psalmen Pahlavi', 'nn': 'salmepahlavi', 'no': 'psalter pahlavi', 'nqo': 'ߛߏߕߌߦߋ߫ ߝߋ߲ߟߋߝ߭ߌ', 'or': 'ସ୍ଲାଟର୍ ପାହାଲାୱୀ', 'pl': 'pahlawi psałterzowy', 'rm': 'pahlavi da psalms', 'ro': 'pahlavi pentru psaltire', 'ru': 'пахлави псалтирная', 'sc': 'psalter pahlavi', 'sl': 'psalmski pahlavi', 'so': 'Qoraalka midig laga bilaabo ee faarsiyiintii', 'sq': 'palavishte psalteri', 'sr': 'псалтер пахлави', 'sr-Cyrl': 'псалтер пахлави', 'sr-Latn': 'psalter pahlavi', 'sv': 'psaltaren-pahlavi', 'ta': 'சால்டர் பஹலவி', 'te': 'సల్టార్ పహ్లావి', 'tg': 'Паҳлавии Псалтирӣ', 'th': 'ปะห์ลาวีซอลเตอร์', 'to': 'tohinima fakapālavi-saame', 'tr': 'Psalter Pehlevi', 'ug': 'پەھلىۋىچە شېئىرىي تىل', 'uk': 'пехлеві релігійний', 'vec': 'palavo saltero', 'vi': 'Chữ Pahlavi Thánh ca', 'xnr': 'सॉल्टर पाहलवी', 'yue': '巴列維文（聖詩體）', 'yue-Hans': '巴列维文（圣诗体）', 'yue-Hant': '巴列維文（聖詩體）', 'zh': '巴列维文（圣诗体）', 'zh-Hans': '巴列维文（圣诗体）', 'zh-Hant': '巴列維文（聖詩體）', 'zu': 'i-Psalter Pahlavi'},
    'Phlv': {'ast': 'pahlavi de llibros', 'az': 'kitab paxlavi', 'az-Latn': 'kitab paxlavi', 'bg': 'Пахлавска', 'bn': 'পুস্তক পাহলভি', 'brx': 'बुक (सालटर) पहलवी', 'bs': 'pahlavi pismo', 'bs-Cyrl': 'пахлави писмо', 'bs-Latn': 'pahlavi pismo', 'ca': 'pahlavi', 'ca-ES-valencia': 'pahlavi', 'ccp': '𑄛𑄪𑄌𑄴𑄖𑄧𑄇𑄴 𑄛𑄳𑄦𑄣𑄧𑄞𑄩', 'cs': 'pahlavské knižní', 'da': 'pahlavi', 'de': 'Pahlavi', 'el': 'Μπουκ Παχλαβί', 'el-polyton': 'Μπουκ Παχλαβί', 'en': 'Book Pahlavi', 'en-Dsrt': '𐐒𐐳𐐿 𐐑𐐪𐑊𐐲𐑂𐐨', 'et': 'pahlavi raamatukiri', 'fa': 'پهلوی کتابی', 'fi': 'kirjapahlavilainen', 'fr': 'pehlevi des livres', 'fy': 'Boek Pahlavi', 'gsw': 'Pahlavi', 'gu': 'બુક પહલવી', 'hi': 'बुक पाहलवी', 'hi-Latn': 'Book Pahlavi', 'hr': 'pahlavi pismo', 'ht': 'pehlevi des livres', 'hu': 'Könyv pahlavi', 'id': 'Kitab Pahlevi', 'it': 'pahlavi book', 'ja': '書物用パフラヴィー文字', 'ka': 'წიგნური ფალაური', 'kgp': 'pahiravi si', 'kn': 'ಬುಕ್ ಪಾಹ್ಲವಿ', 'ko': '북 팔라비 문자', 'ks': 'بوٗک پَہَلوی', 'ks-Arab': 'بوٗک پَہَلوی', 'lb': 'Pahlavi', 'lo': 'ປະຫລາວີບຸກ', 'lt': 'buk pahvali', 'mk': 'книжевно староперсиско', 'ml': 'പഹൽവി ലിപി', 'mr': 'बुक पाहलवी', 'nb': 'pahlavi', 'ne': 'बुक पहल्भी', 'nl': 'Boek Pahlavi', 'nn': 'pahlavi', 'no': 'pahlavi', 'or': 'ବୁକ୍ ପାହାଲାୱୀ', 'pl': 'pahlawi książkowy', 'pt': 'pahlavi antigo', 'rm': 'pahlavi da cudeschs', 'ru': 'пахлави книжная', 'sl': 'knjižno palavanski', 'sr': 'пахлави писмо', 'sr-Cyrl': 'пахлави писмо', 'sr-Latn': 'pahlavi pismo', 'sv': 'bokpahlavi', 'ta': 'புக் பஹலவி', 'te': 'పుస్తక పహ్లావి', 'th': 'ปะห์ลาวีบุ๊ก', 'to': 'tohinima fakapālavi-tohi', 'tr': 'Kitap Pehlevi Dili', 'ug': 'پەھلىۋىچە كىتابى تىل', 'uk': 'пехлеві літературний', 'vi': 'Chữ Pahlavi Sách', 'yrl': 'paravi kuxiímawara', 'yue': '巴列維文（書體）', 'yue-Hans': '巴列维文（书体）', 'yue-Hant': '巴列維文（書體）', 'zh': '巴列维文（书体）', 'zh-Hans': '巴列维文（书体）', 'zh-Hant': '巴列維文（書體）'},
    'Phnx': {'ar': 'الفينيقية', 'ast': 'feniciu', 'az': 'foenik', 'az-Latn': 'foenik', 'bal-Latn': 'Phónisi', 'bg': 'Финикийска', 'bn': 'ফিনিশিয়', 'br': 'fenikianek', 'brx': 'फोनीशीयन', 'bs': 'feničansko pismo', 'bs-Cyrl': 'феничанско писмо', 'bs-Latn': 'feničansko pismo', 'ca': 'fenici', 'ca-ES-valencia': 'fenici', 'ccp': '𑄜𑄨𑄚𑄨𑄥𑄩𑄠𑄧', 'cs': 'fénické', 'da': 'fønikisk', 'de': 'Phönizisch', 'el': 'Φοινικικό', 'el-polyton': 'Φοινικικό', 'en': 'Phoenician', 'en-Dsrt': '𐐙𐐬𐑌𐐨𐑇𐐲𐑌', 'es': 'fenicio', 'et': 'foiniikia', 'eu': 'feniziera', 'fa': 'فنیقی', 'ff-Adlm': '𞤊𞤭𞤲𞤭𞤳𞤭𞤴𞤢𞤲𞤳𞤮', 'fi': 'foinikialainen', 'fr': 'phénicien', 'fy': 'Foenicysk', 'ga': 'Féiníceach', 'gd': 'Pheniceach', 'gsw': 'Phönizisch', 'gu': 'ફોનિશિયન', 'he': 'פיניקי', 'hi': 'फोनिशियन', 'hi-Latn': 'Phoenician', 'hr': 'feničko pismo', 'ht': 'phénicien', 'hu': 'Főniciai', 'id': 'Phoenix', 'ig': 'Phoenician', 'it': 'fenicio', 'ja': 'フェニキア文字', 'ka': 'ფინიკიური', 'kgp': 'fenĩso', 'kk': 'финикия жазуы', 'kk-Cyrl': 'финикия жазуы', 'kn': 'ಫೀನಿಶಿಯನ್', 'ko': '페니키아 문자', 'ks': 'پھونِشِیَن', 'ks-Arab': 'پھونِشِیَن', 'lb': 'Phönizesch', 'lo': 'ຟີນິເຊຍ', 'lt': 'foenikų', 'lv': 'feniķiešu', 'mk': 'феникиско', 'ml': 'ഫിനീഷ്യൻ', 'mn': 'Финик', 'mr': 'फोनिशियन', 'ms': 'Phoenicia', 'nb': 'fønikisk', 'ne': 'फोनिसियन', 'nl': 'Foenicisch', 'nn': 'fønikisk', 'no': 'fønikisk', 'nqo': 'ߝߋߣߌߛߌ', 'or': 'ଫେନୋସିଆନ୍', 'pl': 'fenicki', 'pt': 'fenício', 'rm': 'fenizian', 'ro': 'feniciană', 'ru': 'финикийская', 'sc': 'fenìtziu', 'sl': 'feničanski', 'so': 'Luuqada Phoenicianka', 'sq': 'fenikisht', 'sr': 'Феничанско писмо', 'sr-Cyrl': 'Феничанско писмо', 'sr-Latn': 'Feničansko pismo', 'sv': 'feniciska', 'ta': 'ஃபோனேஷியன்', 'te': 'ఫోనిశియన్', 'tg': 'Финикӣ', 'th': 'ฟินิเชีย', 'to': 'tohinima fakafoinikia', 'tr': 'Fenike', 'ug': 'فىنىكچە', 'uk': 'фінікійський', 'vec': 'feniso', 'vi': 'Chữ Phoenicia', 'xnr': 'फोनिशियन', 'yrl': 'finísiu', 'yue': '腓尼基文', 'yue-Hans': '腓尼基文', 'yue-Hant': '腓尼基文', 'zh': '腓尼基文', 'zh-Hans': '腓尼基文', 'zh-Hant': '腓尼基文', 'zu': 'i-Phoenician'},
    'Piqd': {'en': 'Klingon (KLI pIqaD)'},
    'Plrd': {'ar': 'الصوتيات الجماء', 'ast': 'fonéticu de Pollard', 'az': 'polard fonetik', 'az-Latn': 'polard fonetik', 'bal-Latn': 'Pólli Tawári', 'bg': 'Писменост Полард', 'bn': 'পোলার্ড ধ্বনিক', 'brx': 'पौलार्ड़ फोनेटीक', 'bs': 'polard fonetsko pismo', 'bs-Cyrl': 'поралд фонетско писмо', 'bs-Latn': 'polard fonetsko pismo', 'ca': 'pollard miao', 'ca-ES-valencia': 'pollard miao', 'ccp': '𑄛𑄮𑄣𑄢𑄴𑄓𑄧 𑄙𑄧𑄚𑄨𑄇𑄴', 'cs': 'Pollardova fonetická abeceda', 'da': 'pollardtegn', 'de': 'Pollard Phonetisch', 'el': 'Φωνητικό Πόλαρντ', 'el-polyton': 'Φωνητικό Πόλαρντ', 'en': 'Pollard Phonetic', 'en-Dsrt': '𐐑𐐱𐑊𐐲𐑉𐐼 𐐙𐐬𐑌𐐯𐐻𐐮𐐿', 'es': 'Pollard Miao', 'et': 'Pollardi miao', 'eu': 'polardera fonetikoa', 'ff-Adlm': '𞤖𞤭𞤼𞤮𞤲𞤳𞤮 𞤆𞤮𞤤𞥆𞤢𞤪𞤣', 'fi': 'Pollardin foneettinen', 'fr': 'phonétique de Pollard', 'fy': 'Pollard-fonetysk', 'ga': 'Pollard Foghrach', 'gd': 'Miao Phollard', 'gsw': 'Pollard Phonetisch', 'gu': 'પોલાર્ડ ફોનેટિક', 'he': 'פולארד פונטי', 'hi': 'पॉलार्ड फोनेटिक', 'hi-Latn': 'Pollard Phonetic', 'hr': 'pollard fonetsko pismo', 'ht': 'phonétique de Pollard', 'hu': 'Pollard fonetikus', 'id': 'Fonetik Pollard', 'ig': 'Pollard Phonetic', 'it': 'fonetica di pollard', 'ja': 'ポラード音声記号', 'kgp': 'fonẽtiko porarne', 'kk': 'поллард фонетикалық жазуы', 'kk-Cyrl': 'поллард фонетикалық жазуы', 'km': 'ផូឡាដ', 'kn': 'ಪೊಲ್ಲಾರ್ಡ್ ಫೊನೆಟಿಕ್', 'ko': '폴라드 표음 문자', 'ks': 'پولاڑ پھونِٹِک', 'ks-Arab': 'پولاڑ پھونِٹِک', 'lb': 'Pollard Phonetesch', 'lo': 'ສັດຕະສາດພໍຮລາ', 'lt': 'polard fonetinė', 'mk': 'Полардово', 'ml': 'പൊള്ളാർഡ് ശബ്ദലിപി', 'mn': 'Пирд', 'mr': 'पोलार्ड फोनेटिक', 'ms': 'Fonetik Pollard', 'nb': 'pollard-fonetisk', 'ne': 'पोल्लार्ड फोनेटिक', 'nl': 'Pollard-fonetisch', 'nn': 'pollard-fonetisk', 'no': 'pollard-fonetisk', 'nqo': 'ߔߏߟߊߙ ߞߊ߲ߡߊߛߙߋߡߊ', 'or': 'ପୋଲାର୍ଡ ଫୋନେଟିକ୍', 'pl': 'fonetyczny Pollard’a', 'pt': 'fonético pollard', 'rm': 'fonetica da Pollard', 'ro': 'pollardă fonetică', 'ru': 'поллардовская фонетика', 'sc': 'pollard miao', 'sl': 'Pollardova fonetska pisava', 'so': 'Shibanaha', 'sq': 'polard fonetik', 'sr': 'поралд фонетско писмо', 'sr-Cyrl': 'поралд фонетско писмо', 'sr-Latn': 'porald fonetsko pismo', 'sv': 'pollardtecken', 'ta': 'போலார்ட் ஃபொனெட்டிக்', 'te': 'పోల్లర్డ్ ఫోనెటిక్', 'tg': 'Овоии поллардӣ', 'th': 'สัทศาสตร์พอลลาร์ด', 'to': 'tohinima fakafonētiki-polāti', 'tr': 'Pollard Fonetik', 'ug': 'پوللارد تاۋۇشلىرى', 'uk': 'писемність Полларда', 'vec': 'fonètego de Pollard', 'vi': 'Ngữ âm Pollard', 'xnr': 'पॉलार्ड फोनेटिक', 'yo': 'Fonẹtiiki Polaadi', 'yrl': 'funétiku miau', 'yue': '柏格理拼音符', 'yue-Hans': '柏格理拼音符', 'yue-Hant': '柏格理拼音符', 'zh': '波拉德音标文字', 'zh-Hans': '波拉德音标文字', 'zh-Hant': '柏格理拼音符', 'zu': 'isi-Pollard Phonetic'},
    'Prti': {'ast': 'partu d’inscripciones', 'bal-Latn': 'Kondahi Párti', 'bn': 'পার্থিয়ন', 'bs': 'pisani partian', 'bs-Cyrl': 'писани партиан', 'bs-Latn': 'pisani partian', 'ca': 'parthià inscripcional', 'ca-ES-valencia': 'parthià inscripcional', 'ccp': '𑄛𑄢𑄴𑄗𑄨𑄠𑄧𑄚𑄴', 'cs': 'parthské klínové', 'de': 'Parthisch', 'el': 'Επιγραφικό Παρθιάν', 'el-polyton': 'Επιγραφικό Παρθιάν', 'en': 'Inscriptional Parthian', 'en-Dsrt': '𐐆𐑌𐑅𐐿𐑉𐐮𐐹𐑇𐐲𐑌𐐲𐑊 𐐑𐐱𐑉𐑃𐐨𐐲𐑌', 'et': 'partia raidkiri', 'eu': 'Partiera inskripzioak', 'fa': 'پارتی کتیبه\u200cای', 'ff-Adlm': '𞤄𞤭𞤲𞤣𞤭 𞤆𞤢𞤪𞤧𞤭𞤴𞤢𞤲', 'fi': 'piirtokirjoitusparthialainen', 'fr': 'parthe des inscriptions', 'fy': 'Inscriptioneel Parthysk', 'ga': 'Pairtiach Inscríbhinniúil', 'gd': 'Partais snaidh-sgrìobhte', 'gu': 'ઇન્સ્ક્રિપ્શનલ પાર્થિયન', 'he': 'פרטית כתובה', 'hi': 'इंस्क्रिपश्नल पार्थियन', 'hi-Latn': 'Inscriptional Parthian', 'hr': 'pisani parthian', 'ht': 'parthe des inscriptions', 'hu': 'Feliratos parthian', 'id': 'Prasasti Parthia', 'ig': 'Inscriptional Parthian', 'it': 'partico delle iscrizioni', 'ja': '碑文パルティア文字', 'ka': 'მონუმენტური პართული', 'kgp': 'priti', 'kk': 'жазба парфия', 'kk-Cyrl': 'жазба парфия', 'kn': 'ಇನ್ಸ್\u200cಕ್ರಿಪ್ಶನಲ್ ಪಾರ್ಥಿಯನ್', 'ko': '명문 파라티아 문자', 'lo': 'ພາຮ໌ເທຍອິນສຄຮິປຊັນແນລ', 'lt': 'rašytiniai partų', 'mk': 'натписно партиско', 'ml': 'പൃതി', 'mn': 'Партиан', 'mr': 'इन्स्क्रिप्शनल पर्थियन', 'ms': 'Inskripsi Parthian', 'nb': 'inskripsjonsparthisk', 'ne': 'पिआरटी', 'nl': 'Inscriptioneel Parthisch', 'nn': 'inskripsjonsparthisk', 'no': 'inskripsjonsparthisk', 'nqo': 'ߔߊߙߑߕ ߛߓߍߟߌ', 'or': 'ଇନସ୍କ୍ରୀପସାନଲ୍ ପାର୍ଥିଆନ୍', 'pl': 'partyjski inskrypcyjny', 'rm': 'partic dad inscripziuns', 'ro': 'partă pentru inscripții', 'ru': 'парфянская', 'sc': 'pàrticu de sas iscritziones', 'so': 'Qoraalka Parthian', 'sq': 'persishte mbishkrimesh', 'sr': 'писани партиан', 'sr-Cyrl': 'писани партиан', 'sr-Latn': 'pisani partian', 'sv': 'tidig parthianska', 'ta': 'இன்ஸ்கிரிப்ஷனல் பார்த்தியன்', 'te': 'ఇంస్క్రిప్షనాల్ పార్థియన్', 'tg': 'Парфияи Хаттӣ', 'th': 'พาร์เทียอินสคริปชันแนล', 'to': 'tohinima fakapātia-tongi', 'tr': 'Partça Kitabe Dili', 'ug': 'پارتىئانچە ئويما خەت', 'uk': 'парфянський', 'vec': 'parto de le iscrision', 'vi': 'Chữ Parthia Văn bia', 'xnr': 'इंस्क्रिपश्नल पार्थियन', 'yue': '帕提亞文（碑銘體）', 'yue-Hans': '帕提亚文（碑铭体）', 'yue-Hant': '帕提亞文（碑銘體）', 'zh': '帕提亚文碑铭体', 'zh-Hans': '帕提亚文碑铭体', 'zh-Hant': '帕提亞文（碑銘體）', 'zu': 'i-Inscriptional Parthian'},
    'Psin': {'en': 'Proto-Sinaitic'},
    'QA': {'af': 'Katar', 'agq': 'Katà', 'ak': 'Kata', 'am': 'ኳታር', 'ar': 'قطر', 'as': 'কাটাৰ', 'asa': 'Katari', 'ast': 'Qatar', 'az': 'Qətər', 'az-Cyrl': 'Гәтәр', 'az-Latn': 'Qətər', 'bal-Latn': 'Gatar', 'bas': 'Kàtâr', 'be': 'Катар', 'be-tarask': 'Катар', 'bez': 'Hukatali', 'bg': 'Катар', 'bgn': 'قطر', 'blo': 'Kataa', 'bm': 'Katari', 'bn': 'কাতার', 'br': 'Qatar', 'brx': 'कातार', 'bs': 'Katar', 'bs-Cyrl': 'Катар', 'bs-Latn': 'Katar', 'ca': 'Qatar', 'ca-ES-valencia': 'Qatar', 'ccp': '𑄇𑄖𑄢𑄴', 'ce': 'Катар', 'ceb': 'Qatar', 'cgg': 'Kata', 'chr': 'ᎧᏔᎵ', 'ckb': 'قەتەر', 'cs': 'Katar', 'cv': 'Катар', 'cy': 'Qatar', 'da': 'Qatar', 'dav': 'Katari', 'de': 'Katar', 'dje': 'Kataar', 'dsb': 'Katar', 'dz': 'ཀ་ཊར', 'ebu': 'Katari', 'ee': 'Katar nutome', 'el': 'Κατάρ', 'el-polyton': 'Κατάρ', 'en': 'Qatar', 'en-Dsrt': '𐐗𐐲𐐻𐐪𐑉', 'eo': 'Kataro', 'es': 'Catar', 'et': 'Katar', 'eu': 'Qatar', 'ewo': 'Katár', 'fa': 'قطر', 'ff': 'Kataar', 'ff-Adlm': '𞤊𞤢𞤤𞤫𞤧𞤼𞤭𞥅𞤲', 'ff-Latn': 'Kataar', 'fi': 'Qatar', 'fil': 'Qatar', 'fo': 'Katar', 'fr': 'Qatar', 'fur': 'Qatar', 'fy': 'Qatar', 'ga': 'Catar', 'gd': 'Catar', 'gl': 'Qatar', 'gsw': 'Ggatar', 'gu': 'કતાર', 'guz': 'Katari', 'ha': 'Katar', 'he': 'קטאר', 'hi': 'क़तर', 'hi-Latn': 'Qatar', 'hr': 'Katar', 'hsb': 'Katar', 'ht': 'Qatar', 'hu': 'Katar', 'hy': 'Կատար', 'ia': 'Qatar', 'id': 'Qatar', 'ig': 'Qatar', 'is': 'Katar', 'it': 'Qatar', 'ja': 'カタール', 'jmc': 'Katari', 'jv': 'Katar', 'ka': 'კატარი', 'kaa': 'Қатар', 'kaa-Cyrl': 'Қатар', 'kab': 'Qaṭar', 'kam': 'Katari', 'kde': 'Katali', 'kea': 'Katar', 'kgp': 'Katar', 'khq': 'Kataar', 'ki': 'Katari', 'kk': 'Катар', 'kk-Cyrl': 'Катар', 'kln': 'Emetab Qatar', 'km': 'កាតា', 'kn': 'ಖತಾರ್', 'ko': '카타르', 'kok': 'कतार', 'kok-Deva': 'कतार', 'ks': 'قَطِر', 'ks-Arab': 'قَطِر', 'ksb': 'Katali', 'ksf': 'katá', 'ksh': 'Kataa', 'ku': 'Qeter', 'kxv': 'katar', 'kxv-Deva': 'क़तर', 'kxv-Latn': 'katar', 'kxv-Orya': 'କତର', 'kxv-Telu': 'కతార', 'ky': 'Катар', 'lag': 'Katáari', 'lb': 'Katar', 'lg': 'Kataa', 'lld': 'Qatar', 'ln': 'Katari', 'lo': 'ກາຕາ', 'lt': 'Kataras', 'lu': 'Katari', 'luo': 'Qatar', 'luy': 'Katari', 'lv': 'Katara', 'mai': 'कतर', 'mas': 'Katari', 'mer': 'Kata', 'mfe': 'Katar', 'mg': 'Katar', 'mgh': 'Ukatari', 'mi': 'Katā', 'mk': 'Катар', 'ml': 'ഖത്തർ', 'mn': 'Катар', 'mr': 'कतार', 'ms': 'Qatar', 'mt': 'il-Qatar', 'mua': 'Katar', 'my': 'ကာတာ', 'mzn': 'قطر', 'naq': 'Qatar', 'nb': 'Qatar', 'nd': 'Kathari', 'ne': 'कतार', 'nl': 'Qatar', 'nmg': 'Katar', 'nn': 'Qatar', 'no': 'Qatar', 'nqo': 'ߞߊߕߊ߯ߙ', 'nyn': 'Kata', 'om': 'Kuwaatar', 'or': 'କତାର୍', 'pa': 'ਕਤਰ', 'pa-Guru': 'ਕਤਰ', 'pcm': 'Kata', 'pl': 'Katar', 'ps': 'قطر', 'pt': 'Catar', 'qu': 'Qatar', 'rm': 'Katar', 'rn': 'Katari', 'ro': 'Qatar', 'rof': 'Katari', 'ru': 'Катар', 'rwk': 'Katari', 'saq': 'Katari', 'sbp': 'Katali', 'sc': 'Catar', 'sd': 'قطر', 'sd-Arab': 'قطر', 'se': 'Qatar', 'seh': 'Catar', 'ses': 'Kataar', 'sg': 'Katära', 'shi': 'ⵇⴰⵜⴰⵔ', 'shi-Latn': 'qatar', 'shi-Tfng': 'ⵇⴰⵜⴰⵔ', 'si': 'කටාර්', 'sk': 'Katar', 'sl': 'Katar', 'smn': 'Qatar', 'sn': 'Qatar', 'so': 'Qadar', 'sq': 'Katar', 'sr': 'Катар', 'sr-Cyrl': 'Катар', 'sr-Latn': 'Katar', 'sv': 'Qatar', 'sw': 'Qatar', 'syr': 'ܩܛܪ', 'ta': 'கத்தார்', 'te': 'ఖతార్', 'teo': 'Katari', 'tg': 'Қатар', 'th': 'กาตาร์', 'ti': 'ቐጠር', 'tk': 'Katar', 'to': 'Katā', 'tr': 'Katar', 'tt': 'Катар', 'twq': 'Kataar', 'tzm': 'Qaṭar', 'ug': 'قاتار', 'uk': 'Катар', 'ur': 'قطر', 'uz': 'Qatar', 'uz-Cyrl': 'Қатар', 'uz-Latn': 'Qatar', 'vai': 'ꕪꕚꕌ', 'vai-Latn': 'Kataha', 'vai-Vaii': 'ꕪꕚꕌ', 'vec': 'Catar', 'vi': 'Qatar', 'vun': 'Katari', 'wae': 'Katar', 'wo': 'Kataar', 'xh': 'EQatar', 'xnr': 'क़तर', 'xog': 'Kataa', 'yav': 'katáal', 'yi': 'קאַטאַר', 'yo': 'Kota', 'yrl': 'Katara', 'yue': '卡達', 'yue-Hans': '卡达', 'yue-Hant': '卡達', 'zgh': 'ⵇⴰⵜⴰⵔ', 'zh': '卡塔尔', 'zh-Hans': '卡塔尔', 'zh-Hant': '卡達', 'zu': 'i-Qatar'},
    'QM..QZ': {'en': 'Private use'},
    'QO': {'af': 'Afgeleë Oseanië', 'ak': 'Osiana Ano Ano', 'am': 'የኦሺያንያ ዳርቻ', 'ar': 'أوقيانوسيا النائية', 'as': 'আউটলায়িং অ’চিয়ানিয়া', 'ast': 'Oceanía esterior', 'az': 'Uzaq Okeaniya', 'az-Cyrl': 'Узаг Океанија', 'az-Latn': 'Uzaq Okeaniya', 'bal-Latn': 'Tálánén Zerbahrag', 'be': 'Аддаленыя тэрыторыі Акіяніі', 'be-tarask': 'Аддаленыя тэрыторыі Акіяніі', 'bg': 'Отдалечени острови на Океания', 'blo': 'Oseyanii kasaʊlǝŋka', 'bn': 'আউটলাইনিং ওসানিয়া', 'br': 'Oseania diabell', 'brx': 'बायजोआरि असियानिया', 'bs': 'Vanjska Okeanija', 'bs-Cyrl': 'Остала океанија', 'bs-Latn': 'Vanjska Okeanija', 'ca': 'Territoris allunyats d’Oceania', 'ca-ES-valencia': 'Territoris allunyats d’Oceania', 'ccp': '𑄃𑄅𑄪𑄑𑄣𑄭𑄚𑄨𑄁 𑄃𑄮𑄥𑄚𑄨𑄠', 'ce': 'Арахьара Океани', 'ceb': 'Outlying Oceania', 'chr': 'ᎠᏍᏛ ᎣᏏᏰᏂᎠ', 'ckb': 'دەرەوەی ئۆقیانووسیا', 'cs': 'vnější Oceánie', 'cv': 'Тулаш Океани', 'cy': 'Oceania Bellennig', 'da': 'Ydre Oceanien', 'de': 'Äußeres Ozeanien', 'dsb': 'wenkowna Oceaniska', 'dz': 'ཨོཤི་ཡཱན་ན་གྱི་མཐའ་མཚམས', 'ee': 'Outlaying Oceania nutome', 'el': 'Περιφερειακή Ωκεανία', 'el-polyton': 'Περιφερειακή Ωκεανία', 'en': 'Outlying Oceania', 'en-Dsrt': '𐐍𐐻𐑊𐐴𐐮𐑍 𐐄𐑇𐐨𐐰𐑌𐐨𐐲', 'es': 'Territorios alejados de Oceanía', 'et': 'Okeaania hajasaared', 'eu': 'Mugaz kanpoko Ozeania', 'fa': 'بخش\u200cهای دورافتادهٔ اقیانوسیه', 'ff-Adlm': '𞤚𞤢𞤼𞥆𞤫𞥅𞤪𞤭 𞤌𞤧𞤴𞤢𞤲𞤭𞤴𞤢', 'fi': 'Oseanian erillissaaret', 'fil': 'Outlying Oceania', 'fo': 'fjarskoti Osiania', 'fr': 'régions éloignées de l’Océanie', 'fur': 'Oceanie periferiche', 'fy': 'Oerig Oceanië', 'ga': 'an Aigéine Imeallach', 'gd': 'Roinn Iomallach a’ Chuain Sèimh', 'gl': 'Territorios afastados de Oceanía', 'gsw': 'Üssers Ozeaanie', 'gu': 'આઉટલાઈન્ગ ઓશનિયા', 'ha': 'Bakin Teku', 'he': 'טריטוריות באוקיאניה', 'hi': 'आउटलाइंग ओशिनिया', 'hi-Latn': 'Outlying Oceania', 'hr': 'Vanjska područja Oceanije', 'hsb': 'Wonkowna Oceaniska', 'ht': 'régions éloignées de l’Océanie', 'hu': 'Külső-Óceánia', 'hy': 'Արտաքին Օվկիանիա', 'ia': 'Oceania remote', 'id': 'Oseania Luar', 'ig': 'Outlying Oceania', 'is': 'Ytri Eyjaálfa', 'it': 'Oceania lontana', 'ja': 'オセアニア周辺地域', 'jv': 'Oseania Paling Njaba', 'ka': 'შორეული ოკეანეთი', 'kaa': 'Сыртқы Океания', 'kaa-Cyrl': 'Сыртқы Океания', 'kea': 'Ilhas di Oseania', 'kgp': 'Osiỹnĩja Kuvar-gy', 'kk': 'Сыртқы Океания', 'kk-Cyrl': 'Сыртқы Океания', 'km': 'តំបន់ជាយអូសេអានី', 'kn': 'ಔಟ್ ಲೈಯಿಂಗ್ ಓಷಿಯಾನಿಯಾ', 'ko': '오세아니아 외곽', 'kok': 'आवटलायींग ओशेनिया', 'kok-Deva': 'आवटलायींग ओशेनिया', 'ks': 'آوُٹلاینِگ اوشینِیا', 'ks-Arab': 'آوُٹلاینِگ اوشینِیا', 'ksh': 'Ozejahnije ußerhallef', 'ku': 'Okyanûsyaya Dûr', 'ky': 'Алыскы Океания', 'lb': 'Baussecht Ozeanien', 'lld': 'Ozeania dlafora', 'lo': 'ເຂດຫ່າງໄກໂອຊີເນຍ', 'lt': 'Nuošali Okeanija', 'lv': 'Okeānijas attālās salas', 'mai': 'आउटलाइंग ओसेनिया', 'mi': 'Ngā Moutere-a-Kiwa o Waho atu', 'mk': 'Зависни земји во Океанија', 'ml': 'ദ്വീപസമൂഹം', 'mn': 'Номхон далайг тойрсон улс орнууд', 'mr': 'आउटलाईंग ओशनिया', 'ms': 'Oceania Terpencil', 'my': 'သမုဒ္ဒရာ အပြင်ဘက်ရှိ ကျွန်းနိုင်ငံများ', 'mzn': 'اوقیانوسیه\u200cی ِپرت ِجائون', 'nb': 'Ytre Oseania', 'ne': 'बाह्य ओसनिया', 'nl': 'overig Oceanië', 'nn': 'Ytre Oseania', 'no': 'Ytre Oseania', 'nqo': 'ߟߌ߲ߓߊ߲ߘߎ߯ ߕߌ߲߬ߞߎߘߎ߲ ߦߙߐ߫ ߡߊߕߊ߯ߣߍ߲', 'om': 'Ooshiiniyaa Alaa', 'or': 'ସୀମାନ୍ତବର୍ତ୍ତୀ ଓସିନିଆ', 'pa': 'ਆਊਟਲਾਇੰਗ ਓਸ਼ੀਨੀਆ', 'pa-Guru': 'ਆਊਟਲਾਇੰਗ ਓਸ਼ੀਨੀਆ', 'pcm': 'Rimót Pát ọf Oshẹ́nia', 'pl': 'Oceania — wyspy dalekie', 'ps': 'بهرنۍ اوسيانه', 'pt': 'Oceania Remota', 'qu': 'Oceanía Periférica', 'rm': 'Oceania Periferica', 'ro': 'Oceania Periferică', 'ru': 'Внешняя Океания', 'sc': 'Otzeània perifèrica', 'sd': 'بيروني سامونڊي', 'sd-Arab': 'بيروني سامونڊي', 'si': 'ඈත ඕෂනියාව', 'sk': 'ostatné Tichomorie', 'sl': 'Ostala oceanija', 'so': 'Dhulxeebeedka Osheeniya', 'sq': 'Oqeania e Largët (Lindja e Largët)', 'sr': 'удаљена острва Океаније', 'sr-Cyrl': 'удаљена острва Океаније', 'sr-Latn': 'udaljena ostrva Okeanije', 'sv': 'yttre öar i Oceanien', 'sw': 'Oceania ya Nje', 'syr': 'ܐܘܩܝܢܘܣܝܐ ܒܪܝܬܐ', 'ta': 'வெளிப்புற ஓஷியானியா', 'te': 'ఒషీనియా బయటున్నవి', 'tg': 'Уқёнуси Дур', 'th': 'เอาต์ไลอิงโอเชียเนีย', 'ti': 'ካብ ኦሽያንያ ርሒቖም ግዝኣታት', 'tk': 'Daşky Okeaniýa', 'to': 'ʻOsēnia mamaʻo', 'tr': 'Uzak Okyanusya', 'tt': 'Ерак Океания', 'ug': 'ئوكيانىيە ئەتراپىدىكى ئاراللار', 'uk': 'Віддалена Океанія', 'ur': 'بیرونی اوشیانیا', 'uz': 'Tashqi Okeaniya', 'uz-Cyrl': 'Ёндош Океания', 'uz-Latn': 'Tashqi Okeaniya', 'vec': 'Oseania perifèrega', 'vi': 'Vùng xa xôi thuộc Châu Đại Dương', 'wae': 'Üssers Ozeanie', 'wo': 'Oceanie', 'xh': 'I-Oceania Esemaphandleni', 'xnr': 'आउटलाइंग ओशिनिया', 'yo': 'Agbègbè Òṣọ́ọ́níà', 'yrl': 'Useãniya (R)', 'yue': '大洋洲邊疆群島', 'yue-Hans': '大洋洲边疆群岛', 'yue-Hant': '大洋洲邊疆群島', 'zh': '大洋洲边远群岛', 'zh-Hans': '大洋洲边远群岛', 'zh-Hant': '大洋洲邊疆群島', 'zu': 'i-Outlying Oceania'},
    'Qaaa..Qabx': {'en': 'Private use'},
    'Qaag': {'ar': 'زوجيي', 'bal-Latn': 'Zawgi', 'bs': 'zavgji pismo', 'bs-Cyrl': 'завгји писмо', 'bs-Latn': 'zavgji pismo', 'ca': 'zawgyi', 'ca-ES-valencia': 'zawgyi', 'cs': 'zawgyi', 'da': 'zawgyi', 'de': 'Zawgyi', 'en': 'Zawgyi', 'es': 'zawgyi', 'eu': 'zauagiera', 'ff-Adlm': '𞤟𞤢𞤱𞤺𞤭𞥅𞤴𞤵', 'fi': 'burmalainen zawgyi-toteutus', 'fr': 'zawgyi', 'gd': 'Zawgyi', 'he': 'זאוגיי', 'hi': 'ज़ौजी', 'hi-Latn': 'Zawgyi', 'hr': 'zawgyi', 'ht': 'zawgyi', 'hu': 'Zawgyi', 'id': 'Zawgyi', 'ig': 'Zawgyi', 'it': 'zawgyi', 'kk': 'зоджи жазуы', 'kk-Cyrl': 'зоджи жазуы', 'ko': '저지 문자', 'ku': 'zawgyi', 'mn': 'Каак', 'ms': 'Zawgyi', 'nb': 'zawgyi', 'nl': 'Zawgyi', 'nn': 'zawgyi', 'no': 'zawgyi', 'nqo': 'ߖ߭ߊߎߜ߭ߌ', 'pl': 'zawgyi', 'pt': 'zawgyi', 'ro': 'zawgyi', 'ru': 'зоджи', 'sc': 'zawgyi', 'sk': 'zawgyi', 'so': 'Qoraalka Sawgiga', 'sq': 'zaugi', 'sv': 'zawgyi', 'tg': 'Завгӯйӣ', 'tr': 'Zawgyi', 'uk': 'зоджі', 'vec': 'zovghi', 'vi': 'Chữ Zawgyi', 'zh': '照济文', 'zh-Hans': '照济文', 'zh-Hant': '佐基文', 'zu': 'i-Zawgyi'},
    'RE': {'af': 'Réunion', 'agq': 'Lèyunyɔ̀ŋ', 'ak': 'Reyuniɔn', 'am': 'ሪዩኒየን', 'ar': 'روينيون', 'as': 'ৰিইউনিয়ন', 'asa': 'Riyunioni', 'ast': 'Reunión', 'az': 'Reyunyon', 'az-Cyrl': 'Рејунјон', 'az-Latn': 'Reyunyon', 'bal-Latn': 'Réyunian', 'bas': 'Rèunyɔ̂ŋ', 'be': 'Рэюньён', 'be-tarask': 'Рэюньён', 'bez': 'Huliyunioni', 'bg': 'Реюнион', 'blo': 'Reeniyɔŋ', 'bm': 'Reyuɲɔŋ', 'bn': 'রিইউনিয়ন', 'br': 'Ar Reünion', 'brx': 'रेयूनियन', 'bs': 'Reunion', 'bs-Cyrl': 'Реинион', 'bs-Latn': 'Reunion', 'ca': 'Illa de la Reunió', 'ca-ES-valencia': 'Illa de la Reunió', 'ccp': '𑄢𑄨𑄃𑄨𑄃𑄪𑄚𑄨𑄠𑄧𑄚𑄴', 'ce': 'Реюньон', 'ceb': 'Reunion', 'cgg': 'Riyuniyoni', 'chr': 'ᎴᏳᏂᎠᏂ', 'ckb': 'ڕییوونیەن', 'cs': 'Réunion', 'cv': 'Реюньон', 'cy': 'Réunion', 'da': 'Réunion', 'dav': 'Riyunioni', 'de': 'Réunion', 'dje': 'Reenioŋ', 'dsb': 'Réunion', 'dz': 'རེ་ཡུ་ནི་ཡོན', 'ebu': 'Riyunioni', 'ee': 'Réunion nutome', 'el': 'Ρεϊνιόν', 'el-polyton': 'Ρεϊνιόν', 'en': 'Réunion', 'en-Dsrt': '𐐡𐐨𐐷𐐭𐑌𐐷𐐲𐑌', 'eo': 'Reunio', 'es': 'Reunión', 'et': 'Réunion', 'eu': 'Reunion', 'ewo': 'Reuniɔ́ŋ', 'fa': 'رئونیون', 'ff': 'Rewiñoo', 'ff-Adlm': '𞤈𞤫𞥅𞤲𞤭𞤴𞤮𞤲', 'ff-Latn': 'Rewiñoo', 'fi': 'Réunion', 'fil': 'Réunion', 'fo': 'Réunion', 'fr': 'La Réunion', 'fur': 'Reunion', 'fy': 'Réunion', 'ga': 'La Réunion', 'gd': 'Réunion', 'gl': 'Reunión', 'gsw': 'Reünioon', 'gu': 'રીયુનિયન', 'guz': 'Riyunioni', 'ha': 'Rawuniyan', 'he': 'ראוניון', 'hi': 'रियूनियन', 'hi-Latn': 'Reunion', 'hr': 'Réunion', 'hsb': 'Réunion', 'ht': 'La Réunion', 'hu': 'Réunion', 'hy': 'Ռեյունիոն', 'ia': 'Reunion', 'id': 'Réunion', 'ig': 'Réunion', 'is': 'Réunion', 'it': 'Riunione', 'ja': 'レユニオン', 'jgo': 'Lɛ́uniyɔ̂n', 'jmc': 'Riyunioni', 'jv': 'Réunion', 'ka': 'რეუნიონი', 'kaa': 'Реюньон', 'kaa-Cyrl': 'Реюньон', 'kab': 'Timlilit', 'kam': 'Riyunioni', 'kde': 'Liyunioni', 'kea': 'Runion', 'kgp': 'Hujáv', 'khq': 'Reenioŋ', 'ki': 'Riyunioni', 'kk': 'Реюньон', 'kk-Cyrl': 'Реюньон', 'kln': 'Emetab Réunion', 'km': 'រេអុយញ៉ុង', 'kn': 'ರಿಯೂನಿಯನ್', 'ko': '레위니옹', 'kok': 'रीयूनियन', 'kok-Deva': 'रीयूनियन', 'ks': 'رِیوٗنِیَن', 'ks-Arab': 'رِیوٗنِیَن', 'ksb': 'Liyunioni', 'ksf': 'rɛunyɔŋ', 'ksh': 'Rehunjohn', 'ku': 'Réunion', 'kxv': 'riyuniyn', 'kxv-Deva': 'रियूनियन', 'kxv-Latn': 'riyuniyn', 'kxv-Orya': 'ରିୟୁନିୟନ', 'kxv-Telu': 'రీయూనియన్', 'ky': 'Реюньон', 'lag': 'Reyunióoni', 'lb': 'Réunion', 'lg': 'Leyunyoni', 'lld': 'Réunion', 'ln': 'Lenyo', 'lo': 'ເຣອູນິຍົງ', 'lt': 'Reunjonas', 'lu': 'Lenyo', 'luo': 'Réunion', 'luy': 'Riyunioni', 'lv': 'Reinjona', 'mai': 'रियूनियन', 'mas': 'Riyunioni', 'mer': 'Riyunioni', 'mfe': 'Larenion', 'mg': 'Larenion', 'mgh': 'Uriyunioni', 'mi': 'Reūnio', 'mk': 'Рејунион', 'ml': 'റീയൂണിയൻ', 'mn': 'Реюнион', 'mr': 'रियुनियन', 'ms': 'Reunion', 'mt': 'Réunion', 'mua': 'Sǝr Reunion', 'my': 'ရီယူနီယန်', 'mzn': 'رئونیون', 'naq': 'Réunion', 'nb': 'Réunion', 'nd': 'Réunion', 'ne': 'रियुनियन', 'nl': 'Réunion', 'nmg': 'Réuniɔn', 'nn': 'Réunion', 'no': 'Réunion', 'nqo': 'ߟߊ߬ߘߍ', 'nyn': 'Riyuniyoni', 'om': 'Riyuuniyeen', 'or': 'ରିୟୁନିଅନ୍', 'pa': 'ਰਿਯੂਨੀਅਨ', 'pa-Guru': 'ਰਿਯੂਨੀਅਨ', 'pcm': 'Réyúniọn', 'pl': 'Reunion', 'ps': 'ریونین', 'pt': 'Reunião', 'qu': 'Reunión', 'rm': 'Réunion', 'rn': 'Amazinga ya Reyiniyo', 'ro': 'Réunion', 'rof': 'Riyunioni', 'ru': 'Реюньон', 'rwk': 'Riyunioni', 'saq': 'Riyunioni', 'sbp': 'Liyunioni', 'sc': 'Riunione', 'sd': 'ري يونين', 'sd-Arab': 'ري يونين', 'se': 'Réunion', 'seh': 'Reunião', 'ses': 'Reenioŋ', 'sg': 'Reinïon', 'shi': 'ⵔⵉⵢⵓⵏⵢⵓⵏ', 'shi-Latn': 'riyunyun', 'shi-Tfng': 'ⵔⵉⵢⵓⵏⵢⵓⵏ', 'si': 'රීයුනියන්', 'sk': 'Réunion', 'sl': 'Reunion', 'smn': 'Réunion', 'sn': 'Réunion', 'so': 'Riyuuniyon', 'sq': 'Reunion', 'sr': 'Реинион', 'sr-Cyrl': 'Реинион', 'sr-Latn': 'Reinion', 'sv': 'Réunion', 'sw': 'Reunion', 'syr': 'ܪܝܘܢܝܘܢ', 'ta': 'ரீயூனியன்', 'te': 'రీయూనియన్', 'teo': 'Riyunioni', 'tg': 'Реюнион', 'th': 'เรอูนียง', 'ti': 'ርዩንየን', 'tk': 'Reýunýon', 'to': 'Lēunioni', 'tr': 'Reunion', 'tt': 'Реюньон', 'twq': 'Reenioŋ', 'tzm': 'Riyyunyun', 'ug': 'رېيۇنىيون', 'uk': 'Реюньйон', 'ur': 'ری یونین', 'uz': 'Reyunion', 'uz-Cyrl': 'Реюнион', 'uz-Latn': 'Reyunion', 'vai': 'ꔓꗠꖻ', 'vai-Latn': 'Renyɔɔ̃', 'vai-Vaii': 'ꔓꗠꖻ', 'vec': 'Ìzola Reunion', 'vi': 'Réunion', 'vun': 'Riyunioni', 'wae': 'Réunion', 'wo': 'Reeñoo', 'xh': 'ERéunion', 'xnr': 'रियूनियन', 'xog': 'Leyunyoni', 'yav': 'elewuniɔ́ŋ', 'yi': 'רעאוניאן', 'yo': 'Riuniyan', 'yrl': 'Yumuatirisawa', 'yue': '留尼旺', 'yue-Hans': '留尼旺', 'yue-Hant': '留尼旺', 'zgh': 'ⵔⵉⵢⵓⵏⵢⵓⵏ', 'zh': '留尼汪', 'zh-Hans': '留尼汪', 'zh-Hant': '留尼旺', 'zu': 'i-Réunion'},
    'RO': {'af': 'Roemenië', 'agq': 'Lùmanyìa', 'ak': 'Romenia', 'am': 'ሮሜኒያ', 'ar': 'رومانيا', 'as': 'ৰোমানিয়া', 'asa': 'Romania', 'ast': 'Rumanía', 'az': 'Rumıniya', 'az-Cyrl': 'Румынија', 'az-Latn': 'Rumıniya', 'bal-Latn': 'Rumániá', 'bas': 'Rùmanìà', 'be': 'Румынія', 'be-tarask': 'Румынія', 'bez': 'Hulomania', 'bg': 'Румъния', 'bgn': 'رومانیه', 'blo': 'Romanii', 'bm': 'Rumani', 'bn': 'রোমানিয়া', 'br': 'Roumania', 'brx': 'रमानिया', 'bs': 'Rumunija', 'bs-Cyrl': 'Румунија', 'bs-Latn': 'Rumunija', 'ca': 'Romania', 'ca-ES-valencia': 'Romania', 'ccp': '𑄢𑄮𑄟𑄚𑄨𑄠', 'ce': 'Румыни', 'ceb': 'Romania', 'cgg': 'Romaniya', 'chr': 'ᎶᎹᏂᏯ', 'ckb': 'ڕۆمانیا', 'cs': 'Rumunsko', 'cv': 'Румыни', 'cy': 'Rwmania', 'da': 'Rumænien', 'dav': 'Romania', 'de': 'Rumänien', 'dje': 'Rumaani', 'dsb': 'Rumuńska', 'dz': 'རོ་མེ་ནི་ཡ', 'ebu': 'Romania', 'ee': 'Romania nutome', 'el': 'Ρουμανία', 'el-polyton': 'Ρουμανία', 'en': 'Romania', 'en-Dsrt': '𐐡𐐬𐑋𐐩𐑌𐐨𐐲', 'eo': 'Rumanujo', 'es': 'Rumanía', 'et': 'Rumeenia', 'eu': 'Errumania', 'ewo': 'Rumaní', 'fa': 'رومانی', 'ff': 'Rumanii', 'ff-Adlm': '𞤈𞤵𞤥𞤢𞥄𞤲𞤭𞤴𞤢', 'ff-Latn': 'Rumanii', 'fi': 'Romania', 'fil': 'Romania', 'fo': 'Rumenia', 'fr': 'Roumanie', 'fur': 'Romanie', 'fy': 'Roemenië', 'ga': 'an Rómáin', 'gd': 'Romàinia', 'gl': 'Romanía', 'gsw': 'Rumäänie', 'gu': 'રોમાનિયા', 'guz': 'Romania', 'ha': 'Romaniya', 'he': 'רומניה', 'hi': 'रोमानिया', 'hi-Latn': 'Romania', 'hr': 'Rumunjska', 'hsb': 'Rumunska', 'ht': 'Roumanie', 'hu': 'Románia', 'hy': 'Ռումինիա', 'ia': 'Romania', 'id': 'Rumania', 'ie': 'Rumania', 'ig': 'Romania', 'is': 'Rúmenía', 'it': 'Romania', 'ja': 'ルーマニア', 'jmc': 'Romania', 'jv': 'Ruméni', 'ka': 'რუმინეთი', 'kaa': 'Румыния', 'kaa-Cyrl': 'Румыния', 'kab': 'Rumani', 'kam': 'Romania', 'kde': 'Lomania', 'kea': 'Ruménia', 'kgp': 'Homẽnĩja', 'khq': 'Rumaani', 'ki': 'Romania', 'kk': 'Румыния', 'kk-Cyrl': 'Румыния', 'kln': 'Emetab Romania', 'km': 'រូម៉ានី', 'kn': 'ರೊಮೇನಿಯಾ', 'ko': '루마니아', 'kok': 'रोमानीया', 'kok-Deva': 'रोमानीया', 'ks': 'رومانِیا', 'ks-Arab': 'رومانِیا', 'ksb': 'Lomania', 'ksf': 'rɔmaní', 'ksh': 'Romäänije', 'ku': 'Romanya', 'kxv': 'romaniya', 'kxv-Deva': 'रोमानिया', 'kxv-Latn': 'romaniya', 'kxv-Orya': 'ରମାନିଆ', 'kxv-Telu': 'రోమేనియా', 'ky': 'Румыния', 'lag': 'Romaníia', 'lb': 'Rumänien', 'lg': 'Lomaniya', 'lld': 'Romania', 'ln': 'Romani', 'lo': 'ໂຣແມເນຍ', 'lt': 'Rumunija', 'lu': 'Romani', 'luo': 'Romania', 'luy': 'Romania', 'lv': 'Rumānija', 'mai': 'रोमानिया', 'mas': 'Romania', 'mer': 'Romania', 'mfe': 'Roumani', 'mg': 'Romania', 'mgh': 'Uromania', 'mi': 'Romeinia', 'mk': 'Романија', 'ml': 'റൊമാനിയ', 'mn': 'Румын', 'mr': 'रोमानिया', 'ms': 'Romania', 'mt': 'ir-Rumanija', 'mua': 'Romaniya', 'my': 'ရိုမေးနီးယား', 'mzn': 'رومانی', 'naq': 'Romania', 'nb': 'Romania', 'nd': 'Romania', 'ne': 'रोमेनिया', 'nl': 'Roemenië', 'nmg': 'Roumania', 'nn': 'Romania', 'no': 'Romania', 'nqo': 'ߙߎߡߊߣߌ߫', 'nyn': 'Romaniya', 'om': 'Roomaaniyaa', 'or': 'ରୋମାନିଆ', 'pa': 'ਰੋਮਾਨੀਆ', 'pa-Guru': 'ਰੋਮਾਨੀਆ', 'pcm': 'Ruménia', 'pl': 'Rumunia', 'ps': 'رومانیا', 'pt': 'Romênia', 'qu': 'Rumania', 'rm': 'Rumenia', 'rn': 'Rumaniya', 'ro': 'România', 'rof': 'Romania', 'ru': 'Румыния', 'rwk': 'Romania', 'saq': 'Romania', 'sbp': 'Lomaniya', 'sc': 'Romania', 'sd': 'رومانيا', 'sd-Arab': 'رومانيا', 'se': 'Románia', 'seh': 'Romênia', 'ses': 'Rumaani', 'sg': 'Rumanïi', 'shi': 'ⵔⵓⵎⴰⵏⵢⴰ', 'shi-Latn': 'rumanya', 'shi-Tfng': 'ⵔⵓⵎⴰⵏⵢⴰ', 'si': 'රුමේනියාව', 'sk': 'Rumunsko', 'sl': 'Romunija', 'smn': 'Romania', 'sn': 'Romania', 'so': 'Rumaaniya', 'sq': 'Rumani', 'sr': 'Румунија', 'sr-Cyrl': 'Румунија', 'sr-Latn': 'Rumunija', 'sv': 'Rumänien', 'sw': 'Romania', 'syr': 'ܪܘܡܢܝܐ', 'ta': 'ருமேனியா', 'te': 'రోమేనియా', 'teo': 'Romania', 'tg': 'Руминия', 'th': 'โรมาเนีย', 'ti': 'ሩማንያ', 'tk': 'Rumyniýa', 'to': 'Lomēnia', 'tr': 'Romanya', 'tt': 'Румыния', 'twq': 'Rumaani', 'tzm': 'Ṛumanya', 'ug': 'رومىنىيە', 'uk': 'Румунія', 'ur': 'رومانیہ', 'uz': 'Ruminiya', 'uz-Cyrl': 'Руминия', 'uz-Latn': 'Ruminiya', 'vai': 'ꖄꕆꕇꕰ', 'vai-Latn': 'Romininya', 'vai-Vaii': 'ꖄꕆꕇꕰ', 'vec': 'Romania', 'vi': 'Romania', 'vun': 'Romania', 'wae': 'Rumänie', 'wo': 'Rumani', 'xh': 'ERomania', 'xnr': 'रोमानिया', 'xog': 'Lomaniya', 'yav': 'ulumaní', 'yi': 'רומעניע', 'yo': 'Romaniya', 'yrl': 'Romẽniya', 'yue': '羅馬尼亞', 'yue-Hans': '罗马尼亚', 'yue-Hant': '羅馬尼亞', 'zgh': 'ⵔⵓⵎⴰⵏⵢⴰ', 'zh': '罗马尼亚', 'zh-Hans': '罗马尼亚', 'zh-Hant': '羅馬尼亞', 'zu': 'i-Romania'},
    'RS': {'af': 'Serwië', 'ak': 'Sɛbia', 'am': 'ሰርብያ', 'ar': 'صربيا', 'as': 'ছাৰ্বিয়া', 'ast': 'Serbia', 'az': 'Serbiya', 'az-Cyrl': 'Сербија', 'az-Latn': 'Serbiya', 'bal-Latn': 'Sarbiá', 'be': 'Сербія', 'be-tarask': 'Сербія', 'bg': 'Сърбия', 'bgn': 'سیربستان', 'blo': 'Sɛrbii', 'bn': 'সার্বিয়া', 'br': 'Serbia', 'brx': 'सार्बिया', 'bs': 'Srbija', 'bs-Cyrl': 'Србија', 'bs-Latn': 'Srbija', 'ca': 'Sèrbia', 'ca-ES-valencia': 'Sèrbia', 'ccp': '𑄥𑄢𑄴𑄝𑄨𑄠', 'ce': 'Серби', 'ceb': 'Serbia', 'chr': 'ᏒᏈᏯ', 'ckb': 'سربیا', 'cs': 'Srbsko', 'cv': 'Серби', 'cy': 'Serbia', 'da': 'Serbien', 'de': 'Serbien', 'dsb': 'Serbiska', 'dz': 'སཱར་བྷི་ཡ', 'el': 'Σερβία', 'el-polyton': 'Σερβία', 'en': 'Serbia', 'en-Dsrt': '𐐝𐐲𐑉𐐺𐐨𐐲', 'es': 'Serbia', 'et': 'Serbia', 'eu': 'Serbia', 'fa': 'صربستان', 'ff-Adlm': '𞤅𞤫𞤪𞤦𞤭𞤴𞤢𞥄', 'fi': 'Serbia', 'fil': 'Serbia', 'fo': 'Serbia', 'fr': 'Serbie', 'fur': 'Serbie', 'fy': 'Servië', 'ga': 'an tSeirbia', 'gd': 'An t-Sèirb', 'gl': 'Serbia', 'gsw': 'Särbie', 'gu': 'સર્બિયા', 'ha': 'Sabiya', 'he': 'סרביה', 'hi': 'सर्बिया', 'hi-Latn': 'Serbia', 'hr': 'Srbija', 'hsb': 'Serbiska', 'ht': 'Serbie', 'hu': 'Szerbia', 'hy': 'Սերբիա', 'ia': 'Serbia', 'id': 'Serbia', 'ie': 'Serbia', 'ig': 'Serbia', 'is': 'Serbía', 'it': 'Serbia', 'ja': 'セルビア', 'jgo': 'Sɛlɛbî', 'jv': 'Sèrbi', 'ka': 'სერბეთი', 'kaa': 'Сербия', 'kaa-Cyrl': 'Сербия', 'kea': 'Sérvia', 'kgp': 'Sérvija', 'kk': 'Сербия', 'kk-Cyrl': 'Сербия', 'km': 'សែប៊ី', 'kn': 'ಸೆರ್ಬಿಯಾ', 'ko': '세르비아', 'kok': 'सर्बिया', 'kok-Deva': 'सर्बिया', 'ks': 'سَربِیا', 'ks-Arab': 'سَربِیا', 'ksh': 'Särbije', 'ku': 'Sirbistan', 'kxv': 'serbia', 'kxv-Deva': 'सर्बिया', 'kxv-Latn': 'serbia', 'kxv-Orya': 'ସର୍ବିୟା', 'kxv-Telu': 'సెర్బియా', 'ky': 'Сербия', 'lb': 'Serbien', 'lld': 'Serbia', 'ln': 'Serbie', 'lo': 'ເຊີເບຍ', 'lt': 'Serbija', 'lv': 'Serbija', 'mai': 'सर्बिया', 'mi': 'Hirupia', 'mk': 'Србија', 'ml': 'സെർബിയ', 'mn': 'Серби', 'mr': 'सर्बिया', 'ms': 'Serbia', 'mt': 'is-Serbja', 'my': 'ဆားဘီးယား', 'mzn': 'صربستون', 'nb': 'Serbia', 'ne': 'सर्बिया', 'nl': 'Servië', 'nn': 'Serbia', 'no': 'Serbia', 'nqo': 'ߛߍߙߑߓߌ߫', 'om': 'Serbiyaa', 'or': 'ସର୍ବିଆ', 'pa': 'ਸਰਬੀਆ', 'pa-Guru': 'ਸਰਬੀਆ', 'pcm': 'Sẹ́bia', 'pl': 'Serbia', 'ps': 'سربيا', 'pt': 'Sérvia', 'qu': 'Serbia', 'rm': 'Serbia', 'ro': 'Serbia', 'ru': 'Сербия', 'sc': 'Sèrbia', 'sd': 'سربيا', 'sd-Arab': 'سربيا', 'se': 'Serbia', 'si': 'සර්බියාව', 'sk': 'Srbsko', 'sl': 'Srbija', 'smn': 'Serbia', 'so': 'Seerbiya', 'sq': 'Serbi', 'sr': 'Србија', 'sr-Cyrl': 'Србија', 'sr-Latn': 'Srbija', 'sv': 'Serbien', 'sw': 'Serbia', 'syr': 'ܣܪܒܝܐ', 'ta': 'செர்பியா', 'te': 'సెర్బియా', 'tg': 'Сербия', 'th': 'เซอร์เบีย', 'ti': 'ሰርብያ', 'tk': 'Serbiýa', 'to': 'Sēpia', 'tr': 'Sırbistan', 'tt': 'Сербия', 'ug': 'سېربىيە', 'uk': 'Сербія', 'ur': 'سربیا', 'uz': 'Serbiya', 'uz-Cyrl': 'Сербия', 'uz-Latn': 'Serbiya', 'vai': 'ꗻꗡꔬꕩ', 'vai-Vaii': 'ꗻꗡꔬꕩ', 'vec': 'Serbia', 'vi': 'Serbia', 'wae': 'Serbie', 'wo': 'Serbi', 'xh': 'ESerbia', 'xnr': 'सर्बिया', 'yi': 'סערביע', 'yo': 'Sẹ́bíà', 'yrl': 'Sewiya', 'yue': '塞爾維亞', 'yue-Hans': '塞尔维亚', 'yue-Hant': '塞爾維亞', 'zgh': 'ⵙⵉⵔⴱⵢⴰ', 'zh': '塞尔维亚', 'zh-Hans': '塞尔维亚', 'zh-Hant': '塞爾維亞', 'zu': 'i-Serbia'},
    'RU': {'af': 'Rusland', 'agq': 'Loshìa', 'ak': 'Rɔhyea', 'am': 'ሩስያ', 'ar': 'روسيا', 'as': 'ৰাছিয়া', 'asa': 'Uruthi', 'ast': 'Rusia', 'az': 'Rusiya', 'az-Cyrl': 'Русија', 'az-Latn': 'Rusiya', 'bal-Latn': 'Rus', 'bas': 'Ruslànd', 'be': 'Расія', 'be-tarask': 'Расія', 'bez': 'Huulusi', 'bg': 'Русия', 'blo': 'Rɔɔshɩya', 'bm': 'Irisi', 'bn': 'রাশিয়া', 'bo': 'ཨུ་རུ་སུ་', 'br': 'Rusia', 'brx': 'रासिया', 'bs': 'Rusija', 'bs-Cyrl': 'Русија', 'bs-Latn': 'Rusija', 'ca': 'Rússia', 'ca-ES-valencia': 'Rússia', 'ccp': '𑄢𑄥𑄨𑄠', 'ce': 'Росси', 'ceb': 'Russia', 'cgg': 'Rrasha', 'chr': 'ᏲᏂᎢ', 'ckb': 'ڕووسیا', 'cs': 'Rusko', 'cv': 'Раҫҫей', 'cy': 'Rwsia', 'da': 'Rusland', 'dav': 'Urusi', 'de': 'Russland', 'dje': 'Iriši laboo', 'doi': 'रूस', 'dsb': 'Ruska', 'dz': 'ཨུ་རུ་སུ', 'ebu': 'Urusi', 'ee': 'Russia nutome', 'el': 'Ρωσία', 'el-polyton': 'Ρωσία', 'en': 'Russia', 'en-Dsrt': '𐐡𐐲𐑇𐐲', 'eo': 'Rusujo', 'es': 'Rusia', 'et': 'Venemaa', 'eu': 'Errusia', 'ewo': 'Rúsian', 'fa': 'روسیه', 'ff': 'Riisii', 'ff-Adlm': '𞤈𞤮𞥅𞤧𞤭𞤴𞤢', 'ff-Latn': 'Riisii', 'fi': 'Venäjä', 'fil': 'Russia', 'fo': 'Russland', 'fr': 'Russie', 'fur': 'Russie', 'fy': 'Ruslân', 'ga': 'an Rúis', 'gd': 'An Ruis', 'gl': 'Rusia', 'gsw': 'Russland', 'gu': 'રશિયા', 'guz': 'Urusi', 'ha': 'Rasha', 'haw': 'Lūkia', 'he': 'רוסיה', 'hi': 'रूस', 'hi-Latn': 'Russia', 'hr': 'Rusija', 'hsb': 'Ruska', 'ht': 'Russie', 'hu': 'Oroszország', 'hy': 'Ռուսաստան', 'ia': 'Russia', 'id': 'Rusia', 'ie': 'Russia', 'ig': 'Russia', 'ii': 'ꊉꇆꌦ', 'is': 'Rússland', 'it': 'Russia', 'ja': 'ロシア', 'jgo': 'Lusî', 'jmc': 'Urusi', 'jv': 'Rusia', 'ka': 'რუსეთი', 'kaa': 'Россия', 'kaa-Cyrl': 'Россия', 'kab': 'Rrus', 'kam': 'Urusi', 'kde': 'Ulusi', 'kea': 'Rúsia', 'kgp': 'Husija', 'khq': 'Iriši laboo', 'ki': 'Urusi', 'kk': 'Ресей', 'kk-Cyrl': 'Ресей', 'kln': 'Emetab Russia', 'km': 'រុស្ស៊ី', 'kn': 'ರಷ್ಯಾ', 'ko': '러시아', 'kok': 'रूस', 'kok-Deva': 'रूस', 'kok-Latn': 'Roxya', 'ks': 'روٗس', 'ks-Arab': 'روٗس', 'ks-Deva': 'रूस', 'ksb': 'Ulusi', 'ksf': 'risí', 'ksh': 'Rußland', 'ku': 'Rûsya', 'kxv': 'russia', 'kxv-Deva': 'रूस', 'kxv-Latn': 'russia', 'kxv-Orya': 'ରୁସିଆ', 'kxv-Telu': 'రస్యా', 'ky': 'Россия', 'lag': 'Urúusi', 'lb': 'Russland', 'lg': 'Lasa', 'lld': 'Ruscia', 'ln': 'Risí', 'lo': 'ຣັດເຊຍ', 'lrc': 'روٙسیە', 'lt': 'Rusija', 'lu': 'Risi', 'luo': 'Russia', 'luy': 'Urusi', 'lv': 'Krievija', 'mai': 'रूस', 'mas': 'Urusi', 'mer': 'Rashia', 'mfe': 'Larisi', 'mg': 'Rosia', 'mi': 'Rūhia', 'mk': 'Русија', 'ml': 'റഷ്യ', 'mn': 'Орос', 'mni': 'রুসিয়া', 'mni-Beng': 'রুসিয়া', 'mr': 'रशिया', 'ms': 'Rusia', 'mt': 'ir-Russja', 'mua': 'Russiya', 'my': 'ရုရှား', 'mzn': 'روسیه', 'naq': 'Rasiab', 'nb': 'Russland', 'nd': 'Rashiya', 'ne': 'रूस', 'nl': 'Rusland', 'nmg': 'Russi', 'nn': 'Russland', 'no': 'Russland', 'nqo': 'ߌߙߌ߬ߛߌ߫', 'nyn': 'Rrasha', 'om': 'Raashiyaa', 'or': 'ରୁଷିଆ', 'os': 'Уӕрӕсе', 'pa': 'ਰੂਸ', 'pa-Guru': 'ਰੂਸ', 'pcm': 'Rọ́shia', 'pl': 'Rosja', 'ps': 'روسیه', 'pt': 'Rússia', 'qu': 'Rusia', 'rm': 'Russia', 'rn': 'Uburusiya', 'ro': 'Rusia', 'rof': 'Urusi', 'ru': 'Россия', 'rwk': 'Urusi', 'sa': 'रष्यदेश:', 'sah': 'Арассыыйа', 'saq': 'Urusi', 'sat': 'ᱨᱩᱥ', 'sat-Olck': 'ᱨᱩᱥ', 'sbp': 'Wulusi', 'sc': 'Rùssia', 'sd': 'روس', 'sd-Arab': 'روس', 'sd-Deva': 'रशिया', 'se': 'Ruošša', 'seh': 'Rússia', 'ses': 'Iriši laboo', 'sg': 'Rusïi', 'shi': 'ⵔⵓⵙⵢⴰ', 'shi-Latn': 'rusya', 'shi-Tfng': 'ⵔⵓⵙⵢⴰ', 'si': 'රුසියාව', 'sk': 'Rusko', 'sl': 'Rusija', 'smn': 'Ruoššâ', 'sn': 'Russia', 'so': 'Ruush', 'sq': 'Rusi', 'sr': 'Русија', 'sr-Cyrl': 'Русија', 'sr-Latn': 'Rusija', 'su': 'Rusia', 'su-Latn': 'Rusia', 'sv': 'Ryssland', 'sw': 'Urusi', 'syr': 'ܪܘܣܝܐ', 'ta': 'ரஷ்யா', 'te': 'రష్యా', 'teo': 'Urusi', 'tg': 'Русия', 'th': 'รัสเซีย', 'ti': 'ሩስያ', 'tk': 'Russiýa', 'to': 'Lūsia', 'tr': 'Rusya', 'tt': 'Россия', 'twq': 'Iriši laboo', 'tzm': 'Rusya', 'ug': 'رۇسىيە', 'uk': 'Росія', 'ur': 'روس', 'uz': 'Rossiya', 'uz-Cyrl': 'Россия', 'uz-Latn': 'Rossiya', 'vai': 'ꗐꖺꔻꕩ', 'vai-Latn': 'Rɔshiya', 'vai-Vaii': 'ꗐꖺꔻꕩ', 'vec': 'Rusia', 'vi': 'Nga', 'vun': 'Urusi', 'wae': 'Russland', 'wo': 'Risi', 'xh': 'ERashiya', 'xnr': 'रूस', 'xog': 'Lasa', 'yav': 'ulusí', 'yi': 'רוסלאַנד', 'yo': 'Rọṣia', 'yrl': 'Rusiya', 'yue': '俄羅斯', 'yue-Hans': '俄罗斯', 'yue-Hant': '俄羅斯', 'zgh': 'ⵔⵓⵙⵢⴰ', 'zh': '俄罗斯', 'zh-Hans': '俄罗斯', 'zh-Hant': '俄羅斯', 'zu': 'i-Russia'},
    'RW': {'af': 'Rwanda', 'agq': 'Lùwandà', 'ak': 'Rewanda', 'am': 'ሩዋንዳ', 'ar': 'رواندا', 'as': 'ৰোৱাণ্ডা', 'asa': 'Rwanda', 'ast': 'Ruanda', 'az': 'Ruanda', 'az-Cyrl': 'Руанда', 'az-Latn': 'Ruanda', 'bal-Latn': 'Rwándhá', 'bas': 'Rùandà', 'be': 'Руанда', 'be-tarask': 'Руанда', 'bez': 'Hulwanda', 'bg': 'Руанда', 'bgn': 'روندا', 'blo': 'Rʊwanɖaa', 'bm': 'Ruwanda', 'bn': 'রুয়ান্ডা', 'br': 'Rwanda', 'brx': 'रूआण्डा', 'bs': 'Ruanda', 'bs-Cyrl': 'Руанда', 'bs-Latn': 'Ruanda', 'ca': 'Ruanda', 'ca-ES-valencia': 'Ruanda', 'ccp': '𑄢𑄪𑄠𑄚𑄴𑄓', 'ce': 'Руанда', 'ceb': 'Rwanda', 'cgg': 'Rwanda', 'chr': 'ᎶᏩᏂᏓ', 'ckb': 'ڕواندا', 'cs': 'Rwanda', 'cv': 'Руанда', 'cy': 'Rwanda', 'da': 'Rwanda', 'dav': 'Rwanda', 'de': 'Ruanda', 'dje': 'Rwanda', 'dsb': 'Ruanda', 'dz': 'རུ་ཝན་ཌ', 'ebu': 'Rwanda', 'ee': 'Rwanda nutome', 'el': 'Ρουάντα', 'el-polyton': 'Ρουάντα', 'en': 'Rwanda', 'en-Dsrt': '𐐡𐐲𐐶𐐪𐑌𐐼𐐲', 'eo': 'Ruando', 'es': 'Ruanda', 'et': 'Rwanda', 'eu': 'Ruanda', 'ewo': 'Ruwandá', 'fa': 'رواندا', 'ff': 'Ruwanndaa', 'ff-Adlm': '𞤈𞤵𞤱𞤢𞤲𞤣𞤢𞥄', 'ff-Latn': 'Ruwanndaa', 'fi': 'Ruanda', 'fil': 'Rwanda', 'fo': 'Ruanda', 'fr': 'Rwanda', 'fur': 'Ruande', 'fy': 'Rwanda', 'ga': 'Ruanda', 'gd': 'Rubhanda', 'gl': 'Ruanda', 'gsw': 'Ruanda', 'gu': 'રવાંડા', 'guz': 'Rwanda', 'ha': 'Ruwanda', 'he': 'רואנדה', 'hi': 'रवांडा', 'hi-Latn': 'Rwanda', 'hr': 'Ruanda', 'hsb': 'Ruanda', 'ht': 'Rwanda', 'hu': 'Ruanda', 'hy': 'Ռուանդա', 'ia': 'Ruanda', 'id': 'Rwanda', 'ig': 'Rwanda', 'is': 'Rúanda', 'it': 'Ruanda', 'ja': 'ルワンダ', 'jgo': 'Luwánda', 'jmc': 'Rwanda', 'jv': 'Rwanda', 'ka': 'რუანდა', 'kaa': 'Руанда', 'kaa-Cyrl': 'Руанда', 'kab': 'Ruwanda', 'kam': 'Lwanda', 'kde': 'Lwanda', 'kea': 'Ruanda', 'kgp': 'Huỹna', 'khq': 'Rwanda', 'ki': 'Rwanda', 'kk': 'Руанда', 'kk-Cyrl': 'Руанда', 'kln': 'Emetab Rwanda', 'km': 'រវ៉ាន់ដា', 'kn': 'ರುವಾಂಡಾ', 'ko': '르완다', 'kok': 'रवांडा', 'kok-Deva': 'रवांडा', 'ks': 'روٗوانڈا', 'ks-Arab': 'روٗوانڈا', 'ksb': 'Lwanda', 'ksf': 'rwanda', 'ksh': 'Ruanda', 'ku': 'Rwanda', 'kxv': 'rvanḍa', 'kxv-Deva': 'र्-वांडा', 'kxv-Latn': 'rvanḍa', 'kxv-Orya': 'ର୍\u200dୱାଣ୍ଡା', 'kxv-Telu': 'రువాండా', 'ky': 'Руанда', 'lag': 'Rwáanda', 'lb': 'Ruanda', 'lg': 'Rwanda', 'lld': 'Ruanda', 'ln': 'Rwanda', 'lo': 'ຣວັນດາ', 'lt': 'Ruanda', 'lu': 'Rwanda', 'luo': 'Rwanda', 'luy': 'Rwanda', 'lv': 'Ruanda', 'mai': 'रवांडा', 'mas': 'Rwanda', 'mer': 'Rwanda', 'mfe': 'Rwanda', 'mg': 'Roanda', 'mgh': 'Urwanda', 'mi': 'Rāwana', 'mk': 'Руанда', 'ml': 'റുവാണ്ട', 'mn': 'Руанда', 'mr': 'रवांडा', 'ms': 'Rwanda', 'mt': 'ir-Rwanda', 'mua': 'Rwanda', 'my': 'ရဝန်ဒါ', 'mzn': 'روآندا', 'naq': 'Rwandab', 'nb': 'Rwanda', 'nd': 'Ruwanda', 'ne': 'रवाण्डा', 'nl': 'Rwanda', 'nmg': 'Rwanda', 'nn': 'Rwanda', 'no': 'Rwanda', 'nqo': 'ߙߎߥߊ߲ߘߊ߫', 'nyn': 'Rwanda', 'om': 'Ruwwandaa', 'or': 'ରାୱାଣ୍ଡା', 'pa': 'ਰਵਾਂਡਾ', 'pa-Guru': 'ਰਵਾਂਡਾ', 'pcm': 'Ruwánda', 'pl': 'Rwanda', 'ps': 'روندا', 'pt': 'Ruanda', 'qu': 'Ruanda', 'rm': 'Ruanda', 'rn': 'u Rwanda', 'ro': 'Rwanda', 'rof': 'Rwanda', 'ru': 'Руанда', 'rw': 'U Rwanda', 'rwk': 'Rwanda', 'saq': 'Rwanda', 'sbp': 'Lwanda', 'sc': 'Ruanda', 'sd': 'روانڊا', 'sd-Arab': 'روانڊا', 'se': 'Rwanda', 'seh': 'Ruanda', 'ses': 'Rwanda', 'sg': 'Ruandäa', 'shi': 'ⵔⵡⴰⵏⴷⴰ', 'shi-Latn': 'rwanda', 'shi-Tfng': 'ⵔⵡⴰⵏⴷⴰ', 'si': 'රුවන්ඩාව', 'sk': 'Rwanda', 'sl': 'Ruanda', 'smn': 'Ruanda', 'sn': 'Rwanda', 'so': 'Ruwanda', 'sq': 'Ruandë', 'sr': 'Руанда', 'sr-Cyrl': 'Руанда', 'sr-Latn': 'Ruanda', 'sv': 'Rwanda', 'sw': 'Rwanda', 'syr': 'ܪܘܐܢܕܐ', 'ta': 'ருவாண்டா', 'te': 'రువాండా', 'teo': 'Rwanda', 'tg': 'Руанда', 'th': 'รวันดา', 'ti': 'ርዋንዳ', 'tk': 'Ruanda', 'to': 'Luanitā', 'tr': 'Ruanda', 'tt': 'Руанда', 'twq': 'Rwanda', 'tzm': 'Ruwwanḍa', 'ug': 'رىۋاندا', 'uk': 'Руанда', 'ur': 'روانڈا', 'uz': 'Ruanda', 'uz-Cyrl': 'Руанда', 'uz-Latn': 'Ruanda', 'vai': 'ꕟꖙꕡ', 'vai-Latn': 'Rawunda', 'vai-Vaii': 'ꕟꖙꕡ', 'vec': 'Ruanda', 'vi': 'Rwanda', 'vun': 'Rwanda', 'wae': 'Ruanda', 'wo': 'Ruwànda', 'xh': 'ERwanda', 'xnr': 'रवांडा', 'xog': 'Rwanda', 'yav': 'uluándá', 'yi': 'רוואַנדע', 'yo': 'Ruwanda', 'yrl': 'Huãda', 'yue': '盧安達', 'yue-Hans': '卢安达', 'yue-Hant': '盧安達', 'zgh': 'ⵔⵡⴰⵏⴷⴰ', 'zh': '卢旺达', 'zh-Hans': '卢旺达', 'zh-Hant': '盧安達', 'zu': 'i-Rwanda'},
    'Ranj': {'en': 'Ranjana'},
    'Rjng': {'ast': 'rejang', 'az': 'recəng', 'az-Latn': 'recəng', 'bal-Latn': 'Rejang', 'bn': 'রেজ্যাঙ্গ', 'brx': 'रेजेंग', 'bs': 'rejang pismo', 'bs-Cyrl': 'рејанг писмо', 'bs-Latn': 'rejang pismo', 'ca': 'rejang', 'ca-ES-valencia': 'rejang', 'ccp': '𑄢𑄬𑄎𑄳𑄠𑄋𑄴𑄉𑄧', 'cs': 'redžanské', 'da': 'rejang', 'de': 'Rejang', 'el': 'Ρετζάνγκ', 'el-polyton': 'Ρετζάνγκ', 'en': 'Rejang', 'en-Dsrt': '𐐡𐐲𐐾𐐰𐑍', 'es': 'rejang', 'et': 'redžangi', 'eu': 'Rejang', 'fa': 'رجنگی', 'ff-Adlm': '𞤈𞤭𞤶𞤢𞤲𞤺', 'fi': 'rejang', 'fr': 'rejang', 'fy': 'Rejang', 'ga': 'Rejang', 'gd': 'Rejang', 'gsw': 'Rejang', 'gu': 'રીજાંગ', 'he': 'רג׳אנג', 'hi': 'रीजांग', 'hi-Latn': 'Rejang', 'hr': 'rejang pismo', 'ht': 'rejang', 'hu': 'Redzsang', 'id': 'Rejang', 'ig': 'Rejang', 'it': 'rejang', 'ja': 'ルジャン文字', 'ka': 'რეჯანგი', 'kgp': 'rejỹg', 'kk': 'реджанг жазуы', 'kk-Cyrl': 'реджанг жазуы', 'kn': 'ರೆಜಾಂಗ್', 'ko': '레장 문자', 'ks': 'ریجَنگ', 'ks-Arab': 'ریجَنگ', 'lb': 'Rejang', 'lo': 'ເຮຈັງ', 'lt': 'rejang', 'mk': 'реџаншко', 'ml': 'റെജാംഗ്', 'mn': 'Режанг', 'mr': 'रीजांग', 'ms': 'Rejang', 'nb': 'rejang', 'ne': 'रेजाङ', 'nl': 'Rejang', 'nn': 'rejang', 'no': 'rejang', 'nqo': 'ߙߋߖ߭ߊ߲ߜ߭', 'or': 'ରେଜାଙ୍ଗ', 'pl': 'rejang', 'pt': 'rejang', 'rm': 'rejang', 'ro': 'rejang', 'ru': 'реджангская', 'sc': 'rejang', 'so': 'Dadka Rejan', 'sq': 'rexhangisht', 'sr': 'рејанг писмо', 'sr-Cyrl': 'рејанг писмо', 'sr-Latn': 'rejang pismo', 'sv': 'rejang', 'ta': 'ரெஜெய்ன்', 'te': 'రేజాంగ్', 'tg': 'Реҷанг', 'th': 'เรจัง', 'to': 'tohinima fakalesiangi', 'tr': 'Rejang', 'ug': 'رېجاڭچە', 'uk': 'реджанг', 'vec': 'rejang', 'vi': 'Chữ Rejang', 'xnr': 'रीजांग', 'yrl': 'reyãgi', 'yue': '拉讓文', 'yue-Hans': '拉让文', 'yue-Hant': '拉讓文', 'zh': '拉让文', 'zh-Hans': '拉让文', 'zh-Hant': '拉讓文', 'zu': 'i-Rejang'},
    'Rohg': {'af': 'Hanifi', 'ak': 'Hanifi kasa', 'am': 'ሃኒፊ', 'ar': 'الحنيفي', 'as': 'হানিফি', 'az': 'hanifi', 'az-Latn': 'hanifi', 'bal-Latn': 'Hanipi', 'be': 'ханіфі', 'be-tarask': 'ханіфі', 'bg': 'харифи', 'bn': 'হানিফি', 'brx': 'हानिफि', 'bs': 'hanifi pismo', 'bs-Cyrl': 'ханифи писмо', 'bs-Latn': 'hanifi pismo', 'ca': 'hanifi', 'ca-ES-valencia': 'hanifi', 'chr': 'ᎭᏂᏫ', 'cs': 'hanifi', 'cy': 'Hanifi', 'da': 'hanifi', 'de': 'Hanifi Rohingya', 'dsb': 'hanifiske pismo', 'el': 'Χανίφι', 'el-polyton': 'Χανίφι', 'en': 'Hanifi', 'es': 'hanifi', 'et': 'rohingja', 'eu': 'hanifia', 'fa': 'حنیفی', 'ff-Adlm': '𞤖𞤢𞤲𞤭𞤬𞤭', 'fi': 'rohingalainen hanifi', 'fil': 'Hanifi', 'fo': 'hanifi', 'fr': 'hanifi', 'ga': 'Hanifi', 'gd': 'Hanifi Rohingya', 'gl': 'hanifi', 'gu': 'હનીફી', 'ha': 'Hanifi', 'he': 'חניפי', 'hi': 'हनिफ़ि', 'hi-Latn': 'Hanifi', 'hr': 'hanifi pismo', 'hsb': 'hanifiske pismo', 'ht': 'hanifi', 'hu': 'Hanifi', 'hy': 'հանիֆի', 'id': 'Hanifi', 'ig': 'Hanifi', 'is': 'hanifi', 'it': 'hanifi', 'ja': 'ロヒンギャ文字', 'jv': 'Hanifi', 'ka': 'ჰანიფი', 'kk': 'ханифи жазуы', 'kk-Cyrl': 'ханифи жазуы', 'km': 'ហានីហ្វ៊ី', 'kn': 'ಹನೀಫಿ', 'ko': '하니피 문자', 'kok': 'हानिफी', 'kok-Deva': 'हानिफी', 'ky': 'Ханифи (жазуу)', 'lo': 'ຮານິຟີ', 'lt': 'Hanifi', 'lv': 'hanifi', 'mai': 'हनिफी', 'mi': 'Hāniwhi', 'mk': 'ханифи', 'ml': 'ഹനിഫി', 'mn': 'Ханафи бичиг', 'mr': 'हनीफी', 'ms': 'Hanifi Rohingya', 'my': 'ဟာနီဖီ', 'nb': 'hanifi', 'ne': 'हानिफी', 'nl': 'Hanifi Rohingya', 'nn': 'hanifi', 'no': 'hanifi', 'nqo': 'ߤߊߣߝߌ', 'or': 'ହାନିଫି', 'pa': 'ਹਨੀਫੀ', 'pa-Guru': 'ਹਨੀਫੀ', 'pcm': 'Hanífi Lángwej', 'pl': 'hanifi', 'ps': 'حنفی', 'pt': 'hanifi', 'qu': 'Hanifi Simi', 'ro': 'hanifi', 'ru': 'ханифи', 'sc': 'hanifi rohingya', 'sd': 'ھنيفي', 'sd-Arab': 'ھنيفي', 'si': 'හනිෆි', 'sk': 'hanifi', 'sl': 'hanifi', 'so': 'Hanifi', 'sq': 'hanifi', 'sr': 'ханифи', 'sr-Cyrl': 'ханифи', 'sr-Latn': 'hanifi', 'sv': 'hanifiska', 'sw': 'Kihanifi', 'ta': 'ஹனிஃபி', 'te': 'హనీఫీ', 'tg': 'Ханифӣ', 'th': 'ฮะนีฟี', 'ti': 'ሃኒፊ', 'tk': 'hanifi', 'to': 'tohinima fakahanifi-lohingia', 'tr': 'Hanifi', 'uk': 'ханіфі', 'ur': 'حنیفی', 'uz': 'hanifi', 'uz-Latn': 'hanifi', 'vec': 'hanifi', 'vi': 'Chữ Hanifi', 'yo': 'Èdè Hanifi', 'yue': '哈尼菲文', 'yue-Hans': '哈尼菲文', 'yue-Hant': '哈尼菲文', 'zh': '哈乃斐文', 'zh-Hans': '哈乃斐文', 'zh-Hant': '哈乃斐羅興亞文', 'zu': 'isi-Hanifi Rohingya'},
    'Roro': {'ar': 'رنجورنجو', 'ast': 'rongorongo', 'az': 'ronqoronqo', 'az-Latn': 'ronqoronqo', 'bg': 'Ронго-ронго', 'bn': 'রোঙ্গোরোঙ্গো', 'brx': 'रोंगोरोंगो', 'bs': 'rongorongo pismo', 'bs-Cyrl': 'ронгоронго писмо', 'bs-Latn': 'rongorongo pismo', 'ca': 'rongo-rongo', 'ca-ES-valencia': 'rongo-rongo', 'ccp': '𑄢𑄮𑄋𑄴𑄉𑄮𑄢𑄮𑄋𑄴𑄉𑄮', 'cs': 'rongorongo', 'da': 'rongo-rongo', 'de': 'Rongorongo', 'el': 'Ρονγκορόνγκο', 'el-polyton': 'Ρονγκορόνγκο', 'en': 'Rongorongo', 'en-Dsrt': '𐐡𐐪𐑍𐑀𐐬𐑉𐐪𐑌𐑀𐐬', 'es': 'rongo-rongo', 'et': 'rongorongo', 'fi': 'rongorongo', 'fr': 'rongorongo', 'fy': 'Rongorongo', 'gd': 'Rongorongo', 'gsw': 'Rongorongo', 'gu': 'રોંગોરોંગો', 'hi': 'रोन्गोरोन्गो', 'hi-Latn': 'Rongorongo', 'hr': 'rongorongo pismo', 'ht': 'rongorongo', 'hu': 'Rongorongo', 'id': 'Rongorongo', 'it': 'rongorongo', 'ja': 'ロンゴロンゴ文字', 'ka': 'რონგორონგო', 'kgp': 'rãgorãgo', 'kn': 'ರೋಂಗೋರೋಂಗೋ', 'ko': '롱고롱고', 'ks': 'رونگو رونگو', 'ks-Arab': 'رونگو رونگو', 'lb': 'Rongorongo', 'lo': 'ຮອງໂກຮອງໂກ', 'lt': 'rongorongo', 'lv': 'rongorongo', 'mk': 'ронгоронго', 'ml': 'റൊംഗോറൊംഗോ', 'mr': 'रोन्गोरोन्गो', 'nb': 'rongorongo', 'ne': 'रोङ्गोरोङ्गो', 'nl': 'Rongorongo', 'nn': 'rongorongo', 'no': 'rongorongo', 'or': 'ରୋଙ୍ଗୋରୋଙ୍ଗୋ', 'pl': 'rongorongo', 'pt': 'rongorongo', 'rm': 'rongorongo', 'ru': 'ронго-ронго', 'sl': 'rongorongo', 'sr': 'ронгоронго писмо', 'sr-Cyrl': 'ронгоронго писмо', 'sr-Latn': 'rongorongo pismo', 'sv': 'rongo-rongo', 'ta': 'ரொங்கோரொங்கோ', 'te': 'రోంగో రోంగో', 'th': 'รองโกรองโก', 'to': 'tohinima fakalongolongo', 'tr': 'Rongorongo', 'ug': 'روڭگوروڭگو', 'uk': 'ронго-ронго', 'vi': 'Chữ Rongorongo', 'yrl': 'rũgurũgu', 'yue': '朗格朗格象形文', 'yue-Hans': '朗格朗格象形文', 'yue-Hant': '朗格朗格象形文', 'zh': '朗格朗格文', 'zh-Hans': '朗格朗格文', 'zh-Hant': '朗格朗格象形文'},
    'Runr': {'am': 'ሩኒክ', 'ar': 'الروني', 'ast': 'runes', 'az': 'runik', 'az-Latn': 'runik', 'bal-Latn': 'Runi', 'bg': 'Руническа', 'bn': 'রুনিক', 'br': 'runek', 'brx': 'रूनिक', 'bs': 'runsko pismo', 'bs-Cyrl': 'рунско писмо', 'bs-Latn': 'runsko pismo', 'ca': 'rúnic', 'ca-ES-valencia': 'rúnic', 'ccp': '𑄢𑄪𑄚𑄨𑄇𑄴', 'cs': 'runové', 'da': 'runer', 'de': 'Runenschrift', 'el': 'Ρουνίκ', 'el-polyton': 'Ρουνίκ', 'en': 'Runic', 'en-Dsrt': '𐐡𐐭𐑌𐐮𐐿', 'es': 'rúnico', 'et': 'ruunikiri', 'eu': 'errunikoa', 'fa': 'رونی', 'ff-Adlm': '𞤈𞤵𞤲𞤭𞥅𞤳', 'fi': 'riimukirjoitus', 'fo': 'rúnar', 'fr': 'runique', 'fur': 'runic', 'fy': 'Runic', 'ga': 'Rúnach', 'gd': 'Rùn-sgrìobhadh', 'gsw': 'Runäschrift', 'gu': 'રૂનિક', 'he': 'רוני', 'hi': 'रूनिक', 'hi-Latn': 'Runic', 'hr': 'runsko pismo', 'ht': 'runique', 'hu': 'Runikus', 'id': 'Runik', 'ig': 'Runic', 'it': 'runico', 'ja': 'ルーン文字', 'ka': 'რუნული', 'kgp': 'runĩku', 'kk': 'руна', 'kk-Cyrl': 'руна', 'kn': 'ರೂನಿಕ್', 'ko': '룬 문자', 'ks': 'رَنِک', 'ks-Arab': 'رَنِک', 'lb': 'Runeschrëft', 'lo': 'ຮູນິກ', 'lt': 'runų', 'lv': 'rūnu raksts', 'mk': 'рунско', 'ml': 'റുണിക്', 'mn': 'Руни', 'mr': 'रूनिक', 'ms': 'Runic', 'nb': 'runer', 'ne': 'रूनिक', 'nl': 'Runic', 'nn': 'runer', 'no': 'runer', 'nqo': 'ߙߎߣߌߞ', 'or': 'ରନିକ୍', 'pl': 'runiczne', 'pt': 'rúnico', 'rm': 'runic', 'ro': 'runică', 'ru': 'руническая', 'sc': 'rùnicu', 'sk': 'Runové písmo', 'sl': 'runski', 'so': 'Dadka Rejang', 'sq': 'runike', 'sr': 'рунско писмо', 'sr-Cyrl': 'рунско писмо', 'sr-Latn': 'runsko pismo', 'sv': 'runor', 'ta': 'ருனிக்', 'te': 'రూనిక్', 'tg': 'Руникӣ', 'th': 'รูนิก', 'to': 'tohinima fakaluniki', 'tr': 'Runik', 'ug': 'رۇنىكچە', 'uk': 'рунічний', 'vec': 'rùnego', 'vi': 'Chữ Runic', 'xnr': 'रूनिक', 'yrl': 'rúniku', 'yue': '古北歐文字', 'yue-Hans': '古北欧文字', 'yue-Hant': '古北歐文字', 'zh': '古代北欧文', 'zh-Hans': '古代北欧文', 'zh-Hant': '古北歐文字', 'zu': 'i-Runic'},
    'SA': {'af': 'Saoedi-Arabië', 'agq': 'Sawudi Alabi', 'ak': 'Saudi Arabia', 'am': 'ሳውድአረቢያ', 'ar': 'المملكة العربية السعودية', 'as': 'চৌডি আৰবিয়া', 'asa': 'Thaudi', 'ast': 'Arabia Saudita', 'az': 'Səudiyyə Ərəbistanı', 'az-Cyrl': 'Сәудијјә Әрәбистаны', 'az-Latn': 'Səudiyyə Ərəbistanı', 'bal-Latn': 'Saudi Arab', 'bas': 'Sàudi Àrabìà', 'be': 'Саудаўская Аравія', 'be-tarask': 'Саудаўская Аравія', 'bez': 'Husaudi', 'bg': 'Саудитска Арабия', 'blo': 'Sauɖiya', 'bm': 'Arabiya Sawudiya', 'bn': 'সৌদি আরব', 'br': 'Arabia Saoudat', 'brx': 'साउदि आराबिया', 'bs': 'Saudijska Arabija', 'bs-Cyrl': 'Саудијска Арабија', 'bs-Latn': 'Saudijska Arabija', 'ca': 'Aràbia Saudí', 'ca-ES-valencia': 'Aràbia Saudí', 'ccp': '𑄥𑄯𑄘𑄨 𑄃𑄢𑄧𑄝𑄴', 'ce': 'СаӀудийн Ӏаьрбийчоь', 'ceb': 'Saudi Arabia', 'cgg': 'Saudi Areebiya', 'chr': 'ᏌᎤᏗ ᎡᎴᏈᎠ', 'ckb': 'عەرەبستانی سەعوودی', 'cs': 'Saúdská Arábie', 'cv': 'Сауд Аравийӗ', 'cy': 'Saudi Arabia', 'da': 'Saudi-Arabien', 'dav': 'Saudi', 'de': 'Saudi-Arabien', 'dje': 'Saudiya', 'dsb': 'Saudi-Arabiska', 'dyo': 'Abari Saudi', 'dz': 'སཱཝ་དི་ ཨ་རེ་བྷི་ཡ', 'ebu': 'Saudi', 'ee': 'Saudi Arabia nutome', 'el': 'Σαουδική Αραβία', 'el-polyton': 'Σαουδικὴ Ἀραβία', 'en': 'Saudi Arabia', 'en-Dsrt': '𐐝𐐵𐐼𐐨 𐐊𐑉𐐩𐐺𐐨𐐲', 'eo': 'Sauda Arabujo', 'es': 'Arabia Saudí', 'et': 'Saudi Araabia', 'eu': 'Saudi Arabia', 'ewo': 'Arabí Saudí', 'fa': 'عربستان سعودی', 'ff': 'Arabii Sawdit', 'ff-Adlm': '𞤅𞤢𞤵𞥅𞤣 𞤀𞥄𞤪𞤢𞤦𞤭𞤴𞤢𞥄', 'ff-Latn': 'Arabii Sawdit', 'fi': 'Saudi-Arabia', 'fil': 'Saudi Arabia', 'fo': 'Saudiarabia', 'fr': 'Arabie saoudite', 'fur': 'Arabie Saudide', 'fy': 'Saoedi-Arabië', 'ga': 'an Araib Shádach', 'gd': 'Aràibia nan Sabhd', 'gl': 'Arabia Saudita', 'gsw': 'Saudi-Araabie', 'gu': 'સાઉદી અરેબિયા', 'guz': 'Saudi', 'ha': 'Saudiyya', 'he': 'ערב הסעודית', 'hi': 'सऊदी अरब', 'hi-Latn': 'Saudi Arabia', 'hr': 'Saudijska Arabija', 'hsb': 'Sawdi-Arabska', 'ht': 'Arabie saoudite', 'hu': 'Szaúd-Arábia', 'hy': 'Սաուդյան Արաբիա', 'ia': 'Arabia Saudita', 'id': 'Arab Saudi', 'ig': 'Saudi Arabia', 'is': 'Sádi-Arabía', 'it': 'Arabia Saudita', 'ja': 'サウジアラビア', 'jmc': 'Saudi', 'jv': 'Arab Saudi', 'ka': 'საუდის არაბეთი', 'kaa': 'Саудия Арабстаны', 'kaa-Cyrl': 'Саудия Арабстаны', 'kab': 'Suɛudiya Taɛrabt', 'kam': 'Saudi', 'kde': 'Saudia', 'kea': 'Arábia Saudita', 'kgp': 'Aramija Savnita', 'khq': 'Saudiya', 'ki': 'Saudi', 'kk': 'Сауд Арабиясы', 'kk-Cyrl': 'Сауд Арабиясы', 'kln': 'Emetab Saudi Arabia', 'km': 'អារ៉ាប៊ីសាអូឌីត', 'kn': 'ಸೌದಿ ಅರೇಬಿಯಾ', 'ko': '사우디아라비아', 'kok': 'सऊदी अरेबिया', 'kok-Deva': 'सऊदी अरेबिया', 'ks': 'سعودی عرب', 'ks-Arab': 'سعودی عرب', 'ksb': 'Saudi', 'ksf': 'arabí saodí', 'ksh': 'Saudi Arraabije', 'ku': 'Erebistana Siûdî', 'kxv': 'saūdi arabiya', 'kxv-Deva': 'सऊदी अरब', 'kxv-Latn': 'saūdi arabiya', 'kxv-Orya': 'ସାଉଦି ଆରବ', 'kxv-Telu': 'సౌదీ అరేబియా', 'ky': 'Сауд Арабиясы', 'lag': 'Saudíia Arabíia', 'lb': 'Saudi-Arabien', 'lg': 'Sawudarebya - Buwarabu', 'lld': 'Arabia Saudita', 'ln': 'Alabi Sawuditɛ', 'lo': 'ຊາອຸດິ ອາຣາເບຍ', 'lt': 'Saudo Arabija', 'lu': 'Alabu Nsawudi', 'luo': 'Saudi Arabia', 'luy': 'Saudi', 'lv': 'Saūda Arābija', 'mai': 'सऊदी अरब', 'mas': 'Saudi', 'mer': 'Saudi', 'mfe': 'Larabi Saoudit', 'mg': 'Arabia saodita', 'mgh': 'Usaudi', 'mi': 'Hauri Arāpia', 'mk': 'Саудиска Арабија', 'ml': 'സൗദി അറേബ്യ', 'mn': 'Саудын Араб', 'mr': 'सौदी अरब', 'ms': 'Arab Saudi', 'mt': 'l-Arabja Sawdija', 'mua': 'Sǝr Arabiya', 'my': 'ဆော်ဒီအာရေးဘီးယား', 'mzn': 'عربستون', 'naq': 'Saudi Arabiab', 'nb': 'Saudi-Arabia', 'nd': 'Saudi Arabia', 'ne': 'साउदी अरब', 'nl': 'Saoedi-Arabië', 'nmg': 'Saudi Arabia', 'nn': 'Saudi-Arabia', 'no': 'Saudi-Arabia', 'nqo': 'ߛߎ߰ߘߎ߬ߟߊ߫-ߡߊ߲߬ߛߊ߬ߟߊ', 'nyn': 'Saudi Areebiya', 'om': 'Saawud Arabiyaa', 'or': 'ସାଉଦି ଆରବିଆ', 'pa': 'ਸਾਊਦੀ ਅਰਬ', 'pa-Guru': 'ਸਾਊਦੀ ਅਰਬ', 'pcm': 'Saúdí Arébia', 'pl': 'Arabia Saudyjska', 'ps': 'سعودي عربستان', 'pt': 'Arábia Saudita', 'qu': 'Arabia Saudí', 'rm': 'Arabia Saudita', 'rn': 'Arabiya Sawudite', 'ro': 'Arabia Saudită', 'rof': 'Saudi', 'ru': 'Саудовская Аравия', 'rwk': 'Saudi', 'saq': 'Saudi', 'sbp': 'Sawudi', 'sc': 'Aràbia Saudita', 'sd': 'سعودي عرب', 'sd-Arab': 'سعودي عرب', 'se': 'Saudi-Arábia', 'seh': 'Arábia Saudita', 'ses': 'Saudiya', 'sg': 'Saûdi Arabïi', 'shi': 'ⵙⵙⴰⵄⵓⴷⵉⵢⴰ', 'shi-Latn': 'ssaɛudiya', 'shi-Tfng': 'ⵙⵙⴰⵄⵓⴷⵉⵢⴰ', 'si': 'සෞදි අරාබිය', 'sk': 'Saudská Arábia', 'sl': 'Saudova Arabija', 'smn': 'Saudi Arabia', 'sn': 'Saudi Arabia', 'so': 'Sacuudi Carabiya', 'sq': 'Arabi Saudite', 'sr': 'Саудијска Арабија', 'sr-Cyrl': 'Саудијска Арабија', 'sr-Latn': 'Saudijska Arabija', 'sv': 'Saudiarabien', 'sw': 'Saudia', 'syr': 'ܣܥܘܕܝܐ', 'ta': 'சவுதி அரேபியா', 'te': 'సౌదీ అరేబియా', 'teo': 'Saudi', 'tg': 'Арабистони Саудӣ', 'th': 'ซาอุดีอาระเบีย', 'ti': 'ስዑዲ ዓረብ', 'tk': 'Saud Arabystany', 'to': 'Saute ʻAlepea', 'tr': 'Suudi Arabistan', 'tt': 'Согуд Гарәбстаны', 'twq': 'Saudiya', 'tzm': 'Ssaεudiyya Taεrabt', 'ug': 'سەئۇدىي ئەرەبىستان', 'uk': 'Саудівська Аравія', 'ur': 'سعودی عرب', 'uz': 'Saudiya Arabistoni', 'uz-Cyrl': 'Саудия Арабистони', 'uz-Latn': 'Saudiya Arabistoni', 'vai': 'ꕞꕌꖝ ꕸꖃꔀ', 'vai-Latn': 'Lahabu', 'vai-Vaii': 'ꕞꕌꖝ ꕸꖃꔀ', 'vec': 'Arabia Saudìa', 'vi': 'Ả Rập Xê-út', 'vun': 'Saudi', 'wae': 'Saudi Arabie', 'wo': 'Arabi Sawudi', 'xh': 'ESaudi Arabia', 'xnr': 'सऊदी अरब', 'xog': 'Sawudarebya', 'yav': 'alapísawutíit', 'yo': 'Saudi Arabia', 'yrl': 'Arawia Saudita', 'yue': '沙特阿拉伯', 'yue-Hans': '沙乌地阿拉伯', 'yue-Hant': '沙特阿拉伯', 'zgh': 'ⵙⵙⴰⵄⵓⴷⵉⵢⴰ', 'zh': '沙特阿拉伯', 'zh-Hans': '沙特阿拉伯', 'zh-Hant': '沙烏地阿拉伯', 'zu': 'i-Saudi Arabia'},
    'SB': {'af': 'Salomonseilande', 'agq': 'Chwɨlà fɨ Solomwɨ̀n', 'ak': 'Solomɔn Aeland', 'am': 'ሰለሞን ደሴቶች', 'ar': 'جزر سليمان', 'as': 'চোলোমোন দ্বীপপুঞ্জ', 'asa': 'Vithiwa vya Tholomon', 'ast': 'Islles Salomón', 'az': 'Solomon adaları', 'az-Cyrl': 'Соломон адалары', 'az-Latn': 'Solomon adaları', 'bal-Latn': 'Solomán Islánd', 'bas': 'Bìòn bi Salōmò', 'be': 'Саламонавы астравы', 'be-tarask': 'Саламонавы астравы', 'bez': 'Ifisima fya Solomon', 'bg': 'Соломонови острови', 'blo': 'Salomɔɔn kaBʊtukǝltǝna', 'bm': 'Salomo Gun', 'bn': 'সলোমন দ্বীপপুঞ্জ', 'br': 'Inizi Salomon', 'brx': 'सल’मन द्वीप', 'bs': 'Solomonska Ostrva', 'bs-Cyrl': 'Соломонска Острва', 'bs-Latn': 'Solomonska Ostrva', 'ca': 'Illes Salomó', 'ca-ES-valencia': 'Illes Salomó', 'ccp': '𑄥𑄧𑄣𑄮𑄟𑄚𑄴 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Соломонан гӀайренаш', 'ceb': 'Solomon Islands', 'cgg': 'Ebizinga bya Surimaani', 'chr': 'ᏐᎶᎹᏂ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی سلێمان', 'cs': 'Šalamounovy ostrovy', 'cv': 'Соломон утравӗсем', 'cy': 'Ynysoedd Solomon', 'da': 'Salomonøerne', 'dav': 'Visiwa vya Solomon', 'de': 'Salomonen', 'dje': 'Solomon Gungey', 'dsb': 'Salomony', 'dz': 'སོ་ལོ་མོན་ གླིང་ཚོམ', 'ebu': 'Visiwa vya Solomon', 'ee': 'Solomon ƒudomekpowo nutome', 'el': 'Νήσοι Σολομώντος', 'el-polyton': 'Νῆσοι Σολομῶντος', 'en': 'Solomon Islands', 'en-Dsrt': '𐐝𐐪𐑊𐐲𐑋𐐲𐑌 𐐌𐑊𐐲𐑌𐐼𐑆', 'eo': 'Salomonoj', 'es': 'Islas Salomón', 'et': 'Saalomoni Saared', 'eu': 'Salomon Uharteak', 'ewo': 'Minlán Mí Solomɔ́n', 'fa': 'جزایر سلیمان', 'ff': 'Duuɗe Solomon', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤅𞤵𞤤𞤢𞤴𞤥𞤢𞥄𞤲', 'ff-Latn': 'Duuɗe Solomon', 'fi': 'Salomonsaaret', 'fil': 'Solomon Islands', 'fo': 'Salomonoyggjar', 'fr': 'Îles Salomon', 'fur': 'Isulis Salomon', 'fy': 'Salomonseilannen', 'ga': 'Oileáin Sholaimh', 'gd': 'Eileanan Sholaimh', 'gl': 'Illas Salomón', 'gsw': 'Salomoone', 'gu': 'સોલોમન આઇલેન્ડ્સ', 'guz': 'Visiwa vya Solomon', 'ha': 'Tsibiran Salaman', 'he': 'איי שלמה', 'hi': 'सोलोमन द्वीपसमूह', 'hi-Latn': 'Solomon Islands', 'hr': 'Salomonovi Otoci', 'hsb': 'Salomony', 'ht': 'Îles Salomon', 'hu': 'Salamon-szigetek', 'hy': 'Սողոմոնյան կղզիներ', 'ia': 'Insulas Solomon', 'id': 'Kepulauan Solomon', 'ig': 'Agwaetiti Solomon', 'is': 'Salómonseyjar', 'it': 'Isole Salomone', 'ja': 'ソロモン諸島', 'jmc': 'Visiwa vya Solomon', 'jv': 'Kapuloan Suleman', 'ka': 'სოლომონის კუნძულები', 'kaa': 'Соломон атаўлары', 'kaa-Cyrl': 'Соломон атаўлары', 'kab': 'Tigzirin n Sulumun', 'kam': 'Isiwa sya Solomon', 'kde': 'Chisiwa cha Solomon', 'kea': 'Ilhas Salumãu', 'kgp': 'Saromỹv Goj-vẽso', 'khq': 'Solomon Gungey', 'ki': 'Visiwa vya Solomon', 'kk': 'Соломон аралдары', 'kk-Cyrl': 'Соломон аралдары', 'kln': 'Ikwembeiyotab Solomon', 'km': 'កោះ\u200bសូឡូម៉ុង', 'kn': 'ಸಾಲೊಮನ್ ದ್ವೀಪಗಳು', 'ko': '솔로몬 제도', 'kok': 'सोलोमन आइलँड्स', 'kok-Deva': 'सोलोमन आइलँड्स', 'ks': 'سولامان جٔزیٖرٕ', 'ks-Arab': 'سولامان جٔزیٖرٕ', 'ksb': 'Visiwa vya Solomon', 'ksf': 'zǝ salomɔ́n', 'ksh': 'de Solomone-Enselle', 'ku': 'Giravên Solomonê', 'kxv': 'soloman dīp', 'kxv-Deva': 'सोलोमन द्वीप', 'kxv-Latn': 'soloman dīp', 'kxv-Orya': 'ସଲମନ ଦିପ', 'kxv-Telu': 'సోలమన్ దీవులు', 'ky': 'Соломон аралдары', 'lag': 'Visíiwa vya Solomóoni', 'lb': 'Salomonen', 'lg': 'Bizanga by’eSolomooni', 'lld': 'Isoles Salomon', 'ln': 'Bisanga Solomɔ', 'lo': 'ຫມູ່ເກາະໂຊໂລມອນ', 'lt': 'Saliamono Salos', 'lu': 'Lutanda lua Solomu', 'luo': 'Solomon Islands', 'luy': 'Visiwa vya Solomon', 'lv': 'Zālamana salas', 'mai': 'सोलोमन द्वीपसमूह', 'mas': 'Visiwa vya Solomon', 'mer': 'Airandi Cia Solomon', 'mfe': 'Zil Salomon', 'mg': 'Nosy Salomona', 'mi': 'Ngā Motu Horomona', 'mk': 'Соломонски Острови', 'ml': 'സോളമൻ ദ്വീപുകൾ', 'mn': 'Соломоны арлууд', 'mr': 'सोलोमन बेटे', 'ms': 'Kepulauan Solomon', 'mt': 'il-Gżejjer Solomon', 'mua': 'Sǝr Salomon ma laŋne', 'my': 'ဆော်လမွန်ကျွန်းစု', 'mzn': 'سلیمون جزیره', 'naq': 'Solomon Islands', 'nb': 'Salomonøyene', 'nd': 'Solomon Islands', 'ne': 'सोलोमन टापुहरू', 'nl': 'Salomonseilanden', 'nmg': 'Maŋ́ má Salomɔn', 'nn': 'Salomonøyane', 'no': 'Salomonøyene', 'nqo': 'ߛߊߟߏߡߐ߲߫ ߕߌ߲', 'nyn': 'Ebizinga bya Surimaani', 'om': 'Odoloota Solomoon', 'or': 'ସୋଲୋମନ୍\u200c ଦ୍ୱୀପପୁଞ୍ଜ', 'pa': 'ਸੋਲੋਮਨ ਟਾਪੂ', 'pa-Guru': 'ਸੋਲੋਮਨ ਟਾਪੂ', 'pcm': 'Sólómọ́n Aílands', 'pis': 'Solomon Aelan', 'pl': 'Wyspy Salomona', 'ps': 'سليمان ټاپوګان', 'pt': 'Ilhas Salomão', 'qu': 'Islas Salomón', 'rm': 'Inslas Salomonas', 'rn': 'Amazinga ya Salumoni', 'ro': 'Insulele Solomon', 'rof': 'Visiwa vya Solomon', 'ru': 'Соломоновы о-ва', 'rwk': 'Visiwa vya Solomon', 'saq': 'Visiwa vya Solomon', 'sbp': 'Figunguli fya Solomoni', 'sc': 'Ìsulas Salomone', 'sd': 'سولومون ٻيٽَ', 'sd-Arab': 'سولومون ٻيٽَ', 'se': 'Salomon-sullot', 'seh': 'Ilhas Salomão', 'ses': 'Solomon Gungey', 'sg': 'Zûâ Salomöon', 'shi': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵙⴰⵍⵓⵎⴰⵏ', 'shi-Latn': 'tigzirin n saluman', 'shi-Tfng': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵙⴰⵍⵓⵎⴰⵏ', 'si': 'සොලමන් දූපත්', 'sk': 'Šalamúnove ostrovy', 'sl': 'Salomonovi otoki', 'smn': 'Salomosuolluuh', 'sn': 'Zvitsuwa zvaSolomon', 'so': 'Jasiiradda Solomon', 'sq': 'Ishujt Solomon', 'sr': 'Соломонска Острва', 'sr-Cyrl': 'Соломонска Острва', 'sr-Latn': 'Solomonska Ostrva', 'sv': 'Salomonöarna', 'sw': 'Visiwa vya Solomon', 'syr': 'ܓܙܪܬܐ ܕܫܠܝܡܘܢ', 'ta': 'சாலமன் தீவுகள்', 'te': 'సోలమన్ దీవులు', 'teo': 'Visiwa vya Solomon', 'tg': 'Ҷазираҳои Соломон', 'th': 'หมู่เกาะโซโลมอน', 'ti': 'ደሴታት ሰሎሞን', 'tk': 'Solomon adalary', 'to': 'ʻOtumotu Solomone', 'tr': 'Solomon Adaları', 'tt': 'Сөләйман утраулары', 'twq': 'Solomon Gungey', 'tzm': 'Tigzirin n Salumun', 'ug': 'سولومون ئاراللىرى', 'uk': 'Соломонові Острови', 'ur': 'سولومن آئلینڈز', 'uz': 'Solomon orollari', 'uz-Cyrl': 'Соломон ороллари', 'uz-Latn': 'Solomon orollari', 'vai': 'ꖬꕞꔤꕮꕊꕯ ꔳꘋꗣ', 'vai-Latn': 'Sulaimaãna Tiŋŋɛ', 'vai-Vaii': 'ꖬꕞꔤꕮꕊꕯ ꔳꘋꗣ', 'vec': 'Ìzole Salomon', 'vi': 'Quần đảo Solomon', 'vun': 'Visiwa vya Solomon', 'wae': 'Salomone', 'wo': 'Duni Salmoon', 'xh': 'ESolomon Islands', 'xnr': 'सोलोमन द्वीपसमूह', 'xog': 'Ebizanga bya Solomooni', 'yav': 'il salomɔ́ŋ', 'yi': 'סאלאמאן אינזלען', 'yo': 'Etikun Solomoni', 'yrl': 'Kapuãma-ita Sarumũ', 'yue': '索羅門群島', 'yue-Hans': '索罗门群岛', 'yue-Hant': '索羅門群島', 'zgh': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵙⴰⵍⵓⵎⴰⵏ', 'zh': '所罗门群岛', 'zh-Hans': '所罗门群岛', 'zh-Hant': '索羅門群島', 'zu': 'i-Solomon Islands'},
    'SC': {'af': 'Seychelle', 'agq': 'Sɛchɛ̀lɛ̀s', 'ak': 'Seyhyɛl', 'am': 'ሲሼልስ', 'ar': 'سيشل', 'as': 'ছিচিলিছ', 'asa': 'Shelisheli', 'ast': 'Les Seixeles', 'az': 'Seyşel adaları', 'az-Cyrl': 'Сејшел адалары', 'az-Latn': 'Seyşel adaları', 'bal-Latn': 'Sécheles', 'bas': 'Sèsɛ̂l', 'be': 'Сейшэльскія астравы', 'be-tarask': 'Сейшэльскія астравы', 'bez': 'Hushelisheli', 'bg': 'Сейшели', 'bgn': 'سیشیل', 'blo': 'Seshɛl', 'bm': 'Sesɛli', 'bn': 'সিসিলি', 'br': 'Sechelez', 'brx': 'सैशेल्लेस', 'bs': 'Sejšeli', 'bs-Cyrl': 'Сејшели', 'bs-Latn': 'Sejšeli', 'ca': 'Seychelles', 'ca-ES-valencia': 'Seychelles', 'ccp': '𑄥𑄨𑄥𑄨𑄣𑄨', 'ce': 'Сейшелан гӀайренаш', 'ceb': 'Seychelles', 'cgg': 'Shesheresi', 'chr': 'ᏏᎡᏥᎵᏍ', 'ckb': 'سیشێل', 'cs': 'Seychely', 'cv': 'Сейшел утравӗсем', 'cy': 'Seychelles', 'da': 'Seychellerne', 'dav': 'Shelisheli', 'de': 'Seychellen', 'dje': 'Seešel', 'dsb': 'Seychelle', 'dz': 'སེ་ཤཱལས', 'ebu': 'Shelisheli', 'ee': 'Seshɛls nutome', 'el': 'Σεϋχέλλες', 'el-polyton': 'Σεϋχέλλες', 'en': 'Seychelles', 'en-Dsrt': '𐐝𐐩𐑇𐐯𐑊𐑆', 'eo': 'Sejŝeloj', 'es': 'Seychelles', 'et': 'Seišellid', 'eu': 'Seychelleak', 'ewo': 'Sɛsɛ́l', 'fa': 'سیشل', 'ff': 'Seysel', 'ff-Adlm': '𞤅𞤫𞤴𞤭𞤧𞤫𞤤', 'ff-Latn': 'Seysel', 'fi': 'Seychellit', 'fil': 'Seychelles', 'fo': 'Seyskelloyggjar', 'fr': 'Seychelles', 'fur': 'Seychelles', 'fy': 'Seychellen', 'ga': 'na Séiséil', 'gd': 'Na h-Eileanan Sheiseall', 'gl': 'Seychelles', 'gsw': 'Seischälle', 'gu': 'સેશેલ્સ', 'guz': 'Shelisheli', 'ha': 'Seychelles', 'he': 'איי סיישל', 'hi': 'सेशेल्स', 'hi-Latn': 'Seychelles', 'hr': 'Sejšeli', 'hsb': 'Seychelle', 'ht': 'Seychelles', 'hu': 'Seychelle-szigetek', 'hy': 'Սեյշելներ', 'ia': 'Seychelles', 'id': 'Seychelles', 'ig': 'Seychelles', 'is': 'Seychelles-eyjar', 'it': 'Seychelles', 'ja': 'セーシェル', 'jgo': 'Pɛsɛ́shɛl', 'jmc': 'Shelisheli', 'jv': 'Sésèl', 'ka': 'სეიშელის კუნძულები', 'kaa': 'Сейшель атаўлары', 'kaa-Cyrl': 'Сейшель атаўлары', 'kab': 'Seycel', 'kam': 'Shelisheli', 'kde': 'Shelisheli', 'kea': 'Seixelis', 'kgp': 'Sejserenh', 'khq': 'Seešel', 'ki': 'Shelisheli', 'kk': 'Сейшель аралдары', 'kk-Cyrl': 'Сейшель аралдары', 'kln': 'Emetab Seychelles', 'km': 'សីស្ហែល', 'kn': 'ಸೀಶೆಲ್ಲೆಸ್', 'ko': '세이셸', 'kok': 'सेशेल्स', 'kok-Deva': 'सेशेल्स', 'ks': 'سیشَلِس', 'ks-Arab': 'سیشَلِس', 'ksb': 'Shelisheli', 'ksf': 'sɛcɛl', 'ksh': 'de Seischälle', 'ku': 'Seyşel', 'kxv': 'siselles', 'kxv-Deva': 'सेसेल्स', 'kxv-Latn': 'siselles', 'kxv-Orya': 'ସିସେଲ୍ସ', 'kxv-Telu': 'సీషెల్స్', 'ky': 'Сейшел аралдары', 'lag': 'Shelishéeli', 'lb': 'Seychellen', 'lg': 'Sesere', 'lld': 'Seychelles', 'ln': 'Sɛshɛlɛ', 'lo': 'ເຊເຊວເລສ', 'lt': 'Seišeliai', 'lu': 'Seshele', 'luo': 'Seychelles', 'luy': 'Shelisheli', 'lv': 'Seišelu salas', 'mai': 'सेशेल्स', 'mas': 'Shelisheli', 'mer': 'Shelisheli', 'mfe': 'Sesel', 'mg': 'Seyshela', 'mgh': 'Ushelisheli', 'mi': 'Heikere', 'mk': 'Сејшели', 'ml': 'സീഷെൽസ്', 'mn': 'Сейшелийн арлууд', 'mr': 'सेशेल्स', 'ms': 'Seychelles', 'mt': 'is-Seychelles', 'mua': 'Saichel', 'my': 'ဆေးရှဲ', 'mzn': 'سیشل', 'naq': 'Seychelles', 'nb': 'Seychellene', 'nd': 'Seychelles', 'ne': 'सेचेलेस', 'nl': 'Seychellen', 'nmg': 'Seychɛlle', 'nn': 'Seychellane', 'no': 'Seychellene', 'nqo': 'ߛߋߦߌߛߍߟ', 'nyn': 'Shesheresi', 'om': 'Siisheels', 'or': 'ସେଚେଲସ୍', 'pa': 'ਸੇਸ਼ਲਸ', 'pa-Guru': 'ਸੇਸ਼ਲਸ', 'pcm': 'Sẹ́chẹls', 'pl': 'Seszele', 'ps': 'سیچیلیس', 'pt': 'Seicheles', 'qu': 'Seychelles', 'rm': 'Seychellas', 'rn': 'Amazinga ya Seyisheli', 'ro': 'Seychelles', 'rof': 'Shelisheli', 'ru': 'Сейшельские о-ва', 'rwk': 'Shelisheli', 'saq': 'Shelisheli', 'sbp': 'Shelisheli', 'sc': 'Seychelles', 'sd': 'شي شلز', 'sd-Arab': 'شي شلز', 'se': 'Seychellsullot', 'seh': 'Seychelles', 'ses': 'Seešel', 'sg': 'Sëyshêle', 'shi': 'ⵙⵙⵉⵛⵉⵍ', 'shi-Latn': 'ssicil', 'shi-Tfng': 'ⵙⵙⵉⵛⵉⵍ', 'si': 'සීශෙල්ස්', 'sk': 'Seychely', 'sl': 'Sejšeli', 'smn': 'Seychelleh', 'sn': 'Seychelles', 'so': 'Sishelis', 'sq': 'Sejshelle', 'sr': 'Сејшели', 'sr-Cyrl': 'Сејшели', 'sr-Latn': 'Sejšeli', 'sv': 'Seychellerna', 'sw': 'Ushelisheli', 'syr': 'ܣܐܝܫܝܠ', 'ta': 'சீஷெல்ஸ்', 'te': 'సీషెల్స్', 'teo': 'Shelisheli', 'tg': 'Сейшел', 'th': 'เซเชลส์', 'ti': 'ሲሸልስ', 'tk': 'Seýşel adalary', 'to': 'ʻOtumotu Seiseli', 'tr': 'Seyşeller', 'tt': 'Сейшел утраулары', 'twq': 'Seešel', 'tzm': 'Ssicil', 'ug': 'سېيشېل', 'uk': 'Сейшельські Острови', 'ur': 'سشلیز', 'uz': 'Seyshel orollari', 'uz-Cyrl': 'Сейшел ороллари', 'uz-Latn': 'Seyshel orollari', 'vai': 'ꔖꗼꔷ', 'vai-Latn': 'Seshɛɛ', 'vai-Vaii': 'ꔖꗼꔷ', 'vec': 'Ìzole Seisel', 'vi': 'Seychelles', 'vun': 'Shelisheli', 'wae': 'Sečelle', 'wo': 'Seysel', 'xh': 'ESeychelles', 'xnr': 'सेशेल्स', 'xog': 'Sesere', 'yav': 'sesɛ́ɛl', 'yi': 'סיישעל', 'yo': 'Ṣeṣẹlẹsi', 'yrl': 'Seixeri', 'yue': '塞席爾', 'yue-Hans': '塞席尔', 'yue-Hant': '塞席爾', 'zgh': 'ⵙⵙⵉⵛⵉⵍ', 'zh': '塞舌尔', 'zh-Hans': '塞舌尔', 'zh-Hant': '塞席爾', 'zu': 'i-Seychelles'},
    'SD': {'af': 'Soedan', 'agq': 'Sùdân', 'ak': 'Sudan', 'am': 'ሱዳን', 'ar': 'السودان', 'as': 'চুডান', 'asa': 'Thudani', 'ast': 'Sudán', 'az': 'Sudan', 'az-Cyrl': 'Судан', 'az-Latn': 'Sudan', 'bal-Latn': 'Sudán', 'bas': 'Sùdâŋ', 'be': 'Судан', 'be-tarask': 'Судан', 'bez': 'Husudani', 'bg': 'Судан', 'bgn': 'سوڈان', 'blo': 'Suɖaŋ', 'bm': 'Sudaŋ', 'bn': 'সুদান', 'br': 'Soudan', 'brx': 'सूदान', 'bs': 'Sudan', 'bs-Cyrl': 'Судан', 'bs-Latn': 'Sudan', 'ca': 'Sudan', 'ca-ES-valencia': 'Sudan', 'ccp': '𑄥𑄪𑄘𑄚𑄴', 'ce': 'Судан', 'ceb': 'Sudan', 'cgg': 'Sudani', 'chr': 'ᏑᏕᏂ', 'ckb': 'سوودان', 'cs': 'Súdán', 'cv': 'Судан', 'cy': 'Swdan', 'da': 'Sudan', 'dav': 'Sudani', 'de': 'Sudan', 'dje': 'Suudaŋ', 'dsb': 'Sudan', 'dyo': 'Sudan', 'dz': 'སུ་ཌཱན', 'ebu': 'Sudani', 'ee': 'Sudan nutome', 'el': 'Σουδάν', 'el-polyton': 'Σουδάν', 'en': 'Sudan', 'en-Dsrt': '𐐝𐐭𐐼𐐰𐑌', 'eo': 'Sudano', 'es': 'Sudán', 'et': 'Sudaan', 'eu': 'Sudan', 'ewo': 'Sudáŋ', 'fa': 'سودان', 'ff': 'Sudaan', 'ff-Adlm': '𞤅𞤵𞤣𞤢𞥄𞤲', 'ff-Latn': 'Sudaan', 'fi': 'Sudan', 'fil': 'Sudan', 'fo': 'Sudan', 'fr': 'Soudan', 'fur': 'Sudan', 'fy': 'Soedan', 'ga': 'An tSúdáin', 'gd': 'Sudàn', 'gl': 'Sudán', 'gsw': 'Sudan', 'gu': 'સુદાન', 'guz': 'Sudani', 'ha': 'Sudan', 'he': 'סודן', 'hi': 'सूडान', 'hi-Latn': 'Sudan', 'hr': 'Sudan', 'hsb': 'Sudan', 'ht': 'Soudan', 'hu': 'Szudán', 'hy': 'Սուդան', 'ia': 'Sudan', 'id': 'Sudan', 'ig': 'Sudan', 'is': 'Súdan', 'it': 'Sudan', 'ja': 'スーダン', 'jgo': 'Sundân', 'jmc': 'Sudani', 'jv': 'Sudan', 'ka': 'სუდანი', 'kaa': 'Судан', 'kaa-Cyrl': 'Судан', 'kab': 'Sudan', 'kam': 'Sudani', 'kde': 'Sudani', 'kea': 'Sudon', 'kgp': 'Suná', 'khq': 'Suudaŋ', 'ki': 'Sudani', 'kk': 'Судан', 'kk-Cyrl': 'Судан', 'kln': 'Emetab Sudan', 'km': 'ស៊ូដង់', 'kn': 'ಸುಡಾನ್', 'ko': '수단', 'kok': 'सूडान', 'kok-Deva': 'सूडान', 'ks': 'سوٗڈان', 'ks-Arab': 'سوٗڈان', 'ksb': 'Sudani', 'ksf': 'sudan', 'ksh': 'Noodsudahn', 'ku': 'Sûdan', 'kxv': 'suḍan', 'kxv-Deva': 'सूडान', 'kxv-Latn': 'suḍan', 'kxv-Orya': 'ସୁଡାନ', 'kxv-Telu': 'సూడాన్', 'ky': 'Судан', 'lag': 'Sudáani', 'lb': 'Sudan', 'lg': 'Sudaani', 'lld': 'Sudan', 'ln': 'Sudá', 'lo': 'ຊູດານ', 'lt': 'Sudanas', 'lu': 'Suda', 'luo': 'Sudan', 'luy': 'Sudani', 'lv': 'Sudāna', 'mai': 'सूडान', 'mas': 'Sudani', 'mer': 'Sudani', 'mfe': 'Soudan', 'mg': 'Sodan', 'mgh': 'Usudani', 'mi': 'Hūtāne', 'mk': 'Судан', 'ml': 'സുഡാൻ', 'mn': 'Судан', 'mr': 'सुदान', 'ms': 'Sudan', 'mt': 'is-Sudan', 'mua': 'Sudaŋ', 'my': 'ဆူဒန်', 'mzn': 'سودان', 'naq': 'Sudanni', 'nb': 'Sudan', 'nd': 'Sudani', 'ne': 'सुडान', 'nl': 'Soedan', 'nmg': 'Sudaŋ', 'nn': 'Sudan', 'no': 'Sudan', 'nqo': 'ߛߎߘߊ߲߫', 'nus': 'Sudan', 'nyn': 'Sudani', 'om': 'Sudaan', 'or': 'ସୁଦାନ', 'pa': 'ਸੂਡਾਨ', 'pa-Guru': 'ਸੂਡਾਨ', 'pcm': 'Sudán', 'pl': 'Sudan', 'ps': 'سوډان', 'pt': 'Sudão', 'qu': 'Sudán', 'rm': 'Sudan', 'rn': 'Sudani', 'ro': 'Sudan', 'rof': 'Sudani', 'ru': 'Судан', 'rwk': 'Sudani', 'sah': 'Судаан', 'saq': 'Sudani', 'sbp': 'Sudani', 'sc': 'Sudàn', 'sd': 'سوڊان', 'sd-Arab': 'سوڊان', 'se': 'Davvisudan', 'seh': 'Sudão', 'ses': 'Suudaŋ', 'sg': 'Sudäan', 'shi': 'ⵙⵙⵓⴷⴰⵏ', 'shi-Latn': 'ssudan', 'shi-Tfng': 'ⵙⵙⵓⴷⴰⵏ', 'si': 'සූඩානය', 'sk': 'Sudán', 'sl': 'Sudan', 'smn': 'Sudan', 'sn': 'Sudan', 'so': 'Suudaan', 'sq': 'Sudan', 'sr': 'Судан', 'sr-Cyrl': 'Судан', 'sr-Latn': 'Sudan', 'sv': 'Sudan', 'sw': 'Sudan', 'syr': 'ܣܘܕܐܢ', 'ta': 'சூடான்', 'te': 'సూడాన్', 'teo': 'Sudani', 'tg': 'Судон', 'th': 'ซูดาน', 'ti': 'ሱዳን', 'tk': 'Sudan', 'to': 'Sūteni', 'tr': 'Sudan', 'tt': 'Судан', 'twq': 'Suudaŋ', 'tzm': 'Ssudan', 'ug': 'سۇدان', 'uk': 'Судан', 'ur': 'سوڈان', 'uz': 'Sudan', 'uz-Cyrl': 'Судан', 'uz-Latn': 'Sudan', 'vai': 'ꖬꗵꘋ', 'vai-Latn': 'Sudɛŋ', 'vai-Vaii': 'ꖬꗵꘋ', 'vec': 'Sudan', 'vi': 'Sudan', 'vun': 'Sudani', 'wae': 'Sudan', 'wo': 'Sudaŋ', 'xh': 'ESudan', 'xnr': 'सूडान', 'xog': 'Sudaani', 'yav': 'sutáaŋ', 'yi': 'סודאַן', 'yo': 'Sudani', 'yrl': 'Ausudã', 'yue': '蘇丹', 'yue-Hans': '苏丹', 'yue-Hant': '蘇丹', 'zgh': 'ⵙⵙⵓⴷⴰⵏ', 'zh': '苏丹', 'zh-Hans': '苏丹', 'zh-Hant': '蘇丹', 'zu': 'i-Sudan'},
    'SE': {'af': 'Swede', 'agq': 'Suedɨ̀n', 'ak': 'Sweden', 'am': 'ስዊድን', 'ar': 'السويد', 'as': 'চুইডেন', 'asa': 'Uthwidi', 'ast': 'Suecia', 'az': 'İsveç', 'az-Cyrl': 'Исвеч', 'az-Latn': 'İsveç', 'bal-Latn': 'Swidhan', 'bas': 'Swedɛ̀n', 'be': 'Швецыя', 'be-tarask': 'Швецыя', 'bez': 'Huuswidi', 'bg': 'Швеция', 'blo': 'Sʊwɛɖ', 'bm': 'Suwɛdi', 'bn': 'সুইডেন', 'br': 'Sveden', 'brx': 'स्वीडेन', 'bs': 'Švedska', 'bs-Cyrl': 'Шведска', 'bs-Latn': 'Švedska', 'ca': 'Suècia', 'ca-ES-valencia': 'Suècia', 'ccp': '𑄥𑄭𑄪𑄓𑄬𑄚𑄴', 'ce': 'Швеци', 'ceb': 'Sweden', 'cgg': 'Swideni', 'chr': 'ᏍᏫᏕᏂ', 'ckb': 'سوید', 'cs': 'Švédsko', 'cv': 'Швеци', 'cy': 'Sweden', 'da': 'Sverige', 'dav': 'Uswidi', 'de': 'Schweden', 'dje': 'Sweede', 'dsb': 'Šwedska', 'dz': 'སུའི་ཌེན', 'ebu': 'Uswidi', 'ee': 'Sweden nutome', 'el': 'Σουηδία', 'el-polyton': 'Σουηδία', 'en': 'Sweden', 'en-Dsrt': '𐐝𐐶𐐨𐐼𐑌', 'eo': 'Svedujo', 'es': 'Suecia', 'et': 'Rootsi', 'eu': 'Suedia', 'ewo': 'Suwɛ́d', 'fa': 'سوئد', 'ff': 'Suweed', 'ff-Adlm': '𞤅𞤵𞤱𞤫𞤣𞤭𞤴𞤢𞥄', 'ff-Latn': 'Suweed', 'fi': 'Ruotsi', 'fil': 'Sweden', 'fo': 'Svøríki', 'fr': 'Suède', 'fur': 'Svezie', 'fy': 'Zweden', 'ga': 'an tSualainn', 'gd': 'An t-Suain', 'gl': 'Suecia', 'gsw': 'Schweede', 'gu': 'સ્વીડન', 'guz': 'Uswidi', 'ha': 'Suwedan', 'he': 'שוודיה', 'hi': 'स्वीडन', 'hi-Latn': 'Sweden', 'hr': 'Švedska', 'hsb': 'Šwedska', 'ht': 'Suède', 'hu': 'Svédország', 'hy': 'Շվեդիա', 'ia': 'Svedia', 'id': 'Swedia', 'ie': 'Svedia', 'ig': 'Sweden', 'is': 'Svíþjóð', 'it': 'Svezia', 'ja': 'スウェーデン', 'jmc': 'Uswidi', 'jv': 'Swèdhen', 'ka': 'შვედეთი', 'kaa': 'Швеция', 'kaa-Cyrl': 'Швеция', 'kab': 'Swid', 'kam': 'Uswidi', 'kde': 'Uswidi', 'kea': 'Suésia', 'kgp': 'Suésa', 'khq': 'Sweede', 'ki': 'Uswidi', 'kk': 'Швеция', 'kk-Cyrl': 'Швеция', 'kln': 'Emetab Sweden', 'km': 'ស៊ុយអែត', 'kn': 'ಸ್ವೀಡನ್', 'ko': '스웨덴', 'kok': 'स्वीडन', 'kok-Deva': 'स्वीडन', 'ks': 'سویڈن', 'ks-Arab': 'سویڈن', 'ksb': 'Uswidi', 'ksf': 'swɛdǝ', 'ksh': 'Schweede', 'ku': 'Swêd', 'kxv': 'sviḍen', 'kxv-Deva': 'स्वीडन', 'kxv-Latn': 'sviḍen', 'kxv-Orya': 'ସ୍ୱିଡେନ', 'kxv-Telu': 'స్వీడన్', 'ky': 'Швеция', 'lag': 'Uswíidi', 'lb': 'Schweden', 'lg': 'Swideni', 'lld': 'Svezia', 'ln': 'Swédɛ', 'lo': 'ສະວີເດັນ', 'lt': 'Švedija', 'lu': 'Suwedi', 'luo': 'Sweden', 'luy': 'Uswidi', 'lv': 'Zviedrija', 'mai': 'स्वीडन', 'mas': 'Uswidi', 'mer': 'Swideni', 'mfe': 'Laswed', 'mg': 'Soedy', 'mgh': 'Uswidi', 'mi': 'Huitene', 'mk': 'Шведска', 'ml': 'സ്വീഡൻ', 'mn': 'Швед', 'mr': 'स्वीडन', 'ms': 'Sweden', 'mt': 'l-Iżvezja', 'mua': 'Sǝr Sued', 'my': 'ဆွီဒင်', 'mzn': 'سوئد', 'naq': 'Swedeb', 'nb': 'Sverige', 'nd': 'Sweden', 'ne': 'स्विडेन', 'nl': 'Zweden', 'nmg': 'Suɛd', 'nn': 'Sverige', 'no': 'Sverige', 'nqo': 'ߛߎߥߍߘ', 'nyn': 'Swideni', 'om': 'Siwiidin', 'or': 'ସ୍ୱିଡେନ୍\u200c', 'pa': 'ਸਵੀਡਨ', 'pa-Guru': 'ਸਵੀਡਨ', 'pcm': 'Swídẹn', 'pl': 'Szwecja', 'ps': 'سویډن', 'pt': 'Suécia', 'qu': 'Suecia', 'rm': 'Svezia', 'rn': 'Suwedi', 'ro': 'Suedia', 'rof': 'Uswidi', 'ru': 'Швеция', 'rwk': 'Uswidi', 'sah': 'Швеция', 'saq': 'Uswidi', 'sbp': 'Uswidi', 'sc': 'Isvètzia', 'sd': 'سوئيڊن', 'sd-Arab': 'سوئيڊن', 'se': 'Ruoŧŧa', 'seh': 'Suécia', 'ses': 'Sweede', 'sg': 'Suêde', 'shi': 'ⵙⵙⵡⵉⴷ', 'shi-Latn': 'sswid', 'shi-Tfng': 'ⵙⵙⵡⵉⴷ', 'si': 'ස්වීඩනය', 'sk': 'Švédsko', 'sl': 'Švedska', 'smn': 'Ruotâ', 'sn': 'Sweden', 'so': 'Iswidhan', 'sq': 'Suedi', 'sr': 'Шведска', 'sr-Cyrl': 'Шведска', 'sr-Latn': 'Švedska', 'sv': 'Sverige', 'sw': 'Uswidi', 'syr': 'ܣܘܝܕ', 'ta': 'ஸ்வீடன்', 'te': 'స్వీడన్', 'teo': 'Uswidi', 'tg': 'Шветсия', 'th': 'สวีเดน', 'ti': 'ሽወደን', 'tk': 'Şwesiýa', 'to': 'Suēteni', 'tr': 'İsveç', 'tt': 'Швеция', 'twq': 'Sweede', 'tzm': 'Ssewwid', 'ug': 'شىۋېتسىيە', 'uk': 'Швеція', 'ur': 'سویڈن', 'uz': 'Shvetsiya', 'uz-Cyrl': 'Швеция', 'uz-Latn': 'Shvetsiya', 'vai': 'ꖬꔨꗵꘋ', 'vai-Latn': 'Suwidɛŋ', 'vai-Vaii': 'ꖬꔨꗵꘋ', 'vec': 'Zvesia', 'vi': 'Thụy Điển', 'vun': 'Uswidi', 'wae': 'Schwede', 'wo': 'Suwed', 'xh': 'ESweden', 'xnr': 'स्वीडन', 'xog': 'Swideni', 'yav': 'suɛ́t', 'yi': 'שוועדן', 'yo': 'Swidini', 'yrl': 'Suwesiya', 'yue': '瑞典', 'yue-Hans': '瑞典', 'yue-Hant': '瑞典', 'zgh': 'ⵙⵙⵡⵉⴷ', 'zh': '瑞典', 'zh-Hans': '瑞典', 'zh-Hant': '瑞典', 'zu': 'i-Sweden'},
    'SG': {'af': 'Singapoer', 'agq': 'Siŋgàkpôo', 'ak': 'Singapɔ', 'am': 'ሲንጋፖር', 'ar': 'سنغافورة', 'as': 'ছিংগাপুৰ', 'asa': 'Thingapoo', 'ast': 'Singapur', 'az': 'Sinqapur', 'az-Cyrl': 'Сингапур', 'az-Latn': 'Sinqapur', 'bal-Latn': 'Sengápur', 'bas': 'Sìŋgàpûr', 'be': 'Сінгапур', 'be-tarask': 'Сінгапур', 'bez': 'Husingapoo', 'bg': 'Сингапур', 'bgn': 'سینگاپور', 'blo': 'Siŋgapuur', 'bm': 'Sɛngapuri', 'bn': 'সিঙ্গাপুর', 'br': 'Singapour', 'brx': 'सिंंगाप’र', 'bs': 'Singapur', 'bs-Cyrl': 'Сингапур', 'bs-Latn': 'Singapur', 'ca': 'Singapur', 'ca-ES-valencia': 'Singapur', 'ccp': '𑄥𑄨𑄋𑄴𑄉𑄛𑄪𑄢𑄴', 'ce': 'Сингапур', 'ceb': 'Singapore', 'cgg': 'Singapo', 'chr': 'ᏏᏂᎦᏉᎵ', 'ckb': 'سینگاپور', 'cs': 'Singapur', 'cv': 'Сингапур', 'cy': 'Singapore', 'da': 'Singapore', 'dav': 'Singapoo', 'de': 'Singapur', 'dje': 'Singapur', 'dsb': 'Singapur', 'dyo': 'Singapur', 'dz': 'སིང་ག་པོར', 'ebu': 'Singapoo', 'ee': 'Singapɔr nutome', 'el': 'Σιγκαπούρη', 'el-polyton': 'Σιγκαπούρη', 'en': 'Singapore', 'en-Dsrt': '𐐝𐐮𐑍𐐲𐐹𐐱𐑉', 'eo': 'Singapuro', 'es': 'Singapur', 'et': 'Singapur', 'eu': 'Singapur', 'ewo': 'Singapúr', 'fa': 'سنگاپور', 'ff': 'Sinngapuur', 'ff-Adlm': '𞤅𞤭𞤲𞤺𞤢𞤨𞤵𞥅𞤪', 'ff-Latn': 'Sinngapuur', 'fi': 'Singapore', 'fil': 'Singapore', 'fo': 'Singapor', 'fr': 'Singapour', 'fur': 'Singapore', 'fy': 'Singapore', 'ga': 'Singeapór', 'gd': 'Singeapòr', 'gl': 'Singapur', 'gsw': 'Singapuur', 'gu': 'સિંગાપુર', 'guz': 'Singapoo', 'ha': 'Singapur', 'he': 'סינגפור', 'hi': 'सिंगापुर', 'hi-Latn': 'Singapore', 'hr': 'Singapur', 'hsb': 'Singapur', 'ht': 'Singapour', 'hu': 'Szingapúr', 'hy': 'Սինգապուր', 'ia': 'Singapur', 'id': 'Singapura', 'ig': 'Singapore', 'is': 'Singapúr', 'it': 'Singapore', 'ja': 'シンガポール', 'jmc': 'Singapoo', 'jv': 'Singapura', 'ka': 'სინგაპური', 'kaa': 'Сингапур', 'kaa-Cyrl': 'Сингапур', 'kab': 'Singafur', 'kam': 'Singapoo', 'kde': 'Singapoo', 'kea': 'Singapura', 'kgp': 'Sĩgapura', 'khq': 'Singapur', 'ki': 'Singapoo', 'kk': 'Сингапур', 'kk-Cyrl': 'Сингапур', 'kln': 'Emetab Singapore', 'km': 'សិង្ហបុរី', 'kn': 'ಸಿಂಗಪುರ್', 'ko': '싱가포르', 'kok': 'सिंगापूर', 'kok-Deva': 'सिंगापूर', 'ks': 'سِنگاپوٗر', 'ks-Arab': 'سِنگاپوٗر', 'ksb': 'Singapoo', 'ksf': 'siŋapó', 'ksh': 'Singjapuur', 'ku': 'Sîngapûr', 'kxv': 'singapor', 'kxv-Deva': 'सिंगापुर', 'kxv-Latn': 'singapor', 'kxv-Orya': 'ସିଙ୍ଗାପୁର', 'kxv-Telu': 'సింగపూర్', 'ky': 'Сингапур', 'lag': 'Singapoo', 'lb': 'Singapur', 'lg': 'Singapowa', 'lld': 'Singapur', 'ln': 'Singapurɛ', 'lo': 'ສິງກະໂປ', 'lt': 'Singapūras', 'lu': 'Singapure', 'luo': 'Singapore', 'luy': 'Singapoo', 'lv': 'Singapūra', 'mai': 'सिंगापुर', 'mas': 'Singapoo', 'mer': 'Singapoo', 'mfe': 'Singapour', 'mg': 'Singaporo', 'mgh': 'Usingapoo', 'mi': 'Hingapoa', 'mk': 'Сингапур', 'ml': 'സിംഗപ്പൂർ', 'mn': 'Сингапур', 'mr': 'सिंगापूर', 'ms': 'Singapura', 'mt': 'Singapore', 'mua': 'Singapur', 'my': 'စင်္ကာပူ', 'mzn': 'سنگاپور', 'naq': 'Singapore', 'nb': 'Singapore', 'nd': 'Singapore', 'ne': 'सिङ्गापुर', 'nl': 'Singapore', 'nmg': 'Singapur', 'nn': 'Singapore', 'no': 'Singapore', 'nqo': 'ߛߌ߲ߜ߭ߊߔߎߙ', 'nyn': 'Singapo', 'om': 'Singaapoor', 'or': 'ସିଙ୍ଗାପୁର୍', 'pa': 'ਸਿੰਗਾਪੁਰ', 'pa-Guru': 'ਸਿੰਗਾਪੁਰ', 'pcm': 'Singapọ', 'pl': 'Singapur', 'ps': 'سينگاپور', 'pt': 'Singapura', 'qu': 'Singapur', 'rm': 'Singapur', 'rn': 'Singapuru', 'ro': 'Singapore', 'rof': 'Singapoo', 'ru': 'Сингапур', 'rwk': 'Singapoo', 'saq': 'Singapoo', 'sbp': 'Singapoo', 'sc': 'Singapore', 'sd': 'سنگاپور', 'sd-Arab': 'سنگاپور', 'se': 'Singapore', 'seh': 'Cingapura', 'ses': 'Singapur', 'sg': 'Sïngäpûru', 'shi': 'ⵙⵏⵖⴰⴼⵓⵔⴰ', 'shi-Latn': 'snɣafura', 'shi-Tfng': 'ⵙⵏⵖⴰⴼⵓⵔⴰ', 'si': 'සිංගප්පූරුව', 'sk': 'Singapur', 'sl': 'Singapur', 'smn': 'Singapore', 'sn': 'Singapore', 'so': 'Singaboor', 'sq': 'Singapor', 'sr': 'Сингапур', 'sr-Cyrl': 'Сингапур', 'sr-Latn': 'Singapur', 'sv': 'Singapore', 'sw': 'Singapore', 'syr': 'ܣܝܢܓܐܦܘܪ', 'ta': 'சிங்கப்பூர்', 'te': 'సింగపూర్', 'teo': 'Singapoo', 'tg': 'Сингапур', 'th': 'สิงคโปร์', 'ti': 'ሲንጋፖር', 'tk': 'Singapur', 'to': 'Singapoa', 'tr': 'Singapur', 'tt': 'Сингапур', 'twq': 'Singapur', 'tzm': 'Sanɣafura', 'ug': 'سىنگاپور', 'uk': 'Сінгапур', 'ur': 'سنگاپور', 'uz': 'Singapur', 'uz-Cyrl': 'Сингапур', 'uz-Latn': 'Singapur', 'vai': 'ꔻꕬꕶꕱ', 'vai-Latn': 'Síingapoo', 'vai-Vaii': 'ꔻꕬꕶꕱ', 'vec': 'Singapore', 'vi': 'Singapore', 'vun': 'Singapoo', 'wae': 'Singapur', 'wo': 'Singapuur', 'xh': 'ESingapore', 'xnr': 'सिंगापुर', 'xog': 'Singapowa', 'yav': 'singapúul', 'yi': 'סינגאַפּור', 'yo': 'Singapo', 'yrl': 'Sĩgapura', 'yue': '新加坡', 'yue-Hans': '新加坡', 'yue-Hant': '新加坡', 'zgh': 'ⵙⵏⵖⴰⴼⵓⵔⴰ', 'zh': '新加坡', 'zh-Hans': '新加坡', 'zh-Hant': '新加坡', 'zu': 'i-Singapore'},
    'SH': {'af': 'Sint Helena', 'agq': 'Sɛ̀n Èlenà', 'ak': 'Saint Helena', 'am': 'ሴንት ሄለና', 'ar': 'سانت هيلينا', 'as': 'ছেইণ্ট হেলেনা', 'asa': 'Thantahelena', 'ast': 'Santa Helena', 'az': 'Müqəddəs Yelena', 'az-Cyrl': 'Мүгәддәс Јелена', 'az-Latn': 'Müqəddəs Yelena', 'bal-Latn': 'St. Heléná', 'bas': 'Nûmpubi Ɛlēnà', 'be': 'Востраў Святой Алены', 'be-tarask': 'Востраў Святой Алены', 'bez': 'Husantahelena', 'bg': 'Света Елена', 'blo': 'Sɛŋ Elenaa (kaAtukǝltǝna)', 'bm': 'Ɛlɛni Senu', 'bn': 'সেন্ট হেলেনা', 'br': 'Saint-Helena', 'brx': 'सैन्ट हेलेना', 'bs': 'Sveta Helena', 'bs-Cyrl': 'Света Хелена', 'bs-Latn': 'Sveta Helena', 'ca': 'Santa Helena', 'ca-ES-valencia': 'Santa Helena', 'ccp': '𑄥𑄬𑄚𑄴𑄑𑄴 𑄦𑄬𑄣𑄬𑄚', 'ce': 'Сийлахьчу Еленин гӀайре', 'ceb': 'St. Helena', 'cgg': 'Senti Herena', 'chr': 'ᎤᏓᏅᏘ ᎮᎵᎾ', 'ckb': 'سەن هێلێنا', 'cs': 'Svatá Helena', 'cv': 'Сӑваплӑ Елена утравӗ', 'cy': 'Saint Helena', 'da': 'St. Helena', 'dav': 'Santahelena', 'de': 'St. Helena', 'dje': 'Seŋ Helena', 'dsb': 'St. Helena', 'dz': 'སེནཊ་ ཧེ་ལི་ན', 'ebu': 'Santahelena', 'ee': 'Saint Helena nutome', 'el': 'Αγία Ελένη', 'el-polyton': 'Ἁγία Ἑλένη', 'en': 'St. Helena', 'en-Dsrt': '𐐝𐐩𐑌𐐻 𐐐𐐯𐑊𐐲𐑌𐐲', 'eo': 'Sankta Heleno', 'es': 'Santa Elena', 'et': 'Saint Helena', 'eu': 'Santa Helena', 'ewo': 'Ǹfúfúb-Ɛlɛ́na', 'fa': 'سنت هلن', 'ff': 'Sent Helen', 'ff-Adlm': '𞤅𞤫𞤲-𞤖𞤫𞤤𞤫𞤲𞤢𞥄', 'ff-Latn': 'Sent Helen', 'fi': 'Saint Helena', 'fil': 'St. Helena', 'fo': 'St. Helena', 'fr': 'Sainte-Hélène', 'fur': 'Sante Eline', 'fy': 'Sint-Helena', 'ga': 'San Héilin', 'gd': 'Eilean Naomh Eilidh', 'gl': 'Santa Helena', 'gsw': 'St. Helena', 'gu': 'સેંટ હેલેના', 'guz': 'Santahelena', 'ha': 'San Helena', 'he': 'סנט הלנה', 'hi': 'सेंट हेलेना', 'hi-Latn': 'St. Helena', 'hr': 'Sveta Helena', 'hsb': 'St. Helena', 'ht': 'Sainte-Hélène', 'hu': 'Szent Ilona', 'hy': 'Սուրբ Հեղինեի կղզի', 'ia': 'Sancte Helena', 'id': 'Saint Helena', 'ig': 'St. Helena', 'is': 'Sankti Helena', 'it': 'Sant’Elena', 'ja': 'セントヘレナ', 'jmc': 'Santahelena', 'jv': 'Saint Héléna', 'ka': 'წმინდა ელენეს კუნძული', 'kab': 'Sant Ilina', 'kam': 'Santahelena', 'kde': 'Santahelena', 'kea': 'Santa Ilena', 'kgp': 'Sỹnta Erenỹ', 'khq': 'Seŋ Helena', 'ki': 'Santahelena', 'kk': 'Әулие Елена аралы', 'kk-Cyrl': 'Әулие Елена аралы', 'kln': 'Emetab Helena Ne tilil', 'km': 'សង់\u200bហេឡេណា', 'kn': 'ಸೇಂಟ್ ಹೆಲೆನಾ', 'ko': '세인트헬레나', 'kok': 'सेंट हेलिना', 'kok-Deva': 'सेंट हेलिना', 'ks': 'سینٹ ہؠلِنا', 'ks-Arab': 'سینٹ ہؠلِنا', 'ksb': 'Santahelena', 'ksf': 'sɛntɛ́len', 'ksh': 'de Ensel Zint Hellena', 'ku': 'Saint Helena', 'kxv': 'seint helena', 'kxv-Deva': 'सेंट हेलेना', 'kxv-Latn': 'seint helena', 'kxv-Orya': 'ସେଣ୍ଟ ହେଲେନ', 'kxv-Telu': 'సెయింట్ హెలెనా', 'ky': 'Ыйык Елена', 'lag': 'Mʉtakatíifu Heléena', 'lb': 'St. Helena', 'lg': 'Senti Herena', 'lld': 'St. Helena', 'ln': 'Sántu eleni', 'lo': 'ເຊນ ເຮເລນາ', 'lt': 'Šv. Elenos Sala', 'lu': 'Santu eleni', 'luo': 'Saint Helena', 'luy': 'Santahelena', 'lv': 'Sv.Helēnas sala', 'mai': 'सेंट हेलेना', 'mas': 'Santahelena', 'mer': 'Santahelena', 'mfe': 'Sainte-Hélène', 'mg': 'Sainte-Hélène', 'mgh': 'Usantahelena', 'mi': 'Hato Hērena', 'mk': 'Света Елена', 'ml': 'സെന്റ് ഹെലീന', 'mn': 'Сент Хелена', 'mr': 'सेंट हेलेना', 'ms': 'Saint Helena', 'mt': 'Saint Helena', 'mua': 'Sǝr Helena', 'my': 'စိန့်ဟယ်လယ်နာ', 'mzn': 'سنت هلنا', 'naq': 'Saint Helena', 'nb': 'St. Helena', 'nd': 'Saint Helena', 'ne': 'सेन्ट हेलेना', 'nl': 'Sint-Helena', 'nmg': 'Saint Lina', 'nn': 'Saint Helena', 'no': 'St. Helena', 'nqo': 'ߛߍ߲ߕ-ߋߟߍߣ', 'nyn': 'Senti Herena', 'om': 'St. Helenaa', 'or': 'ସେଣ୍ଟ ହେଲେନା', 'pa': 'ਸੇਂਟ ਹੇਲੇਨਾ', 'pa-Guru': 'ਸੇਂਟ ਹੇਲੇਨਾ', 'pcm': 'Sent Hẹlẹ́na', 'pl': 'Wyspa Świętej Heleny', 'ps': 'سینټ هیلینا', 'pt': 'Santa Helena', 'qu': 'Santa Elena', 'rm': 'Sontg’Elena', 'rn': 'Sehelene', 'ro': 'Sfânta Elena', 'rof': 'Santahelena', 'ru': 'о-в Св. Елены', 'rwk': 'Santahelena', 'saq': 'Santahelena', 'sbp': 'Santahelena', 'sc': 'Santa Elene', 'sd': 'سينٽ ھيلينا', 'sd-Arab': 'سينٽ ھيلينا', 'se': 'Saint Helena', 'seh': 'Santa Helena', 'ses': 'Seŋ Helena', 'sg': 'Sênt-Helêna', 'shi': 'ⵙⴰⵏⵜⵉⵍⵉⵏ', 'shi-Latn': 'santilin', 'shi-Tfng': 'ⵙⴰⵏⵜⵉⵍⵉⵏ', 'si': 'ශාන්ත හෙලේනා', 'sk': 'Svätá Helena', 'sl': 'Sveta Helena', 'smn': 'Saint Helena', 'sn': 'Saint Helena', 'so': 'Saynt Helena', 'sq': 'Shën-Elenë', 'sr': 'Света Јелена', 'sr-Cyrl': 'Света Јелена', 'sr-Latn': 'Sveta Jelena', 'sv': 'S:t Helena', 'sw': 'St. Helena', 'syr': 'ܡܪܬܝ ܗܝܠܝܢܐ', 'ta': 'செயின்ட் ஹெலெனா', 'te': 'సెయింట్ హెలెనా', 'teo': 'Santahelena', 'tg': 'Сент Елена', 'th': 'เซนต์เฮเลนา', 'ti': 'ቅድስቲ ሄለና', 'tk': 'Keramatly Ýelena adasy', 'to': 'Sā Helena', 'tr': 'Saint Helena', 'tt': 'Изге Елена утравы', 'twq': 'Seŋ Helena', 'tzm': 'Santehilin', 'ug': 'ساينىت ھېلېنا', 'uk': 'Острів Святої Єлени', 'ur': 'سینٹ ہیلینا', 'uz': 'Muqaddas Yelena oroli', 'uz-Cyrl': 'Муқаддас Елена ороли', 'uz-Latn': 'Muqaddas Yelena oroli', 'vai': 'ꔻꘋ ꗥꔷꕯ', 'vai-Latn': 'Siŋ Hɛlina', 'vai-Vaii': 'ꔻꘋ ꗥꔷꕯ', 'vec': 'Ìzola S. Èlena', 'vi': 'St. Helena', 'vun': 'Santahelena', 'wae': 'St. Helena', 'wo': 'Saŋ Eleen', 'xh': 'ESt. Helena', 'xnr': 'सेंट हेलेना', 'xog': 'Senti Herena', 'yav': 'sɛ́ŋtɛ́ elɛ́ɛnɛ', 'yi': 'סט העלענע', 'yo': 'Hẹlena', 'yrl': 'Sãta Erena', 'yue': '聖赫勒拿島', 'yue-Hans': '圣赫勒拿岛', 'yue-Hant': '聖赫勒拿島', 'zgh': 'ⵙⴰⵏⵜⵉⵍⵉⵏ', 'zh': '圣赫勒拿', 'zh-Hans': '圣赫勒拿', 'zh-Hant': '聖赫勒拿島', 'zu': 'i-St. Helena'},
    'SI': {'af': 'Slowenië', 'agq': 'Sɨ̀lòvɨnyìa', 'ak': 'Slovinia', 'am': 'ስሎቬኒያ', 'ar': 'سلوفينيا', 'as': 'শ্লোভেনিয়া', 'asa': 'Thlovenia', 'ast': 'Eslovenia', 'az': 'Sloveniya', 'az-Cyrl': 'Словенија', 'az-Latn': 'Sloveniya', 'bal-Latn': 'Slowiniá', 'bas': 'Slòvanìà', 'be': 'Славенія', 'be-tarask': 'Славенія', 'bez': 'Huslovenia', 'bg': 'Словения', 'blo': 'Slofeniya', 'bm': 'Sloveni', 'bn': 'স্লোভানিয়া', 'br': 'Slovenia', 'brx': 'स्ल’भेनिया', 'bs': 'Slovenija', 'bs-Cyrl': 'Словенија', 'bs-Latn': 'Slovenija', 'ca': 'Eslovènia', 'ca-ES-valencia': 'Eslovènia', 'ccp': '𑄥𑄳𑄣𑄮𑄞𑄚𑄨𑄠', 'ce': 'Словени', 'ceb': 'Slovenia', 'cgg': 'Sirovaaniya', 'chr': 'ᏍᎶᏫᏂᎠ', 'ckb': 'سلۆڤێنیا', 'cs': 'Slovinsko', 'cv': 'Словени', 'cy': 'Slofenia', 'da': 'Slovenien', 'dav': 'Slovenia', 'de': 'Slowenien', 'dje': 'Sloveeni', 'dsb': 'Słowjeńska', 'dyo': 'Sloveni', 'dz': 'སུ་ལོ་བི་ནི་ཡ', 'ebu': 'Slovenia', 'ee': 'Slovenia nutome', 'el': 'Σλοβενία', 'el-polyton': 'Σλοβενία', 'en': 'Slovenia', 'en-Dsrt': '𐐝𐑊𐐬𐑂𐐨𐑌𐐨𐐲', 'eo': 'Slovenujo', 'es': 'Eslovenia', 'et': 'Sloveenia', 'eu': 'Eslovenia', 'ewo': 'Səlovénia', 'fa': 'اسلوونی', 'ff': 'Slowenii', 'ff-Adlm': '𞤅𞤵𞤤𞤮𞤾𞤫𞤲𞤭𞤴𞤢𞥄', 'ff-Latn': 'Slowenii', 'fi': 'Slovenia', 'fil': 'Slovenia', 'fo': 'Slovenia', 'fr': 'Slovénie', 'fur': 'Slovenie', 'fy': 'Slovenië', 'ga': 'an tSlóivéin', 'gd': 'An t-Slòbhain', 'gl': 'Eslovenia', 'gsw': 'Sloweenie', 'gu': 'સ્લોવેનિયા', 'guz': 'Slovenia', 'ha': 'Sulobeniya', 'he': 'סלובניה', 'hi': 'स्लोवेनिया', 'hi-Latn': 'Slovenia', 'hr': 'Slovenija', 'hsb': 'Słowjenska', 'ht': 'Slovénie', 'hu': 'Szlovénia', 'hy': 'Սլովենիա', 'ia': 'Slovenia', 'id': 'Slovenia', 'ie': 'Slovenia', 'ig': 'Slovenia', 'is': 'Slóvenía', 'it': 'Slovenia', 'ja': 'スロベニア', 'jmc': 'Slovenia', 'jv': 'Slovénia', 'ka': 'სლოვენია', 'kaa': 'Словения', 'kaa-Cyrl': 'Словения', 'kab': 'Sluvinya', 'kam': 'Slovenia', 'kde': 'Slovenia', 'kea': 'Slovénia', 'kgp': 'Enhrovenĩja', 'khq': 'Sloveeni', 'ki': 'Slovenia', 'kk': 'Словения', 'kk-Cyrl': 'Словения', 'kln': 'Emetab Slovenia', 'km': 'ស្លូវេនី', 'kn': 'ಸ್ಲೋವೇನಿಯಾ', 'ko': '슬로베니아', 'kok': 'स्लोवेनिया', 'kok-Deva': 'स्लोवेनिया', 'ks': 'سَلووینِیا', 'ks-Arab': 'سَلووینِیا', 'ksb': 'Slovenia', 'ksf': 'slovɛní', 'ksh': 'ẞloveenije', 'ku': 'Slovenya', 'kxv': 'slovenia', 'kxv-Deva': 'स्लोवेनिया', 'kxv-Latn': 'slovenia', 'kxv-Orya': 'ସ୍ଲବେନିଆ', 'kxv-Telu': 'స్లోవేనియా', 'ky': 'Словения', 'lag': 'Sulovénia', 'lb': 'Slowenien', 'lg': 'Sirovenya', 'lld': 'Slovenia', 'ln': 'Siloveni', 'lo': 'ສະໂລເວເນຍ', 'lt': 'Slovėnija', 'lu': 'Siloveni', 'luo': 'Slovenia', 'luy': 'Slovenia', 'lv': 'Slovēnija', 'mai': 'स्लोबेनिया', 'mas': 'Slovenia', 'mer': 'Slovenia', 'mfe': 'Sloveni', 'mg': 'Slovenia', 'mgh': 'Uslovenia', 'mi': 'Horowinia', 'mk': 'Словенија', 'ml': 'സ്ലോവേനിയ', 'mn': 'Словени', 'mr': 'स्लोव्हेनिया', 'ms': 'Slovenia', 'mt': 'is-Slovenja', 'mua': 'Sloveniya', 'my': 'ဆလိုဗေးနီးယား', 'mzn': 'اسلوونی', 'naq': 'Slovenia', 'nb': 'Slovenia', 'nd': 'Slovenia', 'ne': 'स्लोभेनिया', 'nl': 'Slovenië', 'nmg': 'Slovenia', 'nn': 'Slovenia', 'no': 'Slovenia', 'nqo': 'ߛߑߟߏߝ߭ߋߣߌ߫', 'nyn': 'Sirovaaniya', 'om': 'Islooveeniyaa', 'or': 'ସ୍ଲୋଭେନିଆ', 'pa': 'ਸਲੋਵੇਨੀਆ', 'pa-Guru': 'ਸਲੋਵੇਨੀਆ', 'pcm': 'Slovẹ́nia', 'pl': 'Słowenia', 'ps': 'سلوانیا', 'pt': 'Eslovênia', 'qu': 'Eslovenia', 'rm': 'Slovenia', 'rn': 'Siloveniya', 'ro': 'Slovenia', 'rof': 'Slovenia', 'ru': 'Словения', 'rwk': 'Slovenia', 'saq': 'Slovenia', 'sbp': 'Siloveniya', 'sc': 'Islovènia', 'sd': 'سلوینیا', 'sd-Arab': 'سلوینیا', 'se': 'Slovenia', 'seh': 'Eslovênia', 'ses': 'Sloveeni', 'sg': 'Solovenïi', 'shi': 'ⵙⵍⵓⴼⵉⵏⵢⴰ', 'shi-Latn': 'slufinya', 'shi-Tfng': 'ⵙⵍⵓⴼⵉⵏⵢⴰ', 'si': 'ස්ලෝවේනියාව', 'sk': 'Slovinsko', 'sl': 'Slovenija', 'smn': 'Slovenia', 'sn': 'Slovenia', 'so': 'Islofeeniya', 'sq': 'Slloveni', 'sr': 'Словенија', 'sr-Cyrl': 'Словенија', 'sr-Latn': 'Slovenija', 'sv': 'Slovenien', 'sw': 'Slovenia', 'syr': 'ܣܠܘܒܢܝܐ', 'ta': 'ஸ்லோவேனியா', 'te': 'స్లోవేనియా', 'teo': 'Slovenia', 'tg': 'Словения', 'th': 'สโลวีเนีย', 'ti': 'ስሎቬንያ', 'tk': 'Sloweniýa', 'to': 'Silōvenia', 'tr': 'Slovenya', 'tt': 'Словения', 'twq': 'Sloveeni', 'tzm': 'Sluvinya', 'ug': 'سىلوۋېنىيە', 'uk': 'Словенія', 'ur': 'سلووینیا', 'uz': 'Sloveniya', 'uz-Cyrl': 'Словения', 'uz-Latn': 'Sloveniya', 'vai': 'ꔻꖃꔍꕇꕰ', 'vai-Latn': 'Suloveninya', 'vai-Vaii': 'ꔻꖃꔍꕇꕰ', 'vec': 'Zlovenia', 'vi': 'Slovenia', 'vun': 'Slovenia', 'wae': 'Slowenie', 'wo': 'Esloweni', 'xh': 'ESlovenia', 'xnr': 'स्लोवेनिया', 'xog': 'Sirovenya', 'yav': 'silofení', 'yi': 'סלאוועניע', 'yo': 'Silofania', 'yrl': 'Esiruwẽniya', 'yue': '斯洛維尼亞', 'yue-Hans': '斯洛维尼亚', 'yue-Hant': '斯洛維尼亞', 'zgh': 'ⵙⵍⵓⴼⵉⵏⵢⴰ', 'zh': '斯洛文尼亚', 'zh-Hans': '斯洛文尼亚', 'zh-Hant': '斯洛維尼亞', 'zu': 'i-Slovenia'},
    'SJ': {'af': 'Spitsbergen en Jan Mayen', 'ak': 'Svalbard ne Jan Mayen', 'am': 'ስቫልባርድ እና ጃን ማየን', 'ar': 'سفالبارد وجان ماين', 'as': 'চাভালবাৰ্ড আৰু জন মেয়ন', 'ast': 'Svalbard ya Islla Jan Mayen', 'az': 'Svalbard və Yan-Mayen', 'az-Cyrl': 'Свалбард вә Јан-Мајен', 'az-Latn': 'Svalbard və Yan-Mayen', 'bal-Latn': 'Swalbard o Jan Mayén', 'be': 'Шпіцберген і Ян-Маен', 'be-tarask': 'Шпіцберген і Ян-Маен', 'bg': 'Свалбард и Ян Майен', 'blo': 'Sǝfalbaaɖ na Yan Mayɛn', 'bn': 'স্বালবার্ড ও জান মেয়েন', 'br': 'Svalbard', 'brx': 'स्भालबार्ड आरो जान मायेन', 'bs': 'Svalbard i Jan Mayen', 'bs-Cyrl': 'Свалбард и Јан Мајен', 'bs-Latn': 'Svalbard i Jan Mayen', 'ca': 'Svalbard i Jan Mayen', 'ca-ES-valencia': 'Svalbard i Jan Mayen', 'ccp': '𑄥𑄣𑄴𑄝𑄢𑄴𑄓𑄴 𑄃𑄮 𑄎𑄚𑄴 𑄟𑄬𑄠𑄬𑄚𑄴', 'ce': 'Шпицберген а, Ян-Майен а', 'ceb': 'Svalbard & Jan Mayen', 'chr': 'ᏍᏩᎵᏆᎵᏗ ᎠᎴ ᏤᏂ ᎹᏰᏂ', 'ckb': 'سڤالبارد و یان مایەن', 'cs': 'Špicberky a Jan Mayen', 'cv': 'Шпицберген тата Ян-Майен', 'cy': 'Svalbard a Jan Mayen', 'da': 'Svalbard og Jan Mayen', 'de': 'Spitzbergen und Jan Mayen', 'dsb': 'Svalbard a Jan Mayen', 'dz': 'སྭཱལ་བྷརྡ་ ཨེནཌ་ ཇཱན་མ་ཡེན', 'ee': 'Svalbard kple Yan Mayen nutome', 'el': 'Σβάλμπαρντ και Γιαν Μαγιέν', 'el-polyton': 'Νῆσοι Σβάλμπαρ καὶ Γιὰν Μαγιέν', 'en': 'Svalbard & Jan Mayen', 'en-Dsrt': '𐐝𐑂𐐪𐑊𐐺𐐪𐑉𐐼 𐐰𐑌𐐼 𐐖𐐰𐑌 𐐣𐐴𐐲𐑌', 'eo': 'Svalbardo kaj Janmajeno', 'es': 'Svalbard y Jan Mayen', 'et': 'Svalbard ja Jan Mayen', 'eu': 'Svalbard eta Jan Mayen uharteak', 'fa': 'سوالبارد و یان ماین', 'ff-Adlm': '𞤅𞤢𞤾𞤢𞤤𞤦𞤢𞤪𞤣 & 𞤔𞤢𞤲 𞤃𞤢𞤴𞤫𞤲', 'fi': 'Huippuvuoret ja Jan Mayen', 'fil': 'Svalbard & Jan Mayen', 'fo': 'Svalbard & Jan Mayen', 'fr': 'Svalbard et Jan Mayen', 'fur': 'Svalbard e Jan Mayen', 'fy': 'Spitsbergen en Jan Mayen', 'ga': 'Svalbard agus Jan Mayen', 'gd': 'Svalbard is Jan Mayen', 'gl': 'Svalbard e Jan Mayen', 'gsw': 'Svalbard und Jaan Määie', 'gu': 'સ્વાલબર્ડ અને જેન મેયન', 'ha': 'Svalbard da Jan Mayen', 'he': 'סבאלברד ויאן מאיין', 'hi': 'स्वालबार्ड और जान मायेन', 'hi-Latn': 'Svalbard & Jan Mayen', 'hr': 'Svalbard i Jan Mayen', 'hsb': 'Svalbard a Jan Mayen', 'ht': 'Svalbard et Jan Mayen', 'hu': 'Svalbard és Jan Mayen', 'hy': 'Սվալբարդ և Յան Մայեն', 'ia': 'Svalbard e Jan Mayen', 'id': 'Kepulauan Svalbard dan Jan Mayen', 'ig': 'Svalbard & Jan Mayen', 'is': 'Svalbarði og Jan Mayen', 'it': 'Svalbard e Jan Mayen', 'ja': 'スバールバル諸島・ヤンマイエン島', 'jv': 'Svalbard lan Jan Mayen', 'ka': 'შპიცბერგენი და იან-მაიენი', 'kaa': 'Шпицберген ҳәм Ян-Майен', 'kaa-Cyrl': 'Шпицберген ҳәм Ян-Майен', 'kea': 'Svalbard i Jan Maien', 'kgp': 'Inhvarmarni kar Jan Mỹjẽn', 'kk': 'Шпицберген және Ян-Майен', 'kk-Cyrl': 'Шпицберген және Ян-Майен', 'km': 'ស្វាលបាដ និង ហ្សង់ម៉ាយេន', 'kn': 'ಸ್ವಾಲ್ಬಾರ್ಡ್ ಮತ್ತು ಜಾನ್ ಮೆಯನ್', 'ko': '스발바르제도-얀마웬섬', 'kok': 'स्वालबार्ड आनी जान मेयन', 'kok-Deva': 'स्वालबार्ड आनी जान मेयन', 'ks': 'سَوالبریڑ تہٕ جان ماییڑ', 'ks-Arab': 'سَوالبریڑ تہٕ جان ماییڑ', 'ksh': 'de Enselle Svalbard un Jan Mayen', 'ku': 'Svalbard û Jan Mayen', 'kxv': 'svalbard aḍe jan mayen', 'kxv-Deva': 'स्वालबार्ड ऑड़े जान मायेन', 'kxv-Latn': 'svalbard aḍe jan mayen', 'kxv-Orya': 'ସ୍ୱାଲବାର୍ଡ ଅଡ଼େ ଜାନ ମାୟେନ', 'kxv-Telu': 'స్వాల్\u200cబార్డ్ మరియు జాన్ మాయెన్', 'ky': 'Шпицберген жана Ян-Майен', 'lb': 'Svalbard a Jan Mayen', 'lld': 'Svalbard y Jan Mayen', 'lo': 'ສະວາບາ ແລະ ແຢນ ມາເຢນ', 'lt': 'Svalbardas ir Janas Majenas', 'lv': 'Svalbāra un Jana Majena sala', 'mai': 'स्वालबार्ड आ जान मायेन', 'mi': 'Heopara me Iana Maiana', 'mk': 'Свалбард и Јан Мајен', 'ml': 'സ്വാൽബാഡും ജാൻ മായേനും', 'mn': 'Свалбард ба Ян Майен', 'mr': 'स्वालबर्ड आणि जान मायेन', 'ms': 'Svalbard dan Jan Mayen', 'mt': 'Svalbard u Jan Mayen', 'my': 'စဗိုလ်ဘတ်နှင့်ဂျန်မေရန်', 'mzn': 'سوالبارد و يان ماين', 'nb': 'Svalbard og Jan Mayen', 'ne': 'सभाल्बार्ड र जान मायेन', 'nl': 'Spitsbergen en Jan Mayen', 'nn': 'Svalbard og Jan Mayen', 'no': 'Svalbard og Jan Mayen', 'nqo': 'ߛߊߟߑߓߊߙ ߣߌ߫ ߖ߭ߊ߲ ߡߊߦߍ߲߫', 'om': 'Isvaalbaard fi Jan Mayeen', 'or': 'ସାଲବାର୍ଡ ଏବଂ ଜାନ୍\u200c ମାୟେନ୍\u200c', 'pa': 'ਸਵਾਲਬਰਡ ਅਤੇ ਜਾਨ ਮਾਯੇਨ', 'pa-Guru': 'ਸਵਾਲਬਰਡ ਅਤੇ ਜਾਨ ਮਾਯੇਨ', 'pcm': 'Sválbad & Jén Meyẹn', 'pl': 'Svalbard i Jan Mayen', 'ps': 'سوالبارد او جان ميين', 'pt': 'Svalbard e Jan Mayen', 'qu': 'Svalbard y Jan Mayen', 'rm': 'Svalbard e Jan Mayen', 'ro': 'Svalbard și Jan Mayen', 'ru': 'Шпицберген и Ян-Майен', 'sc': 'Svalbard e Jan Mayen', 'sd': 'سوالبارڊ ۽ جان ماین', 'sd-Arab': 'سوالبارڊ ۽ جان ماین', 'se': 'Svalbárda ja Jan Mayen', 'si': 'ස්වෙල්බර්ඩ් සහ ජේන් මයේන්', 'sk': 'Svalbard a Jan Mayen', 'sl': 'Svalbard in Jan Mayen', 'smn': 'Čokkeväärih já Jan Mayen', 'so': 'Jasiiradda Sfaldbaad & Jaan Mayen', 'sq': 'Svalbard e Jan-Majen', 'sr': 'Свалбард и Јан Мајен', 'sr-Cyrl': 'Свалбард и Јан Мајен', 'sr-Latn': 'Svalbard i Jan Majen', 'sv': 'Svalbard och Jan Mayen', 'sw': 'Svalbard na Jan Mayen', 'syr': 'ܣܒܠܕܒܪܕ ܘܓܐܢ ܡܐܝܝܢ', 'ta': 'ஸ்வல்பார்டு & ஜான் மேயன்', 'te': 'స్వాల్\u200cబార్డ్ మరియు జాన్ మాయెన్', 'tg': 'Шпитсберген ва Ян Майен', 'th': 'สฟาลบาร์และยานไมเอน', 'ti': 'ስቫልባርድን ጃን ማየንን', 'tk': 'Şpisbergen we Ýan-Maýen', 'to': 'Sivolopāti mo Sani Maieni', 'tr': 'Svalbard ve Jan Mayen', 'tt': 'Шпицберген һәм Ян-Майен', 'ug': 'سىۋالبارد ۋە يان مايېن', 'uk': 'Шпіцберген та Ян-Маєн', 'ur': 'سوالبرڈ اور جان ماین', 'uz': 'Shpitsbergen va Yan-Mayen', 'uz-Cyrl': 'Свалбард ва Ян-Майен', 'uz-Latn': 'Shpitsbergen va Yan-Mayen', 'vai': 'ꔻꕙꕒꔵ ꗪ ꕧꘋ ꕮꘂꘋ', 'vai-Vaii': 'ꔻꕙꕒꔵ ꗪ ꕧꘋ ꕮꘂꘋ', 'vec': 'Zvàlbard e Jan Mayen', 'vi': 'Svalbard và Jan Mayen', 'wae': 'Svalbard und Jan Mayen', 'wo': 'Swalbaar ak Jan Mayen', 'xh': 'ESvalbard & Jan Mayen', 'xnr': 'स्वालबार्ड कने जान मायेन', 'yo': 'Sífábáàdì àti Jánì Máyẹ̀nì', 'yrl': 'Siwaubati asuí Yã Mayeni', 'yue': '斯瓦爾巴特群島同揚馬延島', 'yue-Hans': '斯瓦尔巴特群岛同扬马延岛', 'yue-Hant': '斯瓦爾巴特群島同揚馬延島', 'zh': '斯瓦尔巴和扬马延', 'zh-Hans': '斯瓦尔巴和扬马延', 'zh-Hant': '挪威屬斯瓦巴及尖棉', 'zu': 'i-Svalbard ne-Jan Mayen'},
    'SK': {'af': 'Slowakye', 'agq': 'Sɨ̀lòvɨkɨ̀a', 'ak': 'Slovakia', 'am': 'ስሎቫኪያ', 'ar': 'سلوفاكيا', 'as': 'শ্লোভাকিয়া', 'asa': 'Tholvakia', 'ast': 'Eslovaquia', 'az': 'Slovakiya', 'az-Cyrl': 'Словакија', 'az-Latn': 'Slovakiya', 'bal-Latn': 'Slowákiá', 'bas': 'Slòvakìà', 'be': 'Славакія', 'be-tarask': 'Славакія', 'bez': 'Huslovakia', 'bg': 'Словакия', 'blo': 'Slofakii', 'bm': 'Slowaki', 'bn': 'স্লোভাকিয়া', 'br': 'Slovakia', 'brx': 'स्लभाकिया', 'bs': 'Slovačka', 'bs-Cyrl': 'Словачка', 'bs-Latn': 'Slovačka', 'ca': 'Eslovàquia', 'ca-ES-valencia': 'Eslovàquia', 'ccp': '𑄥𑄳𑄣𑄮𑄞𑄇𑄨𑄠', 'ce': 'Словаки', 'ceb': 'Slovakia', 'cgg': 'Sirovaakiya', 'chr': 'ᏍᎶᏩᎩᎠ', 'ckb': 'سلۆڤاکیا', 'cs': 'Slovensko', 'cv': 'Словаки', 'cy': 'Slofacia', 'da': 'Slovakiet', 'dav': 'Slovakia', 'de': 'Slowakei', 'dje': 'Slovaaki', 'dsb': 'Słowakska', 'dyo': 'Slovaki', 'dz': 'སུ་ལོ་བཱ་ཀི་ཡ', 'ebu': 'Slovakia', 'ee': 'Slovakia nutome', 'el': 'Σλοβακία', 'el-polyton': 'Σλοβακία', 'en': 'Slovakia', 'en-Dsrt': '𐐝𐑊𐐬𐑂𐐪𐐿𐐨𐐲', 'eo': 'Slovakujo', 'es': 'Eslovaquia', 'et': 'Slovakkia', 'eu': 'Eslovakia', 'ewo': 'Səlovakí', 'fa': 'اسلواکی', 'ff': 'Slowakii', 'ff-Adlm': '𞤅𞤵𞤤𞤮𞤾𞤢𞥄𞤳𞤭𞤴𞤢', 'ff-Latn': 'Slowakii', 'fi': 'Slovakia', 'fil': 'Slovakia', 'fo': 'Slovakia', 'fr': 'Slovaquie', 'fur': 'Slovachie', 'fy': 'Slowakije', 'ga': 'an tSlóvaic', 'gd': 'An t-Slòbhac', 'gl': 'Eslovaquia', 'gsw': 'Slowakäi', 'gu': 'સ્લોવેકિયા', 'guz': 'Slovakia', 'ha': 'Sulobakiya', 'he': 'סלובקיה', 'hi': 'स्लोवाकिया', 'hi-Latn': 'Slovakia', 'hr': 'Slovačka', 'hsb': 'Słowakska', 'ht': 'Slovaquie', 'hu': 'Szlovákia', 'hy': 'Սլովակիա', 'ia': 'Slovachia', 'id': 'Slovakia', 'ie': 'Slovakia', 'ig': 'Slovakia', 'is': 'Slóvakía', 'it': 'Slovacchia', 'ja': 'スロバキア', 'jmc': 'Slovakia', 'jv': 'Slowak', 'ka': 'სლოვაკეთი', 'kaa': 'Словакия', 'kaa-Cyrl': 'Словакия', 'kab': 'Sluvakya', 'kam': 'Slovakia', 'kde': 'Slovakia', 'kea': 'Slovákia', 'kgp': 'Enhrovakija', 'khq': 'Slovaaki', 'ki': 'Slovakia', 'kk': 'Словакия', 'kk-Cyrl': 'Словакия', 'kln': 'Emetab Slovakia', 'km': 'ស្លូវ៉ាគី', 'kn': 'ಸ್ಲೊವಾಕಿಯಾ', 'ko': '슬로바키아', 'kok': 'स्लोवाकिया', 'kok-Deva': 'स्लोवाकिया', 'ks': 'سَلوواکِیا', 'ks-Arab': 'سَلوواکِیا', 'ksb': 'Slovakia', 'ksf': 'slovakí', 'ksh': 'ẞlovakei', 'ku': 'Slovakya', 'kxv': 'slovakia', 'kxv-Deva': 'स्लोवाकिया', 'kxv-Latn': 'slovakia', 'kxv-Orya': 'ସ୍ଲବାକିୟା', 'kxv-Telu': 'స్లొవేకియా', 'ky': 'Словакия', 'lag': 'Sulováakia', 'lb': 'Slowakei', 'lg': 'Sirovakya', 'lld': 'Slovachia', 'ln': 'Silovaki', 'lo': 'ສະໂລວາເກຍ', 'lt': 'Slovakija', 'lu': 'Silovaki', 'luo': 'Slovakia', 'luy': 'Slovakia', 'lv': 'Slovākija', 'mai': 'स्लोवाकिया', 'mas': 'Slovakia', 'mer': 'Slovakia', 'mfe': 'Slovaki', 'mg': 'Slovakia', 'mgh': 'Uslovakia', 'mi': 'Horowākia', 'mk': 'Словачка', 'ml': 'സ്ലോവാക്യ', 'mn': 'Словак', 'mr': 'स्लोव्हाकिया', 'ms': 'Slovakia', 'mt': 'is-Slovakkja', 'mua': 'Slovakiya', 'my': 'ဆလိုဗက်ကီးယား', 'mzn': 'اسلواکی', 'naq': 'Slovakia', 'nb': 'Slovakia', 'nd': 'Slovakia', 'ne': 'स्लोभाकिया', 'nl': 'Slowakije', 'nmg': 'Slovakia', 'nn': 'Slovakia', 'no': 'Slovakia', 'nqo': 'ߛߑߟߏߝ߭ߊߞߌ߫', 'nyn': 'Sirovaakiya', 'om': 'Isloovaakiyaa', 'or': 'ସ୍ଲୋଭାକିଆ', 'pa': 'ਸਲੋਵਾਕੀਆ', 'pa-Guru': 'ਸਲੋਵਾਕੀਆ', 'pcm': 'Slovékia', 'pl': 'Słowacja', 'ps': 'سلواکیا', 'pt': 'Eslováquia', 'qu': 'Eslovaquia', 'rm': 'Slovachia', 'rn': 'Silovakiya', 'ro': 'Slovacia', 'rof': 'Slovakia', 'ru': 'Словакия', 'rwk': 'Slovakia', 'saq': 'Slovakia', 'sbp': 'Silovakiya', 'sc': 'Islovàchia', 'sd': 'سلوواڪيا', 'sd-Arab': 'سلوواڪيا', 'se': 'Slovákia', 'seh': 'Eslováquia', 'ses': 'Slovaaki', 'sg': 'Solovakïi', 'shi': 'ⵙⵍⵓⴼⴰⴽⵢⴰ', 'shi-Latn': 'slufakya', 'shi-Tfng': 'ⵙⵍⵓⴼⴰⴽⵢⴰ', 'si': 'ස්ලෝවැකියාව', 'sk': 'Slovensko', 'sl': 'Slovaška', 'smn': 'Slovakia', 'sn': 'Slovakia', 'so': 'Islofaakiya', 'sq': 'Sllovaki', 'sr': 'Словачка', 'sr-Cyrl': 'Словачка', 'sr-Latn': 'Slovačka', 'sv': 'Slovakien', 'sw': 'Slovakia', 'syr': 'ܣܠܘܒܩܝܐ', 'ta': 'ஸ்லோவாகியா', 'te': 'స్లొవేకియా', 'teo': 'Slovakia', 'tg': 'Словакия', 'th': 'สโลวะเกีย', 'ti': 'ስሎቫክያ', 'tk': 'Slowakiýa', 'to': 'Silōvakia', 'tr': 'Slovakya', 'tt': 'Словакия', 'twq': 'Slovaaki', 'tzm': 'Sluvakya', 'ug': 'سىلوۋاكىيە', 'uk': 'Словаччина', 'ur': 'سلوواکیہ', 'uz': 'Slovakiya', 'uz-Cyrl': 'Словакия', 'uz-Latn': 'Slovakiya', 'vai': 'ꔻꖃꕙꕃꕩ', 'vai-Latn': 'Sulovakiya', 'vai-Vaii': 'ꔻꖃꕙꕃꕩ', 'vec': 'Zlovachia', 'vi': 'Slovakia', 'vun': 'Slovakia', 'wae': 'Slowakei', 'wo': 'Eslowaki', 'xh': 'ESlovakia', 'xnr': 'स्लोवाकिया', 'xog': 'Sirovakya', 'yav': 'silofakí', 'yi': 'סלאוואַקיי', 'yo': 'Silofakia', 'yrl': 'Esiruwakiya', 'yue': '斯洛伐克', 'yue-Hans': '斯洛伐克', 'yue-Hant': '斯洛伐克', 'zgh': 'ⵙⵍⵓⴼⴰⴽⵢⴰ', 'zh': '斯洛伐克', 'zh-Hans': '斯洛伐克', 'zh-Hant': '斯洛伐克', 'zu': 'i-Slovakia'},
    'SL': {'af': 'Sierra Leone', 'agq': 'Silìa lûŋ', 'ak': 'Sɛra Liɔn', 'am': 'ሴራሊዮን', 'ar': 'سيراليون', 'as': 'চিয়েৰা লিঅ’ন', 'asa': 'Thiera Leoni', 'ast': 'Sierra Lleona', 'az': 'Syerra-Leone', 'az-Cyrl': 'Сјерра-Леоне', 'az-Latn': 'Syerra-Leone', 'bal-Latn': 'Sierrá Leóne', 'bas': 'Sièra Lèɔ̂n', 'be': 'Сьера-Леонэ', 'be-tarask': 'Сьера-Леонэ', 'bez': 'Husiela Lioni', 'bg': 'Сиера Леоне', 'blo': 'Seraleyɔn', 'bm': 'Siyera Lewɔni', 'bn': 'সিয়েরা লিওন', 'br': 'Sierra Leone', 'brx': 'सियेरा लेओन', 'bs': 'Sijera Leone', 'bs-Cyrl': 'Сијера Леоне', 'bs-Latn': 'Sijera Leone', 'ca': 'Sierra Leone', 'ca-ES-valencia': 'Sierra Leone', 'ccp': '𑄥𑄨𑄠𑄬𑄢𑄣𑄨𑄃𑄮𑄚𑄴', 'ce': 'Сьерра- Леоне', 'ceb': 'Sierra Leone', 'cgg': 'Sirra Riyooni', 'chr': 'ᏏᎡᎳ ᎴᎣᏂ', 'ckb': 'سیەرالیۆن', 'cs': 'Sierra Leone', 'cv': 'Сьерра-Леоне', 'cy': 'Sierra Leone', 'da': 'Sierra Leone', 'dav': 'Siera Leoni', 'de': 'Sierra Leone', 'dje': 'Seera Leon', 'dsb': 'Sierra Leone', 'dyo': 'Serra Leon', 'dz': 'སི་ར་ ལི་འོན', 'ebu': 'Siera Leoni', 'ee': 'Sierra Leone nutome', 'el': 'Σιέρα Λεόνε', 'el-polyton': 'Σιέρα Λεόνε', 'en': 'Sierra Leone', 'en-Dsrt': '𐐝𐐨𐐯𐑉𐐲 𐐢𐐨𐐬𐑌', 'eo': 'Sieraleono', 'es': 'Sierra Leona', 'et': 'Sierra Leone', 'eu': 'Sierra Leona', 'ewo': 'Sierá-leónə', 'fa': 'سیرالئون', 'ff': 'Seraa liyon', 'ff-Adlm': '𞤅𞤢𞤪𞤢𞤤𞤮𞤲', 'ff-Latn': 'Seraa liyon', 'fi': 'Sierra Leone', 'fil': 'Sierra Leone', 'fo': 'Sierra Leona', 'fr': 'Sierra Leone', 'fur': 'Sierra Leone', 'fy': 'Sierra Leone', 'ga': 'Siarra Leon', 'gd': 'Siarra Leòmhann', 'gl': 'Serra Leoa', 'gsw': 'Sierra Leoone', 'gu': 'સીએરા લેઓન', 'guz': 'Siera Leoni', 'ha': 'Salewo', 'he': 'סיירה לאון', 'hi': 'सिएरा लियोन', 'hi-Latn': 'Sierra Leone', 'hr': 'Sijera Leone', 'hsb': 'Sierra Leone', 'ht': 'Sierra Leone', 'hu': 'Sierra Leone', 'hy': 'Սիեռա Լեոնե', 'ia': 'Sierra Leone', 'id': 'Sierra Leone', 'ig': 'Sierra Leone', 'is': 'Síerra Leóne', 'it': 'Sierra Leone', 'ja': 'シエラレオネ', 'jgo': 'Siyɛ́la Lɛɔ̂n', 'jmc': 'Siera Leoni', 'jv': 'Siéra Léoné', 'ka': 'სიერა-ლეონე', 'kaa': 'Сьерра-Леоне', 'kaa-Cyrl': 'Сьерра-Леоне', 'kab': 'Sira Lyun', 'kam': 'Siera Leoni', 'kde': 'Siela Leoni', 'kea': 'Sera Lioa', 'kgp': 'Krĩ Mĩgkusũg-fi', 'khq': 'Seera Leon', 'ki': 'Siera Leoni', 'kk': 'Сьерра-Леоне', 'kk-Cyrl': 'Сьерра-Леоне', 'kln': 'Emetab Sierra Leone', 'km': 'សៀរ៉ាឡេអូន', 'kn': 'ಸಿಯೆರ್ರಾ ಲಿಯೋನ್', 'ko': '시에라리온', 'kok': 'सिएरा लियॉन', 'kok-Deva': 'सिएरा लियॉन', 'ks': 'سیرا لیون', 'ks-Arab': 'سیرا لیون', 'ksb': 'Siela Leoni', 'ksf': 'syɛraleon', 'ksh': 'Sjärra Lejone', 'ku': 'Sierra Leone', 'kxv': 'sierra leyon', 'kxv-Deva': 'सिएरा लियोन', 'kxv-Latn': 'sierra leyon', 'kxv-Orya': 'ସିଏରା ଲିୟନ', 'kxv-Telu': 'సియెర్రా లియాన్', 'ky': 'Сьерра-Леоне', 'lag': 'Seraleóoni', 'lb': 'Sierra Leone', 'lg': 'Siyeralewone', 'lld': 'Sierra Leone', 'ln': 'Siera Leonɛ', 'lo': 'ເຊຍຣາ ລີໂອນ', 'lt': 'Siera Leonė', 'lu': 'Siera Leone', 'luo': 'Sierra Leone', 'luy': 'Siera Leoni', 'lv': 'Sjerraleone', 'mai': 'सियरा लियोन', 'mas': 'Siera Leoni', 'mer': 'Siera Leoni', 'mfe': 'Sierra Leone', 'mg': 'Sierra Leone', 'mi': 'Te Araone', 'mk': 'Сиера Леоне', 'ml': 'സിയെറ ലിയോൺ', 'mn': 'Сьерра-Леоне', 'mr': 'सिएरा लिओन', 'ms': 'Sierra Leone', 'mt': 'Sierra Leone', 'mua': 'Sierra Leonǝ', 'my': 'ဆီယာရာ လီယွန်း', 'mzn': 'سیرالئون', 'naq': 'Sierra Leone', 'nb': 'Sierra Leone', 'nd': 'Sierra Leone', 'ne': 'सिएर्रा लिओन', 'nl': 'Sierra Leone', 'nmg': 'Sierra Leɔn', 'nn': 'Sierra Leone', 'no': 'Sierra Leone', 'nqo': 'ߛߙߊ߬ߟߏ߲߫', 'nyn': 'Sirra Riyooni', 'om': 'Seeraaliyoon', 'or': 'ସିଏରା ଲିଓନ', 'pa': 'ਸਿਏਰਾ ਲਿਓਨ', 'pa-Guru': 'ਸਿਏਰਾ ਲਿਓਨ', 'pcm': 'Siẹ́ra Líon', 'pl': 'Sierra Leone', 'ps': 'سییرا لیون', 'pt': 'Serra Leoa', 'qu': 'Sierra Leona', 'rm': 'Sierra Leone', 'rn': 'Siyeralewone', 'ro': 'Sierra Leone', 'rof': 'Siera Leoni', 'ru': 'Сьерра-Леоне', 'rwk': 'Siera Leoni', 'saq': 'Siera Leoni', 'sbp': 'Siela Liyoni', 'sc': 'Sierra Leone', 'sd': 'سيرا ليون', 'sd-Arab': 'سيرا ليون', 'se': 'Sierra Leone', 'seh': 'Serra Leoa', 'ses': 'Seera Leon', 'sg': 'Sierä-Leône', 'shi': 'ⵙⵙⵉⵔⴰⵍⵢⵓⵏ', 'shi-Latn': 'ssiralyun', 'shi-Tfng': 'ⵙⵙⵉⵔⴰⵍⵢⵓⵏ', 'si': 'සියරාලියෝන්', 'sk': 'Sierra Leone', 'sl': 'Sierra Leone', 'smn': 'Sierra Leone', 'sn': 'Sierra Leone', 'so': 'Siraaliyoon', 'sq': 'Sierra-Leone', 'sr': 'Сијера Леоне', 'sr-Cyrl': 'Сијера Леоне', 'sr-Latn': 'Sijera Leone', 'sv': 'Sierra Leone', 'sw': 'Siera Leoni', 'syr': 'ܣܝܝܪܐ ܠܝܐܘܢܝ', 'ta': 'சியாரா லியோன்', 'te': 'సియెర్రా లియాన్', 'teo': 'Siera Leoni', 'tg': 'Сиерра-Леоне', 'th': 'เซียร์ราลีโอน', 'ti': 'ሴራ ልዮን', 'tk': 'Sýerra-Leone', 'to': 'Siela Leone', 'tr': 'Sierra Leone', 'tt': 'Сьерра-Леоне', 'twq': 'Seera Leon', 'tzm': 'Siralyun', 'ug': 'سېررالېئون', 'uk': 'Сьєрра-Леоне', 'ur': 'سیرالیون', 'uz': 'Syerra-Leone', 'uz-Cyrl': 'Сьерра-Леоне', 'uz-Latn': 'Syerra-Leone', 'vai': 'ꔋꕩ ꕒꕌꖺ ꕸꖃꔀ', 'vai-Latn': 'Gbeya Bahawɔ', 'vai-Vaii': 'ꔋꕩ ꕒꕌꖺ ꕸꖃꔀ', 'vec': 'Siera Leon', 'vi': 'Sierra Leone', 'vun': 'Siera Leoni', 'wae': 'Sierra Leone', 'wo': 'Siyera Lewon', 'xh': 'ESierra Leone', 'xnr': 'सिएरा लियोन', 'xog': 'Siyeralewone', 'yav': 'sieláleyɔ́ɔn', 'yi': 'סיערע לעאנע', 'yo': 'Siria looni', 'yrl': 'Iwitera Leowa', 'yue': '獅子山', 'yue-Hans': '狮子山', 'yue-Hant': '獅子山', 'zgh': 'ⵙⵙⵉⵔⴰⵍⵢⵓⵏ', 'zh': '塞拉利昂', 'zh-Hans': '塞拉利昂', 'zh-Hant': '獅子山', 'zu': 'i-Sierra Leone'},
    'SM': {'af': 'San Marino', 'agq': 'Sàn Màlenù', 'ak': 'San Marino', 'am': 'ሳን ማሪኖ', 'ar': 'سان مارينو', 'as': 'চান মাৰিনো', 'asa': 'Thamarino', 'ast': 'San Marín', 'az': 'San-Marino', 'az-Cyrl': 'Сан-Марино', 'az-Latn': 'San-Marino', 'bal-Latn': 'San Mariánó', 'bas': 'Nûmpubi Māatìn', 'be': 'Сан-Марына', 'be-tarask': 'Сан-Марына', 'bez': 'Husamalino', 'bg': 'Сан Марино', 'blo': 'Sɛŋ Marinoo', 'bm': 'Marini-Senu', 'bn': 'সান মারিনো', 'br': 'San Marino', 'brx': 'सान मेरिन’', 'bs': 'San Marino', 'bs-Cyrl': 'Сан Марино', 'bs-Latn': 'San Marino', 'ca': 'San Marino', 'ca-ES-valencia': 'San Marino', 'ccp': '𑄥𑄚𑄴 𑄟𑄢𑄨𑄚𑄮', 'ce': 'Сан-Марино', 'ceb': 'San Marino', 'cgg': 'Samarino', 'chr': 'ᎤᏓᏅᏘ ᎹᎵᎢᏃ', 'ckb': 'سان مارینۆ', 'cs': 'San Marino', 'cv': 'Сан-Марино', 'cy': 'San Marino', 'da': 'San Marino', 'dav': 'Samarino', 'de': 'San Marino', 'dje': 'San Marino', 'dsb': 'San Marino', 'dz': 'སཱན་མ་རི་ནོ', 'ebu': 'Samarino', 'ee': 'San Marino nutome', 'el': 'Άγιος Μαρίνος', 'el-polyton': 'Ἅγιος Μαρίνος', 'en': 'San Marino', 'en-Dsrt': '𐐝𐐪𐑌 𐐣𐐲𐑉𐐨𐑌𐐬', 'eo': 'Sanmarino', 'es': 'San Marino', 'et': 'San Marino', 'eu': 'San Marino', 'ewo': 'Ǹfúfúb Maríno', 'fa': 'سان\u200cمارینو', 'ff': 'See Maree', 'ff-Adlm': '𞤅𞤢𞤲 𞤃𞤢𞤪𞤭𞤲𞤮𞥅', 'ff-Latn': 'See Maree', 'fi': 'San Marino', 'fil': 'San Marino', 'fo': 'San Marino', 'fr': 'Saint-Marin', 'fur': 'San Marin', 'fy': 'San Marino', 'ga': 'San Mairíne', 'gd': 'San Marino', 'gl': 'San Marino', 'gsw': 'San Mariino', 'gu': 'સૅન મેરિનો', 'guz': 'Samarino', 'ha': 'San Marino', 'he': 'סן מרינו', 'hi': 'सैन मेरीनो', 'hi-Latn': 'San Marino', 'hr': 'San Marino', 'hsb': 'San Marino', 'ht': 'Saint-Marin', 'hu': 'San Marino', 'hy': 'Սան Մարինո', 'ia': 'San Marino', 'id': 'San Marino', 'ie': 'San-Marino', 'ig': 'San Marino', 'is': 'San Marínó', 'it': 'San Marino', 'ja': 'サンマリノ', 'jmc': 'Samarino', 'jv': 'San Marino', 'ka': 'სან-მარინო', 'kaa': 'Сан-Марино', 'kaa-Cyrl': 'Сан-Марино', 'kab': 'San Marinu', 'kam': 'Samarino', 'kde': 'Samalino', 'kea': 'San Marinu', 'kgp': 'Sỹ Mỹrĩnũ', 'khq': 'San Marino', 'ki': 'Samarino', 'kk': 'Сан-Марино', 'kk-Cyrl': 'Сан-Марино', 'kln': 'Emetab San Marino', 'km': 'សាន\u200bម៉ារីណូ', 'kn': 'ಸ್ಯಾನ್ ಮೆರಿನೋ', 'ko': '산마리노', 'kok': 'सॅन मारीनो', 'kok-Deva': 'सॅन मारीनो', 'ks': 'سین میرِنو', 'ks-Arab': 'سین میرِنو', 'ksb': 'Samalino', 'ksf': 'sɛnmarǝn', 'ksh': 'San-Marino', 'ku': 'San Marîno', 'kxv': 'san marino', 'kxv-Deva': 'सेन मेरीनो', 'kxv-Latn': 'san marino', 'kxv-Orya': 'ସନ ମାରିନ', 'kxv-Telu': 'సస్ మారిసో', 'ky': 'Сан Марино', 'lag': 'Samaríino', 'lb': 'San Marino', 'lg': 'Sanimarino', 'lld': 'San Marino', 'ln': 'Sántu Marinɛ', 'lo': 'ແຊນ ມາຣິໂນ', 'lt': 'San Marinas', 'lu': 'Santu Marine', 'luo': 'San Marino', 'luy': 'Samarino', 'lv': 'Sanmarīno', 'mai': 'सेन मरीनो', 'mas': 'Samarino', 'mer': 'Samarino', 'mfe': 'Saint-Marin', 'mg': 'Saint-Marin', 'mgh': 'Usamarino', 'mi': 'Hana Marino', 'mk': 'Сан Марино', 'ml': 'സാൻ മറിനോ', 'mn': 'Сан-Марино', 'mr': 'सॅन मरीनो', 'ms': 'San Marino', 'mt': 'San Marino', 'mua': 'Sǝr Marino', 'my': 'ဆန်မာရီနို', 'mzn': 'سن مارینو', 'naq': 'San Marino', 'nb': 'San Marino', 'nd': 'San Marino', 'ne': 'सान् मारिनो', 'nl': 'San Marino', 'nmg': 'San Marino', 'nn': 'San Marino', 'no': 'San Marino', 'nqo': 'ߛߍߕ-ߡߊߙߍ߲߫', 'nyn': 'Samarino', 'om': 'Saan Mariinoo', 'or': 'ସାନ୍ ମାରିନୋ', 'pa': 'ਸੈਨ ਮਰੀਨੋ', 'pa-Guru': 'ਸੈਨ ਮਰੀਨੋ', 'pcm': 'San Maríno', 'pl': 'San Marino', 'ps': 'سان مارینو', 'pt': 'San Marino', 'qu': 'San Marino', 'rm': 'San Marino', 'rn': 'Sanimarino', 'ro': 'San Marino', 'rof': 'Samarino', 'ru': 'Сан-Марино', 'rwk': 'Samarino', 'saq': 'Samarino', 'sbp': 'Samalino', 'sc': 'Santu Marinu', 'sd': 'سین مرینو', 'sd-Arab': 'سین مرینو', 'se': 'San Marino', 'seh': 'San Marino', 'ses': 'San Marino', 'sg': 'Sên-Marëen', 'shi': 'ⵙⴰⵏⵎⴰⵔⵉⵏⵓ', 'shi-Latn': 'sanmarinu', 'shi-Tfng': 'ⵙⴰⵏⵎⴰⵔⵉⵏⵓ', 'si': 'සැන් මැරිනෝ', 'sk': 'San Maríno', 'sl': 'San Marino', 'smn': 'San Marino', 'sn': 'San Marino', 'so': 'San Marino', 'sq': 'San-Marino', 'sr': 'Сан Марино', 'sr-Cyrl': 'Сан Марино', 'sr-Latn': 'San Marino', 'sv': 'San Marino', 'sw': 'San Marino', 'syr': 'ܣܢ ܡܪܝܢܘ', 'ta': 'சான் மரினோ', 'te': 'శాన్ మారినో', 'teo': 'Samarino', 'tg': 'Сан-Марино', 'th': 'ซานมาริโน', 'ti': 'ሳን ማሪኖ', 'tk': 'San-Marino', 'to': 'Sā Malino', 'tr': 'San Marino', 'tt': 'Сан-Марино', 'twq': 'San Marino', 'tzm': 'Sanmarinu', 'ug': 'سان مارىنو', 'uk': 'Сан-Марино', 'ur': 'سان مارینو', 'uz': 'San-Marino', 'uz-Cyrl': 'Сан-Марино', 'uz-Latn': 'San-Marino', 'vai': 'ꕮꔸꖆ ꕢꘋ', 'vai-Latn': 'Saŋ Marindo', 'vai-Vaii': 'ꕮꔸꖆ ꕢꘋ', 'vec': 'San Marin', 'vi': 'San Marino', 'vun': 'Samarino', 'wae': 'San Marino', 'wo': 'San Marino', 'xh': 'ESan Marino', 'xnr': 'सैन मेरीनो', 'xog': 'Sanimarino', 'yav': 'san malíno', 'yi': 'סאַן מאַרינא', 'yo': 'Sani Marino', 'yrl': 'Sã Marino', 'yue': '聖馬利諾', 'yue-Hans': '圣马利诺', 'yue-Hant': '聖馬利諾', 'zgh': 'ⵙⴰⵏⵎⴰⵔⵉⵏⵓ', 'zh': '圣马力诺', 'zh-Hans': '圣马力诺', 'zh-Hant': '聖馬利諾', 'zu': 'i-San Marino'},
    'SN': {'af': 'Senegal', 'agq': 'Sɛ̀nɛ̀gâa', 'ak': 'Senegal', 'am': 'ሴኔጋል', 'ar': 'السنغال', 'as': 'চেনেগাল', 'asa': 'Thenegali', 'ast': 'Senegal', 'az': 'Seneqal', 'az-Cyrl': 'Сенегал', 'az-Latn': 'Seneqal', 'bal-Latn': 'Senigál', 'bas': 'Sènègâl', 'be': 'Сенегал', 'be-tarask': 'Сенегал', 'bez': 'Husenegali', 'bg': 'Сенегал', 'bgn': 'سینیگال', 'blo': 'Senegal', 'bm': 'Senegali', 'bn': 'সেনেগাল', 'br': 'Senegal', 'brx': 'सेनेगाल', 'bs': 'Senegal', 'bs-Cyrl': 'Сенегал', 'bs-Latn': 'Senegal', 'ca': 'Senegal', 'ca-ES-valencia': 'Senegal', 'ccp': '𑄥𑄬𑄚𑄬𑄉𑄣𑄴', 'ce': 'Сенегал', 'ceb': 'Senegal', 'cgg': 'Senego', 'chr': 'ᏏᏂᎦᎵ', 'ckb': 'سێنێگاڵ', 'cs': 'Senegal', 'cv': 'Сенегал', 'cy': 'Senegal', 'da': 'Senegal', 'dav': 'Senegali', 'de': 'Senegal', 'dje': 'Senegal', 'dsb': 'Senegal', 'dyo': 'Senegal', 'dz': 'སེ་ནི་གྷལ', 'ebu': 'Senegali', 'ee': 'Senegal nutome', 'el': 'Σενεγάλη', 'el-polyton': 'Σενεγάλη', 'en': 'Senegal', 'en-Dsrt': '𐐝𐐯𐑌𐐲𐑀𐐱𐑊', 'eo': 'Senegalo', 'es': 'Senegal', 'et': 'Senegal', 'eu': 'Senegal', 'ewo': 'Senegál', 'fa': 'سنگال', 'ff': 'Senegaal', 'ff-Adlm': '𞤅𞤫𞤲𞤫𞤺𞤢𞥄𞤤', 'ff-Latn': 'Senegaal', 'fi': 'Senegal', 'fil': 'Senegal', 'fo': 'Senegal', 'fr': 'Sénégal', 'fur': 'Senegal', 'fy': 'Senegal', 'ga': 'An tSeineagáil', 'gd': 'Seanagal', 'gl': 'Senegal', 'gsw': 'Senegal', 'gu': 'સેનેગલ', 'guz': 'Senegali', 'ha': 'Sanigal', 'he': 'סנגל', 'hi': 'सेनेगल', 'hi-Latn': 'Senegal', 'hr': 'Senegal', 'hsb': 'Senegal', 'ht': 'Sénégal', 'hu': 'Szenegál', 'hy': 'Սենեգալ', 'ia': 'Senegal', 'id': 'Senegal', 'ig': 'Senegal', 'is': 'Senegal', 'it': 'Senegal', 'ja': 'セネガル', 'jgo': 'Sɛnɛgâl', 'jmc': 'Senegali', 'jv': 'Sénégal', 'ka': 'სენეგალი', 'kaa': 'Сенегал', 'kaa-Cyrl': 'Сенегал', 'kab': 'Sinigal', 'kam': 'Senegali', 'kde': 'Senegali', 'kea': 'Senegal', 'kgp': 'Senẽgar', 'khq': 'Senegal', 'ki': 'Senegali', 'kk': 'Сенегал', 'kk-Cyrl': 'Сенегал', 'kln': 'Emetab Senegal', 'km': 'សេណេហ្គាល់', 'kn': 'ಸೆನೆಗಲ್', 'ko': '세네갈', 'kok': 'सिनिगल', 'kok-Deva': 'सिनिगल', 'ks': 'سینیگَل', 'ks-Arab': 'سینیگَل', 'ksb': 'Senegali', 'ksf': 'sɛnɛgal', 'ksh': 'der Sennejall', 'ku': 'Senegal', 'kxv': 'senegal', 'kxv-Deva': 'सेनेगल', 'kxv-Latn': 'senegal', 'kxv-Orya': 'ସେନେଗାଲ', 'kxv-Telu': 'సెనెగల్', 'ky': 'Сенегал', 'lag': 'Senegáali', 'lb': 'Senegal', 'lg': 'Senegaalo', 'lld': 'Senegal', 'ln': 'Senegalɛ', 'lo': 'ເຊນີໂກລ', 'lt': 'Senegalas', 'lu': 'Senegale', 'luo': 'Senegal', 'luy': 'Senegali', 'lv': 'Senegāla', 'mai': 'सेनेगल', 'mas': 'Senegali', 'mer': 'Senego', 'mfe': 'Senegal', 'mg': 'Senegal', 'mgh': 'Usenegali', 'mi': 'Henekara', 'mk': 'Сенегал', 'ml': 'സെനഗൽ', 'mn': 'Сенегал', 'mr': 'सेनेगल', 'ms': 'Senegal', 'mt': 'is-Senegal', 'mua': 'Senegal', 'my': 'ဆီနီဂေါ', 'mzn': 'سنگال', 'naq': 'Senegali', 'nb': 'Senegal', 'nd': 'Senegali', 'ne': 'सेनेगल', 'nl': 'Senegal', 'nmg': 'Senegal', 'nn': 'Senegal', 'no': 'Senegal', 'nqo': 'ߛߣߍ߬ߜߊ߯ߟߌ߫', 'nyn': 'Senego', 'om': 'Senegaal', 'or': 'ସେନେଗାଲ୍', 'pa': 'ਸੇਨੇਗਲ', 'pa-Guru': 'ਸੇਨੇਗਲ', 'pcm': 'Sẹ́nẹ́gal', 'pl': 'Senegal', 'ps': 'سينيګال', 'pt': 'Senegal', 'qu': 'Senegal', 'rm': 'Senegal', 'rn': 'Senegali', 'ro': 'Senegal', 'rof': 'Senegali', 'ru': 'Сенегал', 'rwk': 'Senegali', 'saq': 'Senegali', 'sbp': 'Senegali', 'sc': 'Senegal', 'sd': 'سينيگال', 'sd-Arab': 'سينيگال', 'se': 'Senegal', 'seh': 'Senegal', 'ses': 'Senegal', 'sg': 'Senegäle', 'shi': 'ⵙⵙⵉⵏⵉⴳⴰⵍ', 'shi-Latn': 'ssinigal', 'shi-Tfng': 'ⵙⵙⵉⵏⵉⴳⴰⵍ', 'si': 'සෙනගාලය', 'sk': 'Senegal', 'sl': 'Senegal', 'smn': 'Senegal', 'sn': 'Senegal', 'so': 'Sinigaal', 'sq': 'Senegal', 'sr': 'Сенегал', 'sr-Cyrl': 'Сенегал', 'sr-Latn': 'Senegal', 'sv': 'Senegal', 'sw': 'Senegali', 'syr': 'ܣܢܓܐܠ', 'ta': 'செனெகல்', 'te': 'సెనెగల్', 'teo': 'Senegali', 'tg': 'Сенегал', 'th': 'เซเนกัล', 'ti': 'ሰነጋል', 'tk': 'Senegal', 'to': 'Senekalo', 'tr': 'Senegal', 'tt': 'Сенегал', 'twq': 'Senegal', 'tzm': 'Ssiniɣal', 'ug': 'سېنېگال', 'uk': 'Сенегал', 'ur': 'سینیگل', 'uz': 'Senegal', 'uz-Cyrl': 'Сенегал', 'uz-Latn': 'Senegal', 'vai': 'ꔻꕇꕭꕌ', 'vai-Latn': 'Sinigaha', 'vai-Vaii': 'ꔻꕇꕭꕌ', 'vec': 'Sènegal', 'vi': 'Senegal', 'vun': 'Senegali', 'wae': 'Senegal', 'wo': 'Senegaal', 'xh': 'ESenegal', 'xnr': 'सेनेगल', 'xog': 'Senegaalo', 'yav': 'senekáal', 'yi': 'סענעגאַל', 'yo': 'Sẹnẹga', 'yrl': 'Senegau', 'yue': '塞內加爾', 'yue-Hans': '塞内加尔', 'yue-Hant': '塞內加爾', 'zgh': 'ⵙⵙⵉⵏⵉⴳⴰⵍ', 'zh': '塞内加尔', 'zh-Hans': '塞内加尔', 'zh-Hant': '塞內加爾', 'zu': 'i-Senegal'},
    'SO': {'af': 'Somalië', 'agq': 'Sòmalìa', 'ak': 'Somalia', 'am': 'ሶማሊያ', 'ar': 'الصومال', 'as': 'চোমালিয়া', 'asa': 'Thomalia', 'ast': 'Somalia', 'az': 'Somali', 'az-Cyrl': 'Сомали', 'az-Latn': 'Somali', 'bal-Latn': 'Sómáliá', 'bas': 'Sòmalìà', 'be': 'Самалі', 'be-tarask': 'Самалі', 'bez': 'Husomalia', 'bg': 'Сомалия', 'bgn': 'سومالیا', 'blo': 'Somalii', 'bm': 'Somali', 'bn': 'সোমালিয়া', 'br': 'Somalia', 'brx': 'समालिया', 'bs': 'Somalija', 'bs-Cyrl': 'Сомалија', 'bs-Latn': 'Somalija', 'ca': 'Somàlia', 'ca-ES-valencia': 'Somàlia', 'ccp': '𑄥𑄮𑄟𑄣𑄨𑄠', 'ce': 'Сомали', 'ceb': 'Somalia', 'cgg': 'Somaariya', 'chr': 'ᏐᎹᎵ', 'ckb': 'سۆمالیا', 'cs': 'Somálsko', 'cv': 'Сомали', 'cy': 'Somalia', 'da': 'Somalia', 'dav': 'Somalia', 'de': 'Somalia', 'dje': 'Somaali', 'dsb': 'Somalija', 'dyo': 'Somali', 'dz': 'སོ་མ་ལི་ཡ', 'ebu': 'Somalia', 'ee': 'Somalia nutome', 'el': 'Σομαλία', 'el-polyton': 'Σομαλία', 'en': 'Somalia', 'en-Dsrt': '𐐝𐐲𐑋𐐪𐑊𐐨𐐲', 'eo': 'Somalujo', 'es': 'Somalia', 'et': 'Somaalia', 'eu': 'Somalia', 'ewo': 'Somália', 'fa': 'سومالی', 'ff': 'Somalii', 'ff-Adlm': '𞤅𞤵𞥅𞤥𞤢𞥄𞤤𞤭', 'ff-Latn': 'Somalii', 'fi': 'Somalia', 'fil': 'Somalia', 'fo': 'Somalia', 'fr': 'Somalie', 'fur': 'Somalie', 'fy': 'Somalië', 'ga': 'an tSomáil', 'gd': 'Somàilia', 'gl': 'Somalia', 'gsw': 'Somaalie', 'gu': 'સોમાલિયા', 'guz': 'Somalia', 'ha': 'Somaliya', 'he': 'סומליה', 'hi': 'सोमालिया', 'hi-Latn': 'Somalia', 'hr': 'Somalija', 'hsb': 'Somalija', 'ht': 'Somalie', 'hu': 'Szomália', 'hy': 'Սոմալի', 'ia': 'Somalia', 'id': 'Somalia', 'ig': 'Somalia', 'is': 'Sómalía', 'it': 'Somalia', 'ja': 'ソマリア', 'jgo': 'Sɔmalî', 'jmc': 'Somalia', 'jv': 'Somalia', 'ka': 'სომალი', 'kaa': 'Сомали', 'kaa-Cyrl': 'Сомали', 'kab': 'Ṣumal', 'kam': 'Somalia', 'kde': 'Somalia', 'kea': 'Sumália', 'kgp': 'Somỹrija', 'khq': 'Somaali', 'ki': 'Somaria', 'kk': 'Сомали', 'kk-Cyrl': 'Сомали', 'kln': 'Emetab Somalia', 'km': 'សូម៉ាលី', 'kn': 'ಸೊಮಾಲಿಯಾ', 'ko': '소말리아', 'kok': 'सोमालिया', 'kok-Deva': 'सोमालिया', 'ks': 'سومالِیا', 'ks-Arab': 'سومالِیا', 'ksb': 'Somalia', 'ksf': 'somalí', 'ksh': 'Somaalije', 'ku': 'Somalya', 'kxv': 'somaliya', 'kxv-Deva': 'सोमालिया', 'kxv-Latn': 'somaliya', 'kxv-Orya': 'ସମାଲିଆ', 'kxv-Telu': 'సోమలియా', 'ky': 'Сомали', 'lag': 'Somália', 'lb': 'Somalia', 'lg': 'Somaliya', 'lld': 'Somalia', 'ln': 'Somali', 'lo': 'ໂຊມາເລຍ', 'lt': 'Somalis', 'lu': 'Somali', 'luo': 'Somalia', 'luy': 'Somalia', 'lv': 'Somālija', 'mai': 'सोमालिया', 'mas': 'Somalia', 'mer': 'Somalia', 'mfe': 'Somali', 'mg': 'Somalia', 'mgh': 'Usomalia', 'mi': 'Hūmārie', 'mk': 'Сомалија', 'ml': 'സോമാലിയ', 'mn': 'Сомали', 'mr': 'सोमालिया', 'ms': 'Somalia', 'mt': 'is-Somalja', 'mua': 'Somaliya', 'my': 'ဆိုမာလီယာ', 'mzn': 'سومالی', 'naq': 'Somaliab', 'nb': 'Somalia', 'nd': 'Somaliya', 'ne': 'सोमालिया', 'nl': 'Somalië', 'nmg': 'Somália', 'nn': 'Somalia', 'no': 'Somalia', 'nqo': 'ߛߏߡߊߟߌ߫', 'nyn': 'Somaariya', 'om': 'Somaaliyaa', 'or': 'ସୋମାଲିଆ', 'pa': 'ਸੋਮਾਲੀਆ', 'pa-Guru': 'ਸੋਮਾਲੀਆ', 'pcm': 'Sọmália', 'pl': 'Somalia', 'ps': 'سومالیا', 'pt': 'Somália', 'qu': 'Somalia', 'rm': 'Somalia', 'rn': 'Somaliya', 'ro': 'Somalia', 'rof': 'Somalia', 'ru': 'Сомали', 'rwk': 'Somalia', 'saq': 'Somalia', 'sbp': 'Somaliya', 'sc': 'Somàlia', 'sd': 'سوماليا', 'sd-Arab': 'سوماليا', 'se': 'Somália', 'seh': 'Somália', 'ses': 'Somaali', 'sg': 'Somalïi', 'shi': 'ⵚⵚⵓⵎⴰⵍ', 'shi-Latn': 'ṣṣumal', 'shi-Tfng': 'ⵚⵚⵓⵎⴰⵍ', 'si': 'සෝමාලියාව', 'sk': 'Somálsko', 'sl': 'Somalija', 'smn': 'Somalia', 'sn': 'Somalia', 'so': 'Soomaaliya', 'sq': 'Somali', 'sr': 'Сомалија', 'sr-Cyrl': 'Сомалија', 'sr-Latn': 'Somalija', 'sv': 'Somalia', 'sw': 'Somalia', 'syr': 'ܨܘܡܐܠ', 'ta': 'சோமாலியா', 'te': 'సోమాలియా', 'teo': 'Somalia', 'tg': 'Сомалӣ', 'th': 'โซมาเลีย', 'ti': 'ሶማልያ', 'tk': 'Somali', 'to': 'Sōmalia', 'tr': 'Somali', 'tt': 'Сомали', 'twq': 'Somaali', 'tzm': 'Ṣṣumal', 'ug': 'سومالى', 'uk': 'Сомалі', 'ur': 'صومالیہ', 'uz': 'Somali', 'uz-Cyrl': 'Сомали', 'uz-Latn': 'Somali', 'vai': 'ꖇꕮꔷꕩ', 'vai-Latn': 'Somaliya', 'vai-Vaii': 'ꖇꕮꔷꕩ', 'vec': 'Somalia', 'vi': 'Somalia', 'vun': 'Somalia', 'wae': 'Somalia', 'wo': 'Somali', 'xh': 'ESomalia', 'xnr': 'सोमालिया', 'xog': 'Somaliya', 'yav': 'somalí', 'yi': 'סאמאַליע', 'yo': 'Somalia', 'yrl': 'Somariya', 'yue': '索馬利亞', 'yue-Hans': '索马利亚', 'yue-Hant': '索馬利亞', 'zgh': 'ⵚⵚⵓⵎⴰⵍ', 'zh': '索马里', 'zh-Hans': '索马里', 'zh-Hant': '索馬利亞', 'zu': 'i-Somalia'},
    'SR': {'af': 'Suriname', 'agq': 'Sulènamè', 'ak': 'Suriname', 'am': 'ሱሪናም', 'ar': 'سورينام', 'as': 'ছুৰিনাম', 'asa': 'Thurinamu', 'ast': 'Surinam', 'az': 'Surinam', 'az-Cyrl': 'Суринам', 'az-Latn': 'Surinam', 'bal-Latn': 'Surinaym', 'bas': 'Sùrinâm', 'be': 'Сурынам', 'be-tarask': 'Сурынам', 'bez': 'Husurinamu', 'bg': 'Суринам', 'bgn': 'سورینامی', 'blo': 'Surinam', 'bm': 'Surinami', 'bn': 'সুরিনাম', 'br': 'Surinam', 'brx': 'सुरिनाम', 'bs': 'Surinam', 'bs-Cyrl': 'Суринам', 'bs-Latn': 'Surinam', 'ca': 'Surinam', 'ca-ES-valencia': 'Surinam', 'ccp': '𑄥𑄪𑄢𑄨𑄚𑄟𑄴', 'ce': 'Суринам', 'ceb': 'Suriname', 'cgg': 'Surinaamu', 'chr': 'ᏒᎵᎾᎻ', 'ckb': 'سورینام', 'cs': 'Surinam', 'cv': 'Суринам', 'cy': 'Suriname', 'da': 'Surinam', 'dav': 'Surinamu', 'de': 'Suriname', 'dje': 'Surinaam', 'dsb': 'Surinamska', 'dz': 'སུ་རི་ནཱམ', 'ebu': 'Surinamu', 'ee': 'Suriname nutome', 'el': 'Σουρινάμ', 'el-polyton': 'Σουρινάμ', 'en': 'Suriname', 'en-Dsrt': '𐐝𐐭𐑉𐐲𐑌𐐪𐑋', 'eo': 'Surinamo', 'es': 'Surinam', 'et': 'Suriname', 'eu': 'Surinam', 'ewo': 'Surinám', 'fa': 'سورینام', 'ff': 'Surinaam', 'ff-Adlm': '𞤅𞤵𞤪𞤭𞤲𞤢𞥄𞤥', 'ff-Latn': 'Surinaam', 'fi': 'Suriname', 'fil': 'Suriname', 'fo': 'Surinam', 'fr': 'Suriname', 'fur': 'Suriname', 'fy': 'Suriname', 'ga': 'Suranam', 'gd': 'Suranam', 'gl': 'Suriname', 'gsw': 'Surinam', 'gu': 'સુરીનામ', 'guz': 'Surinamu', 'ha': 'Suriname', 'he': 'סורינאם', 'hi': 'सूरीनाम', 'hi-Latn': 'Suriname', 'hr': 'Surinam', 'hsb': 'Surinam', 'ht': 'Suriname', 'hu': 'Suriname', 'hy': 'Սուրինամ', 'ia': 'Suriname', 'id': 'Suriname', 'ig': 'Suriname', 'is': 'Súrínam', 'it': 'Suriname', 'ja': 'スリナム', 'jmc': 'Surinamu', 'jv': 'Suriname', 'ka': 'სურინამი', 'kaa': 'Суринам', 'kaa-Cyrl': 'Суринам', 'kab': 'Surinam', 'kam': 'Surinamu', 'kde': 'Sulinamu', 'kea': 'Surinami', 'kgp': 'Surinỹmĩ', 'khq': 'Surinaam', 'ki': 'Surinamu', 'kk': 'Суринам', 'kk-Cyrl': 'Суринам', 'kln': 'Emetab Suriname', 'km': 'សូរីណាម', 'kn': 'ಸುರಿನಾಮ್', 'ko': '수리남', 'kok': 'सुरीनाम', 'kok-Deva': 'सुरीनाम', 'ks': 'سُرِنام', 'ks-Arab': 'سُرِنام', 'ksb': 'Sulinamu', 'ksf': 'surinam', 'ksh': 'Sürinamm', 'ku': 'Surînam', 'kxv': 'surīname', 'kxv-Deva': 'सूरीनाम', 'kxv-Latn': 'surīname', 'kxv-Orya': 'ସୁରିନାମ', 'kxv-Telu': 'సూరినామ్', 'ky': 'Суринам', 'lag': 'Surináamu', 'lb': 'Suriname', 'lg': 'Surinaamu', 'lld': 'Surinam', 'ln': 'Surinamɛ', 'lo': 'ຊູຣິນາມ', 'lt': 'Surinamas', 'lu': 'Suriname', 'luo': 'Suriname', 'luy': 'Surinamu', 'lv': 'Surinama', 'mai': 'सुरीनाम', 'mas': 'Surinamu', 'mer': 'Surinamu', 'mfe': 'Surinam', 'mg': 'Sorinam', 'mgh': 'Usurinamu', 'mi': 'Huriname', 'mk': 'Суринам', 'ml': 'സുരിനാം', 'mn': 'Суринам', 'mr': 'सुरिनाम', 'ms': 'Surinam', 'mt': 'is-Suriname', 'mua': 'Sǝr Surinam', 'my': 'ဆူရာနမ်', 'mzn': 'سورینام', 'naq': 'Suriname', 'nb': 'Surinam', 'nd': 'Suriname', 'ne': 'सुरिनेम', 'nl': 'Suriname', 'nmg': 'Surinam', 'nn': 'Surinam', 'no': 'Surinam', 'nqo': 'ߛߎߙߑߣߊߡ', 'nyn': 'Surinaamu', 'om': 'Suriname', 'or': 'ସୁରିନାମ', 'pa': 'ਸੂਰੀਨਾਮ', 'pa-Guru': 'ਸੂਰੀਨਾਮ', 'pcm': 'Súrínam', 'pl': 'Surinam', 'ps': 'سورینام', 'pt': 'Suriname', 'qu': 'Surinam', 'rm': 'Surinam', 'rn': 'Suriname', 'ro': 'Suriname', 'rof': 'Surinamu', 'ru': 'Суринам', 'rwk': 'Surinamu', 'saq': 'Surinamu', 'sbp': 'Sulinamu', 'sc': 'Suriname', 'sd': 'سورينام', 'sd-Arab': 'سورينام', 'se': 'Surinam', 'seh': 'Suriname', 'ses': 'Surinaam', 'sg': 'Surinäm', 'shi': 'ⵙⵓⵔⵉⵏⴰⵎ', 'shi-Latn': 'surinam', 'shi-Tfng': 'ⵙⵓⵔⵉⵏⴰⵎ', 'si': 'සුරිනාමය', 'sk': 'Surinam', 'sl': 'Surinam', 'smn': 'Surinam', 'sn': 'Suriname', 'so': 'Surineym', 'sq': 'Surinami', 'sr': 'Суринам', 'sr-Cyrl': 'Суринам', 'sr-Latn': 'Surinam', 'sv': 'Surinam', 'sw': 'Suriname', 'syr': 'ܣܘܪܝܢܐܡ', 'ta': 'சுரினாம்', 'te': 'సూరినామ్', 'teo': 'Surinamu', 'tg': 'Суринам', 'th': 'ซูรินาเม', 'ti': 'ሱሪናም', 'tk': 'Surinam', 'to': 'Suliname', 'tr': 'Surinam', 'tt': 'Суринам', 'twq': 'Surinaam', 'tzm': 'Surinam', 'ug': 'سۇرىنام', 'uk': 'Суринам', 'ur': 'سورینام', 'uz': 'Surinam', 'uz-Cyrl': 'Суринам', 'uz-Latn': 'Surinam', 'vai': 'ꖬꔸꕯꔈ', 'vai-Latn': 'Surinambe', 'vai-Vaii': 'ꖬꔸꕯꔈ', 'vec': 'Suriname', 'vi': 'Suriname', 'vun': 'Surinamu', 'wae': 'Suriname', 'wo': 'Sirinam', 'xh': 'ESuriname', 'xnr': 'सूरीनाम', 'xog': 'Surinaamu', 'yav': 'sulináam', 'yi': 'סורינאַם', 'yo': 'Surinami', 'yrl': 'Suriname', 'yue': '蘇利南', 'yue-Hans': '苏利南', 'yue-Hant': '蘇利南', 'zgh': 'ⵙⵓⵔⵉⵏⴰⵎ', 'zh': '苏里南', 'zh-Hans': '苏里南', 'zh-Hant': '蘇利南', 'zu': 'i-Suriname'},
    'SS': {'af': 'Suid-Soedan', 'ak': 'Sudan Anaafoɔ', 'am': 'ደቡብ ሱዳን', 'ar': 'جنوب السودان', 'as': 'দক্ষিণ চুডান', 'ast': 'Sudán del Sur', 'az': 'Cənubi Sudan', 'az-Cyrl': 'Ҹәнуби Судан', 'az-Latn': 'Cənubi Sudan', 'bal-Latn': 'Zerbári Sudán', 'be': 'Паўднёвы Судан', 'be-tarask': 'Паўднёвы Судан', 'bg': 'Южен Судан', 'blo': 'Suɖaŋ gʊnyɩsonolaŋ', 'bn': 'দক্ষিণ সুদান', 'br': 'Susoudan', 'brx': 'खोला सुदान', 'bs': 'Južni Sudan', 'bs-Cyrl': 'Јужни Судан', 'bs-Latn': 'Južni Sudan', 'ca': 'Sudan del Sud', 'ca-ES-valencia': 'Sudan del Sud', 'ccp': '𑄘𑄧𑄉𑄨𑄚𑄴 𑄥𑄪𑄘𑄚𑄴', 'ce': 'Къилба Судан', 'ceb': 'South Sudan', 'chr': 'ᏧᎦᎾᏮ ᏑᏕᏂ', 'ckb': 'سوودانی باشوور', 'cs': 'Jižní Súdán', 'cv': 'Кӑнтӑр Судан', 'cy': 'De Swdan', 'da': 'Sydsudan', 'de': 'Südsudan', 'dsb': 'Pódpołdnjowy Sudan', 'dz': 'སཱའུཐ་ སུ་ཌཱན', 'el': 'Νότιο Σουδάν', 'el-polyton': 'Νότιο Σουδάν', 'en': 'South Sudan', 'es': 'Sudán del Sur', 'et': 'Lõuna-Sudaan', 'eu': 'Hego Sudan', 'fa': 'سودان جنوبی', 'ff-Adlm': '𞤅𞤵𞤣𞤢𞥄𞤲 𞤂𞤫𞤧𞤤𞤫𞤴𞤪𞤭', 'fi': 'Etelä-Sudan', 'fil': 'Timog Sudan', 'fo': 'Suðursudan', 'fr': 'Soudan du Sud', 'fy': 'Sûd-Soedan', 'ga': 'an tSúdáin Theas', 'gd': 'Sudàn a Deas', 'gl': 'Sudán do Sur', 'gu': 'દક્ષિણ સુદાન', 'ha': 'Sudan ta Kudu', 'he': 'דרום סודן', 'hi': 'दक्षिण सूडान', 'hi-Latn': 'South Sudan', 'hr': 'Južni Sudan', 'hsb': 'Južny Sudan', 'ht': 'Soudan du Sud', 'hu': 'Dél-Szudán', 'hy': 'Հարավային Սուդան', 'ia': 'Sudan del Sud', 'id': 'Sudan Selatan', 'ig': 'South Sudan', 'is': 'Suður-Súdan', 'it': 'Sud Sudan', 'ja': '南スーダン', 'jv': 'Sudan Kidul', 'ka': 'სამხრეთ სუდანი', 'kaa': 'Қубла Судан', 'kaa-Cyrl': 'Қубла Судан', 'kea': 'Sudon di Sul', 'kgp': 'Sur-Sunáv', 'kk': 'Оңтүстік Судан', 'kk-Cyrl': 'Оңтүстік Судан', 'km': 'ស៊ូដង់\u200bខាង\u200bត្បូង', 'kn': 'ದಕ್ಷಿಣ ಸುಡಾನ್', 'ko': '남수단', 'kok': 'दक्षिण सुडान', 'kok-Deva': 'दक्षिण सुडान', 'ks': 'جنوبی سوڈان', 'ks-Arab': 'جنوبی سوڈان', 'ksh': 'Södsudahn', 'ku': 'Sûdana Başûr', 'kxv': 'dkiṇ sūdan', 'kxv-Deva': 'दॉकिण सूडान', 'kxv-Latn': 'dkiṇ sūdan', 'kxv-Orya': 'ଦକିଣ ସୁଡାନ', 'kxv-Telu': 'దక్షిణ సూడాన్', 'ky': 'Түштүк Судан', 'lb': 'Südsudan', 'lld': 'Sudan dl Süd', 'lo': 'ຊູດານໃຕ້', 'lt': 'Pietų Sudanas', 'lv': 'Dienvidsudāna', 'mai': 'दक्षिण सुडान', 'mi': 'Hūtāne ki te Tonga', 'mk': 'Јужен Судан', 'ml': 'ദക്ഷിണ സുഡാൻ', 'mn': 'Өмнөд Судан', 'mr': 'दक्षिण सुदान', 'ms': 'Sudan Selatan', 'mt': 'is-Sudan t’Isfel', 'my': 'တောင် ဆူဒန်', 'mzn': 'جنوبی سودان', 'nb': 'Sør-Sudan', 'ne': 'दक्षिण सुडान', 'nl': 'Zuid-Soedan', 'nn': 'Sør-Sudan', 'no': 'Sør-Sudan', 'nqo': 'ߥߙߏ߬ߘߎ߮ ߛߎ߬ߘߊ߲߫', 'om': 'Sudaan Kibbaa', 'or': 'ଦକ୍ଷିଣ ସୁଦାନ', 'pa': 'ਦੱਖਣ ਸੁਡਾਨ', 'pa-Guru': 'ਦੱਖਣ ਸੁਡਾਨ', 'pcm': 'Saút Sudan', 'pl': 'Sudan Południowy', 'ps': 'سويلي سوډان', 'pt': 'Sudão do Sul', 'qu': 'Sudán del Sur', 'rm': 'Sudan dal Sid', 'ro': 'Sudanul de Sud', 'ru': 'Южный Судан', 'sc': 'Sudan de su Sud', 'sd': 'ڏکڻ سوڊان', 'sd-Arab': 'ڏکڻ سوڊان', 'se': 'Máttasudan', 'si': 'දකුණු සුඩානය', 'sk': 'Južný Sudán', 'sl': 'Južni Sudan', 'smn': 'Maadâ-Sudan', 'so': 'Koonfur Suudaan', 'sq': 'Sudani i Jugut', 'sr': 'Јужни Судан', 'sr-Cyrl': 'Јужни Судан', 'sr-Latn': 'Južni Sudan', 'sv': 'Sydsudan', 'sw': 'Sudan Kusini', 'syr': 'ܬܝܡܢ ܣܘܕܐܢ', 'ta': 'தெற்கு சூடான்', 'te': 'దక్షిణ సూడాన్', 'tg': 'Судони Ҷанубӣ', 'th': 'ซูดานใต้', 'ti': 'ደቡብ ሱዳን', 'tk': 'Günorta Sudan', 'to': 'Sūtani fakatonga', 'tr': 'Güney Sudan', 'tt': 'Көньяк Судан', 'ug': 'جەنۇبىي سۇدان', 'uk': 'Південний Судан', 'ur': 'جنوبی سوڈان', 'uz': 'Janubiy Sudan', 'uz-Cyrl': 'Жанубий Судан', 'uz-Latn': 'Janubiy Sudan', 'vai': 'ꖬꕜꘋ ꗛꔤ ꔒꘋꗣ ꗏ', 'vai-Vaii': 'ꖬꕜꘋ ꗛꔤ ꔒꘋꗣ ꗏ', 'vec': 'Sud Sudan', 'vi': 'Nam Sudan', 'wo': 'Sudaŋ di Sid', 'xh': 'ESouth Sudan', 'xnr': 'दखण सूडान', 'yi': 'דרום־סודאַן', 'yo': 'Gúúsù Sudan', 'yrl': 'Ausudã Su suí', 'yue': '南蘇丹', 'yue-Hans': '南苏丹', 'yue-Hant': '南蘇丹', 'zgh': 'ⵙⵙⵓⴷⴰⵏ ⵏ ⵉⴼⴼⵓⵙ', 'zh': '南苏丹', 'zh-Hans': '南苏丹', 'zh-Hant': '南蘇丹', 'zu': 'i-South Sudan'},
    'ST': {'af': 'São Tomé en Príncipe', 'agq': 'Sawo Tɔ̀me à Kpèlènsikpɛ̀', 'ak': 'São Tomé ne Príncipe', 'am': 'ሳኦ ቶሜ እና ፕሪንሲፔ', 'ar': 'ساو تومي وبرينسيبي', 'as': 'চাও টোমে আৰু প্ৰিনচিপে', 'asa': 'Thao Tome na Principe', 'ast': 'Santu Tomé y Príncipe', 'az': 'San-Tome və Prinsipi', 'az-Cyrl': 'Сан-Томе вә Принсипи', 'az-Latn': 'San-Tome və Prinsipi', 'bal-Latn': 'Sáó Tóme o Prensip', 'bas': 'Sào Tòme ɓɔ Prɛ̀ŋcipè', 'be': 'Сан-Тамэ і Прынсіпі', 'be-tarask': 'Сан-Тамэ і Прынсіпі', 'bez': 'Husaotome na Huprinsipe', 'bg': 'Сао Томе и Принсипи', 'blo': 'Saotomee', 'bm': 'Sawo Tome-ni-Prinicipe', 'bn': 'সাওটোমা ও প্রিন্সিপি', 'br': 'São Tomé ha Príncipe', 'brx': 'साऊ तमे आरो प्रिन्सिप', 'bs': 'Sao Tome i Principe', 'bs-Cyrl': 'Сао Томе и Принципе', 'bs-Latn': 'Sao Tome i Principe', 'ca': 'São Tomé i Príncipe', 'ca-ES-valencia': 'São Tomé i Príncipe', 'ccp': '𑄥𑄃𑄮𑄑𑄟 𑄃𑄮 𑄛𑄳𑄢𑄨𑄚𑄴𑄥𑄨𑄛𑄨', 'ce': 'Сан-Томе а, Принсипи а', 'ceb': 'Sao Tome & Principe', 'cgg': 'Sawo Tome na Purinsipo', 'chr': 'ᏌᎣ ᏙᎺ ᎠᎴ ᏈᏂᏏᏇ', 'ckb': 'ساوتۆمێ و پرینسیپی', 'cs': 'Svatý Tomáš a Princův ostrov', 'cv': 'Сан-Томе тата Принсипи', 'cy': 'São Tomé a Príncipe', 'da': 'São Tomé og Príncipe', 'dav': 'Sao Tome na Principe', 'de': 'São Tomé und Príncipe', 'dje': 'Sao Tome nda Prinsipe', 'dsb': 'São Tomé a Príncipe', 'dz': 'སཝ་ ཊོ་མེ་ ཨེནཌ་ པྲྀན་སི་པེ', 'ebu': 'Sao Tome na Principe', 'ee': 'São Tomé kple Príncipe nutome', 'el': 'Σάο Τομέ και Πρίνσιπε', 'el-polyton': 'Σάο Τομὲ καὶ Πρίνσιπε', 'en': 'São Tomé & Príncipe', 'en-Dsrt': '𐐝𐐵 𐐓𐐬𐑋 𐐰𐑌𐐼 𐐑𐑉𐐮𐑌𐐽𐐮𐐹𐐩', 'eo': 'Santomeo kaj Principeo', 'es': 'Santo Tomé y Príncipe', 'et': 'São Tomé ja Príncipe', 'eu': 'Sao Tome eta Principe', 'ewo': 'Saó Tomé ai Pəlinəsípe', 'fa': 'سائوتومه و پرینسیپ', 'ff': 'Sawo Tome e Perensipe', 'ff-Adlm': '𞤅𞤢𞤱𞤵 𞤚𞤵𞤥𞤫𞥅 & 𞤆𞤫𞤪𞤫𞤲𞤧𞤭𞤨𞤫', 'ff-Latn': 'Sawo Tome e Perensipe', 'fi': 'São Tomé ja Príncipe', 'fil': 'São Tomé & Príncipe', 'fo': 'Sao Tome & Prinsipi', 'fr': 'Sao Tomé-et-Principe', 'fur': 'Sao Tomè e Principe', 'fy': 'Sao Tomé en Principe', 'ga': 'São Tomé agus Príncipe', 'gd': 'São Tomé agus Príncipe', 'gl': 'San Tomé e Príncipe', 'gsw': 'Sao Tome und Prinssipe', 'gu': 'સાઓ ટૉમ અને પ્રિંસિપે', 'guz': 'Sao Tome na Principe', 'ha': 'Sawo Tome Da Paransip', 'he': 'סאו טומה ופרינסיפה', 'hi': 'साओ टोम और प्रिंसिपे', 'hi-Latn': 'Sao Tome & Principe', 'hr': 'Sveti Toma i Princip', 'hsb': 'São Tomé a Príncipe', 'ht': 'Sao Tomé-et-Principe', 'hu': 'São Tomé és Príncipe', 'hy': 'Սան Տոմե և Փրինսիպի', 'ia': 'São Tomé e Príncipe', 'id': 'Sao Tome dan Principe', 'ig': 'São Tomé & Príncipe', 'is': 'Saó Tóme og Prinsípe', 'it': 'São Tomé e Príncipe', 'ja': 'サントメ・プリンシペ', 'jgo': 'Sáwɔŋ Tɔmɛ́ nɛ́ Pɛlínsipɛ', 'jmc': 'Sao Tome na Principe', 'jv': 'Sao Tomé lan Principé', 'ka': 'სან-ტომე და პრინსიპი', 'kaa': 'Сан-Томе ҳәм Принсипи', 'kaa-Cyrl': 'Сан-Томе ҳәм Принсипи', 'kab': 'Saw Tumi d Pransip', 'kam': 'Sao Tome na Principe', 'kde': 'Saotome na Prinsipe', 'kea': 'San Tume i Prínsipi', 'kgp': 'Sỹ Tomẽ kar Prĩsipi', 'khq': 'Sao Tome nda Prinsipe', 'ki': 'Sao Tome na Principe', 'kk': 'Сан-Томе және Принсипи', 'kk-Cyrl': 'Сан-Томе және Принсипи', 'kln': 'Emetab São Tomé and Príncipe', 'km': 'សៅតូម៉េ និង ប្រាំងស៊ីប', 'kn': 'ಸಾವೋ ಟೋಮ್ ಮತ್ತು ಪ್ರಿನ್ಸಿಪಿ', 'ko': '상투메 프린시페', 'kok': 'साओ टोम आनी प्रिन्सिप', 'kok-Deva': 'साओ टोम आनी प्रिन्सिप', 'ks': 'ساو توم تہٕ پرنسِپی', 'ks-Arab': 'ساو توم تہٕ پرنسِپی', 'ksb': 'Sao Tome na Plincipe', 'ksf': 'saotomɛ́ ri priŋsib', 'ksh': 'Zint Tommeh un Printschipe', 'ku': 'Sao Tome û Prînsîpe', 'kxv': 'sao tom aḍe prinsipe', 'kxv-Deva': 'साओ टोम ऑड़े प्रिंसिपे', 'kxv-Latn': 'sao tom aḍe prinsipe', 'kxv-Orya': 'ସାଅ ଟମ ଅଡ଼େ ପ୍ରିନ୍ସିପେ', 'kxv-Telu': 'సావో టోమ్ మరియు ప్రిన్సిపి', 'ky': 'Сан-Томе жана Принсипи', 'lag': 'Sao Tóome na Pirinsipe', 'lb': 'São Tomé a Príncipe', 'lg': 'Sanitome ne Purincipe', 'lld': 'São Tomé y Príncipe', 'ln': 'Sao Tomé mpé Presipɛ', 'lo': 'ເຊົາທູເມ ແລະ ພຣິນຊິບ', 'lt': 'San Tomė ir Prinsipė', 'lu': 'Sao Tome ne Presipɛ', 'luo': 'São Tomé gi Príncipe', 'luy': 'Sao Tome na Principe', 'lv': 'Santome un Prinsipi', 'mai': 'साओ टोम आ प्रिंसिप', 'mas': 'Sao Tome na Principe', 'mer': 'Sao Tome na Principe', 'mfe': 'São Tome-ek-Prínsip', 'mg': 'São Tomé-et-Príncipe', 'mgh': 'Usao Tome na Principe', 'mi': 'Hato Tomei me Pirinipei', 'mk': 'Саун Томе и Принсип', 'ml': 'സാവോ ടോമും പ്രിൻസിപെയും', 'mn': 'Сан-Томе ба Принсипи', 'mr': 'साओ टोम आणि प्रिंसिपे', 'ms': 'Sao Tome dan Principe', 'mt': 'São Tomé u Príncipe', 'mua': 'Sao Tome ne Principe', 'my': 'ဆောင်တူမေးနှင့် ပရင်စီပီ', 'mzn': 'سائوتومه و پرینسیپ', 'naq': 'São Tomé and Príncipe', 'nb': 'São Tomé og Príncipe', 'nd': 'São Tomé and Príncipe', 'ne': 'साओ टोमे र प्रिन्सिप', 'nl': 'Sao Tomé en Principe', 'nmg': 'Sao Tomé ba Prinship', 'nn': 'São Tomé og Príncipe', 'no': 'São Tomé og Príncipe', 'nqo': 'ߛߊߥߕߏߡߋ߫ ߣߌ߫ ߔߑߙߍ߲ߛߌߔ', 'nyn': 'Sawo Tome na Purinsipo', 'om': 'Sa’oo Toomee fi Prinsippee', 'or': 'ସାଓ ଟୋମେ ଏବଂ ପ୍ରିନସିପି', 'pa': 'ਸਾਓ ਟੋਮ ਅਤੇ ਪ੍ਰਿੰਸੀਪੇ', 'pa-Guru': 'ਸਾਓ ਟੋਮ ਅਤੇ ਪ੍ਰਿੰਸੀਪੇ', 'pcm': 'Sao Tómé & Prínsípẹ', 'pl': 'Wyspy Świętego Tomasza i Książęca', 'ps': 'ساو ټیم او پرنسیپ', 'pt': 'São Tomé e Príncipe', 'qu': 'Santo Tomé y Príncipe', 'rm': 'São Tomé & Príncipe', 'rn': 'Sawotome na Perensipe', 'ro': 'São Tomé și Príncipe', 'rof': 'Sao Tome na Principe', 'ru': 'Сан-Томе и Принсипи', 'rwk': 'Sao Tome na Principe', 'saq': 'Sao Tome na Principe', 'sbp': 'Sayo Tome ni Pilinikipe', 'sc': 'São Tomé e Príncipe', 'sd': 'سائو ٽوم ۽ پرنسپیي', 'sd-Arab': 'سائو ٽوم ۽ پرنسپیي', 'se': 'São Tomé ja Príncipe', 'seh': 'São Tomé e Príncipe', 'ses': 'Sao Tome nda Prinsipe', 'sg': 'Sâô Tömê na Prinsîpe', 'shi': 'ⵙⴰⵡⵟⵓⵎⵉ ⴷ ⴱⵔⴰⵏⵙⵉⴱ', 'shi-Latn': 'sawṭumi d bransib', 'shi-Tfng': 'ⵙⴰⵡⵟⵓⵎⵉ ⴷ ⴱⵔⴰⵏⵙⵉⴱ', 'si': 'සාඕ තෝම් සහ ප්\u200dරින්සිප්', 'sk': 'Svätý Tomáš a Princov ostrov', 'sl': 'Sao Tome in Principe', 'smn': 'São Tomé já Príncipe', 'sn': 'São Tomé and Príncipe', 'so': 'Sao Tome & Birincibal', 'sq': 'Sao-Tome e Principe', 'sr': 'Сао Томе и Принципе', 'sr-Cyrl': 'Сао Томе и Принципе', 'sr-Latn': 'Sao Tome i Principe', 'sv': 'São Tomé och Príncipe', 'sw': 'Sao Tome na Principe', 'syr': 'ܣܐܘ ܛܘܡܝ ܘܦܪܝܢܣܝܦܝ', 'ta': 'சாவ் தோம் & ப்ரின்சிபி', 'te': 'సావో టోమ్ మరియు ప్రిన్సిపి', 'teo': 'Sao Tome na Principe', 'tg': 'Сан Томе ва Принсипи', 'th': 'เซาตูเมและปรินซิปี', 'ti': 'ሳኦ ቶመን ፕሪንሲፐን', 'tk': 'San-Tome we Prinsipi', 'to': 'Sao Tomē mo Pilinisipe', 'tr': 'Sao Tome ve Principe', 'tt': 'Сан-Томе һәм Принсипи', 'twq': 'Sao Tome nda Prinsipe', 'tzm': 'Sawṭumi d Prinsip', 'ug': 'سان تومې ۋە پرىنسىپې', 'uk': 'Сан-Томе і Принсіпі', 'ur': 'ساؤ ٹومے اور پرنسپے', 'uz': 'San-Tome va Prinsipi', 'uz-Cyrl': 'Сан-Томе ва Принсипи', 'uz-Latn': 'San-Tome va Prinsipi', 'vai': 'ꕢꕴ ꕿꔈ ꗪ ꕉ ꕮꔧ ꕗꕴꔀ', 'vai-Latn': 'Sawo Tombe ɓɛ a Gbawoe', 'vai-Vaii': 'ꕢꕴ ꕿꔈ ꗪ ꕉ ꕮꔧ ꕗꕴꔀ', 'vec': 'S. Tomazo e Prìnsipe', 'vi': 'São Tomé và Príncipe', 'vun': 'Sao Tome na Principe', 'wae': 'São Tomé and Príncipe', 'wo': 'Sawo Tome ak Pirinsipe', 'xh': 'ESão Tomé & Príncipe', 'xnr': 'साओ टोम कने प्रिंसिपे', 'xog': 'Sanitome ni Purincipe', 'yav': 'sáwó tomé e pelensípe', 'yi': 'סאַא טאמע און פּרינסיפּע', 'yo': 'Sao tomi ati piriiṣipi', 'yrl': 'Sã Tumé asuí Pirĩsipe', 'yue': '聖多美同普林西比', 'yue-Hans': '圣多美同普林西比', 'yue-Hant': '聖多美同普林西比', 'zgh': 'ⵙⴰⵡⵟⵓⵎⵉ ⴷ ⴱⵔⴰⵏⵙⵉⴱ', 'zh': '圣多美和普林西比', 'zh-Hans': '圣多美和普林西比', 'zh-Hant': '聖多美普林西比', 'zu': 'i-São Tomé kanye ne-Príncipe'},
    'SU': {'en': 'Union of Soviet Socialist Republics'},
    'SV': {'af': 'El Salvador', 'agq': 'Esàvadò', 'ak': 'Ɛl Salvadɔ', 'am': 'ኤል ሳልቫዶር', 'ar': 'السلفادور', 'as': 'এল ছেলভেড’ৰ', 'asa': 'Elsavado', 'ast': 'El Salvador', 'az': 'Salvador', 'az-Cyrl': 'Салвадор', 'az-Latn': 'Salvador', 'bal-Latn': 'El Salwadór', 'bas': 'Sàlvàdɔ̂r', 'be': 'Сальвадор', 'be-tarask': 'Сальвадор', 'bez': 'Huelsavado', 'bg': 'Салвадор', 'blo': 'Ɛl Salfaɖɔɔr', 'bm': 'Salivadɔr', 'bn': 'এল সালভেদর', 'br': 'Salvador', 'brx': 'एल साल्भाड’र', 'bs': 'Salvador', 'bs-Cyrl': 'Салвадор', 'bs-Latn': 'Salvador', 'ca': 'El Salvador', 'ca-ES-valencia': 'El Salvador', 'ccp': '𑄆𑄣𑄴 𑄥𑄣𑄴𑄞𑄬𑄘𑄧𑄢𑄴', 'ce': 'Сальвадор', 'ceb': 'El Salvador', 'cgg': 'Eri Salivado', 'chr': 'ᎡᎵᏌᎵᏆᏙᎵ', 'ckb': 'ئێلسالڤادۆر', 'cs': 'Salvador', 'cv': 'Сальвадор', 'cy': 'El Salvador', 'da': 'El Salvador', 'dav': 'Elsavado', 'de': 'El Salvador', 'dje': 'Salvador laboo', 'dsb': 'El Salvador', 'dyo': 'Salvadoor', 'dz': 'ཨེལ་སལ་བ་ཌོར', 'ebu': 'Elsavado', 'ee': 'El Salvadɔ nutome', 'el': 'Ελ Σαλβαδόρ', 'el-polyton': 'Ἒλ Σαλβαδόρ', 'en': 'El Salvador', 'en-Dsrt': '𐐇𐑊 𐐝𐐰𐑊𐑂𐐲𐐼𐐱𐑉', 'eo': 'Salvadoro', 'es': 'El Salvador', 'et': 'El Salvador', 'eu': 'El Salvador', 'ewo': 'Saləvadór', 'fa': 'السالوادور', 'ff': 'El Salwador', 'ff-Adlm': '𞤉𞤤 𞤅𞤢𞤤𞤾𞤢𞤣𞤮𞥅𞤪', 'ff-Latn': 'El Salwador', 'fi': 'El Salvador', 'fil': 'El Salvador', 'fo': 'El Salvador', 'fr': 'Salvador', 'fur': 'El Salvador', 'fy': 'El Salvador', 'ga': 'An tSalvadóir', 'gd': 'An Salbhador', 'gl': 'O Salvador', 'gsw': 'El Salvador', 'gu': 'એલ સેલ્વાડોર', 'guz': 'Elsavado', 'ha': 'El Salbador', 'he': 'אל סלבדור', 'hi': 'अल सल्वाडोर', 'hi-Latn': 'El Salvador', 'hr': 'Salvador', 'hsb': 'El Salvador', 'ht': 'Salvador', 'hu': 'Salvador', 'hy': 'Սալվադոր', 'ia': 'El Salvador', 'id': 'El Salvador', 'ig': 'El Salvador', 'is': 'El Salvador', 'it': 'El Salvador', 'ja': 'エルサルバドル', 'jmc': 'Elsavado', 'jv': 'Èl Salvador', 'ka': 'სალვადორი', 'kaa': 'Сальвадор', 'kaa-Cyrl': 'Сальвадор', 'kab': 'Salvadur', 'kam': 'Elsavado', 'kde': 'Elsavado', 'kea': 'El Salvador', 'kgp': 'Er Sarvanor', 'khq': 'Salvador laboo', 'ki': 'Elsavado', 'kk': 'Сальвадор', 'kk-Cyrl': 'Сальвадор', 'kln': 'Emetab El Salvador', 'km': 'អែលសាល់វ៉ាឌ័រ', 'kn': 'ಎಲ್ ಸಾಲ್ವೇಡಾರ್', 'ko': '엘살바도르', 'kok': 'एल साल्वाडोर', 'kok-Deva': 'एल साल्वाडोर', 'ks': 'ایل سلویڈر', 'ks-Arab': 'ایل سلویڈر', 'ksb': 'Elsavado', 'ksf': 'salvadɔr', 'ksh': 'Äl Slavadohr', 'ku': 'El Salvador', 'kxv': 'el salvador', 'kxv-Deva': 'अल सल्वाडोर', 'kxv-Latn': 'el salvador', 'kxv-Orya': 'ଅଲ ସଲବାଡର', 'kxv-Telu': 'ఎల్ సాల్వడోర్', 'ky': 'Эль-Сальвадор', 'lag': 'Elisalivado', 'lb': 'El Salvador', 'lg': 'El salivado', 'lld': 'El Salvador', 'ln': 'Savadɔrɛ', 'lo': 'ເອວ ຊໍວາດໍ', 'lt': 'Salvadoras', 'lu': 'Savadore', 'luo': 'El Salvador', 'luy': 'Elsavado', 'lv': 'Salvadora', 'mai': 'अल सल्वाडोर', 'mas': 'Elsavado', 'mer': 'Elsavado', 'mfe': 'Salvador', 'mg': 'El Salvador', 'mgh': 'Usalavado', 'mi': 'Whakaora', 'mk': 'Ел Салвадор', 'ml': 'എൽ സാൽവദോർ', 'mn': 'Эль Сальвадор', 'mr': 'अल साल्वाडोर', 'ms': 'El Salvador', 'mt': 'El Salvador', 'mua': 'Sǝr Salvador', 'my': 'အယ်လ်ဆာဗေးဒိုး', 'mzn': 'السالوادور', 'naq': 'El Salvadori', 'nb': 'El Salvador', 'nd': 'El Salvador', 'ne': 'एल् साल्भाडोर', 'nl': 'El Salvador', 'nmg': 'Salvadɔr', 'nn': 'El Salvador', 'no': 'El Salvador', 'nqo': 'ߛߊߟߑߝ߭ߊߘߐߙ', 'nyn': 'Eri Salivado', 'om': 'El Salvaadoor', 'or': 'ଏଲ୍ ସାଲଭାଡୋର୍', 'pa': 'ਅਲ ਸਲਵਾਡੋਰ', 'pa-Guru': 'ਅਲ ਸਲਵਾਡੋਰ', 'pcm': 'El Sálvádọ', 'pl': 'Salwador', 'ps': 'سالوېډور', 'pt': 'El Salvador', 'qu': 'El Salvador', 'rm': 'El Salvador', 'rn': 'Eli Saluvatori', 'ro': 'El Salvador', 'rof': 'Elsavado', 'ru': 'Сальвадор', 'rwk': 'Elsavado', 'saq': 'Elsavado', 'sbp': 'Elisavado', 'sc': 'El Salvador', 'sd': 'ال سلواڊور', 'sd-Arab': 'ال سلواڊور', 'se': 'El Salvador', 'seh': 'El Salvador', 'ses': 'Salvador laboo', 'sg': 'Salvadöro', 'shi': 'ⵙⴰⵍⴼⴰⴷⵓⵔ', 'shi-Latn': 'salfadur', 'shi-Tfng': 'ⵙⴰⵍⴼⴰⴷⵓⵔ', 'si': 'එල් සැල්වදෝරය', 'sk': 'Salvádor', 'sl': 'Salvador', 'smn': 'El Salvador', 'sn': 'El Salvador', 'so': 'El Salfadoor', 'sq': 'Salvador', 'sr': 'Салвадор', 'sr-Cyrl': 'Салвадор', 'sr-Latn': 'Salvador', 'sv': 'El Salvador', 'sw': 'El Salvador', 'syr': 'ܐܠ ܣܠܒܐܕܘܪ', 'ta': 'எல் சால்வடார்', 'te': 'ఎల్ సాల్వడోర్', 'teo': 'Elsavado', 'tg': 'Эл-Салвадор', 'th': 'เอลซัลวาดอร์', 'ti': 'ኤል ሳልቫዶር', 'tk': 'Salwador', 'to': 'ʻEle Salavatoa', 'tr': 'El Salvador', 'tt': 'Сальвадор', 'twq': 'Salvador laboo', 'tzm': 'Salvaḍur', 'ug': 'سالۋادور', 'uk': 'Сальвадор', 'ur': 'ال سلواڈور', 'uz': 'Salvador', 'uz-Cyrl': 'Салвадор', 'uz-Latn': 'Salvador', 'vai': 'ꗡꗷ ꕢꔍꗍꖺ', 'vai-Latn': 'Ɛlɛ Sávádɔ', 'vai-Vaii': 'ꗡꗷ ꕢꔍꗍꖺ', 'vec': 'El Salvador', 'vi': 'El Salvador', 'vun': 'Elsavado', 'wae': 'El Salvador', 'wo': 'El Salwadoor', 'xh': 'E-El Salvador', 'xnr': 'अल सल्वाडोर', 'xog': 'El salivado', 'yav': 'salfatɔ́ɔl', 'yi': 'על סאַלוואַדאר', 'yo': 'Ẹẹsáfádò', 'yrl': 'Eru Sawadu', 'yue': '薩爾瓦多', 'yue-Hans': '萨尔瓦多', 'yue-Hant': '薩爾瓦多', 'zgh': 'ⵙⴰⵍⴼⴰⴷⵓⵔ', 'zh': '萨尔瓦多', 'zh-Hans': '萨尔瓦多', 'zh-Hant': '薩爾瓦多', 'zu': 'i-El Salvador'},
    'SX': {'af': 'Sint Maarten', 'ak': 'Sint Maaten', 'am': 'ሲንት ማርተን', 'ar': 'سانت مارتن', 'as': 'চিণ্ট মাৰ্টেন', 'ast': 'Sint Maarten', 'az': 'Sint-Marten', 'az-Cyrl': 'Синт-Мартен', 'az-Latn': 'Sint-Marten', 'bal-Latn': 'Sint Márten', 'be': 'Сінт-Мартэн', 'be-tarask': 'Сінт-Мартэн', 'bg': 'Синт Мартен', 'blo': 'Sɛŋ Martɛɛŋ (Holanɖ kaja)', 'bn': 'সিন্ট মার্টেন', 'br': 'Sint Maarten', 'brx': 'सिन्त मार्तेन', 'bs': 'Sint Marten', 'bs-Cyrl': 'Свети Мартин (Холандија)', 'bs-Latn': 'Sint Marten', 'ca': 'Sint Maarten', 'ca-ES-valencia': 'Sint Maarten', 'ccp': '𑄥𑄨𑄚𑄴𑄑𑄴 𑄟𑄢𑄴𑄑𑄬𑄚𑄴', 'ce': 'Синт-Мартен', 'ceb': 'Sint Maarten', 'chr': 'ᏏᏂᏘ ᎹᏘᏂ', 'ckb': 'سینت مارتن', 'cs': 'Svatý Martin (Nizozemsko)', 'cv': 'Синт-Мартен', 'cy': 'Sint Maarten', 'da': 'Sint Maarten', 'de': 'Sint Maarten', 'dsb': 'Sint Maarten', 'dz': 'སིནཊ་ མཱར་ཊེན', 'el': 'Άγιος Μαρτίνος (Ολλανδικό τμήμα)', 'el-polyton': 'Άγιος Μαρτίνος (Ολλανδικό τμήμα)', 'en': 'Sint Maarten', 'es': 'Sint Maarten', 'et': 'Sint Maarten', 'eu': 'Sint Maarten', 'fa': 'سنت مارتن', 'ff-Adlm': '𞤅𞤫𞤲𞤼𞤵 𞤃𞤢𞥄𞤪𞤼𞤫𞤲', 'fi': 'Sint Maarten', 'fil': 'Sint Maarten', 'fo': 'Sint Maarten', 'fr': 'Saint-Martin (partie néerlandaise)', 'fy': 'Sint-Maarten', 'ga': 'Sint Maarten', 'gd': 'Sint Maarten', 'gl': 'Sint Maarten', 'gu': 'સિંટ માર્ટેન', 'ha': 'San Maarten', 'he': 'סנט מארטן', 'hi': 'सिंट मार्टिन', 'hi-Latn': 'Sint Maarten', 'hr': 'Sint Maarten', 'hsb': 'Sint Maarten', 'ht': 'Saint-Martin (partie néerlandaise)', 'hu': 'Sint Maarten', 'hy': 'Սինտ Մարտեն', 'ia': 'Sancte Martino nederlandese', 'id': 'Sint Maarten', 'ie': 'Sint-Maarten', 'ig': 'Sint Maarten', 'is': 'Sint Maarten', 'it': 'Sint Maarten', 'ja': 'シント・マールテン', 'jv': 'Sint Martén', 'ka': 'სინტ-მარტენი', 'kaa': 'Синт-Мартен', 'kaa-Cyrl': 'Синт-Мартен', 'kea': 'San Martinhu (Olanda)', 'kgp': 'Sĩti Mỹ’artẽn', 'kk': 'Синт-Мартен', 'kk-Cyrl': 'Синт-Мартен', 'km': 'សីង\u200bម៉ាធីន', 'kn': 'ಸಿಂಟ್ ಮಾರ್ಟೆನ್', 'ko': '신트마르턴', 'kok': 'सिंट मार्टेन', 'kok-Deva': 'सिंट मार्टेन', 'ks': 'سِنٹ مارٹِن', 'ks-Arab': 'سِنٹ مارٹِن', 'ksh': 'Zint Maarten', 'ku': 'Sint Marteen', 'kxv': 'sint mārṭen', 'kxv-Deva': 'सिंट माऽरतेन', 'kxv-Latn': 'sint mārṭen', 'kxv-Orya': 'ସିଣ୍ଟ ମାର୍ଟିନ', 'kxv-Telu': 'సింట్ మార్టెన్', 'ky': 'Синт-Мартен', 'lb': 'Sint Maarten', 'lld': 'Sint Maarten', 'lo': 'ຊິນ ມາເທັນ', 'lt': 'Sint Martenas', 'lv': 'Sintmārtena', 'mai': 'सिंट मार्टेन', 'mi': 'Hiti Mātene', 'mk': 'Свети Мартин', 'ml': 'സിന്റ് മാർട്ടെൻ', 'mn': 'Синт Мартен', 'mr': 'सिंट मार्टेन', 'ms': 'Sint Maarten', 'mt': 'Sint Maarten', 'my': 'စင့်မာတင်', 'mzn': 'سنت مارتن', 'nb': 'Sint Maarten', 'ne': 'सिन्ट मार्टेन', 'nl': 'Sint-Maarten', 'nn': 'Sint Maarten', 'no': 'Sint Maarten', 'nqo': 'ߛߍ߲ߕ-ߡߊߙߑߕߍ߲߫ ( ߤߏߟߊ߲ߘ ߝߊ߲߭ߝߍ߬ )', 'om': 'Siint Maarteen', 'or': 'ସିଣ୍ଟ ମାର୍ଟୀନ୍\u200c', 'pa': 'ਸਿੰਟ ਮਾਰਟੀਨ', 'pa-Guru': 'ਸਿੰਟ ਮਾਰਟੀਨ', 'pcm': 'Sint Mátin', 'pl': 'Sint Maarten', 'ps': 'سینټ مارټین', 'pt': 'Sint Maarten', 'qu': 'Sint Maarten', 'rm': 'Sint Maarten', 'ro': 'Sint-Maarten', 'ru': 'Синт-Мартен', 'sc': 'Sint Maarten', 'sd': 'سنٽ مارٽن', 'sd-Arab': 'سنٽ مارٽن', 'se': 'Vuolleeatnamat Saint Martin', 'si': 'ශාන්ත මාර්ටෙන්', 'sk': 'Svätý Martin (hol.)', 'sl': 'Sint Maarten', 'smn': 'Sint Maarten', 'so': 'Siint Maarteen', 'sq': 'Sint-Marten', 'sr': 'Свети Мартин (Холандија)', 'sr-Cyrl': 'Свети Мартин (Холандија)', 'sr-Latn': 'Sveti Martin (Holandija)', 'sv': 'Sint Maarten', 'sw': 'Sint Maarten', 'syr': 'ܣܢܬ ܡܐܪܬܝܢ', 'ta': 'சின்ட் மார்டென்', 'te': 'సింట్ మార్టెన్', 'tg': 'Синт-Маартен', 'th': 'ซินต์มาร์เทน', 'ti': 'ሲንት ማርተን', 'tk': 'Sint-Marten', 'to': 'Sā Mātini (fakahōlani)', 'tr': 'Sint Maarten', 'tt': 'Синт-Мартен', 'ug': 'سىنت مارتېن', 'uk': 'Сінт-Мартен', 'ur': 'سنٹ مارٹن', 'uz': 'Sint-Marten', 'uz-Cyrl': 'Синт-Мартен', 'uz-Latn': 'Sint-Marten', 'vai': 'ꔻꘋꔳ ꕮꕊꗳꘋ', 'vai-Vaii': 'ꔻꘋꔳ ꕮꕊꗳꘋ', 'vec': 'S. Martin (Paezi Basi)', 'vi': 'Sint Maarten', 'wo': 'Sin Marten', 'xh': 'ESint Maarten', 'xnr': 'सिंट मार्टिन', 'yo': 'Síntì Mátẹ́ẹ̀nì', 'yrl': 'Sĩti Maatẽ', 'yue': '荷屬聖馬丁', 'yue-Hans': '荷属圣马丁', 'yue-Hant': '荷屬聖馬丁', 'zh': '荷属圣马丁', 'zh-Hans': '荷属圣马丁', 'zh-Hant': '荷屬聖馬丁', 'zu': 'i-Sint Maarten'},
    'SY': {'af': 'Sirië', 'agq': 'Silîa', 'ak': 'Siria', 'am': 'ሶሪያ', 'ar': 'سوريا', 'as': 'চিৰিয়া', 'asa': 'Thiria', 'ast': 'Siria', 'az': 'Suriya', 'az-Cyrl': 'Сурија', 'az-Latn': 'Suriya', 'bal-Latn': 'Suriá', 'bas': 'Sirìà', 'be': 'Сірыя', 'be-tarask': 'Сірыя', 'bez': 'Husilia', 'bg': 'Сирия', 'bgn': 'سوریه', 'blo': 'Sirii', 'bm': 'Siri', 'bn': 'সিরিয়া', 'br': 'Siria', 'brx': 'सीरिया', 'bs': 'Sirija', 'bs-Cyrl': 'Сирија', 'bs-Latn': 'Sirija', 'ca': 'Síria', 'ca-ES-valencia': 'Síria', 'ccp': '𑄥𑄨𑄢𑄨𑄠', 'ce': 'Шема', 'ceb': 'Syria', 'cgg': 'Siriya', 'chr': 'ᏏᎵᎠ', 'ckb': 'سووریا', 'cs': 'Sýrie', 'cv': 'Сири', 'cy': 'Syria', 'da': 'Syrien', 'dav': 'Siria', 'de': 'Syrien', 'dje': 'Suuria', 'dsb': 'Syriska', 'dz': 'སི་རི་ཡ', 'ebu': 'Siria', 'ee': 'Siria nutome', 'el': 'Συρία', 'el-polyton': 'Συρία, Ἀραβικὴ Δημοκρατία τῆς', 'en': 'Syria', 'en-Dsrt': '𐐝𐐮𐑉𐐨𐐲', 'eo': 'Sirio', 'es': 'Siria', 'et': 'Süüria', 'eu': 'Siria', 'ewo': 'Sirí', 'fa': 'سوریه', 'ff': 'Sirii', 'ff-Adlm': '𞤅𞤵𞥅𞤪𞤭𞤴𞤢𞥄', 'ff-Latn': 'Sirii', 'fi': 'Syyria', 'fil': 'Syria', 'fo': 'Sýria', 'fr': 'Syrie', 'fur': 'Sirie', 'fy': 'Syrië', 'ga': 'an tSiria', 'gd': 'Siridhea', 'gl': 'Siria', 'gsw': 'Süürie', 'gu': 'સીરિયા', 'guz': 'Siria', 'ha': 'Sham, Siriya', 'he': 'סוריה', 'hi': 'सीरिया', 'hi-Latn': 'Syria', 'hr': 'Sirija', 'hsb': 'Syriska', 'ht': 'Syrie', 'hu': 'Szíria', 'hy': 'Սիրիա', 'ia': 'Syria', 'id': 'Suriah', 'ig': 'Syria', 'is': 'Sýrland', 'it': 'Siria', 'ja': 'シリア', 'jmc': 'Siria', 'jv': 'Suriah', 'ka': 'სირია', 'kaa': 'Сирия', 'kaa-Cyrl': 'Сирия', 'kab': 'Surya', 'kam': 'Siria', 'kde': 'Silia', 'kea': 'Síria', 'kgp': 'Sirija', 'khq': 'Suuria', 'ki': 'Siria', 'kk': 'Сирия', 'kk-Cyrl': 'Сирия', 'kln': 'Emetab Syria', 'km': 'ស៊ីរី', 'kn': 'ಸಿರಿಯಾ', 'ko': '시리아', 'kok': 'सिरिया', 'kok-Deva': 'सिरिया', 'ks': 'شام', 'ks-Arab': 'شام', 'ksb': 'Silia', 'ksf': 'sirí', 'ksh': 'Sürije', 'ku': 'Sûrîye', 'kxv': 'sīriya', 'kxv-Deva': 'सीरिया', 'kxv-Latn': 'sīriya', 'kxv-Orya': 'ସିରିୟା', 'kxv-Telu': 'సిరియా', 'ky': 'Сирия', 'lag': 'Síria', 'lb': 'Syrien', 'lg': 'Siriya', 'lld': 'Siria', 'ln': 'Sirí', 'lo': 'ຊີເຣຍ', 'lt': 'Sirija', 'lu': 'Siri', 'luo': 'Syria', 'luy': 'Siria', 'lv': 'Sīrija', 'mai': 'सीरिया', 'mas': 'Siria', 'mer': 'Siria', 'mfe': 'Lasiri', 'mg': 'Syria', 'mgh': 'Usiria', 'mi': 'Hiria', 'mk': 'Сирија', 'ml': 'സിറിയ', 'mn': 'Сири', 'mr': 'सीरिया', 'ms': 'Syria', 'mt': 'is-Sirja', 'mua': 'Syria', 'my': 'ဆီးရီးယား', 'mzn': 'سوریه', 'naq': 'Syriab', 'nb': 'Syria', 'nd': 'Syria', 'ne': 'सिरिया', 'nl': 'Syrië', 'nmg': 'Syria', 'nn': 'Syria', 'no': 'Syria', 'nqo': 'ߛߙߌ߫', 'nyn': 'Siriya', 'om': 'Sooriyaa', 'or': 'ସିରିଆ', 'pa': 'ਸੀਰੀਆ', 'pa-Guru': 'ਸੀਰੀਆ', 'pcm': 'Síria', 'pl': 'Syria', 'ps': 'سوریه', 'pt': 'Síria', 'qu': 'Siria', 'rm': 'Siria', 'rn': 'Siriya', 'ro': 'Siria', 'rof': 'Siria', 'ru': 'Сирия', 'rwk': 'Siria', 'saq': 'Siria', 'sbp': 'Siliya', 'sc': 'Sìria', 'sd': 'شام', 'sd-Arab': 'شام', 'se': 'Syria', 'seh': 'Síria', 'ses': 'Suuria', 'sg': 'Sirïi', 'shi': 'ⵙⵓⵔⵢⴰ', 'shi-Latn': 'surya', 'shi-Tfng': 'ⵙⵓⵔⵢⴰ', 'si': 'සිරියාව', 'sk': 'Sýria', 'sl': 'Sirija', 'smn': 'Syria', 'sn': 'Syria', 'so': 'Suuriya', 'sq': 'Siri', 'sr': 'Сирија', 'sr-Cyrl': 'Сирија', 'sr-Latn': 'Sirija', 'sv': 'Syrien', 'sw': 'Syria', 'syr': 'ܣܘܪܝܐ', 'ta': 'சிரியா', 'te': 'సిరియా', 'teo': 'Siria', 'tg': 'Сурия', 'th': 'ซีเรีย', 'ti': 'ሶርያ', 'tk': 'Siriýa', 'to': 'Sīlia', 'tr': 'Suriye', 'tt': 'Сүрия', 'twq': 'Suuria', 'tzm': 'Surya', 'ug': 'سۇرىيە', 'uk': 'Сирія', 'ur': 'شام', 'uz': 'Suriya', 'uz-Cyrl': 'Сурия', 'uz-Latn': 'Suriya', 'vai': 'ꔻꕩꘋ', 'vai-Latn': 'Síyaŋ', 'vai-Vaii': 'ꔻꕩꘋ', 'vec': 'Siria', 'vi': 'Syria', 'vun': 'Siria', 'wae': 'Sürie', 'wo': 'Siri', 'xh': 'ESiriya', 'xnr': 'सीरिया', 'xog': 'Siriya', 'yi': 'סיריע', 'yo': 'Siria', 'yrl': 'Siriya', 'yue': '敘利亞', 'yue-Hans': '叙利亚', 'yue-Hant': '敘利亞', 'zgh': 'ⵙⵓⵔⵢⴰ', 'zh': '叙利亚', 'zh-Hans': '叙利亚', 'zh-Hant': '敘利亞', 'zu': 'i-Syria'},
    'SZ': {'af': 'Eswatini', 'agq': 'Shǔazìlân', 'ak': 'Swaziland', 'am': 'ኤስዋቲኒ', 'ar': 'إسواتيني', 'as': 'ইচ্চুটিনি', 'asa': 'Uthwadhi', 'ast': 'Suazilandia', 'az': 'Esvatini', 'az-Cyrl': 'Свазиленд', 'az-Latn': 'Esvatini', 'bal-Latn': 'Eswatini', 'bas': 'Swàzìlând', 'be': 'Эсватыні', 'be-tarask': 'Эсватыні', 'bez': 'Huuswazi', 'bg': 'Есватини', 'blo': 'Ɛsʊwatinii', 'bm': 'Swazilandi', 'bn': 'ইসওয়াতিনি', 'br': 'Eswatini', 'brx': 'एस्वातिनि', 'bs': 'Esvatini', 'bs-Cyrl': 'Есватини', 'bs-Latn': 'Esvatini', 'ca': 'Eswatini', 'ca-ES-valencia': 'Eswatini', 'ccp': '𑄥𑄮𑄠𑄎𑄨𑄣𑄳𑄠𑄚𑄴𑄓𑄴', 'ce': 'Свазиленд', 'ceb': 'Eswatini', 'cgg': 'Swazirandi', 'chr': 'ᎡᏍᏩᏘᏂ', 'ckb': 'سوازیلاند', 'cs': 'Eswatini', 'cv': 'Эсватини', 'cy': 'Eswatini', 'da': 'Eswatini', 'dav': 'Uswazi', 'de': 'Eswatini', 'dje': 'Swaziland', 'dsb': 'Swasiska', 'dz': 'སུ་ཝ་ཛི་ལེནཌ', 'ebu': 'Uswazi', 'ee': 'Swaziland nutome', 'el': 'Εσουατίνι', 'el-polyton': 'Εσουατίνι', 'en': 'Eswatini', 'en-Dsrt': '𐐝𐐶𐐪𐑆𐐨𐑊𐐰𐑌𐐼', 'eo': 'Svazilando', 'es': 'Esuatini', 'et': 'Svaasimaa', 'eu': 'Swazilandia', 'ewo': 'Swazilándə', 'fa': 'اسواتینی', 'ff': 'Swaasilannda', 'ff-Adlm': '𞤉𞤧𞤱𞤢𞤼𞤭𞤲𞤭', 'ff-Latn': 'Swaasilannda', 'fi': 'Eswatini', 'fil': 'Swaziland', 'fo': 'Esvatini', 'fr': 'Eswatini', 'fur': 'Swaziland', 'fy': 'Swazilân', 'ga': 'eSuaitíní', 'gd': 'eSwatini', 'gl': 'Eswatini', 'gsw': 'Swasiland', 'gu': 'એસ્વાટીની', 'guz': 'Uswazi', 'ha': 'Eswatini', 'he': 'אסוואטיני', 'hi': 'एस्वाटिनी', 'hi-Latn': 'Eswatini', 'hr': 'Esvatini', 'hsb': 'Swaziska', 'ht': 'Eswatini', 'hu': 'Szváziföld', 'hy': 'Էսվատինի', 'ia': 'Eswatini', 'id': 'eSwatini', 'ig': 'Eswatini', 'is': 'Esvatíní', 'it': 'Eswatini', 'ja': 'エスワティニ', 'jgo': 'Swazilân', 'jmc': 'Uswazi', 'jv': 'Swasiland', 'ka': 'სვაზილენდი', 'kaa': 'Эсватини', 'kaa-Cyrl': 'Эсватини', 'kab': 'Swazilund', 'kam': 'Uswazi', 'kde': 'Uswazi', 'kea': 'Eswatini', 'kgp': 'Esuatinĩ', 'khq': 'Swaziland', 'ki': 'Uswazi', 'kk': 'Эсватини', 'kk-Cyrl': 'Эсватини', 'kln': 'Emetab Swaziland', 'km': 'ស្វាស៊ីឡង់', 'kn': 'ಎಸ್ವಾಟಿನಿ', 'ko': '에스와티니', 'kok': 'इस्वातिनी', 'kok-Deva': 'इस्वातिनी', 'ks': 'ایسواتنی', 'ks-Arab': 'ایسواتنی', 'ksb': 'Uswazi', 'ksf': 'swazilan', 'ksh': 'ẞwaasiland', 'ku': 'Eswatînî', 'kxv': 'esvaṭini', 'kxv-Deva': 'एस्वाटिनी', 'kxv-Latn': 'esvaṭini', 'kxv-Orya': 'ଏସ୍ୱାଟିନି', 'kxv-Telu': 'ఈస్వాటిని', 'ky': 'Свазиленд', 'lag': 'Ʉswáazi', 'lb': 'Swasiland', 'lg': 'Swazirandi', 'lld': 'Eswatini', 'ln': 'Swazilandi', 'lo': '\u200bເອ\u200bສະ\u200bວາ\u200bຕິ\u200bນີ', 'lt': 'Svazilandas', 'lu': 'Swazilandi', 'luo': 'Swaziland', 'luy': 'Uswazi', 'lv': 'Svatini', 'mai': 'एस्वाटिनी', 'mas': 'Uswazi', 'mer': 'Swazilandi', 'mfe': 'Swaziland', 'mg': 'Soazilandy', 'mgh': 'Uswazi', 'mi': 'Ehiwatini', 'mk': 'Свазиленд', 'ml': 'സ്വാസിലൻഡ്', 'mn': 'Эсватини', 'mr': 'इस्वातिनी', 'ms': 'Eswatini', 'mt': 'l-Eswatini', 'mua': 'Sǝr Swaziland', 'my': 'ဆွာဇီလန်', 'mzn': 'سوازیلند', 'naq': 'Swazilandi', 'nb': 'Eswatini', 'nd': 'Swaziland', 'ne': 'स्वाजिल्याण्ड', 'nl': 'Eswatini', 'nmg': 'Swaziland', 'nn': 'Eswatini', 'no': 'Eswatini', 'nqo': 'ߒߛߎߥߊߕߣߌ߫', 'nyn': 'Swazirandi', 'om': 'Iswaatinii', 'or': 'ଇସ୍ୱାତିନୀ', 'pa': 'ਇਸਵਾਤੀਨੀ', 'pa-Guru': 'ਇਸਵਾਤੀਨੀ', 'pcm': 'Ẹswatíni', 'pl': 'Eswatini', 'ps': 'اسواټيني', 'pt': 'Essuatíni', 'qu': 'Esuatini', 'rm': 'Eswatini', 'rn': 'Suwazilandi', 'ro': 'Eswatini', 'rof': 'Uswazi', 'ru': 'Эсватини', 'rwk': 'Uswazi', 'saq': 'Uswazi', 'sbp': 'Uswasi', 'sc': 'Eswatini', 'sd': 'ايسواٽني', 'sd-Arab': 'ايسواٽني', 'se': 'Svazieana', 'seh': 'Suazilândia', 'ses': 'Swaziland', 'sg': 'Swäzïlânde', 'shi': 'ⵙⵡⴰⵣⵉⵍⴰⵏⴷⴰ', 'shi-Latn': 'swazilanda', 'shi-Tfng': 'ⵙⵡⴰⵣⵉⵍⴰⵏⴷⴰ', 'si': 'එස්වාටිනි', 'sk': 'Eswatini', 'sl': 'Esvatini', 'smn': 'Swazieennâm', 'sn': 'Swaziland', 'so': 'Eswaatiini', 'sq': 'Esvatini', 'sr': 'Свазиленд', 'sr-Cyrl': 'Свазиленд', 'sr-Latn': 'Svazilend', 'ss': 'eSwatini', 'sv': 'Eswatini', 'sw': 'Eswatini', 'syr': 'ܐܣܘܐܛܝܢܝ', 'ta': 'எஸ்வாட்டீனி', 'te': 'ఈస్వాటిని', 'teo': 'Uswazi', 'tg': 'Эсватини', 'th': 'เอสวาตีนี', 'ti': 'ኤስዋቲኒ', 'tk': 'Eswatini', 'to': 'ʻEsuatini', 'tok': 'ma Sawasi', 'tr': 'Esvatini', 'tt': 'Свазиленд', 'twq': 'Swaziland', 'tzm': 'Swazilanḍa', 'ug': 'سىۋېزىلاند', 'uk': 'Есватіні', 'ur': 'سواتنی', 'uz': 'Svazilend', 'uz-Cyrl': 'Свазиленд', 'uz-Latn': 'Svazilend', 'vai': 'ꖬꕎꔽ ꖨꕮꕊ', 'vai-Latn': 'Suwazi Lumaã', 'vai-Vaii': 'ꖬꕎꔽ ꖨꕮꕊ', 'vec': 'Esuatini', 'vi': 'Eswatini', 'vun': 'Uswazi', 'wae': 'Swasiland', 'wo': 'Suwasilànd', 'xh': 'ESwatini', 'xnr': 'एस्वाटिनी', 'xog': 'Swazirandi', 'yav': 'suasiláand', 'yi': 'סוואַזילאַנד', 'yo': 'Saṣiland', 'yrl': 'Esuatíni', 'yue': '史瓦濟蘭', 'yue-Hans': '史瓦济兰', 'yue-Hant': '史瓦濟蘭', 'zgh': 'ⵙⵡⴰⵣⵉⵍⴰⵏⴷⴰ', 'zh': '斯威士兰', 'zh-Hans': '斯威士兰', 'zh-Hant': '史瓦帝尼', 'zu': 'i-Swaziland'},
    'Samr': {'ast': 'samaritanu', 'az': 'samaritan', 'az-Latn': 'samaritan', 'bal-Latn': 'Samári', 'bg': 'Самаританска', 'bn': 'সমেরিটন', 'br': 'samaritek', 'brx': 'समारती', 'bs': 'samaritansko pismo', 'bs-Cyrl': 'самаританско писмо', 'bs-Latn': 'samaritansko pismo', 'ca': 'samarità', 'ca-ES-valencia': 'samarità', 'ccp': '𑄥𑄧𑄟𑄬𑄢𑄨𑄑𑄧𑄚𑄴', 'cs': 'samařské', 'da': 'samaritansk', 'de': 'Samaritanisch', 'el': 'Σαμαριτικό', 'el-polyton': 'Σαμαριτικό', 'en': 'Samaritan', 'en-Dsrt': '𐐝𐐲𐑋𐐯𐑉𐐲𐐻𐐲𐑌', 'et': 'Samaaria', 'eu': 'samariera', 'fa': 'سامری', 'ff-Adlm': '𞤅𞤢𞤥𞤢𞤪𞤭𞤼𞤢𞤲', 'fi': 'samarianaramealainen', 'fr': 'samaritain', 'fy': 'Samaritaansk', 'ga': 'Samárach', 'gd': 'Samaritanais', 'gsw': 'Samaritanisch', 'gu': 'સમરિટાન', 'he': 'שומרונית', 'hi': 'समरिटन', 'hi-Latn': 'Samaritan', 'hr': 'samaritansko pismo', 'ht': 'samaritain', 'hu': 'Szamaritán', 'id': 'Samaria', 'ig': 'Samaritan', 'it': 'samaritano', 'ja': 'サマリア文字', 'ka': 'სამარიული', 'kgp': 'samỹritỹnũ', 'kk': 'самария жазуы', 'kk-Cyrl': 'самария жазуы', 'kn': 'ಸಮಾರಿಟನ್', 'ko': '사마리아 문자', 'ks': 'سَمارِٹَن', 'ks-Arab': 'سَمارِٹَن', 'lb': 'Samaritanesch', 'lo': 'ຊາມາເລຍ', 'lt': 'samariečių', 'lv': 'samariešu', 'mk': 'самарјанско', 'ml': 'സമരിയ', 'mn': 'Самаритан', 'mr': 'समरिटान', 'ms': 'Samaritan', 'nb': 'samaritansk', 'ne': 'समारिटन', 'nl': 'Samaritaans', 'nn': 'samaritansk', 'no': 'samaritansk', 'nqo': 'ߛߡߊߙߌߕߍ߲', 'or': 'ସମୌରିଟନ୍', 'pl': 'samarytański', 'pt': 'samaritano', 'rm': 'samaritan', 'ro': 'samariteană', 'ru': 'самаритянская', 'sc': 'samaritanu', 'sl': 'samaritanski', 'so': 'Dadka Samaritan', 'sq': 'samaritanisht', 'sr': 'самаританско писмо', 'sr-Cyrl': 'самаританско писмо', 'sr-Latn': 'samaritansko pismo', 'sv': 'samaritiska', 'ta': 'சமாரிடன்', 'te': 'సమారిటన్', 'tg': 'Самаританӣ', 'th': 'ซามาเรีย', 'to': 'tohinima fakasamalitane', 'tr': 'Samarit', 'ug': 'سامارىچە', 'uk': 'самаритянський', 'vec': 'samaritan', 'vi': 'Chữ Samaritan', 'xnr': 'समरिटन', 'yrl': 'samaritanu', 'yue': '撒馬利亞文', 'yue-Hans': '撒马利亚文', 'yue-Hant': '撒馬利亞文', 'zh': '撒马利亚文', 'zh-Hans': '撒马利亚文', 'zh-Hant': '撒馬利亞文', 'zu': 'i-Samaritan'},
    'Sara': {'ar': 'الساراتي', 'ast': 'sarati', 'az': 'sarati', 'az-Latn': 'sarati', 'bg': 'Сарати', 'bn': 'সারাতি', 'brx': 'सराती', 'bs': 'sarati pismo', 'bs-Cyrl': 'сарати писмо', 'bs-Latn': 'sarati pismo', 'ca': 'sarati', 'ca-ES-valencia': 'sarati', 'ccp': '𑄥𑄢𑄖𑄨', 'cs': 'sarati', 'da': 'sarati', 'de': 'Sarati', 'el': 'Σαράθι', 'el-polyton': 'Σαράθι', 'en': 'Sarati', 'en-Dsrt': '𐐝𐐪𐑉𐐪𐐮𐐻𐐨', 'es': 'sarati', 'et': 'sarati', 'fa': 'ساراتی', 'fi': 'sarati', 'fr': 'sarati', 'fy': 'Sarati', 'gd': 'Sarati', 'gsw': 'Sarati', 'gu': 'સરાતી', 'hi': 'सराती', 'hi-Latn': 'Sarati', 'hr': 'sarati pismo', 'ht': 'sarati', 'hu': 'Szarati', 'id': 'Sarati', 'it': 'sarati', 'ja': 'サラティ文字', 'ka': 'სარატი', 'kgp': 'sarati', 'kn': 'ಸರಾಟಿ', 'ko': '사라티', 'ks': 'سَراتی', 'ks-Arab': 'سَراتی', 'lb': 'Sarati', 'lo': 'ຊາຮາຕິ', 'lt': 'sarati', 'mk': 'сарати', 'ml': 'സരതി', 'mr': 'सराती', 'nb': 'sarati', 'ne': 'सारती', 'nl': 'Sarati', 'nn': 'sarati', 'no': 'sarati', 'or': 'ସାରାତି', 'pl': 'sarati', 'pt': 'sarati', 'rm': 'sarati', 'ru': 'сарати', 'sl': 'saratski', 'sr': 'сарати писмо', 'sr-Cyrl': 'сарати писмо', 'sr-Latn': 'sarati pismo', 'sv': 'sarati', 'ta': 'சாராதி', 'te': 'సరాటి', 'th': 'ซาราติ', 'to': 'tohinima fakasalati', 'tr': 'Sarati', 'ug': 'ساراتىچە', 'uk': 'сараті', 'vi': 'Chữ Sarati', 'yrl': 'sarati', 'yue': '沙拉堤文', 'yue-Hans': '沙拉堤文', 'yue-Hant': '沙拉堤文', 'zh': '沙拉堤文', 'zh-Hans': '沙拉堤文', 'zh-Hant': '沙拉堤文'},
    'Sarb': {'ar': 'العربية الجنوبية القديمة', 'ast': 'árabe del sur antiguu', 'bal-Latn': 'Ahdi Zerbári Arabi', 'br': 'henarabek ar Su', 'bs': 'staro južnoarapsko pismo', 'bs-Cyrl': 'старо јужно арапско писмо', 'bs-Latn': 'staro južnoarapsko pismo', 'ca': 'sud-aràbic antic', 'ca-ES-valencia': 'sud-aràbic antic', 'cs': 'starojihoarabské', 'da': 'oldsørarabisk', 'de': 'Altsüdarabisch', 'en': 'Old South Arabian', 'et': 'Lõuna-Araabia', 'eu': 'hegoaldeko arabiera zaharra', 'fa': 'عربی جنوبی باستان', 'ff-Adlm': '𞤙𞤢𞥄𞤥𞤲𞤢𞥄𞤲𞤺𞤫𞤲𞤳𞤮 𞤀𞥄𞤪𞤢𞤦𞤵 𞤑𞤭𞤯𞤭', 'fi': 'muinaiseteläarabialainen', 'fy': 'Ald Sûd-Arabysk', 'ga': 'Sean-Arabach Theas', 'gd': 'Seann-Arabais Dheasach', 'he': 'דרום ערבית', 'hi-Latn': 'Old South Arabian', 'hr': 'staro južnoarapsko pismo', 'id': 'Arab Selatan Kuno', 'ig': 'Old South Arabian', 'it': 'arabo meridionale antico', 'ja': '古代南アラビア文字', 'ka': 'ძველი სამხრეთ-არაბული', 'kk': 'оңтүстік араб жазуы', 'kk-Cyrl': 'оңтүстік араб жазуы', 'ko': '옛 남부 아라비아 문자', 'lo': 'ອາລະເບຍໃຕ້ໂບຮານ', 'lt': 'senoji pietų Arabijos', 'mk': 'старојужноарапско', 'mn': 'Хуучин Өмнөд Араб', 'ms': 'Arab Selatan Lama', 'nb': 'gammelsørarabisk', 'nl': 'Oud Zuid-Arabisch', 'nn': 'gammelsørarabisk', 'no': 'gammelsørarabisk', 'nqo': 'ߥߙߏ߬ߘߎ߮ ߊߙߊߓߎ߫ ߞߘߐ߬ߡߊ߲', 'rm': 'arab vegl dal sid', 'ro': 'arabă veche din sud', 'ru': 'староюжноарабская', 'sc': 'àrabu meridionale antigu', 'sd': 'قديم ڏاکڻي عربي', 'sd-Arab': 'قديم ڏاکڻي عربي', 'so': 'Crabiyaankii Hore ee Wuqooyi', 'sq': 'arabishte jugore e vjetër', 'sv': 'fornsydarabiska', 'tg': 'Арабии Ҷанубии Куҳна', 'th': 'อาระเบียใต้โบราณ', 'to': 'tohinima fakaʻalepea-tonga-motuʻa', 'tr': 'Eski Güney Arap', 'ug': 'قەدىمكى جەنۇبى ئەرەبچە', 'vec': 'àrabo antigo de’l sud', 'vi': 'Chữ Nam Ả Rập cổ', 'yue': '古南阿拉伯文', 'yue-Hans': '古南阿拉伯文', 'yue-Hant': '古南阿拉伯文', 'zh': '古南阿拉伯文', 'zh-Hans': '古南阿拉伯文', 'zh-Hant': '古南阿拉伯文', 'zu': 'i-Old South Arabian'},
    'Saur': {'ast': 'saurashtra', 'az': 'saurastra', 'az-Latn': 'saurastra', 'bal-Latn': 'Sáuráshtri', 'bg': 'Саураштра', 'bn': 'সৌরাষ্ট্র', 'brx': 'सौराष्ट्र', 'bs': 'sauraštra pismo', 'bs-Cyrl': 'саураштра писмо', 'bs-Latn': 'sauraštra pismo', 'ca': 'saurashtra', 'ca-ES-valencia': 'saurashtra', 'ccp': '𑄥𑄯𑄢𑄌𑄴𑄑𑄳𑄢𑄧', 'cs': 'saurášterské', 'da': 'saurashtra', 'de': 'Saurashtra', 'el': 'Σαουράστρα', 'el-polyton': 'Σαουράστρα', 'en': 'Saurashtra', 'en-Dsrt': '𐐝𐐰𐐭𐑉𐐪𐑇𐐻𐑉𐐪', 'es': 'saurashtra', 'et': 'sauraštra', 'eu': 'saurashtra', 'fa': 'سوراشترایی', 'ff-Adlm': '𞤅𞤢𞤵𞤪𞤢𞥃𞤼𞤪𞤢', 'fi': 'saurashtra', 'fo': 'saurashtra', 'fr': 'saurashtra', 'fy': 'Saurashtra', 'ga': 'Saurashtra', 'gd': 'Saurashtra', 'gsw': 'Saurashtra', 'gu': 'સૌરાષ્ટ્ર', 'he': 'סאוראשטרה', 'hi': 'सौराष्ट्र', 'hi-Latn': 'Saurashtra', 'hr': 'saurashtra pismo', 'ht': 'saurashtra', 'hu': 'Szaurastra', 'id': 'Saurashtra', 'ig': 'Saurashtra', 'it': 'saurashtra', 'ja': 'サウラーシュトラ文字', 'ka': 'საურაშტრა', 'kgp': 'savurajtera', 'kk': 'саураштра жазуы', 'kk-Cyrl': 'саураштра жазуы', 'km': 'សៅរ៉ាសត្រា', 'kn': 'ಸೌರಾಷ್ಟ್ರ', 'ko': '사우라슈트라 문자', 'ks': 'سوراشٹرا', 'ks-Arab': 'سوراشٹرا', 'kxv': 'saurastra', 'kxv-Deva': 'सॉउराष्ट्र', 'kxv-Latn': 'saurastra', 'kxv-Orya': 'ସଉରାଷ୍ଟ୍ର', 'kxv-Telu': 'సౌరాస్ట్ర', 'lb': 'Saurashtra', 'lo': 'ໂສຮັດຕຣະ', 'lt': 'sauraštra', 'mk': 'саураштранско', 'ml': 'സൗരാഷ്ട്ര', 'mn': 'Саураштра', 'mr': 'सौराष्ट्र', 'ms': 'Saurashtra', 'nb': 'saurashtra', 'ne': 'सौराष्ट्र', 'nl': 'Saurashtra', 'nn': 'saurashtra', 'no': 'saurashtra', 'nqo': 'ߛߏߙߊߗߌߙߊ', 'or': 'ସୌରାଷ୍ଟ୍ର', 'pl': 'saurashtra', 'pt': 'saurashtra', 'rm': 'saurashtra', 'ro': 'saurashtra', 'ru': 'саураштра', 'sc': 'saurashtra', 'so': 'Sawrashtra', 'sq': 'saurashtra', 'sr': 'саураштра писмо', 'sr-Cyrl': 'саураштра писмо', 'sr-Latn': 'sauraštra pismo', 'sv': 'saurashtra', 'ta': 'சௌராஷ்ட்ரா', 'te': 'సౌరాష్ట్ర', 'tg': 'Саураштра', 'th': 'โสวรัสตระ', 'to': 'tohinima fakasaulasitā', 'tr': 'Saurashtra', 'ug': 'سائۇراشتىراچە', 'uk': 'саураштра', 'vec': 'saurashtra', 'vi': 'Chữ Saurashtra', 'xnr': 'सौरासटर', 'yo': 'Èdè Saurasitira', 'yrl': 'sauraxitara', 'yue': '索拉什特拉文', 'yue-Hans': '索拉什特拉文', 'yue-Hant': '索拉什特拉文', 'zh': '索拉什特拉文', 'zh-Hans': '索拉什特拉文', 'zh-Hant': '索拉什特拉文', 'zu': 'isi-Saurashtra'},
    'Sgnw': {'ast': 'escritura de signos', 'az': 'işarət yazısı', 'az-Latn': 'işarət yazısı', 'bal-Latn': 'Neshánnebisi', 'bn': 'চিহ্ন লিখন', 'brx': 'सांकेतिक लेख', 'bs': 'znakovno pismo', 'bs-Cyrl': 'знаковно писмо', 'bs-Latn': 'znakovno pismo', 'ca': 'escriptura de signes', 'ca-ES-valencia': 'escriptura de signes', 'ccp': '𑄌𑄨𑄚𑄴𑄦𑄧 𑄣𑄨𑄈𑄧𑄚𑄴', 'cs': 'SignWriting', 'da': 'tegnskrift', 'de': 'Gebärdensprache', 'el': 'Νοηματική γραφή', 'el-polyton': 'Νοηματική γραφή', 'en': 'SignWriting', 'en-Dsrt': '𐐝𐐴𐑌 𐐡𐐴𐐻𐐨𐑍', 'es': 'SignWriting', 'et': 'viipekiri', 'eu': 'zeinu-idazketa', 'ff-Adlm': '𞤄𞤭𞤲𞤣𞤭 𞤊𞤭𞤲𞤣𞤫', 'fi': 'SignWriting', 'fr': 'écriture des signes', 'fy': 'SignWriting', 'ga': 'Litritheoireacht Comharthaí', 'gd': 'Sgrìobhadh cainnte-sanais', 'gsw': 'Gebäärdeschpraach', 'gu': 'સંકેત લિપી', 'he': 'חתימה', 'hi': 'सांकेतिक लेख', 'hi-Latn': 'SignWriting', 'hr': 'znakovno pismo', 'ht': 'écriture des signes', 'hu': 'Jelírás', 'id': 'Tulisan Isyarat', 'ig': 'SignWriting', 'it': 'linguaggio dei segni', 'ja': '手話文字', 'ka': 'ჟესტთა', 'kgp': 'vẽnh-mu vẽnhrád', 'kk': 'жазу', 'kk-Cyrl': 'жазу', 'kn': 'ಸೈನ್\u200cರೈಟಿಂಗ್', 'ko': '수화 문자', 'ks': 'اِشارٲتی لِکھٲے', 'ks-Arab': 'اِشارٲتی لِکھٲے', 'lb': 'Zeechesprooch', 'lo': 'ໄຊນ໌ໄຮຕີ້ງ', 'lt': 'ženklų raštas', 'mk': 'знаковно пишување', 'ml': 'ചിഹ്നലിപി', 'mn': 'Тэмдэгт бичиг', 'mr': 'संकेत लिपी', 'ms': 'Tulisan Isyarat', 'nb': 'tegnskrift', 'ne': 'साइनराइटिङ', 'nl': 'SignWriting', 'nn': 'teiknskrift', 'no': 'tegnskrift', 'nqo': 'ߞߐߙߍߦߊ߫-ߛߓߍߟߌ ߟߎ߬', 'or': 'ସାଙ୍କେତିକ ଲିଖ', 'pl': 'pismo znakowe', 'pt': 'signwriting', 'rm': 'lingua da segns', 'ro': 'scrierea simbolică', 'ru': 'язык знаков', 'sc': 'limba de sos sinnos', 'sl': 'znakovna pisava', 'so': 'Qaabka dhagoolka loola hadlo', 'sq': 'shkrim sing', 'sr': 'знаковно писмо', 'sr-Cyrl': 'знаковно писмо', 'sr-Latn': 'znakovno pismo', 'sv': 'teckningsskrift', 'ta': 'ஸைன்எழுத்து', 'te': 'సంజ్ఞ లిపి', 'tg': 'Аломатнависӣ', 'th': 'ไซน์ไรติ้ง', 'to': 'tohinima fakaʻilonga-tohi', 'tr': 'İşaret Dili', 'ug': 'ئىشارەت تىلى', 'uk': 'знаковий', 'vec': 'scritura de i segni', 'vi': 'Chữ viết Ký hiệu', 'xnr': 'सांकेतिक लेख', 'yrl': 'sãgawa kuatiara', 'yue': '手語書寫符號', 'yue-Hans': '手语书写符号', 'yue-Hant': '手語書寫符號', 'zh': '书写符号', 'zh-Hans': '书写符号', 'zh-Hant': '手語書寫符號', 'zu': 'i-SignWriting'},
    'Shaw': {'am': 'የሻቪያ ፊደል', 'ar': 'الشواني', 'ast': 'shavianu', 'az': 'şavyan', 'az-Latn': 'şavyan', 'bal-Latn': 'Sháwi', 'bn': 'সাভিয়ান', 'brx': 'शेवियन', 'bs': 'šavian pismo', 'bs-Cyrl': 'шавијанско писмо', 'bs-Latn': 'šavian pismo', 'ca': 'shavià', 'ca-ES-valencia': 'shavià', 'ccp': '𑄥𑄞𑄨𑄠𑄚𑄴', 'cs': 'Shawova abeceda', 'da': 'shavisk', 'de': 'Shaw-Alphabet', 'el': 'Σαβιανό', 'el-polyton': 'Σαβιανό', 'en': 'Shavian', 'en-Dsrt': '𐐟𐐩𐑂𐐨𐐲𐑌', 'es': 'shaviano', 'et': 'Shaw’ kiri', 'eu': 'shaviera', 'fa': 'شاوی', 'ff-Adlm': '𞤅𞤢𞤬𞤭𞤴𞤢𞥄𞤲', 'fi': 'shaw’lainen', 'fr': 'shavien', 'fy': 'Shavian', 'ga': 'Shawach', 'gd': 'Sgrìobhadh an t-Seathaich', 'gsw': 'Shaw-Alphabet', 'gu': 'શાવિયાન', 'he': 'שבית', 'hi': 'शावियान', 'hi-Latn': 'Shavian', 'hr': 'shavian pismo', 'ht': 'shavien', 'hu': 'Shaw ábécé', 'id': 'Shavia', 'ig': 'Shavian', 'it': 'shaviano', 'ja': 'ショー文字', 'kgp': 'savijỹnũ', 'kk': 'шоу жазуы', 'kk-Cyrl': 'шоу жазуы', 'kn': 'ಶಾವಿಯಾನ್', 'ko': '샤비안 문자', 'ks': 'شاویَن', 'ks-Arab': 'شاویَن', 'lb': 'Shaw-Alphabet', 'lo': 'ຊອວຽນ', 'lt': 'šavių', 'mk': 'Шоово', 'ml': 'ഷാവിയൻ', 'mn': 'Шави', 'mr': 'शॅव्हियन', 'ms': 'Shavia', 'nb': 'shavisk', 'ne': 'शाभियन', 'nl': 'Shavian', 'nn': 'shavisk', 'no': 'shavisk', 'nqo': 'ߛߊߝ߭ߌ', 'or': 'ସାବିୟାନ୍', 'pl': 'shawa', 'pt': 'shaviano', 'rm': 'shavian', 'ro': 'savă', 'ru': 'алфавит Шоу', 'sc': 'shavianu', 'sl': 'šojevski', 'so': 'calaamad qoris', 'sq': 'shavianisht', 'sr': 'шавијанско писмо', 'sr-Cyrl': 'шавијанско писмо', 'sr-Latn': 'šavijansko pismo', 'sv': 'shawiska', 'ta': 'ஷவியான்', 'te': 'షవియాన్', 'tg': 'Шавианӣ', 'th': 'ซอเวียน', 'to': 'tohinima fakasiavi', 'tr': 'Shavian', 'ug': 'شاۋىيانچە', 'uk': 'шоу', 'vec': 'shavian', 'vi': 'Chữ Shavian', 'xnr': 'शावियान', 'yrl': 'xawianu', 'yue': '簫柏納字符', 'yue-Hans': '箫柏纳字符', 'yue-Hant': '簫柏納字符', 'zh': '萧伯纳式文', 'zh-Hans': '萧伯纳式文', 'zh-Hant': '簫柏納字符', 'zu': 'i-Shavian'},
    'Shrd': {'ast': 'sharada', 'bal-Latn': 'Sharadá', 'bs': 'sharada pismo', 'bs-Cyrl': 'шарада писмо', 'bs-Latn': 'sharada pismo', 'ca': 'shrada', 'ca-ES-valencia': 'shrada', 'cs': 'šáradá', 'da': 'sharada', 'de': 'Sharada', 'en': 'Sharada', 'et': 'šaarada', 'eu': 'sharada', 'ff-Adlm': '𞤡𞤢𞤪𞤢𞤣𞤢', 'fi': 'šarada', 'fy': 'Sharada', 'ga': 'Sharada', 'gd': 'Sharada', 'he': 'שרדה', 'hi-Latn': 'Sharada', 'hr': 'sharada pismo', 'id': 'Sharada', 'ig': 'Sharada', 'it': 'sharada', 'ja': 'シャーラダー文字', 'ka': 'შარადა', 'kk': 'шарада жазуы', 'kk-Cyrl': 'шарада жазуы', 'ko': '사라다 문자', 'lo': 'ຊາຮາດາ', 'lt': 'Šarados', 'mk': 'шарада', 'mn': 'Шарада', 'ms': 'Sharada', 'nb': 'sharada', 'nl': 'Sharada', 'nn': 'sharada', 'no': 'sharada', 'nqo': 'ߛߙߊߘߊ', 'rm': 'sharada', 'ro': 'sharadă', 'ru': 'шарада', 'sc': 'sharada', 'so': 'Sharada', 'sq': 'sharadisht', 'sv': 'sharada', 'tg': 'Шарада', 'th': 'ชาราดา', 'to': 'tohinima fakasiālatā', 'tr': 'Sharada', 'ug': 'شاراداچە', 'vec': 'sharada', 'vi': 'Chữ Sharada', 'yue': '夏拉達文', 'yue-Hans': '夏拉达文', 'yue-Hant': '夏拉達文', 'zh': '夏拉达文', 'zh-Hans': '夏拉达文', 'zh-Hant': '夏拉達文', 'zu': 'i-Sharada'},
    'Shui': {'en': 'Shuishu'},
    'Sidd': {'ast': 'siddham', 'bal-Latn': 'Seddam', 'bs': 'sidam pismo', 'bs-Cyrl': 'сидхам писмо', 'bs-Latn': 'sidam pismo', 'ca': 'siddham', 'ca-ES-valencia': 'siddham', 'cs': 'siddham', 'de': 'Siddham', 'en': 'Siddham', 'et': 'siddhami', 'eu': 'siddham', 'ff-Adlm': '𞤅𞤭𞤣𞥆𞤢𞥄𞤥', 'fi': 'siddham-tavukirjoitus', 'fo': 'siddham', 'ga': 'Siddham', 'gd': 'Siddham', 'he': 'סידהם', 'hi-Latn': 'Siddham', 'hu': 'Sziddham', 'id': 'Siddham', 'ig': 'Siddham', 'ja': '梵字', 'kk': 'сиддхам жазуы', 'kk-Cyrl': 'сиддхам жазуы', 'ko': '실담자', 'lt': 'Siddham', 'mk': 'сидам', 'mn': 'Сиддхам', 'ms': 'Siddham', 'nb': 'siddham', 'nl': 'Siddham', 'nn': 'siddham', 'no': 'siddham', 'nqo': 'ߛߌߘߊߡ', 'ro': 'siddham', 'sc': 'siddham', 'so': 'Siddham', 'sq': 'sidamisht', 'sv': 'siddhamska', 'tg': 'Сиддам', 'th': 'สิทธัม', 'to': 'tohinima fakasititami', 'tr': 'Siddham', 'vec': 'sidam', 'yue': '悉曇文字', 'yue-Hans': '悉昙文字', 'yue-Hant': '悉曇文字', 'zh': '悉昙文', 'zh-Hans': '悉昙文', 'zh-Hant': '悉曇文字', 'zu': 'i-Siddham'},
    'Sind': {'ast': 'khudabadi', 'bal-Latn': 'Kodáwadi', 'bs': 'khudawadi pismo', 'bs-Cyrl': 'кудавади писмо', 'bs-Latn': 'khudawadi pismo', 'ca': 'devangari', 'ca-ES-valencia': 'devangari', 'cs': 'chudábádí', 'da': 'khudawadi', 'de': 'Khudawadi', 'en': 'Khudawadi', 'et': 'hudavadi', 'eu': 'khudawadi', 'ff-Adlm': '𞤑𞤵𞤣𞤢𞤱𞤢𞤣𞤭', 'fi': 'khudabadi', 'fy': 'Sindhi', 'ga': 'Khudawadi', 'gd': 'Khudawadi', 'he': 'חודוואדי', 'hi-Latn': 'Khudawadi', 'hr': 'khudawadi pismo', 'id': 'Khudawadi', 'ig': 'Khudawadi', 'it': 'khudawadi', 'ja': 'クダワディ文字', 'ka': 'ქუდავადი', 'kk': 'кхудавади жазуы', 'kk-Cyrl': 'кхудавади жазуы', 'ko': '쿠다와디 문자', 'lo': 'ດຸດາວາດີ', 'lt': 'Khudawadi', 'mk': 'кудабади', 'mn': 'Кудавади', 'ms': 'Khudawadi', 'nb': 'khudawadi', 'nl': 'Sindhi', 'nn': 'khudawadi', 'no': 'khudawadi', 'nqo': 'ߛߌ߲ߘߌ', 'ro': 'khudawadi', 'ru': 'кхудавади', 'sc': 'khudawadi', 'so': 'khudwadi', 'sq': 'kudavadisht', 'sv': 'sindhiska', 'tg': 'Худовадӣ', 'th': 'คุดาวาดี', 'to': 'tohinima fakakutauāti', 'tr': 'Khudabadi', 'ug': 'كۇداۋادچە', 'vec': 'kudawadi', 'vi': 'Chữ Khudawadi', 'yue': '信德文', 'yue-Hans': '信德文', 'yue-Hant': '信德文', 'zh': '信德文', 'zh-Hans': '信德文', 'zh-Hant': '信德文', 'zu': 'i-Khudawadi'},
    'Sinh': {'af': 'Sinhala', 'ak': 'Sinhala kasa', 'am': 'ሲንሃላ', 'ar': 'السينهالا', 'as': 'সিংহলী', 'ast': 'cingalés', 'az': 'sinhal', 'az-Latn': 'sinhal', 'bal-Latn': 'Senhalá', 'be': 'сінгальскае', 'be-tarask': 'сінгальскае', 'bg': 'синхалска', 'bn': 'সিংহলি', 'br': 'singhalek', 'brx': 'सिंहाली', 'bs': 'pismo sinhala', 'bs-Cyrl': 'синхалско писмо', 'bs-Latn': 'pismo sinhala', 'ca': 'singalès', 'ca-ES-valencia': 'singalès', 'ccp': '𑄥𑄨𑄁𑄦𑄧𑄣𑄨', 'ce': 'сингалхойн', 'chr': 'ᏏᏅᎭᎳ', 'ckb': 'سینھالا', 'cs': 'sinhálské', 'cy': 'Sinhanaidd', 'da': 'singalesisk', 'de': 'Singhalesisch', 'dsb': 'singhaleski', 'dz': 'སིན་ཧ་ལ་རིག་གུ', 'ee': 'sinhalagbeŋɔŋlɔ', 'el': 'Σινχάλα', 'el-polyton': 'Σινχάλα', 'en': 'Sinhala', 'en-Dsrt': '𐐝𐐮𐑌𐐸𐐪𐑊𐐲', 'es': 'cingalés', 'et': 'singali', 'eu': 'sinhala', 'fa': 'سینهالی', 'ff-Adlm': '𞤅𞤭𞤲𞤸𞤢𞤤𞤢', 'fi': 'sinhalilainen', 'fil': 'Sinhala', 'fo': 'sinhala', 'fr': 'cingalais', 'fur': 'sinhala', 'fy': 'Sinhala', 'ga': 'Siolónach', 'gd': 'Sinhala', 'gl': 'cingalés', 'gsw': 'Singhalesisch', 'gu': 'સિંહલી', 'ha': 'Yaren Sinhala', 'he': 'סינהלה', 'hi': 'सिंहली', 'hi-Latn': 'Sinhala', 'hr': 'sinhaleško pismo', 'hsb': 'singhalsce', 'ht': 'cingalais', 'hu': 'Szingaléz', 'hy': 'սինհալական', 'ia': 'cingalese', 'id': 'Sinhala', 'ig': 'Sinhala', 'is': 'sinhala', 'it': 'singalese', 'ja': 'シンハラ文字', 'jv': 'Sinhala', 'ka': 'სინჰალური', 'kea': 'singales', 'kgp': 'sĩgarẽj', 'kk': 'сингаль жазуы', 'kk-Cyrl': 'сингаль жазуы', 'km': 'ស៊ីនហាឡា', 'kn': 'ಸಿಂಹಳ', 'ko': '신할라 문자', 'kok': 'सिंहाला', 'kok-Deva': 'सिंहाला', 'ks': 'سِنہالا', 'ks-Arab': 'سِنہالا', 'ksh': 'singjaleesesche Schreff', 'ku': 'sînhalayî', 'ky': 'Сингала', 'lb': 'Singhalesesch', 'lo': 'ສິນຫາລາ', 'lrc': 'سیناھالا', 'lt': 'sinhalų', 'lv': 'singāļu', 'mai': 'सिंहला', 'mi': 'Hināra', 'mk': 'синхалско писмо', 'ml': 'സിംഹള', 'mn': 'синхала', 'mr': 'सिंहला', 'ms': 'Sinhala', 'my': 'ဆင်ဟာလ', 'nb': 'singalesisk', 'ne': 'सिन्हाला', 'nl': 'Singalees', 'nn': 'singalesisk', 'no': 'singalesisk', 'nqo': 'ߛߌ߲ߜ߭ߟߊ', 'or': 'ସିଂହଳ', 'pa': 'ਸਿੰਹਾਲਾ', 'pa-Guru': 'ਸਿੰਹਾਲਾ', 'pcm': 'Sinhála', 'pl': 'syngaleskie', 'ps': 'سنهالا', 'pt': 'cingalês', 'qu': 'Cingales Simi', 'rm': 'singalais', 'ro': 'singaleză', 'ru': 'сингальская', 'sc': 'singalesu', 'sd': 'سنهالا', 'sd-Arab': 'سنهالا', 'si': 'සිංහල', 'sk': 'sinhálske', 'sl': 'sinhalski', 'so': 'Sinhaala', 'sq': 'sinhal', 'sr': 'синхалско писмо', 'sr-Cyrl': 'синхалско писмо', 'sr-Latn': 'sinhalsko pismo', 'sv': 'singalesiska', 'sw': 'Kisinhala', 'ta': 'சிங்களம்', 'te': 'సింహళం', 'tg': 'Синхала', 'th': 'สิงหล', 'ti': 'ሲንሃላ', 'tk': 'Singal elipbiýi', 'to': 'tohinima fakasingihala', 'tr': 'Seylan', 'ug': 'سىنخالاچە', 'uk': 'сингальська', 'ur': 'سنہالا', 'uz': 'singal', 'uz-Cyrl': 'Синхала', 'uz-Latn': 'singal', 'vec': 'singaleze', 'vi': 'Chữ Sinhala', 'wae': 'Singalesiš', 'xnr': 'सिंहली', 'yo': 'Sìnhálà', 'yrl': 'sĩgarei', 'yue': '錫蘭文', 'yue-Hans': '锡兰文', 'yue-Hant': '錫蘭文', 'zh': '僧伽罗文', 'zh-Hans': '僧伽罗文', 'zh-Hant': '錫蘭文', 'zu': 'isi-Sinhala'},
    'Sogd': {'bal-Latn': 'Sógdiái', 'bs': 'sogdian psmo', 'bs-Cyrl': 'согдиан писмо', 'bs-Latn': 'sogdian psmo', 'en': 'Sogdian', 'et': 'sogdi', 'eu': 'sogdiera', 'ff-Adlm': '𞤅𞤮𞤺𞤮𞤣𞤭𞤴𞤢𞤲', 'fi': 'sogdialainen', 'fo': 'sogdian', 'gd': 'Sogdianais', 'he': 'סוגדית', 'hi-Latn': 'Sogdian', 'hu': 'Szogd', 'id': 'Sogdian', 'ig': 'Sogdian', 'kk': 'соғды жазуы', 'kk-Cyrl': 'соғды жазуы', 'mn': 'Согди', 'ms': 'Sogdia', 'nl': 'Sogdisch', 'nqo': 'ߛߏߜ߭ߑߘߌ', 'ro': 'sogdiană', 'sc': 'sogdianu', 'so': 'Sogdiyaan', 'sq': 'sogdianisht', 'sv': 'sogdiska', 'tg': 'Суғдӣ', 'vec': 'sogdian', 'zh': '粟特文', 'zh-Hans': '粟特文', 'zh-Hant': '粟特文', 'zu': 'i-Sogdian'},
    'Sogo': {'bal-Latn': 'Ahdi Sógdiái', 'bs': 'staro sogdian pismo', 'bs-Cyrl': 'старо согдиан писмо', 'bs-Latn': 'staro sogdian pismo', 'en': 'Old Sogdian', 'et': 'vanasogdi', 'eu': 'sogdiera zaharra', 'ff-Adlm': '𞤅𞤮𞤺𞤮𞤣𞤭𞤴𞤢𞤲 𞤑𞤭𞤯𞥆𞤵𞤲', 'fi': 'muinaissogdialainen', 'fo': 'gamalt sogdian', 'gd': 'Seann-Sogdianais', 'he': 'סוגדית עתיקה', 'hi-Latn': 'Old Sogdian', 'hu': 'Ószogd', 'id': 'Sogdian Lama', 'ig': 'Old Sogdian', 'kk': 'көне соғды жазуы', 'kk-Cyrl': 'көне соғды жазуы', 'mn': 'Хуучин согди', 'ms': 'Sogdia Lama', 'nl': 'Oud Sogdisch', 'nqo': 'ߛߏߜ߭ߑߘߌ ߞߘߐ߬ߡߊ߲', 'ro': 'sogdiană veche', 'sc': 'sogdianu antigu', 'so': 'Sogdiyaankii Hore', 'sq': 'sogdianishte e vjetër', 'sv': 'gammalsogdiska', 'tg': 'Суғдии Куҳна', 'vec': 'sogdian antigo', 'zh': '古粟特文', 'zh-Hans': '古粟特文', 'zh-Hant': '古粟特文', 'zu': 'i-Old Sogdian'},
    'Sora': {'ast': 'sora sompeng', 'bal-Latn': 'Surá Sompeng', 'bs': 'sora sompeng pismo', 'bs-Cyrl': 'сора сомпенг писмо', 'bs-Latn': 'sora sompeng pismo', 'ca': 'sora sompeng', 'ca-ES-valencia': 'sora sompeng', 'cs': 'sora sompeng', 'de': 'Sora Sompeng', 'en': 'Sora Sompeng', 'eu': 'sora sompeng', 'ff-Adlm': '𞤅𞤢𞤪𞤢 𞤅𞤮𞤥𞤨𞤢𞤲𞤺', 'fi': 'sorang sompeng', 'fy': 'Sora Sompeng', 'ga': 'Sora Sompeng', 'gd': 'Sora Sompeng', 'he': 'סורה סומפנג', 'hi-Latn': 'Sora Sompeng', 'hr': 'sora sompeng pismo', 'id': 'Sora Sompeng', 'ig': 'Sora Sompeng', 'it': 'sora sompeng', 'ja': 'ソラング・ソンペング文字', 'ka': 'სორან-სომპენი', 'kk': 'соранг сомпенг жазуы', 'kk-Cyrl': 'соранг сомпенг жазуы', 'ko': '소라 솜펭 문자', 'lo': 'ໂສຮາສົມເປັງ', 'lt': 'Sora Sompeng', 'mk': 'соранг сомпенг', 'mn': 'Сора сомпенг', 'ms': 'Sora Sompeng', 'nb': 'sora sompeng', 'nl': 'Sora Sompeng', 'nn': 'sora sompeng', 'no': 'sora sompeng', 'nqo': 'ߛߏߙߊ ߛߎ߲ߔߍ߲ߜ߭', 'ro': 'sora sompeng', 'ru': 'сора-сонпенг', 'sc': 'sora sompeng', 'so': 'Qoraalka Sora Sompeng', 'sq': 'sorasompengisht', 'sv': 'sora sompeng', 'tg': 'Сора Сомпенг', 'th': 'โสราสมเป็ง', 'to': 'tohinima fakasolasomipengi', 'tr': 'Sora Sompeng', 'ug': 'سورا سامپىڭ', 'vec': 'sora sonpeng', 'vi': 'Chữ Sora Sompeng', 'yue': '索朗桑朋文字', 'yue-Hans': '索朗桑朋文字', 'yue-Hant': '索朗桑朋文字', 'zh': '索朗桑朋文', 'zh-Hans': '索朗桑朋文', 'zh-Hant': '索朗桑朋文字', 'zu': 'i-Sora Sompeng'},
    'Soyo': {'bal-Latn': 'Soyombó', 'bs': 'sojombo pismo', 'bs-Cyrl': 'сојомбо писмо', 'bs-Latn': 'sojombo pismo', 'en': 'Soyombo', 'et': 'sojombo', 'eu': 'soyomboera', 'ff-Adlm': '𞤅𞤮𞤴𞤮𞤥𞤦𞤮', 'fi': 'soyombo-kirjaimisto', 'fo': 'soyombo', 'ga': 'Soyombo', 'gd': 'Soyombo', 'he': 'סויומבו', 'hi-Latn': 'Soyombo', 'id': 'Soyombo', 'ig': 'Soyombo', 'kk': 'соёмбо жазуы', 'kk-Cyrl': 'соёмбо жазуы', 'mn': 'Соёмбо', 'ms': 'Soyombo', 'nl': 'Soyombo', 'nqo': 'ߛߏߦߏ߲ߓߏ', 'ro': 'soyombo', 'sc': 'soyombo', 'so': 'Soyombo', 'sq': 'sojomboisht', 'sv': 'soyombo', 'tg': 'Соёмбо', 'vec': 'soyonbo', 'zh': '索永布文', 'zh-Hans': '索永布文', 'zh-Hant': '索永布文字', 'zu': 'i-Soyombo'},
    'Sund': {'af': 'Soendanees', 'ak': 'Sudanni kasa', 'am': 'ሱዳንኛ', 'ar': 'السوندانية', 'as': 'ছাণ্ডানিজ', 'ast': 'sondanés', 'az': 'sundan', 'az-Latn': 'sundan', 'bal-Latn': 'Sudáni', 'be': 'сунданскае', 'be-tarask': 'сунданскае', 'bg': 'сунданска', 'bn': 'সুন্দানিজ', 'br': 'sundanek', 'brx': 'सून्डानीज', 'bs': 'sundansko pismo', 'bs-Cyrl': 'сунданесе писмо', 'bs-Latn': 'sundansko pismo', 'ca': 'sundanès', 'ca-ES-valencia': 'sundanès', 'ccp': '𑄥𑄚𑄴𑄘𑄚𑄨𑄎𑄴', 'chr': 'ᏚᏓᏂᏎ', 'cs': 'sundské', 'cy': 'Swndaneg', 'da': 'sundanesisk', 'de': 'Sundanesisch', 'dsb': 'sundaneske pismo', 'el': 'Σουνδανικό', 'el-polyton': 'Σουνδανικό', 'en': 'Sundanese', 'en-Dsrt': '𐐝𐐲𐑌𐐼𐐲𐑌𐐨𐑆', 'es': 'sundanés', 'et': 'sunda', 'eu': 'sudandarra', 'fa': 'سوندانی', 'ff-Adlm': '𞤅𞤵𞤲𞤣𞤢𞤲𞤭', 'fi': 'sundalainen', 'fil': 'Sundanese', 'fo': 'sudanskt', 'fr': 'sundanais', 'fy': 'Soendaneesk', 'ga': 'Sundainéis', 'gd': 'Sunda', 'gl': 'sundanés', 'gsw': 'Sundanesisch', 'gu': 'સુદાનીઝ', 'ha': 'Sudananci', 'he': 'סונדאני', 'hi': 'सूडानी', 'hi-Latn': 'Sundanese', 'hr': 'sundansko pismo', 'hsb': 'sundaneske pismo', 'ht': 'sundanais', 'hu': 'Szundanéz', 'hy': 'սունդանական', 'id': 'Sunda', 'ig': 'Sundanese', 'is': 'sundanesískt', 'it': 'sundanese', 'ja': 'スンダ文字', 'jv': 'Sunda', 'ka': 'სუნდანური', 'kgp': 'sũnanẽj', 'kk': 'сунд жазуы', 'kk-Cyrl': 'сунд жазуы', 'km': 'ស៊ូដង់', 'kn': 'ಸುಂಡಾನೀಸ್', 'ko': '순다 문자', 'kok': 'सुंदानीज', 'kok-Deva': 'सुंदानीज', 'ks': 'سَنڈَنیٖز', 'ks-Arab': 'سَنڈَنیٖز', 'ky': 'сундан жазуусу', 'lb': 'Sundanesesch', 'lo': 'ຊຸນດາ', 'lt': 'sundų', 'lv': 'zundu', 'mai': 'सुंडानीज', 'mi': 'Hunanihi', 'mk': 'сунданско', 'ml': 'സന്താനീസ്', 'mn': 'Сундан бичиг', 'mr': 'सूदानी', 'ms': 'Sunda', 'my': 'ဆူဒန်', 'nb': 'sundanesisk', 'ne': 'सुन्डानेली', 'nl': 'Soendanees', 'nn': 'sundanesisk', 'no': 'sundanesisk', 'nqo': 'ߛߎ߲ߘߣߊ', 'or': 'ସୁଦାନିଜ୍', 'pa': 'ਸੂੰਡਾਨੀ', 'pa-Guru': 'ਸੂੰਡਾਨੀ', 'pcm': 'Súndaníz Lángwej', 'pl': 'sundajskie', 'ps': 'سنڈانی', 'pt': 'sundanês', 'qu': 'Sundanese Simi', 'rm': 'sundanais', 'ro': 'sundaneză', 'ru': 'сунданская', 'sc': 'sundanesu', 'sd': 'سوڊاني', 'sd-Arab': 'سوڊاني', 'si': 'සන්ඩනීස්', 'sk': 'sundčina', 'sl': 'sundanski', 'so': 'Dadka Sundaniiska', 'sq': 'sundan', 'sr': 'сунданско писмо', 'sr-Cyrl': 'сунданско писмо', 'sr-Latn': 'sundansko pismo', 'sv': 'sundanesiska', 'sw': 'Kisunda', 'ta': 'சுந்தானீஸ்', 'te': 'సుడానీస్', 'tg': 'Сунданезӣ', 'th': 'ซุนดา', 'ti': 'ሱንዳናዊ', 'tk': 'Sundanez ýazuwy', 'to': 'tohinima fakasunitā', 'tr': 'Sunda', 'ug': 'سۇنداچە', 'uk': 'сунданська', 'ur': 'سوڈانی', 'uz': 'sundan', 'uz-Latn': 'sundan', 'vec': 'sundaneze', 'vi': 'Chữ Xu-đăng', 'xnr': 'सूडानी', 'yo': 'Èdè Sundani', 'yrl': 'sudãnei', 'yue': '巽他文', 'yue-Hans': '巽他文', 'yue-Hant': '巽他文', 'zh': '巽他文', 'zh-Hans': '巽他文', 'zh-Hant': '巽他文', 'zu': 'isi-Sundanese'},
    'Sunu': {'bal-Latn': 'Sunuwar', 'en': 'Sunuwar', 'et': 'sunvari', 'fo': 'sunuwar', 'gd': 'Sunuwar', 'hi-Latn': 'Sunuwar', 'ig': 'Sunuwar', 'vec': 'sunuwar'},
    'Sylo': {'ar': 'السيلوتي ناغري', 'ast': 'syloti nagri', 'az': 'siloti nəqri', 'az-Latn': 'siloti nəqri', 'bal-Latn': 'Siluti Nagri', 'bg': 'Силоти Нагри', 'bn': 'সিলেটি নাগরি', 'brx': 'सील्होटी नागरी', 'bs': 'siloti nagri pismo', 'bs-Cyrl': 'силоти нагри писмо', 'bs-Latn': 'siloti nagri pismo', 'ca': 'syloti nagri', 'ca-ES-valencia': 'syloti nagri', 'ccp': '𑄥𑄨𑄣𑄬𑄑𑄨 𑄚𑄉𑄧𑄢𑄨', 'cs': 'sylhetské', 'da': 'syloti nagri', 'de': 'Syloti Nagri', 'el': 'Συλότι Νάγκρι', 'el-polyton': 'Συλότι Νάγκρι', 'en': 'Syloti Nagri', 'en-Dsrt': '𐐝𐐴𐑊𐐱𐐻𐐨 𐐤𐐰𐑀𐑉𐐨', 'es': 'syloti nagri', 'et': 'siloti', 'eu': 'syloti nagriera', 'fa': 'سیلوتی نگاری', 'ff-Adlm': '𞤅𞤴𞤤𞤮𞤼𞤭𞥅 𞤐𞤢𞤺𞤪𞤭', 'fi': 'syloti nagri', 'fo': 'syloti nagri', 'fr': 'sylotî nâgrî', 'fy': 'Syloti Nagri', 'ga': 'Syloti Nagri', 'gd': 'Syloti Nagri', 'gsw': 'Syloti Nagri', 'gu': 'સિલોતી નાગરી', 'he': 'סילוטי נגרי', 'hi': 'सिलोती नागरी', 'hi-Latn': 'Syloti Nagri', 'hr': 'syloti nagri pismo', 'ht': 'sylotî nâgrî', 'hu': 'Sylheti nagári', 'id': 'Syloti Nagri', 'ig': 'Syloti Nagri', 'it': 'syloti nagri', 'ja': 'シロティ・ナグリ文字', 'ka': 'სილოტი ნაგრი', 'kgp': 'syroti nỹgri', 'kk': 'силхет нагари жазуы', 'kk-Cyrl': 'силхет нагари жазуы', 'km': 'ស៊ីឡូ', 'kn': 'ಸೈಲೋಟಿ ನಗ್ರಿ', 'ko': '실헤티 나가리', 'ks': 'سیلوتی ناگری', 'ks-Arab': 'سیلوتی ناگری', 'lb': 'Syloti Nagri', 'lo': 'ຊີໂລຕິນາກຣີ', 'lt': 'syloti nagri', 'mk': 'силхетско нагари', 'ml': 'സൈലോതി നാഗരി', 'mn': 'Сило', 'mr': 'सिलोती नागरी', 'ms': 'Syloti Nagri', 'nb': 'syloti nagri', 'ne': 'स्ल्योटी नाग्री', 'nl': 'Syloti Nagri', 'nn': 'syloti nagri', 'no': 'syloti nagri', 'nqo': 'ߛߌߟߏߕߌ ߣߊߜ߭ߑߙߌ', 'or': 'ସୀଲିତୋ ନଗରୀ', 'pl': 'syloti nagri', 'pt': 'syloti nagri', 'rm': 'syloti nagri', 'ro': 'syloti nagri', 'ru': 'силоти нагри', 'sc': 'syloti nagri', 'sl': 'siletsko-nagarijski', 'so': 'Qoraalka Luuqada Sylheti', 'sq': 'siloti nagri', 'sr': 'силоти нагри писмо', 'sr-Cyrl': 'силоти нагри писмо', 'sr-Latn': 'siloti nagri pismo', 'sv': 'syloti nagri', 'ta': 'சிலோடி நக்ரி', 'te': 'స్లోటి నాగ్రి', 'tg': 'Силоти Нагрӣ', 'th': 'ซิโลตินากรี', 'to': 'tohinima fakasailoti-nakili', 'tr': 'Syloti Nagri', 'ug': 'سىيولوتى-ناگرىچە', 'uk': 'сілоті нагрі', 'vec': 'syloti nagri', 'vi': 'Chữ Syloti Nagri', 'xnr': 'सिलोती नागरी', 'yrl': 'siruti nagiri', 'yue': '希洛弟納格里文', 'yue-Hans': '希洛弟纳格里文', 'yue-Hant': '希洛弟納格里文', 'zh': '锡尔赫特文', 'zh-Hans': '锡尔赫特文', 'zh-Hant': '希洛弟納格里文', 'zu': 'isi-Syloti Nagri'},
    'Syrc': {'af': 'Siries', 'ak': 'Siiria Tete kasa', 'am': 'ሲሪያክ', 'ar': 'السريانية', 'as': 'ছিৰিয়াক', 'ast': 'siriacu', 'az': 'siryak', 'az-Latn': 'siryak', 'bal-Latn': 'Siriek', 'be': 'сірыйскае', 'be-tarask': 'сірыйскае', 'bg': 'сирийска', 'bn': 'সিরিয়াক', 'br': 'siriek', 'brx': 'सीरीआक', 'bs': 'sirijsko pismo', 'bs-Cyrl': 'сиријско писмо', 'bs-Latn': 'sirijsko pismo', 'ca': 'siríac', 'ca-ES-valencia': 'siríac', 'ccp': '𑄥𑄨𑄢𑄨𑄠𑄇𑄴', 'chr': 'ᏏᎵᎡᎩ', 'cs': 'syrské', 'cy': 'Syrieg', 'da': 'syrisk', 'de': 'Syrisch', 'dsb': 'syriacske pismo', 'el': 'Συριακό', 'el-polyton': 'Συριακό', 'en': 'Syriac', 'en-Dsrt': '𐐝𐐮𐑉𐐨𐐰𐐿', 'es': 'siriaco', 'et': 'süüria', 'eu': 'asiriarra', 'fa': 'سریانی', 'ff-Adlm': '𞤅𞤭𞥅𞤪𞤴𞤢𞤳', 'fi': 'syyrialainen', 'fil': 'Syriac', 'fo': 'syriac', 'fr': 'syriaque', 'fur': 'siriac', 'fy': 'Syriac', 'ga': 'Siriceach', 'gd': 'Suraidheac', 'gl': 'siríaco', 'gsw': 'Syrisch', 'gu': 'સિરિયેક', 'ha': 'Siriyanci', 'he': 'סורי', 'hi': 'सिरिएक', 'hi-Latn': 'Syriac', 'hr': 'sirijsko pismo', 'hsb': 'syriacske pismo', 'ht': 'syriaque', 'hu': 'Szíriai', 'hy': 'ասորական', 'id': 'Suriah', 'ig': 'Siriak', 'is': 'syriakískt', 'it': 'siriaco', 'ja': 'シリア文字', 'jv': 'Siriak', 'ka': 'სირიული', 'kgp': 'sirijaku', 'kk': 'сирия жазуы', 'kk-Cyrl': 'сирия жазуы', 'km': 'ស៊ីរីអែក', 'kn': 'ಸಿರಿಯಾಕ್', 'ko': '시리아 문자', 'kok': 'सिरीयाक', 'kok-Deva': 'सिरीयाक', 'ks': 'سیٖرِیَک', 'ks-Arab': 'سیٖرِیَک', 'ky': 'сириялык жазуу', 'lb': 'Syresch', 'lo': 'ຊີເຣຍ', 'lt': 'sirų', 'lv': 'sīriešu', 'mai': 'सीरियक', 'mi': 'Hīriaka', 'mk': 'сириско', 'ml': 'സിറിയക്ക്', 'mn': 'Сирийк бичиг', 'mr': 'सिरीयाक', 'ms': 'Syria', 'my': 'ဆီရီရက်ခ်', 'nb': 'syrisk', 'ne': 'सिरियाक', 'nl': 'Syriac', 'nn': 'syrisk', 'no': 'syrisk', 'nqo': 'ߛߙߌߦߊߞ', 'or': 'ସିରିୟାକ୍', 'pa': 'ਸੀਰੀਆਈ', 'pa-Guru': 'ਸੀਰੀਆਈ', 'pcm': 'Síriák Lángwej', 'pl': 'syryjskie', 'ps': 'سریانی', 'pt': 'siríaco', 'qu': 'Sirio Simi', 'rm': 'siric', 'ro': 'siriacă', 'ru': 'сирийская', 'sc': 'sirìacu', 'sd': 'شامي', 'sd-Arab': 'شامي', 'si': 'සිරියැක්', 'sk': 'sýrčina', 'sl': 'sirijski', 'so': 'Lahjada Syriac', 'sq': 'siriak', 'sr': 'сиријско писмо', 'sr-Cyrl': 'сиријско писмо', 'sr-Latn': 'sirijsko pismo', 'sv': 'syriska', 'sw': 'Kisiriaki', 'syr': 'ܣܘܪܝܬ', 'ta': 'சிரியாக்', 'te': 'సిరియాక్', 'tg': 'Сурёнӣ', 'th': 'ซีเรีย', 'ti': 'ስይሪክ', 'tk': 'Siriýa ýazuwy', 'to': 'tohinima fakasuliāiā', 'tr': 'Süryani', 'ug': 'سۈرىيەچە', 'uk': 'сирійська', 'ur': 'سریانی', 'uz': 'suryoniy', 'uz-Latn': 'suryoniy', 'vec': 'sirìago', 'vi': 'Chữ Syria', 'xnr': 'सिरियेक', 'yo': 'Èdè Siriaki', 'yrl': 'siríaku', 'yue': '敍利亞文', 'yue-Hans': '敍利亚文', 'yue-Hant': '敍利亞文', 'zh': '叙利亚文', 'zh-Hans': '叙利亚文', 'zh-Hant': '敘利亞文', 'zu': 'isi-Syriac'},
    'Syre': {'ar': 'السريانية الأسترنجيلية', 'ast': 'siriacu estrangelo', 'az': 'estrangela süryanice', 'az-Latn': 'estrangela süryanice', 'bal-Latn': 'Estrangló Siriek', 'bg': 'Сирийска естрангело', 'bn': 'এস্ট্রেঙ্গেলো সিরিয়াক', 'br': 'siriek Estrangelā', 'brx': 'एस्ट्रांगलो सीरीआक', 'bs': 'sirijsko estrangelo pismo', 'bs-Cyrl': 'сиријско естрангело писмо', 'bs-Latn': 'sirijsko estrangelo pismo', 'ca': 'siríac estrangelo', 'ca-ES-valencia': 'siríac estrangelo', 'ccp': '𑄃𑄬𑄌𑄴𑄑𑄳𑄢𑄬𑄋𑄴𑄉𑄬𑄣𑄮 𑄥𑄨𑄢𑄨𑄠𑄇𑄴', 'cs': 'syrské - estrangelo', 'da': 'syrisk - estrangelovariant', 'de': 'Syrisch - Estrangelo-Variante', 'el': 'Εστραντζέλο Συριακό', 'el-polyton': 'Εστραντζέλο Συριακό', 'en': 'Estrangelo Syriac', 'en-Dsrt': '𐐇𐑅𐐻𐑉𐐪𐑍𐐾𐐯𐑊𐐬 𐐝𐐮𐑉𐐨𐐰𐐿', 'es': 'siriaco estrangelo', 'et': 'süüria estrangelo', 'fa': 'سریانی سطرنجیلی', 'fi': 'syyrialainen estrangelo-muunnelma', 'fr': 'syriaque estranghélo', 'fur': 'siriac Estrangelo', 'fy': 'Estrangelo Arameesk', 'gd': 'Suraidheac Estrangela', 'gsw': 'Syrisch - Eschtrangelo-Variante', 'gu': 'એસ્ત્રેન્જેલો સિરિયાક', 'hi': 'एस्त्रेन्जेलो सिरिएक', 'hi-Latn': 'Estrangelo Syriac', 'hr': 'sirijsko estrangelo pismo', 'ht': 'syriaque estranghélo', 'hu': 'Estrangelo szíriai', 'id': 'Suriah Estrangelo', 'ig': 'Estrangelo Syriac', 'it': 'siriaco estrangelo', 'ja': 'シリア文字(エストランゲロ文字)', 'ka': 'სირიული ესტრანგელო', 'kgp': 'sirijaku esitarageru', 'kn': 'ಎಸ್ಟ್ರಾಂಜಿಲೋ ಸಿರಿಯಾಕ್', 'ko': '에스트랑겔로식 시리아 문자', 'ks': 'ایسٹرینجِلو سیٖرِیَک', 'ks-Arab': 'ایسٹرینجِلو سیٖرِیَک', 'lb': 'Syresch-Estrangelo-Variant', 'lo': 'ຊີເຮຍເອສທຮານຈີໂລ', 'lt': 'estrangelo siriečių', 'mk': 'естрангелско сириско', 'ml': 'എസ്റ്റ്രാംഗ്ലോ സിറിയക്', 'mr': 'एस्त्ट्रेन्जेलो सिरियाक', 'nb': 'estrangelosyriakisk', 'ne': 'इस्ट्रेनजेलो सिरियाक', 'nl': 'Estrangelo Aramees', 'nn': 'syriakisk (estrangelo-variant)', 'no': 'estrangelosyriakisk', 'or': 'ଏଷ୍ଟ୍ରାଙ୍ଗେଲୋ ସିରିକ୍', 'pl': 'syriacki estrangelo', 'pt': 'siríaco estrangelo', 'rm': 'siric estrangelo', 'ru': 'сирийская эстрангело', 'sl': 'sirska abeceda estrangelo', 'sr': 'сиријско естрангело писмо', 'sr-Cyrl': 'сиријско естрангело писмо', 'sr-Latn': 'sirijsko estrangelo pismo', 'sv': 'estrangelosyriska', 'ta': 'எஸ்ட்ரெங்கெலோ சிரியாக்', 'te': 'ఎస్ట్రానజీలో సిరియాక్', 'th': 'ซีเรียเอสทรานจีโล', 'to': 'tohinima fakasuliāiā-ʻesitelangelo', 'tr': 'Estrangela Süryani', 'ug': 'سۈرىيەچە ئەبجەت', 'uk': 'давньосирійський естрангело', 'vec': 'sirìago estrànjelo', 'vi': 'Chữ Estrangelo Syriac', 'xnr': 'एस्त्रेन्जेलो सिरिएक', 'yrl': 'siríaku esitarãgeru', 'yue': '敘利亞文（福音體文字變體）', 'yue-Hans': '叙利亚文（福音体文字变体）', 'yue-Hant': '敘利亞文（福音體文字變體）', 'zh': '福音体叙利亚文', 'zh-Hans': '福音体叙利亚文', 'zh-Hant': '敘利亞文（福音體文字變體）'},
    'Syrj': {'ar': 'السريانية الغربية', 'ast': 'siriacu occidental', 'bal-Latn': 'Rónendi Siriek', 'bg': 'Западна сирийска', 'bn': 'পশ্চিমাঞ্চলীয় সিরিয়াক', 'br': 'siriek ar C’hornôg', 'brx': 'पश्चीमी सीरीआक', 'bs': 'zapadnosirijsko pismo', 'bs-Cyrl': 'западносиријско писмо', 'bs-Latn': 'zapadnosirijsko pismo', 'ca': 'siríac occidental', 'ca-ES-valencia': 'siríac occidental', 'ccp': '𑄛𑄧𑄏𑄨𑄟𑄴𑄎𑄉𑄢𑄴 𑄥𑄨𑄢𑄨𑄠𑄇𑄴', 'cs': 'syrské - západní', 'da': 'vestsyrisk', 'de': 'Westsyrisch', 'el': 'Δυτικό Συριακό', 'el-polyton': 'Δυτικό Συριακό', 'en': 'Western Syriac', 'en-Dsrt': '𐐎𐐯𐑅𐐻𐐲𐑉𐑌 𐐝𐐮𐑉𐐨𐐰𐐿', 'es': 'siriaco occidental', 'et': 'läänesüüria', 'fa': 'سریانی غربی', 'fi': 'syyrialainen läntinen muunnelma', 'fr': 'syriaque occidental', 'fur': 'siriac ocidentâl', 'fy': 'West-Arameesk', 'gd': 'Suraidheac Siarach', 'gsw': 'Weschtsyrisch', 'gu': 'પશ્ચિમ સિરિયાક', 'he': 'סורי מערבי', 'hi': 'पश्चिम सिरिएक', 'hi-Latn': 'Western Syriac', 'hr': 'pismo zapadne Sirije', 'ht': 'syriaque occidental', 'hu': 'Nyugat-szíriai', 'id': 'Suriah Barat', 'ig': 'Western Syriac', 'it': 'siriaco occidentale', 'ja': 'シリア文字(西方シリア文字)', 'ka': 'დასავლეთი სირიული', 'kgp': 'sirijaku rãpur', 'kn': 'ಪಶ್ಚಿಮ ಸಿರಿಯಾಕ್', 'ko': '서부 시리아 문자', 'ks': 'مغرِبی سیٖریَک', 'ks-Arab': 'مغرِبی سیٖریَک', 'lb': 'Westsyresch', 'lo': 'ຊີເຮຍຕາເວັນຕົກ', 'lt': 'vakarų sirų', 'lv': 'rietumsīriešu', 'mk': 'западносириско', 'ml': 'പശ്ചിമസുറിയാനി', 'mr': 'पश्चिमी सिरियाक', 'nb': 'vestlig syriakisk', 'ne': 'पश्चिमी सिरियाक', 'nl': 'West-Aramees', 'nn': 'syriakisk (vestleg variant)', 'no': 'vestlig syriakisk', 'or': 'ୱେଷ୍ଟର୍ନ ସିରିକ୍', 'pl': 'syryjski (odmiana zachodnia)', 'pt': 'siríaco ocidental', 'rm': 'siric dal vest', 'ro': 'siriacă occidentală', 'ru': 'западносирийская', 'sl': 'zahodnosirijski', 'sr': 'западносиријско писмо', 'sr-Cyrl': 'западносиријско писмо', 'sr-Latn': 'zapadnosirijsko pismo', 'sv': 'västsyriska', 'ta': 'மேற்கு சிரியாக்', 'te': 'పశ్చిమ సిరియాక్', 'th': 'ซีเรียตะวันตก', 'to': 'tohinima fakasuliāiā-hihifo', 'tr': 'Batı Süryani', 'ug': 'غەربىي سۈرىيەچە', 'uk': 'давньосирійський західний', 'vec': 'sirìago osidentale', 'vi': 'Chữ Tây Syria', 'xnr': 'पश्चिम सिरिएक', 'yrl': 'siriaku usidẽtawara', 'yue': '敘利亞文（西方文字變體）', 'yue-Hans': '叙利亚文（西方文字变体）', 'yue-Hant': '敘利亞文（西方文字變體）', 'zh': '西叙利亚文', 'zh-Hans': '西叙利亚文', 'zh-Hant': '敘利亞文（西方文字變體）'},
    'Syrn': {'ar': 'السريانية الشرقية', 'ast': 'siriacu oriental', 'bal-Latn': 'Ródarátki Siriek', 'bg': 'Източна сирийска', 'bn': 'পূর্বাঞ্চলীয় সিরিয়াক', 'br': 'siriek ar Reter', 'brx': 'पूर्वी सीरीआक', 'bs': 'istočnosirijsko pismo', 'bs-Cyrl': 'писмо источне Сирије', 'bs-Latn': 'istočnosirijsko pismo', 'ca': 'siríac oriental', 'ca-ES-valencia': 'siríac oriental', 'ccp': '𑄛𑄪𑄇𑄴𑄎𑄉𑄧𑄢𑄴 𑄥𑄨𑄢𑄨𑄠𑄇𑄴', 'cs': 'syrské - východní', 'da': 'østsyriakisk', 'de': 'Ostsyrisch', 'el': 'Ανατολικό Συριακό', 'el-polyton': 'Ανατολικό Συριακό', 'en': 'Eastern Syriac', 'en-Dsrt': '𐐀𐑅𐐻𐐲𐑉𐑌 𐐝𐐮𐑉𐐨𐐰𐐿', 'es': 'siriaco oriental', 'et': 'idasüüria', 'fa': 'سریانی شرقی', 'fi': 'syyrialainen itäinen muunnelma', 'fr': 'syriaque oriental', 'fur': 'siriac orientâl', 'fy': 'East-Arameesk', 'gd': 'Suraidheac Earach', 'gsw': 'Oschtsyrisch', 'gu': 'પૂર્વ સિરિયાક', 'he': 'סורי מזרחי', 'hi': 'पूर्व सिरिएक', 'hi-Latn': 'Eastern Syriac', 'hr': 'pismo istočne Sirije', 'ht': 'syriaque oriental', 'hu': 'Kelet-szíriai', 'id': 'Suriah Timur', 'ig': 'Eastern Syriac', 'it': 'siriaco orientale', 'ja': 'シリア文字(東方シリア文字)', 'ka': 'აღმოსავლეთი სირიული', 'kgp': 'sirijaku rãjur', 'kn': 'ಪೂರ್ವ ಸಿರಿಯಾಕ್', 'ko': '동부 시리아 문자', 'ks': 'مشرَقی سیٖریَک', 'ks-Arab': 'مشرَقی سیٖریَک', 'lb': 'Ostsyresch', 'lo': 'ຊີເຮຍຕາເວັນອອກ', 'lt': 'rytų sirų', 'lv': 'austrumsīriešu', 'mk': 'источносириско', 'ml': 'കിഴക്കൻ സിറിയക്', 'mr': 'पूर्वी सिरियाक', 'nb': 'østlig syriakisk', 'ne': 'पूर्वी सिरियाक', 'nl': 'Oost-Aramees', 'nn': 'syriakisk (austleg variant)', 'no': 'østlig syriakisk', 'or': 'ଇଷ୍ଟର୍ନ ସିରିକ୍', 'pl': 'syryjski (odmiana wschodnia)', 'pt': 'siríaco oriental', 'rm': 'siric da l’ost', 'ro': 'siriacă orientală', 'ru': 'восточно-сирийская', 'sl': 'vzhodnosirijski', 'sr': 'писмо источне Сирије', 'sr-Cyrl': 'писмо источне Сирије', 'sr-Latn': 'pismo istočne Sirije', 'sv': 'östsyriska', 'ta': 'கிழக்கு சிரியாக்', 'te': 'తూర్పు సిరియాక్', 'th': 'ซีเรียตะวันออก', 'to': 'tohinima fakasuliāiā-hahake', 'tr': 'Doğu Süryani', 'ug': 'شەرقىي سۈرىيەچە', 'uk': 'давньосирійський східний', 'vec': 'sirìago orientale', 'vi': 'Chữ Đông Syria', 'xnr': 'पूर्व सिरिएक', 'yrl': 'siriaku uriẽtawara', 'yue': '敘利亞文（東方文字變體）', 'yue-Hans': '叙利亚文（东方文字变体）', 'yue-Hant': '敘利亞文（東方文字變體）', 'zh': '东叙利亚文', 'zh-Hans': '东叙利亚文', 'zh-Hant': '敘利亞文（東方文字變體）'},
    'TA': {'af': 'Tristan da Cunha', 'ak': 'Tristan da Kuna', 'am': 'ትሪስታን ዳ ኩንሃ', 'ar': 'تريستان دا كونا', 'as': 'ত্ৰিস্তান দ্যা কুনহা', 'ast': 'Tristán da Cunha', 'az': 'Tristan da Kunya', 'az-Cyrl': 'Тристан да Кунја', 'az-Latn': 'Tristan da Kunya', 'bal-Latn': 'Tristan dá Chonhá', 'be': 'Трыстан-да-Кунья', 'be-tarask': 'Трыстан-да-Кунья', 'bg': 'Тристан да Куня', 'blo': 'Tristan ɖa Kuna', 'bn': 'ট্রিস্টান ডা কুনহা', 'br': 'Tristan da Cunha', 'brx': 'त्रिस्तान दा कुन्हा', 'bs': 'Tristan da Cunha', 'bs-Cyrl': 'Тристан да Куња', 'bs-Latn': 'Tristan da Cunha', 'ca': 'Tristan da Cunha', 'ca-ES-valencia': 'Tristan da Cunha', 'ccp': '𑄑𑄳𑄢𑄌𑄴𑄑𑄚𑄴 𑄓 𑄇𑄪𑄚𑄴𑄦', 'ce': 'Тристан-да- Кунья', 'ceb': 'Tristan da Cunha', 'chr': 'ᏟᏍᏛᏂ Ꮣ ᎫᎾᎭ', 'ckb': 'تریستێن دا کوونا', 'cs': 'Tristan da Cunha', 'cv': 'Тристан-да-Кунья', 'cy': 'Tristan da Cunha', 'da': 'Tristan da Cunha', 'de': 'Tristan da Cunha', 'dsb': 'Tristan da Cunha', 'dz': 'ཏྲིས་ཏན་ད་ཀུན་ཧ', 'ee': 'Tristan da Kunha nutome', 'el': 'Τριστάν ντα Κούνια', 'el-polyton': 'Τριστάν ντα Κούνια', 'en': 'Tristan da Cunha', 'es': 'Tristán de Acuña', 'et': 'Tristan da Cunha', 'eu': 'Tristan da Cunha', 'fa': 'تریستان دا کونا', 'ff-Adlm': '𞤚𞤵𞤪𞤧𞤵𞤼𞤢𞤲 𞤁𞤢𞤳𞤵𞤲𞤸𞤢', 'fi': 'Tristan da Cunha', 'fil': 'Tristan de Cunha', 'fo': 'Tristan da Cunha', 'fr': 'Tristan da Cunha', 'fur': 'Tristan da Cunha', 'fy': 'Tristan da Cunha', 'ga': 'Tristan da Cunha', 'gd': 'Tristan da Cunha', 'gl': 'Tristán da Cunha', 'gu': 'ત્રિસ્તાન દા કુન્હા', 'ha': 'Tritan da Kunha', 'he': 'טריסטן דה קונה', 'hi': 'ट्रिस्टन दा कुना', 'hi-Latn': 'Tristan da Cunha', 'hr': 'Tristan da Cunha', 'hsb': 'Tristan da Cunha', 'ht': 'Tristan da Cunha', 'hu': 'Tristan da Cunha', 'hy': 'Տրիստան դա Կունյա', 'ia': 'Tristan da Cunha', 'id': 'Tristan da Cunha', 'ig': 'Tristan da Cunha', 'is': 'Tristan da Cunha', 'it': 'Tristan da Cunha', 'ja': 'トリスタン・ダ・クーニャ', 'jv': 'Tristan da Cunha', 'ka': 'ტრისტან-და-კუნია', 'kea': 'Tristan da Kunha', 'kgp': 'Trinhtỹv Nakũja', 'kk': 'Тристан-да-Кунья', 'kk-Cyrl': 'Тристан-да-Кунья', 'km': 'ទ្រីស្តង់\u200bដា\u200bចូនហា', 'kn': 'ಟ್ರಿಸ್ತನ್ ಡಾ ಕುನ್ಹಾ', 'ko': '트리스탄다쿠나', 'kok': 'त्रिस्तान दा कुन्हा', 'kok-Deva': 'त्रिस्तान दा कुन्हा', 'ks': 'ٹریسٹن ڈا کنہا', 'ks-Arab': 'ٹریسٹن ڈا کنہا', 'ksh': 'Tristan da Cunja', 'ku': 'Tristan da Cunha', 'ky': 'Тристан-да-Кунья', 'lb': 'Tristan da Cunha', 'lld': 'Tristan da Cunha', 'lo': 'ທຣິສຕັນ ດາ ກັນຮາ', 'lt': 'Tristano da Kunjos', 'lv': 'Tristana da Kuņjas salu teritorijas', 'mai': 'ट्रिस्टन दा कुना', 'mi': 'Tiritana da Kunia', 'mk': 'Тристан да Куња', 'ml': 'ട്രിസ്റ്റൻ ഡ കൂന', 'mn': 'Тристан да Кунъя', 'mr': 'ट्रिस्टन दा कुन्हा', 'ms': 'Tristan da Cunha', 'mt': 'Tristan da Cunha', 'my': 'ထရစ္စတန် ဒါ ကွန်ဟာ', 'mzn': 'تریستان دا جونها', 'nb': 'Tristan da Cunha', 'ne': 'ट्रिस्टान डा कुन्हा', 'nl': 'Tristan da Cunha', 'nn': 'Tristan da Cunha', 'no': 'Tristan da Cunha', 'nqo': 'ߞߎ߲ߓߊ߫ ߕߑߙߌߛߑߕߊ߫', 'om': 'Tiristaan da Kanhaa', 'or': 'ଟ୍ରାଇଷ୍ଟନ୍\u200c ଦା କୁନ୍\u200cଚା', 'pa': 'ਟ੍ਰਿਸਟਾਨ ਦਾ ਕੁੰਹਾ', 'pa-Guru': 'ਟ੍ਰਿਸਟਾਨ ਦਾ ਕੁੰਹਾ', 'pcm': 'Trístán da Kúna', 'pl': 'Tristan da Cunha', 'ps': 'تریستان دا کنها', 'pt': 'Tristão da Cunha', 'qu': 'Tristán de Acuña', 'rm': 'Tristan da Cunha', 'ro': 'Tristan da Cunha', 'ru': 'Тристан-да-Кунья', 'sc': 'Tristan da Cunha', 'sd': 'ٽرسٽن دا ڪوها', 'sd-Arab': 'ٽرسٽن دا ڪوها', 'se': 'Tristan da Cunha', 'si': 'ට්\u200dරිස්ටන් ද කුන්හා', 'sk': 'Tristan da Cunha', 'sl': 'Tristan da Cunha', 'smn': 'Tristan da Cunha', 'so': 'Tiristan da Kunha', 'sq': 'Tristan-da-Kuna', 'sr': 'Тристан да Куња', 'sr-Cyrl': 'Тристан да Куња', 'sr-Latn': 'Tristan da Kunja', 'sv': 'Tristan da Cunha', 'sw': 'Tristan da Cunha', 'ta': 'டிரிஸ்டன் டா குன்ஹா', 'te': 'ట్రిస్టన్ డ కన్హా', 'tg': 'Тристан-да-Куня', 'th': 'ทริสตันดาคูนา', 'ti': 'ትሪስታን ዳ ኩንያ', 'tk': 'Tristan-da-Kunýa', 'to': 'Tulisitani ta Kunuha', 'tr': 'Tristan da Cunha', 'tt': 'Тристан-да-Кунья', 'ug': 'ترىستان داكۇنھا', 'uk': 'Трістан-да-Кунья', 'ur': 'ٹرسٹن ڈا کیونہا', 'uz': 'Tristan-da-Kunya', 'uz-Cyrl': 'Тристан-да-Куня', 'uz-Latn': 'Tristan-da-Kunya', 'vai': 'ꔳꔻꕚꘋ ꕜ ꖴꕯ', 'vai-Vaii': 'ꔳꔻꕚꘋ ꕜ ꖴꕯ', 'vec': 'Tristan de Cugna', 'vi': 'Tristan da Cunha', 'wae': 'Tristan da Cunha', 'wo': 'Tristan da Cunha', 'xh': 'ETristan da Cunha', 'xnr': 'ट्रिस्टन दा कुना', 'yo': 'Tristan da Kunha', 'yrl': 'Tiritãu Kũya', 'yue': '特里斯坦達庫尼亞群島', 'yue-Hans': '特里斯坦达库尼亚群岛', 'yue-Hant': '特里斯坦達庫尼亞群島', 'zh': '特里斯坦-达库尼亚群岛', 'zh-Hans': '特里斯坦-达库尼亚群岛', 'zh-Hant': '特里斯坦達庫尼亞群島', 'zu': 'i-Tristan da Cunha'},
    'TC': {'af': 'Turks- en Caicoseilande', 'agq': 'Chwɨla n Tɨtê à Kaekùs', 'ak': 'Turks ne Caicos Islands', 'am': 'የቱርኮችና የካኢኮስ ደሴቶች', 'ar': 'جزر توركس وكايكوس', 'as': 'টাৰ্কছ অৰু কেইক’ছ দ্বীপপুঞ্জ', 'asa': 'Vithiwa vya Turki na Kaiko', 'ast': 'Islles Turques y Caicos', 'az': 'Törks və Kaykos adaları', 'az-Cyrl': 'Төркс вә Кајкос адалары', 'az-Latn': 'Törks və Kaykos adaları', 'bal-Latn': 'Tork o Káikói Islánd', 'bas': 'Bìòn bi Tûrks nì Kalkòs', 'be': 'Астравы Цёркс і Кайкас', 'be-tarask': 'Астравы Цёркс і Кайкас', 'bez': 'Ifisima fya Turki na Kaiko', 'bg': 'острови Търкс и Кайкос', 'blo': 'Turkisii na Kayɩkɔɔsɩ kaBʊtukǝltǝna', 'bm': 'Turiki Gun ni Kayiki', 'bn': 'তুর্কস ও কাইকোস দ্বীপপুঞ্জ', 'br': 'Inizi Turks ha Caicos', 'brx': 'तुर्की आरो काईक’ज द्वीपफोर', 'bs': 'Ostrva Turks i Kaikos', 'bs-Cyrl': 'Туркс и Кајкос Острва', 'bs-Latn': 'Ostrva Turks i Kaikos', 'ca': 'Illes Turks i Caicos', 'ca-ES-valencia': 'Illes Turks i Caicos', 'ccp': '𑄖𑄪𑄢𑄴𑄇𑄧𑄌𑄴 𑄃𑄮 𑄇𑄭𑄇𑄮𑄌𑄴 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Тёркс а, Кайкос а гӀайренаш', 'ceb': 'Turks & Caicos Islands', 'cgg': 'Ebizinga bya Buturuki na Kaiko', 'chr': 'ᎠᏂᏛᎵᎩ ᎠᎴ ᎨᎢᎪ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی تورکس و کایکۆس', 'cs': 'Turks a Caicos', 'cv': 'Тёркс тата Кайкос утравӗсем', 'cy': 'Ynysoedd Turks a Caicos', 'da': 'Turks- og Caicosøerne', 'dav': 'Visiwa vya Turki na Kaiko', 'de': 'Turks- und Caicosinseln', 'dje': 'Turk nda Kayikos Gungey', 'dsb': 'Turks a Caicos kupy', 'dz': 'ཏུརྐས྄་ ཨེནཌ་ ཀ་ཀོས་གླིང་ཚོམ', 'ebu': 'Visiwa vya Turki na Kaiko', 'ee': 'Tɛks kple Kaikos ƒudomekpowo nutome', 'el': 'Νήσοι Τερκς και Κάικος', 'el-polyton': 'Νῆσοι Τὲρκς καὶ Κάικος', 'en': 'Turks & Caicos Islands', 'en-Dsrt': '𐐓𐐲𐑉𐐿𐑅 𐐰𐑌𐐼 𐐗𐐴𐐿𐐬𐑆 𐐌𐑊𐐲𐑌𐐼𐑆', 'es': 'Islas Turcas y Caicos', 'et': 'Turks ja Caicos', 'eu': 'Turk eta Caico uharteak', 'ewo': 'Minlán Mí túrə́g-ai-Kaíg', 'fa': 'جزایر تورکس و کایکوس', 'ff': 'Duuɗe Turke e Keikoos', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤚𞤵𞤪𞤳𞤵𞤧 & 𞤑𞤢𞤴𞤳𞤮𞥅𞤧', 'ff-Latn': 'Duuɗe Turke e Keikoos', 'fi': 'Turks- ja Caicossaaret', 'fil': 'Turks & Caicos Islands', 'fo': 'Turks- og Caicosoyggjar', 'fr': 'Îles Turques-et-Caïques', 'fur': 'Isulis Turks e Caicos', 'fy': 'Turks- en Caicoseilannen', 'ga': 'Oileáin na dTurcach agus Caicos', 'gd': 'Na h-Eileanan Turcach is Caiceo', 'gl': 'Illas Turks e Caicos', 'gsw': 'Törks- und Gaiggos-Insle', 'gu': 'તુર્ક્સ અને કેકોઝ આઇલેન્ડ્સ', 'guz': 'Visiwa vya Turki na Kaiko', 'ha': 'Turkis Da Tsibiran Kaikwas', 'he': 'איי טרקס וקייקוס', 'hi': 'तुर्क और कैकोज़ द्वीपसमूह', 'hi-Latn': 'Turks & Caicos Islands', 'hr': 'Otoci Turks i Caicos', 'hsb': 'kupy Turks a Caicos', 'ht': 'Îles Turques-et-Caïques', 'hu': 'Turks- és Caicos-szigetek', 'hy': 'Թըրքս և Կայկոս կղզիներ', 'ia': 'Insulas Turcos e Caicos', 'id': 'Kepulauan Turks dan Caicos', 'ie': 'Turks e Caicos', 'ig': 'Turks & Caicos Islands', 'is': 'Turks- og Caicoseyjar', 'it': 'Isole Turks e Caicos', 'ja': 'タークス・カイコス諸島', 'jmc': 'Visiwa vya Turki na Kaiko', 'jv': 'Turks lan Kapuloan Kaikos', 'ka': 'თერქს-ქაიქოსის კუნძულები', 'kaa': 'Тёркс ҳәм Кайкос атаўлары', 'kaa-Cyrl': 'Тёркс ҳәм Кайкос атаўлары', 'kab': 'Ṭurk d Tegzirin n Kaykus', 'kam': 'Visiwa vya Turki na Kaiko', 'kde': 'Chisiwa cha Tuluchi na Kaiko', 'kea': 'Ilhas Turkas i Kaikus', 'kgp': 'Turka kar Kajko Goj-vẽso Ag', 'khq': 'Turk nda Kayikos Gungey', 'ki': 'Visiwa vya Turki na Kaiko', 'kk': 'Теркс және Кайкос аралдары', 'kk-Cyrl': 'Теркс және Кайкос аралдары', 'kln': 'Ikwembeiyotab Turks ak Caicos', 'km': 'កោះ\u200bទួគ និង កៃកូស', 'kn': 'ಟರ್ಕ್ಸ್ ಮತ್ತು ಕೈಕೋಸ್ ದ್ವೀಪಗಳು', 'ko': '터크스 케이커스 제도', 'kok': 'तुर्क्स आनी कॅकोज आयलँड्स', 'kok-Deva': 'तुर्क्स आनी कॅकोज आयलँड्स', 'ks': 'تُرکس تٕہ کیکو جزیرٕ', 'ks-Arab': 'تُرکس تٕہ کیکو جزیرٕ', 'ksb': 'Visiwa vya Tulki na Kaiko', 'ksf': 'zǝ tirk ri kakɔs', 'ksh': 'de Enselle Turks un Kaikos', 'ku': 'Giravên Turks û Kaîkosê', 'kxv': 'turks aḍe keikes dīp', 'kxv-Deva': 'तुर्क ऑड़े केकोज़ द्वीप', 'kxv-Latn': 'turks aḍe keikes dīp', 'kxv-Orya': 'ତୁର୍କ ଅଡ଼େ କାଇକସ ଦିପ', 'kxv-Telu': 'టర్క్స్ మరియు కైకోస్ దీవులు', 'ky': 'Түркс жана Кайкос аралдары', 'lag': 'Visíiwa vya Turíiki na Kaíiko', 'lb': 'Turks- a Caicosinselen', 'lg': 'Bizinga by’eTaaka ne Kayikosi', 'lld': 'Isoles Turks y Caicos', 'ln': 'Bisanga bya Turki mpé Kaiko', 'lo': 'ໝູ່ເກາະ ເທີກ ແລະ ໄຄໂຄສ', 'lt': 'Terkso ir Kaikoso Salos', 'lu': 'Lutanda lua Tuluki ne Kaiko', 'luo': 'Turks gi Caicos Islands', 'luy': 'Visiwa vya Turki na Kaiko', 'lv': 'Tērksas un Kaikosas salas', 'mai': 'तुर्क आ कैकोज द्वीपसमूह', 'mas': 'Visiwa vya Turki na Kaiko', 'mer': 'Aĩrandi cia Takĩ na Kaiko', 'mfe': 'Zil Tirk ek Caïcos', 'mg': 'Nosy Turks sy Caïques', 'mi': 'Koru-Kākoa', 'mk': 'Острови Туркс и Каикос', 'ml': 'ടർക്ക്\u200cസും കെയ്\u200cക്കോ ദ്വീപുകളും', 'mn': 'Турк ба Кайкосын Арлууд', 'mr': 'टर्क्स आणि कैकोस बेटे', 'ms': 'Kepulauan Turks dan Caicos', 'mt': 'il-Gżejjer Turks u Caicos', 'mua': 'Turkiya ne kaicos ma laŋne', 'my': 'တခ်စ်နှင့်ကာအီကိုစ်ကျွန်းစု', 'mzn': 'تورکس و کایکوس جزایر', 'naq': 'Turks and Caicos Islands', 'nb': 'Turks- og Caicosøyene', 'nd': 'Turks and Caicos Islands', 'ne': 'तुर्क र काइकोस टापु', 'nl': 'Turks- en Caicoseilanden', 'nmg': 'Maŋ́ má Turk na Kaiko', 'nn': 'Turks- og Caicosøyane', 'no': 'Turks- og Caicosøyene', 'nqo': 'ߕߎߙߑߞߌ߫ ߣߌ߫ ߞߊߦߌߞ', 'nyn': 'Ebizinga bya Buturuki na Kaiko', 'om': 'Turkis fi Odoloota Kaayikos', 'or': 'ତୁର୍କସ୍\u200c ଏବଂ କାଇକୋସ୍\u200c ଦ୍ୱୀପପୁଞ୍ଜ', 'pa': 'ਟੁਰਕਸ ਅਤੇ ਕੈਕੋਸ ਟਾਪੂ', 'pa-Guru': 'ਟੁਰਕਸ ਅਤੇ ਕੈਕੋਸ ਟਾਪੂ', 'pcm': 'Tọks an Kaíkọ́s Aílands', 'pl': 'Turks i Caicos', 'ps': 'د ترکیې او کیکاسو ټاپو', 'pt': 'Ilhas Turcas e Caicos', 'qu': 'Islas Turcas y Caicos', 'rm': 'Inslas Turks e Caicos', 'rn': 'Amazinga ya Turkisi na Cayikosi', 'ro': 'Insulele Turks și Caicos', 'rof': 'Visiwa vya Turki na Kaiko', 'ru': 'Тёркс и Кайкос', 'rwk': 'Visiwa vya Turki na Kaiko', 'saq': 'Visiwa vya Turki na Kaiko', 'sbp': 'Figunguli fya Tuliki ni Kaiko', 'sc': 'Ìsulas Turks e Caicos', 'sd': 'ترڪ ۽ ڪيڪوس ٻيٽ', 'sd-Arab': 'ترڪ ۽ ڪيڪوس ٻيٽ', 'se': 'Turks ja Caicos-sullot', 'seh': 'Ilhas Turks e Caicos', 'ses': 'Turk nda Kayikos Gungey', 'sg': 'Âzûâ Turku na Kaîki', 'shi': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵜⵓⵔⴽⵢⴰ ⴷ ⴽⴰⵢⴽ', 'shi-Latn': 'tigzirin n turkya d kayk', 'shi-Tfng': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵜⵓⵔⴽⵢⴰ ⴷ ⴽⴰⵢⴽ', 'si': 'ටර්ක්ස් සහ කයිකොස් දූපත්', 'sk': 'Turks a Caicos', 'sl': 'Otoki Turks in Caicos', 'smn': 'Turks- já Caicossuolluuh', 'sn': 'Zvitsuwa zveTurk neCaico', 'so': 'Turks & Kaikos Island', 'sq': 'Ishujt Turks dhe Kaikos', 'sr': 'Острва Туркс и Каикос', 'sr-Cyrl': 'Острва Туркс и Каикос', 'sr-Latn': 'Ostrva Turks i Kaikos', 'sv': 'Turks- och Caicosöarna', 'sw': 'Visiwa vya Turks na Caicos', 'syr': 'ܓܙܝܖ̈ܐ ܕܬܘܪܟܣ ܘܟܐܝܟܘܣ', 'ta': 'டர்க்ஸ் & கைகோஸ் தீவுகள்', 'te': 'టర్క్స్ మరియు కైకోస్ దీవులు', 'teo': 'Visiwa vya Turki na Kaiko', 'tg': 'Ҷазираҳои Теркс ва Кайкос', 'th': 'หมู่เกาะเติกส์และหมู่เกาะเคคอส', 'ti': 'ደሴታት ቱርካትን ካይኮስን', 'tk': 'Terks we Kaýkos adalary', 'to': 'ʻOtumotu Tuki mo Kaikosi', 'tr': 'Turks ve Caicos Adaları', 'tt': 'Теркс һәм Кайкос утраулары', 'twq': 'Turk nda Kayikos Gungey', 'tzm': 'Tigzirin Turkiyyin d Tikaykusin', 'ug': 'تۇركس ۋە كايكوس ئاراللىرى', 'uk': 'Острови Теркс і Кайкос', 'ur': 'ٹرکس اور کیکوس جزائر', 'uz': 'Turks va Kaykos orollari', 'uz-Cyrl': 'Туркс ва Кайкос ороллари', 'uz-Latn': 'Turks va Kaykos orollari', 'vai': 'ꗋꖺꕃꔻ ꗪ ꕪꔤꖏꔻ ꔳꘋꗣ', 'vai-Latn': 'Tukisi ɓɛ̀ Kaikóosi Tiŋŋɛ', 'vai-Vaii': 'ꗋꖺꕃꔻ ꗪ ꕪꔤꖏꔻ ꔳꘋꗣ', 'vec': 'Turks e Caicos', 'vi': 'Quần đảo Turks và Caicos', 'vun': 'Visiwa vya Turki na Kaiko', 'wae': 'Turks- und Caicosinslä', 'wo': 'Duni Tirk ak Kaykos', 'xh': 'ETurks & Caicos Islands', 'xnr': 'तुर्क कने कैकोज़ द्वीपां दा झुंड', 'xog': 'Ebizinga bya Taaka ni Kayikosi', 'yav': 'túluk na káyiik', 'yo': 'Tọọki ati Etikun Kakọsi', 'yrl': 'Kapuãma-ita Tuka-ita asuí Kaiko-ita', 'yue': '土克斯及開科斯群島', 'yue-Hans': '土克斯及开科斯群岛', 'yue-Hant': '土克斯及開科斯群島', 'zgh': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵏ ⵜⵓⵔⴽⵢⴰ ⴷ ⴽⴰⵢⴽ', 'zh': '特克斯和凯科斯群岛', 'zh-Hans': '特克斯和凯科斯群岛', 'zh-Hant': '土克斯及開科斯群島', 'zu': 'i-Turks ne-Caicos Islands'},
    'TD': {'af': 'Tsjad', 'agq': 'Châ', 'ak': 'Kyad', 'am': 'ቻድ', 'ar': 'تشاد', 'as': 'চাড', 'asa': 'Chadi', 'ast': 'Chad', 'az': 'Çad', 'az-Cyrl': 'Чад', 'az-Latn': 'Çad', 'bal-Latn': 'Chád', 'bas': 'Câd', 'be': 'Чад', 'be-tarask': 'Чад', 'bez': 'Huchadi', 'bg': 'Чад', 'bgn': 'چاد', 'blo': 'Caɖ', 'bm': 'Cadi', 'bn': 'চাদ', 'br': 'Tchad', 'brx': 'चाड', 'bs': 'Čad', 'bs-Cyrl': 'Чад', 'bs-Latn': 'Čad', 'ca': 'Txad', 'ca-ES-valencia': 'Txad', 'ccp': '𑄌𑄘𑄴', 'ce': 'Чад', 'ceb': 'Chad', 'cgg': 'Chadi', 'chr': 'ᏣᏗ', 'ckb': 'چاد', 'cs': 'Čad', 'cv': 'Чад', 'cy': 'Tsiad', 'da': 'Tchad', 'dav': 'Chadi', 'de': 'Tschad', 'dje': 'Caadu', 'dsb': 'Čad', 'dyo': 'Cad', 'dz': 'ཅཱཌ', 'ebu': 'Chadi', 'ee': 'Tsad nutome', 'el': 'Τσαντ', 'el-polyton': 'Τσάντ', 'en': 'Chad', 'en-Dsrt': '𐐕𐐰𐐼', 'eo': 'Ĉado', 'es': 'Chad', 'et': 'Tšaad', 'eu': 'Txad', 'ewo': 'Tsád', 'fa': 'چاد', 'ff': 'Caad', 'ff-Adlm': '𞤕𞤢𞥄𞤣', 'ff-Latn': 'Caad', 'fi': 'Tšad', 'fil': 'Chad', 'fo': 'Kjad', 'fr': 'Tchad', 'fur': 'Çad', 'fy': 'Tsjaad', 'ga': 'Sead', 'gd': 'An t-Seàd', 'gl': 'Chad', 'gsw': 'Tschad', 'gu': 'ચાડ', 'guz': 'Chadi', 'ha': 'Cadi', 'he': 'צ׳אד', 'hi': 'चाड', 'hi-Latn': 'Chad', 'hr': 'Čad', 'hsb': 'Čad', 'ht': 'Tchad', 'hu': 'Csád', 'hy': 'Չադ', 'ia': 'Tchad', 'id': 'Chad', 'ie': 'Tchad', 'ig': 'Chad', 'is': 'Tsjad', 'it': 'Ciad', 'ja': 'チャド', 'jgo': 'Cât', 'jmc': 'Chadi', 'jv': 'Chad', 'ka': 'ჩადი', 'kaa': 'Чад', 'kaa-Cyrl': 'Чад', 'kab': 'Čad', 'kam': 'Chadi', 'kde': 'Chadi', 'kea': 'Txadi', 'kgp': 'Sane', 'khq': 'Caadu', 'ki': 'Chadi', 'kk': 'Чад', 'kk-Cyrl': 'Чад', 'kln': 'Emetab Chad', 'km': 'ឆាដ', 'kn': 'ಚಾದ್', 'ko': '차드', 'kok': 'चाड', 'kok-Deva': 'चाड', 'ks': 'چاڑ', 'ks-Arab': 'چاڑ', 'ksb': 'Chadi', 'ksf': 'caád', 'ksh': 'der Tschadd', 'ku': 'Çad', 'kxv': 'cad', 'kxv-Deva': 'चाड', 'kxv-Latn': 'cad', 'kxv-Orya': 'ଚାଡ', 'kxv-Telu': 'చాద్', 'ky': 'Чад', 'lag': 'Cháadi', 'lb': 'Tschad', 'lg': 'Caadi', 'lld': 'Ciad', 'ln': 'Tsádi', 'lo': 'ຊາດ', 'lt': 'Čadas', 'lu': 'Tshadi', 'luo': 'Chad', 'luy': 'Chadi', 'lv': 'Čada', 'mai': 'चाड', 'mas': 'Chadi', 'mer': 'Chadi', 'mfe': 'Tchad', 'mg': 'Tsady', 'mgh': 'Uchadi', 'mi': 'Kāta', 'mk': 'Чад', 'ml': 'ഛാഡ്', 'mn': 'Чад', 'mr': 'चाड', 'ms': 'Chad', 'mt': 'iċ-Chad', 'mua': 'syad', 'my': 'ချဒ်', 'mzn': 'چاد', 'naq': 'Chadi', 'nb': 'Tsjad', 'nd': 'Chadi', 'ne': 'चाड', 'nl': 'Tsjaad', 'nmg': 'Tshad', 'nn': 'Tsjad', 'no': 'Tsjad', 'nqo': 'ߗߊߘ', 'nus': 'Ca̱d', 'nyn': 'Chadi', 'om': 'Chaad', 'or': 'ଚାଦ୍', 'pa': 'ਚਾਡ', 'pa-Guru': 'ਚਾਡ', 'pcm': 'Chád', 'pl': 'Czad', 'ps': 'چاډ', 'pt': 'Chade', 'qu': 'Chad', 'rm': 'Tschad', 'rn': 'Cadi', 'ro': 'Ciad', 'rof': 'Chadi', 'ru': 'Чад', 'rwk': 'Chadi', 'saq': 'Chadi', 'sbp': 'Shadi', 'sc': 'Chad', 'sd': 'چاڊ', 'sd-Arab': 'چاڊ', 'se': 'Tčad', 'seh': 'Chade', 'ses': 'Caadu', 'sg': 'Tyâde', 'shi': 'ⵜⵛⴰⴷ', 'shi-Latn': 'tcad', 'shi-Tfng': 'ⵜⵛⴰⴷ', 'si': 'චැච්', 'sk': 'Čad', 'sl': 'Čad', 'smn': 'Tšad', 'sn': 'Chadi', 'so': 'Jaad', 'sq': 'Çad', 'sr': 'Чад', 'sr-Cyrl': 'Чад', 'sr-Latn': 'Čad', 'sv': 'Tchad', 'sw': 'Chad', 'syr': 'ܬܫܐܕ', 'ta': 'சாட்', 'te': 'చాద్', 'teo': 'Chadi', 'tg': 'Чад', 'th': 'ชาด', 'ti': 'ቻድ', 'tk': 'Çad', 'to': 'Sāti', 'tr': 'Çad', 'tt': 'Чад', 'twq': 'Caadu', 'tzm': 'Tcad', 'ug': 'چاد', 'uk': 'Чад', 'ur': 'چاڈ', 'uz': 'Chad', 'uz-Cyrl': 'Чад', 'uz-Latn': 'Chad', 'vai': 'ꕦꔵ', 'vai-Latn': 'Chádi', 'vai-Vaii': 'ꕦꔵ', 'vec': 'Ciad', 'vi': 'Chad', 'vun': 'Chadi', 'wae': 'Tšad', 'wo': 'Càdd', 'xh': 'EChad', 'xnr': 'चाड', 'xog': 'Caadi', 'yav': 'Sáat', 'yi': 'טשאַד', 'yo': 'Ṣààdì', 'yrl': 'Xade', 'yue': '查德', 'yue-Hans': '查德', 'yue-Hant': '查德', 'zgh': 'ⵜⵛⴰⴷ', 'zh': '乍得', 'zh-Hans': '乍得', 'zh-Hant': '查德', 'zu': 'i-Chad'},
    'TF': {'af': 'Franse Suidelike Gebiede', 'ak': 'Franse Anaafoɔ Nsaase', 'am': 'የፈረንሳይ ደቡባዊ ግዛቶች', 'ar': 'الأقاليم الجنوبية الفرنسية', 'as': 'দক্ষিণ ফ্ৰান্সৰ অঞ্চল', 'ast': 'Tierres Australes Franceses', 'az': 'Fransanın Cənub Əraziləri', 'az-Cyrl': 'Франсанын Ҹәнуб Әразиләри', 'az-Latn': 'Fransanın Cənub Əraziləri', 'bal-Latn': 'Pránsi Zerbári Damag', 'be': 'Французскія паўднёвыя тэрыторыі', 'be-tarask': 'Французскія паўднёвыя тэрыторыі', 'bg': 'Френски южни територии', 'blo': 'Gafɔntǝna kaBʊtǝna gʊnyɩsonolaŋ kabʊja', 'bn': 'ফরাসী দক্ষিণাঞ্চল', 'br': 'Douaroù aostral Frañs', 'brx': 'फ्रान्सनि खोलायारि बाहागोफोर', 'bs': 'Francuske Južne Teritorije', 'bs-Cyrl': 'Француске Јужне Територије', 'bs-Latn': 'Francuske Južne Teritorije', 'ca': 'Terres Australs Antàrtiques Franceses', 'ca-ES-valencia': 'Terres Australs Antàrtiques Franceses', 'ccp': '𑄜𑄢𑄥𑄩 𑄘𑄧𑄉𑄨𑄚𑄧 𑄎𑄉', 'ce': 'Французийн къилба латтанаш', 'ceb': 'French Southern Territories', 'chr': 'ᎠᏂᎦᎸᏥ ᏧᎦᎾᏮ ᎦᏙᎯ ᎤᎵᏍᏛᎢ', 'ckb': 'هەرێمە باشووریەکانی فەرەنسا', 'cs': 'Francouzská jižní území', 'cv': 'Франци Кӑнтӑр территорийӗсем', 'cy': 'Tiroedd Deheuol ac Antarctig Ffrainc', 'da': 'De Franske Besiddelser i Det Sydlige Indiske Ocean og Antarktis', 'de': 'Französische Süd- und Antarktisgebiete', 'dsb': 'Francojski pódpołdnjowy a antarktiski teritorium', 'dz': 'ཕྲནཅ་གི་ལྷོ་ཕྱོགས་མངའ་ཁོངས', 'ee': 'Anyiehe Franseme nutome', 'el': 'Γαλλικά Νότια Εδάφη', 'el-polyton': 'Γαλλικὰ Νότια Ἐδάφη', 'en': 'French Southern Territories', 'en-Dsrt': '𐐙𐑉𐐯𐑌𐐽 𐐝𐐲𐑄𐐲𐑉𐑌 𐐓𐐯𐑉𐐲𐐻𐐱𐑉𐐨𐑆', 'es': 'Territorios Australes Franceses', 'et': 'Prantsuse Lõunaalad', 'eu': 'Hegoaldeko lurralde frantsesak', 'fa': 'سرزمین\u200cهای جنوبی فرانسه', 'ff-Adlm': '𞤚𞤵𞤥𞤦𞤫 𞤂𞤫𞤧𞤤𞤫𞤴𞤶𞤫 𞤊𞤪𞤢𞤲𞤧𞤭', 'fi': 'Ranskan eteläiset ja antarktiset alueet', 'fil': 'French Southern Territories', 'fo': 'Fronsku sunnaru landaøki', 'fr': 'Terres australes françaises', 'fur': 'Teritoris meridionâi francês', 'fy': 'Franse Gebieden in de zuidelijke Indyske Oseaan', 'ga': 'Críocha Francacha Dheisceart an Domhain', 'gd': 'Ranntairean a Deas na Frainge', 'gl': 'Territorios Austrais Franceses', 'gsw': 'Französischi Süüd- und Antarktisgebiet', 'gu': 'ફ્રેંચ સધર્ન ટેરિટરીઝ', 'ha': 'Yankin Faransi ta Kudu', 'he': 'הטריטוריות הדרומיות של צרפת', 'hi': 'फ़्रांसीसी दक्षिणी क्षेत्र', 'hi-Latn': 'French Southern Territories', 'hr': 'Francuski Južni Teritoriji', 'hsb': 'Francoski južny a antarktiski teritorij', 'ht': 'Terres australes françaises', 'hu': 'Francia Déli Területek', 'hy': 'Ֆրանսիական Հարավային Տարածքներ', 'ia': 'Territorios meridional francese', 'id': 'Wilayah Selatan Prancis', 'ig': 'Ụmụ ngalaba Frenchi Southern', 'is': 'Frönsku suðlægu landsvæðin', 'it': 'Terre Australi Francesi', 'ja': '仏領極南諸島', 'jv': 'Wilayah Prancis nang Kutub Kidul', 'ka': 'ფრანგული სამხრეთის ტერიტორიები', 'kaa': 'Француз Қубла территориялары', 'kaa-Cyrl': 'Француз Қубла территориялары', 'kea': 'Terras Franses di Sul', 'kgp': 'Sur Frỹsa Tehitórijo', 'kk': 'Францияның оңтүстік аймақтары', 'kk-Cyrl': 'Францияның оңтүстік аймақтары', 'km': 'ដែនដី\u200bបារាំង\u200bនៅ\u200bភាគខាងត្បូង', 'kn': 'ಫ್ರೆಂಚ್ ದಕ್ಷಿಣ ಪ್ರದೇಶಗಳು', 'ko': '프랑스령 남방 지역', 'kok': 'फ्रेंच दक्षिणी प्रांत', 'kok-Deva': 'फ्रेंच दक्षिणी प्रांत', 'ks': 'فرانسِسی جَنوٗبی عَلاقہٕ', 'ks-Arab': 'فرانسِسی جَنوٗبی عَلاقہٕ', 'ksh': 'Franzüüsesche Södsee-Enselle', 'ku': 'Herêmên Başûr ên Fransayê', 'kxv': 'prēnch dakiṇ teritorī', 'kxv-Deva': 'प्रेंच दकिनी टेरिटोरी', 'kxv-Latn': 'prēnch dakiṇ teritorī', 'kxv-Orya': 'ପ୍ରେଞ୍ଚ ଦକିନୀୟ ଟେରିଟୋରୀ', 'kxv-Telu': 'ప్రెంచ్ దకినియ టెరిటొరి', 'ky': 'Франциянын Түштүктөгү аймактары', 'lb': 'Franséisch Süd- an Antarktisgebidder', 'lld': 'Teritori franzesc dl Süd y dl’Antartica', 'ln': 'Terres australes et antarctiques françaises', 'lo': 'ເຂດແດນທາງໃຕ້ຂອງຝຮັ່ງ', 'lt': 'Prancūzijos Pietų sritys', 'lv': 'Francijas Dienvidjūru teritorija', 'mai': 'फ्रांसीसी दक्षिणी क्षेत्र', 'mi': 'Ngā Rohe o Wīwī ki te Tonga', 'mk': 'Француски Јужни Територии', 'ml': 'ഫ്രഞ്ച് ദക്ഷിണ ഭൂപ്രദേശം', 'mn': 'Францын өмнөд газар нутаг', 'mr': 'फ्रेंच दाक्षिणात्य प्रदेश', 'ms': 'Wilayah Selatan Perancis', 'mt': 'It-Territorji Franċiżi tan-Nofsinhar', 'my': 'ပြင်သစ် တောင်ပိုင်း ပိုင်နက်များ', 'mzn': 'فرانسه\u200cی جنوبی مناطق', 'nb': 'De franske sørterritorier', 'ne': 'फ्रेन्च दक्षिणी राज्यहरू', 'nl': 'Franse Gebieden in de zuidelijke Indische Oceaan', 'nn': 'Dei franske sørterritoria', 'no': 'De franske sørterritorier', 'nqo': 'ߝߊ߬ߙߊ߲߬ߛߌ߫ ߘߎ߰ߞߟߏ ߓߊߙߌ ߘߐ߫', 'om': 'Daangaawwan Kibbaa Faransaay', 'or': 'ଫରାସୀ ଦକ୍ଷିଣ କ୍ଷେତ୍ର', 'pa': 'ਫਰੈਂਚ ਦੱਖਣੀ ਪ੍ਰਦੇਸ਼', 'pa-Guru': 'ਫਰੈਂਚ ਦੱਖਣੀ ਪ੍ਰਦੇਸ਼', 'pcm': 'Frẹ́nch Saútán Tẹ́rẹ́tris', 'pl': 'Francuskie Terytoria Południowe i Antarktyczne', 'ps': 'د فرانسې جنوبي سیمې', 'pt': 'Territórios Franceses do Sul', 'qu': 'Territorios Australes Franceses', 'rm': 'Territoris Franzos Meridiunals', 'ro': 'Teritoriile Australe și Antarctice Franceze', 'ru': 'Французские Южные территории', 'sc': 'Terras australes frantzesas', 'sd': 'فرانسيسي ڏاکڻي علائقا', 'sd-Arab': 'فرانسيسي ڏاکڻي علائقا', 'si': 'දකුණු ප්\u200dරංශ දූපත් සමූහය', 'sk': 'Francúzske južné a antarktické územia', 'sl': 'Francosko južno ozemlje', 'smn': 'Ranska máddááh kuávluh', 'so': 'Dhul xadeedka Koonfureed ee Faransiiska', 'sq': 'Territoret Jugore Franceze', 'sr': 'Француске Јужне Територије', 'sr-Cyrl': 'Француске Јужне Територије', 'sr-Latn': 'Francuske Južne Teritorije', 'sv': 'Franska sydterritorierna', 'sw': 'Himaya za Kusini za Kifaranza', 'syr': 'ܩܠܝܡ̈ܐ ܕܦܪܢܣܐ ܬܝܡܢܝܬܐ', 'ta': 'பிரெஞ்சு தெற்கு பிரதேசங்கள்', 'te': 'ఫ్రెంచ్ దక్షిణ ప్రాంతాలు', 'tg': 'Минтақаҳои Ҷанубии Фаронса', 'th': 'เฟรนช์เซาเทิร์นเทร์ริทอรีส์', 'ti': 'ፈረንሳዊ ደቡባዊ ግዝኣታት', 'tk': 'Fransuz günorta territoriýalary', 'to': 'Potu fonua tonga fakafalanisē', 'tr': 'Fransız Güney Toprakları', 'tt': 'Франциянең Көньяк Территорияләре', 'ug': 'فىرانسىيەنىڭ جەنۇبىي زېمىنى', 'uk': 'Французькі Південні Території', 'ur': 'فرانسیسی جنوبی خطے', 'uz': 'Fransuz Janubiy hududlari', 'uz-Cyrl': 'Француз жанубий ҳудудлари', 'uz-Latn': 'Fransuz Janubiy hududlari', 'vai': 'ꔱꗷꘋꔻ ꗛꔤ ꔒꘋꗣ ꗏ ꕸꖃꔀ ꖸ', 'vai-Vaii': 'ꔱꗷꘋꔻ ꗛꔤ ꔒꘋꗣ ꗏ ꕸꖃꔀ ꖸ', 'vec': 'Teritori fransezi de’l sud', 'vi': 'Lãnh thổ phía Nam Thuộc Pháp', 'wae': 'Französiši Süd- und Antarktisgebiet', 'wo': 'Teer Ostraal gu Fraas', 'xh': 'EFrench Southern Territories', 'xnr': 'फ़्रांसीसी दखणे बखें लाका', 'yo': 'Agbègbè Gúúsù Faranṣé', 'yrl': 'Tetãma Su-ita Frãsa suí', 'yue': '法屬南方屬地', 'yue-Hans': '法属南方属地', 'yue-Hant': '法屬南方屬地', 'zh': '法属南部领地', 'zh-Hans': '法属南部领地', 'zh-Hant': '法屬南部屬地', 'zu': 'i-French Southern Territories'},
    'TG': {'af': 'Togo', 'agq': 'Tugù', 'ak': 'Togo', 'am': 'ቶጐ', 'ar': 'توغو', 'as': 'টোগো', 'asa': 'Togo', 'ast': 'Togu', 'az': 'Toqo', 'az-Cyrl': 'Того', 'az-Latn': 'Toqo', 'bal-Latn': 'Tógó', 'bas': 'Tògo', 'be': 'Тога', 'be-tarask': 'Тога', 'bez': 'Hutogo', 'bg': 'Того', 'blo': 'Togoo', 'bm': 'Togo', 'bn': 'টোগো', 'br': 'Togo', 'brx': 'टग’', 'bs': 'Togo', 'bs-Cyrl': 'Того', 'bs-Latn': 'Togo', 'ca': 'Togo', 'ca-ES-valencia': 'Togo', 'ccp': '𑄑𑄮𑄉𑄮', 'ce': 'Того', 'ceb': 'Togo', 'cgg': 'Togo', 'chr': 'ᏙᎪ', 'ckb': 'تۆگۆ', 'cs': 'Togo', 'cv': 'Того', 'cy': 'Togo', 'da': 'Togo', 'dav': 'Togo', 'de': 'Togo', 'dje': 'Togo', 'dsb': 'Togo', 'dyo': 'Togo', 'dz': 'ཊོ་གྷོ', 'ebu': 'Togo', 'ee': 'Togo nutome', 'el': 'Τόγκο', 'el-polyton': 'Τόγκο', 'en': 'Togo', 'en-Dsrt': '𐐓𐐬𐑀𐐬', 'eo': 'Togolando', 'es': 'Togo', 'et': 'Togo', 'eu': 'Togo', 'ewo': 'Togó', 'fa': 'توگو', 'ff': 'Togoo', 'ff-Adlm': '𞤚𞤮𞤺𞤮', 'ff-Latn': 'Togoo', 'fi': 'Togo', 'fil': 'Togo', 'fo': 'Togo', 'fr': 'Togo', 'fur': 'Togo', 'fy': 'Togo', 'ga': 'Tóga', 'gd': 'Togo', 'gl': 'Togo', 'gsw': 'Toogo', 'gu': 'ટોગો', 'guz': 'Togo', 'ha': 'Togo', 'he': 'טוגו', 'hi': 'टोगो', 'hi-Latn': 'Togo', 'hr': 'Togo', 'hsb': 'Togo', 'ht': 'Togo', 'hu': 'Togo', 'hy': 'Տոգո', 'ia': 'Togo', 'id': 'Togo', 'ig': 'Togo', 'is': 'Tógó', 'it': 'Togo', 'ja': 'トーゴ', 'jgo': 'Tɔ́ŋgɔ', 'jmc': 'Togo', 'jv': 'Togo', 'ka': 'ტოგო', 'kaa': 'Того', 'kaa-Cyrl': 'Того', 'kab': 'Ṭugu', 'kam': 'Togo', 'kde': 'Togo', 'kea': 'Togu', 'kgp': 'Togo', 'khq': 'Togo', 'ki': 'Togo', 'kk': 'Того', 'kk-Cyrl': 'Того', 'kln': 'Emetab Togo', 'km': 'តូហ្គោ', 'kn': 'ಟೋಗೋ', 'ko': '토고', 'kok': 'टोगो', 'kok-Deva': 'टोगो', 'ks': 'ٹوگو', 'ks-Arab': 'ٹوگو', 'ksb': 'Togo', 'ksf': 'togo', 'ksh': 'Toojo', 'ku': 'Togo', 'kxv': 'ṭogo', 'kxv-Deva': 'टोगो', 'kxv-Latn': 'ṭogo', 'kxv-Orya': 'ଟଗ', 'kxv-Telu': 'టోగో', 'ky': 'Того', 'lag': 'Tóogo', 'lb': 'Togo', 'lg': 'Togo', 'lld': 'Togo', 'ln': 'Togo', 'lo': 'ໂຕໂກ', 'lt': 'Togas', 'lu': 'Togu', 'luo': 'Togo', 'luy': 'Togo', 'lv': 'Togo', 'mai': 'टोगो', 'mas': 'Togo', 'mer': 'Togo', 'mfe': 'Togo', 'mg': 'Togo', 'mgh': 'Utogo', 'mi': 'Toko', 'mk': 'Того', 'ml': 'ടോഗോ', 'mn': 'Того', 'mr': 'टोगो', 'ms': 'Togo', 'mt': 'it-Togo', 'mua': 'Sǝr Togo', 'my': 'တိုဂို', 'mzn': 'توگو', 'naq': 'Togob', 'nb': 'Togo', 'nd': 'Thogo', 'ne': 'टोगो', 'nl': 'Togo', 'nmg': 'Togo', 'nn': 'Togo', 'no': 'Togo', 'nqo': 'ߕߜ߭ߏ߫', 'nyn': 'Togo', 'om': 'Toogoo', 'or': 'ଟୋଗୋ', 'pa': 'ਟੋਗੋ', 'pa-Guru': 'ਟੋਗੋ', 'pcm': 'Tógo', 'pl': 'Togo', 'ps': 'ټوګو', 'pt': 'Togo', 'qu': 'Togo', 'rm': 'Togo', 'rn': 'Togo', 'ro': 'Togo', 'rof': 'Togo', 'ru': 'Того', 'rwk': 'Togo', 'saq': 'Togo', 'sbp': 'Togo', 'sc': 'Togo', 'sd': 'ٽوگو', 'sd-Arab': 'ٽوگو', 'se': 'Togo', 'seh': 'Togo', 'ses': 'Togo', 'sg': 'Togö', 'shi': 'ⵟⵓⴳⵓ', 'shi-Latn': 'ṭugu', 'shi-Tfng': 'ⵟⵓⴳⵓ', 'si': 'ටොගෝ', 'sk': 'Togo', 'sl': 'Togo', 'smn': 'Togo', 'sn': 'Togo', 'so': 'Toogo', 'sq': 'Togo', 'sr': 'Того', 'sr-Cyrl': 'Того', 'sr-Latn': 'Togo', 'sv': 'Togo', 'sw': 'Togo', 'syr': 'ܬܘܓܘ', 'ta': 'டோகோ', 'te': 'టోగో', 'teo': 'Togo', 'tg': 'Того', 'th': 'โตโก', 'ti': 'ቶጎ', 'tk': 'Togo', 'to': 'Toko', 'tr': 'Togo', 'tt': 'Того', 'twq': 'Togo', 'tzm': 'Ṭṭugu', 'ug': 'توگو', 'uk': 'Того', 'ur': 'ٹوگو', 'uz': 'Togo', 'uz-Cyrl': 'Того', 'uz-Latn': 'Togo', 'vai': 'ꕿꖑ', 'vai-Latn': 'Togo', 'vai-Vaii': 'ꕿꖑ', 'vec': 'Togo', 'vi': 'Togo', 'vun': 'Togo', 'wae': 'Togo', 'wo': 'Togo', 'xh': 'ETogo', 'xnr': 'टोगो', 'xog': 'Togo', 'yav': 'tokó', 'yi': 'טאגא', 'yo': 'Togo', 'yrl': 'Togu', 'yue': '多哥', 'yue-Hans': '多哥', 'yue-Hant': '多哥', 'zgh': 'ⵟⵓⴳⵓ', 'zh': '多哥', 'zh-Hans': '多哥', 'zh-Hant': '多哥', 'zu': 'i-Togo'},
    'TH': {'af': 'Thailand', 'agq': 'Taelàn', 'ak': 'Taeland', 'am': 'ታይላንድ', 'ar': 'تايلاند', 'as': 'থাইলেণ্ড', 'asa': 'Tailandi', 'ast': 'Tailandia', 'az': 'Tailand', 'az-Cyrl': 'Таиланд', 'az-Latn': 'Tailand', 'bal-Latn': 'Táilánd', 'bas': 'Taylànd', 'be': 'Тайланд', 'be-tarask': 'Тайланд', 'bez': 'Hutailandi', 'bg': 'Тайланд', 'bgn': 'ٹایلنڈ', 'blo': 'Taɩlanɖ', 'bm': 'Tayilandi', 'bn': 'থাইল্যান্ড', 'br': 'Thailand', 'brx': 'थाइलेण्ड', 'bs': 'Tajland', 'bs-Cyrl': 'Тајланд', 'bs-Latn': 'Tajland', 'ca': 'Tailàndia', 'ca-ES-valencia': 'Tailàndia', 'ccp': '𑄗𑄭𑄣𑄳𑄠𑄚𑄴𑄓𑄴', 'ce': 'Таиланд', 'ceb': 'Thailand', 'cgg': 'Tairandi', 'chr': 'ᏔᏯᎴᏂ', 'ckb': 'تایلەند', 'cs': 'Thajsko', 'cv': 'Таиланд', 'cy': 'Gwlad Thai', 'da': 'Thailand', 'dav': 'Tailandi', 'de': 'Thailand', 'dje': 'Taayiland', 'dsb': 'Thailandska', 'dyo': 'Tailand', 'dz': 'ཐཱའི་ལེནཌ', 'ebu': 'Tailandi', 'ee': 'Thailand nutome', 'el': 'Ταϊλάνδη', 'el-polyton': 'Ταϊλάνδη', 'en': 'Thailand', 'en-Dsrt': '𐐓𐐴𐑊𐐰𐑌𐐼', 'eo': 'Tajlando', 'es': 'Tailandia', 'et': 'Tai', 'eu': 'Thailandia', 'ewo': 'Tailán', 'fa': 'تایلند', 'ff': 'Taylannda', 'ff-Adlm': '𞤚𞤢𞥄𞤴𞤤𞤢𞤲𞤣', 'ff-Latn': 'Taylannda', 'fi': 'Thaimaa', 'fil': 'Thailand', 'fo': 'Tailand', 'fr': 'Thaïlande', 'fur': 'Tailandie', 'fy': 'Thailân', 'ga': 'an Téalainn', 'gd': 'Dùthaich nan Tàidh', 'gl': 'Tailandia', 'gsw': 'Thailand', 'gu': 'થાઇલેંડ', 'guz': 'Tailandi', 'ha': 'Tailan', 'he': 'תאילנד', 'hi': 'थाईलैंड', 'hi-Latn': 'Thailand', 'hr': 'Tajland', 'hsb': 'Thailandska', 'ht': 'Thaïlande', 'hu': 'Thaiföld', 'hy': 'Թաիլանդ', 'ia': 'Thailandia', 'id': 'Thailand', 'ig': 'Thailand', 'is': 'Taíland', 'it': 'Thailandia', 'ja': 'タイ', 'jmc': 'Tailandi', 'jv': 'Tanah Thai', 'ka': 'ტაილანდი', 'kaa': 'Таиланд', 'kaa-Cyrl': 'Таиланд', 'kab': 'Ṭayland', 'kam': 'Tailandi', 'kde': 'Tailandi', 'kea': 'Tailándia', 'kgp': 'Taj-Ga', 'khq': 'Taayiland', 'ki': 'Tailandi', 'kk': 'Таиланд', 'kk-Cyrl': 'Таиланд', 'kln': 'Emetab Thailand', 'km': 'ថៃ', 'kn': 'ಥೈಲ್ಯಾಂಡ್', 'ko': '태국', 'kok': 'थायलँड', 'kok-Deva': 'थायलँड', 'ks': 'تھائی لینڈ', 'ks-Arab': 'تھائی لینڈ', 'ksb': 'Tailandi', 'ksf': 'tɛlan', 'ksh': 'Tailand', 'ku': 'Tayland', 'kxv': 'tailanḍ', 'kxv-Deva': 'ताईलेंड', 'kxv-Latn': 'tailanḍ', 'kxv-Orya': 'ତାଇଲାଣ୍ଡ', 'kxv-Telu': 'తయిలాండ్', 'ky': 'Тайланд', 'lag': 'Táilandi', 'lb': 'Thailand', 'lg': 'Tayirandi', 'lld': 'Thailandia', 'ln': 'Tailandɛ', 'lo': 'ໄທ', 'lt': 'Tailandas', 'lu': 'Tayilanda', 'luo': 'Thailand', 'luy': 'Tailandi', 'lv': 'Taizeme', 'mai': 'थाईलैंड', 'mas': 'Tailandi', 'mer': 'Thaĩlandi', 'mfe': 'Thayland', 'mg': 'Thailandy', 'mgh': 'Utailandi', 'mi': 'Tairanga', 'mk': 'Тајланд', 'ml': 'തായ്\u200cലാൻഡ്', 'mn': 'Тайланд', 'mr': 'थायलंड', 'ms': 'Thailand', 'mt': 'it-Tajlandja', 'mua': 'Tailand', 'my': 'ထိုင်း', 'mzn': 'تایلند', 'naq': 'Thailandi', 'nb': 'Thailand', 'nd': 'Thayilandi', 'ne': 'थाइल्याण्ड', 'nl': 'Thailand', 'nmg': 'Taïland', 'nn': 'Thailand', 'no': 'Thailand', 'nqo': 'ߕߊߦߌߘߎ߯', 'nyn': 'Tairandi', 'om': 'Taayilaand', 'or': 'ଥାଇଲ୍ୟାଣ୍ଡ', 'pa': 'ਥਾਈਲੈਂਡ', 'pa-Guru': 'ਥਾਈਲੈਂਡ', 'pcm': 'Taíland', 'pl': 'Tajlandia', 'ps': 'تهايلنډ', 'pt': 'Tailândia', 'qu': 'Tailandia', 'rm': 'Tailanda', 'rn': 'Tayilandi', 'ro': 'Thailanda', 'rof': 'Tailandi', 'ru': 'Таиланд', 'rwk': 'Tailandi', 'saq': 'Tailandi', 'sbp': 'Tailandi', 'sc': 'Tailàndia', 'sd': 'ٿائيليند', 'sd-Arab': 'ٿائيليند', 'se': 'Thaieana', 'seh': 'Tailândia', 'ses': 'Taayiland', 'sg': 'Tailânde', 'shi': 'ⵟⴰⵢⵍⴰⵏⴷ', 'shi-Latn': 'ṭayland', 'shi-Tfng': 'ⵟⴰⵢⵍⴰⵏⴷ', 'shn': 'မိူင်းထႆး', 'si': 'තායිලන්තය', 'sk': 'Thajsko', 'sl': 'Tajska', 'smn': 'Thaieennâm', 'sn': 'Thailand', 'so': 'Taylaand', 'sq': 'Tajlandë', 'sr': 'Тајланд', 'sr-Cyrl': 'Тајланд', 'sr-Latn': 'Tajland', 'sv': 'Thailand', 'sw': 'Tailandi', 'syr': 'ܬܐܝܠܢܕ', 'ta': 'தாய்லாந்து', 'te': 'థాయిలాండ్', 'teo': 'Tailandi', 'tg': 'Таиланд', 'th': 'ไทย', 'ti': 'ታይላንድ', 'tk': 'Taýland', 'to': 'Tailani', 'tr': 'Tayland', 'tt': 'Тайланд', 'twq': 'Taayiland', 'tzm': 'Ṭaylanḍa', 'ug': 'تايلاند', 'uk': 'Таїланд', 'ur': 'تھائی لینڈ', 'uz': 'Tailand', 'uz-Cyrl': 'Таиланд', 'uz-Latn': 'Tailand', 'vai': 'ꕚꔤ ꖨꕮꕊ', 'vai-Latn': 'Tai Lumaã', 'vai-Vaii': 'ꕚꔤ ꖨꕮꕊ', 'vec': 'Tailandia', 'vi': 'Thái Lan', 'vun': 'Tailandi', 'wae': 'Thailand', 'wo': 'Taylànd', 'xh': 'EThailand', 'xnr': 'थाईलैंड', 'xog': 'Tayirandi', 'yav': 'tayiláand', 'yi': 'טיילאַנד', 'yo': 'Tailandi', 'yrl': 'Tairetãma', 'yue': '泰國', 'yue-Hans': '泰国', 'yue-Hant': '泰國', 'zgh': 'ⵟⴰⵢⵍⴰⵏⴷ', 'zh': '泰国', 'zh-Hans': '泰国', 'zh-Hant': '泰國', 'zu': 'i-Thailand'},
    'TJ': {'af': 'Tadjikistan', 'agq': 'Tàdzikìsɨ̀tân', 'ak': 'Tagyikistan', 'am': 'ታጃኪስታን', 'ar': 'طاجيكستان', 'as': 'তাজিকিস্তান', 'asa': 'Tajikithtani', 'ast': 'Taxiquistán', 'az': 'Tacikistan', 'az-Cyrl': 'Таҹикистан', 'az-Latn': 'Tacikistan', 'bal-Latn': 'Tájekestán', 'bas': 'Tàjìkìstaŋ', 'be': 'Таджыкістан', 'be-tarask': 'Таджыкістан', 'bez': 'Hutajikistani', 'bg': 'Таджикистан', 'bgn': 'تاجیکستان', 'blo': 'Tajikistan', 'bm': 'Tajikisitani', 'bn': 'তাজিকিস্তান', 'br': 'Tadjikistan', 'brx': 'ताजिकिस्तान', 'bs': 'Tadžikistan', 'bs-Cyrl': 'Таџикистан', 'bs-Latn': 'Tadžikistan', 'ca': 'Tadjikistan', 'ca-ES-valencia': 'Tadjikistan', 'ccp': '𑄖𑄎𑄨𑄇𑄴𑄥𑄳𑄗𑄚𑄴', 'ce': 'Таджикистан', 'ceb': 'Tajikistan', 'cgg': 'Tajikisitani', 'chr': 'ᏔᏥᎩᏍᏕᏂ', 'ckb': 'تاجیکستان', 'cs': 'Tádžikistán', 'cv': 'Таджикистан', 'cy': 'Tajicistan', 'da': 'Tadsjikistan', 'dav': 'Tajikistani', 'de': 'Tadschikistan', 'dje': 'Taažikistan', 'dsb': 'Tadźikistan', 'dz': 'ཏ་ཇིག་གི་སཏཱན', 'ebu': 'Tajikistani', 'ee': 'Tajikistan nutome', 'el': 'Τατζικιστάν', 'el-polyton': 'Τατζικιστάν', 'en': 'Tajikistan', 'en-Dsrt': '𐐓𐐲𐐾𐐨𐐿𐐲𐑅𐐻𐐰𐑌', 'eo': 'Taĝikujo', 'es': 'Tayikistán', 'et': 'Tadžikistan', 'eu': 'Tajikistan', 'ewo': 'Tadzikisətáŋ', 'fa': 'تاجیکستان', 'ff': 'Tajikistaan', 'ff-Adlm': '𞤚𞤢𞤶𞤭𞤳𞤭𞤧𞤼𞤢𞥄𞤲', 'ff-Latn': 'Tajikistaan', 'fi': 'Tadžikistan', 'fil': 'Tajikistan', 'fo': 'Tadsjikistan', 'fr': 'Tadjikistan', 'fur': 'Tazikistan', 'fy': 'Tadzjikistan', 'ga': 'an Táidsíceastáin', 'gd': 'Taidigeastàn', 'gl': 'Taxiquistán', 'gsw': 'Tadschikischtan', 'gu': 'તાજીકિસ્તાન', 'guz': 'Tajikistani', 'ha': 'Tajikistan', 'he': 'טג׳יקיסטן', 'hi': 'ताजिकिस्तान', 'hi-Latn': 'Tajikistan', 'hr': 'Tadžikistan', 'hsb': 'Tadźikistan', 'ht': 'Tadjikistan', 'hu': 'Tádzsikisztán', 'hy': 'Տաջիկստան', 'ia': 'Tadzhikistan', 'id': 'Tajikistan', 'ig': 'Tajikistan', 'is': 'Tadsíkistan', 'it': 'Tagikistan', 'ja': 'タジキスタン', 'jmc': 'Tajikistani', 'jv': 'Tajikistan', 'ka': 'ტაჯიკეთი', 'kaa': 'Тәжикстан', 'kaa-Cyrl': 'Тәжикстан', 'kab': 'Tajikistan', 'kam': 'Tajikistani', 'kde': 'Tadikistani', 'kea': 'Tadjikistan', 'kgp': 'Tanijikinhtỹv', 'khq': 'Taažikistan', 'ki': 'Tajikistani', 'kk': 'Тәжікстан', 'kk-Cyrl': 'Тәжікстан', 'kln': 'Emetab Tajikistan', 'km': 'តាហ្ស៊ីគីស្ថាន', 'kn': 'ತಜಿಕಿಸ್ತಾನ್', 'ko': '타지키스탄', 'kok': 'तजीकिस्तान', 'kok-Deva': 'तजीकिस्तान', 'ks': 'تاجکِستان', 'ks-Arab': 'تاجکِستان', 'ksb': 'Tajikistani', 'ksf': 'tadjikistaŋ', 'ksh': 'Tadschikistahn', 'ku': 'Tacîkistan', 'kxv': 'tajākīstān', 'kxv-Deva': 'तजाकिस्तान', 'kxv-Latn': 'tajākīstān', 'kxv-Orya': 'ତାଜିକିସ୍ତାନ', 'kxv-Telu': 'తాజాకిస్తాన', 'ky': 'Тажикстан', 'lag': 'Tajikisitáani', 'lb': 'Tadschikistan', 'lg': 'Tajikisitaani', 'lld': 'Tajikistan', 'ln': 'Tazikisitá', 'lo': 'ທາຈິກິດສະຖານ', 'lt': 'Tadžikija', 'lu': 'Tazikisita', 'luo': 'Tajikistan', 'luy': 'Tajikistani', 'lv': 'Tadžikistāna', 'mai': 'ताजिकिस्तान', 'mas': 'Tajikistani', 'mer': 'Tajikistani', 'mfe': 'Tadjikistan', 'mg': 'Tajikistan', 'mgh': 'Ujikistani', 'mi': 'Takiritānga', 'mk': 'Таџикистан', 'ml': 'താജിക്കിസ്ഥാൻ', 'mn': 'Тажикистан', 'mr': 'ताजिकिस्तान', 'ms': 'Tajikistan', 'mt': 'it-Taġikistan', 'mua': 'Tajikistaŋ', 'my': 'တာဂျီကစ္စတန်', 'mzn': 'تاجیکستون', 'naq': 'Tajikistan', 'nb': 'Tadsjikistan', 'nd': 'Tajikistan', 'ne': 'ताजिकिस्तान', 'nl': 'Tadzjikistan', 'nmg': 'Tajikistaŋ', 'nn': 'Tadsjikistan', 'no': 'Tadsjikistan', 'nqo': 'ߕߊߖߞߌߛߑߕߊ߲߫', 'nyn': 'Tajikisitani', 'om': 'Tajikistaan', 'or': 'ତାଜିକିସ୍ଥାନ୍', 'pa': 'ਤਾਜਿਕਿਸਤਾਨ', 'pa-Guru': 'ਤਾਜਿਕਿਸਤਾਨ', 'pcm': 'Tajíkstan', 'pl': 'Tadżykistan', 'ps': 'تاجکستان', 'pt': 'Tadjiquistão', 'qu': 'Tayikistán', 'rm': 'Tadschikistan', 'rn': 'Tajikisitani', 'ro': 'Tadjikistan', 'rof': 'Tajikistani', 'ru': 'Таджикистан', 'rwk': 'Tajikistani', 'saq': 'Tajikistani', 'sbp': 'Tajikisitani', 'sc': 'Tagìkistan', 'sd': 'تاجڪستان', 'sd-Arab': 'تاجڪستان', 'se': 'Tažikistan', 'seh': 'Tadjiquistão', 'ses': 'Taažikistan', 'sg': 'Taazikiistäan', 'shi': 'ⵜⴰⴷⵊⴰⴽⵉⵙⵜⴰⵏ', 'shi-Latn': 'tadjakistan', 'shi-Tfng': 'ⵜⴰⴷⵊⴰⴽⵉⵙⵜⴰⵏ', 'si': 'ටජිකිස්තානය', 'sk': 'Tadžikistan', 'sl': 'Tadžikistan', 'smn': 'Tadžikistan', 'sn': 'Tajikistan', 'so': 'Tajikistan', 'sq': 'Taxhikistan', 'sr': 'Таџикистан', 'sr-Cyrl': 'Таџикистан', 'sr-Latn': 'Tadžikistan', 'sv': 'Tadzjikistan', 'sw': 'Tajikistani', 'syr': 'ܬܐܓܝܟܣܬܐܢ', 'ta': 'தஜிகிஸ்தான்', 'te': 'తజికిస్తాన్', 'teo': 'Tajikistani', 'tg': 'Тоҷикистон', 'th': 'ทาจิกิสถาน', 'ti': 'ታጂኪስታን', 'tk': 'Täjigistan', 'to': 'Tasikitani', 'tr': 'Tacikistan', 'tt': 'Таҗикстан', 'twq': 'Taažikistan', 'tzm': 'Ṭaǧikistan', 'ug': 'تاجىكىستان', 'uk': 'Таджикистан', 'ur': 'تاجکستان', 'uz': 'Tojikiston', 'uz-Cyrl': 'Тожикистон', 'uz-Latn': 'Tojikiston', 'vai': 'ꕚꕀꕃꔻꕚꘋ', 'vai-Latn': 'Tajikisitaŋ', 'vai-Vaii': 'ꕚꕀꕃꔻꕚꘋ', 'vec': 'Tajikistan', 'vi': 'Tajikistan', 'vun': 'Tajikistani', 'wae': 'Tadšikistan', 'wo': 'Tajikistaŋ', 'xh': 'ETajikistan', 'xnr': 'ताज़िकिस्तान', 'xog': 'Tajikisitaani', 'yav': 'tasikistáaŋ', 'yo': 'Takisitani', 'yrl': 'Tayikiretãma', 'yue': '塔吉克', 'yue-Hans': '塔吉克', 'yue-Hant': '塔吉克', 'zgh': 'ⵜⴰⴷⵊⴰⴽⵉⵙⵜⴰⵏ', 'zh': '塔吉克斯坦', 'zh-Hans': '塔吉克斯坦', 'zh-Hant': '塔吉克', 'zu': 'i-Tajikistan'},
    'TK': {'af': 'Tokelau', 'agq': 'Tuwkelawù', 'ak': 'Tokelau', 'am': 'ቶክላው', 'ar': 'توكيلاو', 'as': 'টোকেলাউ', 'asa': 'Tokelau', 'ast': 'Tokeláu', 'az': 'Tokelau', 'az-Cyrl': 'Токелау', 'az-Latn': 'Tokelau', 'bal-Latn': 'Tokéláu', 'bas': 'Tòkèlaò', 'be': 'Такелау', 'be-tarask': 'Такелау', 'bez': 'Hutokelau', 'bg': 'Токелау', 'blo': 'Tokelaʊ', 'bm': 'Tokelo', 'bn': 'টোকেলাউ', 'br': 'Tokelau', 'brx': 'टकेलाऊ', 'bs': 'Tokelau', 'bs-Cyrl': 'Токелау', 'bs-Latn': 'Tokelau', 'ca': 'Tokelau', 'ca-ES-valencia': 'Tokelau', 'ccp': '𑄑𑄮𑄇𑄬𑄣𑄃𑄪', 'ce': 'Токелау', 'ceb': 'Tokelau', 'cgg': 'Tokerawu', 'chr': 'ᏙᎨᎳᏭ', 'ckb': 'تۆکێلاو', 'cs': 'Tokelau', 'cv': 'Токелау', 'cy': 'Tokelau', 'da': 'Tokelau', 'dav': 'Tokelau', 'de': 'Tokelau', 'dje': 'Tokelau', 'dsb': 'Tokelau', 'dz': 'ཏོ་ཀེ་ལའུ་ མཚོ་གླིང', 'ebu': 'Tokelau', 'ee': 'Tokelau nutome', 'el': 'Τοκελάου', 'el-polyton': 'Τοκελάου', 'en': 'Tokelau', 'en-Dsrt': '𐐓𐐬𐐿𐐯𐑊𐐵', 'es': 'Tokelau', 'et': 'Tokelau', 'eu': 'Tokelau', 'ewo': 'Tokeló', 'fa': 'توکلائو', 'ff': 'Tokelaaw', 'ff-Adlm': '𞤚𞤮𞥅𞤳𞤮𞤤𞤢𞥄𞤱𞤵', 'ff-Latn': 'Tokelaaw', 'fi': 'Tokelau', 'fil': 'Tokelau', 'fo': 'Tokelau', 'fr': 'Tokelau', 'fur': 'Tokelau', 'fy': 'Tokelau', 'ga': 'Tócalá', 'gd': 'Tokelau', 'gl': 'Tokelau', 'gsw': 'Tokelau', 'gu': 'ટોકેલાઉ', 'guz': 'Tokelau', 'ha': 'Takelau', 'he': 'טוקלאו', 'hi': 'तोकेलाउ', 'hi-Latn': 'Tokelau', 'hr': 'Tokelau', 'hsb': 'Tokelau', 'ht': 'Tokelau', 'hu': 'Tokelau', 'hy': 'Տոկելաու', 'ia': 'Tokelau', 'id': 'Tokelau', 'ie': 'Tokelau', 'ig': 'Tokelau', 'is': 'Tókelá', 'it': 'Tokelau', 'ja': 'トケラウ', 'jmc': 'Tokelau', 'jv': 'Tokelau', 'ka': 'ტოკელაუ', 'kaa': 'Токелау', 'kaa-Cyrl': 'Токелау', 'kab': 'Ṭuklu', 'kam': 'Tokelau', 'kde': 'Tokelau', 'kea': 'Tokelau', 'kgp': 'Tokeravu', 'khq': 'Tokelau', 'ki': 'Tokelau', 'kk': 'Токелау', 'kk-Cyrl': 'Токелау', 'kln': 'Emetab Tokelau', 'km': 'តូខេឡៅ', 'kn': 'ಟೊಕೆಲಾವ್', 'ko': '토켈라우', 'kok': 'टोकलाऊ', 'kok-Deva': 'टोकलाऊ', 'ks': 'ٹوکلو', 'ks-Arab': 'ٹوکلو', 'ksb': 'Tokelau', 'ksf': 'tokǝlao', 'ksh': 'Tokelau', 'ku': 'Tokelau', 'kxv': 'ṭokelau', 'kxv-Deva': 'तोकेलाउ', 'kxv-Latn': 'ṭokelau', 'kxv-Orya': 'ତକେଲାଉ', 'kxv-Telu': 'టోకెలావ్', 'ky': 'Токелау', 'lag': 'Tokeláau', 'lb': 'Tokelau', 'lg': 'Tokelawu', 'lld': 'Tokelau', 'ln': 'Tokelau', 'lo': 'ໂຕເກເລົາ', 'lt': 'Tokelau', 'lu': 'Tokelau', 'luo': 'Tokelau', 'luy': 'Tokelau', 'lv': 'Tokelau', 'mai': 'टोकेलाऊ', 'mas': 'Tokelau', 'mer': 'Tokelau', 'mfe': 'Tokelau', 'mg': 'Tokelao', 'mgh': 'Utokelau', 'mi': 'Tokerau', 'mk': 'Токелау', 'ml': 'ടോക്കെലൂ', 'mn': 'Токелау', 'mr': 'तोकेलाउ', 'ms': 'Tokelau', 'mt': 'it-Tokelau', 'mua': 'Sǝr Tokelau', 'my': 'တိုကလောင်', 'mzn': 'توکلائو', 'naq': 'Tokelau', 'nb': 'Tokelau', 'nd': 'Thokelawu', 'ne': 'तोकेलाउ', 'nl': 'Tokelau', 'nmg': 'Tokelo', 'nn': 'Tokelau', 'no': 'Tokelau', 'nqo': 'ߕߏߞߋߟߊߏ߫', 'nyn': 'Tokerawu', 'om': 'Tokelau', 'or': 'ଟୋକେଲାଉ', 'pa': 'ਟੋਕੇਲਾਉ', 'pa-Guru': 'ਟੋਕੇਲਾਉ', 'pcm': 'Tókẹ́lau', 'pl': 'Tokelau', 'ps': 'توکیلو', 'pt': 'Tokelau', 'qu': 'Tokelau', 'rm': 'Tokelau', 'rn': 'Tokelawu', 'ro': 'Tokelau', 'rof': 'Tokelau', 'ru': 'Токелау', 'rwk': 'Tokelau', 'saq': 'Tokelau', 'sbp': 'Tokelawu', 'sc': 'Tokelau', 'sd': 'ٽوڪلائو', 'sd-Arab': 'ٽوڪلائو', 'se': 'Tokelau', 'seh': 'Tokelau', 'ses': 'Tokelau', 'sg': 'Tokelau', 'shi': 'ⵟⵓⴽⵍⴰⵡ', 'shi-Latn': 'ṭuklaw', 'shi-Tfng': 'ⵟⵓⴽⵍⴰⵡ', 'si': 'ටොකලාවු', 'sk': 'Tokelau', 'sl': 'Tokelau', 'smn': 'Tokelau', 'sn': 'Tokelau', 'so': 'Tokelaaw', 'sq': 'Tokelau', 'sr': 'Токелау', 'sr-Cyrl': 'Токелау', 'sr-Latn': 'Tokelau', 'sv': 'Tokelauöarna', 'sw': 'Tokelau', 'ta': 'டோகேலோ', 'te': 'టోకెలావ్', 'teo': 'Tokelau', 'tg': 'Токелау', 'th': 'โตเกเลา', 'ti': 'ቶከላው', 'tk': 'Tokelau', 'to': 'Tokelau', 'tr': 'Tokelau', 'tt': 'Токелау', 'twq': 'Tokelau', 'tzm': 'Tuklu', 'ug': 'توكېلاۋ', 'uk': 'Токелау', 'ur': 'ٹوکیلاؤ', 'uz': 'Tokelau', 'uz-Cyrl': 'Токелау', 'uz-Latn': 'Tokelau', 'vai': 'ꕿꔞꖃ', 'vai-Latn': 'Tokelo', 'vai-Vaii': 'ꕿꔞꖃ', 'vec': 'Tokelàu', 'vi': 'Tokelau', 'vun': 'Tokelau', 'wae': 'Tokelau', 'wo': 'Tokoloo', 'xh': 'ETokelau', 'xnr': 'तोकेलाउ', 'xog': 'Tokelawu', 'yav': 'tokeló', 'yo': 'Tokelau', 'yrl': 'Tukerau', 'yue': '托克勞群島', 'yue-Hans': '托克劳群岛', 'yue-Hant': '托克勞群島', 'zgh': 'ⵟⵓⴽⵍⴰⵡ', 'zh': '托克劳', 'zh-Hans': '托克劳', 'zh-Hant': '托克勞群島', 'zu': 'i-Tokelau'},
    'TL': {'af': 'Oos-Timor', 'agq': 'Ês Taemò', 'ak': 'Timɔ Boka', 'am': 'ቲሞር ሌስቴ', 'ar': 'تيمور - ليشتي', 'as': 'টিমোৰ-লেচটে', 'asa': 'Timori ya Mashariki', 'ast': 'Timor Oriental', 'az': 'Şərqi Timor', 'az-Cyrl': 'Шәрги Тимор', 'az-Latn': 'Şərqi Timor', 'bal-Latn': 'Témur-Leste', 'bas': 'Tìmɔ̂r lìkòl', 'be': 'Тымор-Лешці', 'be-tarask': 'Тымор-Лешці', 'bez': 'Hutimori ya Mashariki', 'bg': 'Тимор Лесте', 'blo': 'Timɔɔ gajakalaŋ', 'bm': 'Kɔrɔn Timɔr', 'bn': 'তিমুর-লেস্তে', 'br': 'Timor-Leste', 'brx': 'तिमर-लेस्ते', 'bs': 'Istočni Timor', 'bs-Cyrl': 'Тимор-Лесте', 'bs-Latn': 'Istočni Timor', 'ca': 'Timor-Leste', 'ca-ES-valencia': 'Timor-Leste', 'ccp': '𑄖𑄨𑄟𑄪𑄢𑄴-𑄣𑄬𑄌𑄴𑄖𑄬', 'ce': 'Малхбален Тимор', 'ceb': 'Timor-Leste', 'cgg': 'Burugweizooba bwa Timori', 'chr': 'ᏘᎼᎵ-ᎴᏍᏖ', 'ckb': 'تیمۆری ڕۆژھەڵات', 'cs': 'Východní Timor', 'cv': 'Хӗвелтухӑҫ Тимор', 'cy': 'Timor-Leste', 'da': 'Timor-Leste', 'dav': 'Timori ya Mashariki', 'de': 'Timor-Leste', 'dje': 'Timoor hawsa', 'dsb': 'Timor-Leste', 'dz': 'ཏི་་མོར་ལེ་ཨེསཊ', 'ebu': 'Timori ya Mashariki', 'ee': 'Timor-Leste nutome', 'el': 'Τιμόρ-Λέστε', 'el-polyton': 'Ἀνατολικὸ Τιμόρ', 'en': 'Timor-Leste', 'en-Dsrt': '𐐀𐑅𐐻 𐐓𐐨𐑋𐐱𐑉', 'es': 'Timor-Leste', 'et': 'Ida-Timor', 'eu': 'Ekialdeko Timor', 'ewo': 'Timôr', 'fa': 'تیمور-لسته', 'ff': 'Timoor Fuɗnaange', 'ff-Adlm': '𞤚𞤭𞤥𞤮𞥅𞤪 𞤂𞤫𞤧𞤼𞤫', 'ff-Latn': 'Timoor Fuɗnaange', 'fi': 'Itä-Timor', 'fil': 'Timor-Leste', 'fo': 'Eysturtimor', 'fr': 'Timor oriental', 'fur': 'Timor orientâl', 'fy': 'East-Timor', 'ga': 'Tíomór Thoir', 'gd': 'Timor-Leste', 'gl': 'Timor Leste', 'gsw': 'Oschttimor', 'gu': 'તિમોર-લેસ્તે', 'guz': 'Timori ya Mashariki', 'ha': 'Timor-Leste', 'he': 'טימור-לסטה', 'hi': 'तिमोर-लेस्त', 'hi-Latn': 'Timor-Leste', 'hr': 'Timor-Leste', 'hsb': 'Timor-Leste', 'ht': 'Timor oriental', 'hu': 'Kelet-Timor', 'hy': 'Թիմոր Լեշտի', 'ia': 'Timor del Est', 'id': 'Timor Leste', 'ie': 'Ost-Timor', 'ig': 'Timor-Leste', 'is': 'Tímor-Leste', 'it': 'Timor Est', 'ja': '東ティモール', 'jmc': 'Timori ya Mashariki', 'jv': 'Timor Leste', 'ka': 'ტიმორ-ლესტე', 'kaa': 'Шығыс Тимор', 'kaa-Cyrl': 'Шығыс Тимор', 'kab': 'Tumur Asamar', 'kam': 'Timori ya Mashariki', 'kde': 'Timoli ya Mashaliki', 'kea': 'Timor Lesti', 'kgp': 'Timãr-Rãjur', 'khq': 'Timoor hawsa', 'ki': 'Timori ya Mashariki', 'kk': 'Тимор-Лесте', 'kk-Cyrl': 'Тимор-Лесте', 'kln': 'Emetab Timor nebo Murot tai', 'km': 'ទីម័រលេស្តេ', 'kn': 'ಟಿಮೋರ್ ಲೆಸ್ಟೆ', 'ko': '동티모르', 'kok': 'तिमोर-लेस्ते', 'kok-Deva': 'तिमोर-लेस्ते', 'ks': 'تیمور-لیسٹ', 'ks-Arab': 'تیمور-لیسٹ', 'ksb': 'Timoli ya Mashaliki', 'ksf': 'timor anǝ á ɛst', 'ksh': 'Oß-Timor', 'ku': 'Tîmor-Leste', 'kxv': 'ṭimor - leste', 'kxv-Deva': 'तिमोर-लेस्त', 'kxv-Latn': 'ṭimor - leste', 'kxv-Orya': 'ତିମର-ଲେସ୍ତେ', 'kxv-Telu': 'టిమోర్-లెస్టె', 'ky': 'Тимор-Лесте', 'lag': 'Timóori yi Itʉʉmba', 'lb': 'Osttimor', 'lg': 'Timowa', 'lld': 'Timor-Leste', 'ln': 'Timorɛ ya Moniɛlɛ', 'lo': 'ທິມໍ-ເລສເຕ', 'lt': 'Rytų Timoras', 'lu': 'Timoru wa diboku', 'luo': 'East Timor', 'luy': 'Timori ya Mashariki', 'lv': 'Austrumtimora', 'mai': 'तिमोर-लेस्त', 'mas': 'Timori ya Mashariki', 'mer': 'Timori ya Isti', 'mfe': 'Timor oriantal', 'mg': 'Timor Atsinanana', 'mgh': 'Utimo Mashariki', 'mi': 'Tīmoa ki te Rāwhiti', 'mk': 'Тимор Лесте', 'ml': 'തിമോർ-ലെസ്റ്റെ', 'mn': 'Тимор-Лесте', 'mr': 'तिमोर-लेस्ते', 'ms': 'Timor-Leste', 'mt': 'Timor Leste', 'mua': 'Timoriya', 'my': 'အရှေ့တီမော', 'mzn': 'تیمور شرقی', 'naq': 'East Timor', 'nb': 'Øst-Timor', 'nd': 'East Timor', 'ne': 'टिमोर-लेस्टे', 'nl': 'Oost-Timor', 'nmg': 'Timɔr tsindikēh', 'nn': 'Aust-Timor', 'no': 'Øst-Timor', 'nqo': 'ߕߌߡߐߙ ߕߟߋ߬ߓߐ߬ߝߊ߲', 'nyn': 'Burugweizooba bwa Timori', 'om': 'Tiimoor-Leestee', 'or': 'ତିମୋର୍-ଲେଷ୍ଟେ', 'pa': 'ਤਿਮੋਰ-ਲੇਸਤੇ', 'pa-Guru': 'ਤਿਮੋਰ-ਲੇਸਤੇ', 'pcm': 'Íst Tímọ', 'pl': 'Timor Wschodni', 'ps': 'تيمور-ليسټ', 'pt': 'Timor-Leste', 'qu': 'Timor-Leste', 'rm': 'Timor da l’Ost', 'rn': 'Timoru y’iburasirazuba', 'ro': 'Timor-Leste', 'rof': 'Timori ya Mashariki', 'ru': 'Восточный Тимор', 'rwk': 'Timori ya Mashariki', 'saq': 'Timori ya Mashariki', 'sbp': 'Timoli ya kunena', 'sc': 'Timor-Est', 'sd': 'تيمور ليستي', 'sd-Arab': 'تيمور ليستي', 'se': 'Nuorta-Timor', 'seh': 'Timor Leste', 'ses': 'Timoor hawsa', 'sg': 'Timôro tî Tö', 'shi': 'ⵜⵉⵎⵓⵔ ⵏ ⵍⵇⴱⵍⵜ', 'shi-Latn': 'timur n lqblt', 'shi-Tfng': 'ⵜⵉⵎⵓⵔ ⵏ ⵍⵇⴱⵍⵜ', 'si': 'ටිමෝර් - ලෙස්ත්', 'sk': 'Východný Timor', 'sl': 'Timor-Leste', 'smn': 'Timor-Leste', 'sn': 'East Timor', 'so': 'Timor-Leste', 'sq': 'Timor-Leste', 'sr': 'Тимор-Лесте (Источни Тимор)', 'sr-Cyrl': 'Тимор-Лесте (Источни Тимор)', 'sr-Latn': 'Timor-Leste (Istočni Timor)', 'sv': 'Östtimor', 'sw': 'Timor-Leste', 'syr': 'ܬܝܡܘܪ-ܠܣܬܝ', 'ta': 'திமோர்-லெஸ்தே', 'te': 'టిమోర్-లెస్టె', 'teo': 'Timori ya Mashariki', 'tg': 'Тимор-Лесте', 'th': 'ติมอร์-เลสเต', 'ti': 'ቲሞር-ለስተ', 'tk': 'Timor-Leste', 'to': 'Timoa fakahahake', 'tr': 'Timor-Leste', 'tt': 'Тимор-Лесте', 'twq': 'Timoor hawsa', 'tzm': 'Timur Tagmuṭ', 'ug': 'شەرقىي تىمور', 'uk': 'Тимор-Лешті', 'ur': 'تیمور لیسٹ', 'uz': 'Timor-Leste', 'uz-Cyrl': 'Тимор-Лесте', 'uz-Latn': 'Timor-Leste', 'vai': 'ꔎꔒ ꗃ ꔳꗞꖻ', 'vai-Latn': 'Tele Ɓɔ́ Timɔɔ̃', 'vai-Vaii': 'ꔎꔒ ꗃ ꔳꗞꖻ', 'vec': 'Timor Est', 'vi': 'Timor-Leste', 'vun': 'Timori ya Mashariki', 'wae': 'Ošttimor', 'wo': 'Timor Leste', 'xh': 'ETimor-Leste', 'xnr': 'तिमोर-लेस्त', 'xog': 'Timowa', 'yav': 'timɔ́ɔl u nipálɛ́n', 'yi': 'מזרח טימאר', 'yo': 'Tímọ̀ Lẹsiti', 'yrl': 'Timu-Semusawa', 'yue': '東帝汶', 'yue-Hans': '东帝汶', 'yue-Hant': '東帝汶', 'zgh': 'ⵜⵉⵎⵓⵔ ⵏ ⵍⵇⴱⵍⵜ', 'zh': '东帝汶', 'zh-Hans': '东帝汶', 'zh-Hant': '東帝汶', 'zu': 'i-Timor-Leste'},
    'TM': {'af': 'Turkmenistan', 'agq': 'Tekɨmènèsɨ̀tân', 'ak': 'Tɛkmɛnistan', 'am': 'ቱርክሜኒስታን', 'ar': 'تركمانستان', 'as': 'তুৰ্কমেনিস্তান', 'asa': 'Turukimenithtani', 'ast': 'Turkmenistán', 'az': 'Türkmənistan', 'az-Cyrl': 'Түркмәнистан', 'az-Latn': 'Türkmənistan', 'bal-Latn': 'Torkmenestán', 'bas': 'Tùrgmènìstân', 'be': 'Туркменістан', 'be-tarask': 'Туркменістан', 'bez': 'Huuturukimenistani', 'bg': 'Туркменистан', 'bgn': 'تورکمنستان', 'blo': 'Turkmenistan', 'bm': 'Turikimenisitani', 'bn': 'তুর্কমেনিস্তান', 'br': 'Turkmenistan', 'brx': 'तुर्कमेनीस्तान', 'bs': 'Turkmenistan', 'bs-Cyrl': 'Туркменистан', 'bs-Latn': 'Turkmenistan', 'ca': 'Turkmenistan', 'ca-ES-valencia': 'Turkmenistan', 'ccp': '𑄖𑄪𑄢𑄴𑄇𑄧𑄟𑄬𑄚𑄨𑄌𑄴𑄖𑄚𑄴', 'ce': 'Туркмени', 'ceb': 'Turkmenistan', 'cgg': 'Turukimenisitani', 'chr': 'ᏛᎵᎩᎺᏂᏍᏔᏂ', 'ckb': 'تورکمانستان', 'cs': 'Turkmenistán', 'cv': 'Туркменистан', 'cy': 'Tyrcmenistan', 'da': 'Turkmenistan', 'dav': 'Turukimenistani', 'de': 'Turkmenistan', 'dje': 'Turkmenistaŋ', 'dsb': 'Turkmeniska', 'dz': 'ཊཱརཀ་མེནའི་སཏཱན', 'ebu': 'Turukimenistani', 'ee': 'Tɛkmenistan nutome', 'el': 'Τουρκμενιστάν', 'el-polyton': 'Τουρκμενιστάν', 'en': 'Turkmenistan', 'en-Dsrt': '𐐓𐐲𐑉𐐿𐑋𐐯𐑌𐐲𐑅𐐻𐐰𐑌', 'eo': 'Turkmenujo', 'es': 'Turkmenistán', 'et': 'Türkmenistan', 'eu': 'Turkmenistan', 'ewo': 'Turəkəmənisətáŋ', 'fa': 'ترکمنستان', 'ff': 'Turkmenistaan', 'ff-Adlm': '𞤚𞤵𞤪𞤳𞤵𞤥𞤫𞤲𞤭𞤧𞤼𞤢𞥄𞤲', 'ff-Latn': 'Turkmenistaan', 'fi': 'Turkmenistan', 'fil': 'Turkmenistan', 'fo': 'Turkmenistan', 'fr': 'Turkménistan', 'fur': 'Turkmenistan', 'fy': 'Turkmenistan', 'ga': 'an Tuircméanastáin', 'gd': 'Turcmanastàn', 'gl': 'Turkmenistán', 'gsw': 'Turkmeenischtan', 'gu': 'તુર્કમેનિસ્તાન', 'guz': 'Turukimenistani', 'ha': 'Turkumenistan', 'he': 'טורקמניסטן', 'hi': 'तुर्कमेनिस्तान', 'hi-Latn': 'Turkmenistan', 'hr': 'Turkmenistan', 'hsb': 'Turkmeniska', 'ht': 'Turkménistan', 'hu': 'Türkmenisztán', 'hy': 'Թուրքմենստան', 'ia': 'Turkmenistan', 'id': 'Turkmenistan', 'ig': 'Turkmenistan', 'is': 'Túrkmenistan', 'it': 'Turkmenistan', 'ja': 'トルクメニスタン', 'jmc': 'Turukimenistani', 'jv': 'Turkménistan', 'ka': 'თურქმენეთი', 'kaa': 'Түркменстан', 'kaa-Cyrl': 'Түркменстан', 'kab': 'Ṭurkmanistan', 'kam': 'Turukimenistani', 'kde': 'Tuluchimenistani', 'kea': 'Turkumenistan', 'kgp': 'Turkomẽnĩnhtỹv', 'khq': 'Turkmenistaŋ', 'ki': 'Turukimenistani', 'kk': 'Түрікменстан', 'kk-Cyrl': 'Түрікменстан', 'kln': 'Emetab Turkmenistan', 'km': 'តួកម៉េនីស្ថាន', 'kn': 'ತುರ್ಕಮೆನಿಸ್ತಾನ್', 'ko': '투르크메니스탄', 'kok': 'तुर्कमेनिस्तान', 'kok-Deva': 'तुर्कमेनिस्तान', 'ks': 'تُرکمنستان', 'ks-Arab': 'تُرکمنستان', 'ksb': 'Tulukimenistani', 'ksf': 'tirkmɛnistaŋ', 'ksh': 'Turkmenistahn', 'ku': 'Tirkmenistan', 'kxv': 'turkmenīstān', 'kxv-Deva': 'तुर्कमेनिस्तान', 'kxv-Latn': 'turkmenīstān', 'kxv-Orya': 'ତୁର୍କମେନିସ୍ତାନ', 'kxv-Telu': 'తుర్క్\u200dమెనిస్తన్', 'ky': 'Түркмөнстан', 'lag': 'Uturukimenisitáani', 'lb': 'Turkmenistan', 'lg': 'Takimenesitaani', 'lld': 'Turkmenistan', 'ln': 'Tikɛménisitá', 'lo': 'ເທີກເມນິສະຖານ', 'lt': 'Turkmėnistanas', 'lu': 'Tukemenisita', 'luo': 'Turkmenistan', 'luy': 'Turukimenistani', 'lv': 'Turkmenistāna', 'mai': 'तुर्कमेनिस्तान', 'mas': 'Turukimenistani', 'mer': 'Tukumenistani', 'mfe': 'Turkmenistan', 'mg': 'Torkmenistan', 'mgh': 'Uturukimenistani', 'mi': 'Tukumanatānga', 'mk': 'Туркменистан', 'ml': 'തുർക്ക്മെനിസ്ഥാൻ', 'mn': 'Туркменистан', 'mr': 'तुर्कमेनिस्तान', 'ms': 'Turkmenistan', 'mt': 'it-Turkmenistan', 'mua': 'Turkmenistaŋ', 'my': 'တာ့ခ်မင်နစ္စတန်', 'mzn': 'ترکمونستون', 'naq': 'Turkmenistan', 'nb': 'Turkmenistan', 'nd': 'Turkmenistan', 'ne': 'तुर्कमेनिस्तान', 'nl': 'Turkmenistan', 'nmg': 'Turkmɛnistaŋ', 'nn': 'Turkmenistan', 'no': 'Turkmenistan', 'nqo': 'ߕߎߙߑߞߌߡߋߣߌߛߑߕߊ߲߫', 'nyn': 'Turukimenisitani', 'om': 'Turkimenistaan', 'or': 'ତୁର୍କମେନିସ୍ତାନ', 'pa': 'ਤੁਰਕਮੇਨਿਸਤਾਨ', 'pa-Guru': 'ਤੁਰਕਮੇਨਿਸਤਾਨ', 'pcm': 'Tọkmẹ́nístan', 'pl': 'Turkmenistan', 'ps': 'تورکمنستان', 'pt': 'Turcomenistão', 'qu': 'Turkmenistán', 'rm': 'Turkmenistan', 'rn': 'Turukumenisitani', 'ro': 'Turkmenistan', 'rof': 'Turukimenistani', 'ru': 'Туркменистан', 'rwk': 'Turukimenistani', 'saq': 'Turukimenistani', 'sbp': 'Tulukimenisitani', 'sc': 'Turkmènistan', 'sd': 'ترڪمانستان', 'sd-Arab': 'ترڪمانستان', 'se': 'Turkmenistan', 'seh': 'Turcomenistão', 'ses': 'Turkmenistaŋ', 'sg': 'Turkumenistäan', 'shi': 'ⵜⵓⵔⴽⵎⴰⵏⵙⵜⴰⵏ', 'shi-Latn': 'turkmanstan', 'shi-Tfng': 'ⵜⵓⵔⴽⵎⴰⵏⵙⵜⴰⵏ', 'si': 'ටර්ක්මෙනිස්ථානය', 'sk': 'Turkménsko', 'sl': 'Turkmenistan', 'smn': 'Turkmenistan', 'sn': 'Turkmenistan', 'so': 'Turkmenistan', 'sq': 'Turkmenistan', 'sr': 'Туркменистан', 'sr-Cyrl': 'Туркменистан', 'sr-Latn': 'Turkmenistan', 'sv': 'Turkmenistan', 'sw': 'Turkmenistan', 'syr': 'ܬܘܪܟܡܢܣܬܐܢ', 'ta': 'துர்க்மெனிஸ்தான்', 'te': 'టర్క్\u200cమెనిస్తాన్', 'teo': 'Turukimenistani', 'tg': 'Туркманистон', 'th': 'เติร์กเมนิสถาน', 'ti': 'ቱርክመኒስታን', 'tk': 'Türkmenistan', 'to': 'Tūkimenisitani', 'tr': 'Türkmenistan', 'tt': 'Төркмәнстан', 'twq': 'Turkmenistaŋ', 'tzm': 'Turkmanistan', 'ug': 'تۈركمەنىستان', 'uk': 'Туркменістан', 'ur': 'ترکمانستان', 'uz': 'Turkmaniston', 'uz-Cyrl': 'Туркманистон', 'uz-Latn': 'Turkmaniston', 'vai': 'ꗋꖺꕃꕮꕇꔻꕚꘋ', 'vai-Latn': 'Tukimɛnisitaŋ', 'vai-Vaii': 'ꗋꖺꕃꕮꕇꔻꕚꘋ', 'vec': 'Turkmenistan', 'vi': 'Turkmenistan', 'vun': 'Turukimenistani', 'wae': 'Turkmeništan', 'wo': 'Tirkmenistaŋ', 'xh': 'ETurkmenistan', 'xnr': 'तुर्कमेनिस्तान', 'xog': 'Takimenesitaani', 'yav': 'tulukmenisitáaŋ', 'yi': 'טורקמעניסטאַן', 'yo': 'Tọ́kìmẹ́nísítànì', 'yrl': 'Turkuranaretãma', 'yue': '土庫曼', 'yue-Hans': '土库曼', 'yue-Hant': '土庫曼', 'zgh': 'ⵜⵓⵔⴽⵎⴰⵏⵙⵜⴰⵏ', 'zh': '土库曼斯坦', 'zh-Hans': '土库曼斯坦', 'zh-Hant': '土庫曼', 'zu': 'i-Turkmenistan'},
    'TN': {'af': 'Tunisië', 'agq': 'Tùneshìa', 'ak': 'Tunihyia', 'am': 'ቱኒዚያ', 'ar': 'تونس', 'as': 'টুনিচিয়া', 'asa': 'Tunithia', 'ast': 'Tunicia', 'az': 'Tunis', 'az-Cyrl': 'Тунис', 'az-Latn': 'Tunis', 'bal-Latn': 'Tunishiá', 'bas': 'Tùnisìà', 'be': 'Туніс', 'be-tarask': 'Туніс', 'bez': 'Hutunisia', 'bg': 'Тунис', 'bgn': 'ٹونیس', 'blo': 'Tunisii', 'bm': 'Tunizi', 'bn': 'তিউনিসিয়া', 'br': 'Tunizia', 'brx': 'त्युनिशिया', 'bs': 'Tunis', 'bs-Cyrl': 'Тунис', 'bs-Latn': 'Tunis', 'ca': 'Tunísia', 'ca-ES-valencia': 'Tunísia', 'ccp': '𑄖𑄨𑄃𑄪𑄚𑄨𑄥𑄨𑄠', 'ce': 'Тунис', 'ceb': 'Tunisia', 'cgg': 'Tunizia', 'chr': 'ᏚᏂᏏᏍᎠ', 'ckb': 'توونس', 'cs': 'Tunisko', 'cv': 'Тунис', 'cy': 'Tiwnisia', 'da': 'Tunesien', 'dav': 'Tunisia', 'de': 'Tunesien', 'dje': 'Tunizi', 'dsb': 'Tuneziska', 'dz': 'ཊུ་ནི་ཤི་ཡ', 'ebu': 'Tunisia', 'ee': 'Tunisia nutome', 'el': 'Τυνησία', 'el-polyton': 'Τυνησία', 'en': 'Tunisia', 'eo': 'Tunizio', 'es': 'Túnez', 'et': 'Tuneesia', 'eu': 'Tunisia', 'ewo': 'Tunisí', 'fa': 'تونس', 'ff': 'Tunisii', 'ff-Adlm': '𞤚𞤵𞤲𞤭𞥅𞤧𞤢', 'ff-Latn': 'Tunisii', 'fi': 'Tunisia', 'fil': 'Tunisia', 'fo': 'Tunesia', 'fr': 'Tunisie', 'fur': 'Tunisie', 'fy': 'Tunesië', 'ga': 'An Tuinéis', 'gd': 'Tuinisea', 'gl': 'Tunisia', 'gsw': 'Tuneesie', 'gu': 'ટ્યુનિશિયા', 'guz': 'Tunisia', 'ha': 'Tunisiya', 'he': 'תוניסיה', 'hi': 'ट्यूनीशिया', 'hi-Latn': 'Tunisia', 'hr': 'Tunis', 'hsb': 'Tuneziska', 'ht': 'Tunisie', 'hu': 'Tunézia', 'hy': 'Թունիս', 'ia': 'Tunisia', 'id': 'Tunisia', 'ig': 'Tunisia', 'is': 'Túnis', 'it': 'Tunisia', 'ja': 'チュニジア', 'jgo': 'Tunizî', 'jmc': 'Tunisia', 'jv': 'Tunisia', 'ka': 'ტუნისი', 'kaa': 'Тунис', 'kaa-Cyrl': 'Тунис', 'kab': 'Tunes', 'kam': 'Tunisia', 'kde': 'Tunisia', 'kea': 'Tunízia', 'kgp': 'Tunĩja', 'khq': 'Tunizi', 'ki': 'Tunisia', 'kk': 'Тунис', 'kk-Cyrl': 'Тунис', 'kln': 'Emetab Tunisia', 'km': 'ទុយនីស៊ី', 'kn': 'ಟುನೀಶಿಯ', 'ko': '튀니지', 'kok': 'ट्यूनीशिया', 'kok-Deva': 'ट्यूनीशिया', 'ks': 'ٹونیشِیا', 'ks-Arab': 'ٹونیشِیا', 'ksb': 'Tunisia', 'ksf': 'tunɛsí', 'ksh': 'Tuneesije', 'ku': 'Tûnis', 'kxv': 'ṭunisiaa', 'kxv-Deva': 'ट्यूनीसिया', 'kxv-Latn': 'ṭunisiaa', 'kxv-Orya': 'ଟ୍ୟୁନିସିୟା', 'kxv-Telu': 'ట్యునీషియా', 'ky': 'Тунис', 'lag': 'Tunísia', 'lb': 'Tunesien', 'lg': 'Tunisya', 'lld': 'Tunisia', 'ln': 'Tinizi', 'lo': 'ຕູນິເຊຍ', 'lt': 'Tunisas', 'lu': 'Tinizi', 'luo': 'Tunisia', 'luy': 'Tunisia', 'lv': 'Tunisija', 'mai': 'ट्यूनिशिया', 'mas': 'Tunisia', 'mer': 'Tunisia', 'mfe': 'Tinizi', 'mg': 'Tonizia', 'mgh': 'Utunisia', 'mi': 'Tūnihia', 'mk': 'Тунис', 'ml': 'ടുണീഷ്യ', 'mn': 'Тунис', 'mr': 'ट्यूनिशिया', 'ms': 'Tunisia', 'mt': 'it-Tuneżija', 'mua': 'Tunisiya', 'my': 'တူနီးရှား', 'mzn': 'تونس', 'naq': 'Tunisiab', 'nb': 'Tunisia', 'nd': 'Tunisiya', 'ne': 'ट्युनिसिया', 'nl': 'Tunesië', 'nmg': 'Tunisiá', 'nn': 'Tunisia', 'no': 'Tunisia', 'nqo': 'ߕߎߣߖ߭ߌ߫', 'nyn': 'Tunizia', 'om': 'Tuniiziyaa', 'or': 'ଟ୍ୟୁନିସିଆ', 'pa': 'ਟਿਊਨੀਸ਼ੀਆ', 'pa-Guru': 'ਟਿਊਨੀਸ਼ੀਆ', 'pcm': 'Tuníshia', 'pl': 'Tunezja', 'ps': 'تونس', 'pt': 'Tunísia', 'qu': 'Túnez', 'rm': 'Tunesia', 'rn': 'Tuniziya', 'ro': 'Tunisia', 'rof': 'Tunisia', 'ru': 'Тунис', 'rwk': 'Tunisia', 'saq': 'Tunisia', 'sbp': 'Tunisiya', 'sc': 'Tunisia', 'sd': 'تيونيسيا', 'sd-Arab': 'تيونيسيا', 'se': 'Tunisia', 'seh': 'Tunísia', 'ses': 'Tunizi', 'sg': 'Tunizïi', 'shi': 'ⵜⵓⵏⵙ', 'shi-Latn': 'tuns', 'shi-Tfng': 'ⵜⵓⵏⵙ', 'si': 'ටියුනීසියාව', 'sk': 'Tunisko', 'sl': 'Tunizija', 'smn': 'Tunisia', 'sn': 'Tunisia', 'so': 'Tuniisiya', 'sq': 'Tunizi', 'sr': 'Тунис', 'sr-Cyrl': 'Тунис', 'sr-Latn': 'Tunis', 'sv': 'Tunisien', 'sw': 'Tunisia', 'syr': 'ܬܘܢܣ', 'ta': 'டுனிசியா', 'te': 'ట్యునీషియా', 'teo': 'Tunisia', 'tg': 'Тунис', 'th': 'ตูนิเซีย', 'ti': 'ቱኒዝያ', 'tk': 'Tunis', 'to': 'Tunīsia', 'tr': 'Tunus', 'tt': 'Тунис', 'twq': 'Tunizi', 'tzm': 'Tunes', 'ug': 'تۇنىس', 'uk': 'Туніс', 'ur': 'تونس', 'uz': 'Tunis', 'uz-Cyrl': 'Тунис', 'uz-Latn': 'Tunis', 'vai': 'ꖤꕇꔻꕩ', 'vai-Latn': 'Tunisiya', 'vai-Vaii': 'ꖤꕇꔻꕩ', 'vec': 'Tunizìa', 'vi': 'Tunisia', 'vun': 'Tunisia', 'wae': 'Tunesie', 'wo': 'Tinisi', 'xh': 'ETunisia', 'xnr': 'ट्यूनीशिया', 'xog': 'Tunisya', 'yav': 'tunusí', 'yi': 'טוניסיע', 'yo': 'Tuniṣia', 'yrl': 'Tunisiya', 'yue': '突尼西亞', 'yue-Hans': '突尼西亚', 'yue-Hant': '突尼西亞', 'zgh': 'ⵜⵓⵏⵙ', 'zh': '突尼斯', 'zh-Hans': '突尼斯', 'zh-Hant': '突尼西亞', 'zu': 'i-Tunisia'},
    'TO': {'af': 'Tonga', 'agq': 'Tuŋgà', 'ak': 'Tonga', 'am': 'ቶንጋ', 'ar': 'تونغا', 'as': 'টংগা', 'asa': 'Tonga', 'ast': 'Tonga', 'az': 'Tonqa', 'az-Cyrl': 'Тонга', 'az-Latn': 'Tonqa', 'bal-Latn': 'Tongá', 'bas': 'Tɔŋgà', 'be': 'Тонга', 'be-tarask': 'Тонга', 'bez': 'Hutonga', 'bg': 'Тонга', 'blo': 'Tɔŋga', 'bm': 'Tonga', 'bn': 'টোঙ্গা', 'br': 'Tonga', 'brx': 'टंगा', 'bs': 'Tonga', 'bs-Cyrl': 'Тонга', 'bs-Latn': 'Tonga', 'ca': 'Tonga', 'ca-ES-valencia': 'Tonga', 'ccp': '𑄑𑄮𑄋𑄴𑄉', 'ce': 'Тонга', 'ceb': 'Tonga', 'cgg': 'Tonga', 'chr': 'ᏙᎾᎦ', 'ckb': 'تۆنگا', 'cs': 'Tonga', 'cv': 'Тонга', 'cy': 'Tonga', 'da': 'Tonga', 'dav': 'Tonga', 'de': 'Tonga', 'dje': 'Tonga', 'dsb': 'Tonga', 'dz': 'ཊོང་གྷ', 'ebu': 'Tonga', 'ee': 'Tonga nutome', 'el': 'Τόνγκα', 'el-polyton': 'Τόνγκα', 'en': 'Tonga', 'en-Dsrt': '𐐓𐐪𐑍𐑀𐐲', 'eo': 'Tongo', 'es': 'Tonga', 'et': 'Tonga', 'eu': 'Tonga', 'ewo': 'Tɔngá', 'fa': 'تونگا', 'ff': 'Tonngaa', 'ff-Adlm': '𞤚𞤮𞤲𞤺𞤢', 'ff-Latn': 'Tonngaa', 'fi': 'Tonga', 'fil': 'Tonga', 'fo': 'Tonga', 'fr': 'Tonga', 'fur': 'Tonga', 'fy': 'Tonga', 'ga': 'Tonga', 'gd': 'Tonga', 'gl': 'Tonga', 'gsw': 'Tonga', 'gu': 'ટોંગા', 'guz': 'Tonga', 'ha': 'Tonga', 'he': 'טונגה', 'hi': 'टोंगा', 'hi-Latn': 'Tonga', 'hr': 'Tonga', 'hsb': 'Tonga', 'ht': 'Tonga', 'hu': 'Tonga', 'hy': 'Տոնգա', 'ia': 'Tonga', 'id': 'Tonga', 'ig': 'Tonga', 'is': 'Tonga', 'it': 'Tonga', 'ja': 'トンガ', 'jmc': 'Tonga', 'jv': 'Tonga', 'ka': 'ტონგა', 'kaa': 'Тонга', 'kaa-Cyrl': 'Тонга', 'kab': 'Ṭunga', 'kam': 'Tonga', 'kde': 'Tonga', 'kea': 'Tonga', 'kgp': 'Tãga', 'khq': 'Tonga', 'ki': 'Tonga', 'kk': 'Тонга', 'kk-Cyrl': 'Тонга', 'kln': 'Emetab Tonga', 'km': 'តុងហ្គា', 'kn': 'ಟೊಂಗಾ', 'ko': '통가', 'kok': 'टोंगा', 'kok-Deva': 'टोंगा', 'ks': 'ٹونگا', 'ks-Arab': 'ٹونگا', 'ksb': 'Tonga', 'ksf': 'tɔŋa', 'ksh': 'Tongga', 'ku': 'Tonga', 'kxv': 'ṭonga', 'kxv-Deva': 'टोंगा', 'kxv-Latn': 'ṭonga', 'kxv-Orya': 'ଟଙ୍ଗ', 'kxv-Telu': 'టోంగా', 'ky': 'Тонга', 'lag': 'Tóonga', 'lb': 'Tonga', 'lg': 'Tonga', 'lld': 'Tonga', 'ln': 'Tonga', 'lo': 'ທອງກາ', 'lt': 'Tonga', 'lu': 'Tonga', 'luo': 'Tonga', 'luy': 'Tonga', 'lv': 'Tonga', 'mai': 'टोंगा', 'mas': 'Tonga', 'mer': 'Tonga', 'mfe': 'Tonga', 'mg': 'Tongà', 'mgh': 'Utonga', 'mi': 'Tonga', 'mk': 'Тонга', 'ml': 'ടോംഗ', 'mn': 'Тонга', 'mr': 'टोंगा', 'ms': 'Tonga', 'mt': 'Tonga', 'mua': 'Sǝr Tonga', 'my': 'တွန်ဂါ', 'mzn': 'تونگا', 'naq': 'Tonga', 'nb': 'Tonga', 'nd': 'Thonga', 'ne': 'टोंगा', 'nl': 'Tonga', 'nmg': 'Tɔnga', 'nn': 'Tonga', 'no': 'Tonga', 'nqo': 'ߕߏ߲ߜ߭ߊ߫', 'nyn': 'Tonga', 'om': 'Tonga', 'or': 'ଟୋଙ୍ଗା', 'pa': 'ਟੌਂਗਾ', 'pa-Guru': 'ਟੌਂਗਾ', 'pcm': 'Tónga', 'pl': 'Tonga', 'ps': 'تونګا', 'pt': 'Tonga', 'qu': 'Tonga', 'rm': 'Tonga', 'rn': 'Tonga', 'ro': 'Tonga', 'rof': 'Tonga', 'ru': 'Тонга', 'rw': 'Tonga', 'rwk': 'Tonga', 'saq': 'Tonga', 'sbp': 'Tonga', 'sc': 'Tonga', 'sd': 'ٽونگا', 'sd-Arab': 'ٽونگا', 'se': 'Tonga', 'seh': 'Tonga', 'ses': 'Tonga', 'sg': 'Tonga', 'shi': 'ⵟⵓⵏⴳⴰ', 'shi-Latn': 'ṭunga', 'shi-Tfng': 'ⵟⵓⵏⴳⴰ', 'si': 'ටොංගා', 'sk': 'Tonga', 'sl': 'Tonga', 'smn': 'Tonga', 'sn': 'Tonga', 'so': 'Tonga', 'sq': 'Tonga', 'sr': 'Тонга', 'sr-Cyrl': 'Тонга', 'sr-Latn': 'Tonga', 'sv': 'Tonga', 'sw': 'Tonga', 'syr': 'ܬܘܢܓܐ', 'ta': 'டோங்கா', 'te': 'టోంగా', 'teo': 'Tonga', 'tg': 'Тонга', 'th': 'ตองกา', 'ti': 'ቶንጋ', 'tk': 'Tonga', 'to': 'Tonga', 'tr': 'Tonga', 'tt': 'Тонга', 'twq': 'Tonga', 'tzm': 'Ṭunga', 'ug': 'تونگا', 'uk': 'Тонга', 'ur': 'ٹونگا', 'uz': 'Tonga', 'uz-Cyrl': 'Тонга', 'uz-Latn': 'Tonga', 'vai': 'ꗋꕬ', 'vai-Latn': 'Tɔnga', 'vai-Vaii': 'ꗋꕬ', 'vec': 'Tonga', 'vi': 'Tonga', 'vun': 'Tonga', 'wae': 'Tonga', 'wo': 'Tonga', 'xh': 'ETonga', 'xnr': 'टोंगा', 'xog': 'Tonga', 'yav': 'tɔ́ŋka', 'yi': 'טאנגאַ', 'yo': 'Tonga', 'yrl': 'Tõga', 'yue': '東加', 'yue-Hans': '东加', 'yue-Hant': '東加', 'zgh': 'ⵟⵓⵏⴳⴰ', 'zh': '汤加', 'zh-Hans': '汤加', 'zh-Hant': '東加', 'zu': 'i-Tonga'},
    'TP': {'en': 'East Timor'},
    'TR': {'af': 'Turkye', 'agq': 'Teekì', 'ak': 'Tɛɛki', 'am': 'ቱርክ', 'ar': 'تركيا', 'as': 'তুৰ্কিয়ে', 'asa': 'Uturuki', 'ast': 'Turquía', 'az': 'Türkiyə', 'az-Cyrl': 'Түркијә', 'az-Latn': 'Türkiyə', 'bal-Latn': 'Turkiye', 'bas': 'Tùrkây', 'be': 'Турцыя', 'be-tarask': 'Турцыя', 'bez': 'Huuturuki', 'bg': 'Турция', 'blo': 'Turkii', 'bm': 'Turiki', 'bn': 'তুরস্ক', 'br': 'Turkia', 'brx': 'तुर्की', 'bs': 'Turska', 'bs-Cyrl': 'Турска', 'bs-Latn': 'Turska', 'ca': 'Turquia', 'ca-ES-valencia': 'Turquia', 'ccp': '𑄖𑄪𑄢𑄧𑄌𑄴𑄇𑄧', 'ce': 'Туркойчоь', 'ceb': 'Turkiye', 'cgg': 'Buturuki /Take', 'chr': 'ᎬᏃ', 'ckb': 'تورکیا', 'cs': 'Turecko', 'cv': 'Турци', 'cy': 'Twrci', 'da': 'Tyrkiet', 'dav': 'Uturuki', 'de': 'Türkei', 'dje': 'Turki', 'dsb': 'Turkojska', 'dz': 'ཊཱར་ཀི', 'ebu': 'Uturuki', 'ee': 'Tɛki nutome', 'el': 'Τουρκία', 'el-polyton': 'Τουρκία', 'en': 'Türkiye', 'en-Dsrt': '𐐓𐐲𐑉𐐿𐐨', 'eo': 'Turkujo', 'es': 'Turquía', 'et': 'Türgi', 'eu': 'Turkia', 'ewo': 'Turəkí', 'fa': 'ترکیه', 'ff': 'Turkii', 'ff-Adlm': '𞤚𞤵𞤪𞤳𞤭𞤴𞤢𞥄', 'ff-Latn': 'Turkii', 'fi': 'Turkki', 'fil': 'Türkiye', 'fo': 'Turkaland', 'fr': 'Turquie', 'fur': 'Turchie', 'fy': 'Turkije', 'ga': 'an Tuirc', 'gd': 'An Tuirc', 'gl': 'Turquía', 'gsw': 'Türggei', 'gu': 'તુર્કિયે', 'guz': 'Uturuki', 'ha': 'Turkiyya', 'he': 'טורקיה', 'hi': 'तुर्किये', 'hi-Latn': 'Turkiye', 'hr': 'Turska', 'hsb': 'Turkowska', 'ht': 'Turquie', 'hu': 'Törökország', 'hy': 'Թուրքիա', 'ia': 'Turchia', 'id': 'Turki', 'ig': 'Türkiye', 'is': 'Tyrkland', 'it': 'Turchia', 'ja': 'トルコ', 'jmc': 'Uturuki', 'jv': 'Turki', 'ka': 'თურქეთი', 'kaa': 'Түркия', 'kaa-Cyrl': 'Түркия', 'kab': 'Ṭurk', 'kam': 'Uturuki', 'kde': 'Utuluchi', 'kea': 'Turkia', 'kgp': 'Turkija', 'khq': 'Turki', 'ki': 'Uturuki', 'kk': 'Түркия', 'kk-Cyrl': 'Түркия', 'kln': 'Emetab Turkey', 'km': 'តួកគី', 'kn': 'ತುರ್ಕಿಯೆ', 'ko': '튀르키예', 'kok': 'तुर्की', 'kok-Deva': 'तुर्की', 'ks': 'تُرکی', 'ks-Arab': 'تُرکی', 'ksb': 'Utuluki', 'ksf': 'tirkí', 'ksh': 'de Törkei', 'ku': 'Tirkîye', 'kxv': 'turkī', 'kxv-Deva': 'तुर्की', 'kxv-Latn': 'turkī', 'kxv-Orya': 'ତୁର୍କି', 'kxv-Telu': 'టర్కీ', 'ky': 'Түркия', 'lag': 'Uturúuki', 'lb': 'Tierkei', 'lg': 'Ttake', 'lld': 'Turchia', 'ln': 'Tiliki', 'lo': 'ເທີຄີ', 'lt': 'Turkija', 'lu': 'Tuluki', 'luo': 'Turkey', 'luy': 'Uturuki', 'lv': 'Turcija', 'mai': 'तुर्की', 'mas': 'Uturuki', 'mer': 'Takĩ', 'mfe': 'Tirki', 'mg': 'Torkia', 'mgh': 'Utuki', 'mi': 'Tākei', 'mk': 'Турција', 'ml': 'തുർക്കിയെ', 'mn': 'Турк', 'mr': 'तुर्किये', 'ms': 'Turkiye', 'mt': 'it-Turkija', 'mua': 'Turkiya', 'my': 'တူရကီ', 'mzn': 'ترکیه', 'naq': 'Turkeieb', 'nb': 'Tyrkia', 'nd': 'Thekhi', 'ne': 'तुर्किये', 'nl': 'Turkije', 'nmg': 'Turki', 'nn': 'Tyrkia', 'no': 'Tyrkia', 'nqo': 'ߕߎߙߑߞߌ߫', 'nyn': 'Buturuki /Take', 'om': 'Tarkiye', 'or': 'ତୁର୍କୀ', 'pa': 'ਤੁਰਕੀ', 'pa-Guru': 'ਤੁਰਕੀ', 'pcm': 'Tọ́ki', 'pl': 'Turcja', 'ps': 'ترکي', 'pt': 'Turquia', 'qu': 'Turquía', 'rm': 'Tirchia', 'rn': 'Turukiya', 'ro': 'Turcia', 'rof': 'Uturuki', 'ru': 'Турция', 'rwk': 'Uturuki', 'saq': 'Uturuki', 'sbp': 'Utuluki', 'sc': 'Turchia', 'sd': 'ترڪييي', 'sd-Arab': 'ترڪييي', 'se': 'Durka', 'seh': 'Turquia', 'ses': 'Turki', 'sg': 'Turukïi', 'shi': 'ⵜⵓⵔⴽⵢⴰ', 'shi-Latn': 'turkya', 'shi-Tfng': 'ⵜⵓⵔⴽⵢⴰ', 'si': 'තුර්කිය', 'sk': 'Turecko', 'sl': 'Turčija', 'smn': 'Turkki', 'sn': 'Turkey', 'so': 'Turki', 'sq': 'Turqi', 'sr': 'Турска', 'sr-Cyrl': 'Турска', 'sr-Latn': 'Turska', 'sv': 'Turkiet', 'sw': 'Uturuki', 'syr': 'ܬܘܪܟܝܐ', 'ta': 'துருக்கியே', 'te': 'తుర్కియె', 'teo': 'Uturuki', 'tg': 'Туркия', 'th': 'ตุรกี', 'ti': 'ቱርኪ', 'tk': 'Türkiýe', 'to': 'Toake', 'tr': 'Türkiye', 'tt': 'Төркия', 'twq': 'Turki', 'tzm': 'Turkya', 'ug': 'تۈركىيە', 'uk': 'Туреччина', 'ur': 'ترکیہ', 'uz': 'Turkiya', 'uz-Cyrl': 'Туркия', 'uz-Latn': 'Turkiya', 'vai': 'ꗋꖺꕃ', 'vai-Latn': 'Tɔ́ɔ́ki', 'vai-Vaii': 'ꗋꖺꕃ', 'vec': 'Turchia', 'vi': 'Thổ Nhĩ Kỳ', 'vun': 'Uturuki', 'wae': 'Türkei', 'wo': 'Tirki', 'xh': 'ETurkey', 'xnr': 'तुर्की', 'xog': 'Ttake', 'yav': 'tulukíi', 'yi': 'טערקיי', 'yo': 'Tọọki', 'yrl': 'Tukíya', 'yue': '土耳其', 'yue-Hans': '土耳其', 'yue-Hant': '土耳其', 'zgh': 'ⵜⵓⵔⴽⵢⴰ', 'zh': '土耳其', 'zh-Hans': '土耳其', 'zh-Hant': '土耳其', 'zu': 'i-Turkey'},
    'TT': {'af': 'Trinidad en Tobago', 'agq': 'Tèlenedà à Tòbagù', 'ak': 'Trinidad ne Tobago', 'am': 'ትሪናዳድ እና ቶቤጎ', 'ar': 'ترينيداد وتوباغو', 'as': 'ট্ৰিনিডাড আৰু টোবাগো', 'asa': 'Trinidad na Tobago', 'ast': 'Trinidá y Tobagu', 'az': 'Trinidad və Tobaqo', 'az-Cyrl': 'Тринидад вә Тобаго', 'az-Latn': 'Trinidad və Tobaqo', 'bal-Latn': 'Trinidad o Tobágó', 'bas': 'Trìnidàd ɓɔ Tòbagò', 'be': 'Трынідад і Табага', 'be-tarask': 'Трынідад і Табага', 'bez': 'Hutrinad na Hutobago', 'bg': 'Тринидад и Тобаго', 'blo': 'Triniɖaaɖ na Tobagoo', 'bm': 'Trinite-ni-Tobago', 'bn': 'ত্রিনিনাদ ও টোব্যাগো', 'br': 'Trinidad ha Tobago', 'brx': 'त्रिनिडाड आरो तबेग’', 'bs': 'Trinidad i Tobago', 'bs-Cyrl': 'Тринидад и Тобаго', 'bs-Latn': 'Trinidad i Tobago', 'ca': 'Trinidad i Tobago', 'ca-ES-valencia': 'Trinidad i Tobago', 'ccp': '𑄖𑄳𑄢𑄨𑄚𑄨𑄚𑄘𑄴 𑄃𑄮 𑄑𑄮𑄝𑄳𑄠𑄉𑄮', 'ce': 'Тринидад а, Тобаго а', 'ceb': 'Trinidad & Tobago', 'cgg': 'Turinidad na Tobago', 'chr': 'ᏟᏂᏕᏗ ᎠᎴ ᏙᏆᎪ', 'ckb': 'ترینیداد و تۆباگو', 'cs': 'Trinidad a Tobago', 'cv': 'Тринидад тата Тобаго', 'cy': 'Trinidad a Tobago', 'da': 'Trinidad og Tobago', 'dav': 'Trinidad na Tobago', 'de': 'Trinidad und Tobago', 'dje': 'Trinidad nda Tobaago', 'dsb': 'Trinidad a Tobago', 'dz': 'ཊི་ནི་ཌཱཌ་ ཨེནཌ་ ཊོ་བྷེ་གྷོ', 'ebu': 'Trinidad na Tobago', 'ee': 'Trinidad kple Tobago nutome', 'el': 'Τρινιντάντ και Τομπάγκο', 'el-polyton': 'Τρινιδὰδ καὶ Τομπάγκο', 'en': 'Trinidad & Tobago', 'en-Dsrt': '𐐓𐑉𐐮𐑌𐐮𐐼𐐰𐐼 𐐰𐑌𐐼 𐐓𐐲𐐺𐐩𐑀𐐬', 'eo': 'Trinidado kaj Tobago', 'es': 'Trinidad y Tobago', 'et': 'Trinidad ja Tobago', 'eu': 'Trinidad eta Tobago', 'ewo': 'Təlinité-ai-Tobágo', 'fa': 'ترینیداد و توباگو', 'ff': 'Tirnidaad e Tobaago', 'ff-Adlm': '𞤚𞤭𞤪𞤲𞤭𞤣𞤢𞥄𞤣 & 𞤚𞤮𞤦𞤢𞤺𞤮𞥅', 'ff-Latn': 'Tirnidaad e Tobaago', 'fi': 'Trinidad ja Tobago', 'fil': 'Trinidad & Tobago', 'fo': 'Trinidad & Tobago', 'fr': 'Trinité-et-Tobago', 'fur': 'Trinidad e Tobago', 'fy': 'Trinidad en Tobago', 'ga': 'Oileán na Tríonóide agus Tobága', 'gd': 'Trianaid agus Tobago', 'gl': 'Trinidad e Tobago', 'gsw': 'Trinidad und Tobaago', 'gu': 'ટ્રિનીદાદ અને ટોબેગો', 'guz': 'Trinidad na Tobago', 'ha': 'Tirinidad Da Tobago', 'he': 'טרינידד וטובגו', 'hi': 'त्रिनिदाद और टोबैगो', 'hi-Latn': 'Trinidad & Tobago', 'hr': 'Trinidad i Tobago', 'hsb': 'Trinidad a Tobago', 'ht': 'Trinité-et-Tobago', 'hu': 'Trinidad és Tobago', 'hy': 'Տրինիդադ և Տոբագո', 'ia': 'Trinidad e Tobago', 'id': 'Trinidad dan Tobago', 'ie': 'Trinidad e Tobago', 'ig': 'Trinidad na Tobago', 'is': 'Trínidad og Tóbagó', 'it': 'Trinidad e Tobago', 'ja': 'トリニダード・トバゴ', 'jmc': 'Trinidad na Tobago', 'jv': 'Trinidad lan Tobago', 'ka': 'ტრინიდადი და ტობაგო', 'kaa': 'Тринидад ҳәм Тобаго', 'kaa-Cyrl': 'Тринидад ҳәм Тобаго', 'kab': 'Ṭrindad d Ṭubagu', 'kam': 'Trinidad na Tobago', 'kde': 'Tilinidad na Tobago', 'kea': 'Trinidad i Tobagu', 'kgp': 'Trĩnane kar Tomagu', 'khq': 'Trinidad nda Tobaago', 'ki': 'Trinidad na Tobago', 'kk': 'Тринидад және Тобаго', 'kk-Cyrl': 'Тринидад және Тобаго', 'kln': 'Emetab Trinidad ak Tobago', 'km': 'ទ្រីនីដាត និង\u200bតូបាហ្គោ', 'kn': 'ಟ್ರಿನಿಡಾಡ್ ಮತ್ತು ಟೊಬಾಗೊ', 'ko': '트리니다드 토바고', 'kok': 'ट्रिनीदाद आनी टोबॅगो', 'kok-Deva': 'ट्रिनीदाद आनी टोबॅगो', 'ks': 'ٹرنِنداد تہٕ ٹوبیگو', 'ks-Arab': 'ٹرنِنداد تہٕ ٹوبیگو', 'ksb': 'Tlinidad na Tobago', 'ksf': 'tɛrinitɛ ri tobago', 'ksh': 'Trinidad un Tobähjo', 'ku': 'Trînîdad û Tobago', 'kxv': 'ṭriniḍaḍ aḍe ṭobego', 'kxv-Deva': 'त्रिनिदाद ऑड़े टोबेगो', 'kxv-Latn': 'ṭriniḍaḍ aḍe ṭobego', 'kxv-Orya': 'ତ୍ରିନିଡାଡ ଅଡ଼େ ଟବାଗ', 'kxv-Telu': 'ట్రినిడాడ్ మరియు టొబాగో', 'ky': 'Тринидад жана Тобаго', 'lag': 'Tiriníida ya Tobáago', 'lb': 'Trinidad an Tobago', 'lg': 'Turindaadi ne Tobago', 'lld': 'Trinidad y Tobago', 'ln': 'Tinidadɛ mpé Tobago', 'lo': 'ທຣິນິແດດ ແລະ ໂທແບໂກ', 'lt': 'Trinidadas ir Tobagas', 'lu': 'Tinidade ne Tobago', 'luo': 'Trinidad gi Tobago', 'luy': 'Trinidad na Tobago', 'lv': 'Trinidāda un Tobāgo', 'mai': 'ट्रिनिदाद आ टोबैगो', 'mas': 'Trinidad na Tobago', 'mer': 'Trinidad na Tobago', 'mfe': 'Trinite-ek-Tobago', 'mg': 'Trinidad sy Tobagô', 'mgh': 'Utrinidad na Tobago', 'mi': 'Tirinaki Tōpako', 'mk': 'Тринидад и Тобаго', 'ml': 'ട്രിനിഡാഡും ടുബാഗോയും', 'mn': 'Тринидад ба Тобаго', 'mr': 'त्रिनिदाद आणि टोबॅगो', 'ms': 'Trinidad dan Tobago', 'mt': 'Trinidad u Tobago', 'mua': 'Trinite ne Tobago', 'my': 'ထရီနီဒတ်နှင့် တိုဘက်ဂို', 'mzn': 'ترینیداد و توباگو', 'naq': 'Trinidad and Tobago', 'nb': 'Trinidad og Tobago', 'nd': 'Trinidad le Tobago', 'ne': 'त्रिनिडाड एण्ड टोबागो', 'nl': 'Trinidad en Tobago', 'nmg': 'Trinidad ba Tobágó', 'nn': 'Trinidad og Tobago', 'no': 'Trinidad og Tobago', 'nqo': 'ߕߙߌߣߌߕߋ߫ ߣߌ߫ ߕߏߓߊߜ߭ߏ߫', 'nyn': 'Turinidad na Tobago', 'om': 'Tirinidan fi Tobaagoo', 'or': 'ତ୍ରିନିଦାଦ୍ ଏବଂ ଟୋବାଗୋ', 'pa': 'ਟ੍ਰਿਨੀਡਾਡ ਅਤੇ ਟੋਬਾਗੋ', 'pa-Guru': 'ਟ੍ਰਿਨੀਡਾਡ ਅਤੇ ਟੋਬਾਗੋ', 'pcm': 'Trínídad & Tobágo', 'pl': 'Trynidad i Tobago', 'ps': 'ټرينيډاډ او ټوباګو', 'pt': 'Trinidad e Tobago', 'qu': 'Trinidad y Tobago', 'rm': 'Trinidad e Tobago', 'rn': 'Tirinidadi na Tobago', 'ro': 'Trinidad și Tobago', 'rof': 'Trinidad na Tobago', 'ru': 'Тринидад и Тобаго', 'rwk': 'Trinidad na Tobago', 'saq': 'Trinidad na Tobago', 'sbp': 'Tilinidadi ni Tobago', 'sc': 'Trinidad e Tobago', 'sd': 'ٽريني ڊيڊ ۽ ٽوباگو ٻيٽ', 'sd-Arab': 'ٽريني ڊيڊ ۽ ٽوباگو ٻيٽ', 'se': 'Trinidad ja Tobago', 'seh': 'Trinidad e Tobago', 'ses': 'Trinidad nda Tobaago', 'sg': 'Trinitùee na Tobagö', 'shi': 'ⵜⵔⵉⵏⵉⴷⴰⴷ ⴷ ⵟⵓⴱⴰⴳⵓ', 'shi-Latn': 'trinidad d ṭubagu', 'shi-Tfng': 'ⵜⵔⵉⵏⵉⴷⴰⴷ ⴷ ⵟⵓⴱⴰⴳⵓ', 'si': 'ට්\u200dරිනිඩෑඩ් සහ ටොබැගෝ', 'sk': 'Trinidad a Tobago', 'sl': 'Trinidad in Tobago', 'smn': 'Trinidad já Tobago', 'sn': 'Trinidad and Tobago', 'so': 'Tirinidaad & Tobago', 'sq': 'Trinidad e Tobago', 'sr': 'Тринидад и Тобаго', 'sr-Cyrl': 'Тринидад и Тобаго', 'sr-Latn': 'Trinidad i Tobago', 'sv': 'Trinidad och Tobago', 'sw': 'Trinidad na Tobago', 'syr': 'ܬܪܝܢܝܕܐܕ ܘܬܘܒܐܓܘ', 'ta': 'டிரினிடாட் & டொபாகோ', 'te': 'ట్రినిడాడ్ మరియు టొబాగో', 'teo': 'Trinidad na Tobago', 'tg': 'Тринидад ва Тобаго', 'th': 'ตรินิแดดและโตเบโก', 'ti': 'ትሪኒዳድን ቶባጎን', 'tk': 'Trinidad we Tobago', 'to': 'Tilinitati mo Topako', 'tr': 'Trinidad ve Tobago', 'tt': 'Тринидад һәм Тобаго', 'twq': 'Trinidad nda Tobaago', 'tzm': 'Trinidad d Ṭubagu', 'ug': 'تىرىنىداد ۋە توباگو', 'uk': 'Тринідад і Тобаго', 'ur': 'ترینیداد اور ٹوباگو', 'uz': 'Trinidad va Tobago', 'uz-Cyrl': 'Тринидад ва Тобаго', 'uz-Latn': 'Trinidad va Tobago', 'vai': 'ꖤꔸꔕꕜ ꗪ ꕿꔆꖑ', 'vai-Latn': 'Turindeda ɓɛ́ Tobhego', 'vai-Vaii': 'ꖤꔸꔕꕜ ꗪ ꕿꔆꖑ', 'vec': 'Trinidà e Tabaco', 'vi': 'Trinidad và Tobago', 'vun': 'Trinidad na Tobago', 'wae': 'Trinidad und Tobago', 'wo': 'Tirinite ak Tobago', 'xh': 'ETrinidad & Tobago', 'xnr': 'त्रिनिदाद कने टोबैगो', 'xog': 'Turindaadi ni Tobago', 'yav': 'tilinitáat na tupákɔ', 'yi': 'טרינידאַד און טאבאַגא', 'yo': 'Tirinida ati Tobaga', 'yrl': 'Tirinidadi asuí Tobagu', 'yue': '千里達同多巴哥', 'yue-Hans': '千里达同多巴哥', 'yue-Hant': '千里達同多巴哥', 'zgh': 'ⵜⵔⵉⵏⵉⴷⴰⴷ ⴷ ⵟⵓⴱⴰⴳⵓ', 'zh': '特立尼达和多巴哥', 'zh-Hans': '特立尼达和多巴哥', 'zh-Hant': '千里達及托巴哥', 'zu': 'i-Trinidad ne-Tobago'},
    'TV': {'af': 'Tuvalu', 'agq': 'Tuwvalùw', 'ak': 'Tuvalu', 'am': 'ቱቫሉ', 'ar': 'توفالو', 'as': 'টুভালু', 'asa': 'Tuvalu', 'ast': 'Tuvalu', 'az': 'Tuvalu', 'az-Cyrl': 'Тувалу', 'az-Latn': 'Tuvalu', 'bal-Latn': 'Tuwalu', 'bas': 'Tùvàlù', 'be': 'Тувалу', 'be-tarask': 'Тувалу', 'bez': 'Hutuvalu', 'bg': 'Тувалу', 'blo': 'Tufalu', 'bm': 'Tuvalu', 'bn': 'টুভালু', 'br': 'Tuvalu', 'brx': 'तुभालु', 'bs': 'Tuvalu', 'bs-Cyrl': 'Тувалу', 'bs-Latn': 'Tuvalu', 'ca': 'Tuvalu', 'ca-ES-valencia': 'Tuvalu', 'ccp': '𑄑𑄪𑄞𑄣𑄪', 'ce': 'Тувалу', 'ceb': 'Tuvalu', 'cgg': 'Tuvaru', 'chr': 'ᏚᏩᎷ', 'ckb': 'تووڤالوو', 'cs': 'Tuvalu', 'cv': 'Тувалу', 'cy': 'Tuvalu', 'da': 'Tuvalu', 'dav': 'Tuvalu', 'de': 'Tuvalu', 'dje': 'Tuvalu', 'dsb': 'Tuvalu', 'dz': 'ཏུ་ཝ་ལུ', 'ebu': 'Tuvalu', 'ee': 'Tuvalu nutome', 'el': 'Τουβαλού', 'el-polyton': 'Τουβαλού', 'en': 'Tuvalu', 'en-Dsrt': '𐐓𐐲𐑂𐐪𐑊𐐭', 'eo': 'Tuvalo', 'es': 'Tuvalu', 'et': 'Tuvalu', 'eu': 'Tuvalu', 'ewo': 'Tuvalú', 'fa': 'تووالو', 'ff': 'Tuwaluu', 'ff-Adlm': '𞤚𞤵𞥅𞤾𞤢𞤤𞤵', 'ff-Latn': 'Tuwaluu', 'fi': 'Tuvalu', 'fil': 'Tuvalu', 'fo': 'Tuvalu', 'fr': 'Tuvalu', 'fur': 'Tuvalu', 'fy': 'Tuvalu', 'ga': 'Túvalú', 'gd': 'Tubhalu', 'gl': 'Tuvalu', 'gsw': 'Tuvalu', 'gu': 'તુવાલુ', 'guz': 'Tuvalu', 'ha': 'Tubalu', 'he': 'טובאלו', 'hi': 'तुवालू', 'hi-Latn': 'Tuvalu', 'hr': 'Tuvalu', 'hsb': 'Tuvalu', 'ht': 'Tuvalu', 'hu': 'Tuvalu', 'hy': 'Տուվալու', 'ia': 'Tuvalu', 'id': 'Tuvalu', 'ie': 'Tuvalu', 'ig': 'Tuvalu', 'is': 'Túvalú', 'it': 'Tuvalu', 'ja': 'ツバル', 'jmc': 'Tuvalu', 'jv': 'Tuvalu', 'ka': 'ტუვალუ', 'kaa': 'Тувалу', 'kaa-Cyrl': 'Тувалу', 'kab': 'Ṭuvalu', 'kam': 'Tuvalu', 'kde': 'Tuvalu', 'kea': 'Tuvalu', 'kgp': 'Tuvaru', 'khq': 'Tuvalu', 'ki': 'Tuvalu', 'kk': 'Тувалу', 'kk-Cyrl': 'Тувалу', 'kln': 'Emetab Tuvalu', 'km': 'ទូវ៉ាលូ', 'kn': 'ಟುವಾಲು', 'ko': '투발루', 'kok': 'टुवालू', 'kok-Deva': 'टुवालू', 'ks': 'توٗوالوٗ', 'ks-Arab': 'توٗوالوٗ', 'ksb': 'Tuvalu', 'ksf': 'tuwalu', 'ksh': 'Tuvalu', 'ku': 'Tûvalû', 'kxv': 'tuvala', 'kxv-Deva': 'तुवालू', 'kxv-Latn': 'tuvala', 'kxv-Orya': 'ତୁୱାଲୁ', 'kxv-Telu': 'టువాలు', 'ky': 'Тувалу', 'lag': 'Tuváalu', 'lb': 'Tuvalu', 'lg': 'Tuvalu', 'lld': 'Tuvalu', 'ln': 'Tuvalu', 'lo': 'ຕູວາລູ', 'lt': 'Tuvalu', 'lu': 'Tuvalu', 'luo': 'Tuvalu', 'luy': 'Tuvalu', 'lv': 'Tuvalu', 'mai': 'टुवालू', 'mas': 'Tuvalu', 'mer': 'Tuvalu', 'mfe': 'Tuvalu', 'mg': 'Tovalò', 'mgh': 'Utuvalu', 'mi': 'Tūwaru', 'mk': 'Тувалу', 'ml': 'ടുവാലു', 'mn': 'Тувалу', 'mr': 'टुवालु', 'ms': 'Tuvalu', 'mt': 'Tuvalu', 'mua': 'Sǝr Tuvalu', 'my': 'တူဗားလူ', 'mzn': 'تووالو', 'naq': 'Tuvalu', 'nb': 'Tuvalu', 'nd': 'Thuvalu', 'ne': 'तुभालु', 'nl': 'Tuvalu', 'nmg': 'Tuvalú', 'nn': 'Tuvalu', 'no': 'Tuvalu', 'nqo': 'ߕߎߝ߭ߊߟߎ߫', 'nyn': 'Tuvaru', 'om': 'Tuvalu', 'or': 'ତୁଭାଲୁ', 'pa': 'ਟੁਵਾਲੂ', 'pa-Guru': 'ਟੁਵਾਲੂ', 'pcm': 'Tuválu', 'pl': 'Tuvalu', 'ps': 'توالیو', 'pt': 'Tuvalu', 'qu': 'Tuvalu', 'rm': 'Tuvalu', 'rn': 'Tuvalu', 'ro': 'Tuvalu', 'rof': 'Tuvalu', 'ru': 'Тувалу', 'rwk': 'Tuvalu', 'saq': 'Tuvalu', 'sbp': 'Tuvalu', 'sc': 'Tuvalu', 'sd': 'توالو', 'sd-Arab': 'توالو', 'se': 'Tuvalu', 'seh': 'Tuvalu', 'ses': 'Tuvalu', 'sg': 'Tüvalü', 'shi': 'ⵜⵓⴼⴰⵍⵓ', 'shi-Latn': 'tufalu', 'shi-Tfng': 'ⵜⵓⴼⴰⵍⵓ', 'si': 'ටුවාලූ', 'sk': 'Tuvalu', 'sl': 'Tuvalu', 'smn': 'Tuvalu', 'sn': 'Tuvalu', 'so': 'Tufaalu', 'sq': 'Tuvalu', 'sr': 'Тувалу', 'sr-Cyrl': 'Тувалу', 'sr-Latn': 'Tuvalu', 'sv': 'Tuvalu', 'sw': 'Tuvalu', 'syr': 'ܬܘܒܐܠܘ', 'ta': 'துவாலு', 'te': 'టువాలు', 'teo': 'Tuvalu', 'tg': 'Тувалу', 'th': 'ตูวาลู', 'ti': 'ቱቫሉ', 'tk': 'Tuwalu', 'to': 'Tūvalu', 'tr': 'Tuvalu', 'tt': 'Тувалу', 'twq': 'Tuvalu', 'tzm': 'Ṭuvalu', 'ug': 'تۇۋالۇ', 'uk': 'Тувалу', 'ur': 'ٹووالو', 'uz': 'Tuvalu', 'uz-Cyrl': 'Тувалу', 'uz-Latn': 'Tuvalu', 'vai': 'ꕚꖣꖨ', 'vai-Latn': 'Tuválu', 'vai-Vaii': 'ꕚꖣꖨ', 'vec': 'Tuvalu', 'vi': 'Tuvalu', 'vun': 'Tuvalu', 'wae': 'Tuvalu', 'wo': 'Tuwalo', 'xh': 'ETuvalu', 'xnr': 'तुवालू', 'xog': 'Tuvalu', 'yav': 'tufalú', 'yi': 'טואוואַלו', 'yo': 'Tufalu', 'yrl': 'Tuwaru', 'yue': '吐瓦魯', 'yue-Hans': '吐瓦鲁', 'yue-Hant': '吐瓦魯', 'zgh': 'ⵜⵓⴼⴰⵍⵓ', 'zh': '图瓦卢', 'zh-Hans': '图瓦卢', 'zh-Hant': '吐瓦魯', 'zu': 'i-Tuvalu'},
    'TW': {'af': 'Taiwan', 'agq': 'Taewàn', 'ak': 'Taiwan', 'am': 'ታይዋን', 'ar': 'تايوان', 'as': 'টাইৱান', 'asa': 'Taiwani', 'ast': 'Taiwán', 'az': 'Tayvan', 'az-Cyrl': 'Тајван', 'az-Latn': 'Tayvan', 'bal-Latn': 'Táiwán', 'bas': 'Tàywân', 'be': 'Тайвань', 'be-tarask': 'Тайвань', 'bez': 'Hutaiwani', 'bg': 'Тайван', 'blo': 'Taɩwan', 'bm': 'Tayiwani', 'bn': 'তাইওয়ান', 'br': 'Taiwan', 'brx': 'ताइवान', 'bs': 'Tajvan', 'bs-Cyrl': 'Тајван', 'bs-Latn': 'Tajvan', 'ca': 'Taiwan', 'ca-ES-valencia': 'Taiwan', 'ccp': '𑄖𑄭𑄤𑄚𑄴', 'ce': 'Тайвань', 'ceb': 'Taiwan', 'cgg': 'Tayiwaani', 'chr': 'ᏔᎢᏩᏂ', 'ckb': 'تایوان', 'cs': 'Tchaj-wan', 'cv': 'Тайвань', 'cy': 'Taiwan', 'da': 'Taiwan', 'dav': 'Taiwani', 'de': 'Taiwan', 'dje': 'Taayiwan', 'dsb': 'Taiwan', 'dz': 'ཊཱའི་ཝཱན', 'ebu': 'Taiwani', 'ee': 'Taiwan nutome', 'el': 'Ταϊβάν', 'el-polyton': 'Ταϊβάν', 'en': 'Taiwan', 'en-Dsrt': '𐐓𐐴𐐶𐐪𐑌', 'eo': 'Tajvano', 'es': 'Taiwán', 'et': 'Taiwan', 'eu': 'Taiwan', 'ewo': 'Taiwán', 'fa': 'تایوان', 'ff': 'Taywaan', 'ff-Adlm': '𞤚𞤢𞤴𞤱𞤢𞥄𞤲', 'ff-Latn': 'Taywaan', 'fi': 'Taiwan', 'fil': 'Taiwan', 'fo': 'Taivan', 'fr': 'Taïwan', 'fur': 'Taiwan', 'fy': 'Taiwan', 'ga': 'an Téaváin', 'gd': 'Taidh-Bhàn', 'gl': 'Taiwán', 'gsw': 'Taiwan', 'gu': 'તાઇવાન', 'guz': 'Taiwani', 'ha': 'Taiwan', 'he': 'טייוואן', 'hi': 'ताइवान', 'hi-Latn': 'Taiwan', 'hr': 'Tajvan', 'hsb': 'Taiwan', 'ht': 'Taïwan', 'hu': 'Tajvan', 'hy': 'Թայվան', 'ia': 'Taiwan', 'id': 'Taiwan', 'ig': 'Taiwan', 'is': 'Taívan', 'it': 'Taiwan', 'ja': '台湾', 'jmc': 'Taiwani', 'jv': 'Taiwan', 'ka': 'ტაივანი', 'kaa': 'Тайвань', 'kaa-Cyrl': 'Тайвань', 'kab': 'Ṭaywan', 'kam': 'Taiwani', 'kde': 'Taiwani', 'kea': 'Taiuan', 'kgp': 'Tajuvỹ', 'khq': 'Taayiwan', 'ki': 'Taiwani', 'kk': 'Тайвань', 'kk-Cyrl': 'Тайвань', 'kln': 'Emetab Taiwan', 'km': 'តៃវ៉ាន់', 'kn': 'ತೈವಾನ್', 'ko': '대만', 'kok': 'तायवान', 'kok-Deva': 'तायवान', 'ks': 'تایوان', 'ks-Arab': 'تایوان', 'ksb': 'Taiwani', 'ksf': 'tɛwán', 'ksh': 'Taiwan', 'ku': 'Taywan', 'kxv': 'taivan', 'kxv-Deva': 'ताइवान', 'kxv-Latn': 'taivan', 'kxv-Orya': 'ତାଇୱାନ', 'kxv-Telu': 'తైవాన్', 'ky': 'Тайвань', 'lag': 'Taiwáani', 'lb': 'Taiwan', 'lg': 'Tayiwani', 'lld': 'Taiwan', 'ln': 'Taiwanin', 'lo': 'ໄຕ້ຫວັນ', 'lt': 'Taivanas', 'lu': 'Taiwani', 'luo': 'Taiwan', 'luy': 'Taiwani', 'lv': 'Taivāna', 'mai': 'ताइबान', 'mas': 'Taiwani', 'mer': 'Taiwani', 'mfe': 'Taiwan', 'mg': 'Taioana', 'mi': 'Taiwana', 'mk': 'Тајван', 'ml': 'തായ്\u200cവാൻ', 'mn': 'Тайвань', 'mr': 'तैवान', 'ms': 'Taiwan', 'mt': 'it-Tajwan', 'mua': 'Taiwaŋ', 'my': 'ထိုင်ဝမ်', 'mzn': 'تایوان', 'naq': 'Taiwan', 'nb': 'Taiwan', 'nd': 'Thayiwani', 'ne': 'ताइवान', 'nl': 'Taiwan', 'nmg': 'Taïwan', 'nn': 'Taiwan', 'no': 'Taiwan', 'nqo': 'ߕߊߦߌߥߊ߲߫', 'nyn': 'Tayiwaani', 'om': 'Taayiwwan', 'or': 'ତାଇୱାନ', 'pa': 'ਤਾਇਵਾਨ', 'pa-Guru': 'ਤਾਇਵਾਨ', 'pcm': 'Taíwán', 'pl': 'Tajwan', 'ps': 'تائيوان', 'pt': 'Taiwan', 'qu': 'Taiwán', 'rm': 'Taiwan', 'rn': 'Tayiwani', 'ro': 'Taiwan', 'rof': 'Taiwani', 'ru': 'Тайвань', 'rwk': 'Taiwani', 'saq': 'Taiwani', 'sbp': 'Taiwani', 'sc': 'Taiwàn', 'sd': 'تائیوان', 'sd-Arab': 'تائیوان', 'se': 'Taiwan', 'seh': 'Taiwan', 'ses': 'Taayiwan', 'sg': 'Tâiwâni', 'shi': 'ⵟⴰⵢⵡⴰⵏ', 'shi-Latn': 'ṭaywan', 'shi-Tfng': 'ⵟⴰⵢⵡⴰⵏ', 'si': 'තායිවානය', 'sk': 'Taiwan', 'sl': 'Tajvan', 'smn': 'Taiwan', 'sn': 'Taiwan', 'so': 'Taywaan', 'sq': 'Tajvan', 'sr': 'Тајван', 'sr-Cyrl': 'Тајван', 'sr-Latn': 'Tajvan', 'sv': 'Taiwan', 'sw': 'Taiwan', 'syr': 'ܬܐܝܘܐܢ', 'ta': 'தைவான்', 'te': 'తైవాన్', 'teo': 'Taiwani', 'tg': 'Тайван', 'th': 'ไต้หวัน', 'ti': 'ታይዋን', 'tk': 'Taýwan', 'to': 'Taiuani', 'tr': 'Tayvan', 'tt': 'Тайвань', 'twq': 'Taayiwan', 'tzm': 'Ṭaywan', 'ug': 'تەيۋەن', 'uk': 'Тайвань', 'ur': 'تائیوان', 'uz': 'Tayvan', 'uz-Cyrl': 'Тайван', 'uz-Latn': 'Tayvan', 'vai': 'ꕚꔤꕎꘋ', 'vai-Latn': 'Taiwaŋ', 'vai-Vaii': 'ꕚꔤꕎꘋ', 'vec': 'Taiwan', 'vi': 'Đài Loan', 'vun': 'Taiwani', 'wae': 'Taiwan', 'wo': 'Taywan', 'xh': 'ETaiwan', 'xnr': 'ताइवान', 'xog': 'Tayiwani', 'yav': 'tayiwáan', 'yo': 'Taiwani', 'yrl': 'Taiwã', 'yue': '台灣', 'yue-Hans': '台湾', 'yue-Hant': '台灣', 'zgh': 'ⵟⴰⵢⵡⴰⵏ', 'zh': '台湾', 'zh-Hans': '台湾', 'zh-Hant': '台灣', 'zu': 'i-Taiwan'},
    'TZ': {'af': 'Tanzanië', 'agq': 'Tàanzanyìa', 'ak': 'Tansania', 'am': 'ታንዛኒያ', 'ar': 'تنزانيا', 'as': 'তাঞ্জানিয়া', 'asa': 'Tadhania', 'ast': 'Tanzania', 'az': 'Tanzaniya', 'az-Cyrl': 'Танзанија', 'az-Latn': 'Tanzaniya', 'bal-Latn': 'Tanzániá', 'bas': 'Tànzànià', 'be': 'Танзанія', 'be-tarask': 'Танзанія', 'bez': 'Hutanzania', 'bg': 'Танзания', 'bgn': 'تانزانیا', 'blo': 'Taŋsanii', 'bm': 'Tanzani', 'bn': 'তাঞ্জানিয়া', 'br': 'Tanzania', 'brx': 'तान्जानिया', 'bs': 'Tanzanija', 'bs-Cyrl': 'Танзанија', 'bs-Latn': 'Tanzanija', 'ca': 'Tanzània', 'ca-ES-valencia': 'Tanzània', 'ccp': '𑄖𑄚𑄴𑄎𑄚𑄨𑄠', 'ce': 'Танзани', 'ceb': 'Tanzania', 'cgg': 'Tanzania', 'chr': 'ᏖᏂᏏᏂᏯ', 'ckb': 'تانزانیا', 'cs': 'Tanzanie', 'cv': 'Танзани', 'cy': 'Tanzania', 'da': 'Tanzania', 'dav': 'Tanzania', 'de': 'Tansania', 'dje': 'Tanzaani', 'dsb': 'Tansanija', 'dz': 'ཊཱན་ཛཱ་ནི་ཡ', 'ebu': 'Tanzania', 'ee': 'Tanzania nutome', 'el': 'Τανζανία', 'el-polyton': 'Τανζανία', 'en': 'Tanzania', 'en-Dsrt': '𐐓𐐰𐑌𐑆𐐲𐑌𐐨𐐲', 'eo': 'Tanzanio', 'es': 'Tanzania', 'et': 'Tansaania', 'eu': 'Tanzania', 'ewo': 'Taŋəzaní', 'fa': 'تانزانیا', 'ff': 'Tansanii', 'ff-Adlm': '𞤚𞤢𞤲𞤧𞤢𞤲𞤭𞥅', 'ff-Latn': 'Tansanii', 'fi': 'Tansania', 'fil': 'Tanzania', 'fo': 'Tansania', 'fr': 'Tanzanie', 'fur': 'Tanzanie', 'fy': 'Tanzania', 'ga': 'an Tansáin', 'gd': 'An Tansan', 'gl': 'Tanzania', 'gsw': 'Tansaniia', 'gu': 'તાંઝાનિયા', 'guz': 'Tanzania', 'ha': 'Tanzaniya', 'he': 'טנזניה', 'hi': 'तंज़ानिया', 'hi-Latn': 'Tanzania', 'hr': 'Tanzanija', 'hsb': 'Tansanija', 'ht': 'Tanzanie', 'hu': 'Tanzánia', 'hy': 'Տանզանիա', 'ia': 'Tanzania', 'id': 'Tanzania', 'ig': 'Tanzania', 'is': 'Tansanía', 'it': 'Tanzania', 'ja': 'タンザニア', 'jgo': 'Tanzanî', 'jmc': 'Tanzania', 'jv': 'Tansania', 'ka': 'ტანზანია', 'kaa': 'Танзания', 'kaa-Cyrl': 'Танзания', 'kab': 'Ṭanzanya', 'kam': 'Tanzania', 'kde': 'Tanzania', 'kea': 'Tanzánia', 'kgp': 'Tỹnjỹnĩja', 'khq': 'Tanzaani', 'ki': 'Tanzania', 'kk': 'Танзания', 'kk-Cyrl': 'Танзания', 'kln': 'Emetab Tanzania', 'km': 'តង់សានី', 'kn': 'ತಾಂಜೇನಿಯಾ', 'ko': '탄자니아', 'kok': 'तांझानिया', 'kok-Deva': 'तांझानिया', 'ks': 'تَنجانِیا', 'ks-Arab': 'تَنجانِیا', 'ksb': 'Tanzania', 'ksf': 'tanzaní', 'ksh': 'Tansanija', 'ku': 'Tanzanya', 'kxv': 'ṭanjaniya', 'kxv-Deva': 'तंज़ानिया', 'kxv-Latn': 'ṭanjaniya', 'kxv-Orya': 'ତାଞ୍ଜାନିୟା', 'kxv-Telu': 'టాంజానియా', 'ky': 'Танзания', 'lag': 'Taansanía', 'lb': 'Tansania', 'lg': 'Tanzaniya', 'lld': 'Tanzania', 'ln': 'Tanzani', 'lo': 'ທານຊາເນຍ', 'lt': 'Tanzanija', 'lu': 'Tanzani', 'luo': 'Tanzania', 'luy': 'Tanzania', 'lv': 'Tanzānija', 'mai': 'तंजानिया', 'mas': 'Tansania', 'mer': 'Tanzania', 'mfe': 'Tanzani', 'mg': 'Tanzania', 'mgh': 'Utanzania', 'mi': 'Tānahia', 'mk': 'Танзанија', 'ml': 'ടാൻസാനിയ', 'mn': 'Танзани', 'mr': 'टांझानिया', 'ms': 'Tanzania', 'mt': 'it-Tanzanija', 'mua': 'Tanzaniya', 'my': 'တန်ဇန်းနီးယား', 'mzn': 'تانزانیا', 'naq': 'Tanzaniab', 'nb': 'Tanzania', 'nd': 'Tanzaniya', 'ne': 'तान्जानिया', 'nl': 'Tanzania', 'nmg': 'Tanzánía', 'nn': 'Tanzania', 'no': 'Tanzania', 'nqo': 'ߕߊ߲ߖ߭ߊߣߌ߫', 'nyn': 'Tanzania', 'om': 'Taanzaaniyaa', 'or': 'ତାଞ୍ଜାନିଆ', 'pa': 'ਤਨਜ਼ਾਨੀਆ', 'pa-Guru': 'ਤਨਜ਼ਾਨੀਆ', 'pcm': 'Tanzánia', 'pl': 'Tanzania', 'ps': 'تنزانیا', 'pt': 'Tanzânia', 'qu': 'Tanzania', 'rm': 'Tansania', 'rn': 'Tanzaniya', 'ro': 'Tanzania', 'rof': 'Tanzania', 'ru': 'Танзания', 'rwk': 'Tanzania', 'saq': 'Tanzania', 'sbp': 'Tansaniya', 'sc': 'Tanzània', 'sd': 'تنزانيا', 'sd-Arab': 'تنزانيا', 'se': 'Tanzánia', 'ses': 'Tanzaani', 'sg': 'Tanzanïi', 'shi': 'ⵟⴰⵏⵥⴰⵏⵢⴰ', 'shi-Latn': 'ṭanẓanya', 'shi-Tfng': 'ⵟⴰⵏⵥⴰⵏⵢⴰ', 'si': 'ටැන්සානියාව', 'sk': 'Tanzánia', 'sl': 'Tanzanija', 'smn': 'Tansania', 'sn': 'Tanzania', 'so': 'Tansaaniya', 'sq': 'Tanzani', 'sr': 'Танзанија', 'sr-Cyrl': 'Танзанија', 'sr-Latn': 'Tanzanija', 'sv': 'Tanzania', 'sw': 'Tanzania', 'syr': 'ܛܢܙܢܝܐ', 'ta': 'தான்சானியா', 'te': 'టాంజానియా', 'teo': 'Tanzania', 'tg': 'Танзания', 'th': 'แทนซาเนีย', 'ti': 'ታንዛንያ', 'tk': 'Tanzaniýa', 'to': 'Tenisānia', 'tr': 'Tanzanya', 'tt': 'Танзания', 'twq': 'Tanzaani', 'tzm': 'Ṭanzanya', 'ug': 'تانزانىيە', 'uk': 'Танзанія', 'ur': 'تنزانیہ', 'uz': 'Tanzaniya', 'uz-Cyrl': 'Танзания', 'uz-Latn': 'Tanzaniya', 'vai': 'ꕚꘋꕤꕇꕰ', 'vai-Latn': 'Taŋzaninya', 'vai-Vaii': 'ꕚꘋꕤꕇꕰ', 'vec': 'Tanzania', 'vi': 'Tanzania', 'vun': 'Tanzania', 'wae': 'Tansania', 'wo': 'Taŋsani', 'xh': 'ETanzania', 'xnr': 'तंज़ानिया', 'yav': 'taŋsaní', 'yi': 'טאַנזאַניע', 'yo': 'Tàǹsáníà', 'yrl': 'Tãsaniya', 'yue': '坦尚尼亞', 'yue-Hans': '坦尚尼亚', 'yue-Hant': '坦尚尼亞', 'zgh': 'ⵟⴰⵏⵥⴰⵏⵢⴰ', 'zh': '坦桑尼亚', 'zh-Hans': '坦桑尼亚', 'zh-Hant': '坦尚尼亞', 'zu': 'i-Tanzania'},
    'Tagb': {'am': 'ትአግባንዋ', 'ar': 'التاجبانوا', 'ast': 'tagbanwa', 'az': 'taqbanva', 'az-Latn': 'taqbanva', 'bal-Latn': 'Tagbanwá', 'bg': 'Тагбанва', 'bn': 'টাগোওয়ানা', 'brx': 'तागबानवा', 'bs': 'tagbanva pismo', 'bs-Cyrl': 'тагбанва писмо', 'bs-Latn': 'tagbanva pismo', 'ca': 'tagbanwa', 'ca-ES-valencia': 'tagbanwa', 'ccp': '𑄑𑄉𑄮𑄤𑄚', 'cs': 'tagbanwa', 'da': 'tagbanwa', 'de': 'Tagbanwa', 'el': 'Ταγκμάνγουα', 'el-polyton': 'Ταγκμάνγουα', 'en': 'Tagbanwa', 'en-Dsrt': '𐐓𐐲𐑀𐐺𐐪𐑌𐐶𐐪', 'es': 'tagbanúa', 'et': 'tagbanva', 'eu': 'tagbanwa', 'fa': 'تگبنوایی', 'ff-Adlm': '𞤚𞤢𞤺𞤦𞤢𞤲𞤱𞤢', 'fi': 'tagbanwalainen', 'fo': 'tagbanwa', 'fr': 'tagbanoua', 'fy': 'Tagbanwa', 'ga': 'Tagbanwa', 'gd': 'Tagbanwa', 'gsw': 'Tagbanwa', 'gu': 'તગબન્વા', 'he': 'טגבנווה', 'hi': 'तगबन्वा', 'hi-Latn': 'Tagbanwa', 'hr': 'tagbanwa pismo', 'ht': 'tagbanoua', 'hu': 'Tagbanwa', 'id': 'Tagbanwa', 'ig': 'Tagbanwa', 'it': 'tagbanwa', 'ja': 'タグバンワ文字', 'ka': 'ტაგბანვა', 'kgp': 'tagmánva', 'kk': 'тагбанва жазуы', 'kk-Cyrl': 'тагбанва жазуы', 'kn': 'ಟಾಗ್ಬಾನವಾ', 'ko': '타그반와 문자', 'ks': 'تَگبَنوا', 'ks-Arab': 'تَگبَنوا', 'lo': 'ຕັກບັນວາ', 'lt': 'tagbanva', 'mk': 'тагбанванско', 'ml': 'തഗ്ബൻവാ', 'mn': 'Тагбанва', 'mr': 'तगोआन्वा', 'ms': 'Tagbanwa', 'nb': 'tagbanwa', 'ne': 'टाग्वान्वा', 'nl': 'Tagbanwa', 'nn': 'tagbanwa', 'no': 'tagbanwa', 'nqo': 'ߕߜߊߣߎߥߊ', 'or': 'ତଗବାନ୍ୱା', 'pl': 'tagbanwa', 'pt': 'tagbanwa', 'rm': 'tagbanwa', 'ro': 'tagbanwa', 'ru': 'тагбанва', 'sc': 'tagbanwa', 'sl': 'tagbanski', 'so': 'Tagbanwa', 'sq': 'tagbanvaisht', 'sr': 'тагбанва писмо', 'sr-Cyrl': 'тагбанва писмо', 'sr-Latn': 'tagbanva pismo', 'sv': 'tagbanwa', 'ta': 'தகோவானா', 'te': 'టాగ్బానవా', 'tg': 'Тагбанва', 'th': 'ตักบันวา', 'to': 'tohinima fakatakipaneuā', 'tr': 'Tagbanva', 'ug': 'تاگبانۋاچە', 'uk': 'тагбанва', 'vec': 'tagbanwa', 'vi': 'Chữ Tagbanwa', 'xnr': 'तगबन्वा', 'yrl': 'tagibanua', 'yue': '南島文', 'yue-Hans': '南岛文', 'yue-Hant': '南島文', 'zh': '塔格班瓦文', 'zh-Hans': '塔格班瓦文', 'zh-Hant': '南島文', 'zu': 'i-Tagbanwa'},
    'Takr': {'ast': 'takri', 'bal-Latn': 'Takri', 'bs': 'takri pismo', 'bs-Cyrl': 'такри писмо', 'bs-Latn': 'takri pismo', 'ca': 'takri', 'ca-ES-valencia': 'takri', 'cs': 'takrí', 'da': 'takri', 'de': 'Takri', 'en': 'Takri', 'et': 'taakri', 'eu': 'takriera', 'ff-Adlm': '𞤚𞤢𞤳𞤪𞤭', 'fi': 'takri', 'fo': 'takri', 'fy': 'Takri', 'ga': 'Takri', 'gd': 'Takri', 'he': 'טאקרי', 'hi-Latn': 'Takri', 'hr': 'takri pismo', 'id': 'Takri', 'ig': 'Takri', 'it': 'takri', 'ja': 'タークリー文字', 'ka': 'ტაკრი', 'kk': 'такри жазуы', 'kk-Cyrl': 'такри жазуы', 'ko': '타크리 문자', 'lo': 'ທາຄຮີ', 'lt': 'Takri', 'mk': 'такри', 'mn': 'Такри', 'ms': 'Takri', 'nb': 'takri', 'nl': 'Takri', 'nn': 'takri', 'no': 'takri', 'nqo': 'ߕߊߞߙߌ', 'ro': 'takri', 'ru': 'такри', 'sc': 'takri', 'so': 'Takri', 'sq': 'takri', 'sv': 'takritiska', 'tg': 'Такрӣ', 'th': 'ทาครี', 'to': 'tohinima fakatakili', 'tr': 'Takri', 'ug': 'تاكرىچە', 'vec': 'takri', 'vi': 'Chữ Takri', 'yue': '塔卡里文字', 'yue-Hans': '塔卡里文字', 'yue-Hant': '塔卡里文字', 'zh': '泰克里文', 'zh-Hans': '泰克里文', 'zh-Hant': '塔卡里文字', 'zu': 'i-Takri'},
    'Tale': {'am': 'ታኢ ለ', 'ar': 'التاي لي', 'ast': 'tai le', 'az': 'tay le', 'az-Latn': 'tay le', 'bal-Latn': 'Tái Lé', 'bg': 'Тай Ле', 'bn': 'তাইলে', 'brx': 'ताई ले', 'bs': 'tai le pismo', 'bs-Cyrl': 'таи ле писмо', 'bs-Latn': 'tai le pismo', 'ca': 'tai le', 'ca-ES-valencia': 'tai le', 'ccp': '𑄖𑄭𑄣𑄬', 'cs': 'tai le', 'da': 'tai le', 'de': 'Tai Le', 'el': 'Τάι Λε', 'el-polyton': 'Τάι Λε', 'en': 'Tai Le', 'en-Dsrt': '𐐓𐐴 𐐢𐐯', 'es': 'tai le', 'et': 'tai-löö', 'eu': 'tai le', 'ff-Adlm': '𞤚𞤢𞥄𞤴 𞤂𞤭𞥅', 'fi': 'tailelainen', 'fo': 'tai le', 'fr': 'taï-le', 'fy': 'Tai Le', 'ga': 'Deiheoingis', 'gd': 'Tai Le', 'gsw': 'Tai Le', 'gu': 'તાઇ લી', 'he': 'טאי לה', 'hi': 'ताई ली', 'hi-Latn': 'Tai Le', 'hr': 'tai le pismo', 'ht': 'taï-le', 'hu': 'Tai Le', 'id': 'Tai Le', 'ig': 'Tai Le', 'it': 'tai le', 'ja': 'タイ・レ文字', 'ka': 'ტაი ლე', 'kgp': 'taj-re', 'kk': 'лы жазуы', 'kk-Cyrl': 'лы жазуы', 'km': 'តៃឡេ', 'kn': 'ಥಾಯ್ ಲಿ', 'ko': '타이 레 문자', 'ks': 'تَیلیے', 'ks-Arab': 'تَیلیے', 'lb': 'Tai Le', 'lo': 'ໄທເລ', 'lt': 'tai le', 'mk': 'тај ле', 'ml': 'തായ് ലേ', 'mn': 'Тале', 'mr': 'ताई ली', 'ms': 'Tai Le', 'my': 'တိုင်လီ', 'nb': 'tai le', 'ne': 'टाइले', 'nl': 'Tai Le', 'nn': 'tai le', 'no': 'tai le', 'nqo': 'ߕߊߦߌ-ߟߋ߫', 'or': 'ତାଇ ଲେ', 'pl': 'tai le', 'pt': 'tai Le', 'rm': 'tai le', 'ro': 'tai le', 'ru': 'тайский ле', 'sc': 'tai le', 'so': 'Tai Le', 'sq': 'tai le', 'sr': 'таи ле писмо', 'sr-Cyrl': 'таи ле писмо', 'sr-Latn': 'tai le pismo', 'sv': 'tai le', 'ta': 'தாய் லே', 'te': 'తై లీ', 'tg': 'Тай Ле', 'th': 'ไทเล', 'to': 'tohinima fakatai-lue', 'tr': 'Tai Le', 'ug': 'تاي-لەچە', 'uk': 'тай-лі', 'vec': 'tai le', 'vi': 'Chữ Thái Na', 'xnr': 'ताई ली', 'yrl': 'tai re', 'yue': '傣哪文', 'yue-Hans': '傣哪文', 'yue-Hant': '傣哪文', 'zh': '泰乐文', 'zh-Hans': '泰乐文', 'zh-Hant': '傣哪文', 'zu': 'isi-Tai Le'},
    'Talu': {'am': 'አዲስ ታኢ ሉ', 'ar': 'التاى لى الجديد', 'ast': 'tai lue nuevu', 'az': 'təzə tay lu', 'az-Latn': 'təzə tay lu', 'bal-Latn': 'Nókén Tái Lé', 'bg': 'Нова Тай Ле', 'bn': 'নতুন তাই লু', 'brx': 'नया ताई लुए', 'bs': 'novo tai lue pismo', 'bs-Cyrl': 'нови таи луе', 'bs-Latn': 'novo tai lue pismo', 'ca': 'nou tai lue', 'ca-ES-valencia': 'nou tai lue', 'ccp': '𑄚𑄱 𑄖𑄭 𑄣𑄪', 'cs': 'tai lü nové', 'da': 'tai lue', 'de': 'Tai Lue', 'el': 'Νέο Τάι Λούε', 'el-polyton': 'Νέο Τάι Λούε', 'en': 'New Tai Lue', 'en-Dsrt': '𐐤𐐭 𐐓𐐴 𐐢𐐭𐐯', 'es': 'nuevo tai lue', 'et': 'uus tai-lõõ', 'eu': 'tai lue berria', 'ff-Adlm': '𞤚𞤢𞥄𞤴 𞤂𞤵𞤫 𞤑𞤫𞤧𞤮', 'fi': 'uusi tailuelainen', 'fo': 'nýtt tai lue', 'fr': 'nouveau taï-lue', 'fy': 'Nij Tai Lue', 'ga': 'Tai Lue Nua', 'gd': 'Tai Lue Ùr', 'gsw': 'Tai Lue', 'gu': 'નવીન તાઇ લૂ', 'he': 'טאי ליו חדש', 'hi': 'नया ताई लु', 'hi-Latn': 'Naya Tai Lue', 'hr': 'novo tai lue pismo', 'ht': 'nouveau taï-lue', 'hu': 'Új tai lue', 'id': 'Tai Lue Baru', 'ig': 'New Tai Lue', 'it': 'tai lue', 'ja': '新タイ・ルー文字', 'ka': 'ახალი ტაი ლიუ', 'kgp': 'taj-re tãg', 'kk': 'жаңа лы жазуы', 'kk-Cyrl': 'жаңа лы жазуы', 'km': 'តៃឡឺថ្មី', 'kn': 'ನ್ಯೂ ಥಾಯ್ ಲುಇ', 'ko': '신 타이 루에', 'ks': 'نۆو تیلو', 'ks-Arab': 'نۆو تیلو', 'lb': 'Tai Lue', 'lo': 'ໄທລື້ໃໝ່', 'lt': 'naujasis Tailando lue', 'mk': 'ново тај луе', 'ml': 'പുതിയ തായ് ല്യൂ', 'mn': 'Талу', 'mr': 'नवीन ताई लू', 'ms': 'Tai Lue Baharu', 'nb': 'ny tai lue', 'ne': 'न्यू टाइ लुइ', 'nl': 'Nieuw Tai Lue', 'nn': 'ny tai lue', 'no': 'ny tai lue', 'nqo': 'ߕߊߦߌ-ߟߎ߫ ߞߎߘߊ߫', 'or': 'ନୂତନ ତାଇ ଲୁଏ', 'pl': 'nowy tai lue', 'pt': 'novo tai lue', 'rm': 'tai lue', 'ro': 'tai le nouă', 'ru': 'новый тайский ле', 'sc': 'tai lue nou', 'so': 'Tai Lue cusub', 'sq': 'tai lue i ri', 'sr': 'нови таи луе', 'sr-Cyrl': 'нови таи луе', 'sr-Latn': 'novi tai lue', 'sv': 'tai lue', 'ta': 'புதிய தை லூ', 'te': 'క్రొత్త టై లుఇ', 'tg': 'Тай Леи Нав', 'th': 'ไทลื้อใหม่', 'to': 'tohinima fakatai-lue-foʻou', 'tr': 'New Tai Lue', 'ug': 'يېڭى تاي-لەچە', 'uk': 'новий тайський луе', 'vec': 'tai lu senpio', 'vi': 'Chữ Thái Lặc mới', 'xnr': 'नोआ ताई लु', 'yrl': 'tai rue pisasú', 'yue': '西雙版納新傣文', 'yue-Hans': '西双版纳新傣文', 'yue-Hant': '西雙版納新傣文', 'zh': '新傣文', 'zh-Hans': '新傣文', 'zh-Hant': '西雙版納新傣文', 'zu': 'isi-New Tai Lue'},
    'Taml': {'af': 'Tamil', 'ak': 'Tamil kasa', 'am': 'ታሚል', 'ar': 'التاميلية', 'as': 'তামিল', 'ast': 'tamil', 'az': 'tamil', 'az-Arab': 'تامیلجه', 'az-Latn': 'tamil', 'bal-Latn': 'Támel', 'be': 'тамільскае', 'be-tarask': 'тамільскае', 'bg': 'тамилска', 'bgn': 'تامیلی', 'bn': 'তামিল', 'br': 'tamilek', 'brx': 'तामिल', 'bs': 'tamilsko pismo', 'bs-Cyrl': 'тамилско писмо', 'bs-Latn': 'tamilsko pismo', 'ca': 'tàmil', 'ca-ES-valencia': 'tàmil', 'ccp': '𑄖𑄟𑄨𑄣𑄴', 'ce': 'тамилхойн', 'chr': 'ᏔᎻᎵ', 'ckb': 'تامیلی', 'cs': 'tamilské', 'cy': 'Tamilaidd', 'da': 'tamilsk', 'de': 'Tamilisch', 'dsb': 'tamilski', 'dz': 'ཏ་མིལ་ཡིག་གུ', 'ee': 'tamilgbeŋɔŋlɔ', 'el': 'Ταμίλ', 'el-polyton': 'Ταμίλ', 'en': 'Tamil', 'en-Dsrt': '𐐓𐐰𐑋𐐮𐑊', 'es': 'tamil', 'et': 'tamili', 'eu': 'tamildarra', 'fa': 'تامیلی', 'ff-Adlm': '𞤚𞤢𞤥𞤭𞤤', 'fi': 'tamililainen', 'fil': 'Tamil', 'fo': 'tamilskt', 'fr': 'tamoul', 'fur': 'tamil', 'fy': 'Tamil', 'ga': 'Tamalach', 'gd': 'Taimil', 'gl': 'támil', 'gsw': 'Tamilisch', 'gu': 'તમિલ', 'ha': 'Yaren Tamil', 'he': 'טמיל', 'hi': 'तमिल', 'hi-Latn': 'Tamil', 'hr': 'tamilsko pismo', 'hsb': 'tamilsce', 'ht': 'tamoul', 'hu': 'Tamil', 'hy': 'թամիլական', 'ia': 'tamil', 'id': 'Tamil', 'ig': 'Tamil', 'is': 'tamílskt', 'it': 'tamil', 'ja': 'タミル文字', 'jv': 'Tamil', 'ka': 'ტამილური', 'kea': 'tamil', 'kgp': 'tỹmĩr', 'kk': 'тамиль жазуы', 'kk-Cyrl': 'тамиль жазуы', 'km': 'តាមីល', 'kn': 'ತಮಿಳು', 'ko': '타밀 문자', 'kok': 'तमीळ', 'kok-Deva': 'तमीळ', 'ks': 'تَمِل', 'ks-Arab': 'تَمِل', 'ksh': 'tamiilesche Schreff', 'ku': 'tamîlî', 'kxv': 'tamiḷ', 'kxv-Deva': 'तमिल', 'kxv-Latn': 'tamiḷ', 'kxv-Orya': 'ତାମିଲ୍', 'kxv-Telu': 'తమిళము', 'ky': 'Тамил', 'lb': 'Tamilesch', 'lo': 'ທາມິລ', 'lrc': 'تامیل', 'lt': 'tamilų', 'lv': 'tamilu', 'mai': 'तमिल', 'mi': 'Tamiera', 'mk': 'тамилско писмо', 'ml': 'തമിഴ്', 'mn': 'тамил', 'mr': 'तामिळ', 'ms': 'Tamil', 'my': 'တမီးလ်', 'nb': 'tamilsk', 'ne': 'तामिल', 'nl': 'Tamil', 'nn': 'tamilsk', 'no': 'tamilsk', 'nqo': 'ߕߊߡߎߟ', 'or': 'ତାମିଲ୍', 'pa': 'ਤਮਿਲ', 'pa-Guru': 'ਤਮਿਲ', 'pcm': 'Támil', 'pl': 'tamilskie', 'ps': 'تامل', 'pt': 'tâmil', 'qu': 'Tamil Simi', 'rm': 'tamil', 'ro': 'tamilă', 'ru': 'тамильская', 'sc': 'tamil', 'sd': 'تامل', 'sd-Arab': 'تامل', 'si': 'දෙමළ', 'sk': 'tamilské', 'sl': 'tamilski', 'so': 'Taamiil', 'sq': 'tamil', 'sr': 'тамилско писмо', 'sr-Cyrl': 'тамилско писмо', 'sr-Latn': 'tamilsko pismo', 'sv': 'tamilska', 'sw': 'Kitamil', 'ta': 'தமிழ்', 'te': 'తమిళము', 'tg': 'Тамилӣ', 'th': 'ทมิฬ', 'ti': 'ታሚል', 'tk': 'Tamil elipbiýi', 'to': 'tohinima fakatamili', 'tr': 'Tamil', 'ug': 'تامىلچە', 'uk': 'тамільська', 'ur': 'تمل', 'uz': 'tamil', 'uz-Cyrl': 'Тамил', 'uz-Latn': 'tamil', 'vec': 'tamil', 'vi': 'Chữ Tamil', 'wae': 'Tamiliš', 'xnr': 'तमिल', 'yo': 'Támílì', 'yrl': 'tãmiu', 'yue': '坦米爾文', 'yue-Hans': '坦米尔文', 'yue-Hant': '坦米爾文', 'zh': '泰米尔文', 'zh-Hans': '泰米尔文', 'zh-Hant': '坦米爾文', 'zu': 'isi-Tamil'},
    'Tang': {'ast': 'tangut', 'bal-Latn': 'Tangut', 'bs': 'tangut pismo', 'bs-Cyrl': 'тангут писмо', 'bs-Latn': 'tangut pismo', 'ca': 'tangut', 'ca-ES-valencia': 'tangut', 'cs': 'tangut', 'da': 'tangut', 'de': 'Xixia', 'en': 'Tangut', 'et': 'tanguudi', 'eu': 'tangutera', 'ff-Adlm': '𞤚𞤢𞤲𞤺𞤵𞤼', 'fi': 'tangut', 'fo': 'tangut', 'fy': 'Tangut', 'ga': 'Tangut', 'gd': 'Tangut', 'he': 'טאנגוט', 'hi-Latn': 'Tangut', 'hr': 'tangut pismo', 'id': 'Tangut', 'ig': 'Tangut', 'it': 'tangut', 'ja': '西夏文字', 'ka': 'ტანღუტური', 'kk': 'танғұт жазуы', 'kk-Cyrl': 'танғұт жазуы', 'ko': '탕구트 문자', 'lo': 'ຕັນກັນ', 'lt': 'Tangut', 'mk': 'тангутско', 'mn': 'Тангут', 'ms': 'Tangut', 'nb': 'tangut', 'nl': 'Tangut', 'nn': 'tangut', 'no': 'tangut', 'nqo': 'ߕߊ߲ߜ߭ߎߕ', 'ro': 'tangut', 'ru': 'тангутское меня', 'sc': 'tangut', 'so': 'Luuqada Tangut', 'sq': 'tangut', 'sv': 'tangutiska', 'tg': 'Тангут', 'th': 'ตันกัท', 'to': 'tohinima fakatanguti', 'tr': 'Tangut', 'ug': 'تاڭغۇتچە', 'uk': 'тангут', 'vec': 'tangut', 'vi': 'Chữ Tangut', 'yue': '西夏文', 'yue-Hans': '西夏文', 'yue-Hant': '西夏文', 'zh': '唐古特文', 'zh-Hans': '唐古特文', 'zh-Hant': '西夏文', 'zu': 'i-Tangut'},
    'Tavt': {'ast': 'tai viet', 'bal-Latn': 'Tái Wiet', 'bn': 'তাই ভিয়েৎ', 'bs': 'tai viet pismo', 'bs-Cyrl': 'таи виет писмо', 'bs-Latn': 'tai viet pismo', 'ca': 'tai viet', 'ca-ES-valencia': 'tai viet', 'ccp': '𑄖𑄭 𑄞𑄨𑄠𑄬𑄖𑄴', 'cs': 'tai viet', 'de': 'Tai-Viet', 'el': 'Τάι Βιέτ', 'el-polyton': 'Τάι Βιέτ', 'en': 'Tai Viet', 'en-Dsrt': '𐐓𐐴 𐐚𐐨𐐯𐐻', 'et': 'tai-vieti', 'eu': 'tai viet', 'ff-Adlm': '𞤚𞤢𞥄𞤴 𞤜𞤭𞤫𞥅𞤼', 'fi': 'tai viet', 'fo': 'tai viet', 'fr': 'taï viêt', 'fy': 'Tai Viet', 'ga': 'Tai Viet', 'gd': 'Tai Viet', 'gu': 'તાઇ વેઇત', 'he': 'טאי וייט', 'hi': 'ताई विएत', 'hi-Latn': 'Tai Viet', 'hr': 'tai viet pismo', 'ht': 'taï viêt', 'hu': 'Tai viet', 'id': 'Tai Viet', 'ig': 'Tai Viet', 'it': 'tai viet', 'ja': 'タイ・ヴェト文字', 'ka': 'ტაი-ვიეტი', 'kgp': 'tavuti', 'kk': 'тай вьет жазуы', 'kk-Cyrl': 'тай вьет жазуы', 'km': 'តៃវៀត', 'kn': 'ಥಾಯ್ ವಿಯೆಟ್', 'ko': '태국 베트남 문자', 'lo': 'ໄທຫວຽດ', 'lt': 'tai vet', 'mk': 'тај вјет', 'ml': 'ത്വട്', 'mn': 'Тай Вьетнам', 'mr': 'ताई विएत', 'ms': 'Tai Viet', 'nb': 'tai viet', 'ne': 'टाभ्ट', 'nl': 'Tai Viet', 'nn': 'tai viet', 'no': 'tai viet', 'nqo': 'ߕߊߦߌ-ߝ߭ߌߋ߫', 'or': 'ତାଇ ଭିଏତ୍', 'pl': 'tai viet', 'rm': 'tai viet', 'ro': 'tai viet', 'ru': 'тай-вьет', 'sc': 'tai viet', 'sl': 'tajsko-vietnamski', 'so': 'Farta lagu Qoro Luuqadaha Tai', 'sq': 'tai viet', 'sr': 'таи виет писмо', 'sr-Cyrl': 'таи виет писмо', 'sr-Latn': 'tai viet pismo', 'sv': 'tai viet', 'ta': 'தை வியத்', 'te': 'టై వియట్', 'tg': 'Ветнамии Тайӣ', 'th': 'ไทเวียต', 'to': 'tohinima fakatai-vieti', 'tr': 'Tai Viet', 'ug': 'تايلاندچە-ۋىيېتنامچە', 'uk': 'тай-вʼєт', 'vec': 'tai viet', 'vi': 'Chữ Thái Việt', 'xnr': 'ताई विएत', 'yrl': 'tawiti', 'yue': '傣擔文', 'yue-Hans': '傣担文', 'yue-Hant': '傣擔文', 'zh': '越南傣文', 'zh-Hans': '越南傣文', 'zh-Hant': '傣擔文', 'zu': 'isi-Tai Viet'},
    'Telu': {'af': 'Teloegoe', 'ak': 'Telugu kasa', 'am': 'ተሉጉ', 'ar': 'التيلجو', 'as': 'তেলুগু', 'ast': 'telugu', 'az': 'teluqu', 'az-Latn': 'teluqu', 'bal-Latn': 'Telegó', 'be': 'тэлугу', 'be-tarask': 'тэлугу', 'bg': 'телугу', 'bn': 'তেলেগু', 'br': 'telougou', 'brx': 'तेलुगु', 'bs': 'pismo telugu', 'bs-Cyrl': 'телугу писмо', 'bs-Latn': 'pismo telugu', 'ca': 'telugu', 'ca-ES-valencia': 'telugu', 'ccp': '𑄖𑄬𑄣𑄬𑄉𑄪', 'ce': 'телугу', 'chr': 'ᏖᎷᎦ', 'ckb': 'تێلووگوو', 'cs': 'telugské', 'cy': 'Telugu', 'da': 'telugu', 'de': 'Telugu', 'dsb': 'telugu', 'dz': 'ཏེ་ལུ་གུ་ཡིག་གུ', 'ee': 'telegugbeŋɔŋlɔ', 'el': 'Τελούγκου', 'el-polyton': 'Τελούγκου', 'en': 'Telugu', 'en-Dsrt': '𐐓𐐯𐑊𐐭𐑀𐐭', 'es': 'telugu', 'et': 'telugu', 'eu': 'teluguarra', 'fa': 'تلوگویی', 'ff-Adlm': '𞤚𞤫𞤤𞤵𞤺𞤵', 'fi': 'telugulainen', 'fil': 'Telugu', 'fo': 'telugu', 'fr': 'télougou', 'fur': 'telegu', 'fy': 'Telugu', 'ga': 'Teileagúch', 'gd': 'Telugu', 'gl': 'telugu', 'gsw': 'Telugu', 'gu': 'તેલુગુ', 'ha': 'Yaren Telugu', 'he': 'טלוגו', 'hi': 'तेलुगू', 'hi-Latn': 'Telugu', 'hr': 'teluško pismo', 'hsb': 'telugu', 'ht': 'télougou', 'hu': 'Telugu', 'hy': 'թելուգու', 'ia': 'telugu', 'id': 'Telugu', 'ig': 'Telugu', 'is': 'telúgú', 'it': 'telugu', 'ja': 'テルグ文字', 'jv': 'Telugu', 'ka': 'ტელუგუ', 'kea': 'telugu', 'kgp': 'térugu', 'kk': 'телугу жазуы', 'kk-Cyrl': 'телугу жазуы', 'km': 'តេលុគុ', 'kn': 'ತೆಲುಗು', 'ko': '텔루구 문자', 'kok': 'तेलुगू', 'kok-Deva': 'तेलुगू', 'ks': 'تیلگوٗ', 'ks-Arab': 'تیلگوٗ', 'ksh': 'indesche Telugu-Schreff', 'ku': 'teluguyî', 'kxv': 'telugu', 'kxv-Deva': 'तेलुगू', 'kxv-Latn': 'telugu', 'kxv-Orya': 'ତେଲୁଗୁ', 'kxv-Telu': 'తెలుగు', 'ky': 'Телу', 'lb': 'Telugu', 'lo': 'ເທລູກູ', 'lrc': 'تئلئگو', 'lt': 'telugų', 'lv': 'telugu', 'mai': 'तेलुगू', 'mi': 'Teruku', 'mk': 'телугу', 'ml': 'തെലുങ്ക്', 'mn': 'тэлүгү', 'mr': 'तेलगु', 'ms': 'Telugu', 'my': 'တီလု', 'nb': 'telugu', 'ne': 'तेलुगु', 'nl': 'Telugu', 'nn': 'telugu', 'no': 'telugu', 'nqo': 'ߕߟߋߜ߭ߎ', 'or': 'ତେଲୁଗୁ', 'pa': 'ਤੇਲਗੂ', 'pa-Guru': 'ਤੇਲਗੂ', 'pcm': 'Tẹ́lúgu', 'pl': 'telugu', 'ps': 'تیلیګو', 'pt': 'télugo', 'qu': 'Tegulu Simi', 'rm': 'telugu', 'ro': 'telugu', 'ru': 'телугу', 'sc': 'telugu', 'sd': 'تلگو', 'sd-Arab': 'تلگو', 'si': 'තෙළිඟු', 'sk': 'telugské', 'sl': 'teluški', 'so': 'Teeluguu', 'sq': 'telug', 'sr': 'телугу писмо', 'sr-Cyrl': 'телугу писмо', 'sr-Latn': 'telugu pismo', 'sv': 'telugu', 'sw': 'Kitelugu', 'ta': 'தெலுங்கு', 'te': 'తెలుగు', 'tg': 'Телугу', 'th': 'เตลูกู', 'ti': 'ቴሉጉ', 'tk': 'Telugu elipbiýi', 'to': 'tohinima fakaʻinitia-teluku', 'tr': 'Telugu', 'ug': 'تېلۇگۇچە', 'uk': 'телугу', 'ur': 'تیلگو', 'uz': 'telugu', 'uz-Cyrl': 'Телугу', 'uz-Latn': 'telugu', 'vec': 'telugo', 'vi': 'Chữ Telugu', 'wae': 'Telugu', 'xnr': 'तेलुगू', 'yo': 'Télúgù', 'yrl': 'térugu', 'yue': '泰盧固文', 'yue-Hans': '泰卢固文', 'yue-Hant': '泰盧固文', 'zh': '泰卢固文', 'zh-Hans': '泰卢固文', 'zh-Hant': '泰盧固文', 'zu': 'isi-Telugu'},
    'Teng': {'ar': 'التينجوار', 'ast': 'tengwar', 'az': 'tengvar', 'az-Latn': 'tengvar', 'bn': 'তেঙ্গোয়ার', 'brx': 'तेंगवार', 'bs': 'tengvar pismo', 'bs-Cyrl': 'тенгвар писмо', 'bs-Latn': 'tengvar pismo', 'ca': 'tengwar', 'ca-ES-valencia': 'tengwar', 'ccp': '𑄖𑄬𑄋𑄴𑄉𑄮𑄠𑄢𑄴', 'cs': 'tengwar', 'da': 'tengwar', 'de': 'Tengwar', 'el': 'Τεγνγουάρ', 'el-polyton': 'Τεγνγουάρ', 'en': 'Tengwar', 'en-Dsrt': '𐐓𐐯𐑍𐐶𐐪𐑉', 'es': 'tengwar', 'et': 'Tengwari', 'fa': 'تنگوار', 'fi': 'tengwar', 'fr': 'tengwar', 'fy': 'Tengwar', 'gd': 'Tengwar', 'gsw': 'Tengwar', 'gu': 'તેન્ગવાર', 'hi': 'तेन्गवार', 'hi-Latn': 'Tengwar', 'hr': 'tengwar pismo', 'ht': 'tengwar', 'hu': 'Tengwar', 'id': 'Tenghwar', 'it': 'tengwar', 'ja': 'テングワール文字', 'ka': 'ტენგვარი', 'kgp': 'tẽgvar', 'kn': 'ತೆಂಗ್\u200cವಾರ್', 'ko': '텡과르 문자', 'ks': 'تینگوار', 'ks-Arab': 'تینگوار', 'lb': 'Tengwar', 'lo': 'ເທງກວາຮ', 'lt': 'tengvar', 'mk': 'тенгвар', 'ml': 'തെംഗ്വർ', 'mr': 'तेन्गवार', 'nb': 'tengwar', 'ne': 'टेङ्वार', 'nl': 'Tengwar', 'nn': 'tengwar', 'no': 'tengwar', 'or': 'ତେଙ୍ଗୱାର୍', 'pl': 'tengwar', 'pt': 'tengwar', 'rm': 'tengwar', 'ru': 'тенгварская', 'sl': 'tengvarski', 'sr': 'тенгвар писмо', 'sr-Cyrl': 'тенгвар писмо', 'sr-Latn': 'tengvar pismo', 'sv': 'tengwar', 'ta': 'தெங்வார்', 'te': 'టేంగ్వార్', 'th': 'เทงกวาร์', 'to': 'tohinima fakatengiuali', 'tr': 'Tengvar', 'ug': 'تېڭۋارچە', 'uk': 'тенгвар', 'vi': 'Chữ Tengwar', 'yrl': 'tẽguwari', 'yue': '談格瓦文', 'yue-Hans': '谈格瓦文', 'yue-Hant': '談格瓦文', 'zh': '腾格瓦文字', 'zh-Hans': '腾格瓦文字', 'zh-Hant': '談格瓦文'},
    'Tfng': {'af': 'Tifinagh', 'ak': 'Tifinafo kasa', 'am': 'ቲፊናግህ', 'ar': 'التيفيناغ', 'as': 'টিফিনাঘ', 'ast': 'tifinagh', 'az': 'tifinaq', 'az-Latn': 'tifinaq', 'bal-Latn': 'Tipinag', 'be': 'тыфінаг', 'be-tarask': 'тыфінаг', 'bg': 'тифинаг', 'bn': 'তিফিনাগ', 'brx': 'तिफिनाघ', 'bs': 'tifinag pismo', 'bs-Cyrl': 'тифинаг писмо', 'bs-Latn': 'tifinag pismo', 'ca': 'tifinag', 'ca-ES-valencia': 'tifinag', 'ccp': '𑄖𑄨𑄜𑄨𑄚𑄉𑄴', 'chr': 'ᏘᏫᎾᎦ', 'cs': 'berberské', 'cy': 'Tifinagh', 'da': 'tifinagh', 'de': 'Tifinagh', 'dsb': 'tifinanghske pismo', 'el': 'Τιφινάγκ', 'el-polyton': 'Τιφινάγκ', 'en': 'Tifinagh', 'en-Dsrt': '𐐓𐐮𐑁𐐮𐑌𐐪', 'es': 'tifinagh', 'et': 'tifinagi', 'eu': 'tifinagha', 'fa': 'تیفیناغی', 'ff-Adlm': '𞤚𞤭𞤬𞤭𞤲𞤢𞥄𞤺', 'fi': 'tifinagh', 'fil': 'Tifinagh', 'fo': 'tifinagh', 'fr': 'tifinagh', 'fy': 'Tifinagh', 'ga': 'Tifinagh', 'gd': 'Tifinagh', 'gl': 'tifinagh', 'gsw': 'Tifinagh', 'gu': 'તિફિનાઘ', 'ha': 'Tifinagh', 'he': 'טיפינה', 'hi': 'तिफ़िनाघ', 'hi-Latn': 'Tifinagh', 'hr': 'tifinagh pismo', 'hsb': 'tifinanghske pismo', 'ht': 'tifinagh', 'hu': 'Berber', 'hy': 'տիֆինաղ', 'id': 'Tifinagh', 'ig': 'Tifinagh', 'is': 'tifinagh', 'it': 'tifinagh', 'ja': 'ティフナグ文字', 'jv': 'Tifinak', 'ka': 'ტიფინაღი', 'kgp': 'tifinỹg', 'kk': 'тифинаг жазуы', 'kk-Cyrl': 'тифинаг жазуы', 'km': 'ទីហ្វ៊ីណាហ្វ', 'kn': 'ಟಿಫಿನಾಘ್', 'ko': '티피나그 문자', 'kok': 'टिफीनाघ', 'kok-Deva': 'टिफीनाघ', 'ks': 'تِفِناگ', 'ks-Arab': 'تِفِناگ', 'ky': 'Тифинаг (жазуу)', 'lb': 'Tifinagh', 'lo': 'ທີຟີນາກ', 'lt': 'tifinag', 'lv': 'tifinagu', 'mai': 'टिफिनघ', 'mi': 'Tiwhinā', 'mk': 'тифинаг', 'ml': 'തിഫിനാഗ്', 'mn': 'Тифинаг бичиг', 'mr': 'तिफिनाघ', 'ms': 'Tifinagh', 'my': 'တီဖီနော', 'nb': 'tifinagh', 'ne': 'टिफिनाघ', 'nl': 'Tifinagh', 'nn': 'tifinagh', 'no': 'tifinagh', 'nqo': 'ߕߝߌߣߊߜ߭', 'or': 'ଟିଫିନାଘ୍\u200c', 'pa': 'ਟਿਫੀਨਾਘ', 'pa-Guru': 'ਟਿਫੀਨਾਘ', 'pcm': 'Tífínag Lángwej', 'pl': 'tifinagh (berberski)', 'ps': 'ٹیفناګ', 'pt': 'tifinagh', 'qu': 'Tifinagh Simi', 'rm': 'tifinagh', 'ro': 'tifinagh', 'ru': 'древнеливийская', 'sc': 'tifinagh', 'sd': 'ٽفيناگ', 'sd-Arab': 'ٽفيناگ', 'si': 'ටිෆිනාග්', 'sk': 'tifinagh', 'sl': 'tifinajski', 'so': 'Farta Tifinagh', 'sq': 'tifinag', 'sr': 'тифинаг', 'sr-Cyrl': 'тифинаг', 'sr-Latn': 'tifinag', 'sv': 'tifinaghiska', 'sw': 'Kitifinagh', 'ta': 'டிஃபினாக்', 'te': 'టిఫీనాఘ్', 'tg': 'Тифинаг', 'th': 'ทิฟินาก', 'ti': 'ቲፊንጋ', 'tk': 'Tifinag ýazuwy', 'to': 'tohinima fakatifināki', 'tr': 'Tifinag', 'ug': 'تىفىناغچە', 'uk': 'тифінаг', 'ur': 'ٹفناگ', 'uz': 'tifinag', 'uz-Latn': 'tifinag', 'vec': 'tifinag', 'vi': 'Chữ Tifinagh', 'xnr': 'तिफिनाघ', 'yo': 'Èdè Tifina', 'yrl': 'tifinagi', 'yue': '提非納文', 'yue-Hans': '提非纳文', 'yue-Hant': '提非納文', 'zh': '提非纳文', 'zh-Hans': '提非纳文', 'zh-Hant': '提非納文', 'zu': 'isi-Tifinagh'},
    'Tglg': {'am': 'ታጋሎግ', 'ar': 'التغالوغية', 'ast': 'tagalog', 'az': 'taqaloq', 'az-Latn': 'taqaloq', 'bal-Latn': 'Tagalog', 'bg': 'Тагалог', 'bn': 'টাগালগ', 'br': 'tagalog', 'brx': 'टागालॉग', 'bs': 'tagalog pismo', 'bs-Cyrl': 'тагалог', 'bs-Latn': 'tagalog pismo', 'ca': 'tagàlog', 'ca-ES-valencia': 'tagàlog', 'ccp': '𑄑𑄉𑄣𑄧𑄉𑄴', 'cs': 'tagalské', 'da': 'tagalog', 'de': 'Tagalog', 'el': 'Ταγκαλόγκ', 'el-polyton': 'Ταγκαλόγκ', 'en': 'Tagalog', 'en-Dsrt': '𐐓𐐲𐑀𐐪𐑊𐐲𐑀', 'es': 'tagalo', 'et': 'tagalogi', 'eu': 'tagaloa', 'fa': 'تاگالوگی', 'ff-Adlm': '𞤚𞤢𞤺𞤢𞤤𞤮𞤺', 'fi': 'tagalogilainen', 'fo': 'tagalog', 'fr': 'tagal', 'fur': 'tagalog', 'fy': 'Tagalog', 'ga': 'Tagálagach', 'gd': 'Tagalog', 'gsw': 'Tagalog', 'gu': 'ટેગાલોગ', 'he': 'טגלוג', 'hi': 'टैगालोग', 'hi-Latn': 'Tagalog', 'hr': 'tagalog pismo', 'ht': 'tagal', 'hu': 'Tagalog', 'id': 'Tagalog', 'ig': 'Tagalog', 'it': 'tagalog', 'ja': 'タガログ文字', 'kgp': 'tagaru', 'kk': 'байбайин жазуы', 'kk-Cyrl': 'байбайин жазуы', 'kn': 'ಟ್ಯಾಗಲೋಗ್', 'ko': '타갈로그 문자', 'ks': 'تَگَلوگ', 'ks-Arab': 'تَگَلوگ', 'lb': 'Dagalog', 'lo': 'ຕາກາລອກ', 'lt': 'tagalogų', 'lv': 'tagalu', 'mk': 'тагалошко', 'ml': 'തഗലോഗ്', 'mn': 'Тагалог', 'mr': 'टागालोग', 'ms': 'Tagalog', 'my': 'တဂလော့ဂ်', 'nb': 'tagalog', 'ne': 'टागालोग', 'nl': 'Tagalog', 'nn': 'tagalog', 'no': 'tagalog', 'nqo': 'ߕߜ߭ߊߟ', 'or': 'ଟାଗାଲୋଗ୍', 'pl': 'tagalog', 'pt': 'tagalo', 'rm': 'tagalog', 'ro': 'tagalog', 'ru': 'тагалог', 'sc': 'tagalog', 'sl': 'tagaloški', 'so': 'Luuqada Tagalog', 'sq': 'tagalog', 'sr': 'Тагалог', 'sr-Cyrl': 'Тагалог', 'sr-Latn': 'Tagalog', 'sv': 'tagalog', 'ta': 'தகலாக்', 'te': 'టగలాగ్', 'tg': 'Тагалогӣ', 'th': 'ตากาล็อก', 'to': 'tohinima fakatakaloka', 'tr': 'Takalot', 'ug': 'تاگالوگچە', 'uk': 'тагальський', 'vec': 'tagalog', 'vi': 'Chữ Tagalog', 'xnr': 'टैगालोग', 'yrl': 'tagaru', 'yue': '塔加拉文', 'yue-Hans': '塔加拉文', 'yue-Hant': '塔加拉文', 'zh': '塔加路文', 'zh-Hans': '塔加路文', 'zh-Hant': '塔加拉文', 'zu': 'i-Tagalog'},
    'Thaa': {'af': 'Thaana', 'ak': 'Taana kasa', 'am': 'ታና', 'ar': 'الثعنة', 'as': 'থানা', 'ast': 'thaana', 'az': 'thana', 'az-Latn': 'thana', 'bal-Latn': 'Tána', 'be': 'тана', 'be-tarask': 'тана', 'bg': 'таана', 'bn': 'থানা', 'br': 'thaana', 'brx': 'थाना', 'bs': 'pismo tana', 'bs-Cyrl': 'тана писмо', 'bs-Latn': 'pismo tana', 'ca': 'thaana', 'ca-ES-valencia': 'thaana', 'ccp': '𑄗𑄚', 'ce': 'таана', 'chr': 'ᏔᎠᎾ', 'ckb': 'تانا', 'cs': 'thaana', 'cy': 'Thaana', 'da': 'thaana', 'de': 'Thaana', 'dsb': 'thaana', 'dz': 'ཐཱ་ན་ཡིག་གུ', 'ee': 'thaanagbeŋɔŋlɔ', 'el': 'Θαανά', 'el-polyton': 'Θαανά', 'en': 'Thaana', 'en-Dsrt': '𐐓𐐪𐐱𐑌𐐲', 'es': 'thaana', 'et': 'taana', 'eu': 'thaana', 'fa': 'تانا', 'ff-Adlm': '𞤡𞤢𞥄𞤲𞤢', 'fi': 'thaana', 'fil': 'Thaana', 'fo': 'thaana', 'fr': 'thâna', 'fur': 'thaana', 'fy': 'Thaana', 'ga': 'Tánach', 'gd': 'Thaana', 'gl': 'thaana', 'gsw': 'Thaana', 'gu': 'થાના', 'ha': 'Yaren Thaana', 'he': 'תאנה', 'hi': 'थाना', 'hi-Latn': 'Thaana', 'hr': 'thaana pismo', 'hsb': 'thaana', 'ht': 'thâna', 'hu': 'Thaana', 'hy': 'թաանա', 'ia': 'thaana', 'id': 'Thaana', 'ig': 'Thaana', 'is': 'thaana', 'it': 'thaana', 'ja': 'ターナ文字', 'jv': 'Thaana', 'ka': 'თაანა', 'kea': 'taana', 'kgp': 'ta’anỹ', 'kk': 'тана жазуы', 'kk-Cyrl': 'тана жазуы', 'km': 'ថាណា', 'kn': 'ಥಾನಾ', 'ko': '타나 문자', 'kok': 'थाणा', 'kok-Deva': 'थाणा', 'ks': 'تھانا', 'ks-Arab': 'تھانا', 'ksh': 'malledivesche Taana-Schreff', 'ky': 'Таана', 'lb': 'Thaana', 'lo': 'ທານາ', 'lrc': 'تانا', 'lt': 'hana', 'lv': 'tāna', 'mai': 'थाना', 'mi': 'Tāna', 'mk': 'тана', 'ml': 'ഥാന', 'mn': 'тана', 'mr': 'थाना', 'ms': 'Thaana', 'my': 'သာအ်', 'nb': 'taana', 'ne': 'थाना', 'nl': 'Thaana', 'nn': 'thaana', 'no': 'taana', 'nqo': 'ߕߡߊ', 'or': 'ଥାନା', 'pa': 'ਥਾਨਾ', 'pa-Guru': 'ਥਾਨਾ', 'pcm': 'Tána', 'pl': 'taana', 'ps': 'تهانا', 'pt': 'thaana', 'qu': 'Thaana Simi', 'rm': 'thaana', 'ro': 'thaana', 'ru': 'тана', 'sc': 'thaana', 'sd': 'ٿانا', 'sd-Arab': 'ٿانا', 'si': 'තාන', 'sk': 'tána', 'sl': 'tanajski', 'so': 'Daana', 'sq': 'tanisht', 'sr': 'тана писмо', 'sr-Cyrl': 'тана писмо', 'sr-Latn': 'tana pismo', 'sv': 'taana', 'sw': 'Kithaana', 'ta': 'தானா', 'te': 'థానా', 'tg': 'Таана', 'th': 'ทานา', 'ti': 'ትሃና', 'tk': 'Taana elipbiýi', 'to': 'tohinima fakatāna', 'tr': 'Thaana', 'ug': 'تاناچە', 'uk': 'таана', 'ur': 'تھانا', 'uz': 'taana', 'uz-Cyrl': 'Таана', 'uz-Latn': 'taana', 'vec': 'thana', 'vi': 'Chữ Thaana', 'wae': 'Thána', 'xnr': 'थाना', 'yo': 'Taana', 'yrl': 'ta-ana', 'yue': '塔安那文', 'yue-Hans': '塔安那文', 'yue-Hant': '塔安那文', 'zh': '塔安那文', 'zh-Hans': '塔安那文', 'zh-Hant': '塔安那文', 'zu': 'isi-Thaana'},
    'Thai': {'ak': 'Taelanfoɔ kasa', 'am': 'ታይ', 'ar': 'التايلاندية', 'as': 'থাই', 'ast': 'tailandés', 'az': 'tay', 'az-Latn': 'tay', 'bal-Latn': 'Tái', 'be': 'тайскае', 'be-tarask': 'тайскае', 'bg': 'тайска', 'bgn': 'تایی', 'bn': 'থাই', 'brx': 'थाई', 'bs': 'tajlandsko pismo', 'bs-Cyrl': 'тајландско писмо', 'bs-Latn': 'tajlandsko pismo', 'ca': 'tailandès', 'ca-ES-valencia': 'tailandès', 'ccp': '𑄗𑄭', 'ce': 'тайн', 'chr': 'ᏔᏱ ᏔᏯᎴᏂ', 'ckb': 'تایلەندی', 'cs': 'thajské', 'cy': 'Tai', 'da': 'thailandsk', 'dsb': 'thaiski', 'dz': 'ཐཱའི་ཡིག་གུ', 'ee': 'taigbeŋɔŋlɔ', 'el': 'Ταϊλανδικό', 'el-polyton': 'Ταϊλανδικό', 'en-Dsrt': '𐐓𐐴', 'es': 'tailandés', 'et': 'tai', 'eu': 'thailandiarra', 'fa': 'تایلندی', 'ff-Adlm': '𞤚𞤢𞤱𞤤𞤢𞤲𞤣', 'fi': 'thailainen', 'fo': 'tailendskt', 'fr': 'thaï', 'fy': 'Thais', 'ga': 'Téalannach', 'gd': 'Tàidh', 'gl': 'tailandés', 'gu': 'થાઇ', 'he': 'תאי', 'hi': 'थाई', 'hr': 'tajsko pismo', 'hsb': 'thailandsce', 'ht': 'thaï', 'hy': 'թայական', 'ia': 'thailandese', 'is': 'taílenskt', 'it': 'thailandese', 'ja': 'タイ文字', 'jv': 'Thailand', 'ka': 'ტაილანდური', 'kea': 'tailandes', 'kgp': 'tajrỹnej', 'kk': 'тай жазуы', 'kk-Cyrl': 'тай жазуы', 'km': 'ថៃ', 'kn': 'ಥಾಯ್', 'ko': '타이 문자', 'kok': 'थाई', 'kok-Deva': 'थाई', 'ks': 'تھاے', 'ks-Arab': 'تھاے', 'ksh': 'tailändesche Schreff', 'ku': 'tayî', 'ky': 'Тай', 'lo': 'ໄທ', 'lrc': 'تایلأندی', 'lt': 'tajų', 'lv': 'taju', 'mai': 'थाई', 'mi': 'Tai', 'mk': 'тајландско писмо', 'ml': 'തായ്', 'mn': 'тай', 'mr': 'थाई', 'my': 'ထိုင်း', 'ne': 'थाई', 'nqo': 'ߕߊߦߌ', 'or': 'ଥାଇ', 'pa': 'ਥਾਈ', 'pa-Guru': 'ਥਾਈ', 'pcm': 'Taí', 'pl': 'tajskie', 'ps': 'تایلنډي', 'pt': 'tailandês', 'qu': 'Tailandes Simi', 'rm': 'tailandais', 'ro': 'thailandeză', 'ru': 'тайская', 'sah': 'Таайдыы', 'sc': 'tailandesu', 'sd': 'ٿائي', 'sd-Arab': 'ٿائي', 'si': 'තායි', 'sk': 'thajské', 'sl': 'tajski', 'so': 'Taay', 'sq': 'tajlandez', 'sr': 'тајландско писмо', 'sr-Cyrl': 'тајландско писмо', 'sr-Latn': 'tajlandsko pismo', 'sv': 'thailändska', 'sw': 'Kithai', 'ta': 'தாய்', 'te': 'థాయ్', 'tg': 'Тайӣ', 'th': 'ไทย', 'ti': 'ታይ', 'tk': 'Taý elipbiýi', 'to': 'tohinima fakatailani', 'tr': 'Tay', 'ug': 'تايلاندچە', 'uk': 'тайська', 'ur': 'تھائی', 'uz': 'tay', 'uz-Cyrl': 'Тай', 'uz-Latn': 'tay', 'vec': 'tailandeze', 'vi': 'Chữ Thái', 'xnr': 'थाई', 'yrl': 'tairãdei', 'yue': '泰文', 'yue-Hans': '泰文', 'yue-Hant': '泰文', 'zh': '泰文', 'zh-Hans': '泰文', 'zh-Hant': '泰文', 'zu': 'isi-Thai', 'en': 'Thai'},
    'Tibt': {'af': 'Tibettaans', 'ak': 'Tibɛtanfoɔ kasa', 'am': 'ቲቤትኛ', 'ar': 'التبتية', 'as': 'তিব্বতী', 'ast': 'tibetanu', 'az': 'tibet', 'az-Arab': 'تبتجه', 'az-Latn': 'tibet', 'bal-Latn': 'Tebati', 'be': 'тыбецкае', 'be-tarask': 'тыбецкае', 'bg': 'тибетска', 'bn': 'তিব্বতি', 'bo': 'བོད་ཡིག་', 'br': 'tibetanek', 'brx': 'तिबेतान', 'bs': 'tibetansko pismo', 'bs-Cyrl': 'тибетанско писмо', 'bs-Latn': 'tibetansko pismo', 'ca': 'tibetà', 'ca-ES-valencia': 'tibetà', 'ccp': '𑄖𑄨𑄛𑄴𑄝𑄧𑄖𑄨', 'ce': 'тибетхойн', 'chr': 'ᏘᏇᏔᏂ', 'ckb': 'تەبەتی', 'cs': 'tibetské', 'cy': 'Tibetaidd', 'da': 'tibetansk', 'de': 'Tibetisch', 'dsb': 'tibetski', 'dz': 'ང་བཅས་ཀྱི་ཡིག་གུ', 'ee': 'tibetgbeŋɔŋlɔ', 'el': 'Θιβετιανό', 'el-polyton': 'Θιβετιανό', 'en': 'Tibetan', 'en-Dsrt': '𐐓𐐮𐐺𐐯𐐻𐐲𐑌', 'es': 'tibetano', 'et': 'tiibeti', 'eu': 'tibetarra', 'fa': 'تبتی', 'ff-Adlm': '𞤚𞤭𞤦𞤫𞤼𞤢𞤲', 'fi': 'tiibetiläinen', 'fil': 'Tibetan', 'fo': 'tibetskt', 'fr': 'tibétain', 'fur': 'tibetan', 'fy': 'Tibetaansk', 'ga': 'Tibéadach', 'gd': 'Tibeitis', 'gl': 'tibetano', 'gsw': 'Tibeetisch', 'gu': 'ટિબેટી', 'ha': 'Yaren Tibet', 'he': 'טיבטי', 'hi': 'तिब्बती', 'hi-Latn': 'Tibetan', 'hr': 'tibetansko pismo', 'hsb': 'tibetsce', 'ht': 'tibétain', 'hu': 'Tibeti', 'hy': 'տիբեթական', 'ia': 'tibetan', 'id': 'Tibet', 'ig': 'Tibetan', 'is': 'tíbeskt', 'it': 'tibetano', 'ja': 'チベット文字', 'jv': 'Tibet', 'ka': 'ტიბეტური', 'kea': 'tibetanu', 'kgp': 'timetỹnũ', 'kk': 'тибет жазуы', 'kk-Cyrl': 'тибет жазуы', 'km': 'ទីបេ', 'kn': 'ಟಿಬೇಟನ್', 'ko': '티베트 문자', 'kok': 'तिबेटन', 'kok-Deva': 'तिबेटन', 'ks': 'تِبتی', 'ks-Arab': 'تِبتی', 'ksh': 'tibeetesche Schreff', 'ku': 'tîbetî', 'ky': 'Тибет', 'lb': 'Tibetesch', 'lo': 'ທິເບທັນ', 'lrc': 'تأبأتی', 'lt': 'tibetiečių', 'lv': 'tibetiešu', 'mai': 'तिब्बती', 'mi': 'Tipete', 'mk': 'тибетско писмо', 'ml': 'ടിബറ്റൻ', 'mn': 'төвд', 'mr': 'तिबेटी', 'ms': 'Tibet', 'my': 'တိဘက်', 'nb': 'tibetansk', 'ne': 'तिब्बती', 'nl': 'Tibetaans', 'nn': 'tibetansk', 'no': 'tibetansk', 'nqo': 'ߕߓߌߕߋߕߍ߲', 'or': 'ତିବେତାନ୍', 'pa': 'ਤਿੱਬਤੀ', 'pa-Guru': 'ਤਿੱਬਤੀ', 'pcm': 'Tíbẹt', 'pl': 'tybetańskie', 'ps': 'تبتي', 'pt': 'tibetano', 'qu': 'Tibetano Simi', 'rm': 'tibetan', 'ro': 'tibetană', 'ru': 'тибетская', 'sc': 'tibetanu', 'sd': 'تبيتن', 'sd-Arab': 'تبيتن', 'si': 'ටි\u200dබෙට්', 'sk': 'tibetské', 'sl': 'tibetanski', 'so': 'Tibetaan', 'sq': 'tibetisht', 'sr': 'тибетанско писмо', 'sr-Cyrl': 'тибетанско писмо', 'sr-Latn': 'tibetansko pismo', 'sv': 'tibetanska', 'sw': 'Kitibeti', 'ta': 'திபெத்தியன்', 'te': 'టిబెటన్', 'tg': 'Тибетӣ', 'th': 'ทิเบต', 'ti': 'ቲቤት', 'tk': 'Tibet elipbiýi', 'to': 'tohinima fakataipeti', 'tr': 'Tibet', 'ug': 'تىبەتچە', 'uk': 'тибетська', 'ur': 'تبتی', 'uz': 'tibet', 'uz-Cyrl': 'Тибет', 'uz-Latn': 'tibet', 'vec': 'tibetan', 'vi': 'Chữ Tây Tạng', 'xnr': 'तिब्बती', 'yo': 'Tíbétán', 'yrl': 'tibetanu', 'yue': '西藏文', 'yue-Hans': '西藏文', 'yue-Hant': '西藏文', 'zh': '藏文', 'zh-Hans': '藏文', 'zh-Hant': '藏文', 'zu': 'i-Tibetan'},
    'Tirh': {'ast': 'tirhuta', 'bal-Latn': 'Tirhutá', 'bs': 'tirhuta pismo', 'bs-Cyrl': 'тирхута писмо', 'bs-Latn': 'tirhuta pismo', 'ca': 'tirhut', 'ca-ES-valencia': 'tirhut', 'cs': 'tirhuta', 'da': 'tirhuta', 'de': 'Tirhuta', 'en': 'Tirhuta', 'et': 'tirhuta', 'eu': 'tirhuta', 'ff-Adlm': '𞤚𞤭𞤪𞤸𞤵𞤼𞤢', 'fi': 'tirhuta', 'fo': 'tirhuta', 'fy': 'Tirhuta', 'ga': 'Tirhuta', 'gd': 'Tirhuta', 'he': 'תרותא', 'hi-Latn': 'Tirhuta', 'hr': 'tirhuta pismo', 'id': 'Tirhuta', 'ig': 'Tirhuta', 'it': 'tirhuta', 'ja': 'ティルフータ文字', 'ka': 'ტირჰუტა', 'kk': 'тирхута жазуы', 'kk-Cyrl': 'тирхута жазуы', 'ko': '티르후타 문자', 'lo': 'ເທຮຸທາ', 'lt': 'Tirhuta', 'mk': 'тирхута', 'mn': 'Тирхута', 'ms': 'Tirhuta', 'nb': 'tirhuta', 'nl': 'Tirhuta', 'nn': 'tirhuta', 'no': 'tirhuta', 'nqo': 'ߕߌߙߎߕߊ', 'ro': 'tirhuta', 'ru': 'тирхута', 'sc': 'tirhuta', 'so': 'Qoraalka Luuqada Maithili', 'sq': 'tirhuta', 'sv': 'tirhuta', 'tg': 'Тирхута', 'th': 'เทอฮุทา', 'to': 'tohinima fakatīhuta', 'tr': 'Tirhuta', 'ug': 'تىرخۇتاچە', 'vec': 'tiruta', 'vi': 'Chữ Tirhuta', 'yue': '邁蒂利文', 'yue-Hans': '迈蒂利文', 'yue-Hant': '邁蒂利文', 'zh': '迈蒂利文', 'zh-Hans': '迈蒂利文', 'zh-Hant': '邁蒂利文', 'zu': 'i-Tirhuta'},
    'Tnsa': {'bal-Latn': 'Tangsá', 'bs': 'tangsa pismo', 'bs-Cyrl': 'тангса писмо', 'bs-Latn': 'tangsa pismo', 'en': 'Tangsa', 'et': 'tase', 'eu': 'tangsa', 'ff-Adlm': '𞤚𞤢𞤲𞤺𞤧𞤢', 'fi': 'tangsa', 'fo': 'tangsa', 'gd': 'Tangsa', 'he': 'טנגסה', 'hi-Latn': 'Tangsa', 'id': 'Tangsa', 'ig': 'Tangsa', 'kk': 'тангса жазуы', 'kk-Cyrl': 'тангса жазуы', 'mn': 'Тангса', 'nl': 'Tangsa', 'nqo': 'ߕߊ߲ߜߑߛߊ߫', 'ro': 'tangsa', 'sc': 'tangsa', 'sq': 'tangsa', 'tg': 'Тангса', 'vec': 'tangsa', 'zh': '坦萨文', 'zh-Hans': '坦萨文'},
    'Todr': {'bal-Latn': 'Tódri', 'en': 'Todhri', 'fo': 'todhri', 'gd': 'Todhri', 'hi-Latn': 'Todhri', 'ig': 'Todhri', 'vec': 'todhri'},
    'Toto': {'bal-Latn': 'Tótó', 'bs': 'toto pismo', 'bs-Cyrl': 'тото писмо', 'bs-Latn': 'toto pismo', 'eu': 'totoera', 'ff-Adlm': '𞤚𞤮𞤼𞤮', 'he': 'טוטו', 'id': 'Toto (txo)', 'kk': 'тото жазуы', 'kk-Cyrl': 'тото жазуы', 'mn': 'Тото', 'nl': 'Totoschrift', 'nqo': 'ߕߏߕߏ߫', 'tg': 'Тото', 'zh': '投投文', 'zh-Hans': '投投文', 'en': 'Toto'},
    'Tutg': {'bal-Latn': 'Tulu-Tigálari', 'en': 'Tulu-Tigalari', 'fo': 'tulu-tigalari', 'gd': 'Tulu-Tigalari', 'hi-Latn': 'Tulu-Tigalari', 'ig': 'Tulu-Tigalari', 'vec': 'tigalari'},
    'UA': {'af': 'Oekraïne', 'agq': 'Yùkɛ̀lɛ̂', 'ak': 'Ukren', 'am': 'ዩክሬን', 'ar': 'أوكرانيا', 'as': 'ইউক্ৰেইন', 'ast': 'Ucraína', 'az': 'Ukrayna', 'az-Cyrl': 'Украјна', 'az-Latn': 'Ukrayna', 'bal-Latn': 'Yukrén', 'bas': 'Ùkrɛ̌n', 'be': 'Украіна', 'be-tarask': 'Украіна', 'bez': 'Huukraini', 'bg': 'Украйна', 'blo': 'Ikrɛɛn', 'bm': 'Ukɛrɛni', 'bn': 'ইউক্রেন', 'br': 'Ukraina', 'brx': 'यूक्रेन', 'bs': 'Ukrajina', 'bs-Cyrl': 'Украјина', 'bs-Latn': 'Ukrajina', 'ca': 'Ucraïna', 'ca-ES-valencia': 'Ucraïna', 'ccp': '𑄃𑄨𑄃𑄪𑄇𑄳𑄢𑄬𑄚𑄴', 'ce': 'Украина', 'ceb': 'Ukraine', 'cgg': 'Ukureini', 'chr': 'ᏳᎧᎴᏂ', 'ckb': 'ئۆکرانیا', 'cs': 'Ukrajina', 'cv': 'Украина', 'cy': 'Wcráin', 'da': 'Ukraine', 'dav': 'Ukraini', 'de': 'Ukraine', 'dje': 'Ukreen', 'dsb': 'Ukraina', 'dz': 'ཡུ་ཀརེན', 'ebu': 'Ukraini', 'ee': 'Ukraine nutome', 'el': 'Ουκρανία', 'el-polyton': 'Οὐκρανία', 'en': 'Ukraine', 'eo': 'Ukrainujo', 'es': 'Ucrania', 'et': 'Ukraina', 'eu': 'Ukraina', 'ewo': 'Ukərɛ́n', 'fa': 'اوکراین', 'ff': 'Ukereen', 'ff-Adlm': '𞤓𞤳𞤪𞤫𞥅𞤲𞤭𞤴𞤢', 'ff-Latn': 'Ukereen', 'fi': 'Ukraina', 'fil': 'Ukraine', 'fo': 'Ukraina', 'fr': 'Ukraine', 'fur': 'Ucraine', 'fy': 'Oekraïne', 'ga': 'an Úcráin', 'gd': 'An Ucràin', 'gl': 'Ucraína', 'gsw': 'Ukraiine', 'gu': 'યુક્રેન', 'guz': 'Ukraini', 'ha': 'Yukaran', 'he': 'אוקראינה', 'hi': 'यूक्रेन', 'hi-Latn': 'Ukraine', 'hr': 'Ukrajina', 'hsb': 'Ukraina', 'ht': 'Ukraine', 'hu': 'Ukrajna', 'hy': 'Ուկրաինա', 'ia': 'Ukraina', 'id': 'Ukraina', 'ie': 'Ukraina', 'ig': 'Ukraine', 'is': 'Úkraína', 'it': 'Ucraina', 'ja': 'ウクライナ', 'jmc': 'Ukraini', 'jv': 'Ukrania', 'ka': 'უკრაინა', 'kaa': 'Украина', 'kaa-Cyrl': 'Украина', 'kab': 'Ukran', 'kam': 'Ukraini', 'kde': 'Uklaini', 'kea': 'Ukránia', 'kgp': 'Ukrỹnĩja', 'khq': 'Ukreen', 'ki': 'Ukraini', 'kk': 'Украина', 'kk-Cyrl': 'Украина', 'kln': 'Emetab Ukrainie', 'km': 'អ៊ុយក្រែន', 'kn': 'ಉಕ್ರೈನ್', 'ko': '우크라이나', 'kok': 'युक्रेन', 'kok-Deva': 'युक्रेन', 'ks': 'یوٗرِکین', 'ks-Arab': 'یوٗرِکین', 'ksb': 'Uklaini', 'ksf': 'ukrain', 'ksh': 'de Ukra’iine', 'ku': 'Ûkrayna', 'kxv': 'yūkrain', 'kxv-Deva': 'यूक्रेन', 'kxv-Latn': 'yūkrain', 'kxv-Orya': 'ୟୁକ୍ରେନ୍', 'kxv-Telu': 'ఉక్రెయిన్', 'ky': 'Украина', 'lag': 'Ʉkɨréeni', 'lb': 'Ukrain', 'lg': 'Yukurayine', 'lld': 'Ucraina', 'ln': 'Ikrɛni', 'lo': 'ຢູເຄຣນ', 'lt': 'Ukraina', 'lu': 'Ukreni', 'luo': 'Ukraine', 'luy': 'Ukraini', 'lv': 'Ukraina', 'mai': 'यूक्रेन', 'mas': 'Ukraini', 'mer': 'Ukirĩni', 'mfe': 'Ikrenn', 'mg': 'Okraina', 'mi': 'Ukareinga', 'mk': 'Украина', 'ml': 'ഉക്രെയ്\u200cൻ', 'mn': 'Украин', 'mr': 'युक्रेन', 'ms': 'Ukraine', 'mt': 'l-Ukrajna', 'mua': 'Ukraiŋ', 'my': 'ယူကရိန်း', 'mzn': 'اوکراین', 'naq': 'Ukraine', 'nb': 'Ukraina', 'nd': 'Yukreini', 'ne': 'युक्रेन', 'nl': 'Oekraïne', 'nmg': 'Ukrɛn', 'nn': 'Ukraina', 'no': 'Ukraina', 'nqo': 'ߎ߳ߞߑߙߍߣ', 'nyn': 'Ukureini', 'om': 'Yuukireen', 'or': 'ୟୁକ୍ରେନ୍', 'pa': 'ਯੂਕਰੇਨ', 'pa-Guru': 'ਯੂਕਰੇਨ', 'pcm': 'Yukrein', 'pl': 'Ukraina', 'ps': 'اوکراین', 'pt': 'Ucrânia', 'qu': 'Ucrania', 'rm': 'Ucraina', 'rn': 'Ikerene', 'ro': 'Ucraina', 'rof': 'Ukraini', 'ru': 'Украина', 'rwk': 'Ukraini', 'saq': 'Ukraini', 'sbp': 'Yukileini', 'sc': 'Ucraina', 'sd': 'يوڪرين', 'sd-Arab': 'يوڪرين', 'se': 'Ukraina', 'seh': 'Ucrânia', 'ses': 'Ukreen', 'sg': 'Ukrêni', 'shi': 'ⵓⴽⵔⴰⵏⵢⴰ', 'shi-Latn': 'ukranya', 'shi-Tfng': 'ⵓⴽⵔⴰⵏⵢⴰ', 'si': 'යුක්රේනය', 'sk': 'Ukrajina', 'sl': 'Ukrajina', 'smn': 'Ukraina', 'sn': 'Ukraine', 'so': 'Yukrayn', 'sq': 'Ukrainë', 'sr': 'Украјина', 'sr-Cyrl': 'Украјина', 'sr-Latn': 'Ukrajina', 'sv': 'Ukraina', 'sw': 'Ukraine', 'syr': 'ܐܘܩܪܐܝܢܐ', 'ta': 'உக்ரைன்', 'te': 'ఉక్రెయిన్', 'teo': 'Ukraini', 'tg': 'Украина', 'th': 'ยูเครน', 'ti': 'ዩክሬን', 'tk': 'Ukraina', 'to': 'ʻŪkalaʻine', 'tr': 'Ukrayna', 'tt': 'Украина', 'twq': 'Ukreen', 'tzm': 'Ukranya', 'ug': 'ئۇكرائىنا', 'uk': 'Україна', 'ur': 'یوکرین', 'uz': 'Ukraina', 'uz-Cyrl': 'Украина', 'uz-Latn': 'Ukraina', 'vai': 'ꖳꖴꔓꘋ', 'vai-Latn': 'Yukuréŋ', 'vai-Vaii': 'ꖳꖴꔓꘋ', 'vec': 'Ucraina', 'vi': 'Ukraina', 'vun': 'Ukraini', 'wae': 'Ukraine', 'wo': 'Ikeren', 'xh': 'E-Ukraine', 'xnr': 'यूक्रेन', 'xog': 'Yukurayine', 'yav': 'ukilɛ́ɛn', 'yi': 'אוקראַינע', 'yo': 'Ukarini', 'yrl': 'Ukarãniya', 'yue': '烏克蘭', 'yue-Hans': '乌克兰', 'yue-Hant': '烏克蘭', 'zgh': 'ⵓⴽⵔⴰⵏⵢⴰ', 'zh': '乌克兰', 'zh-Hans': '乌克兰', 'zh-Hant': '烏克蘭', 'zu': 'i-Ukraine'},
    'UG': {'af': 'Uganda', 'agq': 'Yùgandà', 'ak': 'Yuganda', 'am': 'ዩጋንዳ', 'ar': 'أوغندا', 'as': 'উগাণ্ডা', 'asa': 'Uganda', 'ast': 'Uganda', 'az': 'Uqanda', 'az-Cyrl': 'Уганда', 'az-Latn': 'Uqanda', 'bal-Latn': 'Yugandhá', 'bas': 'Ùgandà', 'be': 'Уганда', 'be-tarask': 'Уганда', 'bez': 'Huuganda', 'bg': 'Уганда', 'bgn': 'اوگاندا', 'blo': 'Uganɖaa', 'bm': 'Uganda', 'bn': 'উগান্ডা', 'br': 'Ouganda', 'brx': 'उगाण्डा', 'bs': 'Uganda', 'bs-Cyrl': 'Уганда', 'bs-Latn': 'Uganda', 'ca': 'Uganda', 'ca-ES-valencia': 'Uganda', 'ccp': '𑄅𑄉𑄚𑄴𑄓', 'ce': 'Уганда', 'ceb': 'Uganda', 'cgg': 'Uganda', 'chr': 'ᏳᎦᏂᏓ', 'ckb': 'ئوگاندا', 'cs': 'Uganda', 'cv': 'Уганда', 'cy': 'Uganda', 'da': 'Uganda', 'dav': 'Uganda', 'de': 'Uganda', 'dje': 'Uganda', 'dsb': 'Uganda', 'dz': 'ཡུ་གྷན་ཌ', 'ebu': 'Uganda', 'ee': 'Uganda nutome', 'el': 'Ουγκάντα', 'el-polyton': 'Οὐγκάντα', 'en': 'Uganda', 'en-Dsrt': '𐐏𐐭𐑀𐐰𐑌𐐼𐐲', 'eo': 'Ugando', 'es': 'Uganda', 'et': 'Uganda', 'eu': 'Uganda', 'ewo': 'Ugandá', 'fa': 'اوگاندا', 'ff': 'Unganndaa', 'ff-Adlm': '𞤓𞤺𞤢𞤲𞤣𞤢𞥄', 'ff-Latn': 'Unganndaa', 'fi': 'Uganda', 'fil': 'Uganda', 'fo': 'Uganda', 'fr': 'Ouganda', 'fur': 'Uganda', 'fy': 'Oeganda', 'ga': 'Uganda', 'gd': 'Uganda', 'gl': 'Uganda', 'gsw': 'Uganda', 'gu': 'યુગાંડા', 'guz': 'Uganda', 'ha': 'Yuganda', 'he': 'אוגנדה', 'hi': 'युगांडा', 'hi-Latn': 'Uganda', 'hr': 'Uganda', 'hsb': 'Uganda', 'ht': 'Ouganda', 'hu': 'Uganda', 'hy': 'Ուգանդա', 'ia': 'Uganda', 'id': 'Uganda', 'ig': 'Uganda', 'is': 'Úganda', 'it': 'Uganda', 'ja': 'ウガンダ', 'jgo': 'Uŋgánda', 'jmc': 'Uganda', 'jv': 'Uganda', 'ka': 'უგანდა', 'kaa': 'Уганда', 'kaa-Cyrl': 'Уганда', 'kab': 'Uɣanda', 'kam': 'Ukanda', 'kde': 'Uganda', 'kea': 'Uganda', 'kgp': 'Ugỹna', 'khq': 'Uganda', 'ki': 'Uganda', 'kk': 'Уганда', 'kk-Cyrl': 'Уганда', 'kln': 'Emetab Uganda', 'km': 'អ៊ូហ្គង់ដា', 'kn': 'ಉಗಾಂಡಾ', 'ko': '우간다', 'kok': 'युगांडा', 'kok-Deva': 'युगांडा', 'ks': 'یوٗگانڑا', 'ks-Arab': 'یوٗگانڑا', 'ksb': 'Uganda', 'ksf': 'uganda', 'ksh': 'Ujanda', 'ku': 'Ûganda', 'kxv': 'yūganḍa', 'kxv-Deva': 'युगांडा', 'kxv-Latn': 'yūganḍa', 'kxv-Orya': 'ଉଗାଣ୍ଡା', 'kxv-Telu': 'ఉగాండా', 'ky': 'Уганда', 'lag': 'Ʉgáanda', 'lb': 'Uganda', 'lg': 'Yuganda', 'lld': 'Uganda', 'ln': 'Uganda', 'lo': 'ອູການດາ', 'lt': 'Uganda', 'lu': 'Uganda', 'luo': 'Uganda', 'luy': 'Uganda', 'lv': 'Uganda', 'mai': 'युगांडा', 'mas': 'Uganda', 'mer': 'Uganda', 'mfe': 'Ouganda', 'mg': 'Oganda', 'mi': 'Ukānga', 'mk': 'Уганда', 'ml': 'ഉഗാണ്ട', 'mn': 'Уганда', 'mr': 'युगांडा', 'ms': 'Uganda', 'mt': 'l-Uganda', 'mua': 'Uganda', 'my': 'ယူဂန်းဒါး', 'mzn': 'اوگاندا', 'naq': 'Ugandab', 'nb': 'Uganda', 'nd': 'Uganda', 'ne': 'युगाण्डा', 'nl': 'Oeganda', 'nmg': 'Uganda', 'nn': 'Uganda', 'no': 'Uganda', 'nqo': 'ߎߜ߭ߊ߲ߘߊ߫', 'nyn': 'Uganda', 'om': 'Ugaandaa', 'or': 'ଉଗାଣ୍ଡା', 'pa': 'ਯੂਗਾਂਡਾ', 'pa-Guru': 'ਯੂਗਾਂਡਾ', 'pcm': 'Yugánda', 'pl': 'Uganda', 'ps': 'یوګانډا', 'pt': 'Uganda', 'qu': 'Uganda', 'rm': 'Uganda', 'rn': 'Ubugande', 'ro': 'Uganda', 'rof': 'Uganda', 'ru': 'Уганда', 'rwk': 'Uganda', 'saq': 'Uganda', 'sbp': 'Uganda', 'sc': 'Uganda', 'sd': 'يوگنڊا', 'sd-Arab': 'يوگنڊا', 'se': 'Uganda', 'seh': 'Uganda', 'ses': 'Uganda', 'sg': 'Ugandäa', 'shi': 'ⵓⵖⴰⵏⴷⴰ', 'shi-Latn': 'uɣanda', 'shi-Tfng': 'ⵓⵖⴰⵏⴷⴰ', 'si': 'උගන්ඩාව', 'sk': 'Uganda', 'sl': 'Uganda', 'smn': 'Uganda', 'sn': 'Uganda', 'so': 'Ugaanda', 'sq': 'Ugandë', 'sr': 'Уганда', 'sr-Cyrl': 'Уганда', 'sr-Latn': 'Uganda', 'sv': 'Uganda', 'sw': 'Uganda', 'syr': 'ܐܘܓܢܕܐ', 'ta': 'உகாண்டா', 'te': 'ఉగాండా', 'teo': 'Uganda', 'tg': 'Уганда', 'th': 'ยูกันดา', 'ti': 'ኡጋንዳ', 'tk': 'Uganda', 'to': 'ʻIukanitā', 'tr': 'Uganda', 'tt': 'Уганда', 'twq': 'Uganda', 'tzm': 'Uɣanda', 'ug': 'ئۇگاندا', 'uk': 'Уганда', 'ur': 'یوگنڈا', 'uz': 'Uganda', 'uz-Cyrl': 'Уганда', 'uz-Latn': 'Uganda', 'vai': 'ꖳꕭꕡ', 'vai-Latn': 'Yuganda', 'vai-Vaii': 'ꖳꕭꕡ', 'vec': 'Uganda', 'vi': 'Uganda', 'vun': 'Uganda', 'wae': 'Uganda', 'wo': 'Ugànda', 'xh': 'E-Uganda', 'xnr': 'युगांडा', 'xog': 'Yuganda', 'yav': 'ukánda', 'yi': 'אוגאַנדע', 'yo': 'Uganda', 'yrl': 'Ugãda', 'yue': '烏干達', 'yue-Hans': '乌干达', 'yue-Hant': '烏干達', 'zgh': 'ⵓⵖⴰⵏⴷⴰ', 'zh': '乌干达', 'zh-Hans': '乌干达', 'zh-Hant': '烏干達', 'zu': 'i-Uganda'},
    'UM': {'af': 'Klein afgeleë eilande van die VSA', 'ak': 'U.S. Nkyɛnnkyɛn Supɔ Ahodoɔ', 'am': 'የዩ ኤስ ጠረፍ ላይ ያሉ ደሴቶች', 'ar': 'جزر الولايات المتحدة النائية', 'as': 'ইউ. এছ. আউটলায়িং দ্বীপপুঞ্জ', 'ast': 'Islles Perifériques Menores de los EE.XX.', 'az': 'ABŞ-a bağlı kiçik adacıqlar', 'az-Cyrl': 'АБШ-а бағлы кичик адаҹыглар', 'az-Latn': 'ABŞ-a bağlı kiçik adacıqlar', 'bal-Latn': 'U.S. Daráén Islánd', 'be': 'Малыя Аддаленыя астравы ЗША', 'be-tarask': 'Малыя Аддаленыя астравы ЗША', 'bg': 'Отдалечени острови на САЩ', 'blo': 'Ganɔ gaɖɔŋkɔnɔ kaBʊtǝna Amalɩka nɩ kaBʊtukǝltǝna bʊlǝŋka', 'bn': 'যুক্তরাষ্ট্রের পার্শ্ববর্তী দ্বীপপুঞ্জ', 'br': 'Inizi diabell ar Stadoù-Unanet', 'brx': 'इउ.एस नि बायजोआरि द्वीपफोर', 'bs': 'Američka Vanjska Ostrva', 'bs-Cyrl': 'Мања удаљена острва САД', 'bs-Latn': 'Američka Vanjska Ostrva', 'ca': 'Illes Menors Allunyades dels Estats Units', 'ca-ES-valencia': 'Illes Menors Allunyades dels Estats Units', 'ccp': '𑄎𑄧𑄙𑄢𑄬𑄌𑄴𑄎𑄮𑄢𑄴 𑄦𑄭𑄇𑄪𑄢𑄬 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'АЦШн арахьара кегийн гӀайренаш', 'ceb': 'U.S. Outlying Islands', 'chr': 'U.S. ᎠᏍᏛ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی دەرەوەی ئەمریکا', 'cs': 'Menší odlehlé ostrovy USA', 'cv': 'Тулашӗнчи утравӗсем (АПШ)', 'cy': 'Ynysoedd Pellennig UDA', 'da': 'Amerikanske oversøiske øer', 'de': 'Amerikanische Überseeinseln', 'dsb': 'Ameriska Oceaniska', 'dz': 'ཡུ་ཨེས་གྱི་མཐའ་མཚམས་མཚོ་གླིང་', 'ee': 'U.S. Minor Outlaying ƒudomekpowo nutome', 'el': 'Απομακρυσμένες Νησίδες ΗΠΑ', 'el-polyton': 'Ἀπομακρυσμένες Νησίδες τῶν Ἡνωμένων Πολιτειῶν', 'en': 'U.S. Outlying Islands', 'en-Dsrt': '𐐏𐐭𐑌𐐰𐐮𐐻𐐲𐐼 𐐝𐐻𐐩𐐻𐑅 𐐣𐐴𐑌𐐬𐑉 𐐍𐐻𐑊𐐴𐐨𐑍 𐐌𐑊𐐲𐑌𐐼𐑆', 'eo': 'Usonaj malgrandaj insuloj', 'es': 'Islas menores alejadas de EE. UU.', 'et': 'Ühendriikide hajasaared', 'eu': 'Ameriketako Estatu Batuetako Kanpoaldeko Uharte Txikiak', 'fa': 'جزایر دورافتادهٔ ایالات متحده', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤁𞤢𞥄𞤴𞤭𞥅𞤯𞤫 𞤁𞤂𞤀', 'fi': 'Yhdysvaltain erillissaaret', 'fil': 'U.S. Outlying Islands', 'fo': 'Sambandsríki Amerikas fjarskotnu oyggjar', 'fr': 'Îles mineures éloignées des États-Unis', 'fur': 'Isulis periferichis minôrs dai Stâts Unîts', 'fy': 'Lyts ôflizzen eilannen fan de Ferienigde Staten', 'ga': 'Oileáin Imeallacha S.A.M.', 'gd': 'Meanbh-Eileanan Iomallach nan SA', 'gl': 'Illas Menores Distantes dos Estados Unidos', 'gsw': 'Amerikanisch-Ozeaanie', 'gu': 'યુ.એસ. આઉટલાઇનિંગ આઇલેન્ડ્સ', 'ha': 'Rukunin Tsibirin U.S', 'he': 'האיים המרוחקים הקטנים של ארה״ב', 'hi': 'यू॰एस॰ आउटलाइंग द्वीपसमूह', 'hi-Latn': 'U.S. Outlying Islands', 'hr': 'Mali udaljeni otoci SAD-a', 'hsb': 'Ameriska Oceaniska', 'ht': 'Îles mineures éloignées des États-Unis', 'hu': 'Az USA lakatlan külbirtokai', 'hy': 'Արտաքին կղզիներ (ԱՄՆ)', 'ia': 'Insulas peripheric del SUA', 'id': 'Kepulauan Terluar AS', 'ig': 'Obere Agwaetiti Dị Na Mpụga U.S', 'is': 'Smáeyjar Bandaríkjanna', 'it': 'Isole Minori Esterne degli Stati Uniti', 'ja': '合衆国領有小離島', 'jv': 'Kapuloan AS Paling Njaba', 'ka': 'აშშ-ის შორეული კუნძულები', 'kea': 'Ilhas Minoris Distantis de Stadus Unidus', 'kgp': 'EUA Goj-vẽso Kãsir', 'kk': 'АҚШ-тың сыртқы кіші аралдары', 'kk-Cyrl': 'АҚШ-тың сыртқы кіші аралдары', 'km': 'កោះ\u200bអៅឡាយីង\u200bអាមេរិក', 'kn': 'ಯುಎಸ್\u200c ಔಟ್\u200cಲೇಯಿಂಗ್ ದ್ವೀಪಗಳು', 'ko': '미국령 해외 제도', 'kok': 'यु. एस. मायनर आवटलायींग आयलँड्\u200dस', 'kok-Deva': 'यु. एस. मायनर आवटलायींग आयलँड्\u200dस', 'ks': 'یوٗنایٹِڑ سِٹیٹِس ماینَر آوُٹلییِنگ جٔزیٖرٕ', 'ks-Arab': 'یوٗنایٹِڑ سِٹیٹِس ماینَر آوُٹلییِنگ جٔزیٖرٕ', 'ksh': 'de Vereineschte Schtaate vun Amärrika ier ußerhallef jelääje Enselsche', 'ku': 'Giravên Biçûk ên Derveyî DYAyê', 'kxv': 'yu.es. aautlaing dīp', 'kxv-Deva': 'यू॰एस॰ आउटलाइंग द्वीप', 'kxv-Latn': 'yu.es. aautlaing dīp', 'kxv-Orya': 'ୟୁ ଏସ୍ ଆଉଟଲାଇଙ୍କ ଦିପ', 'kxv-Telu': 'సంయుక్త రాజ్య అమెరికా బయట ఉన్న దీవులు', 'ky': 'АКШнын сырткы аралдары', 'lb': 'Amerikanesch-Ozeanien', 'lld': 'Isoles mëndres dlafora di Stac Unis', 'lo': 'ໝູ່ເກາະຮອບນອກຂອງສະຫະລັດຯ', 'lt': 'Jungtinių Valstijų Mažosios Tolimosios Salos', 'lv': 'ASV Mazās Aizjūras salas', 'mai': 'यू॰एस॰ आउटलाइंग द्वीपसमूह', 'mi': 'Ngā Moutere Amerika o Waho', 'mk': 'Американски територии во Пацификот', 'ml': 'യു.എസ്. ദ്വീപസമൂഹങ്ങൾ', 'mn': 'Америкийн Нэгдсэн Улсын бага арлууд', 'mr': 'यू.एस. आउटलाइंग बेटे', 'ms': 'Kepulauan Terpencil A.S.', 'mt': 'Il-Gżejjer Minuri Mbiegħda tal-Istati Uniti', 'my': 'ယူနိုက်တက်စတိတ် ကျွန်းနိုင်ငံများ', 'mzn': 'آمریکای پَرتِ\u200cپِلا جزیره\u200cئون', 'nb': 'USAs ytre øyer', 'ne': 'संयुक्त राज्यका बाह्य टापुहरु', 'nl': 'Kleine afgelegen eilanden van de Verenigde Staten', 'nn': 'USAs ytre småøyar', 'no': 'USAs ytre øyer', 'nqo': 'ߞߊ߬ߝߏ߫ ߘߍ߬ߣߍ߲ ߘߌߣߍ߲߫ ߕߌ߲߫ ߡߊߕߊ߯ߣߍ߲', 'om': 'U.S. Odoloota Alaa', 'or': 'ଯୁକ୍ତରାଷ୍ଟ୍ର ଆଉଟ୍\u200cଲାଇଙ୍ଗ ଦ୍ଵୀପପୁଞ୍ଜ', 'pa': 'ਯੂ.ਐੱਸ. ਦੂਰ-ਦੁਰਾਡੇ ਟਾਪੂ', 'pa-Guru': 'ਯੂ.ਐੱਸ. ਦੂਰ-ਦੁਰਾਡੇ ਟਾਪੂ', 'pcm': 'U.S. Faá Faá Aílands', 'pl': 'Dalekie Wyspy Mniejsze Stanów Zjednoczonych', 'ps': 'د متحده ایالاتو ټاپوګان', 'pt': 'Ilhas Menores Distantes dos EUA', 'qu': 'Islas menores alejadas de los EE.UU.', 'rm': 'Inslas Pitschnas Perifericas dals Stadis Unids da l’America', 'ro': 'Insulele Îndepărtate ale S.U.A.', 'ru': 'Внешние малые о-ва (США)', 'sc': 'Ìsulas perifèricas de sos Istados Unidos', 'sd': 'آمريڪي خارجي ٻيٽ', 'sd-Arab': 'آمريڪي خارجي ٻيٽ', 'si': 'එක්සත් ජනපද ඈත දූපත්', 'sk': 'Menšie odľahlé ostrovy USA', 'sl': 'Stranski zunanji otoki Združenih držav', 'smn': 'Ovtâstum Staatâi sierânâssuolluuh', 'so': 'Jasiiradaha ka baxsan Maraykanka', 'sq': 'Ishujt Periferikë të SHBA-së', 'sr': 'Удаљена острва САД', 'sr-Cyrl': 'Удаљена острва САД', 'sr-Latn': 'Udaljena ostrva SAD', 'sv': 'USA:s yttre öar', 'sw': 'Visiwa Vidogo vya Nje vya Marekani', 'ta': 'யூ.எஸ். வெளிப்புறத் தீவுகள்', 'te': 'సంయుక్త రాజ్య అమెరికా బయట ఉన్న దీవులు', 'tg': 'Ҷазираҳои Хурди Дурдасти ИМА', 'th': 'หมู่เกาะรอบนอกของสหรัฐอเมริกา', 'ti': 'ካብ ኣመሪካ ርሒቐን ንኣሽቱ ደሴታት', 'tk': 'ABŞ-nyň daşarky adalary', 'to': 'ʻOtumotu siʻi ʻo ʻAmelika', 'tr': 'ABD Küçük Harici Adaları', 'tt': 'АКШ Кече Читтәге утраулары', 'ug': 'ئا ق ش تاشقى ئاراللىرى', 'uk': 'Віддалені острови США', 'ur': 'امریکہ سے باہر کے چھوٹے جزائز', 'uz': 'AQSH yondosh orollari', 'uz-Cyrl': 'АҚШ ёндош ороллари', 'uz-Latn': 'AQSH yondosh orollari', 'vai': 'ꕶꕱ ꕪꘋ ꗅꘋ ꔳꘋꗣ ꖸ', 'vai-Vaii': 'ꕶꕱ ꕪꘋ ꗅꘋ ꔳꘋꗣ ꖸ', 'vec': 'Ìzołe perifèreghe meregane', 'vi': 'Các tiểu đảo xa của Hoa Kỳ', 'wae': 'Amerikaniš Ozeanie', 'wo': 'Duni Amerig Utar meer', 'xh': 'I-U.S. Outlying Islands', 'xnr': 'यू॰एस॰ आउटलाइंग द्वीपां दा झुंड', 'yo': 'Àwọn Erékùsù Kékèké Agbègbè US', 'yrl': 'Kapuãma Kuiriwaita Apekatu EUA suí', 'yue': '美國本土外小島嶼', 'yue-Hans': '美国本土外小岛屿', 'yue-Hant': '美國本土外小島嶼', 'zh': '美国本土外小岛屿', 'zh-Hans': '美国本土外小岛屿', 'zh-Hant': '美國本土外小島嶼', 'zu': 'I-U.S. Outlying Islands'},
    'UN': {'af': 'Verenigde Nasies', 'ak': 'Amansan Nkabomkuo', 'am': 'የተባበሩት መንግስታት', 'ar': 'الأمم المتحدة', 'as': 'ৰাষ্ট্ৰসংঘ', 'ast': 'Naciones Xuníes', 'az': 'Birləşmiş Millətlər Təşkilatı', 'az-Latn': 'Birləşmiş Millətlər Təşkilatı', 'bal-Latn': 'Myánostománi Gal', 'be': 'Арганізацыя Аб’яднаных Нацый', 'be-tarask': 'Арганізацыя Аб’яднаных Нацый', 'bg': 'Организация на обединените нации', 'blo': 'Ganɔ gaɖɔŋkɔnɔ kaBʊtǝna nɖulinya nɩ', 'bn': 'জাতিসংঘ', 'br': 'Broadoù unanet', 'brx': 'जथाय हादोरफोर', 'bs': 'Ujedinjene Nacije', 'bs-Cyrl': 'Уједињене нације', 'bs-Latn': 'Ujedinjene Nacije', 'ca': 'Nacions Unides', 'ca-ES-valencia': 'Nacions Unides', 'ccp': '𑄎𑄘𑄨𑄥𑄧𑄁𑄊𑄧', 'ce': 'Вовшахкхетта Къаьмнийн Организаци', 'ceb': 'Hiniusang Kanasoran', 'chr': 'ᏌᏊ ᎢᏳᎾᎵᏍᏔᏅ ᎠᏰᎵ ᏚᎾᏙᏢᏒ', 'ckb': 'نەتەوە یەکگرتووەکان', 'cs': 'Organizace spojených národů', 'csw': 'ᒫᒫᐏ ᐊᐢᑮᔭ', 'cv': 'Пӗрлешӳллӗ Нацисен Организацийӗ', 'cy': 'Y Cenhedloedd Unedig', 'da': 'De Forenede Nationer', 'de': 'Vereinte Nationen', 'dsb': 'Zjadnośone narody', 'el': 'Ηνωμένα Έθνη', 'el-polyton': 'Ηνωμένα Έθνη', 'en': 'United Nations', 'es': 'Naciones Unidas', 'et': 'Ühinenud Rahvaste Organisatsioon', 'eu': 'Nazio Batuak', 'fa': 'سازمان ملل متحد', 'ff-Adlm': '𞤑𞤢𞤱𞤼𞤢𞤤 𞤘𞤫𞤲𞤯𞤭', 'fi': 'Yhdistyneet kansakunnat', 'fil': 'United Nations', 'fo': 'Sameindu Tjóðir', 'fr': 'Nations Unies', 'ga': 'na Náisiúin Aontaithe', 'gd': 'Na Dùthchannan Aonaichte', 'gl': 'Nacións Unidas', 'gu': 'સંયુક્ત રાષ્ટ્ર', 'ha': 'Majalisar Ɗinkin Duniya', 'he': 'האומות המאוחדות', 'hi': 'संयुक्त राष्ट्र', 'hi-Latn': 'United Nations', 'hr': 'Ujedinjeni narodi', 'hsb': 'Zjednoćene narody', 'ht': 'Nations Unies', 'hu': 'Egyesült Nemzetek Szervezete', 'hy': 'Միավորված ազգերի կազմակերպություն', 'ia': 'Nationes Unite', 'id': 'Perserikatan Bangsa-Bangsa', 'ie': 'Unit Nationes', 'ig': 'Mba Ụwa Jikọrọ Ọnụ', 'is': 'Sameinuðu þjóðirnar', 'it': 'Nazioni Unite', 'ja': '国際連合', 'jv': 'Pasarékatan Bangsa-Bangsa', 'ka': 'გაერო', 'kaa': 'Бирлескен Миллетлер Шөлкеми', 'kaa-Cyrl': 'Бирлескен Миллетлер Шөлкеми', 'kea': 'Nasons Unidas', 'kgp': 'Nỹsãn Unĩna', 'kk': 'Біріккен Ұлттар Ұйымы', 'kk-Cyrl': 'Біріккен Ұлттар Ұйымы', 'km': 'អង្គការសហប្រជាជាតិ', 'kn': 'ಸಂಯುಕ್ತ ಸಂಸ್ಥಾನಗಳು', 'ko': '국제연합', 'kok': 'युनायटेड नेशन्स', 'kok-Deva': 'युनायटेड नेशन्स', 'ks': 'متحدہ مُمٲلک', 'ks-Arab': 'متحدہ مُمٲلک', 'ku': 'Miletên Yekbûyî', 'ky': 'БУ', 'lld': 'Naziuns Unides', 'lo': 'ສະຫະປະຊາຊາດ', 'lt': 'Jungtinės Tautos', 'lv': 'Apvienoto Nāciju Organizācija', 'mai': 'संयुक्त राष्ट्र', 'mi': 'Te Rūnanga Whakakotahi i ngā Iwi o te Ao', 'mk': 'Обединети нации', 'ml': 'ഐക്യരാഷ്ട്രസഭ', 'mn': 'Нэгдсэн Үндэстний Байгууллага', 'mr': 'संयुक्त राष्ट्र', 'ms': 'Bangsa-bangsa Bersatu', 'my': 'ကုလသမဂ္ဂ', 'nb': 'FN', 'ne': 'संयुक्त राष्ट्र संघ', 'nl': 'Verenigde Naties', 'nn': 'Dei sameinte nasjonane', 'no': 'FN', 'nqo': 'ߡߊ߲߬ߕߏ߲߫ ߠߊߘߍ߬ߣߍ߲ ߛߌ߬ߝߏ߲߬ߧߊ߬ߟߌ', 'om': 'Mootummoota Gamtooman', 'or': 'ଜାତିସଂଘ', 'pa': 'ਸੰਯੁਕਤ ਰਾਸ਼ਟਰ', 'pa-Guru': 'ਸੰਯੁਕਤ ਰਾਸ਼ਟਰ', 'pcm': 'Yunaítẹd Néshọns', 'pl': 'Organizacja Narodów Zjednoczonych', 'ps': 'ملگري ملتونه', 'pt': 'Nações Unidas', 'qu': 'Naciones Unidas', 'rm': 'Naziuns Unidas', 'ro': 'Națiunile Unite', 'ru': 'Организация Объединенных Наций', 'sc': 'Natziones Unidas', 'sd': 'گڏيل قومون', 'sd-Arab': 'گڏيل قومون', 'si': 'එක්සත් ජාතීන්', 'sk': 'Organizácia Spojených národov', 'sl': 'Združeni narodi', 'so': 'Qaramada Midoobay', 'sq': 'Organizata e Kombeve të Bashkuara', 'sr': 'Уједињене нације', 'sr-Cyrl': 'Уједињене нације', 'sr-Latn': 'Ujedinjene nacije', 'sv': 'Förenta Nationerna', 'sw': 'Umoja wa Mataifa', 'syr': 'ܐܡ̈ܘܬܐ ܡܚܝ̈ܕܬܐ', 'ta': 'ஐக்கிய நாடுகள்', 'te': 'యునైటెడ్ నేషన్స్', 'tg': 'Созмони Милали Муттаҳид', 'th': 'สหประชาชาติ', 'ti': 'ሕቡራት ሃገራት', 'tk': 'Birleşen Milletler Guramasy', 'to': 'ʻŪ fonua fakatahataha', 'tr': 'Birleşmiş Milletler', 'tt': 'Берләшкән Милләтләр', 'uk': 'Організація Об’єднаних Націй', 'ur': 'اقوام متحدہ', 'uz': 'Birlashgan Millatlar Tashkiloti', 'uz-Latn': 'Birlashgan Millatlar Tashkiloti', 'vec': 'Nasion Unìe', 'vi': 'Liên hợp quốc', 'wo': 'United Nations', 'xh': 'Izizwe Ezimanyeneyo', 'xnr': 'संयुक्त रास्टर', 'yo': 'Ìṣọ̀kan àgbáyé', 'yrl': 'Nasãu Yepewasuwaitá', 'yue': '聯合國', 'yue-Hans': '联合国', 'yue-Hant': '聯合國', 'zh': '联合国', 'zh-Hans': '联合国', 'zh-Hant': '聯合國', 'zu': 'I-United Nations'},
    'US': {'af': 'Verenigde State van Amerika', 'agq': 'USA', 'ak': 'Amɛrika', 'am': 'ዩናይትድ ስቴትስ', 'ar': 'الولايات المتحدة', 'as': 'মাৰ্কিন যুক্তৰাষ্ট্ৰ', 'asa': 'Marekani', 'ast': 'Estaos Xuníos', 'az': 'Amerika Birləşmiş Ştatları', 'az-Cyrl': 'Америка Бирләшмиш Штатлары', 'az-Latn': 'Amerika Birləşmiş Ştatları', 'bal-Latn': 'Amrikáay Tepákén Están', 'bas': 'Àdnà i Bilɔ̀ŋ bi Amerkà', 'be': 'Злучаныя Штаты Амерыкі', 'be-tarask': 'Злучаныя Штаты Амерыкі', 'bez': 'Humalekani', 'bg': 'Съединени щати', 'bgn': 'متحدین ایالات', 'blo': 'Ganɔ gaɖɔŋkɔnɔ kaBʊtǝna Amalɩka nɩ', 'bm': 'Ameriki', 'bn': 'মার্কিন যুক্তরাষ্ট্র', 'bo': 'ཨ་མེ་རི་ཀ།', 'br': 'Stadoù-Unanet', 'brx': 'जथाय हादोर आमेरिका', 'bs': 'Sjedinjene Države', 'bs-Cyrl': 'Сједињене Америчке Државе', 'bs-Latn': 'Sjedinjene Države', 'ca': 'Estats Units', 'ca-ES-valencia': 'Estats Units', 'ccp': '𑄟𑄢𑄴𑄇𑄨𑄚𑄴 𑄎𑄧𑄙𑄢𑄬𑄌𑄴𑄎𑄮', 'ce': 'Цхьанатоьхна Штаташ', 'ceb': 'Estados Unidos', 'cgg': 'Amerika', 'cho': 'United States', 'chr': 'ᏌᏊ ᎢᏳᎾᎵᏍᏔᏅ ᏍᎦᏚᎩ', 'cic': 'United States', 'ckb': 'ویلایەتە یەکگرتووەکان', 'cs': 'Spojené státy', 'csw': 'ᑭᐦᒋ ᒨᑯᒫᓂᕽ', 'cv': 'Пӗрлешӗннӗ Штатсем', 'cy': 'Yr Unol Daleithiau', 'da': 'USA', 'dav': 'Marekani', 'de': 'Vereinigte Staaten', 'dje': 'Ameriki Laabu Margantey', 'doi': 'यूएस', 'dsb': 'Zjadnośone staty Ameriki', 'dz': 'ཡུ་ཨེས་ཨེ', 'ebu': 'Marekani', 'ee': 'USA nutome', 'el': 'Ηνωμένες Πολιτείες', 'el-polyton': 'Ἡνωμένες Πολιτεῖες', 'en': 'United States', 'en-Dsrt': '𐐏𐐭𐑌𐐴𐐻𐐲𐐼 𐐝𐐻𐐩𐐻𐑅', 'eo': 'Usono', 'es': 'Estados Unidos', 'et': 'Ameerika Ühendriigid', 'eu': 'Ameriketako Estatu Batuak', 'ewo': 'Ǹnam Amɛrəkə', 'fa': 'ایالات متحده', 'ff': 'Dowlaaji Dentuɗi Amerik', 'ff-Adlm': '𞤁𞤫𞤲𞤼𞤢𞤤 𞤂𞤢𞤪𞤫', 'ff-Latn': 'Dowlaaji Dentuɗi Amerik', 'fi': 'Yhdysvallat', 'fil': 'Estados Unidos', 'fo': 'Sambandsríki Amerika', 'fr': 'États-Unis', 'fur': 'Stâts Unîts', 'fy': 'Ferienigde Staten', 'ga': 'Stáit Aontaithe Mheiriceá', 'gd': 'Na Stàitean Aonaichte', 'gl': 'Estados Unidos', 'gsw': 'Veräinigti Schtaate', 'gu': 'યુનાઇટેડ સ્ટેટ્સ', 'guz': 'Marekani', 'ha': 'Amurka', 'haw': 'ʻAmelika Hui Pū ʻIa', 'he': 'ארצות הברית', 'hi': 'संयुक्त राज्य', 'hi-Latn': 'United States', 'hnj': '𞄒𞄫𞄱𞄔𞄩𞄴', 'hnj-Hmnp': '𞄒𞄫𞄱𞄔𞄩𞄴', 'hr': 'Sjedinjene Američke Države', 'hsb': 'Zjednoćene staty Ameriki', 'ht': 'États-Unis', 'hu': 'Egyesült Államok', 'hy': 'Միացյալ Նահանգներ', 'ia': 'Statos Unite', 'id': 'Amerika Serikat', 'ig': 'United States', 'ii': 'ꂰꇩ', 'is': 'Bandaríkin', 'it': 'Stati Uniti', 'ja': 'アメリカ合衆国', 'jmc': 'Marekani', 'jv': 'Amérika Sarékat', 'ka': 'ამერიკის შეერთებული შტატები', 'kaa': 'Америка Қурама Штатлары', 'kaa-Cyrl': 'Америка Қурама Штатлары', 'kab': 'WDM', 'kam': 'Marekani', 'kde': 'Malekani', 'kea': 'Stadus Unidos di Merka', 'kgp': 'Enhtano Unĩno', 'khq': 'Ameriki Laabu Margantey', 'ki': 'Amerika', 'kk': 'Америка Құрама Штаттары', 'kk-Cyrl': 'Америка Құрама Штаттары', 'kln': 'Emetab amerika', 'km': 'សហរដ្ឋអាមេរិក', 'kn': 'ಯುನೈಟೆಡ್ ಸ್ಟೇಟ್ಸ್', 'ko': '미국', 'kok': 'युनायटेड स्टेट्स', 'kok-Deva': 'युनायटेड स्टेट्स', 'ks': 'یوٗنایٹِڑ سِٹیٹِس', 'ks-Arab': 'یوٗنایٹِڑ سِٹیٹِس', 'ks-Deva': 'मूतहीद रियासत', 'ksb': 'Malekani', 'ksf': 'amɛrika', 'ksh': 'de vereineschte Schtaate vun Ammärrika', 'ku': 'Dewletên Yekbûyî yên Amerîkayê', 'kxv': 'yunaiṭeḍ stets', 'kxv-Deva': 'आण्डि ति राज्यॉ', 'kxv-Latn': 'yunaiṭeḍ stets', 'kxv-Orya': 'ଆଣ୍ଡିତି ରାଜ୍ୟ', 'kxv-Telu': 'యునైటెడ్ స్టేట్స్', 'ky': 'Кошмо Штаттар', 'lag': 'Amerɨka', 'lb': 'Vereenegt Staaten', 'lg': 'Amerika', 'lkt': 'Mílahaŋska Tȟamákȟočhe', 'lld': 'Stac Unis', 'ln': 'Ameriki', 'lo': 'ສະຫະລັດ', 'lrc': 'ڤولاتیا یأکاگئرتە', 'lt': 'Jungtinės Valstijos', 'lu': 'Ameriki', 'luo': 'USA', 'luy': 'Marekani', 'lv': 'Amerikas Savienotās Valstis', 'mai': 'संयुक्त राज्य', 'mas': 'Marekani', 'mer': 'Amerika', 'mfe': 'Lamerik', 'mg': 'Etazonia', 'mgh': 'Umarekani', 'mi': 'Hononga o Amerika', 'mk': 'Соединети Американски Држави', 'ml': 'അമേരിക്കൻ ഐക്യനാടുകൾ', 'mn': 'Америкийн Нэгдсэн Улс', 'mni': 'য়ুনাইটেদ ষ্টেটস', 'mni-Beng': 'য়ুনাইটেদ ষ্টেটস', 'mr': 'युनायटेड स्टेट्स', 'ms': 'Amerika Syarikat', 'mt': 'l-Istati Uniti', 'mua': 'Amerika', 'mus': 'United States', 'my': 'အမေရိကန် ပြည်ထောင်စု', 'mzn': 'متحده ایالات', 'naq': 'Amerikab', 'nb': 'USA', 'nd': 'Amelika', 'ne': 'संयुक्त राज्य', 'nl': 'Verenigde Staten', 'nmg': 'Amɛŕka', 'nn': 'USA', 'no': 'USA', 'nqo': 'ߞߊ߬ߝߏ߫ ߘߍ߬ߣߍ߲', 'nyn': 'Amerika', 'om': 'Yiinaayitid Isteet', 'or': 'ଯୁକ୍ତ ରାଷ୍ଟ୍ର', 'os': 'АИШ', 'osa': 'United States', 'pa': 'ਸੰਯੁਕਤ ਰਾਜ', 'pa-Guru': 'ਸੰਯੁਕਤ ਰਾਜ', 'pcm': 'Yunaítẹ́d Stets', 'pl': 'Stany Zjednoczone', 'ps': 'متحده آيالات', 'pt': 'Estados Unidos', 'qu': 'Estados Unidos', 'rm': 'Stadis Unids da l’America', 'rn': 'Leta Zunze Ubumwe za Amerika', 'ro': 'Statele Unite ale Americii', 'rof': 'Marekani', 'ru': 'Соединенные Штаты', 'rwk': 'Marekani', 'sa': 'संयुक्त राज्य:', 'sah': 'Америка Холбоһуктаах Штааттара', 'saq': 'Marekani', 'sat': 'ᱭᱩᱱᱟᱭᱴᱮᱰ ᱮᱥᱴᱮᱴ', 'sat-Olck': 'ᱭᱩᱱᱟᱭᱴᱮᱰ ᱮᱥᱴᱮᱴ', 'sbp': 'Malekani', 'sc': 'Istados Unidos', 'sd': 'آمريڪا جون گڏيل رياستون', 'sd-Arab': 'آمريڪا جون گڏيل رياستون', 'sd-Deva': 'अमेरिका', 'se': 'Amerihká ovttastuvvan stáhtat', 'seh': 'Estados Unidos', 'ses': 'Ameriki Laabu Margantey', 'sg': 'ÂLeaa-Ôko tî Amerika', 'shi': 'ⵉⵡⵓⵏⴰⴽ ⵎⵓⵏⵏⵉⵏ ⵏ ⵎⵉⵔⵉⴽⴰⵏ', 'shi-Latn': 'iwunak munnin n mirikan', 'shi-Tfng': 'ⵉⵡⵓⵏⴰⴽ ⵎⵓⵏⵏⵉⵏ ⵏ ⵎⵉⵔⵉⴽⴰⵏ', 'si': 'එක්සත් ජනපදය', 'sk': 'Spojené štáty', 'sl': 'Združene države Amerike', 'smn': 'Ovtâstum Staatah', 'sn': 'Amerika', 'so': 'Maraykanka', 'sq': 'SHBA', 'sr': 'Сједињене Државе', 'sr-Cyrl': 'Сједињене Државе', 'sr-Latn': 'Sjedinjene Države', 'su': 'Amérika Sarikat', 'su-Latn': 'Amérika Sarikat', 'sv': 'USA', 'sw': 'Marekani', 'syr': 'ܐܘܚܕ̈ܢܐ ܡܚܝܕ̈ܐ', 'ta': 'அமெரிக்கா', 'te': 'యునైటెడ్ స్టేట్స్', 'teo': 'Marekani', 'tg': 'Иёлоти Муттаҳида', 'th': 'สหรัฐอเมริกา', 'ti': 'ኣመሪካ', 'tk': 'Amerikanyň Birleşen Ştatlary', 'to': 'Puleʻanga fakatahataha ʻAmelika', 'tr': 'Amerika Birleşik Devletleri', 'tt': 'АКШ', 'twq': 'Ameriki Laabu Margantey', 'tzm': 'Iwunak Idduklen n Amirika', 'ug': 'ئامېرىكا قوشما ئىشتاتلىرى', 'uk': 'Сполучені Штати', 'ur': 'ریاست ہائے متحدہ امریکہ', 'uz': 'Amerika Qo‘shma Shtatlari', 'uz-Cyrl': 'Америка Қўшма Штатлари', 'uz-Latn': 'Amerika Qo‘shma Shtatlari', 'vai': 'ꕶꕱ', 'vai-Latn': 'Poo', 'vai-Vaii': 'ꕶꕱ', 'vec': 'Stadi Unìi', 'vi': 'Hoa Kỳ', 'vun': 'Marekani', 'wae': 'Amerika', 'wo': 'Etaa Sini', 'xh': 'EMelika', 'xnr': 'संयुक्त राज्य', 'xog': 'Amerika', 'yav': 'amálíka', 'yi': 'פֿאַראייניגטע שטאַטן', 'yo': 'Amẹrikà', 'yrl': 'Tetãma-ita Yepewasú', 'yue': '美國', 'yue-Hans': '美国', 'yue-Hant': '美國', 'zgh': 'ⵉⵡⵓⵏⴰⴽ ⵎⵓⵏⵏⵉⵏ ⵏ ⵎⵉⵔⵉⴽⴰⵏ', 'zh': '美国', 'zh-Hans': '美国', 'zh-Hant': '美國', 'zu': 'i-United States'},
    'UY': {'af': 'Uruguay', 'agq': 'Yulùgwɛ̂', 'ak': 'Yurugwae', 'am': 'ኡራጓይ', 'ar': 'أورغواي', 'as': 'উৰুগুৱে', 'asa': 'Urugwai', 'ast': 'Uruguái', 'az': 'Uruqvay', 'az-Cyrl': 'Уругвај', 'az-Latn': 'Uruqvay', 'bal-Latn': 'Yurógóay', 'bas': 'Ùrùgwêy', 'be': 'Уругвай', 'be-tarask': 'Уругвай', 'bez': 'Huulugwai', 'bg': 'Уругвай', 'bgn': 'اوراگوی', 'blo': 'Uruguwee', 'bm': 'Urugwayi', 'bn': 'উরুগুয়ে', 'br': 'Uruguay', 'brx': 'ऊरुगुवे', 'bs': 'Urugvaj', 'bs-Cyrl': 'Уругвај', 'bs-Latn': 'Urugvaj', 'ca': 'Uruguai', 'ca-ES-valencia': 'Uruguai', 'ccp': '𑄅𑄪𑄢𑄪𑄉𑄪𑄠𑄬', 'ce': 'Уругвай', 'ceb': 'Uruguay', 'cgg': 'Urugwai', 'chr': 'ᏳᎷᏇ', 'ckb': 'ئوروگوای', 'cs': 'Uruguay', 'cv': 'Уругвай', 'cy': 'Uruguay', 'da': 'Uruguay', 'dav': 'Urugwai', 'de': 'Uruguay', 'dje': 'Uruguwey', 'dsb': 'Uruguay', 'dz': 'ཡུ་རུ་གུ་ཝའི', 'ebu': 'Urugwai', 'ee': 'uruguaydukɔ', 'el': 'Ουρουγουάη', 'el-polyton': 'Οὐρουγουάη', 'en': 'Uruguay', 'en-Dsrt': '𐐏𐐳𐑉𐐲𐑀𐐶𐐴', 'eo': 'Urugvajo', 'es': 'Uruguay', 'et': 'Uruguay', 'eu': 'Uruguai', 'ewo': 'Urugué', 'fa': 'اروگوئه', 'ff': 'Uruguwaay', 'ff-Adlm': '𞤒𞤵𞤪𞤺𞤮𞤴', 'ff-Latn': 'Uruguwaay', 'fi': 'Uruguay', 'fil': 'Uruguay', 'fo': 'Uruguai', 'fr': 'Uruguay', 'fur': 'Uruguay', 'fy': 'Uruguay', 'ga': 'Uragua', 'gd': 'Uruguaidh', 'gl': 'Uruguai', 'gsw': 'Uruguay', 'gu': 'ઉરુગ્વે', 'guz': 'Urugwai', 'ha': 'Yurigwai', 'he': 'אורוגוואי', 'hi': 'उरूग्वे', 'hi-Latn': 'Uruguay', 'hr': 'Urugvaj', 'hsb': 'Uruguay', 'ht': 'Uruguay', 'hu': 'Uruguay', 'hy': 'Ուրուգվայ', 'ia': 'Uruguay', 'id': 'Uruguay', 'ig': 'Uruguay', 'is': 'Úrúgvæ', 'it': 'Uruguay', 'ja': 'ウルグアイ', 'jmc': 'Urugwai', 'jv': 'Uruguay', 'ka': 'ურუგვაი', 'kaa': 'Уругвай', 'kaa-Cyrl': 'Уругвай', 'kab': 'Urugway', 'kam': 'Urugwai', 'kde': 'Ulugwai', 'kea': 'Uruguai', 'kgp': 'Urugvaj', 'khq': 'Uruguwey', 'ki': 'Urugwai', 'kk': 'Уругвай', 'kk-Cyrl': 'Уругвай', 'kln': 'Emetab Uruguay', 'km': 'អ៊ុយរូហ្គាយ', 'kn': 'ಉರುಗ್ವೆ', 'ko': '우루과이', 'kok': 'उरूग्वे', 'kok-Deva': 'उरूग्वे', 'ks': 'یوٗروگے', 'ks-Arab': 'یوٗروگے', 'ksb': 'Ulugwai', 'ksf': 'urugwɛ́', 'ksh': 'Urrujwai', 'ku': 'Ûrûguay', 'kxv': 'ūrugve', 'kxv-Deva': 'उरूग्वे', 'kxv-Latn': 'ūrugve', 'kxv-Orya': 'ଉରୁଗ୍ୱେ', 'kxv-Telu': 'ఉరుగ్వే', 'ky': 'Уругвай', 'lag': 'Uruguáai', 'lb': 'Uruguay', 'lg': 'Wurugwayi', 'lld': 'Uruguay', 'ln': 'Irigwei', 'lo': 'ອູຣຸກວຍ', 'lt': 'Urugvajus', 'lu': 'Irigwei', 'luo': 'Uruguay', 'luy': 'Urugwai', 'lv': 'Urugvaja', 'mai': 'उरुग्वे', 'mas': 'Urugwai', 'mer': 'Urugwĩ', 'mfe': 'Uruguay', 'mg': 'Orogoay', 'mi': 'Urukoi', 'mk': 'Уругвај', 'ml': 'ഉറുഗ്വേ', 'mn': 'Уругвай', 'mr': 'उरुग्वे', 'ms': 'Uruguay', 'mt': 'l-Urugwaj', 'mua': 'Urugwai', 'my': 'ဥရုဂွေး', 'mzn': 'اروگوئه', 'naq': 'Uruguaib', 'nb': 'Uruguay', 'nd': 'Yurugwai', 'ne': 'उरूग्वे', 'nl': 'Uruguay', 'nmg': 'Uruguay', 'nn': 'Uruguay', 'no': 'Uruguay', 'nqo': 'ߎ߳ߙߑߜ߭ߋߦߌ߫', 'nyn': 'Urugwai', 'om': 'Yuraagaay', 'or': 'ଉରୁଗୁଏ', 'pa': 'ਉਰੂਗਵੇ', 'pa-Guru': 'ਉਰੂਗਵੇ', 'pcm': 'Yúrugwue', 'pl': 'Urugwaj', 'ps': 'یوروګوی', 'pt': 'Uruguai', 'qu': 'Uruguay', 'rm': 'Uruguay', 'rn': 'Irigwe', 'ro': 'Uruguay', 'rof': 'Urugwai', 'ru': 'Уругвай', 'rwk': 'Urugwai', 'saq': 'Urugwai', 'sbp': 'Ulugwayi', 'sc': 'Uruguày', 'sd': 'يوروگوءِ', 'sd-Arab': 'يوروگوءِ', 'se': 'Uruguay', 'seh': 'Uruguai', 'ses': 'Uruguwey', 'sg': 'Uruguëe', 'shi': 'ⵓⵔⵓⴳⵡⴰⵢ', 'shi-Latn': 'urugway', 'shi-Tfng': 'ⵓⵔⵓⴳⵡⴰⵢ', 'si': 'උරුගුවේ', 'sk': 'Uruguaj', 'sl': 'Urugvaj', 'smn': 'Uruguay', 'sn': 'Uruguay', 'so': 'Uruguwaay', 'sq': 'Uruguai', 'sr': 'Уругвај', 'sr-Cyrl': 'Уругвај', 'sr-Latn': 'Urugvaj', 'sv': 'Uruguay', 'sw': 'Uruguay', 'syr': 'ܐܘܪܘܓܘܐܝ', 'ta': 'உருகுவே', 'te': 'ఉరుగ్వే', 'teo': 'Urugwai', 'tg': 'Уругвай', 'th': 'อุรุกวัย', 'ti': 'ኡራጓይ', 'tk': 'Urugwaý', 'to': 'ʻUlukuai', 'tr': 'Uruguay', 'tt': 'Уругвай', 'twq': 'Uruguwey', 'tzm': 'Urugway', 'ug': 'ئۇرۇگۋاي', 'uk': 'Уругвай', 'ur': 'یوروگوئے', 'uz': 'Urugvay', 'uz-Cyrl': 'Уругвай', 'uz-Latn': 'Urugvay', 'vai': 'ꖳꔓꗝꔀ', 'vai-Latn': 'Yuwegɔwe', 'vai-Vaii': 'ꖳꔓꗝꔀ', 'vec': 'Uruguài', 'vi': 'Uruguay', 'vun': 'Urugwai', 'wae': 'Urugauy', 'wo': 'Uruge', 'xh': 'E-Uruguay', 'xnr': 'उरूग्वे', 'xog': 'Wurugwayi', 'yav': 'ulukuéy', 'yi': 'אורוגוויי', 'yo': 'Úrúgúwè', 'yrl': 'Uruguwai', 'yue': '烏拉圭', 'yue-Hans': '乌拉圭', 'yue-Hant': '烏拉圭', 'zgh': 'ⵓⵔⵓⴳⵡⴰⵢ', 'zh': '乌拉圭', 'zh-Hans': '乌拉圭', 'zh-Hant': '烏拉圭', 'zu': 'i-Uruguay'},
    'UZ': {'af': 'Oesbekistan', 'agq': 'Yùzɨ̀bɛkìsɨ̀tân', 'ak': 'Usbɛkistan', 'am': 'ኡዝቤኪስታን', 'ar': 'أوزبكستان', 'as': 'উজবেকিস্তান', 'asa': 'Udhibekithtani', 'ast': 'Uzbequistán', 'az': 'Özbəkistan', 'az-Cyrl': 'Өзбәкистан', 'az-Latn': 'Özbəkistan', 'bal-Latn': 'Ozbekestán', 'bas': 'Ùzbèkìstân', 'be': 'Узбекістан', 'be-tarask': 'Узбекістан', 'bez': 'Huuzibekistani', 'bg': 'Узбекистан', 'bgn': 'اوزبکیستان', 'blo': 'Usbeekistan', 'bm': 'Uzebekisitani', 'bn': 'উজবেকিস্তান', 'br': 'Ouzbekistan', 'brx': 'उजबेकिस्तान', 'bs': 'Uzbekistan', 'bs-Cyrl': 'Узбекистан', 'bs-Latn': 'Uzbekistan', 'ca': 'Uzbekistan', 'ca-ES-valencia': 'Uzbekistan', 'ccp': '𑄅𑄪𑄎𑄴𑄝𑄬𑄇𑄨𑄌𑄴𑄖𑄚𑄴', 'ce': 'Узбекистан', 'ceb': 'Uzbekistan', 'cgg': 'Uzibekisitani', 'chr': 'ᎤᏍᏇᎩᏍᏖᏂ', 'ckb': 'ئوزبەکستان', 'cs': 'Uzbekistán', 'cv': 'Узбекистан', 'cy': 'Uzbekistan', 'da': 'Usbekistan', 'dav': 'Uzibekistani', 'de': 'Usbekistan', 'dje': 'Uzbeekistan', 'dsb': 'Uzbekistan', 'dz': 'ཨུས་བེག་གི་སཏཱན', 'ebu': 'Uzibekistani', 'ee': 'Uzbekistan nutome', 'el': 'Ουζμπεκιστάν', 'el-polyton': 'Οὐζμπεκιστάν', 'en': 'Uzbekistan', 'en-Dsrt': '𐐅𐑆𐐺𐐯𐐿𐐲𐑅𐐻𐐰𐑌', 'eo': 'Uzbekujo', 'es': 'Uzbekistán', 'et': 'Usbekistan', 'eu': 'Uzbekistan', 'ewo': 'Uzubekisətán', 'fa': 'ازبکستان', 'ff': 'Usbekistaan', 'ff-Adlm': '𞤓𞥁𞤦𞤫𞤳𞤭𞤧𞤼𞤢𞥄𞤲', 'ff-Latn': 'Usbekistaan', 'fi': 'Uzbekistan', 'fil': 'Uzbekistan', 'fo': 'Usbekistan', 'fr': 'Ouzbékistan', 'fur': 'Uzbechistan', 'fy': 'Oezbekistan', 'ga': 'an Úisbéiceastáin', 'gd': 'Usbagastàn', 'gl': 'Uzbekistán', 'gsw': 'Uschbeekischtan', 'gu': 'ઉઝ્બેકિસ્તાન', 'guz': 'Uzibekistani', 'ha': 'Uzubekistan', 'he': 'אוזבקיסטן', 'hi': 'उज़्बेकिस्तान', 'hi-Latn': 'Uzbekistan', 'hr': 'Uzbekistan', 'hsb': 'Uzbekistan', 'ht': 'Ouzbékistan', 'hu': 'Üzbegisztán', 'hy': 'Ուզբեկստան', 'ia': 'Uzbekistan', 'id': 'Uzbekistan', 'ig': 'Uzbekistan', 'is': 'Úsbekistan', 'it': 'Uzbekistan', 'ja': 'ウズベキスタン', 'jmc': 'Uzibekistani', 'jv': 'Usbèkistan', 'ka': 'უზბეკეთი', 'kaa': 'Өзбекстан', 'kaa-Cyrl': 'Өзбекстан', 'kab': 'Uzbaxistan', 'kam': 'Uzibekistani', 'kde': 'Uzibechistani', 'kea': 'Uzbekistan', 'kgp': 'Unhmekinhtỹv', 'khq': 'Uzbeekistan', 'ki': 'Uzibekistani', 'kk': 'Өзбекстан', 'kk-Cyrl': 'Өзбекстан', 'kln': 'Emetab Uzibekistani', 'km': 'អ៊ូសបេគីស្ថាន', 'kn': 'ಉಜ್ಬೇಕಿಸ್ಥಾನ್', 'ko': '우즈베키스탄', 'kok': 'उझ्बेकिस्तान', 'kok-Deva': 'उझ्बेकिस्तान', 'ks': 'اُزبِکِستان', 'ks-Arab': 'اُزبِکِستان', 'ksb': 'Uzibekistani', 'ksf': 'usbɛkistaŋ', 'ksh': 'Ußbeekistahn', 'ku': 'Ozbekistan', 'kxv': 'ūjbekistaan', 'kxv-Deva': 'उज़्बेकिस्तान', 'kxv-Latn': 'ūjbekistaan', 'kxv-Orya': 'ଉଜବେକିସ୍ତାନ', 'kxv-Telu': 'ఉజ్బెకిస్తాన్', 'ky': 'Өзбекстан', 'lag': 'Usibekisitáani', 'lb': 'Usbekistan', 'lg': 'Wuzibekisitaani', 'lld': 'Uzbekistan', 'ln': 'Uzibɛkisitá', 'lo': 'ອຸສເບກິສະຖານ', 'lt': 'Uzbekistanas', 'lu': 'Uzibekisita', 'luo': 'Uzbekistan', 'luy': 'Uzibekistani', 'lv': 'Uzbekistāna', 'mai': 'उजबेकिस्तान', 'mas': 'Uzibekistani', 'mer': 'Uzibekistani', 'mfe': 'Ouzbekistan', 'mg': 'Ozbekistan', 'mi': 'Uhipeketāne', 'mk': 'Узбекистан', 'ml': 'ഉസ്\u200cബെക്കിസ്ഥാൻ', 'mn': 'Узбекистан', 'mr': 'उझबेकिस्तान', 'ms': 'Uzbekistan', 'mt': 'l-Użbekistan', 'mua': 'Uzbekistaŋ', 'my': 'ဥဇဘက်ကစ္စတန်', 'mzn': 'ازبکستون', 'naq': 'Uzbekistan', 'nb': 'Usbekistan', 'nd': 'Uzbekistan', 'ne': 'उज्बेकिस्तान', 'nl': 'Oezbekistan', 'nmg': 'Usbǝkistaŋ', 'nn': 'Usbekistan', 'no': 'Usbekistan', 'nqo': 'ߎߖ߭ߑߓߋߞߌߛߑߕߊ߲߫', 'nyn': 'Uzibekisitani', 'om': 'Uzbeekistaan', 'or': 'ଉଜବେକିସ୍ତାନ', 'pa': 'ਉਜ਼ਬੇਕਿਸਤਾਨ', 'pa-Guru': 'ਉਜ਼ਬੇਕਿਸਤਾਨ', 'pcm': 'Uzbẹ́kistan', 'pl': 'Uzbekistan', 'ps': 'اوزبکستان', 'pt': 'Uzbequistão', 'qu': 'Uzbekistán', 'rm': 'Usbekistan', 'rn': 'Uzubekisitani', 'ro': 'Uzbekistan', 'rof': 'Uzibekistani', 'ru': 'Узбекистан', 'rwk': 'Uzibekistani', 'saq': 'Uzibekistani', 'sbp': 'Usibekisitani', 'sc': 'Uzbèkistan', 'sd': 'ازبڪستان', 'sd-Arab': 'ازبڪستان', 'se': 'Usbekistan', 'seh': 'Uzbequistão', 'ses': 'Uzbeekistan', 'sg': 'Uzbekistäan', 'shi': 'ⵓⵣⴱⴰⴽⵉⵙⵜⴰⵏ', 'shi-Latn': 'uzbakistan', 'shi-Tfng': 'ⵓⵣⴱⴰⴽⵉⵙⵜⴰⵏ', 'si': 'උස්බෙකිස්ථානය', 'sk': 'Uzbekistan', 'sl': 'Uzbekistan', 'smn': 'Uzbekistan', 'sn': 'Uzbekistan', 'so': 'Usbakistan', 'sq': 'Uzbekistan', 'sr': 'Узбекистан', 'sr-Cyrl': 'Узбекистан', 'sr-Latn': 'Uzbekistan', 'sv': 'Uzbekistan', 'sw': 'Uzibekistani', 'syr': 'ܐܘܙܒܟܣܬܐܢ', 'ta': 'உஸ்பெகிஸ்தான்', 'te': 'ఉజ్బెకిస్తాన్', 'teo': 'Uzibekistani', 'tg': 'Ӯзбекистон', 'th': 'อุซเบกิสถาน', 'ti': 'ኡዝበኪስታን', 'tk': 'Özbegistan', 'to': 'ʻUsipekitani', 'tok': 'ma Opekisan', 'tr': 'Özbekistan', 'tt': 'Үзбәкстан', 'twq': 'Uzbeekistan', 'tzm': 'Uzbakistan', 'ug': 'ئۆزبېكىستان', 'uk': 'Узбекистан', 'ur': 'ازبکستان', 'uz': 'Oʻzbekiston', 'uz-Cyrl': 'Ўзбекистон', 'uz-Latn': 'Oʻzbekiston', 'vai': 'ꖳꗩꕃꔻꕚꘋ', 'vai-Latn': 'Yubhɛkisitaŋ', 'vai-Vaii': 'ꖳꗩꕃꔻꕚꘋ', 'vec': 'Uzbekistan', 'vi': 'Uzbekistan', 'vun': 'Uzibekistani', 'wae': 'Usbekištan', 'wo': 'Usbekistaŋ', 'xh': 'E-Uzbekistan', 'xnr': 'उज़्बेकिस्तान', 'xog': 'Wuzibekisitaani', 'yav': 'usupekistáaŋ', 'yo': 'Nṣibẹkisitani', 'yrl': 'Yũbuesara-retãma', 'yue': '烏茲別克', 'yue-Hans': '乌兹别克', 'yue-Hant': '烏茲別克', 'zgh': 'ⵓⵣⴱⴰⴽⵉⵙⵜⴰⵏ', 'zh': '乌兹别克斯坦', 'zh-Hans': '乌兹别克斯坦', 'zh-Hant': '烏茲別克', 'zu': 'i-Uzbekistan'},
    'Ugar': {'am': 'ኡጋሪቲክ', 'ar': 'الأجاريتيكية', 'ast': 'ugaríticu', 'az': 'uqarit', 'az-Latn': 'uqarit', 'bal-Latn': 'Yugariti', 'bg': 'Угаритска', 'bn': 'উগারিটিক', 'br': 'ougaritek', 'brx': 'ऊगारीटीक', 'bs': 'ugaritsko pismo', 'bs-Cyrl': 'угаритско писмо', 'bs-Latn': 'ugaritsko pismo', 'ca': 'ugarític', 'ca-ES-valencia': 'ugarític', 'ccp': '𑄅𑄪𑄉𑄢𑄨𑄑𑄨𑄇𑄴', 'cs': 'ugaritské klínové', 'da': 'ugaritisk', 'de': 'Ugaritisch', 'el': 'Ουγκαριτικό', 'el-polyton': 'Ουγκαριτικό', 'en': 'Ugaritic', 'en-Dsrt': '𐐏𐐭𐑀𐐲𐑉𐐮𐐻𐐮𐐿', 'es': 'ugarítico', 'et': 'ugariti', 'eu': 'ugaritiera', 'fa': 'اوگاریتی', 'ff-Adlm': '𞤓𞤺𞤢𞤪𞤭𞤼𞤭𞤳', 'fi': 'ugaritilainen', 'fo': 'ugaritic', 'fr': 'ougaritique', 'fur': 'ugaritic', 'fy': 'Ugaritysk', 'ga': 'Úgairíteach', 'gd': 'Ugariticeach', 'gsw': 'Ugaritisch', 'gu': 'યુગાતિટિક', 'he': 'אוגריתי', 'hi': 'युगारिटिक', 'hi-Latn': 'Ugaritic', 'hr': 'ugaritsko pismo', 'ht': 'ougaritique', 'hu': 'Ugari', 'id': 'Ugaritik', 'ig': 'Ugaritic', 'it': 'ugarita', 'ja': 'ウガリット文字', 'ka': 'უგარითული', 'kgp': 'ugaritiku', 'kk': 'угарит жазуы', 'kk-Cyrl': 'угарит жазуы', 'kn': 'ಉಗಾರಿಟಿಕ್', 'ko': '우가리트 문자', 'ks': 'اُگارِٹِک', 'ks-Arab': 'اُگارِٹِک', 'lb': 'Ugaritesch', 'lo': 'ຍູກາຮິດ', 'lt': 'ugaritik', 'mk': 'угаритско', 'ml': 'ഉഗ്രൈറ്റിക്', 'mn': 'Угаритик', 'mr': 'युगारिटिक', 'ms': 'Ugaritic', 'nb': 'ugaritisk', 'ne': 'युगारिटिक', 'nl': 'Ugaritisch', 'nn': 'ugaritisk', 'no': 'ugaritisk', 'nqo': 'ߎߜ߭ߊߙߕߌߞ', 'or': 'ୟୁଗାରିଟିକ୍', 'pl': 'ugaryckie', 'pt': 'ugarítico', 'rm': 'ugaritic', 'ro': 'ugaritică', 'ru': 'угаритская', 'sc': 'ugarìticu', 'sl': 'ugaritski', 'so': 'Luuqada Ugaritic', 'sq': 'ugaritik', 'sr': 'угаритско писмо', 'sr-Cyrl': 'угаритско писмо', 'sr-Latn': 'ugaritsko pismo', 'sv': 'ugaritiska', 'ta': 'உகாரதிக்', 'te': 'యుగారిటిక్', 'tg': 'Угаритӣ', 'th': 'ยูการิต', 'to': 'tohinima fakaʻūkaliti', 'tr': 'Ugarit Çivi Yazısı', 'ug': 'ئۇگارىتىكچە', 'uk': 'угаритський', 'vec': 'ugarìtego', 'vi': 'Chữ Ugarit', 'xnr': 'युगारिटिक', 'yrl': 'ugarítiku', 'yue': '烏加列文', 'yue-Hans': '乌加列文', 'yue-Hant': '烏加列文', 'zh': '乌加里特文', 'zh-Hans': '乌加里特文', 'zh-Hant': '烏加列文', 'zu': 'i-Ugaritic'},
    'VA': {'af': 'Vatikaanstad', 'agq': 'Vatikàn Sɨ̀tɛ̂', 'ak': 'Vatican Man', 'am': 'ቫቲካን ከተማ', 'ar': 'الفاتيكان', 'as': 'ভেটিকান চিটি', 'asa': 'Vatikani', 'ast': 'Ciudá del Vaticanu', 'az': 'Vatikan', 'az-Cyrl': 'Ватикан', 'az-Latn': 'Vatikan', 'bal-Latn': 'Wátikán Sethi', 'bas': 'Vàtìkâŋ', 'be': 'Ватыкан', 'be-tarask': 'Ватыкан', 'bez': 'Huvatikani', 'bg': 'Ватикан', 'blo': 'Fatikaŋ kaMpá', 'bm': 'Vatikaŋ', 'bn': 'ভ্যাটিকান সিটি', 'br': 'Vatikan', 'brx': 'भेटिकान नोगोर', 'bs': 'Vatikan', 'bs-Cyrl': 'Ватикан', 'bs-Latn': 'Vatikan', 'ca': 'Ciutat del Vaticà', 'ca-ES-valencia': 'Ciutat del Vaticà', 'ccp': '𑄞𑄳𑄠𑄑𑄨𑄇𑄚𑄴 𑄥𑄨𑄑𑄨', 'ce': 'Ватикан', 'ceb': 'Vatican City', 'cgg': 'Vatikani', 'chr': 'ᎠᏥᎳᏁᏠ ᎦᏚᎲ', 'ckb': 'ڤاتیکان', 'cs': 'Vatikán', 'cv': 'Ватикан', 'cy': 'Y Fatican', 'da': 'Vatikanstaten', 'dav': 'Vatikani', 'de': 'Vatikanstadt', 'dje': 'Vaatikan Laama', 'dsb': 'Vatikańske město', 'dz': 'བ་ཊི་ཀཱན་ སི་ཊི', 'ebu': 'Vatikani', 'ee': 'Vatikandu nutome', 'el': 'Βατικανό', 'el-polyton': 'Ἁγία Ἕδρα (Βατικανό)', 'en': 'Vatican City', 'en-Dsrt': '𐐚𐐰𐐼𐐲𐐿𐐲𐑌', 'eo': 'Vatikano', 'es': 'Ciudad del Vaticano', 'et': 'Vatikan', 'eu': 'Vatikano Hiria', 'ewo': 'Ǹnam Vatikán', 'fa': 'واتیکان', 'ff': 'Dowla Waticaan', 'ff-Adlm': '𞤜𞤢𞤼𞤭𞤳𞤢𞥄𞤲', 'ff-Latn': 'Dowla Waticaan', 'fi': 'Vatikaani', 'fil': 'Vatican City', 'fo': 'Vatikanbýur', 'fr': 'État de la Cité du Vatican', 'fur': 'Vatican', 'fy': 'Vaticaanstêd', 'ga': 'Cathair na Vatacáine', 'gd': 'Cathair na Bhatacain', 'gl': 'Cidade do Vaticano', 'gsw': 'Vatikanstadt', 'gu': 'વેટિકન સિટી', 'guz': 'Vatikani', 'ha': 'Birnin Batikan', 'he': 'הוותיקן', 'hi': 'वेटिकन सिटी', 'hi-Latn': 'Vatican City', 'hr': 'Vatikan', 'hsb': 'Vatikanske město', 'ht': 'État de la Cité du Vatican', 'hu': 'Vatikán', 'hy': 'Վատիկան', 'ia': 'Citate del Vaticano', 'id': 'Vatikan', 'ig': 'Vatican City', 'is': 'Vatíkanið', 'it': 'Città del Vaticano', 'ja': 'バチカン市国', 'jmc': 'Vatikani', 'jv': 'Kutha Vatikan', 'ka': 'ქალაქი ვატიკანი', 'kaa': 'Ватикан', 'kaa-Cyrl': 'Ватикан', 'kab': 'Awanek n Vatikan', 'kam': 'Vatikani', 'kde': 'Vatikani', 'kea': 'Vatikanu', 'kgp': 'Vatikỹnũ Emã-mág', 'khq': 'Vaatikan Laama', 'ki': 'Vatikani', 'kk': 'Ватикан', 'kk-Cyrl': 'Ватикан', 'kln': 'Emetab Vatican', 'km': 'បុរី\u200bវ៉ាទីកង់', 'kn': 'ವ್ಯಾಟಿಕನ್ ಸಿಟಿ', 'ko': '바티칸 시국', 'kok': 'वॅटिकन सिटी', 'kok-Deva': 'वॅटिकन सिटी', 'ks': 'ویٹِکَن سِٹی', 'ks-Arab': 'ویٹِکَن سِٹی', 'ksb': 'Vatikani', 'ksf': 'watikáŋ', 'ksh': 'der Vattikahn', 'ku': 'Vatîkan', 'kxv': 'bāṭikān', 'kxv-Deva': 'बाटिकान सिटी', 'kxv-Latn': 'bāṭikān', 'kxv-Orya': 'ବାଟିକାନ୍ ସିଟି', 'kxv-Telu': 'బాటికాన్ సిటి', 'ky': 'Ватикан', 'lag': 'Vatikáani', 'lb': 'Vatikanstad', 'lg': 'Vatikaani', 'lld': 'Cité dl Vatican', 'ln': 'Vatiká', 'lo': 'ນະຄອນ ວາຕິກັນ', 'lt': 'Vatikano Miesto Valstybė', 'lu': 'Nvatika', 'luo': 'Vatican State', 'luy': 'Vatikani', 'lv': 'Vatikāns', 'mai': 'वेटिकन सिटी', 'mas': 'Vatikani', 'mer': 'Vatikani', 'mfe': 'Lata Vatikan', 'mg': 'Firenen’i Vatikana', 'mgh': 'Uvatikani', 'mi': 'Te Poho-o-Pita', 'mk': 'Ватикан', 'ml': 'വത്തിക്കാൻ', 'mn': 'Ватикан хот улс', 'mr': 'व्हॅटिकन सिटी', 'ms': 'Kota Vatican', 'mt': 'l-Istat tal-Belt tal-Vatikan', 'mua': 'Vaticaŋ', 'my': 'ဗာတီကန်စီးတီး', 'mzn': 'واتیکان', 'naq': 'Vatican State', 'nb': 'Vatikanstaten', 'nd': 'Vatican State', 'ne': 'भेटिकन सिटी', 'nl': 'Vaticaanstad', 'nmg': 'Vatikaŋ', 'nn': 'Vatikanstaten', 'no': 'Vatikanstaten', 'nqo': 'ߝ߭ߊߕߌߞߊ߲߫ ߞߊ߬ߝߏ', 'nyn': 'Vatikani', 'om': 'Vaatikaan Siitii', 'or': 'ଭାଟିକାନ୍ ସିଟି', 'pa': 'ਵੈਟੀਕਨ ਸਿਟੀ', 'pa-Guru': 'ਵੈਟੀਕਨ ਸਿਟੀ', 'pcm': 'Vátíkán Síti', 'pl': 'Watykan', 'ps': 'واتیکان ښار', 'pt': 'Cidade do Vaticano', 'qu': 'Santa Sede (Ciudad del Vaticano)', 'rm': 'Citad dal Vatican', 'rn': 'Umurwa wa Vatikani', 'ro': 'Statul Cetății Vaticanului', 'rof': 'Vatikani', 'ru': 'Ватикан', 'rwk': 'Vatikani', 'saq': 'Vatikani', 'sbp': 'Vatikani', 'sc': 'Tzitade de su Vaticanu', 'sd': 'ويٽڪين سٽي', 'sd-Arab': 'ويٽڪين سٽي', 'se': 'Vatikána', 'seh': 'Vaticano', 'ses': 'Vaatikan Laama', 'sg': 'Letëe tî Vatikäan', 'shi': 'ⴰⵡⴰⵏⴽ ⵏ ⴼⴰⵜⵉⴽⴰⵏ', 'shi-Latn': 'awank n fatikan', 'shi-Tfng': 'ⴰⵡⴰⵏⴽ ⵏ ⴼⴰⵜⵉⴽⴰⵏ', 'si': 'වතිකානු නගරය', 'sk': 'Vatikán', 'sl': 'Vatikan', 'smn': 'Vatikan', 'sn': 'Vatican State', 'so': 'Magaalada Faatikaan', 'sq': 'Vatikan', 'sr': 'Ватикан', 'sr-Cyrl': 'Ватикан', 'sr-Latn': 'Vatikan', 'sv': 'Vatikanstaten', 'sw': 'Mji wa Vatican', 'syr': 'ܡܕܝܢܬܐ ܕܘܛܝܩܢ', 'ta': 'வாடிகன் நகரம்', 'te': 'వాటికన్ నగరం', 'teo': 'Vatikani', 'tg': 'Шаҳри Вотикон', 'th': 'นครวาติกัน', 'ti': 'ከተማ ቫቲካን', 'tk': 'Watikan', 'to': 'Kolo Vatikani', 'tr': 'Vatikan', 'tt': 'Ватикан', 'twq': 'Vaatikan Laama', 'tzm': 'Awank iɣrem n Vatikan', 'ug': 'ۋاتىكان', 'uk': 'Ватикан', 'ur': 'ویٹیکن سٹی', 'uz': 'Vatikan', 'uz-Cyrl': 'Ватикан', 'uz-Latn': 'Vatikan', 'vai': 'ꕙꔳꕪꘋ ꕢꕨꕌ', 'vai-Latn': 'Vatikaŋ Ɓoloe', 'vai-Vaii': 'ꕙꔳꕪꘋ ꕢꕨꕌ', 'vec': 'Vategan', 'vi': 'Thành Vatican', 'vun': 'Vatikani', 'wae': 'Vatikan', 'wo': 'Site bu Watikaa', 'xh': 'EVatican City', 'xnr': 'वेटिकन सिटी', 'xog': 'Vatikaani', 'yav': 'fatikáaŋ', 'yi': 'וואַטיקאַן שטאָט', 'yo': 'Ìlú Vatican', 'yrl': 'Watikanu Tawa-wasu', 'yue': '梵蒂岡', 'yue-Hans': '梵蒂冈', 'yue-Hant': '梵蒂岡', 'zgh': 'ⴰⵡⴰⵏⴽ ⵏ ⴼⴰⵜⵉⴽⴰⵏ', 'zh': '梵蒂冈', 'zh-Hans': '梵蒂冈', 'zh-Hant': '梵蒂岡', 'zu': 'i-Vatican City'},
    'VC': {'af': 'Sint Vincent en die Grenadine', 'agq': 'Sɛ̀n Vinsɨ̀n à Gɨlenadi Ù tē', 'ak': 'Saint Vincent ne Grenadines', 'am': 'ሴንት ቪንሴንት እና ግሬናዲንስ', 'ar': 'سانت فنسنت وجزر غرينادين', 'as': 'ছেইণ্ট ভিনচেণ্ট আৰু গ্ৰীণাডাইনছ', 'asa': 'Thantavithenti na Grenadini', 'ast': 'San Vicente y Granadines', 'az': 'Sent-Vinsent və Qrenadinlər', 'az-Cyrl': 'Сент-Винсент вә Гренадинләр', 'az-Latn': 'Sent-Vinsent və Qrenadinlər', 'bal-Latn': 'St. Wensent o Grenádin', 'bas': 'Nûmpubi Vɛ̂ŋsâŋ nì grènàdîn', 'be': 'Сент-Вінсент і Грэнадзіны', 'be-tarask': 'Сент-Вінсент і Грэнадзіны', 'bez': 'Husantavisenti na Hugrenadini', 'bg': 'Сейнт Винсънт и Гренадини', 'blo': 'Sɛŋ Fɩnsaŋ na Grenaɖiniisi', 'bm': 'Vinisɛn-Senu-ni-Grenadini', 'bn': 'সেন্ট ভিনসেন্ট ও গ্রেনাডিনস', 'br': 'Sant Visant hag ar Grenadinez', 'brx': 'सैन्ट भिन्सेन्ट आरो ग्रेनादीन्स', 'bs': 'Sveti Vinsent i Grenadin', 'bs-Cyrl': 'Свети Винсент и Гренадини', 'bs-Latn': 'Sveti Vinsent i Grenadin', 'ca': 'Saint Vincent i les Grenadines', 'ca-ES-valencia': 'Saint Vincent i les Grenadines', 'ccp': '𑄥𑄬𑄚𑄴𑄑𑄴 𑄞𑄨𑄚𑄴𑄥𑄬𑄚𑄴𑄑𑄴 𑄃𑄮 𑄘𑄳𑄠 𑄉𑄳𑄢𑄬𑄚𑄓𑄨𑄚𑄴𑄥𑄴', 'ce': 'Сент-Винсент а, Гренадинаш а', 'ceb': 'St. Vincent & Grenadines', 'cgg': 'Senti Vinsent na Gurenadini', 'chr': 'ᎤᏓᏅᏘ ᏫᏂᏏᏂᏗ ᎠᎴ ᎾᏍᎩ ᏇᎾᏗᏁᏍ', 'ckb': 'سەینت ڤینسەنت و گرینادینز', 'cs': 'Svatý Vincenc a Grenadiny', 'cv': 'Сент-Винсент тата Гренадины', 'cy': 'Saint Vincent a’r Grenadines', 'da': 'Saint Vincent og Grenadinerne', 'dav': 'Santavisenti na Grenadini', 'de': 'St. Vincent und die Grenadinen', 'dje': 'Seŋvinsaŋ nda Grenadine', 'dsb': 'St. Vincent a Grenadiny', 'dz': 'སེནཊ་ཝིན་སེནཌ྄ ཨེནཌ་ གི་རེ་ན་དིནས྄', 'ebu': 'Santavisenti na Grenadini', 'ee': 'Saint Vincent kple Grenadine nutome', 'el': 'Άγιος Βικέντιος και Γρεναδίνες', 'el-polyton': 'Ἅγιος Βικέντιος καὶ Γρεναδίνες', 'en': 'St. Vincent & Grenadines', 'en-Dsrt': '𐐝𐐩𐑌𐐻 𐐚𐐮𐑌𐑅𐐲𐑌𐐻 𐐰𐑌𐐼 𐑄 𐐘𐑉𐐯𐑌𐐲𐐼𐐨𐑌𐑆', 'eo': 'Sankta Vincento kaj Grenadinoj', 'es': 'San Vicente y las Granadinas', 'et': 'Saint Vincent ja Grenadiinid', 'eu': 'Saint Vincent eta Grenadinak', 'ewo': 'Ǹfúfúb-Vɛngəsáŋ-ai-Bə Gələnadín', 'fa': 'سنت وینسنت و گرنادین', 'ff': 'See Weesaa e Garnadiin', 'ff-Adlm': '𞤅𞤼. 𞤜𞤭𞤲𞤧𞤢𞤲 & 𞤘𞤭𞤪𞤲𞤢𞤣𞤭𞥅𞤲', 'ff-Latn': 'See Weesaa e Garnadiin', 'fi': 'Saint Vincent ja Grenadiinit', 'fil': 'St. Vincent & Grenadines', 'fo': 'St. Vinsent & Grenadinoyggjar', 'fr': 'Saint-Vincent-et-les Grenadines', 'fur': 'San Vincent e lis Grenadinis', 'fy': 'Saint Vincent en de Grenadines', 'ga': 'San Uinseann agus na Greanáidíní', 'gd': 'Naomh Bhionsant agus Eileanan Greanadach', 'gl': 'San Vicente e as Granadinas', 'gsw': 'St. Vincent und d’Grönadiine', 'gu': 'સેંટ વિન્સેંટ અને ગ્રેનેડાઇંસ', 'guz': 'Santavisenti na Grenadini', 'ha': 'San Binsan Da Girnadin', 'he': 'סנט וינסנט והגרנדינים', 'hi': 'सेंट विंसेंट और ग्रेनाडाइंस', 'hi-Latn': 'St. Vincent & Grenadines', 'hr': 'Sveti Vincent i Grenadini', 'hsb': 'St. Vincent a Grenadiny', 'ht': 'Saint-Vincent-et-les Grenadines', 'hu': 'Saint Vincent és a Grenadine-szigetek', 'hy': 'Սենթ Վինսենթ և Գրենադիններ', 'ia': 'Sancte Vincente e le Grenadinas', 'id': 'Saint Vincent dan Grenadine', 'ig': 'St. Vincent & Grenadines', 'is': 'Sankti Vinsent og Grenadíneyjar', 'it': 'Saint Vincent e Grenadine', 'ja': 'セントビンセント及びグレナディーン諸島', 'jmc': 'Santavisenti na Grenadini', 'jv': 'Saint Vinsen lan Grénadin', 'ka': 'სენტ-ვინსენტი და გრენადინები', 'kaa': 'Сент-Винсент ҳәм Гренадин атаўлары', 'kaa-Cyrl': 'Сент-Винсент ҳәм Гренадин атаўлары', 'kab': 'San Vansu d Grunadin', 'kam': 'Santavisenti na Grenadini', 'kde': 'Santavisenti na Glenadini', 'kea': 'San Bisenti i Granadinas', 'kgp': 'Sỹ Visenti kar Granỹninỹ', 'khq': 'Seŋvinsaŋ nda Grenadine', 'ki': 'Santavisenti na Grenadini', 'kk': 'Сент-Винсент және Гренадин аралдары', 'kk-Cyrl': 'Сент-Винсент және Гренадин аралдары', 'kln': 'Emetab Vincent netilil ak Grenadines', 'km': 'សាំង\u200bវ៉ាំងសង់ និង ហ្គ្រេណាឌីន', 'kn': 'ಸೇಂಟ್. ವಿನ್ಸೆಂಟ್ ಮತ್ತು ಗ್ರೆನೆಡೈನ್ಸ್', 'ko': '세인트빈센트그레나딘', 'kok': 'सेंट विंसेंट ऐंड द ग्रेनेडाइंस', 'kok-Deva': 'सेंट विंसेंट ऐंड द ग्रेनेडाइंस', 'ks': 'سینٹ وینسؠٹ تہٕ گریناڑاینٕز', 'ks-Arab': 'سینٹ وینسؠٹ تہٕ گریناڑاینٕز', 'ksb': 'Santavisenti na Glenadini', 'ksf': 'sɛnvǝnsǝŋ ri grɛnadín', 'ksh': 'Zint Vinzänz un de Jrenadines-Enselle', 'ku': 'Saint Vincent û Giravên Grenadînê', 'kxv': 'seint vinseṇṭ aḍe grenaḍi', 'kxv-Deva': 'सेंट विंसेंट ऑड़े ग्रेनाडाइंस', 'kxv-Latn': 'seint vinseṇṭ aḍe grenaḍi', 'kxv-Orya': 'ସେଣ୍ଟ ୱିନସେଣ୍ଟ ଅଡ଼େ ଗ୍ରିନାଡାଇନ୍ସ', 'kxv-Telu': 'సెయింట్ విన్సెంట్ మరియు గ్రెనడీన్స్', 'ky': 'Сент-Винсент жана Гренадиндер', 'lag': 'Mʉtakatíifu Viséenti na Gernadíini', 'lb': 'St. Vincent an d’Grenadinnen', 'lg': 'Senti Vinsenti ne Gurendadiini', 'lld': 'St. Vincent y Grenadines', 'ln': 'Sántu vesá mpé Gelenadinɛ', 'lo': 'ເຊນ ວິນເຊນ ແລະ ເກຣເນດິນ', 'lt': 'Šventasis Vincentas ir Grenadinai', 'lu': 'Santu vesa ne Ngelenadine', 'luo': 'Saint Vincent gi Grenadines', 'luy': 'Santavisenti na Grenadini', 'lv': 'Sentvinsenta un Grenadīnas', 'mai': 'सेंट विंसेंट आ ग्रेनाडाइंस', 'mas': 'Santavisenti na Grenadini', 'mer': 'Santavisenti na Grenadini', 'mfe': 'Saint-Vincent-ek-Grenadines', 'mg': 'Saint-Vincent-et-les Grenadines', 'mgh': 'Usantavisenti na Grenadini', 'mi': 'Hato Wēneti me Keretīni', 'mk': 'Сент Винсент и Гренадини', 'ml': 'സെന്റ് വിൻസെന്റും ഗ്രനെഡൈൻസും', 'mn': 'Сент-Винсент ба Гренадин', 'mr': 'सेंट व्हिन्सेंट आणि ग्रेनडाइन्स', 'ms': 'Saint Vincent dan Grenadines', 'mt': 'Saint Vincent u l-Grenadini', 'mua': 'Sǝr Vinceŋ ne Grenadiŋ', 'my': 'စိန့်ဗင်းဆင့်နှင့် ဂရိနေဒိုင်', 'mzn': 'سنت وینسنت و گرنادین', 'naq': 'Saint Vincent and the Grenadines', 'nb': 'St. Vincent og Grenadinene', 'nd': 'Saint Vincent and the Grenadines', 'ne': 'सेन्ट भिन्सेन्ट र ग्रेनाडिन्स', 'nl': 'Saint Vincent en de Grenadines', 'nmg': 'Saint Vincent ba Grenadines', 'nn': 'St. Vincent og Grenadinane', 'no': 'St. Vincent og Grenadinene', 'nqo': 'ߛߍ߲ߕ-ߝ߭ߍߛߊ߲ ߜ߭ߙߋߣߊߘߌ߫', 'nyn': 'Senti Vinsent na Gurenadini', 'om': 'St. Vinseet fi Gireenadines', 'or': 'ସେଣ୍ଟ ଭିନସେଣ୍ଟ ଏବଂ ଗ୍ରେନାଡାଇନ୍ସ', 'pa': 'ਸੇਂਟ ਵਿਨਸੈਂਟ ਐਂਡ ਗ੍ਰੇਨਾਡੀਨਸ', 'pa-Guru': 'ਸੇਂਟ ਵਿਨਸੈਂਟ ਐਂਡ ਗ੍ਰੇਨਾਡੀਨਸ', 'pcm': 'Sent Vínsẹnt & Grẹ́nádians', 'pl': 'Saint Vincent i Grenadyny', 'ps': 'سینټ ویسنټینټ او ګرینډینز', 'pt': 'São Vicente e Granadinas', 'qu': 'San Vicente y las Granadinas', 'rm': 'Saint Vincent e las Grenadinas', 'rn': 'Sevensa na Gerenadine', 'ro': 'Saint Vincent și Grenadinele', 'rof': 'Santavisenti na Grenadini', 'ru': 'Сент-Винсент и Гренадины', 'rwk': 'Santavisenti na Grenadini', 'saq': 'Santavisenti na Grenadini', 'sbp': 'Santavisenti na Gilenadini', 'sc': 'Santu Vissente e sas Grenadinas', 'sd': 'سینٽ ونسنت ۽ گریناڊینز', 'sd-Arab': 'سینٽ ونسنت ۽ گریناڊینز', 'se': 'Saint Vincent ja Grenadine', 'seh': 'São Vicente e Granadinas', 'ses': 'Seŋvinsaŋ nda Grenadine', 'sg': 'Sên-Vensäan na âGrenadîni', 'shi': 'ⵙⴰⵏⴼⴰⵏⵙⴰⵏ ⴷ ⴳⵔⵉⵏⴰⴷⵉⵏ', 'shi-Latn': 'sanfansan d grinadin', 'shi-Tfng': 'ⵙⴰⵏⴼⴰⵏⵙⴰⵏ ⴷ ⴳⵔⵉⵏⴰⴷⵉⵏ', 'si': 'ශාන්ත වින්සන්ට් සහ ග්\u200dරෙනඩින්ස්', 'sk': 'Svätý Vincent a Grenadíny', 'sl': 'Saint Vincent in Grenadine', 'smn': 'St. Vincent já Grenadines', 'sn': 'Saint Vincent and the Grenadines', 'so': 'St. Finsent & Girenadiins', 'sq': 'Shën-Vincent dhe Grenadine', 'sr': 'Сент Винсент и Гренадини', 'sr-Cyrl': 'Сент Винсент и Гренадини', 'sr-Latn': 'Sent Vinsent i Grenadini', 'sv': 'S:t Vincent och Grenadinerna', 'sw': 'St. Vincent na Grenadines', 'syr': 'ܣܐܢܬ ܒܝܢܣܝܢܬ ܘܓܪܝܢܐܕܝܢܐܣ', 'ta': 'செயின்ட் வின்சென்ட் & கிரெனடைன்ஸ்', 'te': 'సెయింట్ విన్సెంట్ మరియు గ్రెనడీన్స్', 'teo': 'Santavisenti na Grenadini', 'tg': 'Сент-Винсент ва Гренадина', 'th': 'เซนต์วินเซนต์และเกรนาดีนส์', 'ti': 'ቅዱስ ቪንሰንትን ግረነዲነዝን', 'tk': 'Sent-Winsent we Grenadinler', 'to': 'Sā Viniseni mo Kulenatini', 'tr': 'Saint Vincent ve Grenadinler', 'tt': 'Сент-Винсент һәм Гренадин', 'twq': 'Seŋvinsaŋ nda Grenadine', 'tzm': 'Santevinsent d Grinadin', 'ug': 'ساينت ۋىنسېنت ۋە گىرېنادىنېس', 'uk': 'Сент-Вінсент і Гренадіни', 'ur': 'سینٹ ونسنٹ اور گرینیڈائنز', 'uz': 'Sent-Vinsent va Grenadin', 'uz-Cyrl': 'Сент-Винсент ва Гренадин', 'uz-Latn': 'Sent-Vinsent va Grenadin', 'vai': 'ꔻꘋ ꔲꘋꔻꘋ ꗪ ꖶꔓꕯꔵꘋ ꖸ', 'vai-Latn': 'Siŋ Viŋsi', 'vai-Vaii': 'ꔻꘋ ꔲꘋꔻꘋ ꗪ ꖶꔓꕯꔵꘋ ꖸ', 'vec': 'S. Vincenso e Granadine', 'vi': 'St. Vincent và Grenadines', 'vun': 'Santavisenti na Grenadini', 'wae': 'St. Vincent und d’Grenadine', 'wo': 'Saŋ Weesaa ak Garanadin', 'xh': 'ESt. Vincent & Grenadines', 'xnr': 'सेंट विंसेंट कने ग्रेनाडाइंस', 'xog': 'Senti Vinsenti ni Gurendadiini', 'yav': 'sɛ́ŋ fɛŋsáŋ elekelenatíin', 'yo': 'Fisẹnnti ati Genadina', 'yrl': 'Sã Wisẽti asuí Garanadĩna-ita', 'yue': '聖文森特同格林納丁斯', 'yue-Hans': '圣文森特同格林纳丁斯', 'yue-Hant': '聖文森特同格林納丁斯', 'zgh': 'ⵙⴰⵏⴼⴰⵏⵙⴰⵏ ⴷ ⴳⵔⵉⵏⴰⴷⵉⵏ', 'zh': '圣文森特和格林纳丁斯', 'zh-Hans': '圣文森特和格林纳丁斯', 'zh-Hant': '聖文森及格瑞那丁', 'zu': 'i-Saint Vincent ne-Grenadines'},
    'VE': {'af': 'Venezuela', 'agq': 'Vɛ̀nɛ̀zǔɛɛlà', 'ak': 'Venezuela', 'am': 'ቬንዙዌላ', 'ar': 'فنزويلا', 'as': 'ভেনিজুৱেলা', 'asa': 'Venezuela', 'ast': 'Venezuela', 'az': 'Venesuela', 'az-Cyrl': 'Венесуела', 'az-Latn': 'Venesuela', 'bal-Latn': 'Wenezwélá', 'bas': 'Vènèzùelà', 'be': 'Венесуэла', 'be-tarask': 'Венесуэла', 'bez': 'Huvenezuela', 'bg': 'Венецуела', 'bgn': 'وینزوویلا', 'blo': 'Fenesuwelaa', 'bm': 'Venezuwela', 'bn': 'ভেনেজুয়েলা', 'br': 'Venezuela', 'brx': 'भेनेजुवेला', 'bs': 'Venecuela', 'bs-Cyrl': 'Венецуела', 'bs-Latn': 'Venecuela', 'ca': 'Veneçuela', 'ca-ES-valencia': 'Veneçuela', 'ccp': '𑄞𑄬𑄚𑄬𑄎𑄪𑄠𑄬𑄣', 'ce': 'Венесуэла', 'ceb': 'Venezuela', 'cgg': 'Venezuwera', 'chr': 'ᏪᏁᏑᏪᎳ', 'ckb': 'ڤەنزوێلا', 'cs': 'Venezuela', 'cv': 'Венесуэла', 'cy': 'Venezuela', 'da': 'Venezuela', 'dav': 'Venezuela', 'de': 'Venezuela', 'dje': 'Veneezuyeela', 'dsb': 'Venezuela', 'dz': 'བེ་ནི་ཛུ་ཝེ་ལ', 'ebu': 'Venezuela', 'ee': 'Venezuela nutome', 'el': 'Βενεζουέλα', 'el-polyton': 'Βενεζουέλα', 'en': 'Venezuela', 'eo': 'Venezuelo', 'es': 'Venezuela', 'et': 'Venezuela', 'eu': 'Venezuela', 'ewo': 'Venezuéla', 'fa': 'ونزوئلا', 'ff': 'Wenesuwelaa', 'ff-Adlm': '𞤜𞤫𞥊𞤲𞤭𞥅𞥁𞤵𞤱𞤫𞤤𞤢𞥄', 'ff-Latn': 'Wenesuwelaa', 'fi': 'Venezuela', 'fil': 'Venezuela', 'fo': 'Venesuela', 'fr': 'Venezuela', 'fur': 'Venezuela', 'fy': 'Venezuela', 'ga': 'Veiniséala', 'gd': 'A’ Bheiniseala', 'gl': 'Venezuela', 'gsw': 'Venezueela', 'gu': 'વેનેઝુએલા', 'guz': 'Venezuela', 'ha': 'Benezuwela', 'he': 'ונצואלה', 'hi': 'वेनेज़ुएला', 'hi-Latn': 'Venezuela', 'hr': 'Venezuela', 'hsb': 'Venezuela', 'ht': 'Venezuela', 'hu': 'Venezuela', 'hy': 'Վենեսուելա', 'ia': 'Venezuela', 'id': 'Venezuela', 'ig': 'Venezuela', 'is': 'Venesúela', 'it': 'Venezuela', 'ja': 'ベネズエラ', 'jgo': 'Vɛnɛzwɛ́la', 'jmc': 'Venezuela', 'jv': 'Vénésuéla', 'ka': 'ვენესუელა', 'kaa': 'Венесуэла', 'kaa-Cyrl': 'Венесуэла', 'kab': 'Venzwila', 'kam': 'Venezuela', 'kde': 'Venezuela', 'kea': 'Vinizuela', 'kgp': 'Venẽjuvéra', 'khq': 'Veneezuyeela', 'ki': 'Venezuela', 'kk': 'Венесуэла', 'kk-Cyrl': 'Венесуэла', 'kln': 'Emetab Venezuela', 'km': 'វ៉េណេស៊ុយអេឡា', 'kn': 'ವೆನೆಜುವೆಲಾ', 'ko': '베네수엘라', 'kok': 'विनेझुएला', 'kok-Deva': 'विनेझुएला', 'ks': 'وینازوٗلا', 'ks-Arab': 'وینازوٗلا', 'ksb': 'Venezuela', 'ksf': 'wɛnǝzwɛla', 'ksh': 'Venezuela', 'ku': 'Venezuela', 'kxv': 'venejuela', 'kxv-Deva': 'वेनेज़ुएला', 'kxv-Latn': 'venejuela', 'kxv-Orya': 'ୱେନେଜୁଏଲ', 'kxv-Telu': 'వెనిజులా', 'ky': 'Венесуэла', 'lag': 'Venezuéela', 'lb': 'Venezuela', 'lg': 'Venzwera', 'lld': 'Venezuela', 'ln': 'Venézuela', 'lo': 'ເວເນຊູເອລາ', 'lt': 'Venesuela', 'lu': 'Venezuela', 'luo': 'Venezuela', 'luy': 'Venezuela', 'lv': 'Venecuēla', 'mai': 'बेनेजुएला', 'mas': 'Venezuela', 'mer': 'Venezuela', 'mfe': 'Venezuela', 'mg': 'Venezoelà', 'mgh': 'Uvenezuela', 'mi': 'Penehūera', 'mk': 'Венецуела', 'ml': 'വെനിസ്വേല', 'mn': 'Венесуэл', 'mr': 'व्हेनेझुएला', 'ms': 'Venezuela', 'mt': 'il-Venezwela', 'mua': 'Sǝr Venezuela', 'my': 'ဗင်နီဇွဲလား', 'mzn': 'ونزوئلا', 'naq': 'Venezuelab', 'nb': 'Venezuela', 'nd': 'Venezuela', 'ne': 'भेनेजुएला', 'nl': 'Venezuela', 'nmg': 'Vǝnǝzuela', 'nn': 'Venezuela', 'no': 'Venezuela', 'nqo': 'ߝ߭ߣߋߖ߭ߎߦߋߟߊ߫', 'nyn': 'Venezuwera', 'om': 'Veenzuweelaa', 'or': 'ଭେନେଜୁଏଲା', 'pa': 'ਵੇਨੇਜ਼ੂਏਲਾ', 'pa-Guru': 'ਵੇਨੇਜ਼ੂਏਲਾ', 'pcm': 'Vẹnẹzuẹ́la', 'pl': 'Wenezuela', 'ps': 'وینزویلا', 'pt': 'Venezuela', 'qu': 'Venezuela', 'rm': 'Venezuela', 'rn': 'Venezuwela', 'ro': 'Venezuela', 'rof': 'Venezuela', 'ru': 'Венесуэла', 'rwk': 'Venezuela', 'saq': 'Venezuela', 'sbp': 'Venesuela', 'sc': 'Venetzuela', 'sd': 'وينزويلا', 'sd-Arab': 'وينزويلا', 'se': 'Venezuela', 'seh': 'Venezuela', 'ses': 'Veneezuyeela', 'sg': 'Venezueläa', 'shi': 'ⴼⵉⵏⵣⵡⵉⵍⴰ', 'shi-Latn': 'finzwila', 'shi-Tfng': 'ⴼⵉⵏⵣⵡⵉⵍⴰ', 'si': 'වෙනිසියුලාව', 'sk': 'Venezuela', 'sl': 'Venezuela', 'smn': 'Venezuela', 'sn': 'Venezuela', 'so': 'Fenisuweela', 'sq': 'Venezuelë', 'sr': 'Венецуела', 'sr-Cyrl': 'Венецуела', 'sr-Latn': 'Venecuela', 'sv': 'Venezuela', 'sw': 'Venezuela', 'syr': 'ܒܢܙܘܝܠܐ', 'ta': 'வெனிசுலா', 'te': 'వెనిజులా', 'teo': 'Venezuela', 'tg': 'Венесуэла', 'th': 'เวเนซุเอลา', 'ti': 'ቬኔዝዌላ', 'tk': 'Wenesuela', 'to': 'Venesuela', 'tr': 'Venezuela', 'tt': 'Венесуэла', 'twq': 'Veneezuyeela', 'tzm': 'Vinzwilla', 'ug': 'ۋېنېسۇئېلا', 'uk': 'Венесуела', 'ur': 'وینزوئیلا', 'uz': 'Venesuela', 'uz-Cyrl': 'Венесуэла', 'uz-Latn': 'Venesuela', 'vai': 'ꕙꔳꕪꘋ ꕸꖃꔀ', 'vai-Latn': 'Vɛnɛzuwela', 'vai-Vaii': 'ꕙꔳꕪꘋ ꕸꖃꔀ', 'vec': 'Venesuela', 'vi': 'Venezuela', 'vun': 'Venezuela', 'wae': 'Venezuela', 'wo': 'Wenesiyela', 'xh': 'EVenezuela', 'xnr': 'वेनेज़ुएला', 'xog': 'Venzwera', 'yav': 'fenesuwelá', 'yi': 'ווענעזועלע', 'yo': 'Fẹnẹṣuẹla', 'yrl': 'Wenesuera', 'yue': '委內瑞拉', 'yue-Hans': '委内瑞拉', 'yue-Hant': '委內瑞拉', 'zgh': 'ⴼⵉⵏⵣⵡⵉⵍⴰ', 'zh': '委内瑞拉', 'zh-Hans': '委内瑞拉', 'zh-Hant': '委內瑞拉', 'zu': 'i-Venezuela'},
    'VG': {'af': 'Britse Maagde-eilande', 'agq': 'Chwɨlà m̀ Vidzinyìa m̀ Bɨ̀letì mò', 'ak': 'Ngresifoɔ Virgin Island', 'am': 'የእንግሊዝ ቨርጂን ደሴቶች', 'ar': 'جزر فيرجن البريطانية', 'as': 'ব্ৰিটিছ ভাৰ্জিন দ্বীপপুঞ্জ', 'asa': 'Vithiwa vya Virgin vya Uingeredha', 'ast': 'Islles Vírxenes Britániques', 'az': 'Britaniyanın Virgin adaları', 'az-Cyrl': 'Британијанын Вирҝин адалары', 'az-Latn': 'Britaniyanın Virgin adaları', 'bal-Latn': 'Bretáni Ajgén Islánd', 'bas': 'Bìòn bi kɔnji bi Ŋgisì', 'be': 'Брытанскія Віргінскія астравы', 'be-tarask': 'Брытанскія Віргінскія астравы', 'bez': 'Ifisima fya Virgin fya Huingereza', 'bg': 'Британски Вирджински острови', 'blo': 'Fɩrjɩɩn kǝBʊtukǝltǝna Gɛɛshɩ kabʊja', 'bm': 'Angilɛ ka Sungurunnin Gun', 'bn': 'ব্রিটিশ ভার্জিন দ্বীপপুঞ্জ', 'br': 'Inizi Gwercʼh Breizh-Veur', 'brx': 'ब्रिटिश भार्जिन द्वीपफोर', 'bs': 'Britanska Djevičanska ostrva', 'bs-Cyrl': 'Британска Дјевичанска острва', 'bs-Latn': 'Britanska Djevičanska ostrva', 'ca': 'Illes Verges Britàniques', 'ca-ES-valencia': 'Illes Verges Britàniques', 'ccp': '𑄝𑄳𑄢𑄨𑄑𑄨𑄌𑄴 𑄞𑄢𑄴𑄎𑄨𑄚𑄴 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Виргинийн гӀайренаш (Британи)', 'ceb': 'British Virgin Islands', 'cgg': 'Ebizinga bya Virigini ebya Bungyereza', 'chr': 'ᏈᏗᏍ ᎠᏒᏂᎸ ᏂᎨᏒᎾ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی ڤیرجنی بەریتانیا', 'cs': 'Britské Panenské ostrovy', 'cv': 'Британин Виргини утравӗсем', 'cy': 'Ynysoedd Gwyryf Prydain', 'da': 'De Britiske Jomfruøer', 'dav': 'Visiwa vya Virgin vya Uingereza', 'de': 'Britische Jungferninseln', 'dje': 'Britiši Virgin gungey', 'dsb': 'Britiske kněžniske kupy', 'dz': 'ཝརཇིན་གླིང་ཚོམ་ བྲཱི་ཊིཤ་མངའ་ཁོངས', 'ebu': 'Visiwa vya Virgin vya Uingereza', 'ee': 'Britaintɔwo ƒe Virgin ƒudomekpowo nutome', 'el': 'Βρετανικές Παρθένες Νήσοι', 'el-polyton': 'Βρετανικὲς Παρθένοι Νῆσοι', 'en': 'British Virgin Islands', 'en-Dsrt': '𐐒𐑉𐐮𐐼𐐮𐑇 𐐚𐐲𐑉𐐾𐐲𐑌 𐐌𐑊𐐲𐑌𐐼𐑆', 'eo': 'Britaj Virgulininsuloj', 'es': 'Islas Vírgenes Británicas', 'et': 'Briti Neitsisaared', 'eu': 'Birjina uharte britainiarrak', 'ewo': 'ńnam Minlán ɛ́ngəlís', 'fa': 'جزایر ویرجین بریتانیا', 'ff': 'duuɗe kecce britanii', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤜𞤭𞤪𞤺𞤭𞥅𞤲 𞤄𞤪𞤭𞤼𞤢𞤲𞤭𞤴𞤢𞤲𞤳𞤮𞥅𞤶𞤫', 'ff-Latn': 'duuɗe kecce britanii', 'fi': 'Brittiläiset Neitsytsaaret', 'fil': 'British Virgin Islands', 'fo': 'Stóra Bretlands Jomfrúoyggjar', 'fr': 'Îles Vierges britanniques', 'fur': 'Isulis vergjinis britanichis', 'fy': 'Britse Maagdeneilannen', 'ga': 'Oileáin Bhriotanacha na Maighdean', 'gd': 'Eileanan Breatannach na Maighdinn', 'gl': 'Illas Virxes Británicas', 'gsw': 'Britischi Jungfere-Insle', 'gu': 'બ્રિટિશ વર્જિન આઇલેન્ડ્સ', 'guz': 'Visiwa vya Virgin vya Uingereza', 'ha': 'Tsibirin Birjin Na Birtaniya', 'he': 'איי הבתולה הבריטיים', 'hi': 'ब्रिटिश वर्जिन द्वीपसमूह', 'hi-Latn': 'British Virgin Islands', 'hr': 'Britanski Djevičanski Otoci', 'hsb': 'Britiske knježniske kupy', 'ht': 'Îles Vierges britanniques', 'hu': 'Brit Virgin-szigetek', 'hy': 'Բրիտանական Վիրջինյան կղզիներ', 'ia': 'Insulas Virgine britannic', 'id': 'Kepulauan Virgin Britania Raya', 'ig': 'British Virgin Islands', 'is': 'Bresku Jómfrúaeyjar', 'it': 'Isole Vergini Britanniche', 'ja': '英領ヴァージン諸島', 'jmc': 'Visiwa vya Virgin vya Uingereza', 'jv': 'Kapuloan Virgin Britania', 'ka': 'ბრიტანეთის ვირჯინის კუნძულები', 'kaa': 'Британия Виргин атаўлары', 'kaa-Cyrl': 'Британия Виргин атаўлары', 'kab': 'Tigzirin Tiverjiniyin Tigliziyin', 'kam': 'Visiwa vya Virgin vya Uingereza', 'kde': 'Chisiwa Chivihi cha Wingalesa', 'kea': 'Ilhas Virjens Británikas', 'kgp': 'Mritỹnĩja Goj-vẽso', 'khq': 'Britiši Virgin gungey', 'ki': 'Visiwa vya Virgin vya Uingereza', 'kk': 'Британдық Виргин аралдары', 'kk-Cyrl': 'Британдық Виргин аралдары', 'kln': 'Ikwembeyotab British Virgin', 'km': 'កោះ\u200bវឺជិន\u200bចក្រភព\u200bអង់គ្លេស', 'kn': 'ಬ್ರಿಟಿಷ್ ವರ್ಜಿನ್ ದ್ವೀಪಗಳು', 'ko': '영국령 버진아일랜드', 'kok': 'ब्रिटिश वर्जिन आयलँड्स', 'kok-Deva': 'ब्रिटिश वर्जिन आयलँड्स', 'ks': 'بَرطانوی ؤرجِن جٔزیٖرٕ', 'ks-Arab': 'بَرطانوی ؤرجِن جٔزیٖرٕ', 'ksb': 'Visiwa vya Vilgin vya Uingeeza', 'ksf': 'zǝ bɛ gɔn inɛ a ingɛrís', 'ksh': 'de brettesche Juffer-Enselle', 'ku': 'Giravên Vîrjînê yên Brîtanyayê', 'kxv': 'briṭis vrjin dīp', 'kxv-Deva': 'ब्रिटिस वर्जिन द्वीप', 'kxv-Latn': 'briṭis vrjin dīp', 'kxv-Orya': 'ବ୍ରିଟିସ୍ ୱିରଜିନ ଦିପ', 'kxv-Telu': 'బ్రిటిష్ వర్జిన్ దీవులు', 'ky': 'Виргин аралдары (Британия)', 'lag': 'Visíiwa vya Vigíini vya Ʉɨngeréesa', 'lb': 'Britesch Joffereninselen', 'lg': 'Bizinga ebya Virigini ebitwalibwa Bungereza', 'lld': 'Isoles Vergines britaniches', 'ln': 'Bisanga bya Vierzi ya Angɛlɛtɛ́lɛ', 'lo': 'ໝູ່ເກາະ ເວີຈິນຂອງອັງກິດ', 'lt': 'Didžiosios Britanijos Mergelių Salos', 'lu': 'Lutanda lua Vierzi wa Angeletele', 'luo': 'British Virgin Islands', 'luy': 'Visiwa vya Virgin vya Uingereza', 'lv': 'Britu Virdžīnas', 'mai': 'ब्रिटिश वर्जिन द्वीपसमूह', 'mas': 'Visiwa vya Virgin vya Uingereza', 'mer': 'Aĩrandi cia Virgin cia Ngeretha', 'mfe': 'Zil vierz britanik', 'mg': 'Nosy britanika virijiny', 'mi': 'Ngā Moutere Puhi Piritene', 'mk': 'Британски Девствени Острови', 'ml': 'ബ്രിട്ടീഷ് വെർജിൻ ദ്വീപുകൾ', 'mn': 'Британийн Виржиний арлууд', 'mr': 'ब्रिटिश व्हर्जिन बेटे', 'ms': 'Kepulauan Virgin British', 'mt': 'il-Gżejjer Verġni Brittaniċi', 'mua': 'ser Anglofon ma laŋne', 'my': 'ဗြိတိသျှ ဗာဂျင်း ကျွန်းစု', 'mzn': 'بریتانیای ویرجین', 'naq': 'British Virgin Islands', 'nb': 'De britiske jomfruøyene', 'nd': 'British Virgin Islands', 'ne': 'बेलायती भर्जिन टापुहरू', 'nl': 'Britse Maagdeneilanden', 'nmg': 'Minsilɛ́ mímaŋ mí ngɛ̄lɛ̄n', 'nn': 'Dei britiske Jomfruøyane', 'no': 'De britiske jomfruøyene', 'nqo': 'ߓߙߌߕߊ߲ߓߊ߫ ߕߌ߲߫ ߞߊߓߊ߲', 'nus': 'Burutic dhuɔ̱ɔ̱l be̱rgin', 'nyn': 'Ebizinga bya Virigini ebya Bungyereza', 'om': 'Odoloota Varjiin Biritish', 'or': 'ବ୍ରିଟିଶ୍\u200c ଭର୍ଜିନ୍ ଦ୍ୱୀପପୁଞ୍ଜ', 'pa': 'ਬ੍ਰਿਟਿਸ਼ ਵਰਜਿਨ ਟਾਪੂ', 'pa-Guru': 'ਬ੍ਰਿਟਿਸ਼ ਵਰਜਿਨ ਟਾਪੂ', 'pcm': 'Brítísh Vájín Aílands', 'pl': 'Brytyjskie Wyspy Dziewicze', 'ps': 'بریتانوی ویګور ټاپوګان', 'pt': 'Ilhas Virgens Britânicas', 'qu': 'Islas Vírgenes Británicas', 'rm': 'Inslas Virginas Britannicas', 'rn': 'Ibirwa by’isugi by’Abongereza', 'ro': 'Insulele Virgine Britanice', 'rof': 'Visiwa vya Virgin vya Uingereza', 'ru': 'Виргинские о-ва (Великобритания)', 'rwk': 'Visiwa vya Virgin vya Uingereza', 'saq': 'Visiwa vya Virgin vya Uingereza', 'sbp': 'Figunguli ifya Viliginiya ifya Uwingelesa', 'sc': 'Ìsulas Vèrgines Britànnicas', 'sd': 'برطانوي ورجن ٻيٽ', 'sd-Arab': 'برطانوي ورجن ٻيٽ', 'se': 'Brittania Virgin-sullot', 'seh': 'Ilhas Virgens Britânicas', 'ses': 'Britiši Virgin gungey', 'sg': 'Âzôâ Viîrîggo tî Anglëe', 'shi': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵜⵉⵎⴳⴰⴷ ⵏ ⵏⵏⴳⵍⵉⵣ', 'shi-Latn': 'tigzirin timgad n nngliz', 'shi-Tfng': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵜⵉⵎⴳⴰⴷ ⵏ ⵏⵏⴳⵍⵉⵣ', 'si': 'බ්\u200dරිතාන්\u200dය වර්ජින් දූපත්', 'sk': 'Britské Panenské ostrovy', 'sl': 'Britanski Deviški otoki', 'smn': 'Brittiliih Nieidâsuolluuh', 'sn': 'Zvitsuwa zveHingirandi', 'so': 'Biritish Farjin Island', 'sq': 'Ishujt e Virgjër Britanikë', 'sr': 'Британска Девичанска Острва', 'sr-Cyrl': 'Британска Девичанска Острва', 'sr-Latn': 'Britanska Devičanska Ostrva', 'sv': 'Brittiska Jungfruöarna', 'sw': 'Visiwa vya Virgin, Uingereza', 'syr': 'ܓܙܖ̈ܝܐ ܒܬܘ̈ܠܐ ܕܒܪܝܛܢܝܐ', 'ta': 'பிரிட்டீஷ் கன்னித் தீவுகள்', 'te': 'బ్రిటిష్ వర్జిన్ దీవులు', 'teo': 'Visiwa vya Virgin vya Uingereza', 'tg': 'Ҷазираҳои Виргини Британия', 'th': 'หมู่เกาะบริติชเวอร์จิน', 'ti': 'ደሴታት ደናግል ብሪጣንያ', 'tk': 'Britan Wirgin adalary', 'to': 'ʻOtumotu Vilikini fakapilitānia', 'tr': 'Britanya Virjin Adaları', 'tt': 'Британия Виргин утраулары', 'twq': 'Britiši Virgin gungey', 'tzm': 'Tigzirin (Virgin) Tibṛiṭaniyin', 'ug': 'ئەنگلىيە ۋىرگىن ئاراللىرى', 'uk': 'Британські Віргінські острови', 'ur': 'برٹش ورجن آئلینڈز', 'uz': 'Britaniya Virgin orollari', 'uz-Cyrl': 'Британия Виргин ороллари', 'uz-Latn': 'Britaniya Virgin orollari', 'vai': 'ꔛꔟꔻ ꗩꗡ ꗏ ꖷꖬ ꔳꘋꗣ', 'vai-Latn': 'Jengéesi Bhɛɛ Lɔ Musu Tiŋŋɛ', 'vai-Vaii': 'ꔛꔟꔻ ꗩꗡ ꗏ ꖷꖬ ꔳꘋꗣ', 'vec': 'Ìzole Vèrjini britàneghe', 'vi': 'Quần đảo Virgin thuộc Anh', 'vun': 'Visiwa vya Virgin vya Uingereza', 'wae': 'Britiši Jungfröiwinslä', 'wo': 'Duni Wirsin yu Brëtaañ', 'xh': 'EBritish Virgin Islands', 'xnr': 'ब्रिटिश वर्जिन द्वीपां दा झुंड', 'xog': 'Ebizinga bya Virigini ebitwalibwa Bungereza', 'yav': 'Filisíin ungilís', 'yo': 'Etíkun Fágínì ti ìlú Bírítísì', 'yrl': 'Kapuã-ita Viyẽ-ita Biritãnika-ita', 'yue': '英屬維京群島', 'yue-Hans': '英属维京群岛', 'yue-Hant': '英屬維京群島', 'zgh': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵜⵉⵎⴳⴰⴷ ⵏ ⵏⵏⴳⵍⵉⵣ', 'zh': '英属维尔京群岛', 'zh-Hans': '英属维尔京群岛', 'zh-Hant': '英屬維京群島', 'zu': 'i-British Virgin Islands'},
    'VI': {'af': 'VSA se Maagde-eilande', 'agq': 'U. S. Chwɨlà fɨ Mbuʔmbu', 'ak': 'Amɛrika Virgin Islands', 'am': 'የአሜሪካ ቨርጂን ደሴቶች', 'ar': 'جزر فيرجن الأمريكية', 'as': 'ইউ. এছ. ভাৰ্জিন দ্বীপপুঞ্জ', 'asa': 'Vithiwa vya Virgin vya Marekani', 'ast': 'Islles Vírxenes Americanes', 'az': 'ABŞ Virgin adaları', 'az-Cyrl': 'АБШ Вирҝин адалары', 'az-Latn': 'ABŞ Virgin adaları', 'bal-Latn': 'Amriki Ajgén Islánd', 'bas': 'Bìòn bi kɔnji bi U.S.', 'be': 'Амерыканскія Віргінскія астравы', 'be-tarask': 'Амерыканскія Віргінскія астравы', 'bez': 'Ifisima fya Virgin fya Humelekani', 'bg': 'Американски Вирджински острови', 'blo': 'Fɩrjɩɩn kaBʊtukǝltǝna Amalɩka kabʊja', 'bm': 'Ameriki ka Sungurunnin Gun', 'bn': 'মার্কিন যুক্তরাষ্ট্রীয় ভার্জিন দ্বীপপুঞ্জ', 'br': 'Inizi Gwercʼh ar Stadoù-Unanet', 'brx': 'इउ. एस भार्जिन द्वीपफोर', 'bs': 'Američka Djevičanska ostrva', 'bs-Cyrl': 'Америчка Дјевичанска острва', 'bs-Latn': 'Američka Djevičanska ostrva', 'ca': 'Illes Verges dels Estats Units', 'ca-ES-valencia': 'Illes Verges dels Estats Units', 'ccp': '𑄟𑄢𑄴𑄇𑄨𑄚𑄴 𑄎𑄧𑄙𑄢𑄬𑄌𑄴𑄎𑄮𑄢𑄴 𑄞𑄢𑄴𑄎𑄨𑄚𑄴 𑄉𑄭 𑄉𑄭 𑄞𑄨𑄘𑄳𑄠', 'ce': 'Виргинийн гӀайренаш (АЦШ)', 'ceb': 'U.S. Virgin Islands', 'cgg': 'Ebizinga bya Virigini ebya Amerika', 'chr': 'U.S. ᎠᏒᏂᎸ ᏂᎨᏒᎾ ᏚᎦᏚᏛᎢ', 'ckb': 'دوورگەکانی ڤیرجنی ئەمەریکا', 'cs': 'Americké Panenské ostrovy', 'cv': 'Виргини утравӗсем (АПШ)', 'cy': 'Ynysoedd Gwyryf yr Unol Daleithiau', 'da': 'De Amerikanske Jomfruøer', 'dav': 'Visiwa vya Virgin vya Marekani', 'de': 'Amerikanische Jungferninseln', 'dje': 'Ameerik Virgin Gungey', 'dsb': 'Ameriske kněžniske kupy', 'dz': 'ཝརཇིན་གླིང་ཚོམ་ ཡུ་ཨེས་ཨེ་མངའ་ཁོངས', 'ebu': 'Visiwa vya Virgin vya Marekani', 'ee': 'U.S. Vɛrgin ƒudomekpowo nutome', 'el': 'Αμερικανικές Παρθένες Νήσοι', 'el-polyton': 'Ἀμερικανικὲς Παρθένοι Νῆσοι', 'en': 'U.S. Virgin Islands', 'en-Dsrt': '𐐏.𐐝. 𐐚𐐲𐑉𐐾𐐲𐑌 𐐌𐑊𐐲𐑌𐐼𐑆', 'eo': 'Usonaj Virgulininsuloj', 'es': 'Islas Vírgenes de EE. UU.', 'et': 'USA Neitsisaared', 'eu': 'Birjina uharte amerikarrak', 'ewo': 'Minlán Mi Amɛrəkə', 'fa': 'جزایر ویرجین ایالات متحده', 'ff': 'Duuɗe Kecce Amerik', 'ff-Adlm': '𞤕𞤵𞤪𞤭𞥅𞤶𞤫 𞤜𞤭𞤪𞤺𞤭𞥅𞤲 𞤁𞤂𞤀', 'ff-Latn': 'Duuɗe Kecce Amerik', 'fi': 'Yhdysvaltain Neitsytsaaret', 'fil': 'U.S. Virgin Islands', 'fo': 'Sambandsríki Amerikas Jomfrúoyggjar', 'fr': 'Îles Vierges des États-Unis', 'fur': 'Isulis vergjinis americanis', 'fy': 'Amerikaanske Maagdeneilannen', 'ga': 'Oileáin Mheiriceánacha na Maighdean', 'gd': 'Eileanan na Maighdinn aig na SA', 'gl': 'Illas Virxes Estadounidenses', 'gsw': 'Amerikaanischi Jungfere-Insle', 'gu': 'યુએસ વર્જિન આઇલેન્ડ્સ', 'guz': 'Visiwa vya Virgin vya Marekani', 'ha': 'Tsibiran Birjin Ta Amurka', 'he': 'איי הבתולה של ארצות הברית', 'hi': 'यू॰एस॰ वर्जिन द्वीपसमूह', 'hi-Latn': 'U.S. Virgin Islands', 'hr': 'Američki Djevičanski Otoci', 'hsb': 'Ameriske knježniske kupy', 'ht': 'Îles Vierges des États-Unis', 'hu': 'Amerikai Virgin-szigetek', 'hy': 'ԱՄՆ Վիրջինյան կղզիներ', 'ia': 'Insulas Virgine statounitese', 'id': 'Kepulauan Virgin Amerika Serikat', 'ig': 'U.S. Virgin Islands', 'is': 'Bandarísku Jómfrúaeyjar', 'it': 'Isole Vergini Americane', 'ja': '米領ヴァージン諸島', 'jmc': 'Visiwa vya Virgin vya Marekani', 'jv': 'Kapuloan Virgin Amérika', 'ka': 'აშშ-ის ვირჯინის კუნძულები', 'kaa': 'Америка Виргин атаўлары', 'kaa-Cyrl': 'Америка Виргин атаўлары', 'kab': 'W.D. Tigzirin n Virginya', 'kam': 'Visiwa vya Virgin vya Marekani', 'kde': 'Chisiwa Chivihi cha Malekani', 'kea': 'Ilhas Virjens Merkanas', 'kgp': 'Virjĩg Goj-vẽso tỹ Amẽrikỹnỹ', 'khq': 'Ameerik Virgin Gungey', 'ki': 'Visiwa vya Virgin vya Marekani', 'kk': 'АҚШ-тың Виргин аралдары', 'kk-Cyrl': 'АҚШ-тың Виргин аралдары', 'kln': 'Ikwemweiyotab Amerika', 'km': 'កោះ\u200bវឺជីន\u200bអាមេរិក', 'kn': 'ಯು.ಎಸ್. ವರ್ಜಿನ್ ದ್ವೀಪಗಳು', 'ko': '미국령 버진아일랜드', 'kok': 'यु. एस. वर्जिन आयलँड्\u200dस', 'kok-Deva': 'यु. एस. वर्जिन आयलँड्\u200dस', 'ks': 'یوٗ ایس ؤرجِن جٔزیٖرٕ', 'ks-Arab': 'یوٗ ایس ؤرجِن جٔزیٖرٕ', 'ksb': 'Visiwa vya Vilgin vya Malekani', 'ksf': 'zǝ bɛ gɔn inɛ á amɛrika', 'ksh': 'de ammärrikahnesche Juffer-Enselle', 'ku': 'Giravên Vîrjînê yên Amerîkayê', 'kxv': 'yu.es. vrjin dīp', 'kxv-Deva': 'यू॰एस॰ वर्जिन द्वीप', 'kxv-Latn': 'yu.es. vrjin dīp', 'kxv-Orya': 'ୟୁ ଏସ ୱିରଜିନ ଦିପ', 'kxv-Telu': 'యు.ఎస్. వర్జిన్ దీవులు', 'ky': 'Виргин аралдары (АКШ)', 'lag': 'Visíiwa vya Vigíini vya Amerɨ́ka', 'lb': 'Amerikanesch Joffereninselen', 'lg': 'Bizinga bya Virigini eby’Amerika', 'lld': 'Isoles Vergines americanes', 'ln': 'Bisanga bya Vierzi ya Ameriki', 'lo': 'ໝູ່ເກາະ ເວີຈິນ ຂອງສະຫະລັດ', 'lt': 'Jungtinių Valstijų Mergelių Salos', 'lu': 'Lutanda lua Vierzi wa Ameriki', 'luo': 'U.S. Virgin Islands', 'luy': 'Visiwa vya Virgin vya Marekani', 'lv': 'ASV Virdžīnas', 'mai': 'यू॰एस॰ वर्जिन द्वीपसमूह', 'mas': 'Visiwa vya Virgin vya Marekani', 'mer': 'Aĩrandi cia Virgin cia Amerika', 'mfe': 'Zil Vierz Lamerik', 'mg': 'Nosy Virijiny Etazonia', 'mi': 'Ngā Moutere Puhi Amerika', 'mk': 'Американски Девствени Острови', 'ml': 'യു.എസ്. വെർജിൻ ദ്വീപുകൾ', 'mn': 'АНУ-ын Виржиний арлууд', 'mr': 'यू.एस. व्हर्जिन बेटे', 'ms': 'Kepulauan Virgin A.S.', 'mt': 'il-Gżejjer Verġni tal-Istati Uniti', 'mua': 'Sǝr amerika ma laŋne', 'my': 'ယူအက်စ် ဗာဂျင်း ကျွန်းစု', 'mzn': 'آمریکای ویرجین', 'naq': 'U.S. Virgin Islands', 'nb': 'De amerikanske jomfruøyene', 'nd': 'U.S. Virgin Islands', 'ne': 'संयुक्त राज्य भर्जिन टापुहरु', 'nl': 'Amerikaanse Maagdeneilanden', 'nmg': 'Minsilɛ mí maŋ́ m´Amɛrka', 'nn': 'Dei amerikanske Jomfruøyane', 'no': 'De amerikanske jomfruøyene', 'nqo': 'ߞߊ߬ߝߏ߫ ߘߍ߬ߣߍ߲ ߕߌ߲߫ ߞߊߓߊ߲߫', 'nyn': 'Ebizinga bya Virigini ebya Amerika', 'om': 'U.S. Odoloota Varjiin', 'or': 'ଯୁକ୍ତରାଷ୍ଟ୍ର ଭର୍ଜିନ୍ ଦ୍ଵୀପପୁଞ୍ଜ', 'pa': 'ਯੂ ਐੱਸ ਵਰਜਿਨ ਟਾਪੂ', 'pa-Guru': 'ਯੂ ਐੱਸ ਵਰਜਿਨ ਟਾਪੂ', 'pcm': 'U.S. Vájín Aílands', 'pl': 'Wyspy Dziewicze Stanów Zjednoczonych', 'ps': 'د متحده آيالاتو ورجن ټاپوګان', 'pt': 'Ilhas Virgens Americanas', 'qu': 'EE.UU. Islas Vírgenes', 'rm': 'Inslas Virginas Americanas', 'rn': 'Amazinga y’Isugi y’Abanyamerika', 'ro': 'Insulele Virgine Americane', 'rof': 'Visiwa vya Virgin vya Marekani', 'ru': 'Виргинские о-ва (США)', 'rwk': 'Visiwa vya Virgin vya Marekani', 'saq': 'Visiwa vya Virgin vya Marekani', 'sbp': 'Figunguli fya Viliginiya ifya Malekani', 'sc': 'Ìsulas Vèrgines de sos Istados Unidos', 'sd': 'آمريڪي ورجن ٻيٽ', 'sd-Arab': 'آمريڪي ورجن ٻيٽ', 'se': 'AOS Virgin-sullot', 'seh': 'Ilhas Virgens dos EUA', 'ses': 'Ameerik Virgin Gungey', 'sg': 'Âzûâ Virîgo tî Amerîka', 'shi': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵜⵉⵎⴳⴰⴷ ⵏ ⵉⵡⵓⵏⴰⴽ ⵎⵓⵏⵏⵉⵏ', 'shi-Latn': 'tigzirin timgad n iwunak munnin', 'shi-Tfng': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵜⵉⵎⴳⴰⴷ ⵏ ⵉⵡⵓⵏⴰⴽ ⵎⵓⵏⵏⵉⵏ', 'si': 'ඇමරිකානු වර්ජින් දූපත්', 'sk': 'Americké Panenské ostrovy', 'sl': 'Ameriški Deviški otoki', 'smn': 'Ovtâstum Staatâi Nieidâsuolluuh', 'sn': 'Zvitsuwa zveAmerika', 'so': 'U.S Fargin Island', 'sq': 'Ishujt e Virgjër të SHBA-së', 'sr': 'Америчка Девичанска Острва', 'sr-Cyrl': 'Америчка Девичанска Острва', 'sr-Latn': 'Američka Devičanska Ostrva', 'sv': 'Amerikanska Jungfruöarna', 'sw': 'Visiwa vya Virgin, Marekani', 'syr': 'ܓܙܖ̈ܝܐ ܒܬܘ̈ܠܐ ܕܐܡܝܪܟܐ', 'ta': 'யூ.எஸ். கன்னித் தீவுகள்', 'te': 'యు.ఎస్. వర్జిన్ దీవులు', 'teo': 'Visiwa vya Virgin vya Marekani', 'tg': 'Ҷазираҳои Виргини ИМА', 'th': 'หมู่เกาะเวอร์จินของสหรัฐอเมริกา', 'ti': 'ደሴታት ደናግል ኣመሪካ', 'tk': 'ABŞ-nyň Wirgin adalary', 'to': 'ʻOtumotu Vilikini fakaʻamelika', 'tr': 'ABD Virjin Adaları', 'tt': 'АКШ Виргин утраулары', 'twq': 'Ameerik Virgin Gungey', 'tzm': 'Tigzirin n Virjin n Iwunak Yedduklen', 'ug': 'ئا ق ش ۋىرگىن ئاراللىرى', 'uk': 'Віргінські Острови (США)', 'ur': 'امریکی ورجن آئلینڈز', 'uz': 'AQSH Virgin orollari', 'uz-Cyrl': 'АҚШ Виргин ороллари', 'uz-Latn': 'AQSH Virgin orollari', 'vai': 'ꕶꕱ ꗩꗡ ꗏ ꖷꖬ ꔳꘋꗣ', 'vai-Latn': 'Poo Bhɛɛ lɔ Musu Tiŋŋɛ', 'vai-Vaii': 'ꕶꕱ ꗩꗡ ꗏ ꖷꖬ ꔳꘋꗣ', 'vec': 'Ìzole Vèrjini meregane', 'vi': 'Quần đảo Virgin thuộc Hoa Kỳ', 'vun': 'Visiwa vya Virgin vya Marekani', 'wae': 'Amerikaniši Jungfröiwinslä', 'wo': 'Duni Wirsin yu Etaa-sini', 'xh': 'E-U.S. Virgin Islands', 'xnr': 'यू॰एस॰ वर्जिन द्वीपां दा झुंड', 'xog': 'Ebizinga bya Virigini eby’Amerika', 'yav': 'pindisúlɛ́ pi amálíka', 'yo': 'Etikun Fagini ti Amẹrika', 'yrl': 'Kapuã-ira Viyẽ-ita Amerikana-ita', 'yue': '美屬維京群島', 'yue-Hans': '美属维京群岛', 'yue-Hant': '美屬維京群島', 'zgh': 'ⵜⵉⴳⵣⵉⵔⵉⵏ ⵜⵉⵎⴳⴰⴷ ⵏ ⵉⵡⵓⵏⴰⴽ ⵎⵓⵏⵏⵉⵏ', 'zh': '美属维尔京群岛', 'zh-Hans': '美属维尔京群岛', 'zh-Hant': '美屬維京群島', 'zu': 'i-U.S. Virgin Islands'},
    'VN': {'af': 'Viëtnam', 'agq': 'Vìyɛnàm', 'ak': 'Viɛtnam', 'am': 'ቬትናም', 'ar': 'فيتنام', 'as': 'ভিয়েটনাম', 'asa': 'Vietinamu', 'ast': 'Vietnam', 'az': 'Vyetnam', 'az-Cyrl': 'Вјетнам', 'az-Latn': 'Vyetnam', 'bal-Latn': 'Wietnám', 'bas': 'Vìɛ̀dnâm', 'be': 'В’етнам', 'be-tarask': 'В’етнам', 'bez': 'Huvietinamu', 'bg': 'Виетнам', 'bgn': 'ویتنام', 'blo': 'Fɛtnam', 'bm': 'Wiyɛtinamu', 'bn': 'ভিয়েতনাম', 'br': 'Viêt Nam', 'brx': 'भियेतनाम', 'bs': 'Vijetnam', 'bs-Cyrl': 'Вијетнам', 'bs-Latn': 'Vijetnam', 'ca': 'Vietnam', 'ca-ES-valencia': 'Vietnam', 'ccp': '𑄞𑄨𑄠𑄬𑄖𑄴𑄚𑄟𑄴', 'ce': 'Вьетнам', 'ceb': 'Vietnam', 'cgg': 'Viyetinaamu', 'chr': 'ᏫᎡᏘᎾᎻ', 'ckb': 'ڤیەتنام', 'cs': 'Vietnam', 'cv': 'Вьетнам', 'cy': 'Fietnam', 'da': 'Vietnam', 'dav': 'Vietinamu', 'de': 'Vietnam', 'dje': 'Vietnaam', 'dsb': 'Vietnam', 'dz': 'བེཊ་ནཱམ', 'ebu': 'Vietinamu', 'ee': 'Vietnam nutome', 'el': 'Βιετνάμ', 'el-polyton': 'Βιετνάμ', 'en': 'Vietnam', 'en-Dsrt': '𐐚𐐨𐐯𐐻𐑌𐐪𐑋', 'eo': 'Vjetnamo', 'es': 'Vietnam', 'et': 'Vietnam', 'eu': 'Vietnam', 'ewo': 'Viɛdənám', 'fa': 'ویتنام', 'ff': 'Wiyetnaam', 'ff-Adlm': '𞤜𞤭𞤴𞤫𞤼𞤲𞤢𞥄𞤥', 'ff-Latn': 'Wiyetnaam', 'fi': 'Vietnam', 'fil': 'Vietnam', 'fo': 'Vjetnam', 'fr': 'Viêt Nam', 'fur': 'Vietnam', 'fy': 'Vietnam', 'ga': 'Vítneam', 'gd': 'Bhiet-Nam', 'gl': 'Vietnam', 'gsw': 'Wietnam', 'gu': 'વિયેતનામ', 'guz': 'Vietinamu', 'ha': 'Biyetinam', 'he': 'וייטנאם', 'hi': 'वियतनाम', 'hi-Latn': 'Vietnam', 'hr': 'Vijetnam', 'hsb': 'Vietnam', 'ht': 'Viêt Nam', 'hu': 'Vietnám', 'hy': 'Վիետնամ', 'ia': 'Vietnam', 'id': 'Vietnam', 'ig': 'Vietnam', 'is': 'Víetnam', 'it': 'Vietnam', 'ja': 'ベトナム', 'jmc': 'Vietinamu', 'jv': 'Viètnam', 'ka': 'ვიეტნამი', 'kaa': 'Вьетнам', 'kaa-Cyrl': 'Вьетнам', 'kab': 'Vyeṭnam', 'kam': 'Vietinamu', 'kde': 'Vietinamu', 'kea': 'Vietnam', 'kgp': 'Vijétinỹ', 'khq': 'Vietnaam', 'ki': 'Vietinamu', 'kk': 'Вьетнам', 'kk-Cyrl': 'Вьетнам', 'kln': 'Emetab Vietnam', 'km': 'វៀតណាម', 'kn': 'ವಿಯೆಟ್ನಾಮ್', 'ko': '베트남', 'kok': 'व्हिएतनाम', 'kok-Deva': 'व्हिएतनाम', 'ks': 'ویتنام', 'ks-Arab': 'ویتنام', 'ksb': 'Vietinamu', 'ksf': 'wyɛtnám', 'ksh': 'Vijätnamm', 'ku': 'Vîetnam', 'kxv': 'viyetnam', 'kxv-Deva': 'वियतनाम', 'kxv-Latn': 'viyetnam', 'kxv-Orya': 'ୱିଏତନାମ', 'kxv-Telu': 'వియత్నాం', 'ky': 'Вьетнам', 'lag': 'Vietináamu', 'lb': 'Vietnam', 'lg': 'Vyetinaamu', 'lld': 'Vietnam', 'ln': 'Viyetinamɛ', 'lo': 'ຫວຽດນາມ', 'lt': 'Vietnamas', 'lu': 'Viyetiname', 'luo': 'Vietnam', 'luy': 'Vietinamu', 'lv': 'Vjetnama', 'mai': 'बियतनाम', 'mas': 'Vietinamu', 'mer': 'Vietinamu', 'mfe': 'Vietnam', 'mg': 'Vietnam', 'mgh': 'Uvietinamu', 'mi': 'Whitināmu', 'mk': 'Виетнам', 'ml': 'വിയറ്റ്നാം', 'mn': 'Вьетнам', 'mr': 'व्हिएतनाम', 'ms': 'Vietnam', 'mt': 'il-Vjetnam', 'mua': 'Sǝr Vietnam', 'my': 'ဗီယက်နမ်', 'mzn': 'ویتنام', 'naq': 'Vietnammi', 'nb': 'Vietnam', 'nd': 'Vietnam', 'ne': 'भिएतनाम', 'nl': 'Vietnam', 'nmg': 'Viɛtnam', 'nn': 'Vietnam', 'no': 'Vietnam', 'nqo': 'ߝ߭ߌߦߍߕߑߣߊߡ', 'nyn': 'Viyetinaamu', 'om': 'Veetinaam', 'or': 'ଭିଏତନାମ୍', 'pa': 'ਵੀਅਤਨਾਮ', 'pa-Guru': 'ਵੀਅਤਨਾਮ', 'pcm': 'Viẹ́tnam', 'pl': 'Wietnam', 'ps': 'وېتنام', 'pt': 'Vietnã', 'qu': 'Vietnam', 'rm': 'Vietnam', 'rn': 'Viyetinamu', 'ro': 'Vietnam', 'rof': 'Vietinamu', 'ru': 'Вьетнам', 'rwk': 'Vietinamu', 'saq': 'Vietinamu', 'sbp': 'Vietinamu', 'sc': 'Vietnam', 'sd': 'ويتنام', 'sd-Arab': 'ويتنام', 'se': 'Vietnam', 'seh': 'Vietnã', 'ses': 'Vietnaam', 'sg': 'Vietnäm', 'shi': 'ⴼⵉⵜⵏⴰⵎ', 'shi-Latn': 'fitnam', 'shi-Tfng': 'ⴼⵉⵜⵏⴰⵎ', 'si': 'වියට්නාමය', 'sk': 'Vietnam', 'sl': 'Vietnam', 'smn': 'Vietnam', 'sn': 'Vietnam', 'so': 'Fiyetnaam', 'sq': 'Vietnam', 'sr': 'Вијетнам', 'sr-Cyrl': 'Вијетнам', 'sr-Latn': 'Vijetnam', 'sv': 'Vietnam', 'sw': 'Vietnamu', 'syr': 'ܒܝܬܢܐܡ', 'ta': 'வியட்நாம்', 'te': 'వియత్నాం', 'teo': 'Vietinamu', 'tg': 'Ветнам', 'th': 'เวียดนาม', 'ti': 'ቬትናም', 'tk': 'Wýetnam', 'to': 'Vietinami', 'tr': 'Vietnam', 'tt': 'Вьетнам', 'twq': 'Vietnaam', 'tzm': 'Viṭnam', 'ug': 'ۋىيېتنام', 'uk': 'Вʼєтнам', 'ur': 'ویتنام', 'uz': 'Vyetnam', 'uz-Cyrl': 'Вьетнам', 'uz-Latn': 'Vyetnam', 'vai': 'ꗲꕇꖮꔃꕞ', 'vai-Latn': 'Viyanami', 'vai-Vaii': 'ꗲꕇꖮꔃꕞ', 'vec': 'Vietnam', 'vi': 'Việt Nam', 'vun': 'Vietinamu', 'wae': 'Vietnam', 'wo': 'Wiyetnam', 'xh': 'EVietnam', 'xnr': 'वियतनाम', 'xog': 'Vyetinaamu', 'yav': 'fiɛtnáam', 'yi': 'וויעטנאַם', 'yo': 'Fẹtinami', 'yrl': 'Wiyetinã', 'yue': '越南', 'yue-Hans': '越南', 'yue-Hant': '越南', 'zgh': 'ⴼⵉⵜⵏⴰⵎ', 'zh': '越南', 'zh-Hans': '越南', 'zh-Hant': '越南', 'zu': 'i-Vietnam'},
    'VU': {'af': 'Vanuatu', 'agq': 'Vànǔatùw', 'ak': 'Vanuatu', 'am': 'ቫኑአቱ', 'ar': 'فانواتو', 'as': 'ভানাটু', 'asa': 'Vanuatu', 'ast': 'Vanuatu', 'az': 'Vanuatu', 'az-Cyrl': 'Вануату', 'az-Latn': 'Vanuatu', 'bal-Latn': 'Wanuátu', 'bas': 'Vànùatù', 'be': 'Вануату', 'be-tarask': 'Вануату', 'bez': 'Huvanuatu', 'bg': 'Вануату', 'blo': 'Fanuwatu', 'bm': 'Vanuwatu', 'bn': 'ভানুয়াটু', 'br': 'Vanuatu', 'brx': 'भानुआटु', 'bs': 'Vanuatu', 'bs-Cyrl': 'Вануату', 'bs-Latn': 'Vanuatu', 'ca': 'Vanuatu', 'ca-ES-valencia': 'Vanuatu', 'ccp': '𑄞𑄚𑄪𑄠𑄑𑄪', 'ce': 'Вануату', 'ceb': 'Vanuatu', 'cgg': 'Vanuatu', 'chr': 'ᏩᏂᎤᏩᏚ', 'ckb': 'ڤانوواتوو', 'cs': 'Vanuatu', 'cv': 'Вануату', 'cy': 'Vanuatu', 'da': 'Vanuatu', 'dav': 'Vanuatu', 'de': 'Vanuatu', 'dje': 'Vanautu', 'dsb': 'Vanuatu', 'dz': 'ཝ་ནུ་ཨ་ཏུ', 'ebu': 'Vanuatu', 'ee': 'Vanuatu nutome', 'el': 'Βανουάτου', 'el-polyton': 'Βανουάτου', 'en': 'Vanuatu', 'en-Dsrt': '𐐚𐐪𐑌𐐳𐐪𐐼𐐭', 'eo': 'Vanuatuo', 'es': 'Vanuatu', 'et': 'Vanuatu', 'eu': 'Vanuatu', 'ewo': 'Vanuátu', 'fa': 'وانواتو', 'ff': 'Wanuwaatuu', 'ff-Adlm': '𞤜𞤢𞤲𞤵𞤱𞤢𞥄𞤼𞤵', 'ff-Latn': 'Wanuwaatuu', 'fi': 'Vanuatu', 'fil': 'Vanuatu', 'fo': 'Vanuatu', 'fr': 'Vanuatu', 'fur': 'Vanuatu', 'fy': 'Vanuatu', 'ga': 'Vanuatú', 'gd': 'Vanuatu', 'gl': 'Vanuatu', 'gsw': 'Wanuatu', 'gu': 'વાનુઆતુ', 'guz': 'Vanuatu', 'ha': 'Banuwatu', 'he': 'ונואטו', 'hi': 'वनुआतू', 'hi-Latn': 'Vanuatu', 'hr': 'Vanuatu', 'hsb': 'Vanuatu', 'ht': 'Vanuatu', 'hu': 'Vanuatu', 'hy': 'Վանուատու', 'ia': 'Vanuatu', 'id': 'Vanuatu', 'ie': 'Vanuatu', 'ig': 'Vanuatu', 'is': 'Vanúatú', 'it': 'Vanuatu', 'ja': 'バヌアツ', 'jmc': 'Vanuatu', 'jv': 'Vanuatu', 'ka': 'ვანუატუ', 'kaa': 'Вануату', 'kaa-Cyrl': 'Вануату', 'kab': 'Vanwatu', 'kam': 'Vanuatu', 'kde': 'Vanuatu', 'kea': 'Vanuatu', 'kgp': 'Vanũvatu', 'khq': 'Vanautu', 'ki': 'Vanuatu', 'kk': 'Вануату', 'kk-Cyrl': 'Вануату', 'kln': 'Emetab Vanuatu', 'km': 'វ៉ានូទូ', 'kn': 'ವನೌಟು', 'ko': '바누아투', 'kok': 'वनौतू', 'kok-Deva': 'वनौतू', 'ks': 'وانوٗتوٗ', 'ks-Arab': 'وانوٗتوٗ', 'ksb': 'Vanuatu', 'ksf': 'wanwatu', 'ksh': 'Vanuatu', 'ku': 'Vanûatû', 'kxv': 'vanuaatu', 'kxv-Deva': 'वनुआतू', 'kxv-Latn': 'vanuaatu', 'kxv-Orya': 'ୱାନୁଆତୁ', 'kxv-Telu': 'వనాటు', 'ky': 'Вануату', 'lag': 'Vanuáatu', 'lb': 'Vanuatu', 'lg': 'Vanawuwatu', 'lld': 'Vanuatu', 'ln': 'Vanuatu', 'lo': 'ວານົວຕູ', 'lt': 'Vanuatu', 'lu': 'Vanuatu', 'luo': 'Vanuatu', 'luy': 'Vanuatu', 'lv': 'Vanuatu', 'mai': 'बनातू', 'mas': 'Vanuatu', 'mer': 'Vanuatu', 'mfe': 'Vanuatu', 'mg': 'Vanoatò', 'mgh': 'Uvanuatu', 'mi': 'Whenuatū', 'mk': 'Вануату', 'ml': 'വന്വാതു', 'mn': 'Вануату', 'mr': 'वानुआतु', 'ms': 'Vanuatu', 'mt': 'Vanuatu', 'mua': 'Sǝr Vanuatu', 'my': 'ဗနွားတူ', 'mzn': 'وانواتو', 'naq': 'Vanuatu', 'nb': 'Vanuatu', 'nd': 'Vhanuatu', 'ne': 'भानुआतु', 'nl': 'Vanuatu', 'nmg': 'Vanuatu', 'nn': 'Vanuatu', 'no': 'Vanuatu', 'nqo': 'ߝ߭ߊߣߎߦߊߕߎ߫', 'nyn': 'Vanuatu', 'om': 'Vanuwaatu', 'or': 'ଭାନୁଆତୁ', 'pa': 'ਵਾਨੂਆਟੂ', 'pa-Guru': 'ਵਾਨੂਆਟੂ', 'pcm': 'Vanuátu', 'pl': 'Vanuatu', 'ps': 'واناتو', 'pt': 'Vanuatu', 'qu': 'Vanuatu', 'rm': 'Vanuatu', 'rn': 'Vanuwatu', 'ro': 'Vanuatu', 'rof': 'Vanuatu', 'ru': 'Вануату', 'rwk': 'Vanuatu', 'saq': 'Vanuatu', 'sbp': 'Vanuatu', 'sc': 'Vanuatu', 'sd': 'وينيٽيو', 'sd-Arab': 'وينيٽيو', 'se': 'Vanuatu', 'seh': 'Vanuatu', 'ses': 'Vanautu', 'sg': 'Vanuatü', 'shi': 'ⴼⴰⵏⵡⴰⵟⵓ', 'shi-Latn': 'fanwaṭu', 'shi-Tfng': 'ⴼⴰⵏⵡⴰⵟⵓ', 'si': 'වනුවාටු', 'sk': 'Vanuatu', 'sl': 'Vanuatu', 'smn': 'Vanuatu', 'sn': 'Vanuatu', 'so': 'Fanuaatu', 'sq': 'Vanuatu', 'sr': 'Вануату', 'sr-Cyrl': 'Вануату', 'sr-Latn': 'Vanuatu', 'sv': 'Vanuatu', 'sw': 'Vanuatu', 'syr': 'ܒܐܢܘܐܛܘ', 'ta': 'வனுவாட்டு', 'te': 'వనాటు', 'teo': 'Vanuatu', 'tg': 'Вануату', 'th': 'วานูอาตู', 'ti': 'ቫንዋቱ', 'tk': 'Wanuatu', 'to': 'Vanuatu', 'tr': 'Vanuatu', 'tt': 'Вануату', 'twq': 'Vanautu', 'tzm': 'Vanwatu', 'ug': 'ۋانۇئاتۇ', 'uk': 'Вануату', 'ur': 'وینوآٹو', 'uz': 'Vanuatu', 'uz-Cyrl': 'Вануату', 'uz-Latn': 'Vanuatu', 'vai': 'ꕙꖸꕎꖤ', 'vai-Latn': 'Vanuwátu', 'vai-Vaii': 'ꕙꖸꕎꖤ', 'vec': 'Vanuatu', 'vi': 'Vanuatu', 'vun': 'Vanuatu', 'wae': 'Vanuatu', 'wo': 'Wanuatu', 'xh': 'EVanuatu', 'xnr': 'वनुआतू', 'xog': 'Vanawuwatu', 'yav': 'fanuatú', 'yi': 'וואַנואַטו', 'yo': 'Faniatu', 'yrl': 'Wanuatu', 'yue': '萬那杜', 'yue-Hans': '万那杜', 'yue-Hant': '萬那杜', 'zgh': 'ⴼⴰⵏⵡⴰⵟⵓ', 'zh': '瓦努阿图', 'zh-Hans': '瓦努阿图', 'zh-Hant': '萬那杜', 'zu': 'i-Vanuatu'},
    'Vaii': {'af': 'Vai', 'ak': 'Vai kasa', 'am': 'ቫይ', 'ar': 'الفاي', 'as': 'ভাই', 'ast': 'vai', 'az': 'vay', 'az-Latn': 'vay', 'bal-Latn': 'Wái', 'be': 'вайскае', 'be-tarask': 'вайскае', 'bg': 'вайска', 'bn': 'ভাই', 'br': 'vai', 'brx': 'भाई', 'bs': 'vai pismo', 'bs-Cyrl': 'ваи писмо', 'bs-Latn': 'vai pismo', 'ca': 'vai', 'ca-ES-valencia': 'vai', 'ccp': '𑄞𑄭', 'chr': 'ᏩᏱ', 'cs': 'vai', 'cy': 'Vai', 'da': 'vai', 'de': 'Vai', 'dsb': 'vaiske pismo', 'el': 'Βάι', 'el-polyton': 'Βάι', 'en': 'Vai', 'en-Dsrt': '𐐚𐐴', 'es': 'vai', 'et': 'vai', 'eu': 'vaiarra', 'fa': 'ویایی', 'ff-Adlm': '𞤜𞤢𞥄𞤴', 'fi': 'vailainen', 'fil': 'Vai', 'fo': 'vai', 'fr': 'vaï', 'fy': 'Vai', 'ga': 'Vadhais', 'gd': 'Vai', 'gl': 'vai', 'gsw': 'Vai', 'gu': 'વાઇ', 'ha': 'Vai', 'he': 'ואי', 'hi': 'वाई', 'hi-Latn': 'Vai', 'hr': 'vai pismo', 'hsb': 'vaiske pismo', 'ht': 'vaï', 'hu': 'Vai', 'hy': 'վայական', 'id': 'Vai', 'ig': 'Vai', 'is': 'vai', 'it': 'vai', 'ja': 'ヴァイ文字', 'jv': 'Vai', 'ka': 'ვაი', 'kgp': 'vaj', 'kk': 'ваи жазуы', 'kk-Cyrl': 'ваи жазуы', 'km': 'វ៉ៃ', 'kn': 'ವಾಯ್', 'ko': '바이 문자', 'kok': 'वाइ', 'kok-Deva': 'वाइ', 'ks': 'واے', 'ks-Arab': 'واے', 'ky': 'Ваи (жазуу)', 'lb': 'Vai', 'lo': 'ໄວ', 'lt': 'vai', 'lv': 'vaju', 'mai': 'बाई', 'mi': 'Wai', 'mk': 'вај', 'ml': 'വൈ', 'mn': 'Ваи бичиг', 'mr': 'वाई', 'ms': 'Vai', 'my': 'ဗိုင်း', 'nb': 'vai', 'ne': 'भाइ', 'nl': 'Vai', 'nn': 'vai', 'no': 'vai', 'nqo': 'ߒߝ߭ߊߦߌ', 'or': 'ୱାଇ', 'pa': 'ਵਾਈ', 'pa-Guru': 'ਵਾਈ', 'pcm': 'Vaí Lángwej', 'pl': 'vai', 'ps': 'وای', 'pt': 'vai', 'qu': 'Vai Simi', 'ro': 'vai', 'ru': 'вайская', 'sc': 'vai', 'sd': 'وائي', 'sd-Arab': 'وائي', 'si': 'වායි', 'sk': 'vai', 'sl': 'zlogovna pisava vai', 'so': 'Dadka Vai', 'sq': 'vai', 'sr': 'ваи', 'sr-Cyrl': 'ваи', 'sr-Latn': 'vai', 'sv': 'vaj', 'sw': 'Kivai', 'ta': 'வை', 'te': 'వాయి', 'tg': 'Вайӣ', 'th': 'ไว', 'ti': 'ቫይ', 'tk': 'Waý ýazuwy', 'to': 'tohinima fakavai', 'tr': 'Vai', 'ug': 'ۋايچە', 'uk': 'ваї', 'ur': 'وائی', 'uz': 'vay', 'uz-Latn': 'vay', 'vec': 'vai', 'vi': 'Chữ Vai', 'xnr': 'वाई', 'yo': 'Èdè Fai', 'yrl': 'wai', 'yue': '瓦依文', 'yue-Hans': '瓦依文', 'yue-Hant': '瓦依文', 'zh': '瓦依文', 'zh-Hans': '瓦依文', 'zh-Hant': '瓦依文', 'zu': 'isi-Vai'},
    'Visp': {'ar': 'الكلام المرئي', 'ast': 'fala visible', 'az': 'danışma səsləri', 'az-Latn': 'danışma səsləri', 'bg': 'Видима реч', 'bn': 'দৃশ্যমান ভাষা', 'brx': 'वीज़ीबल बोली', 'bs': 'vidljivi govor', 'bs-Cyrl': 'видљиви говор', 'bs-Latn': 'vidljivi govor', 'ca': 'llenguatge visible', 'ca-ES-valencia': 'llenguatge visible', 'ccp': '𑄘𑄬𑄉𑄧𑄎𑄭𑄘𑄳𑄠𑄬 𑄞𑄌𑄴', 'cs': 'viditelná řeč', 'da': 'synlig tale', 'de': 'Sichtbare Sprache', 'el': 'Ορατή ομιλία', 'el-polyton': 'Ορατή ομιλία', 'en': 'Visible Speech', 'en-Dsrt': '𐐚𐐱𐑆𐐱𐐺𐐲𐑊 𐐝𐐹𐐨𐐽', 'es': 'lenguaje visible', 'et': 'nähtava kõne', 'fa': 'گفتار قابل مشاهده', 'fi': 'näkyvä puhe', 'fr': 'parole visible', 'fy': 'Sichtbere spraak', 'gsw': 'Sichtbari Schpraach', 'gu': 'વિસિબલ સ્પીચ', 'hi': 'विसिबल स्पीच', 'hi-Latn': 'Visible Speech', 'hr': 'Visible Speech', 'ht': 'parole visible', 'hu': 'Látható beszéd', 'id': 'Ucapan Terlihat', 'it': 'alfabeto visivo', 'ja': '視話法', 'ka': 'ხილული მეტყველება', 'kgp': 'vĩ-ve-há', 'kn': 'ವಿಸಿಬಲ್ ಸ್ಪೀಚ್', 'ko': '시화법', 'ks': 'وِزِبٕل سپیٖچ', 'ks-Arab': 'وِزِبٕل سپیٖچ', 'lb': 'Siichtbar Sprooch', 'lo': 'ຄຳເວົ້າທີ່ເບີ່ງເຫັນໄດ້', 'lt': 'matoma kalba', 'mk': 'видлив говор', 'ml': 'ദൃശ്യഭാഷ', 'mr': 'दृश्य संवाद', 'my': 'မြင်နိုင်သော စကား', 'nb': 'synlig tale', 'ne': 'दृश्यमय वाणी', 'nl': 'Zichtbare spraak', 'nn': 'synleg tale', 'no': 'synlig tale', 'or': 'ଭିଜିବଲ୍ ସ୍ପିଚ୍', 'pl': 'Visible Speech', 'pt': 'visible speech', 'rm': 'alfabet visibel', 'ru': 'видимая речь', 'sl': 'vidni govor', 'sr': 'видљиви говор', 'sr-Cyrl': 'видљиви говор', 'sr-Latn': 'vidljivi govor', 'sv': 'synligt tal', 'ta': 'விசிபிள் ஸ்பீச்', 'te': 'కనిపించే భాష', 'th': 'คำพูดที่มองเห็นได้', 'to': 'tohinima fakafonētiki-hāmai', 'tr': 'Konuşma Sesleri Çizimlemesi', 'ug': 'كۆرۈنۈشچان تاۋۇش', 'uk': 'видиме мовлення', 'vi': 'Tiếng nói Nhìn thấy được', 'yrl': 'nheẽga xipiawera', 'yue': '視覺語音文字', 'yue-Hans': '视觉语音文字', 'yue-Hant': '視覺語音文字', 'zh': '可见语言', 'zh-Hans': '可见语言', 'zh-Hant': '視覺語音文字'},
    'Vith': {'bal-Latn': 'Witkuki', 'bs': 'vitkuki pismo', 'bs-Cyrl': 'виткуки писмо', 'bs-Latn': 'vitkuki pismo', 'en': 'Vithkuqi', 'et': 'Vithkuqi', 'eu': 'vithkuqi', 'ff-Adlm': '𞤜𞤭𞤼𞤳𞤵𞤹𞤭', 'fi': 'vithkuqi', 'fo': 'vithkuqi', 'gd': 'Vithkuqi', 'he': 'ויטוקוקי', 'hi-Latn': 'Vithkuqi', 'id': 'Vithkuqi', 'ig': 'Vithkuqi', 'kk': 'виткути жазуы', 'kk-Cyrl': 'виткути жазуы', 'mn': 'Виткуки', 'nl': 'Vithkuqi', 'nqo': 'ߝ߭ߌߕߑߞߎߜ߭ߌ߫', 'ro': 'vithkuqi', 'sc': 'vithkuqi', 'sq': 'vithkuqi', 'tg': 'Виткукӣ', 'vec': 'beita kukju'},
    'WF': {'af': 'Wallis en Futuna', 'agq': 'Wales à Fùwtuwnà', 'ak': 'Wallis ne Futuna', 'am': 'ዋሊስ እና ፉቱና ደሴቶች', 'ar': 'جزر والس وفوتونا', 'as': 'ৱালিছ আৰু ফুটুনা', 'asa': 'Walith na Futuna', 'ast': 'Wallis y Futuna', 'az': 'Uollis və Futuna', 'az-Cyrl': 'Уоллис вә Футуна', 'az-Latn': 'Uollis və Futuna', 'bal-Latn': 'Wális o Futuná', 'bas': 'Wàlîs nì Fùtunà', 'be': 'Уоліс і Футуна', 'be-tarask': 'Уоліс і Футуна', 'bez': 'Huwalis na Hufutuna', 'bg': 'Уолис и Футуна', 'blo': 'Walis na Futuna', 'bm': 'Walisi-ni-Futuna', 'bn': 'ওয়ালিস ও ফুটুনা', 'br': 'Wallis ha Futuna', 'brx': 'वालिस आरो फ्युचुना', 'bs': 'Ostrva Valis i Futuna', 'bs-Cyrl': 'Валис и Футуна', 'bs-Latn': 'Ostrva Valis i Futuna', 'ca': 'Wallis i Futuna', 'ca-ES-valencia': 'Wallis i Futuna', 'ccp': '𑄤𑄣𑄨𑄌𑄴 𑄃𑄮 𑄜𑄪𑄑𑄪𑄚', 'ce': 'Уоллис а, Футуна а', 'ceb': 'Wallis & Futuna', 'cgg': 'Warris na Futuna', 'chr': 'ᏩᎵᏍ ᎠᎴ ᏊᏚᎾ', 'ckb': 'والیس و فوتونا', 'cs': 'Wallis a Futuna', 'cv': 'Уоллис тата Футуна', 'cy': 'Wallis a Futuna', 'da': 'Wallis og Futuna', 'dav': 'Walis na Futuna', 'de': 'Wallis und Futuna', 'dje': 'Wallis nda Futuna', 'dsb': 'Wallis a Futuna', 'dz': 'ཝལ་ལིས྄་ ཨེནཌ་ ཕུ་ཏུ་ན་', 'ebu': 'Walis na Futuna', 'ee': 'Wallis kple Futuna nutome', 'el': 'Γουάλις και Φουτούνα', 'el-polyton': 'Νῆσοι Οὐάλλις καὶ Φουτουνά', 'en': 'Wallis & Futuna', 'en-Dsrt': '𐐎𐐪𐑊𐐮𐑅 𐐰𐑌𐐼 𐐙𐐭𐐻𐐭𐑌𐐲', 'eo': 'Valiso kaj Futuno', 'es': 'Wallis y Futuna', 'et': 'Wallis ja Futuna', 'eu': 'Wallis eta Futuna', 'ewo': 'Walís-ai-Futúna', 'fa': 'والیس و فوتونا', 'ff': 'Walis e Futuna', 'ff-Adlm': '𞤏𞤢𞤤𞥆𞤭𞥅𞤧 & 𞤊𞤵𞤼𞤵𞤲𞤢', 'ff-Latn': 'Walis e Futuna', 'fi': 'Wallis ja Futuna', 'fil': 'Wallis & Futuna', 'fo': 'Wallis- og Futunaoyggjar', 'fr': 'Wallis-et-Futuna', 'fur': 'Wallis e Futuna', 'fy': 'Wallis en Futuna', 'ga': 'Vailís agus Futúna', 'gd': 'Uallas agus Futuna', 'gl': 'Wallis e Futuna', 'gsw': 'Wallis und Futuuna', 'gu': 'વૉલિસ અને ફ્યુચુના', 'guz': 'Walis na Futuna', 'ha': 'Walis Da Futuna', 'he': 'איי ווליס ופוטונה', 'hi': 'वालिस और फ़्यूचूना', 'hi-Latn': 'Wallis & Futuna', 'hr': 'Wallis i Futuna', 'hsb': 'Wallis a Futuna', 'ht': 'Wallis-et-Futuna', 'hu': 'Wallis és Futuna', 'hy': 'Ուոլիս և Ֆուտունա', 'ia': 'Wallis e Futuna', 'id': 'Kepulauan Wallis dan Futuna', 'ig': 'Wallis & Futuna', 'is': 'Wallis- og Fútúnaeyjar', 'it': 'Wallis e Futuna', 'ja': 'ウォリス・フツナ', 'jmc': 'Walis na Futuna', 'jv': 'Wallis lan Futuna', 'ka': 'უოლისი და ფუტუნა', 'kab': 'Wallis d Futuna', 'kam': 'Walis na Futuna', 'kde': 'Walis na Futuna', 'kea': 'Ualis i Futuna', 'kgp': 'Varinh kar Futunỹ', 'khq': 'Wallis nda Futuna', 'ki': 'Walis na Futuna', 'kk': 'Уоллис және Футуна', 'kk-Cyrl': 'Уоллис және Футуна', 'kln': 'Emetab Walis ak Futuna', 'km': 'វ៉ាលីស និង\u200bហ្វូទូណា', 'kn': 'ವಾಲಿಸ್ ಮತ್ತು ಫುಟುನಾ', 'ko': '왈리스-푸투나 제도', 'kok': 'वालिस आनी फ्यूचूना', 'kok-Deva': 'वालिस आनी फ्यूचूना', 'ks': 'والِس تہٕ فیوٗچوٗنا', 'ks-Arab': 'والِس تہٕ فیوٗچوٗنا', 'ksb': 'Walis na Futuna', 'ksf': 'walis ri futuna', 'ksh': 'Wallis un Futuna', 'ku': 'Wallis û Futuna', 'kxv': 'vallis aḍe puṭuna', 'kxv-Deva': 'वालिस ऑड़े प़्यूचूना', 'kxv-Latn': 'vallis aḍe puṭuna', 'kxv-Orya': 'ୱାଲିସ ଅଡ଼େ ପୁଟୁନା', 'kxv-Telu': 'వాల్లిస్ మరియు ఫుటునా', 'ky': 'Уоллис жана Футуна', 'lag': 'Walíisi na Futúuna', 'lb': 'Wallis a Futuna', 'lg': 'Walisi ne Futuna', 'lld': 'Wallis y Futuna', 'ln': 'Walisɛ mpé Futuna', 'lo': 'ວາລລິສ ແລະ ຟູຕູນາ', 'lt': 'Volisas ir Futūna', 'lu': 'Walise ne Futuna', 'luo': 'Wallis gi Futuna', 'luy': 'Walis na Futuna', 'lv': 'Volisa un Futunas salas', 'mai': 'बालिस आ फुटुना', 'mas': 'Walis na Futuna', 'mer': 'Walis na Futuna', 'mfe': 'Wallis-ek-Futuna', 'mg': 'Wallis sy Futuna', 'mgh': 'Uwalis na Futuna', 'mi': 'Warihi me Whutuna', 'mk': 'Валис и Футуна', 'ml': 'വാലിസ് ആന്റ് ഫ്യൂച്യുന', 'mn': 'Уоллис ба Футуна', 'mr': 'वालिस आणि फ्यूचूना', 'ms': 'Wallis dan Futuna', 'mt': 'Wallis u Futuna', 'mua': 'Wallis ne Futuna', 'my': 'ဝေါလစ်နှင့် ဖူကျူးနား', 'mzn': 'والیس و فوتونا', 'naq': 'Wallis and Futuna', 'nb': 'Wallis og Futuna', 'nd': 'Wallis and Futuna', 'ne': 'वालिस र फुटुना', 'nl': 'Wallis en Futuna', 'nmg': 'Wallis ba Futuna', 'nn': 'Wallis og Futuna', 'no': 'Wallis og Futuna', 'nqo': 'ߥߊߟߌߛߌ߫ ߣߌ߫ ߝߕߎߣߊ߫', 'nyn': 'Warris na Futuna', 'om': 'Waalis fi Futtuuna', 'or': 'ୱାଲିସ୍ ଏବଂ ଫୁତୁନା', 'pa': 'ਵਾਲਿਸ ਅਤੇ ਫੂਟੂਨਾ', 'pa-Guru': 'ਵਾਲਿਸ ਅਤੇ ਫੂਟੂਨਾ', 'pcm': 'Wọ́lis & Fiutúna', 'pl': 'Wallis i Futuna', 'ps': 'والیس او فوتونا', 'pt': 'Wallis e Futuna', 'qu': 'Wallis y Futuna', 'rm': 'Wallis & Futuna', 'rn': 'Walisi na Futuna', 'ro': 'Wallis și Futuna', 'rof': 'Walis na Futuna', 'ru': 'Уоллис и Футуна', 'rwk': 'Walis na Futuna', 'saq': 'Walis na Futuna', 'sbp': 'Walisi ni Futuna', 'sc': 'Wallis e Futuna', 'sd': 'والس ۽ فتونا', 'sd-Arab': 'والس ۽ فتونا', 'se': 'Wallis ja Futuna', 'seh': 'Wallis e Futuna', 'ses': 'Wallis nda Futuna', 'sg': 'Walîsi na Futuna', 'shi': 'ⵡⴰⵍⵉⵙ ⴷ ⴼⵓⵜⵓⵏⴰ', 'shi-Latn': 'walis d futuna', 'shi-Tfng': 'ⵡⴰⵍⵉⵙ ⴷ ⴼⵓⵜⵓⵏⴰ', 'si': 'වැලිස් සහ ෆුටුනා', 'sk': 'Wallis a Futuna', 'sl': 'Wallis in Futuna', 'smn': 'Wallis já Futuna', 'sn': 'Wallis and Futuna', 'so': 'Walis & Futuna', 'sq': 'Uollis e Futuna', 'sr': 'Валис и Футуна', 'sr-Cyrl': 'Валис и Футуна', 'sr-Latn': 'Valis i Futuna', 'sv': 'Wallis- och Futunaöarna', 'sw': 'Wallis na Futuna', 'ta': 'வாலிஸ் மற்றும் ஃபுடுனா', 'te': 'వాల్లిస్ మరియు ఫుటునా', 'teo': 'Walis na Futuna', 'tg': 'Уоллис ва Футуна', 'th': 'วาลลิสและฟุตูนา', 'ti': 'ዋሊስን ፉቱናን', 'tk': 'Uollis we Futuna', 'to': 'ʻUvea mo Futuna', 'tr': 'Wallis ve Futuna', 'tt': 'Уоллис һәм Футуна', 'twq': 'Wallis nda Futuna', 'tzm': 'Walis d Futuna', 'ug': 'ۋاللىس ۋە فۇتۇنا', 'uk': 'Уолліс і Футуна', 'ur': 'ویلیز اور فیوٹیونا', 'uz': 'Uollis va Futuna', 'uz-Cyrl': 'Уоллис ва Футуна', 'uz-Latn': 'Uollis va Futuna', 'vai': 'ꕎꔷꔻ ꗪ ꖢꖤꕯ', 'vai-Latn': 'Walísi', 'vai-Vaii': 'ꕎꔷꔻ ꗪ ꖢꖤꕯ', 'vec': 'Wallis e Futuna', 'vi': 'Wallis và Futuna', 'vun': 'Walis na Futuna', 'wae': 'Wallis und Futuna', 'wo': 'Walis ak Futuna', 'xh': 'EWallis & Futuna', 'xnr': 'वालिस कने फ़्यूचूना', 'xog': 'Walisi ni Futuna', 'yav': 'walíis na futúna', 'yo': 'Wali ati futuna', 'yrl': 'Wari asuí Futuna', 'yue': '瓦利斯同富圖納群島', 'yue-Hans': '瓦利斯同富图纳群岛', 'yue-Hant': '瓦利斯同富圖納群島', 'zgh': 'ⵡⴰⵍⵉⵙ ⴷ ⴼⵓⵜⵓⵏⴰ', 'zh': '瓦利斯和富图纳', 'zh-Hans': '瓦利斯和富图纳', 'zh-Hant': '瓦利斯群島和富圖那群島', 'zu': 'i-Wallis ne-Futuna'},
    'WS': {'af': 'Samoa', 'agq': 'Sàmowà', 'ak': 'Samoa', 'am': 'ሳሞአ', 'ar': 'ساموا', 'as': 'চামোৱা', 'asa': 'Thamoa', 'ast': 'Samoa', 'az': 'Samoa', 'az-Cyrl': 'Самоа', 'az-Latn': 'Samoa', 'bal-Latn': 'Samóá', 'bas': 'Sàmoà', 'be': 'Самоа', 'be-tarask': 'Самоа', 'bez': 'Husamoa', 'bg': 'Самоа', 'blo': 'Samowa', 'bm': 'Samowa', 'bn': 'সামোয়া', 'br': 'Samoa', 'brx': 'सामआ', 'bs': 'Samoa', 'bs-Cyrl': 'Самоа', 'bs-Latn': 'Samoa', 'ca': 'Samoa', 'ca-ES-valencia': 'Samoa', 'ccp': '𑄥𑄟𑄮𑄠', 'ce': 'Самоа', 'ceb': 'Samoa', 'cgg': 'Samowa', 'chr': 'ᏌᎼᎠ', 'ckb': 'ساموا', 'cs': 'Samoa', 'cv': 'Самоа', 'cy': 'Samoa', 'da': 'Samoa', 'dav': 'Samoa', 'de': 'Samoa', 'dje': 'Samoa', 'dsb': 'Samoa', 'dz': 'ས་མོ་ཨ', 'ebu': 'Samoa', 'ee': 'Samoa nutome', 'el': 'Σαμόα', 'el-polyton': 'Σαμόα', 'en': 'Samoa', 'en-Dsrt': '𐐝𐐲𐑋𐐬𐐲', 'eo': 'Samoo', 'es': 'Samoa', 'et': 'Samoa', 'eu': 'Samoa', 'ewo': 'Samoá', 'fa': 'ساموآ', 'ff': 'Samowaa', 'ff-Adlm': '𞤅𞤢𞤥𞤵𞤱𞤢', 'ff-Latn': 'Samowaa', 'fi': 'Samoa', 'fil': 'Samoa', 'fo': 'Samoa', 'fr': 'Samoa', 'fur': 'Samoa', 'fy': 'Samoa', 'ga': 'Samó', 'gd': 'Samotha', 'gl': 'Samoa', 'gsw': 'Samooa', 'gu': 'સમોઆ', 'guz': 'Samoa', 'ha': 'Samoa', 'he': 'סמואה', 'hi': 'समोआ', 'hi-Latn': 'Samoa', 'hr': 'Samoa', 'hsb': 'Samoa', 'ht': 'Samoa', 'hu': 'Szamoa', 'hy': 'Սամոա', 'ia': 'Samoa', 'id': 'Samoa', 'ie': 'Samoa', 'ig': 'Samoa', 'is': 'Samóa', 'it': 'Samoa', 'ja': 'サモア', 'jmc': 'Samoa', 'jv': 'Samoa', 'ka': 'სამოა', 'kaa': 'Самоа', 'kaa-Cyrl': 'Самоа', 'kab': 'Samwa', 'kam': 'Samoa', 'kde': 'Samoa', 'kea': 'Samoa', 'kgp': 'Samãva', 'khq': 'Samoa', 'ki': 'Samoa', 'kk': 'Самоа', 'kk-Cyrl': 'Самоа', 'kln': 'Emetab Samoa', 'km': 'សាម័រ', 'kn': 'ಸಮೋವಾ', 'ko': '사모아', 'kok': 'सामोआ', 'kok-Deva': 'सामोआ', 'ks': 'سامو', 'ks-Arab': 'سامو', 'ksb': 'Samoa', 'ksf': 'samɔa', 'ksh': 'Sammoha', 'ku': 'Samoa', 'kxv': 'samoaa', 'kxv-Deva': 'समोआ', 'kxv-Latn': 'samoaa', 'kxv-Orya': 'ସମୱା', 'kxv-Telu': 'సమోవా', 'ky': 'Самоа', 'lag': 'Samóoa', 'lb': 'Samoa', 'lg': 'Samowa', 'lld': 'Samoa', 'ln': 'Samoa', 'lo': 'ຊາມົວ', 'lt': 'Samoa', 'lu': 'Samoa', 'luo': 'Samoa', 'luy': 'Samoa', 'lv': 'Samoa', 'mai': 'समोआ', 'mas': 'Samoa', 'mer': 'Samoa', 'mfe': 'Samoa', 'mg': 'Samoa', 'mgh': 'Usamoa', 'mi': 'Hāmoa', 'mk': 'Самоа', 'ml': 'സമോവ', 'mn': 'Самоа', 'mr': 'सामोआ', 'ms': 'Samoa', 'mt': 'Samoa', 'mua': 'Sǝr Samoa', 'my': 'ဆမိုးအား', 'mzn': 'ساموآ', 'naq': 'Samoa', 'nb': 'Samoa', 'nd': 'Samowa', 'ne': 'सामोआ', 'nl': 'Samoa', 'nmg': 'Samoa', 'nn': 'Samoa', 'no': 'Samoa', 'nqo': 'ߛߊߡߏߥߊ߫', 'nyn': 'Samowa', 'om': 'Saamowa', 'or': 'ସାମୋଆ', 'pa': 'ਸਾਮੋਆ', 'pa-Guru': 'ਸਾਮੋਆ', 'pcm': 'Samóa', 'pl': 'Samoa', 'ps': 'ساماوا', 'pt': 'Samoa', 'qu': 'Samoa', 'rm': 'Samoa', 'rn': 'Samowa', 'ro': 'Samoa', 'rof': 'Samoa', 'ru': 'Самоа', 'rwk': 'Samoa', 'saq': 'Samoa', 'sbp': 'Samoya', 'sc': 'Samoa', 'sd': 'ساموا', 'sd-Arab': 'ساموا', 'se': 'Samoa', 'seh': 'Samoa', 'ses': 'Samoa', 'sg': 'Samoäa', 'shi': 'ⵙⴰⵎⵡⴰ', 'shi-Latn': 'samwa', 'shi-Tfng': 'ⵙⴰⵎⵡⴰ', 'si': 'සැමෝවා', 'sk': 'Samoa', 'sl': 'Samoa', 'smn': 'Samoa', 'sn': 'Samoa', 'so': 'Samoowa', 'sq': 'Samoa', 'sr': 'Самоа', 'sr-Cyrl': 'Самоа', 'sr-Latn': 'Samoa', 'sv': 'Samoa', 'sw': 'Samoa', 'syr': 'ܣܡܘܐ', 'ta': 'சமோவா', 'te': 'సమోవా', 'teo': 'Samoa', 'tg': 'Самоа', 'th': 'ซามัว', 'ti': 'ሳሞኣ', 'tk': 'Samoa', 'to': 'Haʻamoa', 'tr': 'Samoa', 'tt': 'Самоа', 'twq': 'Samoa', 'tzm': 'Samwa', 'ug': 'ساموئا', 'uk': 'Самоа', 'ur': 'ساموآ', 'uz': 'Samoa', 'uz-Cyrl': 'Самоа', 'uz-Latn': 'Samoa', 'vai': 'ꕢꕹꖙꕉ', 'vai-Latn': 'Samowa', 'vai-Vaii': 'ꕢꕹꖙꕉ', 'vec': 'Samòa', 'vi': 'Samoa', 'vun': 'Samoa', 'wae': 'Samoa', 'wo': 'Samowa', 'xh': 'ESamoa', 'xnr': 'समोआ', 'xog': 'Samowa', 'yav': 'samowá', 'yi': 'סאַמאאַ', 'yo': 'Samọ', 'yrl': 'Samowa', 'yue': '薩摩亞', 'yue-Hans': '萨摩亚', 'yue-Hant': '薩摩亞', 'zgh': 'ⵙⴰⵎⵡⴰ', 'zh': '萨摩亚', 'zh-Hans': '萨摩亚', 'zh-Hant': '薩摩亞', 'zu': 'i-Samoa'},
    'Wara': {'ast': 'varang kshiti', 'bal-Latn': 'Warang Kshiti', 'bs': 'varang kshiti pismo', 'bs-Cyrl': 'варанг кшити писмо', 'bs-Latn': 'varang kshiti pismo', 'ca': 'varang kshiti', 'ca-ES-valencia': 'varang kshiti', 'cs': 'varang kšiti', 'da': 'varang kshiti', 'de': 'Varang Kshiti', 'en': 'Varang Kshiti', 'et': 'hoo', 'eu': 'varang kshiti', 'ff-Adlm': '𞤜𞤢𞤪𞤢𞤲𞤺 𞤑𞤭𞥃𞤼𞤭', 'fi': 'varang kshiti', 'fo': 'varang kshiti', 'fy': 'Varang Kshiti', 'ga': 'Varang Kshiti', 'gd': 'Varang Kshiti', 'he': 'ווראנג סיטי', 'hi-Latn': 'Varang Kshiti', 'hr': 'varang kshiti pismo', 'id': 'Varang Kshiti', 'ig': 'Varang Kshiti', 'it': 'varang kshiti', 'ja': 'バラン・クシティ文字', 'ka': 'ვარანგ-კშიტი', 'kk': 'варанг кшити жазуы', 'kk-Cyrl': 'варанг кшити жазуы', 'ko': '바랑 크시티 문자', 'lo': 'ວາຮັງກສິຕິ', 'lt': 'Varang Kshiti', 'mk': 'варанг кшити', 'mn': 'Варанг Кшити', 'ms': 'Varang Kshiti', 'nb': 'varang kshiti', 'nl': 'Varang Kshiti', 'nn': 'varang kshiti', 'no': 'varang kshiti', 'nqo': 'ߥߊߙߊ߲ߜ߭ ߛߕߌ', 'ro': 'varang kshiti', 'ru': 'варанг-кшити', 'sc': 'varang kshiti', 'so': 'Nidaamka Qoraalka Luuqada Ho', 'sq': 'varang kshiti', 'sv': 'varang kshiti', 'tg': 'Варанг Кшитӣ', 'th': 'วารังกสิติ', 'to': 'tohinima fakavalangi-kisitī', 'tr': 'Varang Kshiti', 'ug': 'ۋاراڭ كىشىتى', 'vec': 'warang siti', 'vi': 'Chữ Varang Kshiti', 'yue': '瓦郎奇蒂文字', 'yue-Hans': '瓦郎奇蒂文字', 'yue-Hant': '瓦郎奇蒂文字', 'zh': '瓦郎奇蒂文字', 'zh-Hans': '瓦郎奇蒂文字', 'zh-Hant': '瓦郎奇蒂文字', 'zu': 'i-Varang Kshiti'},
    'Wcho': {'bal-Latn': 'Wanchó', 'bs': 'vančo pismo', 'bs-Cyrl': 'ванчо', 'bs-Latn': 'vančo pismo', 'en': 'Wancho', 'et': 'vantšo', 'eu': 'wanchoera', 'ff-Adlm': '𞤏𞤢𞤲𞤷𞤮𞥅', 'fi': 'wancholainen', 'fo': 'wancho', 'fr': 'wantcho', 'gd': 'Wancho', 'he': 'וונצ׳ו', 'hi-Latn': 'Wancho', 'ht': 'wantcho', 'id': 'Wancho', 'ig': 'Wancho', 'kk': 'ванчо жазуы', 'kk-Cyrl': 'ванчо жазуы', 'km': 'វ៉ាន់ឆូ', 'mn': 'Ванчу', 'ms': 'Wancho', 'nl': 'Wancho', 'nqo': 'ߥߊ߲ߗߏ', 'ro': 'wancho', 'sc': 'wancho', 'so': 'Dadka wanjo', 'sq': 'vanço', 'sv': 'wancho', 'tg': 'Ванчо', 'vec': 'wancho', 'zh': '万秋文', 'zh-Hans': '万秋文', 'zu': 'isi-Wancho'},
    'Wole': {'ast': 'woleai', 'bs': 'woleai pismo', 'bs-Latn': 'woleai pismo', 'ca': 'woleai', 'ca-ES-valencia': 'woleai', 'cs': 'karolínské (woleai)', 'da': 'woleai', 'de': 'Woleaianisch', 'en': 'Woleai', 'et': 'voleai', 'fi': 'woleai', 'fy': 'Woleai', 'gd': 'Woleai', 'hi-Latn': 'Woleai', 'hr': 'woleai pismo', 'id': 'Woleai', 'it': 'woleai', 'ja': 'ウォレアイ文字', 'ka': 'ვოლეაი', 'ko': '울레아이', 'lo': 'ໂອລີເອ', 'lt': 'Woleai', 'mk': 'волеајско', 'nb': 'woleai', 'nl': 'Woleai', 'nn': 'woleai', 'no': 'woleai', 'ru': 'волеаи', 'sv': 'woleai', 'th': 'โอลีเอ', 'to': 'tohinima fakauoleai', 'tr': 'Woleai', 'ug': 'ۋولىئاي', 'vi': 'Chữ Woleai', 'yue': '沃雷艾文', 'yue-Hans': '沃雷艾文', 'yue-Hant': '沃雷艾文', 'zh': '沃莱艾文', 'zh-Hans': '沃莱艾文', 'zh-Hant': '沃雷艾文'},
    'XA': {'af': 'Pseudoaksente', 'ak': 'Anto Kasa', 'am': 'የሀሰት ትእምርት', 'ar': 'لكنات تجريبية غير أصلية', 'as': 'ছুডু-একচেণ্ট', 'az': 'Psevdo-Aksent', 'az-Latn': 'Psevdo-Aksent', 'bal-Latn': 'Kesási-Gálwár', 'be': 'Псеўдаакцэнты', 'be-tarask': 'Псеўдаакцэнты', 'bg': 'Псевдоакценти', 'blo': 'sǝɖoo-aksaŋ', 'bn': 'সিউডো-অ্যাকসেন্ট', 'br': 'pouez-mouezh gaou', 'brx': 'नंखाइ-गारांबोलो', 'bs': 'Pseudo naglasci', 'bs-Cyrl': 'псеудо акценти', 'bs-Latn': 'Pseudo naglasci', 'ca': 'pseudoaccents', 'ca-ES-valencia': 'pseudoaccents', 'ceb': 'Pseudo-Accents', 'chr': 'ᏡᏙ-ᏄᏍᏛᎢᎥᎧᏁᎬᎢ', 'cs': 'simulovaná diakritika', 'cv': 'псевдакцентсем', 'cy': 'Acenion Ffug', 'da': 'pseudo-accenter', 'de': 'Pseudo-Akzente', 'dsb': 'pseudo-akcenty', 'el': 'Ψευδο-προφορές', 'el-polyton': 'Ψευδο-προφορές', 'en': 'Pseudo-Accents', 'es': 'Pseudoacentos', 'et': 'pseudo-aktsent', 'eu': 'Sasiazentuak', 'fa': 'انگلیسی با لهجه خارجی', 'ff-Adlm': '𞤌𞤻𞤵𞤲𞤣𞤫-𞤃𞤢𞤧𞤫', 'fi': 'pseudoaksentit', 'fil': 'Mga Pseudo-Accent', 'fo': 'óekta tónalag', 'fr': 'pseudo-accents', 'ga': 'Bréagaicinn', 'gd': 'Sràcan fuadain', 'gl': 'Pseudoacentos', 'gu': 'સ્યુડો-ઍક્સન્ટ', 'ha': 'Gurɓataccen Karin harshe', 'he': 'מבטאים חלקיים', 'hi': 'सूडो-एक्सेंट', 'hi-Latn': 'Pseudo-Accents', 'hr': 'pseudo naglasci', 'hsb': 'pseudo-akcenty', 'ht': 'pseudo-accents', 'hu': 'Pszeudo-nyelvjárások', 'hy': 'կեղծ տարբերիչներ', 'ia': 'Pseudo-accentos', 'id': 'Aksen Asing', 'ig': 'Pseudo-Accents', 'is': 'gervihreimur', 'it': 'Pseudo-accenti', 'ja': '疑似アクセント', 'jv': 'Logat Semu', 'ka': 'ფსევდო-აქცენტები', 'kaa': 'қалпеки пәт', 'kaa-Cyrl': 'қалпеки пәт', 'kea': 'Pseudo-sotakis', 'kgp': 'Ón vĩ ag', 'kk': 'жалған акцент', 'kk-Cyrl': 'жалған акцент', 'km': 'Pseudo-Accents', 'kn': 'ಸ್ಯುಡೊ ಉಚ್ಛಾರಣೆಗಳು', 'ko': '유사 억양', 'kok': 'स्युडो-ऍक्सेंट', 'kok-Deva': 'स्युडो-ऍक्सेंट', 'ks': 'سوڈو-ایسنٹس', 'ks-Arab': 'سوڈو-ایسنٹس', 'ku': 'Aksanên Psodoyê', 'ky': 'Жасалма- Акцент', 'lld': 'Pseudo-azënc', 'lo': 'ສຳນຽງຊູໂດ', 'lt': 'pseudo A', 'lv': 'pseidoakcenti', 'mai': 'छद्म उच्चारण', 'mi': 'Mita kikoika', 'mk': 'псевдо-акценти', 'ml': 'കൃത്രിമ ഉച്ചാരണം', 'mn': 'хуурмаг аялга', 'mr': 'नकली-उच्चारणे', 'ms': 'Aksen Pseudo', 'my': 'နိုင်ငံခြားသံ', 'nb': 'pseudospråk – aksenter', 'ne': 'सिउडो-एक्सेन्ट्स', 'nl': 'Pseudo-Accenten', 'nn': 'pseudospråk – aksentar', 'no': 'pseudospråk – aksenter', 'nqo': 'ߔߛߏߘߏ߫ ߊߞߑߛߊ߲', 'om': 'Loqoda Sobaa', 'or': 'ସିୟୁଡୋ-ଏସେଣ୍ଟ', 'pa': 'ਗਲਤ-ਉਚਾਰਨ', 'pa-Guru': 'ਗਲਤ-ਉਚਾਰਨ', 'pcm': 'To yúz atifíshál vọis wẹ́n yu de tọk', 'pl': 'Pseudoakcenty', 'ps': 'جعلي خج', 'pt': 'Pseudossotaques', 'qu': 'Seudo-Acentos', 'rm': 'accents pseudo', 'ro': 'Pseudo-accente', 'ru': 'псевдоакценты', 'sc': 'psèudo-atzentos', 'sd': 'سوڊو-لهجا', 'sd-Arab': 'سوڊو-لهجا', 'si': 'ව්\u200dයාජ-උච්චාරණ', 'sk': 'falošná diakritika', 'sl': 'psevdonaglasi', 'so': 'Lahjadaha Pseudo', 'sq': 'Pseudo-thekse', 'sr': 'симулиране дијакритике', 'sr-Cyrl': 'симулиране дијакритике', 'sr-Latn': 'simulirane dijakritike', 'sv': 'fejkade accenter (för test)', 'sw': 'Lafudhi Bandia', 'ta': 'போலி உச்சரிப்புகள்', 'te': 'సూడో యాక్సెంట్స్', 'tg': 'Псевдо-аксентҳо', 'th': 'สำเนียงไม่จริง', 'ti': 'ናይ ሓሶት ላህጃታት', 'tk': 'psewdo-şiweler', 'to': 'fasiʻalea loi', 'tr': 'Psödo Aksanlar', 'tt': 'Псевдоакцентлар', 'uk': 'Псевдоакценти', 'ur': 'بناوٹی لہجے', 'uz': 'Qalbaki urg‘u', 'uz-Latn': 'Qalbaki urg‘u', 'vec': 'Pseudo asenti', 'vi': 'Pseudo-Accents', 'wo': 'Pseudo-aksan', 'xh': 'I-Pseudo-Accents', 'xnr': 'सूडो-एक्सेंट', 'yo': 'ìsọ̀rọ̀sí irọ́', 'yrl': 'Sutakirana-ita', 'yue': '偽口音', 'yue-Hans': '伪口音', 'yue-Hant': '偽口音', 'zh': '伪地区', 'zh-Hans': '伪地区', 'zh-Hant': '偽區域', 'zu': 'Pseudo-Accents'},
    'XA..XZ': {'en': 'Private use'},
    'XB': {'af': 'Pseudotweerigting', 'ak': 'Anto Bidi', 'am': 'የሀሰት ባለሁለት አቅጣጫ', 'ar': 'لكنات تجريبية ثنائية الاتجاه', 'as': 'ছুডু-বিডি', 'az': 'Psevdo-Bidi', 'az-Latn': 'Psevdo-Bidi', 'bal-Latn': 'Kesási- Bidi', 'be': 'Псеўдадвухнапрамкавы', 'be-tarask': 'Псеўдадвухнапрамкавы', 'bg': 'Псевдодвупосочни', 'blo': 'sǝɖoo-biɖi', 'bn': 'সিউডো-বিডি', 'br': 'BiDi gaou', 'brx': 'नंखाइ- बिदि', 'bs': 'Pseudo bidi', 'bs-Cyrl': 'псеудо двосмјерно', 'bs-Latn': 'Pseudo bidi', 'ca': 'pseudobidi', 'ca-ES-valencia': 'pseudobidi', 'ceb': 'Pseudo-Bidi', 'chr': 'ᏡᏙ-ᏈᏗ', 'cs': 'simulovaný obousměrný zápis', 'cv': 'псевд-Bidi', 'cy': 'Bidi Ffug', 'da': 'pseudo-bidi', 'de': 'Pseudo-Bidi', 'dsb': 'pseudo-bidi', 'el': 'Ψευδο-αμφικατευθυντικό', 'el-polyton': 'Ψευδο-αμφικατευθυντικό', 'en': 'Pseudo-Bidi', 'es': 'Pseudobidi', 'et': 'pseudo-Bidi', 'eu': 'Pseudobidia', 'fa': 'مجازی - دوجهته', 'ff-Adlm': '𞤌𞤻𞤵𞤲𞤣𞤫-𞤄𞤭𞤣𞤭', 'fi': 'kaksisuuntainen pseudo', 'fil': 'Pseudo-Bidi', 'fo': 'óektaður BIDI tekstur', 'fr': 'pseudo-bidi', 'ga': 'Bréag-Bidi', 'gd': 'Dà-chomhaireach fuadain', 'gl': 'Pseudobidireccional', 'gu': 'સ્યુડો-દ્વિદિશ', 'ha': 'Gurɓataccen Bidi', 'he': 'דו-כיווני חלקי', 'hi': 'सूडो-बायडायरेक्शनल', 'hi-Latn': 'Pseudo-Bidirectional', 'hr': 'pseudo bidi', 'hsb': 'pseudo-bidi', 'ht': 'pseudo-bidi', 'hu': 'Pszeudo-kétirányú', 'hy': 'կեղծ երկուղղված', 'ia': 'Pseudo-bidi', 'id': 'Pseudo-Bidi', 'ig': 'Pseudo-Bidi', 'is': 'gervistaður', 'it': 'Pseudo-bidi', 'ja': '疑似 BIDI', 'jv': 'Rong Arah Semu', 'ka': 'ფსევდო-ორმიმართულებიანი', 'kaa': 'қалпеки Bidi', 'kaa-Cyrl': 'қалпеки Bidi', 'kea': 'Pseudo-bidiresional', 'kgp': 'Ón régre mĩ.', 'kk': 'жалған Bidi', 'kk-Cyrl': 'жалған Bidi', 'km': 'Pseudo-Bidi', 'kn': 'ಸ್ಯುಡೊ-ಬಿಡಿ', 'ko': '유사 양방향', 'kok': 'स्युडो-बिडी', 'kok-Deva': 'स्युडो-बिडी', 'ks': 'سوڈو-بیڈی', 'ks-Arab': 'سوڈو-بیڈی', 'ku': 'Psodo Bidî', 'ky': 'Жасалма - эки тараптуу', 'lld': 'Pseudo-bidi', 'lo': 'ຊູໂດບີດີ', 'lt': 'pseudo B', 'lv': 'pseidodivvirzienu', 'mai': 'छद्म बिडि', 'mi': 'Piri Kikoika', 'mk': 'псевдо-биди', 'ml': 'സ്യൂഡോ-ബൈഡി', 'mn': 'хуурмаг хоёр чиглэлт', 'mr': 'नकली-बीडी', 'ms': 'Bidi Pseudo', 'my': 'စာပြောင်းပြန်', 'nb': 'pseudospråk – tekst begge veier', 'ne': 'सिउडो-बिडी', 'nl': 'Pseudo-Bidi', 'nn': 'pseudospråk – RTL', 'no': 'pseudospråk – tekst begge veier', 'nqo': 'ߔߛߔߘߏ߫-ߓߘߌ߫', 'om': 'Biidii Sobaa', 'or': 'ସିୟୁଡୋ-ବିଡି', 'pa': 'ਲਿਖਤ ਦੀ ਗਲਤ ਦਿਸ਼ਾ', 'pa-Guru': 'ਲਿਖਤ ਦੀ ਗਲਤ ਦਿਸ਼ਾ', 'pcm': 'Atífíshál Tú-Wé Dairẹ́kshọn', 'pl': 'Pseudodwukierunkowe', 'ps': 'سیډو بیډی', 'pt': 'Pseudobidirecional', 'qu': 'Seudo-Bidi', 'rm': 'pseudo-bidirecziunal', 'ro': 'Pseudo-Bidi', 'ru': 'псевдо-Bidi', 'sc': 'psèudo-bidi', 'sd': 'سوڊو-بي ڊي', 'sd-Arab': 'سوڊو-بي ڊي', 'si': 'ව්\u200dයාජ-බිඩි', 'sk': 'obrátenie sprava doľava', 'sl': 'psevdobidi', 'so': 'Pseudo-Bidi', 'sq': 'Pseudo-bidi', 'sr': 'Псеудобиди', 'sr-Cyrl': 'Псеудобиди', 'sr-Latn': 'Pseudobidi', 'sv': 'fejkad bidi (för test)', 'sw': 'Lafudhi Bidi', 'ta': 'போலி பீடி', 'te': 'సూడో-బిడి', 'tg': 'Псевдо-бидӣ', 'th': 'Bidi ไม่จริง', 'ti': 'ናይ ሓሶት ክልተ ኣንፈታዊ', 'tk': 'psewdo-bidi', 'to': 'fua-ua loi', 'tr': 'Psödo Bidi', 'tt': 'Псевдо-Биди', 'uk': 'Псевдобіді', 'ur': 'بناوٹی دو سمتی', 'uz': 'Qalbaki Bidi', 'uz-Latn': 'Qalbaki Bidi', 'vec': 'Pseudo bidi', 'vi': 'Pseudo-Bidi', 'wo': 'Pseudo-bidi', 'xh': 'I-Pseudo-Bidi', 'xnr': 'सूडो-बायडायरेक्शनल', 'yo': 'Agbègbè irọ́', 'yrl': 'Bidiresiunaurana', 'yue': '偽 Bidi', 'yue-Hans': '伪 Bidi', 'yue-Hant': '偽 Bidi', 'zh': '伪双向语言地区', 'zh-Hans': '伪双向语言地区', 'zh-Hant': '偽比迪', 'zu': 'Pseudo-Bidi'},
    'XK': {'af': 'Kosovo', 'ak': 'Kosovo', 'am': 'ኮሶቮ', 'ar': 'كوسوفو', 'as': 'কচ’ভ’', 'ast': 'Kosovu', 'az': 'Kosovo', 'az-Cyrl': 'Косово', 'az-Latn': 'Kosovo', 'bal-Latn': 'Kósówó', 'be': 'Косава', 'be-tarask': 'Косава', 'bg': 'Косово', 'blo': 'Kɔsofoo', 'bn': 'কসোভো', 'br': 'Kosovo', 'brx': 'कस’भ’', 'bs': 'Kosovo', 'bs-Cyrl': 'Косово', 'bs-Latn': 'Kosovo', 'ca': 'Kosovo', 'ca-ES-valencia': 'Kosovo', 'ccp': '𑄇𑄧𑄥𑄮𑄞𑄮', 'ce': 'Косово', 'ceb': 'Kosovo', 'chr': 'ᎪᏐᏉ', 'ckb': 'کۆسۆڤۆ', 'cs': 'Kosovo', 'cv': 'Косово', 'cy': 'Kosovo', 'da': 'Kosovo', 'de': 'Kosovo', 'dsb': 'Kosowo', 'el': 'Κοσσυφοπέδιο', 'el-polyton': 'Κοσσυφοπέδιο', 'en': 'Kosovo', 'es': 'Kosovo', 'et': 'Kosovo', 'eu': 'Kosovo', 'fa': 'کوزوو', 'ff-Adlm': '𞤑𞤮𞥅𞤧𞤮𞤾𞤮𞥅', 'fi': 'Kosovo', 'fil': 'Kosovo', 'fo': 'Kosovo', 'fr': 'Kosovo', 'fy': 'Kosovo', 'ga': 'an Chosaiv', 'gd': 'A’ Chosobho', 'gl': 'Kosovo', 'gu': 'કોસોવો', 'ha': 'Kasar Kosovo', 'he': 'קוסובו', 'hi': 'कोसोवो', 'hi-Latn': 'Kosovo', 'hr': 'Kosovo', 'hsb': 'Kosowo', 'ht': 'Kosovo', 'hu': 'Koszovó', 'hy': 'Կոսովո', 'ia': 'Kosovo', 'id': 'Kosovo', 'ie': 'Kosovo', 'ig': 'Kosovo', 'is': 'Kósóvó', 'it': 'Kosovo', 'ja': 'コソボ', 'jv': 'Kosovo', 'ka': 'კოსოვო', 'kaa': 'Косово', 'kaa-Cyrl': 'Косово', 'kea': 'Kozovu', 'kgp': 'Kosovo', 'kk': 'Косово', 'kk-Cyrl': 'Косово', 'km': 'កូសូវ៉ូ', 'kn': 'ಕೊಸೊವೊ', 'ko': '코소보', 'kok': 'कोसोवो', 'kok-Deva': 'कोसोवो', 'ks': 'کوسوو', 'ks-Arab': 'کوسوو', 'ksh': 'Kosovo', 'ku': 'Kosova', 'kxv': 'kosovo', 'kxv-Deva': 'कोसोवो', 'kxv-Latn': 'kosovo', 'kxv-Orya': 'କସୱ', 'kxv-Telu': 'కొసోవో', 'ky': 'Косово', 'lb': 'Kosovo', 'lld': 'Kosovo', 'lo': 'ໂຄໂຊໂວ', 'lt': 'Kosovas', 'lv': 'Kosova', 'mai': 'कोसोवो', 'mi': 'Kōhoro', 'mk': 'Косово', 'ml': 'കൊസോവൊ', 'mn': 'Косово', 'mr': 'कोसोव्हो', 'ms': 'Kosovo', 'mt': 'il-Kosovo', 'my': 'ကိုဆိုဗို', 'mzn': 'کوزوو', 'nb': 'Kosovo', 'ne': 'कोसोभो', 'nl': 'Kosovo', 'nn': 'Kosovo', 'no': 'Kosovo', 'nqo': 'ߞߛߏߝ߭ߏ߫', 'om': 'Kosoovoo', 'or': 'କୋସୋଭୋ', 'pa': 'ਕੋਸੋਵੋ', 'pa-Guru': 'ਕੋਸੋਵੋ', 'pcm': 'Kósóvo', 'pl': 'Kosowo', 'ps': 'کوسوو', 'pt': 'Kosovo', 'qu': 'Kosovo', 'rm': 'Cosovo', 'ro': 'Kosovo', 'ru': 'Косово', 'sc': 'Kòssovo', 'sd': 'ڪوسووو', 'sd-Arab': 'ڪوسووو', 'se': 'Kosovo', 'si': 'කොසෝවෝ', 'sk': 'Kosovo', 'sl': 'Kosovo', 'smn': 'Kosovo', 'so': 'Koosofo', 'sq': 'Kosovë', 'sr': 'Косово', 'sr-Cyrl': 'Косово', 'sr-Latn': 'Kosovo', 'sv': 'Kosovo', 'sw': 'Kosovo', 'syr': 'ܩܘܣܘܒܘ', 'ta': 'கொசோவோ', 'te': 'కొసోవో', 'tg': 'Косово', 'th': 'โคโซโว', 'ti': 'ኮሶቮ', 'tk': 'Kosowo', 'to': 'Kōsovo', 'tr': 'Kosova', 'tt': 'Косово', 'ug': 'كوسوۋو', 'uk': 'Косово', 'ur': 'کوسووو', 'uz': 'Kosovo', 'uz-Cyrl': 'Косово', 'uz-Latn': 'Kosovo', 'vai': 'ꖏꖇꕾ', 'vai-Vaii': 'ꖏꖇꕾ', 'vec': 'Kòsovo', 'vi': 'Kosovo', 'wo': 'Kosowo', 'xh': 'EKosovo', 'xnr': 'कोसोवो', 'yi': 'קאסאווא', 'yo': 'Kòsófò', 'yrl': 'Kusuwu', 'yue': '科索沃', 'yue-Hans': '科索沃', 'yue-Hant': '科索沃', 'zh': '科索沃', 'zh-Hans': '科索沃', 'zh-Hant': '科索沃', 'zu': 'i-Kosovo'},
    'Xpeo': {'ar': 'الفارسية القديمة', 'ast': 'persa antiguu', 'az': 'qədimi fars', 'az-Latn': 'qədimi fars', 'bal-Latn': 'Ahdi Pársi', 'bg': 'Староперсийска', 'bn': 'প্রাচীন ফার্সি', 'br': 'persek kozh', 'brx': 'पुरानी फारसी', 'bs': 'staropersijsko pismo', 'bs-Cyrl': 'староперсијско писмо', 'bs-Latn': 'staropersijsko pismo', 'ca': 'persa antic', 'ca-ES-valencia': 'persa antic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄴 𑄜𑄢𑄴𑄥𑄨', 'cs': 'staroperské klínové písmo', 'da': 'oldpersisk', 'de': 'Altpersisch', 'el': 'Παλαιό Περσικό', 'el-polyton': 'Παλαιό Περσικό', 'en': 'Old Persian', 'en-Dsrt': '𐐄𐑊𐐼 𐐑𐐲𐑉𐑈𐐲𐑌', 'es': 'persa antiguo', 'et': 'vanapärsia', 'eu': 'pertsiera zaharra', 'fa': 'فارسی باستان', 'ff-Adlm': '𞤊𞤢𞥄𞤪𞤭𞤧𞤭𞤴𞤢𞤲𞤳𞤮 𞤑𞤭𞤯𞥆𞤵𞤲', 'fi': 'muinaispersialainen', 'fr': 'cunéiforme persépolitain', 'fur': 'vieri persian', 'fy': 'Aldperzysk', 'ga': 'Sean-Pheirseach', 'gd': 'Seann-Pheirsis', 'gsw': 'Altpersisch', 'gu': 'જુની ફારસી', 'he': 'פרסי עתיק', 'hi': 'पुरानी फारसी', 'hi-Latn': 'Old Persian', 'hr': 'staro perzijsko pismo', 'ht': 'cunéiforme persépolitain', 'hu': 'Óperzsa', 'id': 'Persia Kuno', 'ig': 'Old Persian', 'it': 'persiano antico', 'ja': '古代ペルシア文字', 'ka': 'ძველი სპარსული', 'kgp': 'pérsa si', 'kk': 'көне парсы жазуы', 'kk-Cyrl': 'көне парсы жазуы', 'kn': 'ಪ್ರಾಚೀನ ಪರ್ಶಿಯನ್', 'ko': '고대 페르시아 문자', 'ks': 'پرون فارسی', 'ks-Arab': 'پرون فارسی', 'lb': 'Alpersesch', 'lo': 'ເປຮເຊຍໂບຮານ', 'lt': 'senieji persų', 'lv': 'senperiešu', 'mk': 'староперсиско', 'ml': 'പഴയ പേർഷ്യൻ', 'mn': 'Хуучин перс', 'mr': 'पुरातन फारसी', 'ms': 'Parsi Lama', 'mt': 'Persjan Antik', 'my': 'ပါရှန် အဟောင်း', 'nb': 'gammelpersisk', 'ne': 'पुरानो पर्सियन', 'nl': 'Oudperzisch', 'nn': 'gammalpersisk', 'no': 'gammelpersisk', 'nqo': 'ߝߊ߯ߙߛߌ߫ ߞߘߐߡߊ߲', 'or': 'ପୁରୁଣା ଫରାସୀ', 'pl': 'staroperskie', 'pt': 'persa antigo', 'rm': 'persian vegl', 'ro': 'persană veche', 'ru': 'староперсидская', 'sc': 'persianu antigu', 'sd': 'قديم فارسي', 'sd-Arab': 'قديم فارسي', 'sl': 'staroperzijski', 'so': 'Faarsigii Hore', 'sq': 'persian i vjetër', 'sr': 'староперсијско писмо', 'sr-Cyrl': 'староперсијско писмо', 'sr-Latn': 'staropersijsko pismo', 'sv': 'fornpersiska', 'ta': 'பழைய பெர்ஷியன்', 'te': 'ప్రాచీన పర్షియన్', 'tg': 'Форсии Куҳна', 'th': 'เปอร์เซียโบราณ', 'to': 'tohinima fakapēsiamuʻa', 'tr': 'Eski Fars', 'ug': 'قەدىمكى پارىسچە', 'uk': 'давньоперський', 'vec': 'persian antigo', 'vi': 'Chữ Ba Tư cổ', 'xnr': 'पुरानी फारसी', 'yrl': 'perisa kuxiímawara', 'yue': '古波斯文', 'yue-Hans': '古波斯文', 'yue-Hant': '古波斯文', 'zh': '古波斯文', 'zh-Hans': '古波斯文', 'zh-Hant': '古波斯文', 'zu': 'i-Old Persian'},
    'Xsux': {'ar': 'الكتابة المسمارية الأكدية السومرية', 'ast': 'cuneiforme sumeriu acadiu', 'az': 'sumer-akadyan kuneyform', 'az-Latn': 'sumer-akadyan kuneyform', 'bal-Latn': 'Sumer-Akádi Kyunipárm', 'bg': 'Шумеро-акадски клинопис', 'bn': 'সুমের-আক্কাদীয় কীলকরূপ', 'br': 'gennheñvel', 'brx': 'सुमेरो अक्काड़ी कुनेईफॉर्म', 'bs': 'sumersko-akadsko kuneiform pismo', 'bs-Cyrl': 'сумерско-акадско кунеиформ писмо', 'bs-Latn': 'sumersko-akadsko kuneiform pismo', 'ca': 'cuneïforme sumeri-accadi', 'ca-ES-valencia': 'cuneïforme sumeri-accadi', 'ccp': '𑄥𑄪𑄟𑄬𑄢𑄧-𑄃𑄇𑄳𑄇𑄘𑄩𑄠𑄧 𑄇𑄩𑄣𑄧𑄇𑄴𑄢𑄪𑄛𑄴', 'cs': 'sumero-akkadské klínové písmo', 'da': 'sumero-akkadisk cuneiform', 'de': 'Sumerisch-akkadische Keilschrift', 'el': 'Σούμερο-Ακάντιαν Κουνεϊφόρμ', 'el-polyton': 'Σούμερο-Ακάντιαν Κουνεϊφόρμ', 'en': 'Sumero-Akkadian Cuneiform', 'en-Dsrt': '𐐝𐐭𐑋𐐯𐑉𐐬-𐐊𐐿𐐩𐐼𐐨𐐲𐑌 𐐗𐐷𐐭𐑌𐐨𐐲𐑁𐐱𐑉𐑋', 'es': 'cuneiforme sumerio-acadio', 'et': 'sumeri-akadi kiilkiri', 'eu': 'sumero-akadiera kuneiformea', 'fa': 'میخی سومری‐اکدی', 'ff-Adlm': '𞤅𞤵𞤥𞤫𞤪𞤮 𞤀𞤳𞥆𞤢𞤣𞤭𞤴𞤢𞤲 𞤑𞤵𞤲𞤫𞤬𞤮𞤪𞤥', 'fi': 'sumerilais-akkadilainen nuolenpääkirjoitus', 'fo': 'sumero-akkadian cuneiform', 'fr': 'cunéiforme suméro-akkadien', 'fur': 'cuneiform sumeric-acadic', 'fy': 'Sumero-Akkadian Cuneiform', 'ga': 'Dingchruthach Suiméar-Acádach', 'gd': 'Gèinn-sgrìobhadh Sumer is Akkad', 'gsw': 'Sumerisch-akkadischi Keilschrift', 'gu': 'સુમેરો અક્કાદિયન સુનિફોર્મ', 'he': 'כתב היתדות השומרי-אכדי', 'hi': 'सुमेरो अक्कादियन सुनिफॉर्म', 'hi-Latn': 'Sumero-Akkadian Cuneiform', 'hr': 'sumersko-akadsko cuneiform pismo', 'ht': 'cunéiforme suméro-akkadien', 'hu': 'Ékírásos suméro-akkád', 'id': 'Cuneiform Sumero-Akkadia', 'ig': 'Sumero-Akkadian Cuneiform', 'it': 'sumero-accadiano cuneiforme', 'ja': 'シュメール＝アッカド語楔形文字', 'ka': 'შუმერულ-აქადური ლურსმნული', 'kgp': 'sumẽrijo-akanhỹnũ kafén ja', 'kk': 'шумер-аккад сына жазуы', 'kk-Cyrl': 'шумер-аккад сына жазуы', 'kn': 'ಸುಮೇರೋ-ಅಕ್ಕಾಡಿಯನ್ ಕ್ಯೂನಿಫಾರ್ಮ್', 'ko': '수메르-아카드어 설형문자', 'ks': 'سُمیرو اکادیَن کوٗنِفام', 'ks-Arab': 'سُمیرو اکادیَن کوٗنِفام', 'lb': 'Sumeresch-akkadesch Keilschrëft', 'lo': 'ອັກສອນຮູບປລີ່ມສຸເມເຮຍ-ອັດຄາເດຍ', 'lt': 'Šumero Akado dantiraštis', 'lv': 'šumeru-akadiešu ķīļraksts', 'mk': 'сумероакадско клинесто', 'ml': 'സുമേറോ അക്കാഡിയൻ ക്യുണിഫോം', 'mn': 'Сумеро-Аккадиан шаантаг бичиг', 'mr': 'दृश्यमान भाषा', 'ms': 'Aksara Paku Sumero-Akkadia', 'nb': 'sumersk-akkadisk kileskrift', 'nl': 'Sumero-Akkadian Cuneiform', 'nn': 'sumero-akkadisk kileskrift', 'no': 'sumersk-akkadisk kileskrift', 'nqo': 'ߛߎߡߋߙߏ-ߊߞߊߘߌ', 'or': 'ସୁମେରୋ-ଆକ୍କାଡିଆନ୍ ସୁନିଫର୍ମ', 'pl': 'klinowe sumero-akadyjskie', 'pt': 'sumério-acadiano cuneiforme', 'rm': 'scrittira a cugn sumeric-accadica', 'ro': 'cuneiformă sumero-akkadiană', 'ru': 'шумеро-аккадская клинопись', 'sc': 'cuneiforme sumero-acàdicu', 'sl': 'sumersko-akadski klinopis', 'so': 'Qoraalkii Hore ee dadka Sumaariyiinta ee dhulka mesobataamiya', 'sq': 'kuneiform sumero-akadian', 'sr': 'сумерско-акадско кунеиформ писмо', 'sr-Cyrl': 'сумерско-акадско кунеиформ писмо', 'sr-Latn': 'sumersko-akadsko kuneiform pismo', 'sv': 'sumero-akkadisk kilskrift', 'ta': 'சுமெரோ-அக்கடியன் க்யூனிஃபார்ம்', 'te': 'సుమేరో- అక్కడియన్ క్యునిఫార్మ్', 'tg': 'Хати Сумеро-Аккадӣ', 'th': 'อักษรรูปลิ่มสุเมเรีย-อัคคาเดีย', 'to': 'tohinima fakamataʻingahau-sumelo-akatia', 'tr': 'Sümer-Akad Çivi Yazısı', 'ug': 'سۇمېر-ئاككادىيان مىخ خەت', 'uk': 'шумеро-аккадський клінопис', 'vec': 'cugniforme sumero acadian', 'vi': 'Chữ hình nêm Sumero-Akkadian', 'xnr': 'सुमेरो अक्कादियन सुनिफॉर्म', 'yrl': 'sumériu-akadianu kune-sãgawa', 'yue': '蘇米魯亞甲文楔形文字', 'yue-Hans': '苏米鲁亚甲文楔形文字', 'yue-Hant': '蘇米魯亞甲文楔形文字', 'zh': '苏美尔-阿卡德楔形文字', 'zh-Hans': '苏美尔-阿卡德楔形文字', 'zh-Hant': '蘇米魯亞甲文楔形文字', 'zu': 'i-Sumero-Akkadian Cuneiform'},
    'YD': {'en': 'Democratic Yemen'},
    'YE': {'af': 'Jemen', 'agq': 'Yɛmɛ̀n', 'ak': 'Yɛmɛn', 'am': 'የመን', 'ar': 'اليمن', 'as': 'য়েমেন', 'asa': 'Yemeni', 'ast': 'Yemen', 'az': 'Yəmən', 'az-Cyrl': 'Јәмән', 'az-Latn': 'Yəmən', 'bal-Latn': 'Yaman', 'bas': 'Yèmɛ̂n', 'be': 'Емен', 'be-tarask': 'Емен', 'bez': 'Huyemeni', 'bg': 'Йемен', 'bgn': 'یمن', 'blo': 'Yemɛn', 'bm': 'Yemɛni', 'bn': 'ইয়েমেন', 'br': 'Yemen', 'brx': 'येमेन', 'bs': 'Jemen', 'bs-Cyrl': 'Јемен', 'bs-Latn': 'Jemen', 'ca': 'Iemen', 'ca-ES-valencia': 'Iemen', 'ccp': '𑄃𑄨𑄠𑄬𑄟𑄬𑄚𑄴', 'ce': 'Йемен', 'ceb': 'Yemen', 'cgg': 'Yemeni', 'chr': 'ᏰᎺᏂ', 'ckb': 'یەمەن', 'cs': 'Jemen', 'cv': 'Йемен', 'cy': 'Yemen', 'da': 'Yemen', 'dav': 'Yemeni', 'de': 'Jemen', 'dje': 'Yaman', 'dsb': 'Jemen', 'dz': 'ཡེ་མེན', 'ebu': 'Yemeni', 'ee': 'Yemen nutome', 'el': 'Υεμένη', 'el-polyton': 'Ὑεμένη', 'en': 'Yemen', 'en-Dsrt': '𐐏𐐯𐑋𐐲𐑌', 'eo': 'Jemeno', 'es': 'Yemen', 'et': 'Jeemen', 'eu': 'Yemen', 'ewo': 'Yemɛ́n', 'fa': 'یمن', 'ff': 'Yemen', 'ff-Adlm': '𞤒𞤢𞤥𞤢𞤲', 'ff-Latn': 'Yemen', 'fi': 'Jemen', 'fil': 'Yemen', 'fo': 'Jemen', 'fr': 'Yémen', 'fur': 'Yemen', 'fy': 'Jemen', 'ga': 'Éimin', 'gd': 'An Eaman', 'gl': 'Iemen', 'gsw': 'Jeeme', 'gu': 'યમન', 'guz': 'Yemeni', 'ha': 'Yamen', 'he': 'תימן', 'hi': 'यमन', 'hi-Latn': 'Yemen', 'hr': 'Jemen', 'hsb': 'Jemen', 'ht': 'Yémen', 'hu': 'Jemen', 'hy': 'Եմեն', 'ia': 'Yemen', 'id': 'Yaman', 'ig': 'Yemen', 'is': 'Jemen', 'it': 'Yemen', 'ja': 'イエメン', 'jmc': 'Yemeni', 'jv': 'Yaman', 'ka': 'იემენი', 'kaa': 'Йемен', 'kaa-Cyrl': 'Йемен', 'kab': 'Lyamen', 'kam': 'Yemeni', 'kde': 'Yemeni', 'kea': 'Iémen', 'kgp': 'Jemẽn', 'khq': 'Yaman', 'ki': 'Yemeni', 'kk': 'Йемен', 'kk-Cyrl': 'Йемен', 'kln': 'Emetab Yemen', 'km': 'យេម៉ែន', 'kn': 'ಯೆಮನ್', 'ko': '예멘', 'kok': 'येमेन', 'kok-Deva': 'येमेन', 'ks': 'یَمَن', 'ks-Arab': 'یَمَن', 'ksb': 'Yemeni', 'ksf': 'yɛmɛn', 'ksh': 'Jämme', 'ku': 'Yemen', 'kxv': 'yemen', 'kxv-Deva': 'यमन', 'kxv-Latn': 'yemen', 'kxv-Orya': 'ୟମନ', 'kxv-Telu': 'యెమెన్', 'ky': 'Йемен', 'lag': 'Yémeni', 'lb': 'Jemen', 'lg': 'Yemeni', 'lld': 'Yemen', 'ln': 'Yemɛnɛ', 'lo': 'ເຢເມນ', 'lt': 'Jemenas', 'lu': 'Yemenu', 'luo': 'Yemen', 'luy': 'Yemeni', 'lv': 'Jemena', 'mai': 'यमन', 'mas': 'Yemeni', 'mer': 'Yemeni', 'mfe': 'Yemenn', 'mg': 'Yemen', 'mgh': 'Uyemeni', 'mi': 'Īmene', 'mk': 'Јемен', 'ml': 'യെമൻ', 'mn': 'Йемен', 'mr': 'येमेन', 'ms': 'Yaman', 'mt': 'il-Jemen', 'mua': 'Yemeŋ', 'my': 'ယီမင်', 'mzn': 'یمن', 'naq': 'Yemen', 'nb': 'Jemen', 'nd': 'Yemeni', 'ne': 'येमेन', 'nl': 'Jemen', 'nmg': 'Yǝmɛn', 'nn': 'Jemen', 'no': 'Jemen', 'nqo': 'ߦߡߊߣߌ߲߫', 'nyn': 'Yemeni', 'om': 'Yemen', 'or': 'ୟେମେନ୍', 'pa': 'ਯਮਨ', 'pa-Guru': 'ਯਮਨ', 'pcm': 'Yẹ́mẹn', 'pl': 'Jemen', 'ps': 'یمن', 'pt': 'Iêmen', 'qu': 'Yemen', 'rm': 'Jemen', 'rn': 'Yemeni', 'ro': 'Yemen', 'rof': 'Yemeni', 'ru': 'Йемен', 'rwk': 'Yemeni', 'saq': 'Yemeni', 'sbp': 'Yemeni', 'sc': 'Yemen', 'sd': 'يمن', 'sd-Arab': 'يمن', 'se': 'Jemen', 'seh': 'Iêmen', 'ses': 'Yaman', 'sg': 'Yëmêni', 'shi': 'ⵢⴰⵎⴰⵏ', 'shi-Latn': 'yaman', 'shi-Tfng': 'ⵢⴰⵎⴰⵏ', 'si': 'යේමනය', 'sk': 'Jemen', 'sl': 'Jemen', 'smn': 'Jemen', 'sn': 'Yemen', 'so': 'Yaman', 'sq': 'Jemen', 'sr': 'Јемен', 'sr-Cyrl': 'Јемен', 'sr-Latn': 'Jemen', 'sv': 'Jemen', 'sw': 'Yemeni', 'syr': 'ܝܡܢ', 'ta': 'ஏமன்', 'te': 'యెమెన్', 'teo': 'Yemeni', 'tg': 'Яман', 'th': 'เยเมน', 'ti': 'የመን', 'tk': 'Ýemen', 'to': 'Iemeni', 'tr': 'Yemen', 'tt': 'Йәмән', 'twq': 'Yaman', 'tzm': 'Yaman', 'ug': 'يەمەن', 'uk': 'Ємен', 'ur': 'یمن', 'uz': 'Yaman', 'uz-Cyrl': 'Яман', 'uz-Latn': 'Yaman', 'vai': 'ꔝꘈꘋ', 'vai-Latn': 'Yemɛni', 'vai-Vaii': 'ꔝꘈꘋ', 'vec': 'Yemen', 'vi': 'Yemen', 'vun': 'Yemeni', 'wae': 'Jéme', 'wo': 'Yaman', 'xh': 'EYemen', 'xnr': 'यमन', 'xog': 'Yemeni', 'yav': 'yémɛn', 'yi': 'תימן', 'yo': 'Yemeni', 'yrl': 'Yemẽ', 'yue': '也門', 'yue-Hans': '叶门', 'yue-Hant': '也門', 'zgh': 'ⵢⴰⵎⴰⵏ', 'zh': '也门', 'zh-Hans': '也门', 'zh-Hant': '葉門', 'zu': 'i-Yemen'},
    'YT': {'af': 'Mayotte', 'agq': 'Màyotì', 'ak': 'Mayɔte', 'am': 'ሜይኦቴ', 'ar': 'مايوت', 'as': 'মায়োট্টে', 'asa': 'Mayotte', 'ast': 'Mayotte', 'az': 'Mayot', 'az-Cyrl': 'Мајот', 'az-Latn': 'Mayot', 'bal-Latn': 'Mayotte', 'bas': 'Màyɔ̂t', 'be': 'Маёта', 'be-tarask': 'Маёта', 'bez': 'Humayotte', 'bg': 'Майот', 'blo': 'Mayɔɔtɩ', 'bm': 'Mayoti', 'bn': 'মায়োত্তে', 'br': 'Mayotte', 'brx': 'मायत्ते', 'bs': 'Majote', 'bs-Cyrl': 'Мајоте', 'bs-Latn': 'Majote', 'ca': 'Mayotte', 'ca-ES-valencia': 'Mayotte', 'ccp': '𑄟𑄠𑄮𑄖𑄴𑄖𑄬', 'ce': 'Майотта', 'ceb': 'Mayotte', 'cgg': 'Mayote', 'chr': 'ᎺᏯᏖ', 'ckb': 'مایۆت', 'cs': 'Mayotte', 'cv': 'Майотта', 'cy': 'Mayotte', 'da': 'Mayotte', 'dav': 'Mayotte', 'de': 'Mayotte', 'dje': 'Mayooti', 'dsb': 'Mayotte', 'dz': 'མེ་ཡོཊ', 'ebu': 'Mayotte', 'ee': 'Mayotte nutome', 'el': 'Μαγιότ', 'el-polyton': 'Μαγιότ', 'en': 'Mayotte', 'en-Dsrt': '𐐣𐐪𐐷𐐱𐐻', 'eo': 'Majoto', 'es': 'Mayotte', 'et': 'Mayotte', 'eu': 'Mayotte', 'ewo': 'Mayɔ́d', 'fa': 'مایوت', 'ff': 'Mayoot', 'ff-Adlm': '𞤃𞤢𞤴𞤮𞥅𞤼𞤵', 'ff-Latn': 'Mayoot', 'fi': 'Mayotte', 'fil': 'Mayotte', 'fo': 'Mayotte', 'fr': 'Mayotte', 'fur': 'Mayotte', 'fy': 'Mayotte', 'ga': 'Mayotte', 'gd': 'Mayotte', 'gl': 'Mayotte', 'gsw': 'Majott', 'gu': 'મેયોટ', 'guz': 'Mayotte', 'ha': 'Mayoti', 'he': 'מאיוט', 'hi': 'मायोते', 'hi-Latn': 'Mayotte', 'hr': 'Mayotte', 'hsb': 'Mayotte', 'ht': 'Mayotte', 'hu': 'Mayotte', 'hy': 'Մայոտ', 'ia': 'Mayotte', 'id': 'Mayotte', 'ig': 'Mayotte', 'is': 'Mayotte', 'it': 'Mayotte', 'ja': 'マヨット', 'jgo': 'Mayɔ̂t', 'jmc': 'Mayotte', 'jv': 'Mayotte', 'ka': 'მაიოტა', 'kaa': 'Майотта', 'kaa-Cyrl': 'Майотта', 'kab': 'Mayuṭ', 'kam': 'Mayotte', 'kde': 'Maole', 'kea': 'Maiote', 'kgp': 'Mỹjóte', 'khq': 'Mayooti', 'ki': 'Mayotte', 'kk': 'Майотта', 'kk-Cyrl': 'Майотта', 'kln': 'Emetab Mayotte', 'km': 'ម៉ាយុត', 'kn': 'ಮಯೊಟ್ಟೆ', 'ko': '마요트', 'kok': 'मेयोट', 'kok-Deva': 'मेयोट', 'ks': 'مَییٹ', 'ks-Arab': 'مَییٹ', 'ksb': 'Mayotte', 'ksf': 'mayɔ́t', 'ksh': 'Majott', 'ku': 'Mayotte', 'kxv': 'mayoṭṭ', 'kxv-Deva': 'मायोते', 'kxv-Latn': 'mayoṭṭ', 'kxv-Orya': 'ମାୟତେ', 'kxv-Telu': 'మాయొట్', 'ky': 'Майотта', 'lag': 'Mayóote', 'lb': 'Mayotte', 'lg': 'Mayotte', 'lld': 'Mayotte', 'ln': 'Mayotɛ', 'lo': 'ມາຢັອດ', 'lt': 'Majotas', 'lu': 'Mayote', 'luo': 'Mayotte', 'luy': 'Mayotte', 'lv': 'Majota', 'mai': 'मयोटे', 'mas': 'Mayotte', 'mer': 'Mayotte', 'mfe': 'Mayot', 'mg': 'Mayôty', 'mi': 'Māiota', 'mk': 'Мајот', 'ml': 'മയോട്ടി', 'mn': 'Майотта', 'mr': 'मायोट्टे', 'ms': 'Mayotte', 'mt': 'Mayotte', 'mua': 'Mayot', 'my': 'မေယော့', 'mzn': 'مایوت', 'naq': 'Mayotte', 'nb': 'Mayotte', 'nd': 'Mayotte', 'ne': 'मायोट्ट', 'nl': 'Mayotte', 'nmg': 'Mayɔt', 'nn': 'Mayotte', 'no': 'Mayotte', 'nqo': 'ߡߊߦߐߕ', 'nyn': 'Mayote', 'om': 'Maayootee', 'or': 'ମାୟୋଟେ', 'pa': 'ਮਾਯੋਟੀ', 'pa-Guru': 'ਮਾਯੋਟੀ', 'pcm': 'Meyọt', 'pl': 'Majotta', 'ps': 'مايوټ', 'pt': 'Mayotte', 'qu': 'Mayotte', 'rm': 'Mayotte', 'rn': 'Mayote', 'ro': 'Mayotte', 'rof': 'Mayotte', 'ru': 'Майотта', 'rwk': 'Mayotte', 'saq': 'Mayotte', 'sbp': 'Mayote', 'sc': 'Mayotte', 'sd': 'مياتي', 'sd-Arab': 'مياتي', 'se': 'Mayotte', 'seh': 'Mayotte', 'ses': 'Mayooti', 'sg': 'Mäyôte', 'shi': 'ⵎⴰⵢⵓⵟ', 'shi-Latn': 'mayuṭ', 'shi-Tfng': 'ⵎⴰⵢⵓⵟ', 'si': 'මයෝට්', 'sk': 'Mayotte', 'sl': 'Mayotte', 'smn': 'Mayotte', 'sn': 'Mayotte', 'so': 'Mayotte', 'sq': 'Majotë', 'sr': 'Мајот', 'sr-Cyrl': 'Мајот', 'sr-Latn': 'Majot', 'sv': 'Mayotte', 'sw': 'Mayotte', 'syr': 'ܡܐܝܘܛ', 'ta': 'மயோட்', 'te': 'మాయొట్', 'teo': 'Mayotte', 'tg': 'Майотта', 'th': 'มายอต', 'ti': 'ማዮት', 'tk': 'Maýotta', 'to': 'Maiote', 'tr': 'Mayotte', 'tt': 'Майотта', 'twq': 'Mayooti', 'tzm': 'Mayuṭ', 'ug': 'مايوتى', 'uk': 'Майотта', 'ur': 'مایوٹ', 'uz': 'Mayotta', 'uz-Cyrl': 'Майотта', 'uz-Latn': 'Mayotta', 'vai': 'ꕮꗚꔎ', 'vai-Latn': 'Mavote', 'vai-Vaii': 'ꕮꗚꔎ', 'vec': 'Ìzole Maiote', 'vi': 'Mayotte', 'vun': 'Mayotte', 'wae': 'Moyette', 'wo': 'Mayot', 'xh': 'EMayotte', 'xnr': 'मायोते', 'xog': 'Mayotte', 'yav': 'mayɔ́ɔt', 'yi': 'מאַיאט', 'yo': 'Mayote', 'yrl': 'Mayuti', 'yue': '馬約特', 'yue-Hans': '马约特', 'yue-Hant': '馬約特', 'zgh': 'ⵎⴰⵢⵓⵟ', 'zh': '马约特', 'zh-Hans': '马约特', 'zh-Hant': '馬約特島', 'zu': 'i-Mayotte'},
    'YU': {'en': 'Yugoslavia'},
    'Yezi': {'bal-Latn': 'Yezidi', 'bs': 'jezidi pismo', 'bs-Cyrl': 'језиди писмо', 'bs-Latn': 'jezidi pismo', 'en': 'Yezidi', 'et': 'jeziidi', 'eu': 'yezidiera', 'ff-Adlm': '𞤒𞤢𞤶𞤭𞤣𞤭𞥅𞤴𞤵', 'fi': 'jesidi', 'fo': 'vezidi', 'gd': 'Yezidis', 'he': 'יעזי', 'hi-Latn': 'Yezidi', 'id': 'Yezidi', 'ig': 'Yezidi', 'kk': 'езид жазуы', 'kk-Cyrl': 'езид жазуы', 'ku': 'êzidî', 'mn': 'Езиди', 'nl': 'Jezidi', 'nqo': 'ߦߋߖ߭ߘߌ', 'ro': 'yazidită', 'sc': 'yezidi', 'so': 'Dadka Yesiidiga', 'sq': 'jezidi', 'sv': 'yazidiska', 'tg': 'Язидӣ', 'vec': 'yezidi', 'zh': '雅兹迪文', 'zh-Hans': '雅兹迪文', 'zu': 'i-Yezidi'},
    'Yiii': {'af': 'Yi', 'ak': 'Yifoɔ kasa', 'am': 'ዪ', 'ar': 'اليي', 'as': 'য়ি', 'ast': 'yi', 'az': 'yi', 'az-Latn': 'yi', 'bal-Latn': 'Yi', 'be': 'йі', 'be-tarask': 'йі', 'bg': 'Йи', 'bn': 'ই', 'brx': 'यी', 'bs': 'ji pismo', 'bs-Cyrl': 'ји писмо', 'bs-Latn': 'ji pismo', 'ca': 'yi', 'ca-ES-valencia': 'yi', 'ccp': '𑄅𑄪𑄃𑄨', 'chr': 'Ᏹ', 'cs': 'yi', 'cy': 'Yi', 'da': 'yi', 'de': 'Yi', 'dsb': 'yiske pismo', 'el': 'Γι', 'el-polyton': 'Γι', 'en': 'Yi', 'en-Dsrt': '𐐏𐐨', 'es': 'yi', 'et': 'jii', 'eu': 'yiarra', 'fa': 'یی', 'ff-Adlm': '𞤒𞤭𞥅', 'fi': 'yiläinen', 'fil': 'Yi', 'fo': 'yi', 'fr': 'yi', 'fy': 'Yi', 'ga': 'Ís', 'gd': 'Yi', 'gl': 'yi', 'gsw': 'Yi', 'gu': 'યી', 'ha': 'Yi', 'he': 'יי', 'hi': 'यी', 'hi-Latn': 'Yi', 'hr': 'yi pismo', 'hsb': 'yiske pismo', 'ht': 'yi', 'hu': 'Ji', 'hy': 'ի', 'id': 'Yi', 'ig': 'Yi', 'ii': 'ꆈꌠꁱꂷ', 'is': 'yí', 'it': 'yi', 'ja': 'イ文字', 'jv': 'Yi', 'ka': 'ი', 'kgp': 'yji', 'kk': 'и жазуы', 'kk-Cyrl': 'и жазуы', 'km': 'យី', 'kn': 'ಯಿ', 'ko': '이 문자', 'kok': 'यी', 'kok-Deva': 'यी', 'ks': 'یٖی', 'ks-Arab': 'یٖی', 'ky': 'Йи (жазуу)', 'lb': 'Yi', 'lo': 'ຍີ', 'lt': 'ji', 'lv': 'ji', 'mai': 'यी', 'mi': 'Eī', 'mk': 'ји', 'ml': 'യി', 'mn': 'И бичиг', 'mr': 'यी', 'ms': 'Yi', 'my': 'ရီ', 'nb': 'yi', 'ne': 'यी', 'nl': 'Yi', 'nn': 'yi', 'no': 'yi', 'nqo': 'ߦߌ', 'or': 'ୟୀ', 'pa': 'ਯੀ', 'pa-Guru': 'ਯੀ', 'pcm': 'Yi Lángwej', 'pl': 'yi', 'ps': 'یی', 'pt': 'yi', 'qu': 'Yi Simi', 'rm': 'yi', 'ro': 'yi', 'ru': 'и', 'sc': 'yi', 'sd': 'يي', 'sd-Arab': 'يي', 'si': 'යී', 'sk': 'yi', 'sl': 'jiški', 'so': 'Tiknoolajiyada Yi', 'sq': 'ji', 'sr': 'ји', 'sr-Cyrl': 'ји', 'sr-Latn': 'ji', 'sv': 'yi', 'sw': 'Kiyii', 'ta': 'யீ', 'te': 'యి', 'tg': 'Юйӣ', 'th': 'ยิ', 'ti': 'ዪ', 'tk': 'Ýi ýazuwy', 'to': 'tohinima fakaīī', 'tr': 'Yi', 'ug': 'يىچە', 'uk': 'йї', 'ur': 'یی', 'uz': 'i', 'uz-Latn': 'i', 'vec': 'yi', 'vi': 'Chữ Di', 'xnr': 'यी', 'yo': 'Èdè Yi', 'yrl': 'yi', 'yue': '彞文', 'yue-Hans': '彝文', 'yue-Hant': '彞文', 'zh': '彝文', 'zh-Hans': '彝文', 'zh-Hant': '彞文', 'zu': 'isi-Yi'},
    'ZA': {'af': 'Suid-Afrika', 'agq': 'Afɨlekà ghɨ Emàm ghò', 'ak': 'Abibirem Anaafoɔ', 'am': 'ደቡብ አፍሪካ', 'ar': 'جنوب أفريقيا', 'as': 'দক্ষিণ আফ্রিকা', 'asa': 'Afrika Kuthini', 'ast': 'Sudáfrica', 'az': 'Cənub Afrika', 'az-Cyrl': 'Ҹәнуб Африка', 'az-Latn': 'Cənub Afrika', 'bal-Latn': 'Zerbári Apriká', 'bas': 'Àfrǐkà Sɔ̀', 'be': 'Паўднёва-Афрыканская Рэспубліка', 'be-tarask': 'Паўднёва-Афрыканская Рэспубліка', 'bez': 'Huafrika iya Hukusini', 'bg': 'Южна Африка', 'blo': 'Sautafrika', 'bm': 'Worodugu Afriki', 'bn': 'দক্ষিণ আফ্রিকা', 'br': 'Suafrika', 'brx': 'खोला आफ्रिका', 'bs': 'Južnoafrička Republika', 'bs-Cyrl': 'Јужноафричка Република', 'bs-Latn': 'Južnoafrička Republika', 'ca': 'Sud-àfrica', 'ca-ES-valencia': 'Sud-àfrica', 'ccp': '𑄘𑄧𑄉𑄨𑄚𑄴 𑄃𑄜𑄳𑄢𑄨𑄇', 'ce': 'Къилба-Африкин Республика', 'ceb': 'South Africa', 'cgg': 'Sausi Afirika', 'chr': 'ᏧᎦᎾᏮ ᎬᎿᎨᏍᏛ', 'ckb': 'ئەفریقای باشوور', 'cs': 'Jihoafrická republika', 'cv': 'Кӑнтӑр Африка Республики', 'cy': 'De Affrica', 'da': 'Sydafrika', 'dav': 'Afrika Kusini', 'de': 'Südafrika', 'dje': 'Hawsa Afriki Laboo', 'dsb': 'Pódpołdnjowa Afrika (Republika)', 'dz': 'སཱའུཐ་ ཨཕ་རི་ཀ', 'ebu': 'Afrika Kusini', 'ee': 'Anyiehe Africa nutome', 'el': 'Νότια Αφρική', 'el-polyton': 'Νότια Ἀφρική', 'en': 'South Africa', 'en-Dsrt': '𐐝𐐵𐑃 𐐈𐑁𐑉𐐲𐐿𐐲', 'eo': 'Sud-Afriko', 'es': 'Sudáfrica', 'et': 'Lõuna-Aafrika Vabariik', 'eu': 'Hegoafrika', 'ewo': 'Afiríka yá Súd', 'fa': 'افریقای جنوبی', 'ff': 'Afrik bŋ Worgo', 'ff-Adlm': '𞤀𞤬𞤪𞤭𞤳𞤢 𞤂𞤫𞤧𞤤𞤫𞤴𞤪𞤭', 'ff-Latn': 'Afrik bŋ Worgo', 'fi': 'Etelä-Afrikka', 'fil': 'South Africa', 'fo': 'Suðurafrika', 'fr': 'Afrique du Sud', 'fur': 'Sud Afriche', 'fy': 'Sûd-Afrika', 'ga': 'an Afraic Theas', 'gd': 'Afraga a Deas', 'gl': 'Suráfrica', 'gsw': 'Süüdafrika', 'gu': 'દક્ષિણ આફ્રિકા', 'guz': 'Afrika Kusini', 'ha': 'Afirka Ta Kudu', 'he': 'דרום אפריקה', 'hi': 'दक्षिण अफ़्रीका', 'hi-Latn': 'South Africa', 'hr': 'Južnoafrička Republika', 'hsb': 'Južna Afrika (Republika)', 'ht': 'Afrique du Sud', 'hu': 'Dél-afrikai Köztársaság', 'hy': 'Հարավաֆրիկյան Հանրապետություն', 'ia': 'Africa del Sud', 'id': 'Afrika Selatan', 'ig': 'South Africa', 'is': 'Suður-Afríka', 'it': 'Sudafrica', 'ja': '南アフリカ', 'jmc': 'Afrika Kusini', 'jv': 'Afrika Kidul', 'ka': 'სამხრეთ აფრიკის რესპუბლიკა', 'kaa': 'Қубла Африка Республикасы', 'kaa-Cyrl': 'Қубла Африка Республикасы', 'kab': 'Tafriqt Wadda', 'kam': 'Afrika Kusini', 'kde': 'Afilika Kusini', 'kea': 'Áfrika di Sul', 'kgp': 'Sur-Afrika', 'khq': 'Hawsa Afriki Laboo', 'ki': 'Afrika Kusini', 'kk': 'Оңтүстік Африка', 'kk-Cyrl': 'Оңтүстік Африка', 'kln': 'Emetab Afrika nebo Murot tai', 'km': 'អាហ្វ្រិកខាងត្បូង', 'kn': 'ದಕ್ಷಿಣ ಆಫ್ರಿಕಾ', 'ko': '남아프리카', 'kok': 'दक्षिण आफ्रीका', 'kok-Deva': 'दक्षिण आफ्रीका', 'ks': 'جنوبی افریقہ', 'ks-Arab': 'جنوبی افریقہ', 'ksb': 'Aflika Kusini', 'ksf': 'afrik anǝ a sud', 'ksh': 'de Republik Södaffrika', 'ku': 'Afrîkaya Başûr', 'kxv': 'dkīṇ aaprika', 'kxv-Deva': 'दॉकिण आप़्रीका', 'kxv-Latn': 'dkīṇ aaprika', 'kxv-Orya': 'ଦକିଣ ଆପ୍ରିକା', 'kxv-Telu': 'దకిణ అప్రికా', 'ky': 'Түштүк-Африка Республикасы', 'lag': 'Afɨrɨka ya Saame', 'lb': 'Südafrika', 'lg': 'Sawusafirika', 'lld': 'Südafrica', 'ln': 'Afríka ya Súdi', 'lo': 'ອາຟຣິກາໃຕ້', 'lt': 'Pietų Afrika', 'lu': 'Afrika ya Súdi', 'luo': 'South Africa', 'luy': 'Afrika Kusini', 'lv': 'Dienvidāfrikas Republika', 'mai': 'दक्षिण अफ्रीका', 'mas': 'Afrika Kusini', 'mer': 'Afrika ya Southi', 'mfe': 'Sid-Afrik', 'mg': 'Afrika Atsimo', 'mgh': 'Afrika du Sulu', 'mi': 'Āwherika ki te Tonga', 'mk': 'Јужноафриканска Република', 'ml': 'ദക്ഷിണാഫ്രിക്ക', 'mn': 'Өмнөд Африк', 'mr': 'दक्षिण आफ्रिका', 'ms': 'Afrika Selatan', 'mt': 'l-Afrika t’Isfel', 'mua': 'Afrika nekǝsǝŋ', 'my': 'တောင်အာဖရိက', 'mzn': 'جنوبی افریقا', 'naq': 'Suid Afrikab', 'nb': 'Sør-Afrika', 'nd': 'Mzansi ye Afrika', 'ne': 'दक्षिण अफ्रिका', 'nl': 'Zuid-Afrika', 'nmg': 'Afríka yí sí', 'nn': 'Sør-Afrika', 'no': 'Sør-Afrika', 'nqo': 'ߥߙߏ߬ߘߎ߮ ߝߘߊ߬ߝߌ߲߬ߠߊ߫', 'nso': 'Afrika Borwa', 'nyn': 'Sausi Afirika', 'om': 'Afrikaa Kibbaa', 'or': 'ଦକ୍ଷିଣ ଆଫ୍ରିକା', 'pa': 'ਦੱਖਣੀ ਅਫਰੀਕਾ', 'pa-Guru': 'ਦੱਖਣੀ ਅਫਰੀਕਾ', 'pcm': 'Saút Áfríka', 'pl': 'Republika Południowej Afryki', 'ps': 'سویلي افریقا', 'pt': 'África do Sul', 'qu': 'Sudáfrica', 'rm': 'Africa dal Sid', 'rn': 'Afurika y’Epfo', 'ro': 'Africa de Sud', 'rof': 'Afrika Kusini', 'ru': 'Южно-Африканская Республика', 'rwk': 'Afrika Kusini', 'saq': 'Afrika Kusini', 'sbp': 'Afilika Kusini', 'sc': 'Sudàfrica', 'sd': 'ڏکڻ آفريقا', 'sd-Arab': 'ڏکڻ آفريقا', 'se': 'Mátta-Afrihká', 'seh': 'África do Sul', 'ses': 'Hawsa Afriki Laboo', 'sg': 'Mbongo-Afrîka', 'shi': 'ⴰⴼⵔⵉⵇⵢⴰ ⵏ ⵉⴼⴼⵓⵙ', 'shi-Latn': 'afriqya n iffus', 'shi-Tfng': 'ⴰⴼⵔⵉⵇⵢⴰ ⵏ ⵉⴼⴼⵓⵙ', 'si': 'දකුණු අප්\u200dරිකාව', 'sk': 'Južná Afrika', 'sl': 'Južnoafriška republika', 'smn': 'Maadâ-Afrikka', 'sn': 'South Africa', 'so': 'Koonfur Afrika', 'sq': 'Afrika e Jugut', 'sr': 'Јужноафричка Република', 'sr-Cyrl': 'Јужноафричка Република', 'sr-Latn': 'Južnoafrička Republika', 'st': 'Afrika Borwa', 'sv': 'Sydafrika', 'sw': 'Afrika Kusini', 'syr': 'ܬܝܡܢ ܐܦܪܝܩܐ', 'ta': 'தென் ஆப்பிரிக்கா', 'te': 'దక్షిణ ఆఫ్రికా', 'teo': 'Afrika Kusini', 'tg': 'Африкаи Ҷанубӣ', 'th': 'แอฟริกาใต้', 'ti': 'ደቡብ ኣፍሪቃ', 'tk': 'Günorta Afrika', 'tn': 'Aforika Borwa', 'to': 'ʻAfilika tonga', 'tr': 'Güney Afrika', 'tt': 'Көньяк Африка', 'twq': 'Hawsa Afriki Laboo', 'tzm': 'Tafrikt Tunẓul', 'ug': 'جەنۇبىي ئافرىقا', 'uk': 'Південно-Африканська Республіка', 'ur': 'جنوبی افریقہ', 'uz': 'Janubiy Afrika Respublikasi', 'uz-Cyrl': 'Жанубий Африка Республикаси', 'uz-Latn': 'Janubiy Afrika Respublikasi', 'vai': 'ꕉꔱꔸꕪ ꗛꔤ ꔒꘋꗣ ꗏ ꕸꖃꔀ', 'vai-Latn': 'Afirika Kɔi Leŋŋɛ Lɔ', 'vai-Vaii': 'ꕉꔱꔸꕪ ꗛꔤ ꔒꘋꗣ ꗏ ꕸꖃꔀ', 'vec': 'Sudàfrega', 'vi': 'Nam Phi', 'vun': 'Afrika Kusini', 'wae': 'Südafrika', 'wo': 'Afrik di Sid', 'xh': 'EMzantsi Afrika', 'xnr': 'दखण अफ़्रीका', 'xog': 'Sawusafirika', 'yav': 'afilí mbátɛ́', 'yi': 'דרום־אַפֿריקע', 'yo': 'Gúúṣù Áfíríkà', 'yrl': 'Afirika Su suí', 'yue': '南非', 'yue-Hans': '南非', 'yue-Hant': '南非', 'zgh': 'ⴰⴼⵔⵉⵇⵢⴰ ⵏ ⵉⴼⴼⵓⵙ', 'zh': '南非', 'zh-Hans': '南非', 'zh-Hant': '南非', 'zu': 'iNingizimu Afrika'},
    'ZM': {'af': 'Zambië', 'agq': 'Zambìa', 'ak': 'Zambia', 'am': 'ዛምቢያ', 'ar': 'زامبيا', 'as': 'জাম্বিয়া', 'asa': 'Dhambia', 'ast': 'Zambia', 'az': 'Zambiya', 'az-Cyrl': 'Замбија', 'az-Latn': 'Zambiya', 'bal-Latn': 'Zambiá', 'bas': 'Zàmbià', 'be': 'Замбія', 'be-tarask': 'Замбія', 'bem': 'Zambia', 'bez': 'Huzambia', 'bg': 'Замбия', 'bgn': 'زامبیا', 'blo': 'Sambii', 'bm': 'Zanbi', 'bn': 'জাম্বিয়া', 'br': 'Zambia', 'brx': 'जाम्बिया', 'bs': 'Zambija', 'bs-Cyrl': 'Замбија', 'bs-Latn': 'Zambija', 'ca': 'Zàmbia', 'ca-ES-valencia': 'Zàmbia', 'ccp': '𑄎𑄟𑄴𑄝𑄨𑄠', 'ce': 'Замби', 'ceb': 'Zambia', 'cgg': 'Zambia', 'chr': 'ᏌᎻᏈᏯ', 'ckb': 'زامبیا', 'cs': 'Zambie', 'cv': 'Замби', 'cy': 'Zambia', 'da': 'Zambia', 'dav': 'Zambia', 'de': 'Sambia', 'dje': 'Zambi', 'dsb': 'Sambija', 'dz': 'ཛམ་བྷི་ཡ', 'ebu': 'Zambia', 'ee': 'Zambia nutome', 'el': 'Ζάμπια', 'el-polyton': 'Ζάμπια', 'en': 'Zambia', 'en-Dsrt': '𐐞𐐰𐑋𐐺𐐨𐐲', 'eo': 'Zambio', 'es': 'Zambia', 'et': 'Sambia', 'eu': 'Zambia', 'ewo': 'Zambí', 'fa': 'زامبیا', 'ff': 'Sammbi', 'ff-Adlm': '𞤟𞤢𞤥𞤦𞤭𞤴𞤢', 'ff-Latn': 'Sammbi', 'fi': 'Sambia', 'fil': 'Zambia', 'fo': 'Sambia', 'fr': 'Zambie', 'fur': 'Zambia', 'fy': 'Zambia', 'ga': 'an tSaimbia', 'gd': 'Sàimbia', 'gl': 'Zambia', 'gsw': 'Sambia', 'gu': 'ઝામ્બિયા', 'guz': 'Zambia', 'ha': 'Zambiya', 'he': 'זמביה', 'hi': 'ज़ाम्बिया', 'hi-Latn': 'Zambia', 'hr': 'Zambija', 'hsb': 'Sambija', 'ht': 'Zambie', 'hu': 'Zambia', 'hy': 'Զամբիա', 'ia': 'Zambia', 'id': 'Zambia', 'ig': 'Zambia', 'is': 'Sambía', 'it': 'Zambia', 'ja': 'ザンビア', 'jgo': 'Zambî', 'jmc': 'Zambia', 'jv': 'Sambia', 'ka': 'ზამბია', 'kaa': 'Замбия', 'kaa-Cyrl': 'Замбия', 'kab': 'Zambya', 'kam': 'Nzambia', 'kde': 'Zambia', 'kea': 'Zámbia', 'kgp': 'Jỹmija', 'khq': 'Zambi', 'ki': 'Zambia', 'kk': 'Замбия', 'kk-Cyrl': 'Замбия', 'kln': 'Emetab Zambia', 'km': 'សំប៊ី', 'kn': 'ಜಾಂಬಿಯ', 'ko': '잠비아', 'kok': 'झांबिया', 'kok-Deva': 'झांबिया', 'ks': 'زیمبیا', 'ks-Arab': 'زیمبیا', 'ksb': 'Zambia', 'ksf': 'zambí', 'ksh': 'Sambija', 'ku': 'Zambîya', 'kxv': 'jambiya', 'kxv-Deva': 'ज़ाम्बिया', 'kxv-Latn': 'jambiya', 'kxv-Orya': 'ଜାମ୍ବିୟା', 'kxv-Telu': 'జంబియా', 'ky': 'Замбия', 'lag': 'Sámbia', 'lb': 'Sambia', 'lg': 'Zambya', 'lld': 'Zambia', 'ln': 'Zambi', 'lo': 'ແຊມເບຍ', 'lt': 'Zambija', 'lu': 'Zambi', 'luo': 'Zambia', 'luy': 'Zambia', 'lv': 'Zambija', 'mai': 'जाम्बिया', 'mas': 'Sambia', 'mer': 'Zambia', 'mfe': 'Zambi', 'mg': 'Zambia', 'mgh': 'Uzambia', 'mi': 'Tāmipia', 'mk': 'Замбија', 'ml': 'സാംബിയ', 'mn': 'Замби', 'mr': 'झाम्बिया', 'ms': 'Zambia', 'mt': 'iż-Żambja', 'mua': 'Zambiya', 'my': 'ဇမ်ဘီယာ', 'mzn': 'زامبیا', 'naq': 'Zambiab', 'nb': 'Zambia', 'nd': 'Zambiya', 'ne': 'जाम्बिया', 'nl': 'Zambia', 'nmg': 'Zambia', 'nn': 'Zambia', 'no': 'Zambia', 'nqo': 'ߖ߭ߊ߲ߓߌ߫', 'nyn': 'Zambia', 'om': 'Zaambiyaa', 'or': 'ଜାମ୍ବିଆ', 'pa': 'ਜ਼ਾਮਬੀਆ', 'pa-Guru': 'ਜ਼ਾਮਬੀਆ', 'pcm': 'Zámbia', 'pl': 'Zambia', 'ps': 'زیمبیا', 'pt': 'Zâmbia', 'qu': 'Zambia', 'rm': 'Sambia', 'rn': 'Zambiya', 'ro': 'Zambia', 'rof': 'Zambia', 'ru': 'Замбия', 'rwk': 'Zambia', 'saq': 'Zambia', 'sbp': 'Sambiya', 'sc': 'Zàmbia', 'sd': 'زيمبيا', 'sd-Arab': 'زيمبيا', 'se': 'Zambia', 'seh': 'Zâmbia', 'ses': 'Zambi', 'sg': 'Zambïi', 'shi': 'ⵣⴰⵎⴱⵢⴰ', 'shi-Latn': 'zambya', 'shi-Tfng': 'ⵣⴰⵎⴱⵢⴰ', 'si': 'සැම්බියාව', 'sk': 'Zambia', 'sl': 'Zambija', 'smn': 'Sambia', 'sn': 'Zambia', 'so': 'Saambiya', 'sq': 'Zambi', 'sr': 'Замбија', 'sr-Cyrl': 'Замбија', 'sr-Latn': 'Zambija', 'sv': 'Zambia', 'sw': 'Zambia', 'syr': 'ܙܐܡܒܝܐ', 'ta': 'ஜாம்பியா', 'te': 'జాంబియా', 'teo': 'Zambia', 'tg': 'Замбия', 'th': 'แซมเบีย', 'ti': 'ዛምብያ', 'tk': 'Zambiýa', 'to': 'Semipia', 'tr': 'Zambiya', 'tt': 'Замбия', 'twq': 'Zambi', 'tzm': 'Zambya', 'ug': 'زامبىيە', 'uk': 'Замбія', 'ur': 'زامبیا', 'uz': 'Zambiya', 'uz-Cyrl': 'Замбия', 'uz-Latn': 'Zambiya', 'vai': 'ꕤꔭꕩ', 'vai-Latn': 'Zambiya', 'vai-Vaii': 'ꕤꔭꕩ', 'vec': 'Zanbia', 'vi': 'Zambia', 'vun': 'Zambia', 'wae': 'Sambia', 'wo': 'Sàmbi', 'xh': 'EZambia', 'xnr': 'ज़ाम्बिया', 'xog': 'Zambya', 'yav': 'saambíi', 'yi': 'זאַמביע', 'yo': 'Ṣamibia', 'yrl': 'Sãbiya', 'yue': '尚比亞', 'yue-Hans': '尚比亚', 'yue-Hant': '尚比亞', 'zgh': 'ⵣⴰⵎⴱⵢⴰ', 'zh': '赞比亚', 'zh-Hans': '赞比亚', 'zh-Hant': '尚比亞', 'zu': 'i-Zambia'},
    'ZR': {'en': 'Zaire'},
    'ZW': {'af': 'Zimbabwe', 'agq': 'Zìmbagbɛ̀', 'ak': 'Zimbabue', 'am': 'ዚምቧቤ', 'ar': 'زيمبابوي', 'as': 'জিম্বাবৱে', 'asa': 'Dhimbabwe', 'ast': 'Zimbabue', 'az': 'Zimbabve', 'az-Cyrl': 'Зимбабве', 'az-Latn': 'Zimbabve', 'bal-Latn': 'Zembábwé', 'bas': 'Zìmbàbwê', 'be': 'Зімбабвэ', 'be-tarask': 'Зімбабвэ', 'bez': 'Huzimbabwe', 'bg': 'Зимбабве', 'bgn': 'زیمبابوی', 'blo': 'Simbabʊwee', 'bm': 'Zimbabuwe', 'bn': 'জিম্বাবোয়ে', 'br': 'Zimbabwe', 'brx': 'जिम्बाब्वे', 'bs': 'Zimbabve', 'bs-Cyrl': 'Зимбабве', 'bs-Latn': 'Zimbabve', 'ca': 'Zimbàbue', 'ca-ES-valencia': 'Zimbàbue', 'ccp': '𑄎𑄨𑄟𑄴𑄝𑄝𑄪𑄠𑄬', 'ce': 'Зимбабве', 'ceb': 'Zimbabwe', 'cgg': 'Zimbabwe', 'chr': 'ᏏᎻᏆᏇ', 'ckb': 'زیمبابوی', 'cs': 'Zimbabwe', 'cv': 'Зимбабве', 'cy': 'Zimbabwe', 'da': 'Zimbabwe', 'dav': 'Zimbabwe', 'de': 'Simbabwe', 'dje': 'Zimbabwe', 'dsb': 'Simbabwe', 'dz': 'ཛིམ་བྷབ་ཝེ', 'ebu': 'Zimbabwe', 'ee': 'Zimbabwe nutome', 'el': 'Ζιμπάμπουε', 'el-polyton': 'Ζιμπάμπουε', 'en': 'Zimbabwe', 'en-Dsrt': '𐐞𐐮𐑋𐐺𐐪𐐺𐐶𐐩', 'eo': 'Zimbabvo', 'es': 'Zimbabue', 'et': 'Zimbabwe', 'eu': 'Zimbabwe', 'ewo': 'Zimbabwé', 'fa': 'زیمبابوه', 'ff': 'Simbaabuwe', 'ff-Adlm': '𞤟𞤭𞤥𞤦𞤢𞥄𞤥𞤵𞤴𞤢', 'ff-Latn': 'Simbaabuwe', 'fi': 'Zimbabwe', 'fil': 'Zimbabwe', 'fo': 'Simbabvi', 'fr': 'Zimbabwe', 'fur': 'Zimbabwe', 'fy': 'Zimbabwe', 'ga': 'an tSiombáib', 'gd': 'An t-Sìombab', 'gl': 'Zimbabwe', 'gsw': 'Simbabwe', 'gu': 'ઝિમ્બાબ્વે', 'guz': 'Zimbabwe', 'ha': 'Zimbabuwe', 'he': 'זימבבואה', 'hi': 'ज़िम्बाब्वे', 'hi-Latn': 'Zimbabwe', 'hr': 'Zimbabve', 'hsb': 'Simbabwe', 'ht': 'Zimbabwe', 'hu': 'Zimbabwe', 'hy': 'Զիմբաբվե', 'ia': 'Zimbabwe', 'id': 'Zimbabwe', 'ig': 'Zimbabwe', 'is': 'Simbabve', 'it': 'Zimbabwe', 'ja': 'ジンバブエ', 'jgo': 'Zimbámbwɛ', 'jmc': 'Zimbabwe', 'jv': 'Simbabwe', 'ka': 'ზიმბაბვე', 'kaa': 'Зимбабве', 'kaa-Cyrl': 'Зимбабве', 'kab': 'Zimbabwi', 'kam': 'Nzimbambwe', 'kde': 'Zimbabwe', 'kea': 'Zimbábui', 'kgp': 'Jĩmamuje', 'khq': 'Zimbabwe', 'ki': 'Zimbabwe', 'kk': 'Зимбабве', 'kk-Cyrl': 'Зимбабве', 'kln': 'Emetab Zimbabwe', 'km': 'ស៊ីមបាវ៉េ', 'kn': 'ಜಿಂಬಾಬ್ವೆ', 'ko': '짐바브웨', 'kok': 'जिम्बाब्वे', 'kok-Deva': 'जिम्बाब्वे', 'ks': 'زِمبابے', 'ks-Arab': 'زِمبابے', 'ksb': 'Zimbabwe', 'ksf': 'zimbabwɛ́', 'ksh': 'Simbabwe', 'ku': 'Zîmbabwe', 'kxv': 'jimbabve', 'kxv-Deva': 'ज़िम्बाब्वे', 'kxv-Latn': 'jimbabve', 'kxv-Orya': 'ଜିମ୍ବାୱେ', 'kxv-Telu': 'జింబాబ్వే', 'ky': 'Зимбабве', 'lag': 'Simbáabwe', 'lb': 'Simbabwe', 'lg': 'Zimbabwe', 'lld': 'Zimbabwe', 'ln': 'Zimbabwe', 'lo': 'ຊິມບັບເວ', 'lt': 'Zimbabvė', 'lu': 'Zimbabwe', 'luo': 'Zimbabwe', 'luy': 'Zimbabwe', 'lv': 'Zimbabve', 'mai': 'जिम्बाब्बे', 'mas': 'Simbabwe', 'mer': 'Zimbabwe', 'mfe': 'Zimbabwe', 'mg': 'Zimbaboe', 'mgh': 'Uzimbabwe', 'mi': 'Timuwawe', 'mk': 'Зимбабве', 'ml': 'സിംബാബ്\u200cവേ', 'mn': 'Зимбабве', 'mr': 'झिम्बाब्वे', 'ms': 'Zimbabwe', 'mt': 'iż-Żimbabwe', 'mua': 'Zimbabwe', 'my': 'ဇင်ဘာဘွေ', 'mzn': 'زیمبابوه', 'naq': 'Zimbabweb', 'nb': 'Zimbabwe', 'nd': 'Zimbabwe', 'ne': 'जिम्बाबवे', 'nl': 'Zimbabwe', 'nmg': 'Zimbabwǝ', 'nn': 'Zimbabwe', 'no': 'Zimbabwe', 'nqo': 'ߖ߭ߌ߲ߓߊߓߏߦߋ߫', 'nyn': 'Zimbabwe', 'om': 'Zimbaabuwee', 'or': 'ଜିମ୍ବାୱେ', 'pa': 'ਜ਼ਿੰਬਾਬਵੇ', 'pa-Guru': 'ਜ਼ਿੰਬਾਬਵੇ', 'pcm': 'Zimbábwẹ', 'pl': 'Zimbabwe', 'ps': 'زیمبابوی', 'pt': 'Zimbábue', 'qu': 'Zimbabue', 'rm': 'Simbabwe', 'rn': 'Zimbabwe', 'ro': 'Zimbabwe', 'rof': 'Zimbabwe', 'ru': 'Зимбабве', 'rwk': 'Zimbabwe', 'saq': 'Zimbabwe', 'sbp': 'Simbabwe', 'sc': 'Zimbabwe', 'sd': 'زمبابوي', 'sd-Arab': 'زمبابوي', 'se': 'Zimbabwe', 'seh': 'Zimbábue', 'ses': 'Zimbabwe', 'sg': 'Zimbäbwe', 'shi': 'ⵣⵉⵎⴱⴰⴱⵡⵉ', 'shi-Latn': 'zimbabwi', 'shi-Tfng': 'ⵣⵉⵎⴱⴰⴱⵡⵉ', 'si': 'සිම්බාබ්වේ', 'sk': 'Zimbabwe', 'sl': 'Zimbabve', 'smn': 'Zimbabwe', 'sn': 'Zimbabwe', 'so': 'Simbaabwe', 'sq': 'Zimbabve', 'sr': 'Зимбабве', 'sr-Cyrl': 'Зимбабве', 'sr-Latn': 'Zimbabve', 'sv': 'Zimbabwe', 'sw': 'Zimbabwe', 'syr': 'ܙܝܡܒܐܒܘܝ', 'ta': 'ஜிம்பாப்வே', 'te': 'జింబాబ్వే', 'teo': 'Zimbabwe', 'tg': 'Зимбабве', 'th': 'ซิมบับเว', 'ti': 'ዚምባብዌ', 'tk': 'Zimbabwe', 'to': 'Simipapuei', 'tr': 'Zimbabve', 'tt': 'Зимбабве', 'twq': 'Zimbabwe', 'tzm': 'Zimbabwi', 'ug': 'زىمبابۋې', 'uk': 'Зімбабве', 'ur': 'زمبابوے', 'uz': 'Zimbabve', 'uz-Cyrl': 'Зимбабве', 'uz-Latn': 'Zimbabve', 'vai': 'ꔽꕓꖜꔃ', 'vai-Latn': 'Zimbabhuwe', 'vai-Vaii': 'ꔽꕓꖜꔃ', 'vec': 'Zinbawe', 'vi': 'Zimbabwe', 'vun': 'Zimbabwe', 'wae': 'Simbabwe', 'wo': 'Simbabwe', 'xh': 'EZimbabwe', 'xnr': 'ज़िम्बाब्वे', 'xog': 'Zimbabwe', 'yav': 'simbapuwé', 'yi': 'זימבאַבווע', 'yo': 'Ṣimibabe', 'yrl': 'Sĩbabuwe', 'yue': '辛巴威', 'yue-Hans': '辛巴威', 'yue-Hant': '辛巴威', 'zgh': 'ⵣⵉⵎⴱⴰⴱⵡⵉ', 'zh': '津巴布韦', 'zh-Hans': '津巴布韦', 'zh-Hant': '辛巴威', 'zu': 'iZimbabwe'},
    'ZZ': {'af': 'Onbekende gebied', 'ak': 'Mantam a Yɛnnim', 'am': 'ያልታወቀ ክልል', 'ar': 'منطقة غير معروفة', 'as': 'অজ্ঞাত ক্ষেত্ৰ', 'ast': 'Rexón desconocida', 'az': 'Naməlum Region', 'az-Cyrl': 'Намәлум Реҝион', 'az-Latn': 'Naməlum Region', 'bal-Latn': 'Nagisshetagén damag', 'be': 'Невядомы рэгіён', 'be-tarask': 'Невядомы рэгіён', 'bg': 'непознат регион', 'bgc': 'अणजाण जगह', 'bgn': 'نازانتین سیمسر', 'bho': 'अनजान इलाका', 'blo': 'gʊsaʊɩ kʊyɔʊ ʊ mana ma', 'bn': 'অজানা অঞ্চল', 'bo': 'མིའི་ཤེས་རྟོགས་མ་བྱུང་བའི་ཁོར་ཡུག', 'br': 'Rannved dianav', 'brx': 'मिथियै ओनसोल', 'bs': 'Nepoznata oblast', 'bs-Cyrl': 'Непозната или неважећа област', 'bs-Latn': 'Nepoznata oblast', 'ca': 'regió desconeguda', 'ca-ES-valencia': 'regió desconeguda', 'ccp': '𑄃𑄨𑄌𑄨𑄚𑄴 𑄎𑄉', 'ce': 'Йоьвзуш йоцу регион', 'ceb': 'Wala Mailhing Rehiyon', 'chr': 'ᏄᏬᎵᏍᏛᎾ ᎤᏔᏂᏗᎦᏙᎯ', 'ckb': 'ناوچەی نەناسراو', 'cs': 'neznámá oblast', 'cv': 'паллӑ мар регион', 'cy': 'Rhanbarth Anhysbys', 'da': 'Ukendt område', 'de': 'Unbekannte Region', 'doi': 'अनजांता खेत्तर', 'dsb': 'njeznaty region', 'dz': 'ངོ་མ་ཤེས་པའི་ལུང་ཕྱོགས', 'ee': 'nutome manya', 'el': 'Άγνωστη περιοχή', 'el-polyton': 'Άγνωστη περιοχή', 'en': 'Unknown Region', 'en-Dsrt': '𐐊𐑌𐐬𐑌 𐐬𐑉 𐐆𐑌𐑂𐐰𐑊𐐮𐐼 𐐡𐐨𐐾𐐲𐑌', 'es': 'Región desconocida', 'et': 'tundmatu piirkond', 'eu': 'Eskualde ezezaguna', 'fa': 'ناحیهٔ نامشخص', 'ff-Adlm': '𞤖𞤭𞤤𞥆𞤮 𞤀𞤧-𞤢𞤲𞤣𞤢𞥄𞤲𞤺𞤮', 'fi': 'tuntematon alue', 'fil': 'Hindi Kilalang Rehiyon', 'fo': 'ókent øki', 'fr': 'région indéterminée', 'fur': 'Regjon no cognossude o no valide', 'fy': 'Unbekend gebiet', 'ga': 'Réigiún Anaithnid', 'gd': 'Roinn-dùthcha neo-aithnichte', 'gl': 'Rexión descoñecida', 'gsw': 'Unbekannti oder ungültigi Regioon', 'gu': 'અજ્ઞાત પ્રદેશ', 'ha': 'Yanki da ba a sani ba', 'he': 'אזור לא ידוע', 'hi': 'अज्ञात क्षेत्र', 'hi-Latn': 'Unknown Region', 'hr': 'nepoznato područje', 'hsb': 'njeznaty region', 'ht': 'région indéterminée', 'hu': 'Ismeretlen körzet', 'hy': 'Անհայտ տարածաշրջան', 'ia': 'Region incognite', 'id': 'Wilayah Tidak Dikenal', 'ie': 'ínconosset region', 'ig': 'Mpaghara A na-amaghị', 'ii': 'ꃅꄷꅉꀋꐚꌠ', 'is': 'Óþekkt svæði', 'it': 'Regione sconosciuta', 'ja': '不明な地域', 'jgo': 'ŋgɔŋ yi pɛ́ ká kɛ́ jʉɔ', 'jv': 'Daerah Ora Dikenali', 'ka': 'უცნობი რეგიონი', 'kaa': 'белгисиз аймақ', 'kaa-Cyrl': 'белгисиз аймақ', 'kea': 'Rejion Diskonxedu', 'kgp': 'Reji’ỹv Veja tũ', 'kk': 'Белгісіз аймақ', 'kk-Cyrl': 'Белгісіз аймақ', 'km': 'តំបន់មិនស្គាល់', 'kn': 'ಅಜ್ಞಾತ ಪ್ರದೇಶ', 'ko': '알려지지 않은 지역', 'kok': 'अज्ञात प्रांत', 'kok-Deva': 'अज्ञात प्रांत', 'ks': 'نامولوٗم علاقٕہ', 'ks-Arab': 'نامولوٗم علاقٕہ', 'ks-Deva': 'नामोलुम अलाक़', 'ksh': '- Jähjend onbikannt -', 'ku': 'Herêma Nenas', 'ky': 'Белгисиз чөлкөм', 'lb': 'Onbekannt Regioun', 'lld': 'Raiun nia conesciü', 'lo': 'ຂົງເຂດທີ່ບໍ່ຮູ້ຈັກ', 'lrc': 'راساگە نادیار', 'lt': 'nežinoma sritis', 'lv': 'nezināms reģions', 'mai': 'अज्ञात क्षेत्र', 'mgo': 'aba aben tisɔ̀', 'mi': 'Rohe Tē Mōhiotia', 'mk': 'Непознат регион', 'ml': 'അജ്ഞാത പ്രദേശം', 'mn': 'Тодорхойгүй бүс', 'mni': 'মশকখংদবা লমদম', 'mni-Beng': 'মশকখংদবা লমদম', 'mr': 'अज्ञात प्रदेश', 'ms': 'Wilayah Tidak Diketahui', 'mt': 'Reġjun Mhux Magħruf', 'my': 'မသိ (သို့) မရှိသော ဒေသ', 'mzn': 'نامَیِّن منطقه', 'nb': 'ukjent område', 'ne': 'अज्ञात क्षेत्र', 'nl': 'onbekend gebied', 'nn': 'ukjent område', 'no': 'ukjent område', 'nqo': 'ߕߌ߲߬ߞߎߘߎ߲߫ ߕߊ߲߬ߠߊߕߍ߰ߓߊߟߌ', 'om': 'Naannoo Hin Beekamne', 'or': 'ଅଜଣା ଅଞ୍ଚଳ', 'os': 'Нӕзонгӕ бӕстӕ', 'pa': 'ਅਣਪਛਾਤਾ ਇਲਾਕਾ', 'pa-Guru': 'ਅਣਪਛਾਤਾ ਇਲਾਕਾ', 'pcm': 'Ríjọn Wé Nóbọ́di Sabí', 'pl': 'Nieznany region', 'ps': 'نامعلومه سيمه', 'pt': 'Região desconhecida', 'qu': 'Mana yachasqa Suyu', 'raj': 'अज्ञात', 'rm': 'regiun nunenconuschenta', 'ro': 'Regiune necunoscută', 'ru': 'неизвестный регион', 'sa': 'अज्ञात क्षेत्र:', 'sat': 'ᱵᱟᱝᱩᱨᱩᱢ ᱴᱚᱴᱷᱟ', 'sat-Olck': 'ᱵᱟᱝᱩᱨᱩᱢ ᱴᱚᱴᱷᱟ', 'sc': 'regione disconnota', 'sd': 'اڻڄاتل خطو', 'sd-Arab': 'اڻڄاتل خطو', 'sd-Deva': 'अणजातल इलाइको', 'se': 'dovdameahttun guovlu', 'si': 'හඳුනා නොගත් කළාපය', 'sk': 'neznámy región', 'sl': 'Neznano ali neveljavno območje', 'so': 'Gobol aan la aqoonin', 'sq': 'I panjohur', 'sr': 'непознат регион', 'sr-Cyrl': 'непознат регион', 'sr-Latn': 'nepoznat region', 'su': 'Wilayah Teu Dikenal', 'su-Latn': 'Wilayah Teu Dikenal', 'sv': 'okänd region', 'sw': 'Eneo lisilojulikana', 'ta': 'அறியப்படாத பிரதேசம்', 'te': 'తెలియని ప్రాంతం', 'tg': 'Минтақаи номаълум', 'th': 'ภูมิภาคที่ไม่รู้จัก', 'ti': 'ዘይተፈልጠ ዞባ', 'tk': 'Näbelli sebit', 'to': 'Potu fonua taʻeʻiloa pe hala', 'tpi': 'Rijen i no stap', 'tr': 'Bilinmeyen Bölge', 'tt': 'билгесез төбәк', 'ug': 'يوچۇن جاي', 'uk': 'Невідомий регіон', 'ur': 'نامعلوم علاقہ', 'uz': 'Noma’lum mintaqa', 'uz-Cyrl': 'Номаълум минтақа', 'uz-Latn': 'Noma’lum mintaqa', 'vec': 'Rejon desconosùa', 'vi': 'Vùng không xác định', 'wae': 'Unbekannti Regio', 'wo': 'Gox buñ xamul', 'xh': 'Ingingqi Engaziwayo', 'xnr': 'अणजाण लाका', 'yi': 'אומבאַוואוסטער ראַיאן', 'yo': 'Àgbègbè àìmọ̀', 'yrl': 'Tetãma Ũbawaukuamamẽ', 'yue': '未知區域', 'yue-Hans': '未知区域', 'yue-Hant': '未知區域', 'zh': '未知地区', 'zh-Hans': '未知地区', 'zh-Hant': '未知區域', 'zu': 'iSifunda esingaziwa'},
    'Zanb': {'bal-Latn': 'Zanabazar Eskwáer', 'bs': 'zanabazar četvrtasto pismo', 'bs-Cyrl': 'занабазар клинасто писмо', 'bs-Latn': 'zanabazar četvrtasto pismo', 'en': 'Zanabazar Square', 'et': 'Dzanabadzari ruutkiri', 'eu': 'zanabazar koadroa', 'ff-Adlm': '𞤟𞤢𞤲𞤢𞤦𞤢𞥁𞤢𞥄𞤪 𞤁𞤭𞤲𞤺𞤫𞤪𞤫', 'fi': 'zanabazar-neliökirjaimisto', 'ga': 'Zanabazar Square', 'gd': 'Zanabazar ceàrnagach', 'he': 'כיכר זנבזר', 'hi-Latn': 'Zanabazar Square', 'id': 'Zanabazar Square', 'ig': 'Zanabazar Square', 'kk': 'занабазар шаршы жазуы', 'kk-Cyrl': 'занабазар шаршы жазуы', 'mn': 'Занабазарын дөрвөлжин бичиг', 'ms': 'Segi Empat Zanabazar', 'nl': 'vierkant Zanabazar', 'nqo': 'ߖ߭ߣߊߓߊߖ߭ߊ߯ߙ ߝߘߎ߬ߓߍ߲', 'ro': 'Piața Zanabazar', 'sc': 'zanabar cuadradu', 'so': 'Xarafka laba jibaaran ee kujira Xarfaha Zanabazar', 'sq': 'katror zanabazar', 'sv': 'zanabazar kvadratisk skrift', 'tg': 'Майдони Занабазорӣ', 'vec': 'zanabazar cuadrà', 'zh': '札那巴札尔方块文字', 'zh-Hans': '札那巴札尔方块文字', 'zh-Hant': '札那巴札爾文字', 'zu': 'i-Zanabazar Square'},
    'Zinh': {'am': 'የተወረሰ', 'ar': 'الموروث', 'ast': 'heredáu', 'bal-Latn': 'Enheri', 'bn': 'কাই', 'br': 'hêrezh', 'brx': 'विरासत', 'bs': 'nasljedno pismo', 'bs-Cyrl': 'наследно писмо', 'bs-Latn': 'nasljedno pismo', 'ca': 'heretat', 'ca-ES-valencia': 'heretat', 'ccp': '𑄇𑄭', 'cs': 'zděděné', 'da': 'arvet', 'de': 'Geerbter Schriftwert', 'el': 'Κληρονομημένο', 'el-polyton': 'Κληρονομημένο', 'en': 'Inherited', 'en-Dsrt': '𐐆𐑌𐐸𐐯𐑉𐐮𐐻𐐲𐐼', 'es': 'heredado', 'et': 'päritud', 'eu': 'heredatua', 'fa': 'موروثی', 'ff-Adlm': '𞤁𞤮𞤲𞤣𞤭', 'fi': 'peritty', 'fo': 'arva skrift', 'fr': 'hérité', 'fy': 'Oergeërfd', 'ga': 'Oidhreacht', 'gd': 'Dìleab', 'gsw': 'G’eerbtä Schriftwärt', 'gu': 'વંશાગત', 'he': 'מורש', 'hi': 'विरासत', 'hi-Latn': 'Inherited', 'hr': 'nasljedno pismo', 'ht': 'hérité', 'hu': 'Származtatott', 'ia': 'hereditate', 'id': 'Warisan', 'ig': 'Inherited', 'it': 'ereditato', 'ja': '基底文字の種別を継承する結合文字', 'ka': 'გადაღებული', 'kgp': 'ernanu', 'kk': 'мұра етілген', 'kk-Cyrl': 'мұра етілген', 'km': 'ស្នងកេរ្តិ៍', 'kn': 'ಇನ್\u200dಹೆರಿಟೆಡ್', 'ko': '구전 문자', 'lb': 'Geierfte Schrëftwäert', 'lo': 'ອິນເຮຮິດ', 'lt': 'paveldėtas', 'lv': 'mantotā', 'mk': 'наследено', 'ml': 'പാരമ്പര്യമായ', 'mn': 'Зинх', 'mr': 'वंशपरंपरागत', 'ms': 'Diwarisi', 'nb': 'nedarvet', 'ne': 'इन्हेरिटेड', 'nl': 'Overgeërfd', 'nn': 'nedarva', 'no': 'nedarvet', 'nqo': 'ߤߋߙߌߕߋ', 'or': 'ବଂଶଗତ', 'pl': 'dziedziczone', 'pt': 'herdado', 'rm': 'ertà', 'ro': 'moștenită', 'ru': 'унаследованная', 'sc': 'eredadu', 'sl': 'podedovan', 'so': 'Dhaxlay', 'sq': 'zin', 'sr': 'наследно писмо', 'sr-Cyrl': 'наследно писмо', 'sr-Latn': 'nasledno pismo', 'sv': 'ärvda', 'ta': 'பாரம்பரியமான', 'te': 'వారసత్వం', 'tg': 'Мерос', 'th': 'อินเฮอริต', 'to': 'tohinima hokosi', 'tr': 'Kalıtsal', 'ug': 'ئىرسىيەت ئاتالغۇ', 'uk': 'успадкована', 'vec': 'redità', 'vi': 'Chữ Kế thừa', 'xnr': 'विरासत', 'yo': 'Tí a jogún', 'yrl': 'tauxariwa', 'yue': '繼承文字（Unicode）', 'yue-Hans': '继承文字（Unicode）', 'yue-Hant': '繼承文字（Unicode）', 'zh': '遗传学术语', 'zh-Hans': '遗传学术语', 'zh-Hant': '繼承文字（Unicode）', 'zu': 'Okuthethwe'},
    'Zmth': {'af': 'Wiskundige notasie', 'ak': 'Nkontabudeɛ', 'am': 'የሂሳብ መግለጫ', 'ar': 'تدوين رياضي', 'as': 'গাণিতিক চিহ্ন', 'ast': 'escritura matemática', 'az': 'riyazi notasiya', 'az-Latn': 'riyazi notasiya', 'bal-Latn': 'Hesábi Neshán', 'be': 'матэматычныя знакі', 'be-tarask': 'матэматычныя знакі', 'bg': 'математически символи', 'bn': 'গাণিতিক চিহ্ন', 'br': 'notadur jedoniel', 'brx': 'सानखान्थियारि दैदेन्नाय हांखोफोर', 'bs': 'matematička notacija', 'bs-Cyrl': 'математичка нотација', 'bs-Latn': 'matematička notacija', 'ca': 'notació matemàtica', 'ca-ES-valencia': 'notació matemàtica', 'ccp': '𑄚𑄘𑄞𑄬𑄖𑄴 𑄌𑄨𑄚𑄴𑄦𑄧', 'ce': 'математикан маьӀна', 'chr': 'ᎠᏰᎦᎴᏴᏫᏍᎩ ᎠᎤᏓᏗᏍᏙᏗ', 'ckb': 'نیشانەی بیرکاری', 'cs': 'matematický zápis', 'cy': 'Nodiant Mathemategol', 'da': 'matematisk notation', 'de': 'Mathematische Notation', 'dsb': 'matematiski zapis', 'el': 'Μαθηματική σημειογραφία', 'el-polyton': 'Μαθηματική σημειογραφία', 'en': 'Mathematical Notation', 'en-Dsrt': '𐐣𐐰𐑃𐐲𐑋𐐰𐐻𐐲𐐿𐐲𐑊 𐐤𐐬𐐻𐐩𐑇𐐲𐑌', 'es': 'notación matemática', 'et': 'matemaatiline tähistus', 'eu': 'matematikako notazioa', 'fa': 'علائم ریاضی', 'ff-Adlm': '𞤍𞤵𞤪𞥆𞤢𞥄𞤲𞤺𞤮 𞤂𞤭𞤥𞤭𞤲𞤳𞤮', 'fi': 'matemaattinen', 'fil': 'Mathematical Notation', 'fo': 'støddfrøðilig teknskipan', 'fr': 'notation mathématique', 'fy': 'Wiskundige notatie', 'ga': 'Nodaireacht Mhatamaiticiúil', 'gd': 'Gnìomhairean matamataig', 'gl': 'notación matemática', 'gu': 'ગણિતીય સંકેતલિપિ', 'ha': 'Alamar Lissafi', 'he': 'סימון מתמטי', 'hi': 'गणितीय संकेतन', 'hi-Latn': 'Mathematical Notation', 'hr': 'matematičko znakovlje', 'hsb': 'matematiski zapis', 'ht': 'notation mathématique', 'hu': 'Matematikai jelrendszer', 'hy': 'մաթեմատիկական նշաններ', 'ia': 'notation mathematic', 'id': 'Notasi Matematika', 'ig': 'Mkpụrụ ọkwụ Mgbakọ', 'is': 'stærðfræðitákn', 'it': 'notazione matematica', 'ja': '数学記号', 'jv': 'Notasi Matematika', 'ka': 'მათემატიკური ნოტაცია', 'kea': 'notason matimátiku', 'kgp': 'vẽnhnĩkrén rá', 'kk': 'математикалық жазу', 'kk-Cyrl': 'математикалық жазу', 'km': 'និមិត្តសញ្ញាគណិតវិទ្យា', 'kn': 'ಗಣಿತ ಸಂಕೇತಲಿಪಿ', 'ko': '수학 기호', 'kok': 'गणिताची चिन्नां', 'kok-Deva': 'गणिताची चिन्नां', 'ku': 'nîşandana matematîkî', 'ky': 'Математикалык маани', 'lo': 'ເຄື່ອງໝາຍທາງຄະນິດສາດ', 'lt': 'matematiniai simboliai', 'lv': 'matemātiskais pieraksts', 'mai': 'गणितीय संकेतन', 'mi': 'Reo Tohu Pāngarau', 'mk': 'математичка нотација', 'ml': 'ഗണിത രൂപം', 'mn': 'математик тооллын систем', 'mr': 'गणितीय संकेतलिपी', 'ms': 'Tatatanda matematik', 'my': 'ဂဏန်းသင်္ချာ', 'nb': 'matematisk notasjon', 'ne': 'गणितीय चिन्ह', 'nl': 'wiskundige notatie', 'nn': 'matematisk notasjon', 'no': 'matematisk notasjon', 'nqo': 'ߘߡߊ߬ߟߐ߲ ߛߙߍߛߙߍߦߊߟߌ', 'or': 'ଗାଣିତିକ ନୋଟେସନ୍', 'pa': 'ਗਣਿਤ ਚਿੰਨ੍ਹ-ਲਿਪੀ', 'pa-Guru': 'ਗਣਿਤ ਚਿੰਨ੍ਹ-ਲਿਪੀ', 'pcm': 'Matimátiks Sains', 'pl': 'notacja matematyczna', 'ps': 'د ریاضیاتو نوټیشن', 'pt': 'notação matemática', 'qu': 'Matimatica Willay', 'rm': 'notaziun matematica', 'ro': 'notație matematică', 'ru': 'математические обозначения', 'sc': 'notatzione matemàtica', 'sd': 'رياضي جون نشانيون', 'sd-Arab': 'رياضي جون نشانيون', 'si': 'ගණිතමය සංකේත', 'sk': 'matematický zápis', 'sl': 'matematična znamenja', 'so': 'Aqoonsiga Xisaabta', 'sq': 'simbole matematikore', 'sr': 'математичка нотација', 'sr-Cyrl': 'математичка нотација', 'sr-Latn': 'matematička notacija', 'sv': 'matematisk notation', 'sw': 'Hati za kihisabati', 'ta': 'கணிதக்குறியீடு', 'te': 'గణిత సంకేతలిపి', 'tg': 'Аломати риёзӣ', 'th': 'เครื่องหมายทางคณิตศาสตร์', 'ti': 'ናይ ሒሳብ ምልክት', 'tk': 'Matematiki belgiler', 'to': 'tohinima fakamatematika', 'tr': 'Matematiksel Gösterim', 'ug': 'ماتېماتىكىلىق بەلگە', 'uk': 'математична', 'ur': 'ریاضی کی علامتیں', 'uz': 'matematik ifodalar', 'uz-Latn': 'matematik ifodalar', 'vec': 'notasion matemàtega', 'vi': 'Ký hiệu Toán học', 'xnr': 'गणिते आले संकेतन', 'yo': 'Àmì Ìṣèsìrò', 'yrl': 'matemátika kuatiara', 'yue': '數學符號', 'yue-Hans': '数学符号', 'yue-Hant': '數學符號', 'zh': '数学符号', 'zh-Hans': '数学符号', 'zh-Hant': '數學符號', 'zu': 'i-Mathematical Notation'},
    'Zsye': {'af': 'Emoji', 'ak': 'Yimogyi', 'am': 'ኢሞጂ', 'ar': 'إيموجي', 'as': 'ইম’জি', 'ast': 'emoji', 'az': 'emoji', 'az-Latn': 'emoji', 'bal-Latn': 'Emóji', 'be': 'эмодзі', 'be-tarask': 'эмодзі', 'bg': 'емоджи', 'bn': 'ইমোজি', 'br': 'fromlunioù', 'brx': 'ईम’जी', 'bs': 'emoji sličice', 'bs-Cyrl': 'емоџи', 'bs-Latn': 'emoji sličice', 'ca': 'emoji', 'ca-ES-valencia': 'emoji', 'ccp': '𑄃𑄨𑄟𑄮𑄎𑄨', 'ce': 'эмодзи', 'chr': 'ᎡᎼᏥ', 'ckb': 'ئیمۆجی', 'cs': 'emodži', 'cy': 'Emoji', 'da': 'emoji', 'de': 'Emoji', 'dsb': 'emoji', 'el': 'Emoji', 'el-polyton': 'Emoji', 'en': 'Emoji', 'es': 'emojis', 'et': 'emoji', 'eu': 'emojiak', 'fa': 'اموجی', 'ff-Adlm': '𞤐𞤺𞤮𞤼𞥆𞤭', 'fi': 'emoji-symbolit', 'fil': 'Emoji', 'fo': 'emoji', 'fr': 'emoji', 'ga': 'Emoji', 'gd': 'Emoji', 'gl': 'emojis', 'gu': 'ઇમોજી', 'ha': 'Alama ta hoto', 'he': 'אמוג׳י', 'hi': 'ईमोजी', 'hi-Latn': 'Emoji', 'hr': 'emotikoni', 'hsb': 'emoji', 'ht': 'emoji', 'hu': 'Emoji', 'hy': 'էմոջի', 'ia': 'emoji', 'id': 'Emoji', 'ig': 'Emoji', 'is': 'emoji-tákn', 'it': 'emoji', 'ja': '絵文字', 'jv': 'Emoji', 'ka': 'Emoji', 'kea': 'emoji', 'kgp': 'Emãji', 'kk': 'эмодзи', 'kk-Cyrl': 'эмодзи', 'km': 'សញ្ញាអារម្មណ៍', 'kn': 'ಎಮೋಜಿ', 'ko': '이모티콘', 'kok': 'ईमोजी', 'kok-Deva': 'ईमोजी', 'ku': 'emojî', 'ky': 'Быйтыкча', 'lo': 'ອີໂມຈິ', 'lt': 'jaustukai', 'lv': 'emocijzīmes', 'mai': 'इमोजी', 'mi': 'Emohi', 'mk': 'емоџи', 'ml': 'ഇമോജി', 'mn': 'эможи', 'mr': 'इमोजी', 'ms': 'Emoji', 'my': 'အီမိုဂျီ', 'nb': 'emoji', 'ne': 'इमोजी', 'nl': 'emoji', 'nn': 'emoji', 'no': 'emoji', 'nqo': 'ߋߡߏߖߌ߬', 'or': 'ଇମୋଜି', 'pa': 'ਇਮੋਜੀ', 'pa-Guru': 'ਇਮੋਜੀ', 'pcm': 'Ẹ́móji', 'pl': 'emoji', 'ps': 'ایموجي', 'pt': 'emoji', 'qu': 'Emoji', 'rm': 'emojis', 'ro': 'emoji', 'ru': 'эмодзи', 'sc': 'emoji', 'sd': 'ايموجي', 'sd-Arab': 'ايموجي', 'si': 'ඉමොජි', 'sk': 'emodži', 'sl': 'čustvenček', 'so': 'Calaamad Dareen Muujin', 'sq': 'emoji', 'sr': 'емоџи', 'sr-Cyrl': 'емоџи', 'sr-Latn': 'emodži', 'sv': 'emoji', 'sw': 'Emoji', 'syr': 'ܐܝܡܘܓܝ', 'ta': 'எமோஜி', 'te': 'ఎమోజీ', 'tg': 'Эмоҷи', 'th': 'อีโมจิ', 'ti': 'ኢሞጂ', 'tk': 'Emoji', 'to': 'tohinima fakatātā', 'tr': 'Emoji', 'uk': 'емодзі', 'ur': 'ایموجی', 'uz': 'emoji', 'uz-Latn': 'emoji', 'vec': 'emoji', 'vi': 'Biểu tượng cảm xúc', 'xnr': 'ईमोजी', 'yo': 'Émójì', 'yrl': 'Emuyi', 'yue': '表情符號', 'yue-Hans': '表情符号', 'yue-Hant': '表情符號', 'zh': '表情符号', 'zh-Hans': '表情符号', 'zh-Hant': '表情符號', 'zu': 'i-Emoji'},
    'Zsym': {'af': 'Simbole', 'ak': 'Ahyɛnsodeɛ', 'am': 'ምልክቶች', 'ar': 'رموز', 'as': 'প্ৰতীক', 'ast': 'símbolos', 'az': 'simvollar', 'az-Arab': 'سمبول\u200cلار', 'az-Latn': 'simvollar', 'bal-Latn': 'Neshán', 'be': 'сімвалы', 'be-tarask': 'сімвалы', 'bg': 'символи', 'bgn': 'سمولان', 'blo': 'ɩlamba', 'bn': 'প্রতীক', 'br': 'arouezioù', 'brx': 'नेर्सोनफोर', 'bs': 'simboli', 'bs-Cyrl': 'симболи', 'bs-Latn': 'simboli', 'ca': 'símbols', 'ca-ES-valencia': 'símbols', 'ccp': '𑄍𑄪𑄝𑄨𑄉𑄪𑄚𑄴', 'ce': 'символаш', 'chr': 'ᏗᎬᏟᎶᏍᏙᏗ', 'ckb': 'هێماکان', 'cs': 'symboly', 'cy': 'Symbolau', 'da': 'symboler', 'de': 'Symbole', 'dsb': 'symbole', 'dz': 'མཚན་རྟགས', 'ee': 'ŋɔŋlɔdzesiwo', 'el': 'Σύμβολα', 'el-polyton': 'Σύμβολα', 'en': 'Symbols', 'en-Dsrt': '𐐣𐐰𐑃𐐯𐑋𐐰𐐻𐐮𐐿𐐲𐑊 𐑌𐐬𐐻𐐩𐑇𐐲𐑌', 'es': 'símbolos', 'et': 'sümbolid', 'eu': 'ikurrak', 'fa': 'علائم', 'ff-Adlm': '𞤋𞤥𞥆𞤮𞤪𞤫', 'fi': 'symbolit', 'fil': 'Mga Simbolo', 'fo': 'tekin', 'fr': 'symboles', 'fy': 'Symbolen', 'ga': 'Siombailí', 'gd': 'Samhlaidhean', 'gl': 'símbolos', 'gu': 'પ્રતીકો', 'ha': 'Alamomi', 'he': 'סמלים', 'hi': 'चिह्न', 'hi-Latn': 'Symbols', 'hr': 'simboli', 'hsb': 'symbole', 'ht': 'symboles', 'hu': 'Szimbólum', 'hy': 'նշաններ', 'ia': 'symbolos', 'id': 'Simbol', 'ig': 'Akara', 'is': 'tákn', 'it': 'simboli', 'ja': '記号文字', 'jv': 'Simbol', 'ka': 'სიმბოლოები', 'kea': 'símbulus', 'kk': 'таңбалар', 'kk-Cyrl': 'таңбалар', 'km': 'និមិត្តសញ្ញា', 'kn': 'ಸಂಕೇತಗಳು', 'ko': '기호', 'kok': 'चिन्नां', 'kok-Deva': 'चिन्नां', 'ksh': '-Zeiche ävver kein Schreff-', 'ku': 'sembol', 'ky': 'Белгилер', 'lb': 'Symboler', 'lo': 'ສັນຍາລັກ', 'lrc': 'نئشوٙنە یا', 'lt': 'simbolių', 'lv': 'simboli', 'mai': 'संकेत', 'mi': 'Tohu', 'mk': 'симболи', 'ml': 'ചിഹ്നങ്ങൾ', 'mn': 'тэмдэг', 'mr': 'प्रतीक', 'ms': 'Simbol', 'my': 'သင်္ကေတ', 'nb': 'symboler', 'ne': 'प्रतीकहरू', 'nl': 'symbolen', 'nn': 'symbol', 'no': 'symboler', 'nqo': 'ߕߐ߰ߡߊ߬ߛߙߋ', 'or': 'ସଙ୍କେତଗୁଡ଼ିକ', 'pa': 'ਚਿੰਨ੍ਹ', 'pa-Guru': 'ਚਿੰਨ੍ਹ', 'pcm': 'Símbuls', 'pl': 'symbole', 'ps': 'سمبولونه', 'qu': 'Unanchakuna', 'rm': 'simbols', 'ro': 'simboluri', 'ru': 'символы', 'sc': 'sìmbulos', 'sd': 'نشانيون', 'sd-Arab': 'نشانيون', 'si': 'සංකේත', 'sk': 'symboly', 'sl': 'simboli', 'so': 'Calaamado', 'sq': 'me simbole', 'sr': 'симболи', 'sr-Cyrl': 'симболи', 'sr-Latn': 'simboli', 'sv': 'symboler', 'sw': 'Alama', 'ta': 'சின்னங்கள்', 'te': 'చిహ్నాలు', 'tg': 'Аломатҳо', 'th': 'สัญลักษณ์', 'ti': 'ምልክታት', 'tk': 'Nyşanlar', 'to': 'tohinima fakaʻilonga', 'tr': 'Sembol', 'ug': 'بەلگە', 'uk': 'символьна', 'ur': 'علامات', 'uz': 'belgilar', 'uz-Cyrl': 'Рамзлар', 'uz-Latn': 'belgilar', 'vec': 'sìnboli', 'vi': 'Biểu tượng | Ký hiệu', 'xnr': 'चिह्न', 'yo': 'Àwọn àmì', 'yue': '符號', 'yue-Hans': '符号', 'yue-Hant': '符號', 'zh': '符号', 'zh-Hans': '符号', 'zh-Hant': '符號', 'zu': 'amasimbuli'},
    'Zxxx': {'af': 'Ongeskrewe', 'ak': 'Deɛ wɔntwerɛeɛ', 'am': 'ያልተጻፈ', 'ar': 'غير مكتوب', 'as': 'অলিখিত', 'ast': 'non escritu', 'az': 'yazısız', 'az-Arab': 'یازیلمایان', 'az-Latn': 'yazısız', 'bal-Latn': 'Nebeshtah nabutagén syáhag', 'be': 'беспісьменная', 'be-tarask': 'беспісьменная', 'bg': 'без писменост', 'blo': 'kǝ ba ŋɔn na', 'bn': 'অলিখিত', 'bo': 'སྙན་བརྒྱུད། ཡིག་རིགས་སུ་མ་བཀོད་པའི་ཟིན་ཐོ།', 'br': 'anskrivet', 'brx': 'लिरै', 'bs': 'nepisani jezik', 'bs-Cyrl': 'неписани језик', 'bs-Latn': 'nepisani jezik', 'ca': 'sense escriptura', 'ca-ES-valencia': 'sense escriptura', 'ccp': '𑄚𑄧𑄣𑄬𑄇𑄴𑄈𑄳𑄠𑄬', 'ce': 'йоза доцу', 'ceb': 'Dili Sinulat', 'chr': 'ᎪᏪᎳᏅ ᏂᎨᏒᎾ', 'ckb': 'نەنووسراو', 'cs': 'bez zápisu', 'cv': 'ҫырусӑр', 'cy': 'Anysgrifenedig', 'da': 'uden skriftsprog', 'de': 'Schriftlos', 'doi': 'अनलिखत', 'dsb': 'bźez pisma', 'dz': 'ཡིག་ཐོག་མ་བཀོདཔ', 'ee': 'gbemaŋlɔ', 'el': 'Άγραφο', 'el-polyton': 'Άγραφο', 'en': 'Unwritten', 'en-Dsrt': '𐐊𐑌𐑉𐐮𐐻𐐲𐑌', 'es': 'no escrito', 'et': 'kirjakeeleta', 'eu': 'idatzi gabea', 'fa': 'نانوشته', 'ff-Adlm': '𞤀𞤧𞤱𞤭𞤲𞤣𞤢𞥄𞤯𞤵𞤲', 'fi': 'kirjoittamaton', 'fil': 'Hindi Nakasulat', 'fo': 'óskriva', 'fr': 'non écrit', 'fur': 'codiç pes lenghis no scritis', 'fy': 'Ongeschreven', 'ga': 'Neamhscríofa', 'gd': 'Gun sgrìobhadh', 'gl': 'non escrito', 'gsw': 'Schriftlosi Schpraach', 'gu': 'અલિખિત', 'ha': 'Ba rubutacce ba', 'he': 'לא כתוב', 'hi': 'अलिखित', 'hi-Latn': 'Unwritten', 'hr': 'jezik bez pismenosti', 'hsb': 'bjez pisma', 'ht': 'non écrit', 'hu': 'Íratlan nyelvek kódja', 'hy': 'չգրված', 'ia': 'non scripte', 'id': 'Tidak Tertulis', 'ie': 'ínscrit', 'ig': 'A na-edeghị ede', 'ii': 'ꁱꀋꉆꌠ', 'is': 'óskrifað', 'it': 'non scritto', 'ja': '非表記', 'jgo': 'ntúu yi pɛ́ ká ŋwaꞌnε', 'jv': 'Ora Ketulis', 'ka': 'უმწერლობო', 'kaa': 'жазыўы жоқ', 'kaa-Cyrl': 'жазыўы жоқ', 'kea': 'nãu skritu', 'kgp': 'vẽnhrá-tũ', 'kk': 'жазусыз', 'kk-Cyrl': 'жазусыз', 'km': 'គ្មានការសរសេរ', 'kn': 'ಅಲಿಖಿತ', 'ko': '구전', 'kok': 'अलिखीत', 'kok-Deva': 'अलिखीत', 'ks': 'لیکھنَے', 'ks-Arab': 'لیکھنَے', 'ks-Deva': 'गेर तहरीर', 'ksh': '-jaa keij Schreff-', 'ku': 'nenivîskî', 'kxv': 'raciaahalee', 'kxv-Deva': 'राचा-आ-आत्ति', 'kxv-Latn': 'raciaahalee', 'kxv-Telu': 'లిపి లేని', 'ky': 'Жазылбаган', 'lb': 'Ouni Schrëft', 'lld': 'nia scrit', 'lo': 'ບໍ່ມີພາສາຂຽນ', 'lrc': 'نیسئسە نأبیە', 'lt': 'neparašyta', 'lv': 'bez rakstības', 'mai': 'अलिखित', 'mgo': 'ngam choʼ', 'mi': 'Tuhikore', 'mk': 'без писмо', 'ml': 'എഴുതപ്പെടാത്തത്', 'mn': 'бичигдээгүй', 'mni': 'ইদবা', 'mni-Beng': 'ইদবা', 'mr': 'अलिखित', 'ms': 'Tidak ditulis', 'mt': 'Mhux Miktub', 'my': 'ထုံးတမ်းသဖွယ်လိုက်နာလျက်ရှိသော', 'nb': 'språk uten skrift', 'ne': 'नलेखिएको', 'nl': 'ongeschreven', 'nn': 'språk utan skrift', 'no': 'språk uten skrift', 'nqo': 'ߛߓߍߓߊߟߌ', 'om': 'Kan Hin Barreeffamne', 'or': 'ଅଲିଖିତ', 'os': 'Нӕфысгӕ', 'pa': 'ਅਲਿਖਤ', 'pa-Guru': 'ਅਲਿਖਤ', 'pcm': 'Wétín Dẹm Nó Rait', 'pl': 'język bez systemu pisma', 'ps': 'ناليکلی', 'pt': 'ágrafo', 'qu': 'Mana qillqasqa', 'rm': 'betg scrit', 'ro': 'nescrisă', 'ru': 'нет письменности', 'sa': 'अलिखित:', 'sah': 'Суруллубатах', 'sat': 'ᱵᱟᱝ ᱚᱞ ᱟᱠᱟᱱ', 'sat-Olck': 'ᱵᱟᱝ ᱚᱞ ᱟᱠᱟᱱ', 'sc': 'no iscritu', 'sd': 'اڻ لکيل', 'sd-Arab': 'اڻ لکيل', 'sd-Deva': 'अणलिखियल', 'se': 'orrut chállojuvvot', 'si': 'අලිඛිත', 'sk': 'bez zápisu', 'sl': 'nenapisano', 'so': 'Aan la qorin', 'sq': 'i pashkruar', 'sr': 'неписани језик', 'sr-Cyrl': 'неписани језик', 'sr-Latn': 'nepisani jezik', 'su': 'Non-tulisan', 'su-Latn': 'Non-tulisan', 'sv': 'oskrivet språk', 'sw': 'Haijaandikwa', 'syr': 'ܠܝܬ ܟܬܝܒܘܬܐ', 'ta': 'எழுதப்படாதது', 'te': 'లిపి లేని', 'tg': 'Нонавишта', 'th': 'ไม่มีภาษาเขียน', 'ti': 'ዘይተጻሕፈ', 'tk': 'Ýazuwsyz', 'to': 'tohinima taʻetohitohiʻi', 'tok': 'sitelen ala', 'tpi': 'Tok i no raitim yet', 'tr': 'Yazılı Olmayan', 'tt': 'язусыз', 'ug': 'يېزىلمىغان', 'uk': 'безписемна', 'ur': 'غیر تحریر شدہ', 'uz': 'yozuvsiz', 'uz-Cyrl': 'Ёзилмаган', 'uz-Latn': 'yozuvsiz', 'vec': 'miga scrito', 'vi': 'Chưa có chữ viết', 'wae': 'Schriftlos', 'wo': 'Luñ bindul', 'xh': 'Engabhalwanga', 'xnr': 'नी लखोया', 'yo': 'Aikọsilẹ', 'yrl': 'yũpinimasáĩma', 'yue': '非書寫語言', 'yue-Hans': '非书写语言', 'yue-Hant': '非書寫語言', 'zh': '非书面文字', 'zh-Hans': '非书面文字', 'zh-Hant': '非書寫語言', 'zu': 'okungabhaliwe'},
    'Zyyy': {'af': 'Algemeen', 'ak': 'obiara nim', 'am': 'የጋራ', 'ar': 'عام', 'as': 'কোমোন', 'ast': 'común', 'az': 'ümumi yazı', 'az-Arab': 'اوْرتاق', 'az-Latn': 'ümumi yazı', 'bal-Latn': 'Hórén', 'be': 'агульнае', 'be-tarask': 'агульнае', 'bg': 'обща', 'bn': 'সাধারন', 'br': 'boutin', 'brx': 'साधारन', 'bs': 'zajedničko pismo', 'bs-Cyrl': 'заједничко писмо', 'bs-Latn': 'zajedničko pismo', 'ca': 'comú', 'ca-ES-valencia': 'comú', 'ccp': '𑄃𑄧𑄎𑄬𑄃𑄧𑄌𑄴', 'ce': 'массара а тӀеэцна', 'chr': 'ᏯᏃᏉ ᏱᎬᏍᏛᏭ', 'ckb': 'باو', 'cs': 'obecné', 'cy': 'Cyffredin', 'da': 'fælles', 'de': 'Unbestimmt', 'dsb': 'powšykne', 'dz': 'སྤྱིཡིག', 'ee': 'gbeŋɔŋlɔ bɔbɔ', 'el': 'Κοινό', 'el-polyton': 'Κοινό', 'en': 'Common', 'en-Dsrt': '𐐗𐐱𐑋𐐲𐑌', 'es': 'común', 'et': 'üldine', 'eu': 'ohikoa', 'fa': 'مشترک', 'ff-Adlm': '𞤑𞤢𞤬𞤵', 'fi': 'määrittämätön', 'fil': 'Karaniwan', 'fo': 'vanlig', 'fr': 'commun', 'fur': 'comun', 'fy': 'Algemeen', 'ga': 'Coitianta', 'gd': 'Coitcheann', 'gl': 'común', 'gsw': 'Unbeschtimmt', 'gu': 'સામાન્ય', 'ha': 'Gama-gari', 'he': 'רגיל', 'hi': 'सामान्य', 'hi-Latn': 'Common', 'hr': 'zajedničko pismo', 'hsb': 'powšitkowne', 'ht': 'commun', 'hu': 'Meghatározatlan', 'hy': 'ընդհանուր', 'ia': 'commun', 'id': 'Umum', 'ig': 'Common', 'is': 'almennt', 'it': 'comune', 'ja': '共通文字', 'jv': 'Umum', 'ka': 'ზოგადი', 'kea': 'komun', 'kgp': 'kãmũ', 'kk': 'жалпы', 'kk-Cyrl': 'жалпы', 'km': 'ទូទៅ', 'kn': 'ಸಾಮಾನ್ಯ', 'ko': '일반 문자', 'kok': 'सामान्य', 'kok-Deva': 'सामान्य', 'ks': 'عام', 'ks-Arab': 'عام', 'ksh': '-öhnß en Schreff-', 'ku': 'hevpar', 'ky': 'Жалпы', 'lb': 'Onbestëmmt', 'lo': 'ສາມັນ', 'lrc': 'جائوفتاأ', 'lt': 'bendri', 'lv': 'vispārējā', 'mai': 'सामान्य', 'mi': 'Komona', 'mk': 'општо', 'ml': 'സാധാരണ', 'mn': 'нийтлэг', 'mr': 'सामान्य', 'ms': 'Lazim', 'mt': 'Komuni', 'my': 'အများနှင့်သက်ဆိုင်သော', 'nb': 'felles', 'ne': 'साझा', 'nl': 'algemeen', 'nn': 'felles', 'no': 'felles', 'nqo': 'ߞߏߡߎ߲', 'or': 'ସାଧାରଣ', 'pa': 'ਸਧਾਰਨ', 'pa-Guru': 'ਸਧਾਰਨ', 'pcm': 'Jẹ́náral', 'pl': 'wspólne', 'ps': 'عام', 'pt': 'comum', 'qu': 'Common Simi', 'rm': 'betg determinà', 'ro': 'comună', 'ru': 'общепринятая', 'sc': 'comune', 'sd': 'ڪامن', 'sd-Arab': 'ڪامن', 'si': 'පොදු', 'sk': 'všeobecné', 'sl': 'splošno', 'so': 'Caadi ahaan', 'sq': 'i zakonshëm', 'sr': 'заједничко писмо', 'sr-Cyrl': 'заједничко писмо', 'sr-Latn': 'zajedničko pismo', 'sv': 'gemensamma', 'sw': 'Kawaida', 'ta': 'பொது', 'te': 'సామాన్య', 'tg': 'Умумӣ', 'th': 'สามัญ', 'ti': 'ልሙድ', 'tk': 'Umumy', 'to': 'tohinima fakatatau', 'tr': 'Ortak', 'ug': 'ئورتاق', 'uk': 'звичайна', 'ur': 'عام', 'uz': 'umumiy', 'uz-Cyrl': 'Умумий', 'uz-Latn': 'umumiy', 'vec': 'comun', 'vi': 'Chung', 'xnr': 'सामान्य', 'yo': 'Wọ́pọ̀', 'yrl': 'mayewera', 'yue': '一般文字', 'yue-Hans': '一般文字', 'yue-Hant': '一般文字', 'zh': '通用', 'zh-Hans': '通用', 'zh-Hant': '一般文字', 'zu': 'jwayelekile'},
    'Zzzz': {'af': 'Onbekende skryfstelsel', 'ak': 'Deɛ yɛnnim', 'am': 'ያልታወቀ ፊደል', 'ar': 'نظام كتابة غير معروف', 'as': 'অজ্ঞাত লিপি', 'ast': 'escritura desconocida', 'az': 'tanınmayan yazı', 'az-Arab': 'تانینمایان خط', 'az-Latn': 'tanınmayan yazı', 'bal-Latn': 'Kódh nakortagén syáhag', 'be': 'невядомае пісьмо', 'be-tarask': 'невядомае пісьмо', 'bg': 'непозната писменост', 'bgn': 'نامالومین خط', 'blo': 'ʊŋɔn nɖee kʊyɔʊ ʊ mana ma', 'bn': 'অজানা লিপি', 'br': 'skritur dianav', 'brx': 'मिथियै हांखो', 'bs': 'nepoznato pismo', 'bs-Cyrl': 'непознато писмо', 'bs-Latn': 'nepoznato pismo', 'ca': 'escriptura desconeguda', 'ca-ES-valencia': 'escriptura desconeguda', 'ccp': '𑄦𑄧𑄝𑄧𑄢𑄴𑄚𑄧𑄛𑄨𑄠𑄬 𑄦𑄧𑄢𑄧𑄇𑄴', 'ce': 'доьвзуш доцу йоза', 'ceb': 'Wala Mailhing Script', 'chr': 'ᏄᏬᎵᏍᏛᎾ ᎠᏍᏓᏩᏛᏍᏙᏗ', 'ckb': 'خەتی نەناسراو', 'cs': 'neznámé písmo', 'cv': 'паллӑ мар ҫырулах', 'cy': 'Sgript anhysbys', 'da': 'ukendt skriftsprog', 'de': 'Unbekannte Schrift', 'doi': 'अनजांती लिपि', 'dsb': 'njeznate pismo', 'dz': 'ངོ་མ་ཤེས་པའི་ཡི་གུ', 'ee': 'gbeŋɔŋlɔ manya', 'el': 'Άγνωστη γραφή', 'el-polyton': 'Άγνωστη γραφή', 'en': 'Unknown Script', 'en-Dsrt': '𐐊𐑌𐐬𐑌 𐐬𐑉 𐐆𐑌𐑂𐐰𐑊𐐮𐐼 𐐝𐐿𐑉𐐮𐐹𐐻', 'es': 'alfabeto desconocido', 'et': 'määramata kiri', 'eu': 'idazkera ezezaguna', 'fa': 'خط نامشخص', 'ff-Adlm': '𞤄𞤭𞤲𞤣𞤭 𞤀𞤧-𞤢𞤲𞤣𞤢𞥄𞤯𞤭', 'fi': 'tuntematon kirjoitusjärjestelmä', 'fil': 'Hindi Kilalang Script', 'fo': 'ókend skrift', 'fr': 'écriture inconnue', 'fur': 'codiç par scrituris no codificadis', 'fy': 'Onbekend schriftsysteem', 'ga': 'Script Anaithnid', 'gd': 'Litreadh neo-aithnichte', 'gl': 'sistema de escritura descoñecido', 'gsw': 'Uncodiirti Schrift', 'gu': 'અજ્ઞાત લિપિ', 'ha': 'Rubutun da ba sani ba', 'he': 'כתב שאינו ידוע', 'hi': 'अज्ञात लिपि', 'hi-Latn': 'Unknown Script', 'hr': 'nepoznato pismo', 'hsb': 'njeznate pismo', 'ht': 'écriture inconnue', 'hu': 'Ismeretlen írásrendszer', 'hy': 'անհայտ գիր', 'ia': 'scriptura incognite', 'id': 'Skrip Tidak Dikenal', 'ie': 'ínconosset scritura', 'ig': 'Edemede a na-amaghi', 'ii': 'ꅉꀋꐚꌠꁱꂷ', 'is': 'óþekkt letur', 'it': 'scrittura sconosciuta', 'ja': '不明な文字', 'jgo': 'ntɛ-ŋwaꞌnɛ yí pɛ́ ká kɛ́ jí', 'jv': 'Skrip Ora Dikenali', 'ka': 'უცნობი დამწერლობა', 'kaa': 'белгисиз жазыў', 'kaa-Cyrl': 'белгисиз жазыў', 'kea': 'skrita diskonxedu', 'kgp': 'vẽnhrá ki kagtĩg', 'kk': 'белгісіз жазу', 'kk-Cyrl': 'белгісіз жазу', 'km': 'អក្សរមិនស្គាល់', 'kn': 'ಅಪರಿಚಿತ ಲಿಪಿ', 'ko': '알 수 없는 문자', 'kok': 'अज्ञात लिपी', 'kok-Deva': 'अज्ञात लिपी', 'ks': 'نامولوٗم رَسم الخط', 'ks-Arab': 'نامولوٗم رَسم الخط', 'ks-Deva': 'गेर तहरीर स्क्रिप्ट', 'ksh': '-onbikannte Schreff-', 'ku': 'alfabeya nenas', 'kxv': 'puṇāātiakr', 'kxv-Deva': 'पुण्-आँऽति ऑकाॅराॅ', 'kxv-Latn': 'puṇāātiakr', 'kxv-Telu': 'తెలియని లిపి', 'ky': 'Белгисиз жазуу', 'lb': 'Oncodéiert Schrëft', 'lld': 'scritöra nia conesciüda', 'lo': 'ແບບຂຽນທີ່ບໍ່ຮູ້ຈັກ', 'lrc': 'نیسئسە نادیار', 'lt': 'nežinomi rašmenys', 'lv': 'nezināma rakstība', 'mai': 'अज्ञात लिपि', 'mgo': 'abo ŋwaʼri tisɔʼ', 'mi': 'Momotuhi Tē Mōhiotia', 'mk': 'непознато писмо', 'ml': 'അജ്ഞാത ലിപി', 'mn': 'тодорхойгүй бичиг', 'mni': 'মশকখংদবা স্ক্রিপ্ট', 'mni-Beng': 'মশকখংদবা স্ক্রিপ্ট', 'mr': 'अज्ञात लिपी', 'ms': 'Tulisan Tidak Diketahui', 'mt': 'Kitba Mhux Magħrufa', 'my': 'မသိ သို့မဟုတ် မရှိသော စကားလုံး', 'nb': 'ukjent skrift', 'ne': 'अज्ञात लिपि', 'nl': 'onbekend schriftsysteem', 'nn': 'ukjend skrift', 'no': 'ukjent skrift', 'nqo': 'ߛߓߍߟߌ߫ ߡߊߟߐ߲ߓߊߟߌ', 'om': 'Barreeffama Hin Beekamne', 'or': 'ଅଜଣା ଲିପି', 'os': 'Нӕзонгӕ скрипт', 'pa': 'ਅਣਪਛਾਤੀ ਲਿਪੀ', 'pa-Guru': 'ਅਣਪਛਾਤੀ ਲਿਪੀ', 'pcm': 'Raítín Sístẹm Wé Nóbọ́di Sabí', 'pl': 'nieznane pismo', 'prg': 'niwaīstan skriptan', 'ps': 'نامعلومه سکرېپټ', 'pt': 'escrita desconhecida', 'qu': 'Mana yachasqa Qillqa', 'rm': 'scrittira nunenconuschenta', 'ro': 'scriere necunoscută', 'ru': 'неизвестная письменность', 'sa': 'अज्ञात लिपि:', 'sah': 'Биллибэт сурук', 'sat': 'ᱵᱟᱝ ᱩᱨᱩᱢ ᱪᱤᱠᱤ', 'sat-Olck': 'ᱵᱟᱝ ᱩᱨᱩᱢ ᱪᱤᱠᱤ', 'sc': 'iscritura disconnota', 'sd': 'اڻڄاتل لکت', 'sd-Arab': 'اڻڄاتل لکت', 'sd-Deva': 'अणजा॒तल लिखत', 'se': 'dovdameahttun chállin', 'si': 'නොදත් අකුරු මාලාව', 'sk': 'neznáme písmo', 'sl': 'neznan ali neveljaven zapis', 'so': 'Far aan la aqoon amase aan saxnayn', 'sq': 'i panjohur', 'sr': 'непознато писмо', 'sr-Cyrl': 'непознато писмо', 'sr-Latn': 'nepoznato pismo', 'su': 'Tulisan Teu Dikenal', 'su-Latn': 'Tulisan Teu Dikenal', 'sv': 'okänt skriftsystem', 'sw': 'Hati isiyojulikana', 'syr': 'ܠܝܬ ܛܟܣܐ ܕܟܬܝܒܬܐ', 'ta': 'அறியப்படாத ஸ்கிரிப்ட்', 'te': 'తెలియని లిపి', 'tg': 'Скрипти номаълум', 'th': 'ตัวอักษรที่ไม่รู้จัก', 'ti': 'ዘይፍለጥ ኢደ ጽሑፍ', 'tk': 'Näbelli elipbiý', 'to': 'tohinima taʻeʻiloa', 'tok': 'sitelen ante', 'tpi': 'Tok i no gat kod yet', 'tr': 'Bilinmeyen Alfabe', 'tt': 'билгесез язу', 'ug': 'يوچۇن يېزىق', 'uk': 'невідома система письма', 'ur': 'نامعلوم رسم الخط', 'uz': 'noma’lum yozuv', 'uz-Cyrl': 'Номаълум шрифт', 'uz-Latn': 'noma’lum yozuv', 'vec': 'desconosùo', 'vi': 'Chữ viết không xác định', 'wae': 'Unkodierti Schrift', 'wo': 'Mbind muñ xamul', 'xh': 'Ulwimi Olungaziwayo', 'xnr': 'अणजाण लिपि', 'yo': 'Ìṣọwọ́kọ̀wé àìmọ̀', 'yrl': 'yũpinimasá ũbawa uyukuau', 'yue': '未知文字', 'yue-Hans': '未知文字', 'yue-Hant': '未知文字', 'zh': '未知文字', 'zh-Hans': '未知文字', 'zh-Hant': '未知文字', 'zu': 'iskripthi esingaziwa'},
    'aa': {'af': 'Afar', 'am': 'አፋርኛ', 'ar': 'الأفارية', 'as': 'আফাৰ', 'ast': 'afar', 'az': 'afar', 'az-Cyrl': 'афар', 'az-Latn': 'afar', 'bal-Latn': 'Apar', 'be': 'афарская', 'be-tarask': 'афарская', 'bg': 'афарски', 'bgn': 'افار', 'bn': 'আফার', 'br': 'afar', 'bs': 'afarski', 'bs-Cyrl': 'афарски', 'bs-Latn': 'afarski', 'ca': 'àfar', 'ca-ES-valencia': 'àfar', 'ccp': '𑄃𑄜𑄢𑄴', 'ce': 'афарийн', 'chr': 'ᎠᏩᎳ', 'ckb': 'ئەفار', 'cs': 'afarština', 'cy': 'Affareg', 'da': 'afar', 'de': 'Afar', 'dsb': 'afaršćina', 'dz': 'ཨ་ཕར་ཁ', 'el': 'Αφάρ', 'el-polyton': 'Αφάρ', 'en': 'Afar', 'eo': 'afara', 'es': 'afar', 'et': 'afari', 'eu': 'afarera', 'fa': 'آفاری', 'ff-Adlm': '𞤀𞤬𞤢𞥄𞤪𞤫', 'fi': 'afar', 'fil': 'Afar', 'fo': 'afar', 'fr': 'afar', 'fur': 'afar', 'fy': 'Afar', 'ga': 'Afáiris', 'gd': 'Afar', 'gl': 'afar', 'gsw': 'Afar', 'gu': 'અફાર', 'he': 'אפארית', 'hi': 'अफ़ार', 'hi-Latn': 'Afar', 'hr': 'afarski', 'hsb': 'afaršćina', 'ht': 'afar', 'hu': 'afar', 'hy': 'աֆարերեն', 'ia': 'afar', 'id': 'Afar', 'ig': 'Afar', 'is': 'afár', 'it': 'afar', 'ja': 'アファル語', 'ka': 'აფარი', 'kgp': 'afar', 'kk': 'афар тілі', 'kk-Cyrl': 'афар тілі', 'km': 'អាហ្វារ', 'kn': 'ಅಫಾರ್', 'ko': '아파르어', 'kok': 'अफार', 'kok-Deva': 'अफार', 'ks': 'اَفار', 'ks-Arab': 'اَفار', 'ksh': 'Afahresch', 'ku': 'afarî', 'ky': 'афарча', 'lb': 'Afar', 'lld': 'afar', 'lo': 'ອະຟາ', 'lt': 'afarų', 'lv': 'afāru', 'mk': 'афарски', 'ml': 'അഫാർ', 'mn': 'афар', 'mr': 'अफार', 'ms': 'Afar', 'mt': 'Afar', 'my': 'အာဖာ', 'nb': 'afar', 'ne': 'अफार', 'nl': 'Afar', 'nn': 'afar', 'no': 'afar', 'or': 'ଅଫାର୍', 'pa': 'ਅਫ਼ਾਰ', 'pa-Guru': 'ਅਫ਼ਾਰ', 'pl': 'afar', 'ps': 'افري', 'pt': 'afar', 'rm': 'afar', 'ro': 'afar', 'ru': 'афарский', 'sd': 'افار', 'sd-Arab': 'افار', 'si': 'අෆාර්', 'sk': 'afarčina', 'sl': 'afarščina', 'smn': 'afar', 'sq': 'afarisht', 'sr': 'афарски', 'sr-Cyrl': 'афарски', 'sr-Latn': 'afarski', 'sv': 'afar', 'sw': 'Kiafar', 'ta': 'அஃபார்', 'te': 'అఫార్', 'th': 'อะฟาร์', 'ti': 'አፋር', 'tk': 'afar dili', 'to': 'lea fakaʻafāla', 'tr': 'Afar', 'ug': 'ئافارچە', 'uk': 'афарська', 'ur': 'افار', 'uz': 'afar', 'uz-Cyrl': 'афарча', 'uz-Latn': 'afar', 'vec': 'afar', 'vi': 'Tiếng Afar', 'xnr': 'अफ़ार', 'yi': 'אַפֿאַר', 'yrl': 'afari', 'yue': '阿法文', 'yue-Hans': '阿法文', 'yue-Hant': '阿法文', 'zh': '阿法尔语', 'zh-Hans': '阿法尔语', 'zh-Hant': '阿法文', 'zu': 'isi-Afar'},
    'aaa': {'en': 'Ghotuo'},
    'aab': {'en': 'Alumu-Tesu'},
    'aac': {'en': 'Ari'},
    'aad': {'en': 'Amal'},
    'aae': {'en': 'Arbëreshë Albanian'},
    'aaf': {'en': 'Aranadan'},
    'aag': {'en': 'Ambrak'},
    'aah': {'en': "Abu' Arapesh"},
    'aai': {'en': 'Arifama-Miniafia'},
    'aak': {'en': 'Ankave'},
    'aal': {'en': 'Afade'},
    'aam': {'en': 'Aramanik'},
    'aan': {'en': 'Anambé'},
    'aao': {'en': 'Algerian Saharan Arabic'},
    'aap': {'en': 'Pará Arára'},
    'aaq': {'en': 'Eastern Abnaki'},
    'aas': {'en': 'Aasáx'},
    'aat': {'en': 'Arvanitika Albanian'},
    'aau': {'en': 'Abau'},
    'aav': {'en': 'Austro-Asiatic languages'},
    'aav-ban-pro': {'en': 'Proto-Bahnaric'},
    'aav-khm-pro': {'en': 'Proto-Khmuic'},
    'aav-pro': {'en': 'Proto-Austro-Asiatic'},
    'aaw': {'en': 'Solong'},
    'aax': {'en': 'Mandobo Atas'},
    'aaz': {'en': 'Amarasi'},
    'ab': {'ab': 'Аԥсшәа', 'af': 'Abkasies', 'am': 'አብሐዚኛ', 'ar': 'الأبخازية', 'as': 'আবখাজিয়ান', 'ast': 'abkhazianu', 'az': 'abxaz', 'az-Arab': 'آبخازجا', 'az-Cyrl': 'абхаз', 'az-Latn': 'abxaz', 'bal-Latn': 'Abkházi', 'be': 'абхазская', 'be-tarask': 'абхазская', 'bg': 'абхазки', 'bgn': 'آبخازی', 'bn': 'আবখাজিয়ান', 'br': 'abkhazeg', 'brx': 'आब्खाजियान', 'bs': 'abhaski', 'bs-Cyrl': 'абказијски', 'bs-Latn': 'abhaski', 'ca': 'abkhaz', 'ca-ES-valencia': 'abkhaz', 'ccp': '𑄃𑄝𑄴𑄈𑄎𑄨𑄠𑄚𑄴', 'ce': 'абхазхойн', 'chr': 'ᎠᏆᏏᎠᏂ', 'ckb': 'ئەبخازی', 'cs': 'abcházština', 'cy': 'Abchaseg', 'da': 'abkhasisk', 'de': 'Abchasisch', 'dsb': 'abchazšćina', 'dz': 'ཨཱབ་ཁ་ཟི་ཡ་ཁ', 'ee': 'abkhaziagbe', 'el': 'Αμπχαζικά', 'el-polyton': 'Αμπχαζικά', 'en': 'Abkhazian', 'eo': 'abĥaza', 'es': 'abjasio', 'et': 'abhaasi', 'eu': 'abkhaziera', 'fa': 'آبخازی', 'ff-Adlm': '𞤀𞤦𞤳𞤢𞥄𞤧𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'abhaasi', 'fil': 'Abkhazian', 'fo': 'abkhasiskt', 'fr': 'abkhaze', 'fur': 'abcazian', 'fy': 'Abchazysk', 'ga': 'Abcáisis', 'gd': 'Abchasais', 'gl': 'abkhazo', 'gsw': 'Abchasisch', 'gu': 'અબખાજિયન', 'ha': 'Abkhazian', 'he': 'אבחזית', 'hi': 'अब्ख़ाज़ियन', 'hi-Latn': 'Abkhazian', 'hr': 'abhaski', 'hsb': 'abchazišćina', 'ht': 'abkhaze', 'hu': 'abház', 'hy': 'աբխազերեն', 'ia': 'abkhazo', 'id': 'Abkhaz', 'ig': 'Abkaziani', 'is': 'abkasíska', 'it': 'abcaso', 'ja': 'アブハズ語', 'jv': 'Abkhazian', 'ka': 'აფხაზური', 'kea': 'abkáziu', 'kgp': 'amekaso', 'kk': 'абхаз тілі', 'kk-Cyrl': 'абхаз тілі', 'km': 'អាប់ខាហ៊្សាន', 'kn': 'ಅಬ್ಖಾಜಿಯನ್', 'ko': '압카즈어', 'kok': 'अबखेज़ियन', 'kok-Deva': 'अबखेज़ियन', 'ks': 'اَبخازِیان', 'ks-Arab': 'اَبخازِیان', 'ksh': 'Abchahsesch', 'ku': 'abxazî', 'ky': 'абхазча', 'lb': 'Abchasesch', 'lkt': 'Abkhaz Iyápi', 'lld': 'abkhasich', 'lo': 'ແອບຄາຊຽນ', 'lrc': 'آذأربایئجانی', 'lt': 'abchazų', 'lv': 'abhāzu', 'mai': 'अबखाजियान', 'mi': 'Apakāhiana', 'mk': 'апхаски', 'ml': 'അബ്\u200cഖാസിയൻ', 'mn': 'абхаз', 'mr': 'अबखेजियन', 'ms': 'Abkhazia', 'mt': 'Abkażjan', 'my': 'အဘ်ခါဇီရာ', 'mzn': 'آبخازی', 'nb': 'abkhasisk', 'ne': 'अब्खाजियाली', 'nl': 'Abchazisch', 'nn': 'abkhasisk', 'no': 'abkhasisk', 'or': 'ଆବ୍ଖାଜିଆନ୍', 'os': 'абхазаг', 'pa': 'ਅਬਖਾਜ਼ੀਅਨ', 'pa-Guru': 'ਅਬਖਾਜ਼ੀਅਨ', 'pcm': 'Abkházian', 'pl': 'abchaski', 'ps': 'ابخازي', 'pt': 'abcázio', 'qu': 'Abjasia', 'rm': 'abchasian', 'ro': 'abhază', 'ru': 'абхазский', 'sah': 'Абхаастыы', 'sat': 'ᱟᱹᱵᱠᱷᱟᱡᱤᱭᱟᱹᱱ', 'sat-Olck': 'ᱟᱹᱵᱠᱷᱟᱡᱤᱭᱟᱹᱱ', 'sc': 'abcasu', 'sd': 'ابقازیان', 'sd-Arab': 'ابقازیان', 'si': 'ඇබ්කාසියානු', 'sk': 'abcházčina', 'sl': 'abhaščina', 'smn': 'abhasiakielâ', 'so': 'U dhashay Abkhazia', 'sq': 'abkazisht', 'sr': 'абхаски', 'sr-Cyrl': 'абхаски', 'sr-Latn': 'abhaski', 'sv': 'abchaziska', 'sw': 'Kiabkhazi', 'syr': 'ܐܒܟܐܙܝܬ', 'ta': 'அப்காஜியான்', 'te': 'అబ్ఖాజియన్', 'th': 'อับฮาเซีย', 'ti': 'ኣብካዝኛ', 'tk': 'abhaz dili', 'to': 'lea fakaʻapakasia', 'tr': 'Abhazca', 'ug': 'ئابخازچە', 'uk': 'абхазька', 'ur': 'ابقازیان', 'uz': 'abxaz', 'uz-Cyrl': 'абхазча', 'uz-Latn': 'abxaz', 'vec': 'abcazo', 'vi': 'Tiếng Abkhazia', 'wae': 'Abčasiš', 'xnr': 'अब्ख़ाज़ियन', 'yo': 'Èdè Abasia', 'yrl': 'abikasiyu', 'yue': '阿布哈茲文', 'yue-Hans': '阿布哈兹文', 'yue-Hant': '阿布哈茲文', 'zh': '阿布哈西亚语', 'zh-Hans': '阿布哈西亚语', 'zh-Hant': '阿布哈茲文', 'zu': 'isi-Abkhazian'},
    'aba': {'en': 'Abé'},
    'abb': {'en': 'Bankon'},
    'abc': {'en': 'Ambala Ayta'},
    'abd': {'en': 'Manide'},
    'abe': {'en': 'Western Abnaki'},
    'abf': {'en': 'Abai Sungai'},
    'abg': {'en': 'Abaga'},
    'abh': {'en': 'Tajiki Arabic'},
    'abi': {'en': 'Abidji'},
    'abj': {'en': 'Aka-Bea'},
    'abl': {'en': 'Lampung Nyo'},
    'abm': {'en': 'Abanyom'},
    'abn': {'en': 'Abua'},
    'abo': {'en': 'Abon'},
    'abp': {'en': 'Abellen Ayta'},
    'abq': {'en': 'Abaza'},
    'abr': {'en': 'Abron'},
    'abs': {'en': 'Ambonese Malay'},
    'abt': {'en': 'Ambulas'},
    'abu': {'en': 'Abure'},
    'abv': {'en': 'Baharna Arabic'},
    'abw': {'en': 'Pal'},
    'abx': {'en': 'Inabaknon'},
    'aby': {'en': 'Aneme Wake'},
    'abz': {'en': 'Abui'},
    'aca': {'en': 'Achagua'},
    'acb': {'en': 'Áncá'},
    'acd': {'en': 'Gikyode'},
    'ace': {'af': 'Atsjenees', 'am': 'አቻይንኛ', 'ar': 'الأتشينيزية', 'as': 'আচিনিজ', 'ast': 'achinés', 'az': 'akin', 'az-Cyrl': 'акин', 'az-Latn': 'akin', 'be': 'ачэх', 'be-tarask': 'ачэх', 'bg': 'ачешки', 'bgn': 'آچه\u200cای', 'bn': 'অ্যাচাইনিজ', 'br': 'achineg', 'brx': 'एचेनिज', 'bs': 'ačinski', 'bs-Cyrl': 'ачинески', 'bs-Latn': 'ačinski', 'ca': 'atjeh', 'ca-ES-valencia': 'atjeh', 'ccp': '𑄃𑄳𑄃𑄌𑄳𑄆𑄚𑄨𑄎𑄴', 'ce': 'ачехийн', 'chr': 'ᎠᏥᏂᏏ', 'ckb': 'ئاچەیی', 'cs': 'acehština', 'cy': 'Acehneg', 'da': 'achinesisk', 'de': 'Aceh', 'dsb': 'achinezišćina', 'el': 'Ατσινιζικά', 'el-polyton': 'Ατσινιζικά', 'en': 'Acehnese', 'es': 'achenés', 'et': 'atšehi', 'eu': 'acehnera', 'fa': 'آچئی', 'ff-Adlm': '𞤀𞥄𞤧𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'atšeh', 'fil': 'Achinese', 'fo': 'achinese', 'fr': 'aceh', 'fy': 'Atjeesk', 'ga': 'Aicinéis', 'gd': 'Basa Acèh', 'gl': 'achinés', 'gsw': 'Aceh', 'gu': 'અચીની', 'ha': 'Achinese', 'he': 'אכינזית', 'hi': 'अचाइनीस', 'hi-Latn': 'Acehnese', 'hr': 'ačinski', 'hsb': 'achinezišćina', 'ht': 'aceh', 'hu': 'achinéz', 'hy': 'աչեհերեն', 'ia': 'acehnese', 'id': 'Aceh', 'ig': 'Achinisi', 'is': 'akkíska', 'it': 'accinese', 'ja': 'アチェ語', 'jv': 'Achinese', 'ka': 'აჩეხური', 'kgp': 'agsẽm', 'kk': 'ачех тілі', 'kk-Cyrl': 'ачех тілі', 'km': 'អាកហ៊ីនឺស', 'kn': 'ಅಛಿನೀಸ್', 'ko': '아체어', 'kok': 'अचायनीज', 'kok-Deva': 'अचायनीज', 'ks': 'اَچَےنیٖز', 'ks-Arab': 'اَچَےنیٖز', 'ksh': 'Aschenehsesch', 'ku': 'açehî', 'ky': 'ачехче', 'lb': 'Aceh-Sprooch', 'lo': 'ແອັກຊີເນັສ', 'lt': 'ačinezų', 'lv': 'ačinu', 'mai': 'ऐचिनीज', 'mi': 'Akanīhi', 'mk': 'ачешки', 'ml': 'അചിനീസ്', 'mn': 'ачин', 'mr': 'अचीनी', 'ms': 'Aceh', 'mt': 'Aċiniż', 'my': 'အာချေး', 'nb': 'achinesisk', 'ne': 'अचाइनिज', 'nl': 'Atjehs', 'nn': 'achinesisk', 'no': 'achinesisk', 'or': 'ଆଚାଇନୀଜ୍', 'pa': 'ਅਚੀਨੀ', 'pa-Guru': 'ਅਚੀਨੀ', 'pcm': 'Achínẹ́sẹ', 'pl': 'aceh', 'ps': 'اچيني', 'pt': 'achém', 'qu': 'Achinese', 'rm': 'aceh', 'ro': 'aceh', 'ru': 'ачехский', 'sa': 'अचिनीस्', 'sat': 'ᱟᱠᱷᱤᱱᱤᱡᱽ', 'sat-Olck': 'ᱟᱠᱷᱤᱱᱤᱡᱽ', 'sc': 'acehnesu', 'sd': 'اچائينيز', 'sd-Arab': 'اچائينيز', 'se': 'acehgiella', 'si': 'අචයිනිස්', 'sk': 'acehčina', 'sl': 'ačejščina', 'smn': 'atšehkielâ', 'so': 'Shiinays', 'sq': 'akinezisht', 'sr': 'ацешки', 'sr-Cyrl': 'ацешки', 'sr-Latn': 'aceški', 'sv': 'acehnesiska', 'sw': 'Kiacheni', 'ta': 'ஆச்சினீஸ்', 'te': 'ఆఖినీస్', 'th': 'อาเจะห์', 'ti': 'ኣቸኒዝኛ', 'tk': 'açeh dili', 'to': 'lea fakaʻatisē', 'tr': 'Açece', 'ug': 'ئاتجېچە', 'uk': 'ачехська', 'ur': 'اچائینیز', 'uz': 'achin', 'uz-Cyrl': 'ачин', 'uz-Latn': 'achin', 'vec': 'acineze', 'vi': 'Tiếng Achinese', 'xnr': 'अचाइनीस', 'yo': 'Èdè Akinisi', 'yrl': 'axemi', 'yue': '亞齊文', 'yue-Hans': '亚齐文', 'yue-Hant': '亞齊文', 'zh': '亚齐语', 'zh-Hans': '亚齐语', 'zh-Hant': '亞齊文', 'zu': 'isi-Achinese'},
    'acf': {'en': 'Saint Lucian Creole French'},
    'ach': {'af': 'Akoli', 'am': 'አኮሊኛ', 'ar': 'الأكولية', 'ast': 'acoli', 'az': 'akoli', 'az-Latn': 'akoli', 'bg': 'аколи', 'bn': 'আকোলি', 'br': 'acoli', 'brx': 'अकोली', 'bs': 'akoli', 'bs-Cyrl': 'аколи', 'bs-Latn': 'akoli', 'ca': 'acoli', 'ca-ES-valencia': 'acoli', 'ccp': '𑄃𑄇𑄮𑄣𑄨', 'cs': 'akolština', 'cy': 'Acoli', 'da': 'acoli', 'de': 'Acholi', 'el': 'Ακολί', 'el-polyton': 'Ακολί', 'en': 'Acoli', 'es': 'acoli', 'et': 'atšoli', 'eu': 'acholiera', 'fa': 'آچولیایی', 'fi': 'atšoli', 'fil': 'Acoli', 'fr': 'acoli', 'fy': 'Akoli', 'gd': 'Acoli', 'gl': 'acholí', 'gsw': 'Acholi', 'gu': 'એકોલી', 'he': 'אקצ׳ולי', 'hi': 'अकोली', 'hi-Latn': 'Acoli', 'hr': 'ačoli', 'ht': 'acoli', 'hu': 'akoli', 'hy': 'աչոլի', 'id': 'Acoli', 'is': 'acoli', 'it': 'acioli', 'ja': 'アチョリ語', 'ka': 'აჩოლი', 'kgp': 'akori', 'kn': 'ಅಕೋಲಿ', 'ko': '아콜리어', 'ks': 'اَکولی', 'ks-Arab': 'اَکولی', 'lb': 'Acholi-Sprooch', 'lo': 'ອາໂຄລີ', 'lt': 'akolių', 'lv': 'ačolu', 'mk': 'аколи', 'ml': 'അകോലി', 'mr': 'अकोली', 'ms': 'Akoli', 'mt': 'Akoli', 'nb': 'acoli', 'ne': 'अकोली', 'nl': 'Akoli', 'nn': 'acoli', 'no': 'acoli', 'or': 'ଆକୋଲି', 'pa': 'ਅਕੋਲੀ', 'pa-Guru': 'ਅਕੋਲੀ', 'pl': 'aczoli', 'pt': 'acoli', 'rm': 'acoli', 'ro': 'acoli', 'ru': 'ачоли', 'sa': 'अचोलि', 'sk': 'ačoli', 'sl': 'ačolijščina', 'sr': 'аколи', 'sr-Cyrl': 'аколи', 'sr-Latn': 'akoli', 'sv': 'acholi', 'sw': 'Kiakoli', 'ta': 'அகோலி', 'te': 'అకోలి', 'th': 'อาโคลิ', 'to': 'lea fakaʻakoli', 'tr': 'Acoli', 'ug': 'ئاچولىچە', 'uk': 'ачолі', 'ur': 'اکولی', 'vi': 'Tiếng Acoli', 'yrl': 'akuri', 'yue': '阿僑利文', 'yue-Hans': '阿侨利文', 'yue-Hant': '阿僑利文', 'zh': '阿乔利语', 'zh-Hans': '阿乔利语', 'zh-Hant': '阿僑利文', 'zu': 'isi-Acoli'},
    'aci': {'en': 'Aka-Cari'},
    'ack': {'en': 'Aka-Kora'},
    'acl': {'en': 'Akar-Bale'},
    'acm': {'en': 'Mesopotamian Arabic'},
    'acn': {'en': 'Achang'},
    'acp': {'en': 'Eastern Acipa'},
    'acq': {'en': "Ta'izzi-Adeni Arabic"},
    'acr': {'en': 'Achi'},
    'acs': {'en': 'Acroá'},
    'act': {'en': 'Achterhoeks'},
    'acu': {'en': 'Achuar-Shiwiar'},
    'acv': {'en': 'Achumawi'},
    'acw': {'en': 'Hijazi Arabic'},
    'acx': {'en': 'Omani Arabic'},
    'acy': {'en': 'Cypriot Arabic'},
    'acz': {'en': 'Acheron'},
    'ada': {'af': 'Adangme', 'am': 'አዳንግሜ', 'ar': 'الأدانجمية', 'as': 'আদাংমে', 'ast': 'adangme', 'az': 'adanqme', 'az-Cyrl': 'адангме', 'az-Latn': 'adanqme', 'be': 'адангмэ', 'be-tarask': 'адангмэ', 'bg': 'адангме', 'bgn': 'ادانگمی', 'bn': 'অদাগ্মে', 'br': 'adangme', 'brx': 'एडांगमे', 'bs': 'adangmejski', 'bs-Cyrl': 'адангмејски', 'bs-Latn': 'adangmejski', 'ca': 'adangme', 'ca-ES-valencia': 'adangme', 'ccp': '𑄃𑄧𑄘𑄟𑄳𑄉𑄬', 'ce': 'адангме', 'chr': 'ᎠᏓᎾᎦᎺ', 'ckb': 'دانگمێ', 'cs': 'adangme', 'cy': 'Adangmeg', 'da': 'adangme', 'de': 'Adangme', 'dsb': 'adangmešćina', 'el': 'Αντάνγκμε', 'el-polyton': 'Αντάνγκμε', 'en': 'Adangme', 'es': 'adangme', 'et': 'adangme', 'eu': 'adangmera', 'fa': 'آدانگمه\u200cای', 'ff-Adlm': '𞤀𞤣𞤢𞤲𞤺𞤥𞤫𞥅𞤪𞤫', 'fi': 'adangme', 'fil': 'Adangme', 'fo': 'adangme', 'fr': 'adangme', 'fy': 'Adangme', 'ga': 'Daingmis', 'gd': 'Adangme', 'gl': 'adangme', 'gsw': 'Adangme', 'gu': 'અદાંગ્મી', 'ha': 'Adangme', 'he': 'אדנמה', 'hi': 'अदान्गमे', 'hi-Latn': 'Adangme', 'hr': 'adangme', 'hsb': 'adangmešćina', 'ht': 'adangme', 'hu': 'adangme', 'hy': 'ադանգմերեն', 'ia': 'adangme', 'id': 'Adangme', 'ig': 'Adangme', 'is': 'adangme', 'it': 'adangme', 'ja': 'アダングメ語', 'jv': 'Adangme', 'ka': 'ადანგმე', 'kgp': 'anágme', 'kk': 'адангме тілі', 'kk-Cyrl': 'адангме тілі', 'km': 'អាដេងមី', 'kn': 'ಅಡಂಗ್ಮೆ', 'ko': '아당메어', 'kok': 'अडांग्मे', 'kok-Deva': 'अडांग्मे', 'ks': 'اَدَنگمیے', 'ks-Arab': 'اَدَنگمیے', 'ksh': 'Adangmesch', 'ku': 'adangmeyî', 'ky': 'адаңмече', 'lb': 'Adangme', 'lo': 'ອາແດງມີ', 'lt': 'adangmų', 'lv': 'adangmu', 'mai': 'अदेंग', 'mi': 'Atāngami', 'mk': 'адангме', 'ml': 'അഡാങ്\u200cമി', 'mn': 'адангмэ', 'mr': 'अडांग्मे', 'ms': 'Adangme', 'mt': 'Adangme', 'my': 'ဒန်မဲ', 'nb': 'adangme', 'ne': 'अदाङमे', 'nl': 'Adangme', 'nn': 'adangme', 'no': 'adangme', 'or': 'ଆଦାଙ୍ଗେମ୍', 'pa': 'ਅਡਾਂਗਮੇ', 'pa-Guru': 'ਅਡਾਂਗਮੇ', 'pcm': 'Adángme', 'pl': 'adangme', 'ps': 'ادانگمي', 'pt': 'adangme', 'qu': 'Adangme', 'rm': 'andangme', 'ro': 'adangme', 'ru': 'адангме', 'sat': 'ᱟᱹᱰᱟᱱᱜᱽᱢᱮ', 'sat-Olck': 'ᱟᱹᱰᱟᱱᱜᱽᱢᱮ', 'sc': 'adangme', 'sd': 'ادنگمي', 'sd-Arab': 'ادنگمي', 'si': 'අඩන්ග්මෙ', 'sk': 'adangme', 'sl': 'adangmejščina', 'smn': 'adangme', 'so': 'Adangme', 'sq': 'andangmeisht', 'sr': 'адангме', 'sr-Cyrl': 'адангме', 'sr-Latn': 'adangme', 'sv': 'adangme', 'sw': 'Kiadangme', 'ta': 'அதாங்மே', 'te': 'అడాంగ్మే', 'th': 'อาแดงมี', 'ti': 'ኣዳንግሜ', 'tk': 'adangme dili', 'to': 'lea fakaʻatangimē', 'tr': 'Adangme', 'ug': 'ئاداڭمېچە', 'uk': 'адангме', 'ur': 'ادانگمے', 'uz': 'adangme', 'uz-Cyrl': 'адангмэ', 'uz-Latn': 'adangme', 'vec': 'adangme', 'vi': 'Tiếng Adangme', 'xnr': 'अदान्गमे', 'yo': 'Èdè Adame', 'yrl': 'adãgime', 'yue': '阿當莫文', 'yue-Hans': '阿当莫文', 'yue-Hant': '阿當莫文', 'zh': '阿当梅语', 'zh-Hans': '阿当梅语', 'zh-Hant': '阿當莫文', 'zu': 'isi-Adangme'},
    'adb': {'en': 'Atauran'},
    'add': {'en': 'Lidzonka'},
    'ade': {'en': 'Adele'},
    'adf': {'en': 'Dhofari Arabic'},
    'adg': {'en': 'Andegerebinha'},
    'adh': {'en': 'Adhola'},
    'adi': {'en': 'Adi'},
    'adj': {'en': 'Adioukrou'},
    'adl': {'en': 'Galo'},
    'adn': {'en': 'Adang'},
    'ado': {'en': 'Abu'},
    'adp': {'en': 'Adap'},
    'adq': {'en': 'Adangbe'},
    'adr': {'en': 'Adonara'},
    'ads': {'en': 'Adamorobe Sign Language'},
    'adt': {'en': 'Adnyamathanha'},
    'adu': {'en': 'Aduge'},
    'adw': {'en': 'Amundava'},
    'adx': {'en': 'Amdo Tibetan'},
    'ady': {'af': 'Adyghe', 'am': 'አድይግሄ', 'ar': 'الأديغة', 'as': 'আদিগে', 'ast': 'adyghe', 'az': 'adıgey', 'az-Cyrl': 'адуҝе', 'az-Latn': 'adıgey', 'be': 'адыгейская', 'be-tarask': 'адыгейская', 'bg': 'адигейски', 'bgn': 'ادیغی', 'bn': 'আদেগে', 'br': 'adygeieg', 'brx': 'अदिघे', 'bs': 'adigejski', 'bs-Cyrl': 'адигејски', 'bs-Latn': 'adigejski', 'ca': 'adigué', 'ca-ES-valencia': 'adigué', 'ccp': '𑄃𑄘𑄬𑄉𑄬', 'ce': 'адигейн', 'chr': 'ᎠᏗᎨ', 'ckb': 'ئادیگی', 'cs': 'adygejština', 'cy': 'Circaseg Gorllewinol', 'da': 'adyghe', 'de': 'Adygeisch', 'dsb': 'adyghešćina', 'el': 'Αντιγκέα', 'el-polyton': 'Αντιγκέα', 'en': 'Adyghe', 'es': 'adigué', 'et': 'adõgee', 'eu': 'adigera', 'fa': 'آدیجیایی', 'ff-Adlm': '𞤀𞤣𞤭𞤿𞤭𞥅𞤪𞤫', 'fi': 'adyge', 'fil': 'Adyghe', 'fo': 'adyghe', 'fr': 'adyguéen', 'fy': 'Adyghe', 'ga': 'Adaigéis', 'gd': 'Adyghe', 'gl': 'adigueo', 'gsw': 'Adygai', 'gu': 'અદિઘે', 'ha': 'Adyghe', 'he': 'אדיגית', 'hi': 'अदिघे', 'hi-Latn': 'Adyghe', 'hr': 'adigejski', 'hsb': 'adyghešćina', 'ht': 'adyguéen', 'hu': 'adyghe', 'hy': 'ադիղերեն', 'ia': 'adygeano', 'id': 'Adygei', 'ig': 'Adigi', 'is': 'adýge', 'it': 'adyghe', 'ja': 'アディゲ語', 'jv': 'Adyghe', 'ka': 'ადიღეური', 'kgp': 'anhige', 'kk': 'адыгей тілі', 'kk-Cyrl': 'адыгей тілі', 'km': 'អាឌីហ្គី', 'kn': 'ಅಡೈಘೆ', 'ko': '아디게어', 'kok': 'अडिघे', 'kok-Deva': 'अडिघे', 'ks': 'اَدَیٖگیے', 'ks-Arab': 'اَدَیٖگیے', 'ksh': 'Adygehjsch', 'ku': 'adîgeyî', 'ky': 'адыгейче', 'lb': 'Adygéiesch', 'lkt': 'Adyghe Iyápi', 'lo': 'ເອດີຮິ', 'lt': 'adygėjų', 'lv': 'adigu', 'mai': 'अदिगे', 'mi': 'Āteke', 'mk': 'адигејски', 'ml': 'അഡൈഗേ', 'mn': 'адигэ', 'mr': 'अडिघे', 'ms': 'Adyghe', 'mt': 'Adyghe', 'my': 'အဒိုင်ဂီ', 'nb': 'adygeisk', 'ne': 'अदिघे', 'nl': 'Adygees', 'nn': 'adygeisk', 'no': 'adygeisk', 'or': 'ଅଦ୍ୟଘେ', 'os': 'адыгейаг', 'pa': 'ਅਡਿਗੇ', 'pa-Guru': 'ਅਡਿਗੇ', 'pcm': 'Adyghẹ́', 'pl': 'adygejski', 'ps': 'اديغي', 'pt': 'adigue', 'qu': 'Adyghe', 'rm': 'adygai', 'ro': 'adyghe', 'ru': 'адыгейский', 'sat': 'ᱟᱹᱰᱭᱜᱷᱮ', 'sat-Olck': 'ᱟᱹᱰᱭᱜᱷᱮ', 'sc': 'adighè', 'sd': 'اديگهي', 'sd-Arab': 'اديگهي', 'si': 'අඩිඝෙ', 'sk': 'adygejčina', 'sl': 'adigejščina', 'smn': 'adyge', 'so': 'U dhashay Ady', 'sq': 'adigisht', 'sr': 'адигејски', 'sr-Cyrl': 'адигејски', 'sr-Latn': 'adigejski', 'sv': 'adygeiska', 'sw': 'Kiadyghe', 'ta': 'அதகே', 'te': 'అడిగాబ్జే', 'th': 'อะดืยเก', 'ti': 'ኣዲጊ', 'tk': 'adygeý dili', 'to': 'lea fakaʻatikē', 'tr': 'Adigece', 'ug': 'ئادىگېيچە', 'uk': 'адигейська', 'ur': 'ادیگھے', 'uz': 'adigey', 'uz-Cyrl': 'адигей', 'uz-Latn': 'adigey', 'vec': 'adighe', 'vi': 'Tiếng Adyghe', 'xnr': 'अदिघे', 'yo': 'Èdè Adiji', 'yrl': 'adige', 'yue': '阿迪各文', 'yue-Hans': '阿迪各文', 'yue-Hant': '阿迪各文', 'zh': '阿迪格语', 'zh-Hans': '阿迪格语', 'zh-Hant': '阿迪各文', 'zu': 'isi-Adyghe'},
    'adz': {'en': 'Adzera'},
    'ae': {'am': 'አቬስታን', 'ar': 'الأفستية', 'ast': 'avestanín', 'az': 'avestan', 'az-Latn': 'avestan', 'bg': 'авестски', 'bn': 'আবেস্তীয়', 'br': 'avesteg', 'brx': 'अवस्तन्', 'bs': 'avestanski', 'bs-Cyrl': 'авестански', 'bs-Latn': 'avestanski', 'ca': 'avèstic', 'ca-ES-valencia': 'avèstic', 'ccp': '𑄃𑄝𑄬𑄌𑄴𑄖𑄩𑄠𑄧', 'cs': 'avestánština', 'cy': 'Afestaneg', 'da': 'avestan', 'de': 'Avestisch', 'el': 'Αβεστάν', 'el-polyton': 'Αβεστάν', 'en': 'Avestan', 'es': 'avéstico', 'et': 'avesta', 'fa': 'اوستایی', 'fi': 'avesta', 'fr': 'avestique', 'fur': 'avestan', 'fy': 'Avestysk', 'ga': 'Aivéistis', 'gd': 'Avestanais', 'gsw': 'Avestisch', 'gu': 'અવેસ્તન', 'he': 'אבסטן', 'hi': 'अवस्ताई', 'hi-Latn': 'Avestan', 'hr': 'avestički', 'ht': 'avestique', 'hu': 'avesztán', 'id': 'Avesta', 'is': 'avestíska', 'it': 'avestan', 'ja': 'アヴェスタ語', 'ka': 'ავესტური', 'kgp': 'avétiko', 'km': 'អាវេស្ថាន', 'kn': 'ಅವೆಸ್ಟನ್', 'ko': '아베스타어', 'ks': 'اَویستَن', 'ks-Arab': 'اَویستَن', 'ksh': 'Avästahnesch', 'lb': 'Avestesch', 'lkt': 'Avestan Iyápi', 'lo': 'ອາເວັສແຕນ', 'lt': 'avestų', 'lv': 'avesta', 'mk': 'авестански', 'ml': 'അവസ്റ്റാൻ', 'mr': 'अवेस्तन', 'ms': 'Avestan', 'mt': 'Avestan', 'nb': 'avestisk', 'ne': 'अवेस्तान', 'nl': 'Avestisch', 'nn': 'avestisk', 'no': 'avestisk', 'or': 'ଅବେସ୍ତନ', 'os': 'авестӕ', 'pl': 'awestyjski', 'pt': 'avéstico', 'rm': 'avestic', 'ro': 'avestană', 'ru': 'авестийский', 'sk': 'avestčina', 'sl': 'avestijščina', 'sr': 'авестански', 'sr-Cyrl': 'авестански', 'sr-Latn': 'avestanski', 'sv': 'avestiska', 'ta': 'அவெஸ்தான்', 'te': 'అవేస్టాన్', 'th': 'อเวสตะ', 'to': 'lea fakaʻavesitani', 'tr': 'Avestçe', 'ug': 'ئاۋېستاچە', 'uk': 'авестійська', 'vi': 'Tiếng Avestan', 'yrl': 'awesitiku', 'yue': '阿緯斯陀文', 'yue-Hans': '阿纬斯陀文', 'yue-Hant': '阿緯斯陀文', 'zh': '阿维斯塔语', 'zh-Hans': '阿维斯塔语', 'zh-Hant': '阿維斯塔文'},
    'aea': {'en': 'Areba'},
    'aeb': {'ast': 'árabe de Túnez', 'br': 'arabeg Tunizia', 'cs': 'arabština (tuniská)', 'cy': 'Arabeg Tunisia', 'de': 'Tunesisches Arabisch', 'en': 'Tunisian Arabic', 'et': 'Tuneesia araabia', 'fa': 'عربی تونسی', 'fi': 'tunisianarabia', 'fr': 'arabe tunisien', 'gd': 'Arabais Thuiniseach', 'hi-Latn': 'Tunisian Arabic', 'ht': 'arabe tunisien', 'hy': 'թունիսական արաբերեն', 'id': 'Arab Tunisia', 'it': 'arabo tunisino', 'ja': 'チュニジア・アラビア語', 'ko': '튀니지 아랍어', 'ksh': 'Tonehsesch Arahbesch', 'lb': 'Tunesescht Arabesch', 'lt': 'Tuniso arabų', 'mk': 'туниски арапски', 'ms': 'Arab Tunisia', 'nb': 'tunisisk-arabisk', 'nl': 'Tunesisch Arabisch', 'nn': 'tunisisk-arabisk', 'no': 'tunisisk-arabisk', 'pl': 'tunezyjski arabski', 'si': 'ටියුනිසියනු අරාබි', 'sv': 'tunisisk arabiska', 'ta': 'துனிசிய அரபு', 'te': 'టునీషియా అరబిక్', 'th': 'อาหรับตูนิเซีย', 'to': 'lea fakaʻalepea-tunīsia', 'tr': 'Tunus Arapçası', 'yue': '突尼斯阿拉伯文', 'yue-Hans': '突尼斯阿拉伯文', 'yue-Hant': '突尼斯阿拉伯文', 'zh-Hant': '突尼斯阿拉伯文'},
    'aec': {'en': 'Saidi Arabic'},
    'aed': {'en': 'Argentine Sign Language'},
    'aee': {'en': 'Northeast Pashai'},
    'aek': {'en': 'Haeke'},
    'ael': {'en': 'Ambele'},
    'aem': {'en': 'Arem'},
    'aen': {'en': 'Armenian Sign Language'},
    'aeq': {'en': 'Aer'},
    'aer': {'en': 'Eastern Arrernte'},
    'aes': {'en': 'Alsea'},
    'aeu': {'en': 'Akeu'},
    'aew': {'en': 'Ambakich'},
    'aey': {'en': 'Amele'},
    'aez': {'en': 'Aeka'},
    'af': {'af': 'Afrikaans', 'ak': 'Afrikaans', 'am': 'አፍሪካንኛ', 'ar': 'الأفريقانية', 'as': 'আফ্ৰিকানছ্', 'ast': 'afrikaans', 'az': 'afrikaans', 'az-Arab': 'آفریکانس', 'az-Cyrl': 'африкаанс', 'az-Latn': 'afrikaans', 'bal-Latn': 'Aprikái', 'be': 'афрыкаанс', 'be-tarask': 'афрыкаанс', 'bg': 'африканс', 'bgn': 'افریکانس', 'bn': 'আফ্রিকান', 'br': 'afrikaans', 'brx': 'आफ्रिकान्स', 'bs': 'afrikans', 'bs-Cyrl': 'африканс', 'bs-Latn': 'afrikans', 'ca': 'afrikaans', 'ca-ES-valencia': 'afrikaans', 'ccp': '𑄃𑄜𑄳𑄢𑄨𑄇𑄚𑄴', 'ce': 'африкаанс', 'chr': 'ᎠᎬᎿᎨᏍᏛ', 'ckb': 'ئەفریکانس', 'cs': 'afrikánština', 'cy': 'Affricaneg', 'da': 'afrikaans', 'de': 'Afrikaans', 'dsb': 'afrikans', 'dz': 'ཨཕ་རི་ཀཱནས་ཁ', 'ee': 'afrikaangbe', 'el': 'Αφρικάανς', 'el-polyton': 'Αφρικάανς', 'en': 'Afrikaans', 'eo': 'afrikansa', 'es': 'afrikáans', 'et': 'afrikaani', 'eu': 'afrikaansa', 'fa': 'آفریکانس', 'ff-Adlm': '𞤀𞤬𞤪𞤭𞤳𞤢𞤲𞤪𞤫', 'fi': 'afrikaans', 'fil': 'Afrikaans', 'fo': 'afrikaans', 'fr': 'afrikaans', 'fur': 'afrikaans', 'fy': 'Afrikaansk', 'ga': 'Afracáinis', 'gd': 'Afraganais', 'gl': 'afrikaans', 'gsw': 'Afrikaans', 'gu': 'આફ્રિકન્સ', 'ha': 'Afirkanci', 'he': 'אפריקאנס', 'hi': 'अफ़्रीकी', 'hi-Latn': 'Afreeki', 'hr': 'afrikaans', 'hsb': 'afrikaanšćina', 'ht': 'afrikaans', 'hu': 'afrikaans', 'hy': 'աֆրիկաանս', 'ia': 'afrikaans', 'id': 'Afrikaans', 'ig': 'Afrikaans', 'is': 'afríkanska', 'it': 'afrikaans', 'ja': 'アフリカーンス語', 'jv': 'Afrika', 'ka': 'აფრიკაანსი', 'kea': 'afrikaner', 'kgp': 'afrikỹnẽ', 'kk': 'африкаанс тілі', 'kk-Cyrl': 'африкаанс тілі', 'km': 'អាហ្វ្រិកាន', 'kn': 'ಆಫ್ರಿಕಾನ್ಸ್', 'ko': '아프리칸스어', 'kok': 'अफ्रिकान्स', 'kok-Deva': 'अफ्रिकान्स', 'ks': 'اَفریٖکانز', 'ks-Arab': 'اَفریٖکانز', 'ksh': 'Afrikaans', 'ku': 'afrîkansî', 'kxv': 'aprikaans', 'kxv-Deva': 'आप्रिकान्स', 'kxv-Latn': 'aprikaans', 'kxv-Orya': 'ଆପ୍ରୀକାନ୍\u200dସ', 'kxv-Telu': 'ఆప్రికాన్స్', 'ky': 'африкаанча', 'lb': 'Afrikaans', 'lkt': 'Afrikaans Iyápi', 'lld': 'afrikaans', 'lo': 'ອາຟຣິການ', 'lrc': 'آفریکانس', 'lt': 'afrikanų', 'lv': 'afrikandu', 'mai': 'अफ्रीकांस', 'mi': 'Awherikāna', 'mk': 'африканс', 'ml': 'ആഫ്രിക്കാൻസ്', 'mn': 'африкаанс', 'mr': 'अफ्रिकान्स', 'ms': 'Afrikaans', 'mt': 'Afrikans', 'my': 'အာဖရိကန်', 'mzn': 'آفریکانس', 'nb': 'afrikaans', 'ne': 'अफ्रिकान्स', 'nl': 'Afrikaans', 'nn': 'afrikaans', 'no': 'afrikaans', 'nqo': 'ߊߝߙߌߞߊ߲߯ߛߑߞߊ߲', 'om': 'Afrikoota', 'or': 'ଆଫ୍ରିକୀୟ', 'os': 'африкаанс', 'pa': 'ਅਫ਼ਰੀਕੀ', 'pa-Guru': 'ਅਫ਼ਰੀਕੀ', 'pcm': 'Áfríkaans Lángwej', 'pl': 'afrikaans', 'ps': 'افریکانسي', 'pt': 'africâner', 'qu': 'Afrikaans Simi', 'rm': 'afrikaans', 'ro': 'afrikaans', 'ru': 'африкаанс', 'rw': 'Ikinyafurikaneri', 'sah': 'Аппырыкааныстыы', 'sat': 'ᱟᱯᱷᱨᱤᱠᱟᱱᱥ', 'sat-Olck': 'ᱟᱯᱷᱨᱤᱠᱟᱱᱥ', 'sc': 'afrikaans', 'sd': 'آفريڪي', 'sd-Arab': 'آفريڪي', 'se': 'afrikánsagiella', 'si': 'අෆ්\u200dරිකාන්ස්', 'sk': 'afrikánčina', 'sl': 'afrikanščina', 'smn': 'afrikaans', 'so': 'Afrikaanka', 'sq': 'afrikanisht', 'sr': 'африканс', 'sr-Cyrl': 'африканс', 'sr-Latn': 'afrikans', 'sv': 'afrikaans', 'sw': 'Kiafrikana', 'ta': 'ஆஃப்ரிகான்ஸ்', 'te': 'ఆఫ్రికాన్స్', 'tg': 'африкаанс', 'th': 'แอฟริกานส์', 'ti': 'ኣፍሪካንስ', 'tk': 'afrikaans dili', 'to': 'lea fakaʻafilikana', 'tr': 'Afrikaanca', 'tt': 'африкаанс', 'ug': 'ئافرىكانچە', 'uk': 'африкаанс', 'ur': 'افریقی', 'uz': 'afrikaans', 'uz-Cyrl': 'африкаанс', 'uz-Latn': 'afrikaans', 'vec': 'afregan', 'vi': 'Tiếng Hà Lan (Nam Phi)', 'wae': 'Afrikáns', 'wo': 'Afrikaans', 'xh': 'IsiBhulu', 'xnr': 'अफ़्रीकी', 'yi': 'אַפֿריקאַנס', 'yo': 'Èdè Afrikani', 'yrl': 'afiriķãner', 'yue': '南非荷蘭文', 'yue-Hans': '南非荷兰文', 'yue-Hant': '南非荷蘭文', 'zh': '南非荷兰语', 'zh-Hans': '南非荷兰语', 'zh-Hant': '南非荷蘭文', 'zu': 'i-Afrikaans'},
    'afa': {'en': 'Afro-Asiatic languages'},
    'afa-pro': {'en': 'Proto-Afro-Asiatic'},
    'afb': {'en': 'Gulf Arabic'},
    'afd': {'en': 'Andai'},
    'afe': {'en': 'Putukwam'},
    'afg': {'en': 'Afghan Sign Language'},
    'afh': {'am': 'አፍሪሂሊ', 'ar': 'الأفريهيلية', 'ast': 'afrihili', 'az': 'afrihili', 'az-Latn': 'afrihili', 'bg': 'африхили', 'bn': 'আফ্রিহিলি', 'br': 'afrihili', 'brx': 'अफ्रीहीली', 'bs': 'afrihili', 'bs-Cyrl': 'африхили', 'bs-Latn': 'afrihili', 'ca': 'afrihili', 'ca-ES-valencia': 'afrihili', 'ccp': '𑄃𑄜𑄳𑄢𑄨𑄦𑄨𑄣𑄨', 'cs': 'afrihili', 'cy': 'Affrihili', 'da': 'afrihili', 'de': 'Afrihili', 'el': 'Αφριχίλι', 'el-polyton': 'Αφριχίλι', 'en': 'Afrihili', 'es': 'afrihili', 'et': 'afrihili', 'fa': 'آفریهیلی', 'fi': 'afrihili', 'fr': 'afrihili', 'fy': 'Afrihili', 'gd': 'Afrihili', 'gsw': 'Afrihili', 'gu': 'અફ્રિહિલી', 'he': 'אפריהילי', 'hi': 'अफ्रिहिली', 'hi-Latn': 'Afrihili', 'hr': 'afrihili', 'ht': 'afrihili', 'hu': 'afrihili', 'id': 'Afrihili', 'is': 'afríhílí', 'it': 'afrihili', 'ja': 'アフリヒリ語', 'kgp': 'afrihiri', 'kn': 'ಆಫ್ರಿಹಿಲಿ', 'ko': '아프리힐리어', 'ks': 'اَفرِہِلی', 'ks-Arab': 'اَفرِہِلی', 'ksh': 'Afrehihlesch', 'lb': 'Afrihili', 'lo': 'ແອຟີຮີລີ', 'lt': 'afrihili', 'lv': 'afrihili', 'mk': 'африхили', 'ml': 'ആഫ്രിഹിലി', 'mr': 'अफ्रिहिली', 'mt': 'Afriħili', 'nb': 'afrihili', 'ne': 'अफ्रिहिली', 'nl': 'Afrihili', 'nn': 'afrihili', 'no': 'afrihili', 'or': 'ଆଫ୍ରିହିଲି', 'pl': 'afrihili', 'pt': 'afrihili', 'rm': 'afrihili', 'ro': 'afrihili', 'ru': 'африхили', 'sk': 'afrihili', 'sl': 'afrihili', 'sr': 'африхили', 'sr-Cyrl': 'африхили', 'sr-Latn': 'afrihili', 'sv': 'afrihili', 'ta': 'அஃப்ரிஹிலி', 'te': 'అఫ్రిహిలి', 'th': 'แอฟริฮีลี', 'to': 'lea fakaʻafilihili', 'tr': 'Afrihili', 'ug': 'ئافرىخىلىچە', 'uk': 'африхілі', 'vi': 'Tiếng Afrihili', 'yrl': 'afirihiri', 'yue': '阿弗里希利文', 'yue-Hans': '阿弗里希利文', 'yue-Hant': '阿弗里希利文', 'zh': '阿弗里希利语', 'zh-Hans': '阿弗里希利语', 'zh-Hant': '阿弗里希利文'},
    'afi': {'en': 'Akrukay'},
    'afk': {'en': 'Nanubae'},
    'afn': {'en': 'Defaka'},
    'afo': {'en': 'Eloyi'},
    'afp': {'en': 'Tapei'},
    'afs': {'en': 'Afro-Seminole Creole'},
    'aft': {'en': 'Afitti'},
    'afu': {'en': 'Awutu'},
    'afz': {'en': 'Obokuitai'},
    'aga': {'en': 'Aguano'},
    'agb': {'en': 'Legbo'},
    'agc': {'en': 'Agatu'},
    'agd': {'en': 'Agarabi'},
    'age': {'en': 'Angal'},
    'agf': {'en': 'Arguni'},
    'agg': {'en': 'Angor'},
    'agh': {'en': 'Ngelima'},
    'agi': {'en': 'Agariya'},
    'agj': {'en': 'Argobba'},
    'agk': {'en': 'Isarog Agta'},
    'agl': {'en': 'Fembe'},
    'agm': {'en': 'Angaataha'},
    'agn': {'en': 'Agutaynen'},
    'ago': {'en': 'Tainae'},
    'agp': {'en': 'Paranan'},
    'agq': {'af': 'Aghem', 'agq': 'Aghem', 'am': 'አገም', 'ar': 'الأغم', 'as': 'আঘেম', 'ast': 'aghem', 'az': 'aqhem', 'az-Arab': 'آقمجه', 'az-Cyrl': 'агһем', 'az-Latn': 'aqhem', 'bal-Latn': 'Agem', 'be': 'агем', 'be-tarask': 'агем', 'bg': 'агем', 'bgn': 'اغیمی', 'bn': 'এঘেম', 'br': 'aghem', 'brx': 'आघेम', 'bs': 'aghem', 'bs-Cyrl': 'ахемски', 'bs-Latn': 'aghem', 'ca': 'aghem', 'ca-ES-valencia': 'aghem', 'ccp': '𑄃𑄬𑄊𑄟𑄴', 'ce': 'агхӀем', 'chr': 'ᎠᎨᎹ', 'ckb': 'ئاگێم', 'cs': 'aghem', 'cy': 'Aghemeg', 'da': 'aghem', 'de': 'Aghem', 'dsb': 'aghem', 'ee': 'aghemgbe', 'el': 'Αγκέμ', 'el-polyton': 'Αγκέμ', 'en': 'Aghem', 'es': 'aghem', 'et': 'aghemi', 'eu': 'aghemera', 'fa': 'آگیم', 'ff-Adlm': '𞤀𞤺𞤸𞤫𞤥𞤪𞤫', 'fi': 'aghem', 'fil': 'Aghem', 'fo': 'aghem', 'fr': 'aghem', 'fy': 'Aghem', 'ga': 'Aigeimis', 'gd': 'Aghem', 'gl': 'aghem', 'gu': 'અઘેમ', 'ha': 'Aghem', 'he': 'אע׳ם', 'hi': 'अग्हेम', 'hi-Latn': 'Aghem', 'hr': 'aghem', 'hsb': 'aghemšćina', 'ht': 'aghem', 'hu': 'agem', 'hy': 'աղեմ', 'ia': 'aghem', 'id': 'Aghem', 'ig': 'Aghem', 'is': 'aghem', 'it': 'aghem', 'ja': 'アゲム語', 'jv': 'Aghem', 'ka': 'აღემი', 'kea': 'aghem', 'kgp': 'aghẽm', 'kk': 'агем тілі', 'kk-Cyrl': 'агем тілі', 'km': 'អាហ្គីម', 'kn': 'ಅಘೆಮ್', 'ko': '아그햄어', 'kok': 'अघेम', 'kok-Deva': 'अघेम', 'ksh': 'Aghehmesch', 'ku': 'aghemî', 'ky': 'агемче', 'lb': 'Aghem', 'lo': 'ອາເຮັມ', 'lrc': 'آقئم', 'lt': 'aghemų', 'lv': 'aghemu', 'mai': 'अगेम', 'mi': 'Ākeme', 'mk': 'агемски', 'ml': 'ആഘേം', 'mn': 'агем', 'mr': 'अघेम', 'ms': 'Aghem', 'mt': 'Aghem', 'my': 'အာဂ်ဟိန်း', 'mzn': 'آقم', 'nb': 'aghem', 'ne': 'आघेम', 'nl': 'Aghem', 'nn': 'aghem', 'no': 'aghem', 'nqo': 'ߊߜ߭ߍߡߑߞߊ߲', 'or': 'ଆଘେମ', 'pa': 'ਅਗੇਮ', 'pa-Guru': 'ਅਗੇਮ', 'pcm': 'Aghẹ́m Lángwej', 'pl': 'aghem', 'ps': 'اغیمي', 'pt': 'aghem', 'qu': 'Aghem Simi', 'rm': 'aghem', 'ro': 'aghem', 'ru': 'агем', 'sat': 'ᱟᱜᱷᱮᱢ', 'sat-Olck': 'ᱟᱜᱷᱮᱢ', 'sc': 'aghem', 'sd': 'اگهيم', 'sd-Arab': 'اگهيم', 'si': 'ඇගම්', 'sk': 'aghem', 'sl': 'aghemščina', 'smn': 'aghem', 'so': 'Ageem', 'sq': 'agemisht', 'sr': 'агем', 'sr-Cyrl': 'агем', 'sr-Latn': 'agem', 'sv': 'aghem', 'sw': 'Kiaghem', 'ta': 'அகெம்', 'te': 'అగేమ్', 'th': 'อักเฮม', 'ti': 'ኣገም', 'tk': 'ahem dili', 'to': 'lea fakaʻakihemi', 'tr': 'Aghem', 'ug': 'ئاگەمچە', 'uk': 'агем', 'ur': 'اغم', 'uz': 'agem', 'uz-Cyrl': 'агемча', 'uz-Latn': 'agem', 'vec': 'aghem', 'vi': 'Tiếng Aghem', 'xnr': 'अग्हेम', 'yo': 'Èdè Ágẹ̀ẹ̀mù', 'yrl': 'agẽ', 'yue': '亞罕文', 'yue-Hans': '亚罕文', 'yue-Hant': '亞罕文', 'zh': '亚罕语', 'zh-Hans': '亚罕语', 'zh-Hant': '亞罕文', 'zu': 'isi-Aghem'},
    'agr': {'en': 'Aguaruna'},
    'ags': {'en': 'Esimbi'},
    'agt': {'en': 'Central Cagayan Agta'},
    'agu': {'en': 'Aguacateco'},
    'agv': {'en': 'Remontado Dumagat'},
    'agw': {'en': 'Kahua'},
    'agx': {'en': 'Aghul'},
    'agy': {'en': 'Southern Alta'},
    'agz': {'en': 'Mt. Iriga Agta'},
    'aha': {'en': 'Ahanta'},
    'ahb': {'en': 'Axamb'},
    'ahg': {'en': 'Qimant'},
    'ahh': {'en': 'Aghu'},
    'ahi': {'en': 'Tiagbamrin Aizi'},
    'ahk': {'en': 'Akha'},
    'ahl': {'en': 'Igo'},
    'ahm': {'en': 'Mobumrin Aizi'},
    'ahn': {'en': 'Àhàn'},
    'aho': {'en': 'Ahom'},
    'ahp': {'en': 'Aproumu Aizi'},
    'ahr': {'en': 'Ahirani'},
    'ahs': {'en': 'Ashe'},
    'aht': {'en': 'Ahtena'},
    'aia': {'en': 'Arosi'},
    'aib': {'en': 'Ainu (China)'},
    'aic': {'en': 'Ainbai'},
    'aid': {'en': 'Alngith'},
    'aie': {'en': 'Amara'},
    'aif': {'en': 'Agi'},
    'aig': {'en': 'Antigua and Barbuda Creole English'},
    'aih': {'en': 'Ai-Cham'},
    'aii': {'en': 'Assyrian Neo-Aramaic'},
    'aij': {'en': 'Lishanid Noshan'},
    'aik': {'en': 'Ake'},
    'ail': {'en': 'Aimele'},
    'aim': {'en': 'Aimol'},
    'ain': {'af': 'Ainu', 'am': 'አይኑ', 'ar': 'الآينوية', 'as': 'আইনু', 'ast': 'ainu', 'az': 'aynu', 'az-Cyrl': 'ајну', 'az-Latn': 'aynu', 'be': 'айнская', 'be-tarask': 'айнская', 'bg': 'айну', 'bgn': 'آینو', 'bn': 'আইনু', 'br': 'ainoueg', 'brx': 'ऐनू', 'bs': 'ainu', 'bs-Cyrl': 'аину', 'bs-Latn': 'ainu', 'ca': 'ainu', 'ca-ES-valencia': 'ainu', 'ccp': '𑄃𑄳𑄆𑄚𑄪', 'ce': 'айнийн', 'chr': 'ᎠᏱᏄ', 'ckb': 'ئاینوو', 'cs': 'ainština', 'cy': 'Ainŵeg', 'da': 'ainu', 'de': 'Ainu', 'dsb': 'ainušćina', 'el': 'Αϊνού', 'el-polyton': 'Αϊνού', 'en': 'Ainu', 'es': 'ainu', 'et': 'ainu', 'eu': 'ainuera', 'fa': 'آینویی', 'ff-Adlm': '𞤀𞤴𞤲𞤵𞥅𞤪𞤫', 'fi': 'ainu', 'fil': 'Ainu', 'fo': 'ainu', 'fr': 'aïnou', 'fy': 'Ainu', 'ga': 'Aidhniúis', 'gd': 'Ainu', 'gl': 'ainu', 'gsw': 'Ainu', 'gu': 'ઐનુ', 'ha': 'Ainu', 'he': 'אינו', 'hi': 'ऐनू', 'hi-Latn': 'Ainu', 'hr': 'ainuski', 'hsb': 'ainušćina', 'ht': 'aïnou', 'hu': 'ainu', 'hy': 'այներեն', 'ia': 'ainu', 'id': 'Ainu', 'ig': 'Ainu', 'is': 'aínu (Japan)', 'it': 'ainu', 'ja': 'アイヌ語', 'jv': 'Ainu', 'ka': 'აინუური', 'kgp': 'ajinũ', 'kk': 'айну тілі', 'kk-Cyrl': 'айну тілі', 'km': 'អាយនូ', 'kn': 'ಐನು', 'ko': '아이누어', 'kok': 'आयनू', 'kok-Deva': 'आयनू', 'ks': 'اینوٗ', 'ks-Arab': 'اینوٗ', 'ksh': 'Aijnu', 'ku': 'aynuyî', 'ky': 'айнуча', 'lb': 'Ainu-Sprooch', 'lo': 'ໄອນູ', 'lt': 'ainų', 'lv': 'ainu', 'mai': 'आयू', 'mi': 'Ainu', 'mk': 'ајну', 'ml': 'ഐനു', 'mn': 'айну', 'mr': 'ऐनू', 'ms': 'Ainu', 'mt': 'Ajnu', 'my': 'အိန်နု', 'nb': 'ainu', 'ne': 'अइनु', 'nl': 'Aino', 'nn': 'ainu', 'no': 'ainu', 'or': 'ଆଇନୁ', 'pa': 'ਆਇਨੂ', 'pa-Guru': 'ਆਇਨੂ', 'pcm': 'Ainú', 'pl': 'ajnu', 'ps': 'اينويي', 'pt': 'ainu', 'qu': 'Ainu', 'rm': 'ainu', 'ro': 'ainu', 'ru': 'айнский', 'sat': 'ᱟᱭᱤᱱᱩ', 'sat-Olck': 'ᱟᱭᱤᱱᱩ', 'sc': 'àinu', 'sd': 'آئينو', 'sd-Arab': 'آئينو', 'si': 'අයිනු', 'sk': 'ainčina', 'sl': 'ainujščina', 'smn': 'ainukielâ', 'so': 'U dhashay Ain', 'sq': 'ajnuisht', 'sr': 'аину', 'sr-Cyrl': 'аину', 'sr-Latn': 'ainu', 'sv': 'ainu', 'sw': 'Kiainu', 'ta': 'ஐனு', 'te': 'ఐను', 'th': 'ไอนุ', 'ti': 'ኣይኑ', 'tk': 'aýn dili', 'to': 'lea fakaʻainu', 'tr': 'Aynuca', 'ug': 'ئاينۇچە', 'uk': 'айнська', 'ur': 'اینو', 'uz': 'aynu', 'uz-Cyrl': 'айну', 'uz-Latn': 'aynu', 'vec': 'ainu', 'vi': 'Tiếng Ainu', 'xnr': 'ऐनू', 'yo': 'Èdè Ainu', 'yrl': 'ainú', 'yue': '阿伊努文', 'yue-Hans': '阿伊努文', 'yue-Hant': '阿伊努文', 'zh': '阿伊努语', 'zh-Hans': '阿伊努语', 'zh-Hant': '阿伊努文', 'zu': 'isi-Ainu'},
    'aio': {'en': 'Aiton'},
    'aip': {'en': 'Burumakok'},
    'aiq': {'en': 'Aimaq'},
    'air': {'en': 'Airoran'},
    'ais': {'en': 'Nataoran Amis'},
    'ait': {'en': 'Arikem'},
    'aiw': {'en': 'Aari'},
    'aix': {'en': 'Aighon'},
    'aiy': {'en': 'Ali'},
    'aja': {'en': 'Aja (South Sudan)'},
    'ajg': {'en': 'Aja (Benin)'},
    'aji': {'en': 'Ajië'},
    'ajn': {'en': 'Andajin'},
    'ajp': {'en': 'South Levantine Arabic'},
    'ajt': {'en': 'Judeo-Tunisian Arabic'},
    'aju': {'en': 'Judeo-Moroccan Arabic'},
    'ajw': {'en': 'Ajawa'},
    'ajz': {'en': 'Amri Karbi'},
    'ak': {'af': 'Akan', 'agq': 'Akan', 'ak': 'Akan', 'am': 'አካንኛ', 'ar': 'الأكانية', 'as': 'আকান', 'asa': 'Kiakan', 'ast': 'akan', 'az': 'akan', 'az-Arab': 'آکانجا', 'az-Cyrl': 'акан', 'az-Latn': 'akan', 'bal-Latn': 'Akan', 'bas': 'Hɔp u akan', 'be': 'акан', 'be-tarask': 'акан', 'bem': 'Ichi Akan', 'bez': 'Hiakan', 'bg': 'акан', 'bgn': 'اکانی', 'bm': 'akankan', 'bn': 'আকান', 'br': 'akan', 'brx': 'आकान', 'bs': 'akan', 'bs-Cyrl': 'акан', 'bs-Latn': 'akan', 'ca': 'àkan', 'ca-ES-valencia': 'àkan', 'ccp': '𑄃𑄇𑄚𑄴', 'ce': 'акан', 'cgg': 'Orukani', 'chr': 'ᎠᎧᎾ', 'ckb': 'ئاکان', 'cs': 'akanština', 'cy': 'Acaneg', 'da': 'akan', 'dav': 'Kiakan', 'de': 'Akan', 'dje': 'Akan senni', 'dsb': 'akanšćina', 'dyo': 'akan', 'ebu': 'Kĩakan', 'ee': 'blugbe', 'el': 'Ακάν', 'el-polyton': 'Ακάν', 'en': 'Akan', 'es': 'akan', 'et': 'akani', 'eu': 'akanera', 'ewo': 'Ǹkɔ́bɔ akán', 'fa': 'آکان', 'ff': 'Akaan', 'ff-Adlm': '𞤀𞤳𞤢𞤲𞤪𞤫', 'ff-Latn': 'Akaan', 'fi': 'akan', 'fil': 'Akan', 'fo': 'akan', 'fr': 'akan', 'fy': 'Akan', 'ga': 'Acáinis', 'gd': 'Akan', 'gl': 'akan', 'gsw': 'Akan', 'gu': 'અકાન', 'guz': 'Kiakan', 'ha': 'Akan', 'he': 'אקאן', 'hi': 'अकन', 'hi-Latn': 'Akan', 'hr': 'akanski', 'hsb': 'akanšćina', 'ht': 'akan', 'hu': 'akan', 'hy': 'աքան', 'ia': 'akan', 'id': 'Akan', 'ig': 'Akan', 'is': 'akan', 'it': 'akan', 'ja': 'アカン語', 'jmc': 'Kiakanyi', 'jv': 'Akan', 'ka': 'აკანი', 'kab': 'Takanit', 'kam': 'Kiakan', 'kde': 'Chakan', 'kea': 'akan', 'kgp': 'akỹn', 'khq': 'Akan senni', 'ki': 'Kiakan', 'kk': 'акан тілі', 'kk-Cyrl': 'акан тілі', 'kln': 'kutitab Aka', 'km': 'អាកាន', 'kn': 'ಅಕಾನ್', 'ko': '아칸어', 'kok': 'अकान', 'kok-Deva': 'अकान', 'ks': 'اَکان', 'ks-Arab': 'اَکان', 'ksb': 'Kiakan', 'ksf': 'riakan', 'ksh': 'Akahnesch', 'ku': 'akanî', 'ky': 'аканча', 'lag': 'Kɨakáani', 'lb': 'Akan', 'lg': 'Lu-akaani', 'lld': 'akan', 'ln': 'akan', 'lo': 'ອາການ', 'lrc': 'آکان', 'lt': 'akanų', 'lu': 'Liakan', 'luo': 'Kiakan', 'luy': 'Kiakan', 'lv': 'akanu', 'mai': 'अकेन', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lAkan', 'mer': 'Kĩakani', 'mfe': 'akan', 'mg': 'Akan', 'mgh': 'Ikan', 'mi': 'Ākana', 'mk': 'акански', 'ml': 'അകാൻ\u200c', 'mn': 'акан', 'mr': 'अकान', 'ms': 'Akan', 'mt': 'Akan', 'mua': 'akaŋ', 'my': 'အာကန်', 'mzn': 'آکان', 'naq': 'Akangowab', 'nb': 'akan', 'nd': 'isi-Akhani', 'ne': 'आकान', 'nl': 'Akan', 'nmg': 'Kiɛl akan', 'nn': 'akan', 'no': 'akan', 'nqo': 'ߊߞߊ߲ߞߊ', 'nus': 'Thok aka̱ni', 'nyn': 'Orukani', 'or': 'ଅକନ୍', 'pa': 'ਅਕਾਨ', 'pa-Guru': 'ਅਕਾਨ', 'pcm': 'Akan Lángwej', 'pl': 'akan', 'ps': 'اکاني', 'pt': 'akan', 'qu': 'Akan Simi', 'rm': 'akan', 'rn': 'Igikani', 'ro': 'akan', 'rof': 'Kiakani', 'ru': 'акан', 'rwk': 'Kiakanyi', 'saq': 'Kiakan', 'sat': 'ᱟᱠᱟᱱ', 'sat-Olck': 'ᱟᱠᱟᱱ', 'sbp': 'Ishiyakani', 'sc': 'akan', 'sd': 'اڪان', 'sd-Arab': 'اڪان', 'seh': 'akan', 'ses': 'Akan senni', 'sg': 'Akâan', 'shi': 'ⵜⴰⴽⴰⵏⵜ', 'shi-Latn': 'Takant', 'shi-Tfng': 'ⵜⴰⴽⴰⵏⵜ', 'si': 'අකාන්', 'sk': 'akančina', 'sl': 'akanščina', 'smn': 'akankielâ', 'sn': 'chiAkani', 'so': 'Akan', 'sq': 'akanisht', 'sr': 'акански', 'sr-Cyrl': 'акански', 'sr-Latn': 'akanski', 'sv': 'akan', 'sw': 'Kiakani', 'ta': 'அகான்', 'te': 'అకాన్', 'teo': 'Kiakan', 'th': 'อาคาน', 'ti': 'ኣካን', 'tk': 'akan dili', 'to': 'lea fakaʻakani', 'tr': 'Akan', 'twq': 'Akan senni', 'tzm': 'Takanit', 'ug': 'ئاكانچە', 'uk': 'акан', 'ur': 'اکان', 'uz': 'akan', 'uz-Cyrl': 'аканча', 'uz-Latn': 'akan', 'vai': 'ꕉꕪꘋ', 'vai-Latn': 'Akaŋ', 'vai-Vaii': 'ꕉꕪꘋ', 'vec': 'akan', 'vi': 'Tiếng Akan', 'vun': 'Kiakanyi', 'xnr': 'अकन', 'xog': 'Oluakaani', 'yav': 'akánɛ', 'yo': 'Èdè Akani', 'yrl': 'akã', 'yue': '阿坎文', 'yue-Hans': '阿坎文', 'yue-Hant': '阿坎文', 'zgh': 'ⵜⴰⴽⴰⵏⵜ', 'zh': '阿肯语', 'zh-Hans': '阿肯语', 'zh-Hant': '阿坎文', 'zu': 'isi-Akan'},
    'akb': {'en': 'Batak Angkola'},
    'akc': {'en': 'Mpur'},
    'akd': {'en': 'Ukpet-Ehom'},
    'ake': {'en': 'Akawaio'},
    'akf': {'en': 'Akpa'},
    'akg': {'en': 'Anakalangu'},
    'akh': {'en': 'Angal Heneng'},
    'aki': {'en': 'Aiome'},
    'akj': {'en': 'Aka-Jeru'},
    'akk': {'am': 'አካዲያን', 'ar': 'الأكادية', 'ast': 'acadianu', 'az': 'akkad', 'az-Latn': 'akkad', 'be': 'акадская', 'be-tarask': 'акадская', 'bg': 'акадски', 'bn': 'আক্কাদিয়ান', 'br': 'akadeg', 'brx': 'अकाडिनी', 'bs': 'akadijski', 'bs-Cyrl': 'акадијски', 'bs-Latn': 'akadijski', 'ca': 'accadi', 'ca-ES-valencia': 'accadi', 'ccp': '𑄃𑄇𑄳𑄦𑄴𑄘𑄨𑄠𑄚𑄴', 'cs': 'akkadština', 'cy': 'Acadeg', 'da': 'akkadisk', 'de': 'Akkadisch', 'el': 'Ακάντιαν', 'el-polyton': 'Ακάντιαν', 'en': 'Akkadian', 'es': 'acadio', 'et': 'akadi', 'fa': 'اکدی', 'fi': 'akkadi', 'fr': 'akkadien', 'fy': 'Akkadysk', 'ga': 'Acáidis', 'gd': 'Acadais', 'gsw': 'Akkadisch', 'gu': 'અક્કાદીયાન', 'he': 'אכדית', 'hi': 'अक्कादी', 'hi-Latn': 'Akkadian', 'hr': 'akadski', 'ht': 'akkadien', 'hu': 'akkád', 'hy': 'աքքադերեն', 'id': 'Akkadia', 'is': 'akkadíska', 'it': 'accado', 'ja': 'アッカド語', 'ka': 'აქადური', 'kgp': 'akajỹnũ', 'kn': 'ಅಕ್ಕಾಡಿಯನ್', 'ko': '아카드어', 'ks': 'اَکادِیَن', 'ks-Arab': 'اَکادِیَن', 'ksh': 'Akahdesch', 'lb': 'Akkadesch', 'lo': 'ອັກກາດຽມ', 'lt': 'akadianų', 'lv': 'akadiešu', 'mk': 'акадски', 'ml': 'അക്കാഡിയൻ', 'mr': 'अक्केडियन', 'mt': 'Akkadjen', 'nb': 'akkadisk', 'ne': 'अक्कादियाली', 'nl': 'Akkadisch', 'nn': 'akkadisk', 'no': 'akkadisk', 'or': 'ଆକାଡିଆନ୍', 'pl': 'akadyjski', 'pt': 'acadiano', 'rm': 'accadic', 'ro': 'akkadiană', 'ru': 'аккадский', 'sk': 'akkadčina', 'sl': 'akadščina', 'sr': 'акадијски', 'sr-Cyrl': 'акадијски', 'sr-Latn': 'akadijski', 'sv': 'akkadiska', 'ta': 'அக்கேதியன்', 'te': 'అక్కాడియాన్', 'th': 'อักกาด', 'to': 'lea fakaʻakatia', 'tr': 'Akad Dili', 'ug': 'ئاككادچە', 'uk': 'аккадська', 'vi': 'Tiếng Akkadia', 'yi': 'אַקאַדיש', 'yrl': 'akadiãnu', 'yue': '阿卡德文', 'yue-Hans': '阿卡德文', 'yue-Hant': '阿卡德文', 'zh': '阿卡德语', 'zh-Hans': '阿卡德语', 'zh-Hant': '阿卡德文'},
    'akl': {'en': 'Aklanon'},
    'akm': {'en': 'Aka-Bo'},
    'ako': {'en': 'Akurio'},
    'akp': {'en': 'Siwu'},
    'akq': {'en': 'Ak'},
    'akr': {'en': 'Araki'},
    'aks': {'en': 'Akaselem'},
    'akt': {'en': 'Akolet'},
    'aku': {'en': 'Akum'},
    'akv': {'en': 'Akhvakh'},
    'akw': {'en': 'Akwa'},
    'akx': {'en': 'Aka-Kede'},
    'aky': {'en': 'Aka-Kol'},
    'akz': {'am': 'አላባማ', 'ast': 'alabama', 'br': 'alabamaeg', 'ca': 'alabama', 'ca-ES-valencia': 'alabama', 'cs': 'alabamština', 'cy': 'Alabamäeg', 'de': 'Alabama', 'en': 'Alabama', 'et': 'alabama', 'fa': 'آلابامایی', 'fi': 'alabama', 'fr': 'alabama', 'gd': 'Alabama', 'hi-Latn': 'Alabama', 'ht': 'alabama', 'id': 'Alabama', 'it': 'alabama', 'ja': 'アラバマ語', 'lb': 'Alabama', 'lt': 'alabamiečių', 'mk': 'алабамски', 'nb': 'alabama', 'ne': 'अलाबामा', 'nl': 'Alabama', 'nn': 'alabama', 'no': 'alabama', 'pl': 'alabama', 'sv': 'Alabama-muskogee', 'th': 'แอละแบมา', 'to': 'lea fakaʻalapama', 'tr': 'Alabamaca', 'uk': 'алабама', 'vi': 'Tiếng Alabama', 'yue': '阿拉巴馬文', 'yue-Hans': '阿拉巴马文', 'yue-Hant': '阿拉巴馬文', 'zh-Hant': '阿拉巴馬文'},
    'ala': {'en': 'Alago'},
    'alc': {'en': 'Qawasqar'},
    'ald': {'en': 'Alladian'},
    'ale': {'af': 'Aleut', 'am': 'አልዩት', 'ar': 'الأليوتية', 'as': 'আলেউট', 'ast': 'aleut', 'az': 'aleut', 'az-Cyrl': 'алеут', 'az-Latn': 'aleut', 'be': 'алеуцкая', 'be-tarask': 'алеуцкая', 'bg': 'алеутски', 'bgn': 'الیوتی', 'bn': 'আলেউত', 'br': 'aleouteg', 'brx': 'आलेउट', 'bs': 'aleutski', 'bs-Cyrl': 'аљут', 'bs-Latn': 'aleutski', 'ca': 'aleuta', 'ca-ES-valencia': 'aleuta', 'ccp': '𑄃𑄣𑄬𑄅𑄖𑄴', 'ce': 'алеутийн', 'chr': 'ᎠᎵᎤᏘ', 'ckb': 'ئالیوت', 'cs': 'aleutština', 'cy': 'Alewteg', 'da': 'aleutisk', 'de': 'Aleutisch', 'dsb': 'aleutišćina', 'el': 'Αλεούτ', 'el-polyton': 'Αλεούτ', 'en': 'Aleut', 'es': 'aleutiano', 'et': 'aleuudi', 'eu': 'aleutera', 'fa': 'آلئوتی', 'ff-Adlm': '𞤀𞤤𞤫𞤵𞤼𞤵𞥅𞤪𞤫', 'fi': 'aleutti', 'fil': 'Aleut', 'fo': 'aleut', 'fr': 'aléoute', 'fy': 'Aleut', 'ga': 'Ailiúitis', 'gd': 'Aleutais', 'gl': 'aleutiano', 'gsw': 'Aleutisch', 'gu': 'અલેઉત', 'ha': 'Aleut', 'he': 'אלאוט', 'hi': 'अलेउत', 'hi-Latn': 'Aleut', 'hr': 'aleutski', 'hsb': 'aleutišćina', 'ht': 'aléoute', 'hu': 'aleut', 'hy': 'ալեութերեն', 'ia': 'aleuto', 'id': 'Aleut', 'ig': 'Aleut', 'is': 'aleúska', 'it': 'aleuto', 'ja': 'アレウト語', 'jv': 'Aleut', 'ka': 'ალეუტური', 'kgp': 'arevute', 'kk': 'алеут тілі', 'kk-Cyrl': 'алеут тілі', 'km': 'អាលូត', 'kn': 'ಅಲೆಯುಟ್', 'ko': '알류트어', 'kok': 'आलिट', 'kok-Deva': 'आलिट', 'ks': 'اَلویتی', 'ks-Arab': 'اَلویتی', 'ksh': 'Ale’uhtesch', 'ku': 'alêwîtî', 'ky': 'алеутча', 'lb': 'Aleutesch', 'lo': 'ອາເລີດ', 'lt': 'aleutų', 'lv': 'aleutu', 'mai': 'अल्यूट', 'mi': 'Ariuta', 'mk': 'алеутски', 'ml': 'അലൂട്ട്', 'mn': 'алют', 'mr': 'अलेउत', 'ms': 'Aleut', 'mt': 'Aleut', 'my': 'အာလီယု', 'nb': 'aleutisk', 'ne': 'अलेउट', 'nl': 'Aleoetisch', 'nn': 'aleutisk', 'no': 'aleutisk', 'or': 'ଆଲେଇଟୁ', 'pa': 'ਅਲੇਉਟ', 'pa-Guru': 'ਅਲੇਉਟ', 'pcm': 'Alẹut', 'pl': 'aleucki', 'ps': 'اليوتي', 'pt': 'aleúte', 'qu': 'Aleut', 'rm': 'aleutic', 'ro': 'aleută', 'ru': 'алеутский', 'sah': 'Алеуттуу', 'sat': 'ᱟᱞᱮᱣᱴ', 'sat-Olck': 'ᱟᱞᱮᱣᱴ', 'sc': 'aleutinu', 'sd': 'اليوٽ', 'sd-Arab': 'اليوٽ', 'si': 'ඇලුඑට්', 'sk': 'aleutčina', 'sl': 'aleutščina', 'smn': 'aleutkielâ', 'so': 'U dhashay Ale', 'sq': 'aleutisht', 'sr': 'алеутски', 'sr-Cyrl': 'алеутски', 'sr-Latn': 'aleutski', 'sv': 'aleutiska', 'sw': 'Kialeut', 'ta': 'அலூட்', 'te': 'అలియుట్', 'th': 'อาลิวต์', 'ti': 'ኣለውትኛ', 'tk': 'aleut dili', 'to': 'lea fakaʻaleuti', 'tr': 'Aleut dili', 'ug': 'ئالېيۇتچە', 'uk': 'алеутська', 'ur': 'الیوت', 'uz': 'aleut', 'uz-Cyrl': 'алеут', 'uz-Latn': 'aleut', 'vec': 'aleutian', 'vi': 'Tiếng Aleut', 'xnr': 'अलेउत', 'yo': 'Èdè Aleti', 'yrl': 'areúti', 'yue': '阿留申文', 'yue-Hans': '阿留申文', 'yue-Hant': '阿留申文', 'zh': '阿留申语', 'zh-Hans': '阿留申语', 'zh-Hant': '阿留申文', 'zu': 'isi-Aleut'},
    'alf': {'en': 'Alege'},
    'alg': {'en': 'Algonquian languages'},
    'alg-pro': {'en': 'Proto-Algonquian'},
    'alh': {'en': 'Alawa'},
    'ali': {'en': 'Amaimon'},
    'alj': {'en': 'Alangan'},
    'alk': {'en': 'Alak'},
    'all': {'en': 'Allar'},
    'alm': {'en': 'Amblong'},
    'aln': {'ast': 'gheg d’Albania', 'br': 'gegeg', 'ca': 'albanès geg', 'ca-ES-valencia': 'albanès geg', 'cs': 'albánština (Gheg)', 'cy': 'Ghegeg Albania', 'de': 'Gegisch', 'en': 'Gheg Albanian', 'et': 'geegi', 'fi': 'gegi', 'fr': 'guègue', 'gd': 'Albàinis Ghegeach', 'hi-Latn': 'Gheg Albanian', 'ht': 'guègue', 'it': 'albanese ghego', 'ja': 'ゲグ・アルバニア語', 'lb': 'Gegesch', 'lt': 'albanų kalbos gegų tarmė', 'mk': 'гешки албански', 'nb': 'gegisk-albansk', 'ne': 'घेग अल्बानियाली', 'nl': 'Gegisch', 'nn': 'gegisk-albansk', 'no': 'gegisk-albansk', 'pl': 'albański gegijski', 'sv': 'gegiska', 'th': 'เกกแอลเบเนีย', 'to': 'lea fakaʻalapēnia-keki', 'tr': 'Gheg Arnavutçası', 'vi': 'Tiếng Gheg Albani', 'yue': '蓋格阿爾巴尼亞文', 'yue-Hans': '盖格阿尔巴尼亚文', 'yue-Hant': '蓋格阿爾巴尼亞文', 'zh-Hant': '蓋格阿爾巴尼亞文'},
    'alo': {'en': 'Larike-Wakasihu'},
    'alp': {'en': 'Alune'},
    'alq': {'en': 'Algonquin'},
    'alr': {'en': 'Alutor'},
    'als': {'en': 'Tosk Albanian'},
    'alt': {'af': 'Suid-Altai', 'am': 'ደቡባዊ አልታይ', 'ar': 'الألطائية الجنوبية', 'as': 'দাক্ষিণাত্য আল্টাই', 'ast': 'altai del sur', 'az': 'cənubi altay', 'az-Cyrl': 'ҹәнуби алтај', 'az-Latn': 'cənubi altay', 'be': 'паўднёваалтайская', 'be-tarask': 'паўднёваалтайская', 'bg': 'южноалтайски', 'bgn': 'جهلسرین آلتایی', 'bn': 'দক্ষিন আলতাই', 'br': 'altaieg ar Su', 'brx': 'खोलायारि आल्टाई', 'bs': 'južni altai', 'bs-Cyrl': 'јужни алтаи', 'bs-Latn': 'južni altai', 'ca': 'altaic meridional', 'ca-ES-valencia': 'altaic meridional', 'ccp': '𑄓𑄧𑄉𑄨𑄚𑄴 𑄃𑄣𑄴𑄖𑄭', 'ce': 'къилба алтайн', 'chr': 'ᏧᎦᎾᏮ ᏗᏜ ᎠᎵᏔᎢ', 'ckb': 'ئاڵتایی باشوور', 'cs': 'altajština (jižní)', 'cy': 'Altäeg Deheuol', 'da': 'sydaltaisk', 'de': 'Süd-Altaisch', 'dsb': 'pódpołdnjowa altaišćina', 'el': 'Νότια Αλτάι', 'el-polyton': 'Νότια Αλτάι', 'en': 'Southern Altai', 'es': 'altái meridional', 'et': 'altai', 'eu': 'hegoaldeko altaiera', 'fa': 'آلتایی جنوبی', 'ff-Adlm': '𞤀𞤤𞤼𞤢𞤴𞤪𞤫 𞤙𞤢𞥄𞤥𞤲𞤢𞥄𞤲𞤺𞤫𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'altai', 'fil': 'Southern Altai', 'fo': 'suður altai', 'fr': 'altaï du Sud', 'fy': 'Sûd-Altaïsk', 'ga': 'Altaeis an Deiscirt', 'gd': 'Altais Dheasach', 'gl': 'altai meridional', 'gsw': 'Süüd-Altaisch', 'gu': 'દક્ષિણ અલ્તાઇ', 'ha': 'Altai na Kudanci', 'he': 'אלטאי דרומית', 'hi': 'दक्षिणी अल्ताई', 'hi-Latn': 'Southern Altai', 'hr': 'južni altai', 'hsb': 'južna altaišćina', 'ht': 'altaï du Sud', 'hu': 'dél-altaji', 'hy': 'հարավային ալթայերեն', 'ia': 'altai del sud', 'id': 'Altai Selatan', 'ig': 'Southern Altai', 'is': 'suðuraltaíska', 'it': 'altai meridionale', 'ja': '南アルタイ語', 'jv': 'Altai Sisih Kidul', 'ka': 'სამხრეთ ალთაური', 'kgp': 'artaj sur', 'kk': 'оңтүстік алтай тілі', 'kk-Cyrl': 'оңтүстік алтай тілі', 'km': 'អាល់តៃខាងត្បូង', 'kn': 'ದಕ್ಷಿಣ ಅಲ್ಟಾಯ್', 'ko': '남부 알타이어', 'kok': 'दक्षिणी अल्टाय', 'kok-Deva': 'दक्षिणी अल्टाय', 'ks': 'جنوٗبی اَلتایی', 'ks-Arab': 'جنوٗبی اَلتایی', 'ku': 'altayîya başûrî', 'ky': 'түштүк алтайча', 'lb': 'Süd-Alaesch', 'lkt': 'Itóǧata Altai Iyápi', 'lo': 'ອານໄຕໃຕ້', 'lt': 'pietų Altajaus', 'lv': 'dienvidaltajiešu', 'mai': 'दक्षिणी अलटाई', 'mi': 'Ātai ki te Tonga', 'mk': 'јужноалтајски', 'ml': 'തെക്കൻ അൾത്തായി', 'mn': 'өмнөд алтай', 'mr': 'दक्षिणात्य अल्ताई', 'ms': 'Altai Selatan', 'mt': 'Altai tan-Nofsinhar', 'my': 'တောင် အာလ်တိုင်း', 'nb': 'søraltaisk', 'ne': 'दक्षिणी आल्टाइ', 'nl': 'Zuid-Altaïsch', 'nn': 'sør-altaj', 'no': 'søraltaisk', 'or': 'ଦକ୍ଷିଣ ଆଲ୍ଟାଇ', 'pa': 'ਦੱਖਣੀ ਅਲਤਾਈ', 'pa-Guru': 'ਦੱਖਣੀ ਅਲਤਾਈ', 'pcm': 'Saútán Altai Lángwej', 'pl': 'południowoałtajski', 'ps': 'سویل الټای', 'pt': 'altai meridional', 'qu': 'Altai Meridional', 'rm': 'altaic dal sid', 'ro': 'altaică meridională', 'ru': 'южноалтайский', 'sc': 'altai meridionale', 'sd': 'ڏکڻ التائي', 'sd-Arab': 'ڏکڻ التائي', 'si': 'සතර්න් අල්ටය්', 'sk': 'južná altajčina', 'sl': 'južna altajščina', 'smn': 'maadâaltaikielâ', 'so': 'Southern Altai', 'sq': 'altaishte jugore', 'sr': 'јужноалтајски', 'sr-Cyrl': 'јужноалтајски', 'sr-Latn': 'južnoaltajski', 'sv': 'sydaltaiska', 'sw': 'Kialtai', 'ta': 'தெற்கு அல்தை', 'te': 'దక్షిణ ఆల్టై', 'th': 'อัลไตใต้', 'ti': 'ደቡባዊ ኣልታይ', 'tk': 'günorta Altaý dili', 'to': 'lea fakaʻalitai-tonga', 'tr': 'Güney Altayca', 'ug': 'جەنۇبى ئالتاي تىللىرى', 'uk': 'південноалтайська', 'ur': 'جنوبی الٹائی', 'uz': 'janubiy oltoy', 'uz-Latn': 'janubiy oltoy', 'vec': 'altài meridionale', 'vi': 'Tiếng Altai Miền Nam', 'xnr': 'दखणे बखें अल्ताई', 'yo': 'Èdè Gusu Ata', 'yrl': 'autai meridiunau', 'yue': '南阿爾泰文', 'yue-Hans': '南阿尔泰文', 'yue-Hant': '南阿爾泰文', 'zh': '南阿尔泰语', 'zh-Hans': '南阿尔泰语', 'zh-Hant': '南阿爾泰文', 'zu': 'isi-Southern Altai'},
    'alu': {'en': "'Are'are"},
    'alv': {'en': 'Atlantic-Congo languages'},
    'alv-edo-pro': {'en': 'Proto-Edoid'},
    'alv-pro': {'en': 'Proto-Atlantic-Congo'},
    'alw': {'en': 'Alaba-K’abeena'},
    'alx': {'en': 'Amol'},
    'aly': {'en': 'Alyawarr'},
    'alz': {'en': 'Alur'},
    'am': {'af': 'Amharies', 'agq': 'Àmalì', 'ak': 'Amarik', 'am': 'አማርኛ', 'ar': 'الأمهرية', 'as': 'আমহাৰিক', 'asa': 'Kiamhari', 'ast': 'amháricu', 'az': 'amhar', 'az-Arab': 'آمهارجا', 'az-Cyrl': 'амһар', 'az-Latn': 'amhar', 'bal-Latn': 'Amhari', 'bas': 'Hɔp u amhārìk', 'be': 'амхарская', 'be-tarask': 'амхарская', 'bem': 'Ichi Amhari', 'bez': 'Hiamhari', 'bg': 'амхарски', 'bgn': 'امهاری', 'bm': 'amarikikan', 'bn': 'আমহারিক', 'br': 'amhareg', 'brx': 'आम्हारिक', 'bs': 'amharski', 'bs-Cyrl': 'амхарски', 'bs-Latn': 'amharski', 'ca': 'amhàric', 'ca-ES-valencia': 'amhàric', 'ccp': '𑄃𑄟𑄴𑄦𑄢𑄨𑄇𑄴', 'ce': 'амхаройн', 'cgg': 'Orumariki', 'chr': 'ᎠᎹᎭᎵᎩ', 'ckb': 'ئەمھەری', 'cs': 'amharština', 'cy': 'Amhareg', 'da': 'amharisk', 'dav': 'Kiamhari', 'de': 'Amharisch', 'dje': 'Amhaarik senni', 'dsb': 'amharšćina', 'dyo': 'amharik', 'dz': 'ཨམ་ཧ་རིཀ་ཁ', 'ebu': 'Kĩamhari', 'ee': 'amhariagbe', 'el': 'Αμχαρικά', 'el-polyton': 'Αμχαρικά', 'en': 'Amharic', 'eo': 'amhara', 'es': 'amárico', 'et': 'amhara', 'eu': 'amharera', 'ewo': 'Ǹkɔ́bɔ amária', 'fa': 'امهری', 'ff': 'Amarik', 'ff-Adlm': '𞤀𞤥𞤸𞤢𞤪𞤭𞥅𞤪𞤫', 'ff-Latn': 'Amarik', 'fi': 'amhara', 'fil': 'Amharic', 'fo': 'amhariskt', 'fr': 'amharique', 'fur': 'amaric', 'fy': 'Amhaarsk', 'ga': 'Amáiris', 'gd': 'Amtharais', 'gl': 'amhárico', 'gsw': 'Amharisch', 'gu': 'એમ્હારિક', 'guz': 'Kiamhari', 'ha': 'Amharik', 'he': 'אמהרית', 'hi': 'अम्हेरी', 'hi-Latn': 'Amharic', 'hr': 'amharski', 'hsb': 'amharšćina', 'ht': 'amharique', 'hu': 'amhara', 'hy': 'ամհարերեն', 'ia': 'amharico', 'id': 'Amharik', 'ig': 'Amariikị', 'is': 'amharíska', 'it': 'amarico', 'ja': 'アムハラ語', 'jmc': 'Kiamharyi', 'jv': 'Amharik', 'ka': 'ამჰარული', 'kab': 'Tamahrict', 'kam': 'Kiamhari', 'kde': 'Chamhali', 'kea': 'amáriku', 'kgp': 'amỹriko', 'khq': 'Amhaarik senni', 'ki': 'Kiamhari', 'kk': 'амхар тілі', 'kk-Cyrl': 'амхар тілі', 'kln': 'kutitab Amariek', 'km': 'អាំហារិក', 'kn': 'ಅಂಹರಿಕ್', 'ko': '암하라어', 'kok': 'अमहारिक्', 'kok-Deva': 'अमहारिक्', 'ks': 'اَمہاری', 'ks-Arab': 'اَمہاری', 'ksb': 'Kiamhali', 'ksf': 'riamarik', 'ksh': 'Amhahresch', 'ku': 'amharî', 'kxv': 'aarmenia', 'kxv-Deva': 'अम्हेरी', 'kxv-Latn': 'aarmenia', 'kxv-Orya': 'ଆମ୍\u200dହେରି', 'kxv-Telu': 'ఆమ్హెరి', 'ky': 'амхарча', 'lag': 'Kɨmʉháari', 'lb': 'Amharesch', 'lg': 'Lu-amhariki', 'lkt': 'Amharic Iyápi', 'lld': 'amarich', 'ln': 'liamariki', 'lo': 'ອຳຮາຣິກ', 'lrc': 'أمھأری', 'lt': 'amharų', 'lu': 'Liamhariki', 'luo': 'Kiamhari', 'luy': 'Kiamhari', 'lv': 'amharu', 'mai': 'अम्हरिक', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lAmhari', 'mer': 'Kĩamarĩki', 'mfe': 'amarik', 'mg': 'Amharika', 'mgh': 'Imhari', 'mi': 'Amahereka', 'mk': 'амхарски', 'ml': 'അംഹാരിക്', 'mn': 'амхар', 'mr': 'अम्हारिक', 'ms': 'Amharic', 'mt': 'Amhariku', 'mua': 'amharik', 'my': 'အမ်ဟာရစ်ခ်', 'mzn': 'امهری', 'naq': 'Amharicgowab', 'nb': 'amharisk', 'nd': 'isi-Amaharikhi', 'ne': 'अम्हारिक', 'nl': 'Amhaars', 'nmg': 'Kiɛl amaria', 'nn': 'amharisk', 'no': 'amharisk', 'nqo': 'ߊߡߑߤߊߙߌ', 'nus': 'Thok bunyni', 'nyn': 'Orumariki', 'om': 'Afaan Amaaraa', 'or': 'ଆମହାରକି', 'pa': 'ਅਮਹਾਰਿਕ', 'pa-Guru': 'ਅਮਹਾਰਿਕ', 'pcm': 'Amhárík Lángwej', 'pl': 'amharski', 'ps': 'امهاري', 'pt': 'amárico', 'qu': 'Amarico Simi', 'rm': 'amaric', 'rn': 'Ikimuhariki', 'ro': 'amharică', 'rof': 'Kiamhari', 'ru': 'амхарский', 'rw': 'Inyamuhariki', 'rwk': 'Kiamharyi', 'sah': 'Амхаардыы', 'saq': 'Kiamhari', 'sat': 'ᱟᱢᱦᱟᱨᱤᱠ', 'sat-Olck': 'ᱟᱢᱦᱟᱨᱤᱠ', 'sbp': 'Ishiyamuhali', 'sc': 'amàricu', 'sd': 'امهاري', 'sd-Arab': 'امهاري', 'seh': 'amárico', 'ses': 'Amhaarik senni', 'sg': 'Amarîki', 'shi': 'ⵜⴰⵎⵀⴰⵔⵉⵜ', 'shi-Latn': 'Tamharit', 'shi-Tfng': 'ⵜⴰⵎⵀⴰⵔⵉⵜ', 'si': 'ඇම්හාරික්', 'sk': 'amharčina', 'sl': 'amharščina', 'smn': 'amharakielâ', 'sn': 'chiAmaric', 'so': 'Axmaar', 'sq': 'amarisht', 'sr': 'амхарски', 'sr-Cyrl': 'амхарски', 'sr-Latn': 'amharski', 'sv': 'amhariska', 'sw': 'Kiamhari', 'syr': 'ܐܡܪܢܝܬ', 'ta': 'அம்ஹாரிக்', 'te': 'అమ్హారిక్', 'teo': 'Kiamhari', 'tg': 'амҳарӣ', 'th': 'อัมฮารา', 'ti': 'ኣምሓርኛ', 'tk': 'amhar dili', 'to': 'lea fakaʻameliki', 'tr': 'Amharca', 'tt': 'амхар', 'twq': 'Amhaarik senni', 'tzm': 'Tamharit', 'ug': 'ئامھارچە', 'uk': 'амхарська', 'ur': 'امہاری', 'uz': 'amxar', 'uz-Cyrl': 'амхарча', 'uz-Latn': 'amxar', 'vai': 'ꕉꕆꕌꔸ', 'vai-Latn': 'Amihári', 'vai-Vaii': 'ꕉꕆꕌꔸ', 'vec': 'amàrego', 'vi': 'Tiếng Amharic', 'vun': 'Kiamharyi', 'wae': 'Amhariš', 'wo': 'Amharik', 'xh': 'IsiAmharic', 'xnr': 'अम्हेरी', 'xog': 'Oluamhariki', 'yav': 'amalíke', 'yi': 'אַמהאַריש', 'yo': 'Èdè Amariki', 'yrl': 'amáriku', 'yue': '阿姆哈拉文', 'yue-Hans': '阿姆哈拉文', 'yue-Hant': '阿姆哈拉文', 'zgh': 'ⵜⴰⵎⵀⴰⵔⵉⵜ', 'zh': '阿姆哈拉语', 'zh-Hans': '阿姆哈拉语', 'zh-Hant': '阿姆哈拉文', 'zu': 'isi-Amharic'},
    'ama': {'en': 'Amanayé'},
    'amb': {'en': 'Ambo'},
    'amc': {'en': 'Amahuaca'},
    'ame': {'en': "Yanesha'"},
    'amf': {'en': 'Hamer-Banna'},
    'amg': {'en': 'Amurdak'},
    'ami': {'en': 'Amis'},
    'amj': {'en': 'Amdang'},
    'amk': {'en': 'Ambai'},
    'aml': {'en': 'War-Jaintia'},
    'amm': {'en': 'Ama (Papua New Guinea)'},
    'amn': {'en': 'Amanab'},
    'amo': {'en': 'Amo'},
    'amp': {'en': 'Alamblak'},
    'amq': {'en': 'Amahai'},
    'amr': {'en': 'Amarakaeri'},
    'ams': {'en': 'Southern Amami-Oshima'},
    'amt': {'en': 'Amto'},
    'amu': {'en': 'Guerrero Amuzgo'},
    'amv': {'en': 'Ambelau'},
    'amw': {'en': 'Western Neo-Aramaic'},
    'amx': {'en': 'Anmatyerre'},
    'amy': {'en': 'Ami'},
    'amz': {'en': 'Atampaya'},
    'an': {'af': 'Aragonees', 'am': 'አራጎንስ', 'an': 'aragonés', 'ar': 'الأراغونية', 'as': 'আৰ্গোনিজ', 'ast': 'aragonés', 'az': 'araqon', 'az-Cyrl': 'арагон', 'az-Latn': 'araqon', 'bal-Latn': 'Aragóni', 'be': 'арагонская', 'be-tarask': 'арагонская', 'bg': 'арагонски', 'bgn': 'آراگونی', 'bn': 'আর্গোনিজ', 'br': 'aragoneg', 'brx': 'आरागनिस', 'bs': 'aragonski', 'bs-Cyrl': 'арагонежански', 'bs-Latn': 'aragonski', 'ca': 'aragonès', 'ca-ES-valencia': 'aragonès', 'ccp': '𑄃𑄢𑄴𑄉𑄮𑄚𑄨𑄎𑄴', 'ce': 'арагонойн', 'chr': 'ᎠᏩᎪᏂᏏ', 'ckb': 'ئاراگۆنی', 'cs': 'aragonština', 'cy': 'Aragoneg', 'da': 'aragonsk', 'de': 'Aragonesisch', 'dsb': 'aragonšćina', 'el': 'Αραγονικά', 'el-polyton': 'Αραγονικά', 'en': 'Aragonese', 'en-Dsrt': '𐐈𐑉𐐲𐑀𐐱𐑌𐐨𐑆', 'es': 'aragonés', 'et': 'aragoni', 'eu': 'aragoiera', 'fa': 'آراگونی', 'ff-Adlm': '𞤀𞤪𞤢𞤺𞤮𞤲𞤪𞤫', 'fi': 'aragonia', 'fil': 'Aragonese', 'fo': 'aragoniskt', 'fr': 'aragonais', 'fur': 'aragonês', 'fy': 'Aragoneesk', 'ga': 'Aragóinis', 'gd': 'Aragonais', 'gl': 'aragonés', 'gsw': 'Aragonesisch', 'gu': 'અર્ગોનીઝ', 'ha': 'Aragonesanci', 'he': 'אראגונית', 'hi': 'अरागोनी', 'hi-Latn': 'Aragonese', 'hr': 'aragonski', 'hsb': 'aragonšćina', 'ht': 'aragonais', 'hu': 'aragonéz', 'hy': 'արագոներեն', 'ia': 'aragonese', 'id': 'Aragon', 'ig': 'Aragonisị', 'is': 'aragonska', 'it': 'aragonese', 'ja': 'アラゴン語', 'jv': 'Aragonese', 'ka': 'არაგონული', 'kgp': 'aragonẽ', 'kk': 'арагон тілі', 'kk-Cyrl': 'арагон тілі', 'km': 'អារ៉ាហ្គោន', 'kn': 'ಅರಗೊನೀಸ್', 'ko': '아라곤어', 'kok': 'आरागोनिस', 'kok-Deva': 'आरागोनिस', 'ks': 'اَراگونی', 'ks-Arab': 'اَراگونی', 'ksh': 'Arrajonehsesch', 'ku': 'aragonî', 'ky': 'арагончо', 'lb': 'Aragonesesch', 'lld': 'aragonesc', 'lo': 'ອາຣາໂກເນັດ', 'lt': 'aragonesų', 'lv': 'aragoniešu', 'mai': 'अरगोनीज', 'mi': 'Arakonihi', 'mk': 'арагонски', 'ml': 'അരഗോണീസ്', 'mn': 'арагон', 'mr': 'अर्गोनीज', 'ms': 'Aragon', 'mt': 'Aragoniż', 'my': 'အာရာဂွန်', 'nb': 'aragonsk', 'ne': 'अरागोनी', 'nl': 'Aragonees', 'nn': 'aragonsk', 'no': 'aragonsk', 'or': 'ଆର୍ଗୋନୀ', 'pa': 'ਅਰਾਗੋਨੀ', 'pa-Guru': 'ਅਰਾਗੋਨੀ', 'pcm': 'Aragónẹ́sẹ Lángwej', 'pl': 'aragoński', 'ps': 'اراگونېسي', 'pt': 'aragonês', 'qu': 'Aragonesa', 'rm': 'aragonais', 'ro': 'aragoneză', 'ru': 'арагонский', 'sat': 'ᱟᱨᱟᱜᱚᱱᱤᱡᱽ', 'sat-Olck': 'ᱟᱨᱟᱜᱚᱱᱤᱡᱽ', 'sc': 'aragonesu', 'sd': 'ارگني', 'sd-Arab': 'ارگني', 'se': 'aragoniagiella', 'si': 'ඇරගොනීස්', 'sk': 'aragónčina', 'sl': 'aragonščina', 'smn': 'aragoniakielâ', 'so': 'U dhashay Aragon', 'sq': 'aragonezisht', 'sr': 'арагонски', 'sr-Cyrl': 'арагонски', 'sr-Latn': 'aragonski', 'sv': 'aragonesiska', 'sw': 'Kiaragoni', 'syr': 'ܐܪܐܓܘܢܝܬ', 'ta': 'ஆர்கோனீஸ்', 'te': 'అరగోనిస్', 'th': 'อารากอน', 'ti': 'ኣራጎንኛ', 'tk': 'aragon dili', 'to': 'lea fakaʻalakoni', 'tr': 'Aragonca', 'ug': 'ئاراگونچە', 'uk': 'арагонська', 'ur': 'اراگونیز', 'uz': 'aragon', 'uz-Cyrl': 'арагон', 'uz-Latn': 'aragon', 'vec': 'aragoneze', 'vi': 'Tiếng Aragon', 'xnr': 'अर्गोनी', 'yi': 'אַראַגאניש', 'yo': 'Èdè Aragoni', 'yrl': 'aragunẽi', 'yue': '阿拉貢文', 'yue-Hans': '阿拉贡文', 'yue-Hant': '阿拉貢文', 'zh': '阿拉贡语', 'zh-Hans': '阿拉贡语', 'zh-Hant': '阿拉貢文', 'zu': 'isi-Aragonese'},
    'ana': {'en': 'Andaqui'},
    'anb': {'en': 'Andoa'},
    'anc': {'en': 'Ngas'},
    'and': {'en': 'Ansus'},
    'ane': {'en': 'Xârâcùù'},
    'anf': {'en': 'Animere'},
    'ang': {'ar': 'الإنجليزية القديمة', 'ast': 'inglés antiguu', 'az': 'qədim ingilis', 'az-Latn': 'qədim ingilis', 'be': 'стараанглійская', 'be-tarask': 'стараанглійская', 'bg': 'староанглийски', 'bn': 'প্রাচীন ইংরেজী', 'br': 'hensaozneg', 'brx': 'पुरानी अंग्रेज़ी', 'bs': 'staroengleski', 'bs-Cyrl': 'староенглески', 'bs-Latn': 'staroengleski', 'ca': 'anglès antic', 'ca-ES-valencia': 'anglès antic', 'ccp': '𑄛𑄪𑄢𑄧𑄚𑄨 𑄃𑄟𑄧𑄣𑄧𑄢𑄴 𑄃𑄨𑄁𑄢𑄬𑄎𑄩', 'cs': 'staroangličtina', 'cy': 'Hen Saesneg', 'da': 'oldengelsk', 'de': 'Altenglisch', 'dsb': 'anglosaksojšćina', 'el': 'Παλαιά Αγγλικά', 'el-polyton': 'Παλαιά Αγγλικά', 'en': 'Old English', 'es': 'inglés antiguo', 'et': 'vanainglise', 'fa': 'انگلیسی باستان', 'fi': 'muinaisenglanti', 'fr': 'ancien anglais', 'fur': 'vieri inglês', 'fy': 'âldingelsk', 'ga': 'Sean-Bhéarla', 'gd': 'Seann-Bheurla', 'gsw': 'Altänglisch', 'gu': 'જુની અંગ્રેજી', 'he': 'אנגלית עתיקה', 'hi': 'पुरानी अंग्रेज़ी', 'hi-Latn': 'Old English', 'hr': 'staroengleski', 'hsb': 'anglosakšćina', 'ht': 'ancien anglais', 'hu': 'óangol', 'hy': 'հին անգլերեն', 'id': 'Inggris Kuno', 'is': 'fornenska', 'it': 'inglese antico', 'ja': '古英語', 'ka': 'ძველი ინგლისური', 'kgp': 'ĩnhgrej arkajiku', 'kn': 'ಪ್ರಾಚೀನ ಇಂಗ್ಲೀಷ್', 'ko': '고대 영어', 'ks': 'پرون اَنگریٖزی', 'ks-Arab': 'پرون اَنگریٖزی', 'ksh': 'Ahl Änglesch', 'lb': 'Alenglesch', 'lo': 'ອັງກິດໂບຮານ', 'lt': 'senoji anglų', 'lv': 'senangļu', 'mk': 'староанглиски', 'ml': 'പഴയ ഇംഗ്ലീഷ്', 'mr': 'पुरातन इंग्रजी', 'mt': 'Ingliż Antik', 'my': 'အင်ဂလို ဆက္ကစွန်', 'nb': 'gammelengelsk', 'ne': 'पुरातन अङ्ग्रेजी', 'nl': 'Oudengels', 'nn': 'gammalengelsk', 'no': 'gammelengelsk', 'or': 'ପୁରୁଣା ଇଁରାଜୀ', 'os': 'рагон англисаг', 'pa': 'ਪੁਰਾਣੀ ਅੰਗਰੇਜ਼ੀ', 'pa-Guru': 'ਪੁਰਾਣੀ ਅੰਗਰੇਜ਼ੀ', 'pl': 'staroangielski', 'pt': 'inglês arcaico', 'rm': 'englais vegl', 'ro': 'engleză veche', 'ru': 'староанглийский', 'se': 'boares eaŋgalasgiella', 'sk': 'stará angličtina', 'sl': 'stara angleščina', 'sr': 'староенглески', 'sr-Cyrl': 'староенглески', 'sr-Latn': 'staroengleski', 'sv': 'fornengelska', 'sw': 'Kiingereza cha Kale', 'ta': 'பழைய ஆங்கிலம்', 'te': 'ప్రాచీన ఆంగ్లం', 'th': 'อังกฤษโบราณ', 'to': 'lea fakapālangi-motuʻa', 'tr': 'Eski İngilizce', 'ug': 'قەدىمكى ئىنگلىزچە', 'uk': 'давньоанглійська', 'vi': 'Tiếng Anh cổ', 'yi': 'אַלט ענגליש', 'yrl': 'ẽgirixi arkaiku', 'yue': '古英文', 'yue-Hans': '古英文', 'yue-Hant': '古英文', 'zh': '古英语', 'zh-Hans': '古英语', 'zh-Hant': '古英文', 'ang': 'Ænglisc'},
    'anh': {'en': 'Nend'},
    'ani': {'en': 'Andi'},
    'anj': {'en': 'Anor'},
    'ank': {'en': 'Goemai'},
    'anl': {'en': 'Anu-Hkongso Chin'},
    'anm': {'en': 'Anal'},
    'ann': {'af': 'Obolo', 'am': 'ኦቦሎ', 'ann': 'Obolo', 'ar': 'أوبلو', 'as': 'অ’ব’ল’', 'az': 'obolo', 'az-Latn': 'obolo', 'bal-Latn': 'Obóló', 'be': 'абола', 'be-tarask': 'абола', 'bg': 'оболо', 'bn': 'ওবোলো', 'br': 'obolo', 'brx': 'अबल’', 'bs': 'obolo', 'bs-Cyrl': 'оболо', 'bs-Latn': 'obolo', 'ca': 'obolo', 'ca-ES-valencia': 'obolo', 'chr': 'ᎣᏉᎶ', 'cs': 'obolo', 'cy': 'Obolo', 'da': 'obolo', 'de': 'Obolo', 'dsb': 'obološćina', 'el': 'Ομπόλο', 'el-polyton': 'Ομπόλο', 'en': 'Obolo', 'es': 'obolo', 'et': 'obolo', 'eu': 'oboloera', 'fa': 'اوبولو', 'ff-Adlm': '𞤌𞤦𞤮𞤤𞤮𞥅𞤪𞤫', 'fi': 'obolo', 'fil': 'Obolo', 'fr': 'obolo', 'ga': 'Obolo', 'gd': 'Obolo', 'gl': 'obolo', 'gu': 'ઓબોલો', 'ha': 'Obolo', 'he': 'אובולו', 'hi': 'ओबोलो', 'hi-Latn': 'Obolo', 'hr': 'obolo', 'hsb': 'obološćina', 'ht': 'obolo', 'hu': 'obolo', 'hy': 'օբոլո', 'ia': 'obolo', 'id': 'Obolo', 'ig': 'Obolo', 'is': 'obolo', 'it': 'obolo', 'ja': 'オボロ語', 'jv': 'Obolo', 'ka': 'ობოლო', 'kk': 'оболо тілі', 'kk-Cyrl': 'оболо тілі', 'km': 'អូបូឡូ', 'kn': 'ಒಬೊಲೊ', 'ko': '오볼로어', 'kok': 'ओबोलो', 'kok-Deva': 'ओबोलो', 'ku': 'obolo', 'ky': 'оболочо (Нигерия)', 'lo': 'ໂອໂບໂລ', 'lt': 'obolų', 'lv': 'obolo', 'mai': 'ओबोलो', 'mi': 'Ōporo', 'mk': 'оболо', 'ml': 'ഒബോളോ', 'mn': 'оболо', 'mr': 'ओबोलो', 'ms': 'Obolo', 'my': 'အိုဘိုလို', 'nb': 'obolo', 'ne': 'ओबोलो', 'nl': 'Obolo', 'nn': 'obolo', 'no': 'obolo', 'or': 'ଅବୋଲା', 'pa': 'ਓਬੋਲੋ', 'pa-Guru': 'ਓਬੋਲੋ', 'pcm': 'Óbóló Lángwej', 'pl': 'obolo', 'ps': 'Obo', 'pt': 'obolo', 'qu': 'Obolo Simi', 'ro': 'obolo', 'ru': 'оболо', 'sc': 'obolo', 'sd': 'اوبولو', 'sd-Arab': 'اوبولو', 'si': 'ඔබලෝ', 'sk': 'obolo', 'sl': 'obolo', 'so': 'Obolo', 'sq': 'oboloisht', 'sr': 'оболо', 'sr-Cyrl': 'оболо', 'sr-Latn': 'obolo', 'sv': 'obolo', 'sw': 'Kiobolo', 'ta': 'ஒபோலோ', 'te': 'ఒబోలో', 'th': 'โอโบโล', 'ti': 'ኦቦሎ', 'tk': 'obolo dili', 'to': 'lea fakaʻopolo', 'tr': 'Obolo dili', 'uk': 'оболо', 'ur': 'اوبولو', 'uz': 'obolo', 'uz-Latn': 'obolo', 'vec': 'obolo', 'vi': 'Tiếng Obolo', 'yo': 'Èdè Obolo', 'yue': '奧博洛文', 'yue-Hans': '奥博洛文', 'yue-Hant': '奧博洛文', 'zh': '奥博洛语', 'zh-Hans': '奥博洛语', 'zh-Hant': '奧博洛語', 'zu': 'isi-Obolo'},
    'ano': {'en': 'Andoque'},
    'anp': {'af': 'Angika', 'am': 'አንጊካ', 'ar': 'الأنجيكا', 'as': 'আঙ্গিকা', 'ast': 'angika', 'az': 'angika', 'az-Cyrl': 'анҝика', 'az-Latn': 'angika', 'be': 'ангіка', 'be-tarask': 'ангіка', 'bg': 'ангика', 'bgn': 'انگیکای', 'bn': 'আঙ্গিকা', 'br': 'angika', 'brx': 'आङिका', 'bs': 'angika', 'bs-Cyrl': 'ангика', 'bs-Latn': 'angika', 'ca': 'angika', 'ca-ES-valencia': 'angika', 'ccp': '𑄃𑄋𑄳𑄉𑄨𑄇', 'ce': 'ангика', 'chr': 'ᎠᎾᎩᎧ', 'ckb': 'ئەنگیکا', 'cs': 'angika', 'cy': 'Angika', 'da': 'angika', 'de': 'Angika', 'dsb': 'angikašćina', 'el': 'Ανγκικά', 'el-polyton': 'Ανγκικά', 'en': 'Angika', 'es': 'angika', 'et': 'angika', 'eu': 'angikera', 'fa': 'آنگیکا', 'ff-Adlm': '𞤀𞤲𞤺𞤭𞤳𞤢𞥄𞤪𞤫', 'fi': 'angika', 'fil': 'Angika', 'fo': 'angika', 'fr': 'angika', 'fy': 'Angika', 'ga': 'Aingícis', 'gd': 'Angika', 'gl': 'angika', 'gsw': 'Angika', 'gu': 'અંગીકા', 'ha': 'Angika', 'he': 'אנג׳יקה', 'hi': 'अंगिका', 'hi-Latn': 'Angika', 'hr': 'angika', 'hsb': 'angikašćina', 'ht': 'angika', 'hu': 'angika', 'hy': 'անգիկա', 'ia': 'angika', 'id': 'Angika', 'ig': 'Angika', 'is': 'angíka', 'it': 'angika', 'ja': 'アンギカ語', 'jv': 'Angika', 'ka': 'ანგიკა', 'kgp': 'ỹgika', 'kk': 'ангика тілі', 'kk-Cyrl': 'ангика тілі', 'km': 'អាហ្គីកា', 'kn': 'ಆಂಗಿಕಾ', 'ko': '앙가어', 'kok': 'अंगिका', 'kok-Deva': 'अंगिका', 'ks': 'اَنگِکا', 'ks-Arab': 'اَنگِکا', 'ksh': 'Anjikahnesch', 'ku': 'angîkayî', 'ky': 'ангикача', 'lb': 'Angika', 'lo': 'ແອນຈີກາ', 'lt': 'angikų', 'lv': 'angika', 'mai': 'अंगिका', 'mi': 'Anahika', 'mk': 'ангика', 'ml': 'ആൻഗിക', 'mn': 'ангик', 'mr': 'अंगिका', 'ms': 'Angika', 'mt': 'Angika', 'my': 'အန်ဂီကာ', 'nb': 'angika', 'ne': 'अङ्गिका', 'nl': 'Angika', 'nn': 'angika', 'no': 'angika', 'or': 'ଅଁଗୀକା', 'pa': 'ਅੰਗਿਕਾ', 'pa-Guru': 'ਅੰਗਿਕਾ', 'pcm': 'Angíka', 'pl': 'angika', 'ps': 'انگيکي', 'pt': 'angika', 'qu': 'Angika', 'rm': 'angika', 'ro': 'angika', 'ru': 'ангика', 'sa': 'अङ्गिक', 'sat': 'ᱟᱹᱝᱜᱤᱠᱟ', 'sat-Olck': 'ᱟᱹᱝᱜᱤᱠᱟ', 'sc': 'angika', 'sd': 'انجيڪا', 'sd-Arab': 'انجيڪا', 'si': 'අන්ගික', 'sk': 'angika', 'sl': 'angikaščina', 'smn': 'angika', 'so': 'U dhashay Anp', 'sq': 'angikisht', 'sr': 'ангика', 'sr-Cyrl': 'ангика', 'sr-Latn': 'angika', 'sv': 'angika', 'sw': 'Kiangika', 'ta': 'அங்கிகா', 'te': 'ఆంగిక', 'th': 'อังคิกา', 'ti': 'ኣንጂካ', 'tk': 'angika dili', 'to': 'lea fakaʻangika', 'tr': 'Angika', 'ug': 'ئانگىكاچە', 'uk': 'ангіка', 'ur': 'انگیکا', 'uz': 'angika', 'uz-Cyrl': 'ангика', 'uz-Latn': 'angika', 'vec': 'angika', 'vi': 'Tiếng Angika', 'xnr': 'अंगिका', 'yo': 'Èdè Angika', 'yrl': 'ãyika', 'yue': '昂加文', 'yue-Hans': '昂加文', 'yue-Hant': '昂加文', 'zh': '昂加语', 'zh-Hans': '昂加语', 'zh-Hant': '昂加文', 'zu': 'isi-Angika'},
    'anq': {'en': 'Jarawa (India)'},
    'anr': {'en': 'Andh'},
    'ans': {'en': 'Anserma'},
    'ant': {'en': 'Antakarinya'},
    'anu': {'en': 'Anuak'},
    'anv': {'en': 'Denya'},
    'anw': {'en': 'Anaang'},
    'anx': {'en': 'Andra-Hus'},
    'any': {'en': 'Anyin'},
    'anz': {'en': 'Anem'},
    'aoa': {'en': 'Angolar'},
    'aob': {'en': 'Abom'},
    'aoc': {'en': 'Pemon'},
    'aod': {'en': 'Andarum'},
    'aoe': {'en': 'Angal Enen'},
    'aof': {'en': 'Bragat'},
    'aog': {'en': 'Angoram'},
    'aoh': {'en': 'Arma'},
    'aoi': {'en': 'Anindilyakwa'},
    'aoj': {'en': 'Mufian'},
    'aok': {'en': 'Arhö'},
    'aol': {'en': 'Alor'},
    'aom': {'en': 'Ömie'},
    'aon': {'en': 'Bumbita Arapesh'},
    'aor': {'en': 'Aore'},
    'aos': {'en': 'Taikat'},
    'aot': {'en': 'Atong (India)'},
    'aou': {'en': "A'ou"},
    'aox': {'en': 'Atorada'},
    'aoz': {'en': 'Uab Meto'},
    'apa': {'en': 'Apache languages'},
    'apa-pro': {'en': 'Proto-Apachean'},
    'apb': {'en': "Sa'a"},
    'apc': {'apc': 'العامية', 'bal-Latn': 'Latwiái Arabi', 'et': 'Levandi araabia', 'fi': 'urduni', 'gd': 'Arabais Levantach', 'hu': 'levantei arab', 'ia': 'arabe levantin', 'ig': 'apcc', 'ku': 'erebîya bakurê şamê', 'nl': 'Levantijns-Arabisch', 'sc': 'àrabu levantinu', 'ti': 'ሌቫንቲናዊ ዓረብኛ', 'vec': 'àrabo levantin', 'en': 'North Levantine Arabic'},
    'apd': {'en': 'Sudanese Arabic'},
    'ape': {'en': 'Bukiyip'},
    'apf': {'en': 'Pahanan Agta'},
    'apg': {'en': 'Ampanang'},
    'aph': {'en': 'Athpariya'},
    'api': {'en': 'Apiaká'},
    'apj': {'en': 'Jicarilla Apache'},
    'apk': {'en': 'Kiowa Apache'},
    'apl': {'en': 'Lipan Apache'},
    'apm': {'en': 'Mescalero-Chiricahua Apache'},
    'apn': {'en': 'Apinayé'},
    'apo': {'en': 'Ambul'},
    'app': {'en': 'Apma'},
    'apq': {'en': 'A-Pucikwar'},
    'apr': {'en': 'Arop-Lokep'},
    'aps': {'en': 'Arop-Sissano'},
    'apt': {'en': 'Apatani'},
    'apu': {'en': 'Apurinã'},
    'apv': {'en': 'Alapmunte'},
    'apw': {'en': 'Western Apache'},
    'apx': {'en': 'Aputai'},
    'apy': {'en': 'Apalaí'},
    'apz': {'en': 'Safeyoka'},
    'aqa': {'en': 'Alacalufan languages'},
    'aqc': {'en': 'Archi'},
    'aqd': {'en': 'Ampari Dogon'},
    'aqg': {'en': 'Arigidi'},
    'aqk': {'en': 'Aninka'},
    'aql': {'en': 'Algic languages'},
    'aql-pro': {'en': 'Proto-Algic'},
    'aqm': {'en': 'Atohwaim'},
    'aqn': {'en': 'Northern Alta'},
    'aqp': {'en': 'Atakapa'},
    'aqr': {'en': 'Arhâ'},
    'aqt': {'en': 'Angaité'},
    'aqz': {'en': 'Akuntsu'},
    'ar': {'af': 'Arabies', 'agq': 'Àlabì', 'ak': 'Arabeke', 'am': 'ዓረብኛ', 'ar': 'العربية', 'as': 'আৰবী', 'asa': 'Kiarabu', 'ast': 'árabe', 'az': 'ərəb', 'az-Arab': 'عربجه', 'az-Cyrl': 'әрәб', 'az-Latn': 'ərəb', 'bal-Latn': 'Arabi', 'bas': 'Hɔp u arâb', 'be': 'арабская', 'be-tarask': 'арабская', 'bem': 'Ichi Arab', 'bez': 'Hiharabu', 'bg': 'арабски', 'bgn': 'عربی', 'blo': 'gɩlaaribuja', 'bm': 'larabukan', 'bn': 'আরবি', 'br': 'arabeg', 'brx': 'आराबिक', 'bs': 'arapski', 'bs-Cyrl': 'арапски', 'bs-Latn': 'arapski', 'ca': 'àrab', 'ca-ES-valencia': 'àrab', 'ccp': '𑄃𑄢𑄧𑄝𑄩', 'ce': 'Ӏаьрбийн', 'ceb': 'Arabic', 'cgg': 'Oruharabu', 'chr': 'ᎡᎳᏈ', 'ckb': 'عەرەبی', 'cs': 'arabština', 'cv': 'арап', 'cy': 'Arabeg', 'da': 'arabisk', 'dav': 'Kiarabu', 'de': 'Arabisch', 'dje': 'Laaraw senni', 'dsb': 'arabšćina', 'dyo': 'arab', 'dz': 'ཨེ་ར་བིཀ་ཁ', 'ebu': 'Kĩarabu', 'ee': 'Arabiagbe', 'el': 'Αραβικά', 'el-polyton': 'Ἀραβικά', 'en': 'Arabic', 'en-Dsrt': '𐐇𐑉𐐲𐐺𐐮𐐿', 'eo': 'araba', 'es': 'árabe', 'et': 'araabia', 'eu': 'arabiera', 'ewo': 'Ǹkɔ́bɔ arábia', 'fa': 'عربی', 'ff': 'Aarabeere', 'ff-Adlm': '𞤀𞥄𞤪𞤢𞤦𞤫𞥅𞤪𞤫', 'ff-Latn': 'Aarabeere', 'fi': 'arabia', 'fil': 'Arabic', 'fo': 'arabiskt', 'fr': 'arabe', 'fur': 'arap', 'fy': 'Arabysk', 'ga': 'Araibis', 'gd': 'Arabais', 'gl': 'árabe', 'gsw': 'Arabisch', 'gu': 'અરબી', 'guz': 'Kiarabu', 'ha': 'Larabci', 'haw': 'ʻAlapia', 'he': 'ערבית', 'hi': 'अरबी', 'hi-Latn': 'Arabic', 'hnj': '𞄤𞄣', 'hnj-Hmnp': '𞄤𞄣', 'hr': 'arapski', 'hsb': 'arabšćina', 'ht': 'arabe', 'hu': 'arab', 'hy': 'արաբերեն', 'ia': 'arabe', 'id': 'Arab', 'ie': 'arabic', 'ig': 'Arabiikị', 'ii': 'ꀊꇁꀨꉙ', 'is': 'arabíska', 'it': 'arabo', 'ja': 'アラビア語', 'jgo': 'Alâbɛ', 'jmc': 'Kyiarabu', 'jv': 'Arab', 'ka': 'არაბული', 'kab': 'Taɛrabt', 'kam': 'Kiarabu', 'kde': 'Chalabu', 'kea': 'árabi', 'kgp': 'arame', 'khq': 'Laaraw senni', 'ki': 'Kĩarabu', 'kk': 'араб тілі', 'kk-Cyrl': 'араб тілі', 'kln': 'kutitab Arabuk', 'km': 'អារ៉ាប់', 'kn': 'ಅರೇಬಿಕ್', 'ko': '아랍어', 'kok': 'अरेबिक', 'kok-Deva': 'अरेबिक', 'kok-Latn': 'Arbi', 'ks': 'عربی', 'ks-Arab': 'عربی', 'ksb': 'Kialabu', 'ksf': 'riarab', 'ksh': 'Arahbesch', 'ku': 'erebî', 'kxv': 'arabic', 'kxv-Deva': 'आरॉबिक', 'kxv-Latn': 'arabic', 'kxv-Orya': 'ଆରବିକ', 'kxv-Telu': 'ఆరబిక్', 'ky': 'арабча', 'lag': 'Kɨaráabu', 'lb': 'Arabesch', 'lg': 'Luwarabu', 'lkt': 'Arab Iyápi', 'lld': 'arabich', 'ln': 'lialabo', 'lo': 'ອາຣັບ', 'lrc': 'أرأڤی', 'lt': 'arabų', 'lu': 'Arabi', 'luo': 'Kiarabu', 'luy': 'Kiarabu', 'lv': 'arābu', 'mai': 'अरबी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lmarabu', 'mer': 'Kĩarabu', 'mfe': 'arab', 'mg': 'Arabo', 'mgh': 'Iarabu', 'mi': 'Ārapi', 'mk': 'арапски', 'ml': 'അറബിക്', 'mn': 'араб', 'mni': 'আরাবিক', 'mni-Beng': 'আরাবিক', 'mr': 'अरबी', 'ms': 'Arab', 'mt': 'Għarbi', 'mua': 'arabiya', 'my': 'အာရဗီ', 'mzn': 'عربی', 'naq': 'Arabiǁî gowab', 'nb': 'arabisk', 'nd': 'isi-Alabhu', 'ne': 'अरबी', 'nl': 'Arabisch', 'nmg': 'Kiɛl b’árabe', 'nn': 'arabisk', 'no': 'arabisk', 'nqo': 'ߊߙߓߎߞߊ߲', 'nus': 'Thok Jalabni', 'nyn': 'Oruharabu', 'om': 'Arabiffaa', 'or': 'ଆରବିକ୍', 'os': 'араббаг', 'pa': 'ਅਰਬੀ', 'pa-Guru': 'ਅਰਬੀ', 'pcm': 'Arábík Lángwej', 'pl': 'arabski', 'ps': 'عربي', 'pt': 'árabe', 'qu': 'Arabe Simi', 'rm': 'arab', 'rn': 'Icarabu', 'ro': 'arabă', 'rof': 'Kiarabu', 'ru': 'арабский', 'rw': 'Icyarabu', 'rwk': 'Kyiarabu', 'sah': 'Араабтыы', 'saq': 'Kiarabu', 'sat': 'ᱚᱨᱵᱤ', 'sat-Olck': 'ᱚᱨᱵᱤ', 'sbp': 'Ishiyalabu', 'sc': 'àrabu', 'sd': 'عربي', 'sd-Arab': 'عربي', 'se': 'arábagiella', 'seh': 'árabe', 'ses': 'Laaraw senni', 'sg': 'Arâbo', 'shi': 'ⵜⴰⵄⵔⴰⴱⵜ', 'shi-Latn': 'Taɛrabt', 'shi-Tfng': 'ⵜⴰⵄⵔⴰⴱⵜ', 'si': 'අරාබි', 'sk': 'arabčina', 'sl': 'arabščina', 'smn': 'arabiakielâ', 'sn': 'chiArabu', 'so': 'Carabi', 'sq': 'arabisht', 'sr': 'арапски', 'sr-Cyrl': 'арапски', 'sr-Latn': 'arapski', 'sv': 'arabiska', 'sw': 'Kiarabu', 'syr': 'ܥܪܒܝܬ', 'ta': 'அரபிக்', 'te': 'అరబిక్', 'teo': 'Kiarabu', 'tg': 'арабӣ', 'th': 'อาหรับ', 'ti': 'ዓረብኛ', 'tk': 'arap dili', 'to': 'lea fakaʻalepea', 'tr': 'Arapça', 'tt': 'гарәп', 'twq': 'Laaraw senni', 'tzm': 'Taεrabt', 'ug': 'ئەرەبچە', 'uk': 'арабська', 'ur': 'عربی', 'uz': 'arab', 'uz-Cyrl': 'арабча', 'uz-Latn': 'arab', 'vai': 'ꕞꕌꖝ', 'vai-Latn': 'Lahabu', 'vai-Vaii': 'ꕞꕌꖝ', 'vec': 'àrabo', 'vi': 'Tiếng Ả Rập', 'vun': 'Kyiarabu', 'wae': 'Arabiš', 'wo': 'Arabic', 'xh': 'Isi-Arabhu', 'xnr': 'अरबी', 'xog': 'Oluwarabu', 'yav': '́pakas', 'yi': 'אַראַביש', 'yo': 'Èdè Lárúbáwá', 'yrl': 'arabi', 'yue': '阿拉伯文', 'yue-Hans': '阿拉伯文', 'yue-Hant': '阿拉伯文', 'zgh': 'ⵜⴰⵄⵔⴰⴱⵜ', 'zh': '阿拉伯语', 'zh-Hans': '阿拉伯语', 'zh-Hant': '阿拉伯文', 'zu': 'isi-Arabic'},
    'ar-001': {'af': 'Moderne Standaardarabies', 'ak': 'Arabeke Kasa Nhyehyɛeɛ Foforɔ', 'am': 'ዘመናዊ መደበኛ ዓረብኛ', 'ar': 'العربية الفصحى الحديثة', 'as': 'আধুনিক মানক আৰবী', 'ast': 'árabe estándar modernu', 'az': 'müasir standart ərəb', 'az-Arab': 'استاندارد عربجه', 'az-Cyrl': 'мүасир стандарт әрәб', 'az-Latn': 'müasir standart ərəb', 'bal-Latn': 'Arabi (Donyá)', 'be': 'сучасная стандартная арабская', 'be-tarask': 'сучасная стандартная арабская', 'bg': 'съвременен стандартен арабски', 'bgn': 'پیشرفته\u200cاین عربی', 'blo': 'gɩlaaribuja ŋgɩɖee kǝ ba na fʊba na', 'bn': 'আধুনিক আদর্শ আরবি', 'br': 'arabeg modern', 'brx': 'गोदान मानथाखोआरि आराबिक', 'bs': 'moderni standardni arapski', 'bs-Cyrl': 'арапски (стандардни)', 'bs-Latn': 'moderni standardni arapski', 'ca': 'àrab estàndard modern', 'ca-ES-valencia': 'àrab estàndard modern', 'ccp': '𑄚𑄱 𑄉𑄧𑄟𑄴 𑄃𑄢𑄧𑄝𑄩', 'ce': 'ХӀинца болу стандартан Ӏаьрбийн', 'ceb': 'Modernong Standard nga Arabic', 'chr': 'ᎪᎯᏊ ᎢᎬᏥᎩ ᎠᏟᎶᏍᏗ ᎡᎳᏈ', 'ckb': 'عەرەبیی ستاندارد', 'cs': 'arabština (moderní standardní)', 'cv': 'арап литератури', 'cy': 'Arabeg Modern Safonol', 'da': 'moderne standardarabisk', 'de': 'Modernes Hocharabisch', 'dsb': 'moderna wusokoarabšćina', 'el': 'Σύγχρονα Τυπικά Αραβικά', 'el-polyton': 'Σύγχρονα Τυπικά Αραβικά', 'en': 'Modern Standard Arabic', 'es': 'árabe estándar moderno', 'et': 'tänapäeva araabia kirjakeel', 'eu': 'arabiera moderno estandarra', 'fa': 'عربی رسمی', 'ff-Adlm': '𞤀𞥄𞤪𞤢𞤦𞤫𞥅𞤪𞤫 𞤊𞤵𞤧𞤸𞤢 𞤒𞤫𞤲𞤯𞤵𞤳𞤢', 'fi': 'yleisarabia', 'fil': 'Modernong Karaniwang Arabic', 'fo': 'nútíðar vanligt arabiskt', 'fr': 'arabe standard moderne', 'fy': 'Modern standert Arabysk', 'ga': 'Araibis Chaighdeánach', 'gd': 'Nuadh-Arabais Stannardach', 'gl': 'árabe estándar moderno', 'gu': 'મોડર્ન સ્ટાન્ડર્ડ અરબી', 'ha': 'Larabci Asali Na Zamani', 'he': 'ערבית ספרותית', 'hi': 'आधुनिक मानक अरबी', 'hi-Latn': 'Modern Standard Arabic', 'hr': 'moderni standardni arapski', 'hsb': 'moderna wysokoarabšćina', 'ht': 'arabe standard moderne', 'hu': 'modern szabványos arab', 'hy': 'արդի ընդհանուր արաբերեն', 'ia': 'arabe standard moderne', 'id': 'Arab Standar Modern', 'ig': 'Ụdị Arabiikị nke oge a', 'ii': 'ꀊꇁꀨꉙ（ꋧꃅ）', 'is': 'stöðluð nútímaarabíska', 'it': 'arabo moderno standard', 'ja': '現代標準アラビア語', 'jv': 'Arab Standar Anyar', 'ka': 'თანამედროვე სტანდარტული არაბული', 'kea': 'árabi mudernu', 'kgp': 'arame ta’ũn', 'kk': 'қазіргі стандартты араб тілі', 'kk-Cyrl': 'қазіргі стандартты араб тілі', 'km': 'អារ៉ាប់ស្តង់ដារទំនើប', 'kn': 'ಆಧುನಿಕ ಪ್ರಮಾಣಿತ ಅರೇಬಿಕ್', 'ko': '현대 표준 아랍어', 'kok': 'आधुनिक प्रमाणित अरेबिक', 'kok-Deva': 'आधुनिक प्रमाणित अरेबिक', 'ks': 'ماڈرن معیٲری عربی', 'ks-Arab': 'ماڈرن معیٲری عربی', 'ksh': 'Schtandatt Arahbesch', 'ku': 'erebîya modern a standard', 'kxv': 'punijuga mānānka arabic', 'kxv-Deva': 'पुनिजुगो मानांकॉ आरॉबिक', 'kxv-Latn': 'punijuga mānānka arabic', 'kxv-Orya': 'ପୁନି ଜୁଗ ମାନାଙ୍କ ଆରବିକ୍', 'kxv-Telu': 'అదునిక ప్రామాణిక్ అరబిక్', 'ky': 'азыркы адабий араб тилинде', 'lb': 'Modernt Héicharabesch', 'lld': 'arabich modern standard', 'lo': 'ອາຣາບິກມາດຕະຖານສະໄໝໃໝ່', 'lrc': 'عروی مدرن', 'lt': 'šiuolaikinė standartinė arabų', 'lv': 'mūsdienu standarta arābu', 'mai': 'आधुनिक मानक अरबी', 'mi': 'Ārapi Moroki', 'mk': 'современ стандардизиран арапски', 'ml': 'ആധുനിക സ്റ്റാൻഡേർഡ് അറബിക്', 'mn': 'стандарт араб', 'mni': 'মোর্দর্ন স্তেন্দর্দ আরাবিক', 'mni-Beng': 'মোর্দর্ন স্তেন্দর্দ আরাবিক', 'mr': 'आधुनिक प्रमाणित अरबी', 'ms': 'Arab Standard Moden', 'mt': 'Għarbi Standard Modern', 'my': 'ခေတ်သစ် ရှေ့ဆောင် အာရဗီ', 'mzn': 'مدرن استاندارد عربی', 'nb': 'moderne standardarabisk', 'ne': 'आधुनिक मानक अरबी', 'nl': 'modern standaard Arabisch', 'nn': 'moderne standardarabisk', 'no': 'moderne standardarabisk', 'nqo': 'ߊߙߊߓߎߞߊ߲ ߘߐߞߣߍߣߍ߲', 'om': 'Arabiffa Istaandaardii Ammayyaa', 'or': 'ଆଧୁନିକ ମାନକ ଆରବିକ୍\u200c', 'pa': 'ਆਧੁਨਿਕ ਮਿਆਰੀ ਅਰਬੀ', 'pa-Guru': 'ਆਧੁਨਿਕ ਮਿਆਰੀ ਅਰਬੀ', 'pl': 'współczesny arabski', 'ps': 'نوې معياري عربي', 'pt': 'árabe moderno', 'qu': 'Musuq Estandar Arabe Simi', 'rm': 'arab modern standardisà', 'ro': 'arabă standard modernă', 'ru': 'арабский литературный', 'sat': 'ᱟᱫᱷᱩᱱᱤᱠ ᱢᱟᱱᱚᱠ ᱚᱨᱵᱤ', 'sat-Olck': 'ᱟᱫᱷᱩᱱᱤᱠ ᱢᱟᱱᱚᱠ ᱚᱨᱵᱤ', 'sc': 'àrabu modernu istandard', 'sd': 'جديد معياري عربي', 'sd-Arab': 'جديد معياري عربي', 'si': 'නූතන සම්මත අරාබි', 'sk': 'arabčina (moderná štandardná)', 'sl': 'sodobna standardna arabščina', 'smn': 'standard arabiakielâ', 'so': 'Carabiga rasmiga ah', 'sq': 'arabishte standarde moderne', 'sr': 'савремени стандардни арапски', 'sr-Cyrl': 'савремени стандардни арапски', 'sr-Latn': 'savremeni standardni arapski', 'sv': 'modern standardarabiska', 'sw': 'Kiarabu sanifu', 'syr': 'ܥܪܒܝܬ ܪܘܫܡܝܐ ܚܕܬܐ', 'ta': 'நவீன நிலையான அரபிக்', 'te': 'ఆధునిక ప్రామాణిక అరబిక్', 'tg': 'Стандарти муосири арабӣ', 'th': 'อาหรับมาตรฐานสมัยใหม่', 'ti': 'ዘመናዊ ምዱብ ዓረብኛ', 'tk': 'häzirki zaman standart arap dili', 'to': 'lea fakaʻalepea (māmani)', 'tr': 'Modern Standart Arapça', 'tt': 'Заманча стандарт гарәп', 'ug': 'ھازىرقى زامان ئۆلچەملىك ئەرەبچە', 'uk': 'сучасна стандартна арабська', 'ur': 'ماڈرن اسٹینڈرڈ عربی', 'uz': 'standart arab', 'uz-Cyrl': 'стандарт арабча', 'uz-Latn': 'standart arab', 'vec': 'àrabo moderno standard', 'vi': 'Tiếng Ả Rập Hiện đại', 'wo': 'Araab', 'xh': 'Isi-Arabhu (Sale mihla)', 'xnr': 'आधुनिक मानक अरबी', 'yo': 'Èdè Lárúbáwá (Agbáyé)', 'yrl': 'arabi mudernu', 'yue': '現代標準阿拉伯文', 'yue-Hans': '现代标准阿拉伯文', 'yue-Hant': '現代標準阿拉伯文', 'zh': '现代标准阿拉伯语', 'zh-Hans': '现代标准阿拉伯语', 'zh-Hant': '現代標準阿拉伯文', 'zu': 'Isi-Arabic Esivamile Sesimanje'},
    'arb': {'en': 'Standard Arabic'},
    'arc': {'af': 'Aramees', 'am': 'አራማይክ', 'ar': 'الآرامية', 'ast': 'araméu', 'az': 'aramik', 'az-Latn': 'aramik', 'be': 'арамейская', 'be-tarask': 'арамейская', 'bg': 'арамейски', 'bn': 'আরামাইক', 'br': 'arameeg', 'brx': 'अरामाईक', 'bs': 'aramejski', 'bs-Cyrl': 'армајски', 'bs-Latn': 'aramejski', 'ca': 'arameu', 'ca-ES-valencia': 'arameu', 'ccp': '𑄃𑄢𑄟𑄳𑄆𑄇𑄴', 'cs': 'aramejština', 'cy': 'Aramaeg', 'da': 'aramæisk', 'de': 'Aramäisch', 'el': 'Αραμαϊκά', 'el-polyton': 'Ἀραμαϊκά', 'en': 'Aramaic', 'es': 'arameo', 'et': 'aramea', 'fa': 'آرامی', 'fi': 'valtakunnanaramea', 'fr': 'araméen', 'fur': 'aramaic', 'fy': 'Arameesk', 'ga': 'Aramais', 'gd': 'Aramais', 'gl': 'arameo', 'gsw': 'Aramääisch', 'gu': 'એરમૈક', 'he': 'ארמית', 'hi': 'ऐरेमेक', 'hi-Latn': 'Aramaic', 'hr': 'aramejski', 'ht': 'araméen', 'hu': 'arámi', 'hy': 'արամեերեն', 'id': 'Aram', 'is': 'arameíska', 'it': 'aramaico', 'ja': 'アラム語', 'ka': 'არამეული', 'kgp': 'aramajiko', 'kn': 'ಅರಾಮಿಕ್', 'ko': '아람어', 'ks': 'اَرَمیک', 'ks-Arab': 'اَرَمیک', 'lb': 'Aramäesch', 'lo': 'ອາລາມິກ', 'lt': 'aramaikų', 'lv': 'aramiešu', 'mk': 'арамејски', 'ml': 'അരമായ', 'mr': 'अ\u200dॅरेमाइक', 'mt': 'Aramajk', 'nb': 'arameisk', 'ne': 'अरामाइक', 'nl': 'Aramees', 'nn': 'arameisk', 'no': 'arameisk', 'or': 'ଆରାମାଇକ୍', 'pl': 'aramejski', 'pt': 'aramaico', 'rm': 'arameic', 'ro': 'aramaică', 'ru': 'арамейский', 'sk': 'aramejčina', 'sl': 'aramejščina', 'sr': 'арамејски', 'sr-Cyrl': 'арамејски', 'sr-Latn': 'aramejski', 'sv': 'arameiska', 'sw': 'Kiaramu', 'ta': 'அராமைக்', 'te': 'అరామైక్', 'th': 'อราเมอิก', 'to': 'lea fakaʻalāmiti', 'tr': 'Aramice', 'ug': 'ئارامۇچە', 'uk': 'арамейська', 'vi': 'Tiếng Aramaic', 'yi': 'אַראַמיש', 'yrl': 'aramaiku', 'yue': '阿拉米文', 'yue-Hans': '阿拉米文', 'yue-Hant': '阿拉米文', 'zh': '阿拉米语', 'zh-Hans': '阿拉米语', 'zh-Hant': '阿拉米文'},
    'ard': {'en': 'Arabana'},
    'are': {'en': 'Western Arrarnta'},
    'arh': {'en': 'Arhuaco'},
    'ari': {'en': 'Arikara'},
    'arj': {'en': 'Arapaso'},
    'ark': {'en': 'Arikapú'},
    'arl': {'en': 'Arabela'},
    'arn': {'af': 'Mapuche', 'am': 'ማፑቼ', 'ar': 'المابودونغونية', 'arn': 'Mapudungun', 'as': 'মাপুচে', 'ast': 'mapuche', 'az': 'mapuçe', 'az-Cyrl': 'арауканҹа', 'az-Latn': 'mapuçe', 'bal-Latn': 'Mapuche', 'be': 'мапудунгун', 'be-tarask': 'мапудунгун', 'bg': 'мапуче', 'bgn': 'ماپوچه\u200cای', 'bn': 'মাপুচে', 'br': 'araoukaneg', 'brx': 'मापुचे', 'bs': 'mapuški', 'bs-Cyrl': 'ароканијски', 'bs-Latn': 'mapuški', 'ca': 'mapudungu', 'ca-ES-valencia': 'mapudungu', 'ccp': '𑄟𑄛𑄪𑄌𑄨', 'ce': 'арауканхойн', 'chr': 'ᎹᏊᏤ', 'ckb': 'ماپووچە', 'cs': 'mapudungun', 'cy': 'Arawcaneg', 'da': 'mapudungun', 'de': 'Mapudungun', 'dsb': 'arawkašćina', 'el': 'Αραουκανικά', 'el-polyton': 'Αραουκανικά', 'en': 'Mapuche', 'es': 'mapuche', 'et': 'mapudunguni', 'eu': 'mapudunguna', 'fa': 'ماپوچه\u200cای', 'ff-Adlm': '𞤃𞤢𞤨𞤵𞤷𞤭𞥅𞤪𞤫', 'fi': 'mapudungun', 'fil': 'Mapuche', 'fo': 'mapuche', 'fr': 'mapuche', 'fy': 'Araukaansk', 'ga': 'Mapúitsis', 'gd': 'Mapudungun', 'gl': 'mapuche', 'gsw': 'Araukanisch', 'gu': 'મેપુચે', 'ha': 'Mapuche', 'he': 'אראוקנית', 'hi': 'मापूचे', 'hi-Latn': 'Mapuche', 'hr': 'mapuche', 'hsb': 'arawkanšćina', 'ht': 'mapuche', 'hu': 'mapucse', 'hy': 'մապուչի', 'ia': 'mapuche', 'id': 'Mapuche', 'ig': 'Mapuche', 'is': 'mapuche', 'it': 'mapudungun', 'ja': 'マプチェ語', 'jv': 'Mapushe', 'ka': 'მაპუდუნგუნი', 'kea': 'araukanu', 'kgp': 'mỹpunugũn', 'kk': 'мапуче тілі', 'kk-Cyrl': 'мапуче тілі', 'km': 'ម៉ាពូឈី', 'kn': 'ಮಪುಚೆ', 'ko': '마푸둥군어', 'kok': 'मापुचे', 'kok-Deva': 'मापुचे', 'ks': 'ایرو کونِیَن', 'ks-Arab': 'ایرو کونِیَن', 'ku': 'mapuçî', 'ky': 'мапучече', 'lb': 'Mapudungun', 'lo': 'ມາພຸດຊີ', 'lrc': 'ماپوٙچئ', 'lt': 'mapudungunų', 'lv': 'araukāņu', 'mai': 'मापुचे', 'mi': 'Mapūte', 'mk': 'мапучки', 'ml': 'മാപുചി', 'mn': 'мапүчи', 'mr': 'मापुची', 'ms': 'Mapuche', 'mt': 'Mapuche', 'my': 'မာပုချီ', 'mzn': 'ماپوچه', 'nb': 'mapudungun', 'ne': 'मापुचे', 'nl': 'Mapudungun', 'nn': 'mapudungun', 'no': 'mapudungun', 'or': 'ମାପୁଚେ', 'pa': 'ਮਾਪੁਚੇ', 'pa-Guru': 'ਮਾਪੁਚੇ', 'pcm': 'Mapúchẹ́ Lángwej', 'pl': 'mapudungun', 'ps': 'ماپوچه', 'pt': 'mapudungun', 'qu': 'Mapuche Simi', 'rm': 'araucanic', 'ro': 'mapuche', 'ru': 'мапуче', 'sc': 'mapudungun', 'sd': 'ماپوچي', 'sd-Arab': 'ماپوچي', 'si': 'මපුචෙ', 'sk': 'mapudungun', 'sl': 'mapudungunščina', 'smn': 'mapudungun', 'so': 'Mapuche', 'sq': 'mapuçisht', 'sr': 'мапуче', 'sr-Cyrl': 'мапуче', 'sr-Latn': 'mapuče', 'sv': 'mapudungun', 'sw': 'Kimapuche', 'ta': 'மபுச்சே', 'te': 'మపుచే', 'th': 'มาปูเช', 'ti': 'ማፑቺ', 'tk': 'mapuçe dili', 'to': 'lea fakamapuse', 'tr': 'Mapuçe dili', 'tt': 'мапуче', 'ug': 'ماپۇدۇنگۇنچە', 'uk': 'арауканська', 'ur': 'ماپوچے', 'uz': 'mapuche', 'uz-Cyrl': 'мапудунгун', 'uz-Latn': 'mapuche', 'vec': 'mapudungun', 'vi': 'Tiếng Mapuche', 'xnr': 'मापूचे', 'yo': 'Èdè Mapushe', 'yrl': 'mapudũgũ', 'yue': '馬普切文', 'yue-Hans': '马普切文', 'yue-Hant': '馬普切文', 'zh': '马普切语', 'zh-Hans': '马普切语', 'zh-Hant': '馬普切文', 'zu': 'isi-Mapuche'},
    'aro': {'am': 'አራኦና', 'ast': 'araona', 'br': 'araona', 'ca': 'araona', 'ca-ES-valencia': 'araona', 'cs': 'araonština', 'cy': 'Araonaeg', 'de': 'Araona', 'en': 'Araona', 'et': 'araona', 'fi': 'araona', 'fr': 'araona', 'gd': 'Araona', 'hi-Latn': 'Araona', 'ht': 'araona', 'it': 'araona', 'ja': 'アラオナ語', 'lb': 'Araona', 'lt': 'araonų', 'mk': 'араона', 'nb': 'araona', 'ne': 'अराओना', 'nl': 'Araona', 'nn': 'araona', 'no': 'araona', 'pl': 'araona', 'sv': 'araoniska', 'th': 'อาเรานา', 'to': 'lea fakaʻalaona', 'tr': 'Araona', 'uk': 'араона', 'vi': 'Tiếng Araona', 'yue': '阿拉奧納文', 'yue-Hans': '阿拉奥纳文', 'yue-Hant': '阿拉奧納文', 'zh-Hant': '阿拉奧納文'},
    'arp': {'af': 'Arapaho', 'am': 'አራፓሆ', 'ar': 'الأراباهو', 'as': 'আৰাপাহো', 'ast': 'arapaho', 'az': 'arapaho', 'az-Cyrl': 'арапаһо', 'az-Latn': 'arapaho', 'be': 'арапаха', 'be-tarask': 'арапаха', 'bg': 'арапахо', 'bgn': 'اراپاهو', 'bn': 'আরাপাহো', 'br': 'arapaho', 'brx': 'आरापाह’', 'bs': 'arapaho', 'bs-Cyrl': 'арапахо', 'bs-Latn': 'arapaho', 'ca': 'arapaho', 'ca-ES-valencia': 'arapaho', 'ccp': '𑄃𑄢𑄛𑄦𑄮', 'ce': 'арапахо', 'chr': 'ᎠᏩᏈᎰ', 'ckb': 'ئاراپاهۆ', 'cs': 'arapažština', 'cy': 'Arapaho', 'da': 'arapaho', 'de': 'Arapaho', 'dsb': 'arapahošćina', 'el': 'Αραπάχο', 'el-polyton': 'Αραπάχο', 'en': 'Arapaho', 'es': 'arapaho', 'et': 'arapaho', 'eu': 'arapahoera', 'fa': 'آراپاهویی', 'ff-Adlm': '𞤀𞤪𞤢𞤨𞤢𞤸𞤮𞥅𞤪𞤫', 'fi': 'arapaho', 'fil': 'Arapaho', 'fo': 'arapaho', 'fr': 'arapaho', 'fy': 'Arapaho', 'ga': 'Arapachóis', 'gd': 'Arapaho', 'gl': 'arapaho', 'gsw': 'Arapaho', 'gu': 'અરાપાહો', 'ha': 'Arapaho', 'he': 'אראפהו', 'hi': 'अरापाहो', 'hi-Latn': 'Arapaho', 'hr': 'arapaho', 'hsb': 'arapahošćina', 'ht': 'arapaho', 'hu': 'arapaho', 'hy': 'արապահո', 'ia': 'arapaho', 'id': 'Arapaho', 'ig': 'Arapaho', 'is': 'arapahó', 'it': 'arapaho', 'ja': 'アラパホー語', 'jv': 'Arapaho', 'ka': 'არაპაჰო', 'kgp': 'arapaho', 'kk': 'арапахо тілі', 'kk-Cyrl': 'арапахо тілі', 'km': 'អារ៉ាប៉ាហូ', 'kn': 'ಅರಪಾಹೋ', 'ko': '아라파호어', 'kok': 'अरापाहो', 'kok-Deva': 'अरापाहो', 'ks': 'اَراپاہو', 'ks-Arab': 'اَراپاہو', 'ksh': 'Arrappaho', 'ku': 'arapahoyî', 'ky': 'арапахочо', 'lb': 'Arapaho-Sprooch', 'lkt': 'Maȟpíya Tȟó Iyápi', 'lo': 'ອາຣາປາໂຮ', 'lt': 'arapahų', 'lv': 'arapahu', 'mai': 'अरापहो', 'mi': 'Arapaho', 'mk': 'арапахо', 'ml': 'അറാപഹോ', 'mn': 'арапаго', 'mr': 'आरापाहो', 'ms': 'Arapaho', 'mt': 'Arapaho', 'my': 'အာရာပါဟို', 'nb': 'arapaho', 'ne': 'अरापाहो', 'nl': 'Arapaho', 'nn': 'arapaho', 'no': 'arapaho', 'or': 'ଆରାପାହୋ', 'pa': 'ਅਰਾਫਾਓ', 'pa-Guru': 'ਅਰਾਫਾਓ', 'pcm': 'Arapahó', 'pl': 'arapaho', 'ps': 'اراپاهوي', 'pt': 'arapaho', 'qu': 'Arapaho', 'rm': 'arapaho', 'ro': 'arapaho', 'ru': 'арапахо', 'sat': 'ᱟᱨᱟᱯᱟᱦᱚ', 'sat-Olck': 'ᱟᱨᱟᱯᱟᱦᱚ', 'sc': 'arapaho', 'sd': 'اراپائو', 'sd-Arab': 'اراپائو', 'si': 'ඇරපහො', 'sk': 'arapažština', 'sl': 'arapaščina', 'smn': 'arapahokielâ', 'so': 'U dhashay Arap', 'sq': 'arapahoisht', 'sr': 'арапахо', 'sr-Cyrl': 'арапахо', 'sr-Latn': 'arapaho', 'sv': 'arapaho', 'sw': 'Kiarapaho', 'syr': 'ܐܪܐܦܗܝܬ', 'ta': 'அரபஹோ', 'te': 'అరాపాహో', 'th': 'อาราปาโฮ', 'ti': 'ኣራፓሆ', 'tk': 'arapaho dili', 'to': 'lea fakaʻalapaho', 'tr': 'Arapaho dili', 'ug': 'ئاراپاخوچە', 'uk': 'арапахо', 'ur': 'اراپاہو', 'uz': 'arapaxo', 'uz-Cyrl': 'арапахо', 'uz-Latn': 'arapaxo', 'vec': 'arapào', 'vi': 'Tiếng Arapaho', 'xnr': 'अरापाहो', 'yo': 'Èdè Arapaho', 'yrl': 'araparu', 'yue': '阿拉帕霍文', 'yue-Hans': '阿拉帕霍文', 'yue-Hant': '阿拉帕霍文', 'zh': '阿拉帕霍语', 'zh-Hans': '阿拉帕霍语', 'zh-Hant': '阿拉帕霍文', 'zu': 'isi-Arapaho'},
    'arq': {'am': 'የአልጄሪያ ዓረብኛ', 'ast': 'árabe d’Arxelia', 'br': 'arabeg Aljeria', 'cs': 'arabština (alžírská)', 'cy': 'Arabeg Algeria', 'de': 'Algerisches Arabisch', 'en': 'Algerian Arabic', 'et': 'Alžeeria araabia', 'fa': 'عربی الجزایری', 'fi': 'algerianarabia', 'fr': 'arabe algérien', 'gd': 'Arabais Aildireach', 'gu': 'આલ્જેરિયન અરબી', 'hi-Latn': 'Algerian Arabic', 'ht': 'arabe algérien', 'hy': 'ալժիրական արաբերեն', 'id': 'Arab Aljazair', 'it': 'arabo algerino', 'ja': 'アルジェリア・アラビア語', 'ko': '알제리 아랍어', 'ksh': 'Aljehresch Arahbesch', 'lb': 'Algerescht Arabesch', 'lt': 'Alžyro arabų', 'mk': 'алжирски арапски', 'ms': 'Arab Algeria', 'nb': 'algerisk arabisk', 'ne': 'अल्जेरियाली अरबी', 'nl': 'Algerijns Arabisch', 'nn': 'algerisk arabisk', 'no': 'algerisk arabisk', 'pl': 'algierski arabski', 'sv': 'algerisk arabiska', 'sw': 'Kiarabu cha Algeria', 'th': 'อาหรับแอลจีเรีย', 'to': 'lea fakaʻalepea-ʻaisilia', 'tr': 'Cezayir Arapçası', 'uk': 'алжирська арабська', 'vi': 'Tiếng Ả Rập Algeria', 'yue': '阿爾及利亞阿拉伯文', 'yue-Hans': '阿尔及利亚阿拉伯文', 'yue-Hant': '阿爾及利亞阿拉伯文', 'zh-Hant': '阿爾及利亞阿拉伯文'},
    'arr': {'en': 'Karo (Brazil)'},
    'ars': {'af': 'Najdi-Arabies', 'am': 'ናጅዲ አረብኛ', 'ar': 'اللهجة النجدية', 'as': 'নাজডি আৰবী', 'az': 'Nəcd ərəbcəsi', 'az-Latn': 'Nəcd ərəbcəsi', 'be': 'недждыйская арабская', 'be-tarask': 'недждыйская арабская', 'bg': 'найди арабски', 'bn': 'নজদি আরবি', 'br': 'arabeg nadjiek', 'brx': 'नाजडी आराबिक', 'bs': 'najdski arapski', 'bs-Cyrl': 'најди арапски', 'bs-Latn': 'najdski arapski', 'ca': 'àrab najdi', 'ca-ES-valencia': 'àrab najdi', 'chr': 'ᎾᏣᏗ ᎠᎳᏈ', 'cs': 'arabština (Nadžd)', 'cy': 'Arabeg Najdi', 'da': 'Najd-arabisk', 'de': 'Arabisch (Nadschd)', 'dsb': 'najdi arabšćina', 'el': 'Αραβικά Νάτζντι', 'el-polyton': 'Αραβικά Νάτζντι', 'en': 'Najdi Arabic', 'es': 'árabe najdí', 'et': 'Najdi araabia', 'eu': 'Najdeko arabiera', 'fa': 'عربی نجدی', 'ff-Adlm': '𞤀𞥄𞤪𞤢𞤦𞤫𞥅𞤪𞤫 𞤐𞤢𞤶𞤣𞤭', 'fi': 'arabia – najd', 'fil': 'Najdi Arabic', 'fr': 'arabe najdi', 'ga': 'Araibis Najdi', 'gd': 'Arabais Najdi', 'gl': 'árabe najdi', 'gu': 'નજદી અરેબિક', 'ha': 'Larabcin Najdi', 'he': 'ערבית - נג׳ד', 'hi': 'नज्दी अरबी', 'hi-Latn': 'Najdi Arabic', 'hr': 'najdi arapski', 'hsb': 'najdi arabšćina', 'ht': 'arabe najdi', 'hu': 'nedzsdi arab', 'hy': 'նեջդի արաբերեն', 'ia': 'arabe najdi', 'id': 'Arab Najdi', 'ig': 'Najdi Arabikị', 'is': 'najdi-arabíska', 'it': 'arabo najd', 'ja': 'ナジュド地方・アラビア語', 'jv': 'Arab Najdi', 'ka': 'ნაჯის არაბული', 'kgp': 'arame nẽgene', 'kk': 'араб тілі (Неджда)', 'kk-Cyrl': 'араб тілі (Неджда)', 'km': 'អារ៉ាប់ណាឌី', 'kn': 'ನ್ಯಾಜ್ಡಿ ಅರೇಬಿಕ್', 'ko': '아랍어(나즈디)', 'kok': 'नाझदी अरबी', 'kok-Deva': 'नाझदी अरबी', 'ku': 'erebîya necdî', 'ky': 'арабча (нежди диалекти)', 'lo': 'ນາຈດີ ອາຣາບິກ', 'lt': 'arabų najdi', 'lv': 'ņedžu arābu', 'mai': 'नजदी अरबी', 'mi': 'Arapika Nahāri', 'mk': 'неџдиски арапски', 'ml': 'നജ്\u200cദി അറബിക്', 'mn': 'наджи араб', 'mr': 'नजदी अरबी', 'ms': 'Arab Najdi', 'my': 'နာ့ဂျဒီ အာရဗစ်', 'nb': 'najdi-arabisk', 'ne': 'नाज्दी अरविक', 'nl': 'Nadjdi-Arabisch', 'nn': 'najdi-arabisk', 'no': 'najdi-arabisk', 'or': 'ନାଜଦି ଆରବିକ', 'pa': 'ਨਾਜਦੀ ਅਰਬੀ', 'pa-Guru': 'ਨਾਜਦੀ ਅਰਬੀ', 'pcm': 'Nájdí Arábík Lángwej', 'pl': 'arabski nadżdyjski', 'ps': 'نجدی عربی', 'pt': 'árabe négede', 'qu': 'Árabe Najdi Simi', 'ro': 'arabă najdi', 'ru': 'недждийский арабский', 'sc': 'àrabu najdi', 'sd': 'نجدي عربي', 'sd-Arab': 'نجدي عربي', 'si': 'නජ්ඩි අරාබි', 'sk': 'arabčina (nadždská)', 'sl': 'nadždska arabščina', 'so': 'Najdi Arabic', 'sq': 'arabishte naxhdi', 'sr': 'најдиарапски', 'sr-Cyrl': 'најдиарапски', 'sr-Latn': 'najdiarapski', 'sv': 'najdiarabiska', 'sw': 'Kiarabu cha Najdi', 'ta': 'நஜ்தி அரபிக்', 'te': 'నజ్ది అరబిక్', 'th': 'อาหรับนัจญ์ดี', 'ti': 'ናጅዲ ዓረብኛ', 'tk': 'nejdi arap dili', 'to': 'lea fakaʻalepea-nāsiti', 'tr': 'Necd Arapçası', 'uk': 'надждійська арабська', 'ur': 'نجدی عربی', 'uz': 'najd arab', 'uz-Latn': 'najd arab', 'vec': 'àrabo najdi', 'vi': 'Tiếng Ả Rập Najdi', 'xnr': 'नज्दी अरबी', 'yo': 'Èdè Arabiki ti Najidi', 'yrl': 'arabi negede', 'yue': '納吉迪阿拉伯文', 'yue-Hans': '纳吉迪阿拉伯文', 'yue-Hant': '納吉迪阿拉伯文', 'zh': '纳吉迪阿拉伯语', 'zh-Hans': '纳吉迪阿拉伯语', 'zh-Hant': '納吉迪阿拉伯文', 'zu': 'isi-Najdi Arabic'},
    'art': {'en': 'Artificial languages'},
    'art-blk': {'en': 'Bolak'},
    'art-bsp': {'en': 'Black Speech'},
    'art-com': {'en': 'Communicationssprache'},
    'art-dtk': {'en': 'Dothraki'},
    'art-elo': {'en': 'Eloi'},
    'art-gld': {'en': "Goa'uld"},
    'art-lap': {'en': 'Lapine'},
    'art-man': {'en': 'Mandalorian'},
    'art-mun': {'en': 'Mundolinco'},
    'art-nav': {'en': "Na'vi"},
    'art-nox': {'en': 'Noxilo'},
    'art-top': {'en': 'Toki Pona'},
    'art-una': {'en': 'Unas'},
    'aru': {'en': 'Aruá (Amazonas State)'},
    'arv': {'en': 'Arbore'},
    'arw': {'am': 'አራዋክ', 'ar': 'الأراواكية', 'ast': 'arawak', 'az': 'aravak', 'az-Latn': 'aravak', 'bg': 'аравак', 'bn': 'আরাওয়াক', 'br': 'arawakeg', 'brx': 'अरावाक', 'bs': 'aravak', 'bs-Cyrl': 'аравак', 'bs-Latn': 'aravak', 'ca': 'arauac', 'ca-ES-valencia': 'arauac', 'ccp': '𑄃𑄢𑄤𑄇𑄴', 'cs': 'arawacké jazyky', 'cy': 'Arawaceg', 'da': 'arawak', 'de': 'Arawak', 'el': 'Αραγουάκ', 'el-polyton': 'Αραγουάκ', 'en': 'Arawak', 'es': 'arahuaco', 'et': 'aravaki', 'fa': 'آراواکی', 'fi': 'arawak', 'fr': 'arawak', 'fy': 'Arawak', 'gd': 'Arawak', 'gsw': 'Arawak', 'gu': 'અરાવક', 'he': 'ארוואק', 'hi': 'अरावक', 'hi-Latn': 'Arawak', 'hr': 'aravački', 'ht': 'arawak', 'hu': 'aravak', 'id': 'Arawak', 'is': 'aravakska', 'it': 'aruaco', 'ja': 'アラワク語', 'ka': 'არავაკი', 'kgp': 'aruvaki', 'kn': 'ಅರಾವಾಕ್', 'ko': '아라와크어', 'ks': 'اَراوَک', 'ks-Arab': 'اَراوَک', 'lb': 'Arawak-Sprooch', 'lo': 'ອາຣາແວກ', 'lt': 'aravakų', 'lv': 'aravaku', 'mk': 'аравачки', 'ml': 'അറാവക്', 'mr': 'आरावाक', 'mt': 'Arawak', 'nb': 'arawak', 'ne': 'अरावाक', 'nl': 'Arawak', 'nn': 'arawak', 'no': 'arawak', 'or': 'ଆରୱକ', 'pl': 'arawak', 'pt': 'arauaqui', 'rm': 'arawak', 'ro': 'arawak', 'ru': 'аравакский', 'sk': 'arawačtina', 'sl': 'aravaščina', 'sr': 'аравачки', 'sr-Cyrl': 'аравачки', 'sr-Latn': 'aravački', 'sv': 'arawakiska', 'ta': 'அராவாக்', 'te': 'అరావాక్', 'th': 'อาราวัก', 'to': 'lea fakaʻalauaki', 'tr': 'Arawak Dili', 'ug': 'ئاراۋاكچە', 'uk': 'аравакська', 'vi': 'Tiếng Arawak', 'yrl': 'arawaki', 'yue': '阿拉瓦克文', 'yue-Hans': '阿拉瓦克文', 'yue-Hant': '阿拉瓦克文', 'zh': '阿拉瓦克语', 'zh-Hans': '阿拉瓦克语', 'zh-Hant': '阿拉瓦克文'},
    'arx': {'en': 'Aruá (Rodonia State)'},
    'ary': {'ast': 'árabe de Marruecos', 'br': 'arabeg Maroko', 'cs': 'arabština (marocká)', 'cy': 'Arabeg Moroco', 'de': 'Marokkanisches Arabisch', 'en': 'Moroccan Arabic', 'et': 'Maroko araabia', 'fa': 'عربی مراکشی', 'fi': 'marokonarabia', 'fr': 'arabe marocain', 'gd': 'Arabais Mhorocach', 'gu': 'મોરોક્કન અરબી', 'hi-Latn': 'Moroccan Arabic', 'ht': 'arabe marocain', 'id': 'Arab Maroko', 'it': 'arabo marocchino', 'ja': 'モロッコ・アラビア語', 'ko': '모로코 아랍어', 'ksh': 'Marokahnesch Arahbesch', 'lb': 'Marokkanescht Arabesch', 'lt': 'Maroko arabų', 'mk': 'марокански арапски', 'ms': 'Arab Maghribi', 'nb': 'marokkansk-arabisk', 'ne': 'मोरोक्कोली अरबी', 'nl': 'Marokkaans Arabisch', 'nn': 'marokkansk-arabisk', 'no': 'marokkansk-arabisk', 'pl': 'marokański arabski', 'sv': 'marockansk arabiska', 'th': 'อาหรับโมร็อกโก', 'to': 'lea fakaʻalepea-moloko', 'tr': 'Fas Arapçası', 'yue': '摩洛哥阿拉伯文', 'yue-Hans': '摩洛哥阿拉伯文', 'yue-Hant': '摩洛哥阿拉伯文', 'zh-Hant': '摩洛哥阿拉伯文'},
    'arz': {'ast': 'árabe d’Exiptu', 'br': 'arabeg Egipt', 'ca': 'àrab egipci', 'ca-ES-valencia': 'àrab egipci', 'cs': 'arabština (egyptská)', 'cy': 'Arabeg yr Aifft', 'de': 'Ägyptisches Arabisch', 'en': 'Egyptian Arabic', 'et': 'Egiptuse araabia', 'fa': 'عربی مصری', 'fi': 'egyptinarabia', 'fr': 'arabe égyptien', 'gd': 'Arabais Èipheiteach', 'gu': 'ઈજિપ્શિયન અરબી', 'hi-Latn': 'Egyptian Arabic', 'ht': 'arabe égyptien', 'hy': 'եգիպտական արաբերեն', 'id': 'Arab Mesir', 'it': 'arabo egiziano', 'ja': 'エジプト・アラビア語', 'ko': '이집트 아랍어', 'ksh': 'Äjiptesch Arahbesch', 'lb': 'Egyptescht Arabesch', 'lt': 'Egipto arabų', 'mk': 'египетски арапски', 'ms': 'Arab Mesir', 'nb': 'egyptisk arabisk', 'ne': 'इजिप्ट अरबी', 'nl': 'Egyptisch Arabisch', 'nn': 'egyptisk arabisk', 'no': 'egyptisk arabisk', 'pl': 'egipski arabski', 'sv': 'egyptisk arabiska', 'sw': 'Kiarabu cha Misri', 'te': 'ఈజిప్షియన్ అరబిక్', 'th': 'อาหรับพื้นเมืองอียิปต์', 'to': 'lea fakaʻalepea-ʻisipite', 'tr': 'Mısır Arapçası', 'vi': 'Tiếng Ả Rập Ai Cập', 'yue': '埃及阿拉伯文', 'yue-Hans': '埃及阿拉伯文', 'yue-Hant': '埃及阿拉伯文', 'zh-Hant': '埃及阿拉伯文'},
    'as': {'af': 'Assamees', 'ak': 'Asamese', 'am': 'አሳሜዝ', 'ar': 'الأسامية', 'as': 'অসমীয়া', 'ast': 'asamés', 'az': 'assam', 'az-Arab': 'آسامجا', 'az-Cyrl': 'ассам', 'az-Latn': 'assam', 'bal-Latn': 'Asámi', 'be': 'асамская', 'be-tarask': 'асамская', 'bg': 'асамски', 'bgn': 'آسامی', 'bn': 'অসমীয়া', 'br': 'asameg', 'brx': 'असमिया', 'bs': 'asamski', 'bs-Cyrl': 'асемијски', 'bs-Latn': 'asamski', 'ca': 'assamès', 'ca-ES-valencia': 'assamès', 'ccp': '𑄃𑄥𑄟𑄨', 'ce': 'ассамийн', 'chr': 'ᎠᏌᎻᏏ', 'ckb': 'ئاسامی', 'cs': 'ásámština', 'cy': 'Asameg', 'da': 'assamesisk', 'de': 'Assamesisch', 'dsb': 'asamšćina', 'dz': 'ཨ་ས་མིས་ཁ', 'ee': 'assamegbe', 'el': 'Ασαμικά', 'el-polyton': 'Ασαμικά', 'en': 'Assamese', 'eo': 'asama', 'es': 'asamés', 'et': 'assami', 'eu': 'assamera', 'fa': 'آسامی', 'ff-Adlm': '𞤀𞤧𞤢𞤥𞤫𞥅𞤪𞤫', 'fi': 'assami', 'fil': 'Assamese', 'fo': 'assamesiskt', 'fr': 'assamais', 'fur': 'assamês', 'fy': 'Assameesk', 'ga': 'Asaimis', 'gd': 'Asamais', 'gl': 'assamés', 'gsw': 'Assamesisch', 'gu': 'આસામી', 'ha': 'Asamisanci', 'he': 'אסאמית', 'hi': 'असमिया', 'hi-Latn': 'Assamese', 'hr': 'asamski', 'hsb': 'asamšćina', 'ht': 'assamais', 'hu': 'asszámi', 'hy': 'ասամերեն', 'ia': 'assamese', 'id': 'Assam', 'ig': 'Asamisị', 'is': 'assamska', 'it': 'assamese', 'ja': 'アッサム語', 'jv': 'Assam', 'ka': 'ასამური', 'kea': 'asames', 'kgp': 'ajamẽ', 'kk': 'ассам тілі', 'kk-Cyrl': 'ассам тілі', 'km': 'អាសាមីស', 'kn': 'ಅಸ್ಸಾಮೀಸ್', 'ko': '아삼어', 'kok': 'आसामी', 'kok-Deva': 'आसामी', 'ks': 'اسٲمؠ', 'ks-Arab': 'اسٲمؠ', 'ksh': 'Aßamehsesch', 'ku': 'asamî', 'kxv': 'aasamis', 'kxv-Deva': 'आसामीस्', 'kxv-Latn': 'aasamis', 'kxv-Orya': 'ଆସାମିଜ୍', 'kxv-Telu': 'ఆసమీజ్', 'ky': 'ассамча', 'lb': 'Assamesesch', 'lkt': 'Assamese Iyápi', 'lld': 'assamesc', 'lo': 'ອັສຊາມີສ', 'lrc': 'آسامی', 'lt': 'asamų', 'lv': 'asamiešu', 'mai': 'असमी', 'mi': 'Āhamēhi', 'mk': 'асамски', 'ml': 'ആസ്സാമീസ്', 'mn': 'ассам', 'mr': 'आसामी', 'ms': 'Assam', 'mt': 'Assamiż', 'my': 'အာသံ', 'mzn': 'آسامی', 'nb': 'assamesisk', 'ne': 'आसामी', 'nl': 'Assamees', 'nn': 'assamesisk', 'no': 'assamesisk', 'nqo': 'ߊߛߊߡߍߞߊ߲', 'om': 'Assamese', 'or': 'ଆସାମୀୟ', 'pa': 'ਅਸਾਮੀ', 'pa-Guru': 'ਅਸਾਮੀ', 'pcm': 'Asamíz Lángwej', 'pl': 'asamski', 'ps': 'اسامي', 'pt': 'assamês', 'qu': 'Asames Simi', 'rm': 'assami', 'ro': 'asameză', 'ru': 'ассамский', 'rw': 'Icyasamizi', 'sat': 'ᱟᱥᱟᱢᱤᱡᱽ', 'sat-Olck': 'ᱟᱥᱟᱢᱤᱡᱽ', 'sc': 'assamesu', 'sd': 'آسامي', 'sd-Arab': 'آسامي', 'si': 'ඇසෑම්', 'sk': 'ásamčina', 'sl': 'asamščina', 'smn': 'assamkielâ', 'so': 'Asaamiis', 'sq': 'asamezisht', 'sr': 'асамски', 'sr-Cyrl': 'асамски', 'sr-Latn': 'asamski', 'sv': 'assamesiska', 'sw': 'Kiassam', 'ta': 'அஸ்ஸாமீஸ்', 'te': 'అస్సామీస్', 'tg': 'ассомӣ', 'th': 'อัสสัม', 'ti': 'ኣሳሜዝኛ', 'tk': 'assam dili', 'to': 'lea fakaʻasamia', 'tr': 'Assamca', 'tt': 'ассам', 'ug': 'ئاسامچە', 'uk': 'асамська', 'ur': 'آسامی', 'uz': 'assam', 'uz-Cyrl': 'ассомча', 'uz-Latn': 'assam', 'vec': 'asameze', 'vi': 'Tiếng Assam', 'wae': 'Assamesiš', 'wo': 'Asame', 'xnr': 'असमी', 'yi': 'אַסאַמיש', 'yo': 'Èdè Assam', 'yrl': 'asamei', 'yue': '阿薩姆文', 'yue-Hans': '阿萨姆文', 'yue-Hant': '阿薩姆文', 'zh': '阿萨姆语', 'zh-Hans': '阿萨姆语', 'zh-Hant': '阿薩姆文', 'zu': 'isi-Assamese'},
    'asa': {'af': 'Asu', 'am': 'አሱ', 'ar': 'الآسو', 'as': 'আছু', 'asa': 'Kipare', 'ast': 'asu', 'az': 'asu', 'az-Arab': 'آسوجا', 'az-Cyrl': 'асу', 'az-Latn': 'asu', 'bal-Latn': 'Asu', 'be': 'асу', 'be-tarask': 'асу', 'bg': 'асу', 'bgn': 'آسو', 'bn': 'আসু', 'br': 'asu', 'brx': 'आसु', 'bs': 'asu', 'bs-Cyrl': 'асу', 'bs-Latn': 'asu', 'ca': 'pare', 'ca-ES-valencia': 'pare', 'ccp': '𑄃𑄥𑄪', 'ce': 'асу', 'chr': 'ᎠᏑ', 'ckb': 'ئاسوو', 'cs': 'asu', 'cy': 'Asw', 'da': 'asu', 'de': 'Asu', 'dsb': 'pare', 'ee': 'asagbe', 'el': 'Άσου', 'el-polyton': 'Άσου', 'en': 'Asu', 'es': 'asu', 'et': 'asu', 'eu': 'asua', 'fa': 'آسو', 'ff-Adlm': '𞤀𞤧𞤵𞥅𞤪𞤫', 'fi': 'asu', 'fil': 'Asu', 'fo': 'asu', 'fr': 'asu', 'fy': 'Asu', 'ga': 'Asúis', 'gd': 'Asu', 'gl': 'asu', 'gu': 'અસુ', 'ha': 'Asu', 'he': 'אסו', 'hi': 'असु', 'hi-Latn': 'Asu', 'hr': 'asu', 'hsb': 'pare', 'ht': 'asu', 'hu': 'asu', 'hy': 'ասու', 'ia': 'asu', 'id': 'Asu', 'ig': 'Asụ', 'is': 'asu', 'it': 'asu', 'ja': 'アス語', 'jv': 'Asu', 'ka': 'ასუ', 'kea': 'asu', 'kgp': 'asu', 'kk': 'асу тілі', 'kk-Cyrl': 'асу тілі', 'km': 'អាស៊ូ', 'kn': 'ಅಸು', 'ko': '아수어', 'kok': 'असु', 'kok-Deva': 'असु', 'ksh': 'Pare', 'ku': 'asûyî', 'ky': 'асуча', 'lb': 'Asu (Tanzania)', 'lo': 'ອາຊູ', 'lrc': 'آسوٙ', 'lt': 'asu', 'lv': 'asu', 'mai': 'असु', 'mi': 'Ahu', 'mk': 'асу', 'ml': 'ആസു', 'mn': 'асу', 'mr': 'असु', 'ms': 'Asu', 'mt': 'Asu', 'my': 'အာစု', 'mzn': 'آسو', 'nb': 'asu', 'ne': 'आसु', 'nl': 'Asu', 'nn': 'asu (Tanzania)', 'no': 'asu', 'nqo': 'ߊߖ߭ߎߞߊ߲', 'or': 'ଆସୁ', 'pa': 'ਅਸੂ', 'pa-Guru': 'ਅਸੂ', 'pcm': 'Asu Lángwej', 'pl': 'asu', 'ps': 'اسويي', 'pt': 'asu', 'qu': 'Asu Simi', 'rm': 'asu', 'ro': 'asu', 'ru': 'асу', 'sc': 'asu', 'sd': 'اسو', 'sd-Arab': 'اسو', 'si': 'අසු', 'sk': 'asu', 'sl': 'asujščina', 'smn': 'asukielâ', 'so': 'Asu', 'sq': 'asuisht', 'sr': 'асу', 'sr-Cyrl': 'асу', 'sr-Latn': 'asu', 'sv': 'asu', 'sw': 'Kiasu', 'ta': 'அசு', 'te': 'అసు', 'th': 'อาซู', 'ti': 'ኣሱ', 'tk': 'asu dili', 'to': 'lea fakaʻasu', 'tr': 'Asu', 'ug': 'ئاسۇچە', 'uk': 'асу', 'ur': 'آسو', 'uz': 'asu', 'uz-Cyrl': 'асуча', 'uz-Latn': 'asu', 'vec': 'asu', 'vi': 'Tiếng Asu', 'xnr': 'असु', 'yo': 'Èdè Asu', 'yrl': 'asu', 'yue': '阿蘇文', 'yue-Hans': '阿苏文', 'yue-Hant': '阿蘇文', 'zh': '帕雷语', 'zh-Hans': '帕雷语', 'zh-Hant': '阿蘇文', 'zu': 'isi-Asu'},
    'asb': {'en': 'Assiniboine'},
    'asc': {'en': 'Casuarina Coast Asmat'},
    'asd': {'en': 'Asas'},
    'ase': {'am': 'የአሜሪካ የምልክት ቋንቋ', 'ast': 'llingua de signos americana', 'br': 'yezh sinoù Amerika', 'ca': 'llengua de signes americana', 'ca-ES-valencia': 'llengua de signes americana', 'cs': 'znaková řeč (americká)', 'cy': 'Iaith Arwyddion America', 'de': 'Amerikanische Gebärdensprache', 'en': 'American Sign Language', 'et': 'Ameerika viipekeel', 'fi': 'amerikkalainen viittomakieli', 'fr': 'langue des signes américaine', 'gd': 'Cainnt-shanais na h-Aimeireaga', 'hi-Latn': 'American Sign Language', 'ht': 'langue des signes américaine', 'hy': 'ամերիկյան ժեստերի լեզու', 'id': 'Bahasa Isyarat Amerika', 'it': 'lingua dei segni americana', 'ja': 'アメリカ手話', 'ksh': 'Amärrekahnesche Blendeschprohch', 'lb': 'Amerikanesch Zeechesprooch', 'lt': 'Amerikos ženklų kalba', 'mk': 'американски знаковен јазик', 'nb': 'amerikansk tegnspråk', 'ne': 'अमेरिकी साङ्केतिक भाषा', 'nl': 'Amerikaanse Gebarentaal', 'nn': 'amerikansk tegnspråk', 'no': 'amerikansk tegnspråk', 'pl': 'amerykański język migowy', 'sv': 'amerikanskt teckenspråk', 'th': 'ภาษามืออเมริกัน', 'to': 'lea fakaʻilonga-ʻamelika', 'tr': 'Amerikan İşaret Dili', 'uk': 'американська мова рухів', 'vi': 'Ngôn ngữ Ký hiệu Mỹ', 'yue': '美國手語', 'yue-Hans': '美国手语', 'yue-Hant': '美國手語', 'zh-Hant': '美國手語'},
    'asf': {'en': 'Auslan'},
    'asg': {'en': 'Cishingini'},
    'ash': {'en': 'Abishira'},
    'asi': {'en': 'Buruwai'},
    'asj': {'en': 'Sari'},
    'ask': {'en': 'Ashkun'},
    'asl': {'en': 'Asilulu'},
    'asn': {'en': 'Xingú Asuriní'},
    'aso': {'en': 'Dano'},
    'asp': {'en': 'Algerian Sign Language'},
    'asq': {'en': 'Austrian Sign Language'},
    'asr': {'en': 'Asuri'},
    'ass': {'en': 'Ipulo'},
    'ast': {'af': 'Asturies', 'ak': 'Asturiani', 'am': 'አስቱሪያንኛ', 'ar': 'الأسترية', 'as': 'এষ্টুৰীয়', 'ast': 'asturianu', 'az': 'asturiya', 'az-Cyrl': 'астурија', 'az-Latn': 'asturiya', 'bal-Latn': 'Asturiái', 'be': 'астурыйская', 'be-tarask': 'астурыйская', 'bg': 'астурски', 'bgn': 'آستوری', 'bn': 'আস্তুরিয়', 'br': 'asturianeg', 'brx': 'आस्तुरियान', 'bs': 'asturijski', 'bs-Cyrl': 'астуријски', 'bs-Latn': 'asturijski', 'ca': 'asturià', 'ca-ES-valencia': 'asturià', 'ccp': '𑄃𑄌𑄴𑄖𑄪𑄢𑄨𑄠𑄧', 'ce': 'астурийн', 'chr': 'ᎠᏍᏚᎵᎠᏂ', 'ckb': 'ئاستۆری', 'cs': 'asturština', 'cy': 'Astwrianeg', 'da': 'asturisk', 'de': 'Asturisch', 'dsb': 'asturšćina', 'el': 'Αστουριανά', 'el-polyton': 'Αστουριανά', 'en': 'Asturian', 'es': 'asturiano', 'et': 'astuuria', 'eu': 'asturiera', 'fa': 'آستوری', 'ff-Adlm': '𞤀𞤧𞤼𞤵𞤪𞤭𞥅𞤪𞤫', 'fi': 'asturia', 'fil': 'Asturian', 'fo': 'asturianskt', 'fr': 'asturien', 'fur': 'asturian', 'fy': 'Asturysk', 'ga': 'Astúiris', 'gd': 'Astùrais', 'gl': 'asturiano', 'gsw': 'Aschturianisch', 'gu': 'અસ્તુરિયન', 'ha': 'Asturia', 'he': 'אסטורית', 'hi': 'अस्तुरियन', 'hi-Latn': 'Asturian', 'hr': 'asturijski', 'hsb': 'asturšćina', 'ht': 'asturien', 'hu': 'asztúr', 'hy': 'աստուրերեն', 'ia': 'asturiano', 'id': 'Asturia', 'ig': 'Asturianị', 'is': 'astúríska', 'it': 'asturiano', 'ja': 'アストゥリアス語', 'jv': 'Asturia', 'ka': 'ასტურიული', 'kea': 'asturianu', 'kgp': 'anhturijỹnũ', 'kk': 'астурия тілі', 'kk-Cyrl': 'астурия тілі', 'km': 'អាស្ទូរី', 'kn': 'ಆಸ್ಟುರಿಯನ್', 'ko': '아스투리아어', 'kok': 'अस्टुरियान', 'kok-Deva': 'अस्टुरियान', 'ks': 'ایسٹوٗریَن', 'ks-Arab': 'ایسٹوٗریَن', 'ksh': 'Astuhrejahnesch', 'ku': 'astûrî', 'ky': 'астурийче', 'lb': 'Asturianesch', 'lo': 'ອັສຕູຮຽນ', 'lt': 'asturianų', 'lv': 'astūriešu', 'mai': 'अस्तुरियाई', 'mi': 'Ahitūriana', 'mk': 'астурски', 'ml': 'അസ്ട്ടൂരിയൻ', 'mn': 'астури', 'mr': 'अस्तुरियन', 'ms': 'Asturia', 'mt': 'Asturian', 'my': 'အက်စ်တိုးရီးယား', 'nb': 'asturisk', 'ne': 'अस्टुरियाली', 'nl': 'Asturisch', 'nn': 'asturisk', 'no': 'asturisk', 'nqo': 'ߊߛߑߕߎߙߌߞߊ߲', 'om': 'Astuuriyaan', 'or': 'ଆଷ୍ଟୁରିଆନ୍', 'pa': 'ਅਸਤੂਰੀ', 'pa-Guru': 'ਅਸਤੂਰੀ', 'pcm': 'Astúriá Lángwej', 'pl': 'asturyjski', 'ps': 'استورياني', 'pt': 'asturiano', 'qu': 'Asturiano Simi', 'rm': 'asturian', 'ro': 'asturiană', 'ru': 'астурийский', 'sah': 'Астуурдуу', 'sc': 'asturianu', 'sd': 'اسٽورين', 'sd-Arab': 'اسٽورين', 'se': 'asturiagiella', 'si': 'ඇස්ටියුරියන්', 'sk': 'astúrčina', 'sl': 'asturijščina', 'smn': 'asturiakielâ', 'so': 'Astuuriyaan', 'sq': 'asturisht', 'sr': 'астуријски', 'sr-Cyrl': 'астуријски', 'sr-Latn': 'asturijski', 'sv': 'asturiska', 'sw': 'Kiasturia', 'ta': 'அஸ்துரியன்', 'te': 'ఆస్టూరియన్', 'th': 'อัสตูเรียส', 'ti': 'ኣስቱርያን', 'tk': 'asturiý dili', 'to': 'lea fakaʻasitūlia', 'tr': 'Asturyasça', 'ug': 'ئاستۇرىيەچە', 'uk': 'астурійська', 'ur': 'اسٹوریائی', 'uz': 'asturiy', 'uz-Cyrl': 'астурийча', 'uz-Latn': 'asturiy', 'vec': 'asturian', 'vi': 'Tiếng Asturias', 'xnr': 'अस्तुरियन', 'yo': 'Èdè Asturian', 'yrl': 'asiturianu', 'yue': '阿斯圖里亞文', 'yue-Hans': '阿斯图里亚文', 'yue-Hant': '阿斯圖里亞文', 'zh': '阿斯图里亚斯语', 'zh-Hans': '阿斯图里亚斯语', 'zh-Hant': '阿斯圖里亞文', 'zu': 'isi-Asturian'},
    'asu': {'en': 'Tocantins Asurini'},
    'asv': {'en': 'Asoa'},
    'asw': {'en': 'Australian Aborigines Sign Language'},
    'asx': {'en': 'Muratayak'},
    'asy': {'en': 'Yaosakor Asmat'},
    'asz': {'en': 'As'},
    'ata': {'en': 'Pele-Ata'},
    'atb': {'en': 'Zaiwa'},
    'atc': {'en': 'Atsahuaca'},
    'atd': {'en': 'Ata Manobo'},
    'ate': {'en': 'Atemble'},
    'atg': {'en': 'Ivbie North-Okpela-Arhe'},
    'ath': {'en': 'Athapascan languages'},
    'ath-pro': {'en': 'Proto-Athabaskan'},
    'ati': {'en': 'Attié'},
    'atj': {'af': 'Atikamekw', 'am': 'አቲካምከው', 'ar': 'الأتيكاميكو', 'as': 'আতিকামেক', 'az': 'Atikamek', 'az-Latn': 'Atikamek', 'be': 'атыкамек', 'be-tarask': 'атыкамек', 'bg': 'атикамеку', 'bn': 'আটিকামেকিউ', 'br': 'atikamekweg', 'brx': 'आतिकामेको', 'bs': 'atikamekw', 'bs-Cyrl': 'атикамекв', 'bs-Latn': 'atikamekw', 'ca': 'atacama', 'ca-ES-valencia': 'atacama', 'chr': 'ᎠᏂᏘᎧᎺᏆ', 'cs': 'atikamekština', 'cy': 'Atikamekw', 'da': 'atikamekw', 'de': 'Atikamekw', 'dsb': 'atikamekwišćina', 'el': 'Ατικαμέκ', 'el-polyton': 'Ατικαμέκ', 'en': 'Atikamekw', 'es': 'atikamekw', 'et': 'atikameki', 'eu': 'atikamekwera', 'fa': 'آتیکامکو', 'ff-Adlm': '𞤀𞤼𞤭𞤥𞤫𞤳𞤵𞤱𞤪𞤫', 'fi': 'atikamekw', 'fil': 'Atikamekw', 'fr': 'atikamekw', 'ga': 'Atikamekw', 'gd': 'Atikamekw', 'gl': 'atikamekw', 'gu': 'એટીકામેકવ', 'ha': 'Atikamekw', 'he': 'אטיקמק', 'hi': 'अतिकामेक्व', 'hi-Latn': 'Atikamekw', 'hr': 'atikamekw', 'hsb': 'atikamekwišćina', 'ht': 'atikamekw', 'hu': 'attikamek', 'hy': 'աթիկամեկ', 'ia': 'atikamekw', 'id': 'Atikamekw', 'ig': 'Atikamekw', 'is': 'atikamekw', 'it': 'atikamekw', 'ja': 'アティカメク語', 'jv': 'Atikamekw', 'ka': 'ატიკამეკი', 'kk': 'атикамек тілі', 'kk-Cyrl': 'атикамек тілі', 'km': 'អាទិកាម៉េក', 'kn': 'ಅತಿಕಮೆಕ್', 'ko': '아티카메쿠어', 'kok': 'अटिकामेक्वु', 'kok-Deva': 'अटिकामेक्वु', 'ku': 'atîkamekî', 'ky': 'атикамекче', 'lo': 'ອາຕິກາແມກ', 'lt': 'atikamekų', 'lv': 'atikameku', 'mai': 'अतिकामेक', 'mi': 'Atikameke', 'mk': 'атикамек', 'ml': 'അറ്റികമെക്\u200cവ്', 'mn': 'атикамек', 'mr': 'अटिकमेक', 'ms': 'Atikamekw', 'my': 'အာတီကမဲကု', 'nb': 'atikamekw', 'ne': 'आतिकामिको', 'nl': 'Atikamekw', 'nn': 'atikamekw', 'no': 'atikamekw', 'or': 'ଆଟିକାମେକ୍ୱ', 'pa': 'ਅਤਿਕਾਮੇਕਵ', 'pa-Guru': 'ਅਤਿਕਾਮੇਕਵ', 'pcm': 'Atíkamẹ́kw', 'pl': 'atikamekw', 'ps': 'اتیکامیکو', 'pt': 'atikamekw', 'qu': 'Atikamekw', 'ro': 'atikamekw', 'ru': 'атикамек', 'sc': 'atikamekw', 'sd': 'اٽيڪاميڪو', 'sd-Arab': 'اٽيڪاميڪو', 'si': 'ඇටිකමෙක්ව්', 'sk': 'atikamekwčina', 'sl': 'atikamekwščina', 'so': 'Atikamekw', 'sq': 'atikamekisht', 'sr': 'атикамеку', 'sr-Cyrl': 'атикамеку', 'sr-Latn': 'atikameku', 'sv': 'atikamekw', 'sw': 'Kiatikamekw', 'ta': 'அடிகாமேக்வ்', 'te': 'అతికామెక్వ', 'th': 'อะทิคาเมก', 'ti': 'ኣቲካመክ', 'tk': 'atikamekw dili', 'to': 'lea fakaʻatikameku', 'tr': 'Atikamekçe', 'uk': 'атикамек', 'ur': 'اٹیکامیکو', 'uz': 'atikamek', 'uz-Latn': 'atikamek', 'vec': 'atikamec', 'vi': 'Tiếng Atikamekw', 'yo': 'Èdè Atikameki', 'yue': '阿提卡米克文', 'yue-Hans': '阿提卡米克文', 'yue-Hant': '阿提卡米克文', 'zh': '阿提卡米克语', 'zh-Hans': '阿提卡米克语', 'zh-Hant': '阿提卡梅克語', 'zu': 'isi-Atikamekw'},
    'atk': {'en': 'Ati'},
    'atl': {'en': 'Mt. Iraya Agta'},
    'atm': {'en': 'Ata'},
    'atn': {'en': 'Ashtiani'},
    'ato': {'en': 'Atong (Cameroon)'},
    'atp': {'en': 'Pudtol Atta'},
    'atq': {'en': 'Aralle-Tabulahan'},
    'atr': {'en': 'Waimiri-Atroari'},
    'ats': {'en': 'Gros Ventre'},
    'att': {'en': 'Pamplona Atta'},
    'atu': {'en': 'Reel'},
    'atv': {'en': 'Northern Altai'},
    'atw': {'en': 'Atsugewi'},
    'atx': {'en': 'Arutani'},
    'aty': {'en': 'Aneityum'},
    'atz': {'en': 'Arta'},
    'aua': {'en': 'Asumboa'},
    'aub': {'en': 'Alugu'},
    'auc': {'en': 'Waorani'},
    'aud': {'en': 'Anuta'},
    'aue': {'en': 'ǂKxʼauǁʼein'},
    'auf': {'en': 'Arauan languages'},
    'auf-pro': {'en': 'Proto-Arawa'},
    'aug': {'en': 'Aguna'},
    'auh': {'en': 'Aushi'},
    'aui': {'en': 'Anuki'},
    'auj': {'en': 'Awjilah'},
    'auk': {'en': 'Heyo'},
    'aul': {'en': 'Aulua'},
    'aum': {'en': 'Asu (Nigeria)'},
    'aun': {'en': 'Molmo One'},
    'auo': {'en': 'Auyokawa'},
    'aup': {'en': 'Makayam'},
    'auq': {'en': 'Anus'},
    'aur': {'en': 'Aruek'},
    'aus': {'en': 'Australian languages'},
    'aus-arn-pro': {'en': 'Proto-Arnhem'},
    'aus-bun': {'en': 'Bunurong'},
    'aus-cww-pro': {'en': 'Proto-Central New South Wales'},
    'aus-dal-pro': {'en': 'Proto-Daly'},
    'aus-nyu-pro': {'en': 'Proto-Nyulnyulan'},
    'aus-pam-pro': {'en': 'Proto-Pama-Nyungan'},
    'aus-wdj-pro': {'en': 'Proto-Iwaidjan'},
    'aus-won': {'en': 'Wong-gie'},
    'aut': {'en': 'Austral'},
    'auu': {'en': 'Auye'},
    'auw': {'en': 'Awyi'},
    'aux': {'en': 'Aurá'},
    'auy': {'en': 'Awiyaana'},
    'auz': {'en': 'Uzbeki Arabic'},
    'av': {'af': 'Avaries', 'am': 'አቫሪክ', 'ar': 'الأوارية', 'as': 'আভেৰিক', 'ast': 'aváricu', 'az': 'avar', 'az-Cyrl': 'авар', 'az-Latn': 'avar', 'be': 'аварская', 'be-tarask': 'аварская', 'bg': 'аварски', 'bgn': 'آواری', 'bn': 'আভেরিক', 'br': 'avar', 'brx': 'आभारिक', 'bs': 'avarski', 'bs-Cyrl': 'аварски', 'bs-Latn': 'avarski', 'ca': 'àvar', 'ca-ES-valencia': 'àvar', 'ccp': '𑄃𑄞𑄬𑄢𑄨𑄇𑄴', 'ce': 'суьйлийн', 'chr': 'ᎠᏩᎵᎧ', 'ckb': 'ئەڤاری', 'cs': 'avarština', 'cy': 'Afareg', 'da': 'avarisk', 'de': 'Awarisch', 'dsb': 'awaršćina', 'el': 'Αβαρικά', 'el-polyton': 'Αβαρικά', 'en': 'Avaric', 'es': 'avar', 'et': 'avaari', 'eu': 'avarera', 'fa': 'آواری', 'ff-Adlm': '𞤀𞤬𞤱𞤢𞤪𞤭𞥅𞤪𞤫', 'fi': 'avaari', 'fil': 'Avaric', 'fo': 'avariskt', 'fr': 'avar', 'fur': 'avar', 'fy': 'Avarysk', 'ga': 'Aváiris', 'gd': 'Avarais', 'gl': 'avar', 'gsw': 'Awarisch', 'gu': 'અવેરિક', 'ha': 'Avaric', 'he': 'אווארית', 'hi': 'अवेरिक', 'hi-Latn': 'Avaric', 'hr': 'avarski', 'hsb': 'awaršćina', 'ht': 'avar', 'hu': 'avar', 'hy': 'ավարերեն', 'ia': 'avaro', 'id': 'Avar', 'ig': 'Avarịk', 'is': 'avaríska', 'it': 'avaro', 'ja': 'アヴァル語', 'jv': 'Avaric', 'ka': 'ხუნძური', 'kgp': 'avariko', 'kk': 'авар тілі', 'kk-Cyrl': 'авар тілі', 'km': 'អាវ៉ារីក', 'kn': 'ಅವರಿಕ್', 'ko': '아바릭어', 'kok': 'अवारिक', 'kok-Deva': 'अवारिक', 'ks': 'اَوارِک', 'ks-Arab': 'اَوارِک', 'ksh': 'Avahresch', 'ku': 'avarî', 'ky': 'аварикче', 'lb': 'Awaresch', 'lkt': 'Avaric Iyápi', 'lo': 'ອາວາຣິກ', 'lt': 'avarikų', 'lv': 'avāru', 'mai': 'अवरिक', 'mi': 'Āwhāriki', 'mk': 'аварски', 'ml': 'അവാരിക്', 'mn': 'авар', 'mr': 'अ\u200dॅव्हेरिक', 'ms': 'Avaric', 'mt': 'Avarik', 'my': 'အာဗာရစ်ခ်', 'nb': 'avarisk', 'ne': 'अवारिक', 'nl': 'Avarisch', 'nn': 'avarisk', 'no': 'avarisk', 'or': 'ଆଭାରିକ୍', 'os': 'авайраг', 'pa': 'ਅਵਾਰਿਕ', 'pa-Guru': 'ਅਵਾਰਿਕ', 'pcm': 'Afarík Lángwej', 'pl': 'awarski', 'ps': 'اواري', 'pt': 'avárico', 'qu': 'Avaric', 'rm': 'avaric', 'ro': 'avară', 'ru': 'аварский', 'sah': 'Аваардыы', 'sc': 'avaru', 'sd': 'اويرس', 'sd-Arab': 'اويرس', 'si': 'ඇවරික්', 'sk': 'avarčina', 'sl': 'avarščina', 'smn': 'avarkielâ', 'so': 'U dhashay Avar', 'sq': 'avarikisht', 'sr': 'аварски', 'sr-Cyrl': 'аварски', 'sr-Latn': 'avarski', 'sv': 'avariska', 'sw': 'Kiavari', 'ta': 'அவேரிக்', 'te': 'అవారిక్', 'th': 'อาวาร์', 'ti': 'ኣቫርኛ', 'tk': 'awar dili', 'to': 'lea fakaʻavaliki', 'tr': 'Avar dili', 'ug': 'ئاۋارچە', 'uk': 'аварська', 'ur': 'اواری', 'uz': 'avar', 'uz-Cyrl': 'аварча', 'uz-Latn': 'avar', 'vec': 'avàrego', 'vi': 'Tiếng Avaric', 'xnr': 'अवेरिक', 'yo': 'Èdè Afariki', 'yrl': 'awariku', 'yue': '阿瓦爾文', 'yue-Hans': '阿瓦尔文', 'yue-Hant': '阿瓦爾文', 'zh': '阿瓦尔语', 'zh-Hans': '阿瓦尔语', 'zh-Hant': '阿瓦爾文', 'zu': 'isi-Avaric'},
    'avb': {'en': 'Avau'},
    'avd': {'en': 'Alviri-Vidari'},
    'avi': {'en': 'Avikam'},
    'avk': {'ast': 'kotava', 'cs': 'kotava', 'de': 'Kotava', 'en': 'Kotava', 'fi': 'kotava', 'fr': 'kotava', 'gd': 'Kotava', 'hi-Latn': 'Kotava', 'ht': 'kotava', 'it': 'kotava', 'ja': 'コタヴァ', 'lb': 'Kotava', 'lt': 'kotava', 'mk': 'котава', 'nb': 'kotava', 'ne': 'कोटावा', 'nl': 'Kotava', 'nn': 'kotava', 'no': 'kotava', 'pl': 'kotava', 'sv': 'kotava', 'th': 'โคตาวา', 'to': 'lea fakakotava', 'tr': 'Kotava', 'yue': '科塔瓦文', 'yue-Hans': '科塔瓦文', 'yue-Hant': '科塔瓦文', 'zh-Hant': '科塔瓦文'},
    'avl': {'en': 'Eastern Egyptian Bedawi Arabic'},
    'avm': {'en': 'Angkamuthi'},
    'avn': {'en': 'Avatime'},
    'avo': {'en': 'Agavotaguerra'},
    'avs': {'en': 'Aushiri'},
    'avt': {'en': 'Au'},
    'avu': {'en': 'Avokaya'},
    'avv': {'en': 'Avá-Canoeiro'},
    'awa': {'af': 'Awadhi', 'am': 'አዋድሂ', 'ar': 'الأوادية', 'as': 'অৱধী', 'ast': 'awadhi', 'az': 'avadhi', 'az-Cyrl': 'авадһи', 'az-Latn': 'avadhi', 'be': 'авадхі', 'be-tarask': 'авадхі', 'bg': 'авади', 'bgn': 'آوادی', 'bn': 'আওয়াধি', 'br': 'awadhi', 'brx': 'अवधी', 'bs': 'avadhi', 'bs-Cyrl': 'авадхи', 'bs-Latn': 'avadhi', 'ca': 'awadhi', 'ca-ES-valencia': 'awadhi', 'ccp': '𑄃𑄤𑄙𑄨', 'ce': 'авадхи', 'chr': 'ᎠᏩᏗ', 'ckb': 'ئاوادهی', 'cs': 'awadhština', 'cy': 'Awadhi', 'da': 'awadhi', 'de': 'Awadhi', 'dsb': 'awandhišćina', 'el': 'Αγουαντί', 'el-polyton': 'Αγουαντί', 'en': 'Awadhi', 'es': 'avadhi', 'et': 'avadhi', 'eu': 'awadhiera', 'fa': 'اودهی', 'ff-Adlm': '𞤀𞤱𞤢𞤣𞤭𞥅𞤪𞤫', 'fi': 'awadhi', 'fil': 'Awadhi', 'fo': 'awadhi', 'fr': 'awadhi', 'fy': 'Awadhi', 'ga': 'Avaidis', 'gd': 'Awadhi', 'gl': 'awadhi', 'gsw': 'Awadhi', 'gu': 'અવધી', 'ha': 'Awadhi', 'he': 'אוואדית', 'hi': 'अवधी', 'hi-Latn': 'Awadhi', 'hr': 'awadhi', 'hsb': 'awandhišćina', 'ht': 'awadhi', 'hu': 'awádi', 'hy': 'ավադհի', 'ia': 'awadhi', 'id': 'Awadhi', 'ig': 'Awadhi', 'is': 'avadí', 'it': 'awadhi', 'ja': 'アワディー語', 'jv': 'Awadhi', 'ka': 'ავადი', 'kgp': 'avanhi', 'kk': 'авадхи тілі', 'kk-Cyrl': 'авадхи тілі', 'km': 'អាវ៉ាឌី', 'kn': 'ಅವಧಿ', 'ko': '아와히어', 'kok': 'अवधी', 'kok-Deva': 'अवधी', 'ks': 'اَوَدی', 'ks-Arab': 'اَوَدی', 'ksh': 'Awahdesch', 'ku': 'awadhî', 'ky': 'авадхиче', 'lb': 'Awadhi', 'lo': 'ອາວາຮິ', 'lt': 'avadhi', 'lv': 'avadhu', 'mai': 'अवधी', 'mi': 'Awāti', 'mk': 'авади', 'ml': 'അവാധി', 'mn': 'авадхи', 'mr': 'अवधी', 'ms': 'Awadhi', 'mt': 'Awadhi', 'my': 'အာဝါဒီ', 'nb': 'avadhi', 'ne': 'अवधी', 'nl': 'Awadhi', 'nn': 'avadhi', 'no': 'avadhi', 'or': 'ଆୱାଧି', 'pa': 'ਅਵਧੀ', 'pa-Guru': 'ਅਵਧੀ', 'pcm': 'Awadhí', 'pl': 'awadhi', 'ps': 'اوادي', 'pt': 'awadhi', 'qu': 'Awadhi', 'rm': 'awadhi', 'ro': 'awadhi', 'ru': 'авадхи', 'sc': 'awadhi', 'sd': 'اواڌي', 'sd-Arab': 'اواڌي', 'si': 'අවදි', 'sk': 'awadhi', 'sl': 'avadščina', 'smn': 'awadhikielâ', 'sq': 'auadhisht', 'sr': 'авади', 'sr-Cyrl': 'авади', 'sr-Latn': 'avadi', 'sv': 'awadhi', 'sw': 'Kiawadhi', 'ta': 'அவதி', 'te': 'అవధి', 'th': 'อวธี', 'ti': 'ኣዋዲ', 'tk': 'awadhi dili', 'to': 'lea fakaʻauati', 'tr': 'Awadhi', 'ug': 'ئاۋادىچە', 'uk': 'авадхі', 'ur': 'اوادھی', 'uz': 'avadxi', 'uz-Cyrl': 'авадхи', 'uz-Latn': 'avadxi', 'vec': 'awadi', 'vi': 'Tiếng Awadhi', 'xnr': 'अवधी', 'yo': 'Èdè Awadi', 'yrl': 'awadi', 'yue': '阿瓦文', 'yue-Hans': '阿瓦文', 'yue-Hant': '阿瓦文', 'zh': '阿瓦德语', 'zh-Hans': '阿瓦德语', 'zh-Hant': '阿瓦文', 'zu': 'isi-Awadhi'},
    'awb': {'en': 'Awa (Papua New Guinea)'},
    'awc': {'en': 'Cicipu'},
    'awd': {'en': 'Arawakan languages'},
    'awd-mpr': {'en': 'Maypure'},
    'awd-pro': {'en': 'Proto-Arawakan'},
    'awd-taa-pro': {'en': 'Proto-Ta-Arawakan'},
    'awe': {'en': 'Awetí'},
    'awg': {'en': 'Anguthimri'},
    'awh': {'en': 'Awbono'},
    'awi': {'en': 'Aekyom'},
    'awk': {'en': 'Awabakal'},
    'awm': {'en': 'Arawum'},
    'awn': {'en': 'Awngi'},
    'awo': {'en': 'Awak'},
    'awr': {'en': 'Awera'},
    'aws': {'en': 'South Awyu'},
    'awt': {'en': 'Araweté'},
    'awu': {'en': 'Central Awyu'},
    'awv': {'en': 'Jair Awyu'},
    'aww': {'en': 'Awun'},
    'awx': {'en': 'Awara'},
    'awy': {'en': 'Edera Awyu'},
    'axb': {'en': 'Abipon'},
    'axe': {'en': 'Ayerrerenge'},
    'axg': {'en': 'Mato Grosso Arára'},
    'axk': {'en': 'Yaka (Central African Republic)'},
    'axl': {'en': 'Lower Southern Aranda'},
    'axm': {'en': 'Middle Armenian'},
    'axx': {'en': 'Xârâgurè'},
    'ay': {'af': 'Aymara', 'am': 'አያማርኛ', 'ar': 'الأيمارا', 'as': 'আয়মাৰা', 'ast': 'aymara', 'az': 'aymara', 'az-Cyrl': 'ајмара', 'az-Latn': 'aymara', 'be': 'аймара', 'be-tarask': 'аймара', 'bg': 'аймара', 'bgn': 'ایمارای', 'bn': 'আয়মারা', 'br': 'aymara', 'brx': 'आयमारा', 'bs': 'ajmara', 'bs-Cyrl': 'ајмара', 'bs-Latn': 'ajmara', 'ca': 'aimara', 'ca-ES-valencia': 'aimara', 'ccp': '𑄃𑄠𑄧𑄟𑄢', 'ce': 'аймара', 'chr': 'ᎠᏱᎹᎳ', 'ckb': 'ئایمارا', 'cs': 'ajmarština', 'cy': 'Aymareg', 'da': 'aymara', 'de': 'Aymara', 'dsb': 'aymaršćina', 'ee': 'aymargbe', 'el': 'Αϊμάρα', 'el-polyton': 'Αϊμάρα', 'en': 'Aymara', 'eo': 'ajmara', 'es': 'aimara', 'et': 'aimara', 'eu': 'aimara', 'fa': 'آیمارایی', 'ff-Adlm': '𞤀𞤴𞤥𞤢𞤪𞤢𞥄𞤪𞤫', 'fi': 'aimara', 'fil': 'Aymara', 'fo': 'aymara', 'fr': 'aymara', 'fur': 'aymarà', 'fy': 'Aymara', 'ga': 'Aidhmiris', 'gd': 'Aymara', 'gl': 'aimará', 'gsw': 'Aymara', 'gu': 'આયમારા', 'ha': 'Aymaranci', 'he': 'איימארית', 'hi': 'आयमारा', 'hi-Latn': 'Aymara', 'hr': 'ajmarski', 'hsb': 'aymaršćina', 'ht': 'aymara', 'hu': 'ajmara', 'hy': 'այմարա', 'ia': 'aymara', 'id': 'Aymara', 'ig': 'Aymara', 'is': 'aímara', 'it': 'aymara', 'ja': 'アイマラ語', 'jv': 'Aymara', 'ka': 'აიმარა', 'kea': 'aimara', 'kgp': 'ajimỹra', 'kk': 'аймара тілі', 'kk-Cyrl': 'аймара тілі', 'km': 'អីម៉ារ៉ា', 'kn': 'ಅಯ್ಮಾರಾ', 'ko': '아이마라어', 'kok': 'ऐमरा', 'kok-Deva': 'ऐमरा', 'ks': 'ایمارا', 'ks-Arab': 'ایمارا', 'ksh': 'Aimahresch', 'ku': 'aymarayî', 'ky': 'аймарача', 'lb': 'Aymara', 'lo': 'ອາຍມາລາ', 'lt': 'aimarų', 'lv': 'aimaru', 'mai': 'आयमरा', 'mi': 'Aimāra', 'mk': 'ајмарски', 'ml': 'അയ്മാറ', 'mn': 'аймара', 'mr': 'ऐमरा', 'ms': 'Aymara', 'mt': 'Aymara', 'my': 'အိုင်မာရ', 'nb': 'aymara', 'ne': 'ऐमारा', 'nl': 'Aymara', 'nn': 'aymara', 'no': 'aymara', 'or': 'ଆୟମାରା', 'pa': 'ਅਈਮਾਰਾ', 'pa-Guru': 'ਅਈਮਾਰਾ', 'pcm': 'Aymára', 'pl': 'ajmara', 'ps': 'ایماري', 'pt': 'aimará', 'qu': 'Aymara Simi', 'rm': 'aymara', 'ro': 'aymara', 'ru': 'аймара', 'sc': 'aimara', 'sd': 'ایمارا', 'sd-Arab': 'ایمارا', 'si': 'අයිමරා', 'sk': 'aymarčina', 'sl': 'ajmarščina', 'smn': 'aymarakielâ', 'so': 'U dhashay Aymar', 'sq': 'ajmarisht', 'sr': 'ајмара', 'sr-Cyrl': 'ајмара', 'sr-Latn': 'ajmara', 'sv': 'aymara', 'sw': 'Kiaymara', 'ta': 'அய்மரா', 'te': 'ఐమారా', 'th': 'ไอย์มารา', 'ti': 'ኣይማራ', 'tk': 'aýmara dili', 'to': 'lea fakaʻaimala', 'tr': 'Aymara', 'ug': 'ئايماراچە', 'uk': 'аймара', 'ur': 'ایمارا', 'uz': 'aymara', 'uz-Cyrl': 'аймара', 'uz-Latn': 'aymara', 'vec': 'aimar', 'vi': 'Tiếng Aymara', 'wae': 'Aymara', 'xnr': 'आयमारा', 'yo': 'Èdè Amara', 'yrl': 'aimará', 'yue': '艾馬拉文', 'yue-Hans': '艾马拉文', 'yue-Hant': '艾馬拉文', 'zh': '艾马拉语', 'zh-Hans': '艾马拉语', 'zh-Hant': '艾馬拉文', 'zu': 'isi-Aymara'},
    'aya': {'en': 'Awar'},
    'ayb': {'en': 'Ayizo Gbe'},
    'ayc': {'en': 'Southern Aymara'},
    'ayd': {'en': 'Ayabadhu'},
    'aye': {'en': 'Ayere'},
    'ayg': {'en': 'Ginyanga'},
    'ayh': {'en': 'Hadrami Arabic'},
    'ayi': {'en': 'Leyigha'},
    'ayk': {'en': 'Akuku'},
    'ayl': {'en': 'Libyan Arabic'},
    'ayn': {'en': 'Sanaani Arabic'},
    'ayo': {'en': 'Ayoreo'},
    'ayp': {'en': 'North Mesopotamian Arabic'},
    'ayq': {'en': 'Ayi (Papua New Guinea)'},
    'ayr': {'en': 'Central Aymara'},
    'ays': {'en': 'Sorsogon Ayta'},
    'ayt': {'en': 'Magbukun Ayta'},
    'ayu': {'en': 'Ayu'},
    'ayx': {'en': 'Ayi (China)'},
    'ayy': {'en': 'Tayabas Ayta'},
    'ayz': {'en': 'Mai Brat'},
    'az': {'af': 'Azerbeidjans', 'ak': 'Asabegyanni', 'am': 'አዘርባጃንኛ', 'ar': 'الأذربيجانية', 'as': 'আজেৰবাইজানী', 'ast': 'azerbaixanu', 'az': 'azərbaycan', 'az-Arab': 'آذربایجان تۆرکجه\u200cسی', 'az-Cyrl': 'азәрбајҹан', 'az-Latn': 'azərbaycan', 'bal-Latn': 'Ázerbáijáni', 'be': 'азербайджанская', 'be-tarask': 'азербайджанская', 'bg': 'азербайджански', 'bgn': 'آذربایجانی', 'bn': 'আজারবাইজানী', 'br': 'azerbaidjaneg', 'brx': 'आजारबायजानि', 'bs': 'azerbejdžanski', 'bs-Cyrl': 'азербејџански', 'bs-Latn': 'azerbejdžanski', 'ca': 'azerbaidjanès', 'ca-ES-valencia': 'azerbaidjanès', 'ccp': '𑄃𑄎𑄢𑄴𑄝𑄳𑄆𑄎𑄚𑄩', 'ce': 'азербайджанийн', 'chr': 'ᎠᏎᏆᏣᏂ', 'ckb': 'ئازەربایجانی', 'cs': 'ázerbájdžánština', 'cy': 'Aserbaijaneg', 'da': 'aserbajdsjansk', 'de': 'Aserbaidschanisch', 'dsb': 'azerbajdžanšćina', 'dz': 'ཨ་ཛར་བྷའི་ཇཱན་ཁ', 'ee': 'azerbaijangbe', 'el': 'Αζερμπαϊτζανικά', 'el-polyton': 'Αζερμπαϊτζανικά', 'en': 'Azerbaijani', 'eo': 'azerbajĝana', 'es': 'azerbaiyano', 'et': 'aserbaidžaani', 'eu': 'azerbaijanera', 'fa': 'ترکی آذربایجانی', 'ff-Adlm': '𞤀𞤶𞤢𞤪𞤦𞤢𞤴𞤭𞤶𞤢𞤲𞤭𞥅𞤪𞤫', 'fi': 'azeri', 'fil': 'Azerbaijani', 'fo': 'aserbajdsjanskt', 'fr': 'azerbaïdjanais', 'fur': 'azerbaijani', 'fy': 'Azerbeidzjaansk', 'ga': 'Asarbaiseáinis', 'gd': 'Asarbaideànais', 'gl': 'acerbaixano', 'gsw': 'Aserbaidschanisch', 'gu': 'અઝરબૈજાની', 'ha': 'Azerbaijanci', 'he': 'אזרית', 'hi': 'अज़रबैजानी', 'hi-Latn': 'Azerbaijani', 'hr': 'azerbajdžanski', 'hsb': 'azerbajdźanšćina', 'ht': 'azerbaïdjanais', 'hu': 'azerbajdzsáni', 'hy': 'ադրբեջաներեն', 'ia': 'azerbaidzhano', 'id': 'Azerbaijani', 'ig': 'Azerbaijani', 'is': 'aserska', 'it': 'azerbaigiano', 'ja': 'アゼルバイジャン語', 'jv': 'Azerbaijan', 'ka': 'აზერბაიჯანული', 'kea': 'azerbaijanu', 'kgp': 'ajermajỹnũ', 'kk': 'әзірбайжан тілі', 'kk-Cyrl': 'әзірбайжан тілі', 'km': 'អាស៊ែបៃហ្សង់', 'kn': 'ಅಜೆರ್ಬೈಜಾನಿ', 'ko': '아제르바이잔어', 'kok': 'अझरबैजानी', 'kok-Deva': 'अझरबैजानी', 'ks': 'اَزَربیجانی', 'ks-Arab': 'اَزَربیجانی', 'ksh': 'Asserbaidschahnesch', 'ku': 'azerî', 'kxv': 'ajerbaijani', 'kxv-Deva': 'अज़रबेजानी', 'kxv-Latn': 'ajerbaijani', 'kxv-Orya': 'ଆଜେରବେଇଜାନି', 'kxv-Telu': 'అజరబైజాని', 'ky': 'азербайжанча', 'lb': 'Aserbaidschanesch', 'lkt': 'Azerbaijani Iyápi', 'lld': 'azerbaijan', 'lo': 'ອາເຊີໄບຈານິ', 'lrc': 'آذأربایئجانی ھارگە', 'lt': 'azerbaidžaniečių', 'lv': 'azerbaidžāņu', 'mai': 'अजरबैजानी', 'mi': 'Atepaihānia', 'mk': 'азербејџански', 'ml': 'അസർബൈജാനി', 'mn': 'азербайжан', 'mr': 'अझरबैजानी', 'ms': 'Azerbaijan', 'mt': 'Ażerbajġani', 'my': 'အဇာဘိုင်ဂျန်', 'mzn': 'آذری ِترکی', 'nb': 'aserbajdsjansk', 'ne': 'अजरबैजानी', 'nl': 'Azerbeidzjaans', 'nn': 'aserbajdsjansk', 'no': 'aserbajdsjansk', 'nqo': 'ߊߖߍߙߑߓߊߦߌߖߊߞߊ߲', 'om': 'Afaan Azerbaijani', 'or': 'ଆଜେରବାଇଜାନି', 'os': 'тӕтӕйраг', 'pa': 'ਅਜ਼ਰਬਾਈਜਾਨੀ', 'pa-Guru': 'ਅਜ਼ਰਬਾਈਜਾਨੀ', 'pcm': 'Azẹrbaijáni Lángwej', 'pl': 'azerbejdżański', 'ps': 'اذربایجاني', 'pt': 'azerbaijano', 'qu': 'Azerbaiyano Simi', 'rm': 'aserbeidschanic', 'ro': 'azeră', 'ru': 'азербайджанский', 'rw': 'Inyazeribayijani', 'sah': 'Адьырбайдьаанныы', 'sc': 'azerbaigianu', 'sd': 'آزربائيجاني', 'sd-Arab': 'آزربائيجاني', 'si': 'අසර්බයිජාන්', 'sk': 'azerbajdžančina', 'sl': 'azerbajdžanščina', 'smn': 'azerbaidžankielâ', 'so': 'Asarbayjan', 'sq': 'azerbajxhanisht', 'sr': 'азербејџански', 'sr-Cyrl': 'азербејџански', 'sr-Latn': 'azerbejdžanski', 'sv': 'azerbajdzjanska', 'sw': 'Kiazerbaijani', 'syr': 'ܐܙܪܒܝܓܐܢܝܬ', 'ta': 'அசர்பைஜானி', 'te': 'అజర్బైజాని', 'tg': 'озарбойҷонӣ', 'th': 'อาเซอร์ไบจาน', 'ti': 'ኣዘርባጃንኛ', 'tk': 'azerbaýjan dili', 'to': 'lea fakaʻasepaisani', 'tr': 'Azerbaycan dili', 'tt': 'әзәрбайҗан', 'ug': 'ئەزەربەيجانچە', 'uk': 'азербайджанська', 'ur': 'آذربائیجانی', 'uz': 'ozarbayjon', 'uz-Cyrl': 'озарбайжонча', 'uz-Latn': 'ozarbayjon', 'vec': 'azerbaijan', 'vi': 'Tiếng Azerbaijan', 'wae': 'Serbaidšaniš', 'wo': 'Aserbayjane', 'xnr': 'अज़रबैजानी', 'yi': 'אַזערביידזשאַניש', 'yo': 'Èdè Asabaijani', 'yrl': 'aserbayanu', 'yue': '亞塞拜然文', 'yue-Hans': '亚塞拜然文', 'yue-Hant': '亞塞拜然文', 'zh': '阿塞拜疆语', 'zh-Hans': '阿塞拜疆语', 'zh-Hant': '亞塞拜然文', 'zu': 'isi-Azerbaijani'},
    'az-Arab': {'az': 'cənubi azərbaycan', 'az-Arab': 'تۆرکجه', 'az-Latn': 'cənubi azərbaycan', 'ckb': 'ئازەربایجانی باشووری', 'cy': 'Aserbaijaneg Deheuol', 'fa': 'ترکی آذری جنوبی', 'hr': 'južnoazerbajdžanski', 'kgp': 'aseri sur', 'ksh': 'Södasserbaidschahnesch', 'lrc': 'آذأری ھارگە', 'lv': 'dienvidazerbaidžāņu', 'mzn': 'جنوبی آذری ترکی', 'pt': 'azeri sul', 'tr': 'Güney Azerice', 'uk': 'південноазербайджанська', 'ur': 'آزربائیجانی (عربی)', 'yrl': 'aseri sú', 'zh': '南阿塞拜疆语', 'zh-Hans': '南阿塞拜疆语'},
    'aza': {'en': 'Azha'},
    'azb': {'en': 'South Azerbaijani'},
    'azc': {'en': 'Uto-Aztecan languages'},
    'azc-cup-pro': {'en': 'Proto-Cupan'},
    'azc-nah-pro': {'en': 'Proto-Nahuan'},
    'azc-num-pro': {'en': 'Proto-Numic'},
    'azc-pro': {'en': 'Proto-Uto-Aztecan'},
    'azc-tak-pro': {'en': 'Proto-Takic'},
    'azd': {'en': 'Eastern Durango Nahuatl'},
    'azg': {'en': 'San Pedro Amuzgos Amuzgo'},
    'azj': {'en': 'North Azerbaijani'},
    'azm': {'en': 'Ipalapa Amuzgo'},
    'azn': {'en': 'Western Durango Nahuatl'},
    'azo': {'en': 'Awing'},
    'azt': {'en': 'Faire Atta'},
    'azz': {'en': 'Highland Puebla Nahuatl'},
    'ba': {'af': 'Baskir', 'am': 'ባስኪርኛ', 'ar': 'الباشكيرية', 'as': 'বাছখিৰ', 'ast': 'bashkir', 'az': 'başqırd', 'az-Arab': 'باشقیرجا', 'az-Cyrl': 'башгырд', 'az-Latn': 'başqırd', 'ba': 'башҡорт теле', 'bal-Latn': 'Bashkar', 'be': 'башкірская', 'be-tarask': 'башкірская', 'bg': 'башкирски', 'bgn': 'باشقیری', 'bn': 'বাশকির', 'br': 'bachkir', 'brx': 'बाशकिर', 'bs': 'baškirski', 'bs-Cyrl': 'башкир', 'bs-Latn': 'baškirski', 'ca': 'baixkir', 'ca-ES-valencia': 'baixkir', 'ccp': '𑄝𑄌𑄴𑄇𑄨𑄢𑄴', 'ce': 'башкирийн', 'chr': 'ᏆᏍᎯᎩᎠ', 'ckb': 'باشکیەر', 'cs': 'baškirština', 'cy': 'Bashcorteg', 'da': 'bashkir', 'de': 'Baschkirisch', 'dsb': 'baškiršćina', 'el': 'Μπασκίρ', 'el-polyton': 'Μπασκίρ', 'en': 'Bashkir', 'eo': 'baŝkira', 'es': 'baskir', 'et': 'baškiiri', 'eu': 'baxkirera', 'fa': 'باشقیری', 'ff-Adlm': '𞤄𞤢𞤧𞤳𞤭𞥅𞤪𞤫', 'fi': 'baškiiri', 'fil': 'Bashkir', 'fo': 'bashkir', 'fr': 'bachkir', 'fy': 'Basjkiersk', 'ga': 'Baiscíris', 'gd': 'Bashkir', 'gl': 'baxkir', 'gsw': 'Baschkirisch', 'gu': 'બશ્કીર', 'ha': 'Bashkir', 'he': 'בשקירית', 'hi': 'बशख़िर', 'hi-Latn': 'Bashkir', 'hr': 'baškirski', 'hsb': 'baškiršćina', 'ht': 'bachkir', 'hu': 'baskír', 'hy': 'բաշկիրերեն', 'ia': 'bashkir', 'id': 'Bashkir', 'ig': 'Bashkir', 'is': 'baskír', 'it': 'baschiro', 'ja': 'バシキール語', 'jv': 'Bashkir', 'ka': 'ბაშკირული', 'kea': 'baxkir', 'kgp': 'majkir', 'kk': 'башқұрт тілі', 'kk-Cyrl': 'башқұрт тілі', 'km': 'បាស្គៀ', 'kn': 'ಬಶ್ಕಿರ್', 'ko': '바슈키르어', 'kok': 'बष्किर', 'kok-Deva': 'बष्किर', 'ks': 'بَشکیٖر', 'ks-Arab': 'بَشکیٖر', 'ksh': 'Baschkihresch', 'ku': 'başkîrî', 'ky': 'башкырча', 'lb': 'Baschkiresch', 'lkt': 'Bashkir Iyápi', 'lld': 'bashkir', 'lo': 'ບາຣກີ', 'lrc': 'باشکیری', 'lt': 'baškirų', 'lv': 'baškīru', 'mai': 'बशकीर', 'mi': 'Pākīra', 'mk': 'башкирски', 'ml': 'ബഷ്ഖിർ', 'mn': 'башкир', 'mr': 'बष्किर', 'ms': 'Bashkir', 'mt': 'Bashkir', 'my': 'ဘက်ရှ်ကာ', 'mzn': 'باشقیری', 'nb': 'basjkirsk', 'ne': 'बास्किर', 'nl': 'Basjkiers', 'nn': 'basjkirsk', 'no': 'basjkirsk', 'or': 'ବାଶକିର୍', 'os': 'башкираг', 'pa': 'ਬਸ਼ਕੀਰ', 'pa-Guru': 'ਬਸ਼ਕੀਰ', 'pcm': 'Bashkír', 'pl': 'baszkirski', 'ps': 'باشکير', 'pt': 'bashkir', 'qu': 'Baskir Simi', 'rm': 'baschkir', 'ro': 'bașkiră', 'ru': 'башкирский', 'sc': 'baschiru', 'sd': 'ڪينيڊا', 'sd-Arab': 'ڪينيڊا', 'si': 'බාෂ්කිර්', 'sk': 'baškirčina', 'sl': 'baškirščina', 'smn': 'baškirkielâ', 'so': 'Bashkir', 'sq': 'bashkirisht', 'sr': 'башкирски', 'sr-Cyrl': 'башкирски', 'sr-Latn': 'baškirski', 'sv': 'basjkiriska', 'sw': 'Kibashkiri', 'ta': 'பஷ்கிர்', 'te': 'బాష్కిర్', 'tg': 'бошқирдӣ', 'th': 'บัชคีร์', 'ti': 'ባሽኪር', 'tk': 'başgyrt dili', 'to': 'lea fakapasikili', 'tr': 'Başkırtça', 'tt': 'башкорт', 'ug': 'باشقىرتچە', 'uk': 'башкирська', 'ur': 'باشکیر', 'uz': 'boshqird', 'uz-Cyrl': 'бошқирдча', 'uz-Latn': 'boshqird', 'vec': 'bashkir', 'vi': 'Tiếng Bashkir', 'wo': 'Baskir', 'xnr': 'बशख़िर', 'yo': 'Èdè Bashiri', 'yrl': 'baxikiri', 'yue': '巴什客爾文', 'yue-Hans': '巴什客尔文', 'yue-Hant': '巴什客爾文', 'zh': '巴什基尔语', 'zh-Hans': '巴什基尔语', 'zh-Hant': '巴什喀爾文', 'zu': 'isi-Bashkir'},
    'baa': {'en': 'Babatana'},
    'bab': {'en': 'Bainouk-Gunyuño'},
    'bac': {'en': 'Badui'},
    'bad': {'en': 'Banda languages'},
    'bae': {'en': 'Baré'},
    'baf': {'en': 'Nubaca'},
    'bag': {'en': 'Tuki'},
    'bah': {'en': 'Bahamas Creole English'},
    'bai': {'en': 'Bamileke languages'},
    'baj': {'en': 'Barakai'},
    'bal': {'am': 'ባሉቺ', 'ar': 'البلوشية', 'ast': 'baluchi', 'az': 'baluc', 'az-Latn': 'baluc', 'bal': 'بلۆچی', 'bal-Arab': 'بلۆچی', 'bal-Latn': 'Balóchi', 'bg': 'балучи', 'bn': 'বেলুচী', 'br': 'baloutchi', 'brx': 'बलूची', 'bs': 'baluči', 'bs-Cyrl': 'балучи', 'bs-Latn': 'baluči', 'ca': 'balutxi', 'ca-ES-valencia': 'balutxi', 'ccp': '𑄝𑄬𑄣𑄪𑄌𑄩', 'cs': 'balúčština', 'cy': 'Balwtsi', 'da': 'baluchi', 'de': 'Belutschisch', 'el': 'Μπαλούτσι', 'el-polyton': 'Μπαλούτσι', 'en': 'Baluchi', 'es': 'baluchi', 'et': 'belutši', 'fa': 'بلوچی', 'fi': 'belutši', 'fr': 'baloutchi', 'fy': 'Baloetsjysk', 'gd': 'Baluchì', 'gsw': 'Belutschisch', 'gu': 'બલૂચી', 'he': 'באלוצ׳י', 'hi': 'बलूची', 'hi-Latn': 'Baluchi', 'hr': 'belučki', 'ht': 'baloutchi', 'hu': 'balucsi', 'ia': 'baluchi', 'id': 'Baluchi', 'ig': 'Baluchi', 'is': 'balúkí', 'it': 'beluci', 'ja': 'バルーチー語', 'ka': 'ბელუჯი', 'kgp': 'marusi', 'kn': 'ಬಲೂಚಿ', 'ko': '발루치어', 'ks': 'بَلوٗچی', 'ks-Arab': 'بَلوٗچی', 'ksh': 'Beluhtschesch', 'ku': 'belûcî', 'lb': 'Belutschesch', 'lkt': 'Baluchi Iyápi', 'lo': 'ບາລູຊີ', 'lt': 'baluči', 'lv': 'beludžu', 'mk': 'белуџиски', 'ml': 'ബലൂചി', 'mr': 'बलुची', 'ms': 'Baluchi', 'mt': 'Baluċi', 'nb': 'baluchi', 'ne': 'बालुची', 'nl': 'Beloetsji', 'nn': 'baluchi', 'no': 'baluchi', 'nqo': 'ߓߊߟߎߗߌߞߊ߲', 'or': 'ବାଲୁଚି', 'pl': 'beludżi', 'ps': 'بلوڅي', 'pt': 'balúchi', 'rm': 'belutschi', 'ro': 'baluchi', 'ru': 'белуджский', 'sk': 'balúčtina', 'sl': 'beludžijščina', 'sr': 'белучки', 'sr-Cyrl': 'белучки', 'sr-Latn': 'belučki', 'sv': 'baluchiska', 'ta': 'பலூச்சி', 'te': 'బాలుచి', 'th': 'บาลูชิ', 'ti': 'ባሉቺ', 'to': 'lea fakapalusi', 'tr': 'Beluçça', 'ug': 'بېلۇجىچە', 'uk': 'балучі', 'vec': 'baluci', 'vi': 'Tiếng Baluchi', 'yrl': 'barúxi', 'yue': '俾路支文', 'yue-Hans': '俾路支文', 'yue-Hant': '俾路支文', 'zh': '俾路支语', 'zh-Hans': '俾路支语', 'zh-Hant': '俾路支文'},
    'bal-Latn': {'bal-Latn': 'Balóchi (Látin)'},
    'ban': {'af': 'Balinees', 'am': 'ባሊኔስ', 'ar': 'البالينية', 'as': 'বালিনীজ', 'ast': 'balinés', 'az': 'bali', 'az-Cyrl': 'балли', 'az-Latn': 'bali', 'be': 'балійская', 'be-tarask': 'балійская', 'bg': 'балийски', 'bgn': 'بالینسی', 'bn': 'বালিনীয়', 'br': 'balineg', 'brx': 'बालिनिस', 'bs': 'balinezijski', 'bs-Cyrl': 'балинезијски', 'bs-Latn': 'balinezijski', 'ca': 'balinès', 'ca-ES-valencia': 'balinès', 'ccp': '𑄝𑄣𑄨𑄚𑄩𑄠𑄧', 'ce': 'балийн', 'chr': 'ᏆᎵᏁᏏ', 'ckb': 'بالی', 'cs': 'balijština', 'cy': 'Balïeg', 'da': 'balinesisk', 'de': 'Balinesisch', 'dsb': 'balinezišćina', 'el': 'Μπαλινίζ', 'el-polyton': 'Μπαλινίζ', 'en': 'Balinese', 'es': 'balinés', 'et': 'bali', 'eu': 'baliera', 'fa': 'بالیایی', 'ff-Adlm': '𞤄𞤢𞥄𞤤𞤭𞥅𞤪𞤫', 'fi': 'bali', 'fil': 'Balinese', 'fo': 'balinesiskt', 'fr': 'balinais', 'fy': 'Balineesk', 'ga': 'Bailís', 'gd': 'Cànan Bali', 'gl': 'balinés', 'gsw': 'Balinesisch', 'gu': 'બાલિનીસ', 'ha': 'Balenesanci', 'he': 'באלינזית', 'hi': 'बालिनीस', 'hi-Latn': 'Balinese', 'hr': 'balijski', 'hsb': 'balinezišćina', 'ht': 'balinais', 'hu': 'balinéz', 'hy': 'բալիերեն', 'ia': 'balinese', 'id': 'Bali', 'ig': 'Balinese', 'is': 'balíska', 'it': 'balinese', 'ja': 'バリ語', 'jv': 'Bali', 'ka': 'ბალინური', 'kea': 'balines', 'kgp': 'marinẽj', 'kk': 'бали тілі', 'kk-Cyrl': 'бали тілі', 'km': 'បាលី', 'kn': 'ಬಲಿನೀಸ್', 'ko': '발리어', 'kok': 'बालिनीज', 'kok-Deva': 'बालिनीज', 'ks': 'بالِنیٖز', 'ks-Arab': 'بالِنیٖز', 'ksh': 'Balinehsesch', 'ku': 'balînî', 'ky': 'баличе', 'lb': 'Balinesesch', 'lo': 'ບາລີເນັດ', 'lt': 'baliečių', 'lv': 'baliešu', 'mai': 'बाली', 'mi': 'Pārinīhi', 'mk': 'балиски', 'ml': 'ബാലിനീസ്', 'mn': 'бали', 'mr': 'बालिनीज', 'ms': 'Bali', 'mt': 'Baliniż', 'my': 'ဘာလီ', 'nb': 'balinesisk', 'ne': 'बाली', 'nl': 'Balinees', 'nn': 'balinesisk', 'no': 'balinesisk', 'or': 'ବାଲିନୀଜ୍', 'pa': 'ਬਾਲੀਨੀਜ਼', 'pa-Guru': 'ਬਾਲੀਨੀਜ਼', 'pcm': 'Balinẹẹ́s', 'pl': 'balijski', 'ps': 'بالنی', 'pt': 'balinês', 'qu': 'Balines Simi', 'rm': 'balinais', 'ro': 'balineză', 'ru': 'балийский', 'sc': 'balinesu', 'sd': 'بالينيس', 'sd-Arab': 'بالينيس', 'si': 'බැලිනීස්', 'sk': 'balijčina', 'sl': 'balijščina', 'smn': 'balikielâ', 'so': 'U dhashay Baline', 'sq': 'balinezisht', 'sr': 'балијски', 'sr-Cyrl': 'балијски', 'sr-Latn': 'balijski', 'sv': 'balinesiska', 'sw': 'Kibali', 'ta': 'பலினீஸ்', 'te': 'బాలినీస్', 'tg': 'балинӣ', 'th': 'บาหลี', 'ti': 'ባሊንኛ', 'tk': 'baliý dili', 'to': 'lea fakapali', 'tr': 'Bali dili', 'tt': 'бали', 'ug': 'بالىچە', 'uk': 'балійська', 'ur': 'بالینیز', 'uz': 'bali', 'uz-Cyrl': 'балича', 'uz-Latn': 'bali', 'vec': 'balineze', 'vi': 'Tiếng Bali', 'wo': 'Bali', 'xnr': 'बालिनीस', 'yi': 'באַלינעזיש', 'yo': 'Èdè Balini', 'yrl': 'barinei', 'yue': '峇里文', 'yue-Hans': '峇里文', 'yue-Hant': '峇里文', 'zh': '巴厘语', 'zh-Hans': '巴厘语', 'zh-Hant': '峇里文', 'zu': 'isi-Balinese'},
    'bao': {'en': 'Waimaha'},
    'bap': {'en': 'Bantawa'},
    'bar': {'am': 'ባቫሪያን', 'ast': 'bávaru', 'br': 'bavarieg', 'ca': 'bavarès', 'ca-ES-valencia': 'bavarès', 'cs': 'bavorština', 'de': 'Bairisch', 'en': 'Bavarian', 'et': 'baieri', 'fa': 'باواریایی', 'fi': 'baijeri', 'fr': 'bavarois', 'ga': 'Baváiris', 'he': 'בווארית', 'hi-Latn': 'Bavarian', 'ht': 'bavarois', 'id': 'Bavaria', 'it': 'bavarese', 'ja': 'バイエルン・オーストリア語', 'ksh': 'Bairesch', 'lb': 'Bairesch', 'lt': 'bavarų', 'mk': 'баварски', 'nb': 'bairisk', 'ne': 'बाभारियाली', 'nl': 'Beiers', 'nn': 'bairisk', 'no': 'bairisk', 'pl': 'bawarski', 'sv': 'bayerska', 'th': 'บาวาเรีย', 'to': 'lea fakapavālia', 'tr': 'Bavyera dili', 'uk': 'баеріш', 'vi': 'Tiếng Bavaria', 'yi': 'בײַעריש', 'yue': '巴伐利亞文', 'yue-Hans': '巴伐利亚文', 'yue-Hant': '巴伐利亞文', 'zh-Hant': '巴伐利亞文'},
    'bas': {'af': 'Basaa', 'am': 'ባሳ', 'ar': 'الباسا', 'as': 'বাছা', 'ast': 'basaa', 'az': 'basa', 'az-Cyrl': 'баса', 'az-Latn': 'basa', 'bal-Latn': 'Basá', 'bas': 'Ɓàsàa', 'be': 'басаа', 'be-tarask': 'басаа', 'bg': 'баса', 'bgn': 'باسا', 'bn': 'বাসা', 'br': 'basaa', 'brx': 'बासा', 'bs': 'basa', 'bs-Cyrl': 'баса', 'bs-Latn': 'basa', 'ca': 'basa', 'ca-ES-valencia': 'basa', 'ccp': '𑄝𑄥', 'ce': 'баса', 'chr': 'ᏆᏌᎠ', 'ckb': 'باسا', 'cs': 'basa', 'cy': 'Basâeg', 'da': 'basaa', 'de': 'Bassa', 'dsb': 'basaa', 'el': 'Μπάσα', 'el-polyton': 'Μπάσα', 'en': 'Basaa', 'es': 'basaa', 'et': 'basaa', 'eu': 'basaa', 'fa': 'باسایی', 'ff-Adlm': '𞤄𞤢𞤧𞤢𞥄𞤪𞤫', 'fi': 'basaa', 'fil': 'Basaa', 'fo': 'basaa', 'fr': 'bassa', 'fy': 'Basa', 'ga': 'Basáis', 'gd': 'Basaa', 'gl': 'basaa', 'gsw': 'Basaa', 'gu': 'બસા', 'ha': 'Basaa', 'he': 'בסאא', 'hi': 'बसा', 'hi-Latn': 'Basaa', 'hr': 'basa', 'hsb': 'basaa', 'ht': 'bassa', 'hu': 'basza', 'hy': 'բասաա', 'ia': 'basaa', 'id': 'Basa', 'ig': 'Basaà', 'is': 'basa', 'it': 'basa', 'ja': 'バサ語', 'jv': 'Basaa', 'ka': 'ბასა', 'kea': 'basa', 'kgp': 'masa', 'kk': 'баса тілі', 'kk-Cyrl': 'баса тілі', 'km': 'បាសា', 'kn': 'ಬಸಾ', 'ko': '바사어', 'kok': 'बस्सा', 'kok-Deva': 'बस्सा', 'ks': 'باسا', 'ks-Arab': 'باسا', 'ksh': 'Basaa-Schprohche', 'ku': 'basayî', 'ky': 'басаача', 'lb': 'Basaa-Sprooch', 'lo': 'ບາຊາ', 'lt': 'basų', 'lv': 'basu', 'mai': 'बासा', 'mi': 'Pahā', 'mk': 'баса', 'ml': 'ബാസ', 'mn': 'басаа', 'mr': 'बसा', 'ms': 'Basaa', 'mt': 'Basa', 'my': 'ဘာဆာ', 'nb': 'basaa', 'ne': 'बासा', 'nl': 'Basa', 'nn': 'basa', 'nnh': 'Shwóŋò pʉa mbasǎ', 'no': 'basaa', 'nqo': 'ߓߛߊߞߊ߲', 'or': 'ବାସା', 'pa': 'ਬਾਸਾ', 'pa-Guru': 'ਬਾਸਾ', 'pcm': 'Básaa Lángwej', 'pl': 'basaa', 'ps': 'باسا', 'pt': 'basa', 'qu': 'Basaa Simi', 'rm': 'basaa', 'ro': 'basaa', 'ru': 'баса', 'sc': 'basaa', 'sd': 'باسا', 'sd-Arab': 'باسا', 'si': 'බසා', 'sk': 'basa', 'sl': 'basa', 'smn': 'basaakielâ', 'so': 'Basaa', 'sq': 'basaisht', 'sr': 'баса', 'sr-Cyrl': 'баса', 'sr-Latn': 'basa', 'sv': 'basa', 'sw': 'Kibasaa', 'ta': 'பாஸா', 'te': 'బసా', 'th': 'บาสา', 'ti': 'ባሳ', 'tk': 'basaa dili', 'to': 'lea fakapasaʻa', 'tr': 'Basa Dili', 'ug': 'باساچە', 'uk': 'баса', 'ur': 'باسا', 'uz': 'basa', 'uz-Cyrl': 'басаа', 'uz-Latn': 'basa', 'vec': 'basa', 'vi': 'Tiếng Basaa', 'xnr': 'बसा', 'yo': 'Èdè Basaa', 'yrl': 'basa', 'yue': '巴薩文', 'yue-Hans': '巴萨文', 'yue-Hant': '巴薩文', 'zh': '巴萨语', 'zh-Hans': '巴萨语', 'zh-Hant': '巴薩文', 'zu': 'isi-Basaa'},
    'bat': {'en': 'Baltic languages'},
    'bau': {'en': 'Bada (Nigeria)'},
    'bav': {'en': 'Vengo'},
    'baw': {'en': 'Bambili-Bambui'},
    'bax': {'am': 'ባሙን', 'ar': 'بامن', 'ast': 'bamun', 'bs': 'bamunski', 'bs-Latn': 'bamunski', 'ca': 'bamum', 'ca-ES-valencia': 'bamum', 'cs': 'bamun', 'cy': 'Bamwmeg', 'da': 'bamun', 'de': 'Bamun', 'el': 'Μπαμούν', 'el-polyton': 'Μπαμούν', 'en': 'Bamun', 'es': 'bamún', 'et': 'bamuni', 'fa': 'بمونی', 'fi': 'bamum', 'fr': 'bamoun', 'fy': 'Bamoun', 'gd': 'Bamun', 'gu': 'બામન', 'he': 'במום', 'hi-Latn': 'Bamun', 'hr': 'bamunski', 'ht': 'bamoun', 'hu': 'bamun', 'id': 'Bamun', 'is': 'bamun', 'it': 'bamun', 'ja': 'バムン語', 'ka': 'ბამუნი', 'kgp': 'mamũm', 'ko': '바문어', 'lb': 'Bamun', 'lkt': 'Bamun Iyápi', 'lo': 'ບາມຸນ', 'lt': 'bamunų', 'lv': 'bamumu', 'mk': 'бамунски', 'ml': 'ബാമുൻ', 'ms': 'Bamun', 'nb': 'bamun', 'ne': 'बामुन', 'nl': 'Bamoun', 'nn': 'bamun', 'nnh': 'Shwóŋò pamom', 'no': 'bamun', 'pl': 'bamum', 'pt': 'bamum', 'ro': 'bamun', 'ru': 'бамум', 'sk': 'bamun', 'sv': 'bamunska', 'sw': 'Kibamun', 'th': 'บามัน', 'to': 'lea fakapamuni', 'tr': 'Bamun', 'ug': 'بامۇنچە', 'uk': 'бамум', 'vi': 'Tiếng Bamun', 'yrl': 'bamũ', 'yue': '巴姆穆文', 'yue-Hans': '巴姆穆文', 'yue-Hant': '巴姆穆文', 'zh': '巴姆穆语', 'zh-Hans': '巴姆穆语', 'zh-Hant': '巴姆穆文'},
    'bay': {'en': 'Batuley'},
    'baz': {'en': 'Tunen'},
    'bba': {'en': 'Baatonum'},
    'bbb': {'en': 'Barai'},
    'bbc': {'am': 'ባታካ ቶባ', 'ast': 'batak toba', 'cs': 'batak toba', 'de': 'Batak Toba', 'en': 'Batak Toba', 'et': 'bataki', 'fi': 'batak-toba', 'fr': 'batak toba', 'gd': 'Batak Toba', 'hi-Latn': 'Batak Toba', 'ht': 'batak toba', 'id': 'Batak Toba', 'it': 'batak toba', 'ja': 'トバ・バタク語', 'lb': 'Batak Toba', 'lt': 'batak toba', 'mk': 'тоба', 'nb': 'batak toba', 'ne': 'बाताक तोबा', 'nl': 'Batak Toba', 'nn': 'batak toba', 'no': 'batak toba', 'pl': 'batak toba', 'sv': 'batak-toba', 'th': 'บาตักโทบา', 'to': 'lea fakatōpe-pēteki', 'tr': 'Batak Toba', 'uk': 'батак тоба', 'vi': 'Tiếng Batak Toba', 'yue': '巴塔克托巴文', 'yue-Hans': '巴塔克托巴文', 'yue-Hant': '巴塔克托巴文', 'zh-Hant': '巴塔克托巴文'},
    'bbd': {'en': 'Bau'},
    'bbe': {'en': 'Bangba'},
    'bbf': {'en': 'Baibai'},
    'bbg': {'en': 'Barama'},
    'bbh': {'en': 'Bugan'},
    'bbi': {'en': 'Barombi'},
    'bbj': {'ar': 'لغة الغومالا', 'ast': 'ghomala', 'bs': 'gomala', 'bs-Latn': 'gomala', 'ca': 'ghomala', 'ca-ES-valencia': 'ghomala', 'cs': 'ghomala', 'da': 'ghomala', 'de': 'Ghomala', 'el': 'Γκομάλα', 'el-polyton': 'Γκομάλα', 'en': 'Ghomala', 'es': 'ghomala', 'et': 'ghomala', 'fi': 'ghomala', 'fr': 'ghomalaʼ', 'fy': 'Ghomala’', 'gd': 'Ghomala', 'he': 'גומאלה', 'hi-Latn': 'Ghomala', 'hr': 'ghomala', 'ht': 'ghomalaʼ', 'hu': 'gomala', 'id': 'Ghomala', 'it': 'ghomala', 'ja': 'ゴーマラ語', 'kgp': 'gumỹra', 'ko': '고말라어', 'lb': 'Ghomálá’', 'lo': 'ໂຄມາລາ', 'lt': 'ghomalų', 'lv': 'gomalu', 'mk': 'гомала', 'ml': 'ഘോമാല', 'ms': 'Ghomala', 'nb': 'ghomala', 'ne': 'घोमाला', 'nl': 'Ghomala’', 'nn': 'ghomala', 'nnh': 'Shwóŋò pʉa nzsekàʼa', 'no': 'ghomala', 'pl': 'ghomala', 'pt': 'ghomala’', 'ro': 'ghomala', 'ru': 'гомала', 'sk': 'ghomala', 'sv': 'ghomala', 'sw': 'Kighomala', 'th': 'โคมาลา', 'to': 'lea fakakomala', 'tr': 'Ghomala', 'ug': 'گومالاچە', 'uk': 'гомала', 'vi': 'Tiếng Ghomala', 'yrl': 'gumara', 'yue': '戈馬拉文', 'yue-Hans': '戈马拉文', 'yue-Hant': '戈馬拉文', 'zh': '戈马拉语', 'zh-Hans': '戈马拉语', 'zh-Hant': '戈馬拉文'},
    'bbk': {'en': 'Babanki'},
    'bbl': {'en': 'Bats'},
    'bbm': {'en': 'Babango'},
    'bbn': {'en': 'Uneapa'},
    'bbo': {'en': 'Northern Bobo Madaré'},
    'bbp': {'en': 'West Central Banda'},
    'bbq': {'en': 'Bamali'},
    'bbr': {'en': 'Girawa'},
    'bbs': {'en': 'Bakpinka'},
    'bbt': {'en': 'Mburku'},
    'bbu': {'en': 'Kulung (Nigeria)'},
    'bbv': {'en': 'Karnai'},
    'bbw': {'en': 'Baba'},
    'bbx': {'en': 'Bubia'},
    'bby': {'en': 'Befang'},
    'bbz': {'en': 'Babalia Creole Arabic'},
    'bca': {'en': 'Central Bai'},
    'bcb': {'en': 'Bainouk-Samik'},
    'bcc': {'en': 'Southern Balochi'},
    'bcd': {'en': 'North Babar'},
    'bce': {'en': 'Bamenyam'},
    'bcf': {'en': 'Bamu'},
    'bcg': {'en': 'Baga Pokur'},
    'bch': {'en': 'Bariai'},
    'bci': {'en': 'Baoulé'},
    'bcj': {'en': 'Bardi'},
    'bck': {'en': 'Bunuba'},
    'bcl': {'en': 'Central Bikol'},
    'bcm': {'en': 'Bannoni'},
    'bcn': {'en': 'Bali (Nigeria)'},
    'bco': {'en': 'Kaluli'},
    'bcp': {'en': 'Bali (Democratic Republic of Congo)'},
    'bcq': {'en': 'Bench'},
    'bcr': {'en': 'Babine'},
    'bcs': {'en': 'Kohumono'},
    'bct': {'en': 'Bendi'},
    'bcu': {'en': 'Awad Bing'},
    'bcv': {'en': 'Shoo-Minda-Nye'},
    'bcw': {'en': 'Bana'},
    'bcy': {'en': 'Bacama'},
    'bcz': {'en': 'Bainouk-Gunyaamolo'},
    'bda': {'en': 'Bayot'},
    'bdb': {'en': 'Basap'},
    'bdc': {'en': 'Emberá-Baudó'},
    'bdd': {'en': 'Bunama'},
    'bde': {'en': 'Bade'},
    'bdf': {'en': 'Biage'},
    'bdg': {'en': 'Bonggi'},
    'bdh': {'en': 'Baka (South Sudan)'},
    'bdi': {'en': 'Burun'},
    'bdj': {'en': 'Bai (South Sudan)'},
    'bdk': {'en': 'Budukh'},
    'bdl': {'en': 'Indonesian Bajau'},
    'bdm': {'en': 'Buduma'},
    'bdn': {'en': 'Baldemu'},
    'bdo': {'en': 'Morom'},
    'bdp': {'en': 'Bende'},
    'bdq': {'en': 'Bahnar'},
    'bdr': {'en': 'West Coast Bajau'},
    'bds': {'en': 'Burunge'},
    'bdt': {'en': 'Bokoto'},
    'bdu': {'en': 'Oroko'},
    'bdv': {'en': 'Bodo Parja'},
    'bdw': {'en': 'Baham'},
    'bdx': {'en': 'Budong-Budong'},
    'bdy': {'en': 'Bandjalang'},
    'bdz': {'en': 'Badeshi'},
    'be': {'af': 'Belarussies', 'agq': 'Bɛ̀làlusàn', 'ak': 'Belarus kasa', 'am': 'ቤላራሻኛ', 'ar': 'البيلاروسية', 'as': 'বেলাৰুছীয়', 'asa': 'Kibelarusi', 'ast': 'bielorrusu', 'az': 'belarus', 'az-Arab': 'بئلاروسجا', 'az-Cyrl': 'беларус', 'az-Latn': 'belarus', 'bal-Latn': 'Bélárusi', 'bas': 'Hɔp u bièlòrûs', 'be': 'беларуская', 'be-tarask': 'беларуская', 'bem': 'Ichi Belarus', 'bez': 'Hibelarusi', 'bg': 'беларуски', 'bgn': 'بیلاروسی', 'bm': 'biyelorisikan', 'bn': 'বেলারুশিয়', 'br': 'belaruseg', 'brx': 'बेलारुसियान', 'bs': 'bjeloruski', 'bs-Cyrl': 'бјелоруски', 'bs-Latn': 'bjeloruski', 'ca': 'belarús', 'ca-ES-valencia': 'belarús', 'ccp': '𑄝𑄬𑄣𑄢𑄪𑄥𑄨𑄠𑄧', 'ce': 'белорусийн', 'cgg': 'Oruberarusi', 'chr': 'ᏇᎳᎷᏏ', 'ckb': 'بیلاڕووسی', 'cs': 'běloruština', 'cy': 'Belarwseg', 'da': 'belarusisk', 'dav': 'Kibelarusi', 'de': 'Belarussisch', 'dje': 'Belaruus senni', 'dsb': 'běłorušćina', 'dyo': 'belarus', 'dz': 'བེལ་ཨ་རུས་ཁ', 'ebu': 'Kĩmbelarusi', 'ee': 'belarusiagbe', 'el': 'Λευκορωσικά', 'el-polyton': 'Λευκορωσικά', 'en': 'Belarusian', 'eo': 'belorusa', 'es': 'bielorruso', 'et': 'valgevene', 'eu': 'bielorrusiera', 'ewo': 'Ǹkɔ́bɔ belarúsian', 'fa': 'بلاروسی', 'ff': 'Belaruuse', 'ff-Adlm': '𞤄𞤫𞤤𞤢𞤪𞤭𞥅𞤧𞤭𞥅𞤪𞤫', 'ff-Latn': 'Belaruuse', 'fi': 'valkovenäjä', 'fil': 'Belarusian', 'fo': 'hvitarussiskt', 'fr': 'biélorusse', 'fur': 'bielorùs', 'fy': 'Wyt-Russysk', 'ga': 'Bealarúisis', 'gd': 'Bealaruisis', 'gl': 'belaruso', 'gsw': 'Wiissrussisch', 'gu': 'બેલારુશિયન', 'guz': 'Kibelarusi', 'ha': 'Belarusanci', 'he': 'בלארוסית', 'hi': 'बेलारूसी', 'hi-Latn': 'Belarusian', 'hr': 'bjeloruski', 'hsb': 'běłorušćina', 'ht': 'biélorusse', 'hu': 'belarusz', 'hy': 'բելառուսերեն', 'ia': 'bielorusso', 'id': 'Belarusia', 'ig': 'Belarusian', 'is': 'hvítrússneska', 'it': 'bielorusso', 'ja': 'ベラルーシ語', 'jmc': 'Kyibelarusi', 'jv': 'Belarus', 'ka': 'ბელარუსული', 'kab': 'Tabilarusit', 'kam': 'Kibelarusi', 'kde': 'Chibelalusi', 'kea': 'bielorusu', 'kgp': 'huso-kupri', 'khq': 'Belaruus senni', 'ki': 'Kibelarusi', 'kk': 'беларусь тілі', 'kk-Cyrl': 'беларусь тілі', 'kln': 'kutitab Belarusa', 'km': 'បេឡារុស', 'kn': 'ಬೆಲರೂಸಿಯನ್', 'ko': '벨라루스어', 'kok': 'बेलारुशियन', 'kok-Deva': 'बेलारुशियन', 'ks': 'بیلَروٗشیَن', 'ks-Arab': 'بیلَروٗشیَن', 'ksb': 'Kibelaausi', 'ksf': 'ribɛlɔrís', 'ksh': 'Wiißrußesch', 'ku': 'belarûsî', 'kxv': 'belarusiati', 'kxv-Deva': 'बेलारूसी', 'kxv-Latn': 'belarusiati', 'kxv-Orya': 'ବେଲାରୁଷିଆନ୍', 'kxv-Telu': 'బెలారుసియన్', 'ky': 'беларусча', 'lag': 'Kɨberalúusi', 'lb': 'Wäissrussesch', 'lg': 'Lubelarusi', 'lkt': 'Belarus Iyápi', 'lld': 'belarus', 'ln': 'libyelorisí', 'lo': 'ເບລາຣັສຊຽນ', 'lrc': 'بئلاروٙسی', 'lt': 'baltarusių', 'lu': 'Belarusi', 'luo': 'Kibelarusi', 'luy': 'Kibelarusi', 'lv': 'baltkrievu', 'mai': 'बेलारूसी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lBelarusi', 'mer': 'Kĩbelarusi', 'mfe': 'bieloris', 'mg': 'Bielorosy', 'mgh': 'Ibelausi', 'mi': 'Perarūhiana', 'mk': 'белоруски', 'ml': 'ബെലാറുഷ്യൻ', 'mn': 'беларусь', 'mr': 'बेलारुशियन', 'ms': 'Belarus', 'mt': 'Belarussu', 'mua': 'belarussiya', 'my': 'ဘီလာရုစ်', 'mzn': 'بلاروسی', 'naq': 'Belarusanǁî gowab', 'nb': 'belarusisk', 'nd': 'isi-Bhelarashiyani', 'ne': 'बेलारुसी', 'nl': 'Belarussisch', 'nmg': 'Kiɛl belarussie', 'nn': 'belarusisk', 'no': 'belarusisk', 'nqo': 'ߓߌߟߏߙߎ߳ߛߌߞߊ߲', 'nus': 'Thok bälärutha', 'nyn': 'Oruberarusi', 'om': 'Afaan Belarusia', 'or': 'ବେଲାରୁଷିଆନ୍', 'pa': 'ਬੇਲਾਰੂਸੀ', 'pa-Guru': 'ਬੇਲਾਰੂਸੀ', 'pcm': 'Bẹlarúsiá Lángwej', 'pl': 'białoruski', 'ps': 'بېلاروسي', 'pt': 'bielorrusso', 'qu': 'Bielorruso Simi', 'rm': 'bieloruss', 'rn': 'Ikibelarusiya', 'ro': 'belarusă', 'rof': 'Kibelarusi', 'ru': 'белорусский', 'rw': 'Ikibelarusiya', 'rwk': 'Kyibelarusi', 'sah': 'Бөлөрүүстүү', 'saq': 'Kibelarusi', 'sbp': 'Ishibelalusi', 'sc': 'bielorussu', 'sd': 'بيلاروسي', 'sd-Arab': 'بيلاروسي', 'se': 'vilges-ruoššagiella', 'seh': 'bielo-russo', 'ses': 'Belaruus senni', 'sg': 'Bielörûsi', 'shi': 'ⵜⴰⴱⵉⵍⴰⵔⵓⵙⵜ', 'shi-Latn': 'Tabilarust', 'shi-Tfng': 'ⵜⴰⴱⵉⵍⴰⵔⵓⵙⵜ', 'si': 'බෙලරුසියානු', 'sk': 'bieloruština', 'sl': 'beloruščina', 'smn': 'vielgisruošâkielâ', 'sn': 'chiBelarusi', 'so': 'Beleruusiyaan', 'sq': 'bjellorusisht', 'sr': 'белоруски', 'sr-Cyrl': 'белоруски', 'sr-Latn': 'beloruski', 'sv': 'belarusiska', 'sw': 'Kibelarusi', 'ta': 'பெலாருஷியன்', 'te': 'బెలారుషియన్', 'teo': 'Kibelarusi', 'tg': 'белорусӣ', 'th': 'เบลารุส', 'ti': 'ቤላሩስኛ', 'tk': 'belarus dili', 'to': 'lea fakapelalusi', 'tr': 'Belarusça', 'tt': 'белорус', 'twq': 'Belaruus senni', 'tzm': 'Tabilarusit', 'ug': 'بېلارۇسچە', 'uk': 'білоруська', 'ur': 'بیلاروسی', 'uz': 'belarus', 'uz-Cyrl': 'беларусча', 'uz-Latn': 'belarus', 'vai': 'ꔆꕞꖩꔻ', 'vai-Latn': 'Bhelarusaŋ', 'vai-Vaii': 'ꔆꕞꖩꔻ', 'vec': 'beloruso', 'vi': 'Tiếng Belarus', 'vun': 'Kyibelarusi', 'wae': 'Wísrussiš', 'wo': 'Belaris', 'xnr': 'बेलारूसी', 'xog': 'Olubelarusi', 'yav': 'pielúse', 'yi': 'בעלאַרוסיש', 'yo': 'Èdè Belarusi', 'yrl': 'bierurusu', 'yue': '白俄羅斯文', 'yue-Hans': '白俄罗斯文', 'yue-Hant': '白俄羅斯文', 'zgh': 'ⵜⴰⴱⵉⵍⴰⵔⵓⵙⵜ', 'zh': '白俄罗斯语', 'zh-Hans': '白俄罗斯语', 'zh-Hant': '白俄羅斯文', 'zu': 'isi-Belarusian'},
    'bea': {'en': 'Beaver'},
    'beb': {'en': 'Bebele'},
    'bec': {'en': 'Iceve-Maci'},
    'bed': {'en': 'Bedoanas'},
    'bee': {'en': 'Byangsi'},
    'bef': {'en': 'Benabena'},
    'beg': {'en': 'Belait'},
    'beh': {'en': 'Biali'},
    'bei': {'en': "Bekati'"},
    'bej': {'am': 'ቤጃ', 'ar': 'البيجا', 'ast': 'beja', 'az': 'beja', 'az-Latn': 'beja', 'bg': 'бея', 'bn': 'বেজা', 'br': 'bedawieg', 'brx': 'बेजा', 'bs': 'beja', 'bs-Cyrl': 'беја', 'bs-Latn': 'beja', 'ca': 'beja', 'ca-ES-valencia': 'beja', 'ccp': '𑄝𑄬𑄎', 'cs': 'bedža', 'cy': 'Bejäeg', 'da': 'beja', 'de': 'Bedauye', 'el': 'Μπέζα', 'el-polyton': 'Μπέζα', 'en': 'Beja', 'es': 'beja', 'et': 'bedža', 'fa': 'بجایی', 'fi': 'bedža', 'fr': 'bedja', 'fy': 'Beja', 'gd': 'Beja', 'gsw': 'Bedauye', 'gu': 'બેજા', 'he': 'בז׳ה', 'hi': 'बेजा', 'hi-Latn': 'Beja', 'hr': 'beja', 'ht': 'bedja', 'hu': 'bedzsa', 'id': 'Beja', 'is': 'beja', 'it': 'begia', 'ja': 'ベジャ語', 'ka': 'ბეჯა', 'kgp': 'meja', 'kn': 'ಬೇಜಾ', 'ko': '베자어', 'ks': 'بیجا', 'ks-Arab': 'بیجا', 'lb': 'Bedauye', 'lkt': 'Beja Iyápi', 'lo': 'ບີເຈ', 'lt': 'bėjų', 'lv': 'bedžu', 'mk': 'беџа', 'ml': 'ബേജ', 'mr': 'बेजा', 'ms': 'Beja', 'mt': 'Beja', 'nb': 'beja', 'ne': 'बेजा', 'nl': 'Beja', 'nn': 'beja', 'no': 'beja', 'or': 'ବେଜା', 'pl': 'bedża', 'pt': 'beja', 'rm': 'bedscha', 'ro': 'beja', 'ru': 'беджа', 'sk': 'bedža', 'sl': 'bedža', 'sr': 'беџа', 'sr-Cyrl': 'беџа', 'sr-Latn': 'bedža', 'sv': 'beja', 'sw': 'Kibeja', 'ta': 'பேஜா', 'te': 'బేజా', 'th': 'เบจา', 'to': 'lea fakapesa', 'tr': 'Beja dili', 'ug': 'بېجاچە', 'uk': 'беджа', 'vi': 'Tiếng Beja', 'yrl': 'beya', 'yue': '貝扎文', 'yue-Hans': '贝扎文', 'yue-Hant': '貝扎文', 'zh': '贝沙语', 'zh-Hans': '贝沙语', 'zh-Hant': '貝扎文'},
    'bek': {'en': 'Bebeli'},
    'bem': {'af': 'Bemba', 'am': 'ቤምባ', 'ar': 'البيمبا', 'as': 'বেম্বা', 'ast': 'bemba', 'az': 'bemba', 'az-Arab': 'بئمباجا', 'az-Cyrl': 'бемба', 'az-Latn': 'bemba', 'bal-Latn': 'Bembá', 'be': 'бемба', 'be-tarask': 'бемба', 'bem': 'Ichibemba', 'bg': 'бемба', 'bgn': 'بیمبا', 'bn': 'বেম্বা', 'br': 'bemba', 'brx': 'बेम्बा', 'bs': 'bemba', 'bs-Cyrl': 'бемба', 'bs-Latn': 'bemba', 'ca': 'bemba', 'ca-ES-valencia': 'bemba', 'ccp': '𑄝𑄬𑄟𑄴𑄝', 'ce': 'бемба', 'chr': 'ᏇᎹᏆ', 'ckb': 'بێمبا', 'cs': 'bembština', 'cy': 'Bembeg', 'da': 'bemba', 'de': 'Bemba', 'dsb': 'bemba', 'ee': 'bembagbe', 'el': 'Μπέμπα', 'el-polyton': 'Μπέμπα', 'en': 'Bemba', 'es': 'bemba', 'et': 'bemba', 'eu': 'bembera', 'fa': 'بمبایی', 'ff-Adlm': '𞤄𞤫𞤥𞤦𞤢𞥄𞤪𞤫', 'fi': 'bemba', 'fil': 'Bemba', 'fo': 'bemba', 'fr': 'bemba', 'fy': 'Bemba', 'ga': 'Beimbis', 'gd': 'Bemba', 'gl': 'bemba', 'gsw': 'Bemba', 'gu': 'બેમ્બા', 'ha': 'Bemba', 'he': 'במבה', 'hi': 'बेम्बा', 'hi-Latn': 'Bemba', 'hr': 'bemba', 'hsb': 'bemba', 'ht': 'bemba', 'hu': 'bemba', 'hy': 'բեմբա', 'ia': 'bemba', 'id': 'Bemba', 'ig': 'Bembà', 'is': 'bemba', 'it': 'wemba', 'ja': 'ベンバ語', 'jv': 'Bemba', 'ka': 'ბემბა', 'kea': 'bemba', 'kgp': 'mema', 'kk': 'бемба тілі', 'kk-Cyrl': 'бемба тілі', 'km': 'បេមបា', 'kn': 'ಬೆಂಬಾ', 'ko': '벰바어', 'kok': 'बेम्बा', 'kok-Deva': 'बेम्बा', 'ks': 'بیمبا', 'ks-Arab': 'بیمبا', 'ksh': 'Bemba', 'ku': 'bembayî', 'ky': 'бембача', 'lb': 'Bemba-Sprooch', 'lo': 'ເບັມບາ', 'lrc': 'بیما', 'lt': 'bembų', 'lv': 'bembu', 'mai': 'बेम्बा', 'mi': 'Pema', 'mk': 'бемба', 'ml': 'ബേംബ', 'mn': 'бемба', 'mr': 'बेम्बा', 'ms': 'Bemba', 'mt': 'Bemba', 'my': 'ဘိန်ဘာ', 'mzn': 'بمبایی', 'nb': 'bemba', 'ne': 'बेम्बा', 'nl': 'Bemba', 'nn': 'bemba', 'no': 'bemba', 'nqo': 'ߓߋ߲ߓߊߞߊ߲', 'or': 'ବେମ୍ବା', 'pa': 'ਬੇਂਬਾ', 'pa-Guru': 'ਬੇਂਬਾ', 'pcm': 'Bẹ́mba Lángwej', 'pl': 'bemba', 'ps': 'بيمبا', 'pt': 'bemba', 'qu': 'Bemba Simi', 'rm': 'bemba', 'ro': 'bemba', 'ru': 'бемба', 'sc': 'bemba', 'sd': 'بيمبا', 'sd-Arab': 'بيمبا', 'si': 'බෙම්බා', 'sk': 'bemba', 'sl': 'bemba', 'smn': 'bembakielâ', 'so': 'Bemba', 'sq': 'bembaisht', 'sr': 'бемба', 'sr-Cyrl': 'бемба', 'sr-Latn': 'bemba', 'sv': 'bemba', 'sw': 'Kibemba', 'ta': 'பெம்பா', 'te': 'బెంబా', 'tg': 'бемба', 'th': 'เบมบา', 'ti': 'ቤምባ', 'tk': 'bemba dili', 'to': 'lea fakapēmipa', 'tr': 'Bemba', 'tt': 'бемба', 'ug': 'بېمباچە', 'uk': 'бемба', 'ur': 'بیمبا', 'uz': 'bemba', 'uz-Cyrl': 'бемба', 'uz-Latn': 'bemba', 'vec': 'benba', 'vi': 'Tiếng Bemba', 'wo': 'Bemba', 'xnr': 'बेम्बा', 'yo': 'Èdè Béḿbà', 'yrl': 'bẽba', 'yue': '別姆巴文', 'yue-Hans': '别姆巴文', 'yue-Hant': '別姆巴文', 'zh': '本巴语', 'zh-Hans': '本巴语', 'zh-Hant': '別姆巴文', 'zu': 'isi-Bemba'},
    'beo': {'en': 'Beami'},
    'bep': {'en': 'Besoa'},
    'beq': {'en': 'Beembe'},
    'ber': {'en': 'Berber languages'},
    'ber-pro': {'en': 'Proto-Berber'},
    'bes': {'en': 'Besme'},
    'bet': {'en': 'Guiberoua Béte'},
    'beu': {'en': 'Blagar'},
    'bev': {'en': 'Daloa Bété'},
    'bew': {'am': 'ቤታዊ', 'ast': 'betawi', 'bal-Latn': 'Betawi', 'cs': 'batavština', 'de': 'Betawi', 'en': 'Betawi', 'et': 'betavi', 'fi': 'betawi', 'fr': 'betawi', 'gd': 'Betawi', 'hi': 'बेतावी', 'hi-Latn': 'Betawi', 'ht': 'betawi', 'ia': 'betawi', 'id': 'Betawi', 'ig': 'Betawi', 'it': 'betawi', 'ja': 'ベタウィ語', 'ku': 'betawî', 'lb': 'Betawi', 'lt': 'betavi', 'mk': 'бетавски', 'nb': 'betawi', 'ne': 'बेटावी', 'nl': 'Bataviaans', 'nn': 'betawi', 'no': 'betawi', 'nqo': 'ߓߋߕߊߥߌߞߊ߲', 'pl': 'betawi', 'sr': 'бетави', 'sr-Cyrl': 'бетави', 'sr-Latn': 'betavi', 'sv': 'betawiska', 'th': 'เบตาวี', 'ti': 'ቤታዊ', 'to': 'lea fakapetavi', 'tr': 'Betawi', 'uk': 'бетаві', 'vec': 'betawi', 'vi': 'Tiếng Betawi', 'yue': '貝塔維文', 'yue-Hans': '贝塔维文', 'yue-Hant': '貝塔維文', 'zh-Hant': '貝塔維文'},
    'bex': {'en': 'Jur Modo'},
    'bey': {'en': 'Beli (Papua New Guinea)'},
    'bez': {'af': 'Bena', 'am': 'ቤና', 'ar': 'بينا', 'as': 'বেনা', 'ast': 'bena', 'az': 'bena', 'az-Arab': 'بئناجا', 'az-Cyrl': 'бена', 'az-Latn': 'bena', 'bal-Latn': 'Bená', 'be': 'бена', 'be-tarask': 'бена', 'bez': 'Hibena', 'bg': 'бена', 'bgn': 'بینای', 'bn': 'বেনা', 'br': 'bena', 'brx': 'बेना', 'bs': 'bena', 'bs-Cyrl': 'бена', 'bs-Latn': 'bena', 'ca': 'bena', 'ca-ES-valencia': 'bena', 'ccp': '𑄝𑄬𑄚', 'ce': 'бена', 'chr': 'ᏇᎾ', 'ckb': 'بێنا', 'cs': 'bena', 'cy': 'Bena', 'da': 'bena', 'de': 'Bena', 'dsb': 'bena', 'ee': 'benagbe', 'el': 'Μπένα', 'el-polyton': 'Μπένα', 'en': 'Bena', 'es': 'bena', 'et': 'bena', 'eu': 'benera', 'fa': 'بنایی', 'ff-Adlm': '𞤄𞤫𞤲𞤢𞥄𞤪𞤫', 'fi': 'bena', 'fil': 'Bena', 'fo': 'bena', 'fr': 'bena', 'fy': 'Bena', 'ga': 'Beinis', 'gd': 'Bena', 'gl': 'bena', 'gu': 'બેના', 'ha': 'Bena', 'he': 'בנה', 'hi': 'बेना', 'hi-Latn': 'Bena', 'hr': 'bena', 'hsb': 'bena', 'ht': 'bena', 'hu': 'bena', 'hy': 'բենա', 'ia': 'bena', 'id': 'Bena', 'ig': 'Bena', 'is': 'bena', 'it': 'bena', 'ja': 'ベナ語', 'jv': 'Bena', 'ka': 'ბენა', 'kea': 'bena', 'kgp': 'menỹ', 'kk': 'бена тілі', 'kk-Cyrl': 'бена тілі', 'km': 'បេណា', 'kn': 'ಬೆನ', 'ko': '베나어', 'kok': 'बेना', 'kok-Deva': 'बेना', 'ksh': 'Bena', 'ku': 'benayî', 'ky': 'бенача', 'lb': 'Bena', 'lo': 'ບີນາ', 'lrc': 'بئنا', 'lt': 'benų', 'lv': 'bena', 'mai': 'बेना', 'mi': 'Pena', 'mk': 'бена', 'ml': 'ബെനാ', 'mn': 'бена', 'mr': 'बेना', 'ms': 'Bena', 'mt': 'Bena', 'my': 'ဘီနာ', 'mzn': 'بنایی', 'nb': 'bena', 'ne': 'बेना', 'nl': 'Bena', 'nn': 'bena', 'no': 'bena', 'nqo': 'ߓߋߣߊߞߊ߲', 'or': 'ବେନା', 'pa': 'ਬੇਨਾ', 'pa-Guru': 'ਬੇਨਾ', 'pcm': 'Bẹ́na Lángwej', 'pl': 'bena', 'ps': 'بينا', 'pt': 'bena', 'qu': 'Bena Simi', 'rm': 'bena', 'ro': 'bena', 'ru': 'бена', 'sc': 'bena', 'sd': 'بينا', 'sd-Arab': 'بينا', 'si': 'බෙනා', 'sk': 'bena', 'sl': 'benajščina', 'smn': 'benakielâ', 'so': 'Bena', 'sq': 'benaisht', 'sr': 'бена', 'sr-Cyrl': 'бена', 'sr-Latn': 'bena', 'sv': 'bena', 'sw': 'Kibena', 'ta': 'பெனா', 'te': 'బెనా', 'th': 'เบนา', 'ti': 'በና', 'tk': 'bena dili', 'to': 'lea fakapena', 'tr': 'Bena', 'ug': 'بېناچە', 'uk': 'бена', 'ur': 'بینا', 'uz': 'bena', 'uz-Cyrl': 'бенача', 'uz-Latn': 'bena', 'vec': 'bena', 'vi': 'Tiếng Bena', 'xnr': 'बेना', 'yo': 'Èdè Bẹ́nà', 'yrl': 'bena', 'yue': '貝納文', 'yue-Hans': '贝纳文', 'yue-Hant': '貝納文', 'zh': '贝纳语', 'zh-Hans': '贝纳语', 'zh-Hant': '貝納文', 'zu': 'isi-Bena'},
    'bfa': {'en': 'Bari'},
    'bfb': {'en': 'Pauri Bareli'},
    'bfc': {'en': 'Panyi Bai'},
    'bfd': {'am': 'ባፉት', 'ar': 'لغة البافوت', 'ast': 'bafut', 'bs': 'bafut', 'bs-Latn': 'bafut', 'ca': 'bafut', 'ca-ES-valencia': 'bafut', 'cs': 'bafut', 'cy': 'Baffwteg', 'da': 'bafut', 'de': 'Bafut', 'el': 'Μπαφούτ', 'el-polyton': 'Μπαφούτ', 'en': 'Bafut', 'es': 'bafut', 'et': 'bafuti', 'fi': 'fut', 'fr': 'bafut', 'fy': 'Bafut', 'gd': 'Bafut', 'he': 'באפוט', 'hi-Latn': 'Bafut', 'hr': 'bafut', 'ht': 'bafut', 'hu': 'bafut', 'id': 'Bafut', 'it': 'bafut', 'ja': 'バフット語', 'kgp': 'mafun', 'ko': '바푸트어', 'lb': 'Bafut', 'lo': 'ບາຟັດ', 'lt': 'bafutų', 'lv': 'bafutu', 'mk': 'бафут', 'ml': 'ബാഫട്ട്', 'ms': 'Bafut', 'nb': 'bafut', 'ne': 'बाफुट', 'nl': 'Bafut', 'nn': 'bafut', 'nnh': 'Shwóŋò pafud', 'no': 'bafut', 'pl': 'bafut', 'pt': 'bafut', 'ro': 'bafut', 'ru': 'бафут', 'sk': 'bafut', 'sv': 'bafut', 'sw': 'Kibafut', 'th': 'บาฟัต', 'to': 'lea fakapafuti', 'tr': 'Bafut', 'ug': 'بافۇتچە', 'uk': 'бафут', 'vi': 'Tiếng Bafut', 'yrl': 'bafuti', 'yue': '富特文', 'yue-Hans': '富特文', 'yue-Hant': '富特文', 'zh': '巴非特语', 'zh-Hans': '巴非特语', 'zh-Hant': '富特文'},
    'bfe': {'en': 'Betaf'},
    'bff': {'en': 'Bofi'},
    'bfg': {'en': 'Busang Kayan'},
    'bfh': {'en': 'Blafe'},
    'bfi': {'en': 'British Sign Language'},
    'bfj': {'en': 'Bafanji'},
    'bfk': {'en': 'Ban Khor Sign Language'},
    'bfl': {'en': 'Banda-Ndélé'},
    'bfm': {'en': 'Mmen'},
    'bfn': {'en': 'Bunak'},
    'bfo': {'en': 'Malba Birifor'},
    'bfp': {'en': 'Beba'},
    'bfq': {'am': 'ባዳጋ', 'ast': 'badaga', 'ca': 'badaga', 'ca-ES-valencia': 'badaga', 'cs': 'badagština', 'cy': 'Badaga', 'de': 'Badaga', 'en': 'Badaga', 'et': 'badaga', 'fi': 'badaga', 'fr': 'badaga', 'gd': 'Badaga', 'hi-Latn': 'Badaga', 'ht': 'badaga', 'it': 'badaga', 'ja': 'バダガ語', 'lb': 'Badaga', 'lt': 'badaga', 'mk': 'бадага', 'nb': 'badaga', 'ne': 'बडागा', 'nl': 'Badaga', 'nn': 'badaga', 'no': 'badaga', 'pl': 'badaga', 'sv': 'bagada', 'ta': 'படகா', 'th': 'พทคะ', 'to': 'lea fakapataka', 'tr': 'Badaga', 'uk': 'бадага', 'vi': 'Tiếng Badaga', 'yue': '巴達加文', 'yue-Hans': '巴达加文', 'yue-Hant': '巴達加文', 'zh-Hant': '巴達加文'},
    'bfr': {'en': 'Bazigar'},
    'bfs': {'en': 'Southern Bai'},
    'bft': {'en': 'Balti'},
    'bfu': {'en': 'Gahri'},
    'bfw': {'en': 'Bondo'},
    'bfx': {'en': 'Bantayanon'},
    'bfy': {'en': 'Bagheli'},
    'bfz': {'en': 'Mahasu Pahari'},
    'bg': {'af': 'Bulgaars', 'agq': 'Bùugɨlìa', 'ak': 'Bɔlgeria kasa', 'am': 'ቡልጋሪኛ', 'ar': 'البلغارية', 'as': 'বুলগেৰীয়', 'asa': 'Kibulgaria', 'ast': 'búlgaru', 'az': 'bolqar', 'az-Arab': 'بۇلقارجا', 'az-Cyrl': 'булгар', 'az-Latn': 'bolqar', 'bal-Latn': 'Bolgáriái', 'bas': 'Hɔp u bûlgâr', 'be': 'балгарская', 'be-tarask': 'балгарская', 'bem': 'Ichi Bulgariani', 'bez': 'Hibulgaria', 'bg': 'български', 'bgn': 'بلغاری', 'bm': 'buligarikan', 'bn': 'বুলগেরিয়', 'br': 'bulgareg', 'brx': 'बुल्गारियान', 'bs': 'bugarski', 'bs-Cyrl': 'бугарски', 'bs-Latn': 'bugarski', 'ca': 'búlgar', 'ca-ES-valencia': 'búlgar', 'ccp': '𑄝𑄪𑄣𑄴𑄉𑄬𑄢𑄨𑄠𑄧', 'ce': 'болгарийн', 'cgg': 'Oruburugariya', 'chr': 'ᏊᎵᎨᎵᎠᏂ', 'ckb': 'بۆلگاری', 'cs': 'bulharština', 'cy': 'Bwlgareg', 'da': 'bulgarsk', 'dav': 'Kibulgaria', 'de': 'Bulgarisch', 'dje': 'Bulagaari senni', 'dsb': 'bulgaršćina', 'dyo': 'bulgaari', 'dz': 'བཱལ་གེ་རི་ཡཱན་ཁ', 'ebu': 'Kĩbulgaria', 'ee': 'bulgariagbe', 'el': 'Βουλγαρικά', 'el-polyton': 'Βουλγαρικά', 'en': 'Bulgarian', 'eo': 'bulgara', 'es': 'búlgaro', 'et': 'bulgaaria', 'eu': 'bulgariera', 'ewo': 'Ǹkɔ́bɔ buləgárian', 'fa': 'بلغاری', 'ff': 'Bulgariire', 'ff-Adlm': '𞤄𞤭𞤤𞤺𞤢𞥄𞤪𞤫', 'ff-Latn': 'Bulgariire', 'fi': 'bulgaria', 'fil': 'Bulgarian', 'fo': 'bulgarskt', 'fr': 'bulgare', 'fur': 'bulgar', 'fy': 'Bulgaarsk', 'ga': 'Bulgáiris', 'gd': 'Bulgarais', 'gl': 'búlgaro', 'gsw': 'Bulgaarisch', 'gu': 'બલ્ગેરિયન', 'guz': 'Kibulgaria', 'ha': 'Bulgariyanci', 'he': 'בולגרית', 'hi': 'बुल्गारियाई', 'hi-Latn': 'Bulgarian', 'hr': 'bugarski', 'hsb': 'bołharšćina', 'ht': 'bulgare', 'hu': 'bolgár', 'hy': 'բուլղարերեն', 'ia': 'bulgaro', 'id': 'Bulgaria', 'ig': 'Bulgarian', 'is': 'búlgarska', 'it': 'bulgaro', 'ja': 'ブルガリア語', 'jmc': 'Kyibulgaryia', 'jv': 'Bulgaria', 'ka': 'ბულგარული', 'kab': 'Tabulgarit', 'kam': 'Kibulgaria', 'kde': 'Chibulgalia', 'kea': 'búlgaru', 'kgp': 'mugaru', 'khq': 'Bulagaari senni', 'ki': 'Kibulgaria', 'kk': 'болгар тілі', 'kk-Cyrl': 'болгар тілі', 'kln': 'kutitab Bulgaria', 'km': 'ប៊ុលហ្ការី', 'kn': 'ಬಲ್ಗೇರಿಯನ್', 'ko': '불가리아어', 'kok': 'बल्गेरियन', 'kok-Deva': 'बल्गेरियन', 'ks': 'بینا', 'ks-Arab': 'بینا', 'ksb': 'Kibulgalia', 'ksf': 'ribulgarí', 'ksh': 'Bulljahresch', 'ku': 'bulgarî', 'kxv': 'bulgeriati', 'kxv-Deva': 'बुल्गारियाति', 'kxv-Latn': 'bulgeriati', 'kxv-Orya': 'ବୁଲଗାରିଆନ୍', 'kxv-Telu': 'బుల్గారియన్', 'ky': 'болгарча', 'lag': 'Kɨbulugária', 'lb': 'Bulgaresch', 'lg': 'Lubulugariya', 'lkt': 'Bulgar Iyápi', 'lld': 'bulgar', 'ln': 'libiligali', 'lo': 'ບັງກາຣຽນ', 'lrc': 'بولغاری', 'lt': 'bulgarų', 'lu': 'Bulegari', 'luo': 'Kibulgaria', 'luy': 'Kibulgaria', 'lv': 'bulgāru', 'mai': 'बुलगेरियाई', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lBulgaria', 'mer': 'Kĩbulugĩria', 'mfe': 'bilgar', 'mg': 'Biolgara', 'mgh': 'Ibulgaria', 'mi': 'Purukāriana', 'mk': 'бугарски', 'ml': 'ബൾഗേറിയൻ', 'mn': 'болгар', 'mr': 'बल्गेरियन', 'ms': 'Bulgaria', 'mt': 'Bulgaru', 'mua': 'bulgaria', 'my': 'ဘူလ်ဂေးရီးယား', 'mzn': 'بلغاری', 'naq': 'Bulgariaǁî gowab', 'nb': 'bulgarsk', 'nd': 'isi-Bulgaria', 'ne': 'बुल्गेरियाली', 'nl': 'Bulgaars', 'nmg': 'Kiɛl bulgaria', 'nn': 'bulgarsk', 'no': 'bulgarsk', 'nqo': 'ߓߌߟߑߜ߭ߊߙߌߞߊ߲', 'nus': 'Thok bälga̱a̱riani', 'nyn': 'Oruburugariya', 'om': 'Afaan Bulgariya', 'or': 'ବୁଲଗେରିଆନ୍', 'os': 'болгайраг', 'pa': 'ਬੁਲਗਾਰੀਆਈ', 'pa-Guru': 'ਬੁਲਗਾਰੀਆਈ', 'pcm': 'Bọlgériá Lángwej', 'pl': 'bułgarski', 'ps': 'بلغاري', 'pt': 'búlgaro', 'qu': 'Bulgaro Simi', 'rm': 'bulgar', 'rn': 'Ikinyabuligariya', 'ro': 'bulgară', 'rof': 'Kibulgaria', 'ru': 'болгарский', 'rw': 'Urunyabuligariya', 'rwk': 'Kyibulgaryia', 'sah': 'Булҕаардыы', 'saq': 'Kibulgaria', 'sbp': 'Ishibulugalia', 'sc': 'bùlgaru', 'sd': 'بلغاريائي', 'sd-Arab': 'بلغاريائي', 'se': 'bulgáriagiella', 'seh': 'búlgaro', 'ses': 'Bulagaari senni', 'sg': 'Bulugäri', 'shi': 'ⵜⴰⴱⵍⵖⴰⵔⵉⵜ', 'shi-Latn': 'Tablɣarit', 'shi-Tfng': 'ⵜⴰⴱⵍⵖⴰⵔⵉⵜ', 'si': 'බල්ගේරියානු', 'sk': 'bulharčina', 'sl': 'bolgarščina', 'smn': 'bulgariakielâ', 'sn': 'chiBulgarian', 'so': 'Bulgeeriyaan', 'sq': 'bullgarisht', 'sr': 'бугарски', 'sr-Cyrl': 'бугарски', 'sr-Latn': 'bugarski', 'sv': 'bulgariska', 'sw': 'Kibulgaria', 'ta': 'பல்கேரியன்', 'te': 'బల్గేరియన్', 'teo': 'Kibulgaria', 'tg': 'булғорӣ', 'th': 'บัลแกเรีย', 'ti': 'ቡልጋርኛ', 'tk': 'bolgar dili', 'to': 'lea fakapulukalia', 'tr': 'Bulgarca', 'tt': 'болгар', 'twq': 'Bulagaari senni', 'tzm': 'Tabelɣarit', 'ug': 'بۇلغارچە', 'uk': 'болгарська', 'ur': 'بلغاری', 'uz': 'bolgar', 'uz-Cyrl': 'болгарча', 'uz-Latn': 'bolgar', 'vai': 'ꗂꔠꗸꘋ', 'vai-Latn': 'Bhɔgerɛŋ', 'vai-Vaii': 'ꗂꔠꗸꘋ', 'vec': 'bùlgaro', 'vi': 'Tiếng Bulgaria', 'vun': 'Kyibulgaryia', 'wae': 'Bulgariš', 'wo': 'Bilgaar', 'xnr': 'बुल्गारियाई', 'xog': 'Olubulugariya', 'yav': 'bulgálɛ', 'yi': 'בולגאַריש', 'yo': 'Èdè Bugaria', 'yrl': 'búgaru', 'yue': '保加利亞文', 'yue-Hans': '保加利亚文', 'yue-Hant': '保加利亞文', 'zgh': 'ⵜⴰⴱⵍⵖⴰⵔⵉⵜ', 'zh': '保加利亚语', 'zh-Hans': '保加利亚语', 'zh-Hant': '保加利亞文', 'zu': 'isi-Bulgari'},
    'bga': {'en': 'Gwamhi-Wuri'},
    'bgb': {'en': 'Bobongko'},
    'bgc': {'af': 'Haryanvi', 'ak': 'Harianvi', 'am': 'ሃርያንቪኛ', 'ar': 'الهارينفية', 'as': 'হাৰয়ানভি', 'az': 'Haryanvi', 'az-Latn': 'Haryanvi', 'bal-Latn': 'Haryánui', 'be': 'харыанві', 'be-tarask': 'харыанві', 'bg': 'харианви', 'bgc': 'हरियाणवी', 'bn': 'হরিয়ানভি', 'bs': 'harianvi', 'bs-Latn': 'harianvi', 'ca': 'haryanvi', 'ca-ES-valencia': 'haryanvi', 'chr': 'ᎭᏯᏅᎢ', 'cs': 'harijánština', 'cy': 'Haryanvi', 'da': 'harianvi', 'de': 'Haryanvi', 'dsb': 'haryanvišćina', 'el': 'Χαργιάνβι', 'el-polyton': 'Χαργιάνβι', 'en': 'Haryanvi', 'es': 'haryanvi', 'et': 'harjaanvi', 'eu': 'haryanera', 'fa': 'هارایاناوی', 'ff-Adlm': '𞤖𞤢𞤪𞤴𞤢𞤲𞤾𞤭𞥅𞤪𞤫', 'fi': 'haryanvi', 'fil': 'Haryanvi', 'fo': 'haryanvi', 'fr': 'haryanvi', 'ga': 'Haryanvi', 'gd': 'Haryanvi', 'gl': 'hariani', 'gu': 'હરિયાણવી', 'ha': 'Haryanvi', 'he': 'הריאנבי', 'hi': 'हरियाणवी', 'hi-Latn': 'Hariyaanvi', 'hr': 'haryanvi', 'hsb': 'haryanvi', 'ht': 'haryanvi', 'hu': 'haryanvi', 'hy': 'հարյանվի', 'ia': 'haryanvi', 'id': 'Haryanvi', 'ig': 'Haryanvi', 'is': 'haryanví', 'it': 'haryanvi', 'ja': 'ハリヤーンウィー語', 'jv': 'Haryanvi', 'ka': 'ჰარიანვი', 'kk': 'хариани тілі', 'kk-Cyrl': 'хариани тілі', 'km': 'ហារីយ៉ាន់វី', 'kn': 'ಹರ್ಯಾನ್ವಿ', 'ko': '하리안비어', 'kok': 'हरयाणवी', 'kok-Deva': 'हरयाणवी', 'ku': 'haryanvîyî', 'ky': 'харьянвиче', 'lo': 'ຮາຢານວີ', 'lt': 'harijanvi', 'lv': 'harjanvi', 'mi': 'Herianawhi', 'mk': 'харијанви', 'ml': 'ഹർയാൻവി', 'mn': 'харьянви', 'mr': 'हरियाणवी', 'ms': 'Haryanvi', 'my': 'ဟာယန်ဗီ', 'nb': 'haryanvi', 'ne': 'हरयाणवी', 'nl': 'Haryanvi', 'nn': 'haryanvi', 'no': 'haryanvi', 'om': 'Haryanvi', 'or': 'ହରିୟାନଭି', 'pa': 'ਹਰਿਆਣਵੀ', 'pa-Guru': 'ਹਰਿਆਣਵੀ', 'pcm': 'Haryanvi Lángwej', 'pl': 'haryanvi', 'ps': 'هریانوی', 'pt': 'hariani', 'qu': 'Haryanvi', 'ro': 'haryanvi', 'ru': 'харианви', 'sc': 'haryanvi', 'sd': 'ھريانوي', 'sd-Arab': 'ھريانوي', 'si': 'හර්යාන්වි', 'sk': 'haryanvi', 'sl': 'harjanščina', 'so': 'Haryanvi', 'sq': 'harjanvisht', 'sr': 'харијански', 'sr-Cyrl': 'харијански', 'sr-Latn': 'harijanski', 'sv': 'hariyanvi', 'sw': 'Kiharyanvi', 'ta': 'ஹரியான்வி', 'te': 'హర్యాన్వి', 'th': 'หริยนวี', 'ti': 'ሃርያንቪ', 'tk': 'harýanwi dili', 'to': 'lea fakahalaiānivi', 'tr': 'Haryanvi dili', 'uk': 'харʼянві', 'ur': 'ہریانوی', 'uz': 'harianvi', 'uz-Latn': 'harianvi', 'vec': 'harianvi', 'vi': 'Tiếng Haryana', 'yo': 'Èdè Haryanvi', 'yue': '哈里亞納文', 'yue-Hans': '哈里亚纳文', 'yue-Hant': '哈里亞納文', 'zh': '哈里亚纳语', 'zh-Hans': '哈里亚纳语', 'zh-Hant': '哈里亞納文', 'zu': 'isi-Haryanvi'},
    'bgd': {'en': 'Rathwi Bareli'},
    'bge': {'en': 'Bauria'},
    'bgf': {'en': 'Bangandu'},
    'bgg': {'en': 'Bugun'},
    'bgi': {'en': 'Giangan'},
    'bgj': {'en': 'Bangolan'},
    'bgk': {'en': 'Bit'},
    'bgl': {'en': 'Bo (Laos)'},
    'bgm': {'en': 'Baga Mboteni'},
    'bgn': {'af': 'Wes-Balochi', 'am': 'የምዕራብ ባሎቺ', 'ar': 'البلوشية الغربية', 'ast': 'balochi occidental', 'az': 'qərbi bəluc', 'az-Arab': 'دوْغو بلوچجا', 'az-Latn': 'qərbi bəluc', 'bal-Latn': 'Balóchi (Róbarkati)', 'be': 'заходняя белуджская', 'be-tarask': 'заходняя белуджская', 'bg': 'западен балочи', 'bgn': 'بلوچی (رخشانی)', 'bn': 'পশ্চিম বালোচি', 'br': 'baloutchi ar Cʼhornôg', 'bs': 'zapadni belučki', 'bs-Latn': 'zapadni belučki', 'ca': 'balutxi occidental', 'ca-ES-valencia': 'balutxi occidental', 'ccp': '𑄛𑄧𑄏𑄨𑄟𑄴 𑄝𑄣𑄮𑄌𑄨', 'ce': 'малхбузен-белуджийн', 'cs': 'balúčština (západní)', 'cy': 'Balochi Gorllewinol', 'da': 'vestbaluchi', 'de': 'Westliches Belutschi', 'el': 'Δυτικά Μπαλοχικά', 'el-polyton': 'Δυτικά Μπαλοχικά', 'en': 'Western Balochi', 'es': 'baluchi occidental', 'et': 'läänebelutši', 'fa': 'بلوچی غربی', 'fi': 'länsibelutši', 'fil': 'Kanlurang Balochi', 'fo': 'vestur balochi', 'fr': 'baloutchi occidental', 'gd': 'Balochi Shiarach', 'gl': 'baluchi occidental', 'gu': 'પશ્ચિમી બાલોચી', 'he': 'באלוצ׳י מערבית', 'hi': 'पश्चिमी बलोची', 'hi-Latn': 'Western Balochi', 'hr': 'zapadnobaludžijski', 'ht': 'baloutchi occidental', 'hu': 'nyugati beludzs', 'hy': 'արևմտաբելուջիերեն', 'ia': 'baluchi occidental', 'id': 'Balochi Barat', 'ig': 'Western Balochi', 'is': 'vesturbalotsí', 'it': 'beluci occidentale', 'ja': '西バローチー語', 'ka': 'დასავლეთ ბელუჯი', 'kgp': 'marusi-rãpurja', 'kk': 'батыс балучи тілі', 'kk-Cyrl': 'батыс балучи тілі', 'km': 'បាឡូជីខាងលិច', 'kn': 'ಪಶ್ಚಿಮ ಬಲೊಚಿ', 'ko': '서부 발로치어', 'ku': 'belucîya rojavayî', 'ky': 'чыгыш балучиче', 'lo': 'ບາໂລຈີ ພາກຕາເວັນຕົກ', 'lrc': 'بألوٙچی أقتوٙنئشین', 'lt': 'vakarų beludžių', 'lv': 'rietumbeludžu', 'mk': 'западен балочи', 'ml': 'പശ്ചിമ ബലൂചി', 'mr': 'पश्चिमी बालोची', 'ms': 'Balochi Barat', 'my': 'အနောက် ဘဲလိုချီ', 'mzn': 'غربی بلوچی', 'nb': 'vestbalutsji', 'ne': 'पश्चिम बालोची', 'nl': 'Westers Beloetsji', 'nn': 'vestbalutsji', 'no': 'vestbalutsji', 'pa': 'ਪੱਛਮੀ ਬਲੂਚੀ', 'pa-Guru': 'ਪੱਛਮੀ ਬਲੂਚੀ', 'pl': 'beludżi północny', 'pt': 'balúchi ocidental', 'ro': 'baluchi occidentală', 'ru': 'западный белуджский', 'si': 'බටහිර බලොචි', 'sk': 'západná balúčtina', 'sl': 'zahodnobalučijščina', 'sq': 'balokishte perëndimore', 'sr': 'западни белучки', 'sr-Cyrl': 'западни белучки', 'sr-Latn': 'zapadni belučki', 'sv': 'västbaluchiska', 'sw': 'Kibalochi cha Magharibi', 'ta': 'மேற்கு பலோச்சி', 'te': 'పశ్చిమ బలూచీ', 'th': 'บาลูจิตะวันตก', 'ti': 'ምዕራባዊ ባሎቺ', 'to': 'lea fakapalusi-hihifo', 'tr': 'Batı Balochi', 'uk': 'східнобелуджійська', 'ur': 'مغربی بلوچی', 'uz': 'g‘arbiy baluj', 'uz-Latn': 'g‘arbiy baluj', 'vec': 'baluci osidentale', 'vi': 'Tiếng Tây Balochi', 'xnr': 'पश्चिमी बलोची', 'yrl': 'baruxi usidẽtawara', 'yue': '西俾路支文', 'yue-Hans': '西俾路支文', 'yue-Hant': '西俾路支文', 'zh': '西俾路支语', 'zh-Hans': '西俾路支语', 'zh-Hant': '西俾路支文', 'zu': 'isi-Western Balochi'},
    'bgo': {'en': 'Baga Koga'},
    'bgp': {'en': 'Eastern Balochi'},
    'bgq': {'en': 'Bagri'},
    'bgr': {'en': 'Bawm Chin'},
    'bgs': {'en': 'Tagabawa'},
    'bgt': {'en': 'Bughotu'},
    'bgu': {'en': 'Mbongno'},
    'bgv': {'en': 'Warkay-Bipim'},
    'bgw': {'en': 'Bhatri'},
    'bgx': {'en': 'Balkan Gagauz Turkish'},
    'bgy': {'en': 'Benggoi'},
    'bgz': {'en': 'Banggai'},
    'bh': {'en': 'Bihari languages'},
    'bha': {'en': 'Bharia'},
    'bhb': {'en': 'Bhili'},
    'bhc': {'en': 'Biga'},
    'bhd': {'en': 'Bhadrawahi'},
    'bhe': {'en': 'Bhaya'},
    'bhf': {'en': 'Odiai'},
    'bhg': {'en': 'Binandere'},
    'bhh': {'en': 'Bukharic'},
    'bhi': {'en': 'Bhilali'},
    'bhj': {'en': 'Bahing'},
    'bhk': {'en': 'Albay Bicolano'},
    'bhl': {'en': 'Bimin'},
    'bhm': {'en': 'Bathari'},
    'bhn': {'en': 'Bohtan Neo-Aramaic'},
    'bho': {'af': 'Bhojpuri', 'ak': 'Bopuri', 'am': 'ቦጅፑሪ', 'ar': 'البهوجبورية', 'as': 'ভোজপুৰী', 'ast': 'bhojpuri', 'az': 'bxoçpuri', 'az-Cyrl': 'бхочпури', 'az-Latn': 'bxoçpuri', 'bal-Latn': 'Bójpuri', 'be': 'бхаджпуры', 'be-tarask': 'бхаджпуры', 'bg': 'боджпури', 'bgn': 'بهوجپوری', 'bho': 'भोजपुरी', 'bn': 'ভোজপুরি', 'br': 'bhojpuri', 'brx': 'भजपुरी', 'bs': 'bojpuri', 'bs-Cyrl': 'бојпури', 'bs-Latn': 'bojpuri', 'ca': 'bhojpuri', 'ca-ES-valencia': 'bhojpuri', 'ccp': '𑄞𑄮𑄎𑄴𑄛𑄪𑄢𑄨', 'ce': 'бходжпури', 'chr': 'ᏉᏣᏊᎵ', 'ckb': 'بوجپووری', 'cs': 'bhódžpurština', 'cy': 'Bhojpuri', 'da': 'bhojpuri', 'de': 'Bhodschpuri', 'dsb': 'bhojpurišćina', 'el': 'Μπότζπουρι', 'el-polyton': 'Μπότζπουρι', 'en': 'Bhojpuri', 'es': 'bhoyapurí', 'et': 'bhodžpuri', 'eu': 'bhojpurera', 'fa': 'بوجپوری', 'ff-Adlm': '𞤄𞤮𞤧𞤨𞤵𞤪𞤭𞥅𞤪𞤫', 'fi': 'bhodžpuri', 'fil': 'Bhojpuri', 'fo': 'bhojpuri', 'fr': 'bhodjpouri', 'fy': 'Bhojpuri', 'ga': 'Vóispiris', 'gd': 'Bhojpuri', 'gl': 'bhojpuri', 'gsw': 'Bhodschpuri', 'gu': 'ભોજપુરી', 'ha': 'Bhojpuri', 'he': 'בוג׳פורי', 'hi': 'भोजपुरी', 'hi-Latn': 'Bhojpuri', 'hr': 'bhojpuri', 'hsb': 'bhojpurišćina', 'ht': 'bhodjpouri', 'hu': 'bodzspuri', 'hy': 'բհոպուրի', 'ia': 'bhojpuri', 'id': 'Bhojpuri', 'ig': 'Bojpuri', 'is': 'bojpúrí', 'it': 'bhojpuri', 'ja': 'ボージュプリー語', 'jv': 'Bhojpuri', 'ka': 'ბოჯპური', 'kgp': 'mojpuri', 'kk': 'бходжпури тілі', 'kk-Cyrl': 'бходжпури тілі', 'km': 'បូចពូរី', 'kn': 'ಭೋಜಪುರಿ', 'ko': '호즈푸리어', 'kok': 'भोजपुरी', 'kok-Deva': 'भोजपुरी', 'ks': 'بوجپوٗری', 'ks-Arab': 'بوجپوٗری', 'ksh': 'Bhohdschpuresch', 'ku': 'bojpûrî', 'ky': 'бхожпуриче', 'lb': 'Bhodschpuri', 'lo': 'ໂບພູຣິ', 'lt': 'baučpuri', 'lv': 'bhodžpūru', 'mai': 'भोजपुरी', 'mi': 'Pōhipuri', 'mk': 'боџпури', 'ml': 'ഭോജ്\u200cപുരി', 'mn': 'божпури', 'mr': 'भोजपुरी', 'ms': 'Bhojpuri', 'mt': 'Bhojpuri', 'my': 'ဘို့ဂျ်ပူရီ', 'nb': 'bhojpuri', 'ne': 'भोजपुरी', 'nl': 'Bhojpuri', 'nn': 'bhojpuri', 'no': 'bhojpuri', 'om': 'Bihoojpuurii', 'or': 'ଭୋଜପୁରୀ', 'pa': 'ਭੋਜਪੁਰੀ', 'pa-Guru': 'ਭੋਜਪੁਰੀ', 'pcm': 'Bhojpúri', 'pl': 'bhodżpuri', 'ps': 'بهوجپوري', 'pt': 'bhojpuri', 'qu': 'Bhojpuri', 'rm': 'bhojpuri', 'ro': 'bhojpuri', 'ru': 'бходжпури', 'sc': 'bhojpuri', 'sd': 'ڀوجپوري', 'sd-Arab': 'ڀوجپوري', 'si': 'බොජ්පුරි', 'sk': 'bhódžpurčina', 'sl': 'bodžpuri', 'smn': 'bhožpurikielâ', 'so': 'Bhojpuri', 'sq': 'boxhpurisht', 'sr': 'боџпури', 'sr-Cyrl': 'боџпури', 'sr-Latn': 'bodžpuri', 'sv': 'bhojpuri', 'sw': 'Kibhojpuri', 'ta': 'போஜ்பூரி', 'te': 'భోజ్\u200cపురి', 'th': 'โภชปุรี', 'ti': 'ቦጅፑሪ', 'tk': 'bhojpuri dili', 'to': 'lea fakaposipuli', 'tr': 'Arayanice', 'ug': 'بوجپۇرىچە', 'uk': 'бходжпурі', 'ur': 'بھوجپوری', 'uz': 'bxojpuri', 'uz-Cyrl': 'бхожпури', 'uz-Latn': 'bxojpuri', 'vec': 'bojpuri', 'vi': 'Tiếng Bhojpuri', 'xnr': 'भोजपुरी', 'yo': 'Èdè Bojuri', 'yrl': 'buyipuri', 'yue': '博傑普爾文', 'yue-Hans': '博杰普尔文', 'yue-Hant': '博傑普爾文', 'zh': '博杰普尔语', 'zh-Hans': '博杰普尔语', 'zh-Hant': '博傑普爾文', 'zu': 'isi-Bhojpuri'},
    'bhp': {'en': 'Bima'},
    'bhq': {'en': 'Tukang Besi South'},
    'bhr': {'en': 'Bara Malagasy'},
    'bhs': {'en': 'Buwal'},
    'bht': {'en': 'Bhattiyali'},
    'bhu': {'en': 'Bhunjia'},
    'bhv': {'en': 'Bahau'},
    'bhw': {'en': 'Biak'},
    'bhx': {'en': 'Bhalay'},
    'bhy': {'en': 'Bhele'},
    'bhz': {'en': 'Bada (Indonesia)'},
    'bi': {'af': 'Bislama', 'am': 'ቢስላምኛ', 'ar': 'البيسلامية', 'as': 'বিছলামা', 'ast': 'bislama', 'az': 'bislama', 'az-Cyrl': 'бислама', 'az-Latn': 'bislama', 'be': 'біслама', 'be-tarask': 'біслама', 'bg': 'бислама', 'bgn': 'بیسلامه', 'bn': 'বিসলামা', 'br': 'bislama', 'brx': 'बिस्लामा', 'bs': 'bislama', 'bs-Cyrl': 'бислама', 'bs-Latn': 'bislama', 'ca': 'bislama', 'ca-ES-valencia': 'bislama', 'ccp': '𑄝𑄨𑄥𑄴𑄣𑄟', 'ce': 'бислама', 'chr': 'ᏈᏍᎳᎹ', 'ckb': 'بیسلاما', 'cs': 'bislamština', 'cy': 'Bislama', 'da': 'bislama', 'de': 'Bislama', 'dsb': 'bislamšćina', 'el': 'Μπισλάμα', 'el-polyton': 'Μπισλάμα', 'en': 'Bislama', 'eo': 'bislamo', 'es': 'bislama', 'et': 'bislama', 'eu': 'bislama', 'fa': 'بیسلاما', 'ff-Adlm': '𞤄𞤭𞤧𞤤𞤢𞤥𞤢𞥄𞤪𞤫', 'fi': 'bislama', 'fil': 'Bislama', 'fo': 'bislama', 'fr': 'bichelamar', 'fy': 'Bislama', 'ga': 'Bioslaimis', 'gd': 'Bislama', 'gl': 'bislama', 'gsw': 'Bislama', 'gu': 'બિસ્લામા', 'ha': 'Bislama', 'he': 'ביסלמה', 'hi': 'बिस्लामा', 'hi-Latn': 'Bislama', 'hr': 'bislama', 'hsb': 'bislamšćina', 'ht': 'bichelamar', 'hu': 'bislama', 'hy': 'բիսլամա', 'ia': 'bislama', 'id': 'Bislama', 'ig': 'Bislama', 'is': 'bíslama', 'it': 'bislama', 'ja': 'ビスラマ語', 'jv': 'Bislama', 'ka': 'ბისლამა', 'kgp': 'miramá', 'kk': 'бислама тілі', 'kk-Cyrl': 'бислама тілі', 'km': 'ប៊ីស្លាម៉ា', 'kn': 'ಬಿಸ್ಲಾಮಾ', 'ko': '비슬라마어', 'kok': 'बिसलमा', 'kok-Deva': 'बिसलमा', 'ks': 'بِسلاما', 'ks-Arab': 'بِسلاما', 'ksh': 'Bislahmesch', 'ku': 'bîslamayî', 'ky': 'бисламача', 'lb': 'Bislama', 'lo': 'ບິສລະມາ', 'lt': 'bislama', 'lv': 'bišlamā', 'mai': 'बिसलामा', 'mi': 'Pihirāma', 'mk': 'бислама', 'ml': 'ബിസ്\u200cലാമ', 'mn': 'бислам', 'mr': 'बिस्लामा', 'ms': 'Bislama', 'mt': 'Bislama', 'my': 'ဘစ်စ်လာမာ', 'nb': 'bislama', 'ne': 'बिस्लाम', 'nl': 'Bislama', 'nn': 'bislama', 'no': 'bislama', 'or': 'ବିସଲାମା', 'pa': 'ਬਿਸਲਾਮਾ', 'pa-Guru': 'ਬਿਸਲਾਮਾ', 'pcm': 'Bisláma', 'pl': 'bislama', 'ps': 'بسلاما', 'pt': 'bislamá', 'qu': 'Bislama', 'rm': 'bislama', 'ro': 'bislama', 'ru': 'бислама', 'sc': 'bislama', 'sd': 'بسلاما', 'sd-Arab': 'بسلاما', 'si': 'බිස්ලමා', 'sk': 'bislama', 'sl': 'bislamščina', 'smn': 'bislama', 'so': 'U dhashay Bislam', 'sq': 'bislamisht', 'sr': 'бислама', 'sr-Cyrl': 'бислама', 'sr-Latn': 'bislama', 'sv': 'bislama', 'sw': 'Kibislama', 'ta': 'பிஸ்லாமா', 'te': 'బిస్లామా', 'th': 'บิสลามา', 'ti': 'ቢስላማ', 'tk': 'bislama dili', 'to': 'lea fakapisilama', 'tr': 'Bislama', 'ug': 'بىسلاماچە', 'uk': 'біслама', 'ur': 'بسلاما', 'uz': 'bislama', 'uz-Cyrl': 'бислама', 'uz-Latn': 'bislama', 'vec': 'bizlama', 'vi': 'Tiếng Bislama', 'xnr': 'बिस्लामा', 'yo': 'Èdè Bisilama', 'yrl': 'bisiramá', 'yue': '比斯拉馬文', 'yue-Hans': '比斯拉马文', 'yue-Hant': '比斯拉馬文', 'zh': '比斯拉马语', 'zh-Hans': '比斯拉马语', 'zh-Hant': '比斯拉馬文', 'zu': 'isi-Bislama'},
    'bia': {'en': 'Badimaya'},
    'bib': {'en': 'Bissa'},
    'bic': {'en': 'Bikaru'},
    'bid': {'en': 'Bidiyo'},
    'bie': {'en': 'Bepour'},
    'bif': {'en': 'Biafada'},
    'big': {'en': 'Biangai'},
    'bij': {'en': 'Vaghat-Ya-Bijim-Legeri'},
    'bik': {'am': 'ቢኮል', 'ar': 'البيكولية', 'ast': 'bikol', 'az': 'bikol', 'az-Latn': 'bikol', 'bg': 'биколски', 'bn': 'বিকোল', 'br': 'bikol', 'brx': 'बिकोल्', 'bs': 'bikol', 'bs-Cyrl': 'бикол', 'bs-Latn': 'bikol', 'ca': 'bicol', 'ca-ES-valencia': 'bicol', 'ccp': '𑄝𑄨𑄇𑄮𑄣𑄴', 'cs': 'bikolština', 'da': 'bikol', 'de': 'Bikol', 'el': 'Μπικόλ', 'el-polyton': 'Μπικόλ', 'en': 'Bikol', 'es': 'bicol', 'et': 'bikoli', 'fa': 'بیکولی', 'fi': 'bikol', 'fr': 'bikol', 'fy': 'Bikol', 'gd': 'Bikol', 'gsw': 'Bikolisch', 'gu': 'બિકોલ', 'he': 'ביקול', 'hi': 'बिकोल', 'hi-Latn': 'Bikol', 'hr': 'bikolski', 'ht': 'bikol', 'hu': 'bikol', 'id': 'Bikol', 'is': 'bíkol', 'it': 'bicol', 'ja': 'ビコル語', 'kgp': 'mikor', 'kn': 'ಬಿಕೊಲ್', 'ko': '비콜어', 'ks': 'بِکول', 'ks-Arab': 'بِکول', 'lb': 'Bikol-Sprooch', 'lo': 'ບີຄອນ', 'lt': 'bikolų', 'lv': 'bikolu', 'mk': 'биколски', 'ml': 'ബികോൽ', 'mr': 'बिकोल', 'mt': 'Bikol', 'nb': 'bikol', 'ne': 'बिकोल', 'nl': 'Bikol', 'nn': 'bikol', 'no': 'bikol', 'or': 'ବିକୋଲ୍', 'pl': 'bikol', 'pt': 'bikol', 'rm': 'bikol', 'ro': 'bikol', 'ru': 'бикольский', 'sk': 'bikolčina', 'sl': 'bikolski jezik', 'sr': 'бикол', 'sr-Cyrl': 'бикол', 'sr-Latn': 'bikol', 'sv': 'bikol', 'ta': 'பிகோல்', 'te': 'బికోల్', 'th': 'บิกอล', 'to': 'lea fakapikoli', 'tr': 'Bikol', 'ug': 'بىكولچە', 'uk': 'бікольська', 'vi': 'Tiếng Bikol', 'yrl': 'bikú', 'yue': '比科爾文', 'yue-Hans': '比科尔文', 'yue-Hant': '比科爾文', 'zh': '比科尔语', 'zh-Hans': '比科尔语', 'zh-Hant': '比科爾文'},
    'bil': {'en': 'Bile'},
    'bim': {'en': 'Bimoba'},
    'bin': {'af': 'Bini', 'am': 'ቢኒ', 'ar': 'البينية', 'as': 'বিনি', 'ast': 'bini', 'az': 'bini', 'az-Cyrl': 'бини', 'az-Latn': 'bini', 'be': 'эда', 'be-tarask': 'эда', 'bg': 'бини', 'bgn': 'بینی', 'bn': 'বিনি', 'br': 'bini', 'brx': 'बिनी', 'bs': 'bini', 'bs-Cyrl': 'бини', 'bs-Latn': 'bini', 'ca': 'edo', 'ca-ES-valencia': 'edo', 'ccp': '𑄝𑄨𑄚𑄨', 'ce': 'бини', 'chr': 'ᏈᏂ', 'ckb': 'بینی', 'cs': 'bini', 'cy': 'Bini', 'da': 'bini', 'de': 'Bini', 'dsb': 'binišćina', 'el': 'Μπίνι', 'el-polyton': 'Μπίνι', 'en': 'Bini', 'es': 'bini', 'et': 'edo', 'eu': 'edoera', 'fa': 'بینی', 'ff-Adlm': '𞤄𞤭𞤲𞤭𞥅𞤪𞤫', 'fi': 'bini', 'fil': 'Bini', 'fo': 'bini', 'fr': 'bini', 'fy': 'Bini', 'ga': 'Binis', 'gd': 'Edo', 'gl': 'bini', 'gsw': 'Bini', 'gu': 'બિની', 'ha': 'Bini', 'he': 'ביני', 'hi': 'बिनी', 'hi-Latn': 'Bini', 'hr': 'bini', 'hsb': 'binišćina', 'ht': 'bini', 'hu': 'bini', 'hy': 'բինի', 'ia': 'bini', 'id': 'Bini', 'ig': 'Bini', 'is': 'bíní', 'it': 'bini', 'ja': 'ビニ語', 'jv': 'Bini', 'ka': 'ბინი', 'kgp': 'minĩ', 'kk': 'бини тілі', 'kk-Cyrl': 'бини тілі', 'km': 'ប៊ីនី', 'kn': 'ಬಿನಿ', 'ko': '비니어', 'kok': 'बिनी', 'kok-Deva': 'बिनी', 'ks': 'بِنی', 'ks-Arab': 'بِنی', 'ksh': 'Edo', 'ku': 'bînîyî', 'ky': 'биниче', 'lb': 'Bini-Sprooch', 'lo': 'ບີນີ', 'lt': 'bini', 'lv': 'binu', 'mai': 'बिनी', 'mi': 'Pini', 'mk': 'бини', 'ml': 'ബിനി', 'mn': 'бини', 'mr': 'बिनी', 'ms': 'Bini', 'mt': 'Bini', 'my': 'ဘီနီ', 'nb': 'bini', 'ne': 'बिनी', 'nl': 'Bini', 'nn': 'bini', 'no': 'bini', 'or': 'ବିନି', 'pa': 'ਬਿਨੀ', 'pa-Guru': 'ਬਿਨੀ', 'pcm': 'Biní', 'pl': 'bini', 'ps': 'بینی', 'pt': 'bini', 'qu': 'Bini', 'rm': 'bini', 'ro': 'bini', 'ru': 'бини', 'sc': 'bini', 'sd': 'بني', 'sd-Arab': 'بني', 'si': 'බිනි', 'sk': 'bini', 'sl': 'edo', 'smn': 'binikielâ', 'so': 'U dhashay Bin', 'sq': 'binisht', 'sr': 'бини', 'sr-Cyrl': 'бини', 'sr-Latn': 'bini', 'sv': 'bini', 'sw': 'Kibini', 'ta': 'பினி', 'te': 'బిని', 'th': 'บินี', 'ti': 'ቢኒ', 'tk': 'bini dili', 'to': 'lea fakapini', 'tr': 'Bini', 'ug': 'بىنىچە', 'uk': 'біні', 'ur': 'بینی', 'uz': 'bini', 'uz-Cyrl': 'бини', 'uz-Latn': 'bini', 'vec': 'bini', 'vi': 'Tiếng Bini', 'xnr': 'बिनी', 'yo': 'Èdè Bini', 'yrl': 'biní', 'yue': '比尼文', 'yue-Hans': '比尼文', 'yue-Hant': '比尼文', 'zh': '比尼语', 'zh-Hans': '比尼语', 'zh-Hant': '比尼文', 'zu': 'isi-Bini'},
    'bio': {'en': 'Nai'},
    'bip': {'en': 'Bila'},
    'biq': {'en': 'Bipi'},
    'bir': {'en': 'Bisorio'},
    'bit': {'en': 'Berinomo'},
    'biu': {'en': 'Biete'},
    'biv': {'en': 'Southern Birifor'},
    'biw': {'en': 'Kol (Cameroon)'},
    'bix': {'en': 'Bijori'},
    'biy': {'en': 'Birhor'},
    'biz': {'en': 'Baloi'},
    'bja': {'en': 'Budza'},
    'bjb': {'en': 'Banggarla'},
    'bjc': {'en': 'Bariji'},
    'bjd': {'en': 'Bandjigali'},
    'bje': {'en': 'Biao-Jiao Mien'},
    'bjf': {'en': 'Barzani Jewish Neo-Aramaic'},
    'bjg': {'en': 'Bidyogo'},
    'bjh': {'en': 'Bahinemo'},
    'bji': {'en': 'Burji'},
    'bjj': {'en': 'Kanauji'},
    'bjk': {'en': 'Barok'},
    'bjl': {'en': 'Bulu (Papua New Guinea)'},
    'bjm': {'en': 'Bajelani'},
    'bjn': {'am': 'ባንጃር', 'ast': 'banjar', 'cs': 'bandžarština', 'de': 'Banjaresisch', 'en': 'Banjar', 'et': 'bandžari', 'fi': 'banjar', 'fr': 'banjar', 'gd': 'Banjar', 'hi-Latn': 'Banjar', 'ht': 'banjar', 'id': 'Banjar', 'it': 'banjar', 'ja': 'バンジャル語', 'lb': 'Banjaresesch', 'lt': 'bandžarų', 'mk': 'банџарски', 'nb': 'banjar', 'ne': 'बन्जार', 'nl': 'Banjar', 'nn': 'banjar', 'no': 'banjar', 'pl': 'banjar', 'sv': 'banjariska', 'th': 'บันจาร์', 'to': 'lea fakapanisali', 'tr': 'Banjar Dili', 'uk': 'банджарська', 'vi': 'Tiếng Banjar', 'yue': '班亞爾文', 'yue-Hans': '班亚尔文', 'yue-Hant': '班亞爾文', 'zh-Hant': '班亞爾文'},
    'bjo': {'en': 'Mid-Southern Banda'},
    'bjp': {'en': 'Fanamaket'},
    'bjq': {'en': 'Southern Betsimisaraka Malagasy'},
    'bjr': {'en': 'Binumarien'},
    'bjs': {'en': 'Bajan'},
    'bjt': {'en': 'Balanta-Ganja'},
    'bju': {'en': 'Busuu'},
    'bjv': {'en': 'Bedjond'},
    'bjw': {'en': 'Bakwé'},
    'bjx': {'en': 'Banao Itneg'},
    'bjy': {'en': 'Bayali'},
    'bjz': {'en': 'Baruga'},
    'bka': {'en': 'Kyak'},
    'bkb': {'en': 'Finallig'},
    'bkc': {'en': 'Baka (Cameroon)'},
    'bkd': {'en': 'Binukid'},
    'bkf': {'en': 'Beeke'},
    'bkg': {'en': 'Buraka'},
    'bkh': {'en': 'Bakoko'},
    'bki': {'en': 'Baki'},
    'bkj': {'en': 'Pande'},
    'bkk': {'en': 'Brokskat'},
    'bkl': {'en': 'Berik'},
    'bkm': {'ar': 'لغة الكوم', 'ast': 'kom', 'bs': 'kom', 'bs-Latn': 'kom', 'ca': 'kom', 'ca-ES-valencia': 'kom', 'cs': 'kom', 'cy': 'Comeg', 'da': 'kom', 'de': 'Kom', 'el': 'Κομ', 'el-polyton': 'Κομ', 'en': 'Kom', 'es': 'kom', 'et': 'komi (Aafrika)', 'fi': 'kom', 'fr': 'kom', 'fy': 'Kom', 'gd': 'Kom', 'he': 'קום', 'hi-Latn': 'Kom', 'hr': 'kom', 'ht': 'kom', 'hu': 'kom', 'id': 'Kom', 'it': 'kom', 'ja': 'コム語', 'kgp': 'kãm', 'ko': '콤어', 'lb': 'Kom', 'lo': 'ກົມ', 'lt': 'komų', 'lv': 'komu', 'mk': 'ком', 'ml': 'കോം', 'ms': 'Kom', 'nb': 'kom', 'ne': 'कोम', 'nl': 'Kom', 'nn': 'kom', 'nnh': 'Shwóŋò pʉ̀a njinikom', 'no': 'kom', 'pl': 'kom', 'pt': 'kom', 'ro': 'kom', 'ru': 'ком', 'sk': 'kom', 'sv': 'bamekon', 'sw': 'Kikom', 'th': 'กม', 'to': 'lea fakakome', 'tr': 'Kom', 'ug': 'كومچە', 'uk': 'ком', 'vi': 'Tiếng Kom', 'yrl': 'kũ', 'yue': '康姆文', 'yue-Hans': '康姆文', 'yue-Hant': '康姆文', 'zh': '科姆语', 'zh-Hans': '科姆语', 'zh-Hant': '康姆文'},
    'bkn': {'en': 'Bukitan'},
    'bko': {'en': "Kwa'"},
    'bkp': {'en': 'Boko (Democratic Republic of Congo)'},
    'bkq': {'en': 'Bakairí'},
    'bkr': {'en': 'Bakumpai'},
    'bks': {'en': 'Northern Sorsoganon'},
    'bkt': {'en': 'Boloki'},
    'bku': {'en': 'Buhid'},
    'bkv': {'en': 'Bekwarra'},
    'bkw': {'en': 'Bekwel'},
    'bkx': {'en': 'Baikeno'},
    'bky': {'en': 'Bokyi'},
    'bkz': {'en': 'Bungku'},
    'bla': {'af': 'Siksika', 'am': 'ሲክሲካ', 'ar': 'السيكسيكية', 'as': 'ছিক্সিকা', 'ast': 'siksika', 'az': 'siksikə', 'az-Cyrl': 'сиксикә', 'az-Latn': 'siksikə', 'be': 'блэкфут', 'be-tarask': 'блэкфут', 'bg': 'сиксика', 'bgn': 'سیکسیکا', 'bn': 'সিকসিকা', 'br': 'siksika', 'brx': 'सीकसीका', 'bs': 'siksika', 'bs-Cyrl': 'сисика', 'bs-Latn': 'siksika', 'ca': 'blackfoot', 'ca-ES-valencia': 'blackfoot', 'ccp': '𑄥𑄨𑄇𑄴𑄥𑄨𑄇', 'ce': 'сиксика', 'chr': 'ᏏᎩᏏᎧ', 'ckb': 'سیکسیکا', 'cs': 'siksika', 'cy': 'Siksika', 'da': 'siksika', 'de': 'Blackfoot', 'dsb': 'siksikášćina', 'el': 'Σικσίκα', 'el-polyton': 'Σικσίκα', 'en': 'Siksiká', 'es': 'siksika', 'et': 'mustjalaindiaani', 'eu': 'siksikera', 'fa': 'سیکسیکا', 'ff-Adlm': '𞤅𞤭𞤳𞤧𞤭𞤳𞤢𞥄𞤪𞤫', 'fi': 'siksika', 'fil': 'Siksika', 'fo': 'siksika', 'fr': 'siksika', 'fy': 'Siksika', 'ga': 'Sicsicis', 'gd': 'Siksika', 'gl': 'siksiká', 'gsw': 'Blackfoot-Schpraach', 'gu': 'સિક્સિકા', 'ha': 'Siksiká', 'he': 'סיקסיקה', 'hi': 'सिक्सिका', 'hi-Latn': 'Siksika', 'hr': 'siksika', 'hsb': 'siksikášćina', 'ht': 'siksika', 'hu': 'siksika', 'hy': 'սիկսիկա', 'ia': 'siksika', 'id': 'Siksika', 'ig': 'Siksikà', 'is': 'siksika', 'it': 'siksika', 'ja': 'シクシカ語', 'jv': 'Siksiká', 'ka': 'სიკსიკა', 'kgp': 'sigsika', 'kk': 'сиксика тілі', 'kk-Cyrl': 'сиксика тілі', 'km': 'ស៊ីកស៊ីកា', 'kn': 'ಸಿಕ್ಸಿಕಾ', 'ko': '식시카어', 'kok': 'सिकसिका', 'kok-Deva': 'सिकसिका', 'ks': 'سِکسِکا', 'ks-Arab': 'سِکسِکا', 'ku': 'blakfotî', 'ky': 'сиксикача', 'lb': 'Blackfoot-Sprooch', 'lo': 'ຊິກຊິກາ', 'lt': 'siksikų', 'lv': 'siksiku', 'mai': 'सिकसिका', 'mi': 'Hihika', 'mk': 'сиксика', 'ml': 'സിക്സിക', 'mn': 'сиксика', 'mr': 'सिक्सिका', 'ms': 'Siksika', 'mt': 'Siksika', 'my': 'စစ္စီကာ', 'nb': 'siksika', 'ne': 'सिक्सिका', 'nl': 'Siksika', 'nn': 'siksika', 'no': 'siksika', 'or': 'ସିକସିକା', 'pa': 'ਸਿਕਸਿਕਾ', 'pa-Guru': 'ਸਿਕਸਿਕਾ', 'pcm': 'Siksíká Lángwej', 'pl': 'siksika', 'ps': 'سکسيکا', 'pt': 'siksika', 'qu': 'Siksiká Simi', 'rm': 'siksika', 'ro': 'siksika', 'ru': 'сиксика', 'sc': 'pees nieddos', 'sd': 'سڪسڪا', 'sd-Arab': 'سڪسڪا', 'si': 'සික්සිකා', 'sk': 'siksika', 'sl': 'siksika', 'smn': 'siksikakielâ', 'so': 'Siksiká', 'sq': 'siksikaisht', 'sr': 'сисика', 'sr-Cyrl': 'сисика', 'sr-Latn': 'sisika', 'sv': 'siksika', 'sw': 'Kisiksika', 'ta': 'சிக்சிகா', 'te': 'సిక్సికా', 'th': 'สิกสิกา', 'ti': 'ሲክሲካ', 'tk': 'siksika dili', 'to': 'lea fakasikesikā', 'tr': 'Karaayak dili', 'ug': 'سىكسىكاچە', 'uk': 'сіксіка', 'ur': 'سکسیکا', 'uz': 'siksika', 'uz-Latn': 'siksika', 'vec': 'siksika', 'vi': 'Tiếng Siksika', 'xnr': 'सिक्सिका', 'yo': 'Èdè Sikiska', 'yrl': 'sikisika', 'yue': '錫克錫卡文', 'yue-Hans': '锡克锡卡文', 'yue-Hant': '錫克錫卡文', 'zh': '西克西卡语', 'zh-Hans': '西克西卡语', 'zh-Hant': '錫克錫卡文', 'zu': 'isi-Siksika'},
    'blb': {'en': 'Bilua'},
    'blc': {'en': 'Bella Coola'},
    'bld': {'en': 'Bolango'},
    'ble': {'en': 'Balanta-Kentohe'},
    'blf': {'en': 'Buol'},
    'blg': {'en': 'Balau'},
    'blh': {'en': 'Kuwaa'},
    'bli': {'en': 'Bolia'},
    'blj': {'en': 'Bolongan'},
    'blk': {'en': "Pa'o Karen"},
    'bll': {'en': 'Biloxi'},
    'blm': {'en': 'Beli (South Sudan)'},
    'bln': {'en': 'Southern Catanduanes Bikol'},
    'blo': {'af': 'Anii', 'ak': 'Anii', 'am': 'አኒኛ', 'ar': 'الآنية', 'as': 'আনি', 'az': 'anii', 'az-Latn': 'anii', 'bal-Latn': 'Ani', 'be': 'аніі', 'be-tarask': 'аніі', 'bg': 'ании', 'blo': 'anii kagɩja', 'bn': 'অ্যানি', 'bs': 'anii', 'bs-Latn': 'anii', 'ca': 'anii', 'ca-ES-valencia': 'anii', 'chr': 'ᎠᏂᎵ', 'cs': 'aniiština', 'cy': 'Anii', 'da': 'anii', 'de': 'Anii', 'dsb': 'aniišćina', 'el': 'Ανίι', 'el-polyton': 'Ανίι', 'en': 'Anii', 'es': 'anii', 'et': 'anii', 'eu': 'aniiera', 'fa': 'باسیلا', 'fi': 'anii', 'fil': 'Anii', 'fo': 'anii', 'fr': 'anii', 'ga': 'Anii', 'gd': 'Anii', 'gl': 'anii', 'gu': 'અની', 'ha': 'Anii', 'he': 'אני', 'hi': 'अनी', 'hi-Latn': 'Anii', 'hr': 'anii', 'hsb': 'aniišćina', 'ht': 'anii', 'hu': 'anii', 'hy': 'անիի', 'ia': 'anii', 'id': 'Anii', 'ig': 'Anii', 'is': 'anii', 'it': 'anii', 'ja': 'アニ語 (blo)', 'jv': 'Anii', 'ka': 'ანიი', 'kk': 'ании тілі', 'kk-Cyrl': 'ании тілі', 'km': 'អានី', 'kn': 'ಅನೀ', 'ko': '아니이어', 'kok': 'अनी', 'kok-Deva': 'अनी', 'ku': 'bloyî', 'ky': 'анииче', 'lo': 'ອານີ', 'lt': 'guanų', 'lv': 'anī', 'mk': 'ании', 'ml': 'അനി', 'mn': 'Ани', 'mr': 'ॲनीआय', 'ms': 'Anii', 'my': 'အန်နီ', 'nb': 'anii', 'ne': 'अनी', 'nl': 'Anii', 'nn': 'anii', 'no': 'anii', 'nqo': 'ߊߣߌ߯ߞߊ߲', 'om': 'Anii', 'or': 'ବ୍ଲୋ', 'pa': 'ਅਨੀ', 'pa-Guru': 'ਅਨੀ', 'pcm': 'Anii Lángwej', 'pl': 'anii', 'ps': 'انۍ', 'pt': 'anii', 'ro': 'anii', 'ru': 'ании', 'sc': 'anii', 'sd': 'آنيائي', 'sd-Arab': 'آنيائي', 'si': 'අනී', 'sk': 'anii', 'sl': 'anii', 'so': 'Anii', 'sq': 'anisht', 'sr': 'ании', 'sr-Cyrl': 'ании', 'sr-Latn': 'anii', 'sv': 'anii', 'sw': 'Kianii', 'ta': 'அனீ', 'te': 'అని', 'th': 'อานี', 'ti': 'ኣኒ', 'tk': 'blo dili', 'tr': 'Aniice', 'uk': 'анії', 'ur': 'عانی', 'uz': 'Anii', 'uz-Latn': 'Anii', 'vec': 'anii', 'vi': 'Anii', 'yo': 'Anii', 'yue': '阿尼文', 'yue-Hans': '阿尼文', 'yue-Hant': '阿尼文', 'zh': '阿尼语', 'zh-Hans': '阿尼语', 'zh-Hant': '阿尼文', 'zu': 'isi-Anii'},
    'blp': {'en': 'Blablanga'},
    'blq': {'en': 'Baluan-Pam'},
    'blr': {'en': 'Blang'},
    'bls': {'en': 'Balaesang'},
    'blt': {'bal-Latn': 'Tái Dam', 'blt': 'ꪼꪕꪒꪾ', 'en': 'Tai Dam', 'et': 'tai-dami', 'fi': 'tai dam', 'gd': 'Tai Dam', 'hi-Latn': 'Tai Dam', 'ia': 'tai dam', 'ig': 'Tai Dam', 'ku': 'tay dam', 'ti': 'ታይ ዳም', 'vec': 'tai dam'},
    'blv': {'en': 'Kibala'},
    'blw': {'en': 'Balangao'},
    'blx': {'en': 'Mag-Indi Ayta'},
    'bly': {'en': 'Notre'},
    'blz': {'en': 'Balantak'},
    'bm': {'af': 'Bambara', 'am': 'ባምባርኛ', 'ar': 'البامبارا', 'as': 'বামবাৰা', 'ast': 'bambara', 'az': 'bambara', 'az-Arab': 'بامباراجا', 'az-Cyrl': 'бамбара', 'az-Latn': 'bambara', 'bal-Latn': 'Bambará', 'be': 'бамбара', 'be-tarask': 'бамбара', 'bg': 'бамбара', 'bgn': 'بمبارای', 'bm': 'bamanakan', 'bn': 'বামবারা', 'br': 'bambara', 'brx': 'बाम्बारा', 'bs': 'bambara', 'bs-Cyrl': 'бамбара', 'bs-Latn': 'bambara', 'ca': 'bambara', 'ca-ES-valencia': 'bambara', 'ccp': '𑄝𑄟𑄴𑄝𑄢', 'ce': 'бамбара', 'chr': 'ᏆᎻᏆᎳ', 'ckb': 'بامبارا', 'cs': 'bambarština', 'cy': 'Bambareg', 'da': 'bambara', 'de': 'Bambara', 'dsb': 'bambara', 'ee': 'bambaragbe', 'el': 'Μπαμπάρα', 'el-polyton': 'Μπαμπάρα', 'en': 'Bambara', 'es': 'bambara', 'et': 'bambara', 'eu': 'bambarera', 'fa': 'بامبارایی', 'ff-Adlm': '𞤄𞤢𞤥𞤦𞤢𞤪𞤢𞥄𞤪𞤫', 'fi': 'bambara', 'fil': 'Bambara', 'fo': 'bambara', 'fr': 'bambara', 'fy': 'Bambara', 'ga': 'Bambairis', 'gd': 'Bambara', 'gl': 'bambara', 'gsw': 'Bambara', 'gu': 'બામ્બારા', 'ha': 'Bambara', 'he': 'במבארה', 'hi': 'बाम्बारा', 'hi-Latn': 'Bambara', 'hr': 'bambara', 'hsb': 'bambara', 'ht': 'bambara', 'hu': 'bambara', 'hy': 'բամբարա', 'ia': 'bambara', 'id': 'Bambara', 'ig': 'Bambara', 'is': 'bambara', 'it': 'bambara', 'ja': 'バンバラ語', 'jv': 'Bambara', 'ka': 'ბამბარა', 'kea': 'bambara', 'kgp': 'mámara', 'kk': 'бамбара тілі', 'kk-Cyrl': 'бамбара тілі', 'km': 'បាម្បារា', 'kn': 'ಬಂಬಾರಾ', 'ko': '밤바라어', 'kok': 'बंबारा', 'kok-Deva': 'बंबारा', 'ks': 'بَمبارا', 'ks-Arab': 'بَمبارا', 'ksh': 'Bambara', 'ku': 'bambarayî', 'ky': 'бамбарача', 'lb': 'Bambara-Sprooch', 'lld': 'bambara', 'lo': 'ບາມບາຣາ', 'lrc': 'بامبارا', 'lt': 'bambarų', 'lv': 'bambaru', 'mai': 'बमबारा', 'mi': 'Pāpara', 'mk': 'бамбара', 'ml': 'ബംബാറ', 'mn': 'бамбара', 'mr': 'बाम्बारा', 'ms': 'Bambara', 'mt': 'Bambara', 'my': 'ဘန်ဘာရာ', 'mzn': 'بامبارایی', 'nb': 'bambara', 'ne': 'बाम्बारा', 'nl': 'Bambara', 'nn': 'bambara', 'no': 'bambara', 'nqo': 'ߒߞߏ ߡߊߟߌ', 'or': 'ବାମ୍ବାରା', 'pa': 'ਬੰਬਾਰਾ', 'pa-Guru': 'ਬੰਬਾਰਾ', 'pcm': 'Bambára Lángwej', 'pl': 'bambara', 'ps': 'بمبارا', 'pt': 'bambara', 'qu': 'Bambara Simi', 'rm': 'bambara', 'ro': 'bambara', 'ru': 'бамбара', 'sc': 'bambara', 'sd': 'بمبارا', 'sd-Arab': 'بمبارا', 'si': 'බම්බරා', 'sk': 'bambarčina', 'sl': 'bambarščina', 'smn': 'bambarakielâ', 'so': 'Bambaara', 'sq': 'bambarisht', 'sr': 'бамбара', 'sr-Cyrl': 'бамбара', 'sr-Latn': 'bambara', 'sv': 'bambara', 'sw': 'Kibambara', 'ta': 'பம்பாரா', 'te': 'బంబారా', 'th': 'บัมบารา', 'ti': 'ባምባራ', 'tk': 'bamana', 'to': 'lea fakapamipala', 'tr': 'Bambara', 'ug': 'بامباراچە', 'uk': 'бамбара', 'ur': 'بمبارا', 'uz': 'bambara', 'uz-Cyrl': 'бамбарча', 'uz-Latn': 'bambara', 'vec': 'banbara', 'vi': 'Tiếng Bambara', 'xnr': 'बाम्बारा', 'yo': 'Èdè Báḿbàrà', 'yrl': 'bãbara', 'yue': '班巴拉文', 'yue-Hans': '班巴拉文', 'yue-Hant': '班巴拉文', 'zh': '班巴拉语', 'zh-Hans': '班巴拉语', 'zh-Hant': '班巴拉文', 'zu': 'isi-Bambara'},
    'bma': {'en': 'Lame'},
    'bmb': {'en': 'Bembe'},
    'bmc': {'en': 'Biem'},
    'bmd': {'en': 'Baga Manduri'},
    'bme': {'en': 'Limassa'},
    'bmf': {'en': 'Bom-Kim'},
    'bmg': {'en': 'Bamwe'},
    'bmh': {'en': 'Kein'},
    'bmi': {'en': 'Bagirmi'},
    'bmj': {'en': 'Bote-Majhi'},
    'bmk': {'en': 'Ghayavi'},
    'bml': {'en': 'Bomboli'},
    'bmm': {'en': 'Northern Betsimisaraka Malagasy'},
    'bmn': {'en': 'Bina (Papua New Guinea)'},
    'bmo': {'en': 'Bambalang'},
    'bmp': {'en': 'Bulgebi'},
    'bmq': {'en': 'Bomu'},
    'bmr': {'en': 'Muinane'},
    'bms': {'en': 'Bilma Kanuri'},
    'bmt': {'en': 'Biao Mon'},
    'bmu': {'en': 'Somba-Siawari'},
    'bmv': {'en': 'Bum'},
    'bmw': {'en': 'Bomwali'},
    'bmx': {'en': 'Baimak'},
    'bmy': {'en': 'Bemba (Democratic Republic of Congo)'},
    'bmz': {'en': 'Baramu'},
    'bn': {'af': 'Bengaals', 'agq': 'Bɨ̀ŋgalì', 'ak': 'Bengali kasa', 'am': 'ቤንጋሊኛ', 'ar': 'البنغالية', 'as': 'বাংলা', 'asa': 'Kibangla', 'ast': 'bengalín', 'az': 'benqal', 'az-Arab': 'بنگالجا', 'az-Cyrl': 'бенгал', 'az-Latn': 'benqal', 'bal-Latn': 'Bangáli', 'bas': 'Hɔp u bɛŋgàli', 'be': 'бенгальская', 'be-tarask': 'бенгальская', 'bem': 'Ichi Bengali', 'bez': 'Hibangla', 'bg': 'бенгалски', 'bgn': 'بنگالی', 'blo': 'baŋglaa kagɩja', 'bm': 'bɛngalikan', 'bn': 'বাংলা', 'br': 'bengali', 'brx': 'बांला', 'bs': 'bengalski', 'bs-Cyrl': 'бенгалски', 'bs-Latn': 'bengalski', 'ca': 'bengalí', 'ca-ES-valencia': 'bengalí', 'ccp': '𑄝𑄁𑄣', 'ce': 'бенгалийн', 'ceb': 'Bangla', 'cgg': 'Orubengari', 'chr': 'ᏇᏂᎦᎳ', 'ckb': 'بەنگلادێشی', 'cs': 'bengálština', 'cv': 'бенгал', 'cy': 'Bengaleg', 'da': 'bengali', 'dav': 'Kibangla', 'de': 'Bengalisch', 'dje': 'Bengali senni', 'dsb': 'bengalšćina', 'dyo': 'bengali', 'dz': 'བངྒ་ལ་ཁ', 'ebu': 'Kĩbangla', 'ee': 'Bengaligbe', 'el': 'Βεγγαλικά', 'el-polyton': 'Βεγγαλικά', 'en': 'Bangla', 'eo': 'bengala', 'es': 'bengalí', 'et': 'bengali', 'eu': 'bengalera', 'ewo': 'Ǹkɔ́bɔ bɛngalí', 'fa': 'بنگالی', 'ff': 'Bengali', 'ff-Adlm': '𞤄𞤫𞤲𞤺𞤢𞤤𞤭𞥅𞤪𞤫', 'ff-Latn': 'Bengali', 'fi': 'bengali', 'fil': 'Bangla', 'fo': 'bangla', 'fr': 'bengali', 'fur': 'bengalês', 'fy': 'Bengaalsk', 'ga': 'Beangáilis', 'gd': 'Bangla', 'gl': 'bengalí', 'gsw': 'Bengalisch', 'gu': 'બાંગ્લા', 'guz': 'Kibangla', 'ha': 'Bengali', 'he': 'בנגלית', 'hi': 'बंगाली', 'hi-Latn': 'Bangla', 'hnj': '𞄜𞄤', 'hnj-Hmnp': '𞄜𞄤', 'hr': 'bangla', 'hsb': 'bengalšćina', 'ht': 'bengali', 'hu': 'bangla', 'hy': 'բենգալերեն', 'ia': 'bengalese', 'id': 'Bengali', 'ig': 'Bangla', 'is': 'bengalska', 'it': 'bengalese', 'ja': 'ベンガル語', 'jmc': 'Kyibangla', 'jv': 'Bengali', 'ka': 'ბენგალური', 'kab': 'Tabengalit', 'kam': 'Kibangla', 'kde': 'Chibangla', 'kea': 'bengali', 'kgp': 'megari', 'khq': 'Bengali senni', 'ki': 'Kibangla', 'kk': 'бенгал тілі', 'kk-Cyrl': 'бенгал тілі', 'kln': 'kutitab Bengali', 'km': 'បង់ក្លាដែស', 'kn': 'ಬಾಂಗ್ಲಾ', 'ko': '벵골어', 'kok': 'बांग्ला', 'kok-Deva': 'बांग्ला', 'ks': 'بَنگٲلؠ', 'ks-Arab': 'بَنگٲلؠ', 'ksb': 'Kibangla', 'ksf': 'ribɛngáli', 'ksh': 'Bängjahlesch', 'ku': 'bengalî', 'kxv': 'bangali', 'kxv-Deva': 'बंगाली', 'kxv-Latn': 'bangali', 'kxv-Orya': 'ବଙ୍ଗାଲି', 'kxv-Telu': 'బంగ్లా', 'ky': 'бангладешче', 'lag': 'Kɨbangála', 'lb': 'Bengalesch', 'lg': 'Lubengali', 'lkt': 'Bengali Iyápi', 'lld': 'bengalesc', 'ln': 'libengali', 'lo': 'ເບັງກາລີ', 'lrc': 'بأنگالی', 'lt': 'bengalų', 'lu': 'Bengali', 'luo': 'Kibangla', 'luy': 'Kibangla', 'lv': 'bengāļu', 'mai': 'बंगाली', 'mas': 'lnkʉtʉ́k ɔ́ɔ̄ lBengali', 'mer': 'Kĩbangira', 'mfe': 'bengali', 'mg': 'Bengali', 'mgh': 'Ibangla', 'mi': 'Pākara', 'mk': 'бенгалски', 'ml': 'ബംഗ്ലാ', 'mn': 'бенгал', 'mni': 'বাংলা', 'mni-Beng': 'বাংলা', 'mr': 'बंगाली', 'ms': 'Benggali', 'mt': 'Bengali', 'mua': 'bengalia', 'my': 'ဘင်္ဂါလီ', 'mzn': 'بنگالی', 'naq': 'Bengaliǁî gowab', 'nb': 'bengali', 'nd': 'isi-Bhengali', 'ne': 'बंगाली', 'nl': 'Bengaals', 'nmg': 'Kiɛl bengalia', 'nn': 'bengali', 'no': 'bengali', 'nqo': 'ߓߍ߲ߜ߭ߊߟߌߞߊ߲', 'nus': 'Thok bängali', 'nyn': 'Orubengari', 'om': 'Afaan Baangladeshi', 'or': 'ବଙ୍ଗଳା', 'pa': 'ਬੰਗਾਲੀ', 'pa-Guru': 'ਬੰਗਾਲੀ', 'pcm': 'Bángla Lángwej', 'pl': 'bengalski', 'ps': 'بنگالي', 'pt': 'bengali', 'qu': 'Bangla Simi', 'rm': 'bengal', 'rn': 'Ikibengali', 'ro': 'bengaleză', 'rof': 'Kibangla', 'ru': 'бенгальский', 'rw': 'Ikibengali', 'rwk': 'Kyibangla', 'sah': 'Бенгаллыы', 'saq': 'Kibangla', 'sbp': 'Ishibangila', 'sc': 'bengalesu', 'sd': 'بنگلا', 'sd-Arab': 'بنگلا', 'se': 'bengalgiella', 'seh': 'bengali', 'ses': 'Bengali senni', 'sg': 'Bengäli', 'shi': 'ⵜⴰⴱⵏⵖⴰⵍⵉⵜ', 'shi-Latn': 'Tabnɣalit', 'shi-Tfng': 'ⵜⴰⴱⵏⵖⴰⵍⵉⵜ', 'si': 'බෙංගාලි', 'sk': 'bengálčina', 'sl': 'bengalščina', 'smn': 'banglakielâ', 'sn': 'chiBengali', 'so': 'Bangladesh', 'sq': 'bengalisht', 'sr': 'бенгалски', 'sr-Cyrl': 'бенгалски', 'sr-Latn': 'bengalski', 'sv': 'bengali', 'sw': 'Kibengali', 'syr': 'ܒܢܓܐܠܝܐ', 'ta': 'வங்காளம்', 'te': 'బంగ్లా', 'teo': 'Kibangla', 'tg': 'бинғолӣ', 'th': 'บังกลา', 'ti': 'በንጋሊ', 'tk': 'bengal dili', 'to': 'lea fakapāngilā', 'tr': 'Bengalce', 'tt': 'бенгали', 'twq': 'Bengali senni', 'tzm': 'Tabinɣalit', 'ug': 'بېنگالچە', 'uk': 'бенгальська', 'ur': 'بنگلہ', 'uz': 'bengal', 'uz-Cyrl': 'бенгалча', 'uz-Latn': 'bengal', 'vai': 'ꗩꕭꔷ', 'vai-Latn': 'Bhɛŋgáli', 'vai-Vaii': 'ꗩꕭꔷ', 'vec': 'bangla', 'vi': 'Tiếng Bangla', 'vun': 'Kyibangla', 'wae': 'Bengališ', 'wo': 'Baŋla', 'xh': 'IsiBangla', 'xnr': 'बंगाली', 'xog': 'Olubengali', 'yav': 'pengálɛ́ɛ', 'yi': 'בענגאַליש', 'yo': 'Èdè Bengali', 'yrl': 'bẽgari', 'yue': '孟加拉文', 'yue-Hans': '孟加拉文', 'yue-Hant': '孟加拉文', 'zgh': 'ⵜⴰⴱⵏⵖⴰⵍⵉⵜ', 'zh': '孟加拉语', 'zh-Hans': '孟加拉语', 'zh-Hant': '孟加拉文', 'zu': 'isi-Bengali'},
    'bna': {'en': 'Bonerate'},
    'bnb': {'en': 'Bookan'},
    'bnc': {'en': 'Bontok'},
    'bnd': {'en': 'Banda (Indonesia)'},
    'bne': {'en': 'Bintauna'},
    'bnf': {'en': 'Masiwang'},
    'bng': {'en': 'Benga'},
    'bni': {'en': 'Bangi'},
    'bnj': {'en': 'Eastern Tawbuid'},
    'bnk': {'en': 'Bierebo'},
    'bnl': {'en': 'Boon'},
    'bnm': {'en': 'Batanga'},
    'bnn': {'en': 'Bunun'},
    'bno': {'en': 'Bantoanon'},
    'bnp': {'en': 'Bola'},
    'bnq': {'en': 'Bantik'},
    'bnr': {'en': 'Butmas-Tur'},
    'bns': {'en': 'Bundeli'},
    'bnt': {'en': 'Bantu languages'},
    'bnt-ngu-pro': {'en': 'Proto-Nguni'},
    'bnt-phu': {'en': 'Phuthi'},
    'bnt-pro': {'en': 'Proto-Bantu'},
    'bnu': {'en': 'Bentong'},
    'bnv': {'en': 'Bonerif'},
    'bnw': {'en': 'Bisis'},
    'bnx': {'en': 'Bangubangu'},
    'bny': {'en': 'Bintulu'},
    'bnz': {'en': 'Beezen'},
    'bo': {'af': 'Tibettaans', 'am': 'ቲቤታንኛ', 'ar': 'التبتية', 'as': 'তিব্বতী', 'ast': 'tibetanu', 'az': 'tibet', 'az-Cyrl': 'тибет', 'az-Latn': 'tibet', 'bal-Latn': 'Tebbati', 'be': 'тыбецкая', 'be-tarask': 'тыбецкая', 'bg': 'тибетски', 'bgn': 'تبتی', 'bn': 'তিব্বতি', 'bo': 'བོད་སྐད་', 'br': 'tibetaneg', 'brx': 'तिब्बती', 'bs': 'tibetanski', 'bs-Cyrl': 'тибетански', 'bs-Latn': 'tibetanski', 'ca': 'tibetà', 'ca-ES-valencia': 'tibetà', 'ccp': '𑄖𑄨𑄛𑄴𑄝𑄧𑄖𑄨', 'ce': 'тибетхойн', 'chr': 'ᏘᏇᏔᏂ', 'ckb': 'تەبەتی', 'cs': 'tibetština', 'cy': 'Tibeteg', 'da': 'tibetansk', 'de': 'Tibetisch', 'dsb': 'tibetšćina', 'dz': 'བོད་ཁ', 'ee': 'tibetagbe', 'el': 'Θιβετιανά', 'el-polyton': 'Θιβετιανά', 'en': 'Tibetan', 'eo': 'tibeta', 'es': 'tibetano', 'et': 'tiibeti', 'eu': 'tibetera', 'fa': 'تبتی', 'ff-Adlm': '𞤚𞤭𞤦𞤫𞤼𞤫𞤲𞤪𞤫', 'fi': 'tiibet', 'fil': 'Tibetan', 'fo': 'tibetskt', 'fr': 'tibétain', 'fur': 'tibetan', 'fy': 'Tibetaansk', 'ga': 'Tibéidis', 'gd': 'Tibeitis', 'gl': 'tibetano', 'gsw': 'Tibeetisch', 'gu': 'તિબેટીયન', 'ha': 'Tibetan', 'he': 'טיבטית', 'hi': 'तिब्बती', 'hi-Latn': 'Tibbati', 'hr': 'tibetski', 'hsb': 'tibetšćina', 'ht': 'tibétain', 'hu': 'tibeti', 'hy': 'տիբեթերեն', 'ia': 'tibetano', 'id': 'Tibet', 'ig': 'Tibetan', 'is': 'tíbeska', 'it': 'tibetano', 'ja': 'チベット語', 'jv': 'Tibet', 'ka': 'ტიბეტური', 'kea': 'tibetanu', 'kgp': 'timetỹnũ', 'kk': 'тибет тілі', 'kk-Cyrl': 'тибет тілі', 'km': 'ទីបេ', 'kn': 'ಟಿಬೇಟಿಯನ್', 'ko': '티베트어', 'kok': 'तिबेटी', 'kok-Deva': 'तिबेटी', 'ks': 'تِبتی', 'ks-Arab': 'تِبتی', 'ksh': 'Tibehtesch', 'ku': 'tîbetî', 'kxv': 'tibetī', 'kxv-Deva': 'तिब्बती', 'kxv-Latn': 'tibetī', 'kxv-Orya': 'ତିବତି', 'kxv-Telu': 'తిబ్బతన్', 'ky': 'тибетче', 'lb': 'Tibetesch', 'lkt': 'Tibetan Iyápi', 'lld': 'tibetan', 'lo': 'ທິເບທັນ', 'lrc': 'تأبأتی', 'lt': 'tibetiečių', 'lv': 'tibetiešu', 'mai': 'तिब्बती', 'mi': 'Tipete', 'mk': 'тибетски', 'ml': 'ടിബറ്റൻ', 'mn': 'төвд', 'mr': 'तिबेटी', 'ms': 'Tibet', 'mt': 'Tibetjan', 'my': 'တိဘက်', 'mzn': 'تبتی', 'nb': 'tibetansk', 'ne': 'तिब्बती', 'nl': 'Tibetaans', 'nn': 'tibetansk', 'no': 'tibetansk', 'nqo': 'ߕߌߓߋߕߌߞߊ߲', 'or': 'ତିବ୍ବତୀୟ', 'pa': 'ਤਿੱਬਤੀ', 'pa-Guru': 'ਤਿੱਬਤੀ', 'pcm': 'Tibẹ́tan', 'pl': 'tybetański', 'ps': 'تبتي', 'pt': 'tibetano', 'qu': 'Tibetano Simi', 'rm': 'tibetan', 'ro': 'tibetană', 'ru': 'тибетский', 'sah': 'Тибиэттии', 'sc': 'tibetanu', 'sd': 'تبيتائي', 'sd-Arab': 'تبيتائي', 'se': 'tibetagiella', 'si': 'ටිබෙට්', 'sk': 'tibetčina', 'sl': 'tibetanščina', 'smn': 'tiibetkielâ', 'so': 'Tibeetaan', 'sq': 'tibetisht', 'sr': 'тибетански', 'sr-Cyrl': 'тибетански', 'sr-Latn': 'tibetanski', 'sv': 'tibetanska', 'sw': 'Kitibeti', 'ta': 'திபெத்தியன்', 'te': 'టిబెటన్', 'tg': 'тибетӣ', 'th': 'ทิเบต', 'ti': 'ቲበታንኛ', 'tk': 'tibet dili', 'to': 'lea fakatipeti', 'tr': 'Tibetçe', 'tt': 'тибет', 'ug': 'تىبەتچە', 'uk': 'тибетська', 'ur': 'تبتی', 'uz': 'tibet', 'uz-Cyrl': 'тибетча', 'uz-Latn': 'tibet', 'vec': 'tibetan', 'vi': 'Tiếng Tây Tạng', 'wae': 'Tibetiš', 'wo': 'Tibetan', 'xnr': 'तिब्बती', 'yi': 'טיבעטיש', 'yo': 'Tibetán', 'yrl': 'tibetanu', 'yue': '藏文', 'yue-Hans': '藏文', 'yue-Hant': '藏文', 'zh': '藏语', 'zh-Hans': '藏语', 'zh-Hant': '藏文', 'zu': 'isi-Tibetan'},
    'boa': {'en': 'Bora'},
    'bob': {'en': 'Aweer'},
    'boe': {'en': 'Mundabli'},
    'bof': {'en': 'Bolon'},
    'bog': {'en': 'Bamako Sign Language'},
    'boh': {'en': 'Boma'},
    'boi': {'en': 'Barbareño'},
    'boj': {'en': 'Anjam'},
    'bok': {'en': 'Bonjo'},
    'bol': {'en': 'Bole'},
    'bom': {'en': 'Berom'},
    'bon': {'en': 'Bine'},
    'boo': {'en': 'Tiemacèwè Bozo'},
    'bop': {'en': 'Bonkiman'},
    'boq': {'en': 'Bogaya'},
    'bor': {'en': 'Borôro'},
    'bot': {'en': 'Bongo'},
    'bou': {'en': 'Bondei'},
    'bov': {'en': 'Tuwuli'},
    'bow': {'en': 'Rema'},
    'box': {'en': 'Buamu'},
    'boy': {'en': 'Bodo (Central African Republic)'},
    'boz': {'en': 'Tiéyaxo Bozo'},
    'bpa': {'en': 'Daakaka'},
    'bpb': {'en': 'Barbacoas'},
    'bpd': {'en': 'Banda-Banda'},
    'bpe': {'en': 'Bauni'},
    'bpg': {'en': 'Bonggo'},
    'bph': {'en': 'Botlikh'},
    'bpi': {'en': 'Bagupi'},
    'bpj': {'en': 'Binji'},
    'bpk': {'en': 'Orowe'},
    'bpl': {'en': 'Broome Pearling Lugger Pidgin'},
    'bpm': {'en': 'Biyom'},
    'bpn': {'en': 'Dzao Min'},
    'bpo': {'en': 'Anasi'},
    'bpp': {'en': 'Kaure'},
    'bpq': {'en': 'Banda Malay'},
    'bpr': {'en': 'Koronadal Blaan'},
    'bps': {'en': 'Sarangani Blaan'},
    'bpt': {'en': 'Barrow Point'},
    'bpu': {'en': 'Bongu'},
    'bpv': {'en': 'Bian Marind'},
    'bpw': {'en': 'Bo (Papua New Guinea)'},
    'bpx': {'en': 'Palya Bareli'},
    'bpy': {'am': 'ቢሹንፑሪያ', 'ast': 'bishnupriya', 'cs': 'bišnuprijskomanipurština', 'de': 'Bishnupriya', 'en': 'Bishnupriya', 'et': 'bišnuprija', 'fi': 'bišnupria', 'fr': 'bishnupriya', 'gd': 'Bishnupriya', 'gu': 'બિષ્નુપ્રિયા', 'hi-Latn': 'Bishnupriya', 'ht': 'bishnupriya', 'it': 'bishnupriya', 'ja': 'ビシュヌプリヤ・マニプリ語', 'ksh': 'Bischnuprejahnesch', 'lb': 'Bishnupriya', 'lt': 'bišnuprijos', 'mk': 'бишнуприја', 'ms': 'Bishnupriya', 'nb': 'bishnupriya', 'ne': 'विष्णुप्रिया', 'nl': 'Bishnupriya', 'nn': 'bishnupriya', 'no': 'bishnupriya', 'pl': 'bisznuprija-manipuri', 'sv': 'bishnupriya', 'ta': 'பிஷ்ணுப்பிரியா', 'te': 'బిష్ణుప్రియ', 'th': 'พิศนุปริยะ', 'to': 'lea fakapisinupilia', 'tr': 'Bishnupriya', 'vi': 'Tiếng Bishnupriya', 'yue': '比什奴普萊利亞文', 'yue-Hans': '比什奴普莱利亚文', 'yue-Hant': '比什奴普萊利亞文', 'zh-Hant': '比什奴普萊利亞文'},
    'bpz': {'en': 'Bilba'},
    'bqa': {'en': 'Tchumbuli'},
    'bqb': {'en': 'Bagusa'},
    'bqc': {'en': 'Boko (Benin)'},
    'bqd': {'en': 'Bung'},
    'bqf': {'en': 'Baga Kaloum'},
    'bqg': {'en': 'Bago-Kusuntu'},
    'bqh': {'en': 'Baima'},
    'bqi': {'am': 'ባክህቲያሪ', 'ast': 'bakhtiari', 'cs': 'bachtijárština', 'de': 'Bachtiarisch', 'en': 'Bakhtiari', 'et': 'bahtiari', 'fa': 'لری بختیاری', 'fi': 'bahtiari', 'fr': 'bakhtiari', 'gd': 'Bakhtiari', 'hi-Latn': 'Bakhtiari', 'ht': 'bakhtiari', 'it': 'bakhtiari', 'ja': 'バフティヤーリー語', 'lb': 'Bachtiaresch', 'lt': 'bakhtiari', 'mk': 'бахтијарски', 'nb': 'bakhtiari', 'ne': 'बाख्तिआरी', 'nl': 'Bakhtiari', 'nn': 'bakhtiari', 'no': 'bakhtiari', 'pl': 'bachtiarski', 'sv': 'bakhtiari', 'th': 'บักติยารี', 'to': 'lea fakapakitiāli', 'tr': 'Bahtiyari', 'uk': 'бахтіарі', 'vi': 'Tiếng Bakhtiari', 'yue': '巴赫蒂亞里文', 'yue-Hans': '巴赫蒂亚里文', 'yue-Hant': '巴赫蒂亞里文', 'zh-Hant': '巴赫蒂亞里文'},
    'bqj': {'en': 'Bandial'},
    'bqk': {'en': 'Banda-Mbrès'},
    'bql': {'en': 'Bilakura'},
    'bqm': {'en': 'Wumboko'},
    'bqn': {'en': 'Bulgarian Sign Language'},
    'bqo': {'en': 'Balo'},
    'bqp': {'en': 'Busa'},
    'bqq': {'en': 'Biritai'},
    'bqr': {'en': 'Burusu'},
    'bqs': {'en': 'Bosngun'},
    'bqt': {'en': 'Bamukumbit'},
    'bqu': {'en': 'Boguru'},
    'bqv': {'en': 'Koro Wachi'},
    'bqw': {'en': 'Buru (Nigeria)'},
    'bqx': {'en': 'Baangi'},
    'bqy': {'en': 'Bengkala Sign Language'},
    'bqz': {'en': 'Bakaka'},
    'br': {'af': 'Bretons', 'ak': 'Britenni', 'am': 'ብሬቶንኛ', 'ar': 'البريتونية', 'as': 'ব্ৰেটন', 'ast': 'bretón', 'az': 'breton', 'az-Arab': 'برئتونجا', 'az-Cyrl': 'бретон', 'az-Latn': 'breton', 'bal-Latn': 'Brétón', 'be': 'брэтонская', 'be-tarask': 'брэтонская', 'bg': 'бретонски', 'bgn': 'بریتون', 'bn': 'ব্রেটন', 'br': 'brezhoneg', 'brx': 'ब्रेट’न', 'bs': 'bretonski', 'bs-Cyrl': 'бретонски', 'bs-Latn': 'bretonski', 'ca': 'bretó', 'ca-ES-valencia': 'bretó', 'ccp': '𑄝𑄳𑄢𑄬𑄑𑄧𑄚𑄴', 'ce': 'бретонийн', 'chr': 'ᏇᏙᏂ', 'ckb': 'برێتونی', 'cs': 'bretonština', 'cy': 'Llydaweg', 'da': 'bretonsk', 'de': 'Bretonisch', 'dsb': 'bretonšćina', 'ee': 'bretongbe', 'el': 'Βρετονικά', 'el-polyton': 'Βρετονικά', 'en': 'Breton', 'en-Dsrt': '𐐒𐑉𐐯𐐻𐐲𐑌', 'eo': 'bretona', 'es': 'bretón', 'et': 'bretooni', 'eu': 'bretoiera', 'fa': 'برتون', 'ff-Adlm': '𞤄𞤫𞤪𞤫𞤼𞤮𞤲𞤪𞤫', 'fi': 'bretoni', 'fil': 'Breton', 'fo': 'bretonskt', 'fr': 'breton', 'fur': 'breton', 'fy': 'Bretonsk', 'ga': 'Briotáinis', 'gd': 'Breatannais', 'gl': 'bretón', 'gsw': 'Brötoonisch', 'gu': 'બ્રેટોન', 'ha': 'Buretananci', 'he': 'ברטונית', 'hi': 'ब्रेटन', 'hi-Latn': 'Breton', 'hr': 'bretonski', 'hsb': 'bretonšćina', 'ht': 'breton', 'hu': 'breton', 'hy': 'բրետոներեն', 'ia': 'breton', 'id': 'Breton', 'ig': 'Breton', 'is': 'bretónska', 'it': 'bretone', 'ja': 'ブルトン語', 'jv': 'Breton', 'ka': 'ბრეტონული', 'kea': 'bretãu', 'kgp': 'mretỹ', 'kk': 'бретон тілі', 'kk-Cyrl': 'бретон тілі', 'km': 'ប្រ៊ីស្តុន', 'kn': 'ಬ್ರೆಟನ್', 'ko': '브르타뉴어', 'kok': 'ब्रेटन', 'kok-Deva': 'ब्रेटन', 'ks': 'بریٹَن', 'ks-Arab': 'بریٹَن', 'ksh': 'Bettohnesch', 'ku': 'bretonî', 'ky': 'бретончо', 'lb': 'Bretonesch', 'lld': 'breton', 'lo': 'ເບຣຕັນ', 'lrc': 'بئرئتون', 'lt': 'bretonų', 'lv': 'bretoņu', 'mai': 'ब्रेटन', 'mi': 'Peretana', 'mk': 'бретонски', 'ml': 'ബ്രെട്ടൺ', 'mn': 'бретон', 'mr': 'ब्रेतॉन', 'ms': 'Breton', 'mt': 'Breton', 'my': 'ဘရီတွန်', 'mzn': 'برِتونی', 'nb': 'bretonsk', 'ne': 'ब्रेटन', 'nl': 'Bretons', 'nn': 'bretonsk', 'no': 'bretonsk', 'nqo': 'ߓߙߋߕߐ߲ߞߊ߲', 'om': 'Bireetoon', 'or': 'ବ୍ରେଟନ୍', 'pa': 'ਬਰੇਟਨ', 'pa-Guru': 'ਬਰੇਟਨ', 'pcm': 'Brẹ́tọn Lángwej', 'pl': 'bretoński', 'ps': 'برېتون', 'pt': 'bretão', 'qu': 'Breton Simi', 'rm': 'breton', 'ro': 'bretonă', 'ru': 'бретонский', 'rw': 'Inyebiritoni', 'sc': 'brètone', 'sd': 'بريٽن', 'sd-Arab': 'بريٽن', 'se': 'bretonagiella', 'si': 'බ්\u200dරේටොන්', 'sk': 'bretónčina', 'sl': 'bretonščina', 'smn': 'bretonkielâ', 'so': 'Biriton', 'sq': 'bretonisht', 'sr': 'бретонски', 'sr-Cyrl': 'бретонски', 'sr-Latn': 'bretonski', 'sv': 'bretonska', 'sw': 'Kibretoni', 'ta': 'பிரெட்டன்', 'te': 'బ్రెటన్', 'tg': 'бретонӣ', 'th': 'เบรตัน', 'ti': 'ብረቶንኛ', 'tk': 'breton dili', 'to': 'lea fakapeletoni', 'tr': 'Bretonca', 'tt': 'бретон', 'ug': 'بىرېتونچە', 'uk': 'бретонська', 'ur': 'بریٹن', 'uz': 'breton', 'uz-Cyrl': 'бретонча', 'uz-Latn': 'breton', 'vec': 'brètone', 'vi': 'Tiếng Breton', 'wo': 'Breton', 'xnr': 'ब्रेटन', 'yi': 'ברעטאניש', 'yo': 'Èdè Bretoni', 'yrl': 'beretãu', 'yue': '布列塔尼文', 'yue-Hans': '布列塔尼文', 'yue-Hant': '布列塔尼文', 'zh': '布列塔尼语', 'zh-Hans': '布列塔尼语', 'zh-Hant': '布列塔尼文', 'zu': 'isi-Breton'},
    'bra': {'am': 'ብራጅ', 'ar': 'البراجية', 'ast': 'braj', 'az': 'braj', 'az-Latn': 'braj', 'bg': 'брадж', 'bn': 'ব্রাজ', 'br': 'braj', 'brx': 'ब्रज', 'bs': 'braj', 'bs-Cyrl': 'брај', 'bs-Latn': 'braj', 'ca': 'braj', 'ca-ES-valencia': 'braj', 'ccp': '𑄝𑄳𑄢𑄎𑄴', 'cs': 'bradžština', 'da': 'braj', 'de': 'Braj-Bhakha', 'el': 'Μπρατζ', 'el-polyton': 'Μπρατζ', 'en': 'Braj', 'es': 'braj', 'et': 'bradži', 'fa': 'براج', 'fi': 'bradž', 'fr': 'braj', 'fy': 'Braj', 'gd': 'Braj', 'gsw': 'Braj-Bhakha', 'gu': 'વ્રજ', 'he': 'בראג׳', 'hi': 'ब्रज', 'hi-Latn': 'Braj', 'hr': 'braj', 'ht': 'braj', 'hu': 'braj', 'id': 'Braj', 'is': 'braí', 'it': 'braj', 'ja': 'ブラジ語', 'ka': 'ბრაჯი', 'kgp': 'mraj', 'kn': 'ಬ್ರಜ್', 'ko': '브라지어', 'ks': 'برج', 'ks-Arab': 'برج', 'ksh': 'Brahjesch', 'lb': 'Braj-Bhakha', 'lo': 'ບຣາ', 'lt': 'brajų', 'lv': 'bradžiešu', 'mk': 'брај', 'ml': 'ബ്രജ്', 'mr': 'ब्रज', 'mt': 'Braj', 'nb': 'braj', 'ne': 'ब्रज', 'nl': 'Braj', 'nn': 'braj', 'no': 'braj', 'or': 'ବ୍ରାଜ୍', 'pl': 'bradź', 'pt': 'braj', 'rm': 'braj', 'ro': 'braj', 'ru': 'брауи', 'sk': 'bradžčina', 'sl': 'bradžbakanščina', 'sr': 'брај', 'sr-Cyrl': 'брај', 'sr-Latn': 'braj', 'sv': 'braj', 'ta': 'ப்ராஜ்', 'te': 'బ్రాజ్', 'th': 'พัรช', 'to': 'lea fakapalai', 'tr': 'Braj', 'ug': 'بىراجچە', 'uk': 'брадж', 'vi': 'Tiếng Braj', 'yrl': 'barayi', 'yue': '布拉杰文', 'yue-Hans': '布拉杰文', 'yue-Hant': '布拉杰文', 'zh': '布拉杰语', 'zh-Hans': '布拉杰语', 'zh-Hant': '布拉杰文'},
    'brb': {'en': 'Lave'},
    'brc': {'en': 'Berbice Creole Dutch'},
    'brd': {'en': 'Baraamu'},
    'brf': {'en': 'Bira'},
    'brg': {'en': 'Baure'},
    'brh': {'am': 'ብራሁዪ', 'ast': 'brahui', 'br': 'brahweg', 'ca': 'brahui', 'ca-ES-valencia': 'brahui', 'cs': 'brahujština', 'cy': 'Brahui', 'de': 'Brahui', 'en': 'Brahui', 'et': 'brahui', 'fa': 'براهویی', 'fi': 'brahui', 'fr': 'brahoui', 'gd': 'Brahui', 'gu': 'બ્રાહુઈ', 'hi-Latn': 'Brahui', 'ht': 'brahoui', 'it': 'brahui', 'ja': 'ブラフイ語', 'ko': '브라후이어', 'ksh': 'Brahuijesch', 'lb': 'Brahui', 'lt': 'brahujų', 'mk': 'брахујски', 'ms': 'Brahui', 'nb': 'brahui', 'ne': 'ब्राहुइ', 'nl': 'Brahui', 'nn': 'brahui', 'no': 'brahui', 'pl': 'brahui', 'sv': 'brahuiska', 'th': 'บราฮุย', 'to': 'lea fakapalahui', 'tr': 'Brohice', 'vi': 'Tiếng Brahui', 'yue': '布拉維文', 'yue-Hans': '布拉维文', 'yue-Hant': '布拉維文', 'zh-Hant': '布拉維文'},
    'bri': {'en': 'Mokpwe'},
    'brj': {'en': 'Bieria'},
    'brk': {'en': 'Birked'},
    'brl': {'en': 'Birwa'},
    'brm': {'en': 'Barambu'},
    'brn': {'en': 'Boruca'},
    'bro': {'en': 'Brokkat'},
    'brp': {'en': 'Barapasi'},
    'brq': {'en': 'Breri'},
    'brr': {'en': 'Birao'},
    'brs': {'en': 'Baras'},
    'brt': {'en': 'Bitare'},
    'bru': {'en': 'Eastern Bru'},
    'brv': {'en': 'Western Bru'},
    'brw': {'en': 'Bellari'},
    'brx': {'af': 'Bodo', 'ak': 'Bodo', 'am': 'ቦዶ', 'ar': 'البودو', 'as': 'বড়ো', 'ast': 'bodo', 'az': 'bodo', 'az-Arab': 'بوْدوجا', 'az-Cyrl': 'бодо', 'az-Latn': 'bodo', 'bal-Latn': 'Bodó', 'be': 'бода', 'be-tarask': 'бода', 'bg': 'бодо', 'bgn': 'بۆدو', 'bn': 'বোড়ো', 'br': 'bodo', 'brx': 'बर’', 'bs': 'bodo', 'bs-Cyrl': 'бодо', 'bs-Latn': 'bodo', 'ca': 'bodo', 'ca-ES-valencia': 'bodo', 'ccp': '𑄝𑄮𑄢𑄮', 'ce': 'бодо', 'chr': 'ᏉᏙ', 'ckb': 'بۆدۆ', 'cs': 'bodoština', 'cy': 'Bodo', 'da': 'bodo', 'de': 'Bodo', 'dsb': 'bodo', 'ee': 'bodogbe', 'el': 'Μπόντο', 'el-polyton': 'Μπόντο', 'en': 'Bodo', 'es': 'bodo', 'et': 'bodo', 'eu': 'bodoera', 'fa': 'بودویی', 'ff-Adlm': '𞤄𞤮𞤣𞤮𞥅𞤪𞤫', 'fi': 'bodo', 'fil': 'Bodo', 'fo': 'bodo', 'fr': 'bodo', 'fy': 'Bodo', 'ga': 'Bódóis', 'gd': 'Bodo', 'gl': 'bodo', 'gu': 'બોડો', 'ha': 'Bodo', 'he': 'בודו', 'hi': 'बोडो', 'hi-Latn': 'Bodo', 'hr': 'bodo', 'hsb': 'bodo', 'ht': 'bodo', 'hu': 'bodo', 'hy': 'բոդո', 'ia': 'bodo', 'id': 'Bodo', 'ig': 'Bodo', 'is': 'bódó', 'it': 'bodo', 'ja': 'ボド語', 'jv': 'Bodo', 'ka': 'ბოდო', 'kea': 'bodo', 'kgp': 'mono', 'kk': 'бодо тілі', 'kk-Cyrl': 'бодо тілі', 'km': 'បូដូ', 'kn': 'ಬೋಡೊ', 'ko': '보도어', 'kok': 'बोडो', 'kok-Deva': 'बोडो', 'ksh': 'Boddo', 'ku': 'bodoyî', 'kxv': 'boḍo', 'kxv-Deva': 'बोडो', 'kxv-Latn': 'boḍo', 'kxv-Orya': 'ବଡ', 'kxv-Telu': 'బొడొ', 'ky': 'бодочо', 'lb': 'Bodo', 'lo': 'ໂບດູ', 'lrc': 'بودو', 'lt': 'bodo', 'lv': 'bodo', 'mai': 'बोडो', 'mi': 'Pōto', 'mk': 'бодо', 'ml': 'ബോഡോ', 'mn': 'бодо', 'mr': 'बोडो', 'ms': 'Bodo', 'mt': 'Bodo', 'my': 'ဗိုဒို', 'mzn': 'بدویی', 'nb': 'bodo', 'ne': 'बोडो', 'nl': 'Bodo', 'nn': 'bodo', 'no': 'bodo', 'nqo': 'ߓߏߘߏߞߊ߲', 'om': 'Bodo', 'or': 'ବୋଡୋ', 'pa': 'ਬੋਡੋ', 'pa-Guru': 'ਬੋਡੋ', 'pcm': 'Bódo Lángwej', 'pl': 'bodo', 'ps': 'بودو', 'pt': 'bodo', 'qu': 'Bodo Simi', 'rm': 'bodo', 'ro': 'bodo', 'ru': 'бодо', 'sc': 'bodo', 'sd': 'بودو', 'sd-Arab': 'بودو', 'si': 'බොඩො', 'sk': 'bodo', 'sl': 'bodojščina', 'smn': 'bodokielâ', 'so': 'Bodo', 'sq': 'bodoisht', 'sr': 'бодо', 'sr-Cyrl': 'бодо', 'sr-Latn': 'bodo', 'sv': 'bodo', 'sw': 'Kibodo', 'ta': 'போடோ', 'te': 'బోడో', 'th': 'โพโฑ', 'ti': 'ቦዶ', 'tk': 'bodo dili', 'to': 'lea fakapōto', 'tr': 'Bodo', 'ug': 'بودوچە', 'uk': 'бодо', 'ur': 'بوڈو', 'uz': 'bodo', 'uz-Cyrl': 'бодоча', 'uz-Latn': 'bodo', 'vec': 'bodo', 'vi': 'Tiếng Bodo', 'xnr': 'बोडो', 'yo': 'Èdè Bódò', 'yrl': 'budu', 'yue': '博多文', 'yue-Hans': '博多文', 'yue-Hant': '博多文', 'zh': '博多语', 'zh-Hans': '博多语', 'zh-Hant': '博多文', 'zu': 'isi-Bodo'},
    'bry': {'en': 'Burui'},
    'brz': {'en': 'Bilbil'},
    'bs': {'af': 'Bosnies', 'ak': 'Bosniani', 'am': 'ቦስኒያንኛ', 'ar': 'البوسنية', 'as': 'বছনীয়', 'ast': 'bosniu', 'az': 'bosniya', 'az-Arab': 'بوْسنیاجا', 'az-Cyrl': 'босниак', 'az-Latn': 'bosniya', 'bal-Latn': 'Busniái', 'be': 'баснійская', 'be-tarask': 'баснійская', 'bg': 'босненски', 'bgn': 'بوسنی', 'bn': 'বসনীয়', 'br': 'bosneg', 'brx': 'बस्नियान', 'bs': 'bosanski', 'bs-Cyrl': 'босански', 'bs-Latn': 'bosanski', 'ca': 'bosnià', 'ca-ES-valencia': 'bosnià', 'ccp': '𑄝𑄧𑄥𑄴𑄚𑄩𑄠𑄚𑄴', 'ce': 'боснийн', 'chr': 'ᏆᏍᏂᎠᏂ', 'ckb': 'بۆسنی', 'cs': 'bosenština', 'cy': 'Bosnieg', 'da': 'bosnisk', 'de': 'Bosnisch', 'dsb': 'bosnišćina', 'dz': 'བྷོས་ནི་ཡཱན་ཁ', 'ee': 'bosniagbe', 'el': 'Βοσνιακά', 'el-polyton': 'Βοσνιακά', 'en': 'Bosnian', 'eo': 'bosna', 'es': 'bosnio', 'et': 'bosnia', 'eu': 'bosniera', 'fa': 'بوسنیایی', 'ff-Adlm': '𞤄𞤮𞤧𞤲𞤭𞤴𞤢𞥄𞤪𞤫', 'fi': 'bosnia', 'fil': 'Bosnian', 'fo': 'bosniskt', 'fr': 'bosniaque', 'fur': 'bosniac', 'fy': 'Bosnysk', 'ga': 'Boisnis', 'gd': 'Bosnais', 'gl': 'bosníaco', 'gsw': 'Bosnisch', 'gu': 'બોસ્નિયન', 'ha': 'Bosniyanci', 'he': 'בוסנית', 'hi': 'बोस्नियाई', 'hi-Latn': 'Bosnian', 'hr': 'bosanski', 'hsb': 'bosnišćina', 'ht': 'bosniaque', 'hu': 'bosnyák', 'hy': 'բոսնիերեն', 'ia': 'bosniaco', 'id': 'Bosnia', 'ig': 'Bosnian', 'is': 'bosníska', 'it': 'bosniaco', 'ja': 'ボスニア語', 'jv': 'Bosnia lan Hercegovina', 'ka': 'ბოსნიური', 'kea': 'bosniu', 'kgp': 'mojnia', 'kk': 'босния тілі', 'kk-Cyrl': 'босния тілі', 'km': 'បូស្ន៊ី', 'kn': 'ಬೋಸ್ನಿಯನ್', 'ko': '보스니아어', 'kok': 'बोस्नियन', 'kok-Deva': 'बोस्नियन', 'ks': 'بوسنِیَن', 'ks-Arab': 'بوسنِیَن', 'ksh': 'Boßnesch', 'ku': 'bosnî', 'kxv': 'bajniati', 'kxv-Deva': 'बाॅस्नियाती', 'kxv-Latn': 'bajniati', 'kxv-Orya': 'ବସନିଆନ୍', 'kxv-Telu': 'బొస్ నిఆన్', 'ky': 'боснийче', 'lb': 'Bosnesch', 'lkt': 'Bosnia Iyápi', 'lld': 'bosniach', 'lo': 'ບອສນຽນ', 'lrc': 'بوسنیایی', 'lt': 'bosnių', 'lv': 'bosniešu', 'mai': 'बोसनियाई', 'mi': 'Pōngiana', 'mk': 'босански', 'ml': 'ബോസ്നിയൻ', 'mn': 'босни', 'mr': 'बोस्नियन', 'ms': 'Bosnia', 'mt': 'Bożnijaku', 'my': 'ဘော့စ်နီးယား', 'mzn': 'بوسنیایی', 'nb': 'bosnisk', 'ne': 'बोस्नियाली', 'nl': 'Bosnisch', 'nn': 'bosnisk', 'no': 'bosnisk', 'nqo': 'ߓߐߛߑߣߌߞߊ߲', 'om': 'Afaan Bosniyaa', 'or': 'ବୋସନିଆନ୍', 'os': 'босниаг', 'pa': 'ਬੋਸਨੀਆਈ', 'pa-Guru': 'ਬੋਸਨੀਆਈ', 'pcm': 'Bọ́sniá Lángwej', 'pl': 'bośniacki', 'ps': 'بوسني', 'pt': 'bósnio', 'qu': 'Bosnio Simi', 'rm': 'bosniac', 'ro': 'bosniacă', 'ru': 'боснийский', 'rw': 'Inyebosiniya', 'sah': 'Босныйалыы', 'sc': 'bosnìacu', 'sd': 'بوسنيائي', 'sd-Arab': 'بوسنيائي', 'se': 'bosniagiella', 'si': 'බොස්නියානු', 'sk': 'bosniačtina', 'sl': 'bosanščina', 'smn': 'bosniakielâ', 'so': 'Bosniyaan', 'sq': 'boshnjakisht', 'sr': 'босански', 'sr-Cyrl': 'босански', 'sr-Latn': 'bosanski', 'sv': 'bosniska', 'sw': 'Kibosnia', 'ta': 'போஸ்னியன்', 'te': 'బోస్నియన్', 'tg': 'босниягӣ', 'th': 'บอสเนีย', 'ti': 'ቦዝንኛ', 'tk': 'bosniýa dili', 'to': 'lea fakaposinia', 'tr': 'Boşnakça', 'tt': 'босния', 'ug': 'بوسىنچە', 'uk': 'боснійська', 'ur': 'بوسنیائی', 'uz': 'bosniy', 'uz-Cyrl': 'боснийча', 'uz-Latn': 'bosniy', 'vec': 'boznìago', 'vi': 'Tiếng Bosnia', 'wae': 'Bosniš', 'wo': 'Bosñak', 'xnr': 'बोस्नियाई', 'yi': 'באסניש', 'yo': 'Èdè Bosnia', 'yrl': 'businiu', 'yue': '波士尼亞文', 'yue-Hans': '波士尼亚文', 'yue-Hant': '波士尼亞文', 'zh': '波斯尼亚语', 'zh-Hans': '波斯尼亚语', 'zh-Hant': '波士尼亞文', 'zu': 'isi-Bosnian'},
    'bsa': {'en': 'Abinomn'},
    'bsb': {'en': 'Brunei Bisaya'},
    'bsc': {'en': 'Bassari'},
    'bse': {'en': 'Wushi'},
    'bsf': {'en': 'Bauchi'},
    'bsg': {'en': 'Bashkardi'},
    'bsh': {'en': 'Kati'},
    'bsi': {'en': 'Bassossi'},
    'bsj': {'en': 'Bangwinji'},
    'bsk': {'en': 'Burushaski'},
    'bsl': {'en': 'Basa-Gumna'},
    'bsm': {'en': 'Busami'},
    'bsn': {'en': 'Barasana-Eduria'},
    'bso': {'en': 'Buso'},
    'bsp': {'en': 'Baga Sitemu'},
    'bsq': {'en': 'Bassa'},
    'bsr': {'en': 'Bassa-Kontagora'},
    'bss': {'am': 'አኮስ', 'ar': 'أكوس', 'ast': 'akoose', 'bal-Latn': 'Akuse', 'br': 'akoose', 'bs': 'akoski', 'bs-Latn': 'akoski', 'bss': 'Akoose', 'ca': 'akoose', 'ca-ES-valencia': 'akoose', 'cs': 'akoose', 'cy': 'Acwseg', 'da': 'bakossi', 'de': 'Akoose', 'el': 'Ακόσι', 'el-polyton': 'Ακόσι', 'en': 'Akoose', 'es': 'akoose', 'et': 'akoose', 'fi': 'koose', 'fo': 'bakossi', 'fr': 'akoose', 'fy': 'Akoose', 'gd': 'Akoose', 'he': 'אקוסה', 'hi': 'अकूसे', 'hi-Latn': 'Akoose', 'hr': 'akoose', 'ht': 'akoose', 'hu': 'koszi', 'hy': 'աքուզ', 'ia': 'akoose', 'id': 'Akoose', 'ig': 'Akoose', 'is': 'bakossi', 'it': 'akoose', 'ja': 'アコース語', 'kgp': 'akuse', 'ko': '아쿠즈어', 'ku': 'akooseyî', 'lb': 'Akoose', 'lo': 'ອາຄຸດ', 'lt': 'akūsų', 'lv': 'nkosi', 'mk': 'акосе', 'ml': 'അക്കൂസ്', 'ms': 'Akoose', 'nb': 'akose', 'ne': 'अकुज', 'nl': 'Akoose', 'nn': 'bakossi', 'nnh': 'Shwóŋò pakɔsi', 'no': 'akose', 'pl': 'akoose', 'pt': 'akoose', 'ro': 'akoose', 'ru': 'акоосе', 'sk': 'akoose', 'sr': 'акосе', 'sr-Cyrl': 'акосе', 'sr-Latn': 'akose', 'sv': 'bakossi', 'th': 'อาโคซี', 'ti': 'ኣኮስ', 'to': 'lea fakaʻakōse', 'tr': 'Akoose', 'ug': 'ئاكۇسچە', 'uk': 'акус', 'vec': 'akos', 'vi': 'Tiếng Akoose', 'yrl': 'akusi', 'yue': '阿庫色文', 'yue-Hans': '阿库色文', 'yue-Hant': '阿庫色文', 'zh': '阿库色语', 'zh-Hans': '阿库色语', 'zh-Hant': '阿庫色文'},
    'bst': {'en': 'Basketo'},
    'bsu': {'en': 'Bahonsuai'},
    'bsv': {'en': 'Baga Sobané'},
    'bsw': {'en': 'Baiso'},
    'bsx': {'en': 'Yangkam'},
    'bsy': {'en': 'Sabah Bisaya'},
    'bta': {'en': 'Bata'},
    'btb': {'en': 'Beti (Cameroon)'},
    'btc': {'en': 'Bati (Cameroon)'},
    'btd': {'en': 'Batak Dairi'},
    'bte': {'en': 'Gamo-Ningi'},
    'btf': {'en': 'Birgit'},
    'btg': {'en': 'Gagnoa Bété'},
    'bth': {'en': 'Biatah Bidayuh'},
    'bti': {'en': 'Burate'},
    'btj': {'en': 'Bacanese Malay'},
    'btk': {'en': 'Batak languages'},
    'btk-pro': {'en': 'Proto-Batak'},
    'btl': {'en': 'Bhatola'},
    'btm': {'en': 'Batak Mandailing'},
    'btn': {'en': 'Ratagnon'},
    'bto': {'en': 'Rinconada Bikol'},
    'btp': {'en': 'Budibud'},
    'btq': {'en': 'Batek'},
    'btr': {'en': 'Baetora'},
    'bts': {'en': 'Batak Simalungun'},
    'btt': {'en': 'Bete-Bendi'},
    'btu': {'en': 'Batu'},
    'btv': {'en': 'Bateri'},
    'btw': {'en': 'Butuanon'},
    'btx': {'en': 'Batak Karo'},
    'bty': {'en': 'Bobot'},
    'btz': {'en': 'Batak Alas-Kluet'},
    'bua': {'am': 'ቡሪያት', 'ar': 'البرياتية', 'ast': 'buriat', 'az': 'buryat', 'az-Latn': 'buryat', 'be': 'бурацкая', 'be-tarask': 'бурацкая', 'bg': 'бурятски', 'bn': 'বুরিয়াত', 'br': 'bouriat', 'brx': 'बुरियात', 'bs': 'buriat', 'bs-Cyrl': 'буриат', 'bs-Latn': 'buriat', 'ca': 'buriat', 'ca-ES-valencia': 'buriat', 'ccp': '𑄝𑄪𑄢𑄨𑄠𑄖𑄴', 'cs': 'burjatština', 'cy': 'Bwriateg', 'da': 'buriatisk', 'de': 'Burjatisch', 'el': 'Μπουριάτ', 'el-polyton': 'Μπουριάτ', 'en': 'Buriat', 'es': 'buriato', 'et': 'burjaadi', 'fa': 'بوریاتی', 'fi': 'burjaatti', 'fr': 'bouriate', 'fy': 'Buriat', 'ga': 'Buiriáitis', 'gd': 'Buriat', 'gsw': 'Burjatisch', 'gu': 'બુરિયાત', 'he': 'בוריאט', 'hi': 'बुरियात', 'hi-Latn': 'Buriat', 'hr': 'burjatski', 'ht': 'bouriate', 'hu': 'burját', 'id': 'Buriat', 'is': 'búríat', 'it': 'buriat', 'ja': 'ブリヤート語', 'ka': 'ბურიატული', 'kgp': 'murijato', 'kn': 'ಬುರಿಯಟ್', 'ko': '부리아타', 'ks': 'بُرِیَت', 'ks-Arab': 'بُرِیَت', 'ksh': 'Burejahtesch', 'lb': 'Burjatesch', 'lkt': 'Buriat Iyápi', 'lo': 'ບູຣຽດ', 'lt': 'buriatų', 'lv': 'burjatu', 'mk': 'бурјатски', 'ml': 'ബുറിയത്ത്', 'mr': 'बुरियात', 'ms': 'Buriat', 'mt': 'Burjat', 'nb': 'burjatisk', 'ne': 'बुरिआत', 'nl': 'Boerjatisch', 'nn': 'burjatisk', 'no': 'burjatisk', 'or': 'ବୁରିଆଟ୍', 'os': 'бурятаг', 'pl': 'buriacki', 'pt': 'buriato', 'rm': 'buriat', 'ro': 'buriat', 'ru': 'бурятский', 'sk': 'buriatčina', 'sl': 'burjatščina', 'sr': 'бурјатски', 'sr-Cyrl': 'бурјатски', 'sr-Latn': 'burjatski', 'sv': 'burjätiska', 'ta': 'புரியாத்', 'te': 'బురియట్', 'th': 'บูเรียต', 'to': 'lea fakapuliati', 'tr': 'Buryatça', 'ug': 'بۇرىياتچە', 'uk': 'бурятська', 'vi': 'Tiếng Buriat', 'yrl': 'buriatu', 'yue': '布里阿特文', 'yue-Hans': '布里阿特文', 'yue-Hant': '布里阿特文', 'zh': '布里亚特语', 'zh-Hans': '布里亚特语', 'zh-Hant': '布里阿特文'},
    'bub': {'en': 'Bua'},
    'buc': {'en': 'Bushi'},
    'bud': {'en': 'Ntcham'},
    'bue': {'en': 'Beothuk'},
    'buf': {'en': 'Bushoong'},
    'bug': {'af': 'Buginees', 'am': 'ቡጊኔዝ', 'ar': 'البجينيزية', 'as': 'বগিনীজ', 'ast': 'buginés', 'az': 'bugin', 'az-Cyrl': 'буҝин', 'az-Latn': 'bugin', 'be': 'бугіс', 'be-tarask': 'бугіс', 'bg': 'бугински', 'bgn': 'بوگینسی', 'bn': 'বুগিনিজ', 'br': 'bugi', 'brx': 'बुगिनीज', 'bs': 'bugiški', 'bs-Cyrl': 'бугинежански', 'bs-Latn': 'bugiški', 'ca': 'bugui', 'ca-ES-valencia': 'bugui', 'ccp': '𑄝𑄪𑄉𑄨𑄚𑄨', 'ce': 'бугийн', 'chr': 'ᏈᎥᎩᏂᏍ', 'ckb': 'بووگی', 'cs': 'bugiština', 'cy': 'Bwginaeg', 'da': 'buginesisk', 'de': 'Buginesisch', 'dsb': 'bugišćina', 'el': 'Μπουγκίζ', 'el-polyton': 'Μπουγκίζ', 'en': 'Buginese', 'es': 'buginés', 'et': 'bugi', 'eu': 'buginera', 'fa': 'بوگینس', 'ff-Adlm': '𞤄𞤵𞤺𞤭𞤧𞤢𞥄𞤪𞤫', 'fi': 'bugi', 'fil': 'Buginese', 'fo': 'buginesiskt', 'fr': 'bugi', 'fy': 'Bugineesk', 'ga': 'Buiginis', 'gd': 'Cànan nam Bugis', 'gl': 'buginés', 'gsw': 'Bugineesisch', 'gu': 'બુગિનીસ', 'ha': 'Buginesanci', 'he': 'בוגינזית', 'hi': 'बगिनीस', 'hi-Latn': 'Buginese', 'hr': 'buginski', 'hsb': 'buginezišćina', 'ht': 'bugi', 'hu': 'buginéz', 'hy': 'բուգիերեն', 'ia': 'buginese', 'id': 'Bugis', 'ig': 'Buginese', 'is': 'búgíska', 'it': 'bugi', 'ja': 'ブギ語', 'jv': 'Bugis', 'ka': 'ბუგინური', 'kgp': 'muginẽj', 'kk': 'бугис тілі', 'kk-Cyrl': 'бугис тілі', 'km': 'ប៊ុកហ្គី', 'kn': 'ಬುಗಿನೀಸ್', 'ko': '부기어', 'kok': 'बुगिनिज', 'kok-Deva': 'बुगिनिज', 'ks': 'بَگنیٖز', 'ks-Arab': 'بَگنیٖز', 'ksh': 'Bujinehsesch', 'ku': 'bugî', 'ky': 'бугийче', 'lb': 'Buginesesch', 'lo': 'ບູຈີເນັດ', 'lt': 'buginezų', 'lv': 'bugu', 'mai': 'बगिनीज', 'mi': 'Pukenīhi', 'mk': 'бугиски', 'ml': 'ബുഗിനീസ്', 'mn': 'буги', 'mr': 'बगिनीस', 'ms': 'Bugis', 'mt': 'Buginese', 'my': 'ဘူဂစ်စ်', 'nb': 'buginesisk', 'ne': 'बुगिनियाली', 'nl': 'Buginees', 'nn': 'buginesisk', 'no': 'buginesisk', 'or': 'ବୁଗୀନୀଜ୍', 'pa': 'ਬਗਨੀਜ਼', 'pa-Guru': 'ਬਗਨੀਜ਼', 'pcm': 'Buginiís', 'pl': 'bugijski', 'ps': 'بگنيايي', 'pt': 'buginês', 'qu': 'Buginese', 'rm': 'bugi', 'ro': 'bugineză', 'ru': 'бугийский', 'sc': 'buginesu', 'sd': 'بگنيز', 'sd-Arab': 'بگنيز', 'si': 'බුගිනීස්', 'sk': 'bugiština', 'sl': 'buginščina', 'smn': 'bugikielâ', 'so': 'U dhashay Bugin', 'sq': 'buginezisht', 'sr': 'бугијски', 'sr-Cyrl': 'бугијски', 'sr-Latn': 'bugijski', 'sv': 'buginesiska', 'sw': 'Kibugini', 'ta': 'புகினீஸ்', 'te': 'బుగినీస్', 'th': 'บูกิส', 'ti': 'ቡጊንኛ', 'tk': 'bugiý dili', 'to': 'lea fakapukisi', 'tr': 'Bugis', 'ug': 'بۇگىچە', 'uk': 'бугійська', 'ur': 'بگینیز', 'uz': 'bugi', 'uz-Cyrl': 'бугийча', 'uz-Latn': 'bugi', 'vec': 'bugineze', 'vi': 'Tiếng Bugin', 'xnr': 'बगिनीस', 'yo': 'Èdè Bugini', 'yrl': 'buginei', 'yue': '布吉斯文', 'yue-Hans': '布吉斯文', 'yue-Hant': '布吉斯文', 'zh': '布吉语', 'zh-Hans': '布吉语', 'zh-Hant': '布吉斯文', 'zu': 'isi-Buginese'},
    'buh': {'en': 'Younuo Bunu'},
    'bui': {'en': 'Bongili'},
    'buj': {'en': 'Basa-Gurmana'},
    'buk': {'en': 'Bugawac'},
    'bum': {'am': 'ቡሉ', 'ar': 'لغة البولو', 'ast': 'bulu', 'bs': 'bulu', 'bs-Latn': 'bulu', 'ca': 'bulu', 'ca-ES-valencia': 'bulu', 'cs': 'bulu', 'cy': 'Bwlw', 'da': 'bulu', 'de': 'Bulu', 'el': 'Μπουλού', 'el-polyton': 'Μπουλού', 'en': 'Bulu', 'es': 'bulu', 'et': 'bulu', 'fi': 'bulu', 'fr': 'boulou', 'fy': 'Bulu', 'gd': 'Bulu', 'he': 'בולו', 'hi-Latn': 'Bulu', 'hr': 'bulu', 'ht': 'boulou', 'hu': 'bulu', 'id': 'Bulu', 'it': 'bulu', 'ja': 'ブル語', 'kgp': 'muru', 'ko': '불루어', 'lb': 'Bulu', 'lo': 'ບູລູ', 'lt': 'bulu', 'lv': 'bulu', 'mk': 'булу', 'ml': 'ബുളു', 'ms': 'Bulu', 'nb': 'bulu', 'ne': 'बुलु', 'nl': 'Bulu', 'nn': 'bulu', 'nnh': 'Shwóŋò mbulu', 'no': 'bulu', 'pl': 'bulu', 'pt': 'bulu', 'ro': 'bulu', 'ru': 'булу', 'sk': 'bulu', 'sv': 'boulou', 'sw': 'Kibulu', 'th': 'บูลู', 'to': 'lea fakapulu', 'tr': 'Bulu', 'ug': 'بۇلۇچە', 'uk': 'булу', 'vi': 'Tiếng Bulu', 'yrl': 'buru', 'yue': '布魯文', 'yue-Hans': '布鲁文', 'yue-Hant': '布魯文', 'zh': '布鲁语', 'zh-Hans': '布鲁语', 'zh-Hant': '布魯文'},
    'bun': {'en': 'Sherbro'},
    'buo': {'en': 'Terei'},
    'bup': {'en': 'Busoa'},
    'buq': {'en': 'Brem'},
    'bus': {'en': 'Bokobaru'},
    'but': {'en': 'Bungain'},
    'buu': {'en': 'Budu'},
    'buv': {'en': 'Bun'},
    'buw': {'en': 'Bubi'},
    'bux': {'en': 'Boghom'},
    'buy': {'en': 'Bullom So'},
    'buz': {'en': 'Bukwen'},
    'bva': {'en': 'Barein'},
    'bvb': {'en': 'Bube'},
    'bvc': {'en': 'Baelelea'},
    'bvd': {'en': 'Baeggu'},
    'bve': {'en': 'Berau Malay'},
    'bvf': {'en': 'Boor'},
    'bvg': {'en': 'Bonkeng'},
    'bvh': {'en': 'Bure'},
    'bvi': {'en': 'Belanda Viri'},
    'bvj': {'en': 'Baan'},
    'bvk': {'en': 'Bukat'},
    'bvl': {'en': 'Bolivian Sign Language'},
    'bvm': {'en': 'Bamunka'},
    'bvn': {'en': 'Buna'},
    'bvo': {'en': 'Bolgo'},
    'bvp': {'en': 'Bumang'},
    'bvq': {'en': 'Birri'},
    'bvr': {'en': 'Burarra'},
    'bvt': {'en': 'Bati (Indonesia)'},
    'bvu': {'en': 'Bukit Malay'},
    'bvv': {'en': 'Baniva'},
    'bvw': {'en': 'Boga'},
    'bvx': {'en': 'Dibole'},
    'bvy': {'en': 'Baybayanon'},
    'bvz': {'en': 'Bauzi'},
    'bwa': {'en': 'Bwatoo'},
    'bwb': {'en': 'Namosi-Naitasiri-Serua'},
    'bwc': {'en': 'Bwile'},
    'bwd': {'en': 'Bwaidoka'},
    'bwe': {'en': 'Bwe Karen'},
    'bwf': {'en': 'Boselewa'},
    'bwg': {'en': 'Barwe'},
    'bwh': {'en': 'Bishuo'},
    'bwi': {'en': 'Baniwa'},
    'bwj': {'en': 'Láá Láá Bwamu'},
    'bwk': {'en': 'Bauwaki'},
    'bwl': {'en': 'Bwela'},
    'bwm': {'en': 'Biwat'},
    'bwn': {'en': 'Wunai Bunu'},
    'bwo': {'en': 'Boro (Ethiopia)'},
    'bwp': {'en': 'Mandobo Bawah'},
    'bwq': {'en': 'Southern Bobo Madaré'},
    'bwr': {'en': 'Bura-Pabir'},
    'bws': {'en': 'Bomboma'},
    'bwt': {'en': 'Bafaw-Balong'},
    'bwu': {'en': 'Buli (Ghana)'},
    'bww': {'en': 'Bwa'},
    'bwx': {'en': 'Bu-Nao Bunu'},
    'bwy': {'en': 'Cwi Bwamu'},
    'bwz': {'en': 'Bwisi'},
    'bxa': {'en': 'Tairaha'},
    'bxb': {'en': 'Belanda Bor'},
    'bxc': {'en': 'Molengue'},
    'bxd': {'en': 'Pela'},
    'bxe': {'en': 'Birale'},
    'bxf': {'en': 'Bilur'},
    'bxg': {'en': 'Bangala'},
    'bxh': {'en': 'Buhutu'},
    'bxi': {'en': 'Pirlatapa'},
    'bxj': {'en': 'Bayungu'},
    'bxk': {'en': 'Bukusu'},
    'bxl': {'en': 'Jalkunan'},
    'bxm': {'en': 'Mongolia Buriat'},
    'bxn': {'en': 'Burduna'},
    'bxo': {'en': 'Barikanchi'},
    'bxp': {'en': 'Bebil'},
    'bxq': {'en': 'Beele'},
    'bxr': {'en': 'Russia Buriat'},
    'bxs': {'en': 'Busam'},
    'bxu': {'en': 'China Buriat'},
    'bxv': {'en': 'Berakou'},
    'bxw': {'en': 'Bankagooma'},
    'bxx': {'en': 'Borna (Democratic Republic of Congo)'},
    'bxz': {'en': 'Binahari'},
    'bya': {'en': 'Batak'},
    'byb': {'en': 'Bikya'},
    'byc': {'en': 'Ubaghara'},
    'byd': {'en': "Benyadu'"},
    'bye': {'en': 'Pouye'},
    'byf': {'en': 'Bete'},
    'byg': {'en': 'Baygo'},
    'byh': {'en': 'Bhujel'},
    'byi': {'en': 'Buyu'},
    'byj': {'en': 'Bina (Nigeria)'},
    'byk': {'en': 'Biao'},
    'byl': {'en': 'Bayono'},
    'bym': {'en': 'Bidjara'},
    'byn': {'af': 'Blin', 'am': 'ብሊን', 'ar': 'البلينية', 'as': 'ব্লিন', 'ast': 'blin', 'az': 'blin', 'az-Cyrl': 'блин', 'az-Latn': 'blin', 'bal-Latn': 'Blin', 'be': 'білен', 'be-tarask': 'білен', 'bg': 'биленски', 'bgn': 'بلین', 'bn': 'ব্লিন', 'br': 'blin', 'brx': 'ब्लीन', 'bs': 'blin', 'bs-Cyrl': 'блин', 'bs-Latn': 'blin', 'ca': 'bilin', 'ca-ES-valencia': 'bilin', 'ccp': '𑄝𑄳𑄣𑄨𑄚𑄴', 'ce': 'билийн', 'chr': 'ᏟᏂ', 'ckb': 'بلین', 'cs': 'blinština', 'cy': 'Blin', 'da': 'blin', 'de': 'Blin', 'dsb': 'blinšćina', 'el': 'Μπλιν', 'el-polyton': 'Μπλιν', 'en': 'Blin', 'en-Dsrt': '𐐒𐑊𐐮𐑌', 'es': 'blin', 'et': 'bilini', 'eu': 'bilenera', 'fa': 'بلین', 'ff-Adlm': '𞤄𞤭𞤤𞤭𞤲𞤪𞤫', 'fi': 'bilin', 'fil': 'Blin', 'fo': 'blin', 'fr': 'blin', 'fy': 'Blin', 'ga': 'Blinis', 'gd': 'Blin', 'gl': 'blin', 'gsw': 'Blin', 'gu': 'બ્લિન', 'ha': 'Blin', 'he': 'בלין', 'hi': 'ब्लिन', 'hi-Latn': 'Blin', 'hr': 'blin', 'hsb': 'blinšćina', 'ht': 'blin', 'hu': 'blin', 'hy': 'բիլին', 'ia': 'blin', 'id': 'Blin', 'ig': 'Blin', 'is': 'blín', 'it': 'blin', 'ja': 'ビリン語', 'jv': 'Blin', 'ka': 'ბილინი', 'kgp': 'mrĩn', 'kk': 'блин тілі', 'kk-Cyrl': 'блин тілі', 'km': 'ប្ល៊ីន', 'kn': 'ಬ್ಲಿನ್', 'ko': '브린어', 'kok': 'ब्लीन', 'kok-Deva': 'ब्लीन', 'ks': 'بٕلِن', 'ks-Arab': 'بٕلِن', 'ksh': 'Bilihnesch', 'ku': 'blînî', 'ky': 'блинче', 'lb': 'Blin', 'lo': 'ບລິນ', 'lt': 'blin', 'lv': 'bilinu', 'mai': 'बलिन', 'mi': 'Pirina', 'mk': 'биленски', 'ml': 'ബ്ലിൻ', 'mn': 'блин', 'mr': 'ब्लिन', 'ms': 'Blin', 'mt': 'Blin', 'my': 'ဘလင်', 'nb': 'blin', 'ne': 'ब्लिन', 'nl': 'Blin', 'nn': 'blin', 'no': 'blin', 'or': 'ବ୍ଲିନ୍', 'pa': 'ਬਲਿਨ', 'pa-Guru': 'ਬਲਿਨ', 'pcm': 'Bliní', 'pl': 'blin', 'ps': 'بلین', 'pt': 'blin', 'qu': 'Blin', 'rm': 'blin', 'ro': 'blin', 'ru': 'билин', 'sc': 'blin', 'sd': 'بلن', 'sd-Arab': 'بلن', 'si': 'බ්ලින්', 'sk': 'blin', 'sl': 'blinščina', 'smn': 'blinkielâ', 'so': 'U dhashay Byn', 'sq': 'blinisht', 'sr': 'блински', 'sr-Cyrl': 'блински', 'sr-Latn': 'blinski', 'sv': 'blin', 'sw': 'Kiblin', 'ta': 'ப்லின்', 'te': 'బ్లిన్', 'th': 'บลิน', 'ti': 'ብሊን', 'tk': 'blin dili', 'to': 'lea fakapilini', 'tr': 'Blin', 'ug': 'بىلىنچە', 'uk': 'блін', 'ur': 'بلین', 'uz': 'blin', 'uz-Cyrl': 'блинча', 'uz-Latn': 'blin', 'vec': 'blin', 'vi': 'Tiếng Blin', 'xnr': 'ब्लिन', 'yo': 'Èdè Bilini', 'yrl': 'birĩ', 'yue': '比林文', 'yue-Hans': '比林文', 'yue-Hant': '比林文', 'zh': '比林语', 'zh-Hans': '比林语', 'zh-Hant': '比林文', 'zu': 'isi-Blin'},
    'byo': {'en': 'Biyo'},
    'byp': {'en': 'Bumaji'},
    'byq': {'en': 'Basay'},
    'byr': {'en': 'Baruya'},
    'bys': {'en': 'Burak'},
    'byt': {'en': 'Berti'},
    'byv': {'ar': 'لغة الميدومبا', 'ast': 'medumba', 'bs': 'medumba', 'bs-Latn': 'medumba', 'ca': 'medumba', 'ca-ES-valencia': 'medumba', 'cs': 'medumba', 'da': 'medumba', 'de': 'Medumba', 'el': 'Μεντούμπα', 'el-polyton': 'Μεντούμπα', 'en': 'Medumba', 'es': 'medumba', 'et': 'medumba', 'fi': 'medumba', 'fr': 'médumba', 'fy': 'Medumba', 'gd': 'Medumba', 'he': 'מדומבה', 'hi-Latn': 'Medumba', 'hr': 'medumba', 'ht': 'médumba', 'hu': 'medumba', 'id': 'Medumba', 'it': 'medumba', 'ja': 'メドゥンバ語', 'kgp': 'menuma', 'ko': '메둠바어', 'lb': 'Medumba', 'lo': 'ເມດູມບາ', 'lt': 'medumbų', 'lv': 'medumbu', 'mk': 'медумба', 'ml': 'മെഡുംബ', 'ms': 'Medumba', 'nb': 'medumba', 'ne': 'मेडुम्बा', 'nl': 'Medumba', 'nn': 'medumba', 'nnh': 'Shwóŋò ngáŋtÿɔʼ', 'no': 'medumba', 'pl': 'medumba', 'pt': 'medumba', 'ro': 'medumba', 'ru': 'медумба', 'sk': 'medumba', 'sv': 'bagangte', 'sw': 'Kimedumba', 'th': 'เมดุมบา', 'to': 'lea fakametūmipa', 'tr': 'Medumba', 'ug': 'مېدۇمباچە', 'uk': 'медумба', 'vi': 'Tiếng Medumba', 'yrl': 'medũba', 'yue': '梅敦巴文', 'yue-Hans': '梅敦巴文', 'yue-Hant': '梅敦巴文', 'zh': '梅敦巴语', 'zh-Hans': '梅敦巴语', 'zh-Hant': '梅敦巴文'},
    'byw': {'en': 'Belhariya'},
    'byx': {'en': 'Qaqet'},
    'byy': {'en': 'Buya'},
    'byz': {'en': 'Banaro'},
    'bza': {'en': 'Bandi'},
    'bzb': {'en': 'Andio'},
    'bzc': {'en': 'Southern Betsimisaraka Malagasy'},
    'bzd': {'en': 'Bribri'},
    'bze': {'en': 'Jenaama Bozo'},
    'bzf': {'en': 'Boikin'},
    'bzg': {'en': 'Babuza'},
    'bzh': {'en': 'Mapos Buang'},
    'bzi': {'en': 'Bisu'},
    'bzj': {'en': 'Belize Kriol English'},
    'bzk': {'en': 'Nicaragua Creole English'},
    'bzl': {'en': 'Boano (Sulawesi)'},
    'bzm': {'en': 'Bolondo'},
    'bzn': {'en': 'Boano (Maluku)'},
    'bzo': {'en': 'Bozaba'},
    'bzp': {'en': 'Kemberano'},
    'bzq': {'en': 'Buli (Indonesia)'},
    'bzr': {'en': 'Biri'},
    'bzs': {'en': 'Brazilian Sign Language'},
    'bzt': {'en': 'Brithenig'},
    'bzu': {'en': 'Burmeso'},
    'bzv': {'en': 'Naami'},
    'bzw': {'en': 'Basa (Nigeria)'},
    'bzx': {'en': 'Kɛlɛngaxo Bozo'},
    'bzy': {'en': 'Obanliku'},
    'bzz': {'en': 'Evant'},
    'ca': {'af': 'Katalaans', 'ak': 'Katalan', 'am': 'ካታላንኛ', 'ar': 'الكتالانية', 'as': 'কাতালান', 'ast': 'catalán', 'az': 'katalan', 'az-Arab': 'کاتالانجا', 'az-Cyrl': 'каталан', 'az-Latn': 'katalan', 'bal-Latn': 'Katálan', 'be': 'каталанская', 'be-tarask': 'каталанская', 'bg': 'каталонски', 'bgn': 'کاتالانی', 'bn': 'কাতালান', 'br': 'katalaneg', 'brx': 'कातालान', 'bs': 'katalonski', 'bs-Cyrl': 'каталонски', 'bs-Latn': 'katalonski', 'ca': 'català', 'ca-ES-valencia': 'català', 'ccp': '𑄇𑄖𑄣𑄚𑄴', 'ce': 'каталонийн', 'chr': 'ᎨᏔᎳᏂ', 'ckb': 'كاتالۆنی', 'cs': 'katalánština', 'cy': 'Catalaneg', 'da': 'catalansk', 'de': 'Katalanisch', 'dsb': 'katanlanšćina', 'dz': 'ཀེ་ཊ་ལཱན་ཁ', 'ee': 'katalagbe', 'el': 'Καταλανικά', 'el-polyton': 'Καταλανικά', 'en': 'Catalan', 'en-Dsrt': '𐐗𐐪𐐻𐐲𐑊𐐪𐑌', 'eo': 'kataluna', 'es': 'catalán', 'et': 'katalaani', 'eu': 'katalana', 'fa': 'کاتالان', 'ff-Adlm': '𞤑𞤢𞤼𞤢𞤤𞤢𞤲𞤪𞤫', 'fi': 'katalaani', 'fil': 'Catalan', 'fo': 'katalani', 'fr': 'catalan', 'fur': 'catalan', 'fy': 'Katalaansk', 'ga': 'Catalóinis', 'gd': 'Catalanais', 'gl': 'catalán', 'gsw': 'Katalaanisch', 'gu': 'કતલાન', 'ha': 'Kataloniyanci', 'he': 'קטלאנית', 'hi': 'कातालान', 'hi-Latn': 'Catalan', 'hr': 'katalonski', 'hsb': 'katalanšćina', 'ht': 'catalan', 'hu': 'katalán', 'hy': 'կատալաներեն', 'ia': 'catalano', 'id': 'Katalan', 'ig': 'Catalan', 'is': 'katalónska', 'it': 'catalano', 'ja': 'カタロニア語', 'jv': 'Katala', 'ka': 'კატალანური', 'kea': 'katalãu', 'kgp': 'katarũg', 'kk': 'каталан тілі', 'kk-Cyrl': 'каталан тілі', 'km': 'កាតាឡាន', 'kn': 'ಕೆಟಲಾನ್', 'ko': '카탈로니아어', 'kok': 'कटलान', 'kok-Deva': 'कटलान', 'ks': 'کَتلان', 'ks-Arab': 'کَتلان', 'ksh': 'Kattalahnesch', 'ku': 'katalanî', 'kxv': 'keṭelan', 'kxv-Deva': 'कातालान', 'kxv-Latn': 'keṭelan', 'kxv-Orya': 'କାଟଲାନ୍', 'kxv-Telu': 'కాటాలాన్', 'ky': 'каталончо', 'lb': 'Katalanesch', 'lkt': 'Catalan Iyápi', 'lld': 'catalan', 'lo': 'ຄາຕາລານ', 'lrc': 'کاتالان', 'lt': 'katalonų', 'lv': 'katalāņu', 'mai': 'केटालान', 'mi': 'Katarana', 'mk': 'каталонски', 'ml': 'കറ്റാലാൻ', 'mn': 'каталан', 'mr': 'कातालान', 'ms': 'Catalonia', 'mt': 'Katalan', 'my': 'ကတ်တလန်', 'mzn': 'کاتالونی', 'nb': 'katalansk', 'ne': 'क्याटालन', 'nl': 'Catalaans', 'nn': 'katalansk', 'no': 'katalansk', 'nqo': 'ߞߕߊߟߊ߲ߞߊ߲', 'om': 'Afaan Katalaa', 'or': 'କାଟାଲାନ୍', 'os': 'каталайнаг', 'pa': 'ਕੈਟਾਲਾਨ', 'pa-Guru': 'ਕੈਟਾਲਾਨ', 'pcm': 'Kátála Lángwej', 'pl': 'kataloński', 'ps': 'کټلاني', 'pt': 'catalão', 'qu': 'Catalan Simi', 'rm': 'catalan', 'ro': 'catalană', 'ru': 'каталанский', 'rw': 'Igikatalani', 'sah': 'Каталаанныы', 'sc': 'catalanu', 'sd': 'ڪيٽالان', 'sd-Arab': 'ڪيٽالان', 'se': 'katalánagiella', 'si': 'කැටලන්', 'sk': 'katalánčina', 'sl': 'katalonščina', 'smn': 'katalankielâ', 'so': 'Katalaan', 'sq': 'katalonisht', 'sr': 'каталонски', 'sr-Cyrl': 'каталонски', 'sr-Latn': 'katalonski', 'sv': 'katalanska', 'sw': 'Kikatalani', 'ta': 'கேட்டலான்', 'te': 'కాటలాన్', 'tg': 'каталонӣ', 'th': 'คาตาลัน', 'ti': 'ካታላን', 'tk': 'katalan dili', 'to': 'lea fakakatalani', 'tr': 'Katalanca', 'tt': 'каталан', 'ug': 'كاتالانچە', 'uk': 'каталонська', 'ur': 'کیٹالان', 'uz': 'katalan', 'uz-Cyrl': 'каталонча', 'uz-Latn': 'katalan', 'vec': 'catalan', 'vi': 'Tiếng Catalan', 'wae': 'Katalaniš', 'wo': 'Katalan', 'xnr': 'कातालान', 'yi': 'קאַטאַלאניש', 'yo': 'Èdè Katala', 'yrl': 'katará', 'yue': '加泰羅尼亞文', 'yue-Hans': '加泰罗尼亚文', 'yue-Hant': '加泰羅尼亞文', 'zh': '加泰罗尼亚语', 'zh-Hans': '加泰罗尼亚语', 'zh-Hant': '加泰蘭文', 'zu': 'isi-Catalan'},
    'caa': {'en': 'Chortí'},
    'cab': {'en': 'Garifuna'},
    'cac': {'en': 'Chuj'},
    'cad': {'am': 'ካዶ', 'ar': 'الكادو', 'ast': 'caddo', 'az': 'keddo', 'az-Latn': 'keddo', 'bal-Latn': 'Kaddó-kad', 'bg': 'каддо', 'bn': 'ক্যাডো', 'br': 'caddo', 'brx': 'काद्दौ', 'bs': 'kado', 'bs-Cyrl': 'кадо', 'bs-Latn': 'kado', 'ca': 'caddo', 'ca-ES-valencia': 'caddo', 'ccp': '𑄇𑄳𑄠𑄓𑄮', 'cs': 'caddo', 'cy': 'Cado', 'da': 'caddo', 'de': 'Caddo', 'el': 'Κάντο', 'el-polyton': 'Κάντο', 'en': 'Caddo', 'es': 'caddo', 'et': 'kado', 'fa': 'کادویی', 'fi': 'caddo', 'fr': 'caddo', 'fy': 'Kaddo', 'gd': 'Caddo', 'gsw': 'Caddo', 'gu': 'કડ્ડો', 'he': 'קאדו', 'hi': 'कैड्डो', 'hi-Latn': 'Caddo', 'hr': 'caddo', 'ht': 'caddo', 'hu': 'caddo', 'ia': 'caddo', 'id': 'Kado', 'ig': 'Caddo', 'is': 'kaddó', 'it': 'caddo', 'ja': 'カドー語', 'kgp': 'kano', 'kn': 'ಕ್ಯಾಡ್ಡೋ', 'ko': '카도어', 'ks': 'کاڈو', 'ks-Arab': 'کاڈو', 'ku': 'kadoyî', 'lb': 'Caddo', 'lo': 'ແຄດໂດ', 'lt': 'kado', 'lv': 'kadu', 'mk': 'кадо', 'ml': 'കാഡോ', 'mr': 'कॅड्डो', 'mt': 'Kaddo', 'nb': 'caddo', 'ne': 'काड्डो', 'nl': 'Caddo', 'nn': 'caddo', 'no': 'caddo', 'or': 'କାଡୋ', 'pl': 'kaddo', 'pt': 'caddo', 'rm': 'caddo', 'ro': 'caddo', 'ru': 'каддо', 'sk': 'kaddo', 'sl': 'kadoščina', 'sr': 'кадо', 'sr-Cyrl': 'кадо', 'sr-Latn': 'kado', 'sv': 'caddo', 'ta': 'கேடோ', 'te': 'కేడ్డో', 'th': 'คัดโด', 'ti': 'ካድዶ', 'to': 'lea fakakato', 'tr': 'Kado dili', 'ug': 'كاددوچە', 'uk': 'каддо', 'vec': 'cado', 'vi': 'Tiếng Caddo', 'yrl': 'cadu', 'yue': '卡多文', 'yue-Hans': '卡多文', 'yue-Hant': '卡多文', 'zh': '卡多语', 'zh-Hans': '卡多语', 'zh-Hant': '卡多文'},
    'cae': {'en': 'Lehar'},
    'caf': {'en': 'Southern Carrier'},
    'cag': {'en': 'Nivaclé'},
    'cah': {'en': 'Cahuarano'},
    'cai': {'en': 'Central American Indian languages'},
    'caj': {'en': 'Chané'},
    'cak': {'en': 'Kaqchikel'},
    'cal': {'en': 'Carolinian'},
    'cam': {'en': 'Cemuhî'},
    'can': {'en': 'Chambri'},
    'cao': {'en': 'Chácobo'},
    'cap': {'en': 'Chipaya'},
    'caq': {'en': 'Car Nicobarese'},
    'car': {'am': 'ካሪብ', 'ar': 'الكاريبية', 'ast': 'caribe', 'az': 'karib', 'az-Latn': 'karib', 'bg': 'карибски', 'bn': 'ক্যারিব', 'br': 'karibeg', 'brx': 'कारीब्', 'bs': 'karipski', 'bs-Cyrl': 'карипски', 'bs-Latn': 'karipski', 'ca': 'carib', 'ca-ES-valencia': 'carib', 'ccp': '𑄝𑄳𑄠𑄢𑄨𑄛𑄴', 'cs': 'karibština', 'cy': 'Caribeg', 'da': 'caribisk', 'de': 'Karibisch', 'el': 'Καρίμπ', 'el-polyton': 'Καρίμπ', 'en': 'Carib', 'es': 'caribe', 'et': 'kariibi', 'fa': 'کاریبی', 'fi': 'karibi', 'fr': 'caribe', 'fy': 'Karibysk', 'gd': 'Carib', 'gsw': 'Kariibisch', 'gu': 'કરિબ', 'he': 'קאריב', 'hi': 'कैरिब', 'hi-Latn': 'Carib', 'hr': 'karipski', 'ht': 'caribe', 'hu': 'karib', 'id': 'Karibia', 'is': 'karíbamál', 'it': 'caribico', 'ja': 'カリブ語', 'kgp': 'karime', 'kn': 'ಕಾರಿಬ್', 'ko': '카리브어', 'ks': 'کارِب', 'ks-Arab': 'کارِب', 'lb': 'Karibesch', 'lo': 'ຄາຣິບ', 'lt': 'karibų', 'lv': 'karību', 'mk': 'карипски', 'ml': 'കാരിബ്', 'mr': 'कॅरिब', 'mt': 'Karib', 'nb': 'karibisk', 'ne': 'क्यारिब', 'nl': 'Caribisch', 'nn': 'carib', 'no': 'karibisk', 'or': 'କାରିବ୍', 'pl': 'karaibski', 'pt': 'caribe', 'rm': 'caribic', 'ro': 'carib', 'ru': 'кариб', 'sk': 'karibčina', 'sl': 'karibski jezik', 'sr': 'карипски', 'sr-Cyrl': 'карипски', 'sr-Latn': 'karipski', 'sv': 'karibiska', 'ta': 'கரீப்', 'te': 'కేరిబ్', 'th': 'คาริบ', 'to': 'lea fakakalipa', 'tr': 'Carib', 'ug': 'كارىبچە', 'uk': 'карібська', 'vi': 'Tiếng Carib', 'yrl': 'karibi', 'yue': '加勒比文', 'yue-Hans': '加勒比文', 'yue-Hant': '加勒比文', 'zh': '加勒比语', 'zh-Hans': '加勒比语', 'zh-Hant': '加勒比文'},
    'cas': {'en': 'Tsimané'},
    'cau': {'en': 'Caucasian languages'},
    'cau-abz-pro': {'en': 'Proto-Abkhaz-Abaza'},
    'cau-ava-pro': {'en': 'Proto-Avaro-Andian'},
    'cau-cir-pro': {'en': 'Proto-Circassian'},
    'cau-drg-pro': {'en': 'Proto-Dargwa'},
    'cau-lzg-pro': {'en': 'Proto-Lezghian'},
    'cau-nec-pro': {'en': 'Proto-Northeast Caucasian'},
    'cau-nkh-pro': {'en': 'Proto-Nakh'},
    'cau-nwc-pro': {'en': 'Proto-Northwest Caucasian'},
    'cau-tsz-pro': {'en': 'Proto-Tsezian'},
    'cav': {'en': 'Cavineña'},
    'caw': {'en': 'Callawalla'},
    'cax': {'en': 'Chiquitano'},
    'cay': {'af': 'Cayuga', 'am': 'ካዩጋ', 'ar': 'الكايوجية', 'as': 'কায়ুগা', 'ast': 'cayuga', 'az': 'Kayuqa', 'az-Latn': 'Kayuqa', 'be': 'каюга', 'be-tarask': 'каюга', 'bg': 'каюга', 'bn': 'কায়ুগা', 'br': 'kayougeg', 'brx': 'कायुगा', 'bs': 'kajuga', 'bs-Cyrl': 'Кајуга', 'bs-Latn': 'kajuga', 'ca': 'cayuga', 'ca-ES-valencia': 'cayuga', 'chr': 'ᎧᏳᎦ', 'cs': 'kajugština', 'cy': 'Cayuga', 'da': 'cayuga', 'de': 'Cayuga', 'dsb': 'cayugašćina', 'el': 'Καγιούγκα', 'el-polyton': 'Καγιούγκα', 'en': 'Cayuga', 'es': 'cayuga', 'et': 'kajuka', 'eu': 'cayugera', 'fa': 'کایوگا', 'ff-Adlm': '𞤑𞤢𞤴𞤺𞤢𞥄𞤪𞤫', 'fi': 'cayuga', 'fil': 'Cayuga', 'fr': 'cayuga', 'fy': 'Cayuga', 'ga': 'teanga Cayuga', 'gd': 'Cayuga', 'gl': 'cayuga', 'gu': 'કેયુગા', 'ha': 'Cayuga', 'he': 'קאיוגה', 'hi': 'कायुगा', 'hi-Latn': 'Cayuga', 'hr': 'cayuga', 'hsb': 'cayugašćina', 'ht': 'cayuga', 'hu': 'kajuga', 'hy': 'կայուգա', 'ia': 'cayuga', 'id': 'Cayuga', 'ig': 'Cayuga', 'is': 'kajúga', 'it': 'cayuga', 'ja': 'カユーガ語', 'jv': 'Kayuga', 'ka': 'კაიუგა', 'kgp': 'kajuga', 'kk': 'кайюга тілі', 'kk-Cyrl': 'кайюга тілі', 'km': 'ខាយូហ្កា', 'kn': 'ಕಯುಗಾ', 'ko': '카유가어', 'kok': 'कायुगा', 'kok-Deva': 'कायुगा', 'ku': 'kayugayî', 'ky': 'каюгиче', 'lb': 'Cayuga', 'lo': 'ຄາຢູກາ', 'lt': 'kaijūgų', 'lv': 'kajuga', 'mai': 'केयुगा', 'mi': 'Keiūka', 'mk': 'кајуга', 'ml': 'കയൂഗ', 'mn': 'кайюга хэл', 'mr': 'केयुगा', 'ms': 'Cayuga', 'my': 'ကာယုဂါ', 'nb': 'cayuga', 'ne': 'कायुगा', 'nl': 'Cayuga', 'nn': 'cayuga', 'no': 'cayuga', 'or': 'କାୟୁଗା', 'pa': 'ਕਾਯੁਗਾ', 'pa-Guru': 'ਕਾਯੁਗਾ', 'pcm': 'Kayúga', 'pl': 'kajuga', 'ps': 'Ca', 'pt': 'cayuga', 'qu': 'Cayugá', 'ro': 'cayuga', 'ru': 'кайюга', 'sc': 'cayuga', 'sd': 'ڪايوگا', 'sd-Arab': 'ڪايوگا', 'si': 'කයුගා', 'sk': 'kajugčina', 'sl': 'kajuščina', 'so': 'Cayuga', 'sq': 'kajugaisht', 'sr': 'кајуга', 'sr-Cyrl': 'кајуга', 'sr-Latn': 'kajuga', 'sv': 'cayuga', 'sw': 'Kikayuga', 'ta': 'கேயுகா', 'te': 'సేయుగా', 'th': 'คายูกา', 'ti': 'ካዩጋ', 'tk': 'kaýuga dili', 'to': 'lea fakakaiuka', 'tr': 'Kayuga dili', 'ug': 'كايۇگاچە', 'uk': 'кайюга', 'ur': 'کایوگا', 'uz': 'kayuga', 'uz-Latn': 'kayuga', 'vec': 'cayuga', 'vi': 'Tiếng Cayuga', 'yo': 'Èdè Kayuga', 'yrl': 'kayuga', 'yue': '卡尤加文', 'yue-Hans': '卡尤加文', 'yue-Hant': '卡尤加文', 'zh': '卡尤加语', 'zh-Hans': '卡尤加语', 'zh-Hant': '卡尤加文', 'zu': 'isi-Cayuga'},
    'caz': {'en': 'Canichana'},
    'cba': {'en': 'Chibchan languages'},
    'cba-pro': {'en': 'Proto-Chibchan'},
    'cbb': {'en': 'Cabiyarí'},
    'cbc': {'en': 'Carapana'},
    'cbd': {'en': 'Carijona'},
    'cbe': {'en': 'Chipiajes'},
    'cbg': {'en': 'Chimila'},
    'cbh': {'en': 'Cagua'},
    'cbi': {'en': 'Chachi'},
    'cbj': {'en': 'Ede Cabe'},
    'cbk': {'en': 'Chavacano'},
    'cbl': {'en': 'Bualkhaw Chin'},
    'cbn': {'en': 'Nyahkur'},
    'cbo': {'en': 'Izora'},
    'cbq': {'en': 'Tsucuba'},
    'cbr': {'en': 'Cashibo-Cacataibo'},
    'cbs': {'en': 'Cashinahua'},
    'cbt': {'en': 'Chayahuita'},
    'cbu': {'en': 'Candoshi-Shapra'},
    'cbv': {'en': 'Cacua'},
    'cbw': {'en': 'Kinabalian'},
    'cby': {'en': 'Carabayo'},
    'cca': {'en': 'Cauca'},
    'ccc': {'en': 'Chamicuro'},
    'ccd': {'en': 'Cafundo Creole'},
    'cce': {'en': 'Chopi'},
    'ccg': {'en': 'Samba Daka'},
    'cch': {'am': 'አትሳም', 'ar': 'الأتسام', 'ast': 'atsam', 'az': 'atsam', 'az-Latn': 'atsam', 'bal-Latn': 'Atsam', 'bg': 'атсам', 'bn': 'আত্সাম', 'br': 'atsam', 'brx': 'आत्सम', 'bs': 'atsam', 'bs-Cyrl': 'атсамски', 'bs-Latn': 'atsam', 'ca': 'atsam', 'ca-ES-valencia': 'atsam', 'cch': 'Atsam', 'ccp': '𑄃𑄖𑄴𑄥𑄟𑄴', 'cs': 'atsam', 'cy': 'Atsameg', 'da': 'atsam', 'de': 'Atsam', 'el': 'Ατσάμ', 'el-polyton': 'Ατσάμ', 'en': 'Atsam', 'es': 'atsam', 'et': 'aitšami', 'fi': 'atsam', 'fr': 'atsam', 'fy': 'Atsam', 'gd': 'Atsam', 'gsw': 'Atsam', 'gu': 'અત્સમ', 'he': 'אטסם', 'hi': 'अत्सम', 'hi-Latn': 'Atsam', 'hr': 'atsam', 'ht': 'atsam', 'hu': 'atszam', 'ia': 'atsam', 'id': 'Atsam', 'ig': 'Atsam', 'is': 'atsam', 'it': 'atsam', 'ja': 'チャワイ語', 'kgp': 'ansỹm', 'kn': 'ಅಟ್ಸಮ್', 'ko': '앗삼어', 'ks': 'اتسَم', 'ks-Arab': 'اتسَم', 'ku': 'atsamî', 'lb': 'Atsam', 'lo': 'ອາດແຊມ', 'lt': 'atsamų', 'lv': 'atsamu', 'mk': 'ацам', 'ml': 'അറ്റ്സാം', 'mr': 'अत्सम', 'mt': 'Atsam', 'nb': 'atsam', 'ne': 'अट्साम', 'nl': 'Atsam', 'nn': 'atsam', 'no': 'atsam', 'or': 'ଆତ୍ସମ୍', 'pl': 'atsam', 'pt': 'atsam', 'rm': 'atsam', 'ro': 'atsam', 'ru': 'атсам', 'sk': 'atsam', 'sr': 'атсам', 'sr-Cyrl': 'атсам', 'sr-Latn': 'atsam', 'sv': 'atsam', 'ta': 'ஆட்சம்', 'te': 'అట్సామ్', 'th': 'แอตแซม', 'ti': 'ኣትሳም', 'to': 'lea fakaʻatisami', 'tr': 'Atsam', 'ug': 'ئاتسامچە', 'uk': 'атсам', 'vec': 'atsam', 'vi': 'Tiếng Atsam', 'yrl': 'atisã', 'yue': '阿燦文', 'yue-Hans': '阿灿文', 'yue-Hant': '阿燦文', 'zh': '阿灿语', 'zh-Hans': '阿灿语', 'zh-Hant': '阿燦文'},
    'ccj': {'en': 'Kasanga'},
    'ccl': {'en': 'Cutchi-Swahili'},
    'ccm': {'en': 'Malaccan Creole Malay'},
    'ccn': {'en': 'North Caucasian languages'},
    'ccn-pro': {'en': 'Proto-North Caucasian'},
    'cco': {'en': 'Comaltepec Chinantec'},
    'ccp': {'af': 'Tsjaakma', 'am': 'ቻክማ', 'ar': 'تشاكما', 'as': 'চাক্\u200cমা', 'az': 'Çakma', 'az-Latn': 'Çakma', 'bal-Latn': 'Chakmá', 'be': 'чакма', 'be-tarask': 'чакма', 'bg': 'чакма', 'bn': 'চাকমা', 'br': 'chakmaeg', 'brx': 'चाकमा', 'bs': 'čakma', 'bs-Cyrl': 'чакма', 'bs-Latn': 'čakma', 'ca': 'chakma', 'ca-ES-valencia': 'chakma', 'ccp': '𑄌𑄋𑄴𑄟𑄳𑄦', 'chr': 'ᏣᎧᎹ', 'cs': 'čakma', 'cy': 'Tsiacma', 'da': 'chakma', 'de': 'Chakma', 'dsb': 'čakma', 'el': 'Τσάκμα', 'el-polyton': 'Τσάκμα', 'en': 'Chakma', 'es': 'chakma', 'et': 'tšaakma', 'eu': 'chakmera', 'fa': 'چاکما', 'ff-Adlm': '𞤅𞤢𞤳𞤥𞤢𞥄𞤪𞤫', 'fi': 'chakma', 'fil': 'Chakma', 'fo': 'khakma', 'fr': 'changma kodha', 'ga': 'Seácmais', 'gd': 'Chakma', 'gl': 'chakma', 'gu': 'ચકમા', 'ha': 'Chakma', 'he': 'צ׳אקמה', 'hi': 'चकमा', 'hi-Latn': 'Chakma', 'hr': 'chakma', 'hsb': 'čakma', 'ht': 'changma kodha', 'hu': 'csakma', 'hy': 'չակմա', 'ia': 'chakma', 'id': 'Chakma', 'ig': 'Chakma', 'is': 'tsjakma', 'it': 'chakma', 'ja': 'チャクマ語', 'jv': 'Chakma', 'ka': 'ჩაკმა', 'kea': 'xakma', 'kgp': 'sakimỹ', 'kk': 'чакма тілі', 'kk-Cyrl': 'чакма тілі', 'km': 'ចាក់ម៉ា', 'kn': 'ಚಕ್ಮಾ', 'ko': '차크마어', 'kok': 'चाक्मा', 'kok-Deva': 'चाक्मा', 'ku': 'çakmayî', 'ky': 'чакма', 'lo': 'ຊາກມາ', 'lt': 'Čakma', 'lv': 'čakmu', 'mai': 'चकमा', 'mi': 'Tiakamā', 'mk': 'чакмански', 'ml': 'ചക്\u200cമ', 'mn': 'чакма', 'mr': 'चाकमा', 'ms': 'Chakma', 'my': 'ချတ်ခ်မာ', 'nb': 'chakma', 'ne': 'चाक्मा', 'nl': 'Chakma', 'nn': 'chakma', 'no': 'chakma', 'nqo': 'ߞߏߘߊ߫ ߗߊߜ߭ߑߡߊߞߊ߲', 'or': 'ଚକମା', 'pa': 'ਚਕਮਾ', 'pa-Guru': 'ਚਕਮਾ', 'pcm': 'Chákma Lángwej', 'pl': 'czakma', 'ps': 'چکما', 'pt': 'chakma', 'qu': 'Chakma Simi', 'rm': 'chakma', 'ro': 'chakma', 'ru': 'чакма', 'sc': 'chakma', 'sd': 'چمڪا', 'sd-Arab': 'چمڪا', 'si': 'චක්මා', 'sk': 'čakma', 'sl': 'chakma', 'so': 'Jakma', 'sq': 'çakmaisht', 'sr': 'чакма', 'sr-Cyrl': 'чакма', 'sr-Latn': 'čakma', 'sv': 'chakma', 'sw': 'Kichakma', 'ta': 'சக்மா', 'te': 'చక్మా', 'th': 'จักม่า', 'ti': 'ቻክማ', 'tk': 'çakma dili', 'to': 'lea fakasākima', 'tr': 'Chakma', 'uk': 'чакма', 'ur': 'چکمہ', 'uz': 'chakma', 'uz-Latn': 'chakma', 'vec': 'ciakma', 'vi': 'Tiếng Chakma', 'xnr': 'चकमा', 'yo': 'Èdè Chakma', 'yrl': 'xakima', 'yue': '查克馬文', 'yue-Hans': '查克马文', 'yue-Hant': '查克馬文', 'zh': '查克玛语', 'zh-Hans': '查克玛语', 'zh-Hant': '查克馬文', 'zu': 'i-Chakma'},
    'ccq': {'en': 'Chaungtha'},
    'ccr': {'en': 'Cacaopera'},
    'ccs': {'en': 'South Caucasian languages'},
    'ccs-gzn-pro': {'en': 'Proto-Georgian-Zan'},
    'ccs-pro': {'en': 'Proto-Kartvelian'},
    'cda': {'en': 'Choni'},
    'cdc': {'en': 'Chadic languages'},
    'cdc-cbm-pro': {'en': 'Proto-Central Chadic'},
    'cdc-pro': {'en': 'Proto-Chadic'},
    'cdd': {'en': 'Caddoan languages'},
    'cdd-pro': {'en': 'Proto-Caddoan'},
    'cde': {'en': 'Chenchu'},
    'cdf': {'en': 'Chiru'},
    'cdg': {'en': 'Chamari'},
    'cdh': {'en': 'Chambeali'},
    'cdi': {'en': 'Chodri'},
    'cdj': {'en': 'Churahi'},
    'cdm': {'en': 'Chepang'},
    'cdn': {'en': 'Chaudangsi'},
    'cdo': {'en': 'Min Dong Chinese'},
    'cdr': {'en': 'Cinda-Regi-Tiyal'},
    'cds': {'en': 'Chadian Sign Language'},
    'cdy': {'en': 'Chadong'},
    'cdz': {'en': 'Koda'},
    'ce': {'af': 'Tsjetsjeens', 'am': 'ችችን', 'ar': 'الشيشانية', 'as': 'চেচেন', 'ast': 'chechenu', 'az': 'çeçen', 'az-Arab': 'چئچئنجه', 'az-Cyrl': 'чечен', 'az-Latn': 'çeçen', 'bal-Latn': 'Chechen', 'be': 'чачэнская', 'be-tarask': 'чачэнская', 'bg': 'чеченски', 'bgn': 'چیچینی', 'bn': 'চেচেন', 'br': 'tchetcheneg', 'brx': 'चेचेन', 'bs': 'čečenski', 'bs-Cyrl': 'чеченски', 'bs-Latn': 'čečenski', 'ca': 'txetxè', 'ca-ES-valencia': 'txetxè', 'ccp': '𑄌𑄬𑄌𑄬𑄚𑄴', 'ce': 'нохчийн', 'chr': 'ᏤᏤᏂ', 'ckb': 'چیچانی', 'cs': 'čečenština', 'cy': 'Tsietsieneg', 'da': 'tjetjensk', 'de': 'Tschetschenisch', 'dsb': 'čečenšćina', 'el': 'Τσετσενικά', 'el-polyton': 'Τσετσενικά', 'en': 'Chechen', 'es': 'checheno', 'et': 'tšetšeeni', 'eu': 'txetxenera', 'fa': 'چچنی', 'ff-Adlm': '𞤕𞤫𞤷𞤫𞤲𞤪𞤫', 'fi': 'tšetšeeni', 'fil': 'Chechen', 'fo': 'tjetjenskt', 'fr': 'tchétchène', 'fur': 'cecen', 'fy': 'Tsjetsjeensk', 'ga': 'Seisnis', 'gd': 'Deideanais', 'gl': 'checheno', 'gsw': 'Tschetscheenisch', 'gu': 'ચેચન', 'ha': 'Chechen', 'he': 'צ׳צ׳נית', 'hi': 'चेचन', 'hi-Latn': 'Chechen', 'hr': 'čečenski', 'hsb': 'čečenšćina', 'ht': 'tchétchène', 'hu': 'csecsen', 'hy': 'չեչեներեն', 'ia': 'checheno', 'id': 'Chechen', 'ig': 'Chechen', 'is': 'tsjetsjenska', 'it': 'ceceno', 'ja': 'チェチェン語', 'jv': 'Chechen', 'ka': 'ჩეჩნური', 'kea': 'txetxenu', 'kgp': 'sesẽnũ', 'kk': 'шешен тілі', 'kk-Cyrl': 'шешен тілі', 'km': 'ឈីឆេន', 'kn': 'ಚಚೆನ್', 'ko': '체첸어', 'kok': 'चिचेन', 'kok-Deva': 'चिचेन', 'ks': 'چیچَن', 'ks-Arab': 'چیچَن', 'ksh': 'Tschätschehnesch', 'ku': 'çeçenî', 'ky': 'чеченче', 'lb': 'Tschetschenesch', 'lkt': 'Chechen Iyápi', 'lld': 'cecen', 'lo': 'ຊີເຄນ', 'lrc': 'چئچئنی', 'lt': 'čečėnų', 'lv': 'čečenu', 'mai': 'चेचेन', 'mi': 'Tietiene', 'mk': 'чеченски', 'ml': 'ചെചൻ', 'mn': 'чечень', 'mr': 'चेचेन', 'ms': 'Chechen', 'mt': 'Chechen', 'my': 'ချက်ချန်း', 'mzn': 'چچنی', 'nb': 'tsjetsjensk', 'ne': 'चेचेन', 'nl': 'Tsjetsjeens', 'nn': 'tsjetsjensk', 'no': 'tsjetsjensk', 'nqo': 'ߗߋߗߋߣߌߞߊ߲', 'or': 'ଚେଚନ୍', 'os': 'цӕцӕйнаг', 'pa': 'ਚੇਚਨ', 'pa-Guru': 'ਚੇਚਨ', 'pcm': 'Chẹ́chẹn Lángwej', 'pl': 'czeczeński', 'ps': 'چيچني', 'pt': 'checheno', 'qu': 'Checheno Simi', 'rm': 'tschetschen', 'ro': 'cecenă', 'ru': 'чеченский', 'sah': 'Чэчиэннии', 'sc': 'cecenu', 'sd': 'چیچن', 'sd-Arab': 'چیچن', 'si': 'චෙච්නියානු', 'sk': 'čečenčina', 'sl': 'čečenščina', 'smn': 'tšetšenkielâ', 'so': 'Jejen', 'sq': 'çeçenisht', 'sr': 'чеченски', 'sr-Cyrl': 'чеченски', 'sr-Latn': 'čečenski', 'sv': 'tjetjenska', 'sw': 'Kichechenia', 'ta': 'செச்சென்', 'te': 'చెచెన్', 'th': 'เชเชน', 'ti': 'ቸቸንይና', 'tk': 'çeçen dili', 'to': 'lea fakasese', 'tr': 'Çeçence', 'ug': 'چېچىنچە', 'uk': 'чеченська', 'ur': 'چیچن', 'uz': 'chechen', 'uz-Cyrl': 'чечен тили', 'uz-Latn': 'chechen', 'vec': 'cecen', 'vi': 'Tiếng Chechen', 'xnr': 'चेचन', 'yo': 'Èdè Chechen', 'yrl': 'xexenu', 'yue': '車臣文', 'yue-Hans': '车臣文', 'yue-Hant': '車臣文', 'zh': '车臣语', 'zh-Hans': '车臣语', 'zh-Hant': '車臣文', 'zu': 'isi-Chechen'},
    'cea': {'en': 'Lower Chehalis'},
    'ceb': {'af': 'Cebuano', 'ak': 'Kebuano', 'am': 'ሴብዋኖ', 'ar': 'السيبيوانية', 'as': 'চিবুৱানো', 'ast': 'cebuanu', 'az': 'sebuan', 'az-Cyrl': 'себуан', 'az-Latn': 'sebuan', 'bal-Latn': 'Chebuánó', 'be': 'себуана', 'be-tarask': 'себуана', 'bg': 'себуански', 'bgn': 'سینوگبانونی', 'bn': 'চেবুয়ানো', 'br': 'cebuano', 'brx': 'चेबुआनो', 'bs': 'cebuano', 'bs-Cyrl': 'цебуано', 'bs-Latn': 'cebuano', 'ca': 'cebuà', 'ca-ES-valencia': 'cebuà', 'ccp': '𑄌𑄬𑄝𑄪𑄠𑄚𑄮', 'ce': 'себуано', 'ceb': 'Cebuano', 'chr': 'ᏎᏆᏃ', 'ckb': 'سێبوانۆ', 'cs': 'cebuánština', 'cy': 'Cebuano', 'da': 'cebuano', 'de': 'Cebuano', 'dsb': 'cebuanšćina', 'el': 'Σεμπουάνο', 'el-polyton': 'Σεμπουάνο', 'en': 'Cebuano', 'es': 'cebuano', 'et': 'sebu', 'eu': 'cebuanoera', 'fa': 'سبویی', 'ff-Adlm': '𞤅𞤫𞤦𞤱𞤢𞤲𞤮𞥅𞤪𞤫', 'fi': 'cebuano', 'fil': 'Cebuano', 'fo': 'cebuano', 'fr': 'cebuano', 'fy': 'Cebuano', 'ga': 'Seabúáinis', 'gd': 'Cebuano', 'gl': 'cebuano', 'gsw': 'Cebuano', 'gu': 'સિબુઆનો', 'ha': 'Cebuano', 'he': 'סבואנו', 'hi': 'सिबुआनो', 'hi-Latn': 'Cebuano', 'hr': 'cebuano', 'hsb': 'cebuanšćina', 'ht': 'cebuano', 'hu': 'szebuano', 'hy': 'սեբուերեն', 'ia': 'cebuano', 'id': 'Cebuano', 'ig': 'Cebuano', 'is': 'kebúanó', 'it': 'cebuano', 'ja': 'セブアノ語', 'jv': 'Cebuano', 'ka': 'სებუანო', 'kea': 'sebuanu', 'kgp': 'semujỹnũ', 'kk': 'себуано тілі', 'kk-Cyrl': 'себуано тілі', 'km': 'ស៊ីប៊ូអាណូ', 'kn': 'ಸೆಬುವಾನೊ', 'ko': '세부아노어', 'kok': 'सेबुआनो', 'kok-Deva': 'सेबुआनो', 'ks': 'سیباونو', 'ks-Arab': 'سیباونو', 'ksh': 'Zebuwahnesch', 'ku': 'sebwanoyî', 'ky': 'себуанча', 'lb': 'Cebuano', 'lo': 'ຊີບູໂນ', 'lt': 'sebuanų', 'lv': 'sebuāņu', 'mai': 'सेबुयानो', 'mi': 'Hepuano', 'mk': 'себуански', 'ml': 'സെബുവാനോ', 'mn': 'себуано', 'mr': 'सिबुआनो', 'ms': 'Cebuano', 'mt': 'Cebuano', 'my': 'စီဗူအာနို', 'nb': 'cebuano', 'ne': 'सेबुआनो', 'nl': 'Cebuano', 'nn': 'cebuano', 'no': 'cebuano', 'nqo': 'ߛߋߓߎߥߊߞߊ߲', 'om': 'Kubuwanoo', 'or': 'ସୀବୁଆନୋ', 'pa': 'ਸੀਬੂਆਨੋ', 'pa-Guru': 'ਸੀਬੂਆਨੋ', 'pcm': 'Sẹbuáno Lángwej', 'pl': 'cebuański', 'ps': 'سیبوانوي', 'pt': 'cebuano', 'qu': 'Cebuano Simi', 'rm': 'cebuano', 'ro': 'cebuană', 'ru': 'себуано', 'sc': 'cebuanu', 'sd': 'سبوانو', 'sd-Arab': 'سبوانو', 'si': 'සෙබුඅනො', 'sk': 'cebuánčina', 'sl': 'sebuanščina', 'smn': 'cebuanokielâ', 'so': 'Sebuano', 'sq': 'sebuanisht', 'sr': 'себуански', 'sr-Cyrl': 'себуански', 'sr-Latn': 'sebuanski', 'sv': 'cebuano', 'sw': 'Kichebuano', 'ta': 'செபுவானோ', 'te': 'సెబువానో', 'tg': 'себуано', 'th': 'เซบู', 'ti': 'ሰብዋኖ', 'tk': 'sebuan dili', 'to': 'lea fakasepuano', 'tr': 'Sebuano dili', 'tt': 'себуано', 'ug': 'سېبۇچە', 'uk': 'себуанська', 'ur': 'سیبوآنو', 'uz': 'sebuan', 'uz-Cyrl': 'себуанча', 'uz-Latn': 'sebuan', 'vec': 'sebuan', 'vi': 'Tiếng Cebuano', 'wo': 'Sibiyanoo', 'xnr': 'सिबुआनो', 'yi': 'סעבואַניש', 'yo': 'Èdè Sebuano', 'yrl': 'sebuanu', 'yue': '宿霧文', 'yue-Hans': '宿雾文', 'yue-Hant': '宿霧文', 'zh': '宿务语', 'zh-Hans': '宿务语', 'zh-Hant': '宿霧文', 'zu': 'isi-Cebuano'},
    'ceg': {'en': 'Chamacoco'},
    'cek': {'en': 'Eastern Khumi Chin'},
    'cel': {'en': 'Celtic languages'},
    'cel-bry-pro': {'en': 'Proto-Brythonic'},
    'cel-gau': {'en': 'Gaulish'},
    'cel-pro': {'en': 'Proto-Celtic'},
    'cen': {'en': 'Cen'},
    'cet': {'en': 'Centúúm'},
    'cey': {'en': 'Ekai Chin'},
    'cfa': {'en': 'Dijim-Bwilim'},
    'cfd': {'en': 'Cara'},
    'cfg': {'en': 'Como Karim'},
    'cfm': {'en': 'Falam Chin'},
    'cga': {'en': 'Changriwa'},
    'cgc': {'en': 'Kagayanen'},
    'cgg': {'af': 'Kiga', 'am': 'ቺጋኛ', 'ar': 'تشيغا', 'as': 'চিগা', 'ast': 'chiga', 'az': 'çiqa', 'az-Arab': 'چیقاجا', 'az-Cyrl': 'чига', 'az-Latn': 'çiqa', 'bal-Latn': 'Chigá', 'be': 'чыга', 'be-tarask': 'чыга', 'bg': 'чига', 'bgn': 'شیگی', 'bn': 'চিগা', 'br': 'chigaeg', 'brx': 'चीगा', 'bs': 'čiga', 'bs-Cyrl': 'чига', 'bs-Latn': 'čiga', 'ca': 'chiga', 'ca-ES-valencia': 'chiga', 'ccp': '𑄌𑄨𑄉', 'ce': 'чига', 'cgg': 'Rukiga', 'chr': 'ᏥᎦ', 'ckb': 'کیگا', 'cs': 'kiga', 'cy': 'Tsiga', 'da': 'chiga', 'de': 'Rukiga', 'dsb': 'chiga', 'el': 'Τσίγκα', 'el-polyton': 'Τσίγκα', 'en': 'Chiga', 'es': 'chiga', 'et': 'tšiga', 'eu': 'chiga', 'fa': 'چیگا', 'ff-Adlm': '𞤕𞤭𞤺𞤢𞥄𞤪𞤫', 'fi': 'kiga', 'fil': 'Chiga', 'fo': 'chiga', 'fr': 'kiga', 'fy': 'Chiga', 'ga': 'Cígis', 'gd': 'Chiga', 'gl': 'kiga', 'gu': 'ચિગા', 'ha': 'Chiga', 'he': 'צ׳יגה', 'hi': 'शिगा', 'hi-Latn': 'Chiga', 'hr': 'chiga', 'hsb': 'chiga', 'ht': 'kiga', 'hu': 'kiga', 'hy': 'չիգա', 'ia': 'chiga', 'id': 'Kiga', 'ig': 'Chiga', 'is': 'kíga', 'it': 'chiga', 'ja': 'チガ語', 'jv': 'Chiga', 'ka': 'ჩიგა', 'kea': 'xiga', 'kgp': 'siga', 'kk': 'кига тілі', 'kk-Cyrl': 'кига тілі', 'km': 'ឈីហ្កា', 'kn': 'ಚಿಗಾ', 'ko': '치가어', 'kok': 'चिगा', 'kok-Deva': 'चिगा', 'ksh': 'Kihja-Schprohch', 'ku': 'kîgayî', 'ky': 'чигача', 'lb': 'Kiga', 'lo': 'ຊີກາ', 'lrc': 'چیگا', 'lt': 'čigų', 'lv': 'kiga', 'mai': 'चिजिया', 'mi': 'Tieka', 'mk': 'чига', 'ml': 'ചിഗ', 'mn': 'чига', 'mr': 'किगा', 'ms': 'Chiga', 'mt': 'Chiga', 'my': 'ချီဂါ', 'mzn': 'چیگا', 'nb': 'kiga', 'ne': 'चिगा', 'nl': 'Chiga', 'nn': 'kiga', 'no': 'kiga', 'nqo': 'ߞߌߜ߭ߊߞߊ߲', 'or': 'ଚିଗା', 'pa': 'ਚੀਗਾ', 'pa-Guru': 'ਚੀਗਾ', 'pcm': 'Chíga Lángwej', 'pl': 'chiga', 'ps': 'چيگايي', 'pt': 'chiga', 'qu': 'Kiga Simi', 'rm': 'chiga', 'ro': 'chiga', 'ru': 'кига', 'sc': 'chiga', 'sd': 'چگا', 'sd-Arab': 'چگا', 'si': 'චිගා', 'sk': 'kiga', 'sl': 'čigajščina', 'smn': 'kigakielâ', 'so': 'Jiga', 'sq': 'çigisht', 'sr': 'чига', 'sr-Cyrl': 'чига', 'sr-Latn': 'čiga', 'sv': 'chiga', 'sw': 'Kichiga', 'ta': 'சிகா', 'te': 'చిగా', 'tg': 'Чига', 'th': 'คีกา', 'ti': 'ቺጋ', 'tk': 'kiga', 'to': 'lea fakakika', 'tr': 'Kiga', 'ug': 'چىگاچە', 'uk': 'кіга', 'ur': 'چیگا', 'uz': 'chiga', 'uz-Cyrl': 'чигача', 'uz-Latn': 'chiga', 'vec': 'ciga', 'vi': 'Tiếng Chiga', 'xnr': 'शिगा', 'yo': 'Èdè Chiga', 'yrl': 'xiga', 'yue': '奇加文', 'yue-Hans': '奇加文', 'yue-Hant': '奇加文', 'zh': '奇加语', 'zh-Hans': '奇加语', 'zh-Hant': '奇加文', 'zu': 'isi-Chiga'},
    'cgk': {'en': 'Chocangacakha'},
    'ch': {'af': 'Chamorro', 'am': 'ቻሞሮ', 'ar': 'التشامورو', 'as': 'চামোৰো', 'ast': 'chamorro', 'az': 'çamoro', 'az-Cyrl': 'чаморо', 'az-Latn': 'çamoro', 'be': 'чамора', 'be-tarask': 'чамора', 'bg': 'чаморо', 'bgn': 'چه\u200cمروری', 'bn': 'চামোরো', 'br': 'chamorru', 'brx': 'चामोर्रो', 'bs': 'čamoro', 'bs-Cyrl': 'чаморо', 'bs-Latn': 'čamoro', 'ca': 'chamorro', 'ca-ES-valencia': 'chamorro', 'ccp': '𑄌𑄟𑄮𑄢𑄮', 'ce': 'чаморро', 'chr': 'ᏣᎼᎶ', 'ckb': 'چامۆرۆ', 'cs': 'čamoro', 'cy': 'Tsiamorro', 'da': 'chamorro', 'de': 'Chamorro', 'dsb': 'čamoršćina', 'el': 'Τσαμόρο', 'el-polyton': 'Τσαμόρο', 'en': 'Chamorro', 'es': 'chamorro', 'et': 'tšamorro', 'eu': 'txamorroera', 'fa': 'چامورویی', 'ff-Adlm': '𞤕𞤢𞤥𞤮𞤪𞤮𞥅𞤪𞤫', 'fi': 'tšamorro', 'fil': 'Chamorro', 'fo': 'chamorro', 'fr': 'chamorro', 'fur': 'chamorro', 'fy': 'Chamorro', 'ga': 'Seamóiris', 'gd': 'Chamorro', 'gl': 'chamorro', 'gsw': 'Chamorro', 'gu': 'કેમોરો', 'ha': 'Chamorro', 'he': 'צ׳מורו', 'hi': 'कमोरो', 'hi-Latn': 'Chamorro', 'hr': 'chamorro', 'hsb': 'čamoršćina', 'ht': 'chamorro', 'hu': 'csamoró', 'hy': 'չամոռո', 'ia': 'chamorro', 'id': 'Chamorro', 'ig': 'Chamorro', 'is': 'kamorró', 'it': 'chamorro', 'ja': 'チャモロ語', 'jv': 'Khamorro', 'ka': 'ჩამორო', 'kgp': 'samãhu', 'kk': 'чаморро тілі', 'kk-Cyrl': 'чаморро тілі', 'km': 'ឈីម៉ូរ៉ូ', 'kn': 'ಕಮೊರೊ', 'ko': '차모로어', 'kok': 'चामोर्रो', 'kok-Deva': 'चामोर्रो', 'ks': 'کَمورو', 'ks-Arab': 'کَمورو', 'ksh': 'Chamorru', 'ku': 'çamoroyî', 'ky': 'чаморрочо', 'lb': 'Chamorro-Sprooch', 'lo': 'ຊາມໍໂຣ', 'lt': 'čamorų', 'lv': 'čamorru', 'mai': 'चमोरो', 'mi': 'Tiamoro', 'mk': 'чаморски', 'ml': 'ചമോറോ', 'mn': 'чаморро', 'mr': 'कॅमोरो', 'ms': 'Chamorro', 'mt': 'Chamorro', 'my': 'ချမိုရို', 'nb': 'chamorro', 'ne': 'चामोर्रो', 'nl': 'Chamorro', 'nn': 'chamorro', 'no': 'chamorro', 'or': 'ଚାମୋରୋ', 'pa': 'ਚਮੋਰੋ', 'pa-Guru': 'ਚਮੋਰੋ', 'pcm': 'Chamóro Lángwej', 'pl': 'czamorro', 'ps': 'چمورو', 'pt': 'chamorro', 'qu': 'Chamorro Simi', 'rm': 'chamorro', 'ro': 'chamorro', 'ru': 'чаморро', 'sc': 'chamorru', 'sd': 'چمورو', 'sd-Arab': 'چمورو', 'si': 'චමොරො', 'sk': 'čamorčina', 'sl': 'čamorščina', 'smn': 'chamorrokielâ', 'so': 'Chamorro', 'sq': 'kamoroisht', 'sr': 'чаморо', 'sr-Cyrl': 'чаморо', 'sr-Latn': 'čamoro', 'sv': 'chamorro', 'sw': 'Kichamorro', 'ta': 'சாமோரோ', 'te': 'చమర్రో', 'th': 'ชามอร์โร', 'ti': 'ቻሞሮ', 'tk': 'çamorro', 'to': 'lea fakakamolo', 'tr': 'Çamorro dili', 'ug': 'چامورروچە', 'uk': 'чаморро', 'ur': 'چیمارو', 'uz': 'chamorro', 'uz-Cyrl': 'чаморро', 'uz-Latn': 'chamorro', 'vec': 'ciamoro', 'vi': 'Tiếng Chamorro', 'xnr': 'कमोरो', 'yo': 'Èdè S̩amoro', 'yrl': 'xamuru', 'yue': '查莫洛文', 'yue-Hans': '查莫洛文', 'yue-Hant': '查莫洛文', 'zh': '查莫罗语', 'zh-Hans': '查莫罗语', 'zh-Hant': '查莫洛文', 'zu': 'isi-Chamorro'},
    'chb': {'am': 'ቺብቻ', 'ar': 'التشيبشا', 'ast': 'chibcha', 'az': 'çibça', 'az-Latn': 'çibça', 'be': 'чыбча', 'be-tarask': 'чыбча', 'bg': 'чибча', 'bn': 'চিবচা', 'br': 'chibcha', 'brx': 'चीबचा', 'bs': 'čibča', 'bs-Cyrl': 'чибча', 'bs-Latn': 'čibča', 'ca': 'txibtxa', 'ca-ES-valencia': 'txibtxa', 'ccp': '𑄌𑄨𑄛𑄴𑄌', 'cs': 'čibča', 'da': 'chibcha', 'de': 'Chibcha', 'el': 'Τσίμπτσα', 'el-polyton': 'Τσίμπτσα', 'en': 'Chibcha', 'es': 'chibcha', 'et': 'tšibtša', 'fa': 'چیبچا', 'fi': 'tšibtša', 'fr': 'chibcha', 'fy': 'Chibcha', 'gd': 'Chibcha', 'gsw': 'Tschibtscha', 'gu': 'ચિબ્ચા', 'he': 'צ׳יבצ׳ה', 'hi': 'चिब्चा', 'hi-Latn': 'Chibcha', 'hr': 'čibča', 'ht': 'chibcha', 'hu': 'csibcsa', 'id': 'Chibcha', 'is': 'síbsja', 'it': 'chibcha', 'ja': 'チブチャ語', 'ka': 'ჩიბჩა', 'kgp': 'simsa', 'kn': 'ಚಿಬ್ಚಾ', 'ko': '치브차어', 'ks': 'چیٖبچا', 'ks-Arab': 'چیٖبچا', 'lb': 'Chibcha-Sprooch', 'lo': 'ຊິບຊາ', 'lt': 'čibčų', 'lv': 'čibču', 'mk': 'чибча', 'ml': 'ചിബ്ച', 'mr': 'चिब्चा', 'mt': 'Chibcha', 'nb': 'chibcha', 'ne': 'चिब्चा', 'nl': 'Chibcha', 'nn': 'chibcha', 'no': 'chibcha', 'or': 'ଚିବ୍ଚା', 'pl': 'czibcza', 'pt': 'chibcha', 'rm': 'chibcha', 'ro': 'chibcha', 'ru': 'чибча', 'sk': 'čibča', 'sl': 'čibčevščina', 'sr': 'чипча', 'sr-Cyrl': 'чипча', 'sr-Latn': 'čipča', 'sv': 'chibcha', 'ta': 'சிப்சா', 'te': 'చిబ్చా', 'th': 'ชิบชา', 'to': 'lea fakasīpisa', 'tr': 'Çibça dili', 'ug': 'چىبچاچە', 'uk': 'чібча', 'vi': 'Tiếng Chibcha', 'yrl': 'xibixa', 'yue': '奇布查文', 'yue-Hans': '奇布查文', 'yue-Hant': '奇布查文', 'zh': '奇布查语', 'zh-Hans': '奇布查语', 'zh-Hant': '奇布查文'},
    'chc': {'en': 'Catawba'},
    'chd': {'en': 'Highland Oaxaca Chontal'},
    'chf': {'en': 'Tabasco Chontal'},
    'chg': {'am': 'ቻጋታይ', 'ar': 'التشاجاتاي', 'ast': 'chagatai', 'az': 'çağatay', 'az-Latn': 'çağatay', 'bg': 'чагатай', 'bn': 'চাগাতাই', 'brx': 'चगताई', 'bs': 'čagatai', 'bs-Cyrl': 'чагатаи', 'bs-Latn': 'čagatai', 'ca': 'txagatai', 'ca-ES-valencia': 'txagatai', 'ccp': '𑄌𑄉𑄖𑄳𑄆', 'cs': 'čagatajština', 'da': 'chagatai', 'de': 'Tschagataisch', 'el': 'Τσαγκατάι', 'el-polyton': 'Τσαγκατάι', 'en': 'Chagatai', 'es': 'chagatái', 'et': 'tšagatai', 'fa': 'جغتایی', 'fi': 'tšagatai', 'fr': 'tchaghataï', 'fy': 'Chagatai', 'gd': 'Chagatai', 'gsw': 'Tschagataisch', 'gu': 'છગાતાઇ', 'he': 'צ׳אגאטאי', 'hi': 'छगाताई', 'hi-Latn': 'Chagatai', 'hr': 'čagatajski', 'ht': 'tchaghataï', 'hu': 'csagatáj', 'id': 'Chagatai', 'is': 'sjagataí', 'it': 'ciagataico', 'ja': 'チャガタイ語', 'kgp': 'sagataj', 'kn': 'ಚಗಟಾಯ್', 'ko': '차가타이어', 'ks': 'چھَگتاے', 'ks-Arab': 'چھَگتاے', 'lb': 'Tschagataesch', 'lo': 'ຊາກາໄຕ', 'lt': 'čagatų', 'lv': 'džagatajs', 'mk': 'чагатајски', 'ml': 'ഷാഗതായ്', 'mr': 'छागाताइ', 'mt': 'Chagatai', 'nb': 'tsjagatai', 'ne': 'चागाटाई', 'nl': 'Chagatai', 'nn': 'tsjagataisk', 'no': 'tsjagatai', 'or': 'ଛଗତାଇ', 'pl': 'czagatajski', 'pt': 'chagatai', 'rm': 'tschagataic', 'ro': 'chagatai', 'ru': 'чагатайский', 'sk': 'čagatajčina', 'sl': 'čagatajščina', 'sr': 'чагатај', 'sr-Cyrl': 'чагатај', 'sr-Latn': 'čagataj', 'sv': 'chagatai', 'ta': 'ஷகதை', 'te': 'చాగటై', 'th': 'ชะกะไต', 'to': 'lea fakasakatāi', 'tr': 'Çağatayca', 'ug': 'چاغاتايچە', 'uk': 'чагатайська', 'vi': 'Tiếng Chagatai', 'yrl': 'xagatai', 'yue': '查加文', 'yue-Hans': '查加文', 'yue-Hant': '查加文', 'zh': '察合台语', 'zh-Hans': '察合台语', 'zh-Hant': '查加文'},
    'chh': {'en': 'Chinook'},
    'chi-pro': {'en': 'Proto-Chimakuan'},
    'chj': {'en': 'Ojitlán Chinantec'},
    'chk': {'af': 'Chuukees', 'am': 'ቹክስ', 'ar': 'التشكيزية', 'as': 'চুকিজ', 'ast': 'chuukés', 'az': 'çukiz', 'az-Cyrl': 'чукиз', 'az-Latn': 'çukiz', 'be': 'чуук', 'be-tarask': 'чуук', 'bg': 'чуук', 'bgn': 'تروسکی', 'bn': 'চুকিজ', 'br': 'chuuk', 'brx': 'चुकेसे', 'bs': 'čukeski', 'bs-Cyrl': 'чукески', 'bs-Latn': 'čukeski', 'ca': 'chuuk', 'ca-ES-valencia': 'chuuk', 'ccp': '𑄌𑄪𑄇𑄨', 'ce': 'чукчийн', 'chr': 'ᏧᎨᏎ', 'ckb': 'چووکی', 'cs': 'čukština', 'cy': 'Chuukaeg', 'da': 'chuukese', 'de': 'Chuukesisch', 'dsb': 'chuukezišćina', 'el': 'Τσουκίζι', 'el-polyton': 'Τσουκίζι', 'en': 'Chuukese', 'es': 'trukés', 'et': 'tšuugi', 'eu': 'chuukera', 'fa': 'چوکی', 'ff-Adlm': '𞤕𞤵𞥅𞤳𞤵𞥅𞤪𞤫', 'fi': 'chuuk', 'fil': 'Chuukese', 'fo': 'chuukese', 'fr': 'chuuk', 'fy': 'Chuukeesk', 'ga': 'Siúicísis', 'gd': 'Cànan Chuuk', 'gl': 'chuuk', 'gsw': 'Trukesisch', 'gu': 'ચૂકીસ', 'ha': 'Chuukese', 'he': 'צ׳וקסה', 'hi': 'चूकीस', 'hi-Latn': 'Chuukese', 'hr': 'chuukese', 'hsb': 'chuukezišćina', 'ht': 'chuuk', 'hu': 'csukéz', 'hy': 'տրուկերեն', 'ia': 'chuukese', 'id': 'Chuuke', 'ig': 'Chuukese', 'is': 'sjúkíska', 'it': 'chuukese', 'ja': 'チューク語', 'jv': 'Chuukese', 'ka': 'ჩუკოტკური', 'kgp': 'sukese', 'kk': 'чуук тілі', 'kk-Cyrl': 'чуук тілі', 'km': 'ឈូគី', 'kn': 'ಚೂಕಿಸೆ', 'ko': '추크어', 'kok': 'चुकीज', 'kok-Deva': 'चुकीज', 'ks': 'چُکیٖز', 'ks-Arab': 'چُکیٖز', 'ksh': 'Trukehsesch', 'ku': 'çûkî', 'ky': 'чуукиче', 'lb': 'Trukesesch', 'lo': 'ຊູເກດ', 'lt': 'čukesų', 'lv': 'čūku', 'mai': 'चुकीज', 'mi': 'Tiukīhi', 'mk': 'чучки', 'ml': 'ചൂകീസ്', 'mn': 'чуук', 'mr': 'चूकीसे', 'ms': 'Chukese', 'mt': 'Ċukiż', 'my': 'ချူကီးစ်', 'nb': 'chuukesisk', 'ne': 'चुकेसे', 'nl': 'Chuukees', 'nn': 'chuukesisk', 'no': 'chuukesisk', 'or': 'ଚୁକୀସେ', 'pa': 'ਚੂਕੀਸ', 'pa-Guru': 'ਚੂਕੀਸ', 'pcm': 'Chuukís Lángwej', 'pl': 'chuuk', 'ps': 'چواوکي', 'pt': 'chuukese', 'qu': 'Chuukese Simi', 'rm': 'chuukais', 'ro': 'chuukese', 'ru': 'чукотский', 'sc': 'chuukesu', 'sd': 'چڪيز', 'sd-Arab': 'چڪيز', 'si': 'චූකීස්', 'sk': 'chuuk', 'sl': 'trukeščina', 'smn': 'chuukkielâ', 'so': 'Chuukese', 'sq': 'çukezisht', 'sr': 'чучки', 'sr-Cyrl': 'чучки', 'sr-Latn': 'čučki', 'sv': 'chuukesiska', 'sw': 'Kichukisi', 'ta': 'சூகிசே', 'te': 'చూకీస్', 'th': 'ชูก', 'ti': 'ቹኪዝኛ', 'tk': 'çuuk dili', 'to': 'lea fakatūke', 'tr': 'Chuukese', 'ug': 'چۇكچە', 'uk': 'чуукська', 'ur': 'چوکیز', 'uz': 'chukot', 'uz-Cyrl': 'чуукча', 'uz-Latn': 'chukot', 'vec': 'ciukeze', 'vi': 'Tiếng Chuuk', 'xnr': 'चूकीस', 'yo': 'Èdè Shuki', 'yrl': 'xukisi', 'yue': '處奇斯文', 'yue-Hans': '处奇斯文', 'yue-Hant': '處奇斯文', 'zh': '楚克语', 'zh-Hans': '楚克语', 'zh-Hant': '處奇斯文', 'zu': 'isi-Chuukese'},
    'chl': {'en': 'Cahuilla'},
    'chm': {'af': 'Mari', 'am': 'ማሪ', 'ar': 'الماري', 'as': 'মাৰি', 'ast': 'mari', 'az': 'mari', 'az-Cyrl': 'мари', 'az-Latn': 'mari', 'be': 'мары', 'be-tarask': 'мары', 'bg': 'марийски', 'bgn': 'ماری', 'bn': 'মারি', 'br': 'marieg', 'brx': 'मारि', 'bs': 'mari', 'bs-Cyrl': 'мари', 'bs-Latn': 'mari', 'ca': 'mari', 'ca-ES-valencia': 'mari', 'ccp': '𑄟𑄢𑄨', 'ce': 'марийн', 'chr': 'ᎹᎵ', 'ckb': 'ماری', 'cs': 'marijština', 'cy': 'Marieg', 'da': 'mari', 'de': 'Mari', 'dsb': 'marišćina', 'el': 'Μάρι', 'el-polyton': 'Μάρι', 'en': 'Mari', 'es': 'marí', 'et': 'mari', 'eu': 'mariera', 'fa': 'ماریایی', 'ff-Adlm': '𞤃𞤢𞤪𞤭𞥅𞤪𞤫', 'fi': 'mari', 'fil': 'Mari', 'fo': 'mari', 'fr': 'mari', 'fy': 'Mari', 'ga': 'Mairis', 'gd': 'Mari', 'gl': 'mari', 'gsw': 'Tscheremissisch', 'gu': 'મારી', 'ha': 'Mari', 'he': 'מארי', 'hi': 'मारी', 'hi-Latn': 'Mari', 'hr': 'marijski', 'hsb': 'marišćina', 'ht': 'mari', 'hu': 'mari', 'hy': 'մարի', 'ia': 'mari', 'id': 'Mari', 'ig': 'Mari', 'is': 'marí', 'it': 'mari', 'ja': 'マリ語', 'jv': 'Mari', 'ka': 'მარიული', 'kgp': 'mỹri', 'kk': 'мари тілі', 'kk-Cyrl': 'мари тілі', 'km': 'ម៉ារី', 'kn': 'ಮಾರಿ', 'ko': '마리어', 'kok': 'मारी', 'kok-Deva': 'मारी', 'ks': 'ماری', 'ks-Arab': 'ماری', 'ksh': 'Mahresch', 'ku': 'marî', 'ky': 'мариче', 'lb': 'Mari', 'lkt': 'Mari Iyápi', 'lo': 'ມາຣິ', 'lt': 'marių', 'lv': 'mariešu', 'mai': 'मारी', 'mi': 'Mari', 'mk': 'мариски', 'ml': 'മാരി', 'mn': 'мари хэл', 'mr': 'मारी', 'ms': 'Mari', 'mt': 'Mari', 'my': 'မာရီ', 'nb': 'mari', 'ne': 'मारी', 'nl': 'Mari', 'nn': 'mari', 'no': 'mari', 'or': 'ମାରୀ', 'pa': 'ਮਾਰੀ', 'pa-Guru': 'ਮਾਰੀ', 'pcm': 'Mari Lángwej', 'pl': 'maryjski', 'ps': 'ماري', 'pt': 'mari', 'qu': 'Mari Simi', 'rm': 'mari', 'ro': 'mari', 'ru': 'марийский', 'sc': 'mari', 'sd': 'ماري', 'sd-Arab': 'ماري', 'se': 'marigiella', 'si': 'මරි', 'sk': 'marijčina', 'sl': 'marijščina', 'smn': 'marikielâ', 'so': 'Mari', 'sq': 'marisht', 'sr': 'мари', 'sr-Cyrl': 'мари', 'sr-Latn': 'mari', 'sv': 'mariska', 'sw': 'Kimari', 'ta': 'மாரி', 'te': 'మారి', 'tg': 'марӣ', 'th': 'มารี', 'ti': 'ማሪ', 'tk': 'mariý dili', 'to': 'lea fakamalī', 'tr': 'Mari dili', 'tt': 'мари', 'ug': 'مارىچە', 'uk': 'марійська', 'ur': 'ماری', 'uz': 'mari', 'uz-Cyrl': 'мари', 'uz-Latn': 'mari', 'vec': 'mari', 'vi': 'Tiếng Mari', 'wo': 'Mari', 'xnr': 'मारी', 'yo': 'Èdè Mari', 'yrl': 'mari', 'yue': '馬里文', 'yue-Hans': '马里文', 'yue-Hant': '馬里文', 'zh': '马里语', 'zh-Hans': '马里语', 'zh-Hant': '馬里文', 'zu': 'isi-Mari'},
    'chn': {'am': 'ቺኑክ ጃርጎን', 'ar': 'الشينوك جارجون', 'ast': 'xíriga chinook', 'az': 'çinuk ləhçəsi', 'az-Latn': 'çinuk ləhçəsi', 'bg': 'жаргон чинуук', 'bn': 'চিনুক জার্গন', 'brx': 'चीनूक् जार्गन्', 'bs': 'činukski žargon', 'bs-Cyrl': 'чинукски', 'bs-Latn': 'činukski žargon', 'ca': 'pidgin chinook', 'ca-ES-valencia': 'pidgin chinook', 'ccp': '𑄌𑄨𑄚𑄪𑄇𑄴 𑄎𑄢𑄴𑄉𑄧𑄚𑄴', 'cs': 'činuk pidžin', 'da': 'chinook', 'de': 'Chinook', 'el': 'Ιδιωματικά Σινούκ', 'el-polyton': 'Ιδιωματικά Σινούκ', 'en': 'Chinook Jargon', 'es': 'jerga chinuk', 'et': 'tšinuki žargoon', 'fi': 'chinook-jargon', 'fr': 'jargon chinook', 'fy': 'Chinook-jargon', 'gd': 'Chinuk Wawa', 'gsw': 'Chinook', 'gu': 'ચિનૂક જાર્ગન', 'he': 'ניב צ׳ינוק', 'hi': 'चिनूक जारगॉन', 'hi-Latn': 'Chinook Jargon', 'hr': 'chinook žargon', 'ht': 'jargon chinook', 'hu': 'csinuk zsargon', 'id': 'Jargon Chinook', 'is': 'sínúk', 'it': 'gergo chinook', 'ja': 'チヌーク混成語', 'ka': 'ჩინუკის ჟარგონი', 'kgp': 'jargỹ sinũki', 'kn': 'ಚಿನೂಕ್ ಜಾರ್ಗೋನ್', 'ko': '치누크 자곤', 'ks': 'چِنوٗک جارگَن', 'ks-Arab': 'چِنوٗک جارگَن', 'lb': 'Chinook', 'lo': 'ຊີນຸກຈາກອນ', 'lt': 'činuk žargonas', 'lv': 'činuku žargons', 'mk': 'чинучки жаргон', 'ml': 'ചിനൂഗ് ജാർഗൺ', 'mr': 'चिनूक जारगॉन', 'mt': 'Chinook Jargon', 'nb': 'chinook', 'ne': 'चिनुक जार्गन', 'nl': 'Chinook Jargon', 'nn': 'chinook', 'no': 'chinook', 'or': 'ଚିନୁକ୍ ଜାରଗାଁନ୍', 'pl': 'żargon czinucki', 'pt': 'jargão Chinook', 'rm': 'patuà chinook', 'ro': 'jargon chinook', 'ru': 'чинук жаргон', 'sk': 'činucký žargón', 'sl': 'činuški žargon', 'sr': 'чинучки', 'sr-Cyrl': 'чинучки', 'sr-Latn': 'činučki', 'sv': 'chinook', 'ta': 'சினூக் ஜார்கான்', 'te': 'చినూక్ జార్గన్', 'th': 'ชินุกจาร์กอน', 'to': 'lea fakasinuki-takote', 'tr': 'Çinuk dili', 'ug': 'چىنۇك-ژارگونچە', 'uk': 'чинук жаргон', 'vi': 'Biệt ngữ Chinook', 'yrl': 'yarigãu xinoki', 'yue': '契奴克文', 'yue-Hans': '契奴克文', 'yue-Hant': '契奴克文', 'zh': '奇努克混合语', 'zh-Hans': '奇努克混合语', 'zh-Hant': '契奴克文'},
    'cho': {'af': 'Choctaw', 'am': 'ቾክታዋ', 'ar': 'الشوكتو', 'as': 'চোক্টাউ', 'ast': 'choctaw', 'az': 'çoktau', 'az-Cyrl': 'чоктау', 'az-Latn': 'çoktau', 'bal-Latn': 'Choktaw', 'be': 'чокта', 'be-tarask': 'чокта', 'bg': 'чокто', 'bgn': 'چاکتاوی', 'bn': 'চকটাও', 'br': 'choktaw', 'brx': 'चक्टाऊ', 'bs': 'čoktav', 'bs-Cyrl': 'чоктавски', 'bs-Latn': 'čoktav', 'ca': 'choctaw', 'ca-ES-valencia': 'choctaw', 'ccp': '𑄌𑄧𑄇𑄴𑄑𑄳𑄅𑄧𑄠𑄧', 'ce': 'чоктавийн', 'cho': 'Chahta', 'chr': 'ᎠᏣᏓ', 'ckb': 'چۆکتاو', 'cs': 'čoktština', 'cy': 'Siocto', 'da': 'choctaw', 'de': 'Choctaw', 'dsb': 'choctawšćina', 'el': 'Τσόκτο', 'el-polyton': 'Τσόκτο', 'en': 'Choctaw', 'es': 'choctaw', 'et': 'tšokto', 'eu': 'txoktawera', 'fa': 'چوکتویی', 'ff-Adlm': '𞤕𞤢𞤸𞤼𞤢𞥄𞤪𞤫', 'fi': 'choctaw', 'fil': 'Choctaw', 'fo': 'choctaw', 'fr': 'choctaw', 'fy': 'Choctaw', 'ga': 'Seactáis', 'gd': 'Choctaw', 'gl': 'choctaw', 'gsw': 'Choctaw', 'gu': 'ચોક્તૌ', 'ha': 'Choctaw', 'he': 'צ׳וקטאו', 'hi': 'चोक्तौ', 'hi-Latn': 'Choctaw', 'hr': 'choctaw', 'hsb': 'choctawšćina', 'ht': 'choctaw', 'hu': 'csoktó', 'hy': 'չոկտո', 'ia': 'choctaw', 'id': 'Koktaw', 'ig': 'Choctaw', 'is': 'sjoktá', 'it': 'choctaw', 'ja': 'チョクトー語', 'jv': 'Choctaw', 'ka': 'ჩოკტო', 'kgp': 'sogtavo', 'kk': 'чокто тілі', 'kk-Cyrl': 'чокто тілі', 'km': 'ឆុកតាវ', 'kn': 'ಚೋಕ್ಟಾವ್', 'ko': '촉토어', 'kok': 'चाकता', 'kok-Deva': 'चाकता', 'ks': 'چوکتَو', 'ks-Arab': 'چوکتَو', 'ksh': 'Tschoktoh', 'ku': 'çoktavî', 'ky': 'чокточо', 'lb': 'Choctaw', 'lo': 'ຊອກຕິວ', 'lt': 'čoktau', 'lv': 'čoktavu', 'mai': 'चोकटाउ', 'mi': 'Tiokatō', 'mk': 'чоктавски', 'ml': 'ചോക്റ്റാവ്', 'mn': 'чоктау', 'mr': 'चोक्तौ', 'ms': 'Choctaw', 'mt': 'Choctaw', 'my': 'ချော့တို', 'nb': 'choctaw', 'ne': 'चोक्टाव', 'nl': 'Choctaw', 'nn': 'choctaw', 'no': 'choctaw', 'or': 'ଚୋଟୱା', 'pa': 'ਚੌਕਟੋ', 'pa-Guru': 'ਚੌਕਟੋ', 'pcm': 'Shọ́ktau Lángwej', 'pl': 'czoktawski', 'ps': 'چوکټاوي', 'pt': 'choctaw', 'qu': 'Choctaw Simi', 'rm': 'choctaw', 'ro': 'choctaw', 'ru': 'чоктавский', 'sc': 'choctaw', 'sd': 'چوڪ تو', 'sd-Arab': 'چوڪ تو', 'si': 'චොක්ටොව්', 'sk': 'čoktčina', 'sl': 'čoktavščina', 'smn': 'choctawkielâ', 'so': 'Choctaw', 'sq': 'çoktauisht', 'sr': 'чоктавски', 'sr-Cyrl': 'чоктавски', 'sr-Latn': 'čoktavski', 'sv': 'choctaw', 'sw': 'Kichoktao', 'ta': 'சோக்தௌ', 'te': 'చక్టా', 'th': 'ช็อกทอว์', 'ti': 'ቾክቶ', 'tk': 'çokto', 'to': 'lea fakasokitau', 'tr': 'Çoktav dili', 'ug': 'چوكتاۋچە', 'uk': 'чокто', 'ur': 'چاکٹاؤ', 'uz': 'choktav', 'uz-Cyrl': 'чоктавча', 'uz-Latn': 'choktav', 'vec': 'ciòctaw', 'vi': 'Tiếng Choctaw', 'xnr': 'चोक्तौ', 'yo': 'Èdè Shokita', 'yrl': 'xokitau', 'yue': '喬克托文', 'yue-Hans': '乔克托文', 'yue-Hant': '喬克托文', 'zh': '乔克托语', 'zh-Hans': '乔克托语', 'zh-Hant': '喬克托文', 'zu': 'isi-Choctaw'},
    'chp': {'af': 'Chipewyan', 'am': 'ቺፔውያን', 'ar': 'الشيباوايان', 'as': 'চিপেউয়ান', 'ast': 'chipewyanu', 'az': 'çipevyan', 'az-Latn': 'çipevyan', 'be': 'чыпеваян', 'be-tarask': 'чыпеваян', 'bg': 'чипеуански', 'bn': 'চিপেওয়ান', 'br': 'chipewyan', 'brx': 'चिपेव्यान', 'bs': 'čipvijanski', 'bs-Cyrl': 'чипвијански', 'bs-Latn': 'čipvijanski', 'ca': 'chipewyan', 'ca-ES-valencia': 'chipewyan', 'ccp': '𑄌𑄨𑄛𑄮𑄤𑄚𑄴', 'chr': 'ᎠᏥᏇᏯᏂ', 'cs': 'čipevajština', 'cy': 'Chipewyan', 'da': 'chipewyan', 'de': 'Chipewyan', 'dsb': 'chipewyanšćina', 'el': 'Τσίπιουαν', 'el-polyton': 'Τσίπιουαν', 'en': 'Chipewyan', 'es': 'chipewyan', 'et': 'tšipevai', 'eu': 'chipewyera', 'fa': 'چیپه\u200cویه\u200cای', 'ff-Adlm': '𞤕𞤭𞤨𞤴𞤢𞤲𞤪𞤫', 'fi': 'chipewyan', 'fil': 'Chipewyan', 'fr': 'chipewyan', 'fy': 'Chipewyan', 'ga': 'Siopúáinis', 'gd': 'Chipewyan', 'gl': 'chipewyan', 'gsw': 'Chipewyan', 'gu': 'શિપેવ્યાન', 'ha': 'Chipewyan', 'he': 'צ׳יפוויאן', 'hi': 'शिपेव्यान', 'hi-Latn': 'Chipewyan', 'hr': 'chipewyan', 'hsb': 'chipewyanšćina', 'ht': 'chipewyan', 'hu': 'csipevé', 'hy': 'չիպևայան', 'ia': 'chipewyan', 'id': 'Chipewyan', 'ig': 'Chipewyan', 'is': 'sípevíska', 'it': 'chipewyan', 'ja': 'チペワイアン語', 'jv': 'Chipewyan', 'ka': 'ჩიპევიანი', 'kgp': 'sipevyjỹ', 'kk': 'чипевайан тілі', 'kk-Cyrl': 'чипевайан тілі', 'km': 'ឈីប៉េវ៉ាយអិន', 'kn': 'ಚಿಪೆವ್ಯಾನ್', 'ko': '치페우얀', 'kok': 'चिपेवायन', 'kok-Deva': 'चिपेवायन', 'ks': 'شیپویان', 'ks-Arab': 'شیپویان', 'ku': 'çîpevyayî', 'ky': 'чипевайанча', 'lb': 'Chipewyan', 'lo': 'ຊີພິວຢານ', 'lt': 'čipvėjų', 'lv': 'čipevaianu', 'mai': 'चिपवायन', 'mi': 'Tiepewaiana', 'mk': 'чипевјански', 'ml': 'ചിപേവ്യൻ', 'mn': 'чипевайан', 'mr': 'शिपेव्यान', 'ms': 'Chipewyan', 'mt': 'Ċipewjan', 'my': 'ချီပွေဝိုင်ရန်', 'nb': 'chipewiansk', 'ne': 'चिपेव्यान', 'nl': 'Chipewyan', 'nn': 'chipewiansk', 'no': 'chipewiansk', 'or': 'ଚିପେୱାନ୍', 'pa': 'ਚਿਪਵਿਆਨ', 'pa-Guru': 'ਚਿਪਵਿਆਨ', 'pcm': 'Shípẹwián Lángwej', 'pl': 'czipewiański', 'ps': 'Ch', 'pt': 'chipewyan', 'qu': 'Chipewyan Simi', 'rm': 'chipewyan', 'ro': 'chipewyan', 'ru': 'чипевьян', 'sc': 'chipewyan', 'sd': 'چائپائن', 'sd-Arab': 'چائپائن', 'si': 'චිපෙවියන්', 'sk': 'čipevajčina', 'sl': 'čipevščina', 'so': 'Chipewyan', 'sq': 'çipeuajanisht', 'sr': 'чипевјански', 'sr-Cyrl': 'чипевјански', 'sr-Latn': 'čipevjanski', 'sv': 'chipewyan', 'sw': 'Kichipewyani', 'ta': 'சிபெவ்யான்', 'te': 'చిపెవ్యాన్', 'th': 'ชิพิวยัน', 'ti': 'ቺፐውያን', 'tk': 'çipewýan dili', 'to': 'lea fakasipeuiani', 'tr': 'Çipevya dili', 'ug': 'چىپېۋيانچە', 'uk': 'чипевʼян', 'ur': 'چپوائن', 'uz': 'chipevyan', 'uz-Latn': 'chipevyan', 'vec': 'cìpejan', 'vi': 'Tiếng Chipewyan', 'xnr': 'शिपेव्यान', 'yo': 'Èdè Shipewa', 'yrl': 'xipewiyã', 'yue': '奇佩瓦揚文', 'yue-Hans': '奇佩瓦扬文', 'yue-Hant': '奇佩瓦揚文', 'zh': '奇佩维安语', 'zh-Hans': '奇佩维安语', 'zh-Hant': '奇佩瓦揚文', 'zu': 'isi-Chipewyan'},
    'chq': {'en': 'Quiotepec Chinantec'},
    'chr': {'af': 'Cherokees', 'ak': 'Kiroki', 'am': 'ቼሮኬኛ', 'ar': 'الشيروكي', 'as': 'চেৰোকি', 'ast': 'cheroqui', 'az': 'çeroki', 'az-Arab': 'چئروکیجه', 'az-Cyrl': 'чероки', 'az-Latn': 'çeroki', 'bal-Latn': 'Cheruki', 'be': 'чэрокі', 'be-tarask': 'чэрокі', 'bg': 'черокски', 'bgn': 'چیروکی', 'bn': 'চেরোকী', 'br': 'cherokee', 'brx': 'चेरकी', 'bs': 'čeroki', 'bs-Cyrl': 'чероки', 'bs-Latn': 'čeroki', 'ca': 'cherokee', 'ca-ES-valencia': 'cherokee', 'ccp': '𑄌𑄬𑄢𑄮𑄇𑄩', 'ce': 'чероки', 'chr': 'ᏣᎳᎩ', 'ckb': 'چێرۆکی', 'cs': 'čerokézština', 'csw': 'ᒉᕑᐅᑫᕀ', 'cy': 'Tsierocî', 'da': 'cherokee', 'de': 'Cherokee', 'dsb': 'cherokee', 'el': 'Τσερόκι', 'el-polyton': 'Τσερόκι', 'en': 'Cherokee', 'en-Dsrt': '𐐕𐐯𐑉𐐬𐐿𐐨', 'es': 'cheroqui', 'et': 'tšerokii', 'eu': 'txerokiera', 'fa': 'چروکی', 'ff-Adlm': '𞤕𞤫𞥅𞤪𞤮𞤳𞤭𞥅𞤪𞤫', 'fi': 'cherokee', 'fil': 'Cherokee', 'fo': 'cherokee', 'fr': 'cherokee', 'fy': 'Cherokee', 'ga': 'Seiricis', 'gd': 'Cherokee', 'gl': 'cherokee', 'gsw': 'Cherokee', 'gu': 'શેરોકી', 'ha': 'Cherokee', 'he': 'צ׳רוקי', 'hi': 'चेरोकी', 'hi-Latn': 'Cherokee', 'hr': 'čerokijski', 'hsb': 'cherokee', 'ht': 'cherokee', 'hu': 'cseroki', 'hy': 'չերոկի', 'ia': 'cherokee', 'id': 'Cherokee', 'ig': 'Cherokee', 'is': 'Cherokee-mál', 'it': 'cherokee', 'ja': 'チェロキー語', 'jv': 'Cherokee', 'ka': 'ჩეროკი', 'kea': 'xeroki', 'kgp': 'seroki', 'kk': 'чероки тілі', 'kk-Cyrl': 'чероки тілі', 'km': 'ឆេរូគី', 'kn': 'ಚೆರೋಕಿ', 'ko': '체로키어', 'kok': 'चॅरोकी', 'kok-Deva': 'चॅरोकी', 'ks': 'چیروکی', 'ks-Arab': 'چیروکی', 'ksh': 'Tschärrokeh', 'ku': 'çerokî', 'kxv': 'cerokī', 'kxv-Deva': 'चेरोकी', 'kxv-Latn': 'cerokī', 'kxv-Orya': 'ଚେରକୀ', 'kxv-Telu': 'చెరొకీ', 'ky': 'черокиче', 'lb': 'Cherokee', 'lkt': 'Cherokee Iyápi', 'lld': 'cherokee', 'lo': 'ຊີໂຣກີ', 'lrc': 'چوروٙکی', 'lt': 'čerokių', 'lv': 'čiroku', 'mai': 'चेरोकी', 'mi': 'Tierokī', 'mk': 'чероки', 'ml': 'ഷെരോക്കി', 'mn': 'чероки', 'mr': 'चेरोकी', 'ms': 'Cherokee', 'mt': 'Cherokee', 'my': 'ချာရိုကီ', 'mzn': 'چروکیایی', 'nb': 'cherokesisk', 'ne': 'चेरोकी', 'nl': 'Cherokee', 'nn': 'cherokesisk', 'no': 'cherokesisk', 'nqo': 'ߗߋߙߏߞߌߞߊ߲', 'om': 'Cherokee', 'or': 'ଚେରୋକୀ', 'pa': 'ਚੇਰੋਕੀ', 'pa-Guru': 'ਚੇਰੋਕੀ', 'pcm': 'Chẹ́rókii Lángwej', 'pl': 'czirokeski', 'ps': 'چېروکي', 'pt': 'cheroqui', 'qu': 'Cheroqui Simi', 'rm': 'cherokee', 'ro': 'cherokee', 'ru': 'чероки', 'sc': 'cherokee', 'sd': 'چروڪي', 'sd-Arab': 'چروڪي', 'si': 'චෙරොකී', 'sk': 'čerokí', 'sl': 'čerokeščina', 'smn': 'cherokeekielâ', 'so': 'Jerookee', 'sq': 'çerokisht', 'sr': 'чероки', 'sr-Cyrl': 'чероки', 'sr-Latn': 'čeroki', 'sv': 'cherokesiska', 'sw': 'Kicherokee', 'ta': 'செரோகீ', 'te': 'చెరోకీ', 'tg': 'черокӣ', 'th': 'เชอโรกี', 'ti': 'ቸሮኪ', 'tk': 'çeroki', 'to': 'lea fakaselokī', 'tr': 'Çerokice', 'tt': 'чероки', 'ug': 'چېروكىچە', 'uk': 'черокі', 'ur': 'چیروکی', 'uz': 'cheroki', 'uz-Cyrl': 'чероки', 'uz-Latn': 'cheroki', 'vec': 'cerokee', 'vi': 'Tiếng Cherokee', 'wo': 'Ceroki', 'xnr': 'चेरोकी', 'yo': 'Èdè Ṣẹ́rókiì', 'yrl': 'xerokí', 'yue': '柴羅基文', 'yue-Hans': '柴罗基文', 'yue-Hant': '柴羅基文', 'zh': '切罗基语', 'zh-Hans': '切罗基语', 'zh-Hant': '柴羅基文', 'zu': 'isi-Cherokee'},
    'cht': {'en': 'Cholón'},
    'chw': {'en': 'Chuwabu'},
    'chx': {'en': 'Chantyal'},
    'chy': {'af': 'Cheyennees', 'am': 'ችዬኔ', 'ar': 'الشايان', 'as': 'চাইয়েন', 'ast': 'cheyenne', 'az': 'çeyen', 'az-Cyrl': 'чејен', 'az-Latn': 'çeyen', 'be': 'шэйен', 'be-tarask': 'шэйен', 'bg': 'шайенски', 'bgn': 'شاینی زبان', 'bn': 'চেইয়েন', 'br': 'cheyenne', 'brx': 'चेयेण', 'bs': 'čejenski', 'bs-Cyrl': 'чејенски', 'bs-Latn': 'čejenski', 'ca': 'xeiene', 'ca-ES-valencia': 'xeiene', 'ccp': '𑄥𑄳𑄆𑄠𑄬𑄚𑄴', 'ce': 'шайенийн', 'chr': 'ᏣᏰᏂ', 'ckb': 'شایان', 'cs': 'čejenština', 'cy': 'Cheyenne', 'da': 'cheyenne', 'de': 'Cheyenne', 'dsb': 'cheyennešćina', 'el': 'Τσεγιέν', 'el-polyton': 'Τσεγιέν', 'en': 'Cheyenne', 'en-Dsrt': '𐐟𐐴𐐰𐑌', 'es': 'cheyene', 'et': 'šaieeni', 'eu': 'txeieneera', 'fa': 'شایانی', 'ff-Adlm': '𞤅𞤢𞥄𞤴𞤢𞤲𞤪𞤫', 'fi': 'cheyenne', 'fil': 'Cheyenne', 'fo': 'cheyenne', 'fr': 'cheyenne', 'fy': 'Cheyenne', 'ga': 'Siáinis', 'gd': 'Cheyenne', 'gl': 'cheyenne', 'gsw': 'Cheyenne', 'gu': 'શેયેન્ન', 'ha': 'Cheyenne', 'he': 'שאיין', 'hi': 'शेयेन्न', 'hi-Latn': 'Cheyenne', 'hr': 'čejenski', 'hsb': 'cheyennešćina', 'ht': 'cheyenne', 'hu': 'csejen', 'hy': 'շայեն', 'ia': 'cheyenne', 'id': 'Cheyenne', 'ig': 'Cheyene', 'is': 'sjeyen', 'it': 'cheyenne', 'ja': 'シャイアン語', 'jv': 'Cheyenne', 'ka': 'ჩეიენი', 'kgp': 'sejẽnẽ', 'kk': 'шайен тілі', 'kk-Cyrl': 'шайен тілі', 'km': 'ឈីយីនី', 'kn': 'ಚೀಯೆನ್ನೇ', 'ko': '샤이엔어', 'kok': 'शायान', 'kok-Deva': 'शायान', 'ks': 'شییون', 'ks-Arab': 'شییون', 'ksh': 'Tschäjänn', 'ku': 'çeyenî', 'ky': 'шайеннче', 'lb': 'Cheyenne', 'lkt': 'Šahíyela Iyápi', 'lo': 'ຊີເຢນນີ', 'lt': 'čajenų', 'lv': 'šejenu', 'mai': 'शयेन', 'mi': 'Haiene', 'mk': 'чејенски', 'ml': 'ഷായാൻ', 'mn': 'чэенн', 'mr': 'शेयेन्न', 'ms': 'Cheyenne', 'mt': 'Cheyenne', 'my': 'ချေယန်း', 'nb': 'cheyenne', 'ne': 'चेयेन्ने', 'nl': 'Cheyenne', 'nn': 'cheyenne', 'no': 'cheyenne', 'or': 'ଚେଚେନା', 'pa': 'ਛਾਇਆਨ', 'pa-Guru': 'ਛਾਇਆਨ', 'pcm': 'Shẹínn Lángwej', 'pl': 'czejeński', 'ps': 'شيني', 'pt': 'cheiene', 'qu': 'Cheyenne', 'rm': 'cheyenne', 'ro': 'cheyenne', 'ru': 'шайенский', 'sc': 'cheyenne', 'sd': 'چايان', 'sd-Arab': 'چايان', 'si': 'චෙයෙන්නෙ', 'sk': 'čejenčina', 'sl': 'čejenščina', 'smn': 'cheyennekielâ', 'so': 'Cheyenne', 'sq': 'çejenisht', 'sr': 'чејенски', 'sr-Cyrl': 'чејенски', 'sr-Latn': 'čejenski', 'sv': 'cheyenne', 'sw': 'Kicheyeni', 'ta': 'செயேனி', 'te': 'చేయేన్', 'th': 'เชเยนเน', 'ti': 'ሻያን', 'tk': 'şaýenn dili', 'to': 'lea fakaseiene', 'tr': 'Şayence', 'ug': 'چېيېنچە', 'uk': 'чейєнн', 'ur': 'چینّے', 'uz': 'cheyenn', 'uz-Cyrl': 'шайенн', 'uz-Latn': 'cheyenn', 'vec': 'ceyen', 'vi': 'Tiếng Cheyenne', 'xnr': 'शेयेन्न', 'yo': 'Èdè Sheyeni', 'yrl': 'xeyeni', 'yue': '沙伊安文', 'yue-Hans': '沙伊安文', 'yue-Hant': '沙伊安文', 'zh': '夏延语', 'zh-Hans': '夏延语', 'zh-Hant': '沙伊安文', 'zu': 'isi-Cheyenne'},
    'chz': {'en': 'Ozumacín Chinantec'},
    'cia': {'en': 'Cia-Cia'},
    'cib': {'en': 'Ci Gbe'},
    'cic': {'bal-Latn': 'Chekkásaw', 'cic': 'Chikashshanompaʼ', 'en': 'Chickasaw', 'et': 'tšikasoo', 'fi': 'chickasaw', 'gd': 'Chickasaw', 'hi': 'चिकसॉ', 'hi-Latn': 'Chickasaw', 'ia': 'chickasaw', 'ig': 'Chickasaw', 'ku': 'çîkasawî', 'sr': 'чикасо', 'sr-Cyrl': 'чикасо', 'sr-Latn': 'čikaso', 'ti': 'ቺካሳው', 'vec': 'cìcasow', 'zh-Hant': '契卡索文'},
    'cid': {'en': 'Chimariko'},
    'cie': {'en': 'Cineni'},
    'cih': {'en': 'Chinali'},
    'cik': {'en': 'Chitkuli Kinnauri'},
    'cim': {'en': 'Cimbrian'},
    'cin': {'en': 'Cinta Larga'},
    'cip': {'en': 'Chiapanec'},
    'cir': {'en': 'Tiri'},
    'ciw': {'en': 'Chippewa'},
    'ciy': {'en': 'Chaima'},
    'cja': {'en': 'Western Cham'},
    'cje': {'en': 'Chru'},
    'cjh': {'en': 'Upper Chehalis'},
    'cji': {'en': 'Chamalal'},
    'cjk': {'en': 'Chokwe'},
    'cjm': {'en': 'Eastern Cham'},
    'cjn': {'en': 'Chenapian'},
    'cjo': {'en': 'Ashéninka Pajonal'},
    'cjp': {'en': 'Cabécar'},
    'cjr': {'en': 'Chorotega'},
    'cjs': {'en': 'Shor'},
    'cjv': {'en': 'Chuave'},
    'cjy': {'en': 'Jinyu Chinese'},
    'cka': {'en': 'Khumi Awa Chin'},
    'ckb': {'af': 'Sorani', 'am': 'የሶራኒ ኩርድኛ', 'ar': 'السورانية الكردية', 'as': 'চেণ্ট্ৰেল কুৰ্ডিচ', 'ast': 'kurdu central', 'az': 'Mərkəzi kürdcə', 'az-Arab': 'اوْرتا کۆردجه', 'az-Cyrl': 'соран', 'az-Latn': 'Mərkəzi kürdcə', 'bal-Latn': 'Myáni Kordi', 'be': 'цэнтральнакурдская', 'be-tarask': 'цэнтральнакурдская', 'bg': 'кюрдски (централен)', 'bgn': 'مرکزین کوردی', 'bn': 'মধ্য কুর্দিশ', 'br': 'kurdeg sorani', 'brx': 'मिरु कुर्दिस', 'bs': 'centralnokurdski', 'bs-Cyrl': 'централнокурдски', 'bs-Latn': 'centralnokurdski', 'ca': 'kurd central', 'ca-ES-valencia': 'kurd central', 'ccp': '𑄟𑄧𑄖𑄴𑄙𑄳𑄠𑄧 𑄇𑄪𑄢𑄴𑄘𑄨𑄌𑄴', 'ce': 'юккъерчу курдийн', 'chr': 'ᎠᏰᏟ ᎫᏗᏏ', 'ckb': 'کوردیی ناوەندی', 'cs': 'kurdština (sorání)', 'cy': 'Cwrdeg Sorani', 'da': 'sorani', 'de': 'Zentralkurdisch', 'dsb': 'sorani', 'el': 'Κεντρικά Κουρδικά', 'el-polyton': 'Κεντρικά Κουρδικά', 'en': 'Central Kurdish', 'es': 'kurdo sorani', 'et': 'sorani', 'eu': 'erdialdeko kurduera', 'fa': 'کردی مرکزی', 'ff-Adlm': '𞤑𞤵𞤪𞤣𞤵𞥅𞤪𞤫', 'fi': 'soranî', 'fil': 'Central Kurdish', 'fo': 'miðkurdiskt', 'fr': 'sorani', 'fy': 'Soranî', 'ga': 'Coirdis Lárnach', 'gd': 'Cùrdais Mheadhanach', 'gl': 'kurdo central', 'gu': 'સેન્ટ્રલ કુર્દિશ', 'ha': 'Kurdawa ta Tsakiya', 'he': 'כורדית סוראנית', 'hi': 'सोरानी कुर्दिश', 'hi-Latn': 'Kurdish, Sorani', 'hr': 'soranski kurdski', 'hsb': 'sorani', 'ht': 'sorani', 'hu': 'közép-ázsiai kurd', 'hy': 'սորանի քրդերեն', 'ia': 'kurdo central', 'id': 'Kurdi Sorani', 'ig': 'Central Kurdish', 'is': 'miðkúrdíska', 'it': 'curdo centrale', 'ja': '中央クルド語', 'jv': 'Kurdi Tengah', 'ka': 'ცენტრალური ქურთული', 'kea': 'kurdu sentral', 'kgp': 'kurno kuju', 'kk': 'сорани тілі', 'kk-Cyrl': 'сорани тілі', 'km': 'ឃើដភាគកណ្តាល', 'kn': 'ಮಧ್ಯ ಕುರ್ದಿಶ್', 'ko': '소라니 쿠르드어', 'kok': 'मध्य कर्दिश', 'kok-Deva': 'मध्य कर्दिश', 'ksh': 'Zäntrahl-Kurrdesch', 'ku': 'kurdî (soranî)', 'ky': 'борбордук курдча', 'lb': 'Sorani', 'lo': 'ໂຊຣານິ ເຄີດິຊ', 'lrc': 'کوردی سوٙرانی', 'lt': 'soranių kurdų', 'lv': 'centrālkurdu', 'mai': 'सेंट्रल कुर्दी', 'mi': 'Kūrihi Waenga', 'mk': 'централнокурдски', 'ml': 'സെൻട്രൽ കുർദിഷ്', 'mn': 'төв курд', 'mr': 'मध्य कुर्दिश', 'ms': 'Kurdi Tengah', 'mt': 'Kurd Ċentrali', 'my': 'ဗဟိုဒေသသုံး ကဒ်ဘာသာ', 'mzn': 'میونی کوردی', 'nb': 'sentralkurdisk', 'ne': 'मध्यवर्ती कुर्दिस', 'nl': 'Soranî', 'nn': 'sorani', 'no': 'sentralkurdisk', 'nqo': 'ߛߏߙߊߣߌߞߊ߲', 'or': 'କେନ୍ଦ୍ରୀୟ କୁରଡିସ୍', 'pa': 'ਕੇਂਦਰੀ ਕੁਰਦਿਸ਼', 'pa-Guru': 'ਕੇਂਦਰੀ ਕੁਰਦਿਸ਼', 'pcm': 'Mídúl Kọ́dish Lángwej', 'pl': 'sorani', 'ps': 'منځنۍ کوردي', 'pt': 'curdo central', 'qu': 'Chawpi Kurdo Simi', 'rm': 'curd central', 'ro': 'kurdă centrală', 'ru': 'центральнокурдский', 'sah': 'Киин куурдуу', 'sc': 'curdu tzentrale', 'sd': 'مرڪزي ڪردش', 'sd-Arab': 'مرڪزي ڪردش', 'si': 'සොරානි කුර්දිෂ්', 'sk': 'kurdčina (sorání)', 'sl': 'osrednja kurdščina', 'smn': 'sorani kurdikielâ', 'so': 'Bartamaha Kurdish', 'sq': 'kurdishte qendrore', 'sr': 'централни курдски', 'sr-Cyrl': 'централни курдски', 'sr-Latn': 'centralni kurdski', 'sv': 'centralkurdiska', 'sw': 'Kikurdi cha Sorani', 'syr': 'ܩܪܕܝܬ ܩܢܛܪܘܢܝܐ', 'ta': 'மத்திய குர்திஷ்', 'te': 'సెంట్రల్ కర్డిష్', 'tg': 'курдии марказӣ', 'th': 'เคิร์ดตอนกลาง', 'ti': 'ማእከላይ ኩርዲሽ', 'tk': 'merkezi kürt dili', 'to': 'lea fakakūtisi-loloto', 'tr': 'Orta Kürtçe', 'tt': 'үзәк көрд', 'ug': 'مەركىزىي كۇردچە', 'uk': 'центральнокурдська', 'ur': 'سینٹرل کردش', 'uz': 'sorani-kurd', 'uz-Cyrl': 'сорани-курдча', 'uz-Latn': 'sorani-kurd', 'vec': 'curdo sentrale', 'vi': 'Tiếng Kurd Miền Trung', 'wo': 'Kurdi gu Diggu', 'xnr': 'सोरानी कुर्दिश', 'yo': 'Ààrin Gbùngbùn Kurdish', 'yrl': 'kurdu piterapura', 'yue': '索拉尼庫爾德文', 'yue-Hans': '索拉尼库尔德文', 'yue-Hant': '索拉尼庫爾德文', 'zh': '中库尔德语', 'zh-Hans': '中库尔德语', 'zh-Hant': '中庫德文', 'zu': 'isi-Central Kurdish'},
    'ckh': {'en': 'Chak'},
    'ckl': {'en': 'Cibak'},
    'ckm': {'en': 'Chakavian'},
    'ckn': {'en': 'Kaang Chin'},
    'cko': {'en': 'Anufo'},
    'ckq': {'en': 'Kajakse'},
    'ckr': {'en': 'Kairak'},
    'cks': {'en': 'Tayo'},
    'ckt': {'en': 'Chukot'},
    'cku': {'en': 'Koasati'},
    'ckv': {'en': 'Kavalan'},
    'ckx': {'en': 'Caka'},
    'cky': {'en': 'Cakfem-Mushere'},
    'ckz': {'en': 'Cakchiquel-Quiché Mixed Language'},
    'cla': {'en': 'Ron'},
    'clc': {'af': 'Tzilkotin', 'am': 'ቺልኮቲን', 'ar': 'تسيلكوتين', 'as': 'চিলক’টিন', 'az': 'Çilotin', 'az-Latn': 'Çilotin', 'be': 'чылькатын', 'be-tarask': 'чылькатын', 'bg': 'чилкотин', 'bn': 'চিলকোটিন', 'br': 'chilkotineg', 'brx': 'चिल्कटिन', 'bs': 'chilcotin', 'bs-Cyrl': 'Чилкотин', 'bs-Latn': 'chilcotin', 'ca': 'chilcotin', 'ca-ES-valencia': 'chilcotin', 'chr': 'ᏥᎸᎪᏘᎾ', 'cs': 'čilkotinština', 'cy': 'Chilcotin', 'da': 'chilcotin', 'de': 'Chilcotin', 'dsb': 'chilcotinšćina', 'el': 'Τσιλκότιν', 'el-polyton': 'Τσιλκότιν', 'en': 'Chilcotin', 'es': 'chilcotin', 'et': 'tšilkotini', 'eu': 'chilcotinera', 'fa': 'چیلکوتن', 'ff-Adlm': '𞤕𞤭𞤤𞤳𞤮𞤼𞤭𞤲𞤪𞤫', 'fi': 'chilcotin', 'fil': 'Chilcotin', 'fr': 'chilcotin', 'ga': 'Chilcotin', 'gd': 'Chilcotin', 'gl': 'chilcotin', 'gu': 'ચિલકોટિન', 'ha': 'Chilcotin', 'he': 'צ׳ילקוטין', 'hi': 'चिलकोटिन', 'hi-Latn': 'Chilcotin', 'hr': 'chilcotin', 'hsb': 'chilcotinšćina', 'ht': 'chilcotin', 'hu': 'csilkotin', 'hy': 'չիլկոտիներեն', 'ia': 'chilcotin', 'id': 'Chilcotin', 'ig': 'Chilcotin', 'is': 'chilcotin', 'it': 'chilcotin', 'ja': 'チルコーティン語', 'jv': 'Chilcotin', 'ka': 'ჩილკოტინი', 'kk': 'чилкотин тілі', 'kk-Cyrl': 'чилкотин тілі', 'km': 'ឈីលកូទីន', 'kn': 'ಚಿಲ್ಕೋಟಿನ್', 'ko': '칠코틴어', 'kok': 'चिलकोटिन', 'kok-Deva': 'चिलकोटिन', 'ku': 'çilkotînî', 'ky': 'чилкотинче (британдык колумбиядагы аймак)', 'lo': 'ຊິວໂຄຕິນ', 'lt': 'čilkotinų', 'lv': 'čilkotīnu', 'mai': 'चोकोटिन', 'mi': 'Tiekautini', 'mk': 'чилкотински', 'ml': 'ചിൽകോട്ടിൻ', 'mn': 'чилкотин', 'mr': 'चिलकोटिन', 'ms': 'Chilcotin', 'my': 'ချီကိုလ်တင်', 'nb': 'chilcotin', 'ne': 'चिलकोटिन', 'nl': 'Chilcotin', 'nn': 'chilcotin', 'no': 'chilcotin', 'or': 'ଚିଲକୋଟିନ୍', 'pa': 'ਚਿਲਕੋਟਿਨ', 'pa-Guru': 'ਚਿਲਕੋਟਿਨ', 'pcm': 'Chílkotín Lángwej', 'pl': 'chilcotin', 'ps': 'چیلکوټین', 'pt': 'chilcotin', 'qu': 'Chilcotin Simi', 'ro': 'chilcotin', 'ru': 'чилкотин', 'sc': 'chilcotin', 'sd': 'چلڪوٽن', 'sd-Arab': 'چلڪوٽن', 'si': 'චිල්කොටින්', 'sk': 'chilcotin', 'sl': 'čilkotinščina', 'so': 'Chilcotin', 'sq': 'çilkotinisht', 'sr': 'чилкотин', 'sr-Cyrl': 'чилкотин', 'sr-Latn': 'čilkotin', 'sv': 'chilcotin', 'sw': 'Kichikotini', 'ta': 'சில்கோடின்', 'te': 'చిల్కటిన్', 'th': 'ชิลโคติน', 'ti': 'ቺልኮቲን', 'tk': 'çilkotin dili', 'to': 'lea fakatisilikōtini', 'tr': 'Çilkotince', 'uk': 'чилкотін', 'ur': 'چلکوٹن', 'uz': 'chilkotin', 'uz-Latn': 'chilkotin', 'vec': 'cìlcotin', 'vi': 'Tiếng Chilcotin', 'yo': 'Èdè Shikoti', 'yue': '奇爾科延文', 'yue-Hans': '奇尔科延文', 'yue-Hant': '奇爾科延文', 'zh': '奇尔科廷语', 'zh-Hans': '奇尔科廷语', 'zh-Hant': '齊爾柯廷語', 'zu': 'isi-Chilcotin'},
    'cld': {'en': 'Chaldean Neo-Aramaic'},
    'cle': {'en': 'Lealao Chinantec'},
    'clh': {'en': 'Chilisso'},
    'cli': {'en': 'Chakali'},
    'clj': {'en': 'Laitu Chin'},
    'clk': {'en': 'Idu-Mishmi'},
    'cll': {'en': 'Chala'},
    'clm': {'en': 'Clallam'},
    'clo': {'en': 'Lowland Oaxaca Chontal'},
    'clt': {'en': 'Lautu Chin'},
    'clu': {'en': 'Caluyanun'},
    'clw': {'en': 'Chulym'},
    'cly': {'en': 'Eastern Highland Chatino'},
    'cma': {'en': 'Maa'},
    'cmc': {'en': 'Chamic languages'},
    'cmc-pro': {'en': 'Proto-Chamic'},
    'cme': {'en': 'Cerma'},
    'cmg': {'en': 'Classical Mongolian'},
    'cmi': {'en': 'Emberá-Chamí'},
    'cmk': {'en': 'Chimakum'},
    'cml': {'en': 'Campalagian'},
    'cmm': {'en': 'Michigamea'},
    'cmn': {'en': 'Mandarin Chinese'},
    'cmo': {'en': 'Central Mnong'},
    'cmr': {'en': 'Mro-Khimi Chin'},
    'cms': {'en': 'Messapic'},
    'cmt': {'en': 'Camtho'},
    'cna': {'en': 'Changthang'},
    'cnb': {'en': 'Chinbon Chin'},
    'cnc': {'en': 'Côông'},
    'cng': {'en': 'Northern Qiang'},
    'cnh': {'en': 'Hakha Chin'},
    'cni': {'en': 'Asháninka'},
    'cnk': {'en': 'Khumi Chin'},
    'cnl': {'en': 'Lalana Chinantec'},
    'cno': {'en': 'Con'},
    'cnp': {'en': 'Northern Ping Chinese'},
    'cnr': {'en': 'Montenegrin'},
    'cns': {'en': 'Central Asmat'},
    'cnt': {'en': 'Tepetotutla Chinantec'},
    'cnu': {'en': 'Chenoua'},
    'cnw': {'en': 'Ngawn Chin'},
    'cnx': {'en': 'Middle Cornish'},
    'co': {'af': 'Korsikaans', 'am': 'ኮርሲካኛ', 'ar': 'الكورسيكية', 'as': 'কোৰ্ছিকান', 'ast': 'corsu', 'az': 'korsika', 'az-Arab': 'کوْرسیکاجا', 'az-Cyrl': 'корсика', 'az-Latn': 'korsika', 'bal-Latn': 'Korsiki', 'be': 'карсіканская', 'be-tarask': 'карсіканская', 'bg': 'корсикански', 'bgn': 'کرسی', 'bn': 'কর্সিকান', 'br': 'korseg', 'brx': 'कर्सिकान', 'bs': 'korzikanski', 'bs-Cyrl': 'корзикански', 'bs-Latn': 'korzikanski', 'ca': 'cors', 'ca-ES-valencia': 'cors', 'ccp': '𑄇𑄧𑄢𑄴𑄥𑄨𑄇𑄚𑄴', 'ce': 'корсиканийн', 'chr': 'ᎪᎵᏍᎢᎧᏂ', 'ckb': 'کۆرسیکی', 'co': 'corsu', 'cs': 'korsičtina', 'cy': 'Corseg', 'da': 'korsikansk', 'de': 'Korsisch', 'dsb': 'korsišćina', 'el': 'Κορσικανικά', 'el-polyton': 'Κορσικανικά', 'en': 'Corsican', 'en-Dsrt': '𐐗𐐬𐑉𐑅𐐮𐐿𐐲𐑌', 'eo': 'korsika', 'es': 'corso', 'et': 'korsika', 'eu': 'korsikera', 'fa': 'کورسی', 'ff-Adlm': '𞤑𞤮𞤪𞤧𞤭𞤳𞤢𞥄𞤪𞤫', 'fi': 'korsika', 'fil': 'Corsican', 'fo': 'korsikanskt', 'fr': 'corse', 'fur': 'cors', 'fy': 'Korsikaansk', 'ga': 'Corsaicis', 'gd': 'Corsais', 'gl': 'corso', 'gsw': 'Korsisch', 'gu': 'કોર્સિકન', 'ha': 'Corsican', 'he': 'קורסיקנית', 'hi': 'कोर्सीकन', 'hi-Latn': 'Corsican', 'hr': 'korzički', 'hsb': 'korsišćina', 'ht': 'corse', 'hu': 'korzikai', 'hy': 'կորսիկերեն', 'ia': 'corso', 'id': 'Korsika', 'ig': 'Corsican', 'is': 'korsíska', 'it': 'corso', 'ja': 'コルシカ語', 'jv': 'Korsika', 'ka': 'კორსიკული', 'kea': 'kórsiku', 'kgp': 'korso', 'kk': 'корсика тілі', 'kk-Cyrl': 'корсика тілі', 'km': 'កូស៊ីខាន', 'kn': 'ಕೋರ್ಸಿಕನ್', 'ko': '코르시카어', 'kok': 'कोर्सिकन', 'kok-Deva': 'कोर्सिकन', 'ks': 'کارسِکَن', 'ks-Arab': 'کارسِکَن', 'ksh': 'Korsesch', 'ku': 'korsîkayî', 'ky': 'корсиканча', 'lb': 'Korsesch', 'lld': 'corsich', 'lo': 'ຄໍຊິກາ', 'lrc': 'کوریسکان', 'lt': 'korsikiečių', 'lv': 'korsikāņu', 'mai': 'कोर्शियाई', 'mi': 'Kōhikana', 'mk': 'корзикански', 'ml': 'കോർസിക്കൻ', 'mn': 'корсик', 'mr': 'कॉर्सिकन', 'ms': 'Corsica', 'mt': 'Korsiku', 'my': 'ခိုစီကန်', 'mzn': 'کورسیکان', 'nb': 'korsikansk', 'ne': 'कोर्सिकन', 'nl': 'Corsicaans', 'nn': 'korsikansk', 'no': 'korsikansk', 'nqo': 'ߞߐߙߑߛߌߞߊ߲', 'or': 'କୋର୍ସିକାନ୍', 'pa': 'ਕੋਰਸੀਕਨ', 'pa-Guru': 'ਕੋਰਸੀਕਨ', 'pcm': 'Kọsíkan Lángwej', 'pl': 'korsykański', 'ps': 'کورسيکاني', 'pt': 'corso', 'qu': 'Corso Simi', 'rm': 'cors', 'ro': 'corsicană', 'ru': 'корсиканский', 'sc': 'corsicanu', 'sd': 'ڪارسيڪائي', 'sd-Arab': 'ڪارسيڪائي', 'se': 'corsicagiella', 'si': 'කෝසිකානු', 'sk': 'korzičtina', 'sl': 'korziščina', 'smn': 'korsikakielâ', 'so': 'Korsikan', 'sq': 'korsikisht', 'sr': 'корзикански', 'sr-Cyrl': 'корзикански', 'sr-Latn': 'korzikanski', 'sv': 'korsikanska', 'sw': 'Kikosikani', 'ta': 'கார்சிகன்', 'te': 'కోర్సికన్', 'tg': 'корсиканӣ', 'th': 'คอร์ซิกา', 'ti': 'ኮርስኛ', 'tk': 'korsikan dili', 'to': 'lea fakakōsika', 'tr': 'Korsikaca', 'tt': 'корсика', 'ug': 'كورسۇچە', 'uk': 'корсиканська', 'ur': 'کوراسیکن', 'uz': 'korsikan', 'uz-Cyrl': 'корсиканча', 'uz-Latn': 'korsikan', 'vec': 'corso', 'vi': 'Tiếng Corsica', 'wo': 'Kors', 'xnr': 'कोर्सीकन', 'yo': 'Èdè Corsican', 'yrl': 'curisu', 'yue': '科西嘉文', 'yue-Hans': '科西嘉文', 'yue-Hant': '科西嘉文', 'zh': '科西嘉语', 'zh-Hans': '科西嘉语', 'zh-Hant': '科西嘉文', 'zu': 'isi-Corsican'},
    'coa': {'en': 'Cocos Islands Malay'},
    'cob': {'en': 'Chicomuceltec'},
    'coc': {'en': 'Cocopa'},
    'cod': {'en': 'Cocama-Cocamilla'},
    'coe': {'en': 'Koreguaje'},
    'cof': {'en': 'Colorado'},
    'cog': {'en': 'Chong'},
    'coh': {'en': 'Chonyi-Dzihana-Kauma'},
    'coj': {'en': 'Cochimi'},
    'cok': {'en': 'Santa Teresa Cora'},
    'col': {'en': 'Columbia-Wenatchi'},
    'com': {'en': 'Comanche'},
    'con': {'en': 'Cofán'},
    'coo': {'en': 'Comox'},
    'cop': {'af': 'Kopties', 'am': 'ኮፕቲክ', 'ar': 'القبطية', 'ast': 'cópticu', 'az': 'kopt', 'az-Latn': 'kopt', 'be': 'копцкая', 'be-tarask': 'копцкая', 'bg': 'коптски', 'bn': 'কপটিক', 'br': 'kopteg', 'brx': 'कॉप्टीक्', 'bs': 'koptski', 'bs-Cyrl': 'коптски', 'bs-Latn': 'koptski', 'ca': 'copte', 'ca-ES-valencia': 'copte', 'ccp': '𑄇𑄧𑄛𑄴𑄑𑄨𑄇𑄴', 'cs': 'koptština', 'cy': 'Copteg', 'da': 'koptisk', 'de': 'Koptisch', 'el': 'Κοπτικά', 'el-polyton': 'Κοπτικά', 'en': 'Coptic', 'en-Dsrt': '𐐗𐐬𐐹𐐻𐐮𐐿', 'es': 'copto', 'et': 'kopti', 'fa': 'قبطی', 'fi': 'kopti', 'fr': 'copte', 'fur': 'coptic', 'fy': 'Koptysk', 'ga': 'Coptais', 'gd': 'Coptais', 'gsw': 'Koptisch', 'gu': 'કોપ્ટિક', 'he': 'קופטית', 'hi': 'कॉप्टिक', 'hi-Latn': 'Coptic', 'hr': 'koptski', 'ht': 'copte', 'hu': 'kopt', 'hy': 'ղպտերեն', 'id': 'Koptik', 'is': 'koptíska', 'it': 'copto', 'ja': 'コプト語', 'ka': 'კოპტური', 'kgp': 'komta', 'kn': 'ಕೊಪ್ಟಿಕ್', 'ko': '콥트어', 'ks': 'کاپٹِک', 'ks-Arab': 'کاپٹِک', 'ksh': 'Koptesch', 'lb': 'Koptesch', 'lkt': 'Coptic Iyápi', 'lo': 'ຄອບຕິກ', 'lt': 'koptų', 'lv': 'koptu', 'mk': 'коптски', 'ml': 'കോപ്റ്റിക്', 'mr': 'कॉप्टिक', 'ms': 'Coptic', 'mt': 'Koptiku', 'nb': 'koptisk', 'ne': 'कोप्टिक', 'nl': 'Koptisch', 'nn': 'koptisk', 'no': 'koptisk', 'or': 'କପ୍ଟିକ୍', 'os': 'коптаг', 'pl': 'koptyjski', 'pt': 'copta', 'rm': 'coptic', 'ro': 'coptă', 'ru': 'коптский', 'sk': 'koptčina', 'sl': 'koptščina', 'sr': 'коптски', 'sr-Cyrl': 'коптски', 'sr-Latn': 'koptski', 'sv': 'koptiska', 'sw': 'Kikhufti', 'ta': 'காப்டிக்', 'te': 'కోప్టిక్', 'th': 'คอปติก', 'to': 'lea fakakopitika', 'tr': 'Kıptice', 'ug': 'كوپتىكچە', 'uk': 'коптська', 'vi': 'Tiếng Coptic', 'yrl': 'kupita', 'yue': '科普特文', 'yue-Hans': '科普特文', 'yue-Hant': '科普特文', 'zh': '科普特语', 'zh-Hans': '科普特语', 'zh-Hant': '科普特文'},
    'coq': {'en': 'Coquille'},
    'cot': {'en': 'Caquinte'},
    'cou': {'en': 'Wamey'},
    'cov': {'en': 'Cao Miao'},
    'cow': {'en': 'Cowlitz'},
    'cox': {'en': 'Nanti'},
    'coy': {'en': 'Coyaima'},
    'coz': {'en': 'Chochotec'},
    'cpa': {'en': 'Palantla Chinantec'},
    'cpb': {'en': 'Ucayali-Yurúa Ashéninka'},
    'cpc': {'en': 'Ajyíninka Apurucayali'},
    'cpe': {'en': 'English-based creoles and pidgins'},
    'cpe-mar': {'en': 'Maroon Spirit Language'},
    'cpe-spp': {'en': 'Samoan Plantation Pidgin'},
    'cpf': {'en': 'French-based creoles and pidgins'},
    'cpg': {'en': 'Cappadocian Greek'},
    'cpi': {'en': 'Chinese Pidgin English'},
    'cpn': {'en': 'Cherepon'},
    'cpo': {'en': 'Kpeego'},
    'cpp': {'en': 'Portuguese-based creoles and pidgins'},
    'cps': {'am': 'ካፒዝኖን', 'ast': 'capiznon', 'cs': 'kapiznonština', 'de': 'Capiznon', 'en': 'Capiznon', 'et': 'kapisnoni', 'fi': 'capiznon', 'fr': 'capiznon', 'gd': 'Capiznon', 'hi-Latn': 'Capiznon', 'ht': 'capiznon', 'it': 'capiznon', 'ja': 'カピス語', 'lb': 'Capiznon', 'lt': 'capiznon', 'mk': 'капизнон', 'nb': 'kapiz', 'ne': 'कापिज्नोन', 'nl': 'Capiznon', 'nn': 'kapiz', 'no': 'kapiz', 'pl': 'capiznon', 'sv': 'kapisnon', 'th': 'กาปิซนอน', 'to': 'lea fakakapiseno', 'tr': 'Capiznon', 'vi': 'Tiếng Capiznon', 'yue': '卡皮茲文', 'yue-Hans': '卡皮兹文', 'yue-Hant': '卡皮茲文', 'zh-Hant': '卡皮茲文'},
    'cpu': {'en': 'Pichis Ashéninka'},
    'cpx': {'en': 'Pu-Xian Chinese'},
    'cpy': {'en': 'South Ucayali Ashéninka'},
    'cqd': {'en': 'Chuanqiandian Cluster Miao'},
    'cqu': {'en': 'Chilean Quechua'},
    'cr': {'am': 'ክሪ', 'ar': 'الكرى', 'ast': 'cree', 'az': 'kri', 'az-Latn': 'kri', 'bg': 'крии', 'bn': 'ক্রি', 'br': 'kri', 'brx': 'क्री', 'bs': 'kri', 'bs-Cyrl': 'кри', 'bs-Latn': 'kri', 'ca': 'cree', 'ca-ES-valencia': 'cree', 'ccp': '𑄇𑄳𑄢𑄨', 'cs': 'kríjština', 'cy': 'Cri', 'da': 'cree', 'de': 'Cree', 'dsb': 'kri', 'el': 'Κρι', 'el-polyton': 'Κρι', 'en': 'Cree', 'en-Dsrt': '𐐗𐑉𐐨', 'es': 'cree', 'et': 'krii', 'fa': 'کریایی', 'fi': 'cree', 'fr': 'cree', 'fur': 'cree', 'fy': 'Cree', 'ga': 'Craís', 'gd': 'Cree', 'gsw': 'Cree', 'gu': 'ક્રી', 'he': 'קרי', 'hi': 'क्री', 'hi-Latn': 'Cree', 'hr': 'cree', 'hsb': 'kri', 'ht': 'cree', 'hu': 'krí', 'id': 'Kree', 'is': 'krí', 'it': 'cree', 'ja': 'クリー語', 'ka': 'კრი', 'kgp': 'kri', 'kn': 'ಕ್ರೀ', 'ko': '크리어', 'ks': 'کری', 'ks-Arab': 'کری', 'lb': 'Cree', 'lkt': 'Maštíŋča Oyáte Iyápi', 'lo': 'ຄີ', 'lt': 'kry', 'lv': 'krī', 'mk': 'кри', 'ml': 'ക്രീ', 'mr': 'क्री', 'mt': 'Cree', 'my': 'ခရီး', 'nb': 'cree', 'ne': 'क्री', 'nl': 'Cree', 'nn': 'cree', 'no': 'cree', 'or': 'କ୍ରୀ', 'pl': 'kri', 'pt': 'cree', 'rm': 'cree', 'ro': 'cree', 'ru': 'кри', 'sk': 'krí', 'sl': 'krijščina', 'sr': 'кри', 'sr-Cyrl': 'кри', 'sr-Latn': 'kri', 'sv': 'cree', 'ta': 'க்ரீ', 'te': 'క్రి', 'th': 'ครี', 'to': 'lea fakakelī', 'tr': 'Krice', 'ug': 'كرىچە', 'uk': 'крі', 'vi': 'Tiếng Cree', 'yrl': 'kiri', 'yue': '克裡文', 'yue-Hans': '克里文', 'yue-Hant': '克裡文', 'zh': '克里语', 'zh-Hans': '克里语', 'zh-Hant': '克里文'},
    'cra': {'en': 'Chara'},
    'crb': {'en': 'Island Carib'},
    'crc': {'en': 'Lonwolwol'},
    'crd': {'en': "Coeur d'Alene"},
    'crf': {'en': 'Caramanta'},
    'crg': {'af': 'Michif', 'am': 'ሚቺፍ', 'ar': 'الميتشيف', 'as': 'মিচিফ', 'az': 'miçif', 'az-Latn': 'miçif', 'be': 'мічыф', 'be-tarask': 'мічыф', 'bg': 'мичиф', 'bn': 'মিচিফ', 'br': 'michifeg', 'brx': 'मिचिफ', 'bs': 'mičif', 'bs-Cyrl': 'мичиф', 'bs-Latn': 'mičif', 'ca': 'michif', 'ca-ES-valencia': 'michif', 'chr': 'ᎻᏥᏩ', 'cs': 'mičif', 'cy': 'Michif', 'da': 'michif', 'de': 'Michif', 'dsb': 'michifšćina', 'el': 'Μίτσιφ', 'el-polyton': 'Μίτσιφ', 'en': 'Michif', 'es': 'michif', 'et': 'michifi', 'eu': 'metisera', 'fa': 'میچیف', 'ff-Adlm': '𞤃𞤭𞤷𞤭𞤬𞤪𞤫', 'fi': 'michif', 'fil': 'Michif', 'fr': 'mitchif', 'ga': 'Michif', 'gd': 'Michif', 'gl': 'michif', 'gu': 'મિચિફ', 'ha': 'Michif', 'he': 'מצ׳יף', 'hi': 'मिचिफ़', 'hi-Latn': 'Michif', 'hr': 'michif', 'hsb': 'michifšćina', 'ht': 'mitchif', 'hu': 'micsif', 'hy': 'միչիֆ', 'ia': 'mitchif', 'id': 'Michif', 'ig': 'Michif', 'is': 'michif', 'it': 'métchif', 'ja': 'ミチフ語', 'jv': 'Michif', 'ka': 'მიჩიფი', 'kk': 'мичиф тілі', 'kk-Cyrl': 'мичиф тілі', 'km': 'មីឈីហ្វ', 'kn': 'ಮಿಚಿಫ್', 'ko': '미치프어', 'kok': 'मिचिफ', 'kok-Deva': 'मिचिफ', 'ku': 'mîçîfî', 'ky': 'мичифче (индей тили)', 'lo': 'ມິຊິຟ', 'lt': 'metisų', 'lv': 'mičifu', 'mai': 'मिचिफ', 'mi': 'Mītiwhi', 'mk': 'мичиф', 'ml': 'മിചിഫ്', 'mn': 'мичиф', 'mr': 'मिचिफो', 'ms': 'Michif', 'my': 'မစ်ချစ်ခ်', 'nb': 'michif', 'ne': 'मिचिफ', 'nl': 'Michif', 'nn': 'michif', 'no': 'michif', 'or': 'ମିସିଫ', 'pa': 'ਮਿਚਿਫੋ', 'pa-Guru': 'ਮਿਚਿਫੋ', 'pcm': 'Michíf Lángwej', 'pl': 'miszif', 'ps': 'mc', 'pt': 'michif', 'qu': 'Michif Simi', 'ro': 'michif', 'ru': 'мичиф', 'sc': 'michif', 'sd': 'ميچيف', 'sd-Arab': 'ميچيف', 'si': 'මිචිෆ්', 'sk': 'michif', 'sl': 'mičifščina', 'so': 'Michif', 'sq': 'miçifisht', 'sr': 'мичиф', 'sr-Cyrl': 'мичиф', 'sr-Latn': 'mičif', 'sv': 'michif', 'sw': 'Kimichifu', 'ta': 'மிச்சிஃப்', 'te': 'మిచిఫ్', 'th': 'มิชิฟ', 'ti': 'ሚቺፍ', 'tk': 'miçif dili', 'to': 'lea fakametisifi', 'tr': 'Michif dili', 'uk': 'мічиф', 'ur': 'میچیف', 'uz': 'michif', 'uz-Latn': 'michif', 'vec': 'mecif', 'vi': 'Tiếng Michif', 'yo': 'Èdè Misifu', 'yue': '米奇夫文', 'yue-Hans': '米奇夫文', 'yue-Hant': '米奇夫文', 'zh': '米其芙语', 'zh-Hans': '米其芙语', 'zh-Hant': '米奇夫語', 'zu': 'isi-Michif'},
    'crh': {'am': 'ክሪሚያን ተርኪሽ', 'ar': 'لغة تتار القرم', 'ast': 'turcu de Crimea', 'az': 'krım türkcəsi', 'az-Latn': 'krım türkcəsi', 'bg': 'кримскотатарски', 'bn': 'ক্রিমিয়ান তুর্কি', 'br': 'turkeg Krimea', 'brx': 'तुर्की क्रिमिया', 'bs': 'krimski turski', 'bs-Cyrl': 'кримеански турски', 'bs-Latn': 'krimski turski', 'ca': 'tàtar de Crimea', 'ca-ES-valencia': 'tàtar de Crimea', 'ccp': '𑄇𑄳𑄢𑄨𑄟𑄨𑄠𑄚𑄴 𑄖𑄪𑄢𑄴𑄇𑄨', 'cs': 'tatarština (krymská)', 'cy': 'Tyrceg y Crimea', 'da': 'krimtatarisk', 'de': 'Krimtatarisch', 'el': 'Τουρκικά Κριμαίας', 'el-polyton': 'Τουρκικά Κριμαίας', 'en': 'Crimean Tatar', 'es': 'tártaro de Crimea', 'et': 'krimmitatari', 'fa': 'ترکی کریمه', 'fi': 'krimintataari', 'fr': 'tatar de Crimée', 'fy': 'Krim-Tataarsk', 'gd': 'Turcais Chriomach', 'gsw': 'Krimtatarisch', 'gu': 'ક્રિમિયન તુર્કી', 'he': 'טטרית של קרים', 'hi': 'क्रीमीन तुर्की', 'hi-Latn': 'Crimean Turkish', 'hr': 'krimski turski', 'ht': 'tatar de Crimée', 'hu': 'krími tatár', 'hy': 'ղրիմյան թուրքերեն', 'id': 'Tatar Krimea', 'is': 'krímtyrkneska', 'it': 'turco crimeo', 'ja': 'クリミア・タタール語', 'ka': 'ყირიმულ-თურქული', 'kgp': 'krimẽja tá turko', 'kn': 'ಕ್ರಿಮೀಯನ್ ಟರ್ಕಿಷ್', 'ko': '크리민 터키어; 크리민 타타르어', 'ks': 'کرٕمیٖن تُرکی', 'ks-Arab': 'کرٕمیٖن تُرکی', 'lb': 'Krimtataresch', 'lkt': 'Crimean Turkish Iyápi', 'lo': 'ຄຣີເມນເຕີຄິຊ', 'lt': 'Krymo turkų', 'lv': 'Krimas tatāru', 'mk': 'кримскотурски', 'ml': 'ക്രിമിയൻ ടർക്കിഷ്', 'mr': 'क्राइमीन तुर्की', 'ms': 'Turki Krimea', 'mt': 'Tork tal-Krimea', 'nb': 'krimtatarisk', 'ne': 'क्रिमियाली तुर्क', 'nl': 'Krim-Tataars', 'nn': 'krimtatarisk', 'no': 'krimtatarisk', 'or': 'କ୍ରୀମିନ୍ ତୁର୍କୀସ୍', 'pl': 'krymskotatarski', 'pt': 'tártara da Crimeia', 'rm': 'tirc crimean', 'ro': 'turcă crimeeană', 'ru': 'крымско-татарский', 'sk': 'krymská tatárčina', 'sl': 'krimska tatarščina', 'sr': 'кримскотатарски', 'sr-Cyrl': 'кримскотатарски', 'sr-Latn': 'krimskotatarski', 'sv': 'krimtatariska', 'ta': 'கிரிமியன் துர்க்கி', 'te': 'క్రిమియన్ టర్కిష్', 'th': 'ตุรกีไครเมีย', 'to': 'lea fakatatali-kilimea', 'tr': 'Kırım Tatarcası', 'ug': 'قىرىم تۈركچە', 'uk': 'кримськотатарська', 'vi': 'Tiếng Thổ Nhĩ Kỳ Crimean', 'yi': 'קרים־טערקיש', 'yrl': 'Kirimeya turiku', 'yue': '克里米亞半島的土耳其文；克里米亞半島的塔塔爾文', 'yue-Hans': '克里米亚半岛的土耳其文；克里米亚半岛的塔塔尔文', 'yue-Hant': '克里米亞半島的土耳其文；克里米亞半島的塔塔爾文', 'zh': '克里米亚鞑靼语', 'zh-Hans': '克里米亚鞑靼语', 'zh-Hant': '克里米亞韃靼文'},
    'cri': {'en': 'Sãotomense'},
    'crj': {'af': 'Suidoos-Cree', 'am': 'ደቡብ ምዕራባዊ ክሪ', 'ar': 'الكري الجنوب شرقية', 'as': 'দাক্ষিণাত্য ইষ্ট ক্ৰী', 'az': 'cənub-şərqi kri', 'az-Latn': 'cənub-şərqi kri', 'be': 'паўднёва-ўсходняя кры', 'be-tarask': 'паўднёва-ўсходняя кры', 'bg': 'югоизточен крий', 'bn': 'দক্ষিণ পূর্ব ক্রী', 'br': 'krieg ar Gevred', 'brx': 'खोलायारि सानजा क्री', 'bs': 'jugoistočni kri', 'bs-Cyrl': 'јужноисточни кре', 'bs-Latn': 'jugoistočni kri', 'ca': 'cree sud-oriental', 'ca-ES-valencia': 'cree sud-oriental', 'chr': 'ᏧᎦᏃᏮ ᏗᎧᎸᎬ Ꮯ', 'cs': 'kríjština (jihovýchodní)', 'cy': 'Cree De Ddwyrain', 'da': 'sydøstcree', 'de': 'Südost-Cree', 'dsb': 'krotkozajtšna creešćina', 'el': 'Νοτιοανατολικά Κρι', 'el-polyton': 'Νοτιοανατολικά Κρι', 'en': 'Southern East Cree', 'es': 'cree suroriental', 'et': 'lõuna-idakrii', 'eu': 'hego-ekialdeko creera', 'fa': 'کری جنوب شرقی', 'ff-Adlm': '𞤑𞤪𞤭𞥅𞤪𞤫 𞤙𞤢𞥄𞤥𞤲𞤢𞥄𞤲𞤺𞤫𞤲𞤳𞤮', 'fi': 'east cree (eteläinen)', 'fil': 'Southern East Cree', 'fr': 'cri de l’Est (dialecte du Sud)', 'ga': 'Craís an Deiscirt Thoir', 'gd': 'Cree Ear-dheasach', 'gl': 'cree do sueste', 'gu': 'દક્ષિણ પૂર્વ ક્રી', 'ha': 'Cree na Kusu-Maso-Gabas', 'he': 'קרי דרום מזרחי', 'hi': 'दक्षिण पूर्वी क्री', 'hi-Latn': 'Southern East Cree', 'hr': 'jugoistični cree', 'hsb': 'južnowuchodna creešćina', 'ht': 'cri de l’Est (dialecte du Sud)', 'hu': 'délkeleti krí', 'hy': 'հարավ-արևելյան քրի', 'ia': 'cree del sud-est', 'id': 'East Cree Selatan', 'ig': 'Southern East Cree', 'is': 'suðaustur-cree', 'it': 'cree sud-orientale', 'ja': '東部クリー語(南部)', 'jv': 'Kree Kidul Wetan', 'ka': 'სამხრეთ-აღმოსავლეთის კრი', 'kk': 'оңтүстік-шығыс кри тілі', 'kk-Cyrl': 'оңтүстік-шығыс кри тілі', 'km': 'គ្រីខាងកើត\u200bប៉ែកខាងត្បូង', 'kn': 'ದಕ್ಷಿಣ ಪೂರ್ವ ಕ್ರೀ', 'ko': '남동부 크리어', 'kok': 'आग्नेय क्री', 'kok-Deva': 'आग्नेय क्री', 'ku': 'krîya rojhilat ya başûrî', 'ky': 'түштүк-чыгыш криче (индей тили)', 'lo': 'ຄຣີຕາເວັນອອກສ່ຽງໃຕ້', 'lt': 'pietryčių kri', 'lv': 'dienvidaustrumu krī', 'mai': 'दक्षिण-पुरबी क्री', 'mi': 'Kirī Tonga-mā-Rāwhiti', 'mk': 'југоисточен кријски', 'ml': 'സതേൺ ഈസ്റ്റ് ക്രീ', 'mn': 'зүүн өмнөд кри', 'mr': 'दक्षिणात्य इस्ट क्री', 'ms': 'Cree Tenggara', 'my': 'တောင်ပိုင်း အရှေ့ခရီ', 'nb': 'sørlig østcree', 'ne': 'दक्षिण पूर्वी क्री', 'nl': 'Zuidoost-Cree', 'nn': 'sørleg aust-cree', 'no': 'sørlig østcree', 'or': 'ସାଉଥ୍ ଇଷ୍ଟ କ୍ରି', 'pa': 'ਦੱਖਣੀ ਪੂਰਬੀ ਕ੍ਰੀ', 'pa-Guru': 'ਦੱਖਣੀ ਪੂਰਬੀ ਕ੍ਰੀ', 'pcm': 'Saútán Íst Krii Lángwej', 'pl': 'kri południowo-wschodni', 'ps': 'سویل ختیځ کری', 'pt': 'cree do sudeste', 'qu': 'Cree Este del Sur Simi', 'ro': 'cree de sud-est', 'ru': 'юго-восточный кри', 'sc': 'cree sud-orientale', 'sd': 'ڏکڻ اڀرندو ڪري', 'sd-Arab': 'ڏکڻ اڀرندو ڪري', 'si': 'දකුණු නැගෙනහිර ක්\u200dරී', 'sk': 'cree (juhovýchod)', 'sl': 'jugovzhodna krijščina', 'so': 'Southern East Cree', 'sq': 'krijishte juglindore', 'sr': 'југоисточни кри', 'sr-Cyrl': 'југоисточни кри', 'sr-Latn': 'jugoistočni kri', 'sv': 'sydostcree', 'sw': 'Kikrii cha Kusini Mashariki', 'ta': 'தென்கிழக்கு க்ரீ', 'te': 'దక్షిణ తూర్పు క్రీ', 'th': 'ครีตะวันออกเฉียงใต้', 'ti': 'ደቡባዊ ምብራቕ ክሪ', 'tk': 'günorta-gündogar kri dili', 'to': 'lea fakakilī-tongahahake', 'tr': 'Güney Doğu Kricesi', 'uk': 'південно-східна крі', 'ur': 'جنوب مشرقی کری', 'uz': 'janubi-sharqiy kri', 'uz-Latn': 'janubi-sharqiy kri', 'vec': 'cree sud orientale', 'vi': 'Tiếng Cree Đông Nam', 'yo': 'Èdè Gusu Ila-oorun Kri', 'yue': '東南克里文', 'yue-Hans': '东南克里文', 'yue-Hant': '東南克里文', 'zh': '东南部克里语', 'zh-Hans': '东南部克里语', 'zh-Hant': '東南克里語', 'zu': 'Southern East Cree'},
    'crk': {'af': 'Laagvlakte-Cree', 'am': 'ፕላይንስ ክሪ', 'ar': 'البلينز-كري', 'as': 'প্লেইনছ ক্ৰী', 'az': 'ova kricəsi', 'az-Latn': 'ova kricəsi', 'be': 'раўнінны кры', 'be-tarask': 'раўнінны кры', 'bg': 'плейнс крий', 'bn': 'প্লেনস ক্রী', 'br': 'krieg ar cʼhompezennoù', 'brx': 'हायेनारि क्री', 'bs': 'ravničarski kri', 'bs-Cyrl': 'равничарски кре', 'bs-Latn': 'ravničarski kri', 'ca': 'cree de la plana', 'ca-ES-valencia': 'cree de la plana', 'chr': 'ᏠᎨᏏ Ꮯ', 'cs': 'kríjština (z plání)', 'cy': 'Plains Cree', 'da': 'plains cree', 'de': 'Plains-Cree', 'dsb': 'plains creešćina', 'el': 'Κρι πεδιάδας', 'el-polyton': 'Κρι πεδιάδας', 'en': 'Plains Cree', 'es': 'cree de las llanuras', 'et': 'tasandikukrii', 'eu': 'lautadetako creera', 'fa': 'کری صحرایی', 'ff-Adlm': '𞤆𞤤𞤫𞤭𞤲𞤧 𞤑𞤪𞤭𞥅𞤪𞤫', 'fi': 'plains cree', 'fil': 'Plains Cree', 'fr': 'cri des plaines', 'ga': 'Plains Cree', 'gd': 'Cree nam Machair', 'gl': 'cree das chairas', 'gu': 'પ્લેઇન્સ ક્રી', 'ha': 'Plains Cree', 'he': 'קרי מישורים', 'hi': 'प्लेन्स क्री', 'hi-Latn': 'Plains Cree', 'hr': 'plains cree', 'hsb': 'plains creešćina', 'ht': 'cri des plaines', 'hu': 'síksági krí', 'hy': 'հարթավայրերի քրի', 'ia': 'cree del planas', 'id': 'Cree Dataran', 'ig': 'Plains Cree', 'is': 'nehiyawak', 'it': 'cree delle pianure', 'ja': '平原クリー語', 'jv': 'Kree Polos', 'ka': 'დაბლობის კრი', 'kk': 'жазықтағы кри тілі', 'kk-Cyrl': 'жазықтағы кри тілі', 'km': 'គ្រីតំបន់វាលរាប', 'kn': 'ಪ್ಲೇನ್ಸ್ ಕ್ರೀ', 'ko': '평원 크리어', 'kok': 'प्लेन्स क्री', 'kok-Deva': 'प्लेन्स क्री', 'ku': 'kriya bejayî', 'ky': 'өрөөндүк криче (индей тили)', 'lo': 'ເພລນຄຣີ', 'lt': 'supraprastinta kri', 'lv': 'līdzenumu krī', 'mai': 'प्लेंस क्री', 'mi': 'Pareina Kirī', 'mk': 'прериски кријски', 'ml': 'പ്ലെയ്\u200cൻസ് ക്രീ', 'mn': 'плэйн кри', 'mr': 'प्लेन्स क्री', 'ms': 'Plains Cree', 'my': 'ပလိန်းစ် ခရီ', 'nb': 'prærie-cree', 'ne': 'प्लेन्स क्री', 'nl': 'Plains Cree', 'nn': 'prærie-cree', 'no': 'prærie-cree', 'or': 'ପ୍ଲେନ୍ସ କ୍ରି', 'pa': 'ਪਲੇਨਸ ਕ੍ਰੀ', 'pa-Guru': 'ਪਲੇਨਸ ਕ੍ਰੀ', 'pcm': 'Krii fọ Plén Lángwej', 'pl': 'kri równinny', 'ps': 'پلینز کری', 'pt': 'cree das planícies', 'qu': 'Plains Cree Simi', 'ro': 'cree (Prerii)', 'ru': 'равнинный кри', 'sc': 'cree de sas campuras', 'sd': 'پلينز ڪري', 'sd-Arab': 'پلينز ڪري', 'si': 'ප්ලේන්ස් ක්\u200dරී', 'sk': 'plains cree', 'sl': 'nižinska krijščina', 'so': 'Plains Cree', 'sq': 'krijishte fusharake', 'sr': 'плаинскри', 'sr-Cyrl': 'плаинскри', 'sr-Latn': 'plainskri', 'sv': 'slättcree', 'sw': 'Kikri (Maeneo Tambarare)', 'ta': 'சமவெளி க்ரீ', 'te': 'ప్లెయిన్స్ క్రీ', 'th': 'เพลนส์ครี', 'ti': 'ክሪ ፕሌንስ', 'tk': 'düzdeçi kri dili', 'to': 'lea fakakilī-toafa', 'tr': 'Ova Kricesi', 'uk': 'рівнинна крі', 'ur': 'پلینز کری', 'uz': 'tekislik kri', 'uz-Latn': 'tekislik kri', 'vec': 'cree de le pianure', 'vi': 'Tiếng Plains Cree', 'yo': 'Èdè Papa Kri', 'yue': '平原克里文', 'yue-Hans': '平原克里文', 'yue-Hant': '平原克里文', 'zh': '平原克里语', 'zh-Hans': '平原克里语', 'zh-Hant': '平原克里語', 'zu': 'Plains Cree'},
    'crl': {'af': 'Noordoos-Cree', 'am': 'ሰሜን ምስራቃዊ ክሪ', 'ar': 'الكري شمال الشرقية', 'as': 'নৰ্দাৰ্ন ইষ্ট ক্ৰী', 'az': 'şimal-şəqri kri', 'az-Latn': 'şimal-şəqri kri', 'be': 'паўночна-ўсходняя кры', 'be-tarask': 'паўночна-ўсходняя кры', 'bg': 'североизточен крий', 'bn': 'উত্তর পূর্ব ক্রী', 'br': 'krieg ar Biz', 'brx': 'साहायारि सानजा क्री', 'bs': 'sjeveroistočni kri', 'bs-Cyrl': 'сјеверно источни кре', 'bs-Latn': 'sjeveroistočni kri', 'ca': 'cree nord-oriental', 'ca-ES-valencia': 'cree nord-oriental', 'chr': 'ᏧᏴᏢ ᏗᎧᎸᎬ Ꮯ', 'cs': 'kríjština (severovýchodní)', 'cy': 'Gogledd Dwyrain Cree', 'da': 'nordøstcree', 'de': 'Northern East Cree', 'dsb': 'dłujkozajtšna creešćina', 'el': 'Βορειοανατολικά Κρι', 'el-polyton': 'Βορειοανατολικά Κρι', 'en': 'Northern East Cree', 'es': 'cree nororiental', 'et': 'põhja-idakrii', 'eu': 'ipar-ekialdeko creera', 'fa': 'کری شمال شرقی', 'ff-Adlm': 'Vote 𞤑𞤪𞤭𞥅𞤪𞤫 𞤐𞤢𞤲𞤬𞤵𞤯𞤲𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤫', 'fi': 'east cree (pohjoinen)', 'fil': 'Northern East Cree', 'fr': 'cri de l’Est (dialecte du Nord)', 'ga': 'Craís Thoir Thuaidh', 'gd': 'Cree Ear-thuathach', 'gl': 'cree do nordeste', 'gu': 'ઉત્તરી પૂર્વ ક્રી', 'ha': 'Cree na Arewacin-Gabas', 'he': 'קרי צפון מזרחי', 'hi': 'नॉर्दन ईस्ट क्री', 'hi-Latn': 'Northern East Cree', 'hr': 'sjevernoistočni cree', 'hsb': 'sewjernowuchodna creešćina', 'ht': 'cri de l’Est (dialecte du Nord)', 'hu': 'északkeleti krí', 'hy': 'հյուսիս-արևելյան քրի', 'ia': 'cree del nord-est', 'id': 'East Cree Utara', 'ig': 'Northern East Cree', 'is': 'norðaustur-cree', 'it': 'cree nord-orientale', 'ja': '東部クリー語(北部)', 'jv': 'Kree Lor Segara', 'ka': 'ჩრდილო-აღმოსავლეთის კრი', 'kk': 'солтүстік-шығыс кри тілі', 'kk-Cyrl': 'солтүстік-шығыс кри тілі', 'km': 'គ្រីខាងកើត\u200bប៉ែកខាងជើង', 'kn': 'ಉತ್ತರ ಪೂರ್ವ ಕ್ರೀ', 'ko': '북동부 크리어', 'kok': 'ईशान्य क्री', 'kok-Deva': 'ईशान्य क्री', 'ku': 'krîya rojhilat ya bakurî', 'ky': 'чыгыш криче (индей тилдери)', 'lo': 'ຄຣີຕາເວັນອອກສ່ຽງເໜືອ', 'lt': 'šiaurės rytų kri', 'lv': 'ziemeļaustrumu krī', 'mai': 'उत्तर-पूरबी क्री', 'mi': 'Kirī Raki-mā-Rāwhiti', 'mk': 'североисточен кријски', 'ml': 'നോർത്തേൺ ഈസ്റ്റ് ക്രീ', 'mn': 'зүүн хойд кри', 'mr': 'नॉर्दर्न ईस्ट क्री', 'ms': 'Timur Laut Cree', 'my': 'မြောက်ပိုင်း အရှေ့ခရီ', 'nb': 'nordlig østcree', 'ne': 'उत्तर पूर्वी क्री', 'nl': 'Noordoost-Cree', 'nn': 'nordleg aust-cree', 'no': 'nordlig østcree', 'or': 'ଉତ୍ତର ପୂର୍ବ କ୍ରୀ', 'pa': 'ਉੱਤਰੀ ਪੂਰਬੀ ਕ੍ਰੀ', 'pa-Guru': 'ਉੱਤਰੀ ਪੂਰਬੀ ਕ੍ਰੀ', 'pcm': 'Nọ́tán Íst Krií Lángwej', 'pl': 'kri północno-wschodni', 'ps': 'شمالي ختیځ کری', 'pt': 'cree do nordeste', 'qu': 'Cree del Noreste Simi', 'ro': 'cree de nord-est', 'ru': 'северо-восточный кри', 'sc': 'cree nord-orientale', 'sd': 'اترين اوڀر ڪري', 'sd-Arab': 'اترين اوڀر ڪري', 'si': 'උතුරු නැගෙනහිර ක්\u200dරී', 'sk': 'northern east cree', 'sl': 'severovzhodna krijščina', 'so': 'Northern East Cree', 'sq': 'krijishte verilindore', 'sr': 'североисточни кри', 'sr-Cyrl': 'североисточни кри', 'sr-Latn': 'severoistočni kri', 'sv': 'nordost-cree', 'sw': 'Kikrii cha Kaskazini Mashariki', 'ta': 'வடகிழக்கு க்ரீ', 'te': 'ఉత్తర తూర్పు క్రీ', 'th': 'ครีตะวันออกเฉียงเหนือ', 'ti': 'ሰሜናዊ ምብራቕ ክሪ', 'tk': 'demirgazyk-gündogar kri dili', 'to': 'lea fakakilī-tokelauhahake', 'tr': 'Kuzey Doğu Kricesi', 'uk': 'північна іст-крі', 'ur': 'شمال مشرقی کری', 'uz': 'shomoli-sharqiy kri', 'uz-Latn': 'shomoli-sharqiy kri', 'vec': 'cree nord orientale', 'vi': 'Tiếng Cree Đông Bắc', 'yo': 'Èdè ti Ila oorun Ariwa Kri', 'yue': '東北克里文', 'yue-Hans': '东北克里文', 'yue-Hant': '東北克里文', 'zh': '东北部克里语', 'zh-Hans': '东北部克里语', 'zh-Hant': '北部東克里語', 'zu': 'isi-Northern East Cree'},
    'crm': {'af': 'Moose Cree', 'am': 'ሙዝ ክሪ', 'ar': 'الموس-كري', 'as': 'মুজ ক্ৰী', 'az': 'muz kri', 'az-Latn': 'muz kri', 'be': 'муская кры', 'be-tarask': 'муская кры', 'bg': 'муус крее', 'bn': 'মুস ক্রী', 'br': 'krieg ar cʼhornôg', 'brx': 'मूज क्री', 'bs': 'mus kri', 'bs-Cyrl': 'мосе кре', 'bs-Latn': 'mus kri', 'ca': 'moose cree', 'ca-ES-valencia': 'moose cree', 'chr': 'ᎠᏫ ᎡᏆ Ꮯ', 'cs': 'kríjština (Moose)', 'cy': 'Moose Cree', 'da': 'Moose-cree', 'de': 'Moose Cree', 'dsb': 'moode creešćina', 'el': 'Μους Κρι', 'el-polyton': 'Μους Κρι', 'en': 'Moose Cree', 'es': 'cree moose', 'et': 'põdrakrii', 'eu': 'Mooseko creera', 'fa': 'موس کری', 'ff-Adlm': '𞤃𞤮𞥅𞤧𞤫 𞤑𞤪𞤭𞥅𞤪𞤫', 'fi': 'moose cree', 'fil': 'Moose Cree', 'fr': 'cri de Moose', 'ga': 'Moose Cree', 'gd': 'Moose Cree', 'gl': 'cree de Moose', 'gu': 'મૂઝ ક્રી', 'ha': 'Moose Cree', 'he': 'מוס קרי', 'hi': 'मूस क्री', 'hi-Latn': 'Moose Cree', 'hr': 'moose cree', 'hsb': 'moode creešćina', 'ht': 'cri de Moose', 'hu': 'moose krí', 'hy': 'մուզ քրի', 'ia': 'cree moose', 'id': 'Moose Cree', 'ig': 'Moose Cree', 'is': 'moose cree', 'it': 'cree moose', 'ja': 'ムースクリー語', 'jv': 'Moose Cree', 'ka': 'მუსური კრი', 'kk': 'мус кри тілі', 'kk-Cyrl': 'мус кри тілі', 'km': 'មូសគ្រី', 'kn': 'ಮೂಸ್ ಕ್ರೀ', 'ko': '무스크리어', 'kok': 'मूझ क्री', 'kok-Deva': 'मूझ क्री', 'ku': 'krîya mûsî', 'ky': 'муус криче (индей тили)', 'lo': 'ມູສຄຣີ', 'lt': 'muskri', 'lv': 'mūsu krī', 'mai': 'मूज क्री', 'mi': 'Mūhi Kirī', 'mk': 'лосовски кријски', 'ml': 'മൂസ് ക്രീ', 'mn': 'мүүс кри', 'mr': 'मूस क्री', 'ms': 'Moose Cree', 'my': 'မုစ်ခရီး', 'nb': 'moose cree', 'ne': 'मुज क्री', 'nl': 'Moose Cree', 'nn': 'moose cree', 'no': 'moose cree', 'or': 'ମୁସେ କ୍ରୀ', 'pa': 'ਮੂਜ਼ ਕ੍ਰੀ', 'pa-Guru': 'ਮੂਜ਼ ਕ੍ਰੀ', 'pcm': 'Muse Krií Lángwej', 'pl': 'kri Moose', 'ps': 'mcr', 'pt': 'moose cree', 'qu': 'Moose Cree Simi', 'ro': 'cree (Moose)', 'ru': 'мусский кри', 'sc': 'cree moose', 'sd': 'موس ڪري', 'sd-Arab': 'موس ڪري', 'si': 'මූස් ක්\u200dරී', 'sk': 'moose cree', 'sl': 'moose-krijščina', 'so': 'Moose Cree', 'sq': 'krijishte e Muzit', 'sr': 'музкри', 'sr-Cyrl': 'музкри', 'sr-Latn': 'muzkri', 'sv': 'moose cree', 'sw': 'Kikrii cha Moose', 'ta': 'மூஸ் க்ரீ', 'te': 'మూస్ క్రీ', 'th': 'มูสครี', 'ti': 'ሙስ ክሪ', 'tk': 'los-kri dili', 'to': 'lea fakamose-kilī', 'tr': 'Moose Kricesi', 'uk': 'муз-крі', 'ur': 'موس کری', 'uz': 'mus kri', 'uz-Latn': 'mus kri', 'vec': 'cree mose', 'vi': 'Tiếng Moose Cree', 'yo': 'Èdè Moose Kri', 'yue': '駝鹿克里文', 'yue-Hans': '驼鹿克里文', 'yue-Hant': '駝鹿克里文', 'zh': '穆斯克里语', 'zh-Hans': '穆斯克里语', 'zh-Hant': '穆斯克里文', 'zu': 'isi-Moose Cree'},
    'crn': {'en': 'El Nayar Cora'},
    'cro': {'en': 'Crow'},
    'crp': {'en': 'Creoles and pidgins'},
    'crp-gep': {'en': 'Greenlandic Pidgin'},
    'crp-rsn': {'en': 'Russenorsk'},
    'crp-tpr': {'en': 'Taimyr Pidgin Russian'},
    'crq': {'en': "Iyo'wujwa Chorote"},
    'crr': {'af': 'Carolina-Algonkin', 'am': 'ካሮሊና አልጎንክዊያን', 'ar': 'الألغونكوية كارولينا', 'as': 'কেৰ’লিনা আলগংকিয়ান', 'az': 'Karolina alonkincəsi', 'az-Latn': 'Karolina alonkincəsi', 'be': 'каралінская алганкійская', 'be-tarask': 'каралінская алганкійская', 'bg': 'каролински алгонкин', 'bn': 'ক্যারোলিনা অ্যাল্গঙ্কুইয়ান', 'br': 'algonkeg Carolina', 'brx': 'केर’लिना एल्गंक्यान', 'bs': 'sjevernokarolinški algonkvijski', 'bs-Cyrl': 'алгонкуански за Каролину', 'bs-Latn': 'sjevernokarolinški algonkvijski', 'ca': 'algonquí de Carolina', 'ca-ES-valencia': 'algonquí de Carolina', 'chr': 'ᎠᎵᎦᏂᏈᎠᎾ', 'cs': 'algonkinština (Karolína)', 'cy': 'Carolina Algonquian', 'da': 'carolina algonquisk', 'de': 'Carolina-Algonkin', 'dsb': 'carolina algonquianšćina', 'el': 'Καρολίνα Αλγκονκιάν', 'el-polyton': 'Καρολίνα Αλγκονκιάν', 'en': 'Carolina Algonquian', 'es': 'algonquino de Carolina', 'et': 'Carolina algonkini', 'eu': 'Carolinako algonkinera', 'fa': 'آلگانکوین کارولینا', 'ff-Adlm': '𞤀𞤤𞤺𞤮𞤲𞤳𞤭𞤲𞤪𞤫 𞤑𞤢𞥄𞤪𞤤𞤭𞤲𞤢', 'fi': 'pamlico', 'fil': 'Carolina Algonquian', 'fr': 'algonquin de Caroline', 'ga': 'teanga Algancach Carolina', 'gd': 'Algonquianais Charolina', 'gl': 'algonquino de Carolina', 'gu': 'કેરોલિના એલ્ગોનક્વિઅન', 'ha': 'Carolina Algonquian', 'he': 'אלגונקוויאן בקרוליינה', 'hi': 'कैरोलाइना एल्गोंक्वीयन', 'hi-Latn': 'Carolina Algonquian', 'hr': 'karolinski algonkijski', 'hsb': 'carolina algonquianšćina', 'ht': 'algonquin de Caroline', 'hu': 'karolinai algonkin', 'hy': 'կարոլինական ալգոնկիներեն', 'ia': 'algonquino de Carolina', 'id': 'Carolina Algonquian', 'ig': 'Carolina Algonquian', 'is': 'Karólínu-algonkvínska', 'it': 'algonchino della Carolina', 'ja': 'カロライナ・アルゴンキン語', 'jv': 'Karolina Algonquian', 'ka': 'კაროლინური ალგონკინი', 'kk': 'каролиналық алгонкин тілі', 'kk-Cyrl': 'каролиналық алгонкин тілі', 'km': 'អាល់ហ្គនខ្វៀន ខារ៉ូលីណា', 'kn': 'ಕೆರೊಲಿನಾ ಅಲ್ಗೊಂಕ್ವಿಯನ್', 'ko': '캐롤라이나 알곤킨어', 'kok': 'कॅरोलीना अल्गॉन्क्वियन', 'kok-Deva': 'कॅरोलीना अल्गॉन्क्वियन', 'ku': 'zimanê karolina algonquianî', 'ky': 'каролиналык алгонкинче', 'lo': 'ຄາໂຣລິນາ ອາກອນຄວຽນ', 'lt': 'pamlikų', 'lv': 'Karolīnas algonkinu', 'mai': 'केरोलिना अलगोनकुइन', 'mi': 'Arakōkiana Kararaina', 'mk': 'каролински алгонкински', 'ml': 'കരോലീന അൽഗോൻക്വിയാൻ', 'mn': 'каролина алгонкинск хэл', 'mr': 'कॅरोलिना अल्गोंक्वियन', 'ms': 'Carolina Algonquian', 'my': 'ကယ်ရိုလိုင်းနား အယ်လ်ဂွန်းကီယန်', 'nb': 'carolinsk-algonkinsk', 'ne': 'क्यारोलिना एल्गोनक्वियन', 'nl': 'Carolina Algonkisch', 'nn': 'carolinsk-algonkinsk', 'no': 'carolinsk-algonkinsk', 'or': 'କାରୋଲିନା ଆଲଗୋନ୍କିଆନ୍', 'pa': 'ਕੈਰੋਲੀਨਾ ਐਲਗੋਂਕਵਿਅਨ', 'pa-Guru': 'ਕੈਰੋਲੀਨਾ ਐਲਗੋਂਕਵਿਅਨ', 'pcm': 'Karolína Algónkwían', 'pl': 'algonkiński (Karolina Północna)', 'ps': 'Car Alg', 'pt': 'algonquiano Carolina', 'qu': 'Algonquian Carolina', 'ro': 'algonquiană Carolina', 'ru': 'каролинский алгонкинский', 'sc': 'algonchinu de sa Carolina', 'sd': 'ڪيرولينا الگانڪويئن', 'sd-Arab': 'ڪيرولينا الگانڪويئن', 'si': 'කැරොලිනා ඇල්ගොන්කියන්', 'sk': 'karolínska algonkčina', 'sl': 'karolinska algonkinščina', 'so': 'Carolina Algonquian', 'sq': 'algonkuianishte e Karolinës', 'sr': 'каролиншки алгонквијан', 'sr-Cyrl': 'каролиншки алгонквијан', 'sr-Latn': 'karolinški algonkvijan', 'sv': 'Carolina-algonkinska', 'sw': 'Kipamliko cha Carolina', 'ta': 'கரோலினா அல்கோன்குயன்', 'te': 'కరోలినా అల్గోన్\u200cక్వియన్', 'th': 'อัลกอนควินแคโรไลนา', 'ti': 'ካሮሊና አልጎንጉያኛ', 'tk': 'karolina algonkin dili', 'to': 'lea fakaʻalakonikuia-kalolina', 'tr': 'Carolina Algonkin dili', 'uk': 'каролінська алгонкінська', 'ur': 'کیرولینا الگونکوئن', 'uz': 'karolin algonkin', 'uz-Latn': 'karolin algonkin', 'vec': 'algonchin de la Carolina', 'vi': 'Tiếng Carolina Algonquian', 'yo': 'Èdè Alonkuia ti Karolina', 'yue': '卡羅來納州阿爾岡昆語', 'yue-Hans': '卡罗来纳州阿尔冈昆语', 'yue-Hant': '卡羅來納州阿爾岡昆語', 'zh': '卡罗莱纳州阿尔冈昆语', 'zh-Hans': '卡罗莱纳州阿尔冈昆语', 'zh-Hant': '卡羅萊納阿爾岡昆語', 'zu': 'isi-Carolina Algonquian'},
    'crs': {'af': 'Seselwa Franskreools', 'am': 'ሰሰላዊ ክሬኦሊ ፈረንሳይኛ', 'ar': 'الفرنسية الكريولية السيشيلية', 'as': 'ছেছেলৱা ক্ৰিওল ফ্ৰেন্স', 'ast': 'francés criollu seselwa', 'az': 'Seyşel kreol fransızcası', 'az-Cyrl': 'сејшел креолу', 'az-Latn': 'Seyşel kreol fransızcası', 'be': 'сэсэльва', 'be-tarask': 'сэсэльва', 'bg': 'сеселва, креолски френски', 'bgn': 'سیشلی کریولین فرانسوی', 'bn': 'সেসেলওয়া ক্রেওল ফ্রেঞ্চ', 'br': 'kreoleg Sechelez', 'bs': 'seselva kreolski francuski', 'bs-Latn': 'seselva kreolski francuski', 'ca': 'francès crioll de les Seychelles', 'ca-ES-valencia': 'francès crioll de les Seychelles', 'ccp': '𑄥𑄬𑄥𑄬𑄣𑄧𑄤 𑄇𑄳𑄢𑄬𑄃𑄮𑄣𑄴 𑄜𑄳𑄢𑄬𑄐𑄴𑄌𑄧', 'ce': 'сейшелийн креолийн', 'chr': 'ᏎᏎᎵᏩ ᏟᏲᎵ ᎠᏂᎦᎸ', 'ckb': 'فەرەنسیی سیشێلی', 'cs': 'kreolština (seychelská)', 'cy': 'Ffrangeg Seselwa Creole', 'da': 'seselwa (kreol-fransk)', 'de': 'Seychellenkreol', 'el': 'Κρεολικά Γαλλικά Σεϋχελλών', 'el-polyton': 'Κρεολικά Γαλλικά Σεϋχελλών', 'en': 'Seselwa Creole French', 'es': 'criollo seychelense', 'et': 'seišelli', 'eu': 'Seychelleetako kreolera', 'fa': 'سیشل آمیختهٔ فرانسوی', 'fi': 'seychellienkreoli', 'fil': 'Seselwa Creole French', 'fo': 'seselwa creole franskt', 'fr': 'créole seychellois', 'ga': 'Criól Fraincise Seselwa', 'gd': 'Seiseallais', 'gl': 'seselwa (crioulo das Seychelles)', 'gu': 'સેસેલ્વા ક્રેઓલે ફ્રેન્ચ', 'he': 'קריאולית (סיישל)', 'hi': 'सेसेल्वा क्रिओल फ्रेंच', 'hi-Latn': 'Seselwa Creole French', 'hr': 'sejšelski kreolski', 'ht': 'créole seychellois', 'hu': 'szeszelva kreol francia', 'hy': 'սեյշելյան խառնակերտ ֆրանսերեն', 'ia': 'creolo seychellese', 'id': 'Seselwa Kreol Prancis', 'is': 'seychelles-kreólska', 'it': 'creolo delle Seychelles', 'ja': 'セーシェル・クレオール語', 'ka': 'სესელვა-კრეოლური ფრანგული', 'kgp': 'sejséri krijoro-frỹsej', 'kk': 'сейшельдік креол тілі', 'kk-Cyrl': 'сейшельдік креол тілі', 'km': 'សេសេលវ៉ាគ្រីអូល (បារាំង)', 'kn': 'ಸೆಸೆಲ್ವಾ ಕ್ರಯೋಲ್ ಫ್ರೆಂಚ್', 'ko': '세이셸 크리올 프랑스어', 'kok': 'सेसेल्वा क्रयॉल फ्रेन्च', 'kok-Deva': 'सेसेल्वा क्रयॉल फ्रेन्च', 'ky': 'сеселва креол французча', 'lo': 'ເຊເຊວາ ໂຄຣດ ຝຣັ່ງ', 'lt': 'Seišelių kreolų ir prancūzų', 'lv': 'franciskā kreoliskā valoda (Seišelu salas)', 'mk': 'француски (Сеселва креоли)', 'ml': 'സെഷൽവ ക്രിയോൾ ഫ്രഞ്ച്', 'mn': 'сеселва креол франц', 'mr': 'सेसेल्वा क्रिओल फ्रेंच', 'ms': 'Perancis Seselwa Creole', 'mt': 'Franċiż tas-Seselwa Creole', 'my': 'ခရီအိုလီ', 'nb': 'seselwa', 'ne': 'सेसेल्वा क्रिओल फ्रान्सेली', 'nl': 'Seychellencreools', 'nn': 'seselwa (fransk-kreolsk)', 'no': 'seselwa', 'or': 'ସେସେଲୱା କ୍ରେଓଲେ ଫ୍ରେଞ୍ଚ୍', 'pa': 'ਸੇਸੇਲਵਾ ਕ੍ਰਿਓਲ ਫ੍ਰੈਂਚ', 'pa-Guru': 'ਸੇਸੇਲਵਾ ਕ੍ਰਿਓਲ ਫ੍ਰੈਂਚ', 'pl': 'kreolski seszelski', 'ps': 'سسيلوا ڪروئل فرانسوي', 'pt': 'crioulo francês seichelense', 'ro': 'creolă franceză seselwa', 'ru': 'сейшельский креольский', 'sd': 'سيسلوا ڪريئول فرانسي', 'sd-Arab': 'سيسلوا ڪريئول فرانسي', 'si': 'සෙසෙල්ව ක්\u200dරොල් ෆ්\u200dරෙන්ච්', 'sk': 'seychelská kreolčina', 'sl': 'sejšelska francoska kreolščina', 'smn': 'Seychellij kreoliranska', 'sq': 'frëngjishte kreole seselve', 'sr': 'сејшелски креолски француски', 'sr-Cyrl': 'сејшелски креолски француски', 'sr-Latn': 'sejšelski kreolski francuski', 'sv': 'seychellisk kreol', 'sw': 'Krioli ya Shelisheli', 'ta': 'செசெல்வா க்ரெயோல் பிரெஞ்சு', 'te': 'సెసేల్వా క్రియోల్ ఫ్రెంచ్', 'th': 'ครีโอลเซเซลส์ฝรั่งเศส', 'tk': 'seselwa kreole-fransuz dili', 'to': 'lea fakaseselua-falanisē', 'tr': 'Seselwa Kreole Fransızcası', 'uk': 'сейшельська креольська', 'ur': 'سیسلوا کریولے فرانسیسی', 'uz': 'kreol (Seyshel)', 'uz-Latn': 'kreol (Seyshel)', 'vi': 'Tiếng Pháp Seselwa Creole', 'yrl': 'kiriuru frãsei seixeriwara', 'yue': '法語克里奧爾混合語', 'yue-Hans': '法语克里奥尔混合语', 'yue-Hant': '法語克里奧爾混合語', 'zh': '塞舌尔克里奥尔语', 'zh-Hans': '塞舌尔克里奥尔语', 'zh-Hant': '塞席爾克里奧爾法文', 'zu': 'i-Seselwa Creole French'},
    'crt': {'en': "Iyojwa'ja Chorote"},
    'crv': {'en': 'Chaura'},
    'crw': {'en': 'Chrau'},
    'crx': {'en': 'Carrier'},
    'cry': {'en': 'Cori'},
    'crz': {'en': 'Cruzeño'},
    'cs': {'af': 'Tsjeggies', 'agq': 'Chɛ̂', 'ak': 'Kyɛk kasa', 'am': 'ቼክኛ', 'ar': 'التشيكية', 'as': 'চেক', 'asa': 'Kicheki', 'ast': 'checu', 'az': 'çex', 'az-Arab': 'چئکجه', 'az-Cyrl': 'чех', 'az-Latn': 'çex', 'bal-Latn': 'Chek', 'bas': 'Hɔp u cɛ̂k', 'be': 'чэшская', 'be-tarask': 'чэшская', 'bem': 'Ichi Cheki', 'bez': 'Hicheki', 'bg': 'чешки', 'bgn': 'چیکی', 'bm': 'cɛkikan', 'bn': 'চেক', 'br': 'tchekeg', 'brx': 'चैक', 'bs': 'češki', 'bs-Cyrl': 'чешки', 'bs-Latn': 'češki', 'ca': 'txec', 'ca-ES-valencia': 'txec', 'ccp': '𑄌𑄬𑄇𑄴', 'ce': 'чехийн', 'cgg': 'Oruceeki', 'chr': 'ᏤᎩ', 'ckb': 'چێکی', 'cs': 'čeština', 'cy': 'Tsieceg', 'da': 'tjekkisk', 'dav': 'Kichecki', 'de': 'Tschechisch', 'dje': 'Cek senni', 'dsb': 'češćina', 'dyo': 'sek', 'dz': 'ཅེཀ་ཁ', 'ebu': 'Kĩcheki', 'ee': 'tsɛkgbe', 'el': 'Τσεχικά', 'el-polyton': 'Τσεχικά', 'en': 'Czech', 'en-Dsrt': '𐐕𐐯𐐿', 'eo': 'ĉeĥa', 'es': 'checo', 'et': 'tšehhi', 'eu': 'txekiera', 'ewo': 'Ǹkɔ́bɔ tsɛ́g', 'fa': 'چکی', 'ff': 'Cekkere', 'ff-Adlm': '𞤕𞤫𞤳𞤧𞤭𞤲𞤢𞥄𞤪𞤫', 'ff-Latn': 'Cekkere', 'fi': 'tšekki', 'fil': 'Czech', 'fo': 'kekkiskt', 'fr': 'tchèque', 'fur': 'cec', 'fy': 'Tsjechysk', 'ga': 'Seicis', 'gd': 'Seicis', 'gl': 'checo', 'gsw': 'Tschechisch', 'gu': 'ચેક', 'guz': 'Kichecki', 'ha': 'Cek', 'he': 'צ׳כית', 'hi': 'चेक', 'hi-Latn': 'Czech', 'hr': 'češki', 'hsb': 'čěšćina', 'ht': 'tchèque', 'hu': 'cseh', 'hy': 'չեխերեն', 'ia': 'checo', 'id': 'Ceko', 'ie': 'tchec', 'ig': 'Czech', 'is': 'tékkneska', 'it': 'ceco', 'ja': 'チェコ語', 'jmc': 'Kyichecki', 'jv': 'Ceska', 'ka': 'ჩეხური', 'kab': 'Tačikit', 'kam': 'Kichecki', 'kde': 'Chichechi', 'kea': 'txeku', 'kgp': 'séko', 'khq': 'Cek senni', 'ki': 'Kichecki', 'kk': 'чех тілі', 'kk-Cyrl': 'чех тілі', 'kln': 'kutitab Chek', 'km': 'ឆែក', 'kn': 'ಜೆಕ್', 'ko': '체코어', 'kok': 'चेक', 'kok-Deva': 'चेक', 'ks': 'چیک', 'ks-Arab': 'چیک', 'ksb': 'Kichecki', 'ksf': 'ricɛ́k', 'ksh': 'Tschäschesch', 'ku': 'çekî', 'kxv': 'cek', 'kxv-Deva': 'चेक', 'kxv-Latn': 'cek', 'kxv-Orya': 'ଚେକ୍', 'kxv-Telu': 'చెక్', 'ky': 'чехче', 'lag': 'Kɨchéeki', 'lb': 'Tschechesch', 'lg': 'Luceeke', 'lkt': 'Czech Iyápi', 'lld': 'cech', 'ln': 'litshekɛ', 'lo': 'ເຊກ', 'lt': 'čekų', 'lu': 'Tsheki', 'luo': 'Kichecki', 'luy': 'Kichecki', 'lv': 'čehu', 'mai': 'चेक', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lcheki', 'mer': 'Kĩcheki', 'mfe': 'tchek', 'mg': 'Tseky', 'mgh': 'Icheki', 'mi': 'Tieke', 'mk': 'чешки', 'ml': 'ചെക്ക്', 'mn': 'чех', 'mr': 'झेक', 'ms': 'Czech', 'mt': 'Ċek', 'mua': 'syekya', 'my': 'ချက်', 'mzn': 'چکی', 'naq': 'Czechǁî gowab', 'nb': 'tsjekkisk', 'nd': 'isi-Czech', 'ne': 'चेक', 'nl': 'Tsjechisch', 'nmg': 'Kiɛl bó tchɛk', 'nn': 'tsjekkisk', 'no': 'tsjekkisk', 'nqo': 'ߗߍߞߌߞߊ߲', 'nus': 'Thok cik', 'nyn': 'Oruceeki', 'om': 'Afaan Czech', 'or': 'ଚେକ୍', 'os': 'чехаг', 'pa': 'ਚੈੱਕ', 'pa-Guru': 'ਚੈੱਕ', 'pcm': 'Chẹ́k Lángwej', 'pl': 'czeski', 'ps': 'چېکي', 'pt': 'tcheco', 'qu': 'Checo Simi', 'rm': 'tschec', 'rn': 'Igiceke', 'ro': 'cehă', 'rof': 'Kichecki', 'ru': 'чешский', 'rw': 'Igiceke', 'rwk': 'Kyichecki', 'sah': 'Чиэхтии', 'saq': 'Kichecki', 'sbp': 'Ishisheki', 'sc': 'tzecu', 'sd': 'چيڪ', 'sd-Arab': 'چيڪ', 'se': 'čeahkagiella', 'seh': 'tcheco', 'ses': 'Cek senni', 'sg': 'Tyêki', 'shi': 'ⵜⴰⵜⵛⵉⴽⵉⵜ', 'shi-Latn': 'Tatcikit', 'shi-Tfng': 'ⵜⴰⵜⵛⵉⴽⵉⵜ', 'si': 'චෙක්', 'sk': 'čeština', 'sl': 'češčina', 'smn': 'tšeekikielâ', 'sn': 'chiCzech', 'so': 'Jeeg', 'sq': 'çekisht', 'sr': 'чешки', 'sr-Cyrl': 'чешки', 'sr-Latn': 'češki', 'sv': 'tjeckiska', 'sw': 'Kicheki', 'ta': 'செக்', 'te': 'చెక్', 'teo': 'Kichecki', 'tg': 'чехӣ', 'th': 'เช็ก', 'ti': 'ቸክኛ', 'tk': 'çeh dili', 'to': 'lea fakaseki', 'tr': 'Çekçe', 'tt': 'чех', 'twq': 'Cek senni', 'tzm': 'Tačikt', 'ug': 'چېخچە', 'uk': 'чеська', 'ur': 'چیک', 'uz': 'chex', 'uz-Cyrl': 'чехча', 'uz-Latn': 'chex', 'vai': 'ꗿꗡ', 'vai-Latn': 'Chɛ', 'vai-Vaii': 'ꗿꗡ', 'vec': 'seco', 'vi': 'Tiếng Séc', 'vun': 'Kyichecki', 'wae': 'Tšečiš', 'wo': 'Cek', 'xnr': 'चेक', 'xog': 'Oluceeke', 'yav': 'cɛ́kɛ́ɛ', 'yi': 'טשעכיש', 'yo': 'Èdè Seeki', 'yrl': 'tieku', 'yue': '捷克文', 'yue-Hans': '捷克文', 'yue-Hant': '捷克文', 'zgh': 'ⵜⴰⵜⵛⵉⴽⵉⵜ', 'zh': '捷克语', 'zh-Hans': '捷克语', 'zh-Hant': '捷克文', 'zu': 'isi-Czech'},
    'csa': {'en': 'Chiltepec Chinantec'},
    'csb': {'ar': 'الكاشبايان', 'ast': 'kashubianu', 'az': 'kaşubyan', 'az-Latn': 'kaşubyan', 'bg': 'кашубски', 'bn': 'কাশুবিয়ান', 'br': 'kachoubeg', 'brx': 'काशुबियान्', 'bs': 'kašubijanski', 'bs-Cyrl': 'кашубијански', 'bs-Latn': 'kašubijanski', 'ca': 'caixubi', 'ca-ES-valencia': 'caixubi', 'ccp': '𑄇𑄥𑄪𑄝𑄨𑄠𑄚𑄴', 'cs': 'kašubština', 'da': 'kasjubisk', 'de': 'Kaschubisch', 'el': 'Κασούμπιαν', 'el-polyton': 'Κασούμπιαν', 'en': 'Kashubian', 'es': 'casubio', 'et': 'kašuubi', 'fa': 'کاشوبی', 'fi': 'kašubi', 'fr': 'kachoube', 'fy': 'Kasjoebysk', 'ga': 'Caisiúibis', 'gd': 'Caisiubais', 'gsw': 'Kaschubisch', 'gu': 'કાશુબિયન', 'he': 'קשובית', 'hi': 'काशुबियन', 'hi-Latn': 'Kashubian', 'hr': 'kašupski', 'ht': 'kachoube', 'hu': 'kasub', 'id': 'Kashubia', 'is': 'kasúbíska', 'it': 'kashubian', 'ja': 'カシューブ語', 'ka': 'კაშუბური', 'kgp': 'kasumijỹ', 'kn': 'ಕಶುಬಿಯನ್', 'ko': '카슈비아어', 'ks': 'کَشوٗبِیَن', 'ks-Arab': 'کَشوٗبِیَن', 'ksh': 'Kaschuhbesch', 'lb': 'Kaschubesch', 'lo': 'ກາຊູບຽນ', 'lt': 'kašubų', 'lv': 'kašubu', 'mk': 'кашупски', 'ml': 'കാഷുബിയാൻ', 'mr': 'काशुबियन', 'mt': 'Kashubian', 'nb': 'kasjubisk', 'ne': 'कासुवियन', 'nl': 'Kasjoebisch', 'nn': 'kasjubisk', 'no': 'kasjubisk', 'or': 'କାଶୁବିଆନ୍', 'pl': 'kaszubski', 'pt': 'kashubian', 'rm': 'kaschubic', 'ro': 'cașubiană', 'ru': 'кашубский', 'sk': 'kašubčina', 'sl': 'kašubščina', 'sr': 'кашупски', 'sr-Cyrl': 'кашупски', 'sr-Latn': 'kašupski', 'sv': 'kasjubiska', 'ta': 'கஷுபியன்', 'te': 'కషుబియన్', 'th': 'คาซูเบียน', 'to': 'lea fakakasiupia', 'tr': 'Kashubian', 'ug': 'كاسزۇبىچە', 'uk': 'кашубська', 'vi': 'Tiếng Kashubia', 'yi': 'קאַשוביש', 'yrl': 'kaxubiã', 'yue': '卡舒布文', 'yue-Hans': '卡舒布文', 'yue-Hant': '卡舒布文', 'zh': '卡舒比语', 'zh-Hans': '卡舒比语', 'zh-Hant': '卡舒布文'},
    'csc': {'en': 'Catalan Sign Language'},
    'csd': {'en': 'Chiangmai Sign Language'},
    'cse': {'en': 'Czech Sign Language'},
    'csf': {'en': 'Cuba Sign Language'},
    'csg': {'en': 'Chilean Sign Language'},
    'csh': {'en': 'Asho Chin'},
    'csi': {'en': 'Coast Miwok'},
    'csj': {'en': 'Songlai Chin'},
    'csk': {'en': 'Jola-Kasa'},
    'csl': {'en': 'Chinese Sign Language'},
    'csm': {'en': 'Central Sierra Miwok'},
    'csn': {'en': 'Colombian Sign Language'},
    'cso': {'en': 'Sochiapam Chinantec'},
    'csp': {'en': 'Southern Ping Chinese'},
    'csq': {'en': 'Croatia Sign Language'},
    'csr': {'en': 'Costa Rican Sign Language'},
    'css': {'en': 'Southern Ohlone'},
    'cst': {'en': 'Northern Ohlone'},
    'csu': {'en': 'Central Sudanic languages'},
    'csv': {'en': 'Sumtu Chin'},
    'csw': {'af': 'Swampy Cree', 'ak': 'Tadeɛm Kreefoɔ Kasa', 'am': 'ስዋምፒ ክሪ', 'ar': 'السوامبي-كري', 'as': 'স্বাম্পী ক্ৰী', 'az': 'bataqlıq kricəsi', 'az-Latn': 'bataqlıq kricəsi', 'bal-Latn': 'Swampi Kri', 'be': 'балотная кры', 'be-tarask': 'балотная кры', 'bg': 'суампи крий', 'bn': 'সোয়াম্পি ক্রী', 'br': 'krieg ar gwernioù', 'brx': 'स्वाम्पी क्री', 'bs': 'močvarni kri', 'bs-Cyrl': 'свампи кре', 'bs-Latn': 'močvarni kri', 'ca': 'swampy cree', 'ca-ES-valencia': 'swampy cree', 'chr': 'ᏌᎷᏱ Ꮯ', 'cs': 'kríjština (z bažin)', 'csw': 'ᓀᐦᐃᓇᐍᐏᐣ', 'cy': 'Swampy Cree', 'da': 'swampy cree', 'de': 'Swampy Cree', 'dsb': 'swampy creešćina', 'el': 'Κρι Βάλτου', 'el-polyton': 'Κρι Βάλτου', 'en': 'Swampy Cree', 'es': 'cree de los pantanos', 'et': 'sookrii', 'eu': 'zingiretako creera', 'fa': 'کری سوامپی', 'ff-Adlm': '𞤑𞤪𞤭𞥅𞤪𞤫 𞤅𞤢𞤱𞤨𞤭𞥅', 'fi': 'swampy cree', 'fil': 'Latian na Cree', 'fo': 'swampy cree', 'fr': 'cri des marais', 'ga': 'Swampy Cree', 'gd': 'Omushkego', 'gl': 'cree dos pantanos', 'gu': 'સ્વેમ્પી ક્રી', 'ha': 'Swampy Cree', 'he': 'קרי של אזור הביצות', 'hi': 'स्वॉम्पी क्री', 'hi-Latn': 'Swampy Cree', 'hr': 'močvarni cree', 'hsb': 'swampy creešćina', 'ht': 'cri des marais', 'hu': 'mocsári krí', 'hy': 'ճահճի քրի', 'ia': 'cree del paludes', 'id': 'Cree Rawa', 'ig': 'Asụsụ Swampy Kree', 'is': 'maskekon', 'it': 'cree delle paludi', 'ja': '湿原クリー語', 'jv': 'Kree Rawa', 'ka': 'ჭაობის კრი', 'kk': 'батпақты жердің кри тілі', 'kk-Cyrl': 'батпақты жердің кри тілі', 'km': 'គ្រីតំបន់ភក់ល្បាប់', 'kn': 'ಸ್ವಾಂಪಿ ಕ್ರೀ', 'ko': '습지 크리어', 'kok': 'स्वॉम्पी क्री', 'kok-Deva': 'स्वॉम्पी क्री', 'ku': 'krîya swampî', 'ky': 'суампи криче (индей тили)', 'lo': 'ຊວາມປີຄຣີ', 'lt': 'pelkynų kri', 'lv': 'purvu krī', 'mai': 'स्वम्पी क्री', 'mi': 'Wāpi Kirī', 'mk': 'мочуришен кријски', 'ml': 'സ്വാംപി ക്രീ', 'mn': 'свампи кри', 'mr': 'स्वॅम्पी क्री', 'ms': 'Swampy Cree', 'my': 'ဆွမ်ပီ ခရီ', 'nb': 'myr-cree', 'ne': 'स्वाम्पी क्री', 'nl': 'Swampy Cree', 'nn': 'myr-cree', 'no': 'myr-cree', 'or': 'ସ୍ୱାମ୍ପି କ୍ରି', 'pa': 'ਸਵੈਂਪੀ ਕ੍ਰੀ', 'pa-Guru': 'ਸਵੈਂਪੀ ਕ੍ਰੀ', 'pcm': 'Swampi Krií Lángwej', 'pl': 'kri bagienny', 'ps': 'سومپی کری', 'pt': 'cree swampy', 'qu': 'Swampy Cree Simi', 'ro': 'cree (Mlaștini)', 'ru': 'болотный кри', 'sc': 'cree de sas paludes', 'sd': 'سوامپي ڪري', 'sd-Arab': 'سوامپي ڪري', 'si': 'මඩ වගුරු ක්\u200dරී', 'sk': 'swampy cree', 'sl': 'močvirska krijščina', 'so': 'Swampy Cree', 'sq': 'krijishte e moçaleve (Ontario)', 'sr': 'мочварни кри', 'sr-Cyrl': 'мочварни кри', 'sr-Latn': 'močvarni kri', 'sv': 'träskcree', 'sw': 'Kiomushkego', 'ta': 'சதுப்பு நில க்ரீ', 'te': 'స్వాంపీ క్రీ', 'th': 'สวอมปีครี', 'ti': 'ክሪ ረግረግ', 'tk': 'batgalyk kri dili', 'to': 'lea fakakilī-ano', 'tr': 'Bataklık Kricesi', 'uk': 'свампі-крі', 'ur': 'سوامپی کری', 'uz': 'botqoq kri', 'uz-Latn': 'botqoq kri', 'vec': 'cree de le paludi', 'vi': 'Tiếng Swampy Cree', 'yo': 'Èdè Swampi Kri', 'yue': '沼澤克里文', 'yue-Hans': '沼泽克里文', 'yue-Hant': '沼澤克里文', 'zh': '沼泽克里语', 'zh-Hans': '沼泽克里语', 'zh-Hant': '沼澤克里文', 'zu': 'Swampy Cree'},
    'csx': {'en': 'Cambodian Sign Language'},
    'csy': {'en': 'Siyin Chin'},
    'csz': {'en': 'Coos'},
    'cta': {'en': 'Tataltepec Chatino'},
    'ctc': {'en': 'Chetco'},
    'ctd': {'en': 'Tedim Chin'},
    'cte': {'en': 'Tepinapa Chinantec'},
    'ctg': {'en': 'Chittagonian'},
    'cth': {'en': 'Thaiphum Chin'},
    'ctl': {'en': 'Tlacoatzintepec Chinantec'},
    'ctm': {'en': 'Chitimacha'},
    'ctn': {'en': 'Chhintange'},
    'cto': {'en': 'Emberá-Catío'},
    'ctp': {'en': 'Western Highland Chatino'},
    'cts': {'en': 'Northern Catanduanes Bikol'},
    'ctt': {'en': 'Wayanad Chetti'},
    'ctu': {'en': 'Chol'},
    'cty': {'en': 'Moundadan Chetty'},
    'ctz': {'en': 'Zacatepec Chatino'},
    'cu': {'af': 'Kerkslawies', 'am': 'ቸርች ስላቪክ', 'ar': 'سلافية كنسية', 'as': 'চাৰ্চ শ্লেভিক', 'ast': 'eslávicu eclesiásticu', 'az': 'slavyan', 'az-Cyrl': 'славјан', 'az-Latn': 'slavyan', 'bal-Latn': 'Charch Sláwi', 'be': 'царкоўнаславянская', 'be-tarask': 'царкоўнаславянская', 'bg': 'църковнославянски', 'bgn': 'سلاواکی کلیسایی', 'bn': 'চার্চ স্লাভিক', 'br': 'slavoneg iliz', 'brx': 'चर्च स्लाविक्', 'bs': 'staroslavenski', 'bs-Cyrl': 'старославенски', 'bs-Latn': 'staroslavenski', 'ca': 'eslau eclesiàstic', 'ca-ES-valencia': 'eslau eclesiàstic', 'ccp': '𑄌𑄢𑄴𑄌𑄴 𑄥𑄳𑄣𑄞𑄨𑄇𑄴', 'ce': 'килсславянийн', 'chr': 'ᏧᏂᎳᏫᏍᏗ ᏍᎳᏫᎪ', 'ckb': 'سلاویی کلیسەیی', 'cs': 'staroslověnština', 'cy': 'Hen Slafoneg', 'da': 'kirkeslavisk', 'de': 'Kirchenslawisch', 'dsb': 'cerkwinosłowjańšćina', 'el': 'Εκκλησιαστικά Σλαβικά', 'el-polyton': 'Εκκλησιαστικά Σλαβικά', 'en': 'Church Slavic', 'es': 'eslavo eclesiástico', 'et': 'kirikuslaavi', 'eu': 'elizako eslaviera', 'fa': 'اسلاوی کلیسایی', 'ff-Adlm': '𞤅𞤭𞤤𞤾𞤭𞤳𞤪𞤫 𞤕𞤮𞥅𞤷𞤭', 'fi': 'kirkkoslaavi', 'fil': 'Church Slavic', 'fo': 'kirkju sláviskt', 'fr': 'slavon d’église', 'fur': 'sclâf de glesie', 'fy': 'Kerkslavysk', 'ga': 'Slavais na hEaglaise', 'gd': 'Slàbhais na h-Eaglaise', 'gl': 'eslavo eclesiástico', 'gsw': 'Chileslawisch', 'gu': 'ચર્ચ સ્લાવિક', 'ha': 'Church Slavic', 'he': 'סלאבית כנסייתית עתיקה', 'hi': 'चर्च साल्विक', 'hi-Latn': 'Church Slavic', 'hr': 'crkvenoslavenski', 'hsb': 'cyrkwinosłowjanšćina', 'ht': 'slavon d’église', 'hu': 'egyházi szláv', 'hy': 'սլավոներեն, եկեղեցական', 'ia': 'slavo ecclesiastic', 'id': 'Bahasa Gereja Slavonia', 'ig': 'Church slavic', 'is': 'kirkjuslavneska', 'it': 'slavo ecclesiastico', 'ja': '教会スラブ語', 'jv': 'Slavia Gerejani', 'ka': 'საეკლესიო სლავური', 'kea': 'slavu klériku', 'kgp': 'eravu ekresijatiko', 'kk': 'шіркеулік славян тілі', 'kk-Cyrl': 'шіркеулік славян тілі', 'km': 'ឈើជស្លាវិក', 'kn': 'ಚರ್ಚ್ ಸ್ಲಾವಿಕ್', 'ko': '교회 슬라브어', 'kok': 'चर्च स्लेव्हीक', 'kok-Deva': 'चर्च स्लेव्हीक', 'ks': 'چٔرچ سلاوِک', 'ks-Arab': 'چٔرچ سلاوِک', 'ksh': 'Kerscheßlahwesch', 'ku': 'slavîya kenîseyî', 'ky': 'чиркөө славянча', 'lb': 'Kiercheslawesch', 'lld': 'slav eclesiastich', 'lo': 'ໂບດສລາວິກ', 'lt': 'bažnytinė slavų', 'lv': 'baznīcslāvu', 'mk': 'црковнословенски', 'ml': 'ചർച്ച് സ്ലാവിക്', 'mn': 'сүмийн славян', 'mr': 'चर्च स्लाव्हिक', 'ms': 'Slavik Gereja', 'mt': 'Slaviku tal-Knisja', 'my': 'ချပ်ချ် စလာဗစ်', 'nb': 'kirkeslavisk', 'ne': 'चर्च स्लाभिक', 'nl': 'Kerkslavisch', 'nn': 'kyrkjeslavisk', 'no': 'kirkeslavisk', 'nqo': 'ߛߌߟߊߝ߭ߐ߲ߞߊ߲ ߓߊ߬ߕߏ߬ߓߏ߲߬ߞߊ߲', 'or': 'ଚର୍ଚ୍ଚ ସ୍ଲାଭିକ୍', 'pa': 'ਚਰਚ ਸਲਾਵੀ', 'pa-Guru': 'ਚਰਚ ਸਲਾਵੀ', 'pcm': 'Chọ́ch Slávik', 'pl': 'cerkiewnosłowiański', 'ps': 'د کليسا سلاوي', 'pt': 'eslavo eclesiástico', 'qu': 'Eslavo Eclesiástico Simi', 'rm': 'slav da baselgia', 'ro': 'slavonă', 'ru': 'церковнославянский', 'sc': 'islavu eclesiàsticu', 'sd': 'چرچ سلاوی', 'sd-Arab': 'چرچ سلاوی', 'si': 'චර්ච් ස්ලැවික්', 'sk': 'cirkevná slovančina', 'sl': 'stara cerkvena slovanščina', 'smn': 'kirkkoslaavi', 'so': 'Kaniisadda Islaafik', 'sq': 'sllavishte kishtare', 'sr': 'црквенословенски', 'sr-Cyrl': 'црквенословенски', 'sr-Latn': 'crkvenoslovenski', 'sv': 'kyrkslaviska', 'sw': 'Kislovakia cha Kanisa', 'ta': 'சர்ச் ஸ்லாவிக்', 'te': 'చర్చ్ స్లావిక్', 'th': 'เชอร์ชสลาวิก', 'ti': 'ቤተ-ክርስትያን ስላቭኛ', 'tk': 'buthana slaw dili', 'to': 'lea fakasilavia-fakasiasi', 'tr': 'Kilise Slavcası', 'ug': 'قەدىمكى سلاۋيانچە', 'uk': 'церковнословʼянська', 'ur': 'چرچ سلاوک', 'uz': 'slavyan (cherkov)', 'uz-Cyrl': 'славянча (черков)', 'uz-Latn': 'slavyan (cherkov)', 'vec': 'zlavo de cieza', 'vi': 'Tiếng Slavơ Nhà thờ', 'xnr': 'चर्च साल्विक', 'yi': 'קלויסטער־סלאַוויש', 'yo': 'Èdè Síláfííkì Ilé Ìjọ́sìn', 'yrl': 'isirawu ekeresiatiku', 'yue': '宗教斯拉夫文', 'yue-Hans': '宗教斯拉夫文', 'yue-Hant': '宗教斯拉夫文', 'zh': '教会斯拉夫语', 'zh-Hans': '教会斯拉夫语', 'zh-Hant': '宗教斯拉夫文', 'zu': 'isi-Church Slavic'},
    'cua': {'en': 'Cua'},
    'cub': {'en': 'Cubeo'},
    'cuc': {'en': 'Usila Chinantec'},
    'cug': {'en': 'Chungmboko'},
    'cuh': {'en': 'Chuka'},
    'cui': {'en': 'Cuiba'},
    'cuj': {'en': 'Mashco Piro'},
    'cuk': {'en': 'San Blas Kuna'},
    'cul': {'en': 'Culina'},
    'cum': {'en': 'Cumeral'},
    'cuo': {'en': 'Cumanagoto'},
    'cup': {'en': 'Cupeño'},
    'cuq': {'en': 'Cun'},
    'cur': {'en': 'Chhulung'},
    'cus': {'en': 'Cushitic languages'},
    'cus-pro': {'en': 'Proto-Cushitic'},
    'cut': {'en': 'Teutila Cuicatec'},
    'cuu': {'en': 'Tai Ya'},
    'cuv': {'en': 'Cuvok'},
    'cuw': {'en': 'Chukwa'},
    'cux': {'en': 'Tepeuxila Cuicatec'},
    'cuy': {'en': 'Cuitlatec'},
    'cv': {'af': 'Chuvash', 'ak': 'Kyuvahyi', 'am': 'ቹቫሽኛ', 'ar': 'التشوفاشي', 'as': 'চুভাচ', 'ast': 'chuvash', 'az': 'çuvaş', 'az-Arab': 'چۇواشجا', 'az-Cyrl': 'чуваш', 'az-Latn': 'çuvaş', 'bal-Latn': 'Chuwash', 'be': 'чувашская', 'be-tarask': 'чувашская', 'bg': 'чувашки', 'bgn': 'چواشی', 'bn': 'চুবাস', 'br': 'tchouvatch', 'brx': 'चुभास', 'bs': 'čuvaški', 'bs-Cyrl': 'чувашки', 'bs-Latn': 'čuvaški', 'ca': 'txuvaix', 'ca-ES-valencia': 'txuvaix', 'ccp': '𑄌𑄪𑄝𑄥𑄴', 'ce': 'чувашийн', 'chr': 'ᏧᏩᏏ', 'ckb': 'چووڤاشی', 'cs': 'čuvaština', 'cv': 'чӑваш', 'cy': 'Tshwfasheg', 'da': 'tjuvasjisk', 'de': 'Tschuwaschisch', 'dsb': 'chuvashišćina', 'el': 'Τσουβασικά', 'el-polyton': 'Τσουβασικά', 'en': 'Chuvash', 'es': 'chuvasio', 'et': 'tšuvaši', 'eu': 'txuvaxera', 'fa': 'چوواشی', 'ff-Adlm': '𞤕𞤵𞥅𞤾𞤢𞤧𞤪𞤫', 'fi': 'tšuvassi', 'fil': 'Chuvash', 'fo': 'chuvash', 'fr': 'tchouvache', 'fy': 'Tsjoevasjysk', 'ga': 'Suvaisis', 'gd': 'Chuvash', 'gl': 'chuvaxo', 'gsw': 'Tschuwaschisch', 'gu': 'ચૂવાશ', 'ha': 'Chuvash', 'he': 'צ׳ובאש', 'hi': 'चूवाश', 'hi-Latn': 'Chuvash', 'hr': 'čuvaški', 'hsb': 'chuvashišćina', 'ht': 'tchouvache', 'hu': 'csuvas', 'hy': 'չուվաշերեն', 'ia': 'chuvash', 'id': 'Chuvash', 'ig': 'Chuvash', 'is': 'sjúvas', 'it': 'ciuvascio', 'ja': 'チュヴァシ語', 'jv': 'Khuvash', 'ka': 'ჩუვაშური', 'kea': 'txuvaxi', 'kgp': 'suvase', 'kk': 'чуваш тілі', 'kk-Cyrl': 'чуваш тілі', 'km': 'ឈូវ៉ាស', 'kn': 'ಚುವಾಶ್', 'ko': '추바시어', 'kok': 'चुवाश', 'kok-Deva': 'चुवाश', 'ks': 'چُواش', 'ks-Arab': 'چُواش', 'ksh': 'Tschowaschesch', 'ku': 'çuvaşî', 'ky': 'чувашча', 'lb': 'Tschuwaschesch', 'lkt': 'Chuvash Iyápi', 'lld': 'ciuvascich', 'lo': 'ຊູວາຊ', 'lrc': 'چواشی', 'lt': 'čiuvašų', 'lv': 'čuvašu', 'mai': 'शुवास', 'mi': 'Tiuwhāhi', 'mk': 'чувашки', 'ml': 'ചുവാഷ്', 'mn': 'чуваш', 'mr': 'चूवाश', 'ms': 'Chuvash', 'mt': 'Chuvash', 'my': 'ချူဗက်ရှ်', 'mzn': 'چوواشی', 'nb': 'tsjuvasjisk', 'ne': 'चुभास', 'nl': 'Tsjoevasjisch', 'nn': 'tsjuvansk', 'no': 'tsjuvasjisk', 'om': 'Chuvash', 'or': 'ଚୁଭାଶ୍', 'os': 'чувашаг', 'pa': 'ਚੁਵਾਸ਼', 'pa-Guru': 'ਚੁਵਾਸ਼', 'pcm': 'Chúvash', 'pl': 'czuwaski', 'ps': 'چوواشي', 'pt': 'tchuvache', 'qu': 'Chuvash Simi', 'rm': 'tschuvasch', 'ro': 'ciuvașă', 'ru': 'чувашский', 'sc': 'ciuvàsciu', 'sd': 'چو واش', 'sd-Arab': 'چو واش', 'si': 'චවේෂ්', 'sk': 'čuvaština', 'sl': 'čuvaščina', 'smn': 'tšuvaskielâ', 'so': 'Chuvash', 'sq': 'çuvashisht', 'sr': 'чувашки', 'sr-Cyrl': 'чувашки', 'sr-Latn': 'čuvaški', 'sv': 'tjuvasjiska', 'sw': 'Kichuvash', 'ta': 'சுவாஷ்', 'te': 'చువాష్', 'th': 'ชูวัช', 'ti': 'ቹቫሽኛ', 'tk': 'çuwaş dili', 'to': 'lea fakasuvasa', 'tr': 'Çuvaşça', 'ug': 'چۇۋاشچە', 'uk': 'чуваська', 'ur': 'چوواش', 'uz': 'chuvash', 'uz-Cyrl': 'чуваш тили', 'uz-Latn': 'chuvash', 'vec': 'ciuvash', 'vi': 'Tiếng Chuvash', 'xnr': 'चूवाश', 'yo': 'Èdè Ṣufasi', 'yrl': 'tiuwaxi', 'yue': '楚瓦什文', 'yue-Hans': '楚瓦什文', 'yue-Hant': '楚瓦什文', 'zh': '楚瓦什语', 'zh-Hans': '楚瓦什语', 'zh-Hant': '楚瓦什文', 'zu': 'isi-Chuvash'},
    'cvg': {'en': 'Chug'},
    'cvn': {'en': 'Valle Nacional Chinantec'},
    'cwa': {'en': 'Kabwa'},
    'cwb': {'en': 'Maindo'},
    'cwd': {'en': 'Woods Cree'},
    'cwe': {'en': 'Kwere'},
    'cwg': {'en': 'Chewong'},
    'cwt': {'en': 'Kuwaataay'},
    'cy': {'af': 'Wallies', 'ak': 'Wɛɛhye Kasa', 'am': 'ዌልሽ', 'ar': 'الويلزية', 'as': 'ৱেলচ', 'ast': 'galés', 'az': 'uels', 'az-Cyrl': 'уелс', 'az-Latn': 'uels', 'bal-Latn': 'Wéli', 'be': 'валійская', 'be-tarask': 'валійская', 'bg': 'уелски', 'bgn': 'ولزی', 'bn': 'ওয়েলশ', 'br': 'kembraeg', 'brx': 'वैल्श', 'bs': 'velški', 'bs-Cyrl': 'велшки', 'bs-Latn': 'velški', 'ca': 'gal·lès', 'ca-ES-valencia': 'gal·lès', 'ccp': '𑄃𑄮𑄠𑄬𑄣𑄧𑄌𑄴', 'ce': 'валлийн', 'chr': 'ᏪᎵᏏ', 'ckb': 'وێلزی', 'cs': 'velština', 'cy': 'Cymraeg', 'da': 'walisisk', 'de': 'Walisisch', 'dsb': 'walizišćina', 'dz': 'ཝེལཤ་ཁ', 'ee': 'walesgbe', 'el': 'Ουαλικά', 'el-polyton': 'Οὐαλικά', 'en': 'Welsh', 'en-Dsrt': '𐐎𐐯𐑊𐑇', 'eo': 'kimra', 'es': 'galés', 'et': 'kõmri', 'eu': 'galesa', 'fa': 'ولزی', 'ff-Adlm': '𞤘𞤢𞤤𞤭𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'kymri', 'fil': 'Welsh', 'fo': 'walisiskt', 'fr': 'gallois', 'fur': 'galês', 'fy': 'Welsk', 'ga': 'Breatnais', 'gd': 'Cuimris', 'gl': 'galés', 'gsw': 'Walisisch', 'gu': 'વેલ્શ', 'ha': 'Welsh', 'haw': 'Wale', 'he': 'וולשית', 'hi': 'वेल्श', 'hi-Latn': 'Welsh', 'hr': 'velški', 'hsb': 'walizišćina', 'ht': 'gallois', 'hu': 'walesi', 'hy': 'ուելսերեն', 'ia': 'gallese', 'id': 'Welsh', 'ig': 'Welsh', 'is': 'velska', 'it': 'gallese', 'ja': 'ウェールズ語', 'jv': 'Welsh', 'ka': 'უელსური', 'kea': 'gales', 'kgp': 'garej', 'kk': 'валлий тілі', 'kk-Cyrl': 'валлий тілі', 'km': 'វេល', 'kn': 'ವೆಲ್ಶ್', 'ko': '웨일스어', 'kok': 'वेल्श', 'kok-Deva': 'वेल्श', 'ks': 'ویلش', 'ks-Arab': 'ویلش', 'ksh': 'Wallihsesch', 'ku': 'weylsî', 'ky': 'уелшче', 'lb': 'Walisesch', 'lkt': 'Wales Iyápi', 'lld': 'galesc', 'lo': 'ເວວ', 'lrc': 'ڤئلزی', 'lt': 'valų', 'lv': 'velsiešu', 'mai': 'बेल्श', 'mi': 'Werehi', 'mk': 'велшки', 'ml': 'വെൽഷ്', 'mn': 'уэльс', 'mr': 'वेल्श', 'ms': 'Wales', 'mt': 'Welsh', 'my': 'ဝေလ', 'mzn': 'ولزی', 'nb': 'walisisk', 'ne': 'वेल्श', 'nl': 'Welsh', 'nn': 'walisisk', 'no': 'walisisk', 'nqo': 'ߜ߭ߟߏߥߊߞߊ߲', 'om': 'Welishiffaa', 'or': 'ୱେଲ୍ସ', 'pa': 'ਵੈਲਸ਼', 'pa-Guru': 'ਵੈਲਸ਼', 'pcm': 'Wẹlsh', 'pl': 'walijski', 'ps': 'ويلشي', 'pt': 'galês', 'qu': 'Gales Simi', 'rm': 'kimric', 'ro': 'galeză', 'ru': 'валлийский', 'rw': 'Ikigaluwa', 'sc': 'gallesu', 'sd': 'ويلش', 'sd-Arab': 'ويلش', 'se': 'kymragiella', 'si': 'වෙල්ෂ්', 'sk': 'waleština', 'sl': 'valižanščina', 'smn': 'kymrikielâ', 'so': 'Welsh', 'sq': 'uellsisht', 'sr': 'велшки', 'sr-Cyrl': 'велшки', 'sr-Latn': 'velški', 'sv': 'walesiska', 'sw': 'Kiwelisi', 'ta': 'வேல்ஷ்', 'te': 'వెల్ష్', 'tg': 'валлӣ', 'th': 'เวลส์', 'ti': 'ዌልስኛ', 'tk': 'walliý dili', 'to': 'lea fakauēlesi', 'tr': 'Galce', 'tt': 'уэльс', 'ug': 'ۋېلشچە', 'uk': 'валлійська', 'ur': 'ویلش', 'uz': 'valliy', 'uz-Cyrl': 'уэлсча', 'uz-Latn': 'valliy', 'vec': 'galeze', 'vi': 'Tiếng Wales', 'wae': 'Walisiš', 'wo': 'Wels', 'xnr': 'वेल्श', 'yi': 'וועלשיש', 'yo': 'Èdè Welshi', 'yrl': 'garei', 'yue': '威爾斯文', 'yue-Hans': '威尔斯文', 'yue-Hant': '威爾斯文', 'zh': '威尔士语', 'zh-Hans': '威尔士语', 'zh-Hant': '威爾斯文', 'zu': 'isi-Welsh'},
    'cya': {'en': 'Nopala Chatino'},
    'cyb': {'en': 'Cayubaba'},
    'cyo': {'en': 'Cuyonon'},
    'czh': {'en': 'Huizhou Chinese'},
    'czk': {'en': 'Knaanic'},
    'czn': {'en': 'Zenzontepec Chatino'},
    'czo': {'en': 'Min Zhong Chinese'},
    'czt': {'en': 'Zotung Chin'},
    'da': {'af': 'Deens', 'ak': 'Dane kasa', 'am': 'ዴኒሽ', 'ar': 'الدانمركية', 'as': 'ডেনিচ', 'ast': 'danés', 'az': 'danimarka', 'az-Arab': 'دانجا', 'az-Cyrl': 'данимарка', 'az-Latn': 'danimarka', 'bal-Latn': 'Denmárki', 'be': 'дацкая', 'be-tarask': 'дацкая', 'bg': 'датски', 'bgn': 'ڈینمارکی', 'bn': 'ডেনিশ', 'br': 'daneg', 'brx': 'डानिश', 'bs': 'danski', 'bs-Cyrl': 'дански', 'bs-Latn': 'danski', 'ca': 'danès', 'ca-ES-valencia': 'danès', 'ccp': '𑄓𑄬𑄚𑄨𑄌𑄴', 'ce': 'датхойн', 'chr': 'ᏕᏂᏍ', 'ckb': 'دانماركی', 'cs': 'dánština', 'cy': 'Daneg', 'da': 'dansk', 'de': 'Dänisch', 'dsb': 'danšćina', 'dz': 'ཌེ་ནིཤ་ཁ', 'ee': 'denmarkgbe', 'el': 'Δανικά', 'el-polyton': 'Δανικά', 'en': 'Danish', 'en-Dsrt': '𐐔𐐩𐑌𐐮𐑇', 'eo': 'dana', 'es': 'danés', 'et': 'taani', 'eu': 'daniera', 'fa': 'دانمارکی', 'ff-Adlm': '𞤁𞤢𞥄𞤲𞤭𞤧𞤳𞤮𞥅𞤪𞤫', 'fi': 'tanska', 'fil': 'Danish', 'fo': 'danskt', 'fr': 'danois', 'fur': 'danês', 'fy': 'Deensk', 'ga': 'Danmhairgis', 'gd': 'Danmhairgis', 'gl': 'dinamarqués', 'gsw': 'Tänisch', 'gu': 'ડેનિશ', 'ha': 'Danish', 'haw': 'Kenemaka', 'he': 'דנית', 'hi': 'डेनिश', 'hi-Latn': 'Danish', 'hr': 'danski', 'hsb': 'danšćina', 'ht': 'danois', 'hu': 'dán', 'hy': 'դանիերեն', 'ia': 'danese', 'id': 'Dansk', 'ie': 'danesi', 'ig': 'Danish', 'is': 'danska', 'it': 'danese', 'ja': 'デンマーク語', 'jv': 'Dansk', 'ka': 'დანიური', 'kea': 'dinamarkes', 'kgp': 'nhinỹmỹrkej', 'kk': 'дат тілі', 'kk-Cyrl': 'дат тілі', 'km': 'ដាណឺម៉ាក', 'kn': 'ಡ್ಯಾನಿಶ್', 'ko': '덴마크어', 'kok': 'डॅनिश', 'kok-Deva': 'डॅनिश', 'ks': 'ڈینِش', 'ks-Arab': 'ڈینِش', 'ksh': 'Dähnesch', 'ku': 'danmarkî', 'kxv': 'denis', 'kxv-Deva': 'डेनिस', 'kxv-Latn': 'denis', 'kxv-Orya': 'ଡେନିସ୍', 'kxv-Telu': 'డెనిస్', 'ky': 'датча', 'lb': 'Dänesch', 'lkt': 'Dane Iyápi', 'lld': 'danesc', 'lo': 'ແດນິຊ', 'lrc': 'دانمارکی', 'lt': 'danų', 'lv': 'dāņu', 'mai': 'डेनिश', 'mi': 'Teina', 'mk': 'дански', 'ml': 'ഡാനിഷ്', 'mn': 'дани', 'mr': 'डॅनिश', 'ms': 'Denmark', 'mt': 'Daniż', 'my': 'ဒိန်းမတ်', 'mzn': 'دانمارکی', 'nb': 'dansk', 'ne': 'डेनिस', 'nl': 'Deens', 'nn': 'dansk', 'no': 'dansk', 'nqo': 'ߘߊߣߏߥߊߞߊ߲', 'om': 'Afaan Deenmaark', 'or': 'ଡାନିସ୍\u200c', 'os': 'даниаг', 'pa': 'ਡੈਨਿਸ਼', 'pa-Guru': 'ਡੈਨਿਸ਼', 'pcm': 'Dénísh Lángwej', 'pl': 'duński', 'ps': 'ډنمارکي', 'pt': 'dinamarquês', 'qu': 'Danes Simi', 'rm': 'danais', 'ro': 'daneză', 'ru': 'датский', 'rw': 'Ikidaninwa', 'sah': 'Даатскайдыы', 'sc': 'danesu', 'sd': 'ڊينش', 'sd-Arab': 'ڊينش', 'se': 'dánskkagiella', 'si': 'ඩැනිශ්', 'sk': 'dánčina', 'sl': 'danščina', 'smn': 'tanskakielâ', 'so': 'Dhaanish', 'sq': 'danisht', 'sr': 'дански', 'sr-Cyrl': 'дански', 'sr-Latn': 'danski', 'sv': 'danska', 'sw': 'Kidenmaki', 'ta': 'டேனிஷ்', 'te': 'డానిష్', 'tg': 'даниягӣ', 'th': 'เดนมาร์ก', 'ti': 'ዳኒሽ', 'tk': 'daniýa dili', 'to': 'lea fakatenimaʻake', 'tr': 'Danca', 'tt': 'дания', 'ug': 'دانىشچە', 'uk': 'данська', 'ur': 'ڈینش', 'uz': 'dan', 'uz-Cyrl': 'датча', 'uz-Latn': 'dan', 'vec': 'daneze', 'vi': 'Tiếng Đan Mạch', 'wae': 'Däniš', 'wo': 'Danuwa', 'xnr': 'डेनिश', 'yi': 'דעניש', 'yo': 'Èdè Denmaki', 'yrl': 'dinamarikei', 'yue': '丹麥文', 'yue-Hans': '丹麦文', 'yue-Hant': '丹麥文', 'zh': '丹麦语', 'zh-Hans': '丹麦语', 'zh-Hant': '丹麥文', 'zu': 'isi-Danish'},
    'daa': {'en': 'Dangaléat'},
    'dac': {'en': 'Dambi'},
    'dad': {'en': 'Marik'},
    'dae': {'en': 'Duupa'},
    'daf': {'en': 'Dan'},
    'dag': {'en': 'Dagbani'},
    'dah': {'en': 'Gwahatike'},
    'dai': {'en': 'Day'},
    'daj': {'en': 'Dar Fur Daju'},
    'dak': {'af': 'Dakotaans', 'am': 'ዳኮታ', 'ar': 'الداكوتا', 'as': 'ডাকোটা', 'ast': 'dakota', 'az': 'dakota', 'az-Cyrl': 'дакота', 'az-Latn': 'dakota', 'be': 'дакота', 'be-tarask': 'дакота', 'bg': 'дакотски', 'bgn': 'داکوتی', 'bn': 'ডাকোটা', 'br': 'dakota', 'brx': 'डाक’टा', 'bs': 'dakota', 'bs-Cyrl': 'дакота', 'bs-Latn': 'dakota', 'ca': 'dakota', 'ca-ES-valencia': 'dakota', 'ccp': '𑄓𑄇𑄮𑄑', 'ce': 'дакота', 'chr': 'ᏓᎪᏔ', 'ckb': 'داکۆتایی', 'cs': 'dakotština', 'cy': 'Dacotaeg', 'da': 'dakota', 'de': 'Dakota', 'dsb': 'dakotašćina', 'dz': 'ད་ཀོ་ཏ་ཁ', 'el': 'Ντακότα', 'el-polyton': 'Ντακότα', 'en': 'Dakota', 'en-Dsrt': '𐐔𐐲𐐿𐐬𐐻𐐲', 'es': 'dakota', 'et': 'siuu', 'eu': 'dakotera', 'fa': 'داکوتایی', 'ff-Adlm': '𞤁𞤢𞤳𞤮𞤼𞤢𞥄𞤪𞤫', 'fi': 'dakota', 'fil': 'Dakota', 'fo': 'dakota', 'fr': 'dakota', 'fy': 'Dakota', 'ga': 'Dacótais', 'gd': 'Dakota', 'gl': 'dakota', 'gsw': 'Takota', 'gu': 'દાકોતા', 'ha': 'Dakota', 'he': 'דקוטה', 'hi': 'दाकोता', 'hi-Latn': 'Dakota', 'hr': 'dakota jezik', 'hsb': 'dakotašćina', 'ht': 'dakota', 'hu': 'dakota', 'hy': 'դակոտա', 'ia': 'dakota', 'id': 'Dakota', 'ig': 'Dakota', 'is': 'dakóta', 'it': 'dakota', 'ja': 'ダコタ語', 'jv': 'Dakota', 'ka': 'დაკოტური', 'kgp': 'nakota', 'kk': 'дакота тілі', 'kk-Cyrl': 'дакота тілі', 'km': 'ដាកូតា', 'kn': 'ಡಕೋಟಾ', 'ko': '다코타어', 'kok': 'डाकोटा', 'kok-Deva': 'डाकोटा', 'ks': 'ڈکوٹا', 'ks-Arab': 'ڈکوٹا', 'ksh': 'Dakohtesch', 'ku': 'dakotayî', 'ky': 'дакотача', 'lb': 'Dakota-Sprooch', 'lkt': 'Dakȟótiyapi', 'lo': 'ດາໂກຕາ', 'lt': 'dakotų', 'lv': 'dakotu', 'mai': 'डकोटा', 'mi': 'Takōta', 'mk': 'дакота', 'ml': 'ഡകോട്ട', 'mn': 'дакота', 'mr': 'डाकोटा', 'ms': 'Dakota', 'mt': 'Dakota', 'my': 'ဒါကိုတာ', 'nb': 'dakota', 'ne': 'डाकोटा', 'nl': 'Dakota', 'nn': 'dakota', 'no': 'dakota', 'or': 'ଡାକୋଟା', 'pa': 'ਡਕੋਟਾ', 'pa-Guru': 'ਡਕੋਟਾ', 'pcm': 'Dakótá Lángwej', 'pl': 'dakota', 'ps': 'داکوتا', 'pt': 'dacota', 'qu': 'Dakota Simi', 'rm': 'dakota', 'ro': 'dakota', 'ru': 'дакота', 'sc': 'dakota', 'sd': 'ڊڪوٽا', 'sd-Arab': 'ڊڪوٽا', 'si': 'ඩකොටා', 'sk': 'dakotčina', 'sl': 'dakotščina', 'smn': 'dakotakielâ', 'so': 'Dakota', 'sq': 'dakotisht', 'sr': 'дакота', 'sr-Cyrl': 'дакота', 'sr-Latn': 'dakota', 'sv': 'dakota', 'sw': 'Kidakota', 'ta': 'டகோடா', 'te': 'డకోటా', 'th': 'ดาโกทา', 'ti': 'ዳኮታ', 'tk': 'dakota dili', 'to': 'lea fakatakota', 'tr': 'Dakotaca', 'ug': 'داكوتاچە', 'uk': 'дакота', 'ur': 'ڈاکوٹا', 'uz': 'dakota', 'uz-Cyrl': 'дакотча', 'uz-Latn': 'dakota', 'vec': 'dakota', 'vi': 'Tiếng Dakota', 'xnr': 'दाकोता', 'yo': 'Èdè Dakota', 'yrl': 'dakuta', 'yue': '達科他文', 'yue-Hans': '达科他文', 'yue-Hant': '達科他文', 'zh': '达科他语', 'zh-Hans': '达科他语', 'zh-Hant': '達科他文', 'zu': 'isi-Dakota'},
    'dal': {'en': 'Dahalo'},
    'dam': {'en': 'Damakawa'},
    'dao': {'en': 'Daai Chin'},
    'dap': {'en': 'Nisi (India)'},
    'daq': {'en': 'Dandami Maria'},
    'dar': {'af': 'Dargwa', 'am': 'ዳርግዋ', 'ar': 'الدارجوا', 'as': 'দাৰ্গৱা', 'ast': 'dargwa', 'az': 'darqva', 'az-Cyrl': 'даргва', 'az-Latn': 'darqva', 'be': 'даргінская', 'be-tarask': 'даргінская', 'bg': 'даргински', 'bgn': 'دارگوایی', 'bn': 'দার্গওয়া', 'br': 'dargwa', 'brx': 'दर्गवा', 'bs': 'dargva', 'bs-Cyrl': 'даргва', 'bs-Latn': 'dargva', 'ca': 'darguà', 'ca-ES-valencia': 'darguà', 'ccp': '𑄘𑄢𑄴𑄉𑄧𑄤', 'ce': 'даьргӀойн', 'chr': 'ᏓᎳᏆ', 'ckb': 'دارگینی', 'cs': 'dargština', 'cy': 'Dargwa', 'da': 'dargwa', 'de': 'Darginisch', 'dsb': 'dargwašćina', 'el': 'Ντάργκουα', 'el-polyton': 'Ντάργκουα', 'en': 'Dargwa', 'es': 'dargva', 'et': 'dargi', 'eu': 'darginera', 'fa': 'دارگوا', 'ff-Adlm': '𞤁𞤢𞤪𞤺𞤭𞤲𞤢𞥄𞤪𞤫', 'fi': 'dargi', 'fil': 'Dargwa', 'fo': 'dargwa', 'fr': 'dargwa', 'fy': 'Dargwa', 'ga': 'Dargais', 'gd': 'Dargwa', 'gl': 'dargwa', 'gsw': 'Targiinisch', 'gu': 'દાર્ગવા', 'ha': 'Dargwa', 'he': 'דרגווה', 'hi': 'दार्गवा', 'hi-Latn': 'Dargwa', 'hr': 'dargwa', 'hsb': 'dargwašćina', 'ht': 'dargwa', 'hu': 'dargva', 'hy': 'դարգիներեն', 'ia': 'dargwa', 'id': 'Dargwa', 'ig': 'Dargwa', 'is': 'dargva', 'it': 'dargwa', 'ja': 'ダルグワ語', 'jv': 'Dargwa', 'ka': 'დარგუული', 'kgp': 'narguva', 'kk': 'даргин тілі', 'kk-Cyrl': 'даргин тілі', 'km': 'ដាចវ៉ា', 'kn': 'ದರ್ಗ್ವಾ', 'ko': '다르그와어', 'kok': 'दार्ग्वा', 'kok-Deva': 'दार्ग्वा', 'ks': 'دَرگوا', 'ks-Arab': 'دَرگوا', 'ksh': 'Darjihnesch', 'ku': 'dargînî', 'ky': 'даргинче', 'lb': 'Darginesch', 'lkt': 'Dargwa Iyápi', 'lo': 'ດາກວາ', 'lt': 'dargva', 'lv': 'dargu', 'mai': 'डर्गबा', 'mi': 'Tākawa', 'mk': 'даргва', 'ml': 'ഡർഗ്വാ', 'mn': 'даргва', 'mr': 'दार्गवा', 'ms': 'Dargwa', 'mt': 'Dargwa', 'my': 'ဒါဂ်ဝါ', 'nb': 'dargwa', 'ne': 'दार्ग्वा', 'nl': 'Dargwa', 'nn': 'dargwa', 'no': 'dargwa', 'or': 'ଡାରାଗ୍ୱା', 'pa': 'ਦਾਰਗਵਾ', 'pa-Guru': 'ਦਾਰਗਵਾ', 'pcm': 'Dargwá Lángwej', 'pl': 'dargwijski', 'ps': 'درگوا', 'pt': 'dargwa', 'qu': 'Dargwa Simi', 'rm': 'dargwa', 'ro': 'dargwa', 'ru': 'даргинский', 'sc': 'dargua', 'sd': 'ڊارگوا', 'sd-Arab': 'ڊارگوا', 'si': 'ඩාර්ග්වා', 'sk': 'darginčina', 'sl': 'darginščina', 'smn': 'dargikielâ', 'so': 'Dargwa', 'sq': 'darguaisht', 'sr': 'даргински', 'sr-Cyrl': 'даргински', 'sr-Latn': 'darginski', 'sv': 'darginska', 'sw': 'Kidaragwa', 'ta': 'தார்குவா', 'te': 'డార్గ్వా', 'th': 'ดาร์กิน', 'ti': 'ዳርግዋ', 'tk': 'dargi dili', 'to': 'lea fakatalakuā', 'tr': 'Dargince', 'ug': 'دارگىۋاچە', 'uk': 'даргінська', 'ur': 'درگوا', 'uz': 'dargva', 'uz-Cyrl': 'даргинча', 'uz-Latn': 'dargva', 'vec': 'dargwa', 'vi': 'Tiếng Dargwa', 'xnr': 'दार्गवा', 'yo': 'Èdè Dagiwa', 'yrl': 'darigiwa', 'yue': '達爾格瓦文', 'yue-Hans': '达尔格瓦文', 'yue-Hant': '達爾格瓦文', 'zh': '达尔格瓦语', 'zh-Hans': '达尔格瓦语', 'zh-Hant': '達爾格瓦文', 'zu': 'isi-Dargwa'},
    'das': {'en': 'Daho-Doo'},
    'dau': {'en': 'Dar Sila Daju'},
    'dav': {'af': 'Taita', 'am': 'ታይታ', 'ar': 'تيتا', 'as': 'তেইতা', 'ast': 'taita', 'az': 'taita', 'az-Cyrl': 'таита', 'az-Latn': 'taita', 'bal-Latn': 'Táitá', 'be': 'таіта', 'be-tarask': 'таіта', 'bg': 'таита', 'bgn': 'تایتایی', 'bn': 'তাইতা', 'br': 'taita', 'brx': 'टाइटा', 'bs': 'taita', 'bs-Cyrl': 'таита', 'bs-Latn': 'taita', 'ca': 'taita', 'ca-ES-valencia': 'taita', 'ccp': '𑄖𑄳𑄆𑄖', 'ce': 'таита', 'chr': 'ᏔᎢᏔ', 'ckb': 'تایتا', 'cs': 'taita', 'cy': 'Taita', 'da': 'taita', 'dav': 'Kitaita', 'de': 'Taita', 'dsb': 'taita', 'el': 'Τάιτα', 'el-polyton': 'Τάιτα', 'en': 'Taita', 'es': 'taita', 'et': 'davida', 'eu': 'taitera', 'fa': 'تایتا', 'ff-Adlm': '𞤚𞤢𞤭𞤼𞤢𞥄𞤪𞤫', 'fi': 'taita', 'fil': 'Taita', 'fo': 'taita', 'fr': 'taita', 'fy': 'Taita', 'ga': 'Taita', 'gd': 'Taita', 'gl': 'taita', 'gu': 'તૈતા', 'ha': 'Taita', 'he': 'טאיטה', 'hi': 'तैता', 'hi-Latn': 'Taita', 'hr': 'taita', 'hsb': 'taita', 'ht': 'taita', 'hu': 'taita', 'hy': 'թաիթա', 'ia': 'taita', 'id': 'Taita', 'ig': 'Taita', 'is': 'taíta', 'it': 'taita', 'ja': 'タイタ語', 'jv': 'Taita', 'ka': 'ტაიტა', 'kea': 'taita', 'kgp': 'tajta', 'kk': 'таита тілі', 'kk-Cyrl': 'таита тілі', 'km': 'តៃតា', 'kn': 'ಟೈಟ', 'ko': '타이타어', 'kok': 'तायता', 'kok-Deva': 'तायता', 'ksh': 'Dawedahnesch', 'ku': 'tayîtayî', 'ky': 'таитача', 'lb': 'Taita', 'lo': 'ໄຕຕາ', 'lrc': 'تایتا', 'lt': 'taitų', 'lv': 'taitu', 'mai': 'टैटा', 'mi': 'Taita', 'mk': 'таита', 'ml': 'തൈത', 'mn': 'тайта', 'mr': 'तायता', 'ms': 'Taita', 'mt': 'Taita', 'my': 'တိုင်တာ', 'mzn': 'تایتا', 'nb': 'taita', 'ne': 'ताइता', 'nl': 'Taita', 'nn': 'taita', 'no': 'taita', 'nqo': 'ߕߊߦߌߕߊߞߊ߲', 'or': 'ତାଇତି', 'pa': 'ਟੇਟਾ', 'pa-Guru': 'ਟੇਟਾ', 'pcm': 'Taíta', 'pl': 'taita', 'ps': 'ټایټا', 'pt': 'taita', 'qu': 'Taita Simi', 'rm': 'taita', 'ro': 'taita', 'ru': 'таита', 'sc': 'taita', 'sd': 'تائيتا', 'sd-Arab': 'تائيتا', 'si': 'ටයිටා', 'sk': 'taita', 'sl': 'taitajščina', 'smn': 'taitakielâ', 'so': 'Taiita', 'sq': 'tajtaisht', 'sr': 'таита', 'sr-Cyrl': 'таита', 'sr-Latn': 'taita', 'sv': 'taita', 'sw': 'Kitaita', 'ta': 'டைடா', 'te': 'టైటా', 'th': 'ไททา', 'ti': 'ታይታ', 'tk': 'taita dili', 'to': 'lea fakataita', 'tr': 'Taita', 'ug': 'تايتاچە', 'uk': 'таїта', 'ur': 'تائتا', 'uz': 'taita', 'uz-Cyrl': 'таитача', 'uz-Latn': 'taita', 'vec': 'taita', 'vi': 'Tiếng Taita', 'xnr': 'तैता', 'yo': 'Táítà', 'yrl': 'taita', 'yue': '台塔文', 'yue-Hans': '台塔文', 'yue-Hant': '台塔文', 'zh': '台塔语', 'zh-Hans': '台塔语', 'zh-Hant': '台塔文', 'zu': 'isi-Taita'},
    'daw': {'en': 'Davawenyo'},
    'dax': {'en': 'Dayi'},
    'day': {'en': 'Land Dayak languages'},
    'daz': {'en': 'Dao'},
    'dba': {'en': 'Bangime'},
    'dbb': {'en': 'Deno'},
    'dbd': {'en': 'Dadiya'},
    'dbe': {'en': 'Dabe'},
    'dbf': {'en': 'Edopi'},
    'dbg': {'en': 'Dogul Dom Dogon'},
    'dbi': {'en': 'Doka'},
    'dbj': {'en': "Ida'an"},
    'dbl': {'en': 'Dyirbal'},
    'dbm': {'en': 'Duguri'},
    'dbn': {'en': 'Duriankere'},
    'dbo': {'en': 'Dulbu'},
    'dbp': {'en': 'Duwai'},
    'dbq': {'en': 'Daba'},
    'dbr': {'en': 'Dabarre'},
    'dbt': {'en': 'Ben Tey Dogon'},
    'dbu': {'en': 'Bondum Dom Dogon'},
    'dbv': {'en': 'Dungu'},
    'dbw': {'en': 'Bankan Tey Dogon'},
    'dby': {'en': 'Dibiyaso'},
    'dcc': {'en': 'Deccan'},
    'dcr': {'en': 'Negerhollands'},
    'dda': {'en': 'Dadi Dadi'},
    'ddd': {'en': 'Dongotono'},
    'dde': {'en': 'Doondo'},
    'ddg': {'en': 'Fataluku'},
    'ddi': {'en': 'West Goodenough'},
    'ddj': {'en': 'Jaru'},
    'ddn': {'en': 'Dendi (Benin)'},
    'ddo': {'en': 'Dido'},
    'ddr': {'en': 'Dhudhuroa'},
    'dds': {'en': 'Donno So Dogon'},
    'ddw': {'en': 'Dawera-Daweloor'},
    'de': {'af': 'Duits', 'agq': 'Dzamɛ̀', 'ak': 'Gyaaman', 'am': 'ጀርመንኛ', 'ar': 'الألمانية', 'as': 'জাৰ্মান', 'asa': 'Kijerumani', 'ast': 'alemán', 'az': 'alman', 'az-Arab': 'آلمانجا', 'az-Cyrl': 'алман', 'az-Latn': 'alman', 'bal-Latn': 'Jarman', 'bas': 'Hɔp u jamân', 'be': 'нямецкая', 'be-tarask': 'нямецкая', 'bem': 'Ichi Jemani', 'bez': 'Hijerumani', 'bg': 'немски', 'bgn': 'جرمنی', 'blo': 'gɩjaamaja', 'bm': 'alimaɲikan', 'bn': 'জার্মান', 'br': 'alamaneg', 'brx': 'जार्मान', 'bs': 'njemački', 'bs-Cyrl': 'њемачки', 'bs-Latn': 'njemački', 'ca': 'alemany', 'ca-ES-valencia': 'alemany', 'ccp': '𑄎𑄢𑄴𑄟𑄚𑄴', 'ce': 'немцойн', 'ceb': 'German', 'cgg': 'Orugirimaani', 'chr': 'ᏙᎢᏥ', 'ckb': 'ئەڵمانی', 'cs': 'němčina', 'cv': 'нимӗҫ', 'cy': 'Almaeneg', 'da': 'tysk', 'dav': 'Kijerumani', 'de': 'Deutsch', 'dje': 'Almaŋ senni', 'doi': 'जर्मन', 'dsb': 'nimšćina', 'dyo': 'alman', 'dz': 'ཇཱར་མཱན་ཁ', 'ebu': 'Kĩnjeremani', 'ee': 'Germaniagbe', 'el': 'Γερμανικά', 'el-polyton': 'Γερμανικά', 'en': 'German', 'en-Dsrt': '𐐖𐐲𐑉𐑋𐑌𐐲', 'eo': 'germana', 'es': 'alemán', 'et': 'saksa', 'eu': 'alemana', 'ewo': 'Ǹkɔ́bɔ ndzáman', 'fa': 'آلمانی', 'ff': 'Docceere', 'ff-Adlm': '𞤔𞤫𞤪𞤥𞤢𞤲𞤭𞤲𞤳𞤮𞥅𞤪𞤫', 'ff-Latn': 'Docceere', 'fi': 'saksa', 'fil': 'German', 'fo': 'týskt', 'fr': 'allemand', 'fur': 'todesc', 'fy': 'Dútsk', 'ga': 'Gearmáinis', 'gd': 'Gearmailtis', 'gl': 'alemán', 'gsw': 'Tüütsch', 'gu': 'જર્મન', 'guz': 'Kijerumani', 'ha': 'Jamusanci', 'haw': 'Kelemānia', 'he': 'גרמנית', 'hi': 'जर्मन', 'hi-Latn': 'German', 'hr': 'njemački', 'hsb': 'němčina', 'ht': 'allemand', 'hu': 'német', 'hy': 'գերմաներեն', 'ia': 'germano', 'id': 'Jerman', 'ie': 'german', 'ig': 'German', 'ii': 'ꄓꇩꉙ', 'is': 'þýska', 'it': 'tedesco', 'ja': 'ドイツ語', 'jgo': 'Njáman', 'jmc': 'Kyijerumani', 'jv': 'Jérman', 'ka': 'გერმანული', 'kaa': 'немецше', 'kaa-Cyrl': 'немецше', 'kab': 'Talmant', 'kam': 'Kijerumani', 'kde': 'Chidyelumani', 'kea': 'alemon', 'kgp': 'arimỹv', 'khq': 'Almaŋ senni', 'ki': 'Kĩnjeremani', 'kk': 'неміс тілі', 'kk-Cyrl': 'неміс тілі', 'kln': 'kutitab Chermani', 'km': 'អាល្លឺម៉ង់', 'kn': 'ಜರ್ಮನ್', 'ko': '독일어', 'kok': 'जर्मन', 'kok-Deva': 'जर्मन', 'ks': 'جٔرمَن', 'ks-Arab': 'جٔرمَن', 'ks-Deva': 'जर्मन', 'ksb': 'Kijeumani', 'ksf': 'ridjɛrman', 'ksh': 'Deutsch', 'ku': 'almanî', 'kxv': 'jerman', 'kxv-Deva': 'जर्मन', 'kxv-Latn': 'jerman', 'kxv-Orya': 'ଜର୍ମାନ୍', 'kxv-Telu': 'జర్మన్', 'ky': 'немисче', 'lag': 'Kɨjerʉmáani', 'lb': 'Däitsch', 'lg': 'Ludaaki', 'lkt': 'Iyášiča Iyápi', 'lld': 'todësch', 'ln': 'lialemá', 'lo': 'ເຢຍລະມັນ', 'lrc': 'آلمانی', 'lt': 'vokiečių', 'lu': 'Lizelumani', 'luo': 'Kijerumani', 'luy': 'Kijerumani', 'lv': 'vācu', 'mai': 'जर्मन', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ ljerumani', 'mer': 'Kĩnjamanĩ', 'mfe': 'alman', 'mg': 'Alemanina', 'mgh': 'Ijerimani', 'mi': 'Tiamana', 'mk': 'германски', 'ml': 'ജർമ്മൻ', 'mn': 'герман', 'mni': 'জর্মন', 'mni-Beng': 'জর্মন', 'mr': 'जर्मन', 'ms': 'Jerman', 'mt': 'Ġermaniż', 'mua': 'germaŋ', 'my': 'ဂျာမန်', 'mzn': 'آلمانی', 'naq': 'Duits', 'nb': 'tysk', 'nd': 'isi-Jalimani', 'ne': 'जर्मन', 'nl': 'Duits', 'nmg': 'Jáman', 'nn': 'tysk', 'nnh': 'nzǎmɔ̂ɔn', 'no': 'tysk', 'nqo': 'ߊߟߑߡߊ߲ߞߊ߲', 'nus': 'Thok jarmani', 'nyn': 'Orugirimaani', 'om': 'Afaan Jarmanii', 'or': 'ଜର୍ମାନ', 'os': 'немыцаг', 'pa': 'ਜਰਮਨ', 'pa-Guru': 'ਜਰਮਨ', 'pcm': 'Jámán Lángwej', 'pl': 'niemiecki', 'ps': 'الماني', 'pt': 'alemão', 'qu': 'Aleman Simi', 'rm': 'tudestg', 'rn': 'Ikidage', 'ro': 'germană', 'rof': 'Kijerumani', 'ru': 'немецкий', 'rw': 'Ikidage', 'rwk': 'Kyijerumani', 'sa': 'जर्मनभाषा:', 'sah': 'Ниэмэстии', 'saq': 'Kijerumani', 'sat': 'ᱡᱟᱨᱢᱟᱱ', 'sat-Olck': 'ᱡᱟᱨᱢᱟᱱ', 'sbp': 'Ishijelumani', 'sc': 'tedescu', 'sd': 'جرمن', 'sd-Arab': 'جرمن', 'sd-Deva': 'जर्मन', 'se': 'duiskkagiella', 'seh': 'alemão', 'ses': 'Almaŋ senni', 'sg': 'Zâmani', 'shi': 'ⵜⴰⵍⵉⵎⴰⵏⵜ', 'shi-Latn': 'Talimant', 'shi-Tfng': 'ⵜⴰⵍⵉⵎⴰⵏⵜ', 'si': 'ජර්මන්', 'sk': 'nemčina', 'sl': 'nemščina', 'smn': 'saksakielâ', 'sn': 'chiJerimani', 'so': 'Jarmal', 'sq': 'gjermanisht', 'sr': 'немачки', 'sr-Cyrl': 'немачки', 'sr-Latn': 'nemački', 'su': 'Jérman', 'su-Latn': 'Jérman', 'sv': 'tyska', 'sw': 'Kijerumani', 'syr': 'ܐܠܡܢܝܐ', 'ta': 'ஜெர்மன்', 'te': 'జర్మన్', 'teo': 'Kijerumani', 'tg': 'немисӣ', 'th': 'เยอรมัน', 'ti': 'ጀርመን', 'tk': 'nemes dili', 'to': 'lea fakasiamane', 'tpi': 'Jeman', 'tr': 'Almanca', 'tt': 'алман', 'twq': 'Almaŋ senni', 'tzm': 'Talmanit', 'ug': 'گېرمانچە', 'uk': 'німецька', 'ur': 'جرمن', 'uz': 'nemischa', 'uz-Cyrl': 'немисча', 'uz-Latn': 'nemischa', 'vai': 'ꕧꕮꔧ', 'vai-Latn': 'Jamáĩ', 'vai-Vaii': 'ꕧꕮꔧ', 'vec': 'todesco', 'vi': 'Tiếng Đức', 'vun': 'Kyijerumani', 'wae': 'Titš', 'wo': 'Almaa', 'xh': 'IsiJamani', 'xnr': 'जर्मन', 'xog': 'Oludaaki', 'yav': 'ŋndiáman', 'yi': 'דײַטש', 'yo': 'Èdè Jámánì', 'yrl': 'aremãu', 'yue': '德文', 'yue-Hans': '德文', 'yue-Hant': '德文', 'zgh': 'ⵜⴰⵍⵉⵎⴰⵏⵜ', 'zh': '德语', 'zh-Hans': '德语', 'zh-Hant': '德文', 'zu': 'isi-German'},
    'de-AT': {'ak': 'Ɔstria Gyaaman', 'am': 'የኦስትሪያ ጀርመንኛ', 'ar': 'الألمانية النمساوية', 'as': 'অষ্ট্ৰেলিয়ান জাৰ্মান', 'ast': 'alemán d’Austria', 'az': 'Avstriya almancası', 'az-Cyrl': 'Австрија алманҹасы', 'az-Latn': 'Avstriya almancası', 'be': 'аўстрыйская нямецкая', 'be-tarask': 'аўстрыйская нямецкая', 'bgn': 'استرالیاین جرمنی', 'br': 'alamaneg Aostria', 'brx': 'अस्ट्रेयाआरि जार्मान', 'ca': 'alemany austríac', 'ca-ES-valencia': 'alemany austríac', 'ccp': '𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄨𑄠𑄚𑄴 𑄎𑄢𑄴𑄟𑄚𑄴', 'ce': 'австрин немцойн', 'chr': 'ᎠᏟᏯᏂ ᎠᏂᏓᏥ', 'cv': 'австрин нимӗҫ', 'cy': 'Almaeneg Awstria', 'da': 'østrigsk tysk', 'de': 'Österreichisches Deutsch', 'doi': 'आस्ट्रियाई जर्मन', 'dsb': 'awstriska nimšćina', 'dz': 'ཨཱོས་ཊྲི་ཡཱན་ཇཱར་མཱན་ཁ', 'ee': 'Germaniagbe (Austria)', 'el': 'Γερμανικά Αυστρίας', 'el-polyton': 'Γερμανικά Αυστρίας', 'en': 'Austrian German', 'es': 'alemán austríaco', 'et': 'Austria saksa', 'eu': 'Austriako alemana', 'fa': 'آلمانی اتریش', 'ff-Adlm': '𞤔𞤫𞤪𞤥𞤢𞤲𞤭𞤲𞤳𞤮𞥅𞤪𞤫 𞤌𞤼𞤭𞤪𞤧𞤢', 'fi': 'itävallansaksa', 'fil': 'Austrian German', 'fr': 'allemand autrichien', 'fur': 'todesc de Austrie', 'fy': 'Eastenryks Dútsk', 'ga': 'Gearmáinis na hOstaire', 'gd': 'Gearmailtis na h-Ostaire', 'gl': 'alemán austríaco', 'gsw': 'Öschtriichischs Tüütsch', 'gu': 'ઓસ્ટ્રિઅન જર્મન', 'ha': 'Jamusanci Ostiriya', 'hi': 'ऑस्ट्रियाई जर्मन', 'hi-Latn': 'Austrian German', 'hr': 'austrijski njemački', 'hsb': 'awstriska němčina', 'ht': 'allemand autrichien', 'hu': 'osztrák német', 'hy': 'ավստրիական գերմաներեն', 'ia': 'germano austriac', 'ie': 'austrian german', 'ig': 'Austrian German', 'is': 'austurrísk þýska', 'it': 'tedesco austriaco', 'jv': 'Jérman Ostenrik', 'ka': 'ავსტრიული გერმანული', 'kaa': 'немецше (Австрия)', 'kaa-Cyrl': 'немецше (Австрия)', 'kea': 'alemon austríaku', 'kn': 'ಆಸ್ಟ್ರಿಯನ್ ಜರ್ಮನ್', 'kok': 'ऑस्ट्रियन जर्मन', 'kok-Deva': 'ऑस्ट्रियन जर्मन', 'ks': 'آسٹرِیَن جٔرمَن', 'ks-Arab': 'آسٹرِیَن جٔرمَن', 'ks-Deva': 'आस्ट्रियन जर्मन', 'ksh': 'Deutsch uß Öhßterich', 'kxv': 'ausṭriati jerman', 'kxv-Deva': 'ऑस्ट्रियाति जर्मन', 'kxv-Latn': 'ausṭriati jerman', 'kxv-Orya': 'ଅଷ୍ଟ୍ରିଆ ତି ଜର୍ମାନ', 'kxv-Telu': 'అస్ట్రీయన్ జర్మన్', 'lb': 'Éisträichescht Däitsch', 'lld': 'todësch dl’Austria', 'lo': 'ເຢຍລະມັນ (ໂອສຕຣິດ)', 'lrc': 'آلمانی ئوتریشی', 'lt': 'Austrijos vokiečių', 'mai': 'ऑस्ट्रियाई जर्मन', 'mi': 'Tiamana Ateriana', 'mk': 'австриски германски', 'ml': 'ഓസ്\u200cട്രിയൻ ജർമ്മൻ', 'mni': 'ওষ্ট্রিয়ান জর্মন', 'mni-Beng': 'ওষ্ট্রিয়ান জর্মন', 'mr': 'ऑस्ट्रियन जर्मन', 'ms': 'Jerman Austria', 'mt': 'Ġermaniż Awstrijak', 'my': 'ဩစတြီးယား ဂျာမန်', 'mzn': 'اتریش ِآلمانی', 'ne': 'अस्ट्रिएन जर्मन', 'nqo': 'ߏߕߑߙߌߛߌ߬ ߊߟߑߡߊߞߊ߲', 'om': 'Jarmanii Awustiriyaa', 'or': 'ଅଷ୍ଟ୍ରିଆନ୍ ଜର୍ମାନ', 'os': 'австралиаг немыцаг', 'pa': 'ਜਰਮਨ (ਆਸਟਰੀਆਈ)', 'pa-Guru': 'ਜਰਮਨ (ਆਸਟਰੀਆਈ)', 'pcm': 'Ọ́stria Jámán', 'pl': 'niemiecki austriacki', 'ps': 'اتريشي آلماني', 'rm': 'tudestg austriac', 'ru': 'австрийский немецкий', 'sa': 'ऑस्ट्रियाई जर्मनभाषा:', 'sat': 'ᱚᱥᱴᱨᱤᱭᱟ ᱡᱟᱨᱢᱟᱱ', 'sat-Olck': 'ᱚᱥᱴᱨᱤᱭᱟ ᱡᱟᱨᱢᱟᱱ', 'sc': 'tedescu austrìacu', 'sd': 'آسٽريائي جرمن', 'sd-Arab': 'آسٽريائي جرمن', 'sd-Deva': 'आसट्रियन जर्मन', 'si': 'ඔස්ට්\u200dරියානු ජර්මන්', 'sk': 'nemčina (rakúska)', 'sl': 'avstrijska nemščina', 'smn': 'Nuorttâriijkâ saksakielâ', 'so': 'Jarmal Awsteeriya', 'sq': 'gjermanishte austriake', 'sr': 'аустријски немачки', 'sr-Cyrl': 'аустријски немачки', 'sr-Latn': 'austrijski nemački', 'su': 'Jérman Austria', 'su-Latn': 'Jérman Austria', 'sv': 'österrikisk tyska', 'ta': 'ஆஸ்திரிய ஜெர்மன்', 'te': 'ఆస్ట్రియన్ జర్మన్', 'tg': 'немисии австриягӣ', 'th': 'เยอรมัน - ออสเตรีย', 'to': 'lea fakasiamane-ʻaositulia', 'tpi': 'Austria Jeman', 'tr': 'Avusturya Almancası', 'ug': 'ئاۋستىرىيە گېرمانچە', 'ur': 'آسٹریائی جرمن', 'uz': 'nemis (Avstriya)', 'uz-Latn': 'nemis (Avstriya)', 'vec': 'todesco austrìago', 'wae': 'Öštričišes Titš', 'wo': 'Almaa bu Ótiriis', 'xh': 'IsiJamani Sase-Austria', 'xnr': 'ऑस्ट्रिया आली जर्मन', 'yo': 'Èdè Jámánì (Ọ́síríà )', 'zh': '奥地利德语', 'zh-Hans': '奥地利德语', 'zu': 'isi-Austrian German'},
    'de-CH': {'ak': 'Swisalande Gyaaman', 'am': 'የስዊዝ ከፍተኛ ጀርመንኛ', 'as': 'ছুইচ হাই জাৰ্মান', 'ast': 'altualemán de Suiza', 'az': 'İsveçrə yüksək almancası', 'az-Cyrl': 'Исвечрә јүксәк алманҹасы', 'az-Latn': 'İsveçrə yüksək almancası', 'be': 'швейцарская літаратурная нямецкая', 'be-tarask': 'швейцарская літаратурная нямецкая', 'bgn': 'سویسین جرمنی', 'br': 'alamaneg uhel Suis', 'brx': 'सुइस गोजौ जार्मान', 'bs': 'visoki njemački (Švicarska)', 'bs-Cyrl': 'високи њемачки (Швицарска)', 'bs-Latn': 'visoki njemački (Švicarska)', 'ca': 'alemany estàndard suís', 'ca-ES-valencia': 'alemany estàndard suís', 'ccp': '𑄥𑄪𑄃𑄨𑄌𑄴 𑄦𑄭 𑄎𑄢𑄴𑄟𑄚𑄴', 'ce': 'швейцарин литературин немцойн', 'ceb': 'Swiss High German', 'chr': 'ᏍᏫᏏ ᎦᎸᎳᏗ ᎠᏂᏓᏥ', 'cs': 'němčina standardní (Švýcarsko)', 'cv': 'швейцарин нимӗҫ', 'cy': 'Almaeneg Safonol y Swistir', 'da': 'schweizerhøjtysk', 'de': 'Schweizer Hochdeutsch', 'doi': 'स्विस हाई जर्मन', 'dsb': 'šwicarska wusokonimšćina', 'dz': 'སུ་ཡིས་གི་མཐོ་སའི་ཇཱར་མཱན་ཁ', 'ee': 'Germaniagbe (Switzerland)', 'el': 'Υψηλά Γερμανικά Ελβετίας', 'el-polyton': 'Υψηλά Γερμανικά Ελβετίας', 'en': 'Swiss High German', 'es': 'alto alemán suizo', 'et': 'Šveitsi ülemsaksa', 'eu': 'Suitzako aleman garaia', 'fa': 'آلمانی معیار سوئیس', 'ff-Adlm': '𞤔𞤫𞤪𞤥𞤢𞤲𞤭𞤲𞤳𞤮𞥅𞤪𞤫 𞤅𞤵𞤱𞤭𞥅𞤧', 'fi': 'sveitsinyläsaksa', 'fil': 'Swiss High German', 'fo': 'høgt týskt (Sveis)', 'fr': 'allemand suisse', 'fur': 'alt todesc de Svuizare', 'fy': 'Switsersk Heechdútsk', 'ga': 'Ard-Ghearmáinis Eilvéiseach', 'gd': 'Àrd-Ghearmailtis na h-Eilbheise', 'gl': 'alto alemán suízo', 'gsw': 'Schwiizer Hochtüütsch', 'gu': 'સ્વિસ હાય જર્મન', 'ha': 'Jamusanci Suwizalan', 'hi': 'स्विस उच्च जर्मन', 'hi-Latn': 'Swiss High German', 'hr': 'gornjonjemački (švicarski)', 'hsb': 'šwicarska wysokoněmčina', 'ht': 'allemand suisse', 'hu': 'svájci felnémet', 'hy': 'շվեյցարական վերին գերմաներեն', 'ia': 'alte germano suisse', 'id': 'Jerman Tinggi (Swiss)', 'ie': 'sviss alt-german', 'ig': 'Swiss High German', 'is': 'svissnesk háþýska', 'it': 'alto tedesco svizzero', 'ja': '標準ドイツ語 (スイス)', 'jv': 'Jérman Switserlan', 'ka': 'შვეიცარიული ზემოგერმანული', 'kaa': 'немецше (Швейцария)', 'kaa-Cyrl': 'немецше (Швейцария)', 'kea': 'alemon altu suisu', 'kgp': 'suvisa arimỹv-téj', 'kk': 'швейцариялық әдеби неміс тілі', 'kk-Cyrl': 'швейцариялық әдеби неміс тілі', 'kn': 'ಸ್ವಿಸ್ ಹೈ ಜರ್ಮನ್', 'ko': '고지 독일어(스위스)', 'kok': 'स्विझ हाय जर्मन', 'kok-Deva': 'स्विझ हाय जर्मन', 'ks': 'سٕوِس ہائی جٔرمَن', 'ks-Arab': 'سٕوِس ہائی جٔرمَن', 'ks-Deva': 'स्विस हाई जर्मन', 'ksh': 'Deutsch uß de Schweijz', 'kxv': 'svis high jerman', 'kxv-Deva': 'स्विस हाइ ति जर्मन', 'kxv-Latn': 'svis high jerman', 'kxv-Orya': 'ସ୍ୱିସ୍\u200c ହାଇ ତି ଜର୍ମାନ', 'kxv-Telu': 'స్విస్ హఇ జర్మన్', 'ky': 'адабий немисче (Швейцария)', 'lb': 'Schwäizer Héichdäitsch', 'lld': 'todësch standard dla Svizera', 'lo': 'ສະວິສ ໄຮ ເຈີແມນ', 'lrc': 'آلمانی سوٙییسی', 'lt': 'Šveicarijos aukštutinė vokiečių', 'lv': 'augšvācu (Šveice)', 'mai': 'स्विस उच्च जर्मन', 'mi': 'Tiamana Ōkawa Huiterangi', 'mk': 'швајцарски горногермански', 'ml': 'സ്വിസ് ഹൈ ജർമ്മൻ', 'mn': 'герман (Швейцар)', 'mni': 'স্বিজ হায় জর্মন', 'mni-Beng': 'স্বিজ হায় জর্মন', 'mr': 'स्विस हाय जर्मन', 'ms': 'Jerman Halus Switzerland', 'mt': 'Ġermaniż Żvizzeru', 'my': 'အလီမဲန်နစ် ဂျာမန်', 'mzn': 'سوییس ِآلمانی', 'ne': 'स्वीस हाई जर्मन', 'nqo': 'ߛߥߌߛ ߊߟߑߡߊ߲ߞߊ߲', 'om': 'Jarmanii Siwiiz Haay', 'or': 'ସ୍ୱିସ୍\u200c ହାଇ ଜର୍ମାନ', 'os': 'швйецариаг немыцаг', 'pa': 'ਹਾਈ ਜਰਮਨ (ਸਵਿਟਜ਼ਰਲੈਂਡ)', 'pa-Guru': 'ਹਾਈ ਜਰਮਨ (ਸਵਿਟਜ਼ਰਲੈਂਡ)', 'pcm': 'Swítzaland Haí Jámán', 'pl': 'wysokoniemiecki szwajcarski', 'ps': 'سویس های جرمن', 'pt': 'alto alemão (Suíça)', 'rm': 'tudestg da scrittira svizzer', 'ro': 'germană standard (Elveția)', 'ru': 'литературный швейцарский немецкий', 'sa': 'स्विस उच्च जर्मनभाषा:', 'sat': 'ᱥᱩᱣᱤᱥ ᱦᱟᱤ ᱡᱟᱨᱢᱟᱱ', 'sat-Olck': 'ᱥᱩᱣᱤᱥ ᱦᱟᱤ ᱡᱟᱨᱢᱟᱱ', 'sc': 'tedescu artu isvìtzeru', 'sd': 'سوئس هائي جرمن', 'sd-Arab': 'سوئس هائي جرمن', 'sd-Deva': 'स्विस हाई जर्मन', 'si': 'ස්විස් උසස් ජර්මන්', 'sk': 'nemčina (švajčiarska spisovná)', 'sl': 'visoka nemščina (Švica)', 'smn': 'Sveitsi pajesaksakielâ', 'so': 'Iswiiska Sare ee Jarmal', 'sq': 'gjermanishte zvicerane (dialekti i Alpeve)', 'sr': 'швајцарски високи немачки', 'sr-Cyrl': 'швајцарски високи немачки', 'sr-Latn': 'švajcarski visoki nemački', 'su': 'Jérman Swiss Luhur', 'su-Latn': 'Jérman Swiss Luhur', 'sv': 'schweizisk högtyska', 'ta': 'ஸ்விஸ் ஹை ஜெர்மன்', 'te': 'స్విస్ హై జర్మన్', 'tg': 'немисии швейсарии болоӣ', 'th': 'เยอรมันสูง (สวิส)', 'tk': 'ýokarky nemes dili (Şweýsariýa)', 'to': 'lea fakasiamane-hake-suisilani', 'tpi': 'Swis Jeman', 'tr': 'İsviçre Yüksek Almancası', 'tt': 'югары алман (Швейцария)', 'ug': 'شىۋىتسارىيە ئېگىزلىك گېرمانچە', 'uk': 'верхньонімецька (Швейцарія)', 'ur': 'سوئس ہائی جرمن', 'uz': 'yuqori nemis (Shveytsariya)', 'uz-Latn': 'yuqori nemis (Shveytsariya)', 'vec': 'alto todesco zvìsaro', 'vi': 'Tiếng Thượng Giéc-man (Thụy Sĩ)', 'wae': 'Schwizer Hočtitš', 'wo': 'Almaa bu Kawe bu Swis', 'xh': 'IsiJamani Esiyi-High Swiss', 'xnr': 'स्विस उच्च जर्मन', 'yo': 'Èdè Ilẹ̀ Jámánì (Orílẹ́ède swítsàlandì)', 'yrl': 'aremãu iwaté (Suisa)', 'yue': '高地德文 (瑞士)', 'yue-Hans': '高地德文（瑞士）', 'yue-Hant': '高地德文 (瑞士)', 'zh': '瑞士高地德语', 'zh-Hans': '瑞士高地德语', 'zh-Hant': '高地德文（瑞士）', 'zu': 'Isi-Swiss High German'},
    'dec': {'en': 'Dagik'},
    'ded': {'en': 'Dedua'},
    'dee': {'en': 'Dewoin'},
    'def': {'en': 'Dezfuli'},
    'deg': {'en': 'Degema'},
    'deh': {'en': 'Dehwari'},
    'dei': {'en': 'Demisa'},
    'dek': {'en': 'Dek'},
    'del': {'am': 'ዳላዌር', 'ar': 'الديلوير', 'ast': 'delaware', 'az': 'delaver', 'az-Latn': 'delaver', 'bg': 'делауер', 'bn': 'ডেলাওয়ের', 'br': 'delaware', 'brx': 'दलावार्', 'bs': 'delaver', 'bs-Cyrl': 'делавер', 'bs-Latn': 'delaver', 'ca': 'delaware', 'ca-ES-valencia': 'delaware', 'ccp': '𑄓𑄬𑄣𑄤𑄬𑄢𑄴', 'cs': 'delawarština', 'da': 'delaware', 'de': 'Delaware', 'el': 'Ντέλαγουερ', 'el-polyton': 'Ντέλαγουερ', 'en': 'Delaware', 'es': 'delaware', 'et': 'delavari', 'fa': 'دلاواری', 'fi': 'delaware', 'fr': 'delaware', 'fy': 'Delaware', 'gd': 'Delaware', 'gsw': 'Delaware-Schpraach', 'gu': 'દેલવેર', 'he': 'דלאוור', 'hi': 'डिलैवेयर', 'hi-Latn': 'Delaware', 'hr': 'delavarski', 'ht': 'delaware', 'hu': 'delavár', 'id': 'Delaware', 'is': 'delaver', 'it': 'delaware', 'ja': 'デラウェア語', 'ka': 'დელავერული', 'kgp': 'neravare', 'kn': 'ಡೆಲಾವೇರ್', 'ko': '델라웨어어', 'ks': 'ڈیلوییَر', 'ks-Arab': 'ڈیلوییَر', 'lb': 'Delaware-Sprooch', 'lo': 'ເດລາວາ', 'lt': 'delavero', 'lv': 'delavēru', 'mk': 'делавер', 'ml': 'ദെലവേർ', 'mr': 'डेलावेयर', 'mt': 'Delawerjan', 'my': 'ဒယ်လာဝဲလ်', 'nb': 'delaware', 'ne': 'देलावर', 'nl': 'Delaware', 'nn': 'delaware', 'no': 'delaware', 'or': 'ଡେଲାୱେର୍', 'pl': 'delaware', 'pt': 'delaware', 'rm': 'delaware', 'ro': 'delaware', 'ru': 'делаварский', 'sk': 'delawarčina', 'sl': 'delavarščina', 'sr': 'делаверски', 'sr-Cyrl': 'делаверски', 'sr-Latn': 'delaverski', 'sv': 'delaware', 'ta': 'டெலாவர்', 'te': 'డెలావేర్', 'th': 'เดลาแวร์', 'to': 'lea fakatelauale', 'tr': 'Delaware', 'ug': 'دېلاۋارېچە', 'uk': 'делаварська', 'vi': 'Tiếng Delaware', 'yrl': 'deraware', 'yue': '德拉瓦文', 'yue-Hans': '德拉瓦文', 'yue-Hant': '德拉瓦文', 'zh': '特拉华语', 'zh-Hans': '特拉华语', 'zh-Hant': '德拉瓦文'},
    'dem': {'en': 'Dem'},
    'den': {'ar': 'السلافية', 'ast': 'slave', 'az': 'slavey', 'az-Latn': 'slavey', 'bg': 'слейви', 'bn': 'স্ল্যাভ', 'brx': 'स्लेव्', 'bs': 'slave', 'bs-Cyrl': 'славски', 'bs-Latn': 'slave', 'ca': 'slavi', 'ca-ES-valencia': 'slavi', 'ccp': '𑄥𑄳𑄣𑄳𑄠𑄞𑄴', 'cs': 'slejvština (athabaský jazyk)', 'da': 'athapaskisk', 'de': 'Slave', 'el': 'Σλαβικά', 'el-polyton': 'Σλαβικά', 'en': 'Slave', 'es': 'slave', 'et': 'sleivi', 'fi': 'slevi', 'fr': 'esclave', 'fur': 'sclâf', 'fy': 'Slave', 'gd': 'Slavey', 'gsw': 'Slavey', 'gu': 'સ્લેવ', 'he': 'סלאבית', 'hi': 'स्लेव', 'hi-Latn': 'Slave', 'hr': 'slave', 'ht': 'esclave', 'hu': 'szlevi', 'id': 'Slave', 'is': 'slavneska', 'it': 'slave', 'ja': 'スレイビー語', 'ka': 'სლეივი', 'kgp': 'sirave', 'kn': 'ಸ್ಲೇವ್', 'ko': '슬라브어', 'ks': 'سلیو', 'ks-Arab': 'سلیو', 'lb': 'Slave', 'lo': 'ຊີເລັບ', 'lt': 'slave', 'lv': 'sleivu', 'mk': 'слејви', 'ml': 'സ്ലേവ്', 'mr': 'स्लाव्ह', 'mt': 'Slav', 'nb': 'slavey', 'nl': 'Slavey', 'nn': 'slavej', 'no': 'slavey', 'or': 'ସ୍ଲେଭ୍', 'pl': 'slave', 'pt': 'slave', 'rm': 'slavey', 'ro': 'slave', 'ru': 'слейви', 'sk': 'slavé', 'sl': 'slavejščina', 'sr': 'слејви', 'sr-Cyrl': 'слејви', 'sr-Latn': 'slejvi', 'sv': 'slavej', 'ta': 'ஸ்லாவ்', 'te': 'స్లేవ్', 'th': 'สเลวี', 'to': 'lea fakasilave', 'tr': 'Slavey dili', 'ug': 'سلاۋچە', 'uk': 'слейв', 'vi': 'Tiếng Slave', 'yrl': 'isireivei', 'yue': '斯拉夫', 'yue-Hans': '斯拉夫', 'yue-Hant': '斯拉夫', 'zh': '史拉维语', 'zh-Hans': '史拉维语', 'zh-Hant': '斯拉夫'},
    'dep': {'en': 'Pidgin Delaware'},
    'deq': {'en': 'Dendi (Central African Republic)'},
    'der': {'en': 'Deori'},
    'des': {'en': 'Desano'},
    'dev': {'en': 'Domung'},
    'dez': {'en': 'Dengese'},
    'dga': {'en': 'Southern Dagaare'},
    'dgb': {'en': 'Bunoge Dogon'},
    'dgc': {'en': 'Casiguran Dumagat Agta'},
    'dgd': {'en': 'Dagaari Dioula'},
    'dge': {'en': 'Degenan'},
    'dgg': {'en': 'Doga'},
    'dgh': {'en': 'Dghwede'},
    'dgi': {'en': 'Northern Dagara'},
    'dgk': {'en': 'Dagba'},
    'dgl': {'en': 'Andaandi'},
    'dgn': {'en': 'Dagoman'},
    'dgo': {'en': 'Dogri (individual language)'},
    'dgr': {'af': 'Dogrib', 'am': 'ዶግሪብ', 'ar': 'الدوجريب', 'as': 'ডোগ্ৰিব', 'ast': 'dogrib', 'az': 'doqrib', 'az-Cyrl': 'догриб', 'az-Latn': 'doqrib', 'be': 'догрыб', 'be-tarask': 'догрыб', 'bg': 'догриб', 'bgn': 'داگریبی', 'bn': 'দোগ্রীব', 'br': 'dogrib', 'brx': 'डोगरीब', 'bs': 'dogrib', 'bs-Cyrl': 'догриб', 'bs-Latn': 'dogrib', 'ca': 'dogrib', 'ca-ES-valencia': 'dogrib', 'ccp': '𑄘𑄮𑄉𑄳𑄢𑄨𑄝𑄴', 'ce': 'догриб', 'chr': 'ᎩᏟ ᎤᏄᎳᏥ', 'ckb': 'دۆگریب', 'cs': 'dogrib', 'cy': 'Dogrib', 'da': 'dogrib', 'de': 'Dogrib', 'dsb': 'dogribšćina', 'el': 'Ντόγκριμπ', 'el-polyton': 'Ντόγκριμπ', 'en': 'Dogrib', 'es': 'dogrib', 'et': 'dogribi', 'eu': 'dogribera', 'fa': 'دوگریب', 'ff-Adlm': '𞤁𞤮𞤺𞤪𞤭𞤦𞤪𞤫', 'fi': 'dogrib', 'fil': 'Dogrib', 'fo': 'dogrib', 'fr': 'dogrib', 'fy': 'Dogrib', 'ga': 'Dograibis', 'gd': 'Dogrib', 'gl': 'dogrib', 'gsw': 'Togrib', 'gu': 'ડોગ્રિબ', 'ha': 'Dogrib', 'he': 'דוגריב', 'hi': 'डोग्रिब', 'hi-Latn': 'Dogrib', 'hr': 'dogrib', 'hsb': 'dogribšćina', 'ht': 'dogrib', 'hu': 'dogrib', 'hy': 'դոգրիբ', 'ia': 'dogrib', 'id': 'Dogrib', 'ig': 'Dogrib', 'is': 'dogríb', 'it': 'dogrib', 'ja': 'ドグリブ語', 'jv': 'Dogrib', 'ka': 'დოგრიბი', 'kgp': 'nogrim', 'kk': 'догриб тілі', 'kk-Cyrl': 'догриб тілі', 'km': 'ដូគ្រីប', 'kn': 'ಡೋಗ್ರಿಬ್', 'ko': '도그리브어', 'kok': 'डोगरीब', 'kok-Deva': 'डोगरीब', 'ks': 'ڈاگرِب', 'ks-Arab': 'ڈاگرِب', 'ksh': 'Dohjribesch', 'ku': 'dogrîbî', 'ky': 'догрибче', 'lb': 'Dogrib', 'lo': 'ໂດກຣິບ', 'lt': 'dogribų', 'lv': 'dogribu', 'mai': 'डोग्रीब', 'mi': 'Tōkiripi', 'mk': 'догрипски', 'ml': 'ഡോഗ്രിബ്', 'mn': 'догриб', 'mr': 'डोग्रिब', 'ms': 'Dogrib', 'mt': 'Dogrib', 'my': 'ဒေါ့ဂ်ရစ်ဘ်', 'nb': 'dogrib', 'ne': 'दोग्रिब', 'nl': 'Dogrib', 'nn': 'dogrib', 'no': 'dogrib', 'or': 'ଡୋଗ୍ରିବ୍', 'pa': 'ਡੋਗਰਿੱਬ', 'pa-Guru': 'ਡੋਗਰਿੱਬ', 'pcm': 'Dọgríb Lángwej', 'pl': 'dogrib', 'ps': 'داگرب', 'pt': 'dogrib', 'qu': 'Dogrib Simi', 'rm': 'dogrib', 'ro': 'dogrib', 'ru': 'догриб', 'sc': 'dogrib', 'sd': 'داگرب', 'sd-Arab': 'داگرب', 'si': 'ඩොග්\u200dරිබ්', 'sk': 'dogribčina', 'sl': 'dogrib', 'smn': 'dogribkielâ', 'so': 'Dogrib', 'sq': 'dogribisht', 'sr': 'догрипски', 'sr-Cyrl': 'догрипски', 'sr-Latn': 'dogripski', 'sv': 'dogrib', 'sw': 'Kidogrib', 'ta': 'டோக்ரிப்', 'te': 'డోగ్రిబ్', 'th': 'โดกริบ', 'ti': 'ዶግሪብ', 'tk': 'dogrib dili', 'to': 'lea fakatōkelipi', 'tr': 'Dogrib', 'ug': 'دوگرىبچە', 'uk': 'догрибська', 'ur': 'دوگریب', 'uz': 'dogrib', 'uz-Cyrl': 'догриб', 'uz-Latn': 'dogrib', 'vec': 'dogrib', 'vi': 'Tiếng Dogrib', 'xnr': 'डोग्रिब', 'yo': 'Èdè Dogribu', 'yrl': 'dogiri', 'yue': '多格里布文', 'yue-Hans': '多格里布文', 'yue-Hant': '多格里布文', 'zh': '多格里布语', 'zh-Hans': '多格里布语', 'zh-Hant': '多格里布文', 'zu': 'isi-Dogrib'},
    'dgs': {'en': 'Dogoso'},
    'dgt': {'en': "Ndra'ngith"},
    'dgu': {'en': 'Degaru'},
    'dgw': {'en': 'Daungwurrung'},
    'dgx': {'en': 'Doghoro'},
    'dgz': {'en': 'Daga'},
    'dha': {'en': 'Dhanwar (India)'},
    'dhd': {'en': 'Dhundari'},
    'dhg': {'en': 'Dhangu-Djangu'},
    'dhi': {'en': 'Dhimal'},
    'dhl': {'en': 'Dhalandji'},
    'dhm': {'en': 'Zemba'},
    'dhn': {'en': 'Dhanki'},
    'dho': {'en': 'Dhodia'},
    'dhr': {'en': 'Dhargari'},
    'dhs': {'en': 'Dhaiso'},
    'dhu': {'en': 'Dhurga'},
    'dhv': {'en': 'Dehu'},
    'dhw': {'en': 'Dhanwar (Nepal)'},
    'dhx': {'en': 'Dhungaloo'},
    'dia': {'en': 'Dia'},
    'dib': {'en': 'South Central Dinka'},
    'dic': {'en': 'Lakota Dida'},
    'did': {'en': 'Didinga'},
    'dif': {'en': 'Dieri'},
    'dig': {'en': 'Digo'},
    'dih': {'en': 'Kumiai'},
    'dii': {'en': 'Dimbong'},
    'dij': {'en': 'Dai'},
    'dik': {'en': 'Southwestern Dinka'},
    'dil': {'en': 'Dilling'},
    'dim': {'en': 'Dime'},
    'din': {'am': 'ዲንካ', 'ar': 'الدنكا', 'ast': 'dinka', 'az': 'dinka', 'az-Latn': 'dinka', 'bg': 'динка', 'bn': 'ডিংকা', 'br': 'dinka', 'brx': 'डींगका', 'bs': 'dinka', 'bs-Cyrl': 'динка', 'bs-Latn': 'dinka', 'ca': 'dinka', 'ca-ES-valencia': 'dinka', 'ccp': '𑄓𑄨𑄁𑄇', 'cs': 'dinkština', 'cy': 'Dinca', 'da': 'dinka', 'de': 'Dinka', 'el': 'Ντίνκα', 'el-polyton': 'Ντίνκα', 'en': 'Dinka', 'es': 'dinka', 'et': 'dinka', 'fa': 'دینکایی', 'fi': 'dinka', 'fr': 'dinka', 'fy': 'Dinka', 'gd': 'Dinka', 'gsw': 'Tinka', 'gu': 'દિન્કા', 'he': 'דינקה', 'hi': 'दिन्का', 'hi-Latn': 'Dinka', 'hr': 'dinka', 'ht': 'dinka', 'hu': 'dinka', 'id': 'Dinka', 'is': 'dinka', 'it': 'dinca', 'ja': 'ディンカ語', 'ka': 'დინკა', 'kgp': 'ninka', 'kn': 'ಡಿಂಕಾ', 'ko': '딩카어', 'ks': 'ڈِنکا', 'ks-Arab': 'ڈِنکا', 'lb': 'Dinka-Sprooch', 'lo': 'ດິນກາ', 'lt': 'dinkų', 'lv': 'dinku', 'mk': 'динка', 'ml': 'ദിൻക', 'mr': 'डिन्का', 'mt': 'Dinka', 'nb': 'dinka', 'ne': 'दिन्का', 'nl': 'Dinka', 'nn': 'dinka', 'no': 'dinka', 'or': 'ଦିଙ୍କା', 'pl': 'dinka', 'pt': 'dinka', 'rm': 'dinka', 'ro': 'dinka', 'ru': 'динка', 'sk': 'dinkčina', 'sl': 'dinka', 'sr': 'динка', 'sr-Cyrl': 'динка', 'sr-Latn': 'dinka', 'sv': 'dinka', 'ta': 'டின்கா', 'te': 'డింకా', 'th': 'ดิงกา', 'to': 'lea fakatingikā', 'tr': 'Dinka dili', 'ug': 'دىنكاچە', 'uk': 'дінка', 'vi': 'Tiếng Dinka', 'yrl': 'dĩka', 'yue': '丁卡文', 'yue-Hans': '丁卡文', 'yue-Hant': '丁卡文', 'zh': '丁卡语', 'zh-Hans': '丁卡语', 'zh-Hant': '丁卡文'},
    'dio': {'en': 'Dibo'},
    'dip': {'en': 'Northeastern Dinka'},
    'diq': {'en': 'Dimli (individual language)'},
    'dir': {'en': 'Dirim'},
    'dis': {'en': 'Dimasa'},
    'dit': {'en': 'Dirari'},
    'diu': {'en': 'Diriku'},
    'diw': {'en': 'Northwestern Dinka'},
    'dix': {'en': 'Dixon Reef'},
    'diy': {'en': 'Diuwe'},
    'diz': {'en': 'Ding'},
    'dja': {'en': 'Djadjawurrung'},
    'djb': {'en': 'Djinba'},
    'djc': {'en': 'Dar Daju Daju'},
    'djd': {'en': 'Djamindjung'},
    'dje': {'af': 'Zarma', 'am': 'ዛርማኛ', 'ar': 'الزارمية', 'as': 'ঝাৰ্মা', 'ast': 'zarma', 'az': 'zarma', 'az-Cyrl': 'зарма', 'az-Latn': 'zarma', 'bal-Latn': 'Zarmah', 'be': 'зарма', 'be-tarask': 'зарма', 'bg': 'зарма', 'bgn': 'زرمی', 'bn': 'জার্মা', 'br': 'zarma', 'brx': 'जार्मा', 'bs': 'zarma', 'bs-Cyrl': 'зарма', 'bs-Latn': 'zarma', 'ca': 'zarma', 'ca-ES-valencia': 'zarma', 'ccp': '𑄎𑄢𑄴𑄟', 'ce': 'зарма', 'chr': 'ᏌᎹ', 'ckb': 'زارما', 'cs': 'zarmština', 'cy': 'Sarmaeg', 'da': 'zarma', 'de': 'Zarma', 'dje': 'Zarmaciine', 'dsb': 'zarma', 'el': 'Ζάρμα', 'el-polyton': 'Ζάρμα', 'en': 'Zarma', 'es': 'zarma', 'et': 'zarma', 'eu': 'zarma', 'fa': 'زرما', 'ff-Adlm': '𞤔𞤢𞤪𞤥𞤢𞥄𞤪𞤫', 'fi': 'djerma', 'fil': 'Zarma', 'fo': 'sarma', 'fr': 'zarma', 'fy': 'Zarma', 'ga': 'Zarmais', 'gd': 'Zarma', 'gl': 'zarma', 'gu': 'ઝર્મા', 'ha': 'Zarma', 'he': 'זארמה', 'hi': 'झार्मा', 'hi-Latn': 'Zarma', 'hr': 'zarma', 'hsb': 'zarma', 'ht': 'zarma', 'hu': 'zarma', 'hy': 'զարմա', 'ia': 'zarma', 'id': 'Zarma', 'ig': 'Zarma', 'is': 'zarma', 'it': 'zarma', 'ja': 'ザルマ語', 'jv': 'Zarma', 'ka': 'ზარმა', 'kea': 'zarma', 'kgp': 'jarma', 'kk': 'зарма тілі', 'kk-Cyrl': 'зарма тілі', 'km': 'ហ្សាម៉ា', 'kn': 'ಜರ್ಮಾ', 'ko': '자르마어', 'kok': 'झर्मा', 'kok-Deva': 'झर्मा', 'ksh': 'Djerma', 'ku': 'zarma', 'ky': 'зармача', 'lb': 'Zarma', 'lo': 'ຊາມາ', 'lrc': 'زارما', 'lt': 'zarmų', 'lv': 'zarmu', 'mai': 'जरमा', 'mi': 'Tāma', 'mk': 'зарма', 'ml': 'സാർമ്മ', 'mn': 'зарма', 'mr': 'झार्मा', 'ms': 'Zarma', 'mt': 'Zarma', 'my': 'ဇာမာ', 'mzn': 'زارمایی', 'nb': 'zarma', 'ne': 'जर्मा', 'nl': 'Zarma', 'nn': 'zarma', 'no': 'zarma', 'nqo': 'ߖ߭ߍߙߑߡߊߞߊ߲', 'or': 'ଜର୍ମା', 'pa': 'ਜ਼ਾਰਮਾ', 'pa-Guru': 'ਜ਼ਾਰਮਾ', 'pcm': 'Zármá', 'pl': 'dżerma', 'ps': 'زرما', 'pt': 'zarma', 'qu': 'Zarma Simi', 'rm': 'zarma', 'ro': 'zarma', 'ru': 'джерма', 'sc': 'zarma', 'sd': 'زارما', 'sd-Arab': 'زارما', 'si': 'සර්මා', 'sk': 'zarma', 'sl': 'zarmajščina', 'smn': 'zarmakielâ', 'so': 'Sarma', 'sq': 'zarmaisht', 'sr': 'зарма', 'sr-Cyrl': 'зарма', 'sr-Latn': 'zarma', 'sv': 'zarma', 'sw': 'Kizarma', 'ta': 'ஸார்மா', 'te': 'జార్మా', 'th': 'ซาร์มา', 'ti': 'ዛርማ', 'tk': 'zarma dili', 'to': 'lea fakatisāma', 'tr': 'Zarma', 'ug': 'زارماچە', 'uk': 'джерма', 'ur': 'زرما', 'uz': 'zarma', 'uz-Cyrl': 'зарма', 'uz-Latn': 'zarma', 'vec': 'zarma', 'vi': 'Tiếng Zarma', 'xnr': 'झार्मा', 'yo': 'Ṣárúmà', 'yrl': 'sarima', 'yue': '扎爾馬文', 'yue-Hans': '扎尔马文', 'yue-Hant': '扎爾馬文', 'zh': '哲尔马语', 'zh-Hans': '哲尔马语', 'zh-Hant': '扎爾馬文', 'zu': 'isi-Zarma'},
    'djf': {'en': 'Djangun'},
    'dji': {'en': 'Djinang'},
    'djj': {'en': 'Djeebbana'},
    'djk': {'en': 'Eastern Maroon Creole'},
    'djl': {'en': 'Djiwarli'},
    'djm': {'en': 'Jamsay Dogon'},
    'djn': {'en': 'Jawoyn'},
    'djo': {'en': 'Jangkang'},
    'djr': {'en': 'Djambarrpuyngu'},
    'dju': {'en': 'Kapriman'},
    'djw': {'en': 'Djawi'},
    'dka': {'en': 'Dakpakha'},
    'dkg': {'en': 'Kadung'},
    'dkk': {'en': 'Dakka'},
    'dkl': {'en': 'Kolum So Dogon'},
    'dkr': {'en': 'Kuijau'},
    'dks': {'en': 'Southeastern Dinka'},
    'dkx': {'en': 'Mazagway'},
    'dlc': {'en': 'Elfdalian'},
    'dlg': {'en': 'Dolgan'},
    'dlk': {'en': 'Dahalik'},
    'dlm': {'en': 'Dalmatian'},
    'dln': {'en': 'Darlong'},
    'dma': {'en': 'Duma'},
    'dmb': {'en': 'Mombo Dogon'},
    'dmc': {'en': 'Gavak'},
    'dmd': {'en': 'Madhi Madhi'},
    'dme': {'en': 'Dugwor'},
    'dmf': {'en': 'Medefaidrin'},
    'dmg': {'en': 'Upper Kinabatangan'},
    'dmk': {'en': 'Domaaki'},
    'dml': {'en': 'Dameli'},
    'dmm': {'en': 'Dama'},
    'dmn': {'en': 'Mande languages'},
    'dmo': {'en': 'Kemedzung'},
    'dmr': {'en': 'East Damar'},
    'dms': {'en': 'Dampelas'},
    'dmu': {'en': 'Dubu'},
    'dmv': {'en': 'Dumpas'},
    'dmw': {'en': 'Mudburra'},
    'dmx': {'en': 'Dema'},
    'dmy': {'en': 'Demta'},
    'dna': {'en': 'Upper Grand Valley Dani'},
    'dnd': {'en': 'Daonda'},
    'dne': {'en': 'Ndendeule'},
    'dng': {'en': 'Dungan'},
    'dni': {'en': 'Lower Grand Valley Dani'},
    'dnj': {'en': 'Dan'},
    'dnk': {'en': 'Dengka'},
    'dnn': {'en': 'Dzùùngoo'},
    'dno': {'en': 'Ndrulo'},
    'dnr': {'en': 'Danaru'},
    'dnt': {'en': 'Mid Grand Valley Dani'},
    'dnu': {'en': 'Danau'},
    'dnv': {'en': 'Danu'},
    'dnw': {'en': 'Western Dani'},
    'dny': {'en': 'Dení'},
    'doa': {'en': 'Dom'},
    'dob': {'en': 'Dobu'},
    'doc': {'en': 'Northern Dong'},
    'doe': {'en': 'Doe'},
    'dof': {'en': 'Domu'},
    'doh': {'en': 'Dong'},
    'doi': {'af': 'Dogri', 'ak': 'Dɔgri', 'am': 'ዶግሪ', 'ar': 'الدوجرية', 'as': 'ডোগ্ৰী', 'ast': 'dogri', 'az': 'doqri', 'az-Latn': 'doqri', 'bal-Latn': 'Dogri', 'be': 'догры', 'be-tarask': 'догры', 'bg': 'догри', 'bn': 'ডোগরি', 'br': 'dogri', 'brx': 'डोगरी', 'bs': 'dogri', 'bs-Cyrl': 'догри', 'bs-Latn': 'dogri', 'ca': 'dogri', 'ca-ES-valencia': 'dogri', 'ccp': '𑄓𑄮𑄉𑄧𑄢𑄨', 'chr': 'ᏙᎦᎵ', 'cs': 'dogarština', 'cy': 'Dogri', 'da': 'dogri', 'de': 'Dogri', 'doi': 'डोगरी', 'dsb': 'dogrišćina', 'el': 'Ντόγκρι', 'el-polyton': 'Ντόγκρι', 'en': 'Dogri', 'es': 'dogri', 'et': 'dogri', 'eu': 'dogria', 'fa': 'دوگری', 'ff-Adlm': '𞤁𞤮𞤺𞤪𞤭𞥅𞤪𞤫', 'fi': 'dogri', 'fil': 'Dogri', 'fo': 'dogri', 'fr': 'dogri', 'fy': 'Dogri', 'ga': 'Dóigris', 'gd': 'Dogri', 'gl': 'dogri', 'gsw': 'Togri', 'gu': 'ડોગ્રી', 'ha': 'Harshen Dogri', 'he': 'דוגרי', 'hi': 'डोगरी', 'hi-Latn': 'Dogri', 'hr': 'dogri', 'hsb': 'dogrišćina', 'ht': 'dogri', 'hu': 'dogri', 'hy': 'դոգրի', 'ia': 'dogri', 'id': 'Dogri', 'ig': 'Dogri', 'is': 'dogrí', 'it': 'dogri', 'ja': 'ドーグリー語', 'jv': 'Dogri', 'ka': 'დოგრი', 'kgp': 'nogri', 'kk': 'догри тілі', 'kk-Cyrl': 'догри тілі', 'km': 'ដូហ្គ្រី', 'kn': 'ಡೋಗ್ರಿ', 'ko': '도그리어', 'kok': 'डोग्री', 'kok-Deva': 'डोग्री', 'ks': 'ڈوگری', 'ks-Arab': 'ڈوگری', 'ku': 'dogrîyî', 'kxv': 'ḍogri', 'kxv-Deva': 'डोगरी', 'kxv-Latn': 'ḍogri', 'kxv-Orya': 'ଡଗ୍ରୀ', 'kxv-Telu': 'డోగ్రి', 'ky': 'догриче', 'lb': 'Dogri', 'lkt': 'Dogri Iyápi', 'lo': 'ດອກຣີ', 'lt': 'dogri', 'lv': 'dogru', 'mai': 'डोगरी', 'mi': 'Tōkiri', 'mk': 'догри', 'ml': 'ഡോഗ്രി', 'mn': 'догри', 'mr': 'डोगरी', 'ms': 'Dogri', 'mt': 'Dogri', 'my': 'ဒိုဂရီ', 'nb': 'dogri', 'ne': 'डोगरी', 'nl': 'Dogri', 'nn': 'dogri', 'no': 'dogri', 'nqo': 'ߘߏߜ߭ߙߌߞߊ߲', 'om': 'Dogri', 'or': 'ଡୋଗ୍ରୀ', 'pa': 'ਡੋਗਰੀ', 'pa-Guru': 'ਡੋਗਰੀ', 'pcm': 'Dọgri', 'pl': 'dogri', 'ps': 'ډوګري', 'pt': 'dogri', 'qu': 'Dogri Simi', 'rm': 'dogri', 'ro': 'dogri', 'ru': 'догри', 'sc': 'dogri', 'sd': 'ڊوگري', 'sd-Arab': 'ڊوگري', 'si': 'ඩොග්රි', 'sk': 'dógrí', 'sl': 'dogri', 'so': 'Dogri', 'sq': 'dogrisht', 'sr': 'догри', 'sr-Cyrl': 'догри', 'sr-Latn': 'dogri', 'sv': 'dogri', 'sw': 'Kidogri', 'ta': 'டோக்ரி', 'te': 'డోగ్రి', 'th': 'โฑครี', 'ti': 'ዶግሪ', 'tk': 'dogri', 'to': 'lea fakatokili', 'tr': 'Dogri', 'ug': 'دوگرىچە', 'uk': 'догрі', 'ur': 'ڈوگری', 'uz': 'dogri', 'uz-Latn': 'dogri', 'vec': 'dogri', 'vi': 'Tiếng Dogri', 'xnr': 'डोगरी', 'yo': 'Èdè Dogiri', 'yrl': 'dogiribi', 'yue': '多格來文', 'yue-Hans': '多格来文', 'yue-Hant': '多格來文', 'zh': '多格拉语', 'zh-Hans': '多格拉语', 'zh-Hant': '多格來文', 'zu': 'isi-Dogri'},
    'dok': {'en': 'Dondo'},
    'dol': {'en': 'Doso'},
    'don': {'en': 'Toura (Papua New Guinea)'},
    'doo': {'en': 'Dongo'},
    'dop': {'en': 'Lukpa'},
    'doq': {'en': 'Dominican Sign Language'},
    'dor': {'en': "Dori'o"},
    'dos': {'en': 'Dogosé'},
    'dot': {'en': 'Dass'},
    'dov': {'en': 'Dombe'},
    'dow': {'en': 'Doyayo'},
    'dox': {'en': 'Bussa'},
    'doy': {'en': 'Dompo'},
    'doz': {'en': 'Dorze'},
    'dpp': {'en': 'Papar'},
    'dra': {'en': 'Dravidian languages'},
    'dra-pro': {'en': 'Proto-Dravidian'},
    'drb': {'en': 'Dair'},
    'drc': {'en': 'Minderico'},
    'drd': {'en': 'Darmiya'},
    'dre': {'en': 'Dolpo'},
    'drg': {'en': 'Rungus'},
    'drh': {'en': 'Darkhat'},
    'dri': {'en': "C'Lela"},
    'drl': {'en': 'Paakantyi'},
    'drn': {'en': 'West Damar'},
    'dro': {'en': 'Daro-Matu Melanau'},
    'drq': {'en': 'Dura'},
    'drr': {'en': 'Dororo'},
    'drs': {'en': 'Gedeo'},
    'drt': {'en': 'Drents'},
    'dru': {'en': 'Rukai'},
    'dru-pro': {'en': 'Proto-Rukai'},
    'drw': {'en': 'Darwazi'},
    'dry': {'en': 'Darai'},
    'dsb': {'af': 'Nedersorbies', 'ak': 'Sɔɔbia a ɛwɔ fam', 'am': 'የታችኛው ሶርቢያኛ', 'ar': 'صوربيا السفلى', 'as': 'ল’ৱাৰ ছোৰ্বিয়ান', 'ast': 'baxu sorbiu', 'az': 'aşağı sorb', 'az-Cyrl': 'ашағы сорб', 'az-Latn': 'aşağı sorb', 'bal-Latn': 'Láwar Sorbi', 'be': 'ніжнялужыцкая', 'be-tarask': 'ніжнялужыцкая', 'bg': 'долнолужишки', 'bgn': 'صُربی سفلی', 'bn': 'নিম্নতর সোর্বিয়ান', 'br': 'izelsorabeg', 'brx': 'गाहायसिन सर्बियान', 'bs': 'donjolužičkosrpski', 'bs-Cyrl': 'доњолужичкосрпски', 'bs-Latn': 'donjolužičkosrpski', 'ca': 'baix sòrab', 'ca-ES-valencia': 'baix sòrab', 'ccp': '𑄙𑄮𑄣𑄴𑄚𑄬𑄭𑄙𑄳𑄠𑄬 𑄥𑄮𑄢𑄴𑄝𑄨𑄠𑄚𑄴', 'ce': 'сорбийн', 'chr': 'ᎡᎳᏗ ᏐᏈᎠᏂ', 'ckb': 'سربیی خوارین', 'cs': 'dolnolužická srbština', 'cy': 'Sorbeg Isaf', 'da': 'nedersorbisk', 'de': 'Niedersorbisch', 'dsb': 'dolnoserbšćina', 'el': 'Κάτω Σορβικά', 'el-polyton': 'Κάτω Σορβικά', 'en': 'Lower Sorbian', 'en-Dsrt': '𐐢𐐬𐐲𐑉 𐐝𐐬𐑉𐐺𐐨𐐲𐑌', 'es': 'bajo sorbio', 'et': 'alamsorbi', 'eu': 'behe-sorabiera', 'fa': 'صُربی سفلی', 'ff-Adlm': '𞤂𞤫𞤧 𞤅𞤮𞤪𞤦𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'alasorbi', 'fil': 'Lower Sorbian', 'fo': 'lágt sorbian', 'fr': 'bas-sorabe', 'fy': 'Nedersorbysk', 'ga': 'Sorbais Íochtarach', 'gd': 'Sòrbais Ìochdarach', 'gl': 'baixo sorbio', 'gsw': 'Nidersorbisch', 'gu': 'લોઅર સોર્બિયન', 'ha': 'Sorbianci ta kasa', 'he': 'סורבית תחתית', 'hi': 'निचला सॉर्बियन', 'hi-Latn': 'Lower Sorbian', 'hr': 'donjolužički', 'hsb': 'delnjoserbšćina', 'ht': 'bas-sorabe', 'hu': 'alsó-szorb', 'hy': 'ստորին սորբերեն', 'ia': 'basse sorabo', 'id': 'Sorbia Hilir', 'ig': 'Lower Sorbian', 'is': 'lágsorbneska', 'it': 'basso sorabo', 'ja': '低地ソルブ語', 'jv': 'Sorbia Non Standar', 'ka': 'ქვემოსორბული', 'kea': 'sórbiu baxu', 'kgp': 'soramiv rur', 'kk': 'төменгі лужица тілі', 'kk-Cyrl': 'төменгі лужица тілі', 'km': 'សូប៊ីក្រោម', 'kn': 'ಲೋವರ್ ಸರ್ಬಿಯನ್', 'ko': '저지 소르비아어', 'kok': 'लोवर सोर्बियन', 'kok-Deva': 'लोवर सोर्बियन', 'ks': 'بۆنِم ساربِیَن', 'ks-Arab': 'بۆنِم ساربِیَن', 'ksh': 'Niddersorbesch', 'ku': 'sorbîya jêrîn', 'ky': 'төмөнкү сорбианча', 'lb': 'Niddersorbesch', 'lo': 'ຊໍບຽນຕໍ່ກວ່າ', 'lrc': 'سوربی ھاری', 'lt': 'žemutinių sorbų', 'lv': 'lejassorbu', 'mai': 'लोअर सोर्बियाई', 'mi': 'Hōpiana Hakahaka', 'mk': 'долнолужички', 'ml': 'ലോവർ സോർബിയൻ', 'mn': 'доод сорби', 'mr': 'लोअर सोर्बियन', 'ms': 'Sorbian Rendah', 'mt': 'Sorbjan Komuni', 'my': 'အနိမ့် ဆိုဘီယန်း', 'mzn': 'پایین صربی', 'nb': 'lavsorbisk', 'ne': 'तल्लो सोर्बियन', 'nl': 'Nedersorbisch', 'nn': 'lågsorbisk', 'no': 'lavsorbisk', 'nqo': 'ߛߏߙߊߓ-ߓߊߛߑߞߊ߲', 'or': 'ନିମ୍ନ ସର୍ବିଆନ୍\u200c', 'pa': 'ਲੋਅਰ ਸੋਰਬੀਅਨ', 'pa-Guru': 'ਲੋਅਰ ਸੋਰਬੀਅਨ', 'pcm': 'Lówá Sorbiá', 'pl': 'dolnołużycki', 'ps': 'کښته سربيايي', 'pt': 'baixo sorábio', 'qu': 'Bajo Sorbio Simi', 'rm': 'bass sorb', 'ro': 'sorabă de jos', 'ru': 'нижнелужицкий', 'sc': 'sòrabu bassu', 'sd': 'لوئر سوربين', 'sd-Arab': 'لوئر سوربين', 'si': 'පහළ සෝබියානු', 'sk': 'dolnolužická srbčina', 'sl': 'dolnja lužiška srbščina', 'smn': 'vyelisorbi', 'so': 'Soorbiyaanka Hoose', 'sq': 'sorbishte e poshtme', 'sr': 'доњолужичкосрпски', 'sr-Cyrl': 'доњолужичкосрпски', 'sr-Latn': 'donjolužičkosrpski', 'sv': 'lågsorbiska', 'sw': 'Kisobia cha Chini', 'ta': 'லோயர் சோர்பியன்', 'te': 'లోయర్ సోర్బియన్', 'tg': 'сербии поёнӣ', 'th': 'ซอร์เบียตอนล่าง', 'ti': 'ታሕተዋይ ሶርብኛ', 'tk': 'aşaky lužits dili', 'to': 'lea fakasōpia-hifo', 'tr': 'Aşağı Sorbça', 'tt': 'түбән сорб', 'ug': 'تۆۋەن سوربچە', 'uk': 'нижньолужицька', 'ur': 'ذیلی سربیائی', 'uz': 'quyi sorb', 'uz-Cyrl': 'қуйи-сорбча', 'uz-Latn': 'quyi sorb', 'vec': 'baso sorabo', 'vi': 'Tiếng Hạ Sorbia', 'wo': 'Sorab-Suuf', 'xnr': 'बुरला सॉर्बियन', 'yi': 'אונטער־סארביש', 'yo': 'Ṣóbíánù Apá Ìṣàlẹ̀', 'yrl': 'surábiu yatuka', 'yue': '下索布文', 'yue-Hans': '下索布文', 'yue-Hant': '下索布文', 'zh': '下索布语', 'zh-Hans': '下索布语', 'zh-Hant': '下索布文', 'zu': 'isi-Lower Sorbian'},
    'dse': {'en': 'Dutch Sign Language'},
    'dsh': {'en': 'Daasanach'},
    'dsi': {'en': 'Disa'},
    'dsl': {'en': 'Danish Sign Language'},
    'dsn': {'en': 'Dusner'},
    'dso': {'en': 'Desiya'},
    'dsq': {'en': 'Tadaksahak'},
    'dta': {'en': 'Daur'},
    'dtb': {'en': 'Labuk-Kinabatangan Kadazan'},
    'dtd': {'en': 'Ditidaht'},
    'dth': {'en': 'Adithinngithigh'},
    'dti': {'en': 'Ana Tinga Dogon'},
    'dtk': {'en': 'Tene Kan Dogon'},
    'dtm': {'en': 'Tomo Kan Dogon'},
    'dtn': {'en': 'Daatsʼíin'},
    'dto': {'en': 'Tommo So Dogon'},
    'dtp': {'am': 'ሴንተራል ዱሰን', 'ast': 'dusun central', 'cs': 'kadazandusunština', 'de': 'Zentral-Dusun', 'en': 'Central Dusun', 'et': 'keskdusuni', 'fi': 'dusun', 'fr': 'dusun central', 'gd': 'Dusun Mheadhanach', 'hi-Latn': 'Central Dusun', 'ht': 'dusun central', 'it': 'dusun centrale', 'ja': '中央ドゥスン語', 'lb': 'Zentral-Dusun', 'lt': 'centrinio Dusuno', 'mk': 'дусунски', 'nb': 'sentraldusun', 'ne': 'केन्द्रीय दुसुन', 'nl': 'Dusun', 'nn': 'sentraldusun', 'no': 'sentraldusun', 'pl': 'dusun centralny', 'sv': 'centraldusun', 'th': 'ดูซุนกลาง', 'to': 'lea fakatusuni-loloto', 'tr': 'Orta Kadazan', 'vi': 'Tiếng Dusun Miền Trung', 'yue': '中部杜順文', 'yue-Hans': '中部杜顺文', 'yue-Hant': '中部杜順文', 'zh-Hant': '中部杜順文'},
    'dtr': {'en': 'Lotud'},
    'dts': {'en': 'Toro So Dogon'},
    'dtt': {'en': 'Toro Tegu Dogon'},
    'dtu': {'en': 'Tebul Ure Dogon'},
    'dty': {'en': 'Dotyali'},
    'dua': {'af': 'Duala', 'am': 'ዱዋላኛ', 'ar': 'الديولا', 'as': 'ডুৱালা', 'ast': 'duala', 'az': 'duala', 'az-Arab': 'دۇالاجا', 'az-Cyrl': 'дуала', 'az-Latn': 'duala', 'bal-Latn': 'Duálá', 'be': 'дуала', 'be-tarask': 'дуала', 'bg': 'дуала', 'bgn': 'دوالی', 'bn': 'দুয়ালা', 'br': 'douala', 'brx': 'डुआला', 'bs': 'duala', 'bs-Cyrl': 'дуала', 'bs-Latn': 'duala', 'ca': 'douala', 'ca-ES-valencia': 'douala', 'ccp': '𑄘𑄱𑄣', 'ce': 'дуала', 'chr': 'ᏚᎠᎳ', 'ckb': 'دووالا', 'cs': 'dualština', 'cy': 'Diwaleg', 'da': 'duala', 'de': 'Duala', 'dsb': 'duala', 'dua': 'duálá', 'el': 'Ντουάλα', 'el-polyton': 'Ντουάλα', 'en': 'Duala', 'es': 'duala', 'et': 'duala', 'eu': 'dualera', 'fa': 'دوآلایی', 'ff-Adlm': '𞤁𞤵𞤱𞤢𞤤𞤢𞥄𞤪𞤫', 'fi': 'duala', 'fil': 'Duala', 'fo': 'duala', 'fr': 'douala', 'fy': 'Duala', 'ga': 'Duailis', 'gd': 'Duala', 'gl': 'duala', 'gsw': 'Tuala', 'gu': 'દુઆલા', 'ha': 'Duala', 'he': 'דואלה', 'hi': 'दुआला', 'hi-Latn': 'Duala', 'hr': 'duala', 'hsb': 'duala', 'ht': 'douala', 'hu': 'duala', 'hy': 'դուալա', 'ia': 'duala', 'id': 'Duala', 'ig': 'Duala', 'is': 'dúala', 'it': 'duala', 'ja': 'ドゥアラ語', 'jv': 'Duala', 'ka': 'დუალა', 'kea': 'duala', 'kgp': 'nuvara', 'kk': 'дуала тілі', 'kk-Cyrl': 'дуала тілі', 'km': 'ឌួលឡា', 'kn': 'ಡುವಾಲಾ', 'ko': '두알라어', 'kok': 'डुआला', 'kok-Deva': 'डुआला', 'ks': 'دُوالا', 'ks-Arab': 'دُوالا', 'ksh': 'Du’ala', 'ku': 'diwalayî', 'ky': 'дуалача', 'lb': 'Duala', 'lo': 'ດົວລາ', 'lrc': 'دوٙالا', 'lt': 'dualų', 'lv': 'dualu', 'mai': 'डुआला', 'mi': 'Tuāra', 'mk': 'дуала', 'ml': 'ദ്വാല', 'mn': 'дуала', 'mr': 'दुआला', 'ms': 'Duala', 'mt': 'Dwala', 'my': 'ဒူအလာ', 'mzn': 'دوئالایی', 'nb': 'duala', 'ne': 'दुवाला', 'nl': 'Duala', 'nn': 'duala', 'no': 'duala', 'nqo': 'ߘߎߥߟߊߞߊ߲', 'or': 'ଡୁଆନା', 'pa': 'ਡੂਆਲਾ', 'pa-Guru': 'ਡੂਆਲਾ', 'pcm': 'Duála Lángwej', 'pl': 'duala', 'ps': 'دوالا', 'pt': 'duala', 'qu': 'Duala Simi', 'rm': 'duala', 'ro': 'duala', 'ru': 'дуала', 'sc': 'duala', 'sd': 'ڊيولا', 'sd-Arab': 'ڊيولا', 'si': 'ඩුආලා', 'sk': 'duala', 'sl': 'duala', 'smn': 'dualakielâ', 'so': 'Duaala', 'sq': 'dualaisht', 'sr': 'дуала', 'sr-Cyrl': 'дуала', 'sr-Latn': 'duala', 'sv': 'duala', 'sw': 'Kiduala', 'ta': 'டுவாலா', 'te': 'డ్యూలా', 'th': 'ดัวลา', 'ti': 'ድዋላ', 'tk': 'duala dili', 'to': 'lea fakatuala', 'tr': 'Duala', 'ug': 'دۇئالاچە', 'uk': 'дуала', 'ur': 'دوالا', 'uz': 'duala', 'uz-Cyrl': 'дуалача', 'uz-Latn': 'duala', 'vec': 'duala', 'vi': 'Tiếng Duala', 'xnr': 'दुआला', 'yo': 'Èdè Duala', 'yrl': 'duara', 'yue': '杜亞拉文', 'yue-Hans': '杜亚拉文', 'yue-Hant': '杜亞拉文', 'zh': '杜阿拉语', 'zh-Hans': '杜阿拉语', 'zh-Hant': '杜亞拉文', 'zu': 'isi-Duala'},
    'dub': {'en': 'Dubli'},
    'duc': {'en': 'Duna'},
    'dud': {'en': 'Hun-Saare'},
    'due': {'en': 'Umiray Dumaget Agta'},
    'duf': {'en': 'Dumbea'},
    'dug': {'en': 'Duruma'},
    'duh': {'en': 'Dungra Bhil'},
    'dui': {'en': 'Dumun'},
    'duj': {'en': 'Dhuwal'},
    'duk': {'en': 'Uyajitaya'},
    'dul': {'en': 'Alabat Island Agta'},
    'dum': {'ar': 'الهولندية الوسطى', 'ast': 'neerlandés mediu', 'az': 'orta holland', 'az-Latn': 'orta holland', 'bg': 'средновековен холандски', 'bn': 'মধ্য ডাচ', 'br': 'nederlandeg krenn', 'brx': 'मध्य डच', 'bs': 'srednjovjekovni holandski', 'bs-Cyrl': 'средњи холандски', 'bs-Latn': 'srednjovjekovni holandski', 'ca': 'neerlandès mitjà', 'ca-ES-valencia': 'neerlandès mitjà', 'ccp': '𑄟𑄧𑄖𑄴𑄙𑄳𑄠𑄧 𑄓𑄌𑄴', 'cs': 'holandština (středověká)', 'cy': 'Iseldireg Canol', 'da': 'middelhollandsk', 'de': 'Mittelniederländisch', 'el': 'Μέσα Ολλανδικά', 'el-polyton': 'Μέσα Ολλανδικά', 'en': 'Middle Dutch', 'en-Dsrt': '𐐣𐐮𐐼𐐲𐑊 𐐔𐐲𐐽', 'es': 'neerlandés medio', 'et': 'keskhollandi', 'fa': 'هلندی میانه', 'fi': 'keskihollanti', 'fr': 'moyen néerlandais', 'fy': 'Middelnederlânsk', 'ga': 'Meán-Ollainnis', 'gd': 'Meadhan-Dhuitsis', 'gsw': 'Mittelniderländisch', 'gu': 'મધ્ય ડચ', 'he': 'הולנדית תיכונה', 'hi': 'मध्यकालीन पुर्तगाली', 'hi-Latn': 'Middle Dutch', 'hr': 'srednjonizozemski', 'ht': 'moyen néerlandais', 'hu': 'közép holland', 'id': 'Belanda Abad Pertengahan', 'is': 'miðhollenska', 'it': 'olandese medio', 'ja': '中世オランダ語', 'ka': 'საშუალო ჰოლანდიური', 'kgp': 'orỹnej kuju', 'kn': 'ಮಧ್ಯ ಡಚ್', 'ko': '중세 네덜란드어', 'ks': 'وَستی پُرتُگالی', 'ks-Arab': 'وَستی پُرتُگالی', 'ksh': 'Meddelnehderlängsch', 'lb': 'Mëttelhollännesch', 'lo': 'ດັກກາງ', 'lt': 'Vidurio Vokietijos', 'lv': 'vidusholandiešu', 'mk': 'среднохоландски', 'ml': 'മദ്ധ്യ ഡച്ച്', 'mr': 'मिडल डच', 'mt': 'Olandiż Medjevali', 'my': 'အလယ်ပိုင်း ဒတ်ချ်', 'nb': 'mellomnederlandsk', 'ne': 'मध्य डच', 'nl': 'Middelnederlands', 'nn': 'mellomnederlandsk', 'no': 'mellomnederlandsk', 'or': 'ମଧ୍ୟ ପର୍ତ୍ତୁଗାଲୀ', 'pl': 'średniowieczny niderlandzki', 'pt': 'holandês médio', 'rm': 'ollandais mesaun', 'ro': 'neerlandeză medie', 'ru': 'средненидерландский', 'sk': 'stredná holandčina', 'sl': 'srednja nizozemščina', 'sr': 'средњехоландски', 'sr-Cyrl': 'средњехоландски', 'sr-Latn': 'srednjeholandski', 'sv': 'medelnederländska', 'ta': 'மிடில் டச்சு', 'te': 'మధ్యమ డచ్', 'th': 'ดัตช์กลาง', 'to': 'lea fakahōlani-lotoloto', 'tr': 'Ortaçağ Felemenkçesi', 'ug': 'ئوتتۇرا گوللاندىيەچە', 'uk': 'середньонідерландська', 'vi': 'Tiếng Hà Lan Trung cổ', 'yrl': 'hurãdei médiu', 'yue': '中古荷蘭文', 'yue-Hans': '中古荷兰文', 'yue-Hant': '中古荷蘭文', 'zh': '中古荷兰语', 'zh-Hans': '中古荷兰语', 'zh-Hant': '中古荷蘭文'},
    'dun': {'en': 'Dusun Deyah'},
    'duo': {'en': 'Dupaninan Agta'},
    'dup': {'en': 'Duano'},
    'duq': {'en': 'Dusun Malang'},
    'dur': {'en': 'Dii'},
    'dus': {'en': 'Dumi'},
    'duu': {'en': 'Drung'},
    'duv': {'en': 'Duvle'},
    'duw': {'en': 'Dusun Witu'},
    'dux': {'en': 'Duungooma'},
    'duy': {'en': 'Dicamay Agta'},
    'duz': {'en': 'Duli-Gey'},
    'dv': {'af': 'Divehi', 'am': 'ዲቬሂ', 'ar': 'المالديفية', 'as': 'দিবেহি', 'ast': 'divehi', 'az': 'maldiv', 'az-Cyrl': 'малдив', 'az-Latn': 'maldiv', 'bal-Latn': 'Diwéhi', 'be': 'мальдыўская', 'be-tarask': 'мальдыўская', 'bg': 'дивехи', 'bgn': 'دیوهی', 'bn': 'দিবেহি', 'br': 'divehi', 'brx': 'दीवेही', 'bs': 'divehi', 'bs-Cyrl': 'дивехијски', 'bs-Latn': 'divehi', 'ca': 'divehi', 'ca-ES-valencia': 'divehi', 'ccp': '𑄘𑄨𑄝𑄬𑄦𑄨', 'ce': 'мальдивийн', 'chr': 'ᏗᏪᎯ', 'ckb': 'دیڤێهی', 'cs': 'maledivština', 'cy': 'Difehi', 'da': 'divehi', 'de': 'Dhivehi', 'dsb': 'divehi', 'dz': 'དི་བེ་ཧི་ཁ', 'ee': 'divehgbe', 'el': 'Ντιβέχι', 'el-polyton': 'Ντιβέχι', 'en': 'Divehi', 'eo': 'maldiva', 'es': 'divehi', 'et': 'maldiivi', 'eu': 'dhivehia', 'fa': 'دیوهی', 'ff-Adlm': '𞤁𞤭𞥅𞤬𞤫𞤸𞤭𞥅𞤪𞤫', 'fi': 'divehi', 'fil': 'Divehi', 'fo': 'divehi', 'fr': 'maldivien', 'fy': 'Divehi', 'ga': 'Divéihis', 'gd': 'Divehi', 'gl': 'divehi', 'gsw': 'Malediivisch', 'gu': 'દિવેહી', 'ha': 'Divehi', 'he': 'דיבהי', 'hi': 'दिवेही', 'hi-Latn': 'Divehi', 'hr': 'divehi', 'hsb': 'divehi', 'ht': 'maldivien', 'hu': 'divehi', 'hy': 'դիվեհի', 'ia': 'divehi', 'id': 'Divehi', 'ig': 'Divehi', 'is': 'dívehí', 'it': 'divehi', 'ja': 'ディベヒ語', 'jv': 'Divehi', 'ka': 'დივეჰი', 'kgp': 'nivehi', 'kk': 'дивехи тілі', 'kk-Cyrl': 'дивехи тілі', 'km': 'ទេវីហ៊ី', 'kn': 'ದಿವೆಹಿ', 'ko': '디베히어', 'kok': 'दिवेही', 'kok-Deva': 'दिवेही', 'ks': 'دِویہی', 'ks-Arab': 'دِویہی', 'ksh': 'Divehjesch', 'ku': 'divehî', 'ky': 'дивехиче', 'lb': 'Maldivesch', 'lld': 'divehi', 'lo': 'ດີວີຮີ', 'lt': 'divehų', 'lv': 'maldīviešu', 'mai': 'डिबेही', 'mi': 'Tīwhehi', 'mk': 'дивехи', 'ml': 'ദിവെഹി', 'mn': 'дивехи', 'mr': 'दिवेही', 'ms': 'Divehi', 'mt': 'Divehi', 'my': 'ဒီဗာဟီ', 'nb': 'divehi', 'ne': 'दिबेही', 'nl': 'Divehi', 'nn': 'divehi', 'no': 'divehi', 'or': 'ଡିଭେହୀ', 'pa': 'ਦਿਵੇਹੀ', 'pa-Guru': 'ਦਿਵੇਹੀ', 'pcm': 'Divẹhí', 'pl': 'malediwski', 'ps': 'ديویهی', 'pt': 'divehi', 'qu': 'Divehi Simi', 'rm': 'maledivic', 'ro': 'divehi', 'ru': 'мальдивский', 'sc': 'malvidianu', 'sd': 'دويهي', 'sd-Arab': 'دويهي', 'se': 'divehigiella', 'si': 'ඩිවෙහි', 'sk': 'maldivčina', 'sl': 'diveščina', 'smn': 'divehikielâ', 'so': 'Divehi', 'sq': 'divehisht', 'sr': 'малдивски', 'sr-Cyrl': 'малдивски', 'sr-Latn': 'maldivski', 'sv': 'divehi', 'sw': 'Kidivehi', 'ta': 'திவேஹி', 'te': 'దివేహి', 'tg': 'дивеҳӣ', 'th': 'ธิเวหิ', 'ti': 'ዲቨሂ', 'tk': 'diwehi dili', 'to': 'lea fakativehi', 'tr': 'Divehi dili', 'tt': 'мальдив', 'ug': 'دىۋېخچە', 'uk': 'дивехі', 'ur': 'ڈیویہی', 'uz': 'divexi', 'uz-Cyrl': 'дивехи', 'uz-Latn': 'divexi', 'vec': 'maldivian', 'vi': 'Tiếng Divehi', 'wae': 'Malediwiš', 'wo': 'Diweyi', 'xnr': 'दिवेही', 'yo': 'Èdè Difehi', 'yrl': 'diweí', 'yue': '迪維西文', 'yue-Hans': '迪维西文', 'yue-Hant': '迪維西文', 'zh': '迪维希语', 'zh-Hans': '迪维希语', 'zh-Hant': '迪維西文', 'zu': 'isi-Divehi'},
    'dva': {'en': 'Duau'},
    'dwa': {'en': 'Diri'},
    'dwk': {'en': 'Dawik Kui'},
    'dwl': {'en': 'Walo Kumbe Dogon'},
    'dwr': {'en': 'Dawro'},
    'dws': {'en': 'Dutton World Speedwords'},
    'dwu': {'en': 'Dhuwal'},
    'dww': {'en': 'Dawawa'},
    'dwy': {'en': 'Dhuwaya'},
    'dwz': {'en': 'Dewas Rai'},
    'dya': {'en': 'Dyan'},
    'dyb': {'en': 'Dyaberdyaber'},
    'dyd': {'en': 'Dyugun'},
    'dyg': {'en': 'Villa Viciosa Agta'},
    'dyi': {'en': 'Djimini Senoufo'},
    'dym': {'en': 'Yanda Dom Dogon'},
    'dyn': {'en': 'Dyangadi'},
    'dyo': {'af': 'Jola-Fonyi', 'am': 'ጆላ-ፎንዪ', 'ar': 'جولا فونيا', 'as': 'জোলা-ফ’নি', 'ast': 'jola-fonyi', 'az': 'diola', 'az-Cyrl': 'диола', 'az-Latn': 'diola', 'bal-Latn': 'Jólá-Póni', 'be': 'джола-фоньі', 'be-tarask': 'джола-фоньі', 'bg': 'диола-фони', 'bgn': 'جولا فونی', 'bn': 'জোলা-ফনী', 'br': 'diola', 'brx': 'जला- फन्यी', 'bs': 'jola-foni', 'bs-Cyrl': 'јола-фоњи', 'bs-Latn': 'jola-foni', 'ca': 'diola', 'ca-ES-valencia': 'diola', 'ccp': '𑄎𑄧𑄣-𑄜𑄧𑄚𑄩', 'ce': 'дьола-фоньи', 'chr': 'ᏦᎳ-ᏬᏱ', 'ckb': 'جۆلافۆنی', 'cs': 'jola-fonyi', 'cy': 'Jola-Fonyi', 'da': 'jola-fonyi', 'de': 'Diola', 'dsb': 'jola-fonyi', 'dyo': 'joola', 'el': 'Τζόλα-Φόνι', 'el-polyton': 'Τζόλα-Φόνι', 'en': 'Jola-Fonyi', 'es': 'jola-fonyi', 'et': 'fonji', 'eu': 'fonyi jolera', 'fa': 'دیولا فونی', 'ff-Adlm': '𞤔𞤮𞥅𞤤𞤢𞥄𞤪𞤫', 'fi': 'jola-fonyi', 'fil': 'Jola-Fonyi', 'fo': 'jola-fonyi', 'fr': 'diola-fogny', 'fy': 'Jola-Fonyi', 'ga': 'Jóla-Fainis', 'gd': 'Jola-Fonyi', 'gl': 'jola-fonyi', 'gu': 'જોલા-ફોન્યી', 'ha': 'Jola-Fonyi', 'he': 'ג׳ולה פונית', 'hi': 'जोला-फोंई', 'hi-Latn': 'Jola-Fonyi', 'hr': 'jola-fonyi', 'hsb': 'jola-fonyi', 'ht': 'diola-fogny', 'hu': 'jola-fonyi', 'hy': 'ջոլա-ֆոնյի', 'ia': 'jola-fonyi', 'id': 'Jola-Fonyi', 'ig': 'Jola-Fonyi', 'is': 'jola-fonyi', 'it': 'jola-fony', 'ja': 'ジョラ＝フォニィ語', 'jv': 'Jola-Fonyi', 'ka': 'დიოლა', 'kea': 'jola-fonyi', 'kgp': 'jora-fonyj', 'kk': 'диола тілі', 'kk-Cyrl': 'диола тілі', 'km': 'ចូឡាហ៊្វុនយី', 'kn': 'ಜೊಲ-ಫೊನ್ಯಿ', 'ko': '졸라 포니어', 'kok': 'जोला-फोन्यी', 'kok-Deva': 'जोला-फोन्यी', 'ksh': 'Jola-Fonyisch', 'ku': 'jola-fonyi', 'ky': 'жола-фониче', 'lb': 'Jola-Fonyi', 'lo': 'ໂຈລາ-ຟອນຢີ', 'lrc': 'جولا فوٙنیی', 'lt': 'džiola-foni', 'lv': 'diola-fonjī', 'mai': 'जोला-फोनी', 'mi': 'Hora-Whōni', 'mk': 'јола-фоњи', 'ml': 'ജോല-ഫോൻയി', 'mn': 'жола-фони', 'mr': 'जोला-फॉन्यी', 'ms': 'Jola-Fonyi', 'mt': 'Jola-Fonyi', 'my': 'ဂျိုလာ-ဖွန်ရီ', 'mzn': 'جولا-فونی', 'nb': 'jola-fonyi', 'ne': 'जोला-फोनिल', 'nl': 'Jola-Fonyi', 'nn': 'jola-fonyi', 'no': 'jola-fonyi', 'nqo': 'ߝߐߢߌ߫ ߖߏ߬ߟߊ߬ߞߊ߲', 'or': 'ଜୋଲା-ଫୋନୟି', 'pa': 'ਜੋਲਾ-ਫੋਇਨੀ', 'pa-Guru': 'ਜੋਲਾ-ਫੋਇਨੀ', 'pcm': 'Jóla-Fónyi Lángwej', 'pl': 'diola', 'ps': 'جولا فوني', 'pt': 'jola-fonyi', 'qu': 'Jola-Fonyi Simi', 'rm': 'jola-fonyi', 'ro': 'jola-fonyi', 'ru': 'диола-фоньи', 'sc': 'jola-fonyi', 'sd': 'جولا فوني', 'sd-Arab': 'جولا فوني', 'si': 'ජොල-ෆෝනියි', 'sk': 'jola-fonyi', 'sl': 'jola-fonjiščina', 'smn': 'jola-fonyi', 'so': 'Joola-Foonyi', 'sq': 'xhulafonjisht', 'sr': 'џола фоњи', 'sr-Cyrl': 'џола фоњи', 'sr-Latn': 'džola fonji', 'sv': 'jola-fonyi', 'sw': 'Kijola-Fonyi', 'ta': 'ஜோலா-ஃபோன்யி', 'te': 'జోలా-ఫోనయి', 'th': 'โจลา-ฟอนยี', 'ti': 'ጆላ-ፎኒይ', 'tk': 'ýola-fonýi dili', 'to': 'lea fakaiola-fonī', 'tr': 'Jola-Fonyi', 'ug': 'جولاچە', 'uk': 'дьола-фоні', 'ur': 'جولا فونيا', 'uz': 'diola-fogni', 'uz-Cyrl': 'диола-фогни', 'uz-Latn': 'diola-fogni', 'vec': 'jola foni', 'vi': 'Tiếng Jola-Fonyi', 'xnr': 'जोला-फोंई', 'yi': 'זשאלא־פֿאני', 'yo': 'Jola-Fonyi', 'yrl': 'yora-funiyi', 'yue': '朱拉文', 'yue-Hans': '朱拉文', 'yue-Hant': '朱拉文', 'zh': '朱拉语', 'zh-Hans': '朱拉语', 'zh-Hant': '朱拉文', 'zu': 'isi-Jola-Fonyi'},
    'dyu': {'am': 'ድዩላ', 'ar': 'الدايلا', 'ast': 'dyula', 'az': 'dyula', 'az-Latn': 'dyula', 'bg': 'диула', 'bn': 'ডিউলা', 'br': 'dyula', 'brx': 'द्युआला', 'bs': 'diula', 'bs-Cyrl': 'ђула', 'bs-Latn': 'diula', 'ca': 'jula', 'ca-ES-valencia': 'jula', 'ccp': '𑄓𑄨𑄃𑄪𑄣', 'cs': 'djula', 'da': 'dyula', 'de': 'Dyula', 'el': 'Ντογιούλα', 'el-polyton': 'Ντογιούλα', 'en': 'Dyula', 'es': 'diula', 'et': 'djula', 'fa': 'دایولایی', 'fi': 'djula', 'fr': 'dioula', 'fy': 'Dyula', 'gd': 'Dyula', 'gsw': 'Tiula', 'gu': 'ડ્યુલા', 'he': 'דיולה', 'hi': 'ड्युला', 'hi-Latn': 'Dyula', 'hr': 'dyula', 'ht': 'dioula', 'hu': 'diula', 'id': 'Dyula', 'is': 'djúla', 'it': 'diula', 'ja': 'ジュラ語', 'ka': 'დიულა', 'kgp': 'nhivura', 'kn': 'ಡ್ಯೂಲಾ', 'ko': '드율라어', 'ks': 'ڈِیوٗلا', 'ks-Arab': 'ڈِیوٗلا', 'lb': 'Dyula-Sprooch', 'lo': 'ດູລາ', 'lt': 'dyulų', 'lv': 'diūlu', 'mk': 'џула', 'ml': 'ദ്വൈല', 'mr': 'ड्युला', 'mt': 'Dyula', 'nb': 'dyula', 'ne': 'द्युला', 'nl': 'Dyula', 'nn': 'dyula', 'no': 'dyula', 'or': 'ଡୁଆଲା', 'pl': 'diula', 'pt': 'diúla', 'rm': 'diula', 'ro': 'dyula', 'ru': 'диула', 'sk': 'ďula', 'sl': 'diula', 'sr': 'ђула', 'sr-Cyrl': 'ђула', 'sr-Latn': 'đula', 'sv': 'dyula', 'sw': 'Kijula', 'ta': 'ட்யூலா', 'te': 'డ్యులా', 'th': 'ดิวลา', 'to': 'lea fakatiula', 'tr': 'Dyula', 'ug': 'دىيۇلاچە', 'uk': 'діула', 'vi': 'Tiếng Dyula', 'yrl': 'diura', 'yue': '迪尤拉文', 'yue-Hans': '迪尤拉文', 'yue-Hant': '迪尤拉文', 'zh': '迪尤拉语', 'zh-Hans': '迪尤拉语', 'zh-Hant': '迪尤拉文'},
    'dyy': {'en': 'Djabugay'},
    'dz': {'af': 'Dzongkha', 'am': 'ድዞንግኻኛ', 'ar': 'دزونكا', 'as': 'জোংখা', 'ast': 'dzongkha', 'az': 'dzonqxa', 'az-Arab': 'دزونگخا', 'az-Cyrl': 'дзонга', 'az-Latn': 'dzonqxa', 'bal-Latn': 'Dzongká', 'be': 'дзонг-кэ', 'be-tarask': 'дзонг-кэ', 'bg': 'дзонгкха', 'bgn': 'دزونگخا', 'bn': 'জোংখা', 'bo': 'རྫོང་ཁ', 'br': 'dzongkha', 'brx': 'झंखा', 'bs': 'džonga', 'bs-Cyrl': 'џонга', 'bs-Latn': 'džonga', 'ca': 'dzongka', 'ca-ES-valencia': 'dzongka', 'ccp': '𑄎𑄮𑄋𑄴𑄉', 'ce': 'дзонг-кэ', 'chr': 'ᏓᏐᏅᎧ', 'ckb': 'دزوونگخا', 'cs': 'dzongkä', 'cy': 'Dzongkha', 'da': 'dzongkha', 'de': 'Dzongkha', 'dsb': 'dzongkha', 'dz': 'རྫོང་ཁ', 'ee': 'dzongkhagbe', 'el': 'Ντζόνγκχα', 'el-polyton': 'Ντζόνγκχα', 'en': 'Dzongkha', 'eo': 'dzonko', 'es': 'dzongkha', 'et': 'dzongkha', 'eu': 'dzongkha', 'fa': 'دزونگخا', 'ff-Adlm': '𞤄𞤵𞥅𞤼𞤢𞤲𞤪𞤫', 'fi': 'dzongkha', 'fil': 'Dzongkha', 'fo': 'dzongkha', 'fr': 'dzongkha', 'fy': 'Dzongkha', 'ga': 'Seoinicis', 'gd': 'Dzongkha', 'gl': 'dzongkha', 'gsw': 'Dschongkha', 'gu': 'ડ્ઝોંગ્ખા', 'ha': 'Dzongkha', 'he': 'דזונקה', 'hi': 'ज़ोन्गखा', 'hi-Latn': 'Dzongkha', 'hr': 'dzongkha', 'hsb': 'dzongkha', 'ht': 'dzongkha', 'hu': 'dzsonga', 'hy': 'ջոնգքհա', 'ia': 'dzongkha', 'id': 'Dzongkha', 'ig': 'Dọzngọka', 'is': 'dsongka', 'it': 'dzongkha', 'ja': 'ゾンカ語', 'jv': 'Dzongkha', 'ka': 'ძონგკხა', 'kea': 'dzonka', 'kgp': 'jãnga', 'kk': 'дзонг-кэ тілі', 'kk-Cyrl': 'дзонг-кэ тілі', 'km': 'ដុងខា', 'kn': 'ಜೋಂಗ್\u200cಖಾ', 'ko': '종카어', 'kok': 'झोंग्खा', 'kok-Deva': 'झोंग्खा', 'ks': 'زونگکھا', 'ks-Arab': 'زونگکھا', 'ksh': 'Butahnesch', 'ku': 'conxayî', 'ky': 'жонгуча', 'lb': 'Bhutanesch', 'lld': 'dzongkha', 'lo': 'ດີຊອງຄາ', 'lrc': 'زوٙنگخا', 'lt': 'botijų', 'lv': 'dzongke', 'mai': 'जोनखा', 'mi': 'Tonoka', 'mk': 'ѕонгка', 'ml': 'ദ്\u200cസോങ്ക', 'mn': 'зонха', 'mr': 'झोंगखा', 'ms': 'Dzongkha', 'mt': 'Dzongkha', 'my': 'ဒဇွန်ကာ', 'mzn': 'دزونگخا', 'nb': 'dzongkha', 'ne': 'जोङ्खा', 'nl': 'Dzongkha', 'nn': 'dzongkha', 'no': 'dzongkha', 'nqo': 'ߖ߭ߏ߲ߜ߭ߊߞߊ߲', 'or': 'ଦଡଜୋଙ୍ଗଖା', 'pa': 'ਜ਼ੋਂਗਖਾ', 'pa-Guru': 'ਜ਼ੋਂਗਖਾ', 'pcm': 'Zọ́ngka Lángwej', 'pl': 'dzongkha', 'ps': 'ژونگکه', 'pt': 'dzonga', 'qu': 'Butanés Simi', 'rm': 'dzongkha', 'ro': 'dzongkha', 'ru': 'дзонг-кэ', 'sc': 'dzongkha', 'sd': 'زونخا', 'sd-Arab': 'زونخا', 'se': 'dzongkhagiella', 'si': 'ඩිසොන්කා', 'sk': 'dzongkha', 'sl': 'dzonka', 'smn': 'Dzongkha', 'so': 'D’zongqa', 'sq': 'xhongaisht', 'sr': 'џонга', 'sr-Cyrl': 'џонга', 'sr-Latn': 'džonga', 'sv': 'dzongkha', 'sw': 'Kizongkha', 'ta': 'பூடானி', 'te': 'జోంఖా', 'tg': 'дзонгха', 'th': 'ซองคา', 'ti': 'ድዞንግካ', 'tk': 'dzong-ke dili', 'to': 'lea fakatisōngika', 'tr': 'Dzongkha', 'tt': 'дзонг-кха', 'ug': 'زوڭخاچە', 'uk': 'дзонг-ке', 'ur': 'ژونگکھا', 'uz': 'dzongka', 'uz-Cyrl': 'дзонгка', 'uz-Latn': 'dzongka', 'vec': 'dzongka', 'vi': 'Tiếng Dzongkha', 'wae': 'Butaniš', 'wo': 'Dsongkaa', 'xnr': 'ज़ोन्गखा', 'yo': 'Èdè Dzongkha', 'yrl': 'disũga', 'yue': '宗卡文', 'yue-Hans': '宗卡文', 'yue-Hant': '宗卡文', 'zh': '宗卡语', 'zh-Hans': '宗卡语', 'zh-Hant': '宗卡文', 'zu': 'isi-Dzongkha'},
    'dza': {'en': 'Tunzu'},
    'dzd': {'en': 'Daza'},
    'dze': {'en': 'Djiwarli'},
    'dzg': {'af': 'Dazaga', 'am': 'ዳዛጋ', 'ar': 'القرعانية', 'as': 'দাজাগা', 'ast': 'dazaga', 'az': 'dazaqa', 'az-Cyrl': 'дазага', 'az-Latn': 'dazaqa', 'be': 'дазага', 'be-tarask': 'дазага', 'bg': 'дазага', 'bgn': 'دزازا', 'bn': 'দাজাগা', 'br': 'dazagaeg', 'brx': 'डाजागा', 'bs': 'dazaga', 'bs-Cyrl': 'дазага', 'bs-Latn': 'dazaga', 'ca': 'dazaga', 'ca-ES-valencia': 'dazaga', 'ccp': '𑄘𑄉𑄎', 'ce': 'даза', 'chr': 'ᏓᏌᎦ', 'ckb': 'دازا', 'cs': 'dazaga', 'cy': 'Dazaga', 'da': 'dazaga', 'de': 'Dazaga', 'dsb': 'dazagašćina', 'el': 'Νταζάγκα', 'el-polyton': 'Νταζάγκα', 'en': 'Dazaga', 'es': 'dazaga', 'et': 'daza', 'eu': 'dazaera', 'fa': 'دازاگا', 'ff-Adlm': '𞤁𞤢𞤶𞤢𞤺𞤢𞥄𞤪𞤫', 'fi': 'dazaga', 'fil': 'Dazaga', 'fo': 'dazaga', 'fr': 'dazaga', 'fy': 'Dazaga', 'ga': 'Dazaga', 'gd': 'Dazaga', 'gl': 'dazaga', 'gu': 'દાઝાગા', 'ha': 'Dazaga', 'he': 'דזאנגה', 'hi': 'दज़ागा', 'hi-Latn': 'Dazaga', 'hr': 'dazaga', 'hsb': 'dazagašćina', 'ht': 'dazaga', 'hu': 'dazaga', 'hy': 'դազագա', 'ia': 'dazaga', 'id': 'Dazaga', 'ig': 'Dazaga', 'is': 'dazaga', 'it': 'dazaga', 'ja': 'ダザガ語', 'jv': 'Dazaga', 'ka': 'დაზაგა', 'kgp': 'najaga', 'kk': 'дазага тілі', 'kk-Cyrl': 'дазага тілі', 'km': 'ដាហ្សាហ្គា', 'kn': 'ಡಜಾಗ', 'ko': '다장가어', 'kok': 'डाझागा', 'kok-Deva': 'डाझागा', 'ksh': 'Dassajahnesch', 'ku': 'dazagayî', 'ky': 'дазагача', 'lb': 'Dazaga', 'lo': 'ດາຊາກາ', 'lt': 'dazagų', 'lv': 'dazu', 'mai': 'दजागा', 'mi': 'Tahāka', 'mk': 'дазага', 'ml': 'ഡാസാഗ', 'mn': 'дазага', 'mr': 'दाझागा', 'ms': 'Dazaga', 'mt': 'Dazaga', 'my': 'ဒဇာဂါ', 'nb': 'dazaga', 'ne': 'दाजागा', 'nl': 'Dazaga', 'nn': 'dazaga', 'no': 'dazaga', 'or': 'ଡାଜାଗା', 'pa': 'ਡਜ਼ਾਗਾ', 'pa-Guru': 'ਡਜ਼ਾਗਾ', 'pcm': 'Dazágá Lángwej', 'pl': 'dazaga', 'ps': 'ډزاګا', 'pt': 'dazaga', 'qu': 'Dazaga Simi', 'ro': 'dazaga', 'ru': 'даза', 'sc': 'dazaga', 'sd': 'دزاگا', 'sd-Arab': 'دزاگا', 'si': 'ඩසාගා', 'sk': 'dazaga', 'sl': 'dazaga', 'smn': 'dazakielâ', 'so': 'Dazaga', 'sq': 'dazagauisht', 'sr': 'дазага', 'sr-Cyrl': 'дазага', 'sr-Latn': 'dazaga', 'sv': 'dazaga', 'sw': 'Kidazaga', 'ta': 'டசாகா', 'te': 'డాజాగా', 'th': 'ดาซากา', 'ti': 'ዳዛጋ', 'tk': 'daza dili', 'to': 'lea fakatasaka', 'tr': 'Dazaga', 'ug': 'دازاگاچە', 'uk': 'дазага', 'ur': 'دزاگا', 'uz': 'dazag', 'uz-Cyrl': 'дазага', 'uz-Latn': 'dazag', 'vec': 'dazaga', 'vi': 'Tiếng Dazaga', 'xnr': 'दज़ागा', 'yo': 'Èdè Dasaga', 'yrl': 'dasaga', 'yue': '達薩文', 'yue-Hans': '达萨文', 'yue-Hant': '達薩文', 'zh': '达扎葛语', 'zh-Hans': '达扎葛语', 'zh-Hant': '達薩文', 'zu': 'isi-Dazaga'},
    'dzl': {'en': 'Dzalakha'},
    'dzn': {'en': 'Dzando'},
    'eaa': {'en': 'Karenggapa'},
    'ebc': {'en': 'Beginci'},
    'ebg': {'en': 'Ebughu'},
    'ebk': {'en': 'Eastern Bontok'},
    'ebo': {'en': 'Teke-Ebo'},
    'ebr': {'en': 'Ebrié'},
    'ebu': {'af': 'Embu', 'am': 'ኢምቡ', 'ar': 'إمبو', 'as': 'এম্বু', 'ast': 'embú', 'az': 'embu', 'az-Arab': 'ائمبوجا', 'az-Cyrl': 'ембу', 'az-Latn': 'embu', 'bal-Latn': 'Embó', 'be': 'эмбу', 'be-tarask': 'эмбу', 'bg': 'ембу', 'bgn': 'ایمبو', 'bn': 'এম্বু', 'br': 'embu', 'brx': 'इम्बु', 'bs': 'embu', 'bs-Cyrl': 'ембу', 'bs-Latn': 'embu', 'ca': 'embu', 'ca-ES-valencia': 'embu', 'ccp': '𑄃𑄬𑄟𑄳𑄝𑄪𑄪', 'ce': 'эмбу', 'chr': 'ᎡᎻᏊ', 'ckb': 'ئێمبوو', 'cs': 'embu', 'cy': 'Embw', 'da': 'kiembu', 'de': 'Embu', 'dsb': 'embu', 'ebu': 'Kĩembu', 'ee': 'embugbe', 'el': 'Έμπου', 'el-polyton': 'Έμπου', 'en': 'Embu', 'es': 'embu', 'et': 'embu', 'eu': 'embuera', 'fa': 'امبو', 'ff-Adlm': '𞤉𞤥𞤦𞤵𞥅𞤪𞤫', 'fi': 'embu', 'fil': 'Embu', 'fo': 'embu', 'fr': 'embu', 'fy': 'Embu', 'ga': 'Ciambúis', 'gd': 'Embu', 'gl': 'embu', 'gu': 'ઍમ્બુ', 'ha': 'Embu', 'he': 'אמבו', 'hi': 'एम्बु', 'hi-Latn': 'Embu', 'hr': 'embu', 'hsb': 'embu', 'ht': 'embu', 'hu': 'embu', 'hy': 'էմբու', 'ia': 'embu', 'id': 'Embu', 'ig': 'Embu', 'is': 'embu', 'it': 'embu', 'ja': 'エンブ語', 'jv': 'Embu', 'ka': 'ემბუ', 'kea': 'embu', 'kgp': 'ẽmu', 'kk': 'эмбу тілі', 'kk-Cyrl': 'эмбу тілі', 'km': 'អេមប៊ូ', 'kn': 'ಎಂಬು', 'ko': '엠부어', 'kok': 'एम्बु', 'kok-Deva': 'एम्बु', 'ksh': 'Kîembu', 'ku': 'embuyî', 'ky': 'эмбуча', 'lb': 'Kiembu', 'lo': 'ເອັມບູ', 'lrc': 'ئمبو', 'lt': 'embu', 'lv': 'kjembu', 'mai': 'एमबु', 'mi': 'Emepū', 'mk': 'ембу', 'ml': 'എംബു', 'mn': 'эмбу', 'mr': 'एम्बू', 'ms': 'Embu', 'mt': 'Embu', 'my': 'အမ်ဘူ', 'mzn': 'امبو', 'nb': 'kiembu', 'ne': 'एम्बु', 'nl': 'Embu', 'nn': 'embu', 'no': 'kiembu', 'nqo': 'ߋ߲ߓߎߞߊ߲', 'or': 'ଏମ୍ଵୁ', 'pa': 'ਇੰਬੂ', 'pa-Guru': 'ਇੰਬੂ', 'pcm': 'Ẹmbu Lángwej', 'pl': 'embu', 'ps': 'ايمبو', 'pt': 'embu', 'qu': 'Embu Simi', 'rm': 'embu', 'ro': 'embu', 'ru': 'эмбу', 'sc': 'embu', 'sd': 'ايمبيو', 'sd-Arab': 'ايمبيو', 'si': 'එම්බු', 'sk': 'embu', 'sl': 'embujščina', 'smn': 'embukielâ', 'so': 'Embu', 'sq': 'embuisht', 'sr': 'ембу', 'sr-Cyrl': 'ембу', 'sr-Latn': 'embu', 'sv': 'embu', 'sw': 'Kiembu', 'ta': 'எம்பு', 'te': 'ఇంబు', 'th': 'เอ็มบู', 'ti': 'ኤምቡ', 'tk': 'embu dili', 'to': 'lea fakaʻemipū', 'tr': 'Embu', 'ug': 'ئېمبۇچە', 'uk': 'ембу', 'ur': 'امبو', 'uz': 'embu', 'uz-Cyrl': 'эмбуча', 'uz-Latn': 'embu', 'vec': 'enbu', 'vi': 'Tiếng Embu', 'xnr': 'एम्बु', 'yo': 'Èdè Ẹmbù', 'yrl': 'ẽbu', 'yue': '恩布文', 'yue-Hans': '恩布文', 'yue-Hant': '恩布文', 'zh': '恩布语', 'zh-Hans': '恩布语', 'zh-Hant': '恩布文', 'zu': 'isi-Embu'},
    'ecr': {'en': 'Eteocretan'},
    'ecs': {'en': 'Ecuadorian Sign Language'},
    'ecy': {'en': 'Eteocypriot'},
    'ee': {'af': 'Ewe', 'am': 'ኢዊ', 'ar': 'الإيوي', 'as': 'ইৱে', 'ast': 'ewe', 'az': 'eve', 'az-Arab': 'اِوه\u200cجه', 'az-Cyrl': 'еве', 'az-Latn': 'eve', 'bal-Latn': 'Ewe', 'be': 'эве', 'be-tarask': 'эве', 'bg': 'еве', 'bgn': 'اوه\u200cای', 'bn': 'ইউয়ি', 'br': 'ewe', 'brx': 'एवे', 'bs': 'eve', 'bs-Cyrl': 'еве', 'bs-Latn': 'eve', 'ca': 'ewe', 'ca-ES-valencia': 'ewe', 'ccp': '𑄃𑄨𑄅𑄠𑄨', 'ce': 'эве', 'chr': 'ᎡᏪ', 'ckb': 'ئێوێیی', 'cs': 'eweština', 'cy': 'Ewe', 'da': 'ewe', 'de': 'Ewe', 'dsb': 'ewe', 'ee': 'eʋegbe', 'el': 'Έουε', 'el-polyton': 'Έουε', 'en': 'Ewe', 'es': 'ewé', 'et': 'eve', 'eu': 'eweera', 'fa': 'اوه\u200cای', 'ff-Adlm': '𞤉𞤱𞤫𞥅𞤪𞤫', 'fi': 'ewe', 'fil': 'Ewe', 'fo': 'ewe', 'fr': 'éwé', 'fy': 'Ewe', 'ga': 'Éabhais', 'gd': 'Ewe', 'gl': 'ewe', 'gsw': 'Ewe', 'gu': 'ઈવ', 'ha': 'Ewe', 'he': 'אווה', 'hi': 'ईवे', 'hi-Latn': 'Ewe', 'hr': 'ewe', 'hsb': 'ewe', 'ht': 'éwé', 'hu': 'eve', 'hy': 'էվե', 'ia': 'ewe', 'id': 'Ewe', 'ig': 'Ewe', 'is': 'ewe', 'it': 'ewe', 'ja': 'エウェ語', 'jv': 'Ewe', 'ka': 'ევე', 'kea': 'eve', 'kgp': 'eve', 'kk': 'эве тілі', 'kk-Cyrl': 'эве тілі', 'km': 'អ៊ីវ', 'kn': 'ಈವ್', 'ko': '에웨어', 'kok': 'एव', 'kok-Deva': 'एव', 'ks': 'ایٖو', 'ks-Arab': 'ایٖو', 'ksh': 'Ewe', 'ku': 'eweyî', 'ky': 'эбече', 'lb': 'Ewe-Sprooch', 'lld': 'ewe', 'lo': 'ອິວາ', 'lrc': 'ئڤئ', 'lt': 'evių', 'lv': 'evu', 'mai': 'इयु', 'mi': 'Ewe', 'mk': 'еве', 'ml': 'യൂ', 'mn': 'эвэ', 'mr': 'एवे', 'ms': 'Ewe', 'mt': 'Ewe', 'my': 'အီဝီ', 'mzn': 'اوه\u200cیی', 'nb': 'ewe', 'ne': 'इवी', 'nl': 'Ewe', 'nn': 'ewe', 'no': 'ewe', 'nqo': 'ߋߥߋߞߊ߲', 'or': 'ଇୱେ', 'pa': 'ਈਵਈ', 'pa-Guru': 'ਈਵਈ', 'pcm': 'Ẹ́wẹ́ Lángwej', 'pl': 'ewe', 'ps': 'ايو', 'pt': 'ewe', 'qu': 'Ewé Simi', 'rm': 'ewe', 'ro': 'ewe', 'ru': 'эве', 'sc': 'ewe', 'sd': 'ايو', 'sd-Arab': 'ايو', 'si': 'ඉව්', 'sk': 'eweština', 'sl': 'evenščina', 'smn': 'ewekielâ', 'so': 'Eewe', 'sq': 'eveisht', 'sr': 'еве', 'sr-Cyrl': 'еве', 'sr-Latn': 'eve', 'sv': 'ewe', 'sw': 'Kiewe', 'ta': 'ஈவ்', 'te': 'యూ', 'th': 'เอเว', 'ti': 'ኢው', 'tk': 'ewe dili', 'to': 'lea fakaʻeue', 'tr': 'Ewe', 'ug': 'ئېۋېچە', 'uk': 'еве', 'ur': 'ایو', 'uz': 'eve', 'uz-Cyrl': 'эвеча', 'uz-Latn': 'eve', 'vec': 'ewe', 'vi': 'Tiếng Ewe', 'xnr': 'ईवे', 'yo': 'Èdè Ewè', 'yrl': 'ewe', 'yue': '埃維文', 'yue-Hans': '埃维文', 'yue-Hant': '埃維文', 'zh': '埃维语', 'zh-Hans': '埃维语', 'zh-Hant': '埃維文', 'zu': 'isi-Ewe'},
    'eee': {'en': 'E'},
    'efa': {'en': 'Efai'},
    'efe': {'en': 'Efe'},
    'efi': {'af': 'Efik', 'am': 'ኤፊክ', 'ar': 'الإفيك', 'as': 'এফিক', 'ast': 'efik', 'az': 'efik', 'az-Cyrl': 'ефик', 'az-Latn': 'efik', 'be': 'эфік', 'be-tarask': 'эфік', 'bg': 'ефик', 'bgn': 'ایفیکی', 'bn': 'এফিক', 'br': 'efik', 'brx': 'एफिक', 'bs': 'efik', 'bs-Cyrl': 'ефикски', 'bs-Latn': 'efik', 'ca': 'efik', 'ca-ES-valencia': 'efik', 'ccp': '𑄆𑄜𑄨𑄇𑄴', 'ce': 'эфик', 'chr': 'ᎡᏫᎩ', 'ckb': 'ئێفیک', 'cs': 'efikština', 'cy': 'Efik', 'da': 'efik', 'de': 'Efik', 'dsb': 'efikšćina', 'ee': 'efigbe', 'el': 'Εφίκ', 'el-polyton': 'Εφίκ', 'en': 'Efik', 'eo': 'ibibioefika', 'es': 'efik', 'et': 'efiki', 'eu': 'efikera', 'fa': 'افیکی', 'ff-Adlm': '𞤉𞤬𞤭𞤳𞤪𞤫', 'fi': 'efik', 'fil': 'Efik', 'fo': 'efik', 'fr': 'éfik', 'fy': 'Efik', 'ga': 'Eificis', 'gd': 'Efik', 'gl': 'efik', 'gsw': 'Efikisch', 'gu': 'એફિક', 'ha': 'Efik', 'he': 'אפיק', 'hi': 'एफिक', 'hi-Latn': 'Efik', 'hr': 'efik', 'hsb': 'efikšćina', 'ht': 'éfik', 'hu': 'efik', 'hy': 'էֆիկ', 'ia': 'efik', 'id': 'Efik', 'ig': 'Efik', 'is': 'efík', 'it': 'efik', 'ja': 'エフィク語', 'jv': 'Efik', 'ka': 'ეფიკი', 'kgp': 'efike', 'kk': 'эфик тілі', 'kk-Cyrl': 'эфик тілі', 'km': 'អ៊ីហ្វិក', 'kn': 'ಎಫಿಕ್', 'ko': '이픽어', 'kok': 'एफीक', 'kok-Deva': 'एफीक', 'ks': 'ایفِک', 'ks-Arab': 'ایفِک', 'ksh': 'Äffiksch', 'ku': 'efîkî', 'ky': 'эфикче', 'lb': 'Efik', 'lo': 'ອີຟິກ', 'lt': 'efik', 'lv': 'efiku', 'mai': 'एफिक', 'mi': 'Ewhiki', 'mk': 'ефик', 'ml': 'എഫിക്', 'mn': 'эфик', 'mr': 'एफिक', 'ms': 'Efik', 'mt': 'Efik', 'my': 'အာဖိခ်', 'nb': 'efik', 'ne': 'एफिक', 'nl': 'Efik', 'nn': 'efik', 'no': 'efik', 'or': 'ଏଫିକ୍', 'pa': 'ਐਫਿਕ', 'pa-Guru': 'ਐਫਿਕ', 'pcm': 'Ẹ́fík Lángwej', 'pl': 'efik', 'ps': 'افک', 'pt': 'efique', 'qu': 'Efik Simi', 'rm': 'efik', 'ro': 'efik', 'ru': 'эфик', 'sc': 'efik', 'sd': 'ايفڪ', 'sd-Arab': 'ايفڪ', 'si': 'එෆික්', 'sk': 'efik', 'sl': 'efiščina', 'smn': 'efikkielâ', 'so': 'Efik', 'sq': 'efikisht', 'sr': 'ефички', 'sr-Cyrl': 'ефички', 'sr-Latn': 'efički', 'sv': 'efik', 'sw': 'Kiefik', 'ta': 'எஃபிக்', 'te': 'ఎఫిక్', 'th': 'อีฟิก', 'ti': 'ኤፊክ', 'tk': 'efik dili', 'to': 'lea fakaʻefiki', 'tr': 'Efik', 'ug': 'ئېفىكچە', 'uk': 'ефік', 'ur': 'ایفِک', 'uz': 'efik', 'uz-Cyrl': 'эфик', 'uz-Latn': 'efik', 'vec': 'efik', 'vi': 'Tiếng Efik', 'wae': 'Efik', 'xnr': 'एफिक', 'yo': 'Èdè Efiki', 'yrl': 'efiki', 'yue': '埃菲克文', 'yue-Hans': '埃菲克文', 'yue-Hant': '埃菲克文', 'zh': '埃菲克语', 'zh-Hans': '埃菲克语', 'zh-Hant': '埃菲克文', 'zu': 'isi-Efik'},
    'ega': {'en': 'Ega'},
    'egl': {'ast': 'emilianu', 'ca': 'emilià', 'ca-ES-valencia': 'emilià', 'cs': 'emilijština', 'de': 'Emilianisch', 'en': 'Emilian', 'et': 'emiilia', 'fi': 'emilia', 'fr': 'émilien', 'hi-Latn': 'Emilian', 'ht': 'émilien', 'it': 'emiliano', 'ja': 'エミリア語', 'ksh': 'Emilijahnesch', 'lb': 'Emilianesch', 'lt': 'italų kalbos Emilijos tarmė', 'mk': 'емилијански', 'nb': 'emiliansk', 'ne': 'एमिलियाली', 'nl': 'Emiliano', 'nn': 'emiliansk', 'no': 'emiliansk', 'pl': 'emilijski', 'sv': 'emiliska', 'th': 'เอมีเลีย', 'to': 'lea fakaʻemilia', 'tr': 'Emilia Dili', 'vi': 'Tiếng Emilia', 'yue': '埃米利安文', 'yue-Hans': '埃米利安文', 'yue-Hant': '埃米利安文', 'zh-Hant': '埃米利安文'},
    'ego': {'en': 'Eggon'},
    'egx': {'en': 'Egyptian languages'},
    'egy': {'af': 'Antieke Egipties', 'am': 'የጥንታዊ ግብጽኛ', 'ar': 'المصرية القديمة', 'ast': 'exipciu antiguu', 'az': 'qədim misir', 'az-Latn': 'qədim misir', 'be': 'старажытнаегіпецкая', 'be-tarask': 'старажытнаегіпецкая', 'bg': 'древноегипетски', 'bn': 'প্রাচীন মিশরীয়', 'br': 'henegipteg', 'brx': 'प्राचीन मिस्री', 'bs': 'staroegipatski', 'bs-Cyrl': 'староегипатски', 'bs-Latn': 'staroegipatski', 'ca': 'egipci antic', 'ca-ES-valencia': 'egipci antic', 'ccp': '𑄛𑄪𑄢𑄨𑄚𑄩 𑄟𑄨𑄥𑄧𑄢𑄩𑄠𑄧', 'cs': 'egyptština stará', 'cy': 'Hen Eiffteg', 'da': 'oldegyptisk', 'de': 'Ägyptisch', 'el': 'Αρχαία Αιγυπτιακά', 'el-polyton': 'Αἰγυπτιακὰ (ἀρχαῖα)', 'en': 'Ancient Egyptian', 'en-Dsrt': '𐐁𐑌𐐽𐐲𐑌𐐻 𐐀𐐾𐐮𐐹𐐽𐐲𐑌', 'es': 'egipcio antiguo', 'et': 'egiptuse', 'fa': 'مصری کهن', 'fi': 'muinaisegypti', 'fr': 'égyptien ancien', 'fur': 'vieri egjizian', 'fy': 'Aldegyptysk', 'ga': 'Sean-Éigiptis', 'gd': 'Èipheitis Àrsaidh', 'gl': 'exipcio antigo', 'gsw': 'Altägyptisch', 'gu': 'પ્રાચીન ઇજીપ્શિયન', 'he': 'מצרית עתיקה', 'hi': 'प्राचीन मिस्री', 'hi-Latn': 'Ancient Egyptian', 'hr': 'staroegipatski', 'ht': 'égyptien ancien', 'hu': 'óegyiptomi', 'hy': 'հին եգիպտերեն', 'id': 'Mesir Kuno', 'is': 'fornegypska', 'it': 'egiziano antico', 'ja': '古代エジプト語', 'ka': 'ძველეგვიპტური', 'kgp': 'ejimso arkajku', 'kn': 'ಪ್ರಾಚೀನ ಈಜಿಪ್ಟಿಯನ್', 'ko': '고대 이집트어', 'ks': 'قدیٖمی مِصری', 'ks-Arab': 'قدیٖمی مِصری', 'ksh': 'Ahl Äjiptesch', 'lb': 'Egyptesch', 'lo': 'ອີຢິບບູຮານ', 'lt': 'senovės egiptiečių', 'lv': 'ēģiptiešu', 'mk': 'староегипетски', 'ml': 'പ്രാചീന ഈജിപ്ഷ്യൻ', 'mr': 'प्राचीन इजिप्शियन', 'mt': 'Eġizzjan (Antik)', 'my': 'ရှေးဟောင်း အီဂျစ်', 'nb': 'gammelegyptisk', 'ne': 'पुरातन इजिप्टी', 'nl': 'Oudegyptisch', 'nn': 'gammalegyptisk', 'no': 'gammelegyptisk', 'or': 'ପ୍ରାଚୀନ୍ ମିଶିରି', 'os': 'рагон египтаг', 'pa': 'ਪੁਰਾਤਨ ਮਿਸਰੀ', 'pa-Guru': 'ਪੁਰਾਤਨ ਮਿਸਰੀ', 'pl': 'staroegipski', 'pt': 'egípcio arcaico', 'rm': 'egipzian vegl', 'ro': 'egipteană veche', 'ru': 'древнеегипетский', 'sk': 'staroegyptčina', 'sl': 'stara egipčanščina', 'sr': 'староегипатски', 'sr-Cyrl': 'староегипатски', 'sr-Latn': 'staroegipatski', 'sv': 'fornegyptiska', 'sw': 'Kimisri', 'ta': 'பண்டைய எகிப்தியன்', 'te': 'ప్రాచీన ఈజిప్షియన్', 'th': 'อียิปต์โบราณ', 'to': 'lea fakaʻisipitemuʻa', 'tr': 'Eski Mısır Dili', 'ug': 'قەدىمكى مىسىرچە', 'uk': 'давньоєгипетська', 'vi': 'Tiếng Ai Cập cổ', 'yrl': 'egipisiu arkaiku', 'yue': '古埃及文', 'yue-Hans': '古埃及文', 'yue-Hant': '古埃及文', 'zh': '古埃及语', 'zh-Hans': '古埃及语', 'zh-Hant': '古埃及文'},
    'ehs': {'en': 'Miyakubo Sign Language'},
    'ehu': {'en': 'Ehueun'},
    'eip': {'en': 'Eipomek'},
    'eit': {'en': 'Eitiep'},
    'eiv': {'en': 'Askopan'},
    'eja': {'en': 'Ejamat'},
    'eka': {'af': 'Ekajuk', 'am': 'ኤካጁክ', 'ar': 'الإكاجك', 'as': 'একাজুক', 'ast': 'ekajuk', 'az': 'ekacuk', 'az-Cyrl': 'екаҹук', 'az-Latn': 'ekacuk', 'be': 'экаджук', 'be-tarask': 'экаджук', 'bg': 'екажук', 'bgn': 'ایکاجوکی', 'bn': 'ইকাজুক', 'br': 'ekajuk', 'brx': 'एकाजुक', 'bs': 'ekajuk', 'bs-Cyrl': 'екајук', 'bs-Latn': 'ekajuk', 'ca': 'ekajuk', 'ca-ES-valencia': 'ekajuk', 'ccp': '𑄃𑄨𑄇𑄎𑄪𑄇𑄴', 'ce': 'экаджук', 'chr': 'ᎨᎧᏧᎧ', 'ckb': 'ئێکاجووک', 'cs': 'ekajuk', 'cy': 'Ekajuk', 'da': 'ekajuk', 'de': 'Ekajuk', 'dsb': 'ekajukšćina', 'el': 'Εκατζούκ', 'el-polyton': 'Εκατζούκ', 'en': 'Ekajuk', 'es': 'ekajuk', 'et': 'ekadžuki', 'eu': 'ekajuka', 'fa': 'اکاجوک', 'ff-Adlm': '𞤉𞤳𞤢𞤶𞤵𞤳𞤪𞤫', 'fi': 'ekajuk', 'fil': 'Ekajuk', 'fo': 'ekajuk', 'fr': 'ékadjouk', 'fy': 'Ekajuk', 'ga': 'Acaidiúcais', 'gd': 'Ekajuk', 'gl': 'ekajuk', 'gsw': 'Ekajuk', 'gu': 'એકાજુક', 'ha': 'Ekajuk', 'he': 'אקיוק', 'hi': 'एकाजुक', 'hi-Latn': 'Ekajuk', 'hr': 'ekajuk', 'hsb': 'ekajukšćina', 'ht': 'ékadjouk', 'hu': 'ekadzsuk', 'hy': 'էկաջուկ', 'ia': 'ekajuk', 'id': 'Ekajuk', 'ig': 'Ekajuk', 'is': 'ekajúk', 'it': 'ekajuka', 'ja': 'エカジュク語', 'jv': 'Ekajuk', 'ka': 'ეკაჯუკი', 'kgp': 'ekajuki', 'kk': 'экаджук тілі', 'kk-Cyrl': 'экаджук тілі', 'km': 'អ៊ីកាជុក', 'kn': 'ಎಕಾಜುಕ್', 'ko': '이카죽어', 'kok': 'एकाजुक', 'kok-Deva': 'एकाजुक', 'ks': 'ایکاجُک', 'ks-Arab': 'ایکاجُک', 'ksh': 'Ekajukesch', 'ku': 'ekajukî', 'ky': 'экажукча', 'lb': 'Ekajuk', 'lo': 'ອີກາຈັກ', 'lt': 'ekajuk', 'lv': 'ekadžuku', 'mai': 'एकाजुक', 'mi': 'Ekatika', 'mk': 'екаџук', 'ml': 'എകാജുക്', 'mn': 'экажук', 'mr': 'एकाजुक', 'ms': 'Ekajuk', 'mt': 'Ekajuk', 'my': 'အီကာဂျုခ်', 'nb': 'ekajuk', 'ne': 'एकाजुक', 'nl': 'Ekajuk', 'nn': 'ekajuk', 'no': 'ekajuk', 'or': 'ଏକାଜୁକ୍', 'pa': 'ਏਕਾਜੁਕ', 'pa-Guru': 'ਏਕਾਜੁਕ', 'pcm': 'Ẹkajúk Lángwej', 'pl': 'ekajuk', 'ps': 'اکجک', 'pt': 'ekajuk', 'qu': 'Ekajuk Simi', 'rm': 'ekajuk', 'ro': 'ekajuk', 'ru': 'экаджук', 'sc': 'ekajuk', 'sd': 'ايڪاجڪ', 'sd-Arab': 'ايڪاجڪ', 'si': 'එකජුක්', 'sk': 'ekadžuk', 'sl': 'ekajuk', 'smn': 'ekajuk', 'so': 'Ekajuk', 'sq': 'ekajukisht', 'sr': 'екаџук', 'sr-Cyrl': 'екаџук', 'sr-Latn': 'ekadžuk', 'sv': 'ekajuk', 'sw': 'Kiekajuk', 'ta': 'ஈகாஜுக்', 'te': 'ఏకాజక్', 'th': 'อีกาจุก', 'ti': 'ኤካጁክ', 'tk': 'ekajuk dili', 'to': 'lea fakaʻekaiuki', 'tr': 'Ekajuk', 'ug': 'ئېكاجۇكچە', 'uk': 'екаджук', 'ur': 'ایکاجوک', 'uz': 'ekajuk', 'uz-Cyrl': 'экажук', 'uz-Latn': 'ekajuk', 'vec': 'ekajuk', 'vi': 'Tiếng Ekajuk', 'xnr': 'एकाजुक', 'yo': 'Èdè Ekaju', 'yrl': 'ekayuki', 'yue': '艾卡朱克文', 'yue-Hans': '艾卡朱克文', 'yue-Hant': '艾卡朱克文', 'zh': '艾卡朱克语', 'zh-Hans': '艾卡朱克语', 'zh-Hant': '艾卡朱克文', 'zu': 'isi-Ekajuk'},
    'ekc': {'en': 'Eastern Karnic'},
    'eke': {'en': 'Ekit'},
    'ekg': {'en': 'Ekari'},
    'eki': {'en': 'Eki'},
    'ekk': {'en': 'Standard Estonian'},
    'ekl': {'en': 'Kol (Bangladesh)'},
    'ekm': {'en': 'Elip'},
    'eko': {'en': 'Koti'},
    'ekp': {'en': 'Ekpeye'},
    'ekr': {'en': 'Yace'},
    'eky': {'en': 'Eastern Kayah'},
    'el': {'af': 'Grieks', 'agq': 'Gɨ̀lêʔ', 'ak': 'Greek kasa', 'am': 'ግሪክኛ', 'ar': 'اليونانية', 'as': 'গ্ৰীক', 'asa': 'Kigiriki', 'ast': 'griegu', 'az': 'yunan', 'az-Arab': 'یونانجا', 'az-Cyrl': 'јунан', 'az-Latn': 'yunan', 'bal-Latn': 'Yunáni', 'bas': 'Hɔp u gri ᷇kyà', 'be': 'грэчаская', 'be-tarask': 'грэчаская', 'bem': 'Ichi Griki', 'bez': 'Higiriki', 'bg': 'гръцки', 'bgn': 'یونانی', 'bm': 'gɛrɛsikan', 'bn': 'গ্রিক', 'br': 'gresianeg', 'brx': 'ग्रीक', 'bs': 'grčki', 'bs-Cyrl': 'грчки', 'bs-Latn': 'grčki', 'ca': 'grec', 'ca-ES-valencia': 'grec', 'ccp': '𑄉𑄳𑄢𑄨𑄇𑄴', 'ce': 'грекийн', 'cgg': 'Oruguriiki', 'chr': 'ᎠᏂᎪᎢ', 'ckb': 'یۆنانی', 'cs': 'řečtina', 'cy': 'Groeg', 'da': 'græsk', 'dav': 'Kigiriki', 'de': 'Griechisch', 'dje': 'Grek senni', 'dsb': 'grichišćina', 'dyo': 'greek', 'dz': 'གྲིཀ་ཁ', 'ebu': 'Kĩngriki', 'ee': 'grisigbe', 'el': 'Ελληνικά', 'el-polyton': 'Ἑλληνικά', 'en': 'Greek', 'en-Dsrt': '𐐘𐑉𐐨𐐿', 'eo': 'greka', 'es': 'griego', 'et': 'kreeka', 'eu': 'greziera', 'ewo': 'Ǹkɔ́bɔ gəlɛ́g', 'fa': 'یونانی', 'ff': 'Gerke', 'ff-Adlm': '𞤘𞤭𞥅𞤪𞤧𞤢𞥄𞤪𞤫', 'ff-Latn': 'Gerke', 'fi': 'kreikka', 'fil': 'Greek', 'fo': 'grikskt', 'fr': 'grec', 'fur': 'grêc', 'fy': 'Gryks', 'ga': 'Gréigis', 'gd': 'Greugais', 'gl': 'grego', 'gsw': 'Griechisch', 'gu': 'ગ્રીક', 'guz': 'Kigiriki', 'ha': 'Girkanci', 'haw': 'Helene', 'he': 'יוונית', 'hi': 'यूनानी', 'hi-Latn': 'Greek', 'hr': 'grčki', 'hsb': 'grjekšćina', 'ht': 'grec', 'hu': 'görög', 'hy': 'հունարեն', 'ia': 'greco', 'id': 'Yunani', 'ie': 'grec', 'ig': 'Grik', 'is': 'gríska', 'it': 'greco', 'ja': 'ギリシャ語', 'jgo': 'Ŋgɛlɛ̂k', 'jmc': 'Kyigiriki', 'jv': 'Yunani', 'ka': 'ბერძნული', 'kab': 'Tagrikit', 'kam': 'Kigiriki', 'kde': 'Chigilichi', 'kea': 'gregu', 'kgp': 'gregu', 'khq': 'Grek senni', 'ki': 'Kigiriki', 'kk': 'грек тілі', 'kk-Cyrl': 'грек тілі', 'kln': 'kutitab Greece', 'km': 'ក្រិក', 'kn': 'ಗ್ರೀಕ್', 'ko': '그리스어', 'kok': 'ग्रीक', 'kok-Deva': 'ग्रीक', 'kok-Latn': 'Grik', 'ks': 'یوٗنٲنی', 'ks-Arab': 'یوٗنٲنی', 'ksb': 'Kigiiki', 'ksf': 'rigrɛ́k', 'ksh': 'Jrihschesch', 'ku': 'yûnanî', 'kxv': 'grīk', 'kxv-Deva': 'ग्रीक', 'kxv-Latn': 'grīk', 'kxv-Orya': 'ଗ୍ରୀକ୍', 'kxv-Telu': 'గ్రిక్', 'ky': 'грекче', 'lag': 'Kɨgiríki', 'lb': 'Griichesch', 'lg': 'Lugereeki/Luyonaani', 'lkt': 'Greece Iyápi', 'lld': 'grech', 'ln': 'ligeleki', 'lo': 'ກຣີກ', 'lrc': 'یوٙنانی', 'lt': 'graikų', 'lu': 'Giliki', 'luo': 'Kigiriki', 'luy': 'Kigiriki', 'lv': 'grieķu', 'mai': 'यूनानी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lgiriki', 'mer': 'Kĩngiriki', 'mfe': 'grek', 'mg': 'Grika', 'mgh': 'Igiriki', 'mi': 'Kariki', 'mk': 'грчки', 'ml': 'ഗ്രീക്ക്', 'mn': 'грек', 'mr': 'ग्रीक', 'ms': 'Greek', 'mt': 'Grieg', 'mua': 'grek', 'my': 'ဂရိ', 'mzn': 'یونانی', 'naq': 'Xriks', 'nb': 'gresk', 'nd': 'isi-Giliki', 'ne': 'ग्रीक', 'nl': 'Grieks', 'nmg': 'Kiɛl bó grɛk', 'nn': 'gresk', 'no': 'gresk', 'nqo': 'ߜ߭ߙߍ߬ߞߌ߬ߞߊ߲', 'nus': 'Thok girikni', 'nyn': 'Oruguriiki', 'om': 'Afaan Giriiki', 'or': 'ଗ୍ରୀକ୍', 'os': 'бердзейнаг', 'pa': 'ਯੂਨਾਨੀ', 'pa-Guru': 'ਯੂਨਾਨੀ', 'pcm': 'Grík Lángwej', 'pl': 'grecki', 'ps': 'یوناني', 'pt': 'grego', 'qu': 'Griego Simi', 'rm': 'grec', 'rn': 'Ikigereki', 'ro': 'greacă', 'rof': 'Kigiriki', 'ru': 'греческий', 'rw': 'Ikigereki', 'rwk': 'Kyigiriki', 'sah': 'Гириэктии', 'saq': 'Kigiriki', 'sbp': 'Ishigiliki', 'sc': 'grecu', 'sd': 'يوناني', 'sd-Arab': 'يوناني', 'se': 'greikkagiella', 'seh': 'grego', 'ses': 'Grek senni', 'sg': 'Gerêki', 'shi': 'ⵜⴰⴳⵔⵉⴳⵉⵜ', 'shi-Latn': 'Tagrigit', 'shi-Tfng': 'ⵜⴰⴳⵔⵉⴳⵉⵜ', 'si': 'ග්\u200dරීක', 'sk': 'gréčtina', 'sl': 'grščina', 'smn': 'kreikakielâ', 'sn': 'chiGreek', 'so': 'Giriik', 'sq': 'greqisht', 'sr': 'грчки', 'sr-Cyrl': 'грчки', 'sr-Latn': 'grčki', 'sv': 'grekiska', 'sw': 'Kigiriki', 'syr': 'ܝܘܢܐܝܬ', 'ta': 'கிரேக்கம்', 'te': 'గ్రీక్', 'teo': 'Kigiriki', 'tg': 'юнонӣ', 'th': 'กรีก', 'ti': 'ግሪኽኛ', 'tk': 'grek dili', 'to': 'lea fakakalisi', 'tr': 'Yunanca', 'tt': 'грек', 'twq': 'Grek senni', 'tzm': 'Tayunanit', 'ug': 'گىرېكچە', 'uk': 'грецька', 'ur': 'یونانی', 'uz': 'grek', 'uz-Cyrl': 'грекча', 'uz-Latn': 'grek', 'vai': 'ꗥꗷꘋ', 'vai-Latn': 'Hɛlɛŋ', 'vai-Vaii': 'ꗥꗷꘋ', 'vec': 'grego', 'vi': 'Tiếng Hy Lạp', 'vun': 'Kyigiriki', 'wae': 'Gričiš', 'wo': 'Gereg', 'xnr': 'यूनानी', 'xog': 'Oluyonaani', 'yav': 'yavánɛ', 'yi': 'גריכיש', 'yo': 'Èdè Giriki', 'yrl': 'geregu', 'yue': '希臘文', 'yue-Hans': '希腊文', 'yue-Hant': '希臘文', 'zgh': 'ⵜⴰⴳⵔⵉⴳⵉⵜ', 'zh': '希腊语', 'zh-Hans': '希腊语', 'zh-Hant': '希臘文', 'zu': 'isi-Greek'},
    'ele': {'en': 'Elepi'},
    'elh': {'en': 'El Hugeirat'},
    'eli': {'en': 'Nding'},
    'elk': {'en': 'Elkei'},
    'elm': {'en': 'Eleme'},
    'elo': {'en': 'El Molo'},
    'elp': {'en': 'Elpaputih'},
    'elu': {'en': 'Elu'},
    'elx': {'ar': 'الإمايت', 'ast': 'elamita', 'az': 'elamit', 'az-Latn': 'elamit', 'bg': 'еламитски', 'bn': 'এলামাইট', 'br': 'elameg', 'brx': 'एलामी', 'bs': 'elamitski', 'bs-Cyrl': 'еламитски', 'bs-Latn': 'elamitski', 'ca': 'elamita', 'ca-ES-valencia': 'elamita', 'ccp': '𑄆𑄣𑄟𑄭𑄖𑄴', 'cs': 'elamitština', 'cy': 'Elameg', 'da': 'elamitisk', 'de': 'Elamisch', 'el': 'Ελαμάιτ', 'el-polyton': 'Ελαμάιτ', 'en': 'Elamite', 'es': 'elamita', 'et': 'eelami', 'fa': 'عیلامی', 'fi': 'elami', 'fr': 'élamite', 'fy': 'Elamitysk', 'gd': 'Elamais', 'gsw': 'Elamisch', 'gu': 'એલામાઇટ', 'he': 'עילמית', 'hi': 'एलामाइट', 'hi-Latn': 'Elamite', 'hr': 'elamitski', 'ht': 'élamite', 'hu': 'elamit', 'id': 'Elam', 'is': 'elamít', 'it': 'elamitico', 'ja': 'エラム語', 'kgp': 'eramĩte', 'kn': 'ಎಲಾಮೈಟ್', 'ko': '엘람어', 'ks': 'ایلامایِٹ', 'ks-Arab': 'ایلامایِٹ', 'lb': 'Elamesch', 'lo': 'ອີລາໄມ', 'lt': 'elamitų', 'lv': 'elamiešu', 'mk': 'еламски', 'ml': 'എലാമൈറ്റ്', 'mr': 'एलामाइट', 'mt': 'Elamit', 'nb': 'elamittisk', 'ne': 'एलामाइट', 'nl': 'Elamitisch', 'nn': 'elamite', 'no': 'elamittisk', 'or': 'ଏଲାମାଇଟ୍', 'pl': 'elamicki', 'pt': 'elamite', 'rm': 'elamitic', 'ro': 'elamită', 'ru': 'эламский', 'sk': 'elamčina', 'sl': 'elamščina', 'sr': 'еламитски', 'sr-Cyrl': 'еламитски', 'sr-Latn': 'elamitski', 'sv': 'elamitiska', 'ta': 'எலமைட்', 'te': 'ఎలామైట్', 'th': 'อีลาไมต์', 'to': 'lea fakaʻelamite', 'tr': 'Elam', 'ug': 'ئېلامىتچە', 'uk': 'еламська', 'vi': 'Tiếng Elamite', 'yrl': 'eramite', 'yue': '埃蘭文', 'yue-Hans': '埃兰文', 'yue-Hant': '埃蘭文', 'zh': '埃兰语', 'zh-Hans': '埃兰语', 'zh-Hant': '埃蘭文'},
    'ema': {'en': 'Emai-Iuleha-Ora'},
    'emb': {'en': 'Embaloh'},
    'eme': {'en': 'Emerillon'},
    'emg': {'en': 'Eastern Meohang'},
    'emi': {'en': 'Mussau-Emira'},
    'emk': {'en': 'Eastern Maninkakan'},
    'emm': {'en': 'Mamulique'},
    'emn': {'en': 'Eman'},
    'emo': {'en': 'Emok'},
    'emp': {'en': 'Northern Emberá'},
    'emq': {'en': 'Eastern Minyag'},
    'ems': {'en': 'Pacific Gulf Yupik'},
    'emu': {'en': 'Eastern Muria'},
    'emw': {'en': 'Emplawas'},
    'emx': {'en': 'Erromintxela'},
    'emy': {'en': 'Epigraphic Mayan'},
    'emz': {'en': 'Mbessa'},
    'en': {'af': 'Engels', 'agq': 'Kɨŋgele', 'ak': 'Borɔfo', 'am': 'እንግሊዝኛ', 'ar': 'الإنجليزية', 'as': 'ইংৰাজী', 'asa': 'Kiingeredha', 'ast': 'inglés', 'az': 'ingilis', 'az-Arab': 'اینگیلیزجه', 'az-Cyrl': 'инҝилис', 'az-Latn': 'ingilis', 'bal-Latn': 'Engrézi', 'bas': 'Hɔp u ŋgisì', 'be': 'англійская', 'be-tarask': 'англійская', 'bem': 'Ichi Sungu', 'bez': 'Hiingereza', 'bg': 'английски', 'bgc': 'अंग्रेजी', 'bgn': 'انگریزی', 'bho': 'अंगरेजी', 'blo': 'gɛɛshɩ', 'bm': 'angilɛkan', 'bn': 'ইংরেজি', 'bo': 'དབྱིན་ཇིའི་སྐད།', 'br': 'saozneg', 'brx': 'इंराजी', 'bs': 'engleski', 'bs-Cyrl': 'енглески', 'bs-Latn': 'engleski', 'bss': 'Akáálé', 'ca': 'anglès', 'ca-ES-valencia': 'anglès', 'ccp': '𑄃𑄨𑄁𑄢𑄨𑄎𑄨', 'ce': 'ингалсан', 'ceb': 'English', 'cgg': 'Orungyereza', 'cho': 'English', 'chr': 'ᎩᎵᏏ', 'ckb': 'ئینگلیزی', 'cs': 'angličtina', 'csw': 'ᐊᑲᓈᓰᒧᐏᐣ', 'cv': 'акӑлчан', 'cy': 'Saesneg', 'da': 'engelsk', 'dav': 'Kingereza', 'de': 'Englisch', 'dje': 'Inglisi senni', 'doi': 'अंगरेजी', 'dsb': 'engelšćina', 'dyo': 'angle', 'dz': 'ཨིང་ལིཤ་ཁ', 'ebu': 'Kĩthungu', 'ee': 'iŋlisigbe', 'el': 'Αγγλικά', 'el-polyton': 'Ἀγγλικά', 'en': 'English', 'en-Dsrt': '𐐀𐑍𐑊𐐮𐑇', 'eo': 'angla', 'es': 'inglés', 'et': 'inglise', 'eu': 'ingelesa', 'ewo': 'Ǹkɔ́bɔ éngəlís', 'fa': 'انگلیسی', 'ff': 'Engeleere', 'ff-Adlm': '𞤉𞤲𞤺𞤭𞤤𞤫𞥅𞤪𞤫', 'ff-Latn': 'Engeleere', 'fi': 'englanti', 'fil': 'Ingles', 'fo': 'enskt', 'fr': 'anglais', 'fur': 'inglês', 'fy': 'Ingelsk', 'ga': 'Béarla', 'gaa': 'Blɔfo', 'gd': 'Beurla', 'gl': 'inglés', 'gsw': 'Änglisch', 'gu': 'અંગ્રેજી', 'guz': 'Kingereza', 'ha': 'Turanci', 'haw': 'Pelekānia', 'he': 'אנגלית', 'hi': 'अंग्रेज़ी', 'hi-Latn': 'English', 'hnj': '𞄥𞄴𞄅𞄇𞄉𞄦𞄱𞄊', 'hnj-Hmnp': '𞄥𞄴𞄅𞄇𞄉𞄦𞄱𞄊', 'hr': 'engleski', 'hsb': 'jendźelšćina', 'ht': 'anglais', 'hu': 'angol', 'hy': 'անգլերեն', 'ia': 'anglese', 'id': 'Inggris', 'ie': 'anglesi', 'ig': 'Bekee', 'ii': 'ꑱꇩꉙ', 'is': 'enska', 'it': 'inglese', 'ja': '英語', 'jgo': 'Aŋgɛlúshi', 'jmc': 'Kyingereza', 'jv': 'Inggris', 'ka': 'ინგლისური', 'kab': 'Taglizit', 'kam': 'Kingereza', 'kde': 'Chiingeleza', 'kea': 'ingles', 'kgp': 'ĩnhgrej', 'khq': 'Inglisi senni', 'ki': 'Gĩthungũ', 'kk': 'ағылшын тілі', 'kk-Cyrl': 'ағылшын тілі', 'kkj': 'yaman', 'kln': 'kutitab Uingeresa', 'km': 'អង់គ្លេស', 'kn': 'ಇಂಗ್ಲಿಷ್', 'ko': '영어', 'kok': 'इंग्लीश', 'kok-Deva': 'इंग्लीश', 'kok-Latn': 'Inglix', 'ks': 'اَنگیٖزؠ', 'ks-Arab': 'اَنگیٖزؠ', 'ks-Deva': 'अंगरिज़ी', 'ksb': 'Kiingeeza', 'ksf': 'riingɛrís', 'ksh': 'Änglesch', 'ku': 'îngilîzî', 'kxv': 'ingrajī', 'kxv-Deva': 'इंराजी', 'kxv-Latn': 'ingrajī', 'kxv-Orya': 'ଇଂରାଜୀ', 'kxv-Telu': 'ఇంగ్లిస్', 'ky': 'англисче', 'lag': 'Kɨɨngeréesa', 'lb': 'Englesch', 'lg': 'Lungereza', 'lij': 'ingleise', 'lkt': 'Wašíčuiyapi', 'lld': 'inglesc', 'lmo': 'ingles', 'ln': 'lingɛlɛ́sa', 'lo': 'ອັງກິດ', 'lrc': 'ئینگیلیسی', 'lt': 'anglų', 'lu': 'Lingelesa', 'luo': 'Kingereza', 'luy': 'Lusungu', 'lv': 'angļu', 'mai': 'अंगरेजी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ nkɨ́resa', 'mer': 'Kĩngeretha', 'mfe': 'angle', 'mg': 'Anglisy', 'mgh': 'Ingilishi', 'mi': 'Ingarihi', 'mk': 'англиски', 'ml': 'ഇംഗ്ലീഷ്', 'mn': 'англи', 'mni': 'ইংলিস', 'mni-Beng': 'ইংলিস', 'mr': 'इंग्रजी', 'ms': 'Inggeris', 'mt': 'Ingliż', 'mua': 'zah Anglofoŋ', 'my': 'အင်္ဂလိပ်', 'mzn': 'انگلیسی', 'naq': 'Engels', 'nb': 'engelsk', 'nd': 'isi-Ngisi', 'nds': 'Engelsch', 'ne': 'अङ्ग्रेजी', 'nl': 'Engels', 'nmg': 'Ngɛ̄lɛ̄n', 'nn': 'engelsk', 'nnh': 'ngilísè', 'no': 'engelsk', 'nqo': 'ߊ߲߬ߜߑߟߋ߬ߞߊ߲', 'nso': 'Seisimane', 'nus': 'Thok liŋli̱thni', 'nyn': 'Orungyereza', 'oc': 'anglés', 'om': 'Afaan Ingilizii', 'or': 'ଇଂରାଜୀ', 'os': 'англисаг', 'pa': 'ਅੰਗਰੇਜ਼ੀ', 'pa-Guru': 'ਅੰਗਰੇਜ਼ੀ', 'pcm': 'Ínglish', 'pis': 'Ingglis', 'pl': 'angielski', 'prg': 'ēngliskan', 'ps': 'انګليسي', 'pt': 'inglês', 'qu': 'Ingles Simi', 'raj': 'अंग्रेजी', 'rm': 'englais', 'rn': 'Icongereza', 'ro': 'engleză', 'rof': 'Kiingereza', 'ru': 'английский', 'rw': 'Icyongereza', 'rwk': 'Kyingereza', 'sa': 'आङ्ग्लभाषा', 'sah': 'Ааҥыллыы', 'saq': 'Kingereza', 'sat': 'ᱟᱝᱜᱽᱨᱮᱡᱤ', 'sat-Olck': 'ᱟᱝᱜᱽᱨᱮᱡᱤ', 'sbp': 'Ishingelesa', 'sc': 'inglesu', 'sd': 'انگريزي', 'sd-Arab': 'انگريزي', 'sd-Deva': 'अंगरेज़ी', 'se': 'eaŋgalsgiella', 'seh': 'inglês', 'ses': 'Inglisi senni', 'sg': 'Anglëe', 'shi': 'ⵜⴰⵏⴳⵍⵉⵣⵜ', 'shi-Latn': 'Tanglizt', 'shi-Tfng': 'ⵜⴰⵏⴳⵍⵉⵣⵜ', 'si': 'ඉංග්\u200dරීසි', 'sk': 'angličtina', 'sl': 'angleščina', 'smn': 'eŋgâlâskielâ', 'sn': 'Chirungu', 'so': 'Ingiriisi', 'sq': 'anglisht', 'sr': 'енглески', 'sr-Cyrl': 'енглески', 'sr-Latn': 'engleski', 'st': 'Senyesemane', 'su': 'Inggris', 'su-Latn': 'Inggris', 'sv': 'engelska', 'sw': 'Kiingereza', 'syr': 'ܐܢܓܠܝܬ', 'szl': 'angelski', 'ta': 'ஆங்கிலம்', 'te': 'ఇంగ్లీష్', 'teo': 'Kingereza', 'tg': 'англисӣ', 'th': 'อังกฤษ', 'ti': 'እንግሊዝኛ', 'tk': 'iňlis dili', 'tn': 'Sekgoa', 'to': 'lea fakapālangi', 'tok': 'toki Inli', 'tpi': 'Inglis', 'tr': 'İngilizce', 'tt': 'инглиз', 'twq': 'Inglisi senni', 'tzm': 'Tanglizt', 'ug': 'ئىنگلىزچە', 'uk': 'англійська', 'ur': 'انگریزی', 'uz': 'inglizcha', 'uz-Cyrl': 'инглизча', 'uz-Latn': 'inglizcha', 'vai': 'ꕶꕱ', 'vai-Latn': 'Poo', 'vai-Vaii': 'ꕶꕱ', 'vec': 'ingleze', 'vi': 'Tiếng Anh', 'vmw': 'ingles', 'vun': 'Kyingereza', 'wae': 'Engliš', 'wbp': 'Yinkirliji', 'wo': 'Àngale', 'xh': 'IsiNgesi', 'xnr': 'अंग्रेज़ी', 'xog': 'Olungereza', 'yav': 'íŋgilísé', 'yi': 'ענגליש', 'yo': 'Èdè Gẹ̀ẹ́sì', 'yrl': 'ẽgirixi', 'yue': '英文', 'yue-Hans': '英文', 'yue-Hant': '英文', 'za': 'Yinghyij', 'zgh': 'ⵜⴰⵏⴳⵍⵉⵣⵜ', 'zh': '英语', 'zh-Hans': '英语', 'zh-Hant': '英文', 'zu': 'i-English'},
    'en-AU': {'am': 'የአውስትራሊያ እንግሊዝኛ', 'as': 'অষ্ট্ৰেলিয়ান ইংৰাজী', 'ast': 'inglés d’Australia', 'az': 'Avstraliya ingiliscəsi', 'az-Cyrl': 'Австралија инҝилисҹәси', 'az-Latn': 'Avstraliya ingiliscəsi', 'be': 'аўстралійская англійская', 'be-tarask': 'аўстралійская англійская', 'bg': 'австралийски английски', 'bgn': 'استرالیاین انگریزی', 'br': 'saozneg Aostralia', 'brx': 'अस्ट्रेलियानि इंराजी', 'ca': 'anglès australià', 'ca-ES-valencia': 'anglès australià', 'ccp': '𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄃𑄨𑄁𑄢𑄬𑄎𑄨', 'ce': 'Австралин ингалсан', 'chr': 'ᎡᎳᏗᏜ ᎩᎵᏏ', 'ckb': 'ئینگلیزیی ئۆسترالیایی', 'csw': 'ᐊᑲᓈᓰᒧᐏᐣ (AU)', 'cv': 'австралин акӑлчан', 'cy': 'Saesneg Awstralia', 'da': 'australsk engelsk', 'dsb': 'awstralska engelšćina', 'dz': 'ཨཱོས་ཊྲེ་ལི་ཡཱན་ཨིང་ལིཤ་ཁ', 'ee': 'Yevugbe (Australia)', 'el': 'Αγγλικά Αυστραλίας', 'el-polyton': 'Αγγλικά Αυστραλίας', 'en': 'Australian English', 'es': 'inglés australiano', 'et': 'Austraalia inglise', 'eu': 'Australiako ingelesa', 'fa': 'انگلیسی استرالیا', 'fi': 'australianenglanti', 'fr': 'anglais australien', 'fur': 'inglês australian', 'fy': 'Australysk Ingelsk', 'ga': 'Béarla na hAstráile', 'gd': 'Beurla Astràilia', 'gl': 'inglés australiano', 'gsw': 'Auschtralischs Änglisch', 'gu': 'ઓસ્ટ્રેલિયન અંગ્રેજી', 'ha': 'Turanci Ostareliya', 'haw': 'Pelekāne Nū Hōlani', 'hi': 'ऑस्ट्रेलियाई अंग्रेज़ी', 'hi-Latn': 'Australian English', 'hr': 'australski engleski', 'hsb': 'awstralska jendźelšćina', 'ht': 'anglais australien', 'hu': 'ausztrál angol', 'hy': 'ավստրալիական անգլերեն', 'ia': 'anglese australian', 'ig': 'Bekee ndị Australia', 'is': 'áströlsk enska', 'it': 'inglese australiano', 'ja': 'オーストラリア英語', 'jv': 'Inggris Ostrali', 'ka': 'ავსტრალიური ინგლისური', 'kea': 'ingles australianu', 'kn': 'ಆಸ್ಟ್ರೇಲಿಯನ್ ಇಂಗ್ಲಿಷ್', 'kok': 'ऑस्ट्रेलियन इंग्लीश', 'kok-Deva': 'ऑस्ट्रेलियन इंग्लीश', 'ks': 'آسٹریلیَن اَنگریٖزؠ', 'ks-Arab': 'آسٹریلیَن اَنگریٖزؠ', 'ks-Deva': 'आसट्रेलवी अंगरिज़ी', 'ksh': 'Änglesch uß Außtrahlije', 'kxv': 'ausṭreliati ingarjī', 'kxv-Deva': 'ऑस्ट्रेलियाति इंराजी', 'kxv-Latn': 'ausṭreliati ingarjī', 'kxv-Orya': 'ଅଷ୍ଟ୍ରେଲିୟାତି ଇଂରାଜୀ', 'kxv-Telu': 'అస్ట్రె లియన్ ఇంగ్లిస్', 'lb': 'Australescht Englesch', 'lrc': 'ئینگیلیسی ئوستارالیایی', 'lt': 'Australijos anglų', 'mai': 'आस्ट्रेलियाई अंंगरेजी', 'mi': 'Ingarihi Ahitereiriana', 'mk': 'австралиски англиски', 'ml': 'ഓസ്\u200cട്രേലിയൻ ഇംഗ്ലീഷ്', 'mn': 'австрали-англи', 'mni': 'ওষ্ট্রেলিয়ান ইংলিস', 'mni-Beng': 'ওষ্ট্রেলিয়ান ইংলিস', 'mr': 'ऑस्ट्रेलियन इंग्रजी', 'ms': 'Inggeris Australia', 'mt': 'Ingliż Awstraljan', 'my': 'ဩစတြေးလျ အင်္ဂလိပ်', 'mzn': 'استرالیای ِانگلیسی', 'ne': 'अस्ट्रेलियाली अङ्ग्रेजी', 'nqo': 'ߐߛߑߕߙߊߟߌ߫ ߊ߲߬ߜ߭ߑߟߋ߬ߞߊ߲', 'om': 'Ingiliffa Awustiraaliyaa', 'or': 'ଅଷ୍ଟ୍ରେଲିୟ ଇଂରାଜୀ', 'os': 'австралиаг англисаг', 'pcm': 'Ọstréliá Ínglish', 'pl': 'angielski australijski', 'ps': 'آسټرالياوي انګليسي', 'rm': 'englais australian', 'ru': 'австралийский английский', 'sa': 'ऑस्ट्रेलियादेशः आङ्ग्लभाषा', 'sat': 'ᱚᱥᱴᱨᱮᱞᱤᱭᱟᱱ ᱟᱝᱜᱽᱨᱮᱡᱤ', 'sat-Olck': 'ᱚᱥᱴᱨᱮᱞᱤᱭᱟᱱ ᱟᱝᱜᱽᱨᱮᱡᱤ', 'sc': 'inglesu australianu', 'sd': 'آسٽريليائي انگريزي', 'sd-Arab': 'آسٽريليائي انگريزي', 'sd-Deva': 'ऑसटेलियन अंगरेज़ी', 'si': 'ඕස්ට්\u200dරේලියානු ඉංග්\u200dරීසි', 'sk': 'angličtina (austrálska)', 'sl': 'avstralska angleščina', 'smn': 'Australia eŋgâlâskielâ', 'so': 'Ingiriis Austaraaliyaan', 'sq': 'anglishte australiane', 'su': 'Inggris Australia', 'su-Latn': 'Inggris Australia', 'sv': 'australisk engelska', 'ta': 'ஆஸ்திரேலிய ஆங்கிலம்', 'te': 'ఆస్ట్రేలియన్ ఇంగ్లీష్', 'tg': 'англисии австралиягӣ', 'th': 'อังกฤษ - ออสเตรเลีย', 'to': 'lea fakapālangi-ʻaositelēlia', 'tpi': 'Australian Inglis', 'tr': 'Avustralya İngilizcesi', 'ug': 'ئاۋسترالىيە ئىنگلىزچە', 'ur': 'آسٹریلیائی انگریزی', 'uz': 'ingliz (Avstraliya)', 'uz-Latn': 'ingliz (Avstraliya)', 'vec': 'ingleze australian', 'wae': 'Auštrališes Engliš', 'wo': 'Àngale bu Óstraali', 'xh': 'IsiNgesi Sase-Australia', 'xnr': 'अस्ट्रेलिया आली अंग्रेज़ी', 'yo': 'Èdè Gẹ̀ẹ́sì (órílẹ̀-èdè Ọsirélíà)', 'zh': '澳大利亚英语', 'zh-Hans': '澳大利亚英语', 'zu': 'i-Australian English'},
    'en-CA': {'am': 'የካናዳ እንግሊዝኛ', 'as': 'কানাডিয়ান ইংৰাজী', 'ast': 'inglés de Canadá', 'az': 'Kanada ingiliscəsi', 'az-Cyrl': 'Канада инҝилисҹәси', 'az-Latn': 'Kanada ingiliscəsi', 'bal-Latn': 'Engrézi (Kaynadhá)', 'be': 'канадская англійская', 'be-tarask': 'канадская англійская', 'bg': 'канадски английски', 'bgn': 'کاناڈاین انگریزی', 'bo': 'དབྱིན་ཇིའི་སྐད། (ཁེ་ན་ཌ་)', 'br': 'saozneg Kanada', 'brx': 'कानाडानि इंराजी', 'ca': 'anglès canadenc', 'ca-ES-valencia': 'anglès canadenc', 'ccp': '𑄇𑄚𑄓𑄩𑄠𑄧 𑄃𑄨𑄁𑄢𑄬𑄎𑄨', 'ce': 'канадан ингалсан', 'chr': 'ᎨᎾᏓ ᎩᎵᏏ', 'ckb': 'ئینگلیزیی کەنەدایی', 'csw': 'ᐊᑲᓈᓰᒧᐏᐣ (ᑳᓇᑕ)', 'cv': 'канадӑн акӑлчан', 'cy': 'Saesneg Canada', 'da': 'canadisk engelsk', 'doi': 'कैनेडियन अंगरेजी', 'dsb': 'kanadiska engelšćina', 'dz': 'ཀེ་ན་ཌི་ཡཱན་ཨིང་ལིཤ་ཁ', 'ee': 'Yevugbe (Canada)', 'el': 'Αγγλικά Καναδά', 'el-polyton': 'Αγγλικά Καναδά', 'en': 'Canadian English', 'es': 'inglés canadiense', 'et': 'Kanada inglise', 'eu': 'Kanadako ingelesa', 'fa': 'انگلیسی کانادا', 'fi': 'kanadanenglanti', 'fr': 'anglais canadien', 'fur': 'inglês canadês', 'fy': 'Kanadeesk Ingelsk', 'ga': 'Béarla Cheanada', 'gd': 'Beurla Chanada', 'gl': 'inglés canadense', 'gsw': 'Kanadischs Änglisch', 'gu': 'કેનેડિયન અંગ્રેજી', 'haw': 'Pelekāne Kanakā', 'hi': 'कनाडाई अंग्रेज़ी', 'hi-Latn': 'Canadian English', 'hr': 'kanadski engleski', 'hsb': 'kanadiska jendźelšćina', 'ht': 'anglais canadien', 'hu': 'kanadai angol', 'hy': 'կանադական անգլերեն', 'ia': 'anglese canadian', 'ig': 'Bekee ndị Canada', 'is': 'kanadísk enska', 'it': 'inglese canadese', 'ja': 'カナダ英語', 'jv': 'Inggris Kanada', 'ka': 'კანადური ინგლისური', 'kea': 'ingles kanadianu', 'kn': 'ಕೆನೆಡಿಯನ್ ಇಂಗ್ಲಿಷ್', 'kok': 'कॅनडियन इंग्लीश', 'kok-Deva': 'कॅनडियन इंग्लीश', 'ks': 'کینَڈِیٲیی اَنگریٖزؠ', 'ks-Arab': 'کینَڈِیٲیی اَنگریٖزؠ', 'ks-Deva': 'कनाडियन अंगरिज़ी', 'ksh': 'Änglesch uß Kanada', 'kxv': 'kanaḍati ingarjī', 'kxv-Deva': 'कनाडाति इंराजी', 'kxv-Latn': 'kanaḍati ingarjī', 'kxv-Orya': 'କାନାଡିୟାତି ଇଂରାଜୀ', 'kxv-Telu': 'కనెడయన్ ఇంగ్లిస్', 'lb': 'Kanadescht Englesch', 'lrc': 'ئینگیلیسی کانادایی', 'lt': 'Kanados anglų', 'mai': 'कनाडाई अंग्रेज़ी', 'mi': 'Ingarihi Kānata', 'mk': 'канадски англиски', 'ml': 'കനേഡിയൻ ഇംഗ്ലീഷ്', 'mn': 'канад-англи', 'mni': 'কানাদিয়ান ইংলিস', 'mni-Beng': 'কানাদিয়ান ইংলিস', 'mr': 'कॅनडियन इंग्रजी', 'ms': 'Inggeris Kanada', 'mt': 'Ingliż Kanadiż', 'my': 'ကနေဒါ အင်္ဂလိပ်', 'mzn': 'کانادای ِانگلیسی', 'ne': 'क्यानाडेली अङ्ग्रेजी', 'nqo': 'ߞߣߊߘߊ߫ ߊ߲߬ߜ߭ߑߟߋ߬ߞߊ߲', 'om': 'Ingiliffa Kanaadaa', 'or': 'କାନାଡିୟ ଇଂରାଜୀ', 'os': 'канадӕйаг англисаг', 'pcm': 'Kánáda Ínglish', 'pl': 'angielski kanadyjski', 'ps': 'کاناډايي انګلیسي', 'rm': 'englais canadais', 'ru': 'канадский английский', 'sa': 'कनाडादेशः आङ्ग्लभाषा', 'sat': 'ᱠᱟᱱᱟᱰᱤᱭᱟᱱ ᱟᱝᱜᱽᱨᱮᱡᱤ', 'sat-Olck': 'ᱠᱟᱱᱟᱰᱤᱭᱟᱱ ᱟᱝᱜᱽᱨᱮᱡᱤ', 'sc': 'inglesu canadesu', 'sd': 'ڪينيڊيائي انگريزي', 'sd-Arab': 'ڪينيڊيائي انگريزي', 'sd-Deva': 'केनेडियन अंगरेज़ी', 'si': 'කැනේඩියානු ඉංග්\u200dරීසි', 'sk': 'angličtina (kanadská)', 'sl': 'kanadska angleščina', 'smn': 'Kanada eŋgâlâskielâ', 'so': 'Ingiriis Kanadiyaan', 'sq': 'anglishte kanadeze', 'su': 'Inggris Kanada', 'su-Latn': 'Inggris Kanada', 'sv': 'kanadensisk engelska', 'sw': 'Kiingereza (Canada)', 'ta': 'கனடிய ஆங்கிலம்', 'te': 'కెనడియన్ ఇంగ్లీష్', 'tg': 'англисии канадагӣ', 'th': 'อังกฤษ - แคนาดา', 'to': 'lea fakapālangi-kānata', 'tpi': 'Kenedien Inglis', 'tr': 'Kanada İngilizcesi', 'ug': 'كانادا ئىنگلىزچە', 'ur': 'کینیڈین انگریزی', 'uz': 'ingliz (Kanada)', 'uz-Latn': 'ingliz (Kanada)', 'vec': 'ingleze canadeze', 'wae': 'Kanadišes Engliš', 'wo': 'Àngale bu Kanadaa', 'xh': 'IsiNgesi SaseKhanada', 'xnr': 'कनेडे आली अंग्रेज़ी', 'yo': 'Èdè Gẹ̀ẹ́sì (Orílẹ̀-èdè Kánádà)', 'zh': '加拿大英语', 'zh-Hans': '加拿大英语', 'zu': 'i-Canadian English'},
    'en-GB': {'af': 'Engels (VK)', 'ak': 'Ngresi Borɔfo', 'am': 'የብሪቲሽ እንግሊዝኛ', 'ar': 'الإنجليزية المملكة المتحدة', 'as': 'ব্ৰিটিছ ইংৰাজী', 'ast': 'inglés de Gran Bretaña', 'az': 'Britaniya ingiliscəsi', 'az-Cyrl': 'Британија инҝилисҹәси', 'az-Latn': 'Britaniya ingiliscəsi', 'be': 'брытанская англійская', 'be-tarask': 'брытанская англійская', 'bg': 'британски английски', 'bgn': 'بریتانیاین انگریزی', 'bn': 'ইংরেজি (যুক্তরাজ্য)', 'bo': 'དབྱིན་ཇིའི་སྐད། (དབྱིན་ལན་)', 'br': 'saozneg Breizh-Veur', 'brx': 'ब्रिटिश इंराजी', 'ca': 'anglès britànic', 'ca-ES-valencia': 'anglès britànic', 'ccp': '𑄝𑄳𑄢𑄨𑄑𑄨𑄌𑄴 𑄃𑄨𑄁𑄢𑄬𑄎𑄨', 'ce': 'британин ингалсан', 'ceb': 'Britanikong English', 'chr': 'ᎩᎵᏏᏲ ᎩᎵᏏ', 'ckb': 'ئینگلیزیی بریتانیایی', 'cs': 'angličtina (Velká Británie)', 'csw': 'ᐊᑲᓈᓰᒧᐏᐣ (GB)', 'cv': 'британин акӑлчан', 'cy': 'Saesneg Prydain', 'da': 'britisk engelsk', 'de': 'Englisch (GB)', 'doi': 'ब्रिटिश अंगरेजी', 'dsb': 'britiska engelšćina', 'dz': 'བྲི་ཊིཤ་ཨིང་ལིཤ་ཁ', 'ee': 'Yevugbe (Britain)', 'el': 'Αγγλικά Βρετανίας', 'el-polyton': 'Αγγλικά Βρετανίας', 'en': 'British English', 'es': 'inglés británico', 'et': 'Briti inglise', 'eu': 'Britainia Handiko ingelesa', 'fa': 'انگلیسی بریتانیا', 'ff-Adlm': '𞤉𞤲𞤺𞤭𞤤𞤫𞥅𞤪𞤫 𞤁𞤘', 'fi': 'britannianenglanti', 'fil': 'Ingles sa UK', 'fr': 'anglais britannique', 'fur': 'inglês britanic', 'fy': 'Britsk Ingelsk', 'ga': 'Béarla na Breataine', 'gd': 'Beurla Bhreatainn', 'gl': 'inglés británico', 'gsw': 'Britischs Änglisch', 'gu': 'બ્રિટિશ અંગ્રેજી', 'ha': 'Turanci (Biritaniya)', 'haw': 'Pelekānia Pekekāne', 'hi': 'ब्रिटिश अंग्रेज़ी', 'hi-Latn': 'British English', 'hr': 'britanski engleski', 'hsb': 'britiska jendźelšćina', 'ht': 'anglais britannique', 'hu': 'brit angol', 'hy': 'բրիտանական անգլերեն', 'ia': 'anglese britannic', 'id': 'Inggris (Britania)', 'ig': 'Bekee ndị United Kingdom', 'is': 'bresk enska', 'it': 'inglese britannico', 'ja': 'イギリス英語', 'jv': 'Inggris Karajan Manunggal', 'ka': 'ბრიტანული ინგლისური', 'kea': 'ingles britániku', 'kn': 'ಬ್ರಿಟಿಷ್ ಇಂಗ್ಲಿಷ್', 'kok': 'ब्रिटीश इंग्लीश', 'kok-Deva': 'ब्रिटीश इंग्लीश', 'ks': 'بَرطانوی اَنگریٖزؠ', 'ks-Arab': 'بَرطانوی اَنگریٖزؠ', 'ks-Deva': 'बरतानवी अंगरिज़ी', 'ksh': 'Änglesch uß Jruhßbrettannije', 'ku': 'îngilîzî (Qiralîyeta Yekbûyî)', 'kxv': 'briṭisti ingrajī', 'kxv-Deva': 'ब्रिटिस इंराजी', 'kxv-Latn': 'briṭisti ingrajī', 'kxv-Orya': 'ବ୍ରିଟିଶ୍\u200cତି ଇଂରାଜୀ', 'kxv-Telu': 'బ్రిటిస్ ఇంగ్లిస్', 'lb': 'Britescht Englesch', 'lkt': 'Šagláša Wašíčuiyapi', 'lld': 'inglesc (UK)', 'lo': 'ອັງກິດ (ບຣິດທິຊ)', 'lrc': 'ئینگیلیسی بئریتانیایی', 'lt': 'Didžiosios Britanijos anglų', 'lv': 'angļu (Lielbritānija)', 'mai': 'ब्रिटिश अंग्रेज़ी', 'mi': 'Ingarihi Piritene', 'mk': 'британски англиски', 'ml': 'ബ്രിട്ടീഷ് ഇംഗ്ലീഷ്', 'mn': 'британи-англи', 'mni': 'ব্রিটিশ ইংলিস', 'mni-Beng': 'ব্রিটিশ ইংলিস', 'mr': 'ब्रिटिश इंग्रजी', 'ms': 'Inggeris British', 'mt': 'Ingliż Brittaniku', 'my': 'ဗြိတိသျှ အင်္ဂလိပ်', 'mzn': 'بریتیش انگلیسی', 'ne': 'बेलायती अङ्ग्रेजी', 'nqo': 'ߓߙߌߕߊ߲ߓߊ߫ ߊ߲߬ߜ߭ߑߟߋ߬ߞߊ߲', 'om': 'Ingliffa Biritishii', 'or': 'ବ୍ରିଟିଶ୍\u200c ଇଂରାଜୀ', 'os': 'бритайнаг англисаг', 'pa': 'ਅੰਗਰੇਜ਼ੀ (ਬਰਤਾਨਵੀ)', 'pa-Guru': 'ਅੰਗਰੇਜ਼ੀ (ਬਰਤਾਨਵੀ)', 'pcm': 'Brítísh Ínglish', 'pl': 'angielski brytyjski', 'ps': 'بريتانوی انګلیسي', 'qu': 'Ingles Simi (GB)', 'rm': 'englais britannic', 'ru': 'британский английский', 'sa': 'आङ्ग्लदेशीय आङ्ग्लभाषा:', 'sat': 'ᱵᱨᱤᱴᱤᱥ ᱟᱝᱜᱽᱨᱮᱡᱤ', 'sat-Olck': 'ᱵᱨᱤᱴᱤᱥ ᱟᱝᱜᱽᱨᱮᱡᱤ', 'sc': 'inglesu britànnicu', 'sd': 'برطانوي انگريزي', 'sd-Arab': 'برطانوي انگريزي', 'sd-Deva': 'यूके जी अंगरेज़ी', 'si': 'බ්\u200dරිතාන්\u200dය ඉංග්\u200dරීසි', 'sk': 'angličtina (britská)', 'sl': 'britanska angleščina', 'smn': 'Britannia eŋgâlâskielâ', 'so': 'Ingiriis Biritish', 'sq': 'anglishte britanike', 'sr': 'енглески (Велика Британија)', 'sr-Cyrl': 'енглески (Велика Британија)', 'sr-Latn': 'engleski (Velika Britanija)', 'su': 'Inggris UK', 'su-Latn': 'Inggris UK', 'sv': 'brittisk engelska', 'sw': 'Kiingereza (Uingereza)', 'ta': 'பிரிட்டிஷ் ஆங்கிலம்', 'te': 'బ్రిటిష్ ఇంగ్లీష్', 'tg': 'англисии британӣ', 'th': 'อังกฤษ - สหราชอาณาจักร', 'tk': 'iňlis dili (Beýik Britaniýa)', 'to': 'lea fakapilitānia', 'tpi': 'Britis Inglis', 'tr': 'İngiliz İngilizcesi', 'tt': 'Британия инглизчәсе', 'ug': 'ئەنگلىيە ئىنگلىزچە', 'uk': 'англійська (Брит.)', 'ur': 'برطانوی انگریزی', 'uz': 'ingliz (Britaniya)', 'uz-Cyrl': 'инглизча (Британия)', 'uz-Latn': 'ingliz (Britaniya)', 'vec': 'ingleze britànego', 'vi': 'Tiếng Anh (Anh)', 'wae': 'Britišes Engliš', 'wo': 'Àngale bu Grànd Brëtaañ', 'xh': 'IsiNgesi SaseBritane', 'xnr': 'ब्रिटेने आली अंग्रेज़ी', 'yi': 'ענגליש (GB)', 'yo': 'Èdè òyìnbó Gẹ̀ẹ́sì', 'zh': '英国英语', 'zh-Hans': '英国英语', 'zu': 'i-British English'},
    'en-US': {'af': 'Engels (VSA)', 'ak': 'Amɛrika Borɔfo', 'am': 'የአሜሪካ እንግሊዝኛ', 'ar': 'الإنجليزية الولايات المتحدة', 'as': 'আমেৰিকান ইংৰাজী', 'ast': 'inglés d’Estaos Xuníos', 'az': 'Amerika ingiliscəsi', 'az-Cyrl': 'Америка инҝилисҹәси', 'az-Latn': 'Amerika ingiliscəsi', 'be': 'амерыканская англійская', 'be-tarask': 'амерыканская англійская', 'bg': 'американски английски', 'bgn': 'امریکاین انگریزی', 'blo': 'gɛɛshɩ (Ganɔ gaɖɔŋkɔnɔ kabʊtǝna Amalɩka nɩ)', 'bn': 'ইংরেজি (আমেরিকা)', 'bo': 'དབྱིན་ཇིའི་སྐད། (ཨ་རི་)', 'br': 'saozneg Amerika', 'brx': 'आमेरिकानि इंराजी', 'ca': 'anglès americà', 'ca-ES-valencia': 'anglès americà', 'ccp': '𑄃𑄟𑄬𑄢𑄨𑄇𑄢𑄴 𑄃𑄨𑄁𑄢𑄎𑄨', 'ce': 'американ ингалсан', 'ceb': 'English (America)', 'chr': 'ᎠᎹᏰᏟ ᎩᎵᏏ', 'ckb': 'ئینگلیزیی ئەمەریکایی', 'cs': 'angličtina (USA)', 'csw': 'ᐊᑲᓈᓰᒧᐏᐣ (US)', 'cv': 'америкӑн акӑлчан', 'cy': 'Saesneg America', 'da': 'amerikansk engelsk', 'doi': 'अमरीकी अंगरेजी', 'dsb': 'ameriska engelšćina', 'dz': 'ཡུ་ཨེས་ཨིང་ལིཤ་ཁ', 'ee': 'Yevugbe (America)', 'el': 'Αγγλικά Αμερικής', 'el-polyton': 'Αγγλικά Αμερικής', 'en': 'American English', 'es': 'inglés estadounidense', 'et': 'Ameerika inglise', 'eu': 'ingeles amerikarra', 'fa': 'انگلیسی آمریکا', 'ff-Adlm': '𞤉𞤲𞤺𞤭𞤤𞤫𞥅𞤪𞤫 𞤀𞤥𞤫𞤪𞤭𞤳𞤢𞤲𞤳𞤮𞤪𞤫', 'fi': 'amerikanenglanti', 'fil': 'Ingles (American)', 'fr': 'anglais américain', 'fur': 'ingles merecan', 'fy': 'Amerikaansk Ingelsk', 'ga': 'Béarla Mheiriceá', 'gd': 'Beurla na h-Aimeireaga', 'gl': 'inglés estadounidense', 'gsw': 'Amerikanischs Änglisch', 'gu': 'અમેરિકન અંગ્રેજી', 'ha': 'Turanci (Amurka)', 'haw': 'Pelekānia ʻAmelika', 'hi': 'अमेरिकी अंग्रेज़ी', 'hi-Latn': 'American English', 'hr': 'američki engleski', 'hsb': 'ameriska jendźelšćina', 'ht': 'anglais américain', 'hu': 'amerikai angol', 'hy': 'ամերիկյան անգլերեն', 'ia': 'anglese american', 'ig': 'Bekee ndị America', 'is': 'bandarísk enska', 'it': 'inglese americano', 'ja': 'アメリカ英語', 'jv': 'Inggris Amérika Sarékat', 'ka': 'ამერიკული ინგლისური', 'kea': 'ingles merkanu', 'km': 'អង់គ្លេស (ស.រ.អ.)', 'kn': 'ಅಮೆರಿಕನ್ ಇಂಗ್ಲಿಷ್', 'kok': 'अमेरिकन इंग्लीश', 'kok-Deva': 'अमेरिकन इंग्लीश', 'ks': 'امریٖکی اَنٛگریٖزؠ', 'ks-Arab': 'امریٖکی اَنٛگریٖزؠ', 'ks-Deva': 'अमरीकी अंगरिज़ी', 'ksh': 'Amärrekahnesch Änglesch', 'kxv': 'amerikati ingrajī', 'kxv-Deva': 'अमेरिकी इंराजी', 'kxv-Latn': 'amerikati ingrajī', 'kxv-Orya': 'ଆମେରିକାତି ଇଂରାଜୀ', 'kxv-Telu': 'అమెరికాన్ ఇంగ్లిస్', 'ky': 'англисче (Америка Кошмо Штаттары)', 'lb': 'Amerikanescht Englesch', 'lkt': 'Mílahaŋska Wašíčuiyapi', 'lld': 'inglesc (USA)', 'lo': 'ອັງກິດ (ອາເມລິກັນ)', 'lrc': 'ئینگیلیسی ئمریکایی', 'lt': 'Jungtinių Valstijų anglų', 'mai': 'अमेरिकी अंग्रेज़ी', 'mi': 'Ingarihi Amerikana', 'mk': 'американски англиски', 'ml': 'അമേരിക്കൻ ഇംഗ്ലീഷ്', 'mn': 'америк-англи', 'mni': 'অমেরিকান ইংলিস', 'mni-Beng': 'অমেরিকান ইংলিস', 'mr': 'अमेरिकन इंग्रजी', 'ms': 'Inggeris AS', 'mt': 'Ingliż Amerikan', 'my': 'အမေရိကန် အင်္ဂလိပ်', 'mzn': 'امریکن انگلیسی', 'ne': 'अमेरिकी अङ्ग्रेजी', 'nqo': 'ߊߡߋߙߌߞߌ߬ ߊ߲߬ߜ߭ߑߟߋ߬ߞߊ߲', 'om': 'Ingliffa Ameekiraa', 'or': 'ଆମେରିକୀୟ ଇଂରାଜୀ', 'os': 'америкаг англисаг', 'pa': 'ਅੰਗਰੇਜ਼ੀ (ਅਮਰੀਕੀ)', 'pa-Guru': 'ਅੰਗਰੇਜ਼ੀ (ਅਮਰੀਕੀ)', 'pcm': 'Amẹ́ríka Ínglish', 'pl': 'angielski amerykański', 'ps': 'د متحده آيالاتو انګليسي', 'qu': 'Ingles Simi (US)', 'rm': 'englais american', 'ro': 'engleză (S.U.A)', 'ru': 'американский английский', 'sa': 'अमेरिकादेशीय आङ्ग्लभाषा:', 'sat': 'ᱟᱢᱮᱨᱤᱠᱟᱱ ᱟᱝᱜᱽᱨᱮᱡᱤ', 'sat-Olck': 'ᱟᱢᱮᱨᱤᱠᱟᱱ ᱟᱝᱜᱽᱨᱮᱡᱤ', 'sc': 'inglesu americanu', 'sd': 'آمريڪي انگريزي', 'sd-Arab': 'آمريڪي انگريزي', 'sd-Deva': 'यूएस जी अंगरेज़ी', 'si': 'ඇමෙරිකානු ඉංග්\u200dරීසි', 'sk': 'angličtina (americká)', 'sl': 'ameriška angleščina', 'smn': 'Amerika eŋgâlâskielâ', 'so': 'Ingiriis Maraykan', 'sq': 'anglishte amerikane', 'sr': 'енглески (Сједињене Америчке Државе)', 'sr-Cyrl': 'енглески (Сједињене Америчке Државе)', 'sr-Latn': 'engleski (Sjedinjene Američke Države)', 'su': 'Inggris Amerika', 'su-Latn': 'Inggris Amerika', 'sv': 'amerikansk engelska', 'ta': 'அமெரிக்க ஆங்கிலம்', 'te': 'అమెరికన్ ఇంగ్లీష్', 'tg': 'англисии америкоӣ', 'th': 'อังกฤษ - สหรัฐอเมริกา', 'ti': 'እንግሊዝኛ (ሕ.መ.)', 'tk': 'iňlis dili (Amerika)', 'to': 'lea fakapālangi-ʻamelika', 'tpi': 'Amerikan Inglis', 'tr': 'Amerikan İngilizcesi', 'tt': 'Америка инглизчәсе', 'ug': 'ئامېرىكا ئىنگلىزچە', 'ur': 'امریکی انگریزی', 'uz': 'ingliz (Amerika)', 'uz-Cyrl': 'инглизча (Америка)', 'uz-Latn': 'ingliz (Amerika)', 'vec': 'ingleze meregan', 'vi': 'Tiếng Anh (Mỹ)', 'wae': 'Amerikanišes Engliš', 'wo': 'Àngale bu Amerik', 'xh': 'Isingesi SaseMelika', 'xnr': 'अमरिके आली अंग्रेज़ी', 'yi': 'ענגליש (US)', 'yo': 'Èdè Gẹ̀ẹ́sì (US)', 'zh': '美国英语', 'zh-Hans': '美国英语', 'zu': 'i-American English'},
    'ena': {'en': 'Apali'},
    'enb': {'en': 'Markweeta'},
    'enc': {'en': 'En'},
    'end': {'en': 'Ende'},
    'enf': {'en': 'Forest Enets'},
    'enh': {'en': 'Tundra Enets'},
    'enl': {'en': 'Enlhet'},
    'enm': {'ar': 'الإنجليزية الوسطى', 'ast': 'inglés mediu', 'az': 'orta ingilis', 'az-Latn': 'orta ingilis', 'bg': 'средновековен английски', 'bn': 'মধ্য ইংরেজি', 'br': 'krennsaozneg', 'brx': 'मध्य अंग्रेज़ी', 'bs': 'srednjovjekovni engleski', 'bs-Cyrl': 'средњи енглески', 'bs-Latn': 'srednjovjekovni engleski', 'ca': 'anglès mitjà', 'ca-ES-valencia': 'anglès mitjà', 'ccp': '𑄟𑄧𑄖𑄴𑄙𑄳𑄠𑄧 𑄃𑄨𑄁𑄢𑄬𑄎𑄨', 'cs': 'angličtina (středověká)', 'cy': 'Saesneg Canol', 'da': 'middelengelsk', 'de': 'Mittelenglisch', 'el': 'Μέσα Αγγλικά', 'el-polyton': 'Μέσα Αγγλικά', 'en': 'Middle English', 'en-Dsrt': '𐐣𐐮𐐼𐐲𐑊 𐐀𐑍𐑊𐐮𐑇', 'es': 'inglés medio', 'et': 'keskinglise', 'fa': 'انگلیسی میانه', 'fi': 'keskienglanti', 'fr': 'moyen anglais', 'fy': 'Middelingelsk', 'ga': 'Meán-Bhéarla', 'gd': 'Meadhan-Bheurla', 'gsw': 'Mittelänglisch', 'gu': 'મિડિલ અંગ્રેજી', 'he': 'אנגלית תיכונה', 'hi': 'मध्यकालीन अंग्रेज़ी', 'hi-Latn': 'Middle English', 'hr': 'srednjoengleski', 'ht': 'moyen anglais', 'hu': 'közép angol', 'id': 'Inggris Abad Pertengahan', 'is': 'miðenska', 'it': 'inglese medio', 'ja': '中英語', 'ka': 'საშუალო ინგლისური', 'kgp': 'ĩnhgrej kuju', 'kn': 'ಮಧ್ಯ ಇಂಗ್ಲೀಷ್', 'ko': '중세 영어', 'ks': 'وَسطی اَنگریٖزؠ', 'ks-Arab': 'وَسطی اَنگریٖزؠ', 'ksh': 'Meddelänglesch', 'lb': 'Mëttelenglesch', 'lo': 'ອັງກິດກາງ', 'lt': 'Vidurio Anglijos', 'lv': 'vidusangļu', 'mk': 'средноанглиски', 'ml': 'മദ്ധ്യ ഇംഗ്ലീഷ്', 'mr': 'मिडल इंग्रजी', 'mt': 'Ingliż Medjevali', 'my': 'အလယ်ပိုင်း အင်္ဂလိပ်', 'nb': 'mellomengelsk', 'ne': 'मध्य अङ्ग्रेजी', 'nl': 'Middelengels', 'nn': 'mellomengelsk', 'no': 'mellomengelsk', 'or': 'ମଧ୍ୟ ଇଁରାଜୀ', 'pl': 'średnioangielski', 'pt': 'inglês médio', 'rm': 'englais mesaun', 'ro': 'engleză medie', 'ru': 'среднеанглийский', 'sk': 'stredná angličtina', 'sl': 'srednja angleščina', 'sr': 'средњеенглески', 'sr-Cyrl': 'средњеенглески', 'sr-Latn': 'srednjeengleski', 'sv': 'medelengelska', 'ta': 'மிடில் ஆங்கிலம்', 'te': 'మధ్యమ ఆంగ్లం', 'th': 'อังกฤษกลาง', 'to': 'lea fakapālangi-lotoloto', 'tr': 'Ortaçağ İngilizcesi', 'ug': 'ئوتتۇرا ئەسىر ئىنگلىزچە', 'uk': 'середньоанглійська', 'vi': 'Tiếng Anh Trung cổ', 'yi': 'מיטל ענגליש', 'yrl': 'ẽgirixi médiu', 'yue': '中古英文', 'yue-Hans': '中古英文', 'yue-Hant': '中古英文', 'zh': '中古英语', 'zh-Hans': '中古英语', 'zh-Hant': '中古英文'},
    'enn': {'en': 'Engenni'},
    'eno': {'en': 'Enggano'},
    'enq': {'en': 'Enga'},
    'enr': {'en': 'Emumu'},
    'enu': {'en': 'Enu'},
    'env': {'en': 'Enwan (Edu State)'},
    'enw': {'en': 'Enwan (Akwa Ibom State)'},
    'enx': {'en': 'Enxet'},
    'eo': {'af': 'Esperanto', 'ak': 'Esperanto', 'am': 'ኤስፐራንቶ', 'ar': 'الإسبرانتو', 'as': 'এস্পেৰান্তো', 'ast': 'esperanto', 'az': 'esperanto', 'az-Arab': 'اِسپرانتو', 'az-Cyrl': 'есперанто', 'az-Latn': 'esperanto', 'bal-Latn': 'Esperántu', 'be': 'эсперанта', 'be-tarask': 'эсперанта', 'bg': 'есперанто', 'bgn': 'اسپرانتوی', 'bn': 'এস্পেরান্তো', 'br': 'esperanteg', 'brx': 'एस्पेरान्त’', 'bs': 'esperanto', 'bs-Cyrl': 'есперанто', 'bs-Latn': 'esperanto', 'ca': 'esperanto', 'ca-ES-valencia': 'esperanto', 'ccp': '𑄆𑄥𑄴𑄛𑄬𑄢𑄚𑄴𑄖𑄮', 'ce': 'эсперанто', 'chr': 'ᎡᏍᏇᎳᏂᏙ', 'ckb': 'ئێسپیرانتۆ', 'cs': 'esperanto', 'cy': 'Esperanto', 'da': 'esperanto', 'de': 'Esperanto', 'dsb': 'esperanto', 'dz': 'ཨེས་པ་རཱན་ཏོ་ཁ', 'ee': 'esperantogbe', 'el': 'Εσπεράντο', 'el-polyton': 'Εσπεράντο', 'en': 'Esperanto', 'en-Dsrt': '𐐇𐑅𐐹𐐯𐑉𐐪𐑌𐐻𐐬', 'eo': 'Esperanto', 'es': 'esperanto', 'et': 'esperanto', 'eu': 'esperantoa', 'fa': 'اسپرانتو', 'ff-Adlm': '𞤉𞤧𞤨𞤫𞤪𞤢𞤲𞤼𞤮𞥅𞤪𞤫', 'fi': 'esperanto', 'fil': 'Esperanto', 'fo': 'esperanto', 'fr': 'espéranto', 'fur': 'esperanto', 'fy': 'Esperanto', 'ga': 'Esperanto', 'gd': 'Esperanto', 'gl': 'esperanto', 'gsw': 'Eschperanto', 'gu': 'એસ્પેરાન્ટો', 'ha': 'Esperanto', 'he': 'אספרנטו', 'hi': 'एस्पेरेंतो', 'hi-Latn': 'Esperanto', 'hr': 'esperanto', 'hsb': 'esperanto', 'ht': 'espéranto', 'hu': 'eszperantó', 'hy': 'էսպերանտո', 'ia': 'esperanto', 'id': 'Esperanto', 'ig': 'Esperanto', 'is': 'esperantó', 'it': 'esperanto', 'ja': 'エスペラント語', 'jv': 'Esperanto', 'ka': 'ესპერანტო', 'kea': 'sperantu', 'kgp': 'enhperỹtu', 'kk': 'эсперанто тілі', 'kk-Cyrl': 'эсперанто тілі', 'km': 'អេស្ពេរ៉ាន់តូ', 'kn': 'ಎಸ್ಪೆರಾಂಟೊ', 'ko': '에스페란토어', 'kok': 'इस्परान्टो', 'kok-Deva': 'इस्परान्टो', 'ks': 'ایسپَرینٹو', 'ks-Arab': 'ایسپَرینٹو', 'ksh': 'Esperanto', 'ku': 'esperantoyî', 'ky': 'эсперанто', 'lb': 'Esperanto', 'lkt': 'Esperanto Iyápi', 'lld': 'esperanto', 'lo': 'ເອສປາຍ', 'lrc': 'ئسپئرانتو', 'lt': 'esperanto', 'lv': 'esperanto', 'mai': 'एसपेरेंटो', 'mi': 'Eheperāto', 'mk': 'есперанто', 'ml': 'എസ്\u200cപരാന്റോ', 'mn': 'эсперанто', 'mr': 'एस्परान्टो', 'ms': 'Esperanto', 'mt': 'Esperanto', 'my': 'အက်စ်ပရန်တို', 'mzn': 'اسپرانتو', 'nb': 'esperanto', 'ne': 'एस्पेरान्तो', 'nl': 'Esperanto', 'nn': 'esperanto', 'no': 'esperanto', 'nqo': 'ߍߛߑߔߋߙߊ߲ߕߏߞߊ߲', 'om': 'Afaan Esperantoo', 'or': 'ଏସ୍ପାରେଣ୍ଟୋ', 'os': 'есперанто', 'pa': 'ਇਸਪੇਰਾਂਟੋ', 'pa-Guru': 'ਇਸਪੇਰਾਂਟੋ', 'pcm': 'Ẹsperánto Lángwej', 'pl': 'esperanto', 'ps': 'اسپرانتو', 'pt': 'esperanto', 'qu': 'Esperanto Simi', 'rm': 'esperanto', 'ro': 'esperanto', 'ru': 'эсперанто', 'rw': 'Icyesiperanto', 'sc': 'esperanto', 'sd': 'ايسپرانٽو', 'sd-Arab': 'ايسپرانٽو', 'sdh': 'ئسپرانتو', 'si': 'එස්පැරන්ටෝ', 'sk': 'esperanto', 'sl': 'esperanto', 'smn': 'esperantokielâ', 'so': 'Isberaanto', 'sq': 'esperanto', 'sr': 'есперанто', 'sr-Cyrl': 'есперанто', 'sr-Latn': 'esperanto', 'sv': 'esperanto', 'sw': 'Kiesperanto', 'ta': 'எஸ்பரேன்டோ', 'te': 'ఎస్పెరాంటో', 'tg': 'эсперанто', 'th': 'เอสเปรันโต', 'ti': 'ኤስፐራንቶ', 'tk': 'esperanto dili', 'to': 'lea fakaʻesipulanito', 'tr': 'Esperanto', 'tt': 'эсперанто', 'ug': 'ئېسپرانتوچە', 'uk': 'есперанто', 'ur': 'ایسپرانٹو', 'uz': 'esperanto', 'uz-Cyrl': 'эсперанто', 'uz-Latn': 'esperanto', 'vec': 'esperanto', 'vi': 'Tiếng Quốc Tế Ngữ', 'wo': 'Esperantoo', 'xnr': 'एस्पेरेंतो', 'yi': 'עספּעראַנטא', 'yo': 'Èdè Esperanto', 'yrl': 'esiperãtu', 'yue': '世界文', 'yue-Hans': '世界文', 'yue-Hant': '世界文', 'zh': '世界语', 'zh-Hans': '世界语', 'zh-Hant': '世界文', 'zu': 'isi-Esperanto'},
    'eot': {'en': "Beti (Côte d'Ivoire)"},
    'epi': {'en': 'Epie'},
    'era': {'en': 'Eravallan'},
    'erg': {'en': 'Sie'},
    'erh': {'en': 'Eruwa'},
    'eri': {'en': 'Ogea'},
    'erk': {'en': 'South Efate'},
    'ero': {'en': 'Horpa'},
    'err': {'en': 'Erre'},
    'ers': {'en': 'Ersu'},
    'ert': {'en': 'Eritai'},
    'erw': {'en': 'Erokwanas'},
    'es': {'af': 'Spaans', 'agq': 'Sɨ̀kpanìs', 'ak': 'Spain kasa', 'am': 'ስፓኒሽ', 'ar': 'الإسبانية', 'as': 'স্পেনিচ', 'asa': 'Kihithpania', 'ast': 'español', 'az': 'ispan', 'az-Arab': 'اسپانیاجا', 'az-Cyrl': 'испан', 'az-Latn': 'ispan', 'bal-Latn': 'Espini', 'bas': 'Hɔp u panyā', 'be': 'іспанская', 'be-tarask': 'іспанская', 'bem': 'Ichi Spanishi', 'bez': 'Hihispania', 'bg': 'испански', 'bgn': 'هسپانوی', 'blo': 'gɩspaŋja', 'bm': 'esipaɲolkan', 'bn': 'স্প্যানিশ', 'br': 'spagnoleg', 'brx': 'स्पेनिश', 'bs': 'španski', 'bs-Cyrl': 'шпански', 'bs-Latn': 'španski', 'ca': 'espanyol', 'ca-ES-valencia': 'espanyol', 'ccp': '𑄥𑄳𑄛𑄳𑄠𑄚𑄨𑄌𑄴', 'ce': 'испанхойн', 'ceb': 'Espanyol', 'cgg': 'Orusupaani', 'chr': 'ᏍᏆᏂ', 'ckb': 'ئیسپانی', 'cs': 'španělština', 'cv': 'испани', 'cy': 'Sbaeneg', 'da': 'spansk', 'dav': 'Kihispania', 'de': 'Spanisch', 'dje': 'Espaaɲe senni', 'doi': 'स्पैनिश', 'dsb': 'špańšćina', 'dyo': 'español', 'dz': 'ཨིས་པེ་ནིཤ་ཁ', 'ebu': 'Kĩhispania', 'ee': 'Spanishgbe', 'el': 'Ισπανικά', 'el-polyton': 'Ἱσπανικά', 'en': 'Spanish', 'en-Dsrt': '𐐝𐐹𐐰𐑌𐐮𐑇', 'eo': 'hispana', 'es': 'español', 'et': 'hispaania', 'eu': 'gaztelania', 'ewo': 'ǹkɔ́bɔ kpənyá', 'fa': 'اسپانیایی', 'ff': 'Español', 'ff-Adlm': '𞤅𞤭𞤨𞤢𞤲𞤭𞤴𞤢𞤲𞤳𞤮𞥅𞤪𞤫', 'ff-Latn': 'Español', 'fi': 'espanja', 'fil': 'Spanish', 'fo': 'spanskt', 'fr': 'espagnol', 'fur': 'spagnûl', 'fy': 'Spaansk', 'ga': 'Spáinnis', 'gd': 'Spàinntis', 'gl': 'español', 'gsw': 'Schpanisch', 'gu': 'સ્પેનિશ', 'guz': 'Kihispania', 'ha': 'Sifaniyanci', 'haw': 'Paniolo', 'he': 'ספרדית', 'hi': 'स्पेनिश', 'hi-Latn': 'Spanish', 'hr': 'španjolski', 'hsb': 'španišćina', 'ht': 'espagnol', 'hu': 'spanyol', 'hy': 'իսպաներեն', 'ia': 'espaniol', 'id': 'Spanyol', 'ie': 'hispan', 'ig': 'Spanish', 'ii': 'ꑭꀠꑸꉙ', 'is': 'spænska', 'it': 'spagnolo', 'ja': 'スペイン語', 'jmc': 'Kyihispania', 'jv': 'Spanyol', 'ka': 'ესპანური', 'kab': 'Taspenyulit', 'kam': 'Kihispania', 'kde': 'Chihispania', 'kea': 'spanhol', 'kgp': 'enhpỹjór', 'khq': 'Espaaɲe senni', 'ki': 'Kihispania', 'kk': 'испан тілі', 'kk-Cyrl': 'испан тілі', 'kln': 'kutitab Espianik', 'km': 'អេស្ប៉ាញ', 'kn': 'ಸ್ಪ್ಯಾನಿಷ್', 'ko': '스페인어', 'kok': 'स्पॅनीश', 'kok-Deva': 'स्पॅनीश', 'kok-Latn': 'Ispanhol', 'ks': 'ہسپانوی', 'ks-Arab': 'ہسپانوی', 'ks-Deva': 'हसपानवी', 'ksb': 'Kihispania', 'ksf': 'rikpanyá', 'ksh': 'Schpahnesch', 'ku': 'spanî', 'kxv': 'spenis', 'kxv-Deva': 'स्पानिस', 'kxv-Latn': 'spenis', 'kxv-Orya': 'ସ୍ପାନିସ୍', 'kxv-Telu': 'స్పెనిస్', 'ky': 'испанча', 'lag': 'Kɨhispánia', 'lb': 'Spuenesch', 'lg': 'Lusipanya', 'lkt': 'Spayóla Iyápi', 'lld': 'spagnol', 'ln': 'lisipanye', 'lo': 'ສະແປນນິຊ', 'lrc': 'ئسپانیایی', 'lt': 'ispanų', 'lu': 'Lihispania', 'luo': 'Kihispania', 'luy': 'Kihispania', 'lv': 'spāņu', 'mai': 'स्पेनिश', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lspania', 'mer': 'Kĩspĩni', 'mfe': 'espagnol', 'mg': 'Espaniola', 'mgh': 'Ihispaniola', 'mi': 'Pāniora', 'mk': 'шпански', 'ml': 'സ്\u200cപാനിഷ്', 'mn': 'испани', 'mni': 'স্পেনিস', 'mni-Beng': 'স্পেনিস', 'mr': 'स्पॅनिश', 'ms': 'Sepanyol', 'mt': 'Spanjol', 'mua': 'Espaniya', 'my': 'စပိန်', 'mzn': 'ایسپانیولی', 'naq': 'Spaans', 'nb': 'spansk', 'nd': 'isi-Sipeyini', 'ne': 'स्पेनी', 'nl': 'Spaans', 'nmg': 'Paŋá', 'nn': 'spansk', 'no': 'spansk', 'nqo': 'ߊߛߌߔߞߊ߲', 'nus': 'Thok i̱thpaaniani', 'nyn': 'Orusupaani', 'om': 'Afaan Ispeen', 'or': 'ସ୍ପାନିସ୍\u200c', 'os': 'испайнаг', 'pa': 'ਸਪੇਨੀ', 'pa-Guru': 'ਸਪੇਨੀ', 'pcm': 'Spánish Lángwej', 'pl': 'hiszpański', 'ps': 'هسپانوي', 'pt': 'espanhol', 'qu': 'Español Simi', 'rm': 'spagnol', 'rn': 'Icesipanyolo', 'ro': 'spaniolă', 'rof': 'Kihispania', 'ru': 'испанский', 'rw': 'Icyesipanyolo', 'rwk': 'Kyihispania', 'sa': 'स्पेनीय भाषा:', 'sah': 'Ыспаанныы', 'saq': 'Kihispania', 'sat': 'ᱥᱯᱮᱱᱤᱥ', 'sat-Olck': 'ᱥᱯᱮᱱᱤᱥ', 'sbp': 'Ishihisipaniya', 'sc': 'ispagnolu', 'sd': 'هسپانوي', 'sd-Arab': 'هسپانوي', 'sd-Deva': 'स्पेनिश', 'se': 'spánskkagiella', 'seh': 'espanhol', 'ses': 'Espaaɲe senni', 'sg': 'Espanyöl', 'shi': 'ⵜⴰⵙⴱⵏⵢⵓⵍⵉⵜ', 'shi-Latn': 'Tasbnyulit', 'shi-Tfng': 'ⵜⴰⵙⴱⵏⵢⵓⵍⵉⵜ', 'si': 'ස්පාඤ්ඤ', 'sk': 'španielčina', 'sl': 'španščina', 'smn': 'espanjakielâ', 'sn': 'chiSpanish', 'so': 'Isbaanish', 'sq': 'spanjisht', 'sr': 'шпански', 'sr-Cyrl': 'шпански', 'sr-Latn': 'španski', 'su': 'Spanyol', 'su-Latn': 'Spanyol', 'sv': 'spanska', 'sw': 'Kihispania', 'syr': 'ܣܦܢܝܝܐ', 'ta': 'ஸ்பானிஷ்', 'te': 'స్పానిష్', 'teo': 'Kihispania', 'tg': 'испанӣ', 'th': 'สเปน', 'ti': 'ስጳንኛ', 'tk': 'ispan dili', 'to': 'lea fakasipēnisi', 'tpi': 'Spenis', 'tr': 'İspanyolca', 'tt': 'испан', 'twq': 'Espaaɲe senni', 'tzm': 'tasbelyunit', 'ug': 'ئىسپانچە', 'uk': 'іспанська', 'ur': 'ہسپانوی', 'uz': 'ispancha', 'uz-Cyrl': 'испанча', 'uz-Latn': 'ispancha', 'vai': 'ꕐꘊꔧ', 'vai-Latn': 'Panyɛĩ', 'vai-Vaii': 'ꕐꘊꔧ', 'vec': 'spagnolo', 'vi': 'Tiếng Tây Ban Nha', 'vun': 'Kyihispania', 'wae': 'Schpaniš', 'wo': 'Español', 'xh': 'Isi-Spanish', 'xnr': 'स्पेनी', 'xog': 'Olusipanya', 'yav': 'nuɛspanyɔ́lɛ', 'yi': 'שפּאַניש', 'yo': 'Èdè Sípáníìṣì', 'yrl': 'isipãyu', 'yue': '西班牙文', 'yue-Hans': '西班牙文', 'yue-Hant': '西班牙文', 'zgh': 'ⵜⴰⵙⴱⵏⵢⵓⵍⵉⵜ', 'zh': '西班牙语', 'zh-Hans': '西班牙语', 'zh-Hant': '西班牙文', 'zu': 'isi-Spanish'},
    'es-419': {'ak': 'Spain kasa (Laaten Amɛrika)', 'am': 'የላቲን አሜሪካ ስፓኒሽ', 'ar': 'الإسبانية أمريكا اللاتينية', 'as': 'লেটিন আমেৰিকান স্পেনিচ', 'ast': 'español d’América Llatina', 'az': 'Latın Amerikası ispancası', 'az-Arab': 'لاتین آمریکا اسپانیاجاسی', 'az-Cyrl': 'Латын Америкасы испанҹасы', 'az-Latn': 'Latın Amerikası ispancası', 'bal-Latn': 'Espini (Látini Amriká)', 'be': 'лацінаамерыканская іспанская', 'be-tarask': 'лацінаамерыканская іспанская', 'bgn': 'لاتین امریکایی هسپانوی', 'blo': 'gɩspaŋja (latɛŋ kaAmalɩkatǝna)', 'br': 'spagnoleg Amerika latin', 'brx': 'लेटीन आमेरिकानि स्पेनिश', 'ca': 'espanyol llatinoamericà', 'ca-ES-valencia': 'espanyol llatinoamericà', 'ccp': '𑄣𑄳𑄠𑄑𑄨𑄚𑄴 𑄃𑄟𑄬𑄢𑄨𑄇𑄚𑄴 𑄥𑄳𑄛𑄳𑄠𑄚𑄨𑄌𑄴', 'ce': 'латинан американ испанхойн', 'chr': 'ᏔᏘᏂ ᎠᎹᏰᏟ ᏍᏆᏂ', 'cv': 'латинла америкӑн испани', 'cy': 'Sbaeneg America Ladin', 'da': 'latinamerikansk spansk', 'doi': 'लैटिन अमरीकी स्पैनिश', 'dsb': 'łatyńskoamerikańska špańšćina', 'dz': 'ལེ་ཊིན་ཨ་མེ་རི་ཀཱན་གི་ཨིས་པེ་ནིཤ་ཁ', 'ee': 'Spanishgbe (Latin America)', 'el': 'Ισπανικά Λατινικής Αμερικής', 'el-polyton': 'Ισπανικά Λατινικής Αμερικής', 'en': 'Latin American Spanish', 'es': 'español latinoamericano', 'et': 'Ladina-Ameerika hispaania', 'eu': 'Latinoamerikako gaztelania', 'fa': 'اسپانیایی امریکای لاتین', 'ff-Adlm': '𞤅𞤭𞤨𞤢𞤲𞤭𞤴𞤢𞤲𞤳𞤮𞥅𞤪𞤫 𞤀𞤥𞤭𞤪𞤭𞤳 𞤂𞤢𞤼𞤭𞤲𞤭𞤴𞤢', 'fi': 'amerikanespanja', 'fil': 'Latin American na Espanyol', 'fr': 'espagnol d’Amérique latine', 'fur': 'spagnûl de Americhe Latine', 'fy': 'Latynsk-Amerikaansk Spaansk', 'ga': 'Spáinnis Mheiriceá Laidinigh', 'gd': 'Spàinntis na h-Aimeireaga Laidinneach', 'gl': 'español de América', 'gsw': 'Latiinamerikanischs Schpanisch', 'gu': 'લેટિન અમેરિકન સ્પેનિશ', 'ha': 'Sifaniyancin Latin Amirka', 'hi': 'लैटिन अमेरिकी स्पेनिश', 'hi-Latn': 'Latin American Spanish', 'hr': 'latinoamerički španjolski', 'hsb': 'łaćonskoameriska španišćina', 'ht': 'espagnol d’Amérique latine', 'hu': 'latin-amerikai spanyol', 'hy': 'լատինամերիկյան իսպաներեն', 'ia': 'espaniol latinoamerican', 'ie': 'hispan del latin America', 'ig': 'Spanish ndị Latin America', 'is': 'rómönsk-amerísk spænska', 'it': 'spagnolo latinoamericano', 'jv': 'Spanyol (Amerika Latin)', 'ka': 'ლათინურ ამერიკული ესპანური', 'kea': 'spanhol latinu-merkanu', 'kn': 'ಲ್ಯಾಟಿನ್ ಅಮೇರಿಕನ್ ಸ್ಪ್ಯಾನಿಷ್', 'kok': 'लातीं अमेरिकन स्पॅनीश', 'kok-Deva': 'लातीं अमेरिकन स्पॅनीश', 'ks': 'لاطیٖنی امریٖکی ہسپانوی', 'ks-Arab': 'لاطیٖنی امریٖکی ہسپانوی', 'ks-Deva': 'लातिनी अमरीकी हसपानवी', 'ksh': 'Schpahnesch uß Latting-Ammärrika', 'kxv': 'laṭin americati spenis', 'kxv-Deva': 'लातिन आमेरिका ति स्पेनिस', 'kxv-Latn': 'laṭin americati spenis', 'kxv-Orya': 'ଲାଟିନ୍\u200c ଆମେରିକା ତି ସ୍ପେନିସ୍', 'kxv-Telu': 'లాటిన్ అమెరికన్ స్పెనిస్', 'lb': 'Latäinamerikanescht Spuenesch', 'lkt': 'Wiyóȟpeyata Spayóla Iyápi', 'lo': 'ລາຕິນ ອາເມຣິກັນ ສະແປນນິຊ', 'lrc': 'ئسپانیایی ئمریکا لاتین', 'mai': 'लैटिन अमेरिकी स्पेनिश', 'mi': 'Pāniora Amerikana ki te Tonga', 'mk': 'латиноамерикански шпански', 'ml': 'ലാറ്റിൻ അമേരിക്കൻ സ്\u200cപാനിഷ്', 'mn': 'испани хэл (Латин Америк)', 'mni': 'লেটিন অমেরিকান স্পেনিস', 'mni-Beng': 'লেটিন অমেরিকান স্পেনিস', 'mr': 'लॅटिन अमेरिकन स्पॅनिश', 'ms': 'Sepanyol Amerika Latin', 'mt': 'Spanjol Latin Amerikan', 'mzn': 'جنوبی آمریکای ِایسپانیولی', 'ne': 'ल्याटिन अमेरिकी स्पेनी', 'nqo': 'ߊߡߋߙߌߞߌ߬ ߊߛߌߔߊ߲ߞߊ߲ ߟߊ߬ߕߍ߬ߡߊ', 'om': 'Laatinii Ispaanishii Ameerikaa', 'or': 'ଲାଟିନ୍\u200c ଆମେରିକୀୟ ସ୍ପାନିସ୍\u200c', 'os': 'латинаг америкаг англисаг', 'pa': 'ਸਪੇਨੀ (ਲਾਤੀਨੀ ਅਮਰੀਕੀ)', 'pa-Guru': 'ਸਪੇਨੀ (ਲਾਤੀਨੀ ਅਮਰੀਕੀ)', 'pcm': 'Látín Amẹ́ríka Spánish', 'pl': 'amerykański hiszpański', 'ps': 'لاتيني امريکايي هسپانوي', 'qu': 'Español Simi (Latino América)', 'rm': 'spagnol latinamerican', 'ru': 'латиноамериканский испанский', 'sa': 'लैटिन अमेरिकादेशीय स्पेनीय भाषा:', 'sat': 'ᱞᱮᱴᱤᱱ ᱟᱢᱮᱨᱤᱠᱟᱱ ᱥᱯᱮᱱᱤᱥ', 'sat-Olck': 'ᱞᱮᱴᱤᱱ ᱟᱢᱮᱨᱤᱠᱟᱱ ᱥᱯᱮᱱᱤᱥ', 'sc': 'ispagnolu latinoamericanu', 'sd': 'لاطيني آمريڪي اسپينش', 'sd-Arab': 'لاطيني آمريڪي اسپينش', 'sd-Deva': 'लैटिन अमेरिकन स्पैनिश', 'si': 'ලතින් ඇමරිකානු ස්පාඤ්ඤ', 'sk': 'španielčina (latinskoamerická)', 'sl': 'latinskoameriška španščina', 'smn': 'Läättin-Amerika espanjakielâ', 'so': 'Isbaanishka Laatiin Ameerika', 'sq': 'spanjishte amerikano-latine', 'su': 'Spanyol Amérika Latin', 'su-Latn': 'Spanyol Amérika Latin', 'sv': 'latinamerikansk spanska', 'sw': 'Kihispania (Amerika ya Latini)', 'ta': 'லத்தின் அமெரிக்க ஸ்பானிஷ்', 'te': 'లాటిన్ అమెరికన్ స్పానిష్', 'tg': 'испании америкоии лотинӣ', 'th': 'สเปน - ละตินอเมริกา', 'to': 'lea fakasipeini-lātini-ʻamelika', 'tpi': 'Saut Amerikan Spenis', 'tr': 'Latin Amerika İspanyolcası', 'tt': 'испан (Латин Америкасы)', 'ug': 'لاتىن ئامېرىكا ئىسپانچە', 'ur': 'لاطینی امریکی ہسپانوی', 'uz': 'ispan (Lotin Amerikasi)', 'uz-Latn': 'ispan (Lotin Amerikasi)', 'vi': 'Tiếng Tây Ban Nha (Mỹ La tinh)', 'wae': 'Latiamerikanišes Schpaniš', 'wo': 'Español bu Amerik Latin', 'xh': 'IsiSpanish SaseLatin America', 'xnr': 'लैटिन अमरिकी स्पेनिश', 'yo': 'Èdè Sípáníìṣì (orílẹ̀-èdè Látìn-Amẹ́ríkà)', 'zh': '拉丁美洲西班牙语', 'zh-Hans': '拉丁美洲西班牙语', 'zu': 'isi-Latin American Spanish'},
    'es-ES': {'am': 'የአውሮፓ ስፓኒሽ', 'ar': 'الإسبانية الأوروبية', 'as': 'ইউৰোপীয়ান স্পেনিচ', 'ast': 'español européu', 'az': 'Kastiliya ispancası', 'az-Arab': 'اۇروپا اسپانیاجاسی', 'az-Cyrl': 'Кастилија испанҹасы', 'az-Latn': 'Kastiliya ispancası', 'be': 'еўрапейская іспанская', 'be-tarask': 'еўрапейская іспанская', 'bgn': 'اوروپایین هسپانوی', 'bn': 'স্প্যানিশ (ইউরোপ)', 'br': 'spagnoleg Europa', 'brx': 'युरोपनि स्पेनिश', 'ca': 'espanyol europeu', 'ca-ES-valencia': 'espanyol europeu', 'ccp': '𑄄𑄅𑄢𑄮𑄛𑄩𑄠𑄧 𑄥𑄳𑄛𑄳𑄠𑄚𑄨𑄌𑄴', 'ce': 'европан испанхойн', 'ceb': 'Espanyol (Europa)', 'chr': 'ᎠᏂᏍᏆᏂᏱ ᏍᏆᏂ', 'cs': 'španělština (Evropa)', 'cv': 'европӑн испани', 'cy': 'Sbaeneg Ewrop', 'da': 'europæisk spansk', 'doi': 'यूरोपी स्पैनिश', 'dsb': 'europejska špańšćina', 'dz': 'ཡུ་རོབ་ཀྱི་ཨིས་པེ་ནིཤ་ཁ', 'ee': 'Spanishgbe (Europe)', 'el': 'Ισπανικά Ευρώπης', 'el-polyton': 'Ισπανικά Ευρώπης', 'en': 'European Spanish', 'es': 'español de España', 'et': 'Euroopa hispaania', 'eu': 'Europako gaztelania', 'fa': 'اسپانیایی اروپا', 'ff-Adlm': '𞤅𞤭𞤨𞤢𞤲𞤭𞤴𞤢𞤲𞤳𞤮𞥅𞤪𞤫 𞤀𞤪𞤮𞤦𞤢', 'fi': 'euroopanespanja', 'fil': 'European Spanish', 'fr': 'espagnol d’Espagne', 'fur': 'spagnûl iberic', 'fy': 'Europeesk Spaansk', 'ga': 'Spáinnis Eorpach', 'gd': 'Spàinntis Eòrpach', 'gl': 'español de España', 'gsw': 'Ibeerischs Schpanisch', 'gu': 'યુરોપિયન સ્પેનિશ', 'ha': 'Sifaniyanci Turai', 'hi': 'यूरोपीय स्पेनिश', 'hi-Latn': 'European Spanish', 'hr': 'europski španjolski', 'hsb': 'europska španišćina', 'ht': 'espagnol d’Espagne', 'hu': 'európai spanyol', 'hy': 'եվրոպական իսպաներեն', 'ia': 'espaniol europee', 'id': 'Spanyol (Eropa)', 'ie': 'europan hispan', 'ig': 'Spanish ndị Europe', 'is': 'evrópsk spænska', 'it': 'spagnolo europeo', 'ja': 'スペイン語 (イベリア半島)', 'jv': 'Spanyol (Eropah)', 'ka': 'ევროპული ესპანური', 'kea': 'spanhol europeu', 'km': 'អេស្ប៉ាញ (អ៊ឺរ៉ុប)', 'kn': 'ಯುರೋಪಿಯನ್ ಸ್ಪ್ಯಾನಿಷ್', 'ko': '스페인어(유럽)', 'kok': 'युरोपियन स्पॅनीश', 'kok-Deva': 'युरोपियन स्पॅनीश', 'ks': 'یوٗرپی ہسپانوی', 'ks-Arab': 'یوٗرپی ہسپانوی', 'ks-Deva': 'यूरपी हसपानवी', 'ksh': 'Schpahnesch en Schpahneje', 'ku': 'spanî (Ewropa)', 'kxv': 'yuropīyati spenis', 'kxv-Deva': 'यूरोपीय ति स्पेनिस', 'kxv-Latn': 'yuropīyati spenis', 'kxv-Orya': 'ୟୁରପୀୟାତି ସ୍ପେନିସ୍\u200c', 'kxv-Telu': 'యురోపియన్ స్పెనిస్', 'ky': 'испанча (Европа)', 'lb': 'Europäescht Spuenesch', 'lkt': 'Spayólaȟča Iyápi', 'lo': 'ສະເປັນ ຢຸໂຣບ', 'lrc': 'ئسپانیایی ئوروٙپا', 'mai': 'यूरोपीय स्पेनिश', 'mi': 'Pāniora Ūropi', 'mk': 'шпански (Европа)', 'ml': 'യൂറോപ്യൻ സ്\u200cപാനിഷ്', 'mn': 'испани хэл (Европ)', 'mni': 'য়ুরোপিয়ান স্পেনিস', 'mni-Beng': 'য়ুরোপিয়ান স্পেনিস', 'mr': 'युरोपियन स्पॅनिश', 'ms': 'Sepanyol Eropah', 'mt': 'Spanjol Ewropew', 'my': 'စပိန် (ဥရောပ)', 'mzn': 'اروپای ِایسپانیولی', 'ne': 'युरोपेली स्पेनी', 'nqo': 'ߊߛߌߔߊ߲߫ ߊߛߌߔߊ߲ߞߊ߲', 'om': 'Ispaanishii Awurooppaa', 'or': 'ୟୁରୋପୀୟ ସ୍ପାନିସ୍\u200c', 'os': 'европӕйаг англисаг', 'pa': 'ਸਪੇਨੀ (ਯੂਰਪੀ)', 'pa-Guru': 'ਸਪੇਨੀ (ਯੂਰਪੀ)', 'pcm': 'Yúrop Spánish', 'pl': 'europejski hiszpański', 'ps': 'اروپايي هسپانوي', 'rm': 'spagnol europeic', 'ro': 'spaniolă (Europa)', 'ru': 'европейский испанский', 'sa': 'फिरङ्गिन् स्पेनीय भाषा:', 'sat': 'ᱨᱩᱥᱤᱭᱟᱱ ᱥᱯᱮᱱᱤᱥ', 'sat-Olck': 'ᱨᱩᱥᱤᱭᱟᱱ ᱥᱯᱮᱱᱤᱥ', 'sc': 'ispagnolu europeu', 'sd': 'يورپي اسپيني', 'sd-Arab': 'يورپي اسپيني', 'sd-Deva': 'यूरोपी स्पैनिश', 'si': 'යුරෝපීය ස්පාඤ්ඤ', 'sk': 'španielčina (európska)', 'sl': 'evropska španščina', 'smn': 'Espanja espanjakielâ', 'so': 'Isbaanish Yurub', 'sq': 'spanjishte evropiane', 'sr': 'шпански (Европа)', 'sr-Cyrl': 'шпански (Европа)', 'sr-Latn': 'španski (Evropa)', 'su': 'Spanyol Éropa', 'su-Latn': 'Spanyol Éropa', 'sv': 'europeisk spanska', 'sw': 'Kihispania (Ulaya)', 'ta': 'ஐரோப்பிய ஸ்பானிஷ்', 'te': 'యూరోపియన్ స్పానిష్', 'tg': 'испании аврупоӣ', 'th': 'สเปน - ยุโรป', 'ti': 'ስጳንኛ (ኤውሮጳዊ)', 'tk': 'ispan dili (Ýewropa)', 'to': 'lea fakasipeini-ʻeulope', 'tpi': 'Spenis (Spein)', 'tr': 'Avrupa İspanyolcası', 'tt': 'испан (Европа)', 'ug': 'ياۋروپا ئىسپانچە', 'ur': 'یورپی ہسپانوی', 'uz': 'ispan (Yevropa)', 'uz-Latn': 'ispan (Yevropa)', 'vi': 'Tiếng Tây Ban Nha (Châu Âu)', 'wae': 'Iberišes Schpaniš', 'wo': 'Español bu Tugël', 'xh': 'IsiSpanish SaseYurophu', 'xnr': 'यूरोपे आली स्पेनिश', 'yo': 'Èdè Sípáníìṣì (orílẹ̀-èdè Yúróòpù)', 'zh': '欧洲西班牙语', 'zh-Hans': '欧洲西班牙语', 'zu': 'isi-European Spanish'},
    'es-MX': {'am': 'የሜክሲኮ ስፓኒሽ', 'ar': 'الإسبانية المكسيكية', 'as': 'মেক্সিকান স্পেনিচ', 'ast': 'español de Méxicu', 'az': 'Meksika ispancası', 'az-Arab': 'مکزیک اسپانیاجاسی', 'az-Cyrl': 'Мексика испанҹасы', 'az-Latn': 'Meksika ispancası', 'bal-Latn': 'Espini (Meksikó)', 'be': 'мексіканская іспанская', 'be-tarask': 'мексіканская іспанская', 'bgn': 'مکسیکوین هسپانوی', 'br': 'spagnoleg Mecʼhiko', 'brx': 'मेक्सिक’नि स्पेनिश', 'ca': 'espanyol de Mèxic', 'ca-ES-valencia': 'espanyol de Mèxic', 'ccp': '𑄟𑄳𑄠𑄇𑄴𑄥𑄨𑄇𑄚𑄴 𑄥𑄳𑄛𑄳𑄠𑄚𑄨𑄌𑄴', 'ce': 'мексикан испанхойн', 'chr': 'ᏍᏆᏂᏱ ᏍᏆᏂ', 'cv': 'мексикӑн испани', 'cy': 'Sbaeneg Mecsico', 'da': 'mexicansk spansk', 'doi': 'मैक्सिन स्पैनिश', 'dsb': 'mexikańska špańšćina', 'ee': 'Spanishgbe (Mexico)', 'el': 'Ισπανικά Μεξικού', 'el-polyton': 'Ισπανικά Μεξικού', 'en': 'Mexican Spanish', 'es': 'español de México', 'et': 'Mehhiko hispaania', 'eu': 'Mexikoko gaztelania', 'fa': 'اسپانیایی مکزیک', 'ff-Adlm': '𞤅𞤭𞤨𞤢𞤲𞤭𞤴𞤢𞤲𞤳𞤮𞥅𞤪𞤫 𞤃𞤫𞤳𞤧𞤭𞤳', 'fi': 'meksikonespanja', 'fil': 'Mexican na Espanyol', 'fr': 'espagnol du Mexique', 'fy': 'Meksikaansk Spaansk', 'ga': 'Spáinnis Mheicsiceach', 'gd': 'Spàinntis Mheagsagach', 'gl': 'español de México', 'gu': 'મેક્સિકન સ્પેનિશ', 'ha': 'Sifaniyanci Mesiko', 'hi': 'मैक्सिकन स्पेनिश', 'hi-Latn': 'Mexican Spanish', 'hr': 'meksički španjolski', 'hsb': 'mexiska španišćina', 'ht': 'espagnol du Mexique', 'hu': 'spanyol (mexikói)', 'hy': 'մեքսիկական իսպաներեն', 'ia': 'espaniol mexican', 'ie': 'mexican hispan', 'ig': 'Spanish ndị Mexico', 'is': 'mexíkósk spænska', 'it': 'spagnolo messicano', 'jv': 'Spanyol (Meksiko)', 'ka': 'მექსიკური ესპანური', 'kea': 'spanhol mexikanu', 'kn': 'ಮೆಕ್ಸಿಕನ್ ಸ್ಪ್ಯಾನಿಷ್', 'kok': 'मॅक्सिकन स्पॅनीश', 'kok-Deva': 'मॅक्सिकन स्पॅनीश', 'ks': 'میکسیکن ہسپانوی', 'ks-Arab': 'میکسیکن ہسپانوی', 'ks-Deva': 'मेकसिकी हसपानवी', 'ksh': 'Schpahnesch en Mäxikoh', 'kxv': 'meksikoti spenis', 'kxv-Deva': 'मेक्सिको ति स्पेनिस', 'kxv-Latn': 'meksikoti spenis', 'kxv-Orya': 'ମେକ୍ସିକତି ସ୍ପେନିସ୍\u200c', 'kxv-Telu': 'మెక్సికాన్ స్పాస్పెనిస్', 'lb': 'Mexikanescht Spuenesch', 'lo': 'ເມັກຊິກັນ ສະແປນນິຊ', 'lrc': 'ئسپانیایی مئکزیک', 'mai': 'मैक्सिकन स्पेनिश', 'mi': 'Pāniora Mehikana', 'mk': 'мексикански шпански', 'ml': 'മെക്സിക്കൻ സ്പാനിഷ്', 'mn': 'испани хэл (Мексик)', 'mni': 'মেক্সিকান স্পেনিস', 'mni-Beng': 'মেক্সিকান স্পেনিস', 'mr': 'मेक्सिकन स्पॅनिश', 'ms': 'Sepanyol Mexico', 'mt': 'Spanjol tal-Messiku', 'my': 'စပိန် (မက္ကဆီကို)', 'mzn': 'مکزیک ِایسپانیولی', 'ne': 'मेक्सिकन स्पेनी', 'nqo': 'ߡߍߞߑߛߌߞ ߊߛߌߔߊ߲ߞߊ߲', 'om': 'Ispaanishii Meeksiikoo', 'or': 'ମେକ୍ସିକାନ ସ୍ପାନିସ୍\u200c', 'pcm': 'Mẹ́ksiko Spánish', 'pl': 'meksykański hiszpański', 'ps': 'ميکسيکي هسپانوي', 'rm': 'spagnol mexican', 'ru': 'мексиканский испанский', 'sa': 'मैक्सिकन स्पेनीय भाषा:', 'sat': 'ᱢᱮᱠᱥᱤᱠᱟᱱ ᱥᱯᱮᱱᱤᱥ', 'sat-Olck': 'ᱢᱮᱠᱥᱤᱠᱟᱱ ᱥᱯᱮᱱᱤᱥ', 'sc': 'ispagnolu messicanu', 'sd': 'ميڪسيڪين اسپيني', 'sd-Arab': 'ميڪسيڪين اسپيني', 'sd-Deva': 'मेक्सिकन स्पैनिश', 'si': 'මෙක්සිකානු ස්පාඤ්ඤ', 'sk': 'španielčina (mexická)', 'sl': 'mehiška španščina', 'smn': 'Meksiko espanjakielâ', 'so': 'Isbaanishka Mexico', 'sq': 'spanjishte meksikane', 'su': 'Spanyol Méksiko', 'su-Latn': 'Spanyol Méksiko', 'sv': 'mexikansk spanska', 'ta': 'மெக்ஸிகன் ஸ்பானிஷ்', 'te': 'మెక్సికన్ స్పానిష్', 'tg': 'испании мексикоӣ', 'th': 'สเปน - เม็กซิโก', 'to': 'lea fakasipeini-mekisikou', 'tpi': 'Meksikan Spenis', 'tr': 'Meksika İspanyolcası', 'ug': 'مېكسىكا ئىسپانچە', 'ur': 'میکسیکن ہسپانوی', 'uz': 'ispan (Meksika)', 'uz-Latn': 'ispan (Meksika)', 'wo': 'Español bu Meksik', 'xh': 'IsiSpanish SaseMexico', 'xnr': 'मैक्सिको आली स्पेनिश', 'yo': 'Èdè Sípáníìṣì (orílẹ̀-èdè Mẹ́síkò)', 'zh': '墨西哥西班牙语', 'zh-Hans': '墨西哥西班牙语', 'zu': 'isi-Mexican Spanish'},
    'ese': {'en': 'Ese Ejja'},
    'esg': {'en': 'Aheri Gondi'},
    'esh': {'en': 'Eshtehardi'},
    'esi': {'en': 'North Alaskan Inupiatun'},
    'esk': {'en': 'Northwest Alaska Inupiatun'},
    'esl': {'en': 'Egypt Sign Language'},
    'esm': {'en': 'Esuma'},
    'esn': {'en': 'Salvadoran Sign Language'},
    'eso': {'en': 'Estonian Sign Language'},
    'esq': {'en': 'Esselen'},
    'ess': {'en': 'Central Siberian Yupik'},
    'esu': {'am': 'ሴንተራል ዩፒክ', 'ast': 'yupik central', 'cs': 'jupikština (středoaljašská)', 'de': 'Zentral-Alaska-Yupik', 'en': 'Central Yupik', 'et': 'keskjupiki', 'fi': 'alaskanjupik', 'fr': 'youpik central', 'gd': 'Yupik Mheadhanach', 'hi-Latn': 'Central Yupik', 'ht': 'youpik central', 'it': 'yupik centrale', 'ja': '中央アラスカ・ユピック語', 'ksh': 'Zäntrahl-Juppik', 'lb': 'Yup’ik', 'lt': 'centrinės Aliaskos jupikų', 'mk': 'централнојупички', 'nb': 'sentralyupik', 'ne': 'केन्द्रीय युपिक', 'nl': 'Yupik', 'nn': 'sentralyupik', 'no': 'sentralyupik', 'pl': 'yupik środkowosyberyjski', 'sv': 'centralalaskisk jupiska', 'th': 'ยูพิกกลาง', 'to': 'lea fakaiūpiki-loloto', 'tr': 'Merkezi Yupikçe', 'vi': 'Tiếng Yupik Miền Trung', 'yue': '中尤皮克文', 'yue-Hans': '中尤皮克文', 'yue-Hant': '中尤皮克文', 'zh-Hant': '中尤皮克文'},
    'esx': {'en': 'Eskimo-Aleut languages'},
    'esx-esk-pro': {'en': 'Proto-Eskimo'},
    'esx-inu-pro': {'en': 'Proto-Inuit'},
    'esx-pro': {'en': 'Proto-Eskimo-Aleut'},
    'esy': {'en': 'Eskayan'},
    'et': {'af': 'Estnies', 'ak': 'Estonia kasa', 'am': 'ኢስቶኒያንኛ', 'ar': 'الإستونية', 'as': 'এষ্টোনিয়', 'ast': 'estoniu', 'az': 'eston', 'az-Arab': 'اِستونجا', 'az-Cyrl': 'естон', 'az-Latn': 'eston', 'bal-Latn': 'Estóniái', 'be': 'эстонская', 'be-tarask': 'эстонская', 'bg': 'естонски', 'bgn': 'استونیایی', 'bn': 'এস্তোনীয়', 'br': 'estoneg', 'brx': 'एस्ट’नियान', 'bs': 'estonski', 'bs-Cyrl': 'естонски', 'bs-Latn': 'estonski', 'ca': 'estonià', 'ca-ES-valencia': 'estonià', 'ccp': '𑄆𑄌𑄴𑄖𑄨𑄚𑄩𑄠𑄧', 'ce': 'эстонийн', 'chr': 'ᎡᏍᏙᏂᎠᏂ', 'ckb': 'ئیستۆنی', 'cs': 'estonština', 'cy': 'Estoneg', 'da': 'estisk', 'de': 'Estnisch', 'dsb': 'estišćina', 'dz': 'ཨེས་ཊོ་ནི་ཡཱན་ཁ', 'ee': 'estoniagbe', 'el': 'Εσθονικά', 'el-polyton': 'Ἐσθονικά', 'en': 'Estonian', 'en-Dsrt': '𐐀𐑅𐐻𐐬𐑌𐐨𐐲𐑌', 'eo': 'estona', 'es': 'estonio', 'et': 'eesti', 'eu': 'estoniera', 'fa': 'استونیایی', 'ff-Adlm': '𞤉𞤧𞤼𞤮𞤲𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'viro', 'fil': 'Estonian', 'fo': 'estiskt', 'fr': 'estonien', 'fur': 'eston', 'fy': 'Estlânsk', 'ga': 'Eastóinis', 'gd': 'Eastoinis', 'gl': 'estoniano', 'gsw': 'Eestnisch', 'gu': 'એસ્ટોનિયન', 'ha': 'Istoniyanci', 'he': 'אסטונית', 'hi': 'एस्टोनियाई', 'hi-Latn': 'Estonian', 'hr': 'estonski', 'hsb': 'estišćina', 'ht': 'estonien', 'hu': 'észt', 'hy': 'էստոներեն', 'ia': 'estoniano', 'id': 'Estonia', 'ie': 'estonian', 'ig': 'Estonian', 'is': 'eistneska', 'it': 'estone', 'ja': 'エストニア語', 'jv': 'Estonia', 'ka': 'ესტონური', 'kea': 'stonianu', 'kgp': 'enhtonĩjỹnũ', 'kk': 'эстон тілі', 'kk-Cyrl': 'эстон тілі', 'km': 'អេស្តូនី', 'kn': 'ಎಸ್ಟೊನಿಯನ್', 'ko': '에스토니아어', 'kok': 'इस्टोनियन', 'kok-Deva': 'इस्टोनियन', 'ks': 'ایسٹونیَن', 'ks-Arab': 'ایسٹونیَن', 'ksh': 'Äßnesch', 'ku': 'estonî', 'kxv': 'esṭoniyati', 'kxv-Deva': 'एस्टोनियाति', 'kxv-Latn': 'esṭoniyati', 'kxv-Orya': 'ଏସ୍ଟନିଆ', 'kxv-Telu': 'ఎస్టొ నియన్', 'ky': 'эстончо', 'lb': 'Estnesch', 'lkt': 'Estonia Iyápi', 'lld': 'eston', 'lo': 'ເອສໂຕນຽນ', 'lrc': 'ئستونیایی', 'lt': 'estų', 'lv': 'igauņu', 'mai': 'एसटोनियाई', 'mi': 'Etōniana', 'mk': 'естонски', 'ml': 'എസ്റ്റോണിയൻ', 'mn': 'эстони', 'mr': 'इस्टोनियन', 'ms': 'Estonia', 'mt': 'Estonjan', 'my': 'အက်စ်တိုးနီးယား', 'mzn': 'استونیایی', 'nb': 'estisk', 'ne': 'इस्टोनियन', 'nl': 'Estisch', 'nn': 'estisk', 'no': 'estisk', 'nqo': 'ߍߛߑߕߏߣߌߞߊ߲', 'om': 'Afaan Istooniya', 'or': 'ଏସ୍ତୋନିଆନ୍', 'os': 'естойнаг', 'pa': 'ਇਸਟੋਨੀਆਈ', 'pa-Guru': 'ਇਸਟੋਨੀਆਈ', 'pcm': 'Ẹstóniá Lángwej', 'pl': 'estoński', 'ps': 'حبشي', 'pt': 'estoniano', 'qu': 'Estonio Simi', 'rm': 'eston', 'ro': 'estonă', 'ru': 'эстонский', 'rw': 'Icyesitoniya', 'sah': 'Эстиэнийэлии', 'sc': 'èstone', 'sd': 'ايستونائي', 'sd-Arab': 'ايستونائي', 'se': 'esttegiella', 'si': 'එස්තෝනියානු', 'sk': 'estónčina', 'sl': 'estonščina', 'smn': 'eestikielâ', 'so': 'Istooniyaan', 'sq': 'estonisht', 'sr': 'естонски', 'sr-Cyrl': 'естонски', 'sr-Latn': 'estonski', 'sv': 'estniska', 'sw': 'Kiestonia', 'ta': 'எஸ்டோனியன்', 'te': 'ఎస్టోనియన్', 'tg': 'эстонӣ', 'th': 'เอสโตเนีย', 'ti': 'ኤስቶንኛ', 'tk': 'eston dili', 'to': 'lea fakaʻesitōnia', 'tr': 'Estonca', 'tt': 'эстон', 'ug': 'ئېستونچە', 'uk': 'естонська', 'ur': 'اسٹونین', 'uz': 'estoncha', 'uz-Cyrl': 'эстонча', 'uz-Latn': 'estoncha', 'vec': 'estonian', 'vi': 'Tiếng Estonia', 'wae': 'Estniš', 'wo': 'Estoñiye', 'xnr': 'एस्टोनिये आली', 'yi': 'עסטיש', 'yo': 'Èdè Estonia', 'yrl': 'eituniyanu', 'yue': '愛沙尼亞文', 'yue-Hans': '爱沙尼亚文', 'yue-Hant': '愛沙尼亞文', 'zh': '爱沙尼亚语', 'zh-Hans': '爱沙尼亚语', 'zh-Hant': '愛沙尼亞文', 'zu': 'isi-Estonia'},
    'etb': {'en': 'Etebi'},
    'etc': {'en': 'Etchemin'},
    'eth': {'en': 'Ethiopian Sign Language'},
    'etn': {'en': 'Eton (Vanuatu)'},
    'eto': {'en': 'Eton (Cameroon)'},
    'etr': {'en': 'Edolo'},
    'ets': {'en': 'Yekhee'},
    'ett': {'en': 'Etruscan'},
    'etu': {'en': 'Ejagham'},
    'etx': {'en': 'Eten'},
    'etz': {'en': 'Semimi'},
    'eu': {'af': 'Baskies', 'ak': 'Baske', 'am': 'ባስክኛ', 'ar': 'الباسكية', 'as': 'বাস্ক', 'ast': 'vascu', 'az': 'bask', 'az-Arab': 'باسکجا', 'az-Cyrl': 'баск', 'az-Latn': 'bask', 'bal-Latn': 'Bask', 'be': 'баскская', 'be-tarask': 'баскская', 'bg': 'баски', 'bgn': 'باسکی', 'bn': 'বাস্ক', 'br': 'euskareg', 'brx': 'बास्क', 'bs': 'baskijski', 'bs-Cyrl': 'баскијски', 'bs-Latn': 'baskijski', 'ca': 'basc', 'ca-ES-valencia': 'basc', 'ccp': '𑄝𑄌𑄴𑄇𑄧', 'ce': 'баскийн', 'chr': 'ᏆᏍᎨ', 'ckb': 'باسکی', 'cs': 'baskičtina', 'cy': 'Basgeg', 'da': 'baskisk', 'de': 'Baskisch', 'dsb': 'baskišćina', 'dz': 'བཱསཀ་ཁ', 'ee': 'basqugbe', 'el': 'Βασκικά', 'el-polyton': 'Βασκικά', 'en': 'Basque', 'en-Dsrt': '𐐒𐐰𐑅𐐿', 'eo': 'eŭska', 'es': 'euskera', 'et': 'baski', 'eu': 'euskara', 'fa': 'باسکی', 'ff-Adlm': '𞤄𞤢𞤧𞤳𞤢𞤪𞤢𞥄𞤪𞤫', 'fi': 'baski', 'fil': 'Basque', 'fo': 'baskiskt', 'fr': 'basque', 'fur': 'basc', 'fy': 'Baskysk', 'ga': 'Bascais', 'gd': 'Basgais', 'gl': 'éuscaro', 'gsw': 'Baskisch', 'gu': 'બાસ્ક', 'ha': 'Basque', 'he': 'בסקית', 'hi': 'बास्क', 'hi-Latn': 'Basque', 'hr': 'baskijski', 'hsb': 'baskišćina', 'ht': 'basque', 'hu': 'baszk', 'hy': 'բասկերեն', 'ia': 'basco', 'id': 'Basque', 'ig': 'Basque', 'is': 'baskneska', 'it': 'basco', 'ja': 'バスク語', 'jv': 'Basque', 'ka': 'ბასკური', 'kea': 'basku', 'kgp': 'manhku', 'kk': 'баск тілі', 'kk-Cyrl': 'баск тілі', 'km': 'បាសខ៍', 'kn': 'ಬಾಸ್ಕ್', 'ko': '바스크어', 'kok': 'बास्क', 'kok-Deva': 'बास्क', 'ks': 'باسک', 'ks-Arab': 'باسک', 'ksh': 'Baskesch', 'ku': 'baskî', 'kxv': 'bask', 'kxv-Deva': 'बास्क', 'kxv-Latn': 'bask', 'kxv-Orya': 'ବାସ୍କ', 'kxv-Telu': 'బాస్క్', 'ky': 'баскча', 'lb': 'Baskesch', 'lkt': 'Basque Iyápi', 'lld': 'basch', 'lo': 'ບັສກີ', 'lrc': 'باسکی', 'lt': 'baskų', 'lv': 'basku', 'mai': 'बास्क', 'mi': 'Pākihi', 'mk': 'баскиски', 'ml': 'ബാസ്\u200cക്', 'mn': 'баск', 'mr': 'बास्क', 'ms': 'Basque', 'mt': 'Bask', 'my': 'ဘာစ်ခ်', 'mzn': 'باسکی', 'nb': 'baskisk', 'ne': 'बास्क', 'nl': 'Baskisch', 'nn': 'baskisk', 'no': 'baskisk', 'nqo': 'ߓߊߛߑߞߌߞߊ߲', 'om': 'Afaan Baskuu', 'or': 'ବାସ୍କ୍ୱି', 'os': 'баскаг', 'pa': 'ਬਾਸਕ', 'pa-Guru': 'ਬਾਸਕ', 'pcm': 'Básk Lángwej', 'pl': 'baskijski', 'ps': 'باسکي', 'pt': 'basco', 'qu': 'Euskera Simi', 'rm': 'basc', 'ro': 'bască', 'ru': 'баскский', 'rw': 'Ikibasiki', 'sc': 'bascu', 'sd': 'باسق', 'sd-Arab': 'باسق', 'si': 'බාස්ක්', 'sk': 'baskičtina', 'sl': 'baskovščina', 'smn': 'baskikielâ', 'so': 'Basquu', 'sq': 'baskisht', 'sr': 'баскијски', 'sr-Cyrl': 'баскијски', 'sr-Latn': 'baskijski', 'sv': 'baskiska', 'sw': 'Kibaski', 'ta': 'பாஸ்க்', 'te': 'బాస్క్యూ', 'tg': 'баскӣ', 'th': 'บาสก์', 'ti': 'ባስክኛ', 'tk': 'bask dili', 'to': 'lea fakapāsiki', 'tr': 'Baskça', 'tt': 'баск', 'ug': 'باسكىچە', 'uk': 'баскська', 'ur': 'باسکی', 'uz': 'bask', 'uz-Cyrl': 'баскча', 'uz-Latn': 'bask', 'vec': 'basco', 'vi': 'Tiếng Basque', 'wae': 'Baskiš', 'wo': 'Bask', 'xnr': 'बास्क', 'yi': 'באַסקיש', 'yo': 'Èdè Baski', 'yrl': 'basiku', 'yue': '巴斯克文', 'yue-Hans': '巴斯克文', 'yue-Hant': '巴斯克文', 'zh': '巴斯克语', 'zh-Hans': '巴斯克语', 'zh-Hant': '巴斯克文', 'zu': 'isi-Basque'},
    'euq': {'en': 'Basque (family)'},
    'euq-pro': {'en': 'Proto-Basque'},
    'eve': {'en': 'Even'},
    'evh': {'en': 'Uvbie'},
    'evn': {'en': 'Evenki'},
    'ewo': {'af': 'Ewondo', 'am': 'ኤዎንዶ', 'ar': 'الإيوندو', 'as': 'ইওন্দো', 'ast': 'ewondo', 'az': 'evondo', 'az-Cyrl': 'евондо', 'az-Latn': 'evondo', 'bal-Latn': 'Ewondó', 'be': 'эвонда', 'be-tarask': 'эвонда', 'bg': 'евондо', 'bgn': 'اواندویی', 'bn': 'ইওন্ডো', 'br': 'ewondo', 'brx': 'एव’न्द’', 'bs': 'evondo', 'bs-Cyrl': 'евондо', 'bs-Latn': 'evondo', 'ca': 'ewondo', 'ca-ES-valencia': 'ewondo', 'ccp': '𑄄𑄃𑄮𑄚𑄴𑄓𑄮', 'ce': 'эвондо', 'chr': 'ᎡᏬᏂᏙ', 'ckb': 'ئێوۆندۆ', 'cs': 'ewondo', 'cy': 'Ewondo', 'da': 'ewondo', 'de': 'Ewondo', 'dsb': 'ewondo', 'el': 'Εγουόντο', 'el-polyton': 'Εγουόντο', 'en': 'Ewondo', 'es': 'ewondo', 'et': 'evondo', 'eu': 'ewondoa', 'ewo': 'ewondo', 'fa': 'اواندو', 'ff-Adlm': '𞤉𞤱𞤮𞤲𞤣𞤮𞥅𞤪𞤫', 'fi': 'ewondo', 'fil': 'Ewondo', 'fo': 'ewondo', 'fr': 'éwondo', 'fy': 'Ewondo', 'ga': 'Éabhandóis', 'gd': 'Ewondo', 'gl': 'ewondo', 'gsw': 'Ewondo', 'gu': 'ઇવોન્ડો', 'ha': 'Ewondo', 'he': 'אוונדו', 'hi': 'इवोन्डो', 'hi-Latn': 'Ewondo', 'hr': 'ewondo', 'hsb': 'ewondo', 'ht': 'éwondo', 'hu': 'evondo', 'hy': 'էվոնդո', 'ia': 'ewondo', 'id': 'Ewondo', 'ig': 'Ewondo', 'is': 'evondó', 'it': 'ewondo', 'ja': 'エウォンド語', 'jv': 'Ewondo', 'ka': 'ევონდო', 'kea': 'ewondo', 'kgp': 'evãnu', 'kk': 'эвондо тілі', 'kk-Cyrl': 'эвондо тілі', 'km': 'អ៊ីវ៉ុនដូ', 'kn': 'ಇವಾಂಡೋ', 'ko': '이원도어', 'kok': 'एवोंडो', 'kok-Deva': 'एवोंडो', 'ks': 'ایوونڈو', 'ks-Arab': 'ایوونڈو', 'ksh': 'Ewonndo', 'ku': 'ewondoyî', 'ky': 'эвондочо', 'lb': 'Ewondo', 'lo': 'ອີວອນດູ', 'lt': 'evondo', 'lv': 'evondu', 'mai': 'इवोनडो', 'mi': 'Ewāto', 'mk': 'евондо', 'ml': 'എവോൻഡോ', 'mn': 'эвондо', 'mr': 'इवोन्डो', 'ms': 'Ewondo', 'mt': 'Ewondo', 'my': 'အီဝန်ဒို', 'nb': 'ewondo', 'ne': 'इवोन्डो', 'nl': 'Ewondo', 'nn': 'ewondo', 'nnh': 'Shwóŋò pʉa Yɔɔnmendi', 'no': 'ewondo', 'nqo': 'ߋߥߏ߲ߘߏߞߊ߲', 'or': 'ଇୱୋଣ୍ଡୋ', 'pa': 'ਇਵੋਂਡੋ', 'pa-Guru': 'ਇਵੋਂਡੋ', 'pcm': 'Ẹwondo Lángwej', 'pl': 'ewondo', 'ps': 'اوونڊو', 'pt': 'ewondo', 'qu': 'Ewondo Simi', 'rm': 'ewondo', 'ro': 'ewondo', 'ru': 'эвондо', 'sc': 'ewondo', 'sd': 'اوانڊو', 'sd-Arab': 'اوانڊو', 'si': 'එවොන්ඩො', 'sk': 'ewondo', 'sl': 'evondovščina', 'smn': 'ewondokielâ', 'so': 'Eewondho', 'sq': 'euondoisht', 'sr': 'евондо', 'sr-Cyrl': 'евондо', 'sr-Latn': 'evondo', 'sv': 'ewondo', 'sw': 'Kiewondo', 'ta': 'எவோன்டோ', 'te': 'ఎవోండొ', 'th': 'อีวันโด', 'ti': 'ኤዎንዶ', 'tk': 'ewondo dili', 'to': 'lea fakaʻeuōnito', 'tr': 'Ewondo', 'ug': 'ئېۋوندوچە', 'uk': 'евондо', 'ur': 'ایوانڈو', 'uz': 'evondo', 'uz-Cyrl': 'эвондонча', 'uz-Latn': 'evondo', 'vec': 'ewondo', 'vi': 'Tiếng Ewondo', 'xnr': 'इवोन्डो', 'yo': 'Èdè Èwóǹdò', 'yrl': 'ewũdu', 'yue': '依汪都文', 'yue-Hans': '依汪都文', 'yue-Hant': '依汪都文', 'zh': '埃翁多语', 'zh-Hans': '埃翁多语', 'zh-Hant': '依汪都文', 'zu': 'isi-Ewondo'},
    'ext': {'ast': 'estremeñu', 'ca': 'extremeny', 'ca-ES-valencia': 'extremeny', 'cs': 'extremadurština', 'cy': 'Extremadureg', 'de': 'Extremadurisch', 'en': 'Extremaduran', 'et': 'estremenju', 'fi': 'extremadura', 'fr': 'estrémègne', 'gd': 'Cànan na h-Extremadura', 'hi-Latn': 'Extremaduran', 'ht': 'estrémègne', 'it': 'estremegno', 'ja': 'エストレマドゥーラ語', 'lb': 'Extremaduresch', 'lt': 'ispanų kalbos Ekstremadūros tarmė', 'mk': 'екстремадурски', 'nb': 'ekstremaduransk', 'ne': 'एक्सट्रेमादुराली', 'nl': 'Extremeens', 'nn': 'ekstremaduransk', 'no': 'ekstremaduransk', 'pl': 'estremadurski', 'sv': 'extremaduriska', 'th': 'เอกซ์เตรมาดูรา', 'to': 'lea fakaʻekisitematula', 'tr': 'Ekstremadura Dili', 'vi': 'Tiếng Extremadura', 'yue': '埃斯特雷馬杜拉文', 'yue-Hans': '埃斯特雷马杜拉文', 'yue-Hant': '埃斯特雷馬杜拉文', 'zh-Hant': '埃斯特雷馬杜拉文'},
    'eya': {'en': 'Eyak'},
    'eyo': {'en': 'Keiyo'},
    'eza': {'en': 'Ezaa'},
    'eze': {'en': 'Uzekwe'},
    'fa': {'af': 'Persies', 'agq': 'Kpɛɛshìa', 'ak': 'Pɛɛhyia kasa', 'am': 'ፐርሺያኛ', 'ar': 'الفارسية', 'as': 'ফাৰ্ছী', 'asa': 'Kiajemi', 'ast': 'persa', 'az': 'fars', 'az-Arab': 'فارسجا', 'az-Cyrl': 'фарс', 'az-Latn': 'fars', 'bal-Latn': 'Pársi', 'bas': 'Hɔp u pɛrsìà', 'be': 'фарсі', 'be-tarask': 'фарсі', 'bem': 'Ichi Pesia', 'bez': 'Hiajemi', 'bg': 'персийски', 'bgn': 'پارسی', 'bm': 'perisanikan', 'bn': 'ফার্সি', 'br': 'perseg', 'brx': 'फार्सी', 'bs': 'perzijski', 'bs-Cyrl': 'персијски', 'bs-Latn': 'perzijski', 'ca': 'persa', 'ca-ES-valencia': 'persa', 'ccp': '𑄜𑄢𑄴𑄥𑄨', 'ce': 'гӀажарийн', 'cgg': 'Orupaasiya', 'chr': 'ᏇᏏᎠᏂ', 'ckb': 'فارسی', 'cs': 'perština', 'cy': 'Perseg', 'da': 'persisk', 'dav': 'Kiajemi', 'de': 'Persisch', 'dje': 'Farsi senni', 'dsb': 'persišćina', 'dyo': 'persan', 'dz': 'པར་ཤི་ཡཱན་ཁ', 'ebu': 'Kĩanjemi', 'ee': 'persiagbe', 'el': 'Περσικά', 'el-polyton': 'Περσικά', 'en': 'Persian', 'eo': 'persa', 'es': 'persa', 'et': 'pärsia', 'eu': 'persiera', 'ewo': 'ǹkɔ́bɔ fɛ́rəsian', 'fa': 'فارسی', 'ff': 'Perseere', 'ff-Adlm': '𞤊𞤢𞥄𞤪𞤧𞤭𞤴𞤢𞤲𞤳𞤮𞥅𞤪𞤫', 'ff-Latn': 'Perseere', 'fi': 'persia', 'fil': 'Persian', 'fo': 'persiskt', 'fr': 'persan', 'fur': 'persian', 'fy': 'Perzysk', 'ga': 'Peirsis', 'gd': 'Peirsis', 'gl': 'persa', 'gsw': 'Persisch', 'gu': 'ફારસી', 'guz': 'Kiajemi', 'ha': 'Farisanci', 'he': 'פרסית', 'hi': 'फ़ारसी', 'hi-Latn': 'Faarsi', 'hr': 'perzijski', 'hsb': 'persišćina', 'ht': 'persan', 'hu': 'perzsa', 'hy': 'պարսկերեն', 'ia': 'persa', 'id': 'Persia', 'ie': 'persian', 'ig': 'Asụsụ Persia', 'is': 'persneska', 'it': 'persiano', 'ja': 'ペルシア語', 'jmc': 'Kyiajemi', 'jv': 'Persia', 'ka': 'სპარსული', 'kab': 'Tafarisit', 'kam': 'Kiajemi', 'kde': 'Chiajemi', 'kea': 'persa', 'kgp': 'pérsa', 'khq': 'Farsi senni', 'ki': 'Kiajemi', 'kk': 'парсы тілі', 'kk-Cyrl': 'парсы тілі', 'kln': 'kutitab Persia', 'km': 'ភឺសៀន', 'kn': 'ಪರ್ಶಿಯನ್', 'ko': '페르시아어', 'kok': 'पर्शियन', 'kok-Deva': 'पर्शियन', 'ks': 'فارسی', 'ks-Arab': 'فارسی', 'ksb': 'Kiajemi', 'ksf': 'ripɛrsán', 'ksh': 'Pärsesch', 'ku': 'farisî', 'kxv': 'persiati', 'kxv-Deva': 'पर्सियन', 'kxv-Latn': 'persiati', 'kxv-Orya': 'ପର୍ସିୟାନ୍', 'kxv-Telu': 'పర్సియన్', 'ky': 'фарсча', 'lag': 'Kɨajéemi', 'lb': 'Persesch', 'lg': 'Luperusi', 'lkt': 'Persian Iyápi', 'lld': 'persian', 'ln': 'lipelésanɛ', 'lo': 'ເປີຊຽນ', 'lrc': 'فارسی', 'lt': 'persų', 'lu': 'Mpepajemi', 'luo': 'Kiajemi', 'luy': 'Kiajemi', 'lv': 'persiešu', 'mai': 'परसियन', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lpersia', 'mer': 'Kĩpasia', 'mfe': 'persan', 'mg': 'Persa', 'mgh': 'Iajemi', 'mi': 'Pāhiana', 'mk': 'персиски', 'ml': 'പേർഷ്യൻ', 'mn': 'перс', 'mr': 'फारसी', 'ms': 'Parsi', 'mt': 'Persjan', 'mua': 'Persia', 'my': 'ပါရှန်', 'mzn': 'فارسی', 'naq': 'Persiaǁî gowab', 'nb': 'persisk', 'nd': 'isi-Pheshiyani', 'ne': 'फारसी', 'nl': 'Perzisch', 'nmg': 'Kiɛl pɛrsia', 'nn': 'persisk', 'no': 'persisk', 'nqo': 'ߝߊ߯ߙߛߌߞߊ߲', 'nus': 'Thok perthiani', 'nyn': 'Orupaasiya', 'om': 'Afaan Persia', 'or': 'ପର୍ସିଆନ୍', 'os': 'персайнаг', 'pa': 'ਫ਼ਾਰਸੀ', 'pa-Guru': 'ਫ਼ਾਰਸੀ', 'pcm': 'Pẹ́shiá Lángwej', 'pl': 'perski', 'ps': 'فارسي', 'pt': 'persa', 'qu': 'Persa Simi', 'rm': 'persian', 'rn': 'Igiperisi', 'ro': 'persană', 'rof': 'Kiajemi', 'ru': 'персидский', 'rw': 'Inyeperisi', 'rwk': 'Kyiajemi', 'sah': 'Пиэристии', 'saq': 'Kiajemi', 'sbp': 'Ishiajemi', 'sc': 'persianu', 'sd': 'فارسي', 'sd-Arab': 'فارسي', 'se': 'persijagiella', 'seh': 'persa', 'ses': 'Farsi senni', 'sg': 'Farsî', 'shi': 'ⵜⴰⴼⵓⵔⵙⵉⵜ', 'shi-Latn': 'Tafursit', 'shi-Tfng': 'ⵜⴰⴼⵓⵔⵙⵉⵜ', 'si': 'පර්සියානු', 'sk': 'perzština', 'sl': 'perzijščina', 'smn': 'persiakielâ', 'sn': 'chiPeshiya', 'so': 'Faarisi', 'sq': 'persisht', 'sr': 'персијски', 'sr-Cyrl': 'персијски', 'sr-Latn': 'persijski', 'sv': 'persiska', 'sw': 'Kiajemi', 'syr': 'ܦܪܣܝܬ', 'ta': 'பெர்ஷியன்', 'te': 'పర్షియన్', 'teo': 'Kiajemi', 'tg': 'форсӣ', 'th': 'เปอร์เซีย', 'ti': 'ፋርስኛ', 'tk': 'pars dili', 'to': 'lea fakapēsia', 'tr': 'Farsça', 'tt': 'фарсы', 'twq': 'Farsi senni', 'tzm': 'Tafarisit', 'ug': 'پارسچە', 'uk': 'перська', 'ur': 'فارسی', 'uz': 'fors', 'uz-Arab': 'دری', 'uz-Cyrl': 'форсий', 'uz-Latn': 'fors', 'vai': 'ꗨꗡꔻꘂꘋ', 'vai-Latn': 'Pɛɛsiyɛŋ', 'vai-Vaii': 'ꗨꗡꔻꘂꘋ', 'vec': 'persian', 'vi': 'Tiếng Ba Tư', 'vun': 'Kyiajemi', 'wae': 'Persiš', 'wo': 'Pers', 'xnr': 'फ़ारसी', 'xog': 'Oluperusi', 'yav': 'nupɛ́lisɛ', 'yi': 'פּערסיש', 'yo': 'Èdè Pasia', 'yrl': 'perisa', 'yue': '波斯文', 'yue-Hans': '波斯文', 'yue-Hant': '波斯文', 'zgh': 'ⵜⴰⴼⵓⵔⵙⵉⵜ', 'zh': '波斯语', 'zh-Hans': '波斯语', 'zh-Hant': '波斯文', 'zu': 'isi-Persian'},
    'fa-AF': {'af': 'Dari', 'am': 'ዳሪ', 'ar': 'الدارية', 'as': 'দাৰি', 'az': 'dari', 'az-Latn': 'dari', 'bal-Latn': 'Pársi (AF)', 'be': 'дары', 'be-tarask': 'дары', 'bg': 'дари', 'bn': 'দারি', 'br': 'dareg', 'brx': 'डारी', 'bs': 'dari', 'bs-Latn': 'dari', 'ca': 'dari', 'ca-ES-valencia': 'dari', 'chr': 'ᏓᎵ', 'cs': 'darí', 'cy': 'Dari', 'da': 'dari', 'de': 'Dari', 'dsb': 'dari', 'el': 'Νταρί', 'el-polyton': 'Νταρί', 'en': 'Dari', 'es': 'darí', 'et': 'dari', 'eu': 'daria', 'fa': 'دری', 'ff-Adlm': '𞤁𞤢𞤪𞤭𞥅𞤪𞤫', 'fi': 'dari', 'fil': 'Dari', 'fo': 'dari', 'fr': 'dari', 'ga': 'Dairis', 'gd': 'Dari', 'gl': 'dari', 'gu': 'ડારી', 'ha': 'Farisanci na Afaganistan', 'he': 'דארי', 'hi': 'दारी', 'hi-Latn': 'Dari', 'hr': 'dari', 'hsb': 'dari', 'ht': 'dari', 'hu': 'dari', 'hy': 'դարի', 'ia': 'dari', 'id': 'Persia Dari', 'ig': 'Dari', 'is': 'darí', 'it': 'dari', 'ja': 'ダリー語', 'ka': 'დარი', 'kk': 'дари тілі', 'kk-Cyrl': 'дари тілі', 'km': 'ដារី', 'kn': 'ದರಿ', 'ko': '다리어', 'kok': 'दारी', 'kok-Deva': 'दारी', 'ku': 'derî', 'kxv': 'ḍari', 'kxv-Deva': 'डॉरि', 'kxv-Latn': 'ḍari', 'kxv-Orya': 'ଡରି', 'kxv-Telu': 'డారి', 'ky': 'дари', 'lld': 'dari', 'lo': 'ດາຣີ', 'lv': 'darī', 'mai': 'परसियन डारी', 'mi': 'Tāri', 'mk': 'дари', 'ml': 'ഡാരി', 'mr': 'दारी', 'ms': 'Dari', 'my': 'ဒါရီ', 'nb': 'dari', 'ne': 'दारी', 'nl': 'Dari', 'nn': 'dari', 'no': 'dari', 'nqo': 'ߘߊߙߌߞߊ߲', 'or': 'ଦାରି', 'pa': 'ਦਾਰੀ', 'pa-Guru': 'ਦਾਰੀ', 'pcm': 'Dári', 'pl': 'dari', 'ps': 'دری (افغانستان)', 'pt': 'dari', 'qu': 'Dari Simi', 'rm': 'dari', 'ro': 'dari', 'ru': 'дари', 'sc': 'dari', 'sd': 'دري', 'sd-Arab': 'دري', 'si': 'ඩාරි', 'sk': 'daríjčina', 'sl': 'darijščina', 'so': 'Faarsi', 'sq': 'darisht', 'sr': 'дари', 'sr-Cyrl': 'дари', 'sr-Latn': 'dari', 'sv': 'dari', 'sw': 'Kiajemi (Afganistani)', 'ta': 'தாரி', 'te': 'డారి', 'th': 'ดารี', 'ti': 'ዳሪ', 'tk': 'dari dili', 'to': 'lea fakapēsia (ʻtalī)', 'tr': 'Darice', 'uk': 'дарі', 'ur': 'دری', 'uz': 'dari', 'uz-Latn': 'dari', 'vec': 'dari', 'vi': 'Tiếng Dari', 'xnr': 'दारी', 'yue': '達里文', 'yue-Hans': '达里文', 'yue-Hant': '達里文', 'zh': '达里语', 'zh-Hans': '达里语', 'zu': 'isi-Dari'},
    'faa': {'en': 'Fasu'},
    'fab': {'en': "Fa d'Ambu"},
    'fad': {'en': 'Wagi'},
    'faf': {'en': 'Fagani'},
    'fag': {'en': 'Finongan'},
    'fah': {'en': 'Baissa Fali'},
    'fai': {'en': 'Faiwol'},
    'faj': {'en': 'Faita'},
    'fak': {'en': 'Fang (Cameroon)'},
    'fal': {'en': 'South Fali'},
    'fam': {'en': 'Fam'},
    'fan': {'ar': 'الفانج', 'ast': 'fang', 'az': 'fang', 'az-Latn': 'fang', 'bg': 'фанг', 'bn': 'ফ্যাঙ্গ', 'br': 'fang', 'brx': 'फाँग्', 'bs': 'fang', 'bs-Cyrl': 'фанг', 'bs-Latn': 'fang', 'ca': 'fang', 'ca-ES-valencia': 'fang', 'ccp': '𑄜𑄳𑄠𑄋𑄴𑄉𑄧', 'cs': 'fang', 'da': 'fang', 'de': 'Pangwe', 'el': 'Φανγκ', 'el-polyton': 'Φανγκ', 'en': 'Fang', 'es': 'fang', 'et': 'fangi', 'fa': 'فانگی', 'fi': 'fang', 'fr': 'fang', 'fy': 'Fang', 'gd': 'Fang', 'gsw': 'Pangwe-Schpraach', 'gu': 'ફેંગ', 'he': 'פנג', 'hi': 'फैन्ग', 'hi-Latn': 'Fang', 'hr': 'fang', 'ht': 'fang', 'hu': 'fang', 'id': 'Fang', 'is': 'fang', 'it': 'fang', 'ja': 'ファング語', 'kgp': 'fỹnge', 'kn': 'ಫಾಂಗ್', 'ko': '팡그어', 'ks': 'فینگ', 'ks-Arab': 'فینگ', 'lb': 'Pangwe-Sprooch', 'lo': 'ແຟງ', 'lt': 'fangų', 'lv': 'fangu', 'mk': 'фанг', 'ml': 'ഫങ്', 'mr': 'फँग', 'mt': 'Fang', 'nb': 'fang', 'ne': 'फाङ', 'nl': 'Fang', 'nn': 'fang', 'no': 'fang', 'or': 'ଫାଙ୍ଗ', 'pl': 'fang', 'pt': 'fangue', 'rm': 'fang', 'ro': 'fang', 'ru': 'фанг', 'sk': 'fangčina', 'sl': 'fangijščina', 'sr': 'фанг', 'sr-Cyrl': 'фанг', 'sr-Latn': 'fang', 'sv': 'fang', 'ta': 'ஃபேங்க்', 'te': 'ఫాంగ్', 'th': 'ฟอง', 'to': 'lea fakafangi', 'tr': 'Fang', 'ug': 'فاڭچە', 'uk': 'фанг', 'vi': 'Tiếng Fang', 'yrl': 'fãge', 'yue': '芳族文', 'yue-Hans': '芳族文', 'yue-Hant': '芳族文', 'zh': '芳格语', 'zh-Hans': '芳格语', 'zh-Hant': '芳族文'},
    'fap': {'en': 'Paloor'},
    'far': {'en': 'Fataleka'},
    'fat': {'ar': 'الفانتي', 'ast': 'fanti', 'az': 'fanti', 'az-Latn': 'fanti', 'bg': 'фанти', 'bn': 'ফান্তি', 'br': 'fanti', 'brx': 'फाँटी', 'bs': 'fanti', 'bs-Cyrl': 'фанти', 'bs-Latn': 'fanti', 'ca': 'fanti', 'ca-ES-valencia': 'fanti', 'ccp': '𑄜𑄚𑄴𑄖𑄨', 'cs': 'fantština', 'cy': 'Ffanti', 'da': 'fanti', 'de': 'Fanti', 'el': 'Φάντι', 'el-polyton': 'Φάντι', 'en': 'Fanti', 'es': 'fanti', 'et': 'fanti', 'fa': 'فانتیایی', 'fi': 'fanti', 'fr': 'fanti', 'fy': 'Fanti', 'gd': 'Fanti', 'gsw': 'Fanti-Schpraach', 'gu': 'ફન્ટી', 'he': 'פאנטי', 'hi': 'फन्टी', 'hi-Latn': 'Fanti', 'hr': 'fanti', 'ht': 'fanti', 'hu': 'fanti', 'id': 'Fanti', 'is': 'fantí', 'it': 'fanti', 'ja': 'ファンティー語', 'kgp': 'fỹti', 'kn': 'ಫಾಂಟಿ', 'ko': '판티어', 'ks': 'فانٹی', 'ks-Arab': 'فانٹی', 'lb': 'Fanti-Sprooch', 'lo': 'ແຟນຕີ', 'lt': 'fanti', 'lv': 'fantu', 'mk': 'фанти', 'ml': 'ഫാന്റി', 'mr': 'फन्टी', 'mt': 'Fanti', 'nb': 'fanti', 'ne': 'फान्टी', 'nl': 'Fanti', 'nn': 'fanti', 'no': 'fanti', 'or': 'ଫାଣ୍ଟି', 'pl': 'fanti', 'pt': 'fanti', 'rm': 'fanti', 'ro': 'fanti', 'ru': 'фанти', 'sk': 'fanti', 'sl': 'fantijščina', 'sr': 'фанти', 'sr-Cyrl': 'фанти', 'sr-Latn': 'fanti', 'sv': 'fanti', 'ta': 'ஃபான்டி', 'te': 'ఫాంటి', 'th': 'ฟันติ', 'to': 'lea fakafanitē', 'tr': 'Fanti', 'ug': 'فانتىچە', 'uk': 'фанті', 'vi': 'Tiếng Fanti', 'yrl': 'fãti', 'yue': '芳蒂文', 'yue-Hans': '芳蒂文', 'yue-Hant': '芳蒂文', 'zh': '芳蒂语', 'zh-Hans': '芳蒂语', 'zh-Hant': '芳蒂文'},
    'fau': {'en': 'Fayu'},
    'fax': {'en': 'Fala'},
    'fay': {'en': 'Southwestern Fars'},
    'faz': {'en': 'Northwestern Fars'},
    'fbl': {'en': 'West Albay Bikol'},
    'fcs': {'en': 'Quebec Sign Language'},
    'fer': {'en': 'Feroge'},
    'ff': {'af': 'Fulah', 'ak': 'Fula kasa', 'am': 'ፉላኒኛ', 'ar': 'الفولانية', 'as': 'ফুলাহ', 'ast': 'fulah', 'az': 'fula', 'az-Cyrl': 'фула', 'az-Latn': 'fula', 'bal-Latn': 'Pulá', 'be': 'фула', 'be-tarask': 'фула', 'bg': 'фула', 'bgn': 'فولایی', 'bn': 'ফুলাহ্', 'br': 'fula', 'brx': 'फुलाह', 'bs': 'fulah', 'bs-Cyrl': 'фулах', 'bs-Latn': 'fulah', 'ca': 'ful', 'ca-ES-valencia': 'ful', 'ccp': '𑄜𑄪𑄣𑄳𑄦', 'ce': 'фулах', 'chr': 'ᏊᎳᏂ', 'ckb': 'فوولایی', 'cs': 'fulbština', 'cy': 'Ffwla', 'da': 'fulah', 'de': 'Ful', 'dsb': 'fulbšćina', 'el': 'Φουλά', 'el-polyton': 'Φουλά', 'en': 'Fula', 'es': 'fula', 'et': 'fula', 'eu': 'fula', 'fa': 'فولانی', 'ff': 'Pulaar', 'ff-Adlm': '𞤆𞤵𞤤𞤢𞤪', 'ff-Latn': 'Pulaar', 'fi': 'fulani', 'fil': 'Fulah', 'fo': 'fulah', 'fr': 'peul', 'fur': 'fulah', 'fy': 'Fulah', 'ga': 'Fuláinis', 'gd': 'Fulah', 'gl': 'fula', 'gsw': 'Ful', 'gu': 'ફુલા', 'ha': 'Fula', 'he': 'פולה', 'hi': 'फुलाह', 'hi-Latn': 'Fulah', 'hr': 'fula', 'hsb': 'fulbšćina', 'ht': 'peul', 'hu': 'fulani', 'hy': 'ֆուլահ', 'ia': 'fula', 'id': 'Fula', 'ig': 'Fula', 'is': 'fúla', 'it': 'fulah', 'ja': 'フラ語', 'jv': 'Fula', 'ka': 'ფულა', 'kea': 'fula', 'kgp': 'fura', 'kk': 'фула тілі', 'kk-Cyrl': 'фула тілі', 'km': 'ហ្វ៊ូឡា', 'kn': 'ಫುಲಾ', 'ko': '풀라어', 'kok': 'फुला', 'kok-Deva': 'फुला', 'ks': 'فُلاہ', 'ks-Arab': 'فُلاہ', 'ksh': 'Fulfulde', 'ku': 'fulahî', 'ky': 'фулача', 'lb': 'Ful', 'lld': 'fula', 'lo': 'ຟູລາ', 'lt': 'fulahų', 'lv': 'fulu', 'mai': 'फुलाह', 'mi': 'Whūra', 'mk': 'фула', 'ml': 'ഫുല', 'mn': 'фула', 'mr': 'फुलाह', 'ms': 'Fulah', 'mt': 'Fulah', 'my': 'ဖူလာ', 'nb': 'fulfulde', 'ne': 'फुलाह', 'nl': 'Fulah', 'nn': 'fulfulde', 'nnh': 'Shwóŋò menkesaŋ', 'no': 'fulfulde', 'nqo': 'ߝߎߟߊߞߊ߲', 'om': 'Fula', 'or': 'ଫୁଲାହ', 'pa': 'ਫੁਲਾਹ', 'pa-Guru': 'ਫੁਲਾਹ', 'pcm': 'Fúlaní Lángwej', 'pl': 'fulani', 'ps': 'فولاح', 'pt': 'fula', 'qu': 'Fulah Simi', 'rm': 'fulah', 'ro': 'fulah', 'ru': 'фулах', 'sc': 'fulah', 'sd': 'فلاهه', 'sd-Arab': 'فلاهه', 'si': 'ෆුලාහ්', 'sk': 'fulbčina', 'sl': 'fulščina', 'smn': 'fulakielâ', 'so': 'Fuulah', 'sq': 'fulaisht', 'sr': 'фула', 'sr-Cyrl': 'фула', 'sr-Latn': 'fula', 'sv': 'fulani', 'sw': 'Kifulani', 'syr': 'ܦܘܠܐܗܝܬ', 'ta': 'ஃபுலா', 'te': 'ఫ్యుల', 'tg': 'фулаҳ', 'th': 'ฟูลาห์', 'ti': 'ፉላ', 'tk': 'fula dili', 'to': 'lea fakafulā', 'tr': 'Fula dili', 'tt': 'фула', 'ug': 'فۇلاھچە', 'uk': 'фула', 'ur': 'فولہ', 'uz': 'fula', 'uz-Cyrl': 'фулаҳ', 'uz-Latn': 'fula', 'vec': 'fula', 'vi': 'Tiếng Fulah', 'wo': 'Pël', 'xnr': 'फुलाह', 'yo': 'Èdè Fúlàní', 'yrl': 'fura', 'yue': '富拉文', 'yue-Hans': '富拉文', 'yue-Hant': '富拉文', 'zh': '富拉语', 'zh-Hans': '富拉语', 'zh-Hant': '富拉文', 'zu': 'isi-Fulah'},
    'ffi': {'en': 'Foia Foia'},
    'ffm': {'en': 'Maasina Fulfulde'},
    'fgr': {'en': 'Fongoro'},
    'fi': {'af': 'Fins', 'ak': 'Finlande kasa', 'am': 'ፊንላንድኛ', 'ar': 'الفنلندية', 'as': 'ফিনিচ', 'ast': 'finlandés', 'az': 'fin', 'az-Arab': 'فینجه', 'az-Cyrl': 'фин', 'az-Latn': 'fin', 'bal-Latn': 'Fenlándi', 'be': 'фінская', 'be-tarask': 'фінская', 'bg': 'фински', 'bgn': 'فنلاندی', 'bn': 'ফিনিশ', 'br': 'finneg', 'brx': 'फिन्निस', 'bs': 'finski', 'bs-Cyrl': 'фински', 'bs-Latn': 'finski', 'ca': 'finès', 'ca-ES-valencia': 'finès', 'ccp': '𑄜𑄨𑄚𑄨𑄌𑄴', 'ce': 'финнийн', 'chr': 'ᏈᏂᏍ', 'ckb': 'فینلەندی', 'cs': 'finština', 'cy': 'Ffinneg', 'da': 'finsk', 'de': 'Finnisch', 'dsb': 'finšćina', 'dz': 'ཕི་ནིཤ་ཁ', 'ee': 'finlanɖgbe', 'el': 'Φινλανδικά', 'el-polyton': 'Φινλανδικά', 'en': 'Finnish', 'eo': 'finna', 'es': 'finés', 'et': 'soome', 'eu': 'finlandiera', 'fa': 'فنلاندی', 'ff-Adlm': '𞤊𞤫𞤲𞤭𞤧𞤪𞤫', 'fi': 'suomi', 'fil': 'Finnish', 'fo': 'finskt', 'fr': 'finnois', 'fur': 'finlandês', 'fy': 'Finsk', 'ga': 'Fionlainnis', 'gd': 'Fionnlannais', 'gl': 'finés', 'gsw': 'Finnisch', 'gu': 'ફિનિશ', 'ha': 'Yaren mutanen Finland', 'he': 'פינית', 'hi': 'फ़िनिश', 'hi-Latn': 'Finnish', 'hr': 'finski', 'hsb': 'finšćina', 'ht': 'finnois', 'hu': 'finn', 'hy': 'ֆիններեն', 'ia': 'finnese', 'id': 'Suomi', 'ig': 'Finnish', 'is': 'finnska', 'it': 'finlandese', 'ja': 'フィンランド語', 'jv': 'Suomi', 'ka': 'ფინური', 'kea': 'finlandes', 'kgp': 'fĩranẽj', 'kk': 'фин тілі', 'kk-Cyrl': 'фин тілі', 'km': 'ហ្វាំងឡង់', 'kn': 'ಫಿನ್ನಿಶ್', 'ko': '핀란드어', 'kok': 'फिनिश', 'kok-Deva': 'फिनिश', 'ks': 'فِنِش', 'ks-Arab': 'فِنِش', 'ksh': 'Finnesch', 'ku': 'fînî', 'kxv': 'pinnis', 'kxv-Deva': 'प़िनिस', 'kxv-Latn': 'pinnis', 'kxv-Orya': 'ପିନିସ୍', 'kxv-Telu': 'పినిస్', 'ky': 'финче', 'lb': 'Finnesch', 'lkt': 'Finnish Iyápi', 'lld': 'finlandesc', 'lo': 'ຟິນນິຊ', 'lrc': 'فأنلاندی', 'lt': 'suomių', 'lv': 'somu', 'mai': 'फिनिश', 'mi': 'Whinirānia', 'mk': 'фински', 'ml': 'ഫിന്നിഷ്', 'mn': 'фин', 'mr': 'फिनिश', 'ms': 'Finland', 'mt': 'Finlandiż', 'my': 'ဖင်လန်', 'mzn': 'فینیش', 'nb': 'finsk', 'ne': 'फिनिस', 'nl': 'Fins', 'nn': 'finsk', 'no': 'finsk', 'nqo': 'ߝߍߣߏߥߊߞߊ߲', 'om': 'Afaan Fiilaandi', 'or': 'ଫିନ୍ନିସ୍', 'os': 'финнаг', 'pa': 'ਫਿਨਿਸ਼', 'pa-Guru': 'ਫਿਨਿਸ਼', 'pcm': 'Fínísh Lángwej', 'pl': 'fiński', 'ps': 'فینلنډي', 'pt': 'finlandês', 'qu': 'Fines Simi', 'rm': 'finlandais', 'ro': 'finlandeză', 'ru': 'финский', 'rw': 'Igifinilande', 'sah': 'Пииннии', 'sc': 'finlandesu', 'sd': 'فنش', 'sd-Arab': 'فنش', 'se': 'suomagiella', 'si': 'ෆින්ලන්ත', 'sk': 'fínčina', 'sl': 'finščina', 'smn': 'suomâkielâ', 'so': 'Finishka', 'sq': 'finlandisht', 'sr': 'фински', 'sr-Cyrl': 'фински', 'sr-Latn': 'finski', 'sv': 'finska', 'sw': 'Kifini', 'syr': 'ܦܝܢܠܢܕܝܬ', 'ta': 'ஃபின்னிஷ்', 'te': 'ఫిన్నిష్', 'tg': 'финӣ', 'th': 'ฟินแลนด์', 'ti': 'ፊንላንድኛ', 'tk': 'fin dili', 'to': 'lea fakafinilani', 'tr': 'Fince', 'tt': 'фин', 'ug': 'فىنچە', 'uk': 'фінська', 'ur': 'فینیش', 'uz': 'fincha', 'uz-Cyrl': 'финча', 'uz-Latn': 'fincha', 'vec': 'filandeze', 'vi': 'Tiếng Phần Lan', 'wae': 'Finiš', 'wo': 'Feylànde', 'xnr': 'फ़िनिश', 'yi': 'פֿיניש', 'yo': 'Èdè Finisi', 'yrl': 'firãdes', 'yue': '芬蘭文', 'yue-Hans': '芬兰文', 'yue-Hant': '芬蘭文', 'zh': '芬兰语', 'zh-Hans': '芬兰语', 'zh-Hant': '芬蘭文', 'zu': 'isi-Finnish'},
    'fia': {'en': 'Nobiin'},
    'fie': {'en': 'Fyer'},
    'fif': {'en': 'Faifi'},
    'fil': {'af': 'Filippyns', 'ak': 'Filipin kasa', 'am': 'ፊሊፒንኛ', 'ar': 'الفلبينية', 'as': 'ফিলিপিনো', 'ast': 'filipín', 'az': 'filippin', 'az-Arab': 'فیلیپینجه', 'az-Cyrl': 'филиппин', 'az-Latn': 'filippin', 'bal-Latn': 'Pelpini', 'be': 'філіпінская', 'be-tarask': 'філіпінская', 'bg': 'филипински', 'bgn': 'فلیپینی', 'bn': 'ফিলিপিনো', 'br': 'filipineg', 'brx': 'फिलिपिन’', 'bs': 'filipino', 'bs-Cyrl': 'филипински', 'bs-Latn': 'filipino', 'ca': 'filipí', 'ca-ES-valencia': 'filipí', 'ccp': '𑄜𑄨𑄣𑄨𑄛𑄨𑄚𑄮', 'ce': 'филиппинийн', 'chr': 'ᎠᏈᎵᎩ', 'ckb': 'فیلیپینی', 'cs': 'filipínština', 'cy': 'Ffilipineg', 'da': 'filippinsk', 'de': 'Filipino', 'dsb': 'filipinšćina', 'dz': 'ཕི་ལི་པི་ནོ་ཁ', 'ee': 'filipingbe', 'el': 'Φιλιππινικά', 'el-polyton': 'Φιλιππινικά', 'en': 'Filipino', 'eo': 'filipina', 'es': 'filipino', 'et': 'filipiini', 'eu': 'Filipinera', 'fa': 'فیلیپینی', 'ff-Adlm': '𞤊𞤭𞤤𞤭𞤨𞤭𞤲𞤮𞥅𞤪𞤫', 'fi': 'filipino', 'fil': 'Filipino', 'fo': 'filipiniskt', 'fr': 'filipino', 'fur': 'filipin', 'fy': 'Filipynsk', 'ga': 'Filipínis', 'gd': 'Filipinis', 'gl': 'filipino', 'gsw': 'Filipino', 'gu': 'ફિલિપિનો', 'ha': 'Dan Filifin', 'he': 'פיליפינית', 'hi': 'फ़िलिपीनो', 'hi-Latn': 'Filipino', 'hr': 'filipinski', 'hsb': 'filipinšćina', 'ht': 'filipino', 'hu': 'filippínó', 'hy': 'ֆիլիպիներեն', 'ia': 'filipino', 'id': 'Filipino', 'ig': 'Filipino', 'is': 'filippseyska', 'it': 'filippino', 'ja': 'フィリピノ語', 'jv': 'Tagalog', 'ka': 'ფილიპინური', 'kea': 'filipinu', 'kgp': 'firipĩnũ', 'kk': 'филиппин тілі', 'kk-Cyrl': 'филиппин тілі', 'km': 'ហ្វ៊ីលីពីន', 'kn': 'ಫಿಲಿಪಿನೊ', 'ko': '필리핀어', 'kok': 'फिलिपिनो', 'kok-Deva': 'फिलिपिनो', 'ks': 'فِلِپیٖنو', 'ks-Arab': 'فِلِپیٖنو', 'ksh': 'Fillipihnesch', 'ku': 'fîlîpînoyî', 'kxv': 'pilipino', 'kxv-Deva': 'प़िलिपीनो', 'kxv-Latn': 'pilipino', 'kxv-Orya': 'ପିଲିପିନ', 'kxv-Telu': 'పిలిపినో', 'ky': 'филипинче', 'lb': 'Filipino', 'lkt': 'Filipino Iyápi', 'lld': 'filipin', 'lo': 'ຟີລິປີໂນ', 'lrc': 'فیلیپینی', 'lt': 'filipiniečių', 'lv': 'filipīniešu', 'mai': 'फिलिपीनो', 'mi': 'Piripīno', 'mk': 'филипински', 'ml': 'ഫിലിപ്പിനോ', 'mn': 'филипино', 'mr': 'फिलिपिनो', 'ms': 'Filipina', 'mt': 'Filippin', 'my': 'ဖိလစ်ပိုင်', 'mzn': 'فیلیپینو', 'nb': 'filipino', 'ne': 'फिलिपिनी', 'nl': 'Filipijns', 'nn': 'filipino', 'no': 'filipino', 'nqo': 'ߝߟߌߔߌ߲ߞߊ߲', 'om': 'Afaan Filippinii', 'or': 'ଫିଲିପିନୋ', 'os': 'филиппинаг', 'pa': 'ਫਿਲੀਪਿਨੋ', 'pa-Guru': 'ਫਿਲੀਪਿਨੋ', 'pcm': 'Filipínó Lángwej', 'pl': 'filipiński', 'ps': 'فلیپیني', 'pt': 'filipino', 'qu': 'Filipino Simi', 'rm': 'filippino', 'ro': 'filipineză', 'ru': 'филиппинский', 'rw': 'Ikinyafilipine', 'sah': 'Пилипииннии', 'sc': 'filipinu', 'sd': 'فلپائني', 'sd-Arab': 'فلپائني', 'se': 'filippiinnagiella', 'si': 'පිලිපීන', 'sk': 'filipínčina', 'sl': 'filipinščina', 'smn': 'filipinokielâ', 'so': 'Filibiino', 'sq': 'filipinisht', 'sr': 'филипински', 'sr-Cyrl': 'филипински', 'sr-Latn': 'filipinski', 'sv': 'filippinska', 'sw': 'Kifilipino', 'ta': 'ஃபிலிபினோ', 'te': 'ఫిలిపినో', 'tg': 'филиппинӣ', 'th': 'ฟิลิปปินส์', 'ti': 'ፊሊፒንኛ', 'tk': 'filippin dili', 'to': 'lea fakafilipaini', 'tr': 'Filipince', 'tt': 'филиппин', 'ug': 'فىلىپپىنچە', 'uk': 'філіппінська', 'ur': 'فلیپینو', 'uz': 'filipincha', 'uz-Cyrl': 'филипинча', 'uz-Latn': 'filipincha', 'vec': 'filipin', 'vi': 'Tiếng Philippines', 'wae': 'Filipiniš', 'wo': 'Filipiye', 'xnr': 'फ़िलिपीनो', 'yi': 'פֿיליפּינא', 'yo': 'Èdè Filipino', 'yrl': 'firipinu', 'yue': '菲律賓文', 'yue-Hans': '菲律宾文', 'yue-Hant': '菲律賓文', 'zh': '菲律宾语', 'zh-Hans': '菲律宾语', 'zh-Hant': '菲律賓文', 'zu': 'isi-Filipino'},
    'fip': {'en': 'Fipa'},
    'fir': {'en': 'Firan'},
    'fit': {'ast': 'finlandés de Tornedalen', 'br': 'finneg traoñienn an Torne', 'cs': 'finština (tornedalská)', 'cy': 'Ffinneg Tornedal', 'de': 'Meänkieli', 'en': 'Tornedalen Finnish', 'et': 'meä', 'fi': 'meänkieli', 'fr': 'finnois tornédalien', 'gd': 'Meänkieli', 'hi-Latn': 'Tornedalen Finnish', 'ht': 'finnois tornédalien', 'hy': 'տորնադելեն ֆիններեն', 'it': 'finlandese del Tornedalen', 'ja': 'トルネダール・フィンランド語', 'lb': 'Meänkieli', 'lt': 'suomių kalbos Tornedalio tarmė', 'mk': 'турнедаленски фински', 'nb': 'tornedalsfinsk', 'nl': 'Tornedal-Fins', 'nn': 'tornedalsfinsk', 'no': 'tornedalsfinsk', 'pl': 'meänkieli', 'sv': 'meänkieli', 'th': 'ฟินแลนด์ทอร์เนดาเล็น', 'to': 'lea fakafinilani-tōnetale', 'tr': 'Tornedalin Fincesi', 'yue': '托爾訥芬蘭文', 'yue-Hans': '托尔讷芬兰文', 'yue-Hant': '托爾訥芬蘭文', 'zh-Hant': '托爾訥芬蘭文'},
    'fiu': {'en': 'Finno-Ugrian languages'},
    'fiu-fin-pro': {'en': 'Proto-Finnic'},
    'fiu-mdv-pro': {'en': 'Proto-Mordvinic'},
    'fiu-prm-pro': {'en': 'Proto-Permic'},
    'fiu-ugr-pro': {'en': 'Proto-Ugric'},
    'fiw': {'en': 'Fiwaga'},
    'fj': {'af': 'Fidjiaans', 'am': 'ፊጂኛ', 'ar': 'الفيجية', 'as': 'ফিজিয়ান', 'ast': 'fixanu', 'az': 'fici', 'az-Arab': 'فیجیجه', 'az-Cyrl': 'фиҹи', 'az-Latn': 'fici', 'be': 'фіджыйская', 'be-tarask': 'фіджыйская', 'bg': 'фиджийски', 'bgn': 'فیجی', 'bn': 'ফিজিয়ান', 'br': 'fidjieg', 'brx': 'फिजियान', 'bs': 'fidžijski', 'bs-Cyrl': 'фиджијски', 'bs-Latn': 'fidžijski', 'ca': 'fijià', 'ca-ES-valencia': 'fijià', 'ccp': '𑄜𑄨𑄎𑄨𑄠𑄚𑄴', 'ce': 'фиджи', 'chr': 'ᏫᏥᎠᏂ', 'ckb': 'فیجی', 'cs': 'fidžijština', 'cy': 'Ffijïeg', 'da': 'fijiansk', 'de': 'Fidschi', 'dsb': 'fidžišćina', 'dz': 'ཕི་ཇི་ཡཱན་ཁ', 'ee': 'fidzigbe', 'el': 'Φίτζι', 'el-polyton': 'Φίτζι', 'en': 'Fijian', 'eo': 'fiĝia', 'es': 'fiyiano', 'et': 'fidži', 'eu': 'fijiera', 'fa': 'فیجیایی', 'ff-Adlm': '𞤊𞤭𞥅𞤶𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'fidži', 'fil': 'Fijian', 'fo': 'fijimál', 'fr': 'fidjien', 'fur': 'fizian', 'fy': 'Fijysk', 'ga': 'Fidsis', 'gd': 'Fìdis', 'gl': 'fixiano', 'gsw': 'Fidschianisch', 'gu': 'ફીજીયન', 'ha': 'Fijiyanci', 'haw': 'Pīkī', 'he': 'פיג׳ית', 'hi': 'फिजियन', 'hi-Latn': 'Fijian', 'hr': 'fidžijski', 'hsb': 'fidźišćina', 'ht': 'fidjien', 'hu': 'fidzsi', 'hy': 'ֆիջիերեն', 'ia': 'fijiano', 'id': 'Fiji', 'ig': 'Fijanị', 'is': 'fídjeyska', 'it': 'figiano', 'ja': 'フィジー語', 'jv': 'Fijian', 'ka': 'ფიჯი', 'kea': 'fijianu', 'kgp': 'fijỹnũ', 'kk': 'фиджи тілі', 'kk-Cyrl': 'фиджи тілі', 'km': 'ហ៊្វីជី', 'kn': 'ಫಿಜಿಯನ್', 'ko': '피지어', 'kok': 'फिजी', 'kok-Deva': 'फिजी', 'ks': 'فِجیَن', 'ks-Arab': 'فِجیَن', 'ksh': 'Fihdschesch', 'ku': 'fîjî', 'ky': 'фижиче', 'lb': 'Fidschianesch', 'lkt': 'Fiji Iyápi', 'lo': 'ຟິຈຽນ', 'lrc': 'فیجی', 'lt': 'fidžių', 'lv': 'fidžiešu', 'mai': 'फिजी', 'mi': 'Whītīana', 'mk': 'фиџиски', 'ml': 'ഫിജിയൻ', 'mn': 'фижи', 'mr': 'फिजियन', 'ms': 'Fiji', 'mt': 'Fiġjan', 'my': 'ဖီဂျီ', 'mzn': 'فیجیایی', 'nb': 'fijiansk', 'ne': 'फिजियन', 'nl': 'Fijisch', 'nn': 'fijiansk', 'no': 'fijiansk', 'or': 'ଫିଜି', 'os': 'фиджи', 'pa': 'ਫ਼ਿਜ਼ੀ', 'pa-Guru': 'ਫ਼ਿਜ਼ੀ', 'pcm': 'Fíján Lángwej', 'pl': 'fidżijski', 'ps': 'فجیان', 'pt': 'fijiano', 'qu': 'Fiyiano Simi', 'rm': 'fidschian', 'ro': 'fijiană', 'ru': 'фиджи', 'sc': 'fijianu', 'sd': 'فجي', 'sd-Arab': 'فجي', 'se': 'fidjigiella', 'si': 'ෆීජි', 'sk': 'fidžijčina', 'sl': 'fidžijščina', 'smn': 'fidžikielâ', 'so': 'Fijian', 'sq': 'fixhianisht', 'sr': 'фиџијски', 'sr-Cyrl': 'фиџијски', 'sr-Latn': 'fidžijski', 'sv': 'fijianska', 'sw': 'Kifiji', 'ta': 'ஃபிஜியன்', 'te': 'ఫిజియన్', 'th': 'ฟิจิ', 'ti': 'ፊጅያንኛ', 'tk': 'fiji dili', 'to': 'lea fakafisi', 'tr': 'Fiji dili', 'ug': 'فىجىچە', 'uk': 'фіджі', 'ur': 'فجی', 'uz': 'fiji', 'uz-Cyrl': 'фижича', 'uz-Latn': 'fiji', 'vec': 'fijan', 'vi': 'Tiếng Fiji', 'wae': 'Fidšianiš', 'xnr': 'फिजियन', 'yi': 'פֿידזשי', 'yo': 'Èdè Fiji', 'yrl': 'fiyianu', 'yue': '斐濟文', 'yue-Hans': '斐济文', 'yue-Hant': '斐濟文', 'zh': '斐济语', 'zh-Hans': '斐济语', 'zh-Hant': '斐濟文', 'zu': 'isi-Fijian'},
    'fkk': {'en': 'Kirya-Konzəl'},
    'fkv': {'en': 'Kven Finnish'},
    'fla': {'en': "Kalispel-Pend d'Oreille"},
    'flh': {'en': 'Foau'},
    'fli': {'en': 'Fali'},
    'fll': {'en': 'North Fali'},
    'fln': {'en': 'Flinders Island'},
    'flr': {'en': 'Fuliiru'},
    'fly': {'en': 'Flaaitaal'},
    'fmp': {'en': "Fe'fe'"},
    'fmu': {'en': 'Far Western Muria'},
    'fnb': {'en': 'Fanbak'},
    'fng': {'en': 'Fanagalo'},
    'fni': {'en': 'Fania'},
    'fo': {'af': 'Faroëes', 'ak': 'Farosi', 'am': 'ፋሮኛ', 'ar': 'الفاروية', 'as': 'ফাৰোইজ', 'ast': 'feroés', 'az': 'farer', 'az-Arab': 'فاروئه\u200cجه', 'az-Cyrl': 'фарер', 'az-Latn': 'farer', 'bal-Latn': 'Paróese', 'be': 'фарэрская', 'be-tarask': 'фарэрская', 'bg': 'фарьорски', 'bgn': 'فاروئی', 'bn': 'ফেরোইস', 'br': 'faeroeg', 'brx': 'फेर’वीज', 'bs': 'farski', 'bs-Cyrl': 'фарски', 'bs-Latn': 'farski', 'ca': 'feroès', 'ca-ES-valencia': 'feroès', 'ccp': '𑄜𑄢𑄮𑄌𑄴', 'ce': 'фарерийн', 'chr': 'ᏇᎶᎡᏍ', 'ckb': 'فەرۆیی', 'cs': 'faerština', 'cy': 'Ffaröeg', 'da': 'færøsk', 'de': 'Färöisch', 'dsb': 'ferejšćina', 'dz': 'ཕཱ་རོ་ཨིས་ཁ', 'el': 'Φεροϊκά', 'el-polyton': 'Φεροϊκά', 'en': 'Faroese', 'eo': 'feroa', 'es': 'feroés', 'et': 'fääri', 'eu': 'faroera', 'fa': 'فارویی', 'ff-Adlm': '𞤊𞤫𞤪𞤮𞤱𞤫𞤧𞤪𞤫', 'fi': 'fääri', 'fil': 'Faroese', 'fo': 'føroyskt', 'fr': 'féroïen', 'fur': 'faroês', 'fy': 'Faeröersk', 'ga': 'Faróis', 'gd': 'Fàrothais', 'gl': 'feroés', 'gsw': 'Färöisch', 'gu': 'ફોરિસ્ત', 'ha': 'Faroese', 'he': 'פארואזית', 'hi': 'फ़ैरोइज़', 'hi-Latn': 'Faroese', 'hr': 'ferojski', 'hsb': 'färöšćina', 'ht': 'féroïen', 'hu': 'feröeri', 'hy': 'ֆարյորերեն', 'ia': 'feroese', 'id': 'Faroe', 'ig': 'Faroese', 'is': 'færeyska', 'it': 'faroese', 'ja': 'フェロー語', 'jv': 'Faroe', 'ka': 'ფარერული', 'kea': 'faroes', 'kgp': 'fervej', 'kk': 'фарер тілі', 'kk-Cyrl': 'фарер тілі', 'km': 'ហ្វារូស', 'kn': 'ಫರೋಸಿ', 'ko': '페로어', 'kok': 'फेरोस', 'kok-Deva': 'फेरोस', 'ks': 'فَروس', 'ks-Arab': 'فَروس', 'ksh': 'Färröhresch', 'ku': 'ferî', 'ky': 'фарерче', 'lb': 'Färöesch', 'lkt': 'Faroese Iyápi', 'lld': 'faroesc', 'lo': 'ຟາໂຣສ', 'lrc': 'فاروٙسی', 'lt': 'farerų', 'lv': 'fēru', 'mai': 'फेरोइज', 'mi': 'Wharoīhi', 'mk': 'фарски', 'ml': 'ഫാറോസ്', 'mn': 'фарер', 'mr': 'फरोइज', 'ms': 'Faroe', 'mt': 'Faroese', 'my': 'ဖာရို', 'mzn': 'فارویی', 'nb': 'færøysk', 'ne': 'फारोज', 'nl': 'Faeröers', 'nn': 'færøysk', 'no': 'færøysk', 'nqo': 'ߝߋߙߏߞߊ', 'om': 'Afaan Faroese', 'or': 'ଫାରୋଇଜ୍\u200c', 'os': 'фарераг', 'pa': 'ਫ਼ੇਰੋਸੇ', 'pa-Guru': 'ਫ਼ੇਰੋਸੇ', 'pcm': 'Fáróís Lángwej', 'pl': 'farerski', 'ps': 'فاروئې', 'pt': 'feroês', 'qu': 'Feroes Simi', 'rm': 'feroais', 'ro': 'feroeză', 'ru': 'фарерский', 'rw': 'Inyefaroyizi', 'sc': 'faroesu', 'sd': 'فيروايس', 'sd-Arab': 'فيروايس', 'se': 'fearagiella', 'si': 'ෆාරෝස්', 'sk': 'faerčina', 'sl': 'ferščina', 'smn': 'fäärikielâ', 'so': 'Farowsi', 'sq': 'faroisht', 'sr': 'фарски', 'sr-Cyrl': 'фарски', 'sr-Latn': 'farski', 'sv': 'färöiska', 'sw': 'Kifaroe', 'ta': 'ஃபரோயிஸ்', 'te': 'ఫారోస్', 'tg': 'фарерӣ', 'th': 'แฟโร', 'ti': 'ፋሮእይና', 'tk': 'farer dili', 'to': 'lea fakafaloe', 'tr': 'Faroe dili', 'tt': 'фарер', 'ug': 'فائېروچە', 'uk': 'фарерська', 'ur': 'فیروئیز', 'uz': 'farercha', 'uz-Cyrl': 'фарерча', 'uz-Latn': 'farercha', 'vec': 'faroeze', 'vi': 'Tiếng Faroe', 'wo': 'Feroos', 'xnr': 'फ़ैरोइज़', 'yi': 'פֿאַראיש', 'yo': 'Èdè Faroesi', 'yrl': 'faruwesi', 'yue': '法羅文', 'yue-Hans': '法罗文', 'yue-Hant': '法羅文', 'zh': '法罗语', 'zh-Hans': '法罗语', 'zh-Hant': '法羅文', 'zu': 'isi-Faroese'},
    'fod': {'en': 'Foodo'},
    'foi': {'en': 'Foi'},
    'fom': {'en': 'Foma'},
    'fon': {'am': 'ፎን', 'ar': 'الفون', 'as': 'ফ’ন', 'az-Cyrl': 'фон', 'be': 'фон', 'be-tarask': 'фон', 'bg': 'фон', 'bgn': 'فون', 'bn': 'ফন', 'brx': 'फ’न', 'bs-Cyrl': 'фон', 'ccp': '𑄜𑄧𑄚𑄴', 'ce': 'фон', 'chr': 'ᏠᏂ', 'ckb': 'فۆنی', 'cs': 'fonština', 'dsb': 'fonšćina', 'el': 'Φον', 'el-polyton': 'Φον', 'et': 'foni', 'eu': 'fonera', 'fa': 'فونی', 'ff-Adlm': '𞤊𞤮𞤲𞤪𞤫', 'ga': 'Fonais', 'gu': 'ફોન', 'he': 'פון', 'hi': 'फॉन', 'hsb': 'fonšćina', 'hy': 'ֆոն', 'is': 'fón', 'ja': 'フォン語', 'ka': 'ფონი', 'kgp': 'fãm', 'kk': 'фон тілі', 'kk-Cyrl': 'фон тілі', 'km': 'ហ្វ៊ុន', 'kn': 'ಫೋನ್', 'ko': '폰어', 'kok': 'फोन', 'kok-Deva': 'फोन', 'ks': 'فون', 'ks-Arab': 'فون', 'ksh': 'Fohnesch', 'ku': 'fonî', 'ky': 'фончо', 'lb': 'Fon-Sprooch', 'lo': 'ຟອນ', 'lv': 'fonu', 'mai': 'फोन', 'mi': 'Whāna', 'mk': 'фон', 'ml': 'ഫോൻ', 'mn': 'фон', 'mr': 'फॉन', 'my': 'ဖော်န်', 'ne': 'फोन', 'or': 'ଫନ୍', 'pa': 'ਫੌਨ', 'pa-Guru': 'ਫੌਨ', 'pcm': 'Fọn Lángwej', 'ps': 'فان', 'pt': 'fom', 'qu': 'Fon Simi', 'ru': 'фон', 'sd': 'فون', 'sd-Arab': 'فون', 'si': 'ෆොන්', 'sk': 'fončina', 'sl': 'fonščina', 'smn': 'fonkielâ', 'sq': 'fonisht', 'sr': 'фон', 'sr-Cyrl': 'фон', 'sv': 'fonspråket', 'sw': 'Kifon', 'syr': 'ܦܘܢܝܬ', 'ta': 'ஃபான்', 'te': 'ఫాన్', 'th': 'ฟอน', 'ti': 'ፎን', 'tk': 'fon dili', 'to': 'lea fakafōngi', 'ug': 'فونچە', 'uk': 'фон', 'ur': 'فون', 'uz-Cyrl': 'фон', 'vi': 'Tiếng Fon', 'xnr': 'फॉन', 'yo': 'Èdè Fon', 'yrl': 'fũmu', 'yue': '豐文', 'yue-Hans': '丰文', 'yue-Hant': '豐文', 'zh': '丰语', 'zh-Hans': '丰语', 'zh-Hant': '豐文', 'zu': 'isi-Fon', 'en': 'Fon'},
    'for': {'en': 'Fore'},
    'fos': {'en': 'Siraya'},
    'fox': {'en': 'Formosan languages'},
    'fpe': {'en': 'Fernando Po Creole English'},
    'fqs': {'en': 'Fas'},
    'fr': {'af': 'Frans', 'agq': 'Kɨ̀fàlàŋsi', 'ak': 'Frɛnkye', 'am': 'ፈረንሳይኛ', 'ar': 'الفرنسية', 'as': 'ফ্ৰেন্স', 'asa': 'Kifarantha', 'ast': 'francés', 'az': 'fransız', 'az-Arab': 'فرانساجا', 'az-Cyrl': 'франсыз', 'az-Latn': 'fransız', 'bal-Latn': 'Paránsi', 'bas': 'Hɔp u pulàsi', 'be': 'французская', 'be-tarask': 'французская', 'bem': 'Ichi Frenchi', 'bez': 'Hifaransa', 'bg': 'френски', 'bgn': 'فرانسوی', 'blo': 'gɩfɔnɔ', 'bm': 'tubabukan', 'bn': 'ফরাসি', 'br': 'galleg', 'brx': 'फ्रेन्च', 'bs': 'francuski', 'bs-Cyrl': 'француски', 'bs-Latn': 'francuski', 'bss': 'Frɛnsé', 'ca': 'francès', 'ca-ES-valencia': 'francès', 'ccp': '𑄜𑄧𑄢𑄥𑄨', 'ce': 'французийн', 'ceb': 'Pranses', 'cgg': 'Orufaransa', 'chr': 'ᎦᎸᏥ', 'ckb': 'فەرەنسی', 'cs': 'francouzština', 'csw': 'ᐹᕽᐘᔦᓰᒧᐏᐣ', 'cv': 'франци', 'cy': 'Ffrangeg', 'da': 'fransk', 'dav': 'Kifaransa', 'de': 'Französisch', 'dje': 'Fransee senni', 'doi': 'फ्रेंच', 'dsb': 'francojšćina', 'dyo': 'franse', 'dz': 'ཕྲནཅ་ཁ', 'ebu': 'Kĩfaransa', 'ee': 'Fransegbe', 'el': 'Γαλλικά', 'el-polyton': 'Γαλλικά', 'en': 'French', 'en-Dsrt': '𐐙𐑉𐐯𐑌𐐽', 'eo': 'franca', 'es': 'francés', 'et': 'prantsuse', 'eu': 'frantsesa', 'ewo': 'Ǹkɔ́bɔ fulɛnsí', 'fa': 'فرانسوی', 'ff': 'Farayseere', 'ff-Adlm': '𞤊𞤢𞤪𞤢𞤲𞤧𞤭𞥅𞤪𞤫', 'ff-Latn': 'Farayseere', 'fi': 'ranska', 'fil': 'French', 'fo': 'franskt', 'fr': 'français', 'fur': 'francês', 'fy': 'Frânsk', 'ga': 'Fraincis', 'gd': 'Fraingis', 'gl': 'francés', 'gsw': 'Französisch', 'gu': 'ફ્રેન્ચ', 'guz': 'Kifaransa', 'ha': 'Faransanci', 'haw': 'Palani', 'he': 'צרפתית', 'hi': 'फ़्रेंच', 'hi-Latn': 'French', 'hnj': '𞄕𞄤𞄰𞄎𞄦𞄴', 'hnj-Hmnp': '𞄕𞄤𞄰𞄎𞄦𞄴', 'hr': 'francuski', 'hsb': 'francošćina', 'ht': 'français', 'hu': 'francia', 'hy': 'ֆրանսերեն', 'ia': 'francese', 'id': 'Prancis', 'ie': 'francesi', 'ig': 'French', 'ii': 'ꃔꇩꉙ', 'is': 'franska', 'it': 'francese', 'ja': 'フランス語', 'jgo': 'Fɛlánci', 'jmc': 'Kyifaransa', 'jv': 'Prancis', 'ka': 'ფრანგული', 'kaa': 'французша', 'kaa-Cyrl': 'французша', 'kab': 'Tafransist', 'kam': 'Kifaransa', 'kde': 'Chifalansa', 'kea': 'franses', 'kgp': 'frỹsej', 'khq': 'Fransee senni', 'ki': 'Kĩbaranja', 'kk': 'француз тілі', 'kk-Cyrl': 'француз тілі', 'kkj': 'numbu buy', 'kln': 'kutitab Kifaransa', 'km': 'បារាំង', 'kn': 'ಫ್ರೆಂಚ್', 'ko': '프랑스어', 'kok': 'फ्रेंच', 'kok-Deva': 'फ्रेंच', 'kok-Latn': 'Fransez', 'ks': 'فرانسیسی', 'ks-Arab': 'فرانسیسی', 'ks-Deva': 'फ्रांसीसी', 'ksb': 'Kifalansa', 'ksf': 'ripɛrɛsǝ́', 'ksh': 'Franzühsesch', 'ku': 'fransizî', 'kxv': 'prenc', 'kxv-Deva': 'प़्रेंच', 'kxv-Latn': 'prenc', 'kxv-Orya': 'ପ୍ରେଞ୍ଚ\u200d', 'kxv-Telu': 'ప్రెంచ్', 'ky': 'французча', 'lag': 'Kɨfaráansa', 'lb': 'Franséisch', 'lg': 'Lufalansa', 'lkt': 'Wašíču Ikčéka Iyápi', 'lld': 'franzesc', 'ln': 'lifalansɛ́', 'lo': 'ຝຣັ່ງ', 'lrc': 'فآرانسئ ئی', 'lt': 'prancūzų', 'lu': 'Mfwàlànsa', 'luo': 'Kifaransa', 'luy': 'Kifaransa', 'lv': 'franču', 'mai': 'फ़्रेंच', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ faransa', 'mer': 'Kĩfuransi', 'mfe': 'franse', 'mg': 'Frantsay', 'mgh': 'Ifaransa', 'mi': 'Wīwī', 'mk': 'француски', 'ml': 'ഫ്രഞ്ച്', 'mn': 'франц', 'mni': 'ফ্রেঞ্চ', 'mni-Beng': 'ফ্রেঞ্চ', 'mr': 'फ्रेंच', 'ms': 'Perancis', 'mt': 'Franċiż', 'mua': 'zah sǝr Franssǝ', 'my': 'ပြင်သစ်', 'mzn': 'فرانسوی', 'naq': 'Frans', 'nb': 'fransk', 'nd': 'isi-Fulentshi', 'ne': 'फ्रान्सेली', 'nl': 'Frans', 'nmg': 'Fala', 'nn': 'fransk', 'nnh': 'felaŋsée', 'no': 'fransk', 'nqo': 'ߝߊ߬ߙߊ߲߬ߛߌ߬ߞߊ߲', 'nus': 'Thok pɔrɔthani', 'nyn': 'Orufaransa', 'om': 'Afaan Faransaayii', 'or': 'ଫରାସୀ', 'os': 'францаг', 'pa': 'ਫਰਾਂਸੀਸੀ', 'pa-Guru': 'ਫਰਾਂਸੀਸੀ', 'pcm': 'Frẹ́nch Lángwej', 'pl': 'francuski', 'ps': 'فرانسوي', 'pt': 'francês', 'qu': 'Frances Simi', 'rm': 'franzos', 'rn': 'Igifaransa', 'ro': 'franceză', 'rof': 'Kyifaransa', 'ru': 'французский', 'rw': 'Igifaransa', 'rwk': 'Kyifaransa', 'sa': 'फ़्रांसदेशीय भाषा:', 'sah': 'Боронсуустуу', 'saq': 'Kifaransa', 'sat': 'ᱯᱷᱨᱮᱧᱪ', 'sat-Olck': 'ᱯᱷᱨᱮᱧᱪ', 'sbp': 'Ishifalansa', 'sc': 'frantzesu', 'sd': 'فرانسيسي', 'sd-Arab': 'فرانسيسي', 'sd-Deva': 'फ्रेंच', 'se': 'fránskkagiella', 'seh': 'francês', 'ses': 'Fransee senni', 'sg': 'Farânzi', 'shi': 'ⵜⴰⴼⵔⴰⵏⵙⵉⵙⵜ', 'shi-Latn': 'Tafransist', 'shi-Tfng': 'ⵜⴰⴼⵔⴰⵏⵙⵉⵙⵜ', 'si': 'ප්\u200dරංශ', 'sk': 'francúzština', 'sl': 'francoščina', 'smn': 'ranskakielâ', 'sn': 'chiFurenchi', 'so': 'Faransiis', 'sq': 'frëngjisht', 'sr': 'француски', 'sr-Cyrl': 'француски', 'sr-Latn': 'francuski', 'su': 'Prancis', 'su-Latn': 'Prancis', 'sv': 'franska', 'sw': 'Kifaransa', 'syr': 'ܦܪܢܣܝܬ', 'ta': 'பிரெஞ்சு', 'te': 'ఫ్రెంచ్', 'teo': 'Kifaransa', 'tg': 'франсузӣ', 'th': 'ฝรั่งเศส', 'ti': 'ፈረንሳይኛ', 'tk': 'fransuz dili', 'to': 'lea fakafalanisē', 'tpi': 'Frens', 'tr': 'Fransızca', 'tt': 'француз', 'twq': 'Fransee senni', 'tzm': 'Tafṛansist', 'ug': 'فىرانسۇزچە', 'uk': 'французька', 'ur': 'فرانسیسی', 'uz': 'fransuzcha', 'uz-Cyrl': 'французча', 'uz-Latn': 'fransuzcha', 'vai': 'ꗱꘋꔻ', 'vai-Latn': 'Fɛŋsi', 'vai-Vaii': 'ꗱꘋꔻ', 'vec': 'franseze', 'vi': 'Tiếng Pháp', 'vun': 'Kyifaransa', 'wae': 'Wälš', 'wo': 'Farañse', 'xh': 'IsiFrentshi', 'xnr': 'फ़्रांसीसी', 'xog': 'Olufalansa', 'yav': 'feleŋsí', 'yi': 'פֿראַנצויזיש', 'yo': 'Èdè Faransé', 'yrl': 'frãsei', 'yue': '法文', 'yue-Hans': '法文', 'yue-Hant': '法文', 'zgh': 'ⵜⴰⴼⵔⴰⵏⵙⵉⵙⵜ', 'zh': '法语', 'zh-Hans': '法语', 'zh-Hant': '法文', 'zu': 'isi-French'},
    'fr-CA': {'ak': 'Kanada Frɛnkye', 'am': 'የካናዳ ፈረንሳይኛ', 'as': 'কানাডিয়ান ফ্ৰেন্স', 'ast': 'francés de Canadá', 'az': 'Kanada fransızcası', 'az-Cyrl': 'Канада франсызҹасы', 'az-Latn': 'Kanada fransızcası', 'bal-Latn': 'Paránsi (Kaynadhá)', 'be': 'канадская французская', 'be-tarask': 'канадская французская', 'bgn': 'کاناڈاین فرانسوی', 'br': 'galleg Kanada', 'brx': 'कानाडानि फ्रेन्च', 'ca': 'francès canadenc', 'ca-ES-valencia': 'francès canadenc', 'ccp': '𑄇𑄚𑄓𑄩𑄠𑄧 𑄜𑄧𑄢𑄥𑄨', 'ce': 'канадан французийн', 'chr': 'ᎨᎾᏓ ᎦᎸᏥ', 'csw': 'ᐹᕽᐘᔦᓰᒧᐏᐣ (ᑳᓇᑕ)', 'cv': 'канадӑн франци', 'cy': 'Ffrangeg Canada', 'da': 'canadisk fransk', 'doi': 'कैनेडियन फ्रेंच', 'dsb': 'kanadiska francojšćina', 'dz': 'ཀེ་ན་ཌི་ཡཱན་ཕྲནཅ་ཁ', 'ee': 'Fransegbe (Canada)', 'el': 'Γαλλικά Καναδά', 'el-polyton': 'Γαλλικά Καναδά', 'en': 'Canadian French', 'es': 'francés canadiense', 'et': 'Kanada prantsuse', 'eu': 'Kanadako frantsesa', 'fa': 'فرانسوی کانادا', 'ff-Adlm': '𞤊𞤢𞤪𞤢𞤲𞤧𞤭𞥅𞤪𞤫 𞤑𞤢𞤲𞤢𞤣𞤢𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'kanadanranska', 'fr': 'français canadien', 'fur': 'francês dal Canade', 'fy': 'Kanadeesk Frânsk', 'ga': 'Fraincis Cheanada', 'gd': 'Fraingis Chanada', 'gl': 'francés canadense', 'gsw': 'Kanadischs Französisch', 'ha': 'Farasanci Kanada', 'haw': 'Palani Kanakā', 'hi': 'कनाडाई फ़्रेंच', 'hi-Latn': 'Canadian French', 'hr': 'kanadski francuski', 'hsb': 'kanadiska francošćina', 'ht': 'français canadien', 'hu': 'kanadai francia', 'hy': 'կանադական ֆրանսերեն', 'ia': 'francese canadian', 'ig': 'Canadian French', 'is': 'kanadísk franska', 'it': 'francese canadese', 'jv': 'Prancis Kanada', 'ka': 'კანადური ფრანგული', 'kaa': 'французша (Канада)', 'kaa-Cyrl': 'французша (Канада)', 'kea': 'franses kanadianu', 'kn': 'ಕೆನೆಡಿಯನ್ ಫ್ರೆಂಚ್', 'kok': 'कॅनडियन फ्रेंच', 'kok-Deva': 'कॅनडियन फ्रेंच', 'ks': 'کَنیڈیَن فرانسیسی', 'ks-Arab': 'کَنیڈیَن فرانسیسی', 'ks-Deva': 'कनाडियन फ्रांसीसी', 'ksh': 'Franzühsesch uß Kanada', 'ku': 'fransizî (Kanada)', 'kxv': 'kanaḍati prenc', 'kxv-Deva': 'कनाडाति प़्रेंच', 'kxv-Latn': 'kanaḍati prenc', 'kxv-Orya': 'କାନାଡାତି ପ୍ରେଞ୍ଚ\u200d', 'kxv-Telu': 'కానడియెన్ ప్రేంచ్', 'lb': 'Kanadescht Franséisch', 'lrc': 'فآرانسئ ئی کانادا', 'lt': 'Kanados prancūzų', 'mai': 'कनाडाई फ़्रेंच', 'mi': 'Wīwī Kānata', 'mk': 'канадски француски', 'ml': 'കനേഡിയൻ ഫ്രഞ്ച്', 'mni': 'কানাদিয়ান ফ্রেঞ্চ', 'mni-Beng': 'কানাদিয়ান ফ্রেঞ্চ', 'mr': 'कॅनडियन फ्रेंच', 'ms': 'Perancis Kanada', 'mt': 'Franċiż Kanadiż', 'my': 'ကနေဒါ ပြင်သစ်', 'mzn': 'کانادای ِفرانسوی', 'ne': 'क्यानेडाली फ्रान्सेली', 'nqo': 'ߞߣߊߘߊ߫ ߝߊ߬ߙߊ߲߬ߛߌ߬ߞߊ߲', 'or': 'କାନାଡିୟ ଫ୍ରେଞ୍ଚ', 'os': 'канадӕйаг францаг', 'pa': 'ਫਰਾਂਸੀਸੀ (ਕੈਨੇਡੀਅਨ)', 'pa-Guru': 'ਫਰਾਂਸੀਸੀ (ਕੈਨੇਡੀਅਨ)', 'pcm': 'Kánádá Frẹnch', 'pl': 'francuski kanadyjski', 'ps': 'کاناډايي فرانسوي', 'rm': 'franzos canadais', 'ru': 'канадский французский', 'sa': 'कनाडादेशः फ़्रांसदेशीय भाषा:', 'sat': 'ᱠᱟᱱᱟᱰᱤᱭᱟᱱ ᱯᱷᱨᱮᱧᱪ', 'sat-Olck': 'ᱠᱟᱱᱟᱰᱤᱭᱟᱱ ᱯᱷᱨᱮᱧᱪ', 'sc': 'frantzesu canadesu', 'sd': 'ڪينيڊيائي فرانسيسي', 'sd-Arab': 'ڪينيڊيائي فرانسيسي', 'sd-Deva': 'कैनेडियन फ्रेंच', 'si': 'කැනේඩියානු ප්\u200dරංශ', 'sk': 'francúzština (kanadská)', 'sl': 'kanadska francoščina', 'smn': 'Kanada ranskakielâ', 'so': 'Faransiiska Kanada', 'sq': 'frëngjishte kanadeze', 'su': 'Prancis Kanada', 'su-Latn': 'Prancis Kanada', 'sv': 'kanadensisk franska', 'sw': 'Kifaransa (Canada)', 'ta': 'கனடிய பிரெஞ்சு', 'te': 'కెనడియెన్ ఫ్రెంచ్', 'tg': 'франсузии канадагӣ', 'th': 'ฝรั่งเศส - แคนาดา', 'to': 'lea fakafalanisē-kānata', 'tpi': 'Kenedien Frens', 'tr': 'Kanada Fransızcası', 'ug': 'كانادا فىرانسۇزچە', 'ur': 'کینیڈین فرانسیسی', 'uz': 'fransuz (Kanada)', 'uz-Latn': 'fransuz (Kanada)', 'vec': 'franseze canadeze', 'wae': 'Kanadišes Wälš', 'wo': 'Frañse bu Kanadaa', 'xh': 'IsiFrentshi SaseKhanada', 'xnr': 'कनेडे आली फ़्रांसीसी', 'yo': 'Èdè Faransé (orílẹ̀-èdè Kánádà)', 'zh': '加拿大法语', 'zh-Hans': '加拿大法语', 'zu': 'isi-Canadian French'},
    'fr-CH': {'ak': 'Swisalande Frɛnkye', 'am': 'የስዊዝ ፈረንሳይኛ', 'as': 'ছুইচ ফ্ৰেন্স', 'ast': 'francés de Suiza', 'az': 'İsveçrə fransızcası', 'az-Cyrl': 'Исвечрә франсызҹасы', 'az-Latn': 'İsveçrə fransızcası', 'be': 'швейцарская французская', 'be-tarask': 'швейцарская французская', 'bgn': 'اشاره\u200cاین فرانسوی', 'br': 'galleg Suis', 'brx': 'सुइस फ्रेन्च', 'ca': 'francès suís', 'ca-ES-valencia': 'francès suís', 'ccp': '𑄥𑄪𑄃𑄨𑄌𑄴 𑄜𑄧𑄢𑄥𑄨', 'ce': 'швейцарин французийн', 'chr': 'ᏍᏫᏏ ᎦᎸᏥ', 'csw': 'ᐹᕽᐘᔦᓰᒧᐏᐣ (CH)', 'cv': 'швейцарӗн франци', 'cy': 'Ffrangeg y Swistir', 'da': 'schweizisk fransk', 'doi': 'स्विस फ्रेंच', 'dsb': 'šwicarska francojšćina', 'dz': 'སུ་ཡིས་ཕྲནཅ་ཁ', 'ee': 'Fransegbe (Switzerland)', 'el': 'Γαλλικά Ελβετίας', 'el-polyton': 'Γαλλικά Ελβετίας', 'en': 'Swiss French', 'es': 'francés suizo', 'et': 'Šveitsi prantsuse', 'eu': 'Suitzako frantsesa', 'fa': 'فرانسوی سوئیس', 'ff-Adlm': '𞤊𞤢𞤪𞤢𞤲𞤧𞤭𞥅𞤪𞤫 𞤅𞤵𞤱𞤭𞥅𞤧', 'fi': 'sveitsinranska', 'fil': 'Swiss na French', 'fr': 'français suisse', 'fur': 'francês de Svuizare', 'fy': 'Switserse Frânsk', 'ga': 'Fraincis na hEilvéise', 'gd': 'Fraingis Eilbheiseach', 'gl': 'francés suízo', 'gsw': 'Schwiizer Französisch', 'gu': 'સ્વિસ ફ્રેંચ', 'ha': 'Farasanci Suwizalan', 'haw': 'Kuikilani', 'hi': 'स्विस फ़्रेंच', 'hi-Latn': 'Swiss French', 'hr': 'švicarski francuski', 'hsb': 'šwicarska francošćina', 'ht': 'français suisse', 'hu': 'svájci francia', 'hy': 'շվեյցարական ֆրանսերեն', 'ia': 'francese suisse', 'ie': 'sviss francesi', 'ig': 'Swiss French', 'is': 'svissnesk franska', 'it': 'francese svizzero', 'jv': 'Prancis Switserlan', 'ka': 'შვეიცარიული ფრანგული', 'kaa': 'французша (Швейцария)', 'kaa-Cyrl': 'французша (Швейцария)', 'kea': 'franses suisu', 'kn': 'ಸ್ವಿಸ್ ಫ್ರೆಂಚ್', 'kok': 'स्विझ फ्रेंच', 'kok-Deva': 'स्विझ फ्रेंच', 'ks': 'سٕوٕس فرانسیسی', 'ks-Arab': 'سٕوٕس فرانسیسی', 'ks-Deva': 'स्विस फ्रांसीसी', 'ksh': 'Franzühsesch uß de Schweijz', 'ku': 'fransizî (Swîsre)', 'kxv': 'svis prenc', 'kxv-Deva': 'स्विस प़्रेंच', 'kxv-Latn': 'svis prenc', 'kxv-Orya': 'ସ୍ୱିସ୍ ପ୍ରେଞ୍ଚ\u200d', 'kxv-Telu': 'స్విస్ ప్రెంచ్', 'lb': 'Schwäizer Franséisch', 'lo': 'ຝຣັ່ງ (ສວິສ)', 'lrc': 'فآرانسئ ئی سوٙییس', 'lt': 'Šveicarijos prancūzų', 'mai': 'स्विस फ़्रेंच', 'mi': 'Wīwī Huiterangi', 'mk': 'швајцарски француски', 'ml': 'സ്വിസ് ഫ്രഞ്ച്', 'mn': 'франц (Швейцар)', 'mni': 'স্বিজ ফ্রেঞ্চ', 'mni-Beng': 'স্বিজ ফ্রেঞ্চ', 'mr': 'स्विस फ्रेंच', 'ms': 'Perancis Switzerland', 'mt': 'Franċiż Żvizzeru', 'my': 'ဆွစ် ပြင်သစ်', 'mzn': 'سوییس ِفرانسوی', 'ne': 'स्विस फ्रेन्च', 'nqo': 'ߛߥߌߛ ߝߙߊ߬ߛߌ߬ߞߊ߲', 'or': 'ସ୍ୱିସ୍ ଫ୍ରେଞ୍ଚ', 'os': 'швейцариаг францаг', 'pcm': 'Swízalánd Frẹnch', 'pl': 'francuski szwajcarski', 'ps': 'سويسي فرانسوي', 'rm': 'franzos svizzer', 'ru': 'швейцарский французский', 'sa': 'स्विस फ़्रांसदेशीय भाषा:', 'sat': 'ᱥᱩᱣᱤᱥ ᱯᱷᱨᱮᱧᱪ', 'sat-Olck': 'ᱥᱩᱣᱤᱥ ᱯᱷᱨᱮᱧᱪ', 'sc': 'frantzesu isvìtzeru', 'sd': 'سوئس فرانسيسي', 'sd-Arab': 'سوئس فرانسيسي', 'sd-Deva': 'स्विस फ्रेंच', 'si': 'ස්විස් ප්\u200dරංශ', 'sk': 'francúzština (švajčiarska)', 'sl': 'švicarska francoščina', 'smn': 'Sveitsi ranskakielâ', 'so': 'Faransiis (Iswiiserlaand)', 'sq': 'frëngjishte zvicerane', 'su': 'Prancis Swiss', 'su-Latn': 'Prancis Swiss', 'sv': 'schweizisk franska', 'ta': 'ஸ்விஸ் பிரஞ்சு', 'te': 'స్విస్ ఫ్రెంచ్', 'tg': 'франсузии швейсарӣ', 'th': 'ฝรั่งเศส - สวิส', 'to': 'lea fakafalanisē-suisilani', 'tpi': 'Swis Frens', 'tr': 'İsviçre Fransızcası', 'ug': 'شىۋىتسارىيە فىرانسۇزچە', 'ur': 'سوئس فرینچ', 'uz': 'fransuz (Shveytsariya)', 'uz-Latn': 'fransuz (Shveytsariya)', 'vec': 'franseze zvìsaro', 'wae': 'Schwizer Wälš', 'wo': 'Frañse bu Swis', 'xh': 'IsiFrentshi SaseSwitzerland', 'xnr': 'स्विस फ़्रांसीसी', 'yo': 'Èdè Faranṣé (Súwísàlaǹdì)', 'zh': '瑞士法语', 'zh-Hans': '瑞士法语', 'zu': 'isi-Swiss French'},
    'frc': {'af': 'Cajun', 'am': 'ካጁን ፍሬንች', 'ar': 'الفرنسية الكاجونية', 'as': 'কেজ’ন ফৰাছী', 'ast': 'francés cajun', 'az': 'Kacun fransızcası', 'az-Latn': 'Kacun fransızcası', 'bal-Latn': 'Kájon Pránsi', 'be': 'каджунская французская', 'be-tarask': 'каджунская французская', 'bg': 'каджунски френски', 'bn': 'কাজুন ফরাসি', 'br': 'galleg cajun', 'brx': 'काजुन फ्रेन्च', 'bs': 'kajunski francuski', 'bs-Cyrl': 'кајунски француски', 'bs-Latn': 'kajunski francuski', 'ca': 'francès cajun', 'ca-ES-valencia': 'francès cajun', 'chr': 'ᎨᏨᏂ ᎦᎸᏥ', 'cs': 'francouzština (cajunská)', 'cy': 'Ffrangeg Cajwn', 'da': 'cajunfransk', 'de': 'Cajun', 'dsb': 'cajun francojšćina', 'el': 'Γαλλικά (Λουιζιάνα)', 'el-polyton': 'Γαλλικά (Λουιζιάνα)', 'en': 'Cajun French', 'es': 'francés cajún', 'et': 'cajun’i', 'eu': 'cajun frantsesa', 'fa': 'فرانسوی کاجون', 'ff-Adlm': '𞤊𞤢𞤪𞤢𞤲𞤧𞤭𞥅𞤪𞤫 𞤑𞤢𞤣𞤭𞤴𞤫𞤲𞤪𞤫', 'fi': 'cajunranska', 'fil': 'Cajun French', 'fr': 'français cadien', 'ga': 'Fraincis Cajun', 'gd': 'Fraingis nan Cajun', 'gl': 'francés cajun', 'gu': 'કાજૂન ફ્રેન્ચ', 'ha': 'Faransancin Cajun', 'he': 'צרפתית קייג׳ונית', 'hi': 'केजन फ़्रेंच', 'hi-Latn': 'Cajun French', 'hr': 'kajunski francuski', 'hsb': 'cajun francošćina', 'ht': 'français cadien', 'hu': 'cajun francia', 'hy': 'քաջունական ֆրանսերեն', 'ia': 'francese cajun', 'id': 'Prancis Cajun', 'ig': 'Cajun French', 'is': 'cajun-franska', 'it': 'francese cajun', 'ja': 'ケイジャン・フランス語', 'jv': 'Prancis Cajun', 'ka': 'კაჟუნური ფრანგული', 'kgp': 'frỹsej kajũn', 'kk': 'каджун тілі (француз)', 'kk-Cyrl': 'каджун тілі (француз)', 'km': 'បារាំងកាហ្សង់', 'kn': 'ಕಾಜುನ್ ಫ್ರೆಂಚ್', 'ko': '케이준 프랑스어', 'kok': 'कॅजां फ्रेंच', 'kok-Deva': 'कॅजां फ्रेंच', 'ku': 'fransizîya kajûnê', 'ky': 'французча (кажун диалектиси)', 'lb': 'Cajun', 'lo': 'ຝຣັ່ງເຄຈຸນ', 'lt': 'kadžunų prancūzų', 'lv': 'kadžūnu franču', 'mai': 'केजन फ्रांसीसी', 'mi': 'Wīwī Keihana', 'mk': 'каџунски француски', 'ml': 'കേജൺ ഫ്രഞ്ച്', 'mn': 'франц, кажун', 'mr': 'केजॉन फ्रेंच', 'ms': 'Perancis Cajun', 'my': 'ကေဂျန် ဖရန့်စ်', 'nb': 'cajunfransk', 'ne': 'काहुन फ्रान्सेली', 'nl': 'Cajun-Frans', 'nn': 'cajunfransk', 'no': 'cajunfransk', 'nqo': 'ߞߊߘߌ߫ ߝߊ߬ߙߊ߲߬ߛߌ߬ߞߊ߲', 'or': 'କାଜୁନ୍ ଫରାସୀ', 'pa': 'ਕੇਜੁਨ ਫ਼੍ਰੈਂਚ', 'pa-Guru': 'ਕੇਜੁਨ ਫ਼੍ਰੈਂਚ', 'pcm': 'Kájun Frẹnchi', 'pl': 'cajuński', 'ps': 'کاجون فرانسوی', 'pt': 'francês cajun', 'qu': 'Francés Cajun', 'ro': 'franceză cajun', 'ru': 'каджунский французский', 'sc': 'frantzesu cajun', 'sd': 'ڪيجن فرانسيسي', 'sd-Arab': 'ڪيجن فرانسيسي', 'si': 'කජුන් ප්\u200dරංශ', 'sk': 'francúzština (cajunská)', 'sl': 'cajunska francoščina', 'so': 'Faransiiska Cajun', 'sq': 'frëngjishte kajune', 'sr': 'кајунски француски', 'sr-Cyrl': 'кајунски француски', 'sr-Latn': 'kajunski francuski', 'sv': 'cajun-franska', 'sw': 'Kifaransa cha Kajuni', 'ta': 'கஜுன் பிரெஞ்சு', 'te': 'కాజున్ ఫ్రెంచ్', 'th': 'ฝรั่งเศสเคจัน', 'ti': 'ካጁን ፈረንሳይ', 'tk': 'fransuz diliniň kajun şiwesi', 'to': 'lea fakafalanisē-kasuni', 'tr': 'Cajun Fransızcası', 'uk': 'кажунська французька', 'ur': 'کاجن فرانسیسی', 'uz': 'kajun fransuz', 'uz-Latn': 'kajun fransuz', 'vec': 'cadian', 'vi': 'Tiếng Pháp Cajun', 'xnr': 'केजन फ़्रांसीसी', 'yo': 'Èdè Faranse ti Kajun', 'yrl': 'frãsei kayũ', 'yue': '卡真法文', 'yue-Hans': '卡真法文', 'yue-Hant': '卡真法文', 'zh': '卡真法语', 'zh-Hans': '卡真法语', 'zh-Hant': '卡真法文', 'zu': 'isi-Cajun French'},
    'frd': {'en': 'Fordata'},
    'frk': {'en': 'Frankish'},
    'frm': {'ar': 'الفرنسية الوسطى', 'ast': 'francés mediu', 'az': 'orta fransız', 'az-Latn': 'orta fransız', 'bg': 'средновековен френски', 'bn': 'মধ্য ফরাসি', 'br': 'krenncʼhalleg', 'brx': 'मध्य फ्रांसीसी', 'bs': 'srednjovjekovni francuski', 'bs-Cyrl': 'средњи француски', 'bs-Latn': 'srednjovjekovni francuski', 'ca': 'francès mitjà', 'ca-ES-valencia': 'francès mitjà', 'ccp': '𑄟𑄧𑄖𑄴𑄙𑄳𑄠𑄧 𑄜𑄧𑄢𑄥𑄨', 'cs': 'francouzština (středověká)', 'cy': 'Ffrangeg Canol', 'da': 'middelfransk', 'de': 'Mittelfranzösisch', 'el': 'Μέσα Γαλλικά', 'el-polyton': 'Μέσα Γαλλικά', 'en': 'Middle French', 'en-Dsrt': '𐐣𐐮𐐼𐐲𐑊 𐐙𐑉𐐯𐑌𐐽', 'es': 'francés medio', 'et': 'keskprantsuse', 'fa': 'فرانسوی میانه', 'fi': 'keskiranska', 'fr': 'moyen français', 'fy': 'Middelfrânsk', 'ga': 'Meán-Fhraincis', 'gd': 'Meadhan-Fhraingis', 'gsw': 'Mittelfranzösisch', 'gu': 'મિડિલ ફ્રેંચ', 'he': 'צרפתית תיכונה', 'hi': 'मध्यकालीन फ़्रांसीसी', 'hi-Latn': 'Middle French', 'hr': 'srednjofrancuski', 'ht': 'moyen français', 'hu': 'közép francia', 'id': 'Prancis Abad Pertengahan', 'is': 'miðfranska', 'it': 'francese medio', 'ja': '中期フランス語', 'ka': 'საშუალო ფრანგული', 'kgp': 'frỹsej kuju', 'kn': 'ಮಧ್ಯ ಫ್ರೆಂಚ್', 'ko': '중세 프랑스어', 'ks': 'وسطی فرینچ', 'ks-Arab': 'وسطی فرینچ', 'lb': 'Mëttelfranséisch', 'lo': 'ຟຮັ່ງເສດກາງ', 'lt': 'Vidurio Prancūzijos', 'lv': 'vidusfranču', 'mk': 'среднофранцуски', 'ml': 'മദ്ധ്യ ഫ്രഞ്ച്', 'mr': 'मिडल फ्रेंच', 'mt': 'Franċiż Medjevali', 'my': 'အလယ်ပိုင်း ပြင်သစ်', 'nb': 'mellomfransk', 'ne': 'मध्य फ्रान्सेली', 'nl': 'Middelfrans', 'nn': 'mellomfransk', 'no': 'mellomfransk', 'or': 'ମଧ୍ୟ ଫ୍ରେଞ୍ଚ', 'pl': 'średniofrancuski', 'pt': 'francês médio', 'rm': 'franzos mesaun', 'ro': 'franceză medie', 'ru': 'среднефранцузский', 'sk': 'stredná francúzština', 'sl': 'srednja francoščina', 'sr': 'средњефранцуски', 'sr-Cyrl': 'средњефранцуски', 'sr-Latn': 'srednjefrancuski', 'sv': 'medelfranska', 'ta': 'மிடில் பிரெஞ்சு', 'te': 'మధ్యమ ప్రెంచ్', 'th': 'ฝรั่งเศสกลาง', 'to': 'lea fakafalanisē-lotoloto', 'tr': 'Ortaçağ Fransızcası', 'ug': 'ئوتتۇرا ئەسىر فىرانسۇزچە', 'uk': 'середньофранцузька', 'vi': 'Tiếng Pháp Trung cổ', 'yrl': 'frãsei médiu', 'yue': '中古法文', 'yue-Hans': '中古法文', 'yue-Hant': '中古法文', 'zh': '中古法语', 'zh-Hans': '中古法语', 'zh-Hant': '中古法文'},
    'fro': {'ar': 'الفرنسية القديمة', 'ast': 'francés antiguu', 'az': 'qədim fransız', 'az-Latn': 'qədim fransız', 'be': 'старафранцузская', 'be-tarask': 'старафранцузская', 'bg': 'старофренски', 'bn': 'প্রাচীন ফরাসি', 'br': 'hencʼhalleg', 'brx': 'पुरानी फ्रांसीसी', 'bs': 'starofrancuski', 'bs-Cyrl': 'старофранцуски', 'bs-Latn': 'starofrancuski', 'ca': 'francès antic', 'ca-ES-valencia': 'francès antic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄨 𑄜𑄧𑄢𑄥𑄨', 'cs': 'francouzština (stará)', 'cy': 'Hen Ffrangeg', 'da': 'oldfransk', 'de': 'Altfranzösisch', 'el': 'Παλαιά Γαλλικά', 'el-polyton': 'Παλαιά Γαλλικά', 'en': 'Old French', 'es': 'francés antiguo', 'et': 'vanaprantsuse', 'fa': 'فرانسوی باستان', 'fi': 'muinaisranska', 'fr': 'ancien français', 'fur': 'vieri francês', 'fy': 'Aldfrânsk', 'ga': 'Sean-Fhraincis', 'gd': 'Seann-Fhraingis', 'gsw': 'Altfranzösisch', 'gu': 'જૂની ફ્રેંચ', 'he': 'צרפתית עתיקה', 'hi': 'पुरातन फ़्रांसीसी', 'hi-Latn': 'Old French', 'hr': 'starofrancuski', 'ht': 'ancien français', 'hu': 'ófrancia', 'hy': 'հին ֆրանսերեն', 'id': 'Prancis Kuno', 'is': 'fornfranska', 'it': 'francese antico', 'ja': '古フランス語', 'ka': 'ძველი ფრანგული', 'kgp': 'frỹsej arkajku', 'kn': 'ಪ್ರಾಚೀನ ಫ್ರೆಂಚ್', 'ko': '고대 프랑스어', 'ks': 'پرون فرینچ', 'ks-Arab': 'پرون فرینچ', 'lb': 'Alfranséisch', 'lo': 'ຟຮັ່ງເສດໂບຮານ', 'lt': 'senoji prancūzų', 'lv': 'senfranču', 'mk': 'старофранцуски', 'ml': 'പഴയ ഫ്രഞ്ച്', 'mr': 'पुरातन फ्रेंच', 'mt': 'Franċiż Antik', 'my': 'ဖရန်စီစ်', 'nb': 'gammelfransk', 'ne': 'पुरातन फ्रान्सेली', 'nl': 'Oudfrans', 'nn': 'gammalfransk', 'no': 'gammelfransk', 'or': 'ପୁରୁଣା ଫ୍ରେଞ୍ଚ', 'os': 'рагон францаг', 'pl': 'starofrancuski', 'pt': 'francês arcaico', 'rm': 'franzos vegl', 'ro': 'franceză veche', 'ru': 'старофранцузский', 'sk': 'stará francúzština', 'sl': 'stara francoščina', 'sr': 'старофранцуски', 'sr-Cyrl': 'старофранцуски', 'sr-Latn': 'starofrancuski', 'sv': 'fornfranska', 'sw': 'Kifaransa cha Kale', 'ta': 'பழைய பிரெஞ்சு', 'te': 'ప్రాచీన ఫ్రెంచ్', 'th': 'ฝรั่งเศสโบราณ', 'to': 'lea fakafalanisē-motuʻa', 'tr': 'Eski Fransızca', 'ug': 'قەدىمكى فىرانسۇزچە', 'uk': 'давньофранцузька', 'vi': 'Tiếng Pháp cổ', 'yi': 'אַלט־פֿראַנצויזיש', 'yrl': 'frãsei arkaiku', 'yue': '古法文', 'yue-Hans': '古法文', 'yue-Hant': '古法文', 'zh': '古法语', 'zh-Hans': '古法语', 'zh-Hant': '古法文'},
    'frp': {'am': 'አርፒታን', 'ast': 'arpitanu', 'br': 'arpitaneg', 'cs': 'franko-provensálština', 'cy': 'Arpitaneg', 'de': 'Frankoprovenzalisch', 'en': 'Arpitan', 'et': 'frankoprovansi', 'fi': 'arpitaani', 'fr': 'francoprovençal', 'gd': 'Arpitan', 'hi-Latn': 'Arpitan', 'ht': 'francoprovençal', 'id': 'Arpitan', 'it': 'francoprovenzale', 'ja': 'アルピタン語', 'lb': 'Frankoprovenzalesch', 'lt': 'arpitano', 'mk': 'франкопровансалски', 'nb': 'arpitansk', 'ne': 'अर्पितान', 'nl': 'Arpitaans', 'nn': 'arpitansk', 'no': 'arpitansk', 'pl': 'franko-prowansalski', 'sv': 'frankoprovensalska', 'th': 'อาร์พิตา', 'to': 'lea fakaʻāpitano', 'tr': 'Arpitanca', 'uk': 'арпітанська', 'vi': 'Tiếng Arpitan', 'yue': '法蘭克-普羅旺斯文', 'yue-Hans': '法兰克-普罗旺斯文', 'yue-Hant': '法蘭克-普羅旺斯文', 'zh-Hant': '法蘭克-普羅旺斯文'},
    'frq': {'en': 'Forak'},
    'frr': {'af': 'Noord-Fries', 'am': 'ሰሜናዊ ፍሪስያን', 'ar': 'الفريزينية الشمالية', 'as': 'নৰ্দাৰ্ন ফ্ৰিছিয়ান', 'ast': 'frisón del norte', 'az': 'şimali fris', 'az-Latn': 'şimali fris', 'bal-Latn': 'Shemáli Prési', 'be': 'паўночнафрызская', 'be-tarask': 'паўночнафрызская', 'bg': 'северен фризийски', 'bn': 'উত্তরাঞ্চলীয় ফ্রিসিয়ান', 'br': 'frizeg an Norzh', 'brx': 'उत्तरी फ्रीज़ियन्', 'bs': 'sjeverni frizijski', 'bs-Cyrl': 'северно-фризијски', 'bs-Latn': 'sjeverni frizijski', 'ca': 'frisó septentrional', 'ca-ES-valencia': 'frisó septentrional', 'ccp': '𑄅𑄖𑄴𑄗𑄧𑄢𑄴 𑄎𑄬𑄌𑄴𑄎𑄮𑄢𑄴 𑄜𑄳𑄢𑄨𑄥𑄨𑄠𑄚𑄴', 'chr': 'ᏧᏴᏢ ᎷᏈ', 'cs': 'fríština (severní)', 'cy': 'Ffriseg Gogleddol', 'da': 'nordfrisisk', 'de': 'Nordfriesisch', 'dsb': 'pódpołnocna frizišćina', 'el': 'Βόρεια Φριζιανά', 'el-polyton': 'Βόρεια Φριζιανά', 'en': 'Northern Frisian', 'es': 'frisón septentrional', 'et': 'põhjafriisi', 'eu': 'iparraldeko frisiera', 'fa': 'فریزی شمالی', 'ff-Adlm': '𞤊𞤭𞤪𞤧𞤭𞤴𞤢𞤲𞤪𞤫 𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤫', 'fi': 'pohjoisfriisi', 'fil': 'Hilagang Frisian', 'fr': 'frison septentrional', 'frr': 'Nordfriisk', 'fy': 'Noard-Frysk', 'ga': 'Freaslainnis an Tuaiscirt', 'gd': 'Frìoslannais Thuathach', 'gl': 'frisón setentrional', 'gsw': 'Nordfriesisch', 'gu': 'ઉત્તરી ફ્રિશિયન', 'ha': 'Firisiyanci na Arewaci', 'he': 'פריזית צפונית', 'hi': 'उत्तरी फ़्रीसियाई', 'hi-Latn': 'Northern Frisian', 'hr': 'sjevernofrizijski', 'hsb': 'sewjernofrizišćina', 'ht': 'frison septentrional', 'hu': 'északi fríz', 'hy': 'հյուսիսային ֆրիզերեն', 'ia': 'frison septentrional', 'id': 'Frisia Utara', 'ig': 'Northern Frisian', 'is': 'norðurfrísneska', 'it': 'frisone settentrionale', 'ja': '北フリジア語', 'jv': 'Frisian Lor Segara', 'ka': 'ჩრდილოფრიზიული', 'kgp': 'friso tỹ nãrti', 'kk': 'солтүстік фриз тілі', 'kk-Cyrl': 'солтүстік фриз тілі', 'km': 'ហ្វ្រ៊ីសៀន\u200bខាងជើង', 'kn': 'ಉತ್ತರ ಫ್ರಿಸಿಯನ್', 'ko': '북부 프리지아어', 'kok': 'उत्तरीय फ्रिशियन', 'kok-Deva': 'उत्तरीय फ्रिशियन', 'ks': 'شُمٲلی فرِشیَن', 'ks-Arab': 'شُمٲلی فرِشیَن', 'ku': 'frîsîya bakur', 'ky': 'түндүк фризче (евразиялык тилдер)', 'lb': 'Nordfriesesch', 'lo': 'ຟຣີຊຽນເໜືອ', 'lt': 'šiaurinių fryzų', 'lv': 'ziemeļfrīzu', 'mi': 'Whirīhiana ki te Raki', 'mk': 'севернофризиски', 'ml': 'നോർത്തേൻ ഫ്രിഷ്യൻ', 'mn': 'хойд фриз', 'mr': 'उत्तरी फ्रिशियन', 'ms': 'Frisian Utara', 'my': 'မြောက် ဖရီစီရန်', 'nb': 'nordfrisisk', 'ne': 'उत्तरी फ्रिजी', 'nl': 'Noord-Fries', 'nn': 'nordfrisisk', 'no': 'nordfrisisk', 'or': 'ଉତ୍ତର ଫ୍ରିସିୟାନ୍', 'pa': 'ਉੱਤਰੀ ਫ੍ਰੀਜ਼ੀਅਨ', 'pa-Guru': 'ਉੱਤਰੀ ਫ੍ਰੀਜ਼ੀਅਨ', 'pcm': 'Nọ́tán Frísian', 'pl': 'północnofryzyjski', 'ps': 'شمالي فریسیان', 'pt': 'frísio setentrional', 'qu': 'Frisón del Norte Simi', 'rm': 'fris dal nord', 'ro': 'frizonă nordică', 'ru': 'северный фризский', 'sc': 'frisone setentrionale', 'sd': 'اترين فريسين', 'sd-Arab': 'اترين فريسين', 'si': 'උතුරු ෆ්\u200dරිසියානු', 'sk': 'severná frízština', 'sl': 'severna frizijščina', 'so': 'Northern Frisian', 'sq': 'frisianishte veriore', 'sr': 'севернофризијски', 'sr-Cyrl': 'севернофризијски', 'sr-Latn': 'severnofrizijski', 'sv': 'nordfrisiska', 'sw': 'Kifrisia cha Kaskazini', 'ta': 'வடக்கு ஃப்ரிஸியான்', 'te': 'ఉత్తర ఫ్రిసియన్', 'th': 'ฟริเซียนเหนือ', 'ti': 'ሰሜናዊ ፍሪስኛ', 'tk': 'demirgazyk friz dili', 'to': 'lea fakafilisia-tokelau', 'tr': 'Kuzey Frizce', 'ug': 'شىمالى فىرىزيەچە', 'uk': 'фризька північна', 'ur': 'شمالی فریزئین', 'uz': 'shimoliy friz', 'uz-Latn': 'shimoliy friz', 'vec': 'frìzone setentrionale', 'vi': 'Tiếng Frisia Miền Bắc', 'xnr': 'उत्तरी फ़्रीसियाई', 'yi': 'דרום־פֿריזיש', 'yo': 'Èdè Ariwa Frisa', 'yrl': 'firísiu setẽtiriunau', 'yue': '北弗里西亞文', 'yue-Hans': '北弗里西亚文', 'yue-Hant': '北弗里西亞文', 'zh': '北弗里西亚语', 'zh-Hans': '北弗里西亚语', 'zh-Hant': '北弗里西亞文', 'zu': 'isi-Northern Frisian'},
    'frs': {'ar': 'الفريزينية الشرقية', 'ast': 'frisón oriental', 'bg': 'източнофризийски', 'bn': 'পূর্ব ফ্রিসিয়', 'br': 'frizeg ar Reter', 'brx': 'पूर्वी फ्रीज़ियन्', 'bs': 'istočnofrizijski', 'bs-Cyrl': 'источни фризијски', 'bs-Latn': 'istočnofrizijski', 'ca': 'frisó oriental', 'ca-ES-valencia': 'frisó oriental', 'ccp': '𑄛𑄪𑄉𑄮 𑄜𑄳𑄢𑄨𑄥𑄨𑄠𑄧', 'cs': 'fríština (východní)', 'cy': 'Ffriseg y Dwyrain', 'da': 'østfrisisk', 'de': 'Ostfriesisch', 'el': 'Ανατολικά Φριζιανά', 'el-polyton': 'Ανατολικά Φριζιανά', 'en': 'Eastern Frisian', 'es': 'frisón oriental', 'et': 'idafriisi', 'fa': 'فریزی شرقی', 'fi': 'itäfriisi', 'fr': 'frison oriental', 'fy': 'East-Frysk', 'gd': 'Frìoslannais Earach', 'gsw': 'Oschtfriesisch', 'gu': 'પૂર્વ ફ્રિશિયન', 'he': 'פריזית מזרחית', 'hi': 'पूर्वी फ़्रीसियाई', 'hi-Latn': 'Eastern Frisian', 'hr': 'istočnofrizijski', 'ht': 'frison oriental', 'hu': 'keleti fríz', 'hy': 'արևելաֆրիզերեն', 'id': 'Frisia Timur', 'is': 'austurfrísneska', 'it': 'frisone orientale', 'ja': '東フリジア語', 'ka': 'აღმოსავლეთფრიზიული', 'kgp': 'frisỹ rãjur', 'kn': 'ಪೂರ್ವ ಫ್ರಿಸಿಯನ್', 'ko': '동부 프리슬란드어', 'ks': 'مشرِقی فرِشیَن', 'ks-Arab': 'مشرِقی فرِشیَن', 'lb': 'Ostfriesesch', 'lo': 'ຟຣີຊຽນຕາເວັນອອກ', 'lt': 'rytų fryzų', 'lv': 'austrumfrīzu', 'mk': 'источнофризиски', 'ml': 'ഈസ്റ്റേൺ ഫ്രിഷ്യൻ', 'mr': 'पौर्वात्य फ्रिशियन', 'my': 'အရှေ့ ဖရီစီရန်', 'nb': 'østfrisisk', 'ne': 'पूर्वी फ्रिसियाली', 'nl': 'Oost-Fries', 'nn': 'austfrisisk', 'no': 'østfrisisk', 'or': 'ପୂର୍ବ ଫ୍ରିସିୟାନ୍', 'pl': 'wschodniofryzyjski', 'pt': 'frisão oriental', 'rm': 'fris da l’ost', 'ro': 'frizonă orientală', 'ru': 'восточный фризский', 'sk': 'východofrízština', 'sl': 'vzhodna frizijščina', 'sr': 'источнофризијски', 'sr-Cyrl': 'источнофризијски', 'sr-Latn': 'istočnofrizijski', 'sv': 'östfrisiska', 'sw': 'Kifrisia cha Mashariki', 'ta': 'கிழக்கு ஃப்ரிஸியான்', 'te': 'తూర్పు ఫ్రిసియన్', 'th': 'ฟริเซียนตะวันออก', 'to': 'lea fakafilisia-hahake', 'tr': 'Doğu Frizcesi', 'ug': 'شەرقى فىرىزيەچە', 'uk': 'фризька східна', 'vi': 'Tiếng Frisian Miền Đông', 'yi': 'מזרח־פֿריזיש', 'yrl': 'firísiu usidẽtawara', 'yue': '東弗里西亞文', 'yue-Hans': '东弗里西亚文', 'yue-Hant': '東弗里西亞文', 'zh': '东弗里西亚语', 'zh-Hans': '东弗里西亚语', 'zh-Hant': '東弗里西亞文'},
    'frt': {'en': 'Fortsenal'},
    'fse': {'en': 'Finnish Sign Language'},
    'fsl': {'en': 'French Sign Language'},
    'fss': {'en': 'Finland-Swedish Sign Language'},
    'fub': {'en': 'Adamawa Fulfulde'},
    'fuc': {'en': 'Pulaar'},
    'fud': {'en': 'East Futuna'},
    'fue': {'en': 'Borgu Fulfulde'},
    'fuf': {'en': 'Pular'},
    'fuh': {'en': 'Western Niger Fulfulde'},
    'fui': {'en': 'Bagirmi Fulfulde'},
    'fuj': {'en': 'Ko'},
    'fum': {'en': 'Fum'},
    'fun': {'en': 'Fulniô'},
    'fuq': {'en': 'Central-Eastern Niger Fulfulde'},
    'fur': {'af': 'Friuliaans', 'am': 'ፍሩሊያን', 'ar': 'الفريلايان', 'as': 'ফ্ৰিউলিয়ান', 'ast': 'friulianu', 'az': 'friul', 'az-Cyrl': 'фриул', 'az-Latn': 'friul', 'bal-Latn': 'Priuli', 'be': 'фрыульская', 'be-tarask': 'фрыульская', 'bg': 'фриулски', 'bgn': 'فریولی', 'bn': 'ফ্রিউলিও', 'br': 'frioulaneg', 'brx': 'फ्रिउलियान', 'bs': 'friulijski', 'bs-Cyrl': 'фриулијски', 'bs-Latn': 'friulijski', 'ca': 'friülà', 'ca-ES-valencia': 'friülà', 'ccp': '𑄜𑄳𑄢𑄨𑄃𑄪𑄣𑄨𑄠𑄚𑄴', 'ce': 'фриулийн', 'chr': 'ᏞᎤᎵᎠᏂ', 'ckb': 'فریئوولی', 'cs': 'furlanština', 'cy': 'Ffriwleg', 'da': 'friulisk', 'de': 'Friaulisch', 'dsb': 'friulšćina', 'el': 'Φριουλανικά', 'el-polyton': 'Φριουλανικά', 'en': 'Friulian', 'es': 'friulano', 'et': 'friuuli', 'eu': 'friulera', 'fa': 'فریولیایی', 'ff-Adlm': '𞤊𞤭𞤪𞥇𞤵𞤤𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'friuli', 'fil': 'Friulian', 'fo': 'friuliskt', 'fr': 'frioulan', 'fur': 'furlan', 'fy': 'Friulysk', 'ga': 'Friúilis', 'gd': 'Friùilis', 'gl': 'friulano', 'gsw': 'Friulisch', 'gu': 'ફ્રિયુલિયાન', 'ha': 'Friulian', 'he': 'פריולית', 'hi': 'फ्रीयुलीयान', 'hi-Latn': 'Friulian', 'hr': 'furlanski', 'hsb': 'friulšćina', 'ht': 'frioulan', 'hu': 'friuli', 'hy': 'ֆրիուլիերեն', 'ia': 'friulano', 'id': 'Friuli', 'ig': 'Friulian', 'is': 'fríúlska', 'it': 'friulano', 'ja': 'フリウリ語', 'jv': 'Friulian', 'ka': 'ფრიულური', 'kea': 'friulanu', 'kgp': 'frijurỹnũ', 'kk': 'фриуль тілі', 'kk-Cyrl': 'фриуль тілі', 'km': 'ហ៊្វ្រូលាន', 'kn': 'ಫ್ರಿಯುಲಿಯನ್', 'ko': '프리울리어', 'kok': 'फ्रिलियन', 'kok-Deva': 'फ्रिलियन', 'ks': 'فروٗلِیَن', 'ks-Arab': 'فروٗلِیَن', 'ksh': 'Frijauhlesch', 'ku': 'frîyolî', 'ky': 'фриулча', 'lb': 'Friulesch', 'lld': 'furlan', 'lo': 'ຟຣີລຽນ', 'lt': 'friulių', 'lv': 'friūlu', 'mai': 'फ्रियूयन', 'mi': 'Whiriūriana', 'mk': 'фурлански', 'ml': 'ഫ്രിയുലിയാൻ', 'mn': 'фриулан', 'mr': 'फ्रियुलियान', 'ms': 'Friulian', 'mt': 'Frijuljan', 'my': 'ဖရူလီယန်း', 'nb': 'friuliansk', 'ne': 'फ्रिउलियाली', 'nl': 'Friulisch', 'nn': 'friulisk', 'no': 'friuliansk', 'nqo': 'ߝߙߌߥߎߟߊߞߊ߲', 'or': 'ଫ୍ରିୟୁଲୀୟାନ୍', 'pa': 'ਫਰੀਉਲੀਅਨ', 'pa-Guru': 'ਫਰੀਉਲੀਅਨ', 'pcm': 'Friúlián Lángwej', 'pl': 'friulski', 'ps': 'فرائیلیین', 'pt': 'friulano', 'qu': 'Friulano Simi', 'rm': 'friulan', 'ro': 'friulană', 'ru': 'фриульский', 'sc': 'friulanu', 'sd': 'فرائي لئين', 'sd-Arab': 'فرائي لئين', 'si': 'ෆ්\u200dරියුලියන්', 'sk': 'friulčina', 'sl': 'furlanščina', 'smn': 'friulikielâ', 'so': 'Firiyuuliyaan', 'sq': 'friulianisht', 'sr': 'фриулски', 'sr-Cyrl': 'фриулски', 'sr-Latn': 'friulski', 'sv': 'friulianska', 'sw': 'Kifriulian', 'ta': 'ஃப்ரியூலியன்', 'te': 'ఫ్రియులియన్', 'th': 'ฟรูลี', 'ti': 'ፍርዩልኛ', 'tk': 'friul dili', 'to': 'lea fakafulilāni', 'tr': 'Friuli dili', 'ug': 'فىرىئۇلىچە', 'uk': 'фріульська', 'ur': 'فریولیائی', 'uz': 'friul', 'uz-Cyrl': 'фриулча', 'uz-Latn': 'friul', 'vec': 'furlan', 'vi': 'Tiếng Friulian', 'xnr': 'फ्रीयुलीयान', 'yo': 'Firiúlíànì', 'yrl': 'friuranu', 'yue': '弗留利文', 'yue-Hans': '弗留利文', 'yue-Hant': '弗留利文', 'zh': '弗留利语', 'zh-Hans': '弗留利语', 'zh-Hant': '弗留利文', 'zu': 'isi-Friulian'},
    'fut': {'en': 'Futuna-Aniwa'},
    'fuu': {'en': 'Furu'},
    'fuv': {'en': 'Nigerian Fulfulde'},
    'fuy': {'en': 'Fuyug'},
    'fvr': {'en': 'Fur'},
    'fwa': {'en': 'Fwâi'},
    'fwe': {'en': 'Fwe'},
    'fy': {'af': 'Fries', 'ak': 'Atɔeɛ Fam Frihyia Kasa', 'am': 'ምዕራባዊ ፍሪሲኛ', 'ar': 'الفريزيان', 'as': 'ৱেষ্টাৰ্ণ ফ্ৰিছিয়ান', 'ast': 'frisón occidental', 'az': 'qərbi friz', 'az-Cyrl': 'гәрби фриз', 'az-Latn': 'qərbi friz', 'bal-Latn': 'Ferisi (Róbarkati)', 'be': 'заходняя фрызская', 'be-tarask': 'заходняя фрызская', 'bg': 'западнофризийски', 'bgn': 'روچ\u200cکپتین فریزی', 'bn': 'পশ্চিম ফ্রিসিয়ান', 'br': 'frizeg ar Cʼhornôg', 'brx': 'सोनाबारि फ्रिजियान', 'bs': 'zapadni frizijski', 'bs-Cyrl': 'западни фризијски', 'bs-Latn': 'zapadni frizijski', 'ca': 'frisó occidental', 'ca-ES-valencia': 'frisó occidental', 'ccp': '𑄛𑄧𑄎𑄨𑄟𑄴 𑄜𑄳𑄢𑄨𑄥𑄨𑄠𑄚𑄴', 'ce': 'малхбузен-фризийн', 'chr': 'ᏭᏕᎵᎬ ᏗᏜ ᏟᏏᎠᏂ', 'ckb': 'فریسیی ڕۆژاوا', 'cs': 'fríština (západní)', 'cy': 'Ffriseg y Gorllewin', 'da': 'vestfrisisk', 'de': 'Westfriesisch', 'dsb': 'frizišćina', 'dz': 'ནུབ་ཕྼི་སི་ཡན་ཁ', 'el': 'Δυτικά Φριζικά', 'el-polyton': 'Δυτικά Φριζικά', 'en': 'Western Frisian', 'eo': 'okcident-frisa', 'es': 'frisón occidental', 'et': 'läänefriisi', 'eu': 'mendebaldeko frisiera', 'fa': 'فریسی غربی', 'ff-Adlm': '𞤊𞤭𞤪𞤭𞥅𞤧𞤭𞤴𞤢𞤲𞤳𞤮𞥅𞤪𞤫 𞤖𞤭𞤪𞤲𞤢', 'fi': 'länsifriisi', 'fil': 'Western Frisian', 'fo': 'vestur frísiskt', 'fr': 'frison occidental', 'fur': 'frisian', 'fy': 'Frysk', 'ga': 'Freaslainnis Iartharach', 'gd': 'Frìoslannais Shiarach', 'gl': 'frisón occidental', 'gsw': 'Friesisch', 'gu': 'પશ્ચિમિ ફ્રિશિયન', 'ha': 'Frisian ta Yamma', 'he': 'פריזית מערבית', 'hi': 'पश्चिमी फ़्रिसियाई', 'hi-Latn': 'Western Frisian', 'hr': 'zapadnofrizijski', 'hsb': 'frizišćina', 'ht': 'frison occidental', 'hu': 'nyugati fríz', 'hy': 'արևմտաֆրիզերեն', 'ia': 'frison occidental', 'id': 'Frisia Barat', 'ig': 'Ọdịda anyanwụ Frisian', 'is': 'vesturfrísneska', 'it': 'frisone occidentale', 'ja': '西フリジア語', 'jv': 'Frisia Sisih Kulon', 'ka': 'დასავლეთფრიზიული', 'kea': 'fríziu osidental', 'kgp': 'friso rãpur', 'kk': 'батыс фриз тілі', 'kk-Cyrl': 'батыс фриз тілі', 'km': 'ហ្វ្រីស៊ានខាងលិច', 'kn': 'ಪಶ್ಚಿಮ ಫ್ರಿಸಿಯನ್', 'ko': '서부 프리지아어', 'kok': 'पश्चिमी फ्रिशियन', 'kok-Deva': 'पश्चिमी फ्रिशियन', 'ks': 'مغربی فرِشیَن', 'ks-Arab': 'مغربی فرِشیَن', 'ksh': 'Wäßfrihsesch', 'ku': 'frîsî', 'ky': 'батыш фризче', 'lb': 'Westfriesesch', 'lld': 'frison dl vest', 'lo': 'ຟຣິຊຽນ ຕາເວັນຕົກ', 'lrc': 'فئریسی أفتونئشین', 'lt': 'vakarų fryzų', 'lv': 'rietumfrīzu', 'mai': 'पश्चिमी फ्रीजन', 'mi': 'Whirīhiana ki te Uru', 'mk': 'западнофризиски', 'ml': 'പശ്ചിമ ഫ്രിഷിയൻ', 'mn': 'баруун фриз', 'mr': 'पश्चिमी फ्रिशियन', 'ms': 'Frisian Barat', 'mt': 'Frisian tal-Punent', 'my': 'အနောက် ဖရီစီရန်', 'mzn': 'غربی فیریزی', 'nb': 'vestfrisisk', 'ne': 'पश्चिमी फ्रिसियन', 'nl': 'Fries', 'nn': 'vestfrisisk', 'no': 'vestfrisisk', 'nqo': 'ߕߟߋ߬ߓߋ ߝߙߌߛߐ߲ߞߊ߲', 'om': 'Afaan Firisiyaani', 'or': 'ପାଶ୍ଚାତ୍ୟ ଫ୍ରିସିଆନ୍', 'pa': 'ਪੱਛਮੀ ਫ੍ਰਿਸੀਅਨ', 'pa-Guru': 'ਪੱਛਮੀ ਫ੍ਰਿਸੀਅਨ', 'pcm': 'Wẹ́stán Frísiá Lángwej', 'pl': 'zachodniofryzyjski', 'ps': 'لوېديځ فريشي', 'pt': 'frísio ocidental', 'qu': 'Frison Simi', 'rm': 'fris', 'ro': 'frizonă occidentală', 'ru': 'западнофризский', 'rw': 'Igifiriziyani', 'sc': 'frisone otzidentale', 'sd': 'مغربي فريشن', 'sd-Arab': 'مغربي فريشن', 'se': 'oarjifriisagiella', 'si': 'බටහිර ෆ්\u200dරිසියානු', 'sk': 'západná frízština', 'sl': 'zahodna frizijščina', 'smn': 'viestârfriisi', 'so': 'Firiisiyan Galbeed', 'sq': 'frizianishte perëndimore', 'sr': 'западни фризијски', 'sr-Cyrl': 'западни фризијски', 'sr-Latn': 'zapadni frizijski', 'sv': 'västfrisiska', 'sw': 'Kifrisia cha Magharibi', 'ta': 'மேற்கு ஃப்ரிஷியன்', 'te': 'పశ్చిమ ఫ్రిసియన్', 'tg': 'фризии ғарбӣ', 'th': 'ฟริเซียนตะวันตก', 'ti': 'ምዕራባዊ ፍሪስኛ', 'tk': 'günbatar friz dili', 'to': 'lea fakafilisia-hihifo', 'tr': 'Batı Frizcesi', 'ug': 'غەربىي فىرسچە', 'uk': 'західнофризька', 'ur': 'مغربی فریسیئن', 'uz': 'g‘arbiy friz', 'uz-Cyrl': 'ғарбий фризча', 'uz-Latn': 'g‘arbiy friz', 'vec': 'frìzone osidentale', 'vi': 'Tiếng Frisia', 'xnr': 'पश्चिमी फ़्रिसियाई', 'yi': 'מערב־פֿריזיש', 'yo': 'Èdè Frisia', 'yrl': 'frísiu usidẽtawara', 'yue': '西弗里西亞文', 'yue-Hans': '西弗里西亚文', 'yue-Hant': '西弗里西亞文', 'zh': '西弗里西亚语', 'zh-Hans': '西弗里西亚语', 'zh-Hant': '西弗里西亞文', 'zu': 'isi-Western Frisian'},
    'ga': {'af': 'Iers', 'ak': 'Aerelande kasa', 'am': 'አየርላንድኛ', 'ar': 'الأيرلندية', 'as': 'আইৰিচ', 'ast': 'irlandés', 'az': 'irland', 'az-Cyrl': 'ирланд', 'az-Latn': 'irland', 'bal-Latn': 'Áeri', 'be': 'ірландская', 'be-tarask': 'ірландская', 'bg': 'ирландски', 'bgn': 'ایرلندی', 'bn': 'আইরিশ', 'br': 'iwerzhoneg', 'brx': 'आईरिश', 'bs': 'irski', 'bs-Cyrl': 'ирски', 'bs-Latn': 'irski', 'ca': 'irlandès', 'ca-ES-valencia': 'irlandès', 'ccp': '𑄃𑄭𑄢𑄨𑄌𑄴', 'ce': 'ирландхойн', 'chr': 'ᎨᎵᎩ', 'ckb': 'ئیرلەندی', 'cs': 'irština', 'cy': 'Gwyddeleg', 'da': 'irsk', 'de': 'Irisch', 'dsb': 'iršćina', 'dz': 'ཨཱའི་རིཤ་ཁ', 'ee': 'irelanɖgbe', 'el': 'Ιρλανδικά', 'el-polyton': 'Ἰρλανδικά', 'en': 'Irish', 'en-Dsrt': '𐐌𐑉𐐮𐑇', 'eo': 'irlanda', 'es': 'irlandés', 'et': 'iiri', 'eu': 'irlandera', 'fa': 'ایرلندی', 'ff-Adlm': '𞤋𞤪𞤤𞤢𞤲𞤣𞤫𞥅𞤪𞤫', 'fi': 'iiri', 'fil': 'Irish', 'fo': 'írskt', 'fr': 'irlandais', 'fur': 'gaelic irlandês', 'fy': 'Iersk', 'ga': 'Gaeilge', 'gd': 'Gaeilge', 'gl': 'irlandés', 'gsw': 'Iirisch', 'gu': 'આઇરિશ', 'ha': 'Dan Irish', 'haw': 'ʻAiliki', 'he': 'אירית', 'hi': 'आयरिश', 'hi-Latn': 'Irish', 'hr': 'irski', 'hsb': 'iršćina', 'ht': 'irlandais', 'hu': 'ír', 'hy': 'իռլանդերեն', 'ia': 'irlandese', 'id': 'Irlandia', 'ig': 'Irish', 'is': 'írska', 'it': 'irlandese', 'ja': 'アイルランド語', 'jv': 'Irlandia', 'ka': 'ირლანდიური', 'kea': 'irlandes', 'kgp': 'irỹnej', 'kk': 'ирланд тілі', 'kk-Cyrl': 'ирланд тілі', 'km': 'អៀរឡង់', 'kn': 'ಐರಿಷ್', 'ko': '아일랜드어', 'kok': 'आयरिश', 'kok-Deva': 'आयरिश', 'ks': 'اَیرِش', 'ks-Arab': 'اَیرِش', 'ksh': 'Ihresch', 'ku': 'îrlendî', 'ky': 'ирландча', 'lb': 'Iresch', 'lkt': 'Irish Iyápi', 'lld': 'irlandesc', 'lo': 'ໄອຣິສ', 'lrc': 'ئیرلأندی', 'lt': 'airių', 'lv': 'īru', 'mai': 'आइरिश', 'mi': 'Airihi', 'mk': 'ирски', 'ml': 'ഐറിഷ്', 'mn': 'ирланд', 'mr': 'आयरिश', 'ms': 'Ireland', 'mt': 'Irlandiż', 'my': 'အိုင်းရစ်ရှ်', 'mzn': 'ایریش', 'nb': 'irsk', 'ne': 'आइरिस', 'nl': 'Iers', 'nn': 'irsk', 'no': 'irsk', 'nqo': 'ߌߙߑߟߊ߲ߘߌߞߊ߲', 'om': 'Afaan Ayirishii', 'or': 'ଆଇରିସ୍\u200c', 'os': 'ирландиаг', 'pa': 'ਆਇਰਸ਼', 'pa-Guru': 'ਆਇਰਸ਼', 'pcm': 'Aírísh Lángwej', 'pl': 'irlandzki', 'ps': 'ائيرلېنډي', 'pt': 'irlandês', 'qu': 'Irlandes Simi', 'rm': 'irlandais', 'ro': 'irlandeză', 'ru': 'ирландский', 'rw': 'Ikirilandi', 'sc': 'irlandesu', 'sd': 'آئرش', 'sd-Arab': 'آئرش', 'se': 'iirragiella', 'si': 'අයර්ලන්ත', 'sk': 'írčina', 'sl': 'irščina', 'smn': 'iirikielâ', 'so': 'Ayrish', 'sq': 'irlandisht', 'sr': 'ирски', 'sr-Cyrl': 'ирски', 'sr-Latn': 'irski', 'sv': 'iriska', 'sw': 'Kiayalandi', 'ta': 'ஐரிஷ்', 'te': 'ఐరిష్', 'tg': 'ирландӣ', 'th': 'ไอริช', 'ti': 'ኣየርላንድኛ', 'tk': 'irland dili', 'to': 'lea fakaʻaelani', 'tr': 'İrlandaca', 'tt': 'ирланд', 'ug': 'ئىرېلاندچە', 'uk': 'ірландська', 'ur': 'آئرش', 'uz': 'irland', 'uz-Cyrl': 'ирландча', 'uz-Latn': 'irland', 'vec': 'irlandeze', 'vi': 'Tiếng Ireland', 'wae': 'Iriš', 'wo': 'Irlànde', 'xnr': 'आयरिश', 'yi': 'איריש', 'yo': 'Èdè Ireland', 'yrl': 'irãdeixi médiu', 'yue': '愛爾蘭文', 'yue-Hans': '爱尔兰文', 'yue-Hant': '愛爾蘭文', 'zh': '爱尔兰语', 'zh-Hans': '爱尔兰语', 'zh-Hant': '愛爾蘭文', 'zu': 'isi-Irish'},
    'gaa': {'am': 'ጋ', 'ar': 'الجا', 'as': 'গা', 'ast': 'ga', 'az': 'qa', 'az-Cyrl': 'га', 'az-Latn': 'qa', 'be': 'га', 'be-tarask': 'га', 'bg': 'га', 'bgn': 'گا', 'bn': 'গা', 'br': 'ga', 'brx': 'गा', 'bs': 'ga', 'bs-Cyrl': 'га', 'bs-Latn': 'ga', 'ca': 'ga', 'ca-ES-valencia': 'ga', 'ccp': '𑄉𑄳𑄃', 'ce': 'га', 'chr': 'Ꭶ', 'ckb': 'گایی', 'cs': 'gaština', 'cy': 'Ga', 'da': 'ga', 'de': 'Ga', 'dsb': 'gašćina', 'el': 'Γκα', 'el-polyton': 'Γκα', 'en': 'Ga', 'es': 'ga', 'eu': 'gaera', 'fa': 'گایی', 'ff-Adlm': '𞤘𞤢𞥄𞤪𞤫', 'fi': 'ga', 'fil': 'Ga', 'fo': 'ga', 'fr': 'ga', 'fy': 'Ga', 'ga': 'Geáis', 'gaa': 'Gã', 'gd': 'Ga', 'gl': 'ga', 'gsw': 'Ga', 'gu': 'ગા', 'ha': 'Ga', 'he': 'גא', 'hi': 'गा', 'hi-Latn': 'Ga', 'hr': 'ga', 'hsb': 'gašćina', 'ht': 'ga', 'hu': 'ga', 'hy': 'գայերեն', 'ia': 'ga', 'id': 'Ga', 'ig': 'Ga', 'is': 'ga', 'it': 'ga', 'ja': 'ガ語', 'jv': 'Ga', 'ka': 'გა', 'kgp': 'ga', 'kk': 'га тілі', 'kk-Cyrl': 'га тілі', 'km': 'ហ្គា', 'kn': 'ಗ', 'ko': '가어', 'kok': 'गा', 'kok-Deva': 'गा', 'ks': 'گا', 'ks-Arab': 'گا', 'ksh': 'Jah', 'ku': 'gayî', 'ky': 'гача', 'lb': 'Ga-Sprooch', 'lo': 'ກາ', 'lt': 'ga', 'lv': 'ga', 'mai': 'गा', 'mi': 'Kā', 'mk': 'га', 'ml': 'ഗാ', 'mn': 'га', 'mr': 'गा', 'ms': 'Ga', 'mt': 'Ga', 'my': 'ဂါ', 'nb': 'ga', 'ne': 'गा', 'nl': 'Ga', 'nn': 'ga', 'no': 'ga', 'or': 'ଗା', 'pa': 'ਗਾ', 'pa-Guru': 'ਗਾ', 'pcm': 'Ga Lángwej', 'pl': 'ga', 'ps': 'Ga', 'pt': 'ga', 'qu': 'Ga Simi', 'rm': 'ga', 'ro': 'ga', 'ru': 'га', 'sc': 'ga', 'sd': 'گا', 'sd-Arab': 'گا', 'si': 'ගා', 'sk': 'ga', 'sl': 'ga', 'smn': 'gakielâ', 'so': 'Ga', 'sq': 'gaisht', 'sr': 'га', 'sr-Cyrl': 'га', 'sr-Latn': 'ga', 'sv': 'gã', 'sw': 'Ga', 'syr': 'ܓܐܝܬ', 'ta': 'கா', 'te': 'గా', 'th': 'กา', 'ti': 'ጋ', 'tk': 'ga dili', 'to': 'lea fakakā', 'tr': 'Ga dili', 'ug': 'گاچە', 'uk': 'га', 'ur': 'گا', 'uz': 'ga', 'uz-Cyrl': 'га', 'uz-Latn': 'ga', 'vec': 'ga', 'vi': 'Tiếng Ga', 'xnr': 'गा', 'yo': 'Èdè Gaa', 'yrl': 'ga', 'yue': '加族文', 'yue-Hans': '加族文', 'yue-Hant': '加族文', 'zh': '加族语', 'zh-Hans': '加族语', 'zh-Hant': '加族文', 'zu': 'isi-Ga'},
    'gab': {'en': 'Gabri'},
    'gac': {'en': 'Mixed Great Andamanese'},
    'gad': {'en': 'Gaddang'},
    'gae': {'en': 'Guarequena'},
    'gaf': {'en': 'Gende'},
    'gag': {'af': 'Gagauz', 'am': 'ጋጉዝኛ', 'ar': 'الغاغوز', 'ast': 'gagauz', 'az': 'qaqauz', 'az-Arab': 'قاقائوزجا', 'az-Latn': 'qaqauz', 'be': 'гагаузская', 'be-tarask': 'гагаузская', 'bg': 'гагаузки', 'bn': 'গাগাউজ', 'br': 'gagaouzeg', 'bs': 'gagauški', 'bs-Latn': 'gagauški', 'ca': 'gagaús', 'ca-ES-valencia': 'gagaús', 'ccp': '𑄉𑄉𑄃𑄪𑄌𑄴', 'ce': 'гагаузийн', 'cs': 'gagauzština', 'cy': 'Gagauz', 'da': 'gagauzisk', 'de': 'Gagausisch', 'dsb': 'gagauzšćina', 'el': 'Γκαγκάουζ', 'el-polyton': 'Γκαγκάουζ', 'en': 'Gagauz', 'es': 'gagauzo', 'et': 'gagauusi', 'eu': 'gagauzera', 'fa': 'گاگائوزیایی', 'fi': 'gagauzi', 'fil': 'Gagauz', 'fo': 'gagauz', 'fr': 'gagaouze', 'gd': 'Gagauz', 'gl': 'gagauz', 'gu': 'ગાગાઝ', 'he': 'גגאוזית', 'hi': 'गागौज़', 'hi-Latn': 'Gagauz', 'hr': 'gagauski', 'hsb': 'gagauzišćina', 'ht': 'gagaouze', 'hu': 'gagauz', 'hy': 'գագաուզերեն', 'id': 'Gagauz', 'is': 'gagás', 'it': 'gagauzo', 'ja': 'ガガウズ語', 'ka': 'გაგაუზური', 'kea': 'gagauz', 'kgp': 'gagavuj', 'kk': 'гагауз тілі', 'kk-Cyrl': 'гагауз тілі', 'km': 'កាគូស', 'kn': 'ಗಗೌಜ್', 'ko': '가가우스어', 'ky': 'гагаузча', 'lb': 'Gagausesch', 'lo': 'ກາກາອຸຊ', 'lrc': 'گاگائوز', 'lt': 'gagaūzų', 'lv': 'gagauzu', 'mk': 'гагауски', 'ml': 'ഗാഗൂസ്', 'mn': 'гагуз', 'mr': 'गागाउझ', 'ms': 'Gagauz', 'my': 'ဂါဂုဇ်', 'mzn': 'گاگائوزی', 'nb': 'gagausisk', 'ne': 'गगाउज', 'nl': 'Gagaoezisch', 'nn': 'gagausisk', 'no': 'gagausisk', 'pa': 'ਗਾਗੌਜ਼', 'pa-Guru': 'ਗਾਗੌਜ਼', 'pl': 'gagauski', 'pt': 'gagauz', 'ro': 'găgăuză', 'ru': 'гагаузский', 'si': 'ගගාස්', 'sk': 'gagauzština', 'sl': 'gagavščina', 'sq': 'gagauzisht', 'sr': 'гагауз', 'sr-Cyrl': 'гагауз', 'sr-Latn': 'gagauz', 'sv': 'gagauziska', 'sw': 'Kigagauz', 'ta': 'காகௌஸ்', 'te': 'గాగౌజ్', 'th': 'กากาอุซ', 'to': 'lea fakakakausi', 'tr': 'Gagavuzca', 'uk': 'гагаузька', 'ur': 'غاغاوز', 'uz': 'gagauz', 'uz-Latn': 'gagauz', 'vi': 'Tiếng Gagauz', 'yrl': 'gagausi', 'yue': '加告茲文', 'yue-Hans': '加告兹文', 'yue-Hant': '加告茲文', 'zh': '加告兹语', 'zh-Hans': '加告兹语', 'zh-Hant': '加告茲文', 'zu': 'isi-Gagauz'},
    'gah': {'en': 'Alekano'},
    'gai': {'en': 'Borei'},
    'gaj': {'en': 'Gadsup'},
    'gak': {'en': 'Gamkonora'},
    'gal': {'en': 'Galolen'},
    'gam': {'en': 'Kandawo'},
    'gan': {'af': 'Gan-Sjinees', 'am': 'ጋን ቻይንኛ', 'ar': 'الغان الصينية', 'ast': 'chinu gan', 'az': 'qan', 'az-Latn': 'qan', 'br': 'sinaeg Gan', 'ca': 'xinès gan', 'ca-ES-valencia': 'xinès gan', 'ccp': '𑄉𑄧𑄚𑄴', 'cs': 'čínština (dialekty Gan)', 'da': 'gan-kinesisk', 'en': 'Gan Chinese', 'es': 'chino gan', 'et': 'kani', 'fi': 'gan-kiina', 'fo': 'gan kinesiskt', 'ga': 'Sínis Gan', 'he': 'סינית גאן', 'hi-Latn': 'Gan Chinese', 'hr': 'gan kineski', 'hu': 'gan kínai', 'ja': '贛語', 'kgp': 'gỹn', 'kn': 'ಗಾನ್ ಚೀನೀಸ್', 'ko': '간어', 'ky': 'Гань Кытайча', 'lb': 'Gan-Chinesesch', 'lt': 'kinų kalbos dziangsi tarmė', 'mk': 'ган', 'ml': 'ഗാൻ ചൈനീസ്', 'mr': 'गॅन चिनी', 'ms': 'Cina Gan', 'ne': 'गान चिनियाँ', 'nl': 'Ganyu', 'pa': 'ਚੀਨੀ ਗਾਨ', 'pa-Guru': 'ਚੀਨੀ ਗਾਨ', 'ro': 'chineză gan', 'ru': 'гань', 'si': 'ගැන් චයිනිස්', 'ta': 'கன் சீனம்', 'te': 'గాన్ చైనీస్', 'th': 'จีนกั้น', 'to': 'lea fakasiaina-kani', 'tr': 'Gan Çincesi', 'uk': 'ґань', 'vi': 'Tiếng Cám', 'yrl': 'gã', 'yue': '贛語', 'yue-Hans': '赣语', 'yue-Hant': '贛語', 'zh': '赣语', 'zh-Hans': '赣语', 'zh-Hant': '贛語', 'zu': 'isi-Gan Chinese'},
    'gao': {'en': 'Gants'},
    'gap': {'en': 'Gal'},
    'gaq': {'en': "Gata'"},
    'gar': {'en': 'Galeya'},
    'gas': {'en': 'Adiwasi Garasia'},
    'gat': {'en': 'Kenati'},
    'gau': {'en': 'Mudhili Gadaba'},
    'gav': {'en': 'Gabutamon'},
    'gaw': {'en': 'Nobonob'},
    'gax': {'en': 'Borana-Arsi-Guji Oromo'},
    'gay': {'ar': 'الجايو', 'ast': 'gayo', 'az': 'qayo', 'az-Latn': 'qayo', 'bg': 'гайо', 'bn': 'গায়ো', 'br': 'gayo', 'brx': 'गायो', 'bs': 'gajo', 'bs-Cyrl': 'гајо', 'bs-Latn': 'gajo', 'ca': 'gayo', 'ca-ES-valencia': 'gayo', 'ccp': '𑄉𑄧𑄠𑄮', 'cs': 'gayo', 'cy': 'Gaio', 'da': 'gayo', 'de': 'Gayo', 'el': 'Γκάγιο', 'el-polyton': 'Γκάγιο', 'en': 'Gayo', 'es': 'gayo', 'et': 'gajo', 'fa': 'گایویی', 'fi': 'gajo', 'fr': 'gayo', 'fy': 'Gayo', 'gd': 'Gayo', 'gsw': 'Gayo', 'gu': 'ગાયો', 'he': 'גאיו', 'hi': 'गायो', 'hi-Latn': 'Gayo', 'hr': 'gayo', 'ht': 'gayo', 'hu': 'gajo', 'id': 'Gayo', 'is': 'gajó', 'it': 'gayo', 'ja': 'ガヨ語', 'kgp': 'gajo', 'kn': 'ಗಾಯೋ', 'ko': '가요어', 'ks': 'گیےیو', 'ks-Arab': 'گیےیو', 'lb': 'Gayo', 'lo': 'ກາໂຢ', 'lt': 'gajo', 'lv': 'gajo', 'mk': 'гајо', 'ml': 'ഗയൊ', 'mr': 'गायो', 'mt': 'Gayo', 'nb': 'gayo', 'ne': 'गायो', 'nl': 'Gayo', 'nn': 'gayo', 'no': 'gayo', 'or': 'ଗାୟୋ', 'pl': 'gayo', 'pt': 'gayo', 'rm': 'gayo', 'ro': 'gayo', 'ru': 'гайо', 'sk': 'gayo', 'sl': 'gajščina', 'sr': 'гајо', 'sr-Cyrl': 'гајо', 'sr-Latn': 'gajo', 'sv': 'gayo', 'ta': 'கயோ', 'te': 'గాయో', 'th': 'กาโย', 'to': 'lea fakakaio', 'tr': 'Gayo dili', 'ug': 'گايوچە', 'uk': 'гайо', 'vi': 'Tiếng Gayo', 'yrl': 'gayu', 'yue': '加約文', 'yue-Hans': '加约文', 'yue-Hant': '加約文', 'zh': '迦约语', 'zh-Hans': '迦约语', 'zh-Hant': '加約文'},
    'gaz': {'en': 'West Central Oromo'},
    'gba': {'ar': 'الجبيا', 'ast': 'gbaya', 'az': 'qabaya', 'az-Latn': 'qabaya', 'bg': 'гбая', 'bn': 'বায়া', 'br': 'gbaya', 'brx': 'ग्बाया', 'bs': 'gbaja', 'bs-Cyrl': 'гбаја', 'bs-Latn': 'gbaja', 'ca': 'gbaya', 'ca-ES-valencia': 'gbaya', 'ccp': '𑄝𑄠', 'cs': 'gbaja', 'cy': 'Gbaia', 'da': 'gbaya', 'de': 'Gbaya', 'el': 'Γκμπάγια', 'el-polyton': 'Γκμπάγια', 'en': 'Gbaya', 'es': 'gbaya', 'et': 'gbaja', 'fa': 'گبایایی', 'fi': 'gbaja', 'fr': 'gbaya', 'fy': 'Gbaya', 'gd': 'Gbaya', 'gsw': 'Gbaya', 'gu': 'બાયા', 'he': 'גבאיה', 'hi': 'ग्बाया', 'hi-Latn': 'Gbaya', 'hr': 'gbaya', 'ht': 'gbaya', 'hu': 'gbaja', 'id': 'Gbaya', 'is': 'gbaja', 'it': 'gbaya', 'ja': 'バヤ語', 'ka': 'გბაია', 'kgp': 'gemaja', 'kn': 'ಗ್ಬಾಯಾ', 'ko': '그바야어', 'ks': 'گبایا', 'ks-Arab': 'گبایا', 'lb': 'Gbaya-Sprooch', 'lkt': 'Gbaya Iyápi', 'lo': 'ບາຍາ', 'lt': 'gbaja', 'lv': 'gbaju', 'mk': 'гбаја', 'ml': 'ഗബ്യ', 'mr': 'बाया', 'ms': 'Gbaya', 'mt': 'Gbaya', 'nb': 'gbaya', 'ne': 'ग्बाया', 'nl': 'Gbaya', 'nn': 'gbaya', 'no': 'gbaya', 'or': 'ଗବାୟା', 'pl': 'gbaya', 'pt': 'gbaia', 'rm': 'gbaya', 'ro': 'gbaya', 'ru': 'гбая', 'sk': 'gbaja', 'sl': 'gbajščina', 'sr': 'гбаја', 'sr-Cyrl': 'гбаја', 'sr-Latn': 'gbaja', 'sv': 'gbaya', 'sw': 'Kigbaya', 'ta': 'பயா', 'te': 'గ్బాయా', 'th': 'กบายา', 'to': 'lea fakakapaia', 'tr': 'Gbaya', 'ug': 'گىباياچە', 'uk': 'гбайя', 'vi': 'Tiếng Gbaya', 'yrl': 'gibaya', 'yue': '葛巴亞文', 'yue-Hans': '葛巴亚文', 'yue-Hant': '葛巴亞文', 'zh': '格巴亚语', 'zh-Hans': '格巴亚语', 'zh-Hant': '葛巴亞文'},
    'gbb': {'en': 'Kaytetye'},
    'gbc': {'en': 'Garawa'},
    'gbd': {'en': 'Karajarri'},
    'gbe': {'en': 'Niksek'},
    'gbf': {'en': 'Gaikundi'},
    'gbg': {'en': 'Gbanziri'},
    'gbh': {'en': 'Defi Gbe'},
    'gbi': {'en': 'Galela'},
    'gbj': {'en': 'Bodo Gadaba'},
    'gbk': {'en': 'Gaddi'},
    'gbl': {'en': 'Gamit'},
    'gbm': {'en': 'Garhwali'},
    'gbn': {'en': "Mo'da"},
    'gbo': {'en': 'Northern Grebo'},
    'gbp': {'en': 'Gbaya-Bossangoa'},
    'gbq': {'en': 'Gbaya-Bozoum'},
    'gbr': {'en': 'Gbagyi'},
    'gbs': {'en': 'Gbesi Gbe'},
    'gbu': {'en': 'Gagadu'},
    'gbv': {'en': 'Gbanu'},
    'gbw': {'en': 'Gabi-Gabi'},
    'gbx': {'en': 'Eastern Xwla Gbe'},
    'gby': {'en': 'Gbari'},
    'gbz': {'ast': 'dari zoroastrianu', 'cs': 'daríjština (zoroastrijská)', 'cy': 'Dareg y Zoroastriaid', 'de': 'Gabri', 'en': 'Zoroastrian Dari', 'fa': 'دری زرتشتی', 'fi': 'zoroastrialaisdari', 'fr': 'dari zoroastrien', 'gd': 'Dari Zoroastrach', 'gu': 'ઝોરોસ્ટ્રિઅન દારી', 'hi-Latn': 'Zoroastrian Dari', 'ht': 'dari zoroastrien', 'hy': 'զրադաշտական դարի', 'it': 'dari zoroastriano', 'ja': 'ダリー語(ゾロアスター教)', 'ko': '조로아스터 다리어', 'lb': 'Zoroastrianescht Dari', 'lt': 'zoroastrų dari', 'mk': 'зороастриски дари', 'ms': 'Zoroastrian Dari', 'nb': 'zoroastrisk dari', 'nl': 'Zoroastrisch Dari', 'nn': 'zoroastrisk dari', 'no': 'zoroastrisk dari', 'pl': 'zaratusztriański dari', 'sv': 'zoroastrisk dari', 'th': 'ดารีโซโรอัสเตอร์', 'to': 'lea fakateli-soloasitelia', 'tr': 'Zerdüşt Daricesi', 'yue': '索羅亞斯德教達里文', 'yue-Hans': '索罗亚斯德教达里文', 'yue-Hant': '索羅亞斯德教達里文', 'zh-Hant': '索羅亞斯德教達里文'},
    'gcc': {'en': 'Mali'},
    'gcd': {'en': 'Ganggalida'},
    'gce': {'en': 'Galice'},
    'gcf': {'en': 'Guadeloupean Creole French'},
    'gcl': {'en': 'Grenadian Creole English'},
    'gcn': {'en': 'Gaina'},
    'gcr': {'en': 'Guianese Creole French'},
    'gct': {'en': 'Colonia Tovar German'},
    'gd': {'af': 'Skotse Gallies', 'ak': 'Skotlandfoɔ Galek Kasa', 'am': 'የስኮትላንድ ጌይሊክ', 'ar': 'الغيلية الأسكتلندية', 'as': 'স্কটিচ গেইলিক', 'ast': 'gaélicu escocés', 'az': 'Şotlandiya keltcəsi', 'az-Cyrl': 'шотланд келт', 'az-Latn': 'Şotlandiya keltcəsi', 'bal-Latn': 'Eskáti Géli', 'be': 'шатландская гэльская', 'be-tarask': 'шатландская гэльская', 'bg': 'шотландски келтски', 'bgn': 'اسکاتلندی گیلی', 'bn': 'স্কটিশ-গ্যেলিক', 'br': 'skoseg', 'brx': 'स्कटिस गैलिक', 'bs': 'škotski galski', 'bs-Cyrl': 'шкотски галски', 'bs-Latn': 'škotski galski', 'ca': 'gaèlic escocès', 'ca-ES-valencia': 'gaèlic escocès', 'ccp': '𑄃𑄨𑄌𑄴𑄇𑄧𑄖𑄴𑄥𑄧-𑄉𑄳𑄠𑄬𑄣𑄨𑄇𑄴', 'ce': 'гэлийн', 'chr': 'ᏍᎦᏗ ᎨᎵᎩ', 'ckb': 'گه\u200cلیكی سكۆتله\u200cندی', 'cs': 'skotská gaelština', 'cy': 'Gaeleg yr Alban', 'da': 'skotsk gælisk', 'de': 'Gälisch (Schottland)', 'dsb': 'šotišćina', 'el': 'Σκωτικά Κελτικά', 'el-polyton': 'Σκωτικὰ κελτικά', 'en': 'Scottish Gaelic', 'eo': 'skot-gaela', 'es': 'gaélico escocés', 'et': 'gaeli', 'eu': 'Eskoziako gaelikoa', 'fa': 'گیلی اسکاتلندی', 'ff-Adlm': '𞤅𞤭𞤳𞤮𞤼𞤭𞤧𞤪𞤫 𞤘𞤢𞤫𞤭𞤳', 'fi': 'gaeli', 'fil': 'Scottish Gaelic', 'fo': 'skotskt gæliskt', 'fr': 'gaélique écossais', 'fur': 'gaelic scozês', 'fy': 'Schotsk Gaelic', 'ga': 'Gaeilge na hAlban', 'gd': 'Gàidhlig', 'gl': 'gaélico escocés', 'gsw': 'Schottisch-Gäälisch', 'gu': 'સ્કોટીસ ગેલિક', 'ha': 'Kʼabilan Scots Gaelic', 'he': 'גאלית סקוטית', 'hi': 'स्कॉटिश गाएलिक', 'hi-Latn': 'Scottish Gaelic', 'hr': 'škotski gaelski', 'hsb': 'šotiska gelšćina', 'ht': 'gaélique écossais', 'hu': 'skóciai kelta', 'hy': 'շոտլանդական գաելերեն', 'ia': 'gaelico scotese', 'id': 'Gaelik Skotlandia', 'ig': 'Asụsụ Scottish Gaelic', 'is': 'skosk gelíska', 'it': 'gaelico scozzese', 'ja': 'スコットランド・ゲール語', 'jv': 'Gaulia', 'ka': 'შოტლანდიური გელური', 'kgp': 'gajériko enhkosej', 'kk': 'шотландиялық гэль тілі', 'kk-Cyrl': 'шотландиялық гэль тілі', 'km': 'ស្កុតហ្កែលិគ', 'kn': 'ಸ್ಕಾಟಿಶ್ ಗೆಲಿಕ್', 'ko': '스코틀랜드 게일어', 'kok': 'स्कॉटीश गॅलीक', 'kok-Deva': 'स्कॉटीश गॅलीक', 'ks': 'سکوٹِش گیےلِک', 'ks-Arab': 'سکوٹِش گیےلِک', 'ku': 'gaelîka skotî', 'ky': 'шотладиялык гелча', 'lb': 'Schottescht Gällesch', 'lld': 'gaelich scozesc', 'lo': 'ສະກັອດເກລິກ', 'lt': 'škotų (gėlų)', 'lv': 'skotu gēlu', 'mai': 'स्काउटिस गेलिक', 'mi': 'Keiriki Kotimana', 'mk': 'шкотски гелски', 'ml': 'സ്കോട്ടിഷ് ഗൈലിക്', 'mn': 'шотландын гел', 'mr': 'स्कॉटिश गेलिक', 'ms': 'Scots Gaelic', 'mt': 'Galliku Skoċċiż', 'my': 'စကော့တစ်ရှ် ဂေးလစ်ခ်', 'nb': 'skotsk-gælisk', 'ne': 'स्कटिस गाएलिक', 'nl': 'Schots-Gaelisch', 'nn': 'skotsk-gælisk', 'no': 'skotsk-gælisk', 'nqo': 'ߋߞߐߛߌ߬ ߖ߭ߋߏߟߌߞߊ߲', 'om': 'Scots Gaelic', 'or': 'ସ୍କଟିସ୍ ଗାଏଲିକ୍', 'pa': 'ਸਕਾਟਿਸ਼ ਗੇਲਿਕ', 'pa-Guru': 'ਸਕਾਟਿਸ਼ ਗੇਲਿਕ', 'pcm': 'Gaelík Lángwej ọf Gael Pípol fọ Skọ́tland', 'pl': 'szkocki gaelicki', 'ps': 'سکاټلېنډي ګېلک', 'pt': 'gaélico escocês', 'qu': 'Gaelico Escoces Simi', 'rm': 'gaelic scot', 'ro': 'gaelică scoțiană', 'ru': 'гэльский', 'rw': 'Ikigaluwa cy’Igisweduwa', 'sc': 'gaèlicu iscotzesu', 'sd': 'اسڪاٽش گيلڪ', 'sd-Arab': 'اسڪاٽش گيلڪ', 'si': 'ස්කොට්ටිශ් ගෙලික්', 'sk': 'škótska gaelčina', 'sl': 'škotska gelščina', 'smn': 'skottilâš gaelikielâ', 'so': 'Iskot Giilik', 'sq': 'galishte skoceze', 'sr': 'шкотски гелски', 'sr-Cyrl': 'шкотски гелски', 'sr-Latn': 'škotski gelski', 'sv': 'skotsk gaeliska', 'sw': 'Kigaeli cha Uskoti', 'ta': 'ஸ்காட்ஸ் கேலிக்', 'te': 'స్కాటిష్ గేలిక్', 'tg': 'шотландии гэлӣ', 'th': 'เกลิกสกอต', 'ti': 'ስኮትላንዳዊ ጋኤሊክኛ', 'tk': 'şotland kelt dili', 'to': 'lea fakakaeliki', 'tr': 'İskoç Gaelcesi', 'tt': 'шотланд гэль', 'ug': 'شوتلاندىيە گايلچىسى', 'uk': 'шотландська гельська', 'ur': 'سکاٹش گیلک', 'uz': 'shotland-gel', 'uz-Cyrl': 'шотландча гаелик', 'uz-Latn': 'shotland-gel', 'vec': 'gaèlego scoseze', 'vi': 'Tiếng Gael Scotland', 'wo': 'Galuwaa bu Ekos', 'xnr': 'स्कॉटिश गाएलिक', 'yi': 'סקאטיש געליש', 'yo': 'Èdè Gaelik ti Ilu Scotland', 'yrl': 'gaériku ekusei', 'yue': '蘇格蘭蓋爾文', 'yue-Hans': '苏格兰盖尔文', 'yue-Hant': '蘇格蘭蓋爾文', 'zh': '苏格兰盖尔语', 'zh-Hans': '苏格兰盖尔语', 'zh-Hant': '蘇格蘭蓋爾文', 'zu': 'isi-Scottish Gaelic'},
    'gda': {'en': 'Gade Lohar'},
    'gdb': {'en': 'Pottangi Ollar Gadaba'},
    'gdc': {'en': 'Gugu Badhun'},
    'gdd': {'en': 'Gedaged'},
    'gde': {'en': 'Gude'},
    'gdf': {'en': 'Guduf-Gava'},
    'gdg': {'en': "Ga'dang"},
    'gdh': {'en': 'Gadjerawang'},
    'gdi': {'en': 'Gundi'},
    'gdj': {'en': 'Gurdjar'},
    'gdk': {'en': 'Gadang'},
    'gdl': {'en': 'Dirasha'},
    'gdm': {'en': 'Laal'},
    'gdn': {'en': 'Umanakaina'},
    'gdo': {'en': 'Ghodoberi'},
    'gdq': {'en': 'Mehri'},
    'gdr': {'en': 'Wipi'},
    'gds': {'en': 'Ghandruk Sign Language'},
    'gdt': {'en': 'Kungardutyi'},
    'gdu': {'en': 'Gudu'},
    'gdx': {'en': 'Godwari'},
    'gea': {'en': 'Geruma'},
    'geb': {'en': 'Kire'},
    'gec': {'en': 'Gboloo Grebo'},
    'ged': {'en': 'Gade'},
    'gef': {'en': 'Gerai'},
    'geg': {'en': 'Gengle'},
    'geh': {'en': 'Hutterite German'},
    'gei': {'en': 'Gebe'},
    'gej': {'en': 'Gen'},
    'gek': {'en': 'Ywom'},
    'gel': {'en': "ut-Ma'in"},
    'gem': {'en': 'Germanic languages'},
    'gem-pro': {'en': 'Proto-Germanic'},
    'geq': {'en': 'Geme'},
    'ges': {'en': 'Geser-Gorom'},
    'gev': {'en': 'Eviya'},
    'gew': {'en': 'Gera'},
    'gex': {'en': 'Garre'},
    'gey': {'en': 'Enya'},
    'gez': {'af': 'Geez', 'am': 'ግዕዝኛ', 'ar': 'الجعزية', 'as': 'গীজ', 'ast': 'geez', 'az': 'qez', 'az-Cyrl': 'гез', 'az-Latn': 'qez', 'bal-Latn': 'Géz', 'be': 'геэз', 'be-tarask': 'геэз', 'bg': 'гииз', 'bgn': 'گعزی', 'bn': 'গীজ', 'br': 'gezeg', 'brx': 'गीज', 'bs': 'staroetiopski', 'bs-Cyrl': 'џиз', 'bs-Latn': 'staroetiopski', 'ca': 'gueez', 'ca-ES-valencia': 'gueez', 'ccp': '𑄉𑄩𑄎𑄴', 'ce': 'геэз', 'chr': 'ᎩᏏ', 'ckb': 'گیزی', 'cs': 'geez', 'cy': 'Geez', 'da': 'geez', 'de': 'Geez', 'dsb': 'geezišćina', 'el': 'Γκιζ', 'el-polyton': 'Γκιζ', 'en': 'Geez', 'es': 'geez', 'et': 'etioopia', 'eu': 'ge’eza', 'fa': 'گی\u200cئزی', 'ff-Adlm': '𞤘𞤫𞥅𞤶𞤪𞤫', 'fi': 'ge’ez', 'fil': 'Geez', 'fo': 'geez', 'fr': 'guèze', 'fy': 'Geez', 'ga': 'Aetóipis', 'gd': 'Ge’ez', 'gl': 'ge’ez', 'gsw': 'Geez', 'gu': 'ગીઝ', 'ha': 'Geez', 'he': 'געז', 'hi': 'गीज़', 'hi-Latn': 'Geez', 'hr': 'geez', 'hsb': 'geezišćina', 'ht': 'guèze', 'hu': 'geez', 'hy': 'գեեզ', 'ia': 'ge’ez', 'id': 'Geez', 'ig': 'Geez', 'is': 'gís', 'it': 'geez', 'ja': 'ゲエズ語', 'jv': 'Gees', 'ka': 'გეეზი', 'kgp': 'gij', 'kk': 'геэз тілі', 'kk-Cyrl': 'геэз тілі', 'km': 'ជីស', 'kn': 'ಗೀಝ್', 'ko': '게이즈어', 'kok': 'गेझ', 'kok-Deva': 'गेझ', 'ks': 'گیٖز', 'ks-Arab': 'گیٖز', 'ksh': 'Ji’is-Ahl-Ättejohpesch', 'ku': 'geez', 'ky': 'гиизче', 'lb': 'Geez', 'lo': 'ກີ', 'lt': 'gyz', 'lv': 'gēzu', 'mai': 'गीज', 'mi': 'Kīhi', 'mk': 'гиз', 'ml': 'ഗീസ്', 'mn': 'гийз', 'mr': 'गीझ', 'ms': 'Geez', 'mt': 'Geez', 'my': 'ဂီးဇ်', 'nb': 'geez', 'ne': 'गिज', 'nl': 'Ge’ez', 'nn': 'geez', 'no': 'geez', 'or': 'ଗୀଜ୍', 'pa': 'ਜੀਜ਼', 'pa-Guru': 'ਜੀਜ਼', 'pcm': 'Giiz Lángwej', 'pl': 'gyyz', 'ps': 'ګیز', 'pt': 'geez', 'qu': 'Geez Simi', 'rm': 'geez', 'ro': 'geez', 'ru': 'геэз', 'sc': 'ge’ez', 'sd': 'جيز', 'sd-Arab': 'جيز', 'si': 'ගීස්', 'sk': 'etiópčina', 'sl': 'etiopščina', 'smn': 'ge’ez', 'so': 'Geez', 'sq': 'gizisht', 'sr': 'геез', 'sr-Cyrl': 'геез', 'sr-Latn': 'geez', 'sv': 'etiopiska', 'sw': 'Kige’ez', 'ta': 'கீஜ்', 'te': 'జీజ్', 'th': 'กีซ', 'ti': 'ግእዝ', 'tk': 'geez dili', 'to': 'lea fakasiʻisi', 'tr': 'Geez', 'ug': 'گىزچە', 'uk': 'гєез', 'ur': 'گیز', 'uz': 'geez', 'uz-Cyrl': 'геэз', 'uz-Latn': 'geez', 'vec': 'geez', 'vi': 'Tiếng Geez', 'xnr': 'गीज़', 'yo': 'Ede Gẹ́sì', 'yrl': 'giixi', 'yue': '吉茲文', 'yue-Hans': '吉兹文', 'yue-Hant': '吉茲文', 'zh': '吉兹语', 'zh-Hans': '吉兹语', 'zh-Hant': '吉茲文', 'zu': 'isi-Geez'},
    'gfk': {'en': 'Patpatar'},
    'gft': {'en': 'Gafat'},
    'gfx': {'en': 'Mangetti Dune ǃXung'},
    'gga': {'en': 'Gao'},
    'ggb': {'en': 'Gbii'},
    'ggd': {'en': 'Gugadj'},
    'gge': {'en': 'Gurr-goni'},
    'ggg': {'en': 'Gurgula'},
    'ggk': {'en': 'Kungarakany'},
    'ggl': {'en': 'Ganglau'},
    'ggn': {'en': 'Eastern Gurung'},
    'ggo': {'en': 'Southern Gondi'},
    'ggr': {'en': 'Aghu Tharnggalu'},
    'ggt': {'en': 'Gitua'},
    'ggu': {'en': 'Gagu'},
    'ggw': {'en': 'Gogodala'},
    'gha': {'en': 'Ghadamès'},
    'ghc': {'en': 'Hiberno-Scottish Gaelic'},
    'ghe': {'en': 'Southern Ghale'},
    'ghh': {'en': 'Northern Ghale'},
    'ghk': {'en': 'Geko Karen'},
    'ghl': {'en': 'Ghulfan'},
    'ghn': {'en': 'Ghanongga'},
    'gho': {'en': 'Ghomara'},
    'ghr': {'en': 'Ghera'},
    'ghs': {'en': 'Guhu-Samane'},
    'ght': {'en': 'Kuke'},
    'gia': {'en': 'Kija'},
    'gib': {'en': 'Gibanawa'},
    'gic': {'en': 'Gail'},
    'gid': {'en': 'Gidar'},
    'gie': {'en': 'Gaɓogbo'},
    'gig': {'en': 'Goaria'},
    'gih': {'en': 'Githabul'},
    'gii': {'en': 'Girirra'},
    'gil': {'af': 'Gilbertees', 'am': 'ጅልበርትስ', 'ar': 'لغة أهل جبل طارق', 'as': 'গিলবাৰ্টিছ', 'ast': 'gilbertés', 'az': 'qilbert', 'az-Cyrl': 'гилберт', 'az-Latn': 'qilbert', 'be': 'кірыбаці', 'be-tarask': 'кірыбаці', 'bg': 'гилбертски', 'bgn': 'گیلبیرتی', 'bn': 'গিলবার্টিজ', 'br': 'gilberteg', 'brx': 'गील्बर्टिस', 'bs': 'gilbertski', 'bs-Cyrl': 'гилбертшки', 'bs-Latn': 'gilbertski', 'ca': 'gilbertès', 'ca-ES-valencia': 'gilbertès', 'ccp': '𑄉𑄨𑄣𑄴𑄝𑄢𑄴𑄑𑄨𑄎𑄴', 'ce': 'гильбертийн', 'chr': 'ᎩᏇᏘᏏ', 'ckb': 'گیلبێرتی', 'cs': 'kiribatština', 'cy': 'Gilberteg', 'da': 'gilbertesisk', 'de': 'Kiribatisch', 'dsb': 'gilbertezišćina', 'el': 'Γκιλμπερτίζ', 'el-polyton': 'Γκιλμπερτίζ', 'en': 'Gilbertese', 'en-Dsrt': '𐐘𐐮𐑊𐐺𐐯𐑉𐐻𐐨𐑆', 'es': 'gilbertés', 'et': 'kiribati', 'eu': 'kiribatiera', 'fa': 'گیلبرتی', 'ff-Adlm': '𞤘𞤭𞤤𞤦𞤫𞤪𞤼𞤫𞥅𞤧𞤪𞤫', 'fi': 'kiribati', 'fil': 'Gilbertese', 'fo': 'kiribatiskt', 'fr': 'gilbertin', 'fy': 'Gilberteesk', 'ga': 'Gilbeartais', 'gd': 'Ciribeasais', 'gl': 'kiribatiano', 'gsw': 'Gilbertesisch', 'gu': 'જિલ્બરટીઝ', 'ha': 'Gilbertese', 'he': 'קיריבטית', 'hi': 'गिल्बरतीस', 'hi-Latn': 'Gilbertese', 'hr': 'gilbertski', 'hsb': 'gilbertezišćina', 'ht': 'gilbertin', 'hu': 'ikiribati', 'hy': 'կիրիբատի', 'ia': 'gilbertese', 'id': 'Gilbert', 'ig': 'Gilbertese', 'is': 'gilberska', 'it': 'gilbertese', 'ja': 'キリバス語', 'jv': 'Gilbertese', 'ka': 'გილბერტული', 'kgp': 'gimertej', 'kk': 'гильберт тілі', 'kk-Cyrl': 'гильберт тілі', 'km': 'ហ្គីលបឺទ', 'kn': 'ಗಿಲ್ಬರ್ಟೀಸ್', 'ko': '키리바시어', 'kok': 'गिलबर्टीस', 'kok-Deva': 'गिलबर्टीस', 'ks': 'گِلبٔرٹیٖز', 'ks-Arab': 'گِلبٔرٹیٖز', 'ksh': 'Jillbättehsesch', 'ku': 'kîrîbatî', 'ky': 'гилбертче', 'lb': 'Gilbertesesch', 'lo': 'ກິນເບີເທັດ', 'lt': 'kiribati', 'lv': 'kiribatiešu', 'mai': 'गिलबर्टीज', 'mi': 'Kiripatīhi', 'mk': 'гилбертански', 'ml': 'ഗിൽബർട്ടീസ്', 'mn': 'гилберт', 'mr': 'जिल्बरटीज', 'ms': 'Kiribati', 'mt': 'Gilbertjan', 'my': 'ကာရီဗာတီ', 'nb': 'kiribatisk', 'ne': 'गिल्बर्टी', 'nl': 'Gilbertees', 'nn': 'kiribatisk', 'no': 'kiribatisk', 'or': 'ଜିବ୍ରାଟୀଜ୍', 'pa': 'ਗਿਲਬਰਤੀਜ਼', 'pa-Guru': 'ਗਿਲਬਰਤੀਜ਼', 'pcm': 'Gílbátís Lángwej', 'pl': 'gilbertański', 'ps': 'گلبرتي', 'pt': 'gilbertês', 'qu': 'Gilbertese Simi', 'rm': 'gilbertais', 'ro': 'gilbertină', 'ru': 'гилбертский', 'sc': 'gilbertesu', 'sd': 'گلبرٽيز', 'sd-Arab': 'گلبرٽيز', 'si': 'ගිල්බර්ටීස්', 'sk': 'kiribatčina', 'sl': 'kiribatščina', 'smn': 'kiribatikielâ', 'so': 'Gilbertese', 'sq': 'gilbertazisht', 'sr': 'гилбертски', 'sr-Cyrl': 'гилбертски', 'sr-Latn': 'gilbertski', 'sv': 'gilbertiska', 'sw': 'Kikiribati', 'ta': 'கில்பெர்டீஸ்', 'te': 'గిల్బర్టీస్', 'th': 'กิลเบอร์ต', 'ti': 'ጊልበርትኛ', 'tk': 'gilbert dili', 'to': 'lea fakakilipasi', 'tr': 'Kiribatice', 'ug': 'گىلبېرتچە', 'uk': 'гільбертська', 'ur': 'گلبرتیز', 'uz': 'gilbert', 'uz-Cyrl': 'гилбертча', 'uz-Latn': 'gilbert', 'vec': 'gilberteze', 'vi': 'Tiếng Gilbert', 'xnr': 'गिल्बरतीस', 'yo': 'Èdè Gibaati', 'yrl': 'giubetei', 'yue': '吉爾伯特群島文', 'yue-Hans': '吉尔伯特群岛文', 'yue-Hant': '吉爾伯特群島文', 'zh': '吉尔伯特语', 'zh-Hans': '吉尔伯特语', 'zh-Hant': '吉爾伯特群島文', 'zu': 'isi-Gilbertese'},
    'gim': {'en': 'Gimi (Eastern Highlands)'},
    'gin': {'en': 'Hinukh'},
    'gio': {'en': 'Gelao'},
    'gip': {'en': 'Gimi (West New Britain)'},
    'giq': {'en': 'Green Gelao'},
    'gir': {'en': 'Red Gelao'},
    'gis': {'en': 'North Giziga'},
    'git': {'en': 'Gitxsan'},
    'giu': {'en': 'Mulao'},
    'giw': {'en': 'White Gelao'},
    'gix': {'en': 'Gilima'},
    'giy': {'en': 'Giyug'},
    'giz': {'en': 'South Giziga'},
    'gji': {'en': 'Geji'},
    'gjk': {'en': 'Kachi Koli'},
    'gjm': {'en': 'Gunditjmara'},
    'gjn': {'en': 'Gonja'},
    'gjr': {'en': 'Gurindji Kriol'},
    'gju': {'en': 'Gujari'},
    'gka': {'en': 'Guya'},
    'gkd': {'en': 'Magɨ (Madang Province)'},
    'gke': {'en': 'Ndai'},
    'gkn': {'en': 'Gokana'},
    'gko': {'en': 'Kok-Nar'},
    'gkp': {'en': 'Guinea Kpelle'},
    'gku': {'en': 'ǂUngkue'},
    'gl': {'af': 'Galisies', 'ak': 'Galisia kasa', 'am': 'ጋሊሺያንኛ', 'ar': 'الجاليكية', 'as': 'গেলিচিয়ান', 'ast': 'gallegu', 'az': 'qalisiya', 'az-Arab': 'قالیسیاجا', 'az-Cyrl': 'галисија', 'az-Latn': 'qalisiya', 'bal-Latn': 'Galéki', 'be': 'галісійская', 'be-tarask': 'галісійская', 'bg': 'галисийски', 'bgn': 'گالیسی', 'bn': 'গ্যালিশিয়', 'br': 'galizeg', 'brx': 'गेलिसियान', 'bs': 'galicijski', 'bs-Cyrl': 'галски', 'bs-Latn': 'galicijski', 'ca': 'gallec', 'ca-ES-valencia': 'gallec', 'ccp': '𑄉𑄳𑄠𑄣𑄨𑄥𑄨𑄠𑄧', 'ce': 'галисийн', 'chr': 'ᎦᎵᏏᎠᏂ', 'ckb': 'گالیسی', 'cs': 'galicijština', 'cy': 'Galisieg', 'da': 'galicisk', 'de': 'Galicisch', 'dsb': 'galicišćina', 'dz': 'གལ་ཨིས་ཨི་ཡན་ཁ', 'ee': 'galatagbe', 'el': 'Γαλικιανά', 'el-polyton': 'Γαλικιανά', 'en': 'Galician', 'eo': 'galega', 'es': 'gallego', 'et': 'galeegi', 'eu': 'galiziera', 'fa': 'گالیسیایی', 'ff-Adlm': '𞤘𞤢𞤤𞤭𞥅𞤧𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'galicia', 'fil': 'Galician', 'fo': 'galisiskt', 'fr': 'galicien', 'fur': 'galizian', 'fy': 'Galisysk', 'ga': 'Gailísis', 'gd': 'Gailìsis', 'gl': 'galego', 'gsw': 'Galizisch', 'gu': 'ગેલિશિયન', 'ha': 'Bagalike', 'he': 'גליציאנית', 'hi': 'गैलिशियन', 'hi-Latn': 'Galician', 'hr': 'galicijski', 'hsb': 'galicišćina', 'ht': 'galicien', 'hu': 'gallego', 'hy': 'գալիսերեն', 'ia': 'galleco', 'id': 'Galisia', 'ig': 'Galician', 'is': 'galisíska', 'it': 'galiziano', 'ja': 'ガリシア語', 'jv': 'Galisia', 'ka': 'გალისიური', 'kea': 'galegu', 'kgp': 'garego', 'kk': 'галисия тілі', 'kk-Cyrl': 'галисия тілі', 'km': 'ហ្កាលីស្យាន', 'kn': 'ಗ್ಯಾಲಿಶಿಯನ್', 'ko': '갈리시아어', 'kok': 'गेलीशियन', 'kok-Deva': 'गेलीशियन', 'ks': 'گیلِشِیَن', 'ks-Arab': 'گیلِشِیَن', 'ksh': 'Jallihzesch', 'ku': 'galîsî', 'kxv': 'galesiati', 'kxv-Deva': 'ग्यालिसियन', 'kxv-Latn': 'galesiati', 'kxv-Orya': 'ଗଲସିଆତି', 'kxv-Telu': 'గాలసియన్', 'ky': 'галисияча', 'lb': 'Galizesch', 'lkt': 'Galician Iyápi', 'lld': 'galizian', 'lo': 'ກາລິຊຽນ', 'lrc': 'گالیسی', 'lt': 'galisų', 'lv': 'galisiešu', 'mai': 'गेलिशियन', 'mi': 'Karīhia', 'mk': 'галисиски', 'ml': 'ഗലീഷ്യൻ', 'mn': 'галего', 'mr': 'गॅलिशियन', 'ms': 'Galicia', 'mt': 'Galiċjan', 'my': 'ဂါလီစီယာ', 'mzn': 'گالیک', 'nb': 'galisisk', 'ne': 'गलिसियाली', 'nl': 'Galicisch', 'nn': 'galisisk', 'no': 'galisisk', 'nqo': 'ߜ߭ߊߟߌߛߌߞߊ߲', 'om': 'Afaan Galishii', 'or': 'ଗାଲିସିଆନ୍\u200c', 'pa': 'ਗੈਲਿਸ਼ਿਅਨ', 'pa-Guru': 'ਗੈਲਿਸ਼ਿਅਨ', 'pcm': 'Galísiá Lángwej', 'pl': 'galicyjski', 'ps': 'ګلېشيايي', 'pt': 'galego', 'qu': 'Gallego Simi', 'rm': 'galician', 'ro': 'galiciană', 'ru': 'галисийский', 'rw': 'Ikigalisiya', 'sc': 'galitzianu', 'sd': 'گليشئين', 'sd-Arab': 'گليشئين', 'si': 'ගැලීසියානු', 'sk': 'galícijčina', 'sl': 'galicijščina', 'smn': 'galiciakielâ', 'so': 'Galiisiyaan', 'sq': 'galicisht', 'sr': 'галицијски', 'sr-Cyrl': 'галицијски', 'sr-Latn': 'galicijski', 'sv': 'galiciska', 'sw': 'Kigalisi', 'syr': 'ܓܠܝܩܝܬ', 'ta': 'காலிஸியன்', 'te': 'గాలిషియన్', 'tg': 'галисиягӣ', 'th': 'กาลิเซีย', 'ti': 'ጋሊሽያን', 'tk': 'galisiý dili', 'to': 'lea fakakalisia', 'tr': 'Galiçyaca', 'tt': 'галисия', 'ug': 'گالىچە', 'uk': 'галісійська', 'ur': 'گالیشیائی', 'uz': 'galisiy', 'uz-Cyrl': 'галицийча', 'uz-Latn': 'galisiy', 'vec': 'galisian', 'vi': 'Tiếng Galician', 'wae': 'Galiziš', 'wo': 'Galisiye', 'xnr': 'गैलिशियन', 'yi': 'גאַלישיש', 'yo': 'Èdè Galicia', 'yrl': 'garegu', 'yue': '加里西亞文', 'yue-Hans': '加利西亚文', 'yue-Hant': '加里西亞文', 'zh': '加利西亚语', 'zh-Hans': '加利西亚语', 'zh-Hant': '加利西亞文', 'zu': 'isi-Galicia'},
    'glb': {'en': 'Belning'},
    'glc': {'en': 'Bon Gula'},
    'gld': {'en': 'Nanai'},
    'glh': {'en': 'Northwest Pashai'},
    'gli': {'en': 'Guliguli'},
    'glj': {'en': 'Gula Iro'},
    'glk': {'ast': 'gilaki', 'ca': 'gilaki', 'ca-ES-valencia': 'gilaki', 'cs': 'gilačtina', 'de': 'Gilaki', 'en': 'Gilaki', 'et': 'gilaki', 'fa': 'گیلکی', 'fi': 'gilaki', 'fr': 'gilaki', 'gd': 'Gilaki', 'hi-Latn': 'Gilaki', 'ht': 'gilaki', 'id': 'Gilaki', 'it': 'gilaki', 'ja': 'ギラキ語', 'ko': '길라키어', 'lb': 'Gilaki', 'lt': 'gilaki', 'mk': 'гилански', 'ms': 'Gilaki', 'nb': 'gileki', 'ne': 'गिलाकी', 'nl': 'Gilaki', 'nn': 'gileki', 'no': 'gileki', 'pl': 'giliański', 'sv': 'gilaki', 'th': 'กิลากี', 'to': 'lea fakakilaki', 'tr': 'Gilanice', 'vi': 'Tiếng Gilaki', 'yue': '吉拉基文', 'yue-Hans': '吉拉基文', 'yue-Hant': '吉拉基文', 'zh-Hant': '吉拉基文'},
    'gll': {'en': 'Garlali'},
    'glo': {'en': 'Galambu'},
    'glr': {'en': 'Glaro-Twabo'},
    'glu': {'en': 'Gula (Chad)'},
    'glw': {'en': 'Glavda'},
    'gly': {'en': 'Gule'},
    'gma': {'en': 'Gambera'},
    'gmb': {'en': "Gula'alaa"},
    'gmd': {'en': 'Mághdì'},
    'gme': {'en': 'East Germanic languages'},
    'gme-cgo': {'en': 'Crimean Gothic'},
    'gmg': {'en': 'Magɨyi'},
    'gmh': {'ar': 'الألمانية العليا الوسطى', 'ast': 'altualemán mediu', 'az': 'orta yüksək alman', 'az-Latn': 'orta yüksək alman', 'bg': 'средновисоконемски', 'bn': 'মধ্য-উচ্চ জার্মানি', 'br': 'krennalamaneg uhel', 'brx': 'मध्य उच्चस्तरी जर्मन', 'bs': 'srednjovjekovni gornjonjemački', 'bs-Cyrl': 'средњи високи немачки', 'bs-Latn': 'srednjovjekovni gornjonjemački', 'ca': 'alt alemany mitjà', 'ca-ES-valencia': 'alt alemany mitjà', 'ccp': '𑄟𑄧𑄖𑄴𑄙𑄳𑄠𑄧-𑄅𑄪𑄉𑄪𑄢𑄬 𑄎𑄢𑄴𑄟𑄚𑄩', 'cs': 'hornoněmčina (středověká)', 'cy': 'Almaeneg Uchel Canol', 'da': 'middelhøjtysk', 'de': 'Mittelhochdeutsch', 'el': 'Μέσα Άνω Γερμανικά', 'el-polyton': 'Μέσα Άνω Γερμανικά', 'en': 'Middle High German', 'en-Dsrt': '𐐣𐐮𐐼𐐲𐑊 𐐐𐐴 𐐖𐐲𐑉𐑋𐐲𐑌', 'es': 'alto alemán medio', 'et': 'keskülemsaksa', 'fa': 'آلمانی معیار میانه', 'fi': 'keskiyläsaksa', 'fr': 'moyen haut-allemand', 'fy': 'Middelheechdútsk', 'ga': 'Meán-Ard-Ghearmáinis', 'gd': 'Meadhan-Àrd-Gearmailtis', 'gsw': 'Mittelhochtüütsch', 'gu': 'મધ્ય હાઇ જર્મન', 'he': 'גרמנית בינונית-גבוהה', 'hi': 'मध्यकालीन हाइ जर्मन', 'hi-Latn': 'Middle High German', 'hr': 'srednjogornjonjemački', 'ht': 'moyen haut-allemand', 'hu': 'közép felső német', 'id': 'Jerman Abad Pertengahan', 'is': 'miðháþýska', 'it': 'tedesco medio alto', 'ja': '中高ドイツ語', 'ka': 'საშუალო ზემოგერმანული', 'kgp': 'arimỹv-téj kuju', 'kn': 'ಮಧ್ಯ ಹೈ ಜರ್ಮನ್', 'ko': '중세 고지 독일어', 'ks': 'وَسطی ہاے جٔرمَن', 'ks-Arab': 'وَسطی ہاے جٔرمَن', 'lb': 'Mëttelhéichdäitsch', 'lo': 'ເຢຍລະມັນສູງກາງ', 'lt': 'Vidurio Aukštosios Vokietijos', 'lv': 'vidusaugšvācu', 'mk': 'средногорногермански', 'ml': 'മദ്ധ്യ ഉച്ച ജർമൻ', 'mr': 'मिडल हाय जर्मन', 'mt': 'Ġermaniż Medjevali Pulit', 'my': 'အလယ်ပိုင်း အမြင့် ဂျာမန်', 'nb': 'mellomhøytysk', 'ne': 'मध्य उच्च जर्मन', 'nl': 'Middelhoogduits', 'nn': 'mellomhøgtysk', 'no': 'mellomhøytysk', 'or': 'ମିଡିଲ୍ ହାଇ ଜର୍ମାନ୍', 'pl': 'średnio-wysoko-niemiecki', 'pt': 'alto alemão médio', 'rm': 'tudestg mesaun', 'ro': 'germană înaltă medie', 'ru': 'средневерхненемецкий', 'sk': 'stredná horná nemčina', 'sl': 'srednja visoka nemščina', 'sr': 'средњи високонемачки', 'sr-Cyrl': 'средњи високонемачки', 'sr-Latn': 'srednji visokonemački', 'sv': 'medelhögtyska', 'ta': 'மிடில் ஹை ஜெர்மன்', 'te': 'మధ్యమ హై జర్మన్', 'th': 'เยอรมันสูงกลาง', 'to': 'lea fakasiamane-hake-lotoloto', 'tr': 'Ortaçağ Yüksek Almancası', 'ug': 'ئوتتۇرا ئەسىر ئېگىزلىك گېرمانچە', 'uk': 'середньоверхньонімецька', 'vi': 'Tiếng Thượng Giéc-man Trung cổ', 'yi': 'מיטל הויכדויטש', 'yrl': 'aremãu iwaté médiu', 'yue': '中古高地德文', 'yue-Hans': '中古高地德文', 'yue-Hant': '中古高地德文', 'zh': '中古高地德语', 'zh-Hans': '中古高地德语', 'zh-Hant': '中古高地德文'},
    'gml': {'en': 'Middle Low German'},
    'gmm': {'en': 'Gbaya-Mbodomo'},
    'gmn': {'en': 'Gimnime'},
    'gmq': {'en': 'North Germanic languages'},
    'gmq-gut': {'en': 'Gutnish'},
    'gmq-mno': {'en': 'Middle Norwegian'},
    'gmq-oda': {'en': 'Old Danish'},
    'gmq-osw': {'en': 'Old Swedish'},
    'gmq-pro': {'en': 'Proto-Norse'},
    'gmr': {'en': 'Mirning'},
    'gmu': {'en': 'Gumalu'},
    'gmv': {'en': 'Gamo'},
    'gmw': {'en': 'West Germanic languages'},
    'gmw-cfr': {'en': 'Central Franconian'},
    'gmw-gts': {'en': 'Gottscheerish'},
    'gmw-jdt': {'en': 'Jersey Dutch'},
    'gmw-rfr': {'en': 'Rhine Franconian'},
    'gmw-stm': {'en': 'Sathmar Swabian'},
    'gmw-tsx': {'en': 'Transylvanian Saxon'},
    'gmw-vog': {'en': 'Volga German'},
    'gmw-zps': {'en': 'Zipser German'},
    'gmx': {'en': 'Magoma'},
    'gmy': {'en': 'Mycenaean Greek'},
    'gmz': {'en': 'Mgbolizhia'},
    'gn': {'af': 'Guarani', 'am': 'ጓራኒኛ', 'ar': 'الغوارانية', 'as': 'গুৱাৰাণী', 'ast': 'guaraní', 'az': 'quarani', 'az-Cyrl': 'гуарани', 'az-Latn': 'quarani', 'bal-Latn': 'Guárián', 'be': 'гуарані', 'be-tarask': 'гуарані', 'bg': 'гуарани', 'bgn': 'گوارانی', 'bn': 'গুয়ারানি', 'br': 'guarani', 'brx': 'गुआरानी', 'bs': 'gvarani', 'bs-Cyrl': 'гварани', 'bs-Latn': 'gvarani', 'ca': 'guaraní', 'ca-ES-valencia': 'guaraní', 'ccp': '𑄉𑄪𑄠𑄢𑄚𑄨', 'ce': 'гуарани', 'chr': 'ᏆᎳᏂ', 'ckb': 'گووارانی', 'cs': 'guaranština', 'cy': 'Guaraní', 'da': 'guarani', 'de': 'Guaraní', 'dsb': 'guarani', 'dz': 'གུ་ཝ་ར་ནི་ཁ', 'ee': 'guarangbe', 'el': 'Γκουαρανί', 'el-polyton': 'Γκουαρανί', 'en': 'Guarani', 'eo': 'gvarania', 'es': 'guaraní', 'et': 'guaranii', 'eu': 'guaraniera', 'fa': 'گوارانی', 'ff-Adlm': '𞤘𞤵𞤢𞤪𞤢𞤲𞤭𞥅𞤪𞤫', 'fi': 'guarani', 'fil': 'Guarani', 'fo': 'guarani', 'fr': 'guarani', 'fur': 'guaranì', 'fy': 'Guaraní', 'ga': 'Guaráinis', 'gd': 'Guaraní', 'gl': 'guaraní', 'gn': 'avañe’ẽ', 'gsw': 'Guarani', 'gu': 'ગુઆરાની', 'ha': 'Guwaraniyanci', 'he': 'גוארני', 'hi': 'गुआरानी', 'hi-Latn': 'Guarani', 'hr': 'gvaranski', 'hsb': 'guarani', 'ht': 'guarani', 'hu': 'guarani', 'hy': 'գուարանի', 'ia': 'guarani', 'id': 'Guarani', 'ig': 'Guarani', 'is': 'gvaraní', 'it': 'guaraní', 'ja': 'グアラニー語', 'jv': 'Guarani', 'ka': 'გუარანი', 'kea': 'guarani', 'kgp': 'góranĩ', 'kk': 'гуарани тілі', 'kk-Cyrl': 'гуарани тілі', 'km': 'ហ្គូរ៉ានី', 'kn': 'ಗೌರಾನಿ', 'ko': '과라니어', 'kok': 'ग्वारानी', 'kok-Deva': 'ग्वारानी', 'ks': 'گُوارَنی', 'ks-Arab': 'گُوارَنی', 'ksh': 'Juwarahnesch', 'ku': 'guwaranî', 'ky': 'гуараниче', 'lb': 'Guarani', 'lkt': 'Guarani Iyápi', 'lld': 'guaraní', 'lo': 'ກົວຣານີ', 'lrc': 'گوٙآرانی', 'lt': 'gvaranių', 'lv': 'gvaranu', 'mai': 'गुआरानी', 'mi': 'Kuaranī', 'mk': 'гварански', 'ml': 'ഗ്വരനി', 'mn': 'гуарани', 'mr': 'गुआरनी', 'ms': 'Guarani', 'mt': 'Guarani', 'my': 'ဂူအာရာနီ', 'mzn': 'گورانی', 'nb': 'guarani', 'ne': 'गुवारानी', 'nl': 'Guaraní', 'nn': 'guarani', 'no': 'guarani', 'om': 'Afaan Guarani', 'or': 'ଗୁଆରାନୀ', 'pa': 'ਗੁਆਰਾਨੀ', 'pa-Guru': 'ਗੁਆਰਾਨੀ', 'pcm': 'Guáráni Lángwej', 'pl': 'guarani', 'ps': 'ګوراني', 'pt': 'guarani', 'qu': 'Guaraní Simi', 'rm': 'guarani', 'ro': 'guarani', 'ru': 'гуарани', 'rw': 'Inyaguwarani', 'sc': 'guaranì', 'sd': 'گواراني', 'sd-Arab': 'گواراني', 'si': 'ගුවාරනි', 'sk': 'guaraníjčina', 'sl': 'gvaranijščina', 'smn': 'guaranikielâ', 'so': 'Guarani', 'sq': 'guaranisht', 'sr': 'гварани', 'sr-Cyrl': 'гварани', 'sr-Latn': 'gvarani', 'sv': 'guaraní', 'sw': 'Kiguarani', 'ta': 'க்வாரனி', 'te': 'గ్వారనీ', 'tg': 'гуаранӣ', 'th': 'กัวรานี', 'ti': 'ጓራኒ', 'tk': 'guarani dili', 'to': 'lea fakakualani', 'tr': 'Guarani dili', 'tt': 'гуарани', 'ug': 'گۇئارانىچە', 'uk': 'гуарані', 'ur': 'گُارانی', 'uz': 'guarani', 'uz-Cyrl': 'гуарани', 'uz-Latn': 'guarani', 'vec': 'guaranì', 'vi': 'Tiếng Guarani', 'wae': 'Guarani', 'wo': 'Garani', 'xnr': 'गुआरानी', 'yo': 'Èdè Guarani', 'yrl': 'guwarani', 'yue': '瓜拉尼文', 'yue-Hans': '瓜拉尼文', 'yue-Hant': '瓜拉尼文', 'zh': '瓜拉尼语', 'zh-Hans': '瓜拉尼语', 'zh-Hant': '瓜拉尼文', 'zu': 'isi-Guarani'},
    'gna': {'en': 'Kaansa'},
    'gnb': {'en': 'Gangte'},
    'gnc': {'en': 'Guanche'},
    'gnd': {'en': 'Zulgo-Gemzek'},
    'gne': {'en': 'Ganang'},
    'gng': {'en': 'Ngangam'},
    'gnh': {'en': 'Lere'},
    'gni': {'en': 'Gooniyandi'},
    'gnj': {'en': 'Ngen'},
    'gnk': {'en': 'ǁGana'},
    'gnl': {'en': 'Gangulu'},
    'gnm': {'en': 'Ginuman'},
    'gnn': {'en': 'Gumatj'},
    'gno': {'en': 'Northern Gondi'},
    'gnq': {'en': 'Gana'},
    'gnr': {'en': 'Gureng Gureng'},
    'gnt': {'en': 'Guntai'},
    'gnu': {'en': 'Gnau'},
    'gnw': {'en': 'Western Bolivian Guaraní'},
    'gnz': {'en': 'Ganzi'},
    'goa': {'en': 'Guro'},
    'gob': {'en': 'Playero'},
    'goc': {'en': 'Gorakor'},
    'god': {'en': 'Godié'},
    'goe': {'en': 'Gongduk'},
    'gof': {'en': 'Gofa'},
    'gog': {'en': 'Gogo'},
    'goh': {'ar': 'الألمانية العليا القديمة', 'ast': 'altualemán antiguu', 'az': 'qədim alman', 'az-Latn': 'qədim alman', 'bg': 'старовисоконемски', 'bn': 'প্রাচীন উচ্চ জার্মানি', 'br': 'henalamaneg uhel', 'brx': 'पुरानी उच्चस्तरी जर्मन', 'bs': 'staronjemački', 'bs-Cyrl': 'старонемачки', 'bs-Latn': 'staronjemački', 'ca': 'alt alemany antic', 'ca-ES-valencia': 'alt alemany antic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄴 𑄅𑄪𑄉𑄪𑄢𑄬 𑄎𑄢𑄴𑄟𑄚𑄩', 'cs': 'hornoněmčina (stará)', 'cy': 'Hen Almaeneg Uchel', 'da': 'oldhøjtysk', 'de': 'Althochdeutsch', 'el': 'Παλαιά Άνω Γερμανικά', 'el-polyton': 'Παλαιά Άνω Γερμανικά', 'en': 'Old High German', 'es': 'alto alemán antiguo', 'et': 'vanaülemsaksa', 'fa': 'آلمانی علیای باستان', 'fi': 'muinaisyläsaksa', 'fr': 'ancien haut allemand', 'fy': 'Alsheechdútsk', 'ga': 'Sean-Ard-Ghearmáinis', 'gd': 'Seann-Àrd-Gearmailtis', 'gsw': 'Althochtüütsch', 'gu': 'જૂની હાઇ જર્મન', 'he': 'גרמנית עתיקה גבוהה', 'hi': 'पुरातन हाइ जर्मन', 'hi-Latn': 'Old High German', 'hr': 'starovisokonjemački', 'ht': 'ancien haut allemand', 'hu': 'ófelső német', 'hy': 'հին վերին գերմաներեն', 'id': 'Jerman Kuno', 'is': 'fornháþýska', 'it': 'tedesco antico alto', 'ja': '古高ドイツ語', 'ka': 'ძველი ზემოგერმანული', 'kgp': 'arimỹv-téj arkajku', 'kn': 'ಪ್ರಾಚೀನ ಹೈ ಜರ್ಮನ್', 'ko': '고대 고지 독일어', 'ks': 'پرون ہاے جٔرمَن', 'ks-Arab': 'پرون ہاے جٔرمَن', 'lb': 'Alhéichdäitsch', 'lo': 'ເຢຍລະມັນສູງໂບຮານ', 'lt': 'senoji Aukštosios Vokietijos', 'lv': 'senaugšvācu', 'mk': 'старогорногермански', 'ml': 'ഓൾഡ് ഹൈ ജർമൻ', 'mr': 'पुरातन हाइ जर्मन', 'mt': 'Ġermaniż Antik, Pulit', 'nb': 'gammelhøytysk', 'ne': 'पुरातन उच्च जर्मन', 'nl': 'Oudhoogduits', 'nn': 'gammalhøgtysk', 'no': 'gammelhøytysk', 'or': 'ପୁରୁଣା ହାଇ ଜର୍ମାନ୍', 'pl': 'staro-wysoko-niemiecki', 'pt': 'alemão arcaico alto', 'rm': 'vegl tudestg da scrittira', 'ro': 'germană înaltă veche', 'ru': 'древневерхненемецкий', 'sk': 'stará horná nemčina', 'sl': 'stara visoka nemščina', 'sr': 'старонемачки', 'sr-Cyrl': 'старонемачки', 'sr-Latn': 'staronemački', 'sv': 'fornhögtyska', 'ta': 'பழைய ஹை ஜெர்மன்', 'te': 'ప్రాచీన హై జర్మన్', 'th': 'เยอรมันสูงโบราณ', 'to': 'lea fakasiamane-hake-motuʻa', 'tr': 'Eski Yüksek Almanca', 'ug': 'قەدىمكى ئېگىزلىك گېرمانچە', 'uk': 'давньоверхньонімецька', 'vi': 'Tiếng Thượng Giéc-man cổ', 'yi': 'אַלט־ הויכדויטש', 'yrl': 'aremãu arkaiku iwaté', 'yue': '古高地日耳曼文', 'yue-Hans': '古高地日耳曼文', 'yue-Hant': '古高地日耳曼文', 'zh': '古高地德语', 'zh-Hans': '古高地德语', 'zh-Hant': '古高地德文'},
    'goi': {'en': 'Gobasi'},
    'goj': {'en': 'Gowlan'},
    'gok': {'en': 'Gowli'},
    'gol': {'en': 'Gola'},
    'gom': {'en': 'Goan Konkani'},
    'gon': {'ar': 'الجندي', 'ast': 'gondi', 'az': 'qondi', 'az-Latn': 'qondi', 'bg': 'гонди', 'bn': 'গোন্ডি', 'brx': 'गाँडी', 'bs': 'gondi', 'bs-Cyrl': 'гонди', 'bs-Latn': 'gondi', 'ca': 'gondi', 'ca-ES-valencia': 'gondi', 'ccp': '𑄉𑄮𑄚𑄴𑄓𑄨', 'cs': 'góndština', 'da': 'gondi', 'de': 'Gondi', 'el': 'Γκόντι', 'el-polyton': 'Γκόντι', 'en': 'Gondi', 'es': 'gondi', 'et': 'gondi', 'fa': 'گوندی', 'fi': 'gondi', 'fr': 'gondi', 'fy': 'Gondi', 'gd': 'Gondi', 'gsw': 'Gondi', 'gu': 'ગોંડી', 'he': 'גונדי', 'hi': 'गाँडी', 'hi-Latn': 'Gondi', 'hr': 'gondi', 'ht': 'gondi', 'hu': 'gondi', 'id': 'Gondi', 'is': 'gondí', 'it': 'gondi', 'ja': 'ゴーンディー語', 'ka': 'გონდი', 'kgp': 'gãnni', 'kn': 'ಗೊಂಡಿ', 'ko': '곤디어', 'ks': 'گوندی', 'ks-Arab': 'گوندی', 'lb': 'Gondi-Sprooch', 'lo': 'ກອນດີ', 'lt': 'gondi', 'lv': 'gondu valodas', 'mk': 'гонди', 'ml': 'ഗോണ്ഡി', 'mr': 'गाँडी', 'mt': 'Gondi', 'nb': 'gondi', 'ne': 'गोन्डी', 'nl': 'Gondi', 'nn': 'gondi', 'no': 'gondi', 'or': 'ଗୋଣ୍ଡି', 'pl': 'gondi', 'pt': 'gondi', 'rm': 'gondi', 'ro': 'gondi', 'ru': 'гонди', 'sk': 'góndčina', 'sl': 'gondi', 'sr': 'гонди', 'sr-Cyrl': 'гонди', 'sr-Latn': 'gondi', 'sv': 'gondi', 'ta': 'கோன்டி', 'te': 'గోండి', 'th': 'กอนดิ', 'to': 'lea fakakonitī', 'tr': 'Gondi dili', 'ug': 'گوندىچە', 'uk': 'гонді', 'vi': 'Tiếng Gondi', 'yrl': 'gũdi', 'yue': '岡德文', 'yue-Hans': '冈德文', 'yue-Hant': '岡德文', 'zh': '冈德语', 'zh-Hans': '冈德语', 'zh-Hant': '岡德文'},
    'goo': {'en': 'Gone Dau'},
    'gop': {'en': 'Yeretuar'},
    'goq': {'en': 'Gorap'},
    'gor': {'af': 'Gorontalo', 'am': 'ጎሮንታሎ', 'ar': 'الجورونتالو', 'as': 'গোৰোন্তালো', 'ast': 'gorontalo', 'az': 'qorontalo', 'az-Cyrl': 'горонтало', 'az-Latn': 'qorontalo', 'be': 'гарантала', 'be-tarask': 'гарантала', 'bg': 'горонтало', 'bgn': 'گورونتالو', 'bn': 'গোরোন্তালো', 'br': 'gorontalo', 'brx': 'गर’न्टाल’', 'bs': 'gorontalo', 'bs-Cyrl': 'горонтало', 'bs-Latn': 'gorontalo', 'ca': 'gorontalo', 'ca-ES-valencia': 'gorontalo', 'ccp': '𑄉𑄢𑄮𑄚𑄴𑄖𑄣𑄮', 'ce': 'горонтало', 'chr': 'ᎪᎶᏂᏔᏃ', 'ckb': 'گۆرۆنتالی', 'cs': 'gorontalo', 'cy': 'Gorontalo', 'da': 'gorontalo', 'de': 'Mongondou', 'dsb': 'gorontalošćina', 'el': 'Γκοροντάλο', 'el-polyton': 'Γκοροντάλο', 'en': 'Gorontalo', 'es': 'gorontalo', 'et': 'gorontalo', 'eu': 'gorontaloera', 'fa': 'گورونتالو', 'ff-Adlm': '𞤘𞤮𞤪𞤮𞤲𞤼𞤢𞤤𞤮𞥅𞤪𞤫', 'fi': 'gorontalo', 'fil': 'Gorontalo', 'fo': 'gorontalo', 'fr': 'gorontalo', 'fy': 'Gorontalo', 'ga': 'Gorantalais', 'gd': 'Gorontalo', 'gl': 'gorontalo', 'gsw': 'Mongondou', 'gu': 'ગોરોન્તાલો', 'ha': 'Gorontalo', 'he': 'גורונטאלו', 'hi': 'गोरोन्तालो', 'hi-Latn': 'Gorontalo', 'hr': 'gorontalo', 'hsb': 'gorontalošćina', 'ht': 'gorontalo', 'hu': 'gorontalo', 'hy': 'գորոնտալո', 'ia': 'gorontalo', 'id': 'Gorontalo', 'ig': 'Gorontalo', 'is': 'gorontaló', 'it': 'gorontalo', 'ja': 'ゴロンタロ語', 'jv': 'Gorontalo', 'ka': 'გორონტალო', 'kgp': 'gorãntar', 'kk': 'горонтало тілі', 'kk-Cyrl': 'горонтало тілі', 'km': 'ហ្គូរុនតាឡូ', 'kn': 'ಗೊರೊಂಟಾಲೋ', 'ko': '고론탈로어', 'kok': 'गोरोंटालो', 'kok-Deva': 'गोरोंटालो', 'ks': 'گورینٹیلو', 'ks-Arab': 'گورینٹیلو', 'ksh': 'Jorontalohsch', 'ku': 'gorontaloyî', 'ky': 'горонталочо', 'lb': 'Mongondou', 'lo': 'ໂກຣອນຕາໂຣ', 'lt': 'gorontalo', 'lv': 'gorontalu', 'mai': 'गोरोंटेलो', 'mi': 'Korōtaro', 'mk': 'горонтало', 'ml': 'ഗൊറോന്റാലോ', 'mn': 'горонтало', 'mr': 'गोरोन्तालो', 'ms': 'Gorontalo', 'mt': 'Gorontalo', 'my': 'ဂိုရိုတာလို', 'nb': 'gorontalo', 'ne': 'गोरोन्टालो', 'nl': 'Gorontalo', 'nn': 'gorontalo', 'no': 'gorontalo', 'or': 'ଗୋରୋଣ୍ଟାଲୋ', 'pa': 'ਗੋਰੋਂਤਾਲੋ', 'pa-Guru': 'ਗੋਰੋਂਤਾਲੋ', 'pcm': 'Gorontáló Lángwej', 'pl': 'gorontalo', 'ps': 'ګورن ټالو', 'pt': 'gorontalo', 'qu': 'Gorontalo Simi', 'rm': 'gorontalo', 'ro': 'gorontalo', 'ru': 'горонтало', 'sc': 'gorontalo', 'sd': 'گورنٽلو', 'sd-Arab': 'گورنٽلو', 'si': 'ගොරොන්ටාලො', 'sk': 'gorontalo', 'sl': 'gorontalščina', 'smn': 'gorontalokielâ', 'so': 'Gorontalo', 'sq': 'gorontaloisht', 'sr': 'горонтало', 'sr-Cyrl': 'горонтало', 'sr-Latn': 'gorontalo', 'sv': 'gorontalo', 'sw': 'Kigorontalo', 'ta': 'கோரோன்டலோ', 'te': 'గోరోంటలా', 'th': 'กอรอนทาโล', 'ti': 'ጎሮንታሎ', 'tk': 'gorontalo dili', 'to': 'lea fakakolonitalo', 'tr': 'Gorontalo dili', 'ug': 'گورونتالوچە', 'uk': 'горонтало', 'ur': 'گورانٹالو', 'uz': 'gorontalo', 'uz-Cyrl': 'горонтало', 'uz-Latn': 'gorontalo', 'vec': 'gorontalo', 'vi': 'Tiếng Gorontalo', 'xnr': 'गोरोन्तालो', 'yo': 'Èdè Gorontalo', 'yrl': 'gurũtaru', 'yue': '科隆達羅文', 'yue-Hans': '科隆达罗文', 'yue-Hant': '科隆達羅文', 'zh': '哥伦打洛语', 'zh-Hans': '哥伦打洛语', 'zh-Hant': '科隆達羅文', 'zu': 'isi-Gorontalo'},
    'gos': {'en': 'Gronings'},
    'got': {'af': 'Goties', 'ar': 'القوطية', 'ast': 'góticu', 'az': 'qotika', 'az-Latn': 'qotika', 'bg': 'готически', 'bn': 'গথিক', 'br': 'goteg', 'brx': 'गॉथिक', 'bs': 'gotski', 'bs-Cyrl': 'готски', 'bs-Latn': 'gotski', 'ca': 'gòtic', 'ca-ES-valencia': 'gòtic', 'ccp': '𑄉𑄧𑄗𑄨𑄇𑄴', 'cs': 'gótština', 'cy': 'Gotheg', 'da': 'gotisk', 'de': 'Gotisch', 'dsb': 'gotišćina', 'el': 'Γοτθικά', 'el-polyton': 'Γοτθικά', 'en': 'Gothic', 'en-Dsrt': '𐐘𐐱𐑃𐐮𐐿', 'es': 'gótico', 'et': 'gooti', 'fa': 'گوتی', 'fi': 'gootti', 'fr': 'gotique', 'fur': 'gotic', 'fy': 'Gothysk', 'gd': 'Gotais', 'gsw': 'Gotisch', 'gu': 'ગોથિક', 'he': 'גותית', 'hi': 'गॉथिक', 'hi-Latn': 'Gothic', 'hr': 'gotski', 'hsb': 'gotšćina', 'ht': 'gotique', 'hu': 'gót', 'hy': 'գոթերեն', 'id': 'Gotik', 'is': 'gotneska', 'it': 'gotico', 'ja': 'ゴート語', 'ka': 'გოთური', 'kgp': 'gótiko', 'kn': 'ಗೋಥಿಕ್', 'ko': '고트어', 'ks': 'گوتھِک', 'ks-Arab': 'گوتھِک', 'lb': 'Gotesch', 'lo': 'ກອດຮິກ', 'lt': 'gotų', 'lv': 'gotu', 'mk': 'готски', 'ml': 'ഗോഥിക്ക്', 'mr': 'गॉथिक', 'mt': 'Gotiku', 'nb': 'gotisk', 'ne': 'गोथिक', 'nl': 'Gothisch', 'nn': 'gotisk', 'no': 'gotisk', 'or': 'ଗୋଥିକ୍', 'pl': 'gocki', 'pt': 'gótico', 'rm': 'gotic', 'ro': 'gotică', 'ru': 'готский', 'sk': 'gótčina', 'sl': 'gotščina', 'sr': 'готски', 'sr-Cyrl': 'готски', 'sr-Latn': 'gotski', 'sv': 'gotiska', 'ta': 'கோதிக்', 'te': 'గోథిక్', 'th': 'โกธิก', 'to': 'lea fakakotika', 'tr': 'Gotça', 'ug': 'گوتچە', 'uk': 'готська', 'vi': 'Tiếng Gô-tích', 'yi': 'גאטיש', 'yrl': 'gútiku', 'yue': '哥德文', 'yue-Hans': '哥德文', 'yue-Hant': '哥德文', 'zh': '哥特语', 'zh-Hans': '哥特语', 'zh-Hant': '哥德文'},
    'gou': {'en': 'Gavar'},
    'gow': {'en': 'Gorowa'},
    'gox': {'en': 'Gobu'},
    'goy': {'en': 'Goundo'},
    'goz': {'en': 'Gozarkhani'},
    'gpa': {'en': 'Gupa-Abawa'},
    'gpe': {'en': 'Ghanaian Pidgin English'},
    'gpn': {'en': 'Taiap'},
    'gqa': {'en': "Ga'anda"},
    'gqi': {'en': 'Guiqiong'},
    'gqn': {'en': 'Guana (Brazil)'},
    'gqr': {'en': 'Gor'},
    'gqu': {'en': 'Qau'},
    'gra': {'en': 'Rajput Garasia'},
    'grb': {'ar': 'الجريبو', 'ast': 'grebo', 'az': 'qrebo', 'az-Latn': 'qrebo', 'bg': 'гребо', 'bn': 'গ্রেবো', 'br': 'grebo', 'brx': 'ग्रेबो', 'bs': 'grebo', 'bs-Cyrl': 'гребо', 'bs-Latn': 'grebo', 'ca': 'grebo', 'ca-ES-valencia': 'grebo', 'ccp': '𑄉𑄳𑄢𑄬𑄝𑄮', 'cs': 'grebo', 'da': 'grebo', 'de': 'Grebo', 'el': 'Γκρίμπο', 'el-polyton': 'Γκρίμπο', 'en': 'Grebo', 'es': 'grebo', 'et': 'grebo', 'fa': 'گریبویی', 'fi': 'grebo', 'fr': 'grebo', 'fy': 'Grebo', 'gd': 'Grebo', 'gsw': 'Grebo', 'gu': 'ગ્રેબો', 'he': 'גרבו', 'hi': 'ग्रेबो', 'hi-Latn': 'Grebo', 'hr': 'grebo', 'ht': 'grebo', 'hu': 'grebó', 'id': 'Grebo', 'is': 'gerbó', 'it': 'grebo', 'ja': 'グレボ語', 'kgp': 'gremo', 'kn': 'ಗ್ರೇಬೋ', 'ko': '게르보어', 'ks': 'گرِبو', 'ks-Arab': 'گرِبو', 'lb': 'Grebo-Sprooch', 'lo': 'ກຣີໂບ', 'lt': 'grebo', 'lv': 'grebo', 'mk': 'гребо', 'ml': 'ഗ്രബൊ', 'mr': 'ग्रेबो', 'mt': 'Grebo', 'nb': 'grebo', 'ne': 'ग्रेबो', 'nl': 'Grebo', 'nn': 'grebo', 'no': 'grebo', 'or': 'ଗ୍ରେବୋ', 'pl': 'grebo', 'pt': 'grebo', 'rm': 'grebo', 'ro': 'grebo', 'ru': 'гребо', 'sk': 'grebo', 'sl': 'grebščina', 'sr': 'гребо', 'sr-Cyrl': 'гребо', 'sr-Latn': 'grebo', 'sv': 'grebo', 'ta': 'க்ரேபோ', 'te': 'గ్రేబో', 'th': 'เกรโบ', 'to': 'lea fakakēpo', 'tr': 'Grebo dili', 'ug': 'گرېبوچە', 'uk': 'гребо', 'vi': 'Tiếng Grebo', 'yrl': 'gerebu', 'yue': '格列博文', 'yue-Hans': '格列博文', 'yue-Hant': '格列博文', 'zh': '格列博语', 'zh-Hans': '格列博语', 'zh-Hant': '格列博文'},
    'grc': {'af': 'Antieke Grieks', 'am': 'የጥንታዊ ግሪክ', 'ar': 'اليونانية القديمة', 'ast': 'griegu antiguu', 'az': 'qədim yunan', 'az-Latn': 'qədim yunan', 'be': 'старажытнагрэчаская', 'be-tarask': 'старажытнагрэчаская', 'bg': 'древногръцки', 'bn': 'প্রাচীন গ্রীক', 'br': 'hencʼhresianeg', 'brx': 'प्राचीन यूनानी', 'bs': 'starogrčki', 'bs-Cyrl': 'старогрчки', 'bs-Latn': 'starogrčki', 'ca': 'grec antic', 'ca-ES-valencia': 'grec antic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄴 𑄉𑄳𑄢𑄩𑄇𑄴', 'cs': 'starořečtina', 'cy': 'Hen Roeg', 'da': 'oldgræsk', 'de': 'Altgriechisch', 'el': 'Αρχαία Ελληνικά', 'el-polyton': 'Ἀρχαῖα Ἑλληνικά', 'en': 'Ancient Greek', 'en-Dsrt': '𐐁𐑌𐐽𐐲𐑌𐐻 𐐘𐑉𐐨𐐿', 'es': 'griego antiguo', 'et': 'vanakreeka', 'fa': 'یونانی کهن', 'fi': 'muinaiskreikka', 'fr': 'grec ancien', 'fur': 'vieri grêc', 'fy': 'Aldgryks', 'ga': 'Sean-Ghréigis', 'gd': 'Greugais Àrsaidh', 'gl': 'grego antigo', 'gsw': 'Altgriechisch', 'gu': 'પ્રાચીન ગ્રીક', 'he': 'יוונית עתיקה', 'hi': 'प्राचीन यूनानी', 'hi-Latn': 'Ancient Greek', 'hr': 'starogrčki', 'ht': 'grec ancien', 'hu': 'ógörög', 'hy': 'հին հունարեն', 'id': 'Yunani Kuno', 'is': 'forngríska', 'it': 'greco antico', 'ja': '古代ギリシャ語', 'ka': 'ძველი ბერძნული', 'kgp': 'gregu arkajku', 'kn': 'ಪ್ರಾಚೀನ ಗ್ರೀಕ್', 'ko': '고대 그리스어', 'ks': 'قدیٖم یوٗنٲنی', 'ks-Arab': 'قدیٖم یوٗنٲنی', 'lb': 'Algriichesch', 'lo': 'ແອນຊຽນກຣີກ', 'lt': 'senovės graikų', 'lv': 'sengrieķu', 'mk': 'старогрчки', 'ml': 'പുരാതന ഗ്രീക്ക്', 'mr': 'प्राचीन ग्रीक', 'ms': 'Greek Purba', 'mt': 'Grieg, Antik', 'my': 'ရှေးဟောင်း ဂရိ', 'nb': 'gammelgresk', 'ne': 'पुरातन ग्रिक', 'nl': 'Oudgrieks', 'nn': 'gammalgresk', 'no': 'gammelgresk', 'or': 'ପ୍ରାଚୀନ୍ ୟୁନାନୀ', 'os': 'рагон бердзейнаг', 'pa': 'ਪੁਰਾਤਨ ਯੂਨਾਨੀ', 'pa-Guru': 'ਪੁਰਾਤਨ ਯੂਨਾਨੀ', 'pl': 'starogrecki', 'pt': 'grego arcaico', 'rm': 'grec vegl', 'ro': 'greacă veche', 'ru': 'древнегреческий', 'sk': 'starogréčtina', 'sl': 'stara grščina', 'smn': 'toovláš kreikakielâ', 'sr': 'старогрчки', 'sr-Cyrl': 'старогрчки', 'sr-Latn': 'starogrčki', 'sv': 'forngrekiska', 'sw': 'Kiyunani', 'ta': 'பண்டைய கிரேக்கம்', 'te': 'ప్రాచీన గ్రీక్', 'th': 'กรีกโบราณ', 'to': 'lea fakakalisimuʻa', 'tr': 'Antik Yunanca', 'ug': 'قەدىمكى گىرېكچە', 'uk': 'давньогрецька', 'vi': 'Tiếng Hy Lạp cổ', 'yi': 'אוראַלט־גריכיש', 'yrl': 'geregu arkaiku', 'yue': '古希臘文', 'yue-Hans': '古希腊文', 'yue-Hant': '古希臘文', 'zh': '古希腊语', 'zh-Hans': '古希腊语', 'zh-Hant': '古希臘文'},
    'grd': {'en': 'Guruntum-Mbaaru'},
    'grg': {'en': 'Madi'},
    'grh': {'en': 'Gbiri-Niragu'},
    'gri': {'en': 'Ghari'},
    'grj': {'en': 'Southern Grebo'},
    'grk': {'en': 'Greek languages'},
    'grk-cal': {'en': 'Calabrian Greek'},
    'grk-pro': {'en': 'Proto-Hellenic'},
    'grm': {'en': 'Kota Marudu Talantang'},
    'gro': {'en': 'Groma'},
    'grq': {'en': 'Gorovu'},
    'grr': {'en': 'Taznatit'},
    'grs': {'en': 'Gresi'},
    'grt': {'en': 'Garo'},
    'gru': {'en': 'Kistane'},
    'grv': {'en': 'Central Grebo'},
    'grw': {'en': 'Gweda'},
    'grx': {'en': 'Guriaso'},
    'gry': {'en': 'Barclayville Grebo'},
    'grz': {'en': 'Guramalum'},
    'gse': {'en': 'Ghanaian Sign Language'},
    'gsg': {'en': 'German Sign Language'},
    'gsl': {'en': 'Gusilay'},
    'gsm': {'en': 'Guatemalan Sign Language'},
    'gsn': {'en': 'Nema'},
    'gso': {'en': 'Southwest Gbaya'},
    'gsp': {'en': 'Wasembo'},
    'gss': {'en': 'Greek Sign Language'},
    'gsw': {'af': 'Switserse Duits', 'am': 'ስዊዝ ጀርመንኛ', 'ar': 'الألمانية السويسرية', 'as': 'ছুইচ জাৰ্মান', 'ast': 'alemán de Suiza', 'az': 'İsveçrə almancası', 'az-Arab': 'سویس آلمانجاسی', 'az-Cyrl': 'Исвечрә алманҹасы', 'az-Latn': 'İsveçrə almancası', 'bal-Latn': 'Swiz Jarman', 'be': 'швейцарская нямецкая', 'be-tarask': 'швейцарская нямецкая', 'bg': 'швейцарски немски', 'bgn': 'جرمنین سوئیسی', 'bn': 'সুইস জার্মান', 'br': 'alamaneg Suis', 'brx': 'स्वीस जार्मान', 'bs': 'njemački (Švicarska)', 'bs-Cyrl': 'њемачки (Швицарска)', 'bs-Latn': 'njemački (Švicarska)', 'ca': 'alemany suís', 'ca-ES-valencia': 'alemany suís', 'ccp': '𑄥𑄪𑄃𑄨𑄌𑄴 𑄥𑄢𑄴𑄟𑄚𑄴', 'ce': 'швейцарин немцойн', 'chr': 'ᏍᏫᏏ ᎠᏂᏓᏥ', 'ckb': 'ئەڵمانیی سویسڕا', 'cs': 'němčina (Švýcarsko)', 'cy': 'Almaeneg y Swistir', 'da': 'schweizertysk', 'de': 'Schweizerdeutsch', 'dsb': 'šwicarska nimšćina', 'dz': 'སུ་ཡིས་ཇཱར་མཱན་ཁ', 'ee': 'swizerlanɖtɔwo ƒe germaniagbe', 'el': 'Γερμανικά Ελβετίας', 'el-polyton': 'Γερμανικά Ελβετίας', 'en': 'Swiss German', 'es': 'alemán suizo', 'et': 'šveitsisaksa', 'eu': 'Suitzako alemana', 'fa': 'آلمانی سوئیسی', 'ff-Adlm': '𞤔𞤫𞤪𞤥𞤢𞤲𞤪𞤫 𞤅𞤵𞤱𞤭𞤧', 'fi': 'sveitsinsaksa', 'fil': 'Swiss German', 'fo': 'týskt (Sveis)', 'fr': 'suisse allemand', 'fy': 'Switsers Dútsk', 'ga': 'Gearmáinis Eilvéiseach', 'gd': 'Gearmailtis Eilbheiseach', 'gl': 'alemán suízo', 'gsw': 'Schwiizertüütsch', 'gu': 'સ્વિસ જર્મન', 'ha': 'Jamusanci Swiss', 'haw': 'Kuikilani Kelemānia', 'he': 'גרמנית שוויצרית', 'hi': 'स्विस जर्मन', 'hi-Latn': 'Swiss German', 'hr': 'švicarski njemački', 'hsb': 'šwicarska němčina', 'ht': 'suisse allemand', 'hu': 'svájci német', 'hy': 'շվեյցարական գերմաներեն', 'ia': 'germano suisse', 'id': 'Jerman (Swiss)', 'ig': 'German Swiss', 'is': 'svissnesk þýska', 'it': 'tedesco svizzero', 'ja': 'スイスドイツ語', 'jv': 'Jerman Swiss', 'ka': 'შვეიცარიული გერმანული', 'kea': 'alemon suisu', 'kgp': 'arimỹv (Suvisa)', 'kk': 'неміс тілі (Швейцария)', 'kk-Cyrl': 'неміс тілі (Швейцария)', 'km': 'អាល្លឺម៉ង (ស្វីស)', 'kn': 'ಸ್ವಿಸ್ ಜರ್ಮನ್', 'ko': '독일어(스위스)', 'kok': 'स्विस जर्मन', 'kok-Deva': 'स्विस जर्मन', 'ks': 'سٕوِس جٔرمَن', 'ks-Arab': 'سٕوِس جٔرمَن', 'ksh': 'Schwitzerdütsch', 'ku': 'elmanîşî', 'ky': 'немисче (Швейцария)', 'lb': 'Schwäizerdäitsch', 'lo': 'ສະວິສ ເຈີແມນ', 'lrc': 'آلمانی سوٙئیسی', 'lt': 'Šveicarijos vokiečių', 'lv': 'Šveices vācu', 'mai': 'स्वीस जर्मन', 'mi': 'Tiamana Huiterangi', 'mk': 'швајцарски германски', 'ml': 'സ്വിസ് ജർമ്മൻ', 'mn': 'швейцари-герман', 'mr': 'स्विस जर्मन', 'ms': 'Jerman Switzerland', 'mt': 'Ġermaniż tal-Iżvizzera', 'my': 'ဆွစ် ဂျာမန်', 'mzn': 'سوییس آلمانی', 'nb': 'sveitsertysk', 'ne': 'स्वीस जर्मन', 'nl': 'Zwitserduits', 'nn': 'sveitsertysk', 'no': 'sveitsertysk', 'nqo': 'ߛߥߌߛߌ߬ ߊߟߑߡߊ߲ߞߊ߲', 'or': 'ସୁଇସ୍ ଜର୍ମାନ୍', 'pa': 'ਜਰਮਨ (ਸਵਿਸ)', 'pa-Guru': 'ਜਰਮਨ (ਸਵਿਸ)', 'pcm': 'Jámán Swis', 'pl': 'szwajcarski niemiecki', 'ps': 'سویس جرمن', 'pt': 'alemão (Suíça)', 'qu': 'Alsaciano Simi', 'rm': 'tudestg svizzer', 'ro': 'germană (Elveția)', 'ru': 'швейцарский немецкий', 'sc': 'tedescu isvìtzeru', 'sd': 'سوئس جرمن', 'sd-Arab': 'سوئس جرمن', 'si': 'ස්විස් ජර්මානු', 'sk': 'nemčina (švajčiarska)', 'sl': 'nemščina (Švica)', 'smn': 'Sveitsi saksakielâ', 'so': 'Jarmal Iswiis', 'sq': 'gjermanishte zvicerane', 'sr': 'немачки (Швајцарска)', 'sr-Cyrl': 'немачки (Швајцарска)', 'sr-Latn': 'nemački (Švajcarska)', 'sv': 'schweizertyska', 'sw': 'Kijerumani cha Uswisi', 'ta': 'ஸ்விஸ் ஜெர்மன்', 'te': 'స్విస్ జర్మన్', 'th': 'เยอรมันสวิส', 'ti': 'ስዊዘርላንዳዊ ጀርመን', 'tk': 'nemes dili (Şweýsariýa)', 'to': 'lea fakasiamane-suisilani', 'tr': 'İsviçre Almancası', 'ug': 'گېرمانچە شىۋىتسارىيە', 'uk': 'німецька (Швейцарія)', 'ur': 'سوئس جرمن', 'uz': 'nemis (Shveytsariya)', 'uz-Latn': 'nemis (Shveytsariya)', 'vec': 'todesco zvìsaro', 'vi': 'Tiếng Đức (Thụy Sĩ)', 'xnr': 'स्विस जर्मन', 'yo': 'Súwísì ti Jámánì', 'yrl': 'aremãu (Suisa)', 'yue': '德文（瑞士）', 'yue-Hans': '德文（瑞士）', 'yue-Hant': '德文（瑞士）', 'zh': '瑞士德语', 'zh-Hans': '瑞士德语', 'zh-Hant': '德文（瑞士）', 'zu': 'isi-Swiss German'},
    'gta': {'en': 'Guató'},
    'gti': {'en': 'Gbati-ri'},
    'gtu': {'en': 'Aghu-Tharnggala'},
    'gu': {'af': 'Goedjarati', 'ak': 'Gugyarata', 'am': 'ጉጃርቲኛ', 'ar': 'الغوجاراتية', 'as': 'গুজৰাটী', 'ast': 'guyaratí', 'az': 'qucarat', 'az-Cyrl': 'гуҹарат', 'az-Latn': 'qucarat', 'bal-Latn': 'Gojráti', 'be': 'гуджараці', 'be-tarask': 'гуджараці', 'bg': 'гуджарати', 'bgn': 'گوجراتی', 'bn': 'গুজরাটি', 'br': 'gujarati', 'brx': 'गुजराती', 'bs': 'gudžarati', 'bs-Cyrl': 'гуџарати', 'bs-Latn': 'gudžarati', 'ca': 'gujarati', 'ca-ES-valencia': 'gujarati', 'ccp': '𑄉𑄪𑄎𑄴𑄢𑄖𑄨', 'ce': 'гуджарати', 'chr': 'ᎫᏣᎳᏘ', 'ckb': 'گوجاراتی', 'cs': 'gudžarátština', 'cy': 'Gwjarati', 'da': 'gujarati', 'de': 'Gujarati', 'dsb': 'gudžaratšćina', 'dz': 'གུ་ཇ་ར་ཏི་ཁ', 'ee': 'gujarati', 'el': 'Γκουτζαρατικά', 'el-polyton': 'Γκουτζαρατικά', 'en': 'Gujarati', 'eo': 'guĝarata', 'es': 'guyaratí', 'et': 'gudžarati', 'eu': 'gujaratera', 'fa': 'گجراتی', 'ff-Adlm': '𞤘𞤵𞤶𞤢𞤪𞤢𞤼𞤭𞥅𞤪𞤫', 'fi': 'gudžarati', 'fil': 'Gujarati', 'fo': 'gujarati', 'fr': 'goudjarati', 'fur': 'gujarati', 'fy': 'Gujarati', 'ga': 'Gúisearáitis', 'gd': 'Gujarati', 'gl': 'guxarati', 'gsw': 'Gujarati', 'gu': 'ગુજરાતી', 'ha': 'Gujarati', 'he': 'גוג׳ארטי', 'hi': 'गुजराती', 'hi-Latn': 'Gujarati', 'hr': 'gudžaratski', 'hsb': 'gujarati', 'ht': 'goudjarati', 'hu': 'gudzsaráti', 'hy': 'գուջարաթի', 'ia': 'gujarati', 'id': 'Gujarat', 'ig': 'Gujarati', 'is': 'gújaratí', 'it': 'gujarati', 'ja': 'グジャラート語', 'jv': 'Gujarat', 'ka': 'გუჯარათი', 'kea': 'gujarati', 'kgp': 'gujerati', 'kk': 'гуджарати тілі', 'kk-Cyrl': 'гуджарати тілі', 'km': 'ហ្គុយ៉ារ៉ាទី', 'kn': 'ಗುಜರಾತಿ', 'ko': '구자라트어', 'kok': 'गुजराती', 'kok-Deva': 'गुजराती', 'ks': 'گُجرٲتی', 'ks-Arab': 'گُجرٲتی', 'ksh': 'Gutscharatesch', 'ku': 'gujaratî', 'kxv': 'gujraṭī', 'kxv-Deva': 'गुजराटी', 'kxv-Latn': 'gujraṭī', 'kxv-Orya': 'ଗୁଜୁରାଟି', 'kxv-Telu': 'గుజరాటి', 'ky': 'гужаратча', 'lb': 'Gujarati', 'lkt': 'Gujarati Iyápi', 'lld': 'gujarati', 'lo': 'ກູຈາຣາຕິ', 'lrc': 'گوجأراتی', 'lt': 'gudžaratų', 'lv': 'gudžaratu', 'mai': 'गुजराती', 'mi': 'Kutarāti', 'mk': 'гуџарати', 'ml': 'ഗുജറാത്തി', 'mn': 'гужарати', 'mr': 'गुजराती', 'ms': 'Gujarat', 'mt': 'Gujarati', 'my': 'ဂူဂျာရသီ', 'mzn': 'گجراتی', 'nb': 'gujarati', 'ne': 'गुजराती', 'nl': 'Gujarati', 'nn': 'gujarati', 'no': 'gujarati', 'nqo': 'ߜ߭ߎߖߙߊߞߊ߲', 'om': 'Afaan Gujarati', 'or': 'ଗୁଜରାଟୀ', 'pa': 'ਗੁਜਰਾਤੀ', 'pa-Guru': 'ਗੁਜਰਾਤੀ', 'pcm': 'Gujarátí Lángwej', 'pl': 'gudżarati', 'ps': 'ګجراتي', 'pt': 'guzerate', 'qu': 'Gujarati Simi', 'rm': 'gujarati', 'ro': 'gujarati', 'ru': 'гуджарати', 'rw': 'Inyegujarati', 'sc': 'gujarati', 'sd': 'گجراتي', 'sd-Arab': 'گجراتي', 'se': 'gujaratagiella', 'si': 'ගුජරාටි', 'sk': 'gudžarátčina', 'sl': 'gudžaratščina', 'smn': 'gudžaratikielâ', 'so': 'Gujaraati', 'sq': 'guxharatisht', 'sr': 'гуџарати', 'sr-Cyrl': 'гуџарати', 'sr-Latn': 'gudžarati', 'sv': 'gujarati', 'sw': 'Kigujarati', 'syr': 'ܓܘܓܐܪܝܬ', 'ta': 'குஜராத்தி', 'te': 'గుజరాతీ', 'tg': 'гуҷаротӣ', 'th': 'คุชราต', 'ti': 'ጉጃራቲ', 'tk': 'gujarati dili', 'to': 'lea fakakutalati', 'tr': 'Güceratça', 'tt': 'гуҗарати', 'ug': 'گۇجاراتچە', 'uk': 'гуджараті', 'ur': 'گجراتی', 'uz': 'gujarot', 'uz-Cyrl': 'гужаротча', 'uz-Latn': 'gujarot', 'vec': 'gujarati', 'vi': 'Tiếng Gujarati', 'wae': 'Gujarati', 'wo': 'Gujarati', 'xnr': 'गुजराती', 'yo': 'Èdè Gujarati', 'yrl': 'guserate', 'yue': '古吉拉特文', 'yue-Hans': '古吉拉特文', 'yue-Hant': '古吉拉特文', 'zh': '古吉拉特语', 'zh-Hans': '古吉拉特语', 'zh-Hant': '古吉拉特文', 'zu': 'isi-Gujarati'},
    'gua': {'en': 'Shiki'},
    'gub': {'en': 'Guajajára'},
    'guc': {'ast': 'wayuu', 'ca': 'wayú', 'ca-ES-valencia': 'wayú', 'cs': 'wayúuština', 'de': 'Wayúu', 'en': 'Wayuu', 'et': 'vajuu', 'fi': 'wayuu', 'fr': 'wayuu', 'ga': 'Uaúis', 'gd': 'Wayuu', 'hi-Latn': 'Wayuu', 'ht': 'wayuu', 'hy': 'վայուու', 'it': 'wayuu', 'ja': 'ワユ語', 'lb': 'Wayuu', 'lt': 'vajų', 'mk': 'гвахиро', 'nb': 'wayuu', 'nl': 'Wayuu', 'nn': 'wayuu', 'no': 'wayuu', 'pl': 'wayúu', 'sv': 'wayuu', 'th': 'วายู', 'to': 'lea fakaʻuaiū', 'tr': 'Wayuu dili', 'yue': '瓦尤文', 'yue-Hans': '瓦尤文', 'yue-Hant': '瓦尤文', 'zh-Hant': '瓦尤文'},
    'gud': {'en': 'Yocoboué Dida'},
    'gue': {'en': 'Gurindji'},
    'guf': {'en': 'Gupapuyngu'},
    'gug': {'en': 'Paraguayan Guaraní'},
    'guh': {'en': 'Guahibo'},
    'gui': {'en': 'Eastern Bolivian Guaraní'},
    'guk': {'en': 'Gumuz'},
    'gul': {'en': 'Sea Island Creole English'},
    'gum': {'en': 'Guambiano'},
    'gun': {'en': 'Mbyá Guaraní'},
    'guo': {'en': 'Guayabero'},
    'gup': {'en': 'Gunwinggu'},
    'guq': {'en': 'Aché'},
    'gur': {'ast': 'frafra', 'cs': 'frafra', 'de': 'Farefare', 'en': 'Frafra', 'et': 'farefare', 'fi': 'frafra', 'fr': 'gurenne', 'gd': 'Frafra', 'hi-Latn': 'Frafra', 'ht': 'gurenne', 'ja': 'フラフラ語', 'lb': 'Farefare', 'lt': 'frafra', 'mk': 'фарефаре', 'nb': 'frafra', 'ne': 'फ्राफ्रा', 'nl': 'Gurune', 'nn': 'frafra', 'no': 'frafra', 'pl': 'frafra', 'sv': 'farefare', 'th': 'ฟราฟรา', 'to': 'lea fakafalefale', 'tr': 'Frafra', 'vi': 'Tiếng Frafra', 'yue': '弗拉弗拉文', 'yue-Hans': '弗拉弗拉文', 'yue-Hant': '弗拉弗拉文', 'zh-Hant': '弗拉弗拉文'},
    'gus': {'en': 'Guinean Sign Language'},
    'gut': {'en': 'Maléku Jaíka'},
    'guu': {'en': 'Yanomamö'},
    'guv': {'en': 'Gey'},
    'guw': {'en': 'Gun'},
    'gux': {'en': 'Gourmanchéma'},
    'guz': {'af': 'Gusii', 'am': 'ጉስሊኛ', 'ar': 'الغيزية', 'as': 'গুছি', 'ast': 'gusii', 'az': 'qusi', 'az-Cyrl': 'гуси', 'az-Latn': 'qusi', 'bal-Latn': 'Gusi', 'be': 'гусіі', 'be-tarask': 'гусіі', 'bg': 'гусии', 'bgn': 'گوسی', 'bn': 'গুসী', 'br': 'gusiieg', 'brx': 'गुशि', 'bs': 'gusi', 'bs-Cyrl': 'гуси', 'bs-Latn': 'gusi', 'ca': 'gusí', 'ca-ES-valencia': 'gusí', 'ccp': '𑄉𑄪𑄥𑄩', 'ce': 'гусии', 'chr': 'ᎫᏏ', 'ckb': 'گووسی', 'cs': 'gusii', 'cy': 'Gusii', 'da': 'gusii', 'de': 'Gusii', 'dsb': 'gusii', 'el': 'Γκούσι', 'el-polyton': 'Γκούσι', 'en': 'Gusii', 'es': 'gusii', 'et': 'gusii', 'eu': 'gusiiera', 'fa': 'گوسی', 'ff-Adlm': '𞤘𞤵𞤧𞤭𞥅𞤪𞤫', 'fi': 'gusii', 'fil': 'Gusii', 'fo': 'gusii', 'fr': 'gusii', 'fy': 'Gusii', 'ga': 'Gúsaís', 'gd': 'Gusii', 'gl': 'gusii', 'gu': 'ગુસી', 'guz': 'Ekegusii', 'ha': 'Gusii', 'he': 'גוסי', 'hi': 'गुसी', 'hi-Latn': 'Gusii', 'hr': 'gusii', 'hsb': 'gusii', 'ht': 'gusii', 'hu': 'guszii', 'hy': 'գուսի', 'ia': 'gusii', 'id': 'Gusii', 'ig': 'Gusii', 'is': 'gusii', 'it': 'gusii', 'ja': 'グシイ語', 'jv': 'Gusii', 'ka': 'გუსიი', 'kea': 'gusii', 'kgp': 'gusij', 'kk': 'гусии тілі', 'kk-Cyrl': 'гусии тілі', 'km': 'ហ្គូស៊ី', 'kn': 'ಗುಸಿ', 'ko': '구시어', 'kok': 'गुसी', 'kok-Deva': 'गुसी', 'ku': 'gusîî', 'ky': 'гусиче', 'lb': 'Gusii-Sprooch', 'lo': 'ກູຊິ', 'lrc': 'گوٙسی', 'lt': 'gusi', 'lv': 'gusii', 'mai': 'गुसी', 'mi': 'Kūhī', 'mk': 'гуси', 'ml': 'ഗുസീ', 'mn': 'гузы', 'mr': 'गसी', 'ms': 'Gusii', 'mt': 'Gusii', 'my': 'ဂူစီး', 'mzn': 'گوسی', 'nb': 'gusii', 'ne': 'गुसी', 'nl': 'Gusii', 'nn': 'gusii', 'no': 'gusii', 'nqo': 'ߜߎ߭ߛߌߞߊ߲', 'or': 'ଗୁସି', 'pa': 'ਗੁਸੀ', 'pa-Guru': 'ਗੁਸੀ', 'pcm': 'Gusí Lángwej', 'pl': 'gusii', 'ps': 'ګوسي', 'pt': 'gusii', 'qu': 'Guzí Simi', 'rm': 'gusii', 'ro': 'gusii', 'ru': 'гусии', 'sc': 'gusii', 'sd': 'گشي', 'sd-Arab': 'گشي', 'si': 'ගුසී', 'sk': 'gusii', 'sl': 'gusijščina', 'smn': 'gusiikielâ', 'so': 'Guusii', 'sq': 'gusisht', 'sr': 'гуси', 'sr-Cyrl': 'гуси', 'sr-Latn': 'gusi', 'sv': 'gusii', 'sw': 'Kikisii', 'ta': 'குஸி', 'te': 'గుస్సీ', 'th': 'กุซซี', 'ti': 'ጉሲ', 'tk': 'gusii dili', 'to': 'lea fakakusī', 'tr': 'Gusii', 'ug': 'گۇسىچە', 'uk': 'гусії', 'ur': 'گسی', 'uz': 'gusii', 'uz-Cyrl': 'гусии', 'uz-Latn': 'gusii', 'vec': 'gusi', 'vi': 'Tiếng Gusii', 'xnr': 'गुसी', 'yo': 'Gusii', 'yrl': 'gusiyi', 'yue': '古西文', 'yue-Hans': '古西文', 'yue-Hant': '古西文', 'zh': '古西语', 'zh-Hans': '古西语', 'zh-Hant': '古西文', 'zu': 'isi-Gusli'},
    'gv': {'af': 'Manx', 'am': 'ማንክስ', 'ar': 'المنكية', 'as': 'মেংক্স', 'ast': 'manés', 'az': 'manks', 'az-Cyrl': 'манкс', 'az-Latn': 'manks', 'bal-Latn': 'Manks', 'be': 'мэнская', 'be-tarask': 'мэнская', 'bg': 'манкски', 'bgn': 'مانی', 'bn': 'ম্যাঙ্কস', 'br': 'manaveg', 'brx': 'मेनक्स', 'bs': 'manks', 'bs-Cyrl': 'манкс', 'bs-Latn': 'manks', 'ca': 'manx', 'ca-ES-valencia': 'manx', 'ccp': '𑄟𑄳𑄠𑄇𑄴𑄥𑄧', 'ce': 'мэнийн', 'chr': 'ᎹᎾᎧᏏ', 'ckb': 'مانکی', 'cs': 'manština', 'cy': 'Manaweg', 'da': 'manx', 'de': 'Manx', 'dsb': 'manšćina', 'el': 'Μανξ', 'el-polyton': 'Μανξ', 'en': 'Manx', 'en-Dsrt': '𐐣𐐰𐑌𐐿𐑅', 'es': 'manés', 'et': 'mänksi', 'eu': 'manxera', 'fa': 'مانی', 'ff-Adlm': '𞤃𞤢𞤲𞤳𞤭𞤧𞤪𞤫', 'fi': 'manksi', 'fil': 'Manx', 'fo': 'manx', 'fr': 'mannois', 'fur': 'manx', 'fy': 'Manks', 'ga': 'Manainnis', 'gd': 'Gaelg', 'gl': 'manx', 'gsw': 'Manx-Gäälisch', 'gu': 'માંક્સ', 'gv': 'Gaelg', 'ha': 'Manx', 'he': 'מאנית', 'hi': 'मैंक्स', 'hi-Latn': 'Manx', 'hr': 'manski', 'hsb': 'manšćina', 'ht': 'mannois', 'hu': 'man-szigeti', 'hy': 'մեներեն', 'ia': 'mannese', 'id': 'Manx', 'ig': 'Mansị', 'is': 'manska', 'it': 'mannese', 'ja': 'マン島語', 'jv': 'Manx', 'ka': 'მენური', 'kea': 'manks', 'kgp': 'mỹnsi', 'kk': 'мэн тілі', 'kk-Cyrl': 'мэн тілі', 'km': 'មេន', 'kn': 'ಮ್ಯಾಂಕ್ಸ್', 'ko': '맹크스어', 'kok': 'मांक्स', 'kok-Deva': 'मांक्स', 'ks': 'مینکس', 'ks-Arab': 'مینکس', 'ksh': 'Mangx', 'ku': 'manksî', 'ky': 'мэнксыча', 'lb': 'Manx', 'lld': 'manx', 'lo': 'ແມງຊ໌', 'lrc': 'مانکس', 'lt': 'meniečių', 'lv': 'meniešu', 'mai': 'मेंक्स', 'mi': 'Manaki', 'mk': 'манкс', 'ml': 'മാൻസ്', 'mn': 'манкс', 'mr': 'मांक्स', 'ms': 'Manx', 'mt': 'Manx', 'my': 'မန်းဇ်', 'mzn': 'مانکس', 'nb': 'mansk', 'ne': 'मान्क्स', 'nl': 'Manx', 'nn': 'manx', 'no': 'mansk', 'nqo': 'ߡߊߣߏߥߊߞߊ߲', 'or': 'ମାଁକ୍ସ', 'pa': 'ਮੈਂਕਸ', 'pa-Guru': 'ਮੈਂਕਸ', 'pcm': 'Mánks Lángwej', 'pl': 'manx', 'ps': 'مینکس', 'pt': 'manx', 'qu': 'Manés Simi', 'rm': 'manx', 'ro': 'manx', 'ru': 'мэнский', 'sc': 'mannesu', 'sd': 'مينڪس', 'sd-Arab': 'مينڪس', 'se': 'manksgiella', 'si': 'මැන්ක්ස්', 'sk': 'mančina', 'sl': 'manščina', 'smn': 'manks', 'so': 'Mankis', 'sq': 'manksisht', 'sr': 'манкс', 'sr-Cyrl': 'манкс', 'sr-Latn': 'manks', 'sv': 'manx', 'sw': 'Kimanx', 'ta': 'மேங்க்ஸ்', 'te': 'మాంక్స్', 'th': 'มานซ์', 'ti': 'ማንክስ', 'tk': 'men dili', 'to': 'lea fakamangikī', 'tr': 'Man dili', 'ug': 'مانچە', 'uk': 'менкська', 'ur': 'مینکس', 'uz': 'men', 'uz-Cyrl': 'мэнча', 'uz-Latn': 'men', 'vec': 'maneze', 'vi': 'Tiếng Manx', 'xnr': 'मैंक्स', 'yi': 'מאַנקס', 'yo': 'Máǹkì', 'yrl': 'mãkisi', 'yue': '曼島文', 'yue-Hans': '曼岛文', 'yue-Hant': '曼島文', 'zh': '马恩语', 'zh-Hans': '马恩语', 'zh-Hant': '曼島文', 'zu': 'isi-Manx'},
    'gva': {'en': 'Guana (Paraguay)'},
    'gvc': {'en': 'Guanano'},
    'gve': {'en': 'Duwet'},
    'gvf': {'en': 'Golin'},
    'gvj': {'en': 'Guajá'},
    'gvl': {'en': 'Gulay'},
    'gvm': {'en': 'Gurmana'},
    'gvn': {'en': 'Kuku-Yalanji'},
    'gvo': {'en': 'Gavião Do Jiparaná'},
    'gvp': {'en': 'Pará Gavião'},
    'gvr': {'en': 'Gurung'},
    'gvs': {'en': 'Gumawana'},
    'gvy': {'en': 'Guyani'},
    'gwa': {'en': 'Mbato'},
    'gwb': {'en': 'Gwa'},
    'gwc': {'en': 'Gawri'},
    'gwd': {'en': 'Gawwada'},
    'gwe': {'en': 'Gweno'},
    'gwf': {'en': 'Gowro'},
    'gwg': {'en': 'Moo'},
    'gwi': {'af': 'Gwichʼin', 'am': 'ግዊቺን', 'ar': 'غوتشن', 'as': 'জিউইচিন', 'ast': 'gwichʼin', 'az': 'qviçin', 'az-Cyrl': 'гвичин', 'az-Latn': 'qviçin', 'be': 'гуіч’ін', 'be-tarask': 'гуіч’ін', 'bg': 'гвичин', 'bgn': 'گویچنی', 'bn': 'গওইচ্’ইন', 'br': 'gwich’in', 'brx': 'ग्वीचलीन', 'bs': 'gvičin', 'bs-Cyrl': 'гвич’ин', 'bs-Latn': 'gvičin', 'ca': 'gwich’in', 'ca-ES-valencia': 'gwich’in', 'ccp': '𑄉𑄧𑄃𑄮𑄃𑄨𑄌𑄴𑄃𑄨𑄚𑄴', 'ce': 'гвичин', 'chr': 'ᏈᏥᏂ', 'ckb': 'گویچین', 'cs': 'gwichʼin', 'cy': 'Gwichʼin', 'da': 'gwichin', 'de': 'Kutchin', 'dsb': 'gwichʼinšćina', 'el': 'Γκουίτσιν', 'el-polyton': 'Γκουίτσιν', 'en': 'Gwichʼin', 'es': 'kutchin', 'et': 'gvitšini', 'eu': 'gwich’inera', 'fa': 'گویچ این', 'ff-Adlm': '𞤘𞤭𞤱𞤧𞤭𞤲𞤪𞤫', 'fi': 'gwitšin', 'fil': 'Gwichʼin', 'fo': 'gwich’in', 'fr': 'gwichʼin', 'fy': 'Gwichʼin', 'ga': 'Goitsinis', 'gd': 'Gwichʼin', 'gl': 'gwichʼin', 'gsw': 'Kutchinisch', 'gu': 'ગ્વિચ’ઇન', 'ha': 'Gwichʼin', 'he': 'גוויצ׳ן', 'hi': 'ग्विचइन', 'hi-Latn': 'Gwichʼin', 'hr': 'gwich’in', 'hsb': 'gwichʼinšćina', 'ht': 'gwichʼin', 'hu': 'gvicsin', 'hy': 'գվիչին', 'ia': 'gwich’in', 'id': 'Gwich’in', 'ig': 'Gwichʼin', 'is': 'gvísín', 'it': 'gwichʼin', 'ja': 'グウィッチン語', 'jv': 'Gwichʼin', 'ka': 'გვიჩინი', 'kgp': 'guvisĩn', 'kk': 'гвичин тілі', 'kk-Cyrl': 'гвичин тілі', 'km': 'ហ្គីចឈីន', 'kn': 'ಗ್ವಿಚ್\u200cಇನ್', 'ko': '그위친어', 'kok': 'ग्विच', 'kok-Deva': 'ग्विच', 'ks': 'گُوِچ اِن', 'ks-Arab': 'گُوِچ اِن', 'ku': 'gwichʼin', 'ky': 'гвичинче', 'lb': 'Kutchin-Sprooch', 'lo': 'ວິດອິນ', 'lt': 'gvičino', 'lv': 'kučinu', 'mai': 'गुअचीन', 'mi': 'Kuitīna', 'mk': 'гвичински', 'ml': 'ഗ്വിച്ചിൻ', 'mn': 'гвичин', 'mr': 'ग्विच’इन', 'ms': 'Gwichʼin', 'mt': 'Gwiċin', 'my': 'ဂွစ်ချင်', 'nb': 'gwich’in', 'ne': 'गुइचिन', 'nl': 'Gwichʼin', 'nn': 'gwich’in', 'no': 'gwich’in', 'or': 'ଗୱିଚ’ଇନ୍', 'pa': 'ਗਵਿਚ’ਇਨ', 'pa-Guru': 'ਗਵਿਚ’ਇਨ', 'pcm': 'Gwichín Lángwej', 'pl': 'gwichʼin', 'ps': 'ګیچین', 'pt': 'gwichʼin', 'qu': 'Gwichʼin Simi', 'rm': 'gwichʼin', 'ro': 'gwichʼin', 'ru': 'гвичин', 'sc': 'gwichʼin', 'sd': 'گوچن', 'sd-Arab': 'گوچن', 'si': 'ග්විචින්', 'sk': 'kučinčina', 'sl': 'gvičin', 'smn': 'gwich’inkielâ', 'so': 'Gwichʼin', 'sq': 'guiçinisht', 'sr': 'гвичински', 'sr-Cyrl': 'гвичински', 'sr-Latn': 'gvičinski', 'sv': 'gwichin', 'sw': 'Gwichʼin', 'ta': 'குவிசின்', 'te': 'గ్విచిన్', 'th': 'กวิชอิน', 'ti': 'ጒቺን', 'tk': 'gwiçin dili', 'to': 'lea fakaʻuīsini', 'tr': 'Guçince', 'ug': 'گىۋىچىنچە', 'uk': 'кучін', 'ur': 'گوئچ ان', 'uz': 'gvichin', 'uz-Cyrl': 'гвичин', 'uz-Latn': 'gvichin', 'vec': 'gwichʼin', 'vi': 'Tiếng Gwichʼin', 'xnr': 'ग्विचइन', 'yo': 'Èdè giwisi', 'yrl': 'guwixi-ĩ', 'yue': '圭契文', 'yue-Hans': '圭契文', 'yue-Hant': '圭契文', 'zh': '哥威迅语', 'zh-Hans': '哥威迅语', 'zh-Hant': '圭契文', 'zu': 'isi-Gwichʼin'},
    'gwj': {'en': 'ǀGwi'},
    'gwm': {'en': 'Awngthim'},
    'gwn': {'en': 'Gwandara'},
    'gwr': {'en': 'Gwere'},
    'gwt': {'en': 'Gawar-Bati'},
    'gwu': {'en': 'Guwamu'},
    'gww': {'en': 'Kwini'},
    'gwx': {'en': 'Gua'},
    'gxx': {'en': 'Wè Southern'},
    'gya': {'en': 'Northwest Gbaya'},
    'gyb': {'en': 'Garus'},
    'gyd': {'en': 'Kayardild'},
    'gye': {'en': 'Gyem'},
    'gyf': {'en': 'Gungabula'},
    'gyg': {'en': 'Gbayi'},
    'gyi': {'en': 'Gyele'},
    'gyl': {'en': 'Gayil'},
    'gym': {'en': 'Ngäbere'},
    'gyn': {'en': 'Guyanese Creole English'},
    'gyo': {'en': 'Gyalsumdo'},
    'gyr': {'en': 'Guarayu'},
    'gyy': {'en': 'Gunya'},
    'gyz': {'en': 'Geji'},
    'gza': {'en': 'Ganza'},
    'gzi': {'en': 'Gazi'},
    'gzn': {'en': 'Gane'},
    'ha': {'af': 'Hausa', 'agq': 'Kɨtsɔŋkaŋ', 'ak': 'Hausa', 'am': 'ሃውሳኛ', 'ar': 'الهوسا', 'as': 'হাউছা', 'asa': 'Kihautha', 'ast': 'ḥausa', 'az': 'hausa', 'az-Cyrl': 'һауса', 'az-Latn': 'hausa', 'bal-Latn': 'Hausá', 'bas': 'Hɔp u ɓausa', 'be': 'хауса', 'be-tarask': 'хауса', 'bem': 'Ichi Hausa', 'bez': 'Hihausa', 'bg': 'хауса', 'bgn': 'هوسه\u200cای', 'bm': 'awusakan', 'bn': 'হাউসা', 'br': 'haousa', 'brx': 'हाउसा', 'bs': 'hausa', 'bs-Cyrl': 'хауса', 'bs-Latn': 'hausa', 'ca': 'haussa', 'ca-ES-valencia': 'haussa', 'ccp': '𑄦𑄃𑄪𑄥', 'ce': 'хауса', 'cgg': 'Oruhausa', 'chr': 'ᎭᎤᏌ', 'ckb': 'هائووسا', 'cs': 'hauština', 'cy': 'Hawsa', 'da': 'hausa', 'dav': 'Kihausa', 'de': 'Haussa', 'dje': 'Hawsance senni', 'dsb': 'hausa', 'dyo': 'hausa', 'dz': 'ཧཝ་ས་ཁ', 'ebu': 'Kĩhausa', 'ee': 'hausagbe', 'el': 'Χάουσα', 'el-polyton': 'Χάουσα', 'en': 'Hausa', 'eo': 'haŭsa', 'es': 'hausa', 'et': 'hausa', 'eu': 'hausa', 'ewo': 'Ǹkɔ́bɔ aúsá', 'fa': 'هوسایی', 'ff': 'Hawsaŋkoore', 'ff-Adlm': '𞤖𞤢𞤱𞤧𞤢𞥄𞤪𞤫', 'ff-Latn': 'Hawsaŋkoore', 'fi': 'hausa', 'fil': 'Hausa', 'fo': 'hausa', 'fr': 'haoussa', 'fy': 'Hausa', 'ga': 'Hásais', 'gd': 'Hausa', 'gl': 'hausa', 'gsw': 'Haussa', 'gu': 'હૌસા', 'guz': 'Kihausa', 'ha': 'Hausa', 'he': 'האוסה', 'hi': 'हौसा', 'hi-Latn': 'Hausa', 'hr': 'hausa', 'hsb': 'hausa', 'ht': 'haoussa', 'hu': 'hausza', 'hy': 'հաուսա', 'ia': 'hausa', 'id': 'Hausa', 'ig': 'Hausa', 'is': 'hása', 'it': 'hausa', 'ja': 'ハウサ語', 'jmc': 'Kyihausa', 'jv': 'Hausa', 'ka': 'ჰაუსა', 'kab': 'Tahwasit', 'kam': 'Kihausa', 'kde': 'Chihausa', 'kea': 'auza', 'kgp': 'havusa', 'khq': 'Hawsance senni', 'ki': 'Kihausa', 'kk': 'хауса тілі', 'kk-Cyrl': 'хауса тілі', 'kln': 'kutitab Hausa', 'km': 'ហូសា', 'kn': 'ಹೌಸಾ', 'ko': '하우사어', 'kok': 'हौसा', 'kok-Deva': 'हौसा', 'ks': 'ہاوسا', 'ks-Arab': 'ہاوسا', 'ksb': 'Kihausa', 'ksf': 'rikaksa', 'ksh': 'Haußa', 'ku': 'hawsayî', 'ky': 'хаусача', 'lag': 'Kɨhaúusa', 'lb': 'Hausa', 'lg': 'Luhawuza', 'lkt': 'Hausa Iyápi', 'lld': 'haussa', 'ln': 'hausa', 'lo': 'ເຮົາຊາ', 'lrc': 'ھائوسا', 'lt': 'hausų', 'lu': 'Hausa', 'luo': 'Kihausa', 'luy': 'Kihausa', 'lv': 'hausu', 'mai': 'हउसा', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ hausa', 'mer': 'Kĩhausa', 'mfe': 'haoussa', 'mg': 'haoussa', 'mgh': 'Ihausa', 'mi': 'Hauha', 'mk': 'хауса', 'ml': 'ഹൗസ', 'mn': 'хауса', 'mr': 'हौसा', 'ms': 'Hausa', 'mt': 'Hausa', 'mua': 'haussa', 'my': 'ဟာဥစာ', 'mzn': 'هوسا', 'naq': 'Hausagowab', 'nb': 'hausa', 'nd': 'isi-Hausa', 'ne': 'हाउसा', 'nl': 'Hausa', 'nmg': 'Kiɛl máwúsá', 'nn': 'hausa', 'no': 'hausa', 'nqo': 'ߤߊߥߎߛߊߞߊ߲', 'nus': 'Thok ɣowthani', 'nyn': 'Oruhausa', 'om': 'Hawusaa', 'or': 'ହୌସା', 'pa': 'ਹੌਸਾ', 'pa-Guru': 'ਹੌਸਾ', 'pcm': 'Háusá Lángwej', 'pl': 'hausa', 'ps': 'هوسا', 'pt': 'hauçá', 'qu': 'Hausa Simi', 'rm': 'haussa', 'rn': 'Igihawusa', 'ro': 'hausa', 'rof': 'Kihausa', 'ru': 'хауса', 'rwk': 'Kyihausa', 'saq': 'Kihausa', 'sbp': 'Ishihawusa', 'sc': 'hausa', 'sd': 'هوسا', 'sd-Arab': 'هوسا', 'se': 'haussagiella', 'seh': 'hausa', 'ses': 'Hawsance senni', 'sg': 'Haüsä', 'shi': 'ⵜⴰⵀⴰⵡⵙⴰⵜ', 'shi-Latn': 'Tahawsat', 'shi-Tfng': 'ⵜⴰⵀⴰⵡⵙⴰⵜ', 'si': 'හෝසා', 'sk': 'hauština', 'sl': 'havščina', 'smn': 'hausakielâ', 'sn': 'chiHausa', 'so': 'Hawsa', 'sq': 'hausisht', 'sr': 'хауса', 'sr-Cyrl': 'хауса', 'sr-Latn': 'hausa', 'sv': 'hausa', 'sw': 'Kihausa', 'ta': 'ஹௌஸா', 'te': 'హౌసా', 'teo': 'Kihausa', 'tg': 'ҳауса', 'th': 'เฮาซา', 'ti': 'ሃውሳ', 'tk': 'hausa dili', 'to': 'lea fakahausa', 'tr': 'Hausa dili', 'tt': 'хауса', 'twq': 'Hawsance senni', 'tzm': 'Tahawsat', 'ug': 'خائۇساچە', 'uk': 'хауса', 'ur': 'ہؤسا', 'uz': 'xausa', 'uz-Cyrl': 'хауса', 'uz-Latn': 'xausa', 'vai': 'ꕌꖙꕢ', 'vai-Latn': 'Hawusa', 'vai-Vaii': 'ꕌꖙꕢ', 'vec': 'hausa', 'vi': 'Tiếng Hausa', 'vun': 'Kyihausa', 'wae': 'Hausa', 'wo': 'Hawsa', 'xnr': 'हौसा', 'xog': 'Oluhawuza', 'yav': 'pakas', 'yi': 'האַוסאַ', 'yo': 'Èdè Hausa', 'yrl': 'hausá', 'yue': '豪撒文', 'yue-Hans': '豪撒文', 'yue-Hant': '豪撒文', 'zgh': 'ⵜⴰⵀⴰⵡⵙⴰⵜ', 'zh': '豪萨语', 'zh-Hans': '豪萨语', 'zh-Hant': '豪薩文', 'zu': 'isi-Hausa'},
    'haa': {'en': 'Han'},
    'hab': {'en': 'Hanoi Sign Language'},
    'hac': {'en': 'Gurani'},
    'had': {'en': 'Hatam'},
    'hae': {'en': 'Eastern Oromo'},
    'haf': {'en': 'Haiphong Sign Language'},
    'hag': {'en': 'Hanga'},
    'hah': {'en': 'Hahon'},
    'hai': {'af': 'Haida', 'am': 'ሃይዳ', 'ar': 'الهيدا', 'as': 'হেইডা', 'ast': 'haida', 'az': 'hayda', 'az-Latn': 'hayda', 'be': 'хайда', 'be-tarask': 'хайда', 'bg': 'хайда', 'bn': 'হাইডা', 'br': 'haideg', 'brx': 'हाईडा', 'bs': 'haida', 'bs-Cyrl': 'хаида', 'bs-Latn': 'haida', 'ca': 'haida', 'ca-ES-valencia': 'haida', 'ccp': '𑄦𑄭𑄓', 'chr': 'ᎭᏱᏓ', 'cs': 'haidština', 'cy': 'Haida', 'da': 'haida', 'de': 'Haida', 'dsb': 'haidašćina', 'el': 'Χάιντα', 'el-polyton': 'Χάιντα', 'en': 'Haida', 'es': 'haida', 'et': 'haida', 'eu': 'haidera', 'fa': 'هایدایی', 'ff-Adlm': '𞤖𞤢𞤴𞤣𞤢𞥄𞤪𞤫', 'fi': 'haida', 'fil': 'Haida', 'fr': 'haïda', 'fy': 'Haida', 'ga': 'Haídis', 'gd': 'Haida', 'gl': 'haida', 'gsw': 'Haida', 'gu': 'હૈડા', 'ha': 'Haida', 'he': 'האידה', 'hi': 'हैडा', 'hi-Latn': 'Haida', 'hr': 'haidi', 'hsb': 'haidašćina', 'ht': 'haïda', 'hu': 'haida', 'hy': 'հայդա', 'ia': 'haida', 'id': 'Haida', 'ig': 'Haida', 'is': 'haída', 'it': 'haida', 'ja': 'ハイダ語', 'jv': 'Haida', 'ka': 'ჰაიდა', 'kgp': 'hajna', 'kk': 'хайда тілі', 'kk-Cyrl': 'хайда тілі', 'km': 'ហៃដា', 'kn': 'ಹೈಡಾ', 'ko': '하이다어', 'kok': 'हैदा', 'kok-Deva': 'हैदा', 'ks': 'ہَیدا', 'ks-Arab': 'ہَیدا', 'ku': 'haydayî', 'ky': 'хайдача', 'lb': 'Haida-Sprooch', 'lo': 'ໄຮດາ', 'lt': 'haido', 'lv': 'haidu', 'mai': 'हैदा', 'mi': 'Heira', 'mk': 'хајда', 'ml': 'ഹൈഡ', 'mn': 'хайда', 'mr': 'हैडा', 'ms': 'Haida', 'mt': 'Haida', 'my': 'ဟေဒါ', 'nb': 'haida', 'ne': 'हाइदा', 'nl': 'Haida', 'nn': 'haida', 'no': 'haida', 'or': 'ହାଇଡା', 'pa': 'ਹਾਇਡਾ', 'pa-Guru': 'ਹਾਇਡਾ', 'pcm': 'Haída Lángwej', 'pl': 'haida', 'ps': 'ha', 'pt': 'haida', 'qu': 'Haida Simi', 'rm': 'haida', 'ro': 'haida', 'ru': 'хайда', 'sc': 'haida', 'sd': 'ھائيڊا', 'sd-Arab': 'ھائيڊا', 'si': 'හයිඩා', 'sk': 'haida', 'sl': 'haidščina', 'so': 'Haida', 'sq': 'haidaisht', 'sr': 'хаида', 'sr-Cyrl': 'хаида', 'sr-Latn': 'haida', 'sv': 'haida', 'sw': 'Kihaida', 'ta': 'ஹைடா', 'te': 'హైడా', 'th': 'ไฮดา', 'ti': 'ሃይዳ', 'tk': 'haýda dili', 'to': 'lea fakahaita', 'tr': 'Haydaca', 'ug': 'ھەيدەچە', 'uk': 'хайда', 'ur': 'ہائیڈا', 'uz': 'hayda', 'uz-Latn': 'hayda', 'vec': 'haida', 'vi': 'Tiếng Haida', 'xnr': 'हैडा', 'yo': 'Èdè Haida', 'yrl': 'haida', 'yue': '海達文', 'yue-Hans': '海达文', 'yue-Hant': '海達文', 'zh': '海达语', 'zh-Hans': '海达语', 'zh-Hant': '海達文', 'zu': 'isi-Haida'},
    'haj': {'en': 'Hajong'},
    'hak': {'af': 'Hakka-Sjinees', 'am': 'ሃካ ቻይንኛ', 'ar': 'الهاكا الصينية', 'ast': 'chinu hakka', 'az': 'hakka', 'az-Latn': 'hakka', 'br': 'sinaeg Hakka', 'ca': 'xinès hakka', 'ca-ES-valencia': 'xinès hakka', 'ccp': '𑄦𑄧𑄇𑄴', 'cs': 'čínština (dialekty Hakka)', 'da': 'hakka-kinesisk', 'de': 'Hakka', 'en': 'Hakka Chinese', 'es': 'chino hakka', 'et': 'hakka', 'fi': 'hakka-kiina', 'fo': 'hakka kinesiskt', 'fr': 'hakka', 'ga': 'Haicéis', 'gd': 'Hakka', 'he': 'סינית האקה', 'hi-Latn': 'Hakka Chinese', 'hr': 'hakka kineski', 'ht': 'hakka', 'hu': 'hakka kínai', 'it': 'hakka', 'ja': '客家語', 'kgp': 'haka', 'kn': 'ಹಕ್', 'ko': '하카어', 'ky': 'Хакка кытайча', 'lb': 'Hakka-Chinesesch', 'lt': 'kinų kalbos hakų tarmė', 'mk': 'хака', 'ml': 'ഹാക്ക ചൈനീസ്', 'mr': 'हाक्का चिनी', 'ms': 'Cina Hakka', 'nb': 'hakka', 'ne': 'हक्का चिनियाँ', 'nl': 'Hakka', 'nn': 'hakka', 'no': 'hakka', 'pa': 'ਚੀਨੀ ਹਾਕਾ', 'pa-Guru': 'ਚੀਨੀ ਹਾਕਾ', 'pl': 'hakka', 'pt': 'hacá', 'ro': 'chineză hakka', 'ru': 'хакка', 'si': 'හකා චයිනිස්', 'sv': 'hakka', 'ta': 'ஹக்கா சீனம்', 'te': 'హక్కా చైనీస్', 'th': 'จีนแคะ', 'to': 'lea fakasiaina-haka', 'tr': 'Hakka Çincesi', 'uk': 'хаккаська', 'vi': 'Tiếng Khách Gia', 'yrl': 'haká', 'yue': '客家話', 'yue-Hans': '客家话', 'yue-Hant': '客家話', 'zh': '客家语', 'zh-Hans': '客家语', 'zh-Hant': '客家話', 'zu': 'isi-Hakka Chinese'},
    'hal': {'en': 'Halang'},
    'ham': {'en': 'Hewa'},
    'han': {'en': 'Hangaza'},
    'hao': {'en': 'Hakö'},
    'hap': {'en': 'Hupla'},
    'haq': {'en': 'Ha'},
    'har': {'en': 'Harari'},
    'has': {'en': 'Haisla'},
    'hav': {'en': 'Havu'},
    'haw': {'af': 'Hawais', 'am': 'ሃዊያኛ', 'ar': 'لغة هاواي', 'as': 'হাৱাই', 'ast': 'hawaianu', 'az': 'havay', 'az-Cyrl': 'һавај', 'az-Latn': 'havay', 'bal-Latn': 'Hawái', 'be': 'гавайская', 'be-tarask': 'гавайская', 'bg': 'хавайски', 'bgn': 'هاوایی', 'bn': 'হাওয়াইয়ান', 'br': 'hawaieg', 'brx': 'हावाईयान', 'bs': 'havajski', 'bs-Cyrl': 'хавајски', 'bs-Latn': 'havajski', 'ca': 'hawaià', 'ca-ES-valencia': 'hawaià', 'ccp': '𑄦𑄤𑄃𑄨𑄠𑄚𑄴', 'ce': 'гавайн', 'chr': 'ᎭᏩᎼ', 'ckb': 'هاوایی', 'cs': 'havajština', 'cy': 'Hawäieg', 'da': 'hawaiiansk', 'de': 'Hawaiisch', 'dsb': 'hawaiišćina', 'dz': 'ཧ་ཝ་ཡིའི་ཁ', 'ee': 'hawaigbe', 'el': 'Χαβαϊκά', 'el-polyton': 'Χαβαϊκά', 'en': 'Hawaiian', 'en-Dsrt': '𐐐𐐲𐐶𐐴𐐲𐑌', 'eo': 'havaja', 'es': 'hawaiano', 'et': 'havai', 'eu': 'hawaiiera', 'fa': 'هاوایی', 'ff-Adlm': '𞤖𞤢𞤱𞤢𞥄𞤭𞤴𞤫𞤲𞤪𞤫', 'fi': 'havaiji', 'fil': 'Hawaiian', 'fo': 'hawaiianskt', 'fr': 'hawaïen', 'fy': 'Hawaïaansk', 'ga': 'Haváis', 'gd': 'Cànan Hawai’i', 'gl': 'hawaiano', 'gsw': 'Hawaiianisch', 'gu': 'હવાઇયન', 'ha': 'Hawaiianci', 'haw': 'ʻŌlelo Hawaiʻi', 'he': 'הוואית', 'hi': 'हवाई', 'hi-Latn': 'Hawaiian', 'hr': 'havajski', 'hsb': 'hawaiišćina', 'ht': 'hawaïen', 'hu': 'hawaii', 'hy': 'հավայիերեն', 'ia': 'hawaiano', 'id': 'Hawaii', 'ig': 'Hawaịlịan', 'is': 'havaíska', 'it': 'hawaiano', 'ja': 'ハワイ語', 'jv': 'Hawaii', 'ka': 'ჰავაიური', 'kea': 'avaianu', 'kgp': 'havajỹnũ', 'kk': 'гавайи тілі', 'kk-Cyrl': 'гавайи тілі', 'km': 'ហាវ៉ៃ', 'kn': 'ಹವಾಯಿಯನ್', 'ko': '하와이어', 'kok': 'हवायियान', 'kok-Deva': 'हवायियान', 'ks': 'ہوایِیَن', 'ks-Arab': 'ہوایِیَن', 'ksh': 'Hauajahnesch', 'ku': 'hawayî', 'ky': 'гавайча', 'lb': 'Hawaiesch', 'lkt': 'Hawaiian Iyápi', 'lo': 'ຮາໄວອຽນ', 'lrc': 'ھاڤایی', 'lt': 'havajiečių', 'lv': 'havajiešu', 'mai': 'हवाईयन', 'mi': 'Wāhu', 'mk': 'хавајски', 'ml': 'ഹവായിയൻ', 'mn': 'хавай', 'mr': 'हवाईयन', 'ms': 'Hawaii', 'mt': 'Ħawajjan', 'my': 'ဟာဝိုင်ယီ', 'mzn': 'هاواییایی', 'nb': 'hawaiisk', 'ne': 'हवाइयन', 'nl': 'Hawaïaans', 'nn': 'hawaiisk', 'no': 'hawaiisk', 'nqo': 'ߤߥߊߦߌߞߊ߲', 'or': 'ହାୱାଇନ୍', 'pa': 'ਹਵਾਈ', 'pa-Guru': 'ਹਵਾਈ', 'pcm': 'Hawaii Lángwej', 'pl': 'hawajski', 'ps': 'هوایی', 'pt': 'havaiano', 'qu': 'Hawaiano Simi', 'rm': 'hawaian', 'ro': 'hawaiiană', 'ru': 'гавайский', 'sc': 'hawaianu', 'sd': 'هوائي', 'sd-Arab': 'هوائي', 'se': 'hawaiigiella', 'si': 'හවායි', 'sk': 'havajčina', 'sl': 'havajščina', 'smn': 'hawaijikielâ', 'so': 'Hawaay', 'sq': 'havaisht', 'sr': 'хавајски', 'sr-Cyrl': 'хавајски', 'sr-Latn': 'havajski', 'sv': 'hawaiiska', 'sw': 'Kihawai', 'ta': 'ஹவாயியன்', 'te': 'హవాయియన్', 'tg': 'ҳавайӣ', 'th': 'ฮาวาย', 'ti': 'ሃዋይኛ', 'tk': 'gawaý dili', 'to': 'lea fakahauaiʻi', 'tr': 'Hawaii dili', 'tt': 'гавайи', 'ug': 'ھاۋايچە', 'uk': 'гавайська', 'ur': 'ہوائی', 'uz': 'gavaycha', 'uz-Cyrl': 'гавайча', 'uz-Latn': 'gavaycha', 'vec': 'hawaian', 'vi': 'Tiếng Hawaii', 'wae': 'Hawaíaniš', 'wo': 'Hawaye', 'xnr': 'हवाई', 'yo': 'Hawaiian', 'yrl': 'hawayanu', 'yue': '夏威夷文', 'yue-Hans': '夏威夷文', 'yue-Hant': '夏威夷文', 'zh': '夏威夷语', 'zh-Hans': '夏威夷语', 'zh-Hant': '夏威夷文', 'zu': 'isi-Hawaiian'},
    'hax': {'af': 'Suid-Haida', 'am': 'ደቡባዊ ሃይዳ', 'ar': 'هايدا الجنوبية', 'as': 'দাক্ষিণাত্যৰ হাইডা', 'az': 'cənubi haida', 'az-Latn': 'cənubi haida', 'be': 'паўднёвая хайда', 'be-tarask': 'паўднёвая хайда', 'bg': 'южен хайда', 'bn': 'দক্ষিণী হায়দা', 'br': 'haideg ar Su', 'brx': 'खोलायारि हाईडा', 'bs': 'južni haida', 'bs-Cyrl': 'јужни хаида', 'bs-Latn': 'južni haida', 'ca': 'haida meridional', 'ca-ES-valencia': 'haida meridional', 'chr': 'ᏧᎦᏃᏮ ᏗᏜ ᎭᏱᏓ', 'cs': 'haidština (jižní)', 'cy': 'Haida Deheuol', 'da': 'sydhaida', 'de': 'Süd-Haida', 'dsb': 'pódpołdnjowa haidašćina', 'el': 'Βόρεια Χάιντα', 'el-polyton': 'Βόρεια Χάιντα', 'en': 'Southern Haida', 'es': 'haida meridional', 'et': 'lõunahaida', 'eu': 'hegoaldeko haidera', 'fa': 'هایدا جنوبی', 'ff-Adlm': '𞤖𞤢𞤭𞤣𞤢𞥄𞤪𞤫 𞤙𞤢𞥄𞤥𞤲𞤢𞥄𞤲𞤺𞤫𞤲𞤳𞤮', 'fi': 'haida (eteläinen)', 'fil': 'Katimugang Haida', 'fr': 'haïda du Sud', 'ga': 'Haídis an Deiscirt', 'gd': 'Haida Dheasach', 'gl': 'haida do sur', 'gu': 'દક્ષિણ હૈડા', 'ha': 'Haida na Kudanci', 'he': 'האידה דרומית', 'hi': 'दक्षिणी हाइदा', 'hi-Latn': 'Southern Haida', 'hr': 'južni haida', 'hsb': 'južna haidašćina', 'ht': 'haïda du Sud', 'hu': 'déli haida', 'hy': 'հարավային հայդա', 'ia': 'haida del sud', 'id': 'Haida Selatan', 'ig': 'Southern Haida', 'is': 'suður-haída', 'it': 'haida meridionale', 'ja': '南部ハイダ語', 'jv': 'Haida Sisih Kidul', 'ka': 'სამხრეთული ჰაიდა', 'kk': 'оңтүстік хайда тілі', 'kk-Cyrl': 'оңтүстік хайда тілі', 'km': 'ហៃដាខាងត្បូង', 'kn': 'ದಕ್ಷಿಣ ಹೈಡಾ', 'ko': '남부 하이다어', 'kok': 'दक्षिणी हैदा', 'kok-Deva': 'दक्षिणी हैदा', 'ku': 'haîdaya başûrî', 'ky': 'хайдача (индей тили)', 'lo': 'ໄຮດາໃຕ້', 'lt': 'Pietų Haidos', 'lv': 'dienvidhaidu', 'mai': 'दक्षिणी हैदा', 'mi': 'Haira ki te Tonga', 'mk': 'јужен хајда', 'ml': 'സതേൺ ഹൈഡ', 'mn': 'өмнөд хайда', 'mr': 'दक्षिणात्य हैडा', 'ms': 'Haida Selatan', 'my': 'တောင် ဟိုင်ဒါ', 'nb': 'sørlig haida', 'ne': 'दक्षिणी हैडा', 'nl': 'Zuid-Haida', 'nn': 'sørleg haida', 'no': 'sørlig haida', 'or': 'ସାଉଥ୍ ହାଇଡା', 'pa': 'ਦੱਖਣੀ ਹਾਇਡਾ', 'pa-Guru': 'ਦੱਖਣੀ ਹਾਇਡਾ', 'pcm': 'Saútán Haida', 'pl': 'haida południowy', 'ps': 'جنوبي هایدا', 'pt': 'haida do sul', 'qu': 'Haida Meridional', 'ro': 'haida de sud', 'ru': 'южный хайда', 'sc': 'haida meridionale', 'sd': 'ڏاکڻي ھائڊا', 'sd-Arab': 'ڏاکڻي ھائڊا', 'si': 'දකුණු හයිඩා', 'sk': 'haida (juh)', 'sl': 'južna haidščina', 'so': 'Southern Haida', 'sq': 'haidaishte jugore', 'sr': 'јужни хаида', 'sr-Cyrl': 'јужни хаида', 'sr-Latn': 'južni haida', 'sv': 'sydhaida', 'sw': 'Kihaida cha Kusini', 'ta': 'தெற்கு ஹைடா', 'te': 'దక్షిణ హైదా', 'th': 'เฮดาใต้', 'ti': 'ደቡባዊ ሃይዳ', 'tk': 'günorta haýda dili', 'to': 'lea fakahaita-tonga', 'tr': 'Güney Haydaca', 'uk': 'південна гайда', 'ur': 'جنوبی ہائیڈا', 'uz': 'janubiy hayda', 'uz-Latn': 'janubiy hayda', 'vec': 'haida meridionale', 'vi': 'Tiếng Haida miền Nam', 'yo': 'Èdè Gusu Haida', 'yue': '南海達文', 'yue-Hans': '南海达文', 'yue-Hant': '南海達文', 'zh': '南海达语', 'zh-Hans': '南海达语', 'zh-Hant': '南海達語', 'zu': 'Southern Haida'},
    'hay': {'en': 'Haya'},
    'haz': {'en': 'Hazaragi'},
    'hba': {'en': 'Hamba'},
    'hbb': {'en': 'Huba'},
    'hbn': {'en': 'Heiban'},
    'hbo': {'en': 'Ancient Hebrew'},
    'hbu': {'en': 'Habu'},
    'hca': {'en': 'Andaman Creole Hindi'},
    'hch': {'en': 'Huichol'},
    'hdn': {'en': 'Northern Haida'},
    'hds': {'en': 'Honduras Sign Language'},
    'hdy': {'en': 'Hadiyya'},
    'he': {'af': 'Hebreeus', 'ak': 'Hibri kasa', 'am': 'ዕብራይስጥ', 'ar': 'العبرية', 'as': 'হিব্ৰু', 'ast': 'hebréu', 'az': 'ivrit', 'az-Arab': 'عبریجه', 'az-Cyrl': 'иврит', 'az-Latn': 'ivrit', 'bal-Latn': 'Ebráni', 'be': 'іўрыт', 'be-tarask': 'іўрыт', 'bg': 'иврит', 'bgn': 'عبرانی', 'bn': 'হিব্রু', 'br': 'hebraeg', 'brx': 'हिब्रिउ', 'bs': 'hebrejski', 'bs-Cyrl': 'хебрејски', 'bs-Latn': 'hebrejski', 'ca': 'hebreu', 'ca-ES-valencia': 'hebreu', 'ccp': '𑄦𑄨𑄛𑄴𑄝𑄳𑄢𑄪', 'ce': 'жугтийн', 'chr': 'ᎠᏂᏈᎷ', 'ckb': 'عیبری', 'cs': 'hebrejština', 'csw': 'ᐦᐄᑊᕑᐅᐤ', 'cy': 'Hebraeg', 'da': 'hebraisk', 'de': 'Hebräisch', 'dsb': 'hebrejšćina', 'dz': 'ཧེ་བྲུ་ཁ', 'ee': 'hebrigbe', 'el': 'Εβραϊκά', 'el-polyton': 'Ἑβραϊκά', 'en': 'Hebrew', 'eo': 'hebrea', 'es': 'hebreo', 'et': 'heebrea', 'eu': 'hebreera', 'fa': 'عبری', 'ff-Adlm': '𞤖𞤭𞤦𞤵𞤪𞤵𞥅𞤪𞤫', 'fi': 'heprea', 'fil': 'Hebrew', 'fo': 'hebraiskt', 'fr': 'hébreu', 'fur': 'ebraic', 'fy': 'Hebreeuwsk', 'ga': 'Eabhrais', 'gd': 'Eabhra', 'gl': 'hebreo', 'gsw': 'Hebräisch', 'gu': 'હીબ્રુ', 'ha': 'Ibrananci', 'haw': 'Hebera', 'he': 'עברית', 'hi': 'हिब्रू', 'hi-Latn': 'Hebrew', 'hr': 'hebrejski', 'hsb': 'hebrejšćina', 'ht': 'hébreu', 'hu': 'héber', 'hy': 'եբրայերեն', 'ia': 'hebreo', 'id': 'Ibrani', 'ig': 'Hebrew', 'is': 'hebreska', 'it': 'ebraico', 'ja': 'ヘブライ語', 'jv': 'Ibrani', 'ka': 'ებრაული', 'kea': 'ebraiku', 'kgp': 'emrajko', 'kk': 'иврит тілі', 'kk-Cyrl': 'иврит тілі', 'km': 'ហេប្រឺ', 'kn': 'ಹೀಬ್ರೂ', 'ko': '히브리어', 'kok': 'हिब्रू', 'kok-Deva': 'हिब्रू', 'ks': 'عبرٲنؠ', 'ks-Arab': 'عبرٲنؠ', 'ksh': 'Hebrähjesch', 'ku': 'îbranî', 'kxv': 'hibru', 'kxv-Deva': 'हिब्रू', 'kxv-Latn': 'hibru', 'kxv-Orya': 'ହିବ୍ରୁ', 'kxv-Telu': 'హిబ్రూ', 'ky': 'ивритче', 'lb': 'Hebräesch', 'lkt': 'Hebrew Iyápi', 'lld': 'ebraich', 'lo': 'ຮີບຣິວ', 'lrc': 'عئبری', 'lt': 'hebrajų', 'lv': 'ivrits', 'mai': 'हीब्रू', 'mi': 'Hīperu', 'mk': 'хебрејски', 'ml': 'ഹീബ്രു', 'mn': 'еврей', 'mr': 'हिब्रू', 'ms': 'Ibrani', 'mt': 'Ebrajk', 'my': 'ဟီဘရူး', 'mzn': 'عبری', 'nb': 'hebraisk', 'ne': 'हिब्रु', 'nl': 'Hebreeuws', 'nn': 'hebraisk', 'no': 'hebraisk', 'nqo': 'ߋߓߙߋߞߊ߲', 'om': 'Afaan Hebrew', 'or': 'ହିବ୍ରୁ', 'os': 'уираг', 'pa': 'ਹਿਬਰੂ', 'pa-Guru': 'ਹਿਬਰੂ', 'pcm': 'Híbru Lángwej', 'pl': 'hebrajski', 'ps': 'عبراني', 'pt': 'hebraico', 'qu': 'Hebreo Simi', 'rm': 'ebraic', 'ro': 'ebraică', 'ru': 'иврит', 'rw': 'Igiheburayo', 'sc': 'ebreu', 'sd': 'عبراني', 'sd-Arab': 'عبراني', 'si': 'හීබෲ', 'sk': 'hebrejčina', 'sl': 'hebrejščina', 'smn': 'hepreakielâ', 'so': 'Cibraani', 'sq': 'hebraisht', 'sr': 'хебрејски', 'sr-Cyrl': 'хебрејски', 'sr-Latn': 'hebrejski', 'sv': 'hebreiska', 'sw': 'Kiebrania', 'syr': 'ܥܒܪܐܝܬ', 'ta': 'ஹீப்ரூ', 'te': 'హిబ్రూ', 'tg': 'ибронӣ', 'th': 'ฮิบรู', 'ti': 'እብራይስጢ', 'tk': 'ýewreý dili', 'to': 'lea fakahepelū', 'tr': 'İbranice', 'tt': 'яһүд', 'ug': 'ئىبرانىيچە', 'uk': 'іврит', 'ur': 'عبرانی', 'uz': 'ivrit', 'uz-Cyrl': 'иброний', 'uz-Latn': 'ivrit', 'vec': 'ebràego', 'vi': 'Tiếng Do Thái', 'wae': 'Hebräiš', 'wo': 'Ebrë', 'xnr': 'हिब्रू', 'yi': 'העברעאיש', 'yo': 'Èdè Heberu', 'yrl': 'heburaiku', 'yue': '希伯來文', 'yue-Hans': '希伯来文', 'yue-Hant': '希伯來文', 'zh': '希伯来语', 'zh-Hans': '希伯来语', 'zh-Hant': '希伯來文', 'zu': 'isi-Hebrew'},
    'hea': {'en': 'Northern Qiandong Miao'},
    'hed': {'en': 'Herdé'},
    'heg': {'en': 'Helong'},
    'heh': {'en': 'Hehe'},
    'hei': {'en': 'Heiltsuk'},
    'hem': {'en': 'Hemba'},
    'hgm': {'en': 'Haiǁom'},
    'hgw': {'en': 'Haigwai'},
    'hhi': {'en': 'Hoia Hoia'},
    'hhr': {'en': 'Kerak'},
    'hhy': {'en': 'Hoyahoya'},
    'hi': {'af': 'Hindi', 'agq': 'Endì', 'ak': 'Hindi', 'am': 'ሕንድኛ', 'ar': 'الهندية', 'as': 'হিন্দী', 'asa': 'Kihindi', 'ast': 'hindi', 'az': 'hind', 'az-Arab': 'هیندجه', 'az-Cyrl': 'һинд', 'az-Latn': 'hind', 'bal-Latn': 'Hendi', 'bas': 'Hɔp u hindì', 'be': 'хіндзі', 'be-tarask': 'хіндзі', 'bem': 'Ichi Hindu', 'bez': 'Hihindi', 'bg': 'хинди', 'bgn': 'هندی', 'bm': 'inidikan', 'bn': 'হিন্দি', 'bo': 'ཧིན་དི', 'br': 'hindi', 'brx': 'हिन्दी', 'bs': 'hindi', 'bs-Cyrl': 'хинди', 'bs-Latn': 'hindi', 'ca': 'hindi', 'ca-ES-valencia': 'hindi', 'ccp': '𑄦𑄨𑄚𑄴𑄓𑄨', 'ce': 'хӀинди', 'ceb': 'Hindi', 'cgg': 'Oruhindi', 'chr': 'ᎯᏂᏗ', 'ckb': 'هیندی', 'cs': 'hindština', 'cv': 'хинди', 'cy': 'Hindi', 'da': 'hindi', 'dav': 'Kihindi', 'de': 'Hindi', 'dje': 'Induu senni', 'dsb': 'hindišćina', 'dyo': 'endu', 'dz': 'ཧིན་དི་ཁ', 'ebu': 'Kĩhindĩ', 'ee': 'Hindigbe', 'el': 'Χίντι', 'el-polyton': 'Χίντι', 'en': 'Hindi', 'en-Dsrt': '𐐐𐐮𐑌𐐼𐐨', 'eo': 'hinda', 'es': 'hindi', 'et': 'hindi', 'eu': 'hindia', 'ewo': 'Ǹkɔ́bɔ hindí', 'fa': 'هندی', 'ff': 'Hinndi', 'ff-Adlm': '𞤖𞤭𞤲𞤣𞤭𞥅𞤪𞤫', 'ff-Latn': 'Hinndi', 'fi': 'hindi', 'fil': 'Hindi', 'fo': 'hindi', 'fr': 'hindi', 'fur': 'hindi', 'fy': 'Hindi', 'ga': 'Hiondúis', 'gd': 'Hindis', 'gl': 'hindi', 'gsw': 'Hindi', 'gu': 'હિન્દી', 'guz': 'Kihindi', 'ha': 'Harshen Hindi', 'he': 'הינדי', 'hi': 'हिन्दी', 'hi-Latn': 'Hindi', 'hr': 'hindski', 'hsb': 'hindišćina', 'ht': 'hindi', 'hu': 'hindi', 'hy': 'հինդի', 'ia': 'hindi', 'id': 'Hindi', 'ig': 'Hindi', 'ii': 'ꑴꄃꉙ', 'is': 'hindí', 'it': 'hindi', 'ja': 'ヒンディー語', 'jmc': 'Kyihindi', 'jv': 'India', 'ka': 'ჰინდი', 'kab': 'Tahendit', 'kam': 'Kihindi', 'kde': 'Chihindi', 'kea': 'indi', 'kgp': 'hĩni', 'khq': 'Induu senni', 'ki': 'Kĩhĩndĩ', 'kk': 'хинди тілі', 'kk-Cyrl': 'хинди тілі', 'kln': 'kutitab Maindiik', 'km': 'ហិណ្ឌី', 'kn': 'ಹಿಂದಿ', 'ko': '힌디어', 'kok': 'हिन्दी', 'kok-Deva': 'हिन्दी', 'ks': 'ہِندی', 'ks-Arab': 'ہِندی', 'ksb': 'Kihindi', 'ksf': 'riíndí', 'ksh': 'Hinndi', 'ku': 'hindî', 'kxv-Deva': 'हिन्दी', 'kxv-Orya': 'ହିନ୍ଦି', 'kxv-Telu': 'హిందీ', 'ky': 'хиндиче', 'lag': 'Kɨhíindi', 'lb': 'Hindi', 'lg': 'Luhindu', 'lkt': 'Hindi Iyápi', 'lld': 'hindi', 'ln': 'lihindi', 'lo': 'ຮິນດິ', 'lrc': 'ھئنی', 'lt': 'hindi', 'lu': 'Hindi', 'luo': 'Kihindi', 'luy': 'Luhindi', 'lv': 'hindi', 'mai': 'हिन्दी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lmoindi', 'mer': 'Kĩhĩndi', 'mfe': 'hindi', 'mg': 'hindi', 'mgh': 'Ihindi', 'mi': 'Hīni', 'mk': 'хинди', 'ml': 'ഹിന്ദി', 'mn': 'хинди', 'mni': 'হিন্দী', 'mni-Beng': 'হিন্দী', 'mr': 'हिंदी', 'ms': 'Hindi', 'mt': 'Hindi', 'mua': 'hindi', 'my': 'ဟိန္ဒီ', 'mzn': 'هندی', 'naq': 'Hindigowab', 'nb': 'hindi', 'nd': 'isi-Hindi', 'ne': 'हिन्दी', 'nl': 'Hindi', 'nmg': 'Kiɛl b’indien', 'nn': 'hindi', 'no': 'hindi', 'nqo': 'ߍ߲ߘߎߞߊ߲', 'nus': 'Thok ɣändini', 'nyn': 'Oruhindi', 'om': 'Afaan Hindii', 'or': 'ହିନ୍ଦୀ', 'pa': 'ਹਿੰਦੀ', 'pa-Guru': 'ਹਿੰਦੀ', 'pcm': 'Híndi Lángwej', 'pl': 'hindi', 'ps': 'هندي', 'pt': 'híndi', 'qu': 'Hindi Simi', 'rm': 'hindi', 'rn': 'Igihindi', 'ro': 'hindi', 'rof': 'Kihindi', 'ru': 'хинди', 'rw': 'Igihindi', 'rwk': 'Kyihindi', 'saq': 'Kihindi', 'sbp': 'Ishihindi', 'sc': 'hindi', 'sd': 'هندي', 'sd-Arab': 'هندي', 'se': 'hindigiella', 'seh': 'hindi', 'ses': 'Induu senni', 'sg': 'Hîndi', 'shi': 'ⵜⴰⵀⵉⵏⴷⵉⵜ', 'shi-Latn': 'Tahindit', 'shi-Tfng': 'ⵜⴰⵀⵉⵏⴷⵉⵜ', 'si': 'හින්දි', 'sk': 'hindčina', 'sl': 'hindijščina', 'smn': 'hindikielâ', 'sn': 'chiHindi', 'so': 'Hindi', 'sq': 'indisht', 'sr': 'хинди', 'sr-Cyrl': 'хинди', 'sr-Latn': 'hindi', 'sv': 'hindi', 'sw': 'Kihindi', 'syr': 'ܗܢܕܝܐ', 'ta': 'இந்தி', 'te': 'హిందీ', 'teo': 'Kihindi', 'tg': 'ҳиндӣ', 'th': 'ฮินดี', 'ti': 'ሂንዲ', 'tk': 'hindi dili', 'to': 'lea fakahinitī', 'tr': 'Hintçe', 'tt': 'һинд', 'twq': 'Induu senni', 'tzm': 'Tahindit', 'ug': 'ھىندىچە', 'uk': 'гінді', 'ur': 'ہندی', 'uz': 'hind', 'uz-Cyrl': 'ҳинди', 'uz-Latn': 'hind', 'vai': 'ꔦꔺ', 'vai-Latn': 'Híiŋdi', 'vai-Vaii': 'ꔦꔺ', 'vec': 'hindi', 'vi': 'Tiếng Hindi', 'vun': 'Kyihindi', 'wae': 'Hindi', 'wo': 'Endo', 'xh': 'IsiHindi', 'xnr': 'हिन्दी', 'xog': 'Oluhindu', 'yav': 'índí', 'yi': 'הינדי', 'yo': 'Èdè Híńdì', 'yrl': 'hĩdi', 'yue': '北印度文', 'yue-Hans': '北印度文', 'yue-Hant': '北印度文', 'zgh': 'ⵜⴰⵀⵉⵏⴷⵉⵜ', 'zh': '印地语', 'zh-Hans': '印地语', 'zh-Hant': '印地文', 'zu': 'isi-Hindi'},
    'hi-Latn': {'af': 'Hingels', 'ak': 'Laatenfoɔ Hindi', 'am': 'ሕንድኛ (ላቲን)', 'ar': 'الهنجليزية', 'as': 'হিংলিছ', 'az': 'Hindi (latın)', 'az-Latn': 'Hindi (latın)', 'bal-Latn': 'Hendi (Látin Engrézi (Látin) Engrézi (Látin, Amrikáay Tepákén Están) syáhag: Látini)', 'be': 'хінгліш', 'be-tarask': 'хінгліш', 'bg': 'хинглиш', 'blo': 'hinɖii kagɩja (latɛŋ kʊja)', 'bn': 'হিংলিশ', 'bs': 'hingleski', 'bs-Latn': 'hingleski', 'ca': 'hinglish', 'ca-ES-valencia': 'hinglish', 'ceb': 'Hinglish', 'chr': 'ᎯᏂᏲᏁᎦ', 'cs': 'hingliš', 'cy': 'Hinglish', 'da': 'hinglish', 'de': 'Hindi (lateinisch)', 'el': 'Hinglish', 'el-polyton': 'Hinglish', 'en': 'Hindi (Latin)', 'es': 'hinglish', 'et': 'hinglishi', 'eu': 'hindia (latindarra)', 'fa': 'هندی (انگلیسی)', 'ff-Adlm': '𞤖𞤭𞤲𞤺𞤭𞤤𞤭𞤧𞤭𞥅𞤪𞤫', 'fi': 'hinglish', 'fil': 'Hinglish', 'fo': 'hinglish', 'fr': 'hinglish', 'gd': 'Hinglish', 'gl': 'hindi (alfabeto latino)', 'gu': 'હિંગ્લિશ', 'ha': 'Harshen Hindi (Latin)', 'he': 'הינגליש', 'hi': 'हिंग्लिश', 'hi-Latn': 'Hindi (Latin)', 'hr': 'hinglish', 'ht': 'hinglish', 'hu': 'hindi (latin)', 'hy': 'հինդի (լատինատառ)', 'ia': 'hinglish', 'id': 'Hinglish', 'ig': 'Hinglish', 'ii': 'ꑴꐛꑱꉙ', 'is': 'hinglish', 'it': 'hinglish', 'ja': 'インド英語 (ヒングリッシュ)', 'ka': 'ჰინგლისური', 'kaa': 'ҳиндше (латын)', 'kaa-Cyrl': 'ҳиндше (латын)', 'kk': 'хинглиш', 'kk-Cyrl': 'хинглиш', 'km': 'ហ៊ីងលីង', 'kn': 'ಹಿಂಗ್ಲಿಷ್', 'ko': '힝글리시', 'kok': 'हिंग्लीश', 'kok-Deva': 'हिंग्लीश', 'ku': 'hîngilîzî', 'kxv': 'hindi', 'kxv-Deva': 'हिन्दी (लातिन)', 'kxv-Latn': 'hindi', 'ky': 'хинглиш (англис тили + хинди)', 'lld': 'hindi (latin)', 'lo': 'ຮິນກລີຊ', 'lv': 'hinglišs', 'mai': 'हिन्दी (ल्याटिन)', 'mi': 'Hīngarihi', 'mk': 'хинди англиски', 'ml': 'ഹിംഗ്ലീഷ്', 'mn': 'хинглиш', 'mr': 'हिंग्लिश', 'ms': 'Hinglish', 'my': 'ဟိန္ဒီ (လက်တင်)', 'nb': 'hinglish', 'ne': 'हिङ्लिस', 'nl': 'Hinglish', 'nn': 'hinglish', 'no': 'hinglish', 'om': 'Hindii (Laatiin)', 'or': 'ହିଂଲିସ', 'pa': 'ਹਿੰਗਲਿਸ਼', 'pa-Guru': 'ਹਿੰਗਲਿਸ਼', 'pcm': 'Híndi (Látin)', 'pl': 'hindi (alfabet łaciński)', 'ps': 'هندي (لاتيني)', 'pt': 'hinglish', 'qu': 'Hinglish Simi', 'ro': 'hinglish', 'ru': 'хинглиш', 'sc': 'hindi (caràteres latinos)', 'si': 'හිංග්ලිෂ්', 'sk': 'hingličtina', 'sl': 'hingleščina', 'so': 'Hindi (Latin)', 'sq': 'hindisht (latine)', 'sr': 'хинглиш', 'sr-Cyrl': 'хинглиш', 'sr-Latn': 'hingliš', 'sv': 'hindi (latinsk)', 'sw': 'Kihindi chenye Kiingereza', 'ta': 'ஹிங்கிலிஷ்', 'te': 'హింగ్లీష్', 'th': 'อังกฤษแบบอินเดีย', 'ti': 'ሂንግሊሽ', 'to': 'lea fakahinitī (fakalatina)', 'tr': 'Hindilizce', 'tt': 'Һинглиш', 'uk': 'гінгліш', 'ur': 'ہنگلش', 'uz': 'hinglish', 'uz-Latn': 'hinglish', 'vec': 'hingleze', 'vi': 'Tiếng Hindi (lai tiếng Anh)', 'wo': 'Hindī', 'xh': 'IsiHindi (Latin)', 'xnr': 'हिंग्लिस', 'yo': 'Èdè Híndì (Látìnì)', 'zh': '印地语（拉丁字母）', 'zh-Hans': '印地语（拉丁字母）', 'zh-Hant': '印地語（拉丁文）'},
    'hia': {'en': 'Lamang'},
    'hib': {'en': 'Hibito'},
    'hid': {'en': 'Hidatsa'},
    'hif': {'ast': 'hindi de Fiji', 'ca': 'hindi de Fiji', 'ca-ES-valencia': 'hindi de Fiji', 'cs': 'hindština (Fidži)', 'de': 'Fidschi-Hindi', 'en': 'Fiji Hindi', 'et': 'Fidži hindi', 'fa': 'هندی فیجیایی', 'fi': 'fidžinhindi', 'fr': 'hindi fidjien', 'ga': 'Hiondúis Fhidsí', 'gd': 'Hindis Fhìditheach', 'gu': 'ફીજી હિંદી', 'hi-Latn': 'Fiji Hindi', 'ht': 'hindi fidjien', 'id': 'Hindi Fiji', 'it': 'hindi figiano', 'ja': 'フィジー・ヒンディー語', 'ko': '피지 힌디어', 'lb': 'Fidschi-Hindi', 'lt': 'Fidžio hindi', 'mk': 'фиџиски хинди', 'nb': 'fijiansk hindi', 'ne': 'फिजी हिन्दी', 'nl': 'Fijisch Hindi', 'nn': 'fijiansk hindi', 'no': 'fijiansk hindi', 'pa': 'ਫਿਜੀ ਹਿੰਦੀ', 'pa-Guru': 'ਫਿਜੀ ਹਿੰਦੀ', 'pl': 'hindi fidżyjskie', 'sv': 'Fiji-hindi', 'ta': 'ஃபிஜி இந்தி', 'th': 'ฮินดีฟิจิ', 'to': 'lea fakahinitī-fisi', 'tr': 'Fiji Hintçesi', 'vi': 'Tiếng Fiji Hindi', 'yi': 'פידזשי הינדי', 'yue': '斐濟印地文', 'yue-Hans': '斐济印地文', 'yue-Hant': '斐濟印地文', 'zh-Hant': '斐濟印地文'},
    'hig': {'en': 'Kamwe'},
    'hih': {'en': 'Pamosu'},
    'hii': {'en': 'Hinduri'},
    'hij': {'en': 'Hijuk'},
    'hik': {'en': 'Seit-Kaitetu'},
    'hil': {'af': 'Hiligaynon', 'am': 'ሂሊጋይኖን', 'ar': 'الهيليجينون', 'as': 'হিলিগায়নোন', 'ast': 'hiligaynon', 'az': 'hiliqaynon', 'az-Cyrl': 'һилигајнон', 'az-Latn': 'hiliqaynon', 'be': 'хілігайнон', 'be-tarask': 'хілігайнон', 'bg': 'хилигайнон', 'bgn': 'هیلیگایونی', 'bn': 'হিলিগ্যায়নোন', 'br': 'hiligaynon', 'brx': 'हिलीगैन’न', 'bs': 'hiligajnon', 'bs-Cyrl': 'хилигајнон', 'bs-Latn': 'hiligajnon', 'ca': 'híligaynon', 'ca-ES-valencia': 'híligaynon', 'ccp': '𑄦𑄨𑄣𑄨𑄉𑄳𑄠𑄠𑄧𑄚𑄮𑄚𑄴', 'ce': 'хилигайнон', 'chr': 'ᎯᎵᎨᎾᏂ', 'ckb': 'هیلیگاینۆن', 'cs': 'hiligajnonština', 'cy': 'Hiligaynon', 'da': 'hiligaynon', 'de': 'Hiligaynon', 'dsb': 'hiligaynonšćina', 'el': 'Χιλιγκαϊνόν', 'el-polyton': 'Χιλιγκαϊνόν', 'en': 'Hiligaynon', 'es': 'hiligaynon', 'et': 'hiligainoni', 'eu': 'hiligaynonera', 'fa': 'هیلی\u200cگاینونی', 'ff-Adlm': '𞤖𞤭𞤤𞤭𞤺𞤢𞤴𞤲𞤮𞤲𞤪𞤫', 'fi': 'hiligaino', 'fil': 'Hiligaynon', 'fo': 'hiligaynon', 'fr': 'hiligaynon', 'fy': 'Hiligaynon', 'ga': 'Hilgeanóinis', 'gd': 'Hiligaynon', 'gl': 'hiligaynon', 'gsw': 'Hiligaynonisch', 'gu': 'હિલિગેનોન', 'ha': 'Hiligaynon', 'he': 'היליגאינון', 'hi': 'हिलिगेनन', 'hi-Latn': 'Hiligaynon', 'hr': 'hiligaynonski', 'hsb': 'hiligaynonšćina', 'ht': 'hiligaynon', 'hu': 'ilokano', 'hy': 'հիլիգայնոն', 'ia': 'hiligaynon', 'id': 'Hiligaynon', 'ig': 'Hiligayanon', 'is': 'híligaínon', 'it': 'ilongo', 'ja': 'ヒリガイノン語', 'jv': 'Hiligainon', 'ka': 'ჰილიგაინონი', 'kgp': 'hirigajnãn', 'kk': 'хилигайнон тілі', 'kk-Cyrl': 'хилигайнон тілі', 'km': 'ហ៊ីលីហ្គេណុន', 'kn': 'ಹಿಲಿಗೇನನ್', 'ko': '헤리가뇬어', 'kok': 'हिलीगायनॉन', 'kok-Deva': 'हिलीगायनॉन', 'ks': 'ہِلیٖگینَن', 'ks-Arab': 'ہِلیٖگینَن', 'ksh': 'Hiligaynon', 'ku': 'hîlîgaynonî', 'ky': 'хилигайнончо', 'lb': 'Hiligaynon-Sprooch', 'lo': 'ຮິຣິໄກນອນ', 'lt': 'hiligainonų', 'lv': 'hiligainonu', 'mai': 'हिलिगेनन', 'mi': 'Hirikaina', 'mk': 'хилигајнонски', 'ml': 'ഹിലിഗയ്നോൺ', 'mn': 'хилигайнон', 'mr': 'हिलीगेनॉन', 'ms': 'Hiligaynon', 'mt': 'Hiligaynon', 'my': 'ဟီလီဂေနွန်', 'nb': 'hiligaynon', 'ne': 'हिलिगायनोन', 'nl': 'Hiligaynon', 'nn': 'hiligaynon', 'no': 'hiligaynon', 'or': 'ହିଲିଗୈନନ୍', 'pa': 'ਹਿਲੀਗੇਨਨ', 'pa-Guru': 'ਹਿਲੀਗੇਨਨ', 'pcm': 'Híligaínọn', 'pl': 'hiligaynon', 'ps': 'ھلیګینون', 'pt': 'hiligaynon', 'qu': 'Hiligaynon Simi', 'rm': 'hiligaynon', 'ro': 'hiligaynon', 'ru': 'хилигайнон', 'sc': 'ilongu', 'sd': 'هلي گيانان', 'sd-Arab': 'هلي گيانان', 'si': 'හිලිගෙනන්', 'sk': 'hiligajnončina', 'sl': 'hiligajnonščina', 'smn': 'hiligainokielâ', 'so': 'Hiligaynon', 'sq': 'hiligajnonisht', 'sr': 'хилигајнонски', 'sr-Cyrl': 'хилигајнонски', 'sr-Latn': 'hiligajnonski', 'sv': 'hiligaynon', 'sw': 'Kihiligaynon', 'ta': 'ஹிலிகாய்னான்', 'te': 'హిలిగెనాన్', 'tg': 'ҳилигайнон', 'th': 'ฮีลีกัยนน', 'ti': 'ሂሊጋይኖን', 'tk': 'hiligaýnon dili', 'to': 'lea fakahilikainoni', 'tr': 'Hiligaynon dili', 'tt': 'хилигайнон', 'ug': 'خىلىگاينونچە', 'uk': 'хілігайнон', 'ur': 'ہالیگینون', 'uz': 'hiligaynon', 'uz-Cyrl': 'хилигайнон', 'uz-Latn': 'hiligaynon', 'vec': 'hiligheno', 'vi': 'Tiếng Hiligaynon', 'wo': 'Hiligaynon', 'xnr': 'हिलिगेनन', 'yo': 'Èdè Hilgayo', 'yrl': 'irigainũ', 'yue': '希利蓋農文', 'yue-Hans': '希利盖农文', 'yue-Hant': '希利蓋農文', 'zh': '希利盖农语', 'zh-Hans': '希利盖农语', 'zh-Hant': '希利蓋農文', 'zu': 'isi-Hiligaynon', 'hil': 'Ilonggo'},
    'him': {'en': 'Himachali languages'},
    'hio': {'en': 'Tsoa'},
    'hir': {'en': 'Himarimã'},
    'hit': {'af': 'Hetities', 'ar': 'الحثية', 'ast': 'hitita', 'az': 'hittit', 'az-Latn': 'hittit', 'bg': 'хитски', 'bn': 'হিট্টিট', 'brx': 'हीत्ती', 'bs': 'hitite', 'bs-Cyrl': 'хитите', 'bs-Latn': 'hitite', 'ca': 'hitita', 'ca-ES-valencia': 'hitita', 'ccp': '𑄦𑄨𑄖𑄨𑄨𑄖𑄴', 'cs': 'chetitština', 'cy': 'Hetheg', 'da': 'hittitisk', 'de': 'Hethitisch', 'el': 'Χιτίτε', 'el-polyton': 'Χιτίτε', 'en': 'Hittite', 'en-Dsrt': '𐐐𐐮𐐻𐐴𐐻', 'es': 'hitita', 'et': 'heti', 'fa': 'هیتی', 'fi': 'heetti', 'fr': 'hittite', 'fy': 'Hettitysk', 'ga': 'Hitis', 'gd': 'Cànan Het', 'gsw': 'Hethitisch', 'gu': 'હિટ્ટિતે', 'he': 'חתית', 'hi': 'हिताइत', 'hi-Latn': 'Hittite', 'hr': 'hetitski', 'ht': 'hittite', 'hu': 'hettita', 'id': 'Hitit', 'is': 'hettitíska', 'it': 'hittite', 'ja': 'ヒッタイト語', 'ka': 'ხეთური', 'kgp': 'hitita', 'kn': 'ಹಿಟ್ಟಿಟೆ', 'ko': '하타이트어', 'ks': 'ہِتایِت', 'ks-Arab': 'ہِتایِت', 'lb': 'Hethitesch', 'lo': 'ຮິດໄຕ', 'lt': 'hititų', 'lv': 'hetu', 'mk': 'хетитски', 'ml': 'ഹിറ്റൈറ്റ്', 'mr': 'हिट्टिते', 'mt': 'Hittite', 'nb': 'hettittisk', 'ne': 'हिट्टिटे', 'nl': 'Hettitisch', 'nn': 'hettittisk', 'no': 'hettittisk', 'or': 'ହିତୀତେ', 'pl': 'hetycki', 'pt': 'hitita', 'rm': 'ettitic', 'ro': 'hitită', 'ru': 'хеттский', 'sk': 'chetitčina', 'sl': 'hetitščina', 'sr': 'хетитски', 'sr-Cyrl': 'хетитски', 'sr-Latn': 'hetitski', 'sv': 'hettitiska', 'sw': 'Kihiti', 'ta': 'ஹிட்டைட்', 'te': 'హిట్టిటే', 'th': 'ฮิตไตต์', 'to': 'lea fakahitite', 'tr': 'Hititçe', 'ug': 'خىتتىتچە', 'uk': 'хітіті', 'vi': 'Tiếng Hittite', 'yrl': 'hitita', 'yue': '赫梯文', 'yue-Hans': '赫梯文', 'yue-Hant': '赫梯文', 'zh': '赫梯语', 'zh-Hans': '赫梯语', 'zh-Hant': '赫梯文'},
    'hiw': {'en': 'Hiw'},
    'hix': {'en': 'Hixkaryána'},
    'hji': {'en': 'Haji'},
    'hka': {'en': 'Kahe'},
    'hke': {'en': 'Hunde'},
    'hkh': {'en': 'Khah'},
    'hkk': {'en': 'Hunjara-Kaina Ke'},
    'hkn': {'en': 'Mel-Khaonh'},
    'hks': {'en': 'Hong Kong Sign Language'},
    'hla': {'en': 'Halia'},
    'hlb': {'en': 'Halbi'},
    'hld': {'en': 'Halang Doan'},
    'hle': {'en': 'Hlersu'},
    'hlt': {'en': 'Matu Chin'},
    'hlu': {'en': 'Hieroglyphic Luwian'},
    'hma': {'en': 'Southern Mashan Hmong'},
    'hmb': {'en': 'Humburi Senni Songhay'},
    'hmc': {'en': 'Central Huishui Hmong'},
    'hmd': {'en': 'Large Flowery Miao'},
    'hme': {'en': 'Eastern Huishui Hmong'},
    'hmf': {'en': 'Hmong Don'},
    'hmg': {'en': 'Southwestern Guiyang Hmong'},
    'hmh': {'en': 'Southwestern Huishui Hmong'},
    'hmi': {'en': 'Northern Huishui Hmong'},
    'hmj': {'en': 'Ge'},
    'hmk': {'en': 'Maek'},
    'hml': {'en': 'Luopohe Hmong'},
    'hmm': {'en': 'Central Mashan Hmong'},
    'hmn': {'af': 'Hmong', 'am': 'ህሞንግ', 'ar': 'الهمونجية', 'as': 'হমং', 'ast': 'hmong', 'az': 'monq', 'az-Cyrl': 'монг', 'az-Latn': 'monq', 'be': 'хмонг', 'be-tarask': 'хмонг', 'bg': 'хмонг', 'bgn': 'همونگی', 'bn': 'হ্\u200cমোঙ', 'br': 'hmong', 'brx': 'ह्मंग', 'bs': 'hmong', 'bs-Cyrl': 'хмонг', 'bs-Latn': 'hmong', 'ca': 'hmong', 'ca-ES-valencia': 'hmong', 'ccp': '𑄦𑄳𑄦𑄟𑄮𑄋𑄴', 'ce': 'хмонг', 'chr': 'ᎭᎼᏂᎩ', 'ckb': 'همۆنگ', 'cs': 'hmongština', 'cy': 'Hmongeg', 'da': 'hmong', 'de': 'Miao', 'dsb': 'hmongšćina', 'el': 'Χμονγκ', 'el-polyton': 'Χμονγκ', 'en': 'Hmong', 'es': 'hmong', 'et': 'hmongi', 'eu': 'hmonga', 'fa': 'همونگ', 'ff-Adlm': '𞤖𞤵𞤥𞤺𞤵𞤲𞤪𞤫', 'fi': 'hmong', 'fil': 'Hmong', 'fo': 'hmong', 'fr': 'hmong', 'fy': 'Hmong', 'ga': 'Hmongais', 'gd': 'Hmong', 'gl': 'hmong', 'gsw': 'Miao', 'gu': 'હમોંગ', 'ha': 'Hmong', 'he': 'המונג', 'hi': 'ह्मॉंग', 'hi-Latn': 'Hmong', 'hnj': '𞄀𞄄𞄰𞄩', 'hnj-Hmnp': '𞄀𞄄𞄰𞄩', 'hr': 'hmong', 'hsb': 'hmongšćina', 'ht': 'hmong', 'hu': 'hmong', 'hy': 'հմոնգ', 'ia': 'hmong', 'id': 'Hmong', 'ig': 'Hmong', 'is': 'hmong', 'it': 'hmong', 'ja': 'フモン語', 'jv': 'Hmong', 'ka': 'ჰმონგი', 'kea': 'hmong', 'kgp': 'hymãg', 'kk': 'хмонг тілі', 'kk-Cyrl': 'хмонг тілі', 'km': 'ម៉ុង', 'kn': 'ಮೋಂಗ್', 'ko': '히몸어', 'kok': 'मोंग', 'kok-Deva': 'मोंग', 'ks': 'ہمونگ', 'ks-Arab': 'ہمونگ', 'ksh': 'Hmong', 'ku': 'hmongî', 'ky': 'хмонгчо', 'lb': 'Miao-Sprooch', 'lo': 'ມອງ', 'lt': 'hmong', 'lv': 'hmongu', 'mai': 'मौंग', 'mi': 'Mōnga', 'mk': 'хмонг', 'ml': 'മോങ്', 'mn': 'хмонг', 'mr': 'माँग', 'ms': 'Hmong', 'mt': 'Hmong', 'my': 'မုံ', 'nb': 'hmong', 'ne': 'हमोङ', 'nl': 'Hmong', 'nn': 'hmong', 'no': 'hmong', 'nqo': 'ߡߐ߲ߜ߭ߑߞߊ߲', 'or': 'ହଁଙ୍ଗ', 'pa': 'ਹਮੋਂਗ', 'pa-Guru': 'ਹਮੋਂਗ', 'pcm': 'Mọ́ng Lángwej', 'pl': 'hmong', 'ps': 'همونګ', 'pt': 'hmong', 'qu': 'Hmong Daw Simi', 'rm': 'hmong', 'ro': 'hmong', 'ru': 'хмонг', 'sc': 'hmong', 'sd': 'مونگ', 'sd-Arab': 'مونگ', 'si': 'මොන්ග්', 'sk': 'hmongčina', 'sl': 'hmonščina', 'smn': 'hmongkielâ', 'so': 'Hamong', 'sq': 'hmongisht', 'sr': 'хмоншки', 'sr-Cyrl': 'хмоншки', 'sr-Latn': 'hmonški', 'sv': 'hmongspråk', 'sw': 'Kihmong', 'ta': 'மாங்க்', 'te': 'మోంగ్', 'th': 'ม้ง', 'ti': 'ህሞንግ', 'tk': 'hmong dili', 'to': 'lea fakamōngi', 'tr': 'Hmong', 'ug': 'مۆڭچە', 'uk': 'хмонг', 'ur': 'ہمانگ', 'uz': 'xmong', 'uz-Cyrl': 'хмонгча', 'uz-Latn': 'xmong', 'vec': 'mong', 'vi': 'Tiếng H’Mông', 'xnr': 'ह्मॉंग', 'yo': 'Hmong', 'yrl': 'himũgi', 'yue': '孟文', 'yue-Hans': '孟文', 'yue-Hant': '孟文', 'zh': '苗语', 'zh-Hans': '苗语', 'zh-Hant': '苗語', 'zu': 'isi-Hmong'},
    'hmn-pro': {'en': 'Proto-Hmong'},
    'hmp': {'en': 'Northern Mashan Hmong'},
    'hmq': {'en': 'Eastern Qiandong Miao'},
    'hmr': {'en': 'Hmar'},
    'hms': {'en': 'Southern Qiandong Miao'},
    'hmt': {'en': 'Hamtai'},
    'hmu': {'en': 'Hamap'},
    'hmv': {'en': 'Hmong Dô'},
    'hmw': {'en': 'Western Mashan Hmong'},
    'hmx': {'en': 'Hmong-Mien languages'},
    'hmx-mie-pro': {'en': 'Proto-Mien'},
    'hmx-pro': {'en': 'Proto-Hmong-Mien'},
    'hmy': {'en': 'Southern Guiyang Hmong'},
    'hmz': {'en': 'Hmong Shua'},
    'hna': {'en': 'Mina (Cameroon)'},
    'hnd': {'en': 'Southern Hindko'},
    'hne': {'en': 'Chhattisgarhi'},
    'hng': {'en': 'Hungu'},
    'hnh': {'en': 'ǁAni'},
    'hni': {'en': 'Hani'},
    'hnj': {'bal-Latn': 'Hmang Njuá', 'en': 'Hmong Njua', 'fi': 'hmong njua', 'gd': 'Hmong Njua', 'hi': 'हमोंग नजुआ', 'hi-Latn': 'Hmong Njua', 'hnj': '𞄀𞄄𞄰𞄩𞄍𞄜𞄰', 'hnj-Hmnp': '𞄀𞄄𞄰𞄩𞄍𞄜𞄰', 'ig': 'Hmong Njua', 'ku': 'hmongîya njuayî', 'sr': 'хмонг нџуа', 'sr-Cyrl': 'хмонг нџуа', 'sr-Latn': 'hmong ndžua', 'ti': 'ህሞንግ ንጁዋ', 'vec': 'mong novo'},
    'hnn': {'en': 'Hanunoo'},
    'hno': {'en': 'Northern Hindko'},
    'hns': {'en': 'Caribbean Hindustani'},
    'hnu': {'en': 'Hung'},
    'ho': {'ar': 'الهيري موتو', 'ast': 'hiri motu', 'az': 'hiri motu', 'az-Latn': 'hiri motu', 'bg': 'хири моту', 'bn': 'হিরি মোতু', 'br': 'hiri motu', 'brx': 'हीरी मोटु', 'bs': 'hiri motu', 'bs-Cyrl': 'хири моту', 'bs-Latn': 'hiri motu', 'ca': 'hiri motu', 'ca-ES-valencia': 'hiri motu', 'ccp': '𑄦𑄪𑄢𑄨 𑄟𑄮𑄖𑄪', 'cs': 'hiri motu', 'da': 'hirimotu', 'de': 'Hiri-Motu', 'el': 'Χίρι Μότου', 'el-polyton': 'Χίρι Μότου', 'en': 'Hiri Motu', 'es': 'hiri motu', 'et': 'hirimotu', 'fa': 'موتویی هیری', 'fi': 'hiri-motu', 'fr': 'hiri motu', 'fy': 'Hiri Motu', 'ga': 'Motúis Hírí', 'gd': 'Hiri Motu', 'gsw': 'Hiri-Motu', 'gu': 'હિરી મોટૂ', 'he': 'הירי מוטו', 'hi': 'हिरी मोटू', 'hi-Latn': 'Hiri Motu', 'hr': 'hiri motu', 'ht': 'hiri motu', 'hu': 'hiri motu', 'id': 'Hiri Motu', 'is': 'hírímótú', 'it': 'hiri motu', 'ja': 'ヒリモツ語', 'kgp': 'hiri motu', 'kn': 'ಹಿರಿ ಮೊಟು', 'ko': '히리 모투어', 'ks': 'ہِری موتوٗ', 'ks-Arab': 'ہِری موتوٗ', 'lb': 'Hiri-Motu', 'lo': 'ຮິຣິໂມຕູ', 'lt': 'hiri motu', 'lv': 'hirimotu', 'mk': 'хири моту', 'ml': 'ഹിരി മോതു', 'mr': 'हिरी मॉटू', 'mt': 'Hiri Motu', 'nb': 'hiri motu', 'ne': 'हिरी मोटु', 'nl': 'Hiri Motu', 'nn': 'hiri motu', 'no': 'hiri motu', 'or': 'ହିରି ମୋଟୁ', 'pl': 'hiri motu', 'pt': 'hiri motu', 'rm': 'hiri motu', 'ro': 'hiri motu', 'ru': 'хиримоту', 'sk': 'hiri motu', 'sl': 'hiri motu', 'sr': 'хири моту', 'sr-Cyrl': 'хири моту', 'sr-Latn': 'hiri motu', 'sv': 'hirimotu', 'ta': 'ஹிரி மோட்டு', 'te': 'హిరి మోటు', 'th': 'ฮีรีโมตู', 'to': 'lea fakahili-motu', 'tr': 'Hiri Motu', 'ug': 'ھىرى موتۇچە', 'uk': 'хірі-моту', 'vi': 'Tiếng Hiri Motu', 'yrl': 'hiri mutu', 'yue': '西里莫圖土文', 'yue-Hans': '西里莫图土文', 'yue-Hant': '西里莫圖土文', 'zh': '希里莫图语', 'zh-Hans': '希里莫图语', 'zh-Hant': '西里莫圖土文'},
    'hoa': {'en': 'Hoava'},
    'hob': {'en': 'Mari (Madang Province)'},
    'hoc': {'en': 'Ho'},
    'hod': {'en': 'Holma'},
    'hoe': {'en': 'Horom'},
    'hoh': {'en': 'Hobyót'},
    'hoi': {'en': 'Holikachuk'},
    'hoj': {'en': 'Hadothi'},
    'hok': {'en': 'Hokan languages'},
    'hol': {'en': 'Holu'},
    'hom': {'en': 'Homa'},
    'hoo': {'en': 'Holoholo'},
    'hop': {'en': 'Hopi'},
    'hor': {'en': 'Horo'},
    'hos': {'en': 'Ho Chi Minh City Sign Language'},
    'hot': {'en': 'Hote'},
    'hov': {'en': 'Hovongan'},
    'how': {'en': 'Honi'},
    'hoy': {'en': 'Holiya'},
    'hoz': {'en': 'Hozo'},
    'hpo': {'en': 'Hpon'},
    'hps': {'en': "Hawai'i Sign Language (HSL)"},
    'hr': {'af': 'Kroaties', 'ak': 'Kurowehyia kasa', 'am': 'ክሮሽያንኛ', 'ar': 'الكرواتية', 'as': 'ক্ৰোৱেচিয়ান', 'ast': 'croata', 'az': 'xorvat', 'az-Arab': 'کروواسجا', 'az-Cyrl': 'хорват', 'az-Latn': 'xorvat', 'bal-Latn': 'Króshiái', 'be': 'харвацкая', 'be-tarask': 'харвацкая', 'bg': 'хърватски', 'bgn': 'کراوتی', 'bn': 'ক্রোয়েশীয়', 'br': 'kroateg', 'brx': 'क्रवेशियान', 'bs': 'hrvatski', 'bs-Cyrl': 'хрватски', 'bs-Latn': 'hrvatski', 'ca': 'croat', 'ca-ES-valencia': 'croat', 'ccp': '𑄇𑄳𑄢𑄮𑄠𑄬𑄥𑄩𑄠𑄧', 'ce': 'хорватийн', 'chr': 'ᎧᎶᎡᏏᏂ', 'ckb': 'كرواتی', 'cs': 'chorvatština', 'cy': 'Croateg', 'da': 'kroatisk', 'de': 'Kroatisch', 'dsb': 'chorwatšćina', 'dz': 'ཀྲོ་ཨེ་ཤི་ཡཱན་ཁ', 'ee': 'kroatiagbe', 'el': 'Κροατικά', 'el-polyton': 'Κροατικά', 'en': 'Croatian', 'en-Dsrt': '𐐗𐑉𐐬𐐩𐑇𐐲𐑌', 'eo': 'kroata', 'es': 'croata', 'et': 'horvaadi', 'eu': 'kroaziera', 'fa': 'کروات', 'ff-Adlm': '𞤑𞤮𞤪𞤮𞤱𞤢𞤧𞤭𞥅𞤪𞤫', 'fi': 'kroatia', 'fil': 'Croatian', 'fo': 'kroatiskt', 'fr': 'croate', 'fur': 'cravuat', 'fy': 'Kroatysk', 'ga': 'Cróitis', 'gd': 'Cròthaisis', 'gl': 'croata', 'gsw': 'Kroazisch', 'gu': 'ક્રોએશિયન', 'ha': 'Kuroshiyan', 'he': 'קרואטית', 'hi': 'क्रोएशियाई', 'hi-Latn': 'Croatian', 'hr': 'hrvatski', 'hsb': 'chorwatšćina', 'ht': 'croate', 'hu': 'horvát', 'hy': 'խորվաթերեն', 'ia': 'croato', 'id': 'Kroasia', 'ig': 'Croatian', 'is': 'króatíska', 'it': 'croato', 'ja': 'クロアチア語', 'jv': 'Kroasia', 'ka': 'ხორვატული', 'kea': 'kroata', 'kgp': 'krovata', 'kk': 'хорват тілі', 'kk-Cyrl': 'хорват тілі', 'km': 'ក្រូអាត', 'kn': 'ಕ್ರೊಯೇಶಿಯನ್', 'ko': '크로아티아어', 'kok': 'क्रोएशियन', 'kok-Deva': 'क्रोएशियन', 'ks': 'کروشِیَن', 'ks-Arab': 'کروشِیَن', 'ksh': 'Krowatesch', 'ku': 'xirwatî', 'kxv': 'kroesiati', 'kxv-Deva': 'क्रोएसियाति', 'kxv-Latn': 'kroesiati', 'kxv-Orya': 'କ୍ରଏସିଆତି', 'kxv-Telu': 'క్రొయేసియన్', 'ky': 'хорватча', 'lb': 'Kroatesch', 'lkt': 'Croatian Iyápi', 'lld': 'croat', 'lo': 'ໂຄຣເອທຽນ', 'lrc': 'کوروڤاتی', 'lt': 'kroatų', 'lv': 'horvātu', 'mai': 'क्रोशियाई', 'mi': 'Koroātiana', 'mk': 'хрватски', 'ml': 'ക്രൊയേഷ്യൻ', 'mn': 'хорват', 'mr': 'क्रोएशियन', 'ms': 'Croatia', 'mt': 'Kroat', 'my': 'ခရိုအေးရှား', 'mzn': 'کرواتی', 'nb': 'kroatisk', 'ne': 'क्रोयसियाली', 'nl': 'Kroatisch', 'nn': 'kroatisk', 'no': 'kroatisk', 'nqo': 'ߞߙߏߥߊߛߌߞߊ߲', 'om': 'Afaan Croatian', 'or': 'କ୍ରୋଏସୀୟ', 'os': 'хорватаг', 'pa': 'ਕ੍ਰੋਏਸ਼ਿਆਈ', 'pa-Guru': 'ਕ੍ਰੋਏਸ਼ਿਆਈ', 'pcm': 'Kroéshia Lángwej', 'pl': 'chorwacki', 'ps': 'کروايشيايي', 'pt': 'croata', 'qu': 'Croata Simi', 'rm': 'croat', 'ro': 'croată', 'ru': 'хорватский', 'rw': 'Igikorowasiya', 'sc': 'croatu', 'sd': 'ڪروشيائي', 'sd-Arab': 'ڪروشيائي', 'se': 'kroátiagiella', 'si': 'කෝඒෂියානු', 'sk': 'chorvátčina', 'sl': 'hrvaščina', 'smn': 'kroatiakielâ', 'so': 'Koro’eeshiyaan', 'sq': 'kroatisht', 'sr': 'хрватски', 'sr-Cyrl': 'хрватски', 'sr-Latn': 'hrvatski', 'sv': 'kroatiska', 'sw': 'Kikorasia', 'ta': 'குரோஷியன்', 'te': 'క్రొయేషియన్', 'tg': 'хорватӣ', 'th': 'โครเอเชีย', 'ti': 'ክሮኤሽያን', 'tk': 'horwat dili', 'to': 'lea fakakuloisia', 'tr': 'Hırvatça', 'tt': 'хорват', 'ug': 'كىرودىچە', 'uk': 'хорватська', 'ur': 'کراتی', 'uz': 'xorvat', 'uz-Cyrl': 'хорватча', 'uz-Latn': 'xorvat', 'vec': 'croà', 'vi': 'Tiếng Croatia', 'wae': 'Kroatiš', 'wo': 'Krowat', 'xnr': 'क्रोएशियाई', 'yi': 'קראאַטיש', 'yo': 'Èdè Kroatia', 'yrl': 'kuruata', 'yue': '克羅埃西亞文', 'yue-Hans': '克罗埃西亚文', 'yue-Hant': '克羅埃西亞文', 'zh': '克罗地亚语', 'zh-Hans': '克罗地亚语', 'zh-Hant': '克羅埃西亞文', 'zu': 'isi-Croatian'},
    'hra': {'en': 'Hrangkhol'},
    'hrc': {'en': 'Niwer Mil'},
    'hre': {'en': 'Hre'},
    'hrk': {'en': 'Haruku'},
    'hrm': {'en': 'Horned Miao'},
    'hro': {'en': 'Haroi'},
    'hrp': {'en': 'Nhirrpi'},
    'hrr': {'en': 'Horuru'},
    'hrt': {'en': 'Hértevin'},
    'hru': {'en': 'Hruso'},
    'hrw': {'en': 'Warwar Feni'},
    'hrx': {'en': 'Hunsrik'},
    'hrz': {'en': 'Harzani'},
    'hsb': {'af': 'Oppersorbies', 'ak': 'Atifi fam Sɔɔbia Kasa', 'am': 'የላይኛው ሶርቢያንኛ', 'ar': 'الصوربية العليا', 'as': 'আপাৰ ছোৰ্বিয়ান', 'ast': 'altu sorbiu', 'az': 'yuxarı sorb', 'az-Cyrl': 'јухары сорб', 'az-Latn': 'yuxarı sorb', 'bal-Latn': 'Borzi Sorbiái', 'be': 'верхнялужыцкая', 'be-tarask': 'верхнялужыцкая', 'bg': 'горнолужишки', 'bgn': 'علیای سیربی', 'bn': 'উচ্চ সোর্বিয়ান', 'br': 'uhelsorabeg', 'brx': 'गोजौआरि सर्बियान', 'bs': 'gornjolužičkosrpski', 'bs-Cyrl': 'горњолужичкосрпски', 'bs-Latn': 'gornjolužičkosrpski', 'ca': 'alt sòrab', 'ca-ES-valencia': 'alt sòrab', 'ccp': '𑄅𑄪𑄉𑄪𑄢𑄬 𑄥𑄮𑄢𑄴𑄥𑄨𑄠𑄚𑄴', 'ce': 'лакхара сербийн', 'chr': 'ᎦᎸᎳᏗᎨ ᏐᏈᎠᏂ', 'ckb': 'سربیی سەروو', 'cs': 'hornolužická srbština', 'cy': 'Sorbeg Uchaf', 'da': 'øvresorbisk', 'de': 'Obersorbisch', 'dsb': 'górnoserbšćina', 'el': 'Άνω Σορβικά', 'el-polyton': 'Άνω Σορβικά', 'en': 'Upper Sorbian', 'es': 'alto sorbio', 'et': 'ülemsorbi', 'eu': 'goi-sorabiera', 'fa': 'صُربی علیا', 'ff-Adlm': '𞤅𞤮𞤪𞤦𞤭𞤴𞤢𞤲𞤪𞤫 𞤁𞤮𞤱𞤪𞤭', 'fi': 'yläsorbi', 'fil': 'Upper Sorbian', 'fo': 'ovara sorbian', 'fr': 'haut-sorabe', 'fy': 'Oppersorbysk', 'ga': 'Sorbais Uachtarach', 'gd': 'Sòrbais Uachdarach', 'gl': 'alto sorbio', 'gsw': 'Obersorbisch', 'gu': 'અપર સોર્બિયન', 'ha': 'Sorbianci ta Sama', 'he': 'סורבית עילית', 'hi': 'ऊपरी सॉर्बियन', 'hi-Latn': 'Upper Sorbian', 'hr': 'gornjolužički', 'hsb': 'hornjoserbšćina', 'ht': 'haut-sorabe', 'hu': 'felső-szorb', 'hy': 'վերին սորբերեն', 'ia': 'alte sorabo', 'id': 'Sorbia Hulu', 'ig': 'Upper Sorbian', 'is': 'hásorbneska', 'it': 'alto sorabo', 'ja': '高地ソルブ語', 'jv': 'Sorbia Standar', 'ka': 'ზემოსორბული', 'kea': 'sórbiu altu', 'kgp': 'soramiv téj', 'kk': 'жоғарғы лужица тілі', 'kk-Cyrl': 'жоғарғы лужица тілі', 'km': 'សូប៊ីលើ', 'kn': 'ಅಪ್ಪರ್ ಸರ್ಬಿಯನ್', 'ko': '고지 소르비아어', 'kok': 'अपर सोर्बियन', 'kok-Deva': 'अपर सोर्बियन', 'ks': 'ہیرِم ساربِیَن', 'ks-Arab': 'ہیرِم ساربِیَن', 'ksh': 'Bovversorrbesch', 'ku': 'sorbîya jorîn', 'ky': 'жогорку сорбианча', 'lb': 'Uewersorbesch', 'lo': 'ຊໍບຽນ ທາງຕອນເໜືອ', 'lrc': 'سوربی ڤارو', 'lt': 'aukštutinių sorbų', 'lv': 'augšsorbu', 'mai': 'उच्च सोर्बियाई', 'mi': 'Hōpiana Maunga', 'mk': 'горнолужички', 'ml': 'അപ്പർ സോർബിയൻ', 'mn': 'дээд сорби', 'mr': 'अप्पर सॉर्बियन', 'ms': 'Sorbian Atas', 'mt': 'Sorbjan ta’ Fuq', 'my': 'ဆက္ကဆိုနီ', 'mzn': 'بالایی صربی', 'nb': 'høysorbisk', 'ne': 'माथिल्लो सोर्बियन', 'nl': 'Oppersorbisch', 'nn': 'høgsorbisk', 'no': 'høysorbisk', 'nqo': 'ߛߏߙߊߓߎ߫ ߛߊ߲ߘߐ߫ ߞߊ߲', 'or': 'ଉପର ସର୍ବିଆନ୍', 'pa': 'ਅੱਪਰ ਸੋਰਬੀਅਨ', 'pa-Guru': 'ਅੱਪਰ ਸੋਰਬੀਅਨ', 'pcm': 'Sóbiá Lángwej di ọ́p-ọ́p wan', 'pl': 'górnołużycki', 'ps': 'پورته سربيايي', 'pt': 'alto sorábio', 'qu': 'Alto Sorbio Simi', 'rm': 'aut sorb', 'ro': 'sorabă de sus', 'ru': 'верхнелужицкий', 'sc': 'sòrabu artu', 'sd': 'اپر سربيائي', 'sd-Arab': 'اپر سربيائي', 'si': 'ඉහළ සෝබියානු', 'sk': 'hornolužická srbčina', 'sl': 'gornja lužiška srbščina', 'smn': 'pajesorbi', 'so': 'Sorobiyaanka Sare', 'sq': 'sorbishte e sipërme', 'sr': 'горњолужичкосрпски', 'sr-Cyrl': 'горњолужичкосрпски', 'sr-Latn': 'gornjolužičkosrpski', 'sv': 'högsorbiska', 'sw': 'Kisobia cha Ukanda wa Juu', 'ta': 'அப்பர் சோர்பியான்', 'te': 'అప్పర్ సోర్బియన్', 'tg': 'сербии болоӣ', 'th': 'ซอร์เบียตอนบน', 'ti': 'ላዕለዋይ ሶርብኛ', 'tk': 'ýokarky lužits dili', 'to': 'lea fakasōpia-hake', 'tr': 'Yukarı Sorbça', 'tt': 'югары сорб', 'ug': 'ئۈستۈن سوربچە', 'uk': 'верхньолужицька', 'ur': 'اپر سربیائی', 'uz': 'yuqori sorb', 'uz-Cyrl': 'юқори сорбча', 'uz-Latn': 'yuqori sorb', 'vec': 'alto sorabo', 'vi': 'Tiếng Thượng Sorbia', 'wo': 'Sorab-Kaw', 'xnr': 'उपरली सॉर्बियन', 'yi': 'אייבער־סארביש', 'yo': 'Sorbian Apá Òkè', 'yrl': 'surábiu iwaté', 'yue': '上索布文', 'yue-Hans': '上索布文', 'yue-Hant': '上索布文', 'zh': '上索布语', 'zh-Hans': '上索布语', 'zh-Hant': '上索布文', 'zu': 'isi-Upper Sorbian'},
    'hsh': {'en': 'Hungarian Sign Language'},
    'hsl': {'en': 'Hausa Sign Language'},
    'hsn': {'af': 'Xiang-Sjinees', 'am': 'ዢያንግ ቻይንኛ', 'ar': 'شيانغ الصينية', 'ast': 'chinu xiang', 'az': 'syan', 'az-Latn': 'syan', 'bn': 'Xiang চীনা', 'br': 'sinaeg Xian', 'ca': 'xinès xiang', 'ca-ES-valencia': 'xinès xiang', 'ccp': 'Xiang 𑄌𑄨𑄚', 'cs': 'čínština (dialekty Xiang)', 'da': 'xiang-kinesisk', 'de': 'Xiang', 'en': 'Xiang Chinese', 'es': 'chino xiang', 'et': 'sjangi', 'fi': 'xiang-kiina', 'fo': 'xiang kinesiskt', 'fr': 'xiang', 'ga': 'Sínis Xiang', 'gd': 'Xiang', 'he': 'סינית שיאנג', 'hi-Latn': 'Xiang Chinese', 'hr': 'xiang kineski', 'ht': 'xiang', 'hu': 'xiang kínai', 'hy': 'սյան չինարեն', 'it': 'xiang', 'ja': '湘語', 'kgp': 'sijỹg', 'kn': 'ಶಯಾಂಗ್ ಚೀನೀಸೇ', 'ko': '샹어', 'ky': 'Сянь Кытайча', 'lb': 'Xiang-Chinesesch', 'lt': 'kinų kalbos hunano tarmė', 'mk': 'сјанг', 'ml': 'ഷ്യാങ് ചൈനീസ്', 'mr': 'शियांग चिनी', 'ms': 'Cina Xiang', 'nb': 'xiang', 'ne': 'जियाङ चिनियाँ', 'nl': 'Xiangyu', 'nn': 'xiang', 'no': 'xiang', 'pa': 'ਚੀਨੀ ਜ਼ਿਆਂਗ', 'pa-Guru': 'ਚੀਨੀ ਜ਼ਿਆਂਗ', 'pl': 'xiang', 'pt': 'xiang', 'ro': 'chineză xiang', 'ru': 'сян', 'si': 'සියැන් චීන', 'sv': 'xiang', 'ta': 'சியாங்க் சீனம்', 'te': 'జియాంగ్ చైనీస్', 'th': 'จีนเซียง', 'to': 'lea fakasiaina-siangi', 'tr': 'Xiang Çincesi', 'uk': 'сянська китайська', 'vi': 'Tiếng Tương', 'yrl': 'xiãgi', 'yue': '湘語', 'yue-Hans': '湘语', 'yue-Hant': '湘語', 'zh': '湘语', 'zh-Hans': '湘语', 'zh-Hant': '湘語', 'zu': 'isi-Xiang Chinese'},
    'hss': {'en': 'Harsusi'},
    'ht': {'af': 'Haïtiaans', 'am': 'ሃይትኛ ክሮሌ', 'ar': 'الكريولية الهايتية', 'as': 'হেইটিয়ান ক্ৰিয়ল', 'ast': 'haitianu', 'az': 'haiti kreol', 'az-Cyrl': 'һаити креол', 'az-Latn': 'haiti kreol', 'be': 'гаіцянская крэольская', 'be-tarask': 'гаіцянская крэольская', 'bg': 'хаитянски креолски', 'bgn': 'کریول آییسینی', 'bn': 'হাইতিয়ান ক্রেওল', 'br': 'haitieg', 'brx': 'हाईटियान क्रिअल', 'bs': 'haićanski kreolski', 'bs-Cyrl': 'хаићански креолски', 'bs-Latn': 'haićanski kreolski', 'ca': 'crioll d’Haití', 'ca-ES-valencia': 'crioll d’Haití', 'ccp': '𑄦𑄭𑄖𑄨𑄠𑄚𑄴', 'ce': 'гаитийн', 'chr': 'ᎮᏏᎠᏂ ᏟᏲᎵ', 'ckb': 'کریولی هائیتی', 'cs': 'haitština', 'cy': 'Creol Haiti', 'da': 'haitisk', 'de': 'Haiti-Kreolisch', 'dsb': 'haitišćina', 'dz': 'ཧེ་ཏི་ཡཱན་ཁ', 'ee': 'haitigbe', 'el': 'Αϊτιανά', 'el-polyton': 'Αϊτιανά', 'en': 'Haitian Creole', 'en-Dsrt': '𐐐𐐩𐑇𐐲𐑌', 'eo': 'haitia kreola', 'es': 'criollo haitiano', 'et': 'haiti', 'eu': 'Haitiko kreolera', 'fa': 'هائیتیایی', 'ff-Adlm': '𞤀𞤳𞤵𞥅𞤪𞤫 𞤖𞤢𞤴𞤼𞤭𞥅', 'fi': 'haiti', 'fil': 'Haitian', 'fo': 'haitiskt creole', 'fr': 'créole haïtien', 'fur': 'haitian', 'fy': 'Haïtiaansk', 'ga': 'Críol Háítí', 'gd': 'Crìtheol Haidhti', 'gl': 'crioulo haitiano', 'gsw': 'Haitisch', 'gu': 'હૈતિઅન ક્રેઓલે', 'ha': 'Haitian Creole', 'he': 'קריאולית (האיטי)', 'hi': 'हैतियाई', 'hi-Latn': 'Haitian Creole', 'hr': 'haićanski kreolski', 'hsb': 'haitišćina', 'ht': 'Kreyòl Ayisyen', 'hu': 'haiti kreol', 'hy': 'խառնակերտ հայիթերեն', 'ia': 'creolo haitian', 'id': 'Kreol Haiti', 'ig': 'Haịtịan ndị Cerọle', 'is': 'haítíska', 'it': 'creolo haitiano', 'ja': 'ハイチ・クレオール語', 'jv': 'Kreol Haiti', 'ka': 'ჰაიტიური კრეოლი', 'kea': 'aitianu', 'kgp': 'hajtijỹnũ', 'kk': 'гаити тілі', 'kk-Cyrl': 'гаити тілі', 'km': 'ហៃទី', 'kn': 'ಹೈಟಿಯನ್ ಕ್ರಿಯೋಲಿ', 'ko': '아이티어', 'kok': 'हैतियन क्रेयॉल', 'kok-Deva': 'हैतियन क्रेयॉल', 'ks': 'ہیتِیاں', 'ks-Arab': 'ہیتِیاں', 'ksh': 'Ha’iihtesch', 'ku': 'haîtî', 'ky': 'гаитиче', 'lb': 'Haitianesch', 'lkt': 'Haiti Iyápi', 'lo': 'ໄຮຕຽນ', 'lrc': 'ھاییتی', 'lt': 'Haičio', 'lv': 'haitiešu', 'mai': 'हैतियाई क्रियोल', 'mi': 'Kereō Haiti', 'mk': 'хаитски', 'ml': 'ഹെയ്\u200cതിയൻ ക്രിയോൾ', 'mn': 'Гаитийн креол', 'mr': 'हैतीयन क्रेओल', 'ms': 'Kreol Haiti', 'mt': 'Creole ta’ Haiti', 'my': 'ဟေတီ', 'mzn': 'هائتیایی', 'nb': 'haitisk', 'ne': 'हैटियाली क्रियोल', 'nl': 'Haïtiaans Creools', 'nn': 'haitisk', 'no': 'haitisk', 'nqo': 'ߤߊߦߌߕߌ߫ ߕߊ߬ߓߎ߰ߛߌ߬ߞߊ߲', 'or': 'ହୈତାୟିନ୍', 'pa': 'ਹੈਤੀਆਈ', 'pa-Guru': 'ਹੈਤੀਆਈ', 'pcm': 'Haítí Kriol', 'pl': 'kreolski haitański', 'ps': 'هيټي کريول', 'pt': 'haitiano', 'qu': 'Haitiano Criollo Simi', 'rm': 'creol haitian', 'ro': 'haitiană', 'ru': 'гаитянский', 'sc': 'crèolu haitianu', 'sd': 'هيٽي ڪرولي', 'sd-Arab': 'هيٽي ڪرولي', 'se': 'haitigiella', 'si': 'හයිටි', 'sk': 'haitská kreolčina', 'sl': 'haitijska kreolščina', 'smn': 'Haiti kreoli', 'so': 'Heeytiyaan Karawle', 'sq': 'kreolishte e Haitit', 'sr': 'хаићански', 'sr-Cyrl': 'хаићански', 'sr-Latn': 'haićanski', 'sv': 'haitiska', 'sw': 'Kihaiti', 'ta': 'ஹைத்தியன் க்ரியோலி', 'te': 'హైటియన్ క్రియోల్', 'tg': 'гаитии креолӣ', 'th': 'เฮติครีโอล', 'ti': 'ክርዮል ሃይትኛ', 'tk': 'gaiti kreol dili', 'to': 'lea fakahaiti', 'tr': 'Haiti Kreyolu', 'tt': 'гаити креол', 'ug': 'ھايتىچە', 'uk': 'гаїтянська креольська', 'ur': 'ہیتی', 'uz': 'gaityan', 'uz-Cyrl': 'гаитянча', 'uz-Latn': 'gaityan', 'vec': 'haitian', 'vi': 'Tiếng Haiti', 'wae': 'Haitianiš', 'wo': 'Kereyolu Ayti', 'xnr': 'हैतियाई', 'yo': 'Haitian Creole', 'yrl': 'haitianu', 'yue': '海地文', 'yue-Hans': '海地文', 'yue-Hant': '海地文', 'zh': '海地克里奥尔语', 'zh-Hans': '海地克里奥尔语', 'zh-Hant': '海地文', 'zu': 'isi-Haitian'},
    'hti': {'en': 'Hoti'},
    'hto': {'en': 'Minica Huitoto'},
    'hts': {'en': 'Hadza'},
    'htu': {'en': 'Hitu'},
    'htx': {'en': 'Middle Hittite'},
    'hu': {'af': 'Hongaars', 'agq': 'Hɔŋgalìa', 'ak': 'Hangri kasa', 'am': 'ሀንጋሪኛ', 'ar': 'الهنغارية', 'as': 'হাঙ্গেৰিয়ান', 'asa': 'Kihungari', 'ast': 'húngaru', 'az': 'macar', 'az-Cyrl': 'маҹар', 'az-Latn': 'macar', 'bal-Latn': 'Hangári', 'bas': 'Hɔp u hɔŋgrìi', 'be': 'венгерская', 'be-tarask': 'венгерская', 'bem': 'Ichi Hangarian', 'bez': 'Hihungari', 'bg': 'унгарски', 'bgn': 'مجارستانی', 'bm': 'oŋirikan', 'bn': 'হাঙ্গেরীয়', 'br': 'hungareg', 'brx': 'हांगारियान', 'bs': 'mađarski', 'bs-Cyrl': 'мађарски', 'bs-Latn': 'mađarski', 'ca': 'hongarès', 'ca-ES-valencia': 'hongarès', 'ccp': '𑄦𑄁𑄉𑄬𑄢𑄩𑄠𑄧', 'ce': 'венгрийн', 'cgg': 'Oruhangare', 'chr': 'ᎲᏂᎦᎵᎠᏂ', 'ckb': 'هەنگاری', 'cs': 'maďarština', 'cy': 'Hwngareg', 'da': 'ungarsk', 'dav': 'Kihungari', 'de': 'Ungarisch', 'dje': 'Hungaari senni', 'dsb': 'hungoršćina', 'dyo': 'ongrua', 'dz': 'ཧཱང་གྷ་རི་ཡཱན་ཁ', 'ebu': 'Kĩhungari', 'ee': 'hungarigbe', 'el': 'Ουγγρικά', 'el-polyton': 'Οὑγγρικά', 'en': 'Hungarian', 'eo': 'hungara', 'es': 'húngaro', 'et': 'ungari', 'eu': 'hungariera', 'ewo': 'Ǹkɔ́bɔ ungárían', 'fa': 'مجاری', 'ff': 'Hongariire', 'ff-Adlm': '𞤖𞤵𞤲𞤺𞤢𞤪𞤭𞤴𞤢𞤲𞤪𞤫', 'ff-Latn': 'Hongariire', 'fi': 'unkari', 'fil': 'Hungarian', 'fo': 'ungarskt', 'fr': 'hongrois', 'fur': 'ongjarês', 'fy': 'Hongaarsk', 'ga': 'Ungáiris', 'gd': 'Ungairis', 'gl': 'húngaro', 'gsw': 'Ungarisch', 'gu': 'હંગેરિયન', 'guz': 'Kihungari', 'ha': 'Harshen Hungari', 'he': 'הונגרית', 'hi': 'हंगेरियाई', 'hi-Latn': 'Hungarian', 'hr': 'mađarski', 'hsb': 'madźaršćina', 'ht': 'hongrois', 'hu': 'magyar', 'hy': 'հունգարերեն', 'ia': 'hungaro', 'id': 'Hungaria', 'ie': 'hungarian', 'ig': 'Hungarian', 'is': 'ungverska', 'it': 'ungherese', 'ja': 'ハンガリー語', 'jmc': 'Kyihungari', 'jv': 'Hungaria', 'ka': 'უნგრული', 'kab': 'Tahungarit', 'kam': 'Kihungari', 'kde': 'Chihungali', 'kea': 'úngaru', 'kgp': 'ũgaru', 'khq': 'Hungaari senni', 'ki': 'Kihungari', 'kk': 'венгр тілі', 'kk-Cyrl': 'венгр тілі', 'kln': 'kutitab Hangari', 'km': 'ហុងគ្រី', 'kn': 'ಹಂಗೇರಿಯನ್', 'ko': '헝가리어', 'kok': 'हंगेरियन', 'kok-Deva': 'हंगेरियन', 'ks': 'ہَنگیریَن', 'ks-Arab': 'ہَنگیریَن', 'ksb': 'Kihungai', 'ksf': 'riɔngrɔá', 'ksh': 'Unnjarresch', 'ku': 'mecarî', 'kxv': 'hngeriyati', 'kxv-Deva': 'हंगेरियाति', 'kxv-Latn': 'hngeriyati', 'kxv-Orya': 'ହଙ୍ଗେରୀୟାତି', 'kxv-Telu': 'హంగేరియన్', 'ky': 'венгерче', 'lag': 'Kɨhungári', 'lb': 'Ungaresch', 'lg': 'Luhangare', 'lkt': 'Hungary Iyápi', 'lld': 'ungaresc', 'ln': 'liongili', 'lo': 'ຮັງກາຣຽນ', 'lrc': 'مأجاری', 'lt': 'vengrų', 'lu': 'Hongili', 'luo': 'Kihungari', 'luy': 'Kihungari', 'lv': 'ungāru', 'mai': 'हंगेरियन', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lhungari', 'mer': 'Kĩhangarĩ', 'mfe': 'hongrwa', 'mg': 'hongroà', 'mgh': 'Ihungari', 'mi': 'Hanekari', 'mk': 'унгарски', 'ml': 'ഹംഗേറിയൻ', 'mn': 'мажар', 'mr': 'हंगेरियन', 'ms': 'Hungary', 'mt': 'Ungeriż', 'mua': 'hungariya', 'my': 'ဟန်ဂေရီ', 'mzn': 'مجاری', 'naq': 'Hungariaǁî gowab', 'nb': 'ungarsk', 'nd': 'isi-Hangari', 'ne': 'हङ्गेरियाली', 'nl': 'Hongaars', 'nmg': 'Kiɛl b’ɔ́ngrois', 'nn': 'ungarsk', 'no': 'ungarsk', 'nqo': 'ߤߐ߲ߜ߭ߙߌߞߊ߲', 'nus': 'Thok ɣänga̱a̱riɛni', 'nyn': 'Oruhangare', 'om': 'Afaan Hangaari', 'or': 'ହଙ୍ଗେରୀୟ', 'os': 'венгериаг', 'pa': 'ਹੰਗਰੀਆਈ', 'pa-Guru': 'ਹੰਗਰੀਆਈ', 'pcm': 'Họngári Lángwej', 'pl': 'węgierski', 'ps': 'هنګري', 'pt': 'húngaro', 'qu': 'Hungaro Simi', 'rm': 'ungarais', 'rn': 'Ikinyahongiriya', 'ro': 'maghiară', 'rof': 'Kihungari', 'ru': 'венгерский', 'rw': 'Igihongiriya', 'rwk': 'Kyihungari', 'sah': 'Бэҥгиэрдии', 'saq': 'Kihungari', 'sbp': 'Ishihungali', 'sc': 'ungheresu', 'sd': 'هنگري', 'sd-Arab': 'هنگري', 'se': 'ungárgiella', 'seh': 'húngaro', 'ses': 'Hungaari senni', 'sg': 'Hongruäa', 'shi': 'ⵜⴰⵀⵏⵖⴰⵔⵉⵜ', 'shi-Latn': 'Tahnɣarit', 'shi-Tfng': 'ⵜⴰⵀⵏⵖⴰⵔⵉⵜ', 'si': 'හන්ගේරියානු', 'sk': 'maďarčina', 'sl': 'madžarščina', 'smn': 'uŋgarkielâ', 'sn': 'chiHungari', 'so': 'Hangariyaan', 'sq': 'hungarisht', 'sr': 'мађарски', 'sr-Cyrl': 'мађарски', 'sr-Latn': 'mađarski', 'sv': 'ungerska', 'sw': 'Kihungaria', 'ta': 'ஹங்கேரியன்', 'te': 'హంగేరియన్', 'teo': 'Kihungari', 'tg': 'маҷорӣ', 'th': 'ฮังการี', 'ti': 'ሃንጋርኛ', 'tk': 'wenger dili', 'to': 'lea fakahungakalia', 'tr': 'Macarca', 'tt': 'венгр', 'twq': 'Hungaari senni', 'tzm': 'Tahenɣarit', 'ug': 'ۋېنگىرچە', 'uk': 'угорська', 'ur': 'ہنگیرین', 'uz': 'venger', 'uz-Cyrl': 'венгерча', 'uz-Latn': 'venger', 'vai': 'ꖽꔟꗸꘋ', 'vai-Latn': 'Hɔŋgérɛŋ', 'vai-Vaii': 'ꖽꔟꗸꘋ', 'vec': 'ongareze', 'vi': 'Tiếng Hungary', 'vun': 'Kyihungari', 'wae': 'Ungariš', 'wo': 'Ongruwaa', 'xnr': 'हंगेरियाई', 'xog': 'Oluhangare', 'yav': 'ɔ́ŋgɛ', 'yi': 'אונגעריש', 'yo': 'Èdè Hungaria', 'yrl': 'ũgaru', 'yue': '匈牙利文', 'yue-Hans': '匈牙利文', 'yue-Hant': '匈牙利文', 'zgh': 'ⵜⴰⵀⵏⵖⴰⵔⵉⵜ', 'zh': '匈牙利语', 'zh-Hans': '匈牙利语', 'zh-Hant': '匈牙利文', 'zu': 'isi-Hungarian'},
    'hub': {'en': 'Huambisa'},
    'huc': {'en': 'ǂHua'},
    'hud': {'en': 'Huaulu'},
    'hue': {'en': 'San Francisco Del Mar Huave'},
    'huf': {'en': 'Humene'},
    'hug': {'en': 'Huachipaeri'},
    'huh': {'en': 'Huilliche'},
    'hui': {'en': 'Huli'},
    'huj': {'en': 'Northern Guiyang Hmong'},
    'huk': {'en': 'Hulung'},
    'hul': {'en': 'Hula'},
    'hum': {'en': 'Hungana'},
    'huo': {'en': 'Hu'},
    'hup': {'af': 'Hupa', 'am': 'ሁፓ', 'ar': 'الهبا', 'as': 'হুপা', 'ast': 'hupa', 'az': 'hupa', 'az-Cyrl': 'һупа', 'az-Latn': 'hupa', 'be': 'хупа', 'be-tarask': 'хупа', 'bg': 'хупа', 'bgn': 'هوپی', 'bn': 'হুপা', 'br': 'hupa', 'brx': 'हूपा', 'bs': 'hupa', 'bs-Cyrl': 'хупа', 'bs-Latn': 'hupa', 'ca': 'hupa', 'ca-ES-valencia': 'hupa', 'ccp': '𑄦𑄪𑄛', 'ce': 'хупа', 'chr': 'ᎠᏂᎱᏆ', 'ckb': 'هووپا', 'cs': 'hupa', 'cy': 'Hupa', 'da': 'hupa', 'de': 'Hupa', 'dsb': 'hupašćina', 'el': 'Χούπα', 'el-polyton': 'Χούπα', 'en': 'Hupa', 'es': 'hupa', 'et': 'hupa', 'eu': 'hupera', 'fa': 'هوپا', 'ff-Adlm': '𞤖𞤵𞤨𞤢𞥄𞤪𞤫', 'fi': 'hupa', 'fil': 'Hupa', 'fo': 'hupa', 'fr': 'hupa', 'fy': 'Hupa', 'ga': 'Húipis', 'gd': 'Hupa', 'gl': 'hupa', 'gsw': 'Hupa', 'gu': 'હૂપા', 'ha': 'Hupa', 'he': 'הופה', 'hi': 'हूपा', 'hi-Latn': 'Hupa', 'hr': 'hupa', 'hsb': 'hupašćina', 'ht': 'hupa', 'hu': 'hupa', 'hy': 'հուպա', 'ia': 'hupa', 'id': 'Hupa', 'ig': 'Hupa', 'is': 'húpa', 'it': 'hupa', 'ja': 'フパ語', 'jv': 'Hupa', 'ka': 'ჰუპა', 'kgp': 'hupa', 'kk': 'хупа тілі', 'kk-Cyrl': 'хупа тілі', 'km': 'ហ៊ូប៉ា', 'kn': 'ಹೂಪಾ', 'ko': '후파어', 'kok': 'हुपा', 'kok-Deva': 'हुपा', 'ks': 'ہُپا', 'ks-Arab': 'ہُپا', 'ksh': 'Hupa', 'ku': 'hupayî', 'ky': 'хупача', 'lb': 'Hupa', 'lo': 'ຮູປາ', 'lt': 'hupa', 'lv': 'hupu', 'mai': 'हुपा', 'mi': 'Hupa', 'mk': 'хупа', 'ml': 'ഹൂപ', 'mn': 'хупа', 'mr': 'हूपा', 'ms': 'Hupa', 'mt': 'Hupa', 'my': 'ဟူပါ', 'nb': 'hupa', 'ne': 'हुपा', 'nl': 'Hupa', 'nn': 'hupa', 'no': 'hupa', 'or': 'ହୁପା', 'pa': 'ਹੂਪਾ', 'pa-Guru': 'ਹੂਪਾ', 'pcm': 'Húpá Lángwej', 'pl': 'hupa', 'ps': 'ھوپا', 'pt': 'hupa', 'qu': 'Hupa Simi', 'rm': 'hupa', 'ro': 'hupa', 'ru': 'хупа', 'sc': 'hupa', 'sd': 'هوپا', 'sd-Arab': 'هوپا', 'si': 'හුපා', 'sk': 'hupčina', 'sl': 'hupa', 'smn': 'hupakielâ', 'so': 'Hupa', 'sq': 'hupaisht', 'sr': 'хупа', 'sr-Cyrl': 'хупа', 'sr-Latn': 'hupa', 'sv': 'hupa', 'sw': 'Hupa', 'ta': 'ஹுபா', 'te': 'హుపా', 'th': 'ฮูปา', 'ti': 'ሁፓ', 'tk': 'hupa', 'to': 'lea fakahupa', 'tr': 'Hupaca', 'ug': 'خۇپاچە', 'uk': 'хупа', 'ur': 'ہیوپا', 'uz': 'xupa', 'uz-Cyrl': 'хупа тили', 'uz-Latn': 'xupa', 'vec': 'hupa', 'vi': 'Tiếng Hupa', 'xnr': 'हूपा', 'yo': 'Èdè Hupa', 'yrl': 'hupa', 'yue': '胡帕文', 'yue-Hans': '胡帕文', 'yue-Hant': '胡帕文', 'zh': '胡帕语', 'zh-Hans': '胡帕语', 'zh-Hant': '胡帕文', 'zu': 'isi-Hupa'},
    'huq': {'en': 'Tsat'},
    'hur': {'af': 'Halkomelem', 'am': 'ሃልኮመልም', 'ar': 'الهالكوميليم', 'as': 'হেলকোমেলেম', 'az': 'Halkomelem', 'az-Latn': 'Halkomelem', 'be': 'халкамелем', 'be-tarask': 'халкамелем', 'bg': 'халкомелем', 'bn': 'হাল্কোমেলেম', 'br': 'halkomelemeg', 'brx': 'हाल्क’मेलेम', 'bs': 'halkomelem', 'bs-Cyrl': 'халкомелем', 'bs-Latn': 'halkomelem', 'ca': 'halkomelem', 'ca-ES-valencia': 'halkomelem', 'chr': 'ᎭᎵᎪᎺᎴᎻ', 'cs': 'halkomelem', 'cy': 'Halkomelem', 'da': 'halkomelem', 'de': 'Halkomelem', 'dsb': 'halkomelemšćina', 'el': 'Χαλκομελέμ', 'el-polyton': 'Χαλκομελέμ', 'en': 'Halkomelem', 'es': 'halkomelem', 'et': 'halkomelemi', 'eu': 'halkomelema', 'fa': 'هالکوملم', 'ff-Adlm': '𞤖𞤢𞤤𞤳𞤮𞤥𞤫𞤤𞤫𞤥𞤪𞤫', 'fi': 'halkomelem', 'fil': 'Halkomelem', 'fr': 'halkomelem', 'ga': 'Halkomelem', 'gd': 'Halkomelem', 'gl': 'halkomelem', 'gu': 'હેલ્કોમેલેમ', 'ha': 'Halkomelem', 'he': 'הלקומלם', 'hi': 'हल्कोमेलम', 'hi-Latn': 'Halkomelem', 'hr': 'halkomelem', 'hsb': 'halkomelemšćina', 'ht': 'halkomelem', 'hu': 'halkomelem', 'hy': 'հալքոմելեմ', 'ia': 'halkomelem', 'id': 'Halkomelem', 'ig': 'Halkomelem', 'is': 'halkomelem', 'it': 'halkomelem', 'ja': 'ハルコメレム語', 'jv': 'Halkomelem', 'ka': 'ჰალკომელემი', 'kk': 'халкомелем тілі', 'kk-Cyrl': 'халкомелем тілі', 'km': 'ហាល់កូម៉េឡេម', 'kn': 'ಹಾಲ್ಕೊಮೆಲೆಮ್', 'ko': '할코멜렘어', 'kok': 'हाल्कोमेलेम', 'kok-Deva': 'हाल्कोमेलेम', 'ku': 'halkomelemî', 'ky': 'халкомелемче (индей тили)', 'lo': 'ຮາລໂກເມລຽມ', 'lt': 'halkomelemų', 'lv': 'halkomelenu', 'mai': 'हलकोमैलेम', 'mi': 'Hākomerema', 'mk': 'халкомелем', 'ml': 'ഹോക്കൊമെലം', 'mn': 'халкомелем', 'mr': 'हॉल्कमेलम', 'ms': 'Halkomelem', 'my': 'ဟော့ကမိုလွမ်', 'nb': 'halkomelem', 'ne': 'हाल्कोमेलेम', 'nl': 'Halkomelem', 'nn': 'halkomelem', 'no': 'halkomelem', 'or': 'ହାଲକୋମେଲେମ', 'pa': 'ਹਾਲਕੋਮੇਲਮ', 'pa-Guru': 'ਹਾਲਕੋਮੇਲਮ', 'pcm': 'Halkomẹ́lẹ́m Lángwej', 'pl': 'halkomelem', 'ps': 'هلکومیلم', 'pt': 'halkomelem', 'qu': 'Halkomelem Simi', 'ro': 'halkomelem', 'ru': 'халкомелем', 'sc': 'halkomelem', 'sd': 'ھاڪملم', 'sd-Arab': 'ھاڪملم', 'si': 'හල්කොමලෙම්', 'sk': 'halkomelem', 'sl': 'halkomelenščina', 'so': 'Halkomelem', 'sq': 'halkemejlemisht', 'sr': 'халкомелем', 'sr-Cyrl': 'халкомелем', 'sr-Latn': 'halkomelem', 'sv': 'halkomelem', 'sw': 'Kihalkomelemi', 'ta': 'ஹல்கோமெலம்', 'te': 'హల్కోమెలెమ్', 'th': 'ฮัลโกเมเลม', 'ti': 'ሃልኮመለም', 'tk': 'halkomelem dili', 'to': 'lea fakahalikomele', 'tr': 'Halkomelemce', 'uk': 'галкомелем', 'ur': 'ہالکومیلم', 'uz': 'halkomelem', 'uz-Latn': 'halkomelem', 'vec': 'halkomelem', 'vi': 'Tiếng Halkomelem', 'yo': 'Èdè Hakomelemi', 'yue': '哈爾魁梅林語', 'yue-Hans': '哈尔科梅勒姆文', 'yue-Hant': '哈爾魁梅林語', 'zh': '哈尔魁梅林语', 'zh-Hans': '哈尔魁梅林语', 'zh-Hant': '哈爾魁梅林語', 'zu': 'isi-Halkomelem'},
    'hus': {'en': 'Huastec'},
    'hut': {'en': 'Humla'},
    'huu': {'en': 'Murui Huitoto'},
    'huv': {'en': 'San Mateo Del Mar Huave'},
    'huw': {'en': 'Hukumina'},
    'hux': {'en': 'Nüpode Huitoto'},
    'huy': {'en': 'Hulaulá'},
    'huz': {'en': 'Hunzib'},
    'hvc': {'en': 'Haitian Vodoun Culture Language'},
    'hve': {'en': 'San Dionisio Del Mar Huave'},
    'hvk': {'en': 'Haveke'},
    'hvn': {'en': 'Sabu'},
    'hvv': {'en': 'Santa María Del Mar Huave'},
    'hwa': {'en': 'Wané'},
    'hwc': {'en': "Hawai'i Creole English"},
    'hwo': {'en': 'Hwana'},
    'hy': {'af': 'Armeens', 'ak': 'Aameniani', 'am': 'አርሜንኛ', 'ar': 'الأرمنية', 'as': 'আৰ্মেনীয়', 'ast': 'armeniu', 'az': 'erməni', 'az-Arab': 'ائرمنیجه', 'az-Cyrl': 'ермәни', 'az-Latn': 'erməni', 'bal-Latn': 'Arminiái', 'be': 'армянская', 'be-tarask': 'армянская', 'bg': 'арменски', 'bgn': 'ارمنی', 'bn': 'আর্মেনিয়', 'br': 'armenianeg', 'brx': 'आर्मेनियान', 'bs': 'armenski', 'bs-Cyrl': 'јерменски', 'bs-Latn': 'armenski', 'ca': 'armeni', 'ca-ES-valencia': 'armeni', 'ccp': '𑄃𑄢𑄴𑄟𑄬𑄚𑄨𑄠𑄧', 'ce': 'эрмалойн', 'chr': 'ᎠᎳᎻᎠᏂ', 'ckb': 'ئەرمەنی', 'cs': 'arménština', 'cy': 'Armeneg', 'da': 'armensk', 'de': 'Armenisch', 'dsb': 'armeńšćina', 'dz': 'ཨར་མི་ནི་ཡཱན་ཁ', 'ee': 'armeniagbe', 'el': 'Αρμενικά', 'el-polyton': 'Ἀρμενικά', 'en': 'Armenian', 'en-Dsrt': '𐐂𐑉𐑋𐐨𐑌𐐨𐐲𐑌', 'eo': 'armena', 'es': 'armenio', 'et': 'armeenia', 'eu': 'armeniera', 'fa': 'ارمنی', 'ff-Adlm': '𞤀𞤪𞤥𞤫𞤲𞤭𞥅𞤪𞤫', 'fi': 'armenia', 'fil': 'Armenian', 'fo': 'armenskt', 'fr': 'arménien', 'fur': 'armen', 'fy': 'Armeensk', 'ga': 'Airméinis', 'gd': 'Airmeinis', 'gl': 'armenio', 'gsw': 'Armenisch', 'gu': 'આર્મેનિયન', 'ha': 'Armeniyanci', 'he': 'ארמנית', 'hi': 'आर्मेनियाई', 'hi-Latn': 'Armenian', 'hr': 'armenski', 'hsb': 'armenšćina', 'ht': 'arménien', 'hu': 'örmény', 'hy': 'հայերեն', 'ia': 'armenio', 'id': 'Armenia', 'ig': 'Armenianị', 'is': 'armenska', 'it': 'armeno', 'ja': 'アルメニア語', 'jv': 'Armenia', 'ka': 'სომხური', 'kea': 'arméniu', 'kgp': 'armẽnĩju', 'kk': 'армян тілі', 'kk-Cyrl': 'армян тілі', 'km': 'អាមេនី', 'kn': 'ಅರ್ಮೇನಿಯನ್', 'ko': '아르메니아어', 'kok': 'आर्मेनियन', 'kok-Deva': 'आर्मेनियन', 'ks': 'اَرمینیَن', 'ks-Arab': 'اَرمینیَن', 'ksh': 'Armehnesch', 'ku': 'ermenî', 'kxv': 'aarmeniati', 'kxv-Deva': 'आर्मेनियाति', 'kxv-Latn': 'aarmeniati', 'kxv-Orya': 'ଆର୍ମେନିଆତି', 'kxv-Telu': 'అర్మేనియన్', 'ky': 'армянча', 'lb': 'Armenesch', 'lkt': 'Armenia Iyápi', 'lld': 'armenich', 'lo': 'ອາເມນຽນ', 'lrc': 'أرمأنی', 'lt': 'armėnų', 'lv': 'armēņu', 'mai': 'अर्मेनियाई', 'mi': 'Āmeniana', 'mk': 'ерменски', 'ml': 'അർമേനിയൻ', 'mn': 'армен', 'mr': 'अर्मेनियन', 'ms': 'Armenia', 'mt': 'Armen', 'my': 'အာမေးနီးယား', 'mzn': 'ارمنی', 'nb': 'armensk', 'ne': 'आर्मेनियाली', 'nl': 'Armeens', 'nn': 'armensk', 'no': 'armensk', 'nqo': 'ߊߙߊߡߋߣߌߞߊ߲', 'om': 'Armeeniyaa', 'or': 'ଆର୍ମେନିଆନ୍', 'os': 'сомихаг', 'pa': 'ਅਰਮੀਨੀਆਈ', 'pa-Guru': 'ਅਰਮੀਨੀਆਈ', 'pcm': 'Armẹ́niá Lángwej', 'pl': 'ormiański', 'ps': 'آرمينيايي', 'pt': 'armênio', 'qu': 'Armenio Simi', 'rm': 'armen', 'ro': 'armeană', 'ru': 'армянский', 'rw': 'Ikinyarumeniya', 'sah': 'Эрмээннии', 'sat': 'ᱟᱨᱢᱮᱱᱤᱭᱟᱹᱱ', 'sat-Olck': 'ᱟᱨᱢᱮᱱᱤᱭᱟᱹᱱ', 'sc': 'armenu', 'sd': 'ارماني', 'sd-Arab': 'ارماني', 'se': 'armeenagiella', 'si': 'ආර්මේනියානු', 'sk': 'arménčina', 'sl': 'armenščina', 'smn': 'armeniakielâ', 'so': 'Armeeniyaan', 'sq': 'armenisht', 'sr': 'јерменски', 'sr-Cyrl': 'јерменски', 'sr-Latn': 'jermenski', 'sv': 'armeniska', 'sw': 'Kiarmenia', 'syr': 'ܐܪܡܢܝܬ', 'ta': 'ஆர்மேனியன்', 'te': 'ఆర్మీనియన్', 'tg': 'арманӣ', 'th': 'อาร์เมเนีย', 'ti': 'ኣርሜንኛ', 'tk': 'ermeni dili', 'to': 'lea fakaʻāmenia', 'tr': 'Ermenice', 'tt': 'әрмән', 'ug': 'ئەرمېنچە', 'uk': 'вірменська', 'ur': 'آرمینیائی', 'uz': 'arman', 'uz-Cyrl': 'арманча', 'uz-Latn': 'arman', 'vec': 'armen', 'vi': 'Tiếng Armenia', 'wae': 'Armeniš', 'wo': 'Armaniye', 'xnr': 'आर्मेनियाई', 'yi': 'אַרמעניש', 'yo': 'Èdè Armenia', 'yrl': 'arimẽniu', 'yue': '亞美尼亞文', 'yue-Hans': '亚美尼亚文', 'yue-Hant': '亞美尼亞文', 'zh': '亚美尼亚语', 'zh-Hans': '亚美尼亚语', 'zh-Hant': '亞美尼亞文', 'zu': 'isi-Armenia'},
    'hya': {'en': 'Hya'},
    'hyw': {'en': 'Western Armenian'},
    'hyx': {'en': 'Armenian (family)'},
    'hyx-pro': {'en': 'Proto-Armenian'},
    'hz': {'af': 'Herero', 'am': 'ሄሬሮ', 'ar': 'الهيريرو', 'as': 'হেৰেৰো', 'ast': 'herero', 'az': 'herero', 'az-Cyrl': 'һереро', 'az-Latn': 'herero', 'be': 'герэра', 'be-tarask': 'герэра', 'bg': 'хереро', 'bgn': 'هرویی', 'bn': 'হেরেরো', 'br': 'herero', 'brx': 'हेरेर’', 'bs': 'herero', 'bs-Cyrl': 'хереро', 'bs-Latn': 'herero', 'ca': 'herero', 'ca-ES-valencia': 'herero', 'ccp': '𑄦𑄬𑄢𑄬𑄢𑄮', 'ce': 'гереро', 'chr': 'ᎮᎴᎶ', 'ckb': 'هێرێرۆ', 'cs': 'hererština', 'cy': 'Herero', 'da': 'herero', 'de': 'Herero', 'dsb': 'hererošćina', 'el': 'Χερέρο', 'el-polyton': 'Χερέρο', 'en': 'Herero', 'es': 'herero', 'et': 'herero', 'eu': 'hereroera', 'fa': 'هریرویی', 'ff-Adlm': '𞤖𞤫𞤪𞤫𞤪𞤮𞥅𞤪𞤫', 'fi': 'herero', 'fil': 'Herero', 'fo': 'herero', 'fr': 'héréro', 'fy': 'Herero', 'ga': 'Heiréiris', 'gd': 'Herero', 'gl': 'herero', 'gsw': 'Herero', 'gu': 'હેરેરો', 'ha': 'Herero', 'he': 'הררו', 'hi': 'हरैरो', 'hi-Latn': 'Herero', 'hr': 'herero', 'hsb': 'hererošćina', 'ht': 'héréro', 'hu': 'herero', 'hy': 'հերերո', 'ia': 'herero', 'id': 'Herero', 'ig': 'Herero', 'is': 'hereró', 'it': 'herero', 'ja': 'ヘレロ語', 'jv': 'Herero', 'ka': 'ჰერერო', 'kgp': 'herero', 'kk': 'гереро тілі', 'kk-Cyrl': 'гереро тілі', 'km': 'ហឺរីរ៉ូ', 'kn': 'ಹೆರೆರೊ', 'ko': '헤레로어', 'kok': 'हिरिरो', 'kok-Deva': 'हिरिरो', 'ks': 'ہیریٖرو', 'ks-Arab': 'ہیریٖرو', 'ksh': 'Herrehro', 'ku': 'hereroyî', 'ky': 'герерочо', 'lb': 'Herero-Sprooch', 'lo': 'ເຮິຮິໂຣ', 'lt': 'hererų', 'lv': 'hereru', 'mai': 'हैरेरो', 'mi': 'Herero', 'mk': 'хереро', 'ml': 'ഹെരേരൊ', 'mn': 'хереро', 'mr': 'हरेरो', 'ms': 'Herero', 'mt': 'Herero', 'my': 'ဟီရဲရို', 'nb': 'herero', 'ne': 'हेरेरो', 'nl': 'Herero', 'nn': 'herero', 'no': 'herero', 'or': 'ହେରେରୋ', 'pa': 'ਹਰੇਰੋ', 'pa-Guru': 'ਹਰੇਰੋ', 'pcm': 'Hẹrẹ́ro', 'pl': 'herero', 'ps': 'هیرورو', 'pt': 'herero', 'qu': 'Herero Simi', 'rm': 'herero', 'ro': 'herero', 'ru': 'гереро', 'sc': 'herero', 'sd': 'هريرو', 'sd-Arab': 'هريرو', 'si': 'හෙරෙරො', 'sk': 'herero', 'sl': 'herero', 'smn': 'hererokielâ', 'so': 'Herero', 'sq': 'hereroisht', 'sr': 'хереро', 'sr-Cyrl': 'хереро', 'sr-Latn': 'herero', 'sv': 'herero', 'sw': 'Kiherero', 'ta': 'ஹெரேரோ', 'te': 'హెరెరో', 'tg': 'ҳереро', 'th': 'เฮเรโร', 'ti': 'ሄረሮ', 'tk': 'gerero dili', 'to': 'lea fakahelelo', 'tr': 'Herero dili', 'tt': 'гереро', 'ug': 'خېرېروچە', 'uk': 'гереро', 'ur': 'ہریرو', 'uz': 'gerero', 'uz-Cyrl': 'гереро', 'uz-Latn': 'gerero', 'vec': 'herero', 'vi': 'Tiếng Herero', 'wo': 'Herero', 'xnr': 'हरैरो', 'yo': 'Èdè Herero', 'yrl': 'hereru', 'yue': '赫雷羅文', 'yue-Hans': '赫雷罗文', 'yue-Hant': '赫雷羅文', 'zh': '赫雷罗语', 'zh-Hans': '赫雷罗语', 'zh-Hant': '赫雷羅文', 'zu': 'isi-Herero'},
    'ia': {'af': 'Interlingua', 'ak': 'Kasa ntam', 'am': 'ኢንቴርሊንጓ', 'ar': 'اللّغة الوسيطة', 'as': 'ইণ্টাৰলিংগুৱা', 'ast': 'interlingua', 'az': 'interlinqua', 'az-Cyrl': 'интерлингва', 'az-Latn': 'interlinqua', 'bal-Latn': 'Myánzobáni', 'be': 'інтэрлінгва', 'be-tarask': 'інтэрлінгва', 'bg': 'интерлингва', 'bgn': 'اینترلینگوایی', 'bn': 'ইন্টারলিঙ্গুয়া', 'br': 'interlingua', 'brx': 'ईन्टरलिंग्वा', 'bs': 'interlingva', 'bs-Cyrl': 'интерлингва', 'bs-Latn': 'interlingva', 'ca': 'interlingua', 'ca-ES-valencia': 'interlingua', 'ccp': '𑄃𑄨𑄚𑄴𑄑𑄢𑄴𑄣𑄨𑄁𑄉𑄪𑄠', 'ce': 'интерлингва', 'chr': 'ᎠᏰᏟ ᎦᏬᏂᎯᏍᏗ', 'ckb': 'ئینترلینگووا', 'cs': 'interlingua', 'cy': 'Interlingua', 'da': 'interlingua', 'de': 'Interlingua', 'dsb': 'interlingua', 'el': 'Ιντερλίνγκουα', 'el-polyton': 'Ιντερλίνγκουα', 'en': 'Interlingua', 'en-Dsrt': '𐐆𐑌𐐻𐐲𐑉𐑊𐐮𐑍𐐶𐐲', 'eo': 'Interlingvao', 'es': 'interlingua', 'et': 'interlingua', 'eu': 'interlingua', 'fa': 'اینترلینگوا', 'ff-Adlm': '𞤉𞤲𞤼𞤫𞤪𞤤𞤭𞤺𞤢𞥄𞤪𞤫', 'fi': 'interlingua', 'fil': 'Interlingua', 'fo': 'interlingua', 'fr': 'interlingua', 'fy': 'Interlingua', 'ga': 'Interlingua', 'gd': 'Interlingua', 'gl': 'interlingua', 'gsw': 'Interlingua', 'gu': 'ઇંટરલિંગુઆ', 'ha': 'Yare Tsakanin Kasashe', 'he': '\u200fאינטרלינגואה', 'hi': 'इंटरलिंगुआ', 'hi-Latn': 'Interlingua', 'hr': 'interlingua', 'hsb': 'interlingua', 'ht': 'interlingua', 'hu': 'interlingva', 'hy': 'ինտերլինգուա', 'ia': 'interlingua', 'id': 'Interlingua', 'ig': 'Interlingua', 'is': 'interlingua', 'it': 'interlingua', 'ja': 'インターリングア', 'jv': 'Interlingua', 'ka': 'ინტერლინგუალური', 'kea': 'interlingua', 'kgp': 'vĩ-jãgja', 'kk': 'интерлингва тілі', 'kk-Cyrl': 'интерлингва тілі', 'km': 'អ៊ីនធើលីង', 'kn': 'ಇಂಟರ್\u200cಲಿಂಗ್ವಾ', 'ko': '인터링구아', 'kok': 'इन्टरलिंग्वा', 'kok-Deva': 'इन्टरलिंग्वा', 'ks': 'اِنٹَرلِنگوا', 'ks-Arab': 'اِنٹَرلِنگوا', 'ksh': 'de Interlingua', 'ku': 'înterlîngua', 'ky': 'интерлингва', 'lb': 'Interlingua', 'lld': 'interlingua', 'lo': 'ອິນເຕີລິງລົວ', 'lt': 'tarpinė', 'lv': 'interlingva', 'mai': 'इंटरलिंगुआ', 'mi': 'Inarīngua', 'mk': 'интерлингва', 'ml': 'ഇന്റർലിംഗ്വ', 'mn': 'интерлингво', 'mr': 'इंटरलिंग्वा', 'ms': 'Interlingua', 'mt': 'Interlingua', 'my': 'အင်တာလင်ဂွါ', 'nb': 'interlingua', 'ne': 'इन्टर्लिङ्गुआ', 'nl': 'Interlingua', 'nn': 'interlingua', 'no': 'interlingua', 'nqo': 'ߍ߲ߕߍߙߑߟߌ߲ߜ߭ߏߥߊߞߊ߲', 'om': 'Interlingua', 'or': 'ଇର୍ଣ୍ଟଲିଙ୍ଗୁଆ', 'pa': 'ਇੰਟਰਲਿੰਗੁਆ', 'pa-Guru': 'ਇੰਟਰਲਿੰਗੁਆ', 'pcm': 'Intalíngwuá Lángwej', 'pl': 'interlingua', 'ps': 'انټرلنګوا', 'pt': 'interlíngua', 'qu': 'Interlingua Simi', 'rm': 'interlingua', 'ro': 'interlingua', 'ru': 'интерлингва', 'rw': 'Ururimi Gahuzamiryango', 'sc': 'interlìngua', 'sd': 'انٽرلنگئا', 'sd-Arab': 'انٽرلنگئا', 'si': 'ඉන්ටලින්ගුආ', 'sk': 'interlingua', 'sl': 'interlingva', 'smn': 'interlingua', 'so': 'Interlinguwa', 'sq': 'interlingua', 'sr': 'интерлингва', 'sr-Cyrl': 'интерлингва', 'sr-Latn': 'interlingva', 'sv': 'interlingua', 'sw': 'Kiintalingua', 'ta': 'இன்டர்லிங்வா', 'te': 'ఇంటర్లింగ్వా', 'th': 'อินเตอร์ลิงกัว', 'ti': 'ኢንተርሊንጓ', 'tk': 'interlingwa dili', 'to': 'lea fakavahaʻalea', 'tr': 'İnterlingua', 'ug': 'ئارىلىق تىل', 'uk': 'інтерлінгва', 'ur': 'بین لسانیات', 'uz': 'interlingva', 'uz-Cyrl': 'интерлингва', 'uz-Latn': 'interlingva', 'vec': 'interlengua', 'vi': 'Tiếng Khoa Học Quốc Tế', 'xnr': 'इंटरलिंगुआ', 'yo': 'Èdè pipo', 'yrl': 'neẽgasuí', 'yue': '國際文', 'yue-Hans': '国际文', 'yue-Hant': '國際文', 'zh': '国际语', 'zh-Hans': '国际语', 'zh-Hant': '國際文', 'zu': 'izilimi ezihlangene'},
    'iai': {'en': 'Iaai'},
    'ian': {'en': 'Iatmul'},
    'iap': {'en': 'Iapama'},
    'iar': {'en': 'Purari'},
    'iba': {'af': 'Ibanees', 'am': 'ኢባን', 'ar': 'الإيبان', 'as': 'ইবান', 'ast': 'iban', 'az': 'iban', 'az-Cyrl': 'ибан', 'az-Latn': 'iban', 'be': 'ібан', 'be-tarask': 'ібан', 'bg': 'ибан', 'bgn': 'ایبانگه', 'bn': 'ইবান', 'br': 'iban', 'brx': 'ईबान', 'bs': 'iban', 'bs-Cyrl': 'ибан', 'bs-Latn': 'iban', 'ca': 'iban', 'ca-ES-valencia': 'iban', 'ccp': '𑄃𑄨𑄝𑄚𑄴', 'ce': 'ибанийн', 'chr': 'ᎢᏆᏂ', 'ckb': 'ئیبان', 'cs': 'ibanština', 'cy': 'Ibaneg', 'da': 'iban', 'de': 'Iban', 'dsb': 'ibanšćina', 'el': 'Ιμπάν', 'el-polyton': 'Ιμπάν', 'en': 'Iban', 'es': 'iban', 'et': 'ibani', 'eu': 'ibanera', 'fa': 'ایبانی', 'ff-Adlm': '𞤋𞤦𞤢𞤲𞤪𞤫', 'fi': 'iban', 'fil': 'Iban', 'fo': 'iban', 'fr': 'iban', 'fy': 'Iban', 'ga': 'Ibeainis', 'gd': 'Iban', 'gl': 'iban', 'gsw': 'Ibanisch', 'gu': 'ઇબાન', 'ha': 'Iban', 'he': 'איבאן', 'hi': 'इबान', 'hi-Latn': 'Iban', 'hr': 'iban', 'hsb': 'ibanšćina', 'ht': 'iban', 'hu': 'iban', 'hy': 'իբաներեն', 'ia': 'iban', 'id': 'Iban', 'ig': 'Iban', 'is': 'íban', 'it': 'iban', 'ja': 'イバン語', 'jv': 'Iban', 'ka': 'იბანი', 'kgp': 'iman', 'kk': 'ибан тілі', 'kk-Cyrl': 'ибан тілі', 'km': 'អ៊ីបាន', 'kn': 'ಇಬಾನ್', 'ko': '이반어', 'kok': 'आयबन', 'kok-Deva': 'आयबन', 'ks': 'اِبان', 'ks-Arab': 'اِبان', 'ksh': 'Ibahnesch', 'ku': 'iban', 'ky': 'ибанча', 'lb': 'Iban', 'lo': 'ໄອບານ', 'lt': 'iban', 'lv': 'ibanu', 'mai': 'इबन', 'mi': 'Īpana', 'mk': 'ибан', 'ml': 'ഇബാൻ', 'mn': 'ибан', 'mr': 'इबान', 'ms': 'Iban', 'mt': 'Iban', 'my': 'အီဗန်', 'nb': 'iban', 'ne': 'इबान', 'nl': 'Iban', 'nn': 'iban', 'no': 'iban', 'or': 'ଇବାନ୍', 'pa': 'ਇਬਾਨ', 'pa-Guru': 'ਇਬਾਨ', 'pcm': 'Iban Lángwej', 'pl': 'iban', 'ps': 'ابن', 'pt': 'iban', 'qu': 'Iban Simi', 'rm': 'iban', 'ro': 'iban', 'ru': 'ибанский', 'sc': 'iban', 'sd': 'ايبن', 'sd-Arab': 'ايبن', 'si': 'ඉබන්', 'sk': 'ibančina', 'sl': 'ibanščina', 'smn': 'ibankielâ', 'so': 'Iban', 'sq': 'ibanisht', 'sr': 'ибански', 'sr-Cyrl': 'ибански', 'sr-Latn': 'ibanski', 'sv': 'ibanska', 'sw': 'Kiiban', 'ta': 'இபான்', 'te': 'ఐబాన్', 'th': 'อิบาน', 'ti': 'ኢባን', 'tk': 'iban dili', 'to': 'lea fakaʻipani', 'tr': 'Iban', 'ug': 'ئىبانچە', 'uk': 'ібанська', 'ur': 'ایبان', 'uz': 'iban', 'uz-Cyrl': 'ибан тили', 'uz-Latn': 'iban', 'vec': 'iban', 'vi': 'Tiếng Iban', 'xnr': 'इबान', 'yo': 'Èdè Iba', 'yrl': 'ibã', 'yue': '伊班文', 'yue-Hans': '伊班文', 'yue-Hant': '伊班文', 'zh': '伊班语', 'zh-Hans': '伊班语', 'zh-Hant': '伊班文', 'zu': 'isi-Iban'},
    'ibb': {'af': 'Ibibio', 'am': 'ኢቢብዮ', 'ar': 'الإيبيبيو', 'as': 'ইবিবিও', 'ast': 'ibibio', 'az': 'ibibio', 'az-Cyrl': 'ибибио', 'az-Latn': 'ibibio', 'be': 'ібібія', 'be-tarask': 'ібібія', 'bg': 'ибибио', 'bgn': 'ایبیبیو', 'bn': 'ইবিবিও', 'br': 'ibibio', 'brx': 'ईबिबिय’', 'bs': 'ibibio', 'bs-Cyrl': 'ибибио', 'bs-Latn': 'ibibio', 'ca': 'ibibio', 'ca-ES-valencia': 'ibibio', 'ccp': '𑄃𑄨𑄝𑄨𑄝𑄨𑄠𑄧', 'ce': 'ибибио', 'chr': 'ᎢᏈᏈᎣ', 'ckb': 'ئیبیبۆ', 'cs': 'ibibio', 'cy': 'Ibibio', 'da': 'ibibio', 'de': 'Ibibio', 'dsb': 'ibibiošćina', 'el': 'Ιμπίμπιο', 'el-polyton': 'Ιμπίμπιο', 'en': 'Ibibio', 'es': 'ibibio', 'et': 'ibibio', 'eu': 'ibibioera', 'fa': 'ایبیبیو', 'ff-Adlm': '𞤋𞤦𞤭𞥅𞤦𞤭𞤴𞤮𞥅𞤪𞤫', 'fi': 'ibibio', 'fil': 'Ibibio', 'fo': 'ibibio', 'fr': 'ibibio', 'fy': 'Ibibio', 'ga': 'Ibibis', 'gd': 'Ibibio', 'gl': 'ibibio', 'gu': 'ઇબિબિઓ', 'ha': 'Ibibio', 'he': 'איביביו', 'hi': 'इबिबियो', 'hi-Latn': 'Ibibio', 'hr': 'ibibio', 'hsb': 'ibibiošćina', 'ht': 'ibibio', 'hu': 'ibibio', 'hy': 'իբիբիո', 'ia': 'ibibio', 'id': 'Ibibio', 'ig': 'Ibibio', 'is': 'ibibio', 'it': 'ibibio', 'ja': 'イビビオ語', 'jv': 'Ibibio', 'ka': 'იბიბიო', 'kgp': 'imimijo', 'kk': 'ибибио тілі', 'kk-Cyrl': 'ибибио тілі', 'km': 'អាយប៊ីប៊ីអូ', 'kn': 'ಇಬಿಬಿಯೋ', 'ko': '이비비오어', 'kok': 'ईबिबियो', 'kok-Deva': 'ईबिबियो', 'ksh': 'Ibibio', 'ku': 'îbîbîoyî', 'ky': 'ибибиочо', 'lb': 'Ibibio', 'lo': 'ໄອໄບໄບໂອ', 'lt': 'ibibijų', 'lv': 'ibibio', 'mai': 'इबिबियो', 'mi': 'Ipīpio', 'mk': 'ибибио', 'ml': 'ഇബീബിയോ', 'mn': 'ибибио', 'mr': 'इबिबिओ', 'ms': 'Ibibio', 'mt': 'Ibibio', 'my': 'အီဘီဘီယို', 'nb': 'ibibio', 'ne': 'इबिबियो', 'nl': 'Ibibio', 'nn': 'ibibio', 'no': 'ibibio', 'or': 'ଇବିବିଓ', 'pa': 'ਇਬੀਬੀਓ', 'pa-Guru': 'ਇਬੀਬੀਓ', 'pcm': 'Ibibio Lángwej', 'pl': 'ibibio', 'ps': 'ابیبیو', 'pt': 'ibibio', 'qu': 'Ibibio Simi', 'ro': 'ibibio', 'ru': 'ибибио', 'sc': 'ibibio', 'sd': 'ابيبيو', 'sd-Arab': 'ابيبيو', 'si': 'ඉබිබියො', 'sk': 'ibibio', 'sl': 'ibibijščina', 'smn': 'ibibiokielâ', 'so': 'Ibibio', 'sq': 'ibibioisht', 'sr': 'ибибио', 'sr-Cyrl': 'ибибио', 'sr-Latn': 'ibibio', 'sv': 'ibibio', 'sw': 'Kiibibio', 'ta': 'இபிபியோ', 'te': 'ఇబిబియో', 'tg': 'ибибио', 'th': 'อิบิบิโอ', 'ti': 'ኢቢብዮ', 'tk': 'ibibio dili', 'to': 'lea fakaʻipipio', 'tr': 'İbibio dili', 'tt': 'ибибио', 'ug': 'ئىبىبىئوچە', 'uk': 'ібібіо', 'ur': 'ابی بیو', 'uz': 'ibibio', 'uz-Cyrl': 'ибибо', 'uz-Latn': 'ibibio', 'vec': 'ibibio', 'vi': 'Tiếng Ibibio', 'wo': 'Ibibiyo', 'xnr': 'इबिबियो', 'yo': 'Èdè Ibibio', 'yrl': 'ibibiu', 'yue': '伊比比奧文', 'yue-Hans': '伊比比奥文', 'yue-Hant': '伊比比奧文', 'zh': '伊比比奥语', 'zh-Hans': '伊比比奥语', 'zh-Hant': '伊比比奧文', 'zu': 'isi-Ibibio'},
    'ibd': {'en': 'Iwaidja'},
    'ibe': {'en': 'Akpes'},
    'ibg': {'en': 'Ibanag'},
    'ibh': {'en': 'Bih'},
    'ibi': {'en': 'Ibilo'},
    'ibl': {'en': 'Ibaloi'},
    'ibm': {'en': 'Agoi'},
    'ibn': {'en': 'Ibino'},
    'ibr': {'en': 'Ibuoro'},
    'ibu': {'en': 'Ibu'},
    'iby': {'en': 'Ibani'},
    'ica': {'en': 'Ede Ica'},
    'ich': {'en': 'Etkywan'},
    'icl': {'en': 'Icelandic Sign Language'},
    'icr': {'en': 'Islander Creole English'},
    'id': {'id': 'bahasa Indonesia', 'af': 'Indonesies', 'agq': 'Èndònɛshìa', 'ak': 'Indonihyia kasa', 'am': 'ኢንዶኔዥያኛ', 'ar': 'الإندونيسية', 'as': 'ইণ্ডোনেচিয়', 'asa': 'Kiindonethia', 'ast': 'indonesiu', 'az': 'indoneziya', 'az-Cyrl': 'индонезија', 'az-Latn': 'indoneziya', 'bal-Latn': 'Endónési', 'bas': 'Hɔp u indònesìà', 'be': 'інданезійская', 'be-tarask': 'інданезійская', 'bem': 'Ichi Indonesiani', 'bez': 'Hiindonesia', 'bg': 'индонезийски', 'bgn': 'ایندونیزیایی', 'blo': 'Ɛnɖonosii kagɩja', 'bm': 'Ɛndonezikan', 'bn': 'ইন্দোনেশীয়', 'br': 'indonezeg', 'brx': 'ईण्डनेसियान', 'bs': 'indonezijski', 'bs-Cyrl': 'индонежански', 'bs-Latn': 'indonezijski', 'ca': 'indonesi', 'ca-ES-valencia': 'indonesi', 'ccp': '𑄃𑄨𑄚𑄴𑄘𑄮𑄚𑄬𑄥𑄨𑄠𑄧', 'ce': 'индонезихойн', 'ceb': 'Indonesian', 'cgg': 'Oruindonezia', 'chr': 'ᎢᏂᏙᏂᏏᎠ', 'ckb': 'ئیندۆنیزی', 'cs': 'indonéština', 'cv': 'индонези', 'cy': 'Indoneseg', 'da': 'indonesisk', 'dav': 'Kiindonesia', 'de': 'Indonesisch', 'dje': 'Indoneesi senni', 'dsb': 'indonešćina', 'dyo': 'indoneesi', 'dz': 'ཨིན་ཌོ་ནེ་ཤི་ཡཱན་ཁ', 'ebu': 'Kĩindonesia', 'ee': 'Indonesiagbe', 'el': 'Ινδονησιακά', 'el-polyton': 'Ἰνδονησιακά', 'en': 'Indonesian', 'en-Dsrt': '𐐆𐑌𐐼𐐬𐑌𐐨𐑈𐐲𐑌', 'eo': 'indonezia', 'es': 'indonesio', 'et': 'indoneesia', 'eu': 'indonesiera', 'ewo': 'Ǹkɔ́bɔ ɛndonésian', 'fa': 'اندونزیایی', 'ff': 'Endonesiire', 'ff-Adlm': '𞤋𞤲𞤣𞤮𞤲𞤭𞥅𞤧𞤭𞤴𞤢𞤲𞤳𞤮𞥅𞤪𞤫', 'ff-Latn': 'Endonesiire', 'fi': 'indonesia', 'fil': 'Indonesian', 'fo': 'indonesiskt', 'fr': 'indonésien', 'fur': 'indonesian', 'fy': 'Yndonezysk', 'ga': 'Indinéisis', 'gd': 'Innd-Innsis', 'gl': 'indonesio', 'gsw': 'Indonesisch', 'gu': 'ઇન્ડોનેશિયન', 'guz': 'Kiindonesia', 'ha': 'Harshen Indunusiya', 'he': 'אינדונזית', 'hi': 'इंडोनेशियाई', 'hi-Latn': 'Indonesian', 'hr': 'indonezijski', 'hsb': 'indonešćina', 'ht': 'indonésien', 'hu': 'indonéz', 'hy': 'ինդոնեզերեն', 'ia': 'indonesiano', 'ie': 'indonesian', 'ig': 'Indonesian', 'is': 'indónesíska', 'it': 'indonesiano', 'ja': 'インドネシア語', 'jmc': 'Kyiindonesia', 'jv': 'Indonesia', 'ka': 'ინდონეზიური', 'kab': 'Tandunisit', 'kam': 'Kiindonesia', 'kde': 'Chiiongonesia', 'kea': 'indonéziu', 'kgp': 'ĩnonẽsijo', 'khq': 'Indoneesi senni', 'ki': 'Kiindonesia', 'kk': 'индонезия тілі', 'kk-Cyrl': 'индонезия тілі', 'kln': 'kutitab Indonesia', 'km': 'ឥណ្ឌូណេស៊ី', 'kn': 'ಇಂಡೋನೇಶಿಯನ್', 'ko': '인도네시아어', 'kok': 'इंडोनेशियन', 'kok-Deva': 'इंडोनेशियन', 'ks': 'اِنڈونیشیا', 'ks-Arab': 'اِنڈونیشیا', 'ksb': 'Kiindonesia', 'ksf': 'riindonɛsí', 'ksh': 'Indonehsesch', 'ku': 'endonezyayî', 'kxv': 'inḍonesiyati', 'kxv-Deva': 'इंडोनेसियाति', 'kxv-Latn': 'inḍonesiyati', 'kxv-Orya': 'ଇଣ୍ଡନେସୀୟା ତି', 'kxv-Telu': 'ఇండోనేసియన్', 'ky': 'индонезияча', 'lag': 'Kɨɨndonésia', 'lb': 'Indonesesch', 'lg': 'Luyindonezya', 'lkt': 'Indonesia Iyápi', 'lld': 'indonesian', 'ln': 'lindonezi', 'lo': 'ອິນໂດເນຊຽນ', 'lrc': 'أندونئزیایی', 'lt': 'indoneziečių', 'lu': 'Lindonezia', 'luo': 'Kiindonesia', 'luy': 'Kiindonesia', 'lv': 'indonēziešu', 'mai': 'इंडोनेशियाई', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ Indonesia', 'mer': 'Kĩindonesia', 'mfe': 'indonezien', 'mg': 'Indonezianina', 'mi': 'Initonīhiana', 'mk': 'индонезиски', 'ml': 'ഇന്തോനേഷ്യൻ', 'mn': 'индонези', 'mni': 'ইন্দোনেসিয়া', 'mni-Beng': 'ইন্দোনেসিয়া', 'mr': 'इंडोनेशियन', 'ms': 'Indonesia', 'mt': 'Indoneżjan', 'mua': 'indonesiya', 'my': 'အင်ဒိုနီးရှား', 'mzn': 'اندونزیایی', 'naq': 'Indonesiaǁî gowab', 'nb': 'indonesisk', 'nd': 'isi-Indonesia', 'ne': 'इन्डोनेसियाली', 'nl': 'Indonesisch', 'nmg': 'Kiɛl indonesie', 'nn': 'indonesisk', 'no': 'indonesisk', 'nqo': 'ߍ߲ߘߣߏߛߌߞߊ߲', 'nus': 'Thok indunithiani', 'nyn': 'Oruindonezia', 'om': 'Afaan Indoneziya', 'or': 'ଇଣ୍ଡୋନେସୀୟ', 'pa': 'ਇੰਡੋਨੇਸ਼ੀਆਈ', 'pa-Guru': 'ਇੰਡੋਨੇਸ਼ੀਆਈ', 'pcm': 'Indoníshia Lángwej', 'pl': 'indonezyjski', 'ps': 'انډونېزي', 'pt': 'indonésio', 'qu': 'Indonesio Simi', 'rm': 'indonais', 'rn': 'Ikinyendoziya', 'ro': 'indoneziană', 'rof': 'Kiindonesia', 'ru': 'индонезийский', 'rw': 'Ikinyendoziya', 'rwk': 'Kyiindonesia', 'saq': 'Kiindonesia', 'sbp': 'Ishihindonesia', 'sc': 'indonesianu', 'sd': 'انڊونيشي', 'sd-Arab': 'انڊونيشي', 'se': 'indonesiagiella', 'seh': 'indonésio', 'ses': 'Indoneesi senni', 'sg': 'Enndonezïi', 'shi': 'ⵜⴰⵏⴷⵓⵏⵉⵙⵉⵜ', 'shi-Latn': 'Tandunisit', 'shi-Tfng': 'ⵜⴰⵏⴷⵓⵏⵉⵙⵉⵜ', 'si': 'ඉන්දුනීසියානු', 'sk': 'indonézština', 'sl': 'indonezijščina', 'smn': 'indonesiakielâ', 'sn': 'chiIndonesia', 'so': 'Indunusiyaan', 'sq': 'indonezisht', 'sr': 'индонежански', 'sr-Cyrl': 'индонежански', 'sr-Latn': 'indonežanski', 'sv': 'indonesiska', 'sw': 'Kiindonesia', 'ta': 'இந்தோனேஷியன்', 'te': 'ఇండోనేషియన్', 'teo': 'Kiindonesia', 'tg': 'индонезӣ', 'th': 'อินโดนีเซีย', 'ti': 'ኢንዶነዥኛ', 'tk': 'indonez dili', 'to': 'lea fakaʻinitōnesia', 'tr': 'Endonezce', 'tt': 'индонезия', 'twq': 'Indoneesi senni', 'tzm': 'Tindunisit', 'ug': 'ھىندونېزچە', 'uk': 'індонезійська', 'ur': 'انڈونیثیائی', 'uz': 'indonez', 'uz-Cyrl': 'индонезча', 'uz-Latn': 'indonez', 'vai': 'ꔤꖆꕇꔻꘂꘋ', 'vai-Latn': 'Índonisiyɛŋ', 'vai-Vaii': 'ꔤꖆꕇꔻꘂꘋ', 'vec': 'indonezian', 'vi': 'Tiếng Indonesia', 'vun': 'Kyiindonesia', 'wae': 'Indonesiš', 'wo': 'Endonesiye', 'xh': 'Isi-Indonesia', 'xnr': 'इंडोनेशियाई', 'xog': 'Oluyindonezya', 'yav': 'índonísiɛ', 'yi': 'אינדאנעזיש', 'yo': 'Èdè Indonéṣíà', 'yrl': 'ĩdunésiu', 'yue': '印尼文', 'yue-Hans': '印尼文', 'yue-Hant': '印尼文', 'zgh': 'ⵜⴰⵏⴷⵓⵏⵉⵙⵉⵜ', 'zh': '印度尼西亚语', 'zh-Hans': '印度尼西亚语', 'zh-Hant': '印尼文', 'zu': 'isi-Indonesian'},
    'ida': {'en': 'Idakho-Isukha-Tiriki'},
    'idb': {'en': 'Indo-Portuguese'},
    'idc': {'en': 'Idon'},
    'idd': {'en': 'Ede Idaca'},
    'ide': {'en': 'Idere'},
    'idi': {'en': 'Idi'},
    'idr': {'en': 'Indri'},
    'ids': {'en': 'Idesa'},
    'idt': {'en': 'Idaté'},
    'idu': {'en': 'Idoma'},
    'ie': {'af': 'Interlingue', 'ak': 'Kasa afrafra', 'am': 'እንተርሊንግወ', 'ar': 'الإنترلينج', 'as': 'ইণ্টাৰলিংগুৱে', 'ast': 'interlingue', 'az': 'interlinqve', 'az-Latn': 'interlinqve', 'bal-Latn': 'Myánzobán', 'be': 'інтэрлінгвэ', 'be-tarask': 'інтэрлінгвэ', 'bg': 'интерлингве', 'bn': 'ইন্টারলিঙ্গ', 'br': 'interlingue', 'brx': 'ईन्टरलिंग्वे', 'bs': 'interlingve', 'bs-Cyrl': 'међујезички', 'bs-Latn': 'interlingve', 'ca': 'interlingue', 'ca-ES-valencia': 'interlingue', 'ccp': '𑄃𑄨𑄚𑄴𑄑𑄢𑄴𑄣𑄨𑄁𑄉𑄧', 'chr': 'ᏈᏖᎵᏆ', 'cs': 'interlingue', 'cy': 'Interlingue', 'da': 'interlingue', 'de': 'Interlingue', 'dsb': 'interlinguešćina', 'el': 'Ιντερλίνγκουε', 'el-polyton': 'Ιντερλίνγκουε', 'en': 'Interlingue', 'eo': 'Interlingveo', 'es': 'interlingue', 'et': 'interlingue', 'eu': 'interlinguea', 'fa': 'اکسیدنتال', 'fi': 'interlingue', 'fil': 'Interlingue', 'fo': 'interlingue', 'fr': 'interlingue', 'fy': 'Interlingue', 'ga': 'Interlingue', 'gd': 'Interlingue', 'gl': 'occidental', 'gsw': 'Interlingue', 'gu': 'ઇંટરલિંગ', 'ha': 'Intagulanci', 'he': 'אינטרלינגה', 'hi': 'ईन्टरलिंगुइ', 'hi-Latn': 'Interlingue', 'hr': 'interligua', 'hsb': 'interlinguešćina', 'ht': 'interlingue', 'hu': 'interlingue', 'hy': 'ինտերլինգուե', 'ia': 'interlingue', 'id': 'Interlingue', 'ie': 'Interlingue', 'ig': 'Interlingue', 'is': 'interlingve', 'it': 'interlingue', 'ja': 'インターリング', 'jv': 'Interlingue', 'ka': 'ინტერლინგი', 'kgp': 'vĩ-ag kuju ki', 'kk': 'интерлингве тілі', 'kk-Cyrl': 'интерлингве тілі', 'km': 'អ៊ីនធើលីងវេ', 'kn': 'ಇಂಟರ್\u200cಲಿಂಗ್', 'ko': '인테르링구에', 'kok': 'इन्टरलिंग्', 'kok-Deva': 'इन्टरलिंग्', 'ks': 'اِنٹَر لِننگویے', 'ks-Arab': 'اِنٹَر لِننگویے', 'ku': 'înterlîngue', 'ky': 'интерлинг', 'lb': 'Interlingue', 'lld': 'interlingue', 'lo': 'ອິນເຕີລິງກຣີ', 'lt': 'interkalba', 'lv': 'interlingve', 'mk': 'интерлингве', 'ml': 'ഇന്റർലിംഗ്വേ', 'mn': 'нэгдмэл хэл', 'mr': 'इंटरलिंग', 'ms': 'Interlingue', 'mt': 'Interlingue', 'my': 'အင်တာလင်း', 'nb': 'interlingue', 'ne': 'इन्टरलिङ्ग्वे', 'nl': 'Interlingue', 'nn': 'interlingue', 'no': 'interlingue', 'or': 'ଇର୍ଣ୍ଟରଲିଙ୍ଗୁଇ', 'pa': 'ਇੰਟਰਲਿੰਗੁਈ', 'pa-Guru': 'ਇੰਟਰਲਿੰਗੁਈ', 'pcm': 'Intalíngwe Lángwej', 'pl': 'interlingue', 'ps': 'آسا نا جبة', 'pt': 'interlingue', 'rm': 'interlingue', 'ro': 'interlingue', 'ru': 'интерлингве', 'rw': 'Uruhuzandimi', 'sc': 'interlìngue', 'sd': 'انٽرلنگئي', 'sd-Arab': 'انٽرلنگئي', 'si': 'ඉන්ටර්ලින්ග්', 'sk': 'interlingue', 'sl': 'interlingve', 'so': 'Interlingue', 'sq': 'gjuha oksidentale', 'sr': 'интерлингве', 'sr-Cyrl': 'интерлингве', 'sr-Latn': 'interlingve', 'sv': 'interlingue', 'sw': 'lugha ya kisayansi', 'ta': 'இன்டர்லிங்', 'te': 'ఇంటర్లింగ్', 'th': 'อินเตอร์ลิงกิว', 'ti': 'ኢንተርሊንጔ', 'tk': 'interlingwe dili', 'to': 'lea fakavahaʻalingikē', 'tr': 'Interlingue', 'ug': 'ئىنتىرلىڭچە', 'uk': 'інтерлінгве', 'ur': 'غربی', 'uz': 'interlingve', 'uz-Latn': 'interlingve', 'vec': 'interlengue', 'vi': 'Tiếng Interlingue', 'xnr': 'ईन्टरलिंगुइ', 'yo': 'Èdè àtọwọ́dá', 'yrl': 'neẽgapitera', 'yue': '國際文（E）', 'yue-Hans': '国际文（E）', 'yue-Hant': '國際文（E）', 'zh': '国际文字（E）', 'zh-Hans': '国际文字（E）', 'zh-Hant': '國際文（E）', 'zu': 'izimili'},
    'ifa': {'en': 'Amganad Ifugao'},
    'ifb': {'en': 'Batad Ifugao'},
    'ife': {'en': 'Ifè'},
    'iff': {'en': 'Ifo'},
    'ifk': {'en': 'Tuwali Ifugao'},
    'ifm': {'en': 'Teke-Fuumu'},
    'ifu': {'en': 'Mayoyao Ifugao'},
    'ify': {'en': 'Keley-I Kallahan'},
    'ig': {'af': 'Igbo', 'agq': 'Egbò', 'ak': 'Igbo kasa', 'am': 'ኢግቦኛ', 'ar': 'الإيجبو', 'as': 'ইগ্বো', 'asa': 'Kiigbo', 'ast': 'igbo', 'az': 'iqbo', 'az-Cyrl': 'игбо', 'az-Latn': 'iqbo', 'bal-Latn': 'Igbó', 'bas': 'Hɔp u iɓò', 'be': 'ігба', 'be-tarask': 'ігба', 'bem': 'Ichi Ibo', 'bez': 'Hiibo', 'bg': 'игбо', 'bgn': 'ایگبویی', 'bm': 'igibokan', 'bn': 'ইগ্\u200cবো', 'br': 'igbo', 'brx': 'ईग्ब’', 'bs': 'igbo', 'bs-Cyrl': 'игбо', 'bs-Latn': 'igbo', 'ca': 'igbo', 'ca-ES-valencia': 'igbo', 'ccp': '𑄃𑄨𑄉𑄴𑄝𑄮', 'ce': 'игбо', 'cgg': 'Oruibo', 'chr': 'ᎢᎦᎪ', 'ckb': 'ئیگبۆ', 'cs': 'igboština', 'cy': 'Igbo', 'da': 'igbo', 'dav': 'Kiigbo', 'de': 'Igbo', 'dje': 'Iboo senni', 'dsb': 'igbo', 'dyo': 'igbo', 'dz': 'ཨིག་བོ་ཁ', 'ebu': 'Kĩigbo', 'ee': 'igbogbe', 'el': 'Ίγκμπο', 'el-polyton': 'Ίγκμπο', 'en': 'Igbo', 'es': 'igbo', 'et': 'ibo', 'eu': 'igboera', 'ewo': 'Ǹkɔ́bɔ ibó', 'fa': 'ایگبویی', 'ff': 'Igiboore', 'ff-Adlm': '𞤋𞤦𞤮𞥅𞤪𞤫', 'ff-Latn': 'Igiboore', 'fi': 'igbo', 'fil': 'Igbo', 'fo': 'igbo', 'fr': 'igbo', 'fur': 'igbo', 'fy': 'Igbo', 'ga': 'Íogbóis', 'gd': 'Igbo', 'gl': 'igbo', 'gsw': 'Igbo', 'gu': 'ઇગ્બો', 'guz': 'Kiigbo', 'ha': 'Igbo', 'he': 'איגבו', 'hi': 'ईग्बो', 'hi-Latn': 'Igbo', 'hr': 'igbo', 'hsb': 'igbo', 'ht': 'igbo', 'hu': 'igbó', 'hy': 'իգբո', 'ia': 'igbo', 'id': 'Igbo', 'ig': 'Igbo', 'is': 'ígbó', 'it': 'igbo', 'ja': 'イボ語', 'jmc': 'Kyiigbo', 'jv': 'Iqbo', 'ka': 'იგბო', 'kab': 'Tigbut', 'kam': 'Kiigbo', 'kde': 'Chiigbo', 'kea': 'ibo', 'kgp': 'igmo', 'khq': 'Iboo senni', 'ki': 'Kiigbo', 'kk': 'игбо тілі', 'kk-Cyrl': 'игбо тілі', 'kln': 'kutitab Igbo', 'km': 'អ៊ីកបូ', 'kn': 'ಇಗ್ಬೊ', 'ko': '이그보어', 'kok': 'इग्बो', 'kok-Deva': 'इग्बो', 'ks': 'اِگبو', 'ks-Arab': 'اِگبو', 'ksb': 'Kiigbo', 'ksf': 'riigbo', 'ksh': 'Igbo', 'ku': 'îgboyî', 'ky': 'игбочо', 'lag': 'Kiígibo', 'lb': 'Igbo-Sprooch', 'lg': 'Luyibo', 'lkt': 'Igbo Iyápi', 'lld': 'igbo', 'ln': 'igbo', 'lo': 'ອິກໂບ', 'lrc': 'ئیگبو', 'lt': 'igbų', 'lu': 'Igbo', 'luo': 'Kiigbo', 'luy': 'Kiigbo', 'lv': 'igbo', 'mai': 'इगबो', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ Igbo', 'mer': 'Kĩigbo', 'mfe': 'igbo', 'mg': 'igbo', 'mgh': 'Igbo', 'mi': 'Ikapo', 'mk': 'игбо', 'ml': 'ഇഗ്ബോ', 'mn': 'игбо', 'mr': 'ईग्बो', 'ms': 'Igbo', 'mt': 'Igbo', 'mua': 'igbo', 'my': 'အစ္ဂဘို', 'mzn': 'ایگبو', 'naq': 'Igbogowab', 'nb': 'ibo', 'nd': 'isi-Igbo', 'ne': 'इग्बो', 'nl': 'Igbo', 'nmg': 'Kiɛl ikbo', 'nn': 'ibo', 'no': 'ibo', 'nqo': 'ߊߜߏߞߊ߲', 'nus': 'Thok i̱gboni', 'nyn': 'Oruibo', 'or': 'ଇଗବୋ', 'pa': 'ਇਗਬੋ', 'pa-Guru': 'ਇਗਬੋ', 'pcm': 'Igbo Lángwej', 'pl': 'igbo', 'ps': 'اګبو', 'pt': 'igbo', 'qu': 'Igbo Simi', 'rm': 'igbo', 'rn': 'Ikigubo', 'ro': 'igbo', 'rof': 'Kiigbo', 'ru': 'игбо', 'rwk': 'Kyiigbo', 'saq': 'Kiigbo', 'sbp': 'Ishihigibo', 'sc': 'igbo', 'sd': 'اگبو', 'sd-Arab': 'اگبو', 'seh': 'ibo', 'ses': 'Iboo senni', 'sg': 'Ïgbö', 'shi': 'ⵜⵉⴳⴱⵓⵜ', 'shi-Latn': 'Tigbut', 'shi-Tfng': 'ⵜⵉⴳⴱⵓⵜ', 'si': 'ඉග්බෝ', 'sk': 'igboština', 'sl': 'igboščina', 'smn': 'igbokielâ', 'sn': 'chiIgbo', 'so': 'Igbo', 'sq': 'igboisht', 'sr': 'игбо', 'sr-Cyrl': 'игбо', 'sr-Latn': 'igbo', 'sv': 'igbo', 'sw': 'Kiigbo', 'ta': 'இக்போ', 'te': 'ఇగ్బో', 'teo': 'Kiigbo', 'tg': 'игбо', 'th': 'อิกโบ', 'ti': 'ኢግቦ', 'tk': 'igbo dili', 'to': 'lea fakaʻikipō', 'tr': 'İbo dili', 'tt': 'игбо', 'twq': 'Iboo senni', 'tzm': 'Tigbut', 'ug': 'ئىگبوچە', 'uk': 'ігбо', 'ur': 'اِگبو', 'uz': 'igbo', 'uz-Cyrl': 'игбо', 'uz-Latn': 'igbo', 'vai': 'ꔤꕼ', 'vai-Latn': 'Ígbo', 'vai-Vaii': 'ꔤꕼ', 'vec': 'igbo', 'vi': 'Tiếng Igbo', 'vun': 'Kyiigbo', 'wae': 'Igbo', 'wo': 'Igbo', 'xnr': 'ईग्बो', 'xog': 'Oluyibo', 'yav': 'íbo', 'yo': 'Èdè Yíbò', 'yrl': 'igibu', 'yue': '伊布文', 'yue-Hans': '伊布文', 'yue-Hant': '伊布文', 'zgh': 'ⵜⵉⴳⴱⵓⵜ', 'zh': '伊博语', 'zh-Hans': '伊博语', 'zh-Hant': '伊布文', 'zu': 'isi-Igbo'},
    'igb': {'en': 'Ebira'},
    'ige': {'en': 'Igede'},
    'igg': {'en': 'Igana'},
    'igl': {'en': 'Igala'},
    'igm': {'en': 'Kanggape'},
    'ign': {'en': 'Ignaciano'},
    'igo': {'en': 'Isebe'},
    'igs': {'en': 'Interglossa'},
    'igw': {'en': 'Igwe'},
    'ihb': {'en': 'Iha Based Pidgin'},
    'ihi': {'en': 'Ihievbe'},
    'ihp': {'en': 'Iha'},
    'ihw': {'en': 'Bidhawal'},
    'ii': {'af': 'Sichuan Yi', 'am': 'ሲቹዋን ዪ', 'ar': 'السيتشيون يي', 'as': 'ছিচুৱান ই', 'ast': 'yi de Sichuán', 'az': 'siçuan yi', 'az-Latn': 'siçuan yi', 'bal-Latn': 'Sichuái Yi', 'be': 'сычуаньская йі', 'be-tarask': 'сычуаньская йі', 'bg': 'съчуански йи', 'bgn': 'یی سیچوان', 'bn': 'সিচুয়ান য়ি', 'br': 'yieg Sichuan', 'brx': 'शीचुआन यी', 'bs': 'sičuan ji', 'bs-Cyrl': 'сечуан ји', 'bs-Latn': 'sičuan ji', 'ca': 'yi sichuan', 'ca-ES-valencia': 'yi sichuan', 'ccp': '𑄥𑄨𑄥𑄪𑄠𑄚𑄴𑄠𑄨', 'ce': 'сычуань', 'chr': 'ᏏᏧᏩᏂ Ᏹ', 'ckb': 'سیچوان یی', 'cs': 'iština (sečuánská)', 'cy': 'Nwosw', 'da': 'sichuan yi', 'de': 'Yi', 'dsb': 'sichuan yi', 'el': 'Σίτσουαν Γι', 'el-polyton': 'Σίτσουαν Γι', 'en': 'Sichuan Yi', 'es': 'yi de Sichuán', 'et': 'nuosu', 'eu': 'Sichuango yiera', 'fa': 'یی سیچوان', 'ff-Adlm': '𞤅𞤭𞤧𞤵𞤢𞤲𞤪𞤫 𞤒𞤭𞥅', 'fi': 'sichuanin-yi', 'fil': 'Sichuan Yi', 'fo': 'sichuan yi', 'fr': 'yi du Sichuan', 'fy': 'Sichuan Yi', 'ga': 'Ís Shichuan', 'gd': 'Yi Sichuan', 'gl': 'yi sichuanés', 'gsw': 'Sezuanischs Yi', 'gu': 'સિચુઆન યી', 'ha': 'Sichuan Yi', 'he': 'סצ׳ואן יי', 'hi': 'सिचुआन यी', 'hi-Latn': 'Sichuan Yi', 'hr': 'sichuan ji', 'hsb': 'sichuan yi', 'ht': 'yi du Sichuan', 'hu': 'szecsuán ji', 'hy': 'սիչուան', 'ia': 'yi de Sichuan', 'id': 'Sichuan Yi', 'ig': 'Sichuan Yi', 'ii': 'ꆈꌠꉙ', 'is': 'sísúanjí', 'it': 'sichuan yi', 'ja': '四川イ語', 'jv': 'Sichuan Yi', 'ka': 'სიჩუანის ი', 'kea': 'nuosu', 'kgp': 'sisuvỹ ji', 'kk': 'сычуан и тілі', 'kk-Cyrl': 'сычуан и тілі', 'km': 'ស៊ីឈាន់យី', 'kn': 'ಸಿಚುಅನ್ ಯಿ', 'ko': '쓰촨 이어', 'kok': 'सिच्युआन यी', 'kok-Deva': 'सिच्युआन यी', 'ks': 'سِچوان یٖی', 'ks-Arab': 'سِچوان یٖی', 'ku': 'yîyîya siçuwayî', 'ky': 'сычуань йиче', 'lb': 'Sichuan Yi', 'lld': 'sichuan yi', 'lo': 'ເສສວນ ອີ', 'lrc': 'سی چوان یی', 'lt': 'sičuan ji', 'lv': 'Sičuaņas ji', 'mai': 'सिचुआन यी', 'mi': 'Hīhuana Eī', 'mk': 'сичуан ји', 'ml': 'ഷുവാൻയി', 'mn': 'сычуань и', 'mr': 'सिचुआन यी', 'ms': 'Sichuan Yi', 'mt': 'Sichuan Yi', 'my': 'စီချွမ် ရီ', 'mzn': 'سیچوئان یی', 'nb': 'sichuan-yi', 'ne': 'सिचुआन यि', 'nl': 'Yi', 'nn': 'sichuan-yi', 'no': 'sichuan-yi', 'nqo': 'ߛߌߛߎߥߊ߲߫ ߦߌߞߊ߲', 'or': 'ସିଚୁଆନ୍ ୟୀ', 'pa': 'ਸਿਚੁਆਨ ਯੀ', 'pa-Guru': 'ਸਿਚੁਆਨ ਯੀ', 'pcm': 'Síchuan Yi', 'pl': 'syczuański', 'ps': 'سیچیان یی', 'pt': 'sichuan yi', 'qu': 'Yi Simi', 'rm': 'sichuan yi', 'ro': 'yi din Sichuan', 'ru': 'носу', 'sc': 'sichuan yi', 'sd': 'سچوان يي', 'sd-Arab': 'سچوان يي', 'si': 'සිචුආන් යී', 'sk': 's’čchuanská iovčina', 'sl': 'sečuanska jiščina', 'so': 'Sijuwan Yi', 'sq': 'sishuanisht', 'sr': 'сечуански ји', 'sr-Cyrl': 'сечуански ји', 'sr-Latn': 'sečuanski ji', 'sv': 'szezuan i', 'sw': 'Kiyi cha Sichuan', 'ta': 'சிசுவான் ஈ', 'te': 'శిషువన్ ఈ', 'th': 'เสฉวนยี่', 'ti': 'ሲችዋን ዪ', 'tk': 'syçuan-i dili', 'to': 'lea fakasisiuani-ī', 'tr': 'Sichuan Yi', 'ug': 'يىچە (سىچۈەن)', 'uk': 'сичуаньська ї', 'ur': 'سچوان ای', 'uz': 'sichuan', 'uz-Latn': 'sichuan', 'vec': 'yi de Sichuan', 'vi': 'Tiếng Di Tứ Xuyên', 'xnr': 'सिचुआन यी', 'yo': 'Ṣíkuán Yì', 'yrl': 'sixuã yi', 'yue': '四川彝文', 'yue-Hans': '四川彝文', 'yue-Hant': '四川彝文', 'zh': '凉山彝语', 'zh-Hans': '凉山彝语', 'zh-Hant': '四川彝文', 'zu': 'isi-Sichuan Yi'},
    'iin': {'en': 'Thiin'},
    'iir': {'en': 'Indo-Iranian languages'},
    'iir-nur-pro': {'en': 'Proto-Nuristani'},
    'iir-pro': {'en': 'Proto-Indo-Iranian'},
    'ijc': {'en': 'Izon'},
    'ije': {'en': 'Biseni'},
    'ijj': {'en': 'Ede Ije'},
    'ijn': {'en': 'Kalabari'},
    'ijo': {'en': 'Ijo languages'},
    'ijs': {'en': 'Southeast Ijo'},
    'ik': {'am': 'እኑፒያቅኛ', 'ar': 'الإينبياك', 'ast': 'inupiaq', 'az': 'inupiaq', 'az-Latn': 'inupiaq', 'bg': 'инупиак', 'bn': 'ইনুপিয়াক', 'br': 'inupiaq', 'brx': 'इनुपियाक़्', 'bs': 'inupiak', 'bs-Cyrl': 'унупиак', 'bs-Latn': 'inupiak', 'ca': 'inupiak', 'ca-ES-valencia': 'inupiak', 'ccp': '𑄃𑄨𑄚𑄪𑄛𑄨𑄠𑄇𑄴', 'cs': 'inupiakština', 'cy': 'Inwpiaceg', 'da': 'inupiaq', 'de': 'Inupiak', 'dsb': 'inupiak', 'el': 'Ινουπιάκ', 'el-polyton': 'Ινουπιάκ', 'en': 'Inupiaq', 'eo': 'eskima', 'es': 'inupiaq', 'et': 'injupiaki', 'fa': 'اینوپیک', 'fi': 'inupiaq', 'fr': 'inupiaq', 'fur': 'inupiaq', 'fy': 'Inupiaq', 'ga': 'Iniúipiaicis', 'gd': 'Inupiaq', 'gsw': 'Inupiak', 'gu': 'ઇનુપિયાક', 'he': 'אינופיאק', 'hi': 'इनुपियाक्', 'hi-Latn': 'Inupiaq', 'hr': 'inupiaq', 'hsb': 'inupiak', 'ht': 'inupiaq', 'hu': 'inupiak', 'id': 'Inupiak', 'is': 'ínúpíak', 'it': 'inupiak', 'ja': 'イヌピアック語', 'kgp': 'ĩnũpijake', 'kn': 'ಇನುಪಿಯಾಕ್', 'ko': '이누피아크어', 'kok': 'इनूपेयाक्', 'kok-Deva': 'इनूपेयाक्', 'ks': 'اِنُپِیاک', 'ks-Arab': 'اِنُپِیاک', 'lb': 'Inupiak', 'lo': 'ອິນນູປຽກ', 'lt': 'inupiakų', 'lv': 'inupiaku', 'mk': 'инупијачки', 'ml': 'ഇനുപിയാക്', 'mr': 'इनूपियाक', 'mt': 'Inupjak', 'nb': 'inupiak', 'ne': 'इनुपिआक्', 'nl': 'Inupiaq', 'nn': 'inupiak', 'no': 'inupiak', 'or': 'ଇନୁପିୟାକ୍', 'pl': 'inupiak', 'pt': 'inupiaque', 'rm': 'inupiak', 'ro': 'inupiak', 'ru': 'инупиак', 'sk': 'inupik', 'sl': 'inupiaščina', 'sr': 'инупик', 'sr-Cyrl': 'инупик', 'sr-Latn': 'inupik', 'sv': 'inupiak', 'ta': 'இனுபியாக்', 'te': 'ఇనుపైయాక్', 'th': 'อีนูเปียก', 'to': 'lea fakaʻinupiaki', 'tr': 'İnyupikçe', 'ug': 'ئىنۇپىكچە', 'uk': 'інупіак', 'vi': 'Tiếng Inupiaq', 'yrl': 'inupiaki', 'yue': '依奴皮維克文', 'yue-Hans': '依奴皮维克文', 'yue-Hant': '依奴皮維克文', 'zh': '伊努皮克语', 'zh-Hans': '伊努皮克语', 'zh-Hant': '依奴皮維克文'},
    'ike': {'en': 'Eastern Canadian Inuktitut'},
    'iki': {'en': 'Iko'},
    'ikk': {'en': 'Ika'},
    'ikl': {'en': 'Ikulu'},
    'iko': {'en': 'Olulumo-Ikom'},
    'ikp': {'en': 'Ikpeshi'},
    'ikr': {'en': 'Ikaranggal'},
    'iks': {'en': 'Inuit Sign Language'},
    'ikt': {'af': 'Wes-Kanadese Inoektitoet', 'am': 'የምዕራባዊ ካናዳ ኢኑክቲቱት', 'ar': 'الإنكتيتوتية الكندية الغربية', 'as': 'ৱেষ্টাৰ্ণ কানাডিয়ান ইনক্টিটুট', 'az': 'qərbi Kanada inuktitutu', 'az-Latn': 'qərbi Kanada inuktitutu', 'be': 'заходнеканадская інуктытут', 'be-tarask': 'заходнеканадская інуктытут', 'bg': 'западноканадски инуктитут', 'bn': 'পশ্চিম কানাডিয় ইনুক্টিটুট', 'br': 'inuktitut Kanada ar Cʼhornôg', 'brx': 'सोनाबारि कानाडियान इनुक्तितुत', 'bs': 'zapadnokanadski inuktitut', 'bs-Cyrl': 'западно канадски инуктитут', 'bs-Latn': 'zapadnokanadski inuktitut', 'ca': 'inuktitut occidental canadenc', 'ca-ES-valencia': 'inuktitut occidental canadenc', 'chr': 'ᏭᏕᎵᎬ ᎨᎾᏓ ᎢᏄᎩᏘᏚᏘ', 'cs': 'inuktitutština (západokanadská)', 'cy': 'Inuktitut Canadaidd Gorllewinol', 'da': 'vestcanadisk inuktitut', 'de': 'Westkanadisches Inuktitut', 'dsb': 'pódwjacornokanadiska inuktitutšćina', 'el': 'Ινουκτιτούτ Δυτικού Καναδά', 'el-polyton': 'Ινουκτιτούτ Δυτικού Καναδά', 'en': 'Western Canadian Inuktitut', 'es': 'inuit del oeste de Canadá', 'et': 'Lääne-Kanada inuktituti', 'eu': 'Kanada mendebaldeko inuitera', 'fa': 'اینوکتیتوت غرب کانادا', 'ff-Adlm': '𞤋𞤲𞤵𞤳𞤼𞤵𞤪𞤫 𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫 𞤑𞤢𞤲𞤢𞤣𞤢𞥄', 'fi': 'Länsi-Kanadan inuktitut', 'fil': 'Kanlurang Canadian Inuktitut', 'fr': 'inuktitut de l’Ouest canadien', 'ga': 'Ionúitis Iarthar Cheanada', 'gd': 'Inuktitut Shiarach Chanada', 'gl': 'inuktitut canadense occidental', 'gu': 'પશ્ચિમ કેનેડિયન ઇનુકિટ્યુટ', 'ha': 'Inuktitut na Yammacin Kanada', 'he': 'אינוקטיטוט במערב קנדה', 'hi': 'पश्चिमी कनाडाई इनुक्टिटुट', 'hi-Latn': 'Western Canadian Inuktitut', 'hr': 'zapadnokanadski inuktitut', 'hsb': 'zapadnokanadska inuktitutšćina', 'ht': 'inuktitut de l’Ouest canadien', 'hu': 'nyugat-kanadai inuit', 'hy': 'արևմտականադական ինուկտիտուտ', 'ia': 'inuktitut del west canadian', 'id': 'Inuktitut Kanada Barat', 'ig': 'Westarn Canadian Inuktitut', 'is': 'vestur-kanadískt inúktitút', 'it': 'inuktitut canadese occidentale', 'ja': 'イヌイナクトゥン語', 'jv': 'Kanada Inuktitut Sisih Kulon', 'ka': 'დასავლეთ-კანადური ინუკტიტუტი', 'kk': 'батыс канадалық инуктитут тілі', 'kk-Cyrl': 'батыс канадалық инуктитут тілі', 'km': 'អ៊ីនុកទីទុត\u200bកាណាដា\u200bប៉ែកខាងលិច', 'kn': 'ಪಶ್ಚಿಮ ಕೆನಡಿಯನ್ ಇನುಕ್ಟಿಟುಟ್', 'ko': '캐나다 서부 이누크티투트어', 'kok': 'पश्चिमी कॅनडियन इनक्तितुत', 'kok-Deva': 'पश्चिमी कॅनडियन इनक्तितुत', 'ku': 'inuvialuktun', 'ky': 'инуктитутча (Канада)', 'lo': 'ອິນັກທິທັດຄານາດາຕາເວັນຕົກ', 'lt': 'vakarų kanadiečių inuktitutas', 'lv': 'Rietumkanādas inuītu', 'mai': 'पश्चिमी कनाडाई इनुकतुत', 'mi': 'Inukitetūta Kānata ki te Uru', 'mk': 'западноканадски инуктитут', 'ml': 'വെസ്റ്റേൺ കനേഡിയൻ ഇനുക്ടിറ്റൂറ്റ്', 'mn': 'баруун канадын инуктитут', 'mr': 'वेस्टर्न कॅनेडियन इनुकिटुट', 'ms': 'Inuktitut Kanada Barat', 'my': 'အနောက် ကနေဒီယန် အီနုတီတွတ်', 'nb': 'vestlig kanadisk inuktitut', 'ne': 'पश्चिमी क्यानेडेली इनुक्टिटुट', 'nl': 'Westelijk Canadees Inuktitut', 'nn': 'vestleg kanadisk inuktitut', 'no': 'vestlig kanadisk inuktitut', 'or': 'ପାଶ୍ଚାତ୍ୟ କାନାଡିୟ ଇନୁକ୍ଟିଟ', 'pa': 'ਪੱਛਮੀ ਕੈਨੇਡੀਅਨ ਇਨੂਕਟੀਟੂਟ', 'pa-Guru': 'ਪੱਛਮੀ ਕੈਨੇਡੀਅਨ ਇਨੂਕਟੀਟੂਟ', 'pcm': 'Wẹ́stán Kánádá Inuktítut', 'pl': 'inuktitut zachodniokanadyjski', 'ps': 'مغربی کینیډین انوکټیټ', 'pt': 'inuktitut canadense ocidental', 'qu': 'Inuktitut Simi (Canadá occidental)', 'ro': 'inuktitut canadiană occidentală', 'ru': 'восточноканадский инуктитут', 'sc': 'inuktitut canadesu otzidentale', 'sd': 'مغربي ڪينيڊين انوڪٽيٽ', 'sd-Arab': 'مغربي ڪينيڊين انوڪٽيٽ', 'si': 'බටහිර කැනේඩියානු ඉනුක්ටිටුට්', 'sk': 'inuktitut (západná Kanada)', 'sl': 'zahodna kanadska inuktituščina', 'so': 'Western Canadian Inuktitut', 'sq': 'inuktitutishte kanadeze perëndimore', 'sr': 'западноканадски инуктитут', 'sr-Cyrl': 'западноканадски инуктитут', 'sr-Latn': 'zapadnokanadski inuktitut', 'sv': 'inuktun', 'sw': 'Kiinuktituti cha Kanada Magharibi', 'ta': 'மேற்கு கனடிய இனுக்டிடுட்', 'te': 'పశ్చిమ కెనేడియన్ ఇన్నూక్\u200cటిటూట్', 'th': 'อินุกติตุตแคนาดาตะวันตก', 'ti': 'ምዕራባዊ ካናዳዊ ኢናክቲቱት', 'tk': 'Günorta Kanada iniktitut dili', 'to': 'lea fakaʻinuketītuti-kānata-hihifo', 'tr': 'Batı Kanada İnuktitut dili', 'uk': 'західноканадська інуктитут', 'ur': 'مغربی کینیڈین اینُکٹیٹٹ', 'uz': 'sharqiy-kanada inuktitut', 'uz-Latn': 'sharqiy-kanada inuktitut', 'vec': 'inuktitut canadeze osidentale', 'vi': 'Tiếng Inuktitut miền Tây Canada', 'yo': 'Èdè Iwoorun Inutitu ti Kanada', 'yue': '加拿大西部因紐特文', 'yue-Hans': '加拿大西部因纽特文', 'yue-Hant': '加拿大西部因紐特文', 'zh': '西加拿大因纽特语', 'zh-Hans': '西加拿大因纽特语', 'zh-Hant': '西加拿大因紐特語', 'zu': 'Western Canadian Inuktitut'},
    'ikv': {'en': 'Iku-Gora-Ankwa'},
    'ikw': {'en': 'Ikwere'},
    'ikx': {'en': 'Ik'},
    'ikz': {'en': 'Ikizu'},
    'ila': {'en': 'Ile Ape'},
    'ilb': {'en': 'Ila'},
    'ilg': {'en': 'Garig-Ilgar'},
    'ili': {'en': 'Ili Turki'},
    'ilk': {'en': 'Ilongot'},
    'ill': {'en': 'Iranun'},
    'ilm': {'en': 'Iranun (Malaysia)'},
    'ilo': {'af': 'Iloko', 'am': 'ኢሎኮ', 'ar': 'الإيلوكو', 'as': 'ইলোকো', 'ast': 'iloko', 'az': 'iloko', 'az-Cyrl': 'илоко', 'az-Latn': 'iloko', 'be': 'ілакана', 'be-tarask': 'ілакана', 'bg': 'илоко', 'bgn': 'ایلوکانوی', 'bn': 'ইলোকো', 'br': 'ilokanoeg', 'brx': 'इल’क’', 'bs': 'iloko', 'bs-Cyrl': 'илоко', 'bs-Latn': 'iloko', 'ca': 'ilocano', 'ca-ES-valencia': 'ilocano', 'ccp': '𑄃𑄨𑄣𑄮𑄇𑄮', 'ce': 'илоко', 'chr': 'ᎢᎶᎪ', 'ckb': 'ئیلۆکۆ', 'cs': 'ilokánština', 'cy': 'Ilocaneg', 'da': 'iloko', 'de': 'Ilokano', 'dsb': 'ilokošćina', 'el': 'Ιλόκο', 'el-polyton': 'Ιλόκο', 'en': 'Iloko', 'es': 'ilocano', 'et': 'iloko', 'eu': 'ilocanoera', 'fa': 'ایلوکویی', 'ff-Adlm': '𞤋𞤤𞤮𞤳𞤮𞥅𞤪𞤫', 'fi': 'iloko', 'fil': 'Iloko', 'fo': 'iloko', 'fr': 'ilocano', 'fy': 'Iloko', 'ga': 'Ileacáinis', 'gd': 'Iloko', 'gl': 'ilocano', 'gsw': 'Ilokano', 'gu': 'ઇલોકો', 'ha': 'Ikolo', 'he': 'אילוקו', 'hi': 'इलोको', 'hi-Latn': 'Iloko', 'hr': 'iloko', 'hsb': 'ilokošćina', 'ht': 'ilocano', 'hu': 'ilokó', 'hy': 'իլոկերեն', 'ia': 'ilocano', 'id': 'Iloko', 'ig': 'Iloko', 'is': 'ílokó', 'it': 'ilocano', 'ja': 'イロカノ語', 'jv': 'Iloko', 'ka': 'ილოკო', 'kgp': 'irukỹnũ', 'kk': 'илоко тілі', 'kk-Cyrl': 'илоко тілі', 'km': 'អ៊ីឡូកូ', 'kn': 'ಇಲ್ಲಿಕೋ', 'ko': '이로코어', 'kok': 'इलोको', 'kok-Deva': 'इलोको', 'ks': 'اِلوکو', 'ks-Arab': 'اِلوکو', 'ksh': 'Ilokahnesch', 'ku': 'îlokanoyî', 'ky': 'илокочо', 'lb': 'Ilokano-Sprooch', 'lo': 'ໄອໂລໂກ', 'lt': 'ilokų', 'lv': 'iloku', 'mai': 'इलोको', 'mi': 'Iroko', 'mk': 'илокански', 'ml': 'ഇലോകോ', 'mn': 'илоко', 'mr': 'इलोको', 'ms': 'Iloko', 'mt': 'Iloko', 'my': 'အီလိုကို', 'nb': 'iloko', 'ne': 'इयोको', 'nl': 'Iloko', 'nn': 'iloko', 'no': 'iloko', 'or': 'ଇଲୋକୋ', 'pa': 'ਇਲੋਕੋ', 'pa-Guru': 'ਇਲੋਕੋ', 'pcm': 'Ilokó', 'pl': 'ilokano', 'ps': 'الوکو', 'pt': 'ilocano', 'qu': 'Iloko Simi', 'rm': 'ilocano', 'ro': 'iloko', 'ru': 'илоко', 'sc': 'ilocanu', 'sd': 'الوڪو', 'sd-Arab': 'الوڪو', 'si': 'ඉලොකො', 'sk': 'ilokánčina', 'sl': 'ilokanščina', 'smn': 'ilocano', 'so': 'Iloko', 'sq': 'ilokoisht', 'sr': 'илоко', 'sr-Cyrl': 'илоко', 'sr-Latn': 'iloko', 'sv': 'iloko', 'sw': 'Kiilocano', 'ta': 'இலோகோ', 'te': 'ఐలోకో', 'th': 'อีโลโก', 'ti': 'ኢሎካኖ', 'tk': 'iloko dili', 'to': 'lea fakaʻiloko', 'tr': 'Iloko', 'ug': 'ئىلوكانوچە', 'uk': 'ілоканська', 'ur': 'ایلوکو', 'uz': 'iloko', 'uz-Cyrl': 'илоко', 'uz-Latn': 'iloko', 'vec': 'ilocan', 'vi': 'Tiếng Iloko', 'xnr': 'इलोको', 'yo': 'Èdè Iloko', 'yrl': 'irukanu', 'yue': '伊洛闊文', 'yue-Hans': '伊洛阔文', 'yue-Hant': '伊洛闊文', 'zh': '伊洛卡诺语', 'zh-Hans': '伊洛卡诺语', 'zh-Hant': '伊洛闊文', 'zu': 'isi-Iloko', 'ilo': 'Ilokano'},
    'ilp': {'en': 'Iranun (Philippines)'},
    'ils': {'en': 'International Sign'},
    'ilu': {'en': "Ili'uun"},
    'ilv': {'en': 'Ilue'},
    'ilw': {'en': 'Talur'},
    'ima': {'en': 'Mala Malasar'},
    'ime': {'en': 'Imeraguen'},
    'imi': {'en': 'Anamgura'},
    'iml': {'en': 'Miluk'},
    'imn': {'en': 'Imonda'},
    'imo': {'en': 'Imbongu'},
    'imr': {'en': 'Imroing'},
    'ims': {'en': 'Marsian'},
    'imy': {'en': 'Milyan'},
    'in': {'en': 'Indonesian'},
    'inb': {'en': 'Inga'},
    'inc': {'en': 'Indic languages'},
    'inc-dar-pro': {'en': 'Proto-Dardic'},
    'inc-ogu': {'en': 'Old Gujarati'},
    'inc-pro': {'en': 'Proto-Indo-Aryan'},
    'ine': {'en': 'Indo-European languages'},
    'ine-ana-pro': {'en': 'Proto-Anatolian'},
    'ine-bsl-pro': {'en': 'Proto-Balto-Slavic'},
    'ine-pro': {'en': 'Proto-Indo-European'},
    'ine-toc-pro': {'en': 'Proto-Tocharian'},
    'ing': {'en': "Degexit'an"},
    'inh': {'af': 'Ingush', 'am': 'ኢንጉሽ', 'ar': 'الإنجوشية', 'as': 'ইংগুচ', 'ast': 'ingush', 'az': 'inquş', 'az-Cyrl': 'ингуш', 'az-Latn': 'inquş', 'be': 'інгушская', 'be-tarask': 'інгушская', 'bg': 'ингушетски', 'bgn': 'اینگوشی', 'bn': 'ইঙ্গুশ', 'br': 'ingoucheg', 'brx': 'ईंगुष', 'bs': 'ingušetski', 'bs-Cyrl': 'ингвишки', 'bs-Latn': 'ingušetski', 'ca': 'ingúix', 'ca-ES-valencia': 'ingúix', 'ccp': '𑄃𑄨𑄁𑄉𑄪𑄌𑄴', 'ce': 'гӀалгӀайн', 'chr': 'ᎢᏂᎫᏏ', 'ckb': 'ئینگووش', 'cs': 'inguština', 'cy': 'Ingwsieg', 'da': 'ingush', 'de': 'Inguschisch', 'dsb': 'ingushišćina', 'el': 'Ινγκούς', 'el-polyton': 'Ινγκούς', 'en': 'Ingush', 'es': 'ingush', 'et': 'inguši', 'eu': 'ingushera', 'fa': 'اینگوشی', 'ff-Adlm': '𞤋𞤲𞤺𞤮𞤧𞤫𞥅𞤪𞤫', 'fi': 'inguuši', 'fil': 'Ingush', 'fo': 'inguish', 'fr': 'ingouche', 'fy': 'Ingoesj', 'ga': 'Iongúis', 'gd': 'Ingush', 'gl': 'inguxo', 'gsw': 'Inguschisch', 'gu': 'ઇંગુશ', 'ha': 'Ingush', 'he': 'אינגושית', 'hi': 'इंगुश', 'hi-Latn': 'Ingush', 'hr': 'ingušetski', 'hsb': 'ingushišćina', 'ht': 'ingouche', 'hu': 'ingus', 'hy': 'ինգուշերեն', 'ia': 'ingush', 'id': 'Ingushetia', 'ig': 'Ingush', 'is': 'ingús', 'it': 'ingush', 'ja': 'イングーシ語', 'jv': 'Ingus', 'ka': 'ინგუშური', 'kgp': 'ĩnguse', 'kk': 'ингуш тілі', 'kk-Cyrl': 'ингуш тілі', 'km': 'អ៊ិនហ្គូស', 'kn': 'ಇಂಗುಷ್', 'ko': '인귀시어', 'kok': 'इंगूश', 'kok-Deva': 'इंगूश', 'ks': 'اِنگُش', 'ks-Arab': 'اِنگُش', 'ksh': 'Engjuschesch', 'ku': 'îngûşî', 'ky': 'ингушча', 'lb': 'Inguschesch', 'lkt': 'Ingush Iyápi', 'lo': 'ອິນກັຊ', 'lt': 'ingušų', 'lv': 'ingušu', 'mai': 'इंगुश', 'mi': 'Inguihi', 'mk': 'ингушки', 'ml': 'ഇങ്കുഷ്', 'mn': 'ингуш', 'mr': 'इंगुश', 'ms': 'Ingush', 'mt': 'Ingush', 'my': 'အင်ဂုရှ်', 'nb': 'ingusjisk', 'ne': 'इन्गस', 'nl': 'Ingoesjetisch', 'nn': 'ingusjisk', 'no': 'ingusjisk', 'or': 'ଇଁଙ୍ଗୁଶ୍', 'os': 'мӕхъӕлон', 'pa': 'ਇੰਗੁਸ਼', 'pa-Guru': 'ਇੰਗੁਸ਼', 'pcm': 'Inguísh Lángwej', 'pl': 'inguski', 'ps': 'انگش', 'pt': 'inguche', 'qu': 'Ingush Simi', 'rm': 'ingush', 'ro': 'ingușă', 'ru': 'ингушский', 'sc': 'ingùsciu', 'sd': 'انگش', 'sd-Arab': 'انگش', 'si': 'ඉන්ගුෂ්', 'sk': 'inguština', 'sl': 'inguščina', 'smn': 'inguškielâ', 'so': 'Ingush', 'sq': 'ingushisht', 'sr': 'ингушки', 'sr-Cyrl': 'ингушки', 'sr-Latn': 'inguški', 'sv': 'ingusjiska', 'sw': 'Kiingush', 'ta': 'இங்குஷ்', 'te': 'ఇంగుష్', 'th': 'อินกุช', 'ti': 'ኢንጉሽኛ', 'tk': 'inguş dili', 'to': 'lea fakaʻingusi', 'tr': 'İnguşça', 'ug': 'ئىنگۇشچە', 'uk': 'інгуська', 'ur': 'انگوش', 'uz': 'ingush', 'uz-Cyrl': 'ингушча', 'uz-Latn': 'ingush', 'vec': 'ingush', 'vi': 'Tiếng Ingush', 'xnr': 'इंगुश', 'yo': 'Èdè Ingusi', 'yrl': 'ĩguxi', 'yue': '印古什文', 'yue-Hans': '印古什文', 'yue-Hant': '印古什文', 'zh': '印古什语', 'zh-Hans': '印古什语', 'zh-Hant': '印古什文', 'zu': 'isi-Ingush'},
    'inj': {'en': 'Jungle Inga'},
    'inl': {'en': 'Indonesian Sign Language'},
    'inm': {'en': 'Minaean'},
    'inn': {'en': 'Isinai'},
    'ino': {'en': 'Inoke-Yate'},
    'inp': {'en': 'Iñapari'},
    'ins': {'en': 'Indian Sign Language'},
    'int': {'en': 'Intha'},
    'inz': {'en': 'Ineseño'},
    'io': {'af': 'Ido', 'am': 'ኢዶ', 'ar': 'الإيدو', 'as': 'ইডো', 'ast': 'ido', 'az': 'ido', 'az-Cyrl': 'идо', 'az-Latn': 'ido', 'bal-Latn': 'Idó', 'be': 'іда', 'be-tarask': 'іда', 'bg': 'идо', 'bgn': 'ایدوی', 'bn': 'ইডো', 'br': 'ido', 'brx': 'ईड’', 'bs': 'ido', 'bs-Cyrl': 'идо', 'bs-Latn': 'ido', 'ca': 'ido', 'ca-ES-valencia': 'ido', 'ccp': '𑄃𑄨𑄓𑄮', 'ce': 'идо', 'chr': 'ᎢᏙ', 'ckb': 'ئیدۆ', 'cs': 'ido', 'cy': 'Ido', 'da': 'ido', 'de': 'Ido', 'dsb': 'ido', 'el': 'Ίντο', 'el-polyton': 'Ίντο', 'en': 'Ido', 'es': 'ido', 'et': 'ido', 'eu': 'idoa', 'fa': 'ایدو', 'ff-Adlm': '𞤋𞤣𞤮𞥅𞤪𞤫', 'fi': 'ido', 'fil': 'Ido', 'fo': 'ido', 'fr': 'ido', 'fur': 'ido', 'fy': 'Ido', 'ga': 'Ídis', 'gd': 'Ido', 'gl': 'ido', 'gsw': 'Ido', 'gu': 'ઈડો', 'ha': 'Ido', 'he': 'אידו', 'hi': 'इडौ', 'hi-Latn': 'Ido', 'hr': 'ido', 'hsb': 'ido', 'ht': 'ido', 'hu': 'idó', 'hy': 'իդո', 'ia': 'ido', 'id': 'Ido', 'ig': 'Ido', 'io': 'Ido', 'is': 'ídó', 'it': 'ido', 'ja': 'イド語', 'jv': 'Ido', 'ka': 'იდო', 'kgp': 'ino', 'kk': 'идо тілі', 'kk-Cyrl': 'идо тілі', 'km': 'អ៊ីដូ', 'kn': 'ಇಡೊ', 'ko': '이도어', 'kok': 'इदो', 'kok-Deva': 'इदो', 'ks': 'اِڈو', 'ks-Arab': 'اِڈو', 'ksh': 'Ihdo', 'ku': 'îdoyî', 'ky': 'идочо', 'lb': 'Ido-Sprooch', 'lld': 'ido', 'lo': 'ອີໂດ', 'lt': 'ido', 'lv': 'ido', 'mai': 'ईडो', 'mi': 'Īto', 'mk': 'идо', 'ml': 'ഇഡോ', 'mn': 'идо', 'mr': 'इडौ', 'ms': 'Ido', 'mt': 'Ido', 'my': 'အီဒို', 'nb': 'ido', 'ne': 'इडो', 'nl': 'Ido', 'nn': 'ido', 'no': 'ido', 'or': 'ଇଡୋ', 'pa': 'ਇਡੂ', 'pa-Guru': 'ਇਡੂ', 'pcm': 'Idó Lángwej', 'pl': 'ido', 'ps': 'اڊو', 'pt': 'ido', 'qu': 'Ido Simi', 'rm': 'ido', 'ro': 'ido', 'ru': 'идо', 'sc': 'ido', 'sd': 'ادو', 'sd-Arab': 'ادو', 'si': 'ඉඩො', 'sk': 'ido', 'sl': 'ido', 'smn': 'ido', 'so': 'Ido', 'sq': 'idoisht', 'sr': 'идо', 'sr-Cyrl': 'идо', 'sr-Latn': 'ido', 'sv': 'ido', 'sw': 'Kiido', 'ta': 'இடோ', 'te': 'ఈడో', 'th': 'อีโด', 'ti': 'ኢዶ', 'tk': 'ido dili', 'to': 'lea fakaʻito', 'tr': 'Ido', 'ug': 'ئىدوچە', 'uk': 'ідо', 'ur': 'ایڈو', 'uz': 'ido', 'uz-Cyrl': 'идо', 'uz-Latn': 'ido', 'vec': 'ido', 'vi': 'Tiếng Ido', 'xnr': 'इडौ', 'yi': 'אידא', 'yo': 'Èdè Ido', 'yrl': 'idu', 'yue': '伊多文', 'yue-Hans': '伊多文', 'yue-Hant': '伊多文', 'zh': '伊多语', 'zh-Hans': '伊多语', 'zh-Hant': '伊多文', 'zu': 'isi-Ido'},
    'ior': {'en': 'Inor'},
    'iou': {'en': 'Tuma-Irumu'},
    'iow': {'en': 'Iowa-Oto'},
    'ipi': {'en': 'Ipili'},
    'ipo': {'en': 'Ipiko'},
    'iqu': {'en': 'Iquito'},
    'iqw': {'en': 'Ikwo'},
    'ira': {'en': 'Iranian languages'},
    'ira-azr': {'en': 'Old Azari'},
    'ira-azr-kls': {'en': 'Kalasuri'},
    'ira-azr-klt': {'en': 'Kilit'},
    'ira-pro': {'en': 'Proto-Iranian'},
    'ire': {'en': 'Iresim'},
    'irh': {'en': 'Irarutu'},
    'iri': {'en': 'Rigwe'},
    'irk': {'en': 'Iraqw'},
    'irn': {'en': 'Irántxe'},
    'iro': {'en': 'Iroquoian languages'},
    'iro-min': {'en': 'Mingo'},
    'iro-pro': {'en': 'Proto-Iroquoian'},
    'irr': {'en': 'Ir'},
    'iru': {'en': 'Irula'},
    'irx': {'en': 'Kamberau'},
    'iry': {'en': 'Iraya'},
    'is': {'af': 'Yslands', 'ak': 'Aeslande kasa', 'am': 'አይስላንድኛ', 'ar': 'الأيسلندية', 'as': 'আইচলেণ্ডিক', 'ast': 'islandés', 'az': 'island', 'az-Cyrl': 'исланд', 'az-Latn': 'island', 'bal-Latn': 'Islándi', 'be': 'ісландская', 'be-tarask': 'ісландская', 'bg': 'исландски', 'bgn': 'ایسلندی', 'bn': 'আইসল্যান্ডীয়', 'br': 'islandeg', 'brx': 'आइस्लेन्डिक', 'bs': 'islandski', 'bs-Cyrl': 'исландски', 'bs-Latn': 'islandski', 'ca': 'islandès', 'ca-ES-valencia': 'islandès', 'ccp': '𑄃𑄭𑄌𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄨𑄠𑄧', 'ce': 'исландхойн', 'chr': 'ᏧᏁᏍᏓᎸᎯᎢᎩ', 'ckb': 'ئیسلەندی', 'cs': 'islandština', 'cy': 'Islandeg', 'da': 'islandsk', 'de': 'Isländisch', 'dsb': 'islandšćina', 'dz': 'ཨ་ཡིས་ལེན་ཌིཀ་ཁ', 'ee': 'icelanɖgbe', 'el': 'Ισλανδικά', 'el-polyton': 'Ἰσλανδικά', 'en': 'Icelandic', 'en-Dsrt': '𐐌𐑅𐑊𐐰𐑌𐐼𐐮𐐿', 'eo': 'islanda', 'es': 'islandés', 'et': 'islandi', 'eu': 'islandiera', 'fa': 'ایسلندی', 'ff-Adlm': '𞤀𞤴𞤧𞤭𞤤𞤢𞤲𞤣𞤫𞥅𞤪𞤫', 'fi': 'islanti', 'fil': 'Icelandic', 'fo': 'íslendskt', 'fr': 'islandais', 'fur': 'islandês', 'fy': 'Yslâns', 'ga': 'Íoslainnis', 'gd': 'Innis Tìlis', 'gl': 'islandés', 'gsw': 'Iisländisch', 'gu': 'આઇસલેન્ડિક', 'ha': 'Yaren mutanen Iceland', 'he': 'איסלנדית', 'hi': 'आइसलैंडिक', 'hi-Latn': 'Icelandic', 'hr': 'islandski', 'hsb': 'islandšćina', 'ht': 'islandais', 'hu': 'izlandi', 'hy': 'իսլանդերեն', 'ia': 'islandese', 'id': 'Islandia', 'ig': 'Icelandic', 'is': 'íslenska', 'it': 'islandese', 'ja': 'アイスランド語', 'jv': 'Islandia', 'ka': 'ისლანდიური', 'kea': 'islandes', 'kgp': 'kukryr', 'kk': 'исланд тілі', 'kk-Cyrl': 'исланд тілі', 'km': 'អ៊ីស្លង់', 'kn': 'ಐಸ್\u200cಲ್ಯಾಂಡಿಕ್', 'ko': '아이슬란드어', 'kok': 'आयस्लान्डिक', 'kok-Deva': 'आयस्लान्डिक', 'ks': 'آیِسلینڈِک', 'ks-Arab': 'آیِسلینڈِک', 'ksh': 'Ißlänndesch', 'ku': 'îzlendî', 'kxv': 'aislanḍik', 'kxv-Deva': 'आइसलेंड िक', 'kxv-Latn': 'aislanḍik', 'kxv-Orya': 'ଆଇସଲାଣ୍ଡିକ୍', 'kxv-Telu': 'అఇస్లెండిక్', 'ky': 'исландча', 'lb': 'Islännesch', 'lkt': 'Iceland Iyápi', 'lld': 'islandesc', 'lo': 'ໄອສແລນດິກ', 'lrc': 'ئیسلأندی', 'lt': 'islandų', 'lv': 'islandiešu', 'mai': 'आइसलैंडी', 'mi': 'Tiorangi', 'mk': 'исландски', 'ml': 'ഐസ്\u200cലാൻഡിക്', 'mn': 'исланд', 'mr': 'आईसलँडिक', 'ms': 'Iceland', 'mt': 'Iżlandiż', 'my': 'အိုက်စ်လန်', 'mzn': 'ایسلندی', 'nb': 'islandsk', 'ne': 'आइसल्यान्डियाली', 'nl': 'IJslands', 'nn': 'islandsk', 'no': 'islandsk', 'nqo': 'ߌߛߑߟߊ߲ߘߌߞߊ߲', 'om': 'Ayiislandiffaa', 'or': 'ଆଇସଲାଣ୍ଡିକ୍', 'pa': 'ਆਈਸਲੈਂਡਿਕ', 'pa-Guru': 'ਆਈਸਲੈਂਡਿਕ', 'pcm': 'Aíslánd Lángwej', 'pl': 'islandzki', 'ps': 'ايسلنډي', 'pt': 'islandês', 'qu': 'Islandes Simi', 'rm': 'islandais', 'ro': 'islandeză', 'ru': 'исландский', 'rw': 'Igisilande', 'sc': 'islandesu', 'sd': 'آئيس لينڊڪ', 'sd-Arab': 'آئيس لينڊڪ', 'se': 'islánddagiella', 'si': 'අයිස්ලන්ත', 'sk': 'islandčina', 'sl': 'islandščina', 'smn': 'islandkielâ', 'so': 'Ayslandays', 'sq': 'islandisht', 'sr': 'исландски', 'sr-Cyrl': 'исландски', 'sr-Latn': 'islandski', 'sv': 'isländska', 'sw': 'Kiisilandi', 'ta': 'ஐஸ்லேண்டிக்', 'te': 'ఐస్లాండిక్', 'tg': 'исландӣ', 'th': 'ไอซ์แลนด์', 'ti': 'ኣይስላንድኛ', 'tk': 'island dili', 'to': 'lea fakaʻaisilani', 'tr': 'İzlandaca', 'tt': 'исланд', 'ug': 'ئىسلاندچە', 'uk': 'ісландська', 'ur': 'آئس لینڈک', 'uz': 'island', 'uz-Cyrl': 'исландча', 'uz-Latn': 'island', 'vec': 'izlandeze', 'vi': 'Tiếng Iceland', 'wae': 'Iisländiš', 'wo': 'Islànde', 'xnr': 'आइसलैंडिक', 'yi': 'איסלאַנדיש', 'yo': 'Èdè Icelandic', 'yrl': 'isirãdei', 'yue': '冰島文', 'yue-Hans': '冰岛文', 'yue-Hant': '冰島文', 'zh': '冰岛语', 'zh-Hans': '冰岛语', 'zh-Hant': '冰島文', 'zu': 'isi-Icelandic'},
    'isa': {'en': 'Isabi'},
    'isc': {'en': 'Isconahua'},
    'isd': {'en': 'Isnag'},
    'ise': {'en': 'Italian Sign Language'},
    'isg': {'en': 'Irish Sign Language'},
    'ish': {'en': 'Esan'},
    'isi': {'en': 'Nkem-Nkum'},
    'isk': {'en': 'Ishkashimi'},
    'ism': {'en': 'Masimasi'},
    'isn': {'en': 'Isanzu'},
    'iso': {'en': 'Isoko'},
    'isr': {'en': 'Israeli Sign Language'},
    'ist': {'en': 'Istriot'},
    'isu': {'en': 'Isu (Menchum Division)'},
    'it': {'af': 'Italiaans', 'agq': 'Ètalìa', 'ak': 'Italy kasa', 'am': 'ጣሊያንኛ', 'ar': 'الإيطالية', 'as': 'ইটালিয়ান', 'asa': 'Kiitaliaano', 'ast': 'italianu', 'az': 'italyan', 'az-Arab': 'ایتالیاجا', 'az-Cyrl': 'италјан', 'az-Latn': 'italyan', 'bal-Latn': 'Itáliái', 'bas': 'Hɔp u italìà', 'be': 'італьянская', 'be-tarask': 'італьянская', 'bem': 'Ichi Italiani', 'bez': 'Hiitaliano', 'bg': 'италиански', 'bgn': 'ایتالیایی', 'blo': 'gɩtaliija', 'bm': 'italikan', 'bn': 'ইতালিয়', 'br': 'italianeg', 'brx': 'इटालियान', 'bs': 'italijanski', 'bs-Cyrl': 'италијански', 'bs-Latn': 'italijanski', 'ca': 'italià', 'ca-ES-valencia': 'italià', 'ccp': '𑄃𑄨𑄖𑄣𑄩𑄠𑄧', 'ce': 'итальянийн', 'ceb': 'Italyano', 'cgg': 'Oruyitare', 'chr': 'ᎬᏩᎵᏲᏥᎢ', 'ckb': 'ئیتالی', 'cs': 'italština', 'cv': 'итали', 'cy': 'Eidaleg', 'da': 'italiensk', 'dav': 'Kiitaliano', 'de': 'Italienisch', 'dje': 'Itaali senni', 'doi': 'इटालियन', 'dsb': 'italšćina', 'dyo': 'italien', 'dz': 'ཨི་ཊ་ལི་ཡཱན་ཁ', 'ebu': 'Kĩitaliano', 'ee': 'Italiagbe', 'el': 'Ιταλικά', 'el-polyton': 'Ἰταλικά', 'en': 'Italian', 'en-Dsrt': '𐐆𐐻𐐰𐑊𐐷𐐲𐑌', 'eo': 'itala', 'es': 'italiano', 'et': 'itaalia', 'eu': 'italiera', 'ewo': 'Ǹkɔ́bɔ etáliɛn', 'fa': 'ایتالیایی', 'ff': 'Italiyeere', 'ff-Adlm': '𞤋𞤼𞤢𞤤𞤭𞤲𞤳𞤮𞥅𞤪𞤫', 'ff-Latn': 'Italiyeere', 'fi': 'italia', 'fil': 'Italian', 'fo': 'italskt', 'fr': 'italien', 'fur': 'talian', 'fy': 'Italiaansk', 'ga': 'Iodáilis', 'gd': 'Eadailtis', 'gl': 'italiano', 'gsw': 'Italiänisch', 'gu': 'ઇટાલિયન', 'guz': 'Kiitaliano', 'ha': 'Italiyanci', 'haw': 'ʻĪkālia', 'he': 'איטלקית', 'hi': 'इतालवी', 'hi-Latn': 'Italian', 'hr': 'talijanski', 'hsb': 'italšćina', 'ht': 'italien', 'hu': 'olasz', 'hy': 'իտալերեն', 'ia': 'italiano', 'id': 'Italia', 'ie': 'italian', 'ig': 'Italian', 'ii': 'ꑴꄊꆺꉙ', 'is': 'ítalska', 'it': 'italiano', 'ja': 'イタリア語', 'jmc': 'Kyiitaliano', 'jv': 'Italia', 'ka': 'იტალიური', 'kaa': 'итальянша', 'kaa-Cyrl': 'итальянша', 'kab': 'Taṭalyanit', 'kam': 'Kiitaliano', 'kde': 'Chiitaliano', 'kea': 'italianu', 'kgp': 'itarijỹnũ', 'khq': 'Itaali senni', 'ki': 'Kĩtaliano', 'kk': 'итальян тілі', 'kk-Cyrl': 'итальян тілі', 'kln': 'kutitab Talianek', 'km': 'អ៊ីតាលី', 'kn': 'ಇಟಾಲಿಯನ್', 'ko': '이탈리아어', 'kok': 'इटालियन', 'kok-Deva': 'इटालियन', 'ks': 'اِطالوی', 'ks-Arab': 'اِطالوی', 'ks-Deva': 'इतालवी', 'ksb': 'Kiitaliano', 'ksf': 'riitalyɛ́n', 'ksh': 'Etalljähnesch', 'ku': 'îtalî', 'kxv': 'iṭaliti', 'kxv-Deva': 'इताली ती', 'kxv-Latn': 'iṭaliti', 'kxv-Orya': 'ଇଟାଲି ତି', 'kxv-Telu': 'ఇటాలియన్', 'ky': 'италиянча', 'lag': 'Kɨtaliáano', 'lb': 'Italienesch', 'lg': 'Luyitale', 'lkt': 'Italia Iyápi', 'lld': 'talian', 'ln': 'litaliano', 'lo': 'ອິຕາລຽນ', 'lrc': 'ئیتالیایی', 'lt': 'italų', 'lu': 'Litali', 'luo': 'Kiitaliano', 'luy': 'Kiitaliano', 'lv': 'itāļu', 'mai': 'इतालवी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ ltalian', 'mer': 'Kĩitalĩ', 'mfe': 'italien', 'mg': 'Italianina', 'mgh': 'Italiano', 'mi': 'Itāriana', 'mk': 'италијански', 'ml': 'ഇറ്റാലിയൻ', 'mn': 'итали', 'mni': 'ইটালিয়ন', 'mni-Beng': 'ইটালিয়ন', 'mr': 'इटालियन', 'ms': 'Itali', 'mt': 'Taljan', 'mua': 'italiya', 'my': 'အီတလီ', 'mzn': 'ایتالیایی', 'naq': 'Italians', 'nb': 'italiensk', 'nd': 'isi-Italiano', 'ne': 'इटालेली', 'nl': 'Italiaans', 'nmg': 'Kiɛl italia', 'nn': 'italiensk', 'no': 'italiensk', 'nqo': 'ߌߕߊߟߌߞߊ߲', 'nus': 'Thok i̱taliani', 'nyn': 'Oruyitare', 'om': 'Afaan Xaaliyaani', 'or': 'ଇଟାଲୀୟ', 'os': 'италиаг', 'pa': 'ਇਤਾਲਵੀ', 'pa-Guru': 'ਇਤਾਲਵੀ', 'pcm': 'Ítáli Lángwej', 'pl': 'włoski', 'ps': 'ایټالوي', 'pt': 'italiano', 'qu': 'Italiano Simi', 'rm': 'talian', 'rn': 'Igitaliyani', 'ro': 'italiană', 'rof': 'Kiitaliano', 'ru': 'итальянский', 'rw': 'Igitaliyani', 'rwk': 'Kyiitaliano', 'sa': 'इटलीदेशीय भाषा:', 'sah': 'Ытаалыйалыы', 'saq': 'Kiitaliano', 'sat': 'ᱤᱴᱟᱞᱤᱟᱱ', 'sat-Olck': 'ᱤᱴᱟᱞᱤᱟᱱ', 'sbp': 'Ishihitaliyano', 'sc': 'italianu', 'sd': 'اطالوي', 'sd-Arab': 'اطالوي', 'sd-Deva': 'इटालियनु', 'se': 'itáliagiella', 'seh': 'italiano', 'ses': 'Itaali senni', 'sg': 'Ênnde', 'shi': 'ⵜⴰⵟⴰⵍⵢⴰⵏⵜ', 'shi-Latn': 'Taṭalyant', 'shi-Tfng': 'ⵜⴰⵟⴰⵍⵢⴰⵏⵜ', 'si': 'ඉතාලි', 'sk': 'taliančina', 'sl': 'italijanščina', 'smn': 'italiakielâ', 'sn': 'chiTariana', 'so': 'Talyaani', 'sq': 'italisht', 'sr': 'италијански', 'sr-Cyrl': 'италијански', 'sr-Latn': 'italijanski', 'su': 'Italia', 'su-Latn': 'Italia', 'sv': 'italienska', 'sw': 'Kiitaliano', 'syr': 'ܐܝܛܠܝܬ', 'ta': 'இத்தாலியன்', 'te': 'ఇటాలియన్', 'teo': 'Kiitaliano', 'tg': 'италиявӣ', 'th': 'อิตาลี', 'ti': 'ጥልያን', 'tk': 'italýan dili', 'to': 'lea fakaʻītali', 'tpi': 'Italien', 'tr': 'İtalyanca', 'tt': 'итальян', 'twq': 'Itaali senni', 'tzm': 'Taṭalyant', 'ug': 'ئىتالىيانچە', 'uk': 'італійська', 'ur': 'اطالوی', 'uz': 'italyan', 'uz-Cyrl': 'италянча', 'uz-Latn': 'italyan', 'vai': 'ꔤꕚꔷꘂꘋ', 'vai-Latn': 'Itáliyɛŋ', 'vai-Vaii': 'ꔤꕚꔷꘂꘋ', 'vec': 'italian', 'vi': 'Tiếng Italy', 'vun': 'Kyiitaliano', 'wae': 'Italieniš', 'wo': 'Italiye', 'xh': 'IsiTaliyane', 'xnr': 'इतालवी', 'xog': 'Oluyitale', 'yav': 'itáliɛ', 'yi': 'איטאַליעניש', 'yo': 'Èdè Ítálì', 'yrl': 'itarianu', 'yue': '義大利文', 'yue-Hans': '意大利文', 'yue-Hant': '義大利文', 'zgh': 'ⵜⴰⵟⴰⵍⵢⴰⵏⵜ', 'zh': '意大利语', 'zh-Hans': '意大利语', 'zh-Hant': '義大利文', 'zu': 'isi-Italian'},
    'itb': {'en': 'Binongan Itneg'},
    'itc': {'en': 'Italic languages'},
    'itc-ola': {'en': 'Old Latin'},
    'itc-pro': {'en': 'Proto-Italic'},
    'itd': {'en': 'Southern Tidung'},
    'ite': {'en': 'Itene'},
    'iti': {'en': 'Inlaod Itneg'},
    'itk': {'en': 'Judeo-Italian'},
    'itl': {'en': 'Itelmen'},
    'itm': {'en': 'Itu Mbon Uzo'},
    'ito': {'en': 'Itonama'},
    'itr': {'en': 'Iteri'},
    'its': {'en': 'Isekiri'},
    'itt': {'en': 'Maeng Itneg'},
    'itv': {'en': 'Itawit'},
    'itw': {'en': 'Ito'},
    'itx': {'en': 'Itik'},
    'ity': {'en': 'Moyadan Itneg'},
    'itz': {'en': 'Itzá'},
    'iu': {'af': 'Inoektitoet', 'am': 'እኑክቲቱትኛ', 'ar': 'الإينكتيتت', 'as': 'ইনুক্টিটুট', 'ast': 'inuktitut', 'az': 'inuktitut', 'az-Cyrl': 'инуктитут', 'az-Latn': 'inuktitut', 'bal-Latn': 'Inuktitut', 'be': 'інуктытут', 'be-tarask': 'інуктытут', 'bg': 'инуктитут', 'bgn': 'اینوکتیتوتی', 'bn': 'ইনুক্টিটুট', 'br': 'inuktitut', 'brx': 'इनुक्टीटुत', 'bs': 'inuktitut', 'bs-Cyrl': 'инуктитут', 'bs-Latn': 'inuktitut', 'ca': 'inuktitut', 'ca-ES-valencia': 'inuktitut', 'ccp': '𑄃𑄨𑄚𑄪𑄇𑄴𑄑𑄨𑄑𑄪𑄖𑄴', 'ce': 'инуктитут', 'chr': 'ᎢᏄᎦᏘᏚ', 'ckb': 'ئینوکتیتوت', 'cs': 'inuktitutština', 'cy': 'Inwctitwt', 'da': 'inuktitut', 'de': 'Inuktitut', 'dsb': 'inuitšćina', 'el': 'Ινούκτιτουτ', 'el-polyton': 'Ινούκτιτουτ', 'en': 'Inuktitut', 'eo': 'inuita', 'es': 'inuktitut', 'et': 'inuktituti', 'eu': 'inuitera', 'fa': 'اینوکتیتوت', 'ff-Adlm': '𞤋𞤲𞤵𞤳𞤼𞤫𞥅𞤪𞤫', 'fi': 'inuktitut', 'fil': 'Inuktitut', 'fo': 'inuktitut', 'fr': 'inuktitut', 'fur': 'inuktitut', 'fy': 'Inuktitut', 'ga': 'Ionúitis', 'gd': 'Inuktitut', 'gl': 'inuktitut', 'gsw': 'Inukitut', 'gu': 'ઇનુકિટૂટ', 'ha': 'Inuktitut', 'he': 'אינוקטיטוט', 'hi': 'इनुक्टिटुट', 'hi-Latn': 'Inuktitut', 'hr': 'inuktitut', 'hsb': 'inuitšćina', 'ht': 'inuktitut', 'hu': 'inuktitut', 'hy': 'ինուկտիտուտ', 'ia': 'inuktitut', 'id': 'Inuktitut', 'ig': 'Inuktitut', 'is': 'inúktitút', 'it': 'inuktitut', 'ja': 'イヌクティトット語', 'jv': 'Inuktitut', 'ka': 'ინუკტიტუტი', 'kea': 'inuktitut', 'kgp': 'inugtituti', 'kk': 'инуктитут тілі', 'kk-Cyrl': 'инуктитут тілі', 'km': 'អ៊ីនុកទីទុត', 'kn': 'ಇನುಕ್ಟಿಟುಟ್', 'ko': '이눅티투트어', 'kok': 'इन्युक्तिटुट', 'kok-Deva': 'इन्युक्तिटुट', 'ks': 'اِنُکتِتوٗ', 'ks-Arab': 'اِنُکتِتوٗ', 'ksh': 'Inuktitut', 'ku': 'înuîtî', 'ky': 'инуктитутча', 'lb': 'Inukitut', 'lld': 'inuktitut', 'lo': 'ອິນນຸກຕິຕັດ', 'lrc': 'ئینوکتیتوٙت', 'lt': 'inukitut', 'lv': 'inuītu', 'mai': 'इनुकतितुत', 'mi': 'Inukitetūta', 'mk': 'инуктитут', 'ml': 'ഇനുക്റ്റിറ്റുട്ട്', 'mn': 'инуктитут', 'mr': 'इनुक्तीटुट', 'ms': 'Inuktitut', 'mt': 'Inuktitut', 'my': 'အီနုခ်တီတု', 'mzn': 'انوکتیتوت', 'nb': 'inuktitut', 'ne': 'इनुक्टिटुट', 'nl': 'Inuktitut', 'nn': 'inuktitut', 'no': 'inuktitut', 'or': 'ଇନୁକଟିଟୁତ୍\u200c', 'pa': 'ਇੰਕਟੀਟੂਤ', 'pa-Guru': 'ਇੰਕਟੀਟੂਤ', 'pcm': 'Inuktítut', 'pl': 'inuktitut', 'ps': 'انوکتیتوت', 'pt': 'inuktitut', 'qu': 'Inuktitut Simi', 'rm': 'inuktitut', 'ro': 'inuktitut', 'ru': 'инуктитут', 'sc': 'inuktitut', 'sd': 'انو ڪتوت', 'sd-Arab': 'انو ڪتوت', 'si': 'ඉනුක්ටිටුට්', 'sk': 'inuktitut', 'sl': 'inuktitutščina', 'smn': 'inuktitut', 'so': 'Inuktitut', 'sq': 'inuktitutisht', 'sr': 'инуктитутски', 'sr-Cyrl': 'инуктитутски', 'sr-Latn': 'inuktitutski', 'sv': 'inuktitut', 'sw': 'Kiinuktituti', 'ta': 'இனுகிடூட்', 'te': 'ఇనుక్టిటుట్', 'tg': 'инуктитутӣ', 'th': 'อินุกติตุต', 'ti': 'ኢናክቲቱት', 'tk': 'inuktitut dili', 'to': 'lea fakaʻinuketītuti', 'tr': 'İnuktitut dili', 'tt': 'инуктикут', 'ug': 'ئىنۇكتىتۇتچە', 'uk': 'інуктитут', 'ur': 'اینُکٹیٹٹ', 'uz': 'inuktitut', 'uz-Cyrl': 'инуктитут', 'uz-Latn': 'inuktitut', 'vec': 'inuktitut', 'vi': 'Tiếng Inuktitut', 'wo': 'Inuktitit', 'xnr': 'इनूकीटूत्', 'yo': 'Èdè Inukitu', 'yrl': 'inukitituti', 'yue': '因紐特文', 'yue-Hans': '因纽特文', 'yue-Hant': '因紐特文', 'zh': '因纽特语', 'zh-Hans': '因纽特语', 'zh-Hant': '因紐特文', 'zu': 'isi-Inuktitut'},
    'ium': {'en': 'Iu Mien'},
    'ivb': {'en': 'Ibatan'},
    'ivv': {'en': 'Ivatan'},
    'iw': {'en': 'Hebrew'},
    'iwk': {'en': 'I-Wak'},
    'iwm': {'en': 'Iwam'},
    'iwo': {'en': 'Iwur'},
    'iws': {'en': 'Sepik Iwam'},
    'ixc': {'en': 'Ixcatec'},
    'ixl': {'en': 'Ixil'},
    'iya': {'en': 'Iyayu'},
    'iyo': {'en': 'Mesaka'},
    'iyx': {'en': 'Yaka (Congo)'},
    'izh': {'ast': 'ingrianu', 'cs': 'ingrijština', 'de': 'Ischorisch', 'en': 'Ingrian', 'et': 'isuri', 'fi': 'inkeroinen', 'fr': 'ingrien', 'hi-Latn': 'Ingrian', 'ht': 'ingrien', 'it': 'ingrico', 'ja': 'イングリア語', 'ksh': 'Ingjrijahnesch', 'lb': 'Ischoresch', 'lt': 'ingrų', 'mk': 'ижорски', 'nb': 'ingrisk', 'ne': 'इन्ग्रियाली', 'nl': 'Ingrisch', 'nn': 'ingrisk', 'no': 'ingrisk', 'pl': 'ingryjski', 'sv': 'ingriska', 'th': 'อินเกรียน', 'to': 'lea fakaʻingiliani', 'tr': 'İngriya Dili', 'vi': 'Tiếng Ingria', 'yue': '英格裏亞文', 'yue-Hans': '英格里亚文', 'yue-Hant': '英格裏亞文', 'zh-Hant': '英格里亞文'},
    'izi': {'en': 'Izi-Ezaa-Ikwo-Mgbo'},
    'izr': {'en': 'Izere'},
    'izz': {'en': 'Izii'},
    'ja': {'af': 'Japannees', 'agq': 'Dzàkpànê', 'ak': 'Gyapan kasa', 'am': 'ጃፓንኛ', 'ar': 'اليابانية', 'as': 'জাপানী', 'asa': 'Kijapani', 'ast': 'xaponés', 'az': 'yapon', 'az-Arab': 'ژاپونجا', 'az-Cyrl': 'јапон', 'az-Latn': 'yapon', 'bal-Latn': 'Jápáni', 'bas': 'Hɔp u yapàn', 'be': 'японская', 'be-tarask': 'японская', 'bem': 'Ichi Japanisi', 'bez': 'Hijapani', 'bg': 'японски', 'bgn': 'جاپانی', 'blo': 'gɩjapaŋja', 'bm': 'zapɔnekan', 'bn': 'জাপানি', 'bo': 'ཉི་ཧོང་སྐད་', 'br': 'japaneg', 'brx': 'जापानीज', 'bs': 'japanski', 'bs-Cyrl': 'јапански', 'bs-Latn': 'japanski', 'ca': 'japonès', 'ca-ES-valencia': 'japonès', 'ccp': '𑄎𑄛𑄚𑄨', 'ce': 'японийн', 'ceb': 'Hinapon', 'cgg': 'Orujapaani', 'chr': 'ᏣᏩᏂᏏ', 'ckb': 'ژاپۆنی', 'cs': 'japonština', 'cv': 'япони', 'cy': 'Japaneeg', 'da': 'japansk', 'dav': 'Kijapani', 'de': 'Japanisch', 'dje': 'Japonee senni', 'doi': 'जापानी', 'dsb': 'japańšćina', 'dyo': 'saponee', 'dz': 'ཇཱ་པཱ་ནིས་ཁ', 'ebu': 'Kĩnjapani', 'ee': 'Japangbe', 'el': 'Ιαπωνικά', 'el-polyton': 'Ἰαπωνικά', 'en': 'Japanese', 'en-Dsrt': '𐐖𐐰𐐹𐐲𐑌𐐨𐑆', 'eo': 'japana', 'es': 'japonés', 'et': 'jaapani', 'eu': 'japoniera', 'ewo': 'Ǹkɔ́bɔ hapɔ́n', 'fa': 'ژاپنی', 'ff': 'Saponeere', 'ff-Adlm': '𞤐𞤭𞤨𞤮𞤲𞤪𞤫', 'ff-Latn': 'Saponeere', 'fi': 'japani', 'fil': 'Japanese', 'fo': 'japanskt', 'fr': 'japonais', 'fur': 'gjaponês', 'fy': 'Japans', 'ga': 'Seapáinis', 'gd': 'Seapanais', 'gl': 'xaponés', 'gsw': 'Japanisch', 'gu': 'જાપાનીઝ', 'guz': 'Kijapani', 'ha': 'Japananci', 'haw': 'Kepanī', 'he': 'יפנית', 'hi': 'जापानी', 'hi-Latn': 'Japanese', 'hr': 'japanski', 'hsb': 'japanšćina', 'ht': 'japonais', 'hu': 'japán', 'hy': 'ճապոներեն', 'ia': 'japonese', 'id': 'Jepang', 'ie': 'japanesi', 'ig': 'Japanese', 'ii': 'ꏝꀪꉙ', 'is': 'japanska', 'it': 'giapponese', 'ja': '日本語', 'jmc': 'Kyijapani', 'jv': 'Jepang', 'ka': 'იაპონური', 'kaa': 'японша', 'kaa-Cyrl': 'японша', 'kab': 'Tajapunit', 'kam': 'Kijapani', 'kde': 'Chidyapani', 'kea': 'japones', 'kgp': 'japonẽj', 'khq': 'Japonee senni', 'ki': 'Kĩnjabani', 'kk': 'жапон тілі', 'kk-Cyrl': 'жапон тілі', 'kln': 'kutitap Japan', 'km': 'ជប៉ុន', 'kn': 'ಜಾಪನೀಸ್', 'ko': '일본어', 'kok': 'जपानी', 'kok-Deva': 'जपानी', 'ks': 'جاپٲنؠ', 'ks-Arab': 'جاپٲنؠ', 'ks-Deva': 'जापानी', 'ksb': 'Kijapani', 'ksf': 'rijapɔ́ŋ', 'ksh': 'Japahnesch', 'ku': 'japonî', 'kxv': 'japanij', 'kxv-Deva': 'जापानीज', 'kxv-Latn': 'japanij', 'kxv-Orya': 'ଜାପାନିଜ୍', 'kxv-Telu': 'జపనిస్', 'ky': 'жапончо', 'lag': 'Kɨjapáani', 'lb': 'Japanesch', 'lg': 'Lujapani', 'lkt': 'Kisúŋla Iyápi', 'lld': 'iapanesc', 'ln': 'lizapɔ', 'lo': 'ຍີ່ປຸ່ນ', 'lrc': 'جاپوٙنی', 'lt': 'japonų', 'lu': 'Liyapani', 'luo': 'Kijapani', 'luy': 'Kijapani', 'lv': 'japāņu', 'mai': 'जापानी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ japani', 'mer': 'Kĩjapani', 'mfe': 'zapone', 'mg': 'Japoney', 'mgh': 'Ijapani', 'mi': 'Hapanihi', 'mk': 'јапонски', 'ml': 'ജാപ്പനീസ്', 'mn': 'япон', 'mni': 'জাপানিজ', 'mni-Beng': 'জাপানিজ', 'mr': 'जपानी', 'ms': 'Jepun', 'mt': 'Ġappuniż', 'mua': 'zah sǝr Japoŋ', 'my': 'ဂျပန်', 'mzn': 'جاپونی', 'naq': 'Japanees', 'nb': 'japansk', 'nd': 'isi-Japhani', 'ne': 'जापानी', 'nl': 'Japans', 'nmg': 'Kiɛl bó japonɛ̌', 'nn': 'japansk', 'no': 'japansk', 'nqo': 'ߖ߭ߊߔߐ߲ߞߊ߲', 'nus': 'Thok japanni', 'nyn': 'Orujapaani', 'om': 'Afaan Japanii', 'or': 'ଜାପାନୀ', 'os': 'япойнаг', 'pa': 'ਜਪਾਨੀ', 'pa-Guru': 'ਜਪਾਨੀ', 'pcm': 'Japan Lángwej', 'pl': 'japoński', 'ps': 'جاپاني', 'pt': 'japonês', 'qu': 'Japones Simi', 'rm': 'giapunais', 'rn': 'Ikiyapani', 'ro': 'japoneză', 'rof': 'Kijapani', 'ru': 'японский', 'rw': 'Ikiyapani', 'rwk': 'Kyijapani', 'sa': 'सूर्यमूलीय भाषा:', 'sah': 'Дьоппуоннуу', 'saq': 'Kijapani', 'sat': 'ᱡᱟᱯᱟᱱᱤ', 'sat-Olck': 'ᱡᱟᱯᱟᱱᱤ', 'sbp': 'Ishijapani', 'sc': 'giaponesu', 'sd': 'جاپاني', 'sd-Arab': 'جاپاني', 'sd-Deva': 'जापानी', 'se': 'japánagiella', 'seh': 'japonês', 'ses': 'Japonee senni', 'sg': 'Zaponëe', 'shi': 'ⵜⴰⵊⴰⴱⴱⵓⵏⵉⵜ', 'shi-Latn': 'Tajabbunit', 'shi-Tfng': 'ⵜⴰⵊⴰⴱⴱⵓⵏⵉⵜ', 'si': 'ජපන්', 'sk': 'japončina', 'sl': 'japonščina', 'smn': 'jaapaankielâ', 'sn': 'chiJapani', 'so': 'Jabaaniis', 'sq': 'japonisht', 'sr': 'јапански', 'sr-Cyrl': 'јапански', 'sr-Latn': 'japanski', 'su': 'Jepang', 'su-Latn': 'Jepang', 'sv': 'japanska', 'sw': 'Kijapani', 'ta': 'ஜப்பானியம்', 'te': 'జపనీస్', 'teo': 'Kijapani', 'tg': 'японӣ', 'th': 'ญี่ปุ่น', 'ti': 'ጃፓንኛ', 'tk': 'ýapon dili', 'to': 'lea fakasiapani', 'tpi': 'Japanis', 'tr': 'Japonca', 'tt': 'япон', 'twq': 'Japonee senni', 'tzm': 'Tajappunit', 'ug': 'ياپونچە', 'uk': 'японська', 'ur': 'جاپانی', 'uz': 'yapon', 'uz-Cyrl': 'японча', 'uz-Latn': 'yapon', 'vai': 'ꕧꕐꕇꔧ', 'vai-Latn': 'Japaníĩ', 'vai-Vaii': 'ꕧꕐꕇꔧ', 'vec': 'japoneze', 'vi': 'Tiếng Nhật', 'vun': 'Kyijapani', 'wae': 'Japaniš', 'wo': 'Sapone', 'xh': 'IsiJapan', 'xnr': 'जापानी', 'xog': 'Olujapani', 'yav': 'ndiáman', 'yi': 'יאַפּאַניש', 'yo': 'Èdè Jàpáànù', 'yrl': 'yapunei', 'yue': '日文', 'yue-Hans': '日文', 'yue-Hant': '日文', 'zgh': 'ⵜⴰⵊⴰⴱⴱⵓⵏⵉⵜ', 'zh': '日语', 'zh-Hans': '日语', 'zh-Hant': '日文', 'zu': 'isi-Japanese'},
    'jaa': {'en': 'Jamamadí'},
    'jab': {'en': 'Hyam'},
    'jac': {'en': "Popti'"},
    'jad': {'en': 'Jahanka'},
    'jae': {'en': 'Yabem'},
    'jaf': {'en': 'Jara'},
    'jah': {'en': 'Jah Hut'},
    'jaj': {'en': 'Zazao'},
    'jak': {'en': 'Jakun'},
    'jal': {'en': 'Yalahatan'},
    'jam': {'ast': 'inglés criollu xamaicanu', 'br': 'kreoleg Jamaika', 'ca': 'crioll anglès de Jamaica', 'ca-ES-valencia': 'crioll anglès de Jamaica', 'cs': 'jamajská kreolština', 'de': 'Jamaikanisch-Kreolisch', 'en': 'Jamaican Creole English', 'et': 'Jamaica kreoolkeel', 'fi': 'jamaikankreolienglanti', 'fr': 'créole jamaïcain', 'gd': 'Beurla Crìtheolach Diameuga', 'hi-Latn': 'Jamaican Creole English', 'ht': 'créole jamaïcain', 'it': 'creolo giamaicano', 'ja': 'ジャマイカ・クレオール語', 'ksh': 'Jamaikahnesch-Änglesch', 'lb': 'Jamaikanesch-Kreolesch', 'lt': 'Jamaikos kreolų anglų', 'mk': 'јамајски креолски', 'nb': 'jamaicansk kreolengelsk', 'ne': 'जमैकाली क्रेओले अङ्ग्रेजी', 'nl': 'Jamaicaans Creools', 'nn': 'jamaicansk kreolengelsk', 'no': 'jamaicansk kreolengelsk', 'pl': 'jamajski', 'sv': 'jamaikansk engelsk kreol', 'th': 'อังกฤษคลีโอลจาเมกา', 'to': 'lea fakapālangi-samaika', 'tr': 'Jamaika Patois Dili', 'vi': 'Tiếng Anh Jamaica Creole', 'yue': '牙買加克裏奧爾英文', 'yue-Hans': '牙买加克里奥尔英文', 'yue-Hant': '牙買加克裏奧爾英文', 'zh-Hant': '牙買加克里奧爾英文'},
    'jan': {'en': 'Jandai'},
    'jao': {'en': 'Yanyuwa'},
    'jaq': {'en': 'Yaqay'},
    'jar': {'en': 'Jarawa (Nigeria)'},
    'jas': {'en': 'New Caledonian Javanese'},
    'jat': {'en': 'Jakati'},
    'jau': {'en': 'Yaur'},
    'jax': {'en': 'Jambi Malay'},
    'jay': {'en': 'Yan-nhangu'},
    'jaz': {'en': 'Jawe'},
    'jbe': {'en': 'Judeo-Berber'},
    'jbi': {'en': 'Badjiri'},
    'jbj': {'en': 'Arandai'},
    'jbk': {'en': 'Barikewa'},
    'jbm': {'en': 'Bijim'},
    'jbn': {'en': 'Nafusi'},
    'jbo': {'af': 'Lojban', 'am': 'ሎጅባን', 'ar': 'اللوجبان', 'as': 'লোজ্বান', 'ast': 'lojban', 'az': 'loğban', 'az-Cyrl': 'лоғбан', 'az-Latn': 'loğban', 'bal-Latn': 'Lojban', 'be': 'ложбан', 'be-tarask': 'ложбан', 'bg': 'ложбан', 'bgn': 'لوجبانی', 'bn': 'লোজবান', 'br': 'lojban', 'brx': 'ल’जबान', 'bs': 'lojban', 'bs-Cyrl': 'лојбан', 'bs-Latn': 'lojban', 'ca': 'lojban', 'ca-ES-valencia': 'lojban', 'ccp': '𑄣𑄮𑄌𑄴𑄝𑄚𑄴', 'ce': 'ложбан', 'chr': 'ᎶᏣᏆᏂ', 'ckb': 'لۆژبان', 'cs': 'lojban', 'cy': 'Lojban', 'da': 'lojban', 'de': 'Lojban', 'dsb': 'lojbanšćina', 'el': 'Λόζμπαν', 'el-polyton': 'Λόζμπαν', 'en': 'Lojban', 'es': 'lojban', 'et': 'ložban', 'eu': 'lojbana', 'fa': 'لوجبان', 'ff-Adlm': '𞤂𞤮𞤶𞤦𞤢𞤲𞤪𞤫', 'fi': 'lojban', 'fil': 'Lojban', 'fo': 'lojban', 'fr': 'lojban', 'fy': 'Lojban', 'ga': 'Lojban', 'gd': 'Lojban', 'gl': 'lojban', 'gsw': 'Lojbanisch', 'gu': 'લોજ્બાન', 'ha': 'Lojban', 'he': 'לוז׳באן', 'hi': 'लोज्बान', 'hi-Latn': 'Lojban', 'hr': 'lojban', 'hsb': 'lojbanšćina', 'ht': 'lojban', 'hu': 'lojban', 'hy': 'լոժբան', 'ia': 'lojban', 'id': 'Lojban', 'ig': 'Lojban', 'is': 'lojban', 'it': 'lojban', 'ja': 'ロジバン語', 'jbo': 'la .lojban.', 'jv': 'Lojban', 'ka': 'ლოჟბანი', 'kgp': 'rojmán', 'kk': 'ложбан тілі', 'kk-Cyrl': 'ложбан тілі', 'km': 'លុចបាន', 'kn': 'ಲೊಜ್ಬಾನ್', 'ko': '로반어', 'kok': 'लोजबान', 'kok-Deva': 'लोजबान', 'ks': 'لوجبان', 'ks-Arab': 'لوجبان', 'ksh': 'Lodschbahn', 'ku': 'lojbanî', 'ky': 'ложбанча', 'lb': 'Lojban', 'lo': 'ໂລບບັນ', 'lt': 'loiban', 'lv': 'ložbans', 'mai': 'लोजबन', 'mi': 'Rōpāna', 'mk': 'ложбан', 'ml': 'ലോജ്ബാൻ', 'mn': 'ложбан', 'mr': 'लोज्बान', 'ms': 'Lojban', 'mt': 'Lojban', 'my': 'လိုဂျ်ဘန်', 'nb': 'lojban', 'ne': 'लोज्बान', 'nl': 'Lojban', 'nn': 'lojban', 'no': 'lojban', 'or': 'ଲୋଜବାନ୍', 'pa': 'ਲੋਜਬਾਨ', 'pa-Guru': 'ਲੋਜਬਾਨ', 'pcm': 'Lojban Lángwej', 'pl': 'lojban', 'ps': 'لوجبان', 'pt': 'lojban', 'qu': 'Lojban Simi', 'rm': 'lojban', 'ro': 'lojban', 'ru': 'ложбан', 'sc': 'lojban', 'sd': 'لوجبين', 'sd-Arab': 'لوجبين', 'si': 'ලොජ්බන්', 'sk': 'lojban', 'sl': 'lojban', 'smn': 'lojban', 'so': 'Lojban', 'sq': 'lojbanisht', 'sr': 'ложбан', 'sr-Cyrl': 'ложбан', 'sr-Latn': 'ložban', 'sv': 'lojban', 'sw': 'Lojban', 'ta': 'லோஜ்பன்', 'te': 'లోజ్బాన్', 'th': 'โลชบัน', 'ti': 'ሎጅባን', 'tk': 'lojban dili', 'to': 'lea fakalosipani', 'tr': 'Lojban', 'ug': 'لوجبانچە', 'uk': 'ложбан', 'ur': 'لوجبان', 'uz': 'lojban', 'uz-Latn': 'lojban', 'vec': 'lojban', 'vi': 'Tiếng Lojban', 'xnr': 'लोज्बान', 'yi': 'לאזשבאָן', 'yo': 'Èdè Lobani', 'yrl': 'ruyibã', 'yue': '邏輯文', 'yue-Hans': '逻辑文', 'yue-Hant': '邏輯文', 'zh': '逻辑语', 'zh-Hans': '逻辑语', 'zh-Hant': '邏輯文', 'zu': 'isi-Lojban'},
    'jbr': {'en': 'Jofotek-Bromnya'},
    'jbt': {'en': 'Jabutí'},
    'jbu': {'en': 'Jukun Takum'},
    'jbw': {'en': 'Yawijibaya'},
    'jcs': {'en': 'Jamaican Country Sign Language'},
    'jct': {'en': 'Krymchak'},
    'jda': {'en': 'Jad'},
    'jdg': {'en': 'Jadgali'},
    'jdt': {'en': 'Judeo-Tat'},
    'jeb': {'en': 'Jebero'},
    'jee': {'en': 'Jerung'},
    'jeg': {'en': 'Jeng'},
    'jeh': {'en': 'Jeh'},
    'jei': {'en': 'Yei'},
    'jek': {'en': 'Jeri Kuo'},
    'jel': {'en': 'Yelmek'},
    'jen': {'en': 'Dza'},
    'jer': {'en': 'Jere'},
    'jet': {'en': 'Manem'},
    'jeu': {'en': 'Jonkor Bourmataguil'},
    'jgb': {'en': 'Ngbee'},
    'jge': {'en': 'Judeo-Georgian'},
    'jgk': {'en': 'Gwak'},
    'jgo': {'af': 'Ngomba', 'am': 'ንጎምባ', 'ar': 'نغومبا', 'as': 'নগোম্বা', 'ast': 'ngomba', 'az': 'nqomba', 'az-Cyrl': 'нгомба', 'az-Latn': 'nqomba', 'bal-Latn': 'Ngombá', 'be': 'нгомба', 'be-tarask': 'нгомба', 'bg': 'нгомба', 'bgn': 'نگومبی', 'bn': 'গোম্বা', 'br': 'ngomba', 'brx': 'न्गम्बा', 'bs': 'ngomba', 'bs-Cyrl': 'нгомба', 'bs-Latn': 'ngomba', 'ca': 'ngomba', 'ca-ES-valencia': 'ngomba', 'ccp': '𑄉𑄮𑄟𑄴𑄝', 'ce': 'нгомба', 'chr': 'ᎾᎪᏆ', 'ckb': 'نگۆمبا', 'cs': 'ngomba', 'cy': 'Ngomba', 'da': 'ngomba', 'de': 'Ngomba', 'dsb': 'ngomba', 'el': 'Νγκόμπα', 'el-polyton': 'Νγκόμπα', 'en': 'Ngomba', 'es': 'ngomba', 'et': 'ngomba', 'eu': 'ngomba', 'fa': 'نگومبا', 'ff-Adlm': '𞤐𞤺𞤮𞤥𞤦𞤢𞥄𞤪𞤫', 'fi': 'ngomba', 'fil': 'Ngomba', 'fo': 'ngomba', 'fr': 'ngomba', 'fy': 'Ngomba', 'ga': 'Ngomba', 'gd': 'Ngomba', 'gl': 'ngomba', 'gu': 'નગોમ્બા', 'ha': 'Ngomba', 'he': 'נגומבה', 'hi': 'नगोंबा', 'hi-Latn': 'Ngomba', 'hr': 'ngomba', 'hsb': 'ngomba', 'ht': 'ngomba', 'hu': 'ngomba', 'hy': 'նգոմբա', 'ia': 'ngomba', 'id': 'Ngomba', 'ig': 'Ngomba', 'is': 'ngomba', 'it': 'ngamambo', 'ja': 'ンゴンバ語', 'jgo': 'Ndaꞌa', 'jv': 'Ngomba', 'ka': 'ნგომბა', 'kea': 'ñomba', 'kgp': 'gẽma', 'kk': 'нгомба тілі', 'kk-Cyrl': 'нгомба тілі', 'km': 'ងុំបា', 'kn': 'ನೊಂಬಾ', 'ko': '응곰바어', 'kok': 'गोंबा', 'kok-Deva': 'गोंबा', 'ksh': 'Njomba', 'ku': 'ngomba', 'ky': 'нгомбача', 'lb': 'Ngomba', 'lo': 'ງອມບາ', 'lrc': 'نئگوٙمبا', 'lt': 'ngombų', 'lv': 'ngomba', 'mai': 'नेगोमबा', 'mi': 'Nakōma', 'mk': 'нгомба', 'ml': 'ഗോമ്പ', 'mn': 'нгомба', 'mr': 'गोम्बा', 'ms': 'Ngomba', 'mt': 'Ngomba', 'my': 'ဂွမ်ဘာ', 'mzn': 'نگومبا', 'nb': 'ngomba', 'ne': 'न्गोम्बा', 'nl': 'Ngomba', 'nn': 'ngomba', 'no': 'ngomba', 'nqo': 'ߒߜ߭ߏ߲ߓߊߞߊ߲', 'or': 'ନାଗୋମ୍ଵା', 'pa': 'ਨਗੋਂਬਾ', 'pa-Guru': 'ਨਗੋਂਬਾ', 'pcm': 'Ngómbá Lángwej', 'pl': 'ngombe', 'ps': 'نګومبا', 'pt': 'nguemba', 'qu': 'Ngomba Simi', 'rm': 'ngomba', 'ro': 'ngomba', 'ru': 'нгомба', 'sc': 'ngomba', 'sd': 'نغومبا', 'sd-Arab': 'نغومبا', 'si': 'නොම්බා', 'sk': 'ngomba', 'sl': 'ngomba', 'smn': 'ngomba', 'so': 'Ingoomba', 'sq': 'ngombisht', 'sr': 'нгомба', 'sr-Cyrl': 'нгомба', 'sr-Latn': 'ngomba', 'sv': 'ngomba', 'sw': 'Kingomba', 'ta': 'நகொம்பா', 'te': 'గోంబా', 'th': 'อึนกอมบา', 'ti': 'ኤንጎምባ', 'tk': 'ngomba dili', 'to': 'lea fakanikōmipa', 'tr': 'Ngomba', 'ug': 'نگومباچە', 'uk': 'нгомба', 'ur': 'نگومبا', 'uz': 'ngomba', 'uz-Cyrl': 'нгомба', 'uz-Latn': 'ngomba', 'vec': 'ngonba', 'vi': 'Tiếng Ngomba', 'xnr': 'नगोंबा', 'yo': 'Ńgòmbà', 'yrl': 'ĩgẽba', 'yue': '恩格姆巴文', 'yue-Hans': '恩格姆巴文', 'yue-Hant': '恩格姆巴文', 'zh': '恩艮巴语', 'zh-Hans': '恩艮巴语', 'zh-Hant': '恩格姆巴文', 'zu': 'isi-Ngomba'},
    'jhi': {'en': 'Jehai'},
    'jhs': {'en': 'Jhankot Sign Language'},
    'ji': {'en': 'Yiddish'},
    'jia': {'en': 'Jina'},
    'jib': {'en': 'Jibu'},
    'jic': {'en': 'Tol'},
    'jid': {'en': 'Bu (Kaduna State)'},
    'jie': {'en': 'Jilbe'},
    'jig': {'en': 'Jingulu'},
    'jih': {'en': 'sTodsde'},
    'jii': {'en': 'Jiiddu'},
    'jil': {'en': 'Jilim'},
    'jim': {'en': 'Jimi (Cameroon)'},
    'jio': {'en': 'Jiamao'},
    'jiq': {'en': 'Guanyinqiao'},
    'jit': {'en': 'Jita'},
    'jiu': {'en': 'Youle Jinuo'},
    'jiv': {'en': 'Shuar'},
    'jiy': {'en': 'Buyuan Jinuo'},
    'jje': {'en': 'Jejueo'},
    'jjr': {'en': 'Bankal'},
    'jka': {'en': 'Kaera'},
    'jkm': {'en': 'Mobwa Karen'},
    'jko': {'en': 'Kubo'},
    'jkp': {'en': 'Paku Karen'},
    'jkr': {'en': 'Koro (India)'},
    'jks': {'en': 'Amami Koniya Sign Language'},
    'jku': {'en': 'Labir'},
    'jle': {'en': 'Ngile'},
    'jls': {'en': 'Jamaican Sign Language'},
    'jma': {'en': 'Dima'},
    'jmb': {'en': 'Zumbun'},
    'jmc': {'af': 'Machame', 'am': 'ማቻሜኛ', 'ar': 'الماتشامية', 'as': 'মেকহেম', 'ast': 'machame', 'az': 'maçam', 'az-Cyrl': 'мачам', 'az-Latn': 'maçam', 'bal-Latn': 'Makami', 'be': 'мачамбэ', 'be-tarask': 'мачамбэ', 'bg': 'мачаме', 'bgn': 'ماچامه\u200cای', 'bn': 'মাকামে', 'br': 'machame', 'brx': 'मेचामे', 'bs': 'makame', 'bs-Cyrl': 'мачаме', 'bs-Latn': 'makame', 'ca': 'machame', 'ca-ES-valencia': 'machame', 'ccp': '𑄟𑄇𑄟𑄬', 'ce': 'мачаме', 'chr': 'ᎹᏣᎺ', 'ckb': 'ماچامێ', 'cs': 'mašame', 'cy': 'Matsiame', 'da': 'machame', 'de': 'Machame', 'dsb': 'machame', 'el': 'Ματσάμε', 'el-polyton': 'Ματσάμε', 'en': 'Machame', 'es': 'machame', 'et': 'matšame', 'eu': 'machamea', 'fa': 'ماچامه\u200cای', 'ff-Adlm': '𞤃𞤢𞤳𞤢𞤥𞤫𞥅𞤪𞤫', 'fi': 'machame', 'fil': 'Machame', 'fo': 'machame', 'fr': 'matchamé', 'fy': 'Machame', 'ga': 'Machame', 'gd': 'Machame', 'gl': 'machame', 'gu': 'મકામે', 'ha': 'Machame', 'he': 'מאקאמה', 'hi': 'मैकहैमे', 'hi-Latn': 'Machame', 'hr': 'machame', 'hsb': 'machame', 'ht': 'matchamé', 'hu': 'machame', 'hy': 'մաշամե', 'ia': 'machame', 'id': 'Machame', 'ig': 'Machame', 'is': 'masjáme', 'it': 'machame', 'ja': 'マチャメ語', 'jmc': 'Kimachame', 'jv': 'Machame', 'ka': 'კიმაშამი', 'kea': 'matxame', 'kgp': 'mỹsame', 'kk': 'мачаме тілі', 'kk-Cyrl': 'мачаме тілі', 'km': 'ម៉ាឆាំ', 'kn': 'ಮ್ಯಕಮೆ', 'ko': '마차메어', 'kok': 'मचामे', 'kok-Deva': 'मचामे', 'ku': 'machame', 'ky': 'мачамече', 'lb': 'Machame', 'lo': 'ມາແຊມ', 'lrc': 'ماچامئ', 'lt': 'mačamų', 'lv': 'mačamu', 'mai': 'मैकेम', 'mi': 'Mākame', 'mk': 'мачаме', 'ml': 'മചേം', 'mn': 'мачамэ', 'mr': 'मशामे', 'ms': 'Machame', 'mt': 'Machame', 'my': 'မချာမီ', 'mzn': 'ماچامه', 'nb': 'machame', 'ne': 'माचामे', 'nl': 'Machame', 'nn': 'machame', 'no': 'machame', 'nqo': 'ߡߊߗߊߡߋߞߊ߲', 'or': 'ମାଚେମେ', 'pa': 'ਮਚਾਮੇ', 'pa-Guru': 'ਮਚਾਮੇ', 'pcm': 'Machámẹ́ Lángwej', 'pl': 'machame', 'ps': 'ماچمی', 'pt': 'machame', 'qu': 'Machame Simi', 'rm': 'machame', 'ro': 'machame', 'ru': 'мачаме', 'sc': 'machame', 'sd': 'ميڪم', 'sd-Arab': 'ميڪم', 'si': 'මැකාමී', 'sk': 'mašame', 'sl': 'mačamejščina', 'smn': 'machame', 'so': 'Chaga', 'sq': 'maçamisht', 'sr': 'мачаме', 'sr-Cyrl': 'мачаме', 'sr-Latn': 'mačame', 'sv': 'kimashami', 'sw': 'Kimachame', 'ta': 'மாசெம்', 'te': 'మకొమ్', 'th': 'มาชาเม', 'ti': 'ማኬም', 'tk': 'maçame dili', 'to': 'lea fakamasame', 'tr': 'Machame', 'ug': 'ماچامچە', 'uk': 'мачаме', 'ur': 'ماشیم', 'uz': 'machame', 'uz-Cyrl': 'мачаме тили', 'uz-Latn': 'machame', 'vec': 'machame', 'vi': 'Tiếng Machame', 'xnr': 'मैकहैमे', 'yo': 'Máṣámè', 'yrl': 'maxami', 'yue': '馬恰美文', 'yue-Hans': '马恰美文', 'yue-Hant': '馬恰美文', 'zh': '马切姆语', 'zh-Hans': '马切姆语', 'zh-Hant': '馬恰美文', 'zu': 'isi-Machame'},
    'jmd': {'en': 'Yamdena'},
    'jmi': {'en': 'Jimi (Nigeria)'},
    'jml': {'en': 'Jumli'},
    'jmn': {'en': 'Makuri Naga'},
    'jmr': {'en': 'Kamara'},
    'jms': {'en': 'Mashi (Nigeria)'},
    'jmw': {'en': 'Mouwase'},
    'jmx': {'en': 'Western Juxtlahuaca Mixtec'},
    'jna': {'en': 'Jangshung'},
    'jnd': {'en': 'Jandavra'},
    'jng': {'en': 'Yangman'},
    'jni': {'en': 'Janji'},
    'jnj': {'en': 'Yemsa'},
    'jnl': {'en': 'Rawat'},
    'jns': {'en': 'Jaunsari'},
    'job': {'en': 'Joba'},
    'jod': {'en': 'Wojenaka'},
    'jog': {'en': 'Jogi'},
    'jor': {'en': 'Jorá'},
    'jos': {'en': 'Jordanian Sign Language'},
    'jow': {'en': 'Jowulu'},
    'jpa': {'en': 'Jewish Palestinian Aramaic'},
    'jpr': {'ar': 'الفارسية اليهودية', 'ast': 'xudeo-persa', 'az': 'ivrit-fars', 'az-Latn': 'ivrit-fars', 'bg': 'юдео-персийски', 'bn': 'জুদেও ফার্সি', 'br': 'yuzev-perseg', 'brx': 'यहुदी फ़ारसी', 'bs': 'judeo-perzijski', 'bs-Cyrl': 'јудео-персијски', 'bs-Latn': 'judeo-perzijski', 'ca': 'judeopersa', 'ca-ES-valencia': 'judeopersa', 'ccp': '𑄎𑄪𑄘𑄬𑄃𑄮 𑄜𑄢𑄴𑄥𑄨', 'cs': 'judeoperština', 'cy': 'Iddew-Bersieg', 'da': 'jødisk-persisk', 'de': 'Jüdisch-Persisch', 'el': 'Ιουδαϊκά-Περσικά', 'el-polyton': 'Ιουδαϊκά-Περσικά', 'en': 'Judeo-Persian', 'en-Dsrt': '𐐖𐐭𐐼𐐨𐐬-𐐑𐐯𐑉𐑈𐐲𐑌', 'es': 'judeo-persa', 'et': 'juudipärsia', 'fa': 'فارسی یهودی', 'fi': 'juutalaispersia', 'fr': 'judéo-persan', 'fy': 'Judeo-Perzysk', 'gd': 'Peirsis Iùdhach', 'gsw': 'Jüüdisch-Persisch', 'gu': 'જુદેઓ-પર્શિયન', 'he': 'פרסית יהודית', 'hi': 'जुदेओ-पर्शियन', 'hi-Latn': 'Judeo-Persian', 'hr': 'judejsko-perzijski', 'ht': 'judéo-persan', 'hu': 'zsidó-perzsa', 'id': 'Ibrani-Persia', 'is': 'gyðingapersneska', 'it': 'giudeo persiano', 'ja': 'ユダヤ・ペルシア語', 'ka': 'იუდეო-სპარსული', 'kgp': 'junajko-pérsa', 'kn': 'ಜೂಡಿಯೋ-ಪರ್ಶಿಯನ್', 'ko': '유대-페르시아어', 'ks': 'جوڈیو فارسی', 'ks-Arab': 'جوڈیو فارسی', 'ksh': 'Jühdesch-Pärsesch', 'lb': 'Jiddesch-Persesch', 'lo': 'ຈູແດວ-ເພີຊຽນ', 'lt': 'judėjų persų', 'lv': 'jūdpersiešu', 'mk': 'еврејскоперсиски', 'ml': 'ജൂഡിയോ-പേർഷ്യൻ', 'mr': 'जुदेओ-फारसी', 'mt': 'Lhudi-Persjan', 'my': 'ဂျူဒီယို-ပါရှန်', 'nb': 'jødepersisk', 'ne': 'जुडियो-फारसी', 'nl': 'Judeo-Perzisch', 'nn': 'jødepersisk', 'no': 'jødepersisk', 'or': 'ଜୁଡେଓ-ପର୍ସିଆନ୍', 'pl': 'judeo-perski', 'pt': 'judaico-persa', 'rm': 'giudaic-persian', 'ro': 'iudeo-persană', 'ru': 'еврейско-персидский', 'sk': 'židovská perzština', 'sl': 'judovska perzijščina', 'sr': 'јудео-персијски', 'sr-Cyrl': 'јудео-персијски', 'sr-Latn': 'judeo-persijski', 'sv': 'judisk persiska', 'ta': 'ஜூதேயோ-பெர்ஷியன்', 'te': 'జ్యుడియో-పర్షియన్', 'th': 'ยิว-เปอร์เซีย', 'to': 'lea fakaʻiuteo-pēsia', 'tr': 'Yahudi Farsçası', 'ug': 'ئىبرانى پارسچە', 'uk': 'юдео-перська', 'vi': 'Tiếng Judeo-Ba Tư', 'yi': 'יידיש־פערסיש', 'yrl': 'yudaiku-perisa', 'yue': '猶太教-波斯文', 'yue-Hans': '犹太教-波斯文', 'yue-Hant': '猶太教-波斯文', 'zh': '犹太波斯语', 'zh-Hans': '犹太波斯语', 'zh-Hant': '猶太教-波斯文'},
    'jpx': {'en': 'Japanese (family)'},
    'jpx-pro': {'en': 'Proto-Japonic'},
    'jpx-ryu-pro': {'en': 'Proto-Ryukyuan'},
    'jqr': {'en': 'Jaqaru'},
    'jra': {'en': 'Jarai'},
    'jrb': {'ar': 'العربية اليهودية', 'ast': 'xudeo-árabe', 'az': 'ivrit-ərəb', 'az-Latn': 'ivrit-ərəb', 'bg': 'юдео-арабски', 'bn': 'জুদেও আরবি', 'br': 'yuzev-arabeg', 'brx': 'यहुदी अरबी', 'bs': 'judeo-arapski', 'bs-Cyrl': 'јудео-арапски', 'bs-Latn': 'judeo-arapski', 'ca': 'judeoàrab', 'ca-ES-valencia': 'judeoàrab', 'ccp': '𑄎𑄪𑄘𑄬𑄃𑄮 𑄃𑄢𑄧𑄝𑄨', 'cs': 'judeoarabština', 'cy': 'Iddew-Arabeg', 'da': 'jødisk-arabisk', 'de': 'Jüdisch-Arabisch', 'el': 'Ιουδαϊκά-Αραβικά', 'el-polyton': 'Ιουδαϊκά-Αραβικά', 'en': 'Judeo-Arabic', 'en-Dsrt': '𐐖𐐭𐐼𐐨𐐬-𐐈𐑉𐐲𐐺𐐮𐐿', 'es': 'judeo-árabe', 'et': 'juudiaraabia', 'fa': 'عربی یهودی', 'fi': 'juutalaisarabia', 'fr': 'judéo-arabe', 'fy': 'Judeo-Arabysk', 'gd': 'Arabais Iùdhach', 'gsw': 'Jüüdisch-Arabisch', 'gu': 'જુદેઓ-અરબી', 'he': 'ערבית יהודית', 'hi': 'जुदेओ-अरेबिक', 'hi-Latn': 'Judeo-Arabic', 'hr': 'judejsko-arapski', 'ht': 'judéo-arabe', 'hu': 'zsidó-arab', 'id': 'Ibrani-Arab', 'is': 'gyðingaarabíska', 'it': 'giudeo arabo', 'ja': 'ユダヤ・アラビア語', 'ka': 'იუდეო-არაბული', 'kgp': 'junajko-aramiko', 'kn': 'ಜೂಡಿಯೋ-ಅರೇಬಿಕ್', 'ko': '유대-아라비아어', 'ks': 'جوڈیو عربی', 'ks-Arab': 'جوڈیو عربی', 'lb': 'Jiddesch-Arabesch', 'lo': 'ຈູແດວ-ອາລາບິກ', 'lt': 'judėjų arabų', 'lv': 'jūdarābu', 'mk': 'еврејскоарапски', 'ml': 'ജൂഡിയോ-അറബിക്', 'mr': 'जुदेओ-अरबी', 'mt': 'Lhudi-Għarbi', 'my': 'ဂျူဒီယို-အာရဗီ', 'nb': 'jødearabisk', 'ne': 'जुडियो-अरबी', 'nl': 'Judeo-Arabisch', 'nn': 'jødearabisk', 'no': 'jødearabisk', 'or': 'ଜୁଡେଓ-ଆରବୀକ୍', 'pl': 'judeoarabski', 'pt': 'judaico-arábico', 'rm': 'giudaic-arab', 'ro': 'iudeo-arabă', 'ru': 'еврейско-арабский', 'sk': 'židovská arabčina', 'sl': 'judovska arabščina', 'sr': 'јудео-арапски', 'sr-Cyrl': 'јудео-арапски', 'sr-Latn': 'judeo-arapski', 'sv': 'judisk arabiska', 'ta': 'ஜூதேயோ-அராபிக்', 'te': 'జ్యుడియో-అరబిక్', 'th': 'ยิว-อาหรับ', 'to': 'lea fakaʻiuteo-ʻalepea', 'tr': 'Yahudi Arapçası', 'ug': 'ئىبرانى ئەرەبچە', 'uk': 'юдео-арабська', 'vi': 'Tiếng Judeo-Ả Rập', 'yrl': 'yudaiku-arabiku', 'yue': '猶太阿拉伯文', 'yue-Hans': '犹太阿拉伯文', 'yue-Hant': '猶太阿拉伯文', 'zh': '犹太阿拉伯语', 'zh-Hans': '犹太阿拉伯语', 'zh-Hant': '猶太阿拉伯文'},
    'jrr': {'en': 'Jiru'},
    'jrt': {'en': 'Jakattoe'},
    'jru': {'en': 'Japrería'},
    'jsl': {'en': 'Japanese Sign Language'},
    'jua': {'en': 'Júma'},
    'jub': {'en': 'Wannu'},
    'juc': {'en': 'Jurchen'},
    'jud': {'en': 'Worodougou'},
    'juh': {'en': 'Hõne'},
    'jui': {'en': 'Ngadjuri'},
    'juk': {'en': 'Wapan'},
    'jul': {'en': 'Jirel'},
    'jum': {'en': 'Jumjum'},
    'jun': {'en': 'Juang'},
    'juo': {'en': 'Jiba'},
    'jup': {'en': 'Hupdë'},
    'jur': {'en': 'Jurúna'},
    'jus': {'en': 'Jumla Sign Language'},
    'jut': {'ast': 'jutlandés', 'cs': 'jutština', 'de': 'Jütisch', 'en': 'Jutish', 'et': 'jüüti', 'fi': 'juutti', 'fr': 'jute', 'ga': 'Iútlainnis', 'hi-Latn': 'Jutish', 'ht': 'jute', 'it': 'jutlandico', 'ja': 'ユトランド語', 'ksh': 'Jüttesch', 'lb': 'Jütesch', 'lt': 'danų kalbos jutų tarmė', 'mk': 'јитски', 'nb': 'jysk', 'ne': 'जुटिस', 'nl': 'Jutlands', 'nn': 'jysk', 'no': 'jysk', 'pl': 'jutlandzki', 'sv': 'jylländska', 'th': 'จัท', 'to': 'lea fakaʻiutilani', 'tr': 'Yutland Dili', 'vi': 'Tiếng Jutish', 'yue': '日德蘭文', 'yue-Hans': '日德兰文', 'yue-Hant': '日德蘭文', 'zh-Hant': '日德蘭文'},
    'juu': {'en': 'Ju'},
    'juw': {'en': 'Wãpha'},
    'juy': {'en': 'Juray'},
    'jv': {'af': 'Javaans', 'agq': 'Dzàbvànê', 'ak': 'Gyabanis kasa', 'am': 'ጃቫኛ', 'ar': 'الجاوية', 'as': 'জাভানী', 'asa': 'Kijava', 'ast': 'xavanés', 'az': 'yava', 'az-Cyrl': 'јава', 'az-Latn': 'yava', 'bal-Latn': 'Jáwáni', 'bas': 'Hɔp u yavà', 'be': 'яванская', 'be-tarask': 'яванская', 'bem': 'Ichi Javanisi', 'bez': 'Hijava', 'bg': 'явански', 'bgn': 'جاوه\u200cای', 'bm': 'javanekan', 'bn': 'জাভানিজ', 'br': 'javaneg', 'brx': 'जाभानीस', 'bs': 'javanski', 'bs-Cyrl': 'јавански', 'bs-Latn': 'javanski', 'ca': 'javanès', 'ca-ES-valencia': 'javanès', 'ccp': '𑄎𑄞𑄚𑄨𑄎𑄴', 'ce': 'яванийн', 'cgg': 'Orujava', 'chr': 'ᏆᏌ ᏣᏩ', 'ckb': 'جاڤایی', 'cs': 'javánština', 'cy': 'Jafanaeg', 'da': 'javanesisk', 'dav': 'Kijava', 'de': 'Javanisch', 'dje': 'Javanee senni', 'dsb': 'javašćina', 'dyo': 'savanee', 'dz': 'ཇཱ་བ་ནིས་ཁ', 'ebu': 'Kĩjava', 'ee': 'dzavangbe', 'el': 'Ιαβανικά', 'el-polyton': 'Ιαβανικά', 'en': 'Javanese', 'en-Dsrt': '𐐖𐐪𐑂𐐲𐑌𐐨𐑆', 'eo': 'java', 'es': 'javanés', 'et': 'jaava', 'eu': 'javera', 'ewo': 'Ǹkɔ́bɔ havanís', 'fa': 'جاوه\u200cای', 'ff': 'Sawaneere', 'ff-Adlm': '𞤔𞤢𞥄𞤱𞤢𞤫𞥅𞤪𞤫', 'ff-Latn': 'Sawaneere', 'fi': 'jaava', 'fil': 'Javanese', 'fo': 'javanskt', 'fr': 'javanais', 'fy': 'Javaansk', 'ga': 'Iáivis', 'gd': 'Deàbhanais', 'gl': 'xavanés', 'gsw': 'Javanisch', 'gu': 'જાવાનીસ', 'guz': 'Kijava', 'ha': 'Javananci', 'he': 'יאוואית', 'hi': 'जावानीज़', 'hi-Latn': 'Javanese', 'hr': 'javanski', 'hsb': 'javašćina', 'ht': 'javanais', 'hu': 'jávai', 'hy': 'ճավայերեն', 'ia': 'javanese', 'id': 'Jawa', 'ig': 'Javanese', 'is': 'javanska', 'it': 'giavanese', 'ja': 'ジャワ語', 'jmc': 'Kyijava', 'jv': 'Jawa', 'ka': 'იავური', 'kab': 'Tajavanit', 'kam': 'Kijava', 'kde': 'Chidyava', 'kea': 'javanes', 'kgp': 'javanẽj', 'khq': 'Javanee senni', 'ki': 'Kijava', 'kk': 'ява тілі', 'kk-Cyrl': 'ява тілі', 'kln': 'kutitap Javanese', 'km': 'ជ្វា', 'kn': 'ಜಾವಾನೀಸ್', 'ko': '자바어', 'kok': 'जावनीज', 'kok-Deva': 'जावनीज', 'ks': 'جَوَنیٖز', 'ks-Arab': 'جَوَنیٖز', 'ksb': 'Kijava', 'ksf': 'rijawanɛ́', 'ksh': 'Javahnesch', 'ku': 'javayî', 'ky': 'жаванизче', 'lag': 'Kɨjáava', 'lb': 'Javanesch', 'lg': 'Lunnajjava', 'lkt': 'Java Iyápi', 'lld': 'giavanesc', 'ln': 'lizava', 'lo': 'ຈາແວນີສ', 'lrc': 'جاڤئ یی', 'lt': 'javiečių', 'lu': 'Java', 'luo': 'Kijava', 'luy': 'Kijava', 'lv': 'javiešu', 'mai': 'जावानी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ ljana', 'mer': 'Kĩjava', 'mfe': 'zavane', 'mg': 'Javaney', 'mgh': 'Ijava', 'mi': 'Hāwhanihi', 'mk': 'јавански', 'ml': 'ജാവനീസ്', 'mn': 'ява', 'mr': 'जावानीज', 'ms': 'Jawa', 'mt': 'Ġavaniż', 'mua': 'javaniya', 'my': 'ဂျာဗား', 'mzn': 'جاوایی', 'naq': 'Javanese', 'nb': 'javanesisk', 'nd': 'isi-Java', 'ne': 'जाभानी', 'nl': 'Javaans', 'nmg': 'Kiɛl bó javanɛ̌', 'nn': 'javanesisk', 'no': 'javanesisk', 'nqo': 'ߖ߭ߝ߭ߊߣߊߞߊ߲', 'nus': 'Thok jabanithni', 'nyn': 'Orujava', 'om': 'Afaan Java', 'or': 'ଜାଭାନିଜ୍\u200c', 'pa': 'ਜਾਵਾਨੀਜ਼', 'pa-Guru': 'ਜਾਵਾਨੀਜ਼', 'pcm': 'Javáníz Lángwej', 'pl': 'jawajski', 'ps': 'جاوايي', 'pt': 'javanês', 'qu': 'Javanés Simi', 'rm': 'javanais', 'rn': 'Ikinyejava', 'ro': 'javaneză', 'rof': 'Kijava', 'ru': 'яванский', 'rw': 'Inyejava', 'rwk': 'Kyijava', 'saq': 'Kijava', 'sbp': 'Ishijava', 'sc': 'giavanesu', 'sd': 'جاونيز', 'sd-Arab': 'جاونيز', 'se': 'javagiella', 'seh': 'javanês', 'ses': 'Javanee senni', 'sg': 'Zavanëe', 'shi': 'ⵜⴰⵊⴰⴼⴰⵏⵉⵜ', 'shi-Latn': 'Tajavanit', 'shi-Tfng': 'ⵜⴰⵊⴰⴼⴰⵏⵉⵜ', 'si': 'ජාවා', 'sk': 'jávčina', 'sl': 'javanščina', 'smn': 'jaavakielâ', 'sn': 'chiJava', 'so': 'Jafaaniis', 'sq': 'javanisht', 'sr': 'јавански', 'sr-Cyrl': 'јавански', 'sr-Latn': 'javanski', 'sv': 'javanesiska', 'sw': 'Kijava', 'ta': 'ஜாவனீஸ்', 'te': 'జావనీస్', 'teo': 'Kijava', 'th': 'ชวา', 'ti': 'ጃቫንኛ', 'tk': 'ýawa dili', 'to': 'lea fakasava', 'tr': 'Cava dili', 'twq': 'Javanee senni', 'tzm': 'Tajavanit', 'ug': 'ياۋاچە', 'uk': 'яванська', 'ur': 'جاوی', 'uz': 'yavan', 'uz-Cyrl': 'яванча', 'uz-Latn': 'yavan', 'vai': 'ꕧꕙꕇꔧ', 'vai-Latn': 'Javaníĩ', 'vai-Vaii': 'ꕧꕙꕇꔧ', 'vec': 'javaneze', 'vi': 'Tiếng Java', 'vun': 'Kyijava', 'xnr': 'जावानीज़', 'xog': 'Olunnajjava', 'yav': 'yávanɛ', 'yi': 'יאַוואַנעזיש', 'yo': 'Èdè Javanasi', 'yrl': 'yawanei', 'yue': '爪哇文', 'yue-Hans': '爪哇文', 'yue-Hant': '爪哇文', 'zgh': 'ⵜⴰⵊⴰⴱⴰⵏⵉⵜ', 'zh': '爪哇语', 'zh-Hans': '爪哇语', 'zh-Hant': '爪哇文', 'zu': 'isi-Javanese'},
    'jvd': {'en': 'Javindo'},
    'jvn': {'en': 'Caribbean Javanese'},
    'jw': {'en': 'Javanese'},
    'jwi': {'en': 'Jwira-Pepesa'},
    'jya': {'en': 'Jiarong'},
    'jye': {'en': 'Judeo-Yemeni Arabic'},
    'jyy': {'en': 'Jaya'},
    'ka': {'af': 'Georgies', 'ak': 'Gyɔɔgyia kasa', 'am': 'ጆርጂያንኛ', 'ar': 'الجورجية', 'as': 'জৰ্জিয়ান', 'ast': 'xeorxanu', 'az': 'gürcü', 'az-Arab': 'گۆرجوجه', 'az-Cyrl': 'ҝүрҹү', 'az-Latn': 'gürcü', 'bal-Latn': 'Járjiái', 'be': 'грузінская', 'be-tarask': 'грузінская', 'bg': 'грузински', 'bgn': 'گرجی', 'bn': 'জর্জিয়ান', 'br': 'jorjianeg', 'brx': 'जर्जियान', 'bs': 'gruzijski', 'bs-Cyrl': 'грузијски', 'bs-Latn': 'gruzijski', 'ca': 'georgià', 'ca-ES-valencia': 'georgià', 'ccp': '𑄎𑄧𑄢𑄴𑄎𑄨𑄠𑄚𑄴', 'ce': 'гуьржийн', 'chr': 'ᏦᏥᎠᏂ', 'ckb': 'گۆرجستانی', 'cs': 'gruzínština', 'cy': 'Georgeg', 'da': 'georgisk', 'de': 'Georgisch', 'dsb': 'georgišćina', 'dz': 'ཇཽ་ཇི་ཡཱན་ཁ', 'ee': 'gɔgiagbe', 'el': 'Γεωργιανά', 'el-polyton': 'Γεωργιανά', 'en': 'Georgian', 'en-Dsrt': '𐐖𐐬𐑉𐐾𐐲𐑌', 'eo': 'kartvela', 'es': 'georgiano', 'et': 'gruusia', 'eu': 'georgiera', 'fa': 'گرجی', 'ff-Adlm': '𞤔𞤮𞥅𞤪𞥁𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'georgia', 'fil': 'Georgian', 'fo': 'georgiskt', 'fr': 'géorgien', 'fur': 'gjeorgjian', 'fy': 'Georgysk', 'ga': 'Seoirsis', 'gd': 'Cairtbheilis', 'gl': 'xeorxiano', 'gsw': 'Georgisch', 'gu': 'જ્યોર્જિયન', 'ha': 'Jojiyanci', 'he': 'גאורגית', 'hi': 'जॉर्जियाई', 'hi-Latn': 'Georgian', 'hr': 'gruzijski', 'hsb': 'georgišćina', 'ht': 'géorgien', 'hu': 'grúz', 'hy': 'վրացերեն', 'ia': 'georgiano', 'id': 'Georgia', 'ig': 'Georgian', 'is': 'georgíska', 'it': 'georgiano', 'ja': 'ジョージア語', 'jv': 'Georgia', 'ka': 'ქართული', 'kea': 'jorjianu', 'kgp': 'jiórjijỹnũ', 'kk': 'грузин тілі', 'kk-Cyrl': 'грузин тілі', 'km': 'ហ្សក\u200bហ្ស៊ី', 'kn': 'ಜಾರ್ಜಿಯನ್', 'ko': '조지아어', 'kok': 'जॉर्जियन', 'kok-Deva': 'जॉर्जियन', 'ks': 'جارجِیَن', 'ks-Arab': 'جارجِیَن', 'ksh': 'Je’orjesch', 'ku': 'gurcî', 'kxv': 'jorjiati', 'kxv-Deva': 'जॉर्जियाति', 'kxv-Latn': 'jorjiati', 'kxv-Orya': 'ଜର୍ଜିୟାତି', 'kxv-Telu': 'జర్జియన్', 'ky': 'грузинче', 'lb': 'Georgesch', 'lkt': 'Georia Iyápi', 'lld': 'georgian', 'lo': 'ຈໍຈຽນ', 'lrc': 'گورجی', 'lt': 'gruzinų', 'lv': 'gruzīnu', 'mai': 'जार्जियाई', 'mi': 'Hōriana', 'mk': 'грузиски', 'ml': 'ജോർജിയൻ', 'mn': 'гүрж', 'mr': 'जॉर्जियन', 'ms': 'Georgia', 'mt': 'Ġorġjan', 'my': 'ဂျော်ဂျီယာ', 'mzn': 'گرجی', 'nb': 'georgisk', 'ne': 'जर्जियाली', 'nl': 'Georgisch', 'nn': 'georgisk', 'no': 'georgisk', 'nqo': 'ߖ߭ߋߐߙߑߖ߭ߌߞߊ߲', 'om': 'Afaan Georgian', 'or': 'ଜର୍ଜିଆନ୍\u200c', 'os': 'гуырдзиаг', 'pa': 'ਜਾਰਜੀਆਈ', 'pa-Guru': 'ਜਾਰਜੀਆਈ', 'pcm': 'Jọ́jiá Lángwej', 'pl': 'gruziński', 'ps': 'جورجيائي', 'pt': 'georgiano', 'qu': 'Georgiano Simi', 'rm': 'georgian', 'ro': 'georgiană', 'ru': 'грузинский', 'rw': 'Inyejeworujiya', 'sah': 'Курусууннуу', 'sc': 'georgianu', 'sd': 'جارجيائي', 'sd-Arab': 'جارجيائي', 'se': 'georgiagiella', 'si': 'ජෝර්ජියානු', 'sk': 'gruzínčina', 'sl': 'gruzijščina', 'smn': 'georgiakielâ', 'so': 'Joorijiyaan', 'sq': 'gjeorgjisht', 'sr': 'грузијски', 'sr-Cyrl': 'грузијски', 'sr-Latn': 'gruzijski', 'sv': 'georgiska', 'sw': 'Kijojia', 'syr': 'ܓܘܪܓܝܐܝܬ', 'ta': 'ஜார்ஜியன்', 'te': 'జార్జియన్', 'tg': 'гурҷӣ', 'th': 'จอร์เจีย', 'ti': 'ጆርጅያንኛ', 'tk': 'gruzin dili', 'to': 'lea fakaseōsia', 'tr': 'Gürcüce', 'tt': 'грузин', 'ug': 'گىرۇزچە', 'uk': 'грузинська', 'ur': 'جارجیائی', 'uz': 'gruzincha', 'uz-Cyrl': 'грузинча', 'uz-Latn': 'gruzincha', 'vec': 'jeorjan', 'vi': 'Tiếng Georgia', 'wae': 'Georgiš', 'wo': 'Sorsiye', 'xnr': 'जॉर्जियाई', 'yi': 'גרוזיניש', 'yo': 'Èdè Georgia', 'yrl': 'geurgianu', 'yue': '喬治亞文', 'yue-Hans': '乔治亚文', 'yue-Hant': '喬治亞文', 'zh': '格鲁吉亚语', 'zh-Hans': '格鲁吉亚语', 'zh-Hant': '喬治亞文', 'zu': 'isi-Georgian'},
    'kaa': {'ar': 'الكارا-كالباك', 'ast': 'kara-kalpak', 'az': 'qaraqalpaq', 'az-Latn': 'qaraqalpaq', 'bal-Latn': 'Kara-Kalpak', 'bg': 'каракалпашки', 'bn': 'কারা-কাল্পাক', 'br': 'karakalpak', 'brx': 'कारा कलपक', 'bs': 'kara-kalpak', 'bs-Cyrl': 'кара-калпашки', 'bs-Latn': 'kara-kalpak', 'ca': 'karakalpak', 'ca-ES-valencia': 'karakalpak', 'ccp': '𑄇𑄢-𑄇𑄣𑄴𑄛𑄇𑄴', 'cs': 'karakalpačtina', 'cy': 'Cara-Calpaceg', 'da': 'karakalpakisk', 'de': 'Karakalpakisch', 'el': 'Κάρα-Καλπάκ', 'el-polyton': 'Κάρα-Καλπάκ', 'en': 'Kara-Kalpak', 'es': 'karakalpako', 'et': 'karakalpaki', 'fa': 'قره\u200cقالپاقی', 'fi': 'karakalpakki', 'fr': 'karakalpak', 'fy': 'Karakalpaks', 'ga': 'Cara-Chalpáis', 'gd': 'Kara-Kalpak', 'gsw': 'Karakalpakisch', 'gu': 'કારા-કલ્પક', 'he': 'קארא-קלפאק', 'hi': 'कारा-कल्पक', 'hi-Latn': 'Kara-Kalpak', 'hr': 'kara-kalpak', 'ht': 'karakalpak', 'hu': 'kara-kalpak', 'ia': 'kara-kalpak', 'id': 'Kara-Kalpak', 'ig': 'Kara-Kalpak', 'is': 'karakalpak', 'it': 'kara-kalpak', 'ja': 'カラカルパク語', 'ka': 'ყარაყალფახური', 'kaa': 'қарақалпақша', 'kaa-Cyrl': 'қарақалпақша', 'kgp': 'kara-karkag', 'kn': 'ಕಾರಾ-ಕಲ್ಪಾಕ್', 'ko': '카라칼파크어', 'ks': 'کارا کَلپَک', 'ks-Arab': 'کارا کَلپَک', 'ku': 'kara-kalpakî', 'lb': 'Karakalpakesch', 'lkt': 'Kara-Kalpak Iyápi', 'lo': 'ກາຣາ-ການປາກ', 'lt': 'karakalpakų', 'lv': 'karakalpaku', 'mk': 'каракалпачки', 'ml': 'കര-കാൽപ്പക്', 'mr': 'कारा-कल्पक', 'mt': 'Kara-Kalpak', 'nb': 'karakalpakisk', 'ne': 'कारा-काल्पाक', 'nl': 'Karakalpaks', 'nn': 'karakalpakisk', 'no': 'karakalpakisk', 'or': 'କାରା-କଲ୍ପକ୍', 'pl': 'karakałpacki', 'pt': 'kara-kalpak', 'rm': 'karakalpak', 'ro': 'karakalpak', 'ru': 'каракалпакский', 'sk': 'karakalpačtina', 'sl': 'karakalpaščina', 'sr': 'кара-калпашки', 'sr-Cyrl': 'кара-калпашки', 'sr-Latn': 'kara-kalpaški', 'sv': 'karakalpakiska', 'ta': 'காரா-கல்பாக்', 'te': 'కారా-కల్పాక్', 'th': 'การา-กาลพาก', 'ti': 'ካራ-ካልፓክ', 'to': 'lea fakakala-kalipaki', 'tr': 'Karakalpakça', 'ug': 'قارا-قالپاقچە', 'uk': 'каракалпацька', 'vec': 'caracalpago', 'vi': 'Tiếng Kara-Kalpak', 'xnr': 'कारा-कल्पक', 'yrl': 'kara-kaupaki', 'yue': '卡拉卡爾帕克文', 'yue-Hans': '卡拉卡尔帕克文', 'yue-Hant': '卡拉卡爾帕克文', 'zh': '卡拉卡尔帕克语', 'zh-Hans': '卡拉卡尔帕克语', 'zh-Hant': '卡拉卡爾帕克文'},
    'kab': {'af': 'Kabyle', 'am': 'ካብይል', 'ar': 'القبيلية', 'as': 'কাবাইল', 'ast': 'kabileñu', 'az': 'kabile', 'az-Cyrl': 'кабиле', 'az-Latn': 'kabile', 'bal-Latn': 'Kabáile', 'be': 'кабільская', 'be-tarask': 'кабільская', 'bg': 'кабилски', 'bgn': 'قبایلی', 'bn': 'কাবাইলে', 'br': 'kabileg', 'brx': 'काबील', 'bs': 'kabile', 'bs-Cyrl': 'кабиле', 'bs-Latn': 'kabile', 'ca': 'cabilenc', 'ca-ES-valencia': 'cabilenc', 'ccp': '𑄇𑄝𑄭𑄣𑄬', 'ce': 'кабилийн', 'chr': 'ᎧᏈᎴ', 'ckb': 'کبائیلی', 'cs': 'kabylština', 'cy': 'Cabileg', 'da': 'kabylsk', 'de': 'Kabylisch', 'dsb': 'kabylšćina', 'el': 'Καμπίλε', 'el-polyton': 'Καμπίλε', 'en': 'Kabyle', 'es': 'cabileño', 'et': 'kabiili', 'eu': 'kabiliera', 'fa': 'قبایلی', 'ff-Adlm': '𞤑𞤢𞤦𞤭𞤤𞤭𞥅𞤪𞤫', 'fi': 'kabyyli', 'fil': 'Kabyle', 'fo': 'kabyle', 'fr': 'kabyle', 'fy': 'Kabyle', 'ga': 'Caibílis', 'gd': 'Kabyle', 'gl': 'cabila', 'gsw': 'Kabylisch', 'gu': 'કબાઇલ', 'ha': 'Kabyle', 'he': 'קבילה', 'hi': 'कबाइल', 'hi-Latn': 'Kabyle', 'hr': 'kabilski', 'hsb': 'kabylšćina', 'ht': 'kabyle', 'hu': 'kabije', 'hy': 'կաբիլերեն', 'ia': 'kabylo', 'id': 'Kabyle', 'ig': 'Kabyle', 'is': 'kabíle', 'it': 'cabilo', 'ja': 'カビル語', 'jv': 'Kabyle', 'ka': 'კაბილური', 'kab': 'Taqbaylit', 'kea': 'kabila', 'kgp': 'kamyre', 'kk': 'кабил тілі', 'kk-Cyrl': 'кабил тілі', 'km': 'កាប៊ីឡេ', 'kn': 'ಕಬೈಲ್', 'ko': '커바일어', 'kok': 'काबायले', 'kok-Deva': 'काबायले', 'ks': 'کَبایِل', 'ks-Arab': 'کَبایِل', 'ksh': 'Kabyhlesch', 'ku': 'kabîlî', 'ky': 'кабылча', 'lb': 'Kabylesch', 'lo': 'ກາໄບລ໌', 'lrc': 'کابیلئ', 'lt': 'kebailų', 'lv': 'kabilu', 'mai': 'कबाइल', 'mi': 'Kapāiro', 'mk': 'кабилски', 'ml': 'കബൈൽ', 'mn': 'кабиле', 'mr': 'कबाइल', 'ms': 'Kabyle', 'mt': 'Kabuljan', 'my': 'ကဘိုင်လ်', 'mzn': 'قبایلی', 'nb': 'kabylsk', 'ne': 'काबिल', 'nl': 'Kabylisch', 'nn': 'kabylsk', 'no': 'kabylsk', 'nqo': 'ߞߊߓߟߌߞߊ߲', 'or': 'କବାଇଲ୍', 'pa': 'ਕਬਾਇਲ', 'pa-Guru': 'ਕਬਾਇਲ', 'pcm': 'Kabail Lángwej', 'pl': 'kabylski', 'ps': 'کیبیل', 'pt': 'kabyle', 'qu': 'Cabilio Simi', 'rm': 'kabyle', 'ro': 'kabyle', 'ru': 'кабильский', 'sc': 'cabilu', 'sd': 'ڪبائل', 'sd-Arab': 'ڪبائل', 'si': 'කාබිල්', 'sk': 'kabylčina', 'sl': 'kabilščina', 'smn': 'kabylkielâ', 'so': 'Kabayle', 'sq': 'kabilisht', 'sr': 'кабиле', 'sr-Cyrl': 'кабиле', 'sr-Latn': 'kabile', 'sv': 'kabyliska', 'sw': 'Kikabylia', 'ta': 'கபாய்ல்', 'te': 'కాబిల్', 'th': 'กาไบล', 'ti': 'ካቢልኛ', 'tk': 'kabil dili', 'to': 'lea fakakapile', 'tr': 'Kabiliyece', 'ug': 'كابىلېچە', 'uk': 'кабільська', 'ur': 'قبائلی', 'uz': 'kabil', 'uz-Cyrl': 'кабилча', 'uz-Latn': 'kabil', 'vec': 'cabil', 'vi': 'Tiếng Kabyle', 'xnr': 'कबाइल', 'yo': 'Kabilè', 'yrl': 'kabire', 'yue': '卡比爾文', 'yue-Hans': '卡比尔文', 'yue-Hant': '卡比爾文', 'zh': '卡拜尔语', 'zh-Hans': '卡拜尔语', 'zh-Hant': '卡比爾文', 'zu': 'isi-Kabyle'},
    'kac': {'af': 'Kachin', 'am': 'ካቺን', 'ar': 'الكاتشين', 'as': 'কাচিন', 'ast': 'kachin', 'az': 'kaçin', 'az-Cyrl': 'качин', 'az-Latn': 'kaçin', 'be': 'качынская', 'be-tarask': 'качынская', 'bg': 'качински', 'bgn': 'کاچینی', 'bn': 'কাচিন', 'br': 'kachin', 'brx': 'काचीन', 'bs': 'kačin', 'bs-Cyrl': 'качин', 'bs-Latn': 'kačin', 'ca': 'katxin', 'ca-ES-valencia': 'katxin', 'ccp': '𑄇𑄌𑄨𑄚𑄴', 'ce': 'качинийн', 'chr': 'ᎧᏥᏂ', 'ckb': 'کاچین', 'cs': 'kačijština', 'cy': 'Kachin', 'da': 'kachin', 'de': 'Kachin', 'dsb': 'kachinšćina', 'dz': 'ཀ་ཆིན་ཁ', 'el': 'Κατσίν', 'el-polyton': 'Κατσίν', 'en': 'Kachin', 'es': 'kachin', 'et': 'katšini', 'eu': 'jingphoera', 'fa': 'کاچینی', 'ff-Adlm': '𞤑𞤢𞤧𞤭𞤲𞤪𞤫', 'fi': 'katšin', 'fil': 'Kachin', 'fo': 'kachin', 'fr': 'kachin', 'fy': 'Kachin', 'ga': 'Caitsinis', 'gd': 'Kachin', 'gl': 'kachin', 'gsw': 'Kachin-Schpraach', 'gu': 'કાચિન', 'ha': 'Kachin', 'he': 'קצ׳ין', 'hi': 'काचिन', 'hi-Latn': 'Kachin', 'hr': 'kačinski', 'hsb': 'kachinšćina', 'ht': 'kachin', 'hu': 'kacsin', 'hy': 'կաչիներեն', 'ia': 'kachin', 'id': 'Kachin', 'ig': 'Kachin', 'is': 'kasín', 'it': 'kachin', 'ja': 'カチン語', 'jv': 'Kakhin', 'ka': 'კაჩინი', 'kgp': 'kasĩn', 'kk': 'качин тілі', 'kk-Cyrl': 'качин тілі', 'km': 'កាឈីន', 'kn': 'ಕಚಿನ್', 'ko': '카친어', 'kok': 'काचीन', 'kok-Deva': 'काचीन', 'ks': 'کاچِن', 'ks-Arab': 'کاچِن', 'ku': 'cingphoyî', 'ky': 'кахинче', 'lb': 'Kachin-Sprooch', 'lo': 'ກາຊິນ', 'lt': 'kačinų', 'lv': 'kačinu', 'mai': 'कचिन', 'mi': 'Katīana', 'mk': 'качински', 'ml': 'കാചിൻ', 'mn': 'качин', 'mr': 'काचिन', 'ms': 'Kachin', 'mt': 'Kachin', 'my': 'ကချင်', 'nb': 'kachin', 'ne': 'काचिन', 'nl': 'Kachin', 'nn': 'kachin', 'no': 'kachin', 'or': 'କଚିନ୍', 'pa': 'ਕਾਚਿਨ', 'pa-Guru': 'ਕਾਚਿਨ', 'pcm': 'Kachín Lángwej', 'pl': 'kaczin', 'ps': 'کاچین', 'pt': 'kachin', 'qu': 'Kachin Simi', 'rm': 'kachin', 'ro': 'kachin', 'ru': 'качинский', 'sc': 'kachin', 'sd': 'ڪچن', 'sd-Arab': 'ڪچن', 'si': 'කචින්', 'sk': 'kačjinčina', 'sl': 'kačinščina', 'smn': 'kachin', 'so': 'Kachin', 'sq': 'kaçinisht', 'sr': 'качински', 'sr-Cyrl': 'качински', 'sr-Latn': 'kačinski', 'sv': 'kachin', 'sw': 'Kachin', 'ta': 'காசின்', 'te': 'కాచిన్', 'th': 'กะฉิ่น', 'ti': 'ካቺን', 'tk': 'kaçin dili', 'to': 'lea fakakasini', 'tr': 'Kaçin dili', 'ug': 'كاچىنچە', 'uk': 'качін', 'ur': 'کاچن', 'uz': 'kachin', 'uz-Latn': 'kachin', 'vec': 'kachin', 'vi': 'Tiếng Kachin', 'xnr': 'काचिन', 'yo': 'Èdè Kashini', 'yrl': 'kaxĩ', 'yue': '卡琴文', 'yue-Hans': '卡琴文', 'yue-Hant': '卡琴文', 'zh': '克钦语', 'zh-Hans': '克钦语', 'zh-Hant': '卡琴文', 'zu': 'isi-Kachin'},
    'kad': {'en': 'Adara'},
    'kae': {'en': 'Ketangalan'},
    'kaf': {'en': 'Katso'},
    'kag': {'en': 'Kajaman'},
    'kah': {'en': 'Kara (Central African Republic)'},
    'kai': {'en': 'Karekare'},
    'kaj': {'af': 'Jju', 'am': 'ጅጁ', 'ar': 'الجو', 'as': 'জজু', 'ast': 'jju', 'az': 'ju', 'az-Cyrl': 'жу', 'az-Latn': 'ju', 'bal-Latn': 'Jju', 'be': 'дджу', 'be-tarask': 'дджу', 'bg': 'жжу', 'bgn': 'جیجو', 'bn': 'জজু', 'br': 'jju', 'brx': 'झु', 'bs': 'kaju', 'bs-Cyrl': 'ђу', 'bs-Latn': 'kaju', 'ca': 'jju', 'ca-ES-valencia': 'jju', 'ccp': '𑄃𑄧𑄌𑄴𑄎𑄪', 'ce': 'каджи', 'chr': 'ᏥᏧ', 'ckb': 'کیجوو', 'cs': 'jju', 'cy': 'Jju', 'da': 'jju', 'de': 'Jju', 'dsb': 'jjušćina', 'el': 'Τζου', 'el-polyton': 'Τζου', 'en': 'Jju', 'es': 'jju', 'et': 'jju', 'eu': 'jjua', 'fa': 'جو', 'ff-Adlm': '𞤑𞤢𞤶𞤫𞥅𞤪𞤫', 'fi': 'jju', 'fil': 'Jju', 'fo': 'jju', 'fr': 'jju', 'fy': 'Jju', 'ga': 'Jju', 'gd': 'Jju', 'gl': 'jju', 'gsw': 'Jju', 'gu': 'જ્જુ', 'ha': 'Jju', 'he': 'ג׳ו', 'hi': 'ज्जु', 'hi-Latn': 'Jju', 'hr': 'kaje', 'hsb': 'jjušćina', 'ht': 'jju', 'hu': 'jju', 'hy': 'ջյու', 'ia': 'jju', 'id': 'Jju', 'ig': 'Jju', 'is': 'jju', 'it': 'jju', 'ja': 'カジェ語', 'jv': 'Jju', 'ka': 'კაჯი', 'kaj': 'Kaje', 'kgp': 'ju', 'kk': 'каджи тілі', 'kk-Cyrl': 'каджи тілі', 'km': 'ជូ', 'kn': 'ಜ್ಜು', 'ko': '까꼬토끄어', 'kok': 'जु', 'kok-Deva': 'जु', 'ks': 'جُوٗ', 'ks-Arab': 'جُوٗ', 'ku': 'jju', 'ky': 'джуча', 'lb': 'Jju', 'lo': 'ຈຣູ', 'lt': 'ju', 'lv': 'kadži', 'mai': 'जू', 'mi': 'Heiho', 'mk': 'каџе', 'ml': 'ജ്യൂ', 'mn': 'жжу', 'mr': 'ज्जु', 'ms': 'Jju', 'mt': 'Jju', 'my': 'ဂျူအူ', 'nb': 'jju', 'ne': 'ज्जु', 'nl': 'Jju', 'nn': 'jju', 'no': 'jju', 'or': 'ଜଜୁ', 'pa': 'ਜਜੂ', 'pa-Guru': 'ਜਜੂ', 'pcm': 'Jju Lángwej', 'pl': 'jju', 'ps': 'ججو', 'pt': 'jju', 'qu': 'Jju Simi', 'rm': 'jju', 'ro': 'jju', 'ru': 'каджи', 'sc': 'jju', 'sd': 'پوڪيپسي', 'sd-Arab': 'پوڪيپسي', 'si': 'ජ්ජු', 'sk': 'jju', 'sl': 'jju', 'smn': 'jju', 'so': 'Jju', 'sq': 'kajeisht', 'sr': 'џу', 'sr-Cyrl': 'џу', 'sr-Latn': 'džu', 'sv': 'jju', 'sw': 'Kijju', 'ta': 'ஜ்ஜூ', 'te': 'జ్యూ', 'th': 'คจู', 'ti': 'ጅጁ', 'tk': 'ju dili', 'to': 'lea fakasisū', 'tr': 'Jju', 'ug': 'جۇچە', 'uk': 'йю', 'ur': 'جے جو', 'uz': 'kaji', 'uz-Cyrl': 'кажи', 'uz-Latn': 'kaji', 'vec': 'jju', 'vi': 'Tiếng Jju', 'xnr': 'ज्जु', 'yo': 'Èdè Ju', 'yrl': 'iyu', 'yue': '卡捷文', 'yue-Hans': '卡捷文', 'yue-Hant': '卡捷文', 'zh': '卡捷语', 'zh-Hans': '卡捷语', 'zh-Hant': '卡捷文', 'zu': 'isi-Jju'},
    'kak': {'en': 'Kalanguya'},
    'kam': {'af': 'Kamba', 'am': 'ካምባ', 'ar': 'الكامبا', 'as': 'কাম্বা', 'ast': 'kamba', 'az': 'kamba', 'az-Cyrl': 'камба', 'az-Latn': 'kamba', 'bal-Latn': 'Kambá', 'be': 'камба', 'be-tarask': 'камба', 'bg': 'камба', 'bgn': 'کامبایی', 'bn': 'কাম্বা', 'br': 'kamba', 'brx': 'काम्बा', 'bs': 'kamba', 'bs-Cyrl': 'камба', 'bs-Latn': 'kamba', 'ca': 'kamba', 'ca-ES-valencia': 'kamba', 'ccp': '𑄇𑄟𑄴𑄝', 'ce': 'камба', 'chr': 'ᎧᎻᏆ', 'ckb': 'کامبا', 'cs': 'kambština', 'cy': 'Camba', 'da': 'kamba', 'de': 'Kamba', 'dsb': 'kamba', 'el': 'Κάμπα', 'el-polyton': 'Κάμπα', 'en': 'Kamba', 'es': 'kamba', 'et': 'kamba', 'eu': 'kambera', 'fa': 'کامبایی', 'ff-Adlm': '𞤑𞤢𞤥𞤦𞤢𞥄𞤪𞤫', 'fi': 'kamba', 'fil': 'Kamba', 'fo': 'kamba', 'fr': 'kamba', 'fy': 'Kamba', 'ga': 'Cambais', 'gd': 'Kamba', 'gl': 'kamba', 'gsw': 'Kamba', 'gu': 'કમ્બા', 'ha': 'Kamba', 'he': 'קמבה', 'hi': 'कम्बा', 'hi-Latn': 'Kamba', 'hr': 'kamba', 'hsb': 'kamba', 'ht': 'kamba', 'hu': 'kamba', 'hy': 'կամբա', 'ia': 'kamba', 'id': 'Kamba', 'ig': 'Kamba', 'is': 'kamba', 'it': 'kamba', 'ja': 'カンバ語', 'jv': 'Kamba', 'ka': 'კამბა', 'kam': 'Kikamba', 'kea': 'kamba', 'kgp': 'kỹma', 'kk': 'камба тілі', 'kk-Cyrl': 'камба тілі', 'km': 'កាំបា', 'kn': 'ಕಂಬಾ', 'ko': '캄바어', 'kok': 'कंबा', 'kok-Deva': 'कंबा', 'ks': 'کامبا', 'ks-Arab': 'کامبا', 'ksh': 'Kamba vun Kehnija', 'ku': 'kambayî', 'ky': 'камбача', 'lb': 'Kamba', 'lo': 'ກາມບາ', 'lrc': 'کامبا', 'lt': 'kembų', 'lv': 'kambu', 'mai': 'कामबा', 'mi': 'Kāmapa', 'mk': 'камба', 'ml': 'കംബ', 'mn': 'камба', 'mr': 'काम्बा', 'ms': 'Kamba', 'mt': 'Kamba', 'my': 'ကမ်ဘာ', 'mzn': 'کامبایی', 'nb': 'kamba', 'ne': 'काम्बा', 'nl': 'Kamba', 'nn': 'kamba', 'no': 'kamba', 'nqo': 'ߞߊ߲ߓߊߞߊ߲', 'or': 'କମ୍ବା', 'pa': 'ਕੰਬਾ', 'pa-Guru': 'ਕੰਬਾ', 'pcm': 'Kámbá Lángwej', 'pl': 'kamba', 'ps': 'کامبا', 'pt': 'kamba', 'qu': 'Kamba Simi', 'rm': 'kamba', 'ro': 'kamba', 'ru': 'камба', 'sc': 'kamba', 'sd': 'ڪئمبا', 'sd-Arab': 'ڪئمبا', 'si': 'කැම්බා', 'sk': 'kamba', 'sl': 'kambaščina', 'smn': 'kambakielâ', 'so': 'Kaamba', 'sq': 'kambaisht', 'sr': 'камба', 'sr-Cyrl': 'камба', 'sr-Latn': 'kamba', 'sv': 'kamba', 'sw': 'Kikamba', 'ta': 'கம்பா', 'te': 'కంబా', 'th': 'คัมบา', 'ti': 'ካምባ', 'tk': 'kamba dili', 'to': 'lea fakakamipa', 'tr': 'Kamba', 'ug': 'كامباچە', 'uk': 'камба', 'ur': 'کامبا', 'uz': 'kamba', 'uz-Cyrl': 'камбача', 'uz-Latn': 'kamba', 'vec': 'kanba', 'vi': 'Tiếng Kamba', 'xnr': 'कम्बा', 'yo': 'Káńbà', 'yrl': 'kãba', 'yue': '卡姆巴文', 'yue-Hans': '卡姆巴文', 'yue-Hant': '卡姆巴文', 'zh': '卡姆巴语', 'zh-Hans': '卡姆巴语', 'zh-Hant': '卡姆巴文', 'zu': 'isi-Kamba'},
    'kao': {'en': 'Xaasongaxango'},
    'kap': {'en': 'Bezhta'},
    'kaq': {'en': 'Capanahua'},
    'kar': {'en': 'Karen languages'},
    'kav': {'en': 'Katukína'},
    'kaw': {'ar': 'الكوي', 'ast': 'kawi', 'az': 'kavi', 'az-Latn': 'kavi', 'bg': 'кави', 'bn': 'কাউই', 'brx': 'कावी', 'bs': 'kavi', 'bs-Cyrl': 'кави', 'bs-Latn': 'kavi', 'ca': 'kawi', 'ca-ES-valencia': 'kawi', 'ccp': '𑄇𑄃𑄪𑄃𑄨', 'cs': 'kawi', 'da': 'kawi', 'de': 'Kawi', 'el': 'Κάουι', 'el-polyton': 'Κάουι', 'en': 'Kawi', 'es': 'kawi', 'et': 'kaavi', 'fa': 'کاویایی', 'fi': 'kavi', 'fr': 'kawi', 'fy': 'Kawi', 'gd': 'Kawi', 'gsw': 'Kawi', 'gu': 'કાવી', 'he': 'קאווי', 'hi': 'कावी', 'hi-Latn': 'Kawi', 'hr': 'kawi', 'ht': 'kawi', 'hu': 'kawi', 'id': 'Kawi', 'is': 'kaví', 'it': 'kawi', 'ja': 'カウィ語', 'kgp': 'kavi', 'kn': 'ಕಾವಿ', 'ko': '카위어', 'ks': 'کَوی', 'ks-Arab': 'کَوی', 'lb': 'Kawi', 'lo': 'ກະວີ', 'lt': 'kavių', 'lv': 'kāvi', 'mk': 'кави', 'ml': 'കാവി', 'mr': 'कावी', 'mt': 'Kawi', 'nb': 'kawi', 'ne': 'कावी', 'nl': 'Kawi', 'nn': 'kawi', 'no': 'kawi', 'or': 'କାୱି', 'pl': 'kawi', 'pt': 'kawi', 'rm': 'kawi', 'ro': 'kawi', 'ru': 'кави', 'sk': 'kawi', 'sl': 'kavi', 'sr': 'кави', 'sr-Cyrl': 'кави', 'sr-Latn': 'kavi', 'sv': 'kawi', 'ta': 'காவி', 'te': 'కావి', 'th': 'กวี', 'to': 'lea fakakavi', 'tr': 'Kawi', 'ug': 'كاۋىچە', 'uk': 'каві', 'vi': 'Tiếng Kawi', 'yrl': 'kawi', 'yue': '卡威文', 'yue-Hans': '卡威文', 'yue-Hant': '卡威文', 'zh': '卡威语', 'zh-Hans': '卡威语', 'zh-Hant': '卡威文'},
    'kax': {'en': 'Kao'},
    'kay': {'en': 'Kamayurá'},
    'kba': {'en': 'Kalarko'},
    'kbb': {'en': 'Kaxuiâna'},
    'kbc': {'en': 'Kadiwéu'},
    'kbd': {'af': 'Kabardiaans', 'am': 'ካባርዲያን', 'ar': 'الكاباردايان', 'as': 'কাবাৰ্ডিয়ান', 'ast': 'kabardianu', 'az': 'kabarda-çərkəz', 'az-Cyrl': 'кабарда-чәркәз', 'az-Latn': 'kabarda-çərkəz', 'be': 'кабардзінская', 'be-tarask': 'кабардзінская', 'bg': 'кабардски', 'bgn': 'کاباردینی', 'bn': 'কাবার্ডিয়ান', 'br': 'kabardeg', 'brx': 'काबार्डीयानि', 'bs': 'kabardijski', 'bs-Cyrl': 'кабардијски', 'bs-Latn': 'kabardijski', 'ca': 'kabardí', 'ca-ES-valencia': 'kabardí', 'ccp': '𑄇𑄝𑄢𑄴𑄓𑄨𑄠𑄚𑄴', 'ce': 'гӀебартойн', 'chr': 'ᎧᏆᏗᎠᏂ', 'ckb': 'کاباردی', 'cs': 'kabardinština', 'cy': 'Cabardieg', 'da': 'kabardian', 'de': 'Kabardinisch', 'dsb': 'kabardianšćina', 'el': 'Καμπαρντιανά', 'el-polyton': 'Καμπαρντιανά', 'en': 'Kabardian', 'es': 'kabardiano', 'et': 'kabardi-tšerkessi', 'eu': 'kabardiera', 'fa': 'کاباردینی', 'ff-Adlm': '𞤑𞤢𞤦𞤢𞤪𞤣𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'kabardi', 'fil': 'Kabardian', 'fo': 'kabardinskt', 'fr': 'kabarde', 'fy': 'Kabardysk', 'ga': 'Cabairdis', 'gd': 'Cabardais', 'gl': 'cabardiano', 'gsw': 'Kabardinisch', 'gu': 'કબાર્ડિયન', 'ha': 'Karbadiyanci', 'he': 'קברדית', 'hi': 'कबार्डियन', 'hi-Latn': 'Kabardian', 'hr': 'kabardinski', 'hsb': 'kabardianšćina', 'ht': 'kabarde', 'hu': 'kabardi', 'hy': 'կաբարդերեն', 'ia': 'cabardiano', 'id': 'Kabardi', 'ig': 'Kabadian', 'is': 'kabardíska', 'it': 'cabardino', 'ja': 'カバルド語', 'jv': 'Kabardian', 'ka': 'ყაბარდოული', 'kgp': 'kamarnhijỹnũ', 'kk': 'кабардин тілі', 'kk-Cyrl': 'кабардин тілі', 'km': 'កាបាឌៀ', 'kn': 'ಕಬರ್ಡಿಯನ್', 'ko': '카바르디어', 'kok': 'काबार्डियन', 'kok-Deva': 'काबार्डियन', 'ks': 'کَبارڈِیَن', 'ks-Arab': 'کَبارڈِیَن', 'ksh': 'Kabadihnesch', 'ku': 'kabardî', 'ky': 'кабардинче', 'lb': 'Kabardinesch', 'lkt': 'Kabardian Iyápi', 'lo': 'ກາບາດຽນ', 'lt': 'kabardinų', 'lv': 'kabardiešu', 'mai': 'कबर्दियन', 'mi': 'Kapāriana', 'mk': 'кабардински', 'ml': 'കബർഡിയാൻ', 'mn': 'кабардин', 'mr': 'कबार्डियन', 'ms': 'Kabardia', 'mt': 'Kabardian', 'my': 'ကဘာဒင်', 'nb': 'kabardisk', 'ne': 'काबार्दियाली', 'nl': 'Kabardisch', 'nn': 'kabardisk', 'no': 'kabardisk', 'or': 'କାବାର୍ଡିଆନ୍', 'os': 'кӕсгон', 'pa': 'ਕਬਾਰਦੀ', 'pa-Guru': 'ਕਬਾਰਦੀ', 'pcm': 'Kabárdian', 'pl': 'kabardyjski', 'ps': 'کابیرین', 'pt': 'kabardiano', 'qu': 'Kabardiano Simi', 'rm': 'kabardic', 'ro': 'kabardian', 'ru': 'кабардинский', 'sc': 'cabardianu', 'sd': 'ڪبارڊيئن', 'sd-Arab': 'ڪبارڊيئن', 'si': 'කබාර්ඩියන්', 'sk': 'kabardčina', 'sl': 'kabardinščina', 'smn': 'kabardikielâ', 'so': 'U dhashay Kabardia', 'sq': 'kabardianisht', 'sr': 'кабардијски', 'sr-Cyrl': 'кабардијски', 'sr-Latn': 'kabardijski', 'sv': 'kabardinska', 'sw': 'Kikabardian', 'ta': 'கபார்டியன்', 'te': 'కబార్డియన్', 'th': 'คาบาร์เดีย', 'ti': 'ካባርድኛ', 'tk': 'kabardin dili', 'to': 'lea fakakapālitia', 'tr': 'Kabardeyce', 'ug': 'كاباردەيچە', 'uk': 'кабардинська', 'ur': 'کبارڈین', 'uz': 'kabardin', 'uz-Latn': 'kabardin', 'vec': 'cabardian', 'vi': 'Tiếng Kabardian', 'xnr': 'कबार्डियन', 'yo': 'Èdè Kabadia', 'yrl': 'kabaridianu', 'yue': '卡巴爾達文', 'yue-Hans': '卡巴尔达文', 'yue-Hant': '卡巴爾達文', 'zh': '卡巴尔德语', 'zh-Hans': '卡巴尔德语', 'zh-Hant': '卡巴爾達文', 'zu': 'isi-Kabardian'},
    'kbe': {'en': 'Kanju'},
    'kbf': {'en': 'Kakauhua'},
    'kbg': {'en': 'Khamba'},
    'kbh': {'en': 'Camsá'},
    'kbi': {'en': 'Kaptiau'},
    'kbj': {'en': 'Kari'},
    'kbk': {'en': 'Grass Koiari'},
    'kbl': {'ar': 'كانمبو', 'ast': 'kanembu', 'bs': 'kanembu', 'bs-Latn': 'kanembu', 'ca': 'kanembu', 'ca-ES-valencia': 'kanembu', 'cs': 'kanembu', 'da': 'kanembu', 'de': 'Kanembu', 'el': 'Κανέμπου', 'el-polyton': 'Κανέμπου', 'en': 'Kanembu', 'es': 'kanembu', 'et': 'kanembu', 'fi': 'kanembu', 'fr': 'kanembou', 'fy': 'Kanembu', 'gd': 'Kanembu', 'he': 'קנמבו', 'hi-Latn': 'Kanembu', 'hr': 'kanembu', 'ht': 'kanembou', 'hu': 'kanembu', 'id': 'Kanembu', 'it': 'kanembu', 'ja': 'カネンブ語', 'kgp': 'kanẽnmu', 'ko': '카넴부어', 'lb': 'Kanembu', 'lo': 'ຄາແນມບູ', 'lt': 'kanembų', 'lv': 'kaņembu', 'mk': 'канембу', 'ml': 'കനെംബു', 'ms': 'Kanembu', 'nb': 'kanembu', 'ne': 'कानेम्बु', 'nl': 'Kanembu', 'nn': 'kanembu', 'no': 'kanembu', 'pl': 'kanembu', 'pt': 'kanembu', 'ro': 'kanembu', 'ru': 'канембу', 'sk': 'kanembu', 'sv': 'kanembu', 'sw': 'Kikanembu', 'th': 'คาเนมบู', 'to': 'lea fakakanēmipu', 'tr': 'Kanembu', 'ug': 'كانېمبۇچە', 'uk': 'канембу', 'vi': 'Tiếng Kanembu', 'yrl': 'kanẽbu', 'yue': '卡念布文', 'yue-Hans': '卡念布文', 'yue-Hant': '卡念布文', 'zh': '加涅姆布语', 'zh-Hans': '加涅姆布语', 'zh-Hant': '卡念布文'},
    'kbm': {'en': 'Iwal'},
    'kbn': {'en': 'Kare (Central African Republic)'},
    'kbo': {'en': 'Keliko'},
    'kbp': {'en': 'Kabiyè'},
    'kbq': {'en': 'Kamano'},
    'kbr': {'en': 'Kafa'},
    'kbs': {'en': 'Kande'},
    'kbt': {'en': 'Abadi'},
    'kbu': {'en': 'Kabutra'},
    'kbv': {'en': 'Dera (Indonesia)'},
    'kbw': {'en': 'Kaiep'},
    'kbx': {'en': 'Ap Ma'},
    'kby': {'en': 'Manga Kanuri'},
    'kbz': {'en': 'Duhwa'},
    'kca': {'en': 'Khanty'},
    'kcb': {'en': 'Kawacha'},
    'kcc': {'en': 'Lubila'},
    'kcd': {'en': 'Ngkâlmpw Kanum'},
    'kce': {'en': 'Kaivi'},
    'kcf': {'en': 'Ukaan'},
    'kcg': {'af': 'Tyap', 'am': 'ታያፕ', 'ar': 'التايابية', 'as': 'ত্যাপ', 'ast': 'tyap', 'az': 'tiyap', 'az-Cyrl': 'тви', 'az-Latn': 'tiyap', 'bal-Latn': 'Tyáp', 'be': 'т’яп', 'be-tarask': 'т’яп', 'bg': 'туап', 'bgn': 'تیاپی', 'bn': 'টিয়াপ', 'br': 'tyap', 'brx': 'त्याप', 'bs': 'tjap', 'bs-Cyrl': 'тјап', 'bs-Latn': 'tjap', 'ca': 'tyap', 'ca-ES-valencia': 'tyap', 'ccp': '𑄑𑄃𑄨𑄠𑄛𑄴', 'ce': 'тьяп', 'chr': 'ᏔᏯᏆ', 'ckb': 'تیاپ', 'cs': 'tyap', 'cy': 'Tyapeg', 'da': 'tyap', 'de': 'Tyap', 'dsb': 'tyapšćina', 'el': 'Τιάπ', 'el-polyton': 'Τιάπ', 'en': 'Tyap', 'es': 'tyap', 'et': 'tjapi', 'eu': 'tyapa', 'fa': 'تیاپی', 'ff-Adlm': '𞤚𞤵𞤴𞤢𞤨𞤵𞥅𞤪𞤫', 'fi': 'tyap', 'fil': 'Tyap', 'fo': 'tyap', 'fr': 'tyap', 'fy': 'Tyap', 'ga': 'Tyap', 'gd': 'Tyap', 'gl': 'tyap', 'gsw': 'Tyap', 'gu': 'ત્યાપ', 'ha': 'Tyap', 'he': 'טיאפ', 'hi': 'त्याप', 'hi-Latn': 'Tyap', 'hr': 'tyap', 'hsb': 'tyapšćina', 'ht': 'tyap', 'hu': 'tyap', 'hy': 'տիապ', 'ia': 'tyap', 'id': 'Tyap', 'ig': 'Tyap', 'is': 'tyap', 'it': 'tyap', 'ja': 'カタブ語', 'jv': 'Tyap', 'ka': 'ტიაპი', 'kcg': 'Katab', 'kgp': 'tyjam', 'kk': 'тьяп тілі', 'kk-Cyrl': 'тьяп тілі', 'km': 'យ៉ាប់', 'kn': 'ಟ್ಯಾಪ್', 'ko': '티얍어', 'kok': 'त्याप', 'kok-Deva': 'त्याप', 'ks': 'تَیَپ', 'ks-Arab': 'تَیَپ', 'ku': 'tyap', 'ky': 'тяпча', 'lb': 'Tyap', 'lo': 'ຕີບ', 'lt': 'tyap', 'lv': 'katabu', 'mai': 'टियाप', 'mi': 'Tiapa', 'mk': 'тјап', 'ml': 'ട്യാപ്', 'mn': 'тяп', 'mr': 'त्याप', 'ms': 'Tyap', 'mt': 'Tyap', 'my': 'တိုင်အပ်', 'nb': 'tyap', 'ne': 'टुआप', 'nl': 'Tyap', 'nn': 'tyap', 'no': 'tyap', 'or': 'ତ୍ୟାପ୍', 'pa': 'ਟਾਇਪ', 'pa-Guru': 'ਟਾਇਪ', 'pcm': 'Tyap Lángwej', 'pl': 'tyap', 'ps': 'تایپ', 'pt': 'tyap', 'qu': 'Tyap Simi', 'rm': 'tyap', 'ro': 'tyap', 'ru': 'тьяп', 'sc': 'tyap', 'sd': 'تياپ', 'sd-Arab': 'تياپ', 'si': 'ට්යප්', 'sk': 'tyap', 'sl': 'tjapska nigerijščina', 'smn': 'tyap', 'so': 'Tyap', 'sq': 'tjapisht', 'sr': 'тјап', 'sr-Cyrl': 'тјап', 'sr-Latn': 'tjap', 'sv': 'tyap', 'sw': 'Kityap', 'ta': 'தையாப்', 'te': 'ట్యాప్', 'th': 'ทีแยป', 'ti': 'ታያፕ', 'tk': 'tiap dili', 'to': 'lea fakatiapi', 'tr': 'Tyap', 'ug': 'تياپچە', 'uk': 'тіап', 'ur': 'تیاپ', 'uz': 'tyap', 'uz-Latn': 'tyap', 'vec': 'tyap', 'vi': 'Tiếng Tyap', 'xnr': 'त्याप', 'yo': 'Èdè Tiyapu', 'yrl': 'tiyapi', 'yue': '卡塔布文', 'yue-Hans': '卡塔布文', 'yue-Hant': '卡塔布文', 'zh': '卡塔布语', 'zh-Hans': '卡塔布语', 'zh-Hant': '卡塔布文', 'zu': 'isi-Tyap'},
    'kch': {'en': 'Vono'},
    'kci': {'en': 'Kamantan'},
    'kcj': {'en': 'Kobiana'},
    'kck': {'en': 'Kalanga'},
    'kcl': {'en': 'Kela (Papua New Guinea)'},
    'kcm': {'en': 'Gula (Central African Republic)', 'kcm': 'Kristang'},
    'kcn': {'en': 'Nubi'},
    'kco': {'en': 'Kinalakna'},
    'kcp': {'en': 'Kanga'},
    'kcq': {'en': 'Kamo'},
    'kcr': {'en': 'Katla'},
    'kcs': {'en': 'Koenoem'},
    'kct': {'en': 'Kaian'},
    'kcu': {'en': 'Kami (Tanzania)'},
    'kcv': {'en': 'Kete'},
    'kcw': {'en': 'Kabwari'},
    'kcx': {'en': 'Kachama-Ganjule'},
    'kcy': {'en': 'Korandje'},
    'kcz': {'en': 'Konongo'},
    'kda': {'en': 'Worimi'},
    'kdc': {'en': 'Kutu'},
    'kdd': {'en': 'Yankunytjatjara'},
    'kde': {'af': 'Makonde', 'am': 'ማኮንዴ', 'ar': 'ماكونده', 'as': 'মাকোণ্ড', 'ast': 'makonde', 'az': 'makond', 'az-Cyrl': 'маконде', 'az-Latn': 'makond', 'bal-Latn': 'Makonde', 'be': 'макондэ', 'be-tarask': 'макондэ', 'bg': 'маконде', 'bgn': 'ماکوندی', 'bn': 'মাকোন্দে', 'br': 'makonde', 'brx': 'माक’ण्डे', 'bs': 'makonde', 'bs-Cyrl': 'маконде', 'bs-Latn': 'makonde', 'ca': 'makonde', 'ca-ES-valencia': 'makonde', 'ccp': '𑄟𑄇𑄮𑄚𑄴𑄘𑄬', 'ce': 'маконде', 'chr': 'ᎹᎪᏕ', 'ckb': 'ماکۆندە', 'cs': 'makonde', 'cy': 'Macondeg', 'da': 'makonde', 'de': 'Makonde', 'dsb': 'makonde', 'el': 'Μακόντε', 'el-polyton': 'Μακόντε', 'en': 'Makonde', 'es': 'makonde', 'et': 'makonde', 'eu': 'makondeera', 'fa': 'ماکونده', 'ff-Adlm': '𞤃𞤢𞤳𞤮𞤲𞤣𞤫𞥅𞤪𞤫', 'fi': 'makonde', 'fil': 'Makonde', 'fo': 'makonde', 'fr': 'makondé', 'fy': 'Makonde', 'ga': 'Makonde', 'gd': 'Makonde', 'gl': 'makonde', 'gu': 'મકોન્ડે', 'ha': 'Makonde', 'he': 'מקונדה', 'hi': 'मैकोंड', 'hi-Latn': 'Makonde', 'hr': 'makonde', 'hsb': 'makonde', 'ht': 'makondé', 'hu': 'makonde', 'hy': 'մակոնդե', 'ia': 'makonde', 'id': 'Makonde', 'ig': 'Makonde', 'is': 'makonde', 'it': 'makonde', 'ja': 'マコンデ語', 'jv': 'Makonde', 'ka': 'მაკონდე', 'kde': 'Chimakonde', 'kea': 'makonde', 'kgp': 'mỹkãne', 'kk': 'маконде тілі', 'kk-Cyrl': 'маконде тілі', 'km': 'ម៉ាកូនដេ', 'kn': 'ಮ್ಯಾಕೊಂಡ್', 'ko': '마콘데어', 'kok': 'माकोंडे', 'kok-Deva': 'माकोंडे', 'ksh': 'Chimakonde', 'ku': 'makondeyî', 'ky': 'макондече', 'lb': 'Makonde', 'lo': 'ມາຄອນເດ', 'lrc': 'ماکوٙندئ', 'lt': 'makondų', 'lv': 'makonde', 'mai': 'मकोंडे', 'mi': 'Makonote', 'mk': 'маконде', 'ml': 'മക്കോണ്ടെ', 'mn': 'маконде', 'mr': 'माकोन्दे', 'ms': 'Makonde', 'mt': 'Makonde', 'my': 'မာခွန်ဒီ', 'mzn': 'ماکونده', 'nb': 'makonde', 'ne': 'माकोन्डे', 'nl': 'Makonde', 'nn': 'makonde', 'no': 'makonde', 'nqo': 'ߡߊߞߐ߲ߘߋߞߊ߲', 'or': 'ମାକୋଣ୍ଡେ', 'pa': 'ਮਕੋਂਡ', 'pa-Guru': 'ਮਕੋਂਡ', 'pcm': 'Makọ́ndẹ́ Lángwej', 'pl': 'makonde', 'ps': 'ميکونډي', 'pt': 'maconde', 'qu': 'Makonde Simi', 'rm': 'makonde', 'ro': 'makonde', 'ru': 'маконде', 'sc': 'makonde', 'sd': 'مڪوندي', 'sd-Arab': 'مڪوندي', 'si': 'මැකොන්ඩ්', 'sk': 'makonde', 'sl': 'makondščina', 'smn': 'makonde', 'so': 'Kimakonde', 'sq': 'makondisht', 'sr': 'маконде', 'sr-Cyrl': 'маконде', 'sr-Latn': 'makonde', 'sv': 'makonde', 'sw': 'Kimakonde', 'ta': 'மகொண்டே', 'te': 'మకొండే', 'th': 'มาคอนเด', 'ti': 'ማኮንደ', 'tk': 'makonde dili', 'to': 'lea fakamakōnite', 'tr': 'Makonde', 'ug': 'ماكوندېچە', 'uk': 'маконде', 'ur': 'ماکونده', 'uz': 'makonde', 'uz-Cyrl': 'макондеча', 'uz-Latn': 'makonde', 'vec': 'makonde', 'vi': 'Tiếng Makonde', 'xnr': 'मैकोंड', 'yo': 'Mákondé', 'yrl': 'makũdi', 'yue': '馬孔德文', 'yue-Hans': '马孔德文', 'yue-Hant': '馬孔德文', 'zh': '马孔德语', 'zh-Hans': '马孔德语', 'zh-Hant': '馬孔德文', 'zu': 'isi-Makonde'},
    'kdf': {'en': 'Mamusi'},
    'kdg': {'en': 'Seba'},
    'kdh': {'en': 'Tem'},
    'kdi': {'en': 'Kumam'},
    'kdj': {'en': 'Karamojong'},
    'kdk': {'en': 'Numèè'},
    'kdl': {'en': 'Tsikimba'},
    'kdm': {'en': 'Kagoma'},
    'kdn': {'en': 'Kunda'},
    'kdo': {'en': 'Kordofanian languages'},
    'kdp': {'en': 'Kaningdon-Nindem'},
    'kdq': {'en': 'Koch'},
    'kdr': {'en': 'Karaim'},
    'kdt': {'en': 'Kuy'},
    'kdu': {'en': 'Kadaru'},
    'kdv': {'en': 'Kado'},
    'kdw': {'en': 'Koneraw'},
    'kdx': {'en': 'Kam'},
    'kdy': {'en': 'Keder'},
    'kdz': {'en': 'Kwaja'},
    'kea': {'af': 'Kabuverdianu', 'ak': 'Kabuvadianu', 'am': 'ካቡቨርዲያኑ', 'ar': 'كابوفيرديانو', 'as': 'কাবুভেৰ্ডিয়ানু', 'ast': 'cabuverdianu', 'az': 'kabuverdian', 'az-Cyrl': 'кабувердиан', 'az-Latn': 'kabuverdian', 'bal-Latn': 'Kubuwerdiánó', 'be': 'кабувердыяну', 'be-tarask': 'кабувердыяну', 'bg': 'кабовердиански', 'bgn': 'کابووردیانو', 'bn': 'কাবুভারদিয়ানু', 'br': 'kabuverdianu', 'brx': 'काबुभेर्डीयानु', 'bs': 'zelenortski', 'bs-Cyrl': 'кабовердијански креолски', 'bs-Latn': 'zelenortski', 'ca': 'crioll capverdià', 'ca-ES-valencia': 'crioll capverdià', 'ccp': '𑄇𑄝𑄪𑄞𑄢𑄴𑄘𑄨𑄠𑄚𑄪', 'ce': 'кабувердьяну', 'chr': 'ᎧᏊᏪᏗᎠᏄ', 'ckb': 'کابووڤێردیانۆ', 'cs': 'kapverdština', 'cy': 'Caboferdianeg', 'da': 'kapverdisk', 'de': 'Kabuverdianu', 'dsb': 'kapverdšćina', 'ee': 'cape verdegbe', 'el': 'Γλώσσα του Πράσινου Ακρωτηρίου', 'el-polyton': 'Γλώσσα του Πράσινου Ακρωτηρίου', 'en': 'Kabuverdianu', 'es': 'criollo caboverdiano', 'et': 'kabuverdianu', 'eu': 'Cabo Verdeko kreolera', 'fa': 'کابووردیانو', 'ff-Adlm': '𞤑𞤢𞤦𞤵𞤾𞤫𞤪𞤣𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'kapverdenkreoli', 'fil': 'Kabuverdianu', 'fo': 'grønhøvdaoyggjarskt', 'fr': 'capverdien', 'fy': 'Kaapverdysk Creools', 'ga': 'Criól Cabo Verde', 'gd': 'Kabuverdianu', 'gl': 'caboverdiano', 'gu': 'કાબુવર્ડિઆનુ', 'ha': 'Kabuverdianu', 'he': 'קאבוורדיאנו', 'hi': 'काबुवेर्दियानु', 'hi-Latn': 'Kabuverdianu', 'hr': 'zelenortski', 'hsb': 'kapverdšćina', 'ht': 'capverdien', 'hu': 'kabuverdianu', 'hy': 'կաբուվերդերեն', 'ia': 'capoverdiano', 'id': 'Kabuverdianu', 'ig': 'Kabụverdịanụ', 'is': 'grænhöfðeyska', 'it': 'capoverdiano', 'ja': 'カーボベルデ・クレオール語', 'jv': 'Kabuverdianu', 'ka': 'კაბუვერდიანუ', 'kea': 'kabuverdianu', 'kgp': 'pu-tánh-vĩ', 'kk': 'кабувердьяну тілі', 'kk-Cyrl': 'кабувердьяну тілі', 'km': 'កាប៊ូវឺឌៀនូ', 'kn': 'ಕಬುವೆರ್ಡಿಯನು', 'ko': '크리올어', 'kok': 'काबुवर्डियनु', 'kok-Deva': 'काबुवर्डियनु', 'ksh': 'Kapvärdesch', 'ku': 'kapverdî', 'ky': 'кабувердиче', 'lb': 'Kabuverdianu', 'lo': 'ຄາເວີເດຍນູ', 'lrc': 'کاباردینو', 'lt': 'Žaliojo Kyšulio kreolų', 'lv': 'kaboverdiešu', 'mai': 'काबुबरदियानु', 'mi': 'Kapuwētianu', 'mk': 'кабувердиану', 'ml': 'കബുവെർദിയാനു', 'mn': 'кабүвердиану', 'mr': 'काबवर्दियानु', 'ms': 'Kabuverdianu', 'mt': 'Cape Verdjan', 'my': 'ကဘူဗာဒီအာနူ', 'mzn': 'کیپ وُردی', 'nb': 'kappverdisk', 'ne': 'काबुभेर्डियानु', 'nl': 'Kaapverdisch Creools', 'nn': 'kappverdisk', 'no': 'kappverdisk', 'nqo': 'ߜߙߋߞߎ߲ߝߙߌߛߌߞߊ߲', 'or': 'କାବୁଭେରଡିଆନୁ', 'pa': 'ਕਾਬੁਵੇਰਦਿਆਨੂ', 'pa-Guru': 'ਕਾਬੁਵੇਰਦਿਆਨੂ', 'pcm': 'Kábúvẹrdiánu Lángwej', 'pl': 'kreolski Wysp Zielonego Przylądka', 'ps': 'کابوورډیانو', 'pt': 'crioulo cabo-verdiano', 'qu': 'Caboverdiano Simi', 'rm': 'cabverdian', 'ro': 'kabuverdianu', 'ru': 'кабувердьяну', 'sc': 'cabubirdianu', 'sd': 'ڪيبيو ويرڊيانو', 'sd-Arab': 'ڪيبيو ويرڊيانو', 'si': 'කබුවෙර්ඩියානු', 'sk': 'kapverdčina', 'sl': 'zelenortskootoška kreolščina', 'smn': 'Kap Verde kreoli', 'so': 'Kabuferdiyanu', 'sq': 'kreolishte e Kepit të Gjelbër', 'sr': 'зеленортски', 'sr-Cyrl': 'зеленортски', 'sr-Latn': 'zelenortski', 'sv': 'kapverdiska', 'sw': 'Kikabuverdianu', 'ta': 'கபுவெர்தியானு', 'te': 'కాబువేర్దియను', 'th': 'คาบูเวอร์เดียนู', 'ti': 'ክርዮል ኬፕ ቨርድኛ', 'tk': 'kabuwerdianu dili', 'to': 'lea fakakapuvelitianu', 'tr': 'Kabuverdianu', 'ug': 'كابۇۋېردىيانچە', 'uk': 'кабувердіану', 'ur': 'کابويرديانو', 'uz': 'kabuverdianu', 'uz-Cyrl': 'кабувердиану', 'uz-Latn': 'kabuverdianu', 'vec': 'caoverdian', 'vi': 'Tiếng Kabuverdianu', 'xnr': 'काबुवेर्दियानु', 'yo': 'Kabufadíánù', 'yrl': 'kiriuru kabu-suikiriwara', 'yue': '卡布威爾第文', 'yue-Hans': '卡布威尔第文', 'yue-Hant': '卡布威爾第文', 'zh': '卡布佛得鲁语', 'zh-Hans': '卡布佛得鲁语', 'zh-Hant': '卡布威爾第文', 'zu': 'isi-Kabuverdianu'},
    'keb': {'en': 'Kélé'},
    'kec': {'en': 'Keiga'},
    'ked': {'en': 'Kerewe'},
    'kee': {'en': 'Eastern Keres'},
    'kef': {'en': 'Kpessi'},
    'keg': {'en': 'Tese'},
    'keh': {'en': 'Keak'},
    'kei': {'en': 'Kei'},
    'kej': {'en': 'Kadar'},
    'kek': {'en': 'Kekchí'},
    'kel': {'en': 'Kela (Democratic Republic of Congo)'},
    'kem': {'en': 'Kemak'},
    'ken': {'ast': 'kenyang', 'bal-Latn': 'Kinyang', 'ca': 'kenyang', 'ca-ES-valencia': 'kenyang', 'cs': 'kenyang', 'de': 'Kenyang', 'en': 'Kenyang', 'fi': 'kenyang', 'fr': 'kényang', 'gd': 'Kenyang', 'hi-Latn': 'Kenyang', 'ht': 'kényang', 'ia': 'kenyang', 'id': 'Kenyang', 'ig': 'Kenyang', 'ja': 'ニャン語', 'ken': 'Kɛnyaŋ', 'ku': 'kenyangî', 'lb': 'Kenyang', 'lt': 'kenyang', 'mk': 'кењанг', 'nb': 'kenyang', 'ne': 'केनयाङ', 'nl': 'Kenyang', 'nn': 'kenyang', 'no': 'kenyang', 'pl': 'kenyang', 'sv': 'kenjang', 'th': 'เกินยาง', 'ti': 'ኬንያንግ', 'to': 'lea fakakeniangi', 'tr': 'Kenyang', 'vec': 'keniang', 'yue': '肯揚文', 'yue-Hans': '肯扬文', 'yue-Hant': '肯揚文', 'zh-Hant': '肯揚文'},
    'keo': {'en': 'Kakwa'},
    'kep': {'en': 'Kaikadi'},
    'keq': {'en': 'Kamar'},
    'ker': {'en': 'Kera'},
    'kes': {'en': 'Kugbo'},
    'ket': {'en': 'Ket'},
    'keu': {'en': 'Akebu'},
    'kev': {'en': 'Kanikkaran'},
    'kew': {'en': 'West Kewa'},
    'kex': {'en': 'Kukna'},
    'key': {'en': 'Kupia'},
    'kez': {'en': 'Kukele'},
    'kfa': {'en': 'Kodava'},
    'kfb': {'en': 'Northwestern Kolami'},
    'kfc': {'en': 'Konda-Dora'},
    'kfd': {'en': 'Korra Koraga'},
    'kfe': {'en': 'Kota (India)'},
    'kff': {'en': 'Koya'},
    'kfg': {'en': 'Kudiya'},
    'kfh': {'en': 'Kurichiya'},
    'kfi': {'en': 'Kannada Kurumba'},
    'kfj': {'en': 'Kemiehua'},
    'kfk': {'en': 'Kinnauri'},
    'kfl': {'en': 'Kung'},
    'kfm': {'en': 'Khunsari'},
    'kfn': {'en': 'Kuk'},
    'kfo': {'af': 'Koro', 'am': 'ኮሮ', 'ar': 'الكورو', 'as': 'কোৰো', 'ast': 'koro', 'az': 'koro', 'az-Cyrl': 'коро', 'az-Latn': 'koro', 'be': 'кора', 'be-tarask': 'кора', 'bg': 'коро', 'bgn': 'کورو', 'bn': 'কোরো', 'br': 'koroeg', 'brx': 'कर’', 'bs': 'koro', 'bs-Cyrl': 'коро', 'bs-Latn': 'koro', 'ca': 'koro', 'ca-ES-valencia': 'koro', 'ccp': '𑄇𑄮𑄢𑄮', 'ce': 'коро', 'chr': 'ᎪᎶ', 'ckb': 'کۆرۆ', 'cs': 'koro', 'cy': 'Koro', 'da': 'koro', 'de': 'Koro', 'dsb': 'korošćina', 'dz': 'ཀོ་རོ་ཁ', 'el': 'Κόρο', 'el-polyton': 'Κόρο', 'en': 'Koro', 'es': 'koro', 'et': 'koro', 'eu': 'koroa', 'fa': 'کورو', 'ff-Adlm': '𞤑𞤮𞤪𞤮𞥅𞤪𞤫', 'fi': 'norsunluurannikonkoro', 'fil': 'Koro', 'fo': 'koro', 'fr': 'koro', 'fy': 'Koro', 'ga': 'Koro', 'gd': 'Koro', 'gl': 'koro', 'gsw': 'Koro', 'gu': 'કોરો', 'ha': 'Koro', 'he': 'קורו', 'hi': 'कोरो', 'hi-Latn': 'Koro', 'hr': 'koro', 'hsb': 'korošćina', 'ht': 'koro', 'hu': 'koro', 'hy': 'կորո', 'ia': 'koro', 'id': 'Koro', 'ig': 'Koro', 'is': 'koro', 'it': 'koro', 'ja': 'コロ語', 'jv': 'Koro', 'ka': 'კორო', 'kgp': 'koro', 'kk': 'коро тілі', 'kk-Cyrl': 'коро тілі', 'km': 'គូរូ', 'kn': 'ಕೋರೋ', 'ko': '코로어', 'kok': 'कोरो', 'kok-Deva': 'कोरो', 'ks': 'کورو', 'ks-Arab': 'کورو', 'ksh': 'Koro vun de Älfebeijnköß', 'ku': 'koro', 'ky': 'корочо', 'lb': 'Koro', 'lo': 'ໂຄໂລ', 'lt': 'koro', 'lv': 'koru', 'mai': 'कोरो', 'mi': 'Koro', 'mk': 'коро', 'ml': 'കോറോ', 'mn': 'коро', 'mr': 'कोरो', 'ms': 'Koro', 'mt': 'Koro', 'my': 'ကိုရို', 'nb': 'koro', 'ne': 'कोरो', 'nl': 'Koro', 'nn': 'koro', 'no': 'koro', 'or': 'କୋରୋ', 'pa': 'ਕੋਰੋ', 'pa-Guru': 'ਕੋਰੋ', 'pcm': 'Koro Lángwej', 'pl': 'koro', 'ps': 'کورو', 'pt': 'koro', 'qu': 'Koro Simi', 'rm': 'koro', 'ro': 'koro', 'ru': 'коро', 'sc': 'koro', 'sd': 'ڪورو', 'sd-Arab': 'ڪورو', 'si': 'කොරො', 'sk': 'koro', 'sl': 'koro', 'smn': 'koro', 'so': 'Koro', 'sq': 'koroisht', 'sr': 'коро', 'sr-Cyrl': 'коро', 'sr-Latn': 'koro', 'sv': 'koro', 'sw': 'Kikoro', 'ta': 'கோரோ', 'te': 'కోరో', 'th': 'โคโร', 'ti': 'ኮሮ', 'tk': 'koro dili', 'to': 'lea fakakolo', 'tr': 'Koro', 'ug': 'كوروچە', 'uk': 'коро', 'ur': 'کورو', 'uz': 'koro', 'uz-Latn': 'koro', 'vec': 'koro', 'vi': 'Tiếng Koro', 'xnr': 'कोरो', 'yo': 'Èdè Koro', 'yrl': 'kuru', 'yue': '科羅文', 'yue-Hans': '科罗文', 'yue-Hant': '科羅文', 'zh': '克罗语', 'zh-Hans': '克罗语', 'zh-Hant': '科羅文', 'zu': 'isi-Koro'},
    'kfp': {'en': 'Korwa'},
    'kfq': {'en': 'Korku'},
    'kfr': {'en': 'Kachhi'},
    'kfs': {'en': 'Bilaspuri'},
    'kft': {'en': 'Kanjari'},
    'kfu': {'en': 'Katkari'},
    'kfv': {'en': 'Kurmukar'},
    'kfw': {'en': 'Kharam Naga'},
    'kfx': {'en': 'Kullu Pahari'},
    'kfy': {'en': 'Kumaoni'},
    'kfz': {'en': 'Koromfé'},
    'kg': {'af': 'Kongolees', 'am': 'ኮንጎኛ', 'ar': 'الكونغو', 'ast': 'kongo', 'az': 'konqo', 'az-Latn': 'konqo', 'bg': 'конгоански', 'bn': 'কঙ্গো', 'br': 'kongo', 'brx': 'कॉंगो', 'bs': 'kongo', 'bs-Cyrl': 'конго', 'bs-Latn': 'kongo', 'ca': 'kongo', 'ca-ES-valencia': 'kongo', 'ccp': '𑄇𑄧𑄁𑄉𑄮', 'cs': 'konžština', 'cy': 'Congo', 'da': 'kongo', 'de': 'Kongolesisch', 'el': 'Κονγκό', 'el-polyton': 'Κονγκό', 'en': 'Kongo', 'es': 'kongo', 'et': 'kongo', 'eu': 'kikongoa', 'fa': 'کنگویی', 'fi': 'kongo', 'fil': 'Kongo', 'fr': 'kikongo', 'fy': 'Kongo', 'ga': 'Congóis', 'gd': 'Kongo', 'gl': 'kongo', 'gsw': 'Kongolesisch', 'gu': 'કોંગો', 'he': 'קונגו', 'hi': 'कोंगो', 'hi-Latn': 'Kongo', 'hr': 'kongo', 'ht': 'kikongo', 'hu': 'kongo', 'id': 'Kongo', 'is': 'kongóska', 'it': 'kongo', 'ja': 'コンゴ語', 'ka': 'კონგო', 'kgp': 'kãgorej', 'kn': 'ಕಾಂಗೋ', 'ko': '콩고어', 'ks': 'کونگو', 'ks-Arab': 'کونگو', 'lb': 'Kongolesesch', 'lo': 'ຄອງໂກ', 'lt': 'Kongo', 'lv': 'kongu', 'mk': 'конго', 'ml': 'കോംഗോ', 'mr': 'काँगो', 'ms': 'Kongo', 'mt': 'Kongo', 'my': 'ကွန်ဂို', 'nb': 'kikongo', 'ne': 'कोङ्गो', 'nl': 'Kongo', 'nn': 'kikongo', 'no': 'kikongo', 'or': 'କଙ୍ଗୋ', 'pl': 'kongo', 'pt': 'congolês', 'rm': 'kongo', 'ro': 'congoleză', 'ru': 'конго', 'sk': 'kongčina', 'sl': 'kongovščina', 'sr': 'конго', 'sr-Cyrl': 'конго', 'sr-Latn': 'kongo', 'sv': 'kikongo', 'sw': 'Kikongo', 'ta': 'காங்கோ', 'te': 'కోంగో', 'th': 'คองโก', 'to': 'lea fakakongikō', 'tr': 'Kongo dili', 'ug': 'كونگوچە', 'uk': 'конґолезька', 'ur': 'کانگو', 'vi': 'Tiếng Kongo', 'yrl': 'kũgurei', 'yue': '剛果文', 'yue-Hans': '刚果文', 'yue-Hant': '剛果文', 'zh': '刚果语', 'zh-Hans': '刚果语', 'zh-Hant': '剛果文', 'zu': 'isi-Kongo'},
    'kga': {'en': 'Koyaga'},
    'kgb': {'en': 'Kawe'},
    'kgc': {'en': 'Kasseng'},
    'kgd': {'en': 'Kataang'},
    'kge': {'en': 'Komering'},
    'kgf': {'en': 'Kube'},
    'kgg': {'en': 'Kusunda'},
    'kgh': {'en': 'Upper Tanudan Kalinga'},
    'kgi': {'en': 'Selangor Sign Language'},
    'kgj': {'en': 'Gamale Kham'},
    'kgk': {'en': 'Kaiwá'},
    'kgl': {'en': 'Kunggari'},
    'kgm': {'en': 'Karipúna'},
    'kgn': {'en': 'Karingani'},
    'kgo': {'en': 'Krongo'},
    'kgp': {'af': 'Kaingang', 'ak': 'Kaingang', 'am': 'ካይንጋንግ', 'ar': 'الكاينغانغ', 'as': 'কেইংগাং', 'ast': 'kaingang', 'az': 'kaiqanq', 'az-Latn': 'kaiqanq', 'bal-Latn': 'Káingáng', 'be': 'каінганг', 'be-tarask': 'каінганг', 'bg': 'кайнганг', 'bn': 'কেইনগ্যাং', 'br': 'kaingangeg', 'brx': 'काईङां', 'bs': 'kaingang', 'bs-Cyrl': 'каинганг', 'bs-Latn': 'kaingang', 'ca': 'kaingà', 'ca-ES-valencia': 'kaingà', 'chr': 'ᎨᏂᎨᏂ', 'cs': 'kaingang', 'cy': 'Kaingang', 'da': 'kaingang', 'de': 'Kaingang', 'dsb': 'kaingangšćina', 'el': 'Κάινγκανγκ', 'el-polyton': 'Κάινγκανγκ', 'en': 'Kaingang', 'es': 'káingang', 'et': 'kaingangi', 'eu': 'kaingangera', 'fa': 'کاین گنگ', 'ff-Adlm': '𞤑𞤢𞤭𞤲𞤺𞤢𞤲𞤺𞤪𞤫', 'fi': 'kaingang', 'fil': 'Kaingang', 'fo': 'kaingang', 'fr': 'caingangue', 'ga': 'Kaingang', 'gd': 'Kaingang', 'gl': 'caingangue', 'gu': 'કૈંગાંગ', 'ha': 'Kaingang', 'he': 'קיינגאנג', 'hi': 'काइंगांग', 'hi-Latn': 'Kaingang', 'hr': 'kaingang', 'hsb': 'kaingangšćina', 'ht': 'caingangue', 'hu': 'kaingang', 'hy': 'կաինգան', 'ia': 'kaingang', 'id': 'Kaingang', 'ig': 'Kaingang', 'is': 'kaingang', 'it': 'kaingang', 'ja': 'カインガング語', 'jv': 'Kaingang', 'ka': 'კაინგანგი', 'kgp': 'kanhgág', 'kk': 'кайнганг тілі', 'kk-Cyrl': 'кайнганг тілі', 'km': 'ខាងហ្កេង', 'kn': 'ಕೈಗಂಗ್', 'ko': '카잉강어', 'kok': 'कैनगँग', 'kok-Deva': 'कैनगँग', 'ku': 'kayingangî', 'ky': 'кайнгангча (индей тили)', 'lb': 'Kaingang', 'lld': 'kaingang', 'lo': 'ເຄນກັງ', 'lt': 'kaingang', 'lv': 'kaingangs', 'mai': 'कैनगैंग', 'mi': 'Keinganga', 'mk': 'каинганшки', 'ml': 'കെയിൻഗാംഗ്', 'mn': 'кайнганг', 'mr': 'काइंगांग', 'ms': 'Kaingang', 'my': 'ကိန်းဂန်', 'nb': 'kaingang', 'ne': 'काइनगाङ', 'nl': 'Kaingang', 'nn': 'kaingang', 'no': 'kaingang', 'nqo': 'ߞߍ߲ߜ߭ߊ߲ߞߊ߲', 'or': 'କାଇଙ୍ଗାଂ', 'pa': 'ਕੈਨਗਾਂਗੋ', 'pa-Guru': 'ਕੈਨਗਾਂਗੋ', 'pcm': 'Kaingáng Lángwej', 'pl': 'kaingang', 'ps': 'کینګا', 'pt': 'caingangue', 'qu': 'Kaingang Simi', 'ro': 'kaingang', 'ru': 'каинганг', 'sc': 'kaingang', 'sd': 'ڪئينگينگ', 'sd-Arab': 'ڪئينگينگ', 'si': 'කයිංගං', 'sk': 'kaingang', 'sl': 'kaingangščina', 'so': 'Kaingang', 'sq': 'kaingangisht', 'sr': 'каинганг', 'sr-Cyrl': 'каинганг', 'sr-Latn': 'kaingang', 'sv': 'kaingang', 'sw': 'Kikaingang', 'ta': 'கைன்கேங்', 'te': 'కైన్\u200cగ్యాంగ్', 'th': 'เคนก่าง', 'ti': 'ካይንጋንግ', 'tk': 'kaýngang dili', 'to': 'lea fakakaingangi', 'tr': 'Kaingang', 'uk': 'кайнґанґ', 'ur': 'کینگینگ', 'uz': 'kaingang', 'uz-Latn': 'kaingang', 'vec': 'kaingang', 'vi': 'Tiếng Kaingang', 'xnr': 'कइंगंग', 'yo': 'Èdè Kaigani', 'yrl': 'kaĩgãgi', 'yue': '坎剛文', 'yue-Hans': '坎刚文', 'yue-Hant': '坎剛文', 'zh': '坎刚语', 'zh-Hans': '坎刚语', 'zh-Hant': '坎剛文', 'zu': 'isi-Kaingang'},
    'kgq': {'en': 'Kamoro'},
    'kgr': {'en': 'Abun'},
    'kgs': {'en': 'Kumbainggar'},
    'kgt': {'en': 'Somyev'},
    'kgu': {'en': 'Kobol'},
    'kgv': {'en': 'Karas'},
    'kgw': {'en': 'Karon Dori'},
    'kgx': {'en': 'Kamaru'},
    'kgy': {'en': 'Kyerung'},
    'kha': {'af': 'Khasi', 'am': 'ካሲ', 'ar': 'الكازية', 'as': 'খাচি', 'ast': 'khasi', 'az': 'xazi', 'az-Cyrl': 'хази', 'az-Latn': 'xazi', 'be': 'кхасі', 'be-tarask': 'кхасі', 'bg': 'кхаси', 'bgn': 'خاسی', 'bn': 'খাশি', 'br': 'khasi', 'brx': 'खासि', 'bs': 'kasi', 'bs-Cyrl': 'каси', 'bs-Latn': 'kasi', 'ca': 'khasi', 'ca-ES-valencia': 'khasi', 'ccp': '𑄈𑄥𑄨', 'ce': 'кхаси', 'chr': 'ᎧᏏ', 'ckb': 'کهاسی', 'cs': 'khásí', 'cy': 'Càseg', 'da': 'khasi', 'de': 'Khasi', 'dsb': 'khasišćina', 'el': 'Κάσι', 'el-polyton': 'Κάσι', 'en': 'Khasi', 'es': 'khasi', 'et': 'khasi', 'eu': 'khasiera', 'fa': 'خاسیایی', 'ff-Adlm': '𞤝𞤢𞤧𞤭𞥅𞤪𞤫', 'fi': 'khasi', 'fil': 'Khasi', 'fo': 'khasi', 'fr': 'khasi', 'fy': 'Khasi', 'ga': 'Caisis', 'gd': 'Khasi', 'gl': 'khasi', 'gsw': 'Khasisch', 'gu': 'ખાસી', 'ha': 'Khasi', 'he': 'קהאסי', 'hi': 'खासी', 'hi-Latn': 'Khasi', 'hr': 'khasi', 'hsb': 'khasišćina', 'ht': 'khasi', 'hu': 'kaszi', 'hy': 'քասիերեն', 'ia': 'khasi', 'id': 'Khasi', 'ig': 'Khasi', 'is': 'kasí', 'it': 'khasi', 'ja': 'カシ語', 'jv': 'Khasi', 'ka': 'ხასი', 'kgp': 'kasi', 'kk': 'кхаси тілі', 'kk-Cyrl': 'кхаси тілі', 'km': 'កាស៊ី', 'kn': 'ಖಾಸಿ', 'ko': '카시어', 'kok': 'खासी', 'kok-Deva': 'खासी', 'ks': 'کھاسی', 'ks-Arab': 'کھاسی', 'ksh': 'Khasi', 'ku': 'khasi', 'ky': 'хасиче', 'lb': 'Khasi-Sprooch', 'lo': 'ຄາສິ', 'lt': 'kasi', 'lv': 'khasu', 'mai': 'खासी', 'mi': 'Kahi', 'mk': 'каси', 'ml': 'ഖാസി', 'mn': 'каси', 'mr': 'खासी', 'ms': 'Khasi', 'mt': 'Khasi', 'my': 'ခါစီ', 'nb': 'khasi', 'ne': 'खासी', 'nl': 'Khasi', 'nn': 'khasi', 'no': 'khasi', 'or': 'ଖାସୀ', 'pa': 'ਖਾਸੀ', 'pa-Guru': 'ਖਾਸੀ', 'pcm': 'Khási Lángwej', 'pl': 'khasi', 'ps': 'خاسې', 'pt': 'khasi', 'qu': 'Khasi Simi', 'rm': 'khasi', 'ro': 'khasi', 'ru': 'кхаси', 'sc': 'khasi', 'sd': 'خاسي', 'sd-Arab': 'خاسي', 'si': 'ඛසි', 'sk': 'khasijčina', 'sl': 'kasi', 'smn': 'khasi', 'so': 'Khasi', 'sq': 'kasisht', 'sr': 'каси', 'sr-Cyrl': 'каси', 'sr-Latn': 'kasi', 'sv': 'khasi', 'sw': 'Kikhasi', 'ta': 'காஸி', 'te': 'ఖాసి', 'th': 'กาสี', 'ti': 'ካሲ', 'tk': 'khasi dili', 'to': 'lea fakakāsi', 'tr': 'Khasi dili', 'ug': 'كاسىچە', 'uk': 'кхасі', 'ur': 'کھاسی', 'uz': 'kxasi', 'uz-Latn': 'kxasi', 'vec': 'khasi', 'vi': 'Tiếng Khasi', 'xnr': 'खासी', 'yo': 'Èdè Kasi', 'yrl': 'kasi', 'yue': '卡西文', 'yue-Hans': '卡西文', 'yue-Hant': '卡西文', 'zh': '卡西语', 'zh-Hans': '卡西语', 'zh-Hant': '卡西文', 'zu': 'isi-Khasi'},
    'khb': {'en': 'Lü'},
    'khc': {'en': 'Tukang Besi North'},
    'khd': {'en': 'Bädi Kanum'},
    'khe': {'en': 'Korowai'},
    'khf': {'en': 'Khuen'},
    'khg': {'en': 'Khams Tibetan'},
    'khh': {'en': 'Kehu'},
    'khi': {'en': 'Khoisan languages'},
    'khi-kun': {'en': 'ǃKung'},
    'khj': {'en': 'Kuturmi'},
    'khk': {'en': 'Halh Mongolian'},
    'khl': {'en': 'Lusi'},
    'khn': {'en': 'Khandesi'},
    'kho': {'ar': 'الخوتانيز', 'ast': 'khotanés', 'az': 'xotan', 'az-Latn': 'xotan', 'bg': 'котски', 'bn': 'খোটানিজ', 'br': 'khotaneg', 'brx': 'ख़ोतानी', 'bs': 'kotanizijski', 'bs-Cyrl': 'котанешки', 'bs-Latn': 'kotanizijski', 'ca': 'khotanès', 'ca-ES-valencia': 'khotanès', 'ccp': '𑄈𑄮𑄑𑄚𑄨𑄎𑄴', 'cs': 'chotánština', 'da': 'khotanesisk', 'de': 'Sakisch', 'el': 'Κοτανικά', 'el-polyton': 'Κοτανικά', 'en': 'Khotanese', 'es': 'kotanés', 'et': 'saka', 'fa': 'ختنی', 'fi': 'khotani', 'fr': 'khotanais', 'fy': 'Khotaneesk', 'gd': 'Cànan Khotan', 'gsw': 'Sakisch', 'gu': 'ખોતાનીસ', 'he': 'קוטאנזית', 'hi': 'खोतानीस', 'hi-Latn': 'Khotanese', 'hr': 'khotanese', 'ht': 'khotanais', 'hu': 'kotanéz', 'id': 'Khotan', 'is': 'kotaska', 'it': 'khotanese', 'ja': 'コータン語', 'kgp': 'kotanẽj', 'kn': 'ಖೋಟಾನೀಸ್', 'ko': '호탄어', 'ks': 'کھوتَنیٖز', 'ks-Arab': 'کھوتَنیٖز', 'lb': 'Sakesch', 'lo': 'ໂຄຕັນ', 'lt': 'kotanezų', 'lv': 'hotaniešu', 'mk': 'хотански', 'ml': 'ഘോറ്റാനേസേ', 'mr': 'खोतानीस', 'mt': 'Kotaniż', 'nb': 'khotanesisk', 'ne': 'खोटानी', 'nl': 'Khotanees', 'nn': 'khotanesisk', 'no': 'khotanesisk', 'or': 'ଖୋତାନୀଜ୍', 'pl': 'chotański', 'pt': 'khotanês', 'rm': 'khotanais', 'ro': 'khotaneză', 'ru': 'хотанский', 'sk': 'chotančina', 'sl': 'kotanščina', 'sr': 'котанешки', 'sr-Cyrl': 'котанешки', 'sr-Latn': 'kotaneški', 'sv': 'khotanesiska', 'ta': 'கோதானீஸ்', 'te': 'ఖోటనీస్', 'th': 'โคตัน', 'to': 'lea fakakōtani', 'tr': 'Hotanca', 'ug': 'خوتەنچە', 'uk': 'хотаносакська', 'vi': 'Tiếng Khotan', 'yrl': 'kutanei', 'yue': '和闐文', 'yue-Hans': '和阗文', 'yue-Hant': '和闐文', 'zh': '和田语', 'zh-Hans': '和田语', 'zh-Hant': '和闐文'},
    'khp': {'en': 'Kapori'},
    'khq': {'af': 'Koyra Chiini', 'am': 'ኮይራ ቺኒ', 'ar': 'كويرا تشيني', 'as': 'কোয়াৰ চিনি', 'ast': 'koyra chiini', 'az': 'koyra çiini', 'az-Cyrl': 'којра чиини', 'az-Latn': 'koyra çiini', 'bal-Latn': 'Koirá Chini', 'be': 'койра чыіні', 'be-tarask': 'койра чыіні', 'bg': 'койра чиини', 'bgn': 'کوجراچینی', 'bn': 'কোয়রা চীনি', 'br': 'koyra chiini', 'brx': 'कईरा चीनि', 'bs': 'kojra čini', 'bs-Cyrl': 'којра чини', 'bs-Latn': 'kojra čini', 'ca': 'koyra chiini', 'ca-ES-valencia': 'koyra chiini', 'ccp': '𑄇𑄮𑄠𑄧𑄢 𑄌𑄩𑄚𑄨', 'ce': 'койра чиини', 'chr': 'ᎪᏱᎳ ᏥᏂ', 'ckb': 'کۆیرا چینی', 'cs': 'koyra chiini', 'cy': 'Koyra Chiini', 'da': 'koyra-chiini', 'de': 'Koyra Chiini', 'dsb': 'koyra chiini', 'el': 'Κόιρα Τσίνι', 'el-polyton': 'Κόιρα Τσίνι', 'en': 'Koyra Chiini', 'es': 'koyra chiini', 'et': 'koyra chiini', 'eu': 'koyra chiinia', 'fa': 'کوجراچینی', 'ff-Adlm': '𞤑𞤮𞤴𞤪𞤢𞤷𞤭𞤲𞤪𞤫', 'fi': 'koyra chiini', 'fil': 'Koyra Chiini', 'fo': 'koyra chiini', 'fr': 'koyra chiini', 'fy': 'Koyra Chiini', 'ga': 'Songais Iartharach', 'gd': 'Koyra Chiini', 'gl': 'koyra chiini', 'gu': 'કોયરા ચિનિ', 'ha': 'Koyra Chiini', 'he': 'קוירה צ׳יני', 'hi': 'कोयरा चीनी', 'hi-Latn': 'Koyra Chiini', 'hr': 'koyra chiini', 'hsb': 'koyra chiini', 'ht': 'koyra chiini', 'hu': 'kojra-csíni', 'hy': 'կոյրա չինի', 'ia': 'koyra chiini', 'id': 'Koyra Chiini', 'ig': 'Koyra Chiini', 'is': 'koyra chiini', 'it': 'koyra chiini', 'ja': 'コイラ・チーニ語', 'jv': 'Koyra Chiini', 'ka': 'კოირა-ჩიინი', 'kea': 'koira txiini', 'kgp': 'kujra sĩnĩ', 'khq': 'Koyra ciini', 'kk': 'койра чини тілі', 'kk-Cyrl': 'койра чини тілі', 'km': 'គុយរ៉ាឈីនី', 'kn': 'ಕೊಯ್ರ ಚೀನಿ', 'ko': '코이라 친니어', 'kok': 'कोयरा छिनी', 'kok-Deva': 'कोयरा छिनी', 'ksh': 'Kojra Tschihni', 'ku': 'koyra chiini', 'ky': 'койра чиниче', 'lb': 'Koyra Chiini', 'lo': 'ຄອຍຣາ ຊິນີ', 'lrc': 'کی یورا چینی', 'lt': 'kojra čini', 'lv': 'koiračiinī', 'mai': 'कोयरा चीनी', 'mi': 'Kōira Tīni', 'mk': 'којра чиини', 'ml': 'കൊയ്റ ചീനി', 'mn': 'койра чини', 'mr': 'कोयरा चीनी', 'ms': 'Koyra Chiini', 'mt': 'Koyra Chiini', 'my': 'ကိုရာ ချီအီနီ', 'mzn': 'کویرا چیینی', 'nb': 'koyra chiini', 'ne': 'कोयरा चिनी', 'nl': 'Koyra Chiini', 'nn': 'koyra chiini', 'no': 'koyra chiini', 'nqo': 'ߞߏߙߌߦߊߗߣߌߞߊ߲', 'or': 'କୋୟରା ଚିନି', 'pa': 'ਕੋਯਰਾ ਚੀਨੀ', 'pa-Guru': 'ਕੋਯਰਾ ਚੀਨੀ', 'pcm': 'Koyra Chíní Lángwej', 'pl': 'koyra chiini', 'ps': 'کویرا چینی', 'pt': 'koyra chiini', 'qu': 'Koyra Chiini Simi', 'rm': 'koyra chiini', 'ro': 'koyra chiini', 'ru': 'койра чиини', 'sc': 'koyra chiini', 'sd': 'ڪيورا چني', 'sd-Arab': 'ڪيورا چني', 'si': 'කොයිරා චිනි', 'sk': 'západná songhajčina', 'sl': 'koyra chiini', 'smn': 'koyra chiini', 'so': 'Koyra Jiini', 'sq': 'kojraçinisht', 'sr': 'којра чиини', 'sr-Cyrl': 'којра чиини', 'sr-Latn': 'kojra čiini', 'sv': 'Timbuktu-songhai', 'sw': 'Kikoyra Chiini', 'ta': 'கொய்ரா சீனீ', 'te': 'కొయరా చీన్నీ', 'th': 'โคย์ราชีนี', 'ti': 'ኮይራ ቺኒ', 'tk': 'koýra-çini dili', 'to': 'lea fakakoila-sīni', 'tr': 'Koyra Chiini', 'ug': 'كويرا چىنىچە', 'uk': 'койра чиїні', 'ur': 'کويرا شيني', 'uz': 'koyra-chiini', 'uz-Cyrl': 'койра-чиини', 'uz-Latn': 'koyra-chiini', 'vec': 'koyra cini', 'vi': 'Tiếng Koyra Chiini', 'xnr': 'कोयरा चीनी', 'yo': 'Koira Ṣíínì', 'yrl': 'kuyira xini', 'yue': '西桑海文', 'yue-Hans': '西桑海文', 'yue-Hant': '西桑海文', 'zh': '西桑海语', 'zh-Hans': '西桑海语', 'zh-Hant': '西桑海文', 'zu': 'isi-Koyra Chiini'},
    'khr': {'en': 'Kharia'},
    'khs': {'en': 'Kasua'},
    'kht': {'en': 'Khamti'},
    'khu': {'en': 'Nkhumbi'},
    'khv': {'en': 'Khvarshi'},
    'khw': {'ast': 'khowar', 'cs': 'chovarština', 'cy': 'Chowareg', 'de': 'Khowar', 'en': 'Khowar', 'et': 'khovari', 'fa': 'کهوار', 'fi': 'khowar', 'fr': 'khowar', 'gd': 'Khowar', 'hi-Latn': 'Khowar', 'ht': 'khowar', 'it': 'khowar', 'ja': 'コワール語', 'ko': '코와르어', 'lb': 'Khowar', 'lt': 'khovarų', 'mk': 'коварски', 'ms': 'Khowar', 'nb': 'khowar', 'ne': 'खोवार', 'nl': 'Khowar', 'nn': 'khowar', 'no': 'khowar', 'pl': 'khowar', 'sv': 'khowar', 'th': 'โควาร์', 'to': 'lea fakakouali', 'tr': 'Çitral Dili', 'yue': '科瓦文', 'yue-Hans': '科瓦文', 'yue-Hant': '科瓦文', 'zh-Hant': '科瓦文'},
    'khx': {'en': 'Kanu'},
    'khy': {'en': 'Kele (Democratic Republic of Congo)'},
    'khz': {'en': 'Keapara'},
    'ki': {'af': 'Kikuyu', 'am': 'ኪኩዩ', 'ar': 'الكيكيو', 'as': 'কিকুয়ু', 'ast': 'kikuyu', 'az': 'kikuyu', 'az-Cyrl': 'кикују', 'az-Latn': 'kikuyu', 'bal-Latn': 'Kikuyu', 'be': 'кікуйю', 'be-tarask': 'кікуйю', 'bg': 'кикую', 'bgn': 'کیکویویی', 'bn': 'কিকুয়ু', 'br': 'kikuyu', 'brx': 'किकुयु', 'bs': 'kikuju', 'bs-Cyrl': 'кикују', 'bs-Latn': 'kikuju', 'ca': 'kikuiu', 'ca-ES-valencia': 'kikuiu', 'ccp': '𑄇𑄨𑄇𑄪𑄠𑄪', 'ce': 'кикуйю', 'chr': 'ᎩᎫᏳ', 'ckb': 'کیکوویوو', 'cs': 'kikujština', 'cy': 'Kikuyu', 'da': 'kikuyu', 'de': 'Kikuyu', 'dsb': 'kikuyu', 'el': 'Κικούγιου', 'el-polyton': 'Κικούγιου', 'en': 'Kikuyu', 'es': 'kikuyu', 'et': 'kikuju', 'eu': 'kikuyuera', 'fa': 'کیکویویی', 'ff-Adlm': '𞤑𞤭𞤳𞤵𞤴𞤵𞥅𞤪𞤫', 'fi': 'kikuju', 'fil': 'Kikuyu', 'fo': 'kikuyu', 'fr': 'kikuyu', 'fy': 'Kikuyu', 'ga': 'Ciocúis', 'gd': 'Kikuyu', 'gl': 'kikuyu', 'gsw': 'Kikuyu-Schpraach', 'gu': 'કિકુયૂ', 'ha': 'Kikuyu', 'he': 'קיקויו', 'hi': 'किकुयू', 'hi-Latn': 'Kikuyu', 'hr': 'kikuyu', 'hsb': 'kikuyu', 'ht': 'kikuyu', 'hu': 'kikuju', 'hy': 'կիկույու', 'ia': 'kikuyu', 'id': 'Kikuyu', 'ig': 'Kikuyu', 'is': 'kíkújú', 'it': 'kikuyu', 'ja': 'キクユ語', 'jv': 'Kikuyu', 'ka': 'კიკუიუ', 'kea': 'kikuiu', 'kgp': 'kikuju', 'ki': 'Gikuyu', 'kk': 'кикуйю тілі', 'kk-Cyrl': 'кикуйю тілі', 'km': 'គីគូយូ', 'kn': 'ಕಿಕುಯು', 'ko': '키쿠유어', 'kok': 'किकुयु', 'kok-Deva': 'किकुयु', 'ks': 'کِکُیوٗ', 'ks-Arab': 'کِکُیوٗ', 'ksh': 'Rekohjo', 'ku': 'kîkûyûyî', 'ky': 'кикуйиче', 'lb': 'Kikuyu-Sprooch', 'lld': 'kikuyu', 'lo': 'ຄິຄູຢຸ', 'lrc': 'کیکیوٙ', 'lt': 'kikujų', 'lv': 'kikuju', 'mai': 'किकुयू', 'mi': 'Kikūiu', 'mk': 'кикују', 'ml': 'കികൂയു', 'mn': 'кикуюү', 'mr': 'किकुयू', 'ms': 'Kikuya', 'mt': 'Kikuju', 'my': 'ကီကူယူ', 'mzn': 'کیکویو', 'nb': 'kikuyu', 'ne': 'किकुयु', 'nl': 'Gikuyu', 'nn': 'kikuyu', 'no': 'kikuyu', 'nqo': 'ߞߌߞߌߦߎߞߊ߲', 'or': 'କୀକୁୟୁ', 'pa': 'ਕਿਕੂਯੂ', 'pa-Guru': 'ਕਿਕੂਯੂ', 'pcm': 'Kikúyú Lángwej', 'pl': 'kikuju', 'ps': 'ککوؤو', 'pt': 'quicuio', 'qu': 'Kikuyu Simi', 'rm': 'kikuyu', 'ro': 'kikuyu', 'ru': 'кикуйю', 'sc': 'kikuyu', 'sd': 'اڪويو', 'sd-Arab': 'اڪويو', 'si': 'කිකුයු', 'sk': 'kikujčina', 'sl': 'kikujščina', 'smn': 'kikujukielâ', 'so': 'Kikuuyu', 'sq': 'kikujuisht', 'sr': 'кикују', 'sr-Cyrl': 'кикују', 'sr-Latn': 'kikuju', 'sv': 'kikuyu', 'sw': 'Kikikuyu', 'ta': 'கிகுயூ', 'te': 'కికుయు', 'th': 'กีกูยู', 'ti': 'ኪኩዩ', 'tk': 'kikuýu dili', 'to': 'lea fakakikuiu', 'tr': 'Kikuyu', 'ug': 'كىكۇيۇچە', 'uk': 'кікуйю', 'ur': 'کیکویو', 'uz': 'kikuyu', 'uz-Cyrl': 'кикую', 'uz-Latn': 'kikuyu', 'vec': 'kikuyu', 'vi': 'Tiếng Kikuyu', 'xnr': 'किकुयू', 'yo': 'Kíkúyù', 'yrl': 'kikuyu', 'yue': '吉庫尤文', 'yue-Hans': '吉库尤文', 'yue-Hant': '吉庫尤文', 'zh': '吉库尤语', 'zh-Hans': '吉库尤语', 'zh-Hant': '吉庫尤文', 'zu': 'isi-Kikuyu'},
    'kia': {'en': 'Kim'},
    'kib': {'en': 'Koalib'},
    'kic': {'en': 'Kickapoo'},
    'kid': {'en': 'Koshin'},
    'kie': {'en': 'Kibet'},
    'kif': {'en': 'Eastern Parbate Kham'},
    'kig': {'en': 'Kimaama'},
    'kih': {'en': 'Kilmeri'},
    'kii': {'en': 'Kitsai'},
    'kij': {'en': 'Kilivila'},
    'kil': {'en': 'Kariya'},
    'kim': {'en': 'Karagas'},
    'kio': {'en': 'Kiowa'},
    'kip': {'en': 'Sheshi Kham'},
    'kiq': {'en': 'Kosadle'},
    'kis': {'en': 'Kis'},
    'kit': {'en': 'Agob'},
    'kiu': {'ast': 'kirmanjki', 'cs': 'zazakština', 'de': 'Kirmanjki', 'en': 'Kirmanjki', 'et': 'kõrmandžki', 'fa': 'کرمانجی', 'fi': 'kirmanjki', 'fr': 'kirmanjki', 'gd': 'Kirmanjki', 'hi-Latn': 'Kirmanjki', 'ht': 'kirmanjki', 'it': 'kirmanjki', 'ja': 'キルマンジュキ語', 'lb': 'Kirmanjki', 'lt': 'kirmanjki', 'mk': 'зазаки', 'nb': 'kirmancki', 'ne': 'किर्मान्जकी', 'nl': 'Kirmanckî', 'nn': 'kirmancki', 'no': 'kirmancki', 'pl': 'kirmandżki', 'sv': 'kirmanjki', 'th': 'เคอร์มานิกิ', 'to': 'lea fakakilimanisikī', 'tr': 'Kırmançça', 'yue': '北紮紮其文', 'yue-Hans': '北扎扎其文', 'yue-Hant': '北紮紮其文', 'zh-Hant': '北紮紮其文'},
    'kiv': {'en': 'Kimbu'},
    'kiw': {'en': 'Northeast Kiwai'},
    'kix': {'en': 'Khiamniungan Naga'},
    'kiy': {'en': 'Kirikiri'},
    'kiz': {'en': 'Kisi'},
    'kj': {'af': 'Kuanyama', 'am': 'ኩዋንያማ', 'ar': 'كوانياما', 'as': 'কুয়ানিয়ামা', 'ast': 'kuanyama', 'az': 'kuanyama', 'az-Cyrl': 'куанјама', 'az-Latn': 'kuanyama', 'be': 'куаньяма', 'be-tarask': 'куаньяма', 'bg': 'кваняма', 'bgn': 'کوانیامایی', 'bn': 'কোয়ানিয়ামা', 'br': 'kwanyama', 'brx': 'कुआनयामा', 'bs': 'kuanjama', 'bs-Cyrl': 'куањама', 'bs-Latn': 'kuanjama', 'ca': 'kuanyama', 'ca-ES-valencia': 'kuanyama', 'ccp': '𑄇𑄮𑄠𑄚𑄨𑄠𑄟', 'ce': 'кунама', 'chr': 'ᎫᏩᏂᎠᎹ', 'ckb': 'کوانیاما', 'cs': 'kuaňamština', 'cy': 'Kuanyama', 'da': 'kuanyama', 'de': 'Kwanyama', 'dsb': 'kuanyamašćina', 'el': 'Κουανιάμα', 'el-polyton': 'Κουανιάμα', 'en': 'Kuanyama', 'es': 'kuanyama', 'et': 'kvanjama', 'eu': 'kuanyama', 'fa': 'کوانیاما', 'ff-Adlm': '𞤑𞤵𞤢𞤻𞤢𞤥𞤢𞥄𞤪𞤫', 'fi': 'kuanjama', 'fil': 'Kuanyama', 'fo': 'kuanyama', 'fr': 'kuanyama', 'fy': 'Kuanyama', 'ga': 'Cuainiáimis', 'gd': 'Kuanyama', 'gl': 'kuanyama', 'gsw': 'Kwanyama', 'gu': 'ક્વાન્યામા', 'ha': 'Kuanyama', 'he': 'קואניאמה', 'hi': 'क्वान्यामा', 'hi-Latn': 'Kuanyama', 'hr': 'kuanyama', 'hsb': 'kuanyamašćina', 'ht': 'kuanyama', 'hu': 'kuanyama', 'hy': 'կուանյամա', 'ia': 'kuanyama', 'id': 'Kuanyama', 'ig': 'Kuanyama', 'is': 'kúanjama', 'it': 'kuanyama', 'ja': 'クワニャマ語', 'jv': 'Kuanyama', 'ka': 'კუნამა', 'kgp': 'kuvanhỹmỹ', 'kk': 'кваньяма тілі', 'kk-Cyrl': 'кваньяма тілі', 'km': 'គូនយ៉ាម៉ា', 'kn': 'ಕ್ವಾನ್\u200cಯಾಮಾ', 'ko': '쿠안야마어', 'kok': 'क्वानयामा', 'kok-Deva': 'क्वानयामा', 'ks': 'کُوانیاما', 'ks-Arab': 'کُوانیاما', 'ksh': 'Oschivambo', 'ku': 'kwanyamayî', 'ky': 'куаньямача', 'lb': 'Kwanyama', 'lo': 'ກວນຍາມາ', 'lt': 'kuaniama', 'lv': 'kvaņamu', 'mai': 'कुअनयामा', 'mi': 'Kuoniāma', 'mk': 'квањама', 'ml': 'ക്വാന്യമ', 'mn': 'куаньяма', 'mr': 'क्वान्यामा', 'ms': 'Kuanyama', 'mt': 'Kuanyama', 'my': 'ကွန်းယာမာ', 'nb': 'kuanyama', 'ne': 'कुआन्यामा', 'nl': 'Kuanyama', 'nn': 'kuanyama', 'no': 'kuanyama', 'or': 'କ୍ୱାନ୍ୟାମ୍', 'pa': 'ਕੁਆਨਯਾਮਾ', 'pa-Guru': 'ਕੁਆਨਯਾਮਾ', 'pcm': 'Kuanyáma Lángwej', 'pl': 'kwanyama', 'ps': 'کواناما', 'pt': 'cuanhama', 'qu': 'Kuanyama Simi', 'rm': 'kuanyama', 'ro': 'kuanyama', 'ru': 'кунама', 'sc': 'kuanyama', 'sd': 'ڪنياما', 'sd-Arab': 'ڪنياما', 'si': 'කුයන්යමා', 'sk': 'kuaňama', 'sl': 'kvanjama', 'smn': 'kuanjama', 'so': 'Kuanyama', 'sq': 'kuanjamaisht', 'sr': 'квањама', 'sr-Cyrl': 'квањама', 'sr-Latn': 'kvanjama', 'sv': 'kuanyama', 'sw': 'Kikwanyama', 'ta': 'குவான்யாமா', 'te': 'క్వాన్యామ', 'th': 'กวนยามา', 'ti': 'ክዋንያማ', 'tk': 'kwanýama dili', 'to': 'lea fakakuaniama', 'tr': 'Kuanyama', 'ug': 'كىۋانياماچە', 'uk': 'кунама', 'ur': 'کونیاما', 'uz': 'kvanyama', 'uz-Latn': 'kvanyama', 'vec': 'kuanyama', 'vi': 'Tiếng Kuanyama', 'xnr': 'क्वान्यामा', 'yo': 'Èdè Kuayama', 'yrl': 'kuãyama', 'yue': '廣亞馬文', 'yue-Hans': '广亚马文', 'yue-Hant': '廣亞馬文', 'zh': '宽亚玛语', 'zh-Hans': '宽亚玛语', 'zh-Hant': '廣亞馬文', 'zu': 'isi-Kuanyama'},
    'kja': {'en': 'Mlap'},
    'kjb': {'en': "Q'anjob'al"},
    'kjc': {'en': 'Coastal Konjo'},
    'kjd': {'en': 'Southern Kiwai'},
    'kje': {'en': 'Kisar'},
    'kjf': {'en': 'Khalaj [Indo-Iranian]'},
    'kjg': {'en': 'Khmu'},
    'kjh': {'en': 'Khakas'},
    'kji': {'en': 'Zabana'},
    'kjj': {'en': 'Khinalugh'},
    'kjk': {'en': 'Highland Konjo'},
    'kjl': {'en': 'Western Parbate Kham'},
    'kjm': {'en': 'Kháng'},
    'kjn': {'en': 'Kunjen'},
    'kjo': {'en': 'Harijan Kinnauri'},
    'kjp': {'en': 'Pwo Eastern Karen'},
    'kjq': {'en': 'Western Keres'},
    'kjr': {'en': 'Kurudu'},
    'kjs': {'en': 'East Kewa'},
    'kjt': {'en': 'Phrae Pwo Karen'},
    'kju': {'en': 'Kashaya'},
    'kjv': {'en': 'Kaikavian Literary Language'},
    'kjx': {'en': 'Ramopa'},
    'kjy': {'en': 'Erave'},
    'kjz': {'en': 'Bumthangkha'},
    'kk': {'af': 'Kazaks', 'ak': 'kasaki kasa', 'am': 'ካዛክኛ', 'ar': 'الكازاخستانية', 'as': 'কাজাখ', 'ast': 'kazaquistanín', 'az': 'qazax', 'az-Arab': 'قازاقجا', 'az-Cyrl': 'газах', 'az-Latn': 'qazax', 'bal-Latn': 'Kázák', 'be': 'казахская', 'be-tarask': 'казахская', 'bg': 'казахски', 'bgn': 'قزاقی', 'bn': 'কাজাখ', 'br': 'kazak', 'brx': 'काजाख', 'bs': 'kazaški', 'bs-Cyrl': 'казашки', 'bs-Latn': 'kazaški', 'ca': 'kazakh', 'ca-ES-valencia': 'kazakh', 'ccp': '𑄇𑄎𑄇𑄴', 'ce': 'кхазакхийн', 'chr': 'ᎧᏌᎧ', 'ckb': 'کازاخی', 'cs': 'kazaština', 'cy': 'Casacheg', 'da': 'kasakhisk', 'de': 'Kasachisch', 'dsb': 'kazachšćina', 'dz': 'ཀ་ཛགས་ཁ', 'ee': 'kazakhstangbe', 'el': 'Καζακικά', 'el-polyton': 'Καζακικά', 'en': 'Kazakh', 'eo': 'kazaĥa', 'es': 'kazajo', 'et': 'kasahhi', 'eu': 'kazakhera', 'fa': 'قزاقی', 'ff-Adlm': '𞤑𞤢𞥁𞤢𞤳𞤪𞤫', 'fi': 'kazakki', 'fil': 'Kazakh', 'fo': 'kazakh', 'fr': 'kazakh', 'fur': 'kazac', 'fy': 'Kazachs', 'ga': 'Casaicis', 'gd': 'Casachais', 'gl': 'kazako', 'gsw': 'Kasachisch', 'gu': 'કઝાખ', 'ha': 'Kazakh', 'he': 'קזחית', 'hi': 'कज़ाख़', 'hi-Latn': 'Kazakh', 'hr': 'kazaški', 'hsb': 'kazachšćina', 'ht': 'kazakh', 'hu': 'kazah', 'hy': 'ղազախերեն', 'ia': 'kazakh', 'id': 'Kazakh', 'ig': 'Kazakh', 'is': 'kasakska', 'it': 'kazako', 'ja': 'カザフ語', 'jv': 'Kazakh', 'ka': 'ყაზახური', 'kea': 'kazak', 'kgp': 'kajake', 'kk': 'қазақ тілі', 'kk-Cyrl': 'қазақ тілі', 'km': 'កាហ្សាក់', 'kn': 'ಕಝಕ್', 'ko': '카자흐어', 'kok': 'कझाख', 'kok-Deva': 'कझाख', 'ks': 'کازَخ', 'ks-Arab': 'کازَخ', 'ksh': 'Kassakkesch', 'ku': 'qazaxî', 'kxv': 'kjaak', 'kxv-Deva': 'कज़ाक़', 'kxv-Latn': 'kjaak', 'kxv-Orya': 'କାଜାକ୍', 'kxv-Telu': 'కజక్', 'ky': 'казакча', 'lb': 'Kasachesch', 'lkt': 'Kazakh Iyápi', 'lld': 'kazakh', 'lo': 'ຄາຊັກ', 'lrc': 'قأزاق', 'lt': 'kazachų', 'lv': 'kazahu', 'mai': 'कजाक', 'mi': 'Kahāka', 'mk': 'казашки', 'ml': 'കസാഖ്', 'mn': 'казах', 'mr': 'कझाक', 'ms': 'Kazakhstan', 'mt': 'Każak', 'my': 'ကာဇာချ', 'mzn': 'قزاقی', 'nb': 'kasakhisk', 'ne': 'काजाख', 'nl': 'Kazachs', 'nn': 'kasakhisk', 'no': 'kasakhisk', 'nqo': 'ߞߖ߭ߊߞߌߞߊ߲', 'or': 'କାଜାଖ୍\u200c', 'pa': 'ਕਜ਼ਾਖ਼', 'pa-Guru': 'ਕਜ਼ਾਖ਼', 'pcm': 'Kazák Lángwej', 'pl': 'kazachski', 'ps': 'قازق', 'pt': 'cazaque', 'qu': 'Kazajo Simi', 'rm': 'casac', 'ro': 'kazahă', 'ru': 'казахский', 'sah': 'Хаһаахтыы', 'sc': 'kazacu', 'sd': 'قازق', 'sd-Arab': 'قازق', 'se': 'kazakgiella', 'si': 'කසාඛ්', 'sk': 'kazaština', 'sl': 'kazaščina', 'smn': 'kazakkielâ', 'so': 'Kasaaq', 'sq': 'kazakisht', 'sr': 'казашки', 'sr-Cyrl': 'казашки', 'sr-Latn': 'kazaški', 'sv': 'kazakiska', 'sw': 'Kikazakh', 'ta': 'கசாக்', 'te': 'కజఖ్', 'tg': 'қазоқӣ', 'th': 'คาซัค', 'ti': 'ካዛክ', 'tk': 'gazak dili', 'to': 'lea fakakasaki', 'tr': 'Kazakça', 'tt': 'казакъ', 'ug': 'قازاقچە', 'uk': 'казахська', 'ur': 'قزاخ', 'uz': 'qozoqcha', 'uz-Cyrl': 'қозоқча', 'uz-Latn': 'qozoqcha', 'vec': 'kazako', 'vi': 'Tiếng Kazakh', 'wae': 'Kazačiš', 'wo': 'Kasax', 'xnr': 'कज़ाख़', 'yi': 'קאַזאַכיש', 'yo': 'Kaṣakì', 'yrl': 'kasaki', 'yue': '哈薩克文', 'yue-Hans': '哈萨克文', 'yue-Hant': '哈薩克文', 'zh': '哈萨克语', 'zh-Hans': '哈萨克语', 'zh-Hant': '哈薩克文', 'zu': 'isi-Kazakh'},
    'kk-Arab': {'kk-Arab': 'قازاق ءتىلى (توتە)'},
    'kka': {'en': 'Kakanda'},
    'kkb': {'en': 'Kwerisa'},
    'kkc': {'en': 'Odoodee'},
    'kkd': {'en': 'Kinuku'},
    'kke': {'en': 'Kakabe'},
    'kkf': {'en': 'Kalaktang Monpa'},
    'kkg': {'en': 'Mabaka Valley Kalinga'},
    'kkh': {'en': 'Khün'},
    'kki': {'en': 'Kagulu'},
    'kkj': {'af': 'Kako', 'am': 'ካኮ', 'ar': 'لغة الكاكو', 'as': 'কাকো', 'ast': 'kako', 'az': 'kako', 'az-Cyrl': 'како', 'az-Latn': 'kako', 'bal-Latn': 'Kákó', 'be': 'како', 'be-tarask': 'како', 'bg': 'како', 'bgn': 'کاکویی', 'bn': 'কাকো', 'br': 'kakoeg', 'brx': 'काक’', 'bs': 'kako', 'bs-Cyrl': 'како', 'bs-Latn': 'kako', 'ca': 'kako', 'ca-ES-valencia': 'kako', 'ccp': '𑄇𑄇𑄮', 'ce': 'како', 'chr': 'ᎧᎪ', 'ckb': 'کاکۆ', 'cs': 'kako', 'cy': 'Kako', 'da': 'kako', 'de': 'Kako', 'dsb': 'kako', 'el': 'Κάκο', 'el-polyton': 'Κάκο', 'en': 'Kako', 'es': 'kako', 'et': 'kako', 'eu': 'kakoa', 'fa': 'کاکایی', 'ff-Adlm': '𞤑𞤢𞤳𞤮𞥅𞤪𞤫', 'fi': 'kako', 'fil': 'Kako', 'fo': 'kako', 'fr': 'kako', 'fy': 'Kako', 'ga': 'Cacóis', 'gd': 'Kako', 'gl': 'kako', 'gu': 'કાકો', 'ha': 'Kako', 'he': 'קאקו', 'hi': 'काको', 'hi-Latn': 'Kako', 'hr': 'kako', 'hsb': 'kako', 'ht': 'kako', 'hu': 'kakó', 'hy': 'կակո', 'ia': 'kako', 'id': 'Kako', 'ig': 'Kako', 'is': 'kako', 'it': 'kako', 'ja': 'カコ語', 'jv': 'Kako', 'ka': 'კაკო', 'kea': 'kako', 'kgp': 'kako', 'kk': 'како тілі', 'kk-Cyrl': 'како тілі', 'kkj': 'kakɔ', 'km': 'កាកូ', 'kn': 'ಕಾಕೊ', 'ko': '카코어', 'kok': 'काको', 'kok-Deva': 'काको', 'ksh': 'Kako', 'ku': 'kako', 'ky': 'какочо', 'lb': 'Kako', 'lo': 'ຄາໂກ', 'lt': 'kako', 'lv': 'kako', 'mai': 'काको', 'mi': 'Kako', 'mk': 'како', 'ml': 'കാകോ', 'mn': 'како', 'mr': 'काको', 'ms': 'Kako', 'mt': 'Kako', 'my': 'ကကို', 'nb': 'kako', 'ne': 'काको', 'nl': 'Kako', 'nn': 'kako', 'nnh': 'Shwóŋò pʉa shÿó Bɛgtùa', 'no': 'kako', 'nqo': 'ߞߊߞߏߞߊ߲', 'or': 'କାକୋ', 'pa': 'ਕਾਕੋ', 'pa-Guru': 'ਕਾਕੋ', 'pcm': 'Kákó Lángwej', 'pl': 'kako', 'ps': 'کاکو', 'pt': 'kako', 'qu': 'Kako Simi', 'rm': 'kako', 'ro': 'kako', 'ru': 'како', 'sc': 'kako', 'sd': 'ڪڪو', 'sd-Arab': 'ڪڪو', 'si': 'කකො', 'sk': 'kako', 'sl': 'kako', 'smn': 'kako', 'so': 'Kaako', 'sq': 'kakoisht', 'sr': 'како', 'sr-Cyrl': 'како', 'sr-Latn': 'kako', 'sv': 'mkako', 'sw': 'Lugha ya Kako', 'ta': 'ககோ', 'te': 'కాకో', 'th': 'คาโก', 'ti': 'ካኮ', 'tk': 'kako dili', 'to': 'lea fakakako', 'tr': 'Kako', 'ug': 'كاكوچە', 'uk': 'како', 'ur': 'کاکو', 'uz': 'kako', 'uz-Cyrl': 'како', 'uz-Latn': 'kako', 'vec': 'kako', 'vi': 'Tiếng Kako', 'xnr': 'काको', 'yo': 'Kàkó', 'yrl': 'kaku', 'yue': '卡庫文', 'yue-Hans': '卡库文', 'yue-Hant': '卡庫文', 'zh': '卡库语', 'zh-Hans': '卡库语', 'zh-Hant': '卡庫文', 'zu': 'isi-Kako'},
    'kkk': {'en': 'Kokota'},
    'kkl': {'en': 'Kosarek Yale'},
    'kkm': {'en': 'Kiong'},
    'kkn': {'en': 'Kon Keu'},
    'kko': {'en': 'Karko'},
    'kkp': {'en': 'Gugubera'},
    'kkq': {'en': 'Kaeku'},
    'kkr': {'en': 'Kir-Balar'},
    'kks': {'en': 'Giiwo'},
    'kkt': {'en': 'Koi'},
    'kku': {'en': 'Tumi'},
    'kkv': {'en': 'Kangean'},
    'kkw': {'en': 'Teke-Kukuya'},
    'kkx': {'en': 'Kohin'},
    'kky': {'en': 'Guugu Yimidhirr'},
    'kkz': {'en': 'Kaska'},
    'kl': {'af': 'Kalaallisut', 'am': 'ካላሊሱት', 'ar': 'الكالاليست', 'as': 'কালালিছুট', 'ast': 'kalaallisut', 'az': 'kalaallisut', 'az-Cyrl': 'калааллисут', 'az-Latn': 'kalaallisut', 'bal-Latn': 'Kalállisut', 'be': 'грэнландская', 'be-tarask': 'грэнландская', 'bg': 'гренландски', 'bgn': 'گرینلندی', 'bn': 'কালাল্লিসুট', 'br': 'greunlandeg', 'brx': 'कलालीसुत', 'bs': 'kalalisutski', 'bs-Cyrl': 'калалисут', 'bs-Latn': 'kalalisutski', 'ca': 'groenlandès', 'ca-ES-valencia': 'groenlandès', 'ccp': '𑄇𑄳𑄠𑄣𑄣𑄴𑄣𑄨𑄥𑄪𑄖𑄴', 'ce': 'гренландхойн', 'chr': 'ᎧᎳᎵᏑᏘ', 'ckb': 'کالالیسووت', 'cs': 'grónština', 'cy': 'Kalaallisut', 'da': 'grønlandsk', 'de': 'Grönländisch', 'dsb': 'grönlandšćina', 'el': 'Καλαάλισουτ', 'el-polyton': 'Καλαάλισουτ', 'en': 'Kalaallisut', 'eo': 'gronlanda', 'es': 'groenlandés', 'et': 'grööni', 'eu': 'groenlandiera', 'fa': 'گرینلندی', 'ff-Adlm': '𞤑𞤢𞤤𞤢𞥄𞤤𞤧𞤵𞤼𞤪𞤫', 'fi': 'kalaallisut', 'fil': 'Kalaallisut', 'fo': 'kalaallisut', 'fr': 'groenlandais', 'fur': 'kalaallisut', 'fy': 'Grienlâns', 'ga': 'Kalaallisut', 'gd': 'Kalaallisut', 'gl': 'kalaallisut', 'gsw': 'Gröönländisch', 'gu': 'કલાલ્લિસુત', 'ha': 'Kalaallisut', 'he': 'גרינלנדית', 'hi': 'कलालीसुत', 'hi-Latn': 'Kalaallisut', 'hr': 'kalaallisut', 'hsb': 'gröndlandšćina', 'ht': 'groenlandais', 'hu': 'grönlandi', 'hy': 'կալաալիսուտ', 'ia': 'groenlandese', 'id': 'Kalaallisut', 'ig': 'Kalaallisut', 'is': 'grænlenska', 'it': 'groenlandese', 'ja': 'グリーンランド語', 'jv': 'Kalaallisut', 'ka': 'დასავლეთ გრენლანდიური', 'kea': 'groenlandes', 'kgp': 'grohẽrỹnej', 'kk': 'калаалисут тілі', 'kk-Cyrl': 'калаалисут тілі', 'kl': 'kalaallisut', 'km': 'កាឡាលលីស៊ុត', 'kn': 'ಕಲಾಲ್ಲಿಸುಟ್', 'ko': '그린란드어', 'kok': 'कालाल्लिसुट', 'kok-Deva': 'कालाल्लिसुट', 'ks': 'کَلالِسُت', 'ks-Arab': 'کَلالِسُت', 'ksh': 'Jröhnländesch', 'ku': 'kalalîsûtî', 'ky': 'калаалисутча', 'lb': 'Grönlännesch', 'lld': 'groenlandesc', 'lo': 'ກຣີນແລນລິດ', 'lrc': 'کالالیسوٙت', 'lt': 'kalalisut', 'lv': 'grenlandiešu', 'mai': 'कलालिसुत', 'mi': 'Kararīhutu', 'mk': 'калалисут', 'ml': 'കലാല്ലിസുട്ട്', 'mn': 'калалисут', 'mr': 'कलाल्लिसत', 'ms': 'Kalaallisut', 'mt': 'Kalallisut', 'my': 'ကလာအ်လီဆပ်', 'mzn': 'کالائلیسوت', 'nb': 'grønlandsk', 'ne': 'कालालिसुट', 'nl': 'Groenlands', 'nn': 'grønlandsk (kalaallisut)', 'no': 'grønlandsk', 'nqo': 'ߜ߭ߙߏߟߊ߲ߘߌߞߊ߲', 'or': 'କାଲାଲିସୁଟ୍', 'pa': 'ਕਲਾਅੱਲੀਸੁਟ', 'pa-Guru': 'ਕਲਾਅੱਲੀਸੁਟ', 'pcm': 'Kalálísút Lángwej', 'pl': 'grenlandzki', 'ps': 'کالالیست', 'pt': 'groenlandês', 'qu': 'Groenlandes Simi', 'rm': 'grönlandais', 'ro': 'kalaallisut', 'ru': 'гренландский', 'sc': 'groenlandesu', 'sd': 'ڪالا ليسٽ', 'sd-Arab': 'ڪالا ليسٽ', 'si': 'කලාලිසට්', 'sk': 'grónčina', 'sl': 'grenlandščina', 'smn': 'kalaallisut', 'so': 'Kalaallisuut', 'sq': 'kalalisutisht', 'sr': 'гренландски', 'sr-Cyrl': 'гренландски', 'sr-Latn': 'grenlandski', 'sv': 'grönländska', 'sw': 'Kikalaallisut', 'ta': 'கலாலிசூட்', 'te': 'కలాల్లిసూట్', 'th': 'กรีนแลนด์', 'ti': 'ግሪንላንድኛ', 'tk': 'grenland dili', 'to': 'lea fakakalaʻalisuti', 'tr': 'Grönland dili', 'ug': 'گىرېنلاندچە', 'uk': 'калааллісут', 'ur': 'کالاليست', 'uz': 'grenland', 'uz-Cyrl': 'гренландча', 'uz-Latn': 'grenland', 'vec': 'groelandeze', 'vi': 'Tiếng Kalaallisut', 'xnr': 'कलालीसुत', 'yo': 'Kalaalísùtì', 'yrl': 'guruẽrãdei', 'yue': '格陵蘭文', 'yue-Hans': '格陵兰文', 'yue-Hant': '格陵蘭文', 'zh': '格陵兰语', 'zh-Hans': '格陵兰语', 'zh-Hant': '格陵蘭文', 'zu': 'isi-Kalaallisut'},
    'kla': {'en': 'Klamath-Modoc'},
    'klb': {'en': 'Kiliwa'},
    'klc': {'en': 'Kolbila'},
    'kld': {'en': 'Gamilaraay'},
    'kle': {'en': 'Kulung (Nepal)'},
    'klf': {'en': 'Kendeje'},
    'klg': {'en': 'Tagakaulo'},
    'klh': {'en': 'Weliki'},
    'kli': {'en': 'Kalumpang'},
    'klj': {'en': 'Khalaj'},
    'klk': {'en': 'Kono (Nigeria)'},
    'kll': {'en': 'Kagan Kalagan'},
    'klm': {'en': 'Migum'},
    'kln': {'af': 'Kalenjin', 'am': 'ካለንጂን', 'ar': 'كالينجين', 'as': 'কালেনজিন', 'ast': 'kalenjin', 'az': 'kalencin', 'az-Cyrl': 'каленҹин', 'az-Latn': 'kalencin', 'bal-Latn': 'Kalenjin', 'be': 'календжын', 'be-tarask': 'календжын', 'bg': 'календжин', 'bgn': 'کالنجین', 'bn': 'কালেনজিন', 'br': 'kalendjineg', 'brx': 'कालेनजिन', 'bs': 'kalenjin', 'bs-Cyrl': 'калењин', 'bs-Latn': 'kalenjin', 'ca': 'kalenjin', 'ca-ES-valencia': 'kalenjin', 'ccp': '𑄇𑄣𑄬𑄚𑄴𑄎𑄨𑄚𑄴', 'ce': 'календжин', 'chr': 'ᎧᎴᏂᏥᏂ', 'ckb': 'کالێنجین', 'cs': 'kalendžin', 'cy': 'Kalenjin', 'da': 'kalenjin', 'de': 'Kalenjin', 'dsb': 'kalenjin', 'el': 'Καλεντζίν', 'el-polyton': 'Καλεντζίν', 'en': 'Kalenjin', 'es': 'kalenjin', 'et': 'kalendžini', 'eu': 'kalenjinera', 'fa': 'کالنجین', 'ff-Adlm': '𞤑𞤢𞤤𞤫𞤲𞤶𞤭𞤲𞤪𞤫', 'fi': 'kalenjin', 'fil': 'Kalenjin', 'fo': 'kalenjin', 'fr': 'kalendjin', 'fy': 'Kalenjin', 'ga': 'Kalenjin', 'gd': 'Kalenjin', 'gl': 'kalenjin', 'gu': 'કલેજિન', 'ha': 'Kalenjin', 'he': 'קלנג׳ין', 'hi': 'कलेंजिन', 'hi-Latn': 'Kalenjin', 'hr': 'kalenjin', 'hsb': 'kalenjin', 'ht': 'kalendjin', 'hu': 'kalendzsin', 'hy': 'կալենջին', 'ia': 'kalenjin', 'id': 'Kalenjin', 'ig': 'Kalenjin', 'is': 'kalenjin', 'it': 'kalenjin', 'ja': 'カレンジン語', 'jv': 'Kalenjin', 'ka': 'კალენჯინი', 'kea': 'kalenjin', 'kgp': 'karẽnjĩn', 'kk': 'каленжин тілі', 'kk-Cyrl': 'каленжин тілі', 'kln': 'Kalenjin', 'km': 'កាលែនជីន', 'kn': 'ಕಲೆಂಜಿನ್', 'ko': '칼렌진어', 'kok': 'कालेंजीन', 'kok-Deva': 'कालेंजीन', 'ksh': 'Kaländjihnesch', 'ku': 'kalencînî', 'ky': 'каленжиче', 'lb': 'Kalenjin', 'lo': 'ຄາເລັນຈິນ', 'lrc': 'کالئجین', 'lt': 'kalenjinų', 'lv': 'kalendžīnu', 'mai': 'केलैंजिन', 'mi': 'Karenini', 'mk': 'каленџин', 'ml': 'കലെഞ്ചിൻ', 'mn': 'каленжин', 'mr': 'कालेंजीन', 'ms': 'Kalenjin', 'mt': 'Kalenjin', 'my': 'ကလန်ဂျင်', 'mzn': 'کالنجین', 'nb': 'kalenjin', 'ne': 'कालेन्जिन', 'nl': 'Kalenjin', 'nn': 'kalenjin', 'no': 'kalenjin', 'nqo': 'ߞߊߟߊ߲ߖߌߞߊ߲', 'or': 'କାଲେନଜିନ୍', 'pa': 'ਕਲੇਜਿਨ', 'pa-Guru': 'ਕਲੇਜਿਨ', 'pcm': 'Kálẹ́njín Lángwej', 'pl': 'kalenjin', 'ps': 'کلینجن', 'pt': 'kalenjin', 'qu': 'Kalenjin Simi', 'rm': 'kalenjin', 'ro': 'kalenjin', 'ru': 'календжин', 'sc': 'kalenjin', 'sd': 'ڪيلين جن', 'sd-Arab': 'ڪيلين جن', 'si': 'කලෙන්ජන්', 'sk': 'kalendžin', 'sl': 'kalenjinščina', 'smn': 'kalenjikielâ', 'so': 'Kalenjin', 'sq': 'kalenxhinisht', 'sr': 'каленџински', 'sr-Cyrl': 'каленџински', 'sr-Latn': 'kalendžinski', 'sv': 'kalenjin', 'sw': 'Kikalenjin', 'ta': 'கலின்ஜின்', 'te': 'కలెంజిన్', 'th': 'คาเลนจิน', 'ti': 'ካለንጂን', 'tk': 'kalenjin dili', 'to': 'lea fakakalenisini', 'tr': 'Kalenjin', 'ug': 'كالېنجىنچە', 'uk': 'календжин', 'ur': 'کالينجين', 'uz': 'kalenjin', 'uz-Cyrl': 'каленжинча', 'uz-Latn': 'kalenjin', 'vec': 'kalenjin', 'vi': 'Tiếng Kalenjin', 'xnr': 'कलेंजिन', 'yo': 'Kálẹnjín', 'yrl': 'karẽyĩ', 'yue': '卡倫金文', 'yue-Hans': '卡伦金文', 'yue-Hant': '卡倫金文', 'zh': '卡伦金语', 'zh-Hans': '卡伦金语', 'zh-Hant': '卡倫金文', 'zu': 'isi-Kalenjin'},
    'klo': {'en': 'Kapya'},
    'klp': {'en': 'Kamasa'},
    'klq': {'en': 'Rumu'},
    'klr': {'en': 'Khaling'},
    'kls': {'en': 'Kalasha'},
    'klt': {'en': 'Nukna'},
    'klu': {'en': 'Klao'},
    'klv': {'en': 'Maskelynes'},
    'klw': {'en': 'Tado'},
    'klx': {'en': 'Koluwawa'},
    'kly': {'en': 'Kalao'},
    'klz': {'en': 'Kabola'},
    'km': {'af': 'Khmer', 'agq': 'Kɨmɛ̀', 'ak': 'Kambodia kasa', 'am': 'ክመር', 'ar': 'الخميرية', 'as': 'খমেৰ', 'asa': 'Kikambodia', 'ast': 'ḥemer', 'az': 'kxmer', 'az-Arab': 'خمئرجه', 'az-Cyrl': 'кхмер', 'az-Latn': 'kxmer', 'bal-Latn': 'Khmér', 'bas': 'Hɔp u kmɛ̂r', 'be': 'кхмерская', 'be-tarask': 'кхмерская', 'bem': 'Ichi Khmer', 'bez': 'Hikambodia', 'bg': 'кхмерски', 'bgn': 'خمری', 'bm': 'kambojikan', 'bn': 'খমের', 'br': 'khmer', 'brx': 'ख्मेर', 'bs': 'kmerski', 'bs-Cyrl': 'кмерски', 'bs-Latn': 'kmerski', 'ca': 'khmer', 'ca-ES-valencia': 'khmer', 'ccp': '𑄈𑄧𑄟𑄬𑄢𑄴', 'ce': 'кхмерийн', 'cgg': 'Orukambodiya', 'chr': 'ᎩᎻᎷ', 'ckb': 'خمێر', 'cs': 'khmérština', 'cy': 'Chmereg', 'da': 'khmer', 'dav': 'Kikambodia', 'de': 'Khmer', 'dje': 'Kmeer senni', 'dsb': 'kambodžanšćina', 'dyo': 'kmeer', 'dz': 'ཁེ་མེར་ཁ', 'ebu': 'Kĩkambodia', 'ee': 'khmergbe', 'el': 'Χμερ', 'el-polyton': 'Χμερ', 'en': 'Khmer', 'en-Dsrt': '𐐗𐐲𐑋𐐯𐑉', 'eo': 'kmera', 'es': 'jemer', 'et': 'khmeeri', 'eu': 'khemerera', 'ewo': 'Ǹkɔ́bɔ kəmɛ́r', 'fa': 'خمری', 'ff': 'Kemeere', 'ff-Adlm': '𞤑𞤵𞤥𞤢𞤴𞤪𞤫', 'ff-Latn': 'Kemeere', 'fi': 'khmer', 'fil': 'Khmer', 'fo': 'khmer', 'fr': 'khmer', 'fur': 'khmer', 'fy': 'Khmer', 'ga': 'Ciméiris', 'gd': 'Cmèar', 'gl': 'khmer', 'gsw': 'Kambodschanisch', 'gu': 'ખ્મેર', 'guz': 'Kikambodia', 'ha': 'Harshen Kimar', 'he': 'חמרית', 'hi': 'खमेर', 'hi-Latn': 'Khmer', 'hr': 'kmerski', 'hsb': 'khmeršćina', 'ht': 'khmer', 'hu': 'khmer', 'hy': 'քմերերեն', 'ia': 'khmer', 'id': 'Khmer', 'ig': 'Khmer', 'is': 'kmer', 'it': 'khmer', 'ja': 'クメール語', 'jmc': 'Kyikambodia', 'jv': 'Khmer', 'ka': 'ქმერული', 'kab': 'Takemrit', 'kam': 'Kikambodia', 'kde': 'Chikambodia', 'kea': 'kmer', 'kgp': 'kymẽr', 'khq': 'Kmeer senni, Game here', 'ki': 'Kikambodia', 'kk': 'кхмер тілі', 'kk-Cyrl': 'кхмер тілі', 'kln': 'kutitab Kher nebo Kwen', 'km': 'ខ្មែរ', 'kn': 'ಖಮೇರ್', 'ko': '크메르어', 'kok': 'कंबोडिया', 'kok-Deva': 'कंबोडिया', 'ks': 'خَمیر', 'ks-Arab': 'خَمیر', 'ksb': 'Kikambodia', 'ksf': 'rikmɛr', 'ksh': 'Khmer', 'ku': 'ximêrî', 'kxv': 'kmer', 'kxv-Deva': 'कमेर', 'kxv-Latn': 'kmer', 'kxv-Orya': 'କାମେର୍', 'kxv-Telu': 'కమెర్', 'ky': 'кмерче', 'lag': 'Kɨkambódia', 'lb': 'Kambodschanesch', 'lg': 'Lukme', 'lkt': 'Khmer Iyápi', 'lld': 'khmer', 'ln': 'likambodza', 'lo': 'ຂະເໝນ', 'lrc': 'خئمئر', 'lt': 'khmerų', 'luo': 'Kikambodia', 'luy': 'Kikambodia', 'lv': 'khmeru', 'mai': 'खमेर', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lkambodia', 'mer': 'Kĩkambodia', 'mfe': 'khmer, santral', 'mg': 'khmer', 'mgh': 'Ikambodia', 'mi': 'Kimēra', 'mk': 'кмерски', 'ml': 'ഖമെർ', 'mn': 'кхмер', 'mr': 'ख्मेर', 'ms': 'Khmer', 'mt': 'Khmer', 'mua': 'kmer', 'my': 'ခမာ', 'mzn': 'خمری', 'naq': 'Khmerǁî gowab, Central', 'nb': 'khmer', 'nd': 'isi-Khambodiya', 'ne': 'खमेर', 'nl': 'Khmer', 'nmg': 'Kiɛl bó mɛr', 'nn': 'khmer', 'no': 'khmer', 'nqo': 'ߞߑߡߍߙߑߞߊ߲', 'nus': 'Thok kameeri', 'nyn': 'Orukambodiya', 'or': 'ଖାମେର୍', 'pa': 'ਖਮੇਰ', 'pa-Guru': 'ਖਮੇਰ', 'pcm': 'Kmaí Lángwej', 'pl': 'khmerski', 'ps': 'خمر', 'pt': 'khmer', 'qu': 'Khmer Simi', 'rm': 'cambodschan', 'rn': 'Igikambodiya', 'ro': 'khmeră', 'rof': 'Kikambodia', 'ru': 'кхмерский', 'rw': 'Igikambodiya', 'rwk': 'Kyikambodia', 'saq': 'Kikambodia', 'sbp': 'Ishikambodia', 'sc': 'khmer', 'sd': 'خمر', 'sd-Arab': 'خمر', 'se': 'kambodiagiella', 'seh': 'cmer', 'ses': 'Kmeer senni', 'sg': 'Kmêre', 'shi': 'ⵜⴰⵅⵎⵉⵔⵜ', 'shi-Latn': 'Taxmirt', 'shi-Tfng': 'ⵜⴰⵅⵎⵉⵔⵜ', 'si': 'කමර්', 'sk': 'khmérčina', 'sl': 'kmerščina', 'smn': 'khmerkielâ', 'sn': 'chiKhema', 'so': 'Kamboodhian', 'sq': 'kmerisht', 'sr': 'кмерски', 'sr-Cyrl': 'кмерски', 'sr-Latn': 'kmerski', 'sv': 'kambodjanska', 'sw': 'Kikambodia', 'ta': 'கெமெர்', 'te': 'ఖ్మేర్', 'teo': 'Kikambodia', 'tg': 'кхмерӣ', 'th': 'เขมร', 'ti': 'ክመር', 'tk': 'khmer dili', 'to': 'lea fakakamipōtia', 'tr': 'Khmer dili', 'tt': 'кхмер', 'twq': 'Kmeer senni, Game here', 'tzm': 'Taxmert ,Talammast', 'ug': 'كىمېرچە', 'uk': 'кхмерська', 'ur': 'خمیر', 'uz': 'xmer', 'uz-Cyrl': 'хмерча', 'uz-Latn': 'xmer', 'vai': 'ꕃꘈꗢ', 'vai-Latn': 'Kimɛɛ̃ tɛ', 'vai-Vaii': 'ꕃꘈꗢ', 'vec': 'canbojan', 'vi': 'Tiếng Khmer', 'vun': 'Kyikambodia', 'wae': 'Kambodšaniš', 'wo': 'Xmer', 'xnr': 'खमेर', 'xog': 'Olukme', 'yav': 'kímɛɛ', 'yi': 'כמער', 'yo': 'Èdè kameri', 'yrl': 'kimé', 'yue': '高棉文', 'yue-Hans': '高棉文', 'yue-Hant': '高棉文', 'zgh': 'ⵜⴰⵅⵎⵉⵔⵜ', 'zh': '高棉语', 'zh-Hans': '高棉语', 'zh-Hant': '高棉文', 'zu': 'isi-Khmer'},
    'kma': {'en': 'Konni'},
    'kmb': {'af': 'Kimbundu', 'am': 'ኪምቡንዱ', 'ar': 'الكيمبندو', 'as': 'কিম্বুন্দু', 'ast': 'kimbundu', 'az': 'kimbundu', 'az-Cyrl': 'кимбунду', 'az-Latn': 'kimbundu', 'be': 'кімбунду', 'be-tarask': 'кімбунду', 'bg': 'кимбунду', 'bgn': 'کیمبوندویی', 'bn': 'কিম্বুন্দু', 'br': 'kimbundu', 'brx': 'किम्बुन्डु', 'bs': 'kimbundu', 'bs-Cyrl': 'кимбунду', 'bs-Latn': 'kimbundu', 'ca': 'kimbundu', 'ca-ES-valencia': 'kimbundu', 'ccp': '𑄇𑄨𑄟𑄴𑄝𑄪𑄚𑄴𑄘𑄪', 'ce': 'кимбунду', 'chr': 'ᎩᎻᏊᏚ', 'ckb': 'کیمبووندوو', 'cs': 'kimbundština', 'cy': 'Kimbundu', 'da': 'kimbundu', 'de': 'Kimbundu', 'dsb': 'kimbundušćina', 'el': 'Κιμπούντου', 'el-polyton': 'Κιμπούντου', 'en': 'Kimbundu', 'es': 'kimbundu', 'et': 'mbundu', 'eu': 'kimbundua', 'fa': 'کیمبوندویی', 'ff-Adlm': '𞤑𞤭𞤥𞤦𞤵𞤲𞤣𞤵𞥅𞤪𞤫', 'fi': 'kimbundu', 'fil': 'Kimbundu', 'fo': 'kimbundu', 'fr': 'kimboundou', 'fy': 'Kimbundu', 'ga': 'Ciombundais', 'gd': 'Kimbundu', 'gl': 'kimbundu', 'gsw': 'Kimbundu-Schpraach', 'gu': 'કિમ્બન્દુ', 'ha': 'Kimbundu', 'he': 'קימבונדו', 'hi': 'किम्बन्दु', 'hi-Latn': 'Kimbundu', 'hr': 'kimbundu', 'hsb': 'kimbundušćina', 'ht': 'kimboundou', 'hu': 'kimbundu', 'hy': 'կիմբունդու', 'ia': 'kimbundu', 'id': 'Kimbundu', 'ig': 'Kimbundu', 'is': 'kimbúndú', 'it': 'kimbundu', 'ja': 'キンブンド語', 'jv': 'Kimbundu', 'ka': 'კიმბუნდუ', 'kgp': 'kĩmuno', 'kk': 'кимбунду тілі', 'kk-Cyrl': 'кимбунду тілі', 'km': 'គីមប៊ុនឌូ', 'kn': 'ಕಿಂಬುಂಡು', 'ko': '킴분두어', 'kok': 'किंबुंडु', 'kok-Deva': 'किंबुंडु', 'ks': 'کِمبُندوٗ', 'ks-Arab': 'کِمبُندوٗ', 'ksh': 'Kimbundu', 'ku': 'kîmbunduyî', 'ky': 'кимбундуча', 'lb': 'Kimbundu-Sprooch', 'lo': 'ຄິມບັນດູ', 'lt': 'kimbundu', 'lv': 'kimbundu', 'mai': 'किमबुंडू', 'mi': 'Kimipunu', 'mk': 'кимбунду', 'ml': 'കിംബുണ്ടു', 'mn': 'кимбунду', 'mr': 'किम्बन्दु', 'ms': 'Kimbundu', 'mt': 'Kimbundu', 'my': 'ကင်ဘွန်ဒူ', 'nb': 'kimbundu', 'ne': 'किम्बुन्डु', 'nl': 'Kimbundu', 'nn': 'kimbundu', 'no': 'kimbundu', 'or': 'କିମ୍ବୁଣ୍ଡୁ', 'pa': 'ਕਿਮਬੁੰਦੂ', 'pa-Guru': 'ਕਿਮਬੁੰਦੂ', 'pcm': 'Kimbúndú Lángwej', 'pl': 'kimbundu', 'ps': 'کیمبوندو', 'pt': 'quimbundo', 'qu': 'Kimbundu Simi', 'rm': 'kimbundu', 'ro': 'kimbundu', 'ru': 'кимбунду', 'sc': 'kimbundu', 'sd': 'ڪنمبونڊو', 'sd-Arab': 'ڪنمبونڊو', 'si': 'කිම්බුන්ඩු', 'sk': 'kimbundu', 'sl': 'kimbundu', 'smn': 'kimbundu', 'so': 'Kimbundu', 'sq': 'kimbunduisht', 'sr': 'кимбунду', 'sr-Cyrl': 'кимбунду', 'sr-Latn': 'kimbundu', 'sv': 'kimbundu', 'sw': 'Kimbundu', 'ta': 'கிம்புன்து', 'te': 'కిమ్బుండు', 'th': 'คิมบุนดู', 'ti': 'ኪምቡንዱ', 'tk': 'kimbundu dili', 'to': 'lea fakakimipūnitu', 'tr': 'Kimbundu', 'ug': 'كىمبۇندۇچە', 'uk': 'кімбунду', 'ur': 'کیمبونڈو', 'uz': 'kimbundu', 'uz-Latn': 'kimbundu', 'vec': 'kinbundu', 'vi': 'Tiếng Kimbundu', 'xnr': 'किम्बन्दु', 'yo': 'Èdè Kimbundu', 'yrl': 'kĩbũdu', 'yue': '金邦杜文', 'yue-Hans': '金邦杜文', 'yue-Hant': '金邦杜文', 'zh': '金邦杜语', 'zh-Hans': '金邦杜语', 'zh-Hant': '金邦杜文', 'zu': 'isi-Kimbundu'},
    'kmc': {'en': 'Southern Dong'},
    'kmd': {'en': 'Majukayang Kalinga'},
    'kme': {'en': 'Bakole'},
    'kmf': {'en': 'Kare (Papua New Guinea)'},
    'kmg': {'en': 'Kâte'},
    'kmh': {'en': 'Kalam'},
    'kmi': {'en': 'Kami (Nigeria)'},
    'kmj': {'en': 'Kumarbhag Paharia'},
    'kmk': {'en': 'Limos Kalinga'},
    'kml': {'en': 'Tanudan Kalinga'},
    'kmm': {'en': 'Kom (India)'},
    'kmn': {'en': 'Awtuw'},
    'kmo': {'en': 'Kwoma'},
    'kmp': {'en': 'Gimme'},
    'kmq': {'en': 'Kwama'},
    'kmr': {'en': 'Northern Kurdish'},
    'kms': {'en': 'Kamasau'},
    'kmt': {'en': 'Kemtuik'},
    'kmu': {'en': 'Kanite'},
    'kmv': {'en': 'Karipúna Creole French'},
    'kmw': {'en': 'Komo (Democratic Republic of Congo)'},
    'kmx': {'en': 'Waboda'},
    'kmy': {'en': 'Koma'},
    'kmz': {'en': 'Khorasani Turkish'},
    'kn': {'af': 'Kannada', 'ak': 'Kanada', 'am': 'ካናዳ', 'ar': 'الكانادا', 'as': 'কানাড়া', 'ast': 'canarés', 'az': 'kannada', 'az-Cyrl': 'каннада', 'az-Latn': 'kannada', 'bal-Latn': 'Kannadá', 'be': 'канада', 'be-tarask': 'канада', 'bg': 'каннада', 'bgn': 'کانارا', 'bn': 'কন্নড়', 'br': 'kanareg', 'brx': 'कन्नड', 'bs': 'kanada', 'bs-Cyrl': 'канада', 'bs-Latn': 'kanada', 'ca': 'kannada', 'ca-ES-valencia': 'kannada', 'ccp': '𑄇𑄧𑄚𑄴𑄚𑄧𑄢𑄴', 'ce': 'каннада', 'chr': 'ᎧᎾᏓ', 'ckb': 'کاننادا', 'cs': 'kannadština', 'cy': 'Kannada', 'da': 'kannada', 'de': 'Kannada', 'dsb': 'kannadšćina', 'dz': 'ཀ་ན་ཌ་ཁ', 'ee': 'kannadagbe', 'el': 'Κανάντα', 'el-polyton': 'Κανάντα', 'en': 'Kannada', 'eo': 'kanara', 'es': 'canarés', 'et': 'kannada', 'eu': 'kannada', 'fa': 'کانارا', 'ff-Adlm': '𞤑𞤢𞤲𞥆𞤢𞤣𞤢𞥄𞤪𞤫', 'fi': 'kannada', 'fil': 'Kannada', 'fo': 'kannada', 'fr': 'kannada', 'fur': 'kannada', 'fy': 'Kannada', 'ga': 'Cannadais', 'gd': 'Kannada', 'gl': 'kannará', 'gsw': 'Kannada', 'gu': 'કન્નડ', 'ha': 'Kannada', 'he': 'קנאדה', 'hi': 'कन्नड़', 'hi-Latn': 'Kannada', 'hr': 'karnatački', 'hsb': 'kannadšćina', 'ht': 'kannada', 'hu': 'kannada', 'hy': 'կաննադա', 'ia': 'kannada', 'id': 'Kannada', 'ig': 'Kannada', 'is': 'kannada', 'it': 'kannada', 'ja': 'カンナダ語', 'jv': 'Kannada', 'ka': 'კანადა', 'kea': 'kanares', 'kgp': 'kanỹrim', 'kk': 'каннада тілі', 'kk-Cyrl': 'каннада тілі', 'km': 'ខាណាដា', 'kn': 'ಕನ್ನಡ', 'ko': '칸나다어', 'kok': 'कन्नड', 'kok-Deva': 'कन्नड', 'kok-Latn': 'Kon’nodd', 'ks': 'کَنَڑ', 'ks-Arab': 'کَنَڑ', 'ksh': 'Kannada', 'ku': 'kannadayî', 'kxv': 'knnaḍa', 'kxv-Deva': 'कन्नड़', 'kxv-Latn': 'knnaḍa', 'kxv-Orya': 'କାନ୍ନାଡ଼ା', 'kxv-Telu': 'కన్నడ', 'ky': 'каннадача', 'lb': 'Kannada', 'lkt': 'Kannada Iyápi', 'lld': 'kannada', 'lo': 'ຄັນນາດາ', 'lrc': 'کاناد', 'lt': 'kanadų', 'lv': 'kannadu', 'mai': 'कन्नड़', 'mi': 'Kanara', 'mk': 'каннада', 'ml': 'കന്നഡ', 'mn': 'каннада', 'mr': 'कन्नड', 'ms': 'Kannada', 'mt': 'Kannada', 'my': 'ကန်နာဒါ', 'mzn': 'کانّادا', 'nb': 'kannada', 'ne': 'कन्नाडा', 'nl': 'Kannada', 'nn': 'kannada', 'no': 'kannada', 'nqo': 'ߞߊ߲ߣߊߘߊߞߊ߲', 'om': 'Afaan Kannada', 'or': 'କନ୍ନଡ', 'pa': 'ਕੰਨੜ', 'pa-Guru': 'ਕੰਨੜ', 'pcm': 'Kánnáda Lángwej', 'pl': 'kannada', 'ps': 'کناډا', 'pt': 'canarim', 'qu': 'Kannada Simi', 'rm': 'kannada', 'ro': 'kannada', 'ru': 'каннада', 'rw': 'Igikanada', 'sc': 'kannada', 'sd': 'ڪناڊا', 'sd-Arab': 'ڪناڊا', 'si': 'කණ්ණඩ', 'sk': 'kannadčina', 'sl': 'kanareščina', 'smn': 'kannada', 'so': 'Kannadays', 'sq': 'kanadisht', 'sr': 'канада', 'sr-Cyrl': 'канада', 'sr-Latn': 'kanada', 'sv': 'kannada', 'sw': 'Kikannada', 'ta': 'கன்னடம்', 'te': 'కన్నడ', 'tg': 'каннада', 'th': 'กันนาดา', 'ti': 'ካንናዳ', 'tk': 'kannada dili', 'to': 'lea fakakanata', 'tr': 'Kannada dili', 'tt': 'каннада', 'ug': 'كانناداچە', 'uk': 'каннада', 'ur': 'کنّاڈا', 'uz': 'kannada', 'uz-Cyrl': 'каннада', 'uz-Latn': 'kannada', 'vec': 'canareze', 'vi': 'Tiếng Kannada', 'wae': 'Kannada', 'wo': 'Kannadaa', 'xnr': 'कन्नड़', 'yi': 'קאַנאַדאַ', 'yo': 'Èdè Kannada', 'yrl': 'kanarĩ', 'yue': '坎那達文', 'yue-Hans': '坎那达文', 'yue-Hant': '坎那達文', 'zh': '卡纳达语', 'zh-Hans': '卡纳达语', 'zh-Hant': '坎那達文', 'zu': 'isi-Kannada'},
    'kna': {'en': 'Dera (Nigeria)'},
    'knb': {'en': 'Lubuagan Kalinga'},
    'knc': {'en': 'Central Kanuri'},
    'knd': {'en': 'Konda'},
    'kne': {'en': 'Kankanaey'},
    'knf': {'en': 'Mankanya'},
    'kng': {'en': 'Koongo'},
    'kni': {'en': 'Kanufi'},
    'knj': {'en': 'Western Kanjobal'},
    'knk': {'en': 'Kuranko'},
    'knl': {'en': 'Keninjal'},
    'knm': {'en': 'Kanamarí'},
    'knn': {'en': 'Konkani (individual language)'},
    'kno': {'en': 'Kono (Sierra Leone)'},
    'knp': {'en': 'Kwanja'},
    'knq': {'en': 'Kintaq'},
    'knr': {'en': 'Kaningra'},
    'kns': {'en': 'Kensiu'},
    'knt': {'en': 'Panoan Katukína'},
    'knu': {'en': 'Kono (Guinea)'},
    'knv': {'en': 'Tabo'},
    'knw': {'en': 'Kung-Ekoka'},
    'knx': {'en': 'Kendayan'},
    'kny': {'en': 'Kanyok'},
    'knz': {'en': 'Kalamsé'},
    'ko': {'af': 'Koreaans', 'agq': 'kùulîa', 'ak': 'Korea kasa', 'am': 'ኮሪያኛ', 'ar': 'الكورية', 'as': 'কোৰিয়ান', 'asa': 'Kikorea', 'ast': 'coreanu', 'az': 'koreya', 'az-Arab': 'کوْره\u200cجه', 'az-Cyrl': 'кореја', 'az-Latn': 'koreya', 'bal-Latn': 'Kuriái', 'bas': 'Hɔp u kɔrēà', 'be': 'карэйская', 'be-tarask': 'карэйская', 'bem': 'Ichi Koriani', 'bez': 'Hikorea', 'bg': 'корейски', 'bgn': 'کوریایی', 'blo': 'Koree kagɩja', 'bm': 'korekan', 'bn': 'কোরিয়ান', 'br': 'koreaneg', 'brx': 'करियान', 'bs': 'korejski', 'bs-Cyrl': 'корејски', 'bs-Latn': 'korejski', 'ca': 'coreà', 'ca-ES-valencia': 'coreà', 'ccp': '𑄇𑄮𑄢𑄨𑄠𑄚𑄴', 'ce': 'корейн', 'ceb': 'Korean', 'cgg': 'Orukoreya', 'chr': 'ᎪᎵᎠᏂ', 'ckb': 'كۆری', 'cs': 'korejština', 'cv': 'корей', 'cy': 'Coreeg', 'da': 'koreansk', 'dav': 'Kikorea', 'de': 'Koreanisch', 'dje': 'Koree senni', 'dsb': 'korejańšćina', 'dyo': 'koree', 'dz': 'ཀོ་རི་ཡཱན་ཁ', 'ebu': 'Kĩkorea', 'ee': 'Koreagbe', 'el': 'Κορεατικά', 'el-polyton': 'Κορεατικά', 'en': 'Korean', 'en-Dsrt': '𐐗𐐬𐑉𐐨𐐲𐑌', 'eo': 'korea', 'es': 'coreano', 'et': 'korea', 'eu': 'koreera', 'ewo': 'Ǹkɔ́bɔ koréan', 'fa': 'کره\u200cای', 'ff': 'Koreere', 'ff-Adlm': '𞤑𞤮𞥅𞤪𞤫𞤴𞤢𞤲𞤪𞤫', 'ff-Latn': 'Koreere', 'fi': 'korea', 'fil': 'Korean', 'fo': 'koreanskt', 'fr': 'coréen', 'fur': 'corean', 'fy': 'Koreaansk', 'ga': 'Cóiréis', 'gd': 'Coirèanais', 'gl': 'coreano', 'gsw': 'Koreaanisch', 'gu': 'કોરિયન', 'guz': 'Kikorea', 'ha': 'Harshen Koriya', 'haw': 'Kōlea', 'he': 'קוריאנית', 'hi': 'कोरियाई', 'hi-Latn': 'Korean', 'hr': 'korejski', 'hsb': 'korejšćina', 'ht': 'coréen', 'hu': 'koreai', 'hy': 'կորեերեն', 'ia': 'coreano', 'id': 'Korea', 'ie': 'korean', 'ig': 'Korean', 'is': 'kóreska', 'it': 'coreano', 'ja': '韓国語', 'jmc': 'Kyikorea', 'jv': 'Korea', 'ka': 'კორეული', 'kaa': 'корейсше', 'kaa-Cyrl': 'корейсше', 'kab': 'Takurit', 'kam': 'Kikorea', 'kde': 'Chikolea', 'kea': 'korianu', 'kgp': 'korejỹnũ', 'khq': 'Koree senni', 'ki': 'Kikorea', 'kk': 'корей тілі', 'kk-Cyrl': 'корей тілі', 'kln': 'kutitab Korea', 'km': 'កូរ៉េ', 'kn': 'ಕೊರಿಯನ್', 'ko': '한국어', 'kok': 'कोरियन', 'kok-Deva': 'कोरियन', 'ks': 'کوریَن', 'ks-Arab': 'کوریَن', 'ksb': 'Kikolea', 'ksf': 'rikɔrɛɛ́', 'ksh': 'Korrejaanesch', 'ku': 'koreyî', 'kxv': 'koriati', 'kxv-Deva': 'कोरियाति', 'kxv-Latn': 'koriati', 'kxv-Orya': 'କରିଆନ୍ ତି', 'kxv-Telu': 'కొరియన్', 'ky': 'корейче', 'lag': 'Kɨkoréa', 'lb': 'Koreanesch', 'lg': 'Lukoreya', 'lkt': 'Korea Iyápi', 'lld': 'corean', 'ln': 'likoreya', 'lo': 'ເກົາຫລີ', 'lrc': 'کورئ یی', 'lt': 'korėjiečių', 'lu': 'Likoreya', 'luo': 'Kikorea', 'luy': 'Kikorea', 'lv': 'korejiešu', 'mai': 'कोरियाई', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lkorea', 'mer': 'Kĩkorea', 'mfe': 'koreen', 'mg': 'Koreanina', 'mgh': 'Ikorea', 'mi': 'Kōreana', 'mk': 'корејски', 'ml': 'കൊറിയൻ', 'mn': 'солонгос', 'mni': 'কোরিয়ন', 'mni-Beng': 'কোরিয়ন', 'mr': 'कोरियन', 'ms': 'Korea', 'mt': 'Korean', 'mua': 'korea', 'my': 'ကိုရီးယား', 'mzn': 'کُره\u200cیی', 'naq': 'Koreaǁî gowab', 'nb': 'koreansk', 'nd': 'isi-Koriya', 'ne': 'कोरियाली', 'nl': 'Koreaans', 'nmg': 'Kiɛl koré', 'nn': 'koreansk', 'no': 'koreansk', 'nqo': 'ߞߏߙߋߞߊ߲', 'nus': 'Thok kuriani', 'nyn': 'Orukoreya', 'om': 'Afaan Korea', 'or': 'କୋରିଆନ୍', 'pa': 'ਕੋਰੀਆਈ', 'pa-Guru': 'ਕੋਰੀਆਈ', 'pcm': 'Koriá Lángwej', 'pl': 'koreański', 'ps': 'کوریایی', 'pt': 'coreano', 'qu': 'Coreano Simi', 'rm': 'corean', 'rn': 'Ikinyakoreya', 'ro': 'coreeană', 'rof': 'Kikorea', 'ru': 'корейский', 'rw': 'Igikoreya', 'rwk': 'Kyikorea', 'sah': 'Кэриэйдии', 'saq': 'Kikorea', 'sbp': 'Ishikoleya', 'sc': 'coreanu', 'sd': 'ڪوريائي', 'sd-Arab': 'ڪوريائي', 'se': 'koreagiella', 'seh': 'coreano', 'ses': 'Koree senni', 'sg': 'Koreyëen', 'shi': 'ⵜⴰⴽⵓⵔⵉⵜ', 'shi-Latn': 'Takurit', 'shi-Tfng': 'ⵜⴰⴽⵓⵔⵉⵜ', 'si': 'කොරියානු', 'sk': 'kórejčina', 'sl': 'korejščina', 'smn': 'koreakielâ', 'sn': 'chiKoria', 'so': 'Kuuriyaan', 'sq': 'koreanisht', 'sr': 'корејски', 'sr-Cyrl': 'корејски', 'sr-Latn': 'korejski', 'sv': 'koreanska', 'sw': 'Kikorea', 'ta': 'கொரியன்', 'te': 'కొరియన్', 'teo': 'Kikorea', 'tg': 'кореягӣ', 'th': 'เกาหลี', 'ti': 'ኮርይኛ', 'tk': 'koreý dili', 'to': 'lea fakakōlea', 'tr': 'Korece', 'tt': 'корея', 'twq': 'Koree senni', 'tzm': 'Takurit', 'ug': 'كورېيەچە', 'uk': 'корейська', 'ur': 'کوریائی', 'uz': 'koreyscha', 'uz-Cyrl': 'корейсча', 'uz-Latn': 'koreyscha', 'vai': 'ꖏꔸꘂꘋ', 'vai-Latn': 'Koríyɛŋ', 'vai-Vaii': 'ꖏꔸꘂꘋ', 'vec': 'corean', 'vi': 'Tiếng Hàn', 'vun': 'Kyikorea', 'wae': 'Koreaniš', 'wo': 'Koreye', 'xh': 'Isi-Korean', 'xnr': 'कोरियाई', 'xog': 'Olukoreya', 'yav': 'kolíe', 'yi': 'קארעאיש', 'yo': 'Èdè Kòríà', 'yrl': 'kurianu', 'yue': '韓文', 'yue-Hans': '韩文', 'yue-Hant': '韓文', 'zgh': 'ⵜⴰⴽⵓⵔⵉⵜ', 'zh': '韩语', 'zh-Hans': '韩语', 'zh-Hant': '韓文', 'zu': 'isi-Korean'},
    'koa': {'en': 'Konomala'},
    'koc': {'en': 'Kpati'},
    'kod': {'en': 'Kodi'},
    'koe': {'en': 'Kacipo-Bale Suri'},
    'kof': {'en': 'Kubi'},
    'kog': {'en': 'Cogui'},
    'koh': {'en': 'Koyo'},
    'koi': {'af': 'Komi-Permyaks', 'am': 'ኮሚ ፔርምያክ', 'ar': 'كومي-بيرماياك', 'ast': 'komi-permyak', 'az': 'komi-permyak', 'az-Latn': 'komi-permyak', 'be': 'комі-пярмяцкая', 'be-tarask': 'комі-пярмяцкая', 'bg': 'коми-пермякски', 'bn': 'কমি-পারমিআক', 'bs': 'komi-permski', 'bs-Latn': 'komi-permski', 'ca': 'komi-permiac', 'ca-ES-valencia': 'komi-permiac', 'ccp': '𑄇𑄧𑄟𑄨-𑄛𑄢𑄧𑄟𑄨𑄃𑄇𑄴', 'ce': 'коми-пермякийн', 'cs': 'komi-permjačtina', 'cy': 'Komi-Permyak', 'da': 'komi-permjakisk', 'de': 'Komi-Permjakisch', 'dsb': 'komi-permyak', 'el': 'Κόμι-Περμιάκ', 'el-polyton': 'Κόμι-Περμιάκ', 'en': 'Komi-Permyak', 'es': 'komi permio', 'et': 'permikomi', 'eu': 'komi-permyakera', 'fa': 'کومی پرمیاک', 'fi': 'komipermjakki', 'fil': 'Komi-Permyak', 'fo': 'komi-permyak', 'fr': 'komi-permiak', 'gd': 'Komi-Permyak', 'gl': 'komi permio', 'gu': 'કોમી-પર્મ્યાક', 'he': 'קומי-פרמיאקית', 'hi': 'कोमी-पर्मयाक', 'hi-Latn': 'Komi-Permyak', 'hr': 'komi-permski', 'hsb': 'permska komišćina', 'ht': 'komi-permiak', 'hu': 'komi-permják', 'hy': 'պերմյակ կոմիերեն', 'id': 'Komi-Permyak', 'is': 'kómí-permyak', 'it': 'permiaco', 'ja': 'コミ・ペルミャク語', 'ka': 'კომი-პერმიაკული', 'kea': 'komi-permiak', 'kgp': 'komĩ-permyjag', 'kk': 'коми-пермяк тілі', 'kk-Cyrl': 'коми-пермяк тілі', 'km': 'គូមីភឹមយ៉ាគ', 'kn': 'ಕೋಮಿ-ಪರ್ಮ್ಯಕ್', 'ko': '코미페르먀크어', 'ky': 'коми-пермякча', 'lb': 'Komi-Permiak', 'lo': 'ໂຄມີ-ເພີມຢັກ', 'lrc': 'کومی پئرمیاک', 'lt': 'komių-permių', 'lv': 'komiešu-permiešu', 'mk': 'коми-пермјачки', 'ml': 'കോമി-പെർമ്യാക്ക്', 'mn': 'коми-пермяк', 'mr': 'कोमी-परम्याक', 'ms': 'Komi-Permyak', 'my': 'ကိုမီ-ပါမြက်', 'mzn': 'کومی-پرمیاک', 'nb': 'komipermjakisk', 'ne': 'कोमी-पर्म्याक', 'nl': 'Komi-Permjaaks', 'nn': 'komipermjakisk', 'no': 'komipermjakisk', 'pa': 'ਕੋਮੀ-ਪੇਰਮਿਆਕ', 'pa-Guru': 'ਕੋਮੀ-ਪੇਰਮਿਆਕ', 'pl': 'komi-permiacki', 'pt': 'komi-permyak', 'ro': 'komi-permiak', 'ru': 'коми-пермяцкий', 'si': 'කොමි-පර්මියාක්', 'sk': 'komi-permiačtina', 'sl': 'komi-permjaščina', 'sq': 'komi-parmjakisht', 'sr': 'коми-пермски', 'sr-Cyrl': 'коми-пермски', 'sr-Latn': 'komi-permski', 'sv': 'komi-permjakiska', 'sw': 'Kikomi-Permyak', 'ta': 'கொமி-பெர்ம்யாக்', 'te': 'కోమి-పర్మాక్', 'th': 'โคมิ-เปียร์เมียค', 'to': 'lea fakakomi-pelemiaki', 'tr': 'Komi-Permyak', 'uk': 'комі-перм’яцька', 'ur': 'کومی پرمیاک', 'uz': 'komi-permyak', 'uz-Latn': 'komi-permyak', 'vi': 'Tiếng Komi-Permyak', 'yrl': 'kumi-perimiaki', 'yue': '科米-彼爾米亞克文', 'yue-Hans': '科米-彼尔米亚克文', 'yue-Hant': '科米-彼爾米亞克文', 'zh': '科米-彼尔米亚克语', 'zh-Hans': '科米-彼尔米亚克语', 'zh-Hant': '科米-彼爾米亞克文', 'zu': 'isi-Komi-Permyak'},
    'koj': {'en': 'Sara Dunjo'},
    'kok': {'af': 'Konkani', 'ak': 'Konkani kasa', 'am': 'ኮንካኒ', 'ar': 'الكونكانية', 'as': 'কোংকণী', 'ast': 'konkani', 'az': 'konkani', 'az-Cyrl': 'конкани', 'az-Latn': 'konkani', 'bal-Latn': 'Konkani', 'be': 'канкані', 'be-tarask': 'канкані', 'bg': 'конкани', 'bgn': 'کونکانی', 'bn': 'কোঙ্কানি', 'br': 'konkani', 'brx': 'कोंकणी', 'bs': 'konkani', 'bs-Cyrl': 'конкани', 'bs-Latn': 'konkani', 'ca': 'concani', 'ca-ES-valencia': 'concani', 'ccp': '𑄇𑄮𑄋𑄴𑄇𑄚𑄨', 'ce': 'конкани', 'chr': 'ᎧᏂᎧᏂ', 'ckb': 'کۆنکانی', 'cs': 'konkánština', 'cy': 'Concani', 'da': 'konkani', 'de': 'Konkani', 'dsb': 'konkani', 'el': 'Κονκανικά', 'el-polyton': 'Κονκανικά', 'en': 'Konkani', 'es': 'konkaní', 'et': 'konkani', 'eu': 'konkanera', 'fa': 'کنکانی', 'ff-Adlm': '𞤑𞤮𞤲𞤳𞤢𞤲𞤭𞥅𞤪𞤫', 'fi': 'konkani', 'fil': 'Konkani', 'fo': 'konkani', 'fr': 'konkani', 'fy': 'Konkani', 'ga': 'Concáinis', 'gd': 'Konkani', 'gl': 'konkani', 'gsw': 'Konkani', 'gu': 'કોંકણી', 'ha': 'Konkananci', 'he': 'קונקאני', 'hi': 'कोंकणी', 'hi-Latn': 'Konkani', 'hr': 'konkani', 'hsb': 'konkani', 'ht': 'konkani', 'hu': 'konkani', 'hy': 'կոնկանի', 'ia': 'konkani', 'id': 'Konkani', 'ig': 'Konkani', 'is': 'konkaní', 'it': 'konkani', 'ja': 'コンカニ語', 'jv': 'Konkani', 'ka': 'კონკანი', 'kea': 'konkani', 'kgp': 'kãkani', 'kk': 'конкани тілі', 'kk-Cyrl': 'конкани тілі', 'km': 'គុនកានី', 'kn': 'ಕೊಂಕಣಿ', 'ko': '코카니어', 'kok': 'कोंकणी', 'kok-Deva': 'कोंकणी', 'kok-Latn': 'Konknni', 'ks': 'کونکَنی', 'ks-Arab': 'کونکَنی', 'ksh': 'Kon’kahnesch', 'ku': 'konkanî', 'kxv': 'konkanī', 'kxv-Deva': 'कोंकणी', 'kxv-Latn': 'konkanī', 'kxv-Orya': 'କଂଙ୍କଣି', 'kxv-Telu': 'కొంకణి', 'ky': 'конканиче', 'lb': 'Konkani', 'lo': 'ກອນການີ', 'lrc': 'کوٙنکانی', 'lt': 'konkanių', 'lv': 'konkanu', 'mai': 'कोंकणी', 'mi': 'Kōkani', 'mk': 'конкани', 'ml': 'കൊങ്കണി', 'mn': 'конкани', 'mr': 'कोंकणी', 'ms': 'Konkani', 'mt': 'Konkani', 'my': 'ကွန်ကနီ', 'mzn': 'کونکانی', 'nb': 'konkani', 'ne': 'कोन्कानी', 'nl': 'Konkani', 'nn': 'konkani', 'no': 'konkani', 'nqo': 'ߞߐ߲ߞߊߣߌߞߊ߲', 'or': 'କୋଙ୍କଣି', 'pa': 'ਕੋਂਕਣੀ', 'pa-Guru': 'ਕੋਂਕਣੀ', 'pcm': 'Kónkéní Lángwej', 'pl': 'konkani', 'ps': 'کونکاني', 'pt': 'concani', 'qu': 'Konkani Simi', 'rm': 'konkani', 'ro': 'konkani', 'ru': 'конкани', 'sc': 'konkani', 'sd': 'ڪونڪي', 'sd-Arab': 'ڪونڪي', 'si': 'කොන්කනි', 'sk': 'konkánčina', 'sl': 'konkanščina', 'smn': 'konkani', 'so': 'Konkani', 'sq': 'konkanisht', 'sr': 'конкани', 'sr-Cyrl': 'конкани', 'sr-Latn': 'konkani', 'sv': 'konkani', 'sw': 'Kikonkani', 'ta': 'கொங்கணி', 'te': 'కొంకణి', 'tg': 'конканӣ', 'th': 'กอนกานี', 'ti': 'ኮንካኒ', 'tk': 'konkani dili', 'to': 'lea fakakonikanī', 'tr': 'Konkani dili', 'tt': 'конкани', 'ug': 'كونكانچە', 'uk': 'конкані', 'ur': 'کونکنی', 'uz': 'konkan', 'uz-Cyrl': 'конканча', 'uz-Latn': 'konkan', 'vec': 'konkani', 'vi': 'Tiếng Konkani', 'wo': 'Konkani', 'xnr': 'कोंकणी', 'yo': 'Kónkánì', 'yrl': 'kũkani', 'yue': '貢根文', 'yue-Hans': '贡根文', 'yue-Hant': '貢根文', 'zh': '孔卡尼语', 'zh-Hans': '孔卡尼语', 'zh-Hant': '貢根文', 'zu': 'isi-Konkani'},
    'kol': {'en': 'Kol (Papua New Guinea)'},
    'koo': {'en': 'Konzo'},
    'kop': {'en': 'Waube'},
    'koq': {'en': 'Kota (Gabon)'},
    'kos': {'ar': 'الكوسراين', 'ast': 'kosraeanu', 'az': 'kosreyan', 'az-Latn': 'kosreyan', 'bg': 'косраен', 'bn': 'কোস্রাইন', 'br': 'kosrae', 'brx': 'कोस्राईयन्', 'bs': 'kosrejski', 'bs-Cyrl': 'косреански', 'bs-Latn': 'kosrejski', 'ca': 'kosraeà', 'ca-ES-valencia': 'kosraeà', 'ccp': '𑄇𑄮𑄥𑄳𑄢𑄭𑄚𑄴', 'cs': 'kosrajština', 'da': 'kosraean', 'de': 'Kosraeanisch', 'el': 'Κοσραενικά', 'el-polyton': 'Κοσραενικά', 'en': 'Kosraean', 'es': 'kosraeano', 'et': 'kosrae', 'fi': 'kosrae', 'fr': 'kosraéen', 'fy': 'Kosraeaansk', 'gsw': 'Kosraeanisch', 'gu': 'કોસરિયન', 'he': 'קוסראיאן', 'hi': 'कोसरैन', 'hi-Latn': 'Kosraean', 'hr': 'naurski', 'ht': 'kosraéen', 'hu': 'kosrei', 'id': 'Kosre', 'is': 'kosraska', 'it': 'kosraean', 'ja': 'コスラエ語', 'ka': 'კუსაიე', 'kgp': 'kosirajỹn', 'kn': 'ಕೊಸರಿಯನ್', 'ko': '코스라이엔어', 'ks': 'کوسرییَن', 'ks-Arab': 'کوسرییَن', 'lb': 'Kosraeanesch', 'lo': 'ຄູສໄລ', 'lt': 'kosreanų', 'lv': 'kosrājiešu', 'mk': 'козрејски', 'ml': 'കൊസറേയൻ', 'mr': 'कोसरियन', 'mt': 'Kosrejan', 'nb': 'kosraeansk', 'ne': 'कोस्राली', 'nl': 'Kosraeaans', 'nn': 'kosraeansk', 'no': 'kosraeansk', 'or': 'କୋସରୈନ୍', 'pl': 'kosrae', 'pt': 'kosraean', 'rm': 'kosraean', 'ro': 'kosrae', 'ru': 'косраенский', 'sk': 'kusaie', 'sl': 'kosrajščina', 'sr': 'косренски', 'sr-Cyrl': 'косренски', 'sr-Latn': 'kosrenski', 'sv': 'kosreanska', 'ta': 'கோஸ்ரைன்', 'te': 'కోస్రేయన్', 'th': 'คูสไร', 'to': 'lea fakakosilae', 'tr': 'Kosraean', 'ug': 'كوسرايېچە', 'uk': 'косрае', 'vi': 'Tiếng Kosrae', 'yrl': 'kusirayã', 'yue': '科斯雷恩文', 'yue-Hans': '科斯雷恩文', 'yue-Hant': '科斯雷恩文', 'zh': '科斯拉伊语', 'zh-Hans': '科斯拉伊语', 'zh-Hant': '科斯雷恩文'},
    'kot': {'en': 'Lagwan'},
    'kou': {'en': 'Koke'},
    'kov': {'en': 'Kudu-Camo'},
    'kow': {'en': 'Kugama'},
    'kox': {'en': 'Coxima'},
    'koy': {'en': 'Koyukon'},
    'koz': {'en': 'Korak'},
    'kpa': {'en': 'Kutto'},
    'kpb': {'en': 'Mullu Kurumba'},
    'kpc': {'en': 'Curripaco'},
    'kpd': {'en': 'Koba'},
    'kpe': {'af': 'Kpellees', 'am': 'ክፔሌ', 'ar': 'الكبيل', 'as': 'কেপেল', 'ast': 'kpelle', 'az': 'kpelle', 'az-Cyrl': 'кпелле', 'az-Latn': 'kpelle', 'bal-Latn': 'Kpelle', 'be': 'кпеле', 'be-tarask': 'кпеле', 'bg': 'кпеле', 'bgn': 'کپله\u200cای', 'bn': 'ক্\u200cপেল্লে', 'br': 'kpelle', 'brx': 'क्पेले', 'bs': 'kpele', 'bs-Cyrl': 'кпеле', 'bs-Latn': 'kpele', 'ca': 'kpelle', 'ca-ES-valencia': 'kpelle', 'ccp': '𑄇𑄴𑄛𑄬𑄣𑄳𑄣𑄬', 'ce': 'кпелле', 'chr': 'ᏇᎴ', 'ckb': 'کپێلێ', 'cs': 'kpelle', 'cy': 'Kpelle', 'da': 'kpelle', 'de': 'Kpelle', 'dsb': 'kpellešćina', 'el': 'Κπέλε', 'el-polyton': 'Κπέλε', 'en': 'Kpelle', 'es': 'kpelle', 'et': 'kpelle', 'eu': 'kpelleera', 'fa': 'کپله\u200cای', 'ff-Adlm': '𞤘𞤫𞤪𞤧𞤫𞥅𞤪𞤫', 'fi': 'kpelle', 'fil': 'Kpelle', 'fo': 'kpelle', 'fr': 'kpellé', 'fy': 'Kpelle', 'ga': 'Caipeilis', 'gd': 'Kpelle', 'gl': 'kpelle', 'gsw': 'Kpelle-Schpraach', 'gu': 'ક્પેલ્લે', 'ha': 'Kpelle', 'he': 'קפלה', 'hi': 'क्पेल', 'hi-Latn': 'Kpelle', 'hr': 'kpelle', 'hsb': 'kpellešćina', 'ht': 'kpellé', 'hu': 'kpelle', 'hy': 'կպելլեերեն', 'ia': 'kpelle', 'id': 'Kpelle', 'ig': 'Kpelle', 'is': 'kpelle', 'it': 'kpelle', 'ja': 'クペレ語', 'jv': 'Kpelle', 'ka': 'კპელე', 'kgp': 'kepere', 'kk': 'кпелле тілі', 'kk-Cyrl': 'кпелле тілі', 'km': 'គ្លីប', 'kn': 'ಕಪೆಲ್ಲೆ', 'ko': '크펠레어', 'kok': 'पेल्ले', 'kok-Deva': 'पेल्ले', 'kpe': 'Kpɛlɛɛ', 'ks': 'کَپیلی', 'ks-Arab': 'کَپیلی', 'ksh': 'Kpäle', 'ku': 'kpelleyî', 'ky': 'кпеллече', 'lb': 'Kpelle-Sprooch', 'lo': 'ກາແປຣ', 'lt': 'kpelių', 'lv': 'kpellu', 'mai': 'कपैले', 'mi': 'Kepēre', 'mk': 'кпеле', 'ml': 'പെൽ', 'mn': 'кпелле', 'mr': 'क्पेल्ले', 'ms': 'Kpelle', 'mt': 'Kpelle', 'my': 'ကပ်ပဲလ်', 'nb': 'kpelle', 'ne': 'क्पेल्ले', 'nl': 'Kpelle', 'nn': 'kpelle', 'no': 'kpelle', 'or': 'କୈପେଲେ', 'pa': 'ਕਪੇਲ', 'pa-Guru': 'ਕਪੇਲ', 'pcm': 'Kpẹllẹ Lángwej', 'pl': 'kpelle', 'ps': 'کیلي', 'pt': 'kpelle', 'qu': 'Kpelle Simi', 'rm': 'kpelle', 'ro': 'kpelle', 'ru': 'кпелле', 'sc': 'kpelle', 'sd': 'ڪپيل', 'sd-Arab': 'ڪپيل', 'si': 'ක්පෙලෙ', 'sk': 'kpelle', 'sl': 'kpelejščina', 'smn': 'kpellekielâ', 'so': 'Kpelle', 'sq': 'kpeleisht', 'sr': 'кпеле', 'sr-Cyrl': 'кпеле', 'sr-Latn': 'kpele', 'sv': 'kpelle', 'sw': 'Kikpelle', 'ta': 'க்பெல்லே', 'te': 'పెల్లే', 'th': 'กาแปล', 'ti': 'ክፐለ', 'tk': 'kpelle dili', 'to': 'lea fakakepele', 'tr': 'Kpelle dili', 'ug': 'كىپەللېچە', 'uk': 'кпеллє', 'ur': 'کیپیلّے', 'uz': 'kpelle', 'uz-Latn': 'kpelle', 'vec': 'kpele', 'vi': 'Tiếng Kpelle', 'xnr': 'क्पेल', 'yo': 'Èdè Pele', 'yrl': 'kipere', 'yue': '克佩列文', 'yue-Hans': '克佩列文', 'yue-Hant': '克佩列文', 'zh': '克佩列语', 'zh-Hans': '克佩列语', 'zh-Hant': '克佩列文', 'zu': 'isi-Kpelle'},
    'kpf': {'en': 'Komba'},
    'kpg': {'en': 'Kapingamarangi'},
    'kph': {'en': 'Kplang'},
    'kpi': {'en': 'Kofei'},
    'kpj': {'en': 'Karajá'},
    'kpk': {'en': 'Kpan'},
    'kpl': {'en': 'Kpala'},
    'kpm': {'en': 'Koho'},
    'kpn': {'en': 'Kepkiriwát'},
    'kpo': {'en': 'Ikposo'},
    'kpp': {'en': 'Paku Karen'},
    'kpq': {'en': 'Korupun-Sela'},
    'kpr': {'en': 'Korafe-Yegha'},
    'kps': {'en': 'Tehit'},
    'kpt': {'en': 'Karata'},
    'kpu': {'en': 'Kafoa'},
    'kpv': {'en': 'Komi-Zyrian'},
    'kpw': {'en': 'Kobon'},
    'kpx': {'en': 'Mountain Koiali'},
    'kpy': {'en': 'Koryak'},
    'kpz': {'en': 'Kupsabiny'},
    'kqa': {'en': 'Mum'},
    'kqb': {'en': 'Kovai'},
    'kqc': {'en': 'Doromu-Koki'},
    'kqd': {'en': 'Koy Sanjaq Surat'},
    'kqe': {'en': 'Kalagan'},
    'kqf': {'en': 'Kakabai'},
    'kqg': {'en': 'Khe'},
    'kqh': {'en': 'Kisankasa'},
    'kqi': {'en': 'Koitabu'},
    'kqj': {'en': 'Koromira'},
    'kqk': {'en': 'Kotafon Gbe'},
    'kql': {'en': 'Kyenele'},
    'kqm': {'en': 'Khisa'},
    'kqn': {'en': 'Kaonde'},
    'kqo': {'en': 'Eastern Krahn'},
    'kqp': {'en': 'Kimré'},
    'kqq': {'en': 'Krenak'},
    'kqr': {'en': 'Kimaragang'},
    'kqs': {'en': 'Northern Kissi'},
    'kqt': {'en': 'Klias River Kadazan'},
    'kqu': {'en': 'Seroa'},
    'kqv': {'en': 'Okolod'},
    'kqw': {'en': 'Kandas'},
    'kqx': {'en': 'Mser'},
    'kqy': {'en': 'Koorete'},
    'kqz': {'en': 'Korana'},
    'kr': {'af': 'Kanuri', 'am': 'ካኑሪ', 'ar': 'الكانوري', 'as': 'কানুৰি', 'ast': 'kanuri', 'az': 'kanuri', 'az-Cyrl': 'канури', 'az-Latn': 'kanuri', 'be': 'кануры', 'be-tarask': 'кануры', 'bg': 'канури', 'bgn': 'کانوری', 'bn': 'কানুরি', 'br': 'kanouri', 'brx': 'कनुरी', 'bs': 'kanuri', 'bs-Cyrl': 'канури', 'bs-Latn': 'kanuri', 'ca': 'kanuri', 'ca-ES-valencia': 'kanuri', 'ccp': '𑄇𑄚𑄪𑄢𑄨', 'ce': 'канури', 'chr': 'ᎧᏄᎵ', 'ckb': 'کانووری', 'cs': 'kanuri', 'cy': 'Canwri', 'da': 'kanuri', 'de': 'Kanuri', 'dsb': 'kanurišćina', 'el': 'Κανούρι', 'el-polyton': 'Κανούρι', 'en': 'Kanuri', 'es': 'kanuri', 'et': 'kanuri', 'eu': 'kanuriera', 'fa': 'کانوریایی', 'ff-Adlm': '𞤑𞤮𞥅𞤤𞤫𞥅𞤪𞤫', 'fi': 'kanuri', 'fil': 'Kanuri', 'fo': 'kanuri', 'fr': 'kanouri', 'fy': 'Kanuri', 'ga': 'Canúiris', 'gd': 'Kanuri', 'gl': 'kanuri', 'gsw': 'Kanuri-Schpraach', 'gu': 'કનુરી', 'ha': 'Kanuri', 'he': 'קאנורי', 'hi': 'कनुरी', 'hi-Latn': 'Kanuri', 'hr': 'kanuri', 'hsb': 'kanurišćina', 'ht': 'kanouri', 'hu': 'kanuri', 'hy': 'կանուրի', 'ia': 'kanuri', 'id': 'Kanuri', 'ig': 'Kanuri', 'is': 'kanúrí', 'it': 'kanuri', 'ja': 'カヌリ語', 'jv': 'Kanuri', 'ka': 'კანური', 'kgp': 'kanũri', 'kk': 'канури тілі', 'kk-Cyrl': 'канури тілі', 'km': 'កានូរី', 'kn': 'ಕನುರಿ', 'ko': '칸누리어', 'kok': 'कानुरी', 'kok-Deva': 'कानुरी', 'ks': 'کَنوٗری', 'ks-Arab': 'کَنوٗری', 'ksh': 'Kanuhresch', 'ku': 'kanurîyî', 'ky': 'кануриче', 'lb': 'Kanuri-Sprooch', 'lo': 'ຄານຸລິ', 'lt': 'kanurių', 'lv': 'kanuru', 'mai': 'कनुरी', 'mi': 'Kanuri', 'mk': 'канури', 'ml': 'കനൂറി', 'mn': 'канури', 'mr': 'कनुरी', 'ms': 'Kanuri', 'mt': 'Kanuri', 'my': 'ကနူရီ', 'nb': 'kanuri', 'ne': 'कानुरी', 'nl': 'Kanuri', 'nn': 'kanuri', 'no': 'kanuri', 'or': 'କନୁରୀ', 'pa': 'ਕਨੂਰੀ', 'pa-Guru': 'ਕਨੂਰੀ', 'pcm': 'Kánurí Lángwej', 'pl': 'kanuri', 'ps': 'کنوری', 'pt': 'canúri', 'qu': 'Kanuri Simi', 'rm': 'kanuri', 'ro': 'kanuri', 'ru': 'канури', 'sc': 'kanuri', 'sd': 'ڪنوري', 'sd-Arab': 'ڪنوري', 'si': 'කනුරි', 'sk': 'kanurijčina', 'sl': 'kanurščina', 'smn': 'kanurikielâ', 'so': 'Kanuri', 'sq': 'kanurisht', 'sr': 'канури', 'sr-Cyrl': 'канури', 'sr-Latn': 'kanuri', 'sv': 'kanuri', 'sw': 'Kikanuri', 'ta': 'கனுரி', 'te': 'కానురి', 'tg': 'канурӣ', 'th': 'คานูรี', 'ti': 'ካኑሪ', 'tk': 'kanuri', 'to': 'lea fakakanuli', 'tr': 'Kanuri dili', 'tt': 'канури', 'ug': 'كانۇرچە', 'uk': 'канурі', 'ur': 'کنوری', 'uz': 'kanuri', 'uz-Cyrl': 'канури', 'uz-Latn': 'kanuri', 'vec': 'canuri', 'vi': 'Tiếng Kanuri', 'wo': 'Kanuri', 'xnr': 'कनुरी', 'yo': 'Èdè Kanuri', 'yrl': 'kanúri', 'yue': '卡努裡文', 'yue-Hans': '卡努里文', 'yue-Hant': '卡努裡文', 'zh': '卡努里语', 'zh-Hans': '卡努里语', 'zh-Hant': '卡努里文', 'zu': 'isi-Kanuri'},
    'kra': {'en': 'Kumhali'},
    'krb': {'en': 'Karkin'},
    'krc': {'af': 'Karachay-Balkar', 'am': 'ካራቻይ-ባልካር', 'ar': 'الكاراتشاي-بالكار', 'as': 'কাৰাচে-বাল্কাৰ', 'ast': 'karachay-balkar', 'az': 'qaraçay-balkar', 'az-Cyrl': 'гарачај-балкар', 'az-Latn': 'qaraçay-balkar', 'be': 'карачай-балкарская', 'be-tarask': 'карачай-балкарская', 'bg': 'карачай-балкарски', 'bgn': 'قره\u200cچایی‐بالکاری', 'bn': 'কারচে-বাল্কার', 'br': 'karatchay-balkar', 'brx': 'कराचय-बलकार', 'bs': 'karačaj-balkar', 'bs-Cyrl': 'карачај-балкар', 'bs-Latn': 'karačaj-balkar', 'ca': 'karatxai-balkar', 'ca-ES-valencia': 'karatxai-balkar', 'ccp': '𑄇𑄢𑄴𑄌𑄮-𑄝𑄣𑄴𑄇𑄢𑄴', 'ce': 'кхарачойн-балкхаройн', 'chr': 'ᎧᎳᏣᏱ-ᏆᎵᎧᎵ', 'ckb': 'کاراچای بالکار', 'cs': 'karačajevo-balkarština', 'cy': 'Karachay-Balkar', 'da': 'karatjai-balkar', 'de': 'Karatschaiisch-Balkarisch', 'dsb': 'karachay-balkaršćina', 'el': 'Καρατσάι-Μπαλκάρ', 'el-polyton': 'Καρατσάι-Μπαλκάρ', 'en': 'Karachay-Balkar', 'es': 'karachay-balkar', 'et': 'karatšai-balkaari', 'eu': 'karachayera-balkarera', 'fa': 'قره\u200cچایی‐بالکاری', 'ff-Adlm': '𞤑𞤢𞤪𞤢𞤧𞤢𞤴-𞤄𞤢𞤤𞤳𞤢𞥄𞤪𞤫', 'fi': 'karatšai-balkaari', 'fil': 'Karachay-Balkar', 'fo': 'karachay-balkar', 'fr': 'karatchaï balkar', 'fy': 'Karachay-Balkar', 'ga': 'Caraicí-Balcáiris', 'gd': 'Karachay-Balkar', 'gl': 'carachaio-bálcara', 'gsw': 'Karatschaiisch-Balkarisch', 'gu': 'કરાચય-બલ્કાર', 'ha': 'Karachay-Balkar', 'he': 'קראצ׳י-בלקר', 'hi': 'कराचय-बल्कार', 'hi-Latn': 'Karachay-Balkar', 'hr': 'karachay-balkar', 'hsb': 'karachay-balkaršćina', 'ht': 'karatchaï balkar', 'hu': 'karacsáj-balkár', 'hy': 'կարաչայ-բալկարերեն', 'ia': 'karachay-balkaro', 'id': 'Karachai Balkar', 'ig': 'Karachay-Balkar', 'is': 'karasaíbalkar', 'it': 'karachay-Balkar', 'ja': 'カラチャイ・バルカル語', 'jv': 'Karachai-Balkar', 'ka': 'ყარაჩაულ-ბალყარული', 'kgp': 'karasaj-markar', 'kk': 'қарашай-балқар тілі', 'kk-Cyrl': 'қарашай-балқар тілі', 'km': 'ការ៉ាឆាយបាល់កា', 'kn': 'ಕರಚಯ್-ಬಲ್ಕಾರ್', 'ko': '카라챠이-발카르어', 'kok': 'कराची-बाल्कर', 'kok-Deva': 'कराची-बाल्कर', 'ks': 'کراچیے بَلکار', 'ks-Arab': 'کراچیے بَلکار', 'ksh': 'Karatschaj-Balkahresch-Törkesch', 'ku': 'karaçay-balkarî', 'ky': 'карачай-балкарча', 'lb': 'Karatschaiesch-Balkaresch', 'lo': 'ກາຣາໄຊ-ບານກາ', 'lt': 'karačiajų balkarijos', 'lv': 'karačaju un balkāru', 'mai': 'करचय-बालकर', 'mi': 'Karatai-Pāka', 'mk': 'карачаевско-балкарски', 'ml': 'കരാചൈ-ബാൽകാർ', 'mn': 'карачай-балкар', 'mr': 'कराचय-बाल्कर', 'ms': 'Karachay-Balkar', 'mt': 'Karachay-Balkar', 'my': 'ကရာချေး-ဘာကာ', 'nb': 'karatsjajbalkarsk', 'ne': 'काराचाय-बाल्कर', 'nl': 'Karatsjaj-Balkarisch', 'nn': 'karatsjaiisk-balkarsk', 'no': 'karatsjajbalkarsk', 'or': 'କରାଚୟ-ବଲ୍କାର', 'os': 'бӕлхъӕрон', 'pa': 'ਕਰਾਚੇ ਬਲਕਾਰ', 'pa-Guru': 'ਕਰਾਚੇ ਬਲਕਾਰ', 'pcm': 'Karáchei-Bálkar', 'pl': 'karaczajsko-bałkarski', 'ps': 'کراچی بالکر', 'pt': 'karachay-balkar', 'qu': 'Karachay-Balkar Simi', 'rm': 'karachay-balkar', 'ro': 'karaceai-balkar', 'ru': 'карачаево-балкарский', 'sc': 'caraciai-balcaru', 'sd': 'ڪراچي بالڪر', 'sd-Arab': 'ڪراچي بالڪر', 'si': 'කරන්චි-බාකර්', 'sk': 'karačajevsko-balkarčina', 'sl': 'karačaj-balkarščina', 'smn': 'karachai-balkarkielâ', 'so': 'Karachay-Balkar', 'sq': 'karaçaj-balkarisht', 'sr': 'карачајско-балкарски', 'sr-Cyrl': 'карачајско-балкарски', 'sr-Latn': 'karačajsko-balkarski', 'sv': 'karachay-balkar', 'sw': 'Kikarachay-Balkar', 'ta': 'கராசே-பல்கார்', 'te': 'కరచే-బల్కార్', 'th': 'คาราไช-บัลคาร์', 'ti': 'ካራቻይ-ባልካርኛ', 'tk': 'karaçaý-balkar dili', 'to': 'lea fakakalate-palakili', 'tr': 'Karaçay-Balkarca', 'ug': 'قاراچاي-بالقارچە', 'uk': 'карачаєво-балкарська', 'ur': 'کراچے بالکر', 'uz': 'qorachoy-bolqor', 'uz-Latn': 'qorachoy-bolqor', 'vec': 'karaciài balkar', 'vi': 'Tiếng Karachay-Balkar', 'xnr': 'कराचय-बल्कार', 'yo': 'Èdè Karasha-Baka', 'yrl': 'karaxai-bauká', 'yue': '卡拉柴-包爾卡爾文', 'yue-Hans': '卡拉柴-包尔卡尔文', 'yue-Hant': '卡拉柴-包爾卡爾文', 'zh': '卡拉恰伊巴尔卡尔语', 'zh-Hans': '卡拉恰伊巴尔卡尔语', 'zh-Hant': '卡拉柴-包爾卡爾文', 'zu': 'isi-Karachay-Balkar'},
    'krd': {'en': 'Kairui-Midiki'},
    'kre': {'en': 'Panará'},
    'krf': {'en': 'Koro (Vanuatu)'},
    'krh': {'en': 'Kurama'},
    'kri': {'ast': 'krio', 'br': 'krio', 'bs': 'krio', 'bs-Latn': 'krio', 'ca': 'krio', 'ca-ES-valencia': 'krio', 'cs': 'krio', 'de': 'Krio', 'en': 'Krio', 'et': 'krio', 'fi': 'krio', 'fr': 'krio', 'gd': 'Krio', 'hi-Latn': 'Krio', 'ht': 'krio', 'id': 'Krio', 'ja': 'クリオ語', 'lb': 'Krio', 'lt': 'krio', 'mk': 'крио', 'nb': 'krio', 'ne': 'क्रिओ', 'nl': 'Krio', 'nn': 'krio', 'no': 'krio', 'pl': 'krio', 'sr': 'крио', 'sr-Cyrl': 'крио', 'sr-Latn': 'krio', 'sv': 'krio', 'th': 'คริโอ', 'to': 'lea fakakilio', 'tr': 'Krio', 'yue': '塞拉利昂克裏奧爾文', 'yue-Hans': '塞拉利昂克里奥尔文', 'yue-Hant': '塞拉利昂克裏奧爾文', 'zh-Hant': '塞拉利昂克裏奧爾文'},
    'krj': {'ast': 'kinaray-a', 'cs': 'kinaraj-a', 'de': 'Kinaray-a', 'en': 'Kinaray-a', 'et': 'kinaraia', 'fi': 'kinaray-a', 'fr': 'kinaray-a', 'gd': 'Kinaray-a', 'hi-Latn': 'Kinaray-a', 'ht': 'kinaray-a', 'ja': 'キナライア語', 'lb': 'Kinaray-a', 'lt': 'kinaray-a', 'mk': 'кинарајски', 'nb': 'kinaray-a', 'ne': 'किनाराय-ए', 'nl': 'Kinaray-a', 'nn': 'kinaray-a', 'no': 'kinaray-a', 'pl': 'kinaraya', 'sv': 'kinaray-a', 'th': 'กินารายอา', 'to': 'lea fakakinaraiā', 'tr': 'Kinaray-a', 'yue': '基那來阿文', 'yue-Hans': '基那来阿文', 'yue-Hant': '基那來阿文', 'zh-Hant': '基那來阿文'},
    'krk': {'en': 'Kerek'},
    'krl': {'af': 'Karelies', 'am': 'ካረሊያን', 'ar': 'الكاريلية', 'as': 'কেৰেলিয়ান', 'ast': 'karelianu', 'az': 'karel', 'az-Cyrl': 'карел', 'az-Latn': 'karel', 'be': 'карэльская', 'be-tarask': 'карэльская', 'bg': 'карелски', 'bgn': 'کاریلینی', 'bn': 'কারেলিয়ান', 'br': 'karelieg', 'brx': 'करेलियान', 'bs': 'karelijski', 'bs-Cyrl': 'карелијски', 'bs-Latn': 'karelijski', 'ca': 'carelià', 'ca-ES-valencia': 'carelià', 'ccp': '𑄇𑄢𑄬𑄣𑄨𑄠𑄚𑄴', 'ce': 'карелийн', 'chr': 'ᎧᎴᎵᎠᏂ', 'ckb': 'کارێلی', 'cs': 'karelština', 'cy': 'Careleg', 'da': 'karelsk', 'de': 'Karelisch', 'dsb': 'karelianšćina', 'el': 'Καρελικά', 'el-polyton': 'Καρελικά', 'en': 'Karelian', 'es': 'carelio', 'et': 'karjala', 'eu': 'kareliera', 'fa': 'کاریلیانی', 'ff-Adlm': '𞤑𞤢𞤪𞤫𞤤𞤭𞤢𞤲𞤪𞤫', 'fi': 'karjala', 'fil': 'Karelian', 'fo': 'karelskt', 'fr': 'carélien', 'fy': 'Karelysk', 'ga': 'Cairéilis', 'gd': 'Cairealais', 'gl': 'carelio', 'gsw': 'Karelisch', 'gu': 'કરેલિયન', 'ha': 'Kareliyanci', 'he': 'קארלית', 'hi': 'करेलियन', 'hi-Latn': 'Karelian', 'hr': 'karelijski', 'hsb': 'karelianšćina', 'ht': 'carélien', 'hu': 'karelai', 'hy': 'կարելերեն', 'ia': 'careliano', 'id': 'Karelia', 'ig': 'Karelian', 'is': 'karélska', 'it': 'careliano', 'ja': 'カレリア語', 'jv': 'Karelian', 'ka': 'კარელიური', 'kgp': 'karérijo', 'kk': 'карель тілі', 'kk-Cyrl': 'карель тілі', 'km': 'ការីលា', 'kn': 'ಕರೇಲಿಯನ್', 'ko': '카렐리야어', 'kok': 'कारेलियन', 'kok-Deva': 'कारेलियन', 'ks': 'کَریلِیَن', 'ks-Arab': 'کَریلِیَن', 'ksh': 'Karehlesch', 'ku': 'karelî', 'ky': 'карелче', 'lb': 'Karelesch', 'lo': 'ກາເຣລຽນ', 'lt': 'karelų', 'lv': 'karēļu', 'mai': 'केरैलियन', 'mi': 'Kareriana', 'mk': 'карелски', 'ml': 'കരീലിയൻ', 'mn': 'карель', 'mr': 'करेलियन', 'ms': 'Karelian', 'mt': 'Kareljan', 'my': 'ကာရီလီယန်', 'nb': 'karelsk', 'ne': 'करेलियन', 'nl': 'Karelisch', 'nn': 'karelsk', 'no': 'karelsk', 'or': 'କାରେଲିୟାନ୍', 'pa': 'ਕਰੀਲੀਅਨ', 'pa-Guru': 'ਕਰੀਲੀਅਨ', 'pcm': 'Karẹ́lian', 'pl': 'karelski', 'ps': 'کاریلین', 'pt': 'carélio', 'qu': 'Karelian Simi', 'rm': 'carelian', 'ro': 'kareliană', 'ru': 'карельский', 'sc': 'carelianu', 'sd': 'ڪريلئين', 'sd-Arab': 'ڪريلئين', 'se': 'gárjilgiella', 'si': 'කැරෙලියන්', 'sk': 'karelčina', 'sl': 'karelščina', 'smn': 'kärjilkielâ', 'so': 'Karelian', 'sq': 'karelianisht', 'sr': 'карелски', 'sr-Cyrl': 'карелски', 'sr-Latn': 'karelski', 'sv': 'karelska', 'sw': 'Karjala', 'ta': 'கரேலியன்', 'te': 'కరేలియన్', 'th': 'คารีเลียน', 'ti': 'ካረልኛ', 'tk': 'karel dili', 'to': 'lea fakakalelia', 'tr': 'Karelyaca', 'ug': 'كارەلچە', 'uk': 'карельська', 'ur': 'کیرلین', 'uz': 'karel', 'uz-Latn': 'karel', 'vec': 'carelian', 'vi': 'Tiếng Karelian', 'xnr': 'करेलियन', 'yo': 'Èdè Karelia', 'yrl': 'karériu', 'yue': '卡累利阿文', 'yue-Hans': '卡累利阿文', 'yue-Hant': '卡累利阿文', 'zh': '卡累利阿语', 'zh-Hans': '卡累利阿语', 'zh-Hant': '卡累利阿文', 'zu': 'isi-Karelian'},
    'krm': {'en': 'Krim'},
    'krn': {'en': 'Sapo'},
    'kro': {'en': 'Kru languages'},
    'krp': {'en': 'Korop'},
    'krr': {'en': 'Krung'},
    'krs': {'en': 'Gbaya (Sudan)'},
    'krt': {'en': 'Tumari Kanuri'},
    'kru': {'af': 'Kurukh', 'am': 'ኩሩክ', 'ar': 'الكوروخ', 'as': 'কুৰুখ', 'ast': 'kurukh', 'az': 'kurux', 'az-Cyrl': 'курух', 'az-Latn': 'kurux', 'be': 'курух', 'be-tarask': 'курух', 'bg': 'курук', 'bgn': 'کوروخی', 'bn': 'কুরুখ', 'br': 'kurukh', 'brx': 'कुरुख', 'bs': 'kuruški', 'bs-Cyrl': 'курукх', 'bs-Latn': 'kuruški', 'ca': 'kurukh', 'ca-ES-valencia': 'kurukh', 'ccp': '𑄇𑄪𑄢𑄪𑄇𑄴', 'ce': 'курух', 'chr': 'ᎫᎷᎩ', 'ckb': 'کوورووخ', 'cs': 'kuruchština', 'cy': 'Kurukh', 'da': 'kurukh', 'de': 'Oraon', 'dsb': 'kurukhšćina', 'el': 'Κουρούχ', 'el-polyton': 'Κουρούχ', 'en': 'Kurukh', 'es': 'kurukh', 'et': 'kuruhhi', 'eu': 'kurukhera', 'fa': 'کوروخی', 'ff-Adlm': '𞤑𞤵𞤪𞤵𞤿𞤵𞥅𞤪𞤫', 'fi': 'kurukh', 'fil': 'Kurukh', 'fo': 'kurukh', 'fr': 'kouroukh', 'fy': 'Kurukh', 'ga': 'Curúicis', 'gd': 'Kurukh', 'gl': 'kurukh', 'gsw': 'Oraon-Schpraach', 'gu': 'કુરૂખ', 'ha': 'Kurukh', 'he': 'קורוק', 'hi': 'कुरूख', 'hi-Latn': 'Kurukh', 'hr': 'kuruški', 'hsb': 'kurukhšćina', 'ht': 'kouroukh', 'hu': 'kuruh', 'hy': 'կուրուխ', 'ia': 'kurukh', 'id': 'Kuruk', 'ig': 'Kurukh', 'is': 'kúrúk', 'it': 'kurukh', 'ja': 'クルク語', 'jv': 'Kuruk', 'ka': 'კურუქი', 'kgp': 'kurug', 'kk': 'курух тілі', 'kk-Cyrl': 'курух тілі', 'km': 'គូរូក', 'kn': 'ಕುರುಖ್', 'ko': '쿠르크어', 'kok': 'कुरुख', 'kok-Deva': 'कुरुख', 'ks': 'کُرُکھ', 'ks-Arab': 'کُرُکھ', 'ksh': 'Korochesch', 'ku': 'kurukh', 'ky': 'курухча', 'lb': 'Oraon-Sprooch', 'lo': 'ກູຣູກ', 'lt': 'kuruk', 'lv': 'kuruhu', 'mai': 'कुरूक', 'mi': 'Kuruka', 'mk': 'курух', 'ml': 'കുരുഖ്', 'mn': 'курук', 'mr': 'कुरूख', 'ms': 'Kurukh', 'mt': 'Kurux', 'my': 'ကူရုပ်ခ်', 'nb': 'kurukh', 'ne': 'कुरुख', 'nl': 'Kurukh', 'nn': 'kurukh', 'no': 'kurukh', 'or': 'କୁରୁଖ', 'pa': 'ਕੁਰੁਖ', 'pa-Guru': 'ਕੁਰੁਖ', 'pcm': 'Kurúkh Lángwej', 'pl': 'kurukh', 'ps': 'کورخ', 'pt': 'kurukh', 'qu': 'Kurukh Simi', 'rm': 'kurukh', 'ro': 'kurukh', 'ru': 'курух', 'sc': 'kurukh', 'sd': 'ڪورخ', 'sd-Arab': 'ڪورخ', 'si': 'කුරුඛ්', 'sk': 'kuruchčina', 'sl': 'kuruk', 'smn': 'kurukhkielâ', 'so': 'Kurukh', 'sq': 'kurukisht', 'sr': 'курук', 'sr-Cyrl': 'курук', 'sr-Latn': 'kuruk', 'sv': 'kurukh', 'sw': 'Kurukh', 'ta': 'குருக்', 'te': 'కూరుఖ్', 'tg': 'курукс', 'th': 'กุรุข', 'ti': 'ኩሩክ', 'tk': 'kuruh dili', 'to': 'lea fakakuluki', 'tr': 'Kurukh dili', 'tt': 'курух', 'ug': 'كۇرۇخچە', 'uk': 'курукх', 'ur': 'کوروکھ', 'uz': 'kurux', 'uz-Latn': 'kurux', 'vec': 'kuruk', 'vi': 'Tiếng Kurukh', 'wo': 'Kuruks', 'xnr': 'कुरूख', 'yo': 'Èdè Kuruki', 'yrl': 'kuruki', 'yue': '庫魯科文', 'yue-Hans': '库鲁科文', 'yue-Hant': '庫魯科文', 'zh': '库鲁克语', 'zh-Hans': '库鲁克语', 'zh-Hant': '庫魯科文', 'zu': 'isi-Kurukh'},
    'krv': {'en': 'Kavet'},
    'krw': {'en': 'Western Krahn'},
    'krx': {'en': 'Karon'},
    'kry': {'en': 'Kryts'},
    'krz': {'en': 'Sota Kanum'},
    'ks': {'af': 'Kasjmiri', 'ak': 'Kahyimiɛ', 'am': 'ካሽሚርኛ', 'ar': 'الكشميرية', 'as': 'কাশ্মিৰী', 'ast': 'cachemirés', 'az': 'kəşmir', 'az-Arab': 'کشمیرجه', 'az-Cyrl': 'кәшмир', 'az-Latn': 'kəşmir', 'bal-Latn': 'Kashmiri', 'be': 'кашмірская', 'be-tarask': 'кашмірская', 'bg': 'кашмирски', 'bgn': 'کشمیری', 'bn': 'কাশ্মীরি', 'br': 'kashmiri', 'brx': 'काश्मिरी', 'bs': 'kašmirski', 'bs-Cyrl': 'кашмирски', 'bs-Latn': 'kašmirski', 'ca': 'caixmiri', 'ca-ES-valencia': 'caixmiri', 'ccp': '𑄇𑄌𑄴𑄟𑄨𑄢𑄨', 'ce': 'кашмири', 'chr': 'ᎧᏏᎻᎵ', 'ckb': 'کەشمیری', 'cs': 'kašmírština', 'cy': 'Cashmireg', 'da': 'kashmiri', 'de': 'Kaschmiri', 'dsb': 'kašmiršćina', 'dz': 'ཀཱཤ་མི་རི་ཁ', 'ee': 'kashmirgbe', 'el': 'Κασμιρικά', 'el-polyton': 'Κασμιρικά', 'en': 'Kashmiri', 'eo': 'kaŝmira', 'es': 'cachemir', 'et': 'kašmiiri', 'eu': 'kaxmirera', 'fa': 'کشمیری', 'ff-Adlm': '𞤑𞤢𞥃𞤥𞤭𞥅𞤪𞤫', 'fi': 'kašmiri', 'fil': 'Kashmiri', 'fo': 'kashmiri', 'fr': 'cachemiri', 'fy': 'Kasjmiri', 'ga': 'Caismíris', 'gd': 'Caismiris', 'gl': 'caxemirés', 'gsw': 'Kaschmirisch', 'gu': 'કાશ્મીરી', 'ha': 'Kashmiri', 'he': 'קשמירית', 'hi': 'कश्मीरी', 'hi-Latn': 'Kashmiri', 'hr': 'kašmirski', 'hsb': 'kašmiršćina', 'ht': 'cachemiri', 'hu': 'kasmíri', 'hy': 'քաշմիրերեն', 'ia': 'kashmiri', 'id': 'Kashmir', 'ig': 'Kashmiri', 'is': 'kasmírska', 'it': 'kashmiri', 'ja': 'カシミール語', 'jv': 'Kashmiri', 'ka': 'ქაშმირული', 'kea': 'kaxmira', 'kgp': 'kasemĩra', 'kk': 'кашмир тілі', 'kk-Cyrl': 'кашмир тілі', 'km': 'កាស្មៀរ', 'kn': 'ಕಾಶ್ಮೀರಿ', 'ko': '카슈미르어', 'kok': 'कश्मीरी', 'kok-Deva': 'कश्मीरी', 'ks': 'کٲشُر', 'ks-Arab': 'کٲشُر', 'ks-Deva': 'कॉशुर', 'ksh': 'Kaschmihresch', 'ku': 'keşmîrî', 'kxv': 'kasmīrī', 'kxv-Deva': 'कस्मीरी', 'kxv-Latn': 'kasmīrī', 'kxv-Orya': 'କାସ୍ମିର', 'kxv-Telu': 'కాస్మిరి', 'ky': 'кашмирче', 'lb': 'Kaschmiresch', 'lkt': 'Kashmir Iyápi', 'lld': 'kashmiri', 'lo': 'ຄາສເມຍຣິ', 'lrc': 'کأشمیری', 'lt': 'kašmyrų', 'lv': 'kašmiriešu', 'mai': 'कशमीरी', 'mi': 'Kahimiri', 'mk': 'кашмирски', 'ml': 'കശ്\u200cമീരി', 'mn': 'кашмир', 'mr': 'काश्मीरी', 'ms': 'Kashmir', 'mt': 'Kashmiri', 'my': 'ကက်ရှ်မီးယား', 'mzn': 'کشمیری', 'nb': 'kasjmiri', 'ne': 'कास्मिरी', 'nl': 'Kasjmiri', 'nn': 'kasjmiri', 'no': 'kasjmiri', 'nqo': 'ߞߊߛߑߡߙߌߞߊ߲', 'or': 'କାଶ୍ମିରୀ', 'pa': 'ਕਸ਼ਮੀਰੀ', 'pa-Guru': 'ਕਸ਼ਮੀਰੀ', 'pcm': 'Kashmírí Lángwej', 'pl': 'kaszmirski', 'ps': 'کشمیري', 'pt': 'caxemira', 'qu': 'Cachemir Simi', 'rm': 'kashmiri', 'ro': 'cașmiră', 'ru': 'кашмири', 'sc': 'kashmiri', 'sd': 'ڪشميري', 'sd-Arab': 'ڪشميري', 'si': 'කාෂ්මීර්', 'sk': 'kašmírčina', 'sl': 'kašmirščina', 'smn': 'kashmirkielâ', 'so': 'Kaashmiir', 'sq': 'kashmirisht', 'sr': 'кашмирски', 'sr-Cyrl': 'кашмирски', 'sr-Latn': 'kašmirski', 'sv': 'kashmiriska', 'sw': 'Kikashmiri', 'ta': 'காஷ்மிரி', 'te': 'కాశ్మీరి', 'tg': 'кашмирӣ', 'th': 'แคชเมียร์', 'ti': 'ካሽሚሪ', 'tk': 'kaşmiri dili', 'to': 'lea fakakāsimila', 'tr': 'Keşmir dili', 'tt': 'кашмири', 'ug': 'كەشمىرچە', 'uk': 'кашмірська', 'ur': 'کشمیری', 'uz': 'kashmircha', 'uz-Cyrl': 'кашмирча', 'uz-Latn': 'kashmircha', 'vec': 'cashmiri', 'vi': 'Tiếng Kashmir', 'wae': 'Kašmiriš', 'wo': 'Kashmiri', 'xnr': 'कश्मीरी', 'yo': 'Kaṣímirì', 'yrl': 'kaxemira', 'yue': '喀什米爾文', 'yue-Hans': '喀什米尔文', 'yue-Hant': '喀什米爾文', 'zh': '克什米尔语', 'zh-Hans': '克什米尔语', 'zh-Hant': '喀什米爾文', 'zu': 'isi-Kashmiri'},
    'ksa': {'en': 'Shuwa-Zamani'},
    'ksb': {'af': 'Shambala', 'am': 'ሻምባላ', 'ar': 'شامبالا', 'as': 'চাম্বালা', 'ast': 'shambala', 'az': 'şambala', 'az-Cyrl': 'шамбала', 'az-Latn': 'şambala', 'bal-Latn': 'Shambalá', 'be': 'шамбала', 'be-tarask': 'шамбала', 'bg': 'шамбала', 'bgn': 'شامبالای', 'bn': 'শাম্বালা', 'br': 'shambala', 'brx': 'शामबाला', 'bs': 'šambala', 'bs-Cyrl': 'шамбала', 'bs-Latn': 'šambala', 'ca': 'shambala', 'ca-ES-valencia': 'shambala', 'ccp': '𑄥𑄟𑄴𑄝𑄣', 'ce': 'шамбала', 'chr': 'ᏝᎻᏆᎸ', 'ckb': 'شامابالا', 'cs': 'šambala', 'cy': 'Shambala', 'da': 'shambala', 'de': 'Shambala', 'dsb': 'šambala', 'el': 'Σαμπάλα', 'el-polyton': 'Σαμπάλα', 'en': 'Shambala', 'es': 'shambala', 'et': 'šambala', 'eu': 'shambalera', 'fa': 'شامبالا', 'ff-Adlm': '𞤅𞤢𞤥𞤦𞤢𞤤𞤢𞥄𞤪𞤫', 'fi': 'shambala', 'fil': 'Shambala', 'fo': 'shambala', 'fr': 'shambala', 'fy': 'Shambala', 'ga': 'Shambala', 'gd': 'Shambala', 'gl': 'shambala', 'gu': 'શમ્બાલા', 'ha': 'Shambala', 'he': 'שמבאלה', 'hi': 'शम्बाला', 'hi-Latn': 'Shambala', 'hr': 'shambala', 'hsb': 'šambala', 'ht': 'shambala', 'hu': 'sambala', 'hy': 'շամբալա', 'ia': 'shambala', 'id': 'Shambala', 'ig': 'Shambala', 'is': 'sjambala', 'it': 'shambala', 'ja': 'サンバー語', 'jv': 'Shambala', 'ka': 'შამბალა', 'kgp': 'sỹmara', 'kk': 'шамбала тілі', 'kk-Cyrl': 'шамбала тілі', 'km': 'សាមបាឡា', 'kn': 'ಶಂಬಲ', 'ko': '샴발라어', 'kok': 'शांबाला', 'kok-Deva': 'शांबाला', 'ksb': 'Kishambaa', 'ku': 'shambala', 'ky': 'шамабалача', 'lb': 'Shambala', 'lo': 'ຊຳບາລ້າ', 'lrc': 'شامبالا', 'lt': 'šambalų', 'lv': 'šambalu', 'mai': 'शमबाला', 'mi': 'Hapāra', 'mk': 'шамбала', 'ml': 'ഷംഭാള', 'mn': 'шамбал', 'mr': 'शांबाला', 'ms': 'Shambala', 'mt': 'Shambala', 'my': 'ရှန်ဘာလာ', 'mzn': 'شامبالا', 'nb': 'shambala', 'ne': 'शाम्बाला', 'nl': 'Shambala', 'nn': 'shambala', 'no': 'shambala', 'nqo': 'ߛߊ߲ߓߟߊߞߊ߲', 'or': 'ଶାମବାଲା', 'pa': 'ਸ਼ੰਬਾਲਾ', 'pa-Guru': 'ਸ਼ੰਬਾਲਾ', 'pcm': 'Shámbala', 'pl': 'sambala', 'ps': 'شمبالا', 'pt': 'shambala', 'qu': 'Shambala Simi', 'rm': 'shambala', 'ro': 'shambala', 'ru': 'шамбала', 'sc': 'shambala', 'sd': 'شمبالا', 'sd-Arab': 'شمبالا', 'si': 'ශාම්බලා', 'sk': 'šambala', 'sl': 'šambala', 'smn': 'shambala', 'so': 'Shambaala', 'sq': 'shambalisht', 'sr': 'шамбала', 'sr-Cyrl': 'шамбала', 'sr-Latn': 'šambala', 'sv': 'kisambaa', 'sw': 'Kisambaa', 'ta': 'ஷம்பாலா', 'te': 'శంబాలా', 'th': 'ชัมบาลา', 'ti': 'ሻምባላ', 'tk': 'şambala dili', 'to': 'lea fakasiamipala', 'tr': 'Şambala', 'ug': 'شامبالاچە', 'uk': 'шамбала', 'ur': 'شامبالا', 'uz': 'shambala', 'uz-Cyrl': 'шамбала', 'uz-Latn': 'shambala', 'vec': 'shanbala', 'vi': 'Tiếng Shambala', 'xnr': 'शम्बाला', 'yo': 'Ṣáńbálà', 'yrl': 'xãbara', 'yue': '尚巴拉文', 'yue-Hans': '尚巴拉文', 'yue-Hant': '尚巴拉文', 'zh': '香巴拉语', 'zh-Hans': '香巴拉语', 'zh-Hant': '尚巴拉文', 'zu': 'isiShambala'},
    'ksc': {'en': 'Southern Kalinga'},
    'ksd': {'en': 'Kuanua'},
    'kse': {'en': 'Kuni'},
    'ksf': {'af': 'Bafia', 'am': 'ባፊያ', 'ar': 'لغة البافيا', 'as': 'বাফিয়া', 'ast': 'bafia', 'az': 'bafia', 'az-Arab': 'بافیاجا', 'az-Cyrl': 'бафиа', 'az-Latn': 'bafia', 'bal-Latn': 'Bapiá', 'be': 'бафія', 'be-tarask': 'бафія', 'bg': 'бафия', 'bgn': 'بافیا', 'bn': 'বাফিয়া', 'br': 'bafiaeg', 'brx': 'बाफिया', 'bs': 'bafia', 'bs-Cyrl': 'бафија', 'bs-Latn': 'bafia', 'ca': 'bafia', 'ca-ES-valencia': 'bafia', 'ccp': '𑄝𑄜𑄨𑄠', 'ce': 'бафиа', 'chr': 'ᏆᏫᎠ', 'ckb': 'بافیا', 'cs': 'bafia', 'cy': 'Baffia', 'da': 'bafia', 'de': 'Bafia', 'dsb': 'bafia', 'el': 'Μπάφια', 'el-polyton': 'Μπάφια', 'en': 'Bafia', 'es': 'bafia', 'et': 'bafia', 'eu': 'bafiera', 'fa': 'بافیایی', 'ff-Adlm': '𞤄𞤢𞤬𞤭𞤴𞤢𞥄𞤪𞤫', 'fi': 'bafia', 'fil': 'Bafia', 'fo': 'bafia', 'fr': 'bafia', 'fy': 'Bafia', 'ga': 'Baifiais', 'gd': 'Bafia', 'gl': 'bafia', 'gu': 'બફિયા', 'ha': 'Bafia', 'he': 'באפיה', 'hi': 'बफिआ', 'hi-Latn': 'Bafia', 'hr': 'bafia', 'hsb': 'bafia', 'ht': 'bafia', 'hu': 'bafia', 'hy': 'բաֆիա', 'ia': 'bafia', 'id': 'Bafia', 'ig': 'Bafịa', 'is': 'bafía', 'it': 'bafia', 'ja': 'バフィア語', 'jv': 'Bafia', 'ka': 'ბაფია', 'kea': 'bafia', 'kgp': 'mafija', 'kk': 'бафиа тілі', 'kk-Cyrl': 'бафиа тілі', 'km': 'បាហ្វៀ', 'kn': 'ಬಫಿಯ', 'ko': '바피아어', 'kok': 'बाफिया', 'kok-Deva': 'बाफिया', 'ksf': 'rikpa', 'ksh': 'Bafijahnesch', 'ku': 'bafyayî', 'ky': 'бафияча', 'lb': 'Bafia', 'lo': 'ບາເຟຍ', 'lrc': 'بافیا', 'lt': 'bafų', 'lv': 'bafiju', 'mai': 'बाफिया', 'mi': 'Pāwhia', 'mk': 'бафија', 'ml': 'ബാഫിയ', 'mn': 'бафиа', 'mr': 'बाफिया', 'ms': 'Bafia', 'mt': 'Bafia', 'my': 'ဘာဖီအာ', 'mzn': 'بافیایی', 'nb': 'bafia', 'ne': 'बाफिया', 'nl': 'Bafia', 'nn': 'bafia', 'no': 'bafia', 'nqo': 'ߓߊߝߌߞߊ߲', 'or': 'ବାଫଲା', 'pa': 'ਬਫ਼ੀਆ', 'pa-Guru': 'ਬਫ਼ੀਆ', 'pcm': 'Bafiá Lángwej', 'pl': 'bafia', 'ps': 'بفیا', 'pt': 'bafia', 'qu': 'Bafia Simi', 'rm': 'bafia', 'ro': 'bafia', 'ru': 'бафия', 'sc': 'bafia', 'sd': 'بافيا', 'sd-Arab': 'بافيا', 'si': 'බාෆියා', 'sk': 'bafia', 'sl': 'bafia', 'smn': 'bafia', 'so': 'Bafia', 'sq': 'bafianisht', 'sr': 'бафија', 'sr-Cyrl': 'бафија', 'sr-Latn': 'bafija', 'sv': 'bafia', 'sw': 'Kibafia', 'ta': 'பாஃபியா', 'te': 'బాఫియ', 'th': 'บาเฟีย', 'ti': 'ባፍያ', 'tk': 'bafia dili', 'to': 'lea fakapafia', 'tr': 'Bafia', 'ug': 'بافىياچە', 'uk': 'бафіа', 'ur': 'بافيا', 'uz': 'bafiya', 'uz-Cyrl': 'бафияча', 'uz-Latn': 'bafiya', 'vec': 'bafia', 'vi': 'Tiếng Bafia', 'xnr': 'बफिआ', 'yo': 'Èdè Báfíà', 'yrl': 'bafia', 'yue': '巴菲亞文', 'yue-Hans': '巴菲亚文', 'yue-Hant': '巴菲亞文', 'zh': '巴菲亚语', 'zh-Hans': '巴菲亚语', 'zh-Hant': '巴菲亞文', 'zu': 'isi-Bafia'},
    'ksg': {'en': 'Kusaghe'},
    'ksh': {'af': 'Keuls', 'am': 'ኮሎኝኛ', 'ar': 'لغة الكولونيان', 'as': 'কোলোগনিয়ান', 'ast': 'colonianu', 'az': 'köln', 'az-Cyrl': 'көлн', 'az-Latn': 'köln', 'bal-Latn': 'Kologni', 'be': 'кёльнская', 'be-tarask': 'кёльнская', 'bg': 'кьолнски', 'bgn': 'کولوگنی', 'bn': 'কলোগনিয়ান', 'br': 'koluneg', 'brx': 'कलगन्यान', 'bs': 'kelnski', 'bs-Cyrl': 'келнски', 'bs-Latn': 'kelnski', 'ca': 'kölsch', 'ca-ES-valencia': 'kölsch', 'ccp': '𑄇𑄣𑄴𑄥𑄧', 'ce': 'коьлнийн', 'chr': 'ᎪᎶᏂᎠᏂ', 'ckb': 'کۆلۆنی', 'cs': 'kolínština', 'cy': 'Cwleneg', 'da': 'kölsch', 'de': 'Kölsch', 'dsb': 'kelnšćina', 'el': 'Κολωνικά', 'el-polyton': 'Κολωνικά', 'en': 'Colognian', 'es': 'kölsch', 'et': 'kölni', 'eu': 'koloniera', 'fa': 'کولش', 'ff-Adlm': '𞤑𞤮𞤤𞤮𞤺𞤲𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'kölsch', 'fil': 'Colognian', 'fo': 'kølnskt', 'fr': 'kölsch', 'fy': 'Kölsch', 'ga': 'Coilsis', 'gd': 'Gearmailtis Chologne', 'gl': 'kölsch', 'gu': 'કોલોગ્નિયન', 'ha': 'Colognian', 'he': 'קולוניאן', 'hi': 'कोलोनियाई', 'hi-Latn': 'Colognian', 'hr': 'kelnski', 'hsb': 'kelnšćina', 'ht': 'kölsch', 'hu': 'kölsch', 'hy': 'քյոլներեն', 'ia': 'coloniese', 'id': 'Dialek Kolsch', 'ig': 'Colognian', 'is': 'kölníska', 'it': 'coloniese', 'ja': 'ケルン語', 'jv': 'Colonia', 'ka': 'კიოლში', 'kea': 'kolonhanu', 'kgp': 'kárysi', 'kk': 'кёльн тілі', 'kk-Cyrl': 'кёльн тілі', 'km': 'កូឡូញ', 'kn': 'ಕಲೊಗ್ನಿಯನ್', 'ko': '콜로그니안어', 'kok': 'कोलोनियन', 'kok-Deva': 'कोलोनियन', 'ksh': 'Kölsch', 'ku': 'rîpwarî', 'ky': 'колоньяча', 'lb': 'Kölsch', 'lo': 'ໂຄລອກນຽນ', 'lt': 'kolognų', 'lv': 'Ķelnes vācu', 'mai': 'कोलोनियन', 'mi': 'Korōniana', 'mk': 'колоњски', 'ml': 'കൊളോണിയൻ', 'mn': 'кёльш', 'mr': 'कोलोग्नियन', 'ms': 'Colognian', 'mt': 'Kolonjan', 'my': 'ကိုလိုနီယန်း', 'nb': 'kølnsk', 'ne': 'कोलोग्नियाली', 'nl': 'Kölsch', 'nn': 'kølnsk', 'no': 'kølnsk', 'nqo': 'ߞߐߟߑߗߌߞߊ߲', 'or': 'କୋଲୋବନିୟ', 'pa': 'ਕਲੋਨੀਅਨ', 'pa-Guru': 'ਕਲੋਨੀਅਨ', 'pcm': 'Kọlónián Lángwej', 'pl': 'gwara kolońska', 'ps': 'کولوګنيايي', 'pt': 'kölsch', 'qu': 'Kölsch Simi', 'rm': 'colognais', 'ro': 'kölsch', 'ru': 'кёльнский', 'sc': 'coloniesu', 'sd': 'ڪلونئين', 'sd-Arab': 'ڪلونئين', 'si': 'කොලොග්නියන්', 'sk': 'kolínčina', 'sl': 'kölnsko narečje', 'smn': 'kölnkielâ', 'so': 'Kologniyaan', 'sq': 'këlnisht', 'sr': 'келнски', 'sr-Cyrl': 'келнски', 'sr-Latn': 'kelnski', 'sv': 'kölniska', 'sw': 'Kicologne', 'ta': 'கொலோக்னியன்', 'te': 'కొలోనియన్', 'th': 'โคโลญ', 'ti': 'ኮሎግኒያን', 'tk': 'keln dili', 'to': 'lea fakakolongia', 'tr': 'Köln lehçesi', 'ug': 'كولىشچە', 'uk': 'кельнська', 'ur': 'کولوگنیائی', 'uz': 'kyoln', 'uz-Cyrl': 'кёлнча', 'uz-Latn': 'kyoln', 'vec': 'colognian', 'vi': 'Tiếng Cologne', 'xnr': 'कोलोनियाई', 'yo': 'Èdè Colognian', 'yrl': 'kurixi', 'yue': '科隆文', 'yue-Hans': '科隆文', 'yue-Hant': '科隆文', 'zh': '科隆语', 'zh-Hans': '科隆语', 'zh-Hant': '科隆文', 'zu': 'isi-Colognian'},
    'ksi': {'en': 'Krisa'},
    'ksj': {'en': 'Uare'},
    'ksk': {'en': 'Kansa'},
    'ksl': {'en': 'Kumalu'},
    'ksm': {'en': 'Kumba'},
    'ksn': {'en': 'Kasiguranin'},
    'kso': {'en': 'Kofa'},
    'ksp': {'en': 'Kaba'},
    'ksq': {'en': 'Kwaami'},
    'ksr': {'en': 'Borong'},
    'kss': {'en': 'Southern Kisi'},
    'kst': {'en': 'Winyé'},
    'ksu': {'en': 'Khamyang'},
    'ksv': {'en': 'Kusu'},
    'ksw': {'en': "S'gaw Karen"},
    'ksx': {'en': 'Kedang'},
    'ksy': {'en': 'Kharia Thar'},
    'ksz': {'en': 'Kodaku'},
    'kta': {'en': 'Katua'},
    'ktb': {'en': 'Kambaata'},
    'ktc': {'en': 'Kholok'},
    'ktd': {'en': 'Kokata'},
    'kte': {'en': 'Nubri'},
    'ktf': {'en': 'Kwami'},
    'ktg': {'en': 'Kalkutung'},
    'kth': {'en': 'Karanga'},
    'kti': {'en': 'North Muyu'},
    'ktj': {'en': 'Plapo Krumen'},
    'ktk': {'en': 'Kaniet'},
    'ktl': {'en': 'Koroshi'},
    'ktm': {'en': 'Kurti'},
    'ktn': {'en': 'Karitiâna'},
    'kto': {'en': 'Kuot'},
    'ktp': {'en': 'Kaduo'},
    'ktq': {'en': 'Katabaga'},
    'ktr': {'en': 'Kota Marudu Tinagas'},
    'kts': {'en': 'South Muyu'},
    'ktt': {'en': 'Ketum'},
    'ktu': {'en': 'Kituba (Democratic Republic of Congo)'},
    'ktv': {'en': 'Eastern Katu'},
    'ktw': {'en': 'Kato'},
    'ktx': {'en': 'Kaxararí'},
    'kty': {'en': 'Kango (Bas-Uélé District)'},
    'ktz': {'en': 'Juǀʼhoan'},
    'ku': {'af': 'Koerdies', 'ak': 'Kɛɛde kasa', 'am': 'ኩርድሽ', 'ar': 'الكردية', 'as': 'কুৰ্ডিচ', 'ast': 'curdu', 'az': 'kürd', 'az-Arab': 'کۆردجه', 'az-Cyrl': 'күрд', 'az-Latn': 'kürd', 'bal-Latn': 'Kordi', 'be': 'курдская', 'be-tarask': 'курдская', 'bg': 'кюрдски', 'bgn': 'کوردی', 'bn': 'কুর্দিশ', 'br': 'kurdeg', 'brx': 'कुर्दीश', 'bs': 'kurdski', 'bs-Cyrl': 'курдски', 'bs-Latn': 'kurdski', 'ca': 'kurd', 'ca-ES-valencia': 'kurd', 'ccp': '𑄇𑄪𑄢𑄴𑄘𑄨𑄥𑄴', 'ce': 'курдийн', 'chr': 'ᎫᏗᏏ', 'ckb': 'کوردی', 'cs': 'kurdština', 'cy': 'Cwrdeg', 'da': 'kurdisk', 'de': 'Kurdisch', 'dsb': 'kurdišćina', 'dz': 'ཀར་ཌིཤ་ཁ', 'ee': 'kurdiagbe', 'el': 'Κουρδικά', 'el-polyton': 'Κουρδικά', 'en': 'Kurdish', 'en-Dsrt': '𐐗𐐲𐑉𐐼𐐮𐑇', 'eo': 'kurda', 'es': 'kurdo', 'et': 'kurdi', 'eu': 'kurduera', 'fa': 'کردی', 'ff-Adlm': '𞤑𞤵𞤪𞤣𞤭𞤧𞤭𞥅𞤪𞤫', 'fi': 'kurdi', 'fil': 'Kurdish', 'fo': 'kurdiskt', 'fr': 'kurde', 'fur': 'curd', 'fy': 'Koerdysk', 'ga': 'Coirdis', 'gd': 'Cùrdais', 'gl': 'kurdo', 'gsw': 'Kurdisch', 'gu': 'કુર્દિશ', 'ha': 'Kurdanci', 'he': 'כורדית', 'hi': 'कुर्दिश', 'hi-Latn': 'Kurdish', 'hr': 'kurdski', 'hsb': 'kurdišćina', 'ht': 'kurde', 'hu': 'kurd', 'hy': 'քրդերեն', 'ia': 'kurdo', 'id': 'Kurdi', 'ig': 'Kurdish', 'is': 'kúrdíska', 'it': 'curdo', 'ja': 'クルド語', 'jv': 'Kurdis', 'ka': 'ქურთული', 'kea': 'kurdu', 'kgp': 'kurno', 'kk': 'күрд тілі', 'kk-Cyrl': 'күрд тілі', 'km': 'ឃឺដ', 'kn': 'ಕುರ್ದಿಷ್', 'ko': '쿠르드어', 'kok': 'कर्दिश', 'kok-Deva': 'कर्दिश', 'ks': 'کُردِش', 'ks-Arab': 'کُردِش', 'ksh': 'Kurrdesch', 'ku': 'kurdî (kurmancî)', 'ky': 'курдча', 'lb': 'Kurdesch', 'lkt': 'Kurd Iyápi', 'lld': 'curdich', 'lo': 'ເຄີດິສ', 'lrc': 'کوردی کورمانجی', 'lt': 'kurdų', 'lv': 'kurdu', 'mai': 'कुर्दी', 'mi': 'Kūrihi', 'mk': 'курдски', 'ml': 'കുർദ്ദിഷ്', 'mn': 'курд', 'mr': 'कुर्दिश', 'ms': 'Kurdish', 'mt': 'Kurd', 'my': 'ကဒ်', 'mzn': 'کوردی', 'nb': 'kurdisk', 'ne': 'कुर्दी', 'nl': 'Koerdisch', 'nn': 'kurdisk', 'no': 'kurdisk', 'nqo': 'ߞߎߙߑߘߎߞߊ߲', 'or': 'କୁର୍ଦ୍ଦିଶ୍', 'os': 'курдаг', 'pa': 'ਕੁਰਦਿਸ਼', 'pa-Guru': 'ਕੁਰਦਿਸ਼', 'pcm': 'Kọ́dísh Lángwej', 'pl': 'kurdyjski', 'ps': 'کردي', 'pt': 'curdo', 'qu': 'Kurdo Simi', 'rm': 'curd', 'ro': 'kurdă', 'ru': 'курдский', 'rw': 'Inyekuridishi', 'sc': 'curdu', 'sd': 'ڪردي', 'sd-Arab': 'ڪردي', 'se': 'kurdigiella', 'si': 'කුර්දි', 'sk': 'kurdčina', 'sl': 'kurdščina', 'smn': 'kurdikielâ', 'so': 'Kurdishka', 'sq': 'kurdisht', 'sr': 'курдски', 'sr-Cyrl': 'курдски', 'sr-Latn': 'kurdski', 'sv': 'kurdiska', 'sw': 'Kikurdi', 'syr': 'ܩܘܪܕܝܬ', 'ta': 'குர்திஷ்', 'te': 'కుర్దిష్', 'tg': 'курдӣ', 'th': 'เคิร์ด', 'ti': 'ኩርዲሽ', 'tk': 'kürt dili', 'to': 'lea fakakulitī', 'tr': 'Kürtçe', 'tt': 'көрд', 'ug': 'كۇردچە', 'uk': 'курдська', 'ur': 'کردش', 'uz': 'kurdcha', 'uz-Cyrl': 'курдча', 'uz-Latn': 'kurdcha', 'vec': 'curdo', 'vi': 'Tiếng Kurd', 'wae': 'Kurdiš', 'wo': 'Kurdi', 'xnr': 'कुर्दिश', 'yi': 'קורדיש', 'yo': 'Kọdiṣì', 'yrl': 'kurdu', 'yue': '庫爾德文', 'yue-Hans': '库尔德文', 'yue-Hant': '庫爾德文', 'zh': '库尔德语', 'zh-Hans': '库尔德语', 'zh-Hant': '庫德文', 'zu': 'isi-Kurdish'},
    'kub': {'en': 'Kutep'},
    'kuc': {'en': 'Kwinsu'},
    'kud': {'en': "'Auhelawa"},
    'kue': {'en': 'Kuman (Papua New Guinea)'},
    'kuf': {'en': 'Western Katu'},
    'kug': {'en': 'Kupa'},
    'kuh': {'en': 'Kushi'},
    'kui': {'en': 'Kuikúro-Kalapálo'},
    'kuj': {'en': 'Kuria'},
    'kuk': {'en': "Kepo'"},
    'kul': {'en': 'Kulere'},
    'kum': {'af': 'Kumyk', 'am': 'ኩማይክ', 'ar': 'القموقية', 'as': 'কুমিক', 'ast': 'kumyk', 'az': 'kumık', 'az-Cyrl': 'кумык', 'az-Latn': 'kumık', 'be': 'кумыцкая', 'be-tarask': 'кумыцкая', 'bg': 'кумикски', 'bgn': 'کومیکی', 'bn': 'কুমিয়াক', 'br': 'koumikeg', 'brx': 'कुमीक', 'bs': 'kumik', 'bs-Cyrl': 'кумик', 'bs-Latn': 'kumik', 'ca': 'kúmik', 'ca-ES-valencia': 'kúmik', 'ccp': '𑄇𑄪𑄟𑄨𑄇𑄴', 'ce': 'гӀумкийн', 'chr': 'ᎫᎻᎧ', 'ckb': 'کوومیک', 'cs': 'kumyčtina', 'cy': 'Cwmiceg', 'da': 'kymyk', 'de': 'Kumükisch', 'dsb': 'kumykšćina', 'el': 'Κουμγιούκ', 'el-polyton': 'Κουμγιούκ', 'en': 'Kumyk', 'es': 'kumyk', 'et': 'kumõki', 'eu': 'kumykera', 'fa': 'کومیکی', 'ff-Adlm': '𞤑𞤵𞤥𞤴𞤢𞤳𞤪𞤫', 'fi': 'kumykki', 'fil': 'Kumyk', 'fo': 'kumyk', 'fr': 'koumyk', 'fy': 'Koemuks', 'ga': 'Cúimicis', 'gd': 'Kumyk', 'gl': 'kumyk', 'gsw': 'Kumükisch', 'gu': 'કુમીક', 'ha': 'Kumyk', 'he': 'קומיקית', 'hi': 'कुमीक', 'hi-Latn': 'Kumyk', 'hr': 'kumyk', 'hsb': 'kumykšćina', 'ht': 'koumyk', 'hu': 'kumük', 'hy': 'կումիկերեն', 'ia': 'kumyko', 'id': 'Kumyk', 'ig': 'Kumik', 'is': 'kúmík', 'it': 'kumyk', 'ja': 'クムク語', 'jv': 'Kumik', 'ka': 'ყუმუხური', 'kgp': 'kumyg', 'kk': 'құмық тілі', 'kk-Cyrl': 'құмық тілі', 'km': 'គូមីគ', 'kn': 'ಕುಮೈಕ್', 'ko': '쿠믹어', 'kok': 'कुमयक', 'kok-Deva': 'कुमयक', 'ks': 'کُمِک', 'ks-Arab': 'کُمِک', 'ksh': 'Kumykesch', 'ku': 'kumikî', 'ky': 'кумыкча', 'lb': 'Kumükesch', 'lo': 'ຄູມີກ', 'lt': 'kumikų', 'lv': 'kumiku', 'mai': 'कुमिक', 'mi': 'Kumiki', 'mk': 'кумички', 'ml': 'കൂമിക്ക്', 'mn': 'кумук', 'mr': 'कुमीक', 'ms': 'Kumyk', 'mt': 'Kumyk', 'my': 'ကွမ်မိုက်', 'nb': 'kumykisk', 'ne': 'कुमिक', 'nl': 'Koemuks', 'nn': 'kumyk', 'no': 'kumykisk', 'or': 'କୁମୀକ୍', 'os': 'хъуымыхъхъаг', 'pa': 'ਕੁਮੀਕ', 'pa-Guru': 'ਕੁਮੀਕ', 'pcm': 'Kumyík Lángwej', 'pl': 'kumycki', 'ps': 'کومک', 'pt': 'kumyk', 'qu': 'Kumyk Simi', 'rm': 'kumuk', 'ro': 'kumyk', 'ru': 'кумыкский', 'sc': 'cumucu', 'sd': 'ڪومڪ', 'sd-Arab': 'ڪومڪ', 'si': 'කුමික්', 'sk': 'kumyčtina', 'sl': 'kumiščina', 'smn': 'kumykkielâ', 'so': 'Kumyk', 'sq': 'kumikisht', 'sr': 'кумички', 'sr-Cyrl': 'кумички', 'sr-Latn': 'kumički', 'sv': 'kumykiska', 'sw': 'Kumyk', 'ta': 'கும்யிக்', 'te': 'కుమ్యిక్', 'th': 'คูมืยค์', 'ti': 'ኩሚይክ', 'tk': 'kumyk dili', 'to': 'lea fakakumiki', 'tr': 'Kumukça', 'ug': 'قۇمۇقچە', 'uk': 'кумицька', 'ur': 'کومیک', 'uz': 'qo‘miq', 'uz-Latn': 'qo‘miq', 'vec': 'kumyk', 'vi': 'Tiếng Kumyk', 'xnr': 'कुमीक', 'yo': 'Èdè Kumiki', 'yrl': 'kumiki', 'yue': '庫密克文', 'yue-Hans': '库密克文', 'yue-Hant': '庫密克文', 'zh': '库梅克语', 'zh-Hans': '库梅克语', 'zh-Hant': '庫密克文', 'zu': 'isi-Kumyk'},
    'kun': {'en': 'Kunama'},
    'kuo': {'en': 'Kumukio'},
    'kup': {'en': 'Kunimaipa'},
    'kuq': {'en': 'Karipuna'},
    'kus': {'en': 'Kusaal'},
    'kut': {'ar': 'الكتيناي', 'ast': 'kutenai', 'az': 'kutenay', 'az-Latn': 'kutenay', 'bg': 'кутенай', 'bn': 'কুটেনাই', 'br': 'kutenai', 'brx': 'कुतेनाई', 'bs': 'kutenai', 'bs-Cyrl': 'кутенаи', 'bs-Latn': 'kutenai', 'ca': 'kutenai', 'ca-ES-valencia': 'kutenai', 'ccp': '𑄇𑄪𑄑𑄬𑄚𑄭', 'cs': 'kutenajština', 'da': 'kutenaj', 'de': 'Kutenai', 'el': 'Κουτενάι', 'el-polyton': 'Κουτενάι', 'en': 'Kutenai', 'en-Dsrt': '𐐢𐐰𐐼𐐨𐑌𐐬', 'es': 'kutenai', 'et': 'kutenai', 'fa': 'کوتنی', 'fi': 'kutenai', 'fr': 'kutenai', 'fy': 'Kutenai', 'gd': 'Kutenai', 'gsw': 'Kutenai-Schpraach', 'gu': 'કુતેનાઇ', 'he': 'קוטנאי', 'hi': 'क्यूतनाई', 'hi-Latn': 'Kutenai', 'hr': 'kutenai', 'ht': 'kutenai', 'hu': 'kutenai', 'id': 'Kutenai', 'is': 'kútenaí', 'it': 'kutenai', 'ja': 'クテナイ語', 'ka': 'კუტენაი', 'kgp': 'kutenaj', 'kn': 'ಕುಟೇನಾಯ್', 'ko': '쿠테네어', 'ks': 'کُتینَے', 'ks-Arab': 'کُتینَے', 'lb': 'Kutenai-Sprooch', 'lo': 'ຄູເທໄນ', 'lt': 'kutenai', 'lv': 'kutenaju', 'mk': 'кутенајски', 'ml': 'കുതേനൈ', 'mr': 'कुतेनाई', 'mt': 'Kutenaj', 'nb': 'kutenai', 'ne': 'कुतेनाइ', 'nl': 'Kutenai', 'nn': 'kutenai', 'no': 'kutenai', 'or': 'କୁତେନାଉ', 'pl': 'kutenai', 'pt': 'kutenai', 'rm': 'kutenai', 'ro': 'kutenai', 'ru': 'кутенаи', 'sk': 'kutenajčina', 'sl': 'kutenajščina', 'sr': 'кутенај', 'sr-Cyrl': 'кутенај', 'sr-Latn': 'kutenaj', 'sv': 'kutenaj', 'ta': 'குடேனை', 'te': 'కుటేనై', 'th': 'คูเทไน', 'to': 'lea fakakutenai', 'tr': 'Kutenai dili', 'ug': 'كۇتەنايچە', 'uk': 'кутенаї', 'vi': 'Tiếng Kutenai', 'yrl': 'kutenai', 'yue': '庫特奈文', 'yue-Hans': '库特奈文', 'yue-Hant': '庫特奈文', 'zh': '库特奈语', 'zh-Hans': '库特奈语', 'zh-Hant': '庫特奈文'},
    'kuu': {'en': 'Upper Kuskokwim'},
    'kuv': {'en': 'Kur'},
    'kuw': {'en': 'Kpagua'},
    'kux': {'en': 'Kukatja'},
    'kuy': {'en': "Kuuku-Ya'u"},
    'kuz': {'en': 'Kunza'},
    'kv': {'af': 'Komi', 'am': 'ኮሚ', 'ar': 'الكومي', 'as': 'কোমি', 'ast': 'komi', 'az': 'komi', 'az-Cyrl': 'коми', 'az-Latn': 'komi', 'be': 'комі', 'be-tarask': 'комі', 'bg': 'коми', 'bgn': 'کومی', 'bn': 'কোমি', 'br': 'komieg', 'brx': 'क’मी', 'bs': 'komi', 'bs-Cyrl': 'коми', 'bs-Latn': 'komi', 'ca': 'komi', 'ca-ES-valencia': 'komi', 'ccp': '𑄇𑄮𑄟𑄨', 'ce': 'комийн', 'chr': 'ᎪᎻ', 'ckb': 'کۆمی', 'cs': 'komijština', 'cy': 'Comi', 'da': 'komi', 'de': 'Komi', 'dsb': 'komišćina', 'el': 'Κόμι', 'el-polyton': 'Κόμι', 'en': 'Komi', 'es': 'komi', 'et': 'komi', 'eu': 'komiera', 'fa': 'کومیایی', 'ff-Adlm': '𞤑𞤮𞤥𞤭𞥅𞤪𞤫', 'fi': 'komi', 'fil': 'Komi', 'fo': 'komi', 'fr': 'komi', 'fy': 'Komi', 'ga': 'Coimis', 'gd': 'Komi', 'gl': 'komi', 'gsw': 'Komi-Schpraach', 'gu': 'કોમી', 'ha': 'Komi', 'he': 'קומי', 'hi': 'कोमी', 'hi-Latn': 'Komi', 'hr': 'komi', 'hsb': 'komišćina', 'ht': 'komi', 'hu': 'komi', 'hy': 'կոմիերեն', 'ia': 'komi', 'id': 'Komi', 'ig': 'Komi', 'is': 'komíska', 'it': 'komi', 'ja': 'コミ語', 'jv': 'Komi', 'ka': 'კომი', 'kgp': 'komĩ', 'kk': 'коми тілі', 'kk-Cyrl': 'коми тілі', 'km': 'កូមី', 'kn': 'ಕೋಮಿ', 'ko': '코미어', 'kok': 'कोमी', 'kok-Deva': 'कोमी', 'ks': 'کومی', 'ks-Arab': 'کومی', 'ksh': 'Kohmesch', 'ku': 'komî', 'ky': 'комиче', 'lb': 'Komi-Sprooch', 'lo': 'ໂຄມິ', 'lt': 'komi', 'lv': 'komiešu', 'mai': 'कोमी', 'mi': 'Komi', 'mk': 'коми', 'ml': 'കോമി', 'mn': 'коми', 'mr': 'कोमी', 'ms': 'Komi', 'mt': 'Komi', 'my': 'ကိုမီ', 'nb': 'komi', 'ne': 'कोमी', 'nl': 'Komi', 'nn': 'komi', 'no': 'komi', 'or': 'କୋମି', 'pa': 'ਕੋਮੀ', 'pa-Guru': 'ਕੋਮੀ', 'pcm': 'Komi Lángwej', 'pl': 'komi', 'ps': 'کومی', 'pt': 'komi', 'qu': 'Komi Simi', 'rm': 'komi', 'ro': 'komi', 'ru': 'коми', 'sc': 'komi', 'sd': 'ڪومي', 'sd-Arab': 'ڪومي', 'se': 'komigiella', 'si': 'කොමි', 'sk': 'komijčina', 'sl': 'komijščina', 'smn': 'komikielâ', 'so': 'Komi', 'sq': 'komisht', 'sr': 'коми', 'sr-Cyrl': 'коми', 'sr-Latn': 'komi', 'sv': 'kome', 'sw': 'Kikomi', 'ta': 'கொமி', 'te': 'కోమి', 'th': 'โกมิ', 'ti': 'ኮሚ', 'tk': 'komi dili', 'to': 'lea fakakomi', 'tr': 'Komi', 'ug': 'كومىچە', 'uk': 'комі', 'ur': 'کومی', 'uz': 'komi', 'uz-Latn': 'komi', 'vec': 'komi', 'vi': 'Tiếng Komi', 'xnr': 'कोमी', 'yo': 'Èdè Komi', 'yrl': 'kumi', 'yue': '科米文', 'yue-Hans': '科米文', 'yue-Hant': '科米文', 'zh': '科米语', 'zh-Hans': '科米语', 'zh-Hant': '科米文', 'zu': 'isi-Komi'},
    'kva': {'en': 'Bagvalal'},
    'kvb': {'en': 'Kubu'},
    'kvc': {'en': 'Kove'},
    'kvd': {'en': 'Kui (Indonesia)'},
    'kve': {'en': 'Kalabakan'},
    'kvf': {'en': 'Kabalai'},
    'kvg': {'en': 'Kuni-Boazi'},
    'kvh': {'en': 'Komodo'},
    'kvi': {'en': 'Kwang'},
    'kvj': {'en': 'Psikye'},
    'kvk': {'en': 'Korean Sign Language'},
    'kvl': {'en': 'Kayaw'},
    'kvm': {'en': 'Kendem'},
    'kvn': {'en': 'Border Kuna'},
    'kvo': {'en': 'Dobel'},
    'kvp': {'en': 'Kompane'},
    'kvq': {'en': 'Geba Karen'},
    'kvr': {'en': 'Kerinci'},
    'kvs': {'en': 'Kunggara'},
    'kvt': {'en': 'Lahta Karen'},
    'kvu': {'en': 'Yinbaw Karen'},
    'kvv': {'en': 'Kola'},
    'kvw': {'en': 'Wersing'},
    'kvx': {'en': 'Parkari Koli'},
    'kvy': {'en': 'Yintale Karen'},
    'kvz': {'en': 'Tsakwambo'},
    'kw': {'af': 'Kornies', 'am': 'ኮርኒሽ', 'ar': 'الكورنية', 'as': 'কোৰ্নিচ', 'ast': 'córnicu', 'az': 'korn', 'az-Arab': 'کوْرنجا', 'az-Cyrl': 'корн', 'az-Latn': 'korn', 'bal-Latn': 'Kornesh', 'be': 'корнская', 'be-tarask': 'корнская', 'bg': 'корнуолски', 'bgn': 'کورنی', 'bn': 'কর্ণিশ', 'br': 'kerneveureg', 'brx': 'कर्निस', 'bs': 'kornski', 'bs-Cyrl': 'корнишки', 'bs-Latn': 'kornski', 'ca': 'còrnic', 'ca-ES-valencia': 'còrnic', 'ccp': '𑄇𑄧𑄢𑄴𑄚𑄨𑄌𑄴', 'ce': 'корнуоллийн', 'chr': 'ᏎᎷᎭ', 'ckb': 'کۆڕنی', 'cs': 'kornština', 'cy': 'Cernyweg', 'da': 'cornisk', 'de': 'Kornisch', 'dsb': 'kornišćina', 'el': 'Κορνουαλικά', 'el-polyton': 'Κορνουαλικά', 'en': 'Cornish', 'en-Dsrt': '𐐗𐐬𐑉𐑌𐐮𐑇', 'es': 'córnico', 'et': 'korni', 'eu': 'kornubiera', 'fa': 'کورنی', 'ff-Adlm': '𞤑𞤮𞤪𞤲𞤭𞥅𞤪𞤫', 'fi': 'korni', 'fil': 'Cornish', 'fo': 'corniskt', 'fr': 'cornique', 'fur': 'cornualiês', 'fy': 'Cornish', 'ga': 'Coirnis', 'gd': 'Còrnais', 'gl': 'córnico', 'gsw': 'Kornisch', 'gu': 'કોર્નિશ', 'ha': 'Cornish', 'he': 'קורנית', 'hi': 'कोर्निश', 'hi-Latn': 'Cornish', 'hr': 'kornski', 'hsb': 'kornišćina', 'ht': 'cornique', 'hu': 'korni', 'hy': 'կոռներեն', 'ia': 'cornico', 'id': 'Kornish', 'ig': 'Cornish', 'is': 'kornbreska', 'it': 'cornico', 'ja': 'コーンウォール語', 'jv': 'Kernowek', 'ka': 'კორნული', 'kea': 'kórniku', 'kgp': 'kórnĩko', 'kk': 'корн тілі', 'kk-Cyrl': 'корн тілі', 'km': 'កូនីស', 'kn': 'ಕಾರ್ನಿಷ್', 'ko': '콘월어', 'kok': 'कोर्निश', 'kok-Deva': 'कोर्निश', 'ks': 'کورنِش', 'ks-Arab': 'کورنِش', 'ksh': 'Kornesch', 'ku': 'kornî', 'kw': 'kernewek', 'ky': 'корнишче', 'lb': 'Kornesch', 'lld': 'cornich', 'lo': 'ຄໍນິຊ', 'lrc': 'کورنیش', 'lt': 'kornų', 'lv': 'korniešu', 'mai': 'कोर्निश', 'mi': 'Kōnihi', 'mk': 'корнски', 'ml': 'കോർണിഷ്', 'mn': 'корн', 'mr': 'कोर्निश', 'ms': 'Cornish', 'mt': 'Korniku', 'my': 'ခိုနီရှ်', 'mzn': 'کورنیش', 'nb': 'kornisk', 'ne': 'कोर्निस', 'nl': 'Cornish', 'nn': 'kornisk', 'no': 'kornisk', 'nqo': 'ߞߐߙߑߣߌߞߌߞߊ߲', 'or': 'କୋର୍ନିସ୍', 'pa': 'ਕੋਰਨਿਸ਼', 'pa-Guru': 'ਕੋਰਨਿਸ਼', 'pcm': 'Kọ́nish Lángwej', 'pl': 'kornijski', 'ps': 'کورنيشي', 'pt': 'córnico', 'qu': 'Córnico Simi', 'rm': 'cornic', 'ro': 'cornică', 'ru': 'корнский', 'sc': 'còrnicu', 'sd': 'ڪورنش', 'sd-Arab': 'ڪورنش', 'se': 'kornagiella', 'si': 'කෝනීසියානු', 'sk': 'kornčina', 'sl': 'kornijščina', 'smn': 'kornikielâ', 'so': 'Kornish', 'sq': 'kornisht', 'sr': 'корнволски', 'sr-Cyrl': 'корнволски', 'sr-Latn': 'kornvolski', 'sv': 'korniska', 'sw': 'Kikorni', 'ta': 'கார்னிஷ்', 'te': 'కోర్నిష్', 'th': 'คอร์นิช', 'ti': 'ኮርንኛ', 'tk': 'korn dili', 'to': 'lea fakakoniuali', 'tr': 'Kernevekçe', 'ug': 'كورنىشچە', 'uk': 'корнська', 'ur': 'کورنش', 'uz': 'korn', 'uz-Cyrl': 'корнча', 'uz-Latn': 'korn', 'vec': 'còrnego', 'vi': 'Tiếng Cornwall', 'xnr': 'कोर्निश', 'yi': 'קארניש', 'yo': 'Èdè Kọ́nììṣì', 'yrl': 'kúriniku', 'yue': '康瓦耳文', 'yue-Hans': '康瓦耳文', 'yue-Hant': '康瓦耳文', 'zh': '康沃尔语', 'zh-Hans': '康沃尔语', 'zh-Hant': '康瓦耳文', 'zu': 'isi-Cornish'},
    'kwa': {'en': 'Dâw'},
    'kwb': {'en': 'Kwa'},
    'kwc': {'en': 'Likwala'},
    'kwd': {'en': 'Kwaio'},
    'kwe': {'en': 'Kwerba'},
    'kwf': {'en': "Kwara'ae"},
    'kwg': {'en': 'Sara Kaba Deme'},
    'kwh': {'en': 'Kowiai'},
    'kwi': {'en': 'Awa-Cuaiquer'},
    'kwj': {'en': 'Kwanga'},
    'kwk': {'af': 'Kwak’wala', 'am': 'ክዋክዋላ', 'ar': 'الكواكوالا', 'as': 'ক্বাকৱালা', 'az': 'Kvakvala', 'az-Latn': 'Kvakvala', 'be': 'квакіутль', 'be-tarask': 'квакіутль', 'bg': 'куак’уала', 'bn': 'কোয়াক’ওয়ালা', 'br': 'kwakwaleg', 'brx': 'क्वाकवाला', 'bs': 'kvakvala', 'bs-Cyrl': 'кваквала', 'bs-Latn': 'kvakvala', 'ca': 'kwak’wala', 'ca-ES-valencia': 'kwak’wala', 'chr': 'ᏆᏆᎳ', 'cs': 'kvakiutština', 'cy': 'Kwakʼwala', 'da': 'kwakʼwala', 'de': 'Kwakʼwala', 'dsb': 'kwakʼwalašćina', 'el': 'Κουακουάλα', 'el-polyton': 'Κουακουάλα', 'en': 'Kwakʼwala', 'es': 'kwakʼwala', 'et': 'kvakvala', 'eu': 'kwakwala', 'fa': 'کواک والا', 'ff-Adlm': '𞤑𞤢𞤱𞤳𞥇𞤱𞤢𞤤𞤢𞥄𞤪𞤫', 'fi': 'kwakʼwala', 'fil': 'Kwakʼwala', 'fr': 'kwak’wala', 'ga': 'Kwakʼwala', 'gd': 'Kwakʼwala', 'gl': 'kwakiutl', 'gu': 'ક્વેકવાલા', 'ha': 'Kwakʼwala', 'he': 'קוואקוואלה', 'hi': 'क्वॉकवाला', 'hi-Latn': 'Kwakʼwala', 'hr': 'kwakʼwala', 'hsb': 'kwakʼwalašćina', 'ht': 'kwak’wala', 'hu': 'kwakʼwala', 'hy': 'կվակվալա', 'ia': 'kwakwala', 'id': 'Kwakʼwala', 'ig': 'Kwakʼwala', 'is': 'kwakʼwala', 'it': 'kwakʼwala', 'ja': 'クヮキゥワラ語', 'jv': 'Kwakʼwala', 'ka': 'კვაკვალა', 'kk': 'квакиутль тілі', 'kk-Cyrl': 'квакиутль тілі', 'km': 'ក្វាក់វ៉ាឡា', 'kn': 'ಕ್ವಾಕ್\u200cವಾಲಾ', 'ko': '곽왈라어', 'kok': 'क्वाकवाला', 'kok-Deva': 'क्वाकवाला', 'ku': 'kwak’walayî', 'ky': 'кваквалача (индей тили)', 'lo': 'ຄວາກຄວາກລາ', 'lt': 'kvakvalų', 'lv': 'kvakvala', 'mai': 'कुअकवाला', 'mi': 'Kuakawara', 'mk': 'кваквала', 'ml': 'ക്വാക്വല', 'mn': 'квак вала', 'mr': 'क्वक्क्वाला', 'ms': 'Kwak’wala', 'my': 'ကွပ်ခ်ဝါလာ', 'nb': 'kwak̓wala', 'ne': 'क्वाकवाला', 'nl': 'Kwakʼwala', 'nn': 'kwak̓wala', 'no': 'kwak̓wala', 'or': 'କ୍ଵାକୱାଲା', 'pa': 'ਕਵਾਕ’ਵਾਲਾ', 'pa-Guru': 'ਕਵਾਕ’ਵਾਲਾ', 'pcm': 'Kwakwála Lángwej', 'pl': 'kwakiutl', 'ps': 'Vote kwk', 'pt': 'kwakʼwala', 'qu': 'Kwakʼwala Simi', 'ro': 'kwakʼwala', 'ru': 'квакиутль', 'sc': 'kwakʼwala', 'sd': 'ڪئاڪ ولا', 'sd-Arab': 'ڪئاڪ ولا', 'si': 'ක්වාක්වාලා', 'sk': 'kwakʼwala', 'sl': 'kvakvala', 'so': 'Kwakʼwala', 'sq': 'kuakualaisht', 'sr': 'кваквала', 'sr-Cyrl': 'кваквала', 'sr-Latn': 'kvakvala', 'sv': 'kwakʼwala', 'sw': 'Kikwakʼwala', 'ta': 'குவாக்வாலா', 'te': 'క్వాక్\u200cవాలా', 'th': 'ควักวาลา', 'ti': 'ክዋክዋላ', 'tk': 'kwakwala dili', 'to': 'lea fakakuakuala', 'tr': 'Kwakʼwala dili', 'uk': 'кваквала', 'ur': 'کیواکوالا', 'uz': 'kvakvala', 'uz-Latn': 'kvakvala', 'vec': 'Kwakwala', 'vi': 'Tiếng Kwakʼwala', 'yo': 'Èdè Kwawala', 'yue': '誇誇瓦拉文', 'yue-Hans': '夸夸瓦拉文', 'yue-Hant': '誇誇瓦拉文', 'zh': '夸夸瓦拉语', 'zh-Hans': '夸夸瓦拉语', 'zh-Hant': '誇誇嘉誇語', 'zu': 'Kwakʼwala'},
    'kwl': {'en': 'Kofyar'},
    'kwm': {'en': 'Kwambi'},
    'kwn': {'en': 'Kwangali'},
    'kwo': {'en': 'Kwomtari'},
    'kwp': {'en': 'Kodia'},
    'kwq': {'en': 'Kwak'},
    'kwr': {'en': 'Kwer'},
    'kws': {'en': 'Kwese'},
    'kwt': {'en': 'Kwesten'},
    'kwu': {'en': 'Kwakum'},
    'kwv': {'en': 'Sara Kaba Náà'},
    'kww': {'en': 'Kwinti'},
    'kwx': {'en': 'Khirwar'},
    'kwy': {'en': 'San Salvador Kongo'},
    'kwz': {'en': 'Kwadi'},
    'kxa': {'en': 'Kairiru'},
    'kxb': {'en': 'Krobu'},
    'kxc': {'en': 'Konso'},
    'kxd': {'en': 'Brunei'},
    'kxe': {'en': 'Kakihum'},
    'kxf': {'en': 'Manumanaw Karen'},
    'kxh': {'en': 'Karo (Ethiopia)'},
    'kxi': {'en': 'Keningau Murut'},
    'kxj': {'en': 'Kulfa'},
    'kxk': {'en': 'Zayein Karen'},
    'kxl': {'en': 'Nepali Kurux'},
    'kxm': {'en': 'Northern Khmer'},
    'kxn': {'en': 'Kanowit-Tanjong Melanau'},
    'kxo': {'en': 'Kanoé'},
    'kxp': {'en': 'Wadiyara Koli'},
    'kxq': {'en': 'Smärky Kanum'},
    'kxr': {'en': 'Koro (Papua New Guinea)'},
    'kxs': {'en': 'Kangjia'},
    'kxt': {'en': 'Koiwat'},
    'kxu': {'en': 'Kui (India)'},
    'kxv': {'af': 'Kuvi', 'ak': 'Kuvi kasa', 'am': 'ኩቪኛ', 'ar': 'الكوفية', 'as': 'কুভি', 'az': 'kuvi', 'az-Latn': 'kuvi', 'bal-Latn': 'Kuwi', 'be': 'куві', 'be-tarask': 'куві', 'bg': 'кови', 'bn': 'কুভি', 'bs': 'kuvi', 'bs-Latn': 'kuvi', 'ca': 'kuvi', 'ca-ES-valencia': 'kuvi', 'chr': 'ᎫᏈ', 'cs': 'kúvi', 'cy': 'Kuvi', 'da': 'kuvi', 'de': 'Kuvi', 'dsb': 'kuvišćina', 'el': 'Κούβι', 'el-polyton': 'Κούβι', 'en': 'Kuvi', 'es': 'kuvi', 'et': 'kuvi', 'eu': 'kuvia', 'fa': 'کووی', 'fi': 'kuvi', 'fil': 'Kuvi', 'fo': 'kuvi', 'fr': 'kuvi', 'ga': 'Kuvi', 'gd': 'Kuvi', 'gl': 'kuvi', 'gu': 'કૂવી', 'ha': 'Kavi', 'he': 'קווי', 'hi': 'कुवी', 'hi-Latn': 'Kuvi', 'hr': 'kuvi', 'hsb': 'kuvišćina', 'ht': 'kuvi', 'hu': 'kuvi', 'hy': 'կուվի', 'ia': 'kuvi', 'id': 'Kuvi', 'ig': 'Kuvi', 'is': 'kúví', 'it': 'kuvi', 'ja': 'クーヴィンガ語', 'jv': 'Kuvi', 'ka': 'კუვი', 'kk': 'куви тілі', 'kk-Cyrl': 'куви тілі', 'km': 'គូវី', 'kn': 'ಕುವಿ', 'ko': '쿠비어', 'kok': 'कुवी', 'kok-Deva': 'कुवी', 'ku': 'kuvî', 'kxv': 'kuvi', 'kxv-Deva': 'कुवि', 'kxv-Latn': 'kuvi', 'kxv-Orya': 'କୁୱି', 'kxv-Telu': 'కువి', 'ky': 'куви', 'lo': 'ຄູວີ', 'lt': 'kuvi', 'lv': 'kuvi', 'mk': 'куви', 'ml': 'കുവി', 'mn': 'куви', 'mr': 'कुवी', 'ms': 'Kuvi', 'my': 'ကူဗီ', 'nb': 'kuvi', 'ne': 'कुभी', 'nl': 'Kuvi', 'nn': 'kuvi', 'no': 'kuvi', 'or': 'କୁୱି', 'pa': 'ਕੁਵੀ', 'pa-Guru': 'ਕੁਵੀ', 'pcm': 'Kuvi Lángwej', 'pl': 'kuvi', 'ps': 'کووئ', 'pt': 'kuvi', 'ro': 'kuvi', 'ru': 'куви', 'sc': 'kuvi', 'sd': 'ڪووي', 'sd-Arab': 'ڪووي', 'si': 'කුවි', 'sk': 'kuvi', 'sl': 'kuvi', 'so': 'Kufi', 'sq': 'kuvisht', 'sr': 'куви', 'sr-Cyrl': 'куви', 'sr-Latn': 'kuvi', 'sv': 'kuvi', 'sw': 'Kikuvi', 'ta': 'குவி', 'te': 'కువి', 'th': 'กูวี', 'ti': 'ኩቪ', 'tk': 'kuwi dili', 'tr': 'Kuvi', 'uk': 'куві', 'ur': 'کووی', 'uz': 'kuvi', 'uz-Latn': 'kuvi', 'vec': 'kuvi', 'vi': 'Tiếng Kuvi', 'yo': 'Kufi', 'yue': '庫維文', 'yue-Hans': '库维文', 'yue-Hant': '庫維文', 'zh': '库维语', 'zh-Hans': '库维语', 'zh-Hant': '庫維文', 'zu': 'Kuvi'},
    'kxw': {'en': 'Konai'},
    'kxx': {'en': 'Likuba'},
    'kxy': {'en': 'Kayong'},
    'kxz': {'en': 'Kerewo'},
    'ky': {'af': 'Kirgisies', 'ak': 'Kɛgyese kasa', 'am': 'ክይርግይዝ', 'ar': 'القيرغيزية', 'as': 'কিৰ্গিজ', 'ast': 'kirguistanín', 'az': 'qırğız', 'az-Arab': 'قیرقیزجا', 'az-Cyrl': 'гырғыз', 'az-Latn': 'qırğız', 'bal-Latn': 'Kirgez', 'be': 'кіргізская', 'be-tarask': 'кіргізская', 'bg': 'киргизки', 'bgn': 'قیرغیزی', 'bn': 'কির্গিজ', 'br': 'kirgiz', 'brx': 'किरगीज', 'bs': 'kirgiški', 'bs-Cyrl': 'киргиски', 'bs-Latn': 'kirgiški', 'ca': 'kirguís', 'ca-ES-valencia': 'kirguís', 'ccp': '𑄇𑄨𑄢𑄴𑄉𑄨𑄌𑄴', 'ce': 'гӀиргӀизойн', 'chr': 'ᎩᎵᏣᎢᏍ', 'ckb': 'كرگیزی', 'cs': 'kyrgyzština', 'cy': 'Cirgiseg', 'da': 'kirgisisk', 'de': 'Kirgisisch', 'dsb': 'kirgišćina', 'dz': 'ཀིར་གིས་ཁ', 'ee': 'kirghistangbe', 'el': 'Κιργιζικά', 'el-polyton': 'Κιργιζικά', 'en': 'Kyrgyz', 'eo': 'kirgiza', 'es': 'kirguís', 'et': 'kirgiisi', 'eu': 'kirgizera', 'fa': 'قرقیزی', 'ff-Adlm': '𞤑𞤭𞤪𞤺𞤵𞥅𞤪𞤫', 'fi': 'kirgiisi', 'fil': 'Kirghiz', 'fo': 'kyrgyz', 'fr': 'kirghize', 'fy': 'Kirgizysk', 'ga': 'Cirgisis', 'gd': 'Cìorgasais', 'gl': 'kirguiz', 'gsw': 'Kirgiisisch', 'gu': 'કિર્ગીઝ', 'ha': 'Kirgizanci', 'he': 'קירגיזית', 'hi': 'किर्गीज़', 'hi-Latn': 'Kyrgyz', 'hr': 'kirgiski', 'hsb': 'kirgišćina', 'ht': 'kirghize', 'hu': 'kirgiz', 'hy': 'ղրղզերեն', 'ia': 'kirghizo', 'id': 'Kirgiz', 'ig': 'Kyrgyz', 'is': 'kirgiska', 'it': 'kirghiso', 'ja': 'キルギス語', 'jv': 'Kirgis', 'ka': 'ყირგიზული', 'kea': 'kirgiz', 'kgp': 'kirginh', 'kk': 'қырғыз тілі', 'kk-Cyrl': 'қырғыз тілі', 'km': '\u200bកៀហ្ស៊ីស', 'kn': 'ಕಿರ್ಗಿಜ್', 'ko': '키르기스어', 'kok': 'किर्गिझ', 'kok-Deva': 'किर्गिझ', 'ks': 'کِرگِز', 'ks-Arab': 'کِرگِز', 'ksh': 'Kirjihsesch', 'ku': 'kirgizî', 'kxv': 'kyrgyj', 'kxv-Deva': 'किर्गीज़', 'kxv-Latn': 'kyrgyj', 'kxv-Orya': 'କୀରଗୀଜ୍', 'kxv-Telu': 'కిర్గజ్', 'ky': 'кыргызча', 'lb': 'Kirgisesch', 'lkt': 'Kirghiz Iyápi', 'lld': 'kyrgyz', 'lo': 'ເກຍກີສ', 'lrc': 'قئرقیزی', 'lt': 'kirgizų', 'lv': 'kirgīzu', 'mai': 'किर्गिज', 'mi': 'Kiakihi', 'mk': 'киргиски', 'ml': 'കിർഗിസ്', 'mn': 'киргиз', 'mr': 'किरगीझ', 'ms': 'Kirghiz', 'mt': 'Kirgiż', 'my': 'ကာဂျစ်', 'mzn': 'قرقیزی', 'nb': 'kirgisisk', 'ne': 'किर्गिज', 'nl': 'Kirgizisch', 'nn': 'kirgisisk', 'no': 'kirgisisk', 'nqo': 'ߞߌߙߑߞߌߖ߭ߑߞߊ߲', 'or': 'କୀରଗୀଜ୍', 'pa': 'ਕਿਰਗੀਜ਼', 'pa-Guru': 'ਕਿਰਗੀਜ਼', 'pcm': 'Kiẹ́gíz Lángwej', 'pl': 'kirgiski', 'ps': 'کرغيزي', 'pt': 'quirguiz', 'qu': 'Kirghiz Simi', 'rm': 'kirghis', 'ro': 'kârgâză', 'ru': 'киргизский', 'rw': 'Inkerigizi', 'sah': 'Кыргыстыы', 'sc': 'chirghisu', 'sd': 'ڪرغيز', 'sd-Arab': 'ڪرغيز', 'si': 'කිර්ගිස්', 'sk': 'kirgizština', 'sl': 'kirgiščina', 'smn': 'kirgiskielâ', 'so': 'Kirgiis', 'sq': 'kirgizisht', 'sr': 'киргиски', 'sr-Cyrl': 'киргиски', 'sr-Latn': 'kirgiski', 'sv': 'kirgiziska', 'sw': 'Kikyrgyz', 'ta': 'கிர்கிஸ்', 'te': 'కిర్గిజ్', 'tg': 'қирғизӣ', 'th': 'คีร์กีซ', 'ti': 'ኪርጊዝኛ', 'tk': 'gyrgyz dili', 'to': 'lea fakakīsisi', 'tr': 'Kırgızca', 'tt': 'кыргыз', 'ug': 'قىرغىزچە', 'uk': 'киргизька', 'ur': 'کرغیزی', 'uz': 'qirgʻizcha', 'uz-Cyrl': 'қирғизча', 'uz-Latn': 'qirgʻizcha', 'vec': 'kirghizo', 'vi': 'Tiếng Kyrgyz', 'wae': 'Kirgisiš', 'wo': 'Kirgiis', 'xnr': 'किर्गीज़', 'yi': 'קירגיזיש', 'yo': 'Kírígíìsì', 'yrl': 'kirigixi', 'yue': '吉爾吉斯文', 'yue-Hans': '吉尔吉斯文', 'yue-Hant': '吉爾吉斯文', 'zh': '柯尔克孜语', 'zh-Hans': '柯尔克孜语', 'zh-Hant': '吉爾吉斯文', 'zu': 'isi-Kyrgyz'},
    'kya': {'en': 'Kwaya'},
    'kyb': {'en': 'Butbut Kalinga'},
    'kyc': {'en': 'Kyaka'},
    'kyd': {'en': 'Karey'},
    'kye': {'en': 'Krache'},
    'kyf': {'en': 'Kouya'},
    'kyg': {'en': 'Keyagana'},
    'kyh': {'en': 'Karok'},
    'kyi': {'en': 'Kiput'},
    'kyj': {'en': 'Karao'},
    'kyk': {'en': 'Kamayo'},
    'kyl': {'en': 'Kalapuya'},
    'kym': {'en': 'Kpatili'},
    'kyn': {'en': 'Northern Binukidnon'},
    'kyo': {'en': 'Kelon'},
    'kyp': {'en': 'Kang'},
    'kyq': {'en': 'Kenga'},
    'kyr': {'en': 'Kuruáya'},
    'kys': {'en': 'Baram Kayan'},
    'kyt': {'en': 'Kayagar'},
    'kyu': {'en': 'Western Kayah'},
    'kyv': {'en': 'Kayort'},
    'kyw': {'en': 'Kudmali'},
    'kyx': {'en': 'Rapoisi'},
    'kyy': {'en': 'Kambaira'},
    'kyz': {'en': 'Kayabí'},
    'kza': {'en': 'Western Karaboro'},
    'kzb': {'en': 'Kaibobo'},
    'kzc': {'en': 'Bondoukou Kulango'},
    'kzd': {'en': 'Kadai'},
    'kze': {'en': 'Kosena'},
    'kzf': {'en': "Da'a Kaili"},
    'kzg': {'en': 'Kikai'},
    'kzh': {'en': 'Kenuzi-Dongola'},
    'kzi': {'en': 'Kelabit'},
    'kzj': {'en': 'Coastal Kadazan'},
    'kzk': {'en': 'Kazukuru'},
    'kzl': {'en': 'Kayeli'},
    'kzm': {'en': 'Kais'},
    'kzn': {'en': 'Kokola'},
    'kzo': {'en': 'Kaningi'},
    'kzp': {'en': 'Kaidipang'},
    'kzq': {'en': 'Kaike'},
    'kzr': {'en': 'Karang'},
    'kzs': {'en': 'Sugut Dusun'},
    'kzt': {'en': 'Tambunan Dusun'},
    'kzu': {'en': 'Kayupulau'},
    'kzv': {'en': 'Komyandaret'},
    'kzw': {'en': 'Karirí-Xocó'},
    'kzx': {'en': 'Kamarian'},
    'kzy': {'en': 'Kango (Tshopo District)'},
    'kzz': {'en': 'Kalabra'},
    'la': {'af': 'Latyn', 'am': 'ላቲንኛ', 'ar': 'اللاتينية', 'as': 'লেটিন', 'ast': 'llatín', 'az': 'latın', 'az-Arab': 'لاتینجه', 'az-Cyrl': 'латын', 'az-Latn': 'latın', 'bal-Latn': 'Látini', 'be': 'лацінская', 'be-tarask': 'лацінская', 'bg': 'латински', 'bgn': 'لاتین', 'bn': 'লাতিন', 'br': 'latin', 'brx': 'लेटीन', 'bs': 'latinski', 'bs-Cyrl': 'латински', 'bs-Latn': 'latinski', 'ca': 'llatí', 'ca-ES-valencia': 'llatí', 'ccp': '𑄣𑄑𑄨𑄚𑄴', 'ce': 'латинан', 'chr': 'ᎳᏘᏂ', 'ckb': 'لاتینی', 'cs': 'latina', 'cy': 'Lladin', 'da': 'latin', 'de': 'Latein', 'dsb': 'łatyńšćina', 'dz': 'ལེ་ཊིན་ཁ', 'ee': 'latin', 'el': 'Λατινικά', 'el-polyton': 'Λατινικά', 'en': 'Latin', 'en-Dsrt': '𐐢𐐰𐐻𐐮𐑌', 'eo': 'latino', 'es': 'latín', 'et': 'ladina', 'eu': 'latina', 'fa': 'لاتین', 'ff-Adlm': '𞤂𞤢𞤼𞤫𞤲𞤪𞤫', 'fi': 'latina', 'fil': 'Latin', 'fo': 'latín', 'fr': 'latin', 'fur': 'latin', 'fy': 'Latyn', 'ga': 'Laidin', 'gd': 'Laideann', 'gl': 'latín', 'gsw': 'Latiin', 'gu': 'લેટિન', 'ha': 'Dan Kabilar Latin', 'haw': 'Lākina', 'he': 'לטינית', 'hi': 'लैटिन', 'hi-Latn': 'Latin', 'hr': 'latinski', 'hsb': 'łaćonšćina', 'ht': 'latin', 'hu': 'latin', 'hy': 'լատիներեն', 'ia': 'latino', 'id': 'Latin', 'ig': 'Latin', 'is': 'latína', 'it': 'latino', 'ja': 'ラテン語', 'jv': 'Latin', 'ka': 'ლათინური', 'kea': 'latin', 'kgp': 'ratĩnh', 'kk': 'латын тілі', 'kk-Cyrl': 'латын тілі', 'km': 'ឡាតំាង', 'kn': 'ಲ್ಯಾಟಿನ್', 'ko': '라틴어', 'kok': 'लॅटिन', 'kok-Deva': 'लॅटिन', 'ks': 'لاتیٖنی', 'ks-Arab': 'لاتیٖنی', 'ksh': 'Lateijnesch', 'ku': 'latînî', 'ky': 'латынча', 'lb': 'Latäin', 'lkt': 'Latin Iyápi', 'lld': 'latin', 'lo': 'ລາຕິນ', 'lrc': 'لاتین', 'lt': 'lotynų', 'lv': 'latīņu', 'mai': 'लातिन', 'mi': 'Rātini', 'mk': 'латински', 'ml': 'ലാറ്റിൻ', 'mn': 'латин', 'mr': 'लॅटिन', 'ms': 'Latin', 'mt': 'Latin', 'my': 'လက်တင်', 'mzn': 'لاتین', 'nb': 'latin', 'ne': 'ल्याटिन', 'nl': 'Latijn', 'nn': 'latin', 'no': 'latin', 'nqo': 'ߟߊ߬ߕߍ߲߬ߞߊ߲', 'om': 'Afaan Laatini', 'or': 'ଲାଟିନ୍', 'os': 'латинаг', 'pa': 'ਲਾਤੀਨੀ', 'pa-Guru': 'ਲਾਤੀਨੀ', 'pcm': 'Látín Lángwej', 'pl': 'łaciński', 'ps': 'لاتیني', 'pt': 'latim', 'qu': 'Latín Simi', 'rm': 'latin', 'ro': 'latină', 'ru': 'латинский', 'rw': 'Ikilatini', 'sah': 'Латыынныы', 'sc': 'latinu', 'sd': 'لاطيني', 'sd-Arab': 'لاطيني', 'se': 'láhtengiella', 'si': 'ලතින්', 'sk': 'latinčina', 'sl': 'latinščina', 'smn': 'läättinkielâ', 'so': 'Laatiin', 'sq': 'latinisht', 'sr': 'латински', 'sr-Cyrl': 'латински', 'sr-Latn': 'latinski', 'sv': 'latin', 'sw': 'Kilatini', 'syr': 'ܠܬܝܢܝܬ', 'ta': 'லத்தின்', 'te': 'లాటిన్', 'tg': 'лотинӣ', 'th': 'ละติน', 'ti': 'ላቲን', 'tk': 'latyn dili', 'to': 'lea fakalatina', 'tr': 'Latince', 'tt': 'латин', 'ug': 'لاتىنچە', 'uk': 'латинська', 'ur': 'لاطینی', 'uz': 'lotincha', 'uz-Cyrl': 'лотинча', 'uz-Latn': 'lotincha', 'vec': 'latin', 'vi': 'Tiếng La-tinh', 'wae': 'Latiniš', 'wo': 'Latin', 'xnr': 'लैटिन', 'yi': 'לאטייניש', 'yo': 'Èdè Latini', 'yrl': 'ratĩ', 'yue': '拉丁文', 'yue-Hans': '拉丁文', 'yue-Hant': '拉丁文', 'zh': '拉丁语', 'zh-Hans': '拉丁语', 'zh-Hant': '拉丁文', 'zu': 'isi-Latin', 'la': 'Lingua latina'},
    'laa': {'en': 'Southern Subanen'},
    'lab': {'en': 'Linear A'},
    'lac': {'en': 'Lacandon'},
    'lad': {'af': 'Ladino', 'am': 'ላዲኖ', 'ar': 'اللادينو', 'as': 'লাডিনো', 'ast': 'ladino', 'az': 'sefard', 'az-Cyrl': 'сефард', 'az-Latn': 'sefard', 'be': 'ладына', 'be-tarask': 'ладына', 'bg': 'ладино', 'bgn': 'لادینو', 'bn': 'লাদিনো', 'br': 'ladino', 'brx': 'लाडीन’', 'bs': 'ladino', 'bs-Cyrl': 'ладино', 'bs-Latn': 'ladino', 'ca': 'judeocastellà', 'ca-ES-valencia': 'judeocastellà', 'ccp': '𑄣𑄓𑄨𑄚𑄮', 'ce': 'ладино', 'chr': 'ᎳᏗᏃ', 'ckb': 'لادینۆ', 'cs': 'ladinština', 'cy': 'Iddew-Sbaeneg', 'da': 'ladino', 'de': 'Ladino', 'dsb': 'ladinšćina', 'el': 'Λαδίνο', 'el-polyton': 'Λαδίνο', 'en': 'Ladino', 'es': 'ladino', 'et': 'ladiino', 'eu': 'ladinoa', 'fa': 'لادینو', 'ff-Adlm': '𞤂𞤢𞤣𞤭𞤲𞤮𞥅𞤪𞤫', 'fi': 'ladino', 'fil': 'Ladino', 'fo': 'ladino', 'fr': 'ladino', 'fy': 'Ladino', 'ga': 'Laidínis', 'gd': 'Ladino', 'gl': 'ladino', 'gsw': 'Ladino', 'gu': 'લાદીનો', 'ha': 'Ladino', 'he': 'לדינו', 'hi': 'लादीनो', 'hi-Latn': 'Ladino', 'hr': 'ladino', 'hsb': 'ladinšćina', 'ht': 'ladino', 'hu': 'ladino', 'hy': 'լադինո', 'ia': 'judeo-espaniol', 'id': 'Ladino', 'ig': 'Ladino', 'is': 'ladínska', 'it': 'giudeo-spagnolo', 'ja': 'ラディノ語', 'jv': 'Ladino', 'ka': 'ლადინო', 'kgp': 'raninũ', 'kk': 'ладино тілі', 'kk-Cyrl': 'ладино тілі', 'km': 'ឡាឌីណូ', 'kn': 'ಲ್ಯಾಡಿನೋ', 'ko': '라디노어', 'kok': 'लाडिनो', 'kok-Deva': 'लाडिनो', 'ks': 'لیڈِنو', 'ks-Arab': 'لیڈِنو', 'ksh': 'Ladihnesch', 'ku': 'ladînoyî', 'ky': 'ладиночо', 'lb': 'Ladino', 'lo': 'ລາດີໂນ', 'lt': 'ladino', 'lv': 'ladino', 'mai': 'लडीनो', 'mi': 'Ratino', 'mk': 'ладино', 'ml': 'ലഡീനോ', 'mn': 'ладин', 'mr': 'लादीनो', 'ms': 'Ladino', 'mt': 'Ladino', 'my': 'လာဒီနို', 'nb': 'ladinsk', 'ne': 'लाडिनो', 'nl': 'Ladino', 'nn': 'ladino', 'no': 'ladinsk', 'or': 'ଲାଦିନୋ', 'pa': 'ਲੈਡੀਨੋ', 'pa-Guru': 'ਲੈਡੀਨੋ', 'pcm': 'Ladíno Lángwej', 'pl': 'ladyński', 'ps': 'لاډینو', 'pt': 'ladino', 'qu': 'Ladino Simi', 'rm': 'ladino', 'ro': 'ladino', 'ru': 'ладино', 'sc': 'giudeu-ispagnolu', 'sd': 'لڊينو', 'sd-Arab': 'لڊينو', 'si': 'ලඩිනො', 'sk': 'židovská španielčina', 'sl': 'ladinščina', 'smn': 'ladinokielâ', 'so': 'Ladino', 'sq': 'ladinoisht', 'sr': 'ладино', 'sr-Cyrl': 'ладино', 'sr-Latn': 'ladino', 'sv': 'ladino', 'sw': 'Kiladino', 'ta': 'லடினோ', 'te': 'లాడినో', 'th': 'ลาดิโน', 'ti': 'ላዲኖ', 'tk': 'ladino dili', 'to': 'lea fakalatino', 'tr': 'Ladino', 'ug': 'لادىنوچە', 'uk': 'ладино', 'ur': 'لیڈینو', 'uz': 'ladino', 'uz-Latn': 'ladino', 'vec': 'judezmo', 'vi': 'Tiếng Ladino', 'xnr': 'लादीनो', 'yi': 'לאַדינא', 'yo': 'Èdè Ladino', 'yrl': 'radinu', 'yue': '拉迪諾文', 'yue-Hans': '拉迪诺文', 'yue-Hant': '拉迪諾文', 'zh': '拉迪诺语', 'zh-Hans': '拉迪诺语', 'zh-Hant': '拉迪諾文', 'zu': 'isi-Ladino'},
    'lae': {'en': 'Pattani'},
    'laf': {'en': 'Lafofa'},
    'lag': {'af': 'Langi', 'am': 'ላንጊ', 'ar': 'لانجي', 'as': 'লাংগি', 'ast': 'langi', 'az': 'langi', 'az-Cyrl': 'ланҝи', 'az-Latn': 'langi', 'bal-Latn': 'Langi', 'be': 'лангі', 'be-tarask': 'лангі', 'bg': 'ланги', 'bgn': 'لانگی', 'bn': 'লাঙ্গি', 'br': 'langi', 'brx': 'लाङि', 'bs': 'langi', 'bs-Cyrl': 'ланги', 'bs-Latn': 'langi', 'ca': 'langi', 'ca-ES-valencia': 'langi', 'ccp': '𑄣𑄋𑄴𑄉𑄨', 'ce': 'ланги', 'chr': 'ᎳᏂᎩ', 'ckb': 'لانگی', 'cs': 'langi', 'cy': 'Langi', 'da': 'langi', 'de': 'Langi', 'dsb': 'langi', 'el': 'Λάνγκι', 'el-polyton': 'Λάνγκι', 'en': 'Langi', 'es': 'langi', 'et': 'langi', 'eu': 'langiera', 'fa': 'لانگی', 'ff-Adlm': '𞤂𞤢𞤲𞤺𞤭𞥅𞤪𞤫', 'fi': 'lango', 'fil': 'Langi', 'fo': 'langi', 'fr': 'langi', 'fy': 'Langi', 'ga': 'Ciolaingis', 'gd': 'Langi', 'gl': 'langi', 'gu': 'લંગી', 'ha': 'Langi', 'he': 'לאנגי', 'hi': 'लांगि', 'hi-Latn': 'Langi', 'hr': 'langi', 'hsb': 'langi', 'ht': 'langi', 'hu': 'langi', 'hy': 'լանգի', 'ia': 'langi', 'id': 'Langi', 'ig': 'Langị', 'is': 'langí', 'it': 'langi', 'ja': 'ランギ語', 'jv': 'Langi', 'ka': 'ლანგი', 'kea': 'langi', 'kgp': 'rỹngi', 'kk': 'ланги тілі', 'kk-Cyrl': 'ланги тілі', 'km': 'ឡានហ្គី', 'kn': 'ಲಾಂಗಿ', 'ko': '랑기어', 'kok': 'लांगी', 'kok-Deva': 'लांगी', 'ksh': 'Lango', 'ku': 'langî', 'ky': 'лангиче', 'lag': 'Kɨlaangi', 'lb': 'Langi', 'lo': 'ແລນກິ', 'lrc': 'لانگی', 'lt': 'langi', 'lv': 'langi', 'mai': 'लंगी', 'mi': 'Rangi', 'mk': 'ланги', 'ml': 'ലാംഗി', 'mn': 'ланги', 'mr': 'लांगी', 'ms': 'Langi', 'mt': 'Langi', 'my': 'လန်ဂီ', 'mzn': 'لانگی', 'nb': 'langi', 'ne': 'लाङ्गी', 'nl': 'Langi', 'nn': 'langi', 'no': 'langi', 'nqo': 'ߟߊ߲ߖߌߞߊ߲', 'or': 'ଲାନଗି', 'pa': 'ਲੰਗਾਈ', 'pa-Guru': 'ਲੰਗਾਈ', 'pcm': 'Langi Lángwej', 'pl': 'langi', 'ps': 'لنګی', 'pt': 'langi', 'qu': 'Langi Simi', 'rm': 'langi', 'ro': 'langi', 'ru': 'ланго', 'sc': 'langi', 'sd': 'لانگي', 'sd-Arab': 'لانگي', 'si': 'ලංගි', 'sk': 'langi', 'sl': 'langijščina', 'smn': 'langokielâ', 'so': 'Laangi', 'sq': 'langisht', 'sr': 'ланги', 'sr-Cyrl': 'ланги', 'sr-Latn': 'langi', 'sv': 'langi', 'sw': 'Kirangi', 'ta': 'லங்கி', 'te': 'లాంగీ', 'th': 'แลนจี', 'ti': 'ላንጊ', 'tk': 'langi dili', 'to': 'lea fakalangi', 'tr': 'Langi', 'ug': 'لانگىچە', 'uk': 'лангі', 'ur': 'لانگی', 'uz': 'langi', 'uz-Cyrl': 'лангича', 'uz-Latn': 'langi', 'vec': 'langi', 'vi': 'Tiếng Langi', 'xnr': 'लांगि', 'yo': 'Láńgì', 'yrl': 'rãgi', 'yue': '朗吉文', 'yue-Hans': '朗吉文', 'yue-Hant': '朗吉文', 'zh': '朗吉语', 'zh-Hans': '朗吉语', 'zh-Hant': '朗吉文', 'zu': 'isi-Langi'},
    'lah': {'ar': 'اللاهندا', 'ast': 'lahnda', 'az': 'qərbi pəncab', 'az-Latn': 'qərbi pəncab', 'bg': 'лахнда', 'bn': 'লান্ডা', 'br': 'lahnda', 'brx': 'लाह्डां', 'bs': 'landa', 'bs-Cyrl': 'ланда', 'bs-Latn': 'landa', 'ca': 'panjabi occidental', 'ca-ES-valencia': 'panjabi occidental', 'ccp': '𑄣𑄚𑄴𑄓', 'cs': 'lahndština', 'cy': 'Lahnda', 'da': 'lahnda', 'de': 'Lahnda', 'ee': 'lahndagbe', 'el': 'Λάχδα', 'el-polyton': 'Λάχδα', 'en': 'Western Panjabi', 'es': 'lahnda', 'et': 'lahnda', 'fa': 'لاهندا', 'fi': 'lahnda', 'fo': 'lahnda', 'fr': 'lahnda', 'fy': 'Lahnda', 'ga': 'Puinseáibis Iartharach', 'gd': 'Lahnda', 'gsw': 'Lahndanisch', 'gu': 'લાહન્ડા', 'he': 'לנדה', 'hi': 'लाह्न्डा', 'hi-Latn': 'Lahnda', 'hr': 'lahnda', 'ht': 'lahnda', 'hu': 'lahnda', 'id': 'Lahnda', 'is': 'landa', 'it': 'lahnda', 'ja': 'ラフンダー語', 'ka': 'ლანდა', 'kgp': 'rahina', 'kn': 'ಲಹಂಡಾ', 'ko': '라한다어', 'ks': 'لَہَندا', 'ks-Arab': 'لَہَندا', 'ksh': 'de Landa-Schprooche', 'lb': 'Lahnda', 'lkt': 'Lahnda Iyápi', 'lo': 'ລານດາ', 'lt': 'landa', 'lv': 'landu', 'mk': 'ланда', 'ml': 'ലഹ്\u200cൻഡ', 'mr': 'लाह्न्डा', 'ms': 'Lahnda', 'mt': 'Lahnda', 'nb': 'lahnda', 'ne': 'लाहन्डा', 'nl': 'Lahnda', 'nn': 'lahnda', 'no': 'lahnda', 'or': 'ଲାହାଣ୍ଡା', 'pl': 'lahnda', 'pt': 'lahnda', 'rm': 'lahnda', 'ro': 'lahnda', 'ru': 'лахнда', 'sk': 'lahandčina', 'sl': 'landa', 'sr': 'ланда', 'sr-Cyrl': 'ланда', 'sr-Latn': 'landa', 'sv': 'lahnda', 'ta': 'லஹன்டா', 'te': 'లాహండా', 'th': 'ลาฮ์นดา', 'to': 'lea fakapunisapi-hihifoi', 'tr': 'Lahnda', 'ug': 'لانداچە', 'uk': 'ланда', 'vi': 'Tiếng Lahnda', 'yrl': 'rãda', 'yue': '拉亨達文', 'yue-Hans': '拉亨达文', 'yue-Hant': '拉亨達文', 'zh': '西旁遮普语', 'zh-Hans': '西旁遮普语', 'zh-Hant': '拉亨達文', 'lah': 'لہندا پنجابی'},
    'lai': {'en': 'Lambya'},
    'laj': {'en': 'Lango (Uganda)'},
    'lak': {'en': 'Laka (Nigeria)'},
    'lal': {'en': 'Lalia'},
    'lam': {'ar': 'اللامبا', 'ast': 'lamba', 'az': 'lamba', 'az-Latn': 'lamba', 'bg': 'ламба', 'bn': 'লাম্বা', 'br': 'lamba', 'brx': 'लांबा', 'bs': 'lamba', 'bs-Cyrl': 'ламба', 'bs-Latn': 'lamba', 'ca': 'lamba', 'ca-ES-valencia': 'lamba', 'ccp': '𑄣𑄟𑄴𑄝', 'cs': 'lambština', 'cy': 'Lamba', 'da': 'lamba', 'de': 'Lamba', 'el': 'Λάμπα', 'el-polyton': 'Λάμπα', 'en': 'Lamba', 'es': 'lamba', 'et': 'lamba', 'fa': 'لامبا', 'fi': 'lamba', 'fr': 'lamba', 'fy': 'Lamba', 'gd': 'Lamba', 'gsw': 'Lambanisch', 'gu': 'લામ્બા', 'he': 'למבה', 'hi': 'लाम्बा', 'hi-Latn': 'Lamba', 'hr': 'lamba', 'ht': 'lamba', 'hu': 'lamba', 'id': 'Lamba', 'is': 'lamba', 'it': 'lamba', 'ja': 'ランバ語', 'ka': 'ლამბა', 'kgp': 'rỹma', 'kn': 'ಲಂಬಾ', 'ko': '람바어', 'ks': 'لَمبا', 'ks-Arab': 'لَمبا', 'lb': 'Lamba-Sprooch', 'lo': 'ແລມບາ', 'lt': 'lamba', 'lv': 'lambu', 'mk': 'ламба', 'ml': 'ലംബ', 'mr': 'लाम्बा', 'mt': 'Lamba', 'nb': 'lamba', 'ne': 'लाम्बा', 'nl': 'Lamba', 'nn': 'lamba', 'no': 'lamba', 'or': 'ଲାମ୍ବା', 'pl': 'lamba', 'pt': 'lamba', 'rm': 'lamba', 'ro': 'lamba', 'ru': 'ламба', 'sk': 'lamba', 'sl': 'lamba', 'sr': 'ламба', 'sr-Cyrl': 'ламба', 'sr-Latn': 'lamba', 'sv': 'lamba', 'sw': 'Lamba', 'ta': 'லம்பா', 'te': 'లాంబా', 'th': 'แลมบา', 'to': 'lea fakalamipā', 'tr': 'Lamba dili', 'ug': 'لامباچە', 'uk': 'ламба', 'vi': 'Tiếng Lamba', 'yrl': 'rãba', 'yue': '蘭巴文', 'yue-Hans': '兰巴文', 'yue-Hant': '蘭巴文', 'zh': '兰巴语', 'zh-Hans': '兰巴语', 'zh-Hant': '蘭巴文'},
    'lan': {'en': 'Laru'},
    'lap': {'en': 'Laka (Chad)'},
    'laq': {'en': 'Qabiao'},
    'lar': {'en': 'Larteh'},
    'las': {'en': 'Lama (Togo)'},
    'lau': {'en': 'Laba'},
    'law': {'en': 'Lauje'},
    'lax': {'en': 'Tiwa'},
    'lay': {'en': 'Lama Bai'},
    'laz': {'en': 'Aribwatsa'},
    'lb': {'af': 'Luxemburgs', 'ak': 'Lɔsimbɔge kasa', 'am': 'ሉክሰምበርግኛ', 'ar': 'اللكسمبورغية', 'as': 'লাক্সেমবাৰ্গিচ', 'ast': 'luxemburgués', 'az': 'lüksemburq', 'az-Cyrl': 'лүксембург', 'az-Latn': 'lüksemburq', 'bal-Latn': 'Logzemborgi', 'be': 'люксембургская', 'be-tarask': 'люксембургская', 'bg': 'люксембургски', 'bgn': 'لوگزامبورگی', 'bn': 'লুক্সেমবার্গীয়', 'br': 'luksembourgeg', 'brx': 'लाक्जेमबुर्गीश', 'bs': 'luksemburški', 'bs-Cyrl': 'луксембуршки', 'bs-Latn': 'luksemburški', 'ca': 'luxemburguès', 'ca-ES-valencia': 'luxemburguès', 'ccp': '𑄣𑄪𑄇𑄴𑄥𑄬𑄟𑄴𑄝𑄢𑄴𑄉𑄩𑄠𑄧', 'ce': 'люксембургхойн', 'chr': 'ᎸᎦᏏᎻᏋᎢᏍ', 'ckb': 'لوکسەمبورگی', 'cs': 'lucemburština', 'cy': 'Lwcsembwrgeg', 'da': 'luxembourgsk', 'de': 'Luxemburgisch', 'dsb': 'luxemburgšćina', 'dz': 'ལག་ཛམ་བོརྒ་ཁ', 'ee': 'laksembɔggbe', 'el': 'Λουξεμβουργιανά', 'el-polyton': 'Λουξεμβουργιανά', 'en': 'Luxembourgish', 'en-Dsrt': '𐐢𐐲𐐿𐑅𐐯𐑋𐐺𐐲𐑉𐑀𐐮𐑇', 'eo': 'luksemburga', 'es': 'luxemburgués', 'et': 'letseburgi', 'eu': 'luxenburgera', 'fa': 'لوگزامبورگی', 'ff-Adlm': '𞤂𞤵𞥁𞤫𞤲𞤦𞤵𞥅𞤪𞤺𞤭𞤧𞤪𞤫', 'fi': 'luxemburg', 'fil': 'Luxembourgish', 'fo': 'luksemborgskt', 'fr': 'luxembourgeois', 'fur': 'lussemburghês', 'fy': 'Luxemburgs', 'ga': 'Lucsambuirgis', 'gd': 'Lugsamburgais', 'gl': 'luxemburgués', 'gsw': 'Luxemburgisch', 'gu': 'લક્ઝેમબર્ગિશ', 'ha': 'Luxembourgish', 'he': 'לוקסמבורגית', 'hi': 'लग्ज़मबर्गी', 'hi-Latn': 'Luxembourgish', 'hr': 'luksemburški', 'hsb': 'luxemburgšćina', 'ht': 'luxembourgeois', 'hu': 'luxemburgi', 'hy': 'լյուքսեմբուրգերեն', 'ia': 'luxemburgese', 'id': 'Luksemburg', 'ig': 'Luxembourgish', 'is': 'lúxemborgíska', 'it': 'lussemburghese', 'ja': 'ルクセンブルク語', 'jv': 'Luksemburg', 'ka': 'ლუქსემბურგული', 'kea': 'luxemburges', 'kgp': 'rusẽmurgej', 'kk': 'люксембург тілі', 'kk-Cyrl': 'люксембург тілі', 'km': 'លុចសំបួ', 'kn': 'ಲಕ್ಸಂಬರ್ಗಿಷ್', 'ko': '룩셈부르크어', 'kok': 'लक्झेम्बर्गीश', 'kok-Deva': 'लक्झेम्बर्गीश', 'ks': 'لُکھزیمبورگِش', 'ks-Arab': 'لُکھزیمبورگِش', 'ksh': 'Luxemborjesch', 'ku': 'luksembûrgî', 'ky': 'люксембургча', 'lb': 'Lëtzebuergesch', 'lkt': 'Luxembourg Iyápi', 'lld': 'lussemburghesc', 'lo': 'ລັກເຊມບວກກິຊ', 'lrc': 'لوٙکزامبوٙرگی', 'lt': 'liuksemburgiečių', 'lv': 'luksemburgiešu', 'mai': 'लगजमबर्गी', 'mi': 'Rakapuō', 'mk': 'луксембуршки', 'ml': 'ലക്\u200cസംബർഗിഷ്', 'mn': 'люксембург', 'mr': 'लक्झेंबर्गिश', 'ms': 'Luxembourg', 'mt': 'Lussemburgiż', 'my': 'လူဇင်ဘတ်', 'mzn': 'لوکزامبورگی', 'nb': 'luxemburgsk', 'ne': 'लक्जेम्बर्गी', 'nl': 'Luxemburgs', 'nn': 'luxemburgsk', 'no': 'luxemburgsk', 'nqo': 'ߟߎߞߑߛߊ߲ߓߎ߯ߙߎߞߊ߲', 'or': 'ଲକ୍ସେମବର୍ଗିସ୍', 'pa': 'ਲਕਜ਼ਮਬਰਗਿਸ਼', 'pa-Guru': 'ਲਕਜ਼ਮਬਰਗਿਸ਼', 'pcm': 'Lọ́ksémbọ́g Lángwej', 'pl': 'luksemburski', 'ps': 'لوګزامبورګي', 'pt': 'luxemburguês', 'qu': 'Luxemburgues Simi', 'rm': 'luxemburgais', 'ro': 'luxemburgheză', 'ru': 'люксембургский', 'sc': 'lussemburghesu', 'sd': 'لگزمبرگ', 'sd-Arab': 'لگزمبرگ', 'se': 'luxemburggagiella', 'si': 'ලක්සැම්බර්ග්', 'sk': 'luxemburčina', 'sl': 'luksemburščina', 'smn': 'luxemburgkielâ', 'so': 'Luksaamboorgish', 'sq': 'luksemburgisht', 'sr': 'луксембуршки', 'sr-Cyrl': 'луксембуршки', 'sr-Latn': 'luksemburški', 'sv': 'luxemburgiska', 'sw': 'Kilasembagi', 'ta': 'லக்ஸம்போர்கிஷ்', 'te': 'లక్సెంబర్గిష్', 'tg': 'люксембургӣ', 'th': 'ลักเซมเบิร์ก', 'ti': 'ሉክሰምበርግኛ', 'tk': 'lýuksemburg dili', 'to': 'lea fakalakisimipeki', 'tr': 'Lüksemburgca', 'tt': 'люксембург', 'ug': 'لىيۇكسېمبۇرگچە', 'uk': 'люксембурзька', 'ur': 'لکسمبرگیش', 'uz': 'lyuksemburgcha', 'uz-Cyrl': 'люксембургча', 'uz-Latn': 'lyuksemburgcha', 'vec': 'lusenburgheze', 'vi': 'Tiếng Luxembourg', 'wae': 'Luxemburgiš', 'wo': 'Liksàmbursuwaa', 'xnr': 'लग्ज़मबर्गी', 'yi': 'לוקסעמבורגיש', 'yo': 'Lùṣẹ́mbọ́ọ̀gì', 'yrl': 'ruxẽbugei', 'yue': '盧森堡文', 'yue-Hans': '卢森堡文', 'yue-Hant': '盧森堡文', 'zh': '卢森堡语', 'zh-Hans': '卢森堡语', 'zh-Hant': '盧森堡文', 'zu': 'isi-Luxembourgish'},
    'lba': {'en': 'Lui'},
    'lbb': {'en': 'Label'},
    'lbc': {'en': 'Lakkia'},
    'lbe': {'en': 'Lak'},
    'lbf': {'en': 'Tinani'},
    'lbg': {'en': 'Laopang'},
    'lbi': {'en': "La'bi"},
    'lbj': {'en': 'Ladakhi'},
    'lbk': {'en': 'Central Bontok'},
    'lbl': {'en': 'Libon Bikol'},
    'lbm': {'en': 'Lodhi'},
    'lbn': {'en': 'Rmeet'},
    'lbo': {'en': 'Laven'},
    'lbq': {'en': 'Wampar'},
    'lbr': {'en': 'Lohorung'},
    'lbs': {'en': 'Libyan Sign Language'},
    'lbt': {'en': 'Lachi'},
    'lbu': {'en': 'Labu'},
    'lbv': {'en': 'Lavatbura-Lamusong'},
    'lbw': {'en': 'Tolaki'},
    'lbx': {'en': 'Lawangan'},
    'lby': {'en': 'Lamalama'},
    'lbz': {'en': 'Lardil'},
    'lcc': {'en': 'Legenyem'},
    'lcd': {'en': 'Lola'},
    'lce': {'en': 'Loncong'},
    'lcf': {'en': 'Lubu'},
    'lch': {'en': 'Luchazi'},
    'lcl': {'en': 'Lisela'},
    'lcm': {'en': 'Tungag'},
    'lcp': {'en': 'Western Lawa'},
    'lcq': {'en': 'Luhu'},
    'lcs': {'en': 'Lisabata-Nuniali'},
    'lda': {'en': 'Kla-Dan'},
    'ldb': {'en': 'Dũya'},
    'ldd': {'en': 'Luri'},
    'ldg': {'en': 'Lenyima'},
    'ldh': {'en': 'Lamja-Dengsa-Tola'},
    'ldi': {'en': 'Laari'},
    'ldj': {'en': 'Lemoro'},
    'ldk': {'en': 'Leelau'},
    'ldl': {'en': 'Kaan'},
    'ldm': {'en': 'Landoma'},
    'ldn': {'en': 'Láadan'},
    'ldo': {'en': 'Loo'},
    'ldp': {'en': 'Tso'},
    'ldq': {'en': 'Lufu'},
    'lea': {'en': 'Lega-Shabunda'},
    'leb': {'en': 'Lala-Bisa'},
    'lec': {'en': 'Leco'},
    'led': {'en': 'Lendu'},
    'lee': {'en': 'Lyélé'},
    'lef': {'en': 'Lelemi'},
    'leg': {'en': 'Lengua'},
    'leh': {'en': 'Lenje'},
    'lei': {'en': 'Lemio'},
    'lej': {'en': 'Lengola'},
    'lek': {'en': 'Leipon'},
    'lel': {'en': 'Lele (Democratic Republic of Congo)'},
    'lem': {'en': 'Nomaande'},
    'len': {'en': 'Lenca'},
    'leo': {'en': 'Leti (Cameroon)'},
    'lep': {'en': 'Lepcha'},
    'leq': {'en': 'Lembena'},
    'ler': {'en': 'Lenkau'},
    'les': {'en': 'Lese'},
    'let': {'en': 'Lesing-Gelimi'},
    'leu': {'en': 'Kara (Papua New Guinea)'},
    'lev': {'en': 'Lamma'},
    'lew': {'en': 'Ledo Kaili'},
    'lex': {'en': 'Luang'},
    'ley': {'en': 'Lemolang'},
    'lez': {'af': 'Lezghies', 'am': 'ሌዝጊያን', 'ar': 'الليزجية', 'as': 'লেজঘিয়ান', 'ast': 'lezghianu', 'az': 'ləzgi', 'az-Cyrl': 'ләзҝи', 'az-Latn': 'ləzgi', 'be': 'лезгінская', 'be-tarask': 'лезгінская', 'bg': 'лезгински', 'bgn': 'لزگی', 'bn': 'লেজঘিয়ান', 'br': 'lezgi', 'brx': 'लेजघीयान', 'bs': 'lezgijski', 'bs-Cyrl': 'лезгиан', 'bs-Latn': 'lezgijski', 'ca': 'lesguià', 'ca-ES-valencia': 'lesguià', 'ccp': '𑄣𑄬𑄎𑄴𑄊𑄨𑄠𑄚𑄴', 'ce': 'лаьзгийн', 'chr': 'ᎴᏏᎦᏂ', 'ckb': 'لەزگی', 'cs': 'lezginština', 'cy': 'Lezgheg', 'da': 'lezghian', 'de': 'Lesgisch', 'dsb': 'lezgianšćina', 'el': 'Λεζγκικά', 'el-polyton': 'Λεζγκικά', 'en': 'Lezghian', 'es': 'lezgiano', 'et': 'lesgi', 'eu': 'lezginera', 'fa': 'لزگی', 'ff-Adlm': '𞤂𞤫𞥁𞤺𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'lezgi', 'fil': 'Lezghian', 'fo': 'lezghian', 'fr': 'lezghien', 'fy': 'Lezgysk', 'ga': 'Leisgis', 'gd': 'Leasgais', 'gl': 'lezguio', 'gsw': 'Lesgisch', 'gu': 'લેઝધીયન', 'ha': 'Lezghiniyanci', 'he': 'לזגית', 'hi': 'लेज़्घीयन', 'hi-Latn': 'Lezghian', 'hr': 'lezgiški', 'hsb': 'lezgianšćina', 'ht': 'lezghien', 'hu': 'lezg', 'hy': 'լեզգիերեն', 'ia': 'lezghiano', 'id': 'Lezghia', 'ig': 'Lezghian', 'is': 'lesgíska', 'it': 'lesgo', 'ja': 'レズギ語', 'jv': 'Lesghian', 'ka': 'ლეზგიური', 'kgp': 'resgi', 'kk': 'лезгин тілі', 'kk-Cyrl': 'лезгин тілі', 'km': 'ឡេសហ្គី', 'kn': 'ಲೆಜ್ಘಿಯನ್', 'ko': '레즈기안어', 'kok': 'लेझघियान', 'kok-Deva': 'लेझघियान', 'ks': 'لیزگِیَن', 'ks-Arab': 'لیزگِیَن', 'ksh': 'Lesjesch', 'ku': 'lezgînî', 'ky': 'лезгинче', 'lb': 'Lesgesch', 'lo': 'ລີຊຽນ', 'lt': 'lezginų', 'lv': 'lezgīnu', 'mai': 'लजगियान', 'mi': 'Rēhiana', 'mk': 'лезгински', 'ml': 'ലസ്ഗിയൻ', 'mn': 'лезги', 'mr': 'लेझ्घीयन', 'ms': 'Lezghian', 'mt': 'Leżgjan', 'my': 'လက်ဇ်ဂီးယား', 'nb': 'lesgisk', 'ne': 'लाज्घियाली', 'nl': 'Lezgisch', 'nn': 'lezghian', 'no': 'lesgisk', 'or': 'ଲେଜଗିୟାନ୍', 'os': 'лекъаг', 'pa': 'ਲੈਜ਼ਗੀ', 'pa-Guru': 'ਲੈਜ਼ਗੀ', 'pcm': 'Lẹzghián Lángwej', 'pl': 'lezgijski', 'ps': 'لیګغیان', 'pt': 'lezgui', 'qu': 'Lezghian Simi', 'rm': 'lezghian', 'ro': 'lezghian', 'ru': 'лезгинский', 'sc': 'lezghianu', 'sd': 'ليزگهين', 'sd-Arab': 'ليزگهين', 'si': 'ලෙස්ගියන්', 'sk': 'lezginčina', 'sl': 'lezginščina', 'smn': 'lezgikielâ', 'so': 'Lezghian', 'sq': 'lezgianisht', 'sr': 'лезгински', 'sr-Cyrl': 'лезгински', 'sr-Latn': 'lezginski', 'sv': 'lezghien', 'sw': 'Kilezighian', 'ta': 'லெஜ்ஜியன்', 'te': 'లేజ్ఘియన్', 'th': 'เลซเกีย', 'ti': 'ለዝግኛ', 'tk': 'lezgin dili', 'to': 'lea fakalesikia', 'tr': 'Lezgice', 'ug': 'لېزگىنچە', 'uk': 'лезгінська', 'ur': 'لیزگیان', 'uz': 'lezgin', 'uz-Latn': 'lezgin', 'vec': 'lezghian', 'vi': 'Tiếng Lezghian', 'xnr': 'लेज़्घीयन', 'yo': 'Èdè Lesgina', 'yrl': 'resigi', 'yue': '列茲干文', 'yue-Hans': '列兹干文', 'yue-Hant': '列茲干文', 'zh': '列兹金语', 'zh-Hans': '列兹金语', 'zh-Hant': '列茲干文', 'zu': 'isi-Lezghian'},
    'lfa': {'en': 'Lefa'},
    'lfn': {'ast': 'lingua franca nova', 'br': 'lingua franca nova', 'cs': 'lingua franca nova', 'de': 'Lingua Franca Nova', 'en': 'Lingua Franca Nova', 'fi': 'lingua franca nova', 'fr': 'lingua franca nova', 'gd': 'Lingua Franca Nova', 'gu': 'લિંગ્વા ફેન્કા નોવા', 'hi-Latn': 'Lingua Franca Nova', 'ht': 'lingua franca nova', 'it': 'Lingua Franca Nova', 'ja': 'リングア・フランカ・ノバ', 'ko': '링구아 프랑카 노바', 'lb': 'Lingua Franca Nova', 'lt': 'naujoji frankų kalba', 'mk': 'лингва франка нова', 'nb': 'lingua franca nova', 'ne': 'लिङ्गुवा फ्राङ्का नोभा', 'nl': 'Lingua Franca Nova', 'nn': 'lingua franca nova', 'no': 'lingua franca nova', 'pl': 'Lingua Franca Nova', 'sv': 'lingua franca nova', 'th': 'ลิงกัวฟรังกาโนวา', 'to': 'lea fakakavakava-foʻou', 'tr': 'Lingua Franca Nova', 'yue': '新共同語言', 'yue-Hans': '新共同语言', 'yue-Hant': '新共同語言', 'zh-Hant': '新共同語言'},
    'lg': {'af': 'Ganda', 'am': 'ጋንዳኛ', 'ar': 'الغاندا', 'as': 'গান্দা', 'ast': 'ganda', 'az': 'qanda', 'az-Arab': 'قانداجا', 'az-Cyrl': 'ганда', 'az-Latn': 'qanda', 'bal-Latn': 'Gandá', 'be': 'ганда', 'be-tarask': 'ганда', 'bg': 'ганда', 'bgn': 'گاندایی', 'bn': 'গান্ডা', 'br': 'ganda', 'brx': 'गान्डा', 'bs': 'ganda', 'bs-Cyrl': 'ганда', 'bs-Latn': 'ganda', 'ca': 'ganda', 'ca-ES-valencia': 'ganda', 'ccp': '𑄉𑄚𑄴𑄓', 'ce': 'ганда', 'chr': 'ᎦᏂᏓ', 'ckb': 'گاندا', 'cs': 'gandština', 'cy': 'Ganda', 'da': 'ganda', 'de': 'Ganda', 'dsb': 'gandšćina', 'el': 'Γκάντα', 'el-polyton': 'Γκάντα', 'en': 'Ganda', 'es': 'ganda', 'et': 'ganda', 'eu': 'luganda', 'fa': 'گاندایی', 'ff-Adlm': '𞤘𞤢𞤲𞤣𞤢𞥄𞤪𞤫', 'fi': 'ganda', 'fil': 'Ganda', 'fo': 'ganda', 'fr': 'ganda', 'fy': 'Ganda', 'ga': 'Lugandais', 'gd': 'Ganda', 'gl': 'ganda', 'gsw': 'Ganda-Schpraach', 'gu': 'ગાંડા', 'ha': 'Ganda', 'he': 'גאנדה', 'hi': 'गांडा', 'hi-Latn': 'Ganda', 'hr': 'ganda', 'hsb': 'gandšćina', 'ht': 'ganda', 'hu': 'ganda', 'hy': 'գանդա', 'ia': 'luganda', 'id': 'Ganda', 'ig': 'Ganda', 'is': 'ganda', 'it': 'ganda', 'ja': 'ガンダ語', 'jv': 'Ganda', 'ka': 'განდა', 'kea': 'luganda', 'kgp': 'rugỹna', 'kk': 'ганда тілі', 'kk-Cyrl': 'ганда тілі', 'km': 'ហ្កាន់ដា', 'kn': 'ಗಾಂಡಾ', 'ko': '간다어', 'kok': 'गांडा', 'kok-Deva': 'गांडा', 'ks': 'گاندا', 'ks-Arab': 'گاندا', 'ksh': 'Lugannda', 'ku': 'lugandayî', 'ky': 'гандача', 'lb': 'Ganda-Sprooch', 'lg': 'Luganda', 'lld': 'ganda', 'lo': 'ແກນດາ', 'lrc': 'گاندا', 'lt': 'ganda', 'lv': 'gandu', 'mai': 'गंडा', 'mi': 'Kānata', 'mk': 'ганда', 'ml': 'ഗാണ്ട', 'mn': 'ганда', 'mr': 'गांडा', 'ms': 'Ganda', 'mt': 'Ganda', 'my': 'ဂန်ဒါ', 'mzn': 'گاندا', 'nb': 'ganda', 'ne': 'गान्डा', 'nl': 'Luganda', 'nn': 'ganda', 'no': 'ganda', 'nqo': 'ߜ߭ߊ߲ߘߊߞߊ߲', 'or': 'ଗନ୍ଦା', 'pa': 'ਗਾਂਡਾ', 'pa-Guru': 'ਗਾਂਡਾ', 'pcm': 'Gánda Lángwej', 'pl': 'ganda', 'ps': 'ګانده', 'pt': 'luganda', 'qu': 'Luganda Simi', 'rm': 'ganda', 'ro': 'ganda', 'ru': 'ганда', 'sc': 'ganda', 'sd': 'گاندا', 'sd-Arab': 'گاندا', 'si': 'ගන්ඩා', 'sk': 'gandčina', 'sl': 'ganda', 'smn': 'luganda', 'so': 'Gandha', 'sq': 'gandaisht', 'sr': 'ганда', 'sr-Cyrl': 'ганда', 'sr-Latn': 'ganda', 'sv': 'luganda', 'sw': 'Kiganda', 'syr': 'ܓܢܕܝܬ', 'ta': 'கான்டா', 'te': 'గాండా', 'th': 'ยูกันดา', 'ti': 'ጋንዳ', 'tk': 'ganda dili', 'to': 'lea fakakanita', 'tr': 'Ganda', 'ug': 'گانداچە', 'uk': 'ганда', 'ur': 'گینڈا', 'uz': 'ganda', 'uz-Cyrl': 'гандача', 'uz-Latn': 'ganda', 'vec': 'ganda', 'vi': 'Tiếng Ganda', 'xnr': 'गांडा', 'yo': 'Ganda', 'yrl': 'rugãda', 'yue': '干達文', 'yue-Hans': '干达文', 'yue-Hant': '干達文', 'zh': '卢干达语', 'zh-Hans': '卢干达语', 'zh-Hant': '干達文', 'zu': 'isi-Ganda'},
    'lga': {'en': 'Lungga'},
    'lgb': {'en': 'Laghu'},
    'lgg': {'en': 'Lugbara'},
    'lgh': {'en': 'Laghuu'},
    'lgi': {'en': 'Lengilu'},
    'lgk': {'en': 'Lingarak'},
    'lgl': {'en': 'Wala'},
    'lgm': {'en': 'Lega-Mwenga'},
    'lgn': {'en': "T'apo"},
    'lgq': {'en': 'Logba'},
    'lgr': {'en': 'Lengo'},
    'lgt': {'en': 'Pahi'},
    'lgu': {'en': 'Longgu'},
    'lgz': {'en': 'Ligenza'},
    'lha': {'en': 'Laha (Viet Nam)'},
    'lhh': {'en': 'Laha (Indonesia)'},
    'lhi': {'en': 'Lahu Shi'},
    'lhl': {'en': 'Lahul Lohar'},
    'lhm': {'en': 'Lhomi'},
    'lhn': {'en': 'Lahanan'},
    'lhp': {'en': 'Lhokpu'},
    'lhs': {'en': 'Mlahsö'},
    'lht': {'en': 'Lo-Toga'},
    'lhu': {'en': 'Lahu'},
    'li': {'af': 'Limburgs', 'am': 'ሊምቡርጊሽ', 'ar': 'الليمبورغية', 'as': 'লিম্বুৰ্গিচ', 'ast': 'limburgués', 'az': 'limburq', 'az-Cyrl': 'лимбург', 'az-Latn': 'limburq', 'be': 'лімбургская', 'be-tarask': 'лімбургская', 'bg': 'лимбургски', 'bgn': 'لیمبورگی', 'bn': 'লিম্বুর্গিশ', 'br': 'limbourgeg', 'brx': 'लीम्बुर्गिस', 'bs': 'limburški', 'bs-Cyrl': 'лимбургиш', 'bs-Latn': 'limburški', 'ca': 'limburguès', 'ca-ES-valencia': 'limburguès', 'ccp': '𑄣𑄨𑄟𑄴𑄝𑄪𑄢𑄴𑄉𑄨𑄌𑄴', 'ce': 'лимбургийн', 'chr': 'ᎴᎹᏊᎵᏏ', 'ckb': 'لیمبورگی', 'cs': 'limburština', 'cy': 'Limbwrgeg', 'da': 'limburgsk', 'de': 'Limburgisch', 'dsb': 'limburšćina', 'el': 'Λιμβουργιανά', 'el-polyton': 'Λιμβουργιανά', 'en': 'Limburgish', 'es': 'limburgués', 'et': 'limburgi', 'eu': 'limburgera', 'fa': 'لیمبورگی', 'ff-Adlm': '𞤂𞤭𞤥𞤦𞤵𞤪𞤺𞤵𞤧𞤪𞤫', 'fi': 'limburg', 'fil': 'Limburgish', 'fo': 'limburgiskt', 'fr': 'limbourgeois', 'fur': 'limburghês', 'fy': 'Limburgs', 'ga': 'Liombuirgis', 'gd': 'Cànan Limburg', 'gl': 'limburgués', 'gsw': 'Limburgisch', 'gu': 'લિંબૂર્ગિશ', 'ha': 'Limburgish', 'he': 'לימבורגית', 'hi': 'लिंबर्गिश', 'hi-Latn': 'Limburgish', 'hr': 'limburški', 'hsb': 'limburšćina', 'ht': 'limbourgeois', 'hu': 'limburgi', 'hy': 'լիմբուրգերեն', 'ia': 'limburgese', 'id': 'Limburgia', 'ig': 'Limburgish', 'is': 'limbúrgíska', 'it': 'limburghese', 'ja': 'リンブルフ語', 'jv': 'Limburgish', 'ka': 'ლიმბურგული', 'kgp': 'rĩmurgej', 'kk': 'лимбург тілі', 'kk-Cyrl': 'лимбург тілі', 'km': 'លីមប៊ូស', 'kn': 'ಲಿಂಬರ್ಗಿಶ್', 'ko': '림버거어', 'kok': 'लिंबर्गिश', 'kok-Deva': 'लिंबर्गिश', 'ks': 'لِمبٔرگِش', 'ks-Arab': 'لِمبٔرگِش', 'ksh': 'Lemburjesch', 'ku': 'lîmbûrgî', 'ky': 'лимбургиче', 'lb': 'Limburgesch', 'lo': 'ລິມເບີກີຊ', 'lt': 'limburgiečių', 'lv': 'limburgiešu', 'mai': 'लिमबर्गिश', 'mi': 'Ripūkuihi', 'mk': 'лимбуршки', 'ml': 'ലിംബർഗിഷ്', 'mn': 'лимбург', 'mr': 'लिंबूर्गिश', 'ms': 'Limburgish', 'mt': 'Limburgish', 'my': 'လင်ဘာဂစ်ရှ်', 'nb': 'limburgsk', 'ne': 'लिम्बुर्गी', 'nl': 'Limburgs', 'nn': 'limburgsk', 'no': 'limburgsk', 'or': 'ଲିମ୍ବୁର୍ଗିସ୍', 'pa': 'ਲਿਮਬੁਰਗੀ', 'pa-Guru': 'ਲਿਮਬੁਰਗੀ', 'pcm': 'Limbógísh Lángwej', 'pl': 'limburski', 'ps': 'لمبرگیانی', 'pt': 'limburguês', 'qu': 'Limburgues Simi', 'rm': 'limburgais', 'ro': 'limburgheză', 'ru': 'лимбургский', 'sc': 'limburghesu', 'sd': 'لمبرگش', 'sd-Arab': 'لمبرگش', 'si': 'ලිම්බර්ගිශ්', 'sk': 'limburčina', 'sl': 'limburščina', 'smn': 'limburgkielâ', 'so': 'Limburgish', 'sq': 'limburgisht', 'sr': 'лимбуршки', 'sr-Cyrl': 'лимбуршки', 'sr-Latn': 'limburški', 'sv': 'limburgiska', 'sw': 'Limburgish', 'ta': 'லிம்பர்கிஷ்', 'te': 'లిమ్బర్గిష్', 'th': 'ลิมเบิร์ก', 'ti': 'ሊምበርግኛ', 'tk': 'limburg dili', 'to': 'lea fakalimipūliki', 'tr': 'Limburgca', 'ug': 'لىمبۇرگچە', 'uk': 'лімбургійська', 'ur': 'لیمبرگش', 'uz': 'limburg', 'uz-Latn': 'limburg', 'vec': 'linburgheze', 'vi': 'Tiếng Limburg', 'xnr': 'लिंबर्गिश', 'yo': 'Èdè Limbogishi', 'yrl': 'rĩburgei', 'yue': '林堡文', 'yue-Hans': '林堡文', 'yue-Hant': '林堡文', 'zh': '林堡语', 'zh-Hans': '林堡语', 'zh-Hant': '林堡文', 'zu': 'isi-Limburgish'},
    'lia': {'en': 'West-Central Limba'},
    'lib': {'en': 'Likum'},
    'lic': {'en': 'Hlai'},
    'lid': {'en': 'Nyindrou'},
    'lie': {'en': 'Likila'},
    'lif': {'en': 'Limbu'},
    'lig': {'en': 'Ligbi'},
    'lih': {'en': 'Lihir'},
    'lii': {'en': 'Lingkhim'},
    'lij': {'af': 'Liguries', 'ak': 'Liguria kasa', 'am': 'ሊጓሪያኛ', 'ar': 'الليغورية', 'as': 'লিংগুৰিয়ান', 'ast': 'ligurianu', 'az': 'liquriya dili', 'az-Latn': 'liquriya dili', 'bal-Latn': 'Liguri', 'be': 'лігурская', 'be-tarask': 'лігурская', 'bg': 'лигурски', 'bn': 'লিগুরিয়ান', 'br': 'ligurieg', 'bs': 'ligurski', 'bs-Cyrl': 'линуриан', 'bs-Latn': 'ligurski', 'ca': 'lígur', 'ca-ES-valencia': 'lígur', 'chr': 'ᎵᏇᎢᎠ', 'cs': 'ligurština', 'cy': 'Ligwreg', 'da': 'ligurisk', 'de': 'Ligurisch', 'dsb': 'liguriańšćina', 'el': 'Λιγουριανά', 'el-polyton': 'Λιγουριανά', 'en': 'Ligurian', 'es': 'ligur', 'et': 'liguuri', 'eu': 'liguriera', 'fa': 'لیگوری', 'ff-Adlm': '𞤂𞤳𞤭𞤺𞤵𞥅𞤪𞤫', 'fi': 'liguuri', 'fil': 'Ligurian', 'fo': 'liguriskt', 'fr': 'ligure', 'ga': 'Liogúiris', 'gd': 'Liogùrais', 'gl': 'lígur', 'gu': 'લિગુરીઅન', 'ha': 'Liguriyanci', 'he': 'ליגורית', 'hi': 'लिगुरियन', 'hi-Latn': 'Ligurian', 'hr': 'ligurski', 'hsb': 'ligurianšćina', 'ht': 'ligure', 'hu': 'ligur', 'hy': 'լիգուրերեն', 'ia': 'ligure', 'id': 'Liguria', 'ie': 'ligurian', 'ig': 'Ligurian', 'is': 'lígúríska', 'it': 'ligure', 'ja': 'リグリア語', 'jv': 'Liguria', 'ka': 'ლიგურიული', 'kk': 'лигур тілі', 'kk-Cyrl': 'лигур тілі', 'km': 'លីគូរី', 'kn': 'ಲಿಗುರಿಯನ್', 'ko': '리구리아어', 'kok': 'लिगुरियन', 'kok-Deva': 'लिगुरियन', 'ku': 'lîgûrî', 'ky': 'лигурча', 'lb': 'Liguresch', 'lij': 'ligure', 'lo': 'ລີກູຣຽນ', 'lt': 'ligūrų', 'lv': 'ligūriešu', 'mk': 'лигурски', 'ml': 'ലിഗൂറിയൻ', 'mn': 'лигури', 'mr': 'लिगुरिअन', 'ms': 'Liguria', 'my': 'လက်ဂါးရီရန်', 'nb': 'ligurisk', 'ne': 'लिगुरियाली', 'nl': 'Ligurisch', 'nn': 'ligurisk', 'no': 'ligurisk', 'nqo': 'ߟߌ߭ߎ߳ߙߌߞߊ߲', 'or': 'ଲିଗୁରିଆନ୍\u200c', 'pa': 'ਲਿਗੂਰੀ', 'pa-Guru': 'ਲਿਗੂਰੀ', 'pcm': 'Ligurián Lángwej', 'pl': 'liguryjski', 'ps': 'لینګورین', 'pt': 'ligure', 'ro': 'liguriană', 'ru': 'лигурский', 'sc': 'lìgure', 'sd': 'لگيوريئن', 'sd-Arab': 'لگيوريئن', 'si': 'ලිගුරියන්', 'sk': 'ligurčina', 'sl': 'ligurščina', 'so': 'Liguuriyaan', 'sq': 'ligurianisht', 'sr': 'лигурски', 'sr-Cyrl': 'лигурски', 'sr-Latn': 'ligurski', 'sv': 'liguriska', 'sw': 'Kiliguria', 'ta': 'லிகூரியன்', 'te': 'లిగూరియన్', 'th': 'ลิกูเรีย', 'ti': 'ሊጉርኛ', 'tk': 'ligur dili', 'to': 'lea fakalikulia', 'tr': 'Ligurca', 'uk': 'лігурійська', 'ur': 'لیگوریائی', 'uz': 'liguryan', 'uz-Latn': 'liguryan', 'vec': 'lìguro', 'vi': 'Tiếng Liguria', 'yo': 'Liguriani', 'yue': '利古里亞文', 'yue-Hans': '利古里亚文', 'yue-Hant': '利古里亞文', 'zh': '利古里亚语', 'zh-Hans': '利古里亚语', 'zh-Hant': '利古里亞文', 'zu': 'IsiLigurian'},
    'lik': {'en': 'Lika'},
    'lil': {'af': 'Lillooet', 'am': 'ሊሎኤት', 'ar': 'الليلويتية', 'as': 'লিল্লোৱেট', 'az': 'Liluet', 'az-Latn': 'Liluet', 'be': 'лілуэт', 'be-tarask': 'лілуэт', 'bg': 'лилоует', 'bn': 'লিল্লুয়েট', 'br': 'lillooet', 'brx': 'लिल्लुवेत', 'bs': 'liluet', 'bs-Cyrl': 'лилоет', 'bs-Latn': 'liluet', 'ca': 'lillooet', 'ca-ES-valencia': 'lillooet', 'chr': 'ᎵᎶᎡᏘ', 'cs': 'lillooetština', 'cy': 'Lillooet', 'da': 'lillooet', 'de': 'Lillooet', 'dsb': 'lillooetšćina', 'el': 'Λιλουέτ', 'el-polyton': 'Λιλουέτ', 'en': 'Lillooet', 'es': 'lillooet', 'et': 'lillueti', 'eu': 'lillooetera', 'fa': 'لیلوئت', 'ff-Adlm': '𞤂𞤭𞤤𞥆𞤮𞥅𞤫𞤼𞤪𞤫', 'fi': 'lillooet', 'fil': 'Lillooet', 'fr': 'lillooet', 'ga': 'Lillooet', 'gd': 'Lillooet', 'gl': 'lillooet', 'gu': 'લિલુએટ', 'ha': 'Lillooet', 'he': 'לילואט', 'hi': 'लिलोएट', 'hi-Latn': 'Lillooet', 'hr': 'lillooet', 'hsb': 'lillooetšćina', 'ht': 'lillooet', 'hu': 'lillooet', 'hy': 'լիլուետ', 'ia': 'lillooet', 'id': 'Lillooet', 'ig': 'Liloetị', 'is': 'lillooet', 'it': 'lillooet', 'ja': 'リルエット語', 'jv': 'Lillooet', 'ka': 'ლილიეტი', 'kk': 'лиллуэт тілі', 'kk-Cyrl': 'лиллуэт тілі', 'km': 'លីលលូអេត', 'kn': 'ಲಿಲ್ಲೂವೆಟ್', 'ko': '릴루엣어', 'kok': 'लल्युएत', 'kok-Deva': 'लल्युएत', 'ku': 'lillooet', 'ky': 'лиллуэтче (индей тили)', 'lo': 'ລິນລູເອັດ', 'lt': 'liluetų', 'lv': 'lilluetu', 'mai': 'लिलुइट', 'mi': 'Riruete', 'mk': 'лилуетски', 'ml': 'ലില്ലുവെറ്റ്', 'mn': 'лиллуэт', 'mr': 'लिलूएट', 'ms': 'Lillooet', 'my': 'လာလူးဝစ်တ်', 'nb': 'lillooet', 'ne': 'लिलुएट', 'nl': 'Lillooet', 'nn': 'lillooet', 'no': 'lillooet', 'or': 'ଲିଲ୍ଲୁଏଟ', 'pa': 'ਲਿਲੂਏਟ', 'pa-Guru': 'ਲਿਲੂਏਟ', 'pcm': 'Lillooẹ́t Lángwej', 'pl': 'lillooet', 'ps': 'lill', 'pt': 'lillooet', 'qu': 'Lillooet Simi', 'ro': 'lillooet', 'ru': 'лиллуэт', 'sc': 'lillooet', 'sd': 'ليلوئيٽ', 'sd-Arab': 'ليلوئيٽ', 'si': 'ලිලූට්', 'sk': 'lillooet', 'sl': 'lilovetščina', 'so': 'Lillooet', 'sq': 'lilluetisht', 'sr': 'лилут', 'sr-Cyrl': 'лилут', 'sr-Latn': 'lilut', 'sv': 'lillooet', 'sw': 'Kilillooet', 'ta': 'லில்லூயிட்', 'te': 'లిలూయెట్', 'th': 'ลิลลูเอต', 'ti': 'ሊሉት', 'tk': 'lilluet dili', 'to': 'lea fakalilōeti', 'tr': 'Lillooet dili', 'uk': 'лілуетська', 'ur': 'للوئیٹ', 'uz': 'lilluet', 'uz-Latn': 'lilluet', 'vec': 'liluet', 'vi': 'Tiếng Lillooet', 'yo': 'Èdè Liloeti', 'yue': '利盧埃特文', 'yue-Hans': '利卢埃特文', 'yue-Hant': '利盧埃特文', 'zh': '利洛埃特语', 'zh-Hans': '利洛埃特语', 'zh-Hant': '利洛威特文', 'zu': 'isi-Lillooet'},
    'lio': {'en': 'Liki'},
    'lip': {'en': 'Sekpele'},
    'liq': {'en': 'Libido'},
    'lir': {'en': 'Liberian English'},
    'lis': {'en': 'Lisu'},
    'liu': {'en': 'Logorik'},
    'liv': {'ast': 'livonianu', 'cs': 'livonština', 'de': 'Livisch', 'en': 'Livonian', 'et': 'liivi', 'fi': 'liivi', 'fr': 'livonien', 'ga': 'Liovóinis', 'hi-Latn': 'Livonian', 'ht': 'livonien', 'it': 'livone', 'ja': 'リヴォニア語', 'lb': 'Livesch', 'lt': 'lyvių', 'mk': 'ливонски', 'nb': 'livisk', 'ne': 'लिभोनियाली', 'nl': 'Lijfs', 'nn': 'livisk', 'no': 'livisk', 'pl': 'liwski', 'sv': 'livoniska', 'th': 'ลิโวเนีย', 'to': 'lea fakalivonia', 'tr': 'Livonca', 'yi': 'ליוויש', 'yue': '利伏尼亞文', 'yue-Hans': '利伏尼亚文', 'yue-Hant': '利伏尼亞文', 'zh-Hant': '利伏尼亞文'},
    'liw': {'en': 'Col'},
    'lix': {'en': 'Liabuku'},
    'liy': {'en': 'Banda-Bambari'},
    'liz': {'en': 'Libinza'},
    'lja': {'en': 'Golpa'},
    'lje': {'en': 'Rampi'},
    'lji': {'en': 'Laiyolo'},
    'ljl': {'en': "Li'o"},
    'ljp': {'en': 'Lampung Api'},
    'ljw': {'en': 'Yirandali'},
    'ljx': {'en': 'Yuru'},
    'lka': {'en': 'Lakalei'},
    'lkb': {'en': 'Kabras'},
    'lkc': {'en': 'Kucong'},
    'lkd': {'en': 'Lakondê'},
    'lke': {'en': 'Kenyi'},
    'lkh': {'en': 'Lakha'},
    'lki': {'en': 'Laki'},
    'lkj': {'en': 'Remun'},
    'lkl': {'en': 'Laeko-Libuat'},
    'lkm': {'en': 'Kalaamaya'},
    'lkn': {'en': 'Lakon'},
    'lko': {'en': 'Khayo'},
    'lkr': {'en': 'Päri'},
    'lks': {'en': 'Kisa'},
    'lkt': {'af': 'Lakota', 'am': 'ላኮታ', 'ar': 'لاكوتا', 'as': 'লাকোটা', 'ast': 'lakota', 'az': 'lakota', 'az-Cyrl': 'лакота', 'az-Latn': 'lakota', 'bal-Latn': 'Lakótá', 'be': 'лакота', 'be-tarask': 'лакота', 'bg': 'лакота', 'bgn': 'لاکوتا', 'bn': 'লাকোটা', 'br': 'lakota', 'brx': 'लाक’ता', 'bs': 'lakota', 'bs-Cyrl': 'лакота', 'bs-Latn': 'lakota', 'ca': 'lakota', 'ca-ES-valencia': 'lakota', 'ccp': '𑄣𑄇𑄮𑄑', 'ce': 'лакота', 'chr': 'ᎳᎪᏓ', 'ckb': 'لاکۆتا', 'cs': 'lakotština', 'cy': 'Lakota', 'da': 'lakota', 'de': 'Lakota', 'dsb': 'lakotšćina', 'el': 'Λακότα', 'el-polyton': 'Λακότα', 'en': 'Lakota', 'es': 'lakota', 'et': 'lakota', 'eu': 'lakotera', 'fa': 'لاکوتا', 'ff-Adlm': '𞤂𞤢𞤳𞤮𞤼𞤢𞥄𞤪𞤫', 'fi': 'lakota', 'fil': 'Lakota', 'fo': 'lakota', 'fr': 'lakota', 'fy': 'Lakota', 'ga': 'Lacótais', 'gd': 'Lakhóta', 'gl': 'lakota', 'gu': 'લાકોટા', 'ha': 'Lakota', 'he': 'לקוטה', 'hi': 'लैकोटा', 'hi-Latn': 'Lakota', 'hr': 'lakota', 'hsb': 'lakota', 'ht': 'lakota', 'hu': 'lakota', 'hy': 'լակոտա', 'ia': 'lakota', 'id': 'Lakota', 'ig': 'Lakota', 'is': 'lakóta', 'it': 'lakota', 'ja': 'ラコタ語', 'jv': 'Lakota', 'ka': 'ლაკოტა', 'kea': 'lakota', 'kgp': 'rakóta', 'kk': 'лакота тілі', 'kk-Cyrl': 'лакота тілі', 'km': 'ឡាកូតា', 'kn': 'ಲಕೊಟ', 'ko': '라코타어', 'kok': 'लाकोटा', 'kok-Deva': 'लाकोटा', 'ksh': 'Lakota', 'ku': 'lakotayî', 'ky': 'лакотача', 'lb': 'Lakota-Sprooch', 'lkt': 'Lakȟólʼiyapi', 'lo': 'ລາໂກຕາ', 'lrc': 'لاکوٙتا', 'lt': 'lakotų', 'lv': 'lakotu', 'mai': 'लकोटा', 'mi': 'Rakōta', 'mk': 'лакотски', 'ml': 'ലകൗട്ട', 'mn': 'лакота', 'mr': 'लाकोटा', 'ms': 'Lakota', 'mt': 'Lakota', 'my': 'လာကိုတာ', 'mzn': 'لاکوتا', 'nb': 'lakota', 'ne': 'लाकोता', 'nl': 'Lakota', 'nn': 'lakota', 'no': 'lakota', 'nqo': 'ߟߊߞߏߕߊߞߊ߲', 'or': 'ଲାକୋଟା', 'pa': 'ਲਕੋਟਾ', 'pa-Guru': 'ਲਕੋਟਾ', 'pcm': 'Lakótá Lángwej', 'pl': 'lakota', 'ps': 'لکوټا', 'pt': 'lacota', 'qu': 'Lakota Simi', 'rm': 'lakota', 'ro': 'lakota', 'ru': 'лакота', 'sc': 'lakota', 'sd': 'لڪوٽا', 'sd-Arab': 'لڪوٽا', 'si': 'ලකොට', 'sk': 'lakotčina', 'sl': 'lakotščina', 'smn': 'lakotakielâ', 'so': 'Laakoota', 'sq': 'lakotisht', 'sr': 'лакота', 'sr-Cyrl': 'лакота', 'sr-Latn': 'lakota', 'sv': 'lakota', 'sw': 'Kilakota', 'ta': 'லகோடா', 'te': 'లకొటా', 'th': 'ลาโกตา', 'ti': 'ላኮታ', 'tk': 'lakota dili', 'to': 'lea fakalakota', 'tr': 'Lakotaca', 'uk': 'лакота', 'ur': 'لاکوٹا', 'uz': 'lakota', 'uz-Cyrl': 'лакотачалакотача', 'uz-Latn': 'lakota', 'vec': 'lacota', 'vi': 'Tiếng Lakota', 'xnr': 'लैकोटा', 'yo': 'Lákota', 'yrl': 'rakuta', 'yue': '拉科塔文', 'yue-Hans': '拉科塔文', 'yue-Hant': '拉科塔文', 'zh': '拉科塔语', 'zh-Hans': '拉科塔语', 'zh-Hant': '拉科塔文', 'zu': 'isi-Lakota'},
    'lku': {'en': 'Kungkari'},
    'lky': {'en': 'Lokoya'},
    'lla': {'en': 'Lala-Roba'},
    'llb': {'en': 'Lolo'},
    'llc': {'en': 'Lele (Guinea)'},
    'lld': {'bal-Latn': 'Ladin', 'et': 'ladiini', 'fi': 'ladin', 'gd': 'Ladainis', 'ia': 'ladino', 'ig': 'ID', 'it': 'ladino', 'lld': 'ladin', 'uk': 'ладинська', 'vec': 'ladin', 'en': 'Ladin'},
    'lle': {'en': 'Lele (Papua New Guinea)'},
    'llf': {'en': 'Hermit'},
    'llg': {'en': 'Lole'},
    'llh': {'en': 'Lamu'},
    'lli': {'en': 'Teke-Laali'},
    'llj': {'en': 'Ladji Ladji'},
    'llk': {'en': 'Lelak'},
    'lll': {'en': 'Lilau'},
    'llm': {'en': 'Lasalimu'},
    'lln': {'en': 'Lele (Chad)'},
    'llo': {'en': 'Khlor'},
    'llp': {'en': 'North Efate'},
    'llq': {'en': 'Lolak'},
    'lls': {'en': 'Lithuanian Sign Language'},
    'llu': {'en': 'Lau'},
    'llx': {'en': 'Lauan'},
    'lma': {'en': 'East Limba'},
    'lmb': {'en': 'Merei'},
    'lmc': {'en': 'Limilngan'},
    'lmd': {'en': 'Lumun'},
    'lme': {'en': 'Pévé'},
    'lmf': {'en': 'South Lembata'},
    'lmg': {'en': 'Lamogai'},
    'lmh': {'en': 'Lambichhong'},
    'lmi': {'en': 'Lombi'},
    'lmj': {'en': 'West Lembata'},
    'lmk': {'en': 'Lamkang'},
    'lml': {'en': 'Hano'},
    'lmm': {'en': 'Lamam'},
    'lmn': {'en': 'Lambadi'},
    'lmo': {'af': 'Lombardies', 'ak': 'Lombad kasa', 'am': 'ሎምባርድኛ', 'ar': 'اللومبردية', 'as': 'ল’ম্বাৰ্ড', 'ast': 'lombardu', 'az': 'lombard dili', 'az-Latn': 'lombard dili', 'bal-Latn': 'Lombard', 'be': 'ламбардская', 'be-tarask': 'ламбардская', 'bg': 'ломбардски', 'bn': 'লম্বার্ড', 'bs': 'lombardski', 'bs-Latn': 'lombardski', 'ca': 'llombard', 'ca-ES-valencia': 'llombard', 'chr': 'ᎶᏆᏗ', 'cs': 'lombardština', 'cy': 'Lombardeg', 'da': 'lombardisk', 'de': 'Lombardisch', 'dsb': 'lombardišćina', 'el': 'Λομβαρδικά', 'el-polyton': 'Λομβαρδικά', 'en': 'Lombard', 'es': 'lombardo', 'et': 'lombardi', 'eu': 'lombardiera', 'fa': 'لومبارد', 'ff-Adlm': '𞤂𞤮𞤥𞤦𞤢𞤪𞤣𞤫', 'fi': 'lombardi', 'fil': 'Lombard', 'fo': 'lombard', 'fr': 'lombard', 'ga': 'Lombairdis', 'gd': 'Lombardais', 'gl': 'lombardo', 'gu': 'લોંબાર્ડ', 'ha': 'Lombard', 'he': 'לומברדית', 'hi': 'लॉमबर्ड', 'hi-Latn': 'Lombard', 'hr': 'lombardski', 'hsb': 'lombardšćina', 'ht': 'lombard', 'hu': 'lombard', 'hy': 'լոմբարդերեն', 'ia': 'lombardo', 'id': 'Lombard', 'ig': 'Lombard', 'is': 'lombardíska', 'it': 'lombardo', 'ja': 'ロンバルド語', 'jv': 'Lombard', 'ka': 'ლომბარდიული', 'kk': 'ломбард тілі', 'kk-Cyrl': 'ломбард тілі', 'km': 'ឡំបាត', 'kn': 'ಲೋಂಬರ್ಡ್', 'ko': '롬바르드어', 'kok': 'लोमबार्ड', 'kok-Deva': 'लोमबार्ड', 'ku': 'lombardî', 'ky': 'ломбардча', 'lb': 'Lombardesch', 'lmo': 'Lombard', 'lo': 'ລອມບາດ', 'lt': 'lombardų', 'lv': 'lombardiešu', 'mk': 'ломбардиски', 'ml': 'ലൊംബാർഡ്', 'mn': 'ломбард', 'mr': 'लोंबार्ड', 'ms': 'Lombard', 'my': 'လန်းဘတ်', 'nb': 'lombardisk', 'ne': 'लोम्बार्ड', 'nl': 'Lombardisch', 'nn': 'lombardisk', 'no': 'lombardisk', 'or': 'ଲୋମ୍ବାର୍ଡ୍\u200c', 'pa': 'ਲੰਬਾਰਡ', 'pa-Guru': 'ਲੰਬਾਰਡ', 'pcm': 'Lombárd Lángwej', 'pl': 'lombardzki', 'ps': 'لومبارډ', 'pt': 'lombardo', 'ro': 'lombardă', 'ru': 'ломбардский', 'sc': 'lombardu', 'sd': 'لامبارڊ', 'sd-Arab': 'لامبارڊ', 'si': 'ලොම්බාර්ඩ්', 'sk': 'lombardčina', 'sl': 'lombardščina', 'so': 'Lombard', 'sq': 'lombardisht', 'sr': 'ломбард', 'sr-Cyrl': 'ломбард', 'sr-Latn': 'lombard', 'sv': 'lombardiska', 'sw': 'Kilongobardi', 'ta': 'லொம்பார்டு', 'te': 'లొంబార్ద్', 'th': 'ลอมบาร์ด', 'ti': 'ሎምባርድኛ', 'tk': 'lombard dili', 'to': 'lea fakalomipāti', 'tr': 'Lombardça', 'uk': 'ломбардська', 'ur': 'لومبارڈ', 'uz': 'lombard', 'uz-Latn': 'lombard', 'vec': 'lonbardo', 'vi': 'Tiếng Lombard', 'yo': 'Lombardi', 'yue': '倫巴底文', 'yue-Hans': '伦巴底文', 'yue-Hant': '倫巴底文', 'zh': '伦巴第语', 'zh-Hans': '伦巴第语', 'zh-Hant': '倫巴底文', 'zu': 'IsiLombard'},
    'lmp': {'en': 'Limbum'},
    'lmq': {'en': 'Lamatuka'},
    'lmr': {'en': 'Lamalera'},
    'lmu': {'en': 'Lamenu'},
    'lmv': {'en': 'Lomaiviti'},
    'lmw': {'en': 'Lake Miwok'},
    'lmx': {'en': 'Laimbue'},
    'lmy': {'en': 'Lamboya'},
    'lmz': {'en': 'Lumbee'},
    'ln': {'af': 'Lingaals', 'am': 'ሊንጋላ', 'ar': 'اللينجالا', 'as': 'লিংগালা', 'ast': 'lingala', 'az': 'linqala', 'az-Cyrl': 'лингала', 'az-Latn': 'linqala', 'bal-Latn': 'Lingálá', 'be': 'лінгала', 'be-tarask': 'лінгала', 'bg': 'лингала', 'bgn': 'لینگالایی', 'bn': 'লিঙ্গালা', 'br': 'lingala', 'brx': 'लिंगाला', 'bs': 'lingala', 'bs-Cyrl': 'лингала', 'bs-Latn': 'lingala', 'ca': 'lingala', 'ca-ES-valencia': 'lingala', 'ccp': '𑄣𑄨𑄋𑄴𑄉𑄣', 'ce': 'лингала', 'chr': 'ᎵᏂᎦᎳ', 'ckb': 'لينگالا', 'cs': 'lingalština', 'cy': 'Lingala', 'da': 'lingala', 'de': 'Lingala', 'dsb': 'lingala', 'ee': 'lingala', 'el': 'Λινγκάλα', 'el-polyton': 'Λινγκάλα', 'en': 'Lingala', 'eo': 'lingala', 'es': 'lingala', 'et': 'lingala', 'eu': 'lingala', 'fa': 'لینگالا', 'ff-Adlm': '𞤂𞤭𞤲𞤺𞤢𞤤𞤢𞥄𞤪𞤫', 'fi': 'lingala', 'fil': 'Lingala', 'fo': 'lingala', 'fr': 'lingala', 'fur': 'lingala', 'fy': 'Lingala', 'ga': 'Liongáilis', 'gd': 'Lingala', 'gl': 'lingala', 'gsw': 'Lingala', 'gu': 'લિંગાલા', 'ha': 'Lingala', 'he': 'לינגלה', 'hi': 'लिंगाला', 'hi-Latn': 'Lingala', 'hr': 'lingala', 'hsb': 'lingala', 'ht': 'lingala', 'hu': 'lingala', 'hy': 'լինգալա', 'ia': 'lingala', 'id': 'Lingala', 'ig': 'Lingala', 'is': 'lingala', 'it': 'lingala', 'ja': 'リンガラ語', 'jv': 'Lingala', 'ka': 'ლინგალა', 'kea': 'lingala', 'kgp': 'rĩgara', 'kk': 'лингала тілі', 'kk-Cyrl': 'лингала тілі', 'km': 'លីនកាឡា', 'kn': 'ಲಿಂಗಾಲ', 'ko': '링갈라어', 'kok': 'लिंगाला', 'kok-Deva': 'लिंगाला', 'ks': 'لِنگالا', 'ks-Arab': 'لِنگالا', 'ksh': 'Lingjalla', 'ku': 'lingalayî', 'ky': 'лингалача', 'lb': 'Lingala', 'lld': 'lingala', 'ln': 'lingála', 'lo': 'ລິງກາລາ', 'lrc': 'لینگالا', 'lt': 'ngalų', 'lv': 'lingala', 'mai': 'लिंगाला', 'mi': 'Ringāra', 'mk': 'лингала', 'ml': 'ലിംഗാല', 'mn': 'лингала', 'mr': 'लिंगाला', 'ms': 'Lingala', 'mt': 'Lingaljan', 'my': 'လင်ဂါလာ', 'mzn': 'لینگالا', 'nb': 'lingala', 'ne': 'लिङ्गाला', 'nl': 'Lingala', 'nn': 'lingala', 'no': 'lingala', 'nqo': 'ߟߌ߲ߜ߭ߟߊߞߊ߲', 'or': 'ଲିଙ୍ଗାଲା', 'pa': 'ਲਿੰਗਾਲਾ', 'pa-Guru': 'ਲਿੰਗਾਲਾ', 'pcm': 'Lingálá Lángwej', 'pl': 'lingala', 'ps': 'لنګالا', 'pt': 'lingala', 'qu': 'Lingala Simi', 'rm': 'lingala', 'ro': 'lingala', 'ru': 'лингала', 'rw': 'Ilingala', 'sc': 'lingala', 'sd': 'لنگالا', 'sd-Arab': 'لنگالا', 'si': 'ලින්ගලා', 'sk': 'lingalčina', 'sl': 'lingala', 'smn': 'lingala', 'so': 'Lingala', 'sq': 'lingalisht', 'sr': 'лингала', 'sr-Cyrl': 'лингала', 'sr-Latn': 'lingala', 'sv': 'lingala', 'sw': 'Kilingala', 'ta': 'லிங்காலா', 'te': 'లింగాల', 'th': 'ลิงกาลา', 'ti': 'ሊንጋላ', 'tk': 'lingala dili', 'to': 'lea lingikala', 'tr': 'Lingala', 'ug': 'لىنگالاچە', 'uk': 'лінгала', 'ur': 'لِنگَلا', 'uz': 'lingala', 'uz-Cyrl': 'лингалча', 'uz-Latn': 'lingala', 'vec': 'lingala', 'vi': 'Tiếng Lingala', 'wae': 'Lingala', 'xnr': 'लिंगाला', 'yo': 'Lìǹgálà', 'yrl': 'rĩgana', 'yue': '林加拉文', 'yue-Hans': '林加拉文', 'yue-Hant': '林加拉文', 'zh': '林加拉语', 'zh-Hans': '林加拉语', 'zh-Hant': '林加拉文', 'zu': 'isi-Lingala'},
    'lna': {'en': 'Langbashe'},
    'lnb': {'en': 'Mbalanhu'},
    'lnd': {'en': 'Lundayeh'},
    'lng': {'en': 'Langobardic'},
    'lnh': {'en': 'Lanoh'},
    'lni': {'en': "Daantanai'"},
    'lnj': {'en': 'Leningitij'},
    'lnl': {'en': 'South Central Banda'},
    'lnm': {'en': 'Langam'},
    'lnn': {'en': 'Lorediakarkar'},
    'lno': {'en': 'Lango (South Sudan)'},
    'lns': {'en': "Lamnso'"},
    'lnu': {'en': 'Longuda'},
    'lnw': {'en': 'Lanima'},
    'lnz': {'en': 'Lonzo'},
    'lo': {'af': 'Lao', 'ak': 'Lawo kasa', 'am': 'ላኦኛ', 'ar': 'اللاوية', 'as': 'লাও', 'ast': 'laosianu', 'az': 'laos', 'az-Cyrl': 'лаос', 'az-Latn': 'laos', 'bal-Latn': 'Láó', 'be': 'лаоская', 'be-tarask': 'лаоская', 'bg': 'лаоски', 'bgn': 'لائوسی', 'bn': 'লাও', 'br': 'laoseg', 'brx': 'लाऊ', 'bs': 'laoski', 'bs-Cyrl': 'лаоски', 'bs-Latn': 'laoski', 'ca': 'laosià', 'ca-ES-valencia': 'laosià', 'ccp': '𑄣𑄃𑄮', 'ce': 'лаоссийн', 'chr': 'ᎳᎣ', 'ckb': 'لائۆیی', 'cs': 'laoština', 'cy': 'Laoeg', 'da': 'lao', 'de': 'Laotisch', 'dsb': 'laošćina', 'dz': 'ལཱ་ཝོས་ཁ', 'ee': 'laogbe', 'el': 'Λαοτινά', 'el-polyton': 'Λαοτινά', 'en': 'Lao', 'en-Dsrt': '𐐢𐐵', 'eo': 'laŭa', 'es': 'lao', 'et': 'lao', 'eu': 'laosera', 'fa': 'لائوسی', 'ff-Adlm': '𞤂𞤢𞤮𞥅𞤪𞤫', 'fi': 'lao', 'fil': 'Lao', 'fo': 'laoskt', 'fr': 'lao', 'fur': 'lao', 'fy': 'Laotiaansk', 'ga': 'Laoisis', 'gd': 'Làtho', 'gl': 'laosiano', 'gsw': 'Laozisch', 'gu': 'લાઓ', 'ha': 'Lao', 'he': 'לאו', 'hi': 'लाओ', 'hi-Latn': 'Lao', 'hr': 'laoski', 'hsb': 'laošćina', 'ht': 'lao', 'hu': 'lao', 'hy': 'լաոսերեն', 'ia': 'laotiano', 'id': 'Lao', 'ig': 'Lao', 'is': 'laó', 'it': 'lao', 'ja': 'ラオ語', 'jv': 'Laos', 'ka': 'ლაოსური', 'kea': 'lausianu', 'kgp': 'raosijỹnũ', 'kk': 'лаос тілі', 'kk-Cyrl': 'лаос тілі', 'km': 'ឡាវ', 'kn': 'ಲಾವೋ', 'ko': '라오어', 'kok': 'लाओ', 'kok-Deva': 'लाओ', 'ks': 'لاو', 'ks-Arab': 'لاو', 'ksh': 'Lahootesch', 'ku': 'lawsî', 'kxv': 'lao', 'kxv-Deva': 'लाओ', 'kxv-Latn': 'lao', 'kxv-Orya': 'ଲାଅ', 'kxv-Telu': 'లావో', 'ky': 'лаочо', 'lb': 'Laotesch', 'lkt': 'Lao Iyápi', 'lld': 'lao', 'lo': 'ລາວ', 'lrc': 'لاو', 'lt': 'laosiečių', 'lv': 'laosiešu', 'mai': 'लाओ', 'mi': 'Rao', 'mk': 'лаошки', 'ml': 'ലാവോ', 'mn': 'лаос', 'mr': 'लाओ', 'ms': 'Laos', 'mt': 'Laosjan', 'my': 'လာအို', 'mzn': 'لائویی', 'nb': 'laotisk', 'ne': 'लाओ', 'nl': 'Laotiaans', 'nn': 'laotisk', 'no': 'laotisk', 'nqo': 'ߟߊߏߞߊ߲', 'or': 'ଲାଓ', 'pa': 'ਲਾਓ', 'pa-Guru': 'ਲਾਓ', 'pcm': 'Láo Lángwej', 'pl': 'laotański', 'ps': 'لاو', 'pt': 'laosiano', 'qu': 'Lao Simi', 'rm': 'laot', 'ro': 'laoțiană', 'ru': 'лаосский', 'rw': 'Ikilawotiyani', 'sc': 'laotianu', 'sd': 'لائو', 'sd-Arab': 'لائو', 'se': 'laogiella', 'si': 'ලාඕ', 'sk': 'laoština', 'sl': 'laoščina', 'smn': 'laokielâ', 'so': 'Lao', 'sq': 'laosisht', 'sr': 'лаоски', 'sr-Cyrl': 'лаоски', 'sr-Latn': 'laoski', 'sv': 'laotiska', 'sw': 'Kilaosi', 'ta': 'லாவோ', 'te': 'లావో', 'tg': 'лаосӣ', 'th': 'ลาว', 'ti': 'ላኦ', 'tk': 'laos dili', 'to': 'lea fakalau', 'tr': 'Lao dili', 'tt': 'лаос', 'ug': 'لائوسچە', 'uk': 'лаоська', 'ur': 'لاؤ', 'uz': 'laos', 'uz-Cyrl': 'лаосча', 'uz-Latn': 'laos', 'vec': 'lao', 'vi': 'Tiếng Lào', 'wae': 'Laotiš', 'wo': 'Laaw', 'xnr': 'लाओ', 'yi': 'לאַא', 'yo': 'Láò', 'yrl': 'rausianu', 'yue': '寮文', 'yue-Hans': '寮文', 'yue-Hant': '寮文', 'zh': '老挝语', 'zh-Hans': '老挝语', 'zh-Hant': '寮文', 'zu': 'isi-Lao'},
    'loa': {'en': 'Loloda'},
    'lob': {'en': 'Lobi'},
    'loc': {'en': 'Inonhan'},
    'lod': {'en': 'Berawan'},
    'loe': {'en': 'Saluan'},
    'lof': {'en': 'Logol'},
    'log': {'en': 'Logo'},
    'loh': {'en': 'Narim'},
    'loi': {'en': "Loma (Côte d'Ivoire)"},
    'loj': {'en': 'Lou'},
    'lok': {'en': 'Loko'},
    'lol': {'ar': 'منغولى', 'ast': 'mongo', 'az': 'monqo', 'az-Latn': 'monqo', 'be': 'монга', 'be-tarask': 'монга', 'bg': 'монго', 'bn': 'মোঙ্গো', 'br': 'mongo', 'brx': 'मोंगो', 'bs': 'mongo', 'bs-Cyrl': 'монго', 'bs-Latn': 'mongo', 'ca': 'mongo', 'ca-ES-valencia': 'mongo', 'ccp': '𑄟𑄮𑄋𑄴𑄉𑄮', 'cs': 'mongština', 'cy': 'Mongo', 'da': 'mongo', 'de': 'Mongo', 'el': 'Μόνγκο', 'el-polyton': 'Μόνγκο', 'en': 'Mongo', 'es': 'mongo', 'et': 'mongo', 'fa': 'مونگویی', 'fi': 'mongo', 'fr': 'mongo', 'fy': 'Mongo', 'gd': 'Mongo', 'gsw': 'Mongo', 'gu': 'મોંગો', 'he': 'מונגו', 'hi': 'मोंगो', 'hi-Latn': 'Mongo', 'hr': 'mongo', 'ht': 'mongo', 'hu': 'mongó', 'id': 'Mongo', 'is': 'mongó', 'it': 'lolo bantu', 'ja': 'モンゴ語', 'ka': 'მონგო', 'kgp': 'mãgo', 'kn': 'ಮೊಂಗೋ', 'ko': '몽고어', 'ks': 'مونگو', 'ks-Arab': 'مونگو', 'lb': 'Mongo', 'lo': 'ແມັງໂກ້', 'lt': 'mongų', 'lv': 'mongu', 'mk': 'монго', 'ml': 'മോങ്കോ', 'mr': 'मोंगो', 'mt': 'Mongo', 'nb': 'mongo', 'ne': 'मोङ्गो', 'nl': 'Mongo', 'nn': 'mongo', 'no': 'mongo', 'or': 'ମଙ୍ଗୋ', 'pl': 'mongo', 'pt': 'mongo', 'rm': 'lomongo', 'ro': 'mongo', 'ru': 'монго', 'sk': 'mongo', 'sl': 'mongo', 'sr': 'монго', 'sr-Cyrl': 'монго', 'sr-Latn': 'mongo', 'sv': 'mongo', 'sw': 'Kimongo', 'ta': 'மோங்கோ', 'te': 'మొంగో', 'th': 'มองโก', 'to': 'lea fakamongikō', 'tr': 'Mongo', 'ug': 'مونگوچە', 'uk': 'монго', 'vi': 'Tiếng Mongo', 'yrl': 'mũgu', 'yue': '芒戈文', 'yue-Hans': '芒戈文', 'yue-Hant': '芒戈文', 'zh': '蒙戈语', 'zh-Hans': '蒙戈语', 'zh-Hant': '芒戈文'},
    'lom': {'en': 'Loma (Liberia)'},
    'lon': {'en': 'Malawi Lomwe'},
    'loo': {'en': 'Lombo'},
    'lop': {'en': 'Lopa'},
    'loq': {'en': 'Lobala'},
    'lor': {'en': 'Téén'},
    'los': {'en': 'Loniu'},
    'lot': {'en': 'Otuho'},
    'lou': {'af': 'Louisiana Kreool', 'am': 'ሉዊዚያና ክሬኦል', 'ar': 'الكريولية اللويزيانية', 'as': 'লুইজিয়ানা কেৰ’ল', 'az': 'Luiziana kreolu', 'az-Latn': 'Luiziana kreolu', 'bal-Latn': 'Luisiáná Krióle', 'be': 'луізіянская крэольская', 'be-tarask': 'луізіянская крэольская', 'bg': 'луизиански креолски', 'bn': 'লুইসিয়ানা ক্রেওল', 'br': 'kreoleg Louiziana', 'brx': 'लुईसियाना क्रेवल', 'bs': 'luizijanski kreolski', 'bs-Cyrl': 'луизиански креолски', 'bs-Latn': 'luizijanski kreolski', 'ca': 'crioll francès de Louisiana', 'ca-ES-valencia': 'crioll francès de Louisiana', 'chr': 'ᎷᏫᏏᎡᎾ ᎦᏬᏂᎯᏍᏗ', 'cs': 'kreolština (Louisiana)', 'cy': 'Louisiana Creole', 'da': 'Louisiana-kreolsk', 'de': 'Kreol (Louisiana)', 'dsb': 'Louisiana kreolšćina', 'el': 'Κρεολικά (Λουιζιάνα)', 'el-polyton': 'Κρεολικά (Λουιζιάνα)', 'en': 'Louisiana Creole', 'es': 'criollo de Luisiana', 'et': 'Louisiana kreoolkeel', 'eu': 'Louisianako kreolera', 'fa': 'کرئول لوئیزیانا', 'ff-Adlm': '𞤀𞤳𞤵𞥅𞤪𞤫 𞤂𞤵𞥅𞥁𞤭𞤴𞤢𞥄𞤲𞤢', 'fi': 'louisianankreoli', 'fil': 'Louisiana Creole', 'fr': 'créole louisianais', 'ga': 'Criól Louisiana', 'gd': 'Crìtheol Louisiana', 'gl': 'crioulo de Luisiana', 'gu': 'લ્યુઇસિયાના ક્રેઓલ', 'ha': 'Creole na Louisiana', 'he': 'קריאולית לואיזיאנית', 'hi': 'लुईज़ियाना क्रियोल', 'hi-Latn': 'Louisiana Creole', 'hr': 'lujzijanski kreolski', 'hsb': 'Louisiana kreolšćina', 'ht': 'créole louisianais', 'hu': 'louisianai kreol', 'hy': 'լուիզիանական կրեոլերեն', 'ia': 'creolo louisianese', 'id': 'Kreol Louisiana', 'ig': 'Louisiana Creole', 'is': 'kreólska (Louisiana)', 'it': 'creolo della Louisiana', 'ja': 'ルイジアナ・クレオール語', 'jv': 'Louisiana Creole', 'ka': 'ლუიზიანას კრეოლური', 'kgp': 'rovusijỹnỹ tá ke pẽ', 'kk': 'креоль тілі (Луизиана)', 'kk-Cyrl': 'креоль тілі (Луизиана)', 'km': 'ក្រេអូល លូអ៊ីស៊ីអាណា', 'kn': 'ಲೂಯಿಸಿಯಾನ ಕ್ರಿಯೋಲ್', 'ko': '루이지애나 크리올어', 'kok': 'ल्युइसियाना क्रियोल', 'kok-Deva': 'ल्युइसियाना क्रियोल', 'ku': 'kreyolîya louisianayê', 'ky': 'луизиана креолчо', 'lo': 'ລຸຍຊີອານນາ ຄຣີໂອນ', 'lt': 'Luizianos kreolų', 'lv': 'Luiziānas kreolu', 'mai': 'लुइसियाना क्रियोल', 'mi': 'Kreōro Ruihiana', 'mk': 'луизијански креолски', 'ml': 'ലൂസിയാന ക്രിയോൾ', 'mn': 'луизиана креоле', 'mr': 'ल्युसियाना क्रिओल', 'ms': 'Kreol Louisiana', 'my': 'လူဝီဇီယားနား ခရီးယို', 'nb': 'louisianakreolsk', 'ne': 'लुसियाना क्रियोल', 'nl': 'Louisiana-Creools', 'nn': 'louisianakreolsk', 'no': 'louisianakreolsk', 'nqo': 'ߟߌߖ߭ߌߦߊߣߌ߫ ߕߊ߬ߓߎ߰ߛߌ߬ߞߊ߲', 'or': 'ଲୌସିଆନା କ୍ରେଓଲେ', 'pa': 'ਲੇਉ', 'pa-Guru': 'ਲੇਉ', 'pcm': 'Kriol fọ Luisiána', 'pl': 'kreolski luizjański', 'ps': 'Louis', 'pt': 'crioulo da Louisiana', 'qu': 'Luisiana Criollo', 'ro': 'creolă (Louisiana)', 'ru': 'луизианский креольский', 'sc': 'crèolu de sa Louisiana', 'sd': 'لوئيزيانا ڪريئول', 'sd-Arab': 'لوئيزيانا ڪريئول', 'si': 'ලුසියානා ක්\u200dරියෝල්', 'sk': 'kreolčina (Louisiana)', 'sl': 'louisianska kreolščina', 'so': 'Louisiana Creole', 'sq': 'kreolishte e Luizianës', 'sr': 'луизијански креолски', 'sr-Cyrl': 'луизијански креолски', 'sr-Latn': 'luizijanski kreolski', 'sv': 'louisiana-kreol', 'sw': 'Kikrioli cha Louisiana', 'ta': 'லூசியானா க்ரயோல்', 'te': 'లూసియానా క్రియోల్', 'th': 'ครีโอลุยเซียนา', 'ti': 'ክርዮል ሉዊዝያና', 'tk': 'Luiziana kreol dili', 'to': 'lea fakaluisiana', 'tr': 'Louisiana Kreolcesi', 'uk': 'луїзіанська креольська', 'ur': 'لوزیانا کریول', 'uz': 'luiziana kreol', 'uz-Latn': 'luiziana kreol', 'vec': 'crèolo de la Luiziana', 'vi': 'Tiếng Creole Louisiana', 'xnr': 'लुईज़ियाना क्रियोल', 'yo': 'Èdè Kreoli ti Louisiana', 'yrl': 'kiriuru ruisianawara', 'yue': '路易斯安那克里奧爾文', 'yue-Hans': '路易斯安那克里奥尔文', 'yue-Hant': '路易斯安那克里奧爾文', 'zh': '路易斯安那克里奥尔语', 'zh-Hans': '路易斯安那克里奥尔语', 'zh-Hant': '路易斯安那克里奧爾文', 'zu': 'isi-Louisiana Creole'},
    'lov': {'en': 'Lopi'},
    'low': {'en': 'Tampias Lobu'},
    'lox': {'en': 'Loun'},
    'loy': {'en': 'Loke'},
    'loz': {'af': 'Lozi', 'am': 'ሎዚ', 'ar': 'اللوزي', 'as': 'লোজি', 'ast': 'lozi', 'az': 'lozi', 'az-Cyrl': 'лози', 'az-Latn': 'lozi', 'be': 'лозі', 'be-tarask': 'лозі', 'bg': 'лози', 'bgn': 'لوزی', 'bn': 'লোজি', 'br': 'lozi', 'brx': 'ल’जी', 'bs': 'lozi', 'bs-Cyrl': 'лози', 'bs-Latn': 'lozi', 'ca': 'lozi', 'ca-ES-valencia': 'lozi', 'ccp': '𑄣𑄮𑄎𑄨', 'ce': 'лози', 'chr': 'ᎶᏏ', 'ckb': 'لۆزی', 'cs': 'lozština', 'cy': 'Lozi', 'da': 'lozi', 'de': 'Lozi', 'dsb': 'lozišćina', 'el': 'Λόζι', 'el-polyton': 'Λόζι', 'en': 'Lozi', 'es': 'lozi', 'et': 'lozi', 'eu': 'loziera', 'fa': 'لوزیایی', 'ff-Adlm': '𞤂𞤮𞥁𞤭𞥅𞤪𞤫', 'fi': 'lozi', 'fil': 'Lozi', 'fo': 'lozi', 'fr': 'lozi', 'fy': 'Lozi', 'ga': 'Lóisis', 'gd': 'Lozi', 'gl': 'lozi', 'gsw': 'Rotse-Schpraach', 'gu': 'લોઝી', 'ha': 'Lozi', 'he': 'לוזית', 'hi': 'लोज़ी', 'hi-Latn': 'Lozi', 'hr': 'lozi', 'hsb': 'lozišćina', 'ht': 'lozi', 'hu': 'lozi', 'hy': 'լոզի', 'ia': 'lozi', 'id': 'Lozi', 'ig': 'Lozi', 'is': 'lozi', 'it': 'lozi', 'ja': 'ロジ語', 'jv': 'Losi', 'ka': 'ლოზი', 'kgp': 'roji', 'kk': 'лози тілі', 'kk-Cyrl': 'лози тілі', 'km': 'ឡូហ្ស៊ី', 'kn': 'ಲೋಝಿ', 'ko': '로지어', 'kok': 'लोझीं', 'kok-Deva': 'लोझीं', 'ks': 'لوزی', 'ks-Arab': 'لوزی', 'ksh': 'Silohzi', 'ku': 'lozî', 'ky': 'лозиче', 'lb': 'Rotse-Sprooch', 'lo': 'ໂລຊິ', 'lt': 'lozių', 'lv': 'lozu', 'mai': 'लोजी', 'mi': 'Rohi', 'mk': 'лози', 'ml': 'ലൊസി', 'mn': 'лози', 'mr': 'लोझि', 'ms': 'Lozi', 'mt': 'Lożi', 'my': 'လိုဇီ', 'nb': 'lozi', 'ne': 'लोजी', 'nl': 'Lozi', 'nn': 'lozi', 'no': 'lozi', 'or': 'ଲୋଜି', 'pa': 'ਲੋਜ਼ੀ', 'pa-Guru': 'ਲੋਜ਼ੀ', 'pcm': 'Lózí Lángwej', 'pl': 'lozi', 'ps': 'لوزی', 'pt': 'lozi', 'qu': 'Lozi Simi', 'rm': 'lozi', 'ro': 'lozi', 'ru': 'лози', 'sc': 'lozi', 'sd': 'لوزي', 'sd-Arab': 'لوزي', 'si': 'ලොසි', 'sk': 'lozi', 'sl': 'lozi', 'smn': 'lozi', 'so': 'Lozi', 'sq': 'lozisht', 'sr': 'лози', 'sr-Cyrl': 'лози', 'sr-Latn': 'lozi', 'sv': 'lozi', 'sw': 'Kilozi', 'ta': 'லோசி', 'te': 'లోజి', 'th': 'โลซิ', 'ti': 'ሎዚ', 'tk': 'lozi dili', 'to': 'lea fakalosi', 'tr': 'Lozi', 'ug': 'لوزىچە', 'uk': 'лозі', 'ur': 'لوزی', 'uz': 'lozi', 'uz-Latn': 'lozi', 'vec': 'lozi', 'vi': 'Tiếng Lozi', 'xnr': 'लोज़ी', 'yo': 'Èdè Lozi', 'yrl': 'rusi', 'yue': '洛齊文', 'yue-Hans': '洛齐文', 'yue-Hant': '洛齊文', 'zh': '洛齐语', 'zh-Hans': '洛齐语', 'zh-Hant': '洛齊文', 'zu': 'isi-Lozi'},
    'lpa': {'en': 'Lelepa'},
    'lpe': {'en': 'Lepki'},
    'lpn': {'en': 'Long Phuri Naga'},
    'lpo': {'en': 'Lipo'},
    'lpx': {'en': 'Lopit'},
    'lra': {'en': "Rara Bakati'"},
    'lrc': {'af': 'Noord-Luri', 'am': 'ሰሜናዊ ሉሪ', 'ar': 'اللرية الشمالية', 'as': 'উদীচ্য লুৰি', 'ast': 'luri del norte', 'az': 'şimali luri', 'az-Arab': 'قۇزئی لوْرجه', 'az-Cyrl': 'шимали лури', 'az-Latn': 'şimali luri', 'bal-Latn': 'Shemáli Lori', 'be': 'паўночная луры', 'be-tarask': 'паўночная луры', 'bg': 'северен лури', 'bgn': 'بُرزسرین لوری', 'bn': 'উত্তরাঞ্চলীয় লুরি', 'br': 'loureg an Norzh', 'brx': 'साहायारि लुरी', 'bs': 'sjeverni luri', 'bs-Cyrl': 'сјеверни лури', 'bs-Latn': 'sjeverni luri', 'ca': 'luri septentrional', 'ca-ES-valencia': 'luri septentrional', 'ccp': '𑄅𑄪𑄖𑄴𑄖𑄮𑄢𑄴 𑄣𑄪𑄢𑄨', 'ce': 'къилбаседа лури', 'chr': 'ᏧᏴᏢ ᏗᏜ ᎷᎵ', 'ckb': 'لوڕیی باکوور', 'cs': 'lúrština (severní)', 'cy': 'Luri Gogleddol', 'da': 'nordluri', 'de': 'Nördliches Luri', 'dsb': 'pódpołnocna lurišćina', 'el': 'Βόρεια Λούρι', 'el-polyton': 'Βόρεια Λούρι', 'en': 'Northern Luri', 'es': 'lorí septentrional', 'et': 'põhjaluri', 'eu': 'iparraldeko lurera', 'fa': 'لری شمالی', 'ff-Adlm': '𞤂𞤵𞤪𞤭𞥅𞤪𞤫 𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤫', 'fi': 'pohjoisluri', 'fil': 'Hilagang Luri', 'fo': 'norður luri', 'fr': 'lori du Nord', 'ga': 'Lúiris an Tuaiscirt', 'gd': 'Luri Thuathach', 'gl': 'luri setentrional', 'gu': 'ઉત્તરી લુરી', 'ha': 'Arewacin Luri', 'he': 'לורית צפונית', 'hi': 'उत्तरी लूरी', 'hi-Latn': 'Northern Luri', 'hr': 'sjevernolurski', 'hsb': 'sewjerna lurišćina', 'ht': 'lori du Nord', 'hu': 'északi luri', 'hy': 'հյուսիսային լուրիերեն', 'ia': 'luri del nord', 'id': 'Luri Utara', 'ig': 'Northern Luri', 'is': 'norðurlúrí', 'it': 'luri settentrionale', 'ja': '北ロル語', 'jv': 'Luri Sisih Lor', 'ka': 'ჩრდილოეთ ლური', 'kgp': 'ruri nãrti', 'kk': 'солтүстік люри тілі', 'kk-Cyrl': 'солтүстік люри тілі', 'km': 'លូរីខាងជើង', 'kn': 'ಉತ್ತರ ಲೂರಿ', 'ko': '북부 루리어', 'kok': 'उत्तरीय लुरी', 'kok-Deva': 'उत्तरीय लुरी', 'ksh': 'Nood-Luhresch', 'ku': 'lurîya bakur', 'ky': 'түндүк луриче', 'lo': 'ລູຣິ ທາງຕອນເໜືອ', 'lrc': 'لۊری شومالی', 'lt': 'šiaurės luri', 'lv': 'ziemeļluru', 'mai': 'उत्तरी लूरी', 'mi': 'Ruri ki te Raki', 'mk': 'севернолуриски', 'ml': 'വടക്കൻ ലൂറി', 'mn': 'хойд лури', 'mr': 'उत्तरी ल्युरी', 'ms': 'Luri Utara', 'mt': 'Luri tat-Tramuntana', 'my': 'မြောက်လူရီ', 'mzn': 'شمالی لُری', 'nb': 'nord-luri', 'ne': 'उत्तरी लुरी', 'nl': 'Noordelijk Luri', 'nn': 'nord-lurisk', 'no': 'nord-luri', 'nqo': 'ߕߟߋ߬ߓߐ ߟߏߙߌߞߊ߲', 'or': 'ଉତ୍ତର ଲୁରି', 'pa': 'ਉੱਤਰੀ ਲੁਰੀ', 'pa-Guru': 'ਉੱਤਰੀ ਲੁਰੀ', 'pcm': 'Nọ́tán Lúrí Lángwej', 'pl': 'luryjski północny', 'ps': 'شمالي لوری', 'pt': 'luri setentrional', 'qu': 'Luri septentrional Simi', 'rm': 'luri dal nord', 'ro': 'luri de nord', 'ru': 'севернолурский', 'sc': 'luri setentrionale', 'sd': 'اتر لوري', 'sd-Arab': 'اتر لوري', 'si': 'උතුරු ලුරි', 'sk': 'severné luri', 'sl': 'severna lurijščina', 'smn': 'taveluri', 'so': 'Luri Waqooyi', 'sq': 'lurishte veriore', 'sr': 'северни лури', 'sr-Cyrl': 'северни лури', 'sr-Latn': 'severni luri', 'sv': 'nordluri', 'sw': 'Kiluri cha Kaskazini', 'ta': 'வடக்கு லுரி', 'te': 'ఉత్తర లూరీ', 'th': 'ลูรีเหนือ', 'ti': 'ሰሜናዊ ሉሪ', 'tk': 'demirgazyk luri dili', 'to': 'lea fakaluli-tokelau', 'tr': 'Kuzey Luri', 'uk': 'північнолурська', 'ur': 'شمالی لری', 'uz': 'shimoliy luri', 'uz-Cyrl': 'шимолий лури', 'uz-Latn': 'shimoliy luri', 'vec': 'luri setentrionale', 'vi': 'Tiếng Bắc Luri', 'xnr': 'उत्तरी लूरी', 'yo': 'Apáàríwá Lúrì', 'yrl': 'ruri setẽtiriunau', 'yue': '北盧爾文', 'yue-Hans': '北卢尔文', 'yue-Hant': '北盧爾文', 'zh': '北卢尔语', 'zh-Hans': '北卢尔语', 'zh-Hant': '北盧爾文', 'zu': 'isi-Northern Luri'},
    'lre': {'en': 'Laurentian'},
    'lrg': {'en': 'Laragia'},
    'lri': {'en': 'Marachi'},
    'lrk': {'en': 'Loarki'},
    'lrl': {'en': 'Lari'},
    'lrm': {'en': 'Marama'},
    'lrn': {'en': 'Lorang'},
    'lro': {'en': 'Laro'},
    'lrr': {'en': 'Southern Yamphu'},
    'lrt': {'en': 'Larantuka Malay'},
    'lrv': {'en': 'Larevat'},
    'lrz': {'en': 'Lemerig'},
    'lsa': {'en': 'Lasgerdi'},
    'lsb': {'en': 'Burundian Sign Language'},
    'lsd': {'en': 'Lishana Deni'},
    'lse': {'en': 'Lusengo'},
    'lsg': {'en': 'Lyons Sign Language'},
    'lsh': {'en': 'Lish'},
    'lsi': {'en': 'Lashi'},
    'lsl': {'en': 'Latvian Sign Language'},
    'lsm': {'af': 'Saamia', 'am': 'ሳሚያ', 'ar': 'الساميا', 'as': 'চামিয়া', 'az': 'saamia', 'az-Latn': 'saamia', 'be': 'саамія', 'be-tarask': 'саамія', 'bg': 'саамски', 'bn': 'সামিয়া', 'brx': 'सामीया', 'bs': 'samia', 'bs-Cyrl': 'самиа', 'bs-Latn': 'samia', 'ca': 'saamia', 'ca-ES-valencia': 'saamia', 'chr': 'ᏌᎠᎻᎠ', 'cs': 'samia', 'cy': 'Saamia', 'da': 'saamia', 'de': 'Saamia', 'dsb': 'saamiašćina', 'el': 'Σαάμια', 'el-polyton': 'Σαάμια', 'en': 'Saamia', 'es': 'samia', 'et': 'samia', 'eu': 'saamia', 'fa': 'سامیا', 'ff-Adlm': '𞤅𞤢𞥄𞤥𞤭𞤢𞥄𞤪𞤫', 'fi': 'samia', 'fil': 'Saamia', 'fr': 'samia', 'ga': 'Saamia', 'gd': 'Saamia', 'gl': 'saamia', 'gu': 'સામિયા', 'ha': 'Saamiyanci', 'he': 'סמיה', 'hi': 'सामिया', 'hi-Latn': 'Saamia', 'hr': 'saamia', 'hsb': 'saamiašćina', 'ht': 'samia', 'hu': 'samia', 'hy': 'սաամերեն', 'ia': 'samia', 'id': 'Saamia', 'ig': 'Saamia', 'is': 'saamia', 'it': 'samia', 'ja': 'サーミア語', 'jv': 'Saamia', 'ka': 'სამია', 'kk': 'самия тілі', 'kk-Cyrl': 'самия тілі', 'km': 'សាមៀ', 'kn': 'ಸಾಮಿಯಾ', 'ko': '사미아어', 'kok': 'सामिया', 'kok-Deva': 'सामिया', 'ku': 'saamia', 'ky': 'саамиача (Уганда, Кения)', 'lo': 'ຊາອາເມຍ', 'lt': 'samių', 'lv': 'sāmia', 'mai': 'सामिया', 'mi': 'Hāmia', 'mk': 'самиски', 'ml': 'സാമിയ', 'mn': 'самиа', 'mr': 'सामिया', 'ms': 'Saamia', 'my': 'ဆာမိအာ', 'nb': 'samia', 'ne': 'सामिया', 'nl': 'Saamia', 'nn': 'samia', 'no': 'samia', 'or': 'ସାମିଆ', 'pa': 'ਸਾਮੀਆ', 'pa-Guru': 'ਸਾਮੀਆ', 'pcm': 'Saamiá Lángwej', 'pl': 'saamia', 'ps': 'سامیه', 'pt': 'saamia', 'qu': 'Saamia Simi', 'ro': 'saamia', 'ru': 'саамиа', 'sc': 'sàmia', 'sd': 'ساميا', 'sd-Arab': 'ساميا', 'si': 'සාමියා', 'sk': 'saamia', 'sl': 'saamijščina', 'so': 'Saamia', 'sq': 'samisht', 'sr': 'самиа', 'sr-Cyrl': 'самиа', 'sr-Latn': 'samia', 'sv': 'saamia', 'sw': 'Kisaamia', 'ta': 'சாமியா', 'te': 'సామియా', 'th': 'ซาเมีย', 'ti': 'ሳምያ', 'tk': 'samiýa dili', 'to': 'lea fakasāmia', 'tr': 'Samia dili', 'uk': 'самія', 'ur': 'سامیہ', 'uz': 'saamia', 'uz-Latn': 'saamia', 'vec': 'samia', 'vi': 'Tiếng Saamia', 'yo': 'Èdè Samia', 'yue': '薩米亞文', 'yue-Hans': '萨米亚文', 'yue-Hant': '薩米亞文', 'zh': '萨米亚语', 'zh-Hans': '萨米亚语', 'zh-Hant': '薩米亞文', 'zu': 'isi-Saamia'},
    'lsn': {'en': 'Tibetan Sign Language'},
    'lso': {'en': 'Laos Sign Language'},
    'lsp': {'en': 'Panamanian Sign Language'},
    'lsr': {'en': 'Aruop'},
    'lss': {'en': 'Lasi'},
    'lst': {'en': 'Trinidad and Tobago Sign Language'},
    'lsv': {'en': 'Sivia Sign Language'},
    'lsy': {'en': 'Mauritian Sign Language'},
    'lt': {'af': 'Litaus', 'ak': 'Lituania kasa', 'am': 'ሊቱዌንያኛ', 'ar': 'الليتوانية', 'as': 'লিথুৱানিয়ান', 'ast': 'lituanu', 'az': 'litva', 'az-Cyrl': 'литва', 'az-Latn': 'litva', 'bal-Latn': 'Litwániái', 'be': 'літоўская', 'be-tarask': 'літоўская', 'bg': 'литовски', 'bgn': 'لیتوانی', 'bn': 'লিথুয়েনীয়', 'br': 'lituaneg', 'brx': 'लिथुवानियान', 'bs': 'litvanski', 'bs-Cyrl': 'литвански', 'bs-Latn': 'litvanski', 'ca': 'lituà', 'ca-ES-valencia': 'lituà', 'ccp': '𑄣𑄨𑄗𑄪𑄠𑄬𑄚𑄩𑄠𑄧', 'ce': 'литвахойн', 'chr': 'ᎵᏚᏩᏂᎠᏂ', 'ckb': 'لیتوانی', 'cs': 'litevština', 'cy': 'Lithwaneg', 'da': 'litauisk', 'de': 'Litauisch', 'dsb': 'litawšćina', 'dz': 'ལི་ཐུ་ཝེ་ནི་ཡཱན་ཁ', 'ee': 'lithuaniagbe', 'el': 'Λιθουανικά', 'el-polyton': 'Λιθουανικά', 'en': 'Lithuanian', 'eo': 'litova', 'es': 'lituano', 'et': 'leedu', 'eu': 'lituaniera', 'fa': 'لیتوانیایی', 'ff-Adlm': '𞤂𞤭𞤼𞤮𞤲𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'liettua', 'fil': 'Lithuanian', 'fo': 'litaviskt', 'fr': 'lituanien', 'fur': 'lituan', 'fy': 'Litouws', 'ga': 'Liotuáinis', 'gd': 'Liotuainis', 'gl': 'lituano', 'gsw': 'Litauisch', 'gu': 'લિથુઆનિયન', 'ha': 'Lituweniyanci', 'he': 'ליטאית', 'hi': 'लिथुआनियाई', 'hi-Latn': 'Lithuanian', 'hr': 'litavski', 'hsb': 'litawšćina', 'ht': 'lituanien', 'hu': 'litván', 'hy': 'լիտվերեն', 'ia': 'lithuano', 'id': 'Lituania', 'ig': 'Lithuanian', 'is': 'litháíska', 'it': 'lituano', 'ja': 'リトアニア語', 'jv': 'Lithuania', 'ka': 'ლიეტუვური', 'kea': 'lituanu', 'kgp': 'rituvỹnũ', 'kk': 'литва тілі', 'kk-Cyrl': 'литва тілі', 'km': 'លីទុយអានី', 'kn': 'ಲಿಥುವೇನಿಯನ್', 'ko': '리투아니아어', 'kok': 'लिथुआनियन', 'kok-Deva': 'लिथुआनियन', 'ks': 'لِتھوانِیَن', 'ks-Arab': 'لِتھوانِیَن', 'ksh': 'Littouesch', 'ku': 'lîtwanî', 'kxv': 'lituaniyati', 'kxv-Deva': 'लितुआनियाति', 'kxv-Latn': 'lituaniyati', 'kxv-Orya': 'ଲିଥୁଆନିଆତି', 'kxv-Telu': 'లితువేనియన్', 'ky': 'литовчо', 'lb': 'Litauesch', 'lkt': 'Lithuania Iyápilt', 'lld': 'lituan', 'lo': 'ລິທົວນຽນ', 'lrc': 'لیتوڤانیایی', 'lt': 'lietuvių', 'lv': 'lietuviešu', 'mai': 'लिथुआनियाई', 'mi': 'Rituānia', 'mk': 'литвански', 'ml': 'ലിത്വാനിയൻ', 'mn': 'литва', 'mr': 'लिथुआनियन', 'ms': 'Lithuania', 'mt': 'Litwan', 'my': 'လစ်သူဝေးနီးယား', 'mzn': 'لتونیایی', 'nb': 'litauisk', 'ne': 'लिथुआनियाली', 'nl': 'Litouws', 'nn': 'litauisk', 'no': 'litauisk', 'nqo': 'ߟߌߕߎ߳ߦߊߣߌߞߊ߲', 'om': 'Afaan Liituniyaa', 'or': 'ଲିଥୁଆନିଆନ୍', 'pa': 'ਲਿਥੁਆਨੀਅਨ', 'pa-Guru': 'ਲਿਥੁਆਨੀਅਨ', 'pcm': 'Lituéniá Lángwej', 'pl': 'litewski', 'ps': 'ليتواني', 'pt': 'lituano', 'qu': 'Lituano Simi', 'rm': 'lituan', 'ro': 'lituaniană', 'ru': 'литовский', 'rw': 'Ikilituwaniya', 'sc': 'lituanu', 'sd': 'ليٿونيائي', 'sd-Arab': 'ليٿونيائي', 'se': 'liettuvagiella', 'si': 'ලිතුවේනියානු', 'sk': 'litovčina', 'sl': 'litovščina', 'smn': 'liettuakielâ', 'so': 'Lituwaanays', 'sq': 'lituanisht', 'sr': 'литвански', 'sr-Cyrl': 'литвански', 'sr-Latn': 'litvanski', 'sv': 'litauiska', 'sw': 'Kilithuania', 'ta': 'லிதுவேனியன்', 'te': 'లిథువేనియన్', 'tg': 'литвонӣ', 'th': 'ลิทัวเนีย', 'ti': 'ሊትዌንኛ', 'tk': 'litwa dili', 'to': 'lea fakalituania', 'tr': 'Litvanca', 'tt': 'литва', 'ug': 'لىتۋانىچە', 'uk': 'литовська', 'ur': 'لیتھوینین', 'uz': 'litva', 'uz-Cyrl': 'литвача', 'uz-Latn': 'litva', 'vec': 'lituan', 'vi': 'Tiếng Litva', 'wae': 'Litauiš', 'wo': 'Lituyaniye', 'xnr': 'लिथुआनियाई', 'yi': 'ליטוויש', 'yo': 'Èdè Lithuania', 'yrl': 'rituanu', 'yue': '立陶宛文', 'yue-Hans': '立陶宛文', 'yue-Hant': '立陶宛文', 'zh': '立陶宛语', 'zh-Hans': '立陶宛语', 'zh-Hant': '立陶宛文', 'zu': 'isi-Lithuanian'},
    'ltc': {'en': 'Late Middle Chinese'},
    'ltg': {'ast': 'latgalianu', 'bal-Latn': 'Latgali', 'cs': 'latgalština', 'cy': 'Latgaleg', 'de': 'Lettgallisch', 'en': 'Latgalian', 'et': 'latgali', 'fi': 'latgalli', 'fr': 'latgalien', 'gd': 'Latgailis', 'hi-Latn': 'Latgalian', 'ht': 'latgalien', 'ia': 'latgaliano', 'ig': 'Latgalian', 'it': 'letgallo', 'ja': 'ラトガリア語', 'ku': 'latgalî', 'lb': 'Lettgallesch', 'lt': 'latgalių', 'mk': 'латгалски', 'nb': 'latgallisk', 'ne': 'लाट्गाली', 'nl': 'Letgaals', 'nn': 'latgallisk', 'no': 'latgallisk', 'pl': 'łatgalski', 'sv': 'lettgalliska', 'th': 'ลัตเกล', 'ti': 'ላትጋላዊ', 'to': 'lea fakalatakale', 'tr': 'Latgalian', 'uk': 'латгальська', 'vec': 'letgalian', 'yue': '拉特加萊文', 'yue-Hans': '拉特加莱文', 'yue-Hant': '拉特加萊文', 'zh-Hant': '拉特加萊文'},
    'lth': {'en': 'Thur'},
    'lti': {'en': 'Leti (Indonesia)'},
    'ltn': {'en': 'Latundê'},
    'lto': {'en': 'Tsotso'},
    'lts': {'en': 'Tachoni'},
    'ltu': {'en': 'Latu'},
    'lu': {'af': 'Luba-Katanga', 'am': 'ሉባ-ካታንጋ', 'ar': 'اللوبا كاتانغا', 'as': 'লুবা-কাটাংগা', 'ast': 'luba-katanga', 'az': 'luba-katanqa', 'az-Cyrl': 'луба-катанга', 'az-Latn': 'luba-katanqa', 'bal-Latn': 'Lubá-Katangá', 'be': 'луба-катанга', 'be-tarask': 'луба-катанга', 'bg': 'луба-катанга', 'bgn': 'لوبایی‐کاتانگا', 'bn': 'লুবা-কাটাঙ্গা', 'br': 'luba-katanga', 'brx': 'लुबा-काटाङा', 'bs': 'luba-katanga', 'bs-Cyrl': 'луба-катанга', 'bs-Latn': 'luba-katanga', 'ca': 'luba katanga', 'ca-ES-valencia': 'luba katanga', 'ccp': '𑄣𑄪𑄝-𑄇𑄑𑄋𑄴𑄉', 'ce': 'луба-катанга', 'chr': 'ᎷᏆ-ᎧᏔᎦ', 'ckb': 'لووبا کاتانگا', 'cs': 'lubu-katanžština', 'cy': 'Luba-Katanga', 'da': 'luba-Katanga', 'de': 'Luba-Katanga', 'dsb': 'luba-katanga', 'el': 'Λούμπα-Κατάνγκα', 'el-polyton': 'Λούμπα-Κατάνγκα', 'en': 'Luba-Katanga', 'es': 'luba-katanga', 'et': 'Katanga luba', 'eu': 'Katangako lubera', 'fa': 'لوبایی‐کاتانگا', 'ff-Adlm': '𞤂𞤵𞤦𞤢-𞤑𞤢𞤼𞤢𞤲𞤺𞤢𞥄𞤪𞤫', 'fi': 'katanganluba', 'fil': 'Luba-Katanga', 'fo': 'luba-katanga', 'fr': 'luba-katanga (kiluba)', 'fy': 'Luba-Katanga', 'ga': 'Lúba-Cataingis', 'gd': 'Luba-Katanga', 'gl': 'luba-katanga', 'gsw': 'Luba', 'gu': 'લૂબા-કટાંગા', 'ha': 'Luba-Katanga', 'he': 'לובה-קטנגה', 'hi': 'ल्यूबा-कटांगा', 'hi-Latn': 'Luba-Katanga', 'hr': 'luba-katanga', 'hsb': 'luba-katanga', 'ht': 'luba-katanga (kiluba)', 'hu': 'luba-katanga', 'hy': 'լուբա-կատանգա', 'ia': 'luba-katanga', 'id': 'Luba-Katanga', 'ig': 'Luba-Katanga', 'is': 'lúbakatanga', 'it': 'luba-katanga', 'ja': 'ルバ・カタンガ語', 'jv': 'Luba-Katanga', 'ka': 'ლუბა-კატანგა', 'kea': 'luba-katanga', 'kgp': 'ruma-katỹga', 'kk': 'луба-катанга тілі', 'kk-Cyrl': 'луба-катанга тілі', 'km': 'លូបាកាតានហ្គា', 'kn': 'ಲೂಬಾ-ಕಟಾಂಗಾ', 'ko': '루바-카탄가어', 'kok': 'लुबा-काटांगा', 'kok-Deva': 'लुबा-काटांगा', 'ks': 'لوُبا کَتَنگا', 'ks-Arab': 'لوُبا کَتَنگا', 'ksh': 'Kiluba', 'ku': 'luba-katangayî', 'ky': 'луба-катангача', 'lb': 'Luba-Katanga', 'lld': 'luba-katanga', 'lo': 'ລູບາ-ຄາຕັງກາ', 'lrc': 'لوٙبا کاتانگا', 'lt': 'luba katanga', 'lu': 'Tshiluba', 'lv': 'lubakatanga', 'mai': 'लुबा-कटांगा', 'mi': 'Rupa Katanga', 'mk': 'луба-катанга', 'ml': 'ലുബ-കറ്റംഗ', 'mn': 'луба-катанга', 'mr': 'ल्यूबा-कटांगा', 'ms': 'Luba-Katanga', 'mt': 'Luba-Katanga', 'my': 'လူဘာ-ကတန်ဂါ', 'mzn': 'لوبا-کاتانگا', 'nb': 'luba-katanga', 'ne': 'लुबा-काताङ्गा', 'nl': 'Luba-Katanga', 'nn': 'luba-katanga', 'no': 'luba-katanga', 'nqo': 'ߞߊߕߊ߲ߜ߭ߊ߫-ߗߌ߬ߟߎ߬ߓߊ߬ߞߊ߲', 'or': 'ଲ୍ୟୁବା-କାଟାଙ୍ଗା', 'pa': 'ਲੂਬਾ-ਕਾਟਾਂਗਾ', 'pa-Guru': 'ਲੂਬਾ-ਕਾਟਾਂਗਾ', 'pcm': 'Lúbá-Katángá Lángwej', 'pl': 'luba-katanga', 'ps': 'لوبا-کټنګا', 'pt': 'luba-catanga', 'qu': 'Luba-Katanga Simi', 'rm': 'luba-katanga', 'ro': 'luba-katanga', 'ru': 'луба-катанга', 'sc': 'luba-katanga', 'sd': 'لوبا-ڪتانگا', 'sd-Arab': 'لوبا-ڪتانگا', 'si': 'ලුබා-කටන්ගා', 'sk': 'lubčina (katanžská)', 'sl': 'luba-katanga', 'smn': 'katangaluba', 'so': 'Luuba-kataanga', 'sq': 'luba-katangaisht', 'sr': 'луба-катанга', 'sr-Cyrl': 'луба-катанга', 'sr-Latn': 'luba-katanga', 'sv': 'luba-katanga', 'sw': 'Kiluba-Katanga', 'ta': 'லுபா-கடாங்கா', 'te': 'లూబ-కటాంగ', 'th': 'ลูบา-กาตองกา', 'ti': 'ሉባ-ካታንጋ', 'tk': 'luba-katanga dili', 'to': 'lea fakalupa-katanga', 'tr': 'Luba-Katanga', 'ug': 'لۇبا-كاتانگاچە', 'uk': 'луба-катанга', 'ur': 'لبا-کاتانجا', 'uz': 'luba-katanga', 'uz-Cyrl': 'луба-катанга', 'uz-Latn': 'luba-katanga', 'vec': 'luba katanga', 'vi': 'Tiếng Luba-Katanga', 'xnr': 'ल्यूबा-कटांगा', 'yo': 'Lúbà-Katanga', 'yrl': 'ruba-katãga', 'yue': '魯巴加丹加文', 'yue-Hans': '鲁巴加丹加文', 'yue-Hant': '魯巴加丹加文', 'zh': '鲁巴加丹加语', 'zh-Hans': '鲁巴加丹加语', 'zh-Hant': '魯巴加丹加文', 'zu': 'isi-Luba-Katanga'},
    'lua': {'af': 'Luba-Lulua', 'am': 'ሉባ-ሉሏ', 'ar': 'اللبا-لؤلؤ', 'as': 'লুবা-লুলুৱা', 'ast': 'luba-lulua', 'az': 'luba-lulua', 'az-Cyrl': 'луба-лулуа', 'az-Latn': 'luba-lulua', 'be': 'луба-касаі', 'be-tarask': 'луба-касаі', 'bg': 'луба-лулуа', 'bgn': 'لوبایی‐لولوا', 'bn': 'লুবা-লুলুয়া', 'br': 'luba-lulua', 'brx': 'लुबा लुलुआ', 'bs': 'luba-lulua', 'bs-Cyrl': 'луба-лулуа', 'bs-Latn': 'luba-lulua', 'ca': 'luba-lulua', 'ca-ES-valencia': 'luba-lulua', 'ccp': '𑄣𑄪𑄝-𑄣𑄪𑄣𑄪𑄠', 'ce': 'луба-лулуа', 'chr': 'ᎷᏆ-ᎷᎷᎠ', 'ckb': 'لووبا لوولووا', 'cs': 'luba-luluaština', 'cy': 'Luba-Lulua', 'da': 'luba-Lulua', 'de': 'Luba-Lulua', 'dsb': 'luba-luluašćina', 'el': 'Λούμπα-Λουλούα', 'el-polyton': 'Λούμπα-Λουλούα', 'en': 'Luba-Lulua', 'es': 'luba-lulua', 'et': 'lulua', 'eu': 'Kasai mendebaldeko lubera', 'fa': 'لوبایی‐لولوا', 'ff-Adlm': '𞤂𞤵𞤦𞤢 𞤑𞤢𞤧𞤢𞤭𞤪𞤫', 'fi': 'luluanluba', 'fil': 'Luba-Lulua', 'fo': 'luba-lulua', 'fr': 'luba-kasaï (ciluba)', 'fy': 'Luba-Lulua', 'ga': 'Luba-Lulua', 'gd': 'Luba-Lulua', 'gl': 'luba-lulua', 'gsw': 'Luba-Lulua', 'gu': 'લૂબા-લુલુઆ', 'ha': 'Luba-Lulua', 'he': 'לובה-לולואה', 'hi': 'ल्यूबा-लुलुआ', 'hi-Latn': 'Luba-Lulua', 'hr': 'luba-lulua', 'hsb': 'luba-luluašćina', 'ht': 'luba-kasaï (ciluba)', 'hu': 'luba-lulua', 'hy': 'լուբա-լուլուա', 'ia': 'luba-lulua', 'id': 'Luba-Lulua', 'ig': 'Luba-Lulua', 'is': 'luba-lulua', 'it': 'luba-lulua', 'ja': 'ルバ・ルルア語', 'jv': 'Luba-Lulua', 'ka': 'ლუბა-ლულუა', 'kgp': 'ruma-ruruva', 'kk': 'луба-лулуа тілі', 'kk-Cyrl': 'луба-лулуа тілі', 'km': 'លូបាលូឡា', 'kn': 'ಲುಬ-ಲುಲಾ', 'ko': '루바-룰루아어', 'kok': 'लुबा-लुलुआ', 'kok-Deva': 'लुबा-लुलुआ', 'ks': 'لوٗبا لوٗلُوا', 'ks-Arab': 'لوٗبا لوٗلُوا', 'ksh': 'Tschiluba', 'ku': 'luba-kasayî', 'ky': 'луба-лулуача', 'lb': 'Luba-Lulua', 'lo': 'ລູບາ-ລູລົວ', 'lt': 'luba lulua', 'lv': 'lubalulva', 'mai': 'लुबा-लुलुआ', 'mi': 'Rupa Rurua', 'mk': 'луба-лулуа', 'ml': 'ലൂബ-ലുലുവ', 'mn': 'луба-лулуа', 'mr': 'लुबा-लुलुआ', 'ms': 'Luba-Lulua', 'mt': 'Luba-Luluwa', 'my': 'လူဘာ-လူလူအာ', 'nb': 'luba-lulua', 'ne': 'लुबा-लुलुआ', 'nl': 'Luba-Lulua', 'nn': 'luba-lulua', 'no': 'luba-lulua', 'or': 'ଲୁବା-ଲୁଲୁଆ', 'pa': 'ਲਿਊਬਾ-ਲਿਊਲਿਆ', 'pa-Guru': 'ਲਿਊਬਾ-ਲਿਊਲਿਆ', 'pcm': 'Luba-Lúlua', 'pl': 'luba-lulua', 'ps': 'لبا لولوا', 'pt': 'luba-lulua', 'qu': 'Luba-Lulua Simi', 'rm': 'luba-lulua', 'ro': 'luba-lulua', 'ru': 'луба-лулуа', 'sc': 'tshiluba', 'sd': 'لوبا-لولوا', 'sd-Arab': 'لوبا-لولوا', 'si': 'ලුබ-ලුලුඅ', 'sk': 'lubčina (luluánska)', 'sl': 'luba-lulua', 'smn': 'lulualuba', 'so': 'Luba-Lulua', 'sq': 'luba-luluaisht', 'sr': 'луба-лулуа', 'sr-Cyrl': 'луба-лулуа', 'sr-Latn': 'luba-lulua', 'sv': 'luba-lulua', 'sw': 'Kiluba-Lulua', 'ta': 'லுபா-லுலுலா', 'te': 'లుబా-లులువ', 'th': 'ลูบา-ลูลัว', 'ti': 'ሉባ-ሉልዋ', 'tk': 'luba-Lulua dili', 'to': 'lea fakalupa-lulua', 'tr': 'Luba-Lulua', 'ug': 'لۇبا-لۇئاچە', 'uk': 'луба-лулуа', 'ur': 'لیوبا لولوآ', 'uz': 'luba-lulua', 'uz-Latn': 'luba-lulua', 'vec': 'luba lulua', 'vi': 'Tiếng Luba-Lulua', 'xnr': 'ल्यूबा-लुलुआ', 'yo': 'Èdè Luba Lulua', 'yrl': 'ruba-rurua', 'yue': '魯巴魯魯亞文', 'yue-Hans': '鲁巴鲁鲁亚文', 'yue-Hant': '魯巴魯魯亞文', 'zh': '卢巴-卢拉语', 'zh-Hans': '卢巴-卢拉语', 'zh-Hant': '魯巴魯魯亞文', 'zu': 'isi-Luba-Lulua'},
    'luc': {'en': 'Aringa'},
    'lud': {'en': 'Ludian'},
    'lue': {'en': 'Luvale'},
    'luf': {'en': 'Laua'},
    'lui': {'ar': 'اللوسينو', 'ast': 'luiseno', 'az': 'luyseno', 'az-Latn': 'luyseno', 'bg': 'луисеньо', 'bn': 'লুইসেনো', 'br': 'luiseno', 'brx': 'लुईसेनो', 'bs': 'luiseno', 'bs-Cyrl': 'луисено', 'bs-Latn': 'luiseno', 'ca': 'luisenyo', 'ca-ES-valencia': 'luisenyo', 'ccp': '𑄣𑄭𑄪𑄥𑄬𑄚𑄮', 'cs': 'luiseňo', 'da': 'luiseno', 'de': 'Luiseno', 'el': 'Λουισένο', 'el-polyton': 'Λουισένο', 'en': 'Luiseno', 'es': 'luiseño', 'et': 'luisenjo', 'fa': 'لویسنو', 'fi': 'luiseño', 'fr': 'luiseño', 'fy': 'Luiseno', 'gd': 'Luiseño', 'gsw': 'Luiseno-Schpraach', 'gu': 'લુઇસેનો', 'he': 'לויסנו', 'hi': 'लुइसेनो', 'hi-Latn': 'Luiseno', 'hr': 'luiseno', 'ht': 'luiseño', 'hu': 'luiseno', 'id': 'Luiseno', 'is': 'lúisenó', 'it': 'luiseno', 'ja': 'ルイセーニョ語', 'ka': 'ლუისენიო', 'kgp': 'rujsẽnũ', 'kn': 'ಲೂಯಿಸೆನೋ', 'ko': '루이세노어', 'ks': 'لویِسینو', 'ks-Arab': 'لویِسینو', 'lb': 'Luiseno-Sprooch', 'lo': 'ລູເຊໂນ', 'lt': 'luiseno', 'lv': 'luisenu', 'mk': 'лујсењски', 'ml': 'ലൂയിസെനോ', 'mr': 'लुइसेनो', 'mt': 'Luiseno', 'nb': 'luiseno', 'ne': 'लुइसेनो', 'nl': 'Luiseno', 'nn': 'luiseno', 'no': 'luiseno', 'or': 'ଲୁଇସେନୋ', 'pl': 'luiseno', 'pt': 'luiseno', 'rm': 'luiseno', 'ro': 'luiseno', 'ru': 'луисеньо', 'sk': 'luiseňo', 'sl': 'luisenščina', 'sr': 'луисењо', 'sr-Cyrl': 'луисењо', 'sr-Latn': 'luisenjo', 'sv': 'luiseño', 'ta': 'லுய்சேனோ', 'te': 'లుయిసెనో', 'th': 'ลุยเซโน', 'to': 'lea fakaluiseno', 'tr': 'Luiseno', 'ug': 'لۇيسېنگوچە', 'uk': 'луїсеньо', 'vi': 'Tiếng Luiseno', 'yrl': 'ruisenu', 'yue': '路易塞諾文', 'yue-Hans': '路易塞诺文', 'yue-Hant': '路易塞諾文', 'zh': '卢伊塞诺语', 'zh-Hans': '卢伊塞诺语', 'zh-Hant': '路易塞諾文'},
    'luj': {'en': 'Luna'},
    'luk': {'en': 'Lunanakha'},
    'lul': {'en': "Olu'bo"},
    'lum': {'en': 'Luimbi'},
    'lun': {'af': 'Lunda', 'am': 'ሉንዳ', 'ar': 'اللوندا', 'as': 'লুণ্ডা', 'ast': 'lunda', 'az': 'lunda', 'az-Cyrl': 'лунда', 'az-Latn': 'lunda', 'be': 'лунда', 'be-tarask': 'лунда', 'bg': 'лунда', 'bgn': 'لوندایی', 'bn': 'লুন্ডা', 'br': 'lunda', 'brx': 'लुण्डा', 'bs': 'lunda', 'bs-Cyrl': 'лунда', 'bs-Latn': 'lunda', 'ca': 'lunda', 'ca-ES-valencia': 'lunda', 'ccp': '𑄣𑄪𑄚𑄴𑄓', 'ce': 'лунда', 'chr': 'ᎷᎾᏓ', 'ckb': 'لووندا', 'cs': 'lundština', 'cy': 'Lwnda', 'da': 'lunda', 'de': 'Lunda', 'dsb': 'lundašćina', 'el': 'Λούντα', 'el-polyton': 'Λούντα', 'en': 'Lunda', 'es': 'lunda', 'et': 'lunda', 'eu': 'lundera', 'fa': 'لوندایی', 'ff-Adlm': '𞤂𞤵𞤲𞤣𞤢𞥄𞤪𞤫', 'fi': 'lunda', 'fil': 'Lunda', 'fo': 'lunda', 'fr': 'lunda', 'fy': 'Lunda', 'ga': 'Lundais', 'gd': 'Lunda', 'gl': 'lunda', 'gsw': 'Lunda-Schpraach', 'gu': 'લુન્ડા', 'ha': 'Lunda', 'he': 'לונדה', 'hi': 'लुन्डा', 'hi-Latn': 'Lunda', 'hr': 'lunda', 'hsb': 'lundašćina', 'ht': 'lunda', 'hu': 'lunda', 'hy': 'լունդա', 'ia': 'lunda', 'id': 'Lunda', 'ig': 'Lunda', 'is': 'lúnda', 'it': 'lunda', 'ja': 'ルンダ語', 'jv': 'Lunda', 'ka': 'ლუნდა', 'kgp': 'rũna', 'kk': 'лунда тілі', 'kk-Cyrl': 'лунда тілі', 'km': 'លុនដា', 'kn': 'ಲುಂಡಾ', 'ko': '룬다어', 'kok': 'लुंडा', 'kok-Deva': 'लुंडा', 'ks': 'لُندا', 'ks-Arab': 'لُندا', 'ksh': 'Schilunnda', 'ku': 'lunda', 'ky': 'лундача', 'lb': 'Lunda-Sprooch', 'lo': 'ລຸນດາ', 'lt': 'Lundos', 'lv': 'lundu', 'mai': 'लुंडा', 'mi': 'Runa', 'mk': 'лунда', 'ml': 'ലുൻഡ', 'mn': 'лунда', 'mr': 'लुन्डा', 'ms': 'Lunda', 'mt': 'Lunda', 'my': 'လွန်ဒါ', 'nb': 'lunda', 'ne': 'लुन्डा', 'nl': 'Lunda', 'nn': 'lunda', 'no': 'lunda', 'or': 'ଲୁଣ୍ଡା', 'pa': 'ਲੁੰਡਾ', 'pa-Guru': 'ਲੁੰਡਾ', 'pcm': 'Lunda Lángwej', 'pl': 'lunda', 'ps': 'لندا', 'pt': 'lunda', 'qu': 'Lunda Simi', 'rm': 'lunda', 'ro': 'lunda', 'ru': 'лунда', 'sc': 'lunda', 'sd': 'لنڊا', 'sd-Arab': 'لنڊا', 'si': 'ලුන්ඩ', 'sk': 'lunda', 'sl': 'lunda', 'smn': 'lunda', 'so': 'Lunda', 'sq': 'lundaisht', 'sr': 'лунда', 'sr-Cyrl': 'лунда', 'sr-Latn': 'lunda', 'sv': 'lunda', 'sw': 'Kilunda', 'ta': 'லூன்டா', 'te': 'లుండా', 'th': 'ลันดา', 'ti': 'ሉንዳ', 'tk': 'lunda dili', 'to': 'lea fakalunitā', 'tr': 'Lunda', 'ug': 'لۇنداچە', 'uk': 'лунда', 'ur': 'لونڈا', 'uz': 'lunda', 'uz-Latn': 'lunda', 'vec': 'lunda', 'vi': 'Tiếng Lunda', 'xnr': 'लुन्डा', 'yo': 'Èdè Lunda', 'yrl': 'rũda', 'yue': '盧恩達文', 'yue-Hans': '卢恩达文', 'yue-Hant': '盧恩達文', 'zh': '隆达语', 'zh-Hans': '隆达语', 'zh-Hant': '盧恩達文', 'zu': 'isi-Lunda'},
    'luo': {'am': 'ሉኦ', 'ar': 'اللو', 'as': 'লুও', 'az-Cyrl': 'луо', 'bal-Latn': 'Luó', 'be': 'луо', 'be-tarask': 'луо', 'bg': 'луо', 'bgn': 'لوئویی', 'bn': 'লুয়ো', 'brx': 'लुअ’', 'bs-Cyrl': 'луо', 'ccp': '𑄣𑄪𑄠𑄮', 'ce': 'луо (Кени а, Танзани а)', 'chr': 'ᎷᎣ', 'ckb': 'لووئۆ', 'cs': 'luoština', 'cy': 'Lŵo', 'el': 'Λούο', 'el-polyton': 'Λούο', 'eu': 'luoera', 'fa': 'لوئویی', 'ff-Adlm': '𞤂𞤵𞤮𞥅𞤪𞤫', 'ga': 'Lúóis', 'gsw': 'Luo-Schpraach', 'gu': 'લ્યુઓ', 'he': 'לואו', 'hi': 'ल्युओ', 'hy': 'լուո', 'is': 'lúó', 'ja': 'ルオ語', 'ka': 'ლუო', 'kgp': 'ruvo', 'kk': 'луо тілі', 'kk-Cyrl': 'луо тілі', 'km': 'លូអូ', 'kn': 'ಲುವೋ', 'ko': '루오어', 'kok': 'लुओ', 'kok-Deva': 'लुओ', 'ks': 'لُوو', 'ks-Arab': 'لُوو', 'ku': 'luoyî', 'ky': 'луочо', 'lb': 'Luo-Sprooch', 'lo': 'ລົວ', 'lrc': 'لوٙ', 'luo': 'Dholuo', 'mai': 'लुओ', 'mi': 'Ruo', 'mk': 'луо', 'ml': 'ലുവോ', 'mn': 'луо', 'mr': 'ल्युओ', 'my': 'လူအို', 'mzn': 'لوئو', 'ne': 'लुओ', 'nqo': 'ߟߎ߳ߏߞߊ߲', 'or': 'ଲୁଓ', 'pa': 'ਲੂਓ', 'pa-Guru': 'ਲੂਓ', 'pcm': 'Luó Lángwej', 'ps': 'لو', 'qu': 'Luo Simi', 'ru': 'луо', 'sd': 'لو', 'sd-Arab': 'لو', 'si': 'ලුඔ', 'so': 'Luwada', 'sq': 'luoisht', 'sr': 'луо', 'sr-Cyrl': 'луо', 'sw': 'Kijaluo', 'ta': 'லுயோ', 'te': 'లువో', 'th': 'ลัว', 'ti': 'ሉኦ', 'tk': 'luo dili', 'to': 'lea fakaluo', 'ug': 'لۇئوچە', 'uk': 'луо', 'ur': 'لو', 'vec': 'doluó', 'vi': 'Tiếng Luo', 'xnr': 'ल्युओ', 'yrl': 'ruwu', 'yue': '盧奧文', 'yue-Hans': '卢奥文', 'yue-Hant': '盧奧文', 'zh': '卢奥语', 'zh-Hans': '卢奥语', 'zh-Hant': '盧奧文', 'zu': 'isi-Luo', 'en': 'Luo (Kenya and Tanzania)'},
    'lup': {'en': 'Lumbu'},
    'luq': {'en': 'Lucumi'},
    'lur': {'en': 'Laura'},
    'lus': {'af': 'Mizo', 'am': 'ሚዞ', 'ar': 'الميزو', 'as': 'মিজো', 'ast': 'mizo', 'az': 'mizo', 'az-Cyrl': 'мизо', 'az-Latn': 'mizo', 'be': 'мізо', 'be-tarask': 'мізо', 'bg': 'мизо', 'bgn': 'لوشه\u200cای', 'bn': 'মিজো', 'br': 'lushai', 'brx': 'लुशाई', 'bs': 'mizo', 'bs-Cyrl': 'лушаи', 'bs-Latn': 'mizo', 'ca': 'mizo', 'ca-ES-valencia': 'mizo', 'ccp': '𑄟𑄨𑄎𑄮', 'ce': 'лушей', 'chr': 'ᎻᏐ', 'ckb': 'میزۆ', 'cs': 'mizoština', 'cy': 'Lwshaieg', 'da': 'lushai', 'de': 'Lushai', 'dsb': 'mizošćina', 'el': 'Μίζο', 'el-polyton': 'Μίζο', 'en': 'Mizo', 'es': 'mizo', 'et': 'lušei', 'eu': 'mizoera', 'fa': 'میزو', 'ff-Adlm': '𞤃𞤭𞤧𞤮𞥅𞤪𞤫', 'fi': 'lusai', 'fil': 'Mizo', 'fo': 'mizo', 'fr': 'lushaï', 'fy': 'Lushai', 'ga': 'Míosóis', 'gd': 'Mizo', 'gl': 'mizo', 'gsw': 'Lushai-Schpraach', 'gu': 'મિઝો', 'ha': 'Mizo', 'he': 'מיזו', 'hi': 'मिज़ो', 'hi-Latn': 'Mizo', 'hr': 'lushai', 'hsb': 'mizošćina', 'ht': 'lushaï', 'hu': 'lushai', 'hy': 'միզո', 'ia': 'mizo', 'id': 'Mizo', 'ig': 'Mizo', 'is': 'lúsaí', 'it': 'lushai', 'ja': 'ミゾ語', 'jv': 'Miso', 'ka': 'მიზო', 'kgp': 'rusaj', 'kk': 'мизо тілі', 'kk-Cyrl': 'мизо тілі', 'km': 'មីហ្សូ', 'kn': 'ಮಿಝೋ', 'ko': '루샤이어', 'kok': 'मिझो', 'kok-Deva': 'मिझो', 'ks': 'لُسہاے', 'ks-Arab': 'لُسہاے', 'ku': 'mizoyî', 'ky': 'мизочо', 'lb': 'Lushai-Sprooch', 'lkt': 'Mizo Iyápi', 'lo': 'ມີໂຊ', 'lt': 'mizo', 'lv': 'lušeju', 'mai': 'मिजो', 'mi': 'Mīho', 'mk': 'мизо', 'ml': 'മിസോ', 'mn': 'мизо', 'mr': 'मिझो', 'ms': 'Mizo', 'mt': 'Mizo', 'my': 'မီဇို', 'nb': 'mizo', 'ne': 'मिजो', 'nl': 'Mizo', 'nn': 'lushai', 'no': 'mizo', 'or': 'ମିଜୋ', 'pa': 'ਮਿਜ਼ੋ', 'pa-Guru': 'ਮਿਜ਼ੋ', 'pcm': 'Mizo Lángwej', 'pl': 'mizo', 'ps': 'ميزو', 'pt': 'lushai', 'qu': 'Mizo Simi', 'rm': 'lushai', 'ro': 'mizo', 'ru': 'мизо', 'sc': 'mizo', 'sd': 'ميزو', 'sd-Arab': 'ميزو', 'si': 'මිසො', 'sk': 'mizorámčina', 'sl': 'mizojščina', 'smn': 'lusai', 'so': 'Mizo', 'sq': 'mizoisht', 'sr': 'мизо', 'sr-Cyrl': 'мизо', 'sr-Latn': 'mizo', 'sv': 'lushai', 'sw': 'Kimizo', 'ta': 'மிஸோ', 'te': 'మిజో', 'th': 'มิโซ', 'ti': 'ማይዞ', 'tk': 'mizo dili', 'to': 'lea fakamiso', 'tr': 'Lushai', 'ug': 'مىزوچە', 'uk': 'мізо', 'ur': 'میزو', 'uz': 'lushay', 'uz-Cyrl': 'лушай', 'uz-Latn': 'lushay', 'vec': 'mizo', 'vi': 'Tiếng Lushai', 'xnr': 'मिज़ो', 'yi': 'מיזא', 'yo': 'Èdè Miso', 'yrl': 'ruxai', 'yue': '盧晒文', 'yue-Hans': '卢晒文', 'yue-Hant': '盧晒文', 'zh': '米佐语', 'zh-Hans': '米佐语', 'zh-Hant': '米佐文', 'zu': 'isi-Mizo'},
    'lut': {'en': 'Lushootseed'},
    'luu': {'en': 'Lumba-Yakkha'},
    'luv': {'en': 'Luwati'},
    'luw': {'en': 'Luo (Cameroon)'},
    'luy': {'af': 'Luyia', 'am': 'ሉያ', 'ar': 'لغة اللويا', 'as': 'লুইয়া', 'ast': 'luyia', 'az': 'luyia', 'az-Cyrl': 'лујиа', 'az-Latn': 'luyia', 'bal-Latn': 'Luiá', 'be': 'луйя', 'be-tarask': 'луйя', 'bg': 'лухя', 'bgn': 'لویایی', 'bn': 'লুইয়া', 'br': 'luyia', 'brx': 'लुईया', 'bs': 'luhija', 'bs-Cyrl': 'луја', 'bs-Latn': 'luhija', 'ca': 'luyia', 'ca-ES-valencia': 'luyia', 'ccp': '𑄣𑄭𑄪𑄠', 'ce': 'лухья', 'chr': 'ᎷᏱᎠ', 'ckb': 'لوویا', 'cs': 'luhja', 'cy': 'Lwyia', 'da': 'luyana', 'de': 'Luhya', 'dsb': 'luhya', 'ee': 'luyiagbe', 'el': 'Λούχια', 'el-polyton': 'Λούχια', 'en': 'Luyia', 'es': 'luyia', 'et': 'luhja', 'eu': 'luhyera', 'fa': 'لویا', 'ff-Adlm': '𞤂𞤵𞤴𞤭𞤢𞥄𞤪𞤫', 'fi': 'luhya', 'fil': 'Luyia', 'fo': 'luyia', 'fr': 'luyia', 'fy': 'Luyia', 'ga': 'Luyia', 'gd': 'Luyia', 'gl': 'luyia', 'gu': 'લુઈયા', 'ha': 'Luyia', 'he': 'לויה', 'hi': 'ल्युईआ', 'hi-Latn': 'Luyia', 'hr': 'luyia', 'hsb': 'luhya', 'ht': 'luyia', 'hu': 'lujia', 'hy': 'լույա', 'ia': 'luyia', 'id': 'Luyia', 'ig': 'Luyia', 'is': 'luyia', 'it': 'luyia', 'ja': 'ルヒヤ語', 'jv': 'Luyia', 'ka': 'ლუჰია', 'kea': 'luyia', 'kgp': 'ruja', 'kk': 'лухиа тілі', 'kk-Cyrl': 'лухиа тілі', 'km': 'លូយ៉ា', 'kn': 'ಲುಯಿಯ', 'ko': '루야어', 'kok': 'ल्युइया', 'kok-Deva': 'ल्युइया', 'ksh': 'Luhjesch', 'ku': 'luhyayî', 'ky': 'лухияча', 'lb': 'Olulujia', 'lo': 'ລູໄຍ', 'lrc': 'لوٙئیا', 'lt': 'luja', 'luy': 'Luluhia', 'lv': 'luhju', 'mai': 'लुुइया', 'mi': 'Rūia', 'mk': 'луја', 'ml': 'ലുയിയ', 'mn': 'луяа', 'mr': 'ल्युइया', 'ms': 'Luyia', 'mt': 'Luyia', 'my': 'လူရီအာ', 'mzn': 'لوییا', 'nb': 'luhya', 'ne': 'लुइया', 'nl': 'Luyia', 'nn': 'olulujia', 'no': 'luhya', 'nqo': 'ߟߎ߳ߦߌߞߊ߲', 'or': 'ଲୁୟିଆ', 'pa': 'ਲੂਈਆ', 'pa-Guru': 'ਲੂਈਆ', 'pcm': 'Luyia Lángwej', 'pl': 'luhya', 'ps': 'لویا', 'pt': 'luyia', 'qu': 'Luyia Simi', 'rm': 'luyia', 'ro': 'luyia', 'ru': 'лухья', 'sc': 'luyia', 'sd': 'لوهيا', 'sd-Arab': 'لوهيا', 'si': 'ලුයියා', 'sk': 'luhja', 'sl': 'luhijščina', 'smn': 'luhya', 'so': 'Luyia', 'sq': 'lujaisht', 'sr': 'лујиа', 'sr-Cyrl': 'лујиа', 'sr-Latn': 'lujia', 'sv': 'luhya', 'sw': 'Kiluhya', 'ta': 'லுயியா', 'te': 'లుయియ', 'th': 'ลูเยีย', 'ti': 'ሉይያ', 'tk': 'luýýa dili', 'to': 'lea fakaluīa', 'tr': 'Luyia', 'ug': 'لۇياچە', 'uk': 'луйя', 'ur': 'لویا', 'uz': 'luhya', 'uz-Cyrl': 'луҳя', 'uz-Latn': 'luhya', 'vec': 'luja', 'vi': 'Tiếng Luyia', 'xnr': 'ल्युईआ', 'yo': 'Luyíà', 'yrl': 'ruiya', 'yue': '盧雅文', 'yue-Hans': '卢雅文', 'yue-Hant': '盧雅文', 'zh': '卢雅语', 'zh-Hans': '卢雅语', 'zh-Hant': '盧雅文', 'zu': 'isi-Luyia'},
    'luz': {'en': 'Southern Luri'},
    'lv': {'af': 'Letties', 'ak': 'Latvia kasa', 'am': 'ላትቪያኛ', 'ar': 'اللاتفية', 'as': 'লাটভিয়ান', 'ast': 'letón', 'az': 'latış', 'az-Cyrl': 'латыш', 'az-Latn': 'latış', 'bal-Latn': 'Latwiái', 'be': 'латышская', 'be-tarask': 'латышская', 'bg': 'латвийски', 'bgn': 'لاتوینی', 'bn': 'লাত্\u200cভীয়', 'br': 'latvieg', 'brx': 'लाटभियान', 'bs': 'latvijski', 'bs-Cyrl': 'латвијски', 'bs-Latn': 'latvijski', 'ca': 'letó', 'ca-ES-valencia': 'letó', 'ccp': '𑄣𑄖𑄴𑄞𑄩𑄠𑄧', 'ce': 'латышийн', 'chr': 'ᎳᏘᏫᎠᏂ', 'ckb': 'لێتۆنی', 'cs': 'lotyština', 'cy': 'Latfieg', 'da': 'lettisk', 'de': 'Lettisch', 'dsb': 'letišćina', 'dz': 'ལཊ་བི་ཡཱན་ཁ', 'ee': 'latviagbe', 'el': 'Λετονικά', 'el-polyton': 'Λετονικά', 'en': 'Latvian', 'en-Dsrt': '𐐢𐐰𐐻𐑂𐐨𐐲𐑌', 'eo': 'latva', 'es': 'letón', 'et': 'läti', 'eu': 'letoniera', 'fa': 'لتونیایی', 'ff-Adlm': '𞤂𞤢𞤼𞤾𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'latvia', 'fil': 'Latvian', 'fo': 'lettiskt', 'fr': 'letton', 'fur': 'leton', 'fy': 'Letlâns', 'ga': 'Laitvis', 'gd': 'Laitbheis', 'gl': 'letón', 'gsw': 'Lettisch', 'gu': 'લાતવિયન', 'ha': 'Latbiyanci', 'he': 'לטבית', 'hi': 'लातवियाई', 'hi-Latn': 'Latvian', 'hr': 'latvijski', 'hsb': 'letišćina', 'ht': 'letton', 'hu': 'lett', 'hy': 'լատվիերեն', 'ia': 'letton', 'id': 'Latvia', 'ie': 'lettonian', 'ig': 'Latvian', 'is': 'lettneska', 'it': 'lettone', 'ja': 'ラトビア語', 'jv': 'Latvia', 'ka': 'ლატვიური', 'kea': 'letãu', 'kgp': 'retỹv', 'kk': 'латыш тілі', 'kk-Cyrl': 'латыш тілі', 'km': 'ឡាតវី', 'kn': 'ಲಾಟ್ವಿಯನ್', 'ko': '라트비아어', 'kok': 'लात्वियन', 'kok-Deva': 'लात्वियन', 'ks': 'لَتوِیَن', 'ks-Arab': 'لَتوِیَن', 'ksh': 'Lättesch', 'ku': 'latvîyayî', 'kxv': 'laṭviati', 'kxv-Deva': 'लातवियाति', 'kxv-Latn': 'laṭviati', 'kxv-Orya': 'ଲାଟବିଆତି', 'kxv-Telu': 'లాట్వియన్', 'ky': 'латышча', 'lb': 'Lettesch', 'lkt': 'Latvia Iyápi', 'lld': 'leton', 'lo': 'ລັດວຽນ', 'lrc': 'لاتوڤیایی', 'lt': 'latvių', 'lv': 'latviešu', 'mai': 'लातवियाई', 'mi': 'Rāwhia', 'mk': 'латвиски', 'ml': 'ലാറ്റ്വിയൻ', 'mn': 'латви', 'mr': 'लात्व्हियन', 'ms': 'Latvia', 'mt': 'Latvjan', 'my': 'လတ်ဗီးယား', 'mzn': 'لاتویایی', 'nb': 'latvisk', 'ne': 'लात्भियाली', 'nl': 'Lets', 'nn': 'latvisk', 'no': 'latvisk', 'nqo': 'ߟߋߕߐ߲ߞߊ߲', 'om': 'Afaan Lativiyaa', 'or': 'ଲାଟଭିଆନ୍', 'pa': 'ਲਾਤੀਵੀ', 'pa-Guru': 'ਲਾਤੀਵੀ', 'pcm': 'Látvián Lángwej', 'pl': 'łotewski', 'ps': 'لېټواني', 'pt': 'letão', 'qu': 'Leton Simi', 'rm': 'letton', 'ro': 'letonă', 'ru': 'латышский', 'rw': 'Ikinyaletoviyani', 'sc': 'lètone', 'sd': 'لاتوين', 'sd-Arab': 'لاتوين', 'se': 'látviagiella', 'si': 'ලැට්වියානු', 'sk': 'lotyština', 'sl': 'latvijščina', 'smn': 'latviakielâ', 'so': 'Laatfiyaan', 'sq': 'letonisht', 'sr': 'летонски', 'sr-Cyrl': 'летонски', 'sr-Latn': 'letonski', 'sv': 'lettiska', 'sw': 'Kilatvia', 'ta': 'லாட்வியன்', 'te': 'లాట్వియన్', 'tg': 'латишӣ', 'th': 'ลัตเวีย', 'ti': 'ላትቭኛ', 'tk': 'latyş dili', 'to': 'lea fakalativia', 'tr': 'Letonca', 'tt': 'латыш', 'ug': 'لاتچە', 'uk': 'латиська', 'ur': 'لیٹوین', 'uz': 'latishcha', 'uz-Cyrl': 'латишча', 'uz-Latn': 'latishcha', 'vec': 'lètone', 'vi': 'Tiếng Latvia', 'wae': 'Lettiš', 'wo': 'Letoniye', 'xnr': 'लातवियाई', 'yi': 'לעטיש', 'yo': 'Èdè látífíànì', 'yrl': 'retãu', 'yue': '拉脫維亞文', 'yue-Hans': '拉脱维亚文', 'yue-Hant': '拉脫維亞文', 'zh': '拉脱维亚语', 'zh-Hans': '拉脱维亚语', 'zh-Hant': '拉脫維亞文', 'zu': 'isi-Latvian'},
    'lva': {'en': "Maku'a"},
    'lvi': {'en': 'Lavi'},
    'lvk': {'en': 'Lavukaleve'},
    'lvs': {'en': 'Standard Latvian'},
    'lvu': {'en': 'Levuka'},
    'lwa': {'en': 'Lwalu'},
    'lwe': {'en': 'Lewo Eleng'},
    'lwg': {'en': 'Wanga'},
    'lwh': {'en': 'White Lachi'},
    'lwl': {'en': 'Eastern Lawa'},
    'lwm': {'en': 'Laomian'},
    'lwo': {'en': 'Luwo'},
    'lws': {'en': 'Malawian Sign Language'},
    'lwt': {'en': 'Lewotobi'},
    'lwu': {'en': 'Lawu'},
    'lww': {'en': 'Lewo'},
    'lxm': {'en': 'Lakurumau'},
    'lya': {'en': 'Layakha'},
    'lyg': {'en': 'Lyngngam'},
    'lyn': {'en': 'Luyana'},
    'lzh': {'ast': 'chinu lliterariu', 'br': 'sinaeg lennegel', 'ca': 'xinès clàssic', 'ca-ES-valencia': 'xinès clàssic', 'cs': 'čínština (klasická)', 'de': 'Klassisches Chinesisch', 'en': 'Literary Chinese', 'et': 'klassikaline hiina', 'fa': 'چینی ادبی', 'fi': 'klassinen kiina', 'fr': 'chinois littéraire', 'gd': 'Sìnis an Litreachais', 'hi-Latn': 'Literary Chinese', 'ht': 'chinois littéraire', 'it': 'cinese classico', 'ja': '漢文', 'lb': 'Klassescht Chinesesch', 'lt': 'klasikinė kinų', 'mk': 'книжевен кинески', 'nb': 'klassisk kinesisk', 'ne': 'साहित्यिक चिनियाँ', 'nl': 'Klassiek Chinees', 'nn': 'klassisk kinesisk', 'no': 'klassisk kinesisk', 'pl': 'chiński klasyczny', 'sv': 'litterär kineiska', 'th': 'จีนคลาสสิก', 'to': 'lea fakasiaina-faʻutohi', 'tr': 'Edebi Çince', 'yue': '文言文', 'yue-Hans': '文言文', 'yue-Hant': '文言文', 'zh-Hant': '文言文'},
    'lzl': {'en': 'Litzlitz'},
    'lzn': {'en': 'Leinong Naga'},
    'lzz': {'ast': 'laz', 'ca': 'laz', 'ca-ES-valencia': 'laz', 'cs': 'lazština', 'de': 'Lasisch', 'en': 'Laz', 'et': 'lazi', 'fi': 'lazi', 'fr': 'laze', 'gd': 'Laz', 'hi-Latn': 'Laz', 'ht': 'laze', 'id': 'Laz', 'it': 'laz', 'ja': 'ラズ語', 'lb': 'Lasesch Sprooch', 'lt': 'laz', 'mk': 'ласки', 'nb': 'lazisk', 'ne': 'लाज', 'nl': 'Lazisch', 'nn': 'lazisk', 'no': 'lazisk', 'pl': 'lazyjski', 'sv': 'laziska', 'th': 'แลซ', 'to': 'lea fakalasu', 'tr': 'Lazca', 'yue': '拉茲文', 'yue-Hans': '拉兹文', 'yue-Hant': '拉茲文', 'zh-Hant': '拉茲文'},
    'maa': {'en': 'San Jerónimo Tecóatl Mazatec'},
    'mab': {'en': 'Yutanduchi Mixtec'},
    'mad': {'af': 'Madurees', 'am': 'ማዱረስ', 'ar': 'المادريز', 'as': 'মাদুৰেছে', 'ast': 'madurés', 'az': 'maduriz', 'az-Cyrl': 'мадуриз', 'az-Latn': 'maduriz', 'be': 'мадурская', 'be-tarask': 'мадурская', 'bg': 'мадурски', 'bgn': 'مادورایی', 'bn': 'মাদুরেজ', 'br': 'madoureg', 'brx': 'मादुरीज', 'bs': 'madureški', 'bs-Cyrl': 'мадурешки', 'bs-Latn': 'madureški', 'ca': 'madurès', 'ca-ES-valencia': 'madurès', 'ccp': '𑄟𑄘𑄪𑄢𑄬𑄥𑄬', 'ce': 'мадурийн', 'chr': 'ᎹᏚᎴᏏ', 'ckb': 'مادووری', 'cs': 'madurština', 'cy': 'Madwreg', 'da': 'madurese', 'de': 'Maduresisch', 'dsb': 'madurezišćina', 'el': 'Μαντουρίζ', 'el-polyton': 'Μαντουρίζ', 'en': 'Madurese', 'es': 'madurés', 'et': 'madura', 'eu': 'madurera', 'fa': 'مادورایی', 'ff-Adlm': '𞤃𞤢𞤣𞤵𞤪𞤫𞥅𞤪𞤫', 'fi': 'madura', 'fil': 'Madurese', 'fo': 'maduresiskt', 'fr': 'madurais', 'fy': 'Madurees', 'ga': 'Maidiúiris', 'gd': 'Cànan Madhura', 'gl': 'madurés', 'gsw': 'Maduresisch', 'gu': 'માદુરીસ', 'ha': 'Madurese', 'he': 'מדורזית', 'hi': 'मादुरीस', 'hi-Latn': 'Madurese', 'hr': 'madurski', 'hsb': 'madurezišćina', 'ht': 'madurais', 'hu': 'madurai', 'hy': 'մադուրերեն', 'ia': 'madurese', 'id': 'Madura', 'ig': 'Madurese', 'is': 'madúrska', 'it': 'madurese', 'ja': 'マドゥラ語', 'jv': 'Madura', 'ka': 'მადურული', 'kgp': 'mỹnurej', 'kk': 'мадур тілі', 'kk-Cyrl': 'мадур тілі', 'km': 'ម៉ាឌូរីស', 'kn': 'ಮದುರೀಸ್', 'ko': '마두라어', 'kok': 'मादुरीज', 'kok-Deva': 'मादुरीज', 'ks': 'مَدُریٖز', 'ks-Arab': 'مَدُریٖز', 'ku': 'madurayî', 'ky': 'мадурисче', 'lb': 'Maduresesch', 'lo': 'ມາດູລາ', 'lt': 'madurezų', 'lv': 'maduriešu', 'mai': 'मदुरीज', 'mi': 'Maturīhi', 'mk': 'мадурски', 'ml': 'മദുരേസേ', 'mn': 'мадури хэл', 'mr': 'मादुरीस', 'ms': 'Madura', 'mt': 'Maduriż', 'my': 'မဒူရာ', 'nb': 'maduresisk', 'ne': 'मादुरेसे', 'nl': 'Madoerees', 'nn': 'maduresisk', 'no': 'maduresisk', 'or': 'ମାଦୁରୀସ୍', 'pa': 'ਮਾਡੂਰੀਸ', 'pa-Guru': 'ਮਾਡੂਰੀਸ', 'pcm': 'Madurẹ́sẹ', 'pl': 'madurski', 'ps': 'مدراسی', 'pt': 'madurês', 'qu': 'Madurese Simi', 'rm': 'madurais', 'ro': 'madureză', 'ru': 'мадурский', 'sc': 'maduresu', 'sd': 'مدورائي', 'sd-Arab': 'مدورائي', 'si': 'මදුරීස්', 'sk': 'madurčina', 'sl': 'madurščina', 'smn': 'madurakielâ', 'so': 'Madurese', 'sq': 'madurezisht', 'sr': 'мадурски', 'sr-Cyrl': 'мадурски', 'sr-Latn': 'madurski', 'sv': 'maduresiska', 'sw': 'Kimadura', 'ta': 'மதுரீஸ்', 'te': 'మాదురీస్', 'th': 'มาดูรา', 'ti': 'ማዱሪዝኛ', 'tk': 'madur dili', 'to': 'lea fakamatula', 'tr': 'Madura Dili', 'ug': 'مادۇرېسچە', 'uk': 'мадурська', 'ur': 'مدورسی', 'uz': 'madur', 'uz-Latn': 'madur', 'vec': 'madureze', 'vi': 'Tiếng Madura', 'xnr': 'मादुरीस', 'yo': 'Èdè Maduri', 'yrl': 'madurei', 'yue': '馬都拉文', 'yue-Hans': '马都拉文', 'yue-Hant': '馬都拉文', 'zh': '马都拉语', 'zh-Hans': '马都拉语', 'zh-Hant': '馬都拉文', 'zu': 'isi-Madurese'},
    'mae': {'en': 'Bo-Rukul'},
    'maf': {'ast': 'mafa', 'bs': 'mafa', 'bs-Latn': 'mafa', 'ca': 'mafa', 'ca-ES-valencia': 'mafa', 'cs': 'mafa', 'da': 'mafa', 'de': 'Mafa', 'el': 'Μάφα', 'el-polyton': 'Μάφα', 'en': 'Mafa', 'es': 'mafa', 'et': 'mafa', 'fi': 'mafa', 'fr': 'mafa', 'fy': 'Mafa', 'gd': 'Mafa', 'he': 'מאפאה', 'hi-Latn': 'Mafa', 'hr': 'mafa', 'ht': 'mafa', 'hu': 'mafa', 'id': 'Mafa', 'it': 'mafa', 'ja': 'マファ語', 'ka': 'მაფა', 'kgp': 'mafa', 'ko': '마파어', 'lb': 'Mafa', 'lo': 'ມາຟາ', 'lt': 'mafų', 'lv': 'mafu', 'mk': 'мафа', 'ml': 'മാഫ', 'ms': 'Mafa', 'nb': 'mafa', 'ne': 'माफा', 'nl': 'Mafa', 'nn': 'mafa', 'no': 'mafa', 'pl': 'mafa', 'pt': 'mafa', 'ro': 'mafa', 'ru': 'мафа', 'sk': 'mafa', 'sv': 'mafa', 'sw': 'Kimafa', 'th': 'มาฟา', 'to': 'lea fakamafa', 'tr': 'Mafa', 'ug': 'مافاچە', 'uk': 'мафа', 'vi': 'Tiếng Mafa', 'yrl': 'mafa', 'yue': '馬法文', 'yue-Hans': '马法文', 'yue-Hant': '馬法文', 'zh': '马法语', 'zh-Hans': '马法语', 'zh-Hant': '馬法文'},
    'mag': {'af': 'Magahi', 'am': 'ማጋሂ', 'ar': 'الماجا', 'as': 'মাগাহি', 'ast': 'magahi', 'az': 'maqahi', 'az-Cyrl': 'магаһи', 'az-Latn': 'maqahi', 'be': 'магахі', 'be-tarask': 'магахі', 'bg': 'магахи', 'bgn': 'ماگاهی', 'bn': 'মাগাহি', 'br': 'magahi', 'brx': 'मगाही', 'bs': 'magahi', 'bs-Cyrl': 'магахи', 'bs-Latn': 'magahi', 'ca': 'magahi', 'ca-ES-valencia': 'magahi', 'ccp': '𑄟𑄉𑄦𑄨', 'ce': 'магахи', 'chr': 'ᎹᎦᎯ', 'ckb': 'ماگاهی', 'cs': 'magahijština', 'cy': 'Magahi', 'da': 'magahi', 'de': 'Khotta', 'dsb': 'magahišćina', 'el': 'Μαγκάχι', 'el-polyton': 'Μαγκάχι', 'en': 'Magahi', 'es': 'magahi', 'et': 'magahi', 'eu': 'magadhera', 'fa': 'ماگاهیایی', 'ff-Adlm': '𞤃𞤢𞤺𞤢𞤸𞤭𞥅𞤪𞤫', 'fi': 'magahi', 'fil': 'Magahi', 'fo': 'magahi', 'fr': 'magahi', 'fy': 'Magahi', 'ga': 'Magaidis', 'gd': 'Magahi', 'gl': 'magahi', 'gsw': 'Khotta', 'gu': 'મગહી', 'ha': 'Magahi', 'he': 'מאגאהית', 'hi': 'मगही', 'hi-Latn': 'Magahi', 'hr': 'magahi', 'hsb': 'magahišćina', 'ht': 'magahi', 'hu': 'magahi', 'hy': 'մագահի', 'ia': 'magahi', 'id': 'Magahi', 'ig': 'Magahi', 'is': 'magahí', 'it': 'magahi', 'ja': 'マガヒー語', 'jv': 'Magahi', 'ka': 'მაგაჰი', 'kgp': 'mỹgahi', 'kk': 'магахи тілі', 'kk-Cyrl': 'магахи тілі', 'km': 'ម៉ាហ្គាហ៊ី', 'kn': 'ಮಗಾಹಿ', 'ko': '마가히어', 'kok': 'मगाही', 'kok-Deva': 'मगाही', 'ks': 'مَگاے', 'ks-Arab': 'مَگاے', 'ku': 'magahî', 'ky': 'магахиче', 'lb': 'Khotta', 'lo': 'ມາກາຮິ', 'lt': 'magahi', 'lv': 'magahiešu', 'mai': 'मगही', 'mi': 'Makāhi', 'mk': 'магахи', 'ml': 'മഗാഹി', 'mn': 'магахи', 'mr': 'मगही', 'ms': 'Magahi', 'mt': 'Magahi', 'my': 'မဂါဟီ', 'nb': 'magahi', 'ne': 'मगधी', 'nl': 'Magahi', 'nn': 'magahi', 'no': 'magahi', 'or': 'ମାଗାହୀ', 'pa': 'ਮਗਾਹੀ', 'pa-Guru': 'ਮਗਾਹੀ', 'pcm': 'Magahí Lángwej', 'pl': 'magahi', 'ps': 'مګهي', 'pt': 'magahi', 'qu': 'Magahi Simi', 'rm': 'magahi', 'ro': 'magahi', 'ru': 'магахи', 'sc': 'magahi', 'sd': 'مگاهي', 'sd-Arab': 'مگاهي', 'si': 'මඝහි', 'sk': 'magadhčina', 'sl': 'magadščina', 'smn': 'magahi', 'so': 'Magahi', 'sq': 'magaisht', 'sr': 'магахи', 'sr-Cyrl': 'магахи', 'sr-Latn': 'magahi', 'sv': 'magahi', 'sw': 'Kimagahi', 'ta': 'மகாஹி', 'te': 'మగాహి', 'th': 'มคหี', 'ti': 'ማጋሂ', 'tk': 'magahi dili', 'to': 'lea fakamakahi', 'tr': 'Magahi', 'ug': 'ماگاخىچە', 'uk': 'магадхі', 'ur': 'مگاہی', 'uz': 'magahi', 'uz-Latn': 'magahi', 'vec': 'magahi', 'vi': 'Tiếng Magahi', 'xnr': 'मगही', 'yo': 'Èdè Magahi', 'yrl': 'magarí', 'yue': '馬加伊文', 'yue-Hans': '马加伊文', 'yue-Hant': '馬加伊文', 'zh': '摩揭陀语', 'zh-Hans': '摩揭陀语', 'zh-Hant': '摩揭陀文', 'zu': 'isi-Magahi'},
    'mai': {'af': 'Maithili', 'ak': 'Maetili', 'am': 'ማይቲሊ', 'ar': 'المايثيلي', 'as': 'মৈথিলী', 'ast': 'maithili', 'az': 'maitili', 'az-Cyrl': 'маитили', 'az-Latn': 'maitili', 'bal-Latn': 'Maitéli', 'be': 'майтхілі', 'be-tarask': 'майтхілі', 'bg': 'майтхили', 'bgn': 'مایدیلی', 'bn': 'মৈথিলি', 'br': 'maithili', 'brx': 'मैथीली', 'bs': 'maitili', 'bs-Cyrl': 'маитили', 'bs-Latn': 'maitili', 'ca': 'maithili', 'ca-ES-valencia': 'maithili', 'ccp': '𑄟𑄳𑄆𑄧𑄗𑄨𑄣𑄨', 'ce': 'майтхили', 'chr': 'ᎹᏟᎵ', 'ckb': 'مائیتیلی', 'cs': 'maithiliština', 'cy': 'Maithili', 'da': 'maithili', 'de': 'Maithili', 'dsb': 'maithilšćina', 'el': 'Μαϊτχίλι', 'el-polyton': 'Μαϊτχίλι', 'en': 'Maithili', 'es': 'maithili', 'et': 'maithili', 'eu': 'maithilia', 'fa': 'مایدیلی', 'ff-Adlm': '𞤃𞤢𞤴𞤭𞤼𞤭𞤤𞤭𞥅𞤪𞤫', 'fi': 'maithili', 'fil': 'Maithili', 'fo': 'maithili', 'fr': 'maïthili', 'fy': 'Maithili', 'ga': 'Maitilis', 'gd': 'Maithili', 'gl': 'maithili', 'gsw': 'Maithili', 'gu': 'મૈથિલી', 'ha': 'Maithili', 'he': 'מאיטילית', 'hi': 'मैथिली', 'hi-Latn': 'Maithili', 'hr': 'maithili', 'hsb': 'maithilšćina', 'ht': 'maïthili', 'hu': 'maithili', 'hy': 'մայթիլի', 'ia': 'maithili', 'id': 'Maithili', 'ig': 'Maithili', 'is': 'maítílí', 'it': 'maithili', 'ja': 'マイティリー語', 'jv': 'Maithili', 'ka': 'მაითილი', 'kgp': 'mỹjtiri', 'kk': 'майтхили тілі', 'kk-Cyrl': 'майтхили тілі', 'km': 'ម៉ៃធីលី', 'kn': 'ಮೈಥಿಲಿ', 'ko': '마이틸리어', 'kok': 'मैथिली', 'kok-Deva': 'मैथिली', 'ks': 'میتَھلی', 'ks-Arab': 'میتَھلی', 'ku': 'maithili', 'kxv': 'maitilī', 'kxv-Deva': 'मेतिली', 'kxv-Latn': 'maitilī', 'kxv-Orya': 'ମଇତିଲୀ', 'kxv-Telu': 'మైతలి', 'ky': 'маитиличе', 'lb': 'Maithili', 'lo': 'ໄມທີລິ', 'lt': 'maithili', 'lv': 'maithili', 'mai': 'मैथिली', 'mi': 'Maitiri', 'mk': 'маитили', 'ml': 'മൈഥിലി', 'mn': 'май', 'mr': 'मैथिली', 'ms': 'Maithili', 'mt': 'Maithili', 'my': 'မိုင်သီလီ', 'nb': 'maithili', 'ne': 'मैथिली', 'nl': 'Maithili', 'nn': 'maithili', 'no': 'maithili', 'nqo': 'ߡߊߗߟߌߞߊ߲', 'or': 'ମୈଥିଳୀ', 'pa': 'ਮੈਥਲੀ', 'pa-Guru': 'ਮੈਥਲੀ', 'pcm': 'Maítíli', 'pl': 'maithili', 'ps': 'مایتھلي', 'pt': 'maithili', 'qu': 'Maithili Simi', 'rm': 'maithili', 'ro': 'maithili', 'ru': 'майтхили', 'sc': 'maithili', 'sd': 'ميٿلي', 'sd-Arab': 'ميٿلي', 'si': 'මයිතිලි', 'sk': 'maithilčina', 'sl': 'maitili', 'smn': 'maithili', 'so': 'Maithili', 'sq': 'maitilisht', 'sr': 'маитили', 'sr-Cyrl': 'маитили', 'sr-Latn': 'maitili', 'sv': 'maithili', 'sw': 'Kimaithili', 'ta': 'மைதிலி', 'te': 'మైథిలి', 'th': 'ไมถิลี', 'ti': 'ማይቲሊ', 'tk': 'maýthili dili', 'to': 'lea fakamaitili', 'tr': 'Maithili', 'ug': 'مايتىلىچە', 'uk': 'майтхілі', 'ur': 'میتھیلی', 'uz': 'maythili', 'uz-Latn': 'maythili', 'vec': 'maithili', 'vi': 'Tiếng Maithili', 'xnr': 'मैथिली', 'yo': 'Èdè Matihi', 'yrl': 'maitiri', 'yue': '邁蒂利文', 'yue-Hans': '迈蒂利文', 'yue-Hant': '邁蒂利文', 'zh': '迈蒂利语', 'zh-Hans': '迈蒂利语', 'zh-Hant': '邁蒂利文', 'zu': 'isi-Maithili'},
    'maj': {'en': 'Jalapa De Díaz Mazatec'},
    'mak': {'af': 'Makasar', 'am': 'ማካሳር', 'ar': 'الماكاسار', 'as': 'মাকাছাৰ', 'ast': 'makasar', 'az': 'makasar', 'az-Cyrl': 'макасар', 'az-Latn': 'makasar', 'be': 'макасар', 'be-tarask': 'макасар', 'bg': 'макасар', 'bgn': 'ماکاساری', 'bn': 'ম্যাকাসার', 'br': 'makasar', 'brx': 'माकासार', 'bs': 'makasar', 'bs-Cyrl': 'макасар', 'bs-Latn': 'makasar', 'ca': 'makassar', 'ca-ES-valencia': 'makassar', 'ccp': '𑄟𑄳𑄠𑄇𑄥𑄢𑄴', 'ce': 'макасарийн', 'chr': 'ᎹᎧᏌ', 'ckb': 'ماکاسار', 'cs': 'makasarština', 'cy': 'Macasareg', 'da': 'makasar', 'de': 'Makassarisch', 'dsb': 'makasaršćina', 'el': 'Μακασάρ', 'el-polyton': 'Μακασάρ', 'en': 'Makasar', 'es': 'macasar', 'et': 'makassari', 'eu': 'makassarera', 'fa': 'ماکاسار', 'ff-Adlm': '𞤃𞤢𞤳𞤢𞤧𞤢𞤪𞤢𞥄𞤪𞤫', 'fi': 'makassar', 'fil': 'Makasar', 'fo': 'makasar', 'fr': 'makassar', 'fy': 'Makassaars', 'ga': 'Macasairis', 'gd': 'Makasar', 'gl': 'makasar', 'gsw': 'Makassarisch', 'gu': 'મકાસર', 'ha': 'Makasar', 'he': 'מקסאר', 'hi': 'मकासर', 'hi-Latn': 'Makasar', 'hr': 'makasar', 'hsb': 'makasaršćina', 'ht': 'makassar', 'hu': 'makaszar', 'hy': 'մակասարերեն', 'ia': 'macassarese', 'id': 'Makasar', 'ig': 'Makasar', 'is': 'makasar', 'it': 'makasar', 'ja': 'マカッサル語', 'jv': 'Makasar', 'ka': 'მაკასარი', 'kgp': 'mỹkasar', 'kk': 'макасар тілі', 'kk-Cyrl': 'макасар тілі', 'km': 'ម៉ាកាសា', 'kn': 'ಮಕಾಸರ್', 'ko': '마카사어', 'kok': 'माकासार', 'kok-Deva': 'माकासार', 'ks': 'مَکَسار', 'ks-Arab': 'مَکَسار', 'ku': 'makasarî', 'ky': 'макасарча', 'lb': 'Makassaresch', 'lo': 'ມາກາຊາຣ', 'lt': 'Makasaro', 'lv': 'makasaru', 'mai': 'मकसर', 'mi': 'Makahā', 'mk': 'макасарски', 'ml': 'മകാസർ', 'mn': 'макасар', 'mr': 'मकस्सर', 'ms': 'Makasar', 'mt': 'Makasar', 'my': 'မကာဆာ', 'nb': 'makasar', 'ne': 'माकासार', 'nl': 'Makassaars', 'nn': 'makasar', 'no': 'makasar', 'or': 'ମକାସର୍', 'pa': 'ਮਕਾਸਰ', 'pa-Guru': 'ਮਕਾਸਰ', 'pcm': 'Mákásá Lángwej', 'pl': 'makasar', 'ps': 'مکاسار', 'pt': 'makasar', 'qu': 'Makasar Simi', 'rm': 'makassar', 'ro': 'makasar', 'ru': 'макассарский', 'sc': 'makassaresu', 'sd': 'مڪاسر', 'sd-Arab': 'مڪاسر', 'si': 'මකාසාර්', 'sk': 'makasarčina', 'sl': 'makasarščina', 'smn': 'makasar', 'so': 'Makasar', 'sq': 'makasarisht', 'sr': 'макасарски', 'sr-Cyrl': 'макасарски', 'sr-Latn': 'makasarski', 'sv': 'makasar', 'sw': 'Kimakasar', 'ta': 'மகாசார்', 'te': 'మకాసార్', 'th': 'มากาซาร์', 'ti': 'ማካሳር', 'tk': 'makasar dili', 'to': 'lea fakamakasali', 'tr': 'Makasar', 'ug': 'ماكاسارچە', 'uk': 'макасарська', 'ur': 'مکاسر', 'uz': 'makasar', 'uz-Latn': 'makasar', 'vec': 'makasar', 'vi': 'Tiếng Makasar', 'xnr': 'मकासर', 'yo': 'Èdè Makasa', 'yrl': 'makasá', 'yue': '望加錫文', 'yue-Hans': '望加锡文', 'yue-Hant': '望加錫文', 'zh': '望加锡语', 'zh-Hans': '望加锡语', 'zh-Hant': '望加錫文', 'zu': 'isi-Makasar'},
    'mam': {'en': 'Mam'},
    'man': {'ar': 'الماندينغ', 'ast': 'mandingo', 'az': 'məndinqo', 'az-Latn': 'məndinqo', 'be': 'мандынг', 'be-tarask': 'мандынг', 'bg': 'мандинго', 'bn': 'ম্যান্ডিঙ্গো', 'brx': 'मांडींगो', 'bs': 'mandingo', 'bs-Cyrl': 'мандинго', 'bs-Latn': 'mandingo', 'ca': 'mandinga', 'ca-ES-valencia': 'mandinga', 'ccp': '𑄟𑄳𑄠𑄚𑄴𑄓𑄨𑄁𑄉𑄮', 'cs': 'mandingština', 'cy': 'Mandingo', 'da': 'mandingo', 'de': 'Malinke', 'el': 'Μαντίνγκο', 'el-polyton': 'Μαντίνγκο', 'en': 'Mandingo', 'es': 'mandingo', 'et': 'malinke', 'fa': 'ماندینگویی', 'fi': 'mandingo', 'fr': 'mandingue', 'fy': 'Mandingo', 'gd': 'Mandingo', 'gsw': 'Manding-Schpraach', 'gu': 'મન્ડિન્ગો', 'he': 'מנדינגו', 'hi': 'मन्डिन्गो', 'hi-Latn': 'Mandingo', 'hr': 'mandingo', 'ht': 'mandingue', 'hu': 'mandingó', 'id': 'Mandingo', 'is': 'mandingó', 'it': 'mandingo', 'ja': 'マンディンゴ語', 'kgp': 'mỹnhĩga', 'kn': 'ಮಂಡಿಂಗೊ', 'ko': '만딩고어', 'ks': 'مَندِنگو', 'ks-Arab': 'مَندِنگو', 'lb': 'Manding-Sprooch', 'lo': 'ມັນດິງກາ', 'lt': 'mandingų', 'lv': 'mandingu', 'mk': 'мандинго', 'ml': 'മണ്ഡിൻഗോ', 'mr': 'मन्डिन्गो', 'mt': 'Mandingo', 'nb': 'mandingo', 'ne': 'मान्दिङो', 'nl': 'Mandingo', 'nn': 'mandingo', 'no': 'mandingo', 'or': 'ମାଣ୍ଡିଙ୍ଗୋ', 'pl': 'mandingo', 'pt': 'mandinga', 'rm': 'mandingo', 'ro': 'mandingo', 'ru': 'мандинго', 'sk': 'mandingo', 'sl': 'mandingo', 'sr': 'мандинго', 'sr-Cyrl': 'мандинго', 'sr-Latn': 'mandingo', 'sv': 'mande', 'ta': 'மான்டிங்கோ', 'te': 'మండింగో', 'th': 'มันดิงกา', 'to': 'lea fakamanitīngiko', 'tr': 'Mandingo', 'ug': 'ماندىنگوچە', 'uk': 'мандінго', 'vi': 'Tiếng Mandingo', 'yrl': 'mãdĩga', 'yue': '曼丁哥文', 'yue-Hans': '曼丁哥文', 'yue-Hant': '曼丁哥文', 'zh': '曼丁哥语', 'zh-Hans': '曼丁哥语', 'zh-Hant': '曼丁哥文'},
    'map': {'en': 'Austronesian languages'},
    'map-ata-pro': {'en': 'Proto-Atayalic'},
    'map-bms': {'en': 'Banyumasan'},
    'map-pro': {'en': 'Proto-Austronesian'},
    'maq': {'en': 'Chiquihuitlán Mazatec'},
    'mas': {'af': 'Masai', 'am': 'ማሳይ', 'ar': 'الماساي', 'as': 'মাছাই', 'ast': 'masái', 'az': 'masay', 'az-Cyrl': 'масај', 'az-Latn': 'masay', 'bal-Latn': 'Masai', 'be': 'маасай', 'be-tarask': 'маасай', 'bg': 'масайски', 'bgn': 'ماسایی', 'bn': 'মাসাই', 'br': 'masai', 'brx': 'मासाई', 'bs': 'masai', 'bs-Cyrl': 'масаи', 'bs-Latn': 'masai', 'ca': 'massai', 'ca-ES-valencia': 'massai', 'ccp': '𑄟𑄥𑄭', 'ce': 'масаи', 'chr': 'ᎹᏌᏱ', 'ckb': 'ماسایی', 'cs': 'masajština', 'cy': 'Masai', 'da': 'masai', 'de': 'Massai', 'dsb': 'masaišćina', 'el': 'Μασάι', 'el-polyton': 'Μασάι', 'en': 'Masai', 'es': 'masái', 'et': 'masai', 'eu': 'masaiera', 'fa': 'ماسایی', 'ff-Adlm': '𞤃𞤢𞤧𞤢𞤴𞤭𞥅𞤪𞤫', 'fi': 'maasai', 'fil': 'Masai', 'fo': 'masai', 'fr': 'maasaï', 'fy': 'Masai', 'ga': 'Másais', 'gd': 'Maasai', 'gl': 'masai', 'gsw': 'Massai-Schpraach', 'gu': 'મસાઇ', 'ha': 'Harshen Masai', 'he': 'מסאית', 'hi': 'मसाई', 'hi-Latn': 'Masai', 'hr': 'masajski', 'hsb': 'masaišćina', 'ht': 'maasaï', 'hu': 'masai', 'hy': 'մասաի', 'ia': 'masai', 'id': 'Masai', 'ig': 'Masai', 'is': 'masaí', 'it': 'masai', 'ja': 'マサイ語', 'jv': 'Masai', 'ka': 'მასაი', 'kgp': 'mỹsaj', 'kk': 'масай тілі', 'kk-Cyrl': 'масай тілі', 'km': 'ម៉ាសៃ', 'kn': 'ಮಸಾಯ್', 'ko': '마사이어', 'kok': 'मसाई', 'kok-Deva': 'मसाई', 'ks': 'مَساے', 'ks-Arab': 'مَساے', 'ku': 'masayî', 'ky': 'масайча', 'lb': 'Massai-Sprooch', 'lo': 'ມາໄຊ', 'lrc': 'ماسایی', 'lt': 'masajų', 'lv': 'masaju', 'mai': 'मसाई', 'mas': 'Maa', 'mi': 'Māhai', 'mk': 'масајски', 'ml': 'മസായ്', 'mn': 'масай', 'mr': 'मसाई', 'ms': 'Masai', 'mt': 'Masai', 'my': 'မာဆိုင်', 'mzn': 'ماسایی', 'nb': 'masai', 'ne': 'मसाई', 'nl': 'Maa', 'nn': 'masai', 'no': 'masai', 'nqo': 'ߡߊ߯ߛߊߞߊ߲', 'or': 'ମାସାଇ', 'pa': 'ਮਸਾਈ', 'pa-Guru': 'ਮਸਾਈ', 'pcm': 'Masaí Lángwej', 'pl': 'masajski', 'ps': 'ماسائي', 'pt': 'massai', 'qu': 'Masai Simi', 'rm': 'masai', 'ro': 'masai', 'ru': 'масаи', 'sc': 'masai', 'sd': 'مسائي', 'sd-Arab': 'مسائي', 'si': 'මසායි', 'sk': 'masajčina', 'sl': 'masajščina', 'smn': 'masaikielâ', 'so': 'Masaay', 'sq': 'masaisht', 'sr': 'масајски', 'sr-Cyrl': 'масајски', 'sr-Latn': 'masajski', 'sv': 'massajiska', 'sw': 'Kimaasai', 'ta': 'மாசாய்', 'te': 'మాసై', 'th': 'มาไซ', 'ti': 'ማሳይ', 'tk': 'masai dili', 'to': 'lea fakamasai', 'tr': 'Masai', 'ug': 'ماسايچە', 'uk': 'масаї', 'ur': 'مسائی', 'uz': 'masay', 'uz-Cyrl': 'масайча', 'uz-Latn': 'masay', 'vec': 'masài', 'vi': 'Tiếng Masai', 'xnr': 'मसाई', 'yo': 'Másáì', 'yrl': 'masai', 'yue': '馬賽文', 'yue-Hans': '马赛文', 'yue-Hant': '馬賽文', 'zh': '马赛语', 'zh-Hans': '马赛语', 'zh-Hant': '馬賽文', 'zu': 'isi-Masai'},
    'mat': {'en': 'San Francisco Matlatzinca'},
    'mau': {'en': 'Huautla Mazatec'},
    'mav': {'en': 'Sateré-Mawé'},
    'maw': {'en': 'Mampruli'},
    'max': {'en': 'North Moluccan Malay'},
    'maz': {'en': 'Central Mazahua'},
    'mba': {'en': 'Higaonon'},
    'mbb': {'en': 'Western Bukidnon Manobo'},
    'mbc': {'en': 'Macushi'},
    'mbd': {'en': 'Dibabawon Manobo'},
    'mbe': {'en': 'Molale'},
    'mbf': {'en': 'Baba Malay'},
    'mbh': {'en': 'Mangseng'},
    'mbi': {'en': 'Ilianen Manobo'},
    'mbj': {'en': 'Nadëb'},
    'mbk': {'en': 'Malol'},
    'mbl': {'en': 'Maxakalí'},
    'mbm': {'en': 'Ombamba'},
    'mbn': {'en': 'Macaguán'},
    'mbo': {'en': 'Mbo (Cameroon)'},
    'mbp': {'en': 'Malayo'},
    'mbq': {'en': 'Maisin'},
    'mbr': {'en': 'Nukak Makú'},
    'mbs': {'en': 'Sarangani Manobo'},
    'mbt': {'en': 'Matigsalug Manobo'},
    'mbu': {'en': 'Mbula-Bwazza'},
    'mbv': {'en': 'Mbulungish'},
    'mbw': {'en': 'Maring'},
    'mbx': {'en': 'Mari (East Sepik Province)'},
    'mby': {'en': 'Memoni'},
    'mbz': {'en': 'Amoltepec Mixtec'},
    'mca': {'en': 'Maca'},
    'mcb': {'en': 'Machiguenga'},
    'mcc': {'en': 'Bitur'},
    'mcd': {'en': 'Sharanahua'},
    'mce': {'en': 'Itundujia Mixtec'},
    'mcf': {'en': 'Matsés'},
    'mcg': {'en': 'Mapoyo'},
    'mch': {'en': 'Maquiritari'},
    'mci': {'en': 'Mese'},
    'mcj': {'en': 'Mvanip'},
    'mck': {'en': 'Mbunda'},
    'mcl': {'en': 'Macaguaje'},
    'mcm': {'en': 'Malaccan Creole Portuguese'},
    'mcn': {'en': 'Masana'},
    'mco': {'en': 'Coatlán Mixe'},
    'mcp': {'en': 'Makaa'},
    'mcq': {'en': 'Ese'},
    'mcr': {'en': 'Menya'},
    'mcs': {'en': 'Mambai'},
    'mct': {'en': 'Mengisa'},
    'mcu': {'en': 'Cameroon Mambila'},
    'mcv': {'en': 'Minanibai'},
    'mcw': {'en': 'Mawa (Chad)'},
    'mcx': {'en': 'Mpiemo'},
    'mcy': {'en': 'South Watut'},
    'mcz': {'en': 'Mawan'},
    'mda': {'en': 'Mada (Nigeria)'},
    'mdb': {'en': 'Morigi'},
    'mdc': {'en': 'Male (Papua New Guinea)'},
    'mdd': {'en': 'Mbum'},
    'mde': {'ar': 'مابا', 'ast': 'maba', 'bs': 'maba', 'bs-Latn': 'maba', 'ca': 'maba', 'ca-ES-valencia': 'maba', 'cs': 'maba', 'da': 'maba', 'de': 'Maba', 'el': 'Μάμπα', 'el-polyton': 'Μάμπα', 'en': 'Maba', 'es': 'maba', 'et': 'maba', 'fi': 'maba', 'fr': 'maba', 'fy': 'Maba', 'gd': 'Maba', 'he': 'מאבא', 'hi-Latn': 'Maba', 'hr': 'maba', 'ht': 'maba', 'hu': 'maba', 'id': 'Maba', 'it': 'maba', 'ja': 'マバ語', 'ka': 'მაბა', 'kgp': 'mama', 'ko': '마바어', 'lb': 'Maba', 'lo': 'ມາບາ', 'lt': 'mabų', 'lv': 'mabu', 'mk': 'маба', 'ml': 'മാബ', 'ms': 'Maba', 'nb': 'maba', 'ne': 'माबा', 'nl': 'Maba', 'nn': 'maba', 'no': 'maba', 'pl': 'maba', 'pt': 'maba', 'ro': 'maba', 'ru': 'маба', 'sk': 'maba', 'sv': 'maba', 'sw': 'Kimaba', 'th': 'มาบา', 'to': 'lea fakamapa', 'tr': 'Maba', 'ug': 'ماباچە', 'uk': 'маба', 'vi': 'Tiếng Maba', 'yrl': 'maba', 'yue': '馬巴文', 'yue-Hans': '马巴文', 'yue-Hant': '馬巴文', 'zh': '马坝语', 'zh-Hans': '马坝语', 'zh-Hant': '馬巴文'},
    'mdf': {'af': 'Moksha', 'am': 'ሞክሻ', 'ar': 'الموكشا', 'as': 'মোক্সা', 'ast': 'moksha', 'az': 'mokşa', 'az-Cyrl': 'мокша', 'az-Latn': 'mokşa', 'bal-Latn': 'Moksha', 'be': 'макшанская', 'be-tarask': 'макшанская', 'bg': 'мокша', 'bgn': 'موکشی', 'bn': 'মোকশা', 'br': 'moksha', 'brx': 'मोक्षा', 'bs': 'mokša', 'bs-Cyrl': 'мокша', 'bs-Latn': 'mokša', 'ca': 'mordovià moksa', 'ca-ES-valencia': 'mordovià moksa', 'ccp': '𑄟𑄮𑄇𑄴𑄥', 'ce': 'мокшанойн', 'chr': 'ᎼᎧᏌ', 'ckb': 'مۆکشا', 'cs': 'mokšanština', 'cy': 'Mocsia', 'da': 'moksha', 'de': 'Mokschanisch', 'dsb': 'mokshašćina', 'el': 'Μόκσα', 'el-polyton': 'Μόκσα', 'en': 'Moksha', 'es': 'moksha', 'et': 'mokša', 'eu': 'mokxera', 'fa': 'مکشایی', 'ff-Adlm': '𞤃𞤮𞤳𞤧𞤢𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'mokša', 'fil': 'Moksha', 'fo': 'moksha', 'fr': 'mokcha', 'fy': 'Moksha', 'ga': 'Mocsais', 'gd': 'Moksha', 'gl': 'moksha', 'gsw': 'Mokschamordwinisch', 'gu': 'મોક્ષ', 'ha': 'Moksha', 'he': 'מוקשה', 'hi': 'मोक्ष', 'hi-Latn': 'Moksha', 'hr': 'moksha', 'hsb': 'mokshašćina', 'ht': 'mokcha', 'hu': 'moksán', 'hy': 'մոկշայերեն', 'ia': 'moksha', 'id': 'Moksha', 'ig': 'Moksha', 'is': 'moksa', 'it': 'moksha', 'ja': 'モクシャ語', 'jv': 'Moksha', 'ka': 'მოქშა', 'kgp': 'mogsa', 'kk': 'мокша тілі', 'kk-Cyrl': 'мокша тілі', 'km': 'មុខសា', 'kn': 'ಮೋಕ್ಷ', 'ko': '모크샤어', 'kok': 'मोक्ष', 'kok-Deva': 'मोक्ष', 'ks': 'موکشا', 'ks-Arab': 'موکشا', 'ksh': 'Mokschah', 'ku': 'mokşayî', 'ky': 'мокшача', 'lb': 'Moksha', 'lo': 'ມອກຊາ', 'lt': 'mokša', 'lv': 'mokšu', 'mai': 'मोकशा', 'mdf': 'мокшень кяль', 'mi': 'Mōkaha', 'mk': 'мокшански', 'ml': 'മോക്ഷ', 'mn': 'мокша', 'mr': 'मोक्ष', 'ms': 'Moksha', 'mt': 'Moksha', 'my': 'မို့ခ်ရှာ', 'nb': 'moksja', 'ne': 'मोक्ष', 'nl': 'Moksja', 'nn': 'moksja', 'no': 'moksja', 'or': 'ମୋକ୍ଷ', 'pa': 'ਮੋਕਸ਼ਾ', 'pa-Guru': 'ਮੋਕਸ਼ਾ', 'pcm': 'Móksha Lángwej', 'pl': 'moksza', 'ps': 'موکشا', 'pt': 'mocsa', 'qu': 'Moksha Simi', 'rm': 'moksha', 'ro': 'moksha', 'ru': 'мокшанский', 'sc': 'moksha', 'sd': 'موڪشا', 'sd-Arab': 'موڪشا', 'se': 'mokšagiella', 'si': 'මොක්ශා', 'sk': 'mokšiančina', 'sl': 'mokšavščina', 'smn': 'mokšakielâ', 'so': 'Moksha', 'sq': 'mokshaisht', 'sr': 'мокша', 'sr-Cyrl': 'мокша', 'sr-Latn': 'mokša', 'sv': 'moksja', 'sw': 'Lugha ya Moksha', 'ta': 'மோக்க்ஷா', 'te': 'మోక్ష', 'th': 'มอคชา', 'ti': 'ሞክሻ', 'tk': 'mokşa dili', 'to': 'lea fakamokisiā', 'tr': 'Mokşa dili', 'ug': 'موكشاچە', 'uk': 'мокша', 'ur': 'موکشا', 'uz': 'moksha', 'uz-Cyrl': 'мокша тили', 'uz-Latn': 'moksha', 'vec': 'moksha', 'vi': 'Tiếng Moksha', 'xnr': 'मुगती', 'yo': 'Èdè Mokisa', 'yrl': 'mukisa', 'yue': '莫克沙文', 'yue-Hans': '莫克沙文', 'yue-Hant': '莫克沙文', 'zh': '莫克沙语', 'zh-Hans': '莫克沙语', 'zh-Hant': '莫克沙文', 'zu': 'isi-Moksha'},
    'mdg': {'en': 'Massalat'},
    'mdh': {'en': 'Maguindanaon'},
    'mdi': {'en': 'Mamvu'},
    'mdj': {'en': 'Mangbetu'},
    'mdk': {'en': 'Mangbutu'},
    'mdl': {'en': 'Maltese Sign Language'},
    'mdm': {'en': 'Mayogo'},
    'mdn': {'en': 'Mbati'},
    'mdp': {'en': 'Mbala'},
    'mdq': {'en': 'Mbole'},
    'mdr': {'ar': 'الماندار', 'ast': 'mandar', 'az': 'mandar', 'az-Latn': 'mandar', 'bg': 'мандар', 'bn': 'ম্যাণ্ডার', 'br': 'mandar', 'brx': 'मंदार', 'bs': 'mandar', 'bs-Cyrl': 'мандар', 'bs-Latn': 'mandar', 'ca': 'mandar', 'ca-ES-valencia': 'mandar', 'ccp': '𑄟𑄳𑄠𑄚𑄴𑄓𑄢𑄴', 'cs': 'mandar', 'cy': 'Mandareg', 'da': 'mandar', 'de': 'Mandaresisch', 'el': 'Μανδάρ', 'el-polyton': 'Μανδάρ', 'en': 'Mandar', 'es': 'mandar', 'et': 'mandari', 'fa': 'ماندار', 'fi': 'mandar', 'fr': 'mandar', 'fy': 'Mandar', 'gd': 'Mandar', 'gsw': 'Mandaresisch', 'gu': 'મંદાર', 'he': 'מנדאר', 'hi': 'मंदार', 'hi-Latn': 'Mandar', 'hr': 'mandar', 'ht': 'mandar', 'hu': 'mandar', 'id': 'Mandar', 'is': 'mandar', 'it': 'mandar', 'ja': 'マンダル語', 'kgp': 'mỹnar', 'kn': 'ಮಂದಾರ್', 'ko': '만다르어', 'ks': 'مَندَر', 'ks-Arab': 'مَندَر', 'lb': 'Mandaresesch', 'lo': 'ມານດາຣ', 'lt': 'mandarų', 'lv': 'mandaru', 'mk': 'мандарски', 'ml': 'മണ്ഡാർ', 'mr': 'मंडार', 'mt': 'Mandar', 'nb': 'mandar', 'ne': 'मन्दर', 'nl': 'Mandar', 'nn': 'mandar', 'no': 'mandar', 'or': 'ମନ୍ଦାର', 'pl': 'mandar', 'pt': 'mandar', 'rm': 'mandar', 'ro': 'mandar', 'ru': 'мандарский', 'sk': 'mandarčina', 'sl': 'mandarščina', 'sr': 'мандар', 'sr-Cyrl': 'мандар', 'sr-Latn': 'mandar', 'sv': 'mandar', 'ta': 'மான்டார்', 'te': 'మండార్', 'th': 'มานดาร์', 'to': 'lea fakamanetali', 'tr': 'Mandar', 'ug': 'ماندارچە', 'uk': 'мандарська', 'vi': 'Tiếng Mandar', 'yrl': 'mãdari', 'yue': '曼達文', 'yue-Hans': '曼达文', 'yue-Hant': '曼達文', 'zh': '曼达尔语', 'zh-Hans': '曼达尔语', 'zh-Hant': '曼達文'},
    'mds': {'en': 'Maria (Papua New Guinea)'},
    'mdt': {'en': 'Mbere'},
    'mdu': {'en': 'Mboko'},
    'mdv': {'en': 'Santa Lucía Monteverde Mixtec'},
    'mdw': {'en': 'Mbosi'},
    'mdx': {'en': 'Dizin'},
    'mdy': {'en': 'Male (Ethiopia)'},
    'mdz': {'en': 'Suruí Do Pará'},
    'mea': {'en': 'Menka'},
    'meb': {'en': 'Ikobi'},
    'mec': {'en': 'Marra'},
    'med': {'en': 'Melpa'},
    'mee': {'en': 'Mengen'},
    'mef': {'en': 'Megam'},
    'meg': {'en': 'Mea'},
    'meh': {'en': 'Southwestern Tlaxiaco Mixtec'},
    'mei': {'en': 'Midob'},
    'mej': {'en': 'Meyah'},
    'mek': {'en': 'Mekeo'},
    'mel': {'en': 'Central Melanau'},
    'mem': {'en': 'Mangala'},
    'men': {'af': 'Mende', 'am': 'ሜንዴ', 'ar': 'الميند', 'as': 'মেণ্ডে', 'ast': 'mende', 'az': 'mende', 'az-Cyrl': 'менде', 'az-Latn': 'mende', 'be': 'мендэ', 'be-tarask': 'мендэ', 'bg': 'менде', 'bgn': 'منده\u200cای', 'bn': 'মেন্ডে', 'br': 'mende', 'brx': 'मेन्दे', 'bs': 'mende', 'bs-Cyrl': 'менде', 'bs-Latn': 'mende', 'ca': 'mende', 'ca-ES-valencia': 'mende', 'ccp': '𑄟𑄬𑄚𑄴𑄓𑄬', 'ce': 'менде', 'chr': 'ᎺᎾᏕ', 'ckb': 'مێندێ', 'cs': 'mende', 'cy': 'Mendeg', 'da': 'mende', 'de': 'Mende', 'dsb': 'mendišćina', 'el': 'Μέντε', 'el-polyton': 'Μέντε', 'en': 'Mende', 'es': 'mende', 'et': 'mende', 'eu': 'mendeera', 'fa': 'منده\u200cای', 'ff-Adlm': '𞤃𞤫𞤲𞤣𞤫𞥅𞤪𞤫', 'fi': 'mende', 'fil': 'Mende', 'fo': 'mende', 'fr': 'mendé', 'fy': 'Mende', 'ga': 'Meindis', 'gd': 'Mende', 'gl': 'mende', 'gsw': 'Mende-Schpraach', 'gu': 'મેન્ડે', 'ha': 'Mende', 'he': 'מנדה', 'hi': 'मेन्डे', 'hi-Latn': 'Mende', 'hr': 'mende', 'hsb': 'mendišćina', 'ht': 'mendé', 'hu': 'mende', 'hy': 'մենդե', 'ia': 'mende', 'id': 'Mende', 'ig': 'Mende', 'is': 'mende', 'it': 'mende', 'ja': 'メンデ語', 'jv': 'Mende', 'ka': 'მენდე', 'kgp': 'mẽne', 'kk': 'менде тілі', 'kk-Cyrl': 'менде тілі', 'km': 'មេនឌី', 'kn': 'ಮೆಂಡೆ', 'ko': '멘데어', 'kok': 'मेंडे', 'kok-Deva': 'मेंडे', 'ks': 'میندیے', 'ks-Arab': 'میندیے', 'ku': 'mende', 'ky': 'мендече', 'lb': 'Mende-Sprooch', 'lo': 'ເມນເດ', 'lt': 'mende', 'lv': 'mendu', 'mai': 'मेंडे', 'mi': 'Menēte', 'mk': 'менде', 'ml': 'മെൻഡെ', 'mn': 'менде', 'mr': 'मेन्डे', 'ms': 'Mende', 'mt': 'Mende', 'my': 'မန်ဒဲ', 'nb': 'mende', 'ne': 'मेन्डे', 'nl': 'Mende', 'nn': 'mende', 'no': 'mende', 'or': 'ମେନଡେ', 'pa': 'ਮੇਂਡੇ', 'pa-Guru': 'ਮੇਂਡੇ', 'pcm': 'Mẹndẹ́ Lángwej', 'pl': 'mende', 'ps': 'مینڊي', 'pt': 'mende', 'qu': 'Mende Simi', 'rm': 'mende', 'ro': 'mende', 'ru': 'менде', 'sc': 'mende', 'sd': 'مينڊي', 'sd-Arab': 'مينڊي', 'si': 'මෙන්ඩෙ', 'sk': 'mendejčina', 'sl': 'mende', 'smn': 'mendekielâ', 'so': 'Mende', 'sq': 'mendisht', 'sr': 'менде', 'sr-Cyrl': 'менде', 'sr-Latn': 'mende', 'sv': 'mende', 'sw': 'Kimende', 'ta': 'மென்டீ', 'te': 'మెండే', 'tg': 'менде', 'th': 'เมนเด', 'ti': 'መንዴ', 'tk': 'mende dili', 'to': 'lea fakamenetī', 'tr': 'Mende dili', 'tt': 'менде', 'ug': 'مېندېچە', 'uk': 'менде', 'ur': 'میندے', 'uz': 'mende', 'uz-Cyrl': 'менде', 'uz-Latn': 'mende', 'vec': 'mende', 'vi': 'Tiếng Mende', 'wo': 'Mende', 'xnr': 'मेन्डे', 'yo': 'Èdè Mende', 'yrl': 'mẽde', 'yue': '門德文', 'yue-Hans': '门德文', 'yue-Hant': '門德文', 'zh': '门德语', 'zh-Hans': '门德语', 'zh-Hant': '門德文', 'zu': 'isi-Mende'},
    'meo': {'en': 'Kedah Malay'},
    'mep': {'en': 'Miriwoong'},
    'meq': {'en': 'Merey'},
    'mer': {'af': 'Meru', 'am': 'ሜሩ', 'ar': 'الميرو', 'as': 'মেৰু', 'ast': 'meru', 'az': 'meru', 'az-Cyrl': 'меру', 'az-Latn': 'meru', 'bal-Latn': 'Méru', 'be': 'меру', 'be-tarask': 'меру', 'bg': 'меру', 'bgn': 'مرویی', 'bn': 'মেরু', 'br': 'meru', 'brx': 'मेरु', 'bs': 'meru', 'bs-Cyrl': 'меру', 'bs-Latn': 'meru', 'ca': 'meru', 'ca-ES-valencia': 'meru', 'ccp': '𑄟𑄬𑄢𑄪', 'ce': 'меру', 'chr': 'ᎺᎷ', 'ckb': 'مێروو', 'cs': 'meru', 'cy': 'Mêrw', 'da': 'meru', 'de': 'Meru', 'dsb': 'meru', 'el': 'Μέρου', 'el-polyton': 'Μέρου', 'en': 'Meru', 'es': 'meru', 'et': 'meru', 'eu': 'meruera', 'fa': 'مرویی', 'ff-Adlm': '𞤃𞤫𞤪𞤵𞥅𞤪𞤫', 'fi': 'meru', 'fil': 'Meru', 'fo': 'meru', 'fr': 'meru', 'fy': 'Meru', 'ga': 'Meru', 'gd': 'Meru', 'gl': 'meru', 'gu': 'મેરુ', 'ha': 'Meru', 'he': 'מרו', 'hi': 'मेरु', 'hi-Latn': 'Meru', 'hr': 'meru', 'hsb': 'meru', 'ht': 'meru', 'hu': 'meru', 'hy': 'մերու', 'ia': 'meri', 'id': 'Meru', 'ig': 'Meru', 'is': 'merú', 'it': 'meru', 'ja': 'メル語', 'jv': 'Meru', 'ka': 'მერუ', 'kgp': 'mẽru', 'kk': 'меру тілі', 'kk-Cyrl': 'меру тілі', 'km': 'មេរូ', 'kn': 'ಮೆರು', 'ko': '메루어', 'kok': 'मेरू', 'kok-Deva': 'मेरू', 'ku': 'meruyî', 'ky': 'меруча', 'lb': 'Meru-Sprooch', 'lo': 'ເມຣູ', 'lrc': 'مئرو', 'lt': 'merų', 'lv': 'meru', 'mai': 'मेरू', 'mer': 'Kĩmĩrũ', 'mi': 'Meru', 'mk': 'меру', 'ml': 'മേരു', 'mn': 'меру', 'mr': 'मेरू', 'ms': 'Meru', 'mt': 'Meru', 'my': 'မီရု', 'mzn': 'مِرویی', 'nb': 'meru', 'ne': 'मेरू', 'nl': 'Meru', 'nn': 'meru', 'no': 'meru', 'nqo': 'ߡߋߙߎߞߊ߲', 'or': 'ମେରୁ', 'pa': 'ਮੇਰੂ', 'pa-Guru': 'ਮੇਰੂ', 'pcm': 'Mẹ́rú Lángwej', 'pl': 'meru', 'ps': 'ميرو', 'pt': 'meru', 'qu': 'Meru Simi', 'rm': 'meru', 'ro': 'meru', 'ru': 'меру', 'sc': 'meru', 'sd': 'ميرو', 'sd-Arab': 'ميرو', 'si': 'මෙරු', 'sk': 'meru', 'sl': 'meru', 'smn': 'merukielâ', 'so': 'Meeru', 'sq': 'meruisht', 'sr': 'меру', 'sr-Cyrl': 'меру', 'sr-Latn': 'meru', 'sv': 'meru', 'sw': 'Kimeru', 'ta': 'மெரு', 'te': 'మెరు', 'th': 'เมรู', 'ti': 'መሩ', 'tk': 'meru dili', 'to': 'lea fakamelu', 'tr': 'Meru', 'ug': 'مېرۇچە', 'uk': 'меру', 'ur': 'میرو', 'uz': 'meru', 'uz-Cyrl': 'меруча', 'uz-Latn': 'meru', 'vec': 'meru', 'vi': 'Tiếng Meru', 'xnr': 'मेरु', 'yo': 'Mérù', 'yrl': 'meru', 'yue': '梅魯文', 'yue-Hans': '梅鲁文', 'yue-Hant': '梅魯文', 'zh': '梅鲁语', 'zh-Hans': '梅鲁语', 'zh-Hant': '梅魯文', 'zu': 'isi-Meru'},
    'mes': {'en': 'Masmaje'},
    'met': {'en': 'Mato'},
    'meu': {'en': 'Motu'},
    'mev': {'en': 'Mano'},
    'mew': {'en': 'Maaka'},
    'mey': {'en': 'Hassaniyya'},
    'mez': {'en': 'Menominee'},
    'mfa': {'en': 'Pattani Malay'},
    'mfb': {'en': 'Bangka'},
    'mfc': {'en': 'Mba'},
    'mfd': {'en': 'Mendankwe-Nkwen'},
    'mfe': {'af': 'Morisjen', 'am': 'ሞሪስየን', 'ar': 'المورسيانية', 'as': 'মৰিছিয়ান', 'ast': 'morisyen', 'az': 'morisien', 'az-Cyrl': 'морисиен', 'az-Latn': 'morisien', 'bal-Latn': 'Murisén', 'be': 'марысьен', 'be-tarask': 'марысьен', 'bg': 'морисиен', 'bgn': 'موریسینی', 'bn': 'মরিসিয়েন', 'br': 'moriseg', 'brx': 'मरिछ्येन', 'bs': 'mauricijski kreolski', 'bs-Cyrl': 'мауритански', 'bs-Latn': 'mauricijski kreolski', 'ca': 'mauricià', 'ca-ES-valencia': 'mauricià', 'ccp': '𑄟𑄢𑄨𑄥𑄨𑄠𑄚𑄴', 'ce': 'маврикин креолийн', 'chr': 'ᎼᎵᏏᎡᏂ', 'ckb': 'مۆریسی', 'cs': 'mauricijská kreolština', 'cy': 'Morisyen', 'da': 'morisyen', 'de': 'Morisyen', 'dsb': 'mauriciska kreolšćina', 'el': 'Μορισιέν', 'el-polyton': 'Μορισιέν', 'en': 'Morisyen', 'es': 'criollo mauriciano', 'et': 'Mauritiuse kreoolkeel', 'eu': 'Mauritaniako kreolera', 'fa': 'موریسین', 'ff-Adlm': '𞤃𞤮𞤪𞤭𞥅𞤧𞤭𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'morisyen', 'fil': 'Morisyen', 'fo': 'morisyen', 'fr': 'créole mauricien', 'fy': 'Morisyen', 'ga': 'Morisyen', 'gd': 'Morisyen', 'gl': 'crioulo mauriciano', 'gu': 'મોરીસ્યેન', 'ha': 'Morisyen', 'he': 'קריאולית מאוריציאנית', 'hi': 'मोरीस्येन', 'hi-Latn': 'Morisyen', 'hr': 'mauricijski kreolski', 'hsb': 'mauriciska kreolšćina', 'ht': 'créole mauricien', 'hu': 'mauritiusi kreol', 'hy': 'մորիսյեն', 'ia': 'creolo mauritian', 'id': 'Morisien', 'ig': 'Morisyen', 'is': 'máritíska', 'it': 'creolo mauriziano', 'ja': 'モーリシャス・クレオール語', 'jv': 'Morisyen', 'ka': 'მორისიენი', 'kgp': 'mãrisijẽn', 'kk': 'морисиен тілі', 'kk-Cyrl': 'морисиен тілі', 'km': 'ម៉ូរីស៊ីន', 'kn': 'ಮೊರಿಸನ್', 'ko': '모리스얀어', 'kok': 'मोरिसेन', 'kok-Deva': 'मोरिसेन', 'ku': 'morisyenî', 'ky': 'морисианча', 'lb': 'Morisyen', 'lo': 'ມໍຣິສເຢນ', 'lrc': 'موٙریسی', 'lt': 'morisijų', 'lv': 'Maurīcijas kreolu', 'mai': 'मोरीसेन', 'mfe': 'kreol morisien', 'mi': 'Morihiene', 'mk': 'морисјен', 'ml': 'മൊറിസിൻ', 'mn': 'морисен', 'mr': 'मोरिस्येन', 'ms': 'Morisyen', 'mt': 'Morisyen', 'my': 'မောရစ်ရှ', 'mzn': 'موریسین', 'nb': 'mauritisk-kreolsk', 'ne': 'मोरिसेन', 'nl': 'Morisyen', 'nn': 'morisyen', 'no': 'mauritisk-kreolsk', 'nqo': 'ߡߏߙߌߛ ߕߊ߬ߓߎ߰ߛߌ߬ߞߊ߲', 'or': 'ମୋରିସୟେନ୍', 'pa': 'ਮੋਰੀਸਿਅਨ', 'pa-Guru': 'ਮੋਰੀਸਿਅਨ', 'pcm': 'Morísiẹ́n Lángwej', 'pl': 'kreolski Mauritiusa', 'ps': 'ماریسیسن', 'pt': 'morisyen', 'qu': 'Mauriciano Simi', 'rm': 'morisyen', 'ro': 'morisyen', 'ru': 'маврикийский креольский', 'sc': 'crèolu mauritzianu', 'sd': 'موریسیین', 'sd-Arab': 'موریسیین', 'si': 'මොරිස්යෙම්', 'sk': 'maurícijská kreolčina', 'sl': 'morisjenščina', 'smn': 'morisyen', 'so': 'Moorisayn', 'sq': 'morisjenisht', 'sr': 'морисјен', 'sr-Cyrl': 'морисјен', 'sr-Latn': 'morisjen', 'sv': 'mauritansk kreol', 'sw': 'Kimoriseni', 'ta': 'மொரிசியன்', 'te': 'మొరిస్యేన్', 'th': 'มอริสเยน', 'ti': 'ክርዮል ማውሪሽይና', 'tk': 'morisýen dili', 'to': 'lea fakamolisieni', 'tr': 'Morisyen', 'ug': 'مورىسيېنچە', 'uk': 'маврикійська креольська', 'ur': 'موریسیین', 'uz': 'morisyen', 'uz-Cyrl': 'морисьен', 'uz-Latn': 'morisyen', 'vec': 'maurisian', 'vi': 'Tiếng Morisyen', 'xnr': 'मोरीस्येन', 'yo': 'Morisiyen', 'yrl': 'murisiẽ', 'yue': '克里奧文（模里西斯）', 'yue-Hans': '克里奥文（模里西斯）', 'yue-Hant': '克里奧文（模里西斯）', 'zh': '毛里求斯克里奥尔语', 'zh-Hans': '毛里求斯克里奥尔语', 'zh-Hant': '克里奧文（模里西斯）', 'zu': 'isi-Morisyen'},
    'mff': {'en': 'Naki'},
    'mfg': {'en': 'Mogofin'},
    'mfh': {'en': 'Matal'},
    'mfi': {'en': 'Wandala'},
    'mfj': {'en': 'Mefele'},
    'mfk': {'en': 'North Mofu'},
    'mfl': {'en': 'Putai'},
    'mfm': {'en': 'Marghi South'},
    'mfn': {'en': 'Cross River Mbembe'},
    'mfo': {'en': 'Mbe'},
    'mfp': {'en': 'Makassar Malay'},
    'mfq': {'en': 'Moba'},
    'mfr': {'en': 'Marrithiyel'},
    'mfs': {'en': 'Mexican Sign Language'},
    'mft': {'en': 'Mokerang'},
    'mfu': {'en': 'Mbwela'},
    'mfv': {'en': 'Mandjak'},
    'mfw': {'en': 'Mulaha'},
    'mfx': {'en': 'Melo'},
    'mfy': {'en': 'Mayo'},
    'mfz': {'en': 'Mabaan'},
    'mg': {'af': 'Malgassies', 'am': 'ማላጋስይ', 'ar': 'الملغاشي', 'as': 'মালাগাছী', 'ast': 'malgaxe', 'az': 'malaqas', 'az-Cyrl': 'малагас', 'az-Latn': 'malaqas', 'bal-Latn': 'Malagase', 'be': 'малагасійская', 'be-tarask': 'малагасійская', 'bg': 'малгашки', 'bgn': 'مالاگاسی', 'bn': 'মালাগাসি', 'br': 'malgacheg', 'brx': 'मालागासी', 'bs': 'malgaški', 'bs-Cyrl': 'малагасијски', 'bs-Latn': 'malgaški', 'ca': 'malgaix', 'ca-ES-valencia': 'malgaix', 'ccp': '𑄟𑄣𑄉𑄥𑄨', 'ce': 'малагасийн', 'chr': 'ᎹᎳᎦᏏ', 'ckb': 'مالاگاسی', 'cs': 'malgaština', 'cy': 'Malagaseg', 'da': 'malagassisk', 'de': 'Malagasy', 'dsb': 'malgašćina', 'dz': 'མ་ལ་ག་སི་ཁ', 'ee': 'malagasegbe', 'el': 'Μαλγασικά', 'el-polyton': 'Μαλγασικά', 'en': 'Malagasy', 'eo': 'malagasa', 'es': 'malgache', 'et': 'malagassi', 'eu': 'malgaxea', 'fa': 'مالاگاسی', 'ff-Adlm': '𞤃𞤢𞤤𞤢𞤺𞤢𞤧𞤭𞥅𞤪𞤫', 'fi': 'malagassi', 'fil': 'Malagasy', 'fo': 'malagassiskt', 'fr': 'malgache', 'fur': 'malagasy', 'fy': 'Malagasysk', 'ga': 'Malagáisis', 'gd': 'Malagasais', 'gl': 'malgaxe', 'gsw': 'Madagassisch', 'gu': 'મલાગસી', 'ha': 'Malagasi', 'he': 'מלגשית', 'hi': 'मालागासी', 'hi-Latn': 'Malagasy', 'hr': 'malgaški', 'hsb': 'malagassišćina', 'ht': 'malgache', 'hu': 'malgas', 'hy': 'մալգաշերեն', 'ia': 'malgache', 'id': 'Malagasi', 'ig': 'Malagasy', 'is': 'malagasíska', 'it': 'malgascio', 'ja': 'マダガスカル語', 'jv': 'Malagasi', 'ka': 'მალაგასიური', 'kea': 'malgaxi', 'kgp': 'mỹrgase', 'kk': 'малагаси тілі', 'kk-Cyrl': 'малагаси тілі', 'km': 'ម៉ាឡាហ្គាស៊ី', 'kn': 'ಮಲಗಾಸಿ', 'ko': '말라가시어', 'kok': 'मलागसी', 'kok-Deva': 'मलागसी', 'ks': 'مَلاگَسی', 'ks-Arab': 'مَلاگَسی', 'ksh': 'Madajaßkesch', 'ku': 'malagasî', 'ky': 'малагасча', 'lb': 'Malagassi-Sprooch', 'lkt': 'Malagasy Iyápi', 'lld': 'malgasich', 'lo': 'ມາລາກາສຊີ', 'lrc': 'مالاگاشی', 'lt': 'malagasų', 'lv': 'malagasu', 'mai': 'मलगासी', 'mg': 'Malagasy', 'mi': 'Marakāhi', 'mk': 'малгашки', 'ml': 'മലഗാസി', 'mn': 'малагаси', 'mr': 'मलागसी', 'ms': 'Malagasy', 'mt': 'Malagasy', 'my': 'မာလဂက်စီ', 'mzn': 'مالاگاسی', 'nb': 'gassisk', 'ne': 'मलागासी', 'nl': 'Malagassisch', 'nn': 'madagassisk', 'no': 'gassisk', 'nqo': 'ߡߊߟߑߜ߭ߊߛߌߞߊ߲', 'or': 'ମାଲାଗାସୀ', 'pa': 'ਮਾਲਾਗੈਸੀ', 'pa-Guru': 'ਮਾਲਾਗੈਸੀ', 'pcm': 'Malagásí Lángwej', 'pl': 'malgaski', 'ps': 'ملغاسي', 'pt': 'malgaxe', 'qu': 'Malgache Simi', 'rm': 'malagassi', 'ro': 'malgașă', 'ru': 'малагасийский', 'sc': 'malgàsciu', 'sd': 'ملاگاسي', 'sd-Arab': 'ملاگاسي', 'si': 'මලගාසි', 'sk': 'malgaština', 'sl': 'malgaščina', 'smn': 'malagaskielâ', 'so': 'Malagaasi', 'sq': 'madagaskarisht', 'sr': 'малгашки', 'sr-Cyrl': 'малгашки', 'sr-Latn': 'malgaški', 'sv': 'malagassiska', 'sw': 'Kimalagasi', 'ta': 'மலகாஸி', 'te': 'మలగాసి', 'tg': 'малагасӣ', 'th': 'มาลากาซี', 'ti': 'ማላጋሲ', 'tk': 'malagasiý dili', 'to': 'lea fakamalakasi', 'tr': 'Malgaşça', 'tt': 'малагаси', 'ug': 'مالاگاسچە', 'uk': 'малагасійська', 'ur': 'ملاگاسی', 'uz': 'malagasiy', 'uz-Cyrl': 'малагасийча', 'uz-Latn': 'malagasiy', 'vec': 'malgaso', 'vi': 'Tiếng Malagasy', 'wae': 'Malagási', 'wo': 'Malagasi', 'xnr': 'मालागासी', 'yo': 'Malagasì', 'yrl': 'maugaxe', 'yue': '馬拉加什文', 'yue-Hans': '马拉加什文', 'yue-Hant': '馬拉加什文', 'zh': '马拉加斯语', 'zh-Hans': '马拉加斯语', 'zh-Hant': '馬達加斯加文', 'zu': 'isi-Malagasy'},
    'mga': {'ar': 'الأيرلندية الوسطى', 'ast': 'írlandés mediu', 'az': 'orta irland', 'az-Latn': 'orta irland', 'bg': 'средновековен ирландски', 'bn': 'মধ্য আইরিশ', 'br': 'krenniwerzhoneg', 'brx': 'मध्य आईरीश भाषा', 'bs': 'srednjovjekovni irski', 'bs-Cyrl': 'средњи ирски', 'bs-Latn': 'srednjovjekovni irski', 'ca': 'gaèlic irlandès mitjà', 'ca-ES-valencia': 'gaèlic irlandès mitjà', 'ccp': '𑄟𑄧𑄖𑄴𑄙𑄳𑄠 𑄃𑄭𑄢𑄨𑄌𑄴', 'cs': 'irština (středověká)', 'cy': 'Gwyddeleg Canol', 'da': 'middelirsk', 'de': 'Mittelirisch', 'el': 'Μέσα Ιρλανδικά', 'el-polyton': 'Μέσα Ιρλανδικά', 'en': 'Middle Irish', 'en-Dsrt': '𐐣𐐮𐐼𐐲𐑊 𐐌𐑉𐐮𐑇', 'es': 'irlandés medio', 'et': 'keskiiri', 'fa': 'ایرلندی میانه', 'fi': 'keski-iiri', 'fr': 'moyen irlandais', 'fy': 'Middeliers', 'ga': 'Meán-Ghaeilge', 'gd': 'Meadhan-Ghaeilge', 'gsw': 'Mittelirisch', 'gu': 'મધ્ય આઈરિશ', 'he': 'אירית תיכונה', 'hi': 'मध्यकालीन आइरिश', 'hi-Latn': 'Middle Irish', 'hr': 'srednjoirski', 'ht': 'moyen irlandais', 'hu': 'közép ír', 'id': 'Irlandia Abad Pertengahan', 'is': 'miðírska', 'it': 'irlandese medio', 'ja': '中期アイルランド語', 'ka': 'საშუალო ირლანდიური', 'kgp': 'irỹnej kuju', 'kn': 'ಮಧ್ಯ ಐರಿಷ್', 'ko': '중세 아일랜드어', 'ks': 'وَستی ایرِش', 'ks-Arab': 'وَستی ایرِش', 'lb': 'Mëtteliresch', 'lo': 'ໄອລິດກາງ', 'lt': 'Vidurio Airijos', 'lv': 'vidusīru', 'mk': 'средноирски', 'ml': 'മദ്ധ്യ ഐറിഷ്', 'mr': 'मिडल आयरिश', 'mt': 'Irlandiż Medjevali', 'my': 'အလယ်ပိုင်း အိုင်းရစ်ရှ်', 'nb': 'mellomirsk', 'ne': 'मध्य आयरिस', 'nl': 'Middeliers', 'nn': 'mellomirsk', 'no': 'mellomirsk', 'or': 'ମଧ୍ୟ ଇରିଶ୍', 'pl': 'średnioirlandzki', 'pt': 'irlandês médio', 'rm': 'irlandais mesaun', 'ro': 'irlandeză medie', 'ru': 'среднеирландский', 'sk': 'stredná írčina', 'sl': 'srednja irščina', 'sr': 'средњеирски', 'sr-Cyrl': 'средњеирски', 'sr-Latn': 'srednjeirski', 'sv': 'medeliriska', 'ta': 'மிடில் ஐரிஷ்', 'te': 'మధ్యమ ఐరిష్', 'th': 'ไอริชกลาง', 'to': 'lea fakaʻaelani-lotoloto', 'tr': 'Ortaçağ İrlandacası', 'ug': 'ئوتتۇرا ئەسىر ئىرېلاندچە', 'uk': 'середньоірландська', 'vi': 'Tiếng Ai-len Trung cổ', 'yrl': 'irãdei médiu', 'yue': '中古愛爾蘭文', 'yue-Hans': '中古爱尔兰文', 'yue-Hant': '中古愛爾蘭文', 'zh': '中古爱尔兰语', 'zh-Hans': '中古爱尔兰语', 'zh-Hant': '中古愛爾蘭文'},
    'mgb': {'en': 'Mararit'},
    'mgc': {'en': 'Morokodo'},
    'mgd': {'en': 'Moru'},
    'mge': {'en': 'Mango'},
    'mgf': {'en': 'Maklew'},
    'mgg': {'en': 'Mpumpong'},
    'mgh': {'af': 'Makhuwa-Meetto', 'am': 'ማኩዋ-ሜቶ', 'ar': 'ماخاوا-ميتو', 'as': 'মাখুৱা-মিট্টো', 'ast': 'makhuwa-meetto', 'az': 'maxuva-meetto', 'az-Cyrl': 'махува-меетто', 'az-Latn': 'maxuva-meetto', 'bal-Latn': 'Makuá-Mitó', 'be': 'макуўа-меета', 'be-tarask': 'макуўа-меета', 'bg': 'макуа мето', 'bgn': 'ماکوا متوی', 'bn': 'মাখুয়া-মেত্তো', 'br': 'makhuwa-meetto', 'brx': 'माखुवा- मीट्ट’', 'bs': 'makuva-meto', 'bs-Cyrl': 'макуа-мето', 'bs-Latn': 'makuva-meto', 'ca': 'makhuwa-metto', 'ca-ES-valencia': 'makhuwa-metto', 'ccp': '𑄟𑄈𑄪𑄠-𑄟𑄬𑄖𑄴𑄖𑄮', 'ce': 'макуа-меетто', 'chr': 'ᎹᎫᏩ-ᎻᏙ', 'ckb': 'ماخوامیتۆ', 'cs': 'makhuwa-meetto', 'cy': 'Makhuwa-Meetto', 'da': 'makhuwa-meetto', 'de': 'Makhuwa-Meetto', 'dsb': 'makhuwa-meetto', 'el': 'Μακούβα-Μέτο', 'el-polyton': 'Μακούβα-Μέτο', 'en': 'Makhuwa-Meetto', 'es': 'makhuwa-meetto', 'et': 'makhuwa-meetto', 'eu': 'makhuwa-meettoera', 'fa': 'ماکوا متو', 'ff-Adlm': '𞤃𞤢𞤳𞤵𞤱𞤢𞥄𞤪𞤫', 'fi': 'makua-meetto', 'fil': 'Makhuwa-Meetto', 'fo': 'makhuwa-meetto', 'fr': 'makua', 'fy': 'Makhuwa-Meetto', 'ga': 'Meiteo-Macuais', 'gd': 'Makhuwa-Meetto', 'gl': 'makhuwa-meetto', 'gu': 'માખુવા-મીટ્ટુ', 'ha': 'Makhuwa-Meetto', 'he': 'מאקוואה מטו', 'hi': 'मैखुवा-मीट्टो', 'hi-Latn': 'Makhuwa-Meetto', 'hr': 'makhuwa-meetto', 'hsb': 'makhuwa-meetto', 'ht': 'makua', 'hu': 'makua-metó', 'hy': 'մաքուա-մետտո', 'ia': 'macua de Metto', 'id': 'Makhuwa-Meetto', 'ig': 'Makhuwa-Meetto', 'is': 'makhuwa-meetto', 'it': 'makhuwa-meetto', 'ja': 'マクア・ミート語', 'jv': 'Makhuwa-Meeto', 'ka': 'მაქუვა-მეეტო', 'kea': 'makua', 'kgp': 'mỹkuva', 'kk': 'макуа-меетто тілі', 'kk-Cyrl': 'макуа-меетто тілі', 'km': 'ម៉ាកគូវ៉ាមីតូ', 'kn': 'ಮ್ಯಖುವಾ- ಮೀಟ್ಟೊ', 'ko': '마크후와-메토어', 'kok': 'माखुवा-मिट्टो', 'kok-Deva': 'माखुवा-मिट्टो', 'ku': 'makhuwa-meetto', 'ky': 'макуача', 'lb': 'Makhuwa-Meetto', 'lo': 'ມາຄູວາ-ມີດໂຕ', 'lrc': 'ماخوڤا میتو', 'lt': 'makua-maeto', 'lv': 'makua-mīto', 'mai': 'मखुआ-मेटो', 'mgh': 'Makua', 'mi': 'Makuwa-Mēto', 'mk': 'макува-мито', 'ml': 'മാഖുവാ-മീത്തോ', 'mn': 'макува-мито', 'mr': 'माखुव्हा-मीट्टो', 'ms': 'Makhuwa-Meetto', 'mt': 'Makhuwa-Meetto', 'my': 'မာခူဝါ-မီအီတို', 'mzn': 'ماخوئا-میتو', 'nb': 'makhuwa-meetto', 'ne': 'माखुवा-मिट्टो', 'nl': 'Makhuwa-Meetto', 'nn': 'makhuwa-meetto', 'no': 'makhuwa-meetto', 'nqo': 'ߡߊߞߎߞߊ߲', 'or': 'ମଖୁୱା-ମେଟ୍ଟା', 'pa': 'ਮਖੋਵਾ-ਮਿੱਟੋ', 'pa-Guru': 'ਮਖੋਵਾ-ਮਿੱਟੋ', 'pcm': 'Makúwá-Mító', 'pl': 'makua-meetto', 'ps': 'مکھوامیتو', 'pt': 'macua-mêto', 'qu': 'Makhuwa-Meetto Simi', 'rm': 'makhuwa-meetto', 'ro': 'makhuwa-meetto', 'ru': 'макуа-меетто', 'sc': 'makhuwa-meetto', 'sd': 'مخووا ميتو', 'sd-Arab': 'مخووا ميتو', 'si': 'මඛුවා-මීටෝ', 'sk': 'makua-meetto', 'sl': 'makuva-meto', 'smn': 'makua-meetto', 'so': 'Luuqadda Makhuwa-Meetto', 'sq': 'makua-mitoisht', 'sr': 'макува-мито', 'sr-Cyrl': 'макува-мито', 'sr-Latn': 'makuva-mito', 'sv': 'makhuwa-meetto', 'sw': 'Kimakhuwa-Meetto', 'ta': 'மகுவா-மீட்டோ', 'te': 'మక్వా-మిట్టో', 'th': 'มากัววา-มีทโท', 'ti': 'ማክዋ-ሜቶ', 'tk': 'makuwa-mito dili', 'to': 'lea fakamakūa-meʻeto', 'tr': 'Makhuwa-Meetto', 'ug': 'ماكۇۋاچە', 'uk': 'макува-меето', 'ur': 'ماخاوا-ميتو', 'uz': 'maxuva-mitto', 'uz-Cyrl': 'махува-митто', 'uz-Latn': 'maxuva-mitto', 'vec': 'meeto', 'vi': 'Tiếng Makhuwa-Meetto', 'xnr': 'मैखुवा-मीट्टो', 'yo': 'Makhuwa-Meeto', 'yrl': 'makua', 'yue': '馬夸文', 'yue-Hans': '马夸文', 'yue-Hant': '馬夸文', 'zh': '马库阿-梅托语', 'zh-Hans': '马库阿-梅托语', 'zh-Hant': '馬夸文', 'zu': 'isi-Makhuwa-Meetto'},
    'mgi': {'en': 'Lijili'},
    'mgj': {'en': 'Abureni'},
    'mgk': {'en': 'Mawes'},
    'mgl': {'en': 'Maleu-Kilenge'},
    'mgm': {'en': 'Mambae'},
    'mgn': {'en': 'Mbangi'},
    'mgo': {'af': 'Meta’', 'am': 'ሜታ', 'ar': 'ميتا', 'as': 'মেটা', 'ast': 'meta’', 'az': 'meta’', 'az-Cyrl': 'метаʼ', 'az-Latn': 'meta’', 'bal-Latn': 'Métá', 'be': 'мета', 'be-tarask': 'мета', 'bg': 'мета', 'bgn': 'میٹایی', 'bn': 'মেটা', 'br': 'metaʼ', 'brx': 'मेता', 'bs': 'meta', 'bs-Cyrl': 'мета', 'bs-Latn': 'meta', 'ca': 'meta’', 'ca-ES-valencia': 'meta’', 'ccp': '𑄟𑄬𑄑', 'ce': 'мета', 'chr': 'ᎺᎳ’', 'ckb': 'مێتە', 'cs': 'meta’', 'cy': 'Meta', 'da': 'meta', 'de': 'Meta’', 'dsb': 'meta’', 'el': 'Μέτα', 'el-polyton': 'Μέτα', 'en': 'Metaʼ', 'es': 'meta’', 'et': 'meta', 'eu': 'metaʼera', 'fa': 'متایی', 'ff-Adlm': '𞤃𞤫𞤼𞤢𞥄𞤪𞤫', 'fi': 'meta’', 'fil': 'Meta’', 'fo': 'metaʼ', 'fr': 'metaʼ', 'fy': 'Meta’', 'ga': 'Metaʼ', 'gd': 'Meta’', 'gl': 'meta’', 'gu': 'મેતા', 'ha': 'Metaʼ', 'he': 'מטא', 'hi': 'मेटा', 'hi-Latn': 'Metaʼ', 'hr': 'meta’', 'hsb': 'meta’', 'ht': 'metaʼ', 'hu': 'meta’', 'hy': 'մետա', 'ia': 'metaʼ', 'id': 'Meta’', 'ig': 'Meta', 'is': 'meta’', 'it': 'meta’', 'ja': 'メタ語', 'jv': 'Meta’', 'ka': 'მეტა-ენა', 'kgp': 'mẽta', 'kk': 'мета тілі', 'kk-Cyrl': 'мета тілі', 'km': 'មេតា', 'kn': 'ಮೆಟಾ', 'ko': '메타어', 'kok': 'मेटा', 'kok-Deva': 'मेटा', 'ku': 'meta’', 'ky': 'метача', 'lb': 'Meta’', 'lo': 'ເມທາ', 'lrc': 'مئتاٛ', 'lt': 'meta', 'lv': 'metu', 'mai': 'मेटा', 'mgo': 'metaʼ', 'mi': 'Meta', 'mk': 'мета', 'ml': 'മേത്താ', 'mn': 'мета', 'mr': 'मीटा', 'ms': 'Meta’', 'mt': 'Metà', 'my': 'မီတာ', 'mzn': 'مِتاء', 'nb': 'meta’', 'ne': 'मेटा', 'nl': 'Meta’', 'nn': 'meta’', 'no': 'meta’', 'nqo': 'ߡߋߕߊߞߊ߲', 'or': 'ମେଟା', 'pa': 'ਮੇਟਾ', 'pa-Guru': 'ਮੇਟਾ', 'pcm': 'Mẹta’ Lángwej', 'pl': 'meta', 'ps': 'ميټا', 'pt': 'meta’', 'qu': 'Metaʼ Simi', 'rm': 'meta’', 'ro': 'meta’', 'ru': 'мета', 'sc': 'meta’', 'sd': 'ميتا', 'sd-Arab': 'ميتا', 'si': 'මෙටා', 'sk': 'meta’', 'sl': 'meta', 'smn': 'meta’', 'so': 'Meetaa', 'sq': 'metaisht', 'sr': 'мета', 'sr-Cyrl': 'мета', 'sr-Latn': 'meta', 'sv': 'meta’', 'sw': 'Kimeta', 'ta': 'மேடா', 'te': 'మెటా', 'th': 'เมตา', 'ti': 'መታ', 'tk': 'meta dili', 'to': 'lea fakametā', 'tr': 'Meta’', 'ug': 'مېتاچە’', 'uk': 'мета', 'ur': 'میٹا', 'uz': 'meta', 'uz-Cyrl': 'мета', 'uz-Latn': 'meta', 'vec': 'meta', 'vi': 'Tiếng Meta’', 'xnr': 'मेटा', 'yo': 'Métà', 'yrl': 'metá', 'yue': '美塔文', 'yue-Hans': '美塔文', 'yue-Hant': '美塔文', 'zh': '梅塔语', 'zh-Hans': '梅塔语', 'zh-Hant': '美塔文', 'zu': 'isi-Meta’'},
    'mgp': {'en': 'Eastern Magar'},
    'mgq': {'en': 'Malila'},
    'mgr': {'en': 'Mambwe-Lungu'},
    'mgs': {'en': 'Manda (Tanzania)'},
    'mgt': {'en': 'Mongol (New Guinea)'},
    'mgu': {'en': 'Mailu'},
    'mgv': {'en': 'Matengo'},
    'mgw': {'en': 'Matumbi'},
    'mgx': {'en': 'Omati'},
    'mgy': {'en': 'Mbunga'},
    'mgz': {'en': 'Mbugwe'},
    'mh': {'af': 'Marshallees', 'am': 'ማርሻሊዝ', 'ar': 'المارشالية', 'as': 'মাৰ্চলিজ', 'ast': 'marshallés', 'az': 'marşal', 'az-Cyrl': 'маршал', 'az-Latn': 'marşal', 'be': 'маршальская', 'be-tarask': 'маршальская', 'bg': 'маршалезе', 'bgn': 'مارشالی', 'bn': 'মার্শালিজ', 'br': 'marshall', 'brx': 'मार्शेलिस', 'bs': 'maršalski', 'bs-Cyrl': 'маршалски', 'bs-Latn': 'maršalski', 'ca': 'marshallès', 'ca-ES-valencia': 'marshallès', 'ccp': '𑄟𑄢𑄴𑄥𑄣𑄨𑄎𑄴', 'ce': 'маршаллийн', 'chr': 'ᎹᏌᎵᏏ', 'ckb': 'مارشاڵی', 'cs': 'maršálština', 'cy': 'Marsialeg', 'da': 'marshallese', 'de': 'Marschallesisch', 'dsb': 'marshallezišćina', 'el': 'Μαρσαλέζικα', 'el-polyton': 'Μαρσαλέζικα', 'en': 'Marshallese', 'es': 'marshalés', 'et': 'maršalli', 'eu': 'marshallera', 'fa': 'مارشالی', 'ff-Adlm': '𞤃𞤢𞤪𞤧𞤢𞤤𞤫𞥅𞤪𞤫', 'fi': 'marshall', 'fil': 'Marshallese', 'fo': 'marshallesiskt', 'fr': 'marshallais', 'fy': 'Marshallees', 'ga': 'Mairsillis', 'gd': 'Marshallais', 'gl': 'marshalés', 'gsw': 'Marschallesisch', 'gu': 'માર્શલીઝ', 'ha': 'Marshallese', 'he': 'מרשלית', 'hi': 'मार्शलीज़', 'hi-Latn': 'Marshallese', 'hr': 'maršalski', 'hsb': 'marshallezišćina', 'ht': 'marshallais', 'hu': 'marshalli', 'hy': 'մարշալերեն', 'ia': 'marshallese', 'id': 'Marshall', 'ig': 'Marshallese', 'is': 'marshallska', 'it': 'marshallese', 'ja': 'マーシャル語', 'jv': 'Marshallese', 'ka': 'მარშალური', 'kgp': 'mỹrsarej', 'kk': 'маршалл тілі', 'kk-Cyrl': 'маршалл тілі', 'km': 'ម៉ាស់សល', 'kn': 'ಮಾರ್ಶಲ್ಲೀಸ್', 'ko': '마셜어', 'kok': 'मार्शलीज', 'kok-Deva': 'मार्शलीज', 'ks': 'مارشَلیٖز', 'ks-Arab': 'مارشَلیٖز', 'ksh': 'Maschallehsesch', 'ku': 'marşalî', 'ky': 'маршаллча', 'lb': 'Marschallesesch', 'lo': 'ມາຊານເລັດ', 'lt': 'Maršalo Salų', 'lv': 'māršaliešu', 'mai': 'मार्शेलीज', 'mi': 'Mararīhi', 'mk': 'маршалски', 'ml': 'മാർഷല്ലീസ്', 'mn': 'маршалл', 'mr': 'मार्शलीज', 'ms': 'Marshall', 'mt': 'Marshalljaniż', 'my': 'မာရှယ်လိဇ်', 'nb': 'marshallesisk', 'ne': 'मार्साली', 'nl': 'Marshallees', 'nn': 'marshallesisk', 'no': 'marshallesisk', 'or': 'ମାର୍ଶାଲୀଜ୍', 'pa': 'ਮਾਰਸ਼ਲੀਜ਼', 'pa-Guru': 'ਮਾਰਸ਼ਲੀਜ਼', 'pcm': 'Máshállís Lángwej', 'pl': 'marszalski', 'ps': 'مارشلیز', 'pt': 'marshalês', 'qu': 'Marshallese Simi', 'rm': 'marschallais', 'ro': 'marshalleză', 'ru': 'маршалльский', 'sc': 'marshallesu', 'sd': 'مارشليز', 'sd-Arab': 'مارشليز', 'si': 'මාශලීස්', 'sk': 'marshallčina', 'sl': 'marshallovščina', 'smn': 'marshallkielâ', 'so': 'Marshallese', 'sq': 'marshallisht', 'sr': 'маршалски', 'sr-Cyrl': 'маршалски', 'sr-Latn': 'maršalski', 'sv': 'marshalliska', 'sw': 'Kimashale', 'ta': 'மார்ஷெலீஸ்', 'te': 'మార్షలీస్', 'th': 'มาร์แชลล์', 'ti': 'ማርሻሊዝኛ', 'tk': 'marşall dili', 'to': 'lea fakamāsolo', 'tr': 'Marshall Adaları dili', 'ug': 'مارشالچە', 'uk': 'маршалльська', 'ur': 'مارشلیز', 'uz': 'marshall', 'uz-Cyrl': 'маршалл тили', 'uz-Latn': 'marshall', 'vec': 'marshaleze', 'vi': 'Tiếng Marshall', 'xnr': 'मार्शलीज़', 'yo': 'Èdè Mashali', 'yrl': 'marixarei', 'yue': '馬紹爾文', 'yue-Hans': '马绍尔文', 'yue-Hant': '馬紹爾文', 'zh': '马绍尔语', 'zh-Hans': '马绍尔语', 'zh-Hant': '馬紹爾文', 'zu': 'isi-Marshallese'},
    'mha': {'en': 'Manda (India)'},
    'mhb': {'en': 'Mahongwe'},
    'mhc': {'en': 'Mocho'},
    'mhd': {'en': 'Mbugu'},
    'mhe': {'en': 'Besisi'},
    'mhf': {'en': 'Mamaa'},
    'mhg': {'en': 'Margu'},
    'mhh': {'en': 'Maskoy Pidgin'},
    'mhi': {'en': "Ma'di"},
    'mhj': {'en': 'Mogholi'},
    'mhk': {'en': 'Mungaka'},
    'mhl': {'en': 'Mauwake'},
    'mhm': {'en': 'Makhuwa-Moniga'},
    'mhn': {'bal-Latn': 'Móchénó', 'et': 'mohheni', 'gd': 'Mócheno', 'ia': 'mocheno', 'ig': 'mhnn', 'vec': 'mòcheno', 'en': 'Mócheno'},
    'mho': {'en': 'Mashi (Zambia)'},
    'mhp': {'en': 'Balinese Malay'},
    'mhq': {'en': 'Mandan'},
    'mhr': {'en': 'Eastern Mari'},
    'mhs': {'en': 'Buru (Indonesia)'},
    'mht': {'en': 'Mandahuaca'},
    'mhu': {'en': 'Digaro-Mishmi'},
    'mhw': {'en': 'Mbukushu'},
    'mhx': {'en': 'Maru'},
    'mhy': {'en': "Ma'anyan"},
    'mhz': {'en': 'Mor (Mor Islands)'},
    'mi': {'af': 'Maori', 'ak': 'Mawori', 'am': 'ማኦሪ', 'ar': 'الماورية', 'as': 'মাওৰি', 'ast': 'maorí', 'az': 'māori', 'az-Cyrl': 'маори', 'az-Latn': 'māori', 'bal-Latn': 'Muri', 'be': 'маары', 'be-tarask': 'маары', 'bg': 'маорски', 'bgn': 'مائوری', 'bn': 'মাওরি', 'br': 'maori', 'brx': 'माऔरि', 'bs': 'maorski', 'bs-Cyrl': 'маорски', 'bs-Latn': 'maorski', 'ca': 'maori', 'ca-ES-valencia': 'maori', 'ccp': '𑄟𑄃𑄮𑄢𑄨', 'ce': 'маори', 'chr': 'ᎹᏫ', 'ckb': 'مائۆری', 'cs': 'maorština', 'cy': 'Māori', 'da': 'maori', 'de': 'Māori', 'dsb': 'maorišćina', 'dz': 'མ་ཨོ་རི་ཁ', 'ee': 'maorgbe', 'el': 'Μαορί', 'el-polyton': 'Μαορί', 'en': 'Māori', 'en-Dsrt': '𐐣𐐵𐑉𐐨', 'eo': 'maoria', 'es': 'maorí', 'et': 'maoori', 'eu': 'maoriera', 'fa': 'مائوری', 'ff-Adlm': '𞤃𞤢𞥄𞤮𞤪𞤭𞥅𞤪𞤫', 'fi': 'maori', 'fil': 'Māori', 'fo': 'maori', 'fr': 'maori', 'fur': 'maori', 'fy': 'Maori', 'ga': 'Maorais', 'gd': 'Māori', 'gl': 'maorí', 'gsw': 'Maori', 'gu': 'માઓરી', 'ha': 'Maori', 'haw': 'Māori', 'he': 'מאורית', 'hi': 'माओरी', 'hi-Latn': 'Māori', 'hr': 'maorski', 'hsb': 'maoršćina', 'ht': 'maori', 'hu': 'maori', 'hy': 'մաորի', 'ia': 'maori', 'id': 'Maori', 'ig': 'Māori', 'is': 'maorí', 'it': 'maori', 'ja': 'マオリ語', 'jv': 'Maori', 'ka': 'მაორი', 'kea': 'maori', 'kgp': 'mỹvóri', 'kk': 'маори тілі', 'kk-Cyrl': 'маори тілі', 'km': 'ម៉ោរី', 'kn': 'ಮಾವೋರಿ', 'ko': '마오리어', 'kok': 'माओरी', 'kok-Deva': 'माओरी', 'ks': 'ماوری', 'ks-Arab': 'ماوری', 'ksh': 'Ma’ohresch', 'ku': 'maorî', 'ky': 'маориче', 'lb': 'Maori', 'lkt': 'Maori Iyápi', 'lld': 'maori', 'lo': 'ມາວຣິ', 'lrc': 'مائوری', 'lt': 'maorių', 'lv': 'maoru', 'mai': 'माओरी', 'mi': 'Māori', 'mk': 'маорски', 'ml': 'മവോറി', 'mn': 'маори', 'mr': 'माओरी', 'ms': 'Maori', 'mt': 'Maori', 'my': 'မာအိုရီ', 'mzn': 'مائوری', 'nb': 'maori', 'ne': 'माओरी', 'nl': 'Maori', 'nn': 'maori', 'no': 'maori', 'nqo': 'ߡߊߏߙߌߞߊ߲', 'or': 'ମାଓରୀ', 'pa': 'ਮਾਉਰੀ', 'pa-Guru': 'ਮਾਉਰੀ', 'pcm': 'Maórí Lángwej', 'pl': 'maoryjski', 'ps': 'ماوري', 'pt': 'maori', 'qu': 'Maori Simi', 'rm': 'maori', 'ro': 'maori', 'ru': 'маори', 'sc': 'maori', 'sd': 'مائوري', 'sd-Arab': 'مائوري', 'se': 'maorigiella', 'si': 'මාවොරි', 'sk': 'maorijčina', 'sl': 'maorščina', 'smn': 'maorikielâ', 'so': 'Maaoori', 'sq': 'maorisht', 'sr': 'маорски', 'sr-Cyrl': 'маорски', 'sr-Latn': 'maorski', 'sv': 'maori', 'sw': 'Kimaori', 'ta': 'மௌரி', 'te': 'మావొరీ', 'tg': 'маорӣ', 'th': 'เมารี', 'ti': 'ማኦሪ', 'tk': 'maori dili', 'to': 'lea fakamāuli', 'tr': 'Maori dili', 'tt': 'маори', 'ug': 'ماۋرىچە', 'uk': 'маорі', 'ur': 'ماؤری', 'uz': 'maori', 'uz-Cyrl': 'маори', 'uz-Latn': 'maori', 'vec': 'maori', 'vi': 'Tiếng Māori', 'wae': 'Maori', 'wo': 'Mawri', 'xnr': 'माओरी', 'yi': 'מאַאריש', 'yo': 'Màórì', 'yrl': 'mauri', 'yue': '毛利文', 'yue-Hans': '毛利文', 'yue-Hant': '毛利文', 'zh': '毛利语', 'zh-Hans': '毛利语', 'zh-Hant': '毛利文', 'zu': 'isi-Maori'},
    'mia': {'en': 'Miami'},
    'mib': {'en': 'Atatláhuca Mixtec'},
    'mic': {'af': 'Micmac', 'am': 'ሚክማው', 'ar': 'الميكماكيونية', 'as': 'মিকমেক', 'ast': 'micmac', 'az': 'mikmak', 'az-Cyrl': 'микмак', 'az-Latn': 'mikmak', 'bal-Latn': 'Mikmaw', 'be': 'мікмак', 'be-tarask': 'мікмак', 'bg': 'микмак', 'bgn': 'میکماکی', 'bn': 'মিকম্যাক', 'br': 'mikmakeg', 'brx': 'मिईकमाक', 'bs': 'mikmak', 'bs-Cyrl': 'микмак', 'bs-Latn': 'mikmak', 'ca': 'micmac', 'ca-ES-valencia': 'micmac', 'ccp': '𑄟𑄨𑄇𑄟𑄳𑄠𑄇𑄴', 'ce': 'микмак', 'chr': 'ᎻᎧᎹᎩ', 'ckb': 'میکماک', 'cs': 'micmac', 'cy': 'Micmaceg', 'da': 'micmac', 'de': 'Micmac', 'dsb': "mi'kmaqšćina", 'el': 'Μικμάκ', 'el-polyton': 'Μικμάκ', 'en': "Mi'kmaw", 'es': 'micmac', 'et': 'mikmaki', 'eu': 'mikmakera', 'fa': 'میکماکی', 'ff-Adlm': '𞤃𞤭𞤳𞤥𞤢𞤹𞤵𞥅𞤪𞤫', 'fi': 'micmac', 'fil': 'Micmac', 'fo': 'micmac', 'fr': 'micmac', 'fy': 'Mi’kmaq', 'ga': 'Micmeaicis', 'gd': 'Mi’kmaq', 'gl': 'micmac', 'gsw': 'Micmac-Schpraach', 'gu': 'મિકમેક', 'ha': "Mi'kmaq", 'he': 'מיקמק', 'hi': 'मिकमैक', 'hi-Latn': "Mi'kmaq", 'hr': 'micmac', 'hsb': "mi'kmaqšćina", 'ht': 'micmac', 'hu': 'mikmak', 'hy': 'միկմակ', 'ia': 'micmac', 'id': 'Mikmak', 'ig': "Mi'kmaw", 'is': 'mikmak', 'it': 'micmac', 'ja': 'ミクマク語', 'jv': "Mi'kmak", 'ka': 'მიკმაკი', 'kgp': 'mĩkemỹke', 'kk': 'микмак тілі', 'kk-Cyrl': 'микмак тілі', 'km': 'មិកមេក', 'kn': 'ಮಿಕ್\u200cಮ್ಯಾಕ್', 'ko': '미크맥어', 'kok': 'मिक्माक', 'kok-Deva': 'मिक्माक', 'ks': 'مِکمیک', 'ks-Arab': 'مِکمیک', 'ku': 'mîkmakî', 'ky': 'микмакча', 'lb': 'Micmac-Sprooch', 'lo': 'ມິກແມກ', 'lt': 'mikmakų', 'lv': 'mikmaku', 'mai': 'मिकमक', 'mi': 'Mīkamā', 'mic': 'Lʼnuiʼsuti', 'mk': 'микмак', 'ml': 'മിക്മാക്', 'mn': 'микмак', 'mr': 'मिकमॅक', 'ms': 'Micmac', 'mt': 'Micmac', 'my': 'မစ်ခ်မက်ခ်', 'nb': 'micmac', 'ne': 'मिकमाक', 'nl': 'Mi’kmaq', 'nn': 'micmac', 'no': 'micmac', 'or': 'ମିକମୌକ୍', 'pa': 'ਮਾਇਮੈਕ', 'pa-Guru': 'ਮਾਇਮੈਕ', 'pcm': 'Mikmák Lángwej', 'pl': 'mikmak', 'ps': 'ممکق', 'pt': 'miquemaque', 'qu': "Mi'kmaq Simi", 'rm': 'micmac', 'ro': 'micmac', 'ru': 'микмак', 'sc': 'micmac', 'sd': 'ميڪ مڪ', 'sd-Arab': 'ميڪ مڪ', 'si': 'මික්මැක්', 'sk': 'mikmakčina', 'sl': 'mikmaščina', 'smn': 'micmac', 'so': "Mi'kmaq", 'sq': 'mikmakisht', 'sr': 'микмак', 'sr-Cyrl': 'микмак', 'sr-Latn': 'mikmak', 'sv': 'mi’kmaq', 'sw': 'Mi’kmaq', 'ta': 'மிக்மாக்', 'te': 'మికమాక్', 'th': 'มิกแมก', 'ti': 'ሚክማክ', 'tk': 'mikmak dili', 'to': 'lea fakamikemaki', 'tr': 'Micmac', 'ug': 'مىكماكچە', 'uk': 'мікмак', 'ur': 'مکمیک', 'uz': 'mikmak', 'uz-Cyrl': 'микмак', 'uz-Latn': 'mikmak', 'vec': 'micmac', 'vi': 'Tiếng Micmac', 'xnr': 'मिकमैक', 'yo': 'Èdè Mikmaki', 'yrl': 'mikemake', 'yue': '米克馬克文', 'yue-Hans': '米克马克文', 'yue-Hant': '米克馬克文', 'zh': '密克马克语', 'zh-Hans': '密克马克语', 'zh-Hant': '米克馬克文', 'zu': 'isi-Micmac'},
    'mid': {'en': 'Mandaic'},
    'mie': {'en': 'Ocotepec Mixtec'},
    'mif': {'en': 'Mofu-Gudur'},
    'mig': {'en': 'San Miguel El Grande Mixtec'},
    'mih': {'en': 'Chayuco Mixtec'},
    'mii': {'en': 'Chigmecatitlán Mixtec'},
    'mij': {'en': 'Abar'},
    'mik': {'en': 'Mikasuki'},
    'mil': {'en': 'Peñoles Mixtec'},
    'mim': {'en': 'Alacatlatzala Mixtec'},
    'min': {'af': 'Minangkabaus', 'am': 'ሚናንግካባኡ', 'ar': 'المينانجكاباو', 'as': 'মিনাংকাবাউ', 'ast': 'minangkabau', 'az': 'minanqkaban', 'az-Cyrl': 'минангкабан', 'az-Latn': 'minanqkaban', 'be': 'мінангкабау', 'be-tarask': 'мінангкабау', 'bg': 'минангкабау', 'bgn': 'مینانگ\u200cکابویی', 'bn': 'মিনাংকাবাউ', 'br': 'minangkabau', 'brx': 'मिनांकाबाऊ', 'bs': 'minangkabau', 'bs-Cyrl': 'минангкабау', 'bs-Latn': 'minangkabau', 'ca': 'minangkabau', 'ca-ES-valencia': 'minangkabau', 'ccp': '𑄟𑄨𑄚𑄋𑄴𑄇𑄝𑄃𑄪', 'ce': 'минангкабау', 'chr': 'ᎻᎾᎧᏆᎤ', 'ckb': 'مینانکاباو', 'cs': 'minangkabau', 'cy': 'Minangkabau', 'da': 'minangkabau', 'de': 'Minangkabau', 'dsb': 'minangkabaušćina', 'el': 'Μινανγκαμπάου', 'el-polyton': 'Μινανγκαμπάου', 'en': 'Minangkabau', 'es': 'minangkabau', 'et': 'minangkabau', 'eu': 'minangkabauera', 'fa': 'مینانگ\u200cکابویی', 'ff-Adlm': '𞤃𞤭𞤲𞤢𞤲𞤺𞤳𞤢𞤦𞤢𞤵𞥅𞤪𞤫', 'fi': 'minangkabau', 'fil': 'Minangkabau', 'fo': 'minangkabau', 'fr': 'minangkabau', 'fy': 'Minangkabau', 'ga': 'Míneangcababhais', 'gd': 'Minangkabau', 'gl': 'minangkabau', 'gsw': 'Minangkabau-Schpraach', 'gu': 'મિનાંગ્કાબાઉ', 'ha': 'Minangkabau', 'he': 'מיננגקבאו', 'hi': 'मिनांग्काबाउ', 'hi-Latn': 'Minangkabau', 'hr': 'minangkabau', 'hsb': 'minangkabaušćina', 'ht': 'minangkabau', 'hu': 'minangkabau', 'hy': 'մինանգկաբաու', 'ia': 'minangkabau', 'id': 'Minangkabau', 'ig': 'Minangkabau', 'is': 'mínangkabá', 'it': 'menangkabau', 'ja': 'ミナンカバウ語', 'jv': 'Minangkabau', 'ka': 'მინანგკაბაუ', 'kgp': 'mĩnỹgkamavu', 'kk': 'минангкабау тілі', 'kk-Cyrl': 'минангкабау тілі', 'km': 'មីណាងកាប៊ូ', 'kn': 'ಮಿನಂಗ್\u200cಕಬಾವು', 'ko': '미낭카바우어', 'kok': 'मिनान्गकाबाव', 'kok-Deva': 'मिनान्गकाबाव', 'ks': 'مِنَنگکَباو', 'ks-Arab': 'مِنَنگکَباو', 'ku': 'mînangkabawî', 'ky': 'минанкабауча', 'lb': 'Minangkabau-Sprooch', 'lo': 'ມີແນງກາບູ', 'lt': 'minangkabau', 'lv': 'minangkabavu', 'mai': 'मिननकबाऊ', 'mi': 'Minākapao', 'mk': 'минангкабау', 'ml': 'മിനാങ്കബൗ', 'mn': 'минангкабау', 'mr': 'मिनांग्काबाउ', 'ms': 'Minangkabau', 'mt': 'Minangkabau', 'my': 'စူကူမီနန်', 'nb': 'minangkabau', 'ne': 'मिनाङकाबाउ', 'nl': 'Minangkabau', 'nn': 'minangkabau', 'no': 'minangkabau', 'or': 'ମିନାଙ୍ଗାବାଉ', 'pa': 'ਮਿਨਾਂਗਕਾਬਾਓ', 'pa-Guru': 'ਮਿਨਾਂਗਕਾਬਾਓ', 'pcm': 'Minangkabáu', 'pl': 'minangkabu', 'ps': 'مينيگاباو', 'pt': 'minangkabau', 'qu': 'Minangkabau Simi', 'rm': 'minangkabau', 'ro': 'minangkabau', 'ru': 'минангкабау', 'sc': 'minangkabau', 'sd': 'مناڪابوا', 'sd-Arab': 'مناڪابوا', 'si': 'මිනන්ග්කබාවු', 'sk': 'minangkabaučina', 'sl': 'minangkabau', 'smn': 'minangkabau', 'so': 'Minangkabau', 'sq': 'minangkabauisht', 'sr': 'минангкабау', 'sr-Cyrl': 'минангкабау', 'sr-Latn': 'minangkabau', 'sv': 'minangkabau', 'sw': 'Kiminangkabau', 'ta': 'மின்னாங்கபௌ', 'te': 'మినాంగ్\u200cకాబో', 'th': 'มีนังกาเบา', 'ti': 'ሚናንግካባው', 'tk': 'minangkabau dili', 'to': 'lea fakaminangikapau', 'tr': 'Minangkabau', 'ug': 'مىناڭكابائۇچە', 'uk': 'мінангкабау', 'ur': 'منانگکباؤ', 'uz': 'minangkabau', 'uz-Cyrl': 'минангкабау', 'uz-Latn': 'minangkabau', 'vec': 'minangkabau', 'vi': 'Tiếng Minangkabau', 'xnr': 'मिनांग्काबाउ', 'yo': 'Èdè Minakabau', 'yrl': 'minãgikabau', 'yue': '米南卡堡文', 'yue-Hans': '米南卡堡文', 'yue-Hant': '米南卡堡文', 'zh': '米南佳保语', 'zh-Hans': '米南佳保语', 'zh-Hant': '米南卡堡文', 'zu': 'isi-Minangkabau'},
    'mio': {'en': 'Pinotepa Nacional Mixtec'},
    'mip': {'en': 'Apasco-Apoala Mixtec'},
    'miq': {'en': 'Mískito'},
    'mir': {'en': 'Isthmus Mixe'},
    'mis': {'en': 'Uncoded languages'},
    'mit': {'en': 'Southern Puebla Mixtec'},
    'miu': {'en': 'Cacaloxtepec Mixtec'},
    'miw': {'en': 'Akoye'},
    'mix': {'en': 'Mixtepec Mixtec'},
    'miy': {'en': 'Ayutla Mixtec'},
    'miz': {'en': 'Coatzospan Mixtec'},
    'mja': {'en': 'Mahei'},
    'mjb': {'en': 'Makalero'},
    'mjc': {'en': 'San Juan Colorado Mixtec'},
    'mjd': {'en': 'Northwest Maidu'},
    'mje': {'en': 'Muskum'},
    'mjg': {'en': 'Tu'},
    'mjh': {'en': 'Mwera (Nyasa)'},
    'mji': {'en': 'Kim Mun'},
    'mjj': {'en': 'Mawak'},
    'mjk': {'en': 'Matukar'},
    'mjl': {'en': 'Mandeali'},
    'mjm': {'en': 'Medebur'},
    'mjn': {'en': 'Ma (Papua New Guinea)'},
    'mjo': {'en': 'Malankuravan'},
    'mjp': {'en': 'Malapandaram'},
    'mjq': {'en': 'Malaryan'},
    'mjr': {'en': 'Malavedan'},
    'mjs': {'en': 'Miship'},
    'mjt': {'en': 'Sauria Paharia'},
    'mju': {'en': 'Manna-Dora'},
    'mjv': {'en': 'Mannan'},
    'mjw': {'en': 'Karbi'},
    'mjx': {'en': 'Mahali'},
    'mjy': {'en': 'Mahican'},
    'mjz': {'en': 'Majhi'},
    'mk': {'af': 'Masedonies', 'ak': 'Mɛsidonia kasa', 'am': 'ሜቄዶንኛ', 'ar': 'المقدونية', 'as': 'মেচিডোনীয়', 'ast': 'macedoniu', 'az': 'makedon', 'az-Arab': 'مقدونیجه', 'az-Cyrl': 'македон', 'az-Latn': 'makedon', 'bal-Latn': 'Makduni', 'be': 'македонская', 'be-tarask': 'македонская', 'bg': 'македонски', 'bgn': 'مقدونی', 'bn': 'ম্যাসিডোনীয়', 'br': 'makedoneg', 'brx': 'मेसेड’नियान', 'bs': 'makedonski', 'bs-Cyrl': 'македонски', 'bs-Latn': 'makedonski', 'ca': 'macedoni', 'ca-ES-valencia': 'macedoni', 'ccp': '𑄟𑄳𑄠𑄥𑄨𑄓𑄮𑄚𑄩𑄠𑄧', 'ce': 'македонхойн', 'chr': 'ᎹᏎᏙᏂᎠᏂ', 'ckb': 'ماكێدۆنی', 'cs': 'makedonština', 'cy': 'Macedoneg', 'da': 'makedonsk', 'de': 'Mazedonisch', 'dsb': 'makedońšćina', 'dz': 'མ་སེ་ཌོ་ནི་ཡཱན་ཁ', 'ee': 'makedoniagbe', 'el': 'Σλαβομακεδονικά', 'el-polyton': 'Σλαβομακεδονικά', 'en': 'Macedonian', 'en-Dsrt': '𐐣𐐰𐑅𐐯𐐼𐐬𐑌𐐨𐐲𐑌', 'eo': 'makedona', 'es': 'macedonio', 'et': 'makedoonia', 'eu': 'mazedoniera', 'fa': 'مقدونی', 'ff-Adlm': '𞤃𞤢𞤧𞤫𞤣𞤮𞤲𞤭𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'makedonia', 'fil': 'Macedonian', 'fo': 'makedónskt', 'fr': 'macédonien', 'fur': 'macedon', 'fy': 'Macedonysk', 'ga': 'Macadóinis', 'gd': 'Masadonais', 'gl': 'macedonio', 'gsw': 'Mazedonisch', 'gu': 'મેસેડોનિયન', 'ha': 'Dan Masedoniya', 'he': 'מקדונית', 'hi': 'मकदूनियाई', 'hi-Latn': 'Macedonian', 'hr': 'makedonski', 'hsb': 'makedonšćina', 'ht': 'macédonien', 'hu': 'macedón', 'hy': 'մակեդոներեն', 'ia': 'macedone', 'id': 'Makedonia', 'ig': 'Macedonian', 'is': 'makedónska', 'it': 'macedone', 'ja': 'マケドニア語', 'jv': 'Makedonia', 'ka': 'მაკედონური', 'kea': 'masedóniu', 'kgp': 'mỹsenojũ', 'kk': 'македон тілі', 'kk-Cyrl': 'македон тілі', 'km': 'ម៉ាសេដូនី', 'kn': 'ಮೆಸಿಡೋನಿಯನ್', 'ko': '마케도니아어', 'kok': 'मेसेडोनियन', 'kok-Deva': 'मेसेडोनियन', 'ks': 'میکَڈونیَن', 'ks-Arab': 'میکَڈونیَن', 'ksh': 'Mazedohnesch', 'ku': 'makedonî', 'kxv': 'mesiḍoniyati', 'kxv-Deva': 'मकदूनियाति', 'kxv-Latn': 'mesiḍoniyati', 'kxv-Orya': 'ମାସେଡୋନିଆ ତି', 'kxv-Telu': 'మాసిడోనియన్', 'ky': 'македончо', 'lb': 'Mazedonesch', 'lkt': 'Macedonia Iyápi', 'lld': 'macedonich', 'lo': 'ແມຊິໂດນຽນ', 'lrc': 'مأقدوٙنی', 'lt': 'makedonų', 'lv': 'maķedoniešu', 'mai': 'मेसिडोनियाई', 'mi': 'Makerōnia', 'mk': 'македонски', 'ml': 'മാസിഡോണിയൻ', 'mn': 'македон', 'mr': 'मॅसेडोनियन', 'ms': 'Macedonia', 'mt': 'Maċedonjan', 'my': 'မက်ဆီဒိုးနီးယား', 'mzn': 'مقدونی', 'nb': 'makedonsk', 'ne': 'म्यासेडोनियन', 'nl': 'Macedonisch', 'nn': 'makedonsk', 'no': 'makedonsk', 'nqo': 'ߡߊߛߋߘߏߣߌߞߊ߲', 'om': 'Afaan Macedooniyaa', 'or': 'ମାସେଡୋନିଆନ୍', 'os': 'мӕчъидон', 'pa': 'ਮੈਕਡੋਨੀਆਈ', 'pa-Guru': 'ਮੈਕਡੋਨੀਆਈ', 'pcm': 'Masẹdóniá Lángwej', 'pl': 'macedoński', 'ps': 'مقدوني', 'pt': 'macedônio', 'qu': 'Macedonio Simi', 'rm': 'macedon', 'ro': 'macedoneană', 'ru': 'македонский', 'rw': 'Ikimasedoniya', 'sc': 'matzèdone', 'sd': 'ميسي ڊونيائي', 'sd-Arab': 'ميسي ڊونيائي', 'se': 'makedoniagiella', 'si': 'මැසිඩෝනියානු', 'sk': 'macedónčina', 'sl': 'makedonščina', 'smn': 'makedoniakielâ', 'so': 'Masadooniyaan', 'sq': 'maqedonisht', 'sr': 'македонски', 'sr-Cyrl': 'македонски', 'sr-Latn': 'makedonski', 'sv': 'makedonska', 'sw': 'Kimacedonia', 'ta': 'மாஸிடோனியன்', 'te': 'మాసిడోనియన్', 'tg': 'мақдунӣ', 'th': 'มาซิโดเนีย', 'ti': 'መቄዶንኛ', 'tk': 'makedon dili', 'to': 'lea fakamasitōnia', 'tr': 'Makedonca', 'tt': 'македон', 'ug': 'ماكېدونچە', 'uk': 'македонська', 'ur': 'مقدونیائی', 'uz': 'makedon', 'uz-Cyrl': 'македонча', 'uz-Latn': 'makedon', 'vec': 'masèdone', 'vi': 'Tiếng Macedonia', 'wae': 'Mazedoniš', 'wo': 'Maseduwaane', 'xnr': 'मकदूनियाई', 'yi': 'מאַקעדאניש', 'yo': 'Èdè Masidonia', 'yrl': 'masedũniu', 'yue': '馬其頓文', 'yue-Hans': '马其顿文', 'yue-Hant': '馬其頓文', 'zh': '马其顿语', 'zh-Hans': '马其顿语', 'zh-Hant': '馬其頓文', 'zu': 'isi-Macedonian'},
    'mka': {'en': 'Mbre'},
    'mkb': {'en': 'Mal Paharia'},
    'mkc': {'en': 'Siliput'},
    'mke': {'en': 'Mawchi'},
    'mkf': {'en': 'Miya'},
    'mkg': {'en': 'Mak (China)'},
    'mkh': {'en': 'Mon-Khmer languages'},
    'mkh-mvi': {'en': 'Middle Vietnamese'},
    'mkh-pro': {'en': 'Proto-Mon-Khmer'},
    'mkh-vie-pro': {'en': 'Proto-Vietic'},
    'mki': {'en': 'Dhatki'},
    'mkj': {'en': 'Mokilese'},
    'mkk': {'en': 'Byep'},
    'mkl': {'en': 'Mokole'},
    'mkm': {'en': 'Moklen'},
    'mkn': {'en': 'Kupang Malay'},
    'mko': {'en': 'Mingang Doso'},
    'mkp': {'en': 'Moikodi'},
    'mkq': {'en': 'Bay Miwok'},
    'mkr': {'en': 'Malas'},
    'mks': {'en': 'Silacayoapan Mixtec'},
    'mkt': {'en': 'Vamale'},
    'mku': {'en': 'Konyanka Maninka'},
    'mkv': {'en': 'Mafea'},
    'mkw': {'en': 'Kituba (Congo)'},
    'mkx': {'en': 'Kinamiging Manobo'},
    'mky': {'en': 'East Makian'},
    'mkz': {'en': 'Makasae'},
    'ml': {'af': 'Malabaars', 'ak': 'Malayalam kasa', 'am': 'ማላያላም', 'ar': 'المالايالامية', 'as': 'মালায়ালম', 'ast': 'malayalam', 'az': 'malayalam', 'az-Cyrl': 'малајалам', 'az-Latn': 'malayalam', 'bal-Latn': 'Malyálam', 'be': 'малаялам', 'be-tarask': 'малаялам', 'bg': 'малаялам', 'bgn': 'مالایالامی', 'bn': 'মালায়ালাম', 'br': 'malayalam', 'brx': 'मालायालम', 'bs': 'malajalam', 'bs-Cyrl': 'малајалам', 'bs-Latn': 'malajalam', 'ca': 'malaiàlam', 'ca-ES-valencia': 'malaiàlam', 'ccp': '𑄟𑄣𑄠𑄣𑄟𑄴', 'ce': 'малаялам', 'chr': 'ᎹᎳᏯᎳᎻ', 'ckb': 'مالایالام', 'cs': 'malajálamština', 'cy': 'Malayalam', 'da': 'malayalam', 'de': 'Malayalam', 'dsb': 'malajamšćina', 'dz': 'མ་ལ་ཡ་ལམ་ཁ', 'ee': 'malayagbe', 'el': 'Μαλαγιαλαμικά', 'el-polyton': 'Μαλαγιαλαμικά', 'en': 'Malayalam', 'eo': 'malajalama', 'es': 'malayálam', 'et': 'malajalami', 'eu': 'malabarera', 'fa': 'مالایالامی', 'ff-Adlm': '𞤃𞤢𞤤𞤢𞤴𞤢𞤤𞤢𞤥𞤪𞤫', 'fi': 'malajalam', 'fil': 'Malayalam', 'fo': 'malayalam', 'fr': 'malayalam', 'fur': 'malayalam', 'fy': 'Malayalam', 'ga': 'Mailéalaimis', 'gd': 'Malayalam', 'gl': 'malabar', 'gsw': 'Malayalam', 'gu': 'મલયાલમ', 'ha': 'Malayalamci', 'he': 'מליאלאם', 'hi': 'मलयालम', 'hi-Latn': 'Malayalam', 'hr': 'malajalamski', 'hsb': 'malajamšćina', 'ht': 'malayalam', 'hu': 'malajálam', 'hy': 'մալայալամ', 'ia': 'malayalam', 'id': 'Malayalam', 'ig': 'Malayalam', 'is': 'malajalam', 'it': 'malayalam', 'ja': 'マラヤーラム語', 'jv': 'Malayalam', 'ka': 'მალაიალამური', 'kea': 'malaialam', 'kgp': 'mỹrajara', 'kk': 'малаялам тілі', 'kk-Cyrl': 'малаялам тілі', 'km': 'ម៉ាឡាយ៉ាឡាម', 'kn': 'ಮಲಯಾಳಂ', 'ko': '말라얄람어', 'kok': 'मल्याळम', 'kok-Deva': 'मल्याळम', 'ks': 'مٔلیالَم', 'ks-Arab': 'مٔلیالَم', 'ksh': 'Mallajalam', 'ku': 'malayalamî', 'kxv': 'malyalam', 'kxv-Deva': 'मलयालम', 'kxv-Latn': 'malyalam', 'kxv-Orya': 'ମାଲାୟଲମ୍', 'kxv-Telu': 'మలయాలం', 'ky': 'малайаламча', 'lb': 'Malayalam', 'lkt': 'Malayalam Iyápi', 'lld': 'malayalam', 'lo': 'ມາເລອາລຳ', 'lrc': 'مالایام', 'lt': 'malajalių', 'lv': 'malajalu', 'mai': 'मलयालम', 'mi': 'Mareiārama', 'mk': 'малајалски', 'ml': 'മലയാളം', 'mn': 'малаялам', 'mr': 'मल्याळम', 'ms': 'Malayalam', 'mt': 'Malayalam', 'my': 'မလေယာလမ်', 'mzn': 'مالایالام', 'nb': 'malayalam', 'ne': 'मलयालम', 'nl': 'Malayalam', 'nn': 'malayalam', 'no': 'malayalam', 'nqo': 'ߡߟߊߦߟߊߡߑߞߊ߲', 'om': 'Malayaalamiffaa', 'or': 'ମାଲାୟଲମ୍', 'pa': 'ਮਲਿਆਲਮ', 'pa-Guru': 'ਮਲਿਆਲਮ', 'pcm': 'Maléyálám Lángwej', 'pl': 'malajalam', 'ps': 'مالايالم', 'pt': 'malaiala', 'qu': 'Malayalam Simi', 'rm': 'malayalam', 'ro': 'malayalam', 'ru': 'малаялам', 'rw': 'Ikimalayalami', 'sc': 'malayalam', 'sd': 'مليالم', 'sd-Arab': 'مليالم', 'si': 'මලයාලම්', 'sk': 'malajálamčina', 'sl': 'malajalamščina', 'smn': 'malajam', 'so': 'Malayalam', 'sq': 'malajalamisht', 'sr': 'малајалам', 'sr-Cyrl': 'малајалам', 'sr-Latn': 'malajalam', 'sv': 'malayalam', 'sw': 'Kimalayalamu', 'ta': 'மலையாளம்', 'te': 'మలయాళం', 'tg': 'малаяламӣ', 'th': 'มาลายาลัม', 'ti': 'ማላያላም', 'tk': 'malaýalam dili', 'to': 'lea fakaʻinitia-malāialami', 'tr': 'Malayalam dili', 'tt': 'малаялам', 'ug': 'مالايالامچە', 'uk': 'малаялам', 'ur': 'مالایالم', 'uz': 'malayalam', 'uz-Cyrl': 'малаялам', 'uz-Latn': 'malayalam', 'vec': 'malayàlam', 'vi': 'Tiếng Malayalam', 'wae': 'Malayalam', 'wo': 'Malayalam', 'xnr': 'मलयालम', 'yi': 'מאַלאַיאַלאַם', 'yo': 'Málàyálámù', 'yrl': 'marayara', 'yue': '馬來亞拉姆文', 'yue-Hans': '马来亚拉姆文', 'yue-Hant': '馬來亞拉姆文', 'zh': '马拉雅拉姆语', 'zh-Hans': '马拉雅拉姆语', 'zh-Hant': '馬來亞拉姆文', 'zu': 'isi-Malayalam'},
    'mla': {'en': 'Malo'},
    'mlb': {'en': 'Mbule'},
    'mlc': {'en': 'Cao Lan'},
    'mld': {'en': 'Malakhel'},
    'mle': {'en': 'Manambu'},
    'mlf': {'en': 'Mal'},
    'mlh': {'en': 'Mape'},
    'mli': {'en': 'Malimpung'},
    'mlj': {'en': 'Miltu'},
    'mlk': {'en': 'Ilwana'},
    'mll': {'en': 'Malua Bay'},
    'mlm': {'en': 'Mulam'},
    'mln': {'en': 'Malango'},
    'mlo': {'en': 'Mlomp'},
    'mlp': {'en': 'Bargam'},
    'mlq': {'en': 'Western Maninkakan'},
    'mlr': {'en': 'Vame'},
    'mls': {'en': 'Masalit'},
    'mlu': {'en': "To'abaita"},
    'mlv': {'en': 'Motlav'},
    'mlw': {'en': 'Moloko'},
    'mlx': {'en': 'Malfaxal'},
    'mlz': {'en': 'Malaynon'},
    'mma': {'en': 'Mama'},
    'mmb': {'en': 'Momina'},
    'mmc': {'en': 'Michoacán Mazahua'},
    'mmd': {'en': 'Maonan'},
    'mme': {'en': 'Mae'},
    'mmf': {'en': 'Mundat'},
    'mmg': {'en': 'North Ambrym'},
    'mmh': {'en': 'Mehináku'},
    'mmi': {'en': 'Musar'},
    'mmj': {'en': 'Majhwar'},
    'mmk': {'en': 'Mukha-Dora'},
    'mml': {'en': 'Man Met'},
    'mmm': {'en': 'Maii'},
    'mmn': {'en': 'Mamanwa'},
    'mmo': {'en': 'Mangga Buang'},
    'mmp': {'en': 'Siawi'},
    'mmq': {'en': 'Musak'},
    'mmr': {'en': 'Western Xiangxi Miao'},
    'mmt': {'en': 'Malalamai'},
    'mmu': {'en': 'Mmaala'},
    'mmv': {'en': 'Miriti'},
    'mmw': {'en': 'Emae'},
    'mmx': {'en': 'Madak'},
    'mmy': {'en': 'Migaama'},
    'mmz': {'en': 'Mabaale'},
    'mn': {'af': 'Mongools', 'ak': 'Mongoliafoɔ kasa', 'am': 'ሞንጎሊያኛ', 'ar': 'المنغولية', 'as': 'মংগোলীয়', 'ast': 'mongol', 'az': 'monqol', 'az-Arab': 'موْغولجا', 'az-Cyrl': 'монгол', 'az-Latn': 'monqol', 'bal-Latn': 'Mangóli', 'be': 'мангольская', 'be-tarask': 'мангольская', 'bg': 'монголски', 'bgn': 'منگولی', 'bn': 'মঙ্গোলিয়', 'br': 'mongoleg', 'brx': 'मंगलियान', 'bs': 'mongolski', 'bs-Cyrl': 'монголски', 'bs-Latn': 'mongolski', 'ca': 'mongol', 'ca-ES-valencia': 'mongol', 'ccp': '𑄟𑄧𑄁𑄉𑄮𑄣𑄨𑄠𑄧', 'ce': 'монголийн', 'chr': 'ᎹᏂᎪᎵᎠᏂ', 'ckb': 'مەنگۆلی', 'cs': 'mongolština', 'cy': 'Mongoleg', 'da': 'mongolsk', 'de': 'Mongolisch', 'dsb': 'mongolšćina', 'ee': 'mongoliagbe', 'el': 'Μογγολικά', 'el-polyton': 'Μογγολικά', 'en': 'Mongolian', 'en-Dsrt': '𐐣𐐱𐑍𐐬𐑊𐐨𐐲𐑌', 'eo': 'mongola', 'es': 'mongol', 'et': 'mongoli', 'eu': 'mongoliera', 'fa': 'مغولی', 'ff-Adlm': '𞤃𞤮𞤲𞤺𞤮𞤤𞤭𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'mongoli', 'fil': 'Mongolian', 'fo': 'mongolskt', 'fr': 'mongol', 'fur': 'mongul', 'fy': 'Mongools', 'ga': 'Mongóilis', 'gd': 'Mongolais', 'gl': 'mongol', 'gsw': 'Mongolisch', 'gu': 'મોંગોલિયન', 'ha': 'Mongoliyanci', 'he': 'מונגולית', 'hi': 'मंगोलियाई', 'hi-Latn': 'Mongolian', 'hr': 'mongolski', 'hsb': 'mongolšćina', 'ht': 'mongol', 'hu': 'mongol', 'hy': 'մոնղոլերեն', 'ia': 'mongol', 'id': 'Mongolia', 'ig': 'Mọngolịan', 'is': 'mongólska', 'it': 'mongolo', 'ja': 'モンゴル語', 'jv': 'Mongolia', 'ka': 'მონღოლური', 'kgp': 'mãgór', 'kk': 'моңғол тілі', 'kk-Cyrl': 'моңғол тілі', 'km': 'ម៉ុងហ្គោលី', 'kn': 'ಮಂಗೋಲಿಯನ್', 'ko': '몽골어', 'kok': 'मंगोलियन', 'kok-Deva': 'मंगोलियन', 'ks': 'مَنگولی', 'ks-Arab': 'مَنگولی', 'ksh': 'Mongjohlesch', 'ku': 'moxolî', 'kxv': 'mongoliyati', 'kxv-Deva': 'मंगोलियाति', 'kxv-Latn': 'mongoliyati', 'kxv-Orya': 'ମଙ୍ଗଲିୟ ତି', 'kxv-Telu': 'మగోంలియన్', 'ky': 'монголчо', 'lb': 'Mongolesch', 'lld': 'mongolich', 'lo': 'ມອງໂກເລຍ', 'lrc': 'موغولی', 'lt': 'mongolų', 'lv': 'mongoļu', 'mai': 'मोंगोलियाई', 'mi': 'Mongōria', 'mk': 'монголски', 'ml': 'മംഗോളിയൻ', 'mn': 'монгол', 'mr': 'मंगोलियन', 'ms': 'Mongolia', 'mt': 'Mongoljan', 'my': 'မွန်ဂိုလီးယား', 'mzn': 'مغولی', 'nb': 'mongolsk', 'ne': 'मङ्गोलियाली', 'nl': 'Mongools', 'nn': 'mongolsk', 'no': 'mongolsk', 'nqo': 'ߡߐ߲ߜ߭ߐߟߌߞߊ߲', 'or': 'ମଙ୍ଗୋଲୀୟ', 'pa': 'ਮੰਗੋਲੀ', 'pa-Guru': 'ਮੰਗੋਲੀ', 'pcm': 'Mọngóliá Lángwej', 'pl': 'mongolski', 'ps': 'منګولیایی', 'pt': 'mongol', 'qu': 'Mongol Simi', 'rm': 'mongolic', 'ro': 'mongolă', 'ru': 'монгольский', 'rw': 'Ikimongoli', 'sah': 'Моҕуоллуу', 'sc': 'mòngolu', 'sd': 'منگولي', 'sd-Arab': 'منگولي', 'se': 'mongoliagiella', 'si': 'මොංගෝලියානු', 'sk': 'mongolčina', 'sl': 'mongolščina', 'smn': 'mongoliakielâ', 'so': 'Mangooli', 'sq': 'mongolisht', 'sr': 'монголски', 'sr-Cyrl': 'монголски', 'sr-Latn': 'mongolski', 'sv': 'mongoliska', 'sw': 'Kimongolia', 'ta': 'மங்கோலியன்', 'te': 'మంగోలియన్', 'tg': 'муғулӣ', 'th': 'มองโกเลีย', 'ti': 'ሞንጎልኛ', 'tk': 'mongol dili', 'to': 'lea fakamongokōlia', 'tr': 'Moğolca', 'tt': 'монгол', 'ug': 'موڭغۇلچە', 'uk': 'монгольська', 'ur': 'منگولین', 'uz': 'mongol', 'uz-Cyrl': 'мўғулча', 'uz-Latn': 'mongol', 'vec': 'móngolo', 'vi': 'Tiếng Mông Cổ', 'wae': 'Mongoliš', 'wo': 'Mongoliye', 'xnr': 'मंगोलियाई', 'yi': 'מאנגאליש', 'yo': 'Mòngólíà', 'yrl': 'mũgú', 'yue': '蒙古文', 'yue-Hans': '蒙古文', 'yue-Hant': '蒙古文', 'zh': '蒙古语', 'zh-Hans': '蒙古语', 'zh-Hant': '蒙古文', 'zu': 'isi-Mongolian'},
    'mna': {'en': 'Mbula'},
    'mnb': {'en': 'Muna'},
    'mnc': {'ar': 'المانشو', 'ast': 'manchú', 'az': 'mançu', 'az-Latn': 'mançu', 'bg': 'манджурски', 'bn': 'মাঞ্চু', 'br': 'manchou', 'brx': 'मांचु', 'bs': 'manču', 'bs-Cyrl': 'манчу', 'bs-Latn': 'manču', 'ca': 'manxú', 'ca-ES-valencia': 'manxú', 'ccp': '𑄟𑄚𑄴𑄌𑄪', 'cs': 'mandžuština', 'cy': 'Manshw', 'da': 'manchu', 'de': 'Mandschurisch', 'dz': 'མན་ཇུ་ཁ', 'el': 'Μαντσού', 'el-polyton': 'Μαντσού', 'en': 'Manchu', 'en-Dsrt': '𐐣𐐰𐑌𐐽𐐭', 'es': 'manchú', 'et': 'mandžu', 'fa': 'مانچویی', 'fi': 'mantšu', 'fr': 'mandchou', 'fy': 'Mantsjoe', 'gd': 'Manchu', 'gsw': 'Mandschurisch', 'gu': 'માન્ચુ', 'he': 'מנצ׳ו', 'hi': 'मन्चु', 'hi-Latn': 'Manchu', 'hr': 'mandžurski', 'ht': 'mandchou', 'hu': 'mandzsu', 'id': 'Manchuria', 'is': 'mansjú', 'it': 'manchu', 'ja': '満州語', 'ka': 'მანჯურიული', 'kgp': 'mỹsu', 'kn': 'ಮಂಚು', 'ko': '만주어', 'ks': 'مانچوٗ', 'ks-Arab': 'مانچوٗ', 'lb': 'Mandschuresch', 'lo': 'ແມນຈູ', 'lt': 'manču', 'lv': 'mandžūru', 'mk': 'манџурски', 'ml': 'മാൻ\u200cചു', 'mr': 'मान्चु', 'mt': 'Manchu', 'my': 'မန်ချူး', 'nb': 'mandsju', 'ne': 'मान्चु', 'nl': 'Mantsjoe', 'nn': 'mandsju', 'no': 'mandsju', 'or': 'ମାଞ୍ଚୁ', 'pl': 'manchu', 'pt': 'manchu', 'rm': 'manchu', 'ro': 'manciuriană', 'ru': 'маньчжурский', 'sk': 'mandžuština', 'sl': 'mandžurščina', 'sr': 'манџурски', 'sr-Cyrl': 'манџурски', 'sr-Latn': 'mandžurski', 'sv': 'manchuriska', 'ta': 'மன்சூ', 'te': 'మంచు', 'th': 'แมนจู', 'to': 'lea fakamanisū', 'tr': 'Mançurya dili', 'ug': 'مانجۇچە', 'uk': 'манчжурська', 'vi': 'Tiếng Mãn Châu', 'yrl': 'mãxu', 'yue': '滿族文', 'yue-Hans': '满族文', 'yue-Hant': '滿族文', 'zh': '满语', 'zh-Hans': '满语', 'zh-Hant': '滿族文'},
    'mnd': {'en': 'Mondé'},
    'mne': {'en': 'Naba'},
    'mnf': {'en': 'Mundani'},
    'mng': {'en': 'Eastern Mnong'},
    'mnh': {'en': 'Mono (Democratic Republic of Congo)'},
    'mni': {'af': 'Manipuri', 'ak': 'Manipuri', 'am': 'ማኒፑሪ', 'ar': 'المانيبورية', 'as': 'মণিপুৰী', 'ast': 'manipuri', 'az': 'manipüri', 'az-Cyrl': 'манипүри', 'az-Latn': 'manipüri', 'bal-Latn': 'Manipuri', 'be': 'мейтэй', 'be-tarask': 'мейтэй', 'bg': 'манипурски', 'bgn': 'مانیپوری', 'bn': 'মণিপুরী', 'br': 'manipuri', 'brx': 'मनिपुरी', 'bs': 'manipuri', 'bs-Cyrl': 'манипури', 'bs-Latn': 'manipuri', 'ca': 'manipurí', 'ca-ES-valencia': 'manipurí', 'ccp': '𑄟𑄚𑄨𑄛𑄪𑄢𑄩', 'ce': 'манипурийн', 'chr': 'ᎺᏂᏉᎵ', 'ckb': 'مانیپووری', 'cs': 'manipurština', 'cy': 'Manipwri', 'da': 'manipuri', 'de': 'Meithei', 'dsb': 'manipuršćina', 'el': 'Μανιπούρι', 'el-polyton': 'Μανιπούρι', 'en': 'Manipuri', 'es': 'manipurí', 'et': 'manipuri', 'eu': 'manipurera', 'fa': 'مانیپوری', 'ff-Adlm': '𞤃𞤢𞤲𞤭𞤨𞤵𞥅𞤪𞤫', 'fi': 'manipuri', 'fil': 'Manipuri', 'fo': 'manupuri', 'fr': 'manipuri', 'fy': 'Manipoeri', 'ga': 'Manapúiris', 'gd': 'Manipuri', 'gl': 'manipuri', 'gsw': 'Meithei-Schpraach', 'gu': 'મણિપુરી', 'ha': 'Manipuri', 'he': 'מניפורית', 'hi': 'मणिपुरी', 'hi-Latn': 'Manipuri', 'hr': 'manipurski', 'hsb': 'manipuršćina', 'ht': 'manipuri', 'hu': 'manipuri', 'hy': 'մանիպուրի', 'ia': 'manipuri', 'id': 'Manipuri', 'ig': 'Asụsụ Manipuri', 'is': 'manípúrí', 'it': 'manipuri', 'ja': 'マニプリ語', 'jv': 'Manipuri', 'ka': 'მანიპური', 'kgp': 'mỹnĩpuri', 'kk': 'манипури тілі', 'kk-Cyrl': 'манипури тілі', 'km': 'ម៉ានីពូរី', 'kn': 'ಮಣಿಪುರಿ', 'ko': '마니푸리어', 'kok': 'मणिपुरी', 'kok-Deva': 'मणिपुरी', 'ks': 'مَنیپوٗری', 'ks-Arab': 'مَنیپوٗری', 'ksh': 'Meitei-Manipuhresch', 'ku': 'manipuri', 'kxv': 'maṇipurī', 'kxv-Deva': 'मणिपुरी', 'kxv-Latn': 'maṇipurī', 'kxv-Orya': 'ମଣିପୁରୀ', 'kxv-Telu': 'మణిపురి', 'ky': 'манипуриче', 'lb': 'Meithei-Sprooch', 'lkt': 'Namipuri Iyápi', 'lo': 'ມານີພູຣິ', 'lt': 'manipurių', 'lv': 'manipūru', 'mai': 'मणिपुरी', 'mi': 'Manipuri', 'mk': 'манипурски', 'ml': 'മണിപ്പൂരി', 'mn': 'манипури', 'mni': 'মৈতৈলোন্', 'mni-Beng': 'মৈতৈলোন্', 'mni-Mtei': 'ꯃꯤꯇꯩꯂꯣꯟ', 'mr': 'मणिपुरी', 'ms': 'Manipuri', 'mt': 'Manipuri', 'my': 'မနိပူရ', 'nb': 'manipuri', 'ne': 'मनिपुरी', 'nl': 'Meitei', 'nn': 'manipuri', 'no': 'manipuri', 'nqo': 'ߡߊߣߌߔߎߙߌߞߊ߲', 'or': 'ମଣିପୁରୀ', 'pa': 'ਮਨੀਪੁਰੀ', 'pa-Guru': 'ਮਨੀਪੁਰੀ', 'pcm': 'Manípuri', 'pl': 'manipuri', 'ps': 'مانی پوری', 'pt': 'manipuri', 'qu': 'Manipuri Simi', 'rm': 'manipuri', 'ro': 'manipuri', 'ru': 'манипурский', 'sc': 'manipuri', 'sd': 'ماني پوري', 'sd-Arab': 'ماني پوري', 'si': 'මනිපුරි', 'sk': 'manípurčina', 'sl': 'manipurščina', 'smn': 'manipuri', 'so': 'Maniburi', 'sq': 'manipurisht', 'sr': 'манипурски', 'sr-Cyrl': 'манипурски', 'sr-Latn': 'manipurski', 'sv': 'manipuri', 'sw': 'Kimanipuri', 'ta': 'மணிப்புரி', 'te': 'మణిపురి', 'tg': 'манипурӣ', 'th': 'มณีปุระ', 'ti': 'ማኒፑሪ', 'tk': 'manipuri dili', 'to': 'lea fakamanipuli', 'tr': 'Manipuri dili', 'tt': 'манипури', 'ug': 'مانىپۇرچە', 'uk': 'маніпурі', 'ur': 'منی پوری', 'uz': 'manipur', 'uz-Cyrl': 'манипурча', 'uz-Latn': 'manipur', 'vec': 'manipur', 'vi': 'Tiếng Manipuri', 'wo': 'Manipuri', 'xnr': 'मणिपुरी', 'yo': 'Èdè Manipuri', 'yrl': 'manipuri', 'yue': '曼尼普裡文', 'yue-Hans': '曼尼普里文', 'yue-Hant': '曼尼普裡文', 'zh': '曼尼普尔语', 'zh-Hans': '曼尼普尔语', 'zh-Hant': '曼尼普爾文', 'zu': 'isi-Manipuri'},
    'mnj': {'en': 'Munji'},
    'mnk': {'en': 'Mandinka'},
    'mnl': {'en': 'Tiale'},
    'mnm': {'en': 'Mapena'},
    'mnn': {'en': 'Southern Mnong'},
    'mno': {'en': 'Manobo languages'},
    'mnp': {'en': 'Min Bei Chinese'},
    'mnq': {'en': 'Minriq'},
    'mnr': {'en': 'Mono (USA)'},
    'mns': {'en': 'Mansi'},
    'mnt': {'en': 'Maykulan'},
    'mnu': {'en': 'Mer'},
    'mnv': {'en': 'Rennell-Bellona'},
    'mnw': {'en': 'Mon'},
    'mnx': {'en': 'Manikion'},
    'mny': {'en': 'Manyawa'},
    'mnz': {'en': 'Moni'},
    'mo': {'en': 'Moldavian'},
    'moa': {'en': 'Mwan'},
    'moc': {'en': 'Mocoví'},
    'mod': {'en': 'Mobilian'},
    'moe': {'af': 'Innu-aimun', 'am': 'ኢኑ-አይመን', 'ar': 'إينو-ايمون', 'as': 'ইন্নু-আইমুন', 'az': 'İnnu-aimun', 'az-Latn': 'İnnu-aimun', 'be': 'іну-аймун', 'be-tarask': 'іну-аймун', 'bg': 'инну-аймун', 'bn': 'ইন্নু-এমুন', 'br': 'montagneg', 'brx': 'इन्नु-आइमुन', 'bs': 'innu-aimun', 'bs-Cyrl': 'иму-аимун', 'bs-Latn': 'innu-aimun', 'ca': 'innu-aimun', 'ca-ES-valencia': 'innu-aimun', 'chr': 'ᎢᏄ-ᎠᏱᏵᏂ', 'cs': 'innu-aimun', 'cy': 'Innu-aimun', 'da': 'innu-aimun', 'de': 'Innu-Aimun', 'dsb': 'innu-aimunšćina', 'el': 'Ινου-αϊμούν', 'el-polyton': 'Ινου-αϊμούν', 'en': 'Innu-aimun', 'es': 'innu-aimun', 'et': 'innu', 'eu': 'innuera', 'fa': 'اینوآیموم', 'ff-Adlm': '𞤋𞤲𞥆𞤵-𞤢𞤴𞤥𞤵𞤲𞤪𞤫', 'fi': 'innu-aimun', 'fil': 'Innu-aimun', 'fr': 'innu-aimun', 'ga': 'Innu-aimun', 'gd': 'Innu-aimun', 'gl': 'innu-aimun', 'gu': 'ઇન્નુ-આયમુન', 'ha': 'Innu-aimun', 'he': 'אינו-אמון', 'hi': 'इन्नु-एईमन', 'hi-Latn': 'Innu-aimun', 'hr': 'innu-aimun', 'hsb': 'innu-aimunšćina', 'ht': 'innu-aimun', 'hu': 'innu-aimun', 'hy': 'իննու-այմուն', 'ia': 'innu-aimun', 'id': 'Innu-aimun', 'ig': 'Innu-aimun', 'is': 'innu-aimun', 'it': 'innu-aimun', 'ja': 'イヌー＝アイムン語', 'jv': 'Innu-aimun', 'ka': 'ინუ-აიმუნი', 'kk': 'инну-аймун тілі', 'kk-Cyrl': 'инну-аймун тілі', 'km': 'អ៊ីននូអៃមុន', 'kn': 'ಇನ್ನು-ಐಮುನ್', 'ko': '이누아문', 'kok': 'इन्यु-ऐमुन', 'kok-Deva': 'इन्यु-ऐमुन', 'ku': 'înûyîya rojhilatî', 'ky': 'инну-аймунча (индейлер тили)', 'lo': 'ອິນນຸໄອມັນ', 'lt': 'montanjų', 'lv': 'motanju', 'mai': 'इनु-ऐमन', 'mi': 'Inu-aimuna', 'mk': 'ину-аимунски', 'ml': 'ഇന്നു-ഐമൂൻ', 'mn': 'инну-аймун', 'mr': 'इन्नू-ॲमन', 'ms': 'Innu-aimun', 'my': 'အီနုအေမွန်', 'nb': 'innu-aimun', 'ne': 'इन्नु-ऐमन', 'nl': 'Innu-aimun', 'nn': 'innu-aimun', 'no': 'innu-aimun', 'or': 'ଇନ୍ନୁ-ଏମୁନ', 'pa': 'ਇਨੂੰ-ਏਮੁਨ', 'pa-Guru': 'ਇਨੂੰ-ਏਮੁਨ', 'pcm': 'Innu-aímun Lángwej', 'pl': 'innu-aimun', 'ps': 'mo', 'pt': 'innu-aimun', 'qu': 'Innu-aimun Simi', 'ro': 'innu-aimun', 'ru': 'инну-аймун', 'sc': 'innu-aimun', 'sd': 'انو آئيمن', 'sd-Arab': 'انو آئيمن', 'si': 'ඉනු-අයිමුන්', 'sk': 'innu-aimunčina', 'sl': 'inuaimunščina', 'so': 'Innu-aimun', 'sq': 'inuaimunisht', 'sr': 'инуајмун', 'sr-Cyrl': 'инуајмун', 'sr-Latn': 'inuajmun', 'sv': 'innu-aimun', 'sw': 'Kiinnu-aimun', 'ta': 'இன்னு-ஐமுன்', 'te': 'ఇన్ను-ఐమున్', 'th': 'อินนุ-ไอมุน', 'ti': 'ኢኑ-ኤመን', 'tk': 'innu-aýmun dili', 'to': 'lea fakaʻinuʻaimuni', 'tr': 'Doğu İnnucası', 'uk': 'інну-аймун', 'ur': 'انو ایمن', 'uz': 'innu-aymun', 'uz-Latn': 'innu-aymun', 'vec': 'inu aimun', 'vi': 'Tiếng Innu-aimun', 'yo': 'Èdè Inuamu', 'yue': '因努埃蒙文', 'yue-Hans': '因努埃蒙文', 'yue-Hant': '因努埃蒙文', 'zh': '因努埃蒙语', 'zh-Hans': '因努埃蒙语', 'zh-Hant': '因紐艾蒙語', 'zu': 'isi-Innu-aimun'},
    'mof': {'en': 'Mohegan-Montauk-Narragansett'},
    'mog': {'en': 'Mongondow'},
    'moh': {'af': 'Mohawk', 'am': 'ሞሃውክ', 'ar': 'الموهوك', 'as': 'মোহোক', 'ast': 'mohawk', 'az': 'mohavk', 'az-Cyrl': 'моһавк', 'az-Latn': 'mohavk', 'bal-Latn': 'Mohawk', 'be': 'мохак', 'be-tarask': 'мохак', 'bg': 'мохоук', 'bgn': 'موهاکی', 'bn': 'মোহাওক', 'br': 'mohawk', 'brx': 'महाउक', 'bs': 'mohavk', 'bs-Cyrl': 'махавски', 'bs-Latn': 'mohavk', 'ca': 'mohawk', 'ca-ES-valencia': 'mohawk', 'ccp': '𑄟𑄮𑄦𑄃𑄮𑄇𑄴', 'ce': 'мохаук', 'chr': 'ᎼᎭᎩ', 'ckb': 'مۆهاوک', 'cs': 'mohawkština', 'cy': 'Mohoceg', 'da': 'mohawk', 'de': 'Mohawk', 'dsb': 'mohawkšćina', 'el': 'Μοχόκ', 'el-polyton': 'Μοχόκ', 'en': 'Mohawk', 'en-Dsrt': '𐐐𐐬𐐸𐐪𐐿', 'es': 'mohawk', 'et': 'mohoogi', 'eu': 'mohawkera', 'fa': 'موهاکی', 'ff-Adlm': '𞤃𞤮𞥅𞤸𞤢𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'mohawk', 'fil': 'Mohawk', 'fo': 'mohawk', 'fr': 'mohawk', 'fy': 'Mohawk', 'ga': 'Móháicis', 'gd': 'Mohawk', 'gl': 'mohawk', 'gsw': 'Mohawk-Schpraach', 'gu': 'મોહૌક', 'ha': 'Mohawk', 'he': 'מוהוק', 'hi': 'मोहौक', 'hi-Latn': 'Mohawk', 'hr': 'mohok', 'hsb': 'mohawkšćina', 'ht': 'mohawk', 'hu': 'mohawk', 'hy': 'մոհավք', 'ia': 'mohawk', 'id': 'Mohawk', 'ig': 'Mohawk', 'is': 'móhíska', 'it': 'mohawk', 'ja': 'モーホーク語', 'jv': 'Mohawk', 'ka': 'მოჰაუკური', 'kgp': 'mãjkỹnũ', 'kk': 'могавк тілі', 'kk-Cyrl': 'могавк тілі', 'km': 'ម៊ូហាគ', 'kn': 'ಮೊಹಾವ್ಕ್', 'ko': '모호크어', 'kok': 'मोहाक', 'kok-Deva': 'मोहाक', 'ks': 'موہاک', 'ks-Arab': 'موہاک', 'ku': 'mohawkî', 'ky': 'мохаукча', 'lb': 'Mohawk-Sprooch', 'lo': 'ໂມຫາ', 'lrc': 'موٙھاڤک', 'lt': 'mohok', 'lv': 'mohauku', 'mai': 'मोहौक', 'mi': 'Mauhōka', 'mk': 'мохавски', 'ml': 'മോഹാക്', 'mn': 'мохаук', 'moh': 'Kanienʼkéha', 'mr': 'मोहॉक', 'ms': 'Mohawk', 'mt': 'Mohawk', 'my': 'မိုဟော့ခ်', 'mzn': 'موهاک', 'nb': 'mohawk', 'ne': 'मोहक', 'nl': 'Mohawk', 'nn': 'mohawk', 'no': 'mohawk', 'or': 'ମୋହୌକ', 'pa': 'ਮੋਹਆਕ', 'pa-Guru': 'ਮੋਹਆਕ', 'pcm': 'Móhọ́k Lángwej', 'pl': 'mohawk', 'ps': 'محاواک', 'pt': 'moicano', 'qu': 'Mohawk Simi', 'rm': 'mohawk', 'ro': 'mohawk', 'ru': 'мохаук', 'sc': 'mohawk', 'sd': 'موهاڪ', 'sd-Arab': 'موهاڪ', 'si': 'මොහොව්ක්', 'sk': 'mohawkčina', 'sl': 'mohoščina', 'smn': 'mohawkkielâ', 'so': 'Mohawk', 'sq': 'mohokisht', 'sr': 'мохочки', 'sr-Cyrl': 'мохочки', 'sr-Latn': 'mohočki', 'sv': 'mohawk', 'sw': 'Lugha ya Mohawk', 'ta': 'மொஹாக்', 'te': 'మోహాక్', 'tg': 'моҳок', 'th': 'โมฮอว์ก', 'ti': 'ሞሃውክ', 'tk': 'mogauk dili', 'to': 'lea fakamohauki', 'tr': 'Mohavk dili', 'tt': 'могаук', 'ug': 'موخاۋكچە', 'uk': 'магавк', 'ur': 'موہاک', 'uz': 'mohauk', 'uz-Cyrl': 'могавк', 'uz-Latn': 'mohauk', 'vec': 'mohawk', 'vi': 'Tiếng Mohawk', 'wo': 'Mowak', 'xnr': 'मोहौक', 'yo': 'Èdè Mohaki', 'yrl': 'muikanu', 'yue': '莫霍克文', 'yue-Hans': '莫霍克文', 'yue-Hant': '莫霍克文', 'zh': '摩霍克语', 'zh-Hans': '摩霍克语', 'zh-Hant': '莫霍克文', 'zu': 'isi-Mohawk'},
    'moi': {'en': 'Mboi'},
    'moj': {'en': 'Monzombo'},
    'mok': {'en': 'Morori'},
    'mom': {'en': 'Mangue'},
    'moo': {'en': 'Monom'},
    'mop': {'en': 'Mopán Maya'},
    'moq': {'en': 'Mor (Bomberai Peninsula)'},
    'mor': {'en': 'Moro'},
    'mos': {'af': 'Mossi', 'am': 'ሞሲ', 'ar': 'الموسي', 'as': 'মোছি', 'ast': 'mossi', 'az': 'mosi', 'az-Cyrl': 'моси', 'az-Latn': 'mosi', 'be': 'мосі', 'be-tarask': 'мосі', 'bg': 'моси', 'bgn': 'موسیی', 'bn': 'মসি', 'br': 'more', 'brx': 'मस्सी', 'bs': 'mosi', 'bs-Cyrl': 'моси', 'bs-Latn': 'mosi', 'ca': 'moore', 'ca-ES-valencia': 'moore', 'ccp': '𑄟𑄧𑄥𑄨', 'ce': 'моси', 'chr': 'ᎼᏍᏏ', 'ckb': 'مۆسی', 'cs': 'mosi', 'cy': 'Mosi', 'da': 'mossi', 'de': 'Mossi', 'dsb': 'mossišćina', 'el': 'Μόσι', 'el-polyton': 'Μόσι', 'en': 'Mossi', 'es': 'mossi', 'et': 'more', 'eu': 'mossiera', 'fa': 'ماسیایی', 'ff-Adlm': '𞤃𞤮𞥅𞤧𞤭𞥅𞤪𞤫', 'fi': 'mossi', 'fil': 'Mossi', 'fo': 'mossi', 'fr': 'moré', 'fy': 'Mossi', 'ga': 'Mosais', 'gd': 'Mossi', 'gl': 'mossi', 'gsw': 'Mossi-Schpraach', 'gu': 'મોસ્સી', 'ha': 'Mossi', 'he': 'מוסי', 'hi': 'मोस्सी', 'hi-Latn': 'Mossi', 'hr': 'mossi', 'hsb': 'mossišćina', 'ht': 'moré', 'hu': 'moszi', 'hy': 'մոսսի', 'ia': 'mossi', 'id': 'Mossi', 'ig': 'Mossi', 'is': 'mossí', 'it': 'mossi', 'ja': 'モシ語', 'jv': 'Mossi', 'ka': 'მოსი', 'kgp': 'mosi', 'kk': 'мосси тілі', 'kk-Cyrl': 'мосси тілі', 'km': 'មូស៊ី', 'kn': 'ಮೊಸ್ಸಿ', 'ko': '모시어', 'kok': 'मोस्सी', 'kok-Deva': 'मोस्सी', 'ks': 'موسی', 'ks-Arab': 'موسی', 'ku': 'moreyî', 'ky': 'моссиче', 'lb': 'Mossi-Sprooch', 'lo': 'ມອສຊີ', 'lt': 'mosi', 'lv': 'mosu', 'mai': 'मोसी', 'mi': 'Mohi', 'mk': 'моси', 'ml': 'മൊസ്സി', 'mn': 'мосси', 'mr': 'मोस्सी', 'ms': 'Mossi', 'mt': 'Mossi', 'my': 'မိုစီ', 'nb': 'mossi', 'ne': 'मोस्सी', 'nl': 'Mossi', 'nn': 'mossi', 'no': 'mossi', 'or': 'ମୋସି', 'pa': 'ਮੋਸੀ', 'pa-Guru': 'ਮੋਸੀ', 'pcm': 'Mósí Lángwej', 'pl': 'mossi', 'ps': 'ماسي', 'pt': 'mossi', 'qu': 'Mossi Simi', 'rm': 'mossi', 'ro': 'mossi', 'ru': 'моси', 'sc': 'moore', 'sd': 'موسي', 'sd-Arab': 'موسي', 'si': 'මොස්සි', 'sk': 'mossi', 'sl': 'mosijščina', 'smn': 'moore', 'so': 'Mossi', 'sq': 'mosisht', 'sr': 'моси', 'sr-Cyrl': 'моси', 'sr-Latn': 'mosi', 'sv': 'mossi', 'sw': 'Kimoore', 'ta': 'மோஸ்ஸி', 'te': 'మోస్సి', 'th': 'โมซี', 'ti': 'ሞሲ', 'tk': 'mossi dili', 'to': 'lea fakamosi', 'tr': 'Mossi', 'ug': 'موسسىچە', 'uk': 'моссі', 'ur': 'موسی', 'uz': 'mossi', 'uz-Cyrl': 'мосси', 'uz-Latn': 'mossi', 'vec': 'mosi', 'vi': 'Tiếng Mossi', 'xnr': 'मोस्सी', 'yo': 'Èdè Mosi', 'yrl': 'musi', 'yue': '莫西文', 'yue-Hans': '莫西文', 'yue-Hant': '莫西文', 'zh': '莫西语', 'zh-Hans': '莫西语', 'zh-Hant': '莫西文', 'zu': 'isi-Mossi'},
    'mot': {'en': 'Barí'},
    'mou': {'en': 'Mogum'},
    'mov': {'en': 'Mohave'},
    'mow': {'en': 'Moi (Congo)'},
    'mox': {'en': 'Molima'},
    'moy': {'en': 'Shekkacho'},
    'moz': {'en': 'Mukulu'},
    'mpa': {'en': 'Mpoto'},
    'mpb': {'en': 'Malak Malak'},
    'mpc': {'en': 'Mangarrayi'},
    'mpd': {'en': 'Machinere'},
    'mpe': {'en': 'Majang'},
    'mpg': {'en': 'Marba'},
    'mph': {'en': 'Maung'},
    'mpi': {'en': 'Mpade'},
    'mpj': {'en': 'Martu Wangka'},
    'mpk': {'en': 'Mbara (Chad)'},
    'mpl': {'en': 'Middle Watut'},
    'mpm': {'en': 'Yosondúa Mixtec'},
    'mpn': {'en': 'Mindiri'},
    'mpo': {'en': 'Miu'},
    'mpp': {'en': 'Migabac'},
    'mpq': {'en': 'Matís'},
    'mpr': {'en': 'Vangunu'},
    'mps': {'en': 'Dadibi'},
    'mpt': {'en': 'Mian'},
    'mpu': {'en': 'Makuráp'},
    'mpv': {'en': 'Mungkip'},
    'mpw': {'en': 'Mapidian'},
    'mpx': {'en': 'Misima-Panaeati'},
    'mpy': {'en': 'Mapia'},
    'mpz': {'en': 'Mpi'},
    'mqa': {'en': 'Maba (Indonesia)'},
    'mqb': {'en': 'Mbuko'},
    'mqc': {'en': 'Mangole'},
    'mqe': {'en': 'Matepi'},
    'mqf': {'en': 'Momuna'},
    'mqg': {'en': 'Kota Bangun Kutai Malay'},
    'mqh': {'en': 'Tlazoyaltepec Mixtec'},
    'mqi': {'en': 'Mariri'},
    'mqj': {'en': 'Mamasa'},
    'mqk': {'en': 'Rajah Kabunsuwan Manobo'},
    'mql': {'en': 'Mbelime'},
    'mqm': {'en': 'South Marquesan'},
    'mqn': {'en': 'Moronene'},
    'mqo': {'en': 'Modole'},
    'mqp': {'en': 'Manipa'},
    'mqq': {'en': 'Minokok'},
    'mqr': {'en': 'Mander'},
    'mqs': {'en': 'West Makian'},
    'mqt': {'en': 'Mok'},
    'mqu': {'en': 'Mandari'},
    'mqv': {'en': 'Mosimo'},
    'mqw': {'en': 'Murupi'},
    'mqx': {'en': 'Mamuju'},
    'mqy': {'en': 'Manggarai'},
    'mqz': {'en': 'Pano'},
    'mr': {'af': 'Marathi', 'ak': 'Marati', 'am': 'ማራቲ', 'ar': 'الماراثية', 'as': 'মাৰাঠী', 'ast': 'marathi', 'az': 'marathi', 'az-Cyrl': 'маратһи', 'az-Latn': 'marathi', 'bal-Latn': 'Maráthi', 'be': 'маратхі', 'be-tarask': 'маратхі', 'bg': 'марати', 'bgn': 'مراٹی', 'bn': 'মারাঠি', 'br': 'marathi', 'brx': 'माराथी', 'bs': 'marati', 'bs-Cyrl': 'марати', 'bs-Latn': 'marati', 'ca': 'marathi', 'ca-ES-valencia': 'marathi', 'ccp': '𑄟𑄢𑄒𑄨', 'ce': 'маратхи', 'chr': 'ᎹᎳᏘ', 'ckb': 'ماراتی', 'cs': 'maráthština', 'cy': 'Marathi', 'da': 'marathi', 'de': 'Marathi', 'dsb': 'maratišćina', 'dz': 'མ་ར་ཐི་ཁ', 'ee': 'marathiagbe', 'el': 'Μαραθικά', 'el-polyton': 'Μαραθικά', 'en': 'Marathi', 'eo': 'marata', 'es': 'maratí', 'et': 'marathi', 'eu': 'marathera', 'fa': 'مراتی', 'ff-Adlm': '𞤃𞤢𞤪𞤢𞤼𞤭𞥅𞤪𞤫', 'fi': 'marathi', 'fil': 'Marathi', 'fo': 'marathi', 'fr': 'marathi', 'fur': 'marathi', 'fy': 'Marathi', 'ga': 'Maraitis', 'gd': 'Marathi', 'gl': 'marathi', 'gsw': 'Marathi', 'gu': 'મરાઠી', 'ha': 'Maratinci', 'he': 'מראטהית', 'hi': 'मराठी', 'hi-Latn': 'Marathi', 'hr': 'marathski', 'hsb': 'maratišćina', 'ht': 'marathi', 'hu': 'maráthi', 'hy': 'մարաթի', 'ia': 'marathi', 'id': 'Marathi', 'ig': 'Asụsụ Marathi', 'is': 'maratí', 'it': 'marathi', 'ja': 'マラーティー語', 'jv': 'Marathi', 'ka': 'მარათჰი', 'kea': 'marati', 'kgp': 'marati', 'kk': 'маратхи тілі', 'kk-Cyrl': 'маратхи тілі', 'km': 'ម៉ារ៉ាធី', 'kn': 'ಮರಾಠಿ', 'ko': '마라티어', 'kok': 'मराठी', 'kok-Deva': 'मराठी', 'kok-Latn': 'Moratthi', 'ks': 'مَرٲٹھؠ', 'ks-Arab': 'مَرٲٹھؠ', 'ksh': 'Marrahtesch', 'ku': 'maratî', 'kxv': 'maraṭi', 'kxv-Deva': 'मराठी', 'kxv-Latn': 'maraṭi', 'kxv-Orya': 'ମରାଟୀ', 'kxv-Telu': 'మరాటి', 'ky': 'маратиче', 'lb': 'Marathi', 'lkt': 'Marathi Iyápi', 'lld': 'marathi', 'lo': 'ມາຣາທີ', 'lrc': 'مأراتی', 'lt': 'maratų', 'lv': 'marathu', 'mai': 'मराठी', 'mi': 'Marati', 'mk': 'марати', 'ml': 'മറാത്തി', 'mn': 'марати', 'mr': 'मराठी', 'ms': 'Marathi', 'mt': 'Marathi', 'my': 'မာရသီ', 'mzn': 'ماراتی', 'nb': 'marathi', 'ne': 'मराठी', 'nl': 'Marathi', 'nn': 'marathi', 'no': 'marathi', 'nqo': 'ߡߙߊߕߌߞߊ߲', 'om': 'Afaan Maratii', 'or': 'ମରାଠୀ', 'pa': 'ਮਰਾਠੀ', 'pa-Guru': 'ਮਰਾਠੀ', 'pcm': 'Marátí Lángwej', 'pl': 'marathi', 'ps': 'مراټهي', 'pt': 'marati', 'qu': 'Marathi Simi', 'rm': 'marathi', 'ro': 'marathi', 'ru': 'маратхи', 'rw': 'Ikimarati', 'sc': 'marathi', 'sd': 'مراٺي', 'sd-Arab': 'مراٺي', 'si': 'මරාති', 'sk': 'maráthčina', 'sl': 'maratščina', 'smn': 'marathikielâ', 'so': 'Maarati', 'sq': 'maratisht', 'sr': 'марати', 'sr-Cyrl': 'марати', 'sr-Latn': 'marati', 'sv': 'marathi', 'sw': 'Kimarathi', 'ta': 'மராத்தி', 'te': 'మరాఠీ', 'tg': 'маратҳӣ', 'th': 'มราฐี', 'ti': 'ማራቲ', 'tk': 'marathi dili', 'to': 'lea fakamalati', 'tr': 'Marathi dili', 'tt': 'маратхи', 'ug': 'ماراتىچە', 'uk': 'маратхі', 'ur': 'مراٹھی', 'uz': 'maratxi', 'uz-Cyrl': 'маратхи', 'uz-Latn': 'maratxi', 'vec': 'marati', 'vi': 'Tiếng Marathi', 'wae': 'Marathi', 'wo': 'Marati', 'xnr': 'मराठी', 'yo': 'Èdè marathi', 'yrl': 'marati', 'yue': '馬拉地文', 'yue-Hans': '马拉地文', 'yue-Hant': '馬拉地文', 'zh': '马拉地语', 'zh-Hans': '马拉地语', 'zh-Hant': '馬拉地文', 'zu': 'isi-Marathi'},
    'mra': {'en': 'Mlabri'},
    'mrb': {'en': 'Marino'},
    'mrc': {'en': 'Maricopa'},
    'mrd': {'en': 'Western Magar'},
    'mre': {'en': "Martha's Vineyard Sign Language"},
    'mrf': {'en': 'Elseng'},
    'mrg': {'en': 'Mising'},
    'mrh': {'en': 'Mara Chin'},
    'mrj': {'ast': 'mari occidental', 'br': 'marieg ar Cʼhornôg', 'ca': 'mari occidental', 'ca-ES-valencia': 'mari occidental', 'cs': 'marijština (západní)', 'cy': 'Mari Gorllewinol', 'de': 'Bergmari', 'en': 'Western Mari', 'et': 'mäemari', 'fi': 'vuorimari', 'fr': 'mari occidental', 'ga': 'Mairis Iartharach', 'gd': 'Mari Shiarach', 'gu': 'પશ્ચિમી મારી', 'hi-Latn': 'Western Mari', 'ht': 'mari occidental', 'hy': 'արևմտամարիերեն', 'it': 'mari occidentale', 'ja': '山地マリ語', 'ko': '서부 마리어', 'lb': 'West-Mari', 'lt': 'vakarų mari', 'mk': 'западномариски', 'nb': 'vestmarisk', 'nl': 'West-Mari', 'nn': 'vestmarisk', 'no': 'vestmarisk', 'pl': 'zachodniomaryjski', 'smn': 'viestârmari', 'sv': 'västmariska', 'th': 'มารีตะวันตก', 'to': 'lea fakamali-hihifo', 'tr': 'Ova Çirmişçesi', 'yue': '西馬裏文', 'yue-Hans': '西马里文', 'yue-Hant': '西馬裏文', 'zh-Hant': '西馬里文'},
    'mrk': {'en': 'Hmwaveke'},
    'mrl': {'en': 'Mortlockese'},
    'mrm': {'en': 'Merlav'},
    'mrn': {'en': 'Cheke Holo'},
    'mro': {'en': 'Mru'},
    'mrp': {'en': 'Morouas'},
    'mrq': {'en': 'North Marquesan'},
    'mrr': {'en': 'Maria (India)'},
    'mrs': {'en': 'Maragus'},
    'mrt': {'en': 'Marghi Central'},
    'mru': {'en': 'Mono (Cameroon)'},
    'mrv': {'en': 'Mangareva'},
    'mrw': {'en': 'Maranao'},
    'mrx': {'en': 'Maremgi'},
    'mry': {'en': 'Mandaya'},
    'mrz': {'en': 'Marind'},
    'ms': {'ms': 'bahasa Malaysia', 'af': 'Maleis', 'agq': 'Màlae', 'ak': 'Malay kasa', 'am': 'ማላይ', 'ar': 'الماليزية', 'as': 'মালয়', 'asa': 'Kimalesia', 'ast': 'malayu', 'az': 'malay', 'az-Cyrl': 'малај', 'az-Latn': 'malay', 'bal-Latn': 'Malai', 'bas': 'Hɔp u makɛ᷆', 'be': 'малайская', 'be-tarask': 'малайская', 'bem': 'Ichi Maleshani', 'bez': 'Himalesia', 'bg': 'малайски', 'bgn': 'مالایی', 'bm': 'malɛzikan', 'bn': 'মালয়', 'br': 'malayseg', 'brx': 'मलै', 'bs': 'malajski', 'bs-Cyrl': 'малајски', 'bs-Latn': 'malajski', 'ca': 'malai', 'ca-ES-valencia': 'malai', 'ccp': '𑄟𑄣𑄧𑄠𑄴', 'ce': 'малайн', 'cgg': 'Orumalesiya', 'chr': 'ᎹᎴ', 'ckb': 'مالیزی', 'cs': 'malajština', 'cy': 'Maleieg', 'da': 'malajisk', 'dav': 'Kimalesia', 'de': 'Malaiisch', 'dje': 'Maleezi senni', 'dsb': 'malajšćina', 'dyo': 'maleesi', 'dz': 'མ་ལེ་ཁ', 'ebu': 'Kĩmalesia', 'ee': 'malaygbe', 'el': 'Μαλαισιανά', 'el-polyton': 'Μαλαισιανά', 'en': 'Malay', 'eo': 'malaja', 'es': 'malayo', 'et': 'malai', 'eu': 'malaysiera', 'ewo': 'Ǹkɔ́bɔ malɛ́sian', 'fa': 'مالایی', 'ff': 'Malayeere', 'ff-Adlm': '𞤃𞤢𞤤𞤫𞥅𞤪𞤫', 'ff-Latn': 'Malayeere', 'fi': 'malaiji', 'fil': 'Malay', 'fo': 'malaiiskt', 'fr': 'malais', 'fur': 'malês', 'fy': 'Maleis', 'ga': 'Malaeis', 'gd': 'Malaidhis', 'gl': 'malaio', 'gsw': 'Malaiisch', 'gu': 'મલય', 'guz': 'Kimalesia', 'ha': 'Harshen Malay', 'he': 'מלאית', 'hi': 'मलय', 'hi-Latn': 'Malay', 'hr': 'malajski', 'hsb': 'malajšćina', 'ht': 'malais', 'hu': 'maláj', 'hy': 'մալայերեն', 'ia': 'malay', 'id': 'Melayu', 'ig': 'Malay', 'is': 'malaíska', 'it': 'malese', 'ja': 'マレー語', 'jmc': 'Kyimalesia', 'jv': 'Melayu', 'ka': 'მალაიური', 'kab': 'Tamalawit', 'kam': 'Kimalesia', 'kde': 'Chimalesia', 'kea': 'maláiu', 'kgp': 'mỹrajo', 'khq': 'Maleezi senni', 'ki': 'Kimalesia', 'kk': 'малай тілі', 'kk-Cyrl': 'малай тілі', 'kln': 'kutitab Malay', 'km': 'ម៉ាឡេ', 'kn': 'ಮಲಯ್', 'ko': '말레이어', 'kok': 'मलय', 'kok-Deva': 'मलय', 'ks': 'مَلَے', 'ks-Arab': 'مَلَے', 'ksb': 'Kimalesia', 'ksf': 'rimalaí', 'ksh': 'Mallaijesch', 'ku': 'malezî', 'kxv': 'malei', 'kxv-Deva': 'मलय', 'kxv-Latn': 'malei', 'kxv-Orya': 'ମଲୟ', 'kxv-Telu': 'మలయ్', 'ky': 'малайча', 'lag': 'Kɨmelésia', 'lb': 'Malaiesch', 'lg': 'Lumalayi', 'lkt': 'Malay Iyápi', 'lld': 'malesc', 'ln': 'limalezi', 'lo': 'ມາເລ', 'lrc': 'مالایی', 'lt': 'malajiečių', 'lu': 'Limalezia', 'luo': 'Kimalesia', 'luy': 'Kimalesia', 'lv': 'malajiešu', 'mai': 'मलय', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ malay', 'mer': 'Kĩmalesia', 'mfe': 'male', 'mg': 'Malay', 'mgh': 'Imalesia', 'mi': 'Marei', 'mk': 'малајски', 'ml': 'മലെയ്', 'mn': 'малай', 'mr': 'मलय', 'mt': 'Malay', 'mua': 'malasiya', 'my': 'မလေး', 'mzn': 'مالایی', 'naq': 'Malayǁî gowab', 'nb': 'malayisk', 'nd': 'isi-Malayi', 'ne': 'मलाय', 'nl': 'Maleis', 'nmg': 'Kiɛl Malɛ̌siā', 'nn': 'malayisk', 'no': 'malayisk', 'nqo': 'ߡߊߟߍߞߊ߲', 'nus': 'Thok mayɛyni', 'nyn': 'Orumalesiya', 'om': 'Malaayiffaa', 'or': 'ମାଲୟ', 'pa': 'ਮਲਯ', 'pa-Guru': 'ਮਲਯ', 'pcm': 'Malé Lángwej', 'pl': 'malajski', 'ps': 'ملایا', 'pt': 'malaio', 'qu': 'Malayo Simi', 'rm': 'malaic', 'rn': 'Ikinyamaleziya', 'ro': 'malaeză', 'rof': 'Kimalesia', 'ru': 'малайский', 'rw': 'Ikimalayi', 'rwk': 'Kyimalesia', 'sah': 'Малаайдыы', 'saq': 'Kimalesia', 'sbp': 'Ishimalesiya', 'sc': 'malesu', 'sd': 'ملي', 'sd-Arab': 'ملي', 'seh': 'malaio', 'ses': 'Maleezi senni', 'sg': 'Malëe', 'shi': 'ⵜⴰⵎⴰⵍⴰⵡⵉⵜ', 'shi-Latn': 'Tamalawit', 'shi-Tfng': 'ⵜⴰⵎⴰⵍⴰⵡⵉⵜ', 'si': 'මැලේ', 'sk': 'malajčina', 'sl': 'malajščina', 'smn': 'malaiji', 'sn': 'chiMalay', 'so': 'Malaay', 'sq': 'malajisht', 'sr': 'малајски', 'sr-Cyrl': 'малајски', 'sr-Latn': 'malajski', 'sv': 'malajiska', 'sw': 'Kimalei', 'ta': 'மலாய்', 'te': 'మలయ్', 'teo': 'Kimalesia', 'tg': 'малайӣ', 'th': 'มาเลย์', 'ti': 'ማላይኛ', 'tk': 'malaý dili', 'to': 'lea fakamalei', 'tr': 'Malayca', 'tt': 'малай', 'twq': 'Maleezi senni', 'tzm': 'Tamalizit', 'ug': 'مالايچە', 'uk': 'малайська', 'ur': 'مالے', 'uz': 'malay', 'uz-Cyrl': 'малай тил', 'uz-Latn': 'malay', 'vai': 'ꕮꔒꔀ', 'vai-Latn': 'Maléee', 'vai-Vaii': 'ꕮꔒꔀ', 'vec': 'maleze', 'vi': 'Tiếng Mã Lai', 'vun': 'Kyimalesia', 'wae': 'Malaíš', 'wo': 'Malay', 'xnr': 'मलय', 'xog': 'Olumalayi', 'yav': 'máliɛ', 'yo': 'Èdè Malaya', 'yrl': 'marayu', 'yue': '馬來文', 'yue-Hans': '马来文', 'yue-Hant': '馬來文', 'zgh': 'ⵜⴰⵎⴰⵍⴰⵡⵉⵜ', 'zh': '马来语', 'zh-Hans': '马来语', 'zh-Hant': '馬來文', 'zu': 'isi-Malay'},
    'msb': {'en': 'Masbatenyo'},
    'msc': {'en': 'Sankaran Maninka'},
    'msd': {'en': 'Yucatec Maya Sign Language'},
    'mse': {'en': 'Musey'},
    'msf': {'en': 'Mekwei'},
    'msg': {'en': 'Moraid'},
    'msh': {'en': 'Masikoro Malagasy'},
    'msi': {'en': 'Sabah Malay'},
    'msj': {'en': 'Ma (Democratic Republic of Congo)'},
    'msk': {'en': 'Mansaka'},
    'msl': {'en': 'Molof'},
    'msm': {'en': 'Agusan Manobo'},
    'msn': {'en': 'Vurës'},
    'mso': {'en': 'Mombum'},
    'msp': {'en': 'Maritsauá'},
    'msq': {'en': 'Caac'},
    'msr': {'en': 'Mongolian Sign Language'},
    'mss': {'en': 'West Masela'},
    'mst': {'en': 'Cataelano Mandaya'},
    'msu': {'en': 'Musom'},
    'msv': {'en': 'Maslam'},
    'msw': {'en': 'Mansoanka'},
    'msx': {'en': 'Moresada'},
    'msy': {'en': 'Aruamu'},
    'msz': {'en': 'Momare'},
    'mt': {'af': 'Maltees', 'ak': 'Malta kasa', 'am': 'ማልቲዝኛ', 'ar': 'المالطية', 'as': 'মাল্টিজ', 'ast': 'maltés', 'az': 'malta', 'az-Cyrl': 'малта', 'az-Latn': 'malta', 'bal-Latn': 'Maltiz', 'be': 'мальтыйская', 'be-tarask': 'мальтыйская', 'bg': 'малтийски', 'bgn': 'مالته\u200cای', 'bn': 'মল্টিয়', 'br': 'malteg', 'brx': 'माल्टीज', 'bs': 'malteški', 'bs-Cyrl': 'малтешки', 'bs-Latn': 'malteški', 'ca': 'maltès', 'ca-ES-valencia': 'maltès', 'ccp': '𑄟𑄧𑄣𑄴𑄑𑄨𑄠𑄧', 'ce': 'мальтойн', 'chr': 'ᎹᎵᏘᏍ', 'ckb': 'ماڵتی', 'cs': 'maltština', 'cy': 'Malteg', 'da': 'maltesisk', 'de': 'Maltesisch', 'dsb': 'maltašćina', 'dz': 'མཱལ་ཊ་ཁ', 'ee': 'maltagbe', 'el': 'Μαλτεζικά', 'el-polyton': 'Μαλτεζικά', 'en': 'Maltese', 'eo': 'malta', 'es': 'maltés', 'et': 'malta', 'eu': 'maltera', 'fa': 'مالتی', 'ff-Adlm': '𞤃𞤢𞤤𞤼𞤭𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'malta', 'fil': 'Maltese', 'fo': 'maltiskt', 'fr': 'maltais', 'fur': 'maltês', 'fy': 'Maltees', 'ga': 'Máltais', 'gd': 'Maltais', 'gl': 'maltés', 'gsw': 'Maltesisch', 'gu': 'માલ્ટિઝ', 'ha': 'Harshen Maltis', 'he': 'מלטית', 'hi': 'माल्टीज़', 'hi-Latn': 'Maltese', 'hr': 'malteški', 'hsb': 'maltašćina', 'ht': 'maltais', 'hu': 'máltai', 'hy': 'մալթայերեն', 'ia': 'maltese', 'id': 'Malta', 'ie': 'maltesi', 'ig': 'Asụsụ Malta', 'is': 'maltneska', 'it': 'maltese', 'ja': 'マルタ語', 'jv': 'Malta', 'ka': 'მალტური', 'kea': 'maltes', 'kgp': 'mỹrtej', 'kk': 'мальта тілі', 'kk-Cyrl': 'мальта тілі', 'km': 'ម៉ាល់តា', 'kn': 'ಮಾಲ್ಟೀಸ್', 'ko': '몰타어', 'kok': 'माल्टिज', 'kok-Deva': 'माल्टिज', 'ks': 'مَلتیٖس', 'ks-Arab': 'مَلتیٖس', 'ksh': 'Maltehsesch', 'ku': 'maltayî', 'ky': 'малтизче', 'lb': 'Maltesesch', 'lkt': 'Maltese Iyápi', 'lld': 'maltesc', 'lo': 'ມອລທີສ', 'lrc': 'مالتی', 'lt': 'maltiečių', 'lv': 'maltiešu', 'mai': 'मलटीज', 'mi': 'Mārata', 'mk': 'малтешки', 'ml': 'മാൾട്ടീസ്', 'mn': 'малта', 'mr': 'माल्टिज्', 'ms': 'Malta', 'mt': 'Malti', 'my': 'မော်လ်တာ', 'mzn': 'مالتی', 'nb': 'maltesisk', 'ne': 'माल्टिज', 'nl': 'Maltees', 'nn': 'maltesisk', 'no': 'maltesisk', 'nqo': 'ߡߊߟߑߕߍߞߊ߲', 'om': 'Afaan Maltesii', 'or': 'ମାଲଟୀଜ୍', 'pa': 'ਮਾਲਟੀਜ਼', 'pa-Guru': 'ਮਾਲਟੀਜ਼', 'pcm': 'Mọ́ltá Lángwej', 'pl': 'maltański', 'ps': 'مالټايي', 'pt': 'maltês', 'qu': 'Maltes Simi', 'rm': 'maltais', 'ro': 'malteză', 'ru': 'мальтийский', 'rw': 'Ikimaliteze', 'sc': 'maltesu', 'sd': 'مالٽي', 'sd-Arab': 'مالٽي', 'se': 'maltagiella', 'si': 'මොල්ටිස්', 'sk': 'maltčina', 'sl': 'malteščina', 'smn': 'maltakielâ', 'so': 'Maltiis', 'sq': 'maltisht', 'sr': 'малтешки', 'sr-Cyrl': 'малтешки', 'sr-Latn': 'malteški', 'sv': 'maltesiska', 'sw': 'Kimalta', 'ta': 'மால்டிஸ்', 'te': 'మాల్టీస్', 'tg': 'малтӣ', 'th': 'มอลตา', 'ti': 'ማልትኛ', 'tk': 'malta dili', 'to': 'lea fakamalita', 'tr': 'Maltaca', 'tt': 'мальта', 'ug': 'مالتاچە', 'uk': 'мальтійська', 'ur': 'مالٹی', 'uz': 'maltiy', 'uz-Cyrl': 'малтача', 'uz-Latn': 'maltiy', 'vec': 'malteze', 'vi': 'Tiếng Malta', 'wae': 'Maltesiš', 'wo': 'Malt', 'xnr': 'माल्टीज़', 'yi': 'מאַלטעזיש', 'yo': 'Èdè Malta', 'yrl': 'mautei', 'yue': '馬爾他文', 'yue-Hans': '马尔他文', 'yue-Hant': '馬爾他文', 'zh': '马耳他语', 'zh-Hans': '马耳他语', 'zh-Hant': '馬爾他文', 'zu': 'isi-Maltese'},
    'mta': {'en': 'Cotabato Manobo'},
    'mtb': {'en': 'Anyin Morofo'},
    'mtc': {'en': 'Munit'},
    'mtd': {'en': 'Mualang'},
    'mte': {'en': 'Mono (Solomon Islands)'},
    'mtf': {'en': 'Murik (Papua New Guinea)'},
    'mtg': {'en': 'Una'},
    'mth': {'en': 'Munggui'},
    'mti': {'en': 'Maiwa (Papua New Guinea)'},
    'mtj': {'en': 'Moskona'},
    'mtk': {'en': "Mbe'"},
    'mtl': {'en': 'Montol'},
    'mtm': {'en': 'Mator'},
    'mtn': {'en': 'Matagalpa'},
    'mto': {'en': 'Totontepec Mixe'},
    'mtp': {'en': 'Wichí Lhamtés Nocten'},
    'mtq': {'en': 'Muong'},
    'mtr': {'en': 'Mewari'},
    'mts': {'en': 'Yora'},
    'mtt': {'en': 'Mota'},
    'mtu': {'en': 'Tututepec Mixtec'},
    'mtv': {'en': "Asaro'o"},
    'mtw': {'en': 'Southern Binukidnon'},
    'mtx': {'en': 'Tidaá Mixtec'},
    'mty': {'en': 'Nabi'},
    'mua': {'af': 'Mundang', 'am': 'ሙንዳንግ', 'ar': 'مندنج', 'as': 'মুণ্ডাং', 'ast': 'mundang', 'az': 'mundanq', 'az-Cyrl': 'мунданг', 'az-Latn': 'mundanq', 'bal-Latn': 'Mundang', 'be': 'мунданг', 'be-tarask': 'мунданг', 'bg': 'мунданг', 'bgn': 'ماندانگی', 'bn': 'মুদাঙ্গ', 'br': 'moundangeg', 'brx': 'मुन्दां', 'bs': 'mundang', 'bs-Cyrl': 'мунданг', 'bs-Latn': 'mundang', 'ca': 'mundang', 'ca-ES-valencia': 'mundang', 'ccp': '𑄟𑄪𑄘𑄋𑄴𑄉𑄧', 'ce': 'мунданг', 'chr': 'ᎽᏂᏓᎩ', 'ckb': 'موندانگ', 'cs': 'mundang', 'cy': 'Mundang', 'da': 'mundang', 'de': 'Mundang', 'dsb': 'mundang', 'el': 'Μουντάνγκ', 'el-polyton': 'Μουντάνγκ', 'en': 'Mundang', 'es': 'mundang', 'et': 'mundangi', 'eu': 'mudangera', 'fa': 'ماندانگی', 'ff-Adlm': '𞤃𞤵𞤲𞤣𞤢𞤲𞤪𞤫', 'fi': 'mundang', 'fil': 'Mundang', 'fo': 'mundang', 'fr': 'moundang', 'fy': 'Mundang', 'ga': 'Mundang', 'gd': 'Mundang', 'gl': 'mundang', 'gu': 'મુનડાન્ગ', 'ha': 'Mundang', 'he': 'מונדאנג', 'hi': 'मुंडैंग', 'hi-Latn': 'Mundang', 'hr': 'mundang', 'hsb': 'mundang', 'ht': 'moundang', 'hu': 'mundang', 'hy': 'մունդանգ', 'ia': 'mundang', 'id': 'Mundang', 'ig': 'Mundang', 'is': 'mundang', 'it': 'mundang', 'ja': 'ムンダン語', 'jv': 'Mundang', 'ka': 'მუნდანგი', 'kgp': 'mũnág', 'kk': 'мунданг тілі', 'kk-Cyrl': 'мунданг тілі', 'km': 'មុនដាង', 'kn': 'ಮುಂಡಂಗ್', 'ko': '문당어', 'kok': 'मुडांग', 'kok-Deva': 'मुडांग', 'ksh': 'Mundang', 'ku': 'mundangî', 'ky': 'мундангча', 'lb': 'Mundang', 'lo': 'ມັນດັງ', 'lrc': 'موٙندانگ', 'lt': 'mundangų', 'lv': 'mundangu', 'mai': 'मुंडांग', 'mi': 'Mūnatanga', 'mk': 'мунданг', 'ml': 'മുന്ദാംഗ്', 'mn': 'мунданг', 'mr': 'मुंडांग', 'ms': 'Mundang', 'mt': 'Mundang', 'mua': 'MUNDAŊ', 'my': 'မွန်ဒန်း', 'mzn': 'موندانگ', 'nb': 'mundang', 'ne': 'मुन्डाङ', 'nl': 'Mundang', 'nn': 'mundang', 'no': 'mundang', 'nqo': 'ߡߎ߲ߘߊ߲ߞߊ߲', 'or': 'ମୁନଡାଂ', 'pa': 'ਮੁੰਡੇਂਗ', 'pa-Guru': 'ਮੁੰਡੇਂਗ', 'pcm': 'Mundáng Lángwej', 'pl': 'mundang', 'ps': 'مندانګ', 'pt': 'mundang', 'qu': 'Mundang Simi', 'rm': 'mundang', 'ro': 'mundang', 'ru': 'мунданг', 'sc': 'mundang', 'sd': 'من دانگ', 'sd-Arab': 'من دانگ', 'si': 'මුන්ඩන්', 'sk': 'mundang', 'sl': 'mundang', 'smn': 'mundang', 'so': 'Miyundhaang', 'sq': 'mundangisht', 'sr': 'мунданг', 'sr-Cyrl': 'мунданг', 'sr-Latn': 'mundang', 'sv': 'mundang', 'sw': 'Kimundang', 'ta': 'முன்டாங்', 'te': 'మండాంగ్', 'th': 'มันดัง', 'ti': 'ሙንዳንግ', 'tk': 'mundang dili', 'to': 'lea fakamunitangi', 'tr': 'Mundang', 'ug': 'مۇنداڭچە', 'uk': 'мунданг', 'ur': 'منڈانگ', 'uz': 'mundang', 'uz-Cyrl': 'мунданг', 'uz-Latn': 'mundang', 'vec': 'mundang', 'vi': 'Tiếng Mundang', 'xnr': 'मुंडैंग', 'yo': 'Múndàngì', 'yrl': 'mũdãgi', 'yue': '蒙當文', 'yue-Hans': '蒙当文', 'yue-Hant': '蒙當文', 'zh': '蒙当语', 'zh-Hans': '蒙当语', 'zh-Hant': '蒙當文', 'zu': 'isi-Mundang'},
    'mub': {'en': 'Mubi'},
    'muc': {'en': 'Ajumbu'},
    'mud': {'en': 'Mednyj Aleut'},
    'mue': {'en': 'Media Lengua'},
    'mug': {'en': 'Musgu'},
    'muh': {'en': 'Mündü'},
    'mui': {'en': 'Musi'},
    'muj': {'en': 'Mabire'},
    'muk': {'en': 'Mugom'},
    'mul': {'af': 'Verskeie tale', 'ak': 'Kasa ahodoɔ', 'am': 'ባለብዙ ቋንቋዎች', 'ar': 'لغات متعددة', 'as': 'একাধিক ভাষা', 'ast': 'múltiples llingües', 'az': 'çoxsaylı dillər', 'az-Cyrl': 'чохсајлы дилләр', 'az-Latn': 'çoxsaylı dillər', 'bal-Latn': 'Báz zobán', 'be': 'некалькі моў', 'be-tarask': 'некалькі моў', 'bg': 'многоезични', 'bgn': 'چینکه زبان', 'bn': 'একাধিক ভাষা', 'br': 'yezhoù lies', 'brx': 'गोबां रावफोर', 'bs': 'više jezika', 'bs-Cyrl': 'више језика', 'bs-Latn': 'više jezika', 'ca': 'llengües vàries', 'ca-ES-valencia': 'llengües vàries', 'ccp': '𑄝𑄣𑄧𑄇𑄴𑄇𑄚𑄨 𑄞𑄌𑄴', 'ce': 'тайп-тайпа доьзалан меттанаш', 'chr': 'ᏧᏈᏍᏗ ᏗᎦᏬᏂᎯᏍᏗ', 'ckb': 'چەند زمان', 'cs': 'více jazyků', 'csw': 'ᒥᐦᒉᐟ ᐊᔭᒧᐏᓇ', 'cy': 'Mwy nag un iaith', 'da': 'flere sprog', 'de': 'Mehrsprachig', 'dsb': 'wěcejrěcne', 'ee': 'gbegbɔgblɔ sɔgbɔwo', 'el': 'Πολλαπλές γλώσσες', 'el-polyton': 'Πολλαπλές γλῶσσες', 'en': 'Multiple languages', 'en-Dsrt': '𐐣𐐲𐑊𐐻𐐮𐐹𐐲𐑊 𐐢𐐩𐑍𐐶𐐮𐐾𐐲𐑆', 'es': 'varios idiomas', 'et': 'mitu keelt', 'eu': 'zenbait hizkuntza', 'fa': 'چندین زبان', 'ff-Adlm': '𞤍𞤫𞤲𞤯𞤫 𞤅𞤫𞤪𞤼𞤵𞤯𞤫', 'fi': 'useita kieliä', 'fil': 'Maramihang Wika', 'fo': 'ymisk mál', 'fr': 'multilingue', 'fur': 'lenghis multiplis', 'fy': 'Meardere talen', 'ga': 'Ilteangacha', 'gd': 'Iomadh cànan', 'gl': 'varias linguas', 'gsw': 'Mehrschpraachig', 'gu': 'બહુવિધ ભાષાઓ', 'ha': 'Harsuna masu yawa', 'he': 'מספר שפות', 'hi': 'एकाधिक भाषाएँ', 'hi-Latn': 'Multiple languages', 'hr': 'više jezika', 'hsb': 'wjacerěčne', 'ht': 'multilingue', 'hu': 'többszörös nyelvek', 'hy': 'բազմալեզու', 'ia': 'plure linguas', 'id': 'Beberapa Bahasa', 'ig': 'Ọtụtụ asụsụ', 'is': 'mörg tungumál', 'it': 'multilingua', 'ja': '複数言語', 'jv': 'Basa Multilingua', 'ka': 'სხვადასხვა ენა', 'kgp': 'vẽnhvĩ’e', 'kk': 'бірнеше тіл', 'kk-Cyrl': 'бірнеше тіл', 'km': 'ពហុភាសា', 'kn': 'ಬಹುಸಂಖ್ಯೆಯ ಭಾಷೆಗಳು', 'ko': '다중 언어', 'kok': 'साबार भाशा', 'kok-Deva': 'साबार भाशा', 'ks': 'واریاہ زبان', 'ks-Arab': 'واریاہ زبان', 'ksh': '-ongerscheidlijje Schprohche-', 'ku': 'pirzimanî', 'ky': 'бир нече тилде', 'lb': 'Méisproocheg', 'lo': 'ຫລາຍພາສາ', 'lt': 'kelios kalbos', 'lv': 'vairākas valodas', 'mai': 'एक सं बेसी भाषा', 'mi': 'Ngā reo maha', 'mk': 'повеќе јазици', 'ml': 'പലഭാഷകൾ', 'mn': 'олон хэл', 'mr': 'एकाधिक भाषा', 'ms': 'Pelbagai Bahasa', 'mt': 'Lingwi Diversi', 'my': 'ဘာသာစကား အမျိုးမျိုး', 'nb': 'flere språk', 'ne': 'बहुभाषा', 'nl': 'meerdere talen', 'nn': 'fleire språk', 'no': 'flere språk', 'nqo': 'ߞߊ߲ߥߙߍߞߊ߲', 'or': 'ବିବିଧ ଭାଷା', 'pa': 'ਬਹੁਤੀਆਂ ਬੋਲੀਆਂ', 'pa-Guru': 'ਬਹੁਤੀਆਂ ਬੋਲੀਆਂ', 'pcm': 'Plẹ́ntí Lángwej-dẹm', 'pl': 'wiele języków', 'ps': 'متعدد ژبې', 'pt': 'múltiplos idiomas', 'qu': 'Idiomas M´últiples Simi', 'rm': 'pluriling', 'ro': 'mai multe limbi', 'ru': 'языки разных семей', 'sc': 'limbas mùltiplas', 'sd': 'هڪ کان وڌيڪ ٻوليون', 'sd-Arab': 'هڪ کان وڌيڪ ٻوليون', 'si': 'බහු භාෂා', 'sk': 'viaceré jazyky', 'sl': 'več jezikov', 'smn': 'maŋgâ kielâ', 'so': 'Luuqaddo badan', 'sq': 'gjuhë të shumëfishta', 'sr': 'Више језика', 'sr-Cyrl': 'Више језика', 'sr-Latn': 'Više jezika', 'sv': 'flera språk', 'sw': 'Lugha nyingi', 'syr': 'ܠܫ̈ܢܐ ܦܖ̈ܝܫܐ', 'ta': 'பல மொழிகள்', 'te': 'బహుళ భాషలు', 'th': 'หลายภาษา', 'ti': 'ዝተፈላለዩ ቋንቋታት', 'tk': 'birnäçe dil', 'to': 'lea tuifio', 'tr': 'Birden Fazla Dil', 'ug': 'كۆپ تىللار', 'uk': 'кілька мов', 'ur': 'متعدد زبانیں', 'uz': 'bir nechta til', 'uz-Cyrl': 'бир нечта тил', 'uz-Latn': 'bir nechta til', 'vec': 'multilengua', 'vi': 'Nhiều ngôn ngữ', 'xnr': 'इक ते जादा बोलियाँ', 'yo': 'Ọlọ́pọ̀ èdè', 'yrl': 'siía nheẽga', 'yue': '多種語言', 'yue-Hans': '多种语言', 'yue-Hant': '多種語言', 'zh': '多语种', 'zh-Hans': '多语种', 'zh-Hant': '多種語言', 'zu': 'izilimi ezehlukene'},
    'mum': {'en': 'Maiwala'},
    'mun': {'en': 'Munda languages'},
    'muo': {'en': 'Nyong'},
    'mup': {'en': 'Malvi'},
    'muq': {'en': 'Eastern Xiangxi Miao'},
    'mur': {'en': 'Murle'},
    'mus': {'af': 'Kreek', 'am': 'ሙስኮኪ', 'ar': 'الكريك', 'as': 'ক্ৰীক', 'ast': 'creek', 'az': 'krik', 'az-Cyrl': 'крик', 'az-Latn': 'krik', 'bal-Latn': 'Muskógi', 'be': 'мускогі', 'be-tarask': 'мускогі', 'bg': 'мускогски', 'bgn': 'کریکی', 'bn': 'মুস্কোগী', 'br': 'muskogi', 'brx': 'मुस्क’गी', 'bs': 'kriški', 'bs-Cyrl': 'кришки', 'bs-Latn': 'kriški', 'ca': 'creek', 'ca-ES-valencia': 'creek', 'ccp': '𑄇𑄳𑄢𑄨𑄇𑄴', 'ce': 'крик', 'chr': 'ᎠᎫᏌ', 'ckb': 'کریک', 'cs': 'kríkština', 'cy': 'Creek', 'da': 'creek', 'de': 'Muskogee', 'dsb': 'krik', 'el': 'Κρικ', 'el-polyton': 'Κρικ', 'en': 'Muscogee', 'en-Dsrt': '𐐗𐑉𐐨𐐿', 'es': 'creek', 'et': 'maskogi', 'eu': 'muscogeera', 'fa': 'کریکی', 'ff-Adlm': '𞤃𞤵𞤧𞤳𞤮𞤳𞤭𞥅𞤪𞤫', 'fi': 'creek', 'fil': 'Creek', 'fo': 'creek', 'fr': 'creek', 'fy': 'Creek', 'ga': 'Muscogee', 'gd': 'Creek', 'gl': 'creek', 'gsw': 'Muskogee-Schpraach', 'gu': 'ક્રિક', 'ha': 'Muscogee', 'he': 'קריק', 'hi': 'क्रीक', 'hi-Latn': 'Muscogee', 'hr': 'creek', 'hsb': 'krik', 'ht': 'creek', 'hu': 'krík', 'hy': 'մասքոջի', 'ia': 'creek', 'id': 'Bahasa Muskogee', 'ig': 'Muscogee', 'is': 'krík', 'it': 'creek', 'ja': 'クリーク語', 'jv': 'Muskogee', 'ka': 'კრიკი', 'kgp': 'krig', 'kk': 'крик тілі', 'kk-Cyrl': 'крик тілі', 'km': 'គ្រីក', 'kn': 'ಕ್ರೀಕ್', 'ko': '크리크어', 'kok': 'मस्कोजी', 'kok-Deva': 'मस्कोजी', 'ks': 'کریٖک', 'ks-Arab': 'کریٖک', 'ksh': 'Krihk', 'ku': 'krîkî', 'ky': 'крикче', 'lb': 'Muskogee-Sprooch', 'lo': 'ມັສໂກກີ', 'lt': 'krykų', 'lv': 'krīku', 'mai': 'मसकोगी', 'mi': 'Mukōki', 'mk': 'крик', 'ml': 'ക്രീക്ക്', 'mn': 'крик', 'mr': 'क्रीक', 'ms': 'Creek', 'mt': 'Kriek', 'mus': 'Mvskoke', 'my': 'ခရိချ', 'nb': 'creek', 'ne': 'क्रिक', 'nl': 'Creek', 'nn': 'creek', 'no': 'creek', 'or': 'କ୍ରୀକ୍', 'pa': 'ਕ੍ਰੀਕ', 'pa-Guru': 'ਕ੍ਰੀਕ', 'pcm': 'Múskójii Lángwej', 'pl': 'krik', 'ps': 'کريکي', 'pt': 'creek', 'qu': 'Muscogee Simi', 'rm': 'creek', 'ro': 'creek', 'ru': 'крик', 'sc': 'muscogee', 'sd': 'ڪريڪ', 'sd-Arab': 'ڪريڪ', 'si': 'ක්\u200dරීක්', 'sk': 'kríkčina', 'sl': 'creekovščina', 'smn': 'muskogeekielâ', 'so': 'Muscogee', 'sq': 'krikisht', 'sr': 'кришки', 'sr-Cyrl': 'кришки', 'sr-Latn': 'kriški', 'sv': 'muskogee', 'sw': 'Kikriki', 'ta': 'க்ரீக்', 'te': 'క్రీక్', 'th': 'มัสคีกี', 'ti': 'ክሪክ', 'tk': 'krik dili', 'to': 'lea fakakileki', 'tr': 'Krikçe', 'ug': 'كىرىكچە', 'uk': 'крік', 'ur': 'کریک', 'uz': 'krik', 'uz-Cyrl': 'крикча', 'uz-Latn': 'krik', 'vec': 'crek', 'vi': 'Tiếng Creek', 'xnr': 'क्रीक', 'yo': 'Èdè Muskogi', 'yrl': 'kirik', 'yue': '克里克文', 'yue-Hans': '克里克文', 'yue-Hant': '克里克文', 'zh': '克里克语', 'zh-Hans': '克里克语', 'zh-Hant': '克里克文', 'zu': 'isi-Creek'},
    'mut': {'en': 'Western Muria'},
    'muu': {'en': 'Yaaku'},
    'muv': {'en': 'Muthuvan'},
    'mux': {'en': 'Bo-Ung'},
    'muy': {'en': 'Muyang'},
    'muz': {'en': 'Mursi'},
    'mva': {'en': 'Manam'},
    'mvb': {'en': 'Mattole'},
    'mvd': {'en': 'Mamboru'},
    'mve': {'en': 'Marwari (Pakistan)'},
    'mvf': {'en': 'Peripheral Mongolian'},
    'mvg': {'en': 'Yucuañe Mixtec'},
    'mvh': {'en': 'Mulgi'},
    'mvi': {'en': 'Miyako'},
    'mvk': {'en': 'Mekmek'},
    'mvl': {'en': 'Mbara (Australia)'},
    'mvm': {'en': 'Muya'},
    'mvn': {'en': 'Minaveha'},
    'mvo': {'en': 'Marovo'},
    'mvp': {'en': 'Duri'},
    'mvq': {'en': 'Moere'},
    'mvr': {'en': 'Marau'},
    'mvs': {'en': 'Massep'},
    'mvt': {'en': 'Mpotovoro'},
    'mvu': {'en': 'Marfa'},
    'mvv': {'en': 'Tagal Murut'},
    'mvw': {'en': 'Machinga'},
    'mvx': {'en': 'Meoswar'},
    'mvy': {'en': 'Indus Kohistani'},
    'mvz': {'en': 'Mesqan'},
    'mwa': {'en': 'Mwatebu'},
    'mwb': {'en': 'Juwal'},
    'mwc': {'en': 'Are'},
    'mwd': {'en': 'Mudbura'},
    'mwe': {'en': 'Mwera (Chimwera)'},
    'mwf': {'en': 'Murrinh-Patha'},
    'mwg': {'en': 'Aiklep'},
    'mwh': {'en': 'Mouk-Aria'},
    'mwi': {'en': 'Labo'},
    'mwj': {'en': 'Maligo'},
    'mwk': {'en': 'Kita Maninkakan'},
    'mwl': {'af': 'Mirandees', 'am': 'ሚራንዴዝ', 'ar': 'الميرانديز', 'as': 'মিৰাণ্ডিজ', 'ast': 'mirandés', 'az': 'mirand', 'az-Cyrl': 'миранд', 'az-Latn': 'mirand', 'be': 'мірандыйская', 'be-tarask': 'мірандыйская', 'bg': 'мирандийски', 'bgn': 'میراندیسی', 'bn': 'মিরান্ডিজ', 'br': 'mirandeg', 'brx': 'मिराण्डीज', 'bs': 'mirandeški', 'bs-Cyrl': 'мирандешки', 'bs-Latn': 'mirandeški', 'ca': 'mirandès', 'ca-ES-valencia': 'mirandès', 'ccp': '𑄟𑄨𑄢𑄚𑄴𑄓𑄨𑄎𑄴', 'ce': 'мирандойн', 'chr': 'ᎻᎳᏕᏏ', 'ckb': 'میراندی', 'cs': 'mirandština', 'cy': 'Mirandeg', 'da': 'mirandesisk', 'de': 'Mirandesisch', 'dsb': 'mirandezišćina', 'el': 'Μιραντεζικά', 'el-polyton': 'Μιραντεζικά', 'en': 'Mirandese', 'es': 'mirandés', 'et': 'miranda', 'eu': 'mirandesa', 'fa': 'میراندی', 'ff-Adlm': '𞤃𞤭𞤪𞤢𞤲𞤣𞤫𞥅𞤪𞤫', 'fi': 'mirandeesi', 'fil': 'Mirandese', 'fo': 'mirandesiskt', 'fr': 'mirandais', 'fur': 'mirandês', 'fy': 'Mirandees', 'ga': 'Mioraindéis', 'gd': 'Miorandais', 'gl': 'mirandés', 'gsw': 'Mirandesisch', 'gu': 'મિરાંડી', 'ha': 'Mirandese', 'he': 'מירנדזית', 'hi': 'मिरांडी', 'hi-Latn': 'Mirandese', 'hr': 'mirandski', 'hsb': 'mirandezišćina', 'ht': 'mirandais', 'hu': 'mirandéz', 'hy': 'միրանդերեն', 'ia': 'mirandese', 'id': 'Miranda', 'ig': 'Mirandese', 'is': 'mirandesíska', 'it': 'mirandese', 'ja': 'ミランダ語', 'jv': 'Mirandese', 'ka': 'მირანდული', 'kgp': 'mĩrỹnej', 'kk': 'миранд тілі', 'kk-Cyrl': 'миранд тілі', 'km': 'មីរ៉ានដេស', 'kn': 'ಮಿರಾಂಡೀಸ್', 'ko': '미란데어', 'kok': 'मिरांडीस', 'kok-Deva': 'मिरांडीस', 'ks': 'مِراندیٖز', 'ks-Arab': 'مِراندیٖز', 'ksh': 'Mirandehsesch', 'ku': 'mîrandî', 'ky': 'мирандизче', 'lb': 'Mirandesesch', 'lo': 'ມີລັນດາ', 'lt': 'mirandezų', 'lv': 'mirandiešu', 'mai': 'मिरंडीज', 'mi': 'Miranatīhi', 'mk': 'мирандски', 'ml': 'മിരാൻറസേ', 'mn': 'меранди', 'mr': 'मिरांडिज्', 'ms': 'Mirandese', 'mt': 'Mirandiż', 'my': 'မီရန်ဒီးဇ်', 'nb': 'mirandesisk', 'ne': 'मिरान्डी', 'nl': 'Mirandees', 'nn': 'mirandesisk', 'no': 'mirandesisk', 'or': 'ମିରାଣ୍ଡିଜ୍', 'pa': 'ਮਿਰਾਂਡੀ', 'pa-Guru': 'ਮਿਰਾਂਡੀ', 'pcm': 'Mirándẹ́sẹ Lángwej', 'pl': 'mirandyjski', 'ps': 'مرانديز', 'pt': 'mirandês', 'qu': 'Mirandés Simi', 'rm': 'mirandais', 'ro': 'mirandeză', 'ru': 'мирандский', 'sc': 'mirandesu', 'sd': 'مرانڊيز', 'sd-Arab': 'مرانڊيز', 'si': 'මිරන්ඩීස්', 'sk': 'mirandčina', 'sl': 'mirandeščina', 'smn': 'mirandeskielâ', 'so': 'Mirandese', 'sq': 'mirandisht', 'sr': 'мирандски', 'sr-Cyrl': 'мирандски', 'sr-Latn': 'mirandski', 'sv': 'mirandesiska', 'sw': 'Kimirandi', 'ta': 'மிரான்டீஸ்', 'te': 'మిరాండిస్', 'th': 'มีรันดา', 'ti': 'ሚራንዲዝኛ', 'tk': 'mirand dili', 'to': 'lea fakamilanitēsi', 'tr': 'Miranda dili', 'ug': 'مىراندېسچە', 'uk': 'мірандська', 'ur': 'میرانڈیز', 'uz': 'miranda', 'uz-Cyrl': 'мирандес', 'uz-Latn': 'miranda', 'vec': 'mirandeze', 'vi': 'Tiếng Miranda', 'xnr': 'मिरांडी', 'yo': 'Èdè Mirandisi', 'yrl': 'mirãdei', 'yue': '米蘭德斯文', 'yue-Hans': '米兰德斯文', 'yue-Hant': '米蘭德斯文', 'zh': '米兰德斯语', 'zh-Hans': '米兰德斯语', 'zh-Hant': '米蘭德斯文', 'zu': 'isi-Mirandese'},
    'mwm': {'en': 'Sar'},
    'mwn': {'en': 'Nyamwanga'},
    'mwo': {'en': 'Central Maewo'},
    'mwp': {'en': 'Kala Lagaw Ya'},
    'mwq': {'en': 'Mün Chin'},
    'mwr': {'ar': 'الماروارية', 'ast': 'marwari', 'az': 'maruari', 'az-Latn': 'maruari', 'bg': 'марвари', 'bn': 'মারোয়ারি', 'brx': 'मारवाड़ी', 'bs': 'marvari', 'bs-Cyrl': 'марвари', 'bs-Latn': 'marvari', 'ca': 'marwari', 'ca-ES-valencia': 'marwari', 'ccp': '𑄟𑄢𑄮𑄠𑄢𑄨', 'cs': 'márvárština', 'cy': 'Marwari', 'da': 'marwari', 'de': 'Marwari', 'el': 'Μαργουάρι', 'el-polyton': 'Μαργουάρι', 'en': 'Marwari', 'es': 'marwari', 'et': 'marvari', 'fa': 'مارواری', 'fi': 'marwari', 'fr': 'marwarî', 'fy': 'Marwari', 'ga': 'Marmhairis', 'gd': 'Marwari', 'gsw': 'Marwarisch', 'gu': 'મારવાડી', 'he': 'מרווארי', 'hi': 'मारवाड़ी', 'hi-Latn': 'Marwari', 'hr': 'marwari', 'ht': 'marwarî', 'hu': 'márvári', 'id': 'Marwari', 'is': 'marvarí', 'it': 'marwari', 'ja': 'マールワーリー語', 'ka': 'მარვარი', 'kgp': 'mỹrvari', 'kn': 'ಮಾರ್ವಾಡಿ', 'ko': '마르와리어', 'ks': 'مارواڑی', 'ks-Arab': 'مارواڑی', 'lb': 'Marwari', 'lo': 'ມາວາຣິ', 'lt': 'marvari', 'lv': 'marvaru', 'mk': 'марвари', 'ml': 'മർവാരി', 'mr': 'मारवाडी', 'mt': 'Marwari', 'nb': 'marwari', 'ne': 'माडवारी', 'nl': 'Marwari', 'nn': 'marwari', 'no': 'marwari', 'or': 'ମାରୱାରୀ', 'pl': 'marwari', 'pt': 'marwari', 'rm': 'marwari', 'ro': 'marwari', 'ru': 'марвари', 'sk': 'marwari', 'sl': 'marvarščina', 'sr': 'марвари', 'sr-Cyrl': 'марвари', 'sr-Latn': 'marvari', 'sv': 'marwari', 'ta': 'மார்வாரி', 'te': 'మార్వాడి', 'th': 'มารวาฑี', 'to': 'lea fakamaliwali', 'tr': 'Marvari', 'ug': 'مارۋارىچە', 'uk': 'марварі', 'vi': 'Tiếng Marwari', 'yrl': 'mariwari', 'yue': '馬爾尼裡文', 'yue-Hans': '马尔尼里文', 'yue-Hant': '馬爾尼裡文', 'zh': '马尔瓦里语', 'zh-Hans': '马尔瓦里语', 'zh-Hant': '馬瓦里文'},
    'mws': {'en': 'Mwimbi-Muthambi'},
    'mwt': {'en': 'Moken'},
    'mwu': {'en': 'Mittu'},
    'mwv': {'ast': 'mentawai', 'cs': 'mentavajština', 'de': 'Mentawai', 'en': 'Mentawai', 'et': 'mentavei', 'fi': 'mentawai', 'fr': 'mentawaï', 'gd': 'Mentawai', 'hi-Latn': 'Mentawai', 'ht': 'mentawaï', 'id': 'Mentawai', 'it': 'mentawai', 'ja': 'メンタワイ語', 'lb': 'Mentawai', 'lt': 'mentavai', 'nb': 'mentawai', 'ne': 'मेन्टावाई', 'nl': 'Mentawai', 'nn': 'mentawai', 'no': 'mentawai', 'pl': 'mentawai', 'sv': 'mentawai', 'th': 'เม็นตาไว', 'to': 'lea fakamenitauai', 'tr': 'Mentawai', 'yue': '明打威文', 'yue-Hans': '明打威文', 'yue-Hant': '明打威文', 'zh-Hant': '明打威文'},
    'mww': {'en': 'Hmong Daw'},
    'mwx': {'en': 'Mediak'},
    'mwy': {'en': 'Mosiro'},
    'mwz': {'en': 'Moingi'},
    'mxa': {'en': 'Northwest Oaxaca Mixtec'},
    'mxb': {'en': 'Tezoatlán Mixtec'},
    'mxc': {'en': 'Manyika'},
    'mxd': {'en': 'Modang'},
    'mxe': {'en': 'Mele-Fila'},
    'mxf': {'en': 'Malgbe'},
    'mxg': {'en': 'Mbangala'},
    'mxh': {'en': 'Mvuba'},
    'mxi': {'en': 'Mozarabic'},
    'mxj': {'en': 'Miju-Mishmi'},
    'mxk': {'en': 'Monumbo'},
    'mxl': {'en': 'Maxi Gbe'},
    'mxm': {'en': 'Meramera'},
    'mxn': {'en': 'Moi (Indonesia)'},
    'mxo': {'en': 'Mbowe'},
    'mxp': {'en': 'Tlahuitoltepec Mixe'},
    'mxq': {'en': 'Juquila Mixe'},
    'mxr': {'en': 'Murik (Malaysia)'},
    'mxs': {'en': 'Huitepec Mixtec'},
    'mxt': {'en': 'Jamiltepec Mixtec'},
    'mxu': {'en': 'Mada (Cameroon)'},
    'mxv': {'en': 'Metlatónoc Mixtec'},
    'mxw': {'en': 'Namo'},
    'mxx': {'en': 'Mahou'},
    'mxy': {'en': 'Southeastern Nochixtlán Mixtec'},
    'mxz': {'en': 'Central Masela'},
    'my': {'af': 'Birmaans', 'agq': 'Bùumɛsɛ̀', 'ak': 'Bɛɛmis kasa', 'am': 'ቡርማኛ', 'ar': 'البورمية', 'as': 'বাৰ্মীজ', 'asa': 'Kiburma', 'ast': 'birmanu', 'az': 'birman', 'az-Arab': 'بۇرماجا', 'az-Cyrl': 'бирман', 'az-Latn': 'birman', 'bal-Latn': 'Barmái', 'bas': 'Hɔp u birmàn', 'be': 'бірманская', 'be-tarask': 'бірманская', 'bem': 'Ichi Burma', 'bez': 'Hiburma', 'bg': 'бирмански', 'bgn': 'بورمي', 'bm': 'birimanikan', 'bn': 'বর্মি', 'br': 'birmaneg', 'brx': 'बार्मिस', 'bs': 'burmanski', 'bs-Cyrl': 'бурмански', 'bs-Latn': 'burmanski', 'ca': 'birmà', 'ca-ES-valencia': 'birmà', 'ccp': '𑄝𑄧𑄢𑄴𑄟𑄨', 'ce': 'бирманийн', 'cgg': 'Oruburuma', 'chr': 'ᏋᎻᏍ', 'ckb': 'میانماری', 'cs': 'barmština', 'cy': 'Byrmaneg', 'da': 'burmesisk', 'dav': 'Kiburma', 'de': 'Birmanisch', 'dje': 'Burme senni', 'dsb': 'burmašćina', 'dyo': 'birmani', 'dz': 'བར་མིས་ཁ', 'ebu': 'Kĩburma', 'ee': 'burmagbe', 'el': 'Βιρμανικά', 'el-polyton': 'Βιρμανικά', 'en': 'Burmese', 'en-Dsrt': '𐐒𐐲𐑉𐑋𐐨𐑆', 'eo': 'birma', 'es': 'birmano', 'et': 'birma', 'eu': 'birmaniera', 'ewo': 'Ǹkɔ́bɔ birəmán', 'fa': 'برمه\u200cای', 'ff': 'Burmeese', 'ff-Adlm': '𞤄𞤵𞤪𞤥𞤢𞥄𞤪𞤫', 'ff-Latn': 'Burmeese', 'fi': 'burma', 'fil': 'Burmese', 'fo': 'burmesiskt', 'fr': 'birman', 'fy': 'Birmees', 'ga': 'Burmais', 'gd': 'Burmais', 'gl': 'birmano', 'gsw': 'Birmanisch', 'gu': 'બર્મીઝ', 'guz': 'Kiburma', 'ha': 'Burmanci', 'he': 'בורמזית', 'hi': 'बर्मीज़', 'hi-Latn': 'Burmese', 'hr': 'burmanski', 'hsb': 'burmašćina', 'ht': 'birman', 'hu': 'burmai', 'hy': 'բիրմայերեն', 'ia': 'birmano', 'id': 'Burma', 'ig': 'Burmese', 'is': 'búrmneska', 'it': 'birmano', 'ja': 'ミャンマー語', 'jmc': 'Kyiburma', 'jv': 'Myanmar', 'ka': 'ბირმული', 'kab': 'Taburmisit', 'kam': 'Kiburma', 'kde': 'Chibulma', 'kea': 'birmanes', 'kgp': 'mirmỹnẽj', 'khq': 'Burme senni', 'ki': 'Kiburma', 'kk': 'бирма тілі', 'kk-Cyrl': 'бирма тілі', 'kln': 'kutitab Burma', 'km': 'ភូមា', 'kn': 'ಬರ್ಮೀಸ್', 'ko': '버마어', 'kok': 'बर्मीज', 'kok-Deva': 'बर्मीज', 'ks': 'بٔمیٖز', 'ks-Arab': 'بٔمیٖز', 'ksb': 'Kibulma', 'ksf': 'ribirmán', 'ksh': 'Burmehsesch', 'ku': 'burmayî', 'kxv': 'burmij', 'kxv-Deva': 'बर्मीज़', 'kxv-Latn': 'burmij', 'kxv-Orya': 'ବର୍ମୀଜ୍', 'kxv-Telu': 'బర్మీస్', 'ky': 'бурмача', 'lag': 'Kɨbáama', 'lb': 'Birmanesch', 'lg': 'Lubbama', 'lkt': 'Burmese Iyápi', 'lld': 'birmanich', 'ln': 'libilimá', 'lo': 'ມຽນມາ', 'lrc': 'بئرمئ یی', 'lt': 'birmiečių', 'luo': 'Kiburma', 'luy': 'Kiburma', 'lv': 'birmiešu', 'mai': 'बर्मी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lBurma', 'mer': 'Kĩburma', 'mfe': 'birman', 'mg': 'Birmana', 'mgh': 'Iburma', 'mi': 'Pēmīhi', 'mk': 'бурмански', 'ml': 'ബർമീസ്', 'mn': 'бирм', 'mr': 'बर्मी', 'ms': 'Burma', 'mt': 'Burmiż', 'mua': 'birmania', 'my': 'မြန်မာ', 'mzn': 'برمه\u200cیی', 'naq': 'Burmesǁî gowab', 'nb': 'burmesisk', 'nd': 'isi-Burma', 'ne': 'बर्मेली', 'nl': 'Birmaans', 'nmg': 'Kiɛl birmania', 'nn': 'burmesisk', 'no': 'burmesisk', 'nqo': 'ߓߙߌߡߊ߲ߞߊ߲', 'nus': 'Thok bormi̱thni', 'nyn': 'Oruburuma', 'om': 'Burmeesee', 'or': 'ବର୍ମୀଜ୍', 'pa': 'ਬਰਮੀ', 'pa-Guru': 'ਬਰਮੀ', 'pcm': 'Bọ́ma Lángwej', 'pl': 'birmański', 'ps': 'برمایی', 'pt': 'birmanês', 'qu': 'Birmano Simi', 'rm': 'birman', 'rn': 'Ikinyabirimaniya', 'ro': 'birmană', 'rof': 'Kiburma', 'ru': 'бирманский', 'rwk': 'Kyiburma', 'saq': 'Kiburma', 'sbp': 'Ishibuluma', 'sc': 'burmesu', 'sd': 'برمي', 'sd-Arab': 'برمي', 'se': 'burmagiella', 'seh': 'birmanês', 'ses': 'Burme senni', 'sg': 'Miamära, Birimäni', 'shi': 'ⵜⴰⴱⵉⵔⵎⴰⵏⵉⵜ', 'shi-Latn': 'Tabirmanit', 'shi-Tfng': 'ⵜⴰⴱⵉⵔⵎⴰⵏⵉⵜ', 'si': 'බුරුම', 'sk': 'barmčina', 'sl': 'burmanščina', 'smn': 'burmakielâ', 'sn': 'chiBurma', 'so': 'Burmese', 'sq': 'birmanisht', 'sr': 'бурмански', 'sr-Cyrl': 'бурмански', 'sr-Latn': 'burmanski', 'sv': 'burmesiska', 'sw': 'Kiburma', 'ta': 'பர்மீஸ்', 'te': 'బర్మీస్', 'teo': 'Kiburma', 'tg': 'бирманӣ', 'th': 'พม่า', 'ti': 'በርምኛ', 'tk': 'birma dili', 'to': 'lea fakapema', 'tr': 'Birman dili', 'tt': 'бирма', 'twq': 'Burme senni', 'tzm': 'Taburmanit', 'ug': 'بىرماچە', 'uk': 'бірманська', 'ur': 'برمی', 'uz': 'birman', 'uz-Cyrl': 'бирманча', 'uz-Latn': 'birman', 'vai': 'ꗩꕆꔻ', 'vai-Latn': 'Bhɛmísi', 'vai-Vaii': 'ꗩꕆꔻ', 'vec': 'birman', 'vi': 'Tiếng Miến Điện', 'vun': 'Kyiburma', 'wae': 'Burmesiš', 'wo': 'Birmes', 'xnr': 'बर्मीज़', 'xog': 'Olubbama', 'yav': 'bímanɛ', 'yi': 'בירמאַניש', 'yo': 'Èdè Bumiisi', 'yrl': 'birimanei', 'yue': '緬甸文', 'yue-Hans': '缅甸文', 'yue-Hant': '緬甸文', 'zgh': 'ⵜⴰⴱⵉⵔⵎⴰⵏⵉⵜ', 'zh': '缅甸语', 'zh-Hans': '缅甸语', 'zh-Hant': '緬甸文', 'zu': 'isi-Burmese'},
    'myb': {'en': 'Mbay'},
    'myc': {'en': 'Mayeka'},
    'myd': {'en': 'Maramba'},
    'mye': {'ast': 'myene', 'bs': 'mjene', 'bs-Latn': 'mjene', 'ca': 'myene', 'ca-ES-valencia': 'myene', 'cs': 'myene', 'da': 'myene', 'de': 'Myene', 'el': 'Μιένε', 'el-polyton': 'Μιένε', 'en': 'Myene', 'es': 'myene', 'et': 'mjene', 'fi': 'myene', 'fr': 'myènè', 'fy': 'Myene', 'gd': 'Myene', 'he': 'מאיין', 'hi-Latn': 'Myene', 'hr': 'myene', 'ht': 'myènè', 'hu': 'myene', 'id': 'Myene', 'it': 'myene', 'ja': 'ミエネ語', 'ka': 'მიენე', 'kgp': 'myene', 'ko': '미예네어', 'lb': 'Myene', 'lo': 'ມໍຢິນ', 'lt': 'mjenų', 'lv': 'mjenu', 'mk': 'мјене', 'ml': 'മയീൻ', 'ms': 'Myene', 'nb': 'myene', 'ne': 'म्येने', 'nl': 'Myene', 'nn': 'myene', 'no': 'myene', 'pl': 'myene', 'pt': 'myene', 'ro': 'myene', 'ru': 'миене', 'sk': 'myene', 'sv': 'myene', 'th': 'มยีน', 'to': 'lea fakamiene', 'tr': 'Myene', 'ug': 'ميېنېچە', 'uk': 'миін', 'vi': 'Tiếng Myene', 'yrl': 'miyene', 'yue': '姆耶內文', 'yue-Hans': '姆耶内文', 'yue-Hant': '姆耶內文', 'zh': '姆耶内语', 'zh-Hans': '姆耶内语', 'zh-Hant': '姆耶內文'},
    'myf': {'en': 'Bambassi'},
    'myg': {'en': 'Manta'},
    'myh': {'en': 'Makah'},
    'myi': {'en': 'Mina (India)'},
    'myj': {'en': 'Mangayat'},
    'myk': {'en': 'Mamara Senoufo'},
    'myl': {'en': 'Moma'},
    'mym': {'en': "Me'en"},
    'myn': {'en': 'Mayan languages'},
    'myn-pro': {'en': 'Proto-Mayan'},
    'myo': {'en': 'Anfillo'},
    'myp': {'en': 'Pirahã'},
    'myq': {'en': 'Forest Maninka'},
    'myr': {'en': 'Muniche'},
    'mys': {'en': 'Mesmes'},
    'myt': {'en': 'Sangab Mandaya'},
    'myu': {'en': 'Mundurukú'},
    'myv': {'af': 'Erzya', 'am': 'ኤርዝያ', 'ar': 'الأرزية', 'as': 'এৰজিয়া', 'ast': 'erzya', 'az': 'erzya', 'az-Cyrl': 'ерзја', 'az-Latn': 'erzya', 'bal-Latn': 'Erziá', 'be': 'эрзянская', 'be-tarask': 'эрзянская', 'bg': 'ерзиа', 'bgn': 'ارزیای', 'bn': 'এরজিয়া', 'br': 'erza', 'brx': 'एर्जीया', 'bs': 'erzija', 'bs-Cyrl': 'ерзија', 'bs-Latn': 'erzija', 'ca': 'mordovià erza', 'ca-ES-valencia': 'mordovià erza', 'ccp': '𑄆𑄢𑄧𑄎𑄨𑄠', 'ce': 'эрзянийн', 'chr': 'ᎡᏏᏯ', 'ckb': 'ئێرزیا', 'cs': 'erzjanština', 'cy': 'Erzya', 'da': 'erzya', 'de': 'Ersja-Mordwinisch', 'dsb': 'erzyašćina', 'el': 'Έρζια', 'el-polyton': 'Έρζια', 'en': 'Erzya', 'es': 'erzya', 'et': 'ersa', 'eu': 'erziera', 'fa': 'ارزیایی', 'ff-Adlm': '𞤉𞤪𞤶𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'ersä', 'fil': 'Erzya', 'fo': 'erzya', 'fr': 'erzya', 'fy': 'Erzja', 'ga': 'Éirsis', 'gd': 'Erzya', 'gl': 'erzya', 'gsw': 'Erzya', 'gu': 'એર્ઝયા', 'ha': 'Erzya', 'he': 'ארזיה', 'hi': 'एर्ज़या', 'hi-Latn': 'Erzya', 'hr': 'mordvinski', 'hsb': 'erzyašćina', 'ht': 'erzya', 'hu': 'erzjány', 'hy': 'էրզյա', 'ia': 'erzya', 'id': 'Eryza', 'ig': 'Erzya', 'is': 'ersja', 'it': 'erzya', 'ja': 'エルジャ語', 'jv': 'Ersia', 'ka': 'ერზია', 'kgp': 'érsija', 'kk': 'эрзя тілі', 'kk-Cyrl': 'эрзя тілі', 'km': 'អឺហ្ស៊ីយ៉ា', 'kn': 'ಎರ್ಝ್ಯಾ', 'ko': '엘즈야어', 'kok': 'एरझिया', 'kok-Deva': 'एरझिया', 'ks': 'ایرزِیا', 'ks-Arab': 'ایرزِیا', 'ksh': 'Ärsjahnesch', 'ku': 'erzayî', 'ky': 'эрзянча', 'lb': 'Ersja-Mordwinesch', 'lo': 'ເອີຍາ', 'lt': 'erzyjų', 'lv': 'erzju', 'mai': 'अरजाया', 'mi': 'Erehīa', 'mk': 'ерзјански', 'ml': 'ഏഴ്സ്യ', 'mn': 'эрзя', 'mr': 'एर्झ्या', 'ms': 'Erzya', 'mt': 'Erzya', 'my': 'အီဇယာ', 'myv': 'эрзянь кель', 'nb': 'erzia', 'ne': 'इर्ज्या', 'nl': 'Erzja', 'nn': 'erzia', 'no': 'erzia', 'or': 'ଏର୍ଜୟା', 'pa': 'ਇਰਜ਼ੀਆ', 'pa-Guru': 'ਇਰਜ਼ੀਆ', 'pcm': 'Ẹrziá Lángwej', 'pl': 'erzja', 'ps': 'ارزيا', 'pt': 'erzya', 'qu': 'Erzya Simi', 'rm': 'erzya', 'ro': 'erzya', 'ru': 'эрзянский', 'sc': 'erzya', 'sd': 'ايريزيا', 'sd-Arab': 'ايريزيا', 'se': 'ersagiella', 'si': 'එර්ස්යා', 'sk': 'erzjančina', 'sl': 'erzjanščina', 'smn': 'ersäkielâ', 'so': 'Erzya', 'sq': 'erzjaisht', 'sr': 'ерзја', 'sr-Cyrl': 'ерзја', 'sr-Latn': 'erzja', 'sv': 'erjya', 'sw': 'Kierzya', 'ta': 'ஏர்ஜியா', 'te': 'ఎర్జియా', 'th': 'เอียร์ซยา', 'ti': 'ኤርዝያ', 'tk': 'erzýan dili', 'to': 'lea fakaʻelisia', 'tr': 'Erzya', 'ug': 'ئېرزاچە', 'uk': 'ерзя', 'ur': 'ارزیا', 'uz': 'erzya', 'uz-Cyrl': 'эрзянча', 'uz-Latn': 'erzya', 'vec': 'erzyen', 'vi': 'Tiếng Erzya', 'xnr': 'एर्ज़या', 'yo': 'Èdè Esiya', 'yrl': 'erisia', 'yue': '厄爾茲亞文', 'yue-Hans': '厄尔兹亚文', 'yue-Hant': '厄爾茲亞文', 'zh': '厄尔兹亚语', 'zh-Hans': '厄尔兹亚语', 'zh-Hant': '厄爾茲亞文', 'zu': 'isi-Erzya'},
    'myw': {'en': 'Muyuw'},
    'myx': {'en': 'Masaaba'},
    'myy': {'en': 'Macuna'},
    'myz': {'en': 'Classical Mandaic'},
    'mza': {'en': 'Santa María Zacatepec Mixtec'},
    'mzb': {'en': 'Tumzabt'},
    'mzc': {'en': 'Madagascar Sign Language'},
    'mzd': {'en': 'Malimba'},
    'mze': {'en': 'Morawa'},
    'mzg': {'en': 'Monastic Sign Language'},
    'mzh': {'en': 'Wichí Lhamtés Güisnay'},
    'mzi': {'en': 'Ixcatlán Mazatec'},
    'mzj': {'en': 'Manya'},
    'mzk': {'en': 'Nigeria Mambila'},
    'mzl': {'en': 'Mazatlán Mixe'},
    'mzm': {'en': 'Mumuye'},
    'mzn': {'af': 'Masanderani', 'am': 'ማዛንደራኒ', 'ar': 'المازندرانية', 'as': 'মাজেন্দাৰানি', 'ast': 'mazanderani', 'az': 'mazandaran', 'az-Arab': 'مازنیجه', 'az-Cyrl': 'мазандаран', 'az-Latn': 'mazandaran', 'bal-Latn': 'Mázendaráni', 'be': 'мазандэранская', 'be-tarask': 'мазандэранская', 'bg': 'мазандерански', 'bgn': 'مازندرانی', 'bn': 'মাজানদেরানি', 'br': 'mazanderaneg', 'brx': 'माजान्देरानि', 'bs': 'mazanderanski', 'bs-Cyrl': 'мазандерани', 'bs-Latn': 'mazanderanski', 'ca': 'mazanderani', 'ca-ES-valencia': 'mazanderani', 'ccp': '𑄟𑄎𑄚𑄴𑄘𑄬𑄢𑄚𑄨', 'ce': 'мазандеранхойн', 'chr': 'ᎹᏌᏕᎳᏂ', 'ckb': 'مازەندەرانی', 'cs': 'mázandaránština', 'cy': 'Masanderani', 'da': 'mazenisk', 'de': 'Masanderanisch', 'dsb': 'mazanderanšćina', 'el': 'Μαζαντεράνι', 'el-polyton': 'Μαζαντεράνι', 'en': 'Mazanderani', 'es': 'mazandaraní', 'et': 'mazandaraani', 'eu': 'mazandarandera', 'fa': 'مازندرانی', 'ff-Adlm': '𞤃𞤢𞥁𞤢𞤲𞤣𞤫𞤪𞤢𞤲𞤭𞥅𞤪𞤫', 'fi': 'mazandarani', 'fil': 'Mazanderani', 'fo': 'mazanderani', 'fr': 'mazandérani', 'ga': 'Mázandaráinis', 'gd': 'Mazanderani', 'gl': 'mazandaraní', 'gu': 'મઝાન્દેરાની', 'ha': 'Mazanderani', 'he': 'מאזאנדראני', 'hi': 'माज़न्देरानी', 'hi-Latn': 'Mazanderani', 'hr': 'mazanderanski', 'hsb': 'mazanderanšćina', 'ht': 'mazandérani', 'hu': 'mázanderáni', 'hy': 'մազանդարաներեն', 'ia': 'mazanderani', 'id': 'Mazanderani', 'ig': 'Mazanderani', 'is': 'masanderaní', 'it': 'mazandarani', 'ja': 'マーザンダラーン語', 'jv': 'Mazanderani', 'ka': 'მაზანდერანული', 'kgp': 'mỹsánarỹni', 'kk': 'мазандеран тілі', 'kk-Cyrl': 'мазандеран тілі', 'km': 'ម៉ាហ្សានដឺរេនី', 'kn': 'ಮಜಂದೆರಾನಿ', 'ko': '마잔데라니어', 'kok': 'मझांडेराणी', 'kok-Deva': 'मझांडेराणी', 'ku': 'mazenderanî', 'ky': 'мазандераниче', 'lb': 'Mazandarani', 'lo': 'ມາແຊນເດີລັງ', 'lrc': 'مازأندأرانی', 'lt': 'mazenderanių', 'lv': 'mazanderāņu', 'mai': 'मजनदेरानी', 'mi': 'Mahaterani', 'mk': 'мазендерански', 'ml': 'മസന്ററാനി', 'mn': 'мазандерани', 'mr': 'माझानदेरानी', 'ms': 'Mazanderani', 'mt': 'Mazanderani', 'my': 'မာဇန်ဒါရန်နီ', 'mzn': 'مازرونی', 'nb': 'mazandarani', 'ne': 'मजानडेरानी', 'nl': 'Mazanderani', 'nn': 'mazanderani', 'no': 'mazandarani', 'nqo': 'ߡߊߖ߭ߊ߲ߘߋߙߊߣߞߊ߲', 'or': 'ମାଜାନଡେରାନି', 'pa': 'ਮੇਜ਼ੈਂਡਰਾਨੀ', 'pa-Guru': 'ਮੇਜ਼ੈਂਡਰਾਨੀ', 'pcm': 'Mazandẹrání Lángwej', 'pl': 'mazanderański', 'ps': 'مزاندراني', 'pt': 'mazandarani', 'qu': 'Mazandaraní Simi', 'rm': 'mazanderani', 'ro': 'mazanderani', 'ru': 'мазандеранский', 'sc': 'mazandarani', 'sd': 'مزيندراني', 'sd-Arab': 'مزيندراني', 'si': 'මැසන්ඩරනි', 'sk': 'mázandaránčina', 'sl': 'mazanderanščina', 'smn': 'mazandarani', 'so': 'Masanderaani', 'sq': 'mazanderanisht', 'sr': 'мазандерански', 'sr-Cyrl': 'мазандерански', 'sr-Latn': 'mazanderanski', 'sv': 'mazanderani', 'sw': 'Kimazanderani', 'ta': 'மசந்தேரனி', 'te': 'మాసన్\u200cదెరాని', 'th': 'มาซันดารานี', 'ti': 'ማዛንደራኒ', 'tk': 'mazanderan dili', 'to': 'lea fakamasanitelani', 'tr': 'Mazenderanca', 'uk': 'мазандеранська', 'ur': 'مزندرانی', 'uz': 'mozandaron', 'uz-Cyrl': 'мазандеран', 'uz-Latn': 'mozandaron', 'vec': 'mazandarani', 'vi': 'Tiếng Mazanderani', 'xnr': 'माज़न्देरानी', 'yo': 'Masanderani', 'yrl': 'masãdarani', 'yue': '馬贊德蘭文', 'yue-Hans': '马赞德兰文', 'yue-Hant': '馬贊德蘭文', 'zh': '马赞德兰语', 'zh-Hans': '马赞德兰语', 'zh-Hant': '馬贊德蘭文', 'zu': 'isi-Mazanderani'},
    'mzo': {'en': 'Matipuhy'},
    'mzp': {'en': 'Movima'},
    'mzq': {'en': 'Mori Atas'},
    'mzr': {'en': 'Marúbo'},
    'mzs': {'en': 'Macanese'},
    'mzt': {'en': 'Mintil'},
    'mzu': {'en': 'Inapang'},
    'mzv': {'en': 'Manza'},
    'mzw': {'en': 'Deg'},
    'mzx': {'en': 'Mawayana'},
    'mzy': {'en': 'Mozambican Sign Language'},
    'mzz': {'en': 'Maiadomu'},
    'na': {'af': 'Nauru', 'am': 'ናኡሩ', 'ar': 'النورو', 'as': 'নাউৰু', 'ast': 'nauru', 'az': 'nauru', 'az-Cyrl': 'науру', 'az-Latn': 'nauru', 'be': 'науру', 'be-tarask': 'науру', 'bg': 'науру', 'bgn': 'نائورویی', 'bn': 'নাউরু', 'br': 'naurueg', 'brx': 'नाऊरू', 'bs': 'nauru', 'bs-Cyrl': 'науру', 'bs-Latn': 'nauru', 'ca': 'nauruà', 'ca-ES-valencia': 'nauruà', 'ccp': '𑄚𑄃𑄪𑄢𑄪', 'ce': 'науру', 'chr': 'ᏃᎤᎷ', 'ckb': 'نائوروو', 'cs': 'naurština', 'cy': 'Nawrŵeg', 'da': 'nauru', 'de': 'Nauruisch', 'dsb': 'naurušćina', 'el': 'Ναούρου', 'el-polyton': 'Ναούρου', 'en': 'Nauru', 'eo': 'naura', 'es': 'nauruano', 'et': 'nauru', 'eu': 'nauruera', 'fa': 'نائورویی', 'ff-Adlm': '𞤐𞤢𞤱𞤵𞤪𞤵𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'nauru', 'fil': 'Nauru', 'fo': 'nauru', 'fr': 'nauruan', 'fy': 'Nauruaansk', 'ga': 'Nárúis', 'gd': 'Nabhru', 'gl': 'nauruano', 'gsw': 'Nauruisch', 'gu': 'નાઉરૂ', 'ha': 'Nauru', 'he': 'נאורית', 'hi': 'नाउरू', 'hi-Latn': 'Nauru', 'hr': 'nauru', 'hsb': 'naurušćina', 'ht': 'nauruan', 'hu': 'naurui', 'hy': 'նաուրու', 'ia': 'nauru', 'id': 'Nauru', 'ig': 'Nauru', 'is': 'nárúska', 'it': 'nauru', 'ja': 'ナウル語', 'jv': 'Nauru', 'ka': 'ნაურუ', 'kgp': 'nỹvuruvánũ', 'kk': 'науру тілі', 'kk-Cyrl': 'науру тілі', 'km': 'ណូរូ', 'kn': 'ನೌರು', 'ko': '나우루어', 'kok': 'नौरू', 'kok-Deva': 'नौरू', 'ks': 'ناورُ', 'ks-Arab': 'ناورُ', 'ksh': 'Nauruhesch', 'ku': 'nawrûyî', 'ky': 'науруча', 'lb': 'Nauruesch', 'lo': 'ນາຢູລູ', 'lt': 'naurų', 'lv': 'nauruiešu', 'mai': 'नउरू', 'mi': 'Nauru', 'mk': 'науруански', 'ml': 'നൗറു', 'mn': 'науру', 'mr': 'नउरू', 'ms': 'Nauru', 'mt': 'Naurujan', 'my': 'နော်ရူး', 'nb': 'nauru', 'ne': 'नाउरू', 'nl': 'Nauruaans', 'nn': 'nauru', 'no': 'nauru', 'or': 'ନାଉରୁ', 'pa': 'ਨਾਉਰੂ', 'pa-Guru': 'ਨਾਉਰੂ', 'pcm': 'Naúru Lángwej', 'pl': 'nauruański', 'ps': 'نایرو', 'pt': 'nauruano', 'qu': 'Nauru Simi', 'rm': 'nauru', 'ro': 'nauru', 'ru': 'науру', 'sc': 'nauru', 'sd': 'نائو', 'sd-Arab': 'نائو', 'si': 'නෞරු', 'sk': 'nauruština', 'sl': 'naurujščina', 'smn': 'naurukielâ', 'so': 'Nauru', 'sq': 'nauruisht', 'sr': 'науруски', 'sr-Cyrl': 'науруски', 'sr-Latn': 'nauruski', 'sv': 'nauruanska', 'sw': 'Kinauru', 'ta': 'நவ்ரூ', 'te': 'నౌరు', 'th': 'นาอูรู', 'ti': 'ናውርዋንኛ', 'tk': 'nauru dili', 'to': 'lea fakanaulu', 'tr': 'Nauru dili', 'ug': 'ناۋرۇچە', 'uk': 'науру', 'ur': 'ناؤرو', 'uz': 'nauru', 'uz-Latn': 'nauru', 'vec': 'nauruan', 'vi': 'Tiếng Nauru', 'xnr': 'नाउरू', 'yo': 'Èdè Nauru', 'yrl': 'nauruanu', 'yue': '諾魯文', 'yue-Hans': '诺鲁文', 'yue-Hant': '諾魯文', 'zh': '瑙鲁语', 'zh-Hans': '瑙鲁语', 'zh-Hant': '諾魯文', 'zu': 'isi-Nauru'},
    'naa': {'en': 'Namla'},
    'nab': {'en': 'Southern Nambikuára'},
    'nac': {'en': 'Narak'},
    'nad': {'en': 'Nijadali'},
    'nae': {'en': "Naka'ela"},
    'naf': {'en': 'Nabak'},
    'nag': {'en': 'Naga Pidgin'},
    'nah': {'en': 'Nahuatl languages', 'nah': 'Nāhuatl'},
    'nai': {'en': 'North American Indian languages'},
    'nai-chu-pro': {'en': 'Proto-Chumash'},
    'nai-ckn-pro': {'en': 'Proto-Chinookan'},
    'nai-kat': {'en': 'Kathlamet'},
    'nai-mdu-pro': {'en': 'Proto-Maidun'},
    'nai-pic': {'en': 'Picuris'},
    'nai-pom-pro': {'en': 'Proto-Pomo'},
    'nai-yuc-tip': {'en': 'Tipai'},
    'nai-yuc-yav': {'en': 'Yavapai'},
    'naj': {'en': 'Nalu'},
    'nak': {'en': 'Nakanai'},
    'nal': {'en': 'Nalik'},
    'nam': {'en': "Ngan'gityemerri"},
    'nan': {'af': 'Min Nan-Sjinees', 'am': 'ሚን ኛን ቻይንኛ', 'ar': 'مين-نان الصينية', 'ast': 'chinu min nan', 'az': 'Min Nan', 'az-Latn': 'Min Nan', 'br': 'sinaeg Min Nan', 'ca': 'xinès min del sud', 'ca-ES-valencia': 'xinès min del sud', 'ccp': '𑄚𑄚𑄴', 'cs': 'čínština (dialekty Minnan)', 'da': 'min-kinesisk', 'de': 'Min Nan', 'en': 'Min Nan Chinese', 'es': 'chino min nan', 'et': 'lõunamini', 'fi': 'min nan -kiina', 'fo': 'min nan kinesiskt', 'fr': 'minnan', 'ga': 'Sínis Min Nan', 'gd': 'Min Nan', 'he': 'סינית מין נאן', 'hi': 'मिन नान', 'hi-Latn': 'Min Nan', 'hr': 'min nan kineski', 'ht': 'minnan', 'hu': 'min nan kínai', 'it': 'min nan', 'ja': '閩南語', 'kgp': 'mĩn nỹn', 'kn': 'ನಾನ್', 'ko': '민난어', 'lb': 'Min-Nan-Chinesesch', 'lt': 'kinų kalbos pietų minų tarmė', 'mk': 'јужномински', 'ml': 'മിൻ നാൻ ചൈനീസ്', 'mr': 'मिन नान चिनी', 'ms': 'Cina Min Nan', 'nb': 'minnan', 'ne': 'मिन नान चिनियाँ', 'nl': 'Minnanyu', 'nn': 'minnan', 'no': 'minnan', 'pa': 'ਚੀਨੀ ਮਿਨ ਨਾਨ', 'pa-Guru': 'ਚੀਨੀ ਮਿਨ ਨਾਨ', 'pl': 'minnański', 'pt': 'min nan', 'ro': 'chineză min nan', 'ru': 'миньнань', 'si': 'මින් නන් චයිනිස්', 'sl': 'min nan kitajščina', 'sv': 'min nan', 'ta': 'மின் நான் சீனம்', 'te': 'మిన్ నాన్ చైనీస్', 'th': 'จีนมินหนาน', 'to': 'lea fakasiaina-mininani', 'tr': 'Min Nan Çincesi', 'uk': 'південноміньська', 'vi': 'Tiếng Mân Nam', 'yrl': 'mĩ nã', 'yue': '閩南語', 'yue-Hans': '闽南语', 'yue-Hant': '閩南語', 'zh': '闽南语', 'zh-Hans': '闽南语', 'zh-Hant': '閩南語', 'zu': 'isi-Min Nan Chinese', 'nan-Latn': 'Bân-lâm-gú'},
    'nao': {'en': 'Naaba'},
    'nap': {'af': 'Neapolitaans', 'am': 'ኒአፖሊታን', 'ar': 'النابولية', 'as': 'নিয়াপোলিটেন', 'ast': 'napolitanu', 'az': 'neapolitan', 'az-Cyrl': 'неаполитан', 'az-Latn': 'neapolitan', 'be': 'неапалітанская', 'be-tarask': 'неапалітанская', 'bg': 'неаполитански', 'bgn': 'ناپلی', 'bn': 'নেয়াপোলিটান', 'br': 'napolitaneg', 'brx': 'नियापलिटान', 'bs': 'napolitanski', 'bs-Cyrl': 'неаполитански', 'bs-Latn': 'napolitanski', 'ca': 'napolità', 'ca-ES-valencia': 'napolità', 'ccp': '𑄚𑄬𑄠𑄛𑄮𑄣𑄨𑄑𑄚𑄴', 'ce': 'неаполитанойн', 'chr': 'ᏂᏯᏆᎵᏔᏂ', 'ckb': 'ناپۆلی', 'cs': 'neapolština', 'cy': 'Naplieg', 'da': 'napolitansk', 'de': 'Neapolitanisch', 'dsb': 'neapolitanšćina', 'el': 'Ναπολιτανικά', 'el-polyton': 'Ναπολιτανικά', 'en': 'Neapolitan', 'en-Dsrt': '𐐤𐐨𐐲𐐹𐐱𐑊𐐮𐐻𐐲𐑌', 'es': 'napolitano', 'et': 'napoli', 'eu': 'napoliera', 'fa': 'ناپلی', 'ff-Adlm': '𞤐𞤢𞥄𞤨𞤮𞤤𞤭𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'napoli', 'fil': 'Neapolitan', 'fo': 'napolitanskt', 'fr': 'napolitain', 'fur': 'napoletan', 'fy': 'Napolitaansk', 'ga': 'Napóilis', 'gd': 'Eadailtis Napoli', 'gl': 'napolitano', 'gsw': 'Neapolitanisch', 'gu': 'નેપોલિટાન', 'ha': 'Neapolitan', 'he': 'נפוליטנית', 'hi': 'नीपोलिटन', 'hi-Latn': 'Neapolitan', 'hr': 'napolitanski', 'hsb': 'neapolitanšćina', 'ht': 'napolitain', 'hu': 'nápolyi', 'hy': 'նեապոլերեն', 'ia': 'napolitano', 'id': 'Neapolitan', 'ig': 'Neapolitan', 'is': 'napólíska', 'it': 'napoletano', 'ja': 'ナポリ語', 'jv': 'Neapolitan', 'ka': 'ნეაპოლიტანური', 'kgp': 'nỹporitỹnũ', 'kk': 'неаполитан тілі', 'kk-Cyrl': 'неаполитан тілі', 'km': 'នាប៉ូលីតាន', 'kn': 'ನಿಯಾಪೊಲಿಟನ್', 'ko': '나폴리어', 'kok': 'नियापोलिटन', 'kok-Deva': 'नियापोलिटन', 'ks': 'نیٖپالیٹَن', 'ks-Arab': 'نیٖپالیٹَن', 'ksh': 'Napolletahnesch', 'ku': 'napolîtanî', 'ky': 'неополитанча', 'lb': 'Neapolitanesch', 'lo': 'ນາໂປລີ', 'lt': 'neapoliečių', 'lv': 'neapoliešu', 'mai': 'नेपोलिटन', 'mi': 'Neaporitana', 'mk': 'неаполски', 'ml': 'നെപ്പോളിറ്റാൻ', 'mn': 'неаполитан', 'mr': 'नेपोलिटान', 'ms': 'Neapolitan', 'mt': 'Naplitan', 'my': 'နပိုလီတန်', 'nb': 'napolitansk', 'ne': 'नेपोलिटान', 'nl': 'Napolitaans', 'nn': 'napolitansk', 'no': 'napolitansk', 'or': 'ନୀପୋଲିଟାନ୍', 'pa': 'ਨਿਆਪੋਲੀਟਨ', 'pa-Guru': 'ਨਿਆਪੋਲੀਟਨ', 'pcm': 'Niapolítán Lángwej', 'pl': 'neapolitański', 'ps': 'نيپالين', 'pt': 'napolitano', 'qu': 'Neapolitan Simi', 'rm': 'neapolitan', 'ro': 'napolitană', 'ru': 'неаполитанский', 'sc': 'napoletanu', 'sd': 'نيپولٽن', 'sd-Arab': 'نيپولٽن', 'si': 'නියාපොලිටන්', 'sk': 'neapolčina', 'sl': 'napolitanščina', 'smn': 'napolikielâ', 'so': 'Neapolitan', 'sq': 'napoletanisht', 'sr': 'напуљски', 'sr-Cyrl': 'напуљски', 'sr-Latn': 'napuljski', 'sv': 'napolitanska', 'sw': 'Kinapoli', 'ta': 'நியோபோலிடன்', 'te': 'నియాపోలిటన్', 'th': 'นาโปลี', 'ti': 'ኒያፖሊታንኛ', 'tk': 'neapolitan dili', 'to': 'lea fakanapoletano', 'tr': 'Napolice', 'ug': 'ناپولىچە', 'uk': 'неаполітанська', 'ur': 'نیاپولیٹن', 'uz': 'neapolitan', 'uz-Latn': 'neapolitan', 'vec': 'napoletan', 'vi': 'Tiếng Napoli', 'xnr': 'नीपोलिटन', 'yi': 'נאַפּאליטַניש', 'yo': 'Èdè Neapolita', 'yrl': 'napuritanu', 'yue': '拿波里文', 'yue-Hans': '拿波里文', 'yue-Hant': '拿波里文', 'zh': '那不勒斯语', 'zh-Hans': '那不勒斯语', 'zh-Hant': '拿波里文', 'zu': 'isi-Neapolitan'},
    'naq': {'af': 'Nama', 'am': 'ናማ', 'ar': 'لغة الناما', 'as': 'নামা', 'ast': 'nama', 'az': 'nama', 'az-Cyrl': 'нама', 'az-Latn': 'nama', 'bal-Latn': 'Nama', 'be': 'нама', 'be-tarask': 'нама', 'bg': 'нама', 'bgn': 'نامایی', 'bn': 'নামা', 'br': 'nama', 'brx': 'नामा', 'bs': 'nama', 'bs-Cyrl': 'нама', 'bs-Latn': 'nama', 'ca': 'nama', 'ca-ES-valencia': 'nama', 'ccp': '𑄚𑄟', 'ce': 'нама', 'chr': 'ᎾᎹ', 'ckb': 'ناما', 'cs': 'namaština', 'cy': 'Nama', 'da': 'nama', 'de': 'Nama', 'dsb': 'nama', 'el': 'Νάμα', 'el-polyton': 'Νάμα', 'en': 'Nama', 'es': 'nama', 'et': 'nama', 'eu': 'namera', 'fa': 'نامایی', 'ff-Adlm': '𞤐𞤢𞤥𞤢𞥄𞤪𞤫', 'fi': 'nama', 'fil': 'Nama', 'fo': 'nama', 'fr': 'nama', 'fy': 'Nama', 'ga': 'Nama', 'gd': 'Nama', 'gl': 'nama', 'gu': 'નમા', 'ha': 'Nama', 'he': 'נאמה', 'hi': 'नामा', 'hi-Latn': 'Nama', 'hr': 'nama', 'hsb': 'nama', 'ht': 'nama', 'hu': 'nama', 'hy': 'նամա', 'ia': 'nama', 'id': 'Nama', 'ig': 'Nama', 'is': 'nama', 'it': 'nama', 'ja': 'ナマ語', 'jv': 'Nama', 'ka': 'ნამა', 'kgp': 'nỹmỹ', 'kk': 'нама тілі', 'kk-Cyrl': 'нама тілі', 'km': 'ណាម៉ា', 'kn': 'ನಮ', 'ko': '나마어', 'kok': 'नामा', 'kok-Deva': 'नामा', 'ku': 'namayî', 'ky': 'намача', 'lb': 'Nama', 'lo': 'ນາມາ', 'lrc': 'ناما', 'lt': 'nama', 'lv': 'nama', 'mai': 'नामा', 'mi': 'Nama', 'mk': 'нама', 'ml': 'നാമ', 'mn': 'нама', 'mr': 'नामा', 'ms': 'Nama', 'mt': 'Nama', 'my': 'နာမာ', 'mzn': 'ناما', 'naq': 'Khoekhoegowab', 'nb': 'nama', 'ne': 'नामा', 'nl': 'Nama', 'nn': 'nama', 'no': 'nama', 'nqo': 'ߣߡߊߞߊ߲', 'or': 'ନାମା', 'pa': 'ਨਾਮਾ', 'pa-Guru': 'ਨਾਮਾ', 'pcm': 'Naámá Lángwej', 'pl': 'nama', 'ps': 'ناما', 'pt': 'nama', 'qu': 'Nama Simi', 'rm': 'nama', 'ro': 'nama', 'ru': 'нама', 'sc': 'nama', 'sd': 'ناما', 'sd-Arab': 'ناما', 'si': 'නාමා', 'sk': 'nama', 'sl': 'khoekhoe', 'smn': 'nama', 'so': 'Nama', 'sq': 'namaisht', 'sr': 'нама', 'sr-Cyrl': 'нама', 'sr-Latn': 'nama', 'sv': 'nama', 'sw': 'Kinama', 'ta': 'நாமா', 'te': 'నమ', 'th': 'นามา', 'ti': 'ናማ', 'tk': 'nama dili', 'to': 'lea fakanama', 'tr': 'Nama', 'ug': 'ناماچە', 'uk': 'нама', 'ur': 'ناما', 'uz': 'nama', 'uz-Cyrl': 'нама', 'uz-Latn': 'nama', 'vec': 'nama', 'vi': 'Tiếng Nama', 'xnr': 'नामा', 'yo': 'Námà', 'yrl': 'nama', 'yue': '納馬文', 'yue-Hans': '纳马文', 'yue-Hant': '納馬文', 'zh': '纳马语', 'zh-Hans': '纳马语', 'zh-Hant': '納馬文', 'zu': 'isi-Nama'},
    'nar': {'en': 'Iguta'},
    'nas': {'en': 'Naasioi'},
    'nat': {'en': 'Ca̱hungwa̱rya̱'},
    'naw': {'en': 'Nawuri'},
    'nax': {'en': 'Nakwi'},
    'nay': {'en': 'Ngarrindjeri'},
    'naz': {'en': 'Coatepec Nahuatl'},
    'nb': {'af': 'Boeknoors', 'am': 'የኖርዌይ ቦክማል', 'ar': 'النرويجية بوكمال', 'as': 'নৰৱেজিয়ান বোকমাল', 'ast': 'noruegu Bokmål', 'az': 'bokmal norveç', 'az-Cyrl': 'бокмал норвеч', 'az-Latn': 'bokmal norveç', 'bal-Latn': 'Nárwiji Bokmál', 'be': 'нарвежская (букмол)', 'be-tarask': 'нарвежская (букмол)', 'bg': 'норвежки (букмол)', 'bgn': 'نارویی بوک\u200cمولی', 'bn': 'নরওয়েজিয়ান বোকমাল', 'br': 'norvegeg bokmål', 'brx': 'नरवैजियान बकमाल', 'bs': 'norveški (Bokmal)', 'bs-Cyrl': 'норвешки бокмал', 'bs-Latn': 'norveški (Bokmal)', 'ca': 'noruec bokmål', 'ca-ES-valencia': 'noruec bokmål', 'ccp': '𑄚𑄧𑄢𑄴𑄃𑄮𑄠𑄬𑄎𑄨𑄠𑄚𑄴 𑄝𑄮𑄇𑄴𑄟𑄣𑄴', 'ce': 'норвегийн букмол', 'chr': 'ᏃᎵᏪᏥᏂ ᏉᎧᎹᎵ', 'ckb': 'نەرویژیی بۆکمال', 'cs': 'norština (bokmål)', 'cy': 'Norwyeg Bokmål', 'da': 'bokmål', 'de': 'Norwegisch (Bokmål)', 'dsb': 'norwegske bokmål', 'dz': 'ནོར་ཝེ་ཇི་ཡཱན་བོཀ་མཱལ་ཁ', 'ee': 'nɔweigbe bokmål', 'el': 'Νορβηγικά Μποκμάλ', 'el-polyton': 'Νορβηγικά Μποκμάλ', 'en': 'Norwegian Bokmål', 'eo': 'dannorvega', 'es': 'noruego bokmal', 'et': 'norra bokmål', 'eu': 'bokmål (norvegiera)', 'fa': 'نروژی بوک\u200cمُل', 'ff-Adlm': '𞤐𞤮𞤪𞤱𞤫𞤶𞤭𞤴𞤢𞤲𞤪𞤫 𞤄𞤮𞤳𞤥𞤢𞤤', 'fi': 'norjan bokmål', 'fil': 'Norwegian Bokmål', 'fo': 'norskt bókmál', 'fr': 'norvégien bokmål', 'fur': 'norvegjês bokmål', 'fy': 'Noors - Bokmål', 'ga': 'Bocmál', 'gd': 'Bokmål na Nirribhidh', 'gl': 'noruegués bokmål', 'gsw': 'Norwegisch Bokmål', 'gu': 'નોર્વેજિયન બોકમાલ', 'ha': 'Norwegian Bokmål', 'he': 'נורווגית ספרותית', 'hi': 'नॉर्वेजियाई बोकमाल', 'hi-Latn': 'Norwegian Bokmal', 'hr': 'norveški bokmål', 'hsb': 'norwegšćina (bokmål)', 'ht': 'norvégien bokmål', 'hu': 'norvég (bokmål)', 'hy': 'գրքային նորվեգերեն', 'ia': 'norvegiano bokmål', 'id': 'Bokmål Norwegia', 'ig': 'Norwegian Bokmål', 'is': 'norskt bókmál', 'it': 'norvegese bokmål', 'ja': 'ノルウェー語(ブークモール)', 'jv': 'Bokmål Norwegia', 'ka': 'ნორვეგიული ბუკმოლი', 'kea': 'norueges bokmål', 'kgp': 'mógmỹr nãrovegej', 'kk': 'норвегиялық букмол тілі', 'kk-Cyrl': 'норвегиялық букмол тілі', 'km': 'ន័រវែស បុកម៉ាល់', 'kn': 'ನಾರ್ವೆಜಿಯನ್ ಬೊಕ್ಮಲ್', 'ko': '노르웨이어(보크말)', 'kok': 'नॉर्वेजियन बोकमाल', 'kok-Deva': 'नॉर्वेजियन बोकमाल', 'ks': 'ناروییَن بوکمال', 'ks-Arab': 'ناروییَن بوکمال', 'ksh': 'Norrwehjesch Bokmål', 'ku': 'norwecî (bokmål)', 'kxv': 'norvejiati būkmal', 'kxv-Deva': 'नॉर्वेजियाति बोकमाल', 'kxv-Latn': 'norvejiati būkmal', 'kxv-Orya': 'ନରୱେଜିଆନ୍ ବୋକମଲ୍', 'kxv-Telu': 'సార్వేజియన్ బొకమల్', 'ky': 'норвежче (букмал)', 'lb': 'Norwegesch Bokmål', 'lld': 'norvegesc bokmål', 'lo': 'ນໍເວຈຽນ ບັອກມອລ', 'lrc': 'نورڤئجی بوٙکمال', 'lt': 'norvegų bukmolas', 'lv': 'norvēģu bukmols', 'mai': 'नार्वेजियन बोकमाल', 'mi': 'Pakamō Nōwei', 'mk': 'норвешки букмол', 'ml': 'നോർവീജിയൻ ബുക്\u200cമൽ', 'mn': 'норвегийн букмол', 'mr': 'नॉर्वेजियन बोकमाल', 'ms': 'Bokmal Norway', 'mt': 'Bokmal Norveġiż', 'my': 'နော်ဝေ ဘွတ်ခ်မော်လ်', 'mzn': 'نروژی بوکمال', 'nb': 'norsk bokmål', 'ne': 'नर्वेली बोकमाल', 'nl': 'Noors - Bokmål', 'nn': 'norsk bokmål', 'no': 'norsk bokmål', 'nqo': 'ߣߐߙߑߝ߭ߍߖ߭ ߓߏߞߑߡߊߟߑߞߊ߲', 'or': 'ନରୱେଜିଆନ୍ ବୋକମଲ୍', 'pa': 'ਨਾਰਵੇਜਿਆਈ ਬੋਕਮਲ', 'pa-Guru': 'ਨਾਰਵੇਜਿਆਈ ਬੋਕਮਲ', 'pcm': 'Nọwẹ́jiá Bokmál Lángwej', 'pl': 'norweski (bokmål)', 'ps': 'ناروېئي (بوکمال)', 'pt': 'bokmål norueguês', 'qu': 'Noruego Bokmål Simi', 'rm': 'norvegais bokmål', 'ro': 'norvegiană bokmål', 'ru': 'норвежский букмол', 'sc': 'norvegesu bokmål', 'sd': 'نارويائي بوڪمال', 'sd-Arab': 'نارويائي بوڪمال', 'se': 'girjedárogiella', 'si': 'නෝර්වීජියානු බොක්මල්', 'sk': 'nórčina (bokmal)', 'sl': 'knjižna norveščina', 'smn': 'tárukielâ bokmål', 'so': 'Nawrijii Bokmål', 'sq': 'norvegjishte letrare', 'sr': 'норвешки букмол', 'sr-Cyrl': 'норвешки букмол', 'sr-Latn': 'norveški bukmol', 'sv': 'norskt bokmål', 'sw': 'Kinorwe cha Bokmal', 'ta': 'நார்வேஜியன் பொக்மால்', 'te': 'నార్వేజియన్ బొక్మాల్', 'th': 'นอร์เวย์บุคมอล', 'ti': 'ኖርወያዊ ቦክማል', 'tk': 'norwegiýa bukmol dili', 'to': 'lea fakanouaē-pokimali', 'tr': 'Norveççe Bokmål', 'ug': 'نورۋىگىيە بوكمالچە', 'uk': 'норвезька (букмол)', 'ur': 'نارویجین بوکمل', 'uz': 'norveg-bokmal', 'uz-Cyrl': 'норвегча бокмал', 'uz-Latn': 'norveg-bokmal', 'vec': 'norvejeze bokmal', 'vi': 'Tiếng Na Uy (Bokmål)', 'wae': 'Norwegiš Bokmål', 'xnr': 'नॉर्वेजियाई बोकमाल', 'yo': 'Nọ́ọ́wè Bokímàl', 'yrl': 'bukimau nuruegei', 'yue': '巴克摩挪威文', 'yue-Hans': '巴克摩挪威文', 'yue-Hant': '巴克摩挪威文', 'zh': '书面挪威语', 'zh-Hans': '书面挪威语', 'zh-Hant': '書面挪威文', 'zu': 'isi-Norwegian Bokmål'},
    'nba': {'en': 'Nyemba'},
    'nbb': {'en': 'Ndoe'},
    'nbc': {'en': 'Chang Naga'},
    'nbd': {'en': 'Ngbinda'},
    'nbe': {'en': 'Konyak Naga'},
    'nbf': {'en': 'Naxi'},
    'nbg': {'en': 'Nagarchal'},
    'nbh': {'en': 'Ngamo'},
    'nbi': {'en': 'Mao Naga'},
    'nbj': {'en': 'Ngarinyman'},
    'nbk': {'en': 'Nake'},
    'nbm': {'en': "Ngbaka Ma'bo"},
    'nbn': {'en': 'Kuri'},
    'nbo': {'en': 'Nkukoli'},
    'nbp': {'en': 'Nnam'},
    'nbq': {'en': 'Nggem'},
    'nbr': {'en': 'Numana'},
    'nbs': {'en': 'Namibian Sign Language'},
    'nbt': {'en': 'Na'},
    'nbu': {'en': 'Rongmei Naga'},
    'nbv': {'en': 'Ngamambo'},
    'nbw': {'en': 'Southern Ngbandi'},
    'nbx': {'en': 'Ngura'},
    'nby': {'en': 'Ningera'},
    'nca': {'en': 'Iyo'},
    'ncb': {'en': 'Central Nicobarese'},
    'ncc': {'en': 'Ponam'},
    'ncd': {'en': 'Nachering'},
    'nce': {'en': 'Yale'},
    'ncf': {'en': 'Notsi'},
    'ncg': {'en': "Nisga'a"},
    'nch': {'en': 'Central Huasteca Nahuatl'},
    'nci': {'en': 'Classical Nahuatl'},
    'ncj': {'en': 'Northern Puebla Nahuatl'},
    'nck': {'en': 'Na-kara'},
    'ncl': {'en': 'Michoacán Nahuatl'},
    'ncm': {'en': 'Nambo'},
    'ncn': {'en': 'Nauna'},
    'nco': {'en': 'Sibe'},
    'ncp': {'en': 'Ndaktup'},
    'ncq': {'en': 'Northern Katang'},
    'ncr': {'en': 'Ncane'},
    'ncs': {'en': 'Nicaraguan Sign Language'},
    'nct': {'en': 'Chothe Naga'},
    'ncu': {'en': 'Chumburung'},
    'ncx': {'en': 'Central Puebla Nahuatl'},
    'ncz': {'en': 'Natchez'},
    'nd': {'af': 'Noord-Ndebele', 'am': 'ሰሜን ንዴብሌ', 'ar': 'النديبيل الشمالية', 'as': 'উত্তৰ নিবেবেলে', 'ast': 'ndebele del norte', 'az': 'şimali ndebele', 'az-Cyrl': 'шимали ндебеле', 'az-Latn': 'şimali ndebele', 'bal-Latn': 'Shemáli Nedébéle', 'be': 'паўночная ндэбеле', 'be-tarask': 'паўночная ндэбеле', 'bg': 'северен ндебеле', 'bgn': 'بُرزسرین انده\u200cبله\u200cای', 'bn': 'উত্তর এন্দেবেলে', 'br': 'ndebele an Norzh', 'brx': 'साहा न्देबेले', 'bs': 'sjeverni ndebele', 'bs-Cyrl': 'сјеверни ндебеле', 'bs-Latn': 'sjeverni ndebele', 'ca': 'ndebele septentrional', 'ca-ES-valencia': 'ndebele septentrional', 'ccp': '𑄅𑄖𑄴𑄖𑄧𑄢𑄴 𑄆𑄚𑄴𑄘𑄬𑄝𑄨𑄣𑄨', 'ce': 'къилбаседа ндебели', 'chr': 'ᏧᏴᏢ ᏂᏕᏇᎴ', 'ckb': 'ئندێبێلێی باکوور', 'cs': 'ndebele (Zimbabwe)', 'cy': 'Ndebele Gogleddol', 'da': 'nordndebele', 'de': 'Nord-Ndebele', 'dsb': 'pódpołnocne ndebele', 'ee': 'dziehe ndebelegbe', 'el': 'Βόρεια Ντεμπέλε', 'el-polyton': 'Βόρεια Ντεμπέλε', 'en': 'North Ndebele', 'es': 'ndebele septentrional', 'et': 'põhjandebele', 'eu': 'iparraldeko ndebeleera', 'fa': 'انده\u200cبله\u200cای شمالی', 'ff-Adlm': '𞤐𞤣𞤫𞤦𞤫𞤤𞤫𞥅𞤪𞤫 𞤐𞤢𞤲𞥆𞤢𞥄𞤺𞤫', 'fi': 'pohjois-ndebele', 'fil': 'Hilagang Ndebele', 'fo': 'norður ndebele', 'fr': 'ndébélé du Nord', 'fur': 'ndebele setentrionâl', 'fy': 'Noard-Ndbele', 'ga': 'N-deibéilis an Tuaiscirt', 'gd': 'Ndebele Thuathach', 'gl': 'ndebele setentrional', 'gsw': 'Nord-Ndebele-Schpraach', 'gu': 'ઉત્તર દેબેલ', 'ha': 'North Ndebele', 'he': 'נדבלה צפונית', 'hi': 'उत्तरी देबेल', 'hi-Latn': 'North Ndebele', 'hr': 'sjeverni ndebele', 'hsb': 'sewjero-ndebele', 'ht': 'ndébélé du Nord', 'hu': 'északi ndebele', 'hy': 'հյուսիսային նդեբելե', 'ia': 'ndebele del nord', 'id': 'Ndebele Utara', 'ig': 'North Ndebele', 'is': 'norður-ndebele', 'it': 'ndebele del nord', 'ja': '北ンデベレ語', 'jv': 'Ndebele Lor', 'ka': 'ჩრდილოეთ ნდებელე', 'kgp': 'nemere nãrti', 'kk': 'солтүстік ндебеле тілі', 'kk-Cyrl': 'солтүстік ндебеле тілі', 'km': 'នេបេលេខាងជើង', 'kn': 'ಉತ್ತರ ದೆಬೆಲೆ', 'ko': '북부 은데벨레어', 'kok': 'उत्तर डेबेले', 'kok-Deva': 'उत्तर डेबेले', 'ks': 'شُمال ڈَبیل', 'ks-Arab': 'شُمال ڈَبیل', 'ksh': 'Nood-Ndebele', 'ku': 'ndebelîya bakurî', 'ky': 'түндүк ндыбелче', 'lb': 'Nord-Ndebele-Sprooch', 'lld': 'ndebele dl nord', 'lo': 'ເອັນເດເບເລເໜືອ', 'lrc': 'نئدئبئلئ شومالی', 'lt': 'šiaurės ndebelų', 'lv': 'ziemeļndebelu', 'mai': 'उत्तरी डेबले', 'mi': 'Enetepēra ki te Raki', 'mk': 'северен ндебеле', 'ml': 'നോർത്ത് ഡെബിൾ', 'mn': 'хойд ндебеле', 'mr': 'उत्तर देबेली', 'ms': 'Ndebele Utara', 'mt': 'Ndebeli tat-Tramuntana', 'my': 'မြောက် အွန်န်ဒီဘီလီ', 'mzn': 'شمالی ندبله', 'nb': 'nord-ndebele', 'nd': 'isiNdebele', 'ne': 'उत्तरी न्डेबेले', 'nl': 'Noord-Ndebele', 'nn': 'nord-ndebele', 'no': 'nord-ndebele', 'nqo': 'ߕߟߋ߬ߓߐ ߒߘߓߋߟߋߞߊ߲', 'or': 'ଉତ୍ତର ନେଡବେଲେ', 'pa': 'ਉੱਤਰੀ ਨਡੇਬੇਲੇ', 'pa-Guru': 'ਉੱਤਰੀ ਨਡੇਬੇਲੇ', 'pcm': 'Nọ́tán Ndẹbẹlẹ Lángwej', 'pl': 'ndebele północny', 'ps': 'شمالي نديبل', 'pt': 'ndebele do norte', 'qu': 'Ndebele septentrional Simi', 'rm': 'ndebele dal nord', 'ro': 'ndebele de nord', 'ru': 'северный ндебеле', 'sc': 'ndebele de su nord', 'sd': 'اتر دبيلي', 'sd-Arab': 'اتر دبيلي', 'si': 'උතුරු එන්ඩිබෙලෙ', 'sk': 'ndebelčina (severná)', 'sl': 'severna ndebelščina', 'smn': 'tave-nbedele', 'so': 'Indhebeele Waqooyi', 'sq': 'ndebelishte veriore', 'sr': 'северни ндебеле', 'sr-Cyrl': 'северни ндебеле', 'sr-Latn': 'severni ndebele', 'sv': 'nordndebele', 'sw': 'Kindebele cha Kaskazini', 'ta': 'வடக்கு தெபெலே', 'te': 'ఉత్తర దెబెలె', 'th': 'เอ็นเดเบเลเหนือ', 'ti': 'ሰሜን ኤንደበለ', 'tk': 'demirgazyk ndebele dili', 'to': 'lea fakanetepele-tokelau', 'tr': 'Kuzey Ndebele', 'ug': 'شىمالى ندەبەلەچە', 'uk': 'північна ндебеле', 'ur': 'شمالی دبیل', 'uz': 'shimoliy ndebele', 'uz-Cyrl': 'шимолий ндебеле', 'uz-Latn': 'shimoliy ndebele', 'vec': 'ndebele de’l nord', 'vi': 'Tiếng Ndebele Miền Bắc', 'wae': 'Nordndebele', 'xnr': 'उत्तरी देबेल', 'yo': 'Àríwá Ndebele', 'yrl': 'ĩdebere nutiwara', 'yue': '北地畢列文', 'yue-Hans': '北地毕列文', 'yue-Hant': '北地畢列文', 'zh': '北恩德贝勒语', 'zh-Hans': '北恩德贝勒语', 'zh-Hant': '北地畢列文', 'zu': 'isi-North Ndebele'},
    'nda': {'en': 'Ndasa'},
    'ndb': {'en': 'Kenswei Nsei'},
    'ndc': {'en': 'Ndau'},
    'ndd': {'en': 'Nde-Nsele-Nta'},
    'ndf': {'en': 'Nadruvian'},
    'ndg': {'en': 'Ndengereko'},
    'ndh': {'en': 'Ndali'},
    'ndi': {'en': 'Samba Leko'},
    'ndj': {'en': 'Ndamba'},
    'ndk': {'en': 'Ndaka'},
    'ndl': {'en': 'Ndolo'},
    'ndm': {'en': 'Ndam'},
    'ndn': {'en': 'Ngundi'},
    'ndp': {'en': 'Ndo'},
    'ndq': {'en': 'Ndombe'},
    'ndr': {'en': 'Ndoola'},
    'nds': {'af': 'Nederduits', 'ak': 'Gyaaman kasa a ɛwɔ fam', 'am': 'የታችኛው ጀርመንኛ', 'ar': 'الألمانية السفلى', 'as': 'ল’ জাৰ্মান', 'ast': 'baxu alemán', 'az': 'aşağı alman', 'az-Latn': 'aşağı alman', 'bal-Latn': 'Láw Jarman', 'be': 'ніжненямецкая', 'be-tarask': 'ніжненямецкая', 'bg': 'долнонемски', 'bn': 'নিম্ন জার্মানি', 'br': 'alamaneg izel', 'brx': 'गाहाय जार्मान', 'bs': 'donjonjemački', 'bs-Cyrl': 'ниски њемачки', 'bs-Latn': 'donjonjemački', 'ca': 'baix alemany', 'ca-ES-valencia': 'baix alemany', 'ccp': '𑄖𑄧𑄣𑄬 𑄎𑄢𑄴𑄟𑄚𑄨', 'ce': 'лахара германхойн', 'chr': 'ᎡᎳᏗ ᎠᏂᏓᏥ', 'cs': 'dolnoněmčina', 'cy': 'Almaeneg Isel', 'da': 'nedertysk', 'de': 'Niederdeutsch', 'dsb': 'dolnonimšćina', 'el': 'Κάτω Γερμανικά', 'el-polyton': 'Κάτω Γερμανικά', 'en': 'Low German', 'en-Dsrt': '𐐢𐐬 𐐖𐐯𐑉𐑋𐐲𐑌', 'es': 'bajo alemán', 'et': 'alamsaksa', 'eu': 'behe-alemana', 'fa': 'آلمانی سفلی', 'ff-Adlm': '𞤂𞤫𞤧-𞤀𞤤𞤵𞤥𞤢𞤲𞤪𞤫', 'fi': 'alasaksa', 'fil': 'Low German', 'fo': 'lágt týskt', 'fr': 'bas-allemand', 'fur': 'bas todesc', 'fy': 'Laagduits', 'ga': 'Gearmáinis Íochtarach', 'gd': 'Gearmailtis Ìochdarach', 'gl': 'baixo alemán', 'gsw': 'Nidertüütsch', 'gu': 'લો જર્મન', 'ha': 'Low German', 'he': 'גרמנית תחתית', 'hi': 'निचला जर्मन', 'hi-Latn': 'Low German', 'hr': 'donjonjemački', 'hsb': 'delnjoněmčina', 'ht': 'bas-allemand', 'hu': 'alsónémet', 'hy': 'ստորին գերմաներեն', 'ia': 'basse germano', 'id': 'Jerman Rendah', 'ig': 'Low German', 'ii': 'ꃅꄷꀁꂥꄓꉙ', 'is': 'lágþýska; lágsaxneska', 'it': 'basso tedesco', 'ja': '低地ドイツ語', 'jv': 'Jerman Non Standar', 'ka': 'ქვემოგერმანული', 'kea': 'alemon baxu', 'kgp': 'arimỹv rur', 'kk': 'төменгі неміс тілі', 'kk-Cyrl': 'төменгі неміс тілі', 'km': 'អាល្លឺម៉ង់ក្រោម', 'kn': 'ಲೋ ಜರ್ಮನ್', 'ko': '저지 독일어', 'kok': 'लोवर जर्मन', 'kok-Deva': 'लोवर जर्मन', 'ks': 'بۆنِم جٔرمَن', 'ks-Arab': 'بۆنِم جٔرمَن', 'ku': 'nedersaksî', 'ky': 'төмөнкү немисче', 'lb': 'Nidderdäitsch', 'lo': 'ເຢຍລະມັນ ຕອນໃຕ້', 'lrc': 'آلمانی ھاری', 'lt': 'Žemutinės Vokietijos', 'lv': 'lejasvācu', 'mai': 'लो जर्मन', 'mi': 'Tiamana Hakahaka', 'mk': 'долногермански', 'ml': 'ലോ ജർമൻ', 'mn': 'доод герман', 'mr': 'लो जर्मन', 'ms': 'Jerman Rendah', 'mt': 'Ġermaniż Komuni', 'my': 'အနိမ့် ဂျာမန်', 'mzn': 'پایین آلمانی', 'nb': 'nedertysk', 'nds': 'Neddersass’sch', 'ne': 'तल्लो जर्मन', 'nl': 'Nedersaksisch', 'nn': 'lågtysk', 'no': 'nedertysk', 'nqo': 'ߊߟߑߡߊ߲ߘߎ߯-ߓߊߛߑߞߊ߲', 'or': 'ଲୋ ଜର୍ମାନ୍', 'pa': 'ਲੋ ਜਰਮਨ', 'pa-Guru': 'ਲੋ ਜਰਮਨ', 'pcm': 'Ló Jámán Lángwej', 'pl': 'dolnoniemiecki', 'ps': 'کښته آلماني', 'pt': 'baixo alemão', 'qu': 'Bajo Alemán Simi', 'rm': 'bass tudestg', 'ro': 'germana de jos', 'ru': 'нижненемецкий', 'sc': 'tedescu bassu', 'sd': 'لو جرمن', 'sd-Arab': 'لو جرمن', 'si': 'පහළ ජර්මන්', 'sk': 'dolná nemčina', 'sl': 'nizka nemščina', 'so': 'Jarmal Hooseeya', 'sq': 'gjermanishte e vendeve të ulëta', 'sr': 'нисконемачки', 'sr-Cyrl': 'нисконемачки', 'sr-Latn': 'niskonemački', 'sv': 'lågtyska', 'sw': 'Kisaksoni', 'syr': 'ܐܠܡܢܝܐ ܠܐܠܬܚܬ', 'ta': 'லோ ஜெர்மன்', 'te': 'లో జర్మన్', 'th': 'เยอรมันต่ำ', 'ti': 'ትሑት ጀርመን', 'tk': 'aşaky nemes dili', 'to': 'lea fakasiamane-hifo', 'tr': 'Aşağı Almanca', 'ug': 'تۆۋەن گېرمانچە', 'uk': 'нижньонімецька', 'ur': 'ادنی جرمن', 'uz': 'quyi nemis', 'uz-Latn': 'quyi nemis', 'vec': 'baso todesco', 'vi': 'Tiếng Hạ Đức', 'xnr': 'बुरला जर्मन', 'yi': 'נידערדײַטש', 'yo': 'Jámánì ìpìlẹ̀', 'yrl': 'aremaũ yatuka', 'yue': '低地德文', 'yue-Hans': '低地德文', 'yue-Hant': '低地德文', 'zh': '低地德语', 'zh-Hans': '低地德语', 'zh-Hant': '低地德文', 'zu': 'isi-Low German'},
    'nds-NL': {'af': 'Nedersaksies', 'am': 'የታችኛው ሳክሰን', 'ar': 'السكسونية السفلى', 'ast': 'baxu saxón', 'az': 'aşağı sakson', 'az-Cyrl': 'ашағы саксон', 'az-Latn': 'aşağı sakson', 'bal-Latn': 'Láw Jarman (NL)', 'be': 'ніжнесаксонская', 'be-tarask': 'ніжнесаксонская', 'bg': 'долносаксонски', 'bgn': 'ساکسونی سفلی', 'bn': 'লো স্যাক্সন', 'br': 'saksoneg izel', 'bs': 'donjosaksonski', 'bs-Latn': 'donjosaksonski', 'ca': 'baix saxó', 'ca-ES-valencia': 'baix saxó', 'ccp': '𑄣𑄮𑄥𑄳𑄠𑄇𑄴𑄥𑄧𑄚𑄴', 'ce': 'лахара саксонийн', 'chr': 'ᎡᎳᏗ ᏁᏛᎳᏂ', 'cs': 'dolnosaština', 'cy': 'Sacsoneg Isel', 'de': 'Niedersächsisch', 'el': 'Κάτω Γερμανικά Ολλανδίας', 'el-polyton': 'Κάτω Γερμανικά Ολλανδίας', 'en': 'Low Saxon', 'es': 'bajo sajón', 'et': 'Hollandi alamsaksa', 'eu': 'behe-saxoiera', 'fa': 'ساکسونی سفلی', 'ff-Adlm': '𞤂𞤫𞤧 𞤅𞤢𞤳𞤧𞤮𞤲𞤪𞤫', 'fi': 'alankomaidenalasaksa', 'fil': 'Low Saxon', 'fo': 'lágt saksiskt', 'fr': 'bas-saxon néerlandais', 'ga': 'Sacsainis Íochtarach', 'gd': 'Sagsannais Ìochdarach', 'gl': 'baixo saxón', 'gu': 'લો સેક્સોન', 'he': 'סקסונית תחתית', 'hi': 'निचली सैक्सन', 'hi-Latn': 'Low Saxon', 'hr': 'donjosaksonski', 'ht': 'bas-saxon néerlandais', 'hu': 'alsószász', 'hy': 'ստորին սաքսոներեն', 'ia': 'basse saxone', 'ig': 'Low Saxon', 'is': 'lágsaxneska', 'it': 'basso tedesco olandese', 'ka': 'ქვემოსაქსონური', 'kgp': 'sagsỹv rur', 'kk': 'төменгі саксон тілі', 'kk-Cyrl': 'төменгі саксон тілі', 'km': 'ហ្សាក់ស្យុងក្រោម', 'kn': 'ಲೋ ಸ್ಯಾಕ್ಸನ್', 'ko': '저지 색슨어', 'ksh': 'Nehdersaksesch en de Nederläng', 'ky': 'төмөнкү саксончо', 'lo': 'ຊາຊອນ ຕອນໄຕ', 'lrc': 'آلمانی ھارگە جا', 'lt': 'Žemutinės Saksonijos (Nyderlandai)', 'lv': 'lejassakšu', 'mk': 'долносаксонски', 'ml': 'ലോ സാക്സൺ', 'mr': 'लो सॅक्सन', 'ms': 'Saxon Rendah', 'mt': 'Sassonu Komuni', 'my': 'ဂျာမန် (နယ်သာလန်)', 'mzn': 'پایین ساکسونی', 'nb': 'nedersaksisk', 'ne': 'तल्लो साक्सन', 'nl': 'Nederduits', 'nn': 'lågsaksisk', 'no': 'nedersaksisk', 'nqo': 'ߤߏߟߊ߲ߘߌ߬ ߓߊߛߊߞߑߛߐ߲ߞߊ߲', 'pa': 'ਲੋ ਸੈਕਸਨ', 'pa-Guru': 'ਲੋ ਸੈਕਸਨ', 'pl': 'dolnosaksoński', 'pt': 'baixo saxão', 'ro': 'saxona de jos', 'ru': 'нижнесаксонский', 'sc': 'sàssone bassu', 'si': 'පහළ සැක්සන්', 'sk': 'dolná saština', 'sl': 'nizka saščina', 'smn': 'Vuáládâhenâmij saksakielâ', 'sq': 'gjermanishte saksone e vendeve të ulëta', 'sr': 'нискосаксонски', 'sr-Cyrl': 'нискосаксонски', 'sr-Latn': 'niskosaksonski', 'sv': 'lågsaxiska', 'syr': 'ܗܘܠܢܕܐ ܠܐܠܬܚܬ', 'ta': 'லோ சாக்ஸன்', 'te': 'లో సాక్సన్', 'th': 'แซกซอนใต้', 'ti': 'ትሑት ሳክሰን', 'to': 'lea fakasakisoni-hifo', 'tr': 'Aşağı Saksonca', 'uk': 'нижньосаксонська', 'ur': 'ادنی سیکسن', 'uz': 'quyi sakson', 'uz-Latn': 'quyi sakson', 'vec': 'sàsone baso', 'vi': 'Tiếng Hạ Saxon', 'yrl': 'sakisãu yatuka', 'yue': '低地薩克遜文', 'yue-Hans': '低地萨克逊文', 'yue-Hant': '低地薩克遜文', 'zh': '低萨克森语', 'zh-Hans': '低萨克森语', 'zh-Hant': '低地薩克遜文', 'zu': 'isi-Low Saxon'},
    'nds-de': {'en': 'German Low German'},
    'nds-nl': {'en': 'Dutch Low Saxon'},
    'ndt': {'en': 'Ndunga'},
    'ndu': {'en': 'Dugun'},
    'ndv': {'en': 'Ndut'},
    'ndw': {'en': 'Ndobo'},
    'ndx': {'en': 'Nduga'},
    'ndy': {'en': 'Lutos'},
    'ndz': {'en': 'Ndogo'},
    'ne': {'af': 'Nepalees', 'agq': 'Nɛ̀kpalì', 'ak': 'Nɛpal kasa', 'am': 'ኔፓሊኛ', 'ar': 'النيبالية', 'as': 'নেপালী', 'asa': 'Kinepali', 'ast': 'nepalés', 'az': 'nepal', 'az-Cyrl': 'непал', 'az-Latn': 'nepal', 'bal-Latn': 'Népáli', 'bas': 'Hɔp u nepa᷆l', 'be': 'непальская', 'be-tarask': 'непальская', 'bem': 'Ichi Nepali', 'bez': 'Hinepali', 'bg': 'непалски', 'bgn': 'نیپالی', 'bm': 'nepalekan', 'bn': 'নেপালী', 'bo': 'ནེ་པ་ལི', 'br': 'nepaleg', 'brx': 'नेपाली', 'bs': 'nepalski', 'bs-Cyrl': 'непалски', 'bs-Latn': 'nepalski', 'ca': 'nepalès', 'ca-ES-valencia': 'nepalès', 'ccp': '𑄚𑄬𑄛𑄣𑄨', 'ce': 'непалхойн', 'cgg': 'Orunepali', 'chr': 'ᏁᏆᎵ', 'ckb': 'نیپالی', 'cs': 'nepálština', 'cy': 'Nepaleg', 'da': 'nepalesisk', 'dav': 'Kinepali', 'de': 'Nepalesisch', 'dje': 'Neepal senni', 'dsb': 'nepalšćina', 'dyo': 'nepalees', 'dz': 'ནེ་པཱལི་ཁ', 'ebu': 'Kĩnepali', 'ee': 'nepalgbe', 'el': 'Νεπαλικά', 'el-polyton': 'Νεπαλικά', 'en': 'Nepali', 'eo': 'nepala', 'es': 'nepalí', 'et': 'nepali', 'eu': 'nepalera', 'ewo': 'ǹkɔ́bɔ nefálian', 'fa': 'نپالی', 'ff': 'Nepaaleere', 'ff-Adlm': '𞤐𞤫𞤨𞤢𞤤𞤭𞤲𞤳𞤮𞥅𞤪𞤫', 'ff-Latn': 'Nepaaleere', 'fi': 'nepali', 'fil': 'Nepali', 'fo': 'nepalskt', 'fr': 'népalais', 'fur': 'nepalês', 'fy': 'Nepalees', 'ga': 'Neipeailis', 'gd': 'Neapàlais', 'gl': 'nepalí', 'gsw': 'Nepalesisch', 'gu': 'નેપાળી', 'guz': 'Kinepali', 'ha': 'Nepali', 'he': 'נפאלית', 'hi': 'नेपाली', 'hi-Latn': 'Nepali', 'hr': 'nepalski', 'hsb': 'nepalšćina', 'ht': 'népalais', 'hu': 'nepáli', 'hy': 'նեպալերեն', 'ia': 'nepalese', 'id': 'Nepali', 'ig': 'Nepali', 'is': 'nepalska', 'it': 'nepalese', 'ja': 'ネパール語', 'jmc': 'Kyinepali', 'jv': 'Nepal', 'ka': 'ნეპალური', 'kab': 'Tanipalit', 'kam': 'Kinepali', 'kde': 'Chinepali', 'kea': 'nepales', 'kgp': 'nẽparej', 'khq': 'Neepal senni', 'ki': 'Kinepali', 'kk': 'непал тілі', 'kk-Cyrl': 'непал тілі', 'kln': 'kutitab Nepali', 'km': 'នេប៉ាល់', 'kn': 'ನೇಪಾಳಿ', 'ko': '네팔어', 'kok': 'नेपाळी', 'kok-Deva': 'नेपाळी', 'ks': 'نیپٲلؠ', 'ks-Arab': 'نیپٲلؠ', 'ksb': 'Kinepali', 'ksf': 'rinepalɛ́', 'ksh': 'Nepallehsesch', 'ku': 'nepalî', 'kxv': 'nepaḷī', 'kxv-Deva': 'नेपाली', 'kxv-Latn': 'nepaḷī', 'kxv-Orya': 'ନେପାଳୀ', 'kxv-Telu': 'సేపాలి', 'ky': 'непалча', 'lag': 'Kɨnepáali', 'lb': 'Nepalesesch', 'lg': 'Lunepali', 'lkt': 'Nepal Iyápi', 'lld': 'nepalesc', 'ln': 'linepalɛ', 'lo': 'ເນປາລີ', 'lrc': 'نئپالی', 'lt': 'nepaliečių', 'lu': 'nepali', 'luo': 'Kinepali', 'luy': 'Kinepali', 'lv': 'nepāliešu', 'mai': 'नेपाली', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lnepali', 'mer': 'Kĩnepali', 'mfe': 'nepale', 'mg': 'Nepale', 'mgh': 'Inepali', 'mi': 'Nepari', 'mk': 'непалски', 'ml': 'നേപ്പാളി', 'mn': 'балба', 'mr': 'नेपाळी', 'ms': 'Nepal', 'mt': 'Nepaliż', 'mua': 'Nepaliya', 'my': 'နီပေါ', 'mzn': 'نپالی', 'naq': 'Nepalǁî gowab', 'nb': 'nepali', 'nd': 'isi-Nepali', 'ne': 'नेपाली', 'nl': 'Nepalees', 'nmg': 'Kiɛl nepal', 'nn': 'nepalsk', 'no': 'nepali', 'nqo': 'ߣߋߔߌߟߌߞߊ߲', 'nus': 'Thok napalni', 'nyn': 'Orunepali', 'om': 'Afaan Nepalii', 'or': 'ନେପାଳୀ', 'pa': 'ਨੇਪਾਲੀ', 'pa-Guru': 'ਨੇਪਾਲੀ', 'pcm': 'Nẹpálí Lángwej', 'pl': 'nepalski', 'ps': 'نېپالي', 'pt': 'nepalês', 'qu': 'Nepali Simi', 'rm': 'nepalais', 'rn': 'Ikinepali', 'ro': 'nepaleză', 'rof': 'Kinepali', 'ru': 'непальский', 'rw': 'Ikinepali', 'rwk': 'Kyinepali', 'sah': 'Ньыпааллыы', 'saq': 'Kinepali', 'sbp': 'Ishinepali', 'sc': 'nepalesu', 'sd': 'نيپالي', 'sd-Arab': 'نيپالي', 'se': 'nepaligiella', 'seh': 'nepalês', 'ses': 'Neepal senni', 'sg': 'Nepalëe', 'shi': 'ⵜⴰⵏⵉⴱⴰⵍⵉⵜ', 'shi-Latn': 'Tanibalit', 'shi-Tfng': 'ⵜⴰⵏⵉⴱⴰⵍⵉⵜ', 'si': 'නේපාල', 'sk': 'nepálčina', 'sl': 'nepalščina', 'smn': 'nepalkielâ', 'sn': 'chiNepali', 'so': 'Nebaali', 'sq': 'nepalisht', 'sr': 'непалски', 'sr-Cyrl': 'непалски', 'sr-Latn': 'nepalski', 'sv': 'nepalesiska', 'sw': 'Kinepali', 'ta': 'நேபாளி', 'te': 'నేపాలీ', 'teo': 'Kinepali', 'tg': 'непалӣ', 'th': 'เนปาล', 'ti': 'ኔፓሊ', 'tk': 'nepal dili', 'to': 'lea fakanepali', 'tr': 'Nepalce', 'tt': 'непали', 'twq': 'Neepal senni', 'tzm': 'Tanippalit', 'ug': 'نېپالچە', 'uk': 'непальська', 'ur': 'نیپالی', 'uz': 'nepal', 'uz-Cyrl': 'непалча', 'uz-Latn': 'nepal', 'vai': 'ꕇꕐꔷ', 'vai-Latn': 'Nipali', 'vai-Vaii': 'ꕇꕐꔷ', 'vec': 'nepaleze', 'vi': 'Tiếng Nepal', 'vun': 'Kyinepali', 'wae': 'Nepalesiš', 'wo': 'Nepale', 'xnr': 'नेपाली', 'xog': 'Olunepali', 'yav': 'nunipálɛ', 'yi': 'נעפּאַליש', 'yo': 'Èdè Nepali', 'yrl': 'neparei', 'yue': '尼泊爾文', 'yue-Hans': '尼泊尔文', 'yue-Hant': '尼泊爾文', 'zgh': 'ⵜⴰⵏⵉⴱⴰⵍⵉⵜ', 'zh': '尼泊尔语', 'zh-Hans': '尼泊尔语', 'zh-Hant': '尼泊爾文', 'zu': 'isi-Nepali'},
    'nea': {'en': "Eastern Ngad'a"},
    'neb': {'en': "Toura (Côte d'Ivoire)"},
    'nec': {'en': 'Nedebang'},
    'ned': {'en': 'Nde-Gbite'},
    'nee': {'en': 'Nêlêmwa-Nixumwak'},
    'nef': {'en': 'Nefamese'},
    'neg': {'en': 'Negidal'},
    'neh': {'en': 'Nyenkha'},
    'nei': {'en': 'Neo-Hittite'},
    'nej': {'en': 'Neko'},
    'nek': {'en': 'Neku'},
    'nem': {'en': 'Nemi'},
    'nen': {'en': 'Nengone'},
    'neo': {'en': 'Ná-Meo'},
    'neq': {'en': 'North Central Mixe'},
    'ner': {'en': 'Yahadian'},
    'nes': {'en': 'Bhoti Kinnauri'},
    'net': {'en': 'Nete'},
    'neu': {'en': 'Neo'},
    'nev': {'en': 'Nyaheun'},
    'new': {'af': 'Newari', 'am': 'ኒዋሪ(ኔፓል)', 'ar': 'النوارية', 'as': 'নেৱাৰি', 'ast': 'newari', 'az': 'nevari', 'az-Cyrl': 'невари', 'az-Latn': 'nevari', 'be': 'неўары', 'be-tarask': 'неўары', 'bg': 'неварски', 'bgn': 'نیواری', 'bn': 'নেওয়ারি', 'br': 'newari', 'brx': 'नेवारी', 'bs': 'nevari', 'bs-Cyrl': 'невари', 'bs-Latn': 'nevari', 'ca': 'newari', 'ca-ES-valencia': 'newari', 'ccp': '𑄚𑄬𑄃𑄮𑄠𑄢𑄨', 'ce': 'неваройн', 'chr': 'ᏁᏩᎵ', 'ckb': 'نێواری', 'cs': 'névárština', 'cy': 'Newaeg', 'da': 'newari', 'de': 'Newari', 'dsb': 'newarišćina', 'el': 'Νεγουάρι', 'el-polyton': 'Νεγουάρι', 'en': 'Newari', 'es': 'nevarí', 'et': 'nevari', 'eu': 'newarera', 'fa': 'نواریایی', 'ff-Adlm': '𞤐𞤫𞤱𞤢𞤪𞤭𞥅𞤪𞤫', 'fi': 'newari', 'fil': 'Newari', 'fo': 'newari', 'fr': 'newari', 'fy': 'Newari', 'ga': 'Néamharais', 'gd': 'Newari', 'gl': 'newari', 'gsw': 'Newarisch', 'gu': 'નેવારી', 'ha': 'Newari', 'he': 'נווארי', 'hi': 'नेवाड़ी', 'hi-Latn': 'Newari', 'hr': 'newari', 'hsb': 'newarišćina', 'ht': 'newari', 'hu': 'nevari', 'hy': 'նեվարերեն', 'ia': 'newari', 'id': 'Newari', 'ig': 'Newari', 'is': 'nevarí', 'it': 'newari', 'ja': 'ネワール語', 'jv': 'Newari', 'ka': 'ნევარი', 'kgp': 'nẽvari', 'kk': 'невар тілі', 'kk-Cyrl': 'невар тілі', 'km': 'នេវ៉ាវី', 'kn': 'ನೇವಾರೀ', 'ko': '네와르어', 'kok': 'नेवारी', 'kok-Deva': 'नेवारी', 'ks': 'نیواری', 'ks-Arab': 'نیواری', 'ksh': 'Newahresch', 'ku': 'newarî', 'ky': 'невариче', 'lb': 'Newari', 'lo': 'ນີວາຣິ', 'lt': 'nevari', 'lv': 'nevaru', 'mai': 'नेबारी', 'mi': 'Newari', 'mk': 'неварски', 'ml': 'നേവാരി', 'mn': 'невари', 'mr': 'नेवारी', 'ms': 'Newari', 'mt': 'Newari', 'my': 'နီဝါရီ', 'nb': 'newari', 'ne': 'नेवारी', 'nl': 'Newari', 'nn': 'newari', 'no': 'newari', 'or': 'ନେୱାରୀ', 'pa': 'ਨੇਵਾਰੀ', 'pa-Guru': 'ਨੇਵਾਰੀ', 'pcm': 'Nẹwarí Lángwej', 'pl': 'newarski', 'ps': 'نيواري', 'pt': 'newari', 'qu': 'Newari Simi', 'rm': 'newari', 'ro': 'newari', 'ru': 'неварский', 'sc': 'nepal bhasa', 'sd': 'نيواري', 'sd-Arab': 'نيواري', 'si': 'නෙවාරි', 'sk': 'nevárčina', 'sl': 'nevarščina', 'smn': 'newari', 'so': 'Newari', 'sq': 'neuarisht', 'sr': 'невари', 'sr-Cyrl': 'невари', 'sr-Latn': 'nevari', 'sv': 'newariska', 'sw': 'Kinewari', 'ta': 'நெவாரி', 'te': 'నెవారి', 'th': 'เนวาร์', 'ti': 'ነዋሪ', 'tk': 'newari dili', 'to': 'lea fakaneuali', 'tr': 'Nevari', 'ug': 'نېۋارىچە', 'uk': 'неварі', 'ur': 'نیواری', 'uz': 'nevar', 'uz-Latn': 'nevar', 'vec': 'newari', 'vi': 'Tiếng Newari', 'xnr': 'नेवाड़ी', 'yo': 'Èdè Newari', 'yrl': 'newari', 'yue': '尼瓦爾文', 'yue-Hans': '尼瓦尔文', 'yue-Hant': '尼瓦爾文', 'zh': '尼瓦尔语', 'zh-Hans': '尼瓦尔语', 'zh-Hant': '尼瓦爾文', 'zu': 'isi-Newari'},
    'nex': {'en': 'Neme'},
    'ney': {'en': 'Neyo'},
    'nez': {'en': 'Nez Perce'},
    'nfa': {'en': 'Dhao'},
    'nfd': {'en': 'Ahwai'},
    'nfl': {'en': 'Ayiwo'},
    'nfr': {'en': 'Nafaanra'},
    'nfu': {'en': 'Mfumte'},
    'ng': {'af': 'Ndonga', 'am': 'ንዶንጋ', 'ar': 'الندونجا', 'as': 'এন্দোঙ্গা', 'ast': 'ndonga', 'az': 'ndonqa', 'az-Cyrl': 'ндонга', 'az-Latn': 'ndonqa', 'be': 'ндонга', 'be-tarask': 'ндонга', 'bg': 'ндонга', 'bgn': 'اندونگی', 'bn': 'এন্দোঙ্গা', 'br': 'ndonga', 'brx': 'न्डोंगा', 'bs': 'ndonga', 'bs-Cyrl': 'ндонга', 'bs-Latn': 'ndonga', 'ca': 'ndonga', 'ca-ES-valencia': 'ndonga', 'ccp': '𑄆𑄚𑄴𑄘𑄮𑄋𑄴𑄉', 'ce': 'ндонга', 'chr': 'ᎾᏙᎦ', 'ckb': 'ندۆنگا', 'cs': 'ndondština', 'cy': 'Ndonga', 'da': 'ndonga', 'de': 'Ndonga', 'dsb': 'ndongašćina', 'el': 'Ντόνγκα', 'el-polyton': 'Ντόνγκα', 'en': 'Ndonga', 'es': 'ndonga', 'et': 'ndonga', 'eu': 'ndonga', 'fa': 'اندونگایی', 'ff-Adlm': '𞤐𞤣𞤮𞤲𞤺𞤢𞥄𞤪𞤫', 'fi': 'ndonga', 'fil': 'Ndonga', 'fo': 'ndonga', 'fr': 'ndonga', 'fy': 'Ndonga', 'ga': 'Ndongais', 'gd': 'Ndonga', 'gl': 'ndonga', 'gsw': 'Ndonga', 'gu': 'ડોન્ગા', 'ha': 'Ndonga', 'he': 'נדונגה', 'hi': 'डोन्गा', 'hi-Latn': 'Ndonga', 'hr': 'ndonga', 'hsb': 'ndongašćina', 'ht': 'ndonga', 'hu': 'ndonga', 'hy': 'նդոնգա', 'ia': 'ndonga', 'id': 'Ndonga', 'ig': 'Ndonga', 'is': 'ndonga', 'it': 'ndonga', 'ja': 'ンドンガ語', 'jv': 'Ndonga', 'ka': 'ნდონგა', 'kgp': 'nogã', 'kk': 'ндонга тілі', 'kk-Cyrl': 'ндонга тілі', 'km': 'នុនហ្គា', 'kn': 'ಡೋಂಗಾ', 'ko': '느동가어', 'kok': 'डोंगा', 'kok-Deva': 'डोंगा', 'ks': 'ڈونگا', 'ks-Arab': 'ڈونگا', 'ksh': 'Ndongjahnesch', 'ku': 'ndongayî', 'ky': 'ндонгача', 'lb': 'Ndonga', 'lo': 'ເອັນດອງກາ', 'lt': 'ndongų', 'lv': 'ndongu', 'mai': 'एनडोंगा', 'mi': 'Natōka', 'mk': 'ндонга', 'ml': 'ഡോങ്ക', 'mn': 'ндонга', 'mr': 'डोंगा', 'ms': 'Ndonga', 'mt': 'Ndonga', 'my': 'အွန်ဒွန်ဂါ', 'nb': 'ndonga', 'ne': 'न्दोन्गा', 'nl': 'Ndonga', 'nn': 'ndonga', 'no': 'ndonga', 'or': 'ଡୋଙ୍ଗା', 'pa': 'ਐਂਡੋਂਗਾ', 'pa-Guru': 'ਐਂਡੋਂਗਾ', 'pcm': 'Ndónga Lángwej', 'pl': 'ndonga', 'ps': 'ندونگا', 'pt': 'dongo', 'qu': 'Ndonga Simi', 'rm': 'ndonga', 'ro': 'ndonga', 'ru': 'ндонга', 'sc': 'ndonga', 'sd': 'ڊونگا', 'sd-Arab': 'ڊونگا', 'si': 'න්ඩොන්ගා', 'sk': 'ndonga', 'sl': 'ndonga', 'smn': 'ndonga', 'so': 'Ndonga', 'sq': 'ndongaisht', 'sr': 'ндонга', 'sr-Cyrl': 'ндонга', 'sr-Latn': 'ndonga', 'sv': 'ndonga', 'sw': 'Kindonga', 'ta': 'தோங்கா', 'te': 'డోంగా', 'th': 'ดองกา', 'ti': 'ኤንዶንጋ', 'tk': 'ndonga dili', 'to': 'lea fakanetongikā', 'tr': 'Ndonga', 'ug': 'ندونگاچە', 'uk': 'ндонга', 'ur': 'نڈونگا', 'uz': 'ndonga', 'uz-Latn': 'ndonga', 'vec': 'ndonga', 'vi': 'Tiếng Ndonga', 'xnr': 'डोन्गा', 'yo': 'Èdè Ndonga', 'yrl': 'dũgu', 'yue': '恩東加文', 'yue-Hans': '恩东加文', 'yue-Hant': '恩東加文', 'zh': '恩东加语', 'zh-Hans': '恩东加语', 'zh-Hant': '恩東加文', 'zu': 'isi-Ndonga'},
    'nga': {'en': 'Ngbaka'},
    'ngb': {'en': 'Northern Ngbandi'},
    'ngc': {'en': 'Ngombe (Democratic Republic of Congo)'},
    'ngd': {'en': 'Ngando (Central African Republic)'},
    'nge': {'en': 'Ngemba'},
    'ngf': {'en': 'Trans-New Guinea languages'},
    'ngg': {'en': 'Ngbaka Manza'},
    'ngh': {'en': 'Nǁng'},
    'ngi': {'en': 'Ngizim'},
    'ngj': {'en': 'Ngie'},
    'ngk': {'en': 'Dalabon'},
    'ngl': {'en': 'Lomwe'},
    'ngm': {'en': "Ngatik Men's Creole"},
    'ngn': {'en': 'Ngwo'},
    'ngo': {'en': 'Ngoni'},
    'ngp': {'en': 'Ngulu'},
    'ngq': {'en': 'Ngurimi'},
    'ngr': {'en': 'Engdewu'},
    'ngs': {'en': 'Gvoko'},
    'ngt': {'en': 'Kriang'},
    'ngu': {'en': 'Guerrero Nahuatl'},
    'ngv': {'en': 'Nagumi'},
    'ngw': {'en': 'Ngwaba'},
    'ngx': {'en': 'Nggwahyi'},
    'ngy': {'en': 'Tibea'},
    'ngz': {'en': 'Ngungwel'},
    'nha': {'en': 'Nhanda'},
    'nhb': {'en': 'Beng'},
    'nhc': {'en': 'Tabasco Nahuatl'},
    'nhd': {'en': 'Chiripá'},
    'nhe': {'en': 'Eastern Huasteca Nahuatl'},
    'nhf': {'en': 'Nhuwala'},
    'nhg': {'en': 'Tetelcingo Nahuatl'},
    'nhh': {'en': 'Nahari'},
    'nhi': {'en': 'Zacatlán-Ahuacatlán-Tepetzintla Nahuatl'},
    'nhk': {'en': 'Isthmus-Cosoleacaque Nahuatl'},
    'nhm': {'en': 'Morelos Nahuatl'},
    'nhn': {'en': 'Central Nahuatl', 'de': 'Zentral-Nahuatl'},
    'nho': {'en': 'Takuu'},
    'nhp': {'en': 'Isthmus-Pajapan Nahuatl'},
    'nhq': {'en': 'Huaxcaleca Nahuatl'},
    'nhr': {'en': 'Naro'},
    'nht': {'en': 'Ometepec Nahuatl'},
    'nhu': {'en': 'Noone'},
    'nhv': {'en': 'Temascaltepec Nahuatl'},
    'nhw': {'en': 'Western Huasteca Nahuatl'},
    'nhx': {'en': 'Isthmus-Mecayapan Nahuatl'},
    'nhy': {'en': 'Northern Oaxaca Nahuatl'},
    'nhz': {'en': 'Santa María La Alta Nahuatl'},
    'nia': {'af': 'Nias', 'am': 'ኒአስ', 'ar': 'النياس', 'as': 'নিয়াছ', 'ast': 'nias', 'az': 'nias', 'az-Cyrl': 'ниас', 'az-Latn': 'nias', 'be': 'ніас', 'be-tarask': 'ніас', 'bg': 'ниас', 'bgn': 'نیاسی', 'bn': 'নিয়াস', 'br': 'nias', 'brx': 'नियास', 'bs': 'nias', 'bs-Cyrl': 'ниас', 'bs-Latn': 'nias', 'ca': 'nias', 'ca-ES-valencia': 'nias', 'ccp': '𑄚𑄨𑄠𑄌𑄴', 'ce': 'ниас', 'chr': 'ᏂᎠᏏ', 'ckb': 'نیاس', 'cs': 'nias', 'cy': 'Nias', 'da': 'nias', 'de': 'Nias', 'dsb': 'niazišćina', 'el': 'Νίας', 'el-polyton': 'Νίας', 'en': 'Nias', 'es': 'nias', 'et': 'niasi', 'eu': 'niasera', 'fa': 'نیاسی', 'ff-Adlm': '𞤙𞤢𞤧𞤭𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'nias', 'fil': 'Nias', 'fo': 'nias', 'fr': 'niha', 'fy': 'Nias', 'ga': 'Niaisis', 'gd': 'Nias', 'gl': 'nias', 'gsw': 'Nias-Schpraach', 'gu': 'નિયાસ', 'ha': 'Nias', 'he': 'ניאס', 'hi': 'नियास', 'hi-Latn': 'Nias', 'hr': 'nias', 'hsb': 'niazišćina', 'ht': 'niha', 'hu': 'nias', 'hy': 'նիասերեն', 'ia': 'nias', 'id': 'Nias', 'ig': 'Nias', 'is': 'nías', 'it': 'nias', 'ja': 'ニアス語', 'jv': 'Nias', 'ka': 'ნიასი', 'kgp': 'nĩja', 'kk': 'ниас тілі', 'kk-Cyrl': 'ниас тілі', 'km': 'នីអាស', 'kn': 'ನಿಯಾಸ್', 'ko': '니아스어', 'kok': 'नियास', 'kok-Deva': 'नियास', 'ks': 'نِیاس', 'ks-Arab': 'نِیاس', 'ku': 'nîasî', 'ky': 'ниасча', 'lb': 'Nias-Sprooch', 'lo': 'ນີ່ອັດ', 'lt': 'nias', 'lv': 'njasu', 'mai': 'नियास', 'mi': 'Niāhi', 'mk': 'нијас', 'ml': 'നിയാസ്', 'mn': 'ниас хэл', 'mr': 'नियास', 'ms': 'Nias', 'mt': 'Nijas', 'my': 'နီးရပ်စ်', 'nb': 'nias', 'ne': 'नियास', 'nl': 'Nias', 'nn': 'nias', 'no': 'nias', 'or': 'ନୀୟାସ୍', 'pa': 'ਨਿਆਸ', 'pa-Guru': 'ਨਿਆਸ', 'pcm': 'Nias Lángwej', 'pl': 'nias', 'ps': 'نياس', 'pt': 'nias', 'qu': 'Nias Simi', 'rm': 'nias', 'ro': 'nias', 'ru': 'ниас', 'sc': 'nias', 'sd': 'نياس', 'sd-Arab': 'نياس', 'si': 'නියාස්', 'sk': 'niasánčina', 'sl': 'niaščina', 'smn': 'niaskielâ', 'so': 'Nias', 'sq': 'niasisht', 'sr': 'ниас', 'sr-Cyrl': 'ниас', 'sr-Latn': 'nias', 'sv': 'nias', 'sw': 'Kiniasi', 'ta': 'நியாஸ்', 'te': 'నియాస్', 'th': 'นีอัส', 'ti': 'ንያስ', 'tk': 'nias dili', 'to': 'lea fakaniasi', 'tr': 'Nias', 'ug': 'نىئاسچە', 'uk': 'ніаська', 'ur': 'نیاس', 'uz': 'nias', 'uz-Latn': 'nias', 'vec': 'nias', 'vi': 'Tiếng Nias', 'xnr': 'नियास', 'yo': 'Èdè Nia', 'yrl': 'niyasi', 'yue': '尼亞斯文', 'yue-Hans': '尼亚斯文', 'yue-Hant': '尼亞斯文', 'zh': '尼亚斯语', 'zh-Hans': '尼亚斯语', 'zh-Hant': '尼亞斯文', 'zu': 'isi-Nias'},
    'nib': {'en': 'Nakame'},
    'nic': {'en': 'Niger-Kordofanian languages'},
    'nic-bco-pro': {'en': 'Proto-Benue-Congo'},
    'nic-bod-pro': {'en': 'Proto-Bantoid'},
    'nic-gur-pro': {'en': 'Proto-Gur'},
    'nic-pro': {'en': 'Proto-Niger-Congo'},
    'nic-ubg-pro': {'en': 'Proto-Ubangian'},
    'nid': {'en': 'Ngandi'},
    'nie': {'en': 'Niellim'},
    'nif': {'en': 'Nek'},
    'nig': {'en': 'Ngalakgan'},
    'nih': {'en': 'Nyiha (Tanzania)'},
    'nii': {'en': 'Nii'},
    'nij': {'en': 'Ngaju'},
    'nik': {'en': 'Southern Nicobarese'},
    'nil': {'en': 'Nila'},
    'nim': {'en': 'Nilamba'},
    'nin': {'en': 'Ninzo'},
    'nio': {'en': 'Nganasan'},
    'niq': {'en': 'Nandi'},
    'nir': {'en': 'Nimboran'},
    'nis': {'en': 'Nimi'},
    'nit': {'en': 'Southeastern Kolami'},
    'niu': {'af': 'Niueaans', 'am': 'ኒዩአንኛ', 'ar': 'النيوي', 'as': 'নিয়ুৱান', 'ast': 'niueanu', 'az': 'niyuan', 'az-Cyrl': 'нијуан', 'az-Latn': 'niyuan', 'be': 'ніўэ', 'be-tarask': 'ніўэ', 'bg': 'ниуеан', 'bgn': 'نیویی', 'bn': 'নিউয়ান', 'br': 'niue', 'brx': 'नियुइआन', 'bs': 'niue', 'bs-Cyrl': 'ниуеан', 'bs-Latn': 'niue', 'ca': 'niueà', 'ca-ES-valencia': 'niueà', 'ccp': '𑄚𑄨𑄃𑄪𑄠𑄚𑄴', 'ce': 'ниуэ', 'chr': 'ᏂᏳᏫᏯᏂ', 'ckb': 'نیئوویی', 'cs': 'niueština', 'cy': 'Niuean', 'da': 'niueansk', 'de': 'Niue', 'dsb': 'niueanšćina', 'el': 'Νιούε', 'el-polyton': 'Νιούε', 'en': 'Niuean', 'es': 'niueano', 'et': 'niue', 'eu': 'niueera', 'fa': 'نیویی', 'ff-Adlm': '𞤐𞤭𞤵𞤫𞤴𞤢𞤲𞤪𞤫', 'fi': 'niue', 'fil': 'Niuean', 'fo': 'niuean', 'fr': 'niuéen', 'fy': 'Niueaansk', 'ga': 'Níobhais', 'gd': 'Cànan Niue', 'gl': 'niueano', 'gsw': 'Niue-Schpraach', 'gu': 'નિયુઆન', 'ha': 'Niuean', 'he': 'ניואן', 'hi': 'नियुआन', 'hi-Latn': 'Niuean', 'hr': 'niujski', 'hsb': 'niueanšćina', 'ht': 'niuéen', 'hu': 'niuei', 'hy': 'նիուերեն', 'ia': 'nieuano', 'id': 'Niuea', 'ig': 'Niuean', 'is': 'níveska', 'it': 'niue', 'ja': 'ニウーエイ語', 'jv': 'Niuean', 'ka': 'ნიუე', 'kgp': 'nivuvejỹnũ', 'kk': 'ниуэ тілі', 'kk-Cyrl': 'ниуэ тілі', 'km': 'នូអៀន', 'kn': 'ನಿಯುವನ್', 'ko': '니웨언어', 'kok': 'नायान', 'kok-Deva': 'नायान', 'ks': 'نِیویَن', 'ks-Arab': 'نِیویَن', 'ku': 'nîwî', 'ky': 'ньюанча', 'lb': 'Niue-Sprooch', 'lo': 'ນີ່ອູ', 'lt': 'niujiečių', 'lv': 'niuāņu', 'mai': 'नियुएन', 'mi': 'Niueana', 'mk': 'ниујески', 'ml': 'ന്യുവാൻ', 'mn': 'ниуэ', 'mr': 'नियुआन', 'mt': 'Niuean', 'my': 'နူအဲယန်း', 'nb': 'niueansk', 'ne': 'निउएन', 'nl': 'Niueaans', 'nn': 'niuisk', 'no': 'niueansk', 'or': 'ନିୟୁଆନ୍', 'pa': 'ਨਿਊਏਈ', 'pa-Guru': 'ਨਿਊਏਈ', 'pcm': 'Niúeán Lángwej', 'pl': 'niue', 'ps': 'نیان', 'pt': 'niueano', 'qu': 'Niuean Simi', 'rm': 'niue', 'ro': 'niueană', 'ru': 'ниуэ', 'sc': 'niueanu', 'sd': 'نووي', 'sd-Arab': 'نووي', 'si': 'නියුඑන්', 'sk': 'niueština', 'sl': 'niuejščina', 'smn': 'niuekielâ', 'so': 'Niuean', 'sq': 'niueanisht', 'sr': 'ниуејски', 'sr-Cyrl': 'ниуејски', 'sr-Latn': 'niuejski', 'sv': 'niueanska', 'sw': 'Kiniuea', 'ta': 'நியூவான்', 'te': 'నియాన్', 'tg': 'ниуэӣ', 'th': 'นีวเว', 'ti': 'ንዌንኛ', 'tk': 'niue dili', 'to': 'lea fakaniuē', 'tr': 'Niue dili', 'tt': 'ниуэ', 'ug': 'نيۇئېچە', 'uk': 'ніуе', 'ur': 'نیویائی', 'uz': 'niue', 'uz-Cyrl': 'ниуэча', 'uz-Latn': 'niue', 'vec': 'niuean', 'vi': 'Tiếng Niuean', 'wo': 'Niweyan', 'xnr': 'नियुआन', 'yo': 'Èdè Niu', 'yrl': 'niweanu', 'yue': '紐埃文', 'yue-Hans': '纽埃文', 'yue-Hant': '紐埃文', 'zh': '纽埃语', 'zh-Hans': '纽埃语', 'zh-Hant': '紐埃文', 'zu': 'isi-Niuean'},
    'niv': {'en': 'Gilyak'},
    'niw': {'en': 'Nimo'},
    'nix': {'en': 'Hema'},
    'niy': {'en': 'Ngiti'},
    'niz': {'en': 'Ningil'},
    'nja': {'en': 'Nzanyi'},
    'njb': {'en': 'Nocte Naga'},
    'njd': {'en': 'Ndonde Hamba'},
    'njh': {'en': 'Lotha Naga'},
    'nji': {'en': 'Gudanji'},
    'njj': {'en': 'Njen'},
    'njl': {'en': 'Njalgulgule'},
    'njm': {'en': 'Angami Naga'},
    'njn': {'en': 'Liangmai Naga'},
    'njo': {'am': 'ኦ ናጋ', 'ast': 'ao naga', 'br': 'aoeg', 'cs': 'ao (jazyky Nágálandu)', 'cy': 'Ao Naga', 'de': 'Ao-Naga', 'en': 'Ao Naga', 'et': 'ao', 'fi': 'ao naga', 'fr': 'ao', 'gd': 'Ao Naga', 'hi-Latn': 'Ao Naga', 'ht': 'ao', 'it': 'ao', 'ja': 'アオ・ナガ語', 'lb': 'Ao Naga', 'lt': 'ao naga', 'mk': 'ао нага', 'nb': 'ao naga', 'ne': 'अओ नागा', 'nl': 'Ao Naga', 'nn': 'ao naga', 'no': 'ao naga', 'pl': 'ao', 'sv': 'ao-naga', 'th': 'อ๋าวนากา', 'to': 'lea fakaʻaonasa', 'tr': 'Ao Naga', 'uk': 'ао нага', 'vi': 'Tiếng Ao Naga', 'yue': '阿沃那加文', 'yue-Hans': '阿沃那加文', 'yue-Hant': '阿沃那加文', 'zh-Hant': '阿沃那加文'},
    'njr': {'en': 'Njerep'},
    'njs': {'en': 'Nisa'},
    'njt': {'en': 'Ndyuka-Trio Pidgin'},
    'nju': {'en': 'Ngadjunmaya'},
    'njx': {'en': 'Kunyi'},
    'njy': {'en': 'Njyem'},
    'njz': {'en': 'Nyishi'},
    'nka': {'en': 'Nkoya'},
    'nkb': {'en': 'Khoibu Naga'},
    'nkc': {'en': 'Nkongho'},
    'nkd': {'en': 'Koireng'},
    'nke': {'en': 'Duke'},
    'nkf': {'en': 'Inpui Naga'},
    'nkg': {'en': 'Nekgini'},
    'nkh': {'en': 'Khezha Naga'},
    'nki': {'en': 'Thangal Naga'},
    'nkj': {'en': 'Nakai'},
    'nkk': {'en': 'Nokuku'},
    'nkm': {'en': 'Namat'},
    'nkn': {'en': 'Nkangala'},
    'nko': {'en': 'Nkonya'},
    'nkp': {'en': 'Niuatoputapu'},
    'nkq': {'en': 'Nkami'},
    'nkr': {'en': 'Nukuoro'},
    'nks': {'en': 'North Asmat'},
    'nkt': {'en': 'Nyika (Tanzania)'},
    'nku': {'en': 'Bouna Kulango'},
    'nkv': {'en': 'Nyika (Malawi and Zambia)'},
    'nkw': {'en': 'Nkutu'},
    'nkx': {'en': 'Nkoroo'},
    'nkz': {'en': 'Nkari'},
    'nl': {'af': 'Nederlands', 'agq': 'Dɔ̂s', 'ak': 'Dɛɛkye', 'am': 'ደች', 'ar': 'الهولندية', 'as': 'ডাচ', 'asa': 'Kiholandhi', 'ast': 'neerlandés', 'az': 'holland', 'az-Arab': 'هوْلندجه', 'az-Cyrl': 'һолланд', 'az-Latn': 'holland', 'bal-Latn': 'Dacch', 'bas': 'Hɔp u nlɛ̀ndi', 'be': 'нідэрландская', 'be-tarask': 'нідэрландская', 'bem': 'Ichi Dachi', 'bez': 'Hiholanzi', 'bg': 'нидерландски', 'bgn': 'هالنڈی', 'blo': 'Holanɖ kagɩja', 'bm': 'olandekan', 'bn': 'ওলন্দাজ', 'br': 'nederlandeg', 'brx': 'डात्च', 'bs': 'nizozemski', 'bs-Cyrl': 'холандски', 'bs-Latn': 'nizozemski', 'ca': 'neerlandès', 'ca-ES-valencia': 'neerlandès', 'ccp': '𑄓𑄌𑄴', 'ce': 'голландхойн', 'ceb': 'Dutch', 'cgg': 'Orudaaki', 'chr': 'ᏛᏥ', 'ckb': 'هۆڵەندی', 'cs': 'nizozemština', 'cv': 'голланди', 'cy': 'Iseldireg', 'da': 'nederlandsk', 'dav': 'Kiholanzi', 'de': 'Niederländisch', 'dje': 'Holandee senni', 'dsb': 'nižozemšćina', 'dyo': 'neerlande', 'dz': 'ཌཆ་ཁ', 'ebu': 'Kĩholanzi', 'ee': 'Hollandgbe', 'el': 'Ολλανδικά', 'el-polyton': 'Ὁλλανδικά', 'en': 'Dutch', 'en-Dsrt': '𐐔𐐲𐐽', 'eo': 'nederlanda', 'es': 'neerlandés', 'et': 'hollandi', 'eu': 'nederlandera', 'ewo': 'Ǹkɔ́bɔ nɛrəlándía', 'fa': 'هلندی', 'ff': 'Dacceere', 'ff-Adlm': '𞤁𞤮𞥅𞤷𞤵𞤪𞤫', 'ff-Latn': 'Dacceere', 'fi': 'hollanti', 'fil': 'Dutch', 'fo': 'hálendskt', 'fr': 'néerlandais', 'fur': 'olandês', 'fy': 'Nederlânsk', 'ga': 'Ollainnis', 'gd': 'Duitsis', 'gl': 'neerlandés', 'gsw': 'Niderländisch', 'gu': 'ડચ', 'guz': 'Kiholanzi', 'ha': 'Holanci', 'haw': 'Hōlani', 'he': 'הולנדית', 'hi': 'डच', 'hi-Latn': 'Dutch', 'hr': 'nizozemski', 'hsb': 'nižozemšćina', 'ht': 'néerlandais', 'hu': 'holland', 'hy': 'հոլանդերեն', 'ia': 'nederlandese', 'id': 'Belanda', 'ie': 'hollandesi', 'ig': 'Dutch', 'ii': 'ꉿꇂꉙ', 'is': 'hollenska', 'it': 'olandese', 'ja': 'オランダ語', 'jmc': 'Kyiholanzi', 'jv': 'Walanda', 'ka': 'ნიდერლანდური', 'kab': 'Tadučit', 'kam': 'Kiholanzi', 'kde': 'Chiholanzi', 'kea': 'olandes', 'kgp': 'orỹnej', 'khq': 'Holandee senni', 'ki': 'Kiholanzi', 'kk': 'нидерланд тілі', 'kk-Cyrl': 'нидерланд тілі', 'kln': 'kutitab Boa', 'km': 'ហូឡង់', 'kn': 'ಡಚ್', 'ko': '네덜란드어', 'kok': 'डच', 'kok-Deva': 'डच', 'ks': 'ڈَچ', 'ks-Arab': 'ڈَچ', 'ksb': 'Kiholanzi', 'ksf': 'riɔlándɛ́', 'ksh': 'Holländesch', 'ku': 'holendî', 'kxv': 'ḍc', 'kxv-Deva': 'डच', 'kxv-Latn': 'ḍc', 'kxv-Orya': 'ଡଚ୍', 'kxv-Telu': 'డచ్', 'ky': 'голландча', 'lag': 'Kɨholáanzi', 'lb': 'Hollännesch', 'lg': 'Luholandi', 'lkt': 'Dutch Iyápi', 'lld': 'neerlandesc', 'ln': 'lifalamá', 'lo': 'ດັຊ', 'lrc': 'ھولأندی', 'lt': 'olandų', 'lu': 'olandi', 'luo': 'Kiholanzi', 'luy': 'Kiholanzi', 'lv': 'holandiešu', 'mai': 'डच', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lduchi', 'mer': 'Kĩholandi', 'mfe': 'olande', 'mg': 'Holandey', 'mgh': 'Iholanzi', 'mi': 'Tati', 'mk': 'холандски', 'ml': 'ഡച്ച്', 'mn': 'нидерланд', 'mni': 'দচ', 'mni-Beng': 'দচ', 'mr': 'डच', 'ms': 'Belanda', 'mt': 'Olandiż', 'mua': 'zah sǝr ma kasǝŋ', 'my': 'ဒတ်ခ်ျ', 'mzn': 'هلندی', 'naq': 'Hollands', 'nb': 'nederlandsk', 'nd': 'isi-Dutch', 'ne': 'डच', 'nl': 'Nederlands', 'nmg': 'Kiɛl bóllandais', 'nn': 'nederlandsk', 'no': 'nederlandsk', 'nqo': 'ߣߍ߯ߙߑߟߊ߲ߘߌߞߊ߲', 'nus': 'Thok da̱c', 'nyn': 'Orudaaki', 'om': 'Afaan Dachii', 'or': 'ଡଚ୍', 'pa': 'ਡੱਚ', 'pa-Guru': 'ਡੱਚ', 'pcm': 'Dọch Lángwej', 'pl': 'niderlandzki', 'ps': 'هالېنډي', 'pt': 'holandês', 'qu': 'Neerlandes Simi', 'rm': 'ollandais', 'rn': 'Igiholandi', 'ro': 'neerlandeză', 'rof': 'Kiholanzi', 'ru': 'нидерландский', 'rw': 'Ikinerilande', 'rwk': 'Kyiholanzi', 'saq': 'Kiholanzi', 'sbp': 'Ishiholansi', 'sc': 'olandesu', 'sd': 'ڊچ', 'sd-Arab': 'ڊچ', 'se': 'hollánddagiella', 'seh': 'holandês', 'ses': 'Holandee senni', 'sg': 'Holandëe', 'shi': 'ⵜⴰⵀⵓⵍⴰⵏⴷⵉⵜ', 'shi-Latn': 'Tahulandit', 'shi-Tfng': 'ⵜⴰⵀⵓⵍⴰⵏⴷⵉⵜ', 'si': 'ලන්දේසි', 'sk': 'holandčina', 'sl': 'nizozemščina', 'smn': 'hollandkielâ', 'sn': 'chiDutch', 'so': 'Holandays', 'sq': 'holandisht', 'sr': 'холандски', 'sr-Cyrl': 'холандски', 'sr-Latn': 'holandski', 'sv': 'nederländska', 'sw': 'Kiholanzi', 'syr': 'ܗܘܠܢܕܝܬ', 'ta': 'டச்சு', 'te': 'డచ్', 'teo': 'Kiholanzi', 'tg': 'голландӣ', 'th': 'ดัตช์', 'ti': 'ዳች', 'tk': 'niderland dili', 'to': 'lea fakahōlani', 'tr': 'Felemenkçe', 'tt': 'голланд', 'twq': 'Holandee senni', 'tzm': 'Tahulanḍit', 'ug': 'گوللاندچە', 'uk': 'нідерландська', 'ur': 'ڈچ', 'uz': 'niderland', 'uz-Cyrl': 'голландча', 'uz-Latn': 'niderland', 'vai': 'ꗍꔿ', 'vai-Latn': 'Dɔchi', 'vai-Vaii': 'ꗍꔿ', 'vec': 'olandeze', 'vi': 'Tiếng Hà Lan', 'vun': 'Kyiholanzi', 'wae': 'Holändiš', 'wo': 'Neyerlànde', 'xh': 'IsiDatshi', 'xnr': 'डच', 'xog': 'Oluholandi', 'yav': 'nilándɛ', 'yi': 'האלענדיש', 'yo': 'Èdè Dọ́ọ̀ṣì', 'yrl': 'hurãdei', 'yue': '荷蘭文', 'yue-Hans': '荷兰文', 'yue-Hant': '荷蘭文', 'zgh': 'ⵜⴰⵀⵓⵍⴰⵏⴷⵉⵜ', 'zh': '荷兰语', 'zh-Hans': '荷兰语', 'zh-Hant': '荷蘭文', 'zu': 'isi-Dutch'},
    'nl-BE': {'af': 'Vlaams', 'ak': 'Dɛɛkye (Bɛɛgyiɔm', 'am': 'ፍሌሚሽ', 'as': 'ফ্লেমিচ', 'ast': 'flamencu', 'az': 'flamand', 'az-Cyrl': 'фламанд', 'az-Latn': 'flamand', 'bal-Latn': 'Dacch (Béljiam)', 'be': 'фламандская', 'be-tarask': 'фламандская', 'bg': 'фламандски', 'bgn': 'فلامانی', 'bn': 'ফ্লেমিশ', 'br': 'flandrezeg', 'brx': 'फ्लेमिस', 'bs': 'flamanski', 'bs-Cyrl': 'фламански', 'bs-Latn': 'flamanski', 'ca': 'flamenc', 'ca-ES-valencia': 'flamenc', 'ccp': '𑄜𑄳𑄣𑄬𑄟𑄨𑄌𑄴', 'ce': 'фламандийн', 'ceb': 'Flemish', 'chr': 'ᏊᎵᏥᎥᎻ ᏛᏥ', 'ckb': 'فلێمی', 'cs': 'vlámština', 'cv': 'фламанди', 'cy': 'Fflemeg', 'da': 'flamsk', 'de': 'Flämisch', 'dsb': 'flamšćina', 'dz': 'ཕྷེལེ་མིཤ་ཁ', 'ee': 'Flemishgbe', 'el': 'Φλαμανδικά', 'el-polyton': 'Φλαμανδικά', 'en': 'Flemish', 'es': 'flamenco', 'et': 'flaami', 'eu': 'flandriera', 'fa': 'فلمنگی', 'ff-Adlm': '𞤊𞤭𞤤𞤢𞤥𞤢𞤲𞤪𞤫', 'fi': 'flaami', 'fil': 'Flemish', 'fo': 'flamskt', 'fr': 'flamand', 'fur': 'flamant', 'fy': 'Vlaams', 'ga': 'Pléimeannais', 'gd': 'Flànrais', 'gl': 'flamengo', 'gsw': 'Fläämisch', 'gu': 'ફ્લેમિશ', 'he': 'הולנדית (פלמית)', 'hi': 'फ़्लेमिश', 'hi-Latn': 'Flemish', 'hr': 'flamanski', 'hsb': 'flamšćina', 'ht': 'flamand', 'hu': 'flamand', 'hy': 'ֆլամանդերեն', 'ia': 'flamingo', 'ie': 'flandrian', 'ig': 'Flemish', 'is': 'flæmska', 'it': 'fiammingo', 'ja': 'フラマン語', 'jv': 'Flemis', 'ka': 'ფლამანდიური', 'kea': 'flamengu', 'kgp': 'framẽgo', 'kk': 'фламанд тілі', 'kk-Cyrl': 'фламанд тілі', 'km': 'ផ្លាមីស', 'kn': 'ಫ್ಲೆಮಿಷ್', 'ko': '플라망어', 'kok': 'फ्लेमिश', 'kok-Deva': 'फ्लेमिश', 'ks': 'فلیمِش', 'ks-Arab': 'فلیمِش', 'ksh': 'Flähmesch', 'ku': 'flamî', 'kxv': 'vlaams', 'kxv-Deva': 'प़्लेमिस', 'kxv-Latn': 'vlaams', 'kxv-Orya': 'ପେ୍ଲମିସ୍', 'kxv-Telu': 'ప్లెమిస్', 'ky': 'фламандча', 'lb': 'Flämesch', 'lkt': 'Flemish Iyápi', 'lld': 'flamesc', 'lo': 'ຟລີມິຊ', 'lrc': 'فئلاماندی', 'lt': 'flamandų', 'lv': 'flāmu', 'mai': 'फ्लेमिश', 'mi': 'Tati Whēmirihi', 'mk': 'фламански', 'ml': 'ഫ്ലമിഷ്', 'mn': 'фламанд', 'mni': 'ফ্লেমিশ', 'mni-Beng': 'ফ্লেমিশ', 'mr': 'फ्लेमिश', 'ms': 'Flemish', 'mt': 'Fjamming', 'my': 'ဖလီမစ်ရှ်', 'mzn': 'فلمیش', 'nb': 'flamsk', 'ne': 'फ्लेमिस', 'nl': 'Vlaams', 'nn': 'flamsk', 'no': 'flamsk', 'nqo': 'ߝߌߟߊߡߊ߲ߞߊ߲', 'om': 'Flemish', 'or': 'ଫ୍ଲେମିଶ୍', 'pa': 'ਫਲੈਮਿਸ਼', 'pa-Guru': 'ਫਲੈਮਿਸ਼', 'pcm': 'Flẹ́mish Lángwej', 'pl': 'flamandzki', 'ps': 'فلېمېشي', 'pt': 'flamengo', 'qu': 'Flamenco Simi', 'rm': 'flam', 'ro': 'flamandă', 'ru': 'фламандский', 'sc': 'fiammingu', 'sd': 'فليمش', 'sd-Arab': 'فليمش', 'si': 'ෆ්ලෙමිශ්', 'sk': 'flámčina', 'sl': 'flamščina', 'smn': 'hollandkielâ (flaami)', 'so': 'Af faleemi', 'sq': 'flamandisht', 'sr': 'фламански', 'sr-Cyrl': 'фламански', 'sr-Latn': 'flamanski', 'sv': 'flamländska', 'sw': 'Kiflemi', 'syr': 'ܦܠܡܝܫܝܬ', 'ta': 'ஃப்லெமிஷ்', 'te': 'ఫ్లెమిష్', 'tg': 'Фламандӣ', 'th': 'เฟลมิช', 'ti': 'ፍላሚሽ', 'tk': 'flamand dili', 'to': 'lea fakahōlani-pelesiume', 'tr': 'Flamanca', 'tt': 'фламандча', 'uk': 'фламандська', 'ur': 'فلیمِش', 'uz': 'flamand', 'uz-Cyrl': 'фламандча', 'uz-Latn': 'flamand', 'vec': 'fiamingo', 'wae': 'Flämiš', 'wo': 'Belsig', 'xh': 'IsiFlemish', 'xnr': 'फ़्लेमिश', 'yi': 'פֿלעמיש', 'yo': 'Èdè Flemiṣi', 'yrl': 'faramẽgu', 'yue': '佛蘭芒文', 'yue-Hans': '佛兰芒文', 'yue-Hant': '佛蘭芒文', 'zh': '弗拉芒语', 'zh-Hans': '弗拉芒语', 'zh-Hant': '法蘭德斯文', 'zu': 'isi-Flemish'},
    'nla': {'en': 'Ngombale'},
    'nlc': {'en': 'Nalca'},
    'nle': {'en': 'East Nyala'},
    'nlg': {'en': 'Gela'},
    'nli': {'en': 'Grangali'},
    'nlj': {'en': 'Nyali'},
    'nlk': {'en': 'Ninia Yali'},
    'nll': {'en': 'Nihali'},
    'nlm': {'en': 'Mankiyali'},
    'nln': {'en': 'Durango Nahuatl'},
    'nlo': {'en': 'Ngul'},
    'nlq': {'en': 'Lao Naga'},
    'nlr': {'en': 'Ngarla'},
    'nlu': {'en': 'Nchumbulu'},
    'nlv': {'en': 'Orizaba Nahuatl'},
    'nlw': {'en': 'Walangama'},
    'nlx': {'en': 'Nahali'},
    'nly': {'en': 'Nyamal'},
    'nlz': {'en': 'Nalögo'},
    'nma': {'en': 'Maram Naga'},
    'nmb': {'en': 'Big Nambas'},
    'nmc': {'en': 'Ngam'},
    'nmd': {'en': 'Ndumu'},
    'nme': {'en': 'Mzieme Naga'},
    'nmf': {'en': 'Tangkhul Naga (India)'},
    'nmg': {'af': 'Kwasio', 'am': 'ክዋሲዮ', 'ar': 'كواسيو', 'as': 'কোৱাছিঅ’', 'ast': 'kwasio', 'az': 'kvasio', 'az-Cyrl': 'квасио', 'az-Latn': 'kvasio', 'bal-Latn': 'Kwásiu', 'be': 'нгумба', 'be-tarask': 'нгумба', 'bg': 'квасио', 'bgn': 'کوازیو', 'bn': 'কোয়াসিও', 'br': 'ngoumbeg', 'brx': 'क्वासिअ’', 'bs': 'kvasio', 'bs-Cyrl': 'квасио', 'bs-Latn': 'kvasio', 'ca': 'bissio', 'ca-ES-valencia': 'bissio', 'ccp': '𑄇𑄱𑄥𑄨𑄃𑄮', 'ce': 'квасио', 'chr': 'ᏆᏏᏲ', 'ckb': 'کواسیۆ', 'cs': 'kwasio', 'cy': 'Kwasio', 'da': 'kwasio', 'de': 'Kwasio', 'dsb': 'kwasio', 'el': 'Κβάσιο', 'el-polyton': 'Κβάσιο', 'en': 'Kwasio', 'es': 'kwasio', 'et': 'kwasio', 'eu': 'kwasiera', 'fa': 'کوازیو', 'ff-Adlm': '𞤐𞤺𞤵𞤥𞤦𞤢𞥄𞤪𞤫', 'fi': 'kwasio', 'fil': 'Kwasio', 'fo': 'kwasio', 'fr': 'ngoumba', 'fy': 'Ngumba', 'ga': 'Cuaiseois', 'gd': 'Kwasio', 'gl': 'kwasio', 'gu': 'ક્વાસિઓ', 'ha': 'Kwasio', 'he': 'קוואסיו', 'hi': 'क्वासिओ', 'hi-Latn': 'Kwasio', 'hr': 'kwasio', 'hsb': 'kwasio', 'ht': 'ngoumba', 'hu': 'ngumba', 'hy': 'կվասիո', 'ia': 'kwasio', 'id': 'Kwasio', 'ig': 'Kwasio', 'is': 'kwasio', 'it': 'kwasio', 'ja': 'クワシオ語', 'jv': 'Kwasio', 'ka': 'კვასიო', 'kea': 'kuazio', 'kgp': 'kivasijo', 'kk': 'квасио тілі', 'kk-Cyrl': 'квасио тілі', 'km': 'ក្វាស្យូ', 'kn': 'ಖ್ವಾಸಿಯೊ', 'ko': '크와시오어', 'kok': 'क्वासीयो', 'kok-Deva': 'क्वासीयो', 'ksh': 'Gyele', 'ku': 'kwasio', 'ky': 'квасиочо', 'lb': 'Kwasio', 'lo': 'ກວາຊີໂອ', 'lrc': 'کئڤاسیوٙ', 'lt': 'kvasių', 'lv': 'kvasio', 'mai': 'कवासियो', 'mi': 'Kuahio', 'mk': 'квазио', 'ml': 'ക്വേസിയോ', 'mn': 'квазио', 'mr': 'क्वासिओ', 'ms': 'Kwasio', 'mt': 'Kwasio', 'my': 'ကွာစီအို', 'mzn': 'کوئاسیو', 'nb': 'kwasio', 'ne': 'क्वासियो', 'nl': 'Ngumba', 'nn': 'kwasio', 'no': 'kwasio', 'nqo': 'ߒߜ߭ߎ߲ߓߊߞߊ߲', 'or': 'କୱାସିଓ', 'pa': 'ਕਵਾਸਿਓ', 'pa-Guru': 'ਕਵਾਸਿਓ', 'pcm': 'Kwasió Lángwej', 'pl': 'ngumba', 'ps': 'کواسیو', 'pt': 'kwasio', 'qu': 'Kwasio Ngumba Simi', 'rm': 'kwasio', 'ro': 'kwasio', 'ru': 'квасио', 'sc': 'kwasio', 'sd': 'ڪويسيو', 'sd-Arab': 'ڪويسيو', 'si': 'කුවාසිඔ', 'sk': 'kwasio', 'sl': 'kwasio', 'smn': 'kwasio', 'so': 'Kuwaasiyo', 'sq': 'kuasisht', 'sr': 'квасио', 'sr-Cyrl': 'квасио', 'sr-Latn': 'kvasio', 'sv': 'kwasio', 'sw': 'Kikwasio', 'ta': 'க்வாசியோ', 'te': 'క్వాసియె', 'th': 'กวาซิโอ', 'ti': 'ክዋስዮ', 'tk': 'kwasio dili', 'to': 'lea fakakuasio', 'tr': 'Kwasio', 'ug': 'كۋاسىيوچە', 'uk': 'квазіо', 'ur': 'کوايسو', 'uz': 'kvasio', 'uz-Cyrl': 'квасио', 'uz-Latn': 'kvasio', 'vec': 'kwazio', 'vi': 'Tiếng Kwasio', 'xnr': 'क्वासिओ', 'yo': 'Kíwáṣíò', 'yrl': 'kuwasiu', 'yue': '夸西奧文', 'yue-Hans': '夸西奥文', 'yue-Hant': '夸西奧文', 'zh': '夸西奥语', 'zh-Hans': '夸西奥语', 'zh-Hant': '夸西奧文', 'zu': 'isi-Kwasio'},
    'nmh': {'en': 'Monsang Naga'},
    'nmi': {'en': 'Nyam'},
    'nmj': {'en': 'Ngombe (Central African Republic)'},
    'nmk': {'en': 'Namakura'},
    'nml': {'en': 'Ndemli'},
    'nmm': {'en': 'Manangba'},
    'nmn': {'en': 'ǃXóõ'},
    'nmo': {'en': 'Moyon Naga'},
    'nmp': {'en': 'Nimanbur'},
    'nmq': {'en': 'Nambya'},
    'nmr': {'en': 'Nimbari'},
    'nms': {'en': 'Letemboi'},
    'nmt': {'en': 'Namonuito'},
    'nmu': {'en': 'Northeast Maidu'},
    'nmv': {'en': 'Ngamini'},
    'nmw': {'en': 'Nimoa'},
    'nmx': {'en': 'Nama (Papua New Guinea)'},
    'nmy': {'en': 'Namuyi'},
    'nmz': {'en': 'Nawdm'},
    'nn': {'af': 'Nuwe Noors', 'ak': 'Nɔwefoɔ Ninɔso', 'am': 'የኖርዌይ ናይኖርስክ', 'ar': 'النرويجية نينورسك', 'as': 'নৰৱেজিয়ান নায়নোৰ্স্ক', 'ast': 'noruegu Nynorsk', 'az': 'nünorsk norveç', 'az-Cyrl': 'нүнорск норвеч', 'az-Latn': 'nünorsk norveç', 'bal-Latn': 'Nárwiji Nókén', 'be': 'нарвежская (нюношк)', 'be-tarask': 'нарвежская (нюношк)', 'bg': 'норвежки (нюношк)', 'bgn': 'نارویی نی\u200cنوشکی', 'bn': 'নরওয়েজিয়ান নিনর্স্ক', 'br': 'norvegeg nynorsk', 'brx': 'नरवैजियान निनर्स्क', 'bs': 'norveški (Nynorsk)', 'bs-Cyrl': 'норвешки нинорск', 'bs-Latn': 'norveški (Nynorsk)', 'ca': 'noruec nynorsk', 'ca-ES-valencia': 'noruec nynorsk', 'ccp': '𑄚𑄧𑄢𑄴𑄃𑄮𑄠𑄬𑄎𑄩𑄠𑄚𑄴 𑄚𑄨𑄚𑄧𑄢𑄴𑄥𑄳𑄇𑄴', 'ce': 'норвегийн нюнорск', 'chr': 'ᏃᎵᏪᏥᏂ ᎾᎵᏍᎩ', 'ckb': 'نەرویژیی نینۆرسک', 'cs': 'norština (nynorsk)', 'cy': 'Norwyeg Nynorsk', 'da': 'nynorsk', 'de': 'Norwegisch (Nynorsk)', 'dsb': 'norwegske nynorsk', 'dz': 'ནོར་ཝེ་ཇི་ཡཱན་ནོརསཀ་ཁ', 'ee': 'nɔweigbe ninɔsk', 'el': 'Νορβηγικά Νινόρσκ', 'el-polyton': 'Νορβηγικά Νινόρσκ', 'en': 'Norwegian Nynorsk', 'eo': 'novnorvega', 'es': 'noruego nynorsk', 'et': 'uusnorra', 'eu': 'nynorsk (norvegiera)', 'fa': 'نروژی نی\u200cنُشک', 'ff-Adlm': '𞤐𞤮𞤪𞤱𞤫𞤶𞤭𞤴𞤢𞤲𞤪𞤫 𞤙𞤮𞤪𞤧𞤳', 'fi': 'norjan nynorsk', 'fil': 'Norwegian Nynorsk', 'fo': 'nýnorskt', 'fr': 'norvégien nynorsk', 'fur': 'norvegjês nynorsk', 'fy': 'Noors - Nynorsk', 'ga': 'Nua-Ioruais', 'gd': 'Nynorsk na Nirribhidh', 'gl': 'noruegués nynorsk', 'gsw': 'Norwegisch Nynorsk', 'gu': 'નોર્વેજિયન નાયનૉર્સ્ક', 'ha': 'Norwegian Nynorsk', 'he': 'נורווגית חדשה', 'hi': 'नॉर्वेजियाई नॉयनॉर्स्क', 'hi-Latn': 'Norwegian Nynorsk', 'hr': 'norveški nynorsk', 'hsb': 'norwegšćina (nynorsk)', 'ht': 'norvégien nynorsk', 'hu': 'norvég (nynorsk)', 'hy': 'նոր նորվեգերեն', 'ia': 'norvegiano nynorsk', 'id': 'Nynorsk Norwegia', 'ig': 'Norwegian Nynorsk', 'is': 'nýnorska', 'it': 'norvegese nynorsk', 'ja': 'ノルウェー語(ニーノシュク)', 'jv': 'Nynorsk Norwegia', 'ka': 'ნორვეგიული ნიუნორსკი', 'kea': 'norueges nynorsk', 'kgp': 'nĩnãrsig nãrovegej', 'kk': 'норвегиялық нюнорск тілі', 'kk-Cyrl': 'норвегиялық нюнорск тілі', 'km': 'ន័រវែស នីនូស', 'kn': 'ನಾರ್ವೇಜಿಯನ್ ನೈನಾರ್ಸ್ಕ್', 'ko': '노르웨이어(니노르스크)', 'kok': 'नॉर्वेजियन नायनोर्स्क', 'kok-Deva': 'नॉर्वेजियन नायनोर्स्क', 'ks': 'ناروییَن نَے نورسک', 'ks-Arab': 'ناروییَن نَے نورسک', 'ksh': 'Neu Norrwehjesch', 'ku': 'norwecî (nynorsk)', 'ky': 'норвежче (нинорск)', 'lb': 'Norwegesch Nynorsk', 'lld': 'norvegesc nynorsk', 'lo': 'ນໍເວຈຽນ ນີນອກ', 'lrc': 'نورڤئجی نینورسک', 'lt': 'naujoji norvegų', 'lv': 'jaunnorvēģu', 'mai': 'नार्वेजियन निनोरेस्क', 'mi': 'Nīnōka Nōwei', 'mk': 'норвешки нинорск', 'ml': 'നോർവീജിയൻ നൈനോർക്\u200cസ്', 'mn': 'норвегийн нинорск', 'mr': 'नॉर्वेजियन न्योर्स्क', 'ms': 'Nynorsk Norway', 'mt': 'Ninorsk Norveġiż', 'my': 'နော်ဝေ နီးနောစ်', 'mzn': 'نروژی نینورسک', 'nb': 'norsk nynorsk', 'ne': 'नर्वेली नाइनोर्स्क', 'nl': 'Noors - Nynorsk', 'nn': 'norsk nynorsk', 'no': 'norsk nynorsk', 'nqo': 'ߣߐߙߑߝ߭ߍߖ߭ ߢߙߐߛߌߞߊ߲', 'om': 'Afaan Norwegian', 'or': 'ନରୱେଜିଆନ୍ ନିନର୍ସ୍କ୍\u200c', 'pa': 'ਨਾਰਵੇਜਿਆਈ ਨਿਓਨੌਰਸਕ', 'pa-Guru': 'ਨਾਰਵੇਜਿਆਈ ਨਿਓਨੌਰਸਕ', 'pcm': 'Nọwẹ́jiá Niúnọsk', 'pl': 'norweski (nynorsk)', 'ps': 'ناروېئي (نائنورسک)', 'pt': 'nynorsk norueguês', 'qu': 'Noruego Nynorsk Simi', 'rm': 'norvegiais nynorsk', 'ro': 'norvegiană nynorsk', 'ru': 'нюнорск', 'rw': 'Inyenoruveji (Nyonorusiki)', 'sc': 'norvegesu nynorsk', 'sd': 'نارويائي نيوناسڪ', 'sd-Arab': 'نارويائي نيوناسڪ', 'se': 'ođđadárogiella', 'si': 'නෝර්වීජියානු නයිනෝර්ස්ක්', 'sk': 'nórčina (nynorsk)', 'sl': 'novonorveščina', 'smn': 'tárukielâ nynorsk', 'so': 'Nawriijiga Nynorsk', 'sq': 'norvegjishte nynorsk', 'sr': 'норвешки нинорск', 'sr-Cyrl': 'норвешки нинорск', 'sr-Latn': 'norveški ninorsk', 'sv': 'nynorska', 'sw': 'Kinorwe cha Nynorsk', 'ta': 'நார்வேஜியன் நியூநார்ஸ்க்', 'te': 'నార్వేజియాన్ న్యోర్స్క్', 'th': 'นอร์เวย์นีนอสก์', 'ti': 'ኖርወያዊ ናይኖርስክ', 'tk': 'norwegiýa nýunorsk dili', 'to': 'lea fakanoauē-ninosiki', 'tr': 'Norveççe Nynorsk', 'ug': 'يېڭى نورۋېگچە', 'uk': 'норвезька (нюношк)', 'ur': 'نارویجین نینورسک', 'uz': 'norveg-nyunorsk', 'uz-Cyrl': 'норвегча нюнорск', 'uz-Latn': 'norveg-nyunorsk', 'vec': 'norvejeze nynorsk', 'vi': 'Tiếng Na Uy (Nynorsk)', 'wae': 'Norwegiš Nynorsk', 'xnr': 'नॉर्वेजियाई नॉयनॉर्स्क', 'yi': 'נײַ־נארוועגיש', 'yo': 'Nọ́ọ́wè Nínọ̀sìkì', 'yrl': 'ninorisiki nuruegei', 'yue': '耐諾斯克挪威文', 'yue-Hans': '耐诺斯克挪威文', 'yue-Hant': '耐諾斯克挪威文', 'zh': '挪威尼诺斯克语', 'zh-Hans': '挪威尼诺斯克语', 'zh-Hant': '新挪威文', 'zu': 'isi-Norwegian Nynorsk'},
    'nna': {'en': 'Nyangumarta'},
    'nnb': {'en': 'Nande'},
    'nnc': {'en': 'Nancere'},
    'nnd': {'en': 'West Ambae'},
    'nne': {'en': 'Ngandyera'},
    'nnf': {'en': 'Ngaing'},
    'nng': {'en': 'Maring Naga'},
    'nnh': {'af': 'Ngiemboon', 'am': 'ኒጊምቡን', 'ar': 'لغة النجيمبون', 'as': 'নিয়েম্বোন', 'ast': 'ngiemboon', 'az': 'ngiemboon', 'az-Cyrl': 'нҝиембоон', 'az-Latn': 'ngiemboon', 'bal-Latn': 'Ngembun', 'be': 'нг’ембон', 'be-tarask': 'нг’ембон', 'bg': 'нгиембун', 'bgn': 'انگی\u200cایمبونی', 'bn': 'নগিয়েম্বুন', 'br': 'ngiemboon', 'brx': 'न्गीएम्बून', 'bs': 'ngiembon', 'bs-Cyrl': 'нгиембун', 'bs-Latn': 'ngiembon', 'ca': 'ngiemboon', 'ca-ES-valencia': 'ngiemboon', 'ccp': '𑄚𑄨𑄋𑄴𑄉𑄬𑄟𑄴𑄝𑄪𑄚𑄴', 'ce': 'нгиембунд', 'chr': 'ᎾᏥᏰᎹᏊᏂ', 'ckb': 'نگیمبوون', 'cs': 'ngiemboon', 'cy': 'Ngiemboon', 'da': 'ngiemboon', 'de': 'Ngiemboon', 'dsb': 'ngiemboon', 'el': 'Νγκιεμπούν', 'el-polyton': 'Νγκιεμπούν', 'en': 'Ngiemboon', 'es': 'ngiemboon', 'et': 'ngiembooni', 'eu': 'ngiemboonera', 'fa': 'نیامبون', 'ff-Adlm': '𞤐𞤶𞤢𞤥𞤦𞤵𞥅𞤪𞤫', 'fi': 'ngiemboon', 'fil': 'Ngiemboon', 'fo': 'ngiemboon', 'fr': 'ngiemboon', 'fy': 'Ngiemboon', 'ga': 'Ngiemboon', 'gd': 'Ngiemboon', 'gl': 'ngiemboon', 'gu': 'નીએમબુન', 'ha': 'Ngiemboon', 'he': 'נגיאמבון', 'hi': 'गैम्बू', 'hi-Latn': 'Ngiemboon', 'hr': 'ngiemboon', 'hsb': 'ngiemboon', 'ht': 'ngiemboon', 'hu': 'ngiemboon', 'hy': 'նգիեմբուն', 'ia': 'ngiemboon', 'id': 'Ngiemboon', 'ig': 'Ngiemboon', 'is': 'ngiemboon', 'it': 'ngiemboon', 'ja': 'ンジエムブーン語', 'jv': 'Ngiemboon', 'ka': 'ნგიმბუნი', 'kgp': 'gijẽmun', 'kk': 'нгиембун тілі', 'kk-Cyrl': 'нгиембун тілі', 'km': 'ងៀមប៊ូន', 'kn': 'ನಿಂಬೂನ್', 'ko': '느기엠본어', 'kok': 'गियेम्बून', 'kok-Deva': 'गियेम्बून', 'ksh': 'Njijembohn', 'ku': 'ngiemboon', 'ky': 'нгимбунча', 'lb': 'Ngiemboon', 'lo': 'ຈີ່ມບູນ', 'lt': 'ngiembūnų', 'lv': 'ngjembūnu', 'mai': 'जियमबून', 'mi': 'Nekiepūna', 'mk': 'нгиембун', 'ml': 'ഗീംബൂൺ', 'mn': 'нгиембүүн', 'mr': 'जिएम्बून', 'ms': 'Ngiemboon', 'mt': 'Ngiemboon', 'my': 'အွန်ရဲဘွန်း', 'nb': 'ngiemboon', 'ne': 'न्गिएम्बुन', 'nl': 'Ngiemboon', 'nn': 'ngiemboon', 'nnh': 'Shwóŋò ngiembɔɔn', 'no': 'ngiemboon', 'nqo': 'ߒߖߋ߲ߓߎ߲ߞߊ߲', 'or': 'ନାଗିମବୋନ୍', 'pa': 'ਨਿਓਮਬੂਨ', 'pa-Guru': 'ਨਿਓਮਬੂਨ', 'pcm': 'Ngiẹ́mbọn Lángwej', 'pl': 'ngiemboon', 'ps': 'نایجیمون', 'pt': 'ngiemboon', 'qu': 'Ngiemboon Simi', 'rm': 'ngienboon', 'ro': 'ngiemboon', 'ru': 'нгиембунд', 'sc': 'ngiemboon', 'sd': 'نغيمبون', 'sd-Arab': 'نغيمبون', 'si': 'න්ගියාම්බූන්', 'sk': 'ngiemboon', 'sl': 'ngiemboonščina', 'smn': 'ngiemboon', 'so': 'Ingiyembuun', 'sq': 'ngiembunisht', 'sr': 'нгиембун', 'sr-Cyrl': 'нгиембун', 'sr-Latn': 'ngiembun', 'sv': 'bamileké-ngiemboon', 'sw': 'Lugha ya Ngiemboon', 'ta': 'நெகெய்ம்பூன்', 'te': 'గింబూన్', 'th': 'จีมบูน', 'ti': 'ኤንጌምቡን', 'tk': 'ngembun dili', 'to': 'lea fakangiemipōni', 'tr': 'Ngiemboon', 'ug': 'نگېمبۇنچە', 'uk': 'нгємбун', 'ur': 'نگیمبون', 'uz': 'ngiyembun', 'uz-Cyrl': 'нгиембун', 'uz-Latn': 'ngiyembun', 'vec': 'ngienbon', 'vi': 'Tiếng Ngiemboon', 'xnr': 'गैम्बू', 'yo': 'Ngiembùnù', 'yrl': 'ĩgiẽbũ', 'yue': '恩甘澎文', 'yue-Hans': '恩甘澎文', 'yue-Hant': '恩甘澎文', 'zh': '恩甘澎语', 'zh-Hans': '恩甘澎语', 'zh-Hant': '恩甘澎文', 'zu': 'isi-Ngiemboon'},
    'nni': {'en': 'North Nuaulu'},
    'nnj': {'en': 'Nyangatom'},
    'nnk': {'en': 'Nankina'},
    'nnl': {'en': 'Northern Rengma Naga'},
    'nnm': {'en': 'Namia'},
    'nnn': {'en': 'Ngete'},
    'nnp': {'en': 'Wancho Naga'},
    'nnq': {'en': 'Ngindo'},
    'nnr': {'en': 'Narungga'},
    'nns': {'en': 'Ningye'},
    'nnt': {'en': 'Nanticoke'},
    'nnu': {'en': 'Dwang'},
    'nnv': {'en': 'Nugunu (Australia)'},
    'nnw': {'en': 'Southern Nuni'},
    'nnx': {'en': 'Ngong'},
    'nny': {'en': 'Nyangga'},
    'nnz': {'en': "Nda'nda'"},
    'no': {'af': 'Noors', 'ak': 'Nɔwefoɔ kasa', 'am': 'ኖርዌይኛ', 'ar': 'النرويجية', 'as': 'নৰৱেজিয়ান', 'ast': 'noruegu', 'az': 'norveç', 'az-Latn': 'norveç', 'bal-Latn': 'Nárwiji', 'be': 'нарвежская', 'be-tarask': 'нарвежская', 'bg': 'норвежки', 'bn': 'নরওয়েজীয়', 'br': 'norvegeg', 'brx': 'नरवैजियान', 'bs': 'norveški', 'bs-Cyrl': 'норвешки', 'bs-Latn': 'norveški', 'ca': 'noruec', 'ca-ES-valencia': 'noruec', 'ccp': '𑄚𑄧𑄢𑄴𑄃𑄮𑄠𑄬𑄎𑄩𑄠𑄧', 'chr': 'ᏃᎵᏪᏥᏂ', 'ckb': 'نۆروێژی', 'cs': 'norština', 'cy': 'Norwyeg', 'da': 'norsk', 'de': 'Norwegisch', 'dsb': 'norwegšćina', 'dz': 'ནོར་ཝི་ཇི་ཡན་ཁ', 'ee': 'nɔweigbe', 'el': 'Νορβηγικά', 'el-polyton': 'Νορβηγικά', 'en': 'Norwegian', 'eo': 'norvega', 'es': 'noruego', 'et': 'norra', 'eu': 'norvegiera', 'fa': 'نروژی', 'ff-Adlm': '𞤐𞤮𞤪𞤱𞤫𞤶𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'norja', 'fil': 'Norwegian', 'fo': 'norskt', 'fr': 'norvégien', 'fur': 'norvegjês', 'fy': 'Noors', 'ga': 'Ioruais', 'gd': 'Nirribhis', 'gl': 'noruegués', 'gsw': 'Norwegisch', 'gu': 'નૉર્વેજીયન', 'ha': 'Harhsen Norway', 'he': 'נורווגית', 'hi': 'नॉर्वेजियाई', 'hi-Latn': 'Norwegian', 'hr': 'norveški', 'hsb': 'norwegšćina', 'ht': 'norvégien', 'hu': 'norvég', 'hy': 'նորվեգերեն', 'ia': 'norvegiano', 'id': 'Norwegia', 'ig': 'Norwegian', 'is': 'norska', 'it': 'norvegese', 'ja': 'ノルウェー語', 'jv': 'Norwegia', 'ka': 'ნორვეგიული', 'kgp': 'nãrovegej', 'kk': 'норвег тілі', 'kk-Cyrl': 'норвег тілі', 'km': 'ន័រវែស', 'kn': 'ನಾರ್ವೇಜಿಯನ್', 'ko': '노르웨이어', 'kok': 'नोर्वेजियन', 'kok-Deva': 'नोर्वेजियन', 'ks': 'ناروییَن', 'ks-Arab': 'ناروییَن', 'ksh': 'Norrwehjesch', 'ku': 'norwecî', 'ky': 'норвежче', 'lb': 'Norwegesch', 'lld': 'norvegesc', 'lo': 'ນໍເວຈຽນ', 'lt': 'norvegų', 'lv': 'norvēģu', 'mai': 'नार्वेजियन', 'mi': 'Nōwei', 'mk': 'норвешки', 'ml': 'നോർവീജിയൻ', 'mn': 'норвег', 'mr': 'नॉर्वेजियन', 'ms': 'Norway', 'mt': 'Norveġiż', 'my': 'နော်ဝေ', 'nb': 'norsk', 'ne': 'नर्वेली', 'nl': 'Noors', 'nn': 'norsk', 'no': 'norsk', 'nqo': 'ߣߐߙߑߝ߭ߍߖ߭ߌߞߊ߲', 'om': 'Afaan Norweyii', 'or': 'ନରୱେଜିଆନ୍', 'pa': 'ਨਾਰਵੇਜਿਆਈ', 'pa-Guru': 'ਨਾਰਵੇਜਿਆਈ', 'pcm': 'Nọ́wẹ́jiá Lángwej', 'pl': 'norweski', 'ps': 'ناروېئي', 'pt': 'norueguês', 'qu': 'Noruego Simi', 'rm': 'norvegiais', 'ro': 'norvegiană', 'ru': 'норвежский', 'rw': 'Ikinoruveji', 'sc': 'norvegesu', 'sd': 'نارويجيائي', 'sd-Arab': 'نارويجيائي', 'se': 'dárogiella', 'si': 'නෝර්වීජියානු', 'sk': 'nórčina', 'sl': 'norveščina', 'smn': 'tárukielâ', 'so': 'Nawriiji', 'sq': 'norvegjisht', 'sr': 'норвешки', 'sr-Cyrl': 'норвешки', 'sr-Latn': 'norveški', 'sv': 'norska', 'sw': 'Kinorwe', 'syr': 'ܢܘܪܒܝܓܐܝܬ', 'ta': 'நார்வேஜியன்', 'te': 'నార్వేజియన్', 'tg': 'норвегӣ', 'th': 'นอร์เวย์', 'ti': 'ኖርወይኛ', 'tk': 'norweg dili', 'to': 'lea fakanouaē', 'tr': 'Norveççe', 'ug': 'نورۋېگچە', 'uk': 'норвезька', 'ur': 'نارویجین', 'uz': 'norveg', 'uz-Latn': 'norveg', 'vec': 'norvejeze', 'vi': 'Tiếng Na Uy', 'wo': 'Nerwesiye', 'xnr': 'नॉर्वेजियाई', 'yi': 'נארוועגיש', 'yo': 'Èdè Norway', 'yrl': 'nuruegei', 'yue': '挪威文', 'yue-Hans': '挪威文', 'yue-Hant': '挪威文', 'zh': '挪威语', 'zh-Hans': '挪威语', 'zh-Hant': '挪威文', 'zu': 'isi-Norwegian'},
    'noa': {'en': 'Woun Meu'},
    'noc': {'en': 'Nuk'},
    'nod': {'en': 'Northern Thai'},
    'noe': {'en': 'Nimadi'},
    'nof': {'en': 'Nomane'},
    'nog': {'af': 'Nogai', 'am': 'ኖጋይ', 'ar': 'النوجاي', 'as': 'নোগাই', 'ast': 'nogai', 'az': 'noqay', 'az-Cyrl': 'ногај', 'az-Latn': 'noqay', 'be': 'нагайская', 'be-tarask': 'нагайская', 'bg': 'ногаи', 'bgn': 'نغایی', 'bn': 'নোগাই', 'br': 'nogay', 'brx': 'न’गाई', 'bs': 'nogai', 'bs-Cyrl': 'ногаи', 'bs-Latn': 'nogai', 'ca': 'nogai', 'ca-ES-valencia': 'nogai', 'ccp': '𑄚𑄮𑄉𑄭', 'ce': 'ногӀийн', 'chr': 'ᏃᎦᏱ', 'ckb': 'نۆگای', 'cs': 'nogajština', 'cy': 'Nogai', 'da': 'nogai', 'de': 'Nogai', 'dsb': 'nogaišćina', 'el': 'Νογκάι', 'el-polyton': 'Νογκάι', 'en': 'Nogai', 'es': 'nogai', 'et': 'nogai', 'eu': 'nogaiera', 'fa': 'نغایی', 'ff-Adlm': '𞤐𞤮𞤺𞤢𞤭𞥅𞤪𞤫', 'fi': 'nogai', 'fil': 'Nogai', 'fo': 'nogai', 'fr': 'nogaï', 'fy': 'Nogai', 'ga': 'Nógaeis', 'gd': 'Nogai', 'gl': 'nogai', 'gsw': 'Nogaisch', 'gu': 'નોગાઇ', 'ha': 'Harshen Nogai', 'he': 'נוגאי', 'hi': 'नोगाई', 'hi-Latn': 'Nogai', 'hr': 'nogajski', 'hsb': 'nogaišćina', 'ht': 'nogaï', 'hu': 'nogaj', 'hy': 'նոգայերեն', 'ia': 'nogai', 'id': 'Nogai', 'ig': 'Nogai', 'is': 'nógaí', 'it': 'nogai', 'ja': 'ノガイ語', 'jv': 'Nogai', 'ka': 'ნოღაური', 'kgp': 'nãgaj', 'kk': 'ноғай тілі', 'kk-Cyrl': 'ноғай тілі', 'km': 'ណូហ្គៃ', 'kn': 'ನೊಗಾಯ್', 'ko': '노가이어', 'kok': 'नोगाय', 'kok-Deva': 'नोगाय', 'ks': 'نوگاے', 'ks-Arab': 'نوگاے', 'ksh': 'Nojal', 'ku': 'nogayî', 'ky': 'ногайча', 'lb': 'Nogai', 'lo': 'ນໍໄກ', 'lt': 'nogų', 'lv': 'nogaju', 'mai': 'नोगई', 'mi': 'Nōkai', 'mk': 'ногајски', 'ml': 'നോഗൈ', 'mn': 'ногаи', 'mr': 'नोगाई', 'ms': 'Nogai', 'mt': 'Nogai', 'my': 'နိုဂိုင်', 'nb': 'nogaisk', 'ne': 'नोगाइ', 'nl': 'Nogai', 'nn': 'nogai', 'no': 'nogaisk', 'or': 'ନୋଗାଇ', 'pa': 'ਨੋਗਾਈ', 'pa-Guru': 'ਨੋਗਾਈ', 'pcm': 'Nogái Lángwej', 'pl': 'nogajski', 'ps': 'نوګی', 'pt': 'nogai', 'qu': 'Nogai Simi', 'rm': 'nogai', 'ro': 'nogai', 'ru': 'ногайский', 'sah': 'Нагаайдыы', 'sc': 'nogai', 'sd': 'نوگائي', 'sd-Arab': 'نوگائي', 'si': 'නොගායි', 'sk': 'nogajčina', 'sl': 'nogajščina', 'smn': 'nogaikielâ', 'so': 'Nogai', 'sq': 'nogajisht', 'sr': 'ногајски', 'sr-Cyrl': 'ногајски', 'sr-Latn': 'nogajski', 'sv': 'nogai', 'sw': 'Kinogai', 'ta': 'நோகை', 'te': 'నోగై', 'th': 'โนไก', 'ti': 'ኖጋይኛ', 'tk': 'nogaý dili', 'to': 'lea fakanokai', 'tr': 'Nogayca', 'ug': 'نوغايچە', 'uk': 'ногайська', 'ur': 'نوگائی', 'uz': 'no‘g‘ay', 'uz-Latn': 'no‘g‘ay', 'vec': 'nogài', 'vi': 'Tiếng Nogai', 'xnr': 'नोगाई', 'yo': 'Èdè Nogai', 'yrl': 'nugai', 'yue': '諾蓋文', 'yue-Hans': '诺盖文', 'yue-Hant': '諾蓋文', 'zh': '诺盖语', 'zh-Hans': '诺盖语', 'zh-Hant': '諾蓋文', 'zu': 'isi-Nogai'},
    'noh': {'en': 'Nomu'},
    'noi': {'en': 'Noiri'},
    'noj': {'en': 'Nonuya'},
    'nok': {'en': 'Nooksack'},
    'nol': {'en': 'Nomlaki'},
    'nom': {'en': 'Nocamán'},
    'non': {'ar': 'النورس القديم', 'ast': 'noruegu antiguu', 'az': 'qədim nors', 'az-Latn': 'qədim nors', 'be': 'старанарвежская', 'be-tarask': 'старанарвежская', 'bg': 'старонорвежки', 'bn': 'প্রাচীন নর্স', 'br': 'hennorseg', 'brx': 'पुरानी नॉर्स्', 'bs': 'staronordijski', 'bs-Cyrl': 'стари норски', 'bs-Latn': 'staronordijski', 'ca': 'nòrdic antic', 'ca-ES-valencia': 'nòrdic antic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄴 𑄚𑄧𑄢𑄴𑄥𑄧', 'cs': 'norština historická', 'cy': 'Hen Norseg', 'da': 'oldislandsk', 'de': 'Altnordisch', 'el': 'Παλαιά Νορβηγικά', 'el-polyton': 'Παλαιά Νορβηγικά', 'en': 'Old Norse', 'es': 'nórdico antiguo', 'et': 'vanapõhja', 'fa': 'نرس باستان', 'fi': 'muinaisnorja', 'fr': 'vieux norrois', 'fur': 'vieri norvegjês', 'fy': 'Aldnoarsk', 'ga': 'Sean-Lochlainnis', 'gd': 'Seann-Lochlannais', 'gsw': 'Altnordisch', 'gu': 'જૂની નોર્સ', 'he': '\u200fנורדית עתיקה', 'hi': 'पुराना नॉर्स', 'hi-Latn': 'Old Norse', 'hr': 'staronorveški', 'ht': 'vieux norrois', 'hu': 'óskandináv', 'hy': 'հին նորվեգերեն', 'id': 'Norse Kuno', 'is': 'norræna', 'it': 'norse antico', 'ja': '古ノルド語', 'ka': 'ძველსკანდინავიური', 'kgp': 'nãrniko arkajku', 'kn': 'ಪ್ರಾಚೀನ ನೋರ್ಸ್', 'ko': '고대 노르웨이어', 'ks': 'پرون نارسی', 'ks-Arab': 'پرون نارسی', 'lb': 'Alnordesch', 'lo': 'ນໍໂບຮານ', 'lt': 'senoji norsų', 'lv': 'sennorvēģu', 'mk': 'старонордиски', 'ml': 'പഴയ നോഴ്\u200cസ്', 'mr': 'पुरातन नॉर्स', 'mt': 'Nors Antik', 'nb': 'gammelnorsk', 'ne': 'पुरानो नोर्से', 'nl': 'Oudnoors', 'nn': 'gammalnorsk', 'no': 'gammelnorsk', 'or': 'ପୁରୁଣା ନର୍ସ', 'pl': 'staronordyjski', 'pt': 'nórdico arcaico', 'rm': 'nordic vegl', 'ro': 'nordică veche', 'ru': 'старонорвежский', 'sk': 'stará nórčina', 'sl': 'stara nordijščina', 'smn': 'toovláš tárukielâ', 'sr': 'старонордијски', 'sr-Cyrl': 'старонордијски', 'sr-Latn': 'staronordijski', 'sv': 'fornnordiska', 'ta': 'பழைய நோர்ஸ்', 'te': 'ప్రాచిన నోర్స్', 'th': 'นอร์สโบราณ', 'to': 'lea fakanoauē-motuʻa', 'tr': 'Eski Nors dili', 'ug': 'قەدىمكى نورۋېگچە', 'uk': 'давньонорвезька', 'vi': 'Tiếng Na Uy cổ', 'yrl': 'núridiku arkaiku', 'yue': '古諾爾斯文', 'yue-Hans': '古诺尔斯文', 'yue-Hant': '古諾爾斯文', 'zh': '古诺尔斯语', 'zh-Hans': '古诺尔斯语', 'zh-Hant': '古諾爾斯文'},
    'noo': {'en': 'Nootka'},
    'nop': {'en': 'Numanggang'},
    'noq': {'en': 'Ngongo'},
    'nos': {'en': 'Eastern Nisu'},
    'not': {'en': 'Nomatsiguenga'},
    'nou': {'en': 'Ewage-Notu'},
    'nov': {'ast': 'novial', 'br': 'novial', 'ca': 'novial', 'ca-ES-valencia': 'novial', 'cs': 'novial', 'de': 'Novial', 'en': 'Novial', 'et': 'noviaal', 'fi': 'novial', 'fr': 'novial', 'gd': 'Novial', 'hi-Latn': 'Novial', 'ht': 'novial', 'it': 'novial', 'ja': 'ノヴィアル', 'lb': 'Novial', 'lt': 'novial', 'mk': 'новијал', 'nb': 'novial', 'ne': 'नोभियल', 'nl': 'Novial', 'nn': 'novial', 'no': 'novial', 'pl': 'novial', 'sv': 'novial', 'th': 'โนเวียล', 'to': 'lea fakanoviale', 'tr': 'Novial', 'yue': '諾維亞文', 'yue-Hans': '诺维亚文', 'yue-Hant': '諾維亞文', 'zh-Hant': '諾維亞文'},
    'now': {'en': 'Nyambo'},
    'noy': {'en': 'Noy'},
    'noz': {'en': 'Nayi'},
    'npa': {'en': 'Nar Phu'},
    'npb': {'en': 'Nupbikha'},
    'npg': {'en': 'Ponyo-Gongwang Naga'},
    'nph': {'en': 'Phom Naga'},
    'npi': {'en': 'Nepali (individual language)'},
    'npl': {'en': 'Southeastern Puebla Nahuatl'},
    'npn': {'en': 'Mondropolon'},
    'npo': {'en': 'Pochuri Naga'},
    'nps': {'en': 'Nipsan'},
    'npu': {'en': 'Puimei Naga'},
    'npx': {'en': 'Noipx'},
    'npy': {'en': 'Napu'},
    'nqg': {'en': 'Southern Nago'},
    'nqk': {'en': 'Kura Ede Nago'},
    'nql': {'en': 'Ngendelengo'},
    'nqm': {'en': 'Ndom'},
    'nqn': {'en': 'Nen'},
    'nqo': {'af': 'N’Ko', 'ak': 'Nko', 'am': 'ንኮ', 'ar': 'أنكو', 'as': 'এন্কো', 'ast': 'n’ko', 'az': 'nko', 'az-Cyrl': 'нго', 'az-Latn': 'nko', 'bal-Latn': 'Nko', 'be': 'нко', 'be-tarask': 'нко', 'bg': 'нко', 'bgn': 'نکوی', 'bn': 'এন’কো', 'br': 'nkoeg', 'brx': 'नंक’', 'bs': 'nko', 'bs-Cyrl': 'н’ко', 'bs-Latn': 'nko', 'ca': 'n’Ko', 'ca-ES-valencia': 'n’Ko', 'ccp': '𑄆𑄚𑄴𑄇𑄮', 'ce': 'нко', 'chr': 'ᎾᎪ', 'ckb': 'نکۆ', 'cs': 'n’ko', 'cy': 'N’Ko', 'da': 'n’ko', 'de': 'N’Ko', 'dsb': 'n’ko', 'el': 'Ν’Κο', 'el-polyton': 'Ν’Κο', 'en': 'N’Ko', 'es': 'n’ko', 'et': 'nkoo', 'eu': 'n’koera', 'fa': 'نکو', 'ff-Adlm': '𞤐𞤳𞤮𞥅𞤪𞤫', 'fi': 'n’ko', 'fil': 'N’Ko', 'fo': 'nʼko', 'fr': 'n’ko', 'fy': 'N’ko', 'ga': 'N-cóis', 'gd': 'N’Ko', 'gl': 'n’ko', 'gsw': 'N’Ko', 'gu': 'એન’કો', 'ha': 'N’Ko', 'he': 'נ׳קו', 'hi': 'एन्को', 'hi-Latn': 'N’Ko', 'hr': 'n’ko', 'hsb': 'n’ko', 'ht': 'n’ko', 'hu': 'n’kó', 'hy': 'նկո', 'ia': 'n’ko', 'id': 'N’Ko', 'ig': 'N’Ko', 'is': 'n’ko', 'it': 'n’ko', 'ja': 'ンコ語', 'jv': 'N’Ko', 'ka': 'ნკო', 'kgp': 'nyko', 'kk': 'нко тілі', 'kk-Cyrl': 'нко тілі', 'km': 'នគោ', 'kn': 'ಎನ್\u200cಕೋ', 'ko': '응코어', 'kok': 'नको', 'kok-Deva': 'नको', 'ks': 'این کو', 'ks-Arab': 'این کو', 'ku': 'n’Ko', 'ky': 'нкочо', 'lb': 'N’Ko', 'lo': 'ເອັນໂກ', 'lrc': 'نئکوٙ', 'lt': 'enko', 'lv': 'nko', 'mai': 'एनको', 'mi': 'Unukō', 'mk': 'нко', 'ml': 'ഇൻകോ', 'mn': 'нко', 'mr': 'एन्को', 'ms': 'N’ko', 'mt': 'N’Ko', 'my': 'အွန်ကို', 'mzn': 'نئکو', 'nb': 'nʼko', 'ne': 'नको', 'nl': 'N’Ko', 'nn': 'n’ko', 'no': 'nʼko', 'nqo': 'ߒߞߏ', 'or': 'ଏନକୋ', 'pa': 'ਐਂਕੋ', 'pa-Guru': 'ਐਂਕੋ', 'pcm': 'N’Ko Lángwej', 'pl': 'n’ko', 'ps': 'نکو', 'pt': 'n’ko', 'qu': 'N’Ko Simi', 'rm': 'n’ko', 'ro': 'n’ko', 'ru': 'нко', 'sc': 'n’ko', 'sd': 'نڪو', 'sd-Arab': 'نڪو', 'si': 'එන්‘කෝ', 'sk': 'n’ko', 'sl': 'n’ko', 'smn': 'n’ko', 'so': 'N’Ko', 'sq': 'nkoisht', 'sr': 'нко', 'sr-Cyrl': 'нко', 'sr-Latn': 'nko', 'sv': 'n-kå', 'sw': 'N’Ko', 'ta': 'என்‘கோ', 'te': 'న్కో', 'th': 'เอ็นโก', 'ti': 'ኤንኮ', 'tk': 'nko dili', 'to': 'lea fakanikō', 'tr': 'N’Ko', 'ug': 'نىكوچە', 'uk': 'нко', 'ur': 'اینکو', 'uz': 'nko', 'uz-Cyrl': 'нко', 'uz-Latn': 'nko', 'vec': 'n’ko', 'vi': 'Tiếng N’Ko', 'xnr': 'एन्को', 'yo': 'Èdè Nko', 'yrl': 'nikú', 'yue': '曼德文字 (N’Ko)', 'yue-Hans': '曼德文字 (N’Ko)', 'yue-Hant': '曼德文字 (N’Ko)', 'zh': '西非书面文字', 'zh-Hans': '西非书面文字', 'zh-Hant': '曼德文字 (N’Ko)', 'zu': 'isi-N’Ko'},
    'nqq': {'en': 'Kyan-Karyaw Naga'},
    'nqt': {'en': 'Nteng'},
    'nqy': {'en': 'Akyaung Ari Naga'},
    'nr': {'af': 'Suid-Ndebele', 'am': 'ደቡብ ንደቤሌ', 'ar': 'النديبيل الجنوبي', 'as': 'দক্ষিণ দেবেল', 'ast': 'ndebele del sur', 'az': 'cənubi ndebele', 'az-Cyrl': 'ҹәнуби ндебеле', 'az-Latn': 'cənubi ndebele', 'bal-Latn': 'Zerbári Nedebéli', 'be': 'паўднёвая ндэбеле', 'be-tarask': 'паўднёвая ндэбеле', 'bg': 'южен ндебеле', 'bgn': 'جهلسرین انده\u200cبله\u200cای', 'bn': 'দক্ষিণ এনডেবেলে', 'br': 'ndebele ar Su', 'brx': 'खोला न्देबेले', 'bs': 'južni ndebele', 'bs-Cyrl': 'јужни ндебеле', 'bs-Latn': 'južni ndebele', 'ca': 'ndebele meridional', 'ca-ES-valencia': 'ndebele meridional', 'ccp': '𑄓𑄧𑄉𑄨𑄚𑄴 𑄆𑄚𑄴𑄓𑄬𑄝𑄬𑄣𑄬', 'ce': 'къилба ндебеле', 'chr': 'ᏧᎦᎾᏮ ᏂᏕᏇᎴ', 'ckb': 'ئندێبێلێی باشوور', 'cs': 'ndebele (Jižní Afrika)', 'cy': 'Ndebele Deheuol', 'da': 'sydndebele', 'de': 'Süd-Ndebele', 'dsb': 'pódpołdnjowa ndebelšćina', 'el': 'Νότια Ντεμπέλε', 'el-polyton': 'Νότια Ντεμπέλε', 'en': 'South Ndebele', 'es': 'ndebele meridional', 'et': 'lõunandebele', 'eu': 'hegoaldeko ndebeleera', 'fa': 'انده\u200cبله\u200cای جنوبی', 'ff-Adlm': '𞤐𞤣𞤫𞤦𞤫𞤤𞤫𞥅𞤪𞤫 𞤙𞤢𞥄𞤥𞤲𞤢𞥄𞤲𞤺𞤫', 'fi': 'etelä-ndebele', 'fil': 'South Ndebele', 'fo': 'suður ndebele', 'fr': 'ndébélé du Sud', 'fy': 'Sûd-Ndbele', 'ga': 'Ndeibéilis an Deiscirt', 'gd': 'Ndebele Dheasach', 'gl': 'ndebele meridional', 'gsw': 'Süüd-Ndebele-Schpraach', 'gu': 'દક્ષિણ દેબેલ', 'ha': 'Ndebele na Kudu', 'he': 'נדבלה דרומית', 'hi': 'दक्षिण देबेल', 'hi-Latn': 'South Ndebele', 'hr': 'južni ndebele', 'hsb': 'južna ndebelšćina', 'ht': 'ndébélé du Sud', 'hu': 'déli ndebele', 'hy': 'հարավային նդեբելե', 'ia': 'ndebele del sud', 'id': 'Ndebele Selatan', 'ig': 'South Ndebele', 'is': 'suðurndebele', 'it': 'ndebele del sud', 'ja': '南ンデベレ語', 'jv': 'Ndebele Kidul', 'ka': 'სამხრეთ ნდებელური', 'kgp': 'nemere sur', 'kk': 'оңтүстік ндебеле тілі', 'kk-Cyrl': 'оңтүстік ндебеле тілі', 'km': 'នេប៊េលខាងត្បូង', 'kn': 'ದಕ್ಷಿಣ ದೆಬೆಲೆ', 'ko': '남부 은데벨레어', 'kok': 'दक्षिण डेबेले', 'kok-Deva': 'दक्षिण डेबेले', 'ks': 'جنوب ڈیبیل', 'ks-Arab': 'جنوب ڈیبیل', 'ku': 'ndebelîya başûrî', 'ky': 'түштүк ндебелече', 'lb': 'Süd-Ndebele-Sprooch', 'lld': 'ndebele dl süd', 'lo': 'ນີບີລີໃຕ້', 'lt': 'pietų ndebele', 'lv': 'dienvidndebelu', 'mai': 'दक्षिणी एनडेबेले', 'mi': 'Enetepēra ki te Tonga', 'mk': 'јужен ндебеле', 'ml': 'ദക്ഷിണ നെഡിബിൾ', 'mn': 'өмнөд ндебеле', 'mr': 'दक्षिणात्य देबेली', 'ms': 'Ndebele Selatan', 'mt': 'Ndebele tan-Nofsinhar', 'my': 'တောင် အွန်န်ဘီလီ', 'nb': 'sør-ndebele', 'ne': 'दक्षिण न्देबेले', 'nl': 'Zuid-Ndbele', 'nn': 'sør-ndebele', 'no': 'sør-ndebele', 'or': 'ଦକ୍ଷିଣ ନେଡବେଲେ', 'pa': 'ਸਾਊਥ ਨਡੇਬੇਲੇ', 'pa-Guru': 'ਸਾਊਥ ਨਡੇਬੇਲੇ', 'pcm': 'Sáút Ndẹbẹlẹ Lángwej', 'pl': 'ndebele południowy', 'ps': 'سويلي نديبيل', 'pt': 'ndebele do sul', 'qu': 'Ndebele del Sur Simi', 'rm': 'ndebele dal sid', 'ro': 'ndebele de sud', 'ru': 'южный ндебеле', 'sc': 'ndebele de su sud', 'sd': 'ڏکڻ دبيلي', 'sd-Arab': 'ڏکڻ دبيلي', 'si': 'සෞත් ඩ්බේල්', 'sk': 'ndebelčina (južná)', 'sl': 'južna ndebelščina', 'smn': 'maadâ-nbedele', 'so': 'South Ndebele', 'sq': 'ndebelishte jugore', 'sr': 'јужни ндебеле', 'sr-Cyrl': 'јужни ндебеле', 'sr-Latn': 'južni ndebele', 'sv': 'sydndebele', 'sw': 'Kindebele', 'ta': 'தெற்கு தெபெலே', 'te': 'దక్షిణ దెబెలె', 'th': 'เอ็นเดเบเลใต้', 'ti': 'ደቡብ ኤንደበለ', 'tk': 'günorta ndebele dili', 'to': 'lea fakanetepele-tonga', 'tr': 'Güney Ndebele', 'ug': 'جەنۇبى ندەبەلەچە', 'uk': 'ндебелє південна', 'ur': 'جنوبی نڈیبیلی', 'uz': 'janubiy ndebel', 'uz-Cyrl': 'жанубий ндебелча', 'uz-Latn': 'janubiy ndebel', 'vec': 'ndebele de’l sud', 'vi': 'Tiếng Ndebele Miền Nam', 'xnr': 'दखण देबेल', 'yo': 'Èdè Gusu Ndebele', 'yrl': 'ĩdebere suwara', 'yue': '南地畢列文', 'yue-Hans': '南地毕列文', 'yue-Hant': '南地畢列文', 'zh': '南恩德贝勒语', 'zh-Hans': '南恩德贝勒语', 'zh-Hant': '南地畢列文', 'zu': 'isi-South Ndebele'},
    'nra': {'en': 'Ngom'},
    'nrb': {'en': 'Nara'},
    'nrc': {'en': 'Noric'},
    'nre': {'en': 'Southern Rengma Naga'},
    'nrf': {'en': 'Jèrriais'},
    'nrg': {'en': 'Narango'},
    'nri': {'en': 'Chokri Naga'},
    'nrk': {'en': 'Ngarla'},
    'nrl': {'en': 'Ngarluma'},
    'nrm': {'en': 'Narom'},
    'nrn': {'en': 'Norn'},
    'nrp': {'en': 'North Picene'},
    'nrr': {'en': 'Norra'},
    'nrt': {'en': 'Northern Kalapuya'},
    'nru': {'en': 'Narua'},
    'nrx': {'en': 'Ngurmbur'},
    'nrz': {'en': 'Lala'},
    'nsa': {'en': 'Sangtam Naga'},
    'nsb': {'en': 'Lower Nossob'},
    'nsc': {'en': 'Nshi'},
    'nsd': {'en': 'Southern Nisu'},
    'nse': {'en': 'Nsenga'},
    'nsf': {'en': 'Northwestern Nisu'},
    'nsg': {'en': 'Ngasa'},
    'nsh': {'en': 'Ngoshie'},
    'nsi': {'en': 'Nigerian Sign Language'},
    'nsk': {'en': 'Naskapi'},
    'nsl': {'en': 'Norwegian Sign Language'},
    'nsm': {'en': 'Sumi Naga'},
    'nsn': {'en': 'Nehan'},
    'nso': {'af': 'Noord-Sotho', 'am': 'ሰሜናዊ ሶቶ', 'ar': 'السوتو الشمالية', 'as': 'উদীচ্য ছোথো', 'ast': 'sotho del norte', 'az': 'şimal soto', 'az-Cyrl': 'шимали сото', 'az-Latn': 'şimal soto', 'bal-Latn': 'Shemáli Sotó', 'be': 'паўночная сота', 'be-tarask': 'паўночная сота', 'bg': 'северен сото', 'bgn': 'بُرزسرین سوتویی', 'bn': 'উত্তরাঞ্চলীয় সোথো', 'br': 'sotho an Norzh', 'brx': 'साहायारि सथ’', 'bs': 'sjeverni soto', 'bs-Cyrl': 'сјеверни сото', 'bs-Latn': 'sjeverni soto', 'ca': 'sotho septentrional', 'ca-ES-valencia': 'sotho septentrional', 'ccp': '𑄃𑄪𑄖𑄴𑄗𑄧𑄢𑄴 𑄢𑄬𑄌𑄴𑄎𑄮𑄢𑄴 𑄥𑄮𑄗𑄮', 'ce': 'къилбаседа сото', 'chr': 'ᏧᏴᏢ ᏗᏜ ᏐᏠ', 'ckb': 'سۆتۆی باکوور', 'cs': 'sotština (severní)', 'cy': 'Sotho Gogleddol', 'da': 'nordsotho', 'de': 'Nord-Sotho', 'dsb': 'połnocna sothošćina', 'ee': 'dziehe sothogbe', 'el': 'Βόρεια Σόθο', 'el-polyton': 'Βόρεια Σόθο', 'en': 'Northern Sotho', 'es': 'sotho septentrional', 'et': 'põhjasotho', 'eu': 'pediera', 'fa': 'سوتویی شمالی', 'ff-Adlm': '𞤅𞤮𞤼𞤮𞥅𞤪𞤫 𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤫', 'fi': 'pohjoissotho', 'fil': 'Hilagang Sotho', 'fo': 'norður sotho', 'fr': 'sotho du Nord', 'fur': 'sotho setentrionâl', 'fy': 'Noard-Sotho', 'ga': 'Sútúis an Tuaiscirt', 'gd': 'Sesotho sa Leboa', 'gl': 'sesotho do norte', 'gsw': 'Nord-Sotho-Schpraach', 'gu': 'ઉત્તરી સોથો', 'ha': 'Sotho na Arewaci', 'he': 'סותו צפונית', 'hi': 'उत्तरी सोथो', 'hi-Latn': 'Northern Sotho', 'hr': 'sjeverni sotski', 'hsb': 'sewjerna sothošćina', 'ht': 'sotho du Nord', 'hu': 'északi szeszotó', 'hy': 'հյուսիսային սոթո', 'ia': 'sotho del nord', 'id': 'Sotho Utara', 'ig': 'Northern Sotho', 'is': 'norðursótó', 'it': 'sotho del nord', 'ja': '北部ソト語', 'jv': 'Sotho Sisih Lor', 'ka': 'ჩრდილოეთ სოთო', 'kgp': 'soto nãrti', 'kk': 'солтүстік сото тілі', 'kk-Cyrl': 'солтүстік сото тілі', 'km': 'សូថូខាងជើង', 'kn': 'ಉತ್ತರ ಸೋಥೋ', 'ko': '북부 소토어', 'kok': 'उत्तरीय सोथो', 'kok-Deva': 'उत्तरीय सोथो', 'ks': 'شمالی ستھو', 'ks-Arab': 'شمالی ستھو', 'ksh': 'Nood-Sohto', 'ku': 'sotoyîya bakur', 'ky': 'түндүк соточо', 'lb': 'Nord-Sotho-Sprooch', 'lo': 'ໂຊໂທເໜືອ', 'lt': 'šiaurės Soto', 'lv': 'ziemeļsotu', 'mai': 'उत्तरी सोथो', 'mi': 'Hoto ki te Raki', 'mk': 'северносотски', 'ml': 'നോർത്തേൻ സോതോ', 'mn': 'хойд сото', 'mr': 'उत्तरी सोथो', 'ms': 'Sotho Utara', 'mt': 'Soto tat-Tramuntana', 'my': 'မြောက် ဆိုသို', 'nb': 'nord-sotho', 'ne': 'उत्तरी सोथो', 'nl': 'Noord-Sotho', 'nn': 'nordsotho', 'no': 'nord-sotho', 'nso': 'Sesotho sa Leboa', 'or': 'ଉତ୍ତରୀ ସୋଥୋ', 'pa': 'ਉੱਤਰੀ ਸੋਥੋ', 'pa-Guru': 'ਉੱਤਰੀ ਸੋਥੋ', 'pcm': 'Nọ́tán Sótho Lángwej', 'pl': 'sotho północny', 'ps': 'شمالي سوتو', 'pt': 'soto setentrional', 'qu': 'Sesotho Sa Leboa Simi', 'rm': 'sotho dal nord', 'ro': 'sotho de nord', 'ru': 'северный сото', 'sc': 'sotho setentrionale', 'sd': 'اتر سوٿو', 'sd-Arab': 'اتر سوٿو', 'si': 'නොදර්න් සොතො', 'sk': 'sothčina (severná)', 'sl': 'severna sotščina', 'smn': 'tavesotho', 'so': 'Northern Sotho', 'sq': 'sotoishte veriore', 'sr': 'северни сото', 'sr-Cyrl': 'северни сото', 'sr-Latn': 'severni soto', 'sv': 'nordsotho', 'sw': 'Kisotho cha Kaskazini', 'ta': 'வடக்கு சோதோ', 'te': 'ఉత్తర సోతో', 'th': 'โซโทเหนือ', 'ti': 'ሰሜናዊ ሶቶ', 'tk': 'demirgazyk soto dili', 'to': 'lea fakasoto-tokelau', 'tr': 'Kuzey Sotho dili', 'ug': 'شىمالىي سوتوچە', 'uk': 'північна сото', 'ur': 'شمالی سوتھو', 'uz': 'shimoliy soto', 'uz-Latn': 'shimoliy soto', 'vec': 'soto setentrionale', 'vi': 'Tiếng Sotho Miền Bắc', 'wae': 'Nordsotho', 'xnr': 'उत्तरी सोथो', 'yo': 'Èdè Ariwa Soto', 'yrl': 'sutu setẽtiriunau', 'yue': '北索托文', 'yue-Hans': '北索托文', 'yue-Hant': '北索托文', 'zh': '北索托语', 'zh-Hans': '北索托语', 'zh-Hant': '北索托文', 'zu': 'isi-Northern Sotho'},
    'nsp': {'en': 'Nepalese Sign Language'},
    'nsq': {'en': 'Northern Sierra Miwok'},
    'nsr': {'en': 'Maritime Sign Language'},
    'nss': {'en': 'Nali'},
    'nst': {'en': 'Tase Naga'},
    'nsu': {'en': 'Sierra Negra Nahuatl'},
    'nsv': {'en': 'Southwestern Nisu'},
    'nsw': {'en': 'Navut'},
    'nsx': {'en': 'Nsongo'},
    'nsy': {'en': 'Nasal'},
    'nsz': {'en': 'Nisenan'},
    'ntd': {'en': 'Northern Tidung'},
    'nte': {'en': 'Nathembo'},
    'ntg': {'en': 'Ngantangarra'},
    'nti': {'en': 'Natioro'},
    'ntj': {'en': 'Ngaanyatjarra'},
    'ntk': {'en': 'Ikoma-Nata-Isenye'},
    'ntm': {'en': 'Nateni'},
    'nto': {'en': 'Ntomba'},
    'ntp': {'en': 'Northern Tepehuan'},
    'ntr': {'en': 'Delo'},
    'nts': {'en': 'Natagaimas'},
    'ntu': {'en': 'Natügu'},
    'ntw': {'en': 'Nottoway'},
    'ntx': {'en': 'Tangkhul Naga (Myanmar)'},
    'nty': {'en': 'Mantsi'},
    'ntz': {'en': 'Natanzi'},
    'nua': {'en': 'Yuanga'},
    'nub': {'en': 'Nubian languages'},
    'nuc': {'en': 'Nukuini'},
    'nud': {'en': 'Ngala'},
    'nue': {'en': 'Ngundu'},
    'nuf': {'en': 'Nusu'},
    'nug': {'en': 'Nungali'},
    'nuh': {'en': 'Ndunda'},
    'nui': {'en': 'Ngumbi'},
    'nuj': {'en': 'Nyole'},
    'nuk': {'en': 'Nuu-chah-nulth'},
    'nul': {'en': 'Nusa Laut'},
    'num': {'en': "Niuafo'ou"},
    'nun': {'en': 'Anong'},
    'nuo': {'en': 'Nguôn'},
    'nup': {'en': 'Nupe-Nupe-Tako'},
    'nuq': {'en': 'Nukumanu'},
    'nur': {'en': 'Nukuria'},
    'nus': {'af': 'Nuer', 'am': 'ኑዌር', 'ar': 'النوير', 'as': 'নুয়েৰ', 'ast': 'nuer', 'az': 'nuer', 'az-Cyrl': 'нуер', 'az-Latn': 'nuer', 'bal-Latn': 'Nuér', 'be': 'нуэр', 'be-tarask': 'нуэр', 'bg': 'нуер', 'bgn': 'نویری', 'bn': 'নুয়ার', 'br': 'nouereg', 'brx': 'नूवेर', 'bs': 'nuer', 'bs-Cyrl': 'нуер', 'bs-Latn': 'nuer', 'ca': 'nuer', 'ca-ES-valencia': 'nuer', 'ccp': '𑄚𑄪𑄠𑄢𑄴', 'ce': 'нуэр', 'chr': 'ᏄᏪᎵ', 'ckb': 'نوێر', 'cs': 'nuerština', 'cy': 'Nŵereg', 'da': 'nuer', 'de': 'Nuer', 'dsb': 'nuer', 'el': 'Νούερ', 'el-polyton': 'Νούερ', 'en': 'Nuer', 'es': 'nuer', 'et': 'nueri', 'eu': 'nuerera', 'fa': 'نویر', 'ff-Adlm': '𞤐𞤵𞤫𞤪𞤭𞥅𞤪𞤫', 'fi': 'nuer', 'fil': 'Nuer', 'fo': 'nuer', 'fr': 'nuer', 'fy': 'Nuer', 'ga': 'Nuairis', 'gd': 'Nuer', 'gl': 'nuer', 'gu': 'નુએર', 'ha': 'Nuer', 'he': 'נואר', 'hi': 'नुएर', 'hi-Latn': 'Nuer', 'hr': 'nuerski', 'hsb': 'nuer', 'ht': 'nuer', 'hu': 'nuer', 'hy': 'նուեր', 'ia': 'nuer', 'id': 'Nuer', 'ig': 'Nuer', 'is': 'núer', 'it': 'nuer', 'ja': 'ヌエル語', 'jv': 'Nuer', 'ka': 'ნუერი', 'kgp': 'nũver', 'kk': 'нуэр тілі', 'kk-Cyrl': 'нуэр тілі', 'km': 'នូអ័រ', 'kn': 'ನೂಯರ್', 'ko': '누에르어', 'kok': 'न्युयर', 'kok-Deva': 'न्युयर', 'ku': 'nuer', 'ky': 'нуерче', 'lb': 'Nuer', 'lo': 'ເນີ', 'lrc': 'نیوٙئر', 'lt': 'nuerų', 'lv': 'nueru', 'mai': 'नूअर', 'mi': 'Nua', 'mk': 'нуер', 'ml': 'നുവേർ', 'mn': 'нуер', 'mr': 'नुएर', 'ms': 'Nuer', 'mt': 'Nuer', 'my': 'နူအာ', 'mzn': 'نوئر', 'nb': 'nuer', 'ne': 'नुएर', 'nl': 'Nuer', 'nn': 'nuer', 'no': 'nuer', 'nqo': 'ߣߎߦߋߞߊ߲', 'nus': 'Thok Nath', 'or': 'ନୁଏର', 'pa': 'ਨੁਏਰ', 'pa-Guru': 'ਨੁਏਰ', 'pcm': 'Núa', 'pl': 'nuer', 'ps': 'نویر', 'pt': 'nuer', 'qu': 'Nuer Simi', 'rm': 'nuer', 'ro': 'nuer', 'ru': 'нуэр', 'sc': 'nuer', 'sd': 'نيور', 'sd-Arab': 'نيور', 'si': 'නොයර්', 'sk': 'nuer', 'sl': 'nuerščina', 'smn': 'nuer', 'so': 'Nuweer', 'sq': 'nuerisht', 'sr': 'нуер', 'sr-Cyrl': 'нуер', 'sr-Latn': 'nuer', 'sv': 'nuer', 'sw': 'Kinuer', 'ta': 'நியூர்', 'te': 'న్యుర్', 'th': 'เนือร์', 'ti': 'ንዌር', 'tk': 'nuer dili', 'to': 'lea fakanueli', 'tr': 'Nuer', 'ug': 'مۇئېرچە', 'uk': 'нуер', 'ur': 'نویر', 'uz': 'nuer', 'uz-Cyrl': 'нуэрча', 'uz-Latn': 'nuer', 'vec': 'nuer', 'vi': 'Tiếng Nuer', 'xnr': 'नुएर', 'yo': 'Núẹ̀', 'yrl': 'nuiri', 'yue': '努埃爾文', 'yue-Hans': '努埃尔文', 'yue-Hant': '努埃爾文', 'zh': '努埃尔语', 'zh-Hans': '努埃尔语', 'zh-Hant': '努埃爾文', 'zu': 'isi-Nuer'},
    'nut': {'en': 'Nung (Viet Nam)'},
    'nuu': {'en': 'Ngbundu'},
    'nuv': {'en': 'Northern Nuni'},
    'nuw': {'en': 'Nguluwan'},
    'nux': {'en': 'Mehek'},
    'nuy': {'en': 'Nunggubuyu'},
    'nuz': {'en': 'Tlamacazapa Nahuatl'},
    'nv': {'af': 'Navajo', 'am': 'ናቫጆ', 'ar': 'النافاجو', 'as': 'নাভাজো', 'ast': 'navajo', 'az': 'navayo', 'az-Cyrl': 'навајо', 'az-Latn': 'navayo', 'bal-Latn': 'Nawájó', 'be': 'наваха', 'be-tarask': 'наваха', 'bg': 'навахо', 'bgn': 'ناواهویی', 'bn': 'নাভাজো', 'br': 'navacʼho', 'brx': 'नावाहो', 'bs': 'navaho', 'bs-Cyrl': 'навахо', 'bs-Latn': 'navaho', 'ca': 'navaho', 'ca-ES-valencia': 'navaho', 'ccp': '𑄚𑄞𑄎𑄮', 'ce': 'навахо', 'chr': 'ᎾᏩᎰ', 'ckb': 'ناڤاجۆ', 'cs': 'navažština', 'cy': 'Nafaho', 'da': 'navajo', 'de': 'Navajo', 'dsb': 'navaho', 'el': 'Νάβαχο', 'el-polyton': 'Νάβαχο', 'en': 'Navajo', 'en-Dsrt': '𐐤𐐪𐑂𐐲𐐸𐐬', 'es': 'navajo', 'et': 'navaho', 'eu': 'navajoera', 'fa': 'ناواهویی', 'ff-Adlm': '𞤐𞤢𞤬𞤱𞤢𞤸𞤮𞥅𞤪𞤫', 'fi': 'navajo', 'fil': 'Navajo', 'fo': 'navajo', 'fr': 'navajo', 'fur': 'navajo', 'fy': 'Navajo', 'ga': 'Navachóis', 'gd': 'Navajo', 'gl': 'navajo', 'gsw': 'Navajo-Schpraach', 'gu': 'નાવાજો', 'ha': 'Navajo', 'he': 'נאוואחו', 'hi': 'नवाहो', 'hi-Latn': 'Navajo', 'hr': 'navajo', 'hsb': 'navaho', 'ht': 'navajo', 'hu': 'navahó', 'hy': 'նավախո', 'ia': 'navajo', 'id': 'Navajo', 'ig': 'Navajo', 'is': 'navahó', 'it': 'navajo', 'ja': 'ナバホ語', 'jv': 'Navajo', 'ka': 'ნავახო', 'kgp': 'nỹvaho', 'kk': 'навахо тілі', 'kk-Cyrl': 'навахо тілі', 'km': 'ណាវ៉ាចូ', 'kn': 'ನವಾಜೊ', 'ko': '나바호어', 'kok': 'नावाजो', 'kok-Deva': 'नावाजो', 'ks': 'نَواجو', 'ks-Arab': 'نَواجو', 'ksh': 'Navvacho', 'ku': 'navajoyî', 'ky': 'наваджочо', 'lb': 'Navajo', 'lkt': 'Šináglegleǧa Iyápi', 'lld': 'navajan', 'lo': 'ນາວາໂຈ', 'lt': 'navajų', 'lv': 'navahu', 'mai': 'नवाजो', 'mi': 'Nawahō', 'mk': 'навахо', 'ml': 'നവാജോ', 'mn': 'навахо', 'mr': 'नावाजो', 'ms': 'Navajo', 'mt': 'Navajo', 'my': 'နာဗာဟို', 'nb': 'navajo', 'ne': 'नाभाजो', 'nl': 'Navajo', 'nn': 'navajo', 'no': 'navajo', 'nqo': 'ߣߝ߭ߊߖߏߞߊ߲', 'nv': 'Diné Bizaad', 'or': 'ନାଭାଜୋ', 'pa': 'ਨਵਾਜੋ', 'pa-Guru': 'ਨਵਾਜੋ', 'pcm': 'Navájo Lángwej', 'pl': 'nawaho', 'ps': 'نواجو', 'pt': 'navajo', 'qu': 'Navajo Simi', 'rm': 'navajo', 'ro': 'navajo', 'ru': 'навахо', 'sc': 'navajo', 'sd': 'نواجو', 'sd-Arab': 'نواجو', 'si': 'නවාජො', 'sk': 'navaho', 'sl': 'navajščina', 'smn': 'navajokielâ', 'so': 'Navajo', 'sq': 'navahoisht', 'sr': 'навахо', 'sr-Cyrl': 'навахо', 'sr-Latn': 'navaho', 'sv': 'navaho', 'sw': 'Kinavajo', 'ta': 'நவாஜோ', 'te': 'నవాజొ', 'th': 'นาวาโฮ', 'ti': 'ናቫሆ', 'tk': 'nawaho dili', 'to': 'lea fakanavaho', 'tr': 'Navaho dili', 'ug': 'ناۋاخوچە', 'uk': 'навахо', 'ur': 'نواجو', 'uz': 'navaxo', 'uz-Latn': 'navaxo', 'vec': 'navajo', 'vi': 'Tiếng Navajo', 'xnr': 'नावाजो', 'yo': 'Èdè Nafajo', 'yrl': 'nawayu', 'yue': '納瓦霍文', 'yue-Hans': '纳瓦霍文', 'yue-Hant': '納瓦霍文', 'zh': '纳瓦霍语', 'zh-Hans': '纳瓦霍语', 'zh-Hant': '納瓦荷文', 'zu': 'isi-Navajo'},
    'nvh': {'en': 'Nasarian'},
    'nvm': {'en': 'Namiae'},
    'nvo': {'en': 'Nyokon'},
    'nwa': {'en': 'Nawathinehena'},
    'nwb': {'en': 'Nyabwa'},
    'nwc': {'am': 'ክላሲክ ኔዋሪ', 'ar': 'النوارية التقليدية', 'ast': 'newari clásicu', 'bg': 'класически невари', 'bn': 'প্রাচীন নেওয়ারী', 'br': 'newari klasel', 'brx': 'पुरानी नेवारी', 'bs': 'klasični nevari', 'bs-Cyrl': 'класични невари', 'bs-Latn': 'klasični nevari', 'ca': 'newari clàssic', 'ca-ES-valencia': 'newari clàssic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄴 𑄚𑄬𑄃𑄮𑄠𑄢𑄩', 'cs': 'newarština (klasická)', 'cy': 'Hen Newari', 'da': 'klassisk newarisk', 'de': 'Alt-Newari', 'el': 'Κλασικά Νεουάρι', 'el-polyton': 'Κλασικά Νεουάρι', 'en': 'Classical Newari', 'es': 'newari clásico', 'et': 'vananevari', 'fa': 'نواریایی کلاسیک', 'fi': 'klassinen newari', 'fr': 'newarî classique', 'fy': 'Klassiek Newari', 'gd': 'Newari Chlasaigeach', 'gsw': 'Alt-Newari', 'gu': 'પરંપરાગત નેવારી', 'he': 'נווארית קלאסית', 'hi': 'पारम्परिक नेवारी', 'hi-Latn': 'Classical Newari', 'hr': 'klasični newari', 'ht': 'newarî classique', 'hu': 'klasszikus newari', 'id': 'Newari Klasik', 'is': 'klassísk nevaríska', 'it': 'newari classico', 'ja': '古典ネワール語', 'ka': 'კლასიკური ნევარული', 'kgp': 'nẽvari há tỹvĩ', 'kn': 'ಶಾಸ್ತ್ರೀಯ ನೇವಾರಿ', 'ko': '고전 네와르어', 'ks': 'کلاسِکَل نیواری', 'ks-Arab': 'کلاسِکَل نیواری', 'lb': 'Al-Newari', 'lo': 'ເນວາດັ້ງເດີມ', 'lt': 'klasikinė nevari', 'lv': 'klasiskā nevaru', 'mk': 'класичен неварски', 'ml': 'ക്ലാസിക്കൽ നേവാരി', 'mr': 'अभिजात नेवारी', 'mt': 'Newari Klassiku', 'nb': 'klassisk newari', 'ne': 'परम्परागत नेवारी', 'nl': 'Klassiek Nepalbhasa', 'nn': 'klassisk newarisk', 'no': 'klassisk newari', 'or': 'ପାରମ୍ପରିକ ନେୱାରୀ', 'pl': 'newarski klasyczny', 'pt': 'newari clássico', 'rm': 'newari classic', 'ro': 'newari clasică', 'ru': 'классический невари', 'sk': 'klasická nevárčina', 'sl': 'klasična nevarščina', 'sr': 'класични неварски', 'sr-Cyrl': 'класични неварски', 'sr-Latn': 'klasični nevarski', 'sv': 'klassisk newariska', 'sw': 'Kinewari cha kale', 'ta': 'பாரம்பரிய நேவாரி', 'te': 'సాంప్రదాయ న్యూయారీ', 'th': 'เนวาร์ดั้งเดิม', 'to': 'lea fakaneuali-motuʻa', 'tr': 'Klasik Nevari', 'ug': 'نېۋارچە', 'uk': 'неварі класична', 'vi': 'Tiếng Newari cổ', 'yrl': 'newari katuwa', 'yue': '古尼瓦爾文', 'yue-Hans': '古尼瓦尔文', 'yue-Hant': '古尼瓦爾文', 'zh': '古典尼瓦尔语', 'zh-Hans': '古典尼瓦尔语', 'zh-Hant': '古尼瓦爾文'},
    'nwe': {'en': 'Ngwe'},
    'nwg': {'en': 'Ngayawung'},
    'nwi': {'en': 'Southwest Tanna'},
    'nwm': {'en': 'Nyamusa-Molo'},
    'nwo': {'en': 'Nauo'},
    'nwr': {'en': 'Nawaru'},
    'nwx': {'en': 'Middle Newar'},
    'nwy': {'en': 'Nottoway-Meherrin'},
    'nxa': {'en': 'Nauete'},
    'nxd': {'en': 'Ngando (Democratic Republic of Congo)'},
    'nxe': {'en': 'Nage'},
    'nxg': {'en': "Ngad'a"},
    'nxi': {'en': 'Nindi'},
    'nxk': {'en': 'Koki Naga'},
    'nxl': {'en': 'South Nuaulu'},
    'nxm': {'en': 'Numidian'},
    'nxn': {'en': 'Ngawun'},
    'nxo': {'en': 'Ndambomo'},
    'nxq': {'en': 'Naxi'},
    'nxr': {'en': 'Ninggerum'},
    'nxu': {'en': 'Narau'},
    'nxx': {'en': 'Nafri'},
    'ny': {'af': 'Nyanja', 'am': 'ንያንጃ', 'ar': 'النيانجا', 'as': 'ন্যাঞ্জা', 'ast': 'nyanja', 'az': 'nyanca', 'az-Cyrl': 'нјанҹа', 'az-Latn': 'nyanca', 'bal-Latn': 'Nyanjá', 'be': 'ньянджа', 'be-tarask': 'ньянджа', 'bg': 'нянджа', 'bgn': 'نیانجی', 'bn': 'নায়াঞ্জা', 'br': 'nyanja', 'brx': 'न्यान्जा', 'bs': 'njanja', 'bs-Cyrl': 'њања', 'bs-Latn': 'njanja', 'ca': 'nyanja', 'ca-ES-valencia': 'nyanja', 'ccp': '𑄚𑄠𑄚𑄴𑄎', 'ce': 'ньянджа', 'chr': 'ᏂᏯᏂᏣ', 'ckb': 'نیانجا', 'cs': 'ňandžština', 'cy': 'Nianja', 'da': 'nyanja', 'de': 'Nyanja', 'dsb': 'nyanja', 'ee': 'nyanjagbe', 'el': 'Νιάντζα', 'el-polyton': 'Νιάντζα', 'en': 'Nyanja', 'es': 'nyanja', 'et': 'njandža', 'eu': 'chewera', 'fa': 'نیانجایی', 'ff-Adlm': '𞤙𞤢𞤲𞤶𞤢𞥄𞤪𞤫', 'fi': 'njandža', 'fil': 'Nyanja', 'fo': 'nyanja', 'fr': 'chewa', 'fy': 'Nyanja', 'ga': 'Siséivis', 'gd': 'Nyanja', 'gl': 'chewa', 'gsw': 'Chewa-Schpraach', 'gu': 'ન્યાન્જા', 'ha': 'Nyanja', 'he': 'ניאנג׳ה', 'hi': 'न्यानजा', 'hi-Latn': 'Nyanja', 'hr': 'njandža', 'hsb': 'nyanja', 'ht': 'chewa', 'hu': 'nyandzsa', 'hy': 'նյանջա', 'ia': 'nyanja', 'id': 'Nyanja', 'ig': 'Nyanja', 'is': 'nýanja', 'it': 'nyanja', 'ja': 'ニャンジャ語', 'jv': 'Nyanja', 'ka': 'ნიანჯა', 'kgp': 'nĩjỹnja', 'kk': 'ньянджа тілі', 'kk-Cyrl': 'ньянджа тілі', 'km': 'ណានចា', 'kn': 'ನ್ಯಾಂಜಾ', 'ko': '냔자어', 'kok': 'नांन्जा', 'kok-Deva': 'नांन्जा', 'ks': 'نِیَنجا', 'ks-Arab': 'نِیَنجا', 'ksh': 'Schi-Schewa', 'ku': 'çîçewayî', 'ky': 'ньянджача', 'lb': 'Nyanja-Sprooch', 'lld': 'nyanja', 'lo': 'ນານຈາ', 'lt': 'nianjų', 'lv': 'čičeva', 'mai': 'नींजा', 'mi': 'Niānia', 'mk': 'њанџа', 'ml': 'ന്യൻജ', 'mn': 'нянжа', 'mr': 'न्यान्जा', 'ms': 'Nyanja', 'mt': 'Nyanja', 'my': 'နရန်ဂျာ', 'nb': 'nyanja', 'ne': 'न्यान्जा', 'nl': 'Nyanja', 'nn': 'nyanja', 'no': 'nyanja', 'nqo': 'ߛߋߥߞߊ߲', 'ny': 'Nyanja', 'or': 'ନିୟାଞ୍ଜ', 'pa': 'ਨਯਾਂਜਾ', 'pa-Guru': 'ਨਯਾਂਜਾ', 'pcm': 'Nyánja', 'pl': 'njandża', 'ps': 'نیانجا', 'pt': 'nianja', 'qu': 'Nyanja Simi', 'rm': 'nyanja', 'ro': 'nyanja', 'ru': 'ньянджа', 'sc': 'nyanja', 'sd': 'نيانجا', 'sd-Arab': 'نيانجا', 'si': 'න්යන්ජා', 'sk': 'ňandža', 'sl': 'njanščina', 'smn': 'njanža', 'so': 'Inyaanja', 'sq': 'nianjisht', 'sr': 'њанџа', 'sr-Cyrl': 'њанџа', 'sr-Latn': 'njandža', 'sv': 'nyanja', 'sw': 'Kinyanja', 'ta': 'நயன்ஜா', 'te': 'న్యాన్జా', 'tg': 'нянҷа', 'th': 'เนียนจา', 'ti': 'ንያንጃ', 'tk': 'nýanja dili', 'to': 'lea fakanianisa', 'tr': 'Nyanja', 'tt': 'ньянҗа', 'ug': 'نىيانجاچە', 'uk': 'ньянджа', 'ur': 'نیانجا', 'uz': 'cheva', 'uz-Cyrl': 'чева', 'uz-Latn': 'cheva', 'vec': 'chewa', 'vi': 'Tiếng Nyanja', 'wae': 'Nyanja', 'wo': 'Sewa', 'xnr': 'न्यानजा', 'yo': 'Ńyájà', 'yrl': 'niãya', 'yue': '尼揚賈文', 'yue-Hans': '尼扬贾文', 'yue-Hant': '尼揚賈文', 'zh': '齐切瓦语', 'zh-Hans': '齐切瓦语', 'zh-Hant': '尼揚賈文', 'zu': 'isi-Nyanja'},
    'nyb': {'en': 'Nyangbo'},
    'nyc': {'en': 'Nyanga-li'},
    'nyd': {'en': 'Nyore'},
    'nye': {'en': 'Nyengo'},
    'nyf': {'en': 'Giryama'},
    'nyg': {'en': 'Nyindu'},
    'nyh': {'en': 'Nyikina'},
    'nyi': {'en': 'Ama (Sudan)'},
    'nyj': {'en': 'Nyanga'},
    'nyk': {'en': 'Nyaneka'},
    'nyl': {'en': 'Nyeu'},
    'nym': {'ar': 'النيامويزي', 'ast': 'nyamwezi', 'az': 'nyamvezi', 'az-Latn': 'nyamvezi', 'bg': 'ниамвези', 'bn': 'ন্যায়ামওয়েজি', 'br': 'nyamwezi', 'brx': 'न्यामवेज़ी', 'bs': 'njamvezi', 'bs-Cyrl': 'њамвези', 'bs-Latn': 'njamvezi', 'ca': 'nyamwesi', 'ca-ES-valencia': 'nyamwesi', 'ccp': '𑄚𑄳𑄠𑄠𑄟𑄴𑄃𑄮𑄠𑄬𑄎𑄨', 'cs': 'ňamwežština', 'cy': 'Niamwezi', 'da': 'nyamwezi', 'de': 'Nyamwezi', 'el': 'Νιαμγουέζι', 'el-polyton': 'Νιαμγουέζι', 'en': 'Nyamwezi', 'es': 'nyamwezi', 'et': 'njamvesi', 'fa': 'نیام\u200cوزیایی', 'fi': 'nyamwezi', 'fr': 'nyamwezi', 'fy': 'Nyamwezi', 'gd': 'Nyamwezi', 'gsw': 'Nyamwezi-Schpraach', 'gu': 'ન્યામવેઝી', 'he': 'ניאמווזי', 'hi': 'न्यामवेज़ी', 'hi-Latn': 'Nyamwezi', 'hr': 'nyamwezi', 'ht': 'nyamwezi', 'hu': 'nyamvézi', 'id': 'Nyamwezi', 'is': 'njamvesí', 'it': 'nyamwezi', 'ja': 'ニャムウェジ語', 'ka': 'ნიამვეზი', 'kgp': 'nyjỹm-vesi', 'kn': 'ನ್ಯಾಮ್\u200cವೆಂಜಿ', 'ko': '니암웨지어', 'ks': 'نِیَمویزی', 'ks-Arab': 'نِیَمویزی', 'lb': 'Nyamwezi-Sprooch', 'lo': 'ນາມວີຊິ', 'lt': 'niamvezi', 'lv': 'ņamvezu', 'mk': 'њамвези', 'ml': 'ന്യാംവേസി', 'mr': 'न्यामवेझी', 'mt': 'Njamweżi', 'nb': 'nyamwezi', 'ne': 'न्यामवेजी', 'nl': 'Nyamwezi', 'nn': 'nyamwezi', 'no': 'nyamwezi', 'or': 'ନ୍ୟାମୱେଜୀ', 'pl': 'niamwezi', 'pt': 'nyamwezi', 'rm': 'nyamwezi', 'ro': 'nyamwezi', 'ru': 'ньямвези', 'sk': 'ňamwezi', 'sl': 'njamveščina', 'sr': 'њамвези', 'sr-Cyrl': 'њамвези', 'sr-Latn': 'njamvezi', 'sv': 'nyamwezi', 'sw': 'Kinyamwezi', 'ta': 'நியாம்வேஜி', 'te': 'న్యంవేజి', 'th': 'เนียมเวซี', 'to': 'lea fakaniamiuesi', 'tr': 'Nyamvezi', 'ug': 'نيامۋېزىچە', 'uk': 'ньямвезі', 'vi': 'Tiếng Nyamwezi', 'yrl': 'niãmuwesi', 'yue': '尼揚韋齊文', 'yue-Hans': '尼扬韦齐文', 'yue-Hant': '尼揚韋齊文', 'zh': '尼扬韦齐语', 'zh-Hans': '尼扬韦齐语', 'zh-Hant': '尼揚韋齊文'},
    'nyn': {'af': 'Nyankole', 'am': 'ኒያንኮል', 'ar': 'النيانكول', 'as': 'ন্যানকোল', 'ast': 'nyankole', 'az': 'nyankol', 'az-Cyrl': 'нјанкол', 'az-Latn': 'nyankol', 'bal-Latn': 'Nyankóle', 'be': 'ньянколе', 'be-tarask': 'ньянколе', 'bg': 'нянколе', 'bgn': 'نیانکوله\u200cای', 'bn': 'ন্যায়াঙ্কোলে', 'br': 'nyankole', 'brx': 'न्यानकोले', 'bs': 'njankole', 'bs-Cyrl': 'њанколе', 'bs-Latn': 'njankole', 'ca': 'nyankole', 'ca-ES-valencia': 'nyankole', 'ccp': '𑄚𑄳𑄠𑄠𑄋𑄴𑄇𑄮𑄣𑄬', 'ce': 'ньянколе', 'chr': 'ᏂᏯᎾᎪᎴ', 'ckb': 'نیانکۆلێ', 'cs': 'ňankolština', 'cy': 'Niancole', 'da': 'nyankole', 'de': 'Nyankole', 'dsb': 'nyankole', 'el': 'Νιανκόλε', 'el-polyton': 'Νιανκόλε', 'en': 'Nyankole', 'es': 'nyankole', 'et': 'njankole', 'eu': 'nkoreera', 'fa': 'نیانکوله\u200cای', 'ff-Adlm': '𞤙𞤢𞤲𞤳𞤮𞤤𞤫𞥅𞤪𞤫', 'fi': 'nyankole', 'fil': 'Nyankole', 'fo': 'nyankole', 'fr': 'nyankolé', 'fy': 'Nyankole', 'ga': 'Niancóilis', 'gd': 'Nyankole', 'gl': 'nyankole', 'gsw': 'Nyankole', 'gu': 'ન્યાનકોલ', 'ha': 'Nyankole', 'he': 'ניאנקולה', 'hi': 'न्यानकोल', 'hi-Latn': 'Nyankole', 'hr': 'nyankole', 'hsb': 'nyankole', 'ht': 'nyankolé', 'hu': 'nyankole', 'hy': 'նյանկոլե', 'ia': 'nyankole', 'id': 'Nyankole', 'ig': 'Nyankole', 'is': 'nyankole', 'it': 'nyankole', 'ja': 'ニャンコレ語', 'jv': 'Nyankole', 'ka': 'ნიანკოლე', 'kgp': 'nyjỹmkore', 'kk': 'нианколе тілі', 'kk-Cyrl': 'нианколе тілі', 'km': 'ណានកូលេ', 'kn': 'ನ್ಯಾನ್\u200cಕೋಲೆ', 'ko': '니안콜어', 'kok': 'नानकोले', 'kok-Deva': 'नानकोले', 'ks': 'نِیَنکول', 'ks-Arab': 'نِیَنکول', 'ku': 'nyankole', 'ky': 'ныйанколчо', 'lb': 'Nyankole', 'lo': 'ນານຄອນ', 'lrc': 'نیان کوٙلئ', 'lt': 'niankolų', 'lv': 'ņankolu', 'mai': 'नीयनकोले', 'mi': 'Niānakore', 'mk': 'њанколе', 'ml': 'ന്യാൻകോൾ', 'mn': 'нянколе', 'mr': 'न्यानकोल', 'ms': 'Nyankole', 'mt': 'Nyankole', 'my': 'နရန်ကိုလီ', 'mzn': 'نیانکوله', 'nb': 'nyankole', 'ne': 'न्यान्कोल', 'nl': 'Nyankole', 'nn': 'nyankole', 'no': 'nyankole', 'nqo': 'ߢߊ߲ߞߏߟߋߞߊ߲', 'nyn': 'Runyankore', 'or': 'ନ୍ୟାନକୋଲ୍', 'pa': 'ਨਿਆਂਕੋਲੇ', 'pa-Guru': 'ਨਿਆਂਕੋਲੇ', 'pcm': 'Nyankólẹ', 'pl': 'nyankole', 'ps': 'نینکول', 'pt': 'nyankole', 'qu': 'Nyankole Simi', 'rm': 'nyankole', 'ro': 'nyankole', 'ru': 'ньянколе', 'sc': 'nyankole', 'sd': 'نايانڪول', 'sd-Arab': 'نايانڪول', 'si': 'නයන්කෝලෙ', 'sk': 'ňankole', 'sl': 'njankole', 'smn': 'nyankolekielâ', 'so': 'Inyankoole', 'sq': 'niankolisht', 'sr': 'њанколе', 'sr-Cyrl': 'њанколе', 'sr-Latn': 'njankole', 'sv': 'nyankole', 'sw': 'Kinyankole', 'ta': 'நியான்கோலே', 'te': 'న్యాన్కోలె', 'th': 'เนียนโกเล', 'ti': 'ንያንኮል', 'tk': 'nýankole dili', 'to': 'lea fakanianikole', 'tr': 'Nyankole', 'ug': 'نىيانكولېچە', 'uk': 'ньянколе', 'ur': 'نینکول', 'uz': 'nyankole', 'uz-Cyrl': 'нянколе', 'uz-Latn': 'nyankole', 'vec': 'nyankole', 'vi': 'Tiếng Nyankole', 'xnr': 'न्यानकोल', 'yo': 'Ńyákọ́lè', 'yrl': 'niãkuri', 'yue': '尼揚科萊文', 'yue-Hans': '尼扬科莱文', 'yue-Hant': '尼揚科萊文', 'zh': '尼昂科勒语', 'zh-Hans': '尼昂科勒语', 'zh-Hant': '尼揚科萊文', 'zu': 'isi-Nyankole'},
    'nyo': {'ar': 'النيورو', 'ast': 'nyoro', 'az': 'nyoro', 'az-Latn': 'nyoro', 'bg': 'нуоро', 'bn': 'ন্যোরো', 'br': 'nyoro', 'brx': 'न्यौरो', 'bs': 'njoro', 'bs-Cyrl': 'њоро', 'bs-Latn': 'njoro', 'ca': 'nyoro', 'ca-ES-valencia': 'nyoro', 'ccp': '𑄚𑄧𑄱𑄢𑄮', 'cs': 'ňorština', 'cy': 'Nioro', 'da': 'nyoro-sprog', 'de': 'Nyoro', 'el': 'Νιόρο', 'el-polyton': 'Νιόρο', 'en': 'Nyoro', 'es': 'nyoro', 'et': 'njoro', 'fa': 'نیورویی', 'fi': 'nyoro', 'fr': 'nyoro', 'fy': 'Nyoro', 'gd': 'Nyoro', 'gsw': 'Nyoro', 'gu': 'ન્યોરો', 'he': 'ניורו', 'hi': 'न्योरो', 'hi-Latn': 'Nyoro', 'hr': 'nyoro', 'ht': 'nyoro', 'hu': 'nyoró', 'id': 'Nyoro', 'is': 'njóró', 'it': 'nyoro', 'ja': 'ニョロ語', 'ka': 'ნიორო', 'kgp': 'nyjor', 'kn': 'ನ್ಯೋರೋ', 'ko': '뉴로어', 'ks': 'نِیورو', 'ks-Arab': 'نِیورو', 'lb': 'Nyoro', 'lo': 'ໂນໂຣ', 'lt': 'niorų', 'lv': 'ņoru', 'mk': 'њоро', 'ml': 'ന്യോറോ', 'mr': 'न्योरो', 'mt': 'Nyoro', 'nb': 'nyoro', 'ne': 'न्योरो', 'nl': 'Nyoro', 'nn': 'nyoro', 'no': 'nyoro', 'or': 'ନ୍ୟାରୋ', 'pl': 'nyoro', 'pt': 'nyoro', 'rm': 'nyoro', 'ro': 'nyoro', 'ru': 'ньоро', 'sk': 'ňoro', 'sl': 'njoro', 'sr': 'њоро', 'sr-Cyrl': 'њоро', 'sr-Latn': 'njoro', 'sv': 'nyoro', 'sw': 'Kinyoro', 'ta': 'நியோரோ', 'te': 'నేయోరో', 'th': 'นิโอโร', 'to': 'lea fakaniolo', 'tr': 'Nyoro', 'ug': 'نىئوروچە', 'uk': 'ньоро', 'vi': 'Tiếng Nyoro', 'yrl': 'niyuru', 'yue': '尼奧囉文', 'yue-Hans': '尼奥啰文', 'yue-Hant': '尼奧囉文', 'zh': '尼奥罗语', 'zh-Hans': '尼奥罗语', 'zh-Hant': '尼奧囉文'},
    'nyp': {'en': "Nyang'i"},
    'nyq': {'en': 'Nayini'},
    'nyr': {'en': 'Nyiha (Malawi)'},
    'nys': {'en': 'Nyungar'},
    'nyt': {'en': 'Nyawaygi'},
    'nyu': {'en': 'Nyungwe'},
    'nyv': {'en': 'Nyulnyul'},
    'nyw': {'en': 'Nyaw'},
    'nyx': {'en': 'Nganyaywana'},
    'nyy': {'en': 'Nyakyusa-Ngonde'},
    'nza': {'en': 'Tigon Mbembe'},
    'nzb': {'en': 'Njebi'},
    'nzd': {'en': 'Nzadi'},
    'nzi': {'ar': 'النزيما', 'ast': 'nzima', 'az': 'nzima', 'az-Latn': 'nzima', 'bg': 'нзима', 'bn': 'এনজিমা', 'brx': 'न्ज़ीमा', 'bs': 'nzima', 'bs-Cyrl': 'нзима', 'bs-Latn': 'nzima', 'ca': 'nzema', 'ca-ES-valencia': 'nzema', 'ccp': '𑄆𑄚𑄴𑄎𑄨𑄟', 'cs': 'nzima', 'cy': 'Nzimeg', 'da': 'nzima', 'de': 'Nzima', 'el': 'Νζίμα', 'el-polyton': 'Νζίμα', 'en': 'Nzima', 'es': 'nzima', 'et': 'nzima', 'fa': 'نزیمایی', 'fi': 'nzima', 'fr': 'nzema', 'fy': 'Nzima', 'gd': 'Nzima', 'gsw': 'Nzima', 'gu': 'ન્ઝિમા', 'he': 'נזימה', 'hi': 'न्ज़ीमा', 'hi-Latn': 'Nzima', 'hr': 'nzima', 'ht': 'nzema', 'hu': 'nzima', 'id': 'Nzima', 'is': 'nsíma', 'it': 'nzima', 'ja': 'ンゼマ語', 'ka': 'ნზიმა', 'kgp': 'nĩsimỹ', 'kn': 'ಜೀಮಾ', 'ko': '느지마어', 'ks': 'نَظیٖما', 'ks-Arab': 'نَظیٖما', 'lb': 'Nzima', 'lo': 'ນິມາ', 'lt': 'nzima', 'lv': 'nzemu', 'mk': 'нзима', 'ml': 'സിമ', 'mr': 'न्झिमा', 'mt': 'Nzima', 'nb': 'nzima', 'ne': 'नजिमा', 'nl': 'Nzima', 'nn': 'nzima', 'no': 'nzima', 'or': 'ଞ୍ଜିମା', 'pl': 'nzema', 'pt': 'nzima', 'rm': 'nzima', 'ro': 'nzima', 'ru': 'нзима', 'sk': 'nzima', 'sl': 'nzima', 'sr': 'нзима', 'sr-Cyrl': 'нзима', 'sr-Latn': 'nzima', 'sv': 'nzima', 'sw': 'Kinzema', 'ta': 'நிஜ்மா', 'te': 'జీమా', 'th': 'นซิมา', 'to': 'lea fakanesima', 'tr': 'Nzima dili', 'ug': 'نىزەماچە', 'uk': 'нзіма', 'vi': 'Tiếng Nzima', 'yrl': 'ĩsima', 'yue': '尼茲馬文', 'yue-Hans': '尼兹马文', 'yue-Hant': '尼茲馬文', 'zh': '恩济马语', 'zh-Hans': '恩济马语', 'zh-Hant': '尼茲馬文'},
    'nzk': {'en': 'Nzakara'},
    'nzm': {'en': 'Zeme Naga'},
    'nzs': {'en': 'New Zealand Sign Language'},
    'nzu': {'en': 'Teke-Nzikou'},
    'nzy': {'en': 'Nzakambay'},
    'nzz': {'en': 'Nanga Dama Dogon'},
    'oaa': {'en': 'Orok'},
    'oac': {'en': 'Oroch'},
    'oar': {'en': 'Old Aramaic (up to 700 BCE)'},
    'oav': {'en': 'Old Avar'},
    'obi': {'en': 'Obispeño'},
    'obk': {'en': 'Southern Bontok'},
    'obl': {'en': 'Oblo'},
    'obm': {'en': 'Moabite'},
    'obo': {'en': 'Obo Manobo'},
    'obr': {'en': 'Old Burmese'},
    'obt': {'en': 'Old Breton'},
    'obu': {'en': 'Obulom'},
    'oc': {'af': 'Oksitaans', 'ak': 'Osita kasa', 'am': 'ኦሲታን', 'ar': 'الأوكسيتانية', 'as': 'অ’চিটান', 'ast': 'occitanu', 'az': 'oksitan', 'az-Cyrl': 'окситан', 'az-Latn': 'oksitan', 'bal-Latn': 'Ositi', 'be': 'аксітанская', 'be-tarask': 'аксітанская', 'bg': 'окситански', 'bgn': 'اوکیتایی', 'bn': 'অক্সিটান', 'br': 'okitaneg', 'brx': 'अक्सितान', 'bs': 'oksitanski', 'bs-Cyrl': 'провансалски', 'bs-Latn': 'oksitanski', 'ca': 'occità', 'ca-ES-valencia': 'occità', 'ccp': '𑄃𑄧𑄇𑄴𑄥𑄨𑄑𑄚𑄴', 'ce': 'окситанойн', 'chr': 'ᎠᏏᏔᏂ', 'ckb': 'ئۆکسیتانی', 'cs': 'okcitánština', 'cy': 'Ocsitaneg', 'da': 'occitansk', 'de': 'Okzitanisch', 'dsb': 'okcitanšćina', 'el': 'Οξιτανικά', 'el-polyton': 'Οξιτανικά', 'en': 'Occitan', 'eo': 'okcitana', 'es': 'occitano', 'et': 'oksitaani', 'eu': 'okzitaniera', 'fa': 'اکسیتان', 'ff-Adlm': '𞤌𞤷𞥆𞤭𞤼𞤢𞤲𞤪𞤫', 'fi': 'oksitaani', 'fil': 'Occitan', 'fo': 'occitanskt', 'fr': 'occitan', 'fur': 'ocitan', 'fy': 'Occitaansk', 'ga': 'Ocsatáinis', 'gd': 'Ogsatanais', 'gl': 'occitano', 'gsw': 'Okzitanisch', 'gu': 'ઓક્સિટન', 'ha': 'Ositanci', 'he': 'אוקסיטנית', 'hi': 'ओसीटान', 'hi-Latn': 'Occitan', 'hr': 'okcitanski', 'hsb': 'okcitanšćina', 'ht': 'occitan', 'hu': 'okszitán', 'hy': 'օքսիտաներեն', 'ia': 'occitano', 'id': 'Ositania', 'ig': 'Asụsụ Osịtan', 'is': 'oksítaníska', 'it': 'occitano', 'ja': 'オック語', 'jv': 'Ossitan', 'ka': 'ოქსიტანური', 'kgp': 'ogsitỹnũ', 'kk': 'окситан тілі', 'kk-Cyrl': 'окситан тілі', 'km': 'អូសីតាន់', 'kn': 'ಒಸಿಟನ್', 'ko': '오크어', 'kok': 'ऑक्सिटन', 'kok-Deva': 'ऑक्सिटन', 'ks': 'اوکسیٖٹَن', 'ks-Arab': 'اوکسیٖٹَن', 'ku': 'oksîtanî', 'ky': 'окситанча', 'lb': 'Okzitanesch', 'lld': 'ocitan', 'lo': 'ອັອກຊີຕານ', 'lt': 'očitarų', 'lv': 'oksitāņu', 'mai': 'औकसीटान', 'mi': 'Ōkitana', 'mk': 'окситански', 'ml': 'ഓക്\u200cസിറ്റൻ', 'mn': 'окситан', 'mr': 'ऑक्सितान', 'ms': 'Occitania', 'mt': 'Oċċitan', 'my': 'အိုစီတန်', 'nb': 'oksitansk', 'ne': 'अक्सिटन', 'nl': 'Occitaans', 'nn': 'oksitansk', 'no': 'oksitansk', 'oc': 'occitan', 'om': 'Afaan Occit', 'or': 'ଓସିଟାନ୍', 'pa': 'ਓਕਸੀਟਾਨ', 'pa-Guru': 'ਓਕਸੀਟਾਨ', 'pcm': 'Oksitán Lángwej', 'pl': 'oksytański', 'ps': 'اوکسيټاني', 'pt': 'occitânico', 'qu': 'Occitano Simi', 'rm': 'occitan', 'ro': 'occitană', 'ru': 'окситанский', 'rw': 'Inyogusitani', 'sc': 'otzitanu', 'sd': 'آڪسيٽن', 'sd-Arab': 'آڪسيٽن', 'se': 'oksitánagiella', 'si': 'ඔසිටාන්', 'sk': 'okcitánčina', 'sl': 'okcitanščina', 'smn': 'oksitan', 'so': 'Occitan', 'sq': 'oksitanisht', 'sr': 'окситански', 'sr-Cyrl': 'окситански', 'sr-Latn': 'oksitanski', 'sv': 'occitanska', 'sw': 'Kiokitani', 'ta': 'ஒக்கிடன்', 'te': 'ఆక్సిటన్', 'tg': 'окситанӣ', 'th': 'อ็อกซิตัน', 'ti': 'ኦክሲታንኛ', 'tk': 'oksitan dili', 'to': 'lea fakaʻokitane', 'tr': 'Oksitan dili', 'tt': 'окситан', 'ug': 'ئوكسىتچە', 'uk': 'окситанська', 'ur': 'آکسیٹان', 'uz': 'oksitan', 'uz-Cyrl': 'окситанча', 'uz-Latn': 'oksitan', 'vec': 'ositan', 'vi': 'Tiếng Occitan', 'wo': 'Ositan', 'xnr': 'ओसीटान', 'yi': 'אקסיטאַניש', 'yo': 'Èdè Ọ̀kísítáànì', 'yrl': 'usitãniku', 'yue': '奧克西坦文', 'yue-Hans': '奥克西坦文', 'yue-Hant': '奧克西坦文', 'zh': '奥克语', 'zh-Hans': '奥克语', 'zh-Hant': '奧克西坦文', 'zu': 'isi-Occitan'},
    'oca': {'en': 'Ocaina'},
    'och': {'en': 'Old Chinese'},
    'ocm': {'en': 'Old Cham'},
    'oco': {'en': 'Old Cornish'},
    'ocu': {'en': 'Atzingo Matlatzinca'},
    'oda': {'en': 'Odut'},
    'odk': {'en': 'Od'},
    'odt': {'en': 'Old Dutch'},
    'odu': {'en': 'Odual'},
    'ofo': {'en': 'Ofo'},
    'ofs': {'en': 'Old Frisian'},
    'ofu': {'en': 'Efutop'},
    'ogb': {'en': 'Ogbia'},
    'ogc': {'en': 'Ogbah'},
    'oge': {'en': 'Old Georgian'},
    'ogg': {'en': 'Ogbogolo'},
    'ogo': {'en': 'Khana'},
    'ogu': {'en': 'Ogbronuagum'},
    'oht': {'en': 'Old Hittite'},
    'ohu': {'en': 'Old Hungarian'},
    'oia': {'en': 'Oirata'},
    'oin': {'en': 'Inebu One'},
    'oj': {'ar': 'الأوجيبوا', 'ast': 'ojibwa', 'az': 'ocibva', 'az-Latn': 'ocibva', 'be': 'аджыбвэ', 'be-tarask': 'аджыбвэ', 'bg': 'оджибва', 'bn': 'ওজিবওয়া', 'br': 'ojibweg', 'brx': 'ओहीबवा', 'bs': 'ojibva', 'bs-Cyrl': 'ојибва', 'bs-Latn': 'ojibva', 'ca': 'ojibwa', 'ca-ES-valencia': 'ojibwa', 'ccp': '𑄃𑄮𑄎𑄨𑄝𑄧𑄤', 'cs': 'odžibvejština', 'cy': 'Ojibwa', 'da': 'ojibwa', 'de': 'Ojibwa', 'el': 'Οζιβίγουα', 'el-polyton': 'Οζιβίγουα', 'en': 'Ojibwa', 'es': 'ojibwa', 'et': 'odžibvei', 'fa': 'اوجیبوایی', 'fi': 'odžibwa', 'fr': 'ojibwa', 'fy': 'Ojibwa', 'ga': 'Óisibis', 'gd': 'Ojibwa', 'gsw': 'Ojibwa-Schpraach', 'gu': 'ઓજિબ્વા', 'he': 'אוג׳יבווה', 'hi': 'ओजिब्वा', 'hi-Latn': 'Ojibwa', 'hr': 'ojibwa', 'ht': 'ojibwa', 'hu': 'ojibva', 'hy': 'օջիբվա', 'id': 'Ojibwa', 'is': 'ojibva', 'it': 'ojibwa', 'ja': 'オジブウェー語', 'ka': 'ოჯიბვე', 'kgp': 'ojimva', 'kn': 'ಒಜಿಬ್ವಾ', 'ko': '오지브와어', 'ks': 'اوجِبوا', 'ks-Arab': 'اوجِبوا', 'lb': 'Ojibwa-Sprooch', 'lkt': 'Ȟaȟátȟuŋwaŋ Iyápi', 'lo': 'ໂອຈິບວາ', 'lt': 'ojibva', 'lv': 'odžibvu', 'mk': 'оџибва', 'ml': 'ഓജിബ്വാ', 'mr': 'ओजिब्वा', 'mt': 'Oġibwa', 'nb': 'ojibwa', 'ne': 'ओजिब्वा', 'nl': 'Ojibwa', 'nn': 'ojibwa', 'no': 'ojibwa', 'or': 'ଓଜିୱା', 'pl': 'odżibwa', 'pt': 'ojibwa', 'rm': 'ojibwa', 'ro': 'ojibwa', 'ru': 'оджибва', 'sk': 'odžibva', 'sl': 'anašinabščina', 'sr': 'оџибве', 'sr-Cyrl': 'оџибве', 'sr-Latn': 'odžibve', 'sv': 'odjibwa', 'ta': 'ஒஜிப்வா', 'te': 'చేవా', 'th': 'โอจิบวา', 'to': 'lea fakaʻosipiuā', 'tr': 'Ojibva dili', 'ug': 'ئوجىبۋاچە', 'uk': 'оджібва', 'vi': 'Tiếng Ojibwa', 'yrl': 'uyibua', 'yue': '奧杰布瓦文', 'yue-Hans': '奥杰布瓦文', 'yue-Hant': '奧杰布瓦文', 'zh': '奥吉布瓦语', 'zh-Hans': '奥吉布瓦语', 'zh-Hant': '奧杰布瓦文'},
    'ojb': {'af': 'Noordwes-Ojibwa', 'am': 'ሰሜን ምዕራባዊ ኦጂብዋ', 'ar': 'أوجيبوا الشمالية الغربية', 'as': 'নৰ্থ-ৱেষ্টাৰ্ণ অজিবৱা', 'az': 'şimal-qərbi ocibva', 'az-Latn': 'şimal-qərbi ocibva', 'be': 'паўночна-заходняя аджыбвэ', 'be-tarask': 'паўночна-заходняя аджыбвэ', 'bg': 'северозападен оджибве', 'bn': 'উত্তর পশ্চিম ওজিবোয়া', 'br': 'ojibweg ar Gwalarn', 'brx': 'साहा-सोनाबारि अजिब्वा', 'bs': 'sjeverozapadni ojibva', 'bs-Cyrl': 'сјеверозападни ојибва', 'bs-Latn': 'sjeverozapadni ojibva', 'ca': 'ojibwa septentrional', 'ca-ES-valencia': 'ojibwa septentrional', 'chr': 'ᏧᏴᏢ ᏭᏕᎵᎬ ᎣᏥᏆ', 'cs': 'odžibvejština (severozápadní)', 'cy': 'Ojibwa gogledd-orllewin', 'da': 'nordvestojibwa', 'de': 'Nordwest-Ojibwe', 'dsb': 'dłujkowjacorna ojibwašćina', 'el': 'Βορειοδυτικά Οζιβίγουα', 'el-polyton': 'Βορειοδυτικά Οζιβίγουα', 'en': 'Northwestern Ojibwa', 'es': 'ojibwa noroccidental', 'et': 'loodeodžibvei', 'eu': 'ipar-mendebaldeko ojibwa', 'fa': 'اوجیبوای شمالی', 'ff-Adlm': '𞤌𞤶𞤭𞤦𞤵𞤱𞤢𞥄𞤪𞤫 𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤫', 'fi': 'luoteis-odžibwa', 'fil': 'Hilagang-Kanluran ng Ojibwa', 'fr': 'ojibwé du Nord-Ouest', 'ga': 'Óisibis Iarthuiscirt', 'gd': 'Ojibwa Iar-thuathach', 'gl': 'ojibwa do noroeste', 'gu': 'ઉત્તરપશ્ચિમી ઓઝિબવે', 'ha': 'Ojibwa na Arewa-Maso-Yamma', 'he': 'אוג׳יבווה צפון מערבית', 'hi': 'उत्तरपश्चिमी ओजिब्वे', 'hi-Latn': 'Northwestern Ojibwa', 'hr': 'sjeverozapadni ojibwa', 'hsb': 'sewjernozapadna ojibwašćina', 'ht': 'ojibwé du Nord-Ouest', 'hu': 'északnyugati odzsibva', 'hy': 'հյուսիս-արևմտյան օջիբվե', 'ia': 'ojibwa del nord-west', 'id': 'Ojibwe Barat Laut', 'ig': 'Northwestern Ojibwa', 'is': 'norðvestur-ojibwa', 'it': 'ojibwe nord-occidentale', 'ja': '北西部オジブワ語', 'jv': 'Ojibwa Kulon Segara', 'ka': 'ჩრდილო-დასავლეთის ოჯიბვა', 'kk': 'солтүстік-батыс оджибве тілі', 'kk-Cyrl': 'солтүстік-батыс оджибве тілі', 'km': 'អូជីបវ៉ា\u200bប៉ែកពាយ័ព្យ', 'kn': 'ವಾಯುವ್ಯ ಓಜಿಬ್ವಾ', 'ko': '북서부 오지브와어', 'kok': 'वायव्य ओजिब्वे', 'kok-Deva': 'वायव्य ओजिब्वे', 'ku': 'ojibweyîya bakurî', 'ky': 'түндүк-батыш ожибвече (индей тили)', 'lo': 'ໂອຈິບວາຕາເວັນຕົກສ່ຽງເໜືອ', 'lt': 'šiaurės vakarų odžibvių', 'lv': 'ziemeļrietumu odžibvu', 'mai': 'उत्तरी-पश्चिम ओजिवा', 'mi': 'Ōtīpia Raki-mā-Uru', 'mk': 'северозападен оџибва', 'ml': 'നോർത്ത്\u200cവെസ്റ്റേൺ ഒജീബ്\u200cവെ', 'mn': 'баруун хойд ожибва', 'mr': 'नॉर्थवेस्टर्न ओजिब्वा', 'ms': 'Ojibwa Barat Laut', 'my': 'အိုဂျစ်ဘွာ', 'nb': 'nordvestlig ojibwa', 'ne': 'उत्तरपश्चिम ओजिब्बा', 'nl': 'Noordwest-Ojibwe', 'nn': 'nordvestleg ojibwa', 'no': 'nordvestlig ojibwa', 'or': 'ଉତ୍ତର-ପଶ୍ଚିମ ଓଜିବ୍ଵା', 'pa': 'ਉੱਤਰ-ਪੱਛਮੀ ਔਜਿਬਵਾ', 'pa-Guru': 'ਉੱਤਰ-ਪੱਛਮੀ ਔਜਿਬਵਾ', 'pcm': 'Nọ́tán Ojibwa', 'pl': 'odżibwe północno-zachodni', 'ps': 'شمال لویدیځ اوجیبوا', 'pt': 'ojibwa do noroeste', 'qu': 'Ojibwa del noroeste Simi', 'ro': 'ojibwa de nord-vest', 'ru': 'северо-западный оджибве', 'sc': 'ojibwa nord-otzidentale', 'sd': 'اتر الھندي اوجيبوا', 'sd-Arab': 'اتر الھندي اوجيبوا', 'si': 'වයඹ ඔජිබ්වා', 'sk': 'northwestern ojibwa', 'sl': 'severozahodna očipvejščina', 'so': 'Northwestern Ojibwa', 'sq': 'oxhibuaishte verilindore', 'sr': 'северозападни оџибва', 'sr-Cyrl': 'северозападни оџибва', 'sr-Latn': 'severozapadni odžibva', 'sv': 'nordvästojibwa', 'sw': 'Kiojibwa cha Kaskazini Magharibi', 'ta': 'வடமேற்கு ஓஜிப்வா', 'te': 'వాయువ్య ఓజిబ్వా', 'th': 'โอจิบเวตะวันตกเฉียงเหนือ', 'ti': 'ሰሜናዊ ምዕራብ ኦጂብዋ', 'tk': 'demirgazyk-günbatar ojibwa dili', 'to': 'lea fakaʻosipiuā-tokelauhihifo', 'tr': 'Kuzeybatı Ojibwe dili', 'uk': 'північно-західна оджибве', 'ur': 'شمال مغربی اوجبوا', 'uz': 'shimoli-gʻarbiy ojibva', 'uz-Latn': 'shimoli-gʻarbiy ojibva', 'vec': 'ojibwa nord osidentale', 'vi': 'Tiếng Ojibwe Tây Bắc', 'yo': 'Èdè Ariwa-iwoorun Ojibwa', 'yue': '西北奧吉布瓦文', 'yue-Hans': '西北奥吉布瓦文', 'yue-Hant': '西北奧吉布瓦文', 'zh': '西北部奥吉布瓦语', 'zh-Hans': '西北部奥吉布瓦语', 'zh-Hant': '西北奧吉布瓦語', 'zu': 'Northwestern Ojibwa'},
    'ojc': {'af': 'Sentraal-Ojibwa', 'am': 'ማዕከላዊ ኦጂብዋ', 'ar': 'أوجيبوا الوسطى', 'as': 'চেন্ট্ৰেক অজিবৱা', 'az': 'Mərkəzi ocibva', 'az-Latn': 'Mərkəzi ocibva', 'be': 'цэнтральная аджыбвэ', 'be-tarask': 'цэнтральная аджыбвэ', 'bg': 'централен оджибва', 'bn': 'মধ্য ওজিবুয়া', 'br': 'ojibweg ar cʼhreiz', 'brx': 'मिरु अजिब्वा', 'bs': 'centralni ojibva', 'bs-Cyrl': 'Централни обијва', 'bs-Latn': 'centralni ojibva', 'ca': 'ojibwa central', 'ca-ES-valencia': 'ojibwa central', 'chr': 'ᎠᏰᏟ ᎣᏥᏆ', 'cs': 'odžibvejština (střední)', 'cy': 'Ojibwa Canolog', 'da': 'centralojibwa', 'de': 'Zentral-Ojibwe', 'dsb': 'centralna ojibwašćina', 'el': 'Κεντρικά Οτζίμπουα', 'el-polyton': 'Κεντρικά Οτζίμπουα', 'en': 'Central Ojibwa', 'es': 'ojibwa central', 'et': 'keskodžibvei', 'eu': 'erdialdeko ojibwa', 'fa': 'اوجیبوای مرکزی', 'ff-Adlm': '𞤌𞤶𞤭𞤦𞤵𞤱𞤪𞤫 𞤕𞤢𞤳𞤢', 'fi': 'keskiojibwa', 'fil': 'Central Ojibwa', 'fr': 'ojibwé central', 'ga': 'Óisibis Lárnach', 'gd': 'Ojibwa Mheadhanach', 'gl': 'ojibwa', 'gu': 'સેન્ટ્રલ ઓઝિબ્વા', 'ha': 'Ojibwa na Tsakiya', 'he': 'אוג׳יבווה (מרכז)', 'hi': 'सेंट्रल ओजीब्वे', 'hi-Latn': 'Central Ojibwa', 'hr': 'centralni ojibwa', 'hsb': 'centralna ojibwašćina', 'ht': 'ojibwé central', 'hu': 'középvidéki odzsibva', 'hy': 'կենտրոնական օջիբվե', 'ia': 'ojibwa central', 'id': 'Ojibwe Tengah', 'ig': 'Central Ojibwa', 'is': 'ojibwa', 'it': 'ojibwe centrale', 'ja': '中部オジブワ語', 'jv': 'Ojibwa Tengah', 'ka': 'ცენტრალური ოჯიბვე', 'kk': 'орталық оджибве тілі', 'kk-Cyrl': 'орталық оджибве тілі', 'km': 'អូជីពវ៉ាកណ្ដាល', 'kn': 'ಮಧ್ಯ ಓಜಿಬ್ವಾ', 'ko': '중앙 오지브와어', 'kok': 'मध्य ओजिब्वे', 'kok-Deva': 'मध्य ओजिब्वे', 'ku': 'ojibwayîya navîn', 'ky': 'борбордук ожибвече', 'lo': 'ໂອຈິບວາກາງ', 'lt': 'ojibvų', 'lv': 'centrālā odžibvu', 'mai': 'सेंट्रल ओजिबा', 'mi': 'Ohīpawe Waenga', 'mk': 'централен оџибва', 'ml': 'സെൻട്രൽ ഒജീബ്\u200cവെ', 'mn': 'төв ожибва', 'mr': 'सेंट्रल ओजिब्वा', 'ms': 'Ojibwa Tengah', 'my': 'အိုဂျစ်ဘွေး', 'nb': 'ojibwa (sentral)', 'ne': 'सेन्ट्रल अजिब्बा', 'nl': 'Centraal Ojibwa', 'nn': 'ojibwa (sentral)', 'no': 'ojibwa (sentral)', 'or': 'କେନ୍ଦ୍ରୀୟ ଓଜିବ୍ଵା', 'pa': 'ਸੈਂਟਰਲ ਔਜਿਬਵਾ', 'pa-Guru': 'ਸੈਂਟਰਲ ਔਜਿਬਵਾ', 'pcm': 'Sẹ́ntrál Ojíbwa', 'pl': 'odżibwe centralny', 'ps': 'مرکزي اوجیبوا', 'pt': 'ojibwa central', 'qu': 'Ojibwa Central', 'ro': 'ojibwa centrală', 'ru': 'центральный оджибве', 'sc': 'ojibwa tzentrale', 'sd': 'وچولي اوجيبوي', 'sd-Arab': 'وچولي اوجيبوي', 'si': 'මධ්\u200dයම ඔජිබ්වා', 'sk': 'centrálna odžibvejčina', 'sl': 'osrednja očipvejščina', 'so': 'Central Ojibwa', 'sq': 'oxhibuaishte qendrore', 'sr': 'централни оџибва', 'sr-Cyrl': 'централни оџибва', 'sr-Latn': 'centralni odžibva', 'sv': 'ojibwa', 'sw': 'Kiojibwa cha kati', 'ta': 'மத்திய ஓஜிப்வா', 'te': 'సెంట్రల్ ఓజిబ్వా', 'th': 'โอจิบเวตอนกลาง', 'ti': 'ማእከላይ ኦጂብዋ', 'tk': 'merkezi ojibwa dili', 'to': 'lea fakaʻosipiuā-loto', 'tr': 'Orta Ojibwe dili', 'uk': 'центральна оджибве', 'ur': 'سینٹرل اوجبوا', 'uz': 'markaziy ijibve', 'uz-Latn': 'markaziy ijibve', 'vec': 'ojibwa sentrale', 'vi': 'Tiếng Ojibwe miền Trung', 'yo': 'Èdè Ojibwa Aarin', 'yue': '中奧吉布瓦文', 'yue-Hans': '中奥吉布瓦文', 'yue-Hant': '中奧吉布瓦文', 'zh': '中奥吉布瓦语', 'zh-Hans': '中奥吉布瓦语', 'zh-Hant': '中央奧吉布瓦語', 'zu': 'isi-Central Ojibwa'},
    'ojg': {'en': 'Eastern Ojibwa'},
    'ojp': {'en': 'Old Japanese'},
    'ojs': {'af': 'Oji-Cree', 'am': 'ኦጂ-ክሪ', 'ar': 'الأوجي-كري', 'as': 'অ’জি-ক্ৰী', 'az': 'ocikri', 'az-Latn': 'ocikri', 'be': 'оджы-кры', 'be-tarask': 'оджы-кры', 'bg': 'оджи крий', 'bn': 'ওজি-ক্রী', 'br': 'ojibweg Severn', 'brx': 'अजि क्री', 'bs': 'odži kri', 'bs-Cyrl': 'оји-кре', 'bs-Latn': 'odži kri', 'ca': 'oji-cree', 'ca-ES-valencia': 'oji-cree', 'chr': 'ᎣᏥ-Ꮯ', 'cs': 'odžibvejština (severní)', 'cy': 'Oji-Cree', 'da': 'oji-cree', 'de': 'Oji-Cree', 'dsb': 'oji-creešćina', 'el': 'Ότζι-Κρι', 'el-polyton': 'Ότζι-Κρι', 'en': 'Oji-Cree', 'es': 'oji-cree', 'et': 'Severni odžibvei', 'eu': 'oji-creera', 'fa': 'اوجی-کری', 'ff-Adlm': '𞤌𞤶𞤭-𞤑𞤪𞤭𞥅𞤪𞤫', 'fi': 'oji-cree', 'fil': 'Oji-Cree', 'fr': 'oji-cri', 'ga': 'Oji-Cree', 'gd': 'Oji-Cree', 'gl': 'oji-cree', 'gu': 'ઓજી-ક્રી', 'ha': 'Oji-Cree', 'he': 'אוג׳י-קרי', 'hi': 'ओजी-क्री', 'hi-Latn': 'Oji-Cree', 'hr': 'oji-cree', 'hsb': 'oji-creešćina', 'ht': 'oji-cri', 'hu': 'odzsi-krí', 'hy': 'օջի քրի', 'ia': 'oji-cree', 'id': 'Oji-Cree', 'ig': 'Oji-Cree', 'is': 'oji-cree', 'it': 'oji-cree', 'ja': 'セヴァーン・オジブワ語', 'jv': 'Oji-Kree', 'ka': 'ოჯი-კრი', 'kk': 'оджи-кри тілі', 'kk-Cyrl': 'оджи-кри тілі', 'km': 'អូជីគ្រី', 'kn': 'ಓಜಿ-ಕ್ರೀ', 'ko': '오지 크리어', 'kok': 'ओजि-क्री', 'kok-Deva': 'ओजि-क्री', 'ku': 'oji-cree', 'ky': 'ожи-криче (индей тили)', 'lo': 'ໂອຈິຄຣີ', 'lt': 'odži kri', 'lv': 'odži-krī', 'mai': 'ओजि-क्री', 'mi': 'Ōti-Kirī', 'mk': 'очиски кријски', 'ml': 'ഒജി-ക്രീ', 'mn': 'ожи кри', 'mr': 'ओजी-क्री', 'ms': 'Oji-Cree', 'my': 'အိုဂျီခရီ', 'nb': 'oji-cree', 'ne': 'ओजी क्री', 'nl': 'Oji-Cree', 'nn': 'oji-cree', 'no': 'oji-cree', 'or': 'ଓଜି-କ୍ରି', 'pa': 'ਓਜੀ-ਕ੍ਰੀ', 'pa-Guru': 'ਓਜੀ-ਕ੍ਰੀ', 'pcm': 'Ojí-Krii Lángwej', 'pl': 'odżi-kri', 'ps': 'اوجي-کري', 'pt': 'oji-cree', 'qu': 'Oji-Cree Simi', 'ro': 'oji-cree', 'ru': 'оджи-кри', 'sc': 'oji-Cree', 'sd': 'اوجي ڪري', 'sd-Arab': 'اوجي ڪري', 'si': 'ඔජි-ක්\u200dරී', 'sk': 'oji-cree', 'sl': 'oči-krijščina', 'so': 'Oji-Cree', 'sq': 'oxhikrijisht', 'sr': 'оџикри', 'sr-Cyrl': 'оџикри', 'sr-Latn': 'odžikri', 'sv': 'oji-cree', 'sw': 'Kikrii cha Oji', 'ta': 'ஓஜி-க்ரீ', 'te': 'ఓజి-క్రీ', 'th': 'ออจิ-ครี', 'ti': 'ኦጂ-ክሪ', 'tk': 'oji-kri dili', 'to': 'lea fakakilī-osi', 'tr': 'Anişininice', 'uk': 'оджи-крі', 'ur': 'اوجی کری', 'uz': 'oji-kri', 'uz-Latn': 'oji-kri', 'vec': 'oji cree', 'vi': 'Tiếng Oji-Cree', 'yo': 'Èdè Oji Kri', 'yue': '王子克里文', 'yue-Hans': '王子克里文', 'yue-Hant': '王子克里文', 'zh': '欧吉克里语', 'zh-Hans': '欧吉克里语', 'zh-Hant': '奧吉克里語', 'zu': 'isi-Oji-Cree'},
    'ojv': {'en': 'Ontong Java'},
    'ojw': {'af': 'Wes-Ojibwa', 'am': 'ምዕራባዊ ኦጂቡዋ', 'ar': 'الأوجيبوا الغربية', 'as': 'ৱেষ্টাৰ্ণ অজিবৱা', 'az': 'qərbi ocibva', 'az-Latn': 'qərbi ocibva', 'be': 'заходняя аджыбвэ', 'be-tarask': 'заходняя аджыбвэ', 'bg': 'западен оджибва', 'bn': 'পশ্চিম ওজিবোয়া', 'br': 'ojibweg ar Cʼhornôg', 'brx': 'सोनाबारि अजिब्वा', 'bs': 'zapadni ojibva', 'bs-Cyrl': 'западни ојибва', 'bs-Latn': 'zapadni ojibva', 'ca': 'ojibwa occidental', 'ca-ES-valencia': 'ojibwa occidental', 'chr': 'ᏭᏕᎵᎬ ᏗᏜ ᎣᏥᏆ', 'cs': 'odžibvejština (západní)', 'cy': 'Ojibwa Gorllewinol', 'da': 'vestojibwa', 'de': 'West-Ojibwe', 'dsb': 'pódwjacorna ojibwašćina', 'el': 'Δυτικά Οζιβίγουα', 'el-polyton': 'Δυτικά Οζιβίγουα', 'en': 'Western Ojibwa', 'es': 'ojibwa occidental', 'et': 'lääneodžibvei', 'eu': 'mendebaldeko ojibwa', 'fa': 'اوجیبوای غربی', 'ff-Adlm': '𞤌𞤶𞤭𞤦𞤱𞤢𞥄𞤪𞤫 𞤖𞤭𞥅𞤪𞤲𞤢𞥄𞤲𞤺𞤫', 'fi': 'länsi-odžibwa', 'fil': 'Kanlurang Ojibwa', 'fr': 'ojibwé occidental', 'ga': 'Óisibis an Iarthar', 'gd': 'Ojibwa Shiarach', 'gl': 'ojibwa do oeste', 'gu': 'પશ્ચિમી ઓઝિબ્વા', 'ha': 'Ojibwa na Yammaci', 'he': 'אוביג׳ווה מערבית', 'hi': 'पश्चिमी ओजिब्वा', 'hi-Latn': 'Western Ojibwa', 'hr': 'zapadni ojibwa', 'hsb': 'zapadna ojibwašćina', 'ht': 'ojibwé occidental', 'hu': 'nyugati odzsibva', 'hy': 'արևմտյան օջիբվե', 'ia': 'ojibwa del west', 'id': 'Ojibwe Barat', 'ig': 'Westarn Ojibwa', 'is': 'vestur-ojibwa', 'it': 'ojibwe occidentale', 'ja': '西部オジブワ語', 'jv': 'Ojibwa Sisih Kulon', 'ka': 'დასავლეთის ოჯიბვა', 'kk': 'батыс оджибве тілі', 'kk-Cyrl': 'батыс оджибве тілі', 'km': 'អូជីបវ៉ា\u200bខាងលិច', 'kn': 'ಪಶ್ಚಿಮ ಓಜಿಬ್ವಾ', 'ko': '서부 오지브와어', 'kok': 'पश्चिमी ओजिब्वे', 'kok-Deva': 'पश्चिमी ओजिब्वे', 'ku': 'ojîbweyîya rojavayî', 'ky': 'батыш ожибвече (индей тили)', 'lo': 'ໂອຈິບວາຕາເວັນຕົກ', 'lt': 'vakarų odžibvių', 'lv': 'rietumodžibvu', 'mai': 'पश्चिमी ओजिबा', 'mi': 'Ōhīpiwa ki te Uru', 'mk': 'западен оџибва', 'ml': 'വെസ്റ്റേൺ ഒജീബ്\u200cവെ', 'mn': 'баруун ожибва', 'mr': 'वेस्टर्न ओजिबवा', 'ms': 'Ojibwa Barat', 'my': 'အနောက် အိုဂျီဘွာ', 'nb': 'vestlig ojibwa', 'ne': 'पश्चिमी ओजिबा', 'nl': 'West-Ojibwe', 'nn': 'vestleg ojibwa', 'no': 'vestlig ojibwa', 'or': 'ପାଶ୍ଚାତ୍ୟ ଓଜିବ୍ଵା', 'pa': 'ਪੱਛਮੀ ਓਜਿਬਵਾ', 'pa-Guru': 'ਪੱਛਮੀ ਓਜਿਬਵਾ', 'pcm': 'Wẹ́stán Ojibua', 'pl': 'odżibwe zachodni', 'ps': 'لویدیځ اوجیبوا', 'pt': 'ojibwa ocidental', 'qu': 'Ojibwa Occidental', 'ro': 'ojibwa de vest', 'ru': 'западный оджибве', 'sc': 'ojibwa otzidentale', 'sd': 'مغربي اوجيبو', 'sd-Arab': 'مغربي اوجيبو', 'si': 'බටහිර ඔජිබ්වා', 'sk': 'ojibwa (západ)', 'sl': 'zahodna očipvejščina', 'so': 'Western Ojibwa', 'sq': 'oxhibuaishte perëndimore', 'sr': 'западни оџибва', 'sr-Cyrl': 'западни оџибва', 'sr-Latn': 'zapadni odžibva', 'sv': 'västojibwe', 'sw': 'Kiojibwa cha Magharibi', 'ta': 'மேற்கு ஓஜிப்வா', 'te': 'పశ్చిమ ఓజిబ్వా', 'th': 'โอจิบเวตะวันตก', 'ti': 'ምዕራባዊ ኦጂብዋ', 'tk': 'günbatar ojibwa dili', 'to': 'lea fakaʻosipiuā-hihifo', 'tr': 'Batı Ojibwe dili', 'uk': 'західна оджибве', 'ur': 'مغربی اوجبوا', 'uz': 'gʻarbiy ojibva', 'uz-Latn': 'gʻarbiy ojibva', 'vec': 'ojibwa osidentale', 'vi': 'Tiếng Ojibwe miền Tây', 'yo': 'Èdè Iwoorun Ojibwa', 'yue': '西奧吉布瓦文', 'yue-Hans': '西奥吉布瓦文', 'yue-Hant': '西奧吉布瓦文', 'zh': '西奥吉布瓦语', 'zh-Hans': '西奥吉布瓦语', 'zh-Hant': '西奧吉布瓦語', 'zu': 'Western Ojibwa'},
    'oka': {'af': 'Okanagan', 'am': 'ኦካናጋን', 'ar': 'الأوكاناغانية', 'as': 'অ’কানাগান', 'az': 'okanaqan', 'az-Latn': 'okanaqan', 'be': 'аканаган', 'be-tarask': 'аканаган', 'bg': 'оканаган', 'bn': 'ওকানাগান', 'br': 'okanaganeg', 'brx': 'अकानागान', 'bs': 'okanagan', 'bs-Cyrl': 'оканаган', 'bs-Latn': 'okanagan', 'ca': 'okanagà', 'ca-ES-valencia': 'okanagà', 'chr': 'ᎣᎧᎾᎦᏂ', 'cs': 'okanaganština', 'cy': 'Okanagan', 'da': 'okanagan', 'de': 'Okanagan', 'dsb': 'okanaganšćina', 'el': 'Οκανάγκαν', 'el-polyton': 'Οκανάγκαν', 'en': 'Okanagan', 'es': 'okanagan', 'et': 'okanagani', 'eu': 'okanaganera', 'fa': 'اوکاناگان', 'ff-Adlm': '𞤌𞤳𞤢𞤲𞤢𞤺𞤢𞤲𞤪𞤫', 'fi': 'okanagan', 'fil': 'Okanagan', 'fr': 'colville-okanagan', 'ga': 'Okanagan', 'gd': 'Okanagan', 'gl': 'okanagan', 'gu': 'ઓકાનાગન', 'ha': 'Okanagan', 'he': 'אוקאנגן', 'hi': 'ओकनागन', 'hi-Latn': 'Okanagan', 'hr': 'okanagan', 'hsb': 'okanaganšćina', 'ht': 'colville-okanagan', 'hu': 'okanagan', 'hy': 'օկանագան', 'ia': 'okanagan', 'id': 'Okanagan', 'ig': 'Okanagan', 'is': 'okanagan', 'it': 'okanagan', 'ja': 'オカナガン語', 'jv': 'Okanagan', 'ka': 'ოკანაგანი', 'kk': 'оканаган тілі', 'kk-Cyrl': 'оканаган тілі', 'km': 'អូកាណាហ្កាន', 'kn': 'ಒಕನಾಗನ್', 'ko': '오카나간어', 'kok': 'ओकानागान', 'kok-Deva': 'ओकानागान', 'ku': 'okanagan', 'ky': 'оканаганча (Канада)', 'lo': 'ໂອກະນາກັນ', 'lt': 'okanaganų', 'lv': 'okanaganu', 'mai': 'ओकंगान', 'mi': 'Ōkanakana', 'mk': 'оканагански', 'ml': 'ഒകാനഗൻ', 'mn': 'оканаган', 'mr': 'ओकनागन', 'ms': 'Okanagan', 'my': 'အိုကနဂန်', 'nb': 'okanagansk', 'ne': 'ओकानागान', 'nl': 'Okanagan', 'nn': 'okanagansk', 'no': 'okanagansk', 'or': 'ଓକାନାଗାନ୍', 'pa': 'ਓਕਾਨਾਗਨ', 'pa-Guru': 'ਓਕਾਨਾਗਨ', 'pcm': 'Okanagan Langwej', 'pl': 'okanagan', 'ps': 'اوکاګان', 'pt': 'okanagan', 'qu': 'Okanagan Simi', 'ro': 'okanagan', 'ru': 'оканаган', 'sc': 'okanagan', 'sd': 'اوڪاناگن', 'sd-Arab': 'اوڪاناگن', 'si': 'ඔකනගන්', 'sk': 'okanagan', 'sl': 'okanaganščina', 'so': 'Okanagan', 'sq': 'okanaganisht', 'sr': 'оканган', 'sr-Cyrl': 'оканган', 'sr-Latn': 'okangan', 'sv': 'okanagan', 'sw': 'Kiokanagani', 'ta': 'ஒகானகன்', 'te': 'ఒకానగన్', 'th': 'โอคานากัน', 'ti': 'ኦካናጋን', 'tk': 'okanagan dili', 'to': 'lea faka-ʻokanākani', 'tr': 'Okanagan dili', 'uk': 'оканаганська', 'ur': 'اوکناگن', 'uz': 'okanagan', 'uz-Latn': 'okanagan', 'vec': 'okanagan', 'vi': 'Tiếng Okanagan', 'yo': 'Èdè Okanaga', 'yue': '奧卡納根文', 'yue-Hans': '奥卡纳根文', 'yue-Hant': '奧卡納根文', 'zh': '欧肯那根语', 'zh-Hans': '欧肯那根语', 'zh-Hant': '奧卡諾根語', 'zu': 'isi-Okanagan'},
    'okb': {'en': 'Okobo'},
    'okc': {'en': 'Kobo'},
    'okd': {'en': 'Okodia'},
    'oke': {'en': 'Okpe (Southwestern Edo)'},
    'okg': {'en': 'Koko Babangk'},
    'okh': {'en': 'Koresh-e Rostam'},
    'oki': {'en': 'Okiek'},
    'okj': {'en': 'Oko-Juwoi'},
    'okk': {'en': 'Kwamtim One'},
    'okl': {'en': 'Old Kentish Sign Language'},
    'okm': {'en': 'Middle Korean (10th-16th cent.)'},
    'okn': {'en': 'Oki-No-Erabu'},
    'oko': {'en': 'Old Korean (3rd-9th cent.)'},
    'okr': {'en': 'Kirike'},
    'oks': {'en': 'Oko-Eni-Osayen'},
    'oku': {'en': 'Oku'},
    'okv': {'en': 'Orokaiva'},
    'okx': {'en': 'Okpe (Northwestern Edo)'},
    'okz': {'en': 'Old Khmer'},
    'ola': {'en': 'Walungge'},
    'old': {'en': 'Mochi'},
    'ole': {'en': 'Olekha'},
    'olk': {'en': 'Olkol'},
    'olm': {'en': 'Oloma'},
    'olo': {'en': 'Livvi'},
    'olr': {'en': 'Olrat'},
    'olt': {'en': 'Old Lithuanian'},
    'olu': {'en': 'Kuvale'},
    'om': {'af': 'Oromo', 'am': 'ኦሮሚኛ', 'ar': 'الأورومية', 'as': 'ওৰোমো', 'ast': 'oromo', 'az': 'oromo', 'az-Cyrl': 'оромо', 'az-Latn': 'oromo', 'bal-Latn': 'Oromó', 'be': 'арома', 'be-tarask': 'арома', 'bg': 'оромо', 'bgn': 'اورومویی', 'bn': 'অরোমো', 'br': 'oromoeg', 'brx': 'अरम’', 'bs': 'oromo', 'bs-Cyrl': 'оромо', 'bs-Latn': 'oromo', 'ca': 'oromo', 'ca-ES-valencia': 'oromo', 'ccp': '𑄃𑄧𑄢𑄮𑄟𑄮', 'ce': 'оромо', 'chr': 'ᎣᎶᎼ', 'ckb': 'ئۆرۆمۆ', 'cs': 'oromština', 'cy': 'Oromo', 'da': 'oromo', 'de': 'Oromo', 'dsb': 'oromo', 'el': 'Ορόμο', 'el-polyton': 'Ορόμο', 'en': 'Oromo', 'eo': 'oroma', 'es': 'oromo', 'et': 'oromo', 'eu': 'oromoera', 'fa': 'اورومویی', 'ff-Adlm': '𞤌𞤪𞤮𞤥𞤮𞥅𞤪𞤫', 'fi': 'oromo', 'fil': 'Oromo', 'fo': 'oromo', 'fr': 'oromo', 'fy': 'Oromo', 'ga': 'Oraimis', 'gd': 'Oromo', 'gl': 'oromo', 'gsw': 'Oromo', 'gu': 'ઓરોમો', 'ha': 'Oromo', 'he': 'אורומו', 'hi': 'ओरोमो', 'hi-Latn': 'Oromo', 'hr': 'oromski', 'hsb': 'oromo', 'ht': 'oromo', 'hu': 'oromo', 'hy': 'օրոմո', 'ia': 'oromo', 'id': 'Oromo', 'ig': 'Oromo', 'is': 'oromo', 'it': 'oromo', 'ja': 'オロモ語', 'jv': 'Oromo', 'ka': 'ორომო', 'kea': 'oromo', 'kgp': 'orãmũ', 'kk': 'оромо тілі', 'kk-Cyrl': 'оромо тілі', 'km': 'អូរ៉ូម៉ូ', 'kn': 'ಒರೊಮೊ', 'ko': '오로모어', 'kok': 'ओरोमो', 'kok-Deva': 'ओरोमो', 'ks': 'اۆرومو', 'ks-Arab': 'اۆرومو', 'ku': 'oromoyî', 'ky': 'оромочо', 'lb': 'Oromo', 'lld': 'oromo', 'lo': 'ໂອໂຣໂມ', 'lrc': 'ئوروموٙ', 'lt': 'oromų', 'lv': 'oromu', 'mai': 'ओरोमो', 'mi': 'Ōromo', 'mk': 'оромо', 'ml': 'ഒറോമോ', 'mn': 'оромо', 'mr': 'ओरोमो', 'ms': 'Oromo', 'mt': 'Oromo', 'my': 'အိုရိုမို', 'mzn': 'اورومو', 'nb': 'oromo', 'ne': 'ओरोमो', 'nl': 'Afaan Oromo', 'nn': 'oromo', 'no': 'oromo', 'nqo': 'ߏߙߏߡߏߞߊ߲', 'om': 'Oromoo', 'or': 'ଓରୋମୋ', 'pa': 'ਓਰੋਮੋ', 'pa-Guru': 'ਓਰੋਮੋ', 'pcm': 'Orómó', 'pl': 'oromo', 'ps': 'اورومو', 'pt': 'oromo', 'qu': 'Oromo Simi', 'rm': 'oromo', 'ro': 'oromo', 'ru': 'оромо', 'sc': 'oromo', 'sd': 'اورومو', 'sd-Arab': 'اورومو', 'si': 'ඔරොමෝ', 'sk': 'oromčina', 'sl': 'oromo', 'smn': 'oromokielâ', 'so': 'Oromo', 'sq': 'oromoisht', 'sr': 'оромо', 'sr-Cyrl': 'оромо', 'sr-Latn': 'oromo', 'sv': 'oromo', 'sw': 'Kioromo', 'syr': 'ܐܘܪܘܡܘܐܝܬ', 'ta': 'ஒரோமோ', 'te': 'ఒరోమో', 'tg': 'оромо', 'th': 'โอโรโม', 'ti': 'ኦሮሞ', 'tk': 'oromo dili', 'to': 'lea fakaʻolomo', 'tr': 'Oromo dili', 'tt': 'оромо', 'ug': 'ئوروموچە', 'uk': 'оромо', 'ur': 'اورومو', 'uz': 'oromo', 'uz-Cyrl': 'оромо', 'uz-Latn': 'oromo', 'vec': 'oromo', 'vi': 'Tiếng Oromo', 'wo': 'Oromo', 'xnr': 'ओरोमो', 'yo': 'Òròmọ́', 'yrl': 'urumu', 'yue': '奧羅莫文', 'yue-Hans': '奥罗莫文', 'yue-Hant': '奧羅莫文', 'zh': '奥罗莫语', 'zh-Hans': '奥罗莫语', 'zh-Hant': '奧羅莫文', 'zu': 'isi-Oromo'},
    'oma': {'en': 'Omaha-Ponca'},
    'omb': {'en': 'East Ambae'},
    'omc': {'en': 'Mochica'},
    'ome': {'en': 'Omejes'},
    'omg': {'en': 'Omagua'},
    'omi': {'en': 'Omi'},
    'omk': {'en': 'Omok'},
    'oml': {'en': 'Ombo'},
    'omn': {'en': 'Minoan'},
    'omo': {'en': 'Utarmbung'},
    'omp': {'en': 'Old Manipuri'},
    'omq': {'en': 'Oto-Manguean languages'},
    'omq-maz-pro': {'en': 'Proto-Mazatec'},
    'omq-mix-pro': {'en': 'Proto-Mixtecan'},
    'omq-pro': {'en': 'Proto-Oto-Manguean'},
    'omr': {'en': 'Old Marathi'},
    'omt': {'en': 'Omotik'},
    'omu': {'en': 'Omurano'},
    'omv': {'en': 'Omotic languages'},
    'omw': {'en': 'South Tairora'},
    'omx': {'en': 'Old Mon'},
    'omy': {'en': 'Old Malay'},
    'ona': {'en': 'Ona'},
    'onb': {'en': 'Lingao'},
    'one': {'en': 'Oneida'},
    'ong': {'en': 'Olo'},
    'oni': {'en': 'Onin'},
    'onj': {'en': 'Onjob'},
    'onk': {'en': 'Kabore One'},
    'onn': {'en': 'Onobasulu'},
    'ono': {'en': 'Onondaga'},
    'onp': {'en': 'Sartang'},
    'onr': {'en': 'Northern One'},
    'ons': {'en': 'Ono'},
    'ont': {'en': 'Ontenu'},
    'onu': {'en': 'Unua'},
    'onw': {'en': 'Old Nubian'},
    'onx': {'en': 'Onin Based Pidgin'},
    'ood': {'en': "Tohono O'odham"},
    'oog': {'en': 'Ong'},
    'oon': {'en': 'Önge'},
    'oor': {'en': 'Oorlams'},
    'oos': {'en': 'Old Ossetic'},
    'opa': {'en': 'Okpamheri'},
    'opk': {'en': 'Kopkaka'},
    'opm': {'en': 'Oksapmin'},
    'opo': {'en': 'Opao'},
    'opt': {'en': 'Opata'},
    'opy': {'en': 'Ofayé'},
    'or': {'af': 'Oriya', 'ak': 'Odia', 'am': 'ኦዲያ', 'ar': 'الأورية', 'as': 'ওড়িয়া', 'ast': 'oriya', 'az': 'odiya', 'az-Cyrl': 'одија', 'az-Latn': 'odiya', 'bal-Latn': 'Odi', 'be': 'орыя', 'be-tarask': 'орыя', 'bg': 'ория', 'bgn': 'اودیه\u200cای', 'bn': 'ওড়িয়া', 'br': 'oriya', 'brx': 'उरीया', 'bs': 'odija', 'bs-Cyrl': 'одија', 'bs-Latn': 'odija', 'ca': 'oriya', 'ca-ES-valencia': 'oriya', 'ccp': '𑄃𑄮𑄢𑄨𑄠', 'ce': 'ори', 'chr': 'ᎣᏗᎠ', 'ckb': 'ئۆدیا', 'cs': 'urijština', 'cy': 'Odia', 'da': 'oriya', 'de': 'Oriya', 'dsb': 'orojišćina', 'dz': 'ཨོ་རི་ཡ་ཁ', 'ee': 'oriyagbe', 'el': 'Όντια', 'el-polyton': 'Όντια', 'en': 'Odia', 'eo': 'orijo', 'es': 'oriya', 'et': 'oria', 'eu': 'oriya', 'fa': 'اوریه\u200cای', 'ff-Adlm': '𞤌𞤣𞤭𞤢𞥄𞤪𞤫', 'fi': 'orija', 'fil': 'Odia', 'fo': 'odia', 'fr': 'odia', 'fur': 'oriya', 'fy': 'Odia', 'ga': 'Odia', 'gd': 'Odia', 'gl': 'odiá', 'gsw': 'Orija', 'gu': 'ઉડિયા', 'ha': 'Odiya', 'he': 'אורייה', 'hi': 'ओड़िया', 'hi-Latn': 'Odia', 'hr': 'orijski', 'hsb': 'orijšćina', 'ht': 'odia', 'hu': 'odia', 'hy': 'օրիյա', 'ia': 'oriya', 'id': 'Oriya', 'ig': 'Ọdịa', 'is': 'óría', 'it': 'odia', 'ja': 'オディア語', 'jv': 'Odia', 'ka': 'ორია', 'kea': 'odía', 'kgp': 'orija', 'kk': 'ория тілі', 'kk-Cyrl': 'ория тілі', 'km': 'អូឌៀ', 'kn': 'ಒಡಿಯ', 'ko': '오리야어', 'kok': 'उडीया', 'kok-Deva': 'उडीया', 'ks': 'اۆرِیا', 'ks-Arab': 'اۆرِیا', 'ksh': 'Oriija', 'ku': 'oriyayî', 'kxv': 'oḍiaa', 'kxv-Deva': 'उड़िया', 'kxv-Latn': 'oḍiaa', 'kxv-Orya': 'ଅଡ଼ିଆ', 'kxv-Telu': 'ఒడియా', 'ky': 'орияча', 'lb': 'Orija', 'lkt': 'Oriya Iyápi', 'lld': 'odia', 'lo': 'ໂອເດຍ', 'lrc': 'ئوریا', 'lt': 'odijų', 'lv': 'oriju', 'mai': 'उड़िया', 'mi': 'Ōtia', 'mk': 'одија', 'ml': 'ഒഡിയ', 'mn': 'ория', 'mr': 'उडिया', 'ms': 'Odia', 'mt': 'Odia', 'my': 'အိုရီရာ', 'mzn': 'اوریا', 'nb': 'odia', 'ne': 'उडिया', 'nl': 'Odia', 'nn': 'odia', 'no': 'odia', 'nqo': 'ߏߖߊߞߊ߲', 'or': 'ଓଡ଼ିଆ', 'pa': 'ਉੜੀਆ', 'pa-Guru': 'ਉੜੀਆ', 'pcm': 'Ódiá', 'pl': 'orija', 'ps': 'اوڊيا', 'pt': 'oriá', 'qu': 'Odia Simi', 'rm': 'oriya', 'ro': 'odia', 'ru': 'ория', 'rw': 'Inyoriya', 'sc': 'odia', 'sd': 'اوڊيا', 'sd-Arab': 'اوڊيا', 'si': 'ඔඩියා', 'sk': 'uríjčina', 'sl': 'odijščina', 'smn': 'orija', 'so': 'Oodhiya', 'sq': 'odisht', 'sr': 'одија', 'sr-Cyrl': 'одија', 'sr-Latn': 'odija', 'sv': 'oriya', 'sw': 'Kioriya', 'ta': 'ஒடியா', 'te': 'ఒడియా', 'tg': 'одия', 'th': 'โอดิยา', 'ti': 'ኦድያ', 'tk': 'oriýa dili', 'to': 'lea fakaʻotia', 'tr': 'Oriya dili', 'tt': 'ория', 'ug': 'ئودىياچە', 'uk': 'одія', 'ur': 'اڑیہ', 'uz': 'oriya', 'uz-Cyrl': 'одия', 'uz-Latn': 'oriya', 'vec': 'orija', 'vi': 'Tiếng Odia', 'wae': 'Oriya', 'wo': 'Oja', 'xnr': 'उड़िया', 'yo': 'Èdè Òdíà', 'yrl': 'uriá', 'yue': '歐利亞文', 'yue-Hans': '欧利亚文', 'yue-Hant': '歐利亞文', 'zh': '奥里亚语', 'zh-Hans': '奥里亚语', 'zh-Hant': '歐迪亞文', 'zu': 'isi-Odia'},
    'ora': {'en': 'Oroha'},
    'orc': {'en': 'Orma'},
    'ore': {'en': 'Orejón'},
    'org': {'en': 'Oring'},
    'orh': {'en': 'Oroqen'},
    'orn': {'en': 'Orang Kanaq'},
    'oro': {'en': 'Orokolo'},
    'orr': {'en': 'Oruma'},
    'ors': {'en': 'Orang Seletar'},
    'ort': {'en': 'Adivasi Oriya'},
    'oru': {'en': 'Ormuri'},
    'orv': {'en': 'Old Russian'},
    'orw': {'en': 'Oro Win'},
    'orx': {'en': 'Oro'},
    'ory': {'en': 'Odia (individual language)'},
    'orz': {'en': 'Ormu'},
    'os': {'af': 'Osseties', 'am': 'ኦሴቲክ', 'ar': 'الأوسيتيك', 'as': 'ওছেটিক', 'ast': 'oséticu', 'az': 'osetin', 'az-Cyrl': 'осетин', 'az-Latn': 'osetin', 'bal-Latn': 'Oséti', 'be': 'асецінская', 'be-tarask': 'асецінская', 'bg': 'осетински', 'bgn': 'آسیتینی', 'bn': 'ওসেটিক', 'br': 'oseteg', 'brx': 'अस्सेटीक', 'bs': 'osetski', 'bs-Cyrl': 'осетски', 'bs-Latn': 'osetski', 'ca': 'osseta', 'ca-ES-valencia': 'osseta', 'ccp': '𑄃𑄮𑄥𑄬𑄑𑄨𑄇𑄴', 'ce': 'хӀирийн', 'chr': 'ᎣᏎᏘᎧ', 'ckb': 'ئۆسێتی', 'cs': 'osetština', 'cy': 'Oseteg', 'da': 'ossetisk', 'de': 'Ossetisch', 'dsb': 'osetšćina', 'ee': 'ossetiagbe', 'el': 'Οσετικά', 'el-polyton': 'Οσετικά', 'en': 'Ossetic', 'es': 'osético', 'et': 'osseedi', 'eu': 'osetiera', 'fa': 'آسی', 'ff-Adlm': '𞤌𞤧𞥆𞤫𞤼𞤭𞤳𞤪𞤫', 'fi': 'osseetti', 'fil': 'Ossetic', 'fo': 'ossetiskt', 'fr': 'ossète', 'fur': 'osetic', 'fy': 'Ossetysk', 'ga': 'Oiséitis', 'gd': 'Ossetic', 'gl': 'ossetio', 'gsw': 'Ossezisch', 'gu': 'ઓસ્સેટિક', 'ha': 'Ossetic', 'he': 'אוסטית', 'hi': 'ओस्सेटिक', 'hi-Latn': 'Ossetic', 'hr': 'osetski', 'hsb': 'osetšćina', 'ht': 'ossète', 'hu': 'oszét', 'hy': 'օսերեն', 'ia': 'osseto', 'id': 'Ossetia', 'ig': 'Ossetic', 'is': 'ossetíska', 'it': 'ossetico', 'ja': 'オセット語', 'jv': 'Ossetia', 'ka': 'ოსური', 'kea': 'osétiku', 'kgp': 'oseto', 'kk': 'осетин тілі', 'kk-Cyrl': 'осетин тілі', 'km': 'អូស៊ីទិក', 'kn': 'ಒಸ್ಸೆಟಿಕ್', 'ko': '오세트어', 'kok': 'ओसेटिक', 'kok-Deva': 'ओसेटिक', 'ks': 'اۆسیٹِک', 'ks-Arab': 'اۆسیٹِک', 'ksh': 'Oßeetesch', 'ku': 'osetî', 'ky': 'осетинче', 'lb': 'Ossetesch', 'lld': 'ossetich', 'lo': 'ອອດເຊຕິກ', 'lt': 'osetinų', 'lv': 'osetīnu', 'mai': 'ओसेटिक', 'mi': 'Ōtītiki', 'mk': 'осетски', 'ml': 'ഒസ്സെറ്റിക്', 'mn': 'оссетин', 'mr': 'ओस्सेटिक', 'ms': 'Ossete', 'mt': 'Ossettiku', 'my': 'အိုဆဲတစ်ခ်', 'nb': 'ossetisk', 'ne': 'ओस्सेटिक', 'nl': 'Ossetisch', 'nn': 'ossetisk', 'no': 'ossetisk', 'nqo': 'ߏߛߍߕߌߞߊ߲', 'or': 'ଓସେଟିକ୍', 'os': 'ирон', 'pa': 'ਓਸੈਟਿਕ', 'pa-Guru': 'ਓਸੈਟਿਕ', 'pcm': 'Osẹ́tik', 'pl': 'osetyjski', 'ps': 'اوسيټک', 'pt': 'osseto', 'qu': 'Osetio Simi', 'rm': 'ossetic', 'ro': 'osetă', 'ru': 'осетинский', 'sc': 'ossèticu', 'sd': 'اوسيٽڪ', 'sd-Arab': 'اوسيٽڪ', 'si': 'ඔසිටෙක්', 'sk': 'osetčina', 'sl': 'osetinščina', 'smn': 'ossetkielâ', 'so': 'Oseetic', 'sq': 'osetisht', 'sr': 'осетински', 'sr-Cyrl': 'осетински', 'sr-Latn': 'osetinski', 'sv': 'ossetiska', 'sw': 'Kiosetia', 'ta': 'ஒசெட்டிக்', 'te': 'ఒసేటిక్', 'th': 'ออสเซเตีย', 'ti': 'ኦሰትኛ', 'tk': 'osetin dili', 'to': 'lea fakaʻosetiki', 'tr': 'Osetçe', 'ug': 'ئوسسېتچەچە', 'uk': 'осетинська', 'ur': 'اوسیٹک', 'uz': 'osetin', 'uz-Latn': 'osetin', 'vec': 'osètego', 'vi': 'Tiếng Ossetic', 'wae': 'Osétiš', 'xnr': 'ओस्सेटिक', 'yi': 'אסעטיש', 'yo': 'Ọṣẹ́tíìkì', 'yrl': 'usetu', 'yue': '奧塞提文', 'yue-Hans': '奥塞提文', 'yue-Hant': '奧塞提文', 'zh': '奥塞梯语', 'zh-Hans': '奥塞梯语', 'zh-Hant': '奧塞提文', 'zu': 'isi-Ossetic'},
    'osa': {'ar': 'الأوساج', 'ast': 'osage', 'az': 'osage', 'az-Latn': 'osage', 'bal-Latn': 'Oságá', 'bg': 'осейджи', 'bn': 'ওসেজ', 'br': 'osage', 'brx': 'ओसेज', 'bs': 'osage', 'bs-Cyrl': 'осаге', 'bs-Latn': 'osage', 'ca': 'osage', 'ca-ES-valencia': 'osage', 'ccp': '𑄃𑄮𑄥𑄬𑄌𑄴', 'cs': 'osage', 'cy': 'Osageg', 'da': 'osage', 'de': 'Osage', 'el': 'Οσάζ', 'el-polyton': 'Οσάζ', 'en': 'Osage', 'es': 'osage', 'et': 'oseidži', 'fa': 'اوسیجی', 'fi': 'osage', 'fr': 'osage', 'fy': 'Osage', 'gd': 'Osage', 'gsw': 'Osage-Schpraach', 'gu': 'ઓસેજ', 'he': 'אוסג׳', 'hi': 'ओसेज', 'hi-Latn': 'Osage', 'hr': 'osage', 'ht': 'osage', 'hu': 'osage', 'hy': 'օսեյջ', 'ia': 'osage', 'id': 'Osage', 'ig': 'Osage', 'is': 'ósage', 'it': 'osage', 'ja': 'オセージ語', 'kgp': 'osage', 'kk': 'осейдж тілі', 'kk-Cyrl': 'осейдж тілі', 'kn': 'ಓಸಾಜ್', 'ko': '오세이지어', 'ks': 'اۆسیج', 'ks-Arab': 'اۆسیج', 'ku': 'osageyî', 'lb': 'Osage-Sprooch', 'lo': 'ໂອແຊກ', 'lt': 'osage', 'lv': 'važāžu', 'mk': 'осашки', 'ml': 'ഒസേജ്', 'mr': 'ओसेज', 'mt': 'Osaġjan', 'nb': 'osage', 'ne': 'ओसागे', 'nl': 'Osage', 'nn': 'osage', 'no': 'osage', 'or': 'ୱୌସେଜ୍', 'osa': '𐓏𐓘𐓻𐓘𐓻𐓟', 'pl': 'osage', 'pt': 'osage', 'rm': 'osage', 'ro': 'osage', 'ru': 'оседжи', 'sk': 'osedžština', 'sl': 'osage', 'sr': 'осаге', 'sr-Cyrl': 'осаге', 'sr-Latn': 'osage', 'sv': 'osage', 'ta': 'ஓசேஜ்', 'te': 'ఒసాజ్', 'th': 'โอซากี', 'ti': 'ኦሳጌ', 'to': 'lea fakaʻosēse', 'tr': 'Osage', 'ug': 'ئوساگېلارچە', 'uk': 'осейдж', 'vec': 'osage', 'vi': 'Tiếng Osage', 'xnr': 'ओसेज', 'yrl': 'usayi', 'yue': '歐塞奇文', 'yue-Hans': '欧塞奇文', 'yue-Hant': '歐塞奇文', 'zh': '欧塞奇语', 'zh-Hans': '欧塞奇语', 'zh-Hant': '歐塞奇文'},
    'osc': {'en': 'Oscan'},
    'osi': {'en': 'Osing'},
    'osn': {'en': 'Old Sundanese'},
    'oso': {'en': 'Ososo'},
    'osp': {'en': 'Old Spanish'},
    'ost': {'en': 'Osatu'},
    'osu': {'en': 'Southern One'},
    'osx': {'en': 'Old Saxon'},
    'ota': {'ar': 'التركية العثمانية', 'ast': 'turcu otomanu', 'az': 'osman', 'az-Latn': 'osman', 'bg': 'отомански турски', 'bn': 'অটোমান তুর্কি', 'br': 'turkeg otoman', 'brx': 'तुर्की ओटोमान', 'bs': 'osmanski turski', 'bs-Cyrl': 'отомански турски', 'bs-Latn': 'osmanski turski', 'ca': 'turc otomà', 'ca-ES-valencia': 'turc otomà', 'ccp': '𑄃𑄧𑄑𑄮𑄟𑄚𑄴 𑄖𑄪𑄢𑄴𑄇𑄨', 'cs': 'turečtina (osmanská)', 'cy': 'Tyrceg Otoman', 'da': 'osmannisk tyrkisk', 'de': 'Osmanisch', 'el': 'Οθωμανικά Τουρκικά', 'el-polyton': 'Τουρκικά, ὀθωμανικὰ', 'en': 'Ottoman Turkish', 'es': 'turco otomano', 'et': 'osmanitürgi', 'fa': 'ترکی عثمانی', 'fi': 'osmani', 'fr': 'turc ottoman', 'fur': 'turc otoman', 'fy': 'Ottomaansk-Turks', 'gd': 'Turcais Otomanach', 'gsw': 'Osmanisch', 'gu': 'ઓટોમાન તુર્કિશ', 'he': 'טורקית עות׳מנית', 'hi': 'ओटोमान तुर्किश', 'hi-Latn': 'Ottoman Turkish', 'hr': 'turski - otomanski', 'ht': 'turc ottoman', 'hu': 'oszmán-török', 'hy': 'օսմաներեն', 'id': 'Turki Osmani', 'is': 'tyrkneska, ottóman', 'it': 'turco ottomano', 'ja': 'オスマントルコ語', 'kgp': 'turko otomỹnũ', 'kn': 'ಒಟ್ಟೋಮನ್ ತುರ್ಕಿಷ್', 'ko': '오스만 터키어', 'ks': 'اوٹومَن تُرکِش', 'ks-Arab': 'اوٹومَن تُرکِش', 'lb': 'Osmanesch', 'lo': 'ຕູກີອອດໂຕມັນ', 'lt': 'osmanų turkų', 'lv': 'turku osmaņu', 'mk': 'отомански турски', 'ml': 'ഓട്ടോമൻ തുർക്കിഷ്', 'mr': 'ओटोमान तुर्किश', 'mt': 'Tork Ottoman', 'nb': 'ottomansk tyrkisk', 'ne': 'अटोमन तुर्की', 'nl': 'Ottomaans-Turks', 'nn': 'ottomansk tyrkisk', 'no': 'ottomansk tyrkisk', 'or': 'ଓଟ୍ଟୋମନ୍ ତୁର୍କିସ୍', 'pl': 'osmańsko-turecki', 'pt': 'turco otomano', 'rm': 'tirc ottoman', 'ro': 'turcă otomană', 'ru': 'старотурецкий', 'sk': 'osmanská turečtina', 'sl': 'otomanska turščina', 'sr': 'османски турски', 'sr-Cyrl': 'османски турски', 'sr-Latn': 'osmanski turski', 'sv': 'ottomanska', 'ta': 'ஓட்டோமான் துருக்கிஷ்', 'te': 'ఒట్టోమన్ టర్కిష్', 'th': 'ตุรกีออตโตมัน', 'to': 'lea fakatoake-ʻotomani', 'tr': 'Osmanlı Türkçesi', 'ug': 'ئوسمان تۈركچە', 'uk': 'османська', 'vi': 'Tiếng Thổ Nhĩ Kỳ Ottoman', 'yrl': 'turiku utumanu', 'yue': '鄂圖曼土耳其文', 'yue-Hans': '鄂图曼土耳其文', 'yue-Hant': '鄂圖曼土耳其文', 'zh': '奥斯曼土耳其语', 'zh-Hans': '奥斯曼土耳其语', 'zh-Hant': '鄂圖曼土耳其文'},
    'otb': {'en': 'Old Tibetan'},
    'otd': {'en': 'Ot Danum'},
    'ote': {'en': 'Mezquital Otomi'},
    'oti': {'en': 'Oti'},
    'otk': {'en': 'Old Turkish'},
    'otl': {'en': 'Tilapa Otomi'},
    'otm': {'en': 'Eastern Highland Otomi'},
    'otn': {'en': 'Tenango Otomi'},
    'oto': {'en': 'Otomian languages'},
    'otq': {'en': 'Querétaro Otomi'},
    'otr': {'en': 'Otoro'},
    'ots': {'en': 'Estado de México Otomi'},
    'ott': {'en': 'Temoaya Otomi'},
    'otu': {'en': 'Otuke'},
    'otw': {'en': 'Ottawa'},
    'otx': {'en': 'Texcatepec Otomi'},
    'oty': {'en': 'Old Tamil'},
    'otz': {'en': 'Ixtenco Otomi'},
    'oua': {'en': 'Tagargrent'},
    'oub': {'en': 'Glio-Oubi'},
    'oue': {'en': 'Oune'},
    'oui': {'en': 'Old Uighur'},
    'oum': {'en': 'Ouma'},
    'oun': {'en': 'ǃOǃung'},
    'ovd': {'en': 'Elfdalian'},
    'owi': {'en': 'Owiniga'},
    'owl': {'en': 'Old Welsh'},
    'oyb': {'en': 'Oy'},
    'oyd': {'en': 'Oyda'},
    'oym': {'en': 'Wayampi'},
    'oyy': {'en': "Oya'oya"},
    'ozm': {'en': 'Koonzime'},
    'pa': {'af': 'Pandjabi', 'agq': 'Kpuwndzabì', 'ak': 'Pungyabi kasa', 'am': 'ፑንጃብኛ', 'ar': 'البنجابية', 'as': 'পাঞ্জাবী', 'asa': 'Kipunjabi', 'ast': 'punyabí', 'az': 'pəncab', 'az-Arab': 'پنجابجا', 'az-Cyrl': 'пәнҹаб', 'az-Latn': 'pəncab', 'bal-Latn': 'Panjábi', 'bas': 'Hɔp u pɛnjàbi', 'be': 'панджабі', 'be-tarask': 'панджабі', 'bem': 'Ichi Punjabi', 'bez': 'Hipunjabi', 'bg': 'пенджабски', 'bgn': 'پنجاپی', 'bm': 'pɛnijabikan', 'bn': 'পাঞ্জাবী', 'br': 'punjabi', 'brx': 'पान्जाबी', 'bs': 'pandžapski', 'bs-Cyrl': 'пенџапски', 'bs-Latn': 'pandžapski', 'ca': 'panjabi', 'ca-ES-valencia': 'panjabi', 'ccp': '𑄛𑄚𑄴𑄎𑄝𑄩', 'ce': 'панджаби', 'cgg': 'Orupungyabi', 'chr': 'ᏡᏂᏣᏈ', 'ckb': 'پەنجابی', 'cs': 'paňdžábština', 'cy': 'Pwnjabeg', 'da': 'punjabi', 'dav': 'Kipunjabi', 'de': 'Punjabi', 'dje': 'Punjaabi sennii', 'dsb': 'pandžabšćina', 'dyo': 'penjabi', 'dz': 'པཱན་ཇ་བི་ཁ', 'ebu': 'Kĩpunjabi', 'ee': 'pundzabgbe', 'el': 'Παντζαπικά', 'el-polyton': 'Παντζαπικά', 'en': 'Punjabi', 'eo': 'panĝaba', 'es': 'punyabí', 'et': 'pandžabi', 'eu': 'punjabera', 'ewo': 'ǹkɔ́bɔ funəhábia', 'fa': 'پنجابی', 'ff': 'Punjabeere', 'ff-Adlm': '𞤆𞤵𞤲𞤶𞤢𞥄𞤦𞤭𞥅𞤪𞤫', 'ff-Latn': 'Punjabeere', 'fi': 'pandžabi', 'fil': 'Punjabi', 'fo': 'punjabi', 'fr': 'pendjabi', 'fur': 'punjabi', 'fy': 'Punjabi', 'ga': 'Puinseáibis', 'gd': 'Panjabi', 'gl': 'panxabí', 'gsw': 'Pandschabisch', 'gu': 'પંજાબી', 'guz': 'Kipunjabi', 'ha': 'Punjabi', 'he': 'פנג׳אבי', 'hi': 'पंजाबी', 'hi-Latn': 'Punjabi', 'hr': 'pandžapski', 'hsb': 'pandźabšćina', 'ht': 'pendjabi', 'hu': 'pandzsábi', 'hy': 'փենջաբերեն', 'ia': 'punjabi', 'id': 'Punjabi', 'ig': 'Punjabi', 'is': 'púnjabí', 'it': 'punjabi', 'ja': 'パンジャブ語', 'jmc': 'Kyipunjabi', 'jv': 'Punjab', 'ka': 'პენჯაბური', 'kab': 'Tapunjabit', 'kam': 'Kipunjabi', 'kde': 'Chipunjabi', 'kea': 'pandjabi', 'kgp': 'pỹjami', 'khq': 'Punjaabi sennii', 'ki': 'Kipunjabi', 'kk': 'пенджаб тілі', 'kk-Cyrl': 'пенджаб тілі', 'kln': 'kutitab Punjab', 'km': 'បឹនជាពិ', 'kn': 'ಪಂಜಾಬಿ', 'ko': '펀잡어', 'kok': 'पंजाबी', 'kok-Deva': 'पंजाबी', 'ks': 'پَنجٲبؠ', 'ks-Arab': 'پَنجٲبؠ', 'ksb': 'Kipunjabi', 'ksf': 'ripɛnjabí', 'ksh': 'Panschaabesch', 'ku': 'puncabî', 'kxv': 'pnjabī', 'kxv-Deva': 'पंजाबी', 'kxv-Latn': 'pnjabī', 'kxv-Orya': 'ପଞ୍ଜାବୀ', 'kxv-Telu': 'పంజాబి', 'ky': 'пунжабиче', 'lag': 'Kɨpúnjabi', 'lb': 'Pandschabesch', 'lg': 'Lupunjabi', 'lkt': 'Punjabi Iyápi', 'lld': 'punjabi', 'ln': 'lipendzabi', 'lo': 'ປັນຈາບີ', 'lrc': 'پأنجابی', 'lt': 'pendžabų', 'lu': 'Lipunjabi', 'luo': 'Kipunjabi', 'luy': 'Kipunjabi', 'lv': 'pandžabu', 'mai': 'पंजाबी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lpunjabi', 'mer': 'Kĩpunjabu', 'mfe': 'penjabi', 'mg': 'Penjabi', 'mgh': 'Ipunjabi', 'mi': 'Punutapi', 'mk': 'пенџапски', 'ml': 'പഞ്ചാബി', 'mn': 'панжаби', 'mr': 'पंजाबी', 'ms': 'Punjabi', 'mt': 'Punjabi', 'mua': 'Pǝnjabi', 'my': 'ပန်ချာပီ', 'mzn': 'پنجابی', 'naq': 'Punjabigowab', 'nb': 'panjabi', 'nd': 'isi-Phunjabi', 'ne': 'पंजाबी', 'nl': 'Punjabi', 'nmg': 'Kiɛl pɛndjabi', 'nn': 'panjabi', 'no': 'panjabi', 'nqo': 'ߔߍ߲ߖߊߓߌߞߊ߲', 'nus': 'Thok puɔnjabani', 'nyn': 'Orupungyabi', 'om': 'Afaan Punjabii', 'or': 'ପଞ୍ଜାବୀ', 'pa': 'ਪੰਜਾਬੀ', 'pa-Arab': 'پنجابی', 'pa-Guru': 'ਪੰਜਾਬੀ', 'pcm': 'Punjábi', 'pl': 'pendżabski', 'ps': 'پنجابي', 'pt': 'panjabi', 'qu': 'Punyabi Simi', 'rm': 'punjabi', 'rn': 'Igipunjabi', 'ro': 'punjabi', 'rof': 'Kipunjabi', 'ru': 'панджаби', 'rw': 'Igipunjabi', 'rwk': 'Kyipunjabi', 'sah': 'Пандьаабтыы', 'saq': 'Kipunjabi', 'sbp': 'Ishipunjabi', 'sc': 'punjabi', 'sd': 'پنجابي', 'sd-Arab': 'پنجابي', 'se': 'panjabigiella', 'seh': 'panjabi', 'ses': 'Punjaabi sennii', 'sg': 'Penzäbï', 'shi': 'ⵜⴰⴱⵏⵊⴰⴱⵉⵜ', 'shi-Latn': 'Tabnjabit', 'shi-Tfng': 'ⵜⴰⴱⵏⵊⴰⴱⵉⵜ', 'si': 'පන්ජාබි', 'sk': 'pandžábčina', 'sl': 'pandžabščina', 'smn': 'pandžabi', 'sn': 'chiPunjabi', 'so': 'Bunjaabi', 'sq': 'punxhabisht', 'sr': 'пенџапски', 'sr-Cyrl': 'пенџапски', 'sr-Latn': 'pendžapski', 'sv': 'punjabi', 'sw': 'Kipunjabi', 'ta': 'பஞ்சாபி', 'te': 'పంజాబీ', 'teo': 'Kipunjabi', 'tg': 'панҷобӣ', 'th': 'ปัญจาบ', 'ti': 'ፑንጃቢ', 'tk': 'penjab dili', 'to': 'lea fakapūnusapi', 'tr': 'Pencapça', 'tt': 'пәнҗаби', 'twq': 'Punjaabi sennii', 'tzm': 'Tabenjabit', 'ug': 'پەنجابچە', 'uk': 'панджабі', 'ur': 'پنجابی', 'uz': 'panjobcha', 'uz-Cyrl': 'панжобча', 'uz-Latn': 'panjobcha', 'vai': 'ꖛꕨꔬ', 'vai-Latn': 'Puŋjabhi', 'vai-Vaii': 'ꖛꕨꔬ', 'vec': 'punjabi', 'vi': 'Tiếng Punjab', 'vun': 'Kyipunjabi', 'wae': 'Pandšabiš', 'wo': 'Punjabi', 'xnr': 'पंजाबी', 'xog': 'Olupunjabi', 'yav': 'nupunsapíɛ́', 'yo': 'Èdè Punjabi', 'yrl': 'pãyabi', 'yue': '旁遮普文', 'yue-Hans': '旁遮普文', 'yue-Hant': '旁遮普文', 'zgh': 'ⵜⴰⴱⵏⵊⴰⴱⵉⵜ', 'zh': '旁遮普语', 'zh-Hans': '旁遮普语', 'zh-Hant': '旁遮普文', 'zu': 'isi-Punjabi'},
    'paa': {'en': 'Papuan languages'},
    'pab': {'en': 'Parecís'},
    'pac': {'en': 'Pacoh'},
    'pad': {'en': 'Paumarí'},
    'pae': {'en': 'Pagibete'},
    'paf': {'en': 'Paranawát'},
    'pag': {'af': 'Pangasinan', 'am': 'ፓንጋሲናን', 'ar': 'البانجاسينان', 'as': 'পংগাছিনান', 'ast': 'pangasinan', 'az': 'panqasinan', 'az-Cyrl': 'пангасинан', 'az-Latn': 'panqasinan', 'be': 'пангасінан', 'be-tarask': 'пангасінан', 'bg': 'пангасинан', 'bgn': 'پانگاسینانی', 'bn': 'পাঙ্গাসিনান', 'br': 'pangasinan', 'brx': 'पाङासिनान', 'bs': 'pangasinski', 'bs-Cyrl': 'пангасински', 'bs-Latn': 'pangasinski', 'ca': 'pangasinan', 'ca-ES-valencia': 'pangasinan', 'ccp': '𑄛𑄁𑄉𑄥𑄨𑄚𑄚𑄴', 'ce': 'пангасинан', 'chr': 'ᏇᎦᏏᎠᏂ', 'ckb': 'پانگاسینان', 'cs': 'pangasinanština', 'cy': 'Pangasineg', 'da': 'pangasinan', 'de': 'Pangasinan', 'dsb': 'pangasinanšćina', 'el': 'Πανγκασινάν', 'el-polyton': 'Πανγκασινάν', 'en': 'Pangasinan', 'es': 'pangasinán', 'et': 'pangasinani', 'eu': 'pangasinanera', 'fa': 'پانگاسینانی', 'ff-Adlm': '𞤆𞤢𞤲𞤺𞤢𞤧𞤭𞤲𞤢𞤲𞤪𞤫', 'fi': 'pangasinan', 'fil': 'Pangasinan', 'fo': 'pangasinan', 'fr': 'pangasinan', 'fy': 'Pangasinan', 'ga': 'Pangasaíneánais', 'gd': 'Pangasinan', 'gl': 'pangasinan', 'gsw': 'Pangasinanisch', 'gu': 'પંગાસીનાન', 'ha': 'Pangasinanci', 'he': 'פנגסינאן', 'hi': 'पंगासीनान', 'hi-Latn': 'Pangasinan', 'hr': 'pangasinan', 'hsb': 'pangasinanšćina', 'ht': 'pangasinan', 'hu': 'pangaszinan', 'hy': 'պանգասինաներեն', 'ia': 'pangasinan', 'id': 'Pangasina', 'ig': 'Pangasinan', 'is': 'pangasínmál', 'it': 'pangasinan', 'ja': 'パンガシナン語', 'jv': 'Pangasinan', 'ka': 'პანგასინანი', 'kgp': 'pangasinỹ', 'kk': 'пангасинан тілі', 'kk-Cyrl': 'пангасинан тілі', 'km': 'ភេនហ្គាស៊ីណាន', 'kn': 'ಪಂಗಾಸಿನನ್', 'ko': '판가시난어', 'kok': 'पांगासियनान', 'kok-Deva': 'पांगासियनान', 'ks': 'پَنگاسِنَن', 'ks-Arab': 'پَنگاسِنَن', 'ku': 'pangasînanî', 'ky': 'пангасиче', 'lb': 'Pangasinan-Sprooch', 'lo': 'ປານກາຊີນານ', 'lt': 'pangasinanų', 'lv': 'pangasinanu', 'mai': 'पैंगासिनान', 'mi': 'Pāngahina', 'mk': 'пангасинански', 'ml': 'പങ്കാസിനൻ', 'mn': 'пангасин', 'mr': 'पंगासीनान', 'ms': 'Pangasinan', 'mt': 'Pangasinjan', 'my': 'ပန်ဂါစီနန်', 'nb': 'pangasinan', 'ne': 'पाङ्गासिनान', 'nl': 'Pangasinan', 'nn': 'pangasinan', 'no': 'pangasinan', 'or': 'ପାଙ୍ଗାସିନିଆନ୍', 'pa': 'ਪੰਗਾਸੀਨਾਨ', 'pa-Guru': 'ਪੰਗਾਸੀਨਾਨ', 'pcm': 'Pangasínán Lángwej', 'pl': 'pangasinan', 'ps': 'پانګاسین', 'pt': 'pangasinã', 'qu': 'Pangasinan Simi', 'rm': 'pangasinan', 'ro': 'pangasinan', 'ru': 'пангасинан', 'sc': 'pangasinan', 'sd': 'پانگا سينان', 'sd-Arab': 'پانگا سينان', 'si': 'පන්ගසීනන්', 'sk': 'pangasinančina', 'sl': 'pangasinanščina', 'smn': 'pangasinankielâ', 'so': 'Pangasinan', 'sq': 'pangasinanisht', 'sr': 'пангасинански', 'sr-Cyrl': 'пангасинански', 'sr-Latn': 'pangasinanski', 'sv': 'pangasinan', 'sw': 'Kipangasinan', 'ta': 'பன்காசினன்', 'te': 'పంగాసినాన్', 'th': 'ปังกาซีนัน', 'ti': 'ፓንጋሲናን', 'tk': 'pangansinan dili', 'to': 'lea fakapangasinani', 'tr': 'Pangasinan dili', 'ug': 'پانگاسىنانچە', 'uk': 'пангасінанська', 'ur': 'پنگاسنان', 'uz': 'pangasinan', 'uz-Latn': 'pangasinan', 'vec': 'pangasinan', 'vi': 'Tiếng Pangasinan', 'xnr': 'पंगासीनान', 'yo': 'Èdè Pangasina', 'yrl': 'pãgasinã', 'yue': '潘加辛文', 'yue-Hans': '潘加辛文', 'yue-Hant': '潘加辛文', 'zh': '邦阿西南语', 'zh-Hans': '邦阿西南语', 'zh-Hant': '潘加辛文', 'zu': 'isi-Pangasinan'},
    'pah': {'en': 'Tenharim'},
    'pai': {'en': 'Pe'},
    'pak': {'en': 'Parakanã'},
    'pal': {'ar': 'البهلوية', 'ast': 'pahlavi', 'az': 'pəhləvi', 'az-Latn': 'pəhləvi', 'bg': 'пахлави', 'bn': 'পাহ্লাভি', 'br': 'pahlavi', 'brx': 'पहलवी', 'bs': 'pahlavi', 'bs-Cyrl': 'пахлави', 'bs-Latn': 'pahlavi', 'ca': 'pahlavi', 'ca-ES-valencia': 'pahlavi', 'ccp': '𑄛𑄦𑄳𑄣𑄞𑄨', 'cs': 'pahlavština', 'cy': 'Pahlafi', 'da': 'pahlavi', 'de': 'Mittelpersisch', 'el': 'Παχλάβι', 'el-polyton': 'Παχλάβι', 'en': 'Pahlavi', 'es': 'pahlavi', 'et': 'pahlavi', 'fa': 'پهلوی', 'fi': 'pahlavi', 'fr': 'pahlavi', 'fy': 'Pahlavi', 'gd': 'Pahlavi', 'gsw': 'Mittelpersisch', 'gu': 'પહલવી', 'he': 'פלאבי', 'hi': 'पाह्लावी', 'hi-Latn': 'Pahlavi', 'hr': 'pahlavi', 'ht': 'pahlavi', 'hu': 'pahlavi', 'hy': 'պահլավերեն', 'id': 'Pahlevi', 'is': 'palaví', 'it': 'pahlavi', 'ja': 'パフラヴィー語', 'ka': 'ფალაური', 'kgp': 'paravi', 'kn': 'ಪಹ್ಲವಿ', 'ko': '팔레비어', 'ks': 'پَہلَوی', 'ks-Arab': 'پَہلَوی', 'lb': 'Mëttelpersesch', 'lo': 'ພາລາວີ', 'lt': 'vidurinė persų kalba', 'lv': 'pehlevi', 'mk': 'средноперсиски', 'ml': 'പാഹ്ലവി', 'mr': 'पहलवी', 'mt': 'Pahlavi', 'nb': 'pahlavi', 'ne': 'पाहलावी', 'nl': 'Pahlavi', 'nn': 'pahlavi', 'no': 'pahlavi', 'or': 'ପାହ୍ଲାଭି', 'pl': 'pahlavi', 'pt': 'pálavi', 'rm': 'pahlavi', 'ro': 'pahlavi', 'ru': 'пехлевийский', 'sk': 'pahlaví', 'sr': 'пахлави', 'sr-Cyrl': 'пахлави', 'sr-Latn': 'pahlavi', 'sv': 'medelpersiska', 'ta': 'பாஹ்லவி', 'te': 'పహ్లావి', 'th': 'ปะห์ลาวี', 'to': 'lea fakapālavi', 'tr': 'Pehlevi Dili', 'ug': 'پەھلەۋىچە', 'uk': 'пехлеві', 'vi': 'Tiếng Pahlavi', 'yrl': 'parawi', 'yue': '巴列維文', 'yue-Hans': '巴列维文', 'yue-Hant': '巴列維文', 'zh': '巴拉维语', 'zh-Hans': '巴拉维语', 'zh-Hant': '巴列維文'},
    'pam': {'af': 'Pampanga', 'am': 'ፓምፓንጋ', 'ar': 'البامبانجا', 'as': 'পাম্পান্গা', 'ast': 'pampanga', 'az': 'pampanqa', 'az-Cyrl': 'пампанга', 'az-Latn': 'pampanqa', 'be': 'пампанга', 'be-tarask': 'пампанга', 'bg': 'пампанга', 'bgn': 'پامپانگی', 'bn': 'পাম্পাঙ্গা', 'br': 'pampanga', 'brx': 'पाम्पाङा', 'bs': 'pampanga', 'bs-Cyrl': 'пампанга', 'bs-Latn': 'pampanga', 'ca': 'pampanga', 'ca-ES-valencia': 'pampanga', 'ccp': '𑄛𑄟𑄴𑄛𑄋𑄴𑄉', 'ce': 'пампанга', 'chr': 'ᏆᎹᏆᎾᎦ', 'ckb': 'پامپانگا', 'cs': 'papangau', 'cy': 'Pampanga', 'da': 'pampanga', 'de': 'Pampanggan', 'dsb': 'pampangašćina', 'el': 'Παμπάνγκα', 'el-polyton': 'Παμπάνγκα', 'en': 'Pampanga', 'es': 'pampanga', 'et': 'pampanga', 'eu': 'pampangera', 'fa': 'پامپانگایی', 'ff-Adlm': '𞤆𞤢𞤥𞤨𞤢𞤲𞤺𞤢𞥄𞤪𞤫', 'fi': 'pampanga', 'fil': 'Pampanga', 'fo': 'pampanga', 'fr': 'pampangan', 'fy': 'Pampanga', 'ga': 'Pampaingis', 'gd': 'Pampanga', 'gl': 'pampanga', 'gsw': 'Pampanggan-Schpraach', 'gu': 'પમ્પાન્ગા', 'ha': 'Pampanga', 'he': 'פמפאניה', 'hi': 'पाम्पान्गा', 'hi-Latn': 'Pampanga', 'hr': 'pampanga', 'hsb': 'pampangašćina', 'ht': 'pampangan', 'hu': 'pampangan', 'hy': 'պամպանգաերեն', 'ia': 'pampanga', 'id': 'Pampanga', 'ig': 'Pampanga', 'is': 'pampanga', 'it': 'pampanga', 'ja': 'パンパンガ語', 'jv': 'Pampanga', 'ka': 'პამპანგა', 'kgp': 'pampỹga', 'kk': 'пампанга тілі', 'kk-Cyrl': 'пампанга тілі', 'km': 'ផាមភេនហ្គា', 'kn': 'ಪಂಪಾಂಗಾ', 'ko': '팜팡가어', 'kok': 'पांपान्गा', 'kok-Deva': 'पांपान्गा', 'ks': 'پَمپَنگا', 'ks-Arab': 'پَمپَنگا', 'ku': 'kapampanganî', 'ky': 'пампангача', 'lb': 'Pampanggan-Sprooch', 'lo': 'ປາມປານກາ', 'lt': 'pampangų', 'lv': 'pampanganu', 'mai': 'पैमपंगा', 'mi': 'Pamapaka', 'mk': 'пампанга', 'ml': 'പാംപൻഗ', 'mn': 'пампанга', 'mr': 'पाम्पान्गा', 'ms': 'Pampanga', 'mt': 'Pampanga', 'my': 'ပမ်ပန်ညာ', 'nb': 'pampanga', 'ne': 'पामपाङ्गा', 'nl': 'Pampanga', 'nn': 'pampanga', 'no': 'pampanga', 'or': 'ପାମ୍ପାଙ୍ଗା', 'pa': 'ਪੈਂਪਾਂਗਾ', 'pa-Guru': 'ਪੈਂਪਾਂਗਾ', 'pcm': 'Pampánga Lángwej', 'pl': 'pampango', 'ps': 'پمپانگا', 'pt': 'pampanga', 'qu': 'Pampanga Simi', 'rm': 'pampanga', 'ro': 'pampanga', 'ru': 'пампанга', 'sc': 'pampanga', 'sd': 'پيم پينگا', 'sd-Arab': 'پيم پينگا', 'si': 'පන්පන්ග', 'sk': 'kapampangančina', 'sl': 'pampanščina', 'smn': 'pampangakielâ', 'so': 'Pampanga', 'sq': 'pampangaisht', 'sr': 'пампанга', 'sr-Cyrl': 'пампанга', 'sr-Latn': 'pampanga', 'sv': 'pampanga', 'sw': 'Kipampanga', 'ta': 'பம்பாங்கா', 'te': 'పంపన్గా', 'th': 'ปัมปังกา', 'ti': 'ፓምፓንጋ', 'tk': 'kapampangan dili', 'to': 'lea fakapamipanga', 'tr': 'Pampanga', 'ug': 'پامپانگاچە', 'uk': 'пампанга', 'ur': 'پامپنگا', 'uz': 'pampanga', 'uz-Latn': 'pampanga', 'vec': 'panpanga', 'vi': 'Tiếng Pampanga', 'xnr': 'पाम्पान्गा', 'yo': 'Èdè Pampanga', 'yrl': 'pãpãga', 'yue': '潘帕嘉文', 'yue-Hans': '潘帕嘉文', 'yue-Hant': '潘帕嘉文', 'zh': '邦板牙语', 'zh-Hans': '邦板牙语', 'zh-Hant': '潘帕嘉文', 'zu': 'isi-Pampanga'},
    'pao': {'en': 'Northern Paiute'},
    'pap': {'af': 'Papiamento', 'am': 'ፓፒያሜንቶ', 'ar': 'البابيامينتو', 'as': 'পাপিয়ামেণ্টো', 'ast': 'papiamento', 'az': 'papyamento', 'az-Cyrl': 'папјаменто', 'az-Latn': 'papyamento', 'bal-Latn': 'Pápiámentó', 'be': 'пап’яменту', 'be-tarask': 'пап’яменту', 'bg': 'папиаменто', 'bgn': 'پاپیامنتوی', 'bn': 'পাপিয়ামেন্টো', 'br': 'papiamento', 'brx': 'पापीआमेन्तो', 'bs': 'papiamento', 'bs-Cyrl': 'папиаменто', 'bs-Latn': 'papiamento', 'ca': 'papiament', 'ca-ES-valencia': 'papiament', 'ccp': '𑄛𑄛𑄨𑄠𑄟𑄬𑄚𑄴𑄖𑄮', 'ce': 'папьяменто', 'chr': 'ᏆᏈᏯᎺᎾᏙ', 'ckb': 'پاپیامێنتۆ', 'cs': 'papiamento', 'cy': 'Papiamento', 'da': 'papiamento', 'de': 'Papiamento', 'dsb': 'papiamentošćina', 'el': 'Παπιαμέντο', 'el-polyton': 'Παπιαμέντο', 'en': 'Papiamento', 'es': 'papiamento', 'et': 'papiamento', 'eu': 'papiamentoa', 'fa': 'پاپیامنتو', 'ff-Adlm': '𞤆𞤢𞤨𞤭𞤢𞤥𞤫𞤲𞤼𞤮𞥅𞤪𞤫', 'fi': 'papiamentu', 'fil': 'Papiamento', 'fo': 'papiamento', 'fr': 'papiamento', 'fur': 'papiamento', 'fy': 'Papiaments', 'ga': 'Paipeamaintis', 'gd': 'Papiamentu', 'gl': 'papiamento', 'gsw': 'Papiamento', 'gu': 'પાપિયામેન્ટો', 'ha': 'Papiamento', 'he': 'פפיאמנטו', 'hi': 'पापियामेन्टो', 'hi-Latn': 'Papiamento', 'hr': 'papiamento', 'hsb': 'papiamentošćina', 'ht': 'papiamento', 'hu': 'papiamento', 'hy': 'պապյամենտո', 'ia': 'papiamento', 'id': 'Papiamento', 'ig': 'Papiamento', 'is': 'papíamentó', 'it': 'papiamento', 'ja': 'パピアメント語', 'jv': 'Papiamento', 'ka': 'პაპიამენტო', 'kgp': 'papijamẽto', 'kk': 'папьяменто тілі', 'kk-Cyrl': 'папьяменто тілі', 'km': 'ប៉ាប៉ៃមេនតូ', 'kn': 'ಪಪಿಯಾಮೆಂಟೊ', 'ko': '파피아먼토어', 'kok': 'पाप्यामेंटो', 'kok-Deva': 'पाप्यामेंटो', 'ks': 'پَپِیامینٹو', 'ks-Arab': 'پَپِیامینٹو', 'ku': 'papyamentoyî', 'ky': 'папиаменточо', 'lb': 'Papiamento', 'lo': 'ປາມເປຍເມັນໂທ', 'lt': 'papiamento', 'lv': 'papjamento', 'mai': 'पैपियामेंटो', 'mi': 'Papiamēto', 'mk': 'папијаменто', 'ml': 'പാപിയാമെന്റൊ', 'mn': 'папьяменто', 'mr': 'पापियामेन्टो', 'ms': 'Papiamento', 'mt': 'Papiamento', 'my': 'ပါပီမင်တို', 'nb': 'papiamento', 'ne': 'पापियामेन्तो', 'nl': 'Papiaments', 'nn': 'papiamento', 'no': 'papiamento', 'nqo': 'ߔߊߔߌߦߊߡߊ߲ߕߏߞߊ߲', 'or': 'ପାପିଆମେଣ୍ଟୋ', 'pa': 'ਪਾਪਿਆਮੈਂਟੋ', 'pa-Guru': 'ਪਾਪਿਆਮੈਂਟੋ', 'pap': 'Papiamentu', 'pcm': 'Papiaménto Lángwej', 'pl': 'papiamento', 'ps': 'پاپيامينتو', 'pt': 'papiamento', 'qu': 'Papiamento Simi', 'rm': 'papiamento', 'ro': 'papiamento', 'ru': 'папьяменто', 'sc': 'papiamentu', 'sd': 'پاپي امينٽو', 'sd-Arab': 'پاپي امينٽو', 'si': 'පපියමෙන්ටො', 'sk': 'papiamento', 'sl': 'papiamentu', 'smn': 'papiamentu', 'so': 'Papiamento', 'sq': 'papiamentisht', 'sr': 'папијаменто', 'sr-Cyrl': 'папијаменто', 'sr-Latn': 'papijamento', 'sv': 'papiamento', 'sw': 'Kipapiamento', 'ta': 'பபியாமென்டோ', 'te': 'పపియమేంటో', 'tg': 'папиаменто', 'th': 'ปาเปียเมนโต', 'ti': 'ፓፕያመንቶ', 'tk': 'papýamento dili', 'to': 'lea fakapapiamēnito', 'tr': 'Papiamento', 'tt': 'папьяменто', 'ug': 'پاپىيامېنتوچە', 'uk': 'папʼяменто', 'ur': 'پاپیامینٹو', 'uz': 'papiyamento', 'uz-Cyrl': 'папияменто', 'uz-Latn': 'papiyamento', 'vec': 'papiamento', 'vi': 'Tiếng Papiamento', 'wo': 'Papiyamento', 'xnr': 'पापियामेन्टो', 'yo': 'Èdè Papiamento', 'yrl': 'papiamẽtu', 'yue': '帕皮阿門托文', 'yue-Hans': '帕皮阿门托文', 'yue-Hant': '帕皮阿門托文', 'zh': '帕皮阿门托语', 'zh-Hans': '帕皮阿门托语', 'zh-Hant': '帕皮阿門托文', 'zu': 'isi-Papiamento'},
    'paq': {'en': 'Parya'},
    'par': {'en': 'Panamint'},
    'pas': {'en': 'Papasena'},
    'pat': {'en': 'Papitalai'},
    'pau': {'af': 'Palauaans', 'am': 'ፓሉዋን', 'ar': 'البالوان', 'as': 'পালাউৱান', 'ast': 'palauanu', 'az': 'palayan', 'az-Cyrl': 'палајан', 'az-Latn': 'palayan', 'be': 'палау', 'be-tarask': 'палау', 'bg': 'палауан', 'bgn': 'پالائویی', 'bn': 'পালায়ুয়ান', 'br': 'palau', 'brx': 'पालाऊवान', 'bs': 'palauanski', 'bs-Cyrl': 'палауански', 'bs-Latn': 'palauanski', 'ca': 'palauà', 'ca-ES-valencia': 'palauà', 'ccp': '𑄛𑄣𑄠𑄪𑄠𑄚𑄴', 'ce': 'палау', 'chr': 'ᏆᎳᎤᏩᏂ', 'ckb': 'پالائوویی', 'cs': 'palauština', 'cy': 'Palawan', 'da': 'palauansk', 'de': 'Palau', 'dsb': 'palauanšćina', 'el': 'Παλάουαν', 'el-polyton': 'Παλάουαν', 'en': 'Palauan', 'es': 'palauano', 'et': 'belau', 'eu': 'palauera', 'fa': 'پالائویی', 'ff-Adlm': '𞤆𞤢𞤤𞤢𞤵𞤴𞤢𞤲𞤪𞤫', 'fi': 'palau', 'fil': 'Palauan', 'fo': 'palauan', 'fr': 'palau', 'fy': 'Palauaansk', 'ga': 'Palabhais', 'gd': 'Palabhais', 'gl': 'palauano', 'gsw': 'Palau', 'gu': 'પલાઉઆન', 'ha': 'Palauan', 'he': 'פלוואן', 'hi': 'पलोउआन', 'hi-Latn': 'Palauan', 'hr': 'palauanski', 'hsb': 'palauanšćina', 'ht': 'palau', 'hu': 'palaui', 'hy': 'պալաուերեն', 'ia': 'palauano', 'id': 'Palau', 'ig': 'Palauan', 'is': 'paláska', 'it': 'palau', 'ja': 'パラオ語', 'jv': 'Palauan', 'ka': 'ფალაუანი', 'kgp': 'paravỹnũ', 'kk': 'палау тілі', 'kk-Cyrl': 'палау тілі', 'km': 'ប៉ាលូអាន', 'kn': 'ಪಲುಆನ್', 'ko': '팔라우어', 'kok': 'पालावान', 'kok-Deva': 'पालावान', 'ks': 'پَلااُواں', 'ks-Arab': 'پَلااُواں', 'ku': 'palawî', 'ky': 'палауанча', 'lb': 'Palau', 'lo': 'ປາລົວອານ', 'lt': 'palauliečių', 'lv': 'palaviešu', 'mai': 'पलौआन', 'mi': 'Pārau', 'mk': 'палауански', 'ml': 'പലാവുൻ', 'mn': 'палау', 'mr': 'पालाउआन', 'ms': 'Palauan', 'mt': 'Palawjan', 'my': 'ပလာအို', 'nb': 'palauisk', 'ne': 'पालाउवाली', 'nl': 'Palaus', 'nn': 'palauisk', 'no': 'palauisk', 'or': 'ପାଲାଉଆନ୍', 'pa': 'ਪਲਾਊਵੀ', 'pa-Guru': 'ਪਲਾਊਵੀ', 'pcm': 'Palaúán Lángwej', 'pl': 'palau', 'ps': 'پالان', 'pt': 'palauano', 'qu': 'Palauan Simi', 'rm': 'palau', 'ro': 'palauană', 'ru': 'палау', 'sc': 'palauanu', 'sd': 'پلون', 'sd-Arab': 'پلون', 'si': 'පලවුවන්', 'sk': 'palaučina', 'sl': 'palavanščina', 'smn': 'palaukielâ', 'so': 'Palauan', 'sq': 'paluanisht', 'sr': 'палауски', 'sr-Cyrl': 'палауски', 'sr-Latn': 'palauski', 'sv': 'palau', 'sw': 'Kipalau', 'ta': 'பலௌவன்', 'te': 'పలావెన్', 'th': 'ปาเลา', 'ti': 'ፓላውኛ', 'tk': 'palau dili', 'to': 'lea fakapalau', 'tr': 'Palau dili', 'ug': 'پالاۋچە', 'uk': 'палауанська', 'ur': 'پالاون', 'uz': 'palau', 'uz-Latn': 'palau', 'vec': 'palauan', 'vi': 'Tiếng Palauan', 'xnr': 'पलोउआन', 'yo': 'Èdè Pala', 'yrl': 'parauanu', 'yue': '帛琉文', 'yue-Hans': '帛琉文', 'yue-Hant': '帛琉文', 'zh': '帕劳语', 'zh-Hans': '帕劳语', 'zh-Hant': '帛琉文', 'zu': 'isi-Palauan'},
    'pav': {'en': 'Pakaásnovos'},
    'paw': {'en': 'Pawnee'},
    'pax': {'en': 'Pankararé'},
    'pay': {'en': 'Pech'},
    'paz': {'en': 'Pankararú'},
    'pbb': {'en': 'Páez'},
    'pbc': {'en': 'Patamona'},
    'pbe': {'en': 'Mezontla Popoloca'},
    'pbf': {'en': 'Coyotepec Popoloca'},
    'pbg': {'en': 'Paraujano'},
    'pbh': {'en': "E'ñapa Woromaipu"},
    'pbi': {'en': 'Parkwa'},
    'pbl': {'en': 'Mak (Nigeria)'},
    'pbm': {'en': 'Puebla Mazatec'},
    'pbn': {'en': 'Kpasam'},
    'pbo': {'en': 'Papel'},
    'pbp': {'en': 'Badyara'},
    'pbr': {'en': 'Pangwa'},
    'pbs': {'en': 'Central Pame'},
    'pbt': {'en': 'Southern Pashto'},
    'pbu': {'en': 'Northern Pashto'},
    'pbv': {'en': 'Pnar'},
    'pby': {'en': 'Pyu (Papua New Guinea)'},
    'pbz': {'en': 'Palu'},
    'pca': {'en': 'Santa Inés Ahuatempan Popoloca'},
    'pcb': {'en': 'Pear'},
    'pcc': {'en': 'Bouyei'},
    'pcd': {'ast': 'pícaru', 'br': 'pikardeg', 'ca': 'picard', 'ca-ES-valencia': 'picard', 'cs': 'picardština', 'cy': 'Picardeg', 'de': 'Picardisch', 'en': 'Picard', 'et': 'pikardi', 'fi': 'picardi', 'fr': 'picard', 'gd': 'Picard', 'hi-Latn': 'Picard', 'ht': 'picard', 'hy': 'պիկարդերեն', 'it': 'piccardo', 'ja': 'ピカルディ語', 'lb': 'Picardesch', 'lt': 'pikardų', 'mk': 'пикардски', 'nb': 'pikardisk', 'ne': 'पिकार्ड', 'nl': 'Picardisch', 'nn': 'pikardisk', 'no': 'pikardisk', 'pl': 'pikardyjski', 'sv': 'pikardiska', 'th': 'ปิการ์', 'to': 'lea fakapikāti', 'tr': 'Picard Dili', 'yue': '庇卡底文', 'yue-Hans': '庇卡底文', 'yue-Hant': '庇卡底文', 'zh-Hant': '庇卡底文'},
    'pce': {'en': 'Ruching Palaung'},
    'pcf': {'en': 'Paliyan'},
    'pcg': {'en': 'Paniya'},
    'pch': {'en': 'Pardhan'},
    'pci': {'en': 'Duruwa'},
    'pcj': {'en': 'Parenga'},
    'pck': {'en': 'Paite Chin'},
    'pcl': {'en': 'Pardhi'},
    'pcm': {'af': 'Nigeriese Pidgin', 'ak': 'Nigeriafoɔ Pigyin', 'am': 'የናይጄሪያ ፒጂን', 'ar': 'البدجنية النيجيرية', 'as': 'নাইজেৰিয়ান পিজিন', 'ast': 'nixerianu simplificáu', 'az': 'niger kreol', 'az-Cyrl': 'ниҝер креол', 'az-Latn': 'niger kreol', 'bal-Latn': 'Náijiri Pidgin', 'be': 'нігерыйскі піджын', 'be-tarask': 'нігерыйскі піджын', 'bg': 'нигерийски пиджин', 'bgn': 'نایجیریای پیڈگین', 'bn': 'নাইজেরিয় পিজিন', 'br': 'pidjin Nigeria', 'brx': 'नाइजेरियान पिदजीन', 'bs': 'nigerijski pidžin', 'bs-Cyrl': 'нигеријски пидгин', 'bs-Latn': 'nigerijski pidžin', 'ca': 'pidgin de Nigèria', 'ca-ES-valencia': 'pidgin de Nigèria', 'ccp': '𑄚𑄎𑄬𑄢𑄨𑄠𑄧 𑄛𑄨𑄎𑄨𑄚𑄴', 'ce': 'нигерийн-креолийн', 'chr': 'ᎾᎩᎵᎠᏂ ᏈᏥᏂ', 'ckb': 'پیجینی نیجریا', 'cs': 'nigerijský pidžin', 'cy': 'Pidgin Nigeria', 'da': 'nigeriansk pidgin', 'de': 'Nigerianisches Pidgin', 'dsb': 'nigerijanski pidgin', 'el': 'Πίτζιν Νιγηρίας', 'el-polyton': 'Πίτζιν Νιγηρίας', 'en': 'Nigerian Pidgin', 'es': 'pidgin de Nigeria', 'et': 'Nigeeria pidžinkeel', 'eu': 'Nigeriako pidgina', 'fa': 'نیم\u200cزبان نیجریه\u200cای', 'ff-Adlm': '𞤆𞤭𞤶𞤫𞤲𞤪𞤫 𞤐𞤢𞤶𞤭𞤪𞤭𞤴𞤢𞥄', 'fi': 'nigerianpidgin', 'fil': 'Nigerian Pidgin', 'fo': 'nigeriskt pidgin', 'fr': 'pidgin nigérian', 'ga': 'pidsean na Nigéire', 'gd': 'Beurla Nigèiriach', 'gl': 'pidgin nixeriano', 'gu': 'નાઇજેરિયન પીજીન', 'ha': 'Pidgin na Najeriya', 'he': 'פידגין ניגרי', 'hi': 'नाइजीरियाई पिडगिन', 'hi-Latn': 'Nigerian Pidgin', 'hr': 'nigerijski pidžin', 'hsb': 'nigerijanski pidgin', 'ht': 'pidgin nigérian', 'hu': 'nigériai pidgin', 'hy': 'նիգերիական փիջին', 'ia': 'pidgin nigerian', 'id': 'Pidgin Nigeria', 'ig': 'Pidgin ndị Naijirịa', 'is': 'nígerískt pidgin', 'it': 'pidgin nigeriano', 'ja': 'ナイジェリア・ピジン語', 'jv': 'Nigeria Pidgin', 'ka': 'ნიგერიული კრეოლური', 'kgp': 'pingĩn nĩjerijỹnũ', 'kk': 'нигериялық пиджин тілі', 'kk-Cyrl': 'нигериялық пиджин тілі', 'km': 'ភាសាទំនាក់ទំនងនីហ្សេរីយ៉ា', 'kn': 'ನೈಜೀರಿಯನ್ ಪಿಡ್ಗಿನ್', 'ko': '나이지리아 피진어', 'kok': 'नायजिरियन पिजिन', 'kok-Deva': 'नायजिरियन पिजिन', 'ku': 'pîdgînîya nîjeryayî', 'ky': 'аргындашкан тил (Нигерия)', 'lo': 'ໄນຈີຣຽນພິດກິນ', 'lt': 'Nigerijos pidžinų', 'lv': 'Nigērijas pidžinvaloda', 'mai': 'नाइजीरियाई पिजिन', 'mi': 'Ngāitiriana Kōrapurapu', 'mk': 'нигериски пиџин', 'ml': 'നൈജീരിയൻ പിഡ്\u200cഗിൻ', 'mn': 'нигерийн пиджин', 'mr': 'नायजिरिअन पिजिन', 'ms': 'Nigerian Pidgin', 'mt': 'Pidgin Niġerjan', 'my': 'နိုင်ဂျီးရီးယား ပစ်ဂျင်', 'nb': 'nigeriansk pidginspråk', 'ne': 'नाइजेरियाली पिड्जिन', 'nl': 'Nigeriaans Pidgin', 'nn': 'nigeriansk pidgin', 'no': 'nigeriansk pidginspråk', 'nqo': 'ߖߋ߬ߟߌ߬ߓߊߟߊ߫ ߔߌߘߑߜ߭ߍ߲ߞߊ߲', 'or': 'ନାଇଜେରୀୟ ପିଡଗିନ୍', 'pa': 'ਨਾਇਜੀਰੀਆਈ ਪਿਡਗਿਨ', 'pa-Guru': 'ਨਾਇਜੀਰੀਆਈ ਪਿਡਗਿਨ', 'pcm': 'Naijíriá Píjin', 'pl': 'pidżyn nigeryjski', 'ps': 'نائجیریا پیدجن', 'pt': 'pidgin nigeriano', 'qu': 'Pidgin Nigeriano Simi', 'rm': 'pidgin nigerian', 'ro': 'pidgin nigerian', 'ru': 'нигерийско-креольский', 'sc': 'pidgin nigerianu', 'sd': 'نائيجرين پجن', 'sd-Arab': 'نائيجرين پجن', 'si': 'නෛජීරියන් පෙන්ගින්', 'sk': 'nigerijský pidžin', 'sl': 'nigerijski pidžin', 'smn': 'Nigeria pidgin', 'so': 'Bidjinka Nayjeeriya', 'sq': 'pixhinishte nigeriane', 'sr': 'нигеријски пиџин', 'sr-Cyrl': 'нигеријски пиџин', 'sr-Latn': 'nigerijski pidžin', 'sv': 'nigeriansk pidgin', 'sw': 'Pijini ya Nigeria', 'ta': 'நைஜீரியன் பிட்கின்', 'te': 'నైజీరియన్ పిడ్గిన్', 'th': 'ไนจีเรียนพิดจิน', 'ti': 'ፒጂን ናይጀርያ', 'tk': 'nigeriýa-pijin dili', 'to': 'lea fakanaisilia', 'tr': 'Nijerya Pidgin dili', 'uk': 'нігерійсько-креольська', 'ur': 'نائجیریائی پڈگن', 'uz': 'kreol (Nigeriya)', 'uz-Latn': 'kreol (Nigeriya)', 'vec': 'pidgin nijerian', 'vi': 'Pidgin Nigeria', 'xnr': 'नाइजीरियाई पिडगिन', 'yo': 'Èdè Pijini ti Naijiriya', 'yrl': 'pidigĩ niyerianu', 'yue': '尼日利亞皮欽語', 'yue-Hans': '尼日利亚皮钦语', 'yue-Hant': '尼日利亞皮欽語', 'zh': '尼日利亚皮钦语', 'zh-Hans': '尼日利亚皮钦语', 'zh-Hant': '奈及利亞皮欽文', 'zu': 'isi-Nigerian Pidgin'},
    'pcn': {'en': 'Piti'},
    'pcp': {'en': 'Pacahuara'},
    'pcr': {'en': 'Panang'},
    'pcw': {'en': 'Pyapun'},
    'pda': {'en': 'Anam'},
    'pdc': {'ast': 'alemán de Pennsylvania', 'br': 'alamaneg Pennsylvania', 'ca': 'alemany pennsilvanià', 'ca-ES-valencia': 'alemany pennsilvanià', 'cs': 'němčina (pensylvánská)', 'cy': 'Almaeneg Pensylfania', 'de': 'Pennsylvaniadeutsch', 'en': 'Pennsylvania German', 'et': 'Pennsylvania saksa', 'fa': 'آلمانی پنسیلوانیایی', 'fi': 'pennsylvaniansaksa', 'fr': 'pennsilfaanisch', 'gd': 'Gearmailtis Phennsylvania', 'hi-Latn': 'Pennsylvania German', 'ht': 'pennsilfaanisch', 'hy': 'փենսիլվանական գերմաներեն', 'id': 'Jerman Pennsylvania', 'it': 'tedesco della Pennsylvania', 'ja': 'ペンシルベニア・ドイツ語', 'lb': 'Pennsylvaniadäitsch', 'lt': 'Pensilvanijos vokiečių', 'mk': 'пенсилваниски германски', 'nb': 'pennsylvaniatysk', 'ne': 'पेन्सिलभानियाली जर्मन', 'nl': 'Pennsylvania-Duits', 'nn': 'pennsylvaniatysk', 'no': 'pennsylvaniatysk', 'pl': 'pensylwański', 'sv': 'Pennsylvaniatyska', 'ta': 'பென்சில்வேனிய ஜெர்மன்', 'th': 'เยอรมันเพนซิลเวเนีย', 'to': 'lea fakasiamane-penisilivania', 'tr': 'Pensilvanya Almancası', 'yue': '賓夕法尼亞德文', 'yue-Hans': '宾夕法尼亚德文', 'yue-Hant': '賓夕法尼亞德文', 'zh-Hant': '賓夕法尼亞德文'},
    'pdi': {'en': 'Pa Di'},
    'pdn': {'en': 'Podena'},
    'pdo': {'en': 'Padoe'},
    'pdt': {'ast': 'plautdietsch', 'cs': 'němčina (plautdietsch)', 'de': 'Plautdietsch', 'en': 'Plautdietsch', 'et': 'mennoniidisaksa', 'fi': 'plautdietsch', 'fr': 'bas-prussien', 'gd': 'Plautdietsch', 'hi-Latn': 'Plautdietsch', 'ht': 'bas-prussien', 'hy': 'պլատագերմաներեն', 'ja': 'メノナイト低地ドイツ語', 'lb': 'Plattdäitsch', 'lt': 'vokiečių kalbos žemaičių tarmė', 'mk': 'менонитски долногермански', 'nb': 'plautdietsch', 'nl': 'Plautdietsch', 'nn': 'plautdietsch', 'no': 'plautdietsch', 'pl': 'plautdietsch', 'sv': 'mennonitisk lågtyska', 'th': 'เพลาท์ดิช', 'to': 'lea fakasiamane-lafalafa', 'tr': 'Plautdietsch', 'yue': '門諾低地德文', 'yue-Hans': '门诺低地德文', 'yue-Hant': '門諾低地德文', 'zh-Hant': '門諾低地德文'},
    'pdu': {'en': 'Kayan'},
    'pea': {'en': 'Peranakan Indonesian'},
    'peb': {'en': 'Eastern Pomo'},
    'ped': {'en': 'Mala (Papua New Guinea)'},
    'pee': {'en': 'Taje'},
    'pef': {'en': 'Northeastern Pomo'},
    'peg': {'en': 'Pengo'},
    'peh': {'en': 'Bonan'},
    'pei': {'en': 'Chichimeca-Jonaz'},
    'pej': {'en': 'Northern Pomo'},
    'pek': {'en': 'Penchal'},
    'pel': {'en': 'Pekal'},
    'pem': {'en': 'Phende'},
    'peo': {'ar': 'الفارسية القديمة', 'ast': 'persa antiguu', 'az': 'qədim fars', 'az-Latn': 'qədim fars', 'be': 'стараперсідская', 'be-tarask': 'стараперсідская', 'bg': 'староперсийски', 'bn': 'প্রাচীন ফার্সি', 'br': 'henberseg', 'brx': 'पुरानी फ़ारसी', 'bs': 'staroperzijski', 'bs-Cyrl': 'староперсијски', 'bs-Latn': 'staroperzijski', 'ca': 'persa antic', 'ca-ES-valencia': 'persa antic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄴 𑄜𑄢𑄴𑄥𑄨', 'cs': 'staroperština', 'cy': 'Hen Bersieg', 'da': 'oldpersisk', 'de': 'Altpersisch', 'el': 'Αρχαία Περσικά', 'el-polyton': 'Ἀρχαῖα περσικὰ', 'en': 'Old Persian', 'es': 'persa antiguo', 'et': 'vanapärsia', 'fa': 'فارسی باستان', 'fi': 'muinaispersia', 'fr': 'persan ancien', 'fur': 'vieri persian', 'fy': 'Aldperzysk', 'ga': 'Sean-Pheirsis', 'gd': 'Seann-Pheirsis', 'gsw': 'Altpersisch', 'gu': 'જૂની ફારસી', 'he': 'פרסית עתיקה', 'hi': 'पुरानी फारसी', 'hi-Latn': 'Old Persian', 'hr': 'staroperzijski', 'ht': 'persan ancien', 'hu': 'óperzsa', 'hy': 'հին պարսկերեն', 'id': 'Persia Kuno', 'is': 'fornpersneska', 'it': 'persiano antico', 'ja': '古代ペルシア語', 'ka': 'ძველი სპარსული', 'kgp': 'pérsa arkajku', 'kn': 'ಪ್ರಾಚೀನ ಪರ್ಶಿಯನ್', 'ko': '고대 페르시아어', 'ks': 'پرون فارسی', 'ks-Arab': 'پرون فارسی', 'lb': 'Alpersesch', 'lo': 'ເປີເຊຍໂບຮານ', 'lt': 'senoji persų', 'lv': 'senpersu', 'mk': 'староперсиски', 'ml': 'പഴയ പേർഷ്യൻ', 'mr': 'पुरातन फारसी', 'mt': 'Persjan Antik', 'my': 'ပါရှန် အဟောင်း', 'nb': 'gammelpersisk', 'ne': 'पुरातन फारसी', 'nl': 'Oudperzisch', 'nn': 'gammalpersisk', 'no': 'gammelpersisk', 'or': 'ପୁରୁଣା ପର୍ସିଆନ୍', 'pl': 'staroperski', 'pt': 'persa arcaico', 'rm': 'persian vegl', 'ro': 'persană veche', 'ru': 'староперсидский', 'sk': 'stará perzština', 'sl': 'stara perzijščina', 'sr': 'староперсијски', 'sr-Cyrl': 'староперсијски', 'sr-Latn': 'staropersijski', 'sv': 'fornpersiska', 'sw': 'Kiajemi cha Kale', 'ta': 'பழைய பெர்ஷியன்', 'te': 'ప్రాచీన పర్షియన్', 'th': 'เปอร์เซียโบราณ', 'to': 'lea fakapēsia-motuʻa', 'tr': 'Eski Farsça', 'ug': 'قەدىمكى پارىسچە', 'uk': 'давньоперська', 'vi': 'Tiếng Ba Tư cổ', 'yi': 'אַלט פּערסיש', 'yrl': 'persa arkaiku', 'yue': '古波斯文', 'yue-Hans': '古波斯文', 'yue-Hant': '古波斯文', 'zh': '古波斯语', 'zh-Hans': '古波斯语', 'zh-Hant': '古波斯文'},
    'pep': {'en': 'Kunja'},
    'peq': {'en': 'Southern Pomo'},
    'pes': {'en': 'Iranian Persian'},
    'pev': {'en': 'Pémono'},
    'pex': {'en': 'Petats'},
    'pey': {'en': 'Petjo'},
    'pez': {'en': 'Eastern Penan'},
    'pfa': {'en': 'Pááfang'},
    'pfe': {'en': 'Pere'},
    'pfl': {'ast': 'alemán palatinu', 'ca': 'alemany palatí', 'ca-ES-valencia': 'alemany palatí', 'cs': 'falčtina', 'cy': 'Almaeneg Palatin', 'de': 'Pfälzisch', 'en': 'Palatine German', 'et': 'Pfalzi', 'fi': 'pfaltsi', 'fr': 'allemand palatin', 'hi-Latn': 'Palatine German', 'ht': 'allemand palatin', 'hy': 'պալատինյան գերմաներեն', 'it': 'tedesco palatino', 'ja': 'プファルツ語', 'lb': 'Pfälzesch Däitsch', 'lt': 'vokiečių kalbos Pfalco tarmė', 'mk': 'фалечкогермански', 'nb': 'palatintysk', 'ne': 'पालाटिन जर्मन', 'nl': 'Paltsisch', 'nn': 'palatintysk', 'no': 'palatintysk', 'pl': 'palatynacki', 'sv': 'Pfalz-tyska', 'th': 'เยอรมันพาลาทิเนต', 'to': 'lea fakasiamane-palatine', 'tr': 'Palatin Almancası', 'yue': '普法爾茨德文', 'yue-Hans': '普法尔茨德文', 'yue-Hant': '普法爾茨德文', 'zh-Hant': '普法爾茨德文'},
    'pga': {'en': 'Sudanese Creole Arabic'},
    'pgd': {'en': 'Gāndhārī'},
    'pgg': {'en': 'Pangwali'},
    'pgi': {'en': 'Pagi'},
    'pgk': {'en': 'Rerep'},
    'pgl': {'en': 'Primitive Irish'},
    'pgn': {'en': 'Paelignian'},
    'pgs': {'en': 'Pangseng'},
    'pgu': {'en': 'Pagu'},
    'pgy': {'en': 'Pongyong'},
    'pgz': {'en': 'Papua New Guinean Sign Language'},
    'pha': {'en': 'Pa-Hng'},
    'phd': {'en': 'Phudagi'},
    'phg': {'en': 'Phuong'},
    'phh': {'en': 'Phukha'},
    'phi': {'en': 'Philippine languages'},
    'phi-pro': {'en': 'Proto-Philippine'},
    'phk': {'en': 'Phake'},
    'phl': {'en': 'Phalura'},
    'phm': {'en': 'Phimbi'},
    'phn': {'af': 'Fenisies', 'ar': 'الفينيقية', 'ast': 'feniciu', 'az': 'foyenik', 'az-Latn': 'foyenik', 'be': 'фінікійская', 'be-tarask': 'фінікійская', 'bg': 'финикийски', 'bn': 'ফোনিশীয়ান', 'br': 'fenikianeg', 'brx': 'फीनीसी', 'bs': 'feničanski', 'bs-Cyrl': 'феничански', 'bs-Latn': 'feničanski', 'ca': 'fenici', 'ca-ES-valencia': 'fenici', 'ccp': '𑄜𑄮𑄚𑄨𑄥𑄨𑄠𑄚𑄴', 'cs': 'féničtina', 'cy': 'Phoeniceg', 'da': 'fønikisk', 'de': 'Phönizisch', 'el': 'Φοινικικά', 'el-polyton': 'Φοινικικά', 'en': 'Phoenician', 'es': 'fenicio', 'et': 'foiniikia', 'fa': 'فنیقی', 'fi': 'foinikia', 'fr': 'phénicien', 'fy': 'Foenisysk', 'gd': 'Phenicis', 'gsw': 'Phönikisch', 'gu': 'ફોનિશિયન', 'he': 'פיניקית', 'hi': 'फोएनिशियन', 'hi-Latn': 'Phoenician', 'hr': 'fenički', 'ht': 'phénicien', 'hu': 'főniciai', 'hy': 'փյունիկերեն', 'id': 'Funisia', 'is': 'fönikíska', 'it': 'fenicio', 'ja': 'フェニキア語', 'ka': 'ფინიკიური', 'kgp': 'fenĩso', 'kn': 'ಫೀನಿಷಿಯನ್', 'ko': '페니키아어', 'ks': 'فونیٖشیَن', 'ks-Arab': 'فونیٖشیَن', 'lb': 'Phönikesch', 'lo': 'ຟີນີເຊຍ', 'lt': 'finikiečių', 'lv': 'feniķiešu', 'mk': 'феникиски', 'ml': 'ഫീനിഷ്യൻ', 'mr': 'फोनिशियन', 'mt': 'Feniċju', 'nb': 'fønikisk', 'ne': 'फोनिसियाली', 'nl': 'Foenicisch', 'nn': 'fønikisk', 'no': 'fønikisk', 'or': 'ଫୋନେସିଆନ୍', 'pl': 'fenicki', 'pt': 'fenício', 'rm': 'fenizian', 'ro': 'feniciană', 'ru': 'финикийский', 'sk': 'feničtina', 'sl': 'feničanščina', 'sr': 'феничански', 'sr-Cyrl': 'феничански', 'sr-Latn': 'feničanski', 'sv': 'feniciska', 'ta': 'ஃபொனிஷியன்', 'te': 'ఫోనికన్', 'th': 'ฟินิเชีย', 'to': 'lea fakafoinikia', 'tr': 'Fenike dili', 'ug': 'فىنىكىيەچە', 'uk': 'фінікійсько-пунічна', 'vi': 'Tiếng Phoenicia', 'yrl': 'finísiu', 'yue': '腓尼基文', 'yue-Hans': '腓尼基文', 'yue-Hant': '腓尼基文', 'zh': '腓尼基语', 'zh-Hans': '腓尼基语', 'zh-Hant': '腓尼基文'},
    'pho': {'en': 'Phunoi'},
    'phq': {'en': "Phana'"},
    'phr': {'en': 'Pahari-Potwari'},
    'pht': {'en': 'Phu Thai'},
    'phu': {'en': 'Phuan'},
    'phv': {'en': 'Pahlavani'},
    'phw': {'en': 'Phangduwali'},
    'pi': {'ar': 'البالية', 'ast': 'pali', 'az': 'pali', 'az-Latn': 'pali', 'bg': 'пали', 'bn': 'পালি', 'br': 'pali', 'brx': 'पाली', 'bs': 'pali', 'bs-Cyrl': 'пали', 'bs-Latn': 'pali', 'ca': 'pali', 'ca-ES-valencia': 'pali', 'ccp': '𑄛𑄣𑄨', 'cs': 'pálí', 'cy': 'Pali', 'da': 'pali', 'de': 'Pali', 'el': 'Πάλι', 'el-polyton': 'Πάλι', 'en': 'Pali', 'es': 'pali', 'et': 'paali', 'fa': 'پالی', 'fi': 'paali', 'fr': 'pali', 'fy': 'Pali', 'ga': 'Páilis', 'gd': 'Pali', 'gsw': 'Pali', 'gu': 'પાલી', 'he': 'פאלי', 'hi': 'पाली', 'hi-Latn': 'Pali', 'hr': 'pali', 'ht': 'pali', 'hu': 'pali', 'hy': 'պալի', 'id': 'Pali', 'is': 'palí', 'it': 'pali', 'ja': 'パーリ語', 'ka': 'პალი', 'kgp': 'pari', 'kn': 'ಪಾಲಿ', 'ko': '팔리어', 'ks': 'پالی', 'ks-Arab': 'پالی', 'lb': 'Pali', 'lo': 'ປາລີ', 'lt': 'pali', 'lv': 'pāli', 'mk': 'пали', 'ml': 'പാലി', 'mr': 'पाली', 'mt': 'Pali', 'my': 'ပါဠိ', 'nb': 'pali', 'ne': 'पाली', 'nl': 'Pali', 'nn': 'pali', 'no': 'pali', 'or': 'ପାଲି', 'pa': 'ਪਾਲੀ', 'pa-Guru': 'ਪਾਲੀ', 'pl': 'palijski', 'pt': 'páli', 'rm': 'pali', 'ro': 'pali', 'ru': 'пали', 'sk': 'pálí', 'sl': 'palijščina', 'sr': 'пали', 'sr-Cyrl': 'пали', 'sr-Latn': 'pali', 'sv': 'pali', 'ta': 'பாலி', 'te': 'పాలీ', 'th': 'บาลี', 'to': 'lea fakapāli', 'tr': 'Pali', 'ug': 'پالىچە', 'uk': 'палі', 'vi': 'Tiếng Pali', 'yrl': 'pári', 'yue': '巴利文', 'yue-Hans': '巴利文', 'yue-Hant': '巴利文', 'zh': '巴利语', 'zh-Hans': '巴利语', 'zh-Hant': '巴利文'},
    'pia': {'en': 'Pima Bajo'},
    'pib': {'en': 'Yine'},
    'pic': {'en': 'Pinji'},
    'pid': {'en': 'Piaroa'},
    'pie': {'en': 'Piro'},
    'pif': {'en': 'Pingelapese'},
    'pig': {'en': 'Pisabo'},
    'pih': {'en': 'Pitcairn-Norfolk'},
    'pii': {'en': 'Pini'},
    'pij': {'en': 'Pijao'},
    'pil': {'en': 'Yom'},
    'pim': {'en': 'Powhatan'},
    'pin': {'en': 'Piame'},
    'pio': {'en': 'Piapoco'},
    'pip': {'en': 'Pero'},
    'pir': {'en': 'Piratapuyo'},
    'pis': {'af': 'Pijin', 'am': 'ፒጂን', 'ar': 'بيجين', 'as': 'পিজিন', 'az': 'picin', 'az-Latn': 'picin', 'bal-Latn': 'Pijen', 'be': 'саламонскі піджын', 'be-tarask': 'саламонскі піджын', 'bg': 'пиджин', 'bn': 'পিজিন', 'br': 'pidjin', 'brx': 'पिजिन', 'bs': 'pidžin', 'bs-Cyrl': 'пијин', 'bs-Latn': 'pidžin', 'ca': 'pidgin', 'ca-ES-valencia': 'pidgin', 'chr': 'ᏈᏥᎾ', 'cs': 'pidžin (Šalomounovy ostrovy)', 'cy': 'Pijin', 'da': 'pijin', 'de': 'Pijin', 'dsb': 'pijinšćina', 'el': 'Πιτζίν', 'el-polyton': 'Πιτζίν', 'en': 'Pijin', 'es': 'pidgin salomonense', 'et': 'pijini', 'eu': 'pijina', 'fa': 'پی\u200cجین', 'ff-Adlm': '𞤆𞤭𞤶𞤭𞤲𞤪𞤫', 'fi': 'pijin', 'fil': 'Pijin', 'fr': 'pijin', 'ga': 'Pijin', 'gd': 'Pijin', 'gl': 'pijin', 'gu': 'પિજિન', 'ha': 'Pijin', 'he': 'פייג׳ין', 'hi': 'पाईजिन', 'hi-Latn': 'Pijin', 'hr': 'pijin', 'hsb': 'pijinšćina', 'ht': 'pijin', 'hu': 'pidzsin', 'hy': 'փիջին', 'ia': 'pijin', 'id': 'Pijin', 'ig': 'Pijin', 'is': 'pijin', 'it': 'pijin', 'ja': 'ピジン語', 'jv': 'Pijin', 'ka': 'პიჯინი', 'kk': 'пиджин тілі', 'kk-Cyrl': 'пиджин тілі', 'km': 'ពីជីន', 'kn': 'ಪಿಜಿನ್', 'ko': '피진어', 'kok': 'पिजिन', 'kok-Deva': 'पिजिन', 'ku': 'pijînî', 'ky': 'пижинче (Соломон Аралдары)', 'lo': 'ປິຈິນ', 'lt': 'pidžinų', 'lv': 'pidžinvaloda', 'mai': 'पिजिन', 'mi': 'Pītini', 'mk': 'пиџин', 'ml': 'പിജിൻ', 'mn': 'пижин', 'mr': 'पिजिन', 'ms': 'Pijin', 'my': 'ပီဂျင်', 'nb': 'pijin', 'ne': 'पिजिन', 'nl': 'Pijin', 'nn': 'pijin', 'no': 'pijin', 'or': 'ପିଜିନ୍', 'pa': 'ਪਿਜਿਨ', 'pa-Guru': 'ਪਿਜਿਨ', 'pcm': 'Píjín Lángwej', 'pis': 'Pijin', 'pl': 'pijin', 'ps': 'پیجین', 'pt': 'pijin', 'qu': 'Pijin Simi', 'ro': 'pijin', 'ru': 'соломонский пиджин', 'sc': 'pijin', 'sd': 'پائجن', 'sd-Arab': 'پائجن', 'si': 'පිජින්', 'sk': 'pidžin', 'sl': 'pidžin', 'so': 'Pijin', 'sq': 'pixhinisht', 'sr': 'пиџин', 'sr-Cyrl': 'пиџин', 'sr-Latn': 'pidžin', 'sv': 'pidginspråk', 'sw': 'Kipijini', 'syr': 'ܦܝܓܝܢܝܬ', 'ta': 'பிஜின்', 'te': 'పిజిన్', 'th': 'พิดจิน', 'ti': 'ፒጂን', 'tk': 'pijin dili', 'to': 'lea fakapisini', 'tr': 'Pijin dili', 'uk': 'піджин', 'ur': 'پجن', 'uz': 'pijin', 'uz-Latn': 'pijin', 'vec': 'pijin', 'vi': 'Tiếng Pijin', 'yo': 'Èdè Piji', 'yue': '皮京文', 'yue-Hans': '皮京文', 'yue-Hant': '皮京文', 'zh': '皮京语', 'zh-Hans': '皮京语', 'zh-Hant': '皮金語', 'zu': 'Pijin'},
    'pit': {'en': 'Pitta Pitta'},
    'piu': {'en': 'Pintupi-Luritja'},
    'piv': {'en': 'Pileni'},
    'piw': {'en': 'Pimbwe'},
    'pix': {'en': 'Piu'},
    'piy': {'en': 'Piya-Kwonci'},
    'piz': {'en': 'Pije'},
    'pjt': {'en': 'Pitjantjatjara'},
    'pka': {'en': 'Ardhamāgadhī Prākrit'},
    'pkb': {'en': 'Pokomo'},
    'pkc': {'en': 'Paekche'},
    'pkg': {'en': 'Pak-Tong'},
    'pkh': {'en': 'Pankhu'},
    'pkn': {'en': 'Pakanha'},
    'pko': {'en': 'Pökoot'},
    'pkp': {'en': 'Pukapuka'},
    'pkr': {'en': 'Attapady Kurumba'},
    'pks': {'en': 'Pakistan Sign Language'},
    'pkt': {'en': 'Maleng'},
    'pku': {'en': 'Paku'},
    'pl': {'af': 'Pools', 'agq': 'Kpɔlìs', 'ak': 'Pɔland kasa', 'am': 'ፖሊሽ', 'ar': 'البولندية', 'as': 'প’লিচ', 'asa': 'Kipolandi', 'ast': 'polacu', 'az': 'polyak', 'az-Cyrl': 'полјак', 'az-Latn': 'polyak', 'bal-Latn': 'Pólayndi', 'bas': 'Hɔp u pɔlɔ̄nà', 'be': 'польская', 'be-tarask': 'польская', 'bem': 'Ichi Polishi', 'bez': 'Hipolandi', 'bg': 'полски', 'bgn': 'پولنڈی', 'blo': 'Polanɖ kagɩja', 'bm': 'polonekan', 'bn': 'পোলিশ', 'br': 'poloneg', 'brx': 'पलिश', 'bs': 'poljski', 'bs-Cyrl': 'пољски', 'bs-Latn': 'poljski', 'ca': 'polonès', 'ca-ES-valencia': 'polonès', 'ccp': '𑄛𑄮𑄣𑄨𑄌𑄴', 'ce': 'полякийн', 'ceb': 'Polish', 'cgg': 'Orupoori', 'chr': 'ᏉᎵᏍ', 'ckb': 'پۆڵەندی', 'cs': 'polština', 'cv': 'поляк', 'cy': 'Pwyleg', 'da': 'polsk', 'dav': 'Kipolandi', 'de': 'Polnisch', 'dje': 'Polonee senni', 'dsb': 'pólšćina', 'dyo': 'polonees', 'dz': 'པོ་ལིཤ་ཁ', 'ebu': 'Kĩpolandi', 'ee': 'Polishgbe', 'el': 'Πολωνικά', 'el-polyton': 'Πολωνικά', 'en': 'Polish', 'eo': 'pola', 'es': 'polaco', 'et': 'poola', 'eu': 'poloniera', 'ewo': 'ǹkɔ́bɔ fólis', 'fa': 'لهستانی', 'ff': 'Poloneere', 'ff-Adlm': '𞤆𞤮𞤤𞤢𞤲𞤣𞤭𞥅𞤪𞤫', 'ff-Latn': 'Poloneere', 'fi': 'puola', 'fil': 'Polish', 'fo': 'pólskt', 'fr': 'polonais', 'fur': 'polac', 'fy': 'Poalsk', 'ga': 'Polainnis', 'gd': 'Pòlainnis', 'gl': 'polaco', 'gsw': 'Polnisch', 'gu': 'પોલીશ', 'guz': 'Kipolandi', 'ha': 'Harshen Polan', 'he': 'פולנית', 'hi': 'पोलिश', 'hi-Latn': 'Polish', 'hr': 'poljski', 'hsb': 'pólšćina', 'ht': 'polonais', 'hu': 'lengyel', 'hy': 'լեհերեն', 'ia': 'polonese', 'id': 'Polski', 'ie': 'polonesi', 'ig': 'Asụsụ Polish', 'is': 'pólska', 'it': 'polacco', 'ja': 'ポーランド語', 'jmc': 'Kyipolandi', 'jv': 'Polandia', 'ka': 'პოლონური', 'kab': 'Tapulunit', 'kam': 'Kipolandi', 'kde': 'Chipolandi', 'kea': 'pulaku', 'kgp': 'poronẽj', 'khq': 'Polonee senni', 'ki': 'Kipolandi', 'kk': 'поляк тілі', 'kk-Cyrl': 'поляк тілі', 'kln': 'kutitap Poland', 'km': 'ប៉ូឡូញ', 'kn': 'ಪೊಲಿಶ್', 'ko': '폴란드어', 'kok': 'पोलिश', 'kok-Deva': 'पोलिश', 'ks': 'پالِش', 'ks-Arab': 'پالِش', 'ksb': 'Kipolandi', 'ksf': 'ripɔlɔ́n', 'ksh': 'Pollnesch', 'ku': 'polonî', 'kxv': 'polis', 'kxv-Deva': 'पोलिस', 'kxv-Latn': 'polis', 'kxv-Orya': 'ପଲିସ୍', 'kxv-Telu': 'పోలిస్', 'ky': 'полякча', 'lag': 'Kɨpólandi', 'lb': 'Polnesch', 'lg': 'Lupolandi', 'lkt': 'Polish Iyápi', 'lld': 'polach', 'ln': 'lipolonɛ', 'lo': 'ໂປລິຊ', 'lrc': 'لأھئستانی', 'lt': 'lenkų', 'lu': 'Mpoloni', 'luo': 'Kipolandi', 'luy': 'Kipolandi', 'lv': 'poļu', 'mai': 'पोलिश', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lpoland', 'mer': 'Kĩpolandi', 'mfe': 'polone', 'mg': 'Poloney', 'mgh': 'Ipolandi', 'mi': 'Pōrihi', 'mk': 'полски', 'ml': 'പോളിഷ്', 'mn': 'польш', 'mni': 'পোলিশ', 'mni-Beng': 'পোলিশ', 'mr': 'पोलिश', 'ms': 'Poland', 'mt': 'Pollakk', 'mua': 'Poloniya', 'my': 'ပိုလန်', 'mzn': 'لهستونی', 'naq': 'Poleǁî gowab', 'nb': 'polsk', 'nd': 'isi-Pholoshi', 'ne': 'पोलिस', 'nl': 'Pools', 'nmg': 'Kiɛl pɔlɔŋe', 'nn': 'polsk', 'no': 'polsk', 'nqo': 'ߔߟߏߣߍߞߊ߲', 'nus': 'Thok pölicni', 'nyn': 'Orupoori', 'om': 'Afaan Polandii', 'or': 'ପୋଲିଶ୍', 'pa': 'ਪੋਲੈਂਡੀ', 'pa-Guru': 'ਪੋਲੈਂਡੀ', 'pcm': 'Pólánd Lángwej', 'pl': 'polski', 'ps': 'پولنډي', 'pt': 'polonês', 'qu': 'Polaco Simi', 'rm': 'polac', 'rn': 'Ikinyapolonye', 'ro': 'poloneză', 'rof': 'Kipolandi', 'ru': 'польский', 'rw': 'Igipolone', 'rwk': 'Kyipolandi', 'saq': 'Kipolandi', 'sbp': 'Ishipolandi', 'sc': 'polacu', 'sd': 'پولش', 'sd-Arab': 'پولش', 'se': 'polskkagiella', 'seh': 'polonês', 'ses': 'Polonee senni', 'sg': 'Polonëe', 'shi': 'ⵜⴰⴱⵓⵍⵓⵏⵉⵜ', 'shi-Latn': 'Tabulunit', 'shi-Tfng': 'ⵜⴰⴱⵓⵍⵓⵏⵉⵜ', 'si': 'පෝලන්ත', 'sk': 'poľština', 'sl': 'poljščina', 'smn': 'puolakielâ', 'sn': 'chiPolish', 'so': 'Boolish', 'sq': 'polonisht', 'sr': 'пољски', 'sr-Cyrl': 'пољски', 'sr-Latn': 'poljski', 'sv': 'polska', 'sw': 'Kipolandi', 'syr': 'ܦܘܠܢܕܐܝܬ', 'ta': 'போலிஷ்', 'te': 'పోలిష్', 'teo': 'Kipolandi', 'tg': 'лаҳистонӣ', 'th': 'โปแลนด์', 'ti': 'ፖሊሽ', 'tk': 'polýak dili', 'to': 'lea fakapolani', 'tr': 'Lehçe', 'tt': 'поляк', 'twq': 'Polonee senni', 'tzm': 'Tappulunit', 'ug': 'پولەكچە', 'uk': 'польська', 'ur': 'پولش', 'uz': 'polyakcha', 'uz-Cyrl': 'полякча', 'uz-Latn': 'polyakcha', 'vai': 'ꗁꔒꔻ', 'vai-Latn': 'Pɔ́lési', 'vai-Vaii': 'ꗁꔒꔻ', 'vec': 'polaco', 'vi': 'Tiếng Ba Lan', 'vun': 'Kyipolandi', 'wae': 'Polniš', 'wo': 'Polone', 'xh': 'Isi-Polish', 'xnr': 'पोलिश', 'xog': 'Olupolandi', 'yav': 'nupolonɛ́ɛ', 'yi': 'פּויליש', 'yo': 'Èdè Póláǹdì', 'yrl': 'purunei', 'yue': '波蘭文', 'yue-Hans': '波兰文', 'yue-Hant': '波蘭文', 'zgh': 'ⵜⴰⴱⵓⵍⵓⵏⵉⵜ', 'zh': '波兰语', 'zh-Hans': '波兰语', 'zh-Hant': '波蘭文', 'zu': 'isi-Polish'},
    'pla': {'en': 'Miani'},
    'plb': {'en': 'Polonombauk'},
    'plc': {'en': 'Central Palawano'},
    'pld': {'en': 'Polari'},
    'ple': {'en': "Palu'e"},
    'plf': {'en': 'Central Malayo-Polynesian languages'},
    'plf-pro': {'en': 'Proto-Central Malayo-Polynesian'},
    'plg': {'en': 'Pilagá'},
    'plh': {'en': 'Paulohi'},
    'plj': {'en': 'Polci'},
    'plk': {'en': 'Kohistani Shina'},
    'pll': {'en': 'Shwe Palaung'},
    'pln': {'en': 'Palenquero'},
    'plo': {'en': 'Oluta Popoluca'},
    'plp': {'en': 'Palpa'},
    'plq': {'en': 'Palaic'},
    'plr': {'en': 'Palaka Senoufo'},
    'pls': {'en': 'San Marcos Tlacoyalco Popoloca'},
    'plt': {'en': 'Plateau Malagasy'},
    'plu': {'en': 'Palikúr'},
    'plv': {'en': 'Southwest Palawano'},
    'plw': {'en': "Brooke's Point Palawano"},
    'ply': {'en': 'Bolyu'},
    'plz': {'en': 'Paluan'},
    'pma': {'en': 'Paama'},
    'pmb': {'en': 'Pambia'},
    'pmc': {'en': 'Palumata'},
    'pmd': {'en': 'Pallanganmiddang'},
    'pme': {'en': 'Pwaamei'},
    'pmf': {'en': 'Pamona'},
    'pmh': {'en': 'Māhārāṣṭri Prākrit'},
    'pmi': {'en': 'Northern Pumi'},
    'pmj': {'en': 'Southern Pumi'},
    'pmk': {'en': 'Pamlico'},
    'pml': {'en': 'Lingua Franca'},
    'pmm': {'en': 'Pomo'},
    'pmn': {'en': 'Pam'},
    'pmo': {'en': 'Pom'},
    'pmq': {'en': 'Northern Pame'},
    'pmr': {'en': 'Paynamar'},
    'pms': {'ast': 'piamontés', 'br': 'piemonteg', 'ca': 'piemontès', 'ca-ES-valencia': 'piemontès', 'cs': 'piemonština', 'cy': 'Piedmonteg', 'de': 'Piemontesisch', 'en': 'Piedmontese', 'et': 'piemonte', 'fi': 'piemonte', 'fr': 'piémontais', 'gd': 'Piedmontese', 'hi-Latn': 'Piedmontese', 'ht': 'piémontais', 'hy': 'պիեմոնտերեն', 'it': 'piemontese', 'ja': 'ピエモンテ語', 'lb': 'Piemontesesch', 'lt': 'italų kalbos Pjemonto tarmė', 'mk': 'пиемонтски', 'nb': 'piemontesisk', 'ne': 'पिएडमोन्तेसे', 'nl': 'Piëmontees', 'nn': 'piemontesisk', 'no': 'piemontesisk', 'pl': 'piemoncki', 'sv': 'piemontesiska', 'th': 'พีดมอนต์', 'to': 'lea fakapiemonite', 'tr': 'Piyemontece', 'yue': '皮埃蒙特文', 'yue-Hans': '皮埃蒙特文', 'yue-Hant': '皮埃蒙特文', 'zh-Hant': '皮埃蒙特文'},
    'pmt': {'en': 'Tuamotuan'},
    'pmu': {'en': 'Mirpur Panjabi'},
    'pmw': {'en': 'Plains Miwok'},
    'pmx': {'en': 'Poumei Naga'},
    'pmy': {'en': 'Papuan Malay'},
    'pmz': {'en': 'Southern Pame'},
    'pna': {'en': 'Punan Bah-Biau'},
    'pnb': {'en': 'Western Panjabi'},
    'pnc': {'en': 'Pannei'},
    'pnd': {'en': 'Mpinda'},
    'pne': {'en': 'Western Penan'},
    'png': {'en': 'Pangu'},
    'pnh': {'en': 'Penrhyn'},
    'pni': {'en': 'Aoheng'},
    'pnj': {'en': 'Pinjarup'},
    'pnk': {'en': 'Paunaka'},
    'pnl': {'en': 'Paleni'},
    'pnm': {'en': 'Punan Batu 1'},
    'pnn': {'en': 'Pinai-Hagahai'},
    'pno': {'en': 'Panobo'},
    'pnp': {'en': 'Pancana'},
    'pnq': {'en': 'Pana (Burkina Faso)'},
    'pnr': {'en': 'Panim'},
    'pns': {'en': 'Ponosakan'},
    'pnt': {'ast': 'pónticu', 'br': 'ponteg', 'ca': 'pòntic', 'ca-ES-valencia': 'pòntic', 'cs': 'pontština', 'cy': 'Ponteg', 'de': 'Pontisch', 'en': 'Pontic', 'et': 'pontose', 'fi': 'pontoksenkreikka', 'fr': 'pontique', 'hi-Latn': 'Pontic', 'ht': 'pontique', 'hy': 'պոնտերեն', 'it': 'pontico', 'ja': 'ポントス・ギリシャ語', 'ko': '폰틱어', 'lb': 'Pontesch', 'lt': 'Ponto', 'mk': 'понтски', 'nb': 'pontisk', 'ne': 'पोन्टिक', 'nl': 'Pontisch', 'nn': 'pontisk', 'no': 'pontisk', 'pl': 'pontyjski', 'sv': 'pontiska', 'th': 'พอนติก', 'to': 'lea fakaponitiki', 'tr': 'Kuzeybatı Kafkasya', 'yue': '旁狄希臘文', 'yue-Hans': '旁狄希腊文', 'yue-Hant': '旁狄希臘文', 'zh-Hant': '旁狄希臘文'},
    'pnu': {'en': 'Jiongnai Bunu'},
    'pnv': {'en': 'Pinigura'},
    'pnw': {'en': 'Banyjima'},
    'pnx': {'en': 'Phong-Kniang'},
    'pny': {'en': 'Pinyin'},
    'pnz': {'en': 'Pana (Central African Republic)'},
    'poc': {'en': 'Poqomam'},
    'pod': {'en': 'Ponares'},
    'poe': {'en': 'San Juan Atzingo Popoloca'},
    'pof': {'en': 'Poke'},
    'pog': {'en': 'Potiguára'},
    'poh': {'en': "Poqomchi'"},
    'poi': {'en': 'Highland Popoluca'},
    'pok': {'en': 'Pokangá'},
    'pom': {'en': 'Southeastern Pomo'},
    'pon': {'ar': 'البوهنبيايان', 'ast': 'pohnpeianu', 'az': 'ponpey', 'az-Latn': 'ponpey', 'bg': 'понапеан', 'bn': 'পোহ্নপেইয়ান', 'br': 'pohnpei', 'brx': 'पोहनपी', 'bs': 'ponpejski', 'bs-Cyrl': 'понпејски', 'bs-Latn': 'ponpejski', 'ca': 'ponapeà', 'ca-ES-valencia': 'ponapeà', 'ccp': '𑄛𑄮𑄚𑄴𑄦𑄧𑄛𑄳𑄆𑄬𑄠𑄚𑄴', 'cs': 'pohnpeiština', 'cy': 'Pohnpeianeg', 'da': 'ponape', 'de': 'Ponapeanisch', 'el': 'Πομπηικά', 'el-polyton': 'Πομπηικά', 'en': 'Pohnpeian', 'es': 'pohnpeiano', 'et': 'poonpei', 'fa': 'پانپیی', 'fi': 'pohnpei', 'fr': 'pohnpei', 'fy': 'Pohnpeiaansk', 'gd': 'Cànan Pohnpei', 'gsw': 'Ponapeanisch', 'gu': 'પોહપિએન', 'he': 'פונפיאן', 'hi': 'पोह्नपिएन', 'hi-Latn': 'Pohnpeian', 'hr': 'pohnpeian', 'ht': 'pohnpei', 'hu': 'pohnpei', 'hy': 'պոնպեերեն', 'id': 'Pohnpeia', 'is': 'ponpeiska', 'it': 'ponape', 'ja': 'ポンペイ語', 'kgp': 'pãnhpejỹnũ', 'kn': 'ಪೋನ್\u200c\u200cಪಿಯನ್', 'ko': '폼페이어', 'ks': 'پانپییَن', 'ks-Arab': 'پانپییَن', 'lb': 'Ponapeanesch', 'lo': 'ພອນເພ', 'lt': 'Ponapės', 'lv': 'ponapiešu', 'mk': 'понпејски', 'ml': 'പൊൻപിയൻ', 'mr': 'पोह्नपियन', 'mt': 'Ponpejan', 'nb': 'ponapisk', 'nl': 'Pohnpeiaans', 'nn': 'ponapisk', 'no': 'ponapisk', 'or': 'ପୋହପିଏନ୍', 'pl': 'ponpejski', 'pt': 'pohnpeiano', 'rm': 'ponapean', 'ro': 'pohnpeiană', 'ru': 'понапе', 'sk': 'pohnpeiština', 'sl': 'ponpejščina', 'sr': 'понпејски', 'sr-Cyrl': 'понпејски', 'sr-Latn': 'ponpejski', 'sv': 'pohnpeiska', 'ta': 'ஃபோன்பெயென்', 'te': 'పోహ్న్పెయన్', 'th': 'พอห์นเพ', 'to': 'lea fakaponapē', 'tr': 'Pohnpeian', 'ug': 'پوناپېئانچە', 'uk': 'понапе', 'vi': 'Tiếng Pohnpeian', 'yrl': 'pũpeianu', 'yue': '波那貝文', 'yue-Hans': '波那贝文', 'yue-Hant': '波那貝文', 'zh': '波纳佩语', 'zh-Hans': '波纳佩语', 'zh-Hant': '波那貝文'},
    'poo': {'en': 'Central Pomo'},
    'pop': {'en': 'Pwapwâ'},
    'poq': {'en': 'Texistepec Popoluca'},
    'pos': {'en': 'Sayula Popoluca'},
    'pot': {'en': 'Potawatomi'},
    'pov': {'en': 'Upper Guinea Crioulo'},
    'pow': {'en': 'San Felipe Otlaltepec Popoloca'},
    'pox': {'en': 'Polabian'},
    'poy': {'en': 'Pogolo'},
    'poz': {'en': 'Malayo-Polynesian languages'},
    'poz-btk-pro': {'en': 'Proto-Bungku-Tolaki'},
    'poz-cet-pro': {'en': 'Proto-Central-Eastern Malayo-Polynesian'},
    'poz-hce-pro': {'en': 'Proto-Halmahera-Cenderawasih'},
    'poz-lgx-pro': {'en': 'Proto-Lampungic'},
    'poz-mcm-pro': {'en': 'Proto-Malayo-Chamic'},
    'poz-mly-pro': {'en': 'Proto-Malayic'},
    'poz-msa-pro': {'en': 'Proto-Malayo-Sumbawan'},
    'poz-oce-pro': {'en': 'Proto-Oceanic'},
    'poz-pep-pro': {'en': 'Proto-Eastern Polynesian'},
    'poz-pnp-pro': {'en': 'Proto-Nuclear Polynesian'},
    'poz-pol-pro': {'en': 'Proto-Polynesian'},
    'poz-pro': {'en': 'Proto-Malayo-Polynesian'},
    'poz-ssw-pro': {'en': 'Proto-South Sulawesi'},
    'poz-sus-pro': {'en': 'Proto-Sunda-Sulawesi'},
    'poz-swa-pro': {'en': 'Proto-North Sarawak'},
    'ppa': {'en': 'Pao'},
    'ppe': {'en': 'Papi'},
    'ppi': {'en': 'Paipai'},
    'ppk': {'en': 'Uma'},
    'ppl': {'en': 'Pipil'},
    'ppm': {'en': 'Papuma'},
    'ppn': {'en': 'Papapana'},
    'ppo': {'en': 'Folopa'},
    'ppp': {'en': 'Pelende'},
    'ppq': {'en': 'Pei'},
    'ppr': {'en': 'Piru'},
    'pps': {'en': 'San Luís Temalacayuca Popoloca'},
    'ppt': {'en': 'Pare'},
    'ppu': {'en': 'Papora'},
    'pqa': {'en': "Pa'a"},
    'pqe': {'en': 'Eastern Malayo-Polynesian languages'},
    'pqe-pro': {'en': 'Proto-Eastern Malayo-Polynesian'},
    'pqm': {'af': 'Maliseet-Passamaquoddy', 'am': 'ማሊሴት-ፓሳማኩዎድይ', 'ar': 'الماليزيت-باساماكودي', 'as': 'মালিছীট-পাছামাকু’ডী', 'az': 'malesit-passamakvodi', 'az-Latn': 'malesit-passamakvodi', 'be': 'малесіт-пасамакуоды', 'be-tarask': 'малесіт-пасамакуоды', 'bg': 'малисеет-пасамакуоди', 'bn': 'মালিসেট-পাসামাকুয়োড্ডি', 'br': 'malisiteg-pasamawkodieg', 'brx': 'मालिसीत- पास्सामक्वद्दी', 'bs': 'malisit-pasamakvodi', 'bs-Cyrl': 'малисет-шасамкуоди', 'bs-Latn': 'malisit-pasamakvodi', 'ca': 'maliseet-passamaquoddy', 'ca-ES-valencia': 'maliseet-passamaquoddy', 'chr': 'ᎹᎵᏏᏘ-ᏇᏌᎹᏉᏗ', 'cs': 'malesitština-passamaquoddština', 'cy': 'Maliseet-Passamaquoddy', 'da': 'maliseet-passamaquoddy', 'de': 'Maliseet-Passamaquoddy', 'dsb': 'maliseet-passamaquoddyšćina', 'el': 'Μαλισιτ-Πασσαμακουόντ', 'el-polyton': 'Μαλισιτ-Πασσαμακουόντ', 'en': 'Maliseet-Passamaquoddy', 'es': 'maliseet-passamaquoddy', 'et': 'passamakodi', 'eu': 'maliseet-passamaquoddyera', 'fa': 'ملیسیت - پاسماکودی', 'ff-Adlm': '𞤃𞤢𞤤𞤭𞤧𞤫𞥅𞤼-𞤆𞤢𞤧𞤢𞤥𞤢𞤹𞤵𞤮𞤣𞥆𞤭', 'fi': 'maliseet-passamaquoddy', 'fil': 'Maliseet-Passamaquoddy', 'fr': 'malécite-passamaquoddy', 'ga': 'Maliseet-Passamaquoddy', 'gd': 'Maliseet-Passamaquoddy', 'gl': 'malecite-passamaquoddy', 'gu': 'મલિસીટ-પાસમાક્વોડ્ડી', 'ha': 'Maliseet-Passamaquoddy', 'he': 'מליסיט-פאסמקוודי', 'hi': 'मलेसीट-पासेमेक्वोडी', 'hi-Latn': 'Maliseet-Passamaquoddy', 'hr': 'maliseet-Passamaquoddy', 'hsb': 'maliseet-passamaquoddyšćina', 'ht': 'malécite-passamaquoddy', 'hu': 'maliseet-passamaquoddy', 'hy': 'մալեսիտ-պասամակվոդի', 'ia': 'malecite-passamaquoddy', 'id': 'Maliseet-Passamaquoddy', 'ig': 'Maliseet-Passamaquoddy', 'is': 'maliseet-passamaquoddy', 'it': 'malecite-passamaquoddy', 'ja': 'マリシート＝パサマコディ語', 'jv': 'Maliseet-Passamakuoddi', 'ka': 'მალისეტ-პასამაკვოდი', 'kk': 'малесит-пассамакводди тілі', 'kk-Cyrl': 'малесит-пассамакводди тілі', 'km': 'ម៉ាលីស៊ីត ប៉ាសាម៉ាខ្វូឌី', 'kn': 'ಮ್ಯಾಲಿಸೀಟ್-ಪಸ್ಸಾಮಕ್ವಾಡ್ಡಿ', 'ko': '말리시트 파사마쿼디어', 'kok': 'मालिसीत-पास्सामाखिडी', 'kok-Deva': 'मालिसीत-पास्सामाखिडी', 'ku': 'malecite-passamaquoddy', 'ky': 'малесит-пассамакуоддиче (индей тили)', 'lo': 'ມາລິຊີດ ພາສຊາມາໂຄດດີ', 'lt': 'Maliset-Pasamakvodžio', 'lv': 'malisetu-pasamakvodi', 'mai': 'मलिसीट-पैसमाकोडी', 'mi': 'Marahiti-Pehamakoare', 'mk': 'малиситски пасамакводски', 'ml': 'മലിസീറ്റ്-പസാമക്വുഡി', 'mn': 'малесит-пассамакводди', 'mr': 'मालीसेट-पासामाक्वाड्डी', 'ms': 'Maliseet-Passamaquoddy', 'my': 'မလိဇိ ပါစမ်မကွာဒီ', 'nb': 'maliseet-passamaquoddy', 'ne': 'मालिसीट पासामाक्वेडी', 'nl': 'Maliseet-Passamaquoddy', 'nn': 'maliseet-passamaquoddy', 'no': 'maliseet-passamaquoddy', 'or': 'ମାଲିସୀଟ-ପାସମକୁଅଡ୍ଡି', 'pa': 'ਮਾਲੀਸੇਟ-ਪਾਸਾਮਾਕਵੋਡੀ', 'pa-Guru': 'ਮਾਲੀਸੇਟ-ਪਾਸਾਮਾਕਵੋਡੀ', 'pcm': 'Malisiít Pasamákódí Lángwej', 'pl': 'malecite-passamaquoddy', 'ps': 'mpq', 'pt': 'malecite–passamaquoddy', 'qu': 'Maliseet-Passamaquoddy Simi', 'ro': 'maliseet-passamaquoddy', 'ru': 'малесит-пассамакводди', 'sc': 'malecite-passamaquoddy', 'sd': 'ماليسيٽ پاسماڪئوڊي', 'sd-Arab': 'ماليسيٽ پاسماڪئوڊي', 'si': 'මලිසීට්-පස්සමකුඩි', 'sk': 'maliseet-passamaquoddy', 'sl': 'maliseet-passamaquoddščina', 'so': 'Maliseet-Passamaquoddy', 'sq': 'malisit-pasamakuadisht', 'sr': 'малисепасамакводи', 'sr-Cyrl': 'малисепасамакводи', 'sr-Latn': 'malisepasamakvodi', 'sv': 'maliseet-passamaquoddy', 'sw': 'Kimaliseet-Passamaquoddy', 'ta': 'மலிசீட்-பஸ்ஸமகுவாடி', 'te': 'మలిసీట్-పస్సమాక్వొడ్డీ', 'th': 'มาเลไซท์-ปัสมาโควดี', 'ti': 'ማሊሲት-ፓሳማኳዲ', 'tk': 'malisit-passamakwodi dili', 'to': 'lea fakamaliseti-pasamakuoti', 'tr': 'Malisetçe-Passamaquoddy', 'uk': 'малесіт-пасамакводі', 'ur': 'مالیسیٹ پاساماکوڈی', 'uz': 'maliset-passamakvoddi', 'uz-Latn': 'maliset-passamakvoddi', 'vec': 'malesita pasamacuody', 'vi': 'Tiếng Maliseet-Passamaquoddy', 'yo': 'Èdè Maliseti-Pasamkodi', 'yue': '馬里塞-帕薩馬誇迪文', 'yue-Hans': '马里塞-帕萨马夸迪文', 'yue-Hant': '馬里塞-帕薩馬誇迪文', 'zh': '马利塞-帕萨马科迪语', 'zh-Hans': '马利塞-帕萨马科迪语', 'zh-Hant': '馬里希特帕薩瑪奎迪文', 'zu': 'Maliseet-Passamaquoddy'},
    'pqw': {'en': 'Western Malayo-Polynesian languages'},
    'pqw-pro': {'en': 'Proto-Western Malayo-Polynesian'},
    'pra': {'en': 'Prakrit languages'},
    'prb': {'en': "Lua'"},
    'prc': {'en': 'Parachi'},
    'prd': {'en': 'Parsi-Dari'},
    'pre': {'en': 'Principense'},
    'prf': {'en': 'Paranan'},
    'prg': {'af': 'Pruisies', 'ak': 'Prusia kasa', 'am': 'ፐሩሳንኛ', 'ar': 'البروسياوية', 'as': 'প্ৰুছিয়ান', 'ast': 'prusianu', 'az': 'pruss', 'az-Cyrl': 'прусс', 'az-Latn': 'pruss', 'bal-Latn': 'Prushiái', 'be': 'пруская', 'be-tarask': 'пруская', 'bg': 'пруски', 'bgn': 'پروسی', 'bn': 'প্রুশিয়ান', 'br': 'henbruseg', 'bs': 'pruski', 'bs-Cyrl': 'пруски', 'bs-Latn': 'pruski', 'ca': 'prussià', 'ca-ES-valencia': 'prussià', 'ccp': '𑄛𑄴𑄢𑄪𑄥𑄨𑄠𑄚𑄴', 'ce': 'пруссийн', 'chr': 'ᏡᏏᎠᏂ', 'ckb': 'پڕووسی', 'cs': 'pruština', 'cy': 'Prwseg', 'da': 'preussisk', 'de': 'Altpreußisch', 'dsb': 'prusčina', 'el': 'Πρωσικά', 'el-polyton': 'Πρωσικά', 'en': 'Prussian', 'es': 'prusiano', 'et': 'preisi', 'eu': 'prusiera', 'fa': 'پروسی', 'ff-Adlm': '𞤆𞤵𞤪𞤧𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'muinaispreussi', 'fil': 'Prussian', 'fo': 'prusslanskt', 'fr': 'prussien', 'ga': 'Prúisis', 'gd': 'Pruisis', 'gl': 'prusiano', 'gu': 'પ્રુસ્સીયન', 'ha': 'Ferusawa', 'he': 'פרוסית', 'hi': 'प्रुशियाई', 'hi-Latn': 'Prussian', 'hr': 'pruski', 'hsb': 'prušćina', 'ht': 'prussien', 'hu': 'porosz', 'hy': 'պրուսերեն', 'ia': 'prussiano', 'id': 'Prusia', 'ie': 'prussian', 'ig': 'Prụssịan', 'is': 'prússneska', 'it': 'prussiano', 'ja': 'プロシア語', 'jv': 'Prusia', 'ka': 'პრუსიული', 'kea': 'prusianu', 'kgp': 'prusijỹnũ', 'kk': 'пруссия тілі', 'kk-Cyrl': 'пруссия тілі', 'km': 'ព្រូស៊ាន', 'kn': 'ಪ್ರಶಿಯನ್', 'ko': '프러시아어', 'kok': 'प्रुसियन', 'kok-Deva': 'प्रुसियन', 'ku': 'prûsyayî', 'ky': 'пруссча', 'lb': 'Preisesch', 'lo': 'ປຣັສຊຽນ', 'lt': 'prūsų', 'lv': 'prūšu', 'mk': 'пруски', 'ml': 'പ്രഷ്യൻ', 'mn': 'прусс', 'mr': 'प्रुशियन', 'ms': 'Prusia', 'mt': 'Prussu', 'my': 'ပရူရှန်', 'nb': 'prøyssisk', 'ne': 'प्रसियाली', 'nl': 'Oudpruisisch', 'nn': 'prøyssisk', 'no': 'prøyssisk', 'nqo': 'ߔߎߙߛߌߞߊ߲', 'or': 'ପ୍ରୁସିୟ', 'pa': 'ਪਰੂਸ਼ੀਆ', 'pa-Guru': 'ਪਰੂਸ਼ੀਆ', 'pcm': 'Prúshia', 'pl': 'pruski', 'prg': 'prūsiskan', 'ps': 'پروشين', 'pt': 'prussiano', 'qu': 'Prusiano Simi', 'rm': 'prussian', 'ro': 'prusacă', 'ru': 'прусский', 'sc': 'prussianu', 'sd': 'پرشن', 'sd-Arab': 'پرشن', 'si': 'පෘශියන්', 'sk': 'pruština', 'sl': 'stara pruščina', 'smn': 'toovláš preussikielâ', 'so': 'Brashiyaan', 'sq': 'prusisht', 'sr': 'пруски', 'sr-Cyrl': 'пруски', 'sr-Latn': 'pruski', 'sv': 'fornpreussiska', 'sw': 'Kiprussia', 'ta': 'பிரஷ்யன்', 'te': 'ప్రష్యన్', 'th': 'ปรัสเซีย', 'ti': 'ፕሩስኛ', 'tk': 'prussiýa dili', 'to': 'lea fakapulūsia', 'tr': 'Prusyaca', 'uk': 'прусська', 'ur': 'پارسی', 'uz': 'pruss', 'uz-Latn': 'pruss', 'vec': 'prusian', 'vi': 'Tiếng Prussia', 'xnr': 'प्रुशियाई', 'yi': 'פּרייסיש', 'yo': 'Púrúṣíànù', 'yrl': 'purusianu', 'yue': '普魯士文', 'yue-Hans': '普鲁士文', 'yue-Hant': '普魯士文', 'zh': '普鲁士语', 'zh-Hans': '普鲁士语', 'zh-Hant': '普魯士文', 'zu': 'isi-Prussian'},
    'prh': {'en': 'Porohanon'},
    'pri': {'en': 'Paicî'},
    'prk': {'en': 'Parauk'},
    'prl': {'en': 'Peruvian Sign Language'},
    'prm': {'en': 'Kibiri'},
    'prn': {'en': 'Prasuni'},
    'pro': {'ar': 'البروفانسية القديمة', 'ast': 'provenzal antiguu', 'az': 'qədim provansal', 'az-Latn': 'qədim provansal', 'be': 'стараправансальская', 'be-tarask': 'стараправансальская', 'bg': 'старопровансалски', 'bn': 'প্রাচীন প্রোভেনসাল', 'br': 'henbrovañseg', 'brx': 'पुरानी प्रोवाँसाल', 'bs': 'staroprovansalski', 'bs-Cyrl': 'старопровансалски', 'bs-Latn': 'staroprovansalski', 'ca': 'provençal antic', 'ca-ES-valencia': 'provençal antic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄴 𑄛𑄳𑄢𑄮𑄞𑄬𑄚𑄴𑄥𑄣𑄴', 'cs': 'provensálština', 'cy': 'Hen Brofensaleg', 'da': 'oldprovencalsk', 'de': 'Altprovenzalisch', 'el': 'Παλαιά Προβανσάλ', 'el-polyton': 'Παλαιά Προβανσάλ', 'en': 'Old Provençal', 'es': 'provenzal antiguo', 'et': 'vanaprovansi', 'fa': 'پرووانسی باستان', 'fi': 'muinaisprovensaali', 'fr': 'provençal ancien', 'fur': 'vieri provenzâl', 'fy': 'Aldprovençaals', 'gd': 'Seann-Phrovençal', 'gsw': 'Altprovenzalisch', 'gu': 'જુની પ્રોવેન્સલ', 'he': 'פרובנסאל עתיקה', 'hi': 'पुरानी प्रोवेन्सल', 'hi-Latn': 'Old Provençal', 'hr': 'staroprovansalski', 'ht': 'provençal ancien', 'hu': 'óprovánszi', 'hy': 'հին պրովանսերեն', 'id': 'Provencal Lama', 'is': 'fornpróvensalska', 'it': 'provenzale antico', 'ja': '古期プロバンス語', 'ka': 'ძველი პროვანსული', 'kgp': 'provẽsar arkajku', 'kn': 'ಪ್ರಾಚೀನ ಪ್ರೊವೆನ್ಶಿಯಲ್', 'ko': '고대 프로방스어', 'ks': 'پرون پرووینچَل', 'ks-Arab': 'پرون پرووینچَل', 'lb': 'Alprovenzalesch', 'lo': 'ໂປວອງຊານໂບຮານ', 'lt': 'senovės provansalų', 'lv': 'senprovansiešu', 'mk': 'старопровансалски', 'ml': 'പഴയ പ്രൊവൻഷ്ൽ', 'mr': 'पुरातन प्रोव्हेन्सल', 'mt': 'Provenzal Antik', 'nb': 'gammelprovençalsk', 'ne': 'पुरातन प्रोभेन्काल', 'nl': 'Oudprovençaals', 'nn': 'gammalprovençalsk', 'no': 'gammelprovençalsk', 'or': 'ପୁରୁଣା ପ୍ରେଭେନେସିଆଲ୍', 'pl': 'staroprowansalski', 'pt': 'provençal arcaico', 'rm': 'provenzal vegl', 'ro': 'provensală veche', 'ru': 'старопровансальский', 'sk': 'stará okcitánčina', 'sl': 'stara provansalščina', 'sr': 'староокситански', 'sr-Cyrl': 'староокситански', 'sr-Latn': 'starooksitanski', 'sv': 'fornprovensalska', 'ta': 'பழைய ப்ரோவென்சால்', 'te': 'ప్రాచీన ప్రోవెంసాల్', 'th': 'โปรวองซาลโบราณ', 'to': 'lea fakapolovenisi-motuʻa', 'tr': 'Eski Provensal', 'ug': 'قەدىمكى پروۋېنچالچە', 'uk': 'давньопровансальська', 'vi': 'Tiếng Provençal cổ', 'yrl': 'puruwẽsau arkaiku', 'yue': '古普羅旺斯文', 'yue-Hans': '古普罗旺斯文', 'yue-Hant': '古普羅旺斯文', 'zh': '古普罗文斯语', 'zh-Hans': '古普罗文斯语', 'zh-Hant': '古普羅旺斯文'},
    'prp': {'en': 'Parsi'},
    'prq': {'en': 'Ashéninka Perené'},
    'prr': {'en': 'Puri'},
    'prs': {'en': 'Dari'},
    'prt': {'en': 'Phai'},
    'pru': {'en': 'Puragi'},
    'prw': {'en': 'Parawen'},
    'prx': {'en': 'Purik'},
    'pry': {'en': 'Pray 3'},
    'prz': {'en': 'Providencia Sign Language'},
    'ps': {'af': 'Pasjtoe', 'ak': 'Pahyito', 'am': 'ፓሽቶ', 'ar': 'البشتو', 'as': 'পুস্ত', 'ast': 'pashtu', 'az': 'puştu', 'az-Arab': 'پشتوجه', 'az-Cyrl': 'пушту', 'az-Latn': 'puştu', 'bal-Latn': 'Pashtó', 'be': 'пушту', 'be-tarask': 'пушту', 'bg': 'пущу', 'bgn': 'پشتو', 'bn': 'পাশতু', 'br': 'pachto', 'brx': 'पाश्त’', 'bs': 'paštu', 'bs-Cyrl': 'паштунски', 'bs-Latn': 'paštu', 'ca': 'paixtu', 'ca-ES-valencia': 'paixtu', 'ccp': '𑄛𑄌𑄴𑄑𑄪', 'ce': 'пушту', 'chr': 'ᏆᏍᏙ', 'ckb': 'پەشتوو', 'cs': 'paštština', 'cy': 'Pashto', 'da': 'pashto', 'de': 'Paschtu', 'dsb': 'paštunšćina', 'dz': 'པཱཤ་ཏོ་ཁ', 'ee': 'pashtogbe', 'el': 'Πάστο', 'el-polyton': 'Πάστο', 'en': 'Pashto', 'eo': 'paŝtua', 'es': 'pastún', 'et': 'puštu', 'eu': 'paxtunera', 'fa': 'پشتو', 'ff-Adlm': '𞤆𞤢𞤧𞤼𞤵𞤲𞤪𞤫', 'fi': 'paštu', 'fil': 'Pashto', 'fo': 'pashto', 'fr': 'pachto', 'fur': 'pashto', 'fy': 'Pasjtoe', 'ga': 'Paistis', 'gd': 'Pashto', 'gl': 'paxto', 'gsw': 'Paschtu', 'gu': 'પશ્તો', 'ha': 'Pashtanci', 'he': 'פאשטו', 'hi': 'पश्तो', 'hi-Latn': 'Pashto', 'hr': 'paštunski', 'hsb': 'paštunšćina', 'ht': 'pachto', 'hu': 'pastu', 'hy': 'փուշթու', 'ia': 'pashto', 'id': 'Pashto', 'ig': 'Pashọ', 'is': 'pastú', 'it': 'pashto', 'ja': 'パシュトゥー語', 'jv': 'Pashto', 'ka': 'პუშტუ', 'kea': 'paxto', 'kgp': 'pasito', 'kk': 'пушту тілі', 'kk-Cyrl': 'пушту тілі', 'km': 'បាស្តូ', 'kn': 'ಪಾಷ್ಟೋ', 'ko': '파슈토어', 'kok': 'पश्तो', 'kok-Deva': 'पश्तो', 'ks': 'پَشتوٗ', 'ks-Arab': 'پَشتوٗ', 'ksh': 'Paschtuunesch', 'ku': 'peştûyî', 'ky': 'пуштуча', 'lb': 'Paschtu', 'lkt': 'Pashto Iyápi', 'lld': 'pashto', 'lo': 'ປາສໂຕ', 'lrc': 'پأشتوٙ', 'lt': 'puštūnų', 'lv': 'puštu', 'mai': 'पश्तो', 'mi': 'Pāhitō', 'mk': 'паштунски', 'ml': 'പഷ്\u200cതോ', 'mn': 'пушту', 'mr': 'पश्तो', 'ms': 'Pashto', 'mt': 'Pashto', 'my': 'ပက်ရှ်တွန်း', 'mzn': 'پشتو', 'nb': 'pashto', 'ne': 'पास्तो', 'nl': 'Pasjtoe', 'nn': 'pashto', 'no': 'pashto', 'nqo': 'ߔߊߛߑߕߏߞߊ߲', 'or': 'ପାସ୍ତୋ', 'pa': 'ਪਸ਼ਤੋ', 'pa-Guru': 'ਪਸ਼ਤੋ', 'pcm': 'Páshto', 'pl': 'paszto', 'ps': 'پښتو', 'pt': 'pashto', 'qu': 'Pashto Simi', 'rm': 'paschto', 'ro': 'paștună', 'ru': 'пушту', 'rw': 'Impashito', 'sc': 'pashto', 'sd': 'پشتو', 'sd-Arab': 'پشتو', 'si': 'පෂ්ටො', 'sk': 'paštčina', 'sl': 'paštunščina', 'smn': 'paštu', 'so': 'Bashtuu', 'sq': 'pashtoisht', 'sr': 'паштунски', 'sr-Cyrl': 'паштунски', 'sr-Latn': 'paštunski', 'sv': 'afghanska', 'sw': 'Kipashto', 'ta': 'பஷ்தோ', 'te': 'పాష్టో', 'tg': 'пушту', 'th': 'พัชโต', 'ti': 'ፓሽቶ', 'tk': 'peştun dili', 'to': 'lea fakapasitō', 'tr': 'Peştuca', 'tt': 'пушту', 'ug': 'پۇشتۇچە', 'uk': 'пушту', 'ur': 'پشتو', 'uz': 'pushtu', 'uz-Arab': 'پشتو', 'uz-Cyrl': 'пушту', 'uz-Latn': 'pushtu', 'vec': 'pashto', 'vi': 'Tiếng Pashto', 'wae': 'Paštu', 'wo': 'Pasto', 'xnr': 'पश्तो', 'yi': 'פּאַשטאָ', 'yo': 'Páshítò', 'yrl': 'paxitu', 'yue': '普什圖文', 'yue-Hans': '普什图文', 'yue-Hant': '普什圖文', 'zh': '普什图语', 'zh-Hans': '普什图语', 'zh-Hant': '普什圖文', 'zu': 'isi-Pashto'},
    'psa': {'en': 'Asue Awyu'},
    'psc': {'en': 'Persian Sign Language'},
    'psd': {'en': 'Plains Indian Sign Language'},
    'pse': {'en': 'Central Malay'},
    'psg': {'en': 'Penang Sign Language'},
    'psh': {'en': 'Southwest Pashai'},
    'psi': {'en': 'Southeast Pashai'},
    'psl': {'en': 'Puerto Rican Sign Language'},
    'psm': {'en': 'Pauserna'},
    'psn': {'en': 'Panasuan'},
    'pso': {'en': 'Polish Sign Language'},
    'psp': {'en': 'Philippine Sign Language'},
    'psq': {'en': 'Pasi'},
    'psr': {'en': 'Portuguese Sign Language'},
    'pss': {'en': 'Kaulong'},
    'pst': {'en': 'Central Pashto'},
    'psu': {'en': 'Sauraseni Prākrit'},
    'psw': {'en': 'Port Sandwich'},
    'psy': {'en': 'Piscataway'},
    'pt': {'af': 'Portugees', 'agq': 'Kpotùwgîi', 'ak': 'Pɔɔtugal kasa', 'am': 'ፖርቹጋልኛ', 'ar': 'البرتغالية', 'as': 'পৰ্তুগীজ', 'asa': 'Kireno', 'ast': 'portugués', 'az': 'portuqal', 'az-Arab': 'پوْرتغالجه', 'az-Cyrl': 'португал', 'az-Latn': 'portuqal', 'bal-Latn': 'Portagáli', 'bas': 'Hɔp u pɔtɔ̄kì', 'be': 'партугальская', 'be-tarask': 'партугальская', 'bem': 'Ichi Potogisi', 'bez': 'Hileno', 'bg': 'португалски', 'bgn': 'پورتگالی', 'blo': 'gɩpɔrtigalja', 'bm': 'pɔritigalikan', 'bn': 'পর্তুগীজ', 'br': 'portugaleg', 'brx': 'पर्टुगिज', 'bs': 'portugalski', 'bs-Cyrl': 'португалски', 'bs-Latn': 'portugalski', 'ca': 'portuguès', 'ca-ES-valencia': 'portuguès', 'ccp': '𑄛𑄧𑄢𑄴𑄖𑄪𑄉𑄨𑄎𑄴', 'ce': 'португалихойн', 'ceb': 'Portuguese', 'cgg': 'Orupocugo', 'chr': 'ᏉᏧᎩᏍ', 'ckb': 'پورتوگالی', 'cs': 'portugalština', 'cv': 'португали', 'cy': 'Portiwgaleg', 'da': 'portugisisk', 'dav': 'Kireno', 'de': 'Portugiesisch', 'dje': 'Portugee senni', 'doi': 'पुर्तगाली', 'dsb': 'portugalšćina', 'dyo': 'portugees', 'dz': 'པོར་ཅུ་གིས་ཁ', 'ebu': 'Kĩreno', 'ee': 'Portuguesegbe', 'el': 'Πορτογαλικά', 'el-polyton': 'Πορτογαλικά', 'en': 'Portuguese', 'eo': 'portugala', 'es': 'portugués', 'et': 'portugali', 'eu': 'portugesa', 'ewo': 'ǹkɔ́bɔ fɔtugɛ́s', 'fa': 'پرتغالی', 'ff': 'Purtugeere', 'ff-Adlm': '𞤆𞤮𞤪𞤼𞤮𞤳𞤫𞥅𞤧𞤭𞥅𞤪𞤫', 'ff-Latn': 'Purtugeere', 'fi': 'portugali', 'fil': 'Portuguese', 'fo': 'portugiskiskt', 'fr': 'portugais', 'fur': 'portughês', 'fy': 'Portugeesk', 'ga': 'Portaingéilis', 'gd': 'Portagailis', 'gl': 'portugués', 'gsw': 'Portugiisisch', 'gu': 'પોર્ટુગીઝ', 'guz': 'Kireno', 'ha': 'Harshen Potugis', 'haw': 'Pukikī', 'he': 'פורטוגזית', 'hi': 'पुर्तगाली', 'hi-Latn': 'Portuguese', 'hr': 'portugalski', 'hsb': 'portugalšćina', 'ht': 'portugais', 'hu': 'portugál', 'hy': 'պորտուգալերեն', 'ia': 'portugese', 'id': 'Portugis', 'ie': 'portugalesi', 'ig': 'Pọrtụgụese', 'ii': 'ꁍꄨꑸꉙ', 'is': 'portúgalska', 'it': 'portoghese', 'ja': 'ポルトガル語', 'jmc': 'Kyireno', 'jv': 'Portugis', 'ka': 'პორტუგალიური', 'kab': 'Tapurtugalit', 'kam': 'Kireno', 'kde': 'Chileno', 'kea': 'purtuges', 'kgp': 'fóg-vĩ', 'khq': 'Portugee senni', 'ki': 'Kireno', 'kk': 'португал тілі', 'kk-Cyrl': 'португал тілі', 'kln': 'kutitab Portugal', 'km': 'ព័រទុយហ្គាល់', 'kn': 'ಪೋರ್ಚುಗೀಸ್', 'ko': '포르투갈어', 'kok': 'पोर्तुगीज', 'kok-Deva': 'पोर्तुगीज', 'ks': 'پُرتَگیٖز', 'ks-Arab': 'پُرتَگیٖز', 'ks-Deva': 'पुरतउगाली', 'ksb': 'Kileno', 'ksf': 'ripɔrtugɛ́', 'ksh': 'Pochtojesesch', 'ku': 'portugalî', 'kxv': 'portugīj', 'kxv-Deva': 'पुर्तगाली', 'kxv-Latn': 'portugīj', 'kxv-Orya': 'ପର୍ତୁଗୀଜ୍\u200c', 'kxv-Telu': 'పోర్తుగీస్', 'ky': 'португалча', 'lag': 'Kɨréeno', 'lb': 'Portugisesch', 'lg': 'Lupotugiizi', 'lkt': 'Portuguese Iyápi', 'lld': 'portoghesc', 'ln': 'lipulutugɛ́si', 'lo': 'ປອກຕຸຍກິສ', 'lrc': 'پورتئغالی', 'lt': 'portugalų', 'lu': 'Mputulugɛsi', 'luo': 'Kireno', 'luy': 'Kireno', 'lv': 'portugāļu', 'mai': 'पुर्तगाली', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lportuguese', 'mer': 'Kĩpochogo', 'mfe': 'portige', 'mg': 'Portiogey', 'mgh': 'Nreno', 'mi': 'Pōtukīhi', 'mk': 'португалски', 'ml': 'പോർച്ചുഗീസ്', 'mn': 'португал', 'mni': 'পোর্টুগিজ', 'mni-Beng': 'পোর্টুগিজ', 'mr': 'पोर्तुगीज', 'ms': 'Portugis', 'mt': 'Portugiż', 'mua': 'Zah sǝr Portugal', 'my': 'ပေါ်တူဂီ', 'mzn': 'پرتغالی', 'naq': 'Portugees', 'nb': 'portugisisk', 'nd': 'isi-Potukezi', 'ne': 'पोर्तुगी', 'nl': 'Portugees', 'nmg': 'Kiɛl bó pɔ̄rtugɛ̂', 'nn': 'portugisisk', 'no': 'portugisisk', 'nqo': 'ߔߕߏ߬ߞߌ߬ߞߊ߲', 'nus': 'Thok puɔtigali', 'nyn': 'Orupocugo', 'om': 'Afaan Porchugaal', 'or': 'ପର୍ତ୍ତୁଗୀଜ୍\u200c', 'os': 'португалиаг', 'pa': 'ਪੁਰਤਗਾਲੀ', 'pa-Guru': 'ਪੁਰਤਗਾਲੀ', 'pcm': 'Pọtiugiz', 'pl': 'portugalski', 'ps': 'پورتګالي', 'pt': 'português', 'qu': 'Portugues Simi', 'rm': 'portugais', 'rn': 'Igiporutugari', 'ro': 'portugheză', 'rof': 'Kireno', 'ru': 'португальский', 'rw': 'Igiporutugali', 'rwk': 'Kyireno', 'sa': 'पुर्तगालदेशीय भाषा:', 'sah': 'Португааллыы', 'saq': 'Kireno', 'sat': 'ᱯᱩᱨᱛᱜᱟᱞᱤ', 'sat-Olck': 'ᱯᱩᱨᱛᱜᱟᱞᱤ', 'sbp': 'Ishileno', 'sc': 'portoghesu', 'sd': 'پورٽگليز', 'sd-Arab': 'پورٽگليز', 'sd-Deva': 'पुर्तगाली', 'se': 'portugálagiella', 'seh': 'português', 'ses': 'Portugee senni', 'sg': 'Portugëe, Pûra', 'shi': 'ⵜⴰⴱⵕⵟⵇⵉⵣⵜ', 'shi-Latn': 'Tabṛṭqizt', 'shi-Tfng': 'ⵜⴰⴱⵕⵟⵇⵉⵣⵜ', 'si': 'පෘතුගීසි', 'sk': 'portugalčina', 'sl': 'portugalščina', 'smn': 'portugalkielâ', 'sn': 'chiPutukezi', 'so': 'Boortaqiis', 'sq': 'portugalisht', 'sr': 'португалски', 'sr-Cyrl': 'португалски', 'sr-Latn': 'portugalski', 'su': 'Portugis', 'su-Latn': 'Portugis', 'sv': 'portugisiska', 'sw': 'Kireno', 'syr': 'ܦܘܪܛܘܓܠܐܝܬ', 'ta': 'போர்ச்சுகீஸ்', 'te': 'పోర్చుగీస్', 'teo': 'Kireno', 'tg': 'португалӣ', 'th': 'โปรตุเกส', 'ti': 'ፖርቱጊዝኛ', 'tk': 'portugal dili', 'to': 'lea fakapotukali', 'tpi': 'Potigis', 'tr': 'Portekizce', 'tt': 'португал', 'twq': 'Portugee senni', 'tzm': 'Taburtuɣalit', 'ug': 'پورتۇگالچە', 'uk': 'португальська', 'ur': 'پُرتگالی', 'uz': 'portugalcha', 'uz-Cyrl': 'португалча', 'uz-Latn': 'portugalcha', 'vai': 'ꕶꕿꕃꔤ', 'vai-Latn': 'Potokíi', 'vai-Vaii': 'ꕶꕿꕃꔤ', 'vec': 'portogheze', 'vi': 'Tiếng Bồ Đào Nha', 'vun': 'Kyireno', 'wae': 'Portugisiš', 'wo': 'Purtugees', 'xh': 'IsiPhuthukezi', 'xnr': 'पुर्तगाली', 'xog': 'Olupotugiizi', 'yav': 'nupɔlitukɛ́ɛ', 'yi': 'פּארטוגעזיש', 'yo': 'Èdè Pọtogí', 'yrl': 'putugei', 'yue': '葡萄牙文', 'yue-Hans': '葡萄牙文', 'yue-Hant': '葡萄牙文', 'zgh': 'ⵜⴰⴱⵕⵟⵇⵉⵣⵜ', 'zh': '葡萄牙语', 'zh-Hans': '葡萄牙语', 'zh-Hant': '葡萄牙文', 'zu': 'isi-Portuguese'},
    'pt-BR': {'am': 'የብራዚል ፖርቹጋልኛ', 'ar': 'البرتغالية البرازيلية', 'as': 'ব্ৰাজিলিয়ান পৰ্তুগীজ', 'ast': 'portugués del Brasil', 'az': 'Braziliya portuqalcası', 'az-Cyrl': 'Бразилија португалҹасы', 'az-Latn': 'Braziliya portuqalcası', 'be': 'бразільская партугальская', 'be-tarask': 'бразільская партугальская', 'bgn': 'برازیلین پورتگالی', 'br': 'portugaleg Brazil', 'brx': 'ब्राजिलनि पर्टुगिज', 'ca': 'portuguès del Brasil', 'ca-ES-valencia': 'portuguès del Brasil', 'ccp': '𑄝𑄳𑄢𑄎𑄨𑄣𑄬𑄢𑄴 𑄛𑄧𑄢𑄴𑄖𑄪𑄉𑄨𑄎𑄴', 'ce': 'бразилин португалихойн', 'chr': 'ᏆᏏᎵᎢ ᏉᏧᎩᏍ', 'cv': 'бразилин португали', 'cy': 'Portiwgaleg Brasil', 'da': 'brasiliansk portugisisk', 'doi': 'ब्राजीली पुर्तगाली', 'dsb': 'brazilska portugalšćina', 'dz': 'བྲ་ཛི་ལི་ཡཱན་པོར་ཅུ་གིས་ཁ', 'ee': 'Portuguesegbe (Brazil)', 'el': 'Πορτογαλικά Βραζιλίας', 'el-polyton': 'Πορτογαλικά Βραζιλίας', 'en': 'Brazilian Portuguese', 'eo': 'portugala brazila', 'es': 'portugués de Brasil', 'et': 'Brasiilia portugali', 'eu': 'Brasilgo portugesa', 'fa': 'پرتغالی برزیل', 'ff-Adlm': '𞤆𞤮𞤪𞤼𞤮𞤳𞤫𞥅𞤧𞤭𞥅𞤪𞤫 𞤄𞤪𞤫𞥁𞤭𞤤', 'fi': 'brasilianportugali', 'fil': 'Portuges ng Brasil', 'fo': 'portugiskiskt (Brasilia)', 'fr': 'portugais brésilien', 'fur': 'portughês brasilian', 'fy': 'Brazyljaansk Portugees', 'ga': 'Portaingéilis Bhrasaíleach', 'gd': 'Portagailis Bhraisileach', 'gl': 'portugués do Brasil', 'gsw': 'Brasilianischs Portugiisisch', 'gu': 'બ્રાઝિલીયન પોર્ટુગીઝ', 'ha': 'Harshen Potugis na Birazil', 'haw': 'Pukikī Palakila', 'hi': 'ब्राज़ीली पुर्तगाली', 'hi-Latn': 'Brazilian Portuguese', 'hr': 'brazilski portugalski', 'hsb': 'brazilska portugalšćina', 'ht': 'portugais brésilien', 'hu': 'brazíliai portugál', 'hy': 'բրազիլական պորտուգալերեն', 'ia': 'portugese de Brasil', 'ie': 'brasilian portugalesi', 'ig': 'Asụsụ Portugese ndị Brazil', 'is': 'brasílísk portúgalska', 'it': 'portoghese brasiliano', 'jv': 'Portugis Brasil', 'ka': 'ბრაზილიური პორტუგალიური', 'kea': 'purtuges brazileru', 'kk': 'бразилиялық португал тілі', 'kk-Cyrl': 'бразилиялық португал тілі', 'km': 'ព័រទុយហ្កាល់ (ប្រេស៊ីល)', 'kn': 'ಬ್ರೆಜಿಲಿಯನ್ ಪೋರ್ಚುಗೀಸ್', 'kok': 'ब्राझिलियन पोर्तुगीज', 'kok-Deva': 'ब्राझिलियन पोर्तुगीज', 'ks': 'برازیٖلی پُرتَگیٖز', 'ks-Arab': 'برازیٖلی پُرتَگیٖز', 'ks-Deva': 'ब्राज़िली पुरतउगाली', 'ksh': 'Brasilljaanesch Pochtojesesch', 'kxv': 'brajilian portugīj', 'kxv-Deva': 'ब्राज़ीली पुर्तगाली', 'kxv-Latn': 'brajilian portugīj', 'kxv-Orya': 'ବ୍ରାଜିଲିଆନ୍ ପର୍ତୁଗୀଜ୍', 'kxv-Telu': 'బ్రెజిలియన్ పోర్తుగీస్', 'lb': 'Brasilianescht Portugisesch', 'lrc': 'پورتئغالی بئرئزیل', 'mai': 'ब्राज़ीली पुर्तगाली', 'mi': 'Pōtukīhi Parahi', 'mk': 'бразилски португалски', 'ml': 'ബ്രസീലിയൻ പോർച്ചുഗീസ്', 'mn': 'португал хэл (Бразил)', 'mni': 'ব্রাজিলিয়ান পোর্টুগিজ', 'mni-Beng': 'ব্রাজিলিয়ান পোর্টুগিজ', 'mr': 'ब्राझिलियन पोर्तुगीज', 'ms': 'Portugis Brazil', 'mt': 'Portugiż tal-Brażil', 'my': 'ဘရာဇီး ပေါ်တူဂီ', 'mzn': 'برزیل ِپرتغالی', 'ne': 'ब्राजिली पोर्तुगी', 'nqo': 'ߓߙߋߖ߭ߌߟ ߔߕߏ߬ߞߌ߬ߞߊ߲', 'om': 'Afaan Portugali (Braazil)', 'or': 'ବ୍ରାଜିଲିଆନ୍ ପର୍ତ୍ତୁଗୀଜ୍', 'os': 'бразилиаг португалиаг', 'pa': 'ਪੁਰਤਗਾਲੀ (ਬ੍ਰਾਜ਼ੀਲੀ)', 'pa-Guru': 'ਪੁਰਤਗਾਲੀ (ਬ੍ਰਾਜ਼ੀਲੀ)', 'pcm': 'Brazíl Pọtiugíz', 'pl': 'brazylijski portugalski', 'ps': 'برازیلي پرتګالي', 'rm': 'portugais brasilian', 'ru': 'бразильский португальский', 'rw': 'Inyeporutigali (Brezili)', 'sa': 'ब्राज़ीली पुर्तगालदेशीय भाषा:', 'sat': 'ᱵᱨᱟᱡᱤᱞᱤᱭᱟᱱ ᱯᱩᱨᱛᱜᱟᱞᱤ', 'sat-Olck': 'ᱵᱨᱟᱡᱤᱞᱤᱭᱟᱱ ᱯᱩᱨᱛᱜᱟᱞᱤ', 'sc': 'portoghesu brasilianu', 'sd': 'برازيلي پرتگالي', 'sd-Arab': 'برازيلي پرتگالي', 'sd-Deva': 'ब्राज़ीलियन पुर्तगाली', 'si': 'බ්\u200dරසීල පෘතුගීසි', 'sk': 'portugalčina (brazílska)', 'sl': 'brazilska portugalščina', 'smn': 'Brasilia portugalkielâ', 'so': 'Boortaqiiska Baraasiil', 'sq': 'portugalishte braziliane', 'su': 'Portugis Brasil', 'su-Latn': 'Portugis Brasil', 'sv': 'brasiliansk portugisiska', 'sw': 'Kireno (Brazili)', 'ta': 'பிரேசிலிய போர்ச்சுகீஸ்', 'te': 'బ్రెజీలియన్ పోర్చుగీస్', 'tg': 'португалии бразилиягӣ', 'th': 'โปรตุเกส - บราซิล', 'tpi': 'Brasilien Potigis', 'tr': 'Brezilya Portekizcesi', 'ug': 'بىرازىلىيە پورتۇگالچە', 'ur': 'برازیلی پرتگالی', 'uz': 'portugal (Braziliya)', 'uz-Latn': 'portugal (Braziliya)', 'vec': 'portogheze braziłian', 'wae': 'Brasilianišes Portugisiš', 'wo': 'Purtugees bu Bresil', 'xh': 'IsiPhuthukezi SaseBrazil', 'xnr': 'ब्राज़ीली पुर्तगाली', 'yo': 'Èdè Pọtogí (Orilẹ̀-èdè Bràsíl)', 'zh': '巴西葡萄牙语', 'zh-Hans': '巴西葡萄牙语', 'zu': 'isi-Brazillian Portuguese'},
    'pt-PT': {'am': 'የአውሮፓ ፖርቹጋልኛ', 'ar': 'البرتغالية الأوروبية', 'as': 'ইউৰোপীয়ান পৰ্তুগীজ', 'ast': 'portugués européu', 'az': 'Portuqaliya portuqalcası', 'az-Cyrl': 'Португалија португалҹасы', 'az-Latn': 'Portuqaliya portuqalcası', 'be': 'еўрапейская партугальская', 'be-tarask': 'еўрапейская партугальская', 'bgn': 'اوروپایین پورتگالی', 'bn': 'পর্তুগীজ (ইউরোপ)', 'br': 'portugaleg Europa', 'brx': 'युरोपनि पर्टुगिज', 'ca': 'portuguès de Portugal', 'ca-ES-valencia': 'portuguès de Portugal', 'ccp': '𑄃𑄨𑄃𑄪𑄢𑄮𑄛𑄬𑄢𑄴 𑄛𑄧𑄢𑄴𑄖𑄪𑄉𑄨𑄎𑄴', 'ce': 'европан португалихойн', 'chr': 'ᏳᎳᏈ ᏉᏧᎩᏍ', 'cs': 'portugalština (Evropa)', 'cv': 'европӑн португали', 'cy': 'Portiwgaleg Ewrop', 'da': 'europæisk portugisisk', 'doi': 'यूरोपी पुर्तगाली', 'dsb': 'europejska portugalšćina', 'dz': 'ཨི་བེ་རི་ཡཱན་པོར་ཅུ་གིས་ཁ', 'ee': 'Portuguesegbe (Europe)', 'el': 'Πορτογαλικά Ευρώπης', 'el-polyton': 'Πορτογαλικά Ευρώπης', 'en': 'European Portuguese', 'eo': 'portugala eŭropa', 'es': 'portugués de Portugal', 'et': 'Euroopa portugali', 'eu': 'Europako portugesa', 'fa': 'پرتغالی اروپا', 'ff-Adlm': '𞤆𞤮𞤪𞤼𞤮𞤳𞤫𞥅𞤧𞤭𞥅𞤪𞤫 𞤆𞤮𞤪𞤼𞤭𞤺𞤢𞥄𞤤', 'fi': 'euroopanportugali', 'fil': 'European Portuguese', 'fo': 'portugiskiskt (Evropa)', 'fr': 'portugais européen', 'fur': 'portughês iberic', 'fy': 'Europees Portugees', 'ga': 'Portaingéilis Ibéarach', 'gd': 'Portagailis Eòrpach', 'gl': 'portugués de Portugal', 'gsw': 'Iberischs Portugiisisch', 'gu': 'યુરોપિયન પોર્ટુગીઝ', 'ha': 'Potugis Ƙasashen Turai', 'hi': 'यूरोपीय पुर्तगाली', 'hi-Latn': 'European Portuguese', 'hr': 'europski portugalski', 'hsb': 'europska portugalšćina', 'ht': 'portugais européen', 'hu': 'európai portugál', 'hy': 'եվրոպական պորտուգալերեն', 'ia': 'portugese de Portugal', 'id': 'Portugis (Eropa)', 'ie': 'europan portugalesi', 'ig': 'Asụsụ Portuguese ndị Europe', 'is': 'evrópsk portúgalska', 'it': 'portoghese europeo', 'ja': 'ポルトガル語 (イベリア半島)', 'jv': 'Portugis Portugal', 'ka': 'ევროპული პორტუგალიური', 'kea': 'purtuges europeu', 'kk': 'еуропалық португал тілі', 'kk-Cyrl': 'еуропалық португал тілі', 'km': 'ព័រទុយហ្គាល់ (អឺរ៉ុប)', 'kn': 'ಯೂರೋಪಿಯನ್ ಪೋರ್ಚುಗೀಸ್', 'ko': '포르투갈어(유럽)', 'kok': 'युरोपियन पोर्तुगीज', 'kok-Deva': 'युरोपियन पोर्तुगीज', 'ks': 'یوٗرپی پُرتَگیٖز', 'ks-Arab': 'یوٗرپی پُرتَگیٖز', 'ks-Deva': 'यूरपी पुरतउगाली', 'ksh': 'Pochtojesesch uß Pochtojall', 'ku': 'portugalî (Ewropa)', 'kxv': 'yuropīyati portugīj', 'kxv-Deva': 'यूरोपीय पुर्तगाली', 'kxv-Latn': 'yuropīyati portugīj', 'kxv-Orya': 'ୟୁରପ୍ ତି ପର୍ତୁଗୀଜ୍', 'kxv-Telu': 'యురోపియన్ పోర్తుగిస్', 'ky': 'португалча (Европа)', 'lb': 'Europäescht Portugisesch', 'lo': 'ປອກຕຸຍກິສ ຢຸໂຣບ', 'lrc': 'پورتئغالی ئوروٙپایی', 'mai': 'यूरोपीय पुर्तगाली', 'mi': 'Pōtukīhi Uropi', 'mk': 'португалски (Европа)', 'ml': 'യൂറോപ്യൻ പോർച്ചുഗീസ്', 'mn': 'португал хэл (Европ)', 'mni': 'য়ুরোপিয়ান পোর্টুগিজ', 'mni-Beng': 'য়ুরোপিয়ান পোর্টুগিজ', 'mr': 'युरोपियन पोर्तुगीज', 'ms': 'Portugis Eropah', 'mt': 'Portugiż Ewropew', 'my': 'ဥရောပ ပေါ်တူဂီ', 'mzn': 'اروپای ِپرتغالی', 'ne': 'युरोपेली पोर्तुगी', 'nqo': 'ߋߙߐߔߎ߬ ߔߕߏ߬ߞߌ߬ߞߊ߲', 'om': 'Afaan Protuguese', 'or': 'ୟୁରୋପୀୟ ପର୍ତ୍ତୁଗୀଜ୍\u200c', 'os': 'европӕйаг полтугалиаг', 'pa': 'ਪੁਰਤਗਾਲੀ (ਯੂਰਪੀ)', 'pa-Guru': 'ਪੁਰਤਗਾਲੀ (ਯੂਰਪੀ)', 'pcm': 'Yúróp Pọtiugíz', 'pl': 'europejski portugalski', 'ps': 'اروپايي پرتګالي', 'rm': 'portugais europeic', 'ro': 'portugheză (Europa)', 'ru': 'европейский португальский', 'rw': 'Inyeporutigali (Igiporutigali)', 'sa': 'फिरङ्गिन् पुर्तगालदेशीय भाषा:', 'sat': 'ᱭᱩᱨᱚᱯᱤᱭᱟᱱ ᱯᱩᱨᱛᱜᱟᱞᱤ', 'sat-Olck': 'ᱭᱩᱨᱚᱯᱤᱭᱟᱱ ᱯᱩᱨᱛᱜᱟᱞᱤ', 'sc': 'portoghesu europeu', 'sd': 'يورپي پرتگالي', 'sd-Arab': 'يورپي پرتگالي', 'sd-Deva': 'यूरोपी पुर्तगाली', 'si': 'යුරෝපීය පෘතුගීසි', 'sk': 'portugalčina (európska)', 'sl': 'evropska portugalščina', 'smn': 'Portugal portugalkielâ', 'so': 'Boortaqiiska Yurub', 'sq': 'portugalishte evropiane', 'sr': 'португалски (Португал)', 'sr-Cyrl': 'португалски (Португал)', 'sr-Latn': 'portugalski (Portugal)', 'su': 'Portugis Éropa', 'su-Latn': 'Portugis Éropa', 'sv': 'europeisk portugisiska', 'sw': 'Kireno (Ulaya)', 'ta': 'ஐரோப்பிய போர்ச்சுகீஸ்', 'te': 'యూరోపియన్ పోర్చుగీస్', 'tg': 'португалии аврупоӣ', 'th': 'โปรตุเกส - ยุโรป', 'tk': 'portugal dili (Ýewropa)', 'tpi': 'Yurop Potigis', 'tr': 'Avrupa Portekizcesi', 'tt': 'португал (Европа)', 'ug': 'ياۋروپا پورتۇگالچە', 'ur': 'یورپی پرتگالی', 'uz': 'portugal (Yevropa)', 'uz-Latn': 'portugal (Yevropa)', 'vec': 'portogheze europèo', 'vi': 'Tiếng Bồ Đào Nha (Châu Âu)', 'wae': 'Iberišes Portugisiš', 'wo': 'Portugees bu Tugël', 'xh': 'IsiPhuthukezi SasePortugal', 'xnr': 'यूरोपीय पुर्तगाली', 'yo': 'Èdè Pọtogí (orílẹ̀-èdè Yúróòpù)', 'zh': '欧洲葡萄牙语', 'zh-Hans': '欧洲葡萄牙语', 'zu': 'isi-European Portuguese'},
    'pta': {'en': 'Pai Tavytera'},
    'pth': {'en': 'Pataxó Hã-Ha-Hãe'},
    'pti': {'en': 'Pindiini'},
    'ptn': {'en': 'Patani'},
    'pto': {'en': "Zo'é"},
    'ptp': {'en': 'Patep'},
    'ptq': {'en': 'Pattapu'},
    'ptr': {'en': 'Piamatsina'},
    'ptt': {'en': 'Enrekang'},
    'ptu': {'en': 'Bambam'},
    'ptv': {'en': 'Port Vato'},
    'ptw': {'en': 'Pentlatch'},
    'pty': {'en': 'Pathiya'},
    'pua': {'en': 'Western Highland Purepecha'},
    'pub': {'en': 'Purum'},
    'puc': {'en': 'Punan Merap'},
    'pud': {'en': 'Punan Aput'},
    'pue': {'en': 'Puelche'},
    'puf': {'en': 'Punan Merah'},
    'pug': {'en': 'Phuie'},
    'pui': {'en': 'Puinave'},
    'puj': {'en': 'Punan Tubu'},
    'puk': {'en': 'Pu Ko'},
    'pum': {'en': 'Puma'},
    'puo': {'en': 'Puoc'},
    'pup': {'en': 'Pulabu'},
    'puq': {'en': 'Puquina'},
    'pur': {'en': 'Puruborá'},
    'put': {'en': 'Putoh'},
    'puu': {'en': 'Punu'},
    'puw': {'en': 'Puluwatese'},
    'pux': {'en': 'Puare'},
    'puy': {'en': 'Purisimeño'},
    'puz': {'en': 'Purum Naga'},
    'pwa': {'en': 'Pawaia'},
    'pwb': {'en': 'Panawa'},
    'pwg': {'en': 'Gapapaiwa'},
    'pwi': {'en': 'Patwin'},
    'pwm': {'en': 'Molbog'},
    'pwn': {'en': 'Paiwan'},
    'pwo': {'en': 'Pwo Western Karen'},
    'pwr': {'en': 'Powari'},
    'pww': {'en': 'Pwo Northern Karen'},
    'pxm': {'en': 'Quetzaltepec Mixe'},
    'pye': {'en': 'Pye Krumen'},
    'pym': {'en': 'Fyam'},
    'pyn': {'en': 'Poyanáwa'},
    'pys': {'en': 'Paraguayan Sign Language'},
    'pyu': {'en': 'Puyuma'},
    'pyx': {'en': 'Pyu (Myanmar)'},
    'pyy': {'en': 'Pyen'},
    'pzn': {'en': 'Para Naga'},
    'qaa..qtz': {'en': 'Private use'},
    'qfa-cka-pro': {'en': 'Proto-Chukotko-Kamchatkan'},
    'qfa-ctc-cat': {'en': 'Catacao'},
    'qfa-ctc-col': {'en': 'Colán'},
    'qfa-hur-pro': {'en': 'Proto-Hurro-Urartian'},
    'qfa-kor-pro': {'en': 'Proto-Korean'},
    'qfa-len-slv': {'en': 'Salvadoran Lenca'},
    'qfa-sca-pro': {'en': 'Proto-Siouan-Catawban'},
    'qfa-tak-pro': {'en': 'Proto-Tai-Kadai'},
    'qfa-tot-pro': {'en': 'Proto-Totozoquean'},
    'qfa-und-esm': {'en': 'Esmeralda'},
    'qfa-yen-pro': {'en': 'Proto-Yeniseian'},
    'qot': {'en': 'Sahaptin'},
    'qu': {'af': 'Quechua', 'ak': 'Kwɛkya', 'am': 'ኩዌቹዋ', 'ar': 'كيشوا', 'as': 'কুৱেচুৱা', 'ast': 'quechua', 'az': 'keçua', 'az-Cyrl': 'кечуа', 'az-Latn': 'keçua', 'bal-Latn': 'Kwichu', 'be': 'кечуа', 'be-tarask': 'кечуа', 'bg': 'кечуа', 'bgn': 'کچوایی', 'bn': 'কেচুয়া', 'br': 'kechuaeg', 'brx': 'क्वेचुआ', 'bs': 'kečua', 'bs-Cyrl': 'квенча', 'bs-Latn': 'kečua', 'ca': 'quítxua', 'ca-ES-valencia': 'quítxua', 'ccp': '𑄇𑄬𑄌𑄪𑄠', 'ce': 'кечуа', 'chr': 'ᎨᏧᏩ', 'ckb': 'کێچوا', 'cs': 'kečuánština', 'cy': 'Quechua', 'da': 'quechua', 'de': 'Quechua', 'dsb': 'kečua', 'dz': 'ཀྭེ་ཆུ་ཨ་ཁ', 'ee': 'kwetsuagbe', 'el': 'Κέτσουα', 'el-polyton': 'Κέτσουα', 'en': 'Quechua', 'eo': 'keĉua', 'es': 'quechua', 'et': 'ketšua', 'eu': 'kitxua', 'fa': 'کچوایی', 'ff-Adlm': '𞤗𞤵𞤷𞤵𞤢𞤲𞤪𞤫', 'fi': 'ketšua', 'fil': 'Quechua', 'fo': 'quechua', 'fr': 'quechua', 'fur': 'quechua', 'fy': 'Quechua', 'ga': 'Ceatsuais', 'gd': 'Quechua', 'gl': 'quechua', 'gsw': 'Quechua', 'gu': 'ક્વેચુઆ', 'ha': 'Quechua', 'he': 'קצ׳ואה', 'hi': 'क्वेचुआ', 'hi-Latn': 'Quechua', 'hr': 'kečuanski', 'hsb': 'kečua', 'ht': 'quechua', 'hu': 'kecsua', 'hy': 'կեչուա', 'ia': 'quechua', 'id': 'Quechua', 'ig': 'Asụsụ Quechua', 'is': 'kvesjúa', 'it': 'quechua', 'ja': 'ケチュア語', 'jv': 'Quechua', 'ka': 'კეჩუა', 'kea': 'kexua', 'kgp': 'kinsuva', 'kk': 'кечуа тілі', 'kk-Cyrl': 'кечуа тілі', 'km': 'ហ្គិកឈួ', 'kn': 'ಕ್ವೆಚುವಾ', 'ko': '케추아어', 'kok': 'क्वेच्वा', 'kok-Deva': 'क्वेच्वा', 'ks': 'کُویشُوا', 'ks-Arab': 'کُویشُوا', 'ksh': 'Kättschowa', 'ku': 'keçwayî', 'ky': 'кечуача', 'lb': 'Quechua', 'lkt': 'Quechua Iyápi', 'lld': 'quechua', 'lo': 'ຄີຊົວ', 'lrc': 'کوچوٙا', 'lt': 'kečujų', 'lv': 'kečvu', 'mai': 'कुएचुआ', 'mi': 'Kētua', 'mk': 'кечуански', 'ml': 'ക്വെച്ചുവ', 'mn': 'кечуа', 'mr': 'क्वेचुआ', 'ms': 'Quechua', 'mt': 'Quechua', 'my': 'ခီချူဝါအိုဝါ', 'mzn': 'قوئچوئا', 'nb': 'quechua', 'ne': 'क्वेचुवा', 'nl': 'Quechua', 'nn': 'quechua', 'no': 'quechua', 'nqo': 'ߞߋߛߎߥߊߞߊ߲', 'or': 'କ୍ୱେଚୁଆ', 'pa': 'ਕਕੇਸ਼ੁਆ', 'pa-Guru': 'ਕਕੇਸ਼ੁਆ', 'pcm': 'Kẹchuá', 'pl': 'keczua', 'ps': 'کېچوا', 'pt': 'quíchua', 'qu': 'Runasimi', 'rm': 'quechua', 'ro': 'quechua', 'ru': 'кечуа', 'sc': 'quechua', 'sd': 'ڪيچوا', 'sd-Arab': 'ڪيچوا', 'si': 'ක්වීචුවා', 'sk': 'kečuánčina', 'sl': 'kečuanščina', 'smn': 'quechua', 'so': 'Quechua', 'sq': 'keçuaisht', 'sr': 'кечуа', 'sr-Cyrl': 'кечуа', 'sr-Latn': 'kečua', 'sv': 'quechua', 'sw': 'Kikechua', 'ta': 'க்வெச்சுவா', 'te': 'కెచువా', 'tg': 'кечуа', 'th': 'เคชวา', 'ti': 'ቀችዋ', 'tk': 'keçua dili', 'to': 'lea fakakuetisa', 'tr': 'Keçuva dili', 'tt': 'кечуа', 'ug': 'كېچىۋاچە', 'uk': 'кечуа', 'ur': 'کویچوآ', 'uz': 'kechua', 'uz-Cyrl': 'кечуа', 'uz-Latn': 'kechua', 'vec': 'cuechua', 'vi': 'Tiếng Quechua', 'wae': 'Quečua', 'wo': 'Kesuwa', 'xnr': 'क्वेचुआ', 'yo': 'Kúẹ́ńjùà', 'yrl': 'kíxua', 'yue': '蓋楚瓦文', 'yue-Hans': '盖楚瓦文', 'yue-Hant': '蓋楚瓦文', 'zh': '克丘亚语', 'zh-Hans': '克丘亚语', 'zh-Hant': '蓋楚瓦文', 'zu': 'isi-Quechua'},
    'qua': {'en': 'Quapaw'},
    'qub': {'en': 'Huallaga Huánuco Quechua'},
    'quc': {'af': 'K’iche’', 'am': 'ኪቼ', 'ar': 'الكيشية', 'as': 'কিচিয়ে', 'ast': 'kʼicheʼ', 'az': 'kiçe', 'az-Cyrl': 'киче', 'az-Latn': 'kiçe', 'bal-Latn': 'Kichi', 'be': 'кічэ', 'be-tarask': 'кічэ', 'bg': 'киче', 'bgn': 'کیچه\u200c', 'bn': 'কি‘চে', 'br': 'kʼicheʼ', 'bs': 'kiče', 'bs-Latn': 'kiče', 'ca': 'k’iche’', 'ca-ES-valencia': 'k’iche’', 'ccp': '𑄇𑄳𑄦𑄨𑄌𑄬', 'ce': 'киче', 'chr': 'ᎩᏤ', 'ckb': 'کیچەیی', 'cs': 'kičé', 'cy': 'K’iche’', 'da': 'quiché', 'de': 'K’iche’', 'dsb': 'kʼicheʼ', 'el': 'Κιτσέ', 'el-polyton': 'Κιτσέ', 'en': 'Kʼicheʼ', 'es': 'quiché', 'et': 'kitše', 'eu': 'quicheera', 'fa': 'کیچه\u200c', 'fi': 'kʼicheʼ', 'fil': 'Kʼicheʼ', 'fo': 'kʼicheʼ', 'fr': 'quiché', 'ga': 'Cuitséis', 'gd': 'K’iche’', 'gl': 'quiché', 'gu': 'કિચે', 'he': 'קיצ׳ה', 'hi': 'किश', 'hi-Latn': 'Kʼicheʼ', 'hr': 'kiče', 'hsb': 'kʼicheʼ', 'ht': 'quiché', 'hu': 'kicse', 'hy': 'քիչե', 'ia': 'kʼicheʼ', 'id': 'Kʼicheʼ', 'ig': 'Kʼicheʼ', 'is': 'kiche', 'it': 'k’iche’', 'ja': 'キチェ語', 'ka': 'კიჩე', 'kea': 'kitxe', 'kgp': 'kisé', 'kk': 'киче тілі', 'kk-Cyrl': 'киче тілі', 'km': 'គីចឈី', 'kn': 'ಕಿಷೆ', 'ko': '키체어', 'kok': 'किचे', 'kok-Deva': 'किचे', 'ksh': 'K’ische’', 'ku': 'k’iche’', 'ky': 'кичече', 'lb': 'Quiché-Sprooch', 'lo': 'Kʼicheʼ', 'lrc': 'کیچی', 'lt': 'kičių', 'lv': 'kiče', 'mk': 'киче', 'ml': 'ക്വിച്ചെ', 'mn': 'киче', 'mr': 'कीशेइ', 'ms': 'Kʼicheʼ', 'mt': 'K’iche’', 'my': 'ကီခ်အီချီ', 'mzn': 'کئیچه\u200cئی', 'nb': 'k’iche’', 'ne': 'किचे', 'nl': 'K’iche’', 'nn': 'k’iche', 'no': 'k’iche’', 'or': 'କିଚେ', 'pa': 'ਕੇਸ਼', 'pa-Guru': 'ਕੇਸ਼', 'pl': 'kicze', 'ps': 'کچی', 'pt': 'quiché', 'qu': 'Kʼicheʼ Simi', 'quc': 'Kʼicheʼ', 'ro': 'quiché', 'ru': 'киче', 'sd': 'ڪچي', 'sd-Arab': 'ڪچي', 'si': 'කියිචේ', 'sk': 'quiché', 'sl': 'quiche', 'smn': 'ki’che’', 'sq': 'kiçeisht', 'sr': 'киче', 'sr-Cyrl': 'киче', 'sr-Latn': 'kiče', 'sv': 'quiché', 'sw': 'Kʼicheʼ', 'ta': 'கீசீ', 'te': 'కిచే', 'tg': 'киче', 'th': 'กีเช', 'ti': 'ኪቼ', 'tk': 'kiçe dili', 'to': 'lea fakakīsē', 'tr': 'Kiçece', 'tt': 'киче', 'uk': 'кіче', 'ur': 'کيشی', 'uz': 'kiche', 'uz-Cyrl': 'кичэ', 'uz-Latn': 'kiche', 'vec': 'kish', 'vi': 'Tiếng Kʼicheʼ', 'wo': 'Kishe', 'xnr': 'किश', 'yrl': 'kixé', 'yue': '基切文', 'yue-Hans': '基切文', 'yue-Hant': '基切文', 'zh': '基切语', 'zh-Hans': '基切语', 'zh-Hant': '基切文', 'zu': 'isi-Kʼicheʼ'},
    'qud': {'en': 'Calderón Highland Quichua'},
    'quf': {'en': 'Lambayeque Quechua'},
    'qug': {'am': 'ቺምቦራዞ ሃይላንድ ኩቹዋ', 'ast': 'quichua del altiplanu de Chimborazo', 'br': 'kichuaeg Chimborazo', 'cs': 'kečuánština (chimborazo)', 'de': 'Chimborazo Hochland-Quechua', 'en': 'Chimborazo Highland Quichua', 'fi': 'chimborazonylänköketšua', 'fr': 'quichua du Haut-Chimborazo', 'gd': 'Quichua Àrd-tìr Chimborazo', 'hi-Latn': 'Chimborazo Highland Quichua', 'ht': 'quichua du Haut-Chimborazo', 'it': 'quechua dell’altopiano del Chimborazo', 'ja': 'チンボラソ高地ケチュア語', 'lb': 'Kichwa (Chimborazo-Gebidder)', 'lt': 'Čimboraso aukštumų kečujų', 'mk': 'кичвански', 'nb': 'kichwa (Chimborazo-høylandet)', 'ne': 'चिम्बोराजो उच्चस्थान किचुआ', 'nl': 'Kichwa', 'nn': 'kichwa (Chimborazo-høylandet)', 'no': 'kichwa (Chimborazo-høylandet)', 'pl': 'keczua górski (Chimborazo)', 'sv': 'Chimborazo-höglandskichwa', 'th': 'ควิชัวไฮแลนด์ชิมโบราโซ', 'to': 'lea fakakuitisa-simipolaso', 'tr': 'Chimborazo Highland Quichua', 'vi': 'Tiếng Quechua ở Cao nguyên Chimborazo', 'yue': '欽博拉索海蘭蓋丘亞文', 'yue-Hans': '钦博拉索海兰盖丘亚文', 'yue-Hant': '欽博拉索海蘭蓋丘亞文', 'zh-Hant': '欽博拉索海蘭蓋丘亞文'},
    'quh': {'en': 'South Bolivian Quechua'},
    'qui': {'en': 'Quileute'},
    'quk': {'en': 'Chachapoyas Quechua'},
    'qul': {'en': 'North Bolivian Quechua'},
    'qum': {'en': 'Sipacapense'},
    'qun': {'en': 'Quinault'},
    'qup': {'en': 'Southern Pastaza Quechua'},
    'quq': {'en': 'Quinqui'},
    'qur': {'en': 'Yanahuanca Pasco Quechua'},
    'qus': {'en': 'Santiago del Estero Quichua'},
    'quv': {'en': 'Sacapulteco'},
    'quw': {'en': 'Tena Lowland Quichua'},
    'qux': {'en': 'Yauyos Quechua'},
    'quy': {'en': 'Ayacucho Quechua'},
    'quz': {'en': 'Cusco Quechua'},
    'qva': {'en': 'Ambo-Pasco Quechua'},
    'qvc': {'en': 'Cajamarca Quechua'},
    'qve': {'en': 'Eastern Apurímac Quechua'},
    'qvh': {'en': 'Huamalíes-Dos de Mayo Huánuco Quechua'},
    'qvi': {'en': 'Imbabura Highland Quichua'},
    'qvj': {'en': 'Loja Highland Quichua'},
    'qvl': {'en': 'Cajatambo North Lima Quechua'},
    'qvm': {'en': 'Margos-Yarowilca-Lauricocha Quechua'},
    'qvn': {'en': 'North Junín Quechua'},
    'qvo': {'en': 'Napo Lowland Quechua'},
    'qvp': {'en': 'Pacaraos Quechua'},
    'qvs': {'en': 'San Martín Quechua'},
    'qvw': {'en': 'Huaylla Wanca Quechua'},
    'qvy': {'en': 'Queyu'},
    'qvz': {'en': 'Northern Pastaza Quichua'},
    'qwa': {'en': 'Corongo Ancash Quechua'},
    'qwc': {'en': 'Classical Quechua'},
    'qwe': {'en': 'Quechuan (family)'},
    'qwe-kch': {'en': 'Kichwa'},
    'qwh': {'en': 'Huaylas Ancash Quechua'},
    'qwm': {'en': 'Kuman (Russia)'},
    'qws': {'en': 'Sihuas Ancash Quechua'},
    'qwt': {'en': 'Kwalhioqua-Tlatskanai'},
    'qxa': {'en': 'Chiquián Ancash Quechua'},
    'qxc': {'en': 'Chincha Quechua'},
    'qxh': {'en': 'Panao Huánuco Quechua'},
    'qxl': {'en': 'Salasaca Highland Quichua'},
    'qxn': {'en': 'Northern Conchucos Ancash Quechua'},
    'qxo': {'en': 'Southern Conchucos Ancash Quechua'},
    'qxp': {'en': 'Puno Quechua'},
    'qxq': {'en': "Qashqa'i"},
    'qxr': {'en': 'Cañar Highland Quichua'},
    'qxs': {'en': 'Southern Qiang'},
    'qxt': {'en': 'Santa Ana de Tusi Pasco Quechua'},
    'qxu': {'en': 'Arequipa-La Unión Quechua'},
    'qxw': {'en': 'Jauja Wanca Quechua'},
    'qya': {'en': 'Quenya'},
    'qyp': {'en': 'Quiripi'},
    'raa': {'en': 'Dungmali'},
    'rab': {'en': 'Camling'},
    'rac': {'en': 'Rasawa'},
    'rad': {'en': 'Rade'},
    'raf': {'en': 'Western Meohang'},
    'rag': {'en': 'Logooli'},
    'rah': {'en': 'Rabha'},
    'rai': {'en': 'Ramoaaina'},
    'raj': {'af': 'Rajasthani', 'ak': 'Ragyasitan kasa', 'am': 'ራጃስታኒ', 'ar': 'الراجاسثانية', 'as': 'ৰাজস্থানী', 'ast': 'rajasthanín', 'az': 'racastani', 'az-Latn': 'racastani', 'bal-Latn': 'Rájestáni', 'be': 'раджастханская', 'be-tarask': 'раджастханская', 'bg': 'раджастански', 'bn': 'রাজস্থানী', 'br': 'rajasthani', 'brx': 'राजस्थानी', 'bs': 'rajastani', 'bs-Cyrl': 'рађастани', 'bs-Latn': 'rajastani', 'ca': 'rajasthani', 'ca-ES-valencia': 'rajasthani', 'ccp': '𑄢𑄎𑄴𑄥𑄳𑄦𑄚𑄨', 'chr': 'ᎳᏣᏍᎭᏂ', 'cs': 'rádžastánština', 'cy': 'Rajasthaneg', 'da': 'rajasthani', 'de': 'Rajasthani', 'dsb': 'rajasthanišćina', 'el': 'Ραζασθάνι', 'el-polyton': 'Ραζασθάνι', 'en': 'Rajasthani', 'es': 'rajasthani', 'et': 'radžastani', 'eu': 'rajastanera', 'fa': 'راجستانی', 'ff-Adlm': '𞤈𞤢𞤶𞤢𞤧𞤼𞤢𞤲𞤭𞥅𞤪𞤫', 'fi': 'radžastani', 'fil': 'Rajasthani', 'fo': 'rajasthani', 'fr': 'rajasthani', 'fy': 'Rajasthani', 'ga': 'Rajasthani', 'gd': 'Rajasthani', 'gl': 'rajasthani', 'gsw': 'Rajasthani', 'gu': 'રાજસ્થાની', 'ha': 'Rajasthani', 'he': 'ראג׳סטאני', 'hi': 'राजस्थानी', 'hi-Latn': 'Rajasthani', 'hr': 'rajasthani', 'hsb': 'rajasthani', 'ht': 'rajasthani', 'hu': 'radzsasztáni', 'hy': 'ռաջաստաներեն', 'ia': 'rajasthani', 'id': 'Rajasthani', 'ig': 'Rajastani', 'is': 'rajastaní', 'it': 'rajasthani', 'ja': 'ラージャスターン語', 'jv': 'Rajasthani', 'ka': 'რაჯასთანი', 'kgp': 'hajanhtỹnĩ', 'kk': 'раджастани тілі', 'kk-Cyrl': 'раджастани тілі', 'km': 'រ៉ាចាស់ថានី', 'kn': 'ರಾಜಸ್ಥಾನಿ', 'ko': '라자스탄어', 'kok': 'राजस्थानी', 'kok-Deva': 'राजस्थानी', 'ks': 'راجِستھٲنؠ', 'ks-Arab': 'راجِستھٲنؠ', 'ku': 'rajasthanî', 'ky': 'ражастаниче', 'lb': 'Rajasthani', 'lo': 'ຣາຈັສທານິ', 'lt': 'Radžastano', 'lv': 'radžastāņu', 'mi': 'Ratiahitani', 'mk': 'раџастански', 'ml': 'രാജസ്ഥാനി', 'mn': 'ражастани', 'mr': 'राजस्थानी', 'ms': 'Rajasthani', 'mt': 'Raġastani', 'my': 'ရာဂျာစတာနီ', 'nb': 'rajasthani', 'ne': 'राजस्थानी', 'nl': 'Rajasthani', 'nn': 'rajasthani', 'no': 'rajasthani', 'or': 'ରାଜସ୍ଥାନୀ', 'pa': 'ਰਾਜਸਥਾਨੀ', 'pa-Guru': 'ਰਾਜਸਥਾਨੀ', 'pcm': 'Rajástháni Lángwej', 'pl': 'radźasthani', 'ps': 'راجستھانی', 'pt': 'rajastani', 'qu': 'Rajasthani', 'raj': 'राजस्थानी', 'rm': 'rajasthani', 'ro': 'rajasthani', 'ru': 'раджастхани', 'sc': 'rajasthani', 'sd': 'راجستاني', 'sd-Arab': 'راجستاني', 'si': 'රාජස්ථානි', 'sk': 'radžastančina', 'sl': 'radžastanščina', 'so': 'Rajasthani', 'sq': 'raxhastanisht', 'sr': 'раџастански', 'sr-Cyrl': 'раџастански', 'sr-Latn': 'radžastanski', 'sv': 'rajasthani', 'sw': 'Kirajasthani', 'ta': 'ராஜஸ்தானி', 'te': 'రాజస్తానీ', 'th': 'ราชสถาน', 'ti': 'ራጃስታኒ', 'tk': 'rajastani dili', 'to': 'lea fakalasasitani', 'tr': 'Rajasthani', 'ug': 'راجاستانچە', 'uk': 'раджастхані', 'ur': 'راجستھانی', 'uz': 'rajastani', 'uz-Latn': 'rajastani', 'vec': 'rajastani', 'vi': 'Tiếng Rajasthani', 'xnr': 'राजस्थानी', 'yo': 'Rajastánì', 'yrl': 'rayasitani', 'yue': '拉賈斯坦諸文', 'yue-Hans': '拉贾斯坦诸文', 'yue-Hant': '拉賈斯坦諸文', 'zh': '拉贾斯坦语', 'zh-Hans': '拉贾斯坦语', 'zh-Hant': '拉賈斯坦諸文', 'zu': 'isi-Rajasthani'},
    'rak': {'en': 'Tulu-Bohuai'},
    'ral': {'en': 'Ralte'},
    'ram': {'en': 'Canela'},
    'ran': {'en': 'Riantana'},
    'rao': {'en': 'Rao'},
    'rap': {'af': 'Rapanui', 'am': 'ራፓኑኢ', 'ar': 'الراباني', 'as': 'ৰাপানুই', 'ast': 'rapanui', 'az': 'rapanui', 'az-Cyrl': 'рапануи', 'az-Latn': 'rapanui', 'be': 'рапануі', 'be-tarask': 'рапануі', 'bg': 'рапа нуи', 'bgn': 'راپانویی', 'bn': 'রাপানুই', 'br': 'rapanui', 'brx': 'रापानुई', 'bs': 'rapanui', 'bs-Cyrl': 'рапануи', 'bs-Latn': 'rapanui', 'ca': 'rapanui', 'ca-ES-valencia': 'rapanui', 'ccp': '𑄢𑄛𑄚𑄳𑄆𑄪', 'ce': 'рапануйн', 'chr': 'ᎳᏆᏄᏫ', 'ckb': 'ڕاپانوویی', 'cs': 'rapanujština', 'cy': 'Rapanŵi', 'da': 'rapanui', 'de': 'Rapanui', 'dsb': 'rapanuišćina', 'el': 'Ραπανούι', 'el-polyton': 'Ραπανούι', 'en': 'Rapanui', 'es': 'rapanui', 'et': 'rapanui', 'eu': 'rapanuia', 'fa': 'راپانویی', 'ff-Adlm': '𞤈𞤢𞤨𞤢𞤲𞤵𞤭𞥅𞤪𞤫', 'fi': 'rapanui', 'fil': 'Rapanui', 'fo': 'rapanui', 'fr': 'rapanui', 'fy': 'Rapanui', 'ga': 'Rapanúis', 'gd': 'Rapa Nui', 'gl': 'rapanui', 'gsw': 'Oschterinsel-Schpraach', 'gu': 'રાપાનુઇ', 'ha': 'Rapanui', 'he': 'רפאנוי', 'hi': 'रापानुई', 'hi-Latn': 'Rapanui', 'hr': 'rapa nui', 'hsb': 'rapanuišćina', 'ht': 'rapanui', 'hu': 'rapanui', 'hy': 'ռապանուի', 'ia': 'rapanui', 'id': 'Rapanui', 'ig': 'Rapanui', 'is': 'rapanúí', 'it': 'rapanui', 'ja': 'ラパヌイ語', 'jv': 'Rapanui', 'ka': 'რაპანუი', 'kgp': 'hapanũj', 'kk': 'рапануй тілі', 'kk-Cyrl': 'рапануй тілі', 'km': 'រ៉ាប៉ានូ', 'kn': 'ರಾಪಾನುಯಿ', 'ko': '라파뉴이', 'kok': 'रापान्यु', 'kok-Deva': 'रापान्यु', 'ks': 'رَپانوی', 'ks-Arab': 'رَپانوی', 'ku': 'rapanuyî', 'ky': 'рапаньюча', 'lb': 'Ouschterinsel-Sprooch', 'lo': 'ຣາປານຸຍ', 'lt': 'rapanui', 'lv': 'rapanuju', 'mai': 'रपानुई', 'mi': 'Rapanui', 'mk': 'рапанујски', 'ml': 'രാപനൂയി', 'mn': 'рапануи', 'mr': 'रापानुई', 'ms': 'Rapanui', 'mt': 'Rapanwi', 'my': 'ရပန်နူအီ', 'nb': 'rapanui', 'ne': 'रापानुई', 'nl': 'Rapanui', 'nn': 'rapanui', 'no': 'rapanui', 'or': 'ରାପାନୁଇ', 'pa': 'ਰਾਪਾਨੁਈ', 'pa-Guru': 'ਰਾਪਾਨੁਈ', 'pcm': 'Rapánui Lángwej', 'pl': 'rapanui', 'ps': 'رپانوئي', 'pt': 'rapanui', 'qu': 'Rapanui Simi', 'rm': 'rapanui', 'ro': 'rapanui', 'ru': 'рапануйский', 'sc': 'rapanui', 'sd': 'ريپنوئي', 'sd-Arab': 'ريپنوئي', 'si': 'රපනුයි', 'sk': 'rapanujčina', 'sl': 'rapanujščina', 'smn': 'rapanui', 'so': 'Rapanui', 'sq': 'rapanuisht', 'sr': 'рапануи', 'sr-Cyrl': 'рапануи', 'sr-Latn': 'rapanui', 'sv': 'rapanui', 'sw': 'Kirapanui', 'ta': 'ரபனுய்', 'te': 'రాపన్యుయి', 'th': 'ราปานูอี', 'ti': 'ራፓኑይ', 'tk': 'rapanuý dili', 'to': 'lea fakalapanui', 'tr': 'Rapanui dili', 'ug': 'راپانىيچە', 'uk': 'рапануї', 'ur': 'رپانوی', 'uz': 'rapanui', 'uz-Latn': 'rapanui', 'vec': 'rapanui', 'vi': 'Tiếng Rapanui', 'xnr': 'रापानुई', 'yo': 'Èdè Rapanu', 'yrl': 'rapanui', 'yue': '復活島文', 'yue-Hans': '复活岛文', 'yue-Hant': '復活島文', 'zh': '拉帕努伊语', 'zh-Hans': '拉帕努伊语', 'zh-Hant': '復活島文', 'zu': 'isi-Rapanui'},
    'raq': {'en': 'Saam'},
    'rar': {'af': 'Rarotongaans', 'am': 'ራሮቶንጋን', 'ar': 'الراروتونجاني', 'as': 'ৰাৰোতোঙ্গন', 'ast': 'rarotonganu', 'az': 'rarotonqan', 'az-Cyrl': 'раротонган', 'az-Latn': 'rarotonqan', 'be': 'раратонг', 'be-tarask': 'раратонг', 'bg': 'раротонга', 'bgn': 'راروتونگی', 'bn': 'রারোটোংগান', 'br': 'rarotonga', 'brx': 'रार’टंगान', 'bs': 'rarotongan', 'bs-Cyrl': 'раротонган', 'bs-Latn': 'rarotongan', 'ca': 'rarotongà', 'ca-ES-valencia': 'rarotongà', 'ccp': '𑄢𑄢𑄮𑄑𑄮𑄁𑄉𑄚𑄴', 'ce': 'раротонга', 'chr': 'ᎳᎶᏙᎾᎦᏂ', 'ckb': 'ڕاڕۆتۆنگان', 'cs': 'rarotongánština', 'cy': 'Raratongeg', 'da': 'rarotonga', 'de': 'Rarotonganisch', 'dsb': 'rarotonganšćina', 'el': 'Ραροτονγκάν', 'el-polyton': 'Ραροτονγκάν', 'en': 'Rarotongan', 'es': 'rarotongano', 'et': 'rarotonga', 'eu': 'rarotongera', 'fa': 'راروتونگایی', 'ff-Adlm': '𞤈𞤢𞤪𞤮𞤼𞤮𞤲𞤺𞤢𞤲𞤪𞤫', 'fi': 'rarotonga', 'fil': 'Rarotongan', 'fo': 'rarotongiskt', 'fr': 'rarotongien', 'fy': 'Rarotongan', 'ga': 'Raratongais', 'gd': 'Cànan Rarotonga', 'gl': 'rarotongano', 'gsw': 'Rarotonganisch', 'gu': 'રારોટોંગન', 'ha': 'Rarotongan', 'he': 'ררוטונגאן', 'hi': 'रारोतोंगन', 'hi-Latn': 'Rarotongan', 'hr': 'rarotonški', 'hsb': 'rarotonganšćina', 'ht': 'rarotongien', 'hu': 'rarotongai', 'hy': 'ռարոտոնգաներեն', 'ia': 'rarotongano', 'id': 'Rarotonga', 'ig': 'Rarotongan', 'is': 'rarótongska', 'it': 'rarotonga', 'ja': 'ラロトンガ語', 'jv': 'Rarotongan', 'ka': 'რაროტონგული', 'kgp': 'harotãganũ', 'kk': 'раротонган тілі', 'kk-Cyrl': 'раротонган тілі', 'km': 'រ៉ារ៉ូតុងហ្គាន', 'kn': 'ರಾರೋಟೊಂಗನ್', 'ko': '라로통가어', 'kok': 'रारोटोंगान', 'kok-Deva': 'रारोटोंगान', 'ks': 'رَروٹونگَن', 'ks-Arab': 'رَروٹونگَن', 'ku': 'rarotongî', 'ky': 'раротонгача', 'lb': 'Rarotonganesch', 'lo': 'ຣາໂຣທອນການ', 'lt': 'rarotonganų', 'lv': 'rarotongiešu', 'mai': 'रेरोटोंगन', 'mi': 'Rarotonga', 'mk': 'раротонгански', 'ml': 'രാരോടോങ്കൻ', 'mn': 'раротонг', 'mr': 'रारोटोंगन', 'ms': 'Rarotonga', 'mt': 'Rarotongani', 'my': 'ရရိုတွန်ဂန်', 'nb': 'rarotongansk', 'ne': 'रारोटोङ्गान', 'nl': 'Rarotongan', 'nn': 'rarotongansk', 'no': 'rarotongansk', 'or': 'ରାରୋତୋଙ୍ଗନ୍', 'pa': 'ਰਾਰੋਤੋਂਗਨ', 'pa-Guru': 'ਰਾਰੋਤੋਂਗਨ', 'pcm': 'Rarotóngan', 'pl': 'rarotonga', 'ps': 'راروټانګان', 'pt': 'rarotongano', 'qu': 'Rarotongan Simi', 'rm': 'rarotonga', 'ro': 'rarotongan', 'ru': 'раротонга', 'sc': 'rarotonganu', 'sd': 'ريرو ٽينگو', 'sd-Arab': 'ريرو ٽينگو', 'si': 'රරොටොන්ගන්', 'sk': 'rarotongská maorijčina', 'sl': 'rarotongščina', 'smn': 'rarotonga', 'so': 'Rarotongan', 'sq': 'rarontonganisht', 'sr': 'раротонгански', 'sr-Cyrl': 'раротонгански', 'sr-Latn': 'rarotonganski', 'sv': 'rarotonganska', 'sw': 'Kirarotonga', 'ta': 'ரரோடோங்கன்', 'te': 'రారోటొంగాన్', 'th': 'ราโรทองกา', 'ti': 'ራሮቶንጋንኛ', 'tk': 'kuk dili', 'to': 'lea fakalalotonga', 'tr': 'Rarotongan', 'uk': 'раротонга', 'ur': 'راروتونگان', 'uz': 'rarotongan', 'uz-Latn': 'rarotongan', 'vec': 'rarotongan', 'vi': 'Tiếng Rarotongan', 'xnr': 'रारोतोंगन', 'yo': 'Èdè Rarotonga', 'yrl': 'rurutũganu', 'yue': '拉羅通加文', 'yue-Hans': '拉罗通加文', 'yue-Hant': '拉羅通加文', 'zh': '拉罗汤加语', 'zh-Hans': '拉罗汤加语', 'zh-Hant': '拉羅通加文', 'zu': 'isi-Rarotongan'},
    'ras': {'en': 'Tegali'},
    'rat': {'en': 'Razajerdi'},
    'rau': {'en': 'Raute'},
    'rav': {'en': 'Sampang'},
    'raw': {'en': 'Rawang'},
    'rax': {'en': 'Rang'},
    'ray': {'en': 'Rapa'},
    'raz': {'en': 'Rahambuu'},
    'rbb': {'en': 'Rumai Palaung'},
    'rbk': {'en': 'Northern Bontok'},
    'rbl': {'en': 'Miraya Bikol'},
    'rbp': {'en': 'Barababaraba'},
    'rcf': {'en': 'Réunion Creole French'},
    'rdb': {'en': 'Rudbari'},
    'rea': {'en': 'Rerau'},
    'reb': {'en': 'Rembong'},
    'ree': {'en': 'Rejang Kayan'},
    'reg': {'en': 'Kara (Tanzania)'},
    'rei': {'en': 'Reli'},
    'rej': {'en': 'Rejang'},
    'rel': {'en': 'Rendille'},
    'rem': {'en': 'Remo'},
    'ren': {'en': 'Rengao'},
    'rer': {'en': 'Rer Bare'},
    'res': {'en': 'Reshe'},
    'ret': {'en': 'Retta'},
    'rey': {'en': 'Reyesano'},
    'rga': {'en': 'Roria'},
    'rge': {'en': 'Romano-Greek'},
    'rgk': {'en': 'Rangkas'},
    'rgn': {'ast': 'romañol', 'br': 'romagnoleg', 'ca': 'romanyès', 'ca-ES-valencia': 'romanyès', 'cs': 'romaňolština', 'de': 'Romagnol', 'en': 'Romagnol', 'et': 'romanja', 'fi': 'romagnoli', 'fr': 'romagnol', 'gd': 'Romagnol', 'hi-Latn': 'Romagnol', 'ht': 'romagnol', 'hy': 'ռոմանիոլերեն', 'it': 'romagnolo', 'ja': 'ロマーニャ語', 'lb': 'Romagnol', 'lt': 'italų kalbos Romanijos tarmė', 'mk': 'ромањолски', 'nb': 'romagnolsk', 'nl': 'Romagnol', 'nn': 'romagnolsk', 'no': 'romagnolsk', 'pl': 'romagnol', 'sv': 'romagnol', 'th': 'โรมัณโญ', 'to': 'lea fakalomaniolo', 'tr': 'Romanyolca', 'yue': '羅馬格諾里文', 'yue-Hans': '罗马格诺里文', 'yue-Hant': '羅馬格諾里文', 'zh-Hant': '羅馬格諾里文'},
    'rgr': {'en': 'Resígaro'},
    'rgs': {'en': 'Southern Roglai'},
    'rgu': {'en': 'Ringgou'},
    'rhg': {'af': 'Rohingya', 'am': 'ሮሂንግያ', 'ar': 'الروهينغية', 'as': 'ৰোহিঙ্গিয়া', 'az': 'Rohinca', 'az-Latn': 'Rohinca', 'bal-Latn': 'Róhengiá', 'be': 'рахінджа', 'be-tarask': 'рахінджа', 'bg': 'рохинга', 'bn': 'রোহিঙ্গা', 'br': 'rohingya', 'brx': 'रहिंगीया', 'bs': 'rohindža', 'bs-Cyrl': 'рохимгја', 'bs-Latn': 'rohindža', 'ca': 'rohingya', 'ca-ES-valencia': 'rohingya', 'chr': 'ᎶᎯᏂᏯ', 'cs': 'rohingština', 'cy': 'Rohingya', 'da': 'rohingya', 'de': 'Rohingyalisch', 'dsb': 'rohingyašćina', 'el': 'Ροχίνγκια', 'el-polyton': 'Ροχίνγκια', 'en': 'Rohingya', 'es': 'rohinyá', 'et': 'rohingja', 'eu': 'rohingyera', 'fa': 'روهینگیا', 'ff-Adlm': '𞤈𞤮𞤸𞤭𞤲𞤺𞤢𞥄𞤪𞤫', 'fi': 'rohinga', 'fil': 'Rohingya', 'fr': 'rohingya', 'ga': 'Róihinis', 'gd': 'Rohingya', 'gl': 'rohingya', 'gu': 'રોહિંગ્યા', 'ha': 'Harshen Rohingya', 'he': 'רוהינגיה', 'hi': 'रोहिंग्या', 'hi-Latn': 'Rohingya', 'hr': 'rohindža', 'hsb': 'rohingyašćina', 'ht': 'rohingya', 'hu': 'rohingja', 'hy': 'ռոհինջա', 'ia': 'rohingya', 'id': 'Rohingya', 'ig': 'Rohingya', 'is': 'rohingja', 'it': 'rohingya', 'ja': 'ロヒンギャ語', 'jv': 'Rohingya', 'ka': 'როჰინგა', 'kk': 'рохинджа', 'kk-Cyrl': 'рохинджа', 'km': 'រ៉ូហ៊ីងយ៉ា', 'kn': 'ರೋಹಿಂಗ್ಯಾ', 'ko': '로힝야어', 'kok': 'रोहिंग्या', 'kok-Deva': 'रोहिंग्या', 'ku': 'rohingyayî', 'ky': 'рохинжача', 'lo': 'ໂຣຮິນຢາ', 'lt': 'rochindža', 'lv': 'rohindžu', 'mai': 'रोहिंग्या', 'mi': 'Rohingia', 'mk': 'рохиншки', 'ml': 'റോഹിംഗാ', 'mn': 'рохинжа', 'mr': 'रोहिंग्या', 'ms': 'Rohingya', 'my': 'ရိုဟင်ဂျာ', 'nb': 'rohingya', 'ne': 'रोहिङ्ग्या', 'nl': 'Rohingya', 'nn': 'rohingya', 'no': 'rohingya', 'nqo': 'ߙߏߤߌ߲ߜ߭ߊ', 'or': 'ରୋହିଙ୍ଗ୍ୟା', 'pa': 'ਰੋਹਿੰਗਿਆ', 'pa-Guru': 'ਰੋਹਿੰਗਿਆ', 'pcm': 'Rohínjia', 'pl': 'rohingya', 'ps': 'روهینګیا', 'pt': 'ruainga', 'qu': 'Rohingya Simi', 'rhg': '𐴌𐴗𐴥𐴝𐴙𐴚𐴒𐴙𐴝', 'rhg-Rohg': '𐴌𐴗𐴥𐴝𐴙𐴚𐴒𐴙𐴝', 'ro': 'rohingya', 'ru': 'рохинджа', 'sc': 'rohingya', 'sd': 'روھنگيا', 'sd-Arab': 'روھنگيا', 'si': 'රෝහින්ග්\u200dයා', 'sk': 'rohingčina', 'sl': 'rohinščina', 'so': 'Rohingya', 'sq': 'rohingiaisht', 'sr': 'рохингја', 'sr-Cyrl': 'рохингја', 'sr-Latn': 'rohingja', 'sv': 'ruáingga', 'sw': 'Kirohingya', 'ta': 'ரோகிஞ்சா', 'te': 'రోహింగ్యా', 'th': 'โรฮิงญา', 'ti': 'ሮሂንግያ', 'tk': 'rohinýa dili', 'to': 'lea fakalouhingia', 'tr': 'Rohingya dili', 'uk': 'рогінджа', 'ur': 'روہنگیا', 'uz': 'rohinja', 'uz-Latn': 'rohinja', 'vec': 'roinga', 'vi': 'Tiếng Rohingya', 'yo': 'Èdè Rohinga', 'yue': '羅興亞文', 'yue-Hans': '罗兴亚文', 'yue-Hant': '羅興亞文', 'zh': '罗兴亚语', 'zh-Hans': '罗兴亚语', 'zh-Hant': '羅興亞文', 'zu': 'Rohingya'},
    'rhp': {'en': 'Yahang'},
    'ria': {'en': 'Riang (India)'},
    'rie': {'en': 'Rien'},
    'rif': {'ast': 'rifianu', 'bal-Latn': 'Ripi', 'cs': 'rífština', 'de': 'Tarifit', 'en': 'Riffian', 'et': 'riifi', 'ff-Adlm': '𞤈𞤭𞤬𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'tarifit', 'fr': 'rifain', 'gd': 'Tamaisich an Rif', 'hi-Latn': 'Riffian', 'ht': 'rifain', 'hy': 'ռիֆերեն', 'ia': 'rifeno', 'ig': 'Riffian', 'it': 'tarifit', 'ja': 'リーフ語', 'ku': 'tarifit', 'lb': 'Tarifit', 'lt': 'rifų', 'mk': 'рифски', 'nb': 'riff', 'nl': 'Riffijns', 'nn': 'riff', 'no': 'riff', 'pl': 'tarifit', 'rif': 'Tarifit', 'sc': 'rifenu', 'sv': 'riffianska', 'th': 'ริฟฟิอัน', 'ti': 'ሪፍኛ', 'to': 'lea fakalifi', 'tr': 'Rif Berbericesi', 'vec': 'rifegno', 'yue': '里菲亞諾文', 'yue-Hans': '里菲亚诺文', 'yue-Hant': '里菲亞諾文', 'zh': '里夫语', 'zh-Hans': '里夫语', 'zh-Hant': '里菲亞諾文'},
    'ril': {'en': 'Riang Lang'},
    'rim': {'en': 'Nyaturu'},
    'rin': {'en': 'Nungu'},
    'rir': {'en': 'Ribun'},
    'rit': {'en': 'Ritharrngu'},
    'riu': {'en': 'Riung'},
    'rjg': {'en': 'Rajong'},
    'rji': {'en': 'Raji'},
    'rjs': {'en': 'Rajbanshi'},
    'rka': {'en': 'Kraol'},
    'rkb': {'en': 'Rikbaktsa'},
    'rkh': {'en': 'Rakahanga-Manihiki'},
    'rki': {'en': 'Rakhine'},
    'rkm': {'en': 'Marka'},
    'rkt': {'en': 'Rangpuri'},
    'rkw': {'en': 'Arakwal'},
    'rm': {'af': 'Reto-Romaans', 'ak': 'Romanhye kasa', 'am': 'ሮማንሽ', 'ar': 'الرومانشية', 'as': 'ৰোমানচ', 'ast': 'romanche', 'az': 'romanş', 'az-Cyrl': 'романш', 'az-Latn': 'romanş', 'bal-Latn': 'Rumansh', 'be': 'рэтараманская', 'be-tarask': 'рэтараманская', 'bg': 'реторомански', 'bgn': 'رومانش', 'bn': 'রোমান্স', 'br': 'romañcheg', 'brx': 'रमान्श', 'bs': 'retoromanski', 'bs-Cyrl': 'рето-романски', 'bs-Latn': 'retoromanski', 'ca': 'retoromànic', 'ca-ES-valencia': 'retoromànic', 'ccp': '𑄢𑄮𑄟𑄚𑄴𑄥𑄴', 'ce': 'романшийн', 'chr': 'ᎠᏂᎶᎺᏂ', 'ckb': 'ڕۆمانش', 'cs': 'rétorománština', 'cy': 'Románsh', 'da': 'rætoromansk', 'de': 'Rätoromanisch', 'dsb': 'retoromańšćina', 'dz': 'རོ་མེ་ནིཤ་ཁ', 'ee': 'romanshgbe', 'el': 'Ρομανικά', 'el-polyton': 'Ρομανικά', 'en': 'Romansh', 'eo': 'romanĉa', 'es': 'romanche', 'et': 'romanši', 'eu': 'erretorromaniera', 'fa': 'رومانش', 'ff-Adlm': '𞤈𞤮𞤥𞤢𞤲𞤧𞤪𞤫', 'fi': 'retoromaani', 'fil': 'Romansh', 'fo': 'retoromanskt', 'fr': 'romanche', 'fur': 'rumanç', 'fy': 'Reto-Romaansk', 'ga': 'Rómainis', 'gd': 'Rumains', 'gl': 'romanche', 'gsw': 'Rätoromanisch', 'gu': 'રોમાન્શ', 'ha': 'Romansh', 'he': 'רומאנש', 'hi': 'रोमान्श', 'hi-Latn': 'Romansh', 'hr': 'retoromanski', 'hsb': 'retoromanšćina', 'ht': 'romanche', 'hu': 'rétoromán', 'hy': 'ռոմանշերեն', 'ia': 'romanche', 'id': 'Reto-Roman', 'ig': 'Asụsụ Romansh', 'is': 'rómanska', 'it': 'romancio', 'ja': 'ロマンシュ語', 'jv': 'Roman', 'ka': 'რეტორომანული', 'kea': 'romanxi', 'kgp': 'homỹse', 'kk': 'романш тілі', 'kk-Cyrl': 'романш тілі', 'km': 'រ៉ូម៉ង់', 'kn': 'ರೊಮಾನ್ಶ್', 'ko': '로만시어', 'kok': 'रोमान्श', 'kok-Deva': 'रोमान्श', 'ks': 'رومانش', 'ks-Arab': 'رومانش', 'ksh': 'Rätoromaanesch', 'ku': 'romancî', 'ky': 'романшча', 'lb': 'Rätoromanesch', 'lkt': 'Romansh Iyápi', 'lld': 'rumanc', 'lo': 'ໂຣແມນຊ໌', 'lrc': 'رومانش', 'lt': 'retoromanų', 'lv': 'retoromāņu', 'mai': 'रोमांश', 'mi': 'Romānihi', 'mk': 'реторомански', 'ml': 'റൊമാഞ്ച്', 'mn': 'романш', 'mr': 'रोमान्श', 'ms': 'Romansh', 'mt': 'Romanz', 'my': 'ရောမ', 'mzn': 'رومانش', 'nb': 'retoromansk', 'ne': 'रोमानिस', 'nl': 'Reto-Romaans', 'nn': 'retoromansk', 'no': 'retoromansk', 'nqo': 'ߙߏߡߊ߲ߛߌߞߊ߲', 'or': 'ରୋମାନଶ୍\u200c', 'pa': 'ਰੋਮਾਂਸ਼', 'pa-Guru': 'ਰੋਮਾਂਸ਼', 'pcm': 'Románsh', 'pl': 'retoromański', 'ps': 'رومانیش', 'pt': 'romanche', 'qu': 'Romanche Simi', 'rm': 'rumantsch', 'ro': 'romanșă', 'ru': 'романшский', 'sc': 'romànciu', 'sd': 'رومانش', 'sd-Arab': 'رومانش', 'se': 'romanšgiella', 'si': 'රොමෑන්ශ්', 'sk': 'rétorománčina', 'sl': 'retoromanščina', 'smn': 'retoroomaankielâ', 'so': 'Romaanis', 'sq': 'retoromanisht', 'sr': 'романш', 'sr-Cyrl': 'романш', 'sr-Latn': 'romanš', 'sv': 'rätoromanska', 'sw': 'Kiromanshi', 'ta': 'ரோமான்ஷ்', 'te': 'రోమన్ష్', 'tg': 'ретороманӣ', 'th': 'โรแมนซ์', 'ti': 'ሮማንሽ', 'tk': 'retoroman dili', 'to': 'lea fakalaito-lomēnia', 'tr': 'Romanşça', 'tt': 'ретороман', 'ug': 'رومانسچە', 'uk': 'ретороманська', 'ur': 'رومانش', 'uz': 'romansh', 'uz-Cyrl': 'романшча', 'uz-Latn': 'romansh', 'vec': 'romancio', 'vi': 'Tiếng Romansh', 'wae': 'Rätromaniš', 'wo': 'Romaas', 'xnr': 'रोमान्श', 'yo': 'Rómáǹṣì', 'yrl': 'rumãxi', 'yue': '羅曼斯文', 'yue-Hans': '罗曼斯文', 'yue-Hant': '羅曼斯文', 'zh': '罗曼什语', 'zh-Hans': '罗曼什语', 'zh-Hant': '羅曼斯文', 'zu': 'isi-Romansh'},
    'rma': {'en': 'Rama'},
    'rmb': {'en': 'Rembarrnga'},
    'rmc': {'en': 'Carpathian Romani'},
    'rmd': {'en': 'Traveller Danish'},
    'rme': {'en': 'Angloromani'},
    'rmf': {'en': 'Kalo Finnish Romani'},
    'rmg': {'en': 'Traveller Norwegian'},
    'rmh': {'en': 'Murkim'},
    'rmi': {'en': 'Lomavren'},
    'rmk': {'en': 'Romkun'},
    'rml': {'en': 'Baltic Romani'},
    'rmm': {'en': 'Romang'},
    'rmn': {'en': 'Balkan Romani'},
    'rmo': {'en': 'Sinte Romani'},
    'rmp': {'en': 'Rempi'},
    'rmq': {'en': 'Caló'},
    'rmr': {'en': 'Caló'},
    'rms': {'en': 'Romanian Sign Language'},
    'rmt': {'en': 'Domari'},
    'rmu': {'en': 'Tavringer Romani'},
    'rmv': {'en': 'Romanova'},
    'rmw': {'en': 'Welsh Romani'},
    'rmx': {'en': 'Romam'},
    'rmy': {'en': 'Vlax Romani'},
    'rmz': {'en': 'Marma'},
    'rn': {'af': 'Rundi', 'am': 'ሩንዲ', 'ar': 'الرندي', 'as': 'ৰুন্দি', 'ast': 'rundi', 'az': 'rundi', 'az-Cyrl': 'рунди', 'az-Latn': 'rundi', 'bal-Latn': 'Róndi', 'be': 'рундзі', 'be-tarask': 'рундзі', 'bg': 'рунди', 'bgn': 'رونڈی', 'bn': 'রুন্দি', 'br': 'rundi', 'brx': 'रुन्दी', 'bs': 'rundi', 'bs-Cyrl': 'рунди', 'bs-Latn': 'rundi', 'ca': 'rundi', 'ca-ES-valencia': 'rundi', 'ccp': '𑄢𑄪𑄚𑄴𑄘𑄨', 'ce': 'рунди', 'chr': 'ᎷᏂᏗ', 'ckb': 'ڕووندی', 'cs': 'kirundština', 'cy': 'Rwndi', 'da': 'rundi', 'de': 'Rundi', 'dsb': 'kirundišćina', 'ee': 'rundigbe', 'el': 'Ρούντι', 'el-polyton': 'Ρούντι', 'en': 'Rundi', 'eo': 'burunda', 'es': 'kirundi', 'et': 'rundi', 'eu': 'rundiera', 'fa': 'روندیایی', 'ff-Adlm': '𞤈𞤵𞤲𞤣𞤭𞥅𞤪𞤫', 'fi': 'rundi', 'fil': 'Rundi', 'fo': 'rundi', 'fr': 'roundi', 'fy': 'Kirundi', 'ga': 'Rúindis', 'gd': 'Kirundi', 'gl': 'rundi', 'gsw': 'Rundi-Schpraach', 'gu': 'રૂન્દી', 'ha': 'Rundi', 'he': 'קירונדי', 'hi': 'रुन्दी', 'hi-Latn': 'Rundi', 'hr': 'rundi', 'hsb': 'kirundišćina', 'ht': 'roundi', 'hu': 'kirundi', 'hy': 'ռունդի', 'ia': 'rundi', 'id': 'Rundi', 'ig': 'Rundi', 'is': 'rúndí', 'it': 'rundi', 'ja': 'ルンディ語', 'jv': 'Rundi', 'ka': 'რუნდი', 'kea': 'rundi', 'kgp': 'hũni', 'kk': 'рунди тілі', 'kk-Cyrl': 'рунди тілі', 'km': 'រុណ្ឌី', 'kn': 'ರುಂಡಿ', 'ko': '룬디어', 'kok': 'रुंदी', 'kok-Deva': 'रुंदी', 'ks': 'رُندی', 'ks-Arab': 'رُندی', 'ksh': 'K-Rundesch', 'ku': 'rundî', 'ky': 'рундиче', 'lb': 'Rundi-Sprooch', 'lld': 'rundi', 'lo': 'ຣຸນດິ', 'lrc': 'راندی', 'lt': 'rundi', 'lv': 'rundu', 'mai': 'रूंडी', 'mi': 'Rūniti', 'mk': 'рунди', 'ml': 'റുണ്ടി', 'mn': 'рунди', 'mr': 'रुन्दी', 'ms': 'Rundi', 'mt': 'Rundi', 'my': 'ရွန်ဒီ', 'mzn': 'روندی', 'nb': 'rundi', 'ne': 'रुन्डी', 'nl': 'Kirundi', 'nn': 'rundi', 'no': 'rundi', 'nqo': 'ߙߎ߲ߘߌߞߊ߲', 'or': 'ରୁଣ୍ଡି', 'pa': 'ਰੁੰਡੀ', 'pa-Guru': 'ਰੁੰਡੀ', 'pcm': 'Rúndi', 'pl': 'rundi', 'ps': 'رونډی', 'pt': 'rundi', 'qu': 'Rundi Simi', 'rm': 'rundi', 'rn': 'Ikirundi', 'ro': 'kirundi', 'ru': 'рунди', 'sc': 'rundi', 'sd': 'رونڊي', 'sd-Arab': 'رونڊي', 'si': 'රුන්ඩි', 'sk': 'rundčina', 'sl': 'rundščina', 'smn': 'rundi', 'so': 'Rundhi', 'sq': 'rundisht', 'sr': 'кирунди', 'sr-Cyrl': 'кирунди', 'sr-Latn': 'kirundi', 'sv': 'rundi', 'sw': 'Kirundi', 'ta': 'ருண்டி', 'te': 'రుండి', 'th': 'บุรุนดี', 'ti': 'ኪሩንዲ', 'tk': 'rundi dili', 'to': 'lea fakaluaniti', 'tr': 'Kirundi', 'ug': 'رۇندىچە', 'uk': 'рунді', 'ur': 'رونڈی', 'uz': 'rundi', 'uz-Cyrl': 'рунди', 'uz-Latn': 'rundi', 'vec': 'rundi', 'vi': 'Tiếng Rundi', 'wae': 'Rundi', 'xnr': 'रुन्दी', 'yo': 'Rúńdì', 'yrl': 'rũdi', 'yue': '隆迪文', 'yue-Hans': '隆迪文', 'yue-Hant': '隆迪文', 'zh': '隆迪语', 'zh-Hans': '隆迪语', 'zh-Hant': '隆迪文', 'zu': 'isi-Rundi'},
    'rna': {'en': 'Runa'},
    'rnd': {'en': 'Ruund'},
    'rng': {'en': 'Ronga'},
    'rnl': {'en': 'Ranglong'},
    'rnn': {'en': 'Roon'},
    'rnp': {'en': 'Rongpo'},
    'rnr': {'en': 'Nari Nari'},
    'rnw': {'en': 'Rungwa'},
    'ro': {'af': 'Roemeens', 'agq': 'Lùmanyìa', 'ak': 'Romenia kasa', 'am': 'ሮማኒያኛ', 'ar': 'الرومانية', 'as': 'ৰোমানীয়', 'asa': 'Kiromania', 'ast': 'rumanu', 'az': 'rumın', 'az-Arab': 'رومانیاجا', 'az-Cyrl': 'румын', 'az-Latn': 'rumın', 'bal-Latn': 'Rumániái', 'bas': 'Hɔp u rùmanìà', 'be': 'румынская', 'be-tarask': 'румынская', 'bem': 'Ichi Romaniani', 'bez': 'Hilomania', 'bg': 'румънски', 'bgn': 'رومانی', 'bm': 'rumanikan', 'bn': 'রোমানীয়', 'br': 'roumaneg', 'brx': 'रमानीयान', 'bs': 'rumunski', 'bs-Cyrl': 'румунски', 'bs-Latn': 'rumunski', 'ca': 'romanès', 'ca-ES-valencia': 'romanès', 'ccp': '𑄢𑄮𑄟𑄚𑄩𑄠𑄧', 'ce': 'румынийн', 'cgg': 'Oruromania', 'chr': 'ᎶᎹᏂᎠᏂ', 'ckb': 'ڕۆمانی', 'cs': 'rumunština', 'cy': 'Rwmaneg', 'da': 'rumænsk', 'dav': 'Kiromania', 'de': 'Rumänisch', 'dje': 'Rumaani senni', 'dsb': 'rumunšćina', 'dyo': 'rumeen', 'dz': 'རོ་མེ་ནི་ཡཱན་ཁ', 'ebu': 'Kĩromania', 'ee': 'romaniagbe', 'el': 'Ρουμανικά', 'el-polyton': 'Ρουμανικά', 'en': 'Romanian', 'eo': 'rumana', 'es': 'rumano', 'et': 'rumeenia', 'eu': 'errumaniera', 'ewo': 'ńkɔ́bɔ románía', 'fa': 'رومانیایی', 'ff': 'Romaneere', 'ff-Adlm': '𞤈𞤮𞤥𞤢𞤲𞤭𞤴𞤢𞤲𞤪𞤫', 'ff-Latn': 'Romaneere', 'fi': 'romania', 'fil': 'Romanian', 'fo': 'rumenskt', 'fr': 'roumain', 'fur': 'romen', 'fy': 'Roemeensk', 'ga': 'Rómáinis', 'gd': 'Romàinis', 'gl': 'romanés', 'gsw': 'Rumänisch', 'gu': 'રોમાનિયન', 'guz': 'Kiromania', 'ha': 'Romaniyanci', 'he': 'רומנית', 'hi': 'रोमानियाई', 'hi-Latn': 'Romanian', 'hr': 'rumunjski', 'hsb': 'rumunšćina', 'ht': 'roumain', 'hu': 'román', 'hy': 'ռումիներեն', 'ia': 'romaniano', 'id': 'Rumania', 'ig': 'Asụsụ Romanian', 'ii': 'ꇆꂷꆀꑸꉙ', 'is': 'rúmenska', 'it': 'rumeno', 'ja': 'ルーマニア語', 'jmc': 'Kyiromania', 'jv': 'Rumania', 'ka': 'რუმინული', 'kab': 'Tarumanit', 'kam': 'Kiromania', 'kde': 'Chilomania', 'kea': 'rumenu', 'kgp': 'homẽnũ', 'khq': 'Rumaani senni', 'ki': 'Kiromania', 'kk': 'румын тілі', 'kk-Cyrl': 'румын тілі', 'kln': 'kutitab Romaniek', 'km': 'រូម៉ានី', 'kn': 'ರೊಮೇನಿಯನ್', 'ko': '루마니아어', 'kok': 'रोमानियन', 'kok-Deva': 'रोमानियन', 'ks': 'رومٲنی', 'ks-Arab': 'رومٲنی', 'ksb': 'Kiomania', 'ksf': 'rirɔmán', 'ksh': 'Rumänesch', 'ku': 'romanî', 'kxv': 'romaniyati', 'kxv-Deva': 'रोमानियाति', 'kxv-Latn': 'romaniyati', 'kxv-Orya': 'ରମାନିଆ ତି', 'kxv-Telu': 'రోమేనియన్', 'ky': 'румынча', 'lag': 'Kɨromanía', 'lb': 'Rumänesch', 'lg': 'Lulomaniya', 'lkt': 'Romanian Iyápi', 'lld': 'rumen', 'ln': 'liromani', 'lo': 'ໂຣແມນຽນ', 'lrc': 'رومانیایی', 'lt': 'rumunų', 'lu': 'Liromani', 'luo': 'Kiromania', 'luy': 'Kiromania', 'lv': 'rumāņu', 'mai': 'रोमानियाई', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lromania', 'mer': 'Kĩromania', 'mfe': 'roumin', 'mg': 'Romanianina', 'mgh': 'Iromania', 'mi': 'Romeinia', 'mk': 'романски', 'ml': 'റൊമാനിയൻ', 'mn': 'румын', 'mr': 'रोमानियन', 'ms': 'Romania', 'mt': 'Rumen', 'mua': 'Romaniya', 'my': 'ရိုမေနီယား', 'mzn': 'رومانیایی', 'naq': 'Romaniaǁî gowab', 'nb': 'rumensk', 'nd': 'isi-Romani', 'ne': 'रोमानियाली', 'nl': 'Roemeens', 'nmg': 'Kiɛl bó rumɛ̂n', 'nn': 'rumensk', 'no': 'rumensk', 'nqo': 'ߙߏߡߍߞߊ߲', 'nus': 'Thok ji̱ röm', 'nyn': 'Oruromania', 'om': 'Afaan Romaniyaa', 'or': 'ରୋମାନିଆନ୍', 'pa': 'ਰੋਮਾਨੀਆਈ', 'pa-Guru': 'ਰੋਮਾਨੀਆਈ', 'pcm': 'Romániá Lángwej', 'pl': 'rumuński', 'ps': 'رومانیایی', 'pt': 'romeno', 'qu': 'Rumano Simi', 'rm': 'rumen', 'rn': 'Ikinyarumaniya', 'ro': 'română', 'rof': 'Kiromania', 'ru': 'румынский', 'rw': 'Ikinyarumaniya', 'rwk': 'Kyiromania', 'sah': 'Румыынныы', 'saq': 'Kiromania', 'sbp': 'Ishilomaniya', 'sc': 'rumenu', 'sd': 'روماني', 'sd-Arab': 'روماني', 'se': 'romániagiella', 'seh': 'romeno', 'ses': 'Rumaani senni', 'sg': 'Rumëen', 'shi': 'ⵜⴰⵔⵓⵎⴰⵏⵉⵜ', 'shi-Latn': 'Tarumanit', 'shi-Tfng': 'ⵜⴰⵔⵓⵎⴰⵏⵉⵜ', 'si': 'රොමේනියානු', 'sk': 'rumunčina', 'sl': 'romunščina', 'smn': 'romaniakielâ', 'sn': 'chiRomanian', 'so': 'Romanka', 'sq': 'rumanisht', 'sr': 'румунски', 'sr-Cyrl': 'румунски', 'sr-Latn': 'rumunski', 'sv': 'rumänska', 'sw': 'Kiromania', 'syr': 'ܪܘܡܢܐܝܬ', 'ta': 'ரோமேனியன்', 'te': 'రొమేనియన్', 'teo': 'Kiromania', 'tg': 'руминӣ', 'th': 'โรมาเนีย', 'ti': 'ሩማንኛ', 'tk': 'rumyn dili', 'to': 'lea fakalōmenia', 'tr': 'Rumence', 'tt': 'румын', 'twq': 'Rumaani senni', 'tzm': 'Taṛumanit', 'ug': 'رومىنچە', 'uk': 'румунська', 'ur': 'رومینین', 'uz': 'rumincha', 'uz-Cyrl': 'руминча', 'uz-Latn': 'rumincha', 'vai': 'ꖄꕆꕇꘂꘋ', 'vai-Latn': 'Romíniyɛŋ', 'vai-Vaii': 'ꖄꕆꕇꘂꘋ', 'vec': 'romen', 'vi': 'Tiếng Romania', 'vun': 'Kyiromania', 'wae': 'Rumäniš', 'wo': 'Rumaniyee', 'xnr': 'रोमानियाई', 'xog': 'Olulomaniya', 'yav': 'nulumɛ́ŋɛ', 'yi': 'רומעניש', 'yo': 'Èdè Romania', 'yrl': 'rumenu', 'yue': '羅馬尼亞文', 'yue-Hans': '罗马尼亚文', 'yue-Hant': '羅馬尼亞文', 'zgh': 'ⵜⴰⵔⵓⵎⴰⵏⵉⵜ', 'zh': '罗马尼亚语', 'zh-Hans': '罗马尼亚语', 'zh-Hant': '羅馬尼亞文', 'zu': 'isi-Romanian'},
    'ro-MD': {'am': 'ሞልዳቪያንኛ', 'ar': 'المولدوفية', 'as': 'মোল্ডাভিয়ান', 'ast': 'moldavu', 'az': 'moldav', 'az-Latn': 'moldav', 'bal-Latn': 'Rumániái (MD)', 'be': 'малдаўская', 'be-tarask': 'малдаўская', 'bg': 'молдовски', 'bgn': 'مالداوی', 'bn': 'মলদাভিয়', 'br': 'moldoveg', 'brx': 'मोल्डेवियन्', 'bs': 'moldavski', 'bs-Cyrl': 'молдавски', 'bs-Latn': 'moldavski', 'byn': 'ሞልዳቫዊና', 'ca': 'moldau', 'ca-ES-valencia': 'moldau', 'ccp': '𑄟𑄧𑄣𑄴𑄘𑄞𑄨𑄠𑄧', 'ce': 'молдавийн', 'chr': 'ᎹᎵᏙᏫᎠ ᏣᎹᏂᎠᏂ', 'ckb': 'مۆڵداڤی', 'cs': 'moldavština', 'cy': 'Moldofeg', 'da': 'moldovisk', 'de': 'Moldauisch', 'dsb': 'moldawišćina', 'el': 'Μολδαβικά', 'el-polyton': 'Μολδαβικά', 'en': 'Moldavian', 'en-Dsrt': '𐐣𐐬𐑊𐐼𐐩𐑂𐐨𐐲𐑌', 'es': 'moldavo', 'et': 'moldova', 'eu': 'moldaviera', 'fa': 'مولداویایی', 'ff-Adlm': '𞤃𞤮𞤤𞤣𞤢𞤾𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'moldova', 'fil': 'Moldavian', 'fo': 'moldaviskt', 'fr': 'moldave', 'fur': 'moldâf', 'fy': 'Moldavysk', 'ga': 'Moldáivis', 'gd': 'Moldobhais', 'gl': 'moldavo', 'gsw': 'Moldawisch', 'gu': 'મોલડાવિયન', 'he': 'מולדבית', 'hi': 'मोलडावियन', 'hi-Latn': 'Moldavian', 'hr': 'moldavski', 'hsb': 'moldawšćina', 'ht': 'moldave', 'hu': 'moldvai', 'hy': 'մոլդովերեն', 'ia': 'moldavo', 'id': 'Moldavia', 'ig': 'Moldavian', 'is': 'moldóvska', 'it': 'moldavo', 'ja': 'モルダビア語', 'ka': 'მოლდავური', 'kea': 'rumenu moldáviku', 'kgp': 'mãrnavijo', 'kk': 'молдован тілі', 'kk-Cyrl': 'молдован тілі', 'km': 'ម៉ុលដាវី', 'kn': 'ಮಾಲ್ಡೇವಿಯನ್', 'ko': '몰도바어', 'kok': 'मोल्डावियन्', 'kok-Deva': 'मोल्डावियन्', 'ks': 'مولداوِیَن', 'ks-Arab': 'مولداوِیَن', 'kxv': 'molḍaviati', 'kxv-Deva': 'मोलडावियन', 'kxv-Latn': 'molḍaviati', 'kxv-Orya': 'ମଲଡୋଭିଆନ୍ ତି', 'kxv-Telu': 'మెల్డావియన్', 'ky': 'молдованча', 'lb': 'Moldawesch', 'lld': 'moldavich', 'lo': 'ໂມດາວຽນ', 'lrc': 'رومانیایی مولداڤی', 'lt': 'moldavų', 'lv': 'moldāvu', 'mk': 'молдавски', 'ml': 'മോൾഡാവിയൻ', 'mn': 'молдав', 'mr': 'मोल्डाव्हियन', 'ms': 'Moldavia', 'mt': 'Moldovan', 'my': 'မော်လဒိုဗာ', 'mzn': 'مولداوی', 'nb': 'moldovsk', 'ne': 'मोल्डाभियाली', 'nn': 'moldavisk', 'no': 'moldovsk', 'nqo': 'ߡߐߟߑߘߊߝ߭ߌߞߊ߲', 'or': 'ମୋଲଡୋଭିଆନ୍', 'pa': 'ਮੋਲਡਾਵੀਆਈ', 'pa-Guru': 'ਮੋਲਡਾਵੀਆਈ', 'pl': 'mołdawski', 'ps': 'مولداویایی', 'pt': 'moldávio', 'rm': 'moldav', 'ru': 'молдавский', 'sc': 'moldavu', 'sd': 'مالديوي', 'sd-Arab': 'مالديوي', 'si': 'මොල්ඩවිආනු', 'sk': 'moldavčina', 'sl': 'moldavščina', 'sq': 'moldavisht', 'sr': 'молдавски', 'sr-Cyrl': 'молдавски', 'sr-Latn': 'moldavski', 'sv': 'moldaviska', 'sw': 'Kimoldova cha Romania', 'ta': 'மோல்டாவியன்', 'te': 'మొల్డావియన్', 'th': 'มอลโดวา', 'ti': 'ሞልዶቨኛ', 'tig': 'ሞልዳቫዊና', 'tk': 'moldaw dili', 'to': 'lea fakamolitāvia', 'tr': 'Moldovaca', 'uk': 'молдавська', 'ur': 'مالدووا', 'uz': 'moldovan', 'uz-Latn': 'moldovan', 'vec': 'moldavo', 'vi': 'Tiếng Moldova', 'yrl': 'mudáwiu', 'yue': '摩爾多瓦文', 'yue-Hans': '摩尔多瓦文', 'yue-Hant': '摩爾多瓦文', 'zh': '摩尔多瓦语', 'zh-Hans': '摩尔多瓦语', 'zh-Hant': '摩爾多瓦文', 'zu': 'isi-Moldavian'},
    'roa': {'en': 'Romance languages'},
    'roa-gal': {'en': 'Gallo'},
    'roa-leo': {'en': 'Leonese'},
    'roa-oan': {'en': 'Navarro-Aragonese'},
    'roa-oca': {'en': 'Old Catalan'},
    'roa-oit': {'en': 'Old Italian'},
    'roa-ole': {'en': 'Old Leonese'},
    'roa-opt': {'en': 'Old Portuguese'},
    'roa-tar': {'en': 'Tarantino'},
    'rob': {'en': "Tae'"},
    'roc': {'en': 'Cacgia Roglai'},
    'rod': {'en': 'Rogo'},
    'roe': {'en': 'Ronji'},
    'rof': {'af': 'Rombo', 'am': 'ሮምቦ', 'ar': 'الرومبو', 'as': 'ৰোম্বো', 'ast': 'rombo', 'az': 'rombo', 'az-Cyrl': 'ромбо', 'az-Latn': 'rombo', 'bal-Latn': 'Rombó', 'be': 'ромба', 'be-tarask': 'ромба', 'bg': 'ромбо', 'bgn': 'رومبویی', 'bn': 'রম্বো', 'br': 'rombo', 'brx': 'रम्ब’', 'bs': 'rombo', 'bs-Cyrl': 'ромбо', 'bs-Latn': 'rombo', 'ca': 'rombo', 'ca-ES-valencia': 'rombo', 'ccp': '𑄢𑄧𑄟𑄴𑄝𑄮', 'ce': 'ромбо', 'chr': 'ᎶᎹᏉ', 'ckb': 'ڕۆمبۆ', 'cs': 'rombo', 'cy': 'Rombo', 'da': 'rombo', 'de': 'Rombo', 'dsb': 'rombo', 'ee': 'rombogbe', 'el': 'Ρόμπο', 'el-polyton': 'Ρόμπο', 'en': 'Rombo', 'es': 'rombo', 'et': 'rombo', 'eu': 'romboa', 'fa': 'رومبویی', 'ff-Adlm': '𞤈𞤮𞤥𞤦𞤮𞥅𞤪𞤫', 'fi': 'rombo', 'fil': 'Rombo', 'fo': 'rombo', 'fr': 'rombo', 'fy': 'Rombo', 'ga': 'Rombo', 'gd': 'Rombo', 'gl': 'rombo', 'gu': 'રોમ્બો', 'ha': 'Rombo', 'he': 'רומבו', 'hi': 'रोम्बो', 'hi-Latn': 'Rombo', 'hr': 'rombo', 'hsb': 'rombo', 'ht': 'rombo', 'hu': 'rombo', 'hy': 'ռոմբո', 'ia': 'rombo', 'id': 'Rombo', 'ig': 'Rombo', 'is': 'rombó', 'it': 'rombo', 'ja': 'ロンボ語', 'jv': 'Rombo', 'ka': 'რომბო', 'kea': 'rombu', 'kgp': 'hãmo', 'kk': 'ромбо тілі', 'kk-Cyrl': 'ромбо тілі', 'km': 'រុមបូ', 'kn': 'ರೊಂಬೊ', 'ko': '롬보어', 'kok': 'रोम्बो', 'kok-Deva': 'रोम्बो', 'ksh': 'Kirombo', 'ku': 'rombo', 'ky': 'ромбочо', 'lb': 'Rombo', 'lo': 'ຣົມໂບ', 'lrc': 'رومبو', 'lt': 'rombo', 'lv': 'rombo', 'mai': 'रोमबो', 'mi': 'Romopo', 'mk': 'ромбо', 'ml': 'റോംബോ', 'mn': 'ромбо', 'mr': 'रोम्बो', 'ms': 'Rombo', 'mt': 'Rombo', 'my': 'ရွမ်ဘို', 'mzn': 'رومبو', 'nb': 'rombo', 'ne': 'रोम्बो', 'nl': 'Rombo', 'nn': 'rombo', 'no': 'rombo', 'nqo': 'ߙߏ߲ߓߏߞߊ߲', 'or': 'ରୋମ୍ବୋ', 'pa': 'ਰੋਮਬੋ', 'pa-Guru': 'ਰੋਮਬੋ', 'pcm': 'Rómbo', 'pl': 'rombo', 'ps': 'رومبو', 'pt': 'rombo', 'qu': 'Rombo Simi', 'rm': 'rombo', 'ro': 'rombo', 'rof': 'Kihorombo', 'ru': 'ромбо', 'sc': 'rombo', 'sd': 'رومبو', 'sd-Arab': 'رومبو', 'si': 'රෝම්බෝ', 'sk': 'rombo', 'sl': 'rombo', 'smn': 'rombo', 'so': 'Rombo', 'sq': 'romboisht', 'sr': 'ромбо', 'sr-Cyrl': 'ромбо', 'sr-Latn': 'rombo', 'sv': 'rombo', 'sw': 'Kirombo', 'ta': 'ரோம்போ', 'te': 'రోంబో', 'th': 'รอมโบ', 'ti': 'ሮምቦ', 'tk': 'rombo dili', 'to': 'lea fakalomipō', 'tr': 'Rombo', 'ug': 'رومبوچە', 'uk': 'ромбо', 'ur': 'رومبو', 'uz': 'rombo', 'uz-Cyrl': 'ромбоча', 'uz-Latn': 'rombo', 'vec': 'ronbo', 'vi': 'Tiếng Rombo', 'xnr': 'रोम्बो', 'yo': 'Róńbò', 'yrl': 'rũbu', 'yue': '蘭博文', 'yue-Hans': '兰博文', 'yue-Hant': '蘭博文', 'zh': '兰博语', 'zh-Hans': '兰博语', 'zh-Hant': '蘭博文', 'zu': 'isi-Rombo'},
    'rog': {'en': 'Northern Roglai'},
    'rol': {'en': 'Romblomanon'},
    'rom': {'ar': 'الغجرية', 'ast': 'romaní', 'az': 'roman', 'az-Latn': 'roman', 'bg': 'ромски', 'bn': 'রোমানি', 'br': 'romanieg', 'brx': 'रुमानी', 'bs': 'romani', 'bs-Cyrl': 'романи', 'bs-Latn': 'romani', 'ca': 'romaní', 'ca-ES-valencia': 'romaní', 'ccp': '𑄢𑄮𑄟𑄚𑄨', 'cs': 'romština', 'cy': 'Romani', 'da': 'romani', 'de': 'Romani', 'el': 'Ρομανί', 'el-polyton': 'Ρομανί', 'en': 'Romany', 'es': 'romaní', 'et': 'mustlaskeel', 'fa': 'رومانویی', 'fi': 'romani', 'fr': 'romani', 'fy': 'Romani', 'ga': 'Romainis', 'gd': 'Romanais', 'gsw': 'Zigüünerschpraach', 'gu': 'રોમાની', 'he': 'רומאני', 'hi': 'रोमानी', 'hi-Latn': 'Romany', 'hr': 'romski', 'ht': 'romani', 'hu': 'roma', 'hy': 'ռոմաներեն', 'id': 'Romani', 'is': 'romaní', 'it': 'romani', 'ja': 'ロマーニー語', 'ka': 'ბოშური', 'kgp': 'homỹnĩ', 'kn': 'ರೋಮಾನಿ', 'ko': '집시어', 'ks': 'رومَنی', 'ks-Arab': 'رومَنی', 'lb': 'Romani', 'lo': 'ໂຣເມນີ', 'lt': 'romų', 'lv': 'čigānu', 'mk': 'ромски', 'ml': 'റൊമാനി', 'mr': 'रोमानी', 'mt': 'Romanesk', 'nb': 'romani', 'nl': 'Romani', 'nn': 'romani', 'no': 'romani', 'or': 'ରୋମାନି', 'os': 'цигайнаг', 'pl': 'cygański', 'pt': 'romani', 'rm': 'romani', 'ro': 'romani', 'ru': 'цыганский', 'sk': 'rómčina', 'sl': 'romščina', 'smn': 'roomaankielâ', 'sr': 'ромски', 'sr-Cyrl': 'ромски', 'sr-Latn': 'romski', 'sv': 'romani', 'ta': 'ரோமானி', 'te': 'రోమానీ', 'th': 'โรมานี', 'to': 'lea fakalomani', 'tr': 'Romanca', 'ug': 'سىگانچە', 'uk': 'циганська', 'vi': 'Tiếng Romany', 'yrl': 'rumani', 'yue': '吉普賽文', 'yue-Hans': '吉普赛文', 'yue-Hant': '吉普賽文', 'zh': '吉普赛语', 'zh-Hans': '吉普赛语', 'zh-Hant': '吉普賽文'},
    'roo': {'en': 'Rotokas'},
    'rop': {'en': 'Kriol'},
    'ror': {'en': 'Rongga'},
    'rou': {'en': 'Runga'},
    'row': {'en': 'Dela-Oenale'},
    'rpn': {'en': 'Repanbitip'},
    'rpt': {'en': 'Rapting'},
    'rri': {'en': 'Ririo'},
    'rro': {'en': 'Waima'},
    'rrt': {'en': 'Arritinngithigh'},
    'rsb': {'en': 'Romano-Serbian'},
    'rsi': {'en': 'Rennellese Sign Language'},
    'rsl': {'en': 'Russian Sign Language'},
    'rsm': {'en': 'Miriwoong Sign Language'},
    'rtc': {'en': 'Rungtu Chin'},
    'rth': {'en': 'Ratahan'},
    'rtm': {'ast': 'rotumanu', 'cs': 'rotumanština', 'cy': 'Rotumaneg', 'de': 'Rotumanisch', 'en': 'Rotuman', 'et': 'rotuma', 'fi': 'rotuma', 'fr': 'rotuman', 'hi-Latn': 'Rotuman', 'ht': 'rotuman', 'hy': 'ռոտուման', 'id': 'Rotuma', 'it': 'rotumano', 'ja': 'ロツマ語', 'lb': 'Rotumanesch', 'lt': 'rotumanų', 'mk': 'ротумански', 'nb': 'rotumansk', 'nl': 'Rotumaans', 'nn': 'rotumansk', 'no': 'rotumansk', 'pl': 'rotumański', 'sv': 'rotumänska', 'th': 'โรทูมัน', 'to': 'lea fakalotuma', 'tr': 'Rotuman', 'yue': '羅圖馬島文', 'yue-Hans': '罗图马岛文', 'yue-Hant': '羅圖馬島文', 'zh-Hant': '羅圖馬島文'},
    'rts': {'en': 'Yurats'},
    'rtw': {'en': 'Rathawi'},
    'ru': {'af': 'Russies', 'agq': 'Lushìa', 'ak': 'Rahyia kasa', 'am': 'ራሽያኛ', 'ar': 'الروسية', 'as': 'ৰাছিয়ান', 'asa': 'Kiruthi', 'ast': 'rusu', 'az': 'rus', 'az-Arab': 'روسجا', 'az-Cyrl': 'рус', 'az-Latn': 'rus', 'bal-Latn': 'Rusi', 'bas': 'Hɔp u ruslànd', 'be': 'руская', 'be-tarask': 'руская', 'bem': 'Ichi Rusiani', 'bez': 'Hilusi', 'bg': 'руски', 'bgn': 'اوروسی', 'blo': 'gɩrɔɔshɩyaja', 'bm': 'irisikan', 'bn': 'রুশ', 'bo': 'ཨུ་རུ་སུ་སྐད་', 'br': 'rusianeg', 'brx': 'रासियान', 'bs': 'ruski', 'bs-Cyrl': 'руски', 'bs-Latn': 'ruski', 'ca': 'rus', 'ca-ES-valencia': 'rus', 'ccp': '𑄢𑄪𑄌𑄴', 'ce': 'оьрсийн', 'ceb': 'Russian', 'cgg': 'Orurrasha', 'chr': 'ᏲᏅᎯ', 'ckb': 'ڕووسی', 'cs': 'ruština', 'cv': 'вырӑс', 'cy': 'Rwseg', 'da': 'russisk', 'dav': 'Kirusi', 'de': 'Russisch', 'dje': 'Ruusi senni', 'doi': 'रूसी', 'dsb': 'rušćina', 'dyo': 'rus', 'dz': 'ཨུ་རུ་སུའི་ཁ', 'ebu': 'Kĩrusi', 'ee': 'Russiagbe', 'el': 'Ρωσικά', 'el-polyton': 'Ρωσικά', 'en': 'Russian', 'eo': 'rusa', 'es': 'ruso', 'et': 'vene', 'eu': 'errusiera', 'ewo': 'ǹkɔ́bɔ rúsian', 'fa': 'روسی', 'ff': 'Riis', 'ff-Adlm': '𞤈𞤮𞥅𞤧𞤭𞤴𞤢𞤲𞤪𞤫', 'ff-Latn': 'Riis', 'fi': 'venäjä', 'fil': 'Russian', 'fo': 'russiskt', 'fr': 'russe', 'fur': 'rus', 'fy': 'Russysk', 'ga': 'Rúisis', 'gd': 'Ruisis', 'gl': 'ruso', 'gsw': 'Russisch', 'gu': 'રશિયન', 'guz': 'Kirusi', 'ha': 'Rashanci', 'haw': 'Lūkia', 'he': 'רוסית', 'hi': 'रूसी', 'hi-Latn': 'Russian', 'hr': 'ruski', 'hsb': 'rušćina', 'ht': 'russe', 'hu': 'orosz', 'hy': 'ռուսերեն', 'ia': 'russo', 'id': 'Rusia', 'ie': 'russ', 'ig': 'Asụsụ Russia', 'ii': 'ꊉꇩꉙ', 'is': 'rússneska', 'it': 'russo', 'ja': 'ロシア語', 'jmc': 'Kyirusi', 'jv': 'Rusia', 'ka': 'რუსული', 'kab': 'Tarusit', 'kam': 'Kirusi', 'kde': 'Chilusi', 'kea': 'rusu', 'kgp': 'huso', 'khq': 'Ruusi senni', 'ki': 'Kĩracia', 'kk': 'орыс тілі', 'kk-Cyrl': 'орыс тілі', 'kln': 'kutitab Russia', 'km': 'រុស្ស៊ី', 'kn': 'ರಷ್ಯನ್', 'ko': '러시아어', 'kok': 'रशियन', 'kok-Deva': 'रशियन', 'ks': 'روٗسی', 'ks-Arab': 'روٗسی', 'ks-Deva': 'रूसी', 'ksb': 'Kilusi', 'ksf': 'rirís', 'ksh': 'Rußßesch', 'ku': 'rûsî', 'kxv': 'rusiyati', 'kxv-Deva': 'रुसिया ति', 'kxv-Latn': 'rusiyati', 'kxv-Orya': 'ରୁଷିଆ ତି', 'kxv-Telu': 'రస్వన్', 'ky': 'орусча', 'lag': 'Kɨrúusi', 'lb': 'Russesch', 'lg': 'Lulasa', 'lkt': 'Russia Iyápi', 'lld': 'rus', 'ln': 'lirisí', 'lo': 'ລັດເຊຍ', 'lrc': 'روٙسی', 'lt': 'rusų', 'lu': 'Lirisi', 'luo': 'Kirusi', 'luy': 'Kirusi', 'lv': 'krievu', 'mai': 'रूसी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lrusi', 'mer': 'Kĩrashia', 'mfe': 'ris', 'mg': 'Rosianina', 'mgh': 'Irisi', 'mi': 'Ruhiana', 'mk': 'руски', 'ml': 'റഷ്യൻ', 'mn': 'орос', 'mni': 'রুসিয়ান', 'mni-Beng': 'রুসিয়ান', 'mr': 'रशियन', 'ms': 'Rusia', 'mt': 'Russu', 'mua': 'Russiya', 'my': 'ရုရှ', 'mzn': 'روسی', 'naq': 'Russiaǁî gowab', 'nb': 'russisk', 'nd': 'isi-Rashiya', 'ne': 'रसियाली', 'nl': 'Russisch', 'nmg': 'Kiɛl russia', 'nn': 'russisk', 'no': 'russisk', 'nqo': 'ߌ߬ߙߛߌ߬ߞߊ߲', 'nus': 'Thok ra̱ciaani', 'nyn': 'Orurrasha', 'om': 'Afaan Rushiyaa', 'or': 'ରୁଷିୟ', 'os': 'уырыссаг', 'pa': 'ਰੂਸੀ', 'pa-Guru': 'ਰੂਸੀ', 'pcm': 'Rọshiá Lángwej', 'pl': 'rosyjski', 'ps': 'روسي', 'pt': 'russo', 'qu': 'Ruso Simi', 'rm': 'russ', 'rn': 'Ikirusiya', 'ro': 'rusă', 'rof': 'Kirusi', 'ru': 'русский', 'rw': 'Ikirusiya', 'rwk': 'Kyirusi', 'sa': 'रष्यदेशीय भाषा:', 'sah': 'Нууччалыы', 'saq': 'Kirusi', 'sat': 'ᱨᱩᱥᱤᱭᱟᱱ', 'sat-Olck': 'ᱨᱩᱥᱤᱭᱟᱱ', 'sbp': 'Ishilusi', 'sc': 'russu', 'sd': 'روسي', 'sd-Arab': 'روسي', 'sd-Deva': 'रशियनु', 'se': 'ruoššagiella', 'seh': 'russo', 'ses': 'Ruusi senni', 'sg': 'Rûsi', 'shi': 'ⵜⴰⵔⵓⵙⵉⵜ', 'shi-Latn': 'Tarusit', 'shi-Tfng': 'ⵜⴰⵔⵓⵙⵉⵜ', 'si': 'රුසියානු', 'sk': 'ruština', 'sl': 'ruščina', 'smn': 'ruošâkielâ', 'sn': 'chiRashiya', 'so': 'Ruush', 'sq': 'rusisht', 'sr': 'руски', 'sr-Cyrl': 'руски', 'sr-Latn': 'ruski', 'su': 'Rusia', 'su-Latn': 'Rusia', 'sv': 'ryska', 'sw': 'Kirusi', 'syr': 'ܐܘܪܘܣܢܝܬ', 'ta': 'ரஷியன்', 'te': 'రష్యన్', 'teo': 'Kirusi', 'tg': 'русӣ', 'th': 'รัสเซีย', 'ti': 'ሩስኛ', 'tk': 'rus dili', 'to': 'lea fakalūsia', 'tr': 'Rusça', 'tt': 'рус', 'twq': 'Ruusi senni', 'tzm': 'Tarusit', 'ug': 'رۇسچە', 'uk': 'російська', 'ur': 'روسی', 'uz': 'ruscha', 'uz-Cyrl': 'русча', 'uz-Latn': 'ruscha', 'vai': 'ꗐꖺꔻꘂꘋ', 'vai-Latn': 'Rɔshiyɛŋ', 'vai-Vaii': 'ꗐꖺꔻꘂꘋ', 'vec': 'ruso', 'vi': 'Tiếng Nga', 'vun': 'Kyirusi', 'wae': 'Rusiš', 'wo': 'Rus', 'xh': 'Isi-Russian', 'xnr': 'रूसी', 'xog': 'Olulasa', 'yav': 'nulúse', 'yi': 'רוסיש', 'yo': 'Èdè Rọ́ṣíà', 'yrl': 'rusu', 'yue': '俄文', 'yue-Hans': '俄文', 'yue-Hant': '俄文', 'zgh': 'ⵜⴰⵔⵓⵙⵉⵜ', 'zh': '俄语', 'zh-Hans': '俄语', 'zh-Hant': '俄文', 'zu': 'isi-Russian'},
    'rub': {'en': 'Gungu'},
    'ruc': {'en': 'Ruuli'},
    'rue': {'ast': 'rusyn', 'cs': 'rusínština', 'de': 'Russinisch', 'en': 'Rusyn', 'et': 'russiini', 'fi': 'ruteeni', 'fr': 'ruthène', 'gd': 'Rusyn', 'hi-Latn': 'Rusyn', 'ht': 'ruthène', 'hy': 'ռուսիներեն', 'it': 'ruteno', 'ja': 'ルシン語', 'ko': '루신어', 'lb': 'Russinesch', 'lt': 'rusinų', 'mk': 'русински', 'nb': 'rusinsk', 'nl': 'Roetheens', 'nn': 'rusinsk', 'no': 'rusinsk', 'pl': 'rusiński', 'sv': 'rusyn', 'th': 'รูซิน', 'to': 'lea fakalusini', 'tr': 'Rusince', 'yi': 'רוסיניש', 'yue': '盧森尼亞文', 'yue-Hans': '卢森尼亚文', 'yue-Hant': '盧森尼亞文', 'zh-Hant': '盧森尼亞文'},
    'ruf': {'en': 'Luguru'},
    'rug': {'ast': 'roviana', 'cs': 'rovianština', 'de': 'Roviana', 'en': 'Roviana', 'et': 'roviana', 'fi': 'roviana', 'fr': 'roviana', 'gd': 'Roviana', 'hi-Latn': 'Roviana', 'ht': 'roviana', 'hy': 'ռովիանա', 'it': 'roviana', 'ja': 'ロヴィアナ語', 'lb': 'Roviana', 'lt': 'Rovianos', 'mk': 'ровијански', 'nb': 'roviana', 'nl': 'Roviana', 'nn': 'roviana', 'no': 'roviana', 'pl': 'roviana', 'sv': 'rovianska', 'th': 'โรเวียนา', 'to': 'lea fakaloviana', 'tr': 'Roviana', 'yue': '羅維阿納文', 'yue-Hans': '罗维阿纳文', 'yue-Hant': '羅維阿納文', 'zh-Hant': '羅維阿納文'},
    'ruh': {'en': 'Ruga'},
    'rui': {'en': 'Rufiji'},
    'ruk': {'en': 'Che'},
    'ruo': {'en': 'Istro Romanian'},
    'rup': {'af': 'Aromanies', 'am': 'አሮማንያን', 'ar': 'الأرومانيان', 'as': 'আৰোমানীয়', 'ast': 'aromanianu', 'az': 'aroman', 'az-Cyrl': 'ароман', 'az-Latn': 'aroman', 'be': 'арумунская', 'be-tarask': 'арумунская', 'bg': 'арумънски', 'bgn': 'آرومانی', 'bn': 'আরোমেনিয়', 'br': 'aroumaneg', 'brx': 'आरोमानियान', 'bs': 'arumunski', 'bs-Cyrl': 'ароманијски', 'bs-Latn': 'arumunski', 'ca': 'aromanès', 'ca-ES-valencia': 'aromanès', 'ccp': '𑄃𑄢𑄴𑄟𑄬𑄚𑄨𑄠𑄚𑄴', 'ce': 'аруминийн', 'chr': 'ᎠᏬᎹᏂᎠᏂ', 'ckb': 'ئارمۆمانی', 'cs': 'arumunština', 'cy': 'Aromaneg', 'da': 'arumænsk', 'de': 'Aromunisch', 'dsb': 'armanianšćina', 'el': 'Αρομανικά', 'el-polyton': 'Αρομανικά', 'en': 'Aromanian', 'es': 'arrumano', 'et': 'aromuuni', 'eu': 'aromaniera', 'fa': 'آرومانی', 'ff-Adlm': '𞤀𞤪𞤮𞤥𞤢𞤲𞤭𞥅𞤪𞤫', 'fi': 'aromania', 'fil': 'Aromanian', 'fo': 'aromenskt', 'fr': 'aroumain', 'fy': 'Aromaniaansk', 'ga': 'Arómáinis', 'gd': 'Aromanais', 'gl': 'aromanés', 'gsw': 'Aromunisch', 'gu': 'અરોમેનિયન', 'ha': 'Aromaniyanci', 'he': 'ארומנית', 'hi': 'अरोमानियन', 'hi-Latn': 'Aromanian', 'hr': 'aromunski', 'hsb': 'armanianšćina', 'ht': 'aroumain', 'hu': 'aromán', 'hy': 'արոմաներեն', 'ia': 'aromaniano', 'id': 'Aromania', 'ig': 'Aromanian', 'is': 'arúmenska', 'it': 'arumeno', 'ja': 'アルーマニア語', 'jv': 'Aromanian', 'ka': 'არომანული', 'kgp': 'aromẽnũ', 'kk': 'арумын тілі', 'kk-Cyrl': 'арумын тілі', 'km': 'អារ៉ូម៉ានី', 'kn': 'ಅರೋಮಾನಿಯನ್', 'ko': '아로마니아어', 'kok': 'आरोमेनियन', 'kok-Deva': 'आरोमेनियन', 'ks': 'اَرومانی', 'ks-Arab': 'اَرومانی', 'ksh': 'Arromahnesch', 'ku': 'aromanî', 'ky': 'аромунча', 'lb': 'Aromunesch', 'lo': 'ອາໂຣມານຽນ', 'lt': 'aromanių', 'lv': 'aromūnu', 'mai': 'अरोमानियाई', 'mi': 'Aromeiniana', 'mk': 'влашки', 'ml': 'ആരോമാനിയൻ', 'mn': 'ароманы', 'mr': 'अरोमानियन', 'ms': 'Aromanian', 'mt': 'Aromanjan', 'my': 'အာရိုမန်းနီးယန်း', 'nb': 'aromansk', 'ne': 'अरोमानीयाली', 'nl': 'Aroemeens', 'nn': 'arumensk', 'no': 'aromansk', 'or': 'ଆରୋମାନିଆନ୍', 'pa': 'ਅਰੋਮੀਨੀਆਈ', 'pa-Guru': 'ਅਰੋਮੀਨੀਆਈ', 'pcm': 'Arómánian', 'pl': 'arumuński', 'ps': 'اروماني', 'pt': 'aromeno', 'qu': 'Arrumano', 'rm': 'aromunic', 'ro': 'aromână', 'ru': 'арумынский', 'sat': 'ᱟᱨᱚᱢᱟᱱᱤᱭᱟᱹᱱ', 'sat-Olck': 'ᱟᱨᱚᱢᱟᱱᱤᱭᱟᱹᱱ', 'sc': 'arumenu', 'sd': 'ارومينين', 'sd-Arab': 'ارومينين', 'si': 'ඇරොමානියානු', 'sk': 'arumunčina', 'sl': 'aromunščina', 'smn': 'aromaniakielâ', 'so': 'U dhashay Aromania', 'sq': 'vllahisht', 'sr': 'цинцарски', 'sr-Cyrl': 'цинцарски', 'sr-Latn': 'cincarski', 'sv': 'arumänska', 'sw': 'Kiaromania', 'ta': 'அரோமானியன்', 'te': 'ఆరోమేనియన్', 'th': 'อาโรมาเนีย', 'ti': 'ኣሩማንኛ', 'tk': 'arumyn dili', 'to': 'lea fakaʻalomania', 'tr': 'Ulahça', 'ug': 'ئارومانچە', 'uk': 'арумунська', 'ur': 'ارومانی', 'uz': 'arumin', 'uz-Cyrl': 'арумин', 'uz-Latn': 'arumin', 'vec': 'aromen', 'vi': 'Tiếng Aromania', 'xnr': 'अरोमानियन', 'yo': 'Èdè Aromani', 'yrl': 'arumenu', 'yue': '羅馬尼亞語系', 'yue-Hans': '罗马尼亚语系', 'yue-Hant': '羅馬尼亞語系', 'zh': '阿罗马尼亚语', 'zh-Hans': '阿罗马尼亚语', 'zh-Hant': '羅馬尼亞語系', 'zu': 'isi-Aromanian', 'rup': 'armãneashti'},
    'ruq': {'en': 'Megleno Romanian'},
    'rut': {'en': 'Rutul'},
    'ruu': {'en': 'Lanas Lobu'},
    'ruy': {'en': 'Mala (Nigeria)'},
    'ruz': {'en': 'Ruma'},
    'rw': {'af': 'Rwandees', 'agq': 'Lùwandà', 'ak': 'Rewanda kasa', 'am': 'ኪንያርዋንዳ', 'ar': 'الكينيارواندا', 'as': 'কিনয়াৰোৱাণ্ডা', 'asa': 'Kinyarandwa', 'ast': 'kinyarwanda', 'az': 'kinyarvanda', 'az-Cyrl': 'кинјарванда', 'az-Latn': 'kinyarvanda', 'bal-Latn': 'Kenyarwandá', 'bas': 'Hɔp u ruāndà', 'be': 'руанда', 'be-tarask': 'руанда', 'bem': 'Ichi Rwanda', 'bez': 'Hinyarwanda', 'bg': 'киняруанда', 'bgn': 'کینیارواندی', 'bm': 'ruwandakan', 'bn': 'কিনয়ারোয়ান্ডা', 'br': 'kinyarwanda', 'brx': 'किन्यार्वान्दा', 'bs': 'kinjaruanda', 'bs-Cyrl': 'кинјаруанда', 'bs-Latn': 'kinjaruanda', 'ca': 'ruandès', 'ca-ES-valencia': 'ruandès', 'ccp': '𑄇𑄨𑄚𑄴𑄠𑄢𑄮𑄠𑄚𑄴𑄓', 'ce': 'киньяруанда', 'cgg': 'Orunyarwanda', 'chr': 'ᎩᏂᏯᏩᏂᏓ', 'ckb': 'کینیارواندا', 'cs': 'kiňarwandština', 'cy': 'Ciniarŵandeg', 'da': 'kinyarwanda', 'dav': 'Kinyarwanda', 'de': 'Kinyarwanda', 'dje': 'Rwanda senni', 'dsb': 'kinjarwanda', 'dyo': 'ruanda', 'ebu': 'Kĩnyarwanda', 'ee': 'ruwandagbe', 'el': 'Κινιαρουάντα', 'el-polyton': 'Κινιαρουάντα', 'en': 'Kinyarwanda', 'eo': 'ruanda', 'es': 'kinyarwanda', 'et': 'ruanda', 'eu': 'kinyaruanda', 'ewo': 'ǹkɔ́bɔ ruwandá', 'fa': 'کینیارواندایی', 'ff': 'Ruwaanndeere', 'ff-Adlm': '𞤑𞤭𞤻𞤭𞤪𞤵𞤱𞤢𞤲𞤣𞤫𞥅𞤪𞤫', 'ff-Latn': 'Ruwaanndeere', 'fi': 'ruanda', 'fil': 'Kinyarwanda', 'fo': 'kinyarwanda', 'fr': 'kinyarwanda', 'fy': 'Kinyarwanda', 'ga': 'Ciniaruaindis', 'gd': 'Kinyarwanda', 'gl': 'kiñaruanda', 'gsw': 'Ruandisch', 'gu': 'કિન્યારવાન્ડા', 'guz': 'Kinyarwanda', 'ha': 'Kinyarwanda', 'he': 'קנירואנדית', 'hi': 'किन्यारवांडा', 'hi-Latn': 'Kinyarwanda', 'hr': 'kinyarwanda', 'hsb': 'kinjarwanda', 'ht': 'kinyarwanda', 'hu': 'kinyarvanda', 'hy': 'կինյառուանդա', 'ia': 'kinyarwanda', 'id': 'Kinyarwanda', 'ig': 'Kinyarwanda', 'is': 'kínjarvanda', 'it': 'kinyarwanda', 'ja': 'キニアルワンダ語', 'jmc': 'Kyinyarwanda', 'jv': 'Kinyarwanda', 'ka': 'კინიარუანდა', 'kab': 'Taruwandit', 'kam': 'Kinyarwanda', 'kde': 'Chinyalwanda', 'kea': 'kiniaruanda', 'kgp': 'kinĩjarvỹna', 'khq': 'Rwanda senni', 'ki': 'Kinyarwanda', 'kk': 'киньяруанда тілі', 'kk-Cyrl': 'киньяруанда тілі', 'kln': 'kutitab Kinyarwanda', 'km': 'គិនយ៉ាវ៉ាន់ដា', 'kn': 'ಕಿನ್ಯಾರ್\u200cವಾಂಡಾ', 'ko': '르완다어', 'kok': 'किन्यार्वान्डा', 'kok-Deva': 'किन्यार्वान्डा', 'ks': 'کِنیاوِندا', 'ks-Arab': 'کِنیاوِندا', 'ksb': 'Kinyalwanda', 'ksf': 'rirwanda', 'ksh': 'Kinja-Ruandesch', 'ku': 'kînyariwandayî', 'ky': 'руандача', 'lag': 'Kɨnyarwáanda', 'lb': 'Ruandesch', 'lg': 'Lunarwanda', 'lld': 'kinyarwanda', 'ln': 'kinyarwanda', 'lo': 'ຄິນຢາວານດາ', 'lrc': 'کینیاروآندا', 'lt': 'kinjaruandų', 'lu': 'kinyarwanda', 'luo': 'Kinyarwanda', 'luy': 'Kinyarwanda', 'lv': 'kiņaruanda', 'mai': 'किनयरवांडा', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lruwanda', 'mer': 'Kĩrwanda', 'mfe': 'rwanda', 'mg': 'Roande', 'mgh': 'Inyaranda', 'mi': 'Kiniawāna', 'mk': 'руандски', 'ml': 'കിന്യാർവാണ്ട', 'mn': 'киньяруанда', 'mr': 'किन्यार्वान्डा', 'ms': 'Kinyarwanda', 'mt': 'Kinjarwanda', 'mua': 'Zah sǝr Rwanda', 'my': 'ကင်ရာဝန်ဒါ', 'mzn': 'کنیاروآندایی', 'naq': 'Rwandaǁî gowab', 'nb': 'kinyarwanda', 'nd': 'isi-Ruwanda', 'ne': 'किन्यारवान्डा', 'nl': 'Kinyarwanda', 'nmg': 'Kiɛl rwandā', 'nn': 'kinyarwanda', 'no': 'kinyarwanda', 'nqo': 'ߞߌ߲ߦߊߙߎߥߊ߲ߘߊߞߊ߲', 'nus': 'Thok ruaandani', 'nyn': 'Orunyarwanda', 'or': 'କିନ୍ୟାରୱାଣ୍ଡା', 'pa': 'ਕਿਨਿਆਰਵਾਂਡਾ', 'pa-Guru': 'ਕਿਨਿਆਰਵਾਂਡਾ', 'pcm': 'Kinyarwánda Lángwej', 'pl': 'kinya-ruanda', 'ps': 'کینیارونډا', 'pt': 'quiniaruanda', 'qu': 'Kinyarwanda Simi', 'rm': 'kinyarwanda', 'rn': 'Ikinyarwanda', 'ro': 'kinyarwanda', 'rof': 'Kinyarwanda', 'ru': 'киньяруанда', 'rw': 'Ikinyarwanda', 'rwk': 'Kyinyarwanda', 'saq': 'Kinyarwanda', 'sbp': 'Ishinyalwanda', 'sc': 'kinyarwanda', 'sd': 'ڪنيار وانڊا', 'sd-Arab': 'ڪنيار وانڊا', 'seh': 'kinyarwanda', 'ses': 'Rwanda senni', 'sg': 'Ruandäa', 'shi': 'ⵜⴰⵔⵓⵡⴰⵏⴷⵉⵜ', 'shi-Latn': 'Taruwandit', 'shi-Tfng': 'ⵜⴰⵔⵓⵡⴰⵏⴷⵉⵜ', 'si': 'කින්යර්වන්ඩා', 'sk': 'rwandčina', 'sl': 'ruandščina', 'smn': 'ruandakielâ', 'sn': 'chiRwanda', 'so': 'Ruwaandha', 'sq': 'kiniaruandisht', 'sr': 'кињаруанда', 'sr-Cyrl': 'кињаруанда', 'sr-Latn': 'kinjaruanda', 'sv': 'kinjarwanda', 'sw': 'Kinyarwanda', 'ta': 'கின்யாருவான்டா', 'te': 'కిన్యర్వాండా', 'teo': 'Kinyarwanda', 'tg': 'киняруанда', 'th': 'รวันดา', 'ti': 'ኪንያርዋንዳ', 'tk': 'kinýaruanda dili', 'to': 'lea fakakiniāuanita', 'tr': 'Kinyarwanda', 'tt': 'руанда', 'twq': 'Rwanda senni', 'tzm': 'Tarwandit', 'ug': 'كېنىيەرىۋانداچە', 'uk': 'кіньяруанда', 'ur': 'کینیاروانڈا', 'uz': 'kinyaruanda', 'uz-Cyrl': 'киняруанда', 'uz-Latn': 'kinyaruanda', 'vai': 'ꕟꖙꕡ', 'vai-Latn': 'Rawunda', 'vai-Vaii': 'ꕟꖙꕡ', 'vec': 'kiniarwanda', 'vi': 'Tiếng Kinyarwanda', 'vun': 'Kyinyarwanda', 'wae': 'Ruandiš', 'wo': 'Kinyarwànda', 'xnr': 'किन्यारवांडा', 'xog': 'Olunarwanda', 'yav': 'nuluándɛ́ɛ', 'yo': 'Èdè Ruwanda', 'yrl': 'kiniaruãda', 'yue': '盧安達文', 'yue-Hans': '卢安达文', 'yue-Hant': '盧安達文', 'zgh': 'ⵜⴰⵔⵓⵡⴰⵏⴷⵉⵜ', 'zh': '卢旺达语', 'zh-Hans': '卢旺达语', 'zh-Hant': '盧安達文', 'zu': 'isi-Kinyarwanda'},
    'rwa': {'en': 'Rawo'},
    'rwk': {'af': 'Rwa', 'am': 'ርዋ', 'ar': 'الروا', 'as': 'ৰৱা', 'ast': 'rwa', 'az': 'rua', 'az-Cyrl': 'руа', 'az-Latn': 'rua', 'bal-Latn': 'Rwa', 'be': 'руа', 'be-tarask': 'руа', 'bg': 'рва', 'bgn': 'روایی', 'bn': 'রাওয়া', 'br': 'rwa', 'brx': 'र्वा', 'bs': 'rua', 'bs-Cyrl': 'рва', 'bs-Latn': 'rua', 'ca': 'rwo', 'ca-ES-valencia': 'rwo', 'ccp': '𑄢𑄤', 'ce': 'руанда', 'chr': 'Ꮖ', 'ckb': 'ڕوا', 'cs': 'rwa', 'cy': 'Rwa', 'da': 'rwa', 'de': 'Rwa', 'dsb': 'rwa', 'ee': 'rwagbe', 'el': 'Ρουά', 'el-polyton': 'Ρουά', 'en': 'Rwa', 'es': 'rwa', 'et': 'rvaa', 'eu': 'rwera', 'fa': 'روایی', 'ff-Adlm': '𞤈𞤵𞤱𞤢𞥄𞤪𞤫', 'fi': 'rwa', 'fil': 'Rwa', 'fo': 'rwa', 'fr': 'rwa', 'fy': 'Rwa', 'ga': 'Rwa', 'gd': 'Rwa', 'gl': 'rwa', 'gu': 'રવા', 'ha': 'Rwa', 'he': 'ראווה', 'hi': 'रवा', 'hi-Latn': 'Rwa', 'hr': 'rwa', 'hsb': 'rwa', 'ht': 'rwa', 'hu': 'rwo', 'hy': 'ռվա', 'ia': 'rwa', 'id': 'Rwa', 'ig': 'Rwa', 'is': 'rúa', 'it': 'rwa', 'ja': 'ルワ語', 'jv': 'Rwa', 'ka': 'რუა', 'kea': 'rwa', 'kgp': 'hywa', 'kk': 'руа тілі', 'kk-Cyrl': 'руа тілі', 'km': 'រ៉្វា', 'kn': 'ರುವ', 'ko': '르와어', 'kok': 'रवा', 'kok-Deva': 'रवा', 'ku': 'rwa', 'ky': 'руача', 'lb': 'Rwa', 'lo': 'ອາຣວາ', 'lrc': 'رئڤا', 'lt': 'rua', 'lv': 'ruanda', 'mai': 'रूवा', 'mi': 'Rawa', 'mk': 'руа', 'ml': 'റുവാ', 'mn': 'рва', 'mr': 'रव्हा', 'ms': 'Rwa', 'mt': 'Rwa', 'my': 'ရူဝမ်', 'mzn': 'روآیی', 'nb': 'rwa', 'ne': 'र्\u200cवा', 'nl': 'Rwa', 'nn': 'rwa', 'no': 'rwa', 'nqo': 'ߙߥߊߞߊ߲', 'or': 'ଆରଡବ୍ୟୁଏ', 'pa': 'ਰਵਾ', 'pa-Guru': 'ਰਵਾ', 'pcm': 'Rwá', 'pl': 'rwa', 'ps': 'روا', 'pt': 'rwa', 'qu': 'Rwa Simi', 'rm': 'rwa', 'ro': 'rwa', 'ru': 'руанда', 'rwk': 'Kiruwa', 'sc': 'rwa', 'sd': 'روا', 'sd-Arab': 'روا', 'si': 'ර්වා', 'sk': 'rwa', 'sl': 'rwa', 'smn': 'rwa', 'so': 'Raawa', 'sq': 'ruaisht', 'sr': 'руа', 'sr-Cyrl': 'руа', 'sr-Latn': 'rua', 'sv': 'rwa', 'sw': 'Lugha ya Rwa', 'ta': 'ருவா', 'te': 'ర్వా', 'th': 'รวา', 'ti': 'ርዋ', 'tk': 'rwa dili', 'to': 'lea fakaluā', 'tr': 'Rwa', 'ug': 'رىۋاچە', 'uk': 'рва', 'ur': 'روا', 'uz': 'ruanda', 'uz-Cyrl': 'руанда тили', 'uz-Latn': 'ruanda', 'vec': 'rwa', 'vi': 'Tiếng Rwa', 'xnr': 'रवा', 'yo': 'Riwa', 'yrl': 'ruwa', 'yue': '羅瓦文', 'yue-Hans': '罗瓦文', 'yue-Hant': '羅瓦文', 'zh': '罗瓦语', 'zh-Hans': '罗瓦语', 'zh-Hant': '羅瓦文', 'zu': 'isi-Rwa'},
    'rwl': {'en': 'Ruwila'},
    'rwm': {'en': 'Amba (Uganda)'},
    'rwo': {'en': 'Rawa'},
    'rwr': {'en': 'Marwari (India)'},
    'rxd': {'en': 'Ngardi'},
    'rxw': {'en': 'Karuwali'},
    'ryn': {'en': 'Northern Amami-Oshima'},
    'rys': {'en': 'Yaeyama'},
    'ryu': {'en': 'Central Okinawan'},
    'rzh': {'en': 'Rāziḥī'},
    'sa': {'af': 'Sanskrit', 'ak': 'Sanskrit kasa', 'am': 'ሳንስክሪት', 'ar': 'السنسكريتية', 'as': 'সংস্কৃত', 'ast': 'sánscritu', 'az': 'sanskrit', 'az-Cyrl': 'санскрит', 'az-Latn': 'sanskrit', 'bal-Latn': 'Sanskeret', 'be': 'санскрыт', 'be-tarask': 'санскрыт', 'bg': 'санскрит', 'bgn': 'سانسکریٹ', 'bn': 'সংস্কৃত', 'br': 'sanskriteg', 'brx': 'संस्कृत', 'bs': 'sanskrit', 'bs-Cyrl': 'санскрит', 'bs-Latn': 'sanskrit', 'ca': 'sànscrit', 'ca-ES-valencia': 'sànscrit', 'ccp': '𑄥𑄧𑄁𑄥𑄴𑄇𑄳𑄢𑄨𑄖𑄴', 'ce': 'санскрит', 'chr': 'ᏍᏂᏍᎩᏗ', 'ckb': 'سانسکريت', 'cs': 'sanskrt', 'cy': 'Sansgrit', 'da': 'sanskrit', 'de': 'Sanskrit', 'dsb': 'sanskrit', 'dz': 'སཾསྐྲྀཏ་ཁ', 'ee': 'sanskrigbe', 'el': 'Σανσκριτικά', 'el-polyton': 'Σανσκριτικά', 'en': 'Sanskrit', 'eo': 'sanskrito', 'es': 'sánscrito', 'et': 'sanskriti', 'eu': 'sanskritoa', 'fa': 'سانسکریت', 'ff-Adlm': '𞤅𞤢𞤲𞤧𞤳𞤪𞤭𞤼𞤪𞤫', 'fi': 'sanskrit', 'fil': 'Sanskrit', 'fo': 'sanskrit', 'fr': 'sanskrit', 'fur': 'sanscrit', 'fy': 'Sanskriet', 'ga': 'Sanscrait', 'gd': 'Sanskrit', 'gl': 'sánscrito', 'gsw': 'Sanschkrit', 'gu': 'સંસ્કૃત', 'ha': 'Sanskrit', 'he': 'סנסקריט', 'hi': 'संस्कृत', 'hi-Latn': 'Sanskrit', 'hr': 'sanskrtski', 'hsb': 'sanskrit', 'ht': 'sanskrit', 'hu': 'szanszkrit', 'hy': 'սանսկրիտ', 'ia': 'sanscrito', 'id': 'Sanskerta', 'ig': 'Asụsụ Sanskrit', 'is': 'sanskrít', 'it': 'sanscrito', 'ja': 'サンスクリット語', 'jv': 'Sanskerta', 'ka': 'სანსკრიტი', 'kea': 'sánskritu', 'kgp': 'sỹnhkrito', 'kk': 'санскрит тілі', 'kk-Cyrl': 'санскрит тілі', 'km': 'សំស្ក្រឹត', 'kn': 'ಸಂಸ್ಕೃತ', 'ko': '산스크리트어', 'kok': 'संस्कृत', 'kok-Deva': 'संस्कृत', 'ks': 'سَنسکرٕت', 'ks-Arab': 'سَنسکرٕت', 'ksh': 'Sanskrit', 'ku': 'sanskrîtî', 'kxv': 'sanskrit', 'kxv-Deva': 'संस्कृत', 'kxv-Latn': 'sanskrit', 'kxv-Orya': 'ସଂସ୍କୃତ', 'kxv-Telu': 'సంస్కృతం', 'ky': 'санскритче', 'lb': 'Sanskrit', 'lkt': 'Sanskrit Iyápi', 'lld': 'sanscrit', 'lo': 'ສັນສະກຣິດ', 'lrc': 'سانسکئریت', 'lt': 'sanskritas', 'lv': 'sanskrits', 'mai': 'संस्कृत', 'mi': 'Hanahiti', 'mk': 'санскрит', 'ml': 'സംസ്\u200cകൃതം', 'mn': 'санскрит', 'mr': 'संस्कृत', 'ms': 'Sanskrit', 'mt': 'Sanskrit', 'my': 'သင်္သကရိုက်', 'mzn': 'سانسکریت', 'nb': 'sanskrit', 'ne': 'संस्कृत', 'nl': 'Sanskriet', 'nn': 'sanskrit', 'no': 'sanskrit', 'nqo': 'ߛߊߛߑߞߙߌߞߊ߲', 'or': 'ସଂସ୍କୃତ', 'pa': 'ਸੰਸਕ੍ਰਿਤ', 'pa-Guru': 'ਸੰਸਕ੍ਰਿਤ', 'pcm': 'Sánskrit', 'pl': 'sanskryt', 'ps': 'سنسکریټ', 'pt': 'sânscrito', 'qu': 'Sanscrito Simi', 'rm': 'sanscrit', 'ro': 'sanscrită', 'ru': 'санскрит', 'rw': 'Igisansikiri', 'sa': 'संस्कृत भाषा', 'sc': 'sànscritu', 'sd': 'سنسڪرت', 'sd-Arab': 'سنسڪرت', 'si': 'සංස්කෘත', 'sk': 'sanskrit', 'sl': 'sanskrt', 'smn': 'sanskrit', 'so': 'Sanskrit', 'sq': 'sanskritisht', 'sr': 'санскрит', 'sr-Cyrl': 'санскрит', 'sr-Latn': 'sanskrit', 'sv': 'sanskrit', 'sw': 'Kisanskriti', 'ta': 'சமஸ்கிருதம்', 'te': 'సంస్కృతం', 'tg': 'санскрит', 'th': 'สันสกฤต', 'ti': 'ሳንስክሪት', 'tk': 'sanskrit dili', 'to': 'lea fakasanisukuliti', 'tr': 'Sanskrit', 'tt': 'санскрит', 'ug': 'سانسكرىتچە', 'uk': 'санскрит', 'ur': 'سنسکرت', 'uz': 'sanskrit', 'uz-Cyrl': 'санскрит', 'uz-Latn': 'sanskrit', 'vec': 'sànscrito', 'vi': 'Tiếng Phạn', 'wae': 'Sanskrit', 'wo': 'Sanskrit', 'xnr': 'संस्कृत', 'yi': 'סאַנסקריט', 'yo': 'Èdè awon ara Indo', 'yrl': 'sãsikiritu', 'yue': '梵文', 'yue-Hans': '梵文', 'yue-Hant': '梵文', 'zh': '梵语', 'zh-Hans': '梵语', 'zh-Hant': '梵文', 'zu': 'isi-Sanskrit'},
    'saa': {'en': 'Saba'},
    'sab': {'en': 'Buglere'},
    'sac': {'en': 'Meskwaki'},
    'sad': {'af': 'Sandawees', 'am': 'ሳንዳዌ', 'ar': 'السانداوي', 'as': 'ছান্দাৱে', 'ast': 'sandavés', 'az': 'sandave', 'az-Cyrl': 'сандаве', 'az-Latn': 'sandave', 'be': 'сандаўэ', 'be-tarask': 'сандаўэ', 'bg': 'сандаве', 'bgn': 'سانڈاوه\u200cای', 'bn': 'সান্দাওয়ে', 'br': 'sandawe', 'brx': 'साण्डावे', 'bs': 'sandave', 'bs-Cyrl': 'сандаве', 'bs-Latn': 'sandave', 'ca': 'sandawe', 'ca-ES-valencia': 'sandawe', 'ccp': '𑄥𑄳𑄠𑄚𑄴𑄓𑄃𑄮𑄠𑄬', 'ce': 'сандаве', 'chr': 'ᏌᏅᏓᏫ', 'ckb': 'سانداوێ', 'cs': 'sandawština', 'cy': 'Sandäweg', 'da': 'sandawe', 'de': 'Sandawe', 'dsb': 'sandawešćina', 'el': 'Σαντάγουε', 'el-polyton': 'Σαντάγουε', 'en': 'Sandawe', 'es': 'sandawe', 'et': 'sandave', 'eu': 'sandaweera', 'fa': 'سانداوه\u200cای', 'ff-Adlm': '𞤅𞤢𞤲𞤣𞤢𞤱𞤫𞥅𞤪𞤫', 'fi': 'sandawe', 'fil': 'Sandawe', 'fo': 'sandawe', 'fr': 'sandawe', 'fy': 'Sandawe', 'ga': 'Sandabhais', 'gd': 'Sandawe', 'gl': 'sandawe', 'gsw': 'Sandawe-Schpraach', 'gu': 'સોંડવે', 'ha': 'Sandawe', 'he': 'סנדאווה', 'hi': 'सन्डावे', 'hi-Latn': 'Sandawe', 'hr': 'sandawe', 'hsb': 'sandawešćina', 'ht': 'sandawe', 'hu': 'szandave', 'hy': 'սանդավե', 'ia': 'sandawe', 'id': 'Sandawe', 'ig': 'Sandawe', 'is': 'sandave', 'it': 'sandawe', 'ja': 'サンダウェ語', 'jv': 'Sandawe', 'ka': 'სანდავე', 'kgp': 'sỹnave', 'kk': 'сандаве тілі', 'kk-Cyrl': 'сандаве тілі', 'km': 'សានដាវី', 'kn': 'ಸಂಡಾವೇ', 'ko': '산다웨어', 'kok': 'संडावे', 'kok-Deva': 'संडावे', 'ks': 'سَندَویے', 'ks-Arab': 'سَندَویے', 'ku': 'sandawe', 'ky': 'сандавече', 'lb': 'Sandawe-Sprooch', 'lo': 'ຊັນດາວ', 'lt': 'sandavių', 'lv': 'sandavu', 'mai': 'संदाउ', 'mi': 'Hātawe', 'mk': 'сандаве', 'ml': 'സാൻഡവേ', 'mn': 'сандавэ', 'mr': 'सँडवे', 'ms': 'Sandawe', 'mt': 'Sandawe', 'my': 'ဆန်ဒါဝီ', 'nb': 'sandawe', 'ne': 'सान्डेअ', 'nl': 'Sandawe', 'nn': 'sandawe', 'no': 'sandawe', 'or': 'ସଣ୍ଡାୱେ', 'pa': 'ਸਾਂਡੋ', 'pa-Guru': 'ਸਾਂਡੋ', 'pcm': 'Sandáwẹ́ Lángwej', 'pl': 'sandawe', 'ps': 'سنډاوی', 'pt': 'sandawe', 'qu': 'Sandawe Simi', 'rm': 'sandawe', 'ro': 'sandawe', 'ru': 'сандаве', 'sc': 'sandawe', 'sd': 'سنداوي', 'sd-Arab': 'سنداوي', 'si': 'සන්ඩවෙ', 'sk': 'sandaweština', 'sl': 'sandavščina', 'smn': 'sandawe', 'so': 'Sandawe', 'sq': 'sandauisht', 'sr': 'сандаве', 'sr-Cyrl': 'сандаве', 'sr-Latn': 'sandave', 'sv': 'sandawe', 'sw': 'Kisandawe', 'ta': 'சான்டாவே', 'te': 'సండావి', 'th': 'ซันดาเว', 'ti': 'ሳንዳወ', 'tk': 'sandawe dili', 'to': 'lea fakasanitaue', 'tr': 'Sandave', 'ug': 'سانداۋېچە', 'uk': 'сандаве', 'ur': 'سنڈاوے', 'uz': 'sandave', 'uz-Latn': 'sandave', 'vec': 'sandawe', 'vi': 'Tiếng Sandawe', 'xnr': 'सन्डावे', 'yo': 'Èdè Sandawe', 'yrl': 'sãdawe', 'yue': '桑達韋文', 'yue-Hans': '桑达韦文', 'yue-Hant': '桑達韋文', 'zh': '桑达韦语', 'zh-Hans': '桑达韦语', 'zh-Hant': '桑達韋文', 'zu': 'isi-Sandawe'},
    'sae': {'en': 'Sabanê'},
    'saf': {'en': 'Safaliba'},
    'sah': {'af': 'Jakoeties', 'ak': 'Yakut Kasa', 'am': 'ያኩት', 'ar': 'الساخيّة', 'as': 'ছাখা', 'ast': 'sakha', 'az': 'saxa', 'az-Cyrl': 'саха', 'az-Latn': 'saxa', 'bal-Latn': 'Yakut', 'be': 'якуцкая', 'be-tarask': 'якуцкая', 'bg': 'саха', 'bgn': 'یاقوتی', 'bn': 'শাখা', 'br': 'yakouteg', 'brx': 'साखा', 'bs': 'jakutski', 'bs-Cyrl': 'јакутски', 'bs-Latn': 'jakutski', 'ca': 'iacut', 'ca-ES-valencia': 'iacut', 'ccp': '𑄥𑄈', 'ce': 'якутийн', 'chr': 'ᏌᎧᎾ', 'ckb': 'ساخا', 'cs': 'jakutština', 'cy': 'Sakha', 'da': 'jakutisk', 'de': 'Jakutisch', 'dsb': 'jakutšćina', 'ee': 'sakagbe', 'el': 'Σαχά', 'el-polyton': 'Σαχά', 'en': 'Yakut', 'es': 'sakha', 'et': 'jakuudi', 'eu': 'sakhera', 'fa': 'یاقوتی', 'ff-Adlm': '𞤅𞤢𞤿𞤢𞥄𞤪𞤫', 'fi': 'jakuutti', 'fil': 'Yakut', 'fo': 'sakha', 'fr': 'iakoute', 'fy': 'Jakoets', 'ga': 'Sachais', 'gd': 'Sakha', 'gl': 'iacuto', 'gsw': 'Jakutisch', 'gu': 'સખા', 'ha': 'Yakut', 'he': 'סאחה', 'hi': 'याकूत', 'hi-Latn': 'Yakut', 'hr': 'jakutski', 'hsb': 'jakutšćina', 'ht': 'iakoute', 'hu': 'szaha', 'hy': 'յակուտերեն', 'ia': 'yakuto', 'id': 'Sakha', 'ig': 'Yakut', 'is': 'jakút', 'it': 'sacha', 'ja': 'サハ語', 'jv': 'Yakut', 'ka': 'იაკუტური', 'kgp': 'saka', 'kk': 'саха тілі', 'kk-Cyrl': 'саха тілі', 'km': 'យ៉ាឃុត', 'kn': 'ಸಖಾ', 'ko': '야쿠트어', 'kok': 'साखा', 'kok-Deva': 'साखा', 'ks': 'یاکُت', 'ks-Arab': 'یاکُت', 'ksh': 'Jackutesch', 'ku': 'yakutî', 'ky': 'сахача', 'lb': 'Jakutesch', 'lo': 'ຊາກາ', 'lt': 'jakutų', 'lv': 'jakutu', 'mai': 'सखा', 'mi': 'Hakūta', 'mk': 'јакутски', 'ml': 'സാഖ', 'mn': 'саха', 'mr': 'साखा', 'ms': 'Sakha', 'mt': 'Sakha', 'my': 'ဆခါ', 'nb': 'sakha', 'ne': 'साखा', 'nl': 'Jakoets', 'nn': 'sakha', 'no': 'sakha', 'nqo': 'ߌߦߊߞߎߕߌߞߊ߲', 'or': 'ୟାକୂଟ୍\u200c', 'pa': 'ਸਾਖਾ', 'pa-Guru': 'ਸਾਖਾ', 'pcm': 'Sakhá', 'pl': 'jakucki', 'ps': 'سخا', 'pt': 'sakha', 'qu': 'Sakha Simi', 'rm': 'jakut', 'ro': 'sakha', 'ru': 'саха', 'sah': 'саха тыла', 'sc': 'yakut', 'sd': 'ساخا', 'sd-Arab': 'ساخا', 'si': 'සඛා', 'sk': 'jakutčina', 'sl': 'jakutščina', 'smn': 'jakutkielâ', 'so': 'Yakut', 'sq': 'sakaisht', 'sr': 'саха', 'sr-Cyrl': 'саха', 'sr-Latn': 'saha', 'sv': 'jakutiska', 'sw': 'Kisakha', 'ta': 'சக்கா', 'te': 'సాఖా', 'tg': 'саха', 'th': 'ซาคา', 'ti': 'ሳኻ', 'tk': 'ýakut dili', 'to': 'lea fakaiakuti', 'tr': 'Yakutça', 'tt': 'саха', 'ug': 'ساخاچە', 'uk': 'якутська', 'ur': 'ساکھا', 'uz': 'saxa', 'uz-Cyrl': 'саха', 'uz-Latn': 'saxa', 'vec': 'yakuto', 'vi': 'Tiếng Sakha', 'wae': 'Jakutiš', 'wo': 'Saxa', 'xnr': 'याकूत', 'yo': 'Sàkíhà', 'yrl': 'saka', 'yue': '雅庫特文', 'yue-Hans': '雅库特文', 'yue-Hant': '雅庫特文', 'zh': '萨哈语', 'zh-Hans': '萨哈语', 'zh-Hant': '雅庫特文', 'zu': 'i-Sakha'},
    'sai': {'en': 'South American Indian languages'},
    'saj': {'en': 'Sahu'},
    'sak': {'en': 'Sake'},
    'sal': {'en': 'Salishan languages'},
    'sal-pro': {'en': 'Proto-Salish'},
    'sam': {'ar': 'الآرامية السامرية', 'ast': 'araméu samaritanu', 'az': 'samaritan', 'az-Latn': 'samaritan', 'bg': 'самаритански арамейски', 'bn': 'সামারিটান আরামিক', 'br': 'arameeg ar Samaritaned', 'brx': 'समारीती आरामाईक़', 'bs': 'samaritanski aramejski', 'bs-Cyrl': 'самаритански арамејски', 'bs-Latn': 'samaritanski aramejski', 'ca': 'arameu samarità', 'ca-ES-valencia': 'arameu samarità', 'ccp': '𑄥𑄟𑄢𑄨𑄑𑄚𑄴 𑄃𑄢𑄟𑄨𑄇𑄴', 'cs': 'samarština', 'cy': 'Aramaeg Samaria', 'da': 'samaritansk aramæisk', 'de': 'Samaritanisch', 'el': 'Σαμαρίτικα Αραμαϊκά', 'el-polyton': 'Σαμαρίτικα Αραμαϊκά', 'en': 'Samaritan Aramaic', 'es': 'arameo samaritano', 'et': 'Samaaria aramea', 'fa': 'آرامی سامری', 'fi': 'samarianaramea', 'fr': 'araméen samaritain', 'fy': 'Samaritaansk-Arameesk', 'ga': 'Aramais Shamárach', 'gd': 'Aramais Shamaritanach', 'gsw': 'Samaritanisch', 'gu': 'સામરિટાન અરેમિક', 'he': 'ארמית שומרונית', 'hi': 'सामैरिटन अरैमिक', 'hi-Latn': 'Samaritan Aramaic', 'hr': 'samarijanski aramejski', 'ht': 'araméen samaritain', 'hu': 'szamaritánus arámi', 'id': 'Aram Samaria', 'is': 'samversk arameíska', 'it': 'aramaico samaritano', 'ja': 'サマリア・アラム語', 'ka': 'სამარიულ-არამეული', 'kgp': 'aramỹjko samaritỹnũ', 'kn': 'ಸಮರಿಟನ್ ಅರಾಮಿಕ್', 'ko': '사마리아 아랍어', 'ks': 'سَمارِتَن اَرامیک', 'ks-Arab': 'سَمارِتَن اَرامیک', 'lb': 'Samaritanesch', 'lo': 'ສາມາຣິແຕນ-', 'lt': 'samarėjų aramių', 'lv': 'Samārijas aramiešu', 'mk': 'самарјански арамејски', 'ml': 'സമരിയാക്കാരുടെ അരമായ', 'mr': 'सामरिटान अरॅमिक', 'mt': 'Samaritan Aramajk', 'nb': 'samaritansk arameisk', 'nl': 'Samaritaans-Aramees', 'nn': 'samaritansk arameisk', 'no': 'samaritansk arameisk', 'or': 'ସାମୌରିଟନ୍ ଆରମାଇକ୍', 'pl': 'samarytański aramejski', 'pt': 'aramaico samaritano', 'rm': 'arameic samaritan', 'ro': 'aramaică samariteană', 'ru': 'самаритянский арамейский', 'sk': 'samaritánska aramejčina', 'sl': 'samaritanska aramejščina', 'sr': 'самаријански арамејски', 'sr-Cyrl': 'самаријански арамејски', 'sr-Latn': 'samarijanski aramejski', 'sv': 'samaritanska', 'sw': 'Kiaramu cha Wasamaria', 'ta': 'சமாரிடன் அராமைக்', 'te': 'సమారిటన్ అరామైక్', 'th': 'อราเมอิกซามาเรีย', 'to': 'lea fakasamalitani-ʻalāmiti', 'tr': 'Samarit Aramcası', 'ug': 'سامارىتانچە', 'uk': 'самаритянська арамейська', 'vi': 'Tiếng Samaritan Aramaic', 'yrl': 'aramaiku samaritanu', 'yue': '薩瑪利亞阿拉姆文', 'yue-Hans': '萨玛利亚阿拉姆文', 'yue-Hant': '薩瑪利亞阿拉姆文', 'zh': '萨马利亚阿拉姆语', 'zh-Hans': '萨马利亚阿拉姆语', 'zh-Hant': '薩瑪利亞阿拉姆文'},
    'sao': {'en': 'Sause'},
    'sap': {'en': 'Sanapaná'},
    'saq': {'af': 'Samburu', 'am': 'ሳምቡሩ', 'ar': 'سامبورو', 'as': 'ছাম্বুৰু', 'ast': 'samburu', 'az': 'samburu', 'az-Cyrl': 'самбуру', 'az-Latn': 'samburu', 'bal-Latn': 'Samboró', 'be': 'самбуру', 'be-tarask': 'самбуру', 'bg': 'самбуру', 'bgn': 'سامبوروی', 'bn': 'সামবুরু', 'br': 'samburu', 'brx': 'साम्बुरू', 'bs': 'samburu', 'bs-Cyrl': 'самбуру', 'bs-Latn': 'samburu', 'ca': 'samburu', 'ca-ES-valencia': 'samburu', 'ccp': '𑄥𑄟𑄴𑄝𑄪𑄢𑄪', 'ce': 'самбуру', 'chr': 'ᏌᎹᏊᎷ', 'ckb': 'سامبووروو', 'cs': 'samburu', 'cy': 'Sambŵrw', 'da': 'samburu', 'de': 'Samburu', 'dsb': 'samburu', 'el': 'Σαμπούρου', 'el-polyton': 'Σαμπούρου', 'en': 'Samburu', 'es': 'samburu', 'et': 'samburu', 'eu': 'samburuera', 'fa': 'سامبورو', 'ff-Adlm': '𞤅𞤢𞤥𞤦𞤵𞤪𞤵𞥅𞤪𞤫', 'fi': 'samburu', 'fil': 'Samburu', 'fo': 'samburu', 'fr': 'samburu', 'fy': 'Samburu', 'ga': 'Samburu', 'gd': 'Samburu', 'gl': 'samburu', 'gu': 'સમ્બુરુ', 'ha': 'Samburu', 'he': 'סמבורו', 'hi': 'सैम्बुरु', 'hi-Latn': 'Samburu', 'hr': 'samburu', 'hsb': 'samburu', 'ht': 'samburu', 'hu': 'szamburu', 'hy': 'սամբուրու', 'ia': 'samburu', 'id': 'Samburu', 'ig': 'Samburu', 'is': 'sambúrú', 'it': 'samburu', 'ja': 'サンブル語', 'jv': 'Samburu', 'ka': 'სამბურუ', 'kgp': 'sỹmuru', 'kk': 'самбуру тілі', 'kk-Cyrl': 'самбуру тілі', 'km': 'សាមបូរូ', 'kn': 'ಸಂಬುರು', 'ko': '삼부루어', 'kok': 'साम्बुरू', 'kok-Deva': 'साम्बुरू', 'ku': 'samburuyî', 'ky': 'самбуруча', 'lb': 'Samburu', 'lo': 'ຊຳບູຣູ', 'lrc': 'سامبوٙروٙ', 'lt': 'sambūrų', 'lv': 'samburu', 'mai': 'समबुरू', 'mi': 'Hāmapuru', 'mk': 'самбуру', 'ml': 'സംബുരു', 'mn': 'самбүрү', 'mr': 'सांबुरू', 'ms': 'Samburu', 'mt': 'Samburu', 'my': 'ဆမ်ဘူရူ', 'mzn': 'سامبورو', 'nb': 'samburu', 'ne': 'साम्बुरू', 'nl': 'Samburu', 'nn': 'samburu', 'no': 'samburu', 'nqo': 'ߛߊ߲ߓߙߎߞߊ߲', 'or': 'ସମବୁରୁ', 'pa': 'ਸਮਬੁਰੂ', 'pa-Guru': 'ਸਮਬੁਰੂ', 'pcm': 'Sambúru', 'pl': 'samburu', 'ps': 'سمبورو', 'pt': 'samburu', 'qu': 'Samburu Simi', 'rm': 'samburu', 'ro': 'samburu', 'ru': 'самбуру', 'saq': 'Kisampur', 'sc': 'samburu', 'sd': 'سيمبورو', 'sd-Arab': 'سيمبورو', 'si': 'සම්බුරු', 'sk': 'samburu', 'sl': 'samburščina', 'smn': 'samburukielâ', 'so': 'Sambuuru', 'sq': 'samburisht', 'sr': 'самбуру', 'sr-Cyrl': 'самбуру', 'sr-Latn': 'samburu', 'sv': 'samburu', 'sw': 'Kisamburu', 'ta': 'சம்புரு', 'te': 'సంబురు', 'th': 'แซมบูรู', 'ti': 'ሳምቡሩ', 'tk': 'samburu dili', 'to': 'lea fakasamipulu', 'tr': 'Samburu', 'ug': 'سامبۇرۇچە', 'uk': 'самбуру', 'ur': 'سامبورو', 'uz': 'samburu', 'uz-Cyrl': 'самбуруча', 'uz-Latn': 'samburu', 'vec': 'sanburu', 'vi': 'Tiếng Samburu', 'xnr': 'सैम्बुरु', 'yo': 'Samburu', 'yrl': 'sãburu', 'yue': '薩布魯文', 'yue-Hans': '萨布鲁文', 'yue-Hant': '薩布魯文', 'zh': '桑布鲁语', 'zh-Hans': '桑布鲁语', 'zh-Hant': '薩布魯文', 'zu': 'isi-Samburu'},
    'sar': {'en': 'Saraveca'},
    'sas': {'ar': 'الساساك', 'ast': 'sasak', 'az': 'sasak', 'az-Latn': 'sasak', 'bg': 'сасак', 'bn': 'সাসাক', 'br': 'sasak', 'brx': 'सासक', 'bs': 'sasak', 'bs-Cyrl': 'сасак', 'bs-Latn': 'sasak', 'ca': 'sasak', 'ca-ES-valencia': 'sasak', 'ccp': '𑄥𑄥𑄇𑄴', 'cs': 'sasakština', 'cy': 'Sasaceg', 'da': 'sasak', 'de': 'Sasak', 'el': 'Σασάκ', 'el-polyton': 'Σασάκ', 'en': 'Sasak', 'es': 'sasak', 'et': 'sasaki', 'fa': 'ساساکی', 'fi': 'sasak', 'fr': 'sasak', 'fy': 'Sasak', 'gd': 'Sasak', 'gsw': 'Sasak', 'gu': 'સાસાક', 'he': 'סאסק', 'hi': 'सासाक', 'hi-Latn': 'Sasak', 'hr': 'sasak', 'ht': 'sasak', 'hu': 'sasak', 'id': 'Sasak', 'is': 'sasak', 'it': 'sasak', 'ja': 'ササク語', 'kgp': 'sasag', 'kn': 'ಸಸಾಕ್', 'ko': '사사크어', 'ks': 'سَسَک', 'ks-Arab': 'سَسَک', 'lb': 'Sasak', 'lo': 'ຊາຊັກ', 'lt': 'sasak', 'lv': 'sasaku', 'mk': 'сасачки', 'ml': 'സസാക്', 'mr': 'सासाक', 'mt': 'Sasak', 'nb': 'sasak', 'nl': 'Sasak', 'nn': 'sasak', 'no': 'sasak', 'or': 'ସାସାକ୍', 'pl': 'sasak', 'pt': 'sasak', 'rm': 'sasak', 'ro': 'sasak', 'ru': 'сасакский', 'sk': 'sasačtina', 'sl': 'sasaščina', 'sr': 'сасак', 'sr-Cyrl': 'сасак', 'sr-Latn': 'sasak', 'sv': 'sasak', 'ta': 'சாசாக்', 'te': 'ససక్', 'th': 'ซาซัก', 'to': 'lea fakasasaki', 'tr': 'Sasak', 'ug': 'ساساكچە', 'uk': 'сасакська', 'vi': 'Tiếng Sasak', 'yrl': 'sasak', 'yue': '撒撒克文', 'yue-Hans': '撒撒克文', 'yue-Hant': '撒撒克文', 'zh': '萨萨克语', 'zh-Hans': '萨萨克语', 'zh-Hant': '撒撒克文'},
    'sat': {'af': 'Santalies', 'ak': 'Santal kasa', 'am': 'ሳንታሊ', 'ar': 'السانتالية', 'as': 'চাওতালি', 'ast': 'santali', 'az': 'santal', 'az-Cyrl': 'сантал', 'az-Latn': 'santal', 'bal-Latn': 'Santali', 'be': 'санталі', 'be-tarask': 'санталі', 'bg': 'сантали', 'bgn': 'سانٹالی', 'bn': 'সাঁওতালি', 'br': 'santali', 'brx': 'सान्थाली', 'bs': 'santali', 'bs-Cyrl': 'сантали', 'bs-Latn': 'santali', 'ca': 'santali', 'ca-ES-valencia': 'santali', 'ccp': '𑄥𑄀𑄃𑄮𑄖𑄣𑄨', 'ce': 'сантали', 'chr': 'ᏌᏂᏔᎵ', 'ckb': 'سانتالی', 'cs': 'santálština', 'cy': 'Santali', 'da': 'santali', 'de': 'Santali', 'dsb': 'santalšćina', 'el': 'Σαντάλι', 'el-polyton': 'Σαντάλι', 'en': 'Santali', 'es': 'santali', 'et': 'santali', 'eu': 'santalera', 'fa': 'سانتالی', 'ff-Adlm': '𞤅𞤢𞤲𞤼𞤢𞤤𞤭𞥅𞤪𞤫', 'fi': 'santali', 'fil': 'Santali', 'fo': 'santali', 'fr': 'santali', 'fy': 'Santali', 'ga': 'Santáilis', 'gd': 'Santali', 'gl': 'santali', 'gsw': 'Santali', 'gu': 'સંતાલી', 'ha': 'Santali', 'he': 'סאנטאלי', 'hi': 'संथाली', 'hi-Latn': 'Santali', 'hr': 'santalski', 'hsb': 'santalšćina', 'ht': 'santali', 'hu': 'szantáli', 'hy': 'սանտալի', 'ia': 'santali', 'id': 'Santali', 'ig': 'Asụsụ Santali', 'is': 'santalí', 'it': 'santali', 'ja': 'サンターリー語', 'jv': 'Santali', 'ka': 'სანტალი', 'kgp': 'sỹtari', 'kk': 'сантали тілі', 'kk-Cyrl': 'сантали тілі', 'km': 'សាន់តាលី', 'kn': 'ಸಂತಾಲಿ', 'ko': '산탈리어', 'kok': 'संथाली', 'kok-Deva': 'संथाली', 'ks': 'سَنتالی', 'ks-Arab': 'سَنتالی', 'ku': 'santalî', 'kxv': 'santalī', 'kxv-Deva': 'संताली', 'kxv-Latn': 'santalī', 'kxv-Orya': 'ସାନ୍ତାଲି', 'kxv-Telu': 'సంతాలి', 'ky': 'санталиче', 'lb': 'Santali', 'lo': 'ຊານທາລິ', 'lt': 'santalių', 'lv': 'santalu', 'mai': 'संथाली', 'mi': 'Hatāri', 'mk': 'сантали', 'ml': 'സന്താലി', 'mn': 'сантали', 'mr': 'संताली', 'ms': 'Santali', 'mt': 'Santali', 'my': 'ဆန်တာလီ', 'nb': 'santali', 'ne': 'सान्ताली', 'nl': 'Santali', 'nn': 'santali', 'no': 'santali', 'nqo': 'ߛߊ߲ߕߊߟߌߞߊ߲', 'or': 'ସାନ୍ତାଳି', 'pa': 'ਸੰਥਾਲੀ', 'pa-Guru': 'ਸੰਥਾਲੀ', 'pcm': 'Sántáli', 'pl': 'santali', 'ps': 'سنتالي', 'pt': 'santali', 'qu': 'Santali Simi', 'rm': 'santali', 'ro': 'santali', 'ru': 'сантали', 'sat': 'ᱥᱟᱱᱛᱟᱲᱤ', 'sat-Deva': 'सानताड़ी', 'sat-Olck': 'ᱥᱟᱱᱛᱟᱲᱤ', 'sc': 'santali', 'sd': 'سنتالي', 'sd-Arab': 'سنتالي', 'si': 'සෑන්ටලි', 'sk': 'santalčina', 'sl': 'santalščina', 'smn': 'santalikielâ', 'so': 'Santali', 'sq': 'santalisht', 'sr': 'сантали', 'sr-Cyrl': 'сантали', 'sr-Latn': 'santali', 'sv': 'santali', 'sw': 'Kisantali', 'ta': 'சான்டாலி', 'te': 'సంతాలి', 'tg': 'санталӣ', 'th': 'สันตาลี', 'ti': 'ሳንታሊ', 'tk': 'santali dili', 'to': 'lea fakasanitali', 'tr': 'Santali', 'tt': 'сантали', 'ug': 'سانتالچە', 'uk': 'сантальська', 'ur': 'سنتالی', 'uz': 'santali', 'uz-Cyrl': 'сантали', 'uz-Latn': 'santali', 'vec': 'santali', 'vi': 'Tiếng Santali', 'wo': 'Santali', 'xnr': 'संथाली', 'yo': 'Èdè Santali', 'yrl': 'sãtari', 'yue': '散塔利文', 'yue-Hans': '散塔利文', 'yue-Hant': '散塔利文', 'zh': '桑塔利语', 'zh-Hans': '桑塔利语', 'zh-Hant': '桑塔利文', 'zu': 'isi-Santali'},
    'sau': {'en': 'Saleman'},
    'sav': {'en': 'Saafi-Saafi'},
    'saw': {'en': 'Sawi'},
    'sax': {'en': 'Sa'},
    'say': {'en': 'Saya'},
    'saz': {'ast': 'saurashtra', 'cs': 'saurášterština', 'de': 'Saurashtra', 'en': 'Saurashtra', 'et': 'sauraštra', 'fi': 'sauraštri', 'fr': 'saurashtra', 'gd': 'Saurashtra', 'hi-Latn': 'Saurashtra', 'ht': 'saurashtra', 'it': 'saurashtra', 'ja': 'サウラーシュトラ語', 'lb': 'Saurashtra', 'lt': 'sauraštrų', 'mk': 'саураштра', 'nb': 'saurashtra', 'nl': 'Saurashtra', 'nn': 'saurashtra', 'no': 'saurashtra', 'pl': 'saurasztryjski', 'sv': 'saurashtra', 'ta': 'சௌராஷ்டிரம்', 'th': 'เสาราษฏร์', 'to': 'lea fakasaulasitilā', 'tr': 'Saurashtra', 'yue': '索拉什特拉文', 'yue-Hans': '索拉什特拉文', 'yue-Hant': '索拉什特拉文', 'zh-Hant': '索拉什特拉文'},
    'sba': {'af': 'Ngambay', 'am': 'ንጋምባይ', 'ar': 'نامبي', 'as': 'নাংম্বে', 'ast': 'ngambay', 'az': 'nqambay', 'az-Cyrl': 'нгамбај', 'az-Latn': 'nqambay', 'be': 'нгамбай', 'be-tarask': 'нгамбай', 'bg': 'нгамбай', 'bgn': 'انگامبی', 'bn': 'গাম্বে', 'br': 'ngambayeg', 'brx': 'न्गाम्बै', 'bs': 'ngambaj', 'bs-Cyrl': 'нгамбеј', 'bs-Latn': 'ngambaj', 'ca': 'ngambay', 'ca-ES-valencia': 'ngambay', 'ccp': '𑄚𑄳𑄠𑄉𑄟𑄴𑄝𑄬', 'ce': 'нгамбайн', 'chr': 'ᎾᎦᎹᏇ', 'ckb': 'نگامبای', 'cs': 'ngambay', 'cy': 'Ngambeieg', 'da': 'ngambay', 'de': 'Ngambay', 'dsb': 'ngambayšćina', 'el': 'Νγκαμπέι', 'el-polyton': 'Νγκαμπέι', 'en': 'Ngambay', 'es': 'ngambay', 'et': 'ngambai', 'eu': 'ngambayera', 'fa': 'انگامبایی', 'ff-Adlm': '𞤐𞤺𞤢𞤥𞤦𞤢𞤴𞤪𞤫', 'fi': 'ngambay', 'fil': 'Ngambay', 'fo': 'ngambay', 'fr': 'ngambay', 'fy': 'Ngambay', 'ga': 'Ngambay', 'gd': 'Ngambay', 'gl': 'ngambay', 'gu': 'ન્ગામ્બેય', 'ha': 'Ngambay', 'he': 'נגמבאי', 'hi': 'न्गाम्बे', 'hi-Latn': 'Ngambay', 'hr': 'ngambay', 'hsb': 'ngambayšćina', 'ht': 'ngambay', 'hu': 'ngambay', 'hy': 'նգամբայ', 'ia': 'ngambay', 'id': 'Ngambai', 'ig': 'Ngambay', 'is': 'ngambay', 'it': 'ngambay', 'ja': 'ンガムバイ語', 'jv': 'Ngambai', 'ka': 'ნგამბაი', 'kgp': 'gỹmaji', 'kk': 'нгамбай тілі', 'kk-Cyrl': 'нгамбай тілі', 'km': 'ងាំបេយ', 'kn': 'ನಂಬೇ', 'ko': '느감바이어', 'kok': 'गांबे', 'kok-Deva': 'गांबे', 'ksh': 'Njambaij', 'ku': 'ngambay', 'ky': 'нгамбайча', 'lb': 'Ngambay', 'lo': 'ກຳເບ', 'lt': 'ngambajų', 'lv': 'ngambeju', 'mai': 'नामबे', 'mi': 'Nekāpei', 'mk': 'нгембеј', 'ml': 'ഗംബായ്', 'mn': 'нгамбай', 'mr': 'गाम्बे', 'ms': 'Ngambay', 'mt': 'Ngambay', 'my': 'အွန်ဂမ်းဘေး', 'nb': 'ngambay', 'ne': 'न्गामबाय', 'nl': 'Ngambay', 'nn': 'ngambay', 'no': 'ngambay', 'or': 'ନଗାମବେ', 'pa': 'ਨਗਾਂਬੇ', 'pa-Guru': 'ਨਗਾਂਬੇ', 'pcm': 'Ngambai Lángwej', 'pl': 'ngambay', 'ps': 'نګبای', 'pt': 'ngambay', 'qu': 'Ngambay Simi', 'ro': 'ngambay', 'ru': 'нгамбайский', 'sc': 'ngambay', 'sd': 'نغمبي', 'sd-Arab': 'نغمبي', 'si': 'න්ගම්බෙ', 'sk': 'ngambay', 'sl': 'ngambajščina', 'smn': 'ngambay', 'so': 'Ngambay', 'sq': 'ngambajisht', 'sr': 'нгамбај', 'sr-Cyrl': 'нгамбај', 'sr-Latn': 'ngambaj', 'sv': 'ngambay', 'sw': 'Kingambay', 'ta': 'நெகாம்பே', 'te': 'గాంబే', 'th': 'กัมเบ', 'ti': 'ኤንጋምባይ', 'tk': 'ngambaý dili', 'to': 'lea fakangāmipai', 'tr': 'Ngambay', 'ug': 'نگامبايچە', 'uk': 'нгамбай', 'ur': 'نگامبے', 'uz': 'ngambay', 'uz-Latn': 'ngambay', 'vec': 'nganbay', 'vi': 'Tiếng Ngambay', 'xnr': 'न्गाम्बे', 'yo': 'Èdè Ngambayi', 'yrl': 'ĩgãbai', 'yue': '甘拜文', 'yue-Hans': '甘拜文', 'yue-Hant': '甘拜文', 'zh': '甘拜语', 'zh-Hans': '甘拜语', 'zh-Hant': '甘拜文', 'zu': 'isi-Ngambay'},
    'sbb': {'en': 'Simbo'},
    'sbc': {'en': 'Kele (Papua New Guinea)'},
    'sbd': {'en': 'Southern Samo'},
    'sbe': {'en': 'Saliba'},
    'sbf': {'en': 'Chabu'},
    'sbg': {'en': 'Seget'},
    'sbh': {'en': 'Sori-Harengan'},
    'sbi': {'en': 'Seti'},
    'sbj': {'en': 'Surbakhal'},
    'sbk': {'en': 'Safwa'},
    'sbl': {'en': 'Botolan Sambal'},
    'sbm': {'en': 'Sagala'},
    'sbn': {'en': 'Sindhi Bhil'},
    'sbo': {'en': 'Sabüm'},
    'sbp': {'af': 'Sangu', 'am': 'ሳንጉ', 'ar': 'سانغو', 'as': 'ছাঙ্গু', 'ast': 'sangu', 'az': 'sanqu', 'az-Cyrl': 'сангу', 'az-Latn': 'sanqu', 'bal-Latn': 'Sangu', 'be': 'сангу', 'be-tarask': 'сангу', 'bg': 'сангу', 'bgn': 'سانگویی', 'bn': 'সাঙ্গু', 'br': 'sangu', 'brx': 'साङु', 'bs': 'sangu', 'bs-Cyrl': 'сангу', 'bs-Latn': 'sangu', 'ca': 'sangu', 'ca-ES-valencia': 'sangu', 'ccp': '𑄥𑄁𑄚𑄴𑄉𑄪', 'ce': 'сангу', 'chr': 'ᏌᏁᎫ', 'ckb': 'سانگوو', 'cs': 'sangoština', 'cy': 'Sangw', 'da': 'sangu', 'de': 'Sangu', 'dsb': 'sangu', 'el': 'Σάνγκου', 'el-polyton': 'Σάνγκου', 'en': 'Sangu', 'es': 'sangu', 'et': 'sangu', 'eu': 'sanguera', 'fa': 'سانگویی', 'ff-Adlm': '𞤅𞤢𞤲𞤺𞤵𞥅𞤪𞤫', 'fi': 'sangu', 'fil': 'Sangu', 'fo': 'sangu', 'fr': 'isangu', 'fy': 'Sangu', 'ga': 'Sangu', 'gd': 'Sangu', 'gl': 'sangu', 'gu': 'સાંગુ', 'ha': 'Sangu', 'he': 'סאנגו', 'hi': 'सैंगु', 'hi-Latn': 'Sangu', 'hr': 'sangu', 'hsb': 'sangu', 'ht': 'isangu', 'hu': 'szangu', 'hy': 'սանգու', 'ia': 'sangu', 'id': 'Sangu', 'ig': 'Sangu', 'is': 'sangú', 'it': 'sangu', 'ja': 'サング語', 'jv': 'Sangu', 'ka': 'სანგუ', 'kgp': 'sỹgu', 'kk': 'сангу тілі', 'kk-Cyrl': 'сангу тілі', 'km': 'សានហ្គូ', 'kn': 'ಸಂಗು', 'ko': '상구어', 'kok': 'सांगू', 'kok-Deva': 'सांगू', 'ku': 'sanguyî', 'ky': 'сангуча', 'lb': 'Sangu', 'lo': 'ຊານກູ', 'lrc': 'سانگوٙ', 'lt': 'sangų', 'lv': 'sangu', 'mai': 'सांगू', 'mi': 'Hāngu', 'mk': 'сангу', 'ml': 'സംഗു', 'mn': 'сангү', 'mr': 'सांगु', 'ms': 'Sangu', 'mt': 'Sangu', 'my': 'ဆန်ဂု', 'mzn': 'سانگوو', 'nb': 'sangu', 'ne': 'साङ्गु', 'nl': 'Sangu', 'nn': 'sangu', 'no': 'sangu', 'nqo': 'ߌߛߊ߲ߜ߭ߎߞߊ߲', 'or': 'ସାନଗୁ', 'pa': 'ਸੇਂਗੋ', 'pa-Guru': 'ਸੇਂਗੋ', 'pcm': 'Sangu', 'pl': 'sangu', 'ps': 'سانګوو', 'pt': 'sangu', 'qu': 'Sangu Simi', 'rm': 'sangu', 'ro': 'sangu', 'ru': 'сангу', 'sbp': 'Ishisangu', 'sc': 'sangu', 'sd': 'سانگوو', 'sd-Arab': 'سانگوو', 'si': 'සංගු', 'sk': 'sangu', 'sl': 'sangujščina', 'smn': 'sangu', 'so': 'Sangu', 'sq': 'sanguisht', 'sr': 'сангу', 'sr-Cyrl': 'сангу', 'sr-Latn': 'sangu', 'sv': 'sangu', 'sw': 'Kisangu', 'ta': 'சங்கு', 'te': 'సాంగు', 'th': 'แซงกู', 'ti': 'ሳንጉ', 'tk': 'sangu dili', 'to': 'lea fakasangu', 'tr': 'Sangu', 'ug': 'سانگۇچە', 'uk': 'сангу', 'ur': 'سانگو', 'uz': 'sangu', 'uz-Cyrl': 'сангуча', 'uz-Latn': 'sangu', 'vec': 'sangu', 'vi': 'Tiếng Sangu', 'xnr': 'सैंगु', 'yo': 'Sangu', 'yrl': 'sãgu', 'yue': '桑古文', 'yue-Hans': '桑古文', 'yue-Hant': '桑古文', 'zh': '桑古语', 'zh-Hans': '桑古语', 'zh-Hant': '桑古文', 'zu': 'isi-Sangu'},
    'sbq': {'en': 'Sileibi'},
    'sbr': {'en': 'Sembakung Murut'},
    'sbs': {'en': 'Subiya'},
    'sbt': {'en': 'Kimki'},
    'sbu': {'en': 'Stod Bhoti'},
    'sbv': {'en': 'Sabine'},
    'sbw': {'en': 'Simba'},
    'sbx': {'en': 'Seberuang'},
    'sby': {'en': 'Soli'},
    'sbz': {'en': 'Sara Kaba'},
    'sc': {'af': 'Sardinies', 'ak': 'Saadinia kasa', 'am': 'ሳርዲንያን', 'ar': 'السردينية', 'as': 'ছাৰ্ডিনিয়ান', 'ast': 'sardu', 'az': 'sardin', 'az-Cyrl': 'сардин', 'az-Latn': 'sardin', 'bal-Latn': 'Sardéniái', 'be': 'сардзінская', 'be-tarask': 'сардзінская', 'bg': 'сардински', 'bgn': 'ساردینی', 'bn': 'সার্ডিনিয়ান', 'br': 'sardeg', 'brx': 'सार्दीनीयान', 'bs': 'sardinijski', 'bs-Cyrl': 'сардињаски', 'bs-Latn': 'sardinijski', 'ca': 'sard', 'ca-ES-valencia': 'sard', 'ccp': '𑄥𑄢𑄴𑄓𑄨𑄚𑄨𑄠𑄚𑄴', 'ce': 'сардинийн', 'chr': 'ᏌᏗᏂᎠᏂ', 'ckb': 'ساردینی', 'cs': 'sardština', 'cy': 'Sardeg', 'da': 'sardinsk', 'de': 'Sardisch', 'dsb': 'sardinšćina', 'el': 'Σαρδηνιακά', 'el-polyton': 'Σαρδηνιακά', 'en': 'Sardinian', 'es': 'sardo', 'et': 'sardi', 'eu': 'sardiniera', 'fa': 'ساردینیایی', 'ff-Adlm': '𞤅𞤢𞤪𞤣𞤭𞤲𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'sardi', 'fil': 'Sardinian', 'fo': 'sardiskt', 'fr': 'sarde', 'fur': 'sardegnûl', 'fy': 'Sardinysk', 'ga': 'Sairdínis', 'gd': 'Sàrdais', 'gl': 'sardo', 'gsw': 'Sardisch', 'gu': 'સાર્દિનિયન', 'ha': 'Sardiniyanci', 'he': 'סרדינית', 'hi': 'सार्दिनियन', 'hi-Latn': 'Sardinian', 'hr': 'sardski', 'hsb': 'sardinšćina', 'ht': 'sarde', 'hu': 'szardíniai', 'hy': 'սարդիներեն', 'ia': 'sardo', 'id': 'Sardinia', 'ig': 'Asụsụ Sardini', 'is': 'sardínska', 'it': 'sardo', 'ja': 'サルデーニャ語', 'jv': 'Sardinia', 'ka': 'სარდინიული', 'kgp': 'sarno', 'kk': 'сардин тілі', 'kk-Cyrl': 'сардин тілі', 'km': 'សាឌីនា', 'kn': 'ಸರ್ಡೀನಿಯನ್', 'ko': '사르디니아어', 'kok': 'सार्डिनियान', 'kok-Deva': 'सार्डिनियान', 'ks': 'سراڈیٖنی', 'ks-Arab': 'سراڈیٖنی', 'ku': 'sardînî', 'ky': 'сардинче', 'lb': 'Sardesch', 'lld': 'sard', 'lo': 'ສາດີນຽນ', 'lt': 'sardiniečių', 'lv': 'sardīniešu', 'mai': 'सार्डिनियाई', 'mi': 'Hārinia', 'mk': 'сардински', 'ml': 'സർഡിനിയാൻ', 'mn': 'сардин', 'mr': 'सर्दिनियन', 'ms': 'Sardinia', 'mt': 'Sardinjan', 'my': 'ဆာဒီနီးယား', 'nb': 'sardisk', 'ne': 'सार्डिनियाली', 'nl': 'Sardijns', 'nn': 'sardinsk', 'no': 'sardisk', 'or': 'ସାର୍ଡିନିଆନ୍\u200c', 'pa': 'ਸਾਰਡੀਨੀਆਈ', 'pa-Guru': 'ਸਾਰਡੀਨੀਆਈ', 'pcm': 'Sadínián Lángwej', 'pl': 'sardyński', 'ps': 'سارڊيني', 'pt': 'sardo', 'qu': 'Sardinian Simi', 'rm': 'sard', 'ro': 'sardiniană', 'ru': 'сардинский', 'sc': 'sardu', 'sd': 'سارڊيني', 'sd-Arab': 'سارڊيني', 'se': 'sardigiella', 'si': 'සාර්ඩිනිඅන්', 'sk': 'sardínčina', 'sl': 'sardinščina', 'smn': 'sardiniakielâ', 'so': 'Sardinian', 'sq': 'sardenjisht', 'sr': 'сардински', 'sr-Cyrl': 'сардински', 'sr-Latn': 'sardinski', 'sv': 'sardinska', 'sw': 'Kisardinia', 'ta': 'சார்தீனியன்', 'te': 'సార్డీనియన్', 'th': 'ซาร์เดญา', 'ti': 'ሳርዲንኛ', 'tk': 'sardin dili', 'to': 'lea fakasaletīnia', 'tr': 'Sardunya dili', 'ug': 'ساردىنىيەچە', 'uk': 'сардинська', 'ur': 'سردینین', 'uz': 'sardin', 'uz-Latn': 'sardin', 'vec': 'sardo', 'vi': 'Tiếng Sardinia', 'xnr': 'सार्दिनियन', 'yi': 'סאַרדיש', 'yo': 'Èdè Sadini', 'yrl': 'saridú', 'yue': '撒丁文', 'yue-Hans': '撒丁文', 'yue-Hant': '撒丁文', 'zh': '萨丁语', 'zh-Hans': '萨丁语', 'zh-Hant': '撒丁文', 'zu': 'isi-Sardinian'},
    'sca': {'en': 'Sansu'},
    'scb': {'en': 'Chut'},
    'sce': {'en': 'Dongxiang'},
    'scf': {'en': 'San Miguel Creole French'},
    'scg': {'en': 'Sanggau'},
    'sch': {'en': 'Sakachep'},
    'sci': {'en': 'Sri Lankan Creole Malay'},
    'sck': {'en': 'Sadri'},
    'scl': {'en': 'Shina'},
    'scn': {'af': 'Sisiliaans', 'am': 'ሲሲሊያንኛ', 'ar': 'الصقلية', 'as': 'ছিচিলিয়ান', 'ast': 'sicilianu', 'az': 'siciliya', 'az-Cyrl': 'сиҹилија', 'az-Latn': 'siciliya', 'bal-Latn': 'Sesiliái', 'be': 'сіцылійская', 'be-tarask': 'сіцылійская', 'bg': 'сицилиански', 'bgn': 'سیسیلی', 'bn': 'সিসিলিয়ান', 'br': 'sikilieg', 'brx': 'सिसिलियान', 'bs': 'sicilijanski', 'bs-Cyrl': 'сицилијански', 'bs-Latn': 'sicilijanski', 'ca': 'sicilià', 'ca-ES-valencia': 'sicilià', 'ccp': '𑄥𑄨𑄥𑄨𑄣𑄨𑄠𑄚𑄴', 'ce': 'сицилийн', 'chr': 'ᏏᏏᎵᎠᏂ', 'ckb': 'سیسیلی', 'cs': 'sicilština', 'cy': 'Sisileg', 'da': 'siciliansk', 'de': 'Sizilianisch', 'dsb': 'sicilianišćina', 'el': 'Σικελικά', 'el-polyton': 'Σικελικά', 'en': 'Sicilian', 'es': 'siciliano', 'et': 'sitsiilia', 'eu': 'siziliera', 'fa': 'سیسیلی', 'ff-Adlm': '𞤅𞤭𞤧𞤭𞤤𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'sisilia', 'fil': 'Sicilian', 'fo': 'sisilanskt', 'fr': 'sicilien', 'fur': 'sicilian', 'fy': 'Siciliaansk', 'ga': 'Sicilis', 'gd': 'Sisilis', 'gl': 'siciliano', 'gsw': 'Sizilianisch', 'gu': 'સિસિલિયાન', 'ha': 'Sisiliyanci', 'he': 'סיציליאנית', 'hi': 'सिसिलियन', 'hi-Latn': 'Sicilian', 'hr': 'sicilijski', 'hsb': 'sicilšćina', 'ht': 'sicilien', 'hu': 'szicíliai', 'hy': 'սիցիլիերեն', 'ia': 'siciliano', 'id': 'Sisilia', 'ig': 'Sicilian', 'is': 'sikileyska', 'it': 'siciliano', 'ja': 'シチリア語', 'jv': 'Sisilia', 'ka': 'სიცილიური', 'kgp': 'sisirijỹnũ', 'kk': 'сицилия тілі', 'kk-Cyrl': 'сицилия тілі', 'km': 'ស៊ីស៊ីលាន', 'kn': 'ಸಿಸಿಲಿಯನ್', 'ko': '시칠리아어', 'kok': 'सिसिलियान', 'kok-Deva': 'सिसिलियान', 'ks': 'سِچِلِیَن', 'ks-Arab': 'سِچِلِیَن', 'ku': 'sicîlî', 'ky': 'сицилийче', 'lb': 'Sizilianesch', 'lo': 'ຊີຊິລີນ', 'lt': 'siciliečių', 'lv': 'sicīliešu', 'mai': 'सिसिलियाई', 'mi': 'Hihiriana', 'mk': 'сицилијански', 'ml': 'സിസിലിയൻ', 'mn': 'сицил', 'mr': 'सिसिलियन', 'ms': 'Sicili', 'mt': 'Sqalli', 'my': 'စစ္စလီ', 'nb': 'siciliansk', 'ne': 'सिसिलियाली', 'nl': 'Siciliaans', 'nn': 'siciliansk', 'no': 'siciliansk', 'nqo': 'ߛߌߛߟߌߦߊߞߊ߲', 'or': 'ସିଶିଲିଆନ୍', 'pa': 'ਸਿਸੀਲੀਅਨ', 'pa-Guru': 'ਸਿਸੀਲੀਅਨ', 'pcm': 'Sisílián Lángwej', 'pl': 'sycylijski', 'ps': 'سیلیسي', 'pt': 'siciliano', 'qu': 'Siciliano Simi', 'rm': 'sicilian', 'ro': 'siciliană', 'ru': 'сицилийский', 'sc': 'sitzilianu', 'scn': 'sicilianu', 'sd': 'سسلي', 'sd-Arab': 'سسلي', 'se': 'sisiliagiella', 'si': 'සිසිලියන්', 'sk': 'sicílčina', 'sl': 'sicilijanščina', 'smn': 'sisiliakielâ', 'so': 'Sicilian', 'sq': 'siçilianisht', 'sr': 'сицилијански', 'sr-Cyrl': 'сицилијански', 'sr-Latn': 'sicilijanski', 'sv': 'sicilianska', 'sw': 'Kisicilia', 'ta': 'சிசிலியன்', 'te': 'సిసిలియన్', 'th': 'ซิซิลี', 'ti': 'ሲሲልኛ', 'tk': 'sisiliýa dili', 'to': 'lea fakasisīlia', 'tr': 'Sicilyaca', 'ug': 'سىتسىلىيەچە', 'uk': 'сицилійська', 'ur': 'سیسیلین', 'uz': 'sitsiliya', 'uz-Latn': 'sitsiliya', 'vec': 'sisilian', 'vi': 'Tiếng Sicilia', 'xnr': 'सिसिलियन', 'yi': 'סיציליאַניש', 'yo': 'Èdè Sikila', 'yrl': 'sisirianu', 'yue': '西西里文', 'yue-Hans': '西西里文', 'yue-Hant': '西西里文', 'zh': '西西里语', 'zh-Hans': '西西里语', 'zh-Hant': '西西里文', 'zu': 'isi-Sicilian'},
    'sco': {'af': 'Skots', 'am': 'ስኮትስ', 'ar': 'الأسكتلندية', 'as': 'স্কটছ', 'ast': 'scots', 'az': 'skots', 'az-Cyrl': 'скотс', 'az-Latn': 'skots', 'be': 'шатландская', 'be-tarask': 'шатландская', 'bg': 'шотландски', 'bgn': 'اسکاتلندی', 'bn': 'স্কটস', 'br': 'skoteg', 'brx': 'स्क’टस', 'bs': 'škotski', 'bs-Cyrl': 'шкотски', 'bs-Latn': 'škotski', 'ca': 'escocès', 'ca-ES-valencia': 'escocès', 'ccp': '𑄆𑄌𑄴𑄇𑄧𑄖𑄴𑄥𑄴', 'ce': 'шотландхойн', 'chr': 'ᏍᎦᏗ', 'ckb': 'سکۆتس', 'cs': 'skotština', 'cy': 'Sgoteg', 'da': 'skotsk', 'de': 'Schottisch', 'dsb': 'scotšćina', 'el': 'Σκωτικά', 'el-polyton': 'Σκωτικά', 'en': 'Scots', 'es': 'escocés', 'et': 'šoti', 'eu': 'eskoziera', 'fa': 'اسکاتلندی', 'ff-Adlm': '𞤅𞤭𞤳𞤮𞤼𞤧𞤪𞤫', 'fi': 'skotti', 'fil': 'Scots', 'fo': 'skotskt', 'fr': 'écossais', 'fur': 'scozês', 'fy': 'Schots', 'ga': 'Albainis', 'gd': 'Albais', 'gl': 'escocés', 'gsw': 'Schottisch', 'gu': 'સ્કોટ્સ', 'ha': 'Scots', 'he': 'סקוטית', 'hi': 'स्कॉट्स', 'hi-Latn': 'Scots', 'hr': 'škotski', 'hsb': 'scotšćina', 'ht': 'écossais', 'hu': 'skót', 'hy': 'շոտլանդերեն', 'ia': 'scotese', 'id': 'Skotlandia', 'ig': 'Scots', 'is': 'skoska', 'it': 'scozzese', 'ja': 'スコットランド語', 'jv': 'Skots', 'ka': 'შოტლანდიური', 'kgp': 'isikoti', 'kk': 'шотланд тілі', 'kk-Cyrl': 'шотланд тілі', 'km': 'ស្កុត', 'kn': 'ಸ್ಕೋಟ್ಸ್', 'ko': '스코틀랜드어', 'kok': 'स्कॉट्स', 'kok-Deva': 'स्कॉट्स', 'ks': 'سکاٹس', 'ks-Arab': 'سکاٹس', 'ku': 'skotî', 'ky': 'шотландча', 'lb': 'Schottesch', 'lo': 'ສກອດ', 'lt': 'škotų', 'lv': 'skotu', 'mai': 'स्काउट्स', 'mi': 'Kotimana', 'mk': 'шкотски германски', 'ml': 'സ്കോട്സ്', 'mn': 'шотланд', 'mr': 'स्कॉट्स', 'ms': 'Scots', 'mt': 'Skoċċiż', 'my': 'စကော့တ်', 'nb': 'skotsk', 'ne': 'स्कट्स', 'nl': 'Schots', 'nn': 'skotsk', 'no': 'skotsk', 'or': 'ସ୍କଟସ୍', 'pa': 'ਸਕਾਟਸ', 'pa-Guru': 'ਸਕਾਟਸ', 'pcm': 'Skọ́t Lángwej', 'pl': 'scots', 'ps': 'سکاټس', 'pt': 'scots', 'qu': 'Scots Simi', 'rm': 'scot', 'ro': 'scots', 'ru': 'шотландский', 'sc': 'scots', 'sd': 'اسڪاٽس', 'sd-Arab': 'اسڪاٽس', 'si': 'ස්කොට්ස්', 'sk': 'škótčina', 'sl': 'škotščina', 'smn': 'skootikielâ', 'so': 'Scots', 'sq': 'skotisht', 'sr': 'шкотски', 'sr-Cyrl': 'шкотски', 'sr-Latn': 'škotski', 'sv': 'skotska', 'sw': 'Kiskoti', 'syr': 'ܣܟܘܬܠܢܕܐܝܬ', 'ta': 'ஸ்காட்ஸ்', 'te': 'స్కాట్స్', 'th': 'สกอตส์', 'ti': 'ስኮትኛ', 'tk': 'şotland dili', 'to': 'lea fakasikotilani', 'tr': 'İskoçça', 'ug': 'شوتلاندىيەچە', 'uk': 'шотландська', 'ur': 'سکاٹ', 'uz': 'shotland', 'uz-Latn': 'shotland', 'vec': 'scoseze', 'vi': 'Tiếng Scots', 'xnr': 'स्कॉट्स', 'yi': 'סקאטס', 'yo': 'Èdè Sikoti', 'yrl': 'isiutis', 'yue': '蘇格蘭文', 'yue-Hans': '苏格兰文', 'yue-Hant': '蘇格蘭文', 'zh': '苏格兰语', 'zh-Hans': '苏格兰语', 'zh-Hant': '蘇格蘭文', 'zu': 'isi-Scots'},
    'scp': {'en': 'Hyolmo'},
    'scq': {'en': "Sa'och"},
    'scs': {'en': 'North Slavey'},
    'sct': {'en': 'Southern Katang'},
    'scu': {'en': 'Shumcho'},
    'scv': {'en': 'Sheni'},
    'scw': {'en': 'Sha'},
    'scx': {'en': 'Sicel'},
    'sd': {'af': 'Sindhi', 'ak': 'Sindi', 'am': 'ሲንዲ', 'ar': 'السندية', 'as': 'সিন্ধী', 'ast': 'sindhi', 'az': 'sindhi', 'az-Arab': 'سیندیجه', 'az-Cyrl': 'синдһи', 'az-Latn': 'sindhi', 'bal-Latn': 'Sendi', 'be': 'сіндхі', 'be-tarask': 'сіндхі', 'bg': 'синдхи', 'bgn': 'سیندی', 'bn': 'সিন্ধি', 'br': 'sindhi', 'brx': 'सिन्धी', 'bs': 'sindi', 'bs-Cyrl': 'синди', 'bs-Latn': 'sindi', 'ca': 'sindi', 'ca-ES-valencia': 'sindi', 'ccp': '𑄥𑄨𑄚𑄴𑄙𑄨', 'ce': 'синдхи', 'chr': 'ᏏᏂᏗ', 'ckb': 'سيندی', 'cs': 'sindhština', 'cy': 'Sindhi', 'da': 'sindhi', 'de': 'Sindhi', 'dsb': 'sindšćina', 'dz': 'སིན་དཱི་ཁ', 'ee': 'sindhgbe', 'el': 'Σίντι', 'el-polyton': 'Σίντι', 'en': 'Sindhi', 'eo': 'sinda', 'es': 'sindi', 'et': 'sindhi', 'eu': 'sindhia', 'fa': 'سندی', 'ff-Adlm': '𞤅𞤭𞤲𞤣𞤢𞥄𞤪𞤫', 'fi': 'sindhi', 'fil': 'Sindhi', 'fo': 'sindhi', 'fr': 'sindhi', 'fur': 'sindhi', 'fy': 'Sindhi', 'ga': 'Sindis', 'gd': 'Sindhi', 'gl': 'sindhi', 'gsw': 'Sindhi', 'gu': 'સિંધી', 'ha': 'Sindiyanci', 'he': 'סינדהית', 'hi': 'सिंधी', 'hi-Latn': 'Sindhi', 'hr': 'sindski', 'hsb': 'sindhišćina', 'ht': 'sindhi', 'hu': 'szindhi', 'hy': 'սինդհի', 'ia': 'sindhi', 'id': 'Sindhi', 'ig': 'Asụsụ Sindhi', 'is': 'sindí', 'it': 'sindhi', 'ja': 'シンド語', 'jv': 'Sindhi', 'ka': 'სინდჰური', 'kea': 'sindi', 'kgp': 'sĩni', 'kk': 'синдхи тілі', 'kk-Cyrl': 'синдхи тілі', 'km': 'ស៊ីនឌី', 'kn': 'ಸಿಂಧಿ', 'ko': '신디어', 'kok': 'सिंधी', 'kok-Deva': 'सिंधी', 'ks': 'سِندی', 'ks-Arab': 'سِندی', 'ksh': 'Sinndi', 'ku': 'sindhî', 'kxv': 'sindi', 'kxv-Deva': 'सिंधी', 'kxv-Latn': 'sindi', 'kxv-Orya': 'ସିନ୍ଦୀ', 'kxv-Telu': 'సిందీ', 'ky': 'синдхиче', 'lb': 'Sindhi', 'lkt': 'Sindhi Iyápi', 'lld': 'sindhi', 'lo': 'ສິນທິ', 'lrc': 'سئندی', 'lt': 'sindų', 'lv': 'sindhu', 'mai': 'सिंधी', 'mi': 'Hiniti', 'mk': 'синди', 'ml': 'സിന്ധി', 'mn': 'синдхи', 'mr': 'सिंधी', 'ms': 'Sindhi', 'mt': 'Sindhi', 'my': 'စင်ဒီ', 'mzn': 'سندی', 'nb': 'sindhi', 'ne': 'सिन्धी', 'nl': 'Sindhi', 'nn': 'sindhi', 'no': 'sindhi', 'nqo': 'ߛߌ߲ߘߌߞߊ߲', 'or': 'ସିନ୍ଧୀ', 'pa': 'ਸਿੰਧੀ', 'pa-Guru': 'ਸਿੰਧੀ', 'pcm': 'Síndí', 'pl': 'sindhi', 'ps': 'سندهي', 'pt': 'sindi', 'qu': 'Sindhi Simi', 'rm': 'sindhi', 'ro': 'sindhi', 'ru': 'синдхи', 'rw': 'Igisindi', 'sc': 'sindhi', 'sd': 'سنڌي', 'sd-Arab': 'سنڌي', 'sd-Deva': 'सिन्धी', 'si': 'සින්ධි', 'sk': 'sindhčina', 'sl': 'sindščina', 'smn': 'sindhi', 'so': 'Siindhi', 'sq': 'sindisht', 'sr': 'синди', 'sr-Cyrl': 'синди', 'sr-Latn': 'sindi', 'sv': 'sindhi', 'sw': 'Kisindhi', 'ta': 'சிந்தி', 'te': 'సింధీ', 'tg': 'синдӣ', 'th': 'สินธิ', 'ti': 'ሲንድሂ', 'tk': 'sindhi dili', 'to': 'lea fakasīniti', 'tr': 'Sindhi dili', 'tt': 'синдһи', 'ug': 'سىندىچە', 'uk': 'синдхі', 'ur': 'سندھی', 'uz': 'sindhi', 'uz-Cyrl': 'синдҳи', 'uz-Latn': 'sindhi', 'vec': 'sindi', 'vi': 'Tiếng Sindhi', 'wae': 'Sindhi', 'wo': 'Sindi', 'xnr': 'सिंधी', 'yi': 'סינדהי', 'yo': 'Èdè Sindhi', 'yrl': 'sĩdi', 'yue': '信德文', 'yue-Hans': '信德文', 'yue-Hant': '信德文', 'zh': '信德语', 'zh-Hans': '信德语', 'zh-Hant': '信德文', 'zu': 'isi-Sindhi'},
    'sda': {'en': "Toraja-Sa'dan"},
    'sdb': {'en': 'Shabak'},
    'sdc': {'ast': 'sardu sassarés', 'br': 'sasareseg', 'ca': 'sasserès', 'ca-ES-valencia': 'sasserès', 'cs': 'sassarština', 'cy': 'Sasareseg Sardinia', 'de': 'Sassarisch', 'en': 'Sassarese Sardinian', 'fi': 'sassarinsardi', 'fr': 'sarde sassarais', 'gd': 'Sassarese', 'hi-Latn': 'Sassarese Sardinian', 'ht': 'sarde sassarais', 'it': 'sassarese', 'ja': 'サッサリ・サルデーニャ語', 'lb': 'Sassaresesch', 'lt': 'sasaresų sardinų', 'mk': 'сасарски сардински', 'nb': 'sassaresisk sardisk', 'nl': 'Sassarees', 'nn': 'sassaresisk sardisk', 'no': 'sassaresisk sardisk', 'pl': 'sassarski', 'sv': 'sassaresisk sardiska', 'th': 'ซาร์ดิเนียซาสซารี', 'to': 'lea fakasaletīnia-sasalesu', 'tr': 'Sassari Sarduca', 'yue': '薩丁尼亞-薩薩里文', 'yue-Hans': '萨丁尼亚-萨萨里文', 'yue-Hant': '薩丁尼亞-薩薩里文', 'zh-Hant': '薩丁尼亞-薩薩里文'},
    'sde': {'en': 'Surubu'},
    'sdf': {'en': 'Sarli'},
    'sdg': {'en': 'Savi'},
    'sdh': {'af': 'Suid-Koerdies', 'am': 'ደቡባዊ ኩርዲሽ', 'ar': 'الكردية الجنوبية', 'ast': 'kurdu del sur', 'az': 'cənubi kürd', 'az-Arab': 'گۆنئی کۆردجه', 'az-Latn': 'cənubi kürd', 'bal-Latn': 'Zerbári Kordi', 'be': 'паўднёвакурдская', 'be-tarask': 'паўднёвакурдская', 'bg': 'южнокюрдски', 'bn': 'দক্ষিণ কুর্দিশ', 'bs': 'južni kurdski', 'bs-Latn': 'južni kurdski', 'ca': 'kurd meridional', 'ca-ES-valencia': 'kurd meridional', 'ccp': '𑄘𑄧𑄉𑄨𑄚𑄴 𑄇𑄪𑄢𑄴𑄘𑄨𑄌𑄴', 'ckb': 'کوردیی باشووری', 'cs': 'kurdština (jižní)', 'cy': 'Cwrdeg Deheuol', 'da': 'sydkurdisk', 'de': 'Südkurdisch', 'el': 'Νότια Κουρδικά', 'el-polyton': 'Νότια Κουρδικά', 'en': 'Southern Kurdish', 'es': 'kurdo meridional', 'et': 'lõunakurdi', 'fa': 'کردی جنوبی', 'fi': 'eteläkurdi', 'fil': 'Katimugang Kurdish', 'fo': 'suður kurdiskt', 'fr': 'kurde du Sud', 'gd': 'Cùrdais Dheasach', 'gl': 'kurdo meridional', 'gu': 'સર્ઘન કુર્દીશ', 'he': 'כורדית דרומית', 'hi': 'दक्षिणी कार्डिश', 'hi-Latn': 'Southern Kurdish', 'hr': 'južnokurdski', 'ht': 'kurde du Sud', 'hu': 'dél-kurd', 'hy': 'հարավային քրդերեն', 'ia': 'kurdo del sud', 'id': 'Kurdi Selatan', 'ig': 'Southern Kurdish', 'is': 'suðurkúrdíska', 'it': 'curdo meridionale', 'ja': '南部クルド語', 'ka': 'სამხრეთქურთული', 'kgp': 'kurno sur', 'kk': 'оңтүстік күрд тілі', 'kk-Cyrl': 'оңтүстік күрд тілі', 'km': 'ឃើដភាគខាងត្បូង', 'kn': 'ದಕ್ಷಿಣ ಕುರ್ದಿಶ್', 'ko': '남부 쿠르드어', 'ku': 'kurdîya başûrî', 'ky': 'түштүк курдча', 'lo': 'ພາກໄຕ້ ຂອງ ກູດິດ', 'lrc': 'کوردی ھارگە', 'lt': 'pietų kurdų', 'lv': 'dienvidkurdu', 'mk': 'јужнокурдски', 'ml': 'തെക്കൻ കുർദ്ദിഷ്', 'mr': 'दक्षिणी कुर्दिश', 'ms': 'Kurdish Selatan', 'mzn': 'جنوبی کردی', 'nb': 'sørkurdisk', 'ne': 'दक्षिणी कुर्दिश', 'nl': 'Pahlavani', 'nn': 'sørkurdisk', 'no': 'sørkurdisk', 'pa': 'ਦੱਖਣੀ ਕੁਰਦਿਸ਼', 'pa-Guru': 'ਦੱਖਣੀ ਕੁਰਦਿਸ਼', 'pl': 'południowokurdyjski', 'pt': 'curdo meridional', 'ro': 'kurdă de sud', 'ru': 'южнокурдский', 'sdh': 'کوردی خوارگ', 'si': 'දකුණු කුර්දි', 'sk': 'južná kurdčina', 'sl': 'južna kurdščina', 'sq': 'kurdishte jugore', 'sr': 'јужнокурдски', 'sr-Cyrl': 'јужнокурдски', 'sr-Latn': 'južnokurdski', 'sv': 'sydkurdiska', 'sw': 'Kikurdi cha Kusini', 'ta': 'தெற்கு குர்திஷ்', 'te': 'దక్షిణ కుర్డిష్', 'th': 'เคอร์ดิชใต้', 'ti': 'ደቡባዊ ኩርዲሽ', 'to': 'lea faka-tonga ‘o Ketesi', 'tr': 'Güney Kürtçesi', 'uk': 'південнокурдська', 'ur': 'جنوبی کرد', 'uz': 'janubiy kurd', 'uz-Latn': 'janubiy kurd', 'vec': 'curdo meridionale', 'vi': 'Tiếng Kurd Miền Nam', 'yrl': 'kuridu meridiunau', 'yue': '南庫爾德文', 'yue-Hans': '南库尔德文', 'yue-Hant': '南庫爾德文', 'zh': '南库尔德语', 'zh-Hans': '南库尔德语', 'zh-Hant': '南庫德文', 'zu': 'i-Southern Kurdish'},
    'sdj': {'en': 'Suundi'},
    'sdk': {'en': 'Sos Kundi'},
    'sdl': {'en': 'Saudi Arabian Sign Language'},
    'sdm': {'en': 'Semandang'},
    'sdn': {'en': 'Gallurese Sardinian'},
    'sdo': {'en': 'Bukar-Sadung Bidayuh'},
    'sdp': {'en': 'Sherdukpen'},
    'sdq': {'en': 'Semandang'},
    'sdr': {'en': 'Oraon Sadri'},
    'sds': {'en': 'Sened'},
    'sdt': {'en': 'Shuadit'},
    'sdu': {'en': 'Sarudu'},
    'sdv': {'en': 'Eastern Sudanic languages'},
    'sdx': {'en': 'Sibu Melanau'},
    'sdz': {'en': 'Sallands'},
    'se': {'af': 'Noord-Sami', 'am': 'ሰሜናዊ ሳሚ', 'ar': 'سامي الشمالية', 'as': 'উদীচ্য ছামি', 'ast': 'sami del norte', 'az': 'şimali sami', 'az-Cyrl': 'шимали сами', 'az-Latn': 'şimali sami', 'bal-Latn': 'Shemáli Sámi', 'be': 'паўночнасаамская', 'be-tarask': 'паўночнасаамская', 'bg': 'северносаамски', 'bgn': 'بُرزسرین سامی', 'bn': 'উত্তরাঞ্চলীয় সামি', 'br': 'sámi an Norzh', 'brx': 'साहायारि सामी', 'bs': 'sjeverni sami', 'bs-Cyrl': 'сјеверни сами', 'bs-Latn': 'sjeverni sami', 'ca': 'sami septentrional', 'ca-ES-valencia': 'sami septentrional', 'ccp': '𑄅𑄖𑄴𑄖𑄧𑄢𑄴 𑄢𑄬𑄌𑄴𑄎𑄮𑄢𑄴 𑄥𑄟𑄨', 'ce': 'къилбаседа саамийн', 'chr': 'ᏧᏴᏢ ᏗᏜ ᏌᎻ', 'ckb': 'سامیی باکوور', 'cs': 'sámština (severní)', 'cy': 'Sami Gogleddol', 'da': 'nordsamisk', 'de': 'Nordsamisch', 'dsb': 'lapšćina', 'ee': 'dziehe samigbe', 'el': 'Βόρεια Σάμι', 'el-polyton': 'Βόρεια Σάμι', 'en': 'Northern Sami', 'es': 'sami septentrional', 'et': 'põhjasaami', 'eu': 'iparraldeko samiera', 'fa': 'سامی شمالی', 'ff-Adlm': '𞤅𞤢𞤥𞤭𞥅𞤪𞤫 𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤫', 'fi': 'pohjoissaame', 'fil': 'Hilagang Sami', 'fo': 'norður sámiskt', 'fr': 'same du Nord', 'fur': 'sami setentrionâl', 'fy': 'Noard-Samysk', 'ga': 'Sáimis an Tuaiscirt', 'gd': 'Sàmais Thuathach', 'gl': 'saami setentrional', 'gsw': 'Nord-Samisch', 'gu': 'ઉત્તરી સામી', 'ha': 'Sami ta Arewa', 'he': 'סמי צפונית', 'hi': 'नॉर्दन सामी', 'hi-Latn': 'Northern Sami', 'hr': 'sjeverni sami', 'hsb': 'sewjerosamišćina', 'ht': 'same du Nord', 'hu': 'északi számi', 'hy': 'հյուսիսային սաամի', 'ia': 'sami del nord', 'id': 'Sami Utara', 'ig': 'Northern Sami', 'is': 'norðursamíska', 'it': 'sami del nord', 'ja': '北サーミ語', 'jv': 'Sami Sisih Lor', 'ka': 'ჩრდილოეთ საამური', 'kgp': 'samĩ nãrti', 'kk': 'солтүстік саам тілі', 'kk-Cyrl': 'солтүстік саам тілі', 'km': 'សាមីខាងជើង', 'kn': 'ಉತ್ತರ ಸಾಮಿ', 'ko': '북부 사미어', 'kok': 'उत्तरीय सामी', 'kok-Deva': 'उत्तरीय सामी', 'ks': 'شُمٲلی سَمی', 'ks-Arab': 'شُمٲلی سَمی', 'ksh': 'Nood-Lappländesch', 'ku': 'samîya bakur', 'ky': 'түндүк саамиче', 'lb': 'Nordsamesch', 'lld': 'sami dl nord', 'lo': 'ຊາມິເໜືອ', 'lrc': 'سامی شومالی', 'lt': 'šiaurės samių', 'lv': 'ziemeļsāmu', 'mai': 'उत्तरी सामी', 'mi': 'Hami ki te Raki', 'mk': 'северен сами', 'ml': 'വടക്കൻ സമി', 'mn': 'хойд сами', 'mr': 'उत्तरी सामी', 'ms': 'Sami Utara', 'mt': 'Sami tat-Tramuntana', 'my': 'မြောက် ဆာမိ', 'mzn': 'شمالی سامی', 'nb': 'nordsamisk', 'ne': 'उत्तरी सामी', 'nl': 'Noord-Samisch', 'nn': 'nordsamisk', 'no': 'nordsamisk', 'nqo': 'ߛߋߡߌ߫ ߕߟߋ߬ߓߐ߬ߞߊ߲', 'or': 'ଉତ୍ତର ସାମି', 'pa': 'ਉੱਤਰੀ ਸਾਮੀ', 'pa-Guru': 'ਉੱਤਰੀ ਸਾਮੀ', 'pcm': 'Nọ́tán Sámí Lángwej', 'pl': 'północnolapoński', 'ps': 'شمالي سامي', 'pt': 'sami setentrional', 'qu': 'Chincha Sami Simi', 'rm': 'sami dal nord', 'ro': 'sami de nord', 'ru': 'северносаамский', 'sc': 'sami setentrionale', 'sd': 'اتر سامي', 'sd-Arab': 'اتر سامي', 'se': 'davvisámegiella', 'si': 'උතුරු සාමි', 'sk': 'saamčina (severná)', 'sl': 'severna samijščina', 'smn': 'tavekielâ', 'so': 'Sami Waqooyi', 'sq': 'samishte veriore', 'sr': 'северни сами', 'sr-Cyrl': 'северни сами', 'sr-Latn': 'severni sami', 'sv': 'nordsamiska', 'sw': 'Kisami cha Kaskazini', 'ta': 'வடக்கு சமி', 'te': 'ఉత్తర సామి', 'tg': 'самии шимолӣ', 'th': 'ซามิเหนือ', 'ti': 'ሰሜናዊ ሳሚ', 'tk': 'demirgazyk saam dili', 'to': 'lea fakasami-tokelau', 'tr': 'Kuzey Laponcası', 'tt': 'төньяк саам', 'ug': 'شىمالىي سامىچە', 'uk': 'північносаамська', 'ur': 'شمالی سامی', 'uz': 'shimoliy saam', 'uz-Cyrl': 'шимолий саамча', 'uz-Latn': 'shimoliy saam', 'vec': 'sami setentrionale', 'vi': 'Tiếng Sami Miền Bắc', 'wae': 'Nordsamiš', 'wo': 'Penku Sami', 'xnr': 'उतरी सामी', 'yi': 'נארדסאַמיש', 'yo': 'Apáàríwá Sami', 'yrl': 'sami setẽtiriunau', 'yue': '北方薩米文', 'yue-Hans': '北方萨米文', 'yue-Hant': '北方薩米文', 'zh': '北方萨米语', 'zh-Hans': '北方萨米语', 'zh-Hant': '北薩米文', 'zu': 'isi-Northern Sami'},
    'sea': {'en': 'Semai'},
    'seb': {'en': 'Shempire Senoufo'},
    'sec': {'en': 'Sechelt'},
    'sed': {'en': 'Sedang'},
    'see': {'ar': 'السنيكا', 'ast': 'séneca', 'bs': 'seneka', 'bs-Latn': 'seneka', 'ca': 'seneca', 'ca-ES-valencia': 'seneca', 'chr': 'ᏏᏂᎦ', 'cs': 'seneca', 'cy': 'Seneca', 'da': 'seneca', 'de': 'Seneca', 'el': 'Σένεκα', 'el-polyton': 'Σένεκα', 'en': 'Seneca', 'es': 'seneca', 'et': 'seneka', 'fi': 'seneca', 'fr': 'seneca', 'fy': 'Seneca', 'gd': 'Seneca', 'he': 'סנקה', 'hi-Latn': 'Seneca', 'hr': 'seneca', 'ht': 'seneca', 'hu': 'szeneka', 'id': 'Seneca', 'it': 'seneca', 'ja': 'セネカ語', 'ka': 'სენეკა', 'kgp': 'senẽka', 'ko': '세네카어', 'lb': 'Seneca', 'lo': 'ຊີນີກາ', 'lt': 'senecų', 'lv': 'seneku', 'mk': 'сенека', 'ml': 'സെനേക', 'ms': 'Seneca', 'nb': 'seneca', 'nl': 'Seneca', 'nn': 'seneca', 'no': 'seneca', 'pl': 'seneka', 'pt': 'seneca', 'ro': 'seneca', 'ru': 'сенека', 'sk': 'senekčina', 'sv': 'seneca', 'th': 'เซนิกา', 'to': 'lea fakaseneka', 'tr': 'Seneca dili', 'ug': 'سېكنېكاچە', 'uk': 'сенека', 'vi': 'Tiếng Seneca', 'yrl': 'seneka', 'yue': '塞訥卡文', 'yue-Hans': '塞讷卡文', 'yue-Hant': '塞訥卡文', 'zh': '塞内卡语', 'zh-Hans': '塞内卡语', 'zh-Hant': '塞訥卡文'},
    'sef': {'en': 'Cebaara Senoufo'},
    'seg': {'en': 'Segeju'},
    'seh': {'af': 'Sena', 'am': 'ሴና', 'ar': 'سينا', 'as': 'ছেনা', 'ast': 'sena', 'az': 'sena', 'az-Cyrl': 'сена', 'az-Latn': 'sena', 'bal-Latn': 'Sená', 'be': 'сена', 'be-tarask': 'сена', 'bg': 'сена', 'bgn': 'سینایی', 'bn': 'সেনা', 'br': 'sena', 'brx': 'सेना', 'bs': 'sena', 'bs-Cyrl': 'сена', 'bs-Latn': 'sena', 'ca': 'sena', 'ca-ES-valencia': 'sena', 'ccp': '𑄥𑄬𑄚', 'ce': 'сена', 'chr': 'ᏎᎾ', 'ckb': 'سێنا', 'cs': 'sena', 'cy': 'Sena', 'da': 'sena', 'de': 'Sena', 'dsb': 'sena', 'el': 'Σένα', 'el-polyton': 'Σένα', 'en': 'Sena', 'es': 'sena', 'et': 'sena', 'eu': 'senera', 'fa': 'سنا', 'ff-Adlm': '𞤅𞤫𞤲𞤢𞥄𞤪𞤫', 'fi': 'sena', 'fil': 'Sena', 'fo': 'sena', 'fr': 'cisena', 'fy': 'Sena', 'ga': 'Sena', 'gd': 'Sena', 'gl': 'sena', 'gu': 'સેના', 'ha': 'Sena', 'he': 'סנה', 'hi': 'सेना', 'hi-Latn': 'Sena', 'hr': 'sena', 'hsb': 'sena', 'ht': 'cisena', 'hu': 'szena', 'hy': 'սենա', 'ia': 'sena', 'id': 'Sena', 'ig': 'Sena', 'is': 'sena', 'it': 'sena', 'ja': 'セナ語', 'jv': 'Sena', 'ka': 'სენა', 'kgp': 'senỹ', 'kk': 'сена тілі', 'kk-Cyrl': 'сена тілі', 'km': 'ស៊ីណា', 'kn': 'ಸೆನ', 'ko': '세나어', 'kok': 'सेना', 'kok-Deva': 'सेना', 'ku': 'sena', 'ky': 'сенача', 'lb': 'Sena', 'lo': 'ຊີນາ', 'lrc': 'سئنا', 'lt': 'senų', 'lv': 'senu', 'mai': 'सेना', 'mi': 'Hena', 'mk': 'сена', 'ml': 'സേന', 'mn': 'сена', 'mr': 'सेना', 'ms': 'Sena', 'mt': 'Sena', 'my': 'စီနာ', 'mzn': 'سِنایی', 'nb': 'sena', 'ne': 'सेना', 'nl': 'Sena', 'nn': 'sena', 'no': 'sena', 'nqo': 'ߛߌߛߋߣߊߞߊ߲', 'or': 'ସେନା', 'pa': 'ਸੇਨਾ', 'pa-Guru': 'ਸੇਨਾ', 'pcm': 'Sẹ́ná', 'pl': 'sena', 'ps': 'سینا', 'pt': 'sena', 'qu': 'Sena Simi', 'rm': 'sena', 'ro': 'sena', 'ru': 'сена', 'sc': 'sena', 'sd': 'سينا', 'sd-Arab': 'سينا', 'seh': 'sena', 'si': 'සෙනා', 'sk': 'sena', 'sl': 'sena', 'smn': 'sena', 'so': 'Seena', 'sq': 'senaisht', 'sr': 'сена', 'sr-Cyrl': 'сена', 'sr-Latn': 'sena', 'sv': 'sena', 'sw': 'Kisena', 'ta': 'செனா', 'te': 'సెనా', 'th': 'เซนา', 'ti': 'ሰና', 'tk': 'sena dili', 'to': 'lea fakasena', 'tr': 'Sena', 'ug': 'سېناچە', 'uk': 'сена', 'ur': 'سینا', 'uz': 'sena', 'uz-Cyrl': 'сена', 'uz-Latn': 'sena', 'vec': 'sena', 'vi': 'Tiếng Sena', 'xnr': 'सेना', 'yo': 'Ṣẹnà', 'yrl': 'sena', 'yue': '賽納文', 'yue-Hans': '赛纳文', 'yue-Hant': '賽納文', 'zh': '塞纳语', 'zh-Hans': '塞纳语', 'zh-Hant': '賽納文', 'zu': 'isi-Sena'},
    'sei': {'ast': 'seri', 'cs': 'seriština', 'cy': 'Seri', 'de': 'Seri', 'en': 'Seri', 'et': 'seri', 'fi': 'seri', 'fr': 'séri', 'gd': 'Seri', 'hi-Latn': 'Seri', 'ht': 'séri', 'id': 'Seri', 'it': 'seri', 'ja': 'セリ語', 'lb': 'Seri', 'lt': 'seri', 'mk': 'сери', 'nb': 'seri', 'nl': 'Seri', 'nn': 'seri', 'no': 'seri', 'pl': 'seri', 'sv': 'seri', 'th': 'เซรี', 'to': 'lea fakaseli', 'tr': 'Seri', 'yue': '瑟里文', 'yue-Hans': '瑟里文', 'yue-Hant': '瑟里文', 'zh-Hant': '瑟里文'},
    'sej': {'en': 'Sene'},
    'sek': {'en': 'Sekani'},
    'sel': {'ar': 'السيلكب', 'ast': 'selkup', 'az': 'selkup', 'az-Latn': 'selkup', 'bg': 'селкуп', 'bn': 'সেল্কুপ', 'brx': 'सेलकुप', 'bs': 'selkup', 'bs-Cyrl': 'селкап', 'bs-Latn': 'selkup', 'ca': 'selkup', 'ca-ES-valencia': 'selkup', 'ccp': '𑄥𑄬𑄣𑄴𑄇𑄪𑄛𑄴', 'cs': 'selkupština', 'cy': 'Selcypeg', 'da': 'selkupisk', 'de': 'Selkupisch', 'el': 'Σελκούπ', 'el-polyton': 'Σελκούπ', 'en': 'Selkup', 'es': 'selkup', 'et': 'sölkupi', 'fa': 'سلکوپی', 'fi': 'selkuppi', 'fr': 'selkoupe', 'fy': 'Selkup', 'gd': 'Selkup', 'gsw': 'Selkupisch', 'gu': 'સેલ્કપ', 'he': 'סלקופ', 'hi': 'सेल्कप', 'hi-Latn': 'Selkup', 'hr': 'selkupski', 'ht': 'selkoupe', 'hu': 'szölkup', 'id': 'Selkup', 'is': 'selkúp', 'it': 'selkup', 'ja': 'セリクプ語', 'ka': 'სელკუპური', 'kgp': 'serkum', 'kn': 'ಸೆಲ್ಕಪ್', 'ko': '셀쿠프어', 'ks': 'سیلکُپ', 'ks-Arab': 'سیلکُپ', 'lb': 'Selkupesch', 'lo': 'ເຊນຄັບ', 'lt': 'selkup', 'lv': 'selkupu', 'mk': 'селкупски', 'ml': 'സെൽകപ്', 'mr': 'सेल्कप', 'mt': 'Selkup', 'nb': 'selkupisk', 'nl': 'Selkoeps', 'nn': 'selkupisk', 'no': 'selkupisk', 'or': 'ସେଲ୍କପ୍', 'pl': 'selkupski', 'pt': 'selkup', 'rm': 'selkup', 'ro': 'selkup', 'ru': 'селькупский', 'se': 'selkupagiella', 'sk': 'selkupčina', 'sl': 'selkupščina', 'sr': 'селкупски', 'sr-Cyrl': 'селкупски', 'sr-Latn': 'selkupski', 'sv': 'selkup', 'ta': 'செல்குப்', 'te': 'సేల్కప్', 'th': 'เซลคุป', 'to': 'lea fakaselikupi', 'tr': 'Selkup dili', 'ug': 'سېلكاپچە', 'uk': 'селькупська', 'vi': 'Tiếng Selkup', 'yrl': 'seukupi', 'yue': '瑟爾卡普文', 'yue-Hans': '瑟尔卡普文', 'yue-Hant': '瑟爾卡普文', 'zh': '塞尔库普语', 'zh-Hans': '塞尔库普语', 'zh-Hant': '塞爾庫普文'},
    'sem': {'en': 'Semitic languages'},
    'sem-amm': {'en': 'Ammonite'},
    'sem-pro': {'en': 'Proto-Semitic'},
    'sem-srb': {'en': 'Old South Arabian'},
    'sen': {'en': 'Nanerigé Sénoufo'},
    'seo': {'en': 'Suarmin'},
    'sep': {'en': 'Sìcìté Sénoufo'},
    'seq': {'en': 'Senara Sénoufo'},
    'ser': {'en': 'Serrano'},
    'ses': {'af': 'Koyraboro Senni', 'am': 'ኮይራቦሮ ሴኒ', 'ar': 'كويرابورو سيني', 'as': 'কোইৰাবোৰো চেন্নি', 'ast': 'koyraboro senni', 'az': 'koyraboro senni', 'az-Cyrl': 'којраборо сенни', 'az-Latn': 'koyraboro senni', 'bal-Latn': 'Koryáburó Senni', 'be': 'кайрабора сэні', 'be-tarask': 'кайрабора сэні', 'bg': 'койраборо сени', 'bgn': 'کویرابورا سنی', 'bn': 'কোয়রাবোরো সেন্নি', 'br': 'koyraboro senni', 'brx': 'कईराबर’ सेन्नि', 'bs': 'kojraboro seni', 'bs-Cyrl': 'којраборо сени', 'bs-Latn': 'kojraboro seni', 'ca': 'songhai oriental', 'ca-ES-valencia': 'songhai oriental', 'ccp': '𑄇𑄱𑄢𑄝𑄬𑄚𑄮 𑄥𑄬𑄚𑄳𑄚𑄨', 'ce': 'койраборо сенни', 'chr': 'ᎪᏱᎳᏈᎶ ᏎᏂ', 'ckb': 'کۆیرابۆرۆ سێنی', 'cs': 'koyraboro senni', 'cy': 'Koyraboro Senni', 'da': 'koyraboro senni', 'de': 'Koyra Senni', 'dsb': 'koyra senni', 'el': 'Κοϊραμπόρο Σένι', 'el-polyton': 'Κοϊραμπόρο Σένι', 'en': 'Koyraboro Senni', 'es': 'koyraboro senni', 'et': 'koyraboro senni', 'eu': 'koyraboro sennia', 'fa': 'کویرابورا سنی', 'ff-Adlm': '𞤑𞤮𞤪𞤮𞤦𞤮𞤪𞤮𞥅𞤪𞤫 𞤅𞤫𞤲𞥆𞤭', 'fi': 'koyraboro senni', 'fil': 'Koyraboro Senni', 'fo': 'koyraboro senni', 'fr': 'koyraboro senni', 'fy': 'Koyraboro Senni', 'ga': 'Songais Oirthearach', 'gd': 'Koyraboro Senni', 'gl': 'koyraboro senni', 'gu': 'કોયરાબોરો સેન્ની', 'ha': 'Koyraboro Senni', 'he': 'קויראבורו סני', 'hi': 'कोयराबोरो सेन्नी', 'hi-Latn': 'Koyraboro Senni', 'hr': 'koyraboro senni', 'hsb': 'koyra senni', 'ht': 'koyraboro senni', 'hu': 'kojra-szenni', 'hy': 'կոյրաբորո սեննի', 'ia': 'koyraboro senni', 'id': 'Koyraboro Senni', 'ig': 'Koyraboro Senni', 'is': 'koíraboró-senní', 'it': 'koyraboro senni', 'ja': 'コイラボロ・センニ語', 'jv': 'Koyraboro Senni', 'ka': 'კოირაბორო-სენი', 'kea': 'koiraboro seni', 'kgp': 'kojyramoro senĩ', 'kk': 'койраборо сенни тілі', 'kk-Cyrl': 'койраборо сенни тілі', 'km': 'គុយរ៉ាបូរ៉ុស៊ីនី', 'kn': 'ಕೊಯ್ರಬೊರೊ ಸೆನ್ನಿ', 'ko': '코이야보로 세니어', 'kok': 'कोयराबोरो सेन्नी', 'kok-Deva': 'कोयराबोरो सेन्नी', 'ksh': 'Kojraboro Senni', 'ku': 'sonxayî', 'ky': 'койраборо сенниче', 'lb': 'Koyra Senni', 'lo': 'ໂຄຍຣາໂບໂຣ ເຊນນິ', 'lrc': 'کیارابورو سئنی', 'lt': 'kojraboro seni', 'lv': 'koiraboro senni', 'mai': 'कोयराबोरो सैनी', 'mi': 'Kōiraporo Heni', 'mk': 'којраборо сени', 'ml': 'കൊയ്റാബൊറോ സെന്നി', 'mn': 'кёраборо сени', 'mr': 'कोयराबोरो सेन्नी', 'ms': 'Koyraboro Senni', 'mt': 'Koyraboro Senni', 'my': 'ကိုရာဘိုရို ဆမ်နီ', 'mzn': 'کویرابورا سنی', 'nb': 'koyraboro senni', 'ne': 'कोयराबोरो सेन्नी', 'nl': 'Koyraboro Senni', 'nn': 'koyraboro senni', 'no': 'koyraboro senni', 'nqo': 'ߞߏߦߌߙߊߓߙߏ߫ ߛߋߣߌߞߊ߲', 'or': 'କୋୟରା ସେନ୍ନି', 'pa': 'ਕੋਇਰਾਬੋਰੋ ਸੇਂਨੀ', 'pa-Guru': 'ਕੋਇਰਾਬੋਰੋ ਸੇਂਨੀ', 'pcm': 'Kóiraboró Sẹ́nní Lángwej', 'pl': 'koyraboro senni', 'ps': 'کوییرابورو سینی', 'pt': 'koyraboro senni', 'qu': 'Koyraboro Senni Simi', 'rm': 'koyraboro senni', 'ro': 'koyraboro Senni', 'ru': 'койраборо сенни', 'sc': 'koyraboro senni', 'sd': 'ڪيورابورو سيني', 'sd-Arab': 'ڪيورابورو سيني', 'ses': 'Koyraboro senni', 'si': 'කෝයිරාබොරො සෙන්නි', 'sk': 'koyraboro senni', 'sl': 'koyraboro senni', 'smn': 'koyraboro senni', 'so': 'Koyraboro Seenni', 'sq': 'senishte kojrabore', 'sr': 'којраборо сени', 'sr-Cyrl': 'којраборо сени', 'sr-Latn': 'kojraboro seni', 'sv': 'songhai', 'sw': 'Kikoyraboro Senni', 'ta': 'கொய்ராபோரோ சென்னி', 'te': 'కోయోరాబోరో సెన్నీ', 'th': 'โคย์ราโบโรเซนนี', 'ti': 'ኮይራቦሮ ሰኒ', 'tk': 'koýraboro-senni dili', 'to': 'lea fakakoilapolo-seni', 'tr': 'Koyraboro Senni', 'ug': 'شەرقىي سوڭخايچە', 'uk': 'койраборо сені', 'ur': 'کويرابورو سينی', 'uz': 'koyraboro-senni', 'uz-Cyrl': 'койраборо-сенни', 'uz-Latn': 'koyraboro-senni', 'vec': 'koyraboro seni', 'vi': 'Tiếng Koyraboro Senni', 'xnr': 'कोयराबोरो सेन्नी', 'yo': 'Koiraboro Seni', 'yrl': 'kuiraburu seni', 'yue': '東桑海文', 'yue-Hans': '东桑海文', 'yue-Hant': '東桑海文', 'zh': '东桑海语', 'zh-Hans': '东桑海语', 'zh-Hant': '東桑海文', 'zu': 'isi-Koyraboro Senni'},
    'set': {'en': 'Sentani'},
    'seu': {'en': 'Serui-Laut'},
    'sev': {'en': 'Nyarafolo Senoufo'},
    'sew': {'en': 'Sewa Bay'},
    'sey': {'en': 'Secoya'},
    'sez': {'en': 'Senthang Chin'},
    'sfb': {'en': 'Langue des signes de Belgique Francophone'},
    'sfe': {'en': 'Eastern Subanen'},
    'sfm': {'en': 'Small Flowery Miao'},
    'sfs': {'en': 'South African Sign Language'},
    'sfw': {'en': 'Sehwi'},
    'sg': {'af': 'Sango', 'am': 'ሳንጎ', 'ar': 'السانجو', 'as': 'ছাঙ্গো', 'ast': 'sango', 'az': 'sanqo', 'az-Cyrl': 'санго', 'az-Latn': 'sanqo', 'bal-Latn': 'Sangó', 'be': 'санга', 'be-tarask': 'санга', 'bg': 'санго', 'bgn': 'سانگۆیی', 'bn': 'সাঙ্গো', 'br': 'sango', 'brx': 'साङ’', 'bs': 'sango', 'bs-Cyrl': 'санго', 'bs-Latn': 'sango', 'ca': 'sango', 'ca-ES-valencia': 'sango', 'ccp': '𑄥𑄋𑄴𑄉𑄮', 'ce': 'санго', 'chr': 'ᏌᏂᎪ', 'ckb': 'سانگۆ', 'cs': 'sangština', 'cy': 'Sango', 'da': 'sango', 'de': 'Sango', 'dsb': 'sango', 'ee': 'sangogbe', 'el': 'Σάνγκο', 'el-polyton': 'Σάνγκο', 'en': 'Sango', 'eo': 'sangoa', 'es': 'sango', 'et': 'sango', 'eu': 'sangoa', 'fa': 'سانگو', 'ff-Adlm': '𞤅𞤢𞤲𞤺𞤮𞥅𞤪𞤫', 'fi': 'sango', 'fil': 'Sango', 'fo': 'sango', 'fr': 'sango', 'fur': 'sango', 'fy': 'Sango', 'ga': 'Sangóis', 'gd': 'Sango', 'gl': 'sango', 'gsw': 'Sango', 'gu': 'સાંગો', 'ha': 'Sango', 'he': 'סנגו', 'hi': 'सांगो', 'hi-Latn': 'Sango', 'hr': 'sango', 'hsb': 'sango', 'ht': 'sango', 'hu': 'szangó', 'hy': 'սանգո', 'ia': 'sango', 'id': 'Sango', 'ig': 'Sango', 'is': 'sangó', 'it': 'sango', 'ja': 'サンゴ語', 'jv': 'Sango', 'ka': 'სანგო', 'kgp': 'sỹgo', 'kk': 'санго тілі', 'kk-Cyrl': 'санго тілі', 'km': 'សានហ្គោ', 'kn': 'ಸಾಂಗೋ', 'ko': '산고어', 'kok': 'सांगो', 'kok-Deva': 'सांगो', 'ks': 'سَنگو', 'ks-Arab': 'سَنگو', 'ksh': 'Sangjo', 'ku': 'sangoyî', 'ky': 'сангочо', 'lb': 'Sango', 'lld': 'sango', 'lo': 'ແຊງໂກ', 'lrc': 'سانگو', 'lt': 'sango', 'lv': 'sango', 'mai': 'सांगो', 'mi': 'Hāngo', 'mk': 'санго', 'ml': 'സാംഗോ', 'mn': 'санго', 'mr': 'सांगो', 'ms': 'Sango', 'mt': 'Sango', 'my': 'ဆန်ဂို', 'mzn': 'سانگو', 'nb': 'sango', 'ne': 'साङ्गो', 'nl': 'Sango', 'nn': 'sango', 'no': 'sango', 'nqo': 'ߛߊ߲ߜߵߏߞߊ߲', 'or': 'ସାଙ୍ଗୋ', 'pa': 'ਸਾਂਗੋ', 'pa-Guru': 'ਸਾਂਗੋ', 'pcm': 'sàngo', 'pl': 'sango', 'ps': 'سانګو', 'pt': 'sango', 'qu': 'Sango Simi', 'rm': 'sango', 'ro': 'sango', 'ru': 'санго', 'sc': 'sango', 'sd': 'سانگو', 'sd-Arab': 'سانگو', 'sg': 'Sängö', 'si': 'සන්ග්\u200dරෝ', 'sk': 'sango', 'sl': 'sango', 'smn': 'sango', 'so': 'Sango', 'sq': 'sangoisht', 'sr': 'санго', 'sr-Cyrl': 'санго', 'sr-Latn': 'sango', 'sv': 'sango', 'sw': 'Kisango', 'ta': 'சாங்கோ', 'te': 'సాంగో', 'th': 'ซันโก', 'ti': 'ሳንጎ', 'tk': 'sango dili', 'to': 'lea fakasangikō', 'tr': 'Sango', 'ug': 'سانگوچە', 'uk': 'санго', 'ur': 'ساںغو', 'uz': 'sango', 'uz-Cyrl': 'санго', 'uz-Latn': 'sango', 'vec': 'sango', 'vi': 'Tiếng Sango', 'wae': 'Sango', 'xnr': 'सांगो', 'yo': 'Sango', 'yrl': 'sãgú', 'yue': '桑戈文', 'yue-Hans': '桑戈文', 'yue-Hant': '桑戈文', 'zh': '桑戈语', 'zh-Hans': '桑戈语', 'zh-Hant': '桑戈文', 'zu': 'isi-Sango'},
    'sga': {'ar': 'الأيرلندية القديمة', 'ast': 'irlandés antiguu', 'az': 'qədim irland', 'az-Latn': 'qədim irland', 'be': 'стараірландская', 'be-tarask': 'стараірландская', 'bg': 'староирландски', 'bn': 'প্রাচীন আইরিশ', 'br': 'heniwerzhoneg', 'brx': 'पुरानी आईरीश', 'bs': 'staroirski', 'bs-Cyrl': 'староирски', 'bs-Latn': 'staroirski', 'ca': 'irlandès antic', 'ca-ES-valencia': 'irlandès antic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄴 𑄃𑄭𑄢𑄨𑄌𑄴', 'cs': 'irština (stará)', 'cy': 'Hen Wyddeleg', 'da': 'oldirsk', 'de': 'Altirisch', 'el': 'Παλαιά Ιρλανδικά', 'el-polyton': 'Παλαιά Ιρλανδικά', 'en': 'Old Irish', 'es': 'irlandés antiguo', 'et': 'vanaiiri', 'fa': 'ایرلندی باستان', 'fi': 'muinaisiiri', 'fr': 'ancien irlandais', 'fur': 'vieri irlandês', 'fy': 'Aldyrsk', 'ga': 'Sean-Ghaeilge', 'gd': 'Seann-Ghaeilge', 'gsw': 'Altirisch', 'gu': 'જૂની આયરિશ', 'he': 'אירית עתיקה', 'hi': 'पुरानी आइरिश', 'hi-Latn': 'Old Irish', 'hr': 'staroirski', 'ht': 'ancien irlandais', 'hu': 'óír', 'hy': 'հին իռլանդերեն', 'id': 'Irlandia Kuno', 'is': 'fornírska', 'it': 'irlandese antico', 'ja': '古アイルランド語', 'ka': 'ძველი ირლანდიური', 'kgp': 'irỹnej arkajku', 'kn': 'ಪ್ರಾಚೀನ ಐರಿಷ್', 'ko': '고대 아일랜드어', 'ks': 'پرون ایرِش', 'ks-Arab': 'پرون ایرِش', 'lb': 'Aliresch', 'lo': 'ອີຣິຊເກົ່າ', 'lt': 'senoji airių', 'lv': 'senīru', 'mk': 'староирски', 'ml': 'പഴയ ഐറിഷ്', 'mr': 'पुरातन आयरिश', 'mt': 'Irlandiż Antik', 'my': 'အိုင်းရစ် ဟောင်း', 'nb': 'gammelirsk', 'ne': 'पुरातन आयरीस', 'nl': 'Oudiers', 'nn': 'gammalirsk', 'no': 'gammelirsk', 'or': 'ପୁରୁଣା ଇରିଶ୍', 'pl': 'staroirlandzki', 'pt': 'irlandês arcaico', 'rm': 'irlandais vegl', 'ro': 'irlandeză veche', 'ru': 'староирландский', 'sk': 'stará írčina', 'sl': 'stara irščina', 'sr': 'староирски', 'sr-Cyrl': 'староирски', 'sr-Latn': 'staroirski', 'sv': 'forniriska', 'ta': 'பழைய ஐரிஷ்', 'te': 'ప్రాచీన ఐరిష్', 'th': 'ไอริชโบราณ', 'to': 'lea fakaʻaelani-motuʻa', 'tr': 'Eski İrlandaca', 'ug': 'قەدىمكى ئىرېلاندچە', 'uk': 'давньоірландська', 'vi': 'Tiếng Ai-len cổ', 'yi': 'אַלט־איריש', 'yrl': 'irãdesiarkaiku', 'yue': '古愛爾蘭文', 'yue-Hans': '古爱尔兰文', 'yue-Hant': '古愛爾蘭文', 'zh': '古爱尔兰语', 'zh-Hans': '古爱尔兰语', 'zh-Hant': '古愛爾蘭文'},
    'sgb': {'en': 'Mag-antsi Ayta'},
    'sgc': {'en': 'Kipsigis'},
    'sgd': {'en': 'Surigaonon'},
    'sge': {'en': 'Segai'},
    'sgg': {'en': 'Swiss-German Sign Language'},
    'sgh': {'en': 'Shughni'},
    'sgi': {'en': 'Suga'},
    'sgj': {'en': 'Surgujia'},
    'sgk': {'en': 'Sangkong'},
    'sgl': {'en': 'Sanglechi-Ishkashimi'},
    'sgm': {'en': 'Singa'},
    'sgn': {'en': 'Sign languages'},
    'sgo': {'en': 'Songa'},
    'sgp': {'en': 'Singpho'},
    'sgr': {'en': 'Sangisari'},
    'sgs': {'ast': 'samogitianu', 'cs': 'žemaitština', 'cy': 'Samogiteg', 'de': 'Samogitisch', 'en': 'Samogitian', 'et': 'žemaidi', 'fi': 'samogiitti', 'fr': 'samogitien', 'hi-Latn': 'Samogitian', 'ht': 'samogitien', 'it': 'samogitico', 'ja': 'サモギティア語', 'lb': 'Samogitesch', 'lt': 'žemaičių', 'mk': 'самогитски', 'nb': 'samogitisk', 'nl': 'Samogitisch', 'nn': 'samogitisk', 'no': 'samogitisk', 'pl': 'żmudzki', 'sv': 'samogitiska', 'th': 'ซาโมจิเตียน', 'to': 'lea fakasamositia', 'tr': 'Samogitçe', 'yue': '薩莫吉希亞文', 'yue-Hans': '萨莫吉希亚文', 'yue-Hant': '薩莫吉希亞文', 'zh-Hant': '薩莫吉希亞文'},
    'sgt': {'en': 'Brokpake'},
    'sgu': {'en': 'Salas'},
    'sgw': {'en': 'Sebat Bet Gurage'},
    'sgx': {'en': 'Sierra Leone Sign Language'},
    'sgy': {'en': 'Sanglechi'},
    'sgz': {'en': 'Sursurunga'},
    'sh': {'af': 'Serwo-Kroaties', 'am': 'ሰርቦ-ክሮኤሽያኛ', 'ar': 'صربية-كرواتية', 'ast': 'serbo-croata', 'az': 'serb-xorvat', 'az-Latn': 'serb-xorvat', 'be': 'сербскахарвацкая', 'be-tarask': 'сербскахарвацкая', 'bg': 'сърбохърватски', 'bn': 'সার্বো-ক্রোয়েশিয়', 'br': 'serb-kroateg', 'brx': 'सर्बो-क्रोएशन्', 'bs': 'srpskohrvatski', 'bs-Cyrl': 'српскохрватски', 'bs-Latn': 'srpskohrvatski', 'ca': 'serbocroat', 'ca-ES-valencia': 'serbocroat', 'ccp': '𑄥𑄢𑄴𑄝𑄮-𑄇𑄳𑄢𑄮𑄠𑄬𑄥𑄨𑄠𑄧', 'cs': 'srbochorvatština', 'cy': 'Serbo-Croateg', 'da': 'serbokroatisk', 'de': 'Serbo-Kroatisch', 'dsb': 'serbochorwatšćina', 'ee': 'serbo-croatiagbe', 'el': 'Σερβοκροατικά', 'el-polyton': 'Σερβοκροατικά', 'en': 'Serbo-Croatian', 'eo': 'serbo-Kroata', 'es': 'serbocroata', 'et': 'serbia-horvaadi', 'eu': 'serbokroaziera', 'fa': 'صرب و کرواتی', 'fi': 'serbokroaatti', 'fil': 'Serbo-Croatian', 'fo': 'serbokroatiskt', 'fr': 'serbo-croate', 'fy': 'Servokroatysk', 'ga': 'Seirbea-Chróitis', 'gd': 'Sèirb-Chròthaisis', 'gl': 'serbocroata', 'gsw': 'Serbo-Kroatisch', 'gu': 'સર્બો-ક્રોએશિયન', 'ha': 'Kuroweshiyancin-Sabiya', 'he': 'סרבו-קרואטית', 'hi': 'सेर्बो-क्रोएशियाई', 'hi-Latn': 'Serbo-Croatian', 'hr': 'srpsko-hrvatski', 'hsb': 'serbochorwatšćina', 'ht': 'serbo-croate', 'hu': 'szerbhorvát', 'hy': 'սերբա-խորվաթերեն', 'ia': 'serbocroato', 'id': 'Serbo-Kroasia', 'is': 'serbókróatíska', 'it': 'serbo-croato', 'ja': 'セルボ・クロアチア語', 'ka': 'სერბულ-ხორვატული', 'kgp': 'servo-krovata', 'kk': 'серб-хорват тілі', 'kk-Cyrl': 'серб-хорват тілі', 'km': 'សឺបូក្រូអាត', 'kn': 'ಸರ್ಬೋ-ಕ್ರೊಯೇಶಿಯನ್', 'ko': '세르비아-크로아티아어', 'kok': 'सेर्बो-क्रोयेषियन्', 'kok-Deva': 'सेर्बो-क्रोयेषियन्', 'ks': 'سیربو کروشِیَن', 'ks-Arab': 'سیربو کروشِیَن', 'ksh': 'Särbokowatesch', 'ky': 'серб-хорват', 'lb': 'Serbo-Kroatesch', 'lo': 'ເຊີໂບ-ໂກເຊຍ', 'lt': 'serbų-kroatų', 'lv': 'serbu–horvātu', 'mk': 'српскохрватски', 'ml': 'സെർബോ-ക്രൊയേഷ്യൻ', 'mn': 'хорватын серб', 'mr': 'सर्बो-क्रोएशियन', 'ms': 'SerboCroatia', 'mt': 'Serbo-Kroat', 'nb': 'serbokroatisk', 'nl': 'Servo-Kroatisch', 'nn': 'serbokroatisk', 'no': 'serbokroatisk', 'or': 'ସର୍ବୋ-କ୍ରୋଆଟିଆନ୍', 'pl': 'serbsko-chorwacki', 'pt': 'servo-croata', 'rm': 'serbo-croat', 'ro': 'sârbo-croată', 'ru': 'сербскохорватский', 'rw': 'Inyeseribiya na Korowasiya', 'se': 'serbokroatiagiella', 'sk': 'srbochorvátčina', 'sl': 'srbohrvaščina', 'sq': 'serbo-kroatisht', 'sr': 'српскохрватски', 'sr-Cyrl': 'српскохрватски', 'sr-Latn': 'srpskohrvatski', 'sv': 'serbokroatiska', 'sw': 'Kiserbia-kroeshia', 'ta': 'செர்போ-குரோஷியன்', 'te': 'సేర్బో-క్రొయేషియన్', 'th': 'เซอร์โบ-โครเอเชีย', 'ti': 'ሰርቦ-ክሮኤሽያኛ', 'to': 'lea fakakuloisia-sēpia', 'tr': 'Sırp-Hırvat Dili', 'ug': 'سېرب-كرودىيەچە', 'uk': 'сербсько-хорватська', 'ur': 'سربو-کروئیشین', 'vi': 'Tiếng Serbo-Croatia', 'yi': 'סערבא־קראאַטיש', 'yo': 'Èdè Serbo-Croatiani', 'yrl': 'seriwu-kruata', 'yue': '塞爾維亞克羅埃西亞文', 'yue-Hans': '塞尔维亚克罗埃西亚文', 'yue-Hant': '塞爾維亞克羅埃西亞文', 'zh': '塞尔维亚-克罗地亚语', 'zh-Hans': '塞尔维亚-克罗地亚语', 'zh-Hant': '塞爾維亞克羅埃西亞文', 'zu': 'isi-Serbo-Croatian', 'sh': 'Srpskohrvatski'},
    'sha': {'en': 'Shall-Zwall'},
    'shb': {'en': 'Ninam'},
    'shc': {'en': 'Sonde'},
    'shd': {'en': 'Kundal Shahi'},
    'she': {'en': 'Sheko'},
    'shg': {'en': 'Shua'},
    'shh': {'en': 'Shoshoni'},
    'shi': {'af': 'Tachelhit', 'am': 'ታቼልሂት', 'ar': 'تشلحيت', 'as': 'তাচেলহিট', 'ast': 'tachelhit', 'az': 'taçelit', 'az-Cyrl': 'тачелит', 'az-Latn': 'taçelit', 'bal-Latn': 'Tachelhit', 'be': 'ташэльхіт', 'be-tarask': 'ташэльхіт', 'bg': 'ташелхит', 'bgn': 'تاچل\u200cهیتی', 'bn': 'তাচেলহিত', 'br': 'tacheliteg', 'brx': 'ताचेल्हीट', 'bs': 'tahelhit', 'bs-Cyrl': 'ташелхит', 'bs-Latn': 'tahelhit', 'ca': 'taixelhit', 'ca-ES-valencia': 'taixelhit', 'ccp': '𑄖𑄌𑄬𑄣𑄴𑄦𑄨𑄖𑄴', 'ce': 'тахелхит', 'chr': 'ᏔᏤᎵᎯᏘ', 'ckb': 'شیلها', 'cs': 'tašelhit', 'cy': 'Tachelhit', 'da': 'tachelhit', 'de': 'Taschelhit', 'dsb': 'tašelhit', 'el': 'Τασελχίτ', 'el-polyton': 'Τασελχίτ', 'en': 'Tachelhit', 'es': 'tashelhit', 'et': 'šilha', 'eu': 'tachelhita', 'fa': 'تاچل\u200cهیت', 'ff-Adlm': '𞤚𞤢𞤧𞤭𞤤𞤸𞤭𞤼𞤪𞤫', 'fi': 'tašelhit', 'fil': 'Tachelhit', 'fo': 'tachelhit', 'fr': 'chleuh', 'fy': 'Tashelhiyt', 'ga': 'Tachelhit', 'gd': 'Tachelhit', 'gl': 'tachelhit', 'gu': 'તેશીલહિટ', 'ha': 'Tachelhit', 'he': 'שילה', 'hi': 'तैचेल्हित', 'hi-Latn': 'Tachelhit', 'hr': 'tachelhit', 'hsb': 'tašelhit', 'ht': 'chleuh', 'hu': 'tachelhit', 'hy': 'տաշելհիթ', 'ia': 'tachelhit', 'id': 'Tachelhit', 'ig': 'Tachịkịt', 'is': 'tachelhit', 'it': 'tashelhit', 'ja': 'タシルハイト語', 'jv': 'Tachelhit', 'ka': 'შილჰა', 'kgp': 'tasehiti', 'kk': 'ташелхит тілі', 'kk-Cyrl': 'ташелхит тілі', 'km': 'តាឈីលហ៊ីត', 'kn': 'ಟಷೆಲ್\u200dಹಿಟ್', 'ko': '타셸히트어', 'kok': 'ताछेल्हीट', 'kok-Deva': 'ताछेल्हीट', 'ksh': 'Taschelhittesch', 'ku': 'taşelhîtî', 'ky': 'ташелитче', 'lb': 'Taschelhit', 'lo': 'ທາເຊວຫິດ', 'lrc': 'تاچئلھیت', 'lt': 'tachelhitų', 'lv': 'šilhu', 'mai': 'टेचलहिट', 'mi': 'Tāhehita', 'mk': 'тачелхит', 'ml': 'താച്ചലിറ്റ്', 'mn': 'тачелхит', 'mr': 'ताशेल्हिट', 'ms': 'Tachelhit', 'mt': 'Tachelhit', 'my': 'တာချယ်လ်ဟစ်', 'mzn': 'تاچلهیت', 'nb': 'tachelhit', 'ne': 'टाचेल्हिट', 'nl': 'Tashelhiyt', 'nn': 'tachelhit', 'no': 'tachelhit', 'nqo': 'ߗߑߟߋߥߎߞߊ߲', 'or': 'ତାଚେଲହିଟ୍', 'pa': 'ਟਚੇਲਹਿਟ', 'pa-Guru': 'ਟਚੇਲਹਿਟ', 'pcm': 'Táchẹ́lit', 'pl': 'tashelhiyt', 'ps': 'تاکلهیټ', 'pt': 'tachelhit', 'qu': 'Tashelhit Simi', 'rm': 'tachelit', 'ro': 'tachelhit', 'ru': 'ташельхит', 'sc': 'tashelhit', 'sd': 'تيچل هاتي', 'sd-Arab': 'تيچل هاتي', 'shi': 'ⵜⴰⵛⵍⵃⵉⵜ', 'shi-Latn': 'Tashelḥiyt', 'shi-Tfng': 'ⵜⴰⵛⵍⵃⵉⵜ', 'si': 'ටචේල්හිට්', 'sk': 'tachelhit', 'sl': 'tahelitska berberščina', 'smn': 'tašelhit', 'so': 'Shilha', 'sq': 'taçelitisht', 'sr': 'ташелхит', 'sr-Cyrl': 'ташелхит', 'sr-Latn': 'tašelhit', 'sv': 'tachelhit', 'sw': 'Kitachelhit', 'ta': 'தசேஹித்', 'te': 'టాచెల్\u200cహిట్', 'th': 'ทาเชลีห์ท', 'ti': 'ታቸልሂት', 'tk': 'tahelhit dili', 'to': 'lea fakataselihiti', 'tr': 'Taşelit', 'ug': 'شىلخاچە', 'uk': 'тачеліт', 'ur': 'تشلحيت', 'uz': 'tashelxit', 'uz-Cyrl': 'ташелхит', 'uz-Latn': 'tashelxit', 'vec': 'tashelit', 'vi': 'Tiếng Tachelhit', 'xnr': 'तैचेल्हित', 'yo': 'Taṣelíìtì', 'yrl': 'taxeriti', 'yue': '希爾哈文', 'yue-Hans': '希尔哈文', 'yue-Hant': '希爾哈文', 'zh': '希尔哈语', 'zh-Hans': '希尔哈语', 'zh-Hant': '希爾哈文', 'zu': 'isi-Tachelhit'},
    'shj': {'en': 'Shatt'},
    'shk': {'en': 'Shilluk'},
    'shl': {'en': 'Shendu'},
    'shm': {'en': 'Shahrudi'},
    'shn': {'af': 'Shan', 'am': 'ሻን', 'ar': 'الشان', 'as': 'চান', 'ast': 'shan', 'az': 'şan', 'az-Cyrl': 'шан', 'az-Latn': 'şan', 'bal-Latn': 'Shan', 'be': 'шан', 'be-tarask': 'шан', 'bg': 'шан', 'bgn': 'شانی', 'bn': 'শান', 'br': 'shan', 'brx': 'शान', 'bs': 'šan', 'bs-Cyrl': 'шан', 'bs-Latn': 'šan', 'ca': 'xan', 'ca-ES-valencia': 'xan', 'ccp': '𑄥𑄚𑄴', 'ce': 'шанойн', 'chr': 'ᏝᏂ', 'ckb': 'شان', 'cs': 'šanština', 'cy': 'Shan', 'da': 'shan', 'de': 'Schan', 'dsb': 'shanšćina', 'dz': 'ཤཱན་ཁ', 'el': 'Σαν', 'el-polyton': 'Σαν', 'en': 'Shan', 'es': 'shan', 'et': 'šani', 'eu': 'shanera', 'fa': 'شانی', 'ff-Adlm': '𞤅𞤢𞤲𞤪𞤫', 'fi': 'shan', 'fil': 'Shan', 'fo': 'shan', 'fr': 'shan', 'fy': 'Shan', 'ga': 'Seánais', 'gd': 'Shan', 'gl': 'shan', 'gsw': 'Schan-Schpraach', 'gu': 'શેન', 'ha': 'Shan', 'he': 'שאן', 'hi': 'शैन', 'hi-Latn': 'Shan', 'hr': 'shan', 'hsb': 'shanšćina', 'ht': 'shan', 'hu': 'san', 'hy': 'շաներեն', 'ia': 'shan', 'id': 'Shan', 'ig': 'Shan', 'is': 'sjan', 'it': 'shan', 'ja': 'シャン語', 'jv': 'Shan', 'ka': 'შანი', 'kgp': 'sỹn', 'kk': 'шан тілі', 'kk-Cyrl': 'шан тілі', 'km': 'សាន', 'kn': 'ಶಾನ್', 'ko': '샨어', 'kok': 'शान', 'kok-Deva': 'शान', 'ks': 'شان', 'ks-Arab': 'شان', 'ku': 'şanî', 'ky': 'шанча', 'lb': 'Schan-Sprooch', 'lo': 'ຊານ', 'lt': 'šan', 'lv': 'šanu', 'mai': 'शान', 'mi': 'Hāna', 'mk': 'шан', 'ml': 'ഷാൻ', 'mn': 'шань', 'mr': 'शॅन', 'ms': 'Shan', 'mt': 'Shan', 'my': 'ရှမ်း', 'nb': 'shan', 'ne': 'शान', 'nl': 'Shan', 'nn': 'shan', 'no': 'shan', 'or': 'ଶାନ୍', 'pa': 'ਸ਼ਾਨ', 'pa-Guru': 'ਸ਼ਾਨ', 'pcm': 'Shán Lángwej', 'pl': 'szan', 'ps': 'شان', 'pt': 'shan', 'qu': 'Shan Simi', 'rm': 'shan', 'ro': 'shan', 'ru': 'шанский', 'sc': 'shan', 'sd': 'شان', 'sd-Arab': 'شان', 'shn': 'တႆး', 'si': 'ශාන්', 'sk': 'šančina', 'sl': 'šanščina', 'smn': 'shankielâ', 'so': 'Shan', 'sq': 'shanisht', 'sr': 'шански', 'sr-Cyrl': 'шански', 'sr-Latn': 'šanski', 'sv': 'shan', 'sw': 'Kishan', 'ta': 'ஷான்', 'te': 'షాన్', 'th': 'ไทใหญ่', 'ti': 'ሻን', 'tk': 'şan dili', 'to': 'lea fakasiani', 'tr': 'Shan dili', 'ug': 'شانچە', 'uk': 'шанська', 'ur': 'شان', 'uz': 'shan', 'uz-Latn': 'shan', 'vec': 'shan', 'vi': 'Tiếng Shan', 'xnr': 'शैन', 'yo': 'Èdè Shani', 'yrl': 'xãni', 'yue': '撣文', 'yue-Hans': '掸文', 'yue-Hant': '撣文', 'zh': '掸语', 'zh-Hans': '掸语', 'zh-Hant': '撣文', 'zu': 'isi-Shan'},
    'sho': {'en': 'Shanga'},
    'shp': {'en': 'Shipibo-Conibo'},
    'shq': {'en': 'Sala'},
    'shr': {'en': 'Shi'},
    'shs': {'en': 'Shuswap'},
    'sht': {'en': 'Shasta'},
    'shu': {'am': 'ቻዲያን ዓረብኛ', 'ar': 'العربية التشادية', 'ast': 'árabe chadianu', 'br': 'arabeg Tchad', 'bs': 'čadski arapski', 'bs-Latn': 'čadski arapski', 'ca': 'àrab txadià', 'ca-ES-valencia': 'àrab txadià', 'cs': 'arabština (čadská)', 'cy': 'Arabeg Chad', 'da': 'tchadisk arabisk', 'de': 'Tschadisch-Arabisch', 'el': 'Αραβικά του Τσαντ', 'el-polyton': 'Αραβικά του Τσαντ', 'en': 'Chadian Arabic', 'es': 'árabe chadiano', 'et': 'Tšaadi araabia', 'fa': 'عربی چادی', 'fi': 'tšadinarabia', 'fr': 'arabe tchadien', 'fy': 'Tsjadysk Arabysk', 'gd': 'Arabais Seàdach', 'he': 'ערבית צ׳אדית', 'hi-Latn': 'Chadian Arabic', 'hr': 'čadski arapski', 'ht': 'arabe tchadien', 'hu': 'csádi arab', 'id': 'Arab Suwa', 'it': 'arabo ciadiano', 'ja': 'チャド・アラビア語', 'ka': 'ჩადური არაბული', 'kgp': 'arame sanijỹnũ', 'ko': '차디언 아라비아어', 'lb': 'Tschadesch-Arabesch', 'lo': 'ອາລັບ-ຊາດ', 'lt': 'chadian arabų', 'lv': 'Čadas arābu', 'mk': 'чадски арапски', 'ml': 'ചാഡിയൻ അറബി', 'ms': 'Arab Chadian', 'nb': 'tsjadisk arabisk', 'ne': 'चाड अरबी', 'nl': 'Tsjadisch Arabisch', 'nn': 'tsjadisk arabisk', 'no': 'tsjadisk arabisk', 'pl': 'arabski (Czad)', 'pt': 'árabe chadiano', 'ro': 'arabă ciadiană', 'ru': 'чадский арабский', 'sk': 'čadská arabčina', 'sv': 'Tchad-arabiska', 'sw': 'Kiarabu cha Chad', 'th': 'อาหรับ-ชาด', 'to': 'lea fakaʻalepea-sāti', 'tr': 'Çad Arapçası', 'ug': 'چاد ئەرەبچە', 'uk': 'чадійська арабська', 'vi': 'Tiếng Ả-Rập Chad', 'yrl': 'arabi xadianu', 'yue': '阿拉伯文（查德）', 'yue-Hans': '阿拉伯文（查德）', 'yue-Hant': '阿拉伯文（查德）', 'zh': '乍得阿拉伯语', 'zh-Hans': '乍得阿拉伯语', 'zh-Hant': '阿拉伯文（查德）'},
    'shv': {'en': 'Shehri'},
    'shw': {'en': 'Shwai'},
    'shx': {'en': 'She'},
    'shy': {'en': 'Tachawit'},
    'shz': {'en': 'Syenara Senoufo'},
    'si': {'af': 'Sinhala', 'ak': 'Sinhala', 'am': 'ሲንሃላ', 'ar': 'السنهالية', 'as': 'সিংহলা', 'ast': 'cingalés', 'az': 'sinhala', 'az-Cyrl': 'синһала', 'az-Latn': 'sinhala', 'bal-Latn': 'Senhálá', 'be': 'сінгальская', 'be-tarask': 'сінгальская', 'bg': 'синхалски', 'bgn': 'سینهالی', 'bn': 'সিংহলী', 'br': 'singhaleg', 'brx': 'सींहाला', 'bs': 'sinhaleški', 'bs-Cyrl': 'синхалски', 'bs-Latn': 'sinhaleški', 'ca': 'singalès', 'ca-ES-valencia': 'singalès', 'ccp': '𑄥𑄨𑄁𑄦𑄧𑄣𑄩', 'ce': 'сингалхойн', 'chr': 'ᏏᎾᎭᎳ', 'ckb': 'سینهالی', 'cs': 'sinhálština', 'cy': 'Sinhaleg', 'da': 'singalesisk', 'de': 'Singhalesisch', 'dsb': 'singalšćina', 'dz': 'སིང་ཧ་ལ་ཁ', 'ee': 'sinhalgbe', 'el': 'Σινχαλεζικά', 'el-polyton': 'Σινχαλεζικά', 'en': 'Sinhala', 'eo': 'sinhala', 'es': 'cingalés', 'et': 'singali', 'eu': 'sinhala', 'fa': 'سینهالی', 'ff-Adlm': '𞤅𞤭𞤲𞤸𞤢𞤤𞤢𞥄𞤪𞤫', 'fi': 'sinhala', 'fil': 'Sinhala', 'fo': 'singalesiskt', 'fr': 'cingalais', 'fur': 'sinalês', 'fy': 'Singalees', 'ga': 'Siolóinis', 'gd': 'Sinhala', 'gl': 'cingalés', 'gsw': 'Singhalesisch', 'gu': 'સિંહાલી', 'ha': 'Sinhalanci', 'he': 'סינהלה', 'hi': 'सिंहली', 'hi-Latn': 'Sinhala', 'hr': 'sinhaleški', 'hsb': 'singhalšćina', 'ht': 'cingalais', 'hu': 'szingaléz', 'hy': 'սինհալերեն', 'ia': 'cingalese', 'id': 'Sinhala', 'ig': 'Sinhala', 'is': 'singalíska', 'it': 'singalese', 'ja': 'シンハラ語', 'jv': 'Sinhala', 'ka': 'სინჰალური', 'kea': 'singales', 'kgp': 'sĩgarej', 'kk': 'сингал тілі', 'kk-Cyrl': 'сингал тілі', 'km': 'ស្រីលង្កា', 'kn': 'ಸಿಂಹಳ', 'ko': '싱할라어', 'kok': 'सिंहला', 'kok-Deva': 'सिंहला', 'ks': 'سِنہالا', 'ks-Arab': 'سِنہالا', 'ksh': 'Singjaleesesch', 'ku': 'kîngalî', 'kxv': 'sinhali', 'kxv-Deva': 'सिंहली', 'kxv-Latn': 'sinhali', 'kxv-Orya': 'ସିଂହଲା', 'kxv-Telu': 'సింహళం', 'ky': 'сингалача', 'lb': 'Singhalesesch', 'lkt': 'Sinhala Iyápi', 'lld': 'singalesc', 'lo': 'ສິນຫາລາ', 'lrc': 'سینھالا', 'lt': 'sinhalų', 'lv': 'singāļu', 'mai': 'सिंहला', 'mi': 'Hinihāra', 'mk': 'синхалски', 'ml': 'സിംഹള', 'mn': 'синхала', 'mr': 'सिंहला', 'ms': 'Sinhala', 'mt': 'Sinhala', 'my': 'စင်ဟာလာ', 'mzn': 'سینهالا', 'nb': 'singalesisk', 'ne': 'सिन्हाली', 'nl': 'Singalees', 'nn': 'singalesisk', 'no': 'singalesisk', 'nqo': 'ߛߌ߲ߜ߭ߟߊߞߊ߲', 'om': 'Afaan Sinhalese', 'or': 'ସିଂହଳ', 'pa': 'ਸਿੰਹਾਲਾ', 'pa-Guru': 'ਸਿੰਹਾਲਾ', 'pcm': 'Sínhala', 'pl': 'syngaleski', 'ps': 'سينهالي', 'pt': 'cingalês', 'qu': 'Cingales Simi', 'rm': 'singalais', 'ro': 'singhaleză', 'ru': 'сингальский', 'rw': 'Inyesimpaleze', 'sc': 'singalesu', 'sd': 'سنهالا', 'sd-Arab': 'سنهالا', 'si': 'සිංහල', 'sk': 'sinhalčina', 'sl': 'sinhalščina', 'smn': 'sinhala', 'so': 'Sinhala', 'sq': 'sinhalisht', 'sr': 'синхалешки', 'sr-Cyrl': 'синхалешки', 'sr-Latn': 'sinhaleški', 'sv': 'singalesiska', 'sw': 'Kisinhala', 'ta': 'சிங்களம்', 'te': 'సింహళం', 'tg': 'сингалӣ', 'th': 'สิงหล', 'ti': 'ሲንሃላ', 'tk': 'singal dili', 'to': 'lea fakasingihala', 'tr': 'Sinhali dili', 'tt': 'сингал', 'ug': 'سىنگالچە', 'uk': 'сингальська', 'ur': 'سنہالا', 'uz': 'singal', 'uz-Cyrl': 'сингалча', 'uz-Latn': 'singal', 'vec': 'singaleze', 'vi': 'Tiếng Sinhala', 'wae': 'Singalesiš', 'wo': 'Sinala', 'xnr': 'सिंहली', 'yi': 'סינהאַליש', 'yo': 'Èdè Sinhalese', 'yrl': 'sĩgarei', 'yue': '僧伽羅文', 'yue-Hans': '僧伽罗文', 'yue-Hant': '僧伽羅文', 'zh': '僧伽罗语', 'zh-Hans': '僧伽罗语', 'zh-Hant': '僧伽羅文', 'zu': 'isi-Sinhala'},
    'sia': {'en': 'Akkala Sami'},
    'sib': {'en': 'Sebop'},
    'sid': {'am': 'ሲዳምኛ', 'ar': 'السيدامو', 'ast': 'sidamo', 'az': 'sidamo', 'az-Latn': 'sidamo', 'bal-Latn': 'Sidámó', 'bg': 'сидамо', 'bn': 'সিডামো', 'br': 'sidamo', 'brx': 'सीदामो', 'bs': 'sidamo', 'bs-Cyrl': 'сидамо', 'bs-Latn': 'sidamo', 'ca': 'sidamo', 'ca-ES-valencia': 'sidamo', 'ccp': '𑄥𑄨𑄓𑄟𑄮', 'cs': 'sidamo', 'cy': 'Sidamo', 'da': 'sidamo', 'de': 'Sidamo', 'el': 'Σιντάμο', 'el-polyton': 'Σιντάμο', 'en': 'Sidamo', 'es': 'sidamo', 'et': 'sidamo', 'fa': 'سیدامویی', 'fi': 'sidamo', 'fr': 'sidamo', 'fy': 'Sidamo', 'gd': 'Sidamo', 'gsw': 'Sidamo', 'gu': 'સિદામો', 'he': 'סידאמו', 'hi': 'सिदामो', 'hi-Latn': 'Sidamo', 'hr': 'sidamo', 'ht': 'sidamo', 'hu': 'szidamó', 'ia': 'sidamo', 'id': 'Sidamo', 'ig': 'Sidamo', 'is': 'sídamó', 'it': 'sidamo', 'ja': 'シダモ語', 'kgp': 'sinamũ', 'kn': 'ಸಿಡಾಮೋ', 'ko': '시다모어', 'ks': 'سِدامو', 'ks-Arab': 'سِدامو', 'ku': 'sîdamo', 'lb': 'Sidamo', 'lo': 'ຊິດາໂມ', 'lt': 'sidamų', 'lv': 'sidamu', 'mk': 'сидамо', 'ml': 'സിഡാമോ', 'mr': 'सिदामो', 'mt': 'Sidamo', 'nb': 'sidamo', 'nl': 'Sidamo', 'nn': 'sidamo', 'no': 'sidamo', 'or': 'ସିଦାମୋ', 'pl': 'sidamo', 'pt': 'sidamo', 'rm': 'sidamo', 'ro': 'sidamo', 'ru': 'сидама', 'sk': 'sidamo', 'sl': 'sidamščina', 'sr': 'сидамо', 'sr-Cyrl': 'сидамо', 'sr-Latn': 'sidamo', 'sv': 'sidamo', 'ta': 'சிடாமோ', 'te': 'సిడామో', 'th': 'ซิดาโม', 'ti': 'ሲዳመኛ', 'to': 'lea fakasitamo', 'tr': 'Sidamo dili', 'ug': 'سىداموچە', 'uk': 'сідамо', 'vec': 'sidamo', 'vi': 'Tiếng Sidamo', 'yrl': 'sidamu', 'yue': '希達摩文', 'yue-Hans': '希达摩文', 'yue-Hant': '希達摩文', 'zh': '悉达摩语', 'zh-Hans': '悉达摩语', 'zh-Hant': '希達摩文'},
    'sie': {'en': 'Simaa'},
    'sif': {'en': 'Siamou'},
    'sig': {'en': 'Paasaal'},
    'sih': {'en': 'Zire'},
    'sii': {'en': 'Shom Peng'},
    'sij': {'en': 'Numbami'},
    'sik': {'en': 'Sikiana'},
    'sil': {'en': 'Tumulung Sisaala'},
    'sim': {'en': 'Mende (Papua New Guinea)'},
    'sio': {'en': 'Siouan languages'},
    'sio-pro': {'en': 'Proto-Siouan'},
    'sip': {'en': 'Sikkimese'},
    'siq': {'en': 'Sonia'},
    'sir': {'en': 'Siri'},
    'sis': {'en': 'Siuslaw'},
    'sit': {'en': 'Sino-Tibetan languages'},
    'sit-pro': {'en': 'Proto-Sino-Tibetan'},
    'siu': {'en': 'Sinagen'},
    'siv': {'en': 'Sumariup'},
    'siw': {'en': 'Siwai'},
    'six': {'en': 'Sumau'},
    'siy': {'en': 'Sivandi'},
    'siz': {'en': 'Siwi'},
    'sja': {'en': 'Epena'},
    'sjb': {'en': 'Sajau Basap'},
    'sjd': {'en': 'Kildin Sami'},
    'sje': {'en': 'Pite Sami'},
    'sjg': {'en': 'Assangori'},
    'sjk': {'en': 'Kemi Sami'},
    'sjl': {'en': 'Sajalong'},
    'sjm': {'en': 'Mapun'},
    'sjn': {'en': 'Sindarin'},
    'sjo': {'en': 'Xibe'},
    'sjp': {'en': 'Surjapuri'},
    'sjr': {'en': 'Siar-Lak'},
    'sjs': {'en': 'Senhaja De Srair'},
    'sjt': {'en': 'Ter Sami'},
    'sju': {'en': 'Ume Sami'},
    'sjw': {'en': 'Shawnee'},
    'sk': {'af': 'Slowaaks', 'ak': 'Slovak Kasa', 'am': 'ስሎቫክኛ', 'ar': 'السلوفاكية', 'as': 'শ্লোভাক', 'ast': 'eslovacu', 'az': 'slovak', 'az-Arab': 'اسلواکجا', 'az-Cyrl': 'словак', 'az-Latn': 'slovak', 'bal-Latn': 'Solwák', 'be': 'славацкая', 'be-tarask': 'славацкая', 'bg': 'словашки', 'bgn': 'اسلواکی', 'bn': 'স্লোভাক', 'br': 'slovakeg', 'brx': 'स्ल’भाक', 'bs': 'slovački', 'bs-Cyrl': 'словачки', 'bs-Latn': 'slovački', 'ca': 'eslovac', 'ca-ES-valencia': 'eslovac', 'ccp': '𑄥𑄳𑄣𑄮𑄞𑄇𑄴', 'ce': 'словакийн', 'chr': 'ᏍᎶᏩᎩ', 'ckb': 'سلۆڤاكی', 'cs': 'slovenština', 'cy': 'Slofaceg', 'da': 'slovakisk', 'de': 'Slowakisch', 'dsb': 'słowakšćina', 'dz': 'སུ་ལོ་བཱཀ་ཁ', 'ee': 'slovakiagbe', 'el': 'Σλοβακικά', 'el-polyton': 'Σλοβακικά', 'en': 'Slovak', 'eo': 'slovaka', 'es': 'eslovaco', 'et': 'slovaki', 'eu': 'eslovakiera', 'fa': 'اسلواکی', 'ff-Adlm': '𞤅𞤤𞤮𞤾𞤢𞥄𞤳𞤪𞤫', 'fi': 'slovakki', 'fil': 'Slovak', 'fo': 'slovakiskt', 'fr': 'slovaque', 'fur': 'slovac', 'fy': 'Slowaaks', 'ga': 'Slóvaicis', 'gd': 'Slòbhacais', 'gl': 'eslovaco', 'gsw': 'Slowakisch', 'gu': 'સ્લોવૅક', 'ha': 'Basulke', 'he': 'סלובקית', 'hi': 'स्लोवाक', 'hi-Latn': 'Slovak', 'hr': 'slovački', 'hsb': 'słowakšćina', 'ht': 'slovaque', 'hu': 'szlovák', 'hy': 'սլովակերեն', 'ia': 'slovaco', 'id': 'Slovak', 'ie': 'slovac', 'ig': 'Asụsụ Slovak', 'is': 'slóvakíska', 'it': 'slovacco', 'ja': 'スロバキア語', 'jv': 'Slowakia', 'ka': 'სლოვაკური', 'kea': 'slovaku', 'kgp': 'erovako', 'kk': 'словак тілі', 'kk-Cyrl': 'словак тілі', 'km': 'ស្លូវ៉ាគី', 'kn': 'ಸ್ಲೋವಾಕ್', 'ko': '슬로바키아어', 'kok': 'स्लोवाक', 'kok-Deva': 'स्लोवाक', 'ks': 'سلووَک', 'ks-Arab': 'سلووَک', 'ksh': 'ẞlovakesch', 'ku': 'slovakî', 'kxv': 'slovak', 'kxv-Deva': 'स्लोवाक', 'kxv-Latn': 'slovak', 'kxv-Orya': 'ସ୍ଲବାକ୍', 'kxv-Telu': 'స్లోవక్', 'ky': 'словакча', 'lb': 'Slowakesch', 'lkt': 'Slovak Iyápi', 'lld': 'slovach', 'lo': 'ສະໂລແວັກ', 'lrc': 'ئسلوڤاکی', 'lt': 'slovakų', 'lv': 'slovāku', 'mai': 'स्लोबाक', 'mi': 'Horowākia', 'mk': 'словачки', 'ml': 'സ്ലോവാക്', 'mn': 'словак', 'mr': 'स्लोव्हाक', 'ms': 'Slovak', 'mt': 'Slovakk', 'my': 'ဆလိုဗက်', 'mzn': 'اسلواکی', 'nb': 'slovakisk', 'ne': 'स्लोभाकियाली', 'nl': 'Slowaaks', 'nn': 'slovakisk', 'no': 'slovakisk', 'nqo': 'ߛߑߟߏߝ߭ߊߞߌߞߊ߲', 'om': 'Afaan Slovak', 'or': 'ସ୍ଲୋଭାକ୍', 'pa': 'ਸਲੋਵਾਕ', 'pa-Guru': 'ਸਲੋਵਾਕ', 'pcm': 'Slóvak', 'pl': 'słowacki', 'ps': 'سلوواکي', 'pt': 'eslovaco', 'qu': 'Eslovaco Simi', 'rm': 'slovac', 'ro': 'slovacă', 'ru': 'словацкий', 'rw': 'Igisilovaki', 'sah': 'Словаактыы', 'sc': 'islovacu', 'sd': 'سلواڪي', 'sd-Arab': 'سلواڪي', 'se': 'slovákiagiella', 'si': 'ස්ලෝවැක්', 'sk': 'slovenčina', 'sl': 'slovaščina', 'smn': 'slovakiakielâ', 'so': 'Isloofaak', 'sq': 'sllovakisht', 'sr': 'словачки', 'sr-Cyrl': 'словачки', 'sr-Latn': 'slovački', 'sv': 'slovakiska', 'sw': 'Kislovakia', 'ta': 'ஸ்லோவாக்', 'te': 'స్లోవక్', 'tg': 'словакӣ', 'th': 'สโลวัก', 'ti': 'ስሎቫክኛ', 'tk': 'slowak dili', 'to': 'lea fakasolāvaki', 'tr': 'Slovakça', 'tt': 'словак', 'ug': 'سىلوۋاكچە', 'uk': 'словацька', 'ur': 'سلوواک', 'uz': 'slovakcha', 'uz-Cyrl': 'словакча', 'uz-Latn': 'slovakcha', 'vec': 'zlovaco', 'vi': 'Tiếng Slovak', 'wae': 'Slowakiš', 'wo': 'Eslowaki (Eslowak)', 'xnr': 'स्लोवाक', 'yi': 'סלאוואַקיש', 'yo': 'Èdè Slovaki', 'yrl': 'esiruwaku', 'yue': '斯洛伐克文', 'yue-Hans': '斯洛伐克文', 'yue-Hant': '斯洛伐克文', 'zh': '斯洛伐克语', 'zh-Hans': '斯洛伐克语', 'zh-Hant': '斯洛伐克文', 'zu': 'isi-Slovak'},
    'ska': {'en': 'Skagit'},
    'skb': {'en': 'Saek'},
    'skc': {'en': 'Ma Manda'},
    'skd': {'en': 'Southern Sierra Miwok'},
    'ske': {'en': 'Seke (Vanuatu)'},
    'skf': {'en': 'Sakirabiá'},
    'skg': {'en': 'Sakalava Malagasy'},
    'skh': {'en': 'Sikule'},
    'ski': {'en': 'Sika'},
    'skj': {'en': 'Seke (Nepal)'},
    'skk': {'en': 'Sok'},
    'skm': {'en': 'Kutong'},
    'skn': {'en': 'Kolibugan Subanon'},
    'sko': {'en': 'Seko Tengah'},
    'skp': {'en': 'Sekapan'},
    'skq': {'en': 'Sininkere'},
    'skr': {'bal-Latn': 'Saráeki', 'et': 'seraiki', 'fi': 'saraiki', 'gd': 'Saraiki', 'ia': 'saraiki', 'ig': 'skrr', 'ku': 'seraîkî', 'skr': 'سرائیکی', 'vec': 'saraiki', 'zh': '色莱基语', 'zh-Hans': '色莱基语', 'en': 'Saraiki'},
    'sks': {'en': 'Maia'},
    'skt': {'en': 'Sakata'},
    'sku': {'en': 'Sakao'},
    'skv': {'en': 'Skou'},
    'skw': {'en': 'Skepi Creole Dutch'},
    'skx': {'en': 'Seko Padang'},
    'sky': {'en': 'Sikaiana'},
    'skz': {'en': 'Sekar'},
    'sl': {'af': 'Sloweens', 'ak': 'Slovɛniafoɔ Kasa', 'am': 'ስሎቬንኛ', 'ar': 'السلوفانية', 'as': 'শ্লোভেনিয়ান', 'ast': 'eslovenu', 'az': 'sloven', 'az-Cyrl': 'словен', 'az-Latn': 'sloven', 'bal-Latn': 'Solwiniái', 'be': 'славенская', 'be-tarask': 'славенская', 'bg': 'словенски', 'bgn': 'اسلوانی', 'bn': 'স্লোভেনীয়', 'br': 'sloveneg', 'brx': 'स्लोभेनियान', 'bs': 'slovenski', 'bs-Cyrl': 'словенски', 'bs-Latn': 'slovenski', 'ca': 'eslovè', 'ca-ES-valencia': 'eslovè', 'ccp': '𑄥𑄳𑄣𑄮𑄞𑄬𑄚𑄩𑄠𑄧', 'ce': 'словенийн', 'chr': 'ᏍᎶᏫᏂᎠᏂ', 'ckb': 'سلۆڤێنی', 'cs': 'slovinština', 'cy': 'Slofeneg', 'da': 'slovensk', 'de': 'Slowenisch', 'dsb': 'słowjeńšćina', 'dz': 'སུ་ལོ་བི་ནི་ཡཱན་ཁ', 'ee': 'sloveniagbe', 'el': 'Σλοβενικά', 'el-polyton': 'Σλοβενικά', 'en': 'Slovenian', 'eo': 'slovena', 'es': 'esloveno', 'et': 'sloveeni', 'eu': 'esloveniera', 'fa': 'اسلوونیایی', 'ff-Adlm': '𞤅𞤤𞤮𞤾𞤫𞤲𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'sloveeni', 'fil': 'Slovenian', 'fo': 'slovenskt', 'fr': 'slovène', 'fur': 'sloven', 'fy': 'Sloveensk', 'ga': 'Slóivéinis', 'gd': 'Slòbhainis', 'gl': 'esloveno', 'gsw': 'Slowenisch', 'gu': 'સ્લોવેનિયન', 'ha': 'Basulabe', 'he': 'סלובנית', 'hi': 'स्लोवेनियाई', 'hi-Latn': 'Slovenian', 'hr': 'slovenski', 'hsb': 'słowjenšćina', 'ht': 'slovène', 'hu': 'szlovén', 'hy': 'սլովեներեն', 'ia': 'sloveno', 'id': 'Slovenia', 'ie': 'slovenian', 'ig': 'Asụsụ Slovenia', 'is': 'slóvenska', 'it': 'sloveno', 'ja': 'スロベニア語', 'jv': 'Slovenia', 'ka': 'სლოვენური', 'kea': 'slovéniu', 'kgp': 'erovenũ', 'kk': 'словен тілі', 'kk-Cyrl': 'словен тілі', 'km': 'ស្លូវ៉ានី', 'kn': 'ಸ್ಲೋವೇನಿಯನ್', 'ko': '슬로베니아어', 'kok': 'स्लोवानियन', 'kok-Deva': 'स्लोवानियन', 'ks': 'سلووینیَن', 'ks-Arab': 'سلووینیَن', 'ksh': 'ẞloveenesch', 'ku': 'slovenî', 'kxv': 'sloveniyati', 'kxv-Deva': 'स्लोवेनियाति', 'kxv-Latn': 'sloveniyati', 'kxv-Orya': 'ସ୍ଲବେନିଆ ତି', 'kxv-Telu': 'స్లోవేనియన్', 'ky': 'словенче', 'lb': 'Slowenesch', 'lkt': 'Slovenian Iyápi', 'lld': 'sloven', 'lo': 'ສະໂລເວນຽນ', 'lrc': 'ئسلوڤئنیایی', 'lt': 'slovėnų', 'lv': 'slovēņu', 'mai': 'स्लोबेनियाई', 'mi': 'Horowinia', 'mk': 'словенечки', 'ml': 'സ്ലോവേനിയൻ', 'mn': 'словени', 'mr': 'स्लोव्हेनियन', 'ms': 'Slovenia', 'mt': 'Sloven', 'my': 'ဆလိုဗေးနီးယား', 'mzn': 'اسلوونیایی', 'nb': 'slovensk', 'ne': 'स्लोभेनियाली', 'nl': 'Sloveens', 'nn': 'slovensk', 'no': 'slovensk', 'nqo': 'ߛߑߟߏߝ߭ߋߣߌߞߊ߲', 'om': 'Afaan Islovaniyaa', 'or': 'ସ୍ଲୋଭେନିଆନ୍', 'pa': 'ਸਲੋਵੇਨੀਆਈ', 'pa-Guru': 'ਸਲੋਵੇਨੀਆਈ', 'pcm': 'Slovẹ́niá Lángwej', 'pl': 'słoweński', 'ps': 'سلوواني', 'pt': 'esloveno', 'qu': 'Esloveno Simi', 'rm': 'sloven', 'ro': 'slovenă', 'ru': 'словенский', 'rw': 'Ikinyasiloveniya', 'sc': 'islovenu', 'sd': 'سلوويني', 'sd-Arab': 'سلوويني', 'se': 'slovenagiella', 'si': 'ස්ලෝවේනියානු', 'sk': 'slovinčina', 'sl': 'slovenščina', 'smn': 'sloveniakielâ', 'so': 'Islofeeniyaan', 'sq': 'sllovenisht', 'sr': 'словеначки', 'sr-Cyrl': 'словеначки', 'sr-Latn': 'slovenački', 'sv': 'slovenska', 'sw': 'Kislovenia', 'ta': 'ஸ்லோவேனியன்', 'te': 'స్లోవేనియన్', 'tg': 'словенӣ', 'th': 'สโลวีเนีย', 'ti': 'ስሎቬንኛ', 'tk': 'slowen dili', 'to': 'lea fakasolovenia', 'tr': 'Slovence', 'tt': 'словен', 'ug': 'سىلوۋېنچە', 'uk': 'словенська', 'ur': 'سلووینیائی', 'uz': 'slovencha', 'uz-Cyrl': 'словенча', 'uz-Latn': 'slovencha', 'vec': 'zloven', 'vi': 'Tiếng Slovenia', 'wae': 'Sloweniš', 'wo': 'Esloweniye', 'xnr': 'स्लोवेनियाई', 'yi': 'סלאוועניש', 'yo': 'Èdè Slovenia', 'yrl': 'esiruwenu', 'yue': '斯洛維尼亞文', 'yue-Hans': '斯洛维尼亚文', 'yue-Hant': '斯洛維尼亞文', 'zh': '斯洛文尼亚语', 'zh-Hans': '斯洛文尼亚语', 'zh-Hant': '斯洛維尼亞文', 'zu': 'isi-Slovenian'},
    'sla': {'en': 'Slavic languages'},
    'sla-pro': {'en': 'Proto-Slavic'},
    'slc': {'en': 'Sáliba'},
    'sld': {'en': 'Sissala'},
    'sle': {'en': 'Sholaga'},
    'slf': {'en': 'Swiss-Italian Sign Language'},
    'slg': {'en': 'Selungai Murut'},
    'slh': {'af': 'Suid-Lushootseed', 'am': 'ደቡባዊ ሉሹትሲድ', 'ar': 'لوشوتسيد الجنوبية', 'as': 'দাক্ষিণাত্যৰ লুছুটচীড', 'az': 'cənubi luşusid', 'az-Latn': 'cənubi luşusid', 'be': 'паўднёвая лушуцыд', 'be-tarask': 'паўднёвая лушуцыд', 'bg': 'южен лашутсийд', 'bn': 'দক্ষিণী লুশুটসীড', 'br': 'luchoutsideg ar Su', 'brx': 'खोलायारि लाशुतसीद', 'bs': 'južni lašutsid', 'bs-Cyrl': 'јужни лушотсед', 'bs-Latn': 'južni lašutsid', 'ca': 'lushootseed meridional', 'ca-ES-valencia': 'lushootseed meridional', 'chr': 'ᏧᎦᏃᏮ ᏗᏜ ᎷᏑᏘᏏᏗ', 'cs': 'lushootseed (jižní)', 'cy': 'Lushootseed Deheuol', 'da': 'sydlushootseed', 'de': 'Süd-Lushootseed', 'dsb': 'pódpołdnjowa lushootseedšćina', 'el': 'Νότια Λάσουτσιντ', 'el-polyton': 'Νότια Λάσουτσιντ', 'en': 'Southern Lushootseed', 'es': 'lushootseed meridional', 'et': 'Lõuna-Puget-Soundi sališi', 'eu': 'lushootseeda', 'fa': 'لاشوتسید جنوبی', 'ff-Adlm': '𞤂𞤵𞥃𞤵𞤼𞤧𞤭𞤣𞤪𞤫 𞤙𞤢𞥄𞤥𞤲𞤢𞥄𞤲𞤺𞤫𞤲𞤳𞤮', 'fi': 'lushootseed (eteläinen)', 'fil': 'Katimugang Lushootseed', 'fr': 'lushootseed du Sud', 'ga': 'Lushootseed an Deiscirt', 'gd': 'Lushootseed Dheasach', 'gl': 'lushootseed do sur', 'gu': 'દક્ષિણ લુશુટસીડ', 'ha': 'Lushbootseed na Kudanci', 'he': 'לשוטסיד', 'hi': 'दक्षिणी लशूटसीड', 'hi-Latn': 'Southern Lushootseed', 'hr': 'južni lushootseed', 'hsb': 'južna lushootseedšćina', 'ht': 'lushootseed du Sud', 'hu': 'déli lushootseed', 'hy': 'հարավային լուշուցիդ', 'ia': 'lushootseed del sud', 'id': 'Lushootseed Selatan', 'ig': 'Southern Lushootseed', 'is': 'suður-lushootseed', 'it': 'lushootseed meridionale', 'ja': '南部ルシュツィード語', 'jv': 'Lushootseed Sisih Kidul', 'ka': 'სამხრეთული ლუშუციდი', 'kk': 'оңтүстік лушуцид тілі', 'kk-Cyrl': 'оңтүстік лушуцид тілі', 'km': 'ឡាស៊ូតស៊ីតខាងត្បូង', 'kn': 'ದಕ್ಷಿಣ ಲುಶೂಟ್\u200cಸೀಡ್', 'ko': '남부 루슈트시드어', 'kok': 'दक्षिणी लशूतसीड', 'kok-Deva': 'दक्षिणी लशूतसीड', 'ku': 'lushootseeda başûrî', 'ky': 'түштүк лушуцидче (индей тили)', 'lo': 'ລູຊຸດຊີດໃຕ້', 'lt': 'pietų lushusidų', 'lv': 'dienvidlušucīdu', 'mai': 'दक्षिणी लसूटसीड', 'mi': 'Ratūti ki te Tonga', 'mk': 'јужен лушуцид', 'ml': 'സതേൺ ലുഷൂറ്റ്\u200cസീഡ്', 'mn': 'өмнөд лушуцид', 'mr': 'दक्षिणात्य लुशूटसीड', 'ms': 'Lushootseed Selatan', 'my': 'တောင် လာ့ရှုစတိ', 'nb': 'sørlig lushootseed', 'ne': 'दक्षिनी लुस्होस्टेड', 'nl': 'Zuid-Lushootseed', 'nn': 'sørleg lushootseed', 'no': 'sørlig lushootseed', 'or': 'ସାଉଥ୍ ଲୁଶୋସିଟେଡ୍', 'pa': 'ਦੱਖਣੀ ਲੁਸ਼ੂਟਸੀਡ', 'pa-Guru': 'ਦੱਖਣੀ ਲੁਸ਼ੂਟਸੀਡ', 'pcm': 'Saútan Lushútsid', 'pl': 'lushootseed południowy', 'ps': 'سویلي لوشوټسید', 'pt': 'lushootseed do sul', 'qu': 'Lushootseed Meridional', 'ro': 'lushootseed de usd', 'ru': 'южный лушуцид', 'sc': 'lushootseed meridionale', 'sd': 'ڏاکڻي لشوٽسيڊ', 'sd-Arab': 'ڏاکڻي لشوٽسيڊ', 'si': 'දකුණු ලුෂූට්සීඩ්', 'sk': 'lushootseed (juh)', 'sl': 'južna lušucidščina', 'so': 'Southern Lushootseed', 'sq': 'lashutsidishte jugore', 'sr': 'јужни лашутсид', 'sr-Cyrl': 'јужни лашутсид', 'sr-Latn': 'južni lašutsid', 'sv': 'sydlushootseed', 'sw': 'Lugha ya Lushootseed ya Kusini', 'ta': 'தெற்கு லுஷூட்சீட்', 'te': 'దక్షిణ లూషూట్\u200cసీడ్', 'th': 'ลูชูตซีดใต้', 'ti': 'ደቡባዊ ሉሹትሲድ', 'tk': 'günorta Luşutsid dili', 'to': 'lea fakalusūtisiti', 'tr': 'Güney Lushootseed', 'uk': 'південна лушуцид', 'ur': 'جنوبی لوشوٹسیڈ', 'uz': 'janubiy lushutsid', 'uz-Latn': 'janubiy lushutsid', 'vec': 'lushootseed meridionale', 'vi': 'Tiếng Lushootseed miền Nam', 'yo': 'Èdè Gusu Lushootseed', 'yue': '南盧紹錫德文', 'yue-Hans': '南卢绍锡德文', 'yue-Hant': '南盧紹錫德文', 'zh': '南卢舒特种子语', 'zh-Hans': '南卢舒特种子语', 'zh-Hant': '南盧紹錫德語', 'zu': 'Southern Lushootseed'},
    'sli': {'ast': 'baxu silesianu', 'cs': 'němčina (slezská)', 'cy': 'Is-silesieg', 'de': 'Schlesisch (Niederschlesisch)', 'en': 'Lower Silesian', 'et': 'alamsileesia', 'fa': 'سیلزیایی سفلی', 'fi': 'sleesiansaksa', 'fr': 'bas-silésien', 'hi-Latn': 'Lower Silesian', 'ht': 'bas-silésien', 'id': 'Silesia Rendah', 'it': 'tedesco slesiano', 'ja': '低シレジア語', 'lb': 'Nidderschlesesch', 'lt': 'sileziečių žemaičių', 'mk': 'долношлезиски', 'nb': 'lavschlesisk', 'ne': 'तल्लो सिलेसियाली', 'nl': 'Silezisch Duits', 'nn': 'lavschlesisk', 'no': 'lavschlesisk', 'pl': 'dolnośląski', 'sv': 'lågsilesiska', 'th': 'ไซลีเซียตอนล่าง', 'to': 'lea fakasilesia-hifo', 'tr': 'Aşağı Silezyaca', 'yi': 'אונטער שלעזיש', 'yue': '下西利西亞文', 'yue-Hans': '下西利西亚文', 'yue-Hant': '下西利西亞文', 'zh-Hant': '下西利西亞文'},
    'slj': {'en': 'Salumá'},
    'sll': {'en': 'Salt-Yui'},
    'slm': {'en': 'Pangutaran Sama'},
    'sln': {'en': 'Salinan'},
    'slp': {'en': 'Lamaholot'},
    'slq': {'en': 'Salchuq'},
    'slr': {'en': 'Salar'},
    'sls': {'en': 'Singapore Sign Language'},
    'slt': {'en': 'Sila'},
    'slu': {'en': 'Selaru'},
    'slw': {'en': 'Sialum'},
    'slx': {'en': 'Salampasu'},
    'sly': {'ast': 'selayarés', 'cs': 'selajarština', 'de': 'Selayar', 'en': 'Selayar', 'et': 'selajari', 'fi': 'selayar', 'fr': 'sélayar', 'gd': 'Selayar', 'hi-Latn': 'Selayar', 'ht': 'sélayar', 'id': 'Selayar', 'it': 'selayar', 'ja': 'スラヤール語', 'lb': 'Selayar', 'lt': 'selajarų', 'mk': 'селајарски', 'nb': 'selayar', 'nl': 'Selayar', 'nn': 'selayar', 'no': 'selayar', 'pl': 'selayar', 'sv': 'selayar', 'th': 'เซลายาร์', 'to': 'lea fakaselaiā', 'tr': 'Selayar', 'yue': '塞拉亞文', 'yue-Hans': '塞拉亚文', 'yue-Hant': '塞拉亞文', 'zh-Hant': '塞拉亞文'},
    'slz': {'en': "Ma'ya"},
    'sm': {'af': 'Samoaans', 'am': 'ሳሞኣን', 'ar': 'الساموائية', 'as': 'ছামোন', 'ast': 'samoanu', 'az': 'samoa', 'az-Cyrl': 'самоа', 'az-Latn': 'samoa', 'be': 'самоа', 'be-tarask': 'самоа', 'bg': 'самоански', 'bgn': 'ساموآیی', 'bn': 'সামোয়ান', 'br': 'samoan', 'brx': 'साम’वान', 'bs': 'samoanski', 'bs-Cyrl': 'самоански', 'bs-Latn': 'samoanski', 'ca': 'samoà', 'ca-ES-valencia': 'samoà', 'ccp': '𑄥𑄟𑄮𑄠𑄚𑄴', 'ce': 'самоанойн', 'chr': 'ᏌᎼᏯᏂ', 'ckb': 'سامۆیی', 'cs': 'samojština', 'cy': 'Samöeg', 'da': 'samoansk', 'de': 'Samoanisch', 'dsb': 'samošćina', 'ee': 'samoagbe', 'el': 'Σαμοανά', 'el-polyton': 'Σαμοανά', 'en': 'Samoan', 'eo': 'samoa', 'es': 'samoano', 'et': 'samoa', 'eu': 'samoera', 'fa': 'ساموآیی', 'ff-Adlm': '𞤅𞤢𞤥𞤮𞤢𞤲𞤪𞤫', 'fi': 'samoa', 'fil': 'Samoan', 'fo': 'sámoiskt', 'fr': 'samoan', 'fur': 'samoan', 'fy': 'Samoaansk', 'ga': 'Samóis', 'gd': 'Samothais', 'gl': 'samoano', 'gsw': 'Samoanisch', 'gu': 'સામોન', 'ha': 'Samoan', 'haw': 'Kāmoa', 'he': 'סמואית', 'hi': 'सामोन', 'hi-Latn': 'Samoan', 'hr': 'samoanski', 'hsb': 'samoašćina', 'ht': 'samoan', 'hu': 'szamoai', 'hy': 'սամոաերեն', 'ia': 'samoano', 'id': 'Samoa', 'ig': 'Samoan', 'is': 'samóska', 'it': 'samoano', 'ja': 'サモア語', 'jv': 'Samoa', 'ka': 'სამოა', 'kgp': 'samovỹnũ', 'kk': 'самоа тілі', 'kk-Cyrl': 'самоа тілі', 'km': 'សាម័រ', 'kn': 'ಸಮೋವನ್', 'ko': '사모아어', 'kok': 'सामोअन', 'kok-Deva': 'सामोअन', 'ks': 'سَمواَن', 'ks-Arab': 'سَمواَن', 'ksh': 'Sammohanesch', 'ku': 'samoayî', 'ky': 'самоанча', 'lb': 'Samoanesch', 'lo': 'ຊາມົວ', 'lt': 'Samoa', 'lv': 'samoāņu', 'mai': 'समोअन', 'mi': 'Hāmoa', 'mk': 'самоански', 'ml': 'സമോവൻ', 'mn': 'самоа', 'mr': 'सामोअन', 'ms': 'Samoa', 'mt': 'Samoan', 'my': 'ဆမိုအာ', 'nb': 'samoansk', 'ne': 'सामोआ', 'nl': 'Samoaans', 'nn': 'samoansk', 'no': 'samoansk', 'nqo': 'ߛߊߡߏߥߊߞߊ߲', 'or': 'ସାମୋଆନ୍', 'pa': 'ਸਾਮੋਨ', 'pa-Guru': 'ਸਾਮੋਨ', 'pcm': 'Samóá Lángwej', 'pl': 'samoański', 'ps': 'ساموآن', 'pt': 'samoano', 'qu': 'Samoano Simi', 'rm': 'samoan', 'ro': 'samoană', 'ru': 'самоанский', 'sc': 'samoanu', 'sd': 'سموئا', 'sd-Arab': 'سموئا', 'se': 'samoagiella', 'si': 'සෑමොඅන්', 'sk': 'samojčina', 'sl': 'samoanščina', 'smn': 'samoakielâ', 'so': 'Samowan', 'sq': 'samoanisht', 'sr': 'самоански', 'sr-Cyrl': 'самоански', 'sr-Latn': 'samoanski', 'sv': 'samoanska', 'sw': 'Kisamoa', 'ta': 'சமோவான்', 'te': 'సమోవన్', 'th': 'ซามัว', 'ti': 'ሳሞእኛ', 'tk': 'samoa dili', 'to': 'lea fakahaʻamoa', 'tr': 'Samoa dili', 'ug': 'ساموئاچە', 'uk': 'самоанська', 'ur': 'ساموآن', 'uz': 'samoa', 'uz-Latn': 'samoa', 'vec': 'samoan', 'vi': 'Tiếng Samoa', 'wae': 'Samoaniš', 'xnr': 'सामोन', 'yi': 'סאַמאאַניש', 'yo': 'Sámóánù', 'yrl': 'samuanu', 'yue': '薩摩亞文', 'yue-Hans': '萨摩亚文', 'yue-Hant': '薩摩亞文', 'zh': '萨摩亚语', 'zh-Hans': '萨摩亚语', 'zh-Hant': '薩摩亞文', 'zu': 'isi-Samoan'},
    'sma': {'af': 'Suid-Sami', 'am': 'ደቡባዊ ሳሚ', 'ar': 'السامي الجنوبي', 'as': 'দাক্ষিণাত্য ছামি', 'ast': 'sami del sur', 'az': 'cənubi sami', 'az-Cyrl': 'ҹәнуби сами', 'az-Latn': 'cənubi sami', 'bal-Latn': 'Zerbári Sámi', 'be': 'паўднёвасаамская', 'be-tarask': 'паўднёвасаамская', 'bg': 'южносаамски', 'bgn': 'جهلسرین سامی', 'bn': 'দক্ষিণাঞ্চলীয় সামি', 'br': 'sámi ar Su', 'brx': 'पश्चीमी सामी', 'bs': 'južni sami', 'bs-Cyrl': 'јужни сами', 'bs-Latn': 'južni sami', 'ca': 'sami meridional', 'ca-ES-valencia': 'sami meridional', 'ccp': '𑄘𑄧𑄉𑄨𑄚𑄴 𑄢𑄬𑄌𑄴𑄎𑄮𑄢𑄴 𑄥𑄟𑄨', 'ce': 'саамийн (къилба)', 'chr': 'ᏧᎦᎾᏮ ᏗᏜ ᏌᎻ', 'ckb': 'سامیی باشوور', 'cs': 'sámština (jižní)', 'cy': 'Sami Deheuol', 'da': 'sydsamisk', 'de': 'Südsamisch', 'dsb': 'pódpołdnjowa samišćina', 'el': 'Νότια Σάμι', 'el-polyton': 'Νότια Σάμι', 'en': 'Southern Sami', 'es': 'sami meridional', 'et': 'lõunasaami', 'eu': 'hegoaldeko samiera', 'fa': 'سامی جنوبی', 'fi': 'eteläsaame', 'fil': 'Katimugang Sami', 'fo': 'suður sámiskt', 'fr': 'same du Sud', 'fy': 'Sûd-Samysk', 'ga': 'Sáimis Theas', 'gd': 'Sàmais Dheasach', 'gl': 'saami meridional', 'gsw': 'Süüd-Samisch', 'gu': 'દક્ષિણી સામી', 'he': 'סאמי דרומית', 'hi': 'दक्षिणी सामी', 'hi-Latn': 'Southern Sami', 'hr': 'južni sami', 'hsb': 'južnosamišćina', 'ht': 'same du Sud', 'hu': 'déli számi', 'hy': 'հարավային սաամի', 'ia': 'sami del sud', 'id': 'Sami Selatan', 'ig': 'Southern Sami', 'is': 'suðursamíska', 'it': 'sami del sud', 'ja': '南サーミ語', 'ka': 'სამხრეთსამური', 'kgp': 'samĩ sur', 'kk': 'оңтүстік саам тілі', 'kk-Cyrl': 'оңтүстік саам тілі', 'km': 'សាមីខាងត្បូង', 'kn': 'ದಕ್ಷಿಣ ಸಾಮಿ', 'ko': '남부 사미어', 'kok': 'दक्षिणी सामी', 'kok-Deva': 'दक्षिणी सामी', 'ks': 'جنوٗبی سَمی', 'ks-Arab': 'جنوٗبی سَمی', 'ku': 'samîya başûr', 'ky': 'түштүк саамиче', 'lb': 'Südsamesch', 'lo': 'ຊາມິໃຕ້', 'lrc': 'سامی ھارگە', 'lt': 'pietų samių', 'lv': 'dienvidsāmu', 'mk': 'јужен сами', 'ml': 'തെക്കൻ സമി', 'mn': 'өмнөд сами', 'mr': 'दक्षिणात्य सामी', 'ms': 'Sami Selatan', 'mt': 'Sami tan-Nofsinhar', 'my': 'တောင် ဆာမိ', 'mzn': 'جنوبی سامی', 'nb': 'sørsamisk', 'ne': 'दक्षिणी सामी', 'nl': 'Zuid-Samisch', 'nn': 'sørsamisk', 'no': 'sørsamisk', 'or': 'ଦକ୍ଷିଣ ସାମି', 'pa': 'ਦੱਖਣੀ ਸਾਮੀ', 'pa-Guru': 'ਦੱਖਣੀ ਸਾਮੀ', 'pl': 'południowolapoński', 'ps': 'سویلي سامی', 'pt': 'sami meridional', 'qu': 'Qulla Sami Simi', 'rm': 'sami dal sid', 'ro': 'sami de sud', 'ru': 'южносаамский', 'sd': 'ڏکڻ سامي', 'sd-Arab': 'ڏکڻ سامي', 'se': 'lullisámegiella', 'si': 'දකුණු සාමි', 'sk': 'saamčina (južná)', 'sl': 'južna samijščina', 'sma': 'Åarjelsaemien gïele', 'smn': 'maadâsämikielâ', 'sq': 'samishte jugore', 'sr': 'јужни сами', 'sr-Cyrl': 'јужни сами', 'sr-Latn': 'južni sami', 'sv': 'sydsamiska', 'sw': 'Kisami cha Kusini', 'ta': 'தெற்கு சமி', 'te': 'దక్షిణ సామి', 'tg': 'самии ҷанубӣ', 'th': 'ซามิใต้', 'ti': 'ደቡባዊ ሳሚ', 'tk': 'günorta saam dili', 'to': 'lea fakasami-tonga', 'tr': 'Güney Laponcası', 'tt': 'көньяк саам', 'ug': 'جەنۇبىي سامىچە', 'uk': 'південносаамська', 'ur': 'جنوبی سامی', 'uz': 'janubiy saam', 'uz-Cyrl': 'жанубий саамча', 'uz-Latn': 'janubiy saam', 'vec': 'sami meridionale', 'vi': 'Tiếng Sami Miền Nam', 'wo': 'Sami gu Saalum', 'yrl': 'sami meridiunau', 'yue': '南薩米文', 'yue-Hans': '南萨米文', 'yue-Hant': '南薩米文', 'zh': '南萨米语', 'zh-Hans': '南萨米语', 'zh-Hant': '南薩米文', 'zu': 'isi-Southern Sami'},
    'smb': {'en': 'Simbari'},
    'smc': {'en': 'Som'},
    'smd': {'en': 'Sama'},
    'smf': {'en': 'Auwe'},
    'smg': {'en': 'Simbali'},
    'smh': {'en': 'Samei'},
    'smi': {'en': 'Sami languages'},
    'smi-pro': {'en': 'Proto-Samic'},
    'smj': {'af': 'Lule Sami', 'am': 'ሉሌ ሳሚ', 'ar': 'اللول سامي', 'as': 'লুলে ছামি', 'ast': 'lule sami', 'az': 'lule sami', 'az-Cyrl': 'луле сами', 'az-Latn': 'lule sami', 'bal-Latn': 'Lule Sámi', 'be': 'луле-саамская', 'be-tarask': 'луле-саамская', 'bg': 'луле-саамски', 'bgn': 'لوله سامی', 'bn': 'লুলে সামি', 'br': 'sámi Luleå', 'brx': 'लुले सामी', 'bs': 'lule sami', 'bs-Cyrl': 'луле сами', 'bs-Latn': 'lule sami', 'ca': 'sami lule', 'ca-ES-valencia': 'sami lule', 'ccp': '𑄣𑄪𑄣𑄬 𑄥𑄟𑄨', 'ce': 'луле-саамийн', 'chr': 'ᎷᎴ ᏌᎻ', 'ckb': 'لوولێ سامی', 'cs': 'sámština (lulejská)', 'cy': 'Sami Lwle', 'da': 'lulesamisk', 'de': 'Lule-Samisch', 'dsb': 'lule-samišćina', 'el': 'Λούλε Σάμι', 'el-polyton': 'Λούλε Σάμι', 'en': 'Lule Sami', 'es': 'sami lule', 'et': 'Lule saami', 'eu': 'Luleko samiera', 'fa': 'لوله سامی', 'fi': 'luulajansaame', 'fil': 'Lule Sami', 'fo': 'lule sámiskt', 'fr': 'same de Lule', 'fy': 'Lule Sami', 'ga': 'Sáimis Lule', 'gd': 'Sàmais Lule', 'gl': 'saami de Lule', 'gsw': 'Lule-Samisch', 'gu': 'લુલે સામી', 'he': 'לולה סאמי', 'hi': 'ल्युल सामी', 'hi-Latn': 'Lule Sami', 'hr': 'lule sami', 'hsb': 'lule-samišćina', 'ht': 'same de Lule', 'hu': 'lulei számi', 'hy': 'լուլե սաամի', 'ia': 'sami de Lule', 'id': 'Lule Sami', 'ig': 'Lule Sami', 'is': 'lúlesamíska', 'it': 'sami di Lule', 'ja': 'ルレ・サーミ語', 'ka': 'ლულე-საამური', 'kgp': 'samĩ Rure tá', 'kk': 'луле саам тілі', 'kk-Cyrl': 'луле саам тілі', 'km': 'លូលីសាមី', 'kn': 'ಲೂಲ್ ಸಾಮಿ', 'ko': '룰레 사미어', 'kok': 'लुले सामी', 'kok-Deva': 'लुले सामी', 'ks': 'لولیے سَمی', 'ks-Arab': 'لولیے سَمی', 'ksh': 'Lule-Läppländesch', 'ku': 'samiya lule', 'ky': 'луле саамиче', 'lb': 'Lule-Lappesch', 'lo': 'ລຸນຊາມິ', 'lrc': 'لۉلئ سامی', 'lt': 'Liuleo samių', 'lv': 'Luleo sāmu', 'mk': 'луле сами', 'ml': 'ലൂലീ സമി', 'mn': 'люле сами', 'mr': 'ल्युल सामी', 'ms': 'Lule Sami', 'mt': 'Lule Sami', 'my': 'လူလီ ဆာမိ', 'mzn': 'لوله سامی', 'nb': 'lulesamisk', 'ne': 'लुले सामी', 'nl': 'Lule-Samisch', 'nn': 'lulesamisk', 'no': 'lulesamisk', 'or': 'ଲୁଲେ ସାମି', 'pa': 'ਲਿਊਲ ਸਾਮੀ', 'pa-Guru': 'ਲਿਊਲ ਸਾਮੀ', 'pl': 'lule', 'ps': 'لول سامي', 'pt': 'sami de Lule', 'qu': 'Sami Lule Simi', 'rm': 'sami lule', 'ro': 'sami lule', 'ru': 'луле-саамский', 'sd': 'لولي سامي', 'sd-Arab': 'لولي سامي', 'se': 'julevsámegiella', 'si': 'ලුලේ සාමි', 'sk': 'saamčina (lulská)', 'sl': 'luleška samijščina', 'smj': 'julevsámegiella', 'smn': 'juulevsämikielâ', 'sq': 'samishte lule', 'sr': 'луле сами', 'sr-Cyrl': 'луле сами', 'sr-Latn': 'lule sami', 'sv': 'lulesamiska', 'sw': 'Kisami cha Lule', 'ta': 'லுலே சமி', 'te': 'లులే సామి', 'tg': 'луле самӣ', 'th': 'ซามิลูเล', 'ti': 'ሉለ ሳሚ', 'tk': 'lule-saam dili', 'to': 'lea fakasami-lule', 'tr': 'Lule Laponcası', 'tt': 'луле-саам', 'ug': 'لۇلې سامىچە', 'uk': 'саамська луле', 'ur': 'لول سامی', 'uz': 'lule-saam', 'uz-Cyrl': 'луле-саамча', 'uz-Latn': 'lule-saam', 'vec': 'sami de’l Lule', 'vi': 'Tiếng Lule Sami', 'wo': 'Lule Sami', 'yrl': 'sami Lulewara', 'yue': '魯勒薩米文', 'yue-Hans': '鲁勒萨米文', 'yue-Hant': '魯勒薩米文', 'zh': '吕勒萨米语', 'zh-Hans': '吕勒萨米语', 'zh-Hant': '魯勒薩米文', 'zu': 'isi-Lule Sami'},
    'smk': {'en': 'Bolinao'},
    'sml': {'en': 'Central Sama'},
    'smm': {'en': 'Musasa'},
    'smn': {'af': 'Inari Sami', 'am': 'ኢናሪ ሳሚ', 'ar': 'الإيناري سامي', 'as': 'ইনাৰি ছামি', 'ast': 'inari sami', 'az': 'inari sami', 'az-Cyrl': 'инари сами', 'az-Latn': 'inari sami', 'bal-Latn': 'Inári Sámi', 'be': 'інары-саамская', 'be-tarask': 'інары-саамская', 'bg': 'инари-саамски', 'bgn': 'اناری سمی', 'bn': 'ইনারি সামি', 'br': 'sámi Inari', 'brx': 'ईनारी सामी', 'bs': 'inari sami', 'bs-Cyrl': 'инари сами', 'bs-Latn': 'inari sami', 'ca': 'sami d’Inari', 'ca-ES-valencia': 'sami d’Inari', 'ccp': '𑄃𑄨𑄚𑄢𑄨 𑄥𑄟𑄨', 'ce': 'инари-саамийн', 'chr': 'ᎢᎾᎵ ᏌᎻ', 'ckb': 'ئیناری سامی', 'cs': 'sámština (inarijská)', 'cy': 'Inari Sami', 'da': 'enaresamisk', 'de': 'Inari-Samisch', 'dsb': 'inari-samišćina', 'el': 'Ινάρι Σάμι', 'el-polyton': 'Ινάρι Σάμι', 'en': 'Inari Sami', 'es': 'sami inari', 'et': 'Inari saami', 'eu': 'Inariko samiera', 'fa': 'ایناری سامی', 'ff-Adlm': '𞤋𞤲𞤢𞤪𞤭𞤧𞤳𞤢𞤤𞤭𞥅𞤪𞤫', 'fi': 'inarinsaame', 'fil': 'Inari Sami', 'fo': 'inari sami', 'fr': 'same d’Inari', 'fy': 'Inari Sami', 'ga': 'Sáimis Inari', 'gd': 'Sàmais Inari', 'gl': 'saami de Inari', 'gsw': 'Inari-Samisch', 'gu': 'ઇનારી સામી', 'ha': 'Inari Sami', 'he': 'אינארי סאמי', 'hi': 'इनारी सामी', 'hi-Latn': 'Inari Sami', 'hr': 'inari sami', 'hsb': 'inari-samišćina', 'ht': 'same d’Inari', 'hu': 'inari számi', 'hy': 'ինարի սաամի', 'ia': 'sami de Inari', 'id': 'Inari Sami', 'ig': 'Inari Sami', 'is': 'enaresamíska', 'it': 'sami di Inari', 'ja': 'イナリ・サーミ語', 'jv': 'Inari Sami', 'ka': 'ინარი-საამური', 'kea': 'inari sami', 'kgp': 'samĩ Inari tá', 'kk': 'инари саам тілі', 'kk-Cyrl': 'инари саам тілі', 'km': 'អ៊ីណារីសាមី', 'kn': 'ಇನಾರಿ ಸಮೀ', 'ko': '이나리 사미어', 'kok': 'ईनारी सामी', 'kok-Deva': 'ईनारी सामी', 'ks': 'اِناری سَمی', 'ks-Arab': 'اِناری سَمی', 'ksh': 'Inahri Lappländesch', 'ku': 'samîya înarî', 'ky': 'инари саамиче', 'lb': 'Inari-Lappesch', 'lo': 'ອີນາຣິຊາມິ', 'lrc': 'ئیناری سامی', 'lt': 'Inario samių', 'lv': 'Inari sāmu', 'mai': 'इनारी सामी', 'mi': 'Inari Hami', 'mk': 'инариски сами', 'ml': 'ഇനാരി സമി', 'mn': 'инари сами', 'mr': 'इनारी सामी', 'ms': 'Inari Sami', 'mt': 'Inari Sami', 'my': 'အီနာရီ ဆာမိ', 'mzn': 'ایناری سامی', 'nb': 'enaresamisk', 'ne': 'इनारी सामी', 'nl': 'Inari-Samisch', 'nn': 'enaresamisk', 'no': 'enaresamisk', 'nqo': 'ߌߣߊߙߌ߫ ߛߊߡߌߞߊ߲', 'or': 'ଇନାରୀ ସାମି', 'pa': 'ਇਨਾਰੀ ਸਾਮੀ', 'pa-Guru': 'ਇਨਾਰੀ ਸਾਮੀ', 'pcm': 'Ínárí Sámí Lángwej', 'pl': 'inari', 'ps': 'اناري سميع', 'pt': 'lapão de Inari', 'qu': 'Sami Inari Simi', 'rm': 'sami inari', 'ro': 'sami inari', 'ru': 'инари-саамский', 'sc': 'sami de sos inari', 'sd': 'اناري سامي', 'sd-Arab': 'اناري سامي', 'se': 'anárašgiella', 'si': 'ඉනාරි සාමි', 'sk': 'saamčina (inarijská)', 'sl': 'inarska samijščina', 'smn': 'anarâškielâ', 'so': 'Inaari Saami', 'sq': 'samishte inari', 'sr': 'инари сами', 'sr-Cyrl': 'инари сами', 'sr-Latn': 'inari sami', 'sv': 'enaresamiska', 'sw': 'Kisami cha Inari', 'ta': 'இனாரி சமி', 'te': 'ఇనారి సామి', 'tg': 'инари самӣ', 'th': 'ซามิอีนารี', 'ti': 'ሳሚ ኢናሪ', 'tk': 'inari-saam dili', 'to': 'lea fakasami-ʻinali', 'tr': 'İnari Laponcası', 'tt': 'инари-саам', 'ug': 'ئىنارى سامىچە', 'uk': 'саамська інарі', 'ur': 'اناری سامی', 'uz': 'inari-saam', 'uz-Cyrl': 'инари-саамча', 'uz-Latn': 'inari-saam', 'vec': 'sami inarian', 'vi': 'Tiếng Inari Sami', 'wo': 'Inari Sami', 'xnr': 'इनारी सामी', 'yo': 'Inari Sami', 'yrl': 'sami Inariwara', 'yue': '伊納裡薩米文', 'yue-Hans': '伊纳里萨米文', 'yue-Hant': '伊納裡薩米文', 'zh': '伊纳里萨米语', 'zh-Hans': '伊纳里萨米语', 'zh-Hant': '伊納里薩米文', 'zu': 'isi-Inari Sami'},
    'smp': {'en': 'Samaritan Hebrew'},
    'smq': {'en': 'Samo'},
    'smr': {'en': 'Simeulue'},
    'sms': {'af': 'Skolt Sami', 'am': 'ስኮልት ሳሚ', 'ar': 'السكولت سامي', 'as': 'স্কোল্ট ছামি', 'ast': 'skolt sami', 'az': 'skolt sami', 'az-Cyrl': 'сколт сами', 'az-Latn': 'skolt sami', 'bal-Latn': 'Eskált Sámi', 'be': 'колта-саамская', 'be-tarask': 'колта-саамская', 'bg': 'сколт-саамски', 'bgn': 'اسکولت سامی', 'bn': 'স্কোল্ট সামি', 'br': 'sámi Skolt', 'brx': 'स्कल्ट सामी', 'bs': 'skolt sami', 'bs-Cyrl': 'сколтски језик', 'bs-Latn': 'skolt sami', 'ca': 'sami skolt', 'ca-ES-valencia': 'sami skolt', 'ccp': '𑄥𑄳𑄇𑄧𑄣𑄳𑄑𑄧 𑄥𑄟𑄨', 'ce': 'скольт-саамийн', 'chr': 'ᏍᎪᎵᏘ ᏌᎻ', 'ckb': 'سامیی سکۆڵت', 'cs': 'sámština (skoltská)', 'cy': 'Sami Scolt', 'da': 'skoltesamisk', 'de': 'Skolt-Samisch', 'dsb': 'skolt-samišćina', 'el': 'Σκολτ Σάμι', 'el-polyton': 'Σκολτ Σάμι', 'en': 'Skolt Sami', 'es': 'sami skolt', 'et': 'koltasaami', 'eu': 'skolten samiera', 'fa': 'اسکولت سامی', 'ff-Adlm': '𞤅𞤭𞤳𞤮𞤤𞤼 𞤅𞤢𞤥𞤭𞥅𞤪𞤫', 'fi': 'koltansaame', 'fil': 'Skolt Sami', 'fo': 'skolt sámiskt', 'fr': 'same skolt', 'fy': 'Skolt Sami', 'ga': 'Sáimis Skolt', 'gd': 'Sàmais Skolt', 'gl': 'saami skolt', 'gsw': 'Skolt-Samisch', 'gu': 'સ્કોલ્ટ સામી', 'ha': 'Skolt Sami', 'he': 'סקולט סאמי', 'hi': 'स्कोल्ट सामी', 'hi-Latn': 'Skolt Sami', 'hr': 'skolt sami', 'hsb': 'skolt-samišćina', 'ht': 'same skolt', 'hu': 'kolta számi', 'hy': 'սկոլտ սաամի', 'ia': 'sami skolt', 'id': 'Skolt Sami', 'ig': 'Skolt sami', 'is': 'skoltesamíska', 'it': 'sami skolt', 'ja': 'スコルト・サーミ語', 'jv': 'Skolt Sami', 'ka': 'სკოლტ-საამური', 'kgp': 'samĩ Isikórti tá', 'kk': 'колтта саам тілі', 'kk-Cyrl': 'колтта саам тілі', 'km': 'ស្កុលសាមី', 'kn': 'ಸ್ಕೋಟ್ ಸಾಮಿ', 'ko': '스콜트 사미어', 'kok': 'स्कोल्ट सामी', 'kok-Deva': 'स्कोल्ट सामी', 'ks': 'سکولٹ سَمی', 'ks-Arab': 'سکولٹ سَمی', 'ku': 'samîya skoltî', 'ky': 'сколт саамиче', 'lb': 'Skolt-Lappesch', 'lo': 'ສກອດຊາມິ', 'lrc': 'ئسکولت سامی', 'lt': 'Skolto samių', 'lv': 'skoltsāmu', 'mai': 'स्कोल्ट समी', 'mi': 'Hakoto Hāmi', 'mk': 'сколт сами', 'ml': 'സ്കോൾട്ട് സമി', 'mn': 'сколт сами', 'mr': 'स्कोल्ट सामी', 'ms': 'Skolt Sami', 'mt': 'Skolt Sami', 'my': 'စခိုးလ် ဆာမိ', 'mzn': 'سکولت سامی', 'nb': 'skoltesamisk', 'ne': 'स्कोइट सामी', 'nl': 'Skolt-Samisch', 'nn': 'skoltesamisk', 'no': 'skoltesamisk', 'or': 'ସ୍କୋଲ୍ଟ ସାମି', 'pa': 'ਸਕੌਲਟ ਸਾਮੀ', 'pa-Guru': 'ਸਕੌਲਟ ਸਾਮੀ', 'pcm': 'Skolt Sámí Lángwej', 'pl': 'skolt', 'ps': 'سکولټ سمیع', 'pt': 'sami de Skolt', 'qu': 'Sami Skolt Simi', 'rm': 'sami skolt', 'ro': 'sami skolt', 'ru': 'колтта-саамский', 'sc': 'sami skolt', 'sd': 'اسڪاٽ سامي', 'sd-Arab': 'اسڪاٽ سامي', 'se': 'nuortalašgiella', 'si': 'ස්කොල්ට් සාමි', 'sk': 'saamčina (skoltská)', 'sl': 'skoltska samijščina', 'smn': 'nuorttâlâškielâ', 'sms': 'sääʹmǩiõll', 'so': 'Skolt Sami', 'sq': 'samishte skolti', 'sr': 'сколт сами', 'sr-Cyrl': 'сколт сами', 'sr-Latn': 'skolt sami', 'sv': 'skoltsamiska', 'sw': 'Kisami cha Skolt', 'ta': 'ஸ்கோல்ட் சமி', 'te': 'స్కోల్ట్ సామి', 'tg': 'сколти самӣ', 'th': 'ซามิสคอลต์', 'ti': 'ሳሚ ስኮልት', 'tk': 'skolt-saam dili', 'to': 'lea fakasami-sikolita', 'tr': 'Skolt Laponcası', 'tt': 'колтта-саам', 'ug': 'سكولت سامىچە', 'uk': 'скольт-саамська', 'ur': 'سکولٹ سامی', 'uz': 'skolt-saam', 'uz-Cyrl': 'сколт-саамча', 'uz-Latn': 'skolt-saam', 'vec': 'sami skolt', 'vi': 'Tiếng Skolt Sami', 'wo': 'Eskolt Sami', 'xnr': 'स्कोल्ट सामी', 'yo': 'Èdè Sikoti Smi', 'yrl': 'sami Skoltwara', 'yue': '斯科特薩米文', 'yue-Hans': '斯科特萨米文', 'yue-Hant': '斯科特薩米文', 'zh': '斯科特萨米语', 'zh-Hans': '斯科特萨米语', 'zh-Hant': '斯科特薩米文', 'zu': 'isi-Skolt Sami'},
    'smt': {'en': 'Simte'},
    'smu': {'en': 'Somray'},
    'smv': {'en': 'Samvedi'},
    'smw': {'en': 'Sumbawa'},
    'smx': {'en': 'Samba'},
    'smy': {'en': 'Semnani'},
    'smz': {'en': 'Simeku'},
    'sn': {'af': 'Shona', 'am': 'ሾና', 'ar': 'الشونا', 'as': 'চোনা', 'ast': 'shona', 'az': 'şona', 'az-Cyrl': 'шона', 'az-Latn': 'şona', 'bal-Latn': 'Shoná', 'be': 'шона', 'be-tarask': 'шона', 'bg': 'шона', 'bgn': 'شونی', 'bn': 'শোনা', 'br': 'shona', 'brx': 'शना', 'bs': 'šona', 'bs-Cyrl': 'шона', 'bs-Latn': 'šona', 'ca': 'shona', 'ca-ES-valencia': 'shona', 'ccp': '𑄥𑄮𑄚', 'ce': 'шона', 'chr': 'ᏠᎾ', 'ckb': 'شۆنا', 'cs': 'šonština', 'cy': 'Shona', 'da': 'shona', 'de': 'Shona', 'dsb': 'šonšćina', 'ee': 'shonagbe', 'el': 'Σόνα', 'el-polyton': 'Σόνα', 'en': 'Shona', 'eo': 'ŝona', 'es': 'shona', 'et': 'šona', 'eu': 'shonera', 'fa': 'شونایی', 'ff-Adlm': '𞤅𞤮𞤲𞤢𞥄𞤪𞤫', 'fi': 'šona', 'fil': 'Shona', 'fo': 'shona', 'fr': 'shona', 'fy': 'Shona', 'ga': 'Seoinis', 'gd': 'Shona', 'gl': 'shona', 'gsw': 'Schhona', 'gu': 'શોના', 'ha': 'Shona', 'he': 'שונה', 'hi': 'शोणा', 'hi-Latn': 'Shona', 'hr': 'shona', 'hsb': 'šonašćina', 'ht': 'shona', 'hu': 'sona', 'hy': 'շոնա', 'ia': 'shona', 'id': 'Shona', 'ig': 'Shona', 'is': 'shona', 'it': 'shona', 'ja': 'ショナ語', 'jv': 'Shona', 'ka': 'შონა', 'kgp': 'sãnỹ', 'kk': 'шона тілі', 'kk-Cyrl': 'шона тілі', 'km': 'សូណា', 'kn': 'ಶೋನಾ', 'ko': '쇼나어', 'kok': 'शोना', 'kok-Deva': 'शोना', 'ks': 'شونا', 'ks-Arab': 'شونا', 'ksh': 'Schi-Schona', 'ku': 'şonayî', 'ky': 'шонача', 'lb': 'Shona', 'lld': 'shona', 'lo': 'ໂຊນາ', 'lrc': 'شونا', 'lt': 'šonų', 'lv': 'šonu', 'mai': 'शोना', 'mi': 'Hōna', 'mk': 'шона', 'ml': 'ഷോണ', 'mn': 'шона', 'mr': 'शोना', 'ms': 'Shona', 'mt': 'Shona', 'my': 'ရှိုနာ', 'mzn': 'شونا', 'nb': 'shona', 'ne': 'शोना', 'nl': 'Shona', 'nn': 'shona', 'no': 'shona', 'nqo': 'ߛߏߣߊߞߊ߲', 'or': 'ଶୋନା', 'pa': 'ਸ਼ੋਨਾ', 'pa-Guru': 'ਸ਼ੋਨਾ', 'pcm': 'Shóna', 'pl': 'shona', 'ps': 'شونا', 'pt': 'xona', 'qu': 'Shona Simi', 'rm': 'shona', 'ro': 'shona', 'ru': 'шона', 'sc': 'shona', 'sd': 'شونا', 'sd-Arab': 'شونا', 'si': 'ශෝනා', 'sk': 'šončina', 'sl': 'šonščina', 'smn': 'shona', 'sn': 'chiShona', 'so': 'Shoona', 'sq': 'shonisht', 'sr': 'шона', 'sr-Cyrl': 'шона', 'sr-Latn': 'šona', 'sv': 'shona', 'sw': 'Kishona', 'ta': 'ஷோனா', 'te': 'షోన', 'th': 'โชนา', 'ti': 'ሾና', 'tk': 'şona dili', 'to': 'lea fakasiona', 'tr': 'Şona dili', 'ug': 'شوناچە', 'uk': 'шона', 'ur': 'شونا', 'uz': 'shona', 'uz-Cyrl': 'шона', 'uz-Latn': 'shona', 'vec': 'shona', 'vi': 'Tiếng Shona', 'wae': 'Shona', 'xnr': 'शोणा', 'yi': 'שאנאַ', 'yo': 'Ṣọnà', 'yrl': 'xuna', 'yue': '塞內加爾文', 'yue-Hans': '塞内加尔文', 'yue-Hant': '塞內加爾文', 'zh': '绍纳语', 'zh-Hans': '绍纳语', 'zh-Hant': '紹納文', 'zu': 'isiShona'},
    'snb': {'en': 'Sebuyau'},
    'snc': {'en': 'Sinaugoro'},
    'sne': {'en': 'Bau Bidayuh'},
    'snf': {'en': 'Noon'},
    'sng': {'en': 'Sanga (Democratic Republic of Congo)'},
    'snh': {'en': 'Shinabo'},
    'sni': {'en': 'Sensi'},
    'snj': {'en': 'Riverain Sango'},
    'snk': {'af': 'Soninke', 'am': 'ሶኒንኬ', 'ar': 'السونينك', 'as': 'ছোনিনকে', 'ast': 'soninke', 'az': 'soninke', 'az-Cyrl': 'сонинке', 'az-Latn': 'soninke', 'be': 'санінке', 'be-tarask': 'санінке', 'bg': 'сонинке', 'bgn': 'سونینکه\u200cای', 'bn': 'সোনিঙ্কে', 'br': 'soninke', 'brx': 'स’निन्के', 'bs': 'soninke', 'bs-Cyrl': 'сонинке', 'bs-Latn': 'soninke', 'ca': 'soninke', 'ca-ES-valencia': 'soninke', 'ccp': '𑄥𑄮𑄚𑄨𑄋𑄴𑄇𑄬', 'ce': 'сонинке', 'chr': 'ᏐᏂᏂᎨ', 'ckb': 'سۆنینکێ', 'cs': 'sonikština', 'cy': 'Soninceg', 'da': 'soninke', 'de': 'Soninke', 'dsb': 'soninkešćina', 'el': 'Σονίνκε', 'el-polyton': 'Σονίνκε', 'en': 'Soninke', 'es': 'soninké', 'et': 'soninke', 'eu': 'soninkeera', 'fa': 'سونینکه\u200cای', 'ff-Adlm': '𞤅𞤢𞤪𞤢𞤲𞤳𞤵𞤤𞥆𞤪𞤫', 'fi': 'soninke', 'fil': 'Soninke', 'fo': 'soninke', 'fr': 'soninké', 'fy': 'Soninke', 'ga': 'Soinincéis', 'gd': 'Soninke', 'gl': 'soninke', 'gsw': 'Soninke-Schpraach', 'gu': 'સોનિન્કે', 'ha': 'Soninke', 'he': 'סונינקה', 'hi': 'सोनिन्के', 'hi-Latn': 'Soninke', 'hr': 'soninke', 'hsb': 'soninkešćina', 'ht': 'soninké', 'hu': 'szoninke', 'hy': 'սոնինկե', 'ia': 'soninke', 'id': 'Soninke', 'ig': 'Soninke', 'is': 'sóninke', 'it': 'soninke', 'ja': 'ソニンケ語', 'jv': 'Soninke', 'ka': 'სონინკე', 'kgp': 'sãnĩke', 'kk': 'сонинке тілі', 'kk-Cyrl': 'сонинке тілі', 'km': 'សូនីនគេ', 'kn': 'ಸೋನಿಂಕೆ', 'ko': '소닌케어', 'kok': 'सोनिके', 'kok-Deva': 'सोनिके', 'ks': 'سونِنکیے', 'ks-Arab': 'سونِنکیے', 'ku': 'soninke', 'ky': 'сонинкече', 'lb': 'Soninke-Sprooch', 'lo': 'ໂຊນິນກີ', 'lt': 'soninke', 'lv': 'soninku', 'mai': 'सोनिनके', 'mi': 'Honīke', 'mk': 'сонинке', 'ml': 'സോണിൻകെ', 'mn': 'сонинке', 'mr': 'सोनिन्के', 'ms': 'Soninke', 'mt': 'Soninke', 'my': 'ဆိုနင်ကေး', 'nb': 'soninke', 'ne': 'सोनिन्के', 'nl': 'Soninke', 'nn': 'soninke', 'no': 'soninke', 'or': 'ସୋନିଙ୍କେ', 'pa': 'ਸੋਨਿੰਕੇ', 'pa-Guru': 'ਸੋਨਿੰਕੇ', 'pcm': 'Sonínkẹ́ Lángwej', 'pl': 'soninke', 'ps': 'سونینګ', 'pt': 'soninquê', 'qu': 'Soninke Simi', 'rm': 'soninke', 'ro': 'soninke', 'ru': 'сонинке', 'sc': 'soninke', 'sd': 'سونينڪي', 'sd-Arab': 'سونينڪي', 'si': 'සොනින්කෙ', 'sk': 'soninke', 'sl': 'soninke', 'smn': 'soninke', 'so': 'Soninke', 'sq': 'soninkisht', 'sr': 'сонинке', 'sr-Cyrl': 'сонинке', 'sr-Latn': 'soninke', 'sv': 'soninke', 'sw': 'Kisoninke', 'ta': 'சோனின்கே', 'te': 'సోనింకి', 'th': 'โซนีนเก', 'ti': 'ሶኒንከ', 'tk': 'soninke dili', 'to': 'lea fakasoninekē', 'tr': 'Soninke', 'ug': 'سونىنكەچە', 'uk': 'сонінке', 'ur': 'سوننکے', 'uz': 'soninke', 'uz-Latn': 'soninke', 'vec': 'soninké', 'vi': 'Tiếng Soninke', 'xnr': 'सोनिन्के', 'yo': 'Èdè Sonike', 'yrl': 'sunĩkê', 'yue': '索尼基文', 'yue-Hans': '索尼基文', 'yue-Hant': '索尼基文', 'zh': '索宁克语', 'zh-Hans': '索宁克语', 'zh-Hant': '索尼基文', 'zu': 'isi-Soninke'},
    'snl': {'en': 'Sangil'},
    'snm': {'en': "Southern Ma'di"},
    'snn': {'en': 'Siona'},
    'sno': {'en': 'Snohomish'},
    'snp': {'en': 'Siane'},
    'snq': {'en': 'Sangu (Gabon)'},
    'snr': {'en': 'Sihan'},
    'sns': {'en': 'South West Bay'},
    'snu': {'en': 'Senggi'},
    'snv': {'en': "Sa'ban"},
    'snw': {'en': 'Selee'},
    'snx': {'en': 'Sam'},
    'sny': {'en': 'Saniyo-Hiyewe'},
    'snz': {'en': 'Kou'},
    'so': {'af': 'Somalies', 'agq': 'Sòmalì', 'ak': 'Somalia kasa', 'am': 'ሱማልኛ', 'ar': 'الصومالية', 'as': 'ছোমালি', 'asa': 'Kithomali', 'ast': 'somalín', 'az': 'somali', 'az-Cyrl': 'сомали', 'az-Latn': 'somali', 'bal-Latn': 'Sómáli', 'bas': 'Hɔp u somàlî', 'be': 'самалі', 'be-tarask': 'самалі', 'bem': 'Ichi Somalia', 'bez': 'Hisomali', 'bg': 'сомалийски', 'bgn': 'سومالی', 'bm': 'somalikan', 'bn': 'সোমালি', 'br': 'somali', 'brx': 'स’माली', 'bs': 'somalski', 'bs-Cyrl': 'сомалски', 'bs-Latn': 'somalski', 'ca': 'somali', 'ca-ES-valencia': 'somali', 'ccp': '𑄥𑄮𑄟𑄣𑄨', 'ce': 'сомали', 'cgg': 'Orusomaari', 'chr': 'ᏐᎹᎵ', 'ckb': 'سۆمالی', 'cs': 'somálština', 'cy': 'Somaleg', 'da': 'somali', 'dav': 'Kisomali', 'de': 'Somali', 'dje': 'Somaali senni', 'dsb': 'somališćina', 'dyo': 'somali', 'dz': 'སོ་མ་ལི་ཁ', 'ebu': 'Kĩsomali', 'ee': 'somaliagbe', 'el': 'Σομαλικά', 'el-polyton': 'Σομαλικά', 'en': 'Somali', 'eo': 'somala', 'es': 'somalí', 'et': 'somaali', 'eu': 'somaliera', 'ewo': 'ǹkɔ́bɔ somália', 'fa': 'سومالیایی', 'ff': 'Somalii', 'ff-Adlm': '𞤅𞤮𞤥𞤢𞤤𞤭𞥅𞤪𞤫', 'ff-Latn': 'Somalii', 'fi': 'somali', 'fil': 'Somali', 'fo': 'somaliskt', 'fr': 'somali', 'fur': 'somal', 'fy': 'Somalysk', 'ga': 'Somáilis', 'gd': 'Somàilis', 'gl': 'somalí', 'gsw': 'Somali', 'gu': 'સોમાલી', 'guz': 'Kisomali', 'ha': 'Somalianci', 'he': 'סומלית', 'hi': 'सोमाली', 'hi-Latn': 'Somali', 'hr': 'somalski', 'hsb': 'somališćina', 'ht': 'somali', 'hu': 'szomáli', 'hy': 'սոմալիերեն', 'ia': 'somali', 'id': 'Somalia', 'ig': 'Somali', 'is': 'sómalska', 'it': 'somalo', 'ja': 'ソマリ語', 'jmc': 'Kyisomalyi', 'jv': 'Somalia', 'ka': 'სომალიური', 'kab': 'Taṣumalit', 'kam': 'Kisomali', 'kde': 'Chisomali', 'kea': 'somali', 'kgp': 'somỹri', 'khq': 'Somaali senni', 'ki': 'Kĩcumarĩ', 'kk': 'сомали тілі', 'kk-Cyrl': 'сомали тілі', 'kln': 'kutitab Somaliek', 'km': 'សូម៉ាលី', 'kn': 'ಸೊಮಾಲಿ', 'ko': '소말리아어', 'kok': 'सोमाली', 'kok-Deva': 'सोमाली', 'ks': 'سومٲلی', 'ks-Arab': 'سومٲلی', 'ksb': 'Kisomali', 'ksf': 'risomalí', 'ksh': 'Somahlesch', 'ku': 'somalî', 'ky': 'сомаличе', 'lag': 'Kɨsómáali', 'lb': 'Somali', 'lg': 'Lusomaliya', 'lkt': 'Somali Iyápi', 'lld': 'somalich', 'ln': 'lisomali', 'lo': 'ໂຊມາລີ', 'lrc': 'سوٙمالی', 'lt': 'somaliečių', 'lu': 'Lisomali', 'luo': 'Kisomali', 'luy': 'Kisomali', 'lv': 'somāļu', 'mai': 'सोमाली', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lchumari', 'mer': 'Kĩsomali', 'mfe': 'somali', 'mg': 'Somalianina', 'mgh': 'Isomali', 'mi': 'Hamāri', 'mk': 'сомалиски', 'ml': 'സോമാലി', 'mn': 'сомали', 'mr': 'सोमाली', 'ms': 'Somali', 'mt': 'Somali', 'mua': 'Somaliya', 'my': 'ဆိုမာလီ', 'mzn': 'سومالیایی', 'naq': 'Somaliǁî gowab', 'nb': 'somali', 'nd': 'isi-Somali', 'ne': 'सोमाली', 'nl': 'Somalisch', 'nmg': 'Kiɛl somaliā', 'nn': 'somali', 'no': 'somali', 'nqo': 'ߛߏߡߊߟߌߞߊ߲', 'nus': 'Thok thomaaliani', 'nyn': 'Orusomaari', 'or': 'ସୋମାଲିଆ', 'pa': 'ਸੋਮਾਲੀ', 'pa-Guru': 'ਸੋਮਾਲੀ', 'pcm': 'Sọmáli', 'pl': 'somalijski', 'ps': 'سومالي', 'pt': 'somali', 'qu': 'Somali Simi', 'rm': 'somali', 'rn': 'Igisomali', 'ro': 'somaleză', 'rof': 'Kisomali', 'ru': 'сомали', 'rw': 'Igisomali', 'rwk': 'Kyisomalyi', 'saq': 'Kisomali', 'sbp': 'Ishisomali', 'sc': 'sòmalu', 'sd': 'سومالي', 'sd-Arab': 'سومالي', 'seh': 'somali', 'ses': 'Somaali senni', 'sg': 'Somalïi', 'shi': 'ⵜⴰⵙⵓⵎⴰⵍⵉⵜ', 'shi-Latn': 'Tasumalit', 'shi-Tfng': 'ⵜⴰⵙⵓⵎⴰⵍⵉⵜ', 'si': 'සෝමාලි', 'sk': 'somálčina', 'sl': 'somalščina', 'smn': 'somalikielâ', 'sn': 'chiSomali', 'so': 'Soomaali', 'sq': 'somalisht', 'sr': 'сомалски', 'sr-Cyrl': 'сомалски', 'sr-Latn': 'somalski', 'sv': 'somaliska', 'sw': 'Kisomali', 'ta': 'சோமாலி', 'te': 'సోమాలి', 'teo': 'Kisomali', 'tg': 'сомалӣ', 'th': 'โซมาลี', 'ti': 'ሶማሊ', 'tk': 'somali dili', 'to': 'lea fakasomali', 'tr': 'Somalice', 'tt': 'сомали', 'twq': 'Somaali senni', 'tzm': 'Taṣumalit', 'ug': 'سومالىچە', 'uk': 'сомалі', 'ur': 'صومالی', 'uz': 'somalicha', 'uz-Cyrl': 'сомалича', 'uz-Latn': 'somalicha', 'vai': 'ꖇꕮꔷ', 'vai-Latn': 'Somáli', 'vai-Vaii': 'ꖇꕮꔷ', 'vec': 'sòmalo', 'vi': 'Tiếng Somali', 'vun': 'Kyisomalyi', 'wae': 'Somališ', 'wo': 'Somali (làkk)', 'xnr': 'सोमाली', 'xog': 'Olusomaliya', 'yav': 'nusomalíɛ', 'yi': 'סאמאַליש', 'yo': 'Èdè ara Somalia', 'yrl': 'sumari', 'yue': '索馬利文', 'yue-Hans': '索马利文', 'yue-Hant': '索馬利文', 'zgh': 'ⵜⴰⵙⵓⵎⴰⵍⵉⵜ', 'zh': '索马里语', 'zh-Hans': '索马里语', 'zh-Hant': '索馬利文', 'zu': 'isi-Somali'},
    'soa': {'en': 'Thai Song'},
    'sob': {'en': 'Sobei'},
    'soc': {'en': 'So (Democratic Republic of Congo)'},
    'sod': {'en': 'Songoora'},
    'soe': {'en': 'Songomeno'},
    'sog': {'ar': 'السوجدين', 'ast': 'sogdianu', 'az': 'soqdiyen', 'az-Latn': 'soqdiyen', 'bg': 'согдийски', 'bn': 'সোগডিয়ান', 'br': 'sogdieg', 'brx': 'सोगडीयन', 'bs': 'sogdien', 'bs-Cyrl': 'соџијенски', 'bs-Latn': 'sogdien', 'ca': 'sogdià', 'ca-ES-valencia': 'sogdià', 'ccp': '𑄥𑄮𑄇𑄴𑄓𑄠𑄚𑄴', 'cs': 'sogdština', 'cy': 'Sogdeg', 'da': 'sogdiansk', 'de': 'Sogdisch', 'el': 'Σογκντιέν', 'el-polyton': 'Σογκντιέν', 'en': 'Sogdien', 'es': 'sogdiano', 'et': 'sogdi', 'fa': 'سغدی', 'fi': 'sogdi', 'fr': 'sogdien', 'fy': 'Sogdysk', 'ga': 'Sogdánais', 'gsw': 'Sogdisch', 'gu': 'સોગ્ડિએન', 'he': 'סוגדיאן', 'hi': 'सोग्डिएन', 'hi-Latn': 'Sogdien', 'hr': 'sogdien', 'ht': 'sogdien', 'hu': 'sogdien', 'id': 'Sogdien', 'is': 'sogdíen', 'it': 'sogdiano', 'ja': 'ソグド語', 'kgp': 'sognijỹnũ', 'kn': 'ಸೋಗ್ಡಿಯನ್', 'ko': '소그디엔어', 'ks': 'سوگڈِیَن', 'ks-Arab': 'سوگڈِیَن', 'lb': 'Sogdesch', 'lo': 'ຊອກດິນ', 'lt': 'sogdien', 'lv': 'sogdiešu', 'mk': 'зогдијански', 'ml': 'സോജിഡിയൻ', 'mr': 'सोग्डिएन', 'mt': 'Sogdien', 'nb': 'sogdisk', 'nl': 'Sogdisch', 'nn': 'sogdisk', 'no': 'sogdisk', 'or': 'ସୋଗଡିଏନ୍', 'pl': 'sogdyjski', 'pt': 'sogdiano', 'rm': 'sogdian', 'ro': 'sogdien', 'ru': 'согдийский', 'sk': 'sogdijčina', 'sr': 'согдијски', 'sr-Cyrl': 'согдијски', 'sr-Latn': 'sogdijski', 'sv': 'sogdiska', 'ta': 'சோக்தியன்', 'te': 'సోగ్డియన్', 'th': 'ซอกดีน', 'to': 'lea fakasokitiana', 'tr': 'Sogdiana Dili', 'ug': 'سوغدىچە', 'uk': 'согдійська', 'vi': 'Tiếng Sogdien', 'yrl': 'sugidianu', 'yue': '索格底亞納文', 'yue-Hans': '索格底亚纳文', 'yue-Hant': '索格底亞納文', 'zh': '粟特语', 'zh-Hans': '粟特语', 'zh-Hant': '粟特文'},
    'soh': {'en': 'Aka'},
    'soi': {'en': 'Sonha'},
    'soj': {'en': 'Soi'},
    'sok': {'en': 'Sokoro'},
    'sol': {'en': 'Solos'},
    'son': {'en': 'Songhai languages'},
    'soo': {'en': 'Songo'},
    'sop': {'en': 'Songe'},
    'soq': {'en': 'Kanasi'},
    'sor': {'en': 'Somrai'},
    'sos': {'en': 'Seeku'},
    'sou': {'en': 'Southern Thai'},
    'sov': {'en': 'Sonsorol'},
    'sow': {'en': 'Sowanda'},
    'sox': {'en': 'Swo'},
    'soy': {'en': 'Miyobe'},
    'soz': {'en': 'Temi'},
    'spb': {'en': 'Sepa (Indonesia)'},
    'spc': {'en': 'Sapé'},
    'spd': {'en': 'Saep'},
    'spe': {'en': 'Sepa (Papua New Guinea)'},
    'spg': {'en': 'Sian'},
    'spi': {'en': 'Saponi'},
    'spk': {'en': 'Sengo'},
    'spl': {'en': 'Selepet'},
    'spm': {'en': 'Akukem'},
    'spn': {'en': 'Sanapaná'},
    'spo': {'en': 'Spokane'},
    'spp': {'en': 'Supyire Senoufo'},
    'spq': {'en': 'Loreto-Ucayali Spanish'},
    'spr': {'en': 'Saparua'},
    'sps': {'en': 'Saposa'},
    'spt': {'en': 'Spiti Bhoti'},
    'spu': {'en': 'Sapuan'},
    'spv': {'en': 'Sambalpuri'},
    'spx': {'en': 'South Picene'},
    'spy': {'en': 'Sabaot'},
    'sq': {'af': 'Albanees', 'ak': 'Aabeniani', 'am': 'አልባንያንኛ', 'ar': 'الألبانية', 'as': 'আলবেনীয়', 'ast': 'albanu', 'az': 'alban', 'az-Arab': 'آلبانیاجا', 'az-Cyrl': 'албан', 'az-Latn': 'alban', 'bal-Latn': 'Albániái', 'be': 'албанская', 'be-tarask': 'албанская', 'bg': 'албански', 'bgn': 'البانی', 'bn': 'আলবেনীয়', 'br': 'albaneg', 'brx': 'आल्बानियान', 'bs': 'albanski', 'bs-Cyrl': 'албански', 'bs-Latn': 'albanski', 'ca': 'albanès', 'ca-ES-valencia': 'albanès', 'ccp': '𑄃𑄣𑄴𑄝𑄬𑄚𑄩𑄠𑄧', 'ce': 'албанойн', 'chr': 'ᎠᎵᏇᏂ', 'ckb': 'ئەڵبانی', 'cs': 'albánština', 'cy': 'Albaneg', 'da': 'albansk', 'de': 'Albanisch', 'dsb': 'albanšćina', 'dz': 'ཨཱལ་བེ་ནི་ཡཱན་ཁ', 'ee': 'albaniagbe', 'el': 'Αλβανικά', 'el-polyton': 'Ἀλβανικά', 'en': 'Albanian', 'eo': 'albana', 'es': 'albanés', 'et': 'albaania', 'eu': 'albaniera', 'fa': 'آلبانیایی', 'ff-Adlm': '𞤀𞤤𞤦𞤢𞤲𞤭𞥅𞤪𞤫', 'fi': 'albania', 'fil': 'Albanian', 'fo': 'albanskt', 'fr': 'albanais', 'fur': 'albanês', 'fy': 'Albaneesk', 'ga': 'Albáinis', 'gd': 'Albàinis', 'gl': 'albanés', 'gsw': 'Albanisch', 'gu': 'અલ્બેનિયન', 'ha': 'Albaniyanci', 'he': 'אלבנית', 'hi': 'अल्बानियाई', 'hi-Latn': 'Albanian', 'hr': 'albanski', 'hsb': 'albanšćina', 'ht': 'albanais', 'hu': 'albán', 'hy': 'ալբաներեն', 'ia': 'albanese', 'id': 'Albania', 'ig': 'Asụsụ Albania', 'is': 'albanska', 'it': 'albanese', 'ja': 'アルバニア語', 'jv': 'Albania', 'ka': 'ალბანური', 'kea': 'albanes', 'kgp': 'armánẽj', 'kk': 'албан тілі', 'kk-Cyrl': 'албан тілі', 'km': 'អាល់បានី', 'kn': 'ಅಲ್ಬೇನಿಯನ್', 'ko': '알바니아어', 'kok': 'आल्बेनियन्', 'kok-Deva': 'आल्बेनियन्', 'ks': 'البانِیَن', 'ks-Arab': 'البانِیَن', 'ksh': 'Albahnesch', 'ku': 'arnawidî', 'kxv': 'albaniyati', 'kxv-Deva': 'अल्बानियाति', 'kxv-Latn': 'albaniyati', 'kxv-Orya': 'ଆଲବାନିଆନ୍ ତି', 'kxv-Telu': 'అల్బేనియన్', 'ky': 'албанча', 'lb': 'Albanesch', 'lkt': 'Albanian Iyápi', 'lld': 'albanesc', 'lo': 'ອານບານຽນ', 'lrc': 'آلبانی', 'lt': 'albanų', 'lv': 'albāņu', 'mai': 'अलबानिया', 'mi': 'Arapeiniana', 'mk': 'албански', 'ml': 'അൽബേനിയൻ', 'mn': 'албани', 'mr': 'अल्बानियन', 'ms': 'Albania', 'mt': 'Albaniż', 'my': 'အယ်လ်ဘေးနီးယား', 'mzn': 'آلبانیایی', 'nb': 'albansk', 'ne': 'अल्बानियाली', 'nl': 'Albanees', 'nn': 'albansk', 'no': 'albansk', 'nqo': 'ߊߟߑߓߊߣߌߞߊ߲', 'om': 'Afaan Albaniyaa', 'or': 'ଆଲବାନିଆନ୍', 'pa': 'ਅਲਬਾਨੀਆਈ', 'pa-Guru': 'ਅਲਬਾਨੀਆਈ', 'pcm': 'Albéniá Lángwej', 'pl': 'albański', 'ps': 'الباني', 'pt': 'albanês', 'qu': 'Albanes Simi', 'rm': 'albanais', 'ro': 'albaneză', 'ru': 'албанский', 'rw': 'Icyalubaniya', 'sah': 'Албаанныы', 'sat': 'ᱟᱞᱵᱟᱱᱤᱭᱟᱹᱱ', 'sat-Olck': 'ᱟᱞᱵᱟᱱᱤᱭᱟᱹᱱ', 'sc': 'albanesu', 'sd': 'الباني', 'sd-Arab': 'الباني', 'se': 'albánagiella', 'si': 'ඇල්බේනියානු', 'sk': 'albánčina', 'sl': 'albanščina', 'smn': 'albaniakielâ', 'so': 'Albeeniyaan', 'sq': 'shqip', 'sr': 'албански', 'sr-Cyrl': 'албански', 'sr-Latn': 'albanski', 'sv': 'albanska', 'sw': 'Kialbania', 'syr': 'ܐܠܒܢܝܬ', 'ta': 'அல்பேனியன்', 'te': 'అల్బేనియన్', 'tg': 'албанӣ', 'th': 'แอลเบเนีย', 'ti': 'ኣልባንኛ', 'tk': 'alban dili', 'to': 'lea fakaʻalapēnia', 'tr': 'Arnavutça', 'tt': 'албан', 'ug': 'ئالبانچە', 'uk': 'албанська', 'ur': 'البانی', 'uz': 'alban', 'uz-Cyrl': 'албанча', 'uz-Latn': 'alban', 'vec': 'albaneze', 'vi': 'Tiếng Albania', 'wae': 'Albaniš', 'wo': 'Albane', 'xh': 'IsiAlbania', 'xnr': 'अल्बानियाई', 'yi': 'אַלבאַניש', 'yo': 'Èdè Albania', 'yrl': 'aubanei', 'yue': '阿爾巴尼亞文', 'yue-Hans': '阿尔巴尼亚文', 'yue-Hant': '阿爾巴尼亞文', 'zh': '阿尔巴尼亚语', 'zh-Hans': '阿尔巴尼亚语', 'zh-Hant': '阿爾巴尼亞文', 'zu': 'isi-Albania'},
    'sqa': {'en': 'Shama-Sambuga'},
    'sqh': {'en': 'Shau'},
    'sqj': {'en': 'Albanian languages'},
    'sqj-pro': {'en': 'Proto-Albanian'},
    'sqk': {'en': 'Albanian Sign Language'},
    'sqm': {'en': 'Suma'},
    'sqn': {'en': 'Susquehannock'},
    'sqo': {'en': 'Sorkhei'},
    'sqq': {'en': 'Sou'},
    'sqr': {'en': 'Siculo Arabic'},
    'sqs': {'en': 'Sri Lankan Sign Language'},
    'sqt': {'en': 'Soqotri'},
    'squ': {'en': 'Squamish'},
    'sqx': {'en': 'Kufr Qassem Sign Language (KQSL)'},
    'sr': {'af': 'Serwies', 'ak': 'Sɛbia Kasa', 'am': 'ሰርብያኛ', 'ar': 'الصربية', 'as': 'ছাৰ্বিয়ান', 'ast': 'serbiu', 'az': 'serb', 'az-Arab': 'صربجه', 'az-Cyrl': 'серб', 'az-Latn': 'serb', 'bal-Latn': 'Sarbiái', 'be': 'сербская', 'be-tarask': 'сербская', 'bg': 'сръбски', 'bgn': 'سیربی', 'bn': 'সার্বীয়', 'br': 'serbeg', 'brx': 'सार्बियान', 'bs': 'srpski', 'bs-Cyrl': 'српски', 'bs-Latn': 'srpski', 'ca': 'serbi', 'ca-ES-valencia': 'serbi', 'ccp': '𑄥𑄢𑄴𑄝𑄩𑄠𑄧', 'ce': 'сербийн', 'chr': 'ᏒᏈᎠᏂ', 'ckb': 'سربی', 'cs': 'srbština', 'cy': 'Serbeg', 'da': 'serbisk', 'de': 'Serbisch', 'dsb': 'serbišćina', 'dz': 'སཱར་བྷི་ཡཱན་ཁ', 'ee': 'serbiagbe', 'el': 'Σερβικά', 'el-polyton': 'Σερβικά', 'en': 'Serbian', 'eo': 'serba', 'es': 'serbio', 'et': 'serbia', 'eu': 'serbiera', 'fa': 'صربی', 'ff-Adlm': '𞤅𞤫𞤪𞤦𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'serbia', 'fil': 'Serbian', 'fo': 'serbiskt', 'fr': 'serbe', 'fur': 'serp', 'fy': 'Servysk', 'ga': 'Seirbis', 'gd': 'Sèirbis', 'gl': 'serbio', 'gsw': 'Serbisch', 'gu': 'સર્બિયન', 'ha': 'Sabiyan', 'he': 'סרבית', 'hi': 'सर्बियाई', 'hi-Latn': 'Serbian', 'hr': 'srpski', 'hsb': 'serbišćina', 'ht': 'serbe', 'hu': 'szerb', 'hy': 'սերբերեն', 'ia': 'serbo', 'id': 'Serbia', 'ig': 'Asụsụ Serbia', 'is': 'serbneska', 'it': 'serbo', 'ja': 'セルビア語', 'jv': 'Serbia', 'ka': 'სერბული', 'kea': 'sérviu', 'kgp': 'sérvijo', 'kk': 'серб тілі', 'kk-Cyrl': 'серб тілі', 'km': 'ស៊ែប', 'kn': 'ಸೆರ್ಬಿಯನ್', 'ko': '세르비아어', 'kok': 'सर्बियन', 'kok-Deva': 'सर्बियन', 'ks': 'سٔربِیَن', 'ks-Arab': 'سٔربِیَن', 'ksh': 'Särbesch', 'ku': 'sirbî', 'kxv': 'sarbiyati', 'kxv-Deva': 'सर्बियाति', 'kxv-Latn': 'sarbiyati', 'kxv-Orya': 'ସର୍ବିୟ ତି', 'kxv-Telu': 'సెర్బియన్', 'ky': 'сербче', 'lb': 'Serbesch', 'lkt': 'Serbia Iyápi', 'lld': 'serb', 'lo': 'ເຊີບຽນ', 'lrc': 'سئربی', 'lt': 'serbų', 'lv': 'serbu', 'mai': 'सर्बियाई', 'mi': 'Hirupia', 'mk': 'српски', 'ml': 'സെർബിയൻ', 'mn': 'серб', 'mr': 'सर्बियन', 'ms': 'Serbia', 'mt': 'Serb', 'my': 'ဆားဘီးယား', 'mzn': 'صربی', 'nb': 'serbisk', 'ne': 'सर्बियाली', 'nl': 'Servisch', 'nn': 'serbisk', 'no': 'serbisk', 'nqo': 'ߛߍߙߑߓߌߞߊ߲', 'om': 'Afaan Serbiya', 'or': 'ସର୍ବିୟ', 'pa': 'ਸਰਬੀਆਈ', 'pa-Guru': 'ਸਰਬੀਆਈ', 'pcm': 'Sẹrbiá Lángwej', 'pl': 'serbski', 'ps': 'سربيائي', 'pt': 'sérvio', 'qu': 'Serbio Simi', 'rm': 'serb', 'ro': 'sârbă', 'ru': 'сербский', 'rw': 'Igiseribe', 'sc': 'serbu', 'sd': 'سربيائي', 'sd-Arab': 'سربيائي', 'se': 'serbiagiella', 'si': 'සර්බියානු', 'sk': 'srbčina', 'sl': 'srbščina', 'smn': 'serbiakielâ', 'so': 'Seerbiyaan', 'sq': 'serbisht', 'sr': 'српски', 'sr-Cyrl': 'српски', 'sr-Latn': 'srpski', 'sv': 'serbiska', 'sw': 'Kiserbia', 'ta': 'செர்பியன்', 'te': 'సెర్బియన్', 'tg': 'сербӣ', 'th': 'เซอร์เบีย', 'ti': 'ሰርቢያኛ', 'tk': 'serb dili', 'to': 'lea fakasēpia', 'tr': 'Sırpça', 'tt': 'серб', 'ug': 'سېربچە', 'uk': 'сербська', 'ur': 'سربین', 'uz': 'serbcha', 'uz-Cyrl': 'сербча', 'uz-Latn': 'serbcha', 'vec': 'serbo', 'vi': 'Tiếng Serbia', 'wae': 'Serbiš', 'wo': 'Serb', 'xnr': 'सर्बियाई', 'yi': 'סערביש', 'yo': 'Èdè Serbia', 'yrl': 'sériwiu', 'yue': '塞爾維亞文', 'yue-Hans': '塞尔维亚文', 'yue-Hant': '塞爾維亞文', 'zh': '塞尔维亚语', 'zh-Hans': '塞尔维亚语', 'zh-Hant': '塞爾維亞文', 'zu': 'isi-Serbian'},
    'sr-ME': {'en': 'Montenegrin', 'hi-Latn': 'Montenegrin'},
    'sra': {'en': 'Saruga'},
    'srb': {'en': 'Sora'},
    'src': {'en': 'Logudorese Sardinian'},
    'sre': {'en': 'Sara'},
    'srf': {'en': 'Nafi'},
    'srg': {'en': 'Sulod'},
    'srh': {'en': 'Sarikoli'},
    'sri': {'en': 'Siriano'},
    'srk': {'en': 'Serudung Murut'},
    'srl': {'en': 'Isirawa'},
    'srm': {'en': 'Saramaccan'},
    'srn': {'af': 'Sranan Tongo', 'am': 'ስራናን ቶንጎ', 'ar': 'السرانان تونجو', 'as': 'স্ৰানান টোঙ্গো', 'ast': 'sranan tongo', 'az': 'sranan tonqo', 'az-Cyrl': 'сранан тонго', 'az-Latn': 'sranan tonqo', 'be': 'сранан-тонга', 'be-tarask': 'сранан-тонга', 'bg': 'сранан тонго', 'bgn': 'تاکی\u200cتاکی', 'bn': 'স্রানান টোঙ্গো', 'br': 'sranan tongo', 'brx': 'स्रानान टंग’', 'bs': 'srananski tongo', 'bs-Cyrl': 'сранански тонго', 'bs-Latn': 'srananski tongo', 'ca': 'sranan', 'ca-ES-valencia': 'sranan', 'ccp': '𑄥𑄳𑄢𑄚𑄚𑄴 𑄑𑄮𑄋𑄴𑄉𑄮', 'ce': 'сранан-тонго', 'chr': 'ᏏᎳᎾᏂ ᏙᏃᎪ', 'ckb': 'سرانان تۆنگۆ', 'cs': 'sranan tongo', 'cy': 'Sranan Tongo', 'da': 'sranan tongo', 'de': 'Srananisch', 'dsb': 'sranan tongošćina', 'el': 'Σρανάν Τόνγκο', 'el-polyton': 'Σρανάν Τόνγκο', 'en': 'Sranan Tongo', 'es': 'sranan tongo', 'et': 'sranani', 'eu': 'sranan tongoa', 'fa': 'زبان اسرانان', 'ff-Adlm': '𞤅𞤢𞤪𞤲𞤢𞤲-𞤚𞤮𞤲𞤺𞤮𞥅𞤪𞤫', 'fi': 'sranan', 'fil': 'Sranan Tongo', 'fo': 'sranan tongo', 'fr': 'sranan tongo', 'fy': 'Sranantongo', 'ga': 'Suranaimis', 'gd': 'Sranan Tongo', 'gl': 'sranan tongo', 'gsw': 'Srananisch', 'gu': 'સ્રાનન ટોન્ગો', 'ha': 'Sranan Tongo', 'he': 'סרנאן טונגו', 'hi': 'स्रानान टॉन्गो', 'hi-Latn': 'Sranan Tongo', 'hr': 'sranan tongo', 'hsb': 'sranan tongošćina', 'ht': 'sranan tongo', 'hu': 'szranai tongó', 'hy': 'սրանան տոնգո', 'ia': 'sranan tongo', 'id': 'Sranan Tongo', 'ig': 'Sranan Tongo', 'is': 'sranan tongo', 'it': 'sranan tongo', 'ja': 'スリナム語', 'jv': 'Sranan Tongo', 'ka': 'სრანან ტონგო', 'kgp': 'surinỹmẽj', 'kk': 'сранан тонго тілі', 'kk-Cyrl': 'сранан тонго тілі', 'km': 'ស្រាណានតុងហ្គោ', 'kn': 'ಸ್ರಾನನ್ ಟೋಂಗೋ', 'ko': '스라난 통가어', 'kok': 'श्रानन टोंगो', 'kok-Deva': 'श्रानन टोंगो', 'ks': 'سرانَن ٹونگو', 'ks-Arab': 'سرانَن ٹونگو', 'ku': 'sirananî', 'ky': 'сранан тонгочо', 'lb': 'Srananesch', 'lo': 'ສຣານນານຕອນໂກ', 'lt': 'sranan tongo', 'lv': 'sranantogo', 'mai': 'श्रेनन टोंगो', 'mi': 'Harāna Tongo', 'mk': 'срански тонго', 'ml': 'ശ്രാനൻ ഡോങ്കോ', 'mn': 'сранан тонго', 'mr': 'स्रानान टॉन्गो', 'ms': 'Sranan Tongo', 'mt': 'Sranan Tongo', 'my': 'ဆရာနန် တွန်ဂို', 'nb': 'sranan', 'ne': 'स्रानान टोङ्गो', 'nl': 'Sranantongo', 'nn': 'sranan tongo', 'no': 'sranan', 'or': 'ଶାରାନା ଟୋଙ୍ଗୋ', 'pa': 'ਸ੍ਰਾਨਾਨ ਟੋਂਗੋ', 'pa-Guru': 'ਸ੍ਰਾਨਾਨ ਟੋਂਗੋ', 'pcm': 'Sranán Tóngo', 'pl': 'sranan tongo', 'ps': 'سوران ټونګو', 'pt': 'surinamês', 'qu': 'Sranan Tongo Simi', 'rm': 'sranan tongo', 'ro': 'sranan tongo', 'ru': 'сранан-тонго', 'sc': 'sranan tongo', 'sd': 'سرانن تانگو', 'sd-Arab': 'سرانن تانگو', 'si': 'ස්\u200dරන් ටොන්ගො', 'sk': 'surinamčina', 'sl': 'surinamska kreolščina', 'smn': 'sranantongo', 'so': 'Sranan Tongo', 'sq': 'srananisht (sranantongoisht)', 'sr': 'сранан тонго', 'sr-Cyrl': 'сранан тонго', 'sr-Latn': 'sranan tongo', 'sv': 'sranan tongo', 'sw': 'Lugha ya Sranan Tongo', 'ta': 'ஸ்ரானன் டோங்கோ', 'te': 'స్రానన్ టోంగో', 'th': 'สรานานตองโก', 'ti': 'ስራናን ቶንጎ', 'tk': 'sranan-tongo dili', 'to': 'lea fakasulanane-tongikō', 'tr': 'Sranan Tongo', 'ug': 'سىرانان-توڭوچە', 'uk': 'сранан тонго', 'ur': 'سرانن ٹونگو', 'uz': 'sranan-tongo', 'uz-Latn': 'sranan-tongo', 'vec': 'sranan tongo', 'vi': 'Tiếng Sranan Tongo', 'xnr': 'स्रानान टॉन्गो', 'yo': 'Èdè Sirana Tongo', 'yrl': 'surinamei', 'yue': '蘇拉南東墎文', 'yue-Hans': '苏拉南东墎文', 'yue-Hant': '蘇拉南東墎文', 'zh': '苏里南汤加语', 'zh-Hans': '苏里南汤加语', 'zh-Hant': '蘇拉南東墎文', 'zu': 'isi-Sranan Tongo'},
    'sro': {'en': 'Campidanese Sardinian'},
    'srq': {'en': 'Sirionó'},
    'srr': {'ar': 'السرر', 'ast': 'serer', 'az': 'serer', 'az-Latn': 'serer', 'bg': 'серер', 'bn': 'সেরের', 'br': 'serer', 'brx': 'सेरेर', 'bs': 'serer', 'bs-Cyrl': 'серер', 'bs-Latn': 'serer', 'ca': 'serer', 'ca-ES-valencia': 'serer', 'ccp': '𑄥𑄬𑄢𑄬𑄢𑄴', 'cs': 'sererština', 'cy': 'Serereg', 'da': 'serer', 'de': 'Serer', 'el': 'Σερέρ', 'el-polyton': 'Σερέρ', 'en': 'Serer', 'es': 'serer', 'et': 'sereri', 'fa': 'سریری', 'fi': 'serer', 'fr': 'sérère', 'fy': 'Serer', 'gd': 'Serer', 'gsw': 'Serer-Schpraach', 'gu': 'સેરેર', 'he': 'סרר', 'hi': 'सेरेर', 'hi-Latn': 'Serer', 'hr': 'serer', 'ht': 'sérère', 'hu': 'szerer', 'id': 'Serer', 'is': 'serer', 'it': 'serer', 'ja': 'セレル語', 'kgp': 'serere', 'kn': 'ಸೇರೇರ್', 'ko': '세레르어', 'ks': 'سیریر', 'ks-Arab': 'سیریر', 'lb': 'Serer-Sprooch', 'lo': 'ເຊເລີ', 'lt': 'sererų', 'lv': 'serēru', 'mk': 'серер', 'ml': 'സെറർ', 'mr': 'सेरेर', 'mt': 'Serer', 'nb': 'serer', 'nl': 'Serer', 'nn': 'serer', 'no': 'serer', 'or': 'ଶେରେର୍', 'pl': 'serer', 'pt': 'serere', 'rm': 'serer', 'ro': 'serer', 'ru': 'серер', 'sk': 'sererčina', 'sl': 'sererščina', 'sr': 'серерски', 'sr-Cyrl': 'серерски', 'sr-Latn': 'sererski', 'sv': 'serer', 'ta': 'செரெர்', 'te': 'సెరేర్', 'th': 'เซแรร์', 'to': 'lea fakasēlēle', 'tr': 'Serer dili', 'ug': 'سېرېرچە', 'uk': 'серер', 'vi': 'Tiếng Serer', 'yrl': 'serere', 'yue': '塞雷爾文', 'yue-Hans': '塞雷尔文', 'yue-Hant': '塞雷爾文', 'zh': '塞雷尔语', 'zh-Hans': '塞雷尔语', 'zh-Hant': '塞雷爾文'},
    'srs': {'en': 'Sarsi'},
    'srt': {'en': 'Sauri'},
    'sru': {'en': 'Suruí'},
    'srv': {'en': 'Southern Sorsoganon'},
    'srw': {'en': 'Serua'},
    'srx': {'en': 'Sirmauri'},
    'sry': {'en': 'Sera'},
    'srz': {'en': 'Shahmirzadi'},
    'ss': {'af': 'Swazi', 'am': 'ስዋቲ', 'ar': 'السواتي', 'as': 'স্বাতি', 'ast': 'swati', 'az': 'svati', 'az-Cyrl': 'свати', 'az-Latn': 'svati', 'bal-Latn': 'Swáti', 'be': 'суаці', 'be-tarask': 'суаці', 'bg': 'свати', 'bgn': 'سواتی', 'bn': 'সোয়াতি', 'br': 'swati', 'brx': 'स्वाटि', 'bs': 'svati', 'bs-Cyrl': 'свати', 'bs-Latn': 'svati', 'ca': 'swazi', 'ca-ES-valencia': 'swazi', 'ccp': '𑄥𑄮𑄠𑄖𑄨', 'ce': 'свази', 'chr': 'ᏍᏩᏘ', 'ckb': 'سواتی', 'cs': 'siswatština', 'cy': 'Swati', 'da': 'swati', 'de': 'Swazi', 'dsb': 'siswati', 'ee': 'swatgbe', 'el': 'Σουάτι', 'el-polyton': 'Σουάτι', 'en': 'Swati', 'eo': 'svazia', 'es': 'suazi', 'et': 'svaasi', 'eu': 'swatiera', 'fa': 'سوازیایی', 'ff-Adlm': '𞤅𞤵𞤱𞤢𞤼𞤭𞥅𞤪𞤫', 'fi': 'swazi', 'fil': 'Swati', 'fo': 'swatiskt', 'fr': 'swati', 'fur': 'swati', 'fy': 'Swazi', 'ga': 'Suaisis', 'gd': 'Swati', 'gl': 'suazi', 'gsw': 'Swazi', 'gu': 'સ્વાતી', 'ha': 'Swati', 'he': 'סאווזי', 'hi': 'स्वाती', 'hi-Latn': 'Swati', 'hr': 'svati', 'hsb': 'siswati', 'ht': 'swati', 'hu': 'sziszuati', 'hy': 'սվազերեն', 'ia': 'swati', 'id': 'Swati', 'ig': 'Swati', 'is': 'svatí', 'it': 'swati', 'ja': 'スワジ語', 'jv': 'Swati', 'ka': 'სუატი', 'kgp': 'suvaji', 'kk': 'свати тілі', 'kk-Cyrl': 'свати тілі', 'km': 'ស្វាទី', 'kn': 'ಸ್ವಾತಿ', 'ko': '시스와티어', 'kok': 'स्वाती', 'kok-Deva': 'स्वाती', 'ks': 'سواتی', 'ks-Arab': 'سواتی', 'ksh': 'Si-Swatesch', 'ku': 'swazî', 'ky': 'сватиче', 'lb': 'Swazi', 'lld': 'swati', 'lo': 'ສະວາຕິ', 'lt': 'svatų', 'lv': 'svatu', 'mai': 'स्वाती', 'mi': 'Wāti', 'mk': 'свати', 'ml': 'സ്വാറ്റി', 'mn': 'свати', 'mr': 'स्वाती', 'ms': 'Swati', 'mt': 'Swati', 'my': 'ဆွာဇီလန်', 'nb': 'swati', 'ne': 'स्वाती', 'nl': 'Swazi', 'nn': 'swati', 'no': 'swati', 'or': 'ସ୍ଵାତି', 'pa': 'ਸਵਾਤੀ', 'pa-Guru': 'ਸਵਾਤੀ', 'pcm': 'Swáti Lángwej', 'pl': 'suazi', 'ps': 'سواتی', 'pt': 'suázi', 'qu': 'Swati Simi', 'rm': 'swazi', 'ro': 'swati', 'ru': 'свази', 'sc': 'swati', 'sd': 'سواتي', 'sd-Arab': 'سواتي', 'si': 'ස්වති', 'sk': 'svazijčina', 'sl': 'svazijščina', 'smn': 'swazikielâ', 'so': 'Swati', 'sq': 'suatisht', 'sr': 'свази', 'sr-Cyrl': 'свази', 'sr-Latn': 'svazi', 'ss': 'siSwati', 'sv': 'swati', 'sw': 'Kiswati', 'ta': 'ஸ்வாடீ', 'te': 'స్వాతి', 'th': 'สวาติ', 'ti': 'ስዋዚ', 'tk': 'swati dili', 'to': 'lea fakasuati', 'tr': 'Sisvati', 'ug': 'سىۋاتىچە', 'uk': 'сисваті', 'ur': 'سواتی', 'uz': 'svati', 'uz-Cyrl': 'свати', 'uz-Latn': 'svati', 'vec': 'zwati', 'vi': 'Tiếng Swati', 'wae': 'Swazi', 'xnr': 'स्वाती', 'yo': 'Èdè Suwati', 'yrl': 'suási', 'yue': '斯瓦特文', 'yue-Hans': '斯瓦特文', 'yue-Hant': '斯瓦特文', 'zh': '斯瓦蒂语', 'zh-Hans': '斯瓦蒂语', 'zh-Hant': '斯瓦特文', 'zu': 'isiSwati'},
    'ssa': {'en': 'Nilo-Saharan languages'},
    'ssb': {'en': 'Southern Sama'},
    'ssc': {'en': 'Suba-Simbiti'},
    'ssd': {'en': 'Siroi'},
    'sse': {'en': 'Balangingi'},
    'ssf': {'en': 'Thao'},
    'ssg': {'en': 'Seimat'},
    'ssh': {'en': 'Shihhi Arabic'},
    'ssi': {'en': 'Sansi'},
    'ssj': {'en': 'Sausi'},
    'ssk': {'en': 'Sunam'},
    'ssl': {'en': 'Western Sisaala'},
    'ssm': {'en': 'Semnam'},
    'ssn': {'en': 'Waata'},
    'sso': {'en': 'Sissano'},
    'ssp': {'en': 'Spanish Sign Language'},
    'ssq': {'en': "So'a"},
    'ssr': {'en': 'Swiss-French Sign Language'},
    'sss': {'en': 'Sô'},
    'sst': {'en': 'Sinasina'},
    'ssu': {'en': 'Susuami'},
    'ssv': {'en': 'Shark Bay'},
    'ssx': {'en': 'Samberigi'},
    'ssy': {'af': 'Saho', 'am': 'ሳሆኛ', 'ar': 'لغة الساهو', 'as': 'ছাহো', 'ast': 'saho', 'az': 'saho', 'az-Cyrl': 'саһо', 'az-Latn': 'saho', 'bal-Latn': 'Sahó', 'be': 'саха', 'be-tarask': 'саха', 'bg': 'сахо', 'bgn': 'ساهویی', 'bn': 'সাহো', 'br': 'sahoeg', 'bs': 'saho', 'bs-Latn': 'saho', 'ca': 'saho', 'ca-ES-valencia': 'saho', 'ccp': '𑄥𑄦𑄮', 'ce': 'сахо', 'chr': 'ᏌᎰ', 'ckb': 'ساهۆ', 'cs': 'saho', 'cy': 'Saho', 'da': 'saho', 'de': 'Saho', 'el': 'Σάχο', 'el-polyton': 'Σάχο', 'en': 'Saho', 'es': 'saho', 'et': 'saho', 'eu': 'sahoa', 'fa': 'ساهو', 'fi': 'saho', 'fil': 'Saho', 'fo': 'saho', 'fr': 'saho', 'fy': 'Saho', 'gd': 'Saho', 'gl': 'saho', 'gu': 'સાહો', 'he': 'סאהו', 'hi': 'साहो', 'hi-Latn': 'Saho', 'hr': 'saho', 'ht': 'saho', 'hu': 'szahó', 'hy': 'սահոերեն', 'ia': 'saho', 'id': 'Saho', 'ig': 'Saho', 'is': 'saho', 'it': 'saho', 'ja': 'サホ語', 'ka': 'საჰო', 'kgp': 'saho', 'kk': 'сахо тілі', 'kk-Cyrl': 'сахо тілі', 'km': 'សាហូ', 'kn': 'ಸಹೊ', 'ko': '사호어', 'kok': 'साहो', 'kok-Deva': 'साहो', 'ku': 'sahoyî', 'ky': 'сахочо', 'lb': 'Saho', 'lo': 'ຊາໂຮ', 'lt': 'saho', 'lv': 'saho', 'mk': 'сахо', 'ml': 'സാഹോ', 'mn': 'сахо', 'mr': 'साहो', 'ms': 'Saho', 'mt': 'Saho', 'my': 'ဆာဟို', 'nb': 'saho', 'ne': 'साहो', 'nl': 'Saho', 'nn': 'saho', 'no': 'saho', 'or': 'ସହୋ', 'pa': 'ਸਾਹੋ', 'pa-Guru': 'ਸਾਹੋ', 'pl': 'saho', 'ps': 'سهو', 'pt': 'saho', 'ro': 'saho', 'ru': 'сахо', 'sd': 'سهو', 'sd-Arab': 'سهو', 'si': 'සහො', 'sk': 'saho', 'sl': 'saho', 'smn': 'saho', 'sq': 'sahoisht', 'sr': 'сахо', 'sr-Cyrl': 'сахо', 'sr-Latn': 'saho', 'ssy': 'Saho', 'sv': 'saho', 'sw': 'Kisaho', 'ta': 'சஹோ', 'te': 'సాహో', 'th': 'ซาโฮ', 'ti': 'ሳሆ', 'tk': 'saho dili', 'to': 'lea fakasaho', 'tr': 'Saho', 'ug': 'ساخوچە', 'uk': 'сахо', 'ur': 'ساہو', 'uz': 'saho', 'uz-Cyrl': 'саҳоча', 'uz-Latn': 'saho', 'vec': 'saho', 'vi': 'Tiếng Saho', 'xnr': 'साहो', 'yrl': 'saru', 'yue': '薩霍文', 'yue-Hans': '萨霍文', 'yue-Hant': '薩霍文', 'zh': '萨霍语', 'zh-Hans': '萨霍语', 'zh-Hant': '薩霍文', 'zu': 'isi-Saho'},
    'ssz': {'en': 'Sengseng'},
    'st': {'af': 'Suid-Sotho', 'am': 'ደቡባዊ ሶቶ', 'ar': 'السوتو الجنوبية', 'as': 'দাক্ষিণাত্য ছোথো', 'ast': 'sotho del sur', 'az': 'sesoto', 'az-Cyrl': 'сесото', 'az-Latn': 'sesoto', 'bal-Latn': 'Zerbári Sutó', 'be': 'сесута', 'be-tarask': 'сесута', 'bg': 'сото', 'bgn': 'جهلسرین سوتویی', 'bn': 'দক্ষিন সোথো', 'br': 'sotho ar Su', 'brx': 'खोलायारि सथ’', 'bs': 'južni soto', 'bs-Cyrl': 'сесото', 'bs-Latn': 'južni soto', 'ca': 'sotho meridional', 'ca-ES-valencia': 'sotho meridional', 'ccp': '𑄘𑄧𑄉𑄨𑄚𑄴 𑄥𑄮𑄗𑄮', 'ce': 'къилба сото', 'chr': 'ᏧᎦᎾᏮ ᏗᏜ ᏐᏠ', 'ckb': 'سۆتۆی باشوور', 'cs': 'sotština (jižní)', 'cy': 'Sesotheg Deheuol', 'da': 'sydsotho', 'de': 'Süd-Sotho', 'dsb': 'pódpołdnjowa sotšćina (Sesotho)', 'ee': 'anyiehe sothogbe', 'el': 'Νότια Σόθο', 'el-polyton': 'Νότια Σόθο', 'en': 'Southern Sotho', 'eo': 'sota', 'es': 'sotho meridional', 'et': 'lõunasotho', 'eu': 'hegoaldeko sothoera', 'fa': 'سوتوی جنوبی', 'ff-Adlm': '𞤅𞤮𞤼𞤮𞥅𞤪𞤫 𞤙𞤢𞥄𞤥𞤲𞤢𞥄𞤲𞤺𞤫𞤲𞤳𞤮', 'fi': 'eteläsotho', 'fil': 'Katimugang Sotho', 'fo': 'sesotho', 'fr': 'sotho du Sud', 'fur': 'sotho meridionâl', 'fy': 'Sûd-Sotho', 'ga': 'Sútúis an Deiscirt', 'gd': 'Sesotho', 'gl': 'sesotho', 'gsw': 'Süüd-Sotho-Schpraach', 'gu': 'દક્ષિણ સોથો', 'ha': 'Sesotanci', 'he': 'סותו דרומית', 'hi': 'दक्षिणी सेसेथो', 'hi-Latn': 'Southern Sotho', 'hr': 'sesotski', 'hsb': 'južnosotšćina (Sesotho)', 'ht': 'sotho du Sud', 'hu': 'déli szeszotó', 'hy': 'հարավային սոթո', 'ia': 'sotho del sud', 'id': 'Sotho Selatan', 'ig': 'Southern Sotho', 'is': 'suðursótó', 'it': 'sotho del sud', 'ja': '南部ソト語', 'jv': 'Sotho Sisih Kidul', 'ka': 'სამხრეთ სოთოს ენა', 'kgp': 'soto sur', 'kk': 'оңтүстік сото тілі', 'kk-Cyrl': 'оңтүстік сото тілі', 'km': 'សូថូខាងត្បូង', 'kn': 'ದಕ್ಷಿಣ ಸೋಥೋ', 'ko': '남부 소토어', 'kok': 'दक्षिणी सोथो', 'kok-Deva': 'दक्षिणी सोथो', 'ks': 'جنوبی ستھو', 'ks-Arab': 'جنوبی ستھو', 'ksh': 'Söd-Sooto', 'ku': 'sotoyîya başûr', 'ky': 'сесоточо', 'lb': 'Süd-Sotho-Sprooch', 'lld': 'sotho dl süd', 'lo': 'ໂຊໂທໃຕ້', 'lt': 'pietų Soto', 'lv': 'dienvidsotu', 'mai': 'दक्षिणी सोथो', 'mi': 'Hōto ki te Tonga', 'mk': 'сесото', 'ml': 'തെക്കൻ സോതോ', 'mn': 'сесото', 'mr': 'दक्षिणी सोथो', 'ms': 'Sotho Selatan', 'mt': 'Soto tan-Nofsinhar', 'my': 'တောင်ပိုင်း ဆိုသို', 'nb': 'sør-sotho', 'ne': 'दक्षिणी सोथो', 'nl': 'Zuid-Sotho', 'nn': 'sørsotho', 'no': 'sør-sotho', 'nqo': 'ߛߕߏ߫ ߥߙߏ߬ߘߎ߰ߞߊ߲', 'or': 'ସେସୋଥୋ', 'pa': 'ਦੱਖਣੀ ਸੋਥੋ', 'pa-Guru': 'ਦੱਖਣੀ ਸੋਥੋ', 'pcm': 'Saútán Sóto', 'pl': 'sotho południowy', 'ps': 'سويلي سوتو', 'pt': 'soto do sul', 'qu': 'Soto Meridional Simi', 'rm': 'sotho dal sid', 'ro': 'sesotho', 'ru': 'южный сото', 'rw': 'Inyesesoto', 'sc': 'sotho meridionale', 'sd': 'ڏکڻ سوٿي', 'sd-Arab': 'ڏکڻ سوٿي', 'si': 'සතර්න් සොතො', 'sk': 'sothčina (južná)', 'sl': 'sesoto', 'smn': 'maadâsotho', 'so': 'Sesooto', 'sq': 'sotoishte jugore', 'sr': 'сесото', 'sr-Cyrl': 'сесото', 'sr-Latn': 'sesoto', 'st': 'Sesotho', 'sv': 'sydsotho', 'sw': 'Kisotho', 'ta': 'தெற்கு ஸோதோ', 'te': 'దక్షిణ సోతో', 'th': 'โซโทใต้', 'ti': 'ደቡባዊ ሶቶ', 'tk': 'günorta soto dili', 'to': 'lea fakasoto-tonga', 'tr': 'Güney Sotho dili', 'ug': 'سوتوچە', 'uk': 'південна сото', 'ur': 'جنوبی سوتھو', 'uz': 'janubiy soto', 'uz-Latn': 'janubiy soto', 'vec': 'soto meridionale', 'vi': 'Tiếng Sotho Miền Nam', 'wae': 'Südsotho', 'xnr': 'दखणे बखें सेसेथो', 'yo': 'Èdè Sesoto', 'yrl': 'sutu suwara', 'yue': '塞索托文', 'yue-Hans': '塞索托文', 'yue-Hant': '塞索托文', 'zh': '南索托语', 'zh-Hans': '南索托语', 'zh-Hant': '塞索托文', 'zu': 'isi-Southern Sotho'},
    'sta': {'en': 'Settla'},
    'stb': {'en': 'Northern Subanen'},
    'std': {'en': 'Sentinel'},
    'ste': {'en': 'Liana-Seti'},
    'stf': {'en': 'Seta'},
    'stg': {'en': 'Trieng'},
    'sth': {'en': 'Shelta'},
    'sti': {'en': 'Bulo Stieng'},
    'stj': {'en': 'Matya Samo'},
    'stk': {'en': 'Arammba'},
    'stl': {'en': 'Stellingwerfs'},
    'stm': {'en': 'Setaman'},
    'stn': {'en': 'Owa'},
    'sto': {'en': 'Stoney'},
    'stp': {'en': 'Southeastern Tepehuan'},
    'stq': {'ast': 'frisón de Saterland', 'cs': 'fríština (saterlandská)', 'cy': 'Ffriseg Saterland', 'de': 'Saterfriesisch', 'dsb': 'saterfrizišćina', 'en': 'Saterland Frisian', 'et': 'saterfriisi', 'fi': 'saterlandinfriisi', 'fr': 'saterlandais', 'hi-Latn': 'Saterland Frisian', 'hsb': 'saterfrizišćina', 'ht': 'saterlandais', 'it': 'saterfriesisch', 'ja': 'ザーターフリジア語', 'lb': 'Saterfriesesch', 'lt': 'Saterlendo fryzų', 'mk': 'затерландски фризиски', 'nb': 'saterfrisisk', 'nl': 'Saterfries', 'nn': 'saterfrisisk', 'no': 'saterfrisisk', 'pl': 'fryzyjski saterlandzki', 'sv': 'saterfrisiska', 'th': 'ฟรีเซียนซัทเธอร์แลนด์', 'to': 'lea fakafilisia-satēlani', 'tr': 'Saterland Frizcesi', 'yue': '沙特菲士蘭文', 'yue-Hans': '沙特菲士兰文', 'yue-Hant': '沙特菲士蘭文', 'zh-Hant': '沙特菲士蘭文'},
    'str': {'af': 'Straits Salish', 'am': 'ስትሬይትስ ስታሊሽ', 'ar': 'سترايتس ساليش', 'as': 'ষ্ট্ৰেইটছ ছেলিশ্ব', 'az': 'streyts saliş', 'az-Latn': 'streyts saliş', 'be': 'стрэйтс саліш', 'be-tarask': 'стрэйтс саліш', 'bg': 'стрейтс салиш', 'bn': 'স্ট্রেটস সালিস', 'brx': 'स्ट्रैत्स सालिश', 'bs': 'ravničarski sališki', 'bs-Cyrl': 'страитс салиш', 'bs-Latn': 'ravničarski sališki', 'ca': 'straits salish', 'ca-ES-valencia': 'straits salish', 'chr': 'ᏌᎵᏏ', 'cs': 'saliština (z úžin)', 'cy': 'Straits Salish', 'da': 'straits salish', 'de': 'Straits Salish', 'dsb': 'straits salishšćina', 'el': 'Στρέιτς Σαλίς', 'el-polyton': 'Στρέιτς Σαλίς', 'en': 'Straits Salish', 'es': 'salish de los estrechos', 'et': 'väinasališi', 'eu': 'itsasarteetako salishera', 'fa': 'سالیش استریتز', 'ff-Adlm': '𞤅𞤭𞤼𞤪𞤭𞤼 𞤅𞤢𞤤𞤭𞤧𞤪𞤫', 'fi': 'straits-salish', 'fil': 'Straits Salish', 'fr': 'salish des détroits', 'ga': 'Straits Salish', 'gd': 'Salish a’ Chaolais', 'gl': 'salish dos estreitos', 'gu': 'સ્ટ્રેટ્સ સેલિશ', 'ha': 'Straits Salish', 'he': 'סאליש מיצרי חואן דה פוקה', 'hi': 'स्ट्रेट्स सैलिश', 'hi-Latn': 'Straits Salish', 'hr': 'sjeverni sališki', 'hsb': 'straits salishšćina', 'ht': 'salish des détroits', 'hu': 'szorosmenti salish', 'hy': 'սթրեյթս սալիշերեն', 'ia': 'salish del strictos', 'id': 'Salish Selat', 'ig': 'Straits Salish', 'is': 'Straits-salisíanska', 'it': 'salish straits', 'ja': 'ストレイツセイリッシュ語', 'jv': 'Selat Salish', 'ka': 'სტრეიტს სალიში', 'kk': 'солтүстік стрейтс тілі', 'kk-Cyrl': 'солтүстік стрейтс тілі', 'km': 'សាលីសស្ត្រេតស៍', 'kn': 'ಸ್ಟ್ರೇಟ್ಸ್ ಸೆಲಿಶ್', 'ko': '해안 살리시어', 'kok': 'स्ट्रैट सालिश', 'kok-Deva': 'स्ट्रैट सालिश', 'ku': 'saanîçî', 'ky': 'стрейтс салишче (индей тили)', 'lo': 'ຊ່ອງແຄບເຊລີຊ', 'lt': 'Sališo sąsiaurio', 'lv': 'šauruma sališu', 'mai': 'स्ट्रेट्स सैलिस', 'mi': 'Hārihi Kuititanga', 'mk': 'салишки (Северен Теснец)', 'ml': 'സ്ട്രെയ്റ്റ്സ് സെയ്\u200cലിഷ്', 'mn': 'стрейтс салиш', 'mr': 'स्ट्राइट्स सालीश', 'ms': 'Straits Salish', 'my': 'ဆဲလစ်ရှ် ရေလက်ကြား', 'nb': 'straits-salish', 'ne': 'स्ट्रेट स्यालिस', 'nl': 'Straits Salish', 'nn': 'straits-salish', 'no': 'straits-salish', 'or': 'ଷ୍ଟ୍ରାଇଟ୍ ସାଲିଶ୍', 'pa': 'ਸਟਰੇਟਸ ਸੈਲਿਸ਼', 'pa-Guru': 'ਸਟਰੇਟਸ ਸੈਲਿਸ਼', 'pcm': 'Streti Salísh Lángwej', 'pl': 'salisz', 'ps': 'سټریټ سیلش', 'pt': 'salish do estreito norte', 'qu': 'Straits Salish Simi', 'ro': 'salish (Strâmtori)', 'ru': 'стрейтс салиш', 'sc': 'salish de sas astrinturas', 'sd': 'اسٽريٽ سليش', 'sd-Arab': 'اسٽريٽ سليش', 'si': 'සාලිෂ්ස මුද්ර සන්ධිය', 'sk': 'straits salish', 'sl': 'ožinska sališčina', 'so': 'Straits Salish', 'sq': 'sejlishte e Ngushticave të Rozarios', 'sr': 'стреицсалиш', 'sr-Cyrl': 'стреицсалиш', 'sr-Latn': 'streicsališ', 'sv': 'sundsalishanska', 'sw': 'Kisalishi cha Straiti', 'ta': 'ஸ்ட்ரெய்ட்ஸ் சாலிஷ்', 'te': 'స్ట్రెయిట్స్ సలీష్', 'th': 'สเตรตส์ซาลิช', 'ti': 'ሳሊሽ መጻብቦታት', 'tk': 'demirgazyk bogaz saliş dili', 'to': 'lea fakasalisi-vahatokelau', 'tr': 'Boğazlar Saliş dili', 'uk': 'саліська стрейт', 'ur': 'سٹریٹس سالِش', 'uz': 'streyts salish', 'uz-Latn': 'streyts salish', 'vec': 'salish streto', 'vi': 'Tiếng Straits Salish', 'yo': 'Èdè Sitirati Salisi', 'yue': '海峽薩利希文', 'yue-Hans': '海峡萨利希文', 'yue-Hant': '海峽薩利希文', 'zh': '海峡萨利希语', 'zh-Hans': '海峡萨利希语', 'zh-Hant': '海峽薩利希語', 'zu': 'Straits Salish'},
    'sts': {'en': 'Shumashti'},
    'stt': {'en': 'Budeh Stieng'},
    'stu': {'en': 'Samtao'},
    'stv': {'en': "Silt'e"},
    'stw': {'en': 'Satawalese'},
    'sty': {'en': 'Siberian Tatar'},
    'su': {'af': 'Sundanees', 'ak': 'Sunda Kasa', 'am': 'ሱዳንኛ', 'ar': 'السوندانية', 'as': 'ছুণ্ডানীজ', 'ast': 'sondanés', 'az': 'sundan', 'az-Cyrl': 'сундан', 'az-Latn': 'sundan', 'bal-Latn': 'Sudáni', 'be': 'сунда', 'be-tarask': 'сунда', 'bg': 'сундански', 'bgn': 'سونڈی', 'bn': 'সুদানী', 'br': 'sundaneg', 'brx': 'सुन्दानी', 'bs': 'sundanski', 'bs-Cyrl': 'сундански', 'bs-Latn': 'sundanski', 'ca': 'sondanès', 'ca-ES-valencia': 'sondanès', 'ccp': '𑄥𑄪𑄘𑄚𑄩', 'ce': 'сунданхойн', 'chr': 'ᏑᏂᏓᏂᏏ', 'ckb': 'سوندانی', 'cs': 'sundština', 'cy': 'Swndaneg', 'da': 'sundanesisk', 'de': 'Sundanesisch', 'dsb': 'sundanšćina', 'dz': 'སཱུན་ད་ནིས་ཁ', 'el': 'Σουνδανικά', 'el-polyton': 'Σουνδανικά', 'en': 'Sundanese', 'eo': 'sunda', 'es': 'sundanés', 'et': 'sunda', 'eu': 'sundanera', 'fa': 'سوندایی', 'ff-Adlm': '𞤅𞤵𞤲𞤣𞤢𞤲𞤭𞥅𞤪𞤫', 'fi': 'sunda', 'fil': 'Sundanese', 'fo': 'sundanesiskt', 'fr': 'soundanais', 'fur': 'sundanês', 'fy': 'Soendaneesk', 'ga': 'Sundais', 'gd': 'Cànan Sunda', 'gl': 'sundanés', 'gsw': 'Sundanesisch', 'gu': 'સંડેનીઝ', 'ha': 'Harshen Sundanese', 'he': 'סונדנזית', 'hi': 'सुंडानी', 'hi-Latn': 'Sundanese', 'hr': 'sundanski', 'hsb': 'sundanezišćina', 'ht': 'soundanais', 'hu': 'szundanéz', 'hy': 'սունդաներեն', 'ia': 'sundanese', 'id': 'Sunda', 'ig': 'Asụsụ Sundan', 'is': 'súndanska', 'it': 'sundanese', 'ja': 'スンダ語', 'jv': 'Sunda', 'ka': 'სუნდური', 'kea': 'sundanes', 'kgp': 'sunanẽj', 'kk': 'сундан тілі', 'kk-Cyrl': 'сундан тілі', 'km': 'ស៊ូដង់', 'kn': 'ಸುಂಡಾನೀಸ್', 'ko': '순다어', 'kok': 'सुंदनीस', 'kok-Deva': 'सुंदनीस', 'ks': 'سَنڈَنیٖز', 'ks-Arab': 'سَنڈَنیٖز', 'ksh': 'Sindanehsesch', 'ku': 'sundanî', 'ky': 'сунданча', 'lb': 'Sundanesesch', 'lkt': 'Sundanese Iyápi', 'lld': 'sundanesc', 'lo': 'ຊຸນແດນນີສ', 'lrc': 'سوٙدانی', 'lt': 'sundų', 'lv': 'zundu', 'mai': 'सुंदनेजे', 'mi': 'Hunanīhi', 'mk': 'сундски', 'ml': 'സുണ്ടാനീസ്', 'mn': 'сундан', 'mr': 'सुंदानीज', 'ms': 'Sunda', 'mt': 'Sundaniż', 'my': 'ဆူဒန်', 'mzn': 'سوندانسی', 'nb': 'sundanesisk', 'ne': 'सुडानी', 'nl': 'Soendanees', 'nn': 'sundanesisk', 'no': 'sundanesisk', 'nqo': 'ߛߎ߲ߘߣߊߞߊ߲', 'om': 'Afaan Sudaanii', 'or': 'ସୁଦାନୀଜ୍', 'pa': 'ਸੂੰਡਾਨੀ', 'pa-Guru': 'ਸੂੰਡਾਨੀ', 'pcm': 'Sọ́ndaniz', 'pl': 'sundajski', 'ps': 'سوډاني', 'pt': 'sundanês', 'qu': 'Sundanés Simi', 'rm': 'sundanais', 'ro': 'sundaneză', 'ru': 'сунданский', 'rw': 'Inyesudani', 'sc': 'sundanesu', 'sd': 'سوڊاني', 'sd-Arab': 'سوڊاني', 'si': 'සන්ඩනීසියානු', 'sk': 'sundčina', 'sl': 'sundanščina', 'smn': 'sundakielâ', 'so': 'Suudaaniis', 'sq': 'sundanisht', 'sr': 'сундски', 'sr-Cyrl': 'сундски', 'sr-Latn': 'sundski', 'su': 'Basa Sunda', 'su-Latn': 'Basa Sunda', 'sv': 'sundanesiska', 'sw': 'Kisunda', 'ta': 'சுண்டானீஸ்', 'te': 'సండానీస్', 'th': 'ซุนดา', 'ti': 'ሱዳንኛ', 'tk': 'sundan dili', 'to': 'lea fakasunitā', 'tr': 'Sunda dili', 'ug': 'سۇنداچە', 'uk': 'сунданська', 'ur': 'سنڈانیز', 'uz': 'sundan', 'uz-Cyrl': 'сунданча', 'uz-Latn': 'sundan', 'vec': 'sundaneze', 'vi': 'Tiếng Sunda', 'wae': 'Sundanesiš', 'xnr': 'सुंडानी', 'yo': 'Èdè Sudanísì', 'yrl': 'sũdanei', 'yue': '巽他文', 'yue-Hans': '巽他文', 'yue-Hant': '巽他文', 'zh': '巽他语', 'zh-Hans': '巽他语', 'zh-Hant': '巽他文', 'zu': 'isi-Sundanese'},
    'sua': {'en': 'Sulka'},
    'sub': {'en': 'Suku'},
    'suc': {'en': 'Western Subanon'},
    'sue': {'en': 'Suena'},
    'sug': {'en': 'Suganga'},
    'sui': {'en': 'Suki'},
    'suj': {'en': 'Shubi'},
    'suk': {'af': 'Sukuma', 'am': 'ሱኩማ', 'ar': 'السوكوما', 'as': 'ছুকুমা', 'ast': 'sukuma', 'az': 'sukuma', 'az-Cyrl': 'сукума', 'az-Latn': 'sukuma', 'be': 'сукума', 'be-tarask': 'сукума', 'bg': 'сукума', 'bgn': 'سوکومایی', 'bn': 'সুকুমা', 'br': 'sukuma', 'brx': 'सुकुमा', 'bs': 'sukuma', 'bs-Cyrl': 'сукума', 'bs-Latn': 'sukuma', 'ca': 'sukuma', 'ca-ES-valencia': 'sukuma', 'ccp': '𑄥𑄪𑄇𑄪𑄟', 'ce': 'сукума', 'chr': 'ᏑᎫᎹ', 'ckb': 'سووکووما', 'cs': 'sukuma', 'cy': 'Swcwma', 'da': 'sukuma', 'de': 'Sukuma', 'dsb': 'sukumašćina', 'el': 'Σουκούμα', 'el-polyton': 'Σουκούμα', 'en': 'Sukuma', 'es': 'sukuma', 'et': 'sukuma', 'eu': 'sukumera', 'fa': 'سوکومایی', 'ff-Adlm': '𞤅𞤵𞤳𞤵𞤥𞤢𞥄𞤪𞤫', 'fi': 'sukuma', 'fil': 'Sukuma', 'fo': 'sukuma', 'fr': 'soukouma', 'fy': 'Sukuma', 'ga': 'Sucúimis', 'gd': 'Sukuma', 'gl': 'sukuma', 'gsw': 'Sukuma-Schpraach', 'gu': 'સુકુમા', 'ha': 'Sukuma', 'he': 'סוקומה', 'hi': 'सुकुमा', 'hi-Latn': 'Sukuma', 'hr': 'sukuma', 'hsb': 'sukumašćina', 'ht': 'soukouma', 'hu': 'szukuma', 'hy': 'սուկումա', 'ia': 'sukuma', 'id': 'Sukuma', 'ig': 'Sukuma', 'is': 'súkúma', 'it': 'sukuma', 'ja': 'スクマ語', 'jv': 'Sukuma', 'ka': 'სუკუმა', 'kgp': 'sukumỹ', 'kk': 'сукума тілі', 'kk-Cyrl': 'сукума тілі', 'km': 'ស៊ូគូម៉ា', 'kn': 'ಸುಕುಮಾ', 'ko': '수쿠마어', 'kok': 'सुकुमा', 'kok-Deva': 'सुकुमा', 'ks': 'سُکُما', 'ks-Arab': 'سُکُما', 'ku': 'sukuma', 'ky': 'сукумача', 'lb': 'Sukuma-Sprooch', 'lo': 'ຊູຄູມ້າ', 'lt': 'sukuma', 'lv': 'sukumu', 'mai': 'सुकुमा', 'mi': 'Hukuma', 'mk': 'сукума', 'ml': 'സുകുമ', 'mn': 'сукума', 'mr': 'सुकुमा', 'ms': 'Sukuma', 'mt': 'Sukuma', 'my': 'ဆူကူမာ', 'nb': 'sukuma', 'ne': 'सुकुमा', 'nl': 'Sukuma', 'nn': 'sukuma', 'no': 'sukuma', 'or': 'ସୁକୁମା', 'pa': 'ਸੁਕੁਮਾ', 'pa-Guru': 'ਸੁਕੁਮਾ', 'pcm': 'Sukúma Lángwej', 'pl': 'sukuma', 'ps': 'سکوما', 'pt': 'sukuma', 'qu': 'Sukuma Simi', 'rm': 'sukuma', 'ro': 'sukuma', 'ru': 'сукума', 'sc': 'sukuma', 'sd': 'سڪوما', 'sd-Arab': 'سڪوما', 'si': 'සුකුමා', 'sk': 'sukuma', 'sl': 'sukuma', 'smn': 'sukumakielâ', 'so': 'Sukuma', 'sq': 'sukumaisht', 'sr': 'сукума', 'sr-Cyrl': 'сукума', 'sr-Latn': 'sukuma', 'sv': 'sukuma', 'sw': 'Kisukuma', 'ta': 'சுகுமா', 'te': 'సుకుమా', 'th': 'ซูคูมา', 'ti': 'ሱኩማ', 'tk': 'sukuma dili', 'to': 'lea fakasukuma', 'tr': 'Sukuma dili', 'ug': 'سۇكۇماچە', 'uk': 'сукума', 'ur': 'سکوما', 'uz': 'sukuma', 'uz-Latn': 'sukuma', 'vec': 'sukuma', 'vi': 'Tiếng Sukuma', 'xnr': 'सुकुमा', 'yo': 'Èdè Sukuma', 'yrl': 'sukuma', 'yue': '蘇庫馬文', 'yue-Hans': '苏库马文', 'yue-Hant': '蘇庫馬文', 'zh': '苏库马语', 'zh-Hans': '苏库马语', 'zh-Hant': '蘇庫馬文', 'zu': 'isi-Sukuma'},
    'sul': {'en': 'Surigaonon'},
    'sum': {'en': 'Sumo-Mayangna'},
    'suo': {'en': 'Bouni'},
    'suq': {'en': 'Tirmaga-Chai Suri'},
    'sur': {'en': 'Mwaghavul'},
    'sus': {'ar': 'السوسو', 'ast': 'susu', 'az': 'susu', 'az-Latn': 'susu', 'bg': 'сусу', 'bn': 'সুসু', 'brx': 'सुसु', 'bs': 'susu', 'bs-Cyrl': 'сусу', 'bs-Latn': 'susu', 'ca': 'susú', 'ca-ES-valencia': 'susú', 'ccp': '𑄥𑄪𑄥𑄪', 'cs': 'susu', 'cy': 'Swsŵeg', 'da': 'susu', 'de': 'Susu', 'el': 'Σούσου', 'el-polyton': 'Σούσου', 'en': 'Susu', 'es': 'susu', 'et': 'susu', 'fa': 'سوسویی', 'fi': 'susu', 'fr': 'soussou', 'fy': 'Soesoe', 'gd': 'Susu', 'gsw': 'Susu', 'gu': 'સુસુ', 'he': 'סוסו', 'hi': 'सुसु', 'hi-Latn': 'Susu', 'hr': 'susu', 'ht': 'soussou', 'hu': 'szuszu', 'id': 'Susu', 'is': 'súsú', 'it': 'susu', 'ja': 'スス語', 'kgp': 'susu', 'kn': 'ಸುಸು', 'ko': '수수어', 'ks': 'سُسوٗ', 'ks-Arab': 'سُسوٗ', 'lb': 'Susu', 'lo': 'ຊູຊູ', 'lt': 'susu', 'lv': 'susu', 'mk': 'сусу', 'ml': 'സുസു', 'mr': 'सुसु', 'mt': 'Susu', 'nb': 'susu', 'ne': 'सुसू', 'nl': 'Soesoe', 'nn': 'susu', 'no': 'susu', 'or': 'ଶୁଶୁ', 'pl': 'susu', 'pt': 'susu', 'rm': 'susu', 'ro': 'susu', 'ru': 'сусу', 'sk': 'susu', 'sl': 'susujščina', 'sr': 'сусу', 'sr-Cyrl': 'сусу', 'sr-Latn': 'susu', 'sv': 'susu', 'sw': 'Kisusu', 'ta': 'சுசு', 'te': 'సుసు', 'th': 'ซูซู', 'to': 'lea fakasusū', 'tr': 'Susu', 'ug': 'سۇسۇچە', 'uk': 'сусу', 'vi': 'Tiếng Susu', 'yrl': 'susu', 'yue': '蘇蘇文', 'yue-Hans': '苏苏文', 'yue-Hant': '蘇蘇文', 'zh': '苏苏语', 'zh-Hans': '苏苏语', 'zh-Hant': '蘇蘇文'},
    'sut': {'en': 'Subtiaba'},
    'suv': {'en': 'Puroik'},
    'suw': {'en': 'Sumbwa'},
    'sux': {'ar': 'السومارية', 'ast': 'sumeriu', 'az': 'sumeryan', 'az-Latn': 'sumeryan', 'be': 'шумерская', 'be-tarask': 'шумерская', 'bg': 'шумерски', 'bn': 'সুমেরীয়', 'br': 'sumereg', 'brx': 'सुमेरिअन', 'bs': 'sumerski', 'bs-Cyrl': 'сумерски', 'bs-Latn': 'sumerski', 'ca': 'sumeri', 'ca-ES-valencia': 'sumeri', 'ccp': '𑄥𑄪𑄟𑄬𑄢𑄩𑄠𑄧', 'cs': 'sumerština', 'cy': 'Swmereg', 'da': 'sumerisk', 'de': 'Sumerisch', 'el': 'Σουμερικά', 'el-polyton': 'Σουμερικά', 'en': 'Sumerian', 'es': 'sumerio', 'et': 'sumeri', 'fa': 'سومری', 'fi': 'sumeri', 'fr': 'sumérien', 'fur': 'sumeric', 'fy': 'Soemerysk', 'ga': 'Suiméiris', 'gd': 'Cànan Sumer', 'gsw': 'Sumerisch', 'gu': 'સુમેરિયન', 'he': 'שומרית', 'hi': 'सुमेरियन', 'hi-Latn': 'Sumerian', 'hr': 'sumerski', 'ht': 'sumérien', 'hu': 'sumér', 'id': 'Sumeria', 'is': 'súmerska', 'it': 'sumero', 'ja': 'シュメール語', 'ka': 'შუმერული', 'kgp': 'sumẽrijo', 'kn': 'ಸುಮೇರಿಯನ್', 'ko': '수메르어', 'ks': 'سُمیریَن', 'ks-Arab': 'سُمیریَن', 'lb': 'Sumeresch', 'lo': 'ຊູເມີເລຍ', 'lt': 'šumerų', 'lv': 'šumeru', 'mk': 'сумерски', 'ml': 'സുമേരിയൻ', 'mr': 'सुमेरियन', 'mt': 'Sumerjan', 'nb': 'sumerisk', 'ne': 'सुमेरियाली', 'nl': 'Soemerisch', 'nn': 'sumerisk', 'no': 'sumerisk', 'or': 'ସୁମେରିଆନ୍', 'pl': 'sumeryjski', 'pt': 'sumério', 'rm': 'sumeric', 'ro': 'sumeriană', 'ru': 'шумерский', 'sk': 'sumerčina', 'sl': 'sumerščina', 'sr': 'сумерски', 'sr-Cyrl': 'сумерски', 'sr-Latn': 'sumerski', 'sv': 'sumeriska', 'ta': 'சுமேரியன்', 'te': 'సుమేరియాన్', 'th': 'ซูเมอ', 'to': 'lea fakasumelia', 'tr': 'Sümerce', 'ug': 'سۈمەرچە', 'uk': 'шумерська', 'vi': 'Tiếng Sumeria', 'yi': 'סומעריש', 'yrl': 'sumeriu', 'yue': '蘇美文', 'yue-Hans': '苏美文', 'yue-Hant': '蘇美文', 'zh': '苏美尔语', 'zh-Hans': '苏美尔语', 'zh-Hant': '蘇美文'},
    'suy': {'en': 'Suyá'},
    'suz': {'en': 'Sunwar'},
    'sv': {'af': 'Sweeds', 'agq': 'Suedìs', 'ak': 'Sweden kasa', 'am': 'ስዊድንኛ', 'ar': 'السويدية', 'as': 'ছুইডিচ', 'asa': 'Kithwidi', 'ast': 'suecu', 'az': 'isveç', 'az-Arab': 'سوئدجه', 'az-Cyrl': 'исвеч', 'az-Latn': 'isveç', 'bal-Latn': 'Swidi', 'bas': 'Hɔp u suɛ᷆d', 'be': 'шведская', 'be-tarask': 'шведская', 'bem': 'Ichi Swideni', 'bez': 'Hiswidi', 'bg': 'шведски', 'bgn': 'سویڈنی', 'bm': 'suwɛdikan', 'bn': 'সুইডিশ', 'br': 'svedeg', 'brx': 'स्वीडिश', 'bs': 'švedski', 'bs-Cyrl': 'шведски', 'bs-Latn': 'švedski', 'ca': 'suec', 'ca-ES-valencia': 'suec', 'ccp': '𑄥𑄭𑄪𑄓𑄨𑄥𑄴', 'ce': 'шведийн', 'cgg': 'Oruswidi', 'chr': 'ᏍᏫᏗᏏ', 'ckb': 'سویدی', 'cs': 'švédština', 'cy': 'Swedeg', 'da': 'svensk', 'dav': 'Kiswidi', 'de': 'Schwedisch', 'dje': 'Suweede senni', 'dsb': 'šwedšćina', 'dyo': 'suedi', 'dz': 'སུའི་ཌིཤ་ཁ', 'ebu': 'Kĩswidi', 'ee': 'swedengbe', 'el': 'Σουηδικά', 'el-polyton': 'Σουηδικά', 'en': 'Swedish', 'eo': 'sveda', 'es': 'sueco', 'et': 'rootsi', 'eu': 'suediera', 'ewo': 'ǹkɔ́bɔ suwɛ́d', 'fa': 'سوئدی', 'ff': 'Sweedeere', 'ff-Adlm': '𞤅𞤱𞤫𞤣𞤭𞤲𞤳𞤮𞥅𞤪𞤫', 'ff-Latn': 'Sweedeere', 'fi': 'ruotsi', 'fil': 'Swedish', 'fo': 'svenskt', 'fr': 'suédois', 'fur': 'svedês', 'fy': 'Zweeds', 'ga': 'Sualainnis', 'gd': 'Suainis', 'gl': 'sueco', 'gsw': 'Schwedisch', 'gu': 'સ્વીડિશ', 'guz': 'Kiswidi', 'ha': 'Harshen Suwedan', 'haw': 'Kuekene', 'he': 'שוודית', 'hi': 'स्वीडिश', 'hi-Latn': 'Swedish', 'hr': 'švedski', 'hsb': 'šwedšćina', 'ht': 'suédois', 'hu': 'svéd', 'hy': 'շվեդերեն', 'ia': 'svedese', 'id': 'Swedia', 'ie': 'sved', 'ig': 'Sụwidiishi', 'is': 'sænska', 'it': 'svedese', 'ja': 'スウェーデン語', 'jmc': 'Kyiswidi', 'jv': 'Swedia', 'ka': 'შვედური', 'kab': 'Taswidit', 'kam': 'Kiswidi', 'kde': 'Chiswidi', 'kea': 'sueku', 'kgp': 'suvéko', 'khq': 'Suweede senni', 'ki': 'Kiswidi', 'kk': 'швед тілі', 'kk-Cyrl': 'швед тілі', 'kln': 'kutitab Sweden', 'km': 'ស៊ុយអែត', 'kn': 'ಸ್ವೀಡಿಷ್', 'ko': '스웨덴어', 'kok': 'स्विडीश', 'kok-Deva': 'स्विडीश', 'ks': 'سویٖڈِش', 'ks-Arab': 'سویٖڈِش', 'ksb': 'Kiswidi', 'ksf': 'riswɛ́dǝ', 'ksh': 'Schwehdesch', 'ku': 'swêdî', 'kxv': 'sviḍis', 'kxv-Deva': 'स्वीडिस', 'kxv-Latn': 'sviḍis', 'kxv-Orya': 'ସ୍ୱିଡିସ୍', 'kxv-Telu': 'స్విడిస్', 'ky': 'шведче', 'lag': 'Kɨswíidi', 'lb': 'Schwedesch', 'lg': 'Luswideni', 'lkt': 'Swedish Iyápi', 'lld': 'svedesc', 'ln': 'lisuwedɛ', 'lo': 'ສະວີດິຊ', 'lrc': 'سوٙئدی', 'lt': 'švedų', 'lu': 'Lisuwidi', 'luo': 'Kiswidi', 'luy': 'Kiswidi', 'lv': 'zviedru', 'mai': 'स्वीडिश', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lswidi', 'mer': 'Kĩswideni', 'mfe': 'swedwa', 'mg': 'Soisa', 'mgh': 'Iswidi', 'mi': 'Huitene', 'mk': 'шведски', 'ml': 'സ്വീഡിഷ്', 'mn': 'швед', 'mr': 'स्वीडिश', 'ms': 'Sweden', 'mt': 'Żvediż', 'mua': 'Swedia', 'my': 'ဆွီဒင်', 'mzn': 'سوئدی', 'naq': 'Swedeǁî gowab', 'nb': 'svensk', 'nd': 'isi-Swidishi', 'ne': 'स्विडिस', 'nl': 'Zweeds', 'nmg': 'Kiɛl bó suedois', 'nn': 'svensk', 'no': 'svensk', 'nqo': 'ߛߎߥߍߘߌߞߊ߲', 'nus': 'Thok i̱thwidicni', 'nyn': 'Oruswidi', 'om': 'Afaan Suwidiin', 'or': 'ସ୍ୱେଡିସ୍', 'pa': 'ਸਵੀਡਿਸ਼', 'pa-Guru': 'ਸਵੀਡਿਸ਼', 'pcm': 'Suwídẹ́n Lángwej', 'pl': 'szwedzki', 'ps': 'سویډنی', 'pt': 'sueco', 'qu': 'Sueco Simi', 'rm': 'svedais', 'rn': 'Igisuweduwa', 'ro': 'suedeză', 'rof': 'Kiswidi', 'ru': 'шведский', 'rw': 'Igisuweduwa', 'rwk': 'Kyiswidi', 'saq': 'Kiswidi', 'sbp': 'Ishiswidi', 'sc': 'isvedesu', 'sd': 'سويڊش', 'sd-Arab': 'سويڊش', 'se': 'ruoŧagiella', 'seh': 'sueco', 'ses': 'Suweede senni', 'sg': 'Sueduäa', 'shi': 'ⵜⴰⵙⵡⵉⴷⵉⵜ', 'shi-Latn': 'Taswidit', 'shi-Tfng': 'ⵜⴰⵙⵡⵉⴷⵉⵜ', 'si': 'ස්වීඩන්', 'sk': 'švédčina', 'sl': 'švedščina', 'smn': 'ruotâkielâ', 'sn': 'chiSwedish', 'so': 'Iswiidhish', 'sq': 'suedisht', 'sr': 'шведски', 'sr-Cyrl': 'шведски', 'sr-Latn': 'švedski', 'sv': 'svenska', 'sw': 'Kiswidi', 'syr': 'ܣܘܝܕܐܝܬ', 'ta': 'ஸ்வீடிஷ்', 'te': 'స్వీడిష్', 'teo': 'Kiswidi', 'tg': 'шведӣ', 'th': 'สวีเดน', 'ti': 'ስዊድንኛ', 'tk': 'şwed dili', 'to': 'lea fakasuēteni', 'tr': 'İsveççe', 'tt': 'швед', 'twq': 'Suweede senni', 'tzm': 'Taswidit', 'ug': 'شىۋېدچە', 'uk': 'шведська', 'ur': 'سویڈش', 'uz': 'shved', 'uz-Cyrl': 'шведча', 'uz-Latn': 'shved', 'vai': 'ꖬꔨꗵꘋ', 'vai-Latn': 'Súwídɛŋ', 'vai-Vaii': 'ꖬꔨꗵꘋ', 'vec': 'zvedeze', 'vi': 'Tiếng Thụy Điển', 'vun': 'Kyiswidi', 'wae': 'Schwediš', 'wo': 'Suweduwaa', 'xnr': 'स्वीडिश', 'xog': 'Oluswideni', 'yav': 'nusuetua', 'yi': 'שוועדיש', 'yo': 'Èdè Suwidiisi', 'yrl': 'sueku', 'yue': '瑞典文', 'yue-Hans': '瑞典文', 'yue-Hant': '瑞典文', 'zgh': 'ⵜⴰⵙⵡⵉⴷⵉⵜ', 'zh': '瑞典语', 'zh-Hans': '瑞典语', 'zh-Hant': '瑞典文', 'zu': 'isi-Swedish'},
    'sva': {'en': 'Svan'},
    'svb': {'en': 'Ulau-Suain'},
    'svc': {'en': 'Vincentian Creole English'},
    'sve': {'en': 'Serili'},
    'svk': {'en': 'Slovakian Sign Language'},
    'svm': {'en': 'Slavomolisano'},
    'svr': {'en': 'Savara'},
    'svs': {'en': 'Savosavo'},
    'svx': {'en': 'Skalvian'},
    'sw': {'af': 'Swahili', 'ak': 'Swahili', 'am': 'ስዋሂሊኛ', 'ar': 'السواحلية', 'as': 'স্বাহিলি', 'ast': 'suaḥili', 'az': 'suahili', 'az-Arab': 'سواحیلیجه', 'az-Cyrl': 'суаһили', 'az-Latn': 'suahili', 'bal-Latn': 'Swáhéli', 'be': 'суахілі', 'be-tarask': 'суахілі', 'bg': 'суахили', 'bgn': 'سواحلی', 'bn': 'সোয়াহিলি', 'br': 'swahili', 'brx': 'स्वाहिली', 'bs': 'svahili', 'bs-Cyrl': 'свахили', 'bs-Latn': 'svahili', 'ca': 'suahili', 'ca-ES-valencia': 'suahili', 'ccp': '𑄥𑄱𑄦𑄨𑄣𑄨', 'ce': 'суахили', 'chr': 'ᏍᏩᎯᎵ', 'ckb': 'سواهیلی', 'cs': 'svahilština', 'cy': 'Swahili', 'da': 'swahili', 'de': 'Suaheli', 'dsb': 'swahilišćina', 'dz': 'སྭཱ་ཧི་ལི་ཁ', 'ee': 'swahili', 'el': 'Σουαχίλι', 'el-polyton': 'Σουαχίλι', 'en': 'Swahili', 'eo': 'svahila', 'es': 'suajili', 'et': 'suahiili', 'eu': 'swahilia', 'fa': 'سواحیلی', 'ff-Adlm': '𞤅𞤵𞤱𞤢𞤸𞤭𞤤𞤭𞥅𞤪𞤫', 'fi': 'swahili', 'fil': 'Swahili', 'fo': 'swahili', 'fr': 'swahili', 'fur': 'swahili', 'fy': 'Swahili', 'ga': 'Svahaílis', 'gd': 'Kiswahili', 'gl': 'suahili', 'gsw': 'Suaheli', 'gu': 'સ્વાહિલી', 'ha': 'Harshen Suwahili', 'he': 'סווהילי', 'hi': 'स्वाहिली', 'hi-Latn': 'Swahili', 'hr': 'svahili', 'hsb': 'suahelšćina', 'ht': 'swahili', 'hu': 'szuahéli', 'hy': 'սուահիլի', 'ia': 'swahili', 'id': 'Swahili', 'ie': 'swahili', 'ig': 'Asụsụ Swahili', 'ii': 'ꌖꑟꆺꉙ', 'is': 'svahílí', 'it': 'swahili', 'ja': 'スワヒリ語', 'jv': 'Swahili', 'ka': 'სუაჰილი', 'kea': 'suaíli', 'kgp': 'suvahiri', 'kk': 'суахили тілі', 'kk-Cyrl': 'суахили тілі', 'km': 'ស្វាហ៊ីលី', 'kn': 'ಸ್ವಹಿಲಿ', 'ko': '스와힐리어', 'kok': 'स्वाहिली', 'kok-Deva': 'स्वाहिली', 'ks': 'سواہِلی', 'ks-Arab': 'سواہِلی', 'ksh': 'Suahehlesch', 'ku': 'swahîlî', 'kxv': 'svahili', 'kxv-Deva': 'स्वाहिली', 'kxv-Latn': 'svahili', 'kxv-Orya': 'ସ୍ୱାହିଲି', 'kxv-Telu': 'స్వాహిలి', 'ky': 'суахиличе', 'lb': 'Suaheli', 'lkt': 'Swahili Iyápi', 'lld': 'swahili', 'lo': 'ຊວາຮີລິ', 'lrc': 'سأڤاحیلی', 'lt': 'suahilių', 'lv': 'svahili', 'mai': 'स्वाहिलि', 'mi': 'Wāhīri', 'mk': 'свахили', 'ml': 'സ്വാഹിലി', 'mn': 'свахили', 'mr': 'स्वाहिली', 'ms': 'Swahili', 'mt': 'Swahili', 'my': 'ဆွာဟီလီ', 'mzn': 'سواحیلی', 'nb': 'swahili', 'ne': 'स्वाहिली', 'nl': 'Swahili', 'nn': 'swahili', 'no': 'swahili', 'nqo': 'ߛߎߥߊߤߟߌߞߊ߲', 'om': 'Suwahilii', 'or': 'ସ୍ୱାହିଲି', 'pa': 'ਸਵਾਹਿਲੀ', 'pa-Guru': 'ਸਵਾਹਿਲੀ', 'pcm': 'Swahíli', 'pl': 'suahili', 'ps': 'سواهېلي', 'pt': 'suaíli', 'qu': 'Suajili Simi', 'rm': 'suahili', 'ro': 'swahili', 'ru': 'суахили', 'rw': 'Igiswahili', 'sc': 'swahili', 'sd': 'سواحيلي', 'sd-Arab': 'سواحيلي', 'si': 'ස්වාහිලි', 'sk': 'swahilčina', 'sl': 'svahili', 'smn': 'swahilikielâ', 'so': 'Sawaaxili', 'sq': 'suahilisht', 'sr': 'свахили', 'sr-Cyrl': 'свахили', 'sr-Latn': 'svahili', 'sv': 'swahili', 'sw': 'Kiswahili', 'syr': 'ܣܘܐܗܝܠܐܝܬ', 'ta': 'ஸ்வாஹிலி', 'te': 'స్వాహిలి', 'th': 'สวาฮีลี', 'ti': 'ስዋሂሊ', 'tk': 'suahili dili', 'to': 'lea fakasuahili', 'tr': 'Svahili dili', 'ug': 'سىۋاھىلچە', 'uk': 'суахілі', 'ur': 'سواحلی', 'uz': 'suaxili', 'uz-Cyrl': 'суахили', 'uz-Latn': 'suaxili', 'vec': 'suaili', 'vi': 'Tiếng Swahili', 'wae': 'Suaheliš', 'xnr': 'स्वाहिली', 'yi': 'סוואַהיליש', 'yo': 'Èdè Swahili', 'yrl': 'suaíri', 'yue': '史瓦希里文', 'yue-Hans': '史瓦希里文', 'yue-Hant': '史瓦希里文', 'zh': '斯瓦希里语', 'zh-Hans': '斯瓦希里语', 'zh-Hant': '史瓦希里文', 'zu': 'isiSwahili'},
    'sw-CD': {'am': 'ኮንጎ ስዋሂሊ', 'ar': 'الكونغو السواحلية', 'as': 'কঙ্গো স্বাহিলি', 'ast': 'suaḥili del Congu', 'az': 'Konqo suahilicəsi', 'az-Arab': 'کوْنقو سواحیلیسی', 'az-Cyrl': 'Конго суаһилиҹәси', 'az-Latn': 'Konqo suahilicəsi', 'bal-Latn': 'Swáhéli (CD)', 'be': 'кангалезская суахілі', 'be-tarask': 'кангалезская суахілі', 'bg': 'конгоански суахили', 'bgn': 'کانگویی سواحلی', 'bn': 'কঙ্গো সোয়াহিলি', 'br': 'swahili Kongo', 'ca': 'suahili del Congo', 'ca-ES-valencia': 'suahili del Congo', 'ccp': '𑄇𑄧𑄋𑄴𑄉𑄮 𑄥𑄱𑄦𑄨𑄣𑄨', 'ce': 'суахили (Конго)', 'chr': 'ᎧᏂᎪ ᏍᏩᎯᎵ', 'ckb': 'سواهیلیی کۆنگۆ', 'cs': 'svahilština (Kongo)', 'cy': 'Swahili’r Congo', 'da': 'congolesisk swahili', 'de': 'Kongo-Swahili', 'dsb': 'kongojska swahilišćina', 'el': 'Κονγκό Σουαχίλι', 'el-polyton': 'Κονγκό Σουαχίλι', 'en': 'Congo Swahili', 'es': 'suajili del Congo', 'et': 'Kongo suahiili', 'eu': 'Kongoko swahilia', 'fa': 'سواحیلی کنگو', 'ff-Adlm': '𞤅𞤵𞤱𞤢𞤸𞤭𞤤𞤭𞥅𞤪𞤫 𞤑𞤮𞤲𞤺𞤮 𞤑𞤭𞤲𞤧𞤢𞤧𞤢', 'fi': 'kongonswahili', 'fil': 'Congo Swahili', 'fo': 'kongo svahili', 'fr': 'swahili du Congo', 'fy': 'Congo Swahili', 'ga': 'Svahaílis an Chongó', 'gd': 'Kiswahili na Congo', 'gl': 'suahili congolés', 'gu': 'કોંગો સ્વાહિલી', 'he': 'סווהילי קונגו', 'hi': 'कांगो स्वाहिली', 'hi-Latn': 'Congo Swahili', 'hr': 'kongoanski svahili', 'hsb': 'kongoska suahelšćina', 'ht': 'swahili du Congo', 'hu': 'kongói szuahéli', 'hy': 'կոնգոյի սուահիլի', 'ia': 'swahili del Congo', 'id': 'Swahili (Kongo)', 'is': 'kongósvahílí', 'it': 'swahili del Congo', 'ja': 'コンゴ・スワヒリ語', 'ka': 'კონგოს სუაჰილი', 'kea': 'suaíli kongoles', 'kgp': 'suvahiri Kãgo tá', 'kk': 'конго суахили тілі', 'kk-Cyrl': 'конго суахили тілі', 'km': 'កុងហ្គោស្វាហ៊ីលី', 'kn': 'ಕಾಂಗೊ ಸ್ವಹಿಲಿ', 'ko': '콩고 스와힐리어', 'kok': 'काँगो स्वाहिली', 'kok-Deva': 'काँगो स्वाहिली', 'ku': 'swahîlîya kongoyî', 'kxv': 'kongo svahili', 'kxv-Deva': 'कांगो स्वाहिली', 'kxv-Latn': 'kongo svahili', 'kxv-Orya': 'କଙ୍ଗ ସ୍ୱାହିଲି', 'kxv-Telu': 'కాగోం స్వాహిలి', 'ky': 'конго суахаличе', 'lb': 'Kongo-Swahili', 'lld': 'swahili dl Congo', 'lo': 'ຄອງໂກ ຊວາຮີລິ', 'lrc': 'سأڤاحیلی کونگو', 'lt': 'Kongo suahilių', 'lv': 'svahili (Kongo)', 'mk': 'конгоански свахили', 'ml': 'കോംഗോ സ്വാഹിലി', 'mn': 'конгогийн свахили', 'mr': 'काँगो स्वाहिली', 'ms': 'Congo Swahili', 'mt': 'Swahili tar-Repubblika Demokratika tal-Kongo', 'my': 'ကွန်ဂို ဆွာဟီလီ', 'mzn': 'کنگو سواحیلی', 'nb': 'kongolesisk swahili', 'ne': 'कङ्गो स्वाहिली', 'nn': 'kongolesisk swahili', 'no': 'kongolesisk swahili', 'or': 'କଙ୍ଗୋ ସ୍ୱାହିଲି', 'pa': 'ਕਾਂਗੋ ਸਵਾਇਲੀ', 'pa-Guru': 'ਕਾਂਗੋ ਸਵਾਇਲੀ', 'pl': 'kongijski suahili', 'ps': 'کانګو سواهلی', 'pt': 'suaíli do Congo', 'qu': 'Suajili Simi (Congo (RDC))', 'ro': 'swahili (R.D. Congo)', 'ru': 'конголезский суахили', 'sc': 'swahili de su Congo', 'sd': 'ڪونگو سواحيلي', 'sd-Arab': 'ڪونگو سواحيلي', 'si': 'කොංගෝ ස්වාහිලි', 'sk': 'svahilčina (konžská)', 'sl': 'kongoški svahili', 'smn': 'Kongo swahilikielâ', 'sq': 'suahilishte kongoleze', 'sr': 'кисвахили', 'sr-Cyrl': 'кисвахили', 'sr-Latn': 'kisvahili', 'sv': 'Kongo-swahili', 'ta': 'காங்கோ ஸ்வாஹிலி', 'te': 'కాంగో స్వాహిలి', 'th': 'สวาฮีลี - คองโก', 'ti': 'ስዋሂሊ (ኮንጎ)', 'tk': 'kongo suahili dili', 'to': 'lea fakasuahili-kongikō', 'tr': 'Kongo Svahili', 'ug': 'كونگو سىۋالىچە', 'uk': 'суахілі (Конго)', 'ur': 'کانگو سواحلی', 'uz': 'suaxili (Kongo)', 'uz-Cyrl': 'конго-суахили', 'uz-Latn': 'suaxili (Kongo)', 'vec': 'suaili de’l Congo', 'vi': 'Tiếng Swahili Congo', 'yi': 'קאנגא־סוואַהיליש', 'yrl': 'suairi kũguwara', 'yue': '史瓦希里文（剛果）', 'yue-Hans': '史瓦希里文（刚果）', 'yue-Hant': '史瓦希里文（剛果）', 'zh': '刚果斯瓦希里语', 'zh-Hans': '刚果斯瓦希里语', 'zh-Hant': '史瓦希里文（剛果）', 'zu': 'isi-Congo Swahili'},
    'swb': {'af': 'Comoraans', 'am': 'ኮሞሪያን', 'ar': 'القمرية', 'as': 'কোমোৰিয়ান', 'ast': 'comorianu', 'az': 'komor', 'az-Cyrl': 'комор', 'az-Latn': 'komor', 'be': 'каморская', 'be-tarask': 'каморская', 'bg': 'коморски', 'bgn': 'قمرین', 'bn': 'কমোরিয়ান', 'br': 'komoreg', 'brx': 'कमरियान', 'bs': 'komorski', 'bs-Cyrl': 'коморски', 'bs-Latn': 'komorski', 'ca': 'comorià', 'ca-ES-valencia': 'comorià', 'ccp': '𑄇𑄧𑄟𑄮𑄢𑄨𑄠𑄚𑄴', 'ce': 'коморийн', 'chr': 'ᎪᎼᎵᎠᏂ', 'ckb': 'کۆمۆری', 'cs': 'komorština', 'cy': 'Comoreg', 'da': 'comorisk', 'de': 'Komorisch', 'dsb': 'comorianšćina', 'ee': 'komorogbe', 'el': 'Κομοριανά', 'el-polyton': 'Κομοριανά', 'en': 'Comorian', 'es': 'comorense', 'et': 'komoori', 'eu': 'komoreera', 'fa': 'کوموری', 'ff-Adlm': '𞤑𞤮𞤥𞤮𞤪𞤭𞥅𞤪𞤫', 'fi': 'komori', 'fil': 'Comorian', 'fo': 'komoriskt', 'fr': 'comorien', 'fy': 'Shimaore', 'ga': 'teanga na gComórach', 'gd': 'Comorais', 'gl': 'comoriano', 'gu': 'કોમોરિયન', 'ha': 'Komoriyanci', 'he': 'קומורית', 'hi': 'कोमोरियन', 'hi-Latn': 'Comorian', 'hr': 'komorski', 'hsb': 'comorianšćina', 'ht': 'comorien', 'hu': 'comorei', 'hy': 'կոմորերեն', 'ia': 'comoriano', 'id': 'Komoria', 'ig': 'Comorian', 'is': 'shimaoríska', 'it': 'comoriano', 'ja': 'コモロ語', 'jv': 'Komorian', 'ka': 'კომორული', 'kgp': 'komorijỹnũ', 'kk': 'комор тілі', 'kk-Cyrl': 'комор тілі', 'km': 'កូម៉ូរី', 'kn': 'ಕೊಮೊರಿಯನ್', 'ko': '코모로어', 'kok': 'कोमोरियन', 'kok-Deva': 'कोमोरियन', 'ksh': 'Kommohresch', 'ku': 'komorî', 'ky': 'коморчо', 'lb': 'Komoresch', 'lkt': 'Comonian Iyápi', 'lo': 'ໂຄໂນຣຽນ', 'lt': 'komorų', 'lv': 'komoru', 'mai': 'कोमोरायन', 'mi': 'Komōriana', 'mk': 'коморијански', 'ml': 'കൊമോറിയൻ', 'mn': 'комори', 'mr': 'कोमोरियन', 'ms': 'Comoria', 'mt': 'Komorjan', 'my': 'ကိုမိုရီးယန်း', 'nb': 'komorisk', 'ne': 'कोमोरी', 'nl': 'Shimaore', 'nn': 'shimaore', 'no': 'komorisk', 'or': 'କୋମୋରିୟ', 'pa': 'ਕੋਮੋਰੀਅਨ', 'pa-Guru': 'ਕੋਮੋਰੀਅਨ', 'pcm': 'Komória Lángwej', 'pl': 'komoryjski', 'ps': 'کومورياني', 'pt': 'comoriano', 'qu': 'Comorian Simi', 'ro': 'comoreză', 'ru': 'коморский', 'sc': 'comorianu', 'sd': 'ڪمورين', 'sd-Arab': 'ڪمورين', 'se': 'shimaorigiella', 'si': 'කොමොරියන්', 'sk': 'komorčina', 'sl': 'šikomor', 'smn': 'komorikielâ', 'so': 'Comorian', 'sq': 'kamorianisht', 'sr': 'коморски', 'sr-Cyrl': 'коморски', 'sr-Latn': 'komorski', 'sv': 'shimaoré', 'sw': 'Shikomor', 'ta': 'கொமோரியன்', 'te': 'కొమొరియన్', 'th': 'โคเมอเรีย', 'ti': 'ኮሞርኛ', 'tk': 'komor dili', 'to': 'lea fakakomolo', 'tr': 'Komorca', 'ug': 'كومورىچە', 'uk': 'коморська', 'ur': 'کوموریائی', 'uz': 'qamar', 'uz-Cyrl': 'коморча', 'uz-Latn': 'qamar', 'vec': 'comorian', 'vi': 'Tiếng Cômo', 'xnr': 'कोमोरियन', 'yi': 'קאמאריש', 'yo': 'Èdè Komora', 'yrl': 'kumurianu', 'yue': '葛摩文', 'yue-Hans': '葛摩文', 'yue-Hant': '葛摩文', 'zh': '科摩罗语', 'zh-Hans': '科摩罗语', 'zh-Hant': '葛摩文', 'zu': 'isi-Comorian'},
    'swc': {'en': 'Congo Swahili'},
    'swf': {'en': 'Sere'},
    'swg': {'en': 'Swabian'},
    'swh': {'en': 'Swahili (individual language)'},
    'swi': {'en': 'Sui'},
    'swj': {'en': 'Sira'},
    'swk': {'en': 'Malawi Sena'},
    'swl': {'en': 'Swedish Sign Language'},
    'swm': {'en': 'Samosa'},
    'swn': {'en': 'Sawknah'},
    'swo': {'en': 'Shanenawa'},
    'swp': {'en': 'Suau'},
    'swq': {'en': 'Sharwa'},
    'swr': {'en': 'Saweru'},
    'sws': {'en': 'Seluwasan'},
    'swt': {'en': 'Sawila'},
    'swu': {'en': 'Suwawa'},
    'swv': {'en': 'Shekhawati'},
    'sww': {'en': 'Sowa'},
    'swx': {'en': 'Suruahá'},
    'swy': {'en': 'Sarua'},
    'sxb': {'en': 'Suba'},
    'sxc': {'en': 'Sicanian'},
    'sxe': {'en': 'Sighu'},
    'sxg': {'en': 'Shuhi'},
    'sxk': {'en': 'Southern Kalapuya'},
    'sxl': {'en': 'Selian'},
    'sxm': {'en': 'Samre'},
    'sxn': {'en': 'Sangir'},
    'sxo': {'en': 'Sorothaptic'},
    'sxr': {'en': 'Saaroa'},
    'sxs': {'en': 'Sasaru'},
    'sxu': {'en': 'Upper Saxon'},
    'sxw': {'en': 'Saxwe Gbe'},
    'sya': {'en': 'Siang'},
    'syb': {'en': 'Central Subanen'},
    'syc': {'am': 'ክላሲክ ኔይራ', 'ar': 'سريانية تقليدية', 'ast': 'siriacu clásicu', 'bg': 'класически сирийски', 'bn': 'প্রাচীন সিরিও', 'br': 'sirieg klasel', 'brx': 'पारंपरीक सिरिआक', 'bs': 'klasični sirijski', 'bs-Cyrl': 'класични сиријски', 'bs-Latn': 'klasični sirijski', 'ca': 'siríac clàssic', 'ca-ES-valencia': 'siríac clàssic', 'ccp': '𑄛𑄪𑄢𑄮𑄚𑄴 𑄥𑄨𑄢𑄨𑄃𑄮', 'cs': 'syrština (klasická)', 'cy': 'Hen Syrieg', 'da': 'klassisk syrisk', 'de': 'Altsyrisch', 'el': 'Κλασικά Συριακά', 'el-polyton': 'Κλασικά Συριακά', 'en': 'Classical Syriac', 'es': 'siríaco clásico', 'et': 'vanasüüria', 'fa': 'سریانی کلاسیک', 'fi': 'muinaissyyria', 'fr': 'syriaque classique', 'fy': 'Klassiek Syrysk', 'gd': 'Suraidheac Chlasaigeach', 'gsw': 'Altsyrisch', 'gu': 'પરંપરાગત સિરિએક', 'he': 'סירית קלאסית', 'hi': 'क्लासिकल सिरिएक', 'hi-Latn': 'Classical Syriac', 'hr': 'klasični sirski', 'ht': 'syriaque classique', 'hu': 'klasszikus szír', 'id': 'Suriah Klasik', 'is': 'klassísk sýrlenska', 'it': 'siriaco classico', 'ja': '古典シリア語', 'ka': 'კლასიკური სირიული', 'kgp': 'sirijako há tỹvĩ', 'kn': 'ಶಾಸ್ತ್ರೀಯ ಸಿರಿಯಕ್', 'ko': '고전 시리아어', 'lb': 'Alsyresch', 'lo': 'ຊີເລຍແບບດັ້ງເດີມ', 'lt': 'klasikinė sirų', 'lv': 'klasiskā sīriešu', 'mk': 'класичен сириски', 'ml': 'പുരാതന സുറിയാനിഭാഷ', 'mr': 'अभिजात सिरियाक', 'nb': 'klassisk syrisk', 'ne': 'परम्परागत सिरियाक', 'nl': 'Klassiek Syrisch', 'nn': 'klassisk syrisk', 'no': 'klassisk syrisk', 'or': 'କ୍ଲାସିକାଲ୍ ସିରିକ୍', 'pl': 'syriacki', 'pt': 'siríaco clássico', 'rm': 'siric classic', 'ro': 'siriacă clasică', 'ru': 'классический сирийский', 'sk': 'sýrčina (klasická)', 'sl': 'klasična sirščina', 'sr': 'сиријачки', 'sr-Cyrl': 'сиријачки', 'sr-Latn': 'sirijački', 'sv': 'klassisk syriska', 'ta': 'பாரம்பரிய சிரியாக்', 'te': 'సాంప్రదాయ సిరియాక్', 'th': 'ซีเรียแบบดั้งเดิม', 'to': 'lea fakasuliāiā-muʻa', 'tr': 'Klasik Süryanice', 'ug': 'قەدىمىي سۇرىيەچە', 'uk': 'сирійська класична', 'vi': 'Tiếng Syriac cổ', 'yrl': 'siriaku katuwa', 'yue': '古敘利亞文', 'yue-Hans': '古叙利亚文', 'yue-Hant': '古敘利亞文', 'zh': '古典叙利亚语', 'zh-Hans': '古典叙利亚语', 'zh-Hant': '古敘利亞文'},
    'syd': {'en': 'Samoyedic languages'},
    'syd-fne': {'en': 'Forest Nenets'},
    'syd-pro': {'en': 'Proto-Samoyedic'},
    'syi': {'en': 'Seki'},
    'syk': {'en': 'Sukur'},
    'syl': {'en': 'Sylheti'},
    'sym': {'en': 'Maya Samo'},
    'syn': {'en': 'Senaya'},
    'syo': {'en': 'Suoy'},
    'syr': {'af': 'Siries', 'ak': 'Siiria Kasa', 'am': 'ሲሪያክ', 'ar': 'السريانية', 'as': 'চিৰিয়াক', 'ast': 'siriacu', 'az': 'suriya', 'az-Cyrl': 'сурија', 'az-Latn': 'suriya', 'bal-Latn': 'Siriek', 'be': 'сірыйская', 'be-tarask': 'сірыйская', 'bg': 'сирийски', 'bgn': 'سریانی', 'bn': 'সিরিয়াক', 'br': 'sirieg', 'brx': 'सिरिआक', 'bs': 'sirijski', 'bs-Cyrl': 'сиријски', 'bs-Latn': 'sirijski', 'ca': 'siríac', 'ca-ES-valencia': 'siríac', 'ccp': '𑄥𑄨𑄢𑄨𑄠𑄇𑄴', 'ce': 'шемахойн', 'chr': 'ᏏᎵᎠᎩ', 'ckb': 'سریانی', 'cs': 'syrština', 'cy': 'Syrieg', 'da': 'syrisk', 'de': 'Syrisch', 'dsb': 'syriacšćina', 'el': 'Συριακά', 'el-polyton': 'Συριακά', 'en': 'Syriac', 'es': 'siriaco', 'et': 'süüria', 'eu': 'asiriera', 'fa': 'سریانی', 'ff-Adlm': '𞤅𞤭𞥅𞤪𞤭𞤴𞤢𞤳𞤪𞤫', 'fi': 'syyria', 'fil': 'Syriac', 'fo': 'syriac', 'fr': 'syriaque', 'fy': 'Syrysk', 'ga': 'Siricis', 'gd': 'Suraidheac', 'gl': 'siríaco', 'gsw': 'Syrisch', 'gu': 'સિરિએક', 'ha': 'Syriac', 'he': 'סורית', 'hi': 'सिरिएक', 'hi-Latn': 'Syriac', 'hr': 'sirijski', 'hsb': 'syriacšćina', 'ht': 'syriaque', 'hu': 'szír', 'hy': 'ասորերեն', 'ia': 'syriaco', 'id': 'Suriah', 'ig': 'Sirịak', 'is': 'sýrlenska', 'it': 'siriaco', 'ja': 'シリア語', 'jv': 'Siriak', 'ka': 'სირიული', 'kgp': 'sirijako', 'kk': 'сирия тілі', 'kk-Cyrl': 'сирия тілі', 'km': 'ស៊ីរី', 'kn': 'ಸಿರಿಯಾಕ್', 'ko': '시리아어', 'kok': 'सिरियाक', 'kok-Deva': 'सिरियाक', 'ks': 'سیٖریٲیی', 'ks-Arab': 'سیٖریٲیی', 'ksh': 'Sührejakkesch', 'ku': 'siryanî', 'ky': 'сирияча', 'lb': 'Syresch', 'lo': 'ຊີເລຍ', 'lt': 'sirų', 'lv': 'sīriešu', 'mai': 'सीरियक', 'mi': 'Hīriaka', 'mk': 'сириски', 'ml': 'സുറിയാനി', 'mn': 'сири', 'mr': 'सिरियाक', 'ms': 'Syriac', 'mt': 'Sirjan', 'my': 'ဆီးရီးယား', 'nb': 'syriakisk', 'ne': 'सिरियाक', 'nl': 'Syrisch', 'nn': 'syrisk', 'no': 'syriakisk', 'or': 'ସିରିଆକ୍\u200c', 'pa': 'ਸੀਰੀਆਈ', 'pa-Guru': 'ਸੀਰੀਆਈ', 'pcm': 'Síriák Lángwej', 'pl': 'syryjski', 'ps': 'سوریاني', 'pt': 'siríaco', 'qu': 'Siriaco Simi', 'rm': 'siric', 'ro': 'siriacă', 'ru': 'сирийский', 'sc': 'sirìacu', 'sd': 'شامي', 'sd-Arab': 'شامي', 'si': 'ස්\u200dරයෑක්', 'sk': 'sýrčina', 'sl': 'sirščina', 'smn': 'syyriakielâ', 'so': 'Af-Siriyak', 'sq': 'siriakisht', 'sr': 'сиријски', 'sr-Cyrl': 'сиријски', 'sr-Latn': 'sirijski', 'sv': 'syriska', 'sw': 'Lugha ya Syriac', 'syr': 'ܣܘܪܝܝܐ', 'ta': 'சிரியாக்', 'te': 'సిరియాక్', 'tg': 'суриёнӣ', 'th': 'ซีเรีย', 'ti': 'ሶርያኛ', 'tk': 'siriýa dili', 'to': 'lea fakasuliāiā', 'tr': 'Süryanice', 'tt': 'сүрия', 'ug': 'سۇرىيەچە', 'uk': 'сирійська', 'ur': 'سریانی', 'uz': 'suryoniy', 'uz-Cyrl': 'сурияча', 'uz-Latn': 'suryoniy', 'vec': 'sirìago', 'vi': 'Tiếng Syriac', 'wo': 'Siryak', 'yo': 'Èdè Siriaki', 'yrl': 'siriaku', 'yue': '敘利亞文', 'yue-Hans': '叙利亚文', 'yue-Hant': '敘利亞文', 'zh': '叙利亚语', 'zh-Hans': '叙利亚语', 'zh-Hant': '敘利亞文', 'zu': 'isi-Syriac'},
    'sys': {'en': 'Sinyar'},
    'syw': {'en': 'Kagate'},
    'syx': {'en': 'Samay'},
    'syy': {'en': 'Al-Sayyid Bedouin Sign Language'},
    'sza': {'en': 'Semelai'},
    'szb': {'en': 'Ngalum'},
    'szc': {'en': 'Semaq Beri'},
    'szd': {'en': 'Seru'},
    'sze': {'en': 'Seze'},
    'szg': {'en': 'Sengele'},
    'szl': {'af': 'Silesies', 'ak': 'Silesiafoɔ Kasa', 'am': 'ሲሌሲያኛ', 'ar': 'السيليزية', 'as': 'ছাইলেছিয়ান', 'ast': 'silesianu', 'az': 'silez dili', 'az-Latn': 'silez dili', 'bal-Latn': 'Selisi', 'be': 'сілезская', 'be-tarask': 'сілезская', 'bg': 'силезийски', 'bn': 'সিলেশিয়ান', 'br': 'silezieg', 'bs': 'šleski', 'bs-Latn': 'šleski', 'ca': 'silesià', 'ca-ES-valencia': 'silesià', 'chr': 'ᏏᎴᏏᏂ', 'cs': 'slezština', 'cy': 'Silesieg', 'da': 'schlesisk', 'de': 'Schlesisch (Wasserpolnisch)', 'dsb': 'šlazyńšćina', 'el': 'Σιλεσικά', 'el-polyton': 'Σιλεσικά', 'en': 'Silesian', 'es': 'silesio', 'et': 'sileesia', 'eu': 'silesiera', 'fa': 'سیلزیایی', 'fi': 'sleesia', 'fil': 'Silesian', 'fo': 'silesiskt', 'fr': 'silésien', 'ga': 'Siléisis', 'gd': 'Sileisis', 'gl': 'silesiano', 'gu': 'સિલેસ્યિન', 'ha': 'Silessiyanci', 'he': 'שלזית', 'hi': 'सायलिज़ियन', 'hi-Latn': 'Silesian', 'hr': 'šleski', 'hsb': 'šlezišćina', 'ht': 'silésien', 'hu': 'sziléziai', 'hy': 'սիլեզերեն', 'ia': 'silesiano', 'id': 'Silesia', 'ig': 'Asụsụ Sileshia', 'is': 'slesíska', 'it': 'slesiano', 'ja': 'シレジア語', 'jv': 'Silesia', 'ka': 'სილესიური', 'kk': 'силез тілі', 'kk-Cyrl': 'силез тілі', 'km': 'ស៊ីឡេស៊ី', 'kn': 'ಸಿಲೆಸಿಯನ್', 'ko': '실레시아어', 'kok': 'सिलेसियान', 'kok-Deva': 'सिलेसियान', 'ku': 'silesî', 'ky': 'силесче', 'lb': 'Schlesesch', 'lo': 'ຊີເລສຊຽນ', 'lt': 'sileziečių', 'lv': 'silēziešu', 'mk': 'шлезиски', 'ml': 'സൈലേഷ്യൻ', 'mn': 'силез', 'mr': 'सिलेशियन', 'ms': 'Silesia', 'my': 'စလီရှန်', 'nb': 'schlesisk', 'ne': 'सिलेसियाली', 'nl': 'Silezisch', 'nn': 'schlesisk', 'no': 'schlesisk', 'or': 'ସାଇଲେସିଆନ୍\u200c', 'pa': 'ਸਿਲੇਸੀਅਨ', 'pa-Guru': 'ਸਿਲੇਸੀਅਨ', 'pcm': 'Silesián Lángwej', 'pl': 'śląski', 'ps': 'سیلیسیان', 'pt': 'silesiano', 'ro': 'sileziană', 'ru': 'силезский', 'sc': 'silesianu', 'sd': 'سليسيئن', 'sd-Arab': 'سليسيئن', 'si': 'සිලේසියානු', 'sk': 'sliezština', 'sl': 'šlezijščina', 'so': 'Sileshiyaan', 'sq': 'silesisht', 'sr': 'силежански', 'sr-Cyrl': 'силежански', 'sr-Latn': 'siležanski', 'sv': 'silesiska', 'sw': 'Kisilesia', 'szl': 'ślōnski', 'ta': 'சிலேசியன்', 'te': 'సైలీషియన్', 'th': 'ไซลีเซีย', 'ti': 'ሲሌሲያን', 'tk': 'silez dili', 'to': 'lea fakasilesia', 'tr': 'Silezyaca', 'uk': 'сілезька', 'ur': 'سیلیزیائی', 'uz': 'silez', 'uz-Latn': 'silez', 'vec': 'silezian', 'vi': 'Tiếng Silesia', 'yi': 'שלעזיש', 'yo': 'Silìṣíànì', 'yue': '西利西亞文', 'yue-Hans': '西利西亚文', 'yue-Hant': '西利西亞文', 'zh': '西里西亚语', 'zh-Hans': '西里西亚语', 'zh-Hant': '西利西亞文', 'zu': 'iSilesian'},
    'szn': {'en': 'Sula'},
    'szp': {'en': 'Suabo'},
    'szs': {'en': 'Solomon Islands Sign Language'},
    'szv': {'en': 'Isu (Fako Division)'},
    'szw': {'en': 'Sawai'},
    'szy': {'en': 'Sakizaya'},
    'ta': {'af': 'Tamil', 'agq': 'Tamì', 'ak': 'Tamil kasa', 'am': 'ታሚል', 'ar': 'التاميلية', 'as': 'তামিল', 'asa': 'Kitamil', 'ast': 'tamil', 'az': 'tamil', 'az-Arab': 'تامیلجه', 'az-Cyrl': 'тамил', 'az-Latn': 'tamil', 'bal-Latn': 'Támel', 'bas': 'Hɔp u tamu᷆l', 'be': 'тамільская', 'be-tarask': 'тамільская', 'bem': 'Ichi Tamil', 'bez': 'Hitamil', 'bg': 'тамилски', 'bgn': 'تامیلی', 'bm': 'tamulikan', 'bn': 'তামিল', 'br': 'tamileg', 'brx': 'तामिल', 'bs': 'tamilski', 'bs-Cyrl': 'тамилски', 'bs-Latn': 'tamilski', 'ca': 'tàmil', 'ca-ES-valencia': 'tàmil', 'ccp': '𑄖𑄟𑄨𑄣𑄴', 'ce': 'тамилхойн', 'cgg': 'Orutamiri', 'chr': 'ᏔᎻᎵ', 'ckb': 'تامیلی', 'cs': 'tamilština', 'cy': 'Tamileg', 'da': 'tamil', 'dav': 'Kitamil', 'de': 'Tamil', 'dje': 'Tamil senni', 'dsb': 'tamilšćina', 'dyo': 'tamil', 'dz': 'ཏ་མིལ་ཁ', 'ebu': 'Kĩtamil', 'ee': 'tamilgbe', 'el': 'Ταμιλικά', 'el-polyton': 'Ταμιλικά', 'en': 'Tamil', 'eo': 'tamila', 'es': 'tamil', 'et': 'tamili', 'eu': 'tamilera', 'ewo': 'ǹkɔ́bɔ tamíl', 'fa': 'تامیلی', 'ff': 'Tamil', 'ff-Adlm': '𞤚𞤢𞤥𞤵𞤤𞤪𞤫', 'ff-Latn': 'Tamil', 'fi': 'tamili', 'fil': 'Tamil', 'fo': 'tamilskt', 'fr': 'tamoul', 'fur': 'tamil', 'fy': 'Tamil', 'ga': 'Tamailis', 'gd': 'Taimilis', 'gl': 'támil', 'gsw': 'Tamilisch', 'gu': 'તમિલ', 'guz': 'Kitamil', 'ha': 'Tamil', 'he': 'טמילית', 'hi': 'तमिल', 'hi-Latn': 'Tamil', 'hr': 'tamilski', 'hsb': 'tamilšćina', 'ht': 'tamoul', 'hu': 'tamil', 'hy': 'թամիլերեն', 'ia': 'tamil', 'id': 'Tamil', 'ig': 'Tamil', 'is': 'tamílska', 'it': 'tamil', 'ja': 'タミル語', 'jmc': 'Kyitamil', 'jv': 'Tamil', 'ka': 'ტამილური', 'kab': 'Taṭamulit', 'kam': 'Kitamil', 'kde': 'Chitamil', 'kea': 'tamil', 'kgp': 'támĩr', 'khq': 'Tamil senni', 'ki': 'Kitamil', 'kk': 'тамил тілі', 'kk-Cyrl': 'тамил тілі', 'kln': 'kutitab Tamil', 'km': 'តាមីល', 'kn': 'ತಮಿಳು', 'ko': '타밀어', 'kok': 'तमिळ', 'kok-Deva': 'तमिळ', 'ks': 'تَمِل', 'ks-Arab': 'تَمِل', 'ksb': 'Kitamil', 'ksf': 'ritamúl', 'ksh': 'Tamihlesch', 'ku': 'tamîlî', 'kxv': 'tamiḷ', 'kxv-Deva': 'तमिल', 'kxv-Latn': 'tamiḷ', 'kxv-Orya': 'ତାମିଲ୍', 'kxv-Telu': 'తమిళము', 'ky': 'тамилче', 'lag': 'Kɨtamíili', 'lb': 'Tamilesch', 'lg': 'Lutamiiru', 'lkt': 'Tamil Iyápi', 'lld': 'tamilich', 'ln': 'litamuli', 'lo': 'ທາມິລ', 'lrc': 'تامیل', 'lt': 'tamilų', 'lu': 'Mtamuili', 'luo': 'Kitamil', 'luy': 'Kitamil', 'lv': 'tamilu', 'mai': 'तमिल', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ ltamil', 'mer': 'Kĩtamilu', 'mfe': 'tamoul', 'mg': 'Tamoila', 'mgh': 'Itamil', 'mi': 'Tamira', 'mk': 'тамилски', 'ml': 'തമിഴ്', 'mn': 'тамил', 'mr': 'तामिळ', 'ms': 'Tamil', 'mt': 'Tamil', 'mua': 'Tamul', 'my': 'တမီးလ်', 'mzn': 'تامیلی', 'naq': 'Tamilǁî gowab', 'nb': 'tamil', 'nd': 'isi-Thamil', 'ne': 'तामिल', 'nl': 'Tamil', 'nmg': 'Kiɛl tamul', 'nn': 'tamil', 'no': 'tamil', 'nqo': 'ߕߊߡߎߟߌߞߊ߲', 'nus': 'Thok tamilni', 'nyn': 'Orutamiri', 'om': 'Afaan Tamilii', 'or': 'ତାମିଲ୍', 'pa': 'ਤਮਿਲ', 'pa-Guru': 'ਤਮਿਲ', 'pcm': 'tàmil', 'pl': 'tamilski', 'ps': 'تامل', 'pt': 'tâmil', 'qu': 'Tamil Simi', 'rm': 'tamil', 'rn': 'Igitamili', 'ro': 'tamilă', 'rof': 'Kitamil', 'ru': 'тамильский', 'rw': 'Igitamili', 'rwk': 'Kyitamil', 'sah': 'Тамыллыы', 'saq': 'Kitamil', 'sbp': 'Ishitamili', 'sc': 'tamil', 'sd': 'تامل', 'sd-Arab': 'تامل', 'seh': 'tâmil', 'ses': 'Tamil senni', 'sg': 'Tämûli', 'shi': 'ⵜⴰⵜⴰⵎⵉⵍⵜ', 'shi-Latn': 'Tatamilt', 'shi-Tfng': 'ⵜⴰⵜⴰⵎⵉⵍⵜ', 'si': 'දෙමළ', 'sk': 'tamilčina', 'sl': 'tamilščina', 'smn': 'tamilkielâ', 'sn': 'chiTamil', 'so': 'Tamiil', 'sq': 'tamilisht', 'sr': 'тамилски', 'sr-Cyrl': 'тамилски', 'sr-Latn': 'tamilski', 'sv': 'tamil', 'sw': 'Kitamili', 'ta': 'தமிழ்', 'te': 'తమిళం', 'teo': 'Kitamil', 'tg': 'тамилӣ', 'th': 'ทมิฬ', 'ti': 'ታሚል', 'tk': 'tamil dili', 'to': 'lea fakatamili', 'tr': 'Tamilce', 'tt': 'тамил', 'twq': 'Tamil senni', 'tzm': 'Tatamilt', 'ug': 'تامىلچە', 'uk': 'тамільська', 'ur': 'تمل', 'uz': 'tamil', 'uz-Cyrl': 'тамилча', 'uz-Latn': 'tamil', 'vai': 'ꕚꕆꔷ', 'vai-Latn': 'Tamíli', 'vai-Vaii': 'ꕚꕆꔷ', 'vec': 'tamil', 'vi': 'Tiếng Tamil', 'vun': 'Kyitamil', 'wae': 'Tamiliš', 'wo': 'Tamil', 'xnr': 'तमिल', 'xog': 'Olutamiiru', 'yav': 'nutámule', 'yi': 'טאַמיל', 'yo': 'Èdè Tamili', 'yrl': 'tamiu', 'yue': '坦米爾文', 'yue-Hans': '坦米尔文', 'yue-Hant': '坦米爾文', 'zgh': 'ⵜⴰⵜⴰⵎⵉⵍⵜ', 'zh': '泰米尔语', 'zh-Hans': '泰米尔语', 'zh-Hant': '坦米爾文', 'zu': 'isi-Tamil'},
    'taa': {'en': 'Lower Tanana'},
    'tab': {'en': 'Tabassaran'},
    'tac': {'en': 'Lowland Tarahumara'},
    'tad': {'en': 'Tause'},
    'tae': {'en': 'Tariana'},
    'taf': {'en': 'Tapirapé'},
    'tag': {'en': 'Tagoi'},
    'tai': {'en': 'Tai languages'},
    'tai-pro': {'en': 'Proto-Tai'},
    'tai-swe-pro': {'en': 'Proto-Southwestern Tai'},
    'taj': {'en': 'Eastern Tamang'},
    'tak': {'en': 'Tala'},
    'tal': {'en': 'Tal'},
    'tan': {'en': 'Tangale'},
    'tao': {'en': 'Yami'},
    'tap': {'en': 'Taabwa'},
    'taq': {'en': 'Tamasheq'},
    'tar': {'en': 'Central Tarahumara'},
    'tas': {'en': 'Tay Boi'},
    'tau': {'en': 'Upper Tanana'},
    'tav': {'en': 'Tatuyo'},
    'taw': {'en': 'Kalam-Tai'},
    'tax': {'en': 'Tamki'},
    'tay': {'en': 'Atayal'},
    'taz': {'en': 'Tocho'},
    'tba': {'en': 'Aikanã'},
    'tbb': {'en': 'Tapeba'},
    'tbc': {'en': 'Takia'},
    'tbd': {'en': 'Kaki Ae'},
    'tbe': {'en': 'Tanimbili'},
    'tbf': {'en': 'Mandara'},
    'tbg': {'en': 'North Tairora'},
    'tbh': {'en': 'Dharawal'},
    'tbi': {'en': 'Gaam'},
    'tbj': {'en': 'Tiang'},
    'tbk': {'en': 'Calamian Tagbanwa'},
    'tbl': {'en': 'Tboli'},
    'tbm': {'en': 'Tagbu'},
    'tbn': {'en': 'Barro Negro Tunebo'},
    'tbo': {'en': 'Tawala'},
    'tbp': {'en': 'Taworta'},
    'tbq': {'en': 'Tibeto-Burman languages'},
    'tbq-kuk-pro': {'en': 'Proto-Kuki-Chin'},
    'tbq-lol-pro': {'en': 'Proto-Loloish'},
    'tbq-pro': {'en': 'Proto-Tibeto-Burman'},
    'tbr': {'en': 'Tumtum'},
    'tbs': {'en': 'Tanguat'},
    'tbt': {'en': 'Tembo (Kitembo)'},
    'tbu': {'en': 'Tubar'},
    'tbv': {'en': 'Tobo'},
    'tbw': {'en': 'Tagbanwa'},
    'tbx': {'en': 'Kapin'},
    'tby': {'en': 'Tabaru'},
    'tbz': {'en': 'Ditammari'},
    'tca': {'en': 'Ticuna'},
    'tcb': {'en': 'Tanacross'},
    'tcc': {'en': 'Datooga'},
    'tcd': {'en': 'Tafi'},
    'tce': {'af': 'Suid-Tutchone', 'am': 'ደቡባዊ ቱትቾን', 'ar': 'التوتشون الجنوبية', 'as': 'দাক্ষিণাত্যৰ টুটচ’ন', 'az': 'cənubi tuçon', 'az-Latn': 'cənubi tuçon', 'be': 'паўднёвая тутчонэ', 'be-tarask': 'паўднёвая тутчонэ', 'bg': 'южен тучоне', 'bn': 'দক্ষিণী টুচোন', 'br': 'tutchoneg ar Su', 'brx': 'खोलायारि तात्चने', 'bs': 'južni tučoni', 'bs-Cyrl': 'јужни тутчоне', 'bs-Latn': 'južni tučoni', 'ca': 'tutxone meridional', 'ca-ES-valencia': 'tutxone meridional', 'chr': 'ᏧᎦᏃᏮ ᏚᏦᏁ', 'cs': 'tutčonština (jižní)', 'cy': 'Tutchone Deheuol', 'da': 'sydtutchone', 'de': 'Südliches Tutchone', 'dsb': 'pódpołdnjowa tutchonšćina', 'el': 'Νότια Τουτσόνε', 'el-polyton': 'Νότια Τουτσόνε', 'en': 'Southern Tutchone', 'es': 'tutchone meridional', 'et': 'lõunatutšoni', 'eu': 'hegoaldeko tutchoneera', 'fa': 'توچون جنوبی', 'ff-Adlm': '𞤚𞤵𞤼𞤳𞤮𞤲𞤭𞥅𞤪𞤫 𞤙𞤢𞥄𞤥𞤲𞤢𞥄𞤲𞤺𞤫', 'fi': 'etelätutchone', 'fil': 'Katimugang Tutchone', 'fr': 'tutchone du Sud', 'ga': 'Tutchone an Deiscirt', 'gd': 'Tutchone Dheasach', 'gl': 'tutchone do sur', 'gu': 'દક્ષિણ ટુચૉન', 'ha': 'Tutchone na Kudanci', 'he': 'טצ׳ון דרומית', 'hi': 'दक्षिणी टशोनी', 'hi-Latn': 'Southern Tutchone', 'hr': 'južni tutchone', 'hsb': 'južna tutchonšćina', 'ht': 'tutchone du Sud', 'hu': 'déli tutchone', 'hy': 'հարավային թուտչոնե', 'ia': 'tutchone del sud', 'id': 'Tutchone Selatan', 'ig': 'Southern Tutchone', 'is': 'suður-tutchone', 'it': 'tutchone meridionale', 'ja': '南部トゥショーニ語', 'jv': 'Tutkhone Sisih Kidul', 'ka': 'სამხრეთ ტუჩონი', 'kk': 'оңтүстік тутчоне тілі', 'kk-Cyrl': 'оңтүстік тутчоне тілі', 'km': 'ថុចឆុនខាងត្បូង', 'kn': 'ದಕ್ಷಿಣ ಟಚ್\u200cವನ್', 'ko': '남부 투톤어', 'kok': 'दक्षिणी तुत्चोने', 'kok-Deva': 'दक्षिणी तुत्चोने', 'ku': 'totuçena başûrî', 'ky': 'түштүк тутчонече (индей тили)', 'lo': 'ທຸດຊອນໃຕ້', 'lt': 'pietų tučonų', 'lv': 'dienvidtutčonu', 'mai': 'दक्षिणी टोचेन', 'mi': 'Tatōne ki te Tonga', 'mk': 'јужнотучонски', 'ml': 'സതേൺ ടറ്റ്ഷോൺ', 'mn': 'өмнөд тутчоне', 'mr': 'दक्षिणात्य टचोन', 'ms': 'Tutchone Selatan', 'my': 'တောင် တပ်ချွန်', 'nb': 'sørlig tutchone', 'ne': 'दक्षिनी टुट्चोन', 'nl': 'Zuid-Tutchone', 'nn': 'sørleg tutchone', 'no': 'sørlig tutchone', 'or': 'ସାଉଥ୍ ଟଚୋନ୍', 'pa': 'ਦੱਖਣੀ ਟਚੋਨ', 'pa-Guru': 'ਦੱਖਣੀ ਟਚੋਨ', 'pcm': 'Saútán Tutchónẹ Lángwej', 'pl': 'tutchone południowy', 'ps': 'جنوبي توچون', 'pt': 'tutchone do sul', 'qu': 'Tutchone Meridional', 'ro': 'tutchone de sud', 'ru': 'южный тутчоне', 'sc': 'tutchone meridionale', 'sd': 'ڏاکڻي ٽچون', 'sd-Arab': 'ڏاکڻي ٽچون', 'si': 'දකුණු ටචෝන්', 'sk': 'tutchone (juh)', 'sl': 'južna tučonščina', 'so': 'Southern Tutchone', 'sq': 'tatshonishte jugore', 'sr': 'јужни тачон', 'sr-Cyrl': 'јужни тачон', 'sr-Latn': 'južni tačon', 'sv': 'sydtutchone', 'sw': 'Kitutchone cha Kusini', 'ta': 'தெற்கு டட்சோன்', 'te': 'దక్షిణ టుట్చోన్', 'th': 'ทัชโชนใต้', 'ti': 'ደቡባዊ ታትቾን', 'tk': 'günorta tutçone dili', 'to': 'lea fakatutisone-tonga', 'tr': 'Güney Tuçoncası', 'uk': 'південна тутчон', 'ur': 'جنوبی ٹچون', 'uz': 'janubiy tutchone', 'uz-Latn': 'janubiy tutchone', 'vec': 'tutchone meridionale', 'vi': 'Tiếng Tutchone miền Nam', 'yo': 'Èdè Gusu Tushoni', 'yue': '南塔穹文', 'yue-Hans': '南塔穹文', 'yue-Hant': '南塔穹文', 'zh': '南塔穹语', 'zh-Hans': '南塔穹语', 'zh-Hant': '南塔穹語', 'zu': 'Southern Tutchone'},
    'tcf': {'en': "Malinaltepec Me'phaa"},
    'tcg': {'en': 'Tamagario'},
    'tch': {'en': 'Turks And Caicos Creole English'},
    'tci': {'en': 'Wára'},
    'tck': {'en': 'Tchitchege'},
    'tcl': {'en': 'Taman (Myanmar)'},
    'tcm': {'en': 'Tanahmerah'},
    'tcn': {'en': 'Tichurong'},
    'tco': {'en': 'Taungyo'},
    'tcp': {'en': 'Tawr Chin'},
    'tcq': {'en': 'Kaiy'},
    'tcs': {'en': 'Torres Strait Creole'},
    'tct': {'en': "T'en"},
    'tcu': {'en': 'Southeastern Tarahumara'},
    'tcw': {'en': 'Tecpatlán Totonac'},
    'tcx': {'en': 'Toda'},
    'tcy': {'ast': 'tulu', 'br': 'touloueg', 'cs': 'tuluština', 'cy': 'Tulu', 'de': 'Tulu', 'en': 'Tulu', 'et': 'tulu', 'fi': 'tulu', 'fr': 'toulou', 'gd': 'Tulu', 'gu': 'તુલુ', 'hi-Latn': 'Tulu', 'ht': 'toulou', 'hy': 'տուլու', 'id': 'Tulu', 'it': 'tulu', 'ja': 'トゥル語', 'lb': 'Tulu', 'lt': 'tulų', 'mk': 'тулу', 'nb': 'tulu', 'nl': 'Tulu', 'nn': 'tulu', 'no': 'tulu', 'pl': 'tulu', 'sv': 'tulu', 'te': 'తుళు', 'th': 'ตูลู', 'to': 'lea fakatulu', 'tr': 'Tuluca', 'yue': '圖盧文', 'yue-Hans': '图卢文', 'yue-Hant': '圖盧文', 'zh-Hant': '圖盧文'},
    'tcz': {'en': 'Thado Chin'},
    'tda': {'en': 'Tagdal'},
    'tdb': {'en': 'Panchpargania'},
    'tdc': {'en': 'Emberá-Tadó'},
    'tdd': {'en': 'Tai Nüa'},
    'tde': {'en': 'Tiranige Diga Dogon'},
    'tdf': {'en': 'Talieng'},
    'tdg': {'en': 'Western Tamang'},
    'tdh': {'en': 'Thulung'},
    'tdi': {'en': 'Tomadino'},
    'tdj': {'en': 'Tajio'},
    'tdk': {'en': 'Tambas'},
    'tdl': {'en': 'Sur'},
    'tdm': {'en': 'Taruma'},
    'tdn': {'en': 'Tondano'},
    'tdo': {'en': 'Teme'},
    'tdq': {'en': 'Tita'},
    'tdr': {'en': 'Todrah'},
    'tds': {'en': 'Doutai'},
    'tdt': {'en': 'Tetun Dili'},
    'tdu': {'en': 'Tempasuk Dusun'},
    'tdv': {'en': 'Toro'},
    'tdx': {'en': 'Tandroy-Mahafaly Malagasy'},
    'tdy': {'en': 'Tadyawan'},
    'te': {'af': 'Teloegoe', 'ak': 'Telugu', 'am': 'ተሉጉ', 'ar': 'التيلوغوية', 'as': 'তেলুগু', 'ast': 'telugu', 'az': 'teluqu', 'az-Cyrl': 'телугу', 'az-Latn': 'teluqu', 'bal-Latn': 'Telgó', 'be': 'тэлугу', 'be-tarask': 'тэлугу', 'bg': 'телугу', 'bgn': 'تلوگویی', 'bn': 'তেলুগু', 'br': 'telougou', 'brx': 'तेलुगु', 'bs': 'telugu', 'bs-Cyrl': 'телугу', 'bs-Latn': 'telugu', 'ca': 'telugu', 'ca-ES-valencia': 'telugu', 'ccp': '𑄖𑄬𑄣𑄬𑄉𑄪', 'ce': 'телугу', 'chr': 'ᏖᎷᎦ', 'ckb': 'تێلووگوو', 'cs': 'telugština', 'cy': 'Telugu', 'da': 'telugu', 'de': 'Telugu', 'dsb': 'telugšćina', 'dz': 'ཏེ་ལུ་གུ་ཁ', 'ee': 'telegugbe', 'el': 'Τελούγκου', 'el-polyton': 'Τελούγκου', 'en': 'Telugu', 'eo': 'telugua', 'es': 'telugu', 'et': 'telugu', 'eu': 'telugua', 'fa': 'تلوگویی', 'ff-Adlm': '𞤚𞤫𞤤𞤫𞤺𞤵𞥅𞤪𞤫', 'fi': 'telugu', 'fil': 'Telugu', 'fo': 'telugu', 'fr': 'télougou', 'fur': 'telegu', 'fy': 'Telugu', 'ga': 'Teileagúis', 'gd': 'Telugu', 'gl': 'telugu', 'gsw': 'Telugu', 'gu': 'તેલુગુ', 'ha': 'Telugu', 'he': 'טלוגו', 'hi': 'तेलुगू', 'hi-Latn': 'Telugu', 'hr': 'teluški', 'hsb': 'telugu', 'ht': 'télougou', 'hu': 'telugu', 'hy': 'թելուգու', 'ia': 'telugu', 'id': 'Telugu', 'ig': 'Telugu', 'is': 'telúgú', 'it': 'telugu', 'ja': 'テルグ語', 'jv': 'Telugu', 'ka': 'ტელუგუ', 'kea': 'telugu', 'kgp': 'térugo', 'kk': 'телугу тілі', 'kk-Cyrl': 'телугу тілі', 'km': 'តេលុគុ', 'kn': 'ತೆಲುಗು', 'ko': '텔루구어', 'kok': 'तेलुगू', 'kok-Deva': 'तेलुगू', 'ks': 'تیلگوٗ', 'ks-Arab': 'تیلگوٗ', 'ksh': 'Telluhju', 'ku': 'telûgûyî', 'kxv': 'telugu', 'kxv-Deva': 'तेलुगू', 'kxv-Latn': 'telugu', 'kxv-Orya': 'ତେଲୁଗୁ', 'kxv-Telu': 'తెలుగు', 'ky': 'телугуча', 'lb': 'Telugu', 'lkt': 'Telugu Iyápi', 'lld': 'telugu', 'lo': 'ເຕລູກູ', 'lrc': 'تئلئگو', 'lt': 'telugų', 'lv': 'telugu', 'mai': 'तेलगू', 'mi': 'Teruku', 'mk': 'телугу', 'ml': 'തെലുങ്ക്', 'mn': 'тэлүгү', 'mr': 'तेलगू', 'ms': 'Telugu', 'mt': 'Telugu', 'my': 'တီလီဂူ', 'mzn': 'تلوگویی', 'nb': 'telugu', 'ne': 'तेलुगु', 'nl': 'Telugu', 'nn': 'telugu', 'no': 'telugu', 'nqo': 'ߕߋߟߎߜ߭ߎߞߊ߲', 'om': 'Afaan Telugu', 'or': 'ତେଲୁଗୁ', 'pa': 'ਤੇਲਗੂ', 'pa-Guru': 'ਤੇਲਗੂ', 'pcm': 'Tẹlugu', 'pl': 'telugu', 'ps': 'تېليګو', 'pt': 'télugo', 'qu': 'Telugu Simi', 'rm': 'telugu', 'ro': 'telugu', 'ru': 'телугу', 'rw': 'Igitelugu', 'sah': 'Төлүгүлүү', 'sc': 'telugu', 'sd': 'تلگو', 'sd-Arab': 'تلگو', 'si': 'තෙළිඟු', 'sk': 'telugčina', 'sl': 'telugijščina', 'smn': 'telugu', 'so': 'Teluugu', 'sq': 'teluguisht', 'sr': 'телугу', 'sr-Cyrl': 'телугу', 'sr-Latn': 'telugu', 'sv': 'telugu', 'sw': 'Kitelugu', 'ta': 'தெலுங்கு', 'te': 'తెలుగు', 'tg': 'телугу', 'th': 'เตลูกู', 'ti': 'ተሉጉ', 'tk': 'telugu dili', 'to': 'lea fakaʻinitia-teluku', 'tr': 'Telugu dili', 'tt': 'телугу', 'ug': 'تېلۇگۇچە', 'uk': 'телугу', 'ur': 'تیلگو', 'uz': 'telugu', 'uz-Cyrl': 'телугу', 'uz-Latn': 'telugu', 'vec': 'telugo', 'vi': 'Tiếng Telugu', 'wae': 'Telugu', 'wo': 'Telugu', 'xnr': 'तेलुगू', 'yo': 'Èdè Telugu', 'yrl': 'térugu', 'yue': '泰盧固文', 'yue-Hans': '泰卢固文', 'yue-Hant': '泰盧固文', 'zh': '泰卢固语', 'zh-Hans': '泰卢固语', 'zh-Hant': '泰盧固文', 'zu': 'isi-Telugu'},
    'tea': {'en': 'Temiar'},
    'teb': {'en': 'Tetete'},
    'tec': {'en': 'Terik'},
    'ted': {'en': 'Tepo Krumen'},
    'tee': {'en': 'Huehuetla Tepehua'},
    'tef': {'en': 'Teressa'},
    'teg': {'en': 'Teke-Tege'},
    'teh': {'en': 'Tehuelche'},
    'tei': {'en': 'Torricelli'},
    'tek': {'en': 'Ibali Teke'},
    'tem': {'af': 'Timne', 'am': 'ቲምኔ', 'ar': 'التيمن', 'as': 'টিম্নে', 'ast': 'timne', 'az': 'timne', 'az-Cyrl': 'тимне', 'az-Latn': 'timne', 'be': 'тэмнэ', 'be-tarask': 'тэмнэ', 'bg': 'темне', 'bgn': 'تیمنه\u200cای', 'bn': 'টাইম্নে', 'br': 'temne', 'brx': 'तीमने', 'bs': 'timne', 'bs-Cyrl': 'тимне', 'bs-Latn': 'timne', 'ca': 'temne', 'ca-ES-valencia': 'temne', 'ccp': '𑄑𑄭𑄟𑄴𑄚𑄬', 'ce': 'темне', 'chr': 'ᏘᎹᏁ', 'ckb': 'تیمنێ', 'cs': 'temne', 'cy': 'Timneg', 'da': 'temne', 'de': 'Temne', 'dsb': 'timnešćina', 'el': 'Τίμνε', 'el-polyton': 'Τίμνε', 'en': 'Timne', 'es': 'temne', 'et': 'temne', 'eu': 'temneera', 'fa': 'تمنه\u200cای', 'ff-Adlm': '𞤚𞤫𞤥𞤫𞤲𞤫𞥅𞤪𞤫', 'fi': 'temne', 'fil': 'Timne', 'fo': 'timne', 'fr': 'timné', 'fy': 'Timne', 'ga': 'Teimnis', 'gd': 'Timne', 'gl': 'temne', 'gsw': 'Temne', 'gu': 'ટિમ્ને', 'ha': 'Timne', 'he': 'טימנה', 'hi': 'टिम्ने', 'hi-Latn': 'Timne', 'hr': 'temne', 'hsb': 'timnešćina', 'ht': 'timné', 'hu': 'temne', 'hy': 'տեմնե', 'ia': 'temne', 'id': 'Timne', 'ig': 'Timne', 'is': 'tímne', 'it': 'temne', 'ja': 'テムネ語', 'jv': 'Timne', 'ka': 'ტინმე', 'kgp': 'timnẽ', 'kk': 'темне тілі', 'kk-Cyrl': 'темне тілі', 'km': 'ធីមនី', 'kn': 'ಟಿಮ್ನೆ', 'ko': '팀니어', 'kok': 'तिम्ने', 'kok-Deva': 'तिम्ने', 'ks': 'ٹِمنیے', 'ks-Arab': 'ٹِمنیے', 'ku': 'temne', 'ky': 'тимнече', 'lb': 'Temne', 'lo': 'ທີມເນ', 'lt': 'timne', 'lv': 'temnu', 'mai': 'टिमने', 'mi': 'Tīmene', 'mk': 'тимне', 'ml': 'ടിംനേ', 'mn': 'тимн', 'mr': 'टिम्ने', 'ms': 'Timne', 'mt': 'Timne', 'my': 'တင်မ်နဲ', 'nb': 'temne', 'ne': 'टिम्ने', 'nl': 'Timne', 'nn': 'temne', 'no': 'temne', 'or': 'ତିମନେ', 'pa': 'ਟਿਮਨੇ', 'pa-Guru': 'ਟਿਮਨੇ', 'pcm': 'Tímnẹ Lángwej', 'pl': 'temne', 'ps': 'تیمني', 'pt': 'timne', 'qu': 'Timne Simi', 'rm': 'temne', 'ro': 'timne', 'ru': 'темне', 'sc': 'temne', 'sd': 'تمني', 'sd-Arab': 'تمني', 'si': 'ටිම්නෙ', 'sk': 'temne', 'sl': 'temnejščina', 'smn': 'temnekielâ', 'so': 'Timne', 'sq': 'timneisht', 'sr': 'тимне', 'sr-Cyrl': 'тимне', 'sr-Latn': 'timne', 'sv': 'temne', 'sw': 'Kitemne', 'ta': 'டிம்னே', 'te': 'టిమ్నే', 'th': 'ทิมเน', 'ti': 'ቲምኔ', 'tk': 'temne dili', 'to': 'lea fakatimenē', 'tr': 'Timne', 'ug': 'تېمنېچە', 'uk': 'темне', 'ur': 'ٹمنے', 'uz': 'timne', 'uz-Latn': 'timne', 'vec': 'temne', 'vi': 'Tiếng Timne', 'xnr': 'टिम्ने', 'yo': 'Èdè Timne', 'yrl': 'timine', 'yue': '提姆文', 'yue-Hans': '提姆文', 'yue-Hant': '提姆文', 'zh': '泰姆奈语', 'zh-Hans': '泰姆奈语', 'zh-Hant': '提姆文', 'zu': 'isi-Timne'},
    'ten': {'en': 'Tama (Colombia)'},
    'teo': {'af': 'Teso', 'am': 'ቴሶ', 'ar': 'تيسو', 'as': 'তেছো', 'ast': 'teso', 'az': 'teso', 'az-Cyrl': 'тесо', 'az-Latn': 'teso', 'bal-Latn': 'Tésó', 'be': 'тэсо', 'be-tarask': 'тэсо', 'bg': 'тесо', 'bgn': 'تیسویی', 'bn': 'তেসো', 'br': 'tesoeg', 'brx': 'तेस’', 'bs': 'teso', 'bs-Cyrl': 'тесо', 'bs-Latn': 'teso', 'ca': 'teso', 'ca-ES-valencia': 'teso', 'ccp': '𑄖𑄬𑄥𑄮', 'ce': 'тесо', 'chr': 'ᏖᏐ', 'ckb': 'تێسوو', 'cs': 'teso', 'cy': 'Teso', 'da': 'teso', 'de': 'Teso', 'dsb': 'teso', 'el': 'Τέσο', 'el-polyton': 'Τέσο', 'en': 'Teso', 'es': 'teso', 'et': 'teso', 'eu': 'tesoera', 'fa': 'تسویی', 'ff-Adlm': '𞤚𞤫𞤧𞤮𞥅𞤪𞤫', 'fi': 'teso', 'fil': 'Teso', 'fo': 'teso', 'fr': 'teso', 'fy': 'Teso', 'ga': 'Teso', 'gd': 'Teso', 'gl': 'teso', 'gu': 'તેસો', 'ha': 'Teso', 'he': 'טסו', 'hi': 'टेसो', 'hi-Latn': 'Teso', 'hr': 'teso', 'hsb': 'teso', 'ht': 'teso', 'hu': 'teszó', 'hy': 'տեսո', 'ia': 'teso', 'id': 'Teso', 'ig': 'Teso', 'is': 'tesó', 'it': 'teso', 'ja': 'テソ語', 'jv': 'Teso', 'ka': 'ტესო', 'kgp': 'teso', 'kk': 'тесо тілі', 'kk-Cyrl': 'тесо тілі', 'km': 'តេសូ', 'kn': 'ಟೆಸೊ', 'ko': '테조어', 'kok': 'तेसो', 'kok-Deva': 'तेसो', 'ku': 'teso', 'ky': 'тесочо', 'lb': 'Teso', 'lo': 'ເຕໂຊ', 'lrc': 'تئسو', 'lt': 'teso', 'lv': 'teso', 'mai': 'टेसो', 'mi': 'Teho', 'mk': 'тесо', 'ml': 'ടെസോ', 'mn': 'тэсо', 'mr': 'तेसो', 'ms': 'Teso', 'mt': 'Teso', 'my': 'တီဆို', 'mzn': 'تسویی', 'nb': 'teso', 'ne': 'टेसो', 'nl': 'Teso', 'nn': 'teso', 'no': 'teso', 'nqo': 'ߕߋߛߏߞߊ߲', 'or': 'ତେସା', 'pa': 'ਟੇਸੋ', 'pa-Guru': 'ਟੇਸੋ', 'pcm': 'Tẹ́so', 'pl': 'ateso', 'ps': 'تیسو', 'pt': 'teso', 'qu': 'Teso Simi', 'rm': 'teso', 'ro': 'teso', 'ru': 'тесо', 'sc': 'teso', 'sd': 'تيسو', 'sd-Arab': 'تيسو', 'si': 'ටෙසෝ', 'sk': 'teso', 'sl': 'teso', 'smn': 'ateso', 'so': 'Teeso', 'sq': 'tesoisht', 'sr': 'тесо', 'sr-Cyrl': 'тесо', 'sr-Latn': 'teso', 'sv': 'teso', 'sw': 'Kiteso', 'ta': 'டெசோ', 'te': 'టెసో', 'teo': 'Kiteso', 'th': 'เตโซ', 'ti': 'ተሶ', 'tk': 'teso dili', 'to': 'lea fakateso', 'tr': 'Teso', 'ug': 'تېسوچە', 'uk': 'тесо', 'ur': 'تیسو', 'uz': 'teso', 'uz-Cyrl': 'тесо', 'uz-Latn': 'teso', 'vec': 'teso', 'vi': 'Tiếng Teso', 'xnr': 'टेसो', 'yo': 'Tẹ́sò', 'yrl': 'teso', 'yue': '特索文', 'yue-Hans': '特索文', 'yue-Hant': '特索文', 'zh': '特索语', 'zh-Hans': '特索语', 'zh-Hant': '特索文', 'zu': 'isi-Teso'},
    'tep': {'en': 'Tepecano'},
    'teq': {'en': 'Temein'},
    'ter': {'ar': 'التيرينو', 'ast': 'terena', 'az': 'tereno', 'az-Latn': 'tereno', 'bg': 'терено', 'bn': 'তেরেনো', 'br': 'tereno', 'brx': 'तेरेनो', 'bs': 'tereno', 'bs-Cyrl': 'терено', 'bs-Latn': 'tereno', 'ca': 'terena', 'ca-ES-valencia': 'terena', 'ccp': '𑄖𑄬𑄢𑄬𑄚𑄮', 'cs': 'tereno', 'cy': 'Terena', 'da': 'tereno', 'de': 'Tereno', 'el': 'Τερένο', 'el-polyton': 'Τερένο', 'en': 'Tereno', 'es': 'tereno', 'et': 'tereno', 'fa': 'ترنو', 'fi': 'tereno', 'fr': 'tereno', 'fy': 'Tereno', 'gd': 'Terêna', 'gsw': 'Tereno-Schpraach', 'gu': 'તેરેનો', 'he': 'טרנו', 'hi': 'तेरेनो', 'hi-Latn': 'Tereno', 'hr': 'tereno', 'ht': 'tereno', 'hu': 'terenó', 'hy': 'տերենո', 'id': 'Tereno', 'is': 'terenó', 'it': 'tereno', 'ja': 'テレーノ語', 'kgp': 'terẽnũ', 'kn': 'ಟೆರೆನೋ', 'ko': '테레노어', 'ks': 'ٹیرینو', 'ks-Arab': 'ٹیرینو', 'lb': 'Tereno-Sprooch', 'lo': 'ເຕເລໂນ', 'lt': 'Tereno', 'lv': 'tereno', 'mk': 'терено', 'ml': 'ടെറേനോ', 'mr': 'तेरेनो', 'mt': 'Tereno', 'nb': 'tereno', 'nl': 'Tereno', 'nn': 'tereno', 'no': 'tereno', 'or': 'ତେରେନୋ', 'pl': 'tereno', 'pt': 'tereno', 'rm': 'tereno', 'ro': 'tereno', 'ru': 'терено', 'sk': 'terêna', 'sr': 'терено', 'sr-Cyrl': 'терено', 'sr-Latn': 'tereno', 'sv': 'tereno', 'ta': 'டெரெனோ', 'te': 'టెరెనో', 'th': 'เทเรโน', 'to': 'lea fakateleno', 'tr': 'Tereno', 'ug': 'تېرېناچە', 'uk': 'терено', 'vi': 'Tiếng Tereno', 'yrl': 'terenu', 'yue': '泰雷諾文', 'yue-Hans': '泰雷诺文', 'yue-Hant': '泰雷諾文', 'zh': '特伦诺语', 'zh-Hans': '特伦诺语', 'zh-Hant': '泰雷諾文'},
    'tes': {'en': 'Tengger'},
    'tet': {'af': 'Tetoem', 'am': 'ቴተም', 'ar': 'التيتم', 'as': 'তেতুম', 'ast': 'tetum', 'az': 'tetum', 'az-Cyrl': 'тетум', 'az-Latn': 'tetum', 'be': 'тэтум', 'be-tarask': 'тэтум', 'bg': 'тетум', 'bgn': 'ٹیٹومی', 'bn': 'তেতুম', 'br': 'tetum', 'brx': 'तेतुम', 'bs': 'tetum', 'bs-Cyrl': 'тетум', 'bs-Latn': 'tetum', 'ca': 'tètum', 'ca-ES-valencia': 'tètum', 'ccp': '𑄖𑄬𑄖𑄪𑄟𑄴', 'ce': 'тетум', 'chr': 'ᏖᏚᎼ', 'ckb': 'تێتووم', 'cs': 'tetumština', 'cy': 'Tetumeg', 'da': 'tetum', 'de': 'Tetum', 'dsb': 'tetumšćina', 'ee': 'tetumgbe', 'el': 'Τέτουμ', 'el-polyton': 'Τέτουμ', 'en': 'Tetum', 'es': 'tetún', 'et': 'tetumi', 'eu': 'tetuma', 'fa': 'تتومی', 'ff-Adlm': '𞤚𞤫𞤼𞤵𞤥𞤪𞤫', 'fi': 'tetum', 'fil': 'Tetum', 'fo': 'tetum', 'fr': 'tétoum', 'fur': 'tetum', 'fy': 'Tetun', 'ga': 'Teitimis', 'gd': 'Tetum', 'gl': 'tetun', 'gsw': 'Tetum-Schpraach', 'gu': 'તેતુમ', 'ha': 'Tatum', 'he': 'טטום', 'hi': 'तेतुम', 'hi-Latn': 'Tetum', 'hr': 'tetum', 'hsb': 'tetumšćina', 'ht': 'tétoum', 'hu': 'tetum', 'hy': 'տետում', 'ia': 'tetum', 'id': 'Tetun', 'ig': 'Tetum', 'is': 'tetúm', 'it': 'tetum', 'ja': 'テトゥン語', 'jv': 'Tetum', 'ka': 'ტეტუმი', 'kgp': 'tétũm', 'kk': 'тетум тілі', 'kk-Cyrl': 'тетум тілі', 'km': 'ទីទុំ', 'kn': 'ಟೇಟಮ್', 'ko': '테툼어', 'kok': 'तेतम', 'kok-Deva': 'तेतम', 'ks': 'ٹیٹَم', 'ks-Arab': 'ٹیٹَم', 'ksh': 'Tetumsch', 'ku': 'tetûmî', 'ky': 'тетумча', 'lb': 'Tetum-Sprooch', 'lo': 'ເຕຕູມ', 'lt': 'tetum', 'lv': 'tetumu', 'mai': 'टैटम', 'mi': 'Tetumu', 'mk': 'тетум', 'ml': 'ടെറ്റും', 'mn': 'тетум', 'mr': 'तेतुम', 'ms': 'Tetum', 'mt': 'Tetum', 'my': 'တီတွမ်', 'nb': 'tetum', 'ne': 'टेटुम', 'nl': 'Tetun', 'nn': 'tetum', 'no': 'tetum', 'or': 'ତେତୁମ୍', 'pa': 'ਟੇਟਮ', 'pa-Guru': 'ਟੇਟਮ', 'pcm': 'Tẹ́tum Lángwej', 'pl': 'tetum', 'ps': 'تتوم', 'pt': 'tétum', 'qu': 'Tetum Simi', 'rm': 'tetum', 'ro': 'tetum', 'ru': 'тетум', 'sc': 'tetum', 'sd': 'تيتم', 'sd-Arab': 'تيتم', 'si': 'ටේටම්', 'sk': 'tetumčina', 'sl': 'tetumščina', 'smn': 'tetum', 'so': 'Tetum', 'sq': 'tetumisht', 'sr': 'тетум', 'sr-Cyrl': 'тетум', 'sr-Latn': 'tetum', 'sv': 'tetum', 'sw': 'Kitetum', 'ta': 'டெடும்', 'te': 'టేటం', 'th': 'เตตุม', 'ti': 'ቲተም', 'tk': 'tetum dili', 'to': 'lea fakatetumu', 'tr': 'Tetum', 'ug': 'تېتۇمچە', 'uk': 'тетум', 'ur': 'ٹیٹم', 'uz': 'tetum', 'uz-Latn': 'tetum', 'vec': 'tétum', 'vi': 'Tiếng Tetum', 'wae': 'Tetum', 'xnr': 'तेतुम', 'yo': 'Èdè Tetum', 'yrl': 'tetũ', 'yue': '泰頓文', 'yue-Hans': '泰顿文', 'yue-Hant': '泰頓文', 'zh': '德顿语', 'zh-Hans': '德顿语', 'zh-Hant': '泰頓文', 'zu': 'isi-Tetum'},
    'teu': {'en': 'Soo'},
    'tev': {'en': 'Teor'},
    'tew': {'en': 'Tewa (USA)'},
    'tex': {'en': 'Tennet'},
    'tey': {'en': 'Tulishi'},
    'tez': {'en': 'Tetserret'},
    'tfi': {'en': 'Tofin Gbe'},
    'tfn': {'en': 'Tanaina'},
    'tfo': {'en': 'Tefaro'},
    'tfr': {'en': 'Teribe'},
    'tft': {'en': 'Ternate'},
    'tg': {'af': 'Tadjiks', 'ak': 'Tɛgyeke kasa', 'am': 'ታጂክ', 'ar': 'الطاجيكية', 'as': 'তাজিক', 'ast': 'taxiquistanín', 'az': 'tacik', 'az-Arab': 'تاجیکجه', 'az-Cyrl': 'таҹик', 'az-Latn': 'tacik', 'bal-Latn': 'Tájek', 'be': 'таджыкская', 'be-tarask': 'таджыкская', 'bg': 'таджикски', 'bgn': 'تاجیکی', 'bn': 'তাজিক', 'br': 'tadjik', 'brx': 'ताजिक', 'bs': 'tadžički', 'bs-Cyrl': 'таџички', 'bs-Latn': 'tadžički', 'ca': 'tadjik', 'ca-ES-valencia': 'tadjik', 'ccp': '𑄖𑄎𑄨𑄇𑄴', 'ce': 'таджикийн', 'chr': 'ᏔᏥᎩ', 'ckb': 'تاجیکی', 'cs': 'tádžičtina', 'cy': 'Tajiceg', 'da': 'tadsjikisk', 'de': 'Tadschikisch', 'dsb': 'tadžikišćina', 'dz': 'ཏ་ཇིཀ་ཁ', 'ee': 'tadzikistangbe', 'el': 'Τατζικικά', 'el-polyton': 'Τατζικικά', 'en': 'Tajik', 'eo': 'taĝika', 'es': 'tayiko', 'et': 'tadžiki', 'eu': 'tajikera', 'fa': 'تاجیکی', 'ff-Adlm': '𞤚𞤢𞤶𞤭𞤳𞤪𞤫', 'fi': 'tadžikki', 'fil': 'Tajik', 'fo': 'tajik', 'fr': 'tadjik', 'fur': 'tagic', 'fy': 'Tadzjieks', 'ga': 'Taidsícis', 'gd': 'Taidigis', 'gl': 'taxico', 'gsw': 'Tadschikisch', 'gu': 'તાજીક', 'ha': 'Tajik', 'he': 'טג׳יקית', 'hi': 'ताजिक', 'hi-Latn': 'Tajik', 'hr': 'tadžički', 'hsb': 'tadźikšćina', 'ht': 'tadjik', 'hu': 'tadzsik', 'hy': 'տաջիկերեն', 'ia': 'tajiko', 'id': 'Tajik', 'ig': 'Tajik', 'is': 'tadsjikska', 'it': 'tagico', 'ja': 'タジク語', 'jv': 'Tajik', 'ka': 'ტაჯიკური', 'kea': 'tadjiki', 'kgp': 'tanhike', 'kk': 'тәжік тілі', 'kk-Cyrl': 'тәжік тілі', 'km': 'តាហ្ស៊ីគ', 'kn': 'ತಾಜಿಕ್', 'ko': '타지크어', 'kok': 'ताजिक', 'kok-Deva': 'ताजिक', 'ks': 'تاجِک', 'ks-Arab': 'تاجِک', 'ksh': 'Tadschihkesch', 'ku': 'tacikî', 'ky': 'тажикче', 'lb': 'Tadschikesch', 'lkt': 'Tajik Iyápi', 'lld': 'tajich', 'lo': 'ທາຈິກ', 'lrc': 'تاجیکی', 'lt': 'tadžikų', 'lv': 'tadžiku', 'mai': 'ताजिक', 'mi': 'Tāhiki', 'mk': 'таџикистански', 'ml': 'താജിക്', 'mn': 'тажик', 'mr': 'ताजिक', 'ms': 'Tajik', 'mt': 'Taġik', 'my': 'တာဂျစ်', 'mzn': 'تاجیکی', 'nb': 'tadsjikisk', 'ne': 'ताजिक', 'nl': 'Tadzjieks', 'nn': 'tadsjikisk', 'no': 'tadsjikisk', 'nqo': 'ߕߊߖߞߌߞߊ߲', 'or': 'ତାଜିକ୍', 'pa': 'ਤਾਜਿਕ', 'pa-Guru': 'ਤਾਜਿਕ', 'pcm': 'Tájik', 'pl': 'tadżycki', 'ps': 'تاجکي', 'pt': 'tadjique', 'qu': 'Tayiko Simi', 'rm': 'tadjik', 'ro': 'tadjică', 'ru': 'таджикский', 'sah': 'Тадьыыктыы', 'sc': 'tagicu', 'sd': 'تاجڪ', 'sd-Arab': 'تاجڪ', 'si': 'ටජික්', 'sk': 'tadžičtina', 'sl': 'tadžiščina', 'smn': 'tadžikkielâ', 'so': 'Taajik', 'sq': 'taxhikisht', 'sr': 'таџички', 'sr-Cyrl': 'таџички', 'sr-Latn': 'tadžički', 'sv': 'tadzjikiska', 'sw': 'Kitajiki', 'ta': 'தஜிக்', 'te': 'తజిక్', 'tg': 'тоҷикӣ', 'th': 'ทาจิก', 'ti': 'ታጂክኛ', 'tk': 'täjik dili', 'to': 'lea fakatāsiki', 'tr': 'Tacikçe', 'tt': 'таҗик', 'ug': 'تاجىكچە', 'uk': 'таджицька', 'ur': 'تاجک', 'uz': 'tojik', 'uz-Cyrl': 'тожикча', 'uz-Latn': 'tojik', 'vec': 'tàjego', 'vi': 'Tiếng Tajik', 'wae': 'Tadšikiš', 'wo': 'Tajis', 'xnr': 'ताजिक', 'yo': 'Èdè Tàjíìkì', 'yrl': 'tadiyike', 'yue': '塔吉克文', 'yue-Hans': '塔吉克文', 'yue-Hant': '塔吉克文', 'zh': '塔吉克语', 'zh-Hans': '塔吉克语', 'zh-Hant': '塔吉克文', 'zu': 'isi-Tajik'},
    'tga': {'en': 'Sagalla'},
    'tgb': {'en': 'Tobilung'},
    'tgc': {'en': 'Tigak'},
    'tgd': {'en': 'Ciwogai'},
    'tge': {'en': 'Eastern Gorkha Tamang'},
    'tgf': {'en': 'Chalikha'},
    'tgg': {'en': 'Tangga'},
    'tgh': {'en': 'Tobagonian Creole English'},
    'tgi': {'en': 'Lawunuia'},
    'tgj': {'en': 'Tagin'},
    'tgn': {'en': 'Tandaganon'},
    'tgo': {'en': 'Sudest'},
    'tgp': {'en': 'Tangoa'},
    'tgq': {'en': 'Tring'},
    'tgr': {'en': 'Tareng'},
    'tgs': {'en': 'Nume'},
    'tgt': {'en': 'Central Tagbanwa'},
    'tgu': {'en': 'Tanggu'},
    'tgv': {'en': 'Tingui-Boto'},
    'tgw': {'en': 'Tagwana Senoufo'},
    'tgx': {'af': 'Tagish', 'am': 'ታጊሽ', 'ar': 'التاغيش', 'as': 'টেগিশ্ব', 'az': 'taq', 'az-Latn': 'taq', 'be': 'тагіш', 'be-tarask': 'тагіш', 'bg': 'тагиш', 'bn': 'তাগিশ', 'brx': 'तागीश', 'bs': 'tagiš', 'bs-Cyrl': 'тагиш', 'bs-Latn': 'tagiš', 'ca': 'tagish', 'ca-ES-valencia': 'tagish', 'chr': 'ᏔᎩᏏ', 'cs': 'tagiš', 'cy': 'Tagish', 'da': 'tagish', 'de': 'Tagish', 'dsb': 'tagishšćina', 'el': 'Τατζίς', 'el-polyton': 'Τατζίς', 'en': 'Tagish', 'es': 'tagish', 'et': 'tagishi', 'eu': 'tagishera', 'fa': 'تاگیش', 'ff-Adlm': '𞤚𞤢𞤺𞤭𞥃𞤪𞤫', 'fi': 'tagish', 'fil': 'Tagish', 'fr': 'tagish', 'ga': 'Tagish', 'gd': 'Tagish', 'gl': 'tagish', 'gu': 'ટાગિશ', 'ha': 'Tagish', 'he': 'טגישית', 'hi': 'टैगिश', 'hi-Latn': 'Tagish', 'hr': 'tagish', 'hsb': 'tagishšćina', 'ht': 'tagish', 'hu': 'tagish', 'hy': 'թագիշ', 'ia': 'tagish', 'id': 'Tagish', 'ig': 'Tagish', 'is': 'tagíska', 'it': 'tagish', 'ja': 'タギシュ語', 'jv': 'Tagish', 'ka': 'ტაგიში', 'kk': 'тагиш тілі', 'kk-Cyrl': 'тагиш тілі', 'km': 'តាហ្គីស', 'kn': 'ಟಾಗಿಶ್', 'ko': '타기시어', 'kok': 'टॅगिश', 'kok-Deva': 'टॅगिश', 'ku': 'tagîşî', 'ky': 'тагишче (индей тили)', 'lo': 'ທາກີຊ', 'lt': 'tagišų', 'lv': 'tagišu', 'mai': 'टेगिश', 'mi': 'Tākihi', 'mk': 'тагишки', 'ml': 'ടാഗിഷ്', 'mn': 'тагиш', 'mr': 'टॅगिश', 'ms': 'Tagish', 'my': 'တာဂစ်ရှ်', 'nb': 'tagish', 'ne': 'टागिस', 'nl': 'Tagish', 'nn': 'tagish', 'no': 'tagish', 'or': 'ତାଗିଶ', 'pa': 'ਟੈਗਿਸ਼', 'pa-Guru': 'ਟੈਗਿਸ਼', 'pcm': 'Tágísh Lángwej', 'pl': 'tagish', 'ps': 'ټګش', 'pt': 'tagish', 'qu': 'Tagish Simi', 'ro': 'tagish', 'ru': 'тагиш', 'sc': 'tagish', 'sd': 'ٽئگِش', 'sd-Arab': 'ٽئگِش', 'si': 'ටැගිෂ්', 'sk': 'tagiš', 'sl': 'tagiščina', 'so': 'Tagish', 'sq': 'tagishisht', 'sr': 'тагиш', 'sr-Cyrl': 'тагиш', 'sr-Latn': 'tagiš', 'sv': 'tagish', 'sw': 'Kitagishi', 'ta': 'தகிஷ்', 'te': 'టాగీష్', 'th': 'ทากิช', 'ti': 'ታጊሽ', 'tk': 'tagiş dili', 'to': 'lea fakatākisi', 'tr': 'Tagişçe', 'uk': 'тагіш', 'ur': 'ٹاگش', 'uz': 'tagish', 'uz-Latn': 'tagish', 'vec': 'tagish', 'vi': 'Tiếng Tagish', 'yo': 'Èdè Tagisi', 'yue': '塔吉什文', 'yue-Hans': '塔吉什文', 'yue-Hant': '塔吉什文', 'zh': '塔吉什语', 'zh-Hans': '塔吉什语', 'zh-Hant': '塔吉什語', 'zu': 'isi-Tagish'},
    'tgy': {'en': 'Togoyo'},
    'tgz': {'en': 'Tagalaka'},
    'th': {'af': 'Thai', 'agq': 'Tàe', 'ak': 'Taeland kasa', 'am': 'ታይ', 'ar': 'التايلاندية', 'as': 'থাই', 'asa': 'Kitailandi', 'ast': 'tailandés', 'az': 'tay', 'az-Cyrl': 'тај', 'az-Latn': 'tay', 'bal-Latn': 'Tái', 'bas': 'Hɔp u tây', 'be': 'тайская', 'be-tarask': 'тайская', 'bem': 'Ichi Thai', 'bez': 'Hitailand', 'bg': 'тайски', 'bgn': 'تایلندی', 'blo': 'taɩ kagɩja', 'bm': 'tayikan', 'bn': 'থাই', 'br': 'thai', 'brx': 'थाई', 'bs': 'tajlandski', 'bs-Cyrl': 'тајландски', 'bs-Latn': 'tajlandski', 'ca': 'tai', 'ca-ES-valencia': 'tai', 'ccp': '𑄗𑄭', 'ce': 'тайн', 'ceb': 'Thai', 'cgg': 'Orutailandi', 'chr': 'ᏔᏱ', 'ckb': 'تایلەندی', 'cs': 'thajština', 'cv': 'тай', 'cy': 'Thai', 'da': 'thai', 'dav': 'Kitailandi', 'de': 'Thailändisch', 'dje': 'Taailandu senni', 'dsb': 'thailandšćina', 'dyo': 'tay', 'dz': 'ཐཱའི་ཁ', 'ebu': 'Kĩtailandi', 'ee': 'Thailandgbe', 'el': 'Ταϊλανδικά', 'el-polyton': 'Ταϊλανδικά', 'en': 'Thai', 'eo': 'taja', 'es': 'tailandés', 'et': 'tai', 'eu': 'thailandiera', 'ewo': 'ǹkɔ́bɔ táilan', 'fa': 'تایلندی', 'ff': 'Taay', 'ff-Adlm': '𞤚𞤢𞤴𞤤𞤢𞤲𞤣𞤫𞥅𞤪𞤫', 'ff-Latn': 'Taay', 'fi': 'thai', 'fil': 'Thai', 'fo': 'tailendskt', 'fr': 'thaï', 'fur': 'thai', 'fy': 'Thais', 'ga': 'Téalainnis', 'gd': 'Cànan nan Tàidh', 'gl': 'tailandés', 'gsw': 'Thailändisch', 'gu': 'થાઈ', 'guz': 'Kitailandi', 'ha': 'Thai', 'he': 'תאית', 'hi': 'थाई', 'hi-Latn': 'Thai', 'hr': 'tajlandski', 'hsb': 'thailandšćina', 'ht': 'thaï', 'hu': 'thai', 'hy': 'թայերեն', 'ia': 'thai', 'id': 'Thai', 'ig': 'Thai', 'is': 'taílenska', 'it': 'thailandese', 'ja': 'タイ語', 'jmc': 'Kyitailandi', 'jv': 'Thailand', 'ka': 'ტაილანდური', 'kab': 'Taṭaylundit', 'kam': 'Kitailandi', 'kde': 'Chitailandi', 'kea': 'tailandes', 'kgp': 'tajrỹnej', 'khq': 'Taailandu senni', 'ki': 'Kitailandi', 'kk': 'тай тілі', 'kk-Cyrl': 'тай тілі', 'kln': 'kutitab Thailand', 'km': 'ថៃ', 'kn': 'ಥಾಯ್', 'ko': '태국어', 'kok': 'थाई', 'kok-Deva': 'थाई', 'ks': 'تھاے', 'ks-Arab': 'تھاے', 'ksb': 'Kitailandi', 'ksf': 'ritaí', 'ksh': 'Tailändesch', 'ku': 'tayî', 'kxv': 'tae', 'kxv-Deva': 'ताई', 'kxv-Latn': 'tae', 'kxv-Orya': 'ତାଇ', 'kxv-Telu': 'తాఇ', 'ky': 'тайча', 'lag': 'Kɨtáilandi', 'lb': 'Thailännesch', 'lg': 'Luttaayi', 'lkt': 'Thai Iyápi', 'lld': 'thailandesc', 'ln': 'litaye', 'lo': 'ໄທ', 'lrc': 'تایلأندی', 'lt': 'tajų', 'lu': 'Ntailandi', 'luo': 'Kitailandi', 'luy': 'Kitailandi', 'lv': 'taju', 'mai': 'थाई', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ ltai', 'mer': 'Kĩthailandi', 'mfe': 'thaï', 'mg': 'Taioaney', 'mgh': 'Itailandi', 'mi': 'Tai', 'mk': 'тајландски', 'ml': 'തായ്', 'mn': 'тай', 'mni': 'থাই', 'mni-Beng': 'থাই', 'mr': 'थाई', 'ms': 'Thai', 'mt': 'Tajlandiż', 'my': 'ထိုင်း', 'mzn': 'تایی', 'naq': 'Thaiǁî gowab', 'nb': 'thai', 'nd': 'isi-Thayi', 'ne': 'थाई', 'nl': 'Thai', 'nmg': 'Kiɛl thaï', 'nn': 'thai', 'no': 'thai', 'nqo': 'ߕߊߦߌߞߊ߲', 'nus': 'Thok tayni', 'nyn': 'Orutailandi', 'om': 'Afaan Tayii', 'or': 'ଥାଇ', 'pa': 'ਥਾਈ', 'pa-Guru': 'ਥਾਈ', 'pcm': 'Taí', 'pl': 'tajski', 'ps': 'تايلېنډي', 'pt': 'tailandês', 'qu': 'Tailandes Simi', 'rm': 'tailandais', 'rn': 'Ikinyatayilandi', 'ro': 'thailandeză', 'rof': 'Kitailandi', 'ru': 'тайский', 'rw': 'Igitayi', 'rwk': 'Kyitailandi', 'saq': 'Kitailandi', 'sbp': 'Ishitayilandi', 'sc': 'tailandesu', 'sd': 'ٿائي', 'sd-Arab': 'ٿائي', 'se': 'ŧaigiella', 'seh': 'tailandês', 'ses': 'Taailandu senni', 'sg': 'Thâi', 'shi': 'ⵜⴰⵜⴰⵢⵍⴰⵏⴷⵉⵜ', 'shi-Latn': 'Tataylandit', 'shi-Tfng': 'ⵜⴰⵜⴰⵢⵍⴰⵏⴷⵉⵜ', 'si': 'තායි', 'sk': 'thajčina', 'sl': 'tajščina', 'smn': 'thaikielâ', 'sn': 'chiThai', 'so': 'Taaylandays', 'sq': 'tajlandisht', 'sr': 'тајски', 'sr-Cyrl': 'тајски', 'sr-Latn': 'tajski', 'sv': 'thailändska', 'sw': 'Kithai', 'ta': 'தாய்', 'te': 'థాయ్', 'teo': 'Kitailandi', 'tg': 'тайӣ', 'th': 'ไทย', 'ti': 'ታይኛ', 'tk': 'taý dili', 'to': 'lea fakatailani', 'tr': 'Tayca', 'tt': 'тай', 'twq': 'Taailandu senni', 'tzm': 'Taṭayt', 'ug': 'تايلاندچە', 'uk': 'тайська', 'ur': 'تھائی', 'uz': 'tay', 'uz-Cyrl': 'тайча', 'uz-Latn': 'tay', 'vai': 'ꕚꔤ', 'vai-Latn': 'Tái', 'vai-Vaii': 'ꕚꔤ', 'vec': 'tailandeze', 'vi': 'Tiếng Thái', 'vun': 'Kyitailandi', 'wae': 'Thailändiš', 'wo': 'Tay', 'xh': 'Isi-Thai', 'xnr': 'थाई', 'xog': 'Oluttaayi', 'yav': 'nutáyɛ', 'yo': 'Èdè Tai', 'yrl': 'tairãdei', 'yue': '泰文', 'yue-Hans': '泰文', 'yue-Hant': '泰文', 'zgh': 'ⵜⴰⵜⴰⵢⵍⴰⵏⴷⵉⵜ', 'zh': '泰语', 'zh-Hans': '泰语', 'zh-Hant': '泰文', 'zu': 'isi-Thai'},
    'thc': {'en': 'Tai Hang Tong'},
    'thd': {'en': 'Kuuk Thaayorre'},
    'the': {'en': 'Chitwania Tharu'},
    'thf': {'en': 'Thangmi'},
    'thh': {'en': 'Northern Tarahumara'},
    'thi': {'en': 'Tai Long'},
    'thk': {'en': 'Tharaka'},
    'thl': {'en': 'Dangaura Tharu'},
    'thm': {'en': 'Aheu'},
    'thn': {'en': 'Thachanadan'},
    'thp': {'en': 'Thompson'},
    'thq': {'en': 'Kochila Tharu'},
    'thr': {'en': 'Rana Tharu'},
    'ths': {'en': 'Thakali'},
    'tht': {'af': 'Tahltan', 'am': 'ታህልታን', 'ar': 'التالتان', 'as': 'টাহলটান', 'az': 'taltan', 'az-Latn': 'taltan', 'be': 'тальтан', 'be-tarask': 'тальтан', 'bg': 'талтан', 'bn': 'তাহ্লতান', 'brx': 'टाहल्टान', 'bs': 'tahltanski', 'bs-Cyrl': 'талтан', 'bs-Latn': 'tahltanski', 'ca': 'tahltà', 'ca-ES-valencia': 'tahltà', 'chr': 'ᏔᏝᎾ', 'cs': 'tahltan', 'cy': 'Tahltan', 'da': 'tahltan', 'de': 'Tahltan', 'dsb': 'tahltanšćina', 'el': 'Ταλτάν', 'el-polyton': 'Ταλτάν', 'en': 'Tahltan', 'es': 'tahltan', 'et': 'tahltani', 'eu': 'tahltanera', 'fa': 'تالتان', 'ff-Adlm': '𞤚𞤢𞤸𞤢𞤤𞤼𞤢𞤲𞤪𞤫', 'fi': 'tahlta', 'fil': 'Tahltan', 'fr': 'tahltan', 'ga': 'Tahltan', 'gd': 'Tahltan', 'gl': 'tahltan', 'gu': 'તહલતાન', 'ha': 'Tahltan', 'he': 'טלתנית', 'hi': 'टैल्हटन', 'hi-Latn': 'Tahltan', 'hr': 'tahltan', 'hsb': 'tahltanšćina', 'ht': 'tahltan', 'hu': 'tahltan', 'hy': 'թալթան', 'ia': 'tahltan', 'id': 'Tahltan', 'ig': 'Tahitan', 'is': 'tahltan', 'it': 'tahltan', 'ja': 'タールタン語', 'jv': 'Tahltan', 'ka': 'ტალტანი', 'kk': 'тальтан тілі', 'kk-Cyrl': 'тальтан тілі', 'km': 'តាល់តាន', 'kn': 'ಟಾಹ್ಲ್ಟನ್', 'ko': '탈탄어', 'kok': 'ताहल्तान', 'kok-Deva': 'ताहल्तान', 'ku': 'tahltan', 'ky': 'талтанча (индей тили)', 'lo': 'ທາວທັນ', 'lt': 'taltanų', 'lv': 'tahltanu', 'mai': 'टालटन', 'mi': 'Tātana', 'mk': 'талтански', 'ml': 'ടാഹ്\u200cൽടൻ', 'mn': 'талтан', 'mr': 'तहल्टन', 'ms': 'Tahltan', 'my': 'တဟီတန်', 'nb': 'tahltan', 'ne': 'टाहल्टन', 'nl': 'Tahltan', 'nn': 'tahltan', 'no': 'tahltan', 'or': 'ତହଲତାନ୍', 'pa': 'ਤਹਿਲਟਨ', 'pa-Guru': 'ਤਹਿਲਟਨ', 'pcm': 'Tahltán Lángwej', 'pl': 'tahltan', 'ps': 'طهلتان', 'pt': 'tahltan', 'qu': 'Tahltan Simi', 'ro': 'tahltan', 'ru': 'талтан', 'sc': 'tahltan', 'sd': 'ٽهلٽن', 'sd-Arab': 'ٽهلٽن', 'si': 'ටැල්ටන්', 'sk': 'tahltan', 'sl': 'taltanščina', 'so': 'Tahltan', 'sq': 'taltanisht', 'sr': 'тахлтан', 'sr-Cyrl': 'тахлтан', 'sr-Latn': 'tahltan', 'sv': 'tahltan', 'sw': 'Kitahltani', 'ta': 'தால்டன்', 'te': 'ట్యాల్టాన్', 'th': 'ทาลทาน', 'ti': 'ታልተን', 'tk': 'taltan dili', 'to': 'lea fakatālitāni', 'tr': 'Tahltanca', 'uk': 'талтан', 'ur': 'ٹاہلٹن', 'uz': 'taltan', 'uz-Latn': 'taltan', 'vec': 'taltan', 'vi': 'Tiếng Tahltan', 'yo': 'Èdè Tajiti', 'yue': '塔爾坦文', 'yue-Hans': '塔尔坦文', 'yue-Hant': '塔爾坦文', 'zh': '塔尔坦语', 'zh-Hans': '塔尔坦语', 'zh-Hant': '塔爾坦語', 'zu': 'Tahltan'},
    'thu': {'en': 'Thuri'},
    'thv': {'en': 'Tahaggart Tamahaq'},
    'thw': {'en': 'Thudam'},
    'thx': {'en': 'The'},
    'thy': {'en': 'Tha'},
    'thz': {'en': 'Tayart Tamajeq'},
    'ti': {'af': 'Tigrinja', 'ak': 'Tigrinya kasa', 'am': 'ትግርኛ', 'ar': 'التغرينية', 'as': 'টিগৰিনিয়া', 'ast': 'tigrinya', 'az': 'tiqrin', 'az-Cyrl': 'тигрин', 'az-Latn': 'tiqrin', 'bal-Latn': 'Tigriniá', 'be': 'тыгрынья', 'be-tarask': 'тыгрынья', 'bg': 'тигриня', 'bgn': 'ٹیگرینیایی', 'bn': 'তিগরিনিয়া', 'br': 'tigrigna', 'brx': 'तिग्रीन्या', 'bs': 'tigrinja', 'bs-Cyrl': 'тигриња', 'bs-Latn': 'tigrinja', 'ca': 'tigrinya', 'ca-ES-valencia': 'tigrinya', 'ccp': '𑄖𑄨𑄉𑄧𑄢𑄨𑄚𑄨𑄠', 'ce': 'тигринья', 'chr': 'ᏘᎩᎵᏂᎠ', 'ckb': 'تیگرینیا', 'cs': 'tigrinijština', 'cy': 'Tigrinya', 'da': 'tigrinya', 'de': 'Tigrinya', 'dsb': 'tigrinja', 'dz': 'ཏིག་རི་ཉ་ཁ', 'ee': 'tigrinyagbe', 'el': 'Τιγκρινικά', 'el-polyton': 'Τιγκρινικά', 'en': 'Tigrinya', 'eo': 'tigraja', 'es': 'tigriña', 'et': 'tigrinja', 'eu': 'tigrinyera', 'fa': 'تیگرینیایی', 'ff-Adlm': '𞤚𞤭𞤺𞤭𞤪𞤻𞤢𞥄𞤪𞤫', 'fi': 'tigrinja', 'fil': 'Tigrinya', 'fo': 'tigrinya', 'fr': 'tigrigna', 'fy': 'Tigrinya', 'ga': 'Tigrinis', 'gd': 'Tigrinya', 'gl': 'tigriña', 'gsw': 'Tigrinja', 'gu': 'ટાઇગ્રિનિયા', 'ha': 'Tigrinyanci', 'he': 'תיגרינית', 'hi': 'तिग्रीन्या', 'hi-Latn': 'Tigrinya', 'hr': 'tigrinja', 'hsb': 'tigrinšćina', 'ht': 'tigrigna', 'hu': 'tigrinya', 'hy': 'տիգրինյա', 'ia': 'tigrinya', 'id': 'Tigrinya', 'ig': 'Tigrinya', 'is': 'tígrinja', 'it': 'tigrino', 'ja': 'ティグリニア語', 'jv': 'Tigrinya', 'ka': 'ტიგრინია', 'kea': 'tigrinia', 'kgp': 'tigrinĩja', 'kk': 'тигринья тілі', 'kk-Cyrl': 'тигринья тілі', 'km': 'ទីហ្គ្រីញ៉ា', 'kn': 'ಟಿಗ್ರಿನ್ಯಾ', 'ko': '티그리냐어', 'kok': 'तिग्रिन्या', 'kok-Deva': 'तिग्रिन्या', 'ks': 'ٹِگرِنیا', 'ks-Arab': 'ٹِگرِنیا', 'ksh': 'Tijrenejahnesch', 'ku': 'tigrînî', 'ky': 'тигриниача', 'lb': 'Tigrinja', 'lkt': 'Tigrinya Iyápi', 'lld': 'tigrin', 'lo': 'ຕິກຣິນຢາ', 'lrc': 'تیگرینیا', 'lt': 'tigrajų', 'lv': 'tigrinja', 'mai': 'टिगरिनया', 'mi': 'Tekirinia', 'mk': 'тигриња', 'ml': 'ടൈഗ്രിന്യ', 'mn': 'тигринья', 'mr': 'तिग्रिन्या', 'ms': 'Tigrinya', 'mt': 'Tigrinya', 'my': 'တီဂ်ရင်ယာ', 'mzn': 'تیگرینیایی', 'nb': 'tigrinja', 'ne': 'टिग्रिन्या', 'nl': 'Tigrinya', 'nn': 'tigrinja', 'no': 'tigrinja', 'nqo': 'ߕߜ߭ߌߙߌߢߊߞߊ߲', 'om': 'Afaan Tigiree', 'or': 'ଟାଇଗ୍ରିନିଆ', 'pa': 'ਤਿਗ੍ਰੀਨਿਆ', 'pa-Guru': 'ਤਿਗ੍ਰੀਨਿਆ', 'pcm': 'Tigrínyá', 'pl': 'tigrinia', 'ps': 'تيګريني', 'pt': 'tigrínia', 'qu': 'Tigriña Simi', 'rm': 'tigrinya', 'ro': 'tigrină', 'ru': 'тигринья', 'rw': 'Inyatigirinya', 'sc': 'tigrignu', 'sd': 'تگرينيائي', 'sd-Arab': 'تگرينيائي', 'si': 'ටිග්\u200dරින්යා', 'sk': 'tigriňa', 'sl': 'tigrajščina', 'smn': 'tigrinyakielâ', 'so': 'Tigrinya', 'sq': 'tigrinjaisht', 'sr': 'тигриња', 'sr-Cyrl': 'тигриња', 'sr-Latn': 'tigrinja', 'sv': 'tigrinja', 'sw': 'Kitigrinya', 'ta': 'டிக்ரின்யா', 'te': 'టిగ్రిన్యా', 'tg': 'тигриня', 'th': 'ติกริญญา', 'ti': 'ትግርኛ', 'tk': 'tigrinýa dili', 'to': 'lea fakatikilinia', 'tr': 'Tigrinya dili', 'tt': 'тигринья', 'ug': 'تىگرىنياچە', 'uk': 'тигринья', 'ur': 'ٹگرینیا', 'uz': 'tigrinya', 'uz-Cyrl': 'тигриняча', 'uz-Latn': 'tigrinya', 'vec': 'tigrigna', 'vi': 'Tiếng Tigrinya', 'wae': 'Tigrinja', 'wo': 'Tigriña', 'xnr': 'तिग्रीन्या', 'yo': 'Èdè Tigrinya', 'yrl': 'tigirínia', 'yue': '提格利尼亞文', 'yue-Hans': '提格利尼亚文', 'yue-Hant': '提格利尼亞文', 'zh': '提格利尼亚语', 'zh-Hans': '提格利尼亚语', 'zh-Hant': '提格利尼亞文', 'zu': 'isi-Tigrinya'},
    'tia': {'en': 'Tidikelt Tamazight'},
    'tic': {'en': 'Tira'},
    'tid': {'en': 'Tidong'},
    'tie': {'en': 'Tingal'},
    'tif': {'en': 'Tifal'},
    'tig': {'af': 'Tigre', 'am': 'ትግረ', 'ar': 'التيغرية', 'as': 'তাইগ্ৰে', 'ast': 'tigre', 'az': 'tiqre', 'az-Cyrl': 'тигре', 'az-Latn': 'tiqre', 'bal-Latn': 'Tigré', 'be': 'тыгрэ', 'be-tarask': 'тыгрэ', 'bg': 'тигре', 'bgn': 'ٹایگری', 'bn': 'টাইগ্রে', 'br': 'tigreaneg', 'brx': 'टीग्रे', 'bs': 'tigre', 'bs-Cyrl': 'тигре', 'bs-Latn': 'tigre', 'ca': 'tigre', 'ca-ES-valencia': 'tigre', 'ccp': '𑄑𑄭𑄉𑄳𑄢𑄬', 'ce': 'тигре', 'chr': 'ᏢᏓᏥ', 'ckb': 'تیگرێ', 'cs': 'tigrejština', 'cy': 'Tigreg', 'da': 'tigre', 'de': 'Tigre', 'dsb': 'tigrešćina', 'el': 'Τίγκρε', 'el-polyton': 'Τίγκρε', 'en': 'Tigre', 'es': 'tigré', 'et': 'tigree', 'eu': 'tigreera', 'fa': 'تیگره\u200cای', 'ff-Adlm': '𞤚𞤭𞤺𞤭𞤪𞤴𞤢𞤲𞤪𞤫', 'fi': 'tigre', 'fil': 'Tigre', 'fo': 'tigre', 'fr': 'tigré', 'fy': 'Tigre', 'ga': 'Tigréis', 'gd': 'Tigre', 'gl': 'tigré', 'gsw': 'Tigre', 'gu': 'ટાઇગ્રે', 'ha': 'Tigre', 'he': 'טיגרית', 'hi': 'टाइग्रे', 'hi-Latn': 'Tigre', 'hr': 'tigriški', 'hsb': 'tigrešćina', 'ht': 'tigré', 'hu': 'tigré', 'hy': 'տիգրե', 'ia': 'tigre', 'id': 'Tigre', 'ig': 'Tigre', 'is': 'tígre', 'it': 'tigre', 'ja': 'ティグレ語', 'jv': 'Tigre', 'ka': 'თიგრე', 'kgp': 'tigré', 'kk': 'тигре тілі', 'kk-Cyrl': 'тигре тілі', 'km': 'ធីហ្គ្រា', 'kn': 'ಟೈಗ್ರೆ', 'ko': '티그레어', 'kok': 'टिग्रे', 'kok-Deva': 'टिग्रे', 'ks': 'ٹاےگریے', 'ks-Arab': 'ٹاےگریے', 'ksh': 'Tigre', 'ku': 'tigre', 'ky': 'тигрече', 'lb': 'Tigre', 'lo': 'ໄທກຣີ', 'lt': 'tigre', 'lv': 'tigru', 'mai': 'टिगरे', 'mi': 'Tīkara', 'mk': 'тигре', 'ml': 'ടൈഗ്രി', 'mn': 'тигр', 'mr': 'टाइग्रे', 'ms': 'Tigre', 'mt': 'Tigre', 'my': 'တီဂရီ', 'nb': 'tigré', 'ne': 'टिग्रे', 'nl': 'Tigre', 'nn': 'tigré', 'no': 'tigré', 'nqo': 'ߕߌߜ߭ߙߋߞߊ߲', 'or': 'ଟାଇଗ୍ରେ', 'pa': 'ਟਿਗਰਾ', 'pa-Guru': 'ਟਿਗਰਾ', 'pcm': 'Tígrẹ Lángwej', 'pl': 'tigre', 'ps': 'تیګر', 'pt': 'tigré', 'qu': 'Tigre Simi', 'rm': 'tigre', 'ro': 'tigre', 'ru': 'тигре', 'sc': 'tigrè', 'sd': 'تگري', 'sd-Arab': 'تگري', 'si': 'ටීග්\u200dරෙ', 'sk': 'tigrejčina', 'sl': 'tigrejščina', 'smn': 'tigrekielâ', 'so': 'Tigre', 'sq': 'tigreisht', 'sr': 'тигре', 'sr-Cyrl': 'тигре', 'sr-Latn': 'tigre', 'sv': 'tigré', 'sw': 'Kitigre', 'ta': 'டைக்ரே', 'te': 'టీగ్రె', 'th': 'ตีเกร', 'ti': 'ትግረ', 'tk': 'tigre dili', 'to': 'lea fakatikilē', 'tr': 'Tigre', 'ug': 'تىگرېچە', 'uk': 'тигре', 'ur': 'ٹگرے', 'uz': 'tigre', 'uz-Cyrl': 'тигре', 'uz-Latn': 'tigre', 'vec': 'tigré', 'vi': 'Tiếng Tigre', 'xnr': 'टाइग्रे', 'yi': 'טיגרע', 'yo': 'Èdè Tigre', 'yrl': 'tigiré', 'yue': '蒂格雷文', 'yue-Hans': '蒂格雷文', 'yue-Hant': '蒂格雷文', 'zh': '提格雷语', 'zh-Hans': '提格雷语', 'zh-Hant': '蒂格雷文', 'zu': 'isi-Tigre'},
    'tih': {'en': 'Timugon Murut'},
    'tii': {'en': 'Tiene'},
    'tij': {'en': 'Tilung'},
    'tik': {'en': 'Tikar'},
    'til': {'en': 'Tillamook'},
    'tim': {'en': 'Timbe'},
    'tin': {'en': 'Tindi'},
    'tio': {'en': 'Teop'},
    'tip': {'en': 'Trimuris'},
    'tiq': {'en': 'Tiéfo'},
    'tis': {'en': 'Masadiit Itneg'},
    'tit': {'en': 'Tinigua'},
    'tiu': {'en': 'Adasen'},
    'tiv': {'ar': 'التيف', 'bg': 'тив', 'bn': 'টিভ', 'brx': 'टीव्', 'bs-Cyrl': 'тив', 'ccp': '𑄑𑄨𑄞𑄴', 'cs': 'tivština', 'cy': 'Tifeg', 'da': 'tivi', 'el': 'Τιβ', 'el-polyton': 'Τιβ', 'et': 'tivi', 'fa': 'تیوی', 'gsw': 'Tiv-Schpraach', 'gu': 'તિવ', 'he': 'טיב', 'hi': 'तिव', 'hy': 'տիվերեն', 'is': 'tív', 'ja': 'ティブ語', 'kgp': 'tivi', 'kn': 'ಟಿವ್', 'ko': '티브어', 'ks': 'تیٖو', 'ks-Arab': 'تیٖو', 'lb': 'Tiv-Sprooch', 'lo': 'ຕີວ', 'lv': 'tivu', 'mk': 'тив', 'ml': 'ടിവ്', 'mr': 'तिव', 'nn': 'tivi', 'or': 'ତୀଭ୍', 'pl': 'tiw', 'ru': 'тиви', 'sl': 'tivščina', 'sr': 'тив', 'sr-Cyrl': 'тив', 'sv': 'tivi', 'ta': 'டிவ்', 'te': 'టివ్', 'th': 'ทิฟ', 'to': 'lea fakativi', 'ug': 'تىۋچە', 'uk': 'тів', 'vi': 'Tiếng Tiv', 'yrl': 'tivi', 'yue': '提夫文', 'yue-Hans': '提夫文', 'yue-Hant': '提夫文', 'zh': '蒂夫语', 'zh-Hans': '蒂夫语', 'zh-Hant': '提夫文', 'en': 'Tiv'},
    'tiw': {'en': 'Tiwi'},
    'tix': {'en': 'Southern Tiwa'},
    'tiy': {'en': 'Tiruray'},
    'tiz': {'en': 'Tai Hongjin'},
    'tja': {'en': 'Tajuasohn'},
    'tjg': {'en': 'Tunjung'},
    'tji': {'en': 'Northern Tujia'},
    'tjj': {'en': 'Tjungundji'},
    'tjl': {'en': 'Tai Laing'},
    'tjm': {'en': 'Timucua'},
    'tjn': {'en': 'Tonjon'},
    'tjo': {'en': 'Temacine Tamazight'},
    'tjp': {'en': 'Tjupany'},
    'tjs': {'en': 'Southern Tujia'},
    'tju': {'en': 'Tjurruru'},
    'tjw': {'en': 'Djabwurrung'},
    'tk': {'af': 'Turkmeens', 'ak': 'Tɛkmɛnistan Kasa', 'am': 'ቱርክሜን', 'ar': 'التركمانية', 'as': 'তুৰ্কমেন', 'ast': 'turcomanu', 'az': 'türkmən', 'az-Arab': 'تۆرکمنجه', 'az-Cyrl': 'түркмән', 'az-Latn': 'türkmən', 'bal-Latn': 'Trkm', 'be': 'туркменская', 'be-tarask': 'туркменская', 'bg': 'туркменски', 'bgn': 'تورکمنی', 'bn': 'তুর্কমেনী', 'br': 'turkmeneg', 'brx': 'तुर्कमेन', 'bs': 'turkmenski', 'bs-Cyrl': 'туркменски', 'bs-Latn': 'turkmenski', 'ca': 'turcman', 'ca-ES-valencia': 'turcman', 'ccp': '𑄖𑄪𑄢𑄴𑄇𑄧𑄟𑄬𑄚𑄨', 'ce': 'туркменийн', 'chr': 'ᎠᏂᎬᎾ', 'ckb': 'تورکمانی', 'cs': 'turkmenština', 'cy': 'Tyrcmeneg', 'da': 'turkmensk', 'de': 'Turkmenisch', 'dsb': 'turkmeńšćina', 'dz': 'ཊཱརཀ་མེན་ཁ', 'ee': 'tɛkmengbe', 'el': 'Τουρκμενικά', 'el-polyton': 'Τουρκμενικά', 'en': 'Turkmen', 'eo': 'turkmena', 'es': 'turcomano', 'et': 'türkmeeni', 'eu': 'turkmenera', 'fa': 'ترکمنی', 'ff-Adlm': '𞤼𞤵𞤪𞤳𞤥𞤢𞤲𞤪𞤫', 'fi': 'turkmeeni', 'fil': 'Turkmen', 'fo': 'turkmenskt', 'fr': 'turkmène', 'fur': 'turcmen', 'fy': 'Turkmeens', 'ga': 'Tuircméinis', 'gd': 'Turcmanais', 'gl': 'turkmeno', 'gsw': 'Turkmenisch', 'gu': 'તુર્કમેન', 'ha': 'Tukmenistanci', 'he': 'טורקמנית', 'hi': 'तुर्कमेन', 'hi-Latn': 'Turkmen', 'hr': 'turkmenski', 'hsb': 'turkmenšćina', 'ht': 'turkmène', 'hu': 'türkmén', 'hy': 'թուրքմեներեն', 'ia': 'turkmeno', 'id': 'Turkmen', 'ig': 'Turkmen', 'is': 'túrkmenska', 'it': 'turcomanno', 'ja': 'トルクメン語', 'jv': 'Turkmen', 'ka': 'თურქმენული', 'kea': 'turkmenu', 'kgp': 'turkomẽnũ', 'kk': 'түрікмен тілі', 'kk-Cyrl': 'түрікмен тілі', 'km': 'តួកម៉េន', 'kn': 'ಟರ್ಕ್\u200cಮೆನ್', 'ko': '투르크멘어', 'kok': 'तुर्कमेन', 'kok-Deva': 'तुर्कमेन', 'ks': 'تُرکمین', 'ks-Arab': 'تُرکمین', 'ksh': 'Törkmehnesch', 'ku': 'tirkmenî', 'ky': 'түркмөнчө', 'lb': 'Turkmenesch', 'lkt': 'Turkmen Iyápi', 'lld': 'turcmenich', 'lo': 'ເທີກເມັນ', 'lrc': 'تورکأمأنی', 'lt': 'turkmėnų', 'lv': 'turkmēņu', 'mai': 'तुर्कमैन', 'mi': 'Tākamana', 'mk': 'туркменски', 'ml': 'തുർക്\u200cമെൻ', 'mn': 'туркмен', 'mr': 'तुर्कमेन', 'ms': 'Turkmen', 'mt': 'Turkmeni', 'my': 'တာ့ခ်မင်နစ္စတန်', 'mzn': 'ترکمونی', 'nb': 'turkmensk', 'ne': 'टर्कमेन', 'nl': 'Turkmeens', 'nn': 'turkmensk', 'no': 'turkmensk', 'nqo': 'ߕߎߙߞߌߡߍߣߌߞߊ߲', 'om': 'Lammii Turkii', 'or': 'ତୁର୍କମେନ୍', 'pa': 'ਤੁਰਕਮੇਨ', 'pa-Guru': 'ਤੁਰਕਮੇਨ', 'pcm': 'Tọ́kmẹn', 'pl': 'turkmeński', 'ps': 'ترکمني', 'pt': 'turcomeno', 'qu': 'Turcomano Simi', 'rm': 'turkmen', 'ro': 'turkmenă', 'ru': 'туркменский', 'rw': 'Inyeturukimeni', 'sc': 'turcmenu', 'sd': 'ترڪمين', 'sd-Arab': 'ترڪمين', 'si': 'ටර්ක්මෙන්', 'sk': 'turkménčina', 'sl': 'turkmenščina', 'smn': 'turkmenkielâ', 'so': 'Turkumaanish', 'sq': 'turkmenisht', 'sr': 'туркменски', 'sr-Cyrl': 'туркменски', 'sr-Latn': 'turkmenski', 'sv': 'turkmeniska', 'sw': 'Kiturukimeni', 'ta': 'துருக்மென்', 'te': 'తుర్క్\u200cమెన్', 'tg': 'туркманӣ', 'th': 'เติร์กเมน', 'ti': 'ቱርክመንኛ', 'tk': 'türkmen dili', 'to': 'lea fakatēkimeni', 'tr': 'Türkmence', 'tt': 'төрекмән', 'ug': 'تۈركمەنچە', 'uk': 'туркменська', 'ur': 'ترکمان', 'uz': 'turkman', 'uz-Cyrl': 'туркманча', 'uz-Latn': 'turkman', 'vec': 'turcoman', 'vi': 'Tiếng Turkmen', 'wae': 'Turkmeniš', 'wo': 'Tirkmen', 'xnr': 'तुर्कमेन', 'yi': 'טורקמעניש', 'yo': 'Èdè Turkmen', 'yrl': 'turikumenu', 'yue': '土庫曼文', 'yue-Hans': '土库曼文', 'yue-Hant': '土庫曼文', 'zh': '土库曼语', 'zh-Hans': '土库曼语', 'zh-Hant': '土庫曼文', 'zu': 'isi-Turkmen'},
    'tka': {'en': 'Truká'},
    'tkb': {'en': 'Buksa'},
    'tkd': {'en': 'Tukudede'},
    'tke': {'en': 'Takwane'},
    'tkf': {'en': 'Tukumanféd'},
    'tkg': {'en': 'Tesaka Malagasy'},
    'tkk': {'en': 'Takpa'},
    'tkl': {'ar': 'التوكيلاو', 'ast': 'tokelau', 'az': 'tokelay', 'az-Latn': 'tokelay', 'bg': 'токелайски', 'bn': 'টোকেলাউ', 'br': 'tokelau', 'brx': 'टोकेलौ', 'bs': 'tokelau', 'bs-Cyrl': 'токелау', 'bs-Latn': 'tokelau', 'ca': 'tokelauès', 'ca-ES-valencia': 'tokelauès', 'ccp': '𑄑𑄮𑄇𑄬𑄣𑄃𑄪', 'cs': 'tokelauština', 'cy': 'Tocelaweg', 'da': 'tokelau', 'de': 'Tokelauanisch', 'el': 'Τοκελάου', 'el-polyton': 'Τοκελάου', 'en': 'Tokelau', 'es': 'tokelauano', 'et': 'tokelau', 'fi': 'tokelau', 'fr': 'tokelau', 'fy': 'Tokelaus', 'gd': 'Tokelau', 'gsw': 'Tokelauanisch', 'gu': 'તોકેલાઉ', 'he': 'טוקלאו', 'hi': 'तोकेलाऊ', 'hi-Latn': 'Tokelau', 'hr': 'tokelaunski', 'ht': 'tokelau', 'hu': 'tokelaui', 'hy': 'տոկելաու', 'id': 'Tokelau', 'is': 'tókeláska', 'it': 'tokelau', 'ja': 'トケラウ語', 'kgp': 'tokeravánũ', 'kn': 'ಟೊಕೆಲಾವ್', 'ko': '토켈라우제도어', 'ks': 'ٹوکیٖلاو', 'ks-Arab': 'ٹوکیٖلاو', 'lb': 'Tokelauanesch', 'lo': 'ໂຕເກເລົາ', 'lt': 'Tokelau', 'lv': 'tokelaviešu', 'mk': 'токелауански', 'ml': 'ടൊക്കേലൗ', 'mr': 'टोकेलाऊ', 'mt': 'Tokelau', 'nb': 'tokelauisk', 'nl': 'Tokelaus', 'nn': 'tokelau', 'no': 'tokelauisk', 'or': 'ଟୋକେଲାଉ', 'pl': 'tokelau', 'pt': 'toquelauano', 'rm': 'tokelau', 'ro': 'tokelau', 'ru': 'токелайский', 'sk': 'tokelauština', 'sl': 'tokelavščina', 'sr': 'токелау', 'sr-Cyrl': 'токелау', 'sr-Latn': 'tokelau', 'sv': 'tokelauiska', 'ta': 'டோகேலௌ', 'te': 'టోకెలావ్', 'th': 'โตเกเลา', 'to': 'lea fakatokelau', 'tr': 'Tokelau dili', 'ug': 'توكېلاۋچە', 'uk': 'токелау', 'vi': 'Tiếng Tokelau', 'yrl': 'tukerauanu', 'yue': '托克勞文', 'yue-Hans': '托克劳文', 'yue-Hant': '托克勞文', 'zh': '托克劳语', 'zh-Hans': '托克劳语', 'zh-Hant': '托克勞文'},
    'tkm': {'en': 'Takelma'},
    'tkn': {'en': 'Toku-No-Shima'},
    'tkp': {'en': 'Tikopia'},
    'tkq': {'en': 'Tee'},
    'tkr': {'ast': 'tsakhur', 'ca': 'tsakhur', 'ca-ES-valencia': 'tsakhur', 'cs': 'cachurština', 'cy': 'Tsakhureg', 'de': 'Tsachurisch', 'en': 'Tsakhur', 'et': 'tsahhi', 'fi': 'tsahuri', 'fr': 'tsakhour', 'gd': 'Tsakhur', 'hi-Latn': 'Tsakhur', 'ht': 'tsakhour', 'hy': 'ցախուր', 'it': 'tsakhur', 'ja': 'ツァフル語', 'ko': '차후르어', 'lb': 'Tsachuresch', 'lt': 'tsakurų', 'mk': 'цахурски', 'nb': 'tsakhursk', 'nl': 'Tsakhur', 'nn': 'tsakhursk', 'no': 'tsakhursk', 'pl': 'cachurski', 'sv': 'tsakhur', 'th': 'แซคเซอร์', 'to': 'lea fakasākuli', 'tr': 'Sahurca', 'yue': '查庫爾文', 'yue-Hans': '查库尔文', 'yue-Hant': '查庫爾文', 'zh-Hant': '查庫爾文'},
    'tks': {'en': 'Takestani'},
    'tkt': {'en': 'Kathoriya Tharu'},
    'tku': {'en': 'Upper Necaxa Totonac'},
    'tkv': {'en': 'Mur Pano'},
    'tkw': {'en': 'Teanu'},
    'tkx': {'en': 'Tangko'},
    'tkz': {'en': 'Takua'},
    'tl': {'am': 'ታጋሎገኛ', 'ar': 'التاغالوغية', 'ast': 'tagalog', 'az': 'taqaloq', 'az-Latn': 'taqaloq', 'bg': 'тагалог', 'bn': 'তাগালগ', 'br': 'tagalog', 'brx': 'तागालोग', 'bs': 'tagalog', 'bs-Cyrl': 'тагалски', 'bs-Latn': 'tagalog', 'ca': 'tagal', 'ca-ES-valencia': 'tagal', 'ccp': '𑄖𑄉𑄣𑄧𑄇𑄴', 'cs': 'tagalog', 'cy': 'Tagalog', 'da': 'tagalog', 'de': 'Tagalog', 'dsb': 'tagalog', 'ee': 'tagalogbe', 'el': 'Τάγκαλογκ', 'el-polyton': 'Τάγκαλογκ', 'en': 'Tagalog', 'eo': 'tagaloga', 'es': 'tagalo', 'et': 'tagalogi', 'eu': 'Tagaloga', 'fa': 'تاگالوگی', 'fi': 'tagalog', 'fil': 'Tagalog', 'fo': 'tagalog', 'fr': 'tagalog', 'fur': 'tagalog', 'fy': 'Tagalog', 'ga': 'Tagálaigis', 'gd': 'Tagalog', 'gl': 'tagalo', 'gsw': 'Tagalog', 'gu': 'ટાગાલોગ', 'he': 'טאגאלוג', 'hi': 'टैगलॉग', 'hi-Latn': 'Tagalog', 'hr': 'tagalog', 'hsb': 'tagalog', 'ht': 'tagalog', 'hu': 'tagalog', 'hy': 'տագալերեն', 'id': 'Tagalog', 'is': 'tagalog', 'it': 'tagalog', 'ja': 'タガログ語', 'kgp': 'tagaro', 'kn': 'ಟ್ಯಾಗಲೋಗ್', 'ko': '타갈로그어', 'kok': 'तगालोग', 'kok-Deva': 'तगालोग', 'ks': 'تَماشیک', 'ks-Arab': 'تَماشیک', 'ksh': 'Tajalloch', 'lb': 'Dagalog', 'lo': 'ຕາກາລອກ', 'lt': 'tagalogų', 'lv': 'tagalu', 'mk': 'тагалог', 'ml': 'തഗാലോഗ്', 'mr': 'टागालोग', 'mt': 'Tagalog', 'nb': 'tagalog', 'nl': 'Tagalog', 'nn': 'tagalog', 'no': 'tagalog', 'or': 'ଟାଗାଲଗ୍', 'pl': 'tagalski', 'pt': 'tagalo', 'rm': 'tagalog', 'ro': 'tagalog', 'ru': 'тагалог', 'sk': 'tagalčina', 'sl': 'tagalogščina', 'sr': 'тагалог', 'sr-Cyrl': 'тагалог', 'sr-Latn': 'tagalog', 'sv': 'tagalog', 'ta': 'டாகாலோக்', 'te': 'టగలాగ్', 'th': 'ตากาล็อก', 'to': 'lea fakatakāloka', 'tr': 'Tagalogca', 'ug': 'تاگالوگچە', 'uk': 'тагальська', 'ur': 'ٹیگا لوگ', 'vi': 'Tiếng Tagalog', 'yi': 'טאַגאַלאג', 'yrl': 'tagaru', 'yue': '塔加路族文', 'yue-Hans': '塔加路族文', 'yue-Hant': '塔加路族文', 'zh': '他加禄语', 'zh-Hans': '他加禄语', 'zh-Hant': '塔加路族文'},
    'tla': {'en': 'Southwestern Tepehuan'},
    'tlb': {'en': 'Tobelo'},
    'tlc': {'en': 'Yecuatla Totonac'},
    'tld': {'en': 'Talaud'},
    'tlf': {'en': 'Telefol'},
    'tlg': {'en': 'Tofanma'},
    'tlh': {'af': 'Klingon', 'am': 'ክሊንጎን', 'ar': 'الكلينجون', 'as': 'ক্লিংগন', 'ast': 'klingon', 'az': 'klinqon', 'az-Cyrl': 'клингон', 'az-Latn': 'klinqon', 'be': 'клінган', 'be-tarask': 'клінган', 'bg': 'клингонски', 'bgn': 'کلینگونی', 'bn': 'ক্লিঙ্গন', 'br': 'klingon', 'brx': 'क्लिङ’न', 'bs': 'klingonski', 'bs-Cyrl': 'клингонски', 'bs-Latn': 'klingonski', 'ca': 'klingonià', 'ca-ES-valencia': 'klingonià', 'ccp': '𑄇𑄳𑄣𑄨𑄋𑄴𑄉𑄧𑄚𑄴', 'ce': 'клингонин', 'chr': 'ᏟᎦᎾ', 'ckb': 'كلینگۆن', 'cs': 'klingonština', 'cy': 'Klingon', 'da': 'klingon', 'de': 'Klingonisch', 'dsb': 'klingonšćina', 'el': 'Κλίνγκον', 'el-polyton': 'Κλίνγκον', 'en': 'Klingon', 'en-Dsrt': '𐐗𐑊𐐮𐑍𐐱𐑌', 'eo': 'klingona', 'es': 'klingon', 'et': 'klingoni', 'eu': 'klingonera', 'fa': 'کلینگون', 'ff-Adlm': '𞤑𞤭𞤤𞤭𞤲𞤺𞤮𞤲𞤪𞤫', 'fi': 'klingon', 'fil': 'Klingon', 'fo': 'klingonskt', 'fr': 'klingon', 'fy': 'Klingon', 'ga': 'Klingon', 'gd': 'Klingon', 'gl': 'klingon', 'gsw': 'Klingonisch', 'gu': 'ક્લિન્ગોન', 'ha': 'Klingon', 'he': 'קלינגונית', 'hi': 'क्लिंगन', 'hi-Latn': 'Klingon', 'hr': 'klingonski', 'hsb': 'klingonšćina', 'ht': 'klingon', 'hu': 'klingon', 'hy': 'կլինգոն', 'ia': 'klingon', 'id': 'Klingon', 'ig': 'Klingon', 'is': 'klingonska', 'it': 'klingon', 'ja': 'クリンゴン語', 'jv': 'Klingon', 'ka': 'კლინგონი', 'kgp': 'krĩngãg', 'kk': 'клингон тілі', 'kk-Cyrl': 'клингон тілі', 'km': 'ឃ្លីនហ្គុន', 'kn': 'ಕ್ಲಿಂಗನ್', 'ko': '클링온어', 'kok': 'क्लिंगॉन', 'kok-Deva': 'क्लिंगॉन', 'ks': 'کِلِنگون', 'ks-Arab': 'کِلِنگون', 'ksh': 'Klingjohnesch', 'ku': 'klîngonî', 'ky': 'клингончо', 'lb': 'Klingonesch', 'lo': 'ຄຣິງກອນ', 'lt': 'klingonų', 'lv': 'klingoņu', 'mai': 'किलिंगौन', 'mi': 'Kirīngona', 'mk': 'клингонски', 'ml': 'ക്ലിംഗോൺ', 'mn': 'клингон', 'mr': 'क्लिंगॉन', 'ms': 'Klingon', 'mt': 'Klingon', 'my': 'ကလင်ဂွန်', 'nb': 'klingon', 'ne': 'क्लिङ्गन', 'nl': 'Klingon', 'nn': 'klingon', 'no': 'klingon', 'om': 'Afaan Kilingon', 'or': 'କ୍ଲିଙ୍ଗନ୍', 'pa': 'ਕਲਿੰਗਨ', 'pa-Guru': 'ਕਲਿੰਗਨ', 'pcm': 'Klíngon', 'pl': 'klingoński', 'ps': 'کلينګاني', 'pt': 'klingon', 'qu': 'Klingon Simi', 'rm': 'klingonic', 'ro': 'klingoniană', 'ru': 'клингонский', 'rw': 'Inyekilingoni', 'sc': 'klingon', 'sd': 'ڪلون', 'sd-Arab': 'ڪلون', 'si': 'ක්ලින්ගොන්', 'sk': 'klingónčina', 'sl': 'klingonščina', 'smn': 'klingonkielâ', 'so': 'Klingon', 'sq': 'klingonisht', 'sr': 'клингонски', 'sr-Cyrl': 'клингонски', 'sr-Latn': 'klingonski', 'sv': 'klingonska', 'sw': 'Kiklingoni', 'ta': 'க்ளிங்கோன்', 'te': 'క్లింగాన్', 'th': 'คลิงงอน', 'ti': 'ክሊንጎን', 'tk': 'klingon dili', 'to': 'lea fakakilingoni', 'tr': 'Klingonca', 'ug': 'كىلىنگونچە', 'uk': 'клінгонська', 'ur': 'کلنگن', 'uz': 'klingon', 'uz-Latn': 'klingon', 'vec': 'klingon', 'vi': 'Tiếng Klingon', 'xnr': 'क्लिंगन', 'yo': 'Èdè Klingoni', 'yrl': 'kirĩgũ', 'yue': '克林貢文', 'yue-Hans': '克林贡文', 'yue-Hant': '克林貢文', 'zh': '克林贡语', 'zh-Hans': '克林贡语', 'zh-Hant': '克林貢文', 'zu': 'isi-Klingon'},
    'tli': {'af': 'Tlingit', 'am': 'ትሊንጊት', 'ar': 'التلينغيتية', 'as': 'লিংগিট', 'ast': 'tlingit', 'az': 'tlinqit', 'az-Latn': 'tlinqit', 'be': 'тлінгіт', 'be-tarask': 'тлінгіт', 'bg': 'тлингит', 'bn': 'ত্লিঙ্গিট', 'br': 'tinglit', 'brx': 'त्लिङीत', 'bs': 'tlingit', 'bs-Cyrl': 'тлингит', 'bs-Latn': 'tlingit', 'ca': 'tlingit', 'ca-ES-valencia': 'tlingit', 'ccp': '𑄖𑄴𑄣𑄨𑄋𑄴𑄉𑄨𑄖𑄴', 'chr': 'ᏟᎩᏘ', 'cs': 'tlingit', 'cy': 'Tlingit', 'da': 'tlingit', 'de': 'Tlingit', 'dsb': 'tlingitšćina', 'el': 'Τλίνγκιτ', 'el-polyton': 'Τλίνγκιτ', 'en': 'Tlingit', 'es': 'tlingit', 'et': 'tlingiti', 'eu': 'tlingitera', 'fa': 'تلین\u200cگیتی', 'ff-Adlm': '𞤚𞤤𞤭𞤲𞤺𞤭𞤼𞤪𞤫', 'fi': 'tlingit', 'fil': 'Tlingit', 'fr': 'tlingit', 'fy': 'Tlingit', 'ga': 'Clincitis', 'gd': 'Tlingit', 'gl': 'tlingit', 'gsw': 'Tlingit-Schpraach', 'gu': 'ટ્લિંગિટ', 'ha': 'Tlingit', 'he': 'טלינגיט', 'hi': 'त्लिंगित', 'hi-Latn': 'Tlingit', 'hr': 'tlingit', 'hsb': 'tlingitšćina', 'ht': 'tlingit', 'hu': 'tlingit', 'hy': 'տլինգիտ', 'ia': 'tlingit', 'id': 'Tlingit', 'ig': 'Tlingit', 'is': 'tlingit', 'it': 'tlingit', 'ja': 'トリンギット語', 'jv': 'Tlingit', 'ka': 'ტლინგიტი', 'kgp': 'tiringite', 'kk': 'тлинкит тілі', 'kk-Cyrl': 'тлинкит тілі', 'km': 'ថ្លីងហ្គីត', 'kn': 'ಟ್ಲಿಂಗಿಟ್', 'ko': '틀링깃족어', 'kok': 'लिंगिट', 'kok-Deva': 'लिंगिट', 'ks': 'ٹِلِنگِت', 'ks-Arab': 'ٹِلِنگِت', 'ku': 'tlingit', 'ky': 'тлинкитче (индей тили)', 'lb': 'Tlingit-Sprooch', 'lo': 'ທລີງກິດ', 'lt': 'tlingitų', 'lv': 'tlinkitu', 'mai': 'टिलिंगिट', 'mi': 'Tirīkiti', 'mk': 'тлингит', 'ml': 'ലിംഗ്വിറ്റ്', 'mn': 'тлингит', 'mr': 'लिंगित', 'ms': 'Tlingit', 'mt': 'Tlingit', 'my': 'တလင်းဂစ်', 'nb': 'tlingit', 'ne': 'ट्लिङ्गिट', 'nl': 'Tlingit', 'nn': 'tlingit', 'no': 'tlingit', 'or': 'ତ୍ଲିଙ୍ଗିଟ୍', 'pa': 'ਟਲਿੰਗਿਟ', 'pa-Guru': 'ਟਲਿੰਗਿਟ', 'pcm': 'Tlingit Lángwej', 'pl': 'tlingit', 'ps': 'ټلینګیت', 'pt': 'tlinguite', 'qu': 'Tlingit Simi', 'rm': 'tlingit', 'ro': 'tlingit', 'ru': 'тлингит', 'sc': 'tlingit', 'sd': 'ٽِلنگٽ', 'sd-Arab': 'ٽِلنگٽ', 'si': 'ට්ලින්ගිට්', 'sk': 'tlingitčina', 'sl': 'tlingitščina', 'so': 'Tlingit', 'sq': 'tlingitisht', 'sr': 'тлингит', 'sr-Cyrl': 'тлингит', 'sr-Latn': 'tlingit', 'sv': 'tlingit', 'sw': 'Kitlingiti', 'ta': 'லிங்கிட்', 'te': 'ట్లింగిట్', 'th': 'ทลิงกิต', 'ti': 'ትሊንጊት', 'tk': 'tlinkit dili', 'to': 'lea fakatilingikīte', 'tr': 'Tlingitçe', 'ug': 'تىلىنگىتچە', 'uk': 'тлінгіт', 'ur': 'ٹلنگٹ', 'uz': 'tlingit', 'uz-Latn': 'tlingit', 'vec': 'tlingit', 'vi': 'Tiếng Tlingit', 'xnr': 'त्लिंगित', 'yo': 'Èdè Tlingiti', 'yrl': 'tirĩgiti', 'yue': '特林基特文', 'yue-Hans': '特林基特文', 'yue-Hant': '特林基特文', 'zh': '特林吉特语', 'zh-Hans': '特林吉特语', 'zh-Hant': '特林基特文', 'zu': 'Tlingit'},
    'tlj': {'en': 'Talinga-Bwisi'},
    'tlk': {'en': 'Taloki'},
    'tll': {'en': 'Tetela'},
    'tlm': {'en': 'Tolomako'},
    'tln': {'en': "Talondo'"},
    'tlo': {'en': 'Talodi'},
    'tlp': {'en': 'Filomena Mata-Coahuitlán Totonac'},
    'tlq': {'en': 'Tai Loi'},
    'tlr': {'en': 'Talise'},
    'tls': {'en': 'Tambotalo'},
    'tlt': {'en': 'Sou Nama'},
    'tlu': {'en': 'Tulehu'},
    'tlv': {'en': 'Taliabu'},
    'tlw': {'en': 'South Wemale'},
    'tlx': {'en': 'Khehek'},
    'tly': {'ast': 'talixín', 'ca': 'talix', 'ca-ES-valencia': 'talix', 'cs': 'talyština', 'cy': 'Talysheg', 'de': 'Talisch', 'en': 'Talysh', 'et': 'talõši', 'fi': 'tališi', 'fr': 'talysh', 'gd': 'Talysh', 'hi-Latn': 'Talysh', 'ht': 'talysh', 'hy': 'թալիշերեն', 'it': 'taliscio', 'ja': 'タリシュ語', 'ko': '탈리쉬어', 'lb': 'Talesch', 'lt': 'talyšų', 'mk': 'талишки', 'ms': 'Talysh', 'nb': 'talysj', 'nl': 'Talysh', 'nn': 'talysj', 'no': 'talysj', 'pl': 'tałyski', 'sv': 'talysh', 'th': 'ทาลิช', 'to': 'lea fakatalisi', 'tr': 'Talışça', 'yue': '塔里什文', 'yue-Hans': '塔里什文', 'yue-Hant': '塔里什文', 'zh-Hant': '塔里什文'},
    'tma': {'en': 'Tama (Chad)'},
    'tmb': {'en': 'Katbol'},
    'tmc': {'en': 'Tumak'},
    'tmd': {'en': 'Haruai'},
    'tme': {'en': 'Tremembé'},
    'tmf': {'en': 'Toba-Maskoy'},
    'tmg': {'en': 'Ternateño'},
    'tmh': {'ar': 'التاماشيك', 'ast': 'tamashek', 'az': 'tamaşek', 'az-Latn': 'tamaşek', 'bg': 'тамашек', 'bn': 'তামাশেক', 'br': 'tamacheg', 'brx': 'तमाशेक', 'bs': 'tamašek', 'bs-Cyrl': 'тамашек', 'bs-Latn': 'tamašek', 'ca': 'amazic', 'ca-ES-valencia': 'amazic', 'ccp': '𑄖𑄟𑄥𑄬𑄇𑄴', 'cs': 'tamašek', 'cy': 'Tamasheceg', 'da': 'tamashek', 'de': 'Tamaseq', 'el': 'Ταμασέκ', 'el-polyton': 'Ταμασέκ', 'en': 'Tamashek', 'es': 'tamashek', 'et': 'tamašeki', 'fa': 'تاماشقی', 'fi': 'tamašek', 'fr': 'tamacheq', 'fy': 'Tamashek', 'gd': 'Tamashek', 'gsw': 'Tamaseq', 'gu': 'તામાશેખ', 'he': 'טמאשק', 'hi': 'तामाशेक', 'hi-Latn': 'Tamashek', 'hr': 'tamašečki', 'ht': 'tamacheq', 'hu': 'tamasek', 'hy': 'տամաշեկ', 'id': 'Tamashek', 'is': 'tamasjek', 'it': 'tamashek', 'ja': 'タマシェク語', 'kgp': 'tamỹséke', 'kn': 'ಟಮಾಷೆಕ್', 'ko': '타마섹어', 'ks': 'تاماشیک', 'ks-Arab': 'تاماشیک', 'lb': 'Tamaseq', 'lo': 'ທາມາກເຊກ', 'lt': 'tamašek', 'lv': 'tuaregu', 'mk': 'тамашек', 'ml': 'ടമഷേക്', 'mr': 'तामाशेक', 'mt': 'Tamashek', 'nb': 'tamasjek', 'nl': 'Tamashek', 'nn': 'tamasjek', 'no': 'tamasjek', 'or': 'ତାମାଶେକ୍', 'pl': 'tamaszek', 'pt': 'tamaxeque', 'rm': 'tamasheq', 'ro': 'tamashek', 'ru': 'тамашек', 'sk': 'tuaregčina', 'sl': 'tamajaščina', 'sr': 'тамашек', 'sr-Cyrl': 'тамашек', 'sr-Latn': 'tamašek', 'sv': 'tamashek', 'ta': 'தமஷேக்', 'te': 'టామషేక్', 'th': 'ทามาเชก', 'to': 'lea fakatamasieki', 'tr': 'Tamaşek', 'ug': 'تاماشېكچە', 'uk': 'тамашек', 'vi': 'Tiếng Tamashek', 'yrl': 'tamaxeki', 'yue': '塔馬奇克文', 'yue-Hans': '塔马奇克文', 'yue-Hant': '塔馬奇克文', 'zh': '塔马奇克语', 'zh-Hans': '塔马奇克语', 'zh-Hant': '塔馬奇克文'},
    'tmi': {'en': 'Tutuba'},
    'tmj': {'en': 'Samarokena'},
    'tmk': {'en': 'Northwestern Tamang'},
    'tml': {'en': 'Tamnim Citak'},
    'tmm': {'en': 'Tai Thanh'},
    'tmn': {'en': 'Taman (Indonesia)'},
    'tmo': {'en': 'Temoq'},
    'tmp': {'en': 'Tai Mène'},
    'tmq': {'en': 'Tumleo'},
    'tmr': {'en': 'Jewish Babylonian Aramaic (ca. 200-1200 CE)'},
    'tms': {'en': 'Tima'},
    'tmt': {'en': 'Tasmate'},
    'tmu': {'en': 'Iau'},
    'tmv': {'en': 'Tembo (Motembo)'},
    'tmw': {'en': 'Temuan'},
    'tmy': {'en': 'Tami'},
    'tmz': {'en': 'Tamanaku'},
    'tn': {'af': 'Tswana', 'am': 'ጽዋና', 'ar': 'التسوانية', 'as': 'ছোৱানা', 'ast': 'tswana', 'az': 'svana', 'az-Cyrl': 'свана', 'az-Latn': 'svana', 'bal-Latn': 'Tuswáná', 'be': 'тсвана', 'be-tarask': 'тсвана', 'bg': 'тсвана', 'bgn': 'تسوانی', 'bn': 'সোয়ানা', 'br': 'tswana', 'brx': 'त्स्वाना', 'bs': 'tsvana', 'bs-Cyrl': 'тсвана', 'bs-Latn': 'tsvana', 'ca': 'setswana', 'ca-ES-valencia': 'setswana', 'ccp': '𑄥𑄱𑄚', 'ce': 'тсвана', 'chr': 'ᏧᏩᎾ', 'ckb': 'تسوانا', 'cs': 'setswanština', 'cy': 'Tswana', 'da': 'tswana', 'de': 'Tswana', 'dsb': 'tswana', 'ee': 'tswanagbe', 'el': 'Τσουάνα', 'el-polyton': 'Τσουάνα', 'en': 'Tswana', 'eo': 'cvana', 'es': 'setsuana', 'et': 'tsvana', 'eu': 'tswanera', 'fa': 'تسوانایی', 'ff-Adlm': '𞤚𞤭𞤧𞤱𞤢𞤲𞤢𞥄𞤪𞤫', 'fi': 'tswana', 'fil': 'Tswana', 'fo': 'tswana', 'fr': 'tswana', 'fy': 'Tswana', 'ga': 'Suáinis', 'gd': 'Tswana', 'gl': 'tswana', 'gsw': 'Tswana-Schpraach', 'gu': 'ત્સ્વાના', 'ha': 'Tswana', 'he': 'סוואנה', 'hi': 'सेत्स्वाना', 'hi-Latn': 'Tswana', 'hr': 'cvana', 'hsb': 'tswana', 'ht': 'tswana', 'hu': 'szecsuáni', 'hy': 'ցվանա', 'ia': 'tswana', 'id': 'Tswana', 'ig': 'Tswana', 'is': 'tsúana', 'it': 'tswana', 'ja': 'ツワナ語', 'jv': 'Tswana', 'ka': 'ტსვანა', 'kgp': 'tisuvanỹ', 'kk': 'тсвана тілі', 'kk-Cyrl': 'тсвана тілі', 'km': 'ស្វាណា', 'kn': 'ಸ್ವಾನಾ', 'ko': '츠와나어', 'kok': 'सेत्स्वाना', 'kok-Deva': 'सेत्स्वाना', 'ks': 'سوانا', 'ks-Arab': 'سوانا', 'ksh': 'Se-Zwahnesch', 'ku': 'tswanayî', 'ky': 'тсванача', 'lb': 'Tswana-Sprooch', 'lld': 'tswana', 'lo': 'ເຕສະວານາ', 'lt': 'tsvanų', 'lv': 'cvanu', 'mai': 'टुआना', 'mi': 'Hawāna', 'mk': 'цвана', 'ml': 'സ്വാന', 'mn': 'цвана', 'mr': 'त्स्वाना', 'ms': 'Tswana', 'mt': 'Tswana', 'my': 'တီဆဝါနာ', 'nb': 'setswana', 'ne': 'ट्स्वाना', 'nl': 'Tswana', 'nn': 'tswana', 'no': 'setswana', 'or': 'ସୱାନା', 'pa': 'ਤਸਵਾਨਾ', 'pa-Guru': 'ਤਸਵਾਨਾ', 'pcm': 'Tswána Lángwej', 'pl': 'setswana', 'ps': 'سووانا', 'pt': 'tswana', 'qu': 'Setsuana Simi', 'rm': 'tswana', 'ro': 'setswana', 'ru': 'тсвана', 'sc': 'tswana', 'sd': 'تسوانا', 'sd-Arab': 'تسوانا', 'si': 'ස්වනා', 'sk': 'tswančina', 'sl': 'cvanščina', 'smn': 'tswanakielâ', 'so': 'Tswana', 'sq': 'cuanaisht', 'sr': 'цвана', 'sr-Cyrl': 'цвана', 'sr-Latn': 'cvana', 'sv': 'tswana', 'sw': 'Kitswana', 'ta': 'ஸ்வானா', 'te': 'స్వానా', 'th': 'สวานา', 'ti': 'ስዋና', 'tk': 'tswana dili', 'tn': 'Setswana', 'to': 'lea fakatisuana', 'tr': 'Setsvana', 'ug': 'سىۋاناچە', 'uk': 'тсвана', 'ur': 'سوانا', 'uz': 'tsvana', 'uz-Latn': 'tsvana', 'vec': 'tswana', 'vi': 'Tiếng Tswana', 'wae': 'Tswana', 'xnr': 'सेत्स्वाना', 'yo': 'Èdè Suwana', 'yrl': 'tisuana', 'yue': '突尼西亞文', 'yue-Hans': '突尼西亚文', 'yue-Hant': '突尼西亞文', 'zh': '茨瓦纳语', 'zh-Hans': '茨瓦纳语', 'zh-Hant': '塞茲瓦納文', 'zu': 'isi-Tswana'},
    'tna': {'en': 'Tacana'},
    'tnb': {'en': 'Western Tunebo'},
    'tnc': {'en': 'Tanimuca-Retuarã'},
    'tnd': {'en': 'Angosturas Tunebo'},
    'tne': {'en': 'Tinoc Kallahan'},
    'tnf': {'en': 'Tangshewi'},
    'tng': {'en': 'Tobanga'},
    'tnh': {'en': 'Maiani'},
    'tni': {'en': 'Tandia'},
    'tnk': {'en': 'Kwamera'},
    'tnl': {'en': 'Lenakel'},
    'tnm': {'en': 'Tabla'},
    'tnn': {'en': 'North Tanna'},
    'tno': {'en': 'Toromono'},
    'tnp': {'en': 'Whitesands'},
    'tnq': {'en': 'Taino'},
    'tnr': {'en': 'Ménik'},
    'tns': {'en': 'Tenis'},
    'tnt': {'en': 'Tontemboan'},
    'tnu': {'en': 'Tay Khang'},
    'tnv': {'en': 'Tangchangya'},
    'tnw': {'en': 'Tonsawang'},
    'tnx': {'en': 'Tanema'},
    'tny': {'en': 'Tongwe'},
    'tnz': {'en': "Ten'edn"},
    'to': {'af': 'Tongaans', 'ak': 'Tonga kasa', 'am': 'ቶንጋን', 'ar': 'التونغية', 'as': 'টোঙ্গান', 'ast': 'tonganu', 'az': 'tonqa', 'az-Cyrl': 'тонган', 'az-Latn': 'tonqa', 'bal-Latn': 'Tongan', 'be': 'танганская', 'be-tarask': 'танганская', 'bg': 'тонгански', 'bgn': 'تونگی', 'bn': 'টোঙ্গান', 'br': 'tonga', 'brx': 'टंगान', 'bs': 'tonganski', 'bs-Cyrl': 'тонга', 'bs-Latn': 'tonganski', 'ca': 'tongalès', 'ca-ES-valencia': 'tongalès', 'ccp': '𑄑𑄮𑄋𑄴𑄉𑄚𑄴', 'ce': 'тонганийн', 'chr': 'ᏙᎾᎦᏂ', 'ckb': 'تۆنگان', 'cs': 'tongánština', 'cy': 'Tongeg', 'da': 'tongansk', 'de': 'Tongaisch', 'dsb': 'tonganšćina', 'dz': 'ཊོང་གྷན་ཁ', 'ee': 'tongagbe', 'el': 'Τονγκανικά', 'el-polyton': 'Τονγκανικά', 'en': 'Tongan', 'eo': 'tongana', 'es': 'tongano', 'et': 'tonga', 'eu': 'tongera', 'fa': 'تونگایی', 'ff-Adlm': '𞤚𞤮𞤲𞤺𞤢𞤲𞤪𞤫', 'fi': 'tonga', 'fil': 'Tongan', 'fo': 'tonganskt', 'fr': 'tongien', 'fy': 'Tongaansk', 'ga': 'Tongais', 'gd': 'Tonga', 'gl': 'tongano', 'gsw': 'Tongaisch', 'gu': 'ટોંગાન', 'ha': 'Tonganci', 'haw': 'Tonga', 'he': 'טונגאית', 'hi': 'टोंगन', 'hi-Latn': 'Tongan', 'hr': 'tonganski', 'hsb': 'tongašćina', 'ht': 'tongien', 'hu': 'tongai', 'hy': 'տոնգերեն', 'ia': 'tongano', 'id': 'Tonga', 'ig': 'Tongan', 'is': 'tongverska', 'it': 'tongano', 'ja': 'トンガ語', 'jv': 'Tonga', 'ka': 'ტონგანური', 'kea': 'tonganes', 'kgp': 'tãnganẽj', 'kk': 'тонган тілі', 'kk-Cyrl': 'тонган тілі', 'km': 'តុងហ្គា', 'kn': 'ಟೋಂಗನ್', 'ko': '통가어', 'kok': 'तोंगान', 'kok-Deva': 'तोंगान', 'ks': 'ٹونگا', 'ks-Arab': 'ٹونگا', 'ksh': 'Tongjahnesch', 'ku': 'tongî', 'ky': 'тонгача', 'lb': 'Tongaesch', 'lkt': 'Tongan Iyápi', 'lld': 'tongaich', 'lo': 'ທອງການ', 'lrc': 'توٙنگان', 'lt': 'tonganų', 'lv': 'tongiešu', 'mai': 'टोंगन', 'mi': 'Tonga', 'mk': 'тонгајски', 'ml': 'ടോംഗൻ', 'mn': 'тонга', 'mr': 'टोंगन', 'ms': 'Tonga', 'mt': 'Tongan', 'my': 'တွန်ဂါ', 'mzn': 'تونگانی', 'nb': 'tongansk', 'ne': 'टोङ्गन', 'nl': 'Tongaans', 'nn': 'tongansk', 'no': 'tongansk', 'nqo': 'ߕߏ߲ߖ߭ߌߞߊ߲', 'or': 'ଟୋଙ୍ଗା', 'pa': 'ਟੌਂਗਨ', 'pa-Guru': 'ਟੌਂਗਨ', 'pcm': 'Tóngan', 'pl': 'tonga', 'ps': 'تونګان', 'pt': 'tonganês', 'qu': 'Tongano Simi', 'rm': 'tonga', 'ro': 'tongană', 'ru': 'тонганский', 'sc': 'tonganu', 'sd': 'تونگن', 'sd-Arab': 'تونگن', 'si': 'ටොංගා', 'sk': 'tongčina', 'sl': 'tongščina', 'smn': 'tongakielâ', 'so': 'Toongan', 'sq': 'tonganisht', 'sr': 'тонгански', 'sr-Cyrl': 'тонгански', 'sr-Latn': 'tonganski', 'sv': 'tonganska', 'sw': 'Kitonga', 'ta': 'டோங்கான்', 'te': 'టాంగాన్', 'tg': 'тонганӣ', 'th': 'ตองกา', 'ti': 'ቶንጋንኛ', 'tk': 'tongan dili', 'to': 'lea fakatonga', 'tr': 'Tonga dili', 'tt': 'тонга', 'ug': 'تونگانچە', 'uk': 'тонганська', 'ur': 'ٹونگن', 'uz': 'tongan', 'uz-Cyrl': 'тонганча', 'uz-Latn': 'tongan', 'vec': 'tongan', 'vi': 'Tiếng Tonga', 'wae': 'Tonga', 'wo': 'Tongan', 'xnr': 'टोंगन', 'yo': 'Tóńgàn', 'yrl': 'tũganei', 'yue': '東加文', 'yue-Hans': '东加文', 'yue-Hant': '東加文', 'zh': '汤加语', 'zh-Hans': '汤加语', 'zh-Hant': '東加文', 'zu': 'isi-Tongan'},
    'tob': {'en': 'Toba'},
    'toc': {'en': 'Coyutla Totonac'},
    'tod': {'en': 'Toma'},
    'toe': {'en': 'Tomedes'},
    'tof': {'en': 'Gizrra'},
    'tog': {'ar': 'تونجا - نياسا', 'ast': 'tonga nyasa', 'az': 'nyasa tonqa', 'az-Latn': 'nyasa tonqa', 'bg': 'нианса тонга', 'bn': 'নায়াসা টোঙ্গা', 'br': 'nyasa tonga', 'brx': 'न्यासा टॉंगा', 'bs': 'njasa tonga', 'bs-Cyrl': 'њаса тонга', 'bs-Latn': 'njasa tonga', 'ca': 'tonga', 'ca-ES-valencia': 'tonga', 'ccp': '𑄚𑄠𑄥𑄑𑄮𑄋𑄴𑄉', 'cs': 'tonžština (nyasa)', 'da': 'nyasa tongansk', 'de': 'Nyasa Tonga', 'el': 'Νιάσα Τόνγκα', 'el-polyton': 'Νιάσα Τόνγκα', 'en': 'Nyasa Tonga', 'es': 'tonga del Nyasa', 'et': 'tšitonga', 'fa': 'تونگایی نیاسا', 'fi': 'malawintonga', 'fr': 'tonga nyasa', 'fy': 'Nyasa Tonga', 'gd': 'Nyasa Tonga', 'gsw': 'Tsonga-Schpraach', 'gu': 'ન્યાસા ટોન્ગા', 'he': 'ניאסה טונגה', 'hi': 'न्यासा टोन्गा', 'hi-Latn': 'Nyasa Tonga', 'hr': 'nyasa tonga', 'ht': 'tonga nyasa', 'hu': 'nyugati nyasza', 'id': 'Nyasa Tonga', 'is': 'tongverska (nyasa)', 'it': 'nyasa del Tonga', 'ja': 'トンガ語(ニアサ)', 'kgp': 'tãnganẽj Nyjasa tá', 'kn': 'ನ್ಯಾಸಾ ಟೋಂಗಾ', 'ko': '니아사 통가어', 'ks': 'نیاسا ٹونگا', 'ks-Arab': 'نیاسا ٹونگا', 'lb': 'Tsonga-Sprooch', 'lo': 'ນາຍອາຊາຕອງກາ', 'lt': 'niasa tongų', 'lv': 'Njasas tongu', 'mk': 'њаса тонга', 'ml': 'ന്യാസാ ഡോങ്ക', 'mr': 'न्यासा टोन्गा', 'mt': 'Nyasa Tonga', 'nb': 'nyasa-tongansk', 'ne': 'न्यास टोङ्गा', 'nl': 'Nyasa Tonga', 'nn': 'tonga (Nyasa)', 'no': 'nyasa-tongansk', 'or': 'ନ୍ୟାସା ଟୋଙ୍ଗୋ', 'pl': 'tonga (Niasa)', 'pt': 'tonganês de Nyasa', 'rm': 'lingua tsonga', 'ro': 'nyasa tonga', 'ru': 'тонга', 'sk': 'ňasa tonga', 'sl': 'malavijska tongščina', 'sr': 'њаса тонга', 'sr-Cyrl': 'њаса тонга', 'sr-Latn': 'njasa tonga', 'sv': 'nyasatonganska', 'ta': 'நயாசா டோங்கா', 'te': 'న్యాసా టోన్గా', 'th': 'ไนอะซาตองกา', 'to': 'lea fakaniasa-tonga', 'tr': 'Nyasa Tonga', 'ug': 'نياسا توڭانچە', 'uk': 'ньяса тонга', 'vi': 'Tiếng Nyasa Tonga', 'yrl': 'tũganei Niasawara', 'yue': '東加文（尼亞薩）', 'yue-Hans': '东加文（尼亚萨）', 'yue-Hant': '東加文（尼亞薩）', 'zh': '尼亚萨汤加语', 'zh-Hans': '尼亚萨汤加语', 'zh-Hant': '東加文（尼亞薩）'},
    'toh': {'en': 'Gitonga'},
    'toi': {'en': 'Tonga (Zambia)'},
    'toj': {'en': 'Tojolabal'},
    'tok': {'af': 'Toki Pona', 'am': 'ቶኪ ፖና', 'ar': 'التوكي-بونا', 'as': 'ট’কি প’না', 'az': 'tokipona', 'az-Latn': 'tokipona', 'bal-Latn': 'Tóki Póná', 'be': 'такіпона', 'be-tarask': 'такіпона', 'bg': 'токи пона', 'bn': 'টোকি পোনা', 'br': 'toki pona', 'brx': 'टकि पना', 'bs': 'toki pona', 'bs-Cyrl': 'токи пона', 'bs-Latn': 'toki pona', 'ca': 'toki pona', 'ca-ES-valencia': 'toki pona', 'chr': 'ᏙᎩ ᏉᎾ', 'cs': 'toki pona', 'cy': 'Toki Pona', 'da': 'toki pona', 'de': 'Toki Pona', 'dsb': 'toki ponašćina', 'el': 'Τόκι Πόνα', 'el-polyton': 'Τόκι Πόνα', 'en': 'Toki Pona', 'es': 'toki pona', 'et': 'toki pona', 'eu': 'toki pona', 'fa': 'توکی پونا', 'ff-Adlm': '𞤚𞤮𞤳𞤭 𞤆𞤮𞤲𞤢𞥄𞤪𞤫', 'fi': 'toki pona', 'fil': 'Toki Pona', 'fr': 'toki pona', 'ga': 'Toki Pona', 'gd': 'Toki Pona', 'gl': 'toki pona', 'gu': 'ટોકી પોના', 'ha': 'Toki Pona', 'he': 'טוקי פונה', 'hi': 'टोकी पोना', 'hi-Latn': 'Toki Pona', 'hr': 'toki pona', 'hsb': 'toki ponašćina', 'ht': 'toki pona', 'hu': 'toki pona', 'hy': 'տոկիպոնա', 'ia': 'toki pona', 'id': 'Toki Pona', 'ig': 'Toki Pona', 'is': 'toki pona', 'it': 'toki pona', 'ja': 'トキポナ語', 'jv': 'Toki Pona', 'ka': 'ტოკი-პონა', 'kk': 'токипона тілі', 'kk-Cyrl': 'токипона тілі', 'km': 'តូគីប៉ូណា', 'kn': 'ಟೋಕಿ ಪೋನಾ', 'ko': '도기 보나', 'kok': 'तोकी पोना', 'kok-Deva': 'तोकी पोना', 'ku': 'toki pona', 'ky': 'токипонача (эксперименталдык тил)', 'lo': 'ໂທກີໂພນາ', 'lt': 'Toki Pona', 'lv': 'tokiponu', 'mai': 'टोकी पोना', 'mi': 'Toki Pona', 'mk': 'токи пона', 'ml': 'ടോകി പോന', 'mn': 'токипона', 'mr': 'टोकि पोना', 'ms': 'Toki Pona', 'my': 'တိုကီပိုနာ', 'nb': 'toki pona', 'ne': 'टोकी पोना', 'nl': 'Toki Pona', 'nn': 'toki pona', 'no': 'toki pona', 'or': 'ଟୋକି ପୋନା', 'pa': 'ਤੋਕੀ ਪੋਨਾ', 'pa-Guru': 'ਤੋਕੀ ਪੋਨਾ', 'pcm': 'Tongán Lángwej', 'pl': 'toki pona', 'ps': 'توکی پونا', 'pt': 'toki pona', 'qu': 'Toki Pona Simi', 'ro': 'toki pona', 'ru': 'токипона', 'sc': 'toki pona', 'sd': 'توڪي پونا', 'sd-Arab': 'توڪي پونا', 'si': 'ටෝකි පොනා', 'sk': 'toki pona', 'sl': 'toki pona', 'so': 'Toki Pona', 'sq': 'tokiponaisht', 'sr': 'токипона', 'sr-Cyrl': 'токипона', 'sr-Latn': 'tokipona', 'sv': 'toki pona', 'sw': 'Kitoki Pona', 'ta': 'டோக்கி போனா', 'te': 'టోకి పోనా', 'th': 'โทคิโพนา', 'ti': 'ቶኪ ፖና', 'tk': 'toki pona dili', 'to': 'lea fakatoki-pona', 'tok': 'toki pona', 'tr': 'Toki Pona', 'uk': 'токі-пона', 'ur': 'ٹوکی پونا', 'uz': 'tokipona', 'uz-Latn': 'tokipona', 'vec': 'toki pona', 'vi': 'Tiếng Toki Pona', 'yo': 'Èdè Toki Pona', 'yue': '托基波納文', 'yue-Hans': '托基波纳文', 'yue-Hant': '托基波納文', 'zh': '道本语', 'zh-Hans': '道本语', 'zh-Hant': '道本語', 'zu': 'Toki Pona'},
    'tol': {'en': 'Tolowa'},
    'tom': {'en': 'Tombulu'},
    'too': {'en': 'Xicotepec De Juárez Totonac'},
    'top': {'en': 'Papantla Totonac'},
    'toq': {'en': 'Toposa'},
    'tor': {'en': 'Togbo-Vara Banda'},
    'tos': {'en': 'Highland Totonac'},
    'tou': {'en': 'Tho'},
    'tov': {'en': 'Upper Taromi'},
    'tow': {'en': 'Jemez'},
    'tox': {'en': 'Tobian'},
    'toy': {'en': 'Topoiyo'},
    'toz': {'en': 'To'},
    'tpa': {'en': 'Taupota'},
    'tpc': {'en': "Azoyú Me'phaa"},
    'tpe': {'en': 'Tippera'},
    'tpf': {'en': 'Tarpia'},
    'tpg': {'en': 'Kula'},
    'tpi': {'af': 'Tok Pisin', 'am': 'ቶክ ፒሲን', 'ar': 'التوك بيسين', 'as': 'টোক পিছিন', 'ast': 'tok pisin', 'az': 'tok pisin', 'az-Cyrl': 'ток писин', 'az-Latn': 'tok pisin', 'bal-Latn': 'Tok Pisin', 'be': 'ток-пісін', 'be-tarask': 'ток-пісін', 'bg': 'ток писин', 'bgn': 'توک\u200cپیسینی', 'bn': 'টোক পিসিন', 'br': 'tok pisin', 'brx': 'ट’क पिसीन', 'bs': 'tok pisin', 'bs-Cyrl': 'ток писин', 'bs-Latn': 'tok pisin', 'ca': 'tok pisin', 'ca-ES-valencia': 'tok pisin', 'ccp': '𑄑𑄮𑄇𑄴 𑄛𑄨𑄥𑄨𑄚𑄴', 'ce': 'ток-писин', 'chr': 'ᏙᎩ ᏈᏏᏂ', 'ckb': 'تۆکپیسین', 'cs': 'tok pisin', 'cy': 'Tok Pisin', 'da': 'tok pisin', 'de': 'Neumelanesisch', 'dsb': 'tok pisinšćina', 'ee': 'tok pisigbe', 'el': 'Τοκ Πισίν', 'el-polyton': 'Τοκ Πισίν', 'en': 'Tok Pisin', 'es': 'tok pisin', 'et': 'uusmelaneesia', 'eu': 'tok pisin', 'fa': 'توک\u200cپیسینی', 'ff-Adlm': '𞤚𞤮𞤳 𞤆𞤭𞤧𞤭𞤲𞤪𞤫', 'fi': 'tok-pisin', 'fil': 'Tok Pisin', 'fo': 'tok pisin', 'fr': 'tok pisin', 'fy': 'Tok Pisin', 'ga': 'Tok Pisin', 'gd': 'Tok Pisin', 'gl': 'tok pisin', 'gsw': 'Neumelanesisch', 'gu': 'ટોક પિસિન', 'ha': 'Tok Pisin', 'he': 'טוק פיסין', 'hi': 'टोक पिसिन', 'hi-Latn': 'Tok Pisin', 'hr': 'tok pisin', 'hsb': 'tok pisinšćina', 'ht': 'tok pisin', 'hu': 'tok pisin', 'hy': 'տոկ փիսին', 'ia': 'tok pisin', 'id': 'Tok Pisin', 'ig': 'Tok pisin', 'is': 'tokpisin', 'it': 'tok pisin', 'ja': 'トク・ピシン語', 'jv': 'Tok Pisin', 'ka': 'ტოკ-პისინი', 'kgp': 'tóg-pisĩn', 'kk': 'ток-писин тілі', 'kk-Cyrl': 'ток-писин тілі', 'km': 'ថុកពីស៊ីន', 'kn': 'ಟೋಕ್ ಪಿಸಿನ್', 'ko': '토크 피신어', 'kok': 'तोक पिसीन', 'kok-Deva': 'तोक पिसीन', 'ks': 'ٹاک پِسِن', 'ks-Arab': 'ٹاک پِسِن', 'ksh': 'Tok Pisin', 'ku': 'tokpisinî', 'ky': 'ток-писинче', 'lb': 'Neimelanesesch', 'lo': 'ທອກພີຊິນ', 'lt': 'Papua pidžinų', 'lv': 'tokpisins', 'mai': 'टोक पिसिन', 'mi': 'Toko Pīhini', 'mk': 'ток писин', 'ml': 'ടോക് പിസിൻ', 'mn': 'ток писин', 'mr': 'टोक पिसिन', 'ms': 'Tok Pisin', 'mt': 'Tok Pisin', 'my': 'တော့ခ် ပိစင်', 'nb': 'tok pisin', 'ne': 'टोक पिसिन', 'nl': 'Tok Pisin', 'nn': 'tok pisin', 'no': 'tok pisin', 'or': 'ଟୋକ୍ ପିସିନ୍', 'pa': 'ਟੋਕ ਪਿਸਿਨ', 'pa-Guru': 'ਟੋਕ ਪਿਸਿਨ', 'pcm': 'Tọk Písin', 'pl': 'tok pisin', 'ps': 'توک پیسین', 'pt': 'tok pisin', 'qu': 'Tok Pisin Simi', 'rm': 'tok pisin', 'ro': 'tok pisin', 'ru': 'ток-писин', 'sc': 'tok pisin', 'sd': 'تاڪ پسن', 'sd-Arab': 'تاڪ پسن', 'si': 'ටොක් පිසින්', 'sk': 'novoguinejský pidžin', 'sl': 'tok pisin', 'smn': 'tok pisin', 'so': 'Tok Pisin', 'sq': 'pisinishte toku', 'sr': 'ток писин', 'sr-Cyrl': 'ток писин', 'sr-Latn': 'tok pisin', 'sv': 'tok pisin', 'sw': 'Kitokpisin', 'ta': 'டோக் பிஸின்', 'te': 'టోక్ పిసిన్', 'th': 'ท็อกพิซิน', 'ti': 'ቶክ ፒሲን', 'tk': 'tok-pisin dili', 'to': 'lea fakatoki-pisini', 'tpi': 'Tok Pisin', 'tr': 'Tok Pisin', 'ug': 'توك-پىسىنچە', 'uk': 'ток-пісін', 'ur': 'ٹوک پِسِن', 'uz': 'tok-piksin', 'uz-Latn': 'tok-piksin', 'vec': 'tok pisin', 'vi': 'Tiếng Tok Pisin', 'wae': 'Niwmelanesiš', 'xnr': 'टोक पिसिन', 'yo': 'Èdè Tok Pisini', 'yrl': 'tuki pisĩ', 'yue': '托比辛文', 'yue-Hans': '托比辛文', 'yue-Hant': '托比辛文', 'zh': '托克皮辛语', 'zh-Hans': '托克皮辛语', 'zh-Hant': '托比辛文', 'zu': 'isi-Tok Pisin'},
    'tpj': {'en': 'Tapieté'},
    'tpk': {'en': 'Tupinikin'},
    'tpl': {'en': "Tlacoapa Me'phaa"},
    'tpm': {'en': 'Tampulma'},
    'tpn': {'en': 'Tupinambá'},
    'tpo': {'en': 'Tai Pao'},
    'tpp': {'en': 'Pisaflores Tepehua'},
    'tpq': {'en': 'Tukpa'},
    'tpr': {'en': 'Tuparí'},
    'tpt': {'en': 'Tlachichilco Tepehua'},
    'tpu': {'en': 'Tampuan'},
    'tpv': {'en': 'Tanapag'},
    'tpw': {'en': 'Tupí'},
    'tpx': {'en': "Acatepec Me'phaa"},
    'tpy': {'en': 'Trumai'},
    'tpz': {'en': 'Tinputz'},
    'tqb': {'en': 'Tembé'},
    'tql': {'en': 'Lehali'},
    'tqm': {'en': 'Turumsa'},
    'tqn': {'en': 'Tenino'},
    'tqo': {'en': 'Toaripi'},
    'tqp': {'en': 'Tomoip'},
    'tqq': {'en': 'Tunni'},
    'tqr': {'en': 'Torona'},
    'tqt': {'en': 'Western Totonac'},
    'tqu': {'en': 'Touo'},
    'tqw': {'en': 'Tonkawa'},
    'tr': {'af': 'Turks', 'agq': 'Tʉʉkìs', 'ak': 'Tɛɛki kasa', 'am': 'ቱርክኛ', 'ar': 'التركية', 'as': 'তুৰ্কী', 'asa': 'Kituruki', 'ast': 'turcu', 'az': 'türk', 'az-Arab': 'آنادولو تۆرکجه\u200cسی', 'az-Cyrl': 'түрк', 'az-Latn': 'türk', 'bal-Latn': 'Tork', 'bas': 'Hɔp u tûrk', 'be': 'турэцкая', 'be-tarask': 'турэцкая', 'bem': 'Ichi Takishi', 'bez': 'Hituluki', 'bg': 'турски', 'bgn': 'تورکی', 'blo': 'gɩturkiija', 'bm': 'turikikan', 'bn': 'তুর্কী', 'br': 'turkeg', 'brx': 'तुर्कीस', 'bs': 'turski', 'bs-Cyrl': 'турски', 'bs-Latn': 'turski', 'ca': 'turc', 'ca-ES-valencia': 'turc', 'ccp': '𑄖𑄪𑄢𑄴𑄇𑄩', 'ce': 'туркойн', 'ceb': 'Turkish', 'cgg': 'Orukuruki', 'chr': 'ᎠᎬᎾ', 'ckb': 'تورکی', 'cs': 'turečtina', 'cv': 'турккӑ', 'cy': 'Tyrceg', 'da': 'tyrkisk', 'dav': 'Kituruki', 'de': 'Türkisch', 'dje': 'Turku senni', 'dsb': 'turkojšćina', 'dyo': 'turki', 'dz': 'ཊཱར་ཀིཤ་ཁ', 'ebu': 'Kĩturuki', 'ee': 'Turkishgbe', 'el': 'Τουρκικά', 'el-polyton': 'Τουρκικά', 'en': 'Turkish', 'eo': 'turka', 'es': 'turco', 'et': 'türgi', 'eu': 'turkiera', 'ewo': 'ǹkɔ́bɔ túrəki', 'fa': 'ترکی استانبولی', 'ff': 'Turkeere', 'ff-Adlm': '𞤚𞤵𞥅𞤪𞤢𞤲𞤳𞤮𞥅𞤪𞤫', 'ff-Latn': 'Turkeere', 'fi': 'turkki', 'fil': 'Turkish', 'fo': 'turkiskt', 'fr': 'turc', 'fur': 'turc', 'fy': 'Turks', 'ga': 'Tuircis', 'gd': 'Turcais', 'gl': 'turco', 'gsw': 'Türkisch', 'gu': 'ટર્કિશ', 'guz': 'Kituruki', 'ha': 'Harshen Turkiyya', 'he': 'טורקית', 'hi': 'तुर्की', 'hi-Latn': 'Turkish', 'hr': 'turski', 'hsb': 'turkowšćina', 'ht': 'turc', 'hu': 'török', 'hy': 'թուրքերեն', 'ia': 'turco', 'id': 'Turki', 'ie': 'turc', 'ig': 'Turkish', 'is': 'tyrkneska', 'it': 'turco', 'ja': 'トルコ語', 'jmc': 'Kyiturukyi', 'jv': 'Turki', 'ka': 'თურქული', 'kab': 'Taṭurkit', 'kam': 'Kituruki', 'kde': 'Chituluchi', 'kea': 'turku', 'kgp': 'turko', 'khq': 'Turku senni', 'ki': 'Kituruki', 'kk': 'түрік тілі', 'kk-Cyrl': 'түрік тілі', 'kln': 'kutitab Turkey', 'km': 'ទួរគី', 'kn': 'ಟರ್ಕಿಶ್', 'ko': '튀르키예어', 'kok': 'तुर्की', 'kok-Deva': 'तुर्की', 'ks': 'تُرکِش', 'ks-Arab': 'تُرکِش', 'ksb': 'Kituuki', 'ksf': 'riturk', 'ksh': 'Törkesch', 'ku': 'tirkî', 'kxv': 'turkis', 'kxv-Deva': 'तुर्की', 'kxv-Latn': 'turkis', 'kxv-Orya': 'ତୁର୍କିସ୍', 'kxv-Telu': 'టర్కిస్', 'ky': 'түркчө', 'lag': 'Kɨturúuki', 'lb': 'Tierkesch', 'lg': 'Lutake', 'lkt': 'Turkish Iyápi', 'lld': 'türch', 'ln': 'litiliki', 'lo': 'ເທີຄິຊ', 'lrc': 'تورکی', 'lt': 'turkų', 'lu': 'Ntuluki', 'luo': 'Kituruki', 'luy': 'Kituruki', 'lv': 'turku', 'mai': 'टर्किश', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lturuki', 'mer': 'Kĩtakĩ', 'mfe': 'tirk', 'mg': 'Tiorka', 'mgh': 'Ituruki', 'mi': 'Tākei', 'mk': 'турски', 'ml': 'ടർക്കിഷ്', 'mn': 'турк', 'mni': 'টর্কিশ', 'mni-Beng': 'টর্কিশ', 'mr': 'तुर्की', 'ms': 'Turki', 'mt': 'Tork', 'mua': 'Turk', 'my': 'တူရကီ', 'mzn': 'ترکی', 'naq': 'Turkeǁî gowab', 'nb': 'tyrkisk', 'nd': 'isi-Thekishi', 'ne': 'टर्किश', 'nl': 'Turks', 'nmg': 'Kiɛl bó turk', 'nn': 'tyrkisk', 'no': 'tyrkisk', 'nqo': 'ߕߙߎߞߌߞߊ߲', 'nus': 'Thok turkicni', 'nyn': 'Orukuruki', 'om': 'Afaan Turkii', 'or': 'ତୁର୍କିସ୍', 'pa': 'ਤੁਰਕੀ', 'pa-Guru': 'ਤੁਰਕੀ', 'pcm': 'Tọ́ki', 'pl': 'turecki', 'ps': 'ترکي', 'pt': 'turco', 'qu': 'Turco Simi', 'rm': 'tirc', 'rn': 'Igiturukiya', 'ro': 'turcă', 'rof': 'Kituruki', 'ru': 'турецкий', 'rw': 'Igiturukiya', 'rwk': 'Kyiturukyi', 'saq': 'Kituruki', 'sbp': 'Ishituluki', 'sc': 'turcu', 'sd': 'ترڪي', 'sd-Arab': 'ترڪي', 'se': 'durkagiella', 'seh': 'turco', 'ses': 'Turku senni', 'sg': 'Tûrûku', 'shi': 'ⵜⴰⵜⵓⵔⴽⵉⵜ', 'shi-Latn': 'Taturkit', 'shi-Tfng': 'ⵜⴰⵜⵓⵔⴽⵉⵜ', 'si': 'තුර්කි', 'sk': 'turečtina', 'sl': 'turščina', 'smn': 'tuurkikielâ', 'sn': 'chiTurkish', 'so': 'Turkish', 'sq': 'turqisht', 'sr': 'турски', 'sr-Cyrl': 'турски', 'sr-Latn': 'turski', 'sv': 'turkiska', 'sw': 'Kituruki', 'syr': 'ܬܘܪܟܝܬ', 'ta': 'துருக்கிஷ்', 'te': 'టర్కిష్', 'teo': 'Kituruki', 'tg': 'туркӣ', 'th': 'ตุรกี', 'ti': 'ቱርክኛ', 'tk': 'türk dili', 'to': 'lea fakatoake', 'tr': 'Türkçe', 'tt': 'төрек', 'twq': 'Turku senni', 'tzm': 'Taturkit', 'ug': 'تۈركچە', 'uk': 'турецька', 'ur': 'ترکی', 'uz': 'turk', 'uz-Cyrl': 'туркча', 'uz-Latn': 'turk', 'vai': 'ꗋꕃ', 'vai-Latn': 'Tɔ́ki', 'vai-Vaii': 'ꗋꕃ', 'vec': 'turco', 'vi': 'Tiếng Thổ Nhĩ Kỳ', 'vun': 'Kyiturukyi', 'wae': 'Türkiš', 'wo': 'Tirk', 'xh': 'Isi-Turkish', 'xnr': 'तुर्की', 'xog': 'Olutake', 'yav': 'nutúluke', 'yo': 'Èdè Tọọkisi', 'yrl': 'turku', 'yue': '土耳其文', 'yue-Hans': '土耳其文', 'yue-Hant': '土耳其文', 'zgh': 'ⵜⴰⵜⵓⵔⴽⵉⵜ', 'zh': '土耳其语', 'zh-Hans': '土耳其语', 'zh-Hant': '土耳其文', 'zu': 'isi-Turkish'},
    'tra': {'en': 'Tirahi'},
    'trb': {'en': 'Terebu'},
    'trc': {'en': 'Copala Triqui'},
    'trd': {'en': 'Turi'},
    'tre': {'en': 'East Tarangan'},
    'trf': {'en': 'Trinidadian Creole English'},
    'trg': {'en': 'Lishán Didán'},
    'trh': {'en': 'Turaka'},
    'tri': {'en': 'Trió'},
    'trj': {'en': 'Toram'},
    'trk': {'en': 'Turkic languages'},
    'trk-oat': {'en': 'Old Anatolian Turkish'},
    'trk-ogz-pro': {'en': 'Proto-Oghuz'},
    'trk-pro': {'en': 'Proto-Turkic'},
    'trl': {'en': 'Traveller Scottish'},
    'trm': {'en': 'Tregami'},
    'trn': {'en': 'Trinitario'},
    'tro': {'en': 'Tarao Naga'},
    'trp': {'en': 'Kok Borok'},
    'trq': {'en': 'San Martín Itunyoso Triqui'},
    'trr': {'en': 'Taushiro'},
    'trs': {'en': 'Chicahuaxtla Triqui'},
    'trt': {'en': 'Tunggare'},
    'tru': {'ast': 'turoyo', 'br': 'turoyoeg', 'cs': 'turojština', 'de': 'Turoyo', 'en': 'Turoyo', 'et': 'turojo', 'fi': 'turojo', 'fr': 'touroyo', 'gd': 'Turoyo', 'hi-Latn': 'Turoyo', 'ht': 'touroyo', 'hy': 'տուրոյո', 'id': 'Turoyo', 'it': 'turoyo', 'ja': 'トゥロヨ語', 'lb': 'Turoyo', 'lt': 'turoyo', 'mk': 'туројо', 'nb': 'turoyo', 'nl': 'Turoyo', 'nn': 'turoyo', 'no': 'turoyo', 'pl': 'turoyo', 'ru': 'туройо', 'sv': 'turoyo', 'th': 'ตูโรโย', 'to': 'lea fakatuloio', 'tr': 'Turoyo', 'yue': '圖羅尤文', 'yue-Hans': '图罗尤文', 'yue-Hant': '圖羅尤文', 'zh-Hant': '圖羅尤文'},
    'trv': {'af': 'Taroko', 'am': 'ታሮኮ', 'ar': 'لغة التاروكو', 'as': 'তাৰোকো', 'ast': 'taroko', 'az': 'taroko', 'az-Cyrl': 'тароко', 'az-Latn': 'taroko', 'bal-Latn': 'Torokó', 'be': 'тарока', 'be-tarask': 'тарока', 'bg': 'тароко', 'bgn': 'تاروکویی', 'bn': 'তারোকো', 'br': 'taroko', 'brx': 'तार’क’', 'bs': 'taroko', 'bs-Cyrl': 'тароко', 'bs-Latn': 'taroko', 'ca': 'taroko', 'ca-ES-valencia': 'taroko', 'ccp': '𑄖𑄢𑄮𑄇𑄮', 'ce': 'седекойн', 'chr': 'ᏔᎶᎪ', 'ckb': 'تارۆکۆ', 'cs': 'taroko', 'cy': 'Taroko', 'da': 'taroko', 'de': 'Taroko', 'dsb': 'tarokošćina', 'el': 'Ταρόκο', 'el-polyton': 'Ταρόκο', 'en': 'Taroko', 'es': 'taroko', 'et': 'taroko', 'eu': 'tarokoera', 'fa': 'تاروکویی', 'ff-Adlm': '𞤚𞤢𞤪𞤮𞤳𞤮𞥅𞤪𞤫', 'fi': 'taroko', 'fil': 'Taroko', 'fo': 'taroko', 'fr': 'taroko', 'fy': 'Taroko', 'ga': 'Taroko', 'gd': 'Taroko', 'gl': 'taroko', 'gu': 'ટારોકો', 'ha': 'Taroko', 'he': 'טרוקו', 'hi': 'तारोको', 'hi-Latn': 'Taroko', 'hr': 'taroko', 'hsb': 'tarokošćina', 'ht': 'taroko', 'hu': 'tarokó', 'hy': 'տարոկո', 'ia': 'taroko', 'id': 'Taroko', 'ig': 'Taroko', 'is': 'tarókó', 'it': 'taroko', 'ja': 'タロコ語', 'jv': 'Taroko', 'ka': 'ტაროკო', 'kgp': 'taroko', 'kk': 'тароко тілі', 'kk-Cyrl': 'тароко тілі', 'km': 'តារ៉ូកូ', 'kn': 'ಟರೊಕೊ', 'ko': '타로코어', 'kok': 'तारोको', 'kok-Deva': 'तारोको', 'ku': 'tarokoyî', 'ky': 'тарокочо', 'lb': 'Seediq', 'lo': 'ຕາໂລໂກ', 'lt': 'Taroko', 'lv': 'taroko', 'mai': 'टरोको', 'mi': 'Taroko', 'mk': 'тароко', 'ml': 'തരോക്കോ', 'mn': 'тароко', 'mr': 'तारोको', 'ms': 'Taroko', 'mt': 'Taroko', 'my': 'တရိုကို', 'nb': 'taroko', 'ne': 'टारोको', 'nl': 'Taroko', 'nn': 'taroko', 'no': 'taroko', 'or': 'ତାରୋକୋ', 'pa': 'ਟਾਰੋਕੋ', 'pa-Guru': 'ਟਾਰੋਕੋ', 'pcm': 'Tarókó Lángwej', 'pl': 'taroko', 'ps': 'تاروکو', 'pt': 'taroko', 'qu': 'Taroko Simi', 'ro': 'taroko', 'ru': 'седекский', 'sc': 'taroko', 'sd': 'تاروڪو', 'sd-Arab': 'تاروڪو', 'si': 'ටරොකො', 'sk': 'taroko', 'sl': 'taroko', 'smn': 'taroko', 'so': 'Taroko', 'sq': 'torokoisht', 'sr': 'тароко', 'sr-Cyrl': 'тароко', 'sr-Latn': 'taroko', 'sv': 'taroko', 'sw': 'Kitaroko', 'ta': 'தரோகோ', 'te': 'తరోకో', 'th': 'ทาโรโก', 'ti': 'ታሮኮ', 'tk': 'taroko dili', 'to': 'lea fakataloko', 'tr': 'Taroko', 'trv': 'patas Taroko', 'ug': 'توروكوچە', 'uk': 'тароко', 'ur': 'ٹوروکو', 'uz': 'taroko', 'uz-Latn': 'taroko', 'vec': 'taroko', 'vi': 'Tiếng Taroko', 'xnr': 'तारोको', 'yo': 'Èdè Taroko', 'yrl': 'taruku', 'yue': '太魯閣文', 'yue-Hans': '太鲁阁文', 'yue-Hant': '太魯閣文', 'zh': '赛德克语', 'zh-Hans': '赛德克语', 'zh-Hant': '太魯閣文', 'zu': 'isi-Taroko'},
    'trw': {'bal-Latn': 'Torwáli', 'en': 'Torwali', 'et': 'torvali', 'fi': 'torwali', 'gd': 'Torwali', 'hi-Latn': 'Torwali', 'ia': 'torwali', 'ig': 'Torwali', 'ku': 'torwalî', 'ti': 'ቶርዋሊኛ', 'trw': 'توروالی', 'vec': 'torwali', 'zh': '托尔瓦利语', 'zh-Hans': '托尔瓦利语'},
    'trx': {'en': 'Tringgus-Sembaan Bidayuh'},
    'try': {'en': 'Turung'},
    'trz': {'en': 'Torá'},
    'ts': {'af': 'Tsonga', 'am': 'ጾንጋ', 'ar': 'السونجا', 'as': 'ছোঙ্গা', 'ast': 'tsonga', 'az': 'sonqa', 'az-Cyrl': 'сонга', 'az-Latn': 'sonqa', 'bal-Latn': 'Tesungá', 'be': 'тсонга', 'be-tarask': 'тсонга', 'bg': 'цонга', 'bgn': 'تسونگی', 'bn': 'সঙ্গা', 'br': 'tsonga', 'brx': 'त्संगा', 'bs': 'tsonga', 'bs-Cyrl': 'тсонга', 'bs-Latn': 'tsonga', 'ca': 'tsonga', 'ca-ES-valencia': 'tsonga', 'ccp': '𑄥𑄧𑄋𑄴𑄉', 'ce': 'тсонга', 'chr': 'ᏦᎾᎦ', 'ckb': 'تسۆنگا', 'cs': 'tsonga', 'cy': 'Tsongaeg', 'da': 'tsonga', 'de': 'Tsonga', 'dsb': 'tsonga', 'ee': 'tsongagbe', 'el': 'Τσόνγκα', 'el-polyton': 'Τσόνγκα', 'en': 'Tsonga', 'eo': 'conga', 'es': 'tsonga', 'et': 'tsonga', 'eu': 'tsongera', 'fa': 'تسونگایی', 'ff-Adlm': '𞤚𞤭𞤧𞤮𞤲𞤺𞤢𞥄𞤪𞤫', 'fi': 'tsonga', 'fil': 'Tsonga', 'fo': 'tsonga', 'fr': 'tsonga', 'fy': 'Tsonga', 'ga': 'Songais', 'gd': 'Tsonga', 'gl': 'tsonga', 'gsw': 'Tsonga', 'gu': 'સોંગા', 'ha': 'Tsonga', 'he': 'טסונגה', 'hi': 'सोंगा', 'hi-Latn': 'Tsonga', 'hr': 'tsonga', 'hsb': 'tsonga', 'ht': 'tsonga', 'hu': 'conga', 'hy': 'ցոնգա', 'ia': 'tsonga', 'id': 'Tsonga', 'ig': 'Tsonga', 'is': 'tsonga', 'it': 'tsonga', 'ja': 'ツォンガ語', 'jv': 'Tsonga', 'ka': 'ტსონგა', 'kgp': 'tesãga', 'kk': 'тсонга тілі', 'kk-Cyrl': 'тсонга тілі', 'km': 'សុងហ្គា', 'kn': 'ಸೋಂಗಾ', 'ko': '총가어', 'kok': 'त्सोंगा', 'kok-Deva': 'त्सोंगा', 'ks': 'ژونگا', 'ks-Arab': 'ژونگا', 'ksh': 'Xi-Zongjahnesch', 'ku': 'tsongayî', 'ky': 'тсонгача', 'lb': 'Tsonga', 'lld': 'tsonga', 'lo': 'ເຕຊອງກາ', 'lt': 'tsongų', 'lv': 'congu', 'mai': 'शोंगा', 'mi': 'Honga', 'mk': 'цонга', 'ml': 'സോംഗ', 'mn': 'цонга', 'mr': 'सोंगा', 'ms': 'Tsonga', 'mt': 'Tsonga', 'my': 'ဆွန်ဂါ', 'nb': 'tsonga', 'ne': 'ट्सोङ्गा', 'nl': 'Tsonga', 'nn': 'tsonga', 'no': 'tsonga', 'or': 'ସୋଙ୍ଗା', 'pa': 'ਸੋਂਗਾ', 'pa-Guru': 'ਸੋਂਗਾ', 'pcm': 'Tsónga Lángwej', 'pl': 'tsonga', 'ps': 'سونګا', 'pt': 'tsonga', 'qu': 'Tsonga Simi', 'rm': 'tsonga', 'ro': 'tsonga', 'ru': 'тсонга', 'sc': 'tsonga', 'sd': 'سونگا', 'sd-Arab': 'سونگا', 'si': 'සොන්ග', 'sk': 'tsongčina', 'sl': 'congščina', 'smn': 'tsongakielâ', 'so': 'Tsonga', 'sq': 'congaisht', 'sr': 'цонга', 'sr-Cyrl': 'цонга', 'sr-Latn': 'conga', 'sv': 'tsonga', 'sw': 'Kitsonga', 'ta': 'ஸோங்கா', 'te': 'సోంగా', 'th': 'ซองกา', 'ti': 'ሶንጋ', 'tk': 'tsonga dili', 'to': 'lea fakatisonga', 'tr': 'Tsonga', 'ug': 'سونگاچە', 'uk': 'тсонга', 'ur': 'زونگا', 'uz': 'tsonga', 'uz-Latn': 'tsonga', 'vec': 'tsonga', 'vi': 'Tiếng Tsonga', 'wae': 'Tsonga', 'xnr': 'सोंगा', 'yo': 'Èdè Songa', 'yrl': 'tesũga', 'yue': '特松加文', 'yue-Hans': '特松加文', 'yue-Hant': '特松加文', 'zh': '聪加语', 'zh-Hans': '聪加语', 'zh-Hant': '特松加文', 'zu': 'isi-Tsonga'},
    'tsa': {'en': 'Tsaangi'},
    'tsb': {'en': 'Tsamai'},
    'tsc': {'en': 'Tswa'},
    'tsd': {'ast': 'tsakoniu', 'cs': 'tsakonština', 'cy': 'Tsaconeg', 'de': 'Tsakonisch', 'en': 'Tsakonian', 'et': 'tsakoonia', 'fi': 'tsakonia', 'fr': 'tsakonien', 'hi-Latn': 'Tsakonian', 'ht': 'tsakonien', 'hy': 'ցակոներեն', 'it': 'zaconico', 'ja': 'ツァコン語', 'lb': 'Tsakonesch', 'lt': 'tsakonų', 'mk': 'цаконски', 'nb': 'tsakonisk', 'nl': 'Tsakonisch', 'nn': 'tsakonisk', 'no': 'tsakonisk', 'pl': 'cakoński', 'sv': 'tsakodiska', 'th': 'ซาโคเนีย', 'to': 'lea fakasakōnia', 'tr': 'Tsakonca', 'yue': '特薩克尼恩文', 'yue-Hans': '特萨克尼恩文', 'yue-Hant': '特薩克尼恩文', 'zh-Hant': '特薩克尼恩文'},
    'tse': {'en': 'Tunisian Sign Language'},
    'tsf': {'en': 'Southwestern Tamang'},
    'tsg': {'en': 'Tausug'},
    'tsh': {'en': 'Tsuvan'},
    'tsi': {'ar': 'التسيمشيان', 'ast': 'tsimshian', 'az': 'simşyan', 'az-Latn': 'simşyan', 'bg': 'цимшиански', 'bn': 'সিমশিয়ান', 'br': 'tsimshian', 'brx': 'त्सीमशीआन्', 'bs': 'tsimšian', 'bs-Cyrl': 'тсимшиан', 'bs-Latn': 'tsimšian', 'ca': 'tsimshià', 'ca-ES-valencia': 'tsimshià', 'ccp': '𑄥𑄨𑄟𑄴𑄥𑄨𑄠𑄚𑄴', 'cs': 'tsimšijské jazyky', 'da': 'tsimshisk', 'de': 'Tsimshian', 'el': 'Τσίμσιαν', 'el-polyton': 'Τσίμσιαν', 'en': 'Tsimshian', 'es': 'tsimshiano', 'et': 'tsimši', 'fa': 'تسیم\u200cشیانی', 'fi': 'tsimši', 'fr': 'tsimshian', 'fy': 'Tsimshian', 'gd': 'Tsimshian', 'gsw': 'Tsimshian-Schpraach', 'gu': 'સિમ્શિયન', 'he': 'טסימשיאן', 'hi': 'त्सिमीशियन', 'hi-Latn': 'Tsimshian', 'hr': 'tsimshian', 'ht': 'tsimshian', 'hu': 'csimsiáni', 'hy': 'ցիմշյան', 'id': 'Tsimshia', 'is': 'tsimsíska', 'it': 'tsimshian', 'ja': 'チムシュ語', 'kgp': 'simsijỹnũ', 'kn': 'ಸಿಂಶಿಯನ್', 'ko': '트심시안어', 'ks': 'ژھِمشِیان', 'ks-Arab': 'ژھِمشِیان', 'lb': 'Tsimshian-Sprooch', 'lo': 'ຊີມຊີແອນ', 'lt': 'tsimšian', 'lv': 'cimšiāņu', 'mk': 'цимшијански', 'ml': 'സിംഷ്യൻ', 'mr': 'सिम्शियन', 'mt': 'Tsimshian', 'nb': 'tsimshian', 'nl': 'Tsimshian', 'nn': 'tsimshian', 'no': 'tsimshian', 'or': 'ତିସିମିସିଆନ୍', 'pl': 'tsimshian', 'pt': 'tsimshiano', 'rm': 'tsimshian', 'ro': 'tsimshian', 'ru': 'цимшиан', 'sk': 'cimšjančina', 'sl': 'tsimščina', 'sr': 'цимшиан', 'sr-Cyrl': 'цимшиан', 'sr-Latn': 'cimšian', 'sv': 'tsimshian', 'ta': 'ட்ஸிம்ஷியன்', 'te': 'శింషీయన్', 'th': 'ซิมชีแอน', 'to': 'lea fakatisīmisiani', 'tr': 'Tsimshian', 'ug': 'سىمشيانچە', 'uk': 'цимшиан', 'vi': 'Tiếng Tsimshian', 'yrl': 'tesĩmĩxianu', 'yue': '欽西安文', 'yue-Hans': '钦西安文', 'yue-Hant': '欽西安文', 'zh': '钦西安语', 'zh-Hans': '钦西安语', 'zh-Hant': '欽西安文'},
    'tsj': {'en': 'Tshangla'},
    'tsk': {'en': 'Tseku'},
    'tsl': {'en': "Ts'ün-Lao"},
    'tsm': {'en': 'Turkish Sign Language'},
    'tsp': {'en': 'Northern Toussian'},
    'tsq': {'en': 'Thai Sign Language'},
    'tsr': {'en': 'Akei'},
    'tss': {'en': 'Taiwan Sign Language'},
    'tst': {'en': 'Tondi Songway Kiini'},
    'tsu': {'en': 'Tsou'},
    'tsv': {'en': 'Tsogo'},
    'tsw': {'en': 'Tsishingini'},
    'tsx': {'en': 'Mubami'},
    'tsy': {'en': 'Tebul Sign Language'},
    'tsz': {'en': 'Purepecha'},
    'tt': {'af': 'Tataars', 'ak': 'Tata kasa', 'am': 'ታታር', 'ar': 'التترية', 'as': 'তাতাৰ', 'ast': 'tártaru', 'az': 'tatar', 'az-Arab': 'تاتارجا', 'az-Cyrl': 'татар', 'az-Latn': 'tatar', 'bal-Latn': 'Tátar', 'be': 'татарская', 'be-tarask': 'татарская', 'bg': 'татарски', 'bgn': 'تاتاری', 'bn': 'তাতার', 'br': 'tatar', 'brx': 'टाटार', 'bs': 'tatarski', 'bs-Cyrl': 'татарски', 'bs-Latn': 'tatarski', 'ca': 'tàtar', 'ca-ES-valencia': 'tàtar', 'ccp': '𑄖𑄖𑄢𑄴', 'ce': 'гӀезалойн', 'chr': 'ᏔᏔ', 'ckb': 'تاتاری', 'cs': 'tatarština', 'cy': 'Tatareg', 'da': 'tatarisk', 'de': 'Tatarisch', 'dsb': 'tataršćina', 'dz': 'ཊ་ཊར་ཁ', 'el': 'Ταταρικά', 'el-polyton': 'Ταταρικά', 'en': 'Tatar', 'eo': 'tatara', 'es': 'tártaro', 'et': 'tatari', 'eu': 'tatarera', 'fa': 'تاتاری', 'ff-Adlm': '𞤚𞤢𞤼𞤢𞤪𞥇𞤪𞤫', 'fi': 'tataari', 'fil': 'Tatar', 'fo': 'tatar', 'fr': 'tatar', 'fur': 'tartar', 'fy': 'Tataars', 'ga': 'Tatairis', 'gd': 'Tatarais', 'gl': 'tártaro', 'gsw': 'Tatarisch', 'gu': 'તતાર', 'ha': 'Tatar', 'he': 'טטרית', 'hi': 'तातार', 'hi-Latn': 'Tatar', 'hr': 'tatarski', 'hsb': 'tataršćina', 'ht': 'tatar', 'hu': 'tatár', 'hy': 'թաթարերեն', 'ia': 'tataro', 'id': 'Tatar', 'ig': 'Asụsụ Tatar', 'is': 'tatarska', 'it': 'tataro', 'ja': 'タタール語', 'jv': 'Tatar', 'ka': 'თათრული', 'kea': 'tatar', 'kgp': 'tartaru', 'kk': 'татар тілі', 'kk-Cyrl': 'татар тілі', 'km': 'តាតា', 'kn': 'ಟಾಟರ್', 'ko': '타타르어', 'kok': 'तातर', 'kok-Deva': 'तातर', 'ks': 'تَتار', 'ks-Arab': 'تَتار', 'ksh': 'Tattahresch', 'ku': 'teterî', 'ky': 'татарча', 'lb': 'Tataresch', 'lkt': 'Tatar Iyápi', 'lld': 'tatarich', 'lo': 'ທາທາ', 'lrc': 'تاتار', 'lt': 'totorių', 'lv': 'tatāru', 'mai': 'ततार', 'mi': 'Tatā', 'mk': 'татарски', 'ml': 'ടാട്ടർ', 'mn': 'татар', 'mr': 'तातर', 'ms': 'Tatar', 'mt': 'Tatar', 'my': 'တာတာ', 'mzn': 'تاتاری', 'nb': 'tatarisk', 'ne': 'तातार', 'nl': 'Tataars', 'nn': 'tatarisk', 'no': 'tatarisk', 'nqo': 'ߕߊߕߊߙߌߞߊ߲', 'or': 'ତାତାର୍', 'pa': 'ਤਤਾਰ', 'pa-Guru': 'ਤਤਾਰ', 'pcm': 'Tatá', 'pl': 'tatarski', 'ps': 'تاتار', 'pt': 'tártaro', 'qu': 'Tartaro Simi', 'rm': 'tatar', 'ro': 'tătară', 'ru': 'татарский', 'sah': 'Татаардыы', 'sc': 'tàtaru', 'sd': 'تاتار', 'sd-Arab': 'تاتار', 'si': 'ටාටර්', 'sk': 'tatárčina', 'sl': 'tatarščina', 'smn': 'tatarkielâ', 'so': 'Taatar', 'sq': 'tatarisht', 'sr': 'татарски', 'sr-Cyrl': 'татарски', 'sr-Latn': 'tatarski', 'sv': 'tatariska', 'sw': 'Kitatari', 'ta': 'டாடர்', 'te': 'టాటర్', 'tg': 'тоторӣ', 'th': 'ตาตาร์', 'ti': 'ታታር', 'tk': 'tatar dili', 'to': 'lea fakatatale', 'tr': 'Tatarca', 'tt': 'татар', 'ug': 'تاتارچە', 'uk': 'татарська', 'ur': 'تاتار', 'uz': 'tatar', 'uz-Cyrl': 'татарча', 'uz-Latn': 'tatar', 'vec': 'tàtaro', 'vi': 'Tiếng Tatar', 'wo': 'Tatar', 'xnr': 'तातार', 'yi': 'טאָטעריש', 'yo': 'Tátárì', 'yrl': 'táritaru', 'yue': '韃靼文', 'yue-Hans': '鞑靼文', 'yue-Hant': '韃靼文', 'zh': '鞑靼语', 'zh-Hans': '鞑靼语', 'zh-Hant': '韃靼文', 'zu': 'isi-Tatar'},
    'tta': {'en': 'Tutelo'},
    'ttb': {'en': 'Gaa'},
    'ttc': {'en': 'Tektiteko'},
    'ttd': {'en': 'Tauade'},
    'tte': {'en': 'Bwanabwana'},
    'ttf': {'en': 'Tuotomb'},
    'ttg': {'en': 'Tutong'},
    'tth': {'en': "Upper Ta'oih"},
    'tti': {'en': 'Tobati'},
    'ttj': {'en': 'Tooro'},
    'ttk': {'en': 'Totoro'},
    'ttl': {'en': 'Totela'},
    'ttm': {'af': 'Noord-Tutchone', 'am': 'ሰሜናዊ ቱትቾን', 'ar': 'التوتشون الشمالية', 'as': 'নৰ্দাৰ্ন টুটচ’ন', 'az': 'şimali tuçon', 'az-Latn': 'şimali tuçon', 'be': 'паўночная тутчонэ', 'be-tarask': 'паўночная тутчонэ', 'bg': 'северен тучоне', 'bn': 'উত্তরাঞ্চলীয় টুচোন', 'br': 'tutchoneg an Norzh', 'brx': 'साहायारि तात्च’ने', 'bs': 'sjeverni tučoni', 'bs-Cyrl': 'сјеверни тутчоне', 'bs-Latn': 'sjeverni tučoni', 'ca': 'tutxone septentrional', 'ca-ES-valencia': 'tutxone septentrional', 'chr': 'ᏧᏴᏢ ᏗᏜ ᏚᏦᏁ', 'cs': 'tutčonština (severní)', 'cy': 'Tutchone gogleddol', 'da': 'nordtutchone', 'de': 'Nördliches Tutchone', 'dsb': 'połnocna tutchonšćina', 'el': 'Βόρεια Τουτσόνε', 'el-polyton': 'Βόρεια Τουτσόνε', 'en': 'Northern Tutchone', 'es': 'tutchone septentrional', 'et': 'põhjatutšoni', 'eu': 'iparraldeko tutchoneera', 'fa': 'تاچونی شمالی', 'ff-Adlm': '𞤚𞤵𞤼𞤷𞤮𞤲𞤫𞤲𞤪𞤫 𞤐𞤢𞤲𞥆𞤢𞥄𞤲𞤺𞤫𞥅𞤪𞤫', 'fi': 'pohjoinen tutchone', 'fil': 'Northern Tutchone', 'fr': 'tutchone du Nord', 'ga': 'Northern Tutchone', 'gd': 'Tutchone Thuathach', 'gl': 'tutchone do norte', 'gu': 'ઉત્તરી ટુચૉન', 'ha': 'Tutchone na Arewaci', 'he': 'טצ׳ון צפונית', 'hi': 'उत्तरी टुशोनी', 'hi-Latn': 'Northern Tutchone', 'hr': 'sjeverni tutchone', 'hsb': 'sewjerna tutchonšćina', 'ht': 'tutchone du Nord', 'hu': 'északi tutchone', 'hy': 'հյուսիսային թուտչոնե', 'ia': 'tutchone del nord', 'id': 'Tutchone Utara', 'ig': 'Northern Tutchone', 'is': 'norður-tutchone', 'it': 'tutchone settentrionale', 'ja': '北部トゥショーニ語', 'jv': 'Tutkhone Sisih Lor', 'ka': 'ჩრდილოეთ ტუჩონი', 'kk': 'солтүстік тутчоне тілі', 'kk-Cyrl': 'солтүстік тутчоне тілі', 'km': 'ថុចឆុនខាងជើង', 'kn': 'ಉತ್ತರ ಟಚ್\u200cವನ್', 'ko': '북부 투톤어', 'kok': 'उत्तरीय तुत्चोने', 'kok-Deva': 'उत्तरीय तुत्चोने', 'ku': 'tutoçenîya bakur', 'ky': 'түндүк тутчончо (индей тили)', 'lo': 'ທັດໂຈນເເໜືອ', 'lt': 'šiaurės tutsonų', 'lv': 'ziemeļu tučonu', 'mai': 'उत्तरी टचोन', 'mi': 'Tūtone ki te Raki', 'mk': 'севернотучонски', 'ml': 'നോർത്തേൺ ടറ്റ്ഷോൺ', 'mn': 'хойд тутчоне', 'mr': 'नॉर्दर्न टचोन', 'ms': 'Tutchone Utara', 'my': 'မြောက် တပ်ချွန်', 'nb': 'nordlig tutchone', 'ne': 'उत्तरी टुचोन', 'nl': 'Noord-Tutchone', 'nn': 'nordleg tutchone', 'no': 'nordlig tutchone', 'or': 'ଉତ୍ତର ଟୁଚୋନ୍', 'pa': 'ਉੱਤਰੀ ਟਚੋਨ', 'pa-Guru': 'ਉੱਤਰੀ ਟਚੋਨ', 'pcm': 'Nótán Tuchónẹ Lángwej', 'pl': 'tutchone północny', 'ps': 'شمالي ټچون', 'pt': 'tutchone setentrional', 'qu': 'Tutchone del Norte Simi', 'ro': 'tutchone de nord', 'ru': 'северный тутчоне', 'sc': 'tutchone setentrionale', 'sd': 'اترين ٽچون', 'sd-Arab': 'اترين ٽچون', 'si': 'උතුරු ටචෝන්', 'sk': 'northern tutchone', 'sl': 'severna tučonščina', 'so': 'Northern Tutchone', 'sq': 'taçoneishte veriore', 'sr': 'северни тучон', 'sr-Cyrl': 'северни тучон', 'sr-Latn': 'severni tučon', 'sv': 'nordtutchone', 'sw': 'Kitutchone cha Kaskazini', 'ta': 'வடக்கு டட்சோன்', 'te': 'ఉత్తర టుట్చోన్', 'th': 'ทัชโชนเหนือ', 'ti': 'ሰሜናዊ ታትቾን', 'tk': 'demirgazyk tutçone dili', 'to': 'lea fakatutisone-tokelau', 'tr': 'Kuzey Tuçoncası', 'uk': 'північна тутчон', 'ur': 'شمالی ٹچون', 'uz': 'shimoliy tutchone', 'uz-Latn': 'shimoliy tutchone', 'vec': 'tutchone setentrionale', 'vi': 'Tiếng Tutchone miền Bắc', 'yo': 'Èdè Ariwa Tusoni', 'yue': '北塔奇訥文', 'yue-Hans': '北塔奇讷文', 'yue-Hant': '北塔奇訥文', 'zh': '北塔穹语', 'zh-Hans': '北塔穹语', 'zh-Hant': '北塔穹語', 'zu': 'Northern Tutchone'},
    'ttn': {'en': 'Towei'},
    'tto': {'en': "Lower Ta'oih"},
    'ttp': {'en': 'Tombelala'},
    'ttq': {'en': 'Tawallammat Tamajaq'},
    'ttr': {'en': 'Tera'},
    'tts': {'en': 'Northeastern Thai'},
    'ttt': {'ast': 'tati musulmán', 'ca': 'tat meridional', 'ca-ES-valencia': 'tat meridional', 'cs': 'tatština', 'de': 'Tatisch', 'en': 'Muslim Tat', 'et': 'lõunataadi', 'fi': 'tati', 'fr': 'tati caucasien', 'gd': 'Tati', 'gu': 'મુસ્લિમ તાટ', 'hi-Latn': 'Muslim Tat', 'ht': 'tati caucasien', 'id': 'Tat Muslim', 'it': 'tat islamico', 'ja': 'ムスリム・タタール語', 'lb': 'Tatesch', 'lt': 'musulmonų tatų', 'mk': 'татски', 'nb': 'muslimsk tat', 'ne': 'मुस्लिम टाट', 'nl': 'Moslim Tat', 'nn': 'muslimsk tat', 'no': 'muslimsk tat', 'pl': 'tacki', 'sv': 'muslimsk tatariska', 'th': 'ตัตมุสลิม', 'to': 'lea fakatati-moselemi', 'tr': 'Tatça', 'yue': '穆斯林塔特文', 'yue-Hans': '穆斯林塔特文', 'yue-Hant': '穆斯林塔特文', 'zh-Hant': '穆斯林塔特文'},
    'ttu': {'en': 'Torau'},
    'ttv': {'en': 'Titan'},
    'ttw': {'en': 'Long Wat'},
    'tty': {'en': 'Sikaritai'},
    'ttz': {'en': 'Tsum'},
    'tua': {'en': 'Wiarumus'},
    'tub': {'en': 'Tübatulabal'},
    'tuc': {'en': 'Mutu'},
    'tud': {'en': 'Tuxá'},
    'tue': {'en': 'Tuyuca'},
    'tuf': {'en': 'Central Tunebo'},
    'tug': {'en': 'Tunia'},
    'tuh': {'en': 'Taulil'},
    'tui': {'en': 'Tupuri'},
    'tuj': {'en': 'Tugutil'},
    'tul': {'en': 'Tula'},
    'tum': {'af': 'Toemboeka', 'am': 'ቱምቡካ', 'ar': 'التامبوكا', 'as': 'তুম্বুকা', 'ast': 'tumbuka', 'az': 'tumbuka', 'az-Cyrl': 'тумбука', 'az-Latn': 'tumbuka', 'be': 'тумбука', 'be-tarask': 'тумбука', 'bg': 'тумбука', 'bgn': 'تومبوکی', 'bn': 'তুম্বুকা', 'br': 'tumbuka', 'brx': 'तुम्बुका', 'bs': 'tumbuka', 'bs-Cyrl': 'тумбука', 'bs-Latn': 'tumbuka', 'ca': 'tumbuka', 'ca-ES-valencia': 'tumbuka', 'ccp': '𑄖𑄪𑄟𑄴𑄝𑄪𑄇', 'ce': 'тумбука', 'chr': 'ᏛᎹᏊᎧ', 'ckb': 'تومبووکا', 'cs': 'tumbukština', 'cy': 'Twmbwca', 'da': 'tumbuka', 'de': 'Tumbuka', 'dsb': 'tumbukašćina', 'el': 'Τουμπούκα', 'el-polyton': 'Τουμπούκα', 'en': 'Tumbuka', 'es': 'tumbuka', 'et': 'tumbuka', 'eu': 'tumbukera', 'fa': 'تومبوکایی', 'ff-Adlm': '𞤚𞤵𞤥𞤦𞤵𞤳𞤢𞥄𞤪𞤫', 'fi': 'tumbuka', 'fil': 'Tumbuka', 'fo': 'tumbuka', 'fr': 'tumbuka', 'fy': 'Toemboeka', 'ga': 'Tumbúicis', 'gd': 'Tumbuka', 'gl': 'tumbuka', 'gsw': 'Tumbuka-Schpraach', 'gu': 'તુમ્બુકા', 'ha': 'Tumbuka', 'he': 'טומבוקה', 'hi': 'तम्बूका', 'hi-Latn': 'Tumbuka', 'hr': 'tumbuka', 'hsb': 'tumbukašćina', 'ht': 'tumbuka', 'hu': 'tumbuka', 'hy': 'տումբուկա', 'ia': 'tumbuka', 'id': 'Tumbuka', 'ig': 'Tumbuka', 'is': 'túmbúka', 'it': 'tumbuka', 'ja': 'トゥンブカ語', 'jv': 'Tumbuka', 'ka': 'ტუმბუკა', 'kgp': 'tũmuka', 'kk': 'тумбука тілі', 'kk-Cyrl': 'тумбука тілі', 'km': 'ទុមប៊ូកា', 'kn': 'ತುಂಬುಕಾ', 'ko': '툼부카어', 'kok': 'तुंबुका', 'kok-Deva': 'तुंबुका', 'ks': 'تُمبُکا', 'ks-Arab': 'تُمبُکا', 'ku': 'tumbukayî', 'ky': 'тумбукача', 'lb': 'Tumbuka-Sprooch', 'lo': 'ຕຳບູກາ', 'lt': 'tumbukų', 'lv': 'tumbuku', 'mai': 'टुमबुका', 'mi': 'Tumūka', 'mk': 'тумбука', 'ml': 'ടുംബുക', 'mn': 'тумбула', 'mr': 'तुम्बुका', 'ms': 'Tumbuka', 'mt': 'Tumbuka', 'my': 'တမ်ဘူကာ', 'nb': 'tumbuka', 'ne': 'टुम्बुका', 'nl': 'Toemboeka', 'nn': 'tumbuka', 'no': 'tumbuka', 'or': 'ଟୁମ୍ବୁକା', 'pa': 'ਤੁੰਬੁਕਾ', 'pa-Guru': 'ਤੁੰਬੁਕਾ', 'pcm': 'Tumbúka Lángwej', 'pl': 'tumbuka', 'ps': 'تامبوکا', 'pt': 'tumbuka', 'qu': 'Tumbuka Simi', 'rm': 'tumbuka', 'ro': 'tumbuka', 'ru': 'тумбука', 'sc': 'tumbuka', 'sd': 'تمبوڪا', 'sd-Arab': 'تمبوڪا', 'si': 'ටුම්බුකා', 'sk': 'tumbuka', 'sl': 'tumbukščina', 'smn': 'tumbukakielâ', 'so': 'Tumbuka', 'sq': 'tumbukaisht', 'sr': 'тумбука', 'sr-Cyrl': 'тумбука', 'sr-Latn': 'tumbuka', 'sv': 'tumbuka', 'sw': 'Kitumbuka', 'ta': 'தும்புகா', 'te': 'టుంబుకా', 'th': 'ทุมบูกา', 'ti': 'ተምቡካ', 'tk': 'tumbuka dili', 'to': 'lea fakatumepuka', 'tr': 'Tumbuka', 'ug': 'تۇمبۇكاچە', 'uk': 'тумбука', 'ur': 'ٹمبوکا', 'uz': 'tumbuka', 'uz-Latn': 'tumbuka', 'vec': 'tunbuka', 'vi': 'Tiếng Tumbuka', 'yo': 'Èdè Tumbuka', 'yrl': 'tũbuka', 'yue': '圖姆布卡文', 'yue-Hans': '图姆布卡文', 'yue-Hant': '圖姆布卡文', 'zh': '通布卡语', 'zh-Hans': '通布卡语', 'zh-Hant': '圖姆布卡文', 'zu': 'isi-Tumbuka'},
    'tun': {'en': 'Tunica'},
    'tuo': {'en': 'Tucano'},
    'tup': {'en': 'Tupi languages'},
    'tuq': {'en': 'Tedaga'},
    'tus': {'en': 'Tuscarora'},
    'tut': {'en': 'Altaic languages'},
    'tut-pro': {'en': 'Proto-Altaic'},
    'tuu': {'en': 'Tututni'},
    'tuv': {'en': 'Turkana'},
    'tuw': {'en': 'Tungus languages'},
    'tuw-pro': {'en': 'Proto-Tungusic'},
    'tux': {'en': 'Tuxináwa'},
    'tuy': {'en': 'Tugen'},
    'tuz': {'en': 'Turka'},
    'tva': {'en': 'Vaghua'},
    'tvd': {'en': 'Tsuvadi'},
    'tve': {'en': "Te'un"},
    'tvk': {'en': 'Southeast Ambrym'},
    'tvl': {'af': 'Tuvalu', 'am': 'ቱቫሉ', 'ar': 'التوفالو', 'as': 'টুভালু', 'ast': 'tuvalu', 'az': 'tuvalu', 'az-Cyrl': 'тувалу', 'az-Latn': 'tuvalu', 'be': 'тувалу', 'be-tarask': 'тувалу', 'bg': 'тувалуански', 'bgn': 'ٹووالی', 'bn': 'টুভালু', 'br': 'tuvalu', 'brx': 'टुभालु', 'bs': 'tuvalu', 'bs-Cyrl': 'тувалу', 'bs-Latn': 'tuvalu', 'ca': 'tuvaluà', 'ca-ES-valencia': 'tuvaluà', 'ccp': '𑄑𑄪𑄞𑄣𑄪', 'ce': 'тувалу', 'chr': 'ᏚᏩᎷ', 'ckb': 'تووڤالوو', 'cs': 'tuvalština', 'cy': 'Twfalweg', 'da': 'tuvaluansk', 'de': 'Tuvaluisch', 'dsb': 'tuvalušćina', 'el': 'Τουβαλού', 'el-polyton': 'Τουβαλού', 'en': 'Tuvalu', 'es': 'tuvaluano', 'et': 'tuvalu', 'eu': 'tuvaluera', 'fa': 'تووالویی', 'ff-Adlm': '𞤚𞤵𞤾𞤢𞤤𞤵𞥅𞤪𞤫', 'fi': 'tuvalu', 'fil': 'Tuvalu', 'fo': 'tuvalu', 'fr': 'tuvalu', 'fy': 'Tuvaluaansk', 'ga': 'Tuvalu', 'gd': 'Tubhalu', 'gl': 'tuvalés', 'gsw': 'Elliceanisch', 'gu': 'તુવાલુ', 'ha': 'Tuvalu', 'he': 'טובאלו', 'hi': 'तुवालु', 'hi-Latn': 'Tuvalu', 'hr': 'tuvaluanski', 'hsb': 'tuvalušćina', 'ht': 'tuvalu', 'hu': 'tuvalu', 'hy': 'թուվալուերեն', 'ia': 'tuvaluano', 'id': 'Tuvalu', 'ig': 'Tuvalu', 'is': 'túvalúska', 'it': 'tuvalu', 'ja': 'ツバル語', 'jv': 'Tupalu', 'ka': 'ტუვალუ', 'kgp': 'tuvaruvỹnũ', 'kk': 'тувалу тілі', 'kk-Cyrl': 'тувалу тілі', 'km': 'ទូវ៉ាលូ', 'kn': 'ಟುವಾಲು', 'ko': '투발루어', 'kok': 'तुवालू', 'kok-Deva': 'तुवालू', 'ks': 'تُوالوٗ', 'ks-Arab': 'تُوالوٗ', 'ku': 'tuvalûyî', 'ky': 'тувалуча', 'lb': 'Elliceanesch', 'lo': 'ຕູວາລູ', 'lt': 'Tuvalu', 'lv': 'tuvaliešu', 'mai': 'टुवालू', 'mi': 'Tuwaru', 'mk': 'тувалуански', 'ml': 'ടുവാലു', 'mn': 'тувалу', 'mr': 'टुवालु', 'ms': 'Tuvalu', 'mt': 'Tuvalu', 'my': 'တူဗာလူ', 'nb': 'tuvalsk', 'ne': 'टुभालु', 'nl': 'Tuvaluaans', 'nn': 'tuvalu', 'no': 'tuvalsk', 'or': 'ତୁଭାଲୁ', 'pa': 'ਟਿਊਵਾਲੂ', 'pa-Guru': 'ਟਿਊਵਾਲੂ', 'pcm': 'Tuválu Lángwej', 'pl': 'tuvalu', 'ps': 'تووالو', 'pt': 'tuvaluano', 'qu': 'Tuvalu Simi', 'rm': 'tuvalu', 'ro': 'tuvalu', 'ru': 'тувалу', 'sc': 'tuvalu', 'sd': 'توالو', 'sd-Arab': 'توالو', 'si': 'ටුවාලු', 'sk': 'tuvalčina', 'sl': 'tuvalujščina', 'smn': 'tuvalukielâ', 'so': 'Tuvalu', 'sq': 'tuvaluisht', 'sr': 'тувалу', 'sr-Cyrl': 'тувалу', 'sr-Latn': 'tuvalu', 'sv': 'tuvaluanska', 'sw': 'Kituvalu', 'ta': 'டுவாலு', 'te': 'టువాలు', 'th': 'ตูวาลู', 'ti': 'ቱቫልዋንኛ', 'tk': 'tuwalu dili', 'to': 'lea fakatūvalu', 'tr': 'Tuvalyanca', 'ug': 'تۇۋالۇچە', 'uk': 'тувалу', 'ur': 'تووالو', 'uz': 'tuvalu', 'uz-Latn': 'tuvalu', 'vec': 'tuvaluan', 'vi': 'Tiếng Tuvalu', 'xnr': 'तुवालु', 'yo': 'Èdè Tifalu', 'yrl': 'tuwaruanu', 'yue': '吐瓦魯文', 'yue-Hans': '吐瓦鲁文', 'yue-Hant': '吐瓦魯文', 'zh': '图瓦卢语', 'zh-Hans': '图瓦卢语', 'zh-Hant': '吐瓦魯文', 'zu': 'isi-Tuvalu'},
    'tvm': {'en': 'Tela-Masbuar'},
    'tvn': {'en': 'Tavoyan'},
    'tvo': {'en': 'Tidore'},
    'tvs': {'en': 'Taveta'},
    'tvt': {'en': 'Tutsa Naga'},
    'tvu': {'en': 'Tunen'},
    'tvw': {'en': 'Sedoa'},
    'tvx': {'en': 'Taivoan'},
    'tvy': {'en': 'Timor Pidgin'},
    'tw': {'af': 'Twi', 'am': 'ትዊኛ', 'ar': 'التوي', 'ast': 'twi', 'az': 'tvi', 'az-Latn': 'tvi', 'bg': 'туи', 'bn': 'টোয়াই', 'br': 'twi', 'brx': 'ट्वी', 'bs': 'tvi', 'bs-Cyrl': 'тви', 'bs-Latn': 'tvi', 'ca': 'twi', 'ca-ES-valencia': 'twi', 'ccp': '𑄑𑄮𑄠𑄭', 'cs': 'twi', 'cy': 'Twi', 'da': 'twi', 'de': 'Twi', 'dsb': 'twi', 'el': 'Τούι', 'el-polyton': 'Τούι', 'en': 'Twi', 'es': 'twi', 'et': 'tvii', 'eu': 'twia', 'fa': 'توی\u200cیایی', 'fi': 'twi', 'fil': 'Twi', 'fo': 'twi', 'fr': 'twi', 'fy': 'Twi', 'ga': 'Tíbhis', 'gd': 'Twi', 'gl': 'twi', 'gsw': 'Twi', 'gu': 'ટ્વાઇ', 'ha': 'Tiwiniyanci', 'he': 'טווי', 'hi': 'ट्वी', 'hi-Latn': 'Twi', 'hr': 'twi', 'hsb': 'twi', 'ht': 'twi', 'hu': 'twi', 'hy': 'տուի', 'ia': 'twi', 'id': 'Twi', 'is': 'tví', 'it': 'ci', 'ja': 'トウィ語', 'ka': 'თუი', 'kgp': 'tuvi', 'kk': 'тви тілі', 'kk-Cyrl': 'тви тілі', 'km': 'ទ្វី', 'kn': 'ಟ್ವಿ', 'ko': '트위어', 'kok': 'त्वि', 'kok-Deva': 'त्वि', 'ks': 'توی', 'ks-Arab': 'توی', 'ky': 'тви', 'lb': 'Twi', 'lo': 'ທວີ', 'lt': 'tvi', 'lv': 'tvī', 'mk': 'тви', 'ml': 'ട്വി', 'mn': 'тви', 'mr': 'ट्वी', 'ms': 'Twi', 'mt': 'Twi', 'nb': 'twi', 'nl': 'Twi', 'nn': 'twi', 'no': 'twi', 'or': 'ତ୍ୱି', 'pa': 'ਤ੍ਵਿ', 'pa-Guru': 'ਤ੍ਵਿ', 'pl': 'twi', 'pt': 'twi', 'rm': 'twi', 'ro': 'twi', 'ru': 'тви', 'rw': 'Inyetuwi', 'sk': 'twi', 'sl': 'tvi', 'sq': 'tuisht', 'sr': 'тви', 'sr-Cyrl': 'тви', 'sr-Latn': 'tvi', 'sv': 'twi', 'sw': 'Twi', 'ta': 'ட்வி', 'te': 'ట్వి', 'th': 'ทวิ', 'ti': 'ትዊ', 'to': 'lea fakatusuī', 'tr': 'Tvi', 'ug': 'تىۋىچە', 'uk': 'тві', 'ur': 'توی', 'vi': 'Tiếng Twi', 'yrl': 'tui', 'yue': '特威文', 'yue-Hans': '特威文', 'yue-Hant': '特威文', 'zh': '契维语', 'zh-Hans': '契维语', 'zh-Hant': '特威文', 'zu': 'isi-Twi'},
    'twa': {'en': 'Twana'},
    'twb': {'en': 'Western Tawbuid'},
    'twc': {'en': 'Teshenawa'},
    'twd': {'en': 'Twents'},
    'twe': {'en': 'Tewa (Indonesia)'},
    'twf': {'en': 'Northern Tiwa'},
    'twg': {'en': 'Tereweng'},
    'twh': {'en': 'Tai Dón'},
    'twl': {'en': 'Tawara'},
    'twm': {'en': 'Tawang Monpa'},
    'twn': {'en': 'Twendi'},
    'two': {'en': 'Tswapong'},
    'twp': {'en': 'Ere'},
    'twq': {'af': 'Tasawaq', 'am': 'ታሳዋክ', 'ar': 'تاساواق', 'as': 'টাছাৱাক', 'ast': 'tasawaq', 'az': 'tasavaq', 'az-Cyrl': 'тасаваг', 'az-Latn': 'tasavaq', 'bal-Latn': 'Tasawak', 'be': 'тасаўак', 'be-tarask': 'тасаўак', 'bg': 'тасавак', 'bgn': 'تسواکی', 'bn': 'তাসাওয়াক', 'br': 'tasawakeg', 'brx': 'तेसावाक', 'bs': 'tasavak', 'bs-Cyrl': 'тасавак', 'bs-Latn': 'tasavak', 'ca': 'tasawaq', 'ca-ES-valencia': 'tasawaq', 'ccp': '𑄖𑄥𑄤𑄇𑄴', 'ce': 'тасавак', 'chr': 'ᏔᏌᏩᎩ', 'ckb': 'تاساواک', 'cs': 'tasawaq', 'cy': 'Tasawaq', 'da': 'tasawaq', 'de': 'Tasawaq', 'dsb': 'tasawaq', 'el': 'Τασαβάκ', 'el-polyton': 'Τασαβάκ', 'en': 'Tasawaq', 'es': 'tasawaq', 'et': 'taswaqi', 'eu': 'tasawaqa', 'fa': 'تسواکی', 'ff-Adlm': '𞤚𞤢𞤧𞤢𞥄𞤹𞤪𞤫', 'fi': 'tasawaq', 'fil': 'Tasawaq', 'fo': 'tasawaq', 'fr': 'tasawaq', 'fy': 'Tasawaq', 'ga': 'Tasawaq', 'gd': 'Tasawaq', 'gl': 'tasawaq', 'gu': 'તસાવાક', 'ha': 'Tasawak', 'he': 'טסוואק', 'hi': 'टासवाक', 'hi-Latn': 'Tasawaq', 'hr': 'tasawaq', 'hsb': 'tasawaq', 'ht': 'tasawaq', 'hu': 'szavák', 'hy': 'տասավաք', 'ia': 'tasawaq', 'id': 'Tasawaq', 'ig': 'Tasawaq', 'is': 'tasawaq', 'it': 'tasawaq', 'ja': 'タサワク語', 'jv': 'Tasawaq', 'ka': 'ტასავაქი', 'kgp': 'tasavag', 'kk': 'тасавак тілі', 'kk-Cyrl': 'тасавак тілі', 'km': 'តាសាវ៉ាក់', 'kn': 'ಟಸವಕ್', 'ko': '타사와크어', 'kok': 'तासावाक', 'kok-Deva': 'तासावाक', 'ksh': 'Tasawaq', 'ku': 'tasawaq', 'ky': 'тасабакча', 'lb': 'Tasawaq', 'lo': 'ຕາຊາວັກ', 'lrc': 'تاساڤاق', 'lt': 'tasavakų', 'lv': 'tasavaku', 'mai': 'टासावाक', 'mi': 'Tahawaka', 'mk': 'тазавак', 'ml': 'ടസവാക്ക്', 'mn': 'тасавак', 'mr': 'तासाव्हाक', 'ms': 'Tasawaq', 'mt': 'Tasawaq', 'my': 'တာဆာဝါချ', 'mzn': 'تاساواقی', 'nb': 'tasawaq', 'ne': 'तासावाक', 'nl': 'Tasawaq', 'nn': 'tasawaq', 'no': 'tasawaq', 'nqo': 'ߕߛߊߥߊߜ߭ߌߞߊ߲', 'or': 'ତାସାୱାକ୍', 'pa': 'ਤਾਸਾਵਿਕ', 'pa-Guru': 'ਤਾਸਾਵਿਕ', 'pcm': 'Tasawak', 'pl': 'tasawaq', 'ps': 'تساواق', 'pt': 'tasawaq', 'qu': 'Tasawaq Simi', 'rm': 'tasawaq', 'ro': 'tasawaq', 'ru': 'тасавак', 'sc': 'tasawaq', 'sd': 'تساوڪي', 'sd-Arab': 'تساوڪي', 'si': 'ටසවාක්', 'sk': 'tasawaq', 'sl': 'tasawaq', 'smn': 'tasawaq', 'so': 'Tasaawaq', 'sq': 'tasavakisht', 'sr': 'тасавак', 'sr-Cyrl': 'тасавак', 'sr-Latn': 'tasavak', 'sv': 'tasawaq', 'sw': 'Kitasawak', 'ta': 'டசவாக்', 'te': 'టసావాఖ్', 'th': 'ตัสซาวัค', 'ti': 'ታሳዋቅ', 'tk': 'tasawak dili', 'to': 'lea fakatasauaki', 'tr': 'Tasawaq', 'twq': 'Tasawaq senni', 'ug': 'شىمالىي سوڭخايچە', 'uk': 'тасавак', 'ur': 'تاساواق', 'uz': 'tasavak', 'uz-Cyrl': 'тасавак', 'uz-Latn': 'tasavak', 'vec': 'tasawac', 'vi': 'Tiếng Tasawaq', 'xnr': 'टासवाक', 'yo': 'Tasawak', 'yrl': 'tasawake', 'yue': '北桑海文', 'yue-Hans': '北桑海文', 'yue-Hant': '北桑海文', 'zh': '北桑海语', 'zh-Hans': '北桑海语', 'zh-Hant': '北桑海文', 'zu': 'isi-Tasawaq'},
    'twr': {'en': 'Southwestern Tarahumara'},
    'twt': {'en': 'Turiwára'},
    'twu': {'en': 'Termanu'},
    'tww': {'en': 'Tuwari'},
    'twx': {'en': 'Tewe'},
    'twy': {'en': 'Tawoyan'},
    'txa': {'en': 'Tombonuo'},
    'txb': {'en': 'Tokharian B'},
    'txc': {'en': 'Tsetsaut'},
    'txe': {'en': 'Totoli'},
    'txg': {'en': 'Tangut'},
    'txh': {'en': 'Thracian'},
    'txi': {'en': 'Ikpeng'},
    'txj': {'en': 'Tarjumo'},
    'txm': {'en': 'Tomini'},
    'txn': {'en': 'West Tarangan'},
    'txo': {'en': 'Toto'},
    'txq': {'en': 'Tii'},
    'txr': {'en': 'Tartessian'},
    'txs': {'en': 'Tonsea'},
    'txt': {'en': 'Citak'},
    'txu': {'en': 'Kayapó'},
    'txx': {'en': 'Tatana'},
    'txy': {'en': 'Tanosy Malagasy'},
    'ty': {'af': 'Tahities', 'am': 'ታሂታንኛ', 'ar': 'التاهيتية', 'as': 'তাহিতিয়ান', 'ast': 'tahitianu', 'az': 'taxiti', 'az-Cyrl': 'тахити', 'az-Latn': 'taxiti', 'be': 'таіці', 'be-tarask': 'таіці', 'bg': 'таитянски', 'bgn': 'تاهیتی', 'bn': 'তাহিতিয়ান', 'br': 'tahitianeg', 'brx': 'टाहीटियान', 'bs': 'tahićanski', 'bs-Cyrl': 'тахићански', 'bs-Latn': 'tahićanski', 'ca': 'tahitià', 'ca-ES-valencia': 'tahitià', 'ccp': '𑄖𑄦𑄨𑄖𑄨𑄠𑄚𑄴', 'ce': 'таитянойн', 'chr': 'ᏔᎯᏘᎠᏂ', 'ckb': 'تاهیتی', 'cs': 'tahitština', 'cy': 'Tahitïeg', 'da': 'tahitiansk', 'de': 'Tahitisch', 'dsb': 'tahitišćina', 'ee': 'tahitigbe', 'el': 'Ταϊτιανά', 'el-polyton': 'Ταϊτιανά', 'en': 'Tahitian', 'es': 'tahitiano', 'et': 'tahiti', 'eu': 'tahitiera', 'fa': 'تاهیتیایی', 'ff-Adlm': '𞤚𞤢𞤸𞤭𞤼𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'tahiti', 'fil': 'Tahitian', 'fo': 'tahitiskt', 'fr': 'tahitien', 'fur': 'tahitian', 'fy': 'Tahitysk', 'ga': 'Taihítis', 'gd': 'Cànan Tahiti', 'gl': 'tahitiano', 'gsw': 'Tahitisch', 'gu': 'તાહિતિયન', 'ha': 'Tahitiyanci', 'haw': 'Polapola', 'he': 'טהיטית', 'hi': 'ताहितियन', 'hi-Latn': 'Tahitian', 'hr': 'tahićanski', 'hsb': 'tahitišćina', 'ht': 'tahitien', 'hu': 'tahiti', 'hy': 'թաիտերեն', 'ia': 'tahitiano', 'id': 'Tahiti', 'ig': 'Tahitian', 'is': 'tahítíska', 'it': 'taitiano', 'ja': 'タヒチ語', 'jv': 'Tahiti', 'ka': 'ტაიტური', 'kgp': 'tajtijỹnũ', 'kk': 'таити тілі', 'kk-Cyrl': 'таити тілі', 'km': 'តាហ៊ីទី', 'kn': 'ಟಹೀಟಿಯನ್', 'ko': '타히티어', 'kok': 'ताहीशियन', 'kok-Deva': 'ताहीशियन', 'ks': 'تاہیشِیَن', 'ks-Arab': 'تاہیشِیَن', 'ksh': 'Tahihtesch', 'ku': 'tahîtî', 'ky': 'таитиче', 'lb': 'Tahitesch', 'lo': 'ຕາຮີຕຽນ', 'lt': 'taitiečių', 'lv': 'taitiešu', 'mai': 'ताहितियन', 'mi': 'Tahiti', 'mk': 'тахитски', 'ml': 'താഹിതിയൻ', 'mn': 'таити', 'mr': 'ताहितीयन', 'ms': 'Tahiti', 'mt': 'Taħitjan', 'my': 'တဟီတီ', 'nb': 'tahitisk', 'ne': 'टाहिटियन', 'nl': 'Tahitiaans', 'nn': 'tahitisk', 'no': 'tahitisk', 'or': 'ତାହିତିଆନ୍', 'pa': 'ਤਾਹੀਟੀ', 'pa-Guru': 'ਤਾਹੀਟੀ', 'pcm': 'Tahítián Lángwej', 'pl': 'tahitański', 'ps': 'تاهیتي', 'pt': 'taitiano', 'qu': 'Tahití Simi', 'rm': 'tahitian', 'ro': 'tahitiană', 'ru': 'таитянский', 'sc': 'taitianu', 'sd': 'تاهيتي', 'sd-Arab': 'تاهيتي', 'se': 'tahitigiella', 'si': 'ටහිටියන්', 'sk': 'tahitčina', 'sl': 'tahitščina', 'smn': 'tahitikielâ', 'so': 'Tahitian', 'sq': 'tahitisht', 'sr': 'тахићански', 'sr-Cyrl': 'тахићански', 'sr-Latn': 'tahićanski', 'sv': 'tahitiska', 'sw': 'Kitahiti', 'ta': 'தஹிதியன்', 'te': 'తహితియన్', 'th': 'ตาฮิตี', 'ti': 'ታሂትኛ', 'tk': 'taiti dili', 'to': 'lea fakatahiti', 'tr': 'Tahiti dili', 'ug': 'تاختىچە', 'uk': 'таїтянська', 'ur': 'تاہیتی', 'uz': 'taiti', 'uz-Latn': 'taiti', 'vec': 'taitian', 'vi': 'Tiếng Tahiti', 'wae': 'Taitiš', 'xnr': 'ताहितियन', 'yo': 'Èdè Tahiti', 'yrl': 'taitianu', 'yue': '大溪地文', 'yue-Hans': '大溪地文', 'yue-Hant': '大溪地文', 'zh': '塔希提语', 'zh-Hans': '塔希提语', 'zh-Hant': '大溪地文', 'zu': 'isi-Tahitian'},
    'tya': {'en': 'Tauya'},
    'tye': {'en': 'Kyanga'},
    'tyh': {'en': "O'du"},
    'tyi': {'en': 'Teke-Tsaayi'},
    'tyj': {'en': 'Tai Do'},
    'tyl': {'en': 'Thu Lao'},
    'tyn': {'en': 'Kombai'},
    'typ': {'en': 'Thaypan'},
    'tyr': {'en': 'Tai Daeng'},
    'tys': {'en': 'Tày Sa Pa'},
    'tyt': {'en': 'Tày Tac'},
    'tyu': {'en': 'Kua'},
    'tyv': {'af': 'Tuvinees', 'am': 'ቱቪንያንኛ', 'ar': 'التوفية', 'as': 'তুভিনিয়ান', 'ast': 'tuvinianu', 'az': 'tuvinyan', 'az-Cyrl': 'тувинјан', 'az-Latn': 'tuvinyan', 'bal-Latn': 'Tuwini', 'be': 'тувінская', 'be-tarask': 'тувінская', 'bg': 'тувински', 'bgn': 'ٹووینی', 'bn': 'টুভিনিয়ান', 'br': 'touva', 'brx': 'तुभिनियान', 'bs': 'tuvinijski', 'bs-Cyrl': 'тувинијски', 'bs-Latn': 'tuvinijski', 'ca': 'tuvinià', 'ca-ES-valencia': 'tuvinià', 'ccp': '𑄑𑄪𑄞𑄨𑄚𑄨𑄠𑄚𑄴', 'ce': 'тувинийн', 'chr': 'ᏚᏫᏂᎠᏂ', 'ckb': 'تووڤینی', 'cs': 'tuvinština', 'cy': 'Twfwnieg', 'da': 'tuvinian', 'de': 'Tuwinisch', 'dsb': 'tuvinianšćina', 'el': 'Τουβινικά', 'el-polyton': 'Τουβινικά', 'en': 'Tuvinian', 'es': 'tuviniano', 'et': 'tõva', 'eu': 'tuvera', 'fa': 'تووایی', 'ff-Adlm': '𞤚𞤵𞤾𞤭𞤲𞤭𞤴𞤢𞤲𞤪𞤫', 'fi': 'tuva', 'fil': 'Tuvinian', 'fo': 'tuvinian', 'fr': 'touvain', 'fy': 'Tuvinysk', 'ga': 'Túvainis', 'gd': 'Cànan Tuva', 'gl': 'tuvaniano', 'gsw': 'Tuwinisch', 'gu': 'ટુવીનિયન', 'ha': 'Tuviniyanci', 'he': 'טובינית', 'hi': 'तुवीनियन', 'hi-Latn': 'Tuvinian', 'hr': 'tuvinski', 'hsb': 'tuvinianšćina', 'ht': 'touvain', 'hu': 'tuvai', 'hy': 'տուվերեն', 'ia': 'tuvano', 'id': 'Tuvinia', 'ig': 'Tuvinian', 'is': 'túvínska', 'it': 'tuvinian', 'ja': 'トゥヴァ語', 'jv': 'Tupinian', 'ka': 'ტუვა', 'kgp': 'tuvinijỹnũ', 'kk': 'тувин тілі', 'kk-Cyrl': 'тувин тілі', 'km': 'ទូវីនៀ', 'kn': 'ಟುವಿನಿಯನ್', 'ko': '투비니안어', 'kok': 'तुविनियन', 'kok-Deva': 'तुविनियन', 'ks': 'تُویٖنیَن', 'ks-Arab': 'تُویٖنیَن', 'ksh': 'Tuvinijahnesch', 'ku': 'tuvanî', 'ky': 'тувинче', 'lb': 'Tuwinesch', 'lo': 'ຕູວີນຽນ', 'lt': 'tuvių', 'lv': 'tuviešu', 'mai': 'टुविनियन', 'mi': 'Tuwīniana', 'mk': 'тувански', 'ml': 'തുവിനിയൻ', 'mn': 'тува', 'mr': 'टुवीनियन', 'ms': 'Tuvinian', 'mt': 'Tuvinjan', 'my': 'တူဗန်', 'nb': 'tuvinsk', 'ne': 'टुभिनियाली', 'nl': 'Toevaans', 'nn': 'tuvinisk', 'no': 'tuvinsk', 'or': 'ତୁଭିନିଆନ୍', 'pa': 'ਤੁਵੀਨੀਅਨ', 'pa-Guru': 'ਤੁਵੀਨੀਅਨ', 'pcm': 'Tuvínián Lángwej', 'pl': 'tuwiński', 'ps': 'توینیان', 'pt': 'tuviniano', 'qu': 'Tuviniano Simi', 'rm': 'tuvinian', 'ro': 'tuvană', 'ru': 'тувинский', 'sc': 'tuvanu', 'sd': 'تووينيائي', 'sd-Arab': 'تووينيائي', 'si': 'ටුවිනියන්', 'sk': 'tuviančina', 'sl': 'tuvinščina', 'smn': 'tuvakielâ', 'so': 'Tuvinia', 'sq': 'tuvinianisht', 'sr': 'тувински', 'sr-Cyrl': 'тувински', 'sr-Latn': 'tuvinski', 'sv': 'tuviniska', 'sw': 'Kituva', 'ta': 'டுவினியன்', 'te': 'టువినియన్', 'th': 'ตูวา', 'ti': 'ቱቪንኛ', 'tk': 'tuwa dili', 'to': 'lea fakatuvīnia', 'tr': 'Tuvaca', 'ug': 'توۋاچە', 'uk': 'тувинська', 'ur': 'تووینین', 'uz': 'tuva', 'uz-Latn': 'tuva', 'vec': 'tuvinian', 'vi': 'Tiếng Tuvinian', 'xnr': 'तुवीनियन', 'yo': 'Èdè Tuvini', 'yrl': 'tuwinianu', 'yue': '土凡文', 'yue-Hans': '土凡文', 'yue-Hant': '土凡文', 'zh': '图瓦语', 'zh-Hans': '图瓦语', 'zh-Hant': '圖瓦文', 'zu': 'isi-Tuvinian'},
    'tyx': {'en': 'Teke-Tyee'},
    'tyy': {'en': 'Tiyaa'},
    'tyz': {'en': 'Tày'},
    'tza': {'en': 'Tanzanian Sign Language'},
    'tzh': {'en': 'Tzeltal'},
    'tzj': {'en': "Tz'utujil"},
    'tzl': {'en': 'Talossan'},
    'tzm': {'af': 'Sentraal-Atlas-Tamazight', 'am': 'መካከለኛው አትላስ ታማዚኛ', 'ar': 'الأمازيغية وسط الأطلس', 'as': 'চেণ্ট্ৰেল এটলাছ টামাজাইট', 'ast': 'tamazight del Atles central', 'az': 'Mərkəzi Atlas tamazicəsi', 'az-Cyrl': 'Мәркәзи Атлас тамазиҹәси', 'az-Latn': 'Mərkəzi Atlas tamazicəsi', 'be': 'сярэднеатлаская тамазігхт', 'be-tarask': 'сярэднеатлаская тамазігхт', 'bg': 'централноатласки тамазигт', 'bgn': 'مرکزین اتلسین تامازیگتی', 'bn': 'সেন্ট্রাল আটলাস তামাজিগাত', 'br': 'tamazigteg Kreizatlas', 'brx': 'सेन्ट्रेल आटलास तामाजाईट', 'bs': 'centralnoatlaski tamazigt', 'bs-Cyrl': 'централноатласки тамазихт', 'bs-Latn': 'centralnoatlaski tamazigt', 'ca': 'amazic del Marroc central', 'ca-ES-valencia': 'amazic del Marroc central', 'ccp': '𑄥𑄬𑄚𑄴𑄑𑄳𑄢𑄣𑄴 𑄃𑄣𑄴𑄖𑄌𑄴 𑄖𑄟𑄎𑄨𑄉𑄖𑄴', 'ce': 'тамазигхтийн', 'chr': 'ᎠᏰᏟ ᎡᎶᎯ ᏓᏟᎶᏍᏗᏓᏅᎢ ᏔᎹᏏᏘ', 'ckb': 'ئەمازیغی ناوەڕاست', 'cs': 'tamazight (střední Maroko)', 'cy': 'Tamazight Canol yr Atlas', 'da': 'centralmarokkansk tamazight', 'de': 'Zentralatlas-Tamazight', 'dsb': 'centralnoatlaski tamazight', 'el': 'Ταμαζίτ Κεντρικού Μαρόκο', 'el-polyton': 'Ταμαζίτ Κεντρικού Μαρόκο', 'en': 'Central Atlas Tamazight', 'es': 'tamazight del Atlas Central', 'et': 'tamasikti', 'eu': 'Erdialdeko Atlaseko amazigera', 'fa': 'آمازیغی اطلس مرکزی', 'ff-Adlm': '𞤚𞤢𞤥𞤢𞤶𞤭𞤼𞤪𞤫 𞤅𞤢𞤲𞤼𞤪𞤢𞤤 𞤀𞤼𞤤𞤢𞤧', 'fi': 'keskiatlaksentamazight', 'fil': 'Central Atlas Tamazight', 'fo': 'miðatlasfjøll tamazight', 'fr': 'amazighe de l’Atlas central', 'fy': 'Tamazight (Sintraal-Marokko)', 'ga': 'Tamaisis Atlais Láir', 'gd': 'Tamaisich an Atlais Mheadhanaich', 'gl': 'tamazight de Marrocos central', 'gu': 'સેન્ટ્રલ એટલાસ તામાઝિટ', 'ha': 'Tamazight na Atlas Tsaka', 'he': 'תמאזיגת של מרכז מרוקו', 'hi': 'मध्य एटलस तमाज़ित', 'hi-Latn': 'Central Atlas Tamazight', 'hr': 'tamašek (Srednji Atlas)', 'hsb': 'tamazight (srjedźny Marokko)', 'ht': 'amazighe de l’Atlas central', 'hu': 'közép-atlaszi tamazigt', 'hy': 'կենտրոնատլասյան թամազիղտ', 'ia': 'tamazight del Atlas Central', 'id': 'Tamazight Maroko Tengah', 'ig': 'Central Atlas Tamazight', 'is': 'tamazight', 'it': 'tamazight', 'ja': '中央アトラス・タマジクト語', 'jv': 'Tamazight Atlas Tengah', 'ka': 'ცენტრალური მოროკოს ტამაზიგხტი', 'kea': 'tamaziti di Atlas Sentral', 'kgp': 'tamỹjirte Atara Kuju tá', 'kk': 'орталық атлас тамазигхт тілі', 'kk-Cyrl': 'орталық атлас тамазигхт тілі', 'km': 'តាម៉ាសាយអាត្លាសកណ្តាល', 'kn': 'ಮಧ್ಯ ಅಟ್ಲಾಸ್ ಟಮಜೈಟ್', 'ko': '중앙 모로코 타마지트어', 'kok': 'केंद्रीय अटलास तामाझायट', 'kok-Deva': 'केंद्रीय अटलास तामाझायट', 'ksh': 'Tamasicht ussem meddlere Atlas', 'ku': 'temazîxtî', 'ky': 'борбордук Атлас тамазигтче', 'lb': 'Mëttlert-Atlas-Tamazight', 'lo': 'ອັດລາສ ທາມາຊີກ ກາງ', 'lrc': 'تامازیغ مینجایی', 'lt': 'Centrinio Maroko tamazitų', 'lv': 'Centrālmarokas tamazīts', 'mai': 'सेंट्रल एटलस टेमेजिघ', 'mi': 'Tamahīta Te Puku o Atarihi', 'mk': 'централноатлански тамазитски', 'ml': 'മധ്യ അറ്റ്\u200cലസ് ടമാസൈറ്റ്', 'mn': 'Төв Атласын тамазигхт', 'mr': 'मध्य अ\u200dॅटलास तॅमाझाइट', 'ms': 'Tamazight Atlas Tengah', 'mt': 'Tamazight tal-Atlas Ċentrali', 'my': 'အလယ်အာ့တလာစ် တာမာဇိုက်', 'mzn': 'میونی اطلس تامزیقی', 'nb': 'sentralmarokkansk tamazight', 'ne': 'केन्द्रीय एट्लास टामाजिघट', 'nl': 'Tamazight (Centraal-Marokko)', 'nn': 'sentral-tamazight', 'no': 'sentralmarokkansk tamazight', 'nqo': 'ߊߕߌߟߊ߲ߕߊ߫ ߕߊ߲ߓߊ߲ ߊߡߊ߲ߖ߭ߌ߲ߞߊ߲', 'or': 'କେନ୍ଦ୍ରୀୟ ଆଟଲାସ୍ ଟାମାଜିଘାଟ୍', 'pa': 'ਮੱਧ ਐਟਲਸ ਤਮਾਜ਼ਿਤ', 'pa-Guru': 'ਮੱਧ ਐਟਲਸ ਤਮਾਜ਼ਿਤ', 'pcm': 'Mídúl Atlás Támazígt Lángwej', 'pl': 'tamazight (Atlas Środkowy)', 'ps': 'مرکزی اطلس تمازائيٹ', 'pt': 'tamazight do Atlas Central', 'qu': 'Tamazight Simi', 'rm': 'tamazight', 'ro': 'tamazight din Atlasul Central', 'ru': 'среднеатласский тамазигхтский', 'sc': 'tamazight de s’Atlànte tzentrale', 'sd': 'وچ اٽلس تمازائيٽ', 'sd-Arab': 'وچ اٽلس تمازائيٽ', 'si': 'මධ්\u200dයම ඇට්ලස් ටමසිට්', 'sk': 'tamazight (stredomarocký)', 'sl': 'tamašek (Srednji Atlas)', 'smn': 'Koskâatlas tamazight', 'so': 'Bartamaha Atlaas Tamasayt', 'sq': 'tamazajtisht e Atlasit Qendror', 'sr': 'централноатласки тамашек', 'sr-Cyrl': 'централноатласки тамашек', 'sr-Latn': 'centralnoatlaski tamašek', 'sv': 'centralmarockansk tamazight', 'sw': 'Kitamazighati cha Atlasi ya Kati', 'ta': 'மத்திய அட்லஸ் டமசைட்', 'te': 'సెంట్రల్ అట్లాస్ టామాజైట్', 'tg': 'тамазайти Атласи Марказӣ', 'th': 'ทามาไซต์แอตลาสกลาง', 'ti': 'ማእከላይ ኣትላስ ታማዛይት', 'tk': 'orta-atlas tamazight dili', 'to': 'lea fakatamasaiti-ʻatilasi-loloto', 'tr': 'Orta Atlas Tamazigti', 'tt': 'үзәк атлас тамазигт', 'tzm': 'Tamaziɣt n laṭlaṣ', 'ug': 'مەركىزىي ئاتلاس تامازايتچە', 'uk': 'центральноатласька тамазігт', 'ur': 'سینٹرل ایٹلس ٹمازائٹ', 'uz': 'markaziy atlas tamazigxt', 'uz-Cyrl': 'марказий атлас тамазигхт', 'uz-Latn': 'markaziy atlas tamazigxt', 'vec': 'tamazigh de l’Atlante sentrale', 'vi': 'Tiếng Tamazight Miền Trung Ma-rốc', 'wo': 'Tamasis gu Digg Atlaas', 'xnr': 'बिचला एटलस तमाज़ित', 'yo': 'Ààrin Gbùngbùn Atlas Tamazight', 'yrl': 'tamasiriti Átras katuwa', 'yue': '塔馬齊格特文', 'yue-Hans': '塔马齐格特文', 'yue-Hant': '塔馬齊格特文', 'zh': '塔马齐格特语', 'zh-Hans': '塔马齐格特语', 'zh-Hant': '中阿特拉斯塔馬塞特文', 'zu': 'isi-Central Atlas Tamazight'},
    'tzn': {'en': 'Tugun'},
    'tzo': {'en': 'Tzotzil'},
    'tzx': {'en': 'Tabriak'},
    'uam': {'en': 'Uamué'},
    'uan': {'en': 'Kuan'},
    'uar': {'en': 'Tairuma'},
    'uba': {'en': 'Ubang'},
    'ubi': {'en': 'Ubi'},
    'ubl': {'en': "Buhi'non Bikol"},
    'ubr': {'en': 'Ubir'},
    'ubu': {'en': 'Umbu-Ungu'},
    'uby': {'en': 'Ubykh'},
    'uda': {'en': 'Uda'},
    'ude': {'en': 'Udihe'},
    'udg': {'en': 'Muduga'},
    'udi': {'en': 'Udi'},
    'udj': {'en': 'Ujir'},
    'udl': {'en': 'Wuzlam'},
    'udm': {'af': 'Udmurt', 'am': 'ኡድሙርት', 'ar': 'الأدمرت', 'as': 'উদমুৰ্ত', 'ast': 'udmurt', 'az': 'udmurt', 'az-Cyrl': 'удмурт', 'az-Latn': 'udmurt', 'be': 'удмурцкая', 'be-tarask': 'удмурцкая', 'bg': 'удмуртски', 'bgn': 'اوڈمورتی', 'bn': 'উডমুর্ট', 'br': 'oudmourteg', 'brx': 'उडमुर्त', 'bs': 'udmurt', 'bs-Cyrl': 'удмурт', 'bs-Latn': 'udmurt', 'ca': 'udmurt', 'ca-ES-valencia': 'udmurt', 'ccp': '𑄃𑄪𑄓𑄴𑄟𑄪𑄢𑄴𑄑𑄧', 'ce': 'удмуртийн', 'chr': 'ᎤᏚᎷᏘ', 'ckb': 'ئوودموورت', 'cs': 'udmurtština', 'cy': 'Fotiaceg', 'da': 'udmurt', 'de': 'Udmurtisch', 'dsb': 'udmurtšćina', 'el': 'Ουντμούρτ', 'el-polyton': 'Ουντμούρτ', 'en': 'Udmurt', 'es': 'udmurt', 'et': 'udmurdi', 'eu': 'udmurtera', 'fa': 'اودمورتی', 'ff-Adlm': '𞤓𞤣𞤥𞤵𞤪𞤼𞤪𞤫', 'fi': 'udmurtti', 'fil': 'Udmurt', 'fo': 'udmurt', 'fr': 'oudmourte', 'fy': 'Oedmoerts', 'ga': 'Udmairtis', 'gd': 'Udmurt', 'gl': 'udmurto', 'gsw': 'Udmurtisch', 'gu': 'ઉદમુર્ત', 'ha': 'Udmurt', 'he': 'אודמורט', 'hi': 'उदमुर्त', 'hi-Latn': 'Udmurt', 'hr': 'udmurtski', 'hsb': 'udmurtšćina', 'ht': 'oudmourte', 'hu': 'udmurt', 'hy': 'ուդմուրտերեն', 'ia': 'udmurto', 'id': 'Udmurt', 'ig': 'Udmurt', 'is': 'údmúrt', 'it': 'udmurt', 'ja': 'ウドムルト語', 'jv': 'Udmurt', 'ka': 'უდმურტული', 'kgp': 'unmũrte', 'kk': 'удмурт тілі', 'kk-Cyrl': 'удмурт тілі', 'km': 'អាត់មូដ', 'kn': 'ಉಡ್\u200cಮುರ್ಟ್', 'ko': '우드말트어', 'kok': 'उडमुर्त', 'kok-Deva': 'उडमुर्त', 'ks': 'اُدمُرت', 'ks-Arab': 'اُدمُرت', 'ksh': 'Udmuchtesch', 'ku': 'udmurtî', 'ky': 'удмуртча', 'lb': 'Udmurtesch', 'lo': 'ອຸດມັດ', 'lt': 'udmurtų', 'lv': 'udmurtu', 'mai': 'उदमुर्त', 'mi': 'Ūmutu', 'mk': 'удмуртски', 'ml': 'ഉഡ്മുർട്ട്', 'mn': 'удмурт', 'mr': 'उदमुर्त', 'ms': 'Udmurt', 'mt': 'Udmurt', 'my': 'အူမူရတ်', 'nb': 'udmurtisk', 'ne': 'उड्मुर्ट', 'nl': 'Oedmoerts', 'nn': 'udmurt', 'no': 'udmurtisk', 'or': 'ଉଦମୂର୍ତ୍ତ', 'pa': 'ਉਦਮੁਰਤ', 'pa-Guru': 'ਉਦਮੁਰਤ', 'pcm': 'Údmurt Lángwej', 'pl': 'udmurcki', 'ps': 'ادمورت', 'pt': 'udmurte', 'qu': 'Udmurt Simi', 'rm': 'udmurt', 'ro': 'udmurt', 'ru': 'удмуртский', 'sc': 'udmurtu', 'sd': 'ادمرت', 'sd-Arab': 'ادمرت', 'se': 'udmurtagiella', 'si': 'අඩ්මර්ට්', 'sk': 'udmurtčina', 'sl': 'udmurtščina', 'smn': 'udmurtkielâ', 'so': 'Udmurt', 'sq': 'udmurtisht', 'sr': 'удмуртски', 'sr-Cyrl': 'удмуртски', 'sr-Latn': 'udmurtski', 'sv': 'udmurtiska', 'sw': 'Kiudmurt', 'ta': 'உட்முர்ட்', 'te': 'ఉడ్ముర్ట్', 'th': 'อุดมูร์ต', 'ti': 'ዩድሙርት', 'tk': 'udmurt dili', 'to': 'lea fakaʻutimuliti', 'tr': 'Udmurtça', 'ug': 'ئۇدمۇرتچە', 'uk': 'удмуртська', 'ur': 'ادمورت', 'uz': 'udmurt', 'uz-Latn': 'udmurt', 'vec': 'udmurto', 'vi': 'Tiếng Udmurt', 'xnr': 'उदमुर्त', 'yo': 'Èdè Udmuti', 'yrl': 'udimurite', 'yue': '沃蒂艾克文', 'yue-Hans': '沃蒂艾克文', 'yue-Hant': '沃蒂艾克文', 'zh': '乌德穆尔特语', 'zh-Hans': '乌德穆尔特语', 'zh-Hant': '烏德穆爾特文', 'zu': 'isi-Udmurt'},
    'udu': {'en': 'Uduk'},
    'ues': {'en': 'Kioko'},
    'ufi': {'en': 'Ufim'},
    'ug': {'af': 'Uighur', 'ak': 'Yugaa Kasa', 'am': 'ኡይግሁር', 'ar': 'الأويغورية', 'as': 'উইঘুৰ', 'ast': 'uigur', 'az': 'uyğur', 'az-Arab': 'اۇیغورجا', 'az-Cyrl': 'ујғур', 'az-Latn': 'uyğur', 'bal-Latn': 'Yughor', 'be': 'уйгурская', 'be-tarask': 'уйгурская', 'bg': 'уйгурски', 'bgn': 'اویغوری', 'bn': 'উইঘুর', 'br': 'ouigoureg', 'brx': 'उईघुर', 'bs': 'ujgurski', 'bs-Cyrl': 'ујгурски', 'bs-Latn': 'ujgurski', 'ca': 'uigur', 'ca-ES-valencia': 'uigur', 'ccp': '𑄃𑄪𑄃𑄨𑄊𑄪𑄢𑄴', 'ce': 'уйгурийн', 'chr': 'ᏫᎦ', 'ckb': 'ئۆیخۆری', 'cs': 'ujgurština', 'cy': 'Uighur', 'da': 'uygurisk', 'de': 'Uigurisch', 'dsb': 'ujguršćina', 'dz': 'ཝི་གུར་ཁ', 'ee': 'uighurgbe', 'el': 'Ουιγουρικά', 'el-polyton': 'Ουιγουρικά', 'en': 'Uyghur', 'eo': 'ujgura', 'es': 'uigur', 'et': 'uiguuri', 'eu': 'uigurrera', 'fa': 'اویغوری', 'ff-Adlm': '𞤓𞥅𞤴𞤺𞤵𞥅𞤪𞤫', 'fi': 'uiguuri', 'fil': 'Uyghur', 'fo': 'uyghur', 'fr': 'ouïghour', 'fur': 'uigur', 'fy': 'Oeigoers', 'ga': 'Uigiúiris', 'gd': 'Ùigiurais', 'gl': 'uigur', 'gsw': 'Uigurisch', 'gu': 'ઉઇગુર', 'ha': 'Ugiranci', 'he': 'אויגורית', 'hi': 'उइगर', 'hi-Latn': 'Uighur', 'hr': 'ujgurski', 'hsb': 'ujguršćina', 'ht': 'ouïghour', 'hu': 'ujgur', 'hy': 'ույղուրերեն', 'ia': 'uighur', 'id': 'Uyghur', 'ig': 'Uyghur', 'is': 'úígúr', 'it': 'uiguro', 'ja': 'ウイグル語', 'jv': 'Uighur', 'ka': 'უიღურული', 'kea': 'uigur', 'kgp': 'ujgur', 'kk': 'ұйғыр тілі', 'kk-Cyrl': 'ұйғыр тілі', 'km': 'អ៊ុយហ្គឺរ', 'kn': 'ಉಯಿಘರ್', 'ko': '위구르어', 'kok': 'उयघूर', 'kok-Deva': 'उयघूर', 'ksh': 'Uj’juhresch', 'ku': 'oygurî', 'ky': 'уйгурча', 'lb': 'Uiguresch', 'lkt': 'Uyghur Iyápi', 'lld': 'uigurich', 'lo': 'ອຸຍເຄີ', 'lrc': 'ئویغوٙر', 'lt': 'uigūrų', 'lv': 'uiguru', 'mai': 'उइगर', 'mi': 'Wīkura', 'mk': 'ујгурски', 'ml': 'ഉയ്ഘുർ', 'mn': 'уйгур', 'mr': 'उइगुर', 'ms': 'Uyghur', 'mt': 'Uyghur', 'my': 'ဝီဂါ', 'mzn': 'ئوغوری', 'nb': 'uigurisk', 'ne': 'उइघुर', 'nl': 'Oeigoers', 'nn': 'uigurisk', 'no': 'uigurisk', 'nqo': 'ߥߌߜ߭ߎ߯ߙߎߞߊ߲', 'or': 'ୟୁଘୁର୍', 'pa': 'ਉਇਗੁਰ', 'pa-Guru': 'ਉਇਗੁਰ', 'pcm': 'Wiúgọ', 'pl': 'ujgurski', 'ps': 'اويغوري', 'pt': 'uigur', 'qu': 'Uigur Simi', 'rm': 'uiguric', 'ro': 'uigură', 'ru': 'уйгурский', 'rw': 'Ikiwiguri', 'sah': 'Уйгуурдуу', 'sc': 'uiguru', 'sd': 'يوغور', 'sd-Arab': 'يوغور', 'si': 'උයිගර්', 'sk': 'ujgurčina', 'sl': 'ujgurščina', 'smn': 'uigurkielâ', 'so': 'Uighur', 'sq': 'ujgurisht', 'sr': 'ујгурски', 'sr-Cyrl': 'ујгурски', 'sr-Latn': 'ujgurski', 'sv': 'uiguriska', 'sw': 'Kiuyghur', 'ta': 'உய்குர்', 'te': 'ఉయ్\u200cఘర్', 'tg': 'ӯйғурӣ', 'th': 'อุยกูร์', 'ti': 'ኡይጉር', 'tk': 'uýgur dili', 'to': 'lea fakaʻuikūli', 'tr': 'Uygurca', 'tt': 'уйгыр', 'ug': 'ئۇيغۇرچە', 'uk': 'уйгурська', 'ur': 'یوئگہر', 'uz': 'uyg‘ur', 'uz-Cyrl': 'уйғурча', 'uz-Latn': 'uyg‘ur', 'vec': 'uigur', 'vi': 'Tiếng Uyghur', 'wae': 'Uiguriš', 'wo': 'Uygur', 'xnr': 'उइगर', 'yo': 'Yúgọ̀', 'yrl': 'wiguri', 'yue': '維吾爾文', 'yue-Hans': '维吾尔文', 'yue-Hant': '維吾爾文', 'zh': '维吾尔语', 'zh-Hans': '维吾尔语', 'zh-Hant': '維吾爾文', 'zu': 'isi-Uighur'},
    'uga': {'ar': 'اليجاريتيك', 'ast': 'ugaríticu', 'az': 'uqarit', 'az-Latn': 'uqarit', 'bg': 'угаритски', 'bn': 'উগারিটিক', 'br': 'ougariteg', 'brx': 'उगारिती', 'bs': 'ugaritski', 'bs-Cyrl': 'угаритски', 'bs-Latn': 'ugaritski', 'ca': 'ugarític', 'ca-ES-valencia': 'ugarític', 'ccp': '𑄃𑄪𑄉𑄢𑄨𑄑𑄨𑄇𑄴', 'cs': 'ugaritština', 'cy': 'Wgariteg', 'da': 'ugaristisk', 'de': 'Ugaritisch', 'el': 'Ουγκαριτικά', 'el-polyton': 'Ουγκαριτικά', 'en': 'Ugaritic', 'es': 'ugarítico', 'et': 'ugariti', 'fa': 'اوگاریتی', 'fi': 'ugarit', 'fr': 'ougaritique', 'fy': 'Oegaritysk', 'gsw': 'Ugaritisch', 'gu': 'યુગેરિટિક', 'he': 'אוגריתית', 'hi': 'युगैरिटिक', 'hi-Latn': 'Ugaritic', 'hr': 'ugaritski', 'ht': 'ougaritique', 'hu': 'ugariti', 'hy': 'ուգարիտերեն', 'id': 'Ugarit', 'is': 'úgarítíska', 'it': 'ugaritico', 'ja': 'ウガリト語', 'ka': 'უგარითული', 'kgp': 'ugaritiko', 'kn': 'ಉಗಾರಿಟಿಕ್', 'ko': '유가리틱어', 'ks': 'اُگارتِک', 'ks-Arab': 'اُگارتِک', 'lb': 'Ugaritesch', 'lo': 'ຢູກາລິກ', 'lt': 'ugaritų', 'lv': 'ugaritiešu', 'mk': 'угаритски', 'ml': 'ഉഗറിട്ടിക്', 'mr': 'युगॅरिटिक', 'mt': 'Ugaritiku', 'nb': 'ugaritisk', 'nl': 'Oegaritisch', 'nn': 'ugaritisk', 'no': 'ugaritisk', 'or': 'ୟୁଗୋରଟିକ୍', 'pl': 'ugarycki', 'pt': 'ugarítico', 'rm': 'ugaritic', 'ro': 'ugaritică', 'ru': 'угаритский', 'sk': 'ugaritčina', 'sl': 'ugaritski jezik', 'sr': 'угаритски', 'sr-Cyrl': 'угаритски', 'sr-Latn': 'ugaritski', 'sv': 'ugaritiska', 'ta': 'உகாரிடிக்', 'te': 'ఉగారిటిక్', 'th': 'ยูการิต', 'to': 'lea fakaʻūkaliti', 'tr': 'Ugarit dili', 'ug': 'ئۇگارىتىكچە', 'uk': 'угаритська', 'vi': 'Tiếng Ugaritic', 'yrl': 'ugarítiku', 'yue': '烏加列文', 'yue-Hans': '乌加列文', 'yue-Hant': '烏加列文', 'zh': '乌加里特语', 'zh-Hans': '乌加里特语', 'zh-Hant': '烏加列文'},
    'ugb': {'en': 'Kuku-Ugbanh'},
    'uge': {'en': 'Ughele'},
    'ugn': {'en': 'Ugandan Sign Language'},
    'ugo': {'en': 'Ugong'},
    'ugy': {'en': 'Uruguayan Sign Language'},
    'uha': {'en': 'Uhami'},
    'uhn': {'en': 'Damal'},
    'uis': {'en': 'Uisai'},
    'uiv': {'en': 'Iyive'},
    'uji': {'en': 'Tanjijili'},
    'uk': {'af': 'Oekraïens', 'agq': 'Ùkɛlɛnìa', 'ak': 'Ukren kasa', 'am': 'ዩክሬንኛ', 'ar': 'الأوكرانية', 'as': 'ইউক্ৰেইনীয়', 'asa': 'Kiukrania', 'ast': 'ucraín', 'az': 'ukrayna', 'az-Arab': 'اۇکراینجا', 'az-Cyrl': 'украјна', 'az-Latn': 'ukrayna', 'bal-Latn': 'Yukrini', 'bas': 'Hɔp u ukrǎnìà', 'be': 'украінская', 'be-tarask': 'украінская', 'bem': 'Ichi Ukraniani', 'bez': 'Hiukrania', 'bg': 'украински', 'bgn': 'اوکراینی', 'bm': 'ukɛrɛnikan', 'bn': 'ইউক্রেনীয়', 'br': 'ukraineg', 'brx': 'युक्रेनियान', 'bs': 'ukrajinski', 'bs-Cyrl': 'украјински', 'bs-Latn': 'ukrajinski', 'ca': 'ucraïnès', 'ca-ES-valencia': 'ucraïnès', 'ccp': '𑄃𑄨𑄃𑄪𑄇𑄳𑄢𑄬𑄚𑄩𑄠𑄧', 'ce': 'украинийн', 'cgg': 'Orukuraini', 'chr': 'ᏳᎧᎴᏂᎠᏂ', 'ckb': 'ئۆكراینی', 'cs': 'ukrajinština', 'cy': 'Wcreineg', 'da': 'ukrainsk', 'dav': 'Kiukrania', 'de': 'Ukrainisch', 'dje': 'Ukreen senni', 'dsb': 'ukrainšćina', 'dyo': 'ukrain', 'dz': 'ཡུ་ཀེ་རེ་ནི་ཡཱན་ཁ', 'ebu': 'Kĩukrania', 'ee': 'ukraingbe', 'el': 'Ουκρανικά', 'el-polyton': 'Οὐκρανικά', 'en': 'Ukrainian', 'eo': 'ukraina', 'es': 'ucraniano', 'et': 'ukraina', 'eu': 'ukrainera', 'ewo': 'ǹkɔ́bɔ ukelénia', 'fa': 'اوکراینی', 'ff': 'Ukereneere', 'ff-Adlm': '𞤒𞤵𞤳𞤪𞤫𞤲𞤭𞤴𞤢𞤲𞤪𞤫', 'ff-Latn': 'Ukereneere', 'fi': 'ukraina', 'fil': 'Ukranian', 'fo': 'ukrainskt', 'fr': 'ukrainien', 'fur': 'ucrain', 'fy': 'Oekraïens', 'ga': 'Úcráinis', 'gd': 'Ucràinis', 'gl': 'ucraíno', 'gsw': 'Ukrainisch', 'gu': 'યુક્રેનિયન', 'guz': 'Kiukrania', 'ha': 'Harshen Yukuren', 'he': 'אוקראינית', 'hi': 'यूक्रेनियाई', 'hi-Latn': 'Ukrainian', 'hr': 'ukrajinski', 'hsb': 'ukrainšćina', 'ht': 'ukrainien', 'hu': 'ukrán', 'hy': 'ուկրաիներեն', 'ia': 'ukrainiano', 'id': 'Ukraina', 'ig': 'Asụsụ Ukrain', 'is': 'úkraínska', 'it': 'ucraino', 'ja': 'ウクライナ語', 'jmc': 'Kyiukrania', 'jv': 'Ukraina', 'ka': 'უკრაინული', 'kab': 'Tukranit', 'kam': 'Kiukrania', 'kde': 'Chiuklania', 'kea': 'ukranianu', 'kgp': 'ukranĩjỹnũ', 'khq': 'Ukreen senni', 'ki': 'Kiukrania', 'kk': 'украин тілі', 'kk-Cyrl': 'украин тілі', 'kln': 'kutitab Ukraine', 'km': 'អ៊ុយក្រែន', 'kn': 'ಉಕ್ರೇನಿಯನ್', 'ko': '우크라이나어', 'kok': 'युक्रेनियन', 'kok-Deva': 'युक्रेनियन', 'ks': 'یوٗکرینیٲیی', 'ks-Arab': 'یوٗکرینیٲیی', 'ksb': 'Kiuklania', 'ksf': 'riukrɛ́n', 'ksh': 'Ukrainesch', 'ku': 'ukraynî', 'kxv': 'yukraniyati', 'kxv-Deva': 'यूक्रेनियाति', 'kxv-Latn': 'yukraniyati', 'kxv-Orya': 'ୟୁକ୍ରାନିଆତି', 'kxv-Telu': 'యుక్రెయనియన్', 'ky': 'украинче', 'lag': 'Kɨukɨranía', 'lb': 'Ukrainesch', 'lg': 'Luyukurayine', 'lkt': 'Ukrain Iyápi', 'lld': 'ucrainich', 'ln': 'likrɛni', 'lo': 'ຢູເຄຣນຽນ', 'lrc': 'ئوکراینی', 'lt': 'ukrainiečių', 'lu': 'Nkrani', 'luo': 'Kiukrania', 'luy': 'Kiukrania', 'lv': 'ukraiņu', 'mai': 'यूक्रेनी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lkrania', 'mer': 'Kĩukirĩni', 'mfe': 'ikrenien', 'mg': 'Okrainianina', 'mgh': 'Iukran', 'mi': 'Ukareinga', 'mk': 'украински', 'ml': 'ഉക്രേനിയൻ', 'mn': 'украин', 'mr': 'युक्रेनियन', 'ms': 'Ukraine', 'mt': 'Ukren', 'mua': 'Ukrainia', 'my': 'ယူကရိန်း', 'mzn': 'اوکراینی', 'naq': 'Ukrainiaǁî gowab', 'nb': 'ukrainsk', 'nd': 'isi-Ukrain', 'ne': 'युक्रेनी', 'nl': 'Oekraïens', 'nmg': 'Kiɛl b’ukrɛ̄nien', 'nn': 'ukrainsk', 'no': 'ukrainsk', 'nqo': 'ߎߞߌߙߍߣߌߞߊ߲', 'nus': 'Thok ukeraanini', 'nyn': 'Orukuraini', 'om': 'Afaan Ukreenii', 'or': 'ୟୁକ୍ରେନିଆନ୍', 'pa': 'ਯੂਕਰੇਨੀਆਈ', 'pa-Guru': 'ਯੂਕਰੇਨੀਆਈ', 'pcm': 'Yukrénia', 'pl': 'ukraiński', 'ps': 'اوکرايني', 'pt': 'ucraniano', 'qu': 'Ucraniano Simi', 'rm': 'ucranais', 'rn': 'Ikinyayukereni', 'ro': 'ucraineană', 'rof': 'Kiukrania', 'ru': 'украинский', 'rw': 'Ikinyayukereni', 'rwk': 'Kyiukrania', 'sah': 'Украйыыньыстыы', 'saq': 'Kiukrania', 'sbp': 'Ishiyukilaniya', 'sc': 'ucrainu', 'sd': 'يوڪراني', 'sd-Arab': 'يوڪراني', 'se': 'ukrainagiella', 'seh': 'ucraniano', 'ses': 'Ukreen senni', 'sg': 'Ukrêni', 'shi': 'ⵜⵓⴽⵔⴰⵏⵉⵜ', 'shi-Latn': 'Tukranit', 'shi-Tfng': 'ⵜⵓⴽⵔⴰⵏⵉⵜ', 'si': 'යුක්රේනියානු', 'sk': 'ukrajinčina', 'sl': 'ukrajinščina', 'smn': 'ukrainakielâ', 'sn': 'chiUkrenia', 'so': 'Yukreeniyaan', 'sq': 'ukrainisht', 'sr': 'украјински', 'sr-Cyrl': 'украјински', 'sr-Latn': 'ukrajinski', 'sv': 'ukrainska', 'sw': 'Kiukraini', 'syr': 'ܐܘܟܪܐܝܢܐܝܬ', 'ta': 'உக்ரைனியன்', 'te': 'ఉక్రెయినియన్', 'teo': 'Kiukrania', 'tg': 'украинӣ', 'th': 'ยูเครน', 'ti': 'ዩክረይንኛ', 'tk': 'ukrain dili', 'to': 'lea fakaʻūkalaʻine', 'tr': 'Ukraynaca', 'tt': 'украин', 'twq': 'Ukreen senni', 'tzm': 'Tukranit', 'ug': 'ئۇكرائىنچە', 'uk': 'українська', 'ur': 'یوکرینیائی', 'uz': 'ukrain', 'uz-Cyrl': 'украинча', 'uz-Latn': 'ukrain', 'vai': 'ꖳꖴꔓꕇꘂꘋ', 'vai-Latn': 'Yukureniyɛŋ', 'vai-Vaii': 'ꖳꖴꔓꕇꘂꘋ', 'vec': 'ucrain', 'vi': 'Tiếng Ukraina', 'vun': 'Kyiukrania', 'wae': 'Ukrainiš', 'wo': 'Ikreniye', 'xnr': 'यूक्रेनियाई', 'xog': 'Oluyukurayine', 'yav': 'nukeleniɛ́ŋɛ', 'yi': 'אוקראַאיניש', 'yo': 'Èdè Ukania', 'yrl': 'ukaranianu', 'yue': '烏克蘭文', 'yue-Hans': '乌克兰文', 'yue-Hant': '烏克蘭文', 'zgh': 'ⵜⵓⴽⵔⴰⵏⵉⵜ', 'zh': '乌克兰语', 'zh-Hans': '乌克兰语', 'zh-Hant': '烏克蘭文', 'zu': 'isi-Ukrainian'},
    'uka': {'en': 'Kaburi'},
    'ukg': {'en': 'Ukuriguma'},
    'ukh': {'en': 'Ukhwejo'},
    'uki': {'en': 'Kui (India)'},
    'ukk': {'en': 'Muak Sa-aak'},
    'ukl': {'en': 'Ukrainian Sign Language'},
    'ukp': {'en': 'Ukpe-Bayobiri'},
    'ukq': {'en': 'Ukwa'},
    'uks': {'en': 'Urubú-Kaapor Sign Language'},
    'uku': {'en': 'Ukue'},
    'ukv': {'en': 'Kuku'},
    'ukw': {'en': 'Ukwuani-Aboh-Ndoni'},
    'uky': {'en': 'Kuuk-Yak'},
    'ula': {'en': 'Fungwa'},
    'ulb': {'en': 'Ulukwumi'},
    'ulc': {'en': 'Ulch'},
    'ule': {'en': 'Lule'},
    'ulf': {'en': 'Usku'},
    'uli': {'en': 'Ulithian'},
    'ulk': {'en': 'Meriam Mir'},
    'ull': {'en': 'Ullatan'},
    'ulm': {'en': "Ulumanda'"},
    'uln': {'en': 'Unserdeutsch'},
    'ulu': {'en': "Uma' Lung"},
    'ulw': {'en': 'Ulwa'},
    'uma': {'en': 'Umatilla'},
    'umb': {'af': 'Umbundu', 'am': 'ኡምቡንዱ', 'ar': 'الأمبندو', 'as': 'উম্বুন্দু', 'ast': 'umbundu', 'az': 'umbundu', 'az-Cyrl': 'умбунду', 'az-Latn': 'umbundu', 'be': 'умбунду', 'be-tarask': 'умбунду', 'bg': 'умбунду', 'bgn': 'امبونڈویی', 'bn': 'উম্বুন্দু', 'br': 'umbundu', 'brx': 'उम्बुन्डु', 'bs': 'umbundu', 'bs-Cyrl': 'умбунду', 'bs-Latn': 'umbundu', 'ca': 'umbundu', 'ca-ES-valencia': 'umbundu', 'ccp': '𑄃𑄪𑄟𑄴𑄝𑄪𑄚𑄴𑄘𑄪', 'ce': 'умбунду', 'chr': 'ᎤᎹᏊᏅᏚ', 'ckb': 'ئومبووندوو', 'cs': 'umbundu', 'cy': 'Umbundu', 'da': 'umbundu', 'de': 'Umbundu', 'dsb': 'umbundušćina', 'el': 'Ουμπούντου', 'el-polyton': 'Ουμπούντου', 'en': 'Umbundu', 'es': 'umbundu', 'et': 'umbundu', 'eu': 'umbundua', 'fa': 'امبوندویی', 'ff-Adlm': '𞤓𞤥𞤦𞤵𞤲𞤣𞤵𞥅𞤪𞤫', 'fi': 'mbundu', 'fil': 'Umbundu', 'fo': 'umbundu', 'fr': 'umbundu', 'fy': 'Umbundu', 'ga': 'Umbundais', 'gd': 'Umbundu', 'gl': 'umbundu', 'gsw': 'Mbundu-Schpraach', 'gu': 'ઉમ્બુન્ડૂ', 'ha': 'Umbundu', 'he': 'אומבונדו', 'hi': 'उम्बुन्डु', 'hi-Latn': 'Umbundu', 'hr': 'umbundu', 'hsb': 'umbundušćina', 'ht': 'umbundu', 'hu': 'umbundu', 'hy': 'ումբունդու', 'ia': 'umbundu', 'id': 'Umbundu', 'ig': 'Umbundu', 'is': 'úmbúndú', 'it': 'mbundu', 'ja': 'ムブンドゥ語', 'jv': 'Umbundu', 'ka': 'უმბუნდუ', 'kgp': 'ũmunu', 'kk': 'умбунду тілі', 'kk-Cyrl': 'умбунду тілі', 'km': 'អាម់ប៊ុនឌូ', 'kn': 'ಉಂಬುಂಡು', 'ko': '움분두어', 'kok': 'उंबुंडु', 'kok-Deva': 'उंबुंडु', 'ks': 'یُمبُندوٗ', 'ks-Arab': 'یُمبُندوٗ', 'ksh': 'Umbundesch', 'ku': 'umbunduyî', 'ky': 'умбундуча', 'lb': 'Mbundu-Sprooch', 'lo': 'ອຳບັນດູ', 'lt': 'umbundu', 'lv': 'umbundu', 'mai': 'उमबुंदू', 'mi': 'Ūpunu', 'mk': 'умбунду', 'ml': 'ഉംബുന്ദു', 'mn': 'умбунду', 'mr': 'उम्बुन्डु', 'ms': 'Umbundu', 'mt': 'Umbundu', 'my': 'အူဘန်ဒူ', 'nb': 'umbundu', 'ne': 'उम्बुन्डी', 'nl': 'Umbundu', 'nn': 'umbundu', 'no': 'umbundu', 'or': 'ଉମ୍ବୁଣ୍ଡୁ', 'pa': 'ਉਮਬੁੰਡੂ', 'pa-Guru': 'ਉਮਬੁੰਡੂ', 'pcm': 'Umbúndu Lángwej', 'pl': 'umbundu', 'ps': 'امبوندو', 'pt': 'umbundu', 'qu': 'Umbundu Simi', 'rm': 'mbundu', 'ro': 'umbundu', 'ru': 'умбунду', 'sc': 'umbundu', 'sd': 'اومبنڊو', 'sd-Arab': 'اومبنڊو', 'si': 'උබුන්ඩු', 'sk': 'umbundu', 'sl': 'umbundščina', 'smn': 'umbundu', 'so': 'Umbundu', 'sq': 'umbunduisht', 'sr': 'умбунду', 'sr-Cyrl': 'умбунду', 'sr-Latn': 'umbundu', 'sv': 'umbundu', 'sw': 'Umbundu', 'ta': 'அம்பொண்டு', 'te': 'ఉమ్బుండు', 'th': 'อุมบุนดู', 'ti': 'ኣምቡንዱ', 'tk': 'umbundu dili', 'to': 'lea fakaʻumipūnitu', 'tr': 'Umbundu', 'ug': 'ئۇمبۇندۇچە', 'uk': 'умбунду', 'ur': 'اومبوندو', 'uz': 'umbundu', 'uz-Latn': 'umbundu', 'vec': 'unbundu', 'vi': 'Tiếng Umbundu', 'xnr': 'उम्बुन्डु', 'yo': 'Èdè Umbundu', 'yrl': 'ũbũdu', 'yue': '姆本杜文', 'yue-Hans': '姆本杜文', 'yue-Hant': '姆本杜文', 'zh': '翁本杜语', 'zh-Hans': '翁本杜语', 'zh-Hant': '姆本杜文', 'zu': 'isi-Umbundu'},
    'umc': {'en': 'Marrucinian'},
    'umd': {'en': 'Umbindhamu'},
    'umg': {'en': 'Morrobalama'},
    'umi': {'en': 'Ukit'},
    'umm': {'en': 'Umon'},
    'umn': {'en': 'Makyan Naga'},
    'umo': {'en': 'Umotína'},
    'ump': {'en': 'Umpila'},
    'umr': {'en': 'Umbugarla'},
    'ums': {'en': 'Pendau'},
    'umu': {'en': 'Munsee'},
    'una': {'en': 'North Watut'},
    'und': {'af': 'Onbekende taal', 'ak': 'kasa a yɛnnim', 'am': 'ያልታወቀ ቋንቋ', 'ar': 'لغة غير معروفة', 'as': 'অজ্ঞাত ভাষা', 'ast': 'llingua desconocida', 'az': 'naməlum dil', 'az-Arab': 'تانینمایان دیل', 'az-Cyrl': 'намәлум дил', 'az-Latn': 'naməlum dil', 'bal-Latn': 'Nagisshetagén zobán', 'be': 'невядомая мова', 'be-tarask': 'невядомая мова', 'bg': 'непознат език', 'bgn': 'نازانتین زبان', 'blo': 'gɩkrǝ ŋgɩɖee kʊyɔʊ ʊ mana ma', 'bn': 'অজানা ভাষা', 'br': 'yezh dianav', 'brx': 'मिथियै राव', 'bs': 'nepoznati jezik', 'bs-Cyrl': 'непознати језик', 'bs-Latn': 'nepoznati jezik', 'ca': 'idioma desconegut', 'ca-ES-valencia': 'idioma desconegut', 'ccp': '𑄦𑄧𑄝𑄧𑄢𑄴 𑄚𑄧𑄛𑄬𑄠𑄬 𑄞𑄌𑄴', 'ce': 'боьвзуш боцу мотт', 'ceb': 'Wala Mailhing Pinulongan', 'chr': 'ᏄᏬᎵᏍᏛᎾ ᎦᏬᏂᎯᏍᏗ', 'ckb': 'زمانی نەناسراو', 'cs': 'neznámý jazyk', 'csw': 'ᓇᒨᓇ ᐃᐦᑕᐟᐘᐣ ᐊᔭᒧᐏᐣ', 'cv': 'паллӑ мар чӗлхе', 'cy': 'Iaith anhysbys', 'da': 'ukendt sprog', 'de': 'Unbekannte Sprache', 'doi': 'अनजांती भाशा', 'dsb': 'njeznata rěc', 'dz': 'ཁ་ངོ་མ་ཤེསཔ', 'ee': 'gbegbɔgblɔ manya', 'el': 'Άγνωστη γλώσσα', 'el-polyton': 'Άγνωστη γλώσσα', 'en': 'Unknown language', 'eo': 'nekonata lingvo', 'es': 'lengua desconocida', 'et': 'määramata keel', 'eu': 'hizkuntza ezezaguna', 'fa': 'زبان نامشخص', 'ff-Adlm': '𞤍𞤫𞤲𞤺𞤢𞤤 𞤢𞤧-𞤢𞤲𞤣𞤢𞥄𞤲𞤺𞤢𞤤', 'fi': 'tuntematon kieli', 'fil': 'Hindi Kilalang Wika', 'fo': 'ókent mál', 'fr': 'langue indéterminée', 'fur': 'indeterminade', 'fy': 'Onbekende taal', 'ga': 'Teanga anaithnid', 'gd': 'Cànan neo-aithnichte', 'gl': 'lingua descoñecida', 'gsw': 'Unbeschtimmti Schpraach', 'gu': 'અજ્ઞાત ભાષા', 'ha': 'Harshen da ba a sani ba', 'haw': 'ʻIke ʻole ‘ia a kūpono ʻole paha ka ʻōlelo', 'he': 'שפה לא ידועה', 'hi': 'अज्ञात भाषा', 'hi-Latn': 'Unknown language', 'hr': 'nepoznati jezik', 'hsb': 'njeznata rěč', 'ht': 'langue indéterminée', 'hu': 'ismeretlen nyelv', 'hy': 'անհայտ լեզու', 'ia': 'lingua incognite', 'id': 'Bahasa Tidak Dikenal', 'ie': 'ínconosset lingue', 'ig': 'Asụsụ a na-amaghị', 'ii': 'ꅉꀋꌠꅇꂷ', 'is': 'óþekkt tungumál', 'it': 'lingua imprecisata', 'ja': '言語不明', 'jgo': 'cú-pʉɔ yi pɛ́ ká kɛ́ jí', 'jv': 'Basa Ora Dikenali', 'ka': 'უცნობი ენა', 'kea': 'língua diskonxedu', 'kgp': 'vẽnhvĩ ki kagtĩg', 'kk': 'белгісіз тіл', 'kk-Cyrl': 'белгісіз тіл', 'km': 'ភាសាមិនស្គាល់', 'kn': 'ಅಪರಿಚಿತ ಭಾಷೆ', 'ko': '알 수 없는 언어', 'kok': 'अज्ञात भास', 'kok-Deva': 'अज्ञात भास', 'ks': 'اَنزٲنؠ یا نَہ لَگہٕہار زبان', 'ks-Arab': 'اَنزٲنؠ یا نَہ لَگہٕہار زبان', 'ks-Deva': 'नामोलुम ज़बान', 'ksh': '-onbikannte-Schprooch-', 'ku': 'zimanê nenas', 'ky': 'белгисиз тилде', 'lb': 'Onbestëmmt Sprooch', 'lkt': 'Tukté iyápi tȟaŋíŋ šni', 'lld': 'lingaz nia conesciü', 'lo': 'ບໍ່ສາມາດລະບຸພາສາ', 'lrc': 'زوٙن نادیار', 'lt': 'nežinoma kalba', 'lv': 'nezināma valoda', 'mai': 'अज्ञात भाषा', 'mgo': 'ngam tisɔʼ', 'mi': 'Reo Tē Mōhiotia', 'mk': 'непознат јазик', 'ml': 'അജ്ഞാത ഭാഷ', 'mn': 'үл мэдэгдэх хэл', 'mni': 'মশকখংদবা লোল', 'mni-Beng': 'মশকখংদবা লোল', 'mr': 'अज्ञात भाषा', 'ms': 'Bahasa Tidak Diketahui', 'mt': 'Lingwa Mhix Magħrufa', 'my': 'မသိသော ဘာသာ', 'mzn': 'نشناسی\u200cیه زوون', 'nb': 'ukjent språk', 'ne': 'अज्ञात भाषा', 'nl': 'onbekende taal', 'nn': 'ukjent språk', 'no': 'ukjent språk', 'nqo': 'ߞߊ߲߫ ߘߊ߲߬ߠߊ߬ߕߍ߰ߓߊߟߌ', 'om': 'Afaan hin beekamne', 'or': 'ଅଜଣା ଭାଷା', 'os': 'нӕзонгӕ ӕвзаг', 'pa': 'ਅਣਪਛਾਤੀ ਬੋਲੀ', 'pa-Guru': 'ਅਣਪਛਾਤੀ ਬੋਲੀ', 'pcm': 'Lángwej wé nóbọ́di sabi', 'pl': 'nieznany język', 'ps': 'نامعلومه ژبه', 'pt': 'idioma desconhecido', 'qu': 'Mana Riqsisqa Simi', 'rm': 'lingua nunenconuschenta', 'ro': 'limbă necunoscută', 'ru': 'неизвестный язык', 'sa': 'अज्ञात भाषा:', 'sat': 'ᱵᱟᱝ ᱩᱨᱩᱢ ᱯᱟᱹᱨᱥᱤ', 'sat-Olck': 'ᱵᱟᱝ ᱩᱨᱩᱢ ᱯᱟᱹᱨᱥᱤ', 'sc': 'limba disconnota', 'sd': 'اڻڄاتل ٻولي', 'sd-Arab': 'اڻڄاتل ٻولي', 'sd-Deva': 'अणजा॒तल भाषा', 'se': 'dovdameahttun giella', 'si': 'නොදන්නා භාෂාව', 'sk': 'neznámy jazyk', 'sl': 'neznan jezik', 'smn': 'tubdâmettumis kielâ', 'so': 'Af aan la aqoon ama aan sax ahayn', 'sq': 'E panjohur', 'sr': 'непознат језик', 'sr-Cyrl': 'непознат језик', 'sr-Latn': 'nepoznat jezik', 'su': 'Basa teu dikenal', 'su-Latn': 'Basa teu dikenal', 'sv': 'obestämt språk', 'sw': 'Lugha isiyojulikana', 'syr': 'ܠܫܢܐ ܠܐ ܝܕܝܥܐ', 'ta': 'அறியப்படாத மொழி', 'te': 'తెలియని భాష', 'tg': 'забони номаълум', 'th': 'ภาษาที่ไม่รู้จัก', 'ti': 'ዘይተፈልጠ ቋንቋ', 'tk': 'näbelli dil', 'to': 'lea taʻeʻiloa', 'tpi': 'Tok ples i no stap', 'tr': 'Bilinmeyen Dil', 'tt': 'билгесез тел', 'ug': 'يوچۇن تىل', 'uk': 'невідома мова', 'ur': 'نامعلوم زبان', 'uz': 'noma’lum til', 'uz-Cyrl': 'номаълум тил', 'uz-Latn': 'noma’lum til', 'vec': 'lengua desconoscùa', 'vi': 'Ngôn ngữ không xác định', 'wae': 'Unbekannti Schprač', 'wo': 'Làkk wuñ xamul', 'xh': 'Unknown language', 'xnr': 'अणजाण बोली', 'yi': 'אומבאַוואוסטע שפּראַך', 'yo': 'Èdè àìmọ̀', 'yrl': 'ũba uyukuau nheẽga', 'yue': '未知語言', 'yue-Hans': '未知语言', 'yue-Hant': '未知語言', 'zh': '未知语言', 'zh-Hans': '未知语言', 'zh-Hant': '未知語言', 'zu': 'ulimi olungaziwa'},
    'une': {'en': 'Uneme'},
    'ung': {'en': 'Ngarinyin'},
    'uni': {'en': 'Uni'},
    'unk': {'en': 'Enawené-Nawé'},
    'unm': {'en': 'Unami'},
    'unn': {'en': 'Kurnai'},
    'unp': {'en': 'Worora'},
    'unr': {'en': 'Mundari'},
    'unu': {'en': 'Unubahe'},
    'unx': {'en': 'Munda'},
    'unz': {'en': 'Unde Kaili'},
    'uok': {'en': 'Uokha'},
    'upi': {'en': 'Umeda'},
    'upv': {'en': 'Uripiv-Wala-Rano-Atchin'},
    'ur': {'af': 'Oerdoe', 'agq': 'Uudùw', 'ak': 'Urdu kasa', 'am': 'ኡርዱኛ', 'ar': 'الأوردية', 'as': 'উৰ্দু', 'asa': 'Kiurdu', 'ast': 'urdu', 'az': 'urdu', 'az-Arab': 'اوْردوجا', 'az-Cyrl': 'урду', 'az-Latn': 'urdu', 'bal-Latn': 'Urdu', 'bas': 'Hɔp u urdù', 'be': 'урду', 'be-tarask': 'урду', 'bem': 'Ichi Urudu', 'bez': 'Hiurdu', 'bg': 'урду', 'bgn': 'اوردو', 'bm': 'urudukan', 'bn': 'উর্দু', 'br': 'ourdou', 'brx': 'ऊर्दु', 'bs': 'urdu', 'bs-Cyrl': 'урду', 'bs-Latn': 'urdu', 'ca': 'urdú', 'ca-ES-valencia': 'urdú', 'ccp': '𑄃𑄪𑄢𑄴𑄘𑄪', 'ce': 'урду', 'cgg': 'Oru-Urudu', 'chr': 'ᎤᎵᏚ', 'ckb': 'ئۆردوو', 'cs': 'urdština', 'cy': 'Wrdw', 'da': 'urdu', 'dav': 'Kiurdu', 'de': 'Urdu', 'dje': 'Urdu senni', 'dsb': 'urdušćina', 'dyo': 'urdu', 'dz': 'ཨུར་དུ་ཁ', 'ebu': 'Kĩurdu', 'ee': 'urdugbe', 'el': 'Ούρντου', 'el-polyton': 'Ούρντου', 'en': 'Urdu', 'eo': 'urduo', 'es': 'urdu', 'et': 'urdu', 'eu': 'urdua', 'ewo': 'ǹkɔ́bɔ urudú', 'fa': 'اردو', 'ff': 'Urdu', 'ff-Adlm': '𞤓𞤪𞤣𞤵𞥅𞤪𞤫', 'ff-Latn': 'Urdu', 'fi': 'urdu', 'fil': 'Urdu', 'fo': 'urdu', 'fr': 'ourdou', 'fur': 'urdu', 'fy': 'Urdu', 'ga': 'Urdúis', 'gd': 'Ùrdu', 'gl': 'urdú', 'gsw': 'Urdu', 'gu': 'ઉર્દૂ', 'guz': 'Kiurdu', 'ha': 'Urdanci', 'he': 'אורדו', 'hi': 'उर्दू', 'hi-Latn': 'Urdu', 'hr': 'urdski', 'hsb': 'urdušćina', 'ht': 'ourdou', 'hu': 'urdu', 'hy': 'ուրդու', 'ia': 'urdu', 'id': 'Urdu', 'ig': 'Urdu', 'is': 'úrdú', 'it': 'urdu', 'ja': 'ウルドゥー語', 'jmc': 'Kyiurdu', 'jv': 'Urdu', 'ka': 'ურდუ', 'kab': 'Turdut', 'kam': 'Kiurdu', 'kde': 'Chiuldu', 'kea': 'urdu', 'kgp': 'urnu', 'khq': 'Urdu senni', 'ki': 'Kiurdu', 'kk': 'урду тілі', 'kk-Cyrl': 'урду тілі', 'kln': 'kutitab Urdu', 'km': 'អ៊ូរឌូ', 'kn': 'ಉರ್ದು', 'ko': '우르두어', 'kok': 'उर्दू', 'kok-Deva': 'उर्दू', 'ks': 'اُردوٗ', 'ks-Arab': 'اُردوٗ', 'ksb': 'Kiuldu', 'ksf': 'riurdú', 'ksh': 'Urdu/Hindi', 'ku': 'urdûyî', 'kxv': 'urdu', 'kxv-Deva': 'उर्दू', 'kxv-Latn': 'urdu', 'kxv-Orya': 'ଉର୍ଦୁ', 'kxv-Telu': 'ఉర్దూ', 'ky': 'урдуча', 'lag': 'Kɨúrdu', 'lb': 'Urdu', 'lg': 'Lu-urudu', 'lkt': 'Urdu Iyápi', 'lld': 'urdu', 'ln': 'liurdu', 'lo': 'ອູຣດູ', 'lrc': 'ئوردوٙ', 'lt': 'urdų', 'lu': 'Urdu', 'luo': 'Kiurdu', 'luy': 'Kiurdu', 'lv': 'urdu', 'mai': 'उर्दू', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lurdu', 'mer': 'Kĩurdu', 'mfe': 'ourdou', 'mg': 'Ordò', 'mgh': 'Ihurdu', 'mi': 'Ūrutu', 'mk': 'урду', 'ml': 'ഉറുദു', 'mn': 'урду', 'mr': 'उर्दू', 'ms': 'Urdu', 'mt': 'Urdu', 'mua': 'Urdu', 'my': 'အူရ်ဒူ', 'mzn': 'اردو', 'naq': 'Urduǁî gowab', 'nb': 'urdu', 'nd': 'isi-Udu', 'ne': 'उर्दु', 'nl': 'Urdu', 'nmg': 'Kiɛl úrdu', 'nn': 'urdu', 'no': 'urdu', 'nqo': 'ߎߙߘߎߞߊ߲', 'nus': 'Thok udoni', 'nyn': 'Oru-Urudu', 'om': 'Afaan Urdu', 'or': 'ଉର୍ଦ୍ଦୁ', 'pa': 'ਉਰਦੂ', 'pa-Guru': 'ਉਰਦੂ', 'pcm': 'Úrdú', 'pl': 'urdu', 'ps': 'اردو', 'pt': 'urdu', 'qu': 'Urdu Simi', 'rm': 'urdu', 'rn': 'Inyeyurudu', 'ro': 'urdu', 'rof': 'Kiurdu', 'ru': 'урду', 'rw': 'Inyeyurudu', 'rwk': 'Kyiurdu', 'saq': 'Kiurdu', 'sbp': 'Ishiwuludi', 'sc': 'urdu', 'sd': 'اردو', 'sd-Arab': 'اردو', 'se': 'urdugiella', 'seh': 'urdu', 'ses': 'Urdu senni', 'sg': 'Ûrdu', 'shi': 'ⵜⵓⵔⴷⵓⵜ', 'shi-Latn': 'Turdut', 'shi-Tfng': 'ⵜⵓⵔⴷⵓⵜ', 'si': 'උර්දු', 'sk': 'urdčina', 'sl': 'urdujščina', 'smn': 'urdu', 'sn': 'chiUrdu', 'so': 'Urduu', 'sq': 'urduisht', 'sr': 'урду', 'sr-Cyrl': 'урду', 'sr-Latn': 'urdu', 'sv': 'urdu', 'sw': 'Kiurdu', 'syr': 'ܐܘܪܕܘܝܬ', 'ta': 'உருது', 'te': 'ఉర్దూ', 'teo': 'Kiurdu', 'tg': 'урду', 'th': 'อูรดู', 'ti': 'ኡርዱ', 'tk': 'urdu', 'to': 'lea fakaʻūtū', 'tr': 'Urduca', 'tt': 'урду', 'twq': 'Urdu senni', 'tzm': 'Turdut', 'ug': 'ئوردۇچە', 'uk': 'урду', 'ur': 'اردو', 'uz': 'urdu', 'uz-Cyrl': 'урду', 'uz-Latn': 'urdu', 'vai': 'ꖺꖦ', 'vai-Latn': 'Ɔdu', 'vai-Vaii': 'ꖺꖦ', 'vec': 'urdu', 'vi': 'Tiếng Urdu', 'vun': 'Kyiurdu', 'wae': 'Urdu', 'wo': 'Urdu', 'xnr': 'उर्दू', 'xog': 'Olu-urudu', 'yav': 'nulutú', 'yi': 'אורדו', 'yo': 'Èdè Udu', 'yrl': 'urdu', 'yue': '烏都文', 'yue-Hans': '乌都文', 'yue-Hant': '烏都文', 'zgh': 'ⵜⵓⵔⴷⵓⵜ', 'zh': '乌尔都语', 'zh-Hans': '乌尔都语', 'zh-Hant': '烏都文', 'zu': 'isi-Urdu'},
    'ura': {'en': 'Urarina'},
    'urb': {'en': 'Urubú-Kaapor'},
    'urc': {'en': 'Urningangg'},
    'ure': {'en': 'Uru'},
    'urf': {'en': 'Uradhi'},
    'urg': {'en': 'Urigina'},
    'urh': {'en': 'Urhobo'},
    'uri': {'en': 'Urim'},
    'urj': {'en': 'Uralic languages'},
    'urj-pro': {'en': 'Proto-Uralic'},
    'urk': {'en': "Urak Lawoi'"},
    'url': {'en': 'Urali'},
    'urm': {'en': 'Urapmin'},
    'urn': {'en': 'Uruangnirin'},
    'uro': {'en': 'Ura (Papua New Guinea)'},
    'urp': {'en': 'Uru-Pa-In'},
    'urr': {'en': 'Lehalurup'},
    'urt': {'en': 'Urat'},
    'uru': {'en': 'Urumi'},
    'urv': {'en': 'Uruava'},
    'urw': {'en': 'Sop'},
    'urx': {'en': 'Urimo'},
    'ury': {'en': 'Orya'},
    'urz': {'en': 'Uru-Eu-Wau-Wau'},
    'usa': {'en': 'Usarufa'},
    'ush': {'en': 'Ushojo'},
    'usi': {'en': 'Usui'},
    'usk': {'en': 'Usaghade'},
    'usp': {'en': 'Uspanteco'},
    'uss': {'en': 'us-Saare'},
    'usu': {'en': 'Uya'},
    'uta': {'en': 'Otank'},
    'ute': {'en': 'Ute-Southern Paiute'},
    'uth': {'en': 'ut-Hun'},
    'utp': {'en': 'Amba (Solomon Islands)'},
    'utr': {'en': 'Etulo'},
    'utu': {'en': 'Utu'},
    'uum': {'en': 'Urum'},
    'uun': {'en': 'Kulon-Pazeh'},
    'uur': {'en': 'Ura (Vanuatu)'},
    'uuu': {'en': 'U'},
    'uve': {'en': 'West Uvean'},
    'uvh': {'en': 'Uri'},
    'uvl': {'en': 'Lote'},
    'uwa': {'en': 'Kuku-Uwanh'},
    'uya': {'en': 'Doko-Uyanga'},
    'uz': {'af': 'Oesbekies', 'ak': 'Usbɛkistan Kasa', 'am': 'ኡዝቤክኛ', 'ar': 'الأوزبكية', 'as': 'উজবেক', 'ast': 'uzbequistanín', 'az': 'özbək', 'az-Arab': 'اؤزبکجه', 'az-Cyrl': 'өзбәк', 'az-Latn': 'özbək', 'bal-Latn': 'Ozbek', 'be': 'узбекская', 'be-tarask': 'узбекская', 'bg': 'узбекски', 'bgn': 'اوزبکی', 'bn': 'উজবেক', 'br': 'ouzbekeg', 'brx': 'उजबेक', 'bs': 'uzbečki', 'bs-Cyrl': 'узбечки', 'bs-Latn': 'uzbečki', 'ca': 'uzbek', 'ca-ES-valencia': 'uzbek', 'ccp': '𑄃𑄪𑄎𑄴𑄝𑄬𑄇𑄩𑄠𑄧', 'ce': 'узбекийн', 'chr': 'ᎤᏍᏇᎩ', 'ckb': 'ئوزبەکی', 'cs': 'uzbečtina', 'cy': 'Wsbeceg', 'da': 'usbekisk', 'de': 'Usbekisch', 'dsb': 'usbekšćina', 'dz': 'ཨུས་བེཀ་ཁ', 'ee': 'uzbekistangbe', 'el': 'Ουζμπεκικά', 'el-polyton': 'Ουζμπεκικά', 'en': 'Uzbek', 'eo': 'uzbeka', 'es': 'uzbeko', 'et': 'usbeki', 'eu': 'uzbekera', 'fa': 'ازبکی', 'ff-Adlm': '𞤓𞥅𞤧𞤦𞤫𞤳𞤪𞤫', 'fi': 'uzbekki', 'fil': 'Uzbek', 'fo': 'usbekiskt', 'fr': 'ouzbek', 'fur': 'uzbec', 'fy': 'Oezbeeks', 'ga': 'Úisbéiceastáinis', 'gd': 'Usbagais', 'gl': 'uzbeko', 'gsw': 'Usbekisch', 'gu': 'ઉઝ્બેક', 'ha': 'Uzbek', 'he': 'אוזבקית', 'hi': 'उज़्बेक', 'hi-Latn': 'Uzbek', 'hr': 'uzbečki', 'hsb': 'uzbekšćina', 'ht': 'ouzbek', 'hu': 'üzbég', 'hy': 'ուզբեկերեն', 'ia': 'uzbeko', 'id': 'Uzbek', 'ig': 'Uzbek', 'is': 'úsbekska', 'it': 'uzbeco', 'ja': 'ウズベク語', 'jv': 'Uzbek', 'ka': 'უზბეკური', 'kea': 'uzbeki', 'kgp': 'unhmeke', 'kk': 'өзбек тілі', 'kk-Cyrl': 'өзбек тілі', 'km': 'អ៊ូសបេគ', 'kn': 'ಉಜ್ಬೇಕ್', 'ko': '우즈베크어', 'kok': 'उझबेक', 'kok-Deva': 'उझबेक', 'ks': 'اُزبیک', 'ks-Arab': 'اُزبیک', 'ksh': 'Ußbehkesch', 'ku': 'ozbekî', 'kxv': 'ujbek', 'kxv-Deva': 'उज़्बेक', 'kxv-Latn': 'ujbek', 'kxv-Orya': 'ଉଜବେକ୍', 'kxv-Telu': 'ఉజ్బెక్', 'ky': 'өзбекче', 'lb': 'Usbekesch', 'lkt': 'Uzbek Iyápi', 'lld': 'uzbech', 'lo': 'ອຸສເບກ', 'lrc': 'ئوزبأکی', 'lt': 'uzbekų', 'lv': 'uzbeku', 'mai': 'उजबेक', 'mi': 'Ūpeke', 'mk': 'узбечки', 'ml': 'ഉസ്\u200cബെക്ക്', 'mn': 'узбек', 'mr': 'उझ्बेक', 'ms': 'Uzbekistan', 'mt': 'Uzbek', 'my': 'ဥဇဘတ်', 'mzn': 'ازبکی', 'nb': 'usbekisk', 'ne': 'उज्बेकी', 'nl': 'Oezbeeks', 'nn': 'usbekisk', 'no': 'usbekisk', 'nqo': 'ߎߖ߭ߑߓߋߞߌߞߊ߲', 'om': 'Afaan Uzbek', 'or': 'ଉଜବେକ୍', 'pa': 'ਉਜ਼ਬੇਕ', 'pa-Guru': 'ਉਜ਼ਬੇਕ', 'pcm': 'Úzbẹk', 'pl': 'uzbecki', 'ps': 'اوزبکي', 'pt': 'uzbeque', 'qu': 'Uzbeko Simi', 'rm': 'usbec', 'ro': 'uzbecă', 'ru': 'узбекский', 'rw': 'Inyeyuzubeki', 'sah': 'Үзбиэктии', 'sc': 'uzbecu', 'sd': 'ازبڪ', 'sd-Arab': 'ازبڪ', 'si': 'උස්බෙක්', 'sk': 'uzbečtina', 'sl': 'uzbeščina', 'smn': 'uzbekkielâ', 'so': 'Usbakis', 'sq': 'uzbekisht', 'sr': 'узбечки', 'sr-Cyrl': 'узбечки', 'sr-Latn': 'uzbečki', 'sv': 'uzbekiska', 'sw': 'Kiuzbeki', 'ta': 'உஸ்பெக்', 'te': 'ఉజ్బెక్', 'tg': 'ӯзбекӣ', 'th': 'อุซเบก', 'ti': 'ኡዝበክኛ', 'tk': 'özbek dili', 'to': 'lea fakaʻusipeki', 'tok': 'toki Opeki', 'tr': 'Özbekçe', 'tt': 'үзбәк', 'ug': 'ئۆزبېكچە', 'uk': 'узбецька', 'ur': 'ازبیک', 'uz': 'o‘zbek', 'uz-Arab': 'اوزبیک', 'uz-Cyrl': 'ўзбекча', 'uz-Latn': 'o‘zbek', 'vec': 'uzbego', 'vi': 'Tiếng Uzbek', 'wae': 'Usbekiš', 'wo': 'Usbek', 'xnr': 'उज़्बेक', 'yi': 'אוזבעקיש', 'yo': 'Èdè Uzbek', 'yrl': 'usibeki', 'yue': '烏茲別克文', 'yue-Hans': '乌兹别克文', 'yue-Hant': '烏茲別克文', 'zh': '乌兹别克语', 'zh-Hans': '乌兹别克语', 'zh-Hant': '烏茲別克文', 'zu': 'isi-Uzbek'},
    'uzn': {'en': 'Northern Uzbek'},
    'uzs': {'en': 'Southern Uzbek'},
    'vaa': {'en': 'Vaagri Booli'},
    'vae': {'en': 'Vale'},
    'vaf': {'en': 'Vafsi'},
    'vag': {'en': 'Vagla'},
    'vah': {'en': 'Varhadi-Nagpuri'},
    'vai': {'am': 'ቫይ', 'ar': 'الفاي', 'as': 'ভাই', 'az-Cyrl': 'ваи', 'bal-Latn': 'Wái', 'be': 'ваі', 'be-tarask': 'ваі', 'bg': 'ваи', 'bgn': 'وایی', 'bn': 'ভাই', 'brx': 'भाई', 'bs-Cyrl': 'ваи', 'ccp': '𑄞𑄭', 'ce': 'ваи', 'chr': 'ᏩᏱ', 'ckb': 'ڤایی', 'cy': 'Faieg', 'el': 'Βάι', 'el-polyton': 'Βάι', 'eu': 'vaiera', 'fa': 'ویایی', 'ff-Adlm': '𞤾𞤢𞥄𞤴𞤪𞤫', 'fr': 'vaï', 'ga': 'Vadhais', 'gsw': 'Vai-Schpraach', 'gu': 'વાઇ', 'he': 'וואי', 'hi': 'वाई', 'ht': 'vaï', 'hy': 'վաի', 'is': 'vaí', 'ja': 'ヴァイ語', 'ka': 'ვაი', 'kgp': 'vaj', 'kk': 'вай тілі', 'kk-Cyrl': 'вай тілі', 'km': 'វៃ', 'kn': 'ವಾಯಿ', 'ko': '바이어', 'kok': 'वाई', 'kok-Deva': 'वाई', 'ks': 'واے', 'ks-Arab': 'واے', 'ky': 'вайиче', 'lb': 'Vai-Sprooch', 'lo': 'ໄວ', 'lrc': 'ڤای', 'lv': 'vaju', 'mai': 'बाई', 'mi': 'Wai', 'mk': 'вај', 'ml': 'വൈ', 'mn': 'вай', 'mr': 'वाई', 'my': 'ဗိုင်', 'mzn': 'وایی', 'ne': 'भाइ', 'nqo': 'ߒߝ߭ߊߦߌ߲ߞߊ߲', 'or': 'ଭାଇ', 'pa': 'ਵਾਈ', 'pa-Guru': 'ਵਾਈ', 'pcm': 'Vaí', 'pl': 'wai', 'ps': 'وای', 'qu': 'Vai Simi', 'ru': 'ваи', 'sd': 'يا', 'sd-Arab': 'يا', 'si': 'වයි', 'sl': 'vajščina', 'smn': 'vaikielâ', 'so': 'Faayi', 'sq': 'vaisht', 'sr': 'ваи', 'sr-Cyrl': 'ваи', 'sv': 'vaj', 'sw': 'Kivai', 'ta': 'வை', 'te': 'వాయి', 'th': 'ไว', 'ti': 'ቫይ', 'tk': 'wai dili', 'to': 'lea fakavai', 'ug': 'ۋايچە', 'uk': 'ваї', 'ur': 'وائی', 'uz-Cyrl': 'ваи', 'vai': 'ꕙꔤ', 'vai-Vaii': 'ꕙꔤ', 'vi': 'Tiếng Vai', 'xnr': 'वाई', 'yrl': 'wai', 'yue': '瓦伊文', 'yue-Hans': '瓦伊文', 'yue-Hant': '瓦伊文', 'zh': '瓦伊语', 'zh-Hans': '瓦伊语', 'zh-Hant': '瓦伊文', 'zu': 'isi-Vai', 'en': 'Vai'},
    'vaj': {'en': 'Sekele'},
    'val': {'en': 'Vehes'},
    'vam': {'en': 'Vanimo'},
    'van': {'en': 'Valman'},
    'vao': {'en': 'Vao'},
    'vap': {'en': 'Vaiphei'},
    'var': {'en': 'Huarijio'},
    'vas': {'en': 'Vasavi'},
    'vau': {'en': 'Vanuma'},
    'vav': {'en': 'Varli'},
    'vay': {'en': 'Wayu'},
    'vbb': {'en': 'Southeast Babar'},
    'vbk': {'en': 'Southwestern Bontok'},
    've': {'af': 'Venda', 'am': 'ቬንዳ', 'ar': 'الفيندا', 'as': 'ভেণ্ডা', 'ast': 'venda', 'az': 'venda', 'az-Cyrl': 'венда', 'az-Latn': 'venda', 'bal-Latn': 'Wendá', 'be': 'венда', 'be-tarask': 'венда', 'bg': 'венда', 'bgn': 'وینڈایی', 'bn': 'ভেন্ডা', 'br': 'venda', 'brx': 'भेन्डा', 'bs': 'venda', 'bs-Cyrl': 'венда', 'bs-Latn': 'venda', 'ca': 'venda', 'ca-ES-valencia': 'venda', 'ccp': '𑄞𑄬𑄚𑄴𑄓', 'ce': 'венда', 'chr': 'ᏫᏂᏓ', 'ckb': 'ڤێندا', 'cs': 'venda', 'cy': 'Fendeg', 'da': 'venda', 'de': 'Venda', 'dsb': 'vendašćina', 'ee': 'vendagbe', 'el': 'Βέντα', 'el-polyton': 'Βέντα', 'en': 'Venda', 'es': 'venda', 'et': 'venda', 'eu': 'vendera', 'fa': 'وندایی', 'ff-Adlm': '𞤏𞤫𞤲𞤣𞤢𞥄𞤪𞤫', 'fi': 'venda', 'fil': 'Venda', 'fo': 'venda', 'fr': 'venda', 'fur': 'venda', 'fy': 'Venda', 'ga': 'Veindis', 'gd': 'Venda', 'gl': 'venda', 'gsw': 'Venda-Schpraach', 'gu': 'વેન્દા', 'ha': 'Venda', 'he': 'וונדה', 'hi': 'वेन्दा', 'hi-Latn': 'Venda', 'hr': 'venda', 'hsb': 'vendašćina', 'ht': 'venda', 'hu': 'venda', 'hy': 'վենդա', 'ia': 'venda', 'id': 'Venda', 'ig': 'Venda', 'is': 'venda', 'it': 'venda', 'ja': 'ベンダ語', 'jv': 'Venda', 'ka': 'ვენდა', 'kgp': 'vẽna', 'kk': 'венда тілі', 'kk-Cyrl': 'венда тілі', 'km': 'វេនដា', 'kn': 'ವೆಂಡಾ', 'ko': '벤다어', 'kok': 'वेंदा', 'kok-Deva': 'वेंदा', 'ks': 'ویندا', 'ks-Arab': 'ویندا', 'ksh': 'Wenda', 'ky': 'вендача', 'lb': 'Venda-Sprooch', 'lld': 'venda', 'lo': 'ເວນດາ', 'lt': 'vendų', 'lv': 'vendu', 'mai': 'बेंदा', 'mi': 'Wēnera', 'mk': 'венда', 'ml': 'വെന്ദ', 'mn': 'венда', 'mr': 'व्हेंदा', 'ms': 'Venda', 'mt': 'Venda', 'my': 'ဗင်န်ဒါ', 'nb': 'venda', 'ne': 'भेन्डा', 'nl': 'Venda', 'nn': 'venda', 'no': 'venda', 'or': 'ଭେଣ୍ଡା', 'pa': 'ਵੇਂਡਾ', 'pa-Guru': 'ਵੇਂਡਾ', 'pcm': 'Vẹ́nda Lángwej', 'pl': 'venda', 'ps': 'ویندا', 'pt': 'venda', 'qu': 'Venda Simi', 'rm': 'venda', 'ro': 'venda', 'ru': 'венда', 'sc': 'venda', 'sd': 'وينڊا', 'sd-Arab': 'وينڊا', 'si': 'වෙන්ඩා', 'sk': 'vendčina', 'sl': 'venda', 'smn': 'vendakielâ', 'so': 'Venda', 'sq': 'vendaisht', 'sr': 'венда', 'sr-Cyrl': 'венда', 'sr-Latn': 'venda', 'sv': 'venda', 'sw': 'Kivenda', 'ta': 'வென்டா', 'te': 'వెండా', 'tg': 'венда', 'th': 'เวนดา', 'ti': 'ቨንዳ', 'tk': 'wenda dili', 'to': 'lea fakavenitā', 'tr': 'Venda dili', 'tt': 'венда', 'ug': 'ۋېنداچە', 'uk': 'венда', 'ur': 'وینڈا', 'uz': 'venda', 'uz-Cyrl': 'венда', 'uz-Latn': 'venda', 'vec': 'venda', 'vi': 'Tiếng Venda', 'wae': 'Venda', 'wo': 'Wenda', 'xnr': 'वेन्दा', 'yo': 'Èdè Fenda', 'yrl': 'wẽda', 'yue': '溫達文', 'yue-Hans': '温达文', 'yue-Hant': '溫達文', 'zh': '文达语', 'zh-Hans': '文达语', 'zh-Hant': '溫達文', 'zu': 'isi-Venda', 've': 'tshiVenḓa'},
    'vec': {'af': 'Venesiaans', 'ak': 'Vɛnihyia Kasa', 'am': 'ቬነቲያንኛ', 'ar': 'البندقية', 'as': 'ভেনেছিয়ান', 'ast': 'venecianu', 'az': 'venet dili', 'az-Latn': 'venet dili', 'bal-Latn': 'Weneti', 'be': 'венецыянская', 'be-tarask': 'венецыянская', 'bg': 'венециански', 'bn': 'ভেনেশিয়ান', 'br': 'venezieg', 'bs': 'venecijanski', 'bs-Cyrl': 'венецијански', 'bs-Latn': 'venecijanski', 'ca': 'vènet', 'ca-ES-valencia': 'vènet', 'chr': 'ᏈᏂᏒᏂ', 'cs': 'benátština', 'cy': 'Feniseg', 'da': 'venetiansk', 'de': 'Venetisch', 'dsb': 'venetišćina', 'el': 'Βενετικά', 'el-polyton': 'Βενετικά', 'en': 'Venetian', 'es': 'veneciano', 'et': 'veneti', 'eu': 'veneziera', 'fa': 'ونیزی', 'ff-Adlm': '𞤏𞤫𞤲𞤭𞥅𞤧𞤴𞤢𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'venetsia', 'fil': 'Venetian', 'fo': 'venetiskt', 'fr': 'vénitien', 'ga': 'Veinéisis', 'gd': 'Bheinisis', 'gl': 'véneto', 'gu': 'વેનેશ્યિન', 'ha': 'Veneshiyanci', 'he': 'ונציאנית', 'hi': 'वनीशन', 'hi-Latn': 'Venetian', 'hr': 'venecijanski', 'hsb': 'venetišćina', 'ht': 'vénitien', 'hu': 'velencei', 'hy': 'վենետերեն', 'ia': 'venetiano', 'id': 'Venesia', 'ig': 'Asụsụ Veneshia', 'is': 'feneyska', 'it': 'veneto', 'ja': 'ヴェネト語', 'jv': 'Venesia', 'ka': 'ვენეციური', 'kk': 'венеция тілі', 'kk-Cyrl': 'венеция тілі', 'km': 'វេណេតូ', 'kn': 'ವೆನಿಶಿಯನ್', 'ko': '베네치아어', 'kok': 'वेनेशियन', 'kok-Deva': 'वेनेशियन', 'ku': 'venîsî', 'ky': 'венециянча', 'lb': 'Venezesch', 'lo': 'ເວເນຊຽນ', 'lt': 'venetų', 'lv': 'venēciešu', 'mk': 'венецијански', 'ml': 'വെനീഷ്യൻ', 'mn': 'венец', 'mr': 'व्हेनेशियन', 'ms': 'Venetia', 'my': 'ဗနီးရှန်', 'nb': 'venetiansk', 'ne': 'भेनेसियाली', 'nl': 'Venetiaans', 'nn': 'venetiansk', 'no': 'venetiansk', 'nqo': 'ߝ߭ߣߋߛߌߦߊߞߊ߲', 'or': 'ଭନିଶନ୍\u200c', 'pa': 'ਵੇਨੇਸ਼ੀਅਨ', 'pa-Guru': 'ਵੇਨੇਸ਼ੀਅਨ', 'pcm': 'Venetián Lángwej', 'pl': 'wenecki', 'ps': 'وینټیان', 'pt': 'vêneto', 'ro': 'venetă', 'ru': 'венецианский', 'sc': 'vènetu', 'sd': 'ونيشن', 'sd-Arab': 'ونيشن', 'si': 'වැනේසියානු', 'sk': 'benátčina', 'sl': 'beneščina', 'so': 'Dadka Fenaays', 'sq': 'venetisht', 'sr': 'венецијански', 'sr-Cyrl': 'венецијански', 'sr-Latn': 'venecijanski', 'sv': 'venetianska', 'sw': 'Kivenisi', 'ta': 'வினிசியன்', 'te': 'వెనీషియన్', 'th': 'เวเนโต้', 'ti': 'ቬንቲያንኛ', 'tk': 'wenesian dili', 'to': 'lea fakavenēsia', 'tr': 'Venedikçe', 'uk': 'венеційська', 'ur': 'وینسی', 'uz': 'venet', 'uz-Latn': 'venet', 'vec': 'veneto', 'vi': 'Tiếng Veneto', 'yo': 'Fènéṣìànì', 'yue': '威尼斯文', 'yue-Hans': '威尼斯文', 'yue-Hant': '威尼斯文', 'zh': '威尼斯语', 'zh-Hans': '威尼斯语', 'zh-Hant': '威尼斯文', 'zu': 'IsiVenetian'},
    'ved': {'en': 'Veddah'},
    'vel': {'en': 'Veluws'},
    'vem': {'en': 'Vemgo-Mabas'},
    'veo': {'en': 'Ventureño'},
    'vep': {'ast': 'vepsiu', 'br': 'vepseg', 'ca': 'vepse', 'ca-ES-valencia': 'vepse', 'cs': 'vepština', 'cy': 'Feps', 'de': 'Wepsisch', 'en': 'Veps', 'et': 'vepsa', 'fi': 'vepsä', 'fr': 'vepse', 'gd': 'Veps', 'hi-Latn': 'Veps', 'ht': 'vepse', 'hy': 'վեպսերեն', 'it': 'vepso', 'ja': 'ヴェプス語', 'lb': 'Wepsesch', 'lt': 'vepsų', 'mk': 'вепшки', 'nb': 'vepsisk', 'nl': 'Wepsisch', 'nn': 'vepsisk', 'no': 'vepsisk', 'pl': 'wepski', 'smn': 'vepsäkielâ', 'sv': 'veps', 'th': 'เวปส์', 'to': 'lea fakavepisi', 'tr': 'Veps dili', 'yue': '維普森文', 'yue-Hans': '维普森文', 'yue-Hant': '維普森文', 'zh': '维普森语', 'zh-Hans': '维普森语', 'zh-Hant': '維普森文'},
    'ver': {'en': 'Mom Jango'},
    'vgr': {'en': 'Vaghri'},
    'vgt': {'en': 'Vlaamse Gebarentaal'},
    'vi': {'af': 'Viëtnamees', 'agq': 'Vìyɛtnàmê', 'ak': 'Viɛtnam kasa', 'am': 'ቪየትናምኛ', 'ar': 'الفيتنامية', 'as': 'ভিয়েটনামী', 'asa': 'Kivietinamu', 'ast': 'vietnamín', 'az': 'vyetnam', 'az-Arab': 'ویتنامجا', 'az-Cyrl': 'вјетнам', 'az-Latn': 'vyetnam', 'bal-Latn': 'Wietnámi', 'bas': 'Hɔp u vyɛ̄dnàm', 'be': 'в’етнамская', 'be-tarask': 'в’етнамская', 'bem': 'Ichi Vietinamu', 'bez': 'Hivietinamu', 'bg': 'виетнамски', 'bgn': 'ویتنامی', 'bm': 'wiyɛtinamukan', 'bn': 'ভিয়েতনামী', 'br': 'vietnameg', 'brx': 'भियेतनामिस', 'bs': 'vijetnamski', 'bs-Cyrl': 'вијетнамски', 'bs-Latn': 'vijetnamski', 'ca': 'vietnamita', 'ca-ES-valencia': 'vietnamita', 'ccp': '𑄞𑄨𑄠𑄬𑄖𑄴𑄚𑄟𑄩', 'ce': 'вьетнамхойн', 'cgg': 'Oruviyetinaamu', 'chr': 'ᏫᎡᏘᎾᎻᏍ', 'ckb': 'ڤیەتنامی', 'cs': 'vietnamština', 'cy': 'Fietnameg', 'da': 'vietnamesisk', 'dav': 'Kivietinamu', 'de': 'Vietnamesisch', 'dje': 'Vietnaam senni', 'dsb': 'vietnamšćina', 'dyo': 'vietnam', 'dz': 'བེཊ་ནཱ་མིས་ཁ', 'ebu': 'Kĩvietinamu', 'ee': 'vietnamgbe', 'el': 'Βιετναμικά', 'el-polyton': 'Βιετναμικά', 'en': 'Vietnamese', 'eo': 'vjetnama', 'es': 'vietnamita', 'et': 'vietnami', 'eu': 'vietnamera', 'ewo': 'ǹkɔ́bɔ hiɛdənám', 'fa': 'ویتنامی', 'ff': 'Wiyetnameere', 'ff-Adlm': '𞤏𞤭𞤴𞤫𞤼𞤲𞤢𞤥𞤭𞤲𞤳𞤮𞥅𞤪𞤫', 'ff-Latn': 'Wiyetnameere', 'fi': 'vietnam', 'fil': 'Vietnamese', 'fo': 'vjetnamesiskt', 'fr': 'vietnamien', 'fur': 'vietnamite', 'fy': 'Vietnamees', 'ga': 'Vítneaimis', 'gd': 'Bhiet-Namais', 'gl': 'vietnamita', 'gsw': 'Vietnamesisch', 'gu': 'વિયેતનામીસ', 'guz': 'Kivietinamu', 'ha': 'Harshen Biyetinam', 'haw': 'Wiekanama', 'he': 'וייטנאמית', 'hi': 'वियतनामी', 'hi-Latn': 'Vietnamese', 'hr': 'vijetnamski', 'hsb': 'vietnamšćina', 'ht': 'vietnamien', 'hu': 'vietnámi', 'hy': 'վիետնամերեն', 'ia': 'vietnamese', 'id': 'Vietnam', 'ig': 'Vietnamese', 'is': 'víetnamska', 'it': 'vietnamita', 'ja': 'ベトナム語', 'jmc': 'Kyivietinamu', 'jv': 'Vietnam', 'ka': 'ვიეტნამური', 'kab': 'Tabyiṭnamit', 'kam': 'Kivietinamu', 'kde': 'Chivietinamu', 'kea': 'vietnamita', 'kgp': 'vijétinỹmũ', 'khq': 'Vietnaam senni', 'ki': 'Kivietinamu', 'kk': 'вьетнам тілі', 'kk-Cyrl': 'вьетнам тілі', 'kln': 'kutitab Vietnam', 'km': 'វៀតណាម', 'kn': 'ವಿಯೆಟ್ನಾಮೀಸ್', 'ko': '베트남어', 'kok': 'व्हिएतनामीज', 'kok-Deva': 'व्हिएतनामीज', 'ks': 'وِیَتنَمیٖز', 'ks-Arab': 'وِیَتنَمیٖز', 'ksb': 'Kivietinamu', 'ksf': 'riwyɛtnám', 'ksh': 'Vijätnammehsesch', 'ku': 'vîetnamî', 'kxv': 'vietnaamti', 'kxv-Deva': 'वियतनामी', 'kxv-Latn': 'vietnaamti', 'kxv-Orya': 'ୱିଏତନାମ ତି', 'kxv-Telu': 'వియత్నామీస్', 'ky': 'вьетнамча', 'lag': 'Kɨvietináamu', 'lb': 'Vietnamesesch', 'lg': 'Luvyetinaamu', 'lkt': 'Vietnamese Iyápi', 'lld': 'vietnamesc', 'ln': 'liviyetinámi', 'lo': 'ຫວຽດນາມ', 'lrc': 'ڤییئتنامی', 'lt': 'vietnamiečių', 'lu': 'Liviyetinamu', 'luo': 'Kivietinamu', 'luy': 'Kivietinamu', 'lv': 'vjetnamiešu', 'mai': 'बियतनामी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lvietinamu', 'mer': 'Kĩvietinamu', 'mfe': 'vietnamien', 'mg': 'Vietnamianina', 'mgh': 'Ivyetinamu', 'mi': 'Whitināmu', 'mk': 'виетнамски', 'ml': 'വിയറ്റ്നാമീസ്', 'mn': 'вьетнам', 'mr': 'व्हिएतनामी', 'ms': 'Vietnam', 'mt': 'Vjetnamiż', 'mua': 'Vietnamiya', 'my': 'ဗီယက်နမ်', 'mzn': 'ویتنامی', 'naq': 'Vietnamǁî gowab', 'nb': 'vietnamesisk', 'nd': 'isi-Vietnamese', 'ne': 'भियतनामी', 'nl': 'Vietnamees', 'nmg': 'Kiɛl viɛtnam', 'nn': 'vietnamesisk', 'no': 'vietnamesisk', 'nqo': 'ߝ߭ߌߦߋߕߌߣߊߡߌߞߊ߲', 'nus': 'Thok betnaamni', 'nyn': 'Oruviyetinaamu', 'om': 'Afaan Veetinam', 'or': 'ଭିଏତନାମିଜ୍', 'pa': 'ਵੀਅਤਨਾਮੀ', 'pa-Guru': 'ਵੀਅਤਨਾਮੀ', 'pcm': 'Viẹ́tnám Lángwej', 'pl': 'wietnamski', 'ps': 'وېتنامي', 'pt': 'vietnamita', 'qu': 'Vietnamita Simi', 'rm': 'vietnamais', 'rn': 'Ikinyaviyetinamu', 'ro': 'vietnameză', 'rof': 'Kivietinamu', 'ru': 'вьетнамский', 'rw': 'Ikinyaviyetinamu', 'rwk': 'Kyivietinamu', 'saq': 'Kivietinamu', 'sbp': 'Ishivietinamu', 'sc': 'vietnamita', 'sd': 'ويتنامي', 'sd-Arab': 'ويتنامي', 'se': 'vietnamgiella', 'seh': 'vietnamita', 'ses': 'Vietnaam senni', 'sg': 'Vietnäm', 'shi': 'ⵜⴰⴼⵉⵜⵏⴰⵎⵉⵜ', 'shi-Latn': 'Tafitnamit', 'shi-Tfng': 'ⵜⴰⴼⵉⵜⵏⴰⵎⵉⵜ', 'si': 'වියට්නාම්', 'sk': 'vietnamčina', 'sl': 'vietnamščina', 'smn': 'vietnamkielâ', 'sn': 'chiVietnam', 'so': 'Fiitnaamays', 'sq': 'vietnamisht', 'sr': 'вијетнамски', 'sr-Cyrl': 'вијетнамски', 'sr-Latn': 'vijetnamski', 'sv': 'vietnamesiska', 'sw': 'Kivietinamu', 'syr': 'ܒܝܬܢܐܡܐܝܬ', 'ta': 'வியட்நாமீஸ்', 'te': 'వియత్నామీస్', 'teo': 'Kivietinamu', 'tg': 'ветнамӣ', 'th': 'เวียดนาม', 'ti': 'ቬትናምኛ', 'tk': 'wýetnam dili', 'to': 'lea fakavietinami', 'tr': 'Vietnamca', 'tt': 'вьетнам', 'twq': 'Vietnaam senni', 'tzm': 'Taviṭnamit', 'ug': 'ۋىيېتنامچە', 'uk': 'вʼєтнамська', 'ur': 'ویتنامی', 'uz': 'vyetnam', 'uz-Cyrl': 'ветнамча', 'uz-Latn': 'vyetnam', 'vai': 'ꔲꕩꕯꕆꔧ', 'vai-Latn': 'Viyamíĩ', 'vai-Vaii': 'ꔲꕩꕯꕆꔧ', 'vec': 'vietnameze', 'vi': 'Tiếng Việt', 'vun': 'Kyivietinamu', 'wae': 'Vietnamesiš', 'wo': 'Wiyetnaamiye', 'xnr': 'वियतनामी', 'xog': 'Oluvyetinaamu', 'yav': 'nufiɛtnamíɛŋ', 'yi': 'וויעטנאַמעזיש', 'yo': 'Èdè Jetinamu', 'yrl': 'wietinamita', 'yue': '越南文', 'yue-Hans': '越南文', 'yue-Hant': '越南文', 'zgh': 'ⵜⴰⴱⵉⵜⵏⴰⵎⵉⵜ', 'zh': '越南语', 'zh-Hans': '越南语', 'zh-Hant': '越南文', 'zu': 'isi-Vietnamese'},
    'vic': {'en': 'Virgin Islands Creole English'},
    'vid': {'en': 'Vidunda'},
    'vif': {'en': 'Vili'},
    'vig': {'en': 'Viemo'},
    'vil': {'en': 'Vilela'},
    'vin': {'en': 'Vinza'},
    'vis': {'en': 'Vishavan'},
    'vit': {'en': 'Viti'},
    'viv': {'en': 'Iduna'},
    'vka': {'en': 'Kariyarra'},
    'vki': {'en': 'Ija-Zuba'},
    'vkj': {'en': 'Kujarge'},
    'vkk': {'en': 'Kaur'},
    'vkl': {'en': 'Kulisusu'},
    'vkm': {'en': 'Kamakan'},
    'vkn': {'en': 'Koro Nulu'},
    'vko': {'en': 'Kodeoha'},
    'vkp': {'en': 'Korlai Creole Portuguese'},
    'vkt': {'en': 'Tenggarong Kutai Malay'},
    'vku': {'en': 'Kurrama'},
    'vkz': {'en': 'Koro Zuba'},
    'vlp': {'en': 'Valpei'},
    'vls': {'ast': 'flamencu occidental', 'br': 'flandrezeg ar c’hornôg', 'ca': 'flamenc occidental', 'ca-ES-valencia': 'flamenc occidental', 'cs': 'vlámština (západní)', 'cy': 'Fflemeg Gorllewinol', 'de': 'Westflämisch', 'en': 'West Flemish', 'et': 'lääneflaami', 'fi': 'länsiflaami', 'fr': 'flamand occidental', 'ga': 'Pléimeannais Iartharach', 'gd': 'Flànrais Shiarach', 'hi-Latn': 'West Flemish', 'ht': 'flamand occidental', 'hy': 'արևմտաֆլամանդերեն', 'it': 'fiammingo occidentale', 'ja': '西フラマン語', 'lb': 'Westflämesch', 'lt': 'vakarų flamandų', 'mk': 'западнофламански', 'nb': 'vestflamsk', 'nl': 'West-Vlaams', 'nn': 'vestflamsk', 'no': 'vestflamsk', 'pl': 'zachodnioflamandzki', 'sv': 'västflamländska', 'th': 'เฟลมิชตะวันตก', 'to': 'lea fakavelamingi-hihifo', 'tr': 'Batı Flamanca', 'yi': 'מערב פֿלעמיש', 'yue': '西佛蘭德文', 'yue-Hans': '西佛兰德文', 'yue-Hant': '西佛蘭德文', 'zh-Hant': '西佛蘭德文'},
    'vma': {'en': 'Martuyhunira'},
    'vmb': {'en': 'Barbaram'},
    'vmc': {'en': 'Juxtlahuaca Mixtec'},
    'vmd': {'en': 'Mudu Koraga'},
    'vme': {'en': 'East Masela'},
    'vmf': {'ast': 'franconianu del Main', 'cs': 'němčina (mohansko-franské dialekty)', 'de': 'Mainfränkisch', 'en': 'Main-Franconian', 'et': 'Maini frangi', 'fi': 'maininfrankki', 'fr': 'franconien du Main', 'hi-Latn': 'Main-Franconian', 'ht': 'franconien du Main', 'ja': 'マインフランク語', 'lb': 'Mainfränkesch', 'lt': 'pagrindinė frankonų', 'mk': 'мајнскофранконски', 'nb': 'Main-frankisk', 'ne': 'मुख्य-फ्राङ्कोनियाली', 'nl': 'Opperfrankisch', 'nn': 'Main-frankisk', 'no': 'Main-frankisk', 'pl': 'meński frankoński', 'sv': 'Main-frankiska', 'th': 'เมน-ฟรานโกเนีย', 'to': 'lea fakafalanikoni-loloto', 'tr': 'Main Frankonya Dili', 'yue': '美茵-法蘭克尼亞文', 'yue-Hans': '美茵-法兰克尼亚文', 'yue-Hant': '美茵-法蘭克尼亞文', 'zh-Hant': '美茵-法蘭克尼亞文'},
    'vmg': {'en': 'Lungalunga'},
    'vmh': {'en': 'Maraghei'},
    'vmi': {'en': 'Miwa'},
    'vmj': {'en': 'Ixtayutla Mixtec'},
    'vmk': {'en': 'Makhuwa-Shirima'},
    'vml': {'en': 'Malgana'},
    'vmm': {'en': 'Mitlatongo Mixtec'},
    'vmp': {'en': 'Soyaltepec Mazatec'},
    'vmq': {'en': 'Soyaltepec Mixtec'},
    'vmr': {'en': 'Marenje'},
    'vms': {'en': 'Moksela'},
    'vmu': {'en': 'Muluridyi'},
    'vmv': {'en': 'Valley Maidu'},
    'vmw': {'af': 'Makhuwa', 'ak': 'Makuwa', 'am': 'ማክሁዋኛ', 'ar': 'الماكوا', 'as': 'মাখুৱা', 'az': 'makua dili', 'az-Latn': 'makua dili', 'bal-Latn': 'Makuwá', 'be': 'макуа', 'be-tarask': 'макуа', 'bg': 'макува', 'bn': 'মাখুওয়া', 'bs': 'makua', 'bs-Latn': 'makua', 'ca': 'makua', 'ca-ES-valencia': 'makua', 'chr': 'ᎹᎱᏩ', 'cs': 'makhuwština', 'cy': 'Macua', 'da': 'makhuwa', 'de': 'Makua', 'dsb': 'makhuwašćina', 'el': 'Μακούα', 'el-polyton': 'Μακούα', 'en': 'Makhuwa', 'es': 'makua', 'et': 'makua', 'eu': 'makhuwera', 'fa': 'ماکوا', 'fi': 'makhuwa', 'fil': 'Makhuwa', 'fo': 'makhuwa', 'fr': 'macua', 'ga': 'Macuais', 'gd': 'Makhuwa', 'gl': 'makua', 'gu': 'મખુવા', 'ha': 'Makhuwa', 'he': 'מאקואה', 'hi': 'मखुवा', 'hi-Latn': 'Makhuwa', 'hr': 'makhuwa', 'hsb': 'makhuwašćina', 'ht': 'macua', 'hu': 'makua', 'hy': 'մաքուա', 'ia': 'macua', 'id': 'Makhuwa', 'ig': 'Makhuwa', 'is': 'makúva', 'it': 'macua', 'ja': 'マクア語', 'jv': 'Makhuwa', 'ka': 'მაკჰუვა', 'kk': 'макуа тілі', 'kk-Cyrl': 'макуа тілі', 'km': 'ម៉ាឃូវ៉ា', 'kn': 'ಮಖುವಾ', 'ko': '마쿠와어', 'kok': 'माखुवा', 'kok-Deva': 'माखुवा', 'ku': 'makhuwayî', 'ky': 'махувача', 'lo': 'ມາຄູວາ', 'lt': 'makua', 'lv': 'makua', 'mk': 'макуа', 'ml': 'മഖുവ', 'mn': 'макуа', 'mr': 'मखुवा', 'ms': 'Makhuwa', 'my': 'မတ်ကူးဝါး', 'nb': 'makhuwa', 'ne': 'मखुवा', 'nl': 'Makhuwa', 'nn': 'makhuwa', 'no': 'makhuwa', 'or': 'ମାଖୁୱା', 'pa': 'ਮਖੂਵਾ', 'pa-Guru': 'ਮਖੂਵਾ', 'pcm': 'Mákhuwá Lángwej', 'pl': 'makua', 'ps': 'مکوه', 'pt': 'macua', 'ro': 'makhuwa', 'ru': 'макуа', 'sc': 'macua', 'sd': 'مکووا', 'sd-Arab': 'مکووا', 'si': 'මකුවා', 'sk': 'makhuwčina', 'sl': 'makuva', 'so': 'Af-Makhuwa', 'sq': 'makuvaisht', 'sr': 'макува', 'sr-Cyrl': 'макува', 'sr-Latn': 'makuva', 'sv': 'makua', 'sw': 'Kimakhuwa', 'ta': 'மகுவா', 'te': 'మఖువా', 'th': 'มากัววา', 'ti': 'ማክሁዋ', 'tk': 'mahuwa dili', 'tr': 'Makuaca', 'uk': 'макува', 'ur': 'ماکوائی', 'uz': 'makua', 'uz-Latn': 'makua', 'vec': 'macùa', 'vi': 'Tiếng Makhuwa', 'vmw': 'emakhuwa', 'yo': 'Màkúwà', 'yue': '馬庫瓦文', 'yue-Hans': '马库瓦文', 'yue-Hant': '馬庫瓦文', 'zh': '马库阿语', 'zh-Hans': '马库阿语', 'zh-Hant': '馬庫瓦文', 'zu': 'Makhuwa'},
    'vmx': {'en': 'Tamazola Mixtec'},
    'vmy': {'en': 'Ayautla Mazatec'},
    'vmz': {'en': 'Mazatlán Mazatec'},
    'vnk': {'en': 'Vano'},
    'vnm': {'en': 'Vinmavis'},
    'vnp': {'en': 'Vunapu'},
    'vo': {'af': 'Volapük', 'am': 'ቮላፑክኛ', 'ar': 'لغة الفولابوك', 'as': 'ভোলাপুক', 'ast': 'volapük', 'az': 'volapük', 'az-Cyrl': 'волапүк', 'az-Latn': 'volapük', 'bal-Latn': 'Wolápuk', 'be': 'валапюк', 'be-tarask': 'валапюк', 'bg': 'волапюк', 'bgn': 'ولاپوکی', 'bn': 'ভোলাপুক', 'br': 'volapük', 'brx': 'वोलापोक', 'bs': 'volapuk', 'bs-Cyrl': 'волапук', 'bs-Latn': 'volapuk', 'ca': 'volapük', 'ca-ES-valencia': 'volapük', 'ccp': '𑄞𑄮𑄣𑄛𑄪𑄇𑄴', 'ce': 'волапюк', 'chr': 'ᏬᎳᏊᎩ', 'ckb': 'ڤۆلاپووک', 'cs': 'volapük', 'cy': 'Folapük', 'da': 'volapyk', 'de': 'Volapük', 'dsb': 'volapük', 'el': 'Βολαπιούκ', 'el-polyton': 'Βολαπιούκ', 'en': 'Volapük', 'eo': 'Volapuko', 'es': 'volapük', 'et': 'volapüki', 'eu': 'volapük', 'fa': 'ولاپوک', 'ff-Adlm': '𞤏𞤮𞤤𞤢𞤨𞤵𞤳𞤪𞤫', 'fi': 'volapük', 'fil': 'Volapük', 'fo': 'volapykk', 'fr': 'volapük', 'fy': 'Volapük', 'ga': 'Volapük', 'gd': 'Volapük', 'gl': 'volapuk', 'gsw': 'Volapük', 'gu': 'વોલાપુક', 'ha': 'Volapük', 'he': '\u200fוולאפיק', 'hi': 'वोलापुक', 'hi-Latn': 'Volapük', 'hr': 'volapük', 'hsb': 'volapük', 'ht': 'volapük', 'hu': 'volapük', 'hy': 'վոլապյուկ', 'ia': 'volapük', 'id': 'Volapuk', 'ig': 'Volapük', 'is': 'volapyk', 'it': 'volapük', 'ja': 'ヴォラピュク語', 'jv': 'Volapuk', 'ka': 'ვოლაპუკი', 'kgp': 'vorapuke', 'kk': 'волапюк тілі', 'kk-Cyrl': 'волапюк тілі', 'km': 'វូឡាពូក', 'kn': 'ವೋಲಾಪುಕ್', 'ko': '볼라퓌크어', 'kok': 'ओलापुक', 'kok-Deva': 'ओलापुक', 'ks': 'وولَپُک', 'ks-Arab': 'وولَپُک', 'ksh': 'Volapük', 'ku': 'volapûkî', 'ky': 'волапюкча', 'lb': 'Volapük', 'lld': 'volapük', 'lo': 'ໂວລາພັກ', 'lt': 'volapiuko', 'lv': 'volapiks', 'mk': 'волапик', 'ml': 'വോളാപുക്', 'mn': 'волапюк', 'mr': 'ओलापुक', 'ms': 'Volapük', 'mt': 'Volapuk', 'my': 'ဗိုလာပိုက်', 'nb': 'volapyk', 'ne': 'भोलापिक', 'nl': 'Volapük', 'nn': 'volapyk', 'no': 'volapyk', 'nqo': 'ߝ߭ߏߟߊߔߎߞߊ߲', 'or': 'ବୋଲାପୁକ', 'pa': 'ਵੋਲਾਪੂਕ', 'pa-Guru': 'ਵੋਲਾਪੂਕ', 'pcm': 'Vólapiuk', 'pl': 'wolapik', 'ps': 'والاپوک', 'pt': 'volapuque', 'qu': 'Volapük Simi', 'rm': 'volapuk', 'ro': 'volapuk', 'ru': 'волапюк', 'sc': 'volapük', 'sd': 'والپڪ', 'sd-Arab': 'والپڪ', 'si': 'වොලපූක්', 'sk': 'volapük', 'sl': 'volapik', 'smn': 'volapük', 'so': 'Folabuuk', 'sq': 'volapykisht', 'sr': 'волапик', 'sr-Cyrl': 'волапик', 'sr-Latn': 'volapik', 'sv': 'volapük', 'sw': 'Kivolapuk', 'ta': 'ஒலாபூக்', 'te': 'వోలాపుక్', 'th': 'โวลาพึค', 'ti': 'ቮላፑክ', 'tk': 'wolapýuk dili', 'to': 'lea fakavolapiki', 'tr': 'Volapük', 'ug': 'ۋولاپۇكچە', 'uk': 'волапюк', 'ur': 'وولاپوک', 'uz': 'volapyuk', 'uz-Cyrl': 'волапюк', 'uz-Latn': 'volapyuk', 'vec': 'volapük', 'vi': 'Tiếng Volapük', 'xnr': 'वोलापुक', 'yi': 'וואלאַפּוק', 'yo': 'Fọ́lápùùkù', 'yrl': 'wurapuke', 'yue': '沃拉普克文', 'yue-Hans': '沃拉普克文', 'yue-Hant': '沃拉普克文', 'zh': '沃拉普克语', 'zh-Hans': '沃拉普克语', 'zh-Hant': '沃拉普克文', 'zu': 'isi-Volapük', 'vo': 'Volapük'},
    'vor': {'en': 'Voro'},
    'vot': {'ar': 'الفوتيك', 'ast': 'vóticu', 'az': 'votik', 'az-Latn': 'votik', 'bg': 'вотик', 'bn': 'ভোটিক', 'br': 'votyakeg', 'brx': 'वोटीक', 'bs': 'votski', 'bs-Cyrl': 'вотски', 'bs-Latn': 'votski', 'ca': 'vòtic', 'ca-ES-valencia': 'vòtic', 'ccp': '𑄞𑄮𑄑𑄨𑄇𑄴', 'cs': 'votština', 'cy': 'Foteg', 'da': 'votisk', 'de': 'Wotisch', 'el': 'Βότικ', 'el-polyton': 'Βότικ', 'en': 'Votic', 'es': 'vótico', 'et': 'vadja', 'fa': 'وتی', 'fi': 'vatja', 'fr': 'vote', 'fy': 'Votysk', 'gsw': 'Wotisch', 'gu': 'વોટિક', 'he': 'ווטיק', 'hi': 'वॉटिक', 'hi-Latn': 'Votic', 'hr': 'votski', 'ht': 'vote', 'hu': 'votják', 'hy': 'վոդերեն', 'id': 'Votia', 'is': 'votíska', 'it': 'voto', 'ja': 'ヴォート語', 'kgp': 'vótiko', 'kn': 'ವೋಟಿಕ್', 'ko': '보틱어', 'ks': 'ووتِک', 'ks-Arab': 'ووتِک', 'lb': 'Wotesch', 'lo': 'ໂວຕິກ', 'lt': 'Votik', 'lv': 'votu', 'mk': 'вотски', 'ml': 'വോട്ടിക്', 'mr': 'वॉटिक', 'mt': 'Votik', 'nb': 'votisk', 'nl': 'Votisch', 'nn': 'votisk', 'no': 'votisk', 'or': 'ଭୋଟିକ୍', 'pl': 'wotiacki', 'pt': 'vótico', 'rm': 'votic', 'ro': 'votică', 'ru': 'водский', 'sk': 'vodčina', 'sl': 'votjaščina', 'sr': 'водски', 'sr-Cyrl': 'водски', 'sr-Latn': 'vodski', 'sv': 'votiska', 'ta': 'வோட்க்', 'te': 'వోటిక్', 'th': 'โวทิก', 'to': 'lea fakavotiki', 'tr': 'Votça', 'ug': 'ۋوتېچە', 'uk': 'водська', 'vi': 'Tiếng Votic', 'yrl': 'wútiku', 'yue': '沃提克文', 'yue-Hans': '沃提克文', 'yue-Hant': '沃提克文', 'zh': '沃提克语', 'zh-Hans': '沃提克语', 'zh-Hant': '沃提克文'},
    'vra': {'en': "Vera'a"},
    'vro': {'ast': 'voro', 'br': 'voroeg', 'cs': 'võruština', 'de': 'Võro', 'en': 'Võro', 'et': 'võru', 'fi': 'võro', 'fr': 'võro', 'gd': 'Võro', 'hi-Latn': 'Võro', 'ht': 'võro', 'hy': 'վորո', 'it': 'võro', 'ja': 'ヴォロ語', 'lb': 'Voro', 'lt': 'veru', 'mk': 'виру', 'nb': 'sørestisk', 'nl': 'Võro', 'nn': 'sørestisk', 'no': 'sørestisk', 'pl': 'võro', 'sv': 'võru', 'th': 'โวโร', 'to': 'lea fakavōlo', 'tr': 'Võro', 'yue': '佛羅文', 'yue-Hans': '佛罗文', 'yue-Hant': '佛羅文', 'zh-Hant': '佛羅文'},
    'vrs': {'en': 'Varisi'},
    'vrt': {'en': 'Burmbar'},
    'vsi': {'en': 'Moldova Sign Language'},
    'vsl': {'en': 'Venezuelan Sign Language'},
    'vsv': {'en': 'Valencian Sign Language'},
    'vto': {'en': 'Vitou'},
    'vum': {'en': 'Vumbu'},
    'vun': {'af': 'Vunjo', 'am': 'ቩንጆ', 'ar': 'الفونجو', 'as': 'ভুঞ্জু', 'ast': 'vunjo', 'az': 'vunyo', 'az-Cyrl': 'вунјо', 'az-Latn': 'vunyo', 'bal-Latn': 'Wunjó', 'be': 'вунджо', 'be-tarask': 'вунджо', 'bg': 'вунджо', 'bgn': 'ونجو', 'bn': 'ভুঞ্জো', 'br': 'vunjo', 'brx': 'भुन्ज’', 'bs': 'vunjo', 'bs-Cyrl': 'вунјо', 'bs-Latn': 'vunjo', 'ca': 'vunjo', 'ca-ES-valencia': 'vunjo', 'ccp': '𑄞𑄪𑄚𑄴𑄏𑄮', 'ce': 'вунджо', 'chr': 'ᏭᎾᏦ', 'ckb': 'ڤوونجوو', 'cs': 'vunjo', 'cy': 'Funjo', 'da': 'vunjo', 'de': 'Vunjo', 'dsb': 'vunjo', 'el': 'Βούντζο', 'el-polyton': 'Βούντζο', 'en': 'Vunjo', 'es': 'vunjo', 'et': 'vundžo', 'eu': 'vunjoa', 'fa': 'ونجو', 'ff-Adlm': '𞤏𞤵𞤲𞤶𞤮𞥅𞤪𞤫', 'fi': 'vunjo', 'fil': 'Vunjo', 'fo': 'vunjo', 'fr': 'vunjo', 'fy': 'Vunjo', 'ga': 'Vunjo', 'gd': 'Vunjo', 'gl': 'vunjo', 'gu': 'વુન્જો', 'ha': 'Vunjo', 'he': 'וונג׳ו', 'hi': 'वुंजो', 'hi-Latn': 'Vunjo', 'hr': 'vunjo', 'hsb': 'vunjo', 'ht': 'vunjo', 'hu': 'vunjo', 'hy': 'վունջո', 'ia': 'vunjo', 'id': 'Vunjo', 'ig': 'Vunjo', 'is': 'vunjó', 'it': 'vunjo', 'ja': 'ヴンジョ語', 'jv': 'Vunjo', 'ka': 'ვუნჯო', 'kgp': 'vũjo', 'kk': 'вунджо тілі', 'kk-Cyrl': 'вунджо тілі', 'km': 'វុនចូ', 'kn': 'ವುಂಜೊ', 'ko': '분조어', 'kok': 'वुंजो', 'kok-Deva': 'वुंजो', 'ku': 'vunjo', 'ky': 'вунжочо', 'lb': 'Vunjo', 'lo': 'ວັນໂຈ', 'lrc': 'ڤوٙنجوٙ', 'lt': 'vunjo', 'lv': 'vundžo', 'mai': 'बुंजो', 'mi': 'Whunio', 'mk': 'вунџо', 'ml': 'വുൻജോ', 'mn': 'вунжо', 'mr': 'वुंजो', 'ms': 'Vunjo', 'mt': 'Vunjo', 'my': 'ဗွန်ဂျို', 'mzn': 'وونجویی', 'nb': 'vunjo', 'ne': 'भुन्जो', 'nl': 'Vunjo', 'nn': 'vunjo', 'no': 'vunjo', 'nqo': 'ߝ߭ߎߖߏߞߊ߲', 'or': 'ଭୁନଜୋ', 'pa': 'ਵੂੰਜੋ', 'pa-Guru': 'ਵੂੰਜੋ', 'pcm': 'Vúnjo', 'pl': 'vunjo', 'ps': 'وونجو', 'pt': 'vunjo', 'qu': 'Vunjo Simi', 'rm': 'vunjo', 'ro': 'vunjo', 'ru': 'вунджо', 'sc': 'vunjo', 'sd': 'ونجو', 'sd-Arab': 'ونجو', 'si': 'වුන්ජෝ', 'sk': 'vunjo', 'sl': 'vunjo', 'smn': 'vunjo', 'so': 'Fuunjo', 'sq': 'vunxhoisht', 'sr': 'вунџо', 'sr-Cyrl': 'вунџо', 'sr-Latn': 'vundžo', 'sv': 'vunjo', 'sw': 'Kivunjo', 'ta': 'வுன்ஜோ', 'te': 'వుంజొ', 'th': 'วุนจู', 'ti': 'ቩንጆ', 'tk': 'wunýo dili', 'to': 'lea fakavūniso', 'tr': 'Vunjo', 'ug': 'ۋۇنجوچە', 'uk': 'вуньо', 'ur': 'ونجو', 'uz': 'vunjo', 'uz-Cyrl': 'вунжо', 'uz-Latn': 'vunjo', 'vec': 'vunjo', 'vi': 'Tiếng Vunjo', 'vun': 'Kyivunjo', 'xnr': 'वुंजो', 'yo': 'Funjo', 'yrl': 'wũyu', 'yue': '溫舊文', 'yue-Hans': '温旧文', 'yue-Hant': '溫舊文', 'zh': '温旧语', 'zh-Hans': '温旧语', 'zh-Hant': '溫舊文', 'zu': 'isiVunjo'},
    'vut': {'en': 'Vute'},
    'vwa': {'en': 'Awa (China)'},
    'wa': {'af': 'Walloon', 'am': 'ዋሎን', 'ar': 'الولونية', 'as': 'ৱালুন', 'ast': 'valón', 'az': 'valun', 'az-Cyrl': 'валун', 'az-Latn': 'valun', 'bal-Latn': 'Wallun', 'be': 'валонская', 'be-tarask': 'валонская', 'bg': 'валонски', 'bgn': 'والونی', 'bn': 'ওয়ালুন', 'br': 'walloneg', 'brx': 'वाल्लुन', 'bs': 'valun', 'bs-Cyrl': 'валун', 'bs-Latn': 'valun', 'ca': 'való', 'ca-ES-valencia': 'való', 'ccp': '𑄤𑄣𑄪𑄚𑄴', 'ce': 'валлонойн', 'chr': 'ᏩᎷᎾ', 'ckb': 'والوون', 'cs': 'valonština', 'cy': 'Walwneg', 'da': 'vallonsk', 'de': 'Wallonisch', 'dsb': 'walonšćina', 'el': 'Βαλλωνικά', 'el-polyton': 'Βαλλωνικά', 'en': 'Walloon', 'es': 'valón', 'et': 'vallooni', 'eu': 'valoniera', 'fa': 'والونی', 'ff-Adlm': '𞤏𞤢𞥄𞤤𞤮𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'valloni', 'fil': 'Walloon', 'fo': 'walloon', 'fr': 'wallon', 'fur': 'valon', 'fy': 'Waals', 'ga': 'Vallúnais', 'gd': 'Walloon', 'gl': 'valón', 'gsw': 'Wallonisch', 'gu': 'વાલૂન', 'ha': 'Walloon', 'he': 'ולונית', 'hi': 'वाल्लून', 'hi-Latn': 'Walloon', 'hr': 'valonski', 'hsb': 'walonšćina', 'ht': 'wallon', 'hu': 'vallon', 'hy': 'վալոներեն', 'ia': 'wallon', 'id': 'Walloon', 'ig': 'Walloon', 'is': 'vallónska', 'it': 'vallone', 'ja': 'ワロン語', 'jv': 'Walloon', 'ka': 'ვალონური', 'kgp': 'varỹv', 'kk': 'валлон тілі', 'kk-Cyrl': 'валлон тілі', 'km': 'វ៉ាលូន', 'kn': 'ವಾಲೂನ್', 'ko': '왈론어', 'kok': 'वालून', 'kok-Deva': 'वालून', 'ks': 'وَلوٗن', 'ks-Arab': 'وَلوٗن', 'ksh': 'Walohnesch', 'ku': 'walonî', 'ky': 'валлончо', 'lb': 'Wallounesch', 'lld': 'valonesc', 'lo': 'ວໍລູມ', 'lt': 'valonų', 'lv': 'valoņu', 'mai': 'बलून', 'mi': 'Warūna', 'mk': 'валонски', 'ml': 'വല്ലൂൺ', 'mn': 'уоллун', 'mr': 'वालून', 'ms': 'Walloon', 'mt': 'Walloon', 'my': 'ဝါလူးန်', 'nb': 'vallonsk', 'ne': 'वाल्लुन', 'nl': 'Waals', 'nn': 'vallonsk', 'no': 'vallonsk', 'or': 'ୱାଲୁନ୍', 'pa': 'ਵਲੂਨ', 'pa-Guru': 'ਵਲੂਨ', 'pcm': 'Wálun Lángwej', 'pl': 'waloński', 'ps': 'والون', 'pt': 'valão', 'qu': 'Valona Simi', 'rm': 'vallon', 'ro': 'valonă', 'ru': 'валлонский', 'sc': 'vallonu', 'sd': 'ولون', 'sd-Arab': 'ولون', 'se': 'vallonagiella', 'si': 'වෑලූන්', 'sk': 'valónčina', 'sl': 'valonščina', 'smn': 'walloonkielâ', 'so': 'Walloon', 'sq': 'ualunisht', 'sr': 'валонски', 'sr-Cyrl': 'валонски', 'sr-Latn': 'valonski', 'sv': 'vallonska', 'sw': 'Kiwaloon', 'ta': 'ஒவாலூன்', 'te': 'వాలూన్', 'th': 'วอลลูน', 'ti': 'ዋሎን', 'tk': 'wallon dili', 'to': 'lea fakaʻualonia', 'tr': 'Valonca', 'ug': 'ۋاللۇنچە', 'uk': 'валлонська', 'ur': 'والون', 'uz': 'vallon', 'uz-Latn': 'vallon', 'vec': 'valon', 'vi': 'Tiếng Walloon', 'wa': 'walon', 'xnr': 'वाल्लून', 'yo': 'Èdè Waluni', 'yrl': 'warãu', 'yue': '瓦隆文', 'yue-Hans': '瓦隆文', 'yue-Hant': '瓦隆文', 'zh': '瓦隆语', 'zh-Hans': '瓦隆语', 'zh-Hant': '瓦隆文', 'zu': 'isi-Walloon'},
    'waa': {'en': 'Walla Walla'},
    'wab': {'en': 'Wab'},
    'wac': {'en': 'Wasco-Wishram'},
    'wad': {'en': 'Wamesa'},
    'wae': {'af': 'Walser', 'am': 'ዋልሰር', 'ar': 'الوالسر', 'as': 'ৱালছেৰ', 'ast': 'walser', 'az': 'valles', 'az-Cyrl': 'валлес', 'az-Latn': 'valles', 'bal-Latn': 'Welser', 'be': 'вальшская', 'be-tarask': 'вальшская', 'bg': 'валзерски немски', 'bgn': 'والسیری', 'bn': 'ওয়ালসার', 'br': 'walser', 'brx': 'वाल्सार', 'bs': 'valser', 'bs-Cyrl': 'валсерски', 'bs-Latn': 'valser', 'ca': 'walser', 'ca-ES-valencia': 'walser', 'ccp': '𑄤𑄣𑄧𑄥𑄬𑄢𑄴', 'ce': 'валлисийн', 'chr': 'ᏩᎵᏎᎵ', 'ckb': 'والسێر', 'cs': 'němčina (walser)', 'cy': 'Walsereg', 'da': 'walsertysk', 'de': 'Walliserdeutsch', 'dsb': 'walzeršćina', 'ee': 'walsegbe', 'el': 'Βάλσερ', 'el-polyton': 'Βάλσερ', 'en': 'Walser', 'es': 'walser', 'et': 'valsi', 'eu': 'walserera', 'fa': 'والسر', 'ff-Adlm': '𞤏𞤢𞤤𞤧𞤫𞥅𞤪𞤫', 'fi': 'walser', 'fil': 'Walser', 'fo': 'walser', 'fr': 'walser', 'fy': 'Walser', 'ga': 'Walser', 'gd': 'Gearmailtis Wallis', 'gl': 'walser', 'gu': 'વેલ્સેર', 'ha': 'Walser', 'he': 'וואלסר', 'hi': 'वाल्सर', 'hi-Latn': 'Walser', 'hr': 'walserski', 'hsb': 'walzeršćina', 'ht': 'walser', 'hu': 'walser', 'hy': 'վալսերեն', 'ia': 'walser', 'id': 'Walser', 'ig': 'Walser', 'is': 'valser', 'it': 'walser', 'ja': 'ヴァリス語', 'jv': 'Walser', 'ka': 'ვალსერი', 'kgp': 'varser', 'kk': 'вальзер тілі', 'kk-Cyrl': 'вальзер тілі', 'km': 'វេលសឺ', 'kn': 'ವಾಲ್ಸರ್', 'ko': '월저어', 'kok': 'वाल्सर', 'kok-Deva': 'वाल्सर', 'ksh': 'Walserdütsch', 'ku': 'walserî', 'ky': 'валцерче', 'lb': 'Walliserdäitsch', 'lo': 'ວາເຊີ', 'lt': 'valserų', 'lv': 'Vallisas vācu', 'mai': 'बालसर', 'mi': 'Wāhere', 'mk': 'валсер', 'ml': 'വാൾസർ', 'mn': 'уолсэр', 'mr': 'वालसेर', 'ms': 'Walser', 'mt': 'Walser', 'my': 'ဝေါလ်ဆာ', 'nb': 'walsertysk', 'ne': 'वाल्सर', 'nl': 'Walser', 'nn': 'walsertysk', 'no': 'walsertysk', 'nqo': 'ߥߊߟߑߛߍߙߌߞߊ߲', 'or': 'ୱାଲସେର୍', 'pa': 'ਵਾਲਸਰ', 'pa-Guru': 'ਵਾਲਸਰ', 'pcm': 'Wọ́lsa', 'pl': 'walser', 'ps': 'ولسیر', 'pt': 'walser', 'qu': 'Walser Simi', 'rm': 'gualser', 'ro': 'walser', 'ru': 'валлисский', 'sc': 'walser', 'sd': 'والسر', 'sd-Arab': 'والسر', 'si': 'වොල්සර්', 'sk': 'walserčina', 'sl': 'walser', 'smn': 'walliskielâ', 'so': 'Walseer', 'sq': 'ualserisht', 'sr': 'валсерски', 'sr-Cyrl': 'валсерски', 'sr-Latn': 'valserski', 'sv': 'walsertyska', 'sw': 'Kiwalser', 'ta': 'வால்சேர்', 'te': 'వాల్సర్', 'th': 'วัลเซอร์', 'ti': 'ዋልሰር', 'tk': 'walzer dili', 'to': 'lea fakaʻualiseli', 'tr': 'Walser', 'ug': 'ۋالسېرچە', 'uk': 'вальзерська', 'ur': 'والسر', 'uz': 'valis', 'uz-Cyrl': 'валсерча', 'uz-Latn': 'valis', 'vec': 'walser', 'vi': 'Tiếng Walser', 'wae': 'Walser', 'xnr': 'वाल्सर', 'yo': 'Wọsà', 'yrl': 'wauseri', 'yue': '瓦瑟文', 'yue-Hans': '瓦瑟文', 'yue-Hant': '瓦瑟文', 'zh': '瓦尔瑟语', 'zh-Hans': '瓦尔瑟语', 'zh-Hant': '瓦爾瑟文', 'zu': 'isi-Walser'},
    'waf': {'en': 'Wakoná'},
    'wag': {'en': "Wa'ema"},
    'wah': {'en': 'Watubela'},
    'wai': {'en': 'Wares'},
    'waj': {'en': 'Waffa'},
    'wak': {'en': 'Wakashan languages'},
    'wal': {'af': 'Wolaytta', 'am': 'ወላይትኛ', 'ar': 'الولاياتا', 'as': 'ওলেইটা', 'ast': 'wolaytta', 'az': 'valamo', 'az-Cyrl': 'валамо', 'az-Latn': 'valamo', 'bal-Latn': 'Wolettá', 'be': 'волайта', 'be-tarask': 'волайта', 'bg': 'валамо', 'bgn': 'وولایٹی', 'bn': 'ওলায়ট্টা', 'br': 'walamo', 'brx': 'वलाईत्ता', 'bs': 'valamo', 'bs-Cyrl': 'валамо', 'bs-Latn': 'valamo', 'ca': 'wolaita', 'ca-ES-valencia': 'wolaita', 'ccp': '𑄤𑄣𑄟𑄮', 'ce': 'воламо', 'chr': 'ᏬᎳᏱᏔ', 'ckb': 'وۆلایتا', 'cs': 'wolajtština', 'cy': 'Walamo', 'da': 'walamo', 'de': 'Walamo', 'dsb': 'wolayttašćina', 'el': 'Γουολάιτα', 'el-polyton': 'Γουολάιτα', 'en': 'Wolaytta', 'es': 'wolayta', 'et': 'volaita', 'eu': 'wolayttera', 'fa': 'والامو', 'ff-Adlm': '𞤏𞤮𞥅𞤤𞤢𞤴𞤼𞤢𞥄𞤪𞤫', 'fi': 'wolaitta', 'fil': 'Wolaytta', 'fo': 'wolaytta', 'fr': 'walamo', 'fy': 'Walamo', 'ga': 'Uailéitis', 'gd': 'Wolaytta', 'gl': 'wolaytta', 'gsw': 'Walamo-Schpraach', 'gu': 'વોલાયટ્ટા', 'ha': 'Wolaytta', 'he': 'ווליאטה', 'hi': 'वलामो', 'hi-Latn': 'walamo', 'hr': 'walamo', 'hsb': 'wolayttašćina', 'ht': 'walamo', 'hu': 'valamo', 'hy': 'վոլայտա', 'ia': 'wolaytta', 'id': 'Walamo', 'ig': 'Wolaytta', 'is': 'volayatta', 'it': 'walamo', 'ja': 'ウォライタ語', 'jv': 'Wolaitta', 'ka': 'ველაითა', 'kgp': 'vorajta', 'kk': 'волайта тілі', 'kk-Cyrl': 'волайта тілі', 'km': 'វ៉ូឡាយតា', 'kn': 'ವಲಾಯ್ತಾ', 'ko': '월라이타어', 'kok': 'वोलायटा', 'kok-Deva': 'वोलायटा', 'ks': 'والامو', 'ks-Arab': 'والامو', 'ksh': 'Welahmo', 'ku': 'wolaytta', 'ky': 'вольяттача', 'lb': 'Walamo-Sprooch', 'lo': 'ໂວເລຕາ', 'lt': 'valamo', 'lv': 'valamu', 'mai': 'बोलेटा', 'mi': 'Wareita', 'mk': 'воламо', 'ml': 'വൊലൈറ്റ', 'mn': 'уоллайтта', 'mr': 'वोलायता', 'ms': 'Wolaytta', 'mt': 'Walamo', 'my': 'ဝိုလက်တာ', 'nb': 'wolaytta', 'ne': 'वोलेट्टा', 'nl': 'Wolaytta', 'nn': 'wolaytta', 'no': 'wolaytta', 'or': 'ୱାଲମୋ', 'pa': 'ਵੋਲਾਏਟਾ', 'pa-Guru': 'ਵੋਲਾਏਟਾ', 'pcm': 'Wolaítá Lángwej', 'pl': 'wolayta', 'ps': 'ولایټا', 'pt': 'wolaytta', 'qu': 'Wolaytta Simi', 'rm': 'walamo', 'ro': 'wolaita', 'ru': 'воламо', 'sc': 'wolaita', 'sd': 'وولايٽا', 'sd-Arab': 'وولايٽا', 'si': 'වොලෙට්ට', 'sk': 'walamčina', 'sl': 'valamščina', 'smn': 'wolaitakielâ', 'so': 'Wolaytta', 'sq': 'ulajtaisht', 'sr': 'волајта', 'sr-Cyrl': 'волајта', 'sr-Latn': 'volajta', 'sv': 'walamo', 'sw': 'Kiwolaytta', 'ta': 'வோலாய்ட்டா', 'te': 'వాలేట్టా', 'th': 'โวแลตตา', 'ti': 'ዎላይታኛ', 'tk': 'wolaýta dili', 'to': 'lea fakaʻuolaita', 'tr': 'Valamo', 'ug': 'ۋولايتاچە', 'uk': 'волайтта', 'ur': 'وولایتا', 'uz': 'volamo', 'uz-Cyrl': 'волятта', 'uz-Latn': 'volamo', 'vec': 'wolaita', 'vi': 'Tiếng Walamo', 'xnr': 'वलामो', 'yo': 'Èdè Wolata', 'yrl': 'woraita', 'yue': '瓦拉莫文', 'yue-Hans': '瓦拉莫文', 'yue-Hant': '瓦拉莫文', 'zh': '瓦拉莫语', 'zh-Hans': '瓦拉莫语', 'zh-Hant': '瓦拉莫文', 'zu': 'isi-Wolaytta'},
    'wam': {'en': 'Wampanoag'},
    'wan': {'en': 'Wan'},
    'wao': {'en': 'Wappo'},
    'wap': {'en': 'Wapishana'},
    'waq': {'en': 'Wagiman'},
    'war': {'af': 'Waray', 'am': 'ዋራይ', 'ar': 'الواراي', 'as': 'ৱাৰে', 'ast': 'waray', 'az': 'varay', 'az-Cyrl': 'варај', 'az-Latn': 'varay', 'be': 'варай', 'be-tarask': 'варай', 'bg': 'варай', 'bgn': 'واری', 'bn': 'ওয়ারে', 'br': 'waray', 'brx': 'वारै', 'bs': 'varej', 'bs-Cyrl': 'варај', 'bs-Latn': 'varej', 'ca': 'waray', 'ca-ES-valencia': 'waray', 'ccp': '𑄤𑄢𑄬', 'ce': 'варай', 'chr': 'ᏩᎴ', 'ckb': 'وارای', 'cs': 'warajština', 'cy': 'Winarayeg', 'da': 'waray', 'de': 'Waray', 'dsb': 'warayšćina', 'el': 'Γουάραϊ', 'el-polyton': 'Γουάραϊ', 'en': 'Waray', 'es': 'waray', 'et': 'varai', 'eu': 'warayera', 'fa': 'وارایی', 'ff-Adlm': '𞤏𞤢𞤪𞤢𞤴𞤫𞥅𞤪𞤫', 'fi': 'waray', 'fil': 'Waray', 'fo': 'waray', 'fr': 'waray', 'fy': 'Waray', 'ga': 'Uairéis', 'gd': 'Waray', 'gl': 'waray-waray', 'gsw': 'Waray', 'gu': 'વારેય', 'ha': 'Waray', 'he': 'ווראי', 'hi': 'वारै', 'hi-Latn': 'Waray', 'hr': 'waray', 'hsb': 'warayšćina', 'ht': 'waray', 'hu': 'varaó', 'hy': 'վարայերեն', 'ia': 'waray', 'id': 'Warai', 'ig': 'Waray', 'is': 'varaí', 'it': 'waray', 'ja': 'ワライ語', 'jv': 'Warai', 'ka': 'ვარაი', 'kgp': 'varaj', 'kk': 'варай тілі', 'kk-Cyrl': 'варай тілі', 'km': 'វ៉ារេយ', 'kn': 'ವರಾಯ್', 'ko': '와라이어', 'kok': 'वरय', 'kok-Deva': 'वरय', 'ks': 'وَریے', 'ks-Arab': 'وَریے', 'ksh': 'Waray-Waray', 'ku': 'warayî', 'ky': 'варайча', 'lb': 'Waray', 'lo': 'ວາເລ', 'lt': 'varai', 'lv': 'varaju', 'mai': 'बरय', 'mi': 'Warei', 'mk': 'варајски', 'ml': 'വാരേയ്', 'mn': 'варай', 'mr': 'वारे', 'ms': 'Waray', 'mt': 'Waray', 'my': 'ဝါရေး', 'nb': 'waray-waray', 'ne': 'वारे', 'nl': 'Waray', 'nn': 'waray', 'no': 'waray-waray', 'or': 'ୱାରୈ', 'pa': 'ਵੈਰੇ', 'pa-Guru': 'ਵੈਰੇ', 'pcm': 'Warai Lángwej', 'pl': 'waraj', 'ps': 'وارۍ', 'pt': 'waray', 'qu': 'Waray Simi', 'rm': 'waray', 'ro': 'waray', 'ru': 'варай', 'sc': 'waray', 'sd': 'واري', 'sd-Arab': 'واري', 'si': 'වොරෙය්', 'sk': 'waray', 'sl': 'varajščina', 'smn': 'waraykielâ', 'so': 'Waray', 'sq': 'uarajisht', 'sr': 'варајски', 'sr-Cyrl': 'варајски', 'sr-Latn': 'varajski', 'sv': 'waray', 'sw': 'Kiwaray', 'ta': 'வாரே', 'te': 'వారే', 'th': 'วาเรย์', 'ti': 'ዋራይ', 'tk': 'waraý dili', 'to': 'lea fakaʻualai', 'tr': 'Varay', 'ug': 'ۋارايچە', 'uk': 'варай', 'ur': 'وارے', 'uz': 'varay', 'uz-Latn': 'varay', 'vec': 'waray', 'vi': 'Tiếng Waray', 'xnr': 'वारै', 'yo': 'Èdè Wara', 'yrl': 'warai', 'yue': '瓦瑞文', 'yue-Hans': '瓦瑞文', 'yue-Hant': '瓦瑞文', 'zh': '瓦瑞语', 'zh-Hans': '瓦瑞语', 'zh-Hant': '瓦瑞文', 'zu': 'isi-Waray'},
    'was': {'ar': 'الواشو', 'ast': 'washo', 'az': 'vaşo', 'az-Latn': 'vaşo', 'bg': 'уашо', 'bn': 'ওয়াশো', 'br': 'washo', 'brx': 'वाशो', 'bs': 'vašo', 'bs-Cyrl': 'вашо', 'bs-Latn': 'vašo', 'ca': 'washo', 'ca-ES-valencia': 'washo', 'ccp': '𑄤𑄥𑄮', 'cs': 'waština', 'cy': 'Washo', 'da': 'washo', 'de': 'Washo', 'el': 'Γουασό', 'el-polyton': 'Γουασό', 'en': 'Washo', 'es': 'washo', 'et': 'vašo', 'fa': 'واشویی', 'fi': 'washo', 'fr': 'washo', 'fy': 'Washo', 'gd': 'Washo', 'gsw': 'Washo-Schpraach', 'gu': 'વાશો', 'he': 'וואשו', 'hi': 'वाशो', 'hi-Latn': 'Washo', 'hr': 'washo', 'ht': 'washo', 'hu': 'vasó', 'hy': 'վաշո', 'id': 'Washo', 'is': 'vasjó', 'it': 'washo', 'ja': 'ワショ語', 'kgp': 'vaso', 'kn': 'ವಾಷೋ', 'ko': '와쇼어', 'ks': 'واشو', 'ks-Arab': 'واشو', 'lb': 'Washo-Sprooch', 'lo': 'ວາໂຊ', 'lt': 'Vašo', 'lv': 'vašo', 'mk': 'вашо', 'ml': 'വാഷൊ', 'mr': 'वाशो', 'mt': 'Washo', 'nb': 'washo', 'nl': 'Washo', 'nn': 'washo', 'no': 'washo', 'or': 'ୱାସୋ', 'pl': 'washo', 'pt': 'washo', 'rm': 'washo', 'ro': 'washo', 'ru': 'вашо', 'sk': 'washo', 'sl': 'vašajščina', 'sr': 'вашо', 'sr-Cyrl': 'вашо', 'sr-Latn': 'vašo', 'sv': 'washo', 'ta': 'வாஷோ', 'te': 'వాషో', 'th': 'วาโช', 'to': 'lea fakaʻuasiō', 'tr': 'Vaşo', 'ug': 'ۋاشوچە', 'uk': 'вашо', 'vi': 'Tiếng Washo', 'yrl': 'waxu', 'yue': '瓦紹文', 'yue-Hans': '瓦绍文', 'yue-Hant': '瓦紹文', 'zh': '瓦绍语', 'zh-Hans': '瓦绍语', 'zh-Hant': '瓦紹文'},
    'wat': {'en': 'Kaninuwa'},
    'wau': {'en': 'Waurá'},
    'wav': {'en': 'Waka'},
    'waw': {'en': 'Waiwai'},
    'wax': {'en': 'Watam'},
    'way': {'en': 'Wayana'},
    'waz': {'en': 'Wampur'},
    'wba': {'en': 'Warao'},
    'wbb': {'en': 'Wabo'},
    'wbe': {'en': 'Waritai'},
    'wbf': {'en': 'Wara'},
    'wbh': {'en': 'Wanda'},
    'wbi': {'en': 'Vwanji'},
    'wbj': {'en': 'Alagwa'},
    'wbk': {'en': 'Waigali'},
    'wbl': {'en': 'Wakhi'},
    'wbm': {'en': 'Wa'},
    'wbp': {'af': 'Warlpiri', 'am': 'ዋርልፒሪ', 'ar': 'وارلبيري', 'ast': 'warlpiri', 'az': 'valpiri', 'az-Latn': 'valpiri', 'bal-Latn': 'Warlpiri', 'be': 'варлпіры', 'be-tarask': 'варлпіры', 'bg': 'валпири', 'bn': 'ওয়ার্লপিরি', 'bs': 'varlpiri', 'bs-Latn': 'varlpiri', 'ca': 'warlpiri', 'ca-ES-valencia': 'warlpiri', 'ccp': '𑄤𑄢𑄴𑄣𑄴𑄛𑄨𑄢𑄨', 'ce': 'варлпири', 'cs': 'warlpiri', 'cy': 'Warlpiri', 'da': 'walbiri', 'de': 'Warlpiri', 'el': 'Γουαρλπίρι', 'el-polyton': 'Γουαρλπίρι', 'en': 'Warlpiri', 'es': 'warlpiri', 'et': 'varlpiri', 'fa': 'وارلپیری', 'fi': 'warlpiri', 'fil': 'Warlpiri', 'fo': 'warlpiri', 'fr': 'warlpiri', 'gd': 'Warlpiri', 'gl': 'walrpiri', 'gu': 'વાર્લ્પીરી', 'he': 'וורלפירי', 'hi': 'वॉल्पेरी', 'hi-Latn': 'Warlpiri', 'hr': 'warlpiri', 'ht': 'warlpiri', 'hu': 'warlpiri', 'hy': 'վարլպիրի', 'ia': 'warlpiri', 'id': 'Warlpiri', 'ig': 'Warlpiri', 'is': 'varlpiri', 'it': 'warlpiri', 'ja': 'ワルピリ語', 'ka': 'ვალპირი', 'kgp': 'varpiri', 'kk': 'вальбири тілі', 'kk-Cyrl': 'вальбири тілі', 'km': 'វ៉ារីប៉ារី', 'kn': 'ವಾರ್ಲ್\u200cಪಿರಿ', 'ko': '왈피리어', 'ku': 'warlpiri', 'ky': 'ворлпириче', 'lo': 'ວາຣພິຣິ', 'lrc': 'ڤارلپیری', 'lt': 'valrpiri', 'lv': 'varlpirī', 'mk': 'варлпири', 'ml': 'വൂൾപിരി', 'mr': 'वार्लपिरी', 'ms': 'Warlpiri', 'my': 'ဝေါလ်ပီရီ', 'mzn': 'والرپیری', 'nb': 'warlpiri', 'ne': 'वार्ल्पिरी', 'nl': 'Warlpiri', 'nn': 'warlpiri', 'no': 'warlpiri', 'pa': 'ਵਾਲਪੁਰੀ', 'pa-Guru': 'ਵਾਲਪੁਰੀ', 'pl': 'warlpiri', 'pt': 'warlpiri', 'ro': 'warlpiri', 'ru': 'вальбири', 'si': 'වොපිරි', 'sk': 'warlpiri', 'sl': 'varlpirščina', 'sq': 'uarlpirisht', 'sr': 'варлпири', 'sr-Cyrl': 'варлпири', 'sr-Latn': 'varlpiri', 'sv': 'warlpiri', 'sw': 'Kiwarlpiri', 'ta': 'வல்பிரி', 'te': 'వార్లపిరి', 'th': 'วอล์เพอร์รี', 'ti': 'ዋርልፒሪ', 'to': 'lea fakaʻuālipili', 'tr': 'Warlpiri', 'uk': 'валпірі', 'ur': 'وارلپیری', 'uz': 'valbiri', 'uz-Latn': 'valbiri', 'vec': 'warlpiri', 'vi': 'Tiếng Warlpiri', 'wbp': 'Warlpiri', 'xnr': 'वॉल्पेरी', 'yrl': 'waripiri', 'yue': '沃皮瑞文', 'yue-Hans': '沃皮瑞文', 'yue-Hant': '沃皮瑞文', 'zh': '瓦尔皮瑞语', 'zh-Hans': '瓦尔皮瑞语', 'zh-Hant': '沃皮瑞文', 'zu': 'isi-Warlpiri'},
    'wbq': {'en': 'Waddar'},
    'wbr': {'en': 'Wagdi'},
    'wbs': {'en': 'West Bengal Sign Language'},
    'wbt': {'en': 'Warnman'},
    'wbv': {'en': 'Wajarri'},
    'wbw': {'en': 'Woi'},
    'wca': {'en': 'Yanomámi'},
    'wci': {'en': 'Waci Gbe'},
    'wdd': {'en': 'Wandji'},
    'wdg': {'en': 'Wadaginam'},
    'wdj': {'en': 'Wadjiginy'},
    'wdk': {'en': 'Wadikali'},
    'wdu': {'en': 'Wadjigu'},
    'wdy': {'en': 'Wadjabangayi'},
    'wea': {'en': 'Wewaw'},
    'wec': {'en': 'Wè Western'},
    'wed': {'en': 'Wedau'},
    'weg': {'en': 'Wergaia'},
    'weh': {'en': 'Weh'},
    'wei': {'en': 'Kiunum'},
    'wem': {'en': 'Weme Gbe'},
    'wen': {'en': 'Sorbian languages'},
    'weo': {'en': 'Wemale'},
    'wep': {'en': 'Westphalien'},
    'wer': {'en': 'Weri'},
    'wes': {'en': 'Cameroon Pidgin'},
    'wet': {'en': 'Perai'},
    'weu': {'en': 'Rawngtu Chin'},
    'wew': {'en': 'Wejewa'},
    'wfg': {'en': 'Yafi'},
    'wga': {'en': 'Wagaya'},
    'wgb': {'en': 'Wagawaga'},
    'wgg': {'en': 'Wangkangurru'},
    'wgi': {'en': 'Wahgi'},
    'wgo': {'en': 'Waigeo'},
    'wgu': {'en': 'Wirangu'},
    'wgw': {'en': 'Wagawaga'},
    'wgy': {'en': 'Warrgamay'},
    'wha': {'en': 'Sou Upaa'},
    'whg': {'en': 'North Wahgi'},
    'whk': {'en': 'Wahau Kenyah'},
    'whu': {'en': 'Wahau Kayan'},
    'wib': {'en': 'Southern Toussian'},
    'wic': {'en': 'Wichita'},
    'wie': {'en': 'Wik-Epa'},
    'wif': {'en': 'Wik-Keyangan'},
    'wig': {'en': 'Wik Ngathan'},
    'wih': {'en': "Wik-Me'anha"},
    'wii': {'en': 'Minidien'},
    'wij': {'en': 'Wik-Iiyanh'},
    'wik': {'en': 'Wikalkan'},
    'wil': {'en': 'Wilawila'},
    'wim': {'en': 'Wik-Mungkan'},
    'win': {'en': 'Ho-Chunk'},
    'wir': {'en': 'Wiraféd'},
    'wit': {'en': 'Wintu'},
    'wiu': {'en': 'Wiru'},
    'wiv': {'en': 'Vitu'},
    'wiw': {'en': 'Wirangu'},
    'wiy': {'en': 'Wiyot'},
    'wja': {'en': 'Waja'},
    'wji': {'en': 'Warji'},
    'wka': {'en': "Kw'adza"},
    'wkb': {'en': 'Kumbaran'},
    'wkd': {'en': 'Wakde'},
    'wkl': {'en': 'Kalanadi'},
    'wkr': {'en': 'Keerray-Woorroong'},
    'wku': {'en': 'Kunduvadi'},
    'wkw': {'en': 'Wakawaka'},
    'wky': {'en': 'Wangkayutyuru'},
    'wla': {'en': 'Walio'},
    'wlc': {'en': 'Mwali Comorian'},
    'wle': {'en': 'Wolane'},
    'wlg': {'en': 'Kunbarlang'},
    'wlh': {'en': 'Welaun'},
    'wli': {'en': 'Waioli'},
    'wlk': {'en': 'Wailaki'},
    'wll': {'en': 'Wali (Sudan)'},
    'wlm': {'en': 'Middle Welsh'},
    'wlo': {'en': 'Wolio'},
    'wlr': {'en': 'Wailapa'},
    'wls': {'en': 'Wallisian'},
    'wlu': {'en': 'Wuliwuli'},
    'wlv': {'en': 'Wichí Lhamtés Vejoz'},
    'wlw': {'en': 'Walak'},
    'wlx': {'en': 'Wali (Ghana)'},
    'wly': {'en': 'Waling'},
    'wma': {'en': 'Mawa (Nigeria)'},
    'wmb': {'en': 'Wambaya'},
    'wmc': {'en': 'Wamas'},
    'wmd': {'en': 'Mamaindé'},
    'wme': {'en': 'Wambule'},
    'wmg': {'en': 'Western Minyag'},
    'wmh': {'en': "Waima'a"},
    'wmi': {'en': 'Wamin'},
    'wmm': {'en': 'Maiwa (Indonesia)'},
    'wmn': {'en': 'Waamwang'},
    'wmo': {'en': 'Wom (Papua New Guinea)'},
    'wms': {'en': 'Wambon'},
    'wmt': {'en': 'Walmajarri'},
    'wmw': {'en': 'Mwani'},
    'wmx': {'en': 'Womo'},
    'wnb': {'en': 'Wanambre'},
    'wnc': {'en': 'Wantoat'},
    'wnd': {'en': 'Wandarang'},
    'wne': {'en': 'Waneci'},
    'wng': {'en': 'Wanggom'},
    'wni': {'en': 'Ndzwani Comorian'},
    'wnk': {'en': 'Wanukaka'},
    'wnm': {'en': 'Wanggamala'},
    'wnn': {'en': 'Wunumara'},
    'wno': {'en': 'Wano'},
    'wnp': {'en': 'Wanap'},
    'wnu': {'en': 'Usan'},
    'wnw': {'en': 'Wintu'},
    'wny': {'en': 'Wanyi'},
    'wo': {'af': 'Wolof', 'ak': 'Wolɔfo Kasa', 'am': 'ዎሎፍኛ', 'ar': 'الولوفية', 'as': 'ৱোলাফ', 'ast': 'wolof', 'az': 'volof', 'az-Cyrl': 'волоф', 'az-Latn': 'volof', 'bal-Latn': 'Wolop', 'be': 'валоф', 'be-tarask': 'валоф', 'bg': 'волоф', 'bgn': 'ولوفی', 'bn': 'ওলোফ', 'br': 'wolof', 'brx': 'वल’फ', 'bs': 'volof', 'bs-Cyrl': 'волоф', 'bs-Latn': 'volof', 'ca': 'wòlof', 'ca-ES-valencia': 'wòlof', 'ccp': '𑄤𑄃𑄮𑄣𑄮𑄜𑄴', 'ce': 'волоф', 'chr': 'ᏬᎶᏫ', 'ckb': 'وۆلۆف', 'cs': 'wolofština', 'cy': 'Woloff', 'da': 'wolof', 'de': 'Wolof', 'dsb': 'wolof', 'dz': 'ཝོ་ལོཕ་ཁ', 'ee': 'wolofgbe', 'el': 'Γουόλοφ', 'el-polyton': 'Γουόλοφ', 'en': 'Wolof', 'eo': 'volofa', 'es': 'wólof', 'et': 'volofi', 'eu': 'wolofera', 'fa': 'ولوفی', 'ff-Adlm': '𞤏𞤮𞤤𞤮𞤬𞤪𞤫', 'fi': 'wolof', 'fil': 'Wolof', 'fo': 'wolof', 'fr': 'wolof', 'fur': 'wolof', 'fy': 'Wolof', 'ga': 'Volaifis', 'gd': 'Wolof', 'gl': 'wólof', 'gsw': 'Wolof', 'gu': 'વોલોફ', 'ha': 'Wolof', 'he': 'וולוף', 'hi': 'वोलोफ़', 'hi-Latn': 'Wolof', 'hr': 'volof', 'hsb': 'wolof', 'ht': 'wolof', 'hu': 'volof', 'hy': 'վոլոֆ', 'ia': 'wolof', 'id': 'Wolof', 'ig': 'Wolof', 'is': 'volof', 'it': 'wolof', 'ja': 'ウォロフ語', 'jv': 'Wolof', 'ka': 'ვოლოფური', 'kea': 'uolof', 'kgp': 'vorofe', 'kk': 'волоф тілі', 'kk-Cyrl': 'волоф тілі', 'km': 'វូឡុហ្វ', 'kn': 'ವೋಲೋಫ್', 'ko': '월로프어', 'kok': 'वोलोफ', 'kok-Deva': 'वोलोफ', 'ks': 'وولوف', 'ks-Arab': 'وولوف', 'ksh': 'Woloff', 'ku': 'wolofî', 'ky': 'уолофчо', 'lb': 'Wolof', 'lkt': 'Wolof Iyápi', 'lld': 'wolof', 'lo': 'ວໍລອບ', 'lrc': 'ڤولوف', 'lt': 'volofų', 'lv': 'volofu', 'mai': 'बोलोफ', 'mi': 'Warawhe', 'mk': 'волофски', 'ml': 'വൊളോഫ്', 'mn': 'волоф', 'mr': 'वोलोफ', 'ms': 'Wolof', 'mt': 'Wolof', 'my': 'ဝူလိုဖ်', 'mzn': 'وولفی', 'nb': 'wolof', 'ne': 'वुलुफ', 'nl': 'Wolof', 'nn': 'wolof', 'no': 'wolof', 'nqo': 'ߥߟߐߝߐߞߊ߲', 'or': 'ୱୋଲଫ୍', 'pa': 'ਵੋਲੋਫ', 'pa-Guru': 'ਵੋਲੋਫ', 'pcm': 'Wólof', 'pl': 'wolof', 'ps': 'ولوف', 'pt': 'uolofe', 'qu': 'Wolof Simi', 'rm': 'wolof', 'ro': 'wolof', 'ru': 'волоф', 'sc': 'wolof', 'sd': 'وولوف', 'sd-Arab': 'وولوف', 'si': 'වොලොෆ්', 'sk': 'wolofčina', 'sl': 'volofščina', 'smn': 'wolofkielâ', 'so': 'Woolof', 'sq': 'uolofisht', 'sr': 'волоф', 'sr-Cyrl': 'волоф', 'sr-Latn': 'volof', 'sv': 'wolof', 'sw': 'Kiwolofu', 'ta': 'ஓலோஃப்', 'te': 'ఉలూఫ్', 'tg': 'волоф', 'th': 'โวลอฟ', 'ti': 'ዎሎፍ', 'tk': 'wolof dili', 'to': 'lea fakaʻuolofo', 'tr': 'Volofça', 'tt': 'волоф', 'ug': 'ۋولوفچە', 'uk': 'волоф', 'ur': 'وولوف', 'uz': 'volof', 'uz-Cyrl': 'волофча', 'uz-Latn': 'volof', 'vec': 'wolof', 'vi': 'Tiếng Wolof', 'wae': 'Wolof', 'wo': 'Wolof', 'xnr': 'वोलोफ़', 'yo': 'Wọ́lọ́ọ̀fù', 'yrl': 'worofi', 'yue': '沃洛夫文', 'yue-Hans': '沃洛夫文', 'yue-Hant': '沃洛夫文', 'zh': '沃洛夫语', 'zh-Hans': '沃洛夫语', 'zh-Hant': '沃洛夫文', 'zu': 'isi-Wolof'},
    'woa': {'en': 'Kuwema'},
    'wob': {'en': 'Wè Northern'},
    'woc': {'en': 'Wogeo'},
    'wod': {'en': 'Wolani'},
    'woe': {'en': 'Woleaian'},
    'wof': {'en': 'Gambian Wolof'},
    'wog': {'en': 'Wogamusin'},
    'woi': {'en': 'Kamang'},
    'wok': {'en': 'Longto'},
    'wom': {'en': 'Wom (Nigeria)'},
    'won': {'en': 'Wongo'},
    'woo': {'en': 'Manombai'},
    'wor': {'en': 'Woria'},
    'wos': {'en': 'Hanga Hundi'},
    'wow': {'en': 'Wawonii'},
    'woy': {'en': 'Weyto'},
    'wpc': {'en': 'Maco'},
    'wra': {'en': 'Warapu'},
    'wrb': {'en': 'Waluwarra'},
    'wrd': {'en': 'Warduji'},
    'wrg': {'en': 'Warungu'},
    'wrh': {'en': 'Wiradjuri'},
    'wri': {'en': 'Wariyangga'},
    'wrk': {'en': 'Garrwa'},
    'wrl': {'en': 'Warlmanpa'},
    'wrm': {'en': 'Warumungu'},
    'wrn': {'en': 'Warnang'},
    'wro': {'en': 'Worrorra'},
    'wrp': {'en': 'Waropen'},
    'wrr': {'en': 'Wardaman'},
    'wrs': {'en': 'Waris'},
    'wru': {'en': 'Waru'},
    'wrv': {'en': 'Waruna'},
    'wrw': {'en': 'Gugu Warra'},
    'wrx': {'en': 'Wae Rana'},
    'wry': {'en': 'Merwari'},
    'wrz': {'en': 'Waray (Australia)'},
    'wsa': {'en': 'Warembori'},
    'wsg': {'en': 'Adilabad Gondi'},
    'wsi': {'en': 'Wusi'},
    'wsk': {'en': 'Waskia'},
    'wsr': {'en': 'Owenia'},
    'wss': {'en': 'Wasa'},
    'wsu': {'en': 'Wasu'},
    'wsv': {'en': 'Wotapuri-Katarqalai'},
    'wtf': {'en': 'Watiwa'},
    'wth': {'en': 'Wathawurrung'},
    'wti': {'en': 'Berta'},
    'wtk': {'en': 'Watakataui'},
    'wtm': {'en': 'Mewati'},
    'wtw': {'en': 'Wotu'},
    'wua': {'en': 'Wikngenchera'},
    'wub': {'en': 'Wunambal'},
    'wud': {'en': 'Wudu'},
    'wuh': {'en': 'Wutunhua'},
    'wul': {'en': 'Silimo'},
    'wum': {'en': 'Wumbvu'},
    'wun': {'en': 'Bungu'},
    'wur': {'en': 'Wurrugu'},
    'wut': {'en': 'Wutung'},
    'wuu': {'af': 'Wu-Sjinees', 'am': 'ዉ ቻይንኛ', 'ar': 'الوو الصينية', 'as': 'ৱু চাইনিজ', 'ast': 'chinu wu', 'az': 'vu', 'az-Latn': 'vu', 'be': 'ву', 'be-tarask': 'ву', 'bg': 'ву китайски', 'bn': 'উ চীনা', 'br': 'sinaeg Wu', 'brx': 'वु चाइनीज', 'bs': 'Wu kineski', 'bs-Cyrl': 'ву кинески', 'bs-Latn': 'Wu kineski', 'ca': 'xinès wu', 'ca-ES-valencia': 'xinès wu', 'ccp': '𑄤𑄌𑄨𑄚', 'chr': 'Ꮽ ᏓᎶᏂᎨᏍᏛ', 'cs': 'čínština (dialekty Wu)', 'cy': 'Wu Tsieineaidd', 'da': 'wu-kinesisk', 'de': 'Wu', 'dsb': 'wu chinšćina', 'el': 'Κινεζικά Γου', 'el-polyton': 'Κινεζικά Γου', 'en': 'Wu Chinese', 'es': 'chino wu', 'et': 'uu', 'eu': 'wu txinera', 'fa': 'وو چینی', 'ff-Adlm': '𞤏𞤵𞥅𞤪𞤫 𞤅𞤭𞥅𞤲', 'fi': 'wu-kiina', 'fil': 'Wu Chinese', 'fo': 'wu kinesiskt', 'fr': 'wu', 'ga': 'Sínis Wu', 'gd': 'Wu', 'gl': 'chinés wu', 'gu': 'વુ ચાઈનીઝ', 'ha': 'Sinancin Wu', 'he': 'סינית וו', 'hi': 'वू चीनी', 'hi-Latn': 'Wu Chinese', 'hr': 'wu kineski', 'hsb': 'wu chinšćina', 'ht': 'wu', 'hu': 'wu kínai', 'hy': 'վու չինարեն', 'ia': 'wu', 'id': 'Wu Tionghoa', 'ig': 'Wu Chinese', 'is': 'wu-kínverska', 'it': 'wu', 'ja': '呉語', 'jv': 'Tyonghwa Wu', 'ka': 'ვუ', 'kgp': 'vu', 'kk': 'қытай тілі (У)', 'kk-Cyrl': 'қытай тілі (У)', 'km': 'អ៊ូចិន', 'kn': 'ವು ಚೈನೀಸ್', 'ko': '우어', 'kok': 'व्हु चिनी', 'kok-Deva': 'व्हु चिनी', 'ku': 'çînîya wuyî', 'ky': '"У" диалектинде (Кытай)', 'lb': 'Wu-Chinesesch', 'lo': 'ຈີນອູ', 'lt': 'kinų kalbos vu tarmė', 'lv': 'vu ķīniešu', 'mai': 'ऊ चीनी', 'mi': 'Hainamana Wū', 'mk': 'ву', 'ml': 'വു ചൈനീസ്', 'mn': 'хятад, ву хэл', 'mr': 'व्हू चिनी', 'ms': 'Cina Wu', 'my': 'ဝူ တရုတ်', 'nb': 'wu', 'ne': 'ऊ चिनियाँ', 'nl': 'Wuyu', 'nn': 'wu', 'no': 'wu', 'or': 'ୱୁ ଚାଇନିଜ', 'pa': 'ਚੀਨੀ ਵੂ', 'pa-Guru': 'ਚੀਨੀ ਵੂ', 'pcm': 'Wu Chainiz', 'pl': 'wu', 'ps': 'وو چینایی', 'pt': 'wu', 'qu': 'Wu Chino', 'ro': 'chineză wu', 'ru': 'у', 'sc': 'wu', 'sd': 'وو چيني', 'sd-Arab': 'وو چيني', 'si': 'වූ චයිනිස්', 'sk': 'čínština (wu)', 'sl': 'wu-kitajščina', 'so': 'Wu Chinese', 'sq': 'kinezishte vu', 'sr': 'ву кинески', 'sr-Cyrl': 'ву кинески', 'sr-Latn': 'vu kineski', 'sv': 'wu', 'sw': 'Kichina cha Wu', 'ta': 'வூ சீனம்', 'te': 'వు చైనీస్', 'th': 'จีนอู๋', 'ti': 'ቻይናዊ ዉ', 'tk': 'u hytaý dili', 'to': 'lea fakasiaina-uū', 'tr': 'Wu Çincesi', 'uk': 'китайська уська', 'ur': 'وو چائینیز', 'uz': 'vu xitoy', 'uz-Latn': 'vu xitoy', 'vec': 'wu', 'vi': 'Tiếng Ngô', 'xnr': 'वू चीनी', 'yo': 'Èdè Wu ti Saina', 'yrl': 'wurapuki', 'yue': '吳語', 'yue-Hans': '吴语', 'yue-Hant': '吳語', 'zh': '吴语', 'zh-Hans': '吴语', 'zh-Hant': '吳語', 'zu': 'isi-Wu Chinese'},
    'wuv': {'en': 'Wuvulu-Aua'},
    'wux': {'en': 'Wulna'},
    'wuy': {'en': 'Wauyai'},
    'wwa': {'en': 'Waama'},
    'wwb': {'en': 'Wakabunga'},
    'wwo': {'en': 'Wetamut'},
    'wwr': {'en': 'Warrwa'},
    'www': {'en': 'Wawa'},
    'wxa': {'en': 'Waxianghua'},
    'wxw': {'en': 'Wardandi'},
    'wya': {'en': 'Wyandot'},
    'wyb': {'en': 'Wangaaybuwan-Ngiyambaa'},
    'wyi': {'en': 'Woiwurrung'},
    'wym': {'en': 'Wymysorys'},
    'wyr': {'en': 'Wayoró'},
    'wyy': {'en': 'Western Fijian'},
    'xaa': {'en': 'Andalusian Arabic'},
    'xab': {'en': 'Sambe'},
    'xac': {'en': 'Kachari'},
    'xad': {'en': 'Adai'},
    'xae': {'en': 'Aequian'},
    'xag': {'en': 'Aghwan'},
    'xai': {'en': 'Kaimbé'},
    'xaj': {'en': 'Ararandewára'},
    'xak': {'en': 'Máku'},
    'xal': {'af': 'Kalmyk', 'am': 'ካልማይክ', 'ar': 'الكالميك', 'as': 'কাল্মিক', 'ast': 'calmuco', 'az': 'kalmık', 'az-Cyrl': 'калмык', 'az-Latn': 'kalmık', 'be': 'калмыцкая', 'be-tarask': 'калмыцкая', 'bg': 'калмик', 'bgn': 'کالمیکی', 'bn': 'কাল্মাইক', 'br': 'kalmouk', 'brx': 'कालमीक', 'bs': 'kalmik', 'bs-Cyrl': 'калмик', 'bs-Latn': 'kalmik', 'ca': 'calmuc', 'ca-ES-valencia': 'calmuc', 'ccp': '𑄇𑄣𑄴𑄟𑄳𑄆𑄧𑄇𑄴', 'ce': 'гӀалмакхойн', 'chr': 'ᎧᎳᎻᎧ', 'ckb': 'کالمیک', 'cs': 'kalmyčtina', 'cy': 'Calmyceg', 'da': 'kalmyk', 'de': 'Kalmückisch', 'dsb': 'kalmykšćina', 'el': 'Καλμίκ', 'el-polyton': 'Καλμίκ', 'en': 'Kalmyk', 'es': 'kalmyk', 'et': 'kalmõki', 'eu': 'kalmykera', 'fa': 'قلموقی', 'ff-Adlm': '𞤑𞤢𞤤𞤥𞤵𞤳𞤪𞤫', 'fi': 'kalmukki', 'fil': 'Kalmyk', 'fo': 'kalmyk', 'fr': 'kalmouk', 'fy': 'Kalmyk', 'ga': 'Cailmícis', 'gd': 'Kalmyk', 'gl': 'calmuco', 'gsw': 'Kalmückisch', 'gu': 'કાલ્મિક', 'ha': 'Kalmyk', 'he': 'קלמיקית', 'hi': 'काल्मिक', 'hi-Latn': 'Kalmyk', 'hr': 'kalmyk', 'hsb': 'kalmykšćina', 'ht': 'kalmouk', 'hu': 'kalmük', 'hy': 'կալմիկերեն', 'ia': 'calmuco', 'id': 'Kalmuk', 'ig': 'Kalmyk', 'is': 'kalmúkska', 'it': 'kalmyk', 'ja': 'カルムイク語', 'jv': 'Kalmik', 'ka': 'ყალმუხური', 'kgp': 'karmĩg', 'kk': 'қалмақ тілі', 'kk-Cyrl': 'қалмақ тілі', 'km': 'កាលមីគ', 'kn': 'ಕಲ್ಮೈಕ್', 'ko': '칼미크어', 'kok': 'कालमायक', 'kok-Deva': 'कालमायक', 'ks': 'کالمِک', 'ks-Arab': 'کالمِک', 'ksh': 'Kalmükkesch', 'ku': 'kalmîkî', 'ky': 'калмыкча', 'lb': 'Kalmückesch', 'lo': 'ການມິກ', 'lt': 'kalmukų', 'lv': 'kalmiku', 'mai': 'कलमिक', 'mi': 'Karamiki', 'mk': 'калмички', 'ml': 'കാൽമിക്', 'mn': 'халимаг', 'mr': 'काल्मिक', 'ms': 'Kalmyk', 'mt': 'Kalmyk', 'my': 'ကာလ်မိုက်', 'nb': 'kalmukkisk', 'ne': 'काल्मिक', 'nl': 'Kalmuks', 'nn': 'kalmykisk', 'no': 'kalmukkisk', 'or': 'କାଲ୍ମୀକ୍', 'pa': 'ਕਾਲਮਿਕ', 'pa-Guru': 'ਕਾਲਮਿਕ', 'pcm': 'Kalmik', 'pl': 'kałmucki', 'ps': 'کالمک', 'pt': 'kalmyk', 'qu': 'Kalmyk Simi', 'rm': 'kalmuk', 'ro': 'calmucă', 'ru': 'калмыцкий', 'sc': 'calmucu', 'sd': 'ڪيلمڪ', 'sd-Arab': 'ڪيلمڪ', 'si': 'කල්මික්', 'sk': 'kalmyčtina', 'sl': 'kalmiščina', 'smn': 'kalmukkielâ', 'so': 'Kalmyk', 'sq': 'kalmikisht', 'sr': 'калмички', 'sr-Cyrl': 'калмички', 'sr-Latn': 'kalmički', 'sv': 'kalmuckiska', 'sw': 'Kikalmyk', 'ta': 'கல்மிக்', 'te': 'కల్మిక్', 'th': 'คัลมืยค์', 'ti': 'ካልምይክ', 'tk': 'galmyk dili', 'to': 'lea fakakalimiki', 'tr': 'Kalmıkça', 'ug': 'قالماقچە', 'uk': 'калмицька', 'ur': 'کالمیک', 'uz': 'qalmoq', 'uz-Latn': 'qalmoq', 'vec': 'kalmik', 'vi': 'Tiếng Kalmyk', 'xnr': 'काल्मिक', 'yo': 'Èdè Kalimi', 'yrl': 'kaumiki', 'yue': '卡爾梅克文', 'yue-Hans': '卡尔梅克文', 'yue-Hant': '卡爾梅克文', 'zh': '卡尔梅克语', 'zh-Hans': '卡尔梅克语', 'zh-Hant': '卡爾梅克文', 'zu': 'isi-Kalmyk'},
    'xam': {'en': 'ǀXam'},
    'xan': {'en': 'Xamtanga'},
    'xao': {'en': 'Khao'},
    'xap': {'en': 'Apalachee'},
    'xaq': {'en': 'Aquitanian'},
    'xar': {'en': 'Karami'},
    'xas': {'en': 'Kamas'},
    'xat': {'en': 'Katawixi'},
    'xau': {'en': 'Kauwera'},
    'xav': {'en': 'Xavánte'},
    'xaw': {'en': 'Kawaiisu'},
    'xay': {'en': 'Kayan Mahakam'},
    'xba': {'en': 'Kamba (Brazil)'},
    'xbb': {'en': 'Lower Burdekin'},
    'xbc': {'en': 'Bactrian'},
    'xbd': {'en': 'Bindal'},
    'xbe': {'en': 'Bigambal'},
    'xbg': {'en': 'Bunganditj'},
    'xbi': {'en': 'Kombio'},
    'xbj': {'en': 'Birrpayi'},
    'xbm': {'en': 'Middle Breton'},
    'xbn': {'en': 'Kenaboi'},
    'xbo': {'en': 'Bolgarian'},
    'xbp': {'en': 'Bibbulman'},
    'xbr': {'en': 'Kambera'},
    'xbw': {'en': 'Kambiwá'},
    'xbx': {'en': 'Kabixí'},
    'xby': {'en': 'Batjala'},
    'xcb': {'en': 'Cumbric'},
    'xcc': {'en': 'Camunic'},
    'xce': {'en': 'Celtiberian'},
    'xcg': {'en': 'Cisalpine Gaulish'},
    'xch': {'en': 'Chemakum'},
    'xcl': {'en': 'Classical Armenian'},
    'xcm': {'en': 'Comecrudo'},
    'xcn': {'en': 'Cotoname'},
    'xco': {'en': 'Chorasmian'},
    'xcr': {'en': 'Carian'},
    'xct': {'en': 'Classical Tibetan'},
    'xcu': {'en': 'Curonian'},
    'xcv': {'en': 'Chuvantsy'},
    'xcw': {'en': 'Coahuilteco'},
    'xcy': {'en': 'Cayuse'},
    'xda': {'en': 'Darkinyung'},
    'xdc': {'en': 'Dacian'},
    'xdk': {'en': 'Dharuk'},
    'xdm': {'en': 'Edomite'},
    'xdo': {'en': 'Kwandu'},
    'xdy': {'en': 'Malayic Dayak'},
    'xeb': {'en': 'Eblan'},
    'xed': {'en': 'Hdi'},
    'xeg': {'en': 'ǁXegwi'},
    'xel': {'en': 'Kelo'},
    'xem': {'en': 'Kembayan'},
    'xep': {'en': 'Epi-Olmec'},
    'xer': {'en': 'Xerénte'},
    'xes': {'en': 'Kesawai'},
    'xet': {'en': 'Xetá'},
    'xeu': {'en': 'Keoru-Ahia'},
    'xfa': {'en': 'Faliscan'},
    'xga': {'en': 'Galatian'},
    'xgb': {'en': 'Gbin'},
    'xgd': {'en': 'Gudang'},
    'xgf': {'en': 'Gabrielino-Fernandeño'},
    'xgg': {'en': 'Goreng'},
    'xgi': {'en': 'Garingbal'},
    'xgl': {'en': 'Galindan'},
    'xgm': {'en': 'Dharumbal'},
    'xgn': {'en': 'Mongolian languages'},
    'xgn-pro': {'en': 'Proto-Mongolic'},
    'xgr': {'en': 'Garza'},
    'xgu': {'en': 'Unggumi'},
    'xgw': {'en': 'Guwa'},
    'xh': {'af': 'Xhosa', 'ak': 'Hosa Kasa', 'am': 'ዞሳኛ', 'ar': 'الخوسا', 'as': 'হোছা', 'ast': 'xhosa', 'az': 'xosa', 'az-Cyrl': 'хоса', 'az-Latn': 'xosa', 'bal-Latn': 'Khushá', 'be': 'коса', 'be-tarask': 'коса', 'bg': 'кхоса', 'bgn': 'خوسی', 'bn': 'জোসা', 'br': 'xhosa', 'brx': 'ख’सा', 'bs': 'hosa', 'bs-Cyrl': 'коса', 'bs-Latn': 'hosa', 'ca': 'xosa', 'ca-ES-valencia': 'xosa', 'ccp': '𑄎𑄮𑄥', 'ce': 'коса', 'chr': 'ᏠᏌ', 'ckb': 'سسوسا', 'cs': 'xhoština', 'cy': 'Xhosa', 'da': 'xhosa', 'de': 'Xhosa', 'dsb': 'xhosa', 'dz': 'ཞོ་ས་ཁ', 'ee': 'xhosagbe', 'el': 'Κόσα', 'el-polyton': 'Κόσα', 'en': 'Xhosa', 'eo': 'ksosa', 'es': 'xhosa', 'et': 'koosa', 'eu': 'xhosera', 'fa': 'خوسایی', 'ff-Adlm': '𞤑𞤮𞥅𞤧𞤢𞥄𞤪𞤫', 'fi': 'xhosa', 'fil': 'Xhosa', 'fo': 'xhosa', 'fr': 'xhosa', 'fur': 'xhosa', 'fy': 'Xhosa', 'ga': 'Cóisis', 'gd': 'Xhosa', 'gl': 'xhosa', 'gsw': 'Xhosa', 'gu': 'ખોસા', 'ha': 'Xhosa', 'he': 'קוסה', 'hi': 'ख़ोसा', 'hi-Latn': 'Xhosa', 'hr': 'xhosa', 'hsb': 'xhosa', 'ht': 'xhosa', 'hu': 'xhosza', 'hy': 'քոսա', 'ia': 'xhosa', 'id': 'Xhosa', 'ig': 'Xhosa', 'is': 'sósa', 'it': 'xhosa', 'ja': 'コサ語', 'jv': 'Xhosa', 'ka': 'ქჰოსა', 'kea': 'koza', 'kgp': 'sosa', 'kk': 'кхоса тілі', 'kk-Cyrl': 'кхоса тілі', 'km': 'ឃសា', 'kn': 'ಕ್ಸೋಸ', 'ko': '코사어', 'kok': 'खोसा', 'kok-Deva': 'खोसा', 'ks': 'کھوسا', 'ks-Arab': 'کھوسا', 'ksh': 'Isi-Khohsa', 'ku': 'xosayî', 'ky': 'косача', 'lb': 'Xhosa', 'lkt': 'Xhosa Iyápi', 'lld': 'xhosa', 'lo': 'ໂຮຊາ', 'lrc': 'خوٙسا', 'lt': 'kosų', 'lv': 'khosu', 'mai': 'खोसा', 'mi': 'Tōha', 'mk': 'коса', 'ml': 'ഖോസ', 'mn': 'хоса', 'mr': 'खोसा', 'ms': 'Xhosa', 'mt': 'Xhosa', 'my': 'ဇိုစာ', 'mzn': 'خوسا', 'nb': 'xhosa', 'ne': 'खोसा', 'nl': 'Xhosa', 'nn': 'xhosa', 'no': 'xhosa', 'nqo': 'ߛߏߖ߭ߊߞߊ߲', 'om': 'Afaan Xhosa', 'or': 'ଖୋସା', 'pa': 'ਖੋਸਾ', 'pa-Guru': 'ਖੋਸਾ', 'pcm': 'Kọ́sa', 'pl': 'khosa', 'ps': 'خوسا', 'pt': 'xhosa', 'qu': 'Isixhosa Simi', 'rm': 'xhosa', 'ro': 'xhosa', 'ru': 'коса', 'rw': 'Inyehawusa', 'sc': 'xhosa', 'sd': 'زھوسا', 'sd-Arab': 'زھوسا', 'si': 'ශෝසා', 'sk': 'xhoština', 'sl': 'koščina', 'smn': 'xhosakielâ', 'so': 'Xhosa', 'sq': 'xhosaisht', 'sr': 'коса', 'sr-Cyrl': 'коса', 'sr-Latn': 'kosa', 'sv': 'xhosa', 'sw': 'Kixhosa', 'ta': 'ஹோசா', 'te': 'షోసా', 'th': 'คะห์โอซา', 'ti': 'ኮሳ', 'tk': 'kosa dili', 'to': 'lea fakatōsa', 'tr': 'Zosa dili', 'ug': 'خوساچە', 'uk': 'кхоса', 'ur': 'ژوسا', 'uz': 'kxosa', 'uz-Cyrl': 'хоса', 'uz-Latn': 'kxosa', 'vec': 'xhosa', 'vi': 'Tiếng Xhosa', 'wae': 'Xhosa', 'xh': 'IsiXhosa', 'xnr': 'ख़ोसा', 'yo': 'Èdè Xhosa', 'yrl': 'xosa', 'yue': '科薩文', 'yue-Hans': '科萨文', 'yue-Hant': '科薩文', 'zh': '科萨语', 'zh-Hans': '科萨语', 'zh-Hant': '科薩文', 'zu': 'isiXhosa'},
    'xha': {'en': 'Harami'},
    'xhc': {'en': 'Hunnic'},
    'xhd': {'en': 'Hadrami'},
    'xhe': {'en': 'Khetrani'},
    'xhr': {'en': 'Hernican'},
    'xht': {'en': 'Hattic'},
    'xhu': {'en': 'Hurrian'},
    'xhv': {'en': 'Khua'},
    'xia': {'en': 'Xiandao'},
    'xib': {'en': 'Iberian'},
    'xii': {'en': 'Xiri'},
    'xil': {'en': 'Illyrian'},
    'xin': {'en': 'Xinca'},
    'xip': {'en': 'Xipináwa'},
    'xir': {'en': 'Xiriâna'},
    'xis': {'en': 'Kisan'},
    'xiv': {'en': 'Indus Valley Language'},
    'xiy': {'en': 'Xipaya'},
    'xjb': {'en': 'Minjungbal'},
    'xjt': {'en': 'Jaitmatang'},
    'xka': {'en': 'Kalkoti'},
    'xkb': {'en': 'Northern Nago'},
    'xkc': {'en': "Kho'ini"},
    'xkd': {'en': 'Mendalam Kayan'},
    'xke': {'en': 'Kereho'},
    'xkf': {'en': 'Khengkha'},
    'xkg': {'en': 'Kagoro'},
    'xkh': {'en': 'Karahawyana'},
    'xki': {'en': 'Kenyan Sign Language'},
    'xkj': {'en': 'Kajali'},
    'xkk': {'en': "Kaco'"},
    'xkl': {'en': 'Mainstream Kenyah'},
    'xkn': {'en': 'Kayan River Kayan'},
    'xko': {'en': 'Kiorr'},
    'xkp': {'en': 'Kabatei'},
    'xkq': {'en': 'Koroni'},
    'xkr': {'en': 'Xakriabá'},
    'xks': {'en': 'Kumbewaha'},
    'xkt': {'en': 'Kantosi'},
    'xku': {'en': 'Kaamba'},
    'xkv': {'en': 'Kgalagadi'},
    'xkw': {'en': 'Kembra'},
    'xkx': {'en': 'Karore'},
    'xky': {'en': "Uma' Lasan"},
    'xkz': {'en': 'Kurtokha'},
    'xla': {'en': 'Kamula'},
    'xlb': {'en': 'Loup B'},
    'xlc': {'en': 'Lycian'},
    'xld': {'en': 'Lydian'},
    'xle': {'en': 'Lemnian'},
    'xlg': {'en': 'Ligurian (Ancient)'},
    'xli': {'en': 'Liburnian'},
    'xln': {'en': 'Alanic'},
    'xlo': {'en': 'Loup A'},
    'xlp': {'en': 'Lepontic'},
    'xls': {'en': 'Lusitanian'},
    'xlu': {'en': 'Cuneiform Luwian'},
    'xly': {'en': 'Elymian'},
    'xma': {'en': 'Mushungulu'},
    'xmb': {'en': 'Mbonga'},
    'xmc': {'en': 'Makhuwa-Marrevone'},
    'xmd': {'en': 'Mbudum'},
    'xme': {'en': 'Median'},
    'xmf': {'ast': 'mingrelianu', 'br': 'megreleg', 'ca': 'mingrelià', 'ca-ES-valencia': 'mingrelià', 'cs': 'mingrelština', 'de': 'Mingrelisch', 'en': 'Mingrelian', 'et': 'megreli', 'fi': 'mingreli', 'fr': 'mingrélien', 'hi-Latn': 'Mingrelian', 'ht': 'mingrélien', 'it': 'mengrelio', 'ja': 'メグレル語', 'lb': 'Mingrelesch Sprooch', 'lt': 'megrelų', 'mk': 'мегрелски', 'nb': 'mingrelsk', 'ne': 'मिनग्रेलियाली', 'nl': 'Mingreels', 'nn': 'mingrelsk', 'no': 'mingrelsk', 'pl': 'megrelski', 'sv': 'mingrelianska', 'th': 'เมเกรเลีย', 'to': 'lea fakamingilelia', 'tr': 'Megrelce', 'yue': '明格列爾文', 'yue-Hans': '明格列尔文', 'yue-Hant': '明格列爾文', 'zh-Hant': '明格列爾文'},
    'xmg': {'en': 'Mengaka'},
    'xmh': {'en': 'Kugu-Muminh'},
    'xmj': {'en': 'Majera'},
    'xmk': {'en': 'Ancient Macedonian'},
    'xml': {'en': 'Malaysian Sign Language'},
    'xmm': {'en': 'Manado Malay'},
    'xmn': {'en': 'Manichaean Middle Persian'},
    'xmo': {'en': 'Morerebi'},
    'xmp': {'en': "Kuku-Mu'inh"},
    'xmq': {'en': 'Kuku-Mangk'},
    'xmr': {'en': 'Meroitic'},
    'xms': {'en': 'Moroccan Sign Language'},
    'xmt': {'en': 'Matbat'},
    'xmu': {'en': 'Kamu'},
    'xmv': {'en': 'Antankarana Malagasy'},
    'xmw': {'en': 'Tsimihety Malagasy'},
    'xmx': {'en': 'Maden'},
    'xmy': {'en': 'Mayaguduna'},
    'xmz': {'en': 'Mori Bawah'},
    'xna': {'en': 'Ancient North Arabian'},
    'xnb': {'en': 'Kanakanabu'},
    'xnd': {'en': 'Na-Dene languages'},
    'xnd-pro': {'en': 'Proto-Na-Dene'},
    'xng': {'en': 'Middle Mongolian'},
    'xnh': {'en': 'Kuanhua'},
    'xni': {'en': 'Ngarigu'},
    'xnj': {'en': 'Ngoni (Tanzania)'},
    'xnk': {'en': 'Nganakarti'},
    'xnm': {'en': 'Ngumbarl'},
    'xnn': {'en': 'Northern Kankanay'},
    'xno': {'en': 'Anglo-Norman'},
    'xnq': {'en': 'Ngoni (Mozambique)'},
    'xnr': {'af': 'Kangri', 'ak': 'Kangri', 'am': 'ካንጋሪ', 'ar': 'كانغري', 'as': 'কাংগৰি', 'az': 'kanqri', 'az-Latn': 'kanqri', 'bal-Latn': 'Kangri', 'be': 'кангры', 'be-tarask': 'кангры', 'bg': 'кангри', 'bn': 'কাংরি', 'bs': 'kangri', 'bs-Latn': 'kangri', 'ca': 'kangri', 'ca-ES-valencia': 'kangri', 'chr': 'ᎧᏈᎢ', 'cs': 'kángrí', 'cy': 'Kangri', 'da': 'kangri', 'de': 'Kangri', 'dsb': 'kangrišćina', 'el': 'Κάνγκρι', 'el-polyton': 'Κάνγκρι', 'en': 'Kangri', 'es': 'kangri', 'et': 'kangri', 'eu': 'kangrera', 'fa': 'کانگری', 'fi': 'kangri', 'fil': 'Kangri', 'fo': 'kangri', 'fr': 'kangri', 'ga': 'Kangri', 'gd': 'Kangri', 'gl': 'kangri', 'gu': 'કંગરી', 'ha': 'Kangri', 'he': 'קאנגרי', 'hi': 'कांगड़ी', 'hi-Latn': 'Kangri', 'hr': 'kangri', 'hsb': 'kangrišćina', 'ht': 'kangri', 'hu': 'kangri', 'hy': 'կանգրի', 'ia': 'kangri', 'id': 'Kangri', 'ig': 'Kangri', 'is': 'kangrí', 'it': 'kangri', 'ja': 'カーングリー語', 'jv': 'Kangri', 'ka': 'კანგრი', 'kk': 'кангри тілі', 'kk-Cyrl': 'кангри тілі', 'km': 'ខែងគ្រី', 'kn': 'ಕಂಗ್ರಿ', 'ko': '캉리어', 'kok': 'कांग्री', 'kok-Deva': 'कांग्री', 'ku': 'kangrî', 'kxv': 'kangri', 'kxv-Latn': 'kangri', 'ky': 'кангри', 'lo': 'ຄັງຣີ', 'lt': 'kangri', 'lv': 'kangri', 'mk': 'кангри', 'ml': 'കാങ്ടി', 'mn': 'кангри', 'mr': 'कांगरी', 'ms': 'Kangri', 'my': 'ခန်းဂရီ', 'nb': 'kangri', 'ne': 'काङ्ग्री', 'nl': 'Kangri', 'nn': 'kangri', 'no': 'kangri', 'or': 'କାଙ୍ଗ୍ରି', 'pa': 'ਕਾਂਗੜੀ', 'pa-Guru': 'ਕਾਂਗੜੀ', 'pcm': 'Kangri Lángwej', 'pl': 'kangri', 'ps': 'کانګرو', 'pt': 'kandri', 'ro': 'kangri', 'ru': 'кангри', 'sc': 'kangri', 'sd': 'ڪينگري', 'sd-Arab': 'ڪينگري', 'si': 'කැන්ග්\u200dරි', 'sk': 'kángrí', 'sl': 'kangri', 'so': 'Kangri', 'sq': 'kangrisht', 'sr': 'кангри', 'sr-Cyrl': 'кангри', 'sr-Latn': 'kangri', 'sv': 'kangri', 'sw': 'Kikangri', 'ta': 'காங்கிரி', 'te': 'కాంగ్\u200cడీ', 'th': 'กังกรี', 'ti': 'ካንጋሪኛ', 'tk': 'kangri dili', 'tr': 'Kangrice', 'uk': 'кангрі', 'ur': 'کانگری', 'uz': 'kangri', 'uz-Latn': 'kangri', 'vec': 'kangri', 'vi': 'Tiếng Kangri', 'xnr': 'कांगड़ी', 'yo': 'Kangiri', 'yue': '康格里文', 'yue-Hans': '康格里文', 'yue-Hant': '康格里文', 'zh': '康格里语', 'zh-Hans': '康格里语', 'zh-Hant': '康格里', 'zu': 'Kangri'},
    'xns': {'en': 'Kanashi'},
    'xnt': {'en': 'Narragansett'},
    'xnu': {'en': 'Nukunul'},
    'xny': {'en': 'Nyiyaparli'},
    'xnz': {'en': 'Kenzi'},
    'xoc': {'en': "O'chi'chi'"},
    'xod': {'en': 'Kokoda'},
    'xog': {'af': 'Soga', 'am': 'ሶጋ', 'ar': 'السوغا', 'as': 'ছোগা', 'ast': 'soga', 'az': 'soqa', 'az-Cyrl': 'сога', 'az-Latn': 'soqa', 'bal-Latn': 'Sugá', 'be': 'сога', 'be-tarask': 'сога', 'bg': 'сога', 'bgn': 'سوگایی', 'bn': 'সোগা', 'br': 'sogaeg', 'brx': 'स’गा', 'bs': 'soga', 'bs-Cyrl': 'сога', 'bs-Latn': 'soga', 'ca': 'soga', 'ca-ES-valencia': 'soga', 'ccp': '𑄥𑄮𑄉', 'ce': 'сога', 'chr': 'ᏐᎦ', 'ckb': 'سۆگا', 'cs': 'sogština', 'cy': 'Soga', 'da': 'soga', 'de': 'Soga', 'dsb': 'soga', 'el': 'Σόγκα', 'el-polyton': 'Σόγκα', 'en': 'Soga', 'es': 'soga', 'et': 'soga', 'eu': 'sogera', 'fa': 'سوگایی', 'ff-Adlm': '𞤅𞤮𞤺𞤢𞥄𞤪𞤫', 'fi': 'soga', 'fil': 'Soga', 'fo': 'soga', 'fr': 'soga', 'fy': 'Soga', 'ga': 'Soga', 'gd': 'Soga', 'gl': 'soga', 'gu': 'સોગા', 'ha': 'Soga', 'he': 'סוגה', 'hi': 'सोगा', 'hi-Latn': 'Soga', 'hr': 'soga', 'hsb': 'soga', 'ht': 'soga', 'hu': 'szoga', 'hy': 'սոգա', 'ia': 'soga', 'id': 'Soga', 'ig': 'Soga', 'is': 'sóga', 'it': 'soga', 'ja': 'ソガ語', 'jv': 'Soga', 'ka': 'სოგა', 'kgp': 'rusoga', 'kk': 'сога тілі', 'kk-Cyrl': 'сога тілі', 'km': 'សូហ្គា', 'kn': 'ಸೊಗ', 'ko': '소가어', 'kok': 'सोगा', 'kok-Deva': 'सोगा', 'ku': 'sogayî', 'ky': 'согача', 'lb': 'Soga', 'lo': 'ໂຊກາ', 'lrc': 'سوٙگا', 'lt': 'sogų', 'lv': 'sogu', 'mai': 'सोगा', 'mi': 'Hoka', 'mk': 'сога', 'ml': 'സോഗോ', 'mn': 'сога', 'mr': 'सोगा', 'ms': 'Soga', 'mt': 'Soga', 'my': 'ဆိုဂါ', 'mzn': 'سوگا', 'nb': 'soga', 'ne': 'सोगा', 'nl': 'Soga', 'nn': 'soga', 'no': 'soga', 'nqo': 'ߛߏߜ߭ߊߞߊ߲', 'or': 'ସୋଗା', 'pa': 'ਸੋਗਾ', 'pa-Guru': 'ਸੋਗਾ', 'pcm': 'sóga', 'pl': 'soga', 'ps': 'سوګا', 'pt': 'lusoga', 'qu': 'Soga Simi', 'rm': 'soga', 'ro': 'soga', 'ru': 'сога', 'sc': 'soga', 'sd': 'سوگا', 'sd-Arab': 'سوگا', 'si': 'සොගා', 'sk': 'soga', 'sl': 'sogščina', 'smn': 'soga', 'so': 'Sooga', 'sq': 'sogisht', 'sr': 'сога', 'sr-Cyrl': 'сога', 'sr-Latn': 'soga', 'sv': 'lusoga', 'sw': 'Kisoga', 'ta': 'சோகா', 'te': 'సొగా', 'th': 'โซกา', 'ti': 'ሶጋ', 'tk': 'soga dili', 'to': 'lea fakasoka', 'tr': 'Soga', 'ug': 'سوگاچە', 'uk': 'сога', 'ur': 'سوگا', 'uz': 'soga', 'uz-Cyrl': 'сога', 'uz-Latn': 'soga', 'vec': 'soga', 'vi': 'Tiếng Soga', 'xnr': 'सोगा', 'xog': 'Olusoga', 'yo': 'Ṣógà', 'yrl': 'rusoga', 'yue': '索加文', 'yue-Hans': '索加文', 'yue-Hant': '索加文', 'zh': '索加语', 'zh-Hans': '索加语', 'zh-Hant': '索加文', 'zu': 'isi-Soga'},
    'xoi': {'en': 'Kominimung'},
    'xok': {'en': 'Xokleng'},
    'xom': {'en': 'Komo (Sudan)'},
    'xon': {'en': 'Konkomba'},
    'xoo': {'en': 'Xukurú'},
    'xop': {'en': 'Kopar'},
    'xor': {'en': 'Korubo'},
    'xow': {'en': 'Kowaki'},
    'xpa': {'en': 'Pirriya'},
    'xpb': {'en': 'Northeastern Tasmanian'},
    'xpc': {'en': 'Pecheneg'},
    'xpd': {'en': 'Oyster Bay Tasmanian'},
    'xpe': {'en': 'Liberia Kpelle'},
    'xpf': {'en': 'Southeast Tasmanian'},
    'xpg': {'en': 'Phrygian'},
    'xph': {'en': 'North Midlands Tasmanian'},
    'xpi': {'en': 'Pictish'},
    'xpj': {'en': 'Mpalitjanh'},
    'xpk': {'en': 'Kulina Pano'},
    'xpl': {'en': 'Port Sorell Tasmanian'},
    'xpm': {'en': 'Pumpokol'},
    'xpn': {'en': 'Kapinawá'},
    'xpo': {'en': 'Pochutec'},
    'xpp': {'en': 'Puyo-Paekche'},
    'xpq': {'en': 'Mohegan-Pequot'},
    'xpr': {'en': 'Parthian'},
    'xps': {'en': 'Pisidian'},
    'xpt': {'en': 'Punthamara'},
    'xpu': {'en': 'Punic'},
    'xpv': {'en': 'Northern Tasmanian'},
    'xpw': {'en': 'Northwestern Tasmanian'},
    'xpx': {'en': 'Southwestern Tasmanian'},
    'xpy': {'en': 'Puyo'},
    'xpz': {'en': 'Bruny Island Tasmanian'},
    'xqa': {'en': 'Karakhanid'},
    'xqt': {'en': 'Qatabanian'},
    'xra': {'en': 'Krahô'},
    'xrb': {'en': 'Eastern Karaboro'},
    'xrd': {'en': 'Gundungurra'},
    'xre': {'en': 'Kreye'},
    'xrg': {'en': 'Minang'},
    'xri': {'en': 'Krikati-Timbira'},
    'xrm': {'en': 'Armazic'},
    'xrn': {'en': 'Arin'},
    'xrq': {'en': 'Karranga'},
    'xrr': {'en': 'Raetic'},
    'xrt': {'en': 'Aranama-Tamique'},
    'xru': {'en': 'Marriammu'},
    'xrw': {'en': 'Karawa'},
    'xsa': {'en': 'Sabaean'},
    'xsb': {'en': 'Sambal'},
    'xsc': {'en': 'Scythian'},
    'xsd': {'en': 'Sidetic'},
    'xse': {'en': 'Sempan'},
    'xsh': {'en': 'Shamang'},
    'xsi': {'en': 'Sio'},
    'xsj': {'en': 'Subi'},
    'xsl': {'en': 'South Slavey'},
    'xsm': {'en': 'Kasem'},
    'xsn': {'en': 'Sanga (Nigeria)'},
    'xso': {'en': 'Solano'},
    'xsp': {'en': 'Silopi'},
    'xsq': {'en': 'Makhuwa-Saka'},
    'xsr': {'en': 'Sherpa'},
    'xss': {'en': 'Assan'},
    'xsu': {'en': 'Sanumá'},
    'xsv': {'en': 'Sudovian'},
    'xsy': {'en': 'Saisiyat'},
    'xta': {'en': 'Alcozauca Mixtec'},
    'xtb': {'en': 'Chazumba Mixtec'},
    'xtc': {'en': 'Katcha-Kadugli-Miri'},
    'xtd': {'en': 'Diuxi-Tilantongo Mixtec'},
    'xte': {'en': 'Ketengban'},
    'xtg': {'en': 'Transalpine Gaulish'},
    'xth': {'en': 'Yitha Yitha'},
    'xti': {'en': 'Sinicahua Mixtec'},
    'xtj': {'en': 'San Juan Teita Mixtec'},
    'xtl': {'en': 'Tijaltepec Mixtec'},
    'xtm': {'en': 'Magdalena Peñasco Mixtec'},
    'xtn': {'en': 'Northern Tlaxiaco Mixtec'},
    'xto': {'en': 'Tokharian A'},
    'xtp': {'en': 'San Miguel Piedras Mixtec'},
    'xtq': {'en': 'Tumshuqese'},
    'xtr': {'en': 'Early Tripuri'},
    'xts': {'en': 'Sindihui Mixtec'},
    'xtt': {'en': 'Tacahua Mixtec'},
    'xtu': {'en': 'Cuyamecalco Mixtec'},
    'xtv': {'en': 'Thawa'},
    'xtw': {'en': 'Tawandê'},
    'xty': {'en': 'Yoloxochitl Mixtec'},
    'xtz': {'en': 'Tasmanian'},
    'xua': {'en': 'Alu Kurumba'},
    'xub': {'en': 'Betta Kurumba'},
    'xud': {'en': 'Umiida'},
    'xug': {'en': 'Kunigami'},
    'xuj': {'en': 'Jennu Kurumba'},
    'xul': {'en': 'Ngunawal'},
    'xum': {'en': 'Umbrian'},
    'xun': {'en': 'Unggaranggu'},
    'xuo': {'en': 'Kuo'},
    'xup': {'en': 'Upper Umpqua'},
    'xur': {'en': 'Urartian'},
    'xut': {'en': 'Kuthant'},
    'xuu': {'en': 'Kxoe'},
    'xve': {'en': 'Venetic'},
    'xvi': {'en': 'Kamviri'},
    'xvn': {'en': 'Vandalic'},
    'xvo': {'en': 'Volscian'},
    'xvs': {'en': 'Vestinian'},
    'xwa': {'en': 'Kwaza'},
    'xwc': {'en': 'Woccon'},
    'xwd': {'en': 'Wadi Wadi'},
    'xwe': {'en': 'Xwela Gbe'},
    'xwg': {'en': 'Kwegu'},
    'xwj': {'en': 'Wajuk'},
    'xwk': {'en': 'Wangkumara'},
    'xwl': {'en': 'Western Xwla Gbe'},
    'xwo': {'en': 'Written Oirat'},
    'xwr': {'en': 'Kwerba Mamberamo'},
    'xwt': {'en': 'Wotjobaluk'},
    'xww': {'en': 'Wemba Wemba'},
    'xxb': {'en': 'Boro (Ghana)'},
    'xxk': {'en': "Ke'o"},
    'xxm': {'en': 'Minkin'},
    'xxr': {'en': 'Koropó'},
    'xxt': {'en': 'Tambora'},
    'xya': {'en': 'Yaygir'},
    'xyb': {'en': 'Yandjibara'},
    'xyj': {'en': 'Mayi-Yapi'},
    'xyk': {'en': 'Mayi-Kulan'},
    'xyl': {'en': 'Yalakalore'},
    'xyt': {'en': 'Mayi-Thakurti'},
    'xyy': {'en': 'Yorta Yorta'},
    'xzh': {'en': 'Zhang-Zhung'},
    'xzm': {'en': 'Zemgalian'},
    'xzp': {'en': 'Ancient Zapotec'},
    'yaa': {'en': 'Yaminahua'},
    'yab': {'en': 'Yuhup'},
    'yac': {'en': 'Pass Valley Yali'},
    'yad': {'en': 'Yagua'},
    'yae': {'en': 'Pumé'},
    'yaf': {'en': 'Yaka (Democratic Republic of Congo)'},
    'yag': {'en': 'Yámana'},
    'yah': {'en': 'Yazgulyam'},
    'yai': {'en': 'Yagnobi'},
    'yaj': {'en': 'Banda-Yangere'},
    'yak': {'en': 'Yakama'},
    'yal': {'en': 'Yalunka'},
    'yam': {'en': 'Yamba'},
    'yan': {'en': 'Mayangna'},
    'yao': {'ar': 'الياو', 'bg': 'яо', 'bn': 'ইয়াও', 'brx': 'याओ', 'bs': 'jao', 'bs-Cyrl': 'јао', 'bs-Latn': 'jao', 'ccp': '𑄃𑄨𑄠𑄃𑄮', 'cs': 'jaoština', 'el': 'Γιάο', 'el-polyton': 'Γιάο', 'et': 'jao', 'fa': 'یائویی', 'fi': 'jao', 'gsw': 'Yao-Schpraach', 'gu': 'યાઓ', 'he': 'יאו', 'hi': 'याओ', 'hu': 'jaó', 'hy': 'յաո', 'is': 'jaó', 'it': 'yao (bantu)', 'ja': 'ヤオ語', 'kgp': 'javo', 'kn': 'ಯಾವೊ', 'ko': '야오족어', 'ks': 'یاو', 'ks-Arab': 'یاو', 'lb': 'Yao-Sprooch', 'lo': 'ເຢົ້າ', 'lt': 'jao', 'lv': 'jao', 'mk': 'јао', 'ml': 'യാവോ', 'mr': 'याओ', 'or': 'ୟାଓ', 'ru': 'яо', 'sk': 'jao', 'sl': 'jaojščina', 'sr': 'јао', 'sr-Cyrl': 'јао', 'sr-Latn': 'jao', 'sv': 'kiyao', 'sw': 'Kiyao', 'ta': 'யாவ்', 'te': 'యాయే', 'th': 'เย้า', 'to': 'lea fakaʻiao', 'ug': 'ياۋچە', 'uk': 'яо', 'vi': 'Tiếng Yao', 'yrl': 'yau', 'yue': '瑤文', 'yue-Hans': '瑶文', 'yue-Hant': '瑤文', 'zh': '尧语', 'zh-Hans': '尧语', 'zh-Hant': '瑤文', 'en': 'Yao'},
    'yap': {'ar': 'اليابيز', 'ast': 'yapés', 'az': 'yapiz', 'az-Latn': 'yapiz', 'bg': 'япезе', 'bn': 'ইয়াপেসে', 'br': 'yapeg', 'brx': 'यापीज़', 'bs': 'japeški', 'bs-Cyrl': 'јапешки', 'bs-Latn': 'japeški', 'ca': 'yapeà', 'ca-ES-valencia': 'yapeà', 'ccp': '𑄃𑄨𑄠𑄛𑄬𑄥𑄬', 'cs': 'japština', 'da': 'yapese', 'de': 'Yapesisch', 'el': 'Γιαπίζ', 'el-polyton': 'Γιαπίζ', 'en': 'Yapese', 'es': 'yapés', 'et': 'japi', 'fa': 'یاپی', 'fi': 'japi', 'fr': 'yapois', 'fy': 'Yapees', 'gd': 'Cànan Yap', 'gsw': 'Yapesisch', 'gu': 'યાપીસ', 'he': 'יאפזית', 'hi': 'यापीस', 'hi-Latn': 'Yapese', 'hr': 'japski', 'ht': 'yapois', 'hu': 'japi', 'hy': 'յափերեն', 'id': 'Yapois', 'is': 'japíska', 'it': 'yapese', 'ja': 'ヤップ語', 'kgp': 'japese', 'kn': 'ಯಪೀಸೆ', 'ko': '얍페세어', 'ks': 'یَپیٖز', 'ks-Arab': 'یَپیٖز', 'lb': 'Yapesesch', 'lo': 'ຢັບ', 'lt': 'japezų', 'lv': 'japiešu', 'mk': 'јапски', 'ml': 'യെപ്പീസ്', 'mr': 'यापीस', 'mt': 'Yapese', 'nb': 'yapesisk', 'nl': 'Yapees', 'nn': 'yapesisk', 'no': 'yapesisk', 'or': 'ୟାପୀସ୍', 'pl': 'japski', 'pt': 'yapese', 'rm': 'yapais', 'ro': 'yapeză', 'ru': 'яп', 'sk': 'japčina', 'sl': 'japščina', 'sr': 'јапски', 'sr-Cyrl': 'јапски', 'sr-Latn': 'japski', 'sv': 'japetiska', 'ta': 'யாபேசே', 'te': 'యాపిస్', 'th': 'ยัป', 'to': 'lea fakaʻiapi', 'tr': 'Yapça', 'ug': 'ياپچە', 'uk': 'яп', 'vi': 'Tiếng Yap', 'yrl': 'yapese', 'yue': '雅浦文', 'yue-Hans': '雅浦文', 'yue-Hant': '雅浦文', 'zh': '雅浦语', 'zh-Hans': '雅浦语', 'zh-Hant': '雅浦文'},
    'yaq': {'en': 'Yaqui'},
    'yar': {'en': 'Yabarana'},
    'yas': {'en': 'Nugunu (Cameroon)'},
    'yat': {'en': 'Yambeta'},
    'yau': {'en': 'Yuwana'},
    'yav': {'af': 'Yangben', 'am': 'ያንግቤንኛ', 'ar': 'يانجبن', 'as': 'য়াংবেন', 'ast': 'yangben', 'az': 'yanqben', 'az-Cyrl': 'јангбен', 'az-Latn': 'yanqben', 'bal-Latn': 'Yangben', 'be': 'янгбэн', 'be-tarask': 'янгбэн', 'bg': 'янгбен', 'bgn': 'یانگبینی', 'bn': 'ইয়াঙ্গবেন', 'br': 'yangben', 'brx': 'यांबेन', 'bs': 'jangben', 'bs-Cyrl': 'јангбен', 'bs-Latn': 'jangben', 'ca': 'yangben', 'ca-ES-valencia': 'yangben', 'ccp': '𑄠𑄋𑄴𑄉𑄧𑄝𑄬𑄚𑄴', 'ce': 'янгбен', 'chr': 'ᏰᎾᎦᏇᏂ', 'ckb': 'یانگبێن', 'cs': 'jangbenština', 'cy': 'Iangben', 'da': 'yangben', 'de': 'Yangben', 'dsb': 'yangben', 'el': 'Γιανγκμπέν', 'el-polyton': 'Γιανγκμπέν', 'en': 'Yangben', 'es': 'yangben', 'et': 'yangbeni', 'eu': 'yangbenera', 'fa': 'یانگبنی', 'ff-Adlm': '𞤒𞤢𞤲𞤺𞤦𞤫𞥅𞤪𞤫', 'fi': 'yangben', 'fil': 'Yangben', 'fo': 'yangben', 'fr': 'yangben', 'fy': 'Yangben', 'ga': 'Yangben', 'gd': 'Yangben', 'gl': 'yangben', 'gu': 'યાન્ગબેન', 'ha': 'Yangben', 'he': 'יאנגבן', 'hi': 'यांगबेन', 'hi-Latn': 'Yangben', 'hr': 'yangben', 'hsb': 'yangben', 'ht': 'yangben', 'hu': 'jangben', 'hy': 'յանգբեն', 'ia': 'yangben', 'id': 'Yangben', 'ig': 'Yangben', 'is': 'yangben', 'it': 'yangben', 'ja': 'ヤンベン語', 'jv': 'Yangben', 'ka': 'იანგბენი', 'kgp': 'jỹgmen', 'kk': 'янгбен тілі', 'kk-Cyrl': 'янгбен тілі', 'km': 'យ៉ាងបេន', 'kn': 'ಯಾಂಗ್ಬೆನ್', 'ko': '양본어', 'kok': 'यांगबेन', 'kok-Deva': 'यांगबेन', 'ku': 'yangben', 'ky': 'янгбенче', 'lb': 'Yangben', 'lo': 'ແຍງເບນ', 'lt': 'jangbenų', 'lv': 'janbaņu', 'mai': 'यंगबैन', 'mi': 'Angapene', 'mk': 'јенгбен', 'ml': 'യാംഗ്ബെൻ', 'mn': 'янгбен', 'mr': 'यांगबेन', 'ms': 'Yangben', 'mt': 'Yangben', 'my': 'ရန်ဘဲန်', 'nb': 'yangben', 'ne': 'याङ्बेन', 'nl': 'Yangben', 'nn': 'yangben', 'nnh': 'Shwóŋò pʉa shÿó Mbafìa', 'no': 'yangben', 'nqo': 'ߦߊ߲ߜߌߞߊ߲', 'or': 'ୟାଂବେନ୍', 'pa': 'ਯਾਂਗਬੇਨ', 'pa-Guru': 'ਯਾਂਗਬੇਨ', 'pcm': 'Yangbẹn', 'pl': 'yangben', 'ps': 'ینګبین', 'pt': 'yangben', 'qu': 'Yangben Simi', 'rm': 'yangben', 'ro': 'yangben', 'ru': 'янгбен', 'sc': 'yangben', 'sd': 'يانگ بين', 'sd-Arab': 'يانگ بين', 'si': 'යන්ග්බෙන්', 'sk': 'jangben', 'sl': 'jangben', 'smn': 'yangben', 'so': 'Yaangbeen', 'sq': 'jangbenisht', 'sr': 'јангбен', 'sr-Cyrl': 'јангбен', 'sr-Latn': 'jangben', 'sv': 'yangben', 'sw': 'Kiyangben', 'ta': 'யாங்பென்', 'te': 'యాంగ్\u200cబెన్', 'th': 'แยงเบน', 'ti': 'ያንግበን', 'tk': 'ýangben dili', 'to': 'lea fakaʻiangipeni', 'tr': 'Yangben', 'ug': 'ياڭبەنچە', 'uk': 'янгбен', 'ur': 'یانگبین', 'uz': 'yangben', 'uz-Cyrl': 'янгбен', 'uz-Latn': 'yangben', 'vec': 'yangben', 'vi': 'Tiếng Yangben', 'xnr': 'यांगबेन', 'yav': 'nuasue', 'yo': 'Yangbẹn', 'yrl': 'yãgibẽ', 'yue': '洋卞文', 'yue-Hans': '洋卞文', 'yue-Hant': '洋卞文', 'zh': '洋卞语', 'zh-Hans': '洋卞语', 'zh-Hant': '洋卞文', 'zu': 'isi-Yangben'},
    'yaw': {'en': 'Yawalapití'},
    'yax': {'en': 'Yauma'},
    'yay': {'en': 'Agwagwune'},
    'yaz': {'en': 'Lokaa'},
    'yba': {'en': 'Yala'},
    'ybb': {'af': 'Yemba', 'am': 'የምባ', 'ar': 'يمبا', 'as': 'য়েম্বা', 'ast': 'yemba', 'az': 'yemba', 'az-Cyrl': 'јемба', 'az-Latn': 'yemba', 'be': 'йемба', 'be-tarask': 'йемба', 'bg': 'йемба', 'bgn': 'یمبی', 'bn': 'ইয়েম্বা', 'br': 'yemba', 'brx': 'येम्बा', 'bs': 'jemba', 'bs-Cyrl': 'јемба', 'bs-Latn': 'jemba', 'ca': 'yemba', 'ca-ES-valencia': 'yemba', 'ccp': '𑄠𑄮𑄟𑄴𑄝', 'ce': 'йемба', 'chr': 'ᏰᎹᏋ', 'ckb': 'یێمبا', 'cs': 'yemba', 'cy': 'Iembaeg', 'da': 'yemba', 'de': 'Yemba', 'dsb': 'yembašćina', 'el': 'Γιέμπα', 'el-polyton': 'Γιέμπα', 'en': 'Yemba', 'es': 'yemba', 'et': 'jemba', 'eu': 'yemba', 'fa': 'یمبایی', 'ff-Adlm': '𞤒𞤫𞤥𞤦𞤢𞥄𞤪𞤫', 'fi': 'yemba', 'fil': 'Yemba', 'fo': 'yemba', 'fr': 'yemba', 'fy': 'Yemba', 'ga': 'Yemba', 'gd': 'Yemba', 'gl': 'yemba', 'gu': 'યેમ્બા', 'ha': 'Yemba', 'he': 'ימבה', 'hi': 'येंबा', 'hi-Latn': 'Yemba', 'hr': 'yemba', 'hsb': 'yembašćina', 'ht': 'yemba', 'hu': 'jemba', 'hy': 'եմբա', 'ia': 'yemba', 'id': 'Yemba', 'ig': 'Yemba', 'is': 'yemba', 'it': 'yemba', 'ja': 'イエンバ語', 'jv': 'Yemba', 'ka': 'იემბა', 'kgp': 'jẽma', 'kk': 'йемба тілі', 'kk-Cyrl': 'йемба тілі', 'km': 'យេមបា', 'kn': 'ಯೆಂಬಾ', 'ko': '옘바어', 'kok': 'येम्बा', 'kok-Deva': 'येम्बा', 'ksh': 'Jämmbah', 'ku': 'yemba', 'ky': 'йембача', 'lb': 'Yemba', 'lo': 'ແຢມບາ', 'lt': 'jembų', 'lv': 'jembu', 'mai': 'येमबा', 'mi': 'Emapa', 'mk': 'јемба', 'ml': 'യംബ', 'mn': 'емба', 'mr': 'येमबा', 'ms': 'Yemba', 'mt': 'Yemba', 'my': 'ရမ်ဘာ', 'nb': 'yemba', 'ne': 'येम्बा', 'nl': 'Yemba', 'nn': 'yemba', 'nnh': 'Shwóŋò Tsaŋ', 'no': 'yemba', 'or': 'ୟେମବା', 'pa': 'ਯੇਂਬਾ', 'pa-Guru': 'ਯੇਂਬਾ', 'pcm': 'Yẹmba Lángwej', 'pl': 'yemba', 'ps': 'یمبا', 'pt': 'yemba', 'qu': 'Yemba Simi', 'ro': 'yemba', 'ru': 'йемба', 'sc': 'yemba', 'sd': 'ييمبا', 'sd-Arab': 'ييمبا', 'si': 'යෙම්බා', 'sk': 'yemba', 'sl': 'jembajščina', 'smn': 'yemba', 'so': 'Yemba', 'sq': 'jembaisht', 'sr': 'јемба', 'sr-Cyrl': 'јемба', 'sr-Latn': 'jemba', 'sv': 'bamileké-jemba', 'sw': 'Kiyemba', 'ta': 'யெம்பா', 'te': 'యెంబా', 'th': 'เยมบา', 'ti': 'የምባ', 'tk': 'ýemba dili', 'to': 'lea fakaʻiēmipa', 'tr': 'Yemba', 'ug': 'يېمباچە', 'uk': 'ємба', 'ur': 'یمبا', 'uz': 'yemba', 'uz-Latn': 'yemba', 'vec': 'yenba', 'vi': 'Tiếng Yemba', 'xnr': 'येंबा', 'yo': 'Èdè Yemba', 'yrl': 'yẽba', 'yue': '耶姆巴文', 'yue-Hans': '耶姆巴文', 'yue-Hant': '耶姆巴文', 'zh': '耶姆巴语', 'zh-Hans': '耶姆巴语', 'zh-Hant': '耶姆巴文', 'zu': 'isi-Yemba'},
    'ybd': {'en': 'Yangbye'},
    'ybe': {'en': 'West Yugur'},
    'ybh': {'en': 'Yakha'},
    'ybi': {'en': 'Yamphu'},
    'ybj': {'en': 'Hasha'},
    'ybk': {'en': 'Bokha'},
    'ybl': {'en': 'Yukuben'},
    'ybm': {'en': 'Yaben'},
    'ybn': {'en': 'Yabaâna'},
    'ybo': {'en': 'Yabong'},
    'ybx': {'en': 'Yawiyo'},
    'yby': {'en': 'Yaweyuha'},
    'ych': {'en': 'Chesu'},
    'ycl': {'en': 'Lolopo'},
    'ycn': {'en': 'Yucuna'},
    'ycp': {'en': 'Chepya'},
    'yda': {'en': 'Yanda'},
    'ydd': {'en': 'Eastern Yiddish'},
    'yde': {'en': 'Yangum Dey'},
    'ydg': {'en': 'Yidgha'},
    'ydk': {'en': 'Yoidik'},
    'yds': {'en': 'Yiddish Sign Language'},
    'yea': {'en': 'Ravula'},
    'yec': {'en': 'Yeniche'},
    'yee': {'en': 'Yimas'},
    'yei': {'en': 'Yeni'},
    'yej': {'en': 'Yevanic'},
    'yel': {'en': 'Yela'},
    'yen': {'en': 'Yendang'},
    'yer': {'en': 'Tarok'},
    'yes': {'en': 'Nyankpa'},
    'yet': {'en': 'Yetfa'},
    'yeu': {'en': 'Yerukula'},
    'yev': {'en': 'Yapunda'},
    'yey': {'en': 'Yeyi'},
    'yga': {'en': 'Malyangapa'},
    'ygi': {'en': 'Yiningayi'},
    'ygl': {'en': 'Yangum Gel'},
    'ygm': {'en': 'Yagomi'},
    'ygp': {'en': 'Gepo'},
    'ygr': {'en': 'Yagaria'},
    'ygs': {'en': 'Yolŋu Sign Language'},
    'ygu': {'en': 'Yugul'},
    'ygw': {'en': 'Yagwoia'},
    'yha': {'en': 'Baha Buyang'},
    'yhd': {'en': 'Judeo-Iraqi Arabic'},
    'yhl': {'en': 'Hlepho Phowa'},
    'yhs': {'en': 'Yan-nhaŋu Sign Language'},
    'yi': {'af': 'Jiddisj', 'am': 'ይዲሽኛ', 'ar': 'اليديشية', 'as': 'ইদ্দিছ', 'ast': 'yiddish', 'az': 'idiş', 'az-Cyrl': 'идиш', 'az-Latn': 'idiş', 'bal-Latn': 'Yeddi', 'be': 'ідыш', 'be-tarask': 'ідыш', 'bg': 'идиш', 'bgn': 'یدی', 'bn': 'ইদ্দিশ', 'br': 'yiddish', 'brx': 'यीद्दीश', 'bs': 'jidiš', 'bs-Cyrl': 'јидиш', 'bs-Latn': 'jidiš', 'ca': 'ídix', 'ca-ES-valencia': 'ídix', 'ccp': '𑄠𑄨𑄖𑄴𑄘𑄨𑄥𑄴', 'ce': 'идиш', 'chr': 'ᏱᏗᏍ', 'ckb': 'ییدیش', 'cs': 'jidiš', 'cy': 'Iddew-Almaeneg', 'da': 'jiddisch', 'de': 'Jiddisch', 'dsb': 'jidišćina', 'el': 'Γίντις', 'el-polyton': 'Ἰουδαϊκά', 'en': 'Yiddish', 'eo': 'jida', 'es': 'yidis', 'et': 'jidiši', 'eu': 'yiddisha', 'fa': 'یدی', 'ff-Adlm': '𞤒𞤭𞤣𞤭𞤧𞤢𞤲𞤳𞤮𞥅𞤪𞤫', 'fi': 'jiddiš', 'fil': 'Yiddish', 'fo': 'jiddiskt', 'fr': 'yiddish', 'fur': 'yiddish', 'fy': 'Jiddysk', 'ga': 'Giúdais', 'gd': 'Iùdhais', 'gl': 'yiddish', 'gsw': 'Jiddisch', 'gu': 'યિદ્દિશ', 'ha': 'Yaren Yiddish', 'he': 'יידיש', 'hi': 'यहूदी', 'hi-Latn': 'Yiddish', 'hr': 'jidiš', 'hsb': 'jidišćina', 'ht': 'yiddish', 'hu': 'jiddis', 'hy': 'իդիշ', 'ia': 'yiddish', 'id': 'Yiddish', 'ig': 'Yiddish', 'is': 'jiddíska', 'it': 'yiddish', 'ja': 'イディッシュ語', 'jv': 'Yiddish', 'ka': 'იდიში', 'kgp': 'jinhise', 'kk': 'идиш тілі', 'kk-Cyrl': 'идиш тілі', 'km': 'យ៉ីឌីស', 'kn': 'ಯಿಡ್ಡಿಶ್', 'ko': '이디시어', 'kok': 'यिडिश', 'kok-Deva': 'यिडिश', 'ks': 'یِدِش', 'ks-Arab': 'یِدِش', 'ksh': 'Jiddesch', 'ku': 'yidîşî', 'ky': 'идишче', 'lb': 'Jiddesch', 'lld': 'yiddish', 'lo': 'ຢິວ', 'lt': 'jidiš', 'lv': 'jidišs', 'mai': 'ईडिश', 'mi': 'Irihi', 'mk': 'јидиш', 'ml': 'യിദ്ദിഷ്', 'mn': 'иддиш', 'mr': 'यिद्दिश', 'ms': 'Yiddish', 'mt': 'Yiddish', 'my': 'ရဟူဒီ', 'nb': 'jiddisk', 'ne': 'यिद्दिस', 'nl': 'Jiddisch', 'nn': 'jiddisk', 'no': 'jiddisk', 'nqo': 'ߦߘߌߛߌߞߊ߲', 'or': 'ୟିଡିସ୍', 'pa': 'ਯਿਦਿਸ਼', 'pa-Guru': 'ਯਿਦਿਸ਼', 'pcm': 'Yídish', 'pl': 'jidysz', 'ps': 'يديش', 'pt': 'iídiche', 'qu': 'Yiddish Simi', 'rm': 'jiddic', 'ro': 'idiș', 'ru': 'идиш', 'rw': 'Inyeyidishi', 'sc': 'yiddish', 'sd': 'يدش', 'sd-Arab': 'يدش', 'si': 'යිඩිශ්', 'sk': 'jidiš', 'sl': 'jidiš', 'smn': 'jiddish', 'so': 'Yadhish', 'sq': 'jidisht', 'sr': 'јидиш', 'sr-Cyrl': 'јидиш', 'sr-Latn': 'jidiš', 'sv': 'jiddisch', 'sw': 'Kiyiddi', 'syr': 'ܝܕܝܬܝܬ', 'ta': 'யெட்டிஷ்', 'te': 'ఇడ్డిష్', 'tg': 'идиш', 'th': 'ยิดดิช', 'ti': 'ይሁድኛ', 'tk': 'idiş dili', 'to': 'lea fakaītisi', 'tr': 'Yidiş', 'tt': 'идиш', 'ug': 'يىددىشچە', 'uk': 'їдиш', 'ur': 'یدش', 'uz': 'idish', 'uz-Cyrl': 'иддиш', 'uz-Latn': 'idish', 'vec': 'yidish', 'vi': 'Tiếng Yiddish', 'wo': 'Yidis', 'xnr': 'यहूदी', 'yi': 'ייִדיש', 'yo': 'Èdè Yiddishi', 'yrl': 'yídixi', 'yue': '意第緒文', 'yue-Hans': '意第绪文', 'yue-Hant': '意第緒文', 'zh': '意第绪语', 'zh-Hans': '意第绪语', 'zh-Hant': '意第緒文', 'zu': 'isi-Yiddish'},
    'yia': {'en': 'Yinggarda'},
    'yif': {'en': 'Ache'},
    'yig': {'en': 'Wusa Nasu'},
    'yih': {'en': 'Western Yiddish'},
    'yii': {'en': 'Yidiny'},
    'yij': {'en': 'Yindjibarndi'},
    'yik': {'en': 'Dongshanba Lalo'},
    'yil': {'en': 'Yindjilandji'},
    'yim': {'en': 'Yimchungru Naga'},
    'yin': {'en': 'Riang Lai'},
    'yip': {'en': 'Pholo'},
    'yiq': {'en': 'Miqie'},
    'yir': {'en': 'North Awyu'},
    'yis': {'en': 'Yis'},
    'yit': {'en': 'Eastern Lalu'},
    'yiu': {'en': 'Awu'},
    'yiv': {'en': 'Northern Nisu'},
    'yix': {'en': 'Axi Yi'},
    'yiy': {'en': 'Yir Yoront'},
    'yiz': {'en': 'Azhe'},
    'yka': {'en': 'Yakan'},
    'ykg': {'en': 'Northern Yukaghir'},
    'yki': {'en': 'Yoke'},
    'ykk': {'en': 'Yakaikeke'},
    'ykl': {'en': 'Khlula'},
    'ykm': {'en': 'Kap'},
    'ykn': {'en': 'Kua-nsi'},
    'yko': {'en': 'Yasa'},
    'ykr': {'en': 'Yekora'},
    'ykt': {'en': 'Kathu'},
    'yku': {'en': 'Kuamasi'},
    'yky': {'en': 'Yakoma'},
    'yla': {'en': 'Yaul'},
    'ylb': {'en': 'Yaleba'},
    'yle': {'en': 'Yele'},
    'ylg': {'en': 'Yelogu'},
    'yli': {'en': 'Angguruk Yali'},
    'yll': {'en': 'Yil'},
    'ylm': {'en': 'Limi'},
    'yln': {'en': 'Langnian Buyang'},
    'ylo': {'en': 'Naluo Yi'},
    'ylr': {'en': 'Yalarnnga'},
    'ylu': {'en': 'Aribwaung'},
    'yly': {'en': 'Nyâlayu'},
    'yma': {'en': 'Yamphe'},
    'ymb': {'en': 'Yambes'},
    'ymc': {'en': 'Southern Muji'},
    'ymd': {'en': 'Muda'},
    'yme': {'en': 'Yameo'},
    'ymg': {'en': 'Yamongeri'},
    'ymh': {'en': 'Mili'},
    'ymi': {'en': 'Moji'},
    'ymk': {'en': 'Makwe'},
    'yml': {'en': 'Iamalele'},
    'ymm': {'en': 'Maay'},
    'ymn': {'en': 'Yamna'},
    'ymo': {'en': 'Yangum Mon'},
    'ymp': {'en': 'Yamap'},
    'ymq': {'en': 'Qila Muji'},
    'ymr': {'en': 'Malasar'},
    'yms': {'en': 'Mysian'},
    'ymt': {'en': 'Mator-Taygi-Karagas'},
    'ymx': {'en': 'Northern Muji'},
    'ymz': {'en': 'Muzi'},
    'yna': {'en': 'Aluo'},
    'ynd': {'en': 'Yandruwandha'},
    'yne': {'en': "Lang'e"},
    'yng': {'en': 'Yango'},
    'ynh': {'en': 'Yangho'},
    'ynk': {'en': 'Naukan Yupik'},
    'ynl': {'en': 'Yangulam'},
    'ynn': {'en': 'Yana'},
    'yno': {'en': 'Yong'},
    'ynq': {'en': 'Yendang'},
    'yns': {'en': 'Yansi'},
    'ynu': {'en': 'Yahuna'},
    'yo': {'af': 'Joroeba', 'agq': 'Yulùba', 'ak': 'Yoruba', 'am': 'ዮሩባዊኛ', 'ar': 'اليوروبا', 'as': 'ইউৰুবা', 'asa': 'Kiyoruba', 'ast': 'yoruba', 'az': 'yoruba', 'az-Cyrl': 'јоруба', 'az-Latn': 'yoruba', 'bal-Latn': 'Yorobá', 'bas': 'Hɔp u yorūbà', 'be': 'ёруба', 'be-tarask': 'ёруба', 'bem': 'Ichi Yoruba', 'bez': 'Hiyoruba', 'bg': 'йоруба', 'bgn': 'یوروبایی', 'bm': 'yorubakan', 'bn': 'ইওরুবা', 'br': 'yorouba', 'brx': 'यरूबा', 'bs': 'jorubanski', 'bs-Cyrl': 'јоруба', 'bs-Latn': 'jorubanski', 'ca': 'ioruba', 'ca-ES-valencia': 'ioruba', 'ccp': '𑄃𑄨𑄃𑄮𑄢𑄪𑄝', 'ce': 'йоруба', 'cgg': 'Oruyoruba', 'chr': 'ᏲᏄᏆ', 'ckb': 'یۆرووبا', 'cs': 'jorubština', 'cy': 'Iorwba', 'da': 'yoruba', 'dav': 'Kiyoruba', 'de': 'Yoruba', 'dje': 'Yorbance senni', 'dsb': 'jorubšćina', 'dyo': 'yoruba', 'dz': 'ཡོ་རུ་བ་ཁ', 'ebu': 'Kĩyoruba', 'ee': 'yorubagbe', 'el': 'Γιορούμπα', 'el-polyton': 'Γιορούμπα', 'en': 'Yoruba', 'eo': 'joruba', 'es': 'yoruba', 'et': 'joruba', 'eu': 'jorubera', 'ewo': 'ǹkɔ́bɔ yorúba', 'fa': 'یوروبایی', 'ff': 'Yorrubaa', 'ff-Adlm': '𞤒𞤮𞥅𞤪𞤵𞤦𞤢𞥄𞤪𞤫', 'ff-Latn': 'Yorrubaa', 'fi': 'joruba', 'fil': 'Yoruba', 'fo': 'yoruba', 'fr': 'yoruba', 'fur': 'yoruba', 'fy': 'Yoruba', 'ga': 'Iarúibis', 'gd': 'Yoruba', 'gl': 'ioruba', 'gsw': 'Yoruba', 'gu': 'યોરૂબા', 'guz': 'Kiyoruba', 'ha': 'Yarbanci', 'he': 'יורובה', 'hi': 'योरूबा', 'hi-Latn': 'Yoruba', 'hr': 'jorupski', 'hsb': 'jorubašćina', 'ht': 'yoruba', 'hu': 'joruba', 'hy': 'յորուբա', 'ia': 'yoruba', 'id': 'Yoruba', 'ig': 'Yoruba', 'is': 'jórúba', 'it': 'yoruba', 'ja': 'ヨルバ語', 'jmc': 'Kyiyoruba', 'jv': 'Yoruba', 'ka': 'იორუბა', 'kab': 'Tayurubit', 'kam': 'Kiyoruba', 'kde': 'Chiyoluba', 'kea': 'ioruba', 'kgp': 'joruma', 'khq': 'Yorbance senni', 'ki': 'Kiyoruba', 'kk': 'йоруба тілі', 'kk-Cyrl': 'йоруба тілі', 'kln': 'kutitab Yoruba', 'km': 'យរូបា', 'kn': 'ಯೊರುಬಾ', 'ko': '요루바어', 'kok': 'योरुबा', 'kok-Deva': 'योरुबा', 'ks': 'یورُبا', 'ks-Arab': 'یورُبا', 'ksb': 'Kiyoluba', 'ksf': 'riyúuba', 'ksh': 'Joruhba', 'ku': 'yorubayî', 'ky': 'йорубача', 'lag': 'Kɨyorúuba', 'lb': 'Yoruba', 'lg': 'Luyoruba', 'lkt': 'Yoruba Iyápi', 'lld': 'yoruba', 'ln': 'yoruba', 'lo': 'ໂຢຣູບາ', 'lrc': 'یوروبا', 'lt': 'jorubų', 'lu': 'Nyoruba', 'luo': 'Kiyoruba', 'luy': 'Kiyoruba', 'lv': 'jorubu', 'mai': 'योरूबा', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lyoruba', 'mer': 'Kĩyoruba', 'mfe': 'yoruba', 'mg': 'Yôrobà', 'mgh': 'Iyoruba', 'mi': 'Ōrūpa', 'mk': 'јорупски', 'ml': 'യൊറൂബാ', 'mn': 'ёруба', 'mr': 'योरुबा', 'ms': 'Yoruba', 'mt': 'Yoruba', 'mua': 'Yoruba', 'my': 'ယိုရူဘာ', 'mzn': 'یوروبا', 'naq': 'Yorubab', 'nb': 'joruba', 'nd': 'isi-Yorubha', 'ne': 'योरूवा', 'nl': 'Yoruba', 'nmg': 'Yorúbâ', 'nn': 'joruba', 'no': 'joruba', 'nqo': 'ߙߦߏߓߊߞߊ߲', 'nus': 'Thok yurubani', 'nyn': 'Oruyoruba', 'or': 'ୟୋରୁବା', 'pa': 'ਯੋਰੂਬਾ', 'pa-Guru': 'ਯੋਰੂਬਾ', 'pcm': 'Yorubá', 'pl': 'joruba', 'ps': 'یوروبا', 'pt': 'iorubá', 'qu': 'Yoruba Simi', 'rm': 'yoruba', 'rn': 'Ikiyoruba', 'ro': 'yoruba', 'rof': 'Kiyoruba', 'ru': 'йоруба', 'rwk': 'Kyiyoruba', 'saq': 'Kiyoruba', 'sbp': 'Ishiyoluba', 'sc': 'yoruba', 'sd': 'يوروبا', 'sd-Arab': 'يوروبا', 'seh': 'iorubá', 'ses': 'Yorbance senni', 'sg': 'Yoruba', 'shi': 'ⵜⴰⵢⵔⵓⴱⴰⵜ', 'shi-Latn': 'Tayrubat', 'shi-Tfng': 'ⵜⴰⵢⵔⵓⴱⴰⵜ', 'si': 'යොරූබා', 'sk': 'jorubčina', 'sl': 'jorubščina', 'smn': 'yorubakielâ', 'sn': 'chiYoruba', 'so': 'Yoruuba', 'sq': 'jorubaisht', 'sr': 'јоруба', 'sr-Cyrl': 'јоруба', 'sr-Latn': 'joruba', 'sv': 'yoruba', 'sw': 'Kiyoruba', 'ta': 'யோருபா', 'te': 'యోరుబా', 'teo': 'Kiyoruba', 'tg': 'йоруба', 'th': 'โยรูบา', 'ti': 'ዮሩባ', 'tk': 'ýoruba dili', 'to': 'lea fakaʻiōlupa', 'tr': 'Yorubaca', 'tt': 'йоруба', 'twq': 'Yorbance senni', 'tzm': 'Tayurubat', 'ug': 'يورۇباچە', 'uk': 'йоруба', 'ur': 'یوروبا', 'uz': 'yoruba', 'uz-Cyrl': 'йоруба', 'uz-Latn': 'yoruba', 'vai': 'ꖎꖄꕑ', 'vai-Latn': 'Yóróbha', 'vai-Vaii': 'ꖎꖄꕑ', 'vec': 'yoruba', 'vi': 'Tiếng Yoruba', 'vun': 'Kyiyoruba', 'wae': 'Yoruba', 'wo': 'Yoruba', 'xnr': 'योरूबा', 'xog': 'Oluyoruba', 'yav': 'nuyolúpa', 'yo': 'Èdè Yorùbá', 'yrl': 'yurubá', 'yue': '約魯巴文', 'yue-Hans': '约鲁巴文', 'yue-Hant': '約魯巴文', 'zgh': 'ⵜⴰⵢⵔⵓⴱⴰⵜ', 'zh': '约鲁巴语', 'zh-Hans': '约鲁巴语', 'zh-Hant': '約魯巴文', 'zu': 'isi-Yoruba'},
    'yob': {'en': 'Yoba'},
    'yog': {'en': 'Yogad'},
    'yoi': {'en': 'Yonaguni'},
    'yok': {'en': 'Yokuts'},
    'yol': {'en': 'Yola'},
    'yom': {'en': 'Yombe'},
    'yon': {'en': 'Yongkom'},
    'yos': {'en': 'Yos'},
    'yot': {'en': 'Yotti'},
    'yox': {'en': 'Yoron'},
    'yoy': {'en': 'Yoy'},
    'ypa': {'en': 'Phala'},
    'ypb': {'en': 'Labo Phowa'},
    'ypg': {'en': 'Phola'},
    'yph': {'en': 'Phupha'},
    'ypk': {'en': 'Yupik languages'},
    'ypk-pro': {'en': 'Proto-Yupik'},
    'ypm': {'en': 'Phuma'},
    'ypn': {'en': 'Ani Phowa'},
    'ypo': {'en': 'Alo Phola'},
    'ypp': {'en': 'Phupa'},
    'ypz': {'en': 'Phuza'},
    'yra': {'en': 'Yerakai'},
    'yrb': {'en': 'Yareba'},
    'yre': {'en': 'Yaouré'},
    'yri': {'en': 'Yarí'},
    'yrk': {'en': 'Nenets'},
    'yrl': {'af': 'Nheengatu', 'ak': 'Ningatu', 'am': 'ንሄንጋቱ', 'ar': 'النيينجاتو', 'as': 'হিংগাটো', 'ast': 'nheengatu', 'az': 'nyenqatu', 'az-Latn': 'nyenqatu', 'bal-Latn': 'Ningátu', 'be': 'ньенгату', 'be-tarask': 'ньенгату', 'bg': 'ненгату', 'bn': 'নহিংগাটু', 'br': 'nengatoueg', 'brx': 'न्हीगातू', 'bs': 'ningatu', 'bs-Cyrl': 'ненгату', 'bs-Latn': 'ningatu', 'ca': 'nheengatú', 'ca-ES-valencia': 'nheengatú', 'chr': 'ᏂᎾᎦᏚ', 'cs': 'nheengatu', 'cy': 'Nheengatu', 'da': 'nheengatu', 'de': 'Nheengatu', 'dsb': 'nheengatušćina', 'el': 'Νινγκατού', 'el-polyton': 'Νινγκατού', 'en': 'Nheengatu', 'es': 'ñe’engatú', 'et': 'njengatu', 'eu': 'nheengatua', 'fa': 'نهین گاتو', 'ff-Adlm': '𞤛𞤫𞥅𞤲𞤺𞤢𞤼𞤵𞥅𞤪𞤫', 'fi': 'ñeengatú', 'fil': 'Nheengatu', 'fo': 'nheengatu', 'fr': 'nheengatou', 'ga': 'Nheengatu', 'gd': 'Nheengatu', 'gl': 'nheengatu', 'gu': 'નહેનગાતુ', 'ha': 'Nheengatu', 'he': 'נינגטו', 'hi': 'नेइंगातू', 'hi-Latn': 'Nheengatu', 'hr': 'nheengatu', 'hsb': 'nheengatušćina', 'ht': 'nheengatou', 'hu': 'nheengatu', 'hy': 'նինգաթու', 'ia': 'nheengatu', 'id': 'Nheengatu', 'ig': 'Asụsụ Nheengatu', 'is': 'nheengatu', 'it': 'nheengatu', 'ja': 'ニェエンガトゥ語', 'jv': 'Nheengatu', 'ka': 'ნენგატუ', 'kgp': 'nhẽgatu', 'kk': 'ньенгату тілі', 'kk-Cyrl': 'ньенгату тілі', 'km': 'ញីនហ្កាទូ', 'kn': 'ನಿಂಗಾಟು', 'ko': '넨가투어', 'kok': 'न्हीनगातु', 'kok-Deva': 'न्हीनगातु', 'ku': 'nhêngatûyî', 'ky': 'ньенгатуча (түштүк америка тилдери)', 'lb': 'Nheengatu', 'lld': 'nheengatu', 'lo': 'ນີນກາຕູ', 'lt': 'njengatu', 'lv': 'njengatu', 'mai': 'एनहींगाटु', 'mi': 'Nīkātū', 'mk': 'њенгату', 'ml': 'നീൻഗാറ്റു', 'mn': 'ньенгату', 'mr': 'न्हेंगाटू', 'ms': 'Nheengatu', 'my': 'အန်ဟင်းဂတူ', 'nb': 'nheengatu', 'ne': 'न्हिनगातु', 'nl': 'Nheengatu', 'nn': 'nheengatu', 'no': 'nheengatu', 'or': 'ନିଙ୍ଗାଟୁ', 'pa': 'ਨਹੀਂਗਾਤੂ', 'pa-Guru': 'ਨਹੀਂਗਾਤੂ', 'pcm': 'Nhiingátu Lángwej', 'pl': 'nheengatu', 'ps': 'نینګاتو', 'pt': 'nheengatu', 'qu': 'Nheengatu Simi', 'ro': 'nheengatu', 'ru': 'ньенгату', 'sc': 'nheengatu', 'sd': 'نھين گاٽو', 'sd-Arab': 'نھين گاٽو', 'si': 'නොහීඟටු', 'sk': 'nheengatu', 'sl': 'nheengatu', 'so': 'Nheengatu', 'sq': 'nejengatuisht', 'sr': 'нингату', 'sr-Cyrl': 'нингату', 'sr-Latn': 'ningatu', 'sv': 'nheengatu', 'sw': 'Kinheengatu', 'ta': 'நஹீன்கட்டு', 'te': 'నేహ్\u200cగటు', 'th': 'เหงงกาตุ', 'ti': 'ኒንጋቱ', 'tk': 'nhengatu dili', 'to': 'lea fakaneʻēngatū', 'tr': 'Nheengatu', 'uk': 'ньєнґату', 'ur': 'نینگاٹو', 'uz': 'nyengatu', 'uz-Latn': 'nyengatu', 'vec': 'nengatu', 'vi': 'Tiếng Nheengatu', 'yo': 'Èdè Ningatu', 'yrl': 'nheẽgatu', 'yue': '奈恩加圖文', 'yue-Hans': '奈恩加图文', 'yue-Hant': '奈恩加圖文', 'zh': '恩加图语', 'zh-Hans': '恩加图语', 'zh-Hant': '奈恩加圖文', 'zu': 'isi-Nheengatu'},
    'yrm': {'en': 'Yirrk-Mel'},
    'yrn': {'en': 'Yerong'},
    'yro': {'en': 'Yaroamë'},
    'yrs': {'en': 'Yarsun'},
    'yrw': {'en': 'Yarawata'},
    'yry': {'en': 'Yarluyandi'},
    'ysc': {'en': 'Yassic'},
    'ysd': {'en': 'Samatao'},
    'ysg': {'en': 'Sonaga'},
    'ysl': {'en': 'Yugoslavian Sign Language'},
    'ysm': {'en': 'Myanmar Sign Language'},
    'ysn': {'en': 'Sani'},
    'yso': {'en': 'Nisi (China)'},
    'ysp': {'en': 'Southern Lolopo'},
    'ysr': {'en': 'Sirenik Yupik'},
    'yss': {'en': 'Yessan-Mayo'},
    'ysy': {'en': 'Sanie'},
    'yta': {'en': 'Talu'},
    'ytl': {'en': 'Tanglang'},
    'ytp': {'en': 'Thopho'},
    'ytw': {'en': 'Yout Wam'},
    'yty': {'en': 'Yatay'},
    'yua': {'en': 'Yucateco'},
    'yub': {'en': 'Yugambal'},
    'yuc': {'en': 'Yuchi'},
    'yud': {'en': 'Judeo-Tripolitanian Arabic'},
    'yue': {'af': 'Kantonees', 'ak': 'Kantonese', 'am': 'ካንቶኒዝ', 'ar': 'الكَنْتُونية', 'as': 'কেণ্টোনীজ', 'ast': 'cantonés', 'az': 'kanton', 'az-Cyrl': 'кантон', 'az-Latn': 'kanton', 'bal-Latn': 'Kantóni', 'be': 'кантонскі дыялект кітайскай', 'be-tarask': 'кантонскі дыялект кітайскай', 'bg': 'кантонски', 'bgn': 'کانتونیونی', 'bn': 'ক্যান্টোনিজ', 'br': 'kantoneg', 'brx': 'केन्टनिज', 'bs': 'kantonski', 'bs-Cyrl': 'кантонски', 'bs-Latn': 'kantonski', 'ca': 'cantonès', 'ca-ES-valencia': 'cantonès', 'ccp': '𑄇𑄳𑄠𑄚𑄴𑄑𑄮𑄚𑄩𑄎𑄴', 'ce': 'кантонийн', 'chr': 'ᎨᎾᏙᏂᏏ', 'ckb': 'کانتۆنی', 'cs': 'kantonština', 'cy': 'Cantoneeg', 'da': 'kantonesisk', 'de': 'Kantonesisch', 'dsb': 'kantonšćina', 'ee': 'cantongbe', 'el': 'Καντονέζικα', 'el-polyton': 'Καντονέζικα', 'en': 'Cantonese', 'es': 'cantonés', 'et': 'kantoni', 'eu': 'kantonera', 'fa': 'کانتونی', 'ff-Adlm': '𞤑𞤢𞤲𞤼𞤮𞤲𞤫𞥅𞤪𞤫', 'fi': 'kantoninkiina', 'fil': 'Cantonese', 'fo': 'kantonesiskt', 'fr': 'cantonais', 'fy': 'Kantoneesk', 'ga': 'Cantainis', 'gd': 'Cantonais', 'gl': 'cantonés', 'gu': 'કેંટોનીઝ', 'ha': 'Harshen Cantonese', 'he': 'קנטונזית', 'hi': 'कैंटोनीज़', 'hi-Latn': 'Cantonese', 'hr': 'kantonski', 'hsb': 'kantonšćina', 'ht': 'cantonais', 'hu': 'kantoni', 'hy': 'կանտոներեն', 'ia': 'cantonese', 'id': 'Kanton', 'ig': 'Cantonese', 'is': 'kantónska', 'it': 'cantonese', 'ja': '広東語', 'jv': 'Kanton', 'ka': 'კანტონური', 'kea': 'kantunes', 'kgp': 'kỹtonẽj', 'kk': 'кантон тілі', 'kk-Cyrl': 'кантон тілі', 'km': 'កន្តាំង', 'kn': 'ಕ್ಯಾಂಟನೀಸ್', 'ko': '광둥어', 'kok': 'कॅण्टोनीस', 'kok-Deva': 'कॅण्टोनीस', 'ksh': 'Kanton-Schinehsesch', 'ku': 'kantonî', 'ky': 'кантончо', 'lb': 'Kantonesesch', 'lo': 'ກວາງຕຸ້ງ', 'lt': 'kinų kalbos Kantono tarmė', 'lv': 'kantoniešu', 'mai': 'केंटोनी', 'mi': 'Katonīhi', 'mk': 'кантонски', 'ml': 'കാന്റണീസ്', 'mn': 'кантон', 'mr': 'कँटोनीज', 'ms': 'Kantonis', 'mt': 'Kantoniż', 'my': 'ကွမ်းတုံ', 'nb': 'kantonesisk', 'ne': 'क्यान्टोनिज', 'nl': 'Kantonees', 'nn': 'kantonesisk', 'no': 'kantonesisk', 'nqo': 'ߞߊ߲ߕߏߣߊߞߊ߲', 'om': 'Kantonoosee', 'or': 'କାଣ୍ଟୋନିଜ୍\u200c', 'pa': 'ਕੈਂਟੋਨੀਜ਼', 'pa-Guru': 'ਕੈਂਟੋਨੀਜ਼', 'pcm': 'Kántọn Lángwej', 'pl': 'kantoński', 'ps': 'کانټوني', 'pt': 'cantonês', 'qu': 'Cantonés Simi', 'rm': 'cantonais', 'ro': 'cantoneză', 'ru': 'кантонский', 'sc': 'cantonesu', 'sd': 'ڪينٽونيز', 'sd-Arab': 'ڪينٽونيز', 'se': 'kantongiella', 'si': 'කැන්ටොනීස්', 'sk': 'kantončina', 'sl': 'kantonščina', 'smn': 'kantonkielâ', 'so': 'Kantoneese', 'sq': 'kantonezisht', 'sr': 'кантонски', 'sr-Cyrl': 'кантонски', 'sr-Latn': 'kantonski', 'sv': 'kantonesiska', 'sw': 'Kikantoni', 'ta': 'காண்டோனீஸ்', 'te': 'కాంటనీస్', 'th': 'กวางตุ้ง', 'ti': 'ካንቶንኛ', 'tk': 'kanton dili', 'to': 'lea fakakuangitongi', 'tr': 'Kantonca', 'ug': 'گۇاڭدوڭچە', 'uk': 'кантонська', 'ur': 'کینٹونیز', 'uz': 'kanton', 'uz-Cyrl': 'кантонча', 'uz-Latn': 'kanton', 'vec': 'cantoneze', 'vi': 'Tiếng Quảng Đông', 'xnr': 'कैंटोनीज़', 'yo': 'Èdè Kantonese', 'yrl': 'kãtunei', 'yue': '粵語', 'yue-Hans': '粤语', 'yue-Hant': '粵語', 'zh': '粤语', 'zh-Hans': '粤语', 'zh-Hant': '粵語', 'zu': 'isi-Cantonese'},
    'yuf': {'en': 'Havasupai-Walapai-Yavapai'},
    'yug': {'en': 'Yug'},
    'yui': {'en': 'Yurutí'},
    'yuj': {'en': 'Karkar-Yuri'},
    'yuk': {'en': 'Yuki'},
    'yul': {'en': 'Yulu'},
    'yum': {'en': 'Quechan'},
    'yun': {'en': 'Bena (Nigeria)'},
    'yup': {'en': 'Yukpa'},
    'yuq': {'en': 'Yuqui'},
    'yur': {'en': 'Yurok'},
    'yut': {'en': 'Yopno'},
    'yuu': {'en': 'Yugh'},
    'yuw': {'en': 'Yau (Morobe Province)'},
    'yux': {'en': 'Southern Yukaghir'},
    'yuy': {'en': 'East Yugur'},
    'yuz': {'en': 'Yuracare'},
    'yva': {'en': 'Yawa'},
    'yvt': {'en': 'Yavitero'},
    'ywa': {'en': 'Kalou'},
    'ywg': {'en': 'Yinhawangka'},
    'ywl': {'en': 'Western Lalu'},
    'ywn': {'en': 'Yawanawa'},
    'ywq': {'en': 'Wuding-Luquan Yi'},
    'ywr': {'en': 'Yawuru'},
    'ywt': {'en': 'Xishanba Lalo'},
    'ywu': {'en': 'Wumeng Nasu'},
    'yww': {'en': 'Yawarawarga'},
    'yxa': {'en': 'Mayawali'},
    'yxg': {'en': 'Yagara'},
    'yxl': {'en': 'Yardliyawarra'},
    'yxm': {'en': 'Yinwum'},
    'yxu': {'en': 'Yuyu'},
    'yxy': {'en': 'Yabula Yabula'},
    'yyr': {'en': 'Yir Yoront'},
    'yyu': {'en': 'Yau (Sandaun Province)'},
    'yyz': {'en': 'Ayizi'},
    'yzg': {'en': "E'ma Buyang"},
    'yzk': {'en': 'Zokhuo'},
    'za': {'af': 'Zhuang', 'ak': 'Zuang', 'am': 'ዡዋንግኛ', 'ar': 'الزهيونج', 'as': 'ঝুৱাং', 'ast': 'zhuang', 'az': 'çjuan', 'az-Latn': 'çjuan', 'bal-Latn': 'Zhuáng', 'be': 'чжуанская', 'be-tarask': 'чжуанская', 'bg': 'зуанг', 'bn': 'ঝু্য়াঙ', 'br': 'zhuang', 'brx': 'ज़ुआंग', 'bs': 'zuang', 'bs-Cyrl': 'жуанг', 'bs-Latn': 'zuang', 'ca': 'zhuang', 'ca-ES-valencia': 'zhuang', 'ccp': '𑄏𑄪𑄠𑄋𑄴', 'chr': 'ᎱᎠᏂ', 'cs': 'čuangština', 'cy': 'Zhuang', 'da': 'zhuang', 'de': 'Zhuang', 'dsb': 'zhuang', 'el': 'Ζουάνγκ', 'el-polyton': 'Ζουάνγκ', 'en': 'Zhuang', 'eo': 'ĝuanga', 'es': 'zhuang', 'et': 'tšuangi', 'eu': 'zhuangera', 'fa': 'ژوانگی', 'fi': 'zhuang', 'fil': 'Zhuang', 'fo': 'zhuang', 'fr': 'zhuang', 'fy': 'Zhuang', 'ga': 'Siuáingis', 'gd': 'Zhuang', 'gl': 'zhuang', 'gsw': 'Zhuang', 'gu': 'ઝુઆગ', 'ha': 'Zhuang', 'he': 'זואנג', 'hi': 'ज़ुआंग', 'hi-Latn': 'Zhuang', 'hr': 'zhuang', 'hsb': 'zhuang', 'ht': 'zhuang', 'hu': 'zsuang', 'hy': 'ժուանգ', 'ia': 'zhuang', 'id': 'Zhuang', 'ig': 'Zhuang', 'is': 'súang', 'it': 'zhuang', 'ja': 'チワン語', 'jv': 'Zhuang', 'ka': 'ზჰუანგი', 'kgp': 'juvỹg', 'kk': 'чжуан тілі', 'kk-Cyrl': 'чжуан тілі', 'km': 'ហ្សួង', 'kn': 'ಝೂವಾಂಗ್', 'ko': '주앙어', 'kok': 'झ्हुन्ग', 'kok-Deva': 'झ्हुन्ग', 'ks': 'زُہانگ', 'ks-Arab': 'زُہانگ', 'ku': 'zhuangî', 'ky': 'чжуанча', 'lb': 'Zhuang', 'lld': 'zhuang', 'lo': 'ຊວາງ', 'lt': 'chuang', 'lv': 'džuanu', 'mai': 'जा', 'mk': 'џуаншки', 'ml': 'സ്വാംഗ്', 'mn': 'чжуанг', 'mr': 'झुआंग', 'ms': 'Zhuang', 'mt': 'Zhuang', 'my': 'ဂျွမ်', 'nb': 'zhuang', 'ne': 'झुुआङ्ग', 'nl': 'Zhuang', 'nn': 'zhuang', 'no': 'zhuang', 'or': 'ଜୁଆଙ୍ଗ୍\u200c', 'pa': 'ਜ਼ੁਆਂਗ', 'pa-Guru': 'ਜ਼ੁਆਂਗ', 'pcm': 'Zhuáng Lángwej', 'pl': 'czuang', 'ps': 'ژوانګ', 'pt': 'zhuang', 'rm': 'zhuang', 'ro': 'zhuang', 'ru': 'чжуань', 'sc': 'zhuang', 'sd': 'جوئنگ', 'sd-Arab': 'جوئنگ', 'si': 'ෂුවාං', 'sk': 'čuangčina', 'sl': 'džuangščina', 'so': 'Zhuang', 'sq': 'zhuangisht', 'sr': 'џуаншки', 'sr-Cyrl': 'џуаншки', 'sr-Latn': 'džuanški', 'sv': 'zhuang', 'sw': 'Kizhuang', 'ta': 'ஜுவாங்', 'te': 'జువాన్', 'th': 'จ้วง', 'ti': 'ዙኣንግ', 'tk': 'çžuan dili', 'to': 'lea fakasuangi', 'tr': 'Zhuangca', 'ug': 'جۇاڭچە', 'uk': 'чжуан', 'ur': 'ژوانگی', 'uz': 'Chjuan', 'uz-Latn': 'Chjuan', 'vec': 'zuan', 'vi': 'Tiếng Choang', 'xnr': 'ज़ुआंग', 'yo': 'Ṣúwáànù', 'yrl': 'suãgi', 'yue': '壯文', 'yue-Hans': '壮文', 'yue-Hant': '壯文', 'za': 'Vahcuengh', 'zh': '壮语', 'zh-Hans': '壮语', 'zh-Hant': '壯文', 'zu': 'IsiZhuang'},
    'zaa': {'en': 'Sierra de Juárez Zapotec'},
    'zab': {'en': 'Western Tlacolula Valley Zapotec'},
    'zac': {'en': 'Ocotlán Zapotec'},
    'zad': {'en': 'Cajonos Zapotec'},
    'zae': {'en': 'Yareni Zapotec'},
    'zaf': {'en': 'Ayoquesco Zapotec'},
    'zag': {'en': 'Zaghawa'},
    'zah': {'en': 'Zangwal'},
    'zai': {'en': 'Isthmus Zapotec'},
    'zaj': {'en': 'Zaramo'},
    'zak': {'en': 'Zanaki'},
    'zal': {'en': 'Zauzou'},
    'zam': {'en': 'Miahuatlán Zapotec'},
    'zao': {'en': 'Ozolotepec Zapotec'},
    'zap': {'ar': 'الزابوتيك', 'ast': 'zapoteca', 'az': 'zapotek', 'az-Latn': 'zapotek', 'be': 'сапатэк', 'be-tarask': 'сапатэк', 'bg': 'запотек', 'bn': 'জাপোটেক', 'br': 'zapoteg', 'brx': 'ज़ापोतेक', 'bs': 'zapotečki', 'bs-Cyrl': 'запотечки', 'bs-Latn': 'zapotečki', 'ca': 'zapoteca', 'ca-ES-valencia': 'zapoteca', 'ccp': '𑄎𑄛𑄮𑄑𑄬𑄇𑄴', 'cs': 'zapotéčtina', 'cy': 'Zapoteceg', 'da': 'zapotec', 'de': 'Zapotekisch', 'el': 'Ζάποτεκ', 'el-polyton': 'Ζάποτεκ', 'en': 'Zapotec', 'es': 'zapoteco', 'et': 'sapoteegi', 'fa': 'زاپوتکی', 'fi': 'zapoteekki', 'fr': 'zapotèque', 'fy': 'Zapotec', 'gd': 'Zapotec', 'gsw': 'Zapotekisch', 'gu': 'ઝેપોટેક', 'he': 'זאפוטק', 'hi': 'ज़ेपोटेक', 'hi-Latn': 'Zapotec', 'hr': 'zapotečki', 'ht': 'zapotèque', 'hu': 'zapoték', 'hy': 'սապոտեկերեն', 'id': 'Zapotek', 'is': 'sapótek', 'it': 'zapotec', 'ja': 'サポテカ語', 'kgp': 'japoteko', 'kn': 'ಝೋಪೊಟೆಕ್', 'ko': '사포테크어', 'ks': 'زَپوتیک', 'ks-Arab': 'زَپوتیک', 'lb': 'Zapotekesch', 'lo': 'ຊາໂປແຕບ', 'lt': 'zapotekų', 'lv': 'sapoteku', 'mk': 'запотечки', 'ml': 'സാപ്പോടെക്', 'mr': 'झेपोटेक', 'mt': 'Zapotec', 'nb': 'zapotekisk', 'nl': 'Zapotec', 'nn': 'zapotec', 'no': 'zapotekisk', 'or': 'ଜାପୋଟେକ୍', 'pl': 'zapotecki', 'pt': 'zapoteco', 'rm': 'zapotec', 'ro': 'zapotecă', 'ru': 'сапотекский', 'sk': 'zapotéčtina', 'sl': 'zapoteščina', 'sr': 'запотечки', 'sr-Cyrl': 'запотечки', 'sr-Latn': 'zapotečki', 'sv': 'zapotek', 'ta': 'ஜாபோடெக்', 'te': 'జపోటెక్', 'th': 'ซาโปเตก', 'to': 'lea fakasapoteki', 'tr': 'Zapotek dili', 'ug': 'زاپوتېكچە', 'uk': 'сапотекська', 'vi': 'Tiếng Zapotec', 'yrl': 'saputeku', 'yue': '薩波特克文', 'yue-Hans': '萨波特克文', 'yue-Hant': '薩波特克文', 'zh': '萨波蒂克语', 'zh-Hans': '萨波蒂克语', 'zh-Hant': '薩波特克文'},
    'zaq': {'en': 'Aloápam Zapotec'},
    'zar': {'en': 'Rincón Zapotec'},
    'zas': {'en': 'Santo Domingo Albarradas Zapotec'},
    'zat': {'en': 'Tabaa Zapotec'},
    'zau': {'en': 'Zangskari'},
    'zav': {'en': 'Yatzachi Zapotec'},
    'zaw': {'en': 'Mitla Zapotec'},
    'zax': {'en': 'Xadani Zapotec'},
    'zay': {'en': 'Zayse-Zergulla'},
    'zaz': {'en': 'Zari'},
    'zba': {'en': 'Balaibalan'},
    'zbc': {'en': 'Central Berawan'},
    'zbe': {'en': 'East Berawan'},
    'zbl': {'am': 'ብሊስይምቦልስ', 'ar': 'رموز المعايير الأساسية', 'ast': 'simbólicu Bliss', 'az': 'blisimbols', 'az-Latn': 'blisimbols', 'bg': 'блис символи', 'bn': 'চিত্র ভাষা', 'br': 'arouezioù Bliss', 'brx': 'ब्लीस चिन्ह', 'bs': 'blis simboli', 'bs-Cyrl': 'блисимболи', 'bs-Latn': 'blis simboli', 'ca': 'símbols Bliss', 'ca-ES-valencia': 'símbols Bliss', 'ccp': '𑄃𑄉𑄬𑄠 𑄞𑄌𑄴', 'cs': 'bliss systém', 'cy': 'Blisssymbols', 'da': 'blissymboler', 'de': 'Bliss-Symbole', 'el': 'Σύμβολα Bliss', 'el-polyton': 'Σύμβολα Bliss', 'en': 'Blissymbols', 'en-Dsrt': '𐐒𐑊𐐮𐑅-𐑅𐐮𐑋𐐺𐐲𐑊𐑆', 'es': 'símbolos Bliss', 'et': 'Blissi sümbolid', 'fi': 'blisskieli', 'fr': 'symboles Bliss', 'fy': 'Blissymbolen', 'gd': 'Comharran Bliss', 'gsw': 'Bliss-Symbool', 'gu': 'બ્લિસિમ્બોલ્સ', 'he': 'בליסימבולס', 'hi': 'ब्लिसिम्बॉल्स', 'hi-Latn': 'Blissymbols', 'hr': 'Blissovi simboli', 'ht': 'symboles Bliss', 'hu': 'Bliss jelképrendszer', 'id': 'Blissymbol', 'is': 'blisstákn', 'it': 'blissymbol', 'ja': 'ブリスシンボル', 'ka': 'ბლისსიმბოლოები', 'kgp': 'sĩmoru mrij', 'kn': 'ಬ್ಲಿಸ್ಸಿಂಬಲ್ಸ್', 'ko': '블리스 심볼', 'lb': 'Bliss-Symboler', 'lo': 'ສັນຍາລັກບລີຊິມ', 'lt': 'BLISS simbolių', 'lv': 'blissimbolika', 'mk': 'блиссимболи', 'ml': 'ബ്ലിസ്സിംബൽസ്', 'mr': 'ब्लिसिम्बॉल्स', 'nb': 'blissymboler', 'ne': 'ब्लिससिम्बोल्स', 'nl': 'Blissymbolen', 'nn': 'blissymbol', 'no': 'blissymboler', 'or': 'ବ୍ଲିସିମ୍ବଲସ୍', 'pl': 'bliss', 'pt': 'símbolos blis', 'rm': 'simbols da Bliss', 'ro': 'simboluri Bilss', 'ru': 'блиссимволика', 'sk': 'systém Bliss', 'sl': 'znakovni jezik Bliss', 'sr': 'блисимболи', 'sr-Cyrl': 'блисимболи', 'sr-Latn': 'blisimboli', 'sv': 'blissymboler', 'ta': 'ப்லிஸ்ஸிம்பால்ஸ்', 'te': 'బ్లిసింబల్స్', 'th': 'บลิสซิมโบลส์', 'to': 'lea fakaʻilonga-pilisi', 'tr': 'Blis Sembolleri', 'ug': 'بىلىس بەلگىلىرى', 'uk': 'блісса мова', 'vi': 'Ký hiệu Blissymbols', 'yrl': 'rãgasaitá brisi', 'yue': '布列斯符號', 'yue-Hans': '布列斯符号', 'yue-Hant': '布列斯符號', 'zh': '布里斯符号', 'zh-Hans': '布里斯符号', 'zh-Hant': '布列斯符號'},
    'zbt': {'en': 'Batui'},
    'zbu': {'en': 'Bu (Bauchi State)'},
    'zbw': {'en': 'West Berawan'},
    'zca': {'en': 'Coatecas Altas Zapotec'},
    'zch': {'en': 'Central Hongshuihe Zhuang'},
    'zdj': {'en': 'Ngazidja Comorian'},
    'zea': {'ast': 'zeelandés', 'br': 'zelandeg', 'ca': 'zelandès', 'ca-ES-valencia': 'zelandès', 'cs': 'zélandština', 'cy': 'Zêlandeg', 'de': 'Seeländisch', 'en': 'Zeelandic', 'et': 'zeelandi', 'fi': 'seelanti', 'fr': 'zélandais', 'ga': 'Séalainnis', 'gd': 'Cànan Zeeland', 'hi-Latn': 'Zeelandic', 'ht': 'zélandais', 'hy': 'զեյլանդերեն', 'it': 'zelandese', 'ja': 'ゼーラント語', 'lb': 'Seelännesch', 'lt': 'zelandų', 'mk': 'зеландски', 'nb': 'zeeuws', 'nl': 'Zeeuws', 'nn': 'zeeuws', 'no': 'zeeuws', 'pl': 'zelandzki', 'sv': 'zeeländska', 'th': 'เซแลนด์', 'to': 'lea fakasēlani', 'tr': 'Zelandaca', 'yue': '西蘭文', 'yue-Hans': '西兰文', 'yue-Hant': '西蘭文', 'zh-Hant': '西蘭文'},
    'zeg': {'en': 'Zenag'},
    'zeh': {'en': 'Eastern Hongshuihe Zhuang'},
    'zen': {'ar': 'الزيناجا', 'ast': 'zenaga', 'az': 'zenaqa', 'az-Latn': 'zenaqa', 'bg': 'зенага', 'bn': 'জেনাগা', 'br': 'zenaga', 'brx': 'ज़ेनागा', 'bs': 'zenaga', 'bs-Cyrl': 'зенага', 'bs-Latn': 'zenaga', 'ca': 'zenaga', 'ca-ES-valencia': 'zenaga', 'ccp': '𑄎𑄬𑄚𑄉', 'cs': 'zenaga', 'da': 'zenaga', 'de': 'Zenaga', 'el': 'Ζενάγκα', 'el-polyton': 'Ζενάγκα', 'en': 'Zenaga', 'es': 'zenaga', 'et': 'zenaga', 'fa': 'زناگا', 'fi': 'zenaga', 'fr': 'zenaga', 'fy': 'Zenaga', 'gd': 'Zenaga', 'gsw': 'Zenaga', 'gu': 'ઝેનાગા', 'he': 'זנאגה', 'hi': 'ज़ेनान्गा', 'hi-Latn': 'Zenaga', 'hr': 'zenaga', 'ht': 'zenaga', 'hu': 'zenaga', 'hy': 'զենագա', 'id': 'Zenaga', 'is': 'senaga', 'it': 'zenaga', 'ja': 'ゼナガ語', 'ka': 'ზენაგა', 'kgp': 'senỹga', 'kn': 'ಝೆನಾಗಾ', 'ko': '제나가어', 'ks': 'زیناگا', 'ks-Arab': 'زیناگا', 'lb': 'Zenaga', 'lo': 'ເຊນາກາ', 'lt': 'zenaga', 'lv': 'zenagu', 'mk': 'зенага', 'ml': 'സെനഗ', 'mr': 'झेनान्गा', 'mt': 'Zenaga', 'nb': 'zenaga', 'nl': 'Zenaga', 'nn': 'zenaga', 'no': 'zenaga', 'or': 'ଜେନାଗା', 'pl': 'zenaga', 'pt': 'zenaga', 'rm': 'zenaga', 'ro': 'zenaga', 'ru': 'зенагский', 'sk': 'zenaga', 'sl': 'zenaščina', 'sr': 'зенага', 'sr-Cyrl': 'зенага', 'sr-Latn': 'zenaga', 'sv': 'zenaga', 'ta': 'ஜெனகா', 'te': 'జెనాగా', 'th': 'เซนากา', 'to': 'lea fakasenaka', 'tr': 'Zenaga dili', 'ug': 'زېناگاچە', 'uk': 'зенага', 'vi': 'Tiếng Zenaga', 'yrl': 'senaga', 'yue': '澤納加文', 'yue-Hans': '泽纳加文', 'yue-Hant': '澤納加文', 'zh': '泽纳加语', 'zh-Hans': '泽纳加语', 'zh-Hant': '澤納加文'},
    'zga': {'en': 'Kinga'},
    'zgb': {'en': 'Guibei Zhuang'},
    'zgh': {'af': 'Standaard Marokkaanse Tamazight', 'am': 'መደበኛ የሞሮኮ ታማዚግት', 'ar': 'التمازيغية المغربية القياسية', 'as': 'ষ্টেণ্ডাৰ্ড মোৰোক্কান তামাজাইট', 'ast': 'tamazight estándar de Marruecos', 'az': 'tamazi', 'az-Cyrl': 'тамази', 'az-Latn': 'tamazi', 'bal-Latn': 'Gisshetqagén Moróki Tamázi', 'be': 'стандартная мараканская тамазіхт', 'be-tarask': 'стандартная мараканская тамазіхт', 'bg': 'стандартен марокански тамазигт', 'bgn': 'آمازیغی مراکشی معیار', 'bn': 'আদর্শ মরক্কোন তামাজিগাত', 'br': 'tamacheg Maroko standart', 'brx': 'स्टेन्दार्द मरक्कान तामाजाईट', 'bs': 'standardni marokanski tamazigt', 'bs-Cyrl': 'стандардни марокански тамазихт', 'bs-Latn': 'standardni marokanski tamazigt', 'ca': 'amazic estàndard marroquí', 'ca-ES-valencia': 'amazic estàndard marroquí', 'ccp': '𑄉𑄧𑄟𑄴𑄘𑄮𑄣𑄴 𑄟𑄧𑄢𑄧𑄇𑄧𑄧𑄱𑄚𑄴𑄖𑄟𑄎𑄨𑄉𑄖𑄴', 'ce': 'мороккон стандартан тамазигхтийн', 'chr': 'ᎠᏟᎶᏍᏗ ᎼᎶᎪ ᏔᎹᏏᏘ', 'ckb': 'ئەمازیغیی مەغریب', 'cs': 'tamazight (standardní marocký)', 'cy': 'Tamaseit Moroco Safonol', 'da': 'tamazight', 'de': 'Tamazight', 'dsb': 'standardny marokkański tamazight', 'el': 'Τυπικά Ταμαζίτ Μαρόκου', 'el-polyton': 'Τυπικά Ταμαζίτ Μαρόκου', 'en': 'Standard Moroccan Tamazight', 'es': 'tamazight estándar marroquí', 'et': 'tamasikti (Maroko)', 'eu': 'amazigera estandarra', 'fa': 'آمازیغی معیار مراکش', 'ff-Adlm': '𞤚𞤢𞤥𞤢𞥁𞤭𞤼𞤪𞤫 𞤖𞤢𞤲𞤼𞤵𞤲𞥋𞤣𞤫 𞤃𞤢𞤪𞤮𞥅𞤳', 'fi': 'vakioitu tamazight', 'fil': 'Standard Moroccan Tamazight', 'fo': 'vanligt marokanskt tamazight', 'fr': 'amazighe standard marocain', 'fy': 'Standert Marokkaanske Tamazight', 'ga': 'Tamaisis Chaighdeánach Mharacó', 'gd': 'Tamaisich Stannardach Moroco', 'gl': 'tamazight marroquí estándar', 'gu': 'માનક મોરોક્કન તામાઝિટ', 'ha': 'Daidaitaccen Moroccan Tamazight', 'he': 'תמזיע׳ת מרוקאית תקנית', 'hi': 'मानक मोरक्कन तामाज़ाइट', 'hi-Latn': 'Standard Moroccan Tamazight', 'hr': 'standardni marokanski tamašek', 'hsb': 'tamazight', 'ht': 'amazighe standard marocain', 'hu': 'marokkói tamazight', 'hy': 'ընդհանուր մարոկյան թամազիղտ', 'ia': 'tamazight marocchin standard', 'id': 'Tamazight Maroko Standar', 'ig': 'Standard Moroccan Tamazight', 'is': 'staðlað marokkóskt tamazight', 'it': 'tamazight del Marocco standard', 'ja': '標準モロッコ タマジクト語', 'jv': 'Tamazight Moroko Standar', 'ka': 'სტანდარტული მაროკოული ტამაზიგხტი', 'kea': 'tamazait marokinu padron', 'kgp': 'tamỹjirte mỹhókinũ pã', 'kk': 'марокколық стандартты тамазигхт тілі', 'kk-Cyrl': 'марокколық стандартты тамазигхт тілі', 'km': 'តាម៉ាហ្សៃម៉ារ៉ុកស្តង់ដា', 'kn': 'ಸ್ಟ್ಯಾಂಡರ್ಡ್ ಮೊರೊಕ್ಕನ್ ಟಮಜೈಟ್', 'ko': '표준 모로코 타마지트어', 'kok': 'प्रमाणीत मॉरोक्कन तमाझीट', 'kok-Deva': 'प्रमाणीत मॉरोक्कन तमाझीट', 'ku': 'amazîxîya fasî', 'ky': 'марокко тамазигт адабий тилинде', 'lb': 'Marokkanescht Standard-Tamazight', 'lo': 'ໂມຣັອກແຄນ ທາມາຊີກ ມາດຕະຖານ', 'lrc': 'تامازیغ مأراکئشی', 'lt': 'standartinė Maroko tamazigtų', 'lv': 'standarta tamazigtu (Maroka)', 'mai': 'मानक मोरोक्कन टेमेजिग्थ', 'mi': 'Moroko Tamatai', 'mk': 'стандарден марокански тамазитски', 'ml': 'സ്റ്റാൻഡേർഡ് മൊറോക്കൻ റ്റാമസിയറ്റ്', 'mn': 'стандарт тамазайт (Морокко)', 'mr': 'प्रमाण मोरोक्कन तॅमॅझायट', 'ms': 'Tamazight Maghribi Standard', 'mt': 'Tamazight Standard tal-Marokk', 'my': 'မိုရိုကို တမဇိုက်', 'mzn': 'مراکش ِاستاندارد ِتامازیقتی', 'nb': 'standard marrokansk tamazight', 'ne': 'मानक मोरोक्कोन तामाजिघट', 'nl': 'Standaard Marokkaanse Tamazight', 'nn': 'standard marokkansk tamazight', 'no': 'standard marrokansk tamazight', 'nqo': 'ߡߊ߬ߙߐߞߎ߬ ߢߊߓߘߍߡߊ', 'or': 'ମାନାଙ୍କ ମରୋକିୟ ତାମାଜିଘାଟ୍', 'pa': 'ਮਿਆਰੀ ਮੋਰੋਕੇਨ ਟਾਮਾਜ਼ਿਕ', 'pa-Guru': 'ਮਿਆਰੀ ਮੋਰੋਕੇਨ ਟਾਮਾਜ਼ਿਕ', 'pcm': 'Gẹ́nárál Morókó Támazígt Lángwej', 'pl': 'standardowy marokański tamazight', 'ps': 'معياري مراکشي تمازيټ', 'pt': 'tamazirte marroqino padrão', 'qu': 'Bereber Marroquí Estándar Simi', 'rm': 'marocan tamazight standardisà', 'ro': 'tamazight standard marocană', 'ru': 'тамазигхтский', 'sc': 'tamazight istandard marochinu', 'sd': 'معياري مراڪشي تامازائيٽ', 'sd-Arab': 'معياري مراڪشي تامازائيٽ', 'si': 'සම්මත මොරොක්කෝ ටමසිග්ත්', 'sk': 'tuaregčina (marocká štandardná)', 'sl': 'standardni maroški tamazig', 'smn': 'standard tamazight', 'so': 'Morokaanka Tamasayt Rasmiga', 'sq': 'tamaziatishte standarde marokene', 'sr': 'стандардни марокански тамазигт', 'sr-Cyrl': 'стандардни марокански тамазигт', 'sr-Latn': 'standardni marokanski tamazigt', 'sv': 'marockansk standard-tamazight', 'sw': 'Kiberber Sanifu cha Moroko', 'ta': 'ஸ்டாண்டர்ட் மொராக்கன் தமாசைட்', 'te': 'ప్రామాణిక మొరొకన్ టామజైట్', 'th': 'ทามาไซต์โมร็อกโกมาตรฐาน', 'ti': 'ሞሮካዊ ምዱብ ታማዛይት', 'tk': 'standart Marokko tamazight dili', 'to': 'lea fakatamasaiti-moloko', 'tr': 'Standart Fas Tamazigti', 'ug': 'ئۆلچەملىك ماراكەش تامازىتچە', 'uk': 'стандартна марокканська берберська', 'ur': 'اسٹینڈرڈ مراقشی تمازیقی', 'uz': 'tamazigxt', 'uz-Cyrl': 'тамазигхт', 'uz-Latn': 'tamazigxt', 'vec': 'tamazight standard de’l Maroco', 'vi': 'Tiếng Tamazight Chuẩn của Ma-rốc', 'xnr': 'मानक मोरक्कन तामाज़ाइट', 'yo': 'Àfẹnùkò Támásáìtì ti Mòrókò', 'yrl': 'tamasiriti marukinu padrãu', 'yue': '標準摩洛哥塔馬塞特文', 'yue-Hans': '标准摩洛哥塔马塞特文', 'yue-Hant': '標準摩洛哥塔馬塞特文', 'zgh': 'ⵜⴰⵎⴰⵣⵉⵖⵜ', 'zh': '标准摩洛哥塔马塞特语', 'zh-Hans': '标准摩洛哥塔马塞特语', 'zh-Hant': '標準摩洛哥塔馬塞特文', 'zu': 'isi-Moroccan Tamazight esivamile'},
    'zgm': {'en': 'Minz Zhuang'},
    'zgn': {'en': 'Guibian Zhuang'},
    'zgr': {'en': 'Magori'},
    'zh': {'af': 'Chinees', 'agq': 'Chàenê', 'ak': 'Kyaena kasa', 'am': 'ቻይንኛ', 'ar': 'الصينية', 'as': 'চীনা', 'asa': 'Kichina', 'ast': 'chinu', 'az': 'çin', 'az-Arab': 'چینجه', 'az-Cyrl': 'чин', 'az-Latn': 'çin', 'bal': 'چینی', 'bal-Arab': 'چینی', 'bal-Latn': 'Chini', 'bas': 'Hɔp u kinà', 'be': 'кітайская', 'be-tarask': 'кітайская', 'bem': 'Ichi Chainisi', 'bez': 'Hichina', 'bg': 'китайски', 'bgn': 'چینایی', 'blo': 'gɩcaɩnaja manɖarɛŋ', 'bm': 'siniwakan', 'bn': 'চীনা', 'bo': 'རྒྱ་སྐད་', 'br': 'sinaeg', 'brx': 'चाइनीज', 'bs': 'kineski', 'bs-Cyrl': 'кинески', 'bs-Latn': 'kineski', 'ca': 'xinès', 'ca-ES-valencia': 'xinès', 'ccp': '𑄌𑄩𑄚', 'ce': 'цийн', 'ceb': 'Inintsik', 'cgg': 'Oruchaina', 'chr': 'ᏓᎶᏂᎨ', 'ckb': 'چینی', 'cs': 'čínština', 'cv': 'китай', 'cy': 'Tsieinëeg', 'da': 'kinesisk', 'dav': 'Kichina', 'de': 'Chinesisch', 'dje': 'Sinuwa senni', 'doi': 'चीनी', 'dsb': 'chinšćina', 'dyo': 'sinua', 'dz': 'རྒྱ་མི་ཁ', 'ebu': 'Kĩchina', 'ee': 'Chinagbe', 'el': 'Κινεζικά', 'el-polyton': 'Κινεζικά', 'en': 'Chinese', 'en-Dsrt': '𐐕𐐴𐑌𐐨𐑆', 'eo': 'ĉina', 'es': 'chino', 'et': 'hiina', 'eu': 'txinera', 'ewo': 'Ǹkɔ́bɔ tsainís', 'fa': 'چینی', 'ff': 'Sinuwaare', 'ff-Adlm': '𞤕𞤢𞤴𞤲𞤢𞤲𞤳𞤮𞥅𞤪𞤫', 'ff-Latn': 'Sinuwaare', 'fi': 'kiina', 'fil': 'Chinese', 'fo': 'kinesiskt', 'fr': 'chinois', 'fur': 'cinês', 'fy': 'Sineesk', 'ga': 'Sínis', 'gd': 'Sìnis', 'gl': 'chinés', 'gsw': 'Chineesisch', 'gu': 'ચાઇનીઝ', 'guz': 'Kichina', 'ha': 'Harshen Sinanci', 'haw': 'Pākē', 'he': 'סינית', 'hi': 'चीनी', 'hi-Latn': 'Chinese', 'hnj': '𞄋𞄄', 'hnj-Hmnp': '𞄋𞄄', 'hr': 'kineski', 'hsb': 'chinšćina', 'ht': 'chinois', 'hu': 'kínai', 'hy': 'չինարեն', 'ia': 'chinese', 'id': 'Tionghoa', 'ie': 'chinesi', 'ig': 'Chaịniiz', 'ii': 'ꍏꇩꉙ', 'is': 'kínverska', 'it': 'cinese', 'ja': '中国語', 'jgo': 'Shinwâ', 'jmc': 'Kyichina', 'jv': 'Tyonghwa', 'ka': 'ჩინური', 'kaa': 'қытайша', 'kaa-Cyrl': 'қытайша', 'kab': 'Tacinwat, Tamundarint', 'kam': 'Kichina', 'kde': 'Chichina', 'kea': 'xines', 'kgp': 'sĩnẽj', 'khq': 'Sinuwa senni, Mandareŋ', 'ki': 'Kĩcaina', 'kk': 'қытай тілі', 'kk-Cyrl': 'қытай тілі', 'kln': 'kutitab China', 'km': 'ចិន', 'kn': 'ಚೈನೀಸ್', 'ko': '중국어', 'kok': 'चिनी', 'kok-Deva': 'चिनी', 'kok-Latn': 'Chini', 'ks': 'چیٖنی ﴿ترجمع اِشارٕ: خاص طور، مینڈارن چیٖنی۔﴾', 'ks-Arab': 'چیٖنی ﴿ترجمع اِشارٕ: خاص طور، مینڈارن چیٖنی۔﴾', 'ks-Deva': 'चीनी (तरजुम इशार: खास तोर, मैन्डरिन चीनी।)', 'ksb': 'Kichina', 'ksf': 'ricinɔá', 'ksh': 'Schinehsesch (Mandarin)', 'ku': 'çînî', 'kxv': 'cainati', 'kxv-Deva': 'चीनी', 'kxv-Latn': 'cainati', 'kxv-Orya': 'ଚାଇନା ତି', 'kxv-Telu': 'చైనీస్', 'ky': 'кытайча', 'lag': 'Kɨchíina', 'lb': 'Chinesesch', 'lg': 'Lucayina', 'lkt': 'Pȟečhókaŋ Háŋska Iyápi', 'lld': 'cinesc', 'ln': 'lisinwa', 'lo': 'ຈີນ', 'lrc': 'چینی', 'lt': 'kinų', 'lu': 'shinɛ', 'luo': 'Kichina', 'luy': 'Kichina', 'lv': 'ķīniešu', 'mai': 'चीनी', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lchina', 'mer': 'Kĩchina', 'mfe': 'sinwa, mandarin', 'mg': 'Sinoa, Mandarin', 'mgh': 'Ichina', 'mi': 'Hainamana', 'mk': 'кинески', 'ml': 'ചൈനീസ്', 'mn': 'хятад', 'mni': 'চাইনিজ', 'mni-Beng': 'চাইনিজ', 'mr': 'चीनी', 'ms': 'Cina', 'mt': 'Ċiniż', 'mua': 'zah Syiŋ', 'my': 'တရုတ်', 'mzn': 'چینی', 'naq': 'Chineesǁî gowab, Mandarinni', 'nb': 'kinesisk', 'nd': 'isi-China', 'ne': 'चिनियाँ', 'nl': 'Chinees', 'nmg': 'Kiɛl bó chinois', 'nn': 'kinesisk', 'no': 'kinesisk', 'nqo': 'ߛߣߌߥߊߞߊ߲', 'nus': 'Thok cayna', 'nyn': 'Oruchaina', 'om': 'Chinese', 'or': 'ଚାଇନିଜ୍\u200c', 'os': 'китайаг', 'pa': 'ਚੀਨੀ', 'pa-Guru': 'ਚੀਨੀ', 'pcm': 'Mandarín Chainíz Lángwej', 'pl': 'chiński', 'ps': 'چیني', 'pt': 'chinês', 'qu': 'Chino Simi', 'rm': 'chinais', 'rn': 'Igishinwa', 'ro': 'chineză', 'rof': 'Kichina', 'ru': 'китайский', 'rwk': 'Kyichina', 'sa': 'चीनी', 'sah': 'Кытайдыы', 'saq': 'Kichina', 'sat': 'ᱪᱟᱭᱱᱤᱡᱽ', 'sat-Olck': 'ᱪᱟᱭᱱᱤᱡᱽ', 'sbp': 'Ishishina', 'sc': 'tzinesu', 'sd': 'چيني', 'sd-Arab': 'چيني', 'sd-Deva': 'चीनी (तर्जुमे जो द॒स :खास करे, मैन्डरिन चीनी)', 'se': 'kiinnágiella', 'seh': 'chinês', 'ses': 'Sinuwa senni, Mandareŋ', 'sg': 'Shinuäa', 'shi': 'ⵜⴰⵛⵉⵏⵡⵉⵜ', 'shi-Latn': 'Tacinwit', 'shi-Tfng': 'ⵜⴰⵛⵉⵏⵡⵉⵜ', 'si': 'චීන', 'sk': 'čínština', 'sl': 'kitajščina', 'smn': 'mandarinkiinakielâ', 'sn': 'chiChinese', 'so': 'Shinees', 'sq': 'kinezisht', 'sr': 'кинески', 'sr-Cyrl': 'кинески', 'sr-Latn': 'kineski', 'su': 'Tiongkok', 'su-Latn': 'Tiongkok', 'sv': 'kinesiska', 'sw': 'Kichina', 'syr': 'ܨܝܢܝܬ', 'ta': 'சீனம்', 'te': 'చైనీస్', 'teo': 'Kichina', 'tg': 'хитоӣ', 'th': 'จีน', 'ti': 'ቻይንኛ', 'tk': 'hytaý dili', 'to': 'lea fakasiaina', 'tpi': 'Sainis', 'tr': 'Çince', 'tt': 'кытай', 'twq': 'Sinuwa senni, Mandareŋ', 'tzm': 'Tacinwit,Mandarin', 'ug': 'خەنزۇچە', 'uk': 'китайська', 'ur': 'چینی', 'uz': 'xitoy', 'uz-Cyrl': 'хитойча', 'uz-Latn': 'xitoy', 'vai': 'ꕦꕇꔧ', 'vai-Latn': 'Chaniĩ', 'vai-Vaii': 'ꕦꕇꔧ', 'vec': 'sineze', 'vi': 'Tiếng Trung', 'vun': 'Kyichina', 'wae': 'Chinesiš', 'wo': 'Sinuwaa', 'xh': 'IsiMandarin', 'xnr': 'चीनी', 'xog': 'Olucayina', 'yav': 'sinúɛ', 'yi': 'כינעזיש', 'yo': 'Edè Ṣáínà', 'yrl': 'xinanheẽga', 'yue': '中文', 'yue-Hans': '中文', 'yue-Hant': '中文', 'zgh': 'ⵜⴰⵛⵉⵏⵡⵉⵜ', 'zh': '中文', 'zh-Hans': '中文', 'zh-Hant': '中文', 'zu': 'isi-Chinese'},
    'zh-Hans': {'af': 'Mandarynse Chinees (Vereenvoudig)', 'ak': 'Kyaena kasa a emu yɛ mmrɛ', 'am': 'ቀለል ያለ ቻይንኛ', 'ar': 'الصينية المبسطة', 'as': 'সৰলীকৃত চীনা', 'ast': 'chinu simplificáu', 'az': 'sadələşmiş çin', 'az-Arab': 'ساده\u200cلنمیش چینجه', 'az-Cyrl': 'садәләшмиш чин', 'az-Latn': 'sadələşmiş çin', 'bal': 'ساده کرتگێن چینی', 'bal-Arab': 'ساده کرتگێن چینی', 'be': 'кітайская (спрошчаныя іерогліфы)', 'be-tarask': 'кітайская (спрошчаныя іерогліфы)', 'bg': 'китайски (опростен)', 'bgn': 'ساده\u200cگین چینایی', 'blo': 'gɩcaɩnaja gɩburoka', 'bn': 'চীনা ম্যান্ডারিন সরলীকৃত', 'br': 'sinaeg eeunaet', 'brx': 'गोरलै खालामनाय चाइनीज', 'bs': 'kineski (pojednostavljeni)', 'bs-Cyrl': 'кинески (поједностављен)', 'bs-Latn': 'kineski (pojednostavljeni)', 'ca': 'xinès simplificat', 'ca-ES-valencia': 'xinès simplificat', 'ccp': '𑄅𑄪𑄎𑄪𑄅𑄪𑄏𑄫 𑄌𑄩𑄚', 'ce': 'атта цийн', 'ceb': 'Pinasimple nga Chinese', 'chr': 'ᎠᎯᏗᎨ ᏓᎶᏂᎨ', 'ckb': 'چینی (چینیی ئاسانکراو)', 'cs': 'čínština (zjednodušená)', 'cv': 'китай, ҫӑмӑллатнӑ ҫыру', 'cy': 'Tsieinëeg Symledig', 'da': 'forenklet kinesisk', 'de': 'Chinesisch (vereinfacht)', 'doi': 'सरलीकृत चीनी', 'dsb': 'chinšćina (zjadnorjona)', 'dz': 'རྒྱ་མི་ཁ་འཇམ་སངམ', 'ee': 'tsainagbe', 'el': 'Απλοποιημένα Κινεζικά', 'el-polyton': 'Απλοποιημένα Κινεζικά', 'en': 'Simplified Chinese', 'eo': 'ĉina simpligita', 'es': 'chino simplificado', 'et': 'lihtsustatud hiina', 'eu': 'txinera sinplifikatua', 'fa': 'چینی ساده\u200cشده', 'ff-Adlm': '𞤕𞤢𞤴𞤲𞤢𞤲𞤳𞤮𞥅𞤪𞤫 𞤖𞤮𞤴𞤬𞤭𞤲𞤢𞥄𞤲𞤣𞤫', 'fi': 'mandariinikiina (yksinkertaistettu)', 'fil': 'Pinasimpleng Chinese', 'fo': 'einkult kinesiskt', 'fr': 'chinois simplifié', 'fur': 'cinês semplificât', 'fy': 'Ferienfâldich Sineesk', 'ga': 'Sínis Shimplithe', 'gd': 'Sìnis Shimplichte', 'gl': 'chinés simplificado', 'gsw': 'Veräifachts Chineesisch', 'gu': 'સરળીકૃત ચાઇનીઝ', 'ha': 'Sauƙaƙaƙƙen Sinanci', 'haw': 'Pākē Hoʻomaʻalahi ʻia', 'he': 'סינית פשוטה', 'hi': 'सरलीकृत चीनी', 'hi-Latn': 'Simplified Chinese', 'hr': 'kineski (pojednostavljeni)', 'hsb': 'chinšćina (zjednorjena)', 'ht': 'chinois simplifié', 'hu': 'egyszerűsített kínai', 'hy': 'պարզեցված չինարեն', 'ia': 'chinese simplificate', 'id': 'Tionghoa Mandarin (Sederhana)', 'ig': 'Asụsụ Chaịniiz dị mfe', 'is': 'kínverska (einfölduð)', 'it': 'cinese semplificato', 'ja': '簡体中国語', 'jv': 'Tyonghwa (Ringkes)', 'ka': 'გამარტივებული ჩინური', 'kea': 'xines simplifikadu', 'kgp': 'sĩnẽj mẽ', 'kk': 'жеңілдетілген қытай тілі', 'kk-Cyrl': 'жеңілдетілген қытай тілі', 'km': 'ចិន\u200bអក្សរ\u200bកាត់', 'kn': 'ಸರಳೀಕೃತ ಚೈನೀಸ್', 'ko': '중국어(만다린, 간체)', 'kok': 'सोंपी चिनी', 'kok-Deva': 'सोंपी चिनी', 'ks': 'سَہل چیٖنی', 'ks-Arab': 'سَہل چیٖنی', 'ks-Deva': 'आसान चीनी', 'ksh': 'Schinehsesch (eijfache Schreff)', 'ku': 'çînîya sadekirî', 'kxv': 'sahaj cainati', 'kxv-Deva': 'साॅहाॅजाॅ चीनी', 'kxv-Latn': 'sahaj cainati', 'kxv-Orya': 'ସହଜ ଚୀନ', 'kxv-Telu': 'సరళీకృత చైనీస్', 'ky': 'кытайча (жөнөкөйлөштүрүлгөн)', 'lb': 'Chinesesch (vereinfacht)', 'lkt': 'Pȟečhókaŋ Háŋska Iyápi Ikčéka', 'lld': 'cinesc scemplifiché', 'lo': 'ຈີນແບບຮຽບງ່າຍ', 'lrc': 'چینی سادە بیە', 'lt': 'supaprastintoji kinų', 'lv': 'ķīniešu vienkāršotā', 'mai': 'सरलीकृत चीनी', 'mi': 'Hainamana Māmā', 'mk': 'поедноставен кинески', 'ml': 'ലളിതമാക്കിയ ചൈനീസ്', 'mn': 'хялбаршуулсан хятад', 'mni': 'সিমপ্লিফাইদ চাইনিজ', 'mni-Beng': 'সিমপ্লিফাইদ চাইনিজ', 'mr': 'सरलीकृत चीनी', 'ms': 'Cina Ringkas', 'mt': 'Ċiniż Simplifikat', 'my': 'ပြည်ကြီးတရုတ် (ရိုးရှင်း)', 'mzn': 'ساده چینی', 'nb': 'forenklet kinesisk', 'ne': 'सरलिकृत चिनियाँ', 'nl': 'Mandarijn (vereenvoudigd)', 'nn': 'forenkla kinesisk', 'no': 'forenklet kinesisk', 'nqo': 'ߛߣߌߥߊߞߊ߲ ߘߐߞߣߍߣߍ߲', 'om': 'Chinese Salphifame', 'or': 'ସରଳୀକୃତ ଚାଇନିଜ', 'os': 'ӕнцонгонд китайаг', 'pa': 'ਚੀਨੀ ਮੈਂਡਰਿਨ (ਸਰਲ)', 'pa-Guru': 'ਚੀਨੀ ਮੈਂਡਰਿਨ (ਸਰਲ)', 'pl': 'chiński uproszczony', 'ps': 'چيني ماندرين چيني', 'pt': 'chinês simplificado', 'qu': 'Chino Simplificado Simi', 'rm': 'chinais simplifitgà', 'ro': 'chineză simplificată', 'ru': 'китайский, упрощенное письмо', 'sa': 'सरलीकृत चीनी', 'sat': 'ᱟᱞᱜᱟᱣᱟᱠᱟᱱ ᱪᱟᱭᱱᱤᱡᱽ', 'sat-Olck': 'ᱟᱞᱜᱟᱣᱟᱠᱟᱱ ᱪᱟᱭᱱᱤᱡᱽ', 'sc': 'tzinesu semplificadu', 'sd': 'سادي مندارن چيني', 'sd-Arab': 'سادي مندارن چيني', 'sd-Deva': 'सादी थियल चीनी', 'se': 'álki kiinágiella', 'si': 'සරල චීන', 'sk': 'čínština (zjednodušená)', 'sl': 'poenostavljena kitajščina', 'smn': 'oovtâkiärdánis kiinakielâ', 'so': 'Shiinaha Rasmiga ah', 'sq': 'kinezishte e thjeshtuar', 'sr': 'поједностављени кинески', 'sr-Cyrl': 'поједностављени кинески', 'sr-Latn': 'pojednostavljeni kineski', 'su': 'Tiongkok Sederhana', 'su-Latn': 'Tiongkok Sederhana', 'sv': 'förenklad kinesiska', 'sw': 'Kichina (Kilichorahisishwa)', 'syr': 'ܨܝܢܝܬ (ܦܫܝܛܐ)', 'ta': 'எளிதாக்கப்பட்ட சீனம்', 'te': 'సరళీకృత చైనీస్', 'tg': 'хитоии осонфаҳм', 'th': 'จีนตัวย่อ', 'ti': 'ቀሊል ቻይንኛ', 'tk': 'ýönekeýleşdirilen hytaý dili', 'to': 'lea fakasiaina-fakafaingofua', 'tpi': 'Isipela Sainis', 'tr': 'Basitleştirilmiş Çince', 'tt': 'гадиләштерелгән кытай', 'ug': 'ئاددىي خەنچە', 'uk': 'китайська (спрощене письмо)', 'ur': 'چینی (آسان کردہ)', 'uz': 'xitoy (soddalashgan)', 'uz-Cyrl': 'соддалаштирилган хитойча', 'uz-Latn': 'xitoy (soddalashgan)', 'vec': 'mandarin (senpio)', 'vi': 'Tiếng Trung Phổ thông (Giản thể)', 'wae': 'Vereifačts Chinesiš', 'wo': 'Sinuwaa buñ woyofal', 'xh': 'IsiTshayina Esenziwe Lula', 'xnr': 'सादी चीनी', 'yo': 'Ẹdè Ṣáínà Onírọ̀rùn', 'yrl': 'xinanheẽga iwasuĩma', 'yue': '簡體中文', 'yue-Hans': '简体中文', 'yue-Hant': '簡體中文', 'zh': '简体中文', 'zh-Hans': '简体中文', 'zh-Hant': '簡體中文', 'zu': 'isi-Chinese (esenziwe-lula)'},
    'zh-Hant': {'af': 'Mandarynse Chinees (Tradisioneel)', 'ak': 'Tete Kyaena kasa', 'am': 'ባህላዊ ቻይንኛ', 'ar': 'الصينية التقليدية', 'as': 'পৰম্পৰাগত চীনা', 'ast': 'chinu tradicional', 'az': 'ənənəvi çin', 'az-Arab': 'سنتی چینجه', 'az-Cyrl': 'әнәнәви чин', 'az-Latn': 'ənənəvi çin', 'be': 'кітайская (традыцыйныя іерогліфы)', 'be-tarask': 'кітайская (традыцыйныя іерогліфы)', 'bg': 'китайски (традиционен)', 'bgn': 'غدیمین چینایی', 'blo': 'gɩcaɩnaja tututu', 'bn': 'ম্যান্ডারিন চীনা (ঐতিহ্যবাহী)', 'br': 'sinaeg hengounel', 'brx': 'आसार-खान्थियारि चाइनीज', 'bs': 'kineski (tradicionalni)', 'bs-Cyrl': 'кинески мандарински (традиционални)', 'bs-Latn': 'kineski (tradicionalni)', 'ca': 'xinès tradicional', 'ca-ES-valencia': 'xinès tradicional', 'ccp': '𑄢𑄨𑄘𑄨𑄥𑄪𑄘𑄮𑄟𑄴 𑄌𑄩𑄚', 'ce': 'ламастан цийн', 'ceb': 'Tradisyonal nga Chinese', 'chr': 'ᎤᏦᏍᏗ ᏓᎶᏂᎨ', 'ckb': 'چینی (چینیی دێرین)', 'cs': 'standardní čínština (tradiční)', 'cv': 'китай, традициллӗ ҫыру', 'cy': 'Tsieinëeg Traddodiadol', 'da': 'traditionelt kinesisk', 'de': 'Chinesisch (traditionell)', 'doi': 'रवायती चीनी', 'dsb': 'chinšćina (tradicionalna)', 'dz': 'སྔ་དུས་ཀྱི་རྒྱ་མི་ཁ', 'ee': 'blema tsainagbe', 'el': 'Παραδοσιακά Κινεζικά', 'el-polyton': 'Παραδοσιακά Κινεζικά', 'en': 'Traditional Chinese', 'eo': 'ĉina tradicia', 'es': 'chino tradicional', 'et': 'traditsiooniline hiina', 'eu': 'txinera tradizionala', 'fa': 'چینی سنتی', 'ff-Adlm': '𞤕𞤢𞤴𞤲𞤢𞤲𞤳𞤮𞥅𞤪𞤫 𞤒𞤫𞤷𞥆𞤵𞤲𞥋𞤣𞤫', 'fi': 'mandariinikiina (perinteinen)', 'fil': 'Tradisyonal na Chinese', 'fo': 'vanligt kinesiskt', 'fr': 'chinois traditionnel', 'fur': 'cinês tradizionâl', 'fy': 'Tradisjoneel Sineesk', 'ga': 'Sínis Thraidisiúnta', 'gd': 'Sìnis Thradaiseanta', 'gl': 'chinés tradicional', 'gsw': 'Tradizionells Chineesisch', 'gu': 'પારંપરિક ચાઇનીઝ', 'ha': 'Sinanci na gargajiya', 'haw': 'Pākē Kuʻuna', 'he': 'סינית מסורתית', 'hi': 'पारंपरिक चीनी', 'hi-Latn': 'Traditional Chinese', 'hr': 'kineski (tradicionalni)', 'hsb': 'chinšćina (tradicionalna)', 'ht': 'chinois traditionnel', 'hu': 'hagyományos kínai', 'hy': 'ավանդական չինարեն', 'ia': 'chinese traditional', 'id': 'Tionghoa Mandarin (Tradisional)', 'ig': 'Asụsụ ọdịnala Chaịniiz', 'is': 'kínverska (hefðbundin)', 'it': 'cinese tradizionale', 'ja': '繁体中国語', 'jv': 'Tyonghwa (Tradisional)', 'ka': 'ტრადიციული ჩინური', 'kea': 'xines tradisional', 'kgp': 'sĩnẽj pẽ', 'kk': 'дәстүрлі қытай тілі', 'kk-Cyrl': 'дәстүрлі қытай тілі', 'km': 'ចិន\u200bអក្សរ\u200bពេញ', 'kn': 'ಸಾಂಪ್ರದಾಯಿಕ ಚೈನೀಸ್', 'ko': '중국어(만다린, 번체)', 'kok': 'पारंपारीक चिनी', 'kok-Deva': 'पारंपारीक चिनी', 'ks': 'رِوٲجی چیٖنی', 'ks-Arab': 'رِوٲجی چیٖنی', 'ks-Deva': 'रिवायाती चीनी', 'ksh': 'Schinehsesch (tradizjonälle Schreff)', 'ku': 'çînîya kevneşopî', 'kxv': 'hirudlu cainati', 'kxv-Deva': 'हिरूदोल्लु चीनी', 'kxv-Latn': 'hirudlu cainati', 'kxv-Orya': 'ହିରୁଦଲୁ ଚାଇନିଜ୍\u200c', 'kxv-Telu': 'సాంప్రదాయక చైనీస్', 'ky': 'кытайча (салттуу)', 'lb': 'Chinesesch (traditionell)', 'lkt': 'Pȟečhókaŋ Háŋska Iyápi Ȟče', 'lld': 'cinesc tradizional', 'lo': 'ຈີນແບບດັ້ງເດີມ', 'lrc': 'چینی سونأتی', 'lt': 'tradicinė kinų', 'lv': 'ķīniešu tradicionālā', 'mai': 'पारंपरिक चीनी', 'mi': 'Hainamana Tukuiho', 'mk': 'традиционален кинески', 'ml': 'പരമ്പരാഗത ചൈനീസ്', 'mn': 'уламжлалт хятад', 'mni': 'ত্রেদিস্নেল চাইনিজ', 'mni-Beng': 'ত্রেদিস্নেল চাইনিজ', 'mr': 'पारंपारिक चीनी', 'ms': 'Cina Tradisional', 'mt': 'Ċiniż Tradizzjonali', 'my': 'ပြည်ကြီးတရုတ် (ရိုးရာ)', 'mzn': 'سنتی چینی', 'nb': 'tradisjonell kinesisk', 'ne': 'परम्परागत चिनियाँ', 'nl': 'Mandarijn (traditioneel)', 'nn': 'tradisjonell kinesisk', 'no': 'tradisjonell kinesisk', 'nqo': 'ߛߣߌߥߊߞߊ߲ ߦߋ߲ߢߐ߲߯ߠߊ', 'om': 'Chinese Durii', 'or': 'ପାରମ୍ପରିକ ଚାଇନିଜ୍\u200c', 'os': 'традицион китайаг', 'pa': 'ਚੀਨੀ ਮੈਂਡਰਿਨ (ਰਵਾਇਤੀ)', 'pa-Guru': 'ਚੀਨੀ ਮੈਂਡਰਿਨ (ਰਵਾਇਤੀ)', 'pcm': 'Tradíshọ́nál Mandarín Chainíz Lángwej', 'pl': 'chiński tradycyjny', 'ps': 'دوديزه چيني', 'pt': 'chinês tradicional', 'qu': 'Chino Tradicional Simi', 'rm': 'chinais tradiziunal', 'ro': 'chineză tradițională', 'ru': 'китайский, традиционное письмо', 'sa': 'परम्परागत चीनी', 'sat': 'ᱴᱨᱮᱰᱤᱥᱱᱟᱞ ᱪᱟᱭᱱᱤᱡᱽ', 'sat-Olck': 'ᱴᱨᱮᱰᱤᱥᱱᱟᱞ ᱪᱟᱭᱱᱤᱡᱽ', 'sc': 'tzinesu traditzionale', 'sd': 'روايتي چيني', 'sd-Arab': 'روايتي چيني', 'sd-Deva': 'रवायती चीनी', 'se': 'árbevirolaš kiinnágiella', 'si': 'සාම්ප්\u200dරදායික චීන', 'sk': 'čínština (tradičná)', 'sl': 'tradicionalna kitajščina', 'smn': 'ärbivuáválâš kiinakielâ', 'so': 'Af-Shiineeska Qadiimiga ah', 'sq': 'kinezishte tradicionale', 'sr': 'традиционални кинески', 'sr-Cyrl': 'традиционални кинески', 'sr-Latn': 'tradicionalni kineski', 'su': 'Tiongkok Tradisional', 'su-Latn': 'Tiongkok Tradisional', 'sv': 'traditionell kinesiska', 'sw': 'Kichina cha Jadi', 'syr': 'ܨܝܢܐܝܬ', 'ta': 'பாரம்பரிய சீனம்', 'te': 'సాంప్రదాయక చైనీస్', 'tg': 'хитоии анъанавӣ', 'th': 'จีนตัวเต็ม', 'ti': 'ባህላዊ ቻይንኛ', 'tk': 'adaty hytaý dili', 'to': 'lea fakasiaina-tukufakaholo', 'tpi': 'Tredisinol Sainis', 'tr': 'Geleneksel Çince', 'tt': 'традицион кытай', 'ug': 'مۇرەككەپ خەنچە', 'uk': 'китайська (традиційне письмо)', 'ur': 'روایتی چینی', 'uz': 'xitoy (an’anaviy)', 'uz-Cyrl': 'анъанавий хитойча', 'uz-Latn': 'xitoy (an’anaviy)', 'vec': 'mandarin (tradisionale)', 'vi': 'Tiếng Trung Phổ thông (Phồn thể)', 'wae': 'Traditionells Chinesiš', 'wo': 'Sinuwaa bu cosaan', 'xh': 'IsiTshayina Esiqhelekileyo', 'xnr': 'पारंपरिक चीनी', 'yo': 'Èdè Ṣáínà Ìbílẹ̀', 'yrl': 'xinanheẽga katuwa', 'yue': '繁體中文', 'yue-Hans': '繁体中文', 'yue-Hant': '繁體中文', 'zh': '繁体中文', 'zh-Hans': '繁体中文', 'zh-Hant': '繁體中文', 'zu': 'Isi-Chinese Sasendulo'},
    'zhb': {'en': 'Zhaba'},
    'zhd': {'en': 'Dai Zhuang'},
    'zhi': {'en': 'Zhire'},
    'zhn': {'en': 'Nong Zhuang'},
    'zhw': {'en': 'Zhoa'},
    'zhx': {'en': 'Chinese (family)'},
    'zhx-min-pro': {'en': 'Proto-Min'},
    'zia': {'en': 'Zia'},
    'zib': {'en': 'Zimbabwe Sign Language'},
    'zik': {'en': 'Zimakani'},
    'zil': {'en': 'Zialo'},
    'zim': {'en': 'Mesme'},
    'zin': {'en': 'Zinza'},
    'zir': {'en': 'Ziriya'},
    'ziw': {'en': 'Zigula'},
    'ziz': {'en': 'Zizilivakan'},
    'zka': {'en': 'Kaimbulawa'},
    'zkb': {'en': 'Koibal'},
    'zkd': {'en': 'Kadu'},
    'zkg': {'en': 'Koguryo'},
    'zkh': {'en': 'Khorezmian'},
    'zkk': {'en': 'Karankawa'},
    'zkn': {'en': 'Kanan'},
    'zko': {'en': 'Kott'},
    'zkp': {'en': 'São Paulo Kaingáng'},
    'zkr': {'en': 'Zakhring'},
    'zkt': {'en': 'Kitan'},
    'zku': {'en': 'Kaurna'},
    'zkv': {'en': 'Krevinian'},
    'zkz': {'en': 'Khazar'},
    'zla': {'en': 'Zula'},
    'zle': {'en': 'East Slavic languages'},
    'zle-ono': {'en': 'Old Novgorodian'},
    'zlj': {'en': 'Liujiang Zhuang'},
    'zlm': {'en': 'Malay (individual language)'},
    'zln': {'en': 'Lianshan Zhuang'},
    'zlq': {'en': 'Liuqian Zhuang'},
    'zls': {'en': 'South Slavic languages'},
    'zlw': {'en': 'West Slavic languages'},
    'zlw-ocs': {'en': 'Old Czech'},
    'zlw-opl': {'en': 'Old Polish'},
    'zlw-pom': {'en': 'Pomeranian'},
    'zlw-slv': {'en': 'Slovincian'},
    'zma': {'en': 'Manda (Australia)'},
    'zmb': {'en': 'Zimba'},
    'zmc': {'en': 'Margany'},
    'zmd': {'en': 'Maridan'},
    'zme': {'en': 'Mangerr'},
    'zmf': {'en': 'Mfinu'},
    'zmg': {'en': 'Marti Ke'},
    'zmh': {'en': 'Makolkol'},
    'zmi': {'en': 'Negeri Sembilan Malay'},
    'zmj': {'en': 'Maridjabin'},
    'zmk': {'en': 'Mandandanyi'},
    'zml': {'en': 'Matngala'},
    'zmm': {'en': 'Marimanindji'},
    'zmn': {'en': 'Mbangwe'},
    'zmo': {'en': 'Molo'},
    'zmp': {'en': 'Mpuono'},
    'zmq': {'en': 'Mituku'},
    'zmr': {'en': 'Maranunggu'},
    'zms': {'en': 'Mbesa'},
    'zmt': {'en': 'Maringarr'},
    'zmu': {'en': 'Muruwari'},
    'zmv': {'en': 'Mbariman-Gudhinma'},
    'zmw': {'en': 'Mbo (Democratic Republic of Congo)'},
    'zmx': {'en': 'Bomitaba'},
    'zmy': {'en': 'Mariyedi'},
    'zmz': {'en': 'Mbandja'},
    'zna': {'en': 'Zan Gula'},
    'znd': {'en': 'Zande languages'},
    'zne': {'en': 'Zande (individual language)'},
    'zng': {'en': 'Mang'},
    'znk': {'en': 'Manangkari'},
    'zns': {'en': 'Mangas'},
    'zoc': {'en': 'Copainalá Zoque'},
    'zoh': {'en': 'Chimalapa Zoque'},
    'zom': {'en': 'Zou'},
    'zoo': {'en': 'Asunción Mixtepec Zapotec'},
    'zoq': {'en': 'Tabasco Zoque'},
    'zor': {'en': 'Rayón Zoque'},
    'zos': {'en': 'Francisco León Zoque'},
    'zpa': {'en': 'Lachiguiri Zapotec'},
    'zpb': {'en': 'Yautepec Zapotec'},
    'zpc': {'en': 'Choapan Zapotec'},
    'zpd': {'en': 'Southeastern Ixtlán Zapotec'},
    'zpe': {'en': 'Petapa Zapotec'},
    'zpf': {'en': 'San Pedro Quiatoni Zapotec'},
    'zpg': {'en': 'Guevea De Humboldt Zapotec'},
    'zph': {'en': 'Totomachapan Zapotec'},
    'zpi': {'en': 'Santa María Quiegolani Zapotec'},
    'zpj': {'en': 'Quiavicuzas Zapotec'},
    'zpk': {'en': 'Tlacolulita Zapotec'},
    'zpl': {'en': 'Lachixío Zapotec'},
    'zpm': {'en': 'Mixtepec Zapotec'},
    'zpn': {'en': 'Santa Inés Yatzechi Zapotec'},
    'zpo': {'en': 'Amatlán Zapotec'},
    'zpp': {'en': 'El Alto Zapotec'},
    'zpq': {'en': 'Zoogocho Zapotec'},
    'zpr': {'en': 'Santiago Xanica Zapotec'},
    'zps': {'en': 'Coatlán Zapotec'},
    'zpt': {'en': 'San Vicente Coatlán Zapotec'},
    'zpu': {'en': 'Yalálag Zapotec'},
    'zpv': {'en': 'Chichicapan Zapotec'},
    'zpw': {'en': 'Zaniza Zapotec'},
    'zpx': {'en': 'San Baltazar Loxicha Zapotec'},
    'zpy': {'en': 'Mazaltepec Zapotec'},
    'zpz': {'en': 'Texmelucan Zapotec'},
    'zqe': {'en': 'Qiubei Zhuang'},
    'zra': {'en': 'Kara (Korea)'},
    'zrg': {'en': 'Mirgan'},
    'zrn': {'en': 'Zerenkel'},
    'zro': {'en': 'Záparo'},
    'zrp': {'en': 'Zarphatic'},
    'zrs': {'en': 'Mairasi'},
    'zsa': {'en': 'Sarasira'},
    'zsk': {'en': 'Kaskean'},
    'zsl': {'en': 'Zambian Sign Language'},
    'zsm': {'zsm': 'bahasa Malaysia', 'en': 'Standard Malay'},
    'zsr': {'en': 'Southern Rincon Zapotec'},
    'zsu': {'en': 'Sukurum'},
    'zte': {'en': 'Elotepec Zapotec'},
    'ztg': {'en': 'Xanaguía Zapotec'},
    'ztl': {'en': 'Lapaguía-Guivini Zapotec'},
    'ztm': {'en': 'San Agustín Mixtepec Zapotec'},
    'ztn': {'en': 'Santa Catarina Albarradas Zapotec'},
    'ztp': {'en': 'Loxicha Zapotec'},
    'ztq': {'en': 'Quioquitani-Quierí Zapotec'},
    'zts': {'en': 'Tilquiapan Zapotec'},
    'ztt': {'en': 'Tejalapan Zapotec'},
    'ztu': {'en': 'Güilá Zapotec'},
    'ztx': {'en': 'Zaachila Zapotec'},
    'zty': {'en': 'Yatee Zapotec'},
    'zu': {'af': 'Zoeloe', 'agq': 'Zulù', 'ak': 'Zulu', 'am': 'ዙሉኛ', 'ar': 'الزولو', 'as': 'ঝুলু', 'asa': 'Kidhulu', 'ast': 'zulú', 'az': 'zulu', 'az-Arab': 'زۇلوجا', 'az-Cyrl': 'зулу', 'az-Latn': 'zulu', 'bal-Latn': 'Zulu', 'bas': 'Hɔp u zulù', 'be': 'зулу', 'be-tarask': 'зулу', 'bem': 'Ichi Zulu', 'bez': 'Hizulu', 'bg': 'зулуски', 'bgn': 'زولویی', 'bm': 'zulukan', 'bn': 'জুলু', 'br': 'zouloueg', 'brx': 'जुलू', 'bs': 'zulu', 'bs-Cyrl': 'зулу', 'bs-Latn': 'zulu', 'ca': 'zulu', 'ca-ES-valencia': 'zulu', 'ccp': '𑄎𑄪𑄣𑄪', 'ce': 'зулу', 'cgg': 'Oruzuru', 'chr': 'ᏑᎷ', 'ckb': 'زوولوو', 'cs': 'zuluština', 'cy': 'Swlw', 'da': 'zulu', 'dav': 'Kizulu', 'de': 'Zulu', 'dje': 'Zulu senni', 'dsb': 'zulu', 'dyo': 'sulu', 'dz': 'ཟུ་ལུ་ཁ', 'ebu': 'Kĩzulu', 'ee': 'zulugbe', 'el': 'Ζουλού', 'el-polyton': 'Ζουλού', 'en': 'Zulu', 'eo': 'zulua', 'es': 'zulú', 'et': 'suulu', 'eu': 'zuluera', 'ewo': 'ǹkɔ́bɔ zulú', 'fa': 'زولویی', 'ff': 'Suluŋkoore', 'ff-Adlm': '𞥁𞤵𞤤𞤵𞥅𞤪𞤫', 'ff-Latn': 'Suluŋkoore', 'fi': 'zulu', 'fil': 'Zulu', 'fo': 'sulu', 'fr': 'zoulou', 'fur': 'zulu', 'fy': 'Zulu', 'ga': 'Súlúis', 'gd': 'Zulu', 'gl': 'zulú', 'gsw': 'Zulu', 'gu': 'ઝુલુ', 'guz': 'Kizulu', 'ha': 'Harshen Zulu', 'he': 'זולו', 'hi': 'ज़ुलू', 'hi-Latn': 'Zulu', 'hr': 'zulu', 'hsb': 'zulušćina', 'ht': 'zoulou', 'hu': 'zulu', 'hy': 'զուլուերեն', 'ia': 'zulu', 'id': 'Zulu', 'ig': 'Zulu', 'is': 'súlú', 'it': 'zulu', 'ja': 'ズールー語', 'jmc': 'Kyizulu', 'jv': 'Zulu', 'ka': 'ზულუ', 'kab': 'Tazulut', 'kam': 'Kizulu', 'kde': 'Chizulu', 'kea': 'zulu', 'kgp': 'suru', 'khq': 'Julu senni', 'ki': 'Kizulu', 'kk': 'зулу тілі', 'kk-Cyrl': 'зулу тілі', 'kln': 'kutitab Zulu', 'km': 'ហ្សូលូ', 'kn': 'ಜುಲು', 'ko': '줄루어', 'kok': 'झुलू', 'kok-Deva': 'झुलू', 'ks': 'زُلوٗ', 'ks-Arab': 'زُلوٗ', 'ksb': 'Kizulu', 'ksf': 'rizúlu', 'ksh': 'Suhlu', 'ku': 'zuluyî', 'kxv': 'julu', 'kxv-Deva': 'ज़ुलू', 'kxv-Latn': 'julu', 'kxv-Orya': 'ଜୁଲୁ', 'kxv-Telu': 'జాలూ', 'ky': 'зулуча', 'lag': 'Kɨzúulu', 'lb': 'Zulu', 'lg': 'Luzzulu', 'lkt': 'Zulu Iyápi', 'lld': 'zulu', 'ln': 'zulu', 'lo': 'ຊູລູ', 'lrc': 'زولو', 'lt': 'zulų', 'lu': 'Nzulu', 'luo': 'Kizulu', 'luy': 'Kizulu', 'lv': 'zulu', 'mai': 'जुलू', 'mas': 'nkʉtʉ́k ɔ́ɔ̄ lzulu', 'mer': 'Kĩzulu', 'mfe': 'zoulou', 'mg': 'Zolò', 'mgh': 'Izulu', 'mi': 'Tūru', 'mk': 'зулу', 'ml': 'സുലു', 'mn': 'зулу', 'mr': 'झुलू', 'ms': 'Zulu', 'mt': 'Zulu', 'mua': 'Zulu', 'my': 'ဇူးလူး', 'mzn': 'زولو', 'naq': 'Zulub', 'nb': 'zulu', 'nd': 'isi-Zulu', 'ne': 'जुलु', 'nl': 'Zoeloe', 'nmg': 'Zulu', 'nn': 'zulu', 'no': 'zulu', 'nqo': 'ߖ߭ߟߎߞߊ߲', 'nus': 'Thok dhuluni', 'nyn': 'Oruzuru', 'om': 'Afaan Zuulu', 'or': 'ଜୁଲୁ', 'pa': 'ਜ਼ੁਲੂ', 'pa-Guru': 'ਜ਼ੁਲੂ', 'pcm': 'Zúlu', 'pl': 'zulu', 'ps': 'زولو', 'pt': 'zulu', 'qu': 'Isizulu Simi', 'rm': 'zulu', 'rn': 'Ikizulu', 'ro': 'zulu', 'rof': 'Kizulu', 'ru': 'зулу', 'rw': 'Inyezulu', 'rwk': 'Kyizulu', 'sah': 'Зуулулуу', 'saq': 'Kizulu', 'sbp': 'Ishisulu', 'sc': 'zulu', 'sd': 'زولو', 'sd-Arab': 'زولو', 'seh': 'zulu', 'ses': 'Zulu senni', 'sg': 'Zûlu', 'shi': 'ⵜⴰⵣⵓⵍⵓⵜ', 'shi-Latn': 'Tazulut', 'shi-Tfng': 'ⵜⴰⵣⵓⵍⵓⵜ', 'si': 'සුලු', 'sk': 'zuluština', 'sl': 'zulujščina', 'smn': 'zulukielâ', 'sn': 'chiZulu', 'so': 'Zuulu', 'sq': 'zuluisht', 'sr': 'зулу', 'sr-Cyrl': 'зулу', 'sr-Latn': 'zulu', 'sv': 'zulu', 'sw': 'Kizulu', 'ta': 'ஜுலு', 'te': 'జూలూ', 'teo': 'Kizulu', 'th': 'ซูลู', 'ti': 'ዙሉ', 'tk': 'zulu dili', 'to': 'lea fakasulu', 'tr': 'Zuluca', 'twq': 'Zulu senni', 'tzm': 'tazulut', 'ug': 'زۇلۇچە', 'uk': 'зулуська', 'ur': 'زولو', 'uz': 'zulu', 'uz-Cyrl': 'зулу', 'uz-Latn': 'zulu', 'vai': 'ꖮꖨ', 'vai-Latn': 'Zúlu', 'vai-Vaii': 'ꖮꖨ', 'vec': 'zulu', 'vi': 'Tiếng Zulu', 'vun': 'Kyizulu', 'wae': 'Zulu', 'xh': 'isiZulu', 'xnr': 'ज़ुलू', 'xog': 'Oluzzulu', 'yav': 'nusulú', 'yi': 'זולו', 'yo': 'Èdè Ṣulu', 'yrl': 'suru', 'yue': '祖魯文', 'yue-Hans': '祖鲁文', 'yue-Hant': '祖魯文', 'zgh': 'ⵜⴰⵣⵓⵍⵓⵜ', 'zh': '祖鲁语', 'zh-Hans': '祖鲁语', 'zh-Hant': '祖魯文', 'zu': 'isiZulu'},
    'zua': {'en': 'Zeem'},
    'zuh': {'en': 'Tokano'},
    'zum': {'en': 'Kumzari'},
    'zun': {'af': 'Zuni', 'am': 'ዙኒ', 'ar': 'الزونية', 'as': 'ঝুনি', 'ast': 'zuni', 'az': 'zuni', 'az-Cyrl': 'зуни', 'az-Latn': 'zuni', 'be': 'зуні', 'be-tarask': 'зуні', 'bg': 'зуни', 'bgn': 'زونی', 'bn': 'জুনি', 'br': 'zuni', 'brx': 'जुनी', 'bs': 'zuni', 'bs-Cyrl': 'зуни', 'bs-Latn': 'zuni', 'ca': 'zuni', 'ca-ES-valencia': 'zuni', 'ccp': '𑄎𑄪𑄚𑄨', 'ce': 'зуньи', 'chr': 'ᏑᏂ', 'ckb': 'زوونی', 'cs': 'zunijština', 'cy': 'Swni', 'da': 'zuni', 'de': 'Zuni', 'dsb': 'zunišćina', 'el': 'Ζούνι', 'el-polyton': 'Ζούνι', 'en': 'Zuni', 'es': 'zuñi', 'et': 'sunji', 'eu': 'zuñiera', 'fa': 'زونیایی', 'ff-Adlm': '𞤟𞤵𞤲𞤭𞥅𞤪𞤫', 'fi': 'zuni', 'fil': 'Zuni', 'fo': 'zuni', 'fr': 'zuñi', 'fy': 'Zuni', 'ga': 'Zúinis', 'gd': 'Zuñi', 'gl': 'zuni', 'gsw': 'Zuni-Schpraach', 'gu': 'ઝૂની', 'ha': 'Zuni', 'he': 'זוני', 'hi': 'ज़ूनी', 'hi-Latn': 'Zuni', 'hr': 'zuni', 'hsb': 'zunišćina', 'ht': 'zuñi', 'hu': 'zuni', 'hy': 'զունիերեն', 'ia': 'zuni', 'id': 'Zuni', 'ig': 'Zuni', 'is': 'súní', 'it': 'zuni', 'ja': 'ズニ語', 'jv': 'Zuni', 'ka': 'ზუნი', 'kgp': 'sunhi', 'kk': 'зуни тілі', 'kk-Cyrl': 'зуни тілі', 'km': 'ហ្សូនី', 'kn': 'ಝೂನಿ', 'ko': '주니어', 'kok': 'झुनी', 'kok-Deva': 'झुनी', 'ks': 'زوٗنی', 'ks-Arab': 'زوٗنی', 'ksh': 'Suhñi', 'ku': 'zunîyî', 'ky': 'зуниче', 'lb': 'Zuni-Sprooch', 'lo': 'ຊູນີ', 'lt': 'Zuni', 'lv': 'zunju', 'mai': 'जूनी', 'mi': 'Tuni', 'mk': 'зуни', 'ml': 'സുനി', 'mn': 'зуни', 'mr': 'झुनी', 'ms': 'Zuni', 'mt': 'Zuni', 'my': 'ဇူနီ', 'nb': 'zuni', 'ne': 'जुनी', 'nl': 'Zuni', 'nn': 'zuni', 'no': 'zuni', 'or': 'ଜୁନୀ', 'pa': 'ਜ਼ੂਨੀ', 'pa-Guru': 'ਜ਼ੂਨੀ', 'pcm': 'Zúní Lángwej', 'pl': 'zuni', 'ps': 'زوني', 'pt': 'zunhi', 'qu': 'Zuni Simi', 'rm': 'zuni', 'ro': 'zuni', 'ru': 'зуньи', 'sc': 'zuni', 'sd': 'زوني', 'sd-Arab': 'زوني', 'si': 'සුනි', 'sk': 'zuniština', 'sl': 'zunijščina', 'smn': 'zunikielâ', 'so': 'Zuni', 'sq': 'zunisht', 'sr': 'зуни', 'sr-Cyrl': 'зуни', 'sr-Latn': 'zuni', 'sv': 'zuni', 'sw': 'Kizuni', 'ta': 'ஜூனி', 'te': 'జుని', 'th': 'ซูนิ', 'ti': 'ዙኚ', 'tk': 'zuni dili', 'to': 'lea fakasuni', 'tr': 'Zunice', 'ug': 'زۇنىچە', 'uk': 'зуньї', 'ur': 'زونی', 'uz': 'zuni', 'uz-Latn': 'zuni', 'vec': 'zuni', 'vi': 'Tiếng Zuni', 'xnr': 'ज़ूनी', 'yo': 'Èdè Suni', 'yrl': 'sũyi', 'yue': '祖尼文', 'yue-Hans': '祖尼文', 'yue-Hant': '祖尼文', 'zh': '祖尼语', 'zh-Hans': '祖尼语', 'zh-Hant': '祖尼文', 'zu': 'isi-Zuni'},
    'zuy': {'en': 'Zumaya'},
    'zwa': {'en': 'Zay'},
    'zxx': {'af': 'Geen taalinhoud nie', 'ak': 'Lengwestese biara nnim', 'am': 'ቋንቋዊ ይዘት የለውም', 'ar': 'بدون محتوى لغوي', 'as': 'কোনো ভাষা সমল নাই', 'ast': 'ensin conteníu llingüísticu', 'az': 'dil məzmunu yoxdur', 'az-Cyrl': 'дил мәзмуну јохдур', 'az-Latn': 'dil məzmunu yoxdur', 'bal-Latn': 'Hecch zobán', 'be': 'няма моўнага матэрыялу', 'be-tarask': 'няма моўнага матэрыялу', 'bg': 'без лингвистично съдържание', 'bgn': 'بغیر شه زبانین لڑا', 'bn': 'ভাষাভিত্তিক বিষয়বস্তু নেই', 'br': 'diyezh', 'brx': 'रावआरि आयदाफोर जेबो गैया', 'bs': 'bez lingvističkog sadržaja', 'bs-Cyrl': 'без лингвистичког садржаја', 'bs-Latn': 'bez lingvističkog sadržaja', 'ca': 'sense contingut lingüístic', 'ca-ES-valencia': 'sense contingut lingüístic', 'ccp': '𑄞𑄏𑄧𑄢𑄴𑄘𑄮𑄇𑄳𑄠𑄬 𑄝𑄨𑄥𑄧𑄠𑄴 𑄚𑄳𑄄𑄬', 'ce': 'меттан чулацам боцуш', 'chr': 'Ꮭ ᎦᏬᏂᎯᏍᏗ ᎦᎸᏛᎢ ᏱᎩ', 'ckb': 'هیچ ناوەرۆکی زمانی نیە', 'cs': 'žádný jazykový obsah', 'cy': 'Dim cynnwys ieithyddol', 'da': 'intet sprogligt indhold', 'de': 'Keine Sprachinhalte', 'dsb': 'žedno rěcne wopśimjeśe', 'dz': 'སྐད་རིག་ནང་དོན་མེདཔ', 'ee': 'gbegbɔgblɔ manɔmee', 'el': 'Χωρίς γλωσσολογικό περιεχόμενο', 'el-polyton': 'Χωρίς γλωσσολογικό περιεχόμενο', 'en': 'No linguistic content', 'en-Dsrt': '𐐤𐐬 𐑊𐐨𐑍𐐶𐐮𐑅𐐻𐐮𐐿 𐐿𐐱𐑌𐐻𐐯𐑌𐐻', 'eo': 'nelingvaĵo', 'es': 'sin contenido lingüístico', 'et': 'mittekeeleline', 'eu': 'ez dago eduki linguistikorik', 'fa': 'بدون محتوای زبانی', 'ff-Adlm': '𞤀𞤤𞤢𞥄 𞤦𞤢𞤯𞤮𞤪𞤢𞤤 𞤯𞤫𞤲𞤯𞤢𞤲𞤳𞤮', 'fi': 'ei kielellistä sisältöä', 'fil': 'Walang nilalaman na ukol sa wika', 'fo': 'einki málsligt innihald', 'fr': 'sans contenu linguistique', 'fy': 'Gjin linguïstyske ynhâld', 'ga': 'Gan ábhar teangeolaíoch', 'gd': 'Susbaint nach eil ’na chànan', 'gl': 'sen contido lingüístico', 'gsw': 'Kän schpraachliche Inhalt', 'gu': 'કોઇ ભાષાશાસ્ત્રીય સામગ્રી નથી', 'ha': 'Babu abun cikin yare', 'he': 'ללא תוכן לשוני', 'hi': 'कोई भाषा सामग्री नहीं', 'hi-Latn': 'No linguistic content', 'hr': 'bez jezičnog sadržaja', 'hsb': 'žadyn rěčny wobsah', 'ht': 'sans contenu linguistique', 'hu': 'nincs nyelvészeti tartalom', 'hy': 'առանց լեզվային բովանդակության', 'ia': 'sin contento linguistic', 'id': 'Tidak ada konten linguistik', 'ig': 'Enweghị asụsụ dịnaya', 'is': 'ekkert tungumálaefni', 'it': 'nessun contenuto linguistico', 'ja': '言語的内容なし', 'jv': 'Konten tanpa linguistik', 'ka': 'ლინგვისტური შიგთავსი არ არის', 'kea': 'sen kontiudu linguístiku', 'kgp': 'nén ũ vẽnhvĩ ki tũ', 'kk': 'тілдік мазмұны жоқ', 'kk-Cyrl': 'тілдік мазмұны жоқ', 'km': 'គ្មាន\u200bទិន្នន័យ\u200bភាសា', 'kn': 'ಯಾವುದೇ ಭಾಷಾಸಂಬಂಧಿ ವಿಷಯವಿಲ್ಲ', 'ko': '언어 관련 내용 없음', 'kok': 'अणकार सामुग्री ना', 'kok-Deva': 'अणकार सामुग्री ना', 'ks': 'کانہہ تہِ لِسانیاتی مواد نہٕ', 'ks-Arab': 'کانہہ تہِ لِسانیاتی مواد نہٕ', 'ksh': 'kein Schprooch', 'ku': 'bê naveroka zimanî', 'ky': 'тилдик мазмун жок', 'lb': 'Keng Sproochinhalter', 'lo': 'ບໍ່ມີເນື້ອຫາພາສາ', 'lrc': 'بی نئشوٙ', 'lt': 'nėra kalbinio turinio', 'lv': 'bez lingvistiska satura', 'mai': 'कुनु भाषाई सामग्री नहि', 'mi': 'Wetereo kiko kore', 'mk': 'без лингвистичка содржина', 'ml': 'ഭാഷാപരമായ ഉള്ളടക്കമൊന്നുമില്ല', 'mn': 'хэл зүйн агуулгагүй', 'mr': 'भाषावैज्ञानिक सामग्री नाही', 'ms': 'Tiada kandungan linguistik', 'mt': 'Bla kontenut lingwistiku', 'my': 'ဘာသာစကားနှင့် ပတ်သက်သောအရာ မရှိပါ', 'mzn': 'این زوون بشناسی\u200cیه نیّه', 'nb': 'uten språklig innhold', 'ne': 'भाषिक सामग्री छैन', 'nl': 'geen linguïstische inhoud', 'nn': 'utan språkleg innhald', 'no': 'uten språklig innhold', 'nqo': 'ߞߊ߲߫ ߘߐߞߏߟߏ߲', 'or': 'କୌଣସି ଲିଙ୍ଗୁଇଷ୍ଟ ସାମଗ୍ରୀ ନାହିଁ', 'pa': 'ਬੋਲੀ ਸੰਬੰਧੀ ਕੋਈ ਸਮੱਗਰੀ ਨਹੀਂ', 'pa-Guru': 'ਬੋਲੀ ਸੰਬੰਧੀ ਕੋਈ ਸਮੱਗਰੀ ਨਹੀਂ', 'pcm': 'Nó Lángwéj Kọ́ntẹnt', 'pl': 'brak treści o charakterze językowym', 'ps': 'نه ژبني منځپانګه', 'pt': 'sem conteúdo linguístico', 'qu': 'Manaraq simi yachana', 'rm': 'nagins cuntegns linguistics', 'ro': 'fară conținut lingvistic', 'ru': 'нет языкового материала', 'sc': 'perunu cuntenutu linguìsticu', 'sd': 'ڪوئي ٻولي جو مواد ڪونهي', 'sd-Arab': 'ڪوئي ٻولي جو مواد ڪونهي', 'si': 'වාග් විද්\u200dයා අන්තර්ගතයක් නැත', 'sk': 'bez jazykového obsahu', 'sl': 'brez jezikoslovne vsebine', 'smn': 'ij kielâlâš siskáldâs', 'so': 'Luuqad Looma Hayo', 'sq': 'nuk ka përmbajtje gjuhësore', 'sr': 'без лингвистичког садржаја', 'sr-Cyrl': 'без лингвистичког садржаја', 'sr-Latn': 'bez lingvističkog sadržaja', 'sv': 'inget språkligt innehåll', 'sw': 'Hakuna maudhui ya lugha', 'ta': 'மொழி உள்ளடக்கம் ஏதுமில்லை', 'te': 'లిపి లేదు', 'th': 'ไม่มีข้อมูลภาษา', 'ti': 'ቋንቋዊ ትሕዝቶ የለን', 'tk': 'dilçilige degişli mazmun ýok', 'to': 'ʻikai ha lea', 'tr': 'Dilbilim içeriği yok', 'ug': 'تىل مەزمۇنى يوق', 'uk': 'без мовного вмісту', 'ur': 'کوئی لسانی مواد نہیں', 'uz': 'til tarkibi yo‘q', 'uz-Cyrl': 'Тил таркиби йўқ', 'uz-Latn': 'til tarkibi yo‘q', 'vec': 'gnaun contegnùo lenguìstego', 'vi': 'Không có nội dung ngôn ngữ', 'xnr': 'कोई बोली सामग्री नी', 'yo': 'Kò sí àkóònú elédè', 'yrl': 'ũba aykué nheẽga sesewaraitá', 'yue': '無語言內容', 'yue-Hans': '无语言内容', 'yue-Hant': '無語言內容', 'zh': '无语言内容', 'zh-Hans': '无语言内容', 'zh-Hant': '無語言內容', 'zu': 'akukho okuqukethwe kolimi'},
    'zyb': {'en': 'Yongbei Zhuang'},
    'zyg': {'en': 'Yang Zhuang'},
    'zyj': {'en': 'Youjiang Zhuang'},
    'zyn': {'en': 'Yongnan Zhuang'},
    'zyp': {'en': 'Zyphe Chin'},
    'zza': {'af': 'Zaza', 'am': 'ዛዛ', 'ar': 'زازا', 'as': 'ঝাঝা', 'ast': 'zaza', 'az': 'zaza', 'az-Cyrl': 'заза', 'az-Latn': 'zaza', 'be': 'зазакі', 'be-tarask': 'зазакі', 'bg': 'заза', 'bgn': 'زازایی', 'bn': 'জাজা', 'bo': 'ཟ་ཟའ་སྐད།', 'br': 'zazakeg', 'brx': 'जाजा', 'bs': 'zaza', 'bs-Cyrl': 'заза', 'bs-Latn': 'zaza', 'ca': 'zaza', 'ca-ES-valencia': 'zaza', 'ccp': '𑄎𑄎', 'ce': 'заза', 'chr': 'ᏌᏌ', 'ckb': 'زازا', 'cs': 'zaza', 'cy': 'Sasäeg', 'da': 'zaza', 'de': 'Zaza', 'dsb': 'zazašćina', 'el': 'Ζάζα', 'el-polyton': 'Ζάζα', 'en': 'Zaza', 'es': 'zazaki', 'et': 'zaza', 'eu': 'zazera', 'fa': 'زازایی', 'ff-Adlm': '𞤟𞤢𞥁𞤢𞥄𞤪𞤫', 'fi': 'zaza', 'fil': 'Zaza', 'fo': 'zaza', 'fr': 'zazaki', 'fy': 'Zaza', 'ga': 'Zázá', 'gd': 'Zazaki', 'gl': 'zazaki', 'gsw': 'Zaza', 'gu': 'ઝાઝા', 'ha': 'Zaza', 'he': 'זאזא', 'hi': 'ज़ाज़ा', 'hi-Latn': 'Zaza', 'hr': 'zazaki', 'hsb': 'zazašćina', 'ht': 'zazaki', 'hu': 'zaza', 'hy': 'զազաերեն', 'ia': 'zaza', 'id': 'Zaza', 'ig': 'Zaza', 'is': 'zázáíska', 'it': 'zaza', 'ja': 'ザザ語', 'jv': 'Zaza', 'ka': 'ზაზაკი', 'kgp': 'sasaki', 'kk': 'заза тілі', 'kk-Cyrl': 'заза тілі', 'km': 'ហ្សាហ្សា', 'kn': 'ಜಾಝಾ', 'ko': '자자어', 'kok': 'झाझा', 'kok-Deva': 'झाझा', 'ks': 'زازا', 'ks-Arab': 'زازا', 'ksh': 'Sahsajesch', 'ku': 'zazakî (kirdkî, kirmanckî)', 'ky': 'зазача', 'lb': 'Zaza', 'lkt': 'Zaza Iyápi', 'lo': 'ຊາຊາ', 'lt': 'zaza', 'lv': 'zazaki', 'mai': 'जाजा', 'mi': 'Tātā', 'mk': 'заза', 'ml': 'സാസാ', 'mn': 'заза', 'mr': 'झाझा', 'ms': 'Zaza', 'mt': 'Zaza', 'my': 'ဇာဇာ', 'nb': 'zazaisk', 'ne': 'जाजा', 'nl': 'Zaza', 'nn': 'zazaisk', 'no': 'zazaisk', 'or': 'ଜାଜା', 'pa': 'ਜ਼ਾਜ਼ਾ', 'pa-Guru': 'ਜ਼ਾਜ਼ਾ', 'pcm': 'Zázá Lángwej', 'pl': 'zazaki', 'ps': 'زازا', 'pt': 'zazaki', 'qu': 'Zaza Simi', 'rm': 'zaza', 'ro': 'zaza', 'ru': 'заза', 'sc': 'zazaki', 'sd': 'زازا', 'sd-Arab': 'زازا', 'si': 'සාසා', 'sk': 'zaza', 'sl': 'zazajščina', 'smn': 'zazakielâ', 'so': 'Zaza', 'sq': 'zazaisht', 'sr': 'заза', 'sr-Cyrl': 'заза', 'sr-Latn': 'zaza', 'sv': 'zazaiska', 'sw': 'Kizaza', 'ta': 'ஜாஜா', 'te': 'జాజా', 'th': 'ซาซา', 'ti': 'ዛዛኪ', 'tk': 'zazaki dili', 'to': 'lea fakasāsā', 'tr': 'Zazaca', 'ug': 'زازاچە', 'uk': 'зазакі', 'ur': 'زازا', 'uz': 'zaza', 'uz-Latn': 'zaza', 'vec': 'zazaki', 'vi': 'Tiếng Zaza', 'xnr': 'ज़ाज़ा', 'yo': 'Èdè Sasa', 'yrl': 'sasaki', 'yue': '扎扎文', 'yue-Hans': '扎扎文', 'yue-Hant': '扎扎文', 'zh': '扎扎语', 'zh-Hans': '扎扎语', 'zh-Hant': '扎扎文', 'zu': 'isi-Zaza'},
    'zzj': {'en': 'Zuojiang Zhuang'},
}
